/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 6049
#define LARGE_STATE_COUNT 3811
#define SYMBOL_COUNT 482
#define ALIAS_COUNT 10
#define TOKEN_COUNT 295
#define EXTERNAL_TOKEN_COUNT 43
#define FIELD_COUNT 43
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 152
#define SUPERTYPE_COUNT 4

enum ts_symbol_identifiers {
  sym__identifier = 1,
  anon_sym_LBRACE = 2,
  anon_sym_RBRACE = 3,
  anon_sym_COLON = 4,
  anon_sym_SEMI = 5,
  anon_sym_package = 6,
  anon_sym_class = 7,
  anon_sym_role = 8,
  anon_sym_BUILD = 9,
  anon_sym_ADJUST = 10,
  anon_sym_DOLLAR = 11,
  anon_sym_EQ = 12,
  anon_sym_PIPE_PIPE_EQ = 13,
  anon_sym_SLASH_SLASH_EQ = 14,
  anon_sym_AT = 15,
  anon_sym_COMMA = 16,
  anon_sym_RPAREN = 17,
  anon_sym_my = 18,
  anon_sym_extended = 19,
  anon_sym_async = 20,
  anon_sym_sub = 21,
  anon_sym_method = 22,
  anon_sym_LPAREN = 23,
  anon_sym_continue = 24,
  anon_sym_state = 25,
  anon_sym_our = 26,
  anon_sym_try = 27,
  anon_sym_catch = 28,
  anon_sym_finally = 29,
  anon_sym_defer = 30,
  anon_sym_DOT_DOT_DOT = 31,
  anon_sym_else = 32,
  anon_sym_elsif = 33,
  anon_sym_and = 34,
  anon_sym_or = 35,
  anon_sym_xor = 36,
  anon_sym_DASH_GT = 37,
  anon_sym_STAR = 38,
  anon_sym_LBRACK = 39,
  anon_sym_RBRACK = 40,
  anon_sym_PERCENT = 41,
  anon_sym_GT = 42,
  anon_sym_LT_LT = 43,
  anon_sym_GT_GT = 44,
  anon_sym_STAR_STAR_EQ = 45,
  anon_sym_PLUS_EQ = 46,
  anon_sym_DASH_EQ = 47,
  anon_sym_DOT_EQ = 48,
  anon_sym_STAR_EQ = 49,
  anon_sym_SLASH_EQ = 50,
  anon_sym_PERCENT_EQ = 51,
  anon_sym_x_EQ = 52,
  anon_sym_AMP_EQ = 53,
  anon_sym_PIPE_EQ = 54,
  anon_sym_CARET_EQ = 55,
  anon_sym_LT_LT_EQ = 56,
  anon_sym_GT_GT_EQ = 57,
  anon_sym_AMP_AMP_EQ = 58,
  anon_sym_DOT_DOT = 59,
  anon_sym_STAR_STAR = 60,
  anon_sym_PIPE_PIPE = 61,
  anon_sym_SLASH_SLASH = 62,
  anon_sym_CARET_CARET = 63,
  anon_sym_AMP_AMP = 64,
  anon_sym_PIPE = 65,
  anon_sym_CARET = 66,
  anon_sym_AMP = 67,
  anon_sym_GT_GT2 = 68,
  anon_sym_PLUS = 69,
  anon_sym_DASH = 70,
  anon_sym_DOT = 71,
  anon_sym_SLASH = 72,
  anon_sym_x = 73,
  anon_sym_EQ_TILDE = 74,
  anon_sym_BANG_TILDE = 75,
  anon_sym_EQ_EQ = 76,
  anon_sym_BANG_EQ = 77,
  anon_sym_eq = 78,
  anon_sym_EQ_EQ_EQ = 79,
  anon_sym_equ = 80,
  anon_sym_eqr = 81,
  anon_sym_ne = 82,
  anon_sym_LT_EQ_GT = 83,
  anon_sym_cmp = 84,
  anon_sym_TILDE_TILDE = 85,
  anon_sym_LT = 86,
  anon_sym_LT_EQ = 87,
  anon_sym_GT_EQ = 88,
  anon_sym_lt = 89,
  anon_sym_le = 90,
  anon_sym_ge = 91,
  anon_sym_gt = 92,
  anon_sym_isa = 93,
  anon_sym_TILDE = 94,
  anon_sym_BANG = 95,
  anon_sym_PLUS_PLUS = 96,
  anon_sym_DASH_DASH = 97,
  anon_sym_QMARK = 98,
  anon_sym_BSLASH = 99,
  anon_sym_do = 100,
  anon_sym_eval = 101,
  anon_sym_await = 102,
  anon_sym_field = 103,
  anon_sym_local = 104,
  anon_sym_dynamically = 105,
  anon_sym_DOLLAR_POUND = 106,
  anon_sym_require = 107,
  anon_sym_map = 108,
  anon_sym_grep = 109,
  anon_sym_sort = 110,
  anon_sym_goto = 111,
  anon_sym_return = 112,
  anon_sym_undef = 113,
  anon_sym_PERCENT2 = 114,
  anon_sym_AMP2 = 115,
  anon_sym_STAR2 = 116,
  aux_sym__var_indirob_autoquote_token1 = 117,
  anon_sym_EQ_GT = 118,
  anon_sym_use = 119,
  anon_sym_no = 120,
  anon_sym_for = 121,
  anon_sym_foreach = 122,
  anon_sym_last = 123,
  anon_sym_next = 124,
  anon_sym_redo = 125,
  anon_sym_BEGIN = 126,
  anon_sym_INIT = 127,
  anon_sym_CHECK = 128,
  anon_sym_UNITCHECK = 129,
  anon_sym_END = 130,
  anon_sym___FILE__ = 131,
  anon_sym___LINE__ = 132,
  anon_sym___PACKAGE__ = 133,
  anon_sym___SUB__ = 134,
  anon_sym_break = 135,
  anon_sym_fork = 136,
  anon_sym_getppid = 137,
  anon_sym_time = 138,
  anon_sym_times = 139,
  anon_sym_wait = 140,
  anon_sym_wantarray = 141,
  anon_sym_abs = 142,
  anon_sym_alarm = 143,
  anon_sym_chop = 144,
  anon_sym_chdir = 145,
  anon_sym_close = 146,
  anon_sym_closedir = 147,
  anon_sym_caller = 148,
  anon_sym_chomp = 149,
  anon_sym_chr = 150,
  anon_sym_cos = 151,
  anon_sym_chroot = 152,
  anon_sym_defined = 153,
  anon_sym_delete = 154,
  anon_sym_dbmclose = 155,
  anon_sym_exists = 156,
  anon_sym_exit = 157,
  anon_sym_eof = 158,
  anon_sym_exp = 159,
  anon_sym_each = 160,
  anon_sym_fc = 161,
  anon_sym_fileno = 162,
  anon_sym_gmtime = 163,
  anon_sym_getc = 164,
  anon_sym_getpgrp = 165,
  anon_sym_getprotobyname = 166,
  anon_sym_getpwname = 167,
  anon_sym_getpwuid = 168,
  anon_sym_getpeername = 169,
  anon_sym_getnetbyname = 170,
  anon_sym_getsockname = 171,
  anon_sym_getgrnam = 172,
  anon_sym_getgrgid = 173,
  anon_sym_hex = 174,
  anon_sym_int = 175,
  anon_sym_keys = 176,
  anon_sym_lc = 177,
  anon_sym_lcfirst = 178,
  anon_sym_length = 179,
  anon_sym_localtime = 180,
  anon_sym_log = 181,
  anon_sym_lock = 182,
  anon_sym_lstat = 183,
  anon_sym_oct = 184,
  anon_sym_ord = 185,
  anon_sym_prototype = 186,
  anon_sym_pop = 187,
  anon_sym_pos = 188,
  anon_sym_quotemeta = 189,
  anon_sym_reset = 190,
  anon_sym_rand = 191,
  anon_sym_rmdir = 192,
  anon_sym_readdir = 193,
  anon_sym_readline = 194,
  anon_sym_readpipe = 195,
  anon_sym_rewinddir = 196,
  anon_sym_readlink = 197,
  anon_sym_ref = 198,
  anon_sym_scalar = 199,
  anon_sym_shift = 200,
  anon_sym_sin = 201,
  anon_sym_sleep = 202,
  anon_sym_sqrt = 203,
  anon_sym_srand = 204,
  anon_sym_stat = 205,
  anon_sym_study = 206,
  anon_sym_tell = 207,
  anon_sym_telldir = 208,
  anon_sym_tied = 209,
  anon_sym_uc = 210,
  anon_sym_ucfirst = 211,
  anon_sym_untie = 212,
  anon_sym_umask = 213,
  anon_sym_values = 214,
  anon_sym_write = 215,
  sym_comment = 216,
  anon_sym___DATA__ = 217,
  anon_sym___END__ = 218,
  aux_sym___DATA___token1 = 219,
  anon_sym_EOT = 220,
  anon_sym_q = 221,
  anon_sym_qq = 222,
  anon_sym_DASH_GT2 = 223,
  anon_sym_LBRACK2 = 224,
  anon_sym_LBRACE2 = 225,
  anon_sym_DOLLAR_STAR = 226,
  anon_sym_AT_STAR = 227,
  aux_sym__interpolation_fallbacks_token1 = 228,
  aux_sym__interpolation_fallbacks_token2 = 229,
  anon_sym_qw = 230,
  anon_sym_qx = 231,
  anon_sym_qr = 232,
  anon_sym_m = 233,
  anon_sym_s = 234,
  sym_quoted_regexp_modifiers = 235,
  sym_match_regexp_modifiers = 236,
  sym_substitution_regexp_modifiers = 237,
  sym_transliteration_modifiers = 238,
  aux_sym__interpolated_transliteration_content_token1 = 239,
  anon_sym_tr = 240,
  anon_sym_y = 241,
  sym_version = 242,
  anon_sym_if = 243,
  anon_sym_unless = 244,
  anon_sym_while = 245,
  anon_sym_until = 246,
  aux_sym__ident_special_token1 = 247,
  aux_sym__bareword_token1 = 248,
  sym_number = 249,
  anon_sym_true = 250,
  anon_sym_false = 251,
  sym__single_quote = 252,
  sym__double_quote = 253,
  sym__backtick_quote = 254,
  sym__search_slash_quote = 255,
  sym__no_search_slash_plz = 256,
  sym__open_readline_bracket = 257,
  sym__open_fileglob_bracket = 258,
  sym__PERLY_SEMICOLON = 259,
  sym__PERLY_HEREDOC = 260,
  sym__ctrl_z_hack = 261,
  sym__quotelike_begin_quote = 262,
  sym__quotelike_middle_close_quote = 263,
  sym__quotelike_middle_skip = 264,
  sym__quotelike_end_zw = 265,
  sym__quotelike_end_quote = 266,
  sym__q_string_content = 267,
  sym__qq_string_content = 268,
  sym_escape_sequence = 269,
  sym_escaped_delimiter = 270,
  sym__dollar_in_regexp = 271,
  sym_pod = 272,
  sym__gobbled_content = 273,
  sym__attribute_value_begin = 274,
  sym_attribute_value = 275,
  sym_prototype = 276,
  sym__signature_start = 277,
  sym__heredoc_delimiter = 278,
  sym__command_heredoc_delimiter = 279,
  sym__heredoc_start = 280,
  sym__heredoc_middle = 281,
  sym_heredoc_end = 282,
  sym__fat_comma_autoquoted = 283,
  sym__filetest = 284,
  sym__brace_autoquoted_token = 285,
  sym__brace_end_zw = 286,
  sym__dollar_ident_zw = 287,
  sym__no_interp_whitespace_zw = 288,
  sym__NONASSOC = 289,
  sym__RECOVER_PAREN_CLOSE = 290,
  sym__RECOVER_BRACKET_CLOSE = 291,
  sym__RECOVER_BRACE_CLOSE = 292,
  sym__RECOVER_ARROW = 293,
  sym__ERROR = 294,
  sym_source_file = 295,
  sym__PERLY_BRACE_OPEN = 296,
  sym_block = 297,
  sym__fullstmt = 298,
  sym_statement_label = 299,
  sym__semicolon = 300,
  sym__barestmt = 301,
  sym_package_statement = 302,
  sym_class_statement = 303,
  sym_role_statement = 304,
  sym_class_phaser_statement = 305,
  sym_use_version_statement = 306,
  sym_use_statement = 307,
  sym_mandatory_parameter = 308,
  sym_optional_parameter = 309,
  sym_named_parameter = 310,
  sym_slurpy_parameter = 311,
  sym__signature_vars = 312,
  sym_signature = 313,
  sym_subroutine_declaration_statement = 314,
  sym_method_declaration_statement = 315,
  sym_phaser_statement = 316,
  sym_conditional_statement = 317,
  sym__loop_body = 318,
  sym_loop_statement = 319,
  sym_cstyle_for_statement = 320,
  sym__for_initializer = 321,
  sym_for_statement = 322,
  sym_try_statement = 323,
  sym_defer_statement = 324,
  sym_expression_statement = 325,
  sym__postfix_expressions = 326,
  sym_postfix_conditional_expression = 327,
  sym_postfix_loop_expression = 328,
  sym_postfix_for_expression = 329,
  sym_yadayada = 330,
  sym__else = 331,
  sym_else = 332,
  sym_elsif = 333,
  sym__expr = 334,
  sym_lowprec_logical_expression = 335,
  sym__listexpr = 336,
  sym__term_rightward = 337,
  sym_subscripted = 338,
  sym_container_variable = 339,
  sym_glob_slot_expression = 340,
  sym_array_element_expression = 341,
  sym__hash_key = 342,
  sym_hash_element_expression = 343,
  sym_coderef_call_expression = 344,
  sym_anonymous_slice_expression = 345,
  sym_slices = 346,
  sym_slice_container_variable = 347,
  sym_slice_expression = 348,
  sym_keyval_container_variable = 349,
  sym_keyval_expression = 350,
  sym__term = 351,
  sym_readline_expression = 352,
  sym_fileglob_expression = 353,
  sym_assignment_expression = 354,
  sym_binary_expression = 355,
  sym_equality_expression = 356,
  sym_relational_expression = 357,
  sym_unary_expression = 358,
  sym_preinc_expression = 359,
  sym_postinc_expression = 360,
  sym_conditional_expression = 361,
  sym_refgen_expression = 362,
  sym_anonymous_array_expression = 363,
  sym__tricky_list = 364,
  sym_anonymous_hash_expression = 365,
  sym_anonymous_subroutine_expression = 366,
  sym_anonymous_method_expression = 367,
  sym_do_expression = 368,
  sym_eval_expression = 369,
  sym_await_expression = 370,
  sym__declared_vars = 371,
  sym_variable_declaration = 372,
  sym__decl_variable_list = 373,
  sym_localization_expression = 374,
  sym_stub_expression = 375,
  sym_postfix_deref = 376,
  sym_scalar_deref_expression = 377,
  sym_array_deref_expression = 378,
  sym_arraylen_deref_expression = 379,
  sym_hash_deref_expression = 380,
  sym_amper_deref_expression = 381,
  sym_glob_deref_expression = 382,
  sym_require_expression = 383,
  sym_require_version_expression = 384,
  sym_func0op_call_expression = 385,
  sym_func1op_call_expression = 386,
  sym_map_grep_expression = 387,
  sym__sort_routine = 388,
  sym_sort_expression = 389,
  sym__label_arg = 390,
  sym_loopex_expression = 391,
  sym_goto_expression = 392,
  sym_return_expression = 393,
  sym_undef_expression = 394,
  sym__listop = 395,
  sym_indirect_object = 396,
  sym_function_call_expression = 397,
  sym__tricky_indirob_hashref = 398,
  sym_ambiguous_function_call_expression = 399,
  sym_function = 400,
  sym_method_call_expression = 401,
  sym_method = 402,
  sym__variables = 403,
  sym__signature_varname = 404,
  sym_scalar = 405,
  sym__declare_scalar = 406,
  sym__signature_scalar = 407,
  sym_array = 408,
  sym__declare_array = 409,
  sym__signature_array = 410,
  sym__SUB_AMPER = 411,
  sym__GLOB_STAR = 412,
  sym_hash = 413,
  sym__declare_hash = 414,
  sym__signature_hash = 415,
  sym_arraylen = 416,
  sym_glob = 417,
  sym_amper_sub = 418,
  sym__indirob = 419,
  sym_varname = 420,
  sym__var_indirob_autoquote = 421,
  sym__var_indirob = 422,
  sym_attrlist = 423,
  sym_attribute = 424,
  sym_attribute_name = 425,
  sym___DATA__ = 426,
  sym__literal = 427,
  sym__apostrophe = 428,
  sym__quotation_mark = 429,
  sym__backtick = 430,
  sym__search_slash = 431,
  sym__quotelike_begin = 432,
  sym__quotelike_middle_close = 433,
  sym__quotelike_end = 434,
  sym__quotelike_middle = 435,
  sym_string_literal = 436,
  sym_interpolated_string_literal = 437,
  sym__subscripted_interpolations = 438,
  sym__braced_scalar = 439,
  sym__braced_array = 440,
  sym__interp_arrow = 441,
  sym__array_element_interpolation = 442,
  sym__hash_element_interpolation = 443,
  sym__slice_expression_interpolation = 444,
  sym__scalar_deref_interpolation = 445,
  sym__array_deref_interpolation = 446,
  sym__interpolations = 447,
  aux_sym__noninterpolated_string_content = 448,
  sym__interpolation_fallbacks = 449,
  sym__nonvar_interpolation_fallbacks = 450,
  aux_sym__interpolated_string_content = 451,
  sym_quoted_word_list = 452,
  sym_command_string = 453,
  sym_quoted_regexp = 454,
  sym_match_regexp = 455,
  sym_substitution_regexp = 456,
  aux_sym__interpolated_regexp_content = 457,
  aux_sym__interpolated_transliteration_content = 458,
  sym_transliteration_expression = 459,
  sym_heredoc_token = 460,
  sym_command_heredoc_token = 461,
  sym_heredoc_content = 462,
  sym_package = 463,
  sym__version = 464,
  sym__conditionals = 465,
  sym__loops = 466,
  sym_autoquoted_bareword = 467,
  sym__brace_autoquoted = 468,
  sym_identifier = 469,
  sym__ident_special = 470,
  sym_bareword = 471,
  sym_primitive = 472,
  sym_boolean = 473,
  aux_sym_source_file_repeat1 = 474,
  aux_sym_signature_repeat1 = 475,
  aux_sym_subroutine_declaration_statement_repeat1 = 476,
  aux_sym__for_initializer_repeat1 = 477,
  aux_sym__term_rightward_repeat1 = 478,
  aux_sym__decl_variable_list_repeat1 = 479,
  aux_sym_attrlist_repeat1 = 480,
  aux_sym_heredoc_content_repeat1 = 481,
  alias_sym_block_statement = 482,
  alias_sym_data_section = 483,
  alias_sym_filehandle = 484,
  alias_sym_filename = 485,
  alias_sym_label = 486,
  anon_alias_sym_not_DASHinterpolated = 487,
  alias_sym_regexp_content = 488,
  alias_sym_replacement = 489,
  alias_sym_string_content = 490,
  alias_sym_transliteration_content = 491,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__identifier] = "_identifier",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_COLON] = ":",
  [anon_sym_SEMI] = ";",
  [anon_sym_package] = "package",
  [anon_sym_class] = "class",
  [anon_sym_role] = "role",
  [anon_sym_BUILD] = "BUILD",
  [anon_sym_ADJUST] = "ADJUST",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_EQ] = "=",
  [anon_sym_PIPE_PIPE_EQ] = "||=",
  [anon_sym_SLASH_SLASH_EQ] = "//=",
  [anon_sym_AT] = "@",
  [anon_sym_COMMA] = ",",
  [anon_sym_RPAREN] = ")",
  [anon_sym_my] = "my",
  [anon_sym_extended] = "extended",
  [anon_sym_async] = "async",
  [anon_sym_sub] = "sub",
  [anon_sym_method] = "method",
  [anon_sym_LPAREN] = "(",
  [anon_sym_continue] = "continue",
  [anon_sym_state] = "state",
  [anon_sym_our] = "our",
  [anon_sym_try] = "try",
  [anon_sym_catch] = "catch",
  [anon_sym_finally] = "finally",
  [anon_sym_defer] = "defer",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_else] = "else",
  [anon_sym_elsif] = "elsif",
  [anon_sym_and] = "and",
  [anon_sym_or] = "or",
  [anon_sym_xor] = "xor",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_STAR] = "*",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_PERCENT] = "%",
  [anon_sym_GT] = ">",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_DOT_EQ] = ".=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_x_EQ] = "x=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_AMP_EQ] = "&&=",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_SLASH_SLASH] = "//",
  [anon_sym_CARET_CARET] = "^^",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_GT_GT2] = ">>",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_DOT] = ".",
  [anon_sym_SLASH] = "/",
  [anon_sym_x] = "x",
  [anon_sym_EQ_TILDE] = "=~",
  [anon_sym_BANG_TILDE] = "!~",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_eq] = "eq",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_equ] = "equ",
  [anon_sym_eqr] = "eqr",
  [anon_sym_ne] = "ne",
  [anon_sym_LT_EQ_GT] = "<=>",
  [anon_sym_cmp] = "cmp",
  [anon_sym_TILDE_TILDE] = "~~",
  [anon_sym_LT] = "<",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_lt] = "lt",
  [anon_sym_le] = "le",
  [anon_sym_ge] = "ge",
  [anon_sym_gt] = "gt",
  [anon_sym_isa] = "isa",
  [anon_sym_TILDE] = "~",
  [anon_sym_BANG] = "!",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_QMARK] = "\?",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_do] = "do",
  [anon_sym_eval] = "eval",
  [anon_sym_await] = "await",
  [anon_sym_field] = "field",
  [anon_sym_local] = "local",
  [anon_sym_dynamically] = "dynamically",
  [anon_sym_DOLLAR_POUND] = "$#",
  [anon_sym_require] = "require",
  [anon_sym_map] = "map",
  [anon_sym_grep] = "grep",
  [anon_sym_sort] = "sort",
  [anon_sym_goto] = "goto",
  [anon_sym_return] = "return",
  [anon_sym_undef] = "undef",
  [anon_sym_PERCENT2] = "%",
  [anon_sym_AMP2] = "&",
  [anon_sym_STAR2] = "*",
  [aux_sym__var_indirob_autoquote_token1] = "varname",
  [anon_sym_EQ_GT] = "=>",
  [anon_sym_use] = "use",
  [anon_sym_no] = "no",
  [anon_sym_for] = "for",
  [anon_sym_foreach] = "foreach",
  [anon_sym_last] = "last",
  [anon_sym_next] = "next",
  [anon_sym_redo] = "redo",
  [anon_sym_BEGIN] = "BEGIN",
  [anon_sym_INIT] = "INIT",
  [anon_sym_CHECK] = "CHECK",
  [anon_sym_UNITCHECK] = "UNITCHECK",
  [anon_sym_END] = "END",
  [anon_sym___FILE__] = "__FILE__",
  [anon_sym___LINE__] = "__LINE__",
  [anon_sym___PACKAGE__] = "__PACKAGE__",
  [anon_sym___SUB__] = "__SUB__",
  [anon_sym_break] = "break",
  [anon_sym_fork] = "fork",
  [anon_sym_getppid] = "getppid",
  [anon_sym_time] = "time",
  [anon_sym_times] = "times",
  [anon_sym_wait] = "wait",
  [anon_sym_wantarray] = "wantarray",
  [anon_sym_abs] = "abs",
  [anon_sym_alarm] = "alarm",
  [anon_sym_chop] = "chop",
  [anon_sym_chdir] = "chdir",
  [anon_sym_close] = "close",
  [anon_sym_closedir] = "closedir",
  [anon_sym_caller] = "caller",
  [anon_sym_chomp] = "chomp",
  [anon_sym_chr] = "chr",
  [anon_sym_cos] = "cos",
  [anon_sym_chroot] = "chroot",
  [anon_sym_defined] = "defined",
  [anon_sym_delete] = "delete",
  [anon_sym_dbmclose] = "dbmclose",
  [anon_sym_exists] = "exists",
  [anon_sym_exit] = "exit",
  [anon_sym_eof] = "eof",
  [anon_sym_exp] = "exp",
  [anon_sym_each] = "each",
  [anon_sym_fc] = "fc",
  [anon_sym_fileno] = "fileno",
  [anon_sym_gmtime] = "gmtime",
  [anon_sym_getc] = "getc",
  [anon_sym_getpgrp] = "getpgrp",
  [anon_sym_getprotobyname] = "getprotobyname",
  [anon_sym_getpwname] = "getpwname",
  [anon_sym_getpwuid] = "getpwuid",
  [anon_sym_getpeername] = "getpeername",
  [anon_sym_getnetbyname] = "getnetbyname",
  [anon_sym_getsockname] = "getsockname",
  [anon_sym_getgrnam] = "getgrnam",
  [anon_sym_getgrgid] = "getgrgid",
  [anon_sym_hex] = "hex",
  [anon_sym_int] = "int",
  [anon_sym_keys] = "keys",
  [anon_sym_lc] = "lc",
  [anon_sym_lcfirst] = "lcfirst",
  [anon_sym_length] = "length",
  [anon_sym_localtime] = "localtime",
  [anon_sym_log] = "log",
  [anon_sym_lock] = "lock",
  [anon_sym_lstat] = "lstat",
  [anon_sym_oct] = "oct",
  [anon_sym_ord] = "ord",
  [anon_sym_prototype] = "prototype",
  [anon_sym_pop] = "pop",
  [anon_sym_pos] = "pos",
  [anon_sym_quotemeta] = "quotemeta",
  [anon_sym_reset] = "reset",
  [anon_sym_rand] = "rand",
  [anon_sym_rmdir] = "rmdir",
  [anon_sym_readdir] = "readdir",
  [anon_sym_readline] = "readline",
  [anon_sym_readpipe] = "readpipe",
  [anon_sym_rewinddir] = "rewinddir",
  [anon_sym_readlink] = "readlink",
  [anon_sym_ref] = "ref",
  [anon_sym_scalar] = "scalar",
  [anon_sym_shift] = "shift",
  [anon_sym_sin] = "sin",
  [anon_sym_sleep] = "sleep",
  [anon_sym_sqrt] = "sqrt",
  [anon_sym_srand] = "srand",
  [anon_sym_stat] = "stat",
  [anon_sym_study] = "study",
  [anon_sym_tell] = "tell",
  [anon_sym_telldir] = "telldir",
  [anon_sym_tied] = "tied",
  [anon_sym_uc] = "uc",
  [anon_sym_ucfirst] = "ucfirst",
  [anon_sym_untie] = "untie",
  [anon_sym_umask] = "umask",
  [anon_sym_values] = "values",
  [anon_sym_write] = "write",
  [sym_comment] = "comment",
  [anon_sym___DATA__] = "eof_marker",
  [anon_sym___END__] = "eof_marker",
  [aux_sym___DATA___token1] = "__DATA___token1",
  [anon_sym_EOT] = "eof_marker",
  [anon_sym_q] = "q",
  [anon_sym_qq] = "qq",
  [anon_sym_DASH_GT2] = "->",
  [anon_sym_LBRACK2] = "[",
  [anon_sym_LBRACE2] = "{",
  [anon_sym_DOLLAR_STAR] = "$*",
  [anon_sym_AT_STAR] = "@*",
  [aux_sym__interpolation_fallbacks_token1] = "_interpolation_fallbacks_token1",
  [aux_sym__interpolation_fallbacks_token2] = "_interpolation_fallbacks_token2",
  [anon_sym_qw] = "qw",
  [anon_sym_qx] = "qx",
  [anon_sym_qr] = "qr",
  [anon_sym_m] = "m",
  [anon_sym_s] = "s",
  [sym_quoted_regexp_modifiers] = "quoted_regexp_modifiers",
  [sym_match_regexp_modifiers] = "match_regexp_modifiers",
  [sym_substitution_regexp_modifiers] = "substitution_regexp_modifiers",
  [sym_transliteration_modifiers] = "transliteration_modifiers",
  [aux_sym__interpolated_transliteration_content_token1] = "_interpolated_transliteration_content_token1",
  [anon_sym_tr] = "tr",
  [anon_sym_y] = "y",
  [sym_version] = "version",
  [anon_sym_if] = "if",
  [anon_sym_unless] = "unless",
  [anon_sym_while] = "while",
  [anon_sym_until] = "until",
  [aux_sym__ident_special_token1] = "_ident_special_token1",
  [aux_sym__bareword_token1] = "_bareword_token1",
  [sym_number] = "number",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [sym__single_quote] = "'",
  [sym__double_quote] = "'",
  [sym__backtick_quote] = "'",
  [sym__search_slash_quote] = "'",
  [sym__no_search_slash_plz] = "_no_search_slash_plz",
  [sym__open_readline_bracket] = "<",
  [sym__open_fileglob_bracket] = "<",
  [sym__PERLY_SEMICOLON] = "_PERLY_SEMICOLON",
  [sym__PERLY_HEREDOC] = "_PERLY_HEREDOC",
  [sym__ctrl_z_hack] = "eof_marker",
  [sym__quotelike_begin_quote] = "'",
  [sym__quotelike_middle_close_quote] = "'",
  [sym__quotelike_middle_skip] = "_quotelike_middle_skip",
  [sym__quotelike_end_zw] = "_quotelike_end_zw",
  [sym__quotelike_end_quote] = "'",
  [sym__q_string_content] = "_q_string_content",
  [sym__qq_string_content] = "_qq_string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_escaped_delimiter] = "escaped_delimiter",
  [sym__dollar_in_regexp] = "_dollar_in_regexp",
  [sym_pod] = "pod",
  [sym__gobbled_content] = "_gobbled_content",
  [sym__attribute_value_begin] = "_attribute_value_begin",
  [sym_attribute_value] = "attribute_value",
  [sym_prototype] = "prototype",
  [sym__signature_start] = "(",
  [sym__heredoc_delimiter] = "_heredoc_delimiter",
  [sym__command_heredoc_delimiter] = "_command_heredoc_delimiter",
  [sym__heredoc_start] = "_heredoc_start",
  [sym__heredoc_middle] = "_heredoc_middle",
  [sym_heredoc_end] = "heredoc_end",
  [sym__fat_comma_autoquoted] = "_fat_comma_autoquoted",
  [sym__filetest] = "-x",
  [sym__brace_autoquoted_token] = "varname",
  [sym__brace_end_zw] = "_brace_end_zw",
  [sym__dollar_ident_zw] = "_dollar_ident_zw",
  [sym__no_interp_whitespace_zw] = "_no_interp_whitespace_zw",
  [sym__NONASSOC] = "_NONASSOC",
  [sym__RECOVER_PAREN_CLOSE] = ")",
  [sym__RECOVER_BRACKET_CLOSE] = "]",
  [sym__RECOVER_BRACE_CLOSE] = "}",
  [sym__RECOVER_ARROW] = "_RECOVER_ARROW",
  [sym__ERROR] = "_ERROR",
  [sym_source_file] = "source_file",
  [sym__PERLY_BRACE_OPEN] = "_PERLY_BRACE_OPEN",
  [sym_block] = "block",
  [sym__fullstmt] = "_fullstmt",
  [sym_statement_label] = "statement_label",
  [sym__semicolon] = "_semicolon",
  [sym__barestmt] = "_barestmt",
  [sym_package_statement] = "package_statement",
  [sym_class_statement] = "class_statement",
  [sym_role_statement] = "role_statement",
  [sym_class_phaser_statement] = "class_phaser_statement",
  [sym_use_version_statement] = "use_version_statement",
  [sym_use_statement] = "use_statement",
  [sym_mandatory_parameter] = "mandatory_parameter",
  [sym_optional_parameter] = "optional_parameter",
  [sym_named_parameter] = "named_parameter",
  [sym_slurpy_parameter] = "slurpy_parameter",
  [sym__signature_vars] = "_signature_vars",
  [sym_signature] = "signature",
  [sym_subroutine_declaration_statement] = "subroutine_declaration_statement",
  [sym_method_declaration_statement] = "method_declaration_statement",
  [sym_phaser_statement] = "phaser_statement",
  [sym_conditional_statement] = "conditional_statement",
  [sym__loop_body] = "_loop_body",
  [sym_loop_statement] = "loop_statement",
  [sym_cstyle_for_statement] = "cstyle_for_statement",
  [sym__for_initializer] = "_for_initializer",
  [sym_for_statement] = "for_statement",
  [sym_try_statement] = "try_statement",
  [sym_defer_statement] = "defer_statement",
  [sym_expression_statement] = "expression_statement",
  [sym__postfix_expressions] = "_postfix_expressions",
  [sym_postfix_conditional_expression] = "postfix_conditional_expression",
  [sym_postfix_loop_expression] = "postfix_loop_expression",
  [sym_postfix_for_expression] = "postfix_for_expression",
  [sym_yadayada] = "yadayada",
  [sym__else] = "_else",
  [sym_else] = "else",
  [sym_elsif] = "elsif",
  [sym__expr] = "_expr",
  [sym_lowprec_logical_expression] = "lowprec_logical_expression",
  [sym__listexpr] = "_listexpr",
  [sym__term_rightward] = "list_expression",
  [sym_subscripted] = "subscripted",
  [sym_container_variable] = "container_variable",
  [sym_glob_slot_expression] = "glob_slot_expression",
  [sym_array_element_expression] = "array_element_expression",
  [sym__hash_key] = "_hash_key",
  [sym_hash_element_expression] = "hash_element_expression",
  [sym_coderef_call_expression] = "coderef_call_expression",
  [sym_anonymous_slice_expression] = "anonymous_slice_expression",
  [sym_slices] = "slices",
  [sym_slice_container_variable] = "slice_container_variable",
  [sym_slice_expression] = "slice_expression",
  [sym_keyval_container_variable] = "keyval_container_variable",
  [sym_keyval_expression] = "keyval_expression",
  [sym__term] = "_term",
  [sym_readline_expression] = "readline_expression",
  [sym_fileglob_expression] = "fileglob_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_equality_expression] = "equality_expression",
  [sym_relational_expression] = "relational_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_preinc_expression] = "preinc_expression",
  [sym_postinc_expression] = "postinc_expression",
  [sym_conditional_expression] = "conditional_expression",
  [sym_refgen_expression] = "refgen_expression",
  [sym_anonymous_array_expression] = "anonymous_array_expression",
  [sym__tricky_list] = "list_expression",
  [sym_anonymous_hash_expression] = "anonymous_hash_expression",
  [sym_anonymous_subroutine_expression] = "anonymous_subroutine_expression",
  [sym_anonymous_method_expression] = "anonymous_method_expression",
  [sym_do_expression] = "do_expression",
  [sym_eval_expression] = "eval_expression",
  [sym_await_expression] = "await_expression",
  [sym__declared_vars] = "_declared_vars",
  [sym_variable_declaration] = "variable_declaration",
  [sym__decl_variable_list] = "_decl_variable_list",
  [sym_localization_expression] = "localization_expression",
  [sym_stub_expression] = "stub_expression",
  [sym_postfix_deref] = "postfix_deref",
  [sym_scalar_deref_expression] = "scalar_deref_expression",
  [sym_array_deref_expression] = "array_deref_expression",
  [sym_arraylen_deref_expression] = "arraylen_deref_expression",
  [sym_hash_deref_expression] = "hash_deref_expression",
  [sym_amper_deref_expression] = "amper_deref_expression",
  [sym_glob_deref_expression] = "glob_deref_expression",
  [sym_require_expression] = "require_expression",
  [sym_require_version_expression] = "require_version_expression",
  [sym_func0op_call_expression] = "func0op_call_expression",
  [sym_func1op_call_expression] = "func1op_call_expression",
  [sym_map_grep_expression] = "map_grep_expression",
  [sym__sort_routine] = "_sort_routine",
  [sym_sort_expression] = "sort_expression",
  [sym__label_arg] = "_label_arg",
  [sym_loopex_expression] = "loopex_expression",
  [sym_goto_expression] = "goto_expression",
  [sym_return_expression] = "return_expression",
  [sym_undef_expression] = "undef_expression",
  [sym__listop] = "_listop",
  [sym_indirect_object] = "indirect_object",
  [sym_function_call_expression] = "function_call_expression",
  [sym__tricky_indirob_hashref] = "anonymous_hash_expression",
  [sym_ambiguous_function_call_expression] = "ambiguous_function_call_expression",
  [sym_function] = "function",
  [sym_method_call_expression] = "method_call_expression",
  [sym_method] = "method",
  [sym__variables] = "_variables",
  [sym__signature_varname] = "_signature_varname",
  [sym_scalar] = "scalar",
  [sym__declare_scalar] = "scalar",
  [sym__signature_scalar] = "scalar",
  [sym_array] = "array",
  [sym__declare_array] = "array",
  [sym__signature_array] = "array",
  [sym__SUB_AMPER] = "_SUB_AMPER",
  [sym__GLOB_STAR] = "_GLOB_STAR",
  [sym_hash] = "hash",
  [sym__declare_hash] = "hash",
  [sym__signature_hash] = "hash",
  [sym_arraylen] = "arraylen",
  [sym_glob] = "glob",
  [sym_amper_sub] = "function",
  [sym__indirob] = "varname",
  [sym_varname] = "varname",
  [sym__var_indirob_autoquote] = "_var_indirob_autoquote",
  [sym__var_indirob] = "_var_indirob",
  [sym_attrlist] = "attrlist",
  [sym_attribute] = "attribute",
  [sym_attribute_name] = "attribute_name",
  [sym___DATA__] = "__DATA__",
  [sym__literal] = "_literal",
  [sym__apostrophe] = "_apostrophe",
  [sym__quotation_mark] = "_quotation_mark",
  [sym__backtick] = "_backtick",
  [sym__search_slash] = "_search_slash",
  [sym__quotelike_begin] = "_quotelike_begin",
  [sym__quotelike_middle_close] = "_quotelike_middle_close",
  [sym__quotelike_end] = "_quotelike_end",
  [sym__quotelike_middle] = "_quotelike_middle",
  [sym_string_literal] = "string_literal",
  [sym_interpolated_string_literal] = "interpolated_string_literal",
  [sym__subscripted_interpolations] = "_subscripted_interpolations",
  [sym__braced_scalar] = "scalar",
  [sym__braced_array] = "array",
  [sym__interp_arrow] = "_interp_arrow",
  [sym__array_element_interpolation] = "array_element_expression",
  [sym__hash_element_interpolation] = "hash_element_expression",
  [sym__slice_expression_interpolation] = "slice_expression",
  [sym__scalar_deref_interpolation] = "scalar_deref_expression",
  [sym__array_deref_interpolation] = "array_deref_expression",
  [sym__interpolations] = "_interpolations",
  [aux_sym__noninterpolated_string_content] = "_noninterpolated_string_content",
  [sym__interpolation_fallbacks] = "_interpolation_fallbacks",
  [sym__nonvar_interpolation_fallbacks] = "_nonvar_interpolation_fallbacks",
  [aux_sym__interpolated_string_content] = "_interpolated_string_content",
  [sym_quoted_word_list] = "quoted_word_list",
  [sym_command_string] = "command_string",
  [sym_quoted_regexp] = "quoted_regexp",
  [sym_match_regexp] = "match_regexp",
  [sym_substitution_regexp] = "substitution_regexp",
  [aux_sym__interpolated_regexp_content] = "_interpolated_regexp_content",
  [aux_sym__interpolated_transliteration_content] = "_interpolated_transliteration_content",
  [sym_transliteration_expression] = "transliteration_expression",
  [sym_heredoc_token] = "heredoc_token",
  [sym_command_heredoc_token] = "command_heredoc_token",
  [sym_heredoc_content] = "heredoc_content",
  [sym_package] = "package",
  [sym__version] = "_version",
  [sym__conditionals] = "_conditionals",
  [sym__loops] = "_loops",
  [sym_autoquoted_bareword] = "autoquoted_bareword",
  [sym__brace_autoquoted] = "_brace_autoquoted",
  [sym_identifier] = "identifier",
  [sym__ident_special] = "_ident_special",
  [sym_bareword] = "bareword",
  [sym_primitive] = "primitive",
  [sym_boolean] = "boolean",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_signature_repeat1] = "signature_repeat1",
  [aux_sym_subroutine_declaration_statement_repeat1] = "subroutine_declaration_statement_repeat1",
  [aux_sym__for_initializer_repeat1] = "_for_initializer_repeat1",
  [aux_sym__term_rightward_repeat1] = "_term_rightward_repeat1",
  [aux_sym__decl_variable_list_repeat1] = "_decl_variable_list_repeat1",
  [aux_sym_attrlist_repeat1] = "attrlist_repeat1",
  [aux_sym_heredoc_content_repeat1] = "heredoc_content_repeat1",
  [alias_sym_block_statement] = "block_statement",
  [alias_sym_data_section] = "data_section",
  [alias_sym_filehandle] = "filehandle",
  [alias_sym_filename] = "filename",
  [alias_sym_label] = "label",
  [anon_alias_sym_not_DASHinterpolated] = "not-interpolated",
  [alias_sym_regexp_content] = "regexp_content",
  [alias_sym_replacement] = "replacement",
  [alias_sym_string_content] = "string_content",
  [alias_sym_transliteration_content] = "transliteration_content",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__identifier] = sym__identifier,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_package] = anon_sym_package,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_role] = anon_sym_role,
  [anon_sym_BUILD] = anon_sym_BUILD,
  [anon_sym_ADJUST] = anon_sym_ADJUST,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_PIPE_PIPE_EQ] = anon_sym_PIPE_PIPE_EQ,
  [anon_sym_SLASH_SLASH_EQ] = anon_sym_SLASH_SLASH_EQ,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_my] = anon_sym_my,
  [anon_sym_extended] = anon_sym_extended,
  [anon_sym_async] = anon_sym_async,
  [anon_sym_sub] = anon_sym_sub,
  [anon_sym_method] = anon_sym_method,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_state] = anon_sym_state,
  [anon_sym_our] = anon_sym_our,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_catch] = anon_sym_catch,
  [anon_sym_finally] = anon_sym_finally,
  [anon_sym_defer] = anon_sym_defer,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_elsif] = anon_sym_elsif,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_xor] = anon_sym_xor,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_DOT_EQ] = anon_sym_DOT_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_x_EQ] = anon_sym_x_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_AMP_EQ] = anon_sym_AMP_AMP_EQ,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_SLASH_SLASH] = anon_sym_SLASH_SLASH,
  [anon_sym_CARET_CARET] = anon_sym_CARET_CARET,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_GT_GT2] = anon_sym_GT_GT,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_x] = anon_sym_x,
  [anon_sym_EQ_TILDE] = anon_sym_EQ_TILDE,
  [anon_sym_BANG_TILDE] = anon_sym_BANG_TILDE,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_eq] = anon_sym_eq,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_equ] = anon_sym_equ,
  [anon_sym_eqr] = anon_sym_eqr,
  [anon_sym_ne] = anon_sym_ne,
  [anon_sym_LT_EQ_GT] = anon_sym_LT_EQ_GT,
  [anon_sym_cmp] = anon_sym_cmp,
  [anon_sym_TILDE_TILDE] = anon_sym_TILDE_TILDE,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_lt] = anon_sym_lt,
  [anon_sym_le] = anon_sym_le,
  [anon_sym_ge] = anon_sym_ge,
  [anon_sym_gt] = anon_sym_gt,
  [anon_sym_isa] = anon_sym_isa,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_eval] = anon_sym_eval,
  [anon_sym_await] = anon_sym_await,
  [anon_sym_field] = anon_sym_field,
  [anon_sym_local] = anon_sym_local,
  [anon_sym_dynamically] = anon_sym_dynamically,
  [anon_sym_DOLLAR_POUND] = anon_sym_DOLLAR_POUND,
  [anon_sym_require] = anon_sym_require,
  [anon_sym_map] = anon_sym_map,
  [anon_sym_grep] = anon_sym_grep,
  [anon_sym_sort] = anon_sym_sort,
  [anon_sym_goto] = anon_sym_goto,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_undef] = anon_sym_undef,
  [anon_sym_PERCENT2] = anon_sym_PERCENT,
  [anon_sym_AMP2] = anon_sym_AMP,
  [anon_sym_STAR2] = anon_sym_STAR,
  [aux_sym__var_indirob_autoquote_token1] = sym_varname,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [anon_sym_use] = anon_sym_use,
  [anon_sym_no] = anon_sym_no,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_foreach] = anon_sym_foreach,
  [anon_sym_last] = anon_sym_last,
  [anon_sym_next] = anon_sym_next,
  [anon_sym_redo] = anon_sym_redo,
  [anon_sym_BEGIN] = anon_sym_BEGIN,
  [anon_sym_INIT] = anon_sym_INIT,
  [anon_sym_CHECK] = anon_sym_CHECK,
  [anon_sym_UNITCHECK] = anon_sym_UNITCHECK,
  [anon_sym_END] = anon_sym_END,
  [anon_sym___FILE__] = anon_sym___FILE__,
  [anon_sym___LINE__] = anon_sym___LINE__,
  [anon_sym___PACKAGE__] = anon_sym___PACKAGE__,
  [anon_sym___SUB__] = anon_sym___SUB__,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_fork] = anon_sym_fork,
  [anon_sym_getppid] = anon_sym_getppid,
  [anon_sym_time] = anon_sym_time,
  [anon_sym_times] = anon_sym_times,
  [anon_sym_wait] = anon_sym_wait,
  [anon_sym_wantarray] = anon_sym_wantarray,
  [anon_sym_abs] = anon_sym_abs,
  [anon_sym_alarm] = anon_sym_alarm,
  [anon_sym_chop] = anon_sym_chop,
  [anon_sym_chdir] = anon_sym_chdir,
  [anon_sym_close] = anon_sym_close,
  [anon_sym_closedir] = anon_sym_closedir,
  [anon_sym_caller] = anon_sym_caller,
  [anon_sym_chomp] = anon_sym_chomp,
  [anon_sym_chr] = anon_sym_chr,
  [anon_sym_cos] = anon_sym_cos,
  [anon_sym_chroot] = anon_sym_chroot,
  [anon_sym_defined] = anon_sym_defined,
  [anon_sym_delete] = anon_sym_delete,
  [anon_sym_dbmclose] = anon_sym_dbmclose,
  [anon_sym_exists] = anon_sym_exists,
  [anon_sym_exit] = anon_sym_exit,
  [anon_sym_eof] = anon_sym_eof,
  [anon_sym_exp] = anon_sym_exp,
  [anon_sym_each] = anon_sym_each,
  [anon_sym_fc] = anon_sym_fc,
  [anon_sym_fileno] = anon_sym_fileno,
  [anon_sym_gmtime] = anon_sym_gmtime,
  [anon_sym_getc] = anon_sym_getc,
  [anon_sym_getpgrp] = anon_sym_getpgrp,
  [anon_sym_getprotobyname] = anon_sym_getprotobyname,
  [anon_sym_getpwname] = anon_sym_getpwname,
  [anon_sym_getpwuid] = anon_sym_getpwuid,
  [anon_sym_getpeername] = anon_sym_getpeername,
  [anon_sym_getnetbyname] = anon_sym_getnetbyname,
  [anon_sym_getsockname] = anon_sym_getsockname,
  [anon_sym_getgrnam] = anon_sym_getgrnam,
  [anon_sym_getgrgid] = anon_sym_getgrgid,
  [anon_sym_hex] = anon_sym_hex,
  [anon_sym_int] = anon_sym_int,
  [anon_sym_keys] = anon_sym_keys,
  [anon_sym_lc] = anon_sym_lc,
  [anon_sym_lcfirst] = anon_sym_lcfirst,
  [anon_sym_length] = anon_sym_length,
  [anon_sym_localtime] = anon_sym_localtime,
  [anon_sym_log] = anon_sym_log,
  [anon_sym_lock] = anon_sym_lock,
  [anon_sym_lstat] = anon_sym_lstat,
  [anon_sym_oct] = anon_sym_oct,
  [anon_sym_ord] = anon_sym_ord,
  [anon_sym_prototype] = anon_sym_prototype,
  [anon_sym_pop] = anon_sym_pop,
  [anon_sym_pos] = anon_sym_pos,
  [anon_sym_quotemeta] = anon_sym_quotemeta,
  [anon_sym_reset] = anon_sym_reset,
  [anon_sym_rand] = anon_sym_rand,
  [anon_sym_rmdir] = anon_sym_rmdir,
  [anon_sym_readdir] = anon_sym_readdir,
  [anon_sym_readline] = anon_sym_readline,
  [anon_sym_readpipe] = anon_sym_readpipe,
  [anon_sym_rewinddir] = anon_sym_rewinddir,
  [anon_sym_readlink] = anon_sym_readlink,
  [anon_sym_ref] = anon_sym_ref,
  [anon_sym_scalar] = anon_sym_scalar,
  [anon_sym_shift] = anon_sym_shift,
  [anon_sym_sin] = anon_sym_sin,
  [anon_sym_sleep] = anon_sym_sleep,
  [anon_sym_sqrt] = anon_sym_sqrt,
  [anon_sym_srand] = anon_sym_srand,
  [anon_sym_stat] = anon_sym_stat,
  [anon_sym_study] = anon_sym_study,
  [anon_sym_tell] = anon_sym_tell,
  [anon_sym_telldir] = anon_sym_telldir,
  [anon_sym_tied] = anon_sym_tied,
  [anon_sym_uc] = anon_sym_uc,
  [anon_sym_ucfirst] = anon_sym_ucfirst,
  [anon_sym_untie] = anon_sym_untie,
  [anon_sym_umask] = anon_sym_umask,
  [anon_sym_values] = anon_sym_values,
  [anon_sym_write] = anon_sym_write,
  [sym_comment] = sym_comment,
  [anon_sym___DATA__] = sym__ctrl_z_hack,
  [anon_sym___END__] = sym__ctrl_z_hack,
  [aux_sym___DATA___token1] = aux_sym___DATA___token1,
  [anon_sym_EOT] = sym__ctrl_z_hack,
  [anon_sym_q] = anon_sym_q,
  [anon_sym_qq] = anon_sym_qq,
  [anon_sym_DASH_GT2] = anon_sym_DASH_GT,
  [anon_sym_LBRACK2] = anon_sym_LBRACK,
  [anon_sym_LBRACE2] = anon_sym_LBRACE,
  [anon_sym_DOLLAR_STAR] = anon_sym_DOLLAR_STAR,
  [anon_sym_AT_STAR] = anon_sym_AT_STAR,
  [aux_sym__interpolation_fallbacks_token1] = aux_sym__interpolation_fallbacks_token1,
  [aux_sym__interpolation_fallbacks_token2] = aux_sym__interpolation_fallbacks_token2,
  [anon_sym_qw] = anon_sym_qw,
  [anon_sym_qx] = anon_sym_qx,
  [anon_sym_qr] = anon_sym_qr,
  [anon_sym_m] = anon_sym_m,
  [anon_sym_s] = anon_sym_s,
  [sym_quoted_regexp_modifiers] = sym_quoted_regexp_modifiers,
  [sym_match_regexp_modifiers] = sym_match_regexp_modifiers,
  [sym_substitution_regexp_modifiers] = sym_substitution_regexp_modifiers,
  [sym_transliteration_modifiers] = sym_transliteration_modifiers,
  [aux_sym__interpolated_transliteration_content_token1] = aux_sym__interpolated_transliteration_content_token1,
  [anon_sym_tr] = anon_sym_tr,
  [anon_sym_y] = anon_sym_y,
  [sym_version] = sym_version,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_unless] = anon_sym_unless,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_until] = anon_sym_until,
  [aux_sym__ident_special_token1] = aux_sym__ident_special_token1,
  [aux_sym__bareword_token1] = aux_sym__bareword_token1,
  [sym_number] = sym_number,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [sym__single_quote] = sym__single_quote,
  [sym__double_quote] = sym__single_quote,
  [sym__backtick_quote] = sym__single_quote,
  [sym__search_slash_quote] = sym__single_quote,
  [sym__no_search_slash_plz] = sym__no_search_slash_plz,
  [sym__open_readline_bracket] = anon_sym_LT,
  [sym__open_fileglob_bracket] = anon_sym_LT,
  [sym__PERLY_SEMICOLON] = sym__PERLY_SEMICOLON,
  [sym__PERLY_HEREDOC] = sym__PERLY_HEREDOC,
  [sym__ctrl_z_hack] = sym__ctrl_z_hack,
  [sym__quotelike_begin_quote] = sym__single_quote,
  [sym__quotelike_middle_close_quote] = sym__single_quote,
  [sym__quotelike_middle_skip] = sym__quotelike_middle_skip,
  [sym__quotelike_end_zw] = sym__quotelike_end_zw,
  [sym__quotelike_end_quote] = sym__single_quote,
  [sym__q_string_content] = sym__q_string_content,
  [sym__qq_string_content] = sym__qq_string_content,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_escaped_delimiter] = sym_escaped_delimiter,
  [sym__dollar_in_regexp] = sym__dollar_in_regexp,
  [sym_pod] = sym_pod,
  [sym__gobbled_content] = sym__gobbled_content,
  [sym__attribute_value_begin] = sym__attribute_value_begin,
  [sym_attribute_value] = sym_attribute_value,
  [sym_prototype] = sym_prototype,
  [sym__signature_start] = anon_sym_LPAREN,
  [sym__heredoc_delimiter] = sym__heredoc_delimiter,
  [sym__command_heredoc_delimiter] = sym__command_heredoc_delimiter,
  [sym__heredoc_start] = sym__heredoc_start,
  [sym__heredoc_middle] = sym__heredoc_middle,
  [sym_heredoc_end] = sym_heredoc_end,
  [sym__fat_comma_autoquoted] = sym__fat_comma_autoquoted,
  [sym__filetest] = sym__filetest,
  [sym__brace_autoquoted_token] = sym_varname,
  [sym__brace_end_zw] = sym__brace_end_zw,
  [sym__dollar_ident_zw] = sym__dollar_ident_zw,
  [sym__no_interp_whitespace_zw] = sym__no_interp_whitespace_zw,
  [sym__NONASSOC] = sym__NONASSOC,
  [sym__RECOVER_PAREN_CLOSE] = anon_sym_RPAREN,
  [sym__RECOVER_BRACKET_CLOSE] = anon_sym_RBRACK,
  [sym__RECOVER_BRACE_CLOSE] = anon_sym_RBRACE,
  [sym__RECOVER_ARROW] = sym__RECOVER_ARROW,
  [sym__ERROR] = sym__ERROR,
  [sym_source_file] = sym_source_file,
  [sym__PERLY_BRACE_OPEN] = sym__PERLY_BRACE_OPEN,
  [sym_block] = sym_block,
  [sym__fullstmt] = sym__fullstmt,
  [sym_statement_label] = sym_statement_label,
  [sym__semicolon] = sym__semicolon,
  [sym__barestmt] = sym__barestmt,
  [sym_package_statement] = sym_package_statement,
  [sym_class_statement] = sym_class_statement,
  [sym_role_statement] = sym_role_statement,
  [sym_class_phaser_statement] = sym_class_phaser_statement,
  [sym_use_version_statement] = sym_use_version_statement,
  [sym_use_statement] = sym_use_statement,
  [sym_mandatory_parameter] = sym_mandatory_parameter,
  [sym_optional_parameter] = sym_optional_parameter,
  [sym_named_parameter] = sym_named_parameter,
  [sym_slurpy_parameter] = sym_slurpy_parameter,
  [sym__signature_vars] = sym__signature_vars,
  [sym_signature] = sym_signature,
  [sym_subroutine_declaration_statement] = sym_subroutine_declaration_statement,
  [sym_method_declaration_statement] = sym_method_declaration_statement,
  [sym_phaser_statement] = sym_phaser_statement,
  [sym_conditional_statement] = sym_conditional_statement,
  [sym__loop_body] = sym__loop_body,
  [sym_loop_statement] = sym_loop_statement,
  [sym_cstyle_for_statement] = sym_cstyle_for_statement,
  [sym__for_initializer] = sym__for_initializer,
  [sym_for_statement] = sym_for_statement,
  [sym_try_statement] = sym_try_statement,
  [sym_defer_statement] = sym_defer_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym__postfix_expressions] = sym__postfix_expressions,
  [sym_postfix_conditional_expression] = sym_postfix_conditional_expression,
  [sym_postfix_loop_expression] = sym_postfix_loop_expression,
  [sym_postfix_for_expression] = sym_postfix_for_expression,
  [sym_yadayada] = sym_yadayada,
  [sym__else] = sym__else,
  [sym_else] = sym_else,
  [sym_elsif] = sym_elsif,
  [sym__expr] = sym__expr,
  [sym_lowprec_logical_expression] = sym_lowprec_logical_expression,
  [sym__listexpr] = sym__listexpr,
  [sym__term_rightward] = sym__term_rightward,
  [sym_subscripted] = sym_subscripted,
  [sym_container_variable] = sym_container_variable,
  [sym_glob_slot_expression] = sym_glob_slot_expression,
  [sym_array_element_expression] = sym_array_element_expression,
  [sym__hash_key] = sym__hash_key,
  [sym_hash_element_expression] = sym_hash_element_expression,
  [sym_coderef_call_expression] = sym_coderef_call_expression,
  [sym_anonymous_slice_expression] = sym_anonymous_slice_expression,
  [sym_slices] = sym_slices,
  [sym_slice_container_variable] = sym_slice_container_variable,
  [sym_slice_expression] = sym_slice_expression,
  [sym_keyval_container_variable] = sym_keyval_container_variable,
  [sym_keyval_expression] = sym_keyval_expression,
  [sym__term] = sym__term,
  [sym_readline_expression] = sym_readline_expression,
  [sym_fileglob_expression] = sym_fileglob_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_equality_expression] = sym_equality_expression,
  [sym_relational_expression] = sym_relational_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_preinc_expression] = sym_preinc_expression,
  [sym_postinc_expression] = sym_postinc_expression,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_refgen_expression] = sym_refgen_expression,
  [sym_anonymous_array_expression] = sym_anonymous_array_expression,
  [sym__tricky_list] = sym__term_rightward,
  [sym_anonymous_hash_expression] = sym_anonymous_hash_expression,
  [sym_anonymous_subroutine_expression] = sym_anonymous_subroutine_expression,
  [sym_anonymous_method_expression] = sym_anonymous_method_expression,
  [sym_do_expression] = sym_do_expression,
  [sym_eval_expression] = sym_eval_expression,
  [sym_await_expression] = sym_await_expression,
  [sym__declared_vars] = sym__declared_vars,
  [sym_variable_declaration] = sym_variable_declaration,
  [sym__decl_variable_list] = sym__decl_variable_list,
  [sym_localization_expression] = sym_localization_expression,
  [sym_stub_expression] = sym_stub_expression,
  [sym_postfix_deref] = sym_postfix_deref,
  [sym_scalar_deref_expression] = sym_scalar_deref_expression,
  [sym_array_deref_expression] = sym_array_deref_expression,
  [sym_arraylen_deref_expression] = sym_arraylen_deref_expression,
  [sym_hash_deref_expression] = sym_hash_deref_expression,
  [sym_amper_deref_expression] = sym_amper_deref_expression,
  [sym_glob_deref_expression] = sym_glob_deref_expression,
  [sym_require_expression] = sym_require_expression,
  [sym_require_version_expression] = sym_require_version_expression,
  [sym_func0op_call_expression] = sym_func0op_call_expression,
  [sym_func1op_call_expression] = sym_func1op_call_expression,
  [sym_map_grep_expression] = sym_map_grep_expression,
  [sym__sort_routine] = sym__sort_routine,
  [sym_sort_expression] = sym_sort_expression,
  [sym__label_arg] = sym__label_arg,
  [sym_loopex_expression] = sym_loopex_expression,
  [sym_goto_expression] = sym_goto_expression,
  [sym_return_expression] = sym_return_expression,
  [sym_undef_expression] = sym_undef_expression,
  [sym__listop] = sym__listop,
  [sym_indirect_object] = sym_indirect_object,
  [sym_function_call_expression] = sym_function_call_expression,
  [sym__tricky_indirob_hashref] = sym_anonymous_hash_expression,
  [sym_ambiguous_function_call_expression] = sym_ambiguous_function_call_expression,
  [sym_function] = sym_function,
  [sym_method_call_expression] = sym_method_call_expression,
  [sym_method] = sym_method,
  [sym__variables] = sym__variables,
  [sym__signature_varname] = sym__signature_varname,
  [sym_scalar] = sym_scalar,
  [sym__declare_scalar] = sym_scalar,
  [sym__signature_scalar] = sym_scalar,
  [sym_array] = sym_array,
  [sym__declare_array] = sym_array,
  [sym__signature_array] = sym_array,
  [sym__SUB_AMPER] = sym__SUB_AMPER,
  [sym__GLOB_STAR] = sym__GLOB_STAR,
  [sym_hash] = sym_hash,
  [sym__declare_hash] = sym_hash,
  [sym__signature_hash] = sym_hash,
  [sym_arraylen] = sym_arraylen,
  [sym_glob] = sym_glob,
  [sym_amper_sub] = sym_function,
  [sym__indirob] = sym_varname,
  [sym_varname] = sym_varname,
  [sym__var_indirob_autoquote] = sym__var_indirob_autoquote,
  [sym__var_indirob] = sym__var_indirob,
  [sym_attrlist] = sym_attrlist,
  [sym_attribute] = sym_attribute,
  [sym_attribute_name] = sym_attribute_name,
  [sym___DATA__] = sym___DATA__,
  [sym__literal] = sym__literal,
  [sym__apostrophe] = sym__apostrophe,
  [sym__quotation_mark] = sym__quotation_mark,
  [sym__backtick] = sym__backtick,
  [sym__search_slash] = sym__search_slash,
  [sym__quotelike_begin] = sym__quotelike_begin,
  [sym__quotelike_middle_close] = sym__quotelike_middle_close,
  [sym__quotelike_end] = sym__quotelike_end,
  [sym__quotelike_middle] = sym__quotelike_middle,
  [sym_string_literal] = sym_string_literal,
  [sym_interpolated_string_literal] = sym_interpolated_string_literal,
  [sym__subscripted_interpolations] = sym__subscripted_interpolations,
  [sym__braced_scalar] = sym_scalar,
  [sym__braced_array] = sym_array,
  [sym__interp_arrow] = sym__interp_arrow,
  [sym__array_element_interpolation] = sym_array_element_expression,
  [sym__hash_element_interpolation] = sym_hash_element_expression,
  [sym__slice_expression_interpolation] = sym_slice_expression,
  [sym__scalar_deref_interpolation] = sym_scalar_deref_expression,
  [sym__array_deref_interpolation] = sym_array_deref_expression,
  [sym__interpolations] = sym__interpolations,
  [aux_sym__noninterpolated_string_content] = aux_sym__noninterpolated_string_content,
  [sym__interpolation_fallbacks] = sym__interpolation_fallbacks,
  [sym__nonvar_interpolation_fallbacks] = sym__nonvar_interpolation_fallbacks,
  [aux_sym__interpolated_string_content] = aux_sym__interpolated_string_content,
  [sym_quoted_word_list] = sym_quoted_word_list,
  [sym_command_string] = sym_command_string,
  [sym_quoted_regexp] = sym_quoted_regexp,
  [sym_match_regexp] = sym_match_regexp,
  [sym_substitution_regexp] = sym_substitution_regexp,
  [aux_sym__interpolated_regexp_content] = aux_sym__interpolated_regexp_content,
  [aux_sym__interpolated_transliteration_content] = aux_sym__interpolated_transliteration_content,
  [sym_transliteration_expression] = sym_transliteration_expression,
  [sym_heredoc_token] = sym_heredoc_token,
  [sym_command_heredoc_token] = sym_command_heredoc_token,
  [sym_heredoc_content] = sym_heredoc_content,
  [sym_package] = sym_package,
  [sym__version] = sym__version,
  [sym__conditionals] = sym__conditionals,
  [sym__loops] = sym__loops,
  [sym_autoquoted_bareword] = sym_autoquoted_bareword,
  [sym__brace_autoquoted] = sym__brace_autoquoted,
  [sym_identifier] = sym_identifier,
  [sym__ident_special] = sym__ident_special,
  [sym_bareword] = sym_bareword,
  [sym_primitive] = sym_primitive,
  [sym_boolean] = sym_boolean,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_signature_repeat1] = aux_sym_signature_repeat1,
  [aux_sym_subroutine_declaration_statement_repeat1] = aux_sym_subroutine_declaration_statement_repeat1,
  [aux_sym__for_initializer_repeat1] = aux_sym__for_initializer_repeat1,
  [aux_sym__term_rightward_repeat1] = aux_sym__term_rightward_repeat1,
  [aux_sym__decl_variable_list_repeat1] = aux_sym__decl_variable_list_repeat1,
  [aux_sym_attrlist_repeat1] = aux_sym_attrlist_repeat1,
  [aux_sym_heredoc_content_repeat1] = aux_sym_heredoc_content_repeat1,
  [alias_sym_block_statement] = alias_sym_block_statement,
  [alias_sym_data_section] = alias_sym_data_section,
  [alias_sym_filehandle] = alias_sym_filehandle,
  [alias_sym_filename] = alias_sym_filename,
  [alias_sym_label] = alias_sym_label,
  [anon_alias_sym_not_DASHinterpolated] = anon_alias_sym_not_DASHinterpolated,
  [alias_sym_regexp_content] = alias_sym_regexp_content,
  [alias_sym_replacement] = alias_sym_replacement,
  [alias_sym_string_content] = alias_sym_string_content,
  [alias_sym_transliteration_content] = alias_sym_transliteration_content,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__identifier] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_package] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_role] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BUILD] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ADJUST] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_my] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extended] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_async] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sub] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_method] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_state] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_our] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_catch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_defer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_elsif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_xor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_x_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_x] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_eq] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_equ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_eqr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ne] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_cmp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_le] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ge] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_isa] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_eval] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_await] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_field] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_local] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dynamically] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_require] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_map] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_grep] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sort] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_goto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_undef] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__var_indirob_autoquote_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_use] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_no] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_foreach] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_last] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_next] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_redo] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BEGIN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_INIT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CHECK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_UNITCHECK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_END] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___FILE__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___LINE__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___PACKAGE__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___SUB__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fork] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getppid] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_time] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_times] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_wait] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_wantarray] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_abs] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alarm] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chop] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chdir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_close] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_closedir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_caller] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chomp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_cos] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chroot] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_defined] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_delete] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dbmclose] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exists] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_eof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_each] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fc] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fileno] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_gmtime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getc] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getpgrp] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getprotobyname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getpwname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getpwuid] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getpeername] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getnetbyname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getsockname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getgrnam] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_getgrgid] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_hex] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_keys] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lc] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lcfirst] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_length] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_localtime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_log] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lock] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lstat] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_oct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ord] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_prototype] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_pop] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_pos] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_quotemeta] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_reset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rand] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rmdir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_readdir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_readline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_readpipe] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rewinddir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_readlink] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ref] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_scalar] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_shift] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sin] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sleep] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sqrt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_srand] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_stat] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_study] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_tell] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_telldir] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_tied] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_uc] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ucfirst] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_untie] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_umask] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_values] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_write] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [anon_sym___DATA__] = {
    .visible = true,
    .named = true,
  },
  [anon_sym___END__] = {
    .visible = true,
    .named = true,
  },
  [aux_sym___DATA___token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_EOT] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_q] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_qq] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_STAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__interpolation_fallbacks_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolation_fallbacks_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_qw] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_qx] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_qr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_m] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_s] = {
    .visible = true,
    .named = false,
  },
  [sym_quoted_regexp_modifiers] = {
    .visible = true,
    .named = true,
  },
  [sym_match_regexp_modifiers] = {
    .visible = true,
    .named = true,
  },
  [sym_substitution_regexp_modifiers] = {
    .visible = true,
    .named = true,
  },
  [sym_transliteration_modifiers] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolated_transliteration_content_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_tr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_y] = {
    .visible = true,
    .named = false,
  },
  [sym_version] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unless] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_until] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__ident_special_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__bareword_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [sym__single_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__double_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__backtick_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__search_slash_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__no_search_slash_plz] = {
    .visible = false,
    .named = true,
  },
  [sym__open_readline_bracket] = {
    .visible = true,
    .named = false,
  },
  [sym__open_fileglob_bracket] = {
    .visible = true,
    .named = false,
  },
  [sym__PERLY_SEMICOLON] = {
    .visible = false,
    .named = true,
  },
  [sym__PERLY_HEREDOC] = {
    .visible = false,
    .named = true,
  },
  [sym__ctrl_z_hack] = {
    .visible = true,
    .named = true,
  },
  [sym__quotelike_begin_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__quotelike_middle_close_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__quotelike_middle_skip] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_end_zw] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_end_quote] = {
    .visible = true,
    .named = false,
  },
  [sym__q_string_content] = {
    .visible = false,
    .named = true,
  },
  [sym__qq_string_content] = {
    .visible = false,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_escaped_delimiter] = {
    .visible = true,
    .named = true,
  },
  [sym__dollar_in_regexp] = {
    .visible = false,
    .named = true,
  },
  [sym_pod] = {
    .visible = true,
    .named = true,
  },
  [sym__gobbled_content] = {
    .visible = false,
    .named = true,
  },
  [sym__attribute_value_begin] = {
    .visible = false,
    .named = true,
  },
  [sym_attribute_value] = {
    .visible = true,
    .named = true,
  },
  [sym_prototype] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_start] = {
    .visible = true,
    .named = false,
  },
  [sym__heredoc_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym__command_heredoc_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym__heredoc_start] = {
    .visible = false,
    .named = true,
  },
  [sym__heredoc_middle] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_end] = {
    .visible = true,
    .named = true,
  },
  [sym__fat_comma_autoquoted] = {
    .visible = false,
    .named = true,
  },
  [sym__filetest] = {
    .visible = true,
    .named = false,
  },
  [sym__brace_autoquoted_token] = {
    .visible = true,
    .named = true,
  },
  [sym__brace_end_zw] = {
    .visible = false,
    .named = true,
  },
  [sym__dollar_ident_zw] = {
    .visible = false,
    .named = true,
  },
  [sym__no_interp_whitespace_zw] = {
    .visible = false,
    .named = true,
  },
  [sym__NONASSOC] = {
    .visible = false,
    .named = true,
  },
  [sym__RECOVER_PAREN_CLOSE] = {
    .visible = true,
    .named = false,
  },
  [sym__RECOVER_BRACKET_CLOSE] = {
    .visible = true,
    .named = false,
  },
  [sym__RECOVER_BRACE_CLOSE] = {
    .visible = true,
    .named = false,
  },
  [sym__RECOVER_ARROW] = {
    .visible = false,
    .named = true,
  },
  [sym__ERROR] = {
    .visible = false,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__PERLY_BRACE_OPEN] = {
    .visible = false,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym__fullstmt] = {
    .visible = false,
    .named = true,
  },
  [sym_statement_label] = {
    .visible = true,
    .named = true,
  },
  [sym__semicolon] = {
    .visible = false,
    .named = true,
  },
  [sym__barestmt] = {
    .visible = false,
    .named = true,
  },
  [sym_package_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_class_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_role_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_class_phaser_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_use_version_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_use_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_mandatory_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_optional_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_named_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_slurpy_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_vars] = {
    .visible = false,
    .named = true,
  },
  [sym_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_subroutine_declaration_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_method_declaration_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_phaser_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__loop_body] = {
    .visible = false,
    .named = true,
  },
  [sym_loop_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_cstyle_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_initializer] = {
    .visible = false,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_defer_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__postfix_expressions] = {
    .visible = false,
    .named = true,
  },
  [sym_postfix_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_postfix_loop_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_postfix_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_yadayada] = {
    .visible = true,
    .named = true,
  },
  [sym__else] = {
    .visible = false,
    .named = true,
  },
  [sym_else] = {
    .visible = true,
    .named = true,
  },
  [sym_elsif] = {
    .visible = true,
    .named = true,
  },
  [sym__expr] = {
    .visible = false,
    .named = true,
  },
  [sym_lowprec_logical_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__listexpr] = {
    .visible = false,
    .named = true,
  },
  [sym__term_rightward] = {
    .visible = true,
    .named = true,
  },
  [sym_subscripted] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_container_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_glob_slot_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_element_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__hash_key] = {
    .visible = false,
    .named = true,
  },
  [sym_hash_element_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_coderef_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_slice_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_slices] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_slice_container_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_keyval_container_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_keyval_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__term] = {
    .visible = false,
    .named = true,
  },
  [sym_readline_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_fileglob_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_equality_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_relational_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preinc_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_postinc_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_refgen_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__tricky_list] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_hash_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_subroutine_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_method_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_do_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_eval_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_await_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__declared_vars] = {
    .visible = false,
    .named = true,
  },
  [sym_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__decl_variable_list] = {
    .visible = false,
    .named = true,
  },
  [sym_localization_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_stub_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_postfix_deref] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_scalar_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_arraylen_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_amper_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_glob_deref_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_version_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_func0op_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_func1op_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_map_grep_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sort_routine] = {
    .visible = false,
    .named = true,
  },
  [sym_sort_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__label_arg] = {
    .visible = false,
    .named = true,
  },
  [sym_loopex_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_return_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_undef_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__listop] = {
    .visible = false,
    .named = true,
  },
  [sym_indirect_object] = {
    .visible = true,
    .named = true,
  },
  [sym_function_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__tricky_indirob_hashref] = {
    .visible = true,
    .named = true,
  },
  [sym_ambiguous_function_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function] = {
    .visible = true,
    .named = true,
  },
  [sym_method_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_method] = {
    .visible = true,
    .named = true,
  },
  [sym__variables] = {
    .visible = false,
    .named = true,
  },
  [sym__signature_varname] = {
    .visible = false,
    .named = true,
  },
  [sym_scalar] = {
    .visible = true,
    .named = true,
  },
  [sym__declare_scalar] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_scalar] = {
    .visible = true,
    .named = true,
  },
  [sym_array] = {
    .visible = true,
    .named = true,
  },
  [sym__declare_array] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_array] = {
    .visible = true,
    .named = true,
  },
  [sym__SUB_AMPER] = {
    .visible = false,
    .named = true,
  },
  [sym__GLOB_STAR] = {
    .visible = false,
    .named = true,
  },
  [sym_hash] = {
    .visible = true,
    .named = true,
  },
  [sym__declare_hash] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_hash] = {
    .visible = true,
    .named = true,
  },
  [sym_arraylen] = {
    .visible = true,
    .named = true,
  },
  [sym_glob] = {
    .visible = true,
    .named = true,
  },
  [sym_amper_sub] = {
    .visible = true,
    .named = true,
  },
  [sym__indirob] = {
    .visible = true,
    .named = true,
  },
  [sym_varname] = {
    .visible = true,
    .named = true,
  },
  [sym__var_indirob_autoquote] = {
    .visible = false,
    .named = true,
  },
  [sym__var_indirob] = {
    .visible = false,
    .named = true,
  },
  [sym_attrlist] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_name] = {
    .visible = true,
    .named = true,
  },
  [sym___DATA__] = {
    .visible = false,
    .named = true,
  },
  [sym__literal] = {
    .visible = false,
    .named = true,
  },
  [sym__apostrophe] = {
    .visible = false,
    .named = true,
  },
  [sym__quotation_mark] = {
    .visible = false,
    .named = true,
  },
  [sym__backtick] = {
    .visible = false,
    .named = true,
  },
  [sym__search_slash] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_begin] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_middle_close] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_end] = {
    .visible = false,
    .named = true,
  },
  [sym__quotelike_middle] = {
    .visible = false,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_interpolated_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym__subscripted_interpolations] = {
    .visible = false,
    .named = true,
  },
  [sym__braced_scalar] = {
    .visible = true,
    .named = true,
  },
  [sym__braced_array] = {
    .visible = true,
    .named = true,
  },
  [sym__interp_arrow] = {
    .visible = false,
    .named = true,
  },
  [sym__array_element_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym__hash_element_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym__slice_expression_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym__scalar_deref_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym__array_deref_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym__interpolations] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__noninterpolated_string_content] = {
    .visible = false,
    .named = false,
  },
  [sym__interpolation_fallbacks] = {
    .visible = false,
    .named = true,
  },
  [sym__nonvar_interpolation_fallbacks] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__interpolated_string_content] = {
    .visible = false,
    .named = false,
  },
  [sym_quoted_word_list] = {
    .visible = true,
    .named = true,
  },
  [sym_command_string] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_regexp] = {
    .visible = true,
    .named = true,
  },
  [sym_match_regexp] = {
    .visible = true,
    .named = true,
  },
  [sym_substitution_regexp] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolated_regexp_content] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolated_transliteration_content] = {
    .visible = false,
    .named = false,
  },
  [sym_transliteration_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_token] = {
    .visible = true,
    .named = true,
  },
  [sym_command_heredoc_token] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_content] = {
    .visible = true,
    .named = true,
  },
  [sym_package] = {
    .visible = true,
    .named = true,
  },
  [sym__version] = {
    .visible = false,
    .named = true,
  },
  [sym__conditionals] = {
    .visible = false,
    .named = true,
  },
  [sym__loops] = {
    .visible = false,
    .named = true,
  },
  [sym_autoquoted_bareword] = {
    .visible = true,
    .named = true,
  },
  [sym__brace_autoquoted] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__ident_special] = {
    .visible = false,
    .named = true,
  },
  [sym_bareword] = {
    .visible = true,
    .named = true,
  },
  [sym_primitive] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_subroutine_declaration_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__for_initializer_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__term_rightward_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__decl_variable_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attrlist_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_heredoc_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_block_statement] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_data_section] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_filehandle] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_filename] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label] = {
    .visible = true,
    .named = true,
  },
  [anon_alias_sym_not_DASHinterpolated] = {
    .visible = true,
    .named = false,
  },
  [alias_sym_regexp_content] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_replacement] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_transliteration_content] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alternative = 1,
  field_arguments = 2,
  field_array = 3,
  field_arrayref = 4,
  field_attributes = 5,
  field_block = 6,
  field_body = 7,
  field_callback = 8,
  field_catch_block = 9,
  field_catch_expr = 10,
  field_condition = 11,
  field_consequent = 12,
  field_content = 13,
  field_continue = 14,
  field_default = 15,
  field_finally_block = 16,
  field_function = 17,
  field_hash = 18,
  field_hashref = 19,
  field_index = 20,
  field_initialiser = 21,
  field_invocant = 22,
  field_iterator = 23,
  field_key = 24,
  field_label = 25,
  field_left = 26,
  field_lexical = 27,
  field_list = 28,
  field_loopex = 29,
  field_method = 30,
  field_modifiers = 31,
  field_module = 32,
  field_name = 33,
  field_operand = 34,
  field_operator = 35,
  field_phase = 36,
  field_right = 37,
  field_statement = 38,
  field_try_block = 39,
  field_value = 40,
  field_variable = 41,
  field_variables = 42,
  field_version = 43,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alternative] = "alternative",
  [field_arguments] = "arguments",
  [field_array] = "array",
  [field_arrayref] = "arrayref",
  [field_attributes] = "attributes",
  [field_block] = "block",
  [field_body] = "body",
  [field_callback] = "callback",
  [field_catch_block] = "catch_block",
  [field_catch_expr] = "catch_expr",
  [field_condition] = "condition",
  [field_consequent] = "consequent",
  [field_content] = "content",
  [field_continue] = "continue",
  [field_default] = "default",
  [field_finally_block] = "finally_block",
  [field_function] = "function",
  [field_hash] = "hash",
  [field_hashref] = "hashref",
  [field_index] = "index",
  [field_initialiser] = "initialiser",
  [field_invocant] = "invocant",
  [field_iterator] = "iterator",
  [field_key] = "key",
  [field_label] = "label",
  [field_left] = "left",
  [field_lexical] = "lexical",
  [field_list] = "list",
  [field_loopex] = "loopex",
  [field_method] = "method",
  [field_modifiers] = "modifiers",
  [field_module] = "module",
  [field_name] = "name",
  [field_operand] = "operand",
  [field_operator] = "operator",
  [field_phase] = "phase",
  [field_right] = "right",
  [field_statement] = "statement",
  [field_try_block] = "try_block",
  [field_value] = "value",
  [field_variable] = "variable",
  [field_variables] = "variables",
  [field_version] = "version",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [2] = {.index = 1, .length = 1},
  [5] = {.index = 2, .length = 2},
  [6] = {.index = 4, .length = 2},
  [7] = {.index = 6, .length = 4},
  [8] = {.index = 10, .length = 1},
  [9] = {.index = 11, .length = 2},
  [10] = {.index = 13, .length = 1},
  [11] = {.index = 14, .length = 1},
  [12] = {.index = 15, .length = 1},
  [13] = {.index = 16, .length = 1},
  [14] = {.index = 17, .length = 1},
  [15] = {.index = 18, .length = 1},
  [16] = {.index = 19, .length = 1},
  [17] = {.index = 20, .length = 2},
  [19] = {.index = 22, .length = 1},
  [21] = {.index = 23, .length = 1},
  [23] = {.index = 24, .length = 1},
  [24] = {.index = 25, .length = 2},
  [25] = {.index = 0, .length = 1},
  [26] = {.index = 27, .length = 2},
  [27] = {.index = 29, .length = 2},
  [28] = {.index = 31, .length = 3},
  [29] = {.index = 34, .length = 1},
  [33] = {.index = 35, .length = 1},
  [34] = {.index = 36, .length = 1},
  [35] = {.index = 37, .length = 2},
  [36] = {.index = 39, .length = 2},
  [37] = {.index = 41, .length = 1},
  [39] = {.index = 42, .length = 1},
  [40] = {.index = 43, .length = 3},
  [41] = {.index = 46, .length = 1},
  [42] = {.index = 47, .length = 1},
  [44] = {.index = 48, .length = 2},
  [45] = {.index = 50, .length = 2},
  [46] = {.index = 52, .length = 1},
  [47] = {.index = 53, .length = 1},
  [48] = {.index = 52, .length = 1},
  [49] = {.index = 54, .length = 2},
  [51] = {.index = 56, .length = 1},
  [52] = {.index = 57, .length = 2},
  [54] = {.index = 59, .length = 2},
  [55] = {.index = 61, .length = 3},
  [56] = {.index = 64, .length = 2},
  [57] = {.index = 66, .length = 2},
  [58] = {.index = 68, .length = 1},
  [59] = {.index = 69, .length = 2},
  [60] = {.index = 71, .length = 2},
  [61] = {.index = 73, .length = 2},
  [62] = {.index = 75, .length = 2},
  [63] = {.index = 77, .length = 2},
  [64] = {.index = 79, .length = 2},
  [65] = {.index = 81, .length = 2},
  [66] = {.index = 83, .length = 1},
  [67] = {.index = 84, .length = 1},
  [68] = {.index = 83, .length = 1},
  [69] = {.index = 85, .length = 2},
  [70] = {.index = 87, .length = 2},
  [71] = {.index = 0, .length = 1},
  [72] = {.index = 89, .length = 1},
  [73] = {.index = 90, .length = 1},
  [74] = {.index = 91, .length = 1},
  [75] = {.index = 92, .length = 2},
  [76] = {.index = 94, .length = 2},
  [77] = {.index = 96, .length = 1},
  [78] = {.index = 97, .length = 1},
  [79] = {.index = 98, .length = 2},
  [80] = {.index = 100, .length = 3},
  [81] = {.index = 103, .length = 2},
  [82] = {.index = 105, .length = 1},
  [83] = {.index = 106, .length = 2},
  [84] = {.index = 94, .length = 2},
  [85] = {.index = 92, .length = 2},
  [86] = {.index = 97, .length = 1},
  [87] = {.index = 96, .length = 1},
  [88] = {.index = 108, .length = 2},
  [89] = {.index = 110, .length = 1},
  [91] = {.index = 111, .length = 3},
  [92] = {.index = 114, .length = 3},
  [93] = {.index = 117, .length = 2},
  [94] = {.index = 119, .length = 2},
  [95] = {.index = 121, .length = 3},
  [96] = {.index = 124, .length = 1},
  [97] = {.index = 125, .length = 3},
  [98] = {.index = 128, .length = 1},
  [99] = {.index = 129, .length = 3},
  [100] = {.index = 132, .length = 2},
  [101] = {.index = 134, .length = 3},
  [102] = {.index = 137, .length = 2},
  [103] = {.index = 42, .length = 1},
  [104] = {.index = 87, .length = 2},
  [105] = {.index = 139, .length = 2},
  [106] = {.index = 141, .length = 1},
  [107] = {.index = 142, .length = 1},
  [108] = {.index = 143, .length = 1},
  [109] = {.index = 144, .length = 3},
  [110] = {.index = 147, .length = 2},
  [111] = {.index = 149, .length = 3},
  [112] = {.index = 152, .length = 1},
  [113] = {.index = 153, .length = 2},
  [114] = {.index = 155, .length = 2},
  [115] = {.index = 157, .length = 3},
  [116] = {.index = 160, .length = 1},
  [117] = {.index = 161, .length = 3},
  [118] = {.index = 164, .length = 4},
  [119] = {.index = 168, .length = 3},
  [120] = {.index = 171, .length = 3},
  [121] = {.index = 174, .length = 3},
  [122] = {.index = 177, .length = 2},
  [123] = {.index = 179, .length = 4},
  [124] = {.index = 183, .length = 4},
  [125] = {.index = 187, .length = 5},
  [126] = {.index = 192, .length = 2},
  [127] = {.index = 194, .length = 3},
  [128] = {.index = 87, .length = 2},
  [129] = {.index = 194, .length = 3},
  [130] = {.index = 87, .length = 2},
  [131] = {.index = 197, .length = 2},
  [132] = {.index = 199, .length = 4},
  [133] = {.index = 203, .length = 1},
  [134] = {.index = 204, .length = 3},
  [135] = {.index = 207, .length = 2},
  [136] = {.index = 209, .length = 2},
  [137] = {.index = 211, .length = 3},
  [138] = {.index = 214, .length = 2},
  [139] = {.index = 216, .length = 4},
  [140] = {.index = 220, .length = 3},
  [141] = {.index = 223, .length = 4},
  [142] = {.index = 227, .length = 3},
  [143] = {.index = 230, .length = 2},
  [144] = {.index = 232, .length = 5},
  [145] = {.index = 237, .length = 2},
  [146] = {.index = 239, .length = 3},
  [147] = {.index = 239, .length = 3},
  [148] = {.index = 242, .length = 3},
  [149] = {.index = 245, .length = 3},
  [150] = {.index = 248, .length = 4},
  [151] = {.index = 252, .length = 4},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_function, 0},
  [1] =
    {field_loopex, 0},
  [2] =
    {field_array, 0, .inherited = true},
    {field_index, 0, .inherited = true},
  [4] =
    {field_hash, 0, .inherited = true},
    {field_key, 0, .inherited = true},
  [6] =
    {field_array, 0, .inherited = true},
    {field_arrayref, 0, .inherited = true},
    {field_hash, 0, .inherited = true},
    {field_hashref, 0, .inherited = true},
  [10] =
    {field_arrayref, 0, .inherited = true},
  [11] =
    {field_operator, 0},
    {field_operator, 1},
  [13] =
    {field_phase, 0},
  [14] =
    {field_variable, 1},
  [15] =
    {field_variables, 1},
  [16] =
    {field_body, 1},
  [17] =
    {field_try_block, 1},
  [18] =
    {field_block, 1},
  [19] =
    {field_modifiers, 1},
  [20] =
    {field_operand, 1},
    {field_operator, 0},
  [22] =
    {field_version, 1},
  [23] =
    {field_list, 1},
  [24] =
    {field_variable, 0},
  [25] =
    {field_operand, 0},
    {field_operator, 1},
  [27] =
    {field_arguments, 1},
    {field_function, 0},
  [29] =
    {field_operator, 0},
    {field_operator, 2},
  [31] =
    {field_content, 1},
    {field_operator, 0},
    {field_operator, 2},
  [34] =
    {field_name, 1},
  [35] =
    {field_name, 0},
  [36] =
    {field_body, 2},
  [37] =
    {field_body, 2},
    {field_name, 1},
  [39] =
    {field_callback, 1},
    {field_list, 2},
  [41] =
    {field_module, 1},
  [42] =
    {field_operator, 0},
  [43] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [46] =
    {field_list, 2},
  [47] =
    {field_condition, 2},
  [48] =
    {field_invocant, 0},
    {field_method, 2},
  [50] =
    {field_arguments, 2},
    {field_function, 0},
  [52] =
    {field_content, 1},
  [53] =
    {field_modifiers, 2},
  [54] =
    {field_label, 0},
    {field_statement, 2},
  [56] =
    {field_arrayref, 0},
  [57] =
    {field_name, 1},
    {field_version, 2},
  [59] =
    {field_attributes, 2},
    {field_phase, 0},
  [61] =
    {field_body, 3},
    {field_lexical, 0},
    {field_name, 2},
  [64] =
    {field_attributes, 3},
    {field_variable, 1},
  [66] =
    {field_attributes, 3},
    {field_variables, 1},
  [68] =
    {field_body, 3},
  [69] =
    {field_attributes, 2},
    {field_body, 3},
  [71] =
    {field_body, 3},
    {field_name, 1},
  [73] =
    {field_catch_block, 3},
    {field_try_block, 1},
  [75] =
    {field_finally_block, 3},
    {field_try_block, 1},
  [77] =
    {field_callback, 1},
    {field_list, 3},
  [79] =
    {field_module, 1},
    {field_version, 2},
  [81] =
    {field_variables, 1},
    {field_variables, 2},
  [83] =
    {field_content, 2},
  [84] =
    {field_modifiers, 3},
  [85] =
    {field_modifiers, 3},
    {field_operator, 0},
  [87] =
    {field_content, 2},
    {field_operator, 0},
  [89] =
    {field_key, 2},
  [90] =
    {field_arguments, 2},
  [91] =
    {field_index, 2},
  [92] =
    {field_hash, 0},
    {field_key, 2},
  [94] =
    {field_array, 0},
    {field_index, 2},
  [96] =
    {field_hash, 0},
  [97] =
    {field_array, 0},
  [98] =
    {field_invocant, 0},
    {field_method, 3},
  [100] =
    {field_arguments, 1},
    {field_arguments, 3},
    {field_function, 0},
  [103] =
    {field_content, 1},
    {field_modifiers, 3},
  [105] =
    {field_list, 0},
  [106] =
    {field_body, 3},
    {field_name, 2},
  [108] =
    {field_attributes, 3},
    {field_name, 1},
  [110] =
    {field_default, 2},
  [111] =
    {field_body, 4},
    {field_lexical, 0},
    {field_name, 2},
  [114] =
    {field_body, 4},
    {field_lexical, 0},
    {field_name, 3},
  [117] =
    {field_attributes, 2},
    {field_body, 4},
  [119] =
    {field_body, 4},
    {field_name, 1},
  [121] =
    {field_attributes, 3},
    {field_body, 4},
    {field_name, 1},
  [124] =
    {field_list, 3},
  [125] =
    {field_variables, 1},
    {field_variables, 2},
    {field_variables, 3},
  [128] =
    {field_block, 0},
  [129] =
    {field_block, 4, .inherited = true},
    {field_continue, 4, .inherited = true},
    {field_list, 2},
  [132] =
    {field_content, 2},
    {field_modifiers, 4},
  [134] =
    {field_content, 2},
    {field_modifiers, 4},
    {field_operator, 0},
  [137] =
    {field_modifiers, 4},
    {field_operator, 0},
  [139] =
    {field_arguments, 3},
    {field_function, 0},
  [141] =
    {field_key, 3},
  [142] =
    {field_arguments, 3},
  [143] =
    {field_index, 3},
  [144] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequent, 2},
  [147] =
    {field_block, 4},
    {field_condition, 2},
  [149] =
    {field_block, 4, .inherited = true},
    {field_condition, 2},
    {field_continue, 4, .inherited = true},
  [152] =
    {field_body, 4},
  [153] =
    {field_attributes, 3},
    {field_body, 4},
  [155] =
    {field_body, 4},
    {field_name, 2},
  [157] =
    {field_attributes, 4},
    {field_name, 1},
    {field_version, 2},
  [160] =
    {field_default, 3},
  [161] =
    {field_body, 5},
    {field_lexical, 0},
    {field_name, 2},
  [164] =
    {field_attributes, 4},
    {field_body, 5},
    {field_lexical, 0},
    {field_name, 2},
  [168] =
    {field_body, 5},
    {field_lexical, 0},
    {field_name, 3},
  [171] =
    {field_attributes, 3},
    {field_body, 5},
    {field_name, 1},
  [174] =
    {field_catch_block, 3},
    {field_finally_block, 5},
    {field_try_block, 1},
  [177] =
    {field_callback, 3},
    {field_list, 4},
  [179] =
    {field_variables, 1},
    {field_variables, 2},
    {field_variables, 3},
    {field_variables, 4},
  [183] =
    {field_block, 5, .inherited = true},
    {field_condition, 3},
    {field_continue, 5, .inherited = true},
    {field_initialiser, 2},
  [187] =
    {field_block, 5, .inherited = true},
    {field_continue, 5, .inherited = true},
    {field_list, 3},
    {field_variable, 1, .inherited = true},
    {field_variables, 1, .inherited = true},
  [192] =
    {field_modifiers, 5},
    {field_operator, 0},
  [194] =
    {field_content, 2},
    {field_modifiers, 5},
    {field_operator, 0},
  [197] =
    {field_arguments, 4},
    {field_function, 0},
  [199] =
    {field_left, 0},
    {field_operator, 1},
    {field_operator, 3},
    {field_right, 2},
  [203] =
    {field_hashref, 0},
  [204] =
    {field_arguments, 4},
    {field_invocant, 0},
    {field_method, 2},
  [207] =
    {field_attributes, 3},
    {field_body, 5},
  [209] =
    {field_body, 5},
    {field_name, 2},
  [211] =
    {field_attributes, 4},
    {field_body, 5},
    {field_name, 2},
  [214] =
    {field_name, 0},
    {field_value, 3},
  [216] =
    {field_attributes, 4},
    {field_body, 6},
    {field_lexical, 0},
    {field_name, 2},
  [220] =
    {field_body, 6},
    {field_lexical, 0},
    {field_name, 3},
  [223] =
    {field_attributes, 5},
    {field_body, 6},
    {field_lexical, 0},
    {field_name, 3},
  [227] =
    {field_catch_block, 6},
    {field_catch_expr, 4},
    {field_try_block, 1},
  [230] =
    {field_callback, 3},
    {field_list, 5},
  [232] =
    {field_block, 6, .inherited = true},
    {field_condition, 3},
    {field_continue, 6, .inherited = true},
    {field_initialiser, 2},
    {field_iterator, 4},
  [237] =
    {field_block, 0},
    {field_continue, 2},
  [239] =
    {field_content, 2},
    {field_modifiers, 6},
    {field_operator, 0},
  [242] =
    {field_arguments, 5},
    {field_invocant, 0},
    {field_method, 3},
  [245] =
    {field_attributes, 4},
    {field_body, 6},
    {field_name, 2},
  [248] =
    {field_attributes, 5},
    {field_body, 7},
    {field_lexical, 0},
    {field_name, 3},
  [252] =
    {field_catch_block, 6},
    {field_catch_expr, 4},
    {field_finally_block, 8},
    {field_try_block, 1},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [3] = {
    [0] = alias_sym_block_statement,
  },
  [4] = {
    [0] = anon_alias_sym_not_DASHinterpolated,
  },
  [18] = {
    [1] = alias_sym_filename,
  },
  [20] = {
    [0] = sym_function,
  },
  [22] = {
    [0] = alias_sym_label,
  },
  [25] = {
    [1] = sym_indirect_object,
  },
  [27] = {
    [1] = alias_sym_filehandle,
  },
  [28] = {
    [1] = alias_sym_string_content,
    [2] = anon_sym_GT,
  },
  [30] = {
    [0] = sym_scalar,
  },
  [31] = {
    [0] = sym_array,
  },
  [32] = {
    [0] = sym_hash,
  },
  [38] = {
    [2] = alias_sym_data_section,
  },
  [43] = {
    [0] = sym_autoquoted_bareword,
  },
  [46] = {
    [1] = alias_sym_string_content,
  },
  [48] = {
    [1] = alias_sym_regexp_content,
  },
  [50] = {
    [1] = anon_alias_sym_not_DASHinterpolated,
  },
  [53] = {
    [0] = sym_varname,
  },
  [66] = {
    [2] = alias_sym_string_content,
  },
  [68] = {
    [2] = alias_sym_regexp_content,
  },
  [70] = {
    [2] = alias_sym_regexp_content,
  },
  [71] = {
    [0] = sym_function,
  },
  [81] = {
    [1] = alias_sym_regexp_content,
  },
  [84] = {
    [0] = sym_container_variable,
  },
  [85] = {
    [0] = sym_container_variable,
  },
  [86] = {
    [0] = sym_slice_container_variable,
  },
  [87] = {
    [0] = sym_slice_container_variable,
  },
  [89] = {
    [0] = sym_scalar,
  },
  [90] = {
    [1] = sym_varname,
  },
  [100] = {
    [2] = alias_sym_regexp_content,
  },
  [101] = {
    [2] = alias_sym_regexp_content,
  },
  [103] = {
    [3] = alias_sym_replacement,
  },
  [104] = {
    [2] = alias_sym_transliteration_content,
  },
  [105] = {
    [0] = sym_function,
  },
  [126] = {
    [3] = alias_sym_replacement,
  },
  [127] = {
    [2] = alias_sym_regexp_content,
  },
  [128] = {
    [2] = alias_sym_regexp_content,
    [4] = alias_sym_replacement,
  },
  [129] = {
    [2] = alias_sym_transliteration_content,
  },
  [130] = {
    [2] = alias_sym_transliteration_content,
    [4] = alias_sym_replacement,
  },
  [131] = {
    [0] = sym_function,
  },
  [146] = {
    [2] = alias_sym_regexp_content,
    [4] = alias_sym_replacement,
  },
  [147] = {
    [2] = alias_sym_transliteration_content,
    [4] = alias_sym_replacement,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_block, 3,
    sym_block,
    alias_sym_block_statement,
    sym_indirect_object,
  sym__term, 2,
    sym__term,
    alias_sym_filename,
  sym_scalar, 3,
    sym_scalar,
    alias_sym_filehandle,
    sym_container_variable,
  sym_array, 2,
    sym_array,
    sym_slice_container_variable,
  sym__quotelike_end, 2,
    sym__quotelike_end,
    anon_sym_GT,
  aux_sym__noninterpolated_string_content, 5,
    aux_sym__noninterpolated_string_content,
    alias_sym_regexp_content,
    alias_sym_replacement,
    alias_sym_string_content,
    alias_sym_transliteration_content,
  aux_sym__interpolated_string_content, 3,
    aux_sym__interpolated_string_content,
    alias_sym_replacement,
    alias_sym_string_content,
  aux_sym__interpolated_regexp_content, 2,
    aux_sym__interpolated_regexp_content,
    alias_sym_regexp_content,
  aux_sym__interpolated_transliteration_content, 3,
    aux_sym__interpolated_transliteration_content,
    alias_sym_replacement,
    alias_sym_transliteration_content,
  sym_identifier, 2,
    sym_identifier,
    alias_sym_label,
  sym__ident_special, 2,
    sym__ident_special,
    sym_varname,
  sym_bareword, 2,
    sym_bareword,
    alias_sym_filehandle,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 2,
  [8] = 3,
  [9] = 5,
  [10] = 4,
  [11] = 6,
  [12] = 3,
  [13] = 2,
  [14] = 3,
  [15] = 3,
  [16] = 3,
  [17] = 2,
  [18] = 2,
  [19] = 2,
  [20] = 6,
  [21] = 4,
  [22] = 5,
  [23] = 5,
  [24] = 6,
  [25] = 5,
  [26] = 4,
  [27] = 5,
  [28] = 4,
  [29] = 4,
  [30] = 6,
  [31] = 6,
  [32] = 2,
  [33] = 3,
  [34] = 34,
  [35] = 3,
  [36] = 2,
  [37] = 3,
  [38] = 2,
  [39] = 39,
  [40] = 3,
  [41] = 2,
  [42] = 39,
  [43] = 43,
  [44] = 39,
  [45] = 39,
  [46] = 39,
  [47] = 39,
  [48] = 39,
  [49] = 39,
  [50] = 39,
  [51] = 39,
  [52] = 39,
  [53] = 39,
  [54] = 39,
  [55] = 39,
  [56] = 39,
  [57] = 39,
  [58] = 39,
  [59] = 39,
  [60] = 39,
  [61] = 39,
  [62] = 39,
  [63] = 39,
  [64] = 39,
  [65] = 39,
  [66] = 39,
  [67] = 39,
  [68] = 39,
  [69] = 39,
  [70] = 39,
  [71] = 39,
  [72] = 39,
  [73] = 39,
  [74] = 39,
  [75] = 39,
  [76] = 39,
  [77] = 39,
  [78] = 3,
  [79] = 2,
  [80] = 39,
  [81] = 6,
  [82] = 4,
  [83] = 5,
  [84] = 5,
  [85] = 6,
  [86] = 3,
  [87] = 2,
  [88] = 3,
  [89] = 2,
  [90] = 4,
  [91] = 4,
  [92] = 6,
  [93] = 5,
  [94] = 4,
  [95] = 5,
  [96] = 3,
  [97] = 5,
  [98] = 2,
  [99] = 4,
  [100] = 6,
  [101] = 101,
  [102] = 102,
  [103] = 101,
  [104] = 102,
  [105] = 101,
  [106] = 4,
  [107] = 4,
  [108] = 102,
  [109] = 101,
  [110] = 102,
  [111] = 101,
  [112] = 102,
  [113] = 101,
  [114] = 102,
  [115] = 101,
  [116] = 102,
  [117] = 102,
  [118] = 102,
  [119] = 101,
  [120] = 102,
  [121] = 102,
  [122] = 102,
  [123] = 102,
  [124] = 102,
  [125] = 102,
  [126] = 102,
  [127] = 102,
  [128] = 102,
  [129] = 102,
  [130] = 102,
  [131] = 102,
  [132] = 5,
  [133] = 5,
  [134] = 101,
  [135] = 135,
  [136] = 135,
  [137] = 137,
  [138] = 135,
  [139] = 137,
  [140] = 137,
  [141] = 135,
  [142] = 43,
  [143] = 135,
  [144] = 137,
  [145] = 135,
  [146] = 135,
  [147] = 137,
  [148] = 135,
  [149] = 137,
  [150] = 137,
  [151] = 135,
  [152] = 135,
  [153] = 137,
  [154] = 135,
  [155] = 137,
  [156] = 135,
  [157] = 137,
  [158] = 137,
  [159] = 4,
  [160] = 137,
  [161] = 135,
  [162] = 137,
  [163] = 137,
  [164] = 135,
  [165] = 135,
  [166] = 137,
  [167] = 135,
  [168] = 137,
  [169] = 135,
  [170] = 137,
  [171] = 135,
  [172] = 135,
  [173] = 135,
  [174] = 135,
  [175] = 135,
  [176] = 135,
  [177] = 135,
  [178] = 135,
  [179] = 135,
  [180] = 135,
  [181] = 135,
  [182] = 135,
  [183] = 135,
  [184] = 135,
  [185] = 135,
  [186] = 135,
  [187] = 135,
  [188] = 135,
  [189] = 135,
  [190] = 135,
  [191] = 135,
  [192] = 135,
  [193] = 135,
  [194] = 135,
  [195] = 135,
  [196] = 135,
  [197] = 135,
  [198] = 135,
  [199] = 135,
  [200] = 135,
  [201] = 135,
  [202] = 137,
  [203] = 5,
  [204] = 135,
  [205] = 137,
  [206] = 137,
  [207] = 207,
  [208] = 207,
  [209] = 209,
  [210] = 209,
  [211] = 211,
  [212] = 211,
  [213] = 211,
  [214] = 209,
  [215] = 6,
  [216] = 6,
  [217] = 217,
  [218] = 217,
  [219] = 217,
  [220] = 217,
  [221] = 217,
  [222] = 217,
  [223] = 6,
  [224] = 217,
  [225] = 217,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 6,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 226,
  [241] = 241,
  [242] = 227,
  [243] = 234,
  [244] = 6,
  [245] = 234,
  [246] = 231,
  [247] = 237,
  [248] = 228,
  [249] = 230,
  [250] = 232,
  [251] = 233,
  [252] = 235,
  [253] = 236,
  [254] = 238,
  [255] = 239,
  [256] = 226,
  [257] = 227,
  [258] = 258,
  [259] = 258,
  [260] = 6,
  [261] = 231,
  [262] = 237,
  [263] = 228,
  [264] = 230,
  [265] = 232,
  [266] = 233,
  [267] = 235,
  [268] = 236,
  [269] = 238,
  [270] = 239,
  [271] = 226,
  [272] = 241,
  [273] = 227,
  [274] = 258,
  [275] = 231,
  [276] = 237,
  [277] = 228,
  [278] = 230,
  [279] = 232,
  [280] = 233,
  [281] = 235,
  [282] = 236,
  [283] = 238,
  [284] = 239,
  [285] = 226,
  [286] = 241,
  [287] = 227,
  [288] = 258,
  [289] = 231,
  [290] = 237,
  [291] = 228,
  [292] = 230,
  [293] = 232,
  [294] = 238,
  [295] = 226,
  [296] = 241,
  [297] = 227,
  [298] = 231,
  [299] = 237,
  [300] = 228,
  [301] = 230,
  [302] = 232,
  [303] = 238,
  [304] = 226,
  [305] = 241,
  [306] = 227,
  [307] = 231,
  [308] = 237,
  [309] = 228,
  [310] = 230,
  [311] = 232,
  [312] = 238,
  [313] = 241,
  [314] = 227,
  [315] = 231,
  [316] = 237,
  [317] = 228,
  [318] = 230,
  [319] = 232,
  [320] = 238,
  [321] = 226,
  [322] = 241,
  [323] = 241,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 329,
  [330] = 328,
  [331] = 331,
  [332] = 325,
  [333] = 324,
  [334] = 328,
  [335] = 329,
  [336] = 336,
  [337] = 325,
  [338] = 336,
  [339] = 336,
  [340] = 326,
  [341] = 327,
  [342] = 328,
  [343] = 329,
  [344] = 336,
  [345] = 326,
  [346] = 325,
  [347] = 329,
  [348] = 324,
  [349] = 329,
  [350] = 336,
  [351] = 329,
  [352] = 324,
  [353] = 325,
  [354] = 326,
  [355] = 326,
  [356] = 327,
  [357] = 326,
  [358] = 328,
  [359] = 329,
  [360] = 324,
  [361] = 325,
  [362] = 324,
  [363] = 324,
  [364] = 325,
  [365] = 327,
  [366] = 331,
  [367] = 336,
  [368] = 327,
  [369] = 326,
  [370] = 327,
  [371] = 328,
  [372] = 329,
  [373] = 336,
  [374] = 324,
  [375] = 328,
  [376] = 336,
  [377] = 327,
  [378] = 326,
  [379] = 327,
  [380] = 328,
  [381] = 331,
  [382] = 331,
  [383] = 331,
  [384] = 331,
  [385] = 331,
  [386] = 331,
  [387] = 325,
  [388] = 388,
  [389] = 389,
  [390] = 390,
  [391] = 389,
  [392] = 392,
  [393] = 389,
  [394] = 392,
  [395] = 388,
  [396] = 388,
  [397] = 388,
  [398] = 398,
  [399] = 398,
  [400] = 388,
  [401] = 398,
  [402] = 398,
  [403] = 389,
  [404] = 388,
  [405] = 398,
  [406] = 388,
  [407] = 388,
  [408] = 389,
  [409] = 389,
  [410] = 388,
  [411] = 388,
  [412] = 392,
  [413] = 390,
  [414] = 388,
  [415] = 398,
  [416] = 388,
  [417] = 388,
  [418] = 390,
  [419] = 398,
  [420] = 389,
  [421] = 398,
  [422] = 388,
  [423] = 389,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 6,
  [431] = 431,
  [432] = 432,
  [433] = 424,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 425,
  [438] = 425,
  [439] = 439,
  [440] = 440,
  [441] = 425,
  [442] = 442,
  [443] = 439,
  [444] = 440,
  [445] = 425,
  [446] = 425,
  [447] = 425,
  [448] = 439,
  [449] = 440,
  [450] = 425,
  [451] = 425,
  [452] = 439,
  [453] = 440,
  [454] = 454,
  [455] = 455,
  [456] = 425,
  [457] = 457,
  [458] = 458,
  [459] = 439,
  [460] = 440,
  [461] = 461,
  [462] = 425,
  [463] = 463,
  [464] = 464,
  [465] = 431,
  [466] = 424,
  [467] = 436,
  [468] = 442,
  [469] = 454,
  [470] = 458,
  [471] = 461,
  [472] = 463,
  [473] = 464,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 428,
  [479] = 429,
  [480] = 432,
  [481] = 434,
  [482] = 425,
  [483] = 474,
  [484] = 431,
  [485] = 424,
  [486] = 436,
  [487] = 442,
  [488] = 454,
  [489] = 458,
  [490] = 461,
  [491] = 425,
  [492] = 463,
  [493] = 464,
  [494] = 474,
  [495] = 475,
  [496] = 476,
  [497] = 477,
  [498] = 498,
  [499] = 428,
  [500] = 439,
  [501] = 429,
  [502] = 432,
  [503] = 434,
  [504] = 440,
  [505] = 431,
  [506] = 424,
  [507] = 436,
  [508] = 442,
  [509] = 454,
  [510] = 458,
  [511] = 477,
  [512] = 463,
  [513] = 464,
  [514] = 474,
  [515] = 475,
  [516] = 476,
  [517] = 477,
  [518] = 428,
  [519] = 429,
  [520] = 432,
  [521] = 434,
  [522] = 431,
  [523] = 424,
  [524] = 436,
  [525] = 442,
  [526] = 454,
  [527] = 464,
  [528] = 474,
  [529] = 475,
  [530] = 477,
  [531] = 428,
  [532] = 532,
  [533] = 429,
  [534] = 432,
  [535] = 431,
  [536] = 424,
  [537] = 436,
  [538] = 442,
  [539] = 454,
  [540] = 464,
  [541] = 474,
  [542] = 475,
  [543] = 477,
  [544] = 428,
  [545] = 429,
  [546] = 425,
  [547] = 432,
  [548] = 475,
  [549] = 431,
  [550] = 550,
  [551] = 476,
  [552] = 436,
  [553] = 442,
  [554] = 454,
  [555] = 464,
  [556] = 474,
  [557] = 475,
  [558] = 477,
  [559] = 428,
  [560] = 429,
  [561] = 6,
  [562] = 432,
  [563] = 431,
  [564] = 424,
  [565] = 436,
  [566] = 442,
  [567] = 454,
  [568] = 464,
  [569] = 474,
  [570] = 475,
  [571] = 477,
  [572] = 428,
  [573] = 429,
  [574] = 432,
  [575] = 455,
  [576] = 457,
  [577] = 498,
  [578] = 426,
  [579] = 435,
  [580] = 455,
  [581] = 457,
  [582] = 498,
  [583] = 426,
  [584] = 435,
  [585] = 461,
  [586] = 586,
  [587] = 587,
  [588] = 588,
  [589] = 589,
  [590] = 590,
  [591] = 591,
  [592] = 592,
  [593] = 590,
  [594] = 594,
  [595] = 595,
  [596] = 589,
  [597] = 597,
  [598] = 588,
  [599] = 599,
  [600] = 594,
  [601] = 601,
  [602] = 592,
  [603] = 599,
  [604] = 592,
  [605] = 588,
  [606] = 599,
  [607] = 595,
  [608] = 595,
  [609] = 609,
  [610] = 589,
  [611] = 609,
  [612] = 592,
  [613] = 589,
  [614] = 594,
  [615] = 591,
  [616] = 4,
  [617] = 588,
  [618] = 590,
  [619] = 591,
  [620] = 594,
  [621] = 588,
  [622] = 594,
  [623] = 599,
  [624] = 601,
  [625] = 595,
  [626] = 592,
  [627] = 609,
  [628] = 588,
  [629] = 589,
  [630] = 594,
  [631] = 599,
  [632] = 595,
  [633] = 609,
  [634] = 589,
  [635] = 599,
  [636] = 592,
  [637] = 588,
  [638] = 590,
  [639] = 591,
  [640] = 594,
  [641] = 599,
  [642] = 595,
  [643] = 588,
  [644] = 601,
  [645] = 599,
  [646] = 609,
  [647] = 609,
  [648] = 589,
  [649] = 592,
  [650] = 588,
  [651] = 599,
  [652] = 609,
  [653] = 601,
  [654] = 590,
  [655] = 591,
  [656] = 595,
  [657] = 592,
  [658] = 594,
  [659] = 588,
  [660] = 594,
  [661] = 599,
  [662] = 601,
  [663] = 595,
  [664] = 586,
  [665] = 589,
  [666] = 609,
  [667] = 590,
  [668] = 591,
  [669] = 588,
  [670] = 589,
  [671] = 595,
  [672] = 609,
  [673] = 609,
  [674] = 589,
  [675] = 599,
  [676] = 594,
  [677] = 595,
  [678] = 590,
  [679] = 591,
  [680] = 601,
  [681] = 592,
  [682] = 588,
  [683] = 594,
  [684] = 601,
  [685] = 599,
  [686] = 588,
  [687] = 599,
  [688] = 595,
  [689] = 609,
  [690] = 589,
  [691] = 592,
  [692] = 595,
  [693] = 609,
  [694] = 590,
  [695] = 591,
  [696] = 589,
  [697] = 609,
  [698] = 589,
  [699] = 601,
  [700] = 592,
  [701] = 592,
  [702] = 589,
  [703] = 594,
  [704] = 588,
  [705] = 594,
  [706] = 599,
  [707] = 594,
  [708] = 592,
  [709] = 595,
  [710] = 595,
  [711] = 609,
  [712] = 592,
  [713] = 713,
  [714] = 714,
  [715] = 714,
  [716] = 714,
  [717] = 714,
  [718] = 714,
  [719] = 713,
  [720] = 720,
  [721] = 721,
  [722] = 722,
  [723] = 723,
  [724] = 713,
  [725] = 720,
  [726] = 721,
  [727] = 721,
  [728] = 720,
  [729] = 721,
  [730] = 722,
  [731] = 723,
  [732] = 713,
  [733] = 733,
  [734] = 722,
  [735] = 713,
  [736] = 720,
  [737] = 721,
  [738] = 713,
  [739] = 720,
  [740] = 721,
  [741] = 713,
  [742] = 720,
  [743] = 722,
  [744] = 723,
  [745] = 714,
  [746] = 721,
  [747] = 723,
  [748] = 713,
  [749] = 720,
  [750] = 721,
  [751] = 722,
  [752] = 723,
  [753] = 714,
  [754] = 722,
  [755] = 723,
  [756] = 722,
  [757] = 722,
  [758] = 723,
  [759] = 722,
  [760] = 723,
  [761] = 720,
  [762] = 720,
  [763] = 721,
  [764] = 723,
  [765] = 720,
  [766] = 721,
  [767] = 587,
  [768] = 722,
  [769] = 723,
  [770] = 713,
  [771] = 720,
  [772] = 721,
  [773] = 714,
  [774] = 713,
  [775] = 720,
  [776] = 721,
  [777] = 714,
  [778] = 713,
  [779] = 720,
  [780] = 721,
  [781] = 722,
  [782] = 723,
  [783] = 722,
  [784] = 723,
  [785] = 714,
  [786] = 714,
  [787] = 714,
  [788] = 722,
  [789] = 723,
  [790] = 713,
  [791] = 720,
  [792] = 721,
  [793] = 714,
  [794] = 733,
  [795] = 722,
  [796] = 723,
  [797] = 733,
  [798] = 733,
  [799] = 733,
  [800] = 733,
  [801] = 733,
  [802] = 733,
  [803] = 714,
  [804] = 713,
  [805] = 713,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 810,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 806,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 810,
  [820] = 820,
  [821] = 813,
  [822] = 822,
  [823] = 814,
  [824] = 824,
  [825] = 825,
  [826] = 806,
  [827] = 816,
  [828] = 808,
  [829] = 818,
  [830] = 809,
  [831] = 820,
  [832] = 817,
  [833] = 822,
  [834] = 825,
  [835] = 835,
  [836] = 808,
  [837] = 837,
  [838] = 809,
  [839] = 817,
  [840] = 840,
  [841] = 835,
  [842] = 837,
  [843] = 843,
  [844] = 843,
  [845] = 825,
  [846] = 835,
  [847] = 837,
  [848] = 843,
  [849] = 813,
  [850] = 814,
  [851] = 806,
  [852] = 852,
  [853] = 853,
  [854] = 852,
  [855] = 810,
  [856] = 856,
  [857] = 857,
  [858] = 858,
  [859] = 812,
  [860] = 818,
  [861] = 810,
  [862] = 862,
  [863] = 807,
  [864] = 813,
  [865] = 852,
  [866] = 816,
  [867] = 852,
  [868] = 812,
  [869] = 856,
  [870] = 856,
  [871] = 810,
  [872] = 813,
  [873] = 814,
  [874] = 806,
  [875] = 816,
  [876] = 818,
  [877] = 820,
  [878] = 822,
  [879] = 825,
  [880] = 808,
  [881] = 809,
  [882] = 817,
  [883] = 835,
  [884] = 837,
  [885] = 843,
  [886] = 810,
  [887] = 814,
  [888] = 813,
  [889] = 857,
  [890] = 858,
  [891] = 812,
  [892] = 852,
  [893] = 857,
  [894] = 858,
  [895] = 856,
  [896] = 857,
  [897] = 858,
  [898] = 812,
  [899] = 813,
  [900] = 814,
  [901] = 862,
  [902] = 807,
  [903] = 814,
  [904] = 862,
  [905] = 812,
  [906] = 806,
  [907] = 807,
  [908] = 816,
  [909] = 818,
  [910] = 818,
  [911] = 862,
  [912] = 820,
  [913] = 824,
  [914] = 807,
  [915] = 822,
  [916] = 825,
  [917] = 820,
  [918] = 808,
  [919] = 809,
  [920] = 817,
  [921] = 835,
  [922] = 862,
  [923] = 810,
  [924] = 858,
  [925] = 813,
  [926] = 814,
  [927] = 806,
  [928] = 816,
  [929] = 818,
  [930] = 820,
  [931] = 822,
  [932] = 825,
  [933] = 808,
  [934] = 809,
  [935] = 817,
  [936] = 835,
  [937] = 837,
  [938] = 843,
  [939] = 810,
  [940] = 856,
  [941] = 857,
  [942] = 858,
  [943] = 812,
  [944] = 822,
  [945] = 825,
  [946] = 862,
  [947] = 807,
  [948] = 856,
  [949] = 857,
  [950] = 858,
  [951] = 812,
  [952] = 808,
  [953] = 809,
  [954] = 862,
  [955] = 807,
  [956] = 813,
  [957] = 817,
  [958] = 814,
  [959] = 835,
  [960] = 852,
  [961] = 837,
  [962] = 816,
  [963] = 852,
  [964] = 824,
  [965] = 818,
  [966] = 843,
  [967] = 808,
  [968] = 820,
  [969] = 837,
  [970] = 822,
  [971] = 843,
  [972] = 825,
  [973] = 809,
  [974] = 808,
  [975] = 810,
  [976] = 809,
  [977] = 810,
  [978] = 813,
  [979] = 814,
  [980] = 806,
  [981] = 816,
  [982] = 816,
  [983] = 818,
  [984] = 835,
  [985] = 837,
  [986] = 820,
  [987] = 843,
  [988] = 822,
  [989] = 825,
  [990] = 808,
  [991] = 856,
  [992] = 857,
  [993] = 858,
  [994] = 812,
  [995] = 817,
  [996] = 813,
  [997] = 862,
  [998] = 807,
  [999] = 809,
  [1000] = 817,
  [1001] = 835,
  [1002] = 837,
  [1003] = 843,
  [1004] = 814,
  [1005] = 852,
  [1006] = 806,
  [1007] = 807,
  [1008] = 824,
  [1009] = 816,
  [1010] = 818,
  [1011] = 820,
  [1012] = 822,
  [1013] = 852,
  [1014] = 825,
  [1015] = 852,
  [1016] = 806,
  [1017] = 816,
  [1018] = 856,
  [1019] = 808,
  [1020] = 818,
  [1021] = 856,
  [1022] = 857,
  [1023] = 853,
  [1024] = 809,
  [1025] = 858,
  [1026] = 817,
  [1027] = 812,
  [1028] = 835,
  [1029] = 837,
  [1030] = 852,
  [1031] = 843,
  [1032] = 820,
  [1033] = 824,
  [1034] = 862,
  [1035] = 856,
  [1036] = 857,
  [1037] = 858,
  [1038] = 807,
  [1039] = 812,
  [1040] = 806,
  [1041] = 820,
  [1042] = 862,
  [1043] = 816,
  [1044] = 807,
  [1045] = 810,
  [1046] = 862,
  [1047] = 857,
  [1048] = 822,
  [1049] = 825,
  [1050] = 808,
  [1051] = 809,
  [1052] = 813,
  [1053] = 817,
  [1054] = 856,
  [1055] = 857,
  [1056] = 858,
  [1057] = 810,
  [1058] = 824,
  [1059] = 818,
  [1060] = 835,
  [1061] = 813,
  [1062] = 814,
  [1063] = 806,
  [1064] = 816,
  [1065] = 818,
  [1066] = 820,
  [1067] = 822,
  [1068] = 825,
  [1069] = 810,
  [1070] = 808,
  [1071] = 809,
  [1072] = 813,
  [1073] = 817,
  [1074] = 835,
  [1075] = 814,
  [1076] = 837,
  [1077] = 806,
  [1078] = 820,
  [1079] = 816,
  [1080] = 843,
  [1081] = 818,
  [1082] = 812,
  [1083] = 824,
  [1084] = 820,
  [1085] = 822,
  [1086] = 837,
  [1087] = 852,
  [1088] = 822,
  [1089] = 852,
  [1090] = 856,
  [1091] = 857,
  [1092] = 852,
  [1093] = 843,
  [1094] = 825,
  [1095] = 856,
  [1096] = 808,
  [1097] = 809,
  [1098] = 857,
  [1099] = 858,
  [1100] = 824,
  [1101] = 824,
  [1102] = 824,
  [1103] = 824,
  [1104] = 824,
  [1105] = 824,
  [1106] = 817,
  [1107] = 812,
  [1108] = 858,
  [1109] = 812,
  [1110] = 835,
  [1111] = 822,
  [1112] = 825,
  [1113] = 862,
  [1114] = 837,
  [1115] = 807,
  [1116] = 862,
  [1117] = 862,
  [1118] = 843,
  [1119] = 814,
  [1120] = 806,
  [1121] = 824,
  [1122] = 807,
  [1123] = 835,
  [1124] = 837,
  [1125] = 843,
  [1126] = 856,
  [1127] = 858,
  [1128] = 807,
  [1129] = 840,
  [1130] = 857,
  [1131] = 817,
  [1132] = 1132,
  [1133] = 1133,
  [1134] = 1134,
  [1135] = 1135,
  [1136] = 1136,
  [1137] = 1137,
  [1138] = 1138,
  [1139] = 1139,
  [1140] = 1140,
  [1141] = 1141,
  [1142] = 1136,
  [1143] = 1143,
  [1144] = 1134,
  [1145] = 1141,
  [1146] = 1137,
  [1147] = 1135,
  [1148] = 1140,
  [1149] = 1149,
  [1150] = 1150,
  [1151] = 1149,
  [1152] = 1139,
  [1153] = 1133,
  [1154] = 1154,
  [1155] = 1155,
  [1156] = 1143,
  [1157] = 1143,
  [1158] = 1143,
  [1159] = 1143,
  [1160] = 1143,
  [1161] = 1161,
  [1162] = 587,
  [1163] = 1143,
  [1164] = 1164,
  [1165] = 1143,
  [1166] = 1143,
  [1167] = 1143,
  [1168] = 1143,
  [1169] = 1143,
  [1170] = 1143,
  [1171] = 1143,
  [1172] = 1143,
  [1173] = 1143,
  [1174] = 1143,
  [1175] = 1143,
  [1176] = 1143,
  [1177] = 1143,
  [1178] = 1143,
  [1179] = 1143,
  [1180] = 1143,
  [1181] = 1143,
  [1182] = 1143,
  [1183] = 1143,
  [1184] = 1143,
  [1185] = 1143,
  [1186] = 1143,
  [1187] = 1143,
  [1188] = 1143,
  [1189] = 1143,
  [1190] = 1143,
  [1191] = 1143,
  [1192] = 1143,
  [1193] = 1143,
  [1194] = 1143,
  [1195] = 1132,
  [1196] = 1196,
  [1197] = 1196,
  [1198] = 1196,
  [1199] = 1196,
  [1200] = 1196,
  [1201] = 1196,
  [1202] = 1196,
  [1203] = 1196,
  [1204] = 1196,
  [1205] = 1196,
  [1206] = 1196,
  [1207] = 1196,
  [1208] = 1196,
  [1209] = 1196,
  [1210] = 1196,
  [1211] = 1196,
  [1212] = 1196,
  [1213] = 1196,
  [1214] = 1196,
  [1215] = 1196,
  [1216] = 1196,
  [1217] = 1196,
  [1218] = 1196,
  [1219] = 1196,
  [1220] = 1196,
  [1221] = 1196,
  [1222] = 1196,
  [1223] = 1196,
  [1224] = 1196,
  [1225] = 1196,
  [1226] = 1196,
  [1227] = 1196,
  [1228] = 1196,
  [1229] = 587,
  [1230] = 1196,
  [1231] = 1196,
  [1232] = 1196,
  [1233] = 1196,
  [1234] = 1138,
  [1235] = 1132,
  [1236] = 1141,
  [1237] = 1137,
  [1238] = 1140,
  [1239] = 1139,
  [1240] = 1136,
  [1241] = 1135,
  [1242] = 1133,
  [1243] = 1134,
  [1244] = 1164,
  [1245] = 1137,
  [1246] = 1139,
  [1247] = 1136,
  [1248] = 1134,
  [1249] = 1155,
  [1250] = 587,
  [1251] = 1149,
  [1252] = 1135,
  [1253] = 1161,
  [1254] = 1141,
  [1255] = 1140,
  [1256] = 1133,
  [1257] = 1150,
  [1258] = 1132,
  [1259] = 1154,
  [1260] = 1132,
  [1261] = 1137,
  [1262] = 1141,
  [1263] = 1140,
  [1264] = 1139,
  [1265] = 1133,
  [1266] = 1136,
  [1267] = 1135,
  [1268] = 1134,
  [1269] = 1138,
  [1270] = 1135,
  [1271] = 1132,
  [1272] = 1141,
  [1273] = 587,
  [1274] = 1137,
  [1275] = 1141,
  [1276] = 1149,
  [1277] = 1135,
  [1278] = 1134,
  [1279] = 1141,
  [1280] = 1140,
  [1281] = 1137,
  [1282] = 1139,
  [1283] = 1136,
  [1284] = 1141,
  [1285] = 1140,
  [1286] = 1136,
  [1287] = 1140,
  [1288] = 1137,
  [1289] = 1136,
  [1290] = 1132,
  [1291] = 1139,
  [1292] = 1133,
  [1293] = 1135,
  [1294] = 1134,
  [1295] = 1136,
  [1296] = 1138,
  [1297] = 1133,
  [1298] = 1134,
  [1299] = 1132,
  [1300] = 1133,
  [1301] = 1138,
  [1302] = 1137,
  [1303] = 1140,
  [1304] = 1134,
  [1305] = 1138,
  [1306] = 1155,
  [1307] = 1161,
  [1308] = 1164,
  [1309] = 1133,
  [1310] = 1150,
  [1311] = 1139,
  [1312] = 1154,
  [1313] = 1132,
  [1314] = 1139,
  [1315] = 1135,
  [1316] = 1134,
  [1317] = 1132,
  [1318] = 1149,
  [1319] = 1150,
  [1320] = 1155,
  [1321] = 1161,
  [1322] = 1164,
  [1323] = 587,
  [1324] = 1141,
  [1325] = 1137,
  [1326] = 1140,
  [1327] = 1139,
  [1328] = 1141,
  [1329] = 1154,
  [1330] = 1137,
  [1331] = 1136,
  [1332] = 1140,
  [1333] = 1139,
  [1334] = 1136,
  [1335] = 1134,
  [1336] = 1135,
  [1337] = 1135,
  [1338] = 1133,
  [1339] = 1133,
  [1340] = 1149,
  [1341] = 1132,
  [1342] = 1150,
  [1343] = 1132,
  [1344] = 1161,
  [1345] = 1164,
  [1346] = 1154,
  [1347] = 1149,
  [1348] = 587,
  [1349] = 1141,
  [1350] = 1137,
  [1351] = 1150,
  [1352] = 1134,
  [1353] = 1155,
  [1354] = 1161,
  [1355] = 1140,
  [1356] = 1164,
  [1357] = 1139,
  [1358] = 587,
  [1359] = 1135,
  [1360] = 1136,
  [1361] = 1154,
  [1362] = 1133,
  [1363] = 1155,
  [1364] = 1133,
  [1365] = 1137,
  [1366] = 1134,
  [1367] = 1138,
  [1368] = 1140,
  [1369] = 1139,
  [1370] = 1135,
  [1371] = 1136,
  [1372] = 1132,
  [1373] = 1141,
  [1374] = 1150,
  [1375] = 587,
  [1376] = 1136,
  [1377] = 1138,
  [1378] = 1139,
  [1379] = 1133,
  [1380] = 1155,
  [1381] = 1141,
  [1382] = 1137,
  [1383] = 1161,
  [1384] = 1134,
  [1385] = 1164,
  [1386] = 1154,
  [1387] = 1140,
  [1388] = 1137,
  [1389] = 1132,
  [1390] = 1139,
  [1391] = 1133,
  [1392] = 1135,
  [1393] = 1132,
  [1394] = 1141,
  [1395] = 1135,
  [1396] = 1136,
  [1397] = 1138,
  [1398] = 1149,
  [1399] = 1140,
  [1400] = 1134,
  [1401] = 1154,
  [1402] = 1161,
  [1403] = 1155,
  [1404] = 1134,
  [1405] = 1135,
  [1406] = 1149,
  [1407] = 1154,
  [1408] = 1149,
  [1409] = 1140,
  [1410] = 1133,
  [1411] = 587,
  [1412] = 1136,
  [1413] = 1139,
  [1414] = 1164,
  [1415] = 1150,
  [1416] = 1141,
  [1417] = 1132,
  [1418] = 587,
  [1419] = 1150,
  [1420] = 1155,
  [1421] = 1421,
  [1422] = 1137,
  [1423] = 1423,
  [1424] = 1424,
  [1425] = 1133,
  [1426] = 1134,
  [1427] = 1133,
  [1428] = 1428,
  [1429] = 1134,
  [1430] = 1133,
  [1431] = 1424,
  [1432] = 1432,
  [1433] = 1134,
  [1434] = 1434,
  [1435] = 1423,
  [1436] = 1436,
  [1437] = 1437,
  [1438] = 1438,
  [1439] = 1439,
  [1440] = 1440,
  [1441] = 1441,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 1444,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 1449,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1455,
  [1456] = 1456,
  [1457] = 1457,
  [1458] = 1458,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1463,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1466,
  [1467] = 1467,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1470,
  [1471] = 1471,
  [1472] = 1472,
  [1473] = 1473,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1476,
  [1477] = 1477,
  [1478] = 1478,
  [1479] = 1479,
  [1480] = 1480,
  [1481] = 1481,
  [1482] = 1482,
  [1483] = 1483,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1490,
  [1491] = 1491,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 1494,
  [1495] = 1495,
  [1496] = 1496,
  [1497] = 1497,
  [1498] = 1498,
  [1499] = 1499,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 1502,
  [1503] = 1503,
  [1504] = 1504,
  [1505] = 1505,
  [1506] = 1506,
  [1507] = 1507,
  [1508] = 1508,
  [1509] = 1509,
  [1510] = 1510,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1513,
  [1514] = 1514,
  [1515] = 1515,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1133,
  [1520] = 1134,
  [1521] = 1521,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1428,
  [1529] = 1133,
  [1530] = 1134,
  [1531] = 1134,
  [1532] = 1133,
  [1533] = 1432,
  [1534] = 1133,
  [1535] = 1134,
  [1536] = 1434,
  [1537] = 1469,
  [1538] = 1503,
  [1539] = 1508,
  [1540] = 1517,
  [1541] = 1481,
  [1542] = 1491,
  [1543] = 1492,
  [1544] = 1500,
  [1545] = 1447,
  [1546] = 1453,
  [1547] = 1454,
  [1548] = 1455,
  [1549] = 1456,
  [1550] = 1457,
  [1551] = 1460,
  [1552] = 1461,
  [1553] = 1462,
  [1554] = 1463,
  [1555] = 1464,
  [1556] = 1470,
  [1557] = 1522,
  [1558] = 1524,
  [1559] = 1525,
  [1560] = 1497,
  [1561] = 1443,
  [1562] = 1476,
  [1563] = 1521,
  [1564] = 1442,
  [1565] = 1482,
  [1566] = 1438,
  [1567] = 1502,
  [1568] = 1518,
  [1569] = 1444,
  [1570] = 1445,
  [1571] = 1446,
  [1572] = 1450,
  [1573] = 1458,
  [1574] = 1459,
  [1575] = 1465,
  [1576] = 1466,
  [1577] = 1483,
  [1578] = 1484,
  [1579] = 1486,
  [1580] = 1487,
  [1581] = 1496,
  [1582] = 1501,
  [1583] = 1504,
  [1584] = 1506,
  [1585] = 1493,
  [1586] = 1510,
  [1587] = 1513,
  [1588] = 1514,
  [1589] = 1515,
  [1590] = 1468,
  [1591] = 1439,
  [1592] = 1451,
  [1593] = 1473,
  [1594] = 1478,
  [1595] = 1526,
  [1596] = 1489,
  [1597] = 1511,
  [1598] = 1452,
  [1599] = 1475,
  [1600] = 1477,
  [1601] = 1440,
  [1602] = 1449,
  [1603] = 1467,
  [1604] = 1474,
  [1605] = 1479,
  [1606] = 1480,
  [1607] = 1485,
  [1608] = 1488,
  [1609] = 1490,
  [1610] = 1494,
  [1611] = 1495,
  [1612] = 1498,
  [1613] = 1507,
  [1614] = 1499,
  [1615] = 1512,
  [1616] = 1471,
  [1617] = 1527,
  [1618] = 1134,
  [1619] = 1437,
  [1620] = 1448,
  [1621] = 1441,
  [1622] = 1523,
  [1623] = 1133,
  [1624] = 1516,
  [1625] = 1505,
  [1626] = 1436,
  [1627] = 1472,
  [1628] = 1509,
  [1629] = 1423,
  [1630] = 1424,
  [1631] = 1133,
  [1632] = 1632,
  [1633] = 1133,
  [1634] = 1134,
  [1635] = 1428,
  [1636] = 1134,
  [1637] = 1434,
  [1638] = 1432,
  [1639] = 1133,
  [1640] = 1134,
  [1641] = 1641,
  [1642] = 1480,
  [1643] = 1454,
  [1644] = 1455,
  [1645] = 1456,
  [1646] = 1457,
  [1647] = 1460,
  [1648] = 1461,
  [1649] = 1462,
  [1650] = 1463,
  [1651] = 1464,
  [1652] = 1470,
  [1653] = 1493,
  [1654] = 1522,
  [1655] = 1524,
  [1656] = 1525,
  [1657] = 1497,
  [1658] = 1443,
  [1659] = 1476,
  [1660] = 1521,
  [1661] = 1442,
  [1662] = 1482,
  [1663] = 1438,
  [1664] = 1502,
  [1665] = 1518,
  [1666] = 1444,
  [1667] = 1445,
  [1668] = 1446,
  [1669] = 1450,
  [1670] = 1458,
  [1671] = 1459,
  [1672] = 1465,
  [1673] = 1466,
  [1674] = 1469,
  [1675] = 1483,
  [1676] = 1484,
  [1677] = 1486,
  [1678] = 1487,
  [1679] = 1496,
  [1680] = 1501,
  [1681] = 1504,
  [1682] = 1506,
  [1683] = 1509,
  [1684] = 1510,
  [1685] = 1513,
  [1686] = 1514,
  [1687] = 1515,
  [1688] = 1468,
  [1689] = 1439,
  [1690] = 1451,
  [1691] = 1473,
  [1692] = 1478,
  [1693] = 1526,
  [1694] = 1489,
  [1695] = 1511,
  [1696] = 1452,
  [1697] = 1475,
  [1698] = 1477,
  [1699] = 1440,
  [1700] = 1449,
  [1701] = 1467,
  [1702] = 1474,
  [1703] = 1503,
  [1704] = 1485,
  [1705] = 1488,
  [1706] = 1490,
  [1707] = 1494,
  [1708] = 1495,
  [1709] = 1498,
  [1710] = 1471,
  [1711] = 1507,
  [1712] = 1499,
  [1713] = 1508,
  [1714] = 1512,
  [1715] = 1134,
  [1716] = 1527,
  [1717] = 1437,
  [1718] = 1516,
  [1719] = 1133,
  [1720] = 1517,
  [1721] = 1481,
  [1722] = 1505,
  [1723] = 1491,
  [1724] = 1492,
  [1725] = 1500,
  [1726] = 1453,
  [1727] = 1436,
  [1728] = 1472,
  [1729] = 1523,
  [1730] = 1479,
  [1731] = 1731,
  [1732] = 1732,
  [1733] = 1733,
  [1734] = 1734,
  [1735] = 1735,
  [1736] = 1736,
  [1737] = 1737,
  [1738] = 1738,
  [1739] = 1739,
  [1740] = 1740,
  [1741] = 1139,
  [1742] = 1742,
  [1743] = 1140,
  [1744] = 1744,
  [1745] = 1745,
  [1746] = 1746,
  [1747] = 1747,
  [1748] = 1748,
  [1749] = 1133,
  [1750] = 1750,
  [1751] = 1751,
  [1752] = 1752,
  [1753] = 1753,
  [1754] = 1754,
  [1755] = 1136,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 1760,
  [1761] = 1141,
  [1762] = 1762,
  [1763] = 1135,
  [1764] = 1134,
  [1765] = 1765,
  [1766] = 1766,
  [1767] = 1767,
  [1768] = 1137,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1771,
  [1772] = 1772,
  [1773] = 1773,
  [1774] = 1132,
  [1775] = 1775,
  [1776] = 1776,
  [1777] = 1777,
  [1778] = 1773,
  [1779] = 1779,
  [1780] = 1780,
  [1781] = 1781,
  [1782] = 1782,
  [1783] = 1783,
  [1784] = 1784,
  [1785] = 1785,
  [1786] = 1786,
  [1787] = 1787,
  [1788] = 1788,
  [1789] = 1789,
  [1790] = 1790,
  [1791] = 1791,
  [1792] = 1792,
  [1793] = 1793,
  [1794] = 1794,
  [1795] = 1795,
  [1796] = 1796,
  [1797] = 1797,
  [1798] = 1798,
  [1799] = 1799,
  [1800] = 1800,
  [1801] = 1801,
  [1802] = 1802,
  [1803] = 1803,
  [1804] = 1804,
  [1805] = 1805,
  [1806] = 1806,
  [1807] = 1807,
  [1808] = 1773,
  [1809] = 1784,
  [1810] = 1773,
  [1811] = 1811,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 1814,
  [1815] = 1815,
  [1816] = 1816,
  [1817] = 1817,
  [1818] = 1818,
  [1819] = 1819,
  [1820] = 1820,
  [1821] = 1821,
  [1822] = 1822,
  [1823] = 1823,
  [1824] = 1824,
  [1825] = 1825,
  [1826] = 1826,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1829,
  [1830] = 1830,
  [1831] = 1831,
  [1832] = 1832,
  [1833] = 1833,
  [1834] = 1834,
  [1835] = 1835,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1838,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1841,
  [1842] = 1842,
  [1843] = 1843,
  [1844] = 1844,
  [1845] = 1845,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1849,
  [1850] = 1850,
  [1851] = 1851,
  [1852] = 1852,
  [1853] = 1853,
  [1854] = 1854,
  [1855] = 1855,
  [1856] = 1856,
  [1857] = 1857,
  [1858] = 1858,
  [1859] = 1859,
  [1860] = 1860,
  [1861] = 1861,
  [1862] = 1862,
  [1863] = 1863,
  [1864] = 1864,
  [1865] = 1865,
  [1866] = 1866,
  [1867] = 1867,
  [1868] = 1868,
  [1869] = 1869,
  [1870] = 1641,
  [1871] = 1871,
  [1872] = 1872,
  [1873] = 1873,
  [1874] = 1421,
  [1875] = 1875,
  [1876] = 1876,
  [1877] = 1877,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1883,
  [1884] = 1884,
  [1885] = 1885,
  [1886] = 1886,
  [1887] = 1887,
  [1888] = 1888,
  [1889] = 1889,
  [1890] = 1890,
  [1891] = 1891,
  [1892] = 1892,
  [1893] = 1893,
  [1894] = 1894,
  [1895] = 1895,
  [1896] = 1896,
  [1897] = 1897,
  [1898] = 1898,
  [1899] = 1899,
  [1900] = 1900,
  [1901] = 1901,
  [1902] = 1902,
  [1903] = 1903,
  [1904] = 1904,
  [1905] = 1905,
  [1906] = 1906,
  [1907] = 1907,
  [1908] = 1908,
  [1909] = 1909,
  [1910] = 1910,
  [1911] = 1911,
  [1912] = 1912,
  [1913] = 1913,
  [1914] = 1914,
  [1915] = 1915,
  [1916] = 1916,
  [1917] = 1917,
  [1918] = 1918,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 1921,
  [1922] = 1922,
  [1923] = 1923,
  [1924] = 1924,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1927,
  [1928] = 1928,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1937,
  [1938] = 1938,
  [1939] = 1939,
  [1940] = 1940,
  [1941] = 1941,
  [1942] = 1942,
  [1943] = 1943,
  [1944] = 1944,
  [1945] = 1945,
  [1946] = 1946,
  [1947] = 1947,
  [1948] = 1948,
  [1949] = 1949,
  [1950] = 1950,
  [1951] = 1951,
  [1952] = 1952,
  [1953] = 1953,
  [1954] = 587,
  [1955] = 1955,
  [1956] = 1956,
  [1957] = 1957,
  [1958] = 1958,
  [1959] = 1959,
  [1960] = 1960,
  [1961] = 1961,
  [1962] = 1962,
  [1963] = 1963,
  [1964] = 1964,
  [1965] = 1965,
  [1966] = 1966,
  [1967] = 1967,
  [1968] = 1968,
  [1969] = 1737,
  [1970] = 1736,
  [1971] = 1738,
  [1972] = 1739,
  [1973] = 1740,
  [1974] = 1742,
  [1975] = 1744,
  [1976] = 1738,
  [1977] = 1766,
  [1978] = 1762,
  [1979] = 1737,
  [1980] = 1736,
  [1981] = 1769,
  [1982] = 1740,
  [1983] = 1739,
  [1984] = 1737,
  [1985] = 1737,
  [1986] = 1738,
  [1987] = 1736,
  [1988] = 1737,
  [1989] = 1738,
  [1990] = 1736,
  [1991] = 1738,
  [1992] = 1736,
  [1993] = 1790,
  [1994] = 1814,
  [1995] = 1140,
  [1996] = 1139,
  [1997] = 1736,
  [1998] = 1784,
  [1999] = 1737,
  [2000] = 1738,
  [2001] = 1740,
  [2002] = 1739,
  [2003] = 1794,
  [2004] = 1740,
  [2005] = 1739,
  [2006] = 1744,
  [2007] = 1742,
  [2008] = 1740,
  [2009] = 1739,
  [2010] = 1784,
  [2011] = 1818,
  [2012] = 1939,
  [2013] = 1134,
  [2014] = 1135,
  [2015] = 1766,
  [2016] = 1757,
  [2017] = 1742,
  [2018] = 1765,
  [2019] = 1744,
  [2020] = 1747,
  [2021] = 1752,
  [2022] = 1754,
  [2023] = 1745,
  [2024] = 1748,
  [2025] = 1133,
  [2026] = 1139,
  [2027] = 1750,
  [2028] = 1753,
  [2029] = 1770,
  [2030] = 1137,
  [2031] = 1746,
  [2032] = 1136,
  [2033] = 1742,
  [2034] = 1938,
  [2035] = 1762,
  [2036] = 1140,
  [2037] = 1139,
  [2038] = 1140,
  [2039] = 1139,
  [2040] = 1767,
  [2041] = 1948,
  [2042] = 1824,
  [2043] = 1828,
  [2044] = 1831,
  [2045] = 1832,
  [2046] = 1834,
  [2047] = 1837,
  [2048] = 1140,
  [2049] = 1839,
  [2050] = 1840,
  [2051] = 1842,
  [2052] = 1843,
  [2053] = 1844,
  [2054] = 1845,
  [2055] = 1760,
  [2056] = 1758,
  [2057] = 1769,
  [2058] = 1759,
  [2059] = 1742,
  [2060] = 1751,
  [2061] = 1756,
  [2062] = 1901,
  [2063] = 1744,
  [2064] = 1744,
  [2065] = 1906,
  [2066] = 1856,
  [2067] = 1884,
  [2068] = 1913,
  [2069] = 1855,
  [2070] = 1866,
  [2071] = 1829,
  [2072] = 1881,
  [2073] = 1951,
  [2074] = 1860,
  [2075] = 1903,
  [2076] = 1141,
  [2077] = 1838,
  [2078] = 1750,
  [2079] = 1766,
  [2080] = 1770,
  [2081] = 1746,
  [2082] = 1737,
  [2083] = 1141,
  [2084] = 1137,
  [2085] = 1773,
  [2086] = 1751,
  [2087] = 1738,
  [2088] = 1765,
  [2089] = 1747,
  [2090] = 1752,
  [2091] = 1751,
  [2092] = 1758,
  [2093] = 1759,
  [2094] = 1767,
  [2095] = 1751,
  [2096] = 1754,
  [2097] = 1757,
  [2098] = 1736,
  [2099] = 1765,
  [2100] = 1747,
  [2101] = 1752,
  [2102] = 1758,
  [2103] = 1759,
  [2104] = 1767,
  [2105] = 1141,
  [2106] = 1137,
  [2107] = 1754,
  [2108] = 1757,
  [2109] = 1745,
  [2110] = 1141,
  [2111] = 1137,
  [2112] = 1748,
  [2113] = 1745,
  [2114] = 1136,
  [2115] = 1748,
  [2116] = 1750,
  [2117] = 1753,
  [2118] = 1753,
  [2119] = 1135,
  [2120] = 1760,
  [2121] = 1136,
  [2122] = 1135,
  [2123] = 1756,
  [2124] = 1760,
  [2125] = 1756,
  [2126] = 1770,
  [2127] = 1746,
  [2128] = 1770,
  [2129] = 1746,
  [2130] = 1133,
  [2131] = 1746,
  [2132] = 1133,
  [2133] = 1769,
  [2134] = 1769,
  [2135] = 1134,
  [2136] = 1765,
  [2137] = 1747,
  [2138] = 1752,
  [2139] = 1758,
  [2140] = 1759,
  [2141] = 1767,
  [2142] = 1134,
  [2143] = 1141,
  [2144] = 1137,
  [2145] = 1754,
  [2146] = 1757,
  [2147] = 1762,
  [2148] = 1140,
  [2149] = 1136,
  [2150] = 1762,
  [2151] = 1139,
  [2152] = 1135,
  [2153] = 1745,
  [2154] = 1135,
  [2155] = 1737,
  [2156] = 1133,
  [2157] = 1766,
  [2158] = 1133,
  [2159] = 1134,
  [2160] = 1751,
  [2161] = 1132,
  [2162] = 1771,
  [2163] = 1769,
  [2164] = 1762,
  [2165] = 1766,
  [2166] = 1742,
  [2167] = 1744,
  [2168] = 1136,
  [2169] = 1738,
  [2170] = 1772,
  [2171] = 1765,
  [2172] = 1747,
  [2173] = 1752,
  [2174] = 1758,
  [2175] = 1759,
  [2176] = 1767,
  [2177] = 1748,
  [2178] = 1750,
  [2179] = 1753,
  [2180] = 1754,
  [2181] = 1757,
  [2182] = 1736,
  [2183] = 1134,
  [2184] = 1745,
  [2185] = 1748,
  [2186] = 1750,
  [2187] = 1753,
  [2188] = 1760,
  [2189] = 1756,
  [2190] = 1760,
  [2191] = 1756,
  [2192] = 1770,
  [2193] = 1822,
  [2194] = 1772,
  [2195] = 1785,
  [2196] = 1773,
  [2197] = 1773,
  [2198] = 1739,
  [2199] = 1771,
  [2200] = 1781,
  [2201] = 1132,
  [2202] = 1773,
  [2203] = 1740,
  [2204] = 1739,
  [2205] = 1792,
  [2206] = 1805,
  [2207] = 1740,
  [2208] = 1812,
  [2209] = 1784,
  [2210] = 1802,
  [2211] = 1788,
  [2212] = 1772,
  [2213] = 1806,
  [2214] = 1818,
  [2215] = 1814,
  [2216] = 1780,
  [2217] = 1815,
  [2218] = 1816,
  [2219] = 1776,
  [2220] = 1820,
  [2221] = 1789,
  [2222] = 1793,
  [2223] = 1782,
  [2224] = 1787,
  [2225] = 1773,
  [2226] = 1798,
  [2227] = 1821,
  [2228] = 1783,
  [2229] = 1773,
  [2230] = 1775,
  [2231] = 1800,
  [2232] = 1819,
  [2233] = 1795,
  [2234] = 1132,
  [2235] = 1773,
  [2236] = 1817,
  [2237] = 1771,
  [2238] = 1791,
  [2239] = 1777,
  [2240] = 1771,
  [2241] = 1772,
  [2242] = 1786,
  [2243] = 1773,
  [2244] = 1132,
  [2245] = 1813,
  [2246] = 1804,
  [2247] = 1803,
  [2248] = 1799,
  [2249] = 1794,
  [2250] = 1784,
  [2251] = 1769,
  [2252] = 1762,
  [2253] = 1771,
  [2254] = 1790,
  [2255] = 1796,
  [2256] = 1797,
  [2257] = 1811,
  [2258] = 1766,
  [2259] = 1132,
  [2260] = 1807,
  [2261] = 1772,
  [2262] = 1801,
  [2263] = 1779,
  [2264] = 1812,
  [2265] = 1805,
  [2266] = 1805,
  [2267] = 1134,
  [2268] = 1803,
  [2269] = 1806,
  [2270] = 1784,
  [2271] = 1780,
  [2272] = 1840,
  [2273] = 1782,
  [2274] = 1818,
  [2275] = 1783,
  [2276] = 1894,
  [2277] = 1790,
  [2278] = 1806,
  [2279] = 1788,
  [2280] = 1791,
  [2281] = 1794,
  [2282] = 1796,
  [2283] = 1797,
  [2284] = 1811,
  [2285] = 1807,
  [2286] = 1814,
  [2287] = 1802,
  [2288] = 1820,
  [2289] = 1821,
  [2290] = 1822,
  [2291] = 1813,
  [2292] = 1795,
  [2293] = 1799,
  [2294] = 1816,
  [2295] = 1776,
  [2296] = 1779,
  [2297] = 1842,
  [2298] = 1896,
  [2299] = 1790,
  [2300] = 1777,
  [2301] = 1800,
  [2302] = 1817,
  [2303] = 1922,
  [2304] = 1794,
  [2305] = 1819,
  [2306] = 1805,
  [2307] = 1806,
  [2308] = 1780,
  [2309] = 1843,
  [2310] = 1923,
  [2311] = 1782,
  [2312] = 1783,
  [2313] = 1844,
  [2314] = 1788,
  [2315] = 1791,
  [2316] = 1965,
  [2317] = 1790,
  [2318] = 1872,
  [2319] = 1780,
  [2320] = 1790,
  [2321] = 1789,
  [2322] = 1794,
  [2323] = 1779,
  [2324] = 1784,
  [2325] = 1802,
  [2326] = 1140,
  [2327] = 1845,
  [2328] = 1862,
  [2329] = 1133,
  [2330] = 1846,
  [2331] = 1795,
  [2332] = 1820,
  [2333] = 1821,
  [2334] = 1139,
  [2335] = 1135,
  [2336] = 1822,
  [2337] = 1800,
  [2338] = 1817,
  [2339] = 1935,
  [2340] = 1849,
  [2341] = 1850,
  [2342] = 1796,
  [2343] = 1959,
  [2344] = 1784,
  [2345] = 1826,
  [2346] = 1789,
  [2347] = 1830,
  [2348] = 1851,
  [2349] = 1852,
  [2350] = 1858,
  [2351] = 1871,
  [2352] = 1901,
  [2353] = 1833,
  [2354] = 1785,
  [2355] = 1784,
  [2356] = 1875,
  [2357] = 1799,
  [2358] = 1784,
  [2359] = 1786,
  [2360] = 1792,
  [2361] = 1812,
  [2362] = 1827,
  [2363] = 1775,
  [2364] = 1902,
  [2365] = 1903,
  [2366] = 1907,
  [2367] = 1773,
  [2368] = 1797,
  [2369] = 1776,
  [2370] = 1910,
  [2371] = 1911,
  [2372] = 1785,
  [2373] = 1786,
  [2374] = 1792,
  [2375] = 1775,
  [2376] = 1945,
  [2377] = 1915,
  [2378] = 1916,
  [2379] = 1864,
  [2380] = 1905,
  [2381] = 1811,
  [2382] = 1804,
  [2383] = 1920,
  [2384] = 1793,
  [2385] = 1787,
  [2386] = 1807,
  [2387] = 1929,
  [2388] = 1781,
  [2389] = 1804,
  [2390] = 1773,
  [2391] = 1781,
  [2392] = 1784,
  [2393] = 1782,
  [2394] = 1783,
  [2395] = 1908,
  [2396] = 1784,
  [2397] = 1848,
  [2398] = 1947,
  [2399] = 1793,
  [2400] = 1787,
  [2401] = 1940,
  [2402] = 1943,
  [2403] = 1888,
  [2404] = 1889,
  [2405] = 1893,
  [2406] = 1962,
  [2407] = 1890,
  [2408] = 1773,
  [2409] = 1883,
  [2410] = 1928,
  [2411] = 1891,
  [2412] = 1861,
  [2413] = 1867,
  [2414] = 1924,
  [2415] = 1926,
  [2416] = 1781,
  [2417] = 1936,
  [2418] = 1794,
  [2419] = 1798,
  [2420] = 1801,
  [2421] = 1803,
  [2422] = 1944,
  [2423] = 1950,
  [2424] = 1140,
  [2425] = 1788,
  [2426] = 1791,
  [2427] = 1897,
  [2428] = 1773,
  [2429] = 1779,
  [2430] = 1885,
  [2431] = 1819,
  [2432] = 1949,
  [2433] = 1802,
  [2434] = 1820,
  [2435] = 1773,
  [2436] = 1821,
  [2437] = 1822,
  [2438] = 1895,
  [2439] = 1898,
  [2440] = 1800,
  [2441] = 1817,
  [2442] = 1853,
  [2443] = 1892,
  [2444] = 1770,
  [2445] = 1746,
  [2446] = 1798,
  [2447] = 1887,
  [2448] = 1813,
  [2449] = 1933,
  [2450] = 1819,
  [2451] = 1773,
  [2452] = 1952,
  [2453] = 1789,
  [2454] = 1801,
  [2455] = 1798,
  [2456] = 1957,
  [2457] = 1813,
  [2458] = 1795,
  [2459] = 1799,
  [2460] = 1773,
  [2461] = 1816,
  [2462] = 1776,
  [2463] = 1835,
  [2464] = 1773,
  [2465] = 1836,
  [2466] = 1841,
  [2467] = 1854,
  [2468] = 1958,
  [2469] = 1773,
  [2470] = 1812,
  [2471] = 1804,
  [2472] = 1773,
  [2473] = 1773,
  [2474] = 1796,
  [2475] = 1797,
  [2476] = 1960,
  [2477] = 1773,
  [2478] = 1811,
  [2479] = 1136,
  [2480] = 1807,
  [2481] = 1742,
  [2482] = 1818,
  [2483] = 1744,
  [2484] = 1814,
  [2485] = 1961,
  [2486] = 1863,
  [2487] = 1139,
  [2488] = 1963,
  [2489] = 1938,
  [2490] = 1777,
  [2491] = 1751,
  [2492] = 1785,
  [2493] = 1939,
  [2494] = 1804,
  [2495] = 1805,
  [2496] = 1765,
  [2497] = 1747,
  [2498] = 1752,
  [2499] = 1758,
  [2500] = 1759,
  [2501] = 1806,
  [2502] = 1767,
  [2503] = 1780,
  [2504] = 1781,
  [2505] = 1782,
  [2506] = 1754,
  [2507] = 1757,
  [2508] = 1783,
  [2509] = 1786,
  [2510] = 1745,
  [2511] = 1748,
  [2512] = 1750,
  [2513] = 1753,
  [2514] = 1792,
  [2515] = 1777,
  [2516] = 1760,
  [2517] = 1756,
  [2518] = 1815,
  [2519] = 1796,
  [2520] = 1797,
  [2521] = 1948,
  [2522] = 1899,
  [2523] = 1865,
  [2524] = 1955,
  [2525] = 1737,
  [2526] = 1868,
  [2527] = 1811,
  [2528] = 1824,
  [2529] = 1828,
  [2530] = 1831,
  [2531] = 1738,
  [2532] = 1931,
  [2533] = 1932,
  [2534] = 1934,
  [2535] = 1823,
  [2536] = 1964,
  [2537] = 1801,
  [2538] = 1788,
  [2539] = 1791,
  [2540] = 1784,
  [2541] = 1873,
  [2542] = 1937,
  [2543] = 1906,
  [2544] = 1953,
  [2545] = 1880,
  [2546] = 1956,
  [2547] = 1856,
  [2548] = 1798,
  [2549] = 1918,
  [2550] = 1884,
  [2551] = 1912,
  [2552] = 1913,
  [2553] = 1855,
  [2554] = 1866,
  [2555] = 1869,
  [2556] = 1421,
  [2557] = 1829,
  [2558] = 1859,
  [2559] = 1820,
  [2560] = 1821,
  [2561] = 1822,
  [2562] = 1847,
  [2563] = 1800,
  [2564] = 1817,
  [2565] = 1881,
  [2566] = 1882,
  [2567] = 1942,
  [2568] = 1946,
  [2569] = 1951,
  [2570] = 1801,
  [2571] = 1877,
  [2572] = 1803,
  [2573] = 1807,
  [2574] = 1742,
  [2575] = 1927,
  [2576] = 1832,
  [2577] = 1775,
  [2578] = 1818,
  [2579] = 1789,
  [2580] = 1876,
  [2581] = 1744,
  [2582] = 1814,
  [2583] = 1815,
  [2584] = 1779,
  [2585] = 1904,
  [2586] = 1941,
  [2587] = 1834,
  [2588] = 1785,
  [2589] = 1786,
  [2590] = 1792,
  [2591] = 1775,
  [2592] = 1819,
  [2593] = 1641,
  [2594] = 1886,
  [2595] = 1930,
  [2596] = 1857,
  [2597] = 1803,
  [2598] = 1141,
  [2599] = 1815,
  [2600] = 1909,
  [2601] = 1816,
  [2602] = 1860,
  [2603] = 1793,
  [2604] = 1837,
  [2605] = 1784,
  [2606] = 1787,
  [2607] = 587,
  [2608] = 1878,
  [2609] = 1879,
  [2610] = 1900,
  [2611] = 1921,
  [2612] = 1838,
  [2613] = 1777,
  [2614] = 1914,
  [2615] = 1917,
  [2616] = 1925,
  [2617] = 1919,
  [2618] = 1793,
  [2619] = 1787,
  [2620] = 1839,
  [2621] = 1137,
  [2622] = 1825,
  [2623] = 1849,
  [2624] = 1915,
  [2625] = 1916,
  [2626] = 1940,
  [2627] = 1943,
  [2628] = 1893,
  [2629] = 1962,
  [2630] = 1883,
  [2631] = 1928,
  [2632] = 1861,
  [2633] = 1867,
  [2634] = 1924,
  [2635] = 1926,
  [2636] = 1936,
  [2637] = 1944,
  [2638] = 1950,
  [2639] = 1897,
  [2640] = 1885,
  [2641] = 1949,
  [2642] = 1853,
  [2643] = 1887,
  [2644] = 1825,
  [2645] = 1835,
  [2646] = 1836,
  [2647] = 1841,
  [2648] = 1827,
  [2649] = 1922,
  [2650] = 1923,
  [2651] = 1945,
  [2652] = 1920,
  [2653] = 1929,
  [2654] = 1947,
  [2655] = 1892,
  [2656] = 1854,
  [2657] = 1886,
  [2658] = 1894,
  [2659] = 1896,
  [2660] = 1965,
  [2661] = 1901,
  [2662] = 1827,
  [2663] = 1955,
  [2664] = 1964,
  [2665] = 1873,
  [2666] = 1937,
  [2667] = 1906,
  [2668] = 1953,
  [2669] = 1956,
  [2670] = 1856,
  [2671] = 1918,
  [2672] = 1884,
  [2673] = 1912,
  [2674] = 1913,
  [2675] = 1855,
  [2676] = 1866,
  [2677] = 1869,
  [2678] = 1641,
  [2679] = 1421,
  [2680] = 1829,
  [2681] = 1859,
  [2682] = 1847,
  [2683] = 1881,
  [2684] = 1882,
  [2685] = 1942,
  [2686] = 1946,
  [2687] = 1951,
  [2688] = 1877,
  [2689] = 587,
  [2690] = 1922,
  [2691] = 1923,
  [2692] = 1945,
  [2693] = 1920,
  [2694] = 1929,
  [2695] = 1930,
  [2696] = 1947,
  [2697] = 1909,
  [2698] = 1892,
  [2699] = 1860,
  [2700] = 1854,
  [2701] = 1878,
  [2702] = 1879,
  [2703] = 1900,
  [2704] = 1886,
  [2705] = 1914,
  [2706] = 1917,
  [2707] = 1919,
  [2708] = 1833,
  [2709] = 1848,
  [2710] = 1863,
  [2711] = 1865,
  [2712] = 1868,
  [2713] = 1894,
  [2714] = 1871,
  [2715] = 1864,
  [2716] = 1905,
  [2717] = 1896,
  [2718] = 1908,
  [2719] = 1965,
  [2720] = 1933,
  [2721] = 1901,
  [2722] = 1938,
  [2723] = 1939,
  [2724] = 1948,
  [2725] = 1824,
  [2726] = 1828,
  [2727] = 1831,
  [2728] = 1832,
  [2729] = 1834,
  [2730] = 1837,
  [2731] = 1838,
  [2732] = 1839,
  [2733] = 1840,
  [2734] = 1842,
  [2735] = 1843,
  [2736] = 1844,
  [2737] = 1845,
  [2738] = 1846,
  [2739] = 1849,
  [2740] = 1850,
  [2741] = 1851,
  [2742] = 1852,
  [2743] = 1769,
  [2744] = 1888,
  [2745] = 1889,
  [2746] = 1890,
  [2747] = 1891,
  [2748] = 1895,
  [2749] = 1898,
  [2750] = 1952,
  [2751] = 1957,
  [2752] = 1958,
  [2753] = 1960,
  [2754] = 1961,
  [2755] = 1963,
  [2756] = 1955,
  [2757] = 1964,
  [2758] = 1899,
  [2759] = 1873,
  [2760] = 1937,
  [2761] = 1906,
  [2762] = 1931,
  [2763] = 1932,
  [2764] = 1934,
  [2765] = 1823,
  [2766] = 1953,
  [2767] = 1880,
  [2768] = 1927,
  [2769] = 1876,
  [2770] = 1904,
  [2771] = 1941,
  [2772] = 1956,
  [2773] = 1857,
  [2774] = 1856,
  [2775] = 1918,
  [2776] = 1921,
  [2777] = 1884,
  [2778] = 1925,
  [2779] = 1912,
  [2780] = 1762,
  [2781] = 1913,
  [2782] = 1855,
  [2783] = 1866,
  [2784] = 1872,
  [2785] = 1862,
  [2786] = 1869,
  [2787] = 1935,
  [2788] = 1959,
  [2789] = 1826,
  [2790] = 1830,
  [2791] = 1421,
  [2792] = 1858,
  [2793] = 1829,
  [2794] = 1859,
  [2795] = 1875,
  [2796] = 1902,
  [2797] = 1903,
  [2798] = 1907,
  [2799] = 1910,
  [2800] = 1911,
  [2801] = 1915,
  [2802] = 1916,
  [2803] = 1847,
  [2804] = 1940,
  [2805] = 1943,
  [2806] = 1893,
  [2807] = 1962,
  [2808] = 1883,
  [2809] = 1928,
  [2810] = 1861,
  [2811] = 1867,
  [2812] = 1924,
  [2813] = 1926,
  [2814] = 1881,
  [2815] = 1936,
  [2816] = 1944,
  [2817] = 1950,
  [2818] = 1897,
  [2819] = 1885,
  [2820] = 1949,
  [2821] = 1853,
  [2822] = 1887,
  [2823] = 1766,
  [2824] = 1825,
  [2825] = 1835,
  [2826] = 1836,
  [2827] = 1841,
  [2828] = 1827,
  [2829] = 1882,
  [2830] = 1942,
  [2831] = 1946,
  [2832] = 1951,
  [2833] = 1877,
  [2834] = 1922,
  [2835] = 1923,
  [2836] = 1945,
  [2837] = 1920,
  [2838] = 1929,
  [2839] = 1947,
  [2840] = 1892,
  [2841] = 1854,
  [2842] = 1886,
  [2843] = 1894,
  [2844] = 1896,
  [2845] = 1965,
  [2846] = 1930,
  [2847] = 1815,
  [2848] = 1955,
  [2849] = 1964,
  [2850] = 1873,
  [2851] = 1937,
  [2852] = 1953,
  [2853] = 1956,
  [2854] = 1909,
  [2855] = 1918,
  [2856] = 1912,
  [2857] = 1869,
  [2858] = 1641,
  [2859] = 1421,
  [2860] = 1859,
  [2861] = 1860,
  [2862] = 1847,
  [2863] = 1882,
  [2864] = 1942,
  [2865] = 1946,
  [2866] = 1951,
  [2867] = 1877,
  [2868] = 1878,
  [2869] = 1879,
  [2870] = 1900,
  [2871] = 1914,
  [2872] = 1917,
  [2873] = 1919,
  [2874] = 1930,
  [2875] = 1909,
  [2876] = 1860,
  [2877] = 1878,
  [2878] = 1879,
  [2879] = 1900,
  [2880] = 1914,
  [2881] = 1917,
  [2882] = 1919,
  [2883] = 1813,
  [2884] = 1795,
  [2885] = 1799,
  [2886] = 1816,
  [2887] = 1776,
  [2888] = 1833,
  [2889] = 1848,
  [2890] = 1863,
  [2891] = 1865,
  [2892] = 1868,
  [2893] = 1871,
  [2894] = 1864,
  [2895] = 1905,
  [2896] = 1908,
  [2897] = 1833,
  [2898] = 1933,
  [2899] = 1848,
  [2900] = 1938,
  [2901] = 1939,
  [2902] = 1948,
  [2903] = 1824,
  [2904] = 1828,
  [2905] = 1831,
  [2906] = 1832,
  [2907] = 1834,
  [2908] = 1837,
  [2909] = 1838,
  [2910] = 1839,
  [2911] = 1840,
  [2912] = 1842,
  [2913] = 1843,
  [2914] = 1844,
  [2915] = 1845,
  [2916] = 1846,
  [2917] = 1863,
  [2918] = 1865,
  [2919] = 1850,
  [2920] = 1851,
  [2921] = 1852,
  [2922] = 1868,
  [2923] = 1769,
  [2924] = 1871,
  [2925] = 1802,
  [2926] = 1888,
  [2927] = 1889,
  [2928] = 1890,
  [2929] = 1891,
  [2930] = 1895,
  [2931] = 1898,
  [2932] = 1952,
  [2933] = 1957,
  [2934] = 1958,
  [2935] = 1960,
  [2936] = 1961,
  [2937] = 1963,
  [2938] = 1899,
  [2939] = 1864,
  [2940] = 1905,
  [2941] = 1908,
  [2942] = 1931,
  [2943] = 1932,
  [2944] = 1934,
  [2945] = 1823,
  [2946] = 1880,
  [2947] = 1927,
  [2948] = 1933,
  [2949] = 1876,
  [2950] = 1904,
  [2951] = 1941,
  [2952] = 1857,
  [2953] = 1921,
  [2954] = 1938,
  [2955] = 1925,
  [2956] = 1939,
  [2957] = 1762,
  [2958] = 1948,
  [2959] = 1812,
  [2960] = 1872,
  [2961] = 1862,
  [2962] = 1935,
  [2963] = 1959,
  [2964] = 1826,
  [2965] = 1830,
  [2966] = 1824,
  [2967] = 1858,
  [2968] = 1828,
  [2969] = 1831,
  [2970] = 1875,
  [2971] = 1832,
  [2972] = 1834,
  [2973] = 1837,
  [2974] = 1902,
  [2975] = 1903,
  [2976] = 1907,
  [2977] = 1910,
  [2978] = 1911,
  [2979] = 1915,
  [2980] = 1916,
  [2981] = 1838,
  [2982] = 1940,
  [2983] = 1943,
  [2984] = 1893,
  [2985] = 1962,
  [2986] = 1839,
  [2987] = 1883,
  [2988] = 1928,
  [2989] = 1840,
  [2990] = 1861,
  [2991] = 1867,
  [2992] = 1924,
  [2993] = 1926,
  [2994] = 1842,
  [2995] = 1936,
  [2996] = 1944,
  [2997] = 1950,
  [2998] = 1897,
  [2999] = 1885,
  [3000] = 1949,
  [3001] = 1853,
  [3002] = 1887,
  [3003] = 1766,
  [3004] = 1825,
  [3005] = 1835,
  [3006] = 1836,
  [3007] = 1841,
  [3008] = 1843,
  [3009] = 1844,
  [3010] = 1845,
  [3011] = 1846,
  [3012] = 1849,
  [3013] = 1850,
  [3014] = 1851,
  [3015] = 1852,
  [3016] = 1818,
  [3017] = 1814,
  [3018] = 1641,
  [3019] = 1421,
  [3020] = 1888,
  [3021] = 1889,
  [3022] = 1890,
  [3023] = 1891,
  [3024] = 1895,
  [3025] = 1898,
  [3026] = 1773,
  [3027] = 1952,
  [3028] = 1957,
  [3029] = 1958,
  [3030] = 1939,
  [3031] = 1911,
  [3032] = 1961,
  [3033] = 1963,
  [3034] = 1899,
  [3035] = 1931,
  [3036] = 1932,
  [3037] = 1934,
  [3038] = 1823,
  [3039] = 1880,
  [3040] = 1927,
  [3041] = 1876,
  [3042] = 1904,
  [3043] = 1941,
  [3044] = 1857,
  [3045] = 1921,
  [3046] = 1925,
  [3047] = 1872,
  [3048] = 1862,
  [3049] = 1935,
  [3050] = 1959,
  [3051] = 1826,
  [3052] = 1830,
  [3053] = 1858,
  [3054] = 1875,
  [3055] = 587,
  [3056] = 1902,
  [3057] = 1903,
  [3058] = 1907,
  [3059] = 1910,
  [3060] = 1911,
  [3061] = 1915,
  [3062] = 1916,
  [3063] = 587,
  [3064] = 1940,
  [3065] = 1943,
  [3066] = 1893,
  [3067] = 1962,
  [3068] = 1141,
  [3069] = 1883,
  [3070] = 1137,
  [3071] = 1928,
  [3072] = 587,
  [3073] = 1861,
  [3074] = 1867,
  [3075] = 1924,
  [3076] = 1926,
  [3077] = 1936,
  [3078] = 1944,
  [3079] = 1950,
  [3080] = 1897,
  [3081] = 1885,
  [3082] = 1949,
  [3083] = 1853,
  [3084] = 1887,
  [3085] = 1135,
  [3086] = 1825,
  [3087] = 1136,
  [3088] = 1835,
  [3089] = 1836,
  [3090] = 1841,
  [3091] = 1133,
  [3092] = 1134,
  [3093] = 1132,
  [3094] = 1771,
  [3095] = 1641,
  [3096] = 1772,
  [3097] = 1938,
  [3098] = 1948,
  [3099] = 1824,
  [3100] = 1828,
  [3101] = 1831,
  [3102] = 1832,
  [3103] = 1834,
  [3104] = 1837,
  [3105] = 1838,
  [3106] = 1839,
  [3107] = 1840,
  [3108] = 1842,
  [3109] = 1843,
  [3110] = 1844,
  [3111] = 1845,
  [3112] = 1903,
  [3113] = 1770,
  [3114] = 1746,
  [3115] = 1751,
  [3116] = 1765,
  [3117] = 1747,
  [3118] = 1752,
  [3119] = 1758,
  [3120] = 1759,
  [3121] = 1901,
  [3122] = 1906,
  [3123] = 1856,
  [3124] = 1884,
  [3125] = 1913,
  [3126] = 1855,
  [3127] = 1866,
  [3128] = 1829,
  [3129] = 1881,
  [3130] = 1860,
  [3131] = 1767,
  [3132] = 1951,
  [3133] = 1754,
  [3134] = 1757,
  [3135] = 1745,
  [3136] = 1748,
  [3137] = 1750,
  [3138] = 1753,
  [3139] = 1760,
  [3140] = 1910,
  [3141] = 1756,
  [3142] = 1860,
  [3143] = 1827,
  [3144] = 1922,
  [3145] = 1923,
  [3146] = 1945,
  [3147] = 1920,
  [3148] = 1929,
  [3149] = 1947,
  [3150] = 1892,
  [3151] = 1854,
  [3152] = 1886,
  [3153] = 1894,
  [3154] = 1896,
  [3155] = 1965,
  [3156] = 1901,
  [3157] = 1955,
  [3158] = 1964,
  [3159] = 1873,
  [3160] = 1937,
  [3161] = 1906,
  [3162] = 1953,
  [3163] = 1956,
  [3164] = 1856,
  [3165] = 1918,
  [3166] = 1884,
  [3167] = 1912,
  [3168] = 1913,
  [3169] = 1855,
  [3170] = 1866,
  [3171] = 1869,
  [3172] = 1829,
  [3173] = 1859,
  [3174] = 1847,
  [3175] = 1881,
  [3176] = 1882,
  [3177] = 1942,
  [3178] = 1946,
  [3179] = 1877,
  [3180] = 1930,
  [3181] = 1909,
  [3182] = 1860,
  [3183] = 1878,
  [3184] = 1879,
  [3185] = 1900,
  [3186] = 1914,
  [3187] = 1917,
  [3188] = 1919,
  [3189] = 1833,
  [3190] = 1848,
  [3191] = 1863,
  [3192] = 1865,
  [3193] = 1868,
  [3194] = 1871,
  [3195] = 1864,
  [3196] = 1905,
  [3197] = 1908,
  [3198] = 1907,
  [3199] = 1933,
  [3200] = 1846,
  [3201] = 1849,
  [3202] = 1850,
  [3203] = 1851,
  [3204] = 1852,
  [3205] = 1888,
  [3206] = 1889,
  [3207] = 1890,
  [3208] = 1891,
  [3209] = 1895,
  [3210] = 1898,
  [3211] = 1952,
  [3212] = 1957,
  [3213] = 1958,
  [3214] = 1960,
  [3215] = 1961,
  [3216] = 1963,
  [3217] = 1899,
  [3218] = 1931,
  [3219] = 1932,
  [3220] = 1934,
  [3221] = 1823,
  [3222] = 1880,
  [3223] = 1927,
  [3224] = 1876,
  [3225] = 1904,
  [3226] = 1941,
  [3227] = 1857,
  [3228] = 1921,
  [3229] = 1925,
  [3230] = 1872,
  [3231] = 1862,
  [3232] = 1935,
  [3233] = 1959,
  [3234] = 1826,
  [3235] = 1830,
  [3236] = 1858,
  [3237] = 1875,
  [3238] = 1902,
  [3239] = 1960,
  [3240] = 1829,
  [3241] = 1788,
  [3242] = 1791,
  [3243] = 1820,
  [3244] = 1821,
  [3245] = 1822,
  [3246] = 1800,
  [3247] = 1817,
  [3248] = 1789,
  [3249] = 1785,
  [3250] = 1786,
  [3251] = 1792,
  [3252] = 1775,
  [3253] = 1793,
  [3254] = 1787,
  [3255] = 3255,
  [3256] = 1771,
  [3257] = 1772,
  [3258] = 1881,
  [3259] = 1906,
  [3260] = 1903,
  [3261] = 1856,
  [3262] = 1773,
  [3263] = 1951,
  [3264] = 1884,
  [3265] = 1901,
  [3266] = 1913,
  [3267] = 1855,
  [3268] = 1967,
  [3269] = 1866,
  [3270] = 1968,
  [3271] = 1966,
  [3272] = 1132,
  [3273] = 1804,
  [3274] = 1781,
  [3275] = 1773,
  [3276] = 1845,
  [3277] = 1798,
  [3278] = 1801,
  [3279] = 1803,
  [3280] = 1938,
  [3281] = 1779,
  [3282] = 1819,
  [3283] = 3283,
  [3284] = 1948,
  [3285] = 1796,
  [3286] = 1797,
  [3287] = 1811,
  [3288] = 1807,
  [3289] = 1777,
  [3290] = 1805,
  [3291] = 1806,
  [3292] = 1780,
  [3293] = 1782,
  [3294] = 1783,
  [3295] = 1824,
  [3296] = 1828,
  [3297] = 1831,
  [3298] = 1832,
  [3299] = 1834,
  [3300] = 1837,
  [3301] = 1838,
  [3302] = 1839,
  [3303] = 1840,
  [3304] = 1773,
  [3305] = 1641,
  [3306] = 1842,
  [3307] = 1843,
  [3308] = 1844,
  [3309] = 1773,
  [3310] = 1793,
  [3311] = 1916,
  [3312] = 1940,
  [3313] = 1943,
  [3314] = 1893,
  [3315] = 1962,
  [3316] = 1958,
  [3317] = 1883,
  [3318] = 1928,
  [3319] = 1852,
  [3320] = 1861,
  [3321] = 1867,
  [3322] = 1924,
  [3323] = 1926,
  [3324] = 1936,
  [3325] = 1944,
  [3326] = 1950,
  [3327] = 1897,
  [3328] = 1885,
  [3329] = 1949,
  [3330] = 1853,
  [3331] = 1887,
  [3332] = 1825,
  [3333] = 1835,
  [3334] = 1836,
  [3335] = 1841,
  [3336] = 1804,
  [3337] = 1781,
  [3338] = 1960,
  [3339] = 1815,
  [3340] = 1798,
  [3341] = 1801,
  [3342] = 1803,
  [3343] = 1961,
  [3344] = 1818,
  [3345] = 1814,
  [3346] = 1779,
  [3347] = 1819,
  [3348] = 1641,
  [3349] = 1802,
  [3350] = 1796,
  [3351] = 1797,
  [3352] = 1811,
  [3353] = 1807,
  [3354] = 1816,
  [3355] = 1777,
  [3356] = 1805,
  [3357] = 1806,
  [3358] = 1780,
  [3359] = 1963,
  [3360] = 1782,
  [3361] = 1783,
  [3362] = 3283,
  [3363] = 3255,
  [3364] = 1788,
  [3365] = 1791,
  [3366] = 1802,
  [3367] = 1820,
  [3368] = 1821,
  [3369] = 1822,
  [3370] = 1800,
  [3371] = 1817,
  [3372] = 1888,
  [3373] = 1889,
  [3374] = 1790,
  [3375] = 1813,
  [3376] = 1773,
  [3377] = 1890,
  [3378] = 1795,
  [3379] = 1899,
  [3380] = 1931,
  [3381] = 587,
  [3382] = 1794,
  [3383] = 1932,
  [3384] = 1934,
  [3385] = 1823,
  [3386] = 1784,
  [3387] = 1880,
  [3388] = 1789,
  [3389] = 1927,
  [3390] = 1876,
  [3391] = 1904,
  [3392] = 1812,
  [3393] = 1941,
  [3394] = 1785,
  [3395] = 1786,
  [3396] = 1792,
  [3397] = 1775,
  [3398] = 1784,
  [3399] = 1799,
  [3400] = 1776,
  [3401] = 1784,
  [3402] = 1891,
  [3403] = 1850,
  [3404] = 1787,
  [3405] = 1827,
  [3406] = 1901,
  [3407] = 1906,
  [3408] = 1856,
  [3409] = 1884,
  [3410] = 1913,
  [3411] = 1855,
  [3412] = 1866,
  [3413] = 1829,
  [3414] = 1881,
  [3415] = 1857,
  [3416] = 1951,
  [3417] = 1921,
  [3418] = 1790,
  [3419] = 1922,
  [3420] = 1923,
  [3421] = 1925,
  [3422] = 1851,
  [3423] = 1812,
  [3424] = 1920,
  [3425] = 1929,
  [3426] = 1947,
  [3427] = 1892,
  [3428] = 1854,
  [3429] = 1773,
  [3430] = 1872,
  [3431] = 1862,
  [3432] = 1886,
  [3433] = 1935,
  [3434] = 1894,
  [3435] = 1896,
  [3436] = 1965,
  [3437] = 1959,
  [3438] = 1784,
  [3439] = 1818,
  [3440] = 1814,
  [3441] = 1815,
  [3442] = 1955,
  [3443] = 1964,
  [3444] = 1826,
  [3445] = 1873,
  [3446] = 1937,
  [3447] = 1953,
  [3448] = 1956,
  [3449] = 1773,
  [3450] = 1938,
  [3451] = 1918,
  [3452] = 1912,
  [3453] = 1869,
  [3454] = 1948,
  [3455] = 1794,
  [3456] = 1421,
  [3457] = 1859,
  [3458] = 1847,
  [3459] = 1882,
  [3460] = 1942,
  [3461] = 1946,
  [3462] = 1824,
  [3463] = 1877,
  [3464] = 1930,
  [3465] = 1909,
  [3466] = 1828,
  [3467] = 1831,
  [3468] = 1878,
  [3469] = 1879,
  [3470] = 1900,
  [3471] = 1832,
  [3472] = 1914,
  [3473] = 1917,
  [3474] = 1834,
  [3475] = 1919,
  [3476] = 1813,
  [3477] = 1795,
  [3478] = 1799,
  [3479] = 1816,
  [3480] = 1776,
  [3481] = 1833,
  [3482] = 1848,
  [3483] = 1837,
  [3484] = 1863,
  [3485] = 1865,
  [3486] = 1868,
  [3487] = 1871,
  [3488] = 1864,
  [3489] = 1905,
  [3490] = 1838,
  [3491] = 1908,
  [3492] = 1839,
  [3493] = 1840,
  [3494] = 1933,
  [3495] = 1842,
  [3496] = 1843,
  [3497] = 1846,
  [3498] = 1844,
  [3499] = 1845,
  [3500] = 1830,
  [3501] = 1895,
  [3502] = 1898,
  [3503] = 1858,
  [3504] = 1952,
  [3505] = 1903,
  [3506] = 1957,
  [3507] = 1875,
  [3508] = 1902,
  [3509] = 1907,
  [3510] = 1910,
  [3511] = 1911,
  [3512] = 1915,
  [3513] = 1849,
  [3514] = 1945,
  [3515] = 1940,
  [3516] = 1901,
  [3517] = 1906,
  [3518] = 1856,
  [3519] = 1884,
  [3520] = 1913,
  [3521] = 1855,
  [3522] = 1866,
  [3523] = 1829,
  [3524] = 1881,
  [3525] = 1784,
  [3526] = 1903,
  [3527] = 1951,
  [3528] = 1860,
  [3529] = 1827,
  [3530] = 1922,
  [3531] = 1923,
  [3532] = 1945,
  [3533] = 1784,
  [3534] = 1920,
  [3535] = 1929,
  [3536] = 1947,
  [3537] = 1892,
  [3538] = 1854,
  [3539] = 1886,
  [3540] = 1894,
  [3541] = 1896,
  [3542] = 1965,
  [3543] = 1901,
  [3544] = 1955,
  [3545] = 1964,
  [3546] = 1873,
  [3547] = 1937,
  [3548] = 1906,
  [3549] = 1953,
  [3550] = 1956,
  [3551] = 1856,
  [3552] = 1918,
  [3553] = 1884,
  [3554] = 1912,
  [3555] = 1913,
  [3556] = 1855,
  [3557] = 1866,
  [3558] = 1869,
  [3559] = 1829,
  [3560] = 1859,
  [3561] = 1847,
  [3562] = 1881,
  [3563] = 1882,
  [3564] = 1942,
  [3565] = 1946,
  [3566] = 1877,
  [3567] = 1930,
  [3568] = 1909,
  [3569] = 1860,
  [3570] = 1878,
  [3571] = 1879,
  [3572] = 1900,
  [3573] = 1914,
  [3574] = 1917,
  [3575] = 1919,
  [3576] = 1833,
  [3577] = 1848,
  [3578] = 1863,
  [3579] = 1865,
  [3580] = 1868,
  [3581] = 1871,
  [3582] = 1864,
  [3583] = 1905,
  [3584] = 1908,
  [3585] = 1933,
  [3586] = 1938,
  [3587] = 1939,
  [3588] = 1948,
  [3589] = 1824,
  [3590] = 1828,
  [3591] = 1831,
  [3592] = 1832,
  [3593] = 1834,
  [3594] = 1837,
  [3595] = 1838,
  [3596] = 1839,
  [3597] = 1840,
  [3598] = 1842,
  [3599] = 1843,
  [3600] = 1844,
  [3601] = 1845,
  [3602] = 1846,
  [3603] = 1849,
  [3604] = 1850,
  [3605] = 1851,
  [3606] = 1852,
  [3607] = 1888,
  [3608] = 1889,
  [3609] = 1890,
  [3610] = 1891,
  [3611] = 1895,
  [3612] = 1898,
  [3613] = 1952,
  [3614] = 1957,
  [3615] = 1958,
  [3616] = 1961,
  [3617] = 1963,
  [3618] = 1899,
  [3619] = 1931,
  [3620] = 1932,
  [3621] = 1934,
  [3622] = 1823,
  [3623] = 1784,
  [3624] = 1790,
  [3625] = 1794,
  [3626] = 1880,
  [3627] = 1927,
  [3628] = 1876,
  [3629] = 1904,
  [3630] = 1941,
  [3631] = 1857,
  [3632] = 587,
  [3633] = 1921,
  [3634] = 1925,
  [3635] = 1872,
  [3636] = 1862,
  [3637] = 1935,
  [3638] = 1959,
  [3639] = 1826,
  [3640] = 1830,
  [3641] = 1858,
  [3642] = 1875,
  [3643] = 1902,
  [3644] = 1903,
  [3645] = 1907,
  [3646] = 1910,
  [3647] = 1911,
  [3648] = 1915,
  [3649] = 1916,
  [3650] = 1943,
  [3651] = 1893,
  [3652] = 1962,
  [3653] = 1883,
  [3654] = 1928,
  [3655] = 1861,
  [3656] = 1867,
  [3657] = 1924,
  [3658] = 1926,
  [3659] = 1936,
  [3660] = 1944,
  [3661] = 1950,
  [3662] = 1897,
  [3663] = 1885,
  [3664] = 1949,
  [3665] = 1853,
  [3666] = 1887,
  [3667] = 1825,
  [3668] = 1835,
  [3669] = 1836,
  [3670] = 1841,
  [3671] = 1938,
  [3672] = 1939,
  [3673] = 1948,
  [3674] = 1824,
  [3675] = 1828,
  [3676] = 1831,
  [3677] = 1832,
  [3678] = 1834,
  [3679] = 1837,
  [3680] = 1838,
  [3681] = 1839,
  [3682] = 1840,
  [3683] = 1842,
  [3684] = 1843,
  [3685] = 1844,
  [3686] = 1845,
  [3687] = 1421,
  [3688] = 1951,
  [3689] = 1960,
  [3690] = 1860,
  [3691] = 1903,
  [3692] = 1860,
  [3693] = 1818,
  [3694] = 1814,
  [3695] = 1860,
  [3696] = 1938,
  [3697] = 1939,
  [3698] = 1948,
  [3699] = 1824,
  [3700] = 1828,
  [3701] = 1831,
  [3702] = 1832,
  [3703] = 1834,
  [3704] = 1837,
  [3705] = 1838,
  [3706] = 1839,
  [3707] = 1840,
  [3708] = 1842,
  [3709] = 1843,
  [3710] = 1844,
  [3711] = 1845,
  [3712] = 1901,
  [3713] = 1906,
  [3714] = 1856,
  [3715] = 1884,
  [3716] = 1913,
  [3717] = 1855,
  [3718] = 1866,
  [3719] = 1829,
  [3720] = 1881,
  [3721] = 1951,
  [3722] = 1951,
  [3723] = 1832,
  [3724] = 1951,
  [3725] = 1834,
  [3726] = 3726,
  [3727] = 1837,
  [3728] = 1838,
  [3729] = 1839,
  [3730] = 1840,
  [3731] = 3726,
  [3732] = 1901,
  [3733] = 1906,
  [3734] = 1856,
  [3735] = 1884,
  [3736] = 1913,
  [3737] = 1855,
  [3738] = 1866,
  [3739] = 1829,
  [3740] = 1881,
  [3741] = 3726,
  [3742] = 1842,
  [3743] = 3726,
  [3744] = 3726,
  [3745] = 3745,
  [3746] = 3745,
  [3747] = 3726,
  [3748] = 1843,
  [3749] = 3726,
  [3750] = 3726,
  [3751] = 1844,
  [3752] = 3726,
  [3753] = 1845,
  [3754] = 1903,
  [3755] = 1938,
  [3756] = 3726,
  [3757] = 3745,
  [3758] = 3726,
  [3759] = 3726,
  [3760] = 3726,
  [3761] = 3745,
  [3762] = 1948,
  [3763] = 1824,
  [3764] = 1828,
  [3765] = 1831,
  [3766] = 3745,
  [3767] = 3745,
  [3768] = 3745,
  [3769] = 3745,
  [3770] = 3726,
  [3771] = 3771,
  [3772] = 1881,
  [3773] = 3771,
  [3774] = 3771,
  [3775] = 1901,
  [3776] = 3771,
  [3777] = 1906,
  [3778] = 3771,
  [3779] = 3771,
  [3780] = 1838,
  [3781] = 3771,
  [3782] = 1856,
  [3783] = 1884,
  [3784] = 3771,
  [3785] = 1839,
  [3786] = 1913,
  [3787] = 1855,
  [3788] = 3771,
  [3789] = 1866,
  [3790] = 1842,
  [3791] = 1843,
  [3792] = 1844,
  [3793] = 3771,
  [3794] = 3771,
  [3795] = 1845,
  [3796] = 1829,
  [3797] = 1840,
  [3798] = 1903,
  [3799] = 1938,
  [3800] = 1951,
  [3801] = 1948,
  [3802] = 1824,
  [3803] = 3771,
  [3804] = 1828,
  [3805] = 1831,
  [3806] = 3771,
  [3807] = 1834,
  [3808] = 3771,
  [3809] = 1837,
  [3810] = 1832,
  [3811] = 3811,
  [3812] = 3812,
  [3813] = 3811,
  [3814] = 3811,
  [3815] = 3811,
  [3816] = 3812,
  [3817] = 3811,
  [3818] = 3812,
  [3819] = 3812,
  [3820] = 3811,
  [3821] = 3811,
  [3822] = 3812,
  [3823] = 3812,
  [3824] = 3811,
  [3825] = 3812,
  [3826] = 3812,
  [3827] = 3827,
  [3828] = 3828,
  [3829] = 3829,
  [3830] = 3827,
  [3831] = 3829,
  [3832] = 3828,
  [3833] = 3827,
  [3834] = 3834,
  [3835] = 3835,
  [3836] = 3836,
  [3837] = 3835,
  [3838] = 3829,
  [3839] = 3829,
  [3840] = 3828,
  [3841] = 3834,
  [3842] = 3827,
  [3843] = 3834,
  [3844] = 3836,
  [3845] = 3836,
  [3846] = 3834,
  [3847] = 3835,
  [3848] = 3835,
  [3849] = 3835,
  [3850] = 3829,
  [3851] = 3828,
  [3852] = 3827,
  [3853] = 3834,
  [3854] = 3836,
  [3855] = 3835,
  [3856] = 3829,
  [3857] = 3829,
  [3858] = 3828,
  [3859] = 3828,
  [3860] = 3827,
  [3861] = 3829,
  [3862] = 3828,
  [3863] = 3828,
  [3864] = 3827,
  [3865] = 3834,
  [3866] = 3836,
  [3867] = 3835,
  [3868] = 3834,
  [3869] = 3836,
  [3870] = 3835,
  [3871] = 3827,
  [3872] = 3834,
  [3873] = 3836,
  [3874] = 3836,
  [3875] = 3875,
  [3876] = 3876,
  [3877] = 3877,
  [3878] = 3878,
  [3879] = 3879,
  [3880] = 3880,
  [3881] = 3876,
  [3882] = 3882,
  [3883] = 3883,
  [3884] = 3884,
  [3885] = 3885,
  [3886] = 3879,
  [3887] = 3878,
  [3888] = 3876,
  [3889] = 3880,
  [3890] = 3890,
  [3891] = 3882,
  [3892] = 3883,
  [3893] = 3884,
  [3894] = 3877,
  [3895] = 3885,
  [3896] = 3882,
  [3897] = 3897,
  [3898] = 3884,
  [3899] = 3875,
  [3900] = 3900,
  [3901] = 3879,
  [3902] = 3876,
  [3903] = 3903,
  [3904] = 3890,
  [3905] = 3885,
  [3906] = 3878,
  [3907] = 3883,
  [3908] = 3880,
  [3909] = 3878,
  [3910] = 3882,
  [3911] = 3883,
  [3912] = 3885,
  [3913] = 3903,
  [3914] = 3890,
  [3915] = 3897,
  [3916] = 3884,
  [3917] = 3890,
  [3918] = 3897,
  [3919] = 3875,
  [3920] = 3900,
  [3921] = 3897,
  [3922] = 3900,
  [3923] = 3879,
  [3924] = 3876,
  [3925] = 3903,
  [3926] = 3890,
  [3927] = 3880,
  [3928] = 3885,
  [3929] = 3875,
  [3930] = 3880,
  [3931] = 3880,
  [3932] = 3884,
  [3933] = 3903,
  [3934] = 3878,
  [3935] = 3903,
  [3936] = 3882,
  [3937] = 3882,
  [3938] = 3883,
  [3939] = 3878,
  [3940] = 3882,
  [3941] = 3885,
  [3942] = 3883,
  [3943] = 3878,
  [3944] = 3897,
  [3945] = 3880,
  [3946] = 3883,
  [3947] = 3882,
  [3948] = 3883,
  [3949] = 3875,
  [3950] = 3900,
  [3951] = 3885,
  [3952] = 3879,
  [3953] = 3897,
  [3954] = 3875,
  [3955] = 3884,
  [3956] = 3879,
  [3957] = 3900,
  [3958] = 3900,
  [3959] = 3897,
  [3960] = 3875,
  [3961] = 3900,
  [3962] = 3879,
  [3963] = 3876,
  [3964] = 3903,
  [3965] = 3890,
  [3966] = 3876,
  [3967] = 3903,
  [3968] = 3890,
  [3969] = 3879,
  [3970] = 3876,
  [3971] = 3884,
  [3972] = 3903,
  [3973] = 3890,
  [3974] = 3897,
  [3975] = 3875,
  [3976] = 3900,
  [3977] = 3880,
  [3978] = 3885,
  [3979] = 3884,
  [3980] = 3878,
  [3981] = 3981,
  [3982] = 3982,
  [3983] = 3983,
  [3984] = 3983,
  [3985] = 3983,
  [3986] = 3983,
  [3987] = 3983,
  [3988] = 3983,
  [3989] = 3983,
  [3990] = 3983,
  [3991] = 3991,
  [3992] = 3992,
  [3993] = 1794,
  [3994] = 1790,
  [3995] = 3995,
  [3996] = 3996,
  [3997] = 3997,
  [3998] = 3995,
  [3999] = 3997,
  [4000] = 3995,
  [4001] = 3995,
  [4002] = 3995,
  [4003] = 3995,
  [4004] = 3997,
  [4005] = 3997,
  [4006] = 3996,
  [4007] = 3995,
  [4008] = 3995,
  [4009] = 3997,
  [4010] = 3996,
  [4011] = 3997,
  [4012] = 3997,
  [4013] = 3997,
  [4014] = 4014,
  [4015] = 4014,
  [4016] = 4014,
  [4017] = 4014,
  [4018] = 4018,
  [4019] = 4019,
  [4020] = 4020,
  [4021] = 4021,
  [4022] = 4022,
  [4023] = 4023,
  [4024] = 4024,
  [4025] = 4023,
  [4026] = 4023,
  [4027] = 4027,
  [4028] = 4023,
  [4029] = 4024,
  [4030] = 4030,
  [4031] = 1134,
  [4032] = 4023,
  [4033] = 1135,
  [4034] = 1135,
  [4035] = 4035,
  [4036] = 4036,
  [4037] = 1134,
  [4038] = 1133,
  [4039] = 1133,
  [4040] = 4023,
  [4041] = 4023,
  [4042] = 4023,
  [4043] = 4035,
  [4044] = 4044,
  [4045] = 4027,
  [4046] = 4022,
  [4047] = 4022,
  [4048] = 4048,
  [4049] = 4048,
  [4050] = 1135,
  [4051] = 4027,
  [4052] = 4048,
  [4053] = 4048,
  [4054] = 4054,
  [4055] = 4055,
  [4056] = 4048,
  [4057] = 1133,
  [4058] = 4048,
  [4059] = 4055,
  [4060] = 4048,
  [4061] = 4061,
  [4062] = 4048,
  [4063] = 4048,
  [4064] = 4061,
  [4065] = 4048,
  [4066] = 4048,
  [4067] = 4048,
  [4068] = 1140,
  [4069] = 4048,
  [4070] = 1139,
  [4071] = 4048,
  [4072] = 4048,
  [4073] = 4048,
  [4074] = 4048,
  [4075] = 4061,
  [4076] = 4048,
  [4077] = 4048,
  [4078] = 4024,
  [4079] = 4079,
  [4080] = 4048,
  [4081] = 1134,
  [4082] = 4048,
  [4083] = 1135,
  [4084] = 4048,
  [4085] = 1141,
  [4086] = 1137,
  [4087] = 4087,
  [4088] = 1134,
  [4089] = 4089,
  [4090] = 1136,
  [4091] = 4091,
  [4092] = 4089,
  [4093] = 4091,
  [4094] = 4048,
  [4095] = 1141,
  [4096] = 4055,
  [4097] = 4091,
  [4098] = 4098,
  [4099] = 4048,
  [4100] = 1135,
  [4101] = 4061,
  [4102] = 1137,
  [4103] = 4091,
  [4104] = 1139,
  [4105] = 4061,
  [4106] = 4048,
  [4107] = 4054,
  [4108] = 1133,
  [4109] = 4087,
  [4110] = 4055,
  [4111] = 4048,
  [4112] = 4055,
  [4113] = 4048,
  [4114] = 1133,
  [4115] = 4055,
  [4116] = 1136,
  [4117] = 4061,
  [4118] = 4061,
  [4119] = 4091,
  [4120] = 4061,
  [4121] = 4079,
  [4122] = 4048,
  [4123] = 4035,
  [4124] = 4048,
  [4125] = 4079,
  [4126] = 4079,
  [4127] = 4091,
  [4128] = 4055,
  [4129] = 4048,
  [4130] = 4091,
  [4131] = 4091,
  [4132] = 4048,
  [4133] = 4048,
  [4134] = 4055,
  [4135] = 1134,
  [4136] = 4136,
  [4137] = 4048,
  [4138] = 1140,
  [4139] = 4139,
  [4140] = 4140,
  [4141] = 4141,
  [4142] = 4142,
  [4143] = 4141,
  [4144] = 4144,
  [4145] = 4145,
  [4146] = 4145,
  [4147] = 4141,
  [4148] = 4148,
  [4149] = 4035,
  [4150] = 4150,
  [4151] = 4024,
  [4152] = 4152,
  [4153] = 4153,
  [4154] = 4139,
  [4155] = 4155,
  [4156] = 4156,
  [4157] = 4141,
  [4158] = 4027,
  [4159] = 4140,
  [4160] = 4160,
  [4161] = 4141,
  [4162] = 4162,
  [4163] = 4150,
  [4164] = 4164,
  [4165] = 1141,
  [4166] = 1794,
  [4167] = 4167,
  [4168] = 4054,
  [4169] = 4169,
  [4170] = 4156,
  [4171] = 4148,
  [4172] = 4172,
  [4173] = 4155,
  [4174] = 4145,
  [4175] = 4150,
  [4176] = 4145,
  [4177] = 4152,
  [4178] = 4164,
  [4179] = 4167,
  [4180] = 4141,
  [4181] = 4141,
  [4182] = 4182,
  [4183] = 1134,
  [4184] = 4184,
  [4185] = 4141,
  [4186] = 4140,
  [4187] = 4164,
  [4188] = 4144,
  [4189] = 4148,
  [4190] = 4160,
  [4191] = 4148,
  [4192] = 1140,
  [4193] = 4139,
  [4194] = 4155,
  [4195] = 4182,
  [4196] = 4140,
  [4197] = 4145,
  [4198] = 4160,
  [4199] = 4184,
  [4200] = 4141,
  [4201] = 4140,
  [4202] = 4202,
  [4203] = 4203,
  [4204] = 4155,
  [4205] = 4182,
  [4206] = 4145,
  [4207] = 4160,
  [4208] = 4152,
  [4209] = 4209,
  [4210] = 4087,
  [4211] = 4141,
  [4212] = 4212,
  [4213] = 4141,
  [4214] = 1135,
  [4215] = 4144,
  [4216] = 4148,
  [4217] = 4139,
  [4218] = 4140,
  [4219] = 1133,
  [4220] = 4160,
  [4221] = 4182,
  [4222] = 1139,
  [4223] = 4150,
  [4224] = 4164,
  [4225] = 4167,
  [4226] = 4141,
  [4227] = 4203,
  [4228] = 4145,
  [4229] = 4229,
  [4230] = 4152,
  [4231] = 4141,
  [4232] = 4182,
  [4233] = 4144,
  [4234] = 4182,
  [4235] = 4235,
  [4236] = 4141,
  [4237] = 4167,
  [4238] = 4152,
  [4239] = 4152,
  [4240] = 4148,
  [4241] = 4209,
  [4242] = 4141,
  [4243] = 4139,
  [4244] = 1137,
  [4245] = 4155,
  [4246] = 4229,
  [4247] = 4141,
  [4248] = 4152,
  [4249] = 4140,
  [4250] = 4160,
  [4251] = 1790,
  [4252] = 4141,
  [4253] = 4152,
  [4254] = 4212,
  [4255] = 4160,
  [4256] = 4169,
  [4257] = 4172,
  [4258] = 4202,
  [4259] = 4145,
  [4260] = 4141,
  [4261] = 4155,
  [4262] = 4139,
  [4263] = 4150,
  [4264] = 4150,
  [4265] = 1134,
  [4266] = 4164,
  [4267] = 4167,
  [4268] = 4167,
  [4269] = 4155,
  [4270] = 4164,
  [4271] = 4141,
  [4272] = 4150,
  [4273] = 4141,
  [4274] = 4144,
  [4275] = 4141,
  [4276] = 4148,
  [4277] = 4144,
  [4278] = 4139,
  [4279] = 4148,
  [4280] = 4140,
  [4281] = 4160,
  [4282] = 4142,
  [4283] = 4150,
  [4284] = 4139,
  [4285] = 4141,
  [4286] = 4164,
  [4287] = 4167,
  [4288] = 1135,
  [4289] = 4145,
  [4290] = 4182,
  [4291] = 4167,
  [4292] = 1136,
  [4293] = 1133,
  [4294] = 4141,
  [4295] = 4164,
  [4296] = 4141,
  [4297] = 4155,
  [4298] = 4144,
  [4299] = 4182,
  [4300] = 4162,
  [4301] = 4153,
  [4302] = 4144,
  [4303] = 4303,
  [4304] = 4304,
  [4305] = 4305,
  [4306] = 4306,
  [4307] = 4307,
  [4308] = 4162,
  [4309] = 4153,
  [4310] = 4142,
  [4311] = 4156,
  [4312] = 4312,
  [4313] = 4313,
  [4314] = 4229,
  [4315] = 4315,
  [4316] = 4306,
  [4317] = 4307,
  [4318] = 4303,
  [4319] = 4312,
  [4320] = 4313,
  [4321] = 4306,
  [4322] = 4307,
  [4323] = 4306,
  [4324] = 4307,
  [4325] = 1141,
  [4326] = 1135,
  [4327] = 4312,
  [4328] = 4313,
  [4329] = 1139,
  [4330] = 1140,
  [4331] = 1137,
  [4332] = 4332,
  [4333] = 4305,
  [4334] = 4306,
  [4335] = 4307,
  [4336] = 4312,
  [4337] = 4313,
  [4338] = 4312,
  [4339] = 4313,
  [4340] = 4340,
  [4341] = 4304,
  [4342] = 4306,
  [4343] = 4307,
  [4344] = 4312,
  [4345] = 4313,
  [4346] = 4346,
  [4347] = 4347,
  [4348] = 4348,
  [4349] = 4349,
  [4350] = 4350,
  [4351] = 4351,
  [4352] = 4352,
  [4353] = 4351,
  [4354] = 4315,
  [4355] = 4303,
  [4356] = 4312,
  [4357] = 4313,
  [4358] = 1133,
  [4359] = 4306,
  [4360] = 4307,
  [4361] = 1134,
  [4362] = 4352,
  [4363] = 4332,
  [4364] = 4305,
  [4365] = 4312,
  [4366] = 4313,
  [4367] = 4351,
  [4368] = 4340,
  [4369] = 4304,
  [4370] = 1136,
  [4371] = 4087,
  [4372] = 4054,
  [4373] = 4306,
  [4374] = 4350,
  [4375] = 4346,
  [4376] = 4347,
  [4377] = 4346,
  [4378] = 4347,
  [4379] = 4348,
  [4380] = 4349,
  [4381] = 4307,
  [4382] = 4202,
  [4383] = 4203,
  [4384] = 4209,
  [4385] = 4212,
  [4386] = 4348,
  [4387] = 4315,
  [4388] = 4349,
  [4389] = 4389,
  [4390] = 4169,
  [4391] = 4172,
  [4392] = 4352,
  [4393] = 4340,
  [4394] = 4332,
  [4395] = 4395,
  [4396] = 4396,
  [4397] = 4397,
  [4398] = 4398,
  [4399] = 1140,
  [4400] = 4400,
  [4401] = 4401,
  [4402] = 4402,
  [4403] = 4403,
  [4404] = 4404,
  [4405] = 4405,
  [4406] = 4406,
  [4407] = 4407,
  [4408] = 4408,
  [4409] = 4162,
  [4410] = 4410,
  [4411] = 4169,
  [4412] = 4412,
  [4413] = 4413,
  [4414] = 4414,
  [4415] = 4415,
  [4416] = 4153,
  [4417] = 4417,
  [4418] = 4418,
  [4419] = 4419,
  [4420] = 1773,
  [4421] = 4421,
  [4422] = 4405,
  [4423] = 4423,
  [4424] = 4400,
  [4425] = 1139,
  [4426] = 4426,
  [4427] = 4403,
  [4428] = 4412,
  [4429] = 4397,
  [4430] = 4398,
  [4431] = 4142,
  [4432] = 1135,
  [4433] = 4156,
  [4434] = 4404,
  [4435] = 4413,
  [4436] = 4414,
  [4437] = 4403,
  [4438] = 4404,
  [4439] = 4395,
  [4440] = 4440,
  [4441] = 4415,
  [4442] = 4419,
  [4443] = 4423,
  [4444] = 4396,
  [4445] = 4402,
  [4446] = 1135,
  [4447] = 4410,
  [4448] = 1133,
  [4449] = 1773,
  [4450] = 4398,
  [4451] = 4229,
  [4452] = 1134,
  [4453] = 4172,
  [4454] = 1133,
  [4455] = 1134,
  [4456] = 1141,
  [4457] = 1137,
  [4458] = 4202,
  [4459] = 4203,
  [4460] = 1140,
  [4461] = 1139,
  [4462] = 1141,
  [4463] = 4209,
  [4464] = 4212,
  [4465] = 1137,
  [4466] = 4421,
  [4467] = 4405,
  [4468] = 4423,
  [4469] = 4469,
  [4470] = 4407,
  [4471] = 4401,
  [4472] = 4406,
  [4473] = 4440,
  [4474] = 4474,
  [4475] = 4417,
  [4476] = 4474,
  [4477] = 1136,
  [4478] = 4397,
  [4479] = 4389,
  [4480] = 4421,
  [4481] = 4418,
  [4482] = 4426,
  [4483] = 1136,
  [4484] = 4408,
  [4485] = 1738,
  [4486] = 4406,
  [4487] = 1133,
  [4488] = 4488,
  [4489] = 4489,
  [4490] = 4489,
  [4491] = 4491,
  [4492] = 4492,
  [4493] = 4401,
  [4494] = 1736,
  [4495] = 1134,
  [4496] = 4491,
  [4497] = 4497,
  [4498] = 4491,
  [4499] = 4406,
  [4500] = 4491,
  [4501] = 4410,
  [4502] = 4502,
  [4503] = 1794,
  [4504] = 4412,
  [4505] = 4413,
  [4506] = 4414,
  [4507] = 4415,
  [4508] = 4419,
  [4509] = 1136,
  [4510] = 4491,
  [4511] = 4511,
  [4512] = 4440,
  [4513] = 4400,
  [4514] = 1738,
  [4515] = 4019,
  [4516] = 4491,
  [4517] = 4492,
  [4518] = 1737,
  [4519] = 4395,
  [4520] = 4491,
  [4521] = 1773,
  [4522] = 4491,
  [4523] = 4523,
  [4524] = 4408,
  [4525] = 4417,
  [4526] = 4489,
  [4527] = 4418,
  [4528] = 4489,
  [4529] = 4529,
  [4530] = 4530,
  [4531] = 4502,
  [4532] = 4532,
  [4533] = 4523,
  [4534] = 4396,
  [4535] = 4402,
  [4536] = 4407,
  [4537] = 1141,
  [4538] = 1736,
  [4539] = 1737,
  [4540] = 4540,
  [4541] = 1137,
  [4542] = 4542,
  [4543] = 4543,
  [4544] = 4491,
  [4545] = 1140,
  [4546] = 4546,
  [4547] = 1139,
  [4548] = 4491,
  [4549] = 1790,
  [4550] = 4523,
  [4551] = 4551,
  [4552] = 1135,
  [4553] = 4410,
  [4554] = 1737,
  [4555] = 1790,
  [4556] = 4406,
  [4557] = 4415,
  [4558] = 4395,
  [4559] = 4417,
  [4560] = 4396,
  [4561] = 4402,
  [4562] = 4407,
  [4563] = 4408,
  [4564] = 4419,
  [4565] = 4418,
  [4566] = 1134,
  [4567] = 1794,
  [4568] = 1133,
  [4569] = 1794,
  [4570] = 4570,
  [4571] = 1736,
  [4572] = 4572,
  [4573] = 4573,
  [4574] = 4414,
  [4575] = 1790,
  [4576] = 4401,
  [4577] = 1136,
  [4578] = 1773,
  [4579] = 1794,
  [4580] = 1139,
  [4581] = 4570,
  [4582] = 4570,
  [4583] = 4583,
  [4584] = 1790,
  [4585] = 4400,
  [4586] = 4412,
  [4587] = 1141,
  [4588] = 4540,
  [4589] = 4532,
  [4590] = 1137,
  [4591] = 4413,
  [4592] = 1738,
  [4593] = 1135,
  [4594] = 4440,
  [4595] = 1140,
  [4596] = 4596,
  [4597] = 4597,
  [4598] = 4598,
  [4599] = 4599,
  [4600] = 4600,
  [4601] = 4601,
  [4602] = 4602,
  [4603] = 4603,
  [4604] = 4604,
  [4605] = 4599,
  [4606] = 4606,
  [4607] = 4607,
  [4608] = 4596,
  [4609] = 4609,
  [4610] = 4599,
  [4611] = 4611,
  [4612] = 4609,
  [4613] = 4596,
  [4614] = 4614,
  [4615] = 4596,
  [4616] = 4616,
  [4617] = 4616,
  [4618] = 4604,
  [4619] = 4619,
  [4620] = 4597,
  [4621] = 4611,
  [4622] = 4604,
  [4623] = 4596,
  [4624] = 4598,
  [4625] = 4598,
  [4626] = 4626,
  [4627] = 4627,
  [4628] = 4598,
  [4629] = 4599,
  [4630] = 4630,
  [4631] = 4600,
  [4632] = 4607,
  [4633] = 4597,
  [4634] = 4616,
  [4635] = 4604,
  [4636] = 4614,
  [4637] = 4609,
  [4638] = 4601,
  [4639] = 4598,
  [4640] = 4611,
  [4641] = 4607,
  [4642] = 4599,
  [4643] = 4596,
  [4644] = 4607,
  [4645] = 4606,
  [4646] = 4609,
  [4647] = 4647,
  [4648] = 4611,
  [4649] = 1742,
  [4650] = 4607,
  [4651] = 4598,
  [4652] = 4609,
  [4653] = 4614,
  [4654] = 4626,
  [4655] = 4602,
  [4656] = 4611,
  [4657] = 4626,
  [4658] = 4627,
  [4659] = 4596,
  [4660] = 4611,
  [4661] = 4598,
  [4662] = 4607,
  [4663] = 4627,
  [4664] = 4599,
  [4665] = 4611,
  [4666] = 4616,
  [4667] = 4604,
  [4668] = 4607,
  [4669] = 4611,
  [4670] = 4606,
  [4671] = 4647,
  [4672] = 4614,
  [4673] = 4609,
  [4674] = 4614,
  [4675] = 4616,
  [4676] = 4614,
  [4677] = 1742,
  [4678] = 4630,
  [4679] = 4600,
  [4680] = 1744,
  [4681] = 4596,
  [4682] = 4614,
  [4683] = 4607,
  [4684] = 4616,
  [4685] = 4604,
  [4686] = 4598,
  [4687] = 4616,
  [4688] = 4630,
  [4689] = 4616,
  [4690] = 1744,
  [4691] = 4604,
  [4692] = 4619,
  [4693] = 4604,
  [4694] = 4599,
  [4695] = 4609,
  [4696] = 4599,
  [4697] = 4601,
  [4698] = 4614,
  [4699] = 4609,
  [4700] = 4619,
  [4701] = 4602,
  [4702] = 4603,
  [4703] = 4603,
  [4704] = 4647,
  [4705] = 4705,
  [4706] = 4706,
  [4707] = 4019,
  [4708] = 4708,
  [4709] = 4709,
  [4710] = 4710,
  [4711] = 1769,
  [4712] = 4712,
  [4713] = 4713,
  [4714] = 4714,
  [4715] = 4715,
  [4716] = 4705,
  [4717] = 4717,
  [4718] = 4718,
  [4719] = 1762,
  [4720] = 4720,
  [4721] = 4706,
  [4722] = 4722,
  [4723] = 4722,
  [4724] = 4724,
  [4725] = 4724,
  [4726] = 4705,
  [4727] = 4717,
  [4728] = 4717,
  [4729] = 4714,
  [4730] = 4708,
  [4731] = 4731,
  [4732] = 4710,
  [4733] = 4714,
  [4734] = 4734,
  [4735] = 4735,
  [4736] = 4720,
  [4737] = 4737,
  [4738] = 4724,
  [4739] = 4739,
  [4740] = 4706,
  [4741] = 4724,
  [4742] = 4714,
  [4743] = 4720,
  [4744] = 4731,
  [4745] = 4745,
  [4746] = 4706,
  [4747] = 4715,
  [4748] = 4748,
  [4749] = 4749,
  [4750] = 4749,
  [4751] = 4731,
  [4752] = 4752,
  [4753] = 4715,
  [4754] = 4754,
  [4755] = 4748,
  [4756] = 4731,
  [4757] = 4757,
  [4758] = 4709,
  [4759] = 4759,
  [4760] = 4760,
  [4761] = 4761,
  [4762] = 4739,
  [4763] = 4745,
  [4764] = 4709,
  [4765] = 4765,
  [4766] = 4766,
  [4767] = 4739,
  [4768] = 4768,
  [4769] = 4712,
  [4770] = 1766,
  [4771] = 4771,
  [4772] = 4713,
  [4773] = 4715,
  [4774] = 4731,
  [4775] = 4724,
  [4776] = 4754,
  [4777] = 4713,
  [4778] = 4717,
  [4779] = 4705,
  [4780] = 4717,
  [4781] = 4720,
  [4782] = 4782,
  [4783] = 4722,
  [4784] = 4720,
  [4785] = 4705,
  [4786] = 4757,
  [4787] = 4722,
  [4788] = 4771,
  [4789] = 4731,
  [4790] = 4759,
  [4791] = 4791,
  [4792] = 4760,
  [4793] = 4793,
  [4794] = 4794,
  [4795] = 4795,
  [4796] = 4713,
  [4797] = 4797,
  [4798] = 4713,
  [4799] = 4731,
  [4800] = 4761,
  [4801] = 4705,
  [4802] = 4717,
  [4803] = 4720,
  [4804] = 4722,
  [4805] = 4805,
  [4806] = 4806,
  [4807] = 4708,
  [4808] = 4722,
  [4809] = 4710,
  [4810] = 4708,
  [4811] = 4714,
  [4812] = 4710,
  [4813] = 4706,
  [4814] = 4714,
  [4815] = 4724,
  [4816] = 4735,
  [4817] = 4717,
  [4818] = 4739,
  [4819] = 4706,
  [4820] = 4820,
  [4821] = 4724,
  [4822] = 4749,
  [4823] = 4708,
  [4824] = 4722,
  [4825] = 4710,
  [4826] = 4714,
  [4827] = 4754,
  [4828] = 4806,
  [4829] = 4706,
  [4830] = 4735,
  [4831] = 4724,
  [4832] = 4745,
  [4833] = 4706,
  [4834] = 4749,
  [4835] = 4754,
  [4836] = 4715,
  [4837] = 4709,
  [4838] = 4749,
  [4839] = 4754,
  [4840] = 4766,
  [4841] = 4768,
  [4842] = 4757,
  [4843] = 4759,
  [4844] = 4749,
  [4845] = 4754,
  [4846] = 4760,
  [4847] = 4720,
  [4848] = 4757,
  [4849] = 4757,
  [4850] = 4757,
  [4851] = 4757,
  [4852] = 4759,
  [4853] = 4768,
  [4854] = 4749,
  [4855] = 4760,
  [4856] = 4759,
  [4857] = 4759,
  [4858] = 4754,
  [4859] = 4765,
  [4860] = 4860,
  [4861] = 4761,
  [4862] = 4862,
  [4863] = 4860,
  [4864] = 4862,
  [4865] = 4748,
  [4866] = 4739,
  [4867] = 4766,
  [4868] = 4760,
  [4869] = 4869,
  [4870] = 4760,
  [4871] = 4749,
  [4872] = 4795,
  [4873] = 4745,
  [4874] = 4782,
  [4875] = 4737,
  [4876] = 4734,
  [4877] = 4718,
  [4878] = 4709,
  [4879] = 4708,
  [4880] = 4754,
  [4881] = 4766,
  [4882] = 4765,
  [4883] = 4705,
  [4884] = 4820,
  [4885] = 4768,
  [4886] = 4782,
  [4887] = 4761,
  [4888] = 4888,
  [4889] = 4820,
  [4890] = 1762,
  [4891] = 4739,
  [4892] = 4795,
  [4893] = 4737,
  [4894] = 4745,
  [4895] = 4768,
  [4896] = 4713,
  [4897] = 4709,
  [4898] = 4860,
  [4899] = 4862,
  [4900] = 4752,
  [4901] = 4765,
  [4902] = 4734,
  [4903] = 4766,
  [4904] = 4768,
  [4905] = 4718,
  [4906] = 4712,
  [4907] = 4712,
  [4908] = 4766,
  [4909] = 4715,
  [4910] = 4768,
  [4911] = 4712,
  [4912] = 4757,
  [4913] = 1766,
  [4914] = 4759,
  [4915] = 4712,
  [4916] = 4916,
  [4917] = 4715,
  [4918] = 4761,
  [4919] = 4760,
  [4920] = 4761,
  [4921] = 4921,
  [4922] = 4761,
  [4923] = 4705,
  [4924] = 4739,
  [4925] = 4715,
  [4926] = 4745,
  [4927] = 4709,
  [4928] = 4717,
  [4929] = 4765,
  [4930] = 4930,
  [4931] = 4791,
  [4932] = 4766,
  [4933] = 4933,
  [4934] = 4768,
  [4935] = 4935,
  [4936] = 4713,
  [4937] = 1742,
  [4938] = 4938,
  [4939] = 4765,
  [4940] = 4760,
  [4941] = 4712,
  [4942] = 4722,
  [4943] = 4720,
  [4944] = 4708,
  [4945] = 4712,
  [4946] = 4806,
  [4947] = 4731,
  [4948] = 1744,
  [4949] = 4708,
  [4950] = 4739,
  [4951] = 4791,
  [4952] = 4710,
  [4953] = 4710,
  [4954] = 1632,
  [4955] = 4714,
  [4956] = 4765,
  [4957] = 4745,
  [4958] = 4805,
  [4959] = 4709,
  [4960] = 4752,
  [4961] = 4713,
  [4962] = 4710,
  [4963] = 4805,
  [4964] = 4021,
  [4965] = 4745,
  [4966] = 4761,
  [4967] = 4765,
  [4968] = 4766,
  [4969] = 4020,
  [4970] = 4771,
  [4971] = 1769,
  [4972] = 4759,
  [4973] = 4973,
  [4974] = 4974,
  [4975] = 4975,
  [4976] = 4974,
  [4977] = 4977,
  [4978] = 4978,
  [4979] = 4979,
  [4980] = 4980,
  [4981] = 4973,
  [4982] = 4021,
  [4983] = 4983,
  [4984] = 4984,
  [4985] = 4980,
  [4986] = 4986,
  [4987] = 4987,
  [4988] = 4988,
  [4989] = 4989,
  [4990] = 4984,
  [4991] = 4974,
  [4992] = 4977,
  [4993] = 4978,
  [4994] = 4979,
  [4995] = 4980,
  [4996] = 4986,
  [4997] = 4987,
  [4998] = 4984,
  [4999] = 4986,
  [5000] = 4987,
  [5001] = 5001,
  [5002] = 5001,
  [5003] = 5003,
  [5004] = 5004,
  [5005] = 5005,
  [5006] = 5006,
  [5007] = 4973,
  [5008] = 4988,
  [5009] = 4989,
  [5010] = 5003,
  [5011] = 5004,
  [5012] = 5005,
  [5013] = 1794,
  [5014] = 5001,
  [5015] = 5003,
  [5016] = 5004,
  [5017] = 5005,
  [5018] = 5006,
  [5019] = 4973,
  [5020] = 4988,
  [5021] = 5006,
  [5022] = 4989,
  [5023] = 4973,
  [5024] = 4988,
  [5025] = 4989,
  [5026] = 4984,
  [5027] = 4978,
  [5028] = 4986,
  [5029] = 4987,
  [5030] = 5030,
  [5031] = 4973,
  [5032] = 4984,
  [5033] = 4988,
  [5034] = 4983,
  [5035] = 5035,
  [5036] = 5030,
  [5037] = 5006,
  [5038] = 5001,
  [5039] = 4983,
  [5040] = 5040,
  [5041] = 1790,
  [5042] = 5003,
  [5043] = 5004,
  [5044] = 5044,
  [5045] = 5005,
  [5046] = 4983,
  [5047] = 4977,
  [5048] = 4020,
  [5049] = 4019,
  [5050] = 5050,
  [5051] = 4020,
  [5052] = 5006,
  [5053] = 4988,
  [5054] = 4989,
  [5055] = 5055,
  [5056] = 4020,
  [5057] = 4989,
  [5058] = 1766,
  [5059] = 5059,
  [5060] = 4019,
  [5061] = 4974,
  [5062] = 4977,
  [5063] = 4979,
  [5064] = 4978,
  [5065] = 4979,
  [5066] = 4980,
  [5067] = 4983,
  [5068] = 4984,
  [5069] = 4986,
  [5070] = 4986,
  [5071] = 4987,
  [5072] = 5044,
  [5073] = 4989,
  [5074] = 4987,
  [5075] = 5001,
  [5076] = 5003,
  [5077] = 5077,
  [5078] = 5078,
  [5079] = 5004,
  [5080] = 5005,
  [5081] = 5006,
  [5082] = 4973,
  [5083] = 4988,
  [5084] = 4989,
  [5085] = 5077,
  [5086] = 4980,
  [5087] = 4974,
  [5088] = 4977,
  [5089] = 4978,
  [5090] = 4979,
  [5091] = 4980,
  [5092] = 5077,
  [5093] = 5044,
  [5094] = 4021,
  [5095] = 5095,
  [5096] = 5077,
  [5097] = 5078,
  [5098] = 4984,
  [5099] = 4986,
  [5100] = 4987,
  [5101] = 1794,
  [5102] = 5044,
  [5103] = 4983,
  [5104] = 5044,
  [5105] = 4983,
  [5106] = 5001,
  [5107] = 5003,
  [5108] = 5004,
  [5109] = 4974,
  [5110] = 4977,
  [5111] = 1790,
  [5112] = 4978,
  [5113] = 5005,
  [5114] = 5044,
  [5115] = 5115,
  [5116] = 4979,
  [5117] = 1762,
  [5118] = 4980,
  [5119] = 5044,
  [5120] = 4974,
  [5121] = 4977,
  [5122] = 5006,
  [5123] = 4019,
  [5124] = 4974,
  [5125] = 4977,
  [5126] = 4978,
  [5127] = 4979,
  [5128] = 4980,
  [5129] = 5129,
  [5130] = 4973,
  [5131] = 4988,
  [5132] = 4989,
  [5133] = 4978,
  [5134] = 4984,
  [5135] = 4986,
  [5136] = 4987,
  [5137] = 5078,
  [5138] = 5035,
  [5139] = 5095,
  [5140] = 5001,
  [5141] = 4979,
  [5142] = 5001,
  [5143] = 5044,
  [5144] = 5003,
  [5145] = 4983,
  [5146] = 5003,
  [5147] = 1769,
  [5148] = 5004,
  [5149] = 5035,
  [5150] = 5095,
  [5151] = 5005,
  [5152] = 4021,
  [5153] = 5004,
  [5154] = 5006,
  [5155] = 4973,
  [5156] = 4988,
  [5157] = 5005,
  [5158] = 5158,
  [5159] = 5159,
  [5160] = 5160,
  [5161] = 5161,
  [5162] = 5162,
  [5163] = 5161,
  [5164] = 1137,
  [5165] = 5165,
  [5166] = 5159,
  [5167] = 5167,
  [5168] = 5168,
  [5169] = 5169,
  [5170] = 5170,
  [5171] = 5171,
  [5172] = 5172,
  [5173] = 5173,
  [5174] = 5174,
  [5175] = 5175,
  [5176] = 1133,
  [5177] = 5161,
  [5178] = 5178,
  [5179] = 5170,
  [5180] = 5162,
  [5181] = 5181,
  [5182] = 5182,
  [5183] = 5183,
  [5184] = 5184,
  [5185] = 5185,
  [5186] = 5186,
  [5187] = 5184,
  [5188] = 5171,
  [5189] = 5172,
  [5190] = 5170,
  [5191] = 5167,
  [5192] = 5168,
  [5193] = 5193,
  [5194] = 5194,
  [5195] = 5171,
  [5196] = 5184,
  [5197] = 5167,
  [5198] = 5162,
  [5199] = 5182,
  [5200] = 5185,
  [5201] = 5186,
  [5202] = 5173,
  [5203] = 5168,
  [5204] = 5178,
  [5205] = 5205,
  [5206] = 5169,
  [5207] = 5170,
  [5208] = 5161,
  [5209] = 5209,
  [5210] = 5205,
  [5211] = 5194,
  [5212] = 5160,
  [5213] = 5213,
  [5214] = 5213,
  [5215] = 5170,
  [5216] = 5162,
  [5217] = 5182,
  [5218] = 5185,
  [5219] = 5186,
  [5220] = 5182,
  [5221] = 5169,
  [5222] = 5209,
  [5223] = 5175,
  [5224] = 5224,
  [5225] = 5169,
  [5226] = 5174,
  [5227] = 5161,
  [5228] = 5224,
  [5229] = 5229,
  [5230] = 5165,
  [5231] = 5162,
  [5232] = 5182,
  [5233] = 5185,
  [5234] = 5186,
  [5235] = 5172,
  [5236] = 5169,
  [5237] = 5173,
  [5238] = 5193,
  [5239] = 5213,
  [5240] = 5184,
  [5241] = 1141,
  [5242] = 5167,
  [5243] = 5184,
  [5244] = 5184,
  [5245] = 5175,
  [5246] = 5193,
  [5247] = 5178,
  [5248] = 5055,
  [5249] = 5178,
  [5250] = 5250,
  [5251] = 1135,
  [5252] = 5182,
  [5253] = 5185,
  [5254] = 5186,
  [5255] = 5167,
  [5256] = 5181,
  [5257] = 5167,
  [5258] = 5168,
  [5259] = 5168,
  [5260] = 5169,
  [5261] = 5183,
  [5262] = 5205,
  [5263] = 5167,
  [5264] = 5168,
  [5265] = 5169,
  [5266] = 1139,
  [5267] = 5162,
  [5268] = 5182,
  [5269] = 5185,
  [5270] = 5186,
  [5271] = 5159,
  [5272] = 5168,
  [5273] = 5160,
  [5274] = 5184,
  [5275] = 5185,
  [5276] = 5276,
  [5277] = 4410,
  [5278] = 5193,
  [5279] = 5178,
  [5280] = 5194,
  [5281] = 5161,
  [5282] = 5181,
  [5283] = 5250,
  [5284] = 5178,
  [5285] = 5285,
  [5286] = 1136,
  [5287] = 5171,
  [5288] = 5229,
  [5289] = 5184,
  [5290] = 5183,
  [5291] = 5170,
  [5292] = 5165,
  [5293] = 5178,
  [5294] = 5174,
  [5295] = 5209,
  [5296] = 5167,
  [5297] = 5297,
  [5298] = 5168,
  [5299] = 5299,
  [5300] = 5161,
  [5301] = 5276,
  [5302] = 5194,
  [5303] = 5224,
  [5304] = 5170,
  [5305] = 5184,
  [5306] = 5205,
  [5307] = 5169,
  [5308] = 5170,
  [5309] = 5178,
  [5310] = 5184,
  [5311] = 5184,
  [5312] = 1134,
  [5313] = 5184,
  [5314] = 5276,
  [5315] = 5186,
  [5316] = 4530,
  [5317] = 5161,
  [5318] = 5297,
  [5319] = 5276,
  [5320] = 1140,
  [5321] = 5276,
  [5322] = 5229,
  [5323] = 5276,
  [5324] = 5276,
  [5325] = 5162,
  [5326] = 5276,
  [5327] = 5182,
  [5328] = 5185,
  [5329] = 5186,
  [5330] = 5224,
  [5331] = 5162,
  [5332] = 5332,
  [5333] = 5333,
  [5334] = 5334,
  [5335] = 5333,
  [5336] = 5336,
  [5337] = 5337,
  [5338] = 5338,
  [5339] = 5339,
  [5340] = 5340,
  [5341] = 5341,
  [5342] = 5342,
  [5343] = 5336,
  [5344] = 5344,
  [5345] = 5345,
  [5346] = 5346,
  [5347] = 5347,
  [5348] = 5348,
  [5349] = 5349,
  [5350] = 5350,
  [5351] = 5338,
  [5352] = 5339,
  [5353] = 5353,
  [5354] = 5354,
  [5355] = 5355,
  [5356] = 5332,
  [5357] = 5355,
  [5358] = 5358,
  [5359] = 5337,
  [5360] = 5347,
  [5361] = 5361,
  [5362] = 5362,
  [5363] = 5363,
  [5364] = 5364,
  [5365] = 5342,
  [5366] = 5366,
  [5367] = 5367,
  [5368] = 5340,
  [5369] = 5369,
  [5370] = 5370,
  [5371] = 5371,
  [5372] = 5372,
  [5373] = 5358,
  [5374] = 5374,
  [5375] = 5375,
  [5376] = 5376,
  [5377] = 5377,
  [5378] = 5378,
  [5379] = 5349,
  [5380] = 5378,
  [5381] = 5381,
  [5382] = 5382,
  [5383] = 5383,
  [5384] = 5384,
  [5385] = 5333,
  [5386] = 5336,
  [5387] = 5355,
  [5388] = 5363,
  [5389] = 5389,
  [5390] = 5390,
  [5391] = 5347,
  [5392] = 5392,
  [5393] = 5367,
  [5394] = 5394,
  [5395] = 5395,
  [5396] = 5396,
  [5397] = 5337,
  [5398] = 5398,
  [5399] = 5399,
  [5400] = 5363,
  [5401] = 5401,
  [5402] = 5402,
  [5403] = 5403,
  [5404] = 5378,
  [5405] = 5340,
  [5406] = 5344,
  [5407] = 5338,
  [5408] = 5408,
  [5409] = 5341,
  [5410] = 5342,
  [5411] = 5332,
  [5412] = 5412,
  [5413] = 5338,
  [5414] = 5358,
  [5415] = 5339,
  [5416] = 5374,
  [5417] = 5350,
  [5418] = 5349,
  [5419] = 5339,
  [5420] = 5338,
  [5421] = 5421,
  [5422] = 5339,
  [5423] = 5353,
  [5424] = 5349,
  [5425] = 5378,
  [5426] = 5396,
  [5427] = 5427,
  [5428] = 5428,
  [5429] = 5390,
  [5430] = 5430,
  [5431] = 5431,
  [5432] = 5345,
  [5433] = 5355,
  [5434] = 5377,
  [5435] = 5363,
  [5436] = 5436,
  [5437] = 5437,
  [5438] = 5344,
  [5439] = 5369,
  [5440] = 5332,
  [5441] = 5338,
  [5442] = 5383,
  [5443] = 5412,
  [5444] = 5408,
  [5445] = 5333,
  [5446] = 5366,
  [5447] = 5447,
  [5448] = 5358,
  [5449] = 5370,
  [5450] = 5371,
  [5451] = 5354,
  [5452] = 5332,
  [5453] = 5358,
  [5454] = 5333,
  [5455] = 5361,
  [5456] = 5346,
  [5457] = 5376,
  [5458] = 5337,
  [5459] = 5412,
  [5460] = 5346,
  [5461] = 5337,
  [5462] = 5336,
  [5463] = 5362,
  [5464] = 5403,
  [5465] = 5336,
  [5466] = 5340,
  [5467] = 5467,
  [5468] = 5372,
  [5469] = 5362,
  [5470] = 5355,
  [5471] = 5334,
  [5472] = 5374,
  [5473] = 5337,
  [5474] = 5340,
  [5475] = 5363,
  [5476] = 5372,
  [5477] = 5467,
  [5478] = 5381,
  [5479] = 5436,
  [5480] = 5383,
  [5481] = 5384,
  [5482] = 5389,
  [5483] = 5333,
  [5484] = 5392,
  [5485] = 5336,
  [5486] = 5398,
  [5487] = 5399,
  [5488] = 5401,
  [5489] = 5339,
  [5490] = 5336,
  [5491] = 5395,
  [5492] = 5350,
  [5493] = 5340,
  [5494] = 5494,
  [5495] = 5375,
  [5496] = 5355,
  [5497] = 5382,
  [5498] = 5430,
  [5499] = 5341,
  [5500] = 5338,
  [5501] = 5337,
  [5502] = 5421,
  [5503] = 5339,
  [5504] = 5353,
  [5505] = 5332,
  [5506] = 5427,
  [5507] = 5358,
  [5508] = 5381,
  [5509] = 5428,
  [5510] = 5377,
  [5511] = 5384,
  [5512] = 5349,
  [5513] = 5390,
  [5514] = 5340,
  [5515] = 5389,
  [5516] = 5394,
  [5517] = 5436,
  [5518] = 5332,
  [5519] = 5392,
  [5520] = 5437,
  [5521] = 5349,
  [5522] = 5369,
  [5523] = 5378,
  [5524] = 5366,
  [5525] = 5447,
  [5526] = 5332,
  [5527] = 5354,
  [5528] = 5361,
  [5529] = 5376,
  [5530] = 5398,
  [5531] = 5378,
  [5532] = 5467,
  [5533] = 5399,
  [5534] = 5428,
  [5535] = 5402,
  [5536] = 5401,
  [5537] = 5333,
  [5538] = 5538,
  [5539] = 5334,
  [5540] = 5349,
  [5541] = 5375,
  [5542] = 5378,
  [5543] = 5336,
  [5544] = 5358,
  [5545] = 5355,
  [5546] = 5363,
  [5547] = 5378,
  [5548] = 5447,
  [5549] = 5363,
  [5550] = 5382,
  [5551] = 5408,
  [5552] = 5337,
  [5553] = 5340,
  [5554] = 5431,
  [5555] = 5494,
  [5556] = 5345,
  [5557] = 5421,
  [5558] = 5370,
  [5559] = 5349,
  [5560] = 5430,
  [5561] = 5403,
  [5562] = 5437,
  [5563] = 5355,
  [5564] = 5494,
  [5565] = 5367,
  [5566] = 5394,
  [5567] = 5395,
  [5568] = 5431,
  [5569] = 5363,
  [5570] = 5338,
  [5571] = 5339,
  [5572] = 5333,
  [5573] = 5427,
  [5574] = 5402,
  [5575] = 5371,
  [5576] = 5358,
  [5577] = 5577,
  [5578] = 5578,
  [5579] = 5579,
  [5580] = 5578,
  [5581] = 5581,
  [5582] = 5578,
  [5583] = 5583,
  [5584] = 5584,
  [5585] = 5585,
  [5586] = 5586,
  [5587] = 5587,
  [5588] = 5581,
  [5589] = 5587,
  [5590] = 5577,
  [5591] = 5591,
  [5592] = 5592,
  [5593] = 5579,
  [5594] = 5578,
  [5595] = 5595,
  [5596] = 5596,
  [5597] = 5597,
  [5598] = 5581,
  [5599] = 5599,
  [5600] = 5600,
  [5601] = 5587,
  [5602] = 5592,
  [5603] = 5583,
  [5604] = 5591,
  [5605] = 5595,
  [5606] = 5606,
  [5607] = 5583,
  [5608] = 5608,
  [5609] = 5583,
  [5610] = 5584,
  [5611] = 5585,
  [5612] = 5586,
  [5613] = 5585,
  [5614] = 5614,
  [5615] = 5578,
  [5616] = 5597,
  [5617] = 5587,
  [5618] = 5618,
  [5619] = 5586,
  [5620] = 5584,
  [5621] = 5585,
  [5622] = 5586,
  [5623] = 5600,
  [5624] = 5577,
  [5625] = 5579,
  [5626] = 5587,
  [5627] = 5592,
  [5628] = 5587,
  [5629] = 5591,
  [5630] = 5591,
  [5631] = 5587,
  [5632] = 5579,
  [5633] = 5584,
  [5634] = 5584,
  [5635] = 5585,
  [5636] = 5586,
  [5637] = 5579,
  [5638] = 5585,
  [5639] = 5586,
  [5640] = 5581,
  [5641] = 5578,
  [5642] = 5595,
  [5643] = 5581,
  [5644] = 5581,
  [5645] = 5597,
  [5646] = 5595,
  [5647] = 5583,
  [5648] = 5584,
  [5649] = 5585,
  [5650] = 5586,
  [5651] = 5578,
  [5652] = 5597,
  [5653] = 5600,
  [5654] = 5581,
  [5655] = 5591,
  [5656] = 5577,
  [5657] = 5597,
  [5658] = 5591,
  [5659] = 5592,
  [5660] = 5600,
  [5661] = 5661,
  [5662] = 5579,
  [5663] = 5577,
  [5664] = 5595,
  [5665] = 5665,
  [5666] = 5581,
  [5667] = 5597,
  [5668] = 5668,
  [5669] = 5600,
  [5670] = 5579,
  [5671] = 5600,
  [5672] = 5583,
  [5673] = 5577,
  [5674] = 5578,
  [5675] = 5595,
  [5676] = 5595,
  [5677] = 5600,
  [5678] = 5587,
  [5679] = 5592,
  [5680] = 5680,
  [5681] = 5618,
  [5682] = 5682,
  [5683] = 5680,
  [5684] = 5682,
  [5685] = 5583,
  [5686] = 5577,
  [5687] = 5606,
  [5688] = 5583,
  [5689] = 5592,
  [5690] = 5584,
  [5691] = 5691,
  [5692] = 5597,
  [5693] = 5591,
  [5694] = 5618,
  [5695] = 5682,
  [5696] = 5680,
  [5697] = 5579,
  [5698] = 5577,
  [5699] = 5592,
  [5700] = 5591,
  [5701] = 5606,
  [5702] = 5595,
  [5703] = 5618,
  [5704] = 5682,
  [5705] = 5680,
  [5706] = 5592,
  [5707] = 5606,
  [5708] = 5618,
  [5709] = 5682,
  [5710] = 5680,
  [5711] = 5584,
  [5712] = 5585,
  [5713] = 5618,
  [5714] = 5682,
  [5715] = 5680,
  [5716] = 5586,
  [5717] = 5618,
  [5718] = 5682,
  [5719] = 5680,
  [5720] = 5597,
  [5721] = 5618,
  [5722] = 5682,
  [5723] = 5680,
  [5724] = 5661,
  [5725] = 5600,
  [5726] = 5726,
  [5727] = 5727,
  [5728] = 5728,
  [5729] = 5729,
  [5730] = 5730,
  [5731] = 5731,
  [5732] = 5732,
  [5733] = 5733,
  [5734] = 5734,
  [5735] = 5733,
  [5736] = 1735,
  [5737] = 5737,
  [5738] = 5738,
  [5739] = 5739,
  [5740] = 5740,
  [5741] = 5741,
  [5742] = 5734,
  [5743] = 5743,
  [5744] = 5744,
  [5745] = 5744,
  [5746] = 5738,
  [5747] = 5743,
  [5748] = 5748,
  [5749] = 5741,
  [5750] = 5743,
  [5751] = 5608,
  [5752] = 5740,
  [5753] = 5753,
  [5754] = 5691,
  [5755] = 5755,
  [5756] = 5741,
  [5757] = 5757,
  [5758] = 5741,
  [5759] = 1731,
  [5760] = 5730,
  [5761] = 5737,
  [5762] = 5741,
  [5763] = 5763,
  [5764] = 5728,
  [5765] = 5744,
  [5766] = 5729,
  [5767] = 5767,
  [5768] = 5768,
  [5769] = 5741,
  [5770] = 5770,
  [5771] = 5771,
  [5772] = 5728,
  [5773] = 5773,
  [5774] = 5774,
  [5775] = 5726,
  [5776] = 5741,
  [5777] = 5740,
  [5778] = 5778,
  [5779] = 5779,
  [5780] = 5737,
  [5781] = 5763,
  [5782] = 5768,
  [5783] = 5755,
  [5784] = 5784,
  [5785] = 5785,
  [5786] = 1499,
  [5787] = 5726,
  [5788] = 5771,
  [5789] = 5779,
  [5790] = 5741,
  [5791] = 5744,
  [5792] = 5741,
  [5793] = 5737,
  [5794] = 5768,
  [5795] = 5795,
  [5796] = 5741,
  [5797] = 5743,
  [5798] = 5755,
  [5799] = 5799,
  [5800] = 5741,
  [5801] = 5741,
  [5802] = 5740,
  [5803] = 5763,
  [5804] = 5741,
  [5805] = 5741,
  [5806] = 5733,
  [5807] = 5785,
  [5808] = 5743,
  [5809] = 5728,
  [5810] = 5763,
  [5811] = 5768,
  [5812] = 5795,
  [5813] = 5726,
  [5814] = 5741,
  [5815] = 5726,
  [5816] = 5771,
  [5817] = 5771,
  [5818] = 5743,
  [5819] = 5727,
  [5820] = 5730,
  [5821] = 5739,
  [5822] = 5727,
  [5823] = 5737,
  [5824] = 5753,
  [5825] = 5763,
  [5826] = 5732,
  [5827] = 5744,
  [5828] = 5773,
  [5829] = 5734,
  [5830] = 5741,
  [5831] = 5755,
  [5832] = 5768,
  [5833] = 5737,
  [5834] = 5741,
  [5835] = 5835,
  [5836] = 5730,
  [5837] = 5778,
  [5838] = 1512,
  [5839] = 5743,
  [5840] = 5771,
  [5841] = 5755,
  [5842] = 5740,
  [5843] = 5843,
  [5844] = 5737,
  [5845] = 5741,
  [5846] = 5846,
  [5847] = 5755,
  [5848] = 5771,
  [5849] = 5737,
  [5850] = 5779,
  [5851] = 5785,
  [5852] = 5755,
  [5853] = 5853,
  [5854] = 5737,
  [5855] = 5741,
  [5856] = 5795,
  [5857] = 5755,
  [5858] = 5728,
  [5859] = 5737,
  [5860] = 5757,
  [5861] = 5861,
  [5862] = 5755,
  [5863] = 5743,
  [5864] = 5737,
  [5865] = 5741,
  [5866] = 5757,
  [5867] = 5785,
  [5868] = 5737,
  [5869] = 5767,
  [5870] = 5795,
  [5871] = 5737,
  [5872] = 5741,
  [5873] = 5737,
  [5874] = 5753,
  [5875] = 5741,
  [5876] = 5737,
  [5877] = 5741,
  [5878] = 5726,
  [5879] = 5737,
  [5880] = 5835,
  [5881] = 5733,
  [5882] = 5737,
  [5883] = 5728,
  [5884] = 5763,
  [5885] = 5737,
  [5886] = 5740,
  [5887] = 5741,
  [5888] = 5737,
  [5889] = 5730,
  [5890] = 5778,
  [5891] = 5737,
  [5892] = 5741,
  [5893] = 5726,
  [5894] = 5737,
  [5895] = 5895,
  [5896] = 5732,
  [5897] = 5737,
  [5898] = 5755,
  [5899] = 5733,
  [5900] = 5737,
  [5901] = 5773,
  [5902] = 5785,
  [5903] = 5737,
  [5904] = 5779,
  [5905] = 5741,
  [5906] = 5737,
  [5907] = 5773,
  [5908] = 5741,
  [5909] = 5737,
  [5910] = 5910,
  [5911] = 5734,
  [5912] = 5737,
  [5913] = 5741,
  [5914] = 5744,
  [5915] = 5737,
  [5916] = 5727,
  [5917] = 5763,
  [5918] = 5737,
  [5919] = 5744,
  [5920] = 5920,
  [5921] = 5737,
  [5922] = 5727,
  [5923] = 5923,
  [5924] = 5737,
  [5925] = 5733,
  [5926] = 5732,
  [5927] = 5737,
  [5928] = 5928,
  [5929] = 5733,
  [5930] = 5737,
  [5931] = 5771,
  [5932] = 5932,
  [5933] = 5737,
  [5934] = 5741,
  [5935] = 5935,
  [5936] = 5737,
  [5937] = 5937,
  [5938] = 5785,
  [5939] = 5795,
  [5940] = 5795,
  [5941] = 5757,
  [5942] = 5768,
  [5943] = 5732,
  [5944] = 5733,
  [5945] = 5784,
  [5946] = 5730,
  [5947] = 5767,
  [5948] = 5732,
  [5949] = 5728,
  [5950] = 5757,
  [5951] = 5741,
  [5952] = 5731,
  [5953] = 5784,
  [5954] = 5730,
  [5955] = 5955,
  [5956] = 1139,
  [5957] = 5755,
  [5958] = 5763,
  [5959] = 5730,
  [5960] = 5767,
  [5961] = 5734,
  [5962] = 5767,
  [5963] = 5741,
  [5964] = 5739,
  [5965] = 5835,
  [5966] = 5773,
  [5967] = 5955,
  [5968] = 5741,
  [5969] = 5741,
  [5970] = 5773,
  [5971] = 5738,
  [5972] = 5737,
  [5973] = 5727,
  [5974] = 5757,
  [5975] = 5726,
  [5976] = 5726,
  [5977] = 5734,
  [5978] = 5737,
  [5979] = 5740,
  [5980] = 1140,
  [5981] = 5738,
  [5982] = 5771,
  [5983] = 5741,
  [5984] = 5955,
  [5985] = 5757,
  [5986] = 5767,
  [5987] = 5726,
  [5988] = 5734,
  [5989] = 5989,
  [5990] = 5767,
  [5991] = 5739,
  [5992] = 5740,
  [5993] = 5728,
  [5994] = 5955,
  [5995] = 5727,
  [5996] = 5753,
  [5997] = 5785,
  [5998] = 5795,
  [5999] = 5778,
  [6000] = 5768,
  [6001] = 5955,
  [6002] = 5726,
  [6003] = 5755,
  [6004] = 5773,
  [6005] = 5731,
  [6006] = 5731,
  [6007] = 5835,
  [6008] = 5955,
  [6009] = 5299,
  [6010] = 5741,
  [6011] = 5734,
  [6012] = 5989,
  [6013] = 5741,
  [6014] = 5727,
  [6015] = 5955,
  [6016] = 5989,
  [6017] = 5773,
  [6018] = 5726,
  [6019] = 5732,
  [6020] = 5989,
  [6021] = 6021,
  [6022] = 5955,
  [6023] = 5785,
  [6024] = 5955,
  [6025] = 5955,
  [6026] = 5955,
  [6027] = 5955,
  [6028] = 5795,
  [6029] = 5920,
  [6030] = 5932,
  [6031] = 5784,
  [6032] = 6032,
  [6033] = 5732,
  [6034] = 5937,
  [6035] = 5744,
  [6036] = 5920,
  [6037] = 5932,
  [6038] = 5768,
  [6039] = 6032,
  [6040] = 5937,
  [6041] = 5757,
  [6042] = 5767,
  [6043] = 6032,
  [6044] = 5910,
  [6045] = 5740,
  [6046] = 5741,
  [6047] = 6047,
  [6048] = 6048,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym_postfix_deref,
  sym_primitive,
  sym_slices,
  sym_subscripted,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym_postfix_deref] = {.index = 0, .length = 6},
  [sym_primitive] = {.index = 6, .length = 2},
  [sym_slices] = {.index = 8, .length = 2},
  [sym_subscripted] = {.index = 10, .length = 5},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_amper_deref_expression,
    sym_array_deref_expression,
    sym_arraylen_deref_expression,
    sym_glob_deref_expression,
    sym_hash_deref_expression,
    sym_scalar_deref_expression,
  [6] =
    sym_boolean,
    sym_number,
  [8] =
    sym_keyval_expression,
    sym_slice_expression,
  [10] =
    sym_anonymous_slice_expression,
    sym_array_element_expression,
    sym_coderef_call_expression,
    sym_glob_slot_expression,
    sym_hash_element_expression,
};

static const TSCharacterRange extras_character_set_1[] = {
  {'\t', '\r'}, {' ', ' '}, {0x85, 0x85}, {0xa0, 0xa0}, {0x1680, 0x1680}, {0x2000, 0x200a}, {0x2028, 0x2029}, {0x202f, 0x202f},
  {0x205f, 0x205f}, {0x3000, 0x3000},
};

static const TSCharacterRange aux_sym__interpolation_fallbacks_token1_character_set_1[] = {
  {0, '#'}, {'%', '&'}, {'(', '*'}, {',', ','}, {'.', '/'}, {';', '@'}, {'[', '^'}, {'`', '`'},
  {'{', 0x10ffff},
};

static const TSCharacterRange sym_substitution_regexp_modifiers_character_set_1[] = {
  {'a', 'a'}, {'c', 'e'}, {'g', 'g'}, {'i', 'i'}, {'l', 'm'}, {'o', 'p'}, {'r', 's'}, {'u', 'u'},
  {'x', 'x'},
};

static const TSCharacterRange sym__identifier_character_set_1[] = {
  {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6},
  {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374}, {0x376, 0x377}, {0x37b, 0x37d},
  {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x48a, 0x52f},
  {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a}, {0x66e, 0x66f}, {0x671, 0x6d3},
  {0x6d5, 0x6d5}, {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710}, {0x712, 0x72f}, {0x74d, 0x7a5},
  {0x7b1, 0x7b1}, {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a}, {0x824, 0x824}, {0x828, 0x828},
  {0x840, 0x858}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939}, {0x93d, 0x93d}, {0x950, 0x950},
  {0x958, 0x961}, {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bd, 0x9bd}, {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc}, {0xa05, 0xa0a}, {0xa0f, 0xa10},
  {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa72, 0xa74},
  {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabd, 0xabd}, {0xad0, 0xad0},
  {0xae0, 0xae1}, {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39},
  {0xb3d, 0xb3d}, {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95},
  {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbd0, 0xbd0}, {0xc05, 0xc0c},
  {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc61}, {0xc80, 0xc80},
  {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd}, {0xcdd, 0xcde}, {0xce0, 0xce1},
  {0xcf1, 0xcf2}, {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e}, {0xd54, 0xd56}, {0xd5f, 0xd61},
  {0xd7a, 0xd7f}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xe01, 0xe30}, {0xe32, 0xe32},
  {0xe40, 0xe46}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xeb0}, {0xeb2, 0xeb2},
  {0xebd, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47}, {0xf49, 0xf6c}, {0xf88, 0xf8c},
  {0x1000, 0x102a}, {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066}, {0x106e, 0x1070}, {0x1075, 0x1081},
  {0x108e, 0x108e}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd},
  {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711}, {0x171f, 0x1731}, {0x1740, 0x1751},
  {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878}, {0x1880, 0x18a8}, {0x18aa, 0x18aa},
  {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x1a00, 0x1a16}, {0x1a20, 0x1a54},
  {0x1aa7, 0x1aa7}, {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5}, {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f},
  {0x1c5a, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3}, {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa},
  {0x1d00, 0x1dbf}, {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x2102, 0x2102},
  {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25},
  {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6},
  {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007}, {0x3021, 0x3029}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf},
  {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f}, {0xa62a, 0xa62b}, {0xa640, 0xa66e},
  {0xa67f, 0xa69d}, {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc},
  {0xa7f2, 0xa801}, {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3}, {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb},
  {0xa8fd, 0xa8fe}, {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf}, {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef},
  {0xa9fa, 0xa9fe}, {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a}, {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1},
  {0xaab5, 0xaab6}, {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea}, {0xaaf2, 0xaaf4}, {0xab01, 0xab06},
  {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabe2}, {0xac00, 0xd7a3},
  {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28},
  {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d},
  {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b},
  {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xff9d}, {0xffa0, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf},
  {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d},
  {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x10375}, {0x10380, 0x1039d},
  {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563},
  {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc},
  {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805},
  {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2},
  {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00}, {0x10a10, 0x10a13}, {0x10a15, 0x10a17},
  {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9},
  {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6},
  {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126}, {0x11144, 0x11144}, {0x11147, 0x11147},
  {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x1122b},
  {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112de}, {0x11305, 0x1130c},
  {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d}, {0x11350, 0x11350}, {0x1135d, 0x11361},
  {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1}, {0x113d3, 0x113d3}, {0x11400, 0x11434},
  {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae}, {0x115d8, 0x115db}, {0x11600, 0x1162f},
  {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b}, {0x118a0, 0x118df}, {0x118ff, 0x11906},
  {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941}, {0x119a0, 0x119a7}, {0x119aa, 0x119d0},
  {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50}, {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d},
  {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09},
  {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98}, {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02},
  {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f},
  {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a70, 0x16abe}, {0x16ad0, 0x16aed},
  {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f50, 0x16f50},
  {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb},
  {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a},
  {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6},
  {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c},
  {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da},
  {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2},
  {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad},
  {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe},
  {0x1e800, 0x1e8c4}, {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27},
  {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b},
  {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f},
  {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89},
  {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1},
  {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym__identifier_character_set_2[] = {
  {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7}, {0xba, 0xba},
  {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x300, 0x374},
  {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481},
  {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf}, {0x5c1, 0x5c2},
  {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3}, {0x6d5, 0x6dc},
  {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa}, {0x7fd, 0x7fd},
  {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963}, {0x966, 0x96f},
  {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bc, 0x9c4},
  {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc}, {0x9fe, 0x9fe},
  {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39},
  {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa66, 0xa75},
  {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabc, 0xac5},
  {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03}, {0xb05, 0xb0c},
  {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48}, {0xb4b, 0xb4d},
  {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90},
  {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbbe, 0xbc2},
  {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28},
  {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc63},
  {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbc, 0xcc4},
  {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3}, {0xd00, 0xd0c},
  {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f}, {0xd7a, 0xd7f},
  {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca}, {0xdcf, 0xdd4},
  {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59}, {0xe81, 0xe82},
  {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xec8, 0xece},
  {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37}, {0xf39, 0xf39},
  {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049}, {0x1050, 0x109d},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f}, {0x13a0, 0x13f5},
  {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715}, {0x171f, 0x1734},
  {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd}, {0x17e0, 0x17e9},
  {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b}, {0x1930, 0x193b},
  {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c},
  {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59}, {0x1b6b, 0x1b73},
  {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2},
  {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054}, {0x2071, 0x2071},
  {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113},
  {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149},
  {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67},
  {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce},
  {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c}, {0x3041, 0x3096},
  {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf},
  {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1}, {0xa717, 0xa71f},
  {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c}, {0xa840, 0xa873},
  {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c}, {0xa980, 0xa9c0},
  {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2}, {0xaadb, 0xaadd},
  {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a},
  {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d},
  {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41},
  {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe00, 0xfe0f},
  {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b},
  {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe}, {0xffc2, 0xffc7},
  {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d},
  {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0}, {0x10300, 0x1031f},
  {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9},
  {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595},
  {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767},
  {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c},
  {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7},
  {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f},
  {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91},
  {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85},
  {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50}, {0x10f70, 0x10f85},
  {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9},
  {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da},
  {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d},
  {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330},
  {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357}, {0x1135d, 0x11363},
  {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0}, {0x113c2, 0x113c2},
  {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461}, {0x11480, 0x114c5},
  {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644}, {0x11650, 0x11659},
  {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746}, {0x11800, 0x1183a},
  {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938}, {0x1193b, 0x11943},
  {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47}, {0x11a50, 0x11a99},
  {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40}, {0x11c50, 0x11c59},
  {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d},
  {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98}, {0x11da0, 0x11da9},
  {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e},
  {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139}, {0x16800, 0x16a38},
  {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36}, {0x16b40, 0x16b43},
  {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87},
  {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3},
  {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb},
  {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46},
  {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c},
  {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505},
  {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550},
  {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e},
  {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75},
  {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021},
  {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e},
  {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe},
  {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24},
  {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49},
  {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d},
  {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e},
  {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739},
  {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 224,
        '#', 260,
        '$', 126,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 197,
        '/', 198,
        '0', 430,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 134,
        '[', 266,
        '\\', 228,
        ']', 156,
        '^', 186,
        'a', 282,
        'c', 313,
        'd', 285,
        'e', 355,
        'g', 334,
        'i', 290,
        'l', 285,
        'm', 279,
        'n', 285,
        'o', 291,
        'p', 286,
        'r', 336,
        's', 281,
        'u', 289,
        'v', 396,
        'x', 285,
        '{', 267,
        '|', 185,
        '}', 118,
        '~', 222,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(109);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        0x04, 264,
        '!', 224,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 197,
        '/', 198,
        '0', 430,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
        '\\', 228,
        ']', 156,
        '^', 186,
        'a', 352,
        'c', 315,
        'd', 364,
        'e', 395,
        'g', 334,
        'i', 392,
        'l', 324,
        'm', 276,
        'o', 378,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        'x', 199,
        '{', 117,
        '|', 185,
        '}', 118,
        '~', 222,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(1);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        0x04, 264,
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 190,
        ',', 135,
        '-', 193,
        '.', 105,
        '/', 85,
        '0', 430,
        ':', 87,
        ';', 121,
        '<', 88,
        '=', 92,
        '@', 133,
        '[', 155,
        '\\', 228,
        ']', 156,
        'a', 352,
        'c', 316,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'o', 378,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        'x', 368,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(2);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '$', 82,
        '%', 158,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 134,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(15);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(201);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(4);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(15);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(201);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(4);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(17);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(201);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(5);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(17);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'c') ADVANCE(309);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(201);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          lookahead == 'r' ||
          lookahead == 's') ADVANCE(309);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(5);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(201);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(7);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(308);
      if (lookahead == 'c') ADVANCE(303);
      if (lookahead == 'g') ADVANCE(302);
      if (lookahead == 'i') ADVANCE(307);
      if (lookahead == 'l') ADVANCE(302);
      if (lookahead == 'o') ADVANCE(306);
      if (lookahead == 'x') ADVANCE(304);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          lookahead == 'e' ||
          lookahead == 'm' ||
          lookahead == 'p' ||
          lookahead == 'r' ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(308);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(7);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(296);
      if (lookahead == 'c') ADVANCE(295);
      if (lookahead == 'g') ADVANCE(300);
      if (lookahead == 'i') ADVANCE(299);
      if (lookahead == 'l') ADVANCE(300);
      if (lookahead == 'o') ADVANCE(300);
      if (lookahead == 'x') ADVANCE(297);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(300);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(7);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(19);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(283);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(290);
      if (lookahead == 'l') ADVANCE(292);
      if (lookahead == 'o') ADVANCE(292);
      if (lookahead == 'x') ADVANCE(285);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(292);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(7);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        0x04, 264,
        '#', 260,
        '$', 124,
        '%', 234,
        '(', 137,
        ')', 136,
        '*', 153,
        ',', 135,
        '-', 192,
        '/', 86,
        ':', 119,
        ';', 121,
        '=', 130,
        '>', 159,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(37);
      if (lookahead == 'a') ADVANCE(403);
      if (lookahead == 'o') ADVANCE(406);
      if (lookahead == 'x') ADVANCE(404);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(99);
      if (lookahead == '}') ADVANCE(118);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(11);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        0x04, 264,
        '#', 260,
        '$', 124,
        '%', 234,
        '(', 137,
        ')', 136,
        ',', 135,
        '-', 192,
        '/', 86,
        ':', 119,
        ';', 121,
        '=', 128,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(39);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(99);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(12);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        0x04, 264,
        '#', 260,
        '$', 124,
        '%', 234,
        '(', 137,
        ')', 136,
        ',', 135,
        '-', 195,
        '/', 86,
        ':', 119,
        ';', 121,
        '=', 128,
        '>', 93,
        '@', 133,
        '[', 266,
      );
      if (lookahead == '\\') SKIP(39);
      if (lookahead == '{') ADVANCE(267);
      if (lookahead == '|') ADVANCE(99);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(12);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(4);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(4);
      if (lookahead == '\r') SKIP(14);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(5);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(5);
      if (lookahead == '\r') SKIP(16);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(7);
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(7);
      if (lookahead == '\r') SKIP(18);
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(56);
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(56);
      if (lookahead == '\r') SKIP(20);
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(65);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(65);
      if (lookahead == '\r') SKIP(22);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(57);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(57);
      if (lookahead == '\r') SKIP(24);
      END_STATE();
    case 26:
      if (lookahead == '\n') SKIP(58);
      END_STATE();
    case 27:
      if (lookahead == '\n') SKIP(58);
      if (lookahead == '\r') SKIP(26);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(64);
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(64);
      if (lookahead == '\r') SKIP(28);
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(61);
      END_STATE();
    case 31:
      if (lookahead == '\n') SKIP(61);
      if (lookahead == '\r') SKIP(30);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(66);
      END_STATE();
    case 33:
      if (lookahead == '\n') SKIP(66);
      if (lookahead == '\r') SKIP(32);
      END_STATE();
    case 34:
      if (lookahead == '\n') SKIP(69);
      END_STATE();
    case 35:
      if (lookahead == '\n') SKIP(69);
      if (lookahead == '\r') SKIP(34);
      END_STATE();
    case 36:
      if (lookahead == '\n') SKIP(11);
      END_STATE();
    case 37:
      if (lookahead == '\n') SKIP(11);
      if (lookahead == '\r') SKIP(36);
      END_STATE();
    case 38:
      if (lookahead == '\n') SKIP(12);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(12);
      if (lookahead == '\r') SKIP(38);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(72);
      END_STATE();
    case 41:
      if (lookahead == '\n') SKIP(72);
      if (lookahead == '\r') SKIP(40);
      END_STATE();
    case 42:
      if (lookahead == '\n') SKIP(80);
      END_STATE();
    case 43:
      if (lookahead == '\n') SKIP(80);
      if (lookahead == '\r') SKIP(42);
      END_STATE();
    case 44:
      if (lookahead == '\n') SKIP(73);
      END_STATE();
    case 45:
      if (lookahead == '\n') SKIP(73);
      if (lookahead == '\r') SKIP(44);
      END_STATE();
    case 46:
      if (lookahead == '\n') SKIP(74);
      END_STATE();
    case 47:
      if (lookahead == '\n') SKIP(74);
      if (lookahead == '\r') SKIP(46);
      END_STATE();
    case 48:
      if (lookahead == '\n') SKIP(81);
      END_STATE();
    case 49:
      if (lookahead == '\n') SKIP(81);
      if (lookahead == '\r') SKIP(48);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(50);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '\\') ADVANCE(310);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(311);
      if (lookahead != 0) ADVANCE(310);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '!', 224,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 197,
        '/', 198,
        '0', 430,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
        '\\', 228,
        '^', 186,
        'a', 353,
        'c', 315,
        'd', 364,
        'e', 395,
        'g', 334,
        'i', 392,
        'l', 324,
        'm', 276,
        'o', 388,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        'x', 200,
        '{', 117,
        '|', 185,
        '}', 118,
        '~', 222,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(51);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 190,
        ',', 135,
        '-', 193,
        '.', 105,
        '/', 85,
        '0', 430,
        ':', 120,
        ';', 121,
        '<', 88,
        '=', 92,
        '@', 133,
        '[', 155,
        '\\', 228,
        ']', 156,
        'a', 353,
        'c', 316,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'o', 388,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(52);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 53:
      ADVANCE_MAP(
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 105,
        '/', 85,
        '0', 430,
        ':', 87,
        '<', 88,
        '@', 133,
        '[', 155,
        '\\', 228,
        'a', 353,
        'c', 316,
        'd', 364,
        'e', 355,
        'g', 377,
        'l', 325,
        'm', 276,
        'o', 388,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        '{', 117,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(53);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 54:
      ADVANCE_MAP(
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 419,
        '/', 420,
        '0', 417,
        ':', 421,
        ';', 121,
        '<', 422,
        '@', 133,
        '[', 155,
        '\\', 228,
        '^', 425,
        'a', 353,
        'c', 314,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'n', 365,
        'o', 388,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(418);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(413);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0) ADVANCE(412);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 105,
        '/', 85,
        '0', 430,
        ':', 87,
        ';', 121,
        '<', 88,
        '@', 133,
        '[', 155,
        '\\', 228,
        'a', 353,
        'c', 316,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'o', 388,
        'p', 366,
        'r', 337,
        's', 277,
        'u', 327,
        'v', 396,
        '{', 117,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(55);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 56:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 120,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
      );
      if (lookahead == '\\') SKIP(21);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(360);
      if (lookahead == 'g') ADVANCE(335);
      if (lookahead == 'i') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(338);
      if (lookahead == 'x') ADVANCE(200);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(56);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 57:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(25);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(57);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 58:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(27);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(58);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 59:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(27);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(309);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          lookahead == 'r' ||
          lookahead == 's') ADVANCE(309);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(58);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 60:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(290);
      if (lookahead == 'l') ADVANCE(292);
      if (lookahead == 'x') ADVANCE(292);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(292);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(61);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 61:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(61);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 62:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(303);
      if (lookahead == 'g') ADVANCE(302);
      if (lookahead == 'i') ADVANCE(307);
      if (lookahead == 'l') ADVANCE(302);
      if (lookahead == 'x') ADVANCE(308);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'e' ||
          lookahead == 'm' ||
          lookahead == 'o' ||
          lookahead == 'p' ||
          lookahead == 'r' ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(308);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(61);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 63:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 234,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
      );
      if (lookahead == '\\') SKIP(31);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(295);
      if (lookahead == 'g') ADVANCE(300);
      if (lookahead == 'i') ADVANCE(299);
      if (lookahead == 'l') ADVANCE(300);
      if (lookahead == 'x') ADVANCE(300);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(300);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(61);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 64:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(29);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(64);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 65:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(23);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(360);
      if (lookahead == 'g') ADVANCE(335);
      if (lookahead == 'i') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(338);
      if (lookahead == 'x') ADVANCE(200);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(65);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 66:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(66);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 67:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(33);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(309);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'd' ||
          lookahead == 'r' ||
          lookahead == 's') ADVANCE(309);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(66);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 68:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(35);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(290);
      if (lookahead == 'l') ADVANCE(292);
      if (lookahead == 'x') ADVANCE(292);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(292);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(69);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 69:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(35);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(402);
      if (lookahead == 'g') ADVANCE(400);
      if (lookahead == 'i') ADVANCE(409);
      if (lookahead == 'l') ADVANCE(401);
      if (lookahead == 'x') ADVANCE(202);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(69);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 70:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(35);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(303);
      if (lookahead == 'g') ADVANCE(302);
      if (lookahead == 'i') ADVANCE(307);
      if (lookahead == 'l') ADVANCE(302);
      if (lookahead == 'x') ADVANCE(308);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'e' ||
          lookahead == 'm' ||
          lookahead == 'o' ||
          lookahead == 'p' ||
          lookahead == 'r' ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(308);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(69);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 71:
      ADVANCE_MAP(
        '!', 89,
        '#', 260,
        '%', 158,
        '&', 188,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 119,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
      );
      if (lookahead == '\\') SKIP(35);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(295);
      if (lookahead == 'g') ADVANCE(300);
      if (lookahead == 'i') ADVANCE(299);
      if (lookahead == 'l') ADVANCE(300);
      if (lookahead == 'x') ADVANCE(300);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u') ADVANCE(300);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(69);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 72:
      ADVANCE_MAP(
        '#', 260,
        '$', 125,
        '%', 157,
        '&', 187,
        '(', 137,
        '*', 153,
        ':', 120,
        ';', 121,
        '@', 133,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(41);
      if (lookahead == '{') ADVANCE(117);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(72);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 73:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '(') ADVANCE(137);
      if (lookahead == '\\') SKIP(45);
      if (lookahead == 'o') ADVANCE(410);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(73);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 74:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '*') ADVANCE(153);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == '>') ADVANCE(159);
      if (lookahead == '\\') SKIP(47);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(74);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 75:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '*') ADVANCE(153);
      if (lookahead == ':') ADVANCE(421);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (lookahead == '{') ADVANCE(117);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(75);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(414);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0) ADVANCE(412);
      END_STATE();
    case 76:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == ':') ADVANCE(421);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (lookahead == '{') ADVANCE(117);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(76);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(415);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0) ADVANCE(412);
      END_STATE();
    case 77:
      ADVANCE_MAP(
        '#', 260,
        '$', 124,
        ':', 421,
        '\\', 270,
        '^', 272,
        '{', 117,
        '\'', 412,
        '+', 412,
        '-', 412,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(271);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(273);
      if (lookahead != 0) ADVANCE(270);
      END_STATE();
    case 78:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(78);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(416);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0) ADVANCE(412);
      END_STATE();
    case 79:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(127);
      if (lookahead == '@') ADVANCE(134);
      if (lookahead == '[') ADVANCE(155);
      if (lookahead == '\\') ADVANCE(274);
      if (lookahead == '{') ADVANCE(117);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(275);
      if (lookahead != 0) ADVANCE(274);
      END_STATE();
    case 80:
      ADVANCE_MAP(
        '#', 260,
        ')', 136,
        ',', 135,
        '.', 105,
        '0', 430,
        ':', 119,
        ';', 121,
        '=', 92,
      );
      if (lookahead == '\\') SKIP(43);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == 'a') ADVANCE(95);
      if (lookahead == 'o') ADVANCE(97);
      if (lookahead == 'v') ADVANCE(107);
      if (lookahead == 'x') ADVANCE(96);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '}') ADVANCE(118);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(80);
      END_STATE();
    case 81:
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '.') ADVANCE(105);
      if (lookahead == '0') ADVANCE(430);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == '\\') SKIP(49);
      if (lookahead == 'v') ADVANCE(396);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(81);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 82:
      if (lookahead == '*') ADVANCE(268);
      END_STATE();
    case 83:
      if (lookahead == '.') ADVANCE(140);
      END_STATE();
    case 84:
      if (lookahead == '.') ADVANCE(83);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    case 85:
      if (lookahead == '/') ADVANCE(181);
      END_STATE();
    case 86:
      if (lookahead == '/') ADVANCE(90);
      END_STATE();
    case 87:
      if (lookahead == ':') ADVANCE(429);
      END_STATE();
    case 88:
      if (lookahead == '<') ADVANCE(161);
      END_STATE();
    case 89:
      if (lookahead == '=') ADVANCE(206);
      if (lookahead == '~') ADVANCE(204);
      END_STATE();
    case 90:
      if (lookahead == '=') ADVANCE(132);
      END_STATE();
    case 91:
      if (lookahead == '=') ADVANCE(131);
      END_STATE();
    case 92:
      if (lookahead == '>') ADVANCE(238);
      END_STATE();
    case 93:
      if (lookahead == '>') ADVANCE(163);
      END_STATE();
    case 94:
      if (lookahead == 'd') ADVANCE(142);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(94);
      END_STATE();
    case 96:
      if (lookahead == 'o') ADVANCE(98);
      END_STATE();
    case 97:
      if (lookahead == 'r') ADVANCE(145);
      END_STATE();
    case 98:
      if (lookahead == 'r') ADVANCE(149);
      END_STATE();
    case 99:
      if (lookahead == '|') ADVANCE(91);
      END_STATE();
    case 100:
      if (lookahead == '~') ADVANCE(211);
      END_STATE();
    case 101:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(432);
      END_STATE();
    case 102:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(433);
      END_STATE();
    case 103:
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(432);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(435);
      END_STATE();
    case 104:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(431);
      END_STATE();
    case 105:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    case 106:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(432);
      END_STATE();
    case 107:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(312);
      END_STATE();
    case 108:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(434);
      END_STATE();
    case 109:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 224,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 197,
        '/', 198,
        '0', 430,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
        '\\', 228,
        ']', 156,
        '^', 186,
        'a', 352,
        'c', 313,
        'd', 364,
        'e', 355,
        'g', 334,
        'i', 392,
        'l', 324,
        'm', 276,
        'n', 365,
        'o', 378,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        'v', 396,
        'x', 199,
        '{', 117,
        '|', 185,
        '}', 118,
        '~', 222,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(109);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 110:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 224,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        ')', 136,
        '*', 236,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 197,
        '/', 198,
        '0', 430,
        ':', 87,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '@', 133,
        '[', 155,
        '\\', 228,
        '^', 186,
        'a', 352,
        'c', 313,
        'd', 364,
        'e', 395,
        'g', 334,
        'i', 392,
        'l', 324,
        'm', 276,
        'n', 365,
        'o', 378,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        'x', 199,
        '{', 117,
        '|', 185,
        '}', 118,
        '~', 222,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(110);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 111:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 84,
        '/', 85,
        '0', 430,
        ':', 87,
        ';', 121,
        '<', 88,
        '@', 133,
        '[', 155,
        '\\', 228,
        'a', 353,
        'c', 314,
        'd', 364,
        'e', 355,
        'g', 377,
        'l', 325,
        'm', 276,
        'n', 365,
        'o', 388,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(111);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 112:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 84,
        '/', 85,
        '0', 430,
        ':', 87,
        ';', 121,
        '<', 88,
        '@', 133,
        '[', 155,
        '\\', 228,
        'a', 353,
        'c', 314,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'n', 365,
        'o', 388,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(431);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(112);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 113:
      if (eof) ADVANCE(116);
      ADVANCE_MAP(
        0x04, 264,
        '!', 89,
        '#', 260,
        '$', 124,
        '%', 158,
        '&', 188,
        '(', 137,
        ')', 136,
        '*', 154,
        '+', 191,
        ',', 135,
        '-', 194,
        '.', 196,
        '/', 198,
        ':', 120,
        ';', 121,
        '<', 212,
        '=', 129,
        '>', 160,
        '?', 227,
        '[', 155,
      );
      if (lookahead == '\\') SKIP(115);
      if (lookahead == ']') ADVANCE(156);
      if (lookahead == '^') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(363);
      if (lookahead == 'c') ADVANCE(360);
      if (lookahead == 'g') ADVANCE(335);
      if (lookahead == 'i') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(338);
      if (lookahead == 'o') ADVANCE(387);
      if (lookahead == 'x') ADVANCE(199);
      if (lookahead == '{') ADVANCE(117);
      if (lookahead == '|') ADVANCE(185);
      if (lookahead == '}') ADVANCE(118);
      if (lookahead == '~') ADVANCE(100);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(113);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      END_STATE();
    case 114:
      if (eof) ADVANCE(116);
      if (lookahead == '\n') SKIP(113);
      END_STATE();
    case 115:
      if (eof) ADVANCE(116);
      if (lookahead == '\n') SKIP(113);
      if (lookahead == '\r') SKIP(114);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(429);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_class);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_role);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '#') ADVANCE(231);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '#') ADVANCE(231);
      if (lookahead == '*') ADVANCE(268);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '*') ADVANCE(268);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(205);
      if (lookahead == '>') ADVANCE(238);
      if (lookahead == '~') ADVANCE(203);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '>') ADVANCE(238);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE_EQ);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH_EQ);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == '*') ADVANCE(269);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_our);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_our);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_else);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(250);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_or);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_xor);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_xor);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_xor);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(179);
      if (lookahead == '=') ADVANCE(168);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(170);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(214);
      if (lookahead == '>') ADVANCE(189);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_DOT_EQ);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_x_EQ);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_AMP_AMP_EQ);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '.') ADVANCE(140);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(164);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (lookahead == '=') ADVANCE(131);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      if (lookahead == '=') ADVANCE(132);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_CARET_CARET);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(173);
      if (lookahead == '|') ADVANCE(180);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(174);
      if (lookahead == '^') ADVANCE(183);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(184);
      if (lookahead == '=') ADVANCE(172);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_GT_GT2);
      if (lookahead == '=') ADVANCE(176);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(225);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(225);
      if (lookahead == '=') ADVANCE(165);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(226);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(226);
      if (lookahead == '=') ADVANCE(166);
      if (lookahead == '>') ADVANCE(152);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(265);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(178);
      if (lookahead == '=') ADVANCE(167);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(178);
      if (lookahead == '=') ADVANCE(167);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(182);
      if (lookahead == '=') ADVANCE(169);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_x);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == 'o') ADVANCE(380);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_x);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == '=') ADVANCE(171);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_x);
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == 'o') ADVANCE(407);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_x);
      if (lookahead == '=') ADVANCE(171);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_BANG_TILDE);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(207);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_LT_EQ_GT);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_cmp);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_cmp);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_TILDE_TILDE);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(162);
      if (lookahead == '=') ADVANCE(213);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '>') ADVANCE(208);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_le);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_le);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_ge);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_ge);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_isa);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_isa);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (lookahead == '~') ADVANCE(211);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(206);
      if (lookahead == '~') ADVANCE(204);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_local);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_DOLLAR_POUND);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_map);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_grep);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_PERCENT2);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_AMP2);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_STAR2);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym__var_indirob_autoquote_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(237);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_use);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_no);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_redo);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_alarm);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_close);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(351);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_closedir);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_caller);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_cos);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_exp);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_lc);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_log);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_ord);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_pop);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_pos);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_rmdir);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_readdir);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_readpipe);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_scalar);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_sin);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_sleep);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_uc);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(260);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym___DATA___token1);
      if (lookahead == '\r') ADVANCE(263);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(263);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(aux_sym___DATA___token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '\\') ADVANCE(261);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          lookahead != '\n') ADVANCE(262);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(263);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(aux_sym___DATA___token1);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(263);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(anon_sym_EOT);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(anon_sym_DASH_GT2);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(anon_sym_LBRACK2);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(anon_sym_LBRACE2);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(anon_sym_DOLLAR_STAR);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(anon_sym_AT_STAR);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token1);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '\\') ADVANCE(270);
      if (lookahead == '^') ADVANCE(272);
      if (lookahead == '{') ADVANCE(117);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(271);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(271);
      if ((set_contains(sym__identifier_character_set_1, 685, lookahead)) &&
          (lookahead < 'A' || 'Z' < lookahead) &&
          lookahead != '_' &&
          (lookahead < 'a' || 'z' < lookahead)) ADVANCE(273);
      if ((!eof && set_contains(aux_sym__interpolation_fallbacks_token1_character_set_1, 9, lookahead))) ADVANCE(270);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token1);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|') ADVANCE(412);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token1);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token2);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym__interpolation_fallbacks_token2);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '\\') ADVANCE(274);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(275);
      if (lookahead != 0 &&
          lookahead != '#' &&
          lookahead != '$' &&
          lookahead != '@' &&
          lookahead != '[' &&
          lookahead != '\\' &&
          lookahead != '{') ADVANCE(274);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(anon_sym_m);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(373);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_s);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(319);
      if (lookahead == 'i') ADVANCE(362);
      if (lookahead == 'l') ADVANCE(347);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'a') ADVANCE(292);
      if (lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'a') ADVANCE(288);
      if (lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'd') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'i') ADVANCE(284);
      if (lookahead == 'l') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          ('m' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      ADVANCE_MAP(
        'l', 278,
        'n', 280,
        'a', 292,
        'd', 292,
        'i', 292,
        'm', 292,
        'o', 292,
        'p', 292,
        's', 292,
        'u', 292,
        'x', 292,
      );
      END_STATE();
    case 283:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      ADVANCE_MAP(
        'n', 280,
        'a', 292,
        'd', 292,
        'i', 292,
        'l', 292,
        'm', 292,
        'o', 292,
        'p', 292,
        's', 292,
        'u', 292,
        'x', 292,
      );
      END_STATE();
    case 284:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      ADVANCE_MAP(
        'n', 292,
        'a', 292,
        'd', 292,
        'i', 292,
        'l', 292,
        'm', 292,
        'o', 292,
        'p', 292,
        's', 292,
        'u', 292,
        'x', 292,
      );
      END_STATE();
    case 285:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'o') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'o') ADVANCE(287);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'p') ADVANCE(292);
      if (lookahead == 's') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'o') ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'p') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'o') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 's') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 's') ADVANCE(278);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'u') ADVANCE(292);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(sym_quoted_regexp_modifiers);
      if (lookahead == 'a' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(292);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 'a') ADVANCE(300);
      if (lookahead == 'c' ||
          lookahead == 'd' ||
          lookahead == 'g' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 's' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(300);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 'd') ADVANCE(300);
      if ((set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead) ||
          lookahead == 'n') &&
          lookahead != 'd' &&
          lookahead != 'e' &&
          lookahead != 'r') ADVANCE(300);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 'm') ADVANCE(298);
      if ((set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead) ||
          lookahead == 'n') &&
          lookahead != 'e' &&
          lookahead != 'r') ADVANCE(300);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      ADVANCE_MAP(
        'n', 294,
        'a', 300,
        'c', 300,
        'd', 300,
        'g', 300,
        'i', 300,
        'l', 300,
        'm', 300,
        'o', 300,
        'p', 300,
        's', 300,
        'u', 300,
        'x', 300,
      );
      END_STATE();
    case 297:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 'o') ADVANCE(300);
      if ((set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead) ||
          lookahead == 'n') &&
          lookahead != 'e' &&
          lookahead != 'r') ADVANCE(300);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 'p') ADVANCE(300);
      if ((set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead) ||
          lookahead == 'n') &&
          lookahead != 'e' &&
          lookahead != 'r') ADVANCE(300);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if (lookahead == 's') ADVANCE(293);
      if (lookahead == 'a' ||
          lookahead == 'c' ||
          lookahead == 'd' ||
          lookahead == 'g' ||
          lookahead == 'i' ||
          ('l' <= lookahead && lookahead <= 'p') ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(300);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(sym_match_regexp_modifiers);
      if ((set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead) ||
          lookahead == 'n') &&
          lookahead != 'e' &&
          lookahead != 'r') ADVANCE(300);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 'a') ADVANCE(308);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      ADVANCE_MAP(
        'e', 308,
        'a', 308,
        'c', 308,
        'd', 308,
        'g', 308,
        'i', 308,
        'l', 308,
        'm', 308,
        'o', 308,
        'p', 308,
        'r', 308,
        's', 308,
        'u', 308,
        'x', 308,
      );
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 'm') ADVANCE(305);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 'o') ADVANCE(306);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 'p') ADVANCE(308);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 'r') ADVANCE(308);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (lookahead == 's') ADVANCE(301);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(sym_substitution_regexp_modifiers);
      if (set_contains(sym_substitution_regexp_modifiers_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(sym_transliteration_modifiers);
      if (lookahead == 'c' ||
          lookahead == 'd' ||
          lookahead == 'r' ||
          lookahead == 's') ADVANCE(309);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym__interpolated_transliteration_content_token1);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(aux_sym__interpolated_transliteration_content_token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '\\') ADVANCE(310);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          lookahead != '\n') ADVANCE(311);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(310);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_version);
      if (lookahead == '.') ADVANCE(107);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(312);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(359);
      if (lookahead == 'l') ADVANCE(321);
      if (lookahead == 'm') ADVANCE(370);
      if (lookahead == 'o') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(359);
      if (lookahead == 'l') ADVANCE(321);
      if (lookahead == 'o') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(359);
      if (lookahead == 'l') ADVANCE(369);
      if (lookahead == 'm') ADVANCE(370);
      if (lookahead == 'o') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(359);
      if (lookahead == 'l') ADVANCE(369);
      if (lookahead == 'o') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(219);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(330);
      if (lookahead == 'd') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(358);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(354);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(393);
      if (lookahead == 'o') ADVANCE(394);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(382);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'a') ADVANCE(385);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(248);
      if (lookahead == 'e') ADVANCE(215);
      if (lookahead == 'o') ADVANCE(328);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(248);
      if (lookahead == 'o') ADVANCE(328);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(259);
      if (lookahead == 's') ADVANCE(341);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(259);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'g') ADVANCE(249);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(348);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(333);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(143);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(250);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'd') ADVANCE(350);
      if (lookahead == 'p') ADVANCE(349);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(217);
      if (lookahead == 'r') ADVANCE(344);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(217);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(318);
      if (lookahead == 'm') ADVANCE(329);
      if (lookahead == 'o') ADVANCE(356);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(318);
      if (lookahead == 'm') ADVANCE(329);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(215);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(141);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(123);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(239);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(243);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(255);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(375);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(383);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'e') ADVANCE(345);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'i') ADVANCE(381);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'i') ADVANCE(376);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'i') ADVANCE(384);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'i') ADVANCE(386);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(322);
      if (lookahead == 'n') ADVANCE(331);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(322);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(230);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(391);
      if (lookahead == 'x') ADVANCE(371);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(340);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(346);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(323);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'l') ADVANCE(357);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'm') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'm') ADVANCE(242);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'n') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 363:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'n') ADVANCE(331);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 364:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(229);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 365:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(240);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 366:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(374);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(241);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(380);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'o') ADVANCE(394);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 370:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(209);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(247);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(233);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(232);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(251);
      if (lookahead == 's') ADVANCE(252);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(258);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 376:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'p') ADVANCE(343);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(344);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 378:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(146);
      if (lookahead == 'u') ADVANCE(379);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 379:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 380:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(150);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(253);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(361);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(245);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(254);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(256);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(244);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(147);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(332);
      if (lookahead == 'u') ADVANCE(379);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(246);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(122);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(339);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 392:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(317);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 393:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(390);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 's') ADVANCE(342);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (lookahead == 'x') ADVANCE(371);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 396:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(312);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(397);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'a') ADVANCE(220);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'd') ADVANCE(144);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'e') ADVANCE(218);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'e') ADVANCE(216);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'm') ADVANCE(405);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'n') ADVANCE(399);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 404:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'o') ADVANCE(407);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 405:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'p') ADVANCE(210);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'r') ADVANCE(148);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'r') ADVANCE(151);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 408:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'r') ADVANCE(139);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 409:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 's') ADVANCE(398);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 410:
      ACCEPT_TOKEN(sym__identifier);
      if (lookahead == 'u') ADVANCE(408);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 411:
      ACCEPT_TOKEN(sym__identifier);
      if (set_contains(sym__identifier_character_set_2, 800, lookahead)) ADVANCE(411);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      ADVANCE_MAP(
        '!', 223,
        '#', 260,
        '$', 125,
        '%', 234,
        '&', 235,
        '(', 137,
        '*', 236,
        '+', 190,
        '-', 193,
        '.', 419,
        '/', 420,
        '0', 417,
        ':', 421,
        ';', 121,
        '<', 422,
        '@', 133,
        '[', 155,
        '\\', 228,
        '^', 425,
        'a', 353,
        'c', 314,
        'd', 364,
        'e', 395,
        'g', 377,
        'l', 325,
        'm', 276,
        'n', 365,
        'o', 388,
        'p', 366,
        'r', 336,
        's', 277,
        'u', 326,
        '{', 117,
        '}', 118,
        '~', 221,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(418);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(413);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '!') ADVANCE(412);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '*') ADVANCE(153);
      if (lookahead == ':') ADVANCE(421);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (lookahead == '{') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(414);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(412);
      END_STATE();
    case 415:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == ':') ADVANCE(421);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (lookahead == '{') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(415);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(397);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(412);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '#') ADVANCE(260);
      if (lookahead == '$') ADVANCE(124);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '^') ADVANCE(427);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(416);
      if (set_contains(sym__identifier_character_set_1, 685, lookahead)) ADVANCE(411);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(412);
      END_STATE();
    case 417:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      ADVANCE_MAP(
        '.', 437,
        '_', 106,
        'B', 102,
        'b', 102,
        'E', 101,
        'e', 101,
        'X', 108,
        'x', 108,
        '8', 423,
        '9', 423,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(424);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '.') ADVANCE(437);
      if (lookahead == '_') ADVANCE(104);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(101);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(418);
      END_STATE();
    case 419:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '.') ADVANCE(83);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '/') ADVANCE(181);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == ':') ADVANCE(429);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '<') ADVANCE(161);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '_') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(423);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '_') ADVANCE(103);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(423);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(424);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '?' ||
          lookahead == '^' ||
          lookahead == '|') ADVANCE(412);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(237);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(237);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(426);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(aux_sym__ident_special_token1);
      if (lookahead == '?' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|') ADVANCE(412);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(aux_sym__bareword_token1);
      if (lookahead == ':') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(428);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(428);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(aux_sym__bareword_token1);
      if (lookahead == ':') ADVANCE(87);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(428);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(sym_number);
      ADVANCE_MAP(
        '.', 437,
        '_', 106,
        'B', 102,
        'b', 102,
        'E', 101,
        'e', 101,
        'X', 108,
        'x', 108,
        '8', 432,
        '9', 432,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(435);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '.') ADVANCE(437);
      if (lookahead == '_') ADVANCE(104);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(101);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(431);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(432);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(102);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(433);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(108);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(434);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(103);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(432);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(435);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(105);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(101);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(101);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(436);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == 'A') ADVANCE(1);
      if (lookahead == 'B') ADVANCE(2);
      if (lookahead == 'C') ADVANCE(3);
      if (lookahead == 'E') ADVANCE(4);
      if (lookahead == 'I') ADVANCE(5);
      if (lookahead == 'U') ADVANCE(6);
      if (lookahead == '\\') SKIP(7);
      if (lookahead == '_') ADVANCE(8);
      if (lookahead == 'a') ADVANCE(9);
      if (lookahead == 'b') ADVANCE(10);
      if (lookahead == 'c') ADVANCE(11);
      if (lookahead == 'd') ADVANCE(12);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'f') ADVANCE(14);
      if (lookahead == 'g') ADVANCE(15);
      if (lookahead == 'h') ADVANCE(16);
      if (lookahead == 'i') ADVANCE(17);
      if (lookahead == 'k') ADVANCE(18);
      if (lookahead == 'l') ADVANCE(19);
      if (lookahead == 'm') ADVANCE(20);
      if (lookahead == 'n') ADVANCE(21);
      if (lookahead == 'o') ADVANCE(22);
      if (lookahead == 'p') ADVANCE(23);
      if (lookahead == 'q') ADVANCE(24);
      if (lookahead == 'r') ADVANCE(25);
      if (lookahead == 's') ADVANCE(26);
      if (lookahead == 't') ADVANCE(27);
      if (lookahead == 'u') ADVANCE(28);
      if (lookahead == 'v') ADVANCE(29);
      if (lookahead == 'w') ADVANCE(30);
      if (lookahead == 'y') ADVANCE(31);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'D') ADVANCE(32);
      END_STATE();
    case 2:
      if (lookahead == 'E') ADVANCE(33);
      if (lookahead == 'U') ADVANCE(34);
      END_STATE();
    case 3:
      if (lookahead == 'H') ADVANCE(35);
      END_STATE();
    case 4:
      if (lookahead == 'N') ADVANCE(36);
      END_STATE();
    case 5:
      if (lookahead == 'N') ADVANCE(37);
      END_STATE();
    case 6:
      if (lookahead == 'N') ADVANCE(38);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(0);
      if (lookahead == '\r') SKIP(39);
      END_STATE();
    case 8:
      if (lookahead == '_') ADVANCE(40);
      END_STATE();
    case 9:
      if (lookahead == 'b') ADVANCE(41);
      if (lookahead == 's') ADVANCE(42);
      if (lookahead == 'w') ADVANCE(43);
      END_STATE();
    case 10:
      if (lookahead == 'r') ADVANCE(44);
      END_STATE();
    case 11:
      if (lookahead == 'a') ADVANCE(45);
      if (lookahead == 'h') ADVANCE(46);
      if (lookahead == 'o') ADVANCE(47);
      END_STATE();
    case 12:
      if (lookahead == 'b') ADVANCE(48);
      if (lookahead == 'e') ADVANCE(49);
      if (lookahead == 'y') ADVANCE(50);
      END_STATE();
    case 13:
      if (lookahead == 'a') ADVANCE(51);
      if (lookahead == 'l') ADVANCE(52);
      if (lookahead == 'o') ADVANCE(53);
      if (lookahead == 'q') ADVANCE(54);
      if (lookahead == 'v') ADVANCE(55);
      if (lookahead == 'x') ADVANCE(56);
      END_STATE();
    case 14:
      if (lookahead == 'a') ADVANCE(57);
      if (lookahead == 'c') ADVANCE(58);
      if (lookahead == 'i') ADVANCE(59);
      if (lookahead == 'o') ADVANCE(60);
      END_STATE();
    case 15:
      if (lookahead == 'e') ADVANCE(61);
      if (lookahead == 'm') ADVANCE(62);
      if (lookahead == 'o') ADVANCE(63);
      if (lookahead == 't') ADVANCE(64);
      END_STATE();
    case 16:
      if (lookahead == 'e') ADVANCE(65);
      END_STATE();
    case 17:
      if (lookahead == 'f') ADVANCE(66);
      if (lookahead == 'n') ADVANCE(67);
      END_STATE();
    case 18:
      if (lookahead == 'e') ADVANCE(68);
      END_STATE();
    case 19:
      if (lookahead == 'a') ADVANCE(69);
      if (lookahead == 'c') ADVANCE(70);
      if (lookahead == 'e') ADVANCE(71);
      if (lookahead == 'o') ADVANCE(72);
      if (lookahead == 's') ADVANCE(73);
      if (lookahead == 't') ADVANCE(74);
      END_STATE();
    case 20:
      if (lookahead == 'e') ADVANCE(75);
      if (lookahead == 'y') ADVANCE(76);
      END_STATE();
    case 21:
      if (lookahead == 'e') ADVANCE(77);
      END_STATE();
    case 22:
      if (lookahead == 'c') ADVANCE(78);
      END_STATE();
    case 23:
      if (lookahead == 'a') ADVANCE(79);
      if (lookahead == 'r') ADVANCE(80);
      END_STATE();
    case 24:
      ACCEPT_TOKEN(anon_sym_q);
      if (lookahead == 'q') ADVANCE(81);
      if (lookahead == 'r') ADVANCE(82);
      if (lookahead == 'u') ADVANCE(83);
      if (lookahead == 'w') ADVANCE(84);
      if (lookahead == 'x') ADVANCE(85);
      END_STATE();
    case 25:
      if (lookahead == 'a') ADVANCE(86);
      if (lookahead == 'e') ADVANCE(87);
      END_STATE();
    case 26:
      if (lookahead == 'h') ADVANCE(88);
      if (lookahead == 'o') ADVANCE(89);
      if (lookahead == 'q') ADVANCE(90);
      if (lookahead == 'r') ADVANCE(91);
      if (lookahead == 't') ADVANCE(92);
      if (lookahead == 'u') ADVANCE(93);
      END_STATE();
    case 27:
      if (lookahead == 'e') ADVANCE(94);
      if (lookahead == 'i') ADVANCE(95);
      if (lookahead == 'r') ADVANCE(96);
      END_STATE();
    case 28:
      if (lookahead == 'c') ADVANCE(97);
      if (lookahead == 'm') ADVANCE(98);
      if (lookahead == 'n') ADVANCE(99);
      END_STATE();
    case 29:
      if (lookahead == 'a') ADVANCE(100);
      END_STATE();
    case 30:
      if (lookahead == 'a') ADVANCE(101);
      if (lookahead == 'h') ADVANCE(102);
      if (lookahead == 'r') ADVANCE(103);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_y);
      END_STATE();
    case 32:
      if (lookahead == 'J') ADVANCE(104);
      END_STATE();
    case 33:
      if (lookahead == 'G') ADVANCE(105);
      END_STATE();
    case 34:
      if (lookahead == 'I') ADVANCE(106);
      END_STATE();
    case 35:
      if (lookahead == 'E') ADVANCE(107);
      END_STATE();
    case 36:
      if (lookahead == 'D') ADVANCE(108);
      END_STATE();
    case 37:
      if (lookahead == 'I') ADVANCE(109);
      END_STATE();
    case 38:
      if (lookahead == 'I') ADVANCE(110);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(0);
      END_STATE();
    case 40:
      if (lookahead == 'D') ADVANCE(111);
      if (lookahead == 'E') ADVANCE(112);
      if (lookahead == 'F') ADVANCE(113);
      if (lookahead == 'L') ADVANCE(114);
      if (lookahead == 'P') ADVANCE(115);
      if (lookahead == 'S') ADVANCE(116);
      END_STATE();
    case 41:
      if (lookahead == 's') ADVANCE(117);
      END_STATE();
    case 42:
      if (lookahead == 'y') ADVANCE(118);
      END_STATE();
    case 43:
      if (lookahead == 'a') ADVANCE(119);
      END_STATE();
    case 44:
      if (lookahead == 'e') ADVANCE(120);
      END_STATE();
    case 45:
      if (lookahead == 't') ADVANCE(121);
      END_STATE();
    case 46:
      if (lookahead == 'd') ADVANCE(122);
      if (lookahead == 'o') ADVANCE(123);
      if (lookahead == 'r') ADVANCE(124);
      END_STATE();
    case 47:
      if (lookahead == 'n') ADVANCE(125);
      END_STATE();
    case 48:
      if (lookahead == 'm') ADVANCE(126);
      END_STATE();
    case 49:
      if (lookahead == 'f') ADVANCE(127);
      if (lookahead == 'l') ADVANCE(128);
      END_STATE();
    case 50:
      if (lookahead == 'n') ADVANCE(129);
      END_STATE();
    case 51:
      if (lookahead == 'c') ADVANCE(130);
      END_STATE();
    case 52:
      if (lookahead == 's') ADVANCE(131);
      END_STATE();
    case 53:
      if (lookahead == 'f') ADVANCE(132);
      END_STATE();
    case 54:
      ACCEPT_TOKEN(anon_sym_eq);
      if (lookahead == 'r') ADVANCE(133);
      if (lookahead == 'u') ADVANCE(134);
      END_STATE();
    case 55:
      if (lookahead == 'a') ADVANCE(135);
      END_STATE();
    case 56:
      if (lookahead == 'i') ADVANCE(136);
      if (lookahead == 't') ADVANCE(137);
      END_STATE();
    case 57:
      if (lookahead == 'l') ADVANCE(138);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(anon_sym_fc);
      END_STATE();
    case 59:
      if (lookahead == 'e') ADVANCE(139);
      if (lookahead == 'l') ADVANCE(140);
      if (lookahead == 'n') ADVANCE(141);
      END_STATE();
    case 60:
      if (lookahead == 'r') ADVANCE(142);
      END_STATE();
    case 61:
      if (lookahead == 't') ADVANCE(143);
      END_STATE();
    case 62:
      if (lookahead == 't') ADVANCE(144);
      END_STATE();
    case 63:
      if (lookahead == 't') ADVANCE(145);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(anon_sym_gt);
      END_STATE();
    case 65:
      if (lookahead == 'x') ADVANCE(146);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 67:
      if (lookahead == 't') ADVANCE(147);
      END_STATE();
    case 68:
      if (lookahead == 'y') ADVANCE(148);
      END_STATE();
    case 69:
      if (lookahead == 's') ADVANCE(149);
      END_STATE();
    case 70:
      if (lookahead == 'f') ADVANCE(150);
      END_STATE();
    case 71:
      if (lookahead == 'n') ADVANCE(151);
      END_STATE();
    case 72:
      if (lookahead == 'c') ADVANCE(152);
      END_STATE();
    case 73:
      if (lookahead == 't') ADVANCE(153);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(anon_sym_lt);
      END_STATE();
    case 75:
      if (lookahead == 't') ADVANCE(154);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(anon_sym_my);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(anon_sym_ne);
      if (lookahead == 'x') ADVANCE(155);
      END_STATE();
    case 78:
      if (lookahead == 't') ADVANCE(156);
      END_STATE();
    case 79:
      if (lookahead == 'c') ADVANCE(157);
      END_STATE();
    case 80:
      if (lookahead == 'o') ADVANCE(158);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_qq);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_qr);
      END_STATE();
    case 83:
      if (lookahead == 'o') ADVANCE(159);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(anon_sym_qw);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_qx);
      END_STATE();
    case 86:
      if (lookahead == 'n') ADVANCE(160);
      END_STATE();
    case 87:
      if (lookahead == 'a') ADVANCE(161);
      if (lookahead == 'f') ADVANCE(162);
      if (lookahead == 'q') ADVANCE(163);
      if (lookahead == 's') ADVANCE(164);
      if (lookahead == 't') ADVANCE(165);
      if (lookahead == 'w') ADVANCE(166);
      END_STATE();
    case 88:
      if (lookahead == 'i') ADVANCE(167);
      END_STATE();
    case 89:
      if (lookahead == 'r') ADVANCE(168);
      END_STATE();
    case 90:
      if (lookahead == 'r') ADVANCE(169);
      END_STATE();
    case 91:
      if (lookahead == 'a') ADVANCE(170);
      END_STATE();
    case 92:
      if (lookahead == 'a') ADVANCE(171);
      if (lookahead == 'u') ADVANCE(172);
      END_STATE();
    case 93:
      if (lookahead == 'b') ADVANCE(173);
      END_STATE();
    case 94:
      if (lookahead == 'l') ADVANCE(174);
      END_STATE();
    case 95:
      if (lookahead == 'e') ADVANCE(175);
      if (lookahead == 'm') ADVANCE(176);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_tr);
      if (lookahead == 'u') ADVANCE(177);
      if (lookahead == 'y') ADVANCE(178);
      END_STATE();
    case 97:
      if (lookahead == 'f') ADVANCE(179);
      END_STATE();
    case 98:
      if (lookahead == 'a') ADVANCE(180);
      END_STATE();
    case 99:
      if (lookahead == 'd') ADVANCE(181);
      if (lookahead == 'l') ADVANCE(182);
      if (lookahead == 't') ADVANCE(183);
      END_STATE();
    case 100:
      if (lookahead == 'l') ADVANCE(184);
      END_STATE();
    case 101:
      if (lookahead == 'i') ADVANCE(185);
      if (lookahead == 'n') ADVANCE(186);
      END_STATE();
    case 102:
      if (lookahead == 'i') ADVANCE(187);
      END_STATE();
    case 103:
      if (lookahead == 'i') ADVANCE(188);
      END_STATE();
    case 104:
      if (lookahead == 'U') ADVANCE(189);
      END_STATE();
    case 105:
      if (lookahead == 'I') ADVANCE(190);
      END_STATE();
    case 106:
      if (lookahead == 'L') ADVANCE(191);
      END_STATE();
    case 107:
      if (lookahead == 'C') ADVANCE(192);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_END);
      END_STATE();
    case 109:
      if (lookahead == 'T') ADVANCE(193);
      END_STATE();
    case 110:
      if (lookahead == 'T') ADVANCE(194);
      END_STATE();
    case 111:
      if (lookahead == 'A') ADVANCE(195);
      END_STATE();
    case 112:
      if (lookahead == 'N') ADVANCE(196);
      END_STATE();
    case 113:
      if (lookahead == 'I') ADVANCE(197);
      END_STATE();
    case 114:
      if (lookahead == 'I') ADVANCE(198);
      END_STATE();
    case 115:
      if (lookahead == 'A') ADVANCE(199);
      END_STATE();
    case 116:
      if (lookahead == 'U') ADVANCE(200);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_abs);
      END_STATE();
    case 118:
      if (lookahead == 'n') ADVANCE(201);
      END_STATE();
    case 119:
      if (lookahead == 'i') ADVANCE(202);
      END_STATE();
    case 120:
      if (lookahead == 'a') ADVANCE(203);
      END_STATE();
    case 121:
      if (lookahead == 'c') ADVANCE(204);
      END_STATE();
    case 122:
      if (lookahead == 'i') ADVANCE(205);
      END_STATE();
    case 123:
      if (lookahead == 'm') ADVANCE(206);
      if (lookahead == 'p') ADVANCE(207);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_chr);
      if (lookahead == 'o') ADVANCE(208);
      END_STATE();
    case 125:
      if (lookahead == 't') ADVANCE(209);
      END_STATE();
    case 126:
      if (lookahead == 'c') ADVANCE(210);
      END_STATE();
    case 127:
      if (lookahead == 'e') ADVANCE(211);
      if (lookahead == 'i') ADVANCE(212);
      END_STATE();
    case 128:
      if (lookahead == 'e') ADVANCE(213);
      END_STATE();
    case 129:
      if (lookahead == 'a') ADVANCE(214);
      END_STATE();
    case 130:
      if (lookahead == 'h') ADVANCE(215);
      END_STATE();
    case 131:
      if (lookahead == 'i') ADVANCE(216);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_eof);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_eqr);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_equ);
      END_STATE();
    case 135:
      if (lookahead == 'l') ADVANCE(217);
      END_STATE();
    case 136:
      if (lookahead == 's') ADVANCE(218);
      if (lookahead == 't') ADVANCE(219);
      END_STATE();
    case 137:
      if (lookahead == 'e') ADVANCE(220);
      END_STATE();
    case 138:
      if (lookahead == 's') ADVANCE(221);
      END_STATE();
    case 139:
      if (lookahead == 'l') ADVANCE(222);
      END_STATE();
    case 140:
      if (lookahead == 'e') ADVANCE(223);
      END_STATE();
    case 141:
      if (lookahead == 'a') ADVANCE(224);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_for);
      if (lookahead == 'e') ADVANCE(225);
      if (lookahead == 'k') ADVANCE(226);
      END_STATE();
    case 143:
      if (lookahead == 'c') ADVANCE(227);
      if (lookahead == 'g') ADVANCE(228);
      if (lookahead == 'n') ADVANCE(229);
      if (lookahead == 'p') ADVANCE(230);
      if (lookahead == 's') ADVANCE(231);
      END_STATE();
    case 144:
      if (lookahead == 'i') ADVANCE(232);
      END_STATE();
    case 145:
      if (lookahead == 'o') ADVANCE(233);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_hex);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_int);
      END_STATE();
    case 148:
      if (lookahead == 's') ADVANCE(234);
      END_STATE();
    case 149:
      if (lookahead == 't') ADVANCE(235);
      END_STATE();
    case 150:
      if (lookahead == 'i') ADVANCE(236);
      END_STATE();
    case 151:
      if (lookahead == 'g') ADVANCE(237);
      END_STATE();
    case 152:
      if (lookahead == 'a') ADVANCE(238);
      if (lookahead == 'k') ADVANCE(239);
      END_STATE();
    case 153:
      if (lookahead == 'a') ADVANCE(240);
      END_STATE();
    case 154:
      if (lookahead == 'h') ADVANCE(241);
      END_STATE();
    case 155:
      if (lookahead == 't') ADVANCE(242);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_oct);
      END_STATE();
    case 157:
      if (lookahead == 'k') ADVANCE(243);
      END_STATE();
    case 158:
      if (lookahead == 't') ADVANCE(244);
      END_STATE();
    case 159:
      if (lookahead == 't') ADVANCE(245);
      END_STATE();
    case 160:
      if (lookahead == 'd') ADVANCE(246);
      END_STATE();
    case 161:
      if (lookahead == 'd') ADVANCE(247);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_ref);
      END_STATE();
    case 163:
      if (lookahead == 'u') ADVANCE(248);
      END_STATE();
    case 164:
      if (lookahead == 'e') ADVANCE(249);
      END_STATE();
    case 165:
      if (lookahead == 'u') ADVANCE(250);
      END_STATE();
    case 166:
      if (lookahead == 'i') ADVANCE(251);
      END_STATE();
    case 167:
      if (lookahead == 'f') ADVANCE(252);
      END_STATE();
    case 168:
      if (lookahead == 't') ADVANCE(253);
      END_STATE();
    case 169:
      if (lookahead == 't') ADVANCE(254);
      END_STATE();
    case 170:
      if (lookahead == 'n') ADVANCE(255);
      END_STATE();
    case 171:
      if (lookahead == 't') ADVANCE(256);
      END_STATE();
    case 172:
      if (lookahead == 'd') ADVANCE(257);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_sub);
      END_STATE();
    case 174:
      if (lookahead == 'l') ADVANCE(258);
      END_STATE();
    case 175:
      if (lookahead == 'd') ADVANCE(259);
      END_STATE();
    case 176:
      if (lookahead == 'e') ADVANCE(260);
      END_STATE();
    case 177:
      if (lookahead == 'e') ADVANCE(261);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 179:
      if (lookahead == 'i') ADVANCE(262);
      END_STATE();
    case 180:
      if (lookahead == 's') ADVANCE(263);
      END_STATE();
    case 181:
      if (lookahead == 'e') ADVANCE(264);
      END_STATE();
    case 182:
      if (lookahead == 'e') ADVANCE(265);
      END_STATE();
    case 183:
      if (lookahead == 'i') ADVANCE(266);
      END_STATE();
    case 184:
      if (lookahead == 'u') ADVANCE(267);
      END_STATE();
    case 185:
      if (lookahead == 't') ADVANCE(268);
      END_STATE();
    case 186:
      if (lookahead == 't') ADVANCE(269);
      END_STATE();
    case 187:
      if (lookahead == 'l') ADVANCE(270);
      END_STATE();
    case 188:
      if (lookahead == 't') ADVANCE(271);
      END_STATE();
    case 189:
      if (lookahead == 'S') ADVANCE(272);
      END_STATE();
    case 190:
      if (lookahead == 'N') ADVANCE(273);
      END_STATE();
    case 191:
      if (lookahead == 'D') ADVANCE(274);
      END_STATE();
    case 192:
      if (lookahead == 'K') ADVANCE(275);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_INIT);
      END_STATE();
    case 194:
      if (lookahead == 'C') ADVANCE(276);
      END_STATE();
    case 195:
      if (lookahead == 'T') ADVANCE(277);
      END_STATE();
    case 196:
      if (lookahead == 'D') ADVANCE(278);
      END_STATE();
    case 197:
      if (lookahead == 'L') ADVANCE(279);
      END_STATE();
    case 198:
      if (lookahead == 'N') ADVANCE(280);
      END_STATE();
    case 199:
      if (lookahead == 'C') ADVANCE(281);
      END_STATE();
    case 200:
      if (lookahead == 'B') ADVANCE(282);
      END_STATE();
    case 201:
      if (lookahead == 'c') ADVANCE(283);
      END_STATE();
    case 202:
      if (lookahead == 't') ADVANCE(284);
      END_STATE();
    case 203:
      if (lookahead == 'k') ADVANCE(285);
      END_STATE();
    case 204:
      if (lookahead == 'h') ADVANCE(286);
      END_STATE();
    case 205:
      if (lookahead == 'r') ADVANCE(287);
      END_STATE();
    case 206:
      if (lookahead == 'p') ADVANCE(288);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_chop);
      END_STATE();
    case 208:
      if (lookahead == 'o') ADVANCE(289);
      END_STATE();
    case 209:
      if (lookahead == 'i') ADVANCE(290);
      END_STATE();
    case 210:
      if (lookahead == 'l') ADVANCE(291);
      END_STATE();
    case 211:
      if (lookahead == 'r') ADVANCE(292);
      END_STATE();
    case 212:
      if (lookahead == 'n') ADVANCE(293);
      END_STATE();
    case 213:
      if (lookahead == 't') ADVANCE(294);
      END_STATE();
    case 214:
      if (lookahead == 'm') ADVANCE(295);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_each);
      END_STATE();
    case 216:
      if (lookahead == 'f') ADVANCE(296);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_eval);
      END_STATE();
    case 218:
      if (lookahead == 't') ADVANCE(297);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_exit);
      END_STATE();
    case 220:
      if (lookahead == 'n') ADVANCE(298);
      END_STATE();
    case 221:
      if (lookahead == 'e') ADVANCE(299);
      END_STATE();
    case 222:
      if (lookahead == 'd') ADVANCE(300);
      END_STATE();
    case 223:
      if (lookahead == 'n') ADVANCE(301);
      END_STATE();
    case 224:
      if (lookahead == 'l') ADVANCE(302);
      END_STATE();
    case 225:
      if (lookahead == 'a') ADVANCE(303);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_fork);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_getc);
      END_STATE();
    case 228:
      if (lookahead == 'r') ADVANCE(304);
      END_STATE();
    case 229:
      if (lookahead == 'e') ADVANCE(305);
      END_STATE();
    case 230:
      if (lookahead == 'e') ADVANCE(306);
      if (lookahead == 'g') ADVANCE(307);
      if (lookahead == 'p') ADVANCE(308);
      if (lookahead == 'r') ADVANCE(309);
      if (lookahead == 'w') ADVANCE(310);
      END_STATE();
    case 231:
      if (lookahead == 'o') ADVANCE(311);
      END_STATE();
    case 232:
      if (lookahead == 'm') ADVANCE(312);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_goto);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_keys);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_last);
      END_STATE();
    case 236:
      if (lookahead == 'r') ADVANCE(313);
      END_STATE();
    case 237:
      if (lookahead == 't') ADVANCE(314);
      END_STATE();
    case 238:
      if (lookahead == 'l') ADVANCE(315);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_lock);
      END_STATE();
    case 240:
      if (lookahead == 't') ADVANCE(316);
      END_STATE();
    case 241:
      if (lookahead == 'o') ADVANCE(317);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_next);
      END_STATE();
    case 243:
      if (lookahead == 'a') ADVANCE(318);
      END_STATE();
    case 244:
      if (lookahead == 'o') ADVANCE(319);
      END_STATE();
    case 245:
      if (lookahead == 'e') ADVANCE(320);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_rand);
      END_STATE();
    case 247:
      if (lookahead == 'l') ADVANCE(321);
      END_STATE();
    case 248:
      if (lookahead == 'i') ADVANCE(322);
      END_STATE();
    case 249:
      if (lookahead == 't') ADVANCE(323);
      END_STATE();
    case 250:
      if (lookahead == 'r') ADVANCE(324);
      END_STATE();
    case 251:
      if (lookahead == 'n') ADVANCE(325);
      END_STATE();
    case 252:
      if (lookahead == 't') ADVANCE(326);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_sort);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_sqrt);
      END_STATE();
    case 255:
      if (lookahead == 'd') ADVANCE(327);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_stat);
      if (lookahead == 'e') ADVANCE(328);
      END_STATE();
    case 257:
      if (lookahead == 'y') ADVANCE(329);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_tell);
      if (lookahead == 'd') ADVANCE(330);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_tied);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_time);
      if (lookahead == 's') ADVANCE(331);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 262:
      if (lookahead == 'r') ADVANCE(332);
      END_STATE();
    case 263:
      if (lookahead == 'k') ADVANCE(333);
      END_STATE();
    case 264:
      if (lookahead == 'f') ADVANCE(334);
      END_STATE();
    case 265:
      if (lookahead == 's') ADVANCE(335);
      END_STATE();
    case 266:
      if (lookahead == 'e') ADVANCE(336);
      if (lookahead == 'l') ADVANCE(337);
      END_STATE();
    case 267:
      if (lookahead == 'e') ADVANCE(338);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(anon_sym_wait);
      END_STATE();
    case 269:
      if (lookahead == 'a') ADVANCE(339);
      END_STATE();
    case 270:
      if (lookahead == 'e') ADVANCE(340);
      END_STATE();
    case 271:
      if (lookahead == 'e') ADVANCE(341);
      END_STATE();
    case 272:
      if (lookahead == 'T') ADVANCE(342);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(anon_sym_BEGIN);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(anon_sym_BUILD);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(anon_sym_CHECK);
      END_STATE();
    case 276:
      if (lookahead == 'H') ADVANCE(343);
      END_STATE();
    case 277:
      if (lookahead == 'A') ADVANCE(344);
      END_STATE();
    case 278:
      if (lookahead == '_') ADVANCE(345);
      END_STATE();
    case 279:
      if (lookahead == 'E') ADVANCE(346);
      END_STATE();
    case 280:
      if (lookahead == 'E') ADVANCE(347);
      END_STATE();
    case 281:
      if (lookahead == 'K') ADVANCE(348);
      END_STATE();
    case 282:
      if (lookahead == '_') ADVANCE(349);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(anon_sym_async);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(anon_sym_await);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_catch);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(anon_sym_chdir);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(anon_sym_chomp);
      END_STATE();
    case 289:
      if (lookahead == 't') ADVANCE(350);
      END_STATE();
    case 290:
      if (lookahead == 'n') ADVANCE(351);
      END_STATE();
    case 291:
      if (lookahead == 'o') ADVANCE(352);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(anon_sym_defer);
      END_STATE();
    case 293:
      if (lookahead == 'e') ADVANCE(353);
      END_STATE();
    case 294:
      if (lookahead == 'e') ADVANCE(354);
      END_STATE();
    case 295:
      if (lookahead == 'i') ADVANCE(355);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(anon_sym_elsif);
      END_STATE();
    case 297:
      if (lookahead == 's') ADVANCE(356);
      END_STATE();
    case 298:
      if (lookahead == 'd') ADVANCE(357);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(anon_sym_field);
      END_STATE();
    case 301:
      if (lookahead == 'o') ADVANCE(358);
      END_STATE();
    case 302:
      if (lookahead == 'l') ADVANCE(359);
      END_STATE();
    case 303:
      if (lookahead == 'c') ADVANCE(360);
      END_STATE();
    case 304:
      if (lookahead == 'g') ADVANCE(361);
      if (lookahead == 'n') ADVANCE(362);
      END_STATE();
    case 305:
      if (lookahead == 't') ADVANCE(363);
      END_STATE();
    case 306:
      if (lookahead == 'e') ADVANCE(364);
      END_STATE();
    case 307:
      if (lookahead == 'r') ADVANCE(365);
      END_STATE();
    case 308:
      if (lookahead == 'i') ADVANCE(366);
      END_STATE();
    case 309:
      if (lookahead == 'o') ADVANCE(367);
      END_STATE();
    case 310:
      if (lookahead == 'n') ADVANCE(368);
      if (lookahead == 'u') ADVANCE(369);
      END_STATE();
    case 311:
      if (lookahead == 'c') ADVANCE(370);
      END_STATE();
    case 312:
      if (lookahead == 'e') ADVANCE(371);
      END_STATE();
    case 313:
      if (lookahead == 's') ADVANCE(372);
      END_STATE();
    case 314:
      if (lookahead == 'h') ADVANCE(373);
      END_STATE();
    case 315:
      if (lookahead == 't') ADVANCE(374);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(anon_sym_lstat);
      END_STATE();
    case 317:
      if (lookahead == 'd') ADVANCE(375);
      END_STATE();
    case 318:
      if (lookahead == 'g') ADVANCE(376);
      END_STATE();
    case 319:
      if (lookahead == 't') ADVANCE(377);
      END_STATE();
    case 320:
      if (lookahead == 'm') ADVANCE(378);
      END_STATE();
    case 321:
      if (lookahead == 'i') ADVANCE(379);
      END_STATE();
    case 322:
      if (lookahead == 'r') ADVANCE(380);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(anon_sym_reset);
      END_STATE();
    case 324:
      if (lookahead == 'n') ADVANCE(381);
      END_STATE();
    case 325:
      if (lookahead == 'd') ADVANCE(382);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(anon_sym_shift);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(anon_sym_srand);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(anon_sym_state);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(anon_sym_study);
      END_STATE();
    case 330:
      if (lookahead == 'i') ADVANCE(383);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(anon_sym_times);
      END_STATE();
    case 332:
      if (lookahead == 's') ADVANCE(384);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(anon_sym_umask);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(anon_sym_undef);
      END_STATE();
    case 335:
      if (lookahead == 's') ADVANCE(385);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(anon_sym_untie);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(anon_sym_until);
      END_STATE();
    case 338:
      if (lookahead == 's') ADVANCE(386);
      END_STATE();
    case 339:
      if (lookahead == 'r') ADVANCE(387);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(anon_sym_write);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(anon_sym_ADJUST);
      END_STATE();
    case 343:
      if (lookahead == 'E') ADVANCE(388);
      END_STATE();
    case 344:
      if (lookahead == '_') ADVANCE(389);
      END_STATE();
    case 345:
      if (lookahead == '_') ADVANCE(390);
      END_STATE();
    case 346:
      if (lookahead == '_') ADVANCE(391);
      END_STATE();
    case 347:
      if (lookahead == '_') ADVANCE(392);
      END_STATE();
    case 348:
      if (lookahead == 'A') ADVANCE(393);
      END_STATE();
    case 349:
      if (lookahead == '_') ADVANCE(394);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(anon_sym_chroot);
      END_STATE();
    case 351:
      if (lookahead == 'u') ADVANCE(395);
      END_STATE();
    case 352:
      if (lookahead == 's') ADVANCE(396);
      END_STATE();
    case 353:
      if (lookahead == 'd') ADVANCE(397);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(anon_sym_delete);
      END_STATE();
    case 355:
      if (lookahead == 'c') ADVANCE(398);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(anon_sym_exists);
      END_STATE();
    case 357:
      if (lookahead == 'e') ADVANCE(399);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(anon_sym_fileno);
      END_STATE();
    case 359:
      if (lookahead == 'y') ADVANCE(400);
      END_STATE();
    case 360:
      if (lookahead == 'h') ADVANCE(401);
      END_STATE();
    case 361:
      if (lookahead == 'i') ADVANCE(402);
      END_STATE();
    case 362:
      if (lookahead == 'a') ADVANCE(403);
      END_STATE();
    case 363:
      if (lookahead == 'b') ADVANCE(404);
      END_STATE();
    case 364:
      if (lookahead == 'r') ADVANCE(405);
      END_STATE();
    case 365:
      if (lookahead == 'p') ADVANCE(406);
      END_STATE();
    case 366:
      if (lookahead == 'd') ADVANCE(407);
      END_STATE();
    case 367:
      if (lookahead == 't') ADVANCE(408);
      END_STATE();
    case 368:
      if (lookahead == 'a') ADVANCE(409);
      END_STATE();
    case 369:
      if (lookahead == 'i') ADVANCE(410);
      END_STATE();
    case 370:
      if (lookahead == 'k') ADVANCE(411);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(anon_sym_gmtime);
      END_STATE();
    case 372:
      if (lookahead == 't') ADVANCE(412);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(anon_sym_length);
      END_STATE();
    case 374:
      if (lookahead == 'i') ADVANCE(413);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(anon_sym_method);
      END_STATE();
    case 376:
      if (lookahead == 'e') ADVANCE(414);
      END_STATE();
    case 377:
      if (lookahead == 'y') ADVANCE(415);
      END_STATE();
    case 378:
      if (lookahead == 'e') ADVANCE(416);
      END_STATE();
    case 379:
      if (lookahead == 'n') ADVANCE(417);
      END_STATE();
    case 380:
      if (lookahead == 'e') ADVANCE(418);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 382:
      if (lookahead == 'd') ADVANCE(419);
      END_STATE();
    case 383:
      if (lookahead == 'r') ADVANCE(420);
      END_STATE();
    case 384:
      if (lookahead == 't') ADVANCE(421);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(anon_sym_unless);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(anon_sym_values);
      END_STATE();
    case 387:
      if (lookahead == 'r') ADVANCE(422);
      END_STATE();
    case 388:
      if (lookahead == 'C') ADVANCE(423);
      END_STATE();
    case 389:
      if (lookahead == '_') ADVANCE(424);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(anon_sym___END__);
      END_STATE();
    case 391:
      if (lookahead == '_') ADVANCE(425);
      END_STATE();
    case 392:
      if (lookahead == '_') ADVANCE(426);
      END_STATE();
    case 393:
      if (lookahead == 'G') ADVANCE(427);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(anon_sym___SUB__);
      END_STATE();
    case 395:
      if (lookahead == 'e') ADVANCE(428);
      END_STATE();
    case 396:
      if (lookahead == 'e') ADVANCE(429);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(anon_sym_defined);
      END_STATE();
    case 398:
      if (lookahead == 'a') ADVANCE(430);
      END_STATE();
    case 399:
      if (lookahead == 'd') ADVANCE(431);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(anon_sym_finally);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(anon_sym_foreach);
      END_STATE();
    case 402:
      if (lookahead == 'd') ADVANCE(432);
      END_STATE();
    case 403:
      if (lookahead == 'm') ADVANCE(433);
      END_STATE();
    case 404:
      if (lookahead == 'y') ADVANCE(434);
      END_STATE();
    case 405:
      if (lookahead == 'n') ADVANCE(435);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(anon_sym_getpgrp);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(anon_sym_getppid);
      END_STATE();
    case 408:
      if (lookahead == 'o') ADVANCE(436);
      END_STATE();
    case 409:
      if (lookahead == 'm') ADVANCE(437);
      END_STATE();
    case 410:
      if (lookahead == 'd') ADVANCE(438);
      END_STATE();
    case 411:
      if (lookahead == 'n') ADVANCE(439);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(anon_sym_lcfirst);
      END_STATE();
    case 413:
      if (lookahead == 'm') ADVANCE(440);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(anon_sym_package);
      END_STATE();
    case 415:
      if (lookahead == 'p') ADVANCE(441);
      END_STATE();
    case 416:
      if (lookahead == 't') ADVANCE(442);
      END_STATE();
    case 417:
      if (lookahead == 'e') ADVANCE(443);
      if (lookahead == 'k') ADVANCE(444);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(anon_sym_require);
      END_STATE();
    case 419:
      if (lookahead == 'i') ADVANCE(445);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(anon_sym_telldir);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(anon_sym_ucfirst);
      END_STATE();
    case 422:
      if (lookahead == 'a') ADVANCE(446);
      END_STATE();
    case 423:
      if (lookahead == 'K') ADVANCE(447);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(anon_sym___DATA__);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(anon_sym___FILE__);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(anon_sym___LINE__);
      END_STATE();
    case 427:
      if (lookahead == 'E') ADVANCE(448);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(anon_sym_dbmclose);
      END_STATE();
    case 430:
      if (lookahead == 'l') ADVANCE(449);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(anon_sym_extended);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(anon_sym_getgrgid);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(anon_sym_getgrnam);
      END_STATE();
    case 434:
      if (lookahead == 'n') ADVANCE(450);
      END_STATE();
    case 435:
      if (lookahead == 'a') ADVANCE(451);
      END_STATE();
    case 436:
      if (lookahead == 'b') ADVANCE(452);
      END_STATE();
    case 437:
      if (lookahead == 'e') ADVANCE(453);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(anon_sym_getpwuid);
      END_STATE();
    case 439:
      if (lookahead == 'a') ADVANCE(454);
      END_STATE();
    case 440:
      if (lookahead == 'e') ADVANCE(455);
      END_STATE();
    case 441:
      if (lookahead == 'e') ADVANCE(456);
      END_STATE();
    case 442:
      if (lookahead == 'a') ADVANCE(457);
      END_STATE();
    case 443:
      ACCEPT_TOKEN(anon_sym_readline);
      END_STATE();
    case 444:
      ACCEPT_TOKEN(anon_sym_readlink);
      END_STATE();
    case 445:
      if (lookahead == 'r') ADVANCE(458);
      END_STATE();
    case 446:
      if (lookahead == 'y') ADVANCE(459);
      END_STATE();
    case 447:
      ACCEPT_TOKEN(anon_sym_UNITCHECK);
      END_STATE();
    case 448:
      if (lookahead == '_') ADVANCE(460);
      END_STATE();
    case 449:
      if (lookahead == 'l') ADVANCE(461);
      END_STATE();
    case 450:
      if (lookahead == 'a') ADVANCE(462);
      END_STATE();
    case 451:
      if (lookahead == 'm') ADVANCE(463);
      END_STATE();
    case 452:
      if (lookahead == 'y') ADVANCE(464);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(anon_sym_getpwname);
      END_STATE();
    case 454:
      if (lookahead == 'm') ADVANCE(465);
      END_STATE();
    case 455:
      ACCEPT_TOKEN(anon_sym_localtime);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(anon_sym_prototype);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(anon_sym_quotemeta);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(anon_sym_rewinddir);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(anon_sym_wantarray);
      END_STATE();
    case 460:
      if (lookahead == '_') ADVANCE(466);
      END_STATE();
    case 461:
      if (lookahead == 'y') ADVANCE(467);
      END_STATE();
    case 462:
      if (lookahead == 'm') ADVANCE(468);
      END_STATE();
    case 463:
      if (lookahead == 'e') ADVANCE(469);
      END_STATE();
    case 464:
      if (lookahead == 'n') ADVANCE(470);
      END_STATE();
    case 465:
      if (lookahead == 'e') ADVANCE(471);
      END_STATE();
    case 466:
      ACCEPT_TOKEN(anon_sym___PACKAGE__);
      END_STATE();
    case 467:
      ACCEPT_TOKEN(anon_sym_dynamically);
      END_STATE();
    case 468:
      if (lookahead == 'e') ADVANCE(472);
      END_STATE();
    case 469:
      ACCEPT_TOKEN(anon_sym_getpeername);
      END_STATE();
    case 470:
      if (lookahead == 'a') ADVANCE(473);
      END_STATE();
    case 471:
      ACCEPT_TOKEN(anon_sym_getsockname);
      END_STATE();
    case 472:
      ACCEPT_TOKEN(anon_sym_getnetbyname);
      END_STATE();
    case 473:
      if (lookahead == 'm') ADVANCE(474);
      END_STATE();
    case 474:
      if (lookahead == 'e') ADVANCE(475);
      END_STATE();
    case 475:
      ACCEPT_TOKEN(anon_sym_getprotobyname);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 112, .external_lex_state = 2},
  [2] = {.lex_state = 1, .external_lex_state = 3},
  [3] = {.lex_state = 1, .external_lex_state = 3},
  [4] = {.lex_state = 1, .external_lex_state = 3},
  [5] = {.lex_state = 1, .external_lex_state = 3},
  [6] = {.lex_state = 1, .external_lex_state = 3},
  [7] = {.lex_state = 1, .external_lex_state = 3},
  [8] = {.lex_state = 1, .external_lex_state = 3},
  [9] = {.lex_state = 1, .external_lex_state = 3},
  [10] = {.lex_state = 1, .external_lex_state = 3},
  [11] = {.lex_state = 1, .external_lex_state = 3},
  [12] = {.lex_state = 1, .external_lex_state = 4},
  [13] = {.lex_state = 1, .external_lex_state = 4},
  [14] = {.lex_state = 1, .external_lex_state = 5},
  [15] = {.lex_state = 1, .external_lex_state = 6},
  [16] = {.lex_state = 1, .external_lex_state = 7},
  [17] = {.lex_state = 1, .external_lex_state = 6},
  [18] = {.lex_state = 1, .external_lex_state = 5},
  [19] = {.lex_state = 1, .external_lex_state = 7},
  [20] = {.lex_state = 1, .external_lex_state = 4},
  [21] = {.lex_state = 1, .external_lex_state = 4},
  [22] = {.lex_state = 1, .external_lex_state = 4},
  [23] = {.lex_state = 1, .external_lex_state = 6},
  [24] = {.lex_state = 1, .external_lex_state = 7},
  [25] = {.lex_state = 1, .external_lex_state = 7},
  [26] = {.lex_state = 1, .external_lex_state = 7},
  [27] = {.lex_state = 1, .external_lex_state = 5},
  [28] = {.lex_state = 1, .external_lex_state = 6},
  [29] = {.lex_state = 1, .external_lex_state = 5},
  [30] = {.lex_state = 1, .external_lex_state = 6},
  [31] = {.lex_state = 1, .external_lex_state = 5},
  [32] = {.lex_state = 51, .external_lex_state = 8},
  [33] = {.lex_state = 51, .external_lex_state = 8},
  [34] = {.lex_state = 112, .external_lex_state = 2},
  [35] = {.lex_state = 51, .external_lex_state = 4},
  [36] = {.lex_state = 51, .external_lex_state = 4},
  [37] = {.lex_state = 51, .external_lex_state = 4},
  [38] = {.lex_state = 51, .external_lex_state = 4},
  [39] = {.lex_state = 54, .external_lex_state = 9},
  [40] = {.lex_state = 51, .external_lex_state = 4},
  [41] = {.lex_state = 51, .external_lex_state = 4},
  [42] = {.lex_state = 54, .external_lex_state = 9},
  [43] = {.lex_state = 112, .external_lex_state = 2},
  [44] = {.lex_state = 54, .external_lex_state = 9},
  [45] = {.lex_state = 54, .external_lex_state = 9},
  [46] = {.lex_state = 54, .external_lex_state = 9},
  [47] = {.lex_state = 54, .external_lex_state = 9},
  [48] = {.lex_state = 54, .external_lex_state = 9},
  [49] = {.lex_state = 54, .external_lex_state = 9},
  [50] = {.lex_state = 54, .external_lex_state = 9},
  [51] = {.lex_state = 54, .external_lex_state = 9},
  [52] = {.lex_state = 54, .external_lex_state = 9},
  [53] = {.lex_state = 54, .external_lex_state = 9},
  [54] = {.lex_state = 54, .external_lex_state = 9},
  [55] = {.lex_state = 54, .external_lex_state = 9},
  [56] = {.lex_state = 54, .external_lex_state = 9},
  [57] = {.lex_state = 54, .external_lex_state = 9},
  [58] = {.lex_state = 54, .external_lex_state = 9},
  [59] = {.lex_state = 54, .external_lex_state = 9},
  [60] = {.lex_state = 54, .external_lex_state = 9},
  [61] = {.lex_state = 54, .external_lex_state = 9},
  [62] = {.lex_state = 54, .external_lex_state = 9},
  [63] = {.lex_state = 54, .external_lex_state = 9},
  [64] = {.lex_state = 54, .external_lex_state = 9},
  [65] = {.lex_state = 54, .external_lex_state = 9},
  [66] = {.lex_state = 54, .external_lex_state = 9},
  [67] = {.lex_state = 54, .external_lex_state = 9},
  [68] = {.lex_state = 54, .external_lex_state = 9},
  [69] = {.lex_state = 54, .external_lex_state = 9},
  [70] = {.lex_state = 54, .external_lex_state = 9},
  [71] = {.lex_state = 54, .external_lex_state = 9},
  [72] = {.lex_state = 54, .external_lex_state = 9},
  [73] = {.lex_state = 54, .external_lex_state = 9},
  [74] = {.lex_state = 54, .external_lex_state = 9},
  [75] = {.lex_state = 54, .external_lex_state = 9},
  [76] = {.lex_state = 54, .external_lex_state = 9},
  [77] = {.lex_state = 54, .external_lex_state = 9},
  [78] = {.lex_state = 51, .external_lex_state = 4},
  [79] = {.lex_state = 51, .external_lex_state = 4},
  [80] = {.lex_state = 54, .external_lex_state = 9},
  [81] = {.lex_state = 51, .external_lex_state = 4},
  [82] = {.lex_state = 51, .external_lex_state = 4},
  [83] = {.lex_state = 51, .external_lex_state = 4},
  [84] = {.lex_state = 51, .external_lex_state = 8},
  [85] = {.lex_state = 51, .external_lex_state = 8},
  [86] = {.lex_state = 51, .external_lex_state = 4},
  [87] = {.lex_state = 51, .external_lex_state = 4},
  [88] = {.lex_state = 51, .external_lex_state = 4},
  [89] = {.lex_state = 51, .external_lex_state = 4},
  [90] = {.lex_state = 51, .external_lex_state = 8},
  [91] = {.lex_state = 51, .external_lex_state = 4},
  [92] = {.lex_state = 51, .external_lex_state = 4},
  [93] = {.lex_state = 51, .external_lex_state = 4},
  [94] = {.lex_state = 51, .external_lex_state = 4},
  [95] = {.lex_state = 51, .external_lex_state = 4},
  [96] = {.lex_state = 51, .external_lex_state = 4},
  [97] = {.lex_state = 51, .external_lex_state = 4},
  [98] = {.lex_state = 51, .external_lex_state = 4},
  [99] = {.lex_state = 51, .external_lex_state = 4},
  [100] = {.lex_state = 51, .external_lex_state = 4},
  [101] = {.lex_state = 112, .external_lex_state = 4},
  [102] = {.lex_state = 112, .external_lex_state = 4},
  [103] = {.lex_state = 112, .external_lex_state = 4},
  [104] = {.lex_state = 112, .external_lex_state = 4},
  [105] = {.lex_state = 112, .external_lex_state = 4},
  [106] = {.lex_state = 51, .external_lex_state = 4},
  [107] = {.lex_state = 51, .external_lex_state = 4},
  [108] = {.lex_state = 112, .external_lex_state = 4},
  [109] = {.lex_state = 112, .external_lex_state = 4},
  [110] = {.lex_state = 112, .external_lex_state = 4},
  [111] = {.lex_state = 112, .external_lex_state = 4},
  [112] = {.lex_state = 112, .external_lex_state = 4},
  [113] = {.lex_state = 112, .external_lex_state = 4},
  [114] = {.lex_state = 112, .external_lex_state = 4},
  [115] = {.lex_state = 112, .external_lex_state = 4},
  [116] = {.lex_state = 112, .external_lex_state = 4},
  [117] = {.lex_state = 112, .external_lex_state = 4},
  [118] = {.lex_state = 112, .external_lex_state = 4},
  [119] = {.lex_state = 112, .external_lex_state = 4},
  [120] = {.lex_state = 112, .external_lex_state = 4},
  [121] = {.lex_state = 112, .external_lex_state = 4},
  [122] = {.lex_state = 112, .external_lex_state = 4},
  [123] = {.lex_state = 112, .external_lex_state = 4},
  [124] = {.lex_state = 112, .external_lex_state = 4},
  [125] = {.lex_state = 112, .external_lex_state = 4},
  [126] = {.lex_state = 112, .external_lex_state = 4},
  [127] = {.lex_state = 112, .external_lex_state = 4},
  [128] = {.lex_state = 112, .external_lex_state = 4},
  [129] = {.lex_state = 112, .external_lex_state = 4},
  [130] = {.lex_state = 112, .external_lex_state = 4},
  [131] = {.lex_state = 112, .external_lex_state = 4},
  [132] = {.lex_state = 51, .external_lex_state = 4},
  [133] = {.lex_state = 51, .external_lex_state = 4},
  [134] = {.lex_state = 112, .external_lex_state = 4},
  [135] = {.lex_state = 112, .external_lex_state = 4},
  [136] = {.lex_state = 112, .external_lex_state = 4},
  [137] = {.lex_state = 112, .external_lex_state = 4},
  [138] = {.lex_state = 112, .external_lex_state = 4},
  [139] = {.lex_state = 112, .external_lex_state = 4},
  [140] = {.lex_state = 112, .external_lex_state = 4},
  [141] = {.lex_state = 112, .external_lex_state = 4},
  [142] = {.lex_state = 112, .external_lex_state = 4},
  [143] = {.lex_state = 112, .external_lex_state = 4},
  [144] = {.lex_state = 112, .external_lex_state = 4},
  [145] = {.lex_state = 112, .external_lex_state = 4},
  [146] = {.lex_state = 112, .external_lex_state = 4},
  [147] = {.lex_state = 112, .external_lex_state = 4},
  [148] = {.lex_state = 112, .external_lex_state = 4},
  [149] = {.lex_state = 112, .external_lex_state = 4},
  [150] = {.lex_state = 112, .external_lex_state = 4},
  [151] = {.lex_state = 112, .external_lex_state = 4},
  [152] = {.lex_state = 112, .external_lex_state = 4},
  [153] = {.lex_state = 112, .external_lex_state = 4},
  [154] = {.lex_state = 112, .external_lex_state = 4},
  [155] = {.lex_state = 112, .external_lex_state = 4},
  [156] = {.lex_state = 112, .external_lex_state = 4},
  [157] = {.lex_state = 112, .external_lex_state = 4},
  [158] = {.lex_state = 112, .external_lex_state = 4},
  [159] = {.lex_state = 51, .external_lex_state = 4},
  [160] = {.lex_state = 112, .external_lex_state = 4},
  [161] = {.lex_state = 112, .external_lex_state = 4},
  [162] = {.lex_state = 112, .external_lex_state = 4},
  [163] = {.lex_state = 112, .external_lex_state = 4},
  [164] = {.lex_state = 112, .external_lex_state = 4},
  [165] = {.lex_state = 112, .external_lex_state = 4},
  [166] = {.lex_state = 112, .external_lex_state = 4},
  [167] = {.lex_state = 112, .external_lex_state = 4},
  [168] = {.lex_state = 112, .external_lex_state = 4},
  [169] = {.lex_state = 112, .external_lex_state = 4},
  [170] = {.lex_state = 112, .external_lex_state = 4},
  [171] = {.lex_state = 112, .external_lex_state = 4},
  [172] = {.lex_state = 112, .external_lex_state = 4},
  [173] = {.lex_state = 112, .external_lex_state = 4},
  [174] = {.lex_state = 112, .external_lex_state = 4},
  [175] = {.lex_state = 112, .external_lex_state = 4},
  [176] = {.lex_state = 112, .external_lex_state = 4},
  [177] = {.lex_state = 112, .external_lex_state = 4},
  [178] = {.lex_state = 112, .external_lex_state = 4},
  [179] = {.lex_state = 112, .external_lex_state = 4},
  [180] = {.lex_state = 112, .external_lex_state = 4},
  [181] = {.lex_state = 112, .external_lex_state = 4},
  [182] = {.lex_state = 112, .external_lex_state = 4},
  [183] = {.lex_state = 112, .external_lex_state = 4},
  [184] = {.lex_state = 112, .external_lex_state = 4},
  [185] = {.lex_state = 112, .external_lex_state = 4},
  [186] = {.lex_state = 112, .external_lex_state = 4},
  [187] = {.lex_state = 112, .external_lex_state = 4},
  [188] = {.lex_state = 112, .external_lex_state = 4},
  [189] = {.lex_state = 112, .external_lex_state = 4},
  [190] = {.lex_state = 112, .external_lex_state = 4},
  [191] = {.lex_state = 112, .external_lex_state = 4},
  [192] = {.lex_state = 112, .external_lex_state = 4},
  [193] = {.lex_state = 112, .external_lex_state = 4},
  [194] = {.lex_state = 112, .external_lex_state = 4},
  [195] = {.lex_state = 112, .external_lex_state = 4},
  [196] = {.lex_state = 112, .external_lex_state = 4},
  [197] = {.lex_state = 112, .external_lex_state = 4},
  [198] = {.lex_state = 112, .external_lex_state = 4},
  [199] = {.lex_state = 112, .external_lex_state = 4},
  [200] = {.lex_state = 112, .external_lex_state = 4},
  [201] = {.lex_state = 112, .external_lex_state = 4},
  [202] = {.lex_state = 112, .external_lex_state = 4},
  [203] = {.lex_state = 51, .external_lex_state = 4},
  [204] = {.lex_state = 112, .external_lex_state = 4},
  [205] = {.lex_state = 112, .external_lex_state = 4},
  [206] = {.lex_state = 112, .external_lex_state = 4},
  [207] = {.lex_state = 112, .external_lex_state = 4},
  [208] = {.lex_state = 112, .external_lex_state = 4},
  [209] = {.lex_state = 112, .external_lex_state = 4},
  [210] = {.lex_state = 112, .external_lex_state = 4},
  [211] = {.lex_state = 112, .external_lex_state = 4},
  [212] = {.lex_state = 112, .external_lex_state = 4},
  [213] = {.lex_state = 112, .external_lex_state = 4},
  [214] = {.lex_state = 112, .external_lex_state = 4},
  [215] = {.lex_state = 2, .external_lex_state = 3},
  [216] = {.lex_state = 2, .external_lex_state = 3},
  [217] = {.lex_state = 52, .external_lex_state = 5},
  [218] = {.lex_state = 52, .external_lex_state = 5},
  [219] = {.lex_state = 52, .external_lex_state = 5},
  [220] = {.lex_state = 52, .external_lex_state = 5},
  [221] = {.lex_state = 52, .external_lex_state = 5},
  [222] = {.lex_state = 52, .external_lex_state = 5},
  [223] = {.lex_state = 2, .external_lex_state = 4},
  [224] = {.lex_state = 52, .external_lex_state = 5},
  [225] = {.lex_state = 52, .external_lex_state = 5},
  [226] = {.lex_state = 52, .external_lex_state = 9},
  [227] = {.lex_state = 52, .external_lex_state = 9},
  [228] = {.lex_state = 52, .external_lex_state = 9},
  [229] = {.lex_state = 2, .external_lex_state = 7},
  [230] = {.lex_state = 52, .external_lex_state = 9},
  [231] = {.lex_state = 52, .external_lex_state = 9},
  [232] = {.lex_state = 52, .external_lex_state = 9},
  [233] = {.lex_state = 52, .external_lex_state = 9},
  [234] = {.lex_state = 55, .external_lex_state = 8},
  [235] = {.lex_state = 52, .external_lex_state = 9},
  [236] = {.lex_state = 52, .external_lex_state = 9},
  [237] = {.lex_state = 52, .external_lex_state = 9},
  [238] = {.lex_state = 52, .external_lex_state = 9},
  [239] = {.lex_state = 52, .external_lex_state = 9},
  [240] = {.lex_state = 52, .external_lex_state = 9},
  [241] = {.lex_state = 52, .external_lex_state = 9},
  [242] = {.lex_state = 52, .external_lex_state = 9},
  [243] = {.lex_state = 55, .external_lex_state = 8},
  [244] = {.lex_state = 2, .external_lex_state = 5},
  [245] = {.lex_state = 55, .external_lex_state = 8},
  [246] = {.lex_state = 52, .external_lex_state = 9},
  [247] = {.lex_state = 52, .external_lex_state = 9},
  [248] = {.lex_state = 52, .external_lex_state = 9},
  [249] = {.lex_state = 52, .external_lex_state = 9},
  [250] = {.lex_state = 52, .external_lex_state = 9},
  [251] = {.lex_state = 52, .external_lex_state = 9},
  [252] = {.lex_state = 52, .external_lex_state = 9},
  [253] = {.lex_state = 52, .external_lex_state = 9},
  [254] = {.lex_state = 52, .external_lex_state = 9},
  [255] = {.lex_state = 52, .external_lex_state = 9},
  [256] = {.lex_state = 52, .external_lex_state = 9},
  [257] = {.lex_state = 52, .external_lex_state = 9},
  [258] = {.lex_state = 52, .external_lex_state = 9},
  [259] = {.lex_state = 52, .external_lex_state = 9},
  [260] = {.lex_state = 2, .external_lex_state = 6},
  [261] = {.lex_state = 52, .external_lex_state = 9},
  [262] = {.lex_state = 52, .external_lex_state = 9},
  [263] = {.lex_state = 52, .external_lex_state = 9},
  [264] = {.lex_state = 52, .external_lex_state = 9},
  [265] = {.lex_state = 52, .external_lex_state = 9},
  [266] = {.lex_state = 52, .external_lex_state = 9},
  [267] = {.lex_state = 52, .external_lex_state = 9},
  [268] = {.lex_state = 52, .external_lex_state = 9},
  [269] = {.lex_state = 52, .external_lex_state = 9},
  [270] = {.lex_state = 52, .external_lex_state = 9},
  [271] = {.lex_state = 52, .external_lex_state = 9},
  [272] = {.lex_state = 52, .external_lex_state = 9},
  [273] = {.lex_state = 52, .external_lex_state = 9},
  [274] = {.lex_state = 52, .external_lex_state = 9},
  [275] = {.lex_state = 52, .external_lex_state = 9},
  [276] = {.lex_state = 52, .external_lex_state = 9},
  [277] = {.lex_state = 52, .external_lex_state = 9},
  [278] = {.lex_state = 52, .external_lex_state = 9},
  [279] = {.lex_state = 52, .external_lex_state = 9},
  [280] = {.lex_state = 52, .external_lex_state = 9},
  [281] = {.lex_state = 52, .external_lex_state = 9},
  [282] = {.lex_state = 52, .external_lex_state = 9},
  [283] = {.lex_state = 52, .external_lex_state = 9},
  [284] = {.lex_state = 52, .external_lex_state = 9},
  [285] = {.lex_state = 52, .external_lex_state = 9},
  [286] = {.lex_state = 52, .external_lex_state = 9},
  [287] = {.lex_state = 52, .external_lex_state = 9},
  [288] = {.lex_state = 52, .external_lex_state = 9},
  [289] = {.lex_state = 52, .external_lex_state = 9},
  [290] = {.lex_state = 52, .external_lex_state = 9},
  [291] = {.lex_state = 52, .external_lex_state = 9},
  [292] = {.lex_state = 52, .external_lex_state = 9},
  [293] = {.lex_state = 52, .external_lex_state = 9},
  [294] = {.lex_state = 52, .external_lex_state = 9},
  [295] = {.lex_state = 52, .external_lex_state = 9},
  [296] = {.lex_state = 52, .external_lex_state = 9},
  [297] = {.lex_state = 52, .external_lex_state = 9},
  [298] = {.lex_state = 52, .external_lex_state = 9},
  [299] = {.lex_state = 52, .external_lex_state = 9},
  [300] = {.lex_state = 52, .external_lex_state = 9},
  [301] = {.lex_state = 52, .external_lex_state = 9},
  [302] = {.lex_state = 52, .external_lex_state = 9},
  [303] = {.lex_state = 52, .external_lex_state = 9},
  [304] = {.lex_state = 52, .external_lex_state = 9},
  [305] = {.lex_state = 52, .external_lex_state = 9},
  [306] = {.lex_state = 52, .external_lex_state = 9},
  [307] = {.lex_state = 52, .external_lex_state = 9},
  [308] = {.lex_state = 52, .external_lex_state = 9},
  [309] = {.lex_state = 52, .external_lex_state = 9},
  [310] = {.lex_state = 52, .external_lex_state = 9},
  [311] = {.lex_state = 52, .external_lex_state = 9},
  [312] = {.lex_state = 52, .external_lex_state = 9},
  [313] = {.lex_state = 52, .external_lex_state = 9},
  [314] = {.lex_state = 52, .external_lex_state = 9},
  [315] = {.lex_state = 52, .external_lex_state = 9},
  [316] = {.lex_state = 52, .external_lex_state = 9},
  [317] = {.lex_state = 52, .external_lex_state = 9},
  [318] = {.lex_state = 52, .external_lex_state = 9},
  [319] = {.lex_state = 52, .external_lex_state = 9},
  [320] = {.lex_state = 52, .external_lex_state = 9},
  [321] = {.lex_state = 52, .external_lex_state = 9},
  [322] = {.lex_state = 52, .external_lex_state = 9},
  [323] = {.lex_state = 52, .external_lex_state = 9},
  [324] = {.lex_state = 52, .external_lex_state = 4},
  [325] = {.lex_state = 52, .external_lex_state = 10},
  [326] = {.lex_state = 52, .external_lex_state = 5},
  [327] = {.lex_state = 52, .external_lex_state = 5},
  [328] = {.lex_state = 52, .external_lex_state = 5},
  [329] = {.lex_state = 52, .external_lex_state = 7},
  [330] = {.lex_state = 52, .external_lex_state = 5},
  [331] = {.lex_state = 52, .external_lex_state = 10},
  [332] = {.lex_state = 52, .external_lex_state = 10},
  [333] = {.lex_state = 52, .external_lex_state = 4},
  [334] = {.lex_state = 52, .external_lex_state = 5},
  [335] = {.lex_state = 52, .external_lex_state = 7},
  [336] = {.lex_state = 52, .external_lex_state = 5},
  [337] = {.lex_state = 52, .external_lex_state = 10},
  [338] = {.lex_state = 52, .external_lex_state = 5},
  [339] = {.lex_state = 52, .external_lex_state = 5},
  [340] = {.lex_state = 52, .external_lex_state = 5},
  [341] = {.lex_state = 52, .external_lex_state = 5},
  [342] = {.lex_state = 52, .external_lex_state = 5},
  [343] = {.lex_state = 52, .external_lex_state = 7},
  [344] = {.lex_state = 52, .external_lex_state = 5},
  [345] = {.lex_state = 52, .external_lex_state = 5},
  [346] = {.lex_state = 52, .external_lex_state = 10},
  [347] = {.lex_state = 52, .external_lex_state = 7},
  [348] = {.lex_state = 52, .external_lex_state = 4},
  [349] = {.lex_state = 52, .external_lex_state = 7},
  [350] = {.lex_state = 52, .external_lex_state = 5},
  [351] = {.lex_state = 52, .external_lex_state = 7},
  [352] = {.lex_state = 52, .external_lex_state = 4},
  [353] = {.lex_state = 52, .external_lex_state = 10},
  [354] = {.lex_state = 52, .external_lex_state = 5},
  [355] = {.lex_state = 52, .external_lex_state = 5},
  [356] = {.lex_state = 52, .external_lex_state = 5},
  [357] = {.lex_state = 52, .external_lex_state = 5},
  [358] = {.lex_state = 52, .external_lex_state = 5},
  [359] = {.lex_state = 52, .external_lex_state = 7},
  [360] = {.lex_state = 52, .external_lex_state = 4},
  [361] = {.lex_state = 52, .external_lex_state = 10},
  [362] = {.lex_state = 52, .external_lex_state = 4},
  [363] = {.lex_state = 52, .external_lex_state = 4},
  [364] = {.lex_state = 52, .external_lex_state = 10},
  [365] = {.lex_state = 52, .external_lex_state = 5},
  [366] = {.lex_state = 52, .external_lex_state = 10},
  [367] = {.lex_state = 52, .external_lex_state = 5},
  [368] = {.lex_state = 52, .external_lex_state = 5},
  [369] = {.lex_state = 52, .external_lex_state = 5},
  [370] = {.lex_state = 52, .external_lex_state = 5},
  [371] = {.lex_state = 52, .external_lex_state = 5},
  [372] = {.lex_state = 52, .external_lex_state = 7},
  [373] = {.lex_state = 52, .external_lex_state = 5},
  [374] = {.lex_state = 52, .external_lex_state = 4},
  [375] = {.lex_state = 52, .external_lex_state = 5},
  [376] = {.lex_state = 52, .external_lex_state = 5},
  [377] = {.lex_state = 52, .external_lex_state = 5},
  [378] = {.lex_state = 52, .external_lex_state = 5},
  [379] = {.lex_state = 52, .external_lex_state = 5},
  [380] = {.lex_state = 52, .external_lex_state = 5},
  [381] = {.lex_state = 52, .external_lex_state = 10},
  [382] = {.lex_state = 52, .external_lex_state = 10},
  [383] = {.lex_state = 52, .external_lex_state = 10},
  [384] = {.lex_state = 52, .external_lex_state = 10},
  [385] = {.lex_state = 52, .external_lex_state = 10},
  [386] = {.lex_state = 52, .external_lex_state = 10},
  [387] = {.lex_state = 52, .external_lex_state = 10},
  [388] = {.lex_state = 52, .external_lex_state = 4},
  [389] = {.lex_state = 52, .external_lex_state = 4},
  [390] = {.lex_state = 52, .external_lex_state = 4},
  [391] = {.lex_state = 52, .external_lex_state = 4},
  [392] = {.lex_state = 52, .external_lex_state = 8},
  [393] = {.lex_state = 52, .external_lex_state = 4},
  [394] = {.lex_state = 52, .external_lex_state = 8},
  [395] = {.lex_state = 52, .external_lex_state = 4},
  [396] = {.lex_state = 52, .external_lex_state = 4},
  [397] = {.lex_state = 52, .external_lex_state = 4},
  [398] = {.lex_state = 52, .external_lex_state = 4},
  [399] = {.lex_state = 52, .external_lex_state = 4},
  [400] = {.lex_state = 52, .external_lex_state = 4},
  [401] = {.lex_state = 52, .external_lex_state = 4},
  [402] = {.lex_state = 52, .external_lex_state = 4},
  [403] = {.lex_state = 52, .external_lex_state = 4},
  [404] = {.lex_state = 52, .external_lex_state = 4},
  [405] = {.lex_state = 52, .external_lex_state = 4},
  [406] = {.lex_state = 52, .external_lex_state = 4},
  [407] = {.lex_state = 52, .external_lex_state = 4},
  [408] = {.lex_state = 52, .external_lex_state = 4},
  [409] = {.lex_state = 52, .external_lex_state = 4},
  [410] = {.lex_state = 52, .external_lex_state = 4},
  [411] = {.lex_state = 52, .external_lex_state = 4},
  [412] = {.lex_state = 52, .external_lex_state = 8},
  [413] = {.lex_state = 52, .external_lex_state = 4},
  [414] = {.lex_state = 52, .external_lex_state = 4},
  [415] = {.lex_state = 52, .external_lex_state = 4},
  [416] = {.lex_state = 52, .external_lex_state = 4},
  [417] = {.lex_state = 52, .external_lex_state = 4},
  [418] = {.lex_state = 52, .external_lex_state = 4},
  [419] = {.lex_state = 52, .external_lex_state = 4},
  [420] = {.lex_state = 52, .external_lex_state = 4},
  [421] = {.lex_state = 52, .external_lex_state = 4},
  [422] = {.lex_state = 52, .external_lex_state = 4},
  [423] = {.lex_state = 52, .external_lex_state = 4},
  [424] = {.lex_state = 52, .external_lex_state = 4},
  [425] = {.lex_state = 52, .external_lex_state = 4},
  [426] = {.lex_state = 52, .external_lex_state = 4},
  [427] = {.lex_state = 52, .external_lex_state = 4},
  [428] = {.lex_state = 52, .external_lex_state = 4},
  [429] = {.lex_state = 52, .external_lex_state = 4},
  [430] = {.lex_state = 52, .external_lex_state = 4},
  [431] = {.lex_state = 52, .external_lex_state = 4},
  [432] = {.lex_state = 52, .external_lex_state = 4},
  [433] = {.lex_state = 52, .external_lex_state = 4},
  [434] = {.lex_state = 52, .external_lex_state = 4},
  [435] = {.lex_state = 52, .external_lex_state = 4},
  [436] = {.lex_state = 52, .external_lex_state = 4},
  [437] = {.lex_state = 52, .external_lex_state = 4},
  [438] = {.lex_state = 52, .external_lex_state = 4},
  [439] = {.lex_state = 52, .external_lex_state = 4},
  [440] = {.lex_state = 52, .external_lex_state = 4},
  [441] = {.lex_state = 52, .external_lex_state = 4},
  [442] = {.lex_state = 52, .external_lex_state = 4},
  [443] = {.lex_state = 52, .external_lex_state = 4},
  [444] = {.lex_state = 52, .external_lex_state = 4},
  [445] = {.lex_state = 52, .external_lex_state = 4},
  [446] = {.lex_state = 52, .external_lex_state = 4},
  [447] = {.lex_state = 52, .external_lex_state = 4},
  [448] = {.lex_state = 52, .external_lex_state = 4},
  [449] = {.lex_state = 52, .external_lex_state = 4},
  [450] = {.lex_state = 52, .external_lex_state = 4},
  [451] = {.lex_state = 52, .external_lex_state = 4},
  [452] = {.lex_state = 52, .external_lex_state = 4},
  [453] = {.lex_state = 52, .external_lex_state = 4},
  [454] = {.lex_state = 52, .external_lex_state = 4},
  [455] = {.lex_state = 52, .external_lex_state = 4},
  [456] = {.lex_state = 52, .external_lex_state = 4},
  [457] = {.lex_state = 52, .external_lex_state = 4},
  [458] = {.lex_state = 52, .external_lex_state = 4},
  [459] = {.lex_state = 52, .external_lex_state = 4},
  [460] = {.lex_state = 52, .external_lex_state = 4},
  [461] = {.lex_state = 52, .external_lex_state = 4},
  [462] = {.lex_state = 52, .external_lex_state = 4},
  [463] = {.lex_state = 52, .external_lex_state = 4},
  [464] = {.lex_state = 52, .external_lex_state = 4},
  [465] = {.lex_state = 52, .external_lex_state = 4},
  [466] = {.lex_state = 52, .external_lex_state = 4},
  [467] = {.lex_state = 52, .external_lex_state = 4},
  [468] = {.lex_state = 52, .external_lex_state = 4},
  [469] = {.lex_state = 52, .external_lex_state = 4},
  [470] = {.lex_state = 52, .external_lex_state = 4},
  [471] = {.lex_state = 52, .external_lex_state = 4},
  [472] = {.lex_state = 52, .external_lex_state = 4},
  [473] = {.lex_state = 52, .external_lex_state = 4},
  [474] = {.lex_state = 52, .external_lex_state = 4},
  [475] = {.lex_state = 52, .external_lex_state = 4},
  [476] = {.lex_state = 52, .external_lex_state = 4},
  [477] = {.lex_state = 52, .external_lex_state = 4},
  [478] = {.lex_state = 52, .external_lex_state = 4},
  [479] = {.lex_state = 52, .external_lex_state = 4},
  [480] = {.lex_state = 52, .external_lex_state = 4},
  [481] = {.lex_state = 52, .external_lex_state = 4},
  [482] = {.lex_state = 52, .external_lex_state = 4},
  [483] = {.lex_state = 52, .external_lex_state = 4},
  [484] = {.lex_state = 52, .external_lex_state = 4},
  [485] = {.lex_state = 52, .external_lex_state = 4},
  [486] = {.lex_state = 52, .external_lex_state = 4},
  [487] = {.lex_state = 52, .external_lex_state = 4},
  [488] = {.lex_state = 52, .external_lex_state = 4},
  [489] = {.lex_state = 52, .external_lex_state = 4},
  [490] = {.lex_state = 52, .external_lex_state = 4},
  [491] = {.lex_state = 52, .external_lex_state = 4},
  [492] = {.lex_state = 52, .external_lex_state = 4},
  [493] = {.lex_state = 52, .external_lex_state = 4},
  [494] = {.lex_state = 52, .external_lex_state = 4},
  [495] = {.lex_state = 52, .external_lex_state = 4},
  [496] = {.lex_state = 52, .external_lex_state = 4},
  [497] = {.lex_state = 52, .external_lex_state = 4},
  [498] = {.lex_state = 52, .external_lex_state = 4},
  [499] = {.lex_state = 52, .external_lex_state = 4},
  [500] = {.lex_state = 52, .external_lex_state = 4},
  [501] = {.lex_state = 52, .external_lex_state = 4},
  [502] = {.lex_state = 52, .external_lex_state = 4},
  [503] = {.lex_state = 52, .external_lex_state = 4},
  [504] = {.lex_state = 52, .external_lex_state = 4},
  [505] = {.lex_state = 52, .external_lex_state = 4},
  [506] = {.lex_state = 52, .external_lex_state = 4},
  [507] = {.lex_state = 52, .external_lex_state = 4},
  [508] = {.lex_state = 52, .external_lex_state = 4},
  [509] = {.lex_state = 52, .external_lex_state = 4},
  [510] = {.lex_state = 52, .external_lex_state = 4},
  [511] = {.lex_state = 52, .external_lex_state = 4},
  [512] = {.lex_state = 52, .external_lex_state = 4},
  [513] = {.lex_state = 52, .external_lex_state = 4},
  [514] = {.lex_state = 52, .external_lex_state = 4},
  [515] = {.lex_state = 52, .external_lex_state = 4},
  [516] = {.lex_state = 52, .external_lex_state = 4},
  [517] = {.lex_state = 52, .external_lex_state = 4},
  [518] = {.lex_state = 52, .external_lex_state = 4},
  [519] = {.lex_state = 52, .external_lex_state = 4},
  [520] = {.lex_state = 52, .external_lex_state = 4},
  [521] = {.lex_state = 52, .external_lex_state = 4},
  [522] = {.lex_state = 52, .external_lex_state = 4},
  [523] = {.lex_state = 52, .external_lex_state = 4},
  [524] = {.lex_state = 52, .external_lex_state = 4},
  [525] = {.lex_state = 52, .external_lex_state = 4},
  [526] = {.lex_state = 52, .external_lex_state = 4},
  [527] = {.lex_state = 52, .external_lex_state = 4},
  [528] = {.lex_state = 52, .external_lex_state = 4},
  [529] = {.lex_state = 52, .external_lex_state = 4},
  [530] = {.lex_state = 52, .external_lex_state = 4},
  [531] = {.lex_state = 52, .external_lex_state = 4},
  [532] = {.lex_state = 52, .external_lex_state = 4},
  [533] = {.lex_state = 52, .external_lex_state = 4},
  [534] = {.lex_state = 52, .external_lex_state = 4},
  [535] = {.lex_state = 52, .external_lex_state = 4},
  [536] = {.lex_state = 52, .external_lex_state = 4},
  [537] = {.lex_state = 52, .external_lex_state = 4},
  [538] = {.lex_state = 52, .external_lex_state = 4},
  [539] = {.lex_state = 52, .external_lex_state = 4},
  [540] = {.lex_state = 52, .external_lex_state = 4},
  [541] = {.lex_state = 52, .external_lex_state = 4},
  [542] = {.lex_state = 52, .external_lex_state = 4},
  [543] = {.lex_state = 52, .external_lex_state = 4},
  [544] = {.lex_state = 52, .external_lex_state = 4},
  [545] = {.lex_state = 52, .external_lex_state = 4},
  [546] = {.lex_state = 52, .external_lex_state = 4},
  [547] = {.lex_state = 52, .external_lex_state = 4},
  [548] = {.lex_state = 52, .external_lex_state = 4},
  [549] = {.lex_state = 52, .external_lex_state = 4},
  [550] = {.lex_state = 52, .external_lex_state = 4},
  [551] = {.lex_state = 52, .external_lex_state = 4},
  [552] = {.lex_state = 52, .external_lex_state = 4},
  [553] = {.lex_state = 52, .external_lex_state = 4},
  [554] = {.lex_state = 52, .external_lex_state = 4},
  [555] = {.lex_state = 52, .external_lex_state = 4},
  [556] = {.lex_state = 52, .external_lex_state = 4},
  [557] = {.lex_state = 52, .external_lex_state = 4},
  [558] = {.lex_state = 52, .external_lex_state = 4},
  [559] = {.lex_state = 52, .external_lex_state = 4},
  [560] = {.lex_state = 52, .external_lex_state = 4},
  [561] = {.lex_state = 52, .external_lex_state = 8},
  [562] = {.lex_state = 52, .external_lex_state = 4},
  [563] = {.lex_state = 52, .external_lex_state = 4},
  [564] = {.lex_state = 52, .external_lex_state = 4},
  [565] = {.lex_state = 52, .external_lex_state = 4},
  [566] = {.lex_state = 52, .external_lex_state = 4},
  [567] = {.lex_state = 52, .external_lex_state = 4},
  [568] = {.lex_state = 52, .external_lex_state = 4},
  [569] = {.lex_state = 52, .external_lex_state = 4},
  [570] = {.lex_state = 52, .external_lex_state = 4},
  [571] = {.lex_state = 52, .external_lex_state = 4},
  [572] = {.lex_state = 52, .external_lex_state = 4},
  [573] = {.lex_state = 52, .external_lex_state = 4},
  [574] = {.lex_state = 52, .external_lex_state = 4},
  [575] = {.lex_state = 52, .external_lex_state = 4},
  [576] = {.lex_state = 52, .external_lex_state = 4},
  [577] = {.lex_state = 52, .external_lex_state = 4},
  [578] = {.lex_state = 52, .external_lex_state = 4},
  [579] = {.lex_state = 52, .external_lex_state = 4},
  [580] = {.lex_state = 52, .external_lex_state = 4},
  [581] = {.lex_state = 52, .external_lex_state = 4},
  [582] = {.lex_state = 52, .external_lex_state = 4},
  [583] = {.lex_state = 52, .external_lex_state = 4},
  [584] = {.lex_state = 52, .external_lex_state = 4},
  [585] = {.lex_state = 52, .external_lex_state = 4},
  [586] = {.lex_state = 52, .external_lex_state = 4},
  [587] = {.lex_state = 110, .external_lex_state = 3},
  [588] = {.lex_state = 52, .external_lex_state = 4},
  [589] = {.lex_state = 52, .external_lex_state = 4},
  [590] = {.lex_state = 52, .external_lex_state = 4},
  [591] = {.lex_state = 52, .external_lex_state = 4},
  [592] = {.lex_state = 55, .external_lex_state = 4},
  [593] = {.lex_state = 52, .external_lex_state = 4},
  [594] = {.lex_state = 52, .external_lex_state = 4},
  [595] = {.lex_state = 52, .external_lex_state = 4},
  [596] = {.lex_state = 52, .external_lex_state = 4},
  [597] = {.lex_state = 52, .external_lex_state = 4},
  [598] = {.lex_state = 52, .external_lex_state = 4},
  [599] = {.lex_state = 52, .external_lex_state = 4},
  [600] = {.lex_state = 52, .external_lex_state = 4},
  [601] = {.lex_state = 52, .external_lex_state = 4},
  [602] = {.lex_state = 55, .external_lex_state = 4},
  [603] = {.lex_state = 52, .external_lex_state = 4},
  [604] = {.lex_state = 55, .external_lex_state = 4},
  [605] = {.lex_state = 52, .external_lex_state = 4},
  [606] = {.lex_state = 52, .external_lex_state = 4},
  [607] = {.lex_state = 52, .external_lex_state = 4},
  [608] = {.lex_state = 52, .external_lex_state = 4},
  [609] = {.lex_state = 52, .external_lex_state = 4},
  [610] = {.lex_state = 52, .external_lex_state = 4},
  [611] = {.lex_state = 52, .external_lex_state = 4},
  [612] = {.lex_state = 55, .external_lex_state = 4},
  [613] = {.lex_state = 52, .external_lex_state = 4},
  [614] = {.lex_state = 52, .external_lex_state = 4},
  [615] = {.lex_state = 52, .external_lex_state = 4},
  [616] = {.lex_state = 52, .external_lex_state = 4},
  [617] = {.lex_state = 52, .external_lex_state = 4},
  [618] = {.lex_state = 52, .external_lex_state = 4},
  [619] = {.lex_state = 52, .external_lex_state = 4},
  [620] = {.lex_state = 52, .external_lex_state = 4},
  [621] = {.lex_state = 52, .external_lex_state = 4},
  [622] = {.lex_state = 52, .external_lex_state = 4},
  [623] = {.lex_state = 52, .external_lex_state = 4},
  [624] = {.lex_state = 52, .external_lex_state = 4},
  [625] = {.lex_state = 52, .external_lex_state = 4},
  [626] = {.lex_state = 55, .external_lex_state = 4},
  [627] = {.lex_state = 52, .external_lex_state = 4},
  [628] = {.lex_state = 52, .external_lex_state = 4},
  [629] = {.lex_state = 52, .external_lex_state = 4},
  [630] = {.lex_state = 52, .external_lex_state = 4},
  [631] = {.lex_state = 52, .external_lex_state = 4},
  [632] = {.lex_state = 52, .external_lex_state = 4},
  [633] = {.lex_state = 52, .external_lex_state = 4},
  [634] = {.lex_state = 52, .external_lex_state = 4},
  [635] = {.lex_state = 52, .external_lex_state = 4},
  [636] = {.lex_state = 55, .external_lex_state = 4},
  [637] = {.lex_state = 52, .external_lex_state = 4},
  [638] = {.lex_state = 52, .external_lex_state = 4},
  [639] = {.lex_state = 52, .external_lex_state = 4},
  [640] = {.lex_state = 52, .external_lex_state = 4},
  [641] = {.lex_state = 52, .external_lex_state = 4},
  [642] = {.lex_state = 52, .external_lex_state = 4},
  [643] = {.lex_state = 52, .external_lex_state = 4},
  [644] = {.lex_state = 52, .external_lex_state = 4},
  [645] = {.lex_state = 52, .external_lex_state = 4},
  [646] = {.lex_state = 52, .external_lex_state = 4},
  [647] = {.lex_state = 52, .external_lex_state = 4},
  [648] = {.lex_state = 52, .external_lex_state = 4},
  [649] = {.lex_state = 55, .external_lex_state = 4},
  [650] = {.lex_state = 52, .external_lex_state = 4},
  [651] = {.lex_state = 52, .external_lex_state = 4},
  [652] = {.lex_state = 52, .external_lex_state = 4},
  [653] = {.lex_state = 52, .external_lex_state = 4},
  [654] = {.lex_state = 52, .external_lex_state = 4},
  [655] = {.lex_state = 52, .external_lex_state = 4},
  [656] = {.lex_state = 52, .external_lex_state = 4},
  [657] = {.lex_state = 55, .external_lex_state = 4},
  [658] = {.lex_state = 52, .external_lex_state = 4},
  [659] = {.lex_state = 52, .external_lex_state = 4},
  [660] = {.lex_state = 52, .external_lex_state = 4},
  [661] = {.lex_state = 52, .external_lex_state = 4},
  [662] = {.lex_state = 52, .external_lex_state = 4},
  [663] = {.lex_state = 52, .external_lex_state = 4},
  [664] = {.lex_state = 52, .external_lex_state = 4},
  [665] = {.lex_state = 52, .external_lex_state = 4},
  [666] = {.lex_state = 52, .external_lex_state = 4},
  [667] = {.lex_state = 52, .external_lex_state = 4},
  [668] = {.lex_state = 52, .external_lex_state = 4},
  [669] = {.lex_state = 52, .external_lex_state = 4},
  [670] = {.lex_state = 52, .external_lex_state = 4},
  [671] = {.lex_state = 52, .external_lex_state = 4},
  [672] = {.lex_state = 52, .external_lex_state = 4},
  [673] = {.lex_state = 52, .external_lex_state = 4},
  [674] = {.lex_state = 52, .external_lex_state = 4},
  [675] = {.lex_state = 52, .external_lex_state = 4},
  [676] = {.lex_state = 52, .external_lex_state = 4},
  [677] = {.lex_state = 52, .external_lex_state = 4},
  [678] = {.lex_state = 52, .external_lex_state = 4},
  [679] = {.lex_state = 52, .external_lex_state = 4},
  [680] = {.lex_state = 52, .external_lex_state = 4},
  [681] = {.lex_state = 55, .external_lex_state = 4},
  [682] = {.lex_state = 52, .external_lex_state = 4},
  [683] = {.lex_state = 52, .external_lex_state = 4},
  [684] = {.lex_state = 52, .external_lex_state = 4},
  [685] = {.lex_state = 52, .external_lex_state = 4},
  [686] = {.lex_state = 52, .external_lex_state = 4},
  [687] = {.lex_state = 52, .external_lex_state = 4},
  [688] = {.lex_state = 52, .external_lex_state = 4},
  [689] = {.lex_state = 52, .external_lex_state = 4},
  [690] = {.lex_state = 52, .external_lex_state = 4},
  [691] = {.lex_state = 55, .external_lex_state = 4},
  [692] = {.lex_state = 52, .external_lex_state = 4},
  [693] = {.lex_state = 52, .external_lex_state = 4},
  [694] = {.lex_state = 52, .external_lex_state = 4},
  [695] = {.lex_state = 52, .external_lex_state = 4},
  [696] = {.lex_state = 52, .external_lex_state = 4},
  [697] = {.lex_state = 52, .external_lex_state = 4},
  [698] = {.lex_state = 52, .external_lex_state = 4},
  [699] = {.lex_state = 52, .external_lex_state = 4},
  [700] = {.lex_state = 55, .external_lex_state = 4},
  [701] = {.lex_state = 55, .external_lex_state = 4},
  [702] = {.lex_state = 52, .external_lex_state = 4},
  [703] = {.lex_state = 52, .external_lex_state = 4},
  [704] = {.lex_state = 52, .external_lex_state = 4},
  [705] = {.lex_state = 52, .external_lex_state = 4},
  [706] = {.lex_state = 52, .external_lex_state = 4},
  [707] = {.lex_state = 52, .external_lex_state = 4},
  [708] = {.lex_state = 55, .external_lex_state = 4},
  [709] = {.lex_state = 52, .external_lex_state = 4},
  [710] = {.lex_state = 52, .external_lex_state = 4},
  [711] = {.lex_state = 52, .external_lex_state = 4},
  [712] = {.lex_state = 55, .external_lex_state = 4},
  [713] = {.lex_state = 52, .external_lex_state = 10},
  [714] = {.lex_state = 52, .external_lex_state = 4},
  [715] = {.lex_state = 52, .external_lex_state = 4},
  [716] = {.lex_state = 52, .external_lex_state = 4},
  [717] = {.lex_state = 52, .external_lex_state = 4},
  [718] = {.lex_state = 52, .external_lex_state = 4},
  [719] = {.lex_state = 52, .external_lex_state = 10},
  [720] = {.lex_state = 52, .external_lex_state = 10},
  [721] = {.lex_state = 52, .external_lex_state = 10},
  [722] = {.lex_state = 52, .external_lex_state = 4},
  [723] = {.lex_state = 52, .external_lex_state = 4},
  [724] = {.lex_state = 52, .external_lex_state = 10},
  [725] = {.lex_state = 52, .external_lex_state = 10},
  [726] = {.lex_state = 52, .external_lex_state = 10},
  [727] = {.lex_state = 52, .external_lex_state = 10},
  [728] = {.lex_state = 52, .external_lex_state = 10},
  [729] = {.lex_state = 52, .external_lex_state = 10},
  [730] = {.lex_state = 52, .external_lex_state = 4},
  [731] = {.lex_state = 52, .external_lex_state = 4},
  [732] = {.lex_state = 52, .external_lex_state = 10},
  [733] = {.lex_state = 52, .external_lex_state = 4},
  [734] = {.lex_state = 52, .external_lex_state = 4},
  [735] = {.lex_state = 52, .external_lex_state = 10},
  [736] = {.lex_state = 52, .external_lex_state = 10},
  [737] = {.lex_state = 52, .external_lex_state = 10},
  [738] = {.lex_state = 52, .external_lex_state = 10},
  [739] = {.lex_state = 52, .external_lex_state = 10},
  [740] = {.lex_state = 52, .external_lex_state = 10},
  [741] = {.lex_state = 52, .external_lex_state = 10},
  [742] = {.lex_state = 52, .external_lex_state = 10},
  [743] = {.lex_state = 52, .external_lex_state = 4},
  [744] = {.lex_state = 52, .external_lex_state = 4},
  [745] = {.lex_state = 52, .external_lex_state = 4},
  [746] = {.lex_state = 52, .external_lex_state = 10},
  [747] = {.lex_state = 52, .external_lex_state = 4},
  [748] = {.lex_state = 52, .external_lex_state = 10},
  [749] = {.lex_state = 52, .external_lex_state = 10},
  [750] = {.lex_state = 52, .external_lex_state = 10},
  [751] = {.lex_state = 52, .external_lex_state = 4},
  [752] = {.lex_state = 52, .external_lex_state = 4},
  [753] = {.lex_state = 52, .external_lex_state = 4},
  [754] = {.lex_state = 52, .external_lex_state = 4},
  [755] = {.lex_state = 52, .external_lex_state = 4},
  [756] = {.lex_state = 52, .external_lex_state = 4},
  [757] = {.lex_state = 52, .external_lex_state = 4},
  [758] = {.lex_state = 52, .external_lex_state = 4},
  [759] = {.lex_state = 52, .external_lex_state = 4},
  [760] = {.lex_state = 52, .external_lex_state = 4},
  [761] = {.lex_state = 52, .external_lex_state = 10},
  [762] = {.lex_state = 52, .external_lex_state = 10},
  [763] = {.lex_state = 52, .external_lex_state = 10},
  [764] = {.lex_state = 52, .external_lex_state = 4},
  [765] = {.lex_state = 52, .external_lex_state = 10},
  [766] = {.lex_state = 52, .external_lex_state = 10},
  [767] = {.lex_state = 110, .external_lex_state = 3},
  [768] = {.lex_state = 52, .external_lex_state = 4},
  [769] = {.lex_state = 52, .external_lex_state = 4},
  [770] = {.lex_state = 52, .external_lex_state = 10},
  [771] = {.lex_state = 52, .external_lex_state = 10},
  [772] = {.lex_state = 52, .external_lex_state = 10},
  [773] = {.lex_state = 52, .external_lex_state = 4},
  [774] = {.lex_state = 52, .external_lex_state = 10},
  [775] = {.lex_state = 52, .external_lex_state = 10},
  [776] = {.lex_state = 52, .external_lex_state = 10},
  [777] = {.lex_state = 52, .external_lex_state = 4},
  [778] = {.lex_state = 52, .external_lex_state = 10},
  [779] = {.lex_state = 52, .external_lex_state = 10},
  [780] = {.lex_state = 52, .external_lex_state = 10},
  [781] = {.lex_state = 52, .external_lex_state = 4},
  [782] = {.lex_state = 52, .external_lex_state = 4},
  [783] = {.lex_state = 52, .external_lex_state = 4},
  [784] = {.lex_state = 52, .external_lex_state = 4},
  [785] = {.lex_state = 52, .external_lex_state = 4},
  [786] = {.lex_state = 52, .external_lex_state = 4},
  [787] = {.lex_state = 52, .external_lex_state = 4},
  [788] = {.lex_state = 52, .external_lex_state = 4},
  [789] = {.lex_state = 52, .external_lex_state = 4},
  [790] = {.lex_state = 52, .external_lex_state = 10},
  [791] = {.lex_state = 52, .external_lex_state = 10},
  [792] = {.lex_state = 52, .external_lex_state = 10},
  [793] = {.lex_state = 52, .external_lex_state = 4},
  [794] = {.lex_state = 52, .external_lex_state = 4},
  [795] = {.lex_state = 52, .external_lex_state = 4},
  [796] = {.lex_state = 52, .external_lex_state = 4},
  [797] = {.lex_state = 52, .external_lex_state = 4},
  [798] = {.lex_state = 52, .external_lex_state = 4},
  [799] = {.lex_state = 52, .external_lex_state = 4},
  [800] = {.lex_state = 52, .external_lex_state = 4},
  [801] = {.lex_state = 52, .external_lex_state = 4},
  [802] = {.lex_state = 52, .external_lex_state = 4},
  [803] = {.lex_state = 52, .external_lex_state = 4},
  [804] = {.lex_state = 52, .external_lex_state = 10},
  [805] = {.lex_state = 52, .external_lex_state = 10},
  [806] = {.lex_state = 52, .external_lex_state = 4},
  [807] = {.lex_state = 52, .external_lex_state = 4},
  [808] = {.lex_state = 52, .external_lex_state = 4},
  [809] = {.lex_state = 52, .external_lex_state = 4},
  [810] = {.lex_state = 52, .external_lex_state = 4},
  [811] = {.lex_state = 52, .external_lex_state = 4},
  [812] = {.lex_state = 52, .external_lex_state = 4},
  [813] = {.lex_state = 52, .external_lex_state = 4},
  [814] = {.lex_state = 52, .external_lex_state = 4},
  [815] = {.lex_state = 52, .external_lex_state = 4},
  [816] = {.lex_state = 52, .external_lex_state = 4},
  [817] = {.lex_state = 52, .external_lex_state = 4},
  [818] = {.lex_state = 52, .external_lex_state = 4},
  [819] = {.lex_state = 52, .external_lex_state = 4},
  [820] = {.lex_state = 52, .external_lex_state = 4},
  [821] = {.lex_state = 52, .external_lex_state = 4},
  [822] = {.lex_state = 52, .external_lex_state = 4},
  [823] = {.lex_state = 52, .external_lex_state = 4},
  [824] = {.lex_state = 52, .external_lex_state = 4},
  [825] = {.lex_state = 52, .external_lex_state = 4},
  [826] = {.lex_state = 52, .external_lex_state = 4},
  [827] = {.lex_state = 52, .external_lex_state = 4},
  [828] = {.lex_state = 52, .external_lex_state = 4},
  [829] = {.lex_state = 52, .external_lex_state = 4},
  [830] = {.lex_state = 52, .external_lex_state = 4},
  [831] = {.lex_state = 52, .external_lex_state = 4},
  [832] = {.lex_state = 52, .external_lex_state = 4},
  [833] = {.lex_state = 52, .external_lex_state = 4},
  [834] = {.lex_state = 52, .external_lex_state = 4},
  [835] = {.lex_state = 52, .external_lex_state = 4},
  [836] = {.lex_state = 52, .external_lex_state = 4},
  [837] = {.lex_state = 52, .external_lex_state = 4},
  [838] = {.lex_state = 52, .external_lex_state = 4},
  [839] = {.lex_state = 52, .external_lex_state = 4},
  [840] = {.lex_state = 52, .external_lex_state = 4},
  [841] = {.lex_state = 52, .external_lex_state = 4},
  [842] = {.lex_state = 52, .external_lex_state = 4},
  [843] = {.lex_state = 52, .external_lex_state = 4},
  [844] = {.lex_state = 52, .external_lex_state = 4},
  [845] = {.lex_state = 52, .external_lex_state = 4},
  [846] = {.lex_state = 52, .external_lex_state = 4},
  [847] = {.lex_state = 52, .external_lex_state = 4},
  [848] = {.lex_state = 52, .external_lex_state = 4},
  [849] = {.lex_state = 52, .external_lex_state = 4},
  [850] = {.lex_state = 52, .external_lex_state = 4},
  [851] = {.lex_state = 52, .external_lex_state = 4},
  [852] = {.lex_state = 52, .external_lex_state = 4},
  [853] = {.lex_state = 52, .external_lex_state = 4},
  [854] = {.lex_state = 52, .external_lex_state = 4},
  [855] = {.lex_state = 52, .external_lex_state = 4},
  [856] = {.lex_state = 52, .external_lex_state = 4},
  [857] = {.lex_state = 52, .external_lex_state = 4},
  [858] = {.lex_state = 52, .external_lex_state = 4},
  [859] = {.lex_state = 52, .external_lex_state = 4},
  [860] = {.lex_state = 52, .external_lex_state = 4},
  [861] = {.lex_state = 52, .external_lex_state = 4},
  [862] = {.lex_state = 52, .external_lex_state = 4},
  [863] = {.lex_state = 52, .external_lex_state = 4},
  [864] = {.lex_state = 52, .external_lex_state = 4},
  [865] = {.lex_state = 52, .external_lex_state = 4},
  [866] = {.lex_state = 52, .external_lex_state = 4},
  [867] = {.lex_state = 52, .external_lex_state = 4},
  [868] = {.lex_state = 52, .external_lex_state = 4},
  [869] = {.lex_state = 52, .external_lex_state = 4},
  [870] = {.lex_state = 52, .external_lex_state = 4},
  [871] = {.lex_state = 52, .external_lex_state = 4},
  [872] = {.lex_state = 52, .external_lex_state = 4},
  [873] = {.lex_state = 52, .external_lex_state = 4},
  [874] = {.lex_state = 52, .external_lex_state = 4},
  [875] = {.lex_state = 52, .external_lex_state = 4},
  [876] = {.lex_state = 52, .external_lex_state = 4},
  [877] = {.lex_state = 52, .external_lex_state = 4},
  [878] = {.lex_state = 52, .external_lex_state = 4},
  [879] = {.lex_state = 52, .external_lex_state = 4},
  [880] = {.lex_state = 52, .external_lex_state = 4},
  [881] = {.lex_state = 52, .external_lex_state = 4},
  [882] = {.lex_state = 52, .external_lex_state = 4},
  [883] = {.lex_state = 52, .external_lex_state = 4},
  [884] = {.lex_state = 52, .external_lex_state = 4},
  [885] = {.lex_state = 52, .external_lex_state = 4},
  [886] = {.lex_state = 52, .external_lex_state = 4},
  [887] = {.lex_state = 52, .external_lex_state = 4},
  [888] = {.lex_state = 52, .external_lex_state = 4},
  [889] = {.lex_state = 52, .external_lex_state = 4},
  [890] = {.lex_state = 52, .external_lex_state = 4},
  [891] = {.lex_state = 52, .external_lex_state = 4},
  [892] = {.lex_state = 52, .external_lex_state = 4},
  [893] = {.lex_state = 52, .external_lex_state = 4},
  [894] = {.lex_state = 52, .external_lex_state = 4},
  [895] = {.lex_state = 52, .external_lex_state = 4},
  [896] = {.lex_state = 52, .external_lex_state = 4},
  [897] = {.lex_state = 52, .external_lex_state = 4},
  [898] = {.lex_state = 52, .external_lex_state = 4},
  [899] = {.lex_state = 52, .external_lex_state = 4},
  [900] = {.lex_state = 52, .external_lex_state = 4},
  [901] = {.lex_state = 52, .external_lex_state = 4},
  [902] = {.lex_state = 52, .external_lex_state = 4},
  [903] = {.lex_state = 52, .external_lex_state = 4},
  [904] = {.lex_state = 52, .external_lex_state = 4},
  [905] = {.lex_state = 52, .external_lex_state = 4},
  [906] = {.lex_state = 52, .external_lex_state = 4},
  [907] = {.lex_state = 52, .external_lex_state = 4},
  [908] = {.lex_state = 52, .external_lex_state = 4},
  [909] = {.lex_state = 52, .external_lex_state = 4},
  [910] = {.lex_state = 52, .external_lex_state = 4},
  [911] = {.lex_state = 52, .external_lex_state = 4},
  [912] = {.lex_state = 52, .external_lex_state = 4},
  [913] = {.lex_state = 52, .external_lex_state = 4},
  [914] = {.lex_state = 52, .external_lex_state = 4},
  [915] = {.lex_state = 52, .external_lex_state = 4},
  [916] = {.lex_state = 52, .external_lex_state = 4},
  [917] = {.lex_state = 52, .external_lex_state = 4},
  [918] = {.lex_state = 52, .external_lex_state = 4},
  [919] = {.lex_state = 52, .external_lex_state = 4},
  [920] = {.lex_state = 52, .external_lex_state = 4},
  [921] = {.lex_state = 52, .external_lex_state = 4},
  [922] = {.lex_state = 52, .external_lex_state = 4},
  [923] = {.lex_state = 52, .external_lex_state = 4},
  [924] = {.lex_state = 52, .external_lex_state = 4},
  [925] = {.lex_state = 52, .external_lex_state = 4},
  [926] = {.lex_state = 52, .external_lex_state = 4},
  [927] = {.lex_state = 52, .external_lex_state = 4},
  [928] = {.lex_state = 52, .external_lex_state = 4},
  [929] = {.lex_state = 52, .external_lex_state = 4},
  [930] = {.lex_state = 52, .external_lex_state = 4},
  [931] = {.lex_state = 52, .external_lex_state = 4},
  [932] = {.lex_state = 52, .external_lex_state = 4},
  [933] = {.lex_state = 52, .external_lex_state = 4},
  [934] = {.lex_state = 52, .external_lex_state = 4},
  [935] = {.lex_state = 52, .external_lex_state = 4},
  [936] = {.lex_state = 52, .external_lex_state = 4},
  [937] = {.lex_state = 52, .external_lex_state = 4},
  [938] = {.lex_state = 52, .external_lex_state = 4},
  [939] = {.lex_state = 52, .external_lex_state = 4},
  [940] = {.lex_state = 52, .external_lex_state = 4},
  [941] = {.lex_state = 52, .external_lex_state = 4},
  [942] = {.lex_state = 52, .external_lex_state = 4},
  [943] = {.lex_state = 52, .external_lex_state = 4},
  [944] = {.lex_state = 52, .external_lex_state = 4},
  [945] = {.lex_state = 52, .external_lex_state = 4},
  [946] = {.lex_state = 52, .external_lex_state = 4},
  [947] = {.lex_state = 52, .external_lex_state = 4},
  [948] = {.lex_state = 52, .external_lex_state = 4},
  [949] = {.lex_state = 52, .external_lex_state = 4},
  [950] = {.lex_state = 52, .external_lex_state = 4},
  [951] = {.lex_state = 52, .external_lex_state = 4},
  [952] = {.lex_state = 52, .external_lex_state = 4},
  [953] = {.lex_state = 52, .external_lex_state = 4},
  [954] = {.lex_state = 52, .external_lex_state = 4},
  [955] = {.lex_state = 52, .external_lex_state = 4},
  [956] = {.lex_state = 52, .external_lex_state = 4},
  [957] = {.lex_state = 52, .external_lex_state = 4},
  [958] = {.lex_state = 52, .external_lex_state = 4},
  [959] = {.lex_state = 52, .external_lex_state = 4},
  [960] = {.lex_state = 52, .external_lex_state = 4},
  [961] = {.lex_state = 52, .external_lex_state = 4},
  [962] = {.lex_state = 52, .external_lex_state = 4},
  [963] = {.lex_state = 52, .external_lex_state = 4},
  [964] = {.lex_state = 52, .external_lex_state = 4},
  [965] = {.lex_state = 52, .external_lex_state = 4},
  [966] = {.lex_state = 52, .external_lex_state = 4},
  [967] = {.lex_state = 52, .external_lex_state = 4},
  [968] = {.lex_state = 52, .external_lex_state = 4},
  [969] = {.lex_state = 52, .external_lex_state = 4},
  [970] = {.lex_state = 52, .external_lex_state = 4},
  [971] = {.lex_state = 52, .external_lex_state = 4},
  [972] = {.lex_state = 52, .external_lex_state = 4},
  [973] = {.lex_state = 52, .external_lex_state = 4},
  [974] = {.lex_state = 52, .external_lex_state = 4},
  [975] = {.lex_state = 52, .external_lex_state = 4},
  [976] = {.lex_state = 52, .external_lex_state = 4},
  [977] = {.lex_state = 52, .external_lex_state = 4},
  [978] = {.lex_state = 52, .external_lex_state = 4},
  [979] = {.lex_state = 52, .external_lex_state = 4},
  [980] = {.lex_state = 52, .external_lex_state = 4},
  [981] = {.lex_state = 52, .external_lex_state = 4},
  [982] = {.lex_state = 52, .external_lex_state = 4},
  [983] = {.lex_state = 52, .external_lex_state = 4},
  [984] = {.lex_state = 52, .external_lex_state = 4},
  [985] = {.lex_state = 52, .external_lex_state = 4},
  [986] = {.lex_state = 52, .external_lex_state = 4},
  [987] = {.lex_state = 52, .external_lex_state = 4},
  [988] = {.lex_state = 52, .external_lex_state = 4},
  [989] = {.lex_state = 52, .external_lex_state = 4},
  [990] = {.lex_state = 52, .external_lex_state = 4},
  [991] = {.lex_state = 52, .external_lex_state = 4},
  [992] = {.lex_state = 52, .external_lex_state = 4},
  [993] = {.lex_state = 52, .external_lex_state = 4},
  [994] = {.lex_state = 52, .external_lex_state = 4},
  [995] = {.lex_state = 52, .external_lex_state = 4},
  [996] = {.lex_state = 52, .external_lex_state = 4},
  [997] = {.lex_state = 52, .external_lex_state = 4},
  [998] = {.lex_state = 52, .external_lex_state = 4},
  [999] = {.lex_state = 52, .external_lex_state = 4},
  [1000] = {.lex_state = 52, .external_lex_state = 4},
  [1001] = {.lex_state = 52, .external_lex_state = 4},
  [1002] = {.lex_state = 52, .external_lex_state = 4},
  [1003] = {.lex_state = 52, .external_lex_state = 4},
  [1004] = {.lex_state = 52, .external_lex_state = 4},
  [1005] = {.lex_state = 52, .external_lex_state = 4},
  [1006] = {.lex_state = 52, .external_lex_state = 4},
  [1007] = {.lex_state = 52, .external_lex_state = 4},
  [1008] = {.lex_state = 52, .external_lex_state = 4},
  [1009] = {.lex_state = 52, .external_lex_state = 4},
  [1010] = {.lex_state = 52, .external_lex_state = 4},
  [1011] = {.lex_state = 52, .external_lex_state = 4},
  [1012] = {.lex_state = 52, .external_lex_state = 4},
  [1013] = {.lex_state = 52, .external_lex_state = 4},
  [1014] = {.lex_state = 52, .external_lex_state = 4},
  [1015] = {.lex_state = 52, .external_lex_state = 4},
  [1016] = {.lex_state = 52, .external_lex_state = 4},
  [1017] = {.lex_state = 52, .external_lex_state = 4},
  [1018] = {.lex_state = 52, .external_lex_state = 4},
  [1019] = {.lex_state = 52, .external_lex_state = 4},
  [1020] = {.lex_state = 52, .external_lex_state = 4},
  [1021] = {.lex_state = 52, .external_lex_state = 4},
  [1022] = {.lex_state = 52, .external_lex_state = 4},
  [1023] = {.lex_state = 52, .external_lex_state = 4},
  [1024] = {.lex_state = 52, .external_lex_state = 4},
  [1025] = {.lex_state = 52, .external_lex_state = 4},
  [1026] = {.lex_state = 52, .external_lex_state = 4},
  [1027] = {.lex_state = 52, .external_lex_state = 4},
  [1028] = {.lex_state = 52, .external_lex_state = 4},
  [1029] = {.lex_state = 52, .external_lex_state = 4},
  [1030] = {.lex_state = 52, .external_lex_state = 4},
  [1031] = {.lex_state = 52, .external_lex_state = 4},
  [1032] = {.lex_state = 52, .external_lex_state = 4},
  [1033] = {.lex_state = 52, .external_lex_state = 4},
  [1034] = {.lex_state = 52, .external_lex_state = 4},
  [1035] = {.lex_state = 52, .external_lex_state = 4},
  [1036] = {.lex_state = 52, .external_lex_state = 4},
  [1037] = {.lex_state = 52, .external_lex_state = 4},
  [1038] = {.lex_state = 52, .external_lex_state = 4},
  [1039] = {.lex_state = 52, .external_lex_state = 4},
  [1040] = {.lex_state = 52, .external_lex_state = 4},
  [1041] = {.lex_state = 52, .external_lex_state = 4},
  [1042] = {.lex_state = 52, .external_lex_state = 4},
  [1043] = {.lex_state = 52, .external_lex_state = 4},
  [1044] = {.lex_state = 52, .external_lex_state = 4},
  [1045] = {.lex_state = 52, .external_lex_state = 4},
  [1046] = {.lex_state = 52, .external_lex_state = 4},
  [1047] = {.lex_state = 52, .external_lex_state = 4},
  [1048] = {.lex_state = 52, .external_lex_state = 4},
  [1049] = {.lex_state = 52, .external_lex_state = 4},
  [1050] = {.lex_state = 52, .external_lex_state = 4},
  [1051] = {.lex_state = 52, .external_lex_state = 4},
  [1052] = {.lex_state = 52, .external_lex_state = 4},
  [1053] = {.lex_state = 52, .external_lex_state = 4},
  [1054] = {.lex_state = 52, .external_lex_state = 4},
  [1055] = {.lex_state = 52, .external_lex_state = 4},
  [1056] = {.lex_state = 52, .external_lex_state = 4},
  [1057] = {.lex_state = 52, .external_lex_state = 4},
  [1058] = {.lex_state = 52, .external_lex_state = 4},
  [1059] = {.lex_state = 52, .external_lex_state = 4},
  [1060] = {.lex_state = 52, .external_lex_state = 4},
  [1061] = {.lex_state = 52, .external_lex_state = 4},
  [1062] = {.lex_state = 52, .external_lex_state = 4},
  [1063] = {.lex_state = 52, .external_lex_state = 4},
  [1064] = {.lex_state = 52, .external_lex_state = 4},
  [1065] = {.lex_state = 52, .external_lex_state = 4},
  [1066] = {.lex_state = 52, .external_lex_state = 4},
  [1067] = {.lex_state = 52, .external_lex_state = 4},
  [1068] = {.lex_state = 52, .external_lex_state = 4},
  [1069] = {.lex_state = 52, .external_lex_state = 4},
  [1070] = {.lex_state = 52, .external_lex_state = 4},
  [1071] = {.lex_state = 52, .external_lex_state = 4},
  [1072] = {.lex_state = 52, .external_lex_state = 4},
  [1073] = {.lex_state = 52, .external_lex_state = 4},
  [1074] = {.lex_state = 52, .external_lex_state = 4},
  [1075] = {.lex_state = 52, .external_lex_state = 4},
  [1076] = {.lex_state = 52, .external_lex_state = 4},
  [1077] = {.lex_state = 52, .external_lex_state = 4},
  [1078] = {.lex_state = 52, .external_lex_state = 4},
  [1079] = {.lex_state = 52, .external_lex_state = 4},
  [1080] = {.lex_state = 52, .external_lex_state = 4},
  [1081] = {.lex_state = 52, .external_lex_state = 4},
  [1082] = {.lex_state = 52, .external_lex_state = 4},
  [1083] = {.lex_state = 52, .external_lex_state = 4},
  [1084] = {.lex_state = 52, .external_lex_state = 4},
  [1085] = {.lex_state = 52, .external_lex_state = 4},
  [1086] = {.lex_state = 52, .external_lex_state = 4},
  [1087] = {.lex_state = 52, .external_lex_state = 4},
  [1088] = {.lex_state = 52, .external_lex_state = 4},
  [1089] = {.lex_state = 52, .external_lex_state = 4},
  [1090] = {.lex_state = 52, .external_lex_state = 4},
  [1091] = {.lex_state = 52, .external_lex_state = 4},
  [1092] = {.lex_state = 52, .external_lex_state = 4},
  [1093] = {.lex_state = 52, .external_lex_state = 4},
  [1094] = {.lex_state = 52, .external_lex_state = 4},
  [1095] = {.lex_state = 52, .external_lex_state = 4},
  [1096] = {.lex_state = 52, .external_lex_state = 4},
  [1097] = {.lex_state = 52, .external_lex_state = 4},
  [1098] = {.lex_state = 52, .external_lex_state = 4},
  [1099] = {.lex_state = 52, .external_lex_state = 4},
  [1100] = {.lex_state = 52, .external_lex_state = 4},
  [1101] = {.lex_state = 52, .external_lex_state = 4},
  [1102] = {.lex_state = 52, .external_lex_state = 4},
  [1103] = {.lex_state = 52, .external_lex_state = 4},
  [1104] = {.lex_state = 52, .external_lex_state = 4},
  [1105] = {.lex_state = 52, .external_lex_state = 4},
  [1106] = {.lex_state = 52, .external_lex_state = 4},
  [1107] = {.lex_state = 52, .external_lex_state = 4},
  [1108] = {.lex_state = 52, .external_lex_state = 4},
  [1109] = {.lex_state = 52, .external_lex_state = 4},
  [1110] = {.lex_state = 52, .external_lex_state = 4},
  [1111] = {.lex_state = 52, .external_lex_state = 4},
  [1112] = {.lex_state = 52, .external_lex_state = 4},
  [1113] = {.lex_state = 52, .external_lex_state = 4},
  [1114] = {.lex_state = 52, .external_lex_state = 4},
  [1115] = {.lex_state = 52, .external_lex_state = 4},
  [1116] = {.lex_state = 52, .external_lex_state = 4},
  [1117] = {.lex_state = 52, .external_lex_state = 4},
  [1118] = {.lex_state = 52, .external_lex_state = 4},
  [1119] = {.lex_state = 52, .external_lex_state = 4},
  [1120] = {.lex_state = 52, .external_lex_state = 4},
  [1121] = {.lex_state = 52, .external_lex_state = 4},
  [1122] = {.lex_state = 52, .external_lex_state = 4},
  [1123] = {.lex_state = 52, .external_lex_state = 4},
  [1124] = {.lex_state = 52, .external_lex_state = 4},
  [1125] = {.lex_state = 52, .external_lex_state = 4},
  [1126] = {.lex_state = 52, .external_lex_state = 4},
  [1127] = {.lex_state = 52, .external_lex_state = 4},
  [1128] = {.lex_state = 52, .external_lex_state = 4},
  [1129] = {.lex_state = 52, .external_lex_state = 4},
  [1130] = {.lex_state = 52, .external_lex_state = 4},
  [1131] = {.lex_state = 52, .external_lex_state = 4},
  [1132] = {.lex_state = 1, .external_lex_state = 11},
  [1133] = {.lex_state = 1, .external_lex_state = 11},
  [1134] = {.lex_state = 1, .external_lex_state = 11},
  [1135] = {.lex_state = 1, .external_lex_state = 11},
  [1136] = {.lex_state = 1, .external_lex_state = 11},
  [1137] = {.lex_state = 1, .external_lex_state = 11},
  [1138] = {.lex_state = 1, .external_lex_state = 11},
  [1139] = {.lex_state = 1, .external_lex_state = 11},
  [1140] = {.lex_state = 1, .external_lex_state = 11},
  [1141] = {.lex_state = 1, .external_lex_state = 11},
  [1142] = {.lex_state = 1, .external_lex_state = 3},
  [1143] = {.lex_state = 1, .external_lex_state = 12},
  [1144] = {.lex_state = 1, .external_lex_state = 3},
  [1145] = {.lex_state = 1, .external_lex_state = 3},
  [1146] = {.lex_state = 1, .external_lex_state = 3},
  [1147] = {.lex_state = 1, .external_lex_state = 3},
  [1148] = {.lex_state = 1, .external_lex_state = 3},
  [1149] = {.lex_state = 1, .external_lex_state = 3},
  [1150] = {.lex_state = 1, .external_lex_state = 3},
  [1151] = {.lex_state = 1, .external_lex_state = 3},
  [1152] = {.lex_state = 1, .external_lex_state = 3},
  [1153] = {.lex_state = 1, .external_lex_state = 3},
  [1154] = {.lex_state = 1, .external_lex_state = 3},
  [1155] = {.lex_state = 1, .external_lex_state = 3},
  [1156] = {.lex_state = 1, .external_lex_state = 12},
  [1157] = {.lex_state = 1, .external_lex_state = 12},
  [1158] = {.lex_state = 1, .external_lex_state = 12},
  [1159] = {.lex_state = 1, .external_lex_state = 12},
  [1160] = {.lex_state = 1, .external_lex_state = 12},
  [1161] = {.lex_state = 1, .external_lex_state = 3},
  [1162] = {.lex_state = 1, .external_lex_state = 3},
  [1163] = {.lex_state = 1, .external_lex_state = 12},
  [1164] = {.lex_state = 1, .external_lex_state = 3},
  [1165] = {.lex_state = 1, .external_lex_state = 12},
  [1166] = {.lex_state = 1, .external_lex_state = 12},
  [1167] = {.lex_state = 1, .external_lex_state = 12},
  [1168] = {.lex_state = 1, .external_lex_state = 12},
  [1169] = {.lex_state = 1, .external_lex_state = 12},
  [1170] = {.lex_state = 1, .external_lex_state = 12},
  [1171] = {.lex_state = 1, .external_lex_state = 12},
  [1172] = {.lex_state = 1, .external_lex_state = 12},
  [1173] = {.lex_state = 1, .external_lex_state = 12},
  [1174] = {.lex_state = 1, .external_lex_state = 12},
  [1175] = {.lex_state = 1, .external_lex_state = 12},
  [1176] = {.lex_state = 1, .external_lex_state = 12},
  [1177] = {.lex_state = 1, .external_lex_state = 12},
  [1178] = {.lex_state = 1, .external_lex_state = 12},
  [1179] = {.lex_state = 1, .external_lex_state = 12},
  [1180] = {.lex_state = 1, .external_lex_state = 12},
  [1181] = {.lex_state = 1, .external_lex_state = 12},
  [1182] = {.lex_state = 1, .external_lex_state = 12},
  [1183] = {.lex_state = 1, .external_lex_state = 12},
  [1184] = {.lex_state = 1, .external_lex_state = 12},
  [1185] = {.lex_state = 1, .external_lex_state = 12},
  [1186] = {.lex_state = 1, .external_lex_state = 12},
  [1187] = {.lex_state = 1, .external_lex_state = 12},
  [1188] = {.lex_state = 1, .external_lex_state = 12},
  [1189] = {.lex_state = 1, .external_lex_state = 12},
  [1190] = {.lex_state = 1, .external_lex_state = 12},
  [1191] = {.lex_state = 1, .external_lex_state = 12},
  [1192] = {.lex_state = 1, .external_lex_state = 12},
  [1193] = {.lex_state = 1, .external_lex_state = 12},
  [1194] = {.lex_state = 1, .external_lex_state = 12},
  [1195] = {.lex_state = 1, .external_lex_state = 3},
  [1196] = {.lex_state = 1, .external_lex_state = 12},
  [1197] = {.lex_state = 1, .external_lex_state = 12},
  [1198] = {.lex_state = 1, .external_lex_state = 12},
  [1199] = {.lex_state = 1, .external_lex_state = 12},
  [1200] = {.lex_state = 1, .external_lex_state = 12},
  [1201] = {.lex_state = 1, .external_lex_state = 12},
  [1202] = {.lex_state = 1, .external_lex_state = 12},
  [1203] = {.lex_state = 1, .external_lex_state = 12},
  [1204] = {.lex_state = 1, .external_lex_state = 12},
  [1205] = {.lex_state = 1, .external_lex_state = 12},
  [1206] = {.lex_state = 1, .external_lex_state = 12},
  [1207] = {.lex_state = 1, .external_lex_state = 12},
  [1208] = {.lex_state = 1, .external_lex_state = 12},
  [1209] = {.lex_state = 1, .external_lex_state = 12},
  [1210] = {.lex_state = 1, .external_lex_state = 12},
  [1211] = {.lex_state = 1, .external_lex_state = 12},
  [1212] = {.lex_state = 1, .external_lex_state = 12},
  [1213] = {.lex_state = 1, .external_lex_state = 12},
  [1214] = {.lex_state = 1, .external_lex_state = 12},
  [1215] = {.lex_state = 1, .external_lex_state = 12},
  [1216] = {.lex_state = 1, .external_lex_state = 12},
  [1217] = {.lex_state = 1, .external_lex_state = 12},
  [1218] = {.lex_state = 1, .external_lex_state = 12},
  [1219] = {.lex_state = 1, .external_lex_state = 12},
  [1220] = {.lex_state = 1, .external_lex_state = 12},
  [1221] = {.lex_state = 1, .external_lex_state = 12},
  [1222] = {.lex_state = 1, .external_lex_state = 12},
  [1223] = {.lex_state = 1, .external_lex_state = 12},
  [1224] = {.lex_state = 1, .external_lex_state = 12},
  [1225] = {.lex_state = 1, .external_lex_state = 12},
  [1226] = {.lex_state = 1, .external_lex_state = 12},
  [1227] = {.lex_state = 1, .external_lex_state = 12},
  [1228] = {.lex_state = 1, .external_lex_state = 12},
  [1229] = {.lex_state = 1, .external_lex_state = 3},
  [1230] = {.lex_state = 1, .external_lex_state = 12},
  [1231] = {.lex_state = 1, .external_lex_state = 12},
  [1232] = {.lex_state = 1, .external_lex_state = 12},
  [1233] = {.lex_state = 1, .external_lex_state = 12},
  [1234] = {.lex_state = 1, .external_lex_state = 11},
  [1235] = {.lex_state = 1, .external_lex_state = 11},
  [1236] = {.lex_state = 1, .external_lex_state = 11},
  [1237] = {.lex_state = 1, .external_lex_state = 11},
  [1238] = {.lex_state = 1, .external_lex_state = 11},
  [1239] = {.lex_state = 1, .external_lex_state = 11},
  [1240] = {.lex_state = 1, .external_lex_state = 11},
  [1241] = {.lex_state = 1, .external_lex_state = 11},
  [1242] = {.lex_state = 1, .external_lex_state = 11},
  [1243] = {.lex_state = 1, .external_lex_state = 11},
  [1244] = {.lex_state = 1, .external_lex_state = 3},
  [1245] = {.lex_state = 1, .external_lex_state = 3},
  [1246] = {.lex_state = 1, .external_lex_state = 3},
  [1247] = {.lex_state = 1, .external_lex_state = 3},
  [1248] = {.lex_state = 1, .external_lex_state = 3},
  [1249] = {.lex_state = 1, .external_lex_state = 3},
  [1250] = {.lex_state = 1, .external_lex_state = 3},
  [1251] = {.lex_state = 1, .external_lex_state = 3},
  [1252] = {.lex_state = 1, .external_lex_state = 3},
  [1253] = {.lex_state = 1, .external_lex_state = 3},
  [1254] = {.lex_state = 1, .external_lex_state = 3},
  [1255] = {.lex_state = 1, .external_lex_state = 3},
  [1256] = {.lex_state = 1, .external_lex_state = 3},
  [1257] = {.lex_state = 1, .external_lex_state = 3},
  [1258] = {.lex_state = 1, .external_lex_state = 3},
  [1259] = {.lex_state = 1, .external_lex_state = 3},
  [1260] = {.lex_state = 1, .external_lex_state = 13},
  [1261] = {.lex_state = 1, .external_lex_state = 13},
  [1262] = {.lex_state = 1, .external_lex_state = 13},
  [1263] = {.lex_state = 1, .external_lex_state = 13},
  [1264] = {.lex_state = 1, .external_lex_state = 13},
  [1265] = {.lex_state = 1, .external_lex_state = 13},
  [1266] = {.lex_state = 1, .external_lex_state = 13},
  [1267] = {.lex_state = 1, .external_lex_state = 13},
  [1268] = {.lex_state = 1, .external_lex_state = 13},
  [1269] = {.lex_state = 1, .external_lex_state = 13},
  [1270] = {.lex_state = 1, .external_lex_state = 14},
  [1271] = {.lex_state = 1, .external_lex_state = 15},
  [1272] = {.lex_state = 1, .external_lex_state = 16},
  [1273] = {.lex_state = 1, .external_lex_state = 4},
  [1274] = {.lex_state = 1, .external_lex_state = 16},
  [1275] = {.lex_state = 1, .external_lex_state = 14},
  [1276] = {.lex_state = 1, .external_lex_state = 4},
  [1277] = {.lex_state = 1, .external_lex_state = 4},
  [1278] = {.lex_state = 1, .external_lex_state = 15},
  [1279] = {.lex_state = 1, .external_lex_state = 15},
  [1280] = {.lex_state = 1, .external_lex_state = 16},
  [1281] = {.lex_state = 1, .external_lex_state = 15},
  [1282] = {.lex_state = 1, .external_lex_state = 16},
  [1283] = {.lex_state = 1, .external_lex_state = 4},
  [1284] = {.lex_state = 1, .external_lex_state = 4},
  [1285] = {.lex_state = 1, .external_lex_state = 14},
  [1286] = {.lex_state = 1, .external_lex_state = 16},
  [1287] = {.lex_state = 1, .external_lex_state = 15},
  [1288] = {.lex_state = 1, .external_lex_state = 4},
  [1289] = {.lex_state = 1, .external_lex_state = 14},
  [1290] = {.lex_state = 1, .external_lex_state = 4},
  [1291] = {.lex_state = 1, .external_lex_state = 15},
  [1292] = {.lex_state = 1, .external_lex_state = 14},
  [1293] = {.lex_state = 1, .external_lex_state = 15},
  [1294] = {.lex_state = 1, .external_lex_state = 4},
  [1295] = {.lex_state = 1, .external_lex_state = 15},
  [1296] = {.lex_state = 1, .external_lex_state = 14},
  [1297] = {.lex_state = 1, .external_lex_state = 4},
  [1298] = {.lex_state = 1, .external_lex_state = 14},
  [1299] = {.lex_state = 1, .external_lex_state = 14},
  [1300] = {.lex_state = 1, .external_lex_state = 16},
  [1301] = {.lex_state = 1, .external_lex_state = 16},
  [1302] = {.lex_state = 1, .external_lex_state = 14},
  [1303] = {.lex_state = 1, .external_lex_state = 4},
  [1304] = {.lex_state = 1, .external_lex_state = 16},
  [1305] = {.lex_state = 1, .external_lex_state = 15},
  [1306] = {.lex_state = 1, .external_lex_state = 4},
  [1307] = {.lex_state = 1, .external_lex_state = 4},
  [1308] = {.lex_state = 1, .external_lex_state = 4},
  [1309] = {.lex_state = 1, .external_lex_state = 15},
  [1310] = {.lex_state = 1, .external_lex_state = 4},
  [1311] = {.lex_state = 1, .external_lex_state = 14},
  [1312] = {.lex_state = 1, .external_lex_state = 4},
  [1313] = {.lex_state = 1, .external_lex_state = 16},
  [1314] = {.lex_state = 1, .external_lex_state = 4},
  [1315] = {.lex_state = 1, .external_lex_state = 16},
  [1316] = {.lex_state = 1, .external_lex_state = 6},
  [1317] = {.lex_state = 1, .external_lex_state = 6},
  [1318] = {.lex_state = 1, .external_lex_state = 6},
  [1319] = {.lex_state = 1, .external_lex_state = 5},
  [1320] = {.lex_state = 1, .external_lex_state = 5},
  [1321] = {.lex_state = 1, .external_lex_state = 5},
  [1322] = {.lex_state = 1, .external_lex_state = 5},
  [1323] = {.lex_state = 1, .external_lex_state = 7},
  [1324] = {.lex_state = 1, .external_lex_state = 6},
  [1325] = {.lex_state = 1, .external_lex_state = 6},
  [1326] = {.lex_state = 1, .external_lex_state = 6},
  [1327] = {.lex_state = 1, .external_lex_state = 6},
  [1328] = {.lex_state = 1, .external_lex_state = 5},
  [1329] = {.lex_state = 1, .external_lex_state = 5},
  [1330] = {.lex_state = 1, .external_lex_state = 5},
  [1331] = {.lex_state = 1, .external_lex_state = 6},
  [1332] = {.lex_state = 1, .external_lex_state = 5},
  [1333] = {.lex_state = 1, .external_lex_state = 5},
  [1334] = {.lex_state = 1, .external_lex_state = 5},
  [1335] = {.lex_state = 1, .external_lex_state = 5},
  [1336] = {.lex_state = 1, .external_lex_state = 6},
  [1337] = {.lex_state = 1, .external_lex_state = 5},
  [1338] = {.lex_state = 1, .external_lex_state = 6},
  [1339] = {.lex_state = 1, .external_lex_state = 5},
  [1340] = {.lex_state = 1, .external_lex_state = 5},
  [1341] = {.lex_state = 1, .external_lex_state = 7},
  [1342] = {.lex_state = 1, .external_lex_state = 7},
  [1343] = {.lex_state = 1, .external_lex_state = 5},
  [1344] = {.lex_state = 1, .external_lex_state = 7},
  [1345] = {.lex_state = 1, .external_lex_state = 7},
  [1346] = {.lex_state = 1, .external_lex_state = 7},
  [1347] = {.lex_state = 1, .external_lex_state = 7},
  [1348] = {.lex_state = 1, .external_lex_state = 5},
  [1349] = {.lex_state = 1, .external_lex_state = 7},
  [1350] = {.lex_state = 1, .external_lex_state = 7},
  [1351] = {.lex_state = 1, .external_lex_state = 6},
  [1352] = {.lex_state = 1, .external_lex_state = 7},
  [1353] = {.lex_state = 1, .external_lex_state = 6},
  [1354] = {.lex_state = 1, .external_lex_state = 6},
  [1355] = {.lex_state = 1, .external_lex_state = 7},
  [1356] = {.lex_state = 1, .external_lex_state = 6},
  [1357] = {.lex_state = 1, .external_lex_state = 7},
  [1358] = {.lex_state = 1, .external_lex_state = 6},
  [1359] = {.lex_state = 1, .external_lex_state = 7},
  [1360] = {.lex_state = 1, .external_lex_state = 7},
  [1361] = {.lex_state = 1, .external_lex_state = 6},
  [1362] = {.lex_state = 1, .external_lex_state = 7},
  [1363] = {.lex_state = 1, .external_lex_state = 7},
  [1364] = {.lex_state = 51, .external_lex_state = 13},
  [1365] = {.lex_state = 51, .external_lex_state = 13},
  [1366] = {.lex_state = 51, .external_lex_state = 13},
  [1367] = {.lex_state = 51, .external_lex_state = 13},
  [1368] = {.lex_state = 51, .external_lex_state = 13},
  [1369] = {.lex_state = 51, .external_lex_state = 13},
  [1370] = {.lex_state = 51, .external_lex_state = 13},
  [1371] = {.lex_state = 51, .external_lex_state = 13},
  [1372] = {.lex_state = 51, .external_lex_state = 13},
  [1373] = {.lex_state = 51, .external_lex_state = 13},
  [1374] = {.lex_state = 51, .external_lex_state = 4},
  [1375] = {.lex_state = 51, .external_lex_state = 4},
  [1376] = {.lex_state = 51, .external_lex_state = 17},
  [1377] = {.lex_state = 51, .external_lex_state = 13},
  [1378] = {.lex_state = 51, .external_lex_state = 17},
  [1379] = {.lex_state = 51, .external_lex_state = 4},
  [1380] = {.lex_state = 51, .external_lex_state = 4},
  [1381] = {.lex_state = 51, .external_lex_state = 4},
  [1382] = {.lex_state = 51, .external_lex_state = 4},
  [1383] = {.lex_state = 51, .external_lex_state = 4},
  [1384] = {.lex_state = 51, .external_lex_state = 4},
  [1385] = {.lex_state = 51, .external_lex_state = 4},
  [1386] = {.lex_state = 51, .external_lex_state = 4},
  [1387] = {.lex_state = 51, .external_lex_state = 4},
  [1388] = {.lex_state = 51, .external_lex_state = 17},
  [1389] = {.lex_state = 51, .external_lex_state = 4},
  [1390] = {.lex_state = 51, .external_lex_state = 4},
  [1391] = {.lex_state = 51, .external_lex_state = 17},
  [1392] = {.lex_state = 51, .external_lex_state = 17},
  [1393] = {.lex_state = 51, .external_lex_state = 17},
  [1394] = {.lex_state = 51, .external_lex_state = 17},
  [1395] = {.lex_state = 51, .external_lex_state = 4},
  [1396] = {.lex_state = 51, .external_lex_state = 4},
  [1397] = {.lex_state = 51, .external_lex_state = 17},
  [1398] = {.lex_state = 51, .external_lex_state = 4},
  [1399] = {.lex_state = 51, .external_lex_state = 17},
  [1400] = {.lex_state = 51, .external_lex_state = 17},
  [1401] = {.lex_state = 51, .external_lex_state = 8},
  [1402] = {.lex_state = 51, .external_lex_state = 8},
  [1403] = {.lex_state = 51, .external_lex_state = 4},
  [1404] = {.lex_state = 51, .external_lex_state = 8},
  [1405] = {.lex_state = 51, .external_lex_state = 8},
  [1406] = {.lex_state = 51, .external_lex_state = 8},
  [1407] = {.lex_state = 51, .external_lex_state = 4},
  [1408] = {.lex_state = 51, .external_lex_state = 4},
  [1409] = {.lex_state = 51, .external_lex_state = 8},
  [1410] = {.lex_state = 51, .external_lex_state = 8},
  [1411] = {.lex_state = 51, .external_lex_state = 8},
  [1412] = {.lex_state = 51, .external_lex_state = 8},
  [1413] = {.lex_state = 51, .external_lex_state = 8},
  [1414] = {.lex_state = 51, .external_lex_state = 8},
  [1415] = {.lex_state = 51, .external_lex_state = 8},
  [1416] = {.lex_state = 51, .external_lex_state = 8},
  [1417] = {.lex_state = 51, .external_lex_state = 8},
  [1418] = {.lex_state = 51, .external_lex_state = 4},
  [1419] = {.lex_state = 51, .external_lex_state = 4},
  [1420] = {.lex_state = 51, .external_lex_state = 8},
  [1421] = {.lex_state = 51, .external_lex_state = 8},
  [1422] = {.lex_state = 51, .external_lex_state = 8},
  [1423] = {.lex_state = 111, .external_lex_state = 2},
  [1424] = {.lex_state = 111, .external_lex_state = 2},
  [1425] = {.lex_state = 111, .external_lex_state = 2},
  [1426] = {.lex_state = 112, .external_lex_state = 2},
  [1427] = {.lex_state = 112, .external_lex_state = 2},
  [1428] = {.lex_state = 112, .external_lex_state = 2},
  [1429] = {.lex_state = 111, .external_lex_state = 2},
  [1430] = {.lex_state = 112, .external_lex_state = 2},
  [1431] = {.lex_state = 111, .external_lex_state = 4},
  [1432] = {.lex_state = 112, .external_lex_state = 2},
  [1433] = {.lex_state = 112, .external_lex_state = 2},
  [1434] = {.lex_state = 112, .external_lex_state = 2},
  [1435] = {.lex_state = 111, .external_lex_state = 4},
  [1436] = {.lex_state = 112, .external_lex_state = 2},
  [1437] = {.lex_state = 112, .external_lex_state = 2},
  [1438] = {.lex_state = 112, .external_lex_state = 2},
  [1439] = {.lex_state = 112, .external_lex_state = 2},
  [1440] = {.lex_state = 112, .external_lex_state = 2},
  [1441] = {.lex_state = 112, .external_lex_state = 2},
  [1442] = {.lex_state = 112, .external_lex_state = 2},
  [1443] = {.lex_state = 112, .external_lex_state = 2},
  [1444] = {.lex_state = 112, .external_lex_state = 2},
  [1445] = {.lex_state = 112, .external_lex_state = 2},
  [1446] = {.lex_state = 112, .external_lex_state = 2},
  [1447] = {.lex_state = 112, .external_lex_state = 2},
  [1448] = {.lex_state = 112, .external_lex_state = 2},
  [1449] = {.lex_state = 112, .external_lex_state = 2},
  [1450] = {.lex_state = 112, .external_lex_state = 2},
  [1451] = {.lex_state = 112, .external_lex_state = 2},
  [1452] = {.lex_state = 112, .external_lex_state = 2},
  [1453] = {.lex_state = 112, .external_lex_state = 2},
  [1454] = {.lex_state = 112, .external_lex_state = 2},
  [1455] = {.lex_state = 112, .external_lex_state = 2},
  [1456] = {.lex_state = 112, .external_lex_state = 2},
  [1457] = {.lex_state = 112, .external_lex_state = 2},
  [1458] = {.lex_state = 112, .external_lex_state = 2},
  [1459] = {.lex_state = 112, .external_lex_state = 2},
  [1460] = {.lex_state = 112, .external_lex_state = 2},
  [1461] = {.lex_state = 112, .external_lex_state = 2},
  [1462] = {.lex_state = 112, .external_lex_state = 2},
  [1463] = {.lex_state = 112, .external_lex_state = 2},
  [1464] = {.lex_state = 112, .external_lex_state = 2},
  [1465] = {.lex_state = 112, .external_lex_state = 2},
  [1466] = {.lex_state = 112, .external_lex_state = 2},
  [1467] = {.lex_state = 112, .external_lex_state = 2},
  [1468] = {.lex_state = 112, .external_lex_state = 2},
  [1469] = {.lex_state = 112, .external_lex_state = 2},
  [1470] = {.lex_state = 112, .external_lex_state = 2},
  [1471] = {.lex_state = 112, .external_lex_state = 2},
  [1472] = {.lex_state = 112, .external_lex_state = 2},
  [1473] = {.lex_state = 112, .external_lex_state = 2},
  [1474] = {.lex_state = 112, .external_lex_state = 2},
  [1475] = {.lex_state = 112, .external_lex_state = 2},
  [1476] = {.lex_state = 112, .external_lex_state = 2},
  [1477] = {.lex_state = 112, .external_lex_state = 2},
  [1478] = {.lex_state = 112, .external_lex_state = 2},
  [1479] = {.lex_state = 112, .external_lex_state = 2},
  [1480] = {.lex_state = 112, .external_lex_state = 2},
  [1481] = {.lex_state = 112, .external_lex_state = 2},
  [1482] = {.lex_state = 112, .external_lex_state = 2},
  [1483] = {.lex_state = 112, .external_lex_state = 2},
  [1484] = {.lex_state = 112, .external_lex_state = 2},
  [1485] = {.lex_state = 112, .external_lex_state = 2},
  [1486] = {.lex_state = 112, .external_lex_state = 2},
  [1487] = {.lex_state = 112, .external_lex_state = 2},
  [1488] = {.lex_state = 112, .external_lex_state = 2},
  [1489] = {.lex_state = 112, .external_lex_state = 2},
  [1490] = {.lex_state = 112, .external_lex_state = 2},
  [1491] = {.lex_state = 112, .external_lex_state = 2},
  [1492] = {.lex_state = 112, .external_lex_state = 2},
  [1493] = {.lex_state = 112, .external_lex_state = 2},
  [1494] = {.lex_state = 112, .external_lex_state = 2},
  [1495] = {.lex_state = 112, .external_lex_state = 2},
  [1496] = {.lex_state = 112, .external_lex_state = 2},
  [1497] = {.lex_state = 112, .external_lex_state = 2},
  [1498] = {.lex_state = 112, .external_lex_state = 2},
  [1499] = {.lex_state = 112, .external_lex_state = 2},
  [1500] = {.lex_state = 112, .external_lex_state = 2},
  [1501] = {.lex_state = 112, .external_lex_state = 2},
  [1502] = {.lex_state = 112, .external_lex_state = 2},
  [1503] = {.lex_state = 112, .external_lex_state = 2},
  [1504] = {.lex_state = 112, .external_lex_state = 2},
  [1505] = {.lex_state = 112, .external_lex_state = 2},
  [1506] = {.lex_state = 112, .external_lex_state = 2},
  [1507] = {.lex_state = 112, .external_lex_state = 2},
  [1508] = {.lex_state = 112, .external_lex_state = 2},
  [1509] = {.lex_state = 112, .external_lex_state = 2},
  [1510] = {.lex_state = 112, .external_lex_state = 2},
  [1511] = {.lex_state = 112, .external_lex_state = 2},
  [1512] = {.lex_state = 112, .external_lex_state = 2},
  [1513] = {.lex_state = 112, .external_lex_state = 2},
  [1514] = {.lex_state = 112, .external_lex_state = 2},
  [1515] = {.lex_state = 112, .external_lex_state = 2},
  [1516] = {.lex_state = 112, .external_lex_state = 2},
  [1517] = {.lex_state = 112, .external_lex_state = 2},
  [1518] = {.lex_state = 112, .external_lex_state = 2},
  [1519] = {.lex_state = 112, .external_lex_state = 2},
  [1520] = {.lex_state = 112, .external_lex_state = 2},
  [1521] = {.lex_state = 112, .external_lex_state = 2},
  [1522] = {.lex_state = 112, .external_lex_state = 2},
  [1523] = {.lex_state = 112, .external_lex_state = 2},
  [1524] = {.lex_state = 112, .external_lex_state = 2},
  [1525] = {.lex_state = 112, .external_lex_state = 2},
  [1526] = {.lex_state = 112, .external_lex_state = 2},
  [1527] = {.lex_state = 54, .external_lex_state = 9},
  [1528] = {.lex_state = 112, .external_lex_state = 4},
  [1529] = {.lex_state = 112, .external_lex_state = 4},
  [1530] = {.lex_state = 111, .external_lex_state = 4},
  [1531] = {.lex_state = 112, .external_lex_state = 4},
  [1532] = {.lex_state = 111, .external_lex_state = 4},
  [1533] = {.lex_state = 112, .external_lex_state = 4},
  [1534] = {.lex_state = 112, .external_lex_state = 4},
  [1535] = {.lex_state = 112, .external_lex_state = 4},
  [1536] = {.lex_state = 112, .external_lex_state = 4},
  [1537] = {.lex_state = 112, .external_lex_state = 4},
  [1538] = {.lex_state = 112, .external_lex_state = 4},
  [1539] = {.lex_state = 112, .external_lex_state = 4},
  [1540] = {.lex_state = 112, .external_lex_state = 4},
  [1541] = {.lex_state = 112, .external_lex_state = 4},
  [1542] = {.lex_state = 112, .external_lex_state = 4},
  [1543] = {.lex_state = 112, .external_lex_state = 4},
  [1544] = {.lex_state = 112, .external_lex_state = 4},
  [1545] = {.lex_state = 112, .external_lex_state = 4},
  [1546] = {.lex_state = 112, .external_lex_state = 4},
  [1547] = {.lex_state = 112, .external_lex_state = 4},
  [1548] = {.lex_state = 112, .external_lex_state = 4},
  [1549] = {.lex_state = 112, .external_lex_state = 4},
  [1550] = {.lex_state = 112, .external_lex_state = 4},
  [1551] = {.lex_state = 112, .external_lex_state = 4},
  [1552] = {.lex_state = 112, .external_lex_state = 4},
  [1553] = {.lex_state = 112, .external_lex_state = 4},
  [1554] = {.lex_state = 112, .external_lex_state = 4},
  [1555] = {.lex_state = 112, .external_lex_state = 4},
  [1556] = {.lex_state = 112, .external_lex_state = 4},
  [1557] = {.lex_state = 112, .external_lex_state = 4},
  [1558] = {.lex_state = 112, .external_lex_state = 4},
  [1559] = {.lex_state = 112, .external_lex_state = 4},
  [1560] = {.lex_state = 112, .external_lex_state = 4},
  [1561] = {.lex_state = 112, .external_lex_state = 4},
  [1562] = {.lex_state = 112, .external_lex_state = 4},
  [1563] = {.lex_state = 112, .external_lex_state = 4},
  [1564] = {.lex_state = 112, .external_lex_state = 4},
  [1565] = {.lex_state = 112, .external_lex_state = 4},
  [1566] = {.lex_state = 112, .external_lex_state = 4},
  [1567] = {.lex_state = 112, .external_lex_state = 4},
  [1568] = {.lex_state = 112, .external_lex_state = 4},
  [1569] = {.lex_state = 112, .external_lex_state = 4},
  [1570] = {.lex_state = 112, .external_lex_state = 4},
  [1571] = {.lex_state = 112, .external_lex_state = 4},
  [1572] = {.lex_state = 112, .external_lex_state = 4},
  [1573] = {.lex_state = 112, .external_lex_state = 4},
  [1574] = {.lex_state = 112, .external_lex_state = 4},
  [1575] = {.lex_state = 112, .external_lex_state = 4},
  [1576] = {.lex_state = 112, .external_lex_state = 4},
  [1577] = {.lex_state = 112, .external_lex_state = 4},
  [1578] = {.lex_state = 112, .external_lex_state = 4},
  [1579] = {.lex_state = 112, .external_lex_state = 4},
  [1580] = {.lex_state = 112, .external_lex_state = 4},
  [1581] = {.lex_state = 112, .external_lex_state = 4},
  [1582] = {.lex_state = 112, .external_lex_state = 4},
  [1583] = {.lex_state = 112, .external_lex_state = 4},
  [1584] = {.lex_state = 112, .external_lex_state = 4},
  [1585] = {.lex_state = 112, .external_lex_state = 4},
  [1586] = {.lex_state = 112, .external_lex_state = 4},
  [1587] = {.lex_state = 112, .external_lex_state = 4},
  [1588] = {.lex_state = 112, .external_lex_state = 4},
  [1589] = {.lex_state = 112, .external_lex_state = 4},
  [1590] = {.lex_state = 112, .external_lex_state = 4},
  [1591] = {.lex_state = 112, .external_lex_state = 4},
  [1592] = {.lex_state = 112, .external_lex_state = 4},
  [1593] = {.lex_state = 112, .external_lex_state = 4},
  [1594] = {.lex_state = 112, .external_lex_state = 4},
  [1595] = {.lex_state = 112, .external_lex_state = 4},
  [1596] = {.lex_state = 112, .external_lex_state = 4},
  [1597] = {.lex_state = 112, .external_lex_state = 4},
  [1598] = {.lex_state = 112, .external_lex_state = 4},
  [1599] = {.lex_state = 112, .external_lex_state = 4},
  [1600] = {.lex_state = 112, .external_lex_state = 4},
  [1601] = {.lex_state = 112, .external_lex_state = 4},
  [1602] = {.lex_state = 112, .external_lex_state = 4},
  [1603] = {.lex_state = 112, .external_lex_state = 4},
  [1604] = {.lex_state = 112, .external_lex_state = 4},
  [1605] = {.lex_state = 112, .external_lex_state = 4},
  [1606] = {.lex_state = 112, .external_lex_state = 4},
  [1607] = {.lex_state = 112, .external_lex_state = 4},
  [1608] = {.lex_state = 112, .external_lex_state = 4},
  [1609] = {.lex_state = 112, .external_lex_state = 4},
  [1610] = {.lex_state = 112, .external_lex_state = 4},
  [1611] = {.lex_state = 112, .external_lex_state = 4},
  [1612] = {.lex_state = 112, .external_lex_state = 4},
  [1613] = {.lex_state = 112, .external_lex_state = 4},
  [1614] = {.lex_state = 112, .external_lex_state = 4},
  [1615] = {.lex_state = 112, .external_lex_state = 4},
  [1616] = {.lex_state = 112, .external_lex_state = 4},
  [1617] = {.lex_state = 112, .external_lex_state = 4},
  [1618] = {.lex_state = 112, .external_lex_state = 4},
  [1619] = {.lex_state = 112, .external_lex_state = 4},
  [1620] = {.lex_state = 112, .external_lex_state = 4},
  [1621] = {.lex_state = 112, .external_lex_state = 4},
  [1622] = {.lex_state = 112, .external_lex_state = 4},
  [1623] = {.lex_state = 112, .external_lex_state = 4},
  [1624] = {.lex_state = 112, .external_lex_state = 4},
  [1625] = {.lex_state = 112, .external_lex_state = 4},
  [1626] = {.lex_state = 112, .external_lex_state = 4},
  [1627] = {.lex_state = 112, .external_lex_state = 4},
  [1628] = {.lex_state = 112, .external_lex_state = 4},
  [1629] = {.lex_state = 53, .external_lex_state = 4},
  [1630] = {.lex_state = 53, .external_lex_state = 4},
  [1631] = {.lex_state = 52, .external_lex_state = 4},
  [1632] = {.lex_state = 55, .external_lex_state = 8},
  [1633] = {.lex_state = 53, .external_lex_state = 4},
  [1634] = {.lex_state = 52, .external_lex_state = 4},
  [1635] = {.lex_state = 52, .external_lex_state = 4},
  [1636] = {.lex_state = 53, .external_lex_state = 4},
  [1637] = {.lex_state = 52, .external_lex_state = 4},
  [1638] = {.lex_state = 52, .external_lex_state = 4},
  [1639] = {.lex_state = 52, .external_lex_state = 4},
  [1640] = {.lex_state = 52, .external_lex_state = 4},
  [1641] = {.lex_state = 52, .external_lex_state = 8},
  [1642] = {.lex_state = 52, .external_lex_state = 4},
  [1643] = {.lex_state = 52, .external_lex_state = 4},
  [1644] = {.lex_state = 52, .external_lex_state = 4},
  [1645] = {.lex_state = 52, .external_lex_state = 4},
  [1646] = {.lex_state = 52, .external_lex_state = 4},
  [1647] = {.lex_state = 52, .external_lex_state = 4},
  [1648] = {.lex_state = 52, .external_lex_state = 4},
  [1649] = {.lex_state = 52, .external_lex_state = 4},
  [1650] = {.lex_state = 52, .external_lex_state = 4},
  [1651] = {.lex_state = 52, .external_lex_state = 4},
  [1652] = {.lex_state = 52, .external_lex_state = 4},
  [1653] = {.lex_state = 52, .external_lex_state = 4},
  [1654] = {.lex_state = 52, .external_lex_state = 4},
  [1655] = {.lex_state = 52, .external_lex_state = 4},
  [1656] = {.lex_state = 52, .external_lex_state = 4},
  [1657] = {.lex_state = 52, .external_lex_state = 4},
  [1658] = {.lex_state = 52, .external_lex_state = 4},
  [1659] = {.lex_state = 52, .external_lex_state = 4},
  [1660] = {.lex_state = 52, .external_lex_state = 4},
  [1661] = {.lex_state = 52, .external_lex_state = 4},
  [1662] = {.lex_state = 52, .external_lex_state = 4},
  [1663] = {.lex_state = 52, .external_lex_state = 4},
  [1664] = {.lex_state = 52, .external_lex_state = 4},
  [1665] = {.lex_state = 52, .external_lex_state = 4},
  [1666] = {.lex_state = 52, .external_lex_state = 4},
  [1667] = {.lex_state = 52, .external_lex_state = 4},
  [1668] = {.lex_state = 52, .external_lex_state = 4},
  [1669] = {.lex_state = 52, .external_lex_state = 4},
  [1670] = {.lex_state = 52, .external_lex_state = 4},
  [1671] = {.lex_state = 52, .external_lex_state = 4},
  [1672] = {.lex_state = 52, .external_lex_state = 4},
  [1673] = {.lex_state = 52, .external_lex_state = 4},
  [1674] = {.lex_state = 52, .external_lex_state = 4},
  [1675] = {.lex_state = 52, .external_lex_state = 4},
  [1676] = {.lex_state = 52, .external_lex_state = 4},
  [1677] = {.lex_state = 52, .external_lex_state = 4},
  [1678] = {.lex_state = 52, .external_lex_state = 4},
  [1679] = {.lex_state = 52, .external_lex_state = 4},
  [1680] = {.lex_state = 52, .external_lex_state = 4},
  [1681] = {.lex_state = 52, .external_lex_state = 4},
  [1682] = {.lex_state = 52, .external_lex_state = 4},
  [1683] = {.lex_state = 52, .external_lex_state = 4},
  [1684] = {.lex_state = 52, .external_lex_state = 4},
  [1685] = {.lex_state = 52, .external_lex_state = 4},
  [1686] = {.lex_state = 52, .external_lex_state = 4},
  [1687] = {.lex_state = 52, .external_lex_state = 4},
  [1688] = {.lex_state = 52, .external_lex_state = 4},
  [1689] = {.lex_state = 52, .external_lex_state = 4},
  [1690] = {.lex_state = 52, .external_lex_state = 4},
  [1691] = {.lex_state = 52, .external_lex_state = 4},
  [1692] = {.lex_state = 52, .external_lex_state = 4},
  [1693] = {.lex_state = 52, .external_lex_state = 4},
  [1694] = {.lex_state = 52, .external_lex_state = 4},
  [1695] = {.lex_state = 52, .external_lex_state = 4},
  [1696] = {.lex_state = 52, .external_lex_state = 4},
  [1697] = {.lex_state = 52, .external_lex_state = 4},
  [1698] = {.lex_state = 52, .external_lex_state = 4},
  [1699] = {.lex_state = 52, .external_lex_state = 4},
  [1700] = {.lex_state = 52, .external_lex_state = 4},
  [1701] = {.lex_state = 52, .external_lex_state = 4},
  [1702] = {.lex_state = 52, .external_lex_state = 4},
  [1703] = {.lex_state = 52, .external_lex_state = 4},
  [1704] = {.lex_state = 52, .external_lex_state = 4},
  [1705] = {.lex_state = 52, .external_lex_state = 4},
  [1706] = {.lex_state = 52, .external_lex_state = 4},
  [1707] = {.lex_state = 52, .external_lex_state = 4},
  [1708] = {.lex_state = 52, .external_lex_state = 4},
  [1709] = {.lex_state = 52, .external_lex_state = 4},
  [1710] = {.lex_state = 52, .external_lex_state = 4},
  [1711] = {.lex_state = 52, .external_lex_state = 4},
  [1712] = {.lex_state = 52, .external_lex_state = 4},
  [1713] = {.lex_state = 52, .external_lex_state = 4},
  [1714] = {.lex_state = 52, .external_lex_state = 4},
  [1715] = {.lex_state = 52, .external_lex_state = 4},
  [1716] = {.lex_state = 52, .external_lex_state = 4},
  [1717] = {.lex_state = 52, .external_lex_state = 4},
  [1718] = {.lex_state = 52, .external_lex_state = 4},
  [1719] = {.lex_state = 52, .external_lex_state = 4},
  [1720] = {.lex_state = 52, .external_lex_state = 4},
  [1721] = {.lex_state = 52, .external_lex_state = 4},
  [1722] = {.lex_state = 52, .external_lex_state = 4},
  [1723] = {.lex_state = 52, .external_lex_state = 4},
  [1724] = {.lex_state = 52, .external_lex_state = 4},
  [1725] = {.lex_state = 52, .external_lex_state = 4},
  [1726] = {.lex_state = 52, .external_lex_state = 4},
  [1727] = {.lex_state = 52, .external_lex_state = 4},
  [1728] = {.lex_state = 52, .external_lex_state = 4},
  [1729] = {.lex_state = 52, .external_lex_state = 4},
  [1730] = {.lex_state = 52, .external_lex_state = 4},
  [1731] = {.lex_state = 52, .external_lex_state = 4},
  [1732] = {.lex_state = 52, .external_lex_state = 4},
  [1733] = {.lex_state = 52, .external_lex_state = 4},
  [1734] = {.lex_state = 52, .external_lex_state = 4},
  [1735] = {.lex_state = 52, .external_lex_state = 4},
  [1736] = {.lex_state = 113, .external_lex_state = 18},
  [1737] = {.lex_state = 113, .external_lex_state = 18},
  [1738] = {.lex_state = 113, .external_lex_state = 18},
  [1739] = {.lex_state = 113, .external_lex_state = 18},
  [1740] = {.lex_state = 113, .external_lex_state = 18},
  [1741] = {.lex_state = 3, .external_lex_state = 18},
  [1742] = {.lex_state = 113, .external_lex_state = 19},
  [1743] = {.lex_state = 3, .external_lex_state = 18},
  [1744] = {.lex_state = 113, .external_lex_state = 19},
  [1745] = {.lex_state = 3, .external_lex_state = 18},
  [1746] = {.lex_state = 3, .external_lex_state = 18},
  [1747] = {.lex_state = 3, .external_lex_state = 18},
  [1748] = {.lex_state = 3, .external_lex_state = 18},
  [1749] = {.lex_state = 3, .external_lex_state = 18},
  [1750] = {.lex_state = 3, .external_lex_state = 18},
  [1751] = {.lex_state = 3, .external_lex_state = 18},
  [1752] = {.lex_state = 3, .external_lex_state = 18},
  [1753] = {.lex_state = 3, .external_lex_state = 18},
  [1754] = {.lex_state = 3, .external_lex_state = 18},
  [1755] = {.lex_state = 3, .external_lex_state = 18},
  [1756] = {.lex_state = 3, .external_lex_state = 18},
  [1757] = {.lex_state = 3, .external_lex_state = 18},
  [1758] = {.lex_state = 3, .external_lex_state = 18},
  [1759] = {.lex_state = 3, .external_lex_state = 18},
  [1760] = {.lex_state = 3, .external_lex_state = 18},
  [1761] = {.lex_state = 3, .external_lex_state = 18},
  [1762] = {.lex_state = 113, .external_lex_state = 18},
  [1763] = {.lex_state = 3, .external_lex_state = 18},
  [1764] = {.lex_state = 3, .external_lex_state = 18},
  [1765] = {.lex_state = 3, .external_lex_state = 18},
  [1766] = {.lex_state = 113, .external_lex_state = 18},
  [1767] = {.lex_state = 3, .external_lex_state = 18},
  [1768] = {.lex_state = 3, .external_lex_state = 18},
  [1769] = {.lex_state = 113, .external_lex_state = 18},
  [1770] = {.lex_state = 3, .external_lex_state = 18},
  [1771] = {.lex_state = 3, .external_lex_state = 18},
  [1772] = {.lex_state = 3, .external_lex_state = 18},
  [1773] = {.lex_state = 6, .external_lex_state = 18},
  [1774] = {.lex_state = 3, .external_lex_state = 18},
  [1775] = {.lex_state = 6, .external_lex_state = 18},
  [1776] = {.lex_state = 3, .external_lex_state = 18},
  [1777] = {.lex_state = 3, .external_lex_state = 18},
  [1778] = {.lex_state = 8, .external_lex_state = 18},
  [1779] = {.lex_state = 3, .external_lex_state = 18},
  [1780] = {.lex_state = 9, .external_lex_state = 18},
  [1781] = {.lex_state = 9, .external_lex_state = 18},
  [1782] = {.lex_state = 3, .external_lex_state = 18},
  [1783] = {.lex_state = 3, .external_lex_state = 18},
  [1784] = {.lex_state = 3, .external_lex_state = 18},
  [1785] = {.lex_state = 8, .external_lex_state = 18},
  [1786] = {.lex_state = 8, .external_lex_state = 18},
  [1787] = {.lex_state = 6, .external_lex_state = 18},
  [1788] = {.lex_state = 9, .external_lex_state = 18},
  [1789] = {.lex_state = 3, .external_lex_state = 18},
  [1790] = {.lex_state = 3, .external_lex_state = 18},
  [1791] = {.lex_state = 3, .external_lex_state = 18},
  [1792] = {.lex_state = 6, .external_lex_state = 18},
  [1793] = {.lex_state = 8, .external_lex_state = 18},
  [1794] = {.lex_state = 3, .external_lex_state = 18},
  [1795] = {.lex_state = 3, .external_lex_state = 18},
  [1796] = {.lex_state = 3, .external_lex_state = 18},
  [1797] = {.lex_state = 3, .external_lex_state = 18},
  [1798] = {.lex_state = 3, .external_lex_state = 18},
  [1799] = {.lex_state = 3, .external_lex_state = 18},
  [1800] = {.lex_state = 8, .external_lex_state = 18},
  [1801] = {.lex_state = 3, .external_lex_state = 18},
  [1802] = {.lex_state = 3, .external_lex_state = 18},
  [1803] = {.lex_state = 3, .external_lex_state = 18},
  [1804] = {.lex_state = 3, .external_lex_state = 18},
  [1805] = {.lex_state = 3, .external_lex_state = 18},
  [1806] = {.lex_state = 10, .external_lex_state = 18},
  [1807] = {.lex_state = 3, .external_lex_state = 18},
  [1808] = {.lex_state = 10, .external_lex_state = 18},
  [1809] = {.lex_state = 3, .external_lex_state = 18},
  [1810] = {.lex_state = 9, .external_lex_state = 18},
  [1811] = {.lex_state = 9, .external_lex_state = 18},
  [1812] = {.lex_state = 3, .external_lex_state = 18},
  [1813] = {.lex_state = 3, .external_lex_state = 18},
  [1814] = {.lex_state = 3, .external_lex_state = 18},
  [1815] = {.lex_state = 3, .external_lex_state = 18},
  [1816] = {.lex_state = 3, .external_lex_state = 18},
  [1817] = {.lex_state = 6, .external_lex_state = 18},
  [1818] = {.lex_state = 3, .external_lex_state = 18},
  [1819] = {.lex_state = 3, .external_lex_state = 18},
  [1820] = {.lex_state = 3, .external_lex_state = 18},
  [1821] = {.lex_state = 10, .external_lex_state = 18},
  [1822] = {.lex_state = 9, .external_lex_state = 18},
  [1823] = {.lex_state = 3, .external_lex_state = 18},
  [1824] = {.lex_state = 3, .external_lex_state = 18},
  [1825] = {.lex_state = 3, .external_lex_state = 18},
  [1826] = {.lex_state = 3, .external_lex_state = 18},
  [1827] = {.lex_state = 3, .external_lex_state = 18},
  [1828] = {.lex_state = 3, .external_lex_state = 18},
  [1829] = {.lex_state = 3, .external_lex_state = 18},
  [1830] = {.lex_state = 3, .external_lex_state = 18},
  [1831] = {.lex_state = 3, .external_lex_state = 18},
  [1832] = {.lex_state = 3, .external_lex_state = 18},
  [1833] = {.lex_state = 3, .external_lex_state = 18},
  [1834] = {.lex_state = 3, .external_lex_state = 18},
  [1835] = {.lex_state = 3, .external_lex_state = 18},
  [1836] = {.lex_state = 3, .external_lex_state = 18},
  [1837] = {.lex_state = 3, .external_lex_state = 18},
  [1838] = {.lex_state = 3, .external_lex_state = 18},
  [1839] = {.lex_state = 3, .external_lex_state = 18},
  [1840] = {.lex_state = 3, .external_lex_state = 18},
  [1841] = {.lex_state = 3, .external_lex_state = 18},
  [1842] = {.lex_state = 3, .external_lex_state = 18},
  [1843] = {.lex_state = 3, .external_lex_state = 18},
  [1844] = {.lex_state = 3, .external_lex_state = 18},
  [1845] = {.lex_state = 3, .external_lex_state = 18},
  [1846] = {.lex_state = 3, .external_lex_state = 18},
  [1847] = {.lex_state = 3, .external_lex_state = 18},
  [1848] = {.lex_state = 3, .external_lex_state = 18},
  [1849] = {.lex_state = 3, .external_lex_state = 18},
  [1850] = {.lex_state = 3, .external_lex_state = 18},
  [1851] = {.lex_state = 3, .external_lex_state = 18},
  [1852] = {.lex_state = 3, .external_lex_state = 18},
  [1853] = {.lex_state = 3, .external_lex_state = 18},
  [1854] = {.lex_state = 3, .external_lex_state = 18},
  [1855] = {.lex_state = 3, .external_lex_state = 18},
  [1856] = {.lex_state = 3, .external_lex_state = 18},
  [1857] = {.lex_state = 3, .external_lex_state = 18},
  [1858] = {.lex_state = 3, .external_lex_state = 18},
  [1859] = {.lex_state = 3, .external_lex_state = 18},
  [1860] = {.lex_state = 3, .external_lex_state = 18},
  [1861] = {.lex_state = 3, .external_lex_state = 18},
  [1862] = {.lex_state = 3, .external_lex_state = 18},
  [1863] = {.lex_state = 3, .external_lex_state = 18},
  [1864] = {.lex_state = 3, .external_lex_state = 18},
  [1865] = {.lex_state = 3, .external_lex_state = 18},
  [1866] = {.lex_state = 3, .external_lex_state = 18},
  [1867] = {.lex_state = 3, .external_lex_state = 18},
  [1868] = {.lex_state = 3, .external_lex_state = 18},
  [1869] = {.lex_state = 3, .external_lex_state = 18},
  [1870] = {.lex_state = 3, .external_lex_state = 18},
  [1871] = {.lex_state = 3, .external_lex_state = 18},
  [1872] = {.lex_state = 3, .external_lex_state = 18},
  [1873] = {.lex_state = 3, .external_lex_state = 18},
  [1874] = {.lex_state = 3, .external_lex_state = 18},
  [1875] = {.lex_state = 3, .external_lex_state = 18},
  [1876] = {.lex_state = 3, .external_lex_state = 18},
  [1877] = {.lex_state = 3, .external_lex_state = 18},
  [1878] = {.lex_state = 3, .external_lex_state = 18},
  [1879] = {.lex_state = 3, .external_lex_state = 18},
  [1880] = {.lex_state = 3, .external_lex_state = 18},
  [1881] = {.lex_state = 3, .external_lex_state = 18},
  [1882] = {.lex_state = 3, .external_lex_state = 18},
  [1883] = {.lex_state = 3, .external_lex_state = 18},
  [1884] = {.lex_state = 3, .external_lex_state = 18},
  [1885] = {.lex_state = 3, .external_lex_state = 18},
  [1886] = {.lex_state = 3, .external_lex_state = 18},
  [1887] = {.lex_state = 3, .external_lex_state = 18},
  [1888] = {.lex_state = 3, .external_lex_state = 18},
  [1889] = {.lex_state = 3, .external_lex_state = 18},
  [1890] = {.lex_state = 3, .external_lex_state = 18},
  [1891] = {.lex_state = 3, .external_lex_state = 18},
  [1892] = {.lex_state = 3, .external_lex_state = 18},
  [1893] = {.lex_state = 3, .external_lex_state = 18},
  [1894] = {.lex_state = 3, .external_lex_state = 18},
  [1895] = {.lex_state = 3, .external_lex_state = 18},
  [1896] = {.lex_state = 3, .external_lex_state = 18},
  [1897] = {.lex_state = 3, .external_lex_state = 18},
  [1898] = {.lex_state = 3, .external_lex_state = 18},
  [1899] = {.lex_state = 3, .external_lex_state = 18},
  [1900] = {.lex_state = 3, .external_lex_state = 18},
  [1901] = {.lex_state = 3, .external_lex_state = 18},
  [1902] = {.lex_state = 3, .external_lex_state = 18},
  [1903] = {.lex_state = 3, .external_lex_state = 18},
  [1904] = {.lex_state = 3, .external_lex_state = 18},
  [1905] = {.lex_state = 3, .external_lex_state = 18},
  [1906] = {.lex_state = 3, .external_lex_state = 18},
  [1907] = {.lex_state = 3, .external_lex_state = 18},
  [1908] = {.lex_state = 3, .external_lex_state = 18},
  [1909] = {.lex_state = 3, .external_lex_state = 18},
  [1910] = {.lex_state = 3, .external_lex_state = 18},
  [1911] = {.lex_state = 3, .external_lex_state = 18},
  [1912] = {.lex_state = 3, .external_lex_state = 18},
  [1913] = {.lex_state = 3, .external_lex_state = 18},
  [1914] = {.lex_state = 3, .external_lex_state = 18},
  [1915] = {.lex_state = 3, .external_lex_state = 18},
  [1916] = {.lex_state = 3, .external_lex_state = 18},
  [1917] = {.lex_state = 3, .external_lex_state = 18},
  [1918] = {.lex_state = 3, .external_lex_state = 18},
  [1919] = {.lex_state = 3, .external_lex_state = 18},
  [1920] = {.lex_state = 3, .external_lex_state = 18},
  [1921] = {.lex_state = 3, .external_lex_state = 18},
  [1922] = {.lex_state = 3, .external_lex_state = 18},
  [1923] = {.lex_state = 3, .external_lex_state = 18},
  [1924] = {.lex_state = 3, .external_lex_state = 18},
  [1925] = {.lex_state = 3, .external_lex_state = 18},
  [1926] = {.lex_state = 3, .external_lex_state = 18},
  [1927] = {.lex_state = 3, .external_lex_state = 18},
  [1928] = {.lex_state = 3, .external_lex_state = 18},
  [1929] = {.lex_state = 3, .external_lex_state = 18},
  [1930] = {.lex_state = 3, .external_lex_state = 18},
  [1931] = {.lex_state = 3, .external_lex_state = 18},
  [1932] = {.lex_state = 3, .external_lex_state = 18},
  [1933] = {.lex_state = 3, .external_lex_state = 18},
  [1934] = {.lex_state = 3, .external_lex_state = 18},
  [1935] = {.lex_state = 3, .external_lex_state = 18},
  [1936] = {.lex_state = 3, .external_lex_state = 18},
  [1937] = {.lex_state = 3, .external_lex_state = 18},
  [1938] = {.lex_state = 3, .external_lex_state = 18},
  [1939] = {.lex_state = 3, .external_lex_state = 18},
  [1940] = {.lex_state = 3, .external_lex_state = 18},
  [1941] = {.lex_state = 3, .external_lex_state = 18},
  [1942] = {.lex_state = 3, .external_lex_state = 18},
  [1943] = {.lex_state = 3, .external_lex_state = 18},
  [1944] = {.lex_state = 3, .external_lex_state = 18},
  [1945] = {.lex_state = 3, .external_lex_state = 18},
  [1946] = {.lex_state = 3, .external_lex_state = 18},
  [1947] = {.lex_state = 3, .external_lex_state = 18},
  [1948] = {.lex_state = 3, .external_lex_state = 18},
  [1949] = {.lex_state = 3, .external_lex_state = 18},
  [1950] = {.lex_state = 3, .external_lex_state = 18},
  [1951] = {.lex_state = 3, .external_lex_state = 18},
  [1952] = {.lex_state = 3, .external_lex_state = 18},
  [1953] = {.lex_state = 3, .external_lex_state = 18},
  [1954] = {.lex_state = 3, .external_lex_state = 18},
  [1955] = {.lex_state = 3, .external_lex_state = 18},
  [1956] = {.lex_state = 3, .external_lex_state = 18},
  [1957] = {.lex_state = 3, .external_lex_state = 18},
  [1958] = {.lex_state = 3, .external_lex_state = 18},
  [1959] = {.lex_state = 3, .external_lex_state = 18},
  [1960] = {.lex_state = 3, .external_lex_state = 18},
  [1961] = {.lex_state = 3, .external_lex_state = 18},
  [1962] = {.lex_state = 3, .external_lex_state = 18},
  [1963] = {.lex_state = 3, .external_lex_state = 18},
  [1964] = {.lex_state = 3, .external_lex_state = 18},
  [1965] = {.lex_state = 3, .external_lex_state = 18},
  [1966] = {.lex_state = 3, .external_lex_state = 18},
  [1967] = {.lex_state = 3, .external_lex_state = 18},
  [1968] = {.lex_state = 3, .external_lex_state = 18},
  [1969] = {.lex_state = 113, .external_lex_state = 18},
  [1970] = {.lex_state = 113, .external_lex_state = 18},
  [1971] = {.lex_state = 113, .external_lex_state = 18},
  [1972] = {.lex_state = 113, .external_lex_state = 18},
  [1973] = {.lex_state = 113, .external_lex_state = 18},
  [1974] = {.lex_state = 113, .external_lex_state = 19},
  [1975] = {.lex_state = 113, .external_lex_state = 19},
  [1976] = {.lex_state = 113, .external_lex_state = 20},
  [1977] = {.lex_state = 113, .external_lex_state = 18},
  [1978] = {.lex_state = 113, .external_lex_state = 18},
  [1979] = {.lex_state = 113, .external_lex_state = 20},
  [1980] = {.lex_state = 113, .external_lex_state = 20},
  [1981] = {.lex_state = 113, .external_lex_state = 18},
  [1982] = {.lex_state = 113, .external_lex_state = 20},
  [1983] = {.lex_state = 113, .external_lex_state = 20},
  [1984] = {.lex_state = 113, .external_lex_state = 21},
  [1985] = {.lex_state = 113, .external_lex_state = 22},
  [1986] = {.lex_state = 113, .external_lex_state = 22},
  [1987] = {.lex_state = 113, .external_lex_state = 22},
  [1988] = {.lex_state = 113, .external_lex_state = 23},
  [1989] = {.lex_state = 113, .external_lex_state = 23},
  [1990] = {.lex_state = 113, .external_lex_state = 23},
  [1991] = {.lex_state = 113, .external_lex_state = 21},
  [1992] = {.lex_state = 113, .external_lex_state = 21},
  [1993] = {.lex_state = 3, .external_lex_state = 18},
  [1994] = {.lex_state = 3, .external_lex_state = 18},
  [1995] = {.lex_state = 3, .external_lex_state = 20},
  [1996] = {.lex_state = 3, .external_lex_state = 20},
  [1997] = {.lex_state = 56, .external_lex_state = 20},
  [1998] = {.lex_state = 3, .external_lex_state = 18},
  [1999] = {.lex_state = 56, .external_lex_state = 20},
  [2000] = {.lex_state = 56, .external_lex_state = 20},
  [2001] = {.lex_state = 113, .external_lex_state = 22},
  [2002] = {.lex_state = 113, .external_lex_state = 22},
  [2003] = {.lex_state = 3, .external_lex_state = 18},
  [2004] = {.lex_state = 113, .external_lex_state = 21},
  [2005] = {.lex_state = 113, .external_lex_state = 21},
  [2006] = {.lex_state = 113, .external_lex_state = 24},
  [2007] = {.lex_state = 113, .external_lex_state = 24},
  [2008] = {.lex_state = 113, .external_lex_state = 23},
  [2009] = {.lex_state = 113, .external_lex_state = 23},
  [2010] = {.lex_state = 3, .external_lex_state = 18},
  [2011] = {.lex_state = 3, .external_lex_state = 18},
  [2012] = {.lex_state = 3, .external_lex_state = 18},
  [2013] = {.lex_state = 3, .external_lex_state = 20},
  [2014] = {.lex_state = 3, .external_lex_state = 20},
  [2015] = {.lex_state = 113, .external_lex_state = 20},
  [2016] = {.lex_state = 3, .external_lex_state = 20},
  [2017] = {.lex_state = 113, .external_lex_state = 25},
  [2018] = {.lex_state = 3, .external_lex_state = 20},
  [2019] = {.lex_state = 113, .external_lex_state = 25},
  [2020] = {.lex_state = 3, .external_lex_state = 20},
  [2021] = {.lex_state = 3, .external_lex_state = 20},
  [2022] = {.lex_state = 3, .external_lex_state = 20},
  [2023] = {.lex_state = 3, .external_lex_state = 20},
  [2024] = {.lex_state = 3, .external_lex_state = 20},
  [2025] = {.lex_state = 3, .external_lex_state = 20},
  [2026] = {.lex_state = 3, .external_lex_state = 21},
  [2027] = {.lex_state = 3, .external_lex_state = 20},
  [2028] = {.lex_state = 3, .external_lex_state = 20},
  [2029] = {.lex_state = 3, .external_lex_state = 20},
  [2030] = {.lex_state = 3, .external_lex_state = 20},
  [2031] = {.lex_state = 3, .external_lex_state = 20},
  [2032] = {.lex_state = 3, .external_lex_state = 20},
  [2033] = {.lex_state = 113, .external_lex_state = 26},
  [2034] = {.lex_state = 3, .external_lex_state = 18},
  [2035] = {.lex_state = 113, .external_lex_state = 20},
  [2036] = {.lex_state = 3, .external_lex_state = 22},
  [2037] = {.lex_state = 3, .external_lex_state = 22},
  [2038] = {.lex_state = 3, .external_lex_state = 23},
  [2039] = {.lex_state = 3, .external_lex_state = 23},
  [2040] = {.lex_state = 3, .external_lex_state = 20},
  [2041] = {.lex_state = 3, .external_lex_state = 18},
  [2042] = {.lex_state = 3, .external_lex_state = 18},
  [2043] = {.lex_state = 3, .external_lex_state = 18},
  [2044] = {.lex_state = 3, .external_lex_state = 18},
  [2045] = {.lex_state = 3, .external_lex_state = 18},
  [2046] = {.lex_state = 3, .external_lex_state = 18},
  [2047] = {.lex_state = 3, .external_lex_state = 18},
  [2048] = {.lex_state = 3, .external_lex_state = 21},
  [2049] = {.lex_state = 3, .external_lex_state = 18},
  [2050] = {.lex_state = 3, .external_lex_state = 18},
  [2051] = {.lex_state = 3, .external_lex_state = 18},
  [2052] = {.lex_state = 3, .external_lex_state = 18},
  [2053] = {.lex_state = 3, .external_lex_state = 18},
  [2054] = {.lex_state = 3, .external_lex_state = 18},
  [2055] = {.lex_state = 3, .external_lex_state = 20},
  [2056] = {.lex_state = 3, .external_lex_state = 20},
  [2057] = {.lex_state = 113, .external_lex_state = 20},
  [2058] = {.lex_state = 3, .external_lex_state = 20},
  [2059] = {.lex_state = 113, .external_lex_state = 27},
  [2060] = {.lex_state = 3, .external_lex_state = 20},
  [2061] = {.lex_state = 3, .external_lex_state = 20},
  [2062] = {.lex_state = 3, .external_lex_state = 18},
  [2063] = {.lex_state = 113, .external_lex_state = 27},
  [2064] = {.lex_state = 113, .external_lex_state = 26},
  [2065] = {.lex_state = 3, .external_lex_state = 18},
  [2066] = {.lex_state = 3, .external_lex_state = 18},
  [2067] = {.lex_state = 3, .external_lex_state = 18},
  [2068] = {.lex_state = 3, .external_lex_state = 18},
  [2069] = {.lex_state = 3, .external_lex_state = 18},
  [2070] = {.lex_state = 3, .external_lex_state = 18},
  [2071] = {.lex_state = 3, .external_lex_state = 18},
  [2072] = {.lex_state = 3, .external_lex_state = 18},
  [2073] = {.lex_state = 3, .external_lex_state = 18},
  [2074] = {.lex_state = 3, .external_lex_state = 18},
  [2075] = {.lex_state = 3, .external_lex_state = 18},
  [2076] = {.lex_state = 3, .external_lex_state = 20},
  [2077] = {.lex_state = 3, .external_lex_state = 18},
  [2078] = {.lex_state = 3, .external_lex_state = 23},
  [2079] = {.lex_state = 113, .external_lex_state = 21},
  [2080] = {.lex_state = 3, .external_lex_state = 21},
  [2081] = {.lex_state = 3, .external_lex_state = 21},
  [2082] = {.lex_state = 65, .external_lex_state = 20},
  [2083] = {.lex_state = 3, .external_lex_state = 21},
  [2084] = {.lex_state = 3, .external_lex_state = 21},
  [2085] = {.lex_state = 6, .external_lex_state = 20},
  [2086] = {.lex_state = 3, .external_lex_state = 23},
  [2087] = {.lex_state = 65, .external_lex_state = 20},
  [2088] = {.lex_state = 3, .external_lex_state = 23},
  [2089] = {.lex_state = 3, .external_lex_state = 23},
  [2090] = {.lex_state = 3, .external_lex_state = 23},
  [2091] = {.lex_state = 3, .external_lex_state = 21},
  [2092] = {.lex_state = 3, .external_lex_state = 23},
  [2093] = {.lex_state = 3, .external_lex_state = 23},
  [2094] = {.lex_state = 3, .external_lex_state = 23},
  [2095] = {.lex_state = 3, .external_lex_state = 22},
  [2096] = {.lex_state = 3, .external_lex_state = 23},
  [2097] = {.lex_state = 3, .external_lex_state = 23},
  [2098] = {.lex_state = 65, .external_lex_state = 20},
  [2099] = {.lex_state = 3, .external_lex_state = 21},
  [2100] = {.lex_state = 3, .external_lex_state = 21},
  [2101] = {.lex_state = 3, .external_lex_state = 21},
  [2102] = {.lex_state = 3, .external_lex_state = 21},
  [2103] = {.lex_state = 3, .external_lex_state = 21},
  [2104] = {.lex_state = 3, .external_lex_state = 21},
  [2105] = {.lex_state = 3, .external_lex_state = 22},
  [2106] = {.lex_state = 3, .external_lex_state = 22},
  [2107] = {.lex_state = 3, .external_lex_state = 21},
  [2108] = {.lex_state = 3, .external_lex_state = 21},
  [2109] = {.lex_state = 3, .external_lex_state = 23},
  [2110] = {.lex_state = 3, .external_lex_state = 23},
  [2111] = {.lex_state = 3, .external_lex_state = 23},
  [2112] = {.lex_state = 3, .external_lex_state = 23},
  [2113] = {.lex_state = 3, .external_lex_state = 21},
  [2114] = {.lex_state = 3, .external_lex_state = 22},
  [2115] = {.lex_state = 3, .external_lex_state = 21},
  [2116] = {.lex_state = 3, .external_lex_state = 21},
  [2117] = {.lex_state = 3, .external_lex_state = 21},
  [2118] = {.lex_state = 3, .external_lex_state = 23},
  [2119] = {.lex_state = 3, .external_lex_state = 22},
  [2120] = {.lex_state = 3, .external_lex_state = 23},
  [2121] = {.lex_state = 3, .external_lex_state = 23},
  [2122] = {.lex_state = 3, .external_lex_state = 23},
  [2123] = {.lex_state = 3, .external_lex_state = 23},
  [2124] = {.lex_state = 3, .external_lex_state = 21},
  [2125] = {.lex_state = 3, .external_lex_state = 21},
  [2126] = {.lex_state = 3, .external_lex_state = 22},
  [2127] = {.lex_state = 3, .external_lex_state = 22},
  [2128] = {.lex_state = 3, .external_lex_state = 23},
  [2129] = {.lex_state = 3, .external_lex_state = 23},
  [2130] = {.lex_state = 3, .external_lex_state = 22},
  [2131] = {.lex_state = 57, .external_lex_state = 20},
  [2132] = {.lex_state = 3, .external_lex_state = 23},
  [2133] = {.lex_state = 113, .external_lex_state = 21},
  [2134] = {.lex_state = 113, .external_lex_state = 22},
  [2135] = {.lex_state = 3, .external_lex_state = 22},
  [2136] = {.lex_state = 3, .external_lex_state = 22},
  [2137] = {.lex_state = 3, .external_lex_state = 22},
  [2138] = {.lex_state = 3, .external_lex_state = 22},
  [2139] = {.lex_state = 3, .external_lex_state = 22},
  [2140] = {.lex_state = 3, .external_lex_state = 22},
  [2141] = {.lex_state = 3, .external_lex_state = 22},
  [2142] = {.lex_state = 3, .external_lex_state = 23},
  [2143] = {.lex_state = 57, .external_lex_state = 20},
  [2144] = {.lex_state = 57, .external_lex_state = 20},
  [2145] = {.lex_state = 3, .external_lex_state = 22},
  [2146] = {.lex_state = 3, .external_lex_state = 22},
  [2147] = {.lex_state = 113, .external_lex_state = 21},
  [2148] = {.lex_state = 57, .external_lex_state = 20},
  [2149] = {.lex_state = 57, .external_lex_state = 20},
  [2150] = {.lex_state = 113, .external_lex_state = 22},
  [2151] = {.lex_state = 57, .external_lex_state = 20},
  [2152] = {.lex_state = 3, .external_lex_state = 21},
  [2153] = {.lex_state = 3, .external_lex_state = 22},
  [2154] = {.lex_state = 57, .external_lex_state = 20},
  [2155] = {.lex_state = 65, .external_lex_state = 28},
  [2156] = {.lex_state = 3, .external_lex_state = 21},
  [2157] = {.lex_state = 113, .external_lex_state = 22},
  [2158] = {.lex_state = 57, .external_lex_state = 20},
  [2159] = {.lex_state = 57, .external_lex_state = 20},
  [2160] = {.lex_state = 57, .external_lex_state = 20},
  [2161] = {.lex_state = 3, .external_lex_state = 20},
  [2162] = {.lex_state = 3, .external_lex_state = 20},
  [2163] = {.lex_state = 113, .external_lex_state = 23},
  [2164] = {.lex_state = 113, .external_lex_state = 23},
  [2165] = {.lex_state = 113, .external_lex_state = 23},
  [2166] = {.lex_state = 56, .external_lex_state = 24},
  [2167] = {.lex_state = 56, .external_lex_state = 24},
  [2168] = {.lex_state = 3, .external_lex_state = 21},
  [2169] = {.lex_state = 65, .external_lex_state = 28},
  [2170] = {.lex_state = 3, .external_lex_state = 20},
  [2171] = {.lex_state = 57, .external_lex_state = 20},
  [2172] = {.lex_state = 57, .external_lex_state = 20},
  [2173] = {.lex_state = 57, .external_lex_state = 20},
  [2174] = {.lex_state = 57, .external_lex_state = 20},
  [2175] = {.lex_state = 57, .external_lex_state = 20},
  [2176] = {.lex_state = 57, .external_lex_state = 20},
  [2177] = {.lex_state = 3, .external_lex_state = 22},
  [2178] = {.lex_state = 3, .external_lex_state = 22},
  [2179] = {.lex_state = 3, .external_lex_state = 22},
  [2180] = {.lex_state = 57, .external_lex_state = 20},
  [2181] = {.lex_state = 57, .external_lex_state = 20},
  [2182] = {.lex_state = 65, .external_lex_state = 28},
  [2183] = {.lex_state = 3, .external_lex_state = 21},
  [2184] = {.lex_state = 57, .external_lex_state = 20},
  [2185] = {.lex_state = 57, .external_lex_state = 20},
  [2186] = {.lex_state = 57, .external_lex_state = 20},
  [2187] = {.lex_state = 57, .external_lex_state = 20},
  [2188] = {.lex_state = 57, .external_lex_state = 20},
  [2189] = {.lex_state = 57, .external_lex_state = 20},
  [2190] = {.lex_state = 3, .external_lex_state = 22},
  [2191] = {.lex_state = 3, .external_lex_state = 22},
  [2192] = {.lex_state = 57, .external_lex_state = 20},
  [2193] = {.lex_state = 9, .external_lex_state = 20},
  [2194] = {.lex_state = 3, .external_lex_state = 23},
  [2195] = {.lex_state = 8, .external_lex_state = 20},
  [2196] = {.lex_state = 10, .external_lex_state = 20},
  [2197] = {.lex_state = 9, .external_lex_state = 20},
  [2198] = {.lex_state = 65, .external_lex_state = 28},
  [2199] = {.lex_state = 3, .external_lex_state = 23},
  [2200] = {.lex_state = 9, .external_lex_state = 20},
  [2201] = {.lex_state = 57, .external_lex_state = 20},
  [2202] = {.lex_state = 8, .external_lex_state = 20},
  [2203] = {.lex_state = 65, .external_lex_state = 20},
  [2204] = {.lex_state = 65, .external_lex_state = 20},
  [2205] = {.lex_state = 6, .external_lex_state = 20},
  [2206] = {.lex_state = 3, .external_lex_state = 20},
  [2207] = {.lex_state = 65, .external_lex_state = 28},
  [2208] = {.lex_state = 3, .external_lex_state = 20},
  [2209] = {.lex_state = 3, .external_lex_state = 20},
  [2210] = {.lex_state = 3, .external_lex_state = 20},
  [2211] = {.lex_state = 9, .external_lex_state = 20},
  [2212] = {.lex_state = 3, .external_lex_state = 21},
  [2213] = {.lex_state = 10, .external_lex_state = 20},
  [2214] = {.lex_state = 3, .external_lex_state = 20},
  [2215] = {.lex_state = 3, .external_lex_state = 20},
  [2216] = {.lex_state = 9, .external_lex_state = 20},
  [2217] = {.lex_state = 3, .external_lex_state = 20},
  [2218] = {.lex_state = 3, .external_lex_state = 20},
  [2219] = {.lex_state = 3, .external_lex_state = 20},
  [2220] = {.lex_state = 3, .external_lex_state = 20},
  [2221] = {.lex_state = 3, .external_lex_state = 20},
  [2222] = {.lex_state = 8, .external_lex_state = 20},
  [2223] = {.lex_state = 3, .external_lex_state = 20},
  [2224] = {.lex_state = 6, .external_lex_state = 20},
  [2225] = {.lex_state = 6, .external_lex_state = 21},
  [2226] = {.lex_state = 3, .external_lex_state = 20},
  [2227] = {.lex_state = 10, .external_lex_state = 20},
  [2228] = {.lex_state = 3, .external_lex_state = 20},
  [2229] = {.lex_state = 6, .external_lex_state = 23},
  [2230] = {.lex_state = 6, .external_lex_state = 20},
  [2231] = {.lex_state = 8, .external_lex_state = 20},
  [2232] = {.lex_state = 3, .external_lex_state = 20},
  [2233] = {.lex_state = 3, .external_lex_state = 20},
  [2234] = {.lex_state = 3, .external_lex_state = 21},
  [2235] = {.lex_state = 6, .external_lex_state = 22},
  [2236] = {.lex_state = 6, .external_lex_state = 20},
  [2237] = {.lex_state = 57, .external_lex_state = 20},
  [2238] = {.lex_state = 3, .external_lex_state = 20},
  [2239] = {.lex_state = 3, .external_lex_state = 20},
  [2240] = {.lex_state = 3, .external_lex_state = 21},
  [2241] = {.lex_state = 57, .external_lex_state = 20},
  [2242] = {.lex_state = 8, .external_lex_state = 20},
  [2243] = {.lex_state = 59, .external_lex_state = 20},
  [2244] = {.lex_state = 3, .external_lex_state = 22},
  [2245] = {.lex_state = 3, .external_lex_state = 20},
  [2246] = {.lex_state = 3, .external_lex_state = 20},
  [2247] = {.lex_state = 3, .external_lex_state = 20},
  [2248] = {.lex_state = 3, .external_lex_state = 20},
  [2249] = {.lex_state = 3, .external_lex_state = 20},
  [2250] = {.lex_state = 3, .external_lex_state = 20},
  [2251] = {.lex_state = 56, .external_lex_state = 20},
  [2252] = {.lex_state = 56, .external_lex_state = 20},
  [2253] = {.lex_state = 3, .external_lex_state = 22},
  [2254] = {.lex_state = 3, .external_lex_state = 20},
  [2255] = {.lex_state = 3, .external_lex_state = 20},
  [2256] = {.lex_state = 3, .external_lex_state = 20},
  [2257] = {.lex_state = 9, .external_lex_state = 20},
  [2258] = {.lex_state = 56, .external_lex_state = 20},
  [2259] = {.lex_state = 3, .external_lex_state = 23},
  [2260] = {.lex_state = 3, .external_lex_state = 20},
  [2261] = {.lex_state = 3, .external_lex_state = 22},
  [2262] = {.lex_state = 3, .external_lex_state = 20},
  [2263] = {.lex_state = 3, .external_lex_state = 20},
  [2264] = {.lex_state = 3, .external_lex_state = 23},
  [2265] = {.lex_state = 3, .external_lex_state = 22},
  [2266] = {.lex_state = 3, .external_lex_state = 23},
  [2267] = {.lex_state = 64, .external_lex_state = 20},
  [2268] = {.lex_state = 3, .external_lex_state = 22},
  [2269] = {.lex_state = 10, .external_lex_state = 23},
  [2270] = {.lex_state = 3, .external_lex_state = 23},
  [2271] = {.lex_state = 9, .external_lex_state = 23},
  [2272] = {.lex_state = 3, .external_lex_state = 20},
  [2273] = {.lex_state = 3, .external_lex_state = 23},
  [2274] = {.lex_state = 3, .external_lex_state = 23},
  [2275] = {.lex_state = 3, .external_lex_state = 23},
  [2276] = {.lex_state = 3, .external_lex_state = 20},
  [2277] = {.lex_state = 3, .external_lex_state = 22},
  [2278] = {.lex_state = 10, .external_lex_state = 22},
  [2279] = {.lex_state = 9, .external_lex_state = 23},
  [2280] = {.lex_state = 3, .external_lex_state = 23},
  [2281] = {.lex_state = 3, .external_lex_state = 22},
  [2282] = {.lex_state = 3, .external_lex_state = 21},
  [2283] = {.lex_state = 3, .external_lex_state = 21},
  [2284] = {.lex_state = 9, .external_lex_state = 21},
  [2285] = {.lex_state = 3, .external_lex_state = 21},
  [2286] = {.lex_state = 3, .external_lex_state = 23},
  [2287] = {.lex_state = 3, .external_lex_state = 23},
  [2288] = {.lex_state = 3, .external_lex_state = 23},
  [2289] = {.lex_state = 10, .external_lex_state = 23},
  [2290] = {.lex_state = 9, .external_lex_state = 23},
  [2291] = {.lex_state = 3, .external_lex_state = 21},
  [2292] = {.lex_state = 3, .external_lex_state = 21},
  [2293] = {.lex_state = 3, .external_lex_state = 21},
  [2294] = {.lex_state = 3, .external_lex_state = 21},
  [2295] = {.lex_state = 3, .external_lex_state = 21},
  [2296] = {.lex_state = 3, .external_lex_state = 23},
  [2297] = {.lex_state = 3, .external_lex_state = 20},
  [2298] = {.lex_state = 3, .external_lex_state = 20},
  [2299] = {.lex_state = 3, .external_lex_state = 23},
  [2300] = {.lex_state = 3, .external_lex_state = 21},
  [2301] = {.lex_state = 8, .external_lex_state = 23},
  [2302] = {.lex_state = 6, .external_lex_state = 23},
  [2303] = {.lex_state = 3, .external_lex_state = 20},
  [2304] = {.lex_state = 3, .external_lex_state = 23},
  [2305] = {.lex_state = 3, .external_lex_state = 23},
  [2306] = {.lex_state = 3, .external_lex_state = 21},
  [2307] = {.lex_state = 10, .external_lex_state = 21},
  [2308] = {.lex_state = 9, .external_lex_state = 21},
  [2309] = {.lex_state = 3, .external_lex_state = 20},
  [2310] = {.lex_state = 3, .external_lex_state = 20},
  [2311] = {.lex_state = 3, .external_lex_state = 21},
  [2312] = {.lex_state = 3, .external_lex_state = 21},
  [2313] = {.lex_state = 3, .external_lex_state = 20},
  [2314] = {.lex_state = 9, .external_lex_state = 21},
  [2315] = {.lex_state = 3, .external_lex_state = 21},
  [2316] = {.lex_state = 3, .external_lex_state = 20},
  [2317] = {.lex_state = 57, .external_lex_state = 20},
  [2318] = {.lex_state = 3, .external_lex_state = 20},
  [2319] = {.lex_state = 9, .external_lex_state = 22},
  [2320] = {.lex_state = 3, .external_lex_state = 21},
  [2321] = {.lex_state = 3, .external_lex_state = 23},
  [2322] = {.lex_state = 57, .external_lex_state = 20},
  [2323] = {.lex_state = 3, .external_lex_state = 22},
  [2324] = {.lex_state = 57, .external_lex_state = 20},
  [2325] = {.lex_state = 3, .external_lex_state = 21},
  [2326] = {.lex_state = 64, .external_lex_state = 28},
  [2327] = {.lex_state = 3, .external_lex_state = 20},
  [2328] = {.lex_state = 3, .external_lex_state = 20},
  [2329] = {.lex_state = 64, .external_lex_state = 20},
  [2330] = {.lex_state = 3, .external_lex_state = 20},
  [2331] = {.lex_state = 3, .external_lex_state = 22},
  [2332] = {.lex_state = 3, .external_lex_state = 21},
  [2333] = {.lex_state = 10, .external_lex_state = 21},
  [2334] = {.lex_state = 64, .external_lex_state = 28},
  [2335] = {.lex_state = 64, .external_lex_state = 20},
  [2336] = {.lex_state = 9, .external_lex_state = 21},
  [2337] = {.lex_state = 8, .external_lex_state = 21},
  [2338] = {.lex_state = 6, .external_lex_state = 21},
  [2339] = {.lex_state = 3, .external_lex_state = 20},
  [2340] = {.lex_state = 3, .external_lex_state = 20},
  [2341] = {.lex_state = 3, .external_lex_state = 20},
  [2342] = {.lex_state = 3, .external_lex_state = 22},
  [2343] = {.lex_state = 3, .external_lex_state = 20},
  [2344] = {.lex_state = 57, .external_lex_state = 20},
  [2345] = {.lex_state = 3, .external_lex_state = 20},
  [2346] = {.lex_state = 3, .external_lex_state = 21},
  [2347] = {.lex_state = 3, .external_lex_state = 20},
  [2348] = {.lex_state = 3, .external_lex_state = 20},
  [2349] = {.lex_state = 3, .external_lex_state = 20},
  [2350] = {.lex_state = 3, .external_lex_state = 20},
  [2351] = {.lex_state = 3, .external_lex_state = 20},
  [2352] = {.lex_state = 3, .external_lex_state = 20},
  [2353] = {.lex_state = 3, .external_lex_state = 20},
  [2354] = {.lex_state = 8, .external_lex_state = 23},
  [2355] = {.lex_state = 57, .external_lex_state = 20},
  [2356] = {.lex_state = 3, .external_lex_state = 20},
  [2357] = {.lex_state = 3, .external_lex_state = 22},
  [2358] = {.lex_state = 3, .external_lex_state = 22},
  [2359] = {.lex_state = 8, .external_lex_state = 23},
  [2360] = {.lex_state = 6, .external_lex_state = 23},
  [2361] = {.lex_state = 3, .external_lex_state = 21},
  [2362] = {.lex_state = 3, .external_lex_state = 20},
  [2363] = {.lex_state = 6, .external_lex_state = 23},
  [2364] = {.lex_state = 3, .external_lex_state = 20},
  [2365] = {.lex_state = 3, .external_lex_state = 20},
  [2366] = {.lex_state = 3, .external_lex_state = 20},
  [2367] = {.lex_state = 9, .external_lex_state = 22},
  [2368] = {.lex_state = 3, .external_lex_state = 22},
  [2369] = {.lex_state = 3, .external_lex_state = 22},
  [2370] = {.lex_state = 3, .external_lex_state = 20},
  [2371] = {.lex_state = 3, .external_lex_state = 20},
  [2372] = {.lex_state = 8, .external_lex_state = 21},
  [2373] = {.lex_state = 8, .external_lex_state = 21},
  [2374] = {.lex_state = 6, .external_lex_state = 21},
  [2375] = {.lex_state = 6, .external_lex_state = 21},
  [2376] = {.lex_state = 3, .external_lex_state = 20},
  [2377] = {.lex_state = 3, .external_lex_state = 20},
  [2378] = {.lex_state = 3, .external_lex_state = 20},
  [2379] = {.lex_state = 3, .external_lex_state = 20},
  [2380] = {.lex_state = 3, .external_lex_state = 20},
  [2381] = {.lex_state = 9, .external_lex_state = 22},
  [2382] = {.lex_state = 3, .external_lex_state = 22},
  [2383] = {.lex_state = 3, .external_lex_state = 20},
  [2384] = {.lex_state = 8, .external_lex_state = 23},
  [2385] = {.lex_state = 6, .external_lex_state = 23},
  [2386] = {.lex_state = 3, .external_lex_state = 22},
  [2387] = {.lex_state = 3, .external_lex_state = 20},
  [2388] = {.lex_state = 9, .external_lex_state = 23},
  [2389] = {.lex_state = 57, .external_lex_state = 20},
  [2390] = {.lex_state = 9, .external_lex_state = 23},
  [2391] = {.lex_state = 63, .external_lex_state = 20},
  [2392] = {.lex_state = 3, .external_lex_state = 23},
  [2393] = {.lex_state = 3, .external_lex_state = 22},
  [2394] = {.lex_state = 3, .external_lex_state = 22},
  [2395] = {.lex_state = 3, .external_lex_state = 20},
  [2396] = {.lex_state = 3, .external_lex_state = 21},
  [2397] = {.lex_state = 3, .external_lex_state = 20},
  [2398] = {.lex_state = 3, .external_lex_state = 20},
  [2399] = {.lex_state = 8, .external_lex_state = 21},
  [2400] = {.lex_state = 6, .external_lex_state = 21},
  [2401] = {.lex_state = 3, .external_lex_state = 20},
  [2402] = {.lex_state = 3, .external_lex_state = 20},
  [2403] = {.lex_state = 3, .external_lex_state = 20},
  [2404] = {.lex_state = 3, .external_lex_state = 20},
  [2405] = {.lex_state = 3, .external_lex_state = 20},
  [2406] = {.lex_state = 3, .external_lex_state = 20},
  [2407] = {.lex_state = 3, .external_lex_state = 20},
  [2408] = {.lex_state = 8, .external_lex_state = 21},
  [2409] = {.lex_state = 3, .external_lex_state = 20},
  [2410] = {.lex_state = 3, .external_lex_state = 20},
  [2411] = {.lex_state = 3, .external_lex_state = 20},
  [2412] = {.lex_state = 3, .external_lex_state = 20},
  [2413] = {.lex_state = 3, .external_lex_state = 20},
  [2414] = {.lex_state = 3, .external_lex_state = 20},
  [2415] = {.lex_state = 3, .external_lex_state = 20},
  [2416] = {.lex_state = 9, .external_lex_state = 22},
  [2417] = {.lex_state = 3, .external_lex_state = 20},
  [2418] = {.lex_state = 3, .external_lex_state = 21},
  [2419] = {.lex_state = 57, .external_lex_state = 20},
  [2420] = {.lex_state = 57, .external_lex_state = 20},
  [2421] = {.lex_state = 57, .external_lex_state = 20},
  [2422] = {.lex_state = 3, .external_lex_state = 20},
  [2423] = {.lex_state = 3, .external_lex_state = 20},
  [2424] = {.lex_state = 64, .external_lex_state = 20},
  [2425] = {.lex_state = 9, .external_lex_state = 22},
  [2426] = {.lex_state = 3, .external_lex_state = 22},
  [2427] = {.lex_state = 3, .external_lex_state = 20},
  [2428] = {.lex_state = 10, .external_lex_state = 22},
  [2429] = {.lex_state = 57, .external_lex_state = 20},
  [2430] = {.lex_state = 3, .external_lex_state = 20},
  [2431] = {.lex_state = 57, .external_lex_state = 20},
  [2432] = {.lex_state = 3, .external_lex_state = 20},
  [2433] = {.lex_state = 3, .external_lex_state = 22},
  [2434] = {.lex_state = 3, .external_lex_state = 22},
  [2435] = {.lex_state = 10, .external_lex_state = 23},
  [2436] = {.lex_state = 10, .external_lex_state = 22},
  [2437] = {.lex_state = 9, .external_lex_state = 22},
  [2438] = {.lex_state = 3, .external_lex_state = 20},
  [2439] = {.lex_state = 3, .external_lex_state = 20},
  [2440] = {.lex_state = 8, .external_lex_state = 22},
  [2441] = {.lex_state = 6, .external_lex_state = 22},
  [2442] = {.lex_state = 3, .external_lex_state = 20},
  [2443] = {.lex_state = 3, .external_lex_state = 20},
  [2444] = {.lex_state = 64, .external_lex_state = 20},
  [2445] = {.lex_state = 64, .external_lex_state = 20},
  [2446] = {.lex_state = 3, .external_lex_state = 22},
  [2447] = {.lex_state = 3, .external_lex_state = 20},
  [2448] = {.lex_state = 3, .external_lex_state = 22},
  [2449] = {.lex_state = 3, .external_lex_state = 20},
  [2450] = {.lex_state = 3, .external_lex_state = 22},
  [2451] = {.lex_state = 63, .external_lex_state = 20},
  [2452] = {.lex_state = 3, .external_lex_state = 20},
  [2453] = {.lex_state = 3, .external_lex_state = 22},
  [2454] = {.lex_state = 3, .external_lex_state = 22},
  [2455] = {.lex_state = 3, .external_lex_state = 23},
  [2456] = {.lex_state = 3, .external_lex_state = 20},
  [2457] = {.lex_state = 3, .external_lex_state = 23},
  [2458] = {.lex_state = 3, .external_lex_state = 23},
  [2459] = {.lex_state = 3, .external_lex_state = 23},
  [2460] = {.lex_state = 10, .external_lex_state = 21},
  [2461] = {.lex_state = 3, .external_lex_state = 22},
  [2462] = {.lex_state = 3, .external_lex_state = 23},
  [2463] = {.lex_state = 3, .external_lex_state = 20},
  [2464] = {.lex_state = 8, .external_lex_state = 22},
  [2465] = {.lex_state = 3, .external_lex_state = 20},
  [2466] = {.lex_state = 3, .external_lex_state = 20},
  [2467] = {.lex_state = 3, .external_lex_state = 20},
  [2468] = {.lex_state = 3, .external_lex_state = 20},
  [2469] = {.lex_state = 8, .external_lex_state = 23},
  [2470] = {.lex_state = 3, .external_lex_state = 22},
  [2471] = {.lex_state = 3, .external_lex_state = 23},
  [2472] = {.lex_state = 60, .external_lex_state = 20},
  [2473] = {.lex_state = 62, .external_lex_state = 20},
  [2474] = {.lex_state = 57, .external_lex_state = 20},
  [2475] = {.lex_state = 57, .external_lex_state = 20},
  [2476] = {.lex_state = 3, .external_lex_state = 20},
  [2477] = {.lex_state = 9, .external_lex_state = 21},
  [2478] = {.lex_state = 63, .external_lex_state = 20},
  [2479] = {.lex_state = 64, .external_lex_state = 20},
  [2480] = {.lex_state = 57, .external_lex_state = 20},
  [2481] = {.lex_state = 65, .external_lex_state = 29},
  [2482] = {.lex_state = 3, .external_lex_state = 22},
  [2483] = {.lex_state = 65, .external_lex_state = 29},
  [2484] = {.lex_state = 3, .external_lex_state = 22},
  [2485] = {.lex_state = 3, .external_lex_state = 20},
  [2486] = {.lex_state = 3, .external_lex_state = 20},
  [2487] = {.lex_state = 64, .external_lex_state = 20},
  [2488] = {.lex_state = 3, .external_lex_state = 20},
  [2489] = {.lex_state = 3, .external_lex_state = 20},
  [2490] = {.lex_state = 57, .external_lex_state = 20},
  [2491] = {.lex_state = 64, .external_lex_state = 20},
  [2492] = {.lex_state = 8, .external_lex_state = 22},
  [2493] = {.lex_state = 3, .external_lex_state = 20},
  [2494] = {.lex_state = 3, .external_lex_state = 21},
  [2495] = {.lex_state = 57, .external_lex_state = 20},
  [2496] = {.lex_state = 64, .external_lex_state = 20},
  [2497] = {.lex_state = 64, .external_lex_state = 20},
  [2498] = {.lex_state = 64, .external_lex_state = 20},
  [2499] = {.lex_state = 64, .external_lex_state = 20},
  [2500] = {.lex_state = 64, .external_lex_state = 20},
  [2501] = {.lex_state = 60, .external_lex_state = 20},
  [2502] = {.lex_state = 64, .external_lex_state = 20},
  [2503] = {.lex_state = 63, .external_lex_state = 20},
  [2504] = {.lex_state = 9, .external_lex_state = 21},
  [2505] = {.lex_state = 57, .external_lex_state = 20},
  [2506] = {.lex_state = 64, .external_lex_state = 20},
  [2507] = {.lex_state = 64, .external_lex_state = 20},
  [2508] = {.lex_state = 57, .external_lex_state = 20},
  [2509] = {.lex_state = 8, .external_lex_state = 22},
  [2510] = {.lex_state = 64, .external_lex_state = 20},
  [2511] = {.lex_state = 64, .external_lex_state = 20},
  [2512] = {.lex_state = 64, .external_lex_state = 20},
  [2513] = {.lex_state = 64, .external_lex_state = 20},
  [2514] = {.lex_state = 6, .external_lex_state = 22},
  [2515] = {.lex_state = 3, .external_lex_state = 22},
  [2516] = {.lex_state = 64, .external_lex_state = 20},
  [2517] = {.lex_state = 64, .external_lex_state = 20},
  [2518] = {.lex_state = 3, .external_lex_state = 23},
  [2519] = {.lex_state = 3, .external_lex_state = 23},
  [2520] = {.lex_state = 3, .external_lex_state = 23},
  [2521] = {.lex_state = 3, .external_lex_state = 20},
  [2522] = {.lex_state = 3, .external_lex_state = 20},
  [2523] = {.lex_state = 3, .external_lex_state = 20},
  [2524] = {.lex_state = 3, .external_lex_state = 20},
  [2525] = {.lex_state = 65, .external_lex_state = 20},
  [2526] = {.lex_state = 3, .external_lex_state = 20},
  [2527] = {.lex_state = 9, .external_lex_state = 23},
  [2528] = {.lex_state = 3, .external_lex_state = 20},
  [2529] = {.lex_state = 3, .external_lex_state = 20},
  [2530] = {.lex_state = 3, .external_lex_state = 20},
  [2531] = {.lex_state = 65, .external_lex_state = 20},
  [2532] = {.lex_state = 3, .external_lex_state = 20},
  [2533] = {.lex_state = 3, .external_lex_state = 20},
  [2534] = {.lex_state = 3, .external_lex_state = 20},
  [2535] = {.lex_state = 3, .external_lex_state = 20},
  [2536] = {.lex_state = 3, .external_lex_state = 20},
  [2537] = {.lex_state = 3, .external_lex_state = 23},
  [2538] = {.lex_state = 63, .external_lex_state = 20},
  [2539] = {.lex_state = 57, .external_lex_state = 20},
  [2540] = {.lex_state = 3, .external_lex_state = 21},
  [2541] = {.lex_state = 3, .external_lex_state = 20},
  [2542] = {.lex_state = 3, .external_lex_state = 20},
  [2543] = {.lex_state = 3, .external_lex_state = 20},
  [2544] = {.lex_state = 3, .external_lex_state = 20},
  [2545] = {.lex_state = 3, .external_lex_state = 20},
  [2546] = {.lex_state = 3, .external_lex_state = 20},
  [2547] = {.lex_state = 3, .external_lex_state = 20},
  [2548] = {.lex_state = 3, .external_lex_state = 21},
  [2549] = {.lex_state = 3, .external_lex_state = 20},
  [2550] = {.lex_state = 3, .external_lex_state = 20},
  [2551] = {.lex_state = 3, .external_lex_state = 20},
  [2552] = {.lex_state = 3, .external_lex_state = 20},
  [2553] = {.lex_state = 3, .external_lex_state = 20},
  [2554] = {.lex_state = 3, .external_lex_state = 20},
  [2555] = {.lex_state = 3, .external_lex_state = 20},
  [2556] = {.lex_state = 3, .external_lex_state = 20},
  [2557] = {.lex_state = 3, .external_lex_state = 20},
  [2558] = {.lex_state = 3, .external_lex_state = 20},
  [2559] = {.lex_state = 57, .external_lex_state = 20},
  [2560] = {.lex_state = 60, .external_lex_state = 20},
  [2561] = {.lex_state = 63, .external_lex_state = 20},
  [2562] = {.lex_state = 3, .external_lex_state = 20},
  [2563] = {.lex_state = 62, .external_lex_state = 20},
  [2564] = {.lex_state = 59, .external_lex_state = 20},
  [2565] = {.lex_state = 3, .external_lex_state = 20},
  [2566] = {.lex_state = 3, .external_lex_state = 20},
  [2567] = {.lex_state = 3, .external_lex_state = 20},
  [2568] = {.lex_state = 3, .external_lex_state = 20},
  [2569] = {.lex_state = 3, .external_lex_state = 20},
  [2570] = {.lex_state = 3, .external_lex_state = 21},
  [2571] = {.lex_state = 3, .external_lex_state = 20},
  [2572] = {.lex_state = 3, .external_lex_state = 21},
  [2573] = {.lex_state = 3, .external_lex_state = 23},
  [2574] = {.lex_state = 65, .external_lex_state = 24},
  [2575] = {.lex_state = 3, .external_lex_state = 20},
  [2576] = {.lex_state = 3, .external_lex_state = 20},
  [2577] = {.lex_state = 6, .external_lex_state = 22},
  [2578] = {.lex_state = 3, .external_lex_state = 21},
  [2579] = {.lex_state = 57, .external_lex_state = 20},
  [2580] = {.lex_state = 3, .external_lex_state = 20},
  [2581] = {.lex_state = 65, .external_lex_state = 24},
  [2582] = {.lex_state = 3, .external_lex_state = 21},
  [2583] = {.lex_state = 3, .external_lex_state = 21},
  [2584] = {.lex_state = 3, .external_lex_state = 21},
  [2585] = {.lex_state = 3, .external_lex_state = 20},
  [2586] = {.lex_state = 3, .external_lex_state = 20},
  [2587] = {.lex_state = 3, .external_lex_state = 20},
  [2588] = {.lex_state = 62, .external_lex_state = 20},
  [2589] = {.lex_state = 62, .external_lex_state = 20},
  [2590] = {.lex_state = 59, .external_lex_state = 20},
  [2591] = {.lex_state = 59, .external_lex_state = 20},
  [2592] = {.lex_state = 3, .external_lex_state = 21},
  [2593] = {.lex_state = 3, .external_lex_state = 20},
  [2594] = {.lex_state = 3, .external_lex_state = 20},
  [2595] = {.lex_state = 3, .external_lex_state = 20},
  [2596] = {.lex_state = 3, .external_lex_state = 20},
  [2597] = {.lex_state = 3, .external_lex_state = 23},
  [2598] = {.lex_state = 64, .external_lex_state = 20},
  [2599] = {.lex_state = 3, .external_lex_state = 22},
  [2600] = {.lex_state = 3, .external_lex_state = 20},
  [2601] = {.lex_state = 3, .external_lex_state = 23},
  [2602] = {.lex_state = 3, .external_lex_state = 20},
  [2603] = {.lex_state = 62, .external_lex_state = 20},
  [2604] = {.lex_state = 3, .external_lex_state = 20},
  [2605] = {.lex_state = 3, .external_lex_state = 22},
  [2606] = {.lex_state = 59, .external_lex_state = 20},
  [2607] = {.lex_state = 3, .external_lex_state = 20},
  [2608] = {.lex_state = 3, .external_lex_state = 20},
  [2609] = {.lex_state = 3, .external_lex_state = 20},
  [2610] = {.lex_state = 3, .external_lex_state = 20},
  [2611] = {.lex_state = 3, .external_lex_state = 20},
  [2612] = {.lex_state = 3, .external_lex_state = 20},
  [2613] = {.lex_state = 3, .external_lex_state = 23},
  [2614] = {.lex_state = 3, .external_lex_state = 20},
  [2615] = {.lex_state = 3, .external_lex_state = 20},
  [2616] = {.lex_state = 3, .external_lex_state = 20},
  [2617] = {.lex_state = 3, .external_lex_state = 20},
  [2618] = {.lex_state = 8, .external_lex_state = 22},
  [2619] = {.lex_state = 6, .external_lex_state = 22},
  [2620] = {.lex_state = 3, .external_lex_state = 20},
  [2621] = {.lex_state = 64, .external_lex_state = 20},
  [2622] = {.lex_state = 3, .external_lex_state = 20},
  [2623] = {.lex_state = 57, .external_lex_state = 20},
  [2624] = {.lex_state = 3, .external_lex_state = 22},
  [2625] = {.lex_state = 3, .external_lex_state = 22},
  [2626] = {.lex_state = 3, .external_lex_state = 22},
  [2627] = {.lex_state = 3, .external_lex_state = 22},
  [2628] = {.lex_state = 3, .external_lex_state = 22},
  [2629] = {.lex_state = 3, .external_lex_state = 22},
  [2630] = {.lex_state = 3, .external_lex_state = 22},
  [2631] = {.lex_state = 3, .external_lex_state = 22},
  [2632] = {.lex_state = 3, .external_lex_state = 22},
  [2633] = {.lex_state = 3, .external_lex_state = 22},
  [2634] = {.lex_state = 3, .external_lex_state = 22},
  [2635] = {.lex_state = 3, .external_lex_state = 22},
  [2636] = {.lex_state = 3, .external_lex_state = 22},
  [2637] = {.lex_state = 3, .external_lex_state = 22},
  [2638] = {.lex_state = 3, .external_lex_state = 22},
  [2639] = {.lex_state = 3, .external_lex_state = 22},
  [2640] = {.lex_state = 3, .external_lex_state = 22},
  [2641] = {.lex_state = 3, .external_lex_state = 22},
  [2642] = {.lex_state = 3, .external_lex_state = 22},
  [2643] = {.lex_state = 3, .external_lex_state = 22},
  [2644] = {.lex_state = 3, .external_lex_state = 22},
  [2645] = {.lex_state = 3, .external_lex_state = 22},
  [2646] = {.lex_state = 3, .external_lex_state = 22},
  [2647] = {.lex_state = 3, .external_lex_state = 22},
  [2648] = {.lex_state = 3, .external_lex_state = 23},
  [2649] = {.lex_state = 3, .external_lex_state = 23},
  [2650] = {.lex_state = 3, .external_lex_state = 23},
  [2651] = {.lex_state = 3, .external_lex_state = 23},
  [2652] = {.lex_state = 3, .external_lex_state = 23},
  [2653] = {.lex_state = 3, .external_lex_state = 23},
  [2654] = {.lex_state = 3, .external_lex_state = 23},
  [2655] = {.lex_state = 3, .external_lex_state = 23},
  [2656] = {.lex_state = 3, .external_lex_state = 23},
  [2657] = {.lex_state = 3, .external_lex_state = 23},
  [2658] = {.lex_state = 3, .external_lex_state = 23},
  [2659] = {.lex_state = 3, .external_lex_state = 23},
  [2660] = {.lex_state = 3, .external_lex_state = 23},
  [2661] = {.lex_state = 3, .external_lex_state = 23},
  [2662] = {.lex_state = 3, .external_lex_state = 21},
  [2663] = {.lex_state = 3, .external_lex_state = 23},
  [2664] = {.lex_state = 3, .external_lex_state = 23},
  [2665] = {.lex_state = 3, .external_lex_state = 23},
  [2666] = {.lex_state = 3, .external_lex_state = 23},
  [2667] = {.lex_state = 3, .external_lex_state = 23},
  [2668] = {.lex_state = 3, .external_lex_state = 23},
  [2669] = {.lex_state = 3, .external_lex_state = 23},
  [2670] = {.lex_state = 3, .external_lex_state = 23},
  [2671] = {.lex_state = 3, .external_lex_state = 23},
  [2672] = {.lex_state = 3, .external_lex_state = 23},
  [2673] = {.lex_state = 3, .external_lex_state = 23},
  [2674] = {.lex_state = 3, .external_lex_state = 23},
  [2675] = {.lex_state = 3, .external_lex_state = 23},
  [2676] = {.lex_state = 3, .external_lex_state = 23},
  [2677] = {.lex_state = 3, .external_lex_state = 23},
  [2678] = {.lex_state = 3, .external_lex_state = 22},
  [2679] = {.lex_state = 3, .external_lex_state = 22},
  [2680] = {.lex_state = 3, .external_lex_state = 23},
  [2681] = {.lex_state = 3, .external_lex_state = 23},
  [2682] = {.lex_state = 3, .external_lex_state = 23},
  [2683] = {.lex_state = 3, .external_lex_state = 23},
  [2684] = {.lex_state = 3, .external_lex_state = 23},
  [2685] = {.lex_state = 3, .external_lex_state = 23},
  [2686] = {.lex_state = 3, .external_lex_state = 23},
  [2687] = {.lex_state = 3, .external_lex_state = 22},
  [2688] = {.lex_state = 3, .external_lex_state = 23},
  [2689] = {.lex_state = 3, .external_lex_state = 21},
  [2690] = {.lex_state = 3, .external_lex_state = 21},
  [2691] = {.lex_state = 3, .external_lex_state = 21},
  [2692] = {.lex_state = 3, .external_lex_state = 21},
  [2693] = {.lex_state = 3, .external_lex_state = 21},
  [2694] = {.lex_state = 3, .external_lex_state = 21},
  [2695] = {.lex_state = 3, .external_lex_state = 23},
  [2696] = {.lex_state = 3, .external_lex_state = 21},
  [2697] = {.lex_state = 3, .external_lex_state = 23},
  [2698] = {.lex_state = 3, .external_lex_state = 21},
  [2699] = {.lex_state = 3, .external_lex_state = 23},
  [2700] = {.lex_state = 3, .external_lex_state = 21},
  [2701] = {.lex_state = 3, .external_lex_state = 23},
  [2702] = {.lex_state = 3, .external_lex_state = 23},
  [2703] = {.lex_state = 3, .external_lex_state = 23},
  [2704] = {.lex_state = 3, .external_lex_state = 21},
  [2705] = {.lex_state = 3, .external_lex_state = 23},
  [2706] = {.lex_state = 3, .external_lex_state = 23},
  [2707] = {.lex_state = 3, .external_lex_state = 23},
  [2708] = {.lex_state = 3, .external_lex_state = 23},
  [2709] = {.lex_state = 3, .external_lex_state = 23},
  [2710] = {.lex_state = 3, .external_lex_state = 23},
  [2711] = {.lex_state = 3, .external_lex_state = 23},
  [2712] = {.lex_state = 3, .external_lex_state = 23},
  [2713] = {.lex_state = 3, .external_lex_state = 21},
  [2714] = {.lex_state = 3, .external_lex_state = 23},
  [2715] = {.lex_state = 3, .external_lex_state = 23},
  [2716] = {.lex_state = 3, .external_lex_state = 23},
  [2717] = {.lex_state = 3, .external_lex_state = 21},
  [2718] = {.lex_state = 3, .external_lex_state = 23},
  [2719] = {.lex_state = 3, .external_lex_state = 21},
  [2720] = {.lex_state = 3, .external_lex_state = 23},
  [2721] = {.lex_state = 3, .external_lex_state = 21},
  [2722] = {.lex_state = 3, .external_lex_state = 22},
  [2723] = {.lex_state = 3, .external_lex_state = 22},
  [2724] = {.lex_state = 3, .external_lex_state = 22},
  [2725] = {.lex_state = 3, .external_lex_state = 22},
  [2726] = {.lex_state = 3, .external_lex_state = 22},
  [2727] = {.lex_state = 3, .external_lex_state = 22},
  [2728] = {.lex_state = 3, .external_lex_state = 22},
  [2729] = {.lex_state = 3, .external_lex_state = 22},
  [2730] = {.lex_state = 3, .external_lex_state = 22},
  [2731] = {.lex_state = 3, .external_lex_state = 22},
  [2732] = {.lex_state = 3, .external_lex_state = 22},
  [2733] = {.lex_state = 3, .external_lex_state = 22},
  [2734] = {.lex_state = 3, .external_lex_state = 22},
  [2735] = {.lex_state = 3, .external_lex_state = 22},
  [2736] = {.lex_state = 3, .external_lex_state = 22},
  [2737] = {.lex_state = 3, .external_lex_state = 22},
  [2738] = {.lex_state = 3, .external_lex_state = 23},
  [2739] = {.lex_state = 3, .external_lex_state = 23},
  [2740] = {.lex_state = 3, .external_lex_state = 23},
  [2741] = {.lex_state = 3, .external_lex_state = 23},
  [2742] = {.lex_state = 3, .external_lex_state = 23},
  [2743] = {.lex_state = 65, .external_lex_state = 20},
  [2744] = {.lex_state = 3, .external_lex_state = 23},
  [2745] = {.lex_state = 3, .external_lex_state = 23},
  [2746] = {.lex_state = 3, .external_lex_state = 23},
  [2747] = {.lex_state = 3, .external_lex_state = 23},
  [2748] = {.lex_state = 3, .external_lex_state = 23},
  [2749] = {.lex_state = 3, .external_lex_state = 23},
  [2750] = {.lex_state = 3, .external_lex_state = 23},
  [2751] = {.lex_state = 3, .external_lex_state = 23},
  [2752] = {.lex_state = 3, .external_lex_state = 23},
  [2753] = {.lex_state = 3, .external_lex_state = 23},
  [2754] = {.lex_state = 3, .external_lex_state = 23},
  [2755] = {.lex_state = 3, .external_lex_state = 23},
  [2756] = {.lex_state = 3, .external_lex_state = 21},
  [2757] = {.lex_state = 3, .external_lex_state = 21},
  [2758] = {.lex_state = 3, .external_lex_state = 23},
  [2759] = {.lex_state = 3, .external_lex_state = 21},
  [2760] = {.lex_state = 3, .external_lex_state = 21},
  [2761] = {.lex_state = 3, .external_lex_state = 21},
  [2762] = {.lex_state = 3, .external_lex_state = 23},
  [2763] = {.lex_state = 3, .external_lex_state = 23},
  [2764] = {.lex_state = 3, .external_lex_state = 23},
  [2765] = {.lex_state = 3, .external_lex_state = 23},
  [2766] = {.lex_state = 3, .external_lex_state = 21},
  [2767] = {.lex_state = 3, .external_lex_state = 23},
  [2768] = {.lex_state = 3, .external_lex_state = 23},
  [2769] = {.lex_state = 3, .external_lex_state = 23},
  [2770] = {.lex_state = 3, .external_lex_state = 23},
  [2771] = {.lex_state = 3, .external_lex_state = 23},
  [2772] = {.lex_state = 3, .external_lex_state = 21},
  [2773] = {.lex_state = 3, .external_lex_state = 23},
  [2774] = {.lex_state = 3, .external_lex_state = 21},
  [2775] = {.lex_state = 3, .external_lex_state = 21},
  [2776] = {.lex_state = 3, .external_lex_state = 23},
  [2777] = {.lex_state = 3, .external_lex_state = 21},
  [2778] = {.lex_state = 3, .external_lex_state = 23},
  [2779] = {.lex_state = 3, .external_lex_state = 21},
  [2780] = {.lex_state = 65, .external_lex_state = 20},
  [2781] = {.lex_state = 3, .external_lex_state = 21},
  [2782] = {.lex_state = 3, .external_lex_state = 21},
  [2783] = {.lex_state = 3, .external_lex_state = 21},
  [2784] = {.lex_state = 3, .external_lex_state = 23},
  [2785] = {.lex_state = 3, .external_lex_state = 23},
  [2786] = {.lex_state = 3, .external_lex_state = 21},
  [2787] = {.lex_state = 3, .external_lex_state = 23},
  [2788] = {.lex_state = 3, .external_lex_state = 23},
  [2789] = {.lex_state = 3, .external_lex_state = 23},
  [2790] = {.lex_state = 3, .external_lex_state = 23},
  [2791] = {.lex_state = 3, .external_lex_state = 21},
  [2792] = {.lex_state = 3, .external_lex_state = 23},
  [2793] = {.lex_state = 3, .external_lex_state = 21},
  [2794] = {.lex_state = 3, .external_lex_state = 21},
  [2795] = {.lex_state = 3, .external_lex_state = 23},
  [2796] = {.lex_state = 3, .external_lex_state = 23},
  [2797] = {.lex_state = 3, .external_lex_state = 22},
  [2798] = {.lex_state = 3, .external_lex_state = 23},
  [2799] = {.lex_state = 3, .external_lex_state = 23},
  [2800] = {.lex_state = 3, .external_lex_state = 23},
  [2801] = {.lex_state = 3, .external_lex_state = 23},
  [2802] = {.lex_state = 3, .external_lex_state = 23},
  [2803] = {.lex_state = 3, .external_lex_state = 21},
  [2804] = {.lex_state = 3, .external_lex_state = 23},
  [2805] = {.lex_state = 3, .external_lex_state = 23},
  [2806] = {.lex_state = 3, .external_lex_state = 23},
  [2807] = {.lex_state = 3, .external_lex_state = 23},
  [2808] = {.lex_state = 3, .external_lex_state = 23},
  [2809] = {.lex_state = 3, .external_lex_state = 23},
  [2810] = {.lex_state = 3, .external_lex_state = 23},
  [2811] = {.lex_state = 3, .external_lex_state = 23},
  [2812] = {.lex_state = 3, .external_lex_state = 23},
  [2813] = {.lex_state = 3, .external_lex_state = 23},
  [2814] = {.lex_state = 3, .external_lex_state = 21},
  [2815] = {.lex_state = 3, .external_lex_state = 23},
  [2816] = {.lex_state = 3, .external_lex_state = 23},
  [2817] = {.lex_state = 3, .external_lex_state = 23},
  [2818] = {.lex_state = 3, .external_lex_state = 23},
  [2819] = {.lex_state = 3, .external_lex_state = 23},
  [2820] = {.lex_state = 3, .external_lex_state = 23},
  [2821] = {.lex_state = 3, .external_lex_state = 23},
  [2822] = {.lex_state = 3, .external_lex_state = 23},
  [2823] = {.lex_state = 65, .external_lex_state = 20},
  [2824] = {.lex_state = 3, .external_lex_state = 23},
  [2825] = {.lex_state = 3, .external_lex_state = 23},
  [2826] = {.lex_state = 3, .external_lex_state = 23},
  [2827] = {.lex_state = 3, .external_lex_state = 23},
  [2828] = {.lex_state = 57, .external_lex_state = 20},
  [2829] = {.lex_state = 3, .external_lex_state = 21},
  [2830] = {.lex_state = 3, .external_lex_state = 21},
  [2831] = {.lex_state = 3, .external_lex_state = 21},
  [2832] = {.lex_state = 3, .external_lex_state = 21},
  [2833] = {.lex_state = 3, .external_lex_state = 21},
  [2834] = {.lex_state = 57, .external_lex_state = 20},
  [2835] = {.lex_state = 57, .external_lex_state = 20},
  [2836] = {.lex_state = 57, .external_lex_state = 20},
  [2837] = {.lex_state = 57, .external_lex_state = 20},
  [2838] = {.lex_state = 57, .external_lex_state = 20},
  [2839] = {.lex_state = 57, .external_lex_state = 20},
  [2840] = {.lex_state = 57, .external_lex_state = 20},
  [2841] = {.lex_state = 57, .external_lex_state = 20},
  [2842] = {.lex_state = 57, .external_lex_state = 20},
  [2843] = {.lex_state = 57, .external_lex_state = 20},
  [2844] = {.lex_state = 57, .external_lex_state = 20},
  [2845] = {.lex_state = 57, .external_lex_state = 20},
  [2846] = {.lex_state = 3, .external_lex_state = 21},
  [2847] = {.lex_state = 57, .external_lex_state = 20},
  [2848] = {.lex_state = 57, .external_lex_state = 20},
  [2849] = {.lex_state = 57, .external_lex_state = 20},
  [2850] = {.lex_state = 57, .external_lex_state = 20},
  [2851] = {.lex_state = 57, .external_lex_state = 20},
  [2852] = {.lex_state = 57, .external_lex_state = 20},
  [2853] = {.lex_state = 57, .external_lex_state = 20},
  [2854] = {.lex_state = 3, .external_lex_state = 21},
  [2855] = {.lex_state = 57, .external_lex_state = 20},
  [2856] = {.lex_state = 57, .external_lex_state = 20},
  [2857] = {.lex_state = 57, .external_lex_state = 20},
  [2858] = {.lex_state = 3, .external_lex_state = 23},
  [2859] = {.lex_state = 3, .external_lex_state = 23},
  [2860] = {.lex_state = 57, .external_lex_state = 20},
  [2861] = {.lex_state = 3, .external_lex_state = 21},
  [2862] = {.lex_state = 57, .external_lex_state = 20},
  [2863] = {.lex_state = 57, .external_lex_state = 20},
  [2864] = {.lex_state = 57, .external_lex_state = 20},
  [2865] = {.lex_state = 57, .external_lex_state = 20},
  [2866] = {.lex_state = 3, .external_lex_state = 23},
  [2867] = {.lex_state = 57, .external_lex_state = 20},
  [2868] = {.lex_state = 3, .external_lex_state = 21},
  [2869] = {.lex_state = 3, .external_lex_state = 21},
  [2870] = {.lex_state = 3, .external_lex_state = 21},
  [2871] = {.lex_state = 3, .external_lex_state = 21},
  [2872] = {.lex_state = 3, .external_lex_state = 21},
  [2873] = {.lex_state = 3, .external_lex_state = 21},
  [2874] = {.lex_state = 57, .external_lex_state = 20},
  [2875] = {.lex_state = 57, .external_lex_state = 20},
  [2876] = {.lex_state = 57, .external_lex_state = 20},
  [2877] = {.lex_state = 57, .external_lex_state = 20},
  [2878] = {.lex_state = 57, .external_lex_state = 20},
  [2879] = {.lex_state = 57, .external_lex_state = 20},
  [2880] = {.lex_state = 57, .external_lex_state = 20},
  [2881] = {.lex_state = 57, .external_lex_state = 20},
  [2882] = {.lex_state = 57, .external_lex_state = 20},
  [2883] = {.lex_state = 57, .external_lex_state = 20},
  [2884] = {.lex_state = 57, .external_lex_state = 20},
  [2885] = {.lex_state = 57, .external_lex_state = 20},
  [2886] = {.lex_state = 57, .external_lex_state = 20},
  [2887] = {.lex_state = 57, .external_lex_state = 20},
  [2888] = {.lex_state = 57, .external_lex_state = 20},
  [2889] = {.lex_state = 57, .external_lex_state = 20},
  [2890] = {.lex_state = 57, .external_lex_state = 20},
  [2891] = {.lex_state = 57, .external_lex_state = 20},
  [2892] = {.lex_state = 57, .external_lex_state = 20},
  [2893] = {.lex_state = 57, .external_lex_state = 20},
  [2894] = {.lex_state = 57, .external_lex_state = 20},
  [2895] = {.lex_state = 57, .external_lex_state = 20},
  [2896] = {.lex_state = 57, .external_lex_state = 20},
  [2897] = {.lex_state = 3, .external_lex_state = 21},
  [2898] = {.lex_state = 57, .external_lex_state = 20},
  [2899] = {.lex_state = 3, .external_lex_state = 21},
  [2900] = {.lex_state = 3, .external_lex_state = 23},
  [2901] = {.lex_state = 3, .external_lex_state = 23},
  [2902] = {.lex_state = 3, .external_lex_state = 23},
  [2903] = {.lex_state = 3, .external_lex_state = 23},
  [2904] = {.lex_state = 3, .external_lex_state = 23},
  [2905] = {.lex_state = 3, .external_lex_state = 23},
  [2906] = {.lex_state = 3, .external_lex_state = 23},
  [2907] = {.lex_state = 3, .external_lex_state = 23},
  [2908] = {.lex_state = 3, .external_lex_state = 23},
  [2909] = {.lex_state = 3, .external_lex_state = 23},
  [2910] = {.lex_state = 3, .external_lex_state = 23},
  [2911] = {.lex_state = 3, .external_lex_state = 23},
  [2912] = {.lex_state = 3, .external_lex_state = 23},
  [2913] = {.lex_state = 3, .external_lex_state = 23},
  [2914] = {.lex_state = 3, .external_lex_state = 23},
  [2915] = {.lex_state = 3, .external_lex_state = 23},
  [2916] = {.lex_state = 57, .external_lex_state = 20},
  [2917] = {.lex_state = 3, .external_lex_state = 21},
  [2918] = {.lex_state = 3, .external_lex_state = 21},
  [2919] = {.lex_state = 57, .external_lex_state = 20},
  [2920] = {.lex_state = 57, .external_lex_state = 20},
  [2921] = {.lex_state = 57, .external_lex_state = 20},
  [2922] = {.lex_state = 3, .external_lex_state = 21},
  [2923] = {.lex_state = 65, .external_lex_state = 28},
  [2924] = {.lex_state = 3, .external_lex_state = 21},
  [2925] = {.lex_state = 57, .external_lex_state = 20},
  [2926] = {.lex_state = 57, .external_lex_state = 20},
  [2927] = {.lex_state = 57, .external_lex_state = 20},
  [2928] = {.lex_state = 57, .external_lex_state = 20},
  [2929] = {.lex_state = 57, .external_lex_state = 20},
  [2930] = {.lex_state = 57, .external_lex_state = 20},
  [2931] = {.lex_state = 57, .external_lex_state = 20},
  [2932] = {.lex_state = 57, .external_lex_state = 20},
  [2933] = {.lex_state = 57, .external_lex_state = 20},
  [2934] = {.lex_state = 57, .external_lex_state = 20},
  [2935] = {.lex_state = 57, .external_lex_state = 20},
  [2936] = {.lex_state = 57, .external_lex_state = 20},
  [2937] = {.lex_state = 57, .external_lex_state = 20},
  [2938] = {.lex_state = 57, .external_lex_state = 20},
  [2939] = {.lex_state = 3, .external_lex_state = 21},
  [2940] = {.lex_state = 3, .external_lex_state = 21},
  [2941] = {.lex_state = 3, .external_lex_state = 21},
  [2942] = {.lex_state = 57, .external_lex_state = 20},
  [2943] = {.lex_state = 57, .external_lex_state = 20},
  [2944] = {.lex_state = 57, .external_lex_state = 20},
  [2945] = {.lex_state = 57, .external_lex_state = 20},
  [2946] = {.lex_state = 57, .external_lex_state = 20},
  [2947] = {.lex_state = 57, .external_lex_state = 20},
  [2948] = {.lex_state = 3, .external_lex_state = 21},
  [2949] = {.lex_state = 57, .external_lex_state = 20},
  [2950] = {.lex_state = 57, .external_lex_state = 20},
  [2951] = {.lex_state = 57, .external_lex_state = 20},
  [2952] = {.lex_state = 57, .external_lex_state = 20},
  [2953] = {.lex_state = 57, .external_lex_state = 20},
  [2954] = {.lex_state = 3, .external_lex_state = 21},
  [2955] = {.lex_state = 57, .external_lex_state = 20},
  [2956] = {.lex_state = 3, .external_lex_state = 21},
  [2957] = {.lex_state = 65, .external_lex_state = 28},
  [2958] = {.lex_state = 3, .external_lex_state = 21},
  [2959] = {.lex_state = 57, .external_lex_state = 20},
  [2960] = {.lex_state = 57, .external_lex_state = 20},
  [2961] = {.lex_state = 57, .external_lex_state = 20},
  [2962] = {.lex_state = 57, .external_lex_state = 20},
  [2963] = {.lex_state = 57, .external_lex_state = 20},
  [2964] = {.lex_state = 57, .external_lex_state = 20},
  [2965] = {.lex_state = 57, .external_lex_state = 20},
  [2966] = {.lex_state = 3, .external_lex_state = 21},
  [2967] = {.lex_state = 57, .external_lex_state = 20},
  [2968] = {.lex_state = 3, .external_lex_state = 21},
  [2969] = {.lex_state = 3, .external_lex_state = 21},
  [2970] = {.lex_state = 57, .external_lex_state = 20},
  [2971] = {.lex_state = 3, .external_lex_state = 21},
  [2972] = {.lex_state = 3, .external_lex_state = 21},
  [2973] = {.lex_state = 3, .external_lex_state = 21},
  [2974] = {.lex_state = 57, .external_lex_state = 20},
  [2975] = {.lex_state = 3, .external_lex_state = 23},
  [2976] = {.lex_state = 57, .external_lex_state = 20},
  [2977] = {.lex_state = 57, .external_lex_state = 20},
  [2978] = {.lex_state = 57, .external_lex_state = 20},
  [2979] = {.lex_state = 57, .external_lex_state = 20},
  [2980] = {.lex_state = 57, .external_lex_state = 20},
  [2981] = {.lex_state = 3, .external_lex_state = 21},
  [2982] = {.lex_state = 57, .external_lex_state = 20},
  [2983] = {.lex_state = 57, .external_lex_state = 20},
  [2984] = {.lex_state = 57, .external_lex_state = 20},
  [2985] = {.lex_state = 57, .external_lex_state = 20},
  [2986] = {.lex_state = 3, .external_lex_state = 21},
  [2987] = {.lex_state = 57, .external_lex_state = 20},
  [2988] = {.lex_state = 57, .external_lex_state = 20},
  [2989] = {.lex_state = 3, .external_lex_state = 21},
  [2990] = {.lex_state = 57, .external_lex_state = 20},
  [2991] = {.lex_state = 57, .external_lex_state = 20},
  [2992] = {.lex_state = 57, .external_lex_state = 20},
  [2993] = {.lex_state = 57, .external_lex_state = 20},
  [2994] = {.lex_state = 3, .external_lex_state = 21},
  [2995] = {.lex_state = 57, .external_lex_state = 20},
  [2996] = {.lex_state = 57, .external_lex_state = 20},
  [2997] = {.lex_state = 57, .external_lex_state = 20},
  [2998] = {.lex_state = 57, .external_lex_state = 20},
  [2999] = {.lex_state = 57, .external_lex_state = 20},
  [3000] = {.lex_state = 57, .external_lex_state = 20},
  [3001] = {.lex_state = 57, .external_lex_state = 20},
  [3002] = {.lex_state = 57, .external_lex_state = 20},
  [3003] = {.lex_state = 65, .external_lex_state = 28},
  [3004] = {.lex_state = 57, .external_lex_state = 20},
  [3005] = {.lex_state = 57, .external_lex_state = 20},
  [3006] = {.lex_state = 57, .external_lex_state = 20},
  [3007] = {.lex_state = 57, .external_lex_state = 20},
  [3008] = {.lex_state = 3, .external_lex_state = 21},
  [3009] = {.lex_state = 3, .external_lex_state = 21},
  [3010] = {.lex_state = 3, .external_lex_state = 21},
  [3011] = {.lex_state = 3, .external_lex_state = 21},
  [3012] = {.lex_state = 3, .external_lex_state = 21},
  [3013] = {.lex_state = 3, .external_lex_state = 21},
  [3014] = {.lex_state = 3, .external_lex_state = 21},
  [3015] = {.lex_state = 3, .external_lex_state = 21},
  [3016] = {.lex_state = 57, .external_lex_state = 20},
  [3017] = {.lex_state = 57, .external_lex_state = 20},
  [3018] = {.lex_state = 57, .external_lex_state = 20},
  [3019] = {.lex_state = 57, .external_lex_state = 20},
  [3020] = {.lex_state = 3, .external_lex_state = 21},
  [3021] = {.lex_state = 3, .external_lex_state = 21},
  [3022] = {.lex_state = 3, .external_lex_state = 21},
  [3023] = {.lex_state = 3, .external_lex_state = 21},
  [3024] = {.lex_state = 3, .external_lex_state = 21},
  [3025] = {.lex_state = 3, .external_lex_state = 21},
  [3026] = {.lex_state = 67, .external_lex_state = 20},
  [3027] = {.lex_state = 3, .external_lex_state = 21},
  [3028] = {.lex_state = 3, .external_lex_state = 21},
  [3029] = {.lex_state = 3, .external_lex_state = 21},
  [3030] = {.lex_state = 57, .external_lex_state = 20},
  [3031] = {.lex_state = 3, .external_lex_state = 22},
  [3032] = {.lex_state = 3, .external_lex_state = 21},
  [3033] = {.lex_state = 3, .external_lex_state = 21},
  [3034] = {.lex_state = 3, .external_lex_state = 21},
  [3035] = {.lex_state = 3, .external_lex_state = 21},
  [3036] = {.lex_state = 3, .external_lex_state = 21},
  [3037] = {.lex_state = 3, .external_lex_state = 21},
  [3038] = {.lex_state = 3, .external_lex_state = 21},
  [3039] = {.lex_state = 3, .external_lex_state = 21},
  [3040] = {.lex_state = 3, .external_lex_state = 21},
  [3041] = {.lex_state = 3, .external_lex_state = 21},
  [3042] = {.lex_state = 3, .external_lex_state = 21},
  [3043] = {.lex_state = 3, .external_lex_state = 21},
  [3044] = {.lex_state = 3, .external_lex_state = 21},
  [3045] = {.lex_state = 3, .external_lex_state = 21},
  [3046] = {.lex_state = 3, .external_lex_state = 21},
  [3047] = {.lex_state = 3, .external_lex_state = 21},
  [3048] = {.lex_state = 3, .external_lex_state = 21},
  [3049] = {.lex_state = 3, .external_lex_state = 21},
  [3050] = {.lex_state = 3, .external_lex_state = 21},
  [3051] = {.lex_state = 3, .external_lex_state = 21},
  [3052] = {.lex_state = 3, .external_lex_state = 21},
  [3053] = {.lex_state = 3, .external_lex_state = 21},
  [3054] = {.lex_state = 3, .external_lex_state = 21},
  [3055] = {.lex_state = 3, .external_lex_state = 22},
  [3056] = {.lex_state = 3, .external_lex_state = 21},
  [3057] = {.lex_state = 3, .external_lex_state = 21},
  [3058] = {.lex_state = 3, .external_lex_state = 21},
  [3059] = {.lex_state = 3, .external_lex_state = 21},
  [3060] = {.lex_state = 3, .external_lex_state = 21},
  [3061] = {.lex_state = 3, .external_lex_state = 21},
  [3062] = {.lex_state = 3, .external_lex_state = 21},
  [3063] = {.lex_state = 3, .external_lex_state = 23},
  [3064] = {.lex_state = 3, .external_lex_state = 21},
  [3065] = {.lex_state = 3, .external_lex_state = 21},
  [3066] = {.lex_state = 3, .external_lex_state = 21},
  [3067] = {.lex_state = 3, .external_lex_state = 21},
  [3068] = {.lex_state = 64, .external_lex_state = 28},
  [3069] = {.lex_state = 3, .external_lex_state = 21},
  [3070] = {.lex_state = 64, .external_lex_state = 28},
  [3071] = {.lex_state = 3, .external_lex_state = 21},
  [3072] = {.lex_state = 57, .external_lex_state = 20},
  [3073] = {.lex_state = 3, .external_lex_state = 21},
  [3074] = {.lex_state = 3, .external_lex_state = 21},
  [3075] = {.lex_state = 3, .external_lex_state = 21},
  [3076] = {.lex_state = 3, .external_lex_state = 21},
  [3077] = {.lex_state = 3, .external_lex_state = 21},
  [3078] = {.lex_state = 3, .external_lex_state = 21},
  [3079] = {.lex_state = 3, .external_lex_state = 21},
  [3080] = {.lex_state = 3, .external_lex_state = 21},
  [3081] = {.lex_state = 3, .external_lex_state = 21},
  [3082] = {.lex_state = 3, .external_lex_state = 21},
  [3083] = {.lex_state = 3, .external_lex_state = 21},
  [3084] = {.lex_state = 3, .external_lex_state = 21},
  [3085] = {.lex_state = 64, .external_lex_state = 28},
  [3086] = {.lex_state = 3, .external_lex_state = 21},
  [3087] = {.lex_state = 64, .external_lex_state = 28},
  [3088] = {.lex_state = 3, .external_lex_state = 21},
  [3089] = {.lex_state = 3, .external_lex_state = 21},
  [3090] = {.lex_state = 3, .external_lex_state = 21},
  [3091] = {.lex_state = 64, .external_lex_state = 28},
  [3092] = {.lex_state = 64, .external_lex_state = 28},
  [3093] = {.lex_state = 64, .external_lex_state = 20},
  [3094] = {.lex_state = 64, .external_lex_state = 20},
  [3095] = {.lex_state = 3, .external_lex_state = 21},
  [3096] = {.lex_state = 64, .external_lex_state = 20},
  [3097] = {.lex_state = 57, .external_lex_state = 20},
  [3098] = {.lex_state = 57, .external_lex_state = 20},
  [3099] = {.lex_state = 57, .external_lex_state = 20},
  [3100] = {.lex_state = 57, .external_lex_state = 20},
  [3101] = {.lex_state = 57, .external_lex_state = 20},
  [3102] = {.lex_state = 57, .external_lex_state = 20},
  [3103] = {.lex_state = 57, .external_lex_state = 20},
  [3104] = {.lex_state = 57, .external_lex_state = 20},
  [3105] = {.lex_state = 57, .external_lex_state = 20},
  [3106] = {.lex_state = 57, .external_lex_state = 20},
  [3107] = {.lex_state = 57, .external_lex_state = 20},
  [3108] = {.lex_state = 57, .external_lex_state = 20},
  [3109] = {.lex_state = 57, .external_lex_state = 20},
  [3110] = {.lex_state = 57, .external_lex_state = 20},
  [3111] = {.lex_state = 57, .external_lex_state = 20},
  [3112] = {.lex_state = 57, .external_lex_state = 20},
  [3113] = {.lex_state = 64, .external_lex_state = 28},
  [3114] = {.lex_state = 64, .external_lex_state = 28},
  [3115] = {.lex_state = 64, .external_lex_state = 28},
  [3116] = {.lex_state = 64, .external_lex_state = 28},
  [3117] = {.lex_state = 64, .external_lex_state = 28},
  [3118] = {.lex_state = 64, .external_lex_state = 28},
  [3119] = {.lex_state = 64, .external_lex_state = 28},
  [3120] = {.lex_state = 64, .external_lex_state = 28},
  [3121] = {.lex_state = 57, .external_lex_state = 20},
  [3122] = {.lex_state = 57, .external_lex_state = 20},
  [3123] = {.lex_state = 57, .external_lex_state = 20},
  [3124] = {.lex_state = 57, .external_lex_state = 20},
  [3125] = {.lex_state = 57, .external_lex_state = 20},
  [3126] = {.lex_state = 57, .external_lex_state = 20},
  [3127] = {.lex_state = 57, .external_lex_state = 20},
  [3128] = {.lex_state = 57, .external_lex_state = 20},
  [3129] = {.lex_state = 57, .external_lex_state = 20},
  [3130] = {.lex_state = 57, .external_lex_state = 20},
  [3131] = {.lex_state = 64, .external_lex_state = 28},
  [3132] = {.lex_state = 57, .external_lex_state = 20},
  [3133] = {.lex_state = 64, .external_lex_state = 28},
  [3134] = {.lex_state = 64, .external_lex_state = 28},
  [3135] = {.lex_state = 64, .external_lex_state = 28},
  [3136] = {.lex_state = 64, .external_lex_state = 28},
  [3137] = {.lex_state = 64, .external_lex_state = 28},
  [3138] = {.lex_state = 64, .external_lex_state = 28},
  [3139] = {.lex_state = 64, .external_lex_state = 28},
  [3140] = {.lex_state = 3, .external_lex_state = 22},
  [3141] = {.lex_state = 64, .external_lex_state = 28},
  [3142] = {.lex_state = 57, .external_lex_state = 20},
  [3143] = {.lex_state = 3, .external_lex_state = 22},
  [3144] = {.lex_state = 3, .external_lex_state = 22},
  [3145] = {.lex_state = 3, .external_lex_state = 22},
  [3146] = {.lex_state = 3, .external_lex_state = 22},
  [3147] = {.lex_state = 3, .external_lex_state = 22},
  [3148] = {.lex_state = 3, .external_lex_state = 22},
  [3149] = {.lex_state = 3, .external_lex_state = 22},
  [3150] = {.lex_state = 3, .external_lex_state = 22},
  [3151] = {.lex_state = 3, .external_lex_state = 22},
  [3152] = {.lex_state = 3, .external_lex_state = 22},
  [3153] = {.lex_state = 3, .external_lex_state = 22},
  [3154] = {.lex_state = 3, .external_lex_state = 22},
  [3155] = {.lex_state = 3, .external_lex_state = 22},
  [3156] = {.lex_state = 3, .external_lex_state = 22},
  [3157] = {.lex_state = 3, .external_lex_state = 22},
  [3158] = {.lex_state = 3, .external_lex_state = 22},
  [3159] = {.lex_state = 3, .external_lex_state = 22},
  [3160] = {.lex_state = 3, .external_lex_state = 22},
  [3161] = {.lex_state = 3, .external_lex_state = 22},
  [3162] = {.lex_state = 3, .external_lex_state = 22},
  [3163] = {.lex_state = 3, .external_lex_state = 22},
  [3164] = {.lex_state = 3, .external_lex_state = 22},
  [3165] = {.lex_state = 3, .external_lex_state = 22},
  [3166] = {.lex_state = 3, .external_lex_state = 22},
  [3167] = {.lex_state = 3, .external_lex_state = 22},
  [3168] = {.lex_state = 3, .external_lex_state = 22},
  [3169] = {.lex_state = 3, .external_lex_state = 22},
  [3170] = {.lex_state = 3, .external_lex_state = 22},
  [3171] = {.lex_state = 3, .external_lex_state = 22},
  [3172] = {.lex_state = 3, .external_lex_state = 22},
  [3173] = {.lex_state = 3, .external_lex_state = 22},
  [3174] = {.lex_state = 3, .external_lex_state = 22},
  [3175] = {.lex_state = 3, .external_lex_state = 22},
  [3176] = {.lex_state = 3, .external_lex_state = 22},
  [3177] = {.lex_state = 3, .external_lex_state = 22},
  [3178] = {.lex_state = 3, .external_lex_state = 22},
  [3179] = {.lex_state = 3, .external_lex_state = 22},
  [3180] = {.lex_state = 3, .external_lex_state = 22},
  [3181] = {.lex_state = 3, .external_lex_state = 22},
  [3182] = {.lex_state = 3, .external_lex_state = 22},
  [3183] = {.lex_state = 3, .external_lex_state = 22},
  [3184] = {.lex_state = 3, .external_lex_state = 22},
  [3185] = {.lex_state = 3, .external_lex_state = 22},
  [3186] = {.lex_state = 3, .external_lex_state = 22},
  [3187] = {.lex_state = 3, .external_lex_state = 22},
  [3188] = {.lex_state = 3, .external_lex_state = 22},
  [3189] = {.lex_state = 3, .external_lex_state = 22},
  [3190] = {.lex_state = 3, .external_lex_state = 22},
  [3191] = {.lex_state = 3, .external_lex_state = 22},
  [3192] = {.lex_state = 3, .external_lex_state = 22},
  [3193] = {.lex_state = 3, .external_lex_state = 22},
  [3194] = {.lex_state = 3, .external_lex_state = 22},
  [3195] = {.lex_state = 3, .external_lex_state = 22},
  [3196] = {.lex_state = 3, .external_lex_state = 22},
  [3197] = {.lex_state = 3, .external_lex_state = 22},
  [3198] = {.lex_state = 3, .external_lex_state = 22},
  [3199] = {.lex_state = 3, .external_lex_state = 22},
  [3200] = {.lex_state = 3, .external_lex_state = 22},
  [3201] = {.lex_state = 3, .external_lex_state = 22},
  [3202] = {.lex_state = 3, .external_lex_state = 22},
  [3203] = {.lex_state = 3, .external_lex_state = 22},
  [3204] = {.lex_state = 3, .external_lex_state = 22},
  [3205] = {.lex_state = 3, .external_lex_state = 22},
  [3206] = {.lex_state = 3, .external_lex_state = 22},
  [3207] = {.lex_state = 3, .external_lex_state = 22},
  [3208] = {.lex_state = 3, .external_lex_state = 22},
  [3209] = {.lex_state = 3, .external_lex_state = 22},
  [3210] = {.lex_state = 3, .external_lex_state = 22},
  [3211] = {.lex_state = 3, .external_lex_state = 22},
  [3212] = {.lex_state = 3, .external_lex_state = 22},
  [3213] = {.lex_state = 3, .external_lex_state = 22},
  [3214] = {.lex_state = 3, .external_lex_state = 22},
  [3215] = {.lex_state = 3, .external_lex_state = 22},
  [3216] = {.lex_state = 3, .external_lex_state = 22},
  [3217] = {.lex_state = 3, .external_lex_state = 22},
  [3218] = {.lex_state = 3, .external_lex_state = 22},
  [3219] = {.lex_state = 3, .external_lex_state = 22},
  [3220] = {.lex_state = 3, .external_lex_state = 22},
  [3221] = {.lex_state = 3, .external_lex_state = 22},
  [3222] = {.lex_state = 3, .external_lex_state = 22},
  [3223] = {.lex_state = 3, .external_lex_state = 22},
  [3224] = {.lex_state = 3, .external_lex_state = 22},
  [3225] = {.lex_state = 3, .external_lex_state = 22},
  [3226] = {.lex_state = 3, .external_lex_state = 22},
  [3227] = {.lex_state = 3, .external_lex_state = 22},
  [3228] = {.lex_state = 3, .external_lex_state = 22},
  [3229] = {.lex_state = 3, .external_lex_state = 22},
  [3230] = {.lex_state = 3, .external_lex_state = 22},
  [3231] = {.lex_state = 3, .external_lex_state = 22},
  [3232] = {.lex_state = 3, .external_lex_state = 22},
  [3233] = {.lex_state = 3, .external_lex_state = 22},
  [3234] = {.lex_state = 3, .external_lex_state = 22},
  [3235] = {.lex_state = 3, .external_lex_state = 22},
  [3236] = {.lex_state = 3, .external_lex_state = 22},
  [3237] = {.lex_state = 3, .external_lex_state = 22},
  [3238] = {.lex_state = 3, .external_lex_state = 22},
  [3239] = {.lex_state = 3, .external_lex_state = 21},
  [3240] = {.lex_state = 57, .external_lex_state = 20},
  [3241] = {.lex_state = 71, .external_lex_state = 20},
  [3242] = {.lex_state = 64, .external_lex_state = 20},
  [3243] = {.lex_state = 64, .external_lex_state = 20},
  [3244] = {.lex_state = 68, .external_lex_state = 20},
  [3245] = {.lex_state = 71, .external_lex_state = 20},
  [3246] = {.lex_state = 70, .external_lex_state = 20},
  [3247] = {.lex_state = 67, .external_lex_state = 20},
  [3248] = {.lex_state = 64, .external_lex_state = 20},
  [3249] = {.lex_state = 70, .external_lex_state = 20},
  [3250] = {.lex_state = 70, .external_lex_state = 20},
  [3251] = {.lex_state = 67, .external_lex_state = 20},
  [3252] = {.lex_state = 67, .external_lex_state = 20},
  [3253] = {.lex_state = 70, .external_lex_state = 20},
  [3254] = {.lex_state = 67, .external_lex_state = 20},
  [3255] = {.lex_state = 57, .external_lex_state = 20},
  [3256] = {.lex_state = 64, .external_lex_state = 28},
  [3257] = {.lex_state = 64, .external_lex_state = 28},
  [3258] = {.lex_state = 57, .external_lex_state = 20},
  [3259] = {.lex_state = 57, .external_lex_state = 20},
  [3260] = {.lex_state = 57, .external_lex_state = 20},
  [3261] = {.lex_state = 57, .external_lex_state = 20},
  [3262] = {.lex_state = 71, .external_lex_state = 20},
  [3263] = {.lex_state = 57, .external_lex_state = 20},
  [3264] = {.lex_state = 57, .external_lex_state = 20},
  [3265] = {.lex_state = 57, .external_lex_state = 20},
  [3266] = {.lex_state = 57, .external_lex_state = 20},
  [3267] = {.lex_state = 57, .external_lex_state = 20},
  [3268] = {.lex_state = 3, .external_lex_state = 20},
  [3269] = {.lex_state = 57, .external_lex_state = 20},
  [3270] = {.lex_state = 3, .external_lex_state = 20},
  [3271] = {.lex_state = 3, .external_lex_state = 20},
  [3272] = {.lex_state = 64, .external_lex_state = 28},
  [3273] = {.lex_state = 64, .external_lex_state = 20},
  [3274] = {.lex_state = 71, .external_lex_state = 20},
  [3275] = {.lex_state = 67, .external_lex_state = 28},
  [3276] = {.lex_state = 57, .external_lex_state = 20},
  [3277] = {.lex_state = 64, .external_lex_state = 20},
  [3278] = {.lex_state = 64, .external_lex_state = 20},
  [3279] = {.lex_state = 64, .external_lex_state = 20},
  [3280] = {.lex_state = 57, .external_lex_state = 20},
  [3281] = {.lex_state = 64, .external_lex_state = 20},
  [3282] = {.lex_state = 64, .external_lex_state = 20},
  [3283] = {.lex_state = 57, .external_lex_state = 20},
  [3284] = {.lex_state = 57, .external_lex_state = 20},
  [3285] = {.lex_state = 64, .external_lex_state = 20},
  [3286] = {.lex_state = 64, .external_lex_state = 20},
  [3287] = {.lex_state = 71, .external_lex_state = 20},
  [3288] = {.lex_state = 64, .external_lex_state = 20},
  [3289] = {.lex_state = 64, .external_lex_state = 20},
  [3290] = {.lex_state = 64, .external_lex_state = 20},
  [3291] = {.lex_state = 68, .external_lex_state = 20},
  [3292] = {.lex_state = 71, .external_lex_state = 20},
  [3293] = {.lex_state = 64, .external_lex_state = 20},
  [3294] = {.lex_state = 64, .external_lex_state = 20},
  [3295] = {.lex_state = 57, .external_lex_state = 20},
  [3296] = {.lex_state = 57, .external_lex_state = 20},
  [3297] = {.lex_state = 57, .external_lex_state = 20},
  [3298] = {.lex_state = 57, .external_lex_state = 20},
  [3299] = {.lex_state = 57, .external_lex_state = 20},
  [3300] = {.lex_state = 57, .external_lex_state = 20},
  [3301] = {.lex_state = 57, .external_lex_state = 20},
  [3302] = {.lex_state = 57, .external_lex_state = 20},
  [3303] = {.lex_state = 57, .external_lex_state = 20},
  [3304] = {.lex_state = 68, .external_lex_state = 20},
  [3305] = {.lex_state = 64, .external_lex_state = 28},
  [3306] = {.lex_state = 57, .external_lex_state = 20},
  [3307] = {.lex_state = 57, .external_lex_state = 20},
  [3308] = {.lex_state = 57, .external_lex_state = 20},
  [3309] = {.lex_state = 70, .external_lex_state = 20},
  [3310] = {.lex_state = 70, .external_lex_state = 28},
  [3311] = {.lex_state = 64, .external_lex_state = 20},
  [3312] = {.lex_state = 64, .external_lex_state = 20},
  [3313] = {.lex_state = 64, .external_lex_state = 20},
  [3314] = {.lex_state = 64, .external_lex_state = 20},
  [3315] = {.lex_state = 64, .external_lex_state = 20},
  [3316] = {.lex_state = 64, .external_lex_state = 20},
  [3317] = {.lex_state = 64, .external_lex_state = 20},
  [3318] = {.lex_state = 64, .external_lex_state = 20},
  [3319] = {.lex_state = 64, .external_lex_state = 20},
  [3320] = {.lex_state = 64, .external_lex_state = 20},
  [3321] = {.lex_state = 64, .external_lex_state = 20},
  [3322] = {.lex_state = 64, .external_lex_state = 20},
  [3323] = {.lex_state = 64, .external_lex_state = 20},
  [3324] = {.lex_state = 64, .external_lex_state = 20},
  [3325] = {.lex_state = 64, .external_lex_state = 20},
  [3326] = {.lex_state = 64, .external_lex_state = 20},
  [3327] = {.lex_state = 64, .external_lex_state = 20},
  [3328] = {.lex_state = 64, .external_lex_state = 20},
  [3329] = {.lex_state = 64, .external_lex_state = 20},
  [3330] = {.lex_state = 64, .external_lex_state = 20},
  [3331] = {.lex_state = 64, .external_lex_state = 20},
  [3332] = {.lex_state = 64, .external_lex_state = 20},
  [3333] = {.lex_state = 64, .external_lex_state = 20},
  [3334] = {.lex_state = 64, .external_lex_state = 20},
  [3335] = {.lex_state = 64, .external_lex_state = 20},
  [3336] = {.lex_state = 64, .external_lex_state = 28},
  [3337] = {.lex_state = 71, .external_lex_state = 28},
  [3338] = {.lex_state = 64, .external_lex_state = 20},
  [3339] = {.lex_state = 64, .external_lex_state = 20},
  [3340] = {.lex_state = 64, .external_lex_state = 28},
  [3341] = {.lex_state = 64, .external_lex_state = 28},
  [3342] = {.lex_state = 64, .external_lex_state = 28},
  [3343] = {.lex_state = 64, .external_lex_state = 20},
  [3344] = {.lex_state = 64, .external_lex_state = 28},
  [3345] = {.lex_state = 64, .external_lex_state = 28},
  [3346] = {.lex_state = 64, .external_lex_state = 28},
  [3347] = {.lex_state = 64, .external_lex_state = 28},
  [3348] = {.lex_state = 64, .external_lex_state = 20},
  [3349] = {.lex_state = 64, .external_lex_state = 20},
  [3350] = {.lex_state = 64, .external_lex_state = 28},
  [3351] = {.lex_state = 64, .external_lex_state = 28},
  [3352] = {.lex_state = 71, .external_lex_state = 28},
  [3353] = {.lex_state = 64, .external_lex_state = 28},
  [3354] = {.lex_state = 64, .external_lex_state = 28},
  [3355] = {.lex_state = 64, .external_lex_state = 28},
  [3356] = {.lex_state = 64, .external_lex_state = 28},
  [3357] = {.lex_state = 68, .external_lex_state = 28},
  [3358] = {.lex_state = 71, .external_lex_state = 28},
  [3359] = {.lex_state = 64, .external_lex_state = 20},
  [3360] = {.lex_state = 64, .external_lex_state = 28},
  [3361] = {.lex_state = 64, .external_lex_state = 28},
  [3362] = {.lex_state = 57, .external_lex_state = 20},
  [3363] = {.lex_state = 57, .external_lex_state = 20},
  [3364] = {.lex_state = 71, .external_lex_state = 28},
  [3365] = {.lex_state = 64, .external_lex_state = 28},
  [3366] = {.lex_state = 64, .external_lex_state = 28},
  [3367] = {.lex_state = 64, .external_lex_state = 28},
  [3368] = {.lex_state = 68, .external_lex_state = 28},
  [3369] = {.lex_state = 71, .external_lex_state = 28},
  [3370] = {.lex_state = 70, .external_lex_state = 28},
  [3371] = {.lex_state = 67, .external_lex_state = 28},
  [3372] = {.lex_state = 64, .external_lex_state = 20},
  [3373] = {.lex_state = 64, .external_lex_state = 20},
  [3374] = {.lex_state = 64, .external_lex_state = 20},
  [3375] = {.lex_state = 64, .external_lex_state = 20},
  [3376] = {.lex_state = 71, .external_lex_state = 28},
  [3377] = {.lex_state = 64, .external_lex_state = 20},
  [3378] = {.lex_state = 64, .external_lex_state = 20},
  [3379] = {.lex_state = 64, .external_lex_state = 20},
  [3380] = {.lex_state = 64, .external_lex_state = 20},
  [3381] = {.lex_state = 64, .external_lex_state = 20},
  [3382] = {.lex_state = 64, .external_lex_state = 20},
  [3383] = {.lex_state = 64, .external_lex_state = 20},
  [3384] = {.lex_state = 64, .external_lex_state = 20},
  [3385] = {.lex_state = 64, .external_lex_state = 20},
  [3386] = {.lex_state = 64, .external_lex_state = 20},
  [3387] = {.lex_state = 64, .external_lex_state = 20},
  [3388] = {.lex_state = 64, .external_lex_state = 28},
  [3389] = {.lex_state = 64, .external_lex_state = 20},
  [3390] = {.lex_state = 64, .external_lex_state = 20},
  [3391] = {.lex_state = 64, .external_lex_state = 20},
  [3392] = {.lex_state = 64, .external_lex_state = 28},
  [3393] = {.lex_state = 64, .external_lex_state = 20},
  [3394] = {.lex_state = 70, .external_lex_state = 28},
  [3395] = {.lex_state = 70, .external_lex_state = 28},
  [3396] = {.lex_state = 67, .external_lex_state = 28},
  [3397] = {.lex_state = 67, .external_lex_state = 28},
  [3398] = {.lex_state = 64, .external_lex_state = 28},
  [3399] = {.lex_state = 64, .external_lex_state = 20},
  [3400] = {.lex_state = 64, .external_lex_state = 20},
  [3401] = {.lex_state = 64, .external_lex_state = 28},
  [3402] = {.lex_state = 64, .external_lex_state = 20},
  [3403] = {.lex_state = 64, .external_lex_state = 20},
  [3404] = {.lex_state = 67, .external_lex_state = 28},
  [3405] = {.lex_state = 64, .external_lex_state = 20},
  [3406] = {.lex_state = 57, .external_lex_state = 20},
  [3407] = {.lex_state = 57, .external_lex_state = 20},
  [3408] = {.lex_state = 57, .external_lex_state = 20},
  [3409] = {.lex_state = 57, .external_lex_state = 20},
  [3410] = {.lex_state = 57, .external_lex_state = 20},
  [3411] = {.lex_state = 57, .external_lex_state = 20},
  [3412] = {.lex_state = 57, .external_lex_state = 20},
  [3413] = {.lex_state = 57, .external_lex_state = 20},
  [3414] = {.lex_state = 57, .external_lex_state = 20},
  [3415] = {.lex_state = 64, .external_lex_state = 20},
  [3416] = {.lex_state = 57, .external_lex_state = 20},
  [3417] = {.lex_state = 64, .external_lex_state = 20},
  [3418] = {.lex_state = 64, .external_lex_state = 28},
  [3419] = {.lex_state = 64, .external_lex_state = 20},
  [3420] = {.lex_state = 64, .external_lex_state = 20},
  [3421] = {.lex_state = 64, .external_lex_state = 20},
  [3422] = {.lex_state = 64, .external_lex_state = 20},
  [3423] = {.lex_state = 64, .external_lex_state = 20},
  [3424] = {.lex_state = 64, .external_lex_state = 20},
  [3425] = {.lex_state = 64, .external_lex_state = 20},
  [3426] = {.lex_state = 64, .external_lex_state = 20},
  [3427] = {.lex_state = 64, .external_lex_state = 20},
  [3428] = {.lex_state = 64, .external_lex_state = 20},
  [3429] = {.lex_state = 68, .external_lex_state = 28},
  [3430] = {.lex_state = 64, .external_lex_state = 20},
  [3431] = {.lex_state = 64, .external_lex_state = 20},
  [3432] = {.lex_state = 64, .external_lex_state = 20},
  [3433] = {.lex_state = 64, .external_lex_state = 20},
  [3434] = {.lex_state = 64, .external_lex_state = 20},
  [3435] = {.lex_state = 64, .external_lex_state = 20},
  [3436] = {.lex_state = 64, .external_lex_state = 20},
  [3437] = {.lex_state = 64, .external_lex_state = 20},
  [3438] = {.lex_state = 64, .external_lex_state = 20},
  [3439] = {.lex_state = 64, .external_lex_state = 20},
  [3440] = {.lex_state = 64, .external_lex_state = 20},
  [3441] = {.lex_state = 64, .external_lex_state = 28},
  [3442] = {.lex_state = 64, .external_lex_state = 20},
  [3443] = {.lex_state = 64, .external_lex_state = 20},
  [3444] = {.lex_state = 64, .external_lex_state = 20},
  [3445] = {.lex_state = 64, .external_lex_state = 20},
  [3446] = {.lex_state = 64, .external_lex_state = 20},
  [3447] = {.lex_state = 64, .external_lex_state = 20},
  [3448] = {.lex_state = 64, .external_lex_state = 20},
  [3449] = {.lex_state = 70, .external_lex_state = 28},
  [3450] = {.lex_state = 57, .external_lex_state = 20},
  [3451] = {.lex_state = 64, .external_lex_state = 20},
  [3452] = {.lex_state = 64, .external_lex_state = 20},
  [3453] = {.lex_state = 64, .external_lex_state = 20},
  [3454] = {.lex_state = 57, .external_lex_state = 20},
  [3455] = {.lex_state = 64, .external_lex_state = 28},
  [3456] = {.lex_state = 64, .external_lex_state = 20},
  [3457] = {.lex_state = 64, .external_lex_state = 20},
  [3458] = {.lex_state = 64, .external_lex_state = 20},
  [3459] = {.lex_state = 64, .external_lex_state = 20},
  [3460] = {.lex_state = 64, .external_lex_state = 20},
  [3461] = {.lex_state = 64, .external_lex_state = 20},
  [3462] = {.lex_state = 57, .external_lex_state = 20},
  [3463] = {.lex_state = 64, .external_lex_state = 20},
  [3464] = {.lex_state = 64, .external_lex_state = 20},
  [3465] = {.lex_state = 64, .external_lex_state = 20},
  [3466] = {.lex_state = 57, .external_lex_state = 20},
  [3467] = {.lex_state = 57, .external_lex_state = 20},
  [3468] = {.lex_state = 64, .external_lex_state = 20},
  [3469] = {.lex_state = 64, .external_lex_state = 20},
  [3470] = {.lex_state = 64, .external_lex_state = 20},
  [3471] = {.lex_state = 57, .external_lex_state = 20},
  [3472] = {.lex_state = 64, .external_lex_state = 20},
  [3473] = {.lex_state = 64, .external_lex_state = 20},
  [3474] = {.lex_state = 57, .external_lex_state = 20},
  [3475] = {.lex_state = 64, .external_lex_state = 20},
  [3476] = {.lex_state = 64, .external_lex_state = 28},
  [3477] = {.lex_state = 64, .external_lex_state = 28},
  [3478] = {.lex_state = 64, .external_lex_state = 28},
  [3479] = {.lex_state = 64, .external_lex_state = 20},
  [3480] = {.lex_state = 64, .external_lex_state = 28},
  [3481] = {.lex_state = 64, .external_lex_state = 20},
  [3482] = {.lex_state = 64, .external_lex_state = 20},
  [3483] = {.lex_state = 57, .external_lex_state = 20},
  [3484] = {.lex_state = 64, .external_lex_state = 20},
  [3485] = {.lex_state = 64, .external_lex_state = 20},
  [3486] = {.lex_state = 64, .external_lex_state = 20},
  [3487] = {.lex_state = 64, .external_lex_state = 20},
  [3488] = {.lex_state = 64, .external_lex_state = 20},
  [3489] = {.lex_state = 64, .external_lex_state = 20},
  [3490] = {.lex_state = 57, .external_lex_state = 20},
  [3491] = {.lex_state = 64, .external_lex_state = 20},
  [3492] = {.lex_state = 57, .external_lex_state = 20},
  [3493] = {.lex_state = 57, .external_lex_state = 20},
  [3494] = {.lex_state = 64, .external_lex_state = 20},
  [3495] = {.lex_state = 57, .external_lex_state = 20},
  [3496] = {.lex_state = 57, .external_lex_state = 20},
  [3497] = {.lex_state = 64, .external_lex_state = 20},
  [3498] = {.lex_state = 57, .external_lex_state = 20},
  [3499] = {.lex_state = 57, .external_lex_state = 20},
  [3500] = {.lex_state = 64, .external_lex_state = 20},
  [3501] = {.lex_state = 64, .external_lex_state = 20},
  [3502] = {.lex_state = 64, .external_lex_state = 20},
  [3503] = {.lex_state = 64, .external_lex_state = 20},
  [3504] = {.lex_state = 64, .external_lex_state = 20},
  [3505] = {.lex_state = 57, .external_lex_state = 20},
  [3506] = {.lex_state = 64, .external_lex_state = 20},
  [3507] = {.lex_state = 64, .external_lex_state = 20},
  [3508] = {.lex_state = 64, .external_lex_state = 20},
  [3509] = {.lex_state = 64, .external_lex_state = 20},
  [3510] = {.lex_state = 64, .external_lex_state = 20},
  [3511] = {.lex_state = 64, .external_lex_state = 20},
  [3512] = {.lex_state = 64, .external_lex_state = 20},
  [3513] = {.lex_state = 64, .external_lex_state = 20},
  [3514] = {.lex_state = 64, .external_lex_state = 20},
  [3515] = {.lex_state = 64, .external_lex_state = 28},
  [3516] = {.lex_state = 64, .external_lex_state = 20},
  [3517] = {.lex_state = 64, .external_lex_state = 20},
  [3518] = {.lex_state = 64, .external_lex_state = 20},
  [3519] = {.lex_state = 64, .external_lex_state = 20},
  [3520] = {.lex_state = 64, .external_lex_state = 20},
  [3521] = {.lex_state = 64, .external_lex_state = 20},
  [3522] = {.lex_state = 64, .external_lex_state = 20},
  [3523] = {.lex_state = 64, .external_lex_state = 20},
  [3524] = {.lex_state = 64, .external_lex_state = 20},
  [3525] = {.lex_state = 64, .external_lex_state = 20},
  [3526] = {.lex_state = 64, .external_lex_state = 20},
  [3527] = {.lex_state = 64, .external_lex_state = 20},
  [3528] = {.lex_state = 64, .external_lex_state = 20},
  [3529] = {.lex_state = 64, .external_lex_state = 28},
  [3530] = {.lex_state = 64, .external_lex_state = 28},
  [3531] = {.lex_state = 64, .external_lex_state = 28},
  [3532] = {.lex_state = 64, .external_lex_state = 28},
  [3533] = {.lex_state = 64, .external_lex_state = 20},
  [3534] = {.lex_state = 64, .external_lex_state = 28},
  [3535] = {.lex_state = 64, .external_lex_state = 28},
  [3536] = {.lex_state = 64, .external_lex_state = 28},
  [3537] = {.lex_state = 64, .external_lex_state = 28},
  [3538] = {.lex_state = 64, .external_lex_state = 28},
  [3539] = {.lex_state = 64, .external_lex_state = 28},
  [3540] = {.lex_state = 64, .external_lex_state = 28},
  [3541] = {.lex_state = 64, .external_lex_state = 28},
  [3542] = {.lex_state = 64, .external_lex_state = 28},
  [3543] = {.lex_state = 64, .external_lex_state = 28},
  [3544] = {.lex_state = 64, .external_lex_state = 28},
  [3545] = {.lex_state = 64, .external_lex_state = 28},
  [3546] = {.lex_state = 64, .external_lex_state = 28},
  [3547] = {.lex_state = 64, .external_lex_state = 28},
  [3548] = {.lex_state = 64, .external_lex_state = 28},
  [3549] = {.lex_state = 64, .external_lex_state = 28},
  [3550] = {.lex_state = 64, .external_lex_state = 28},
  [3551] = {.lex_state = 64, .external_lex_state = 28},
  [3552] = {.lex_state = 64, .external_lex_state = 28},
  [3553] = {.lex_state = 64, .external_lex_state = 28},
  [3554] = {.lex_state = 64, .external_lex_state = 28},
  [3555] = {.lex_state = 64, .external_lex_state = 28},
  [3556] = {.lex_state = 64, .external_lex_state = 28},
  [3557] = {.lex_state = 64, .external_lex_state = 28},
  [3558] = {.lex_state = 64, .external_lex_state = 28},
  [3559] = {.lex_state = 64, .external_lex_state = 28},
  [3560] = {.lex_state = 64, .external_lex_state = 28},
  [3561] = {.lex_state = 64, .external_lex_state = 28},
  [3562] = {.lex_state = 64, .external_lex_state = 28},
  [3563] = {.lex_state = 64, .external_lex_state = 28},
  [3564] = {.lex_state = 64, .external_lex_state = 28},
  [3565] = {.lex_state = 64, .external_lex_state = 28},
  [3566] = {.lex_state = 64, .external_lex_state = 28},
  [3567] = {.lex_state = 64, .external_lex_state = 28},
  [3568] = {.lex_state = 64, .external_lex_state = 28},
  [3569] = {.lex_state = 64, .external_lex_state = 28},
  [3570] = {.lex_state = 64, .external_lex_state = 28},
  [3571] = {.lex_state = 64, .external_lex_state = 28},
  [3572] = {.lex_state = 64, .external_lex_state = 28},
  [3573] = {.lex_state = 64, .external_lex_state = 28},
  [3574] = {.lex_state = 64, .external_lex_state = 28},
  [3575] = {.lex_state = 64, .external_lex_state = 28},
  [3576] = {.lex_state = 64, .external_lex_state = 28},
  [3577] = {.lex_state = 64, .external_lex_state = 28},
  [3578] = {.lex_state = 64, .external_lex_state = 28},
  [3579] = {.lex_state = 64, .external_lex_state = 28},
  [3580] = {.lex_state = 64, .external_lex_state = 28},
  [3581] = {.lex_state = 64, .external_lex_state = 28},
  [3582] = {.lex_state = 64, .external_lex_state = 28},
  [3583] = {.lex_state = 64, .external_lex_state = 28},
  [3584] = {.lex_state = 64, .external_lex_state = 28},
  [3585] = {.lex_state = 64, .external_lex_state = 28},
  [3586] = {.lex_state = 64, .external_lex_state = 28},
  [3587] = {.lex_state = 64, .external_lex_state = 28},
  [3588] = {.lex_state = 64, .external_lex_state = 28},
  [3589] = {.lex_state = 64, .external_lex_state = 28},
  [3590] = {.lex_state = 64, .external_lex_state = 28},
  [3591] = {.lex_state = 64, .external_lex_state = 28},
  [3592] = {.lex_state = 64, .external_lex_state = 28},
  [3593] = {.lex_state = 64, .external_lex_state = 28},
  [3594] = {.lex_state = 64, .external_lex_state = 28},
  [3595] = {.lex_state = 64, .external_lex_state = 28},
  [3596] = {.lex_state = 64, .external_lex_state = 28},
  [3597] = {.lex_state = 64, .external_lex_state = 28},
  [3598] = {.lex_state = 64, .external_lex_state = 28},
  [3599] = {.lex_state = 64, .external_lex_state = 28},
  [3600] = {.lex_state = 64, .external_lex_state = 28},
  [3601] = {.lex_state = 64, .external_lex_state = 28},
  [3602] = {.lex_state = 64, .external_lex_state = 28},
  [3603] = {.lex_state = 64, .external_lex_state = 28},
  [3604] = {.lex_state = 64, .external_lex_state = 28},
  [3605] = {.lex_state = 64, .external_lex_state = 28},
  [3606] = {.lex_state = 64, .external_lex_state = 28},
  [3607] = {.lex_state = 64, .external_lex_state = 28},
  [3608] = {.lex_state = 64, .external_lex_state = 28},
  [3609] = {.lex_state = 64, .external_lex_state = 28},
  [3610] = {.lex_state = 64, .external_lex_state = 28},
  [3611] = {.lex_state = 64, .external_lex_state = 28},
  [3612] = {.lex_state = 64, .external_lex_state = 28},
  [3613] = {.lex_state = 64, .external_lex_state = 28},
  [3614] = {.lex_state = 64, .external_lex_state = 28},
  [3615] = {.lex_state = 64, .external_lex_state = 28},
  [3616] = {.lex_state = 64, .external_lex_state = 28},
  [3617] = {.lex_state = 64, .external_lex_state = 28},
  [3618] = {.lex_state = 64, .external_lex_state = 28},
  [3619] = {.lex_state = 64, .external_lex_state = 28},
  [3620] = {.lex_state = 64, .external_lex_state = 28},
  [3621] = {.lex_state = 64, .external_lex_state = 28},
  [3622] = {.lex_state = 64, .external_lex_state = 28},
  [3623] = {.lex_state = 64, .external_lex_state = 20},
  [3624] = {.lex_state = 64, .external_lex_state = 20},
  [3625] = {.lex_state = 64, .external_lex_state = 20},
  [3626] = {.lex_state = 64, .external_lex_state = 28},
  [3627] = {.lex_state = 64, .external_lex_state = 28},
  [3628] = {.lex_state = 64, .external_lex_state = 28},
  [3629] = {.lex_state = 64, .external_lex_state = 28},
  [3630] = {.lex_state = 64, .external_lex_state = 28},
  [3631] = {.lex_state = 64, .external_lex_state = 28},
  [3632] = {.lex_state = 64, .external_lex_state = 28},
  [3633] = {.lex_state = 64, .external_lex_state = 28},
  [3634] = {.lex_state = 64, .external_lex_state = 28},
  [3635] = {.lex_state = 64, .external_lex_state = 28},
  [3636] = {.lex_state = 64, .external_lex_state = 28},
  [3637] = {.lex_state = 64, .external_lex_state = 28},
  [3638] = {.lex_state = 64, .external_lex_state = 28},
  [3639] = {.lex_state = 64, .external_lex_state = 28},
  [3640] = {.lex_state = 64, .external_lex_state = 28},
  [3641] = {.lex_state = 64, .external_lex_state = 28},
  [3642] = {.lex_state = 64, .external_lex_state = 28},
  [3643] = {.lex_state = 64, .external_lex_state = 28},
  [3644] = {.lex_state = 64, .external_lex_state = 28},
  [3645] = {.lex_state = 64, .external_lex_state = 28},
  [3646] = {.lex_state = 64, .external_lex_state = 28},
  [3647] = {.lex_state = 64, .external_lex_state = 28},
  [3648] = {.lex_state = 64, .external_lex_state = 28},
  [3649] = {.lex_state = 64, .external_lex_state = 28},
  [3650] = {.lex_state = 64, .external_lex_state = 28},
  [3651] = {.lex_state = 64, .external_lex_state = 28},
  [3652] = {.lex_state = 64, .external_lex_state = 28},
  [3653] = {.lex_state = 64, .external_lex_state = 28},
  [3654] = {.lex_state = 64, .external_lex_state = 28},
  [3655] = {.lex_state = 64, .external_lex_state = 28},
  [3656] = {.lex_state = 64, .external_lex_state = 28},
  [3657] = {.lex_state = 64, .external_lex_state = 28},
  [3658] = {.lex_state = 64, .external_lex_state = 28},
  [3659] = {.lex_state = 64, .external_lex_state = 28},
  [3660] = {.lex_state = 64, .external_lex_state = 28},
  [3661] = {.lex_state = 64, .external_lex_state = 28},
  [3662] = {.lex_state = 64, .external_lex_state = 28},
  [3663] = {.lex_state = 64, .external_lex_state = 28},
  [3664] = {.lex_state = 64, .external_lex_state = 28},
  [3665] = {.lex_state = 64, .external_lex_state = 28},
  [3666] = {.lex_state = 64, .external_lex_state = 28},
  [3667] = {.lex_state = 64, .external_lex_state = 28},
  [3668] = {.lex_state = 64, .external_lex_state = 28},
  [3669] = {.lex_state = 64, .external_lex_state = 28},
  [3670] = {.lex_state = 64, .external_lex_state = 28},
  [3671] = {.lex_state = 64, .external_lex_state = 20},
  [3672] = {.lex_state = 64, .external_lex_state = 20},
  [3673] = {.lex_state = 64, .external_lex_state = 20},
  [3674] = {.lex_state = 64, .external_lex_state = 20},
  [3675] = {.lex_state = 64, .external_lex_state = 20},
  [3676] = {.lex_state = 64, .external_lex_state = 20},
  [3677] = {.lex_state = 64, .external_lex_state = 20},
  [3678] = {.lex_state = 64, .external_lex_state = 20},
  [3679] = {.lex_state = 64, .external_lex_state = 20},
  [3680] = {.lex_state = 64, .external_lex_state = 20},
  [3681] = {.lex_state = 64, .external_lex_state = 20},
  [3682] = {.lex_state = 64, .external_lex_state = 20},
  [3683] = {.lex_state = 64, .external_lex_state = 20},
  [3684] = {.lex_state = 64, .external_lex_state = 20},
  [3685] = {.lex_state = 64, .external_lex_state = 20},
  [3686] = {.lex_state = 64, .external_lex_state = 20},
  [3687] = {.lex_state = 64, .external_lex_state = 28},
  [3688] = {.lex_state = 64, .external_lex_state = 28},
  [3689] = {.lex_state = 64, .external_lex_state = 28},
  [3690] = {.lex_state = 64, .external_lex_state = 20},
  [3691] = {.lex_state = 64, .external_lex_state = 20},
  [3692] = {.lex_state = 64, .external_lex_state = 20},
  [3693] = {.lex_state = 64, .external_lex_state = 20},
  [3694] = {.lex_state = 64, .external_lex_state = 20},
  [3695] = {.lex_state = 64, .external_lex_state = 20},
  [3696] = {.lex_state = 64, .external_lex_state = 20},
  [3697] = {.lex_state = 64, .external_lex_state = 20},
  [3698] = {.lex_state = 64, .external_lex_state = 20},
  [3699] = {.lex_state = 64, .external_lex_state = 20},
  [3700] = {.lex_state = 64, .external_lex_state = 20},
  [3701] = {.lex_state = 64, .external_lex_state = 20},
  [3702] = {.lex_state = 64, .external_lex_state = 20},
  [3703] = {.lex_state = 64, .external_lex_state = 20},
  [3704] = {.lex_state = 64, .external_lex_state = 20},
  [3705] = {.lex_state = 64, .external_lex_state = 20},
  [3706] = {.lex_state = 64, .external_lex_state = 20},
  [3707] = {.lex_state = 64, .external_lex_state = 20},
  [3708] = {.lex_state = 64, .external_lex_state = 20},
  [3709] = {.lex_state = 64, .external_lex_state = 20},
  [3710] = {.lex_state = 64, .external_lex_state = 20},
  [3711] = {.lex_state = 64, .external_lex_state = 20},
  [3712] = {.lex_state = 64, .external_lex_state = 20},
  [3713] = {.lex_state = 64, .external_lex_state = 20},
  [3714] = {.lex_state = 64, .external_lex_state = 20},
  [3715] = {.lex_state = 64, .external_lex_state = 20},
  [3716] = {.lex_state = 64, .external_lex_state = 20},
  [3717] = {.lex_state = 64, .external_lex_state = 20},
  [3718] = {.lex_state = 64, .external_lex_state = 20},
  [3719] = {.lex_state = 64, .external_lex_state = 20},
  [3720] = {.lex_state = 64, .external_lex_state = 20},
  [3721] = {.lex_state = 64, .external_lex_state = 20},
  [3722] = {.lex_state = 64, .external_lex_state = 20},
  [3723] = {.lex_state = 64, .external_lex_state = 20},
  [3724] = {.lex_state = 64, .external_lex_state = 20},
  [3725] = {.lex_state = 64, .external_lex_state = 20},
  [3726] = {.lex_state = 64, .external_lex_state = 20},
  [3727] = {.lex_state = 64, .external_lex_state = 20},
  [3728] = {.lex_state = 64, .external_lex_state = 20},
  [3729] = {.lex_state = 64, .external_lex_state = 20},
  [3730] = {.lex_state = 64, .external_lex_state = 20},
  [3731] = {.lex_state = 64, .external_lex_state = 20},
  [3732] = {.lex_state = 64, .external_lex_state = 20},
  [3733] = {.lex_state = 64, .external_lex_state = 20},
  [3734] = {.lex_state = 64, .external_lex_state = 20},
  [3735] = {.lex_state = 64, .external_lex_state = 20},
  [3736] = {.lex_state = 64, .external_lex_state = 20},
  [3737] = {.lex_state = 64, .external_lex_state = 20},
  [3738] = {.lex_state = 64, .external_lex_state = 20},
  [3739] = {.lex_state = 64, .external_lex_state = 20},
  [3740] = {.lex_state = 64, .external_lex_state = 20},
  [3741] = {.lex_state = 64, .external_lex_state = 20},
  [3742] = {.lex_state = 64, .external_lex_state = 20},
  [3743] = {.lex_state = 64, .external_lex_state = 20},
  [3744] = {.lex_state = 64, .external_lex_state = 20},
  [3745] = {.lex_state = 64, .external_lex_state = 20},
  [3746] = {.lex_state = 64, .external_lex_state = 20},
  [3747] = {.lex_state = 64, .external_lex_state = 20},
  [3748] = {.lex_state = 64, .external_lex_state = 20},
  [3749] = {.lex_state = 64, .external_lex_state = 20},
  [3750] = {.lex_state = 64, .external_lex_state = 20},
  [3751] = {.lex_state = 64, .external_lex_state = 20},
  [3752] = {.lex_state = 64, .external_lex_state = 20},
  [3753] = {.lex_state = 64, .external_lex_state = 20},
  [3754] = {.lex_state = 64, .external_lex_state = 20},
  [3755] = {.lex_state = 64, .external_lex_state = 20},
  [3756] = {.lex_state = 64, .external_lex_state = 20},
  [3757] = {.lex_state = 64, .external_lex_state = 20},
  [3758] = {.lex_state = 64, .external_lex_state = 20},
  [3759] = {.lex_state = 64, .external_lex_state = 20},
  [3760] = {.lex_state = 64, .external_lex_state = 20},
  [3761] = {.lex_state = 64, .external_lex_state = 20},
  [3762] = {.lex_state = 64, .external_lex_state = 20},
  [3763] = {.lex_state = 64, .external_lex_state = 20},
  [3764] = {.lex_state = 64, .external_lex_state = 20},
  [3765] = {.lex_state = 64, .external_lex_state = 20},
  [3766] = {.lex_state = 64, .external_lex_state = 20},
  [3767] = {.lex_state = 64, .external_lex_state = 20},
  [3768] = {.lex_state = 64, .external_lex_state = 20},
  [3769] = {.lex_state = 64, .external_lex_state = 20},
  [3770] = {.lex_state = 64, .external_lex_state = 20},
  [3771] = {.lex_state = 64, .external_lex_state = 20},
  [3772] = {.lex_state = 64, .external_lex_state = 20},
  [3773] = {.lex_state = 64, .external_lex_state = 20},
  [3774] = {.lex_state = 64, .external_lex_state = 20},
  [3775] = {.lex_state = 64, .external_lex_state = 20},
  [3776] = {.lex_state = 64, .external_lex_state = 20},
  [3777] = {.lex_state = 64, .external_lex_state = 20},
  [3778] = {.lex_state = 64, .external_lex_state = 20},
  [3779] = {.lex_state = 64, .external_lex_state = 20},
  [3780] = {.lex_state = 64, .external_lex_state = 20},
  [3781] = {.lex_state = 64, .external_lex_state = 20},
  [3782] = {.lex_state = 64, .external_lex_state = 20},
  [3783] = {.lex_state = 64, .external_lex_state = 20},
  [3784] = {.lex_state = 64, .external_lex_state = 20},
  [3785] = {.lex_state = 64, .external_lex_state = 20},
  [3786] = {.lex_state = 64, .external_lex_state = 20},
  [3787] = {.lex_state = 64, .external_lex_state = 20},
  [3788] = {.lex_state = 64, .external_lex_state = 20},
  [3789] = {.lex_state = 64, .external_lex_state = 20},
  [3790] = {.lex_state = 64, .external_lex_state = 20},
  [3791] = {.lex_state = 64, .external_lex_state = 20},
  [3792] = {.lex_state = 64, .external_lex_state = 20},
  [3793] = {.lex_state = 64, .external_lex_state = 20},
  [3794] = {.lex_state = 64, .external_lex_state = 20},
  [3795] = {.lex_state = 64, .external_lex_state = 20},
  [3796] = {.lex_state = 64, .external_lex_state = 20},
  [3797] = {.lex_state = 64, .external_lex_state = 20},
  [3798] = {.lex_state = 64, .external_lex_state = 20},
  [3799] = {.lex_state = 64, .external_lex_state = 20},
  [3800] = {.lex_state = 64, .external_lex_state = 20},
  [3801] = {.lex_state = 64, .external_lex_state = 20},
  [3802] = {.lex_state = 64, .external_lex_state = 20},
  [3803] = {.lex_state = 64, .external_lex_state = 20},
  [3804] = {.lex_state = 64, .external_lex_state = 20},
  [3805] = {.lex_state = 64, .external_lex_state = 20},
  [3806] = {.lex_state = 64, .external_lex_state = 20},
  [3807] = {.lex_state = 64, .external_lex_state = 20},
  [3808] = {.lex_state = 64, .external_lex_state = 20},
  [3809] = {.lex_state = 64, .external_lex_state = 20},
  [3810] = {.lex_state = 64, .external_lex_state = 20},
  [3811] = {.lex_state = 11, .external_lex_state = 30},
  [3812] = {.lex_state = 11, .external_lex_state = 30},
  [3813] = {.lex_state = 11, .external_lex_state = 30},
  [3814] = {.lex_state = 11, .external_lex_state = 30},
  [3815] = {.lex_state = 11, .external_lex_state = 30},
  [3816] = {.lex_state = 11, .external_lex_state = 30},
  [3817] = {.lex_state = 11, .external_lex_state = 30},
  [3818] = {.lex_state = 11, .external_lex_state = 30},
  [3819] = {.lex_state = 11, .external_lex_state = 30},
  [3820] = {.lex_state = 11, .external_lex_state = 30},
  [3821] = {.lex_state = 11, .external_lex_state = 30},
  [3822] = {.lex_state = 11, .external_lex_state = 30},
  [3823] = {.lex_state = 11, .external_lex_state = 30},
  [3824] = {.lex_state = 11, .external_lex_state = 30},
  [3825] = {.lex_state = 11, .external_lex_state = 30},
  [3826] = {.lex_state = 11, .external_lex_state = 30},
  [3827] = {.lex_state = 11, .external_lex_state = 31},
  [3828] = {.lex_state = 11, .external_lex_state = 31},
  [3829] = {.lex_state = 11, .external_lex_state = 31},
  [3830] = {.lex_state = 11, .external_lex_state = 31},
  [3831] = {.lex_state = 11, .external_lex_state = 31},
  [3832] = {.lex_state = 11, .external_lex_state = 31},
  [3833] = {.lex_state = 11, .external_lex_state = 31},
  [3834] = {.lex_state = 11, .external_lex_state = 31},
  [3835] = {.lex_state = 11, .external_lex_state = 31},
  [3836] = {.lex_state = 11, .external_lex_state = 31},
  [3837] = {.lex_state = 11, .external_lex_state = 31},
  [3838] = {.lex_state = 11, .external_lex_state = 31},
  [3839] = {.lex_state = 11, .external_lex_state = 31},
  [3840] = {.lex_state = 11, .external_lex_state = 31},
  [3841] = {.lex_state = 11, .external_lex_state = 31},
  [3842] = {.lex_state = 11, .external_lex_state = 31},
  [3843] = {.lex_state = 11, .external_lex_state = 31},
  [3844] = {.lex_state = 11, .external_lex_state = 31},
  [3845] = {.lex_state = 11, .external_lex_state = 31},
  [3846] = {.lex_state = 11, .external_lex_state = 31},
  [3847] = {.lex_state = 11, .external_lex_state = 31},
  [3848] = {.lex_state = 11, .external_lex_state = 31},
  [3849] = {.lex_state = 11, .external_lex_state = 31},
  [3850] = {.lex_state = 11, .external_lex_state = 31},
  [3851] = {.lex_state = 11, .external_lex_state = 31},
  [3852] = {.lex_state = 11, .external_lex_state = 31},
  [3853] = {.lex_state = 11, .external_lex_state = 31},
  [3854] = {.lex_state = 11, .external_lex_state = 31},
  [3855] = {.lex_state = 11, .external_lex_state = 31},
  [3856] = {.lex_state = 11, .external_lex_state = 31},
  [3857] = {.lex_state = 11, .external_lex_state = 31},
  [3858] = {.lex_state = 11, .external_lex_state = 31},
  [3859] = {.lex_state = 11, .external_lex_state = 31},
  [3860] = {.lex_state = 11, .external_lex_state = 31},
  [3861] = {.lex_state = 11, .external_lex_state = 31},
  [3862] = {.lex_state = 11, .external_lex_state = 31},
  [3863] = {.lex_state = 11, .external_lex_state = 31},
  [3864] = {.lex_state = 11, .external_lex_state = 31},
  [3865] = {.lex_state = 11, .external_lex_state = 31},
  [3866] = {.lex_state = 11, .external_lex_state = 31},
  [3867] = {.lex_state = 11, .external_lex_state = 31},
  [3868] = {.lex_state = 11, .external_lex_state = 31},
  [3869] = {.lex_state = 11, .external_lex_state = 31},
  [3870] = {.lex_state = 11, .external_lex_state = 31},
  [3871] = {.lex_state = 11, .external_lex_state = 31},
  [3872] = {.lex_state = 11, .external_lex_state = 31},
  [3873] = {.lex_state = 11, .external_lex_state = 31},
  [3874] = {.lex_state = 11, .external_lex_state = 31},
  [3875] = {.lex_state = 11, .external_lex_state = 32},
  [3876] = {.lex_state = 11, .external_lex_state = 32},
  [3877] = {.lex_state = 11, .external_lex_state = 30},
  [3878] = {.lex_state = 11, .external_lex_state = 32},
  [3879] = {.lex_state = 11, .external_lex_state = 32},
  [3880] = {.lex_state = 11, .external_lex_state = 32},
  [3881] = {.lex_state = 11, .external_lex_state = 32},
  [3882] = {.lex_state = 11, .external_lex_state = 32},
  [3883] = {.lex_state = 11, .external_lex_state = 32},
  [3884] = {.lex_state = 11, .external_lex_state = 32},
  [3885] = {.lex_state = 11, .external_lex_state = 32},
  [3886] = {.lex_state = 11, .external_lex_state = 32},
  [3887] = {.lex_state = 11, .external_lex_state = 32},
  [3888] = {.lex_state = 11, .external_lex_state = 32},
  [3889] = {.lex_state = 11, .external_lex_state = 32},
  [3890] = {.lex_state = 11, .external_lex_state = 32},
  [3891] = {.lex_state = 11, .external_lex_state = 32},
  [3892] = {.lex_state = 11, .external_lex_state = 32},
  [3893] = {.lex_state = 11, .external_lex_state = 32},
  [3894] = {.lex_state = 11, .external_lex_state = 31},
  [3895] = {.lex_state = 11, .external_lex_state = 32},
  [3896] = {.lex_state = 11, .external_lex_state = 32},
  [3897] = {.lex_state = 11, .external_lex_state = 32},
  [3898] = {.lex_state = 11, .external_lex_state = 32},
  [3899] = {.lex_state = 11, .external_lex_state = 32},
  [3900] = {.lex_state = 11, .external_lex_state = 32},
  [3901] = {.lex_state = 11, .external_lex_state = 32},
  [3902] = {.lex_state = 11, .external_lex_state = 32},
  [3903] = {.lex_state = 11, .external_lex_state = 32},
  [3904] = {.lex_state = 11, .external_lex_state = 32},
  [3905] = {.lex_state = 11, .external_lex_state = 32},
  [3906] = {.lex_state = 11, .external_lex_state = 32},
  [3907] = {.lex_state = 11, .external_lex_state = 32},
  [3908] = {.lex_state = 11, .external_lex_state = 32},
  [3909] = {.lex_state = 11, .external_lex_state = 32},
  [3910] = {.lex_state = 11, .external_lex_state = 32},
  [3911] = {.lex_state = 11, .external_lex_state = 32},
  [3912] = {.lex_state = 11, .external_lex_state = 32},
  [3913] = {.lex_state = 11, .external_lex_state = 32},
  [3914] = {.lex_state = 11, .external_lex_state = 32},
  [3915] = {.lex_state = 11, .external_lex_state = 32},
  [3916] = {.lex_state = 11, .external_lex_state = 32},
  [3917] = {.lex_state = 11, .external_lex_state = 32},
  [3918] = {.lex_state = 11, .external_lex_state = 32},
  [3919] = {.lex_state = 11, .external_lex_state = 32},
  [3920] = {.lex_state = 11, .external_lex_state = 32},
  [3921] = {.lex_state = 11, .external_lex_state = 32},
  [3922] = {.lex_state = 11, .external_lex_state = 32},
  [3923] = {.lex_state = 11, .external_lex_state = 32},
  [3924] = {.lex_state = 11, .external_lex_state = 32},
  [3925] = {.lex_state = 11, .external_lex_state = 32},
  [3926] = {.lex_state = 11, .external_lex_state = 32},
  [3927] = {.lex_state = 11, .external_lex_state = 32},
  [3928] = {.lex_state = 11, .external_lex_state = 32},
  [3929] = {.lex_state = 11, .external_lex_state = 32},
  [3930] = {.lex_state = 11, .external_lex_state = 32},
  [3931] = {.lex_state = 11, .external_lex_state = 32},
  [3932] = {.lex_state = 11, .external_lex_state = 32},
  [3933] = {.lex_state = 11, .external_lex_state = 32},
  [3934] = {.lex_state = 11, .external_lex_state = 32},
  [3935] = {.lex_state = 11, .external_lex_state = 32},
  [3936] = {.lex_state = 11, .external_lex_state = 32},
  [3937] = {.lex_state = 11, .external_lex_state = 32},
  [3938] = {.lex_state = 11, .external_lex_state = 32},
  [3939] = {.lex_state = 11, .external_lex_state = 32},
  [3940] = {.lex_state = 11, .external_lex_state = 32},
  [3941] = {.lex_state = 11, .external_lex_state = 32},
  [3942] = {.lex_state = 11, .external_lex_state = 32},
  [3943] = {.lex_state = 11, .external_lex_state = 32},
  [3944] = {.lex_state = 11, .external_lex_state = 32},
  [3945] = {.lex_state = 11, .external_lex_state = 32},
  [3946] = {.lex_state = 11, .external_lex_state = 32},
  [3947] = {.lex_state = 11, .external_lex_state = 32},
  [3948] = {.lex_state = 11, .external_lex_state = 32},
  [3949] = {.lex_state = 11, .external_lex_state = 32},
  [3950] = {.lex_state = 11, .external_lex_state = 32},
  [3951] = {.lex_state = 11, .external_lex_state = 32},
  [3952] = {.lex_state = 11, .external_lex_state = 32},
  [3953] = {.lex_state = 11, .external_lex_state = 32},
  [3954] = {.lex_state = 11, .external_lex_state = 32},
  [3955] = {.lex_state = 11, .external_lex_state = 32},
  [3956] = {.lex_state = 11, .external_lex_state = 32},
  [3957] = {.lex_state = 11, .external_lex_state = 32},
  [3958] = {.lex_state = 11, .external_lex_state = 32},
  [3959] = {.lex_state = 11, .external_lex_state = 32},
  [3960] = {.lex_state = 11, .external_lex_state = 32},
  [3961] = {.lex_state = 11, .external_lex_state = 32},
  [3962] = {.lex_state = 11, .external_lex_state = 32},
  [3963] = {.lex_state = 11, .external_lex_state = 32},
  [3964] = {.lex_state = 11, .external_lex_state = 32},
  [3965] = {.lex_state = 11, .external_lex_state = 32},
  [3966] = {.lex_state = 11, .external_lex_state = 32},
  [3967] = {.lex_state = 11, .external_lex_state = 32},
  [3968] = {.lex_state = 11, .external_lex_state = 32},
  [3969] = {.lex_state = 11, .external_lex_state = 32},
  [3970] = {.lex_state = 11, .external_lex_state = 32},
  [3971] = {.lex_state = 11, .external_lex_state = 32},
  [3972] = {.lex_state = 11, .external_lex_state = 32},
  [3973] = {.lex_state = 11, .external_lex_state = 32},
  [3974] = {.lex_state = 11, .external_lex_state = 32},
  [3975] = {.lex_state = 11, .external_lex_state = 32},
  [3976] = {.lex_state = 11, .external_lex_state = 32},
  [3977] = {.lex_state = 11, .external_lex_state = 32},
  [3978] = {.lex_state = 11, .external_lex_state = 32},
  [3979] = {.lex_state = 11, .external_lex_state = 32},
  [3980] = {.lex_state = 11, .external_lex_state = 32},
  [3981] = {.lex_state = 11, .external_lex_state = 32},
  [3982] = {.lex_state = 11, .external_lex_state = 33},
  [3983] = {.lex_state = 11, .external_lex_state = 34},
  [3984] = {.lex_state = 11, .external_lex_state = 34},
  [3985] = {.lex_state = 11, .external_lex_state = 34},
  [3986] = {.lex_state = 11, .external_lex_state = 34},
  [3987] = {.lex_state = 11, .external_lex_state = 34},
  [3988] = {.lex_state = 11, .external_lex_state = 34},
  [3989] = {.lex_state = 11, .external_lex_state = 34},
  [3990] = {.lex_state = 11, .external_lex_state = 34},
  [3991] = {.lex_state = 11, .external_lex_state = 33},
  [3992] = {.lex_state = 11, .external_lex_state = 33},
  [3993] = {.lex_state = 11, .external_lex_state = 18},
  [3994] = {.lex_state = 11, .external_lex_state = 18},
  [3995] = {.lex_state = 11, .external_lex_state = 18},
  [3996] = {.lex_state = 11, .external_lex_state = 18},
  [3997] = {.lex_state = 11, .external_lex_state = 18},
  [3998] = {.lex_state = 11, .external_lex_state = 18},
  [3999] = {.lex_state = 11, .external_lex_state = 18},
  [4000] = {.lex_state = 11, .external_lex_state = 18},
  [4001] = {.lex_state = 11, .external_lex_state = 18},
  [4002] = {.lex_state = 11, .external_lex_state = 18},
  [4003] = {.lex_state = 11, .external_lex_state = 18},
  [4004] = {.lex_state = 11, .external_lex_state = 18},
  [4005] = {.lex_state = 11, .external_lex_state = 18},
  [4006] = {.lex_state = 11, .external_lex_state = 18},
  [4007] = {.lex_state = 11, .external_lex_state = 18},
  [4008] = {.lex_state = 11, .external_lex_state = 18},
  [4009] = {.lex_state = 11, .external_lex_state = 18},
  [4010] = {.lex_state = 11, .external_lex_state = 18},
  [4011] = {.lex_state = 11, .external_lex_state = 18},
  [4012] = {.lex_state = 11, .external_lex_state = 18},
  [4013] = {.lex_state = 11, .external_lex_state = 18},
  [4014] = {.lex_state = 77, .external_lex_state = 35},
  [4015] = {.lex_state = 77, .external_lex_state = 35},
  [4016] = {.lex_state = 77, .external_lex_state = 35},
  [4017] = {.lex_state = 77, .external_lex_state = 35},
  [4018] = {.lex_state = 11, .external_lex_state = 18},
  [4019] = {.lex_state = 11, .external_lex_state = 18},
  [4020] = {.lex_state = 11, .external_lex_state = 18},
  [4021] = {.lex_state = 11, .external_lex_state = 18},
  [4022] = {.lex_state = 13, .external_lex_state = 20},
  [4023] = {.lex_state = 72, .external_lex_state = 36},
  [4024] = {.lex_state = 13, .external_lex_state = 31},
  [4025] = {.lex_state = 72, .external_lex_state = 36},
  [4026] = {.lex_state = 72, .external_lex_state = 36},
  [4027] = {.lex_state = 13, .external_lex_state = 37},
  [4028] = {.lex_state = 72, .external_lex_state = 36},
  [4029] = {.lex_state = 13, .external_lex_state = 30},
  [4030] = {.lex_state = 57, .external_lex_state = 20},
  [4031] = {.lex_state = 13, .external_lex_state = 37},
  [4032] = {.lex_state = 72, .external_lex_state = 36},
  [4033] = {.lex_state = 13, .external_lex_state = 37},
  [4034] = {.lex_state = 13, .external_lex_state = 38},
  [4035] = {.lex_state = 13, .external_lex_state = 37},
  [4036] = {.lex_state = 57, .external_lex_state = 20},
  [4037] = {.lex_state = 13, .external_lex_state = 38},
  [4038] = {.lex_state = 13, .external_lex_state = 37},
  [4039] = {.lex_state = 13, .external_lex_state = 38},
  [4040] = {.lex_state = 72, .external_lex_state = 36},
  [4041] = {.lex_state = 72, .external_lex_state = 36},
  [4042] = {.lex_state = 72, .external_lex_state = 36},
  [4043] = {.lex_state = 13, .external_lex_state = 38},
  [4044] = {.lex_state = 57, .external_lex_state = 20},
  [4045] = {.lex_state = 13, .external_lex_state = 38},
  [4046] = {.lex_state = 13, .external_lex_state = 20},
  [4047] = {.lex_state = 13, .external_lex_state = 20},
  [4048] = {.lex_state = 76, .external_lex_state = 39},
  [4049] = {.lex_state = 76, .external_lex_state = 39},
  [4050] = {.lex_state = 13, .external_lex_state = 40},
  [4051] = {.lex_state = 13, .external_lex_state = 40},
  [4052] = {.lex_state = 76, .external_lex_state = 39},
  [4053] = {.lex_state = 76, .external_lex_state = 39},
  [4054] = {.lex_state = 13, .external_lex_state = 38},
  [4055] = {.lex_state = 11, .external_lex_state = 41},
  [4056] = {.lex_state = 76, .external_lex_state = 39},
  [4057] = {.lex_state = 13, .external_lex_state = 31},
  [4058] = {.lex_state = 76, .external_lex_state = 39},
  [4059] = {.lex_state = 11, .external_lex_state = 41},
  [4060] = {.lex_state = 76, .external_lex_state = 39},
  [4061] = {.lex_state = 11, .external_lex_state = 41},
  [4062] = {.lex_state = 76, .external_lex_state = 39},
  [4063] = {.lex_state = 76, .external_lex_state = 39},
  [4064] = {.lex_state = 11, .external_lex_state = 41},
  [4065] = {.lex_state = 76, .external_lex_state = 39},
  [4066] = {.lex_state = 76, .external_lex_state = 39},
  [4067] = {.lex_state = 76, .external_lex_state = 39},
  [4068] = {.lex_state = 13, .external_lex_state = 30},
  [4069] = {.lex_state = 76, .external_lex_state = 39},
  [4070] = {.lex_state = 13, .external_lex_state = 30},
  [4071] = {.lex_state = 76, .external_lex_state = 39},
  [4072] = {.lex_state = 76, .external_lex_state = 39},
  [4073] = {.lex_state = 76, .external_lex_state = 39},
  [4074] = {.lex_state = 76, .external_lex_state = 39},
  [4075] = {.lex_state = 11, .external_lex_state = 41},
  [4076] = {.lex_state = 76, .external_lex_state = 39},
  [4077] = {.lex_state = 76, .external_lex_state = 39},
  [4078] = {.lex_state = 13, .external_lex_state = 32},
  [4079] = {.lex_state = 76, .external_lex_state = 42},
  [4080] = {.lex_state = 76, .external_lex_state = 39},
  [4081] = {.lex_state = 13, .external_lex_state = 40},
  [4082] = {.lex_state = 76, .external_lex_state = 39},
  [4083] = {.lex_state = 13, .external_lex_state = 30},
  [4084] = {.lex_state = 76, .external_lex_state = 39},
  [4085] = {.lex_state = 13, .external_lex_state = 31},
  [4086] = {.lex_state = 13, .external_lex_state = 31},
  [4087] = {.lex_state = 13, .external_lex_state = 37},
  [4088] = {.lex_state = 13, .external_lex_state = 31},
  [4089] = {.lex_state = 76, .external_lex_state = 42},
  [4090] = {.lex_state = 13, .external_lex_state = 31},
  [4091] = {.lex_state = 75, .external_lex_state = 20},
  [4092] = {.lex_state = 76, .external_lex_state = 42},
  [4093] = {.lex_state = 75, .external_lex_state = 20},
  [4094] = {.lex_state = 76, .external_lex_state = 39},
  [4095] = {.lex_state = 13, .external_lex_state = 30},
  [4096] = {.lex_state = 11, .external_lex_state = 41},
  [4097] = {.lex_state = 75, .external_lex_state = 20},
  [4098] = {.lex_state = 57, .external_lex_state = 20},
  [4099] = {.lex_state = 76, .external_lex_state = 39},
  [4100] = {.lex_state = 13, .external_lex_state = 31},
  [4101] = {.lex_state = 11, .external_lex_state = 41},
  [4102] = {.lex_state = 13, .external_lex_state = 30},
  [4103] = {.lex_state = 75, .external_lex_state = 20},
  [4104] = {.lex_state = 13, .external_lex_state = 31},
  [4105] = {.lex_state = 11, .external_lex_state = 41},
  [4106] = {.lex_state = 76, .external_lex_state = 39},
  [4107] = {.lex_state = 13, .external_lex_state = 37},
  [4108] = {.lex_state = 13, .external_lex_state = 40},
  [4109] = {.lex_state = 13, .external_lex_state = 38},
  [4110] = {.lex_state = 11, .external_lex_state = 41},
  [4111] = {.lex_state = 76, .external_lex_state = 39},
  [4112] = {.lex_state = 11, .external_lex_state = 41},
  [4113] = {.lex_state = 76, .external_lex_state = 39},
  [4114] = {.lex_state = 13, .external_lex_state = 30},
  [4115] = {.lex_state = 11, .external_lex_state = 41},
  [4116] = {.lex_state = 13, .external_lex_state = 30},
  [4117] = {.lex_state = 11, .external_lex_state = 41},
  [4118] = {.lex_state = 11, .external_lex_state = 41},
  [4119] = {.lex_state = 75, .external_lex_state = 20},
  [4120] = {.lex_state = 11, .external_lex_state = 41},
  [4121] = {.lex_state = 76, .external_lex_state = 42},
  [4122] = {.lex_state = 76, .external_lex_state = 39},
  [4123] = {.lex_state = 13, .external_lex_state = 40},
  [4124] = {.lex_state = 76, .external_lex_state = 39},
  [4125] = {.lex_state = 76, .external_lex_state = 42},
  [4126] = {.lex_state = 76, .external_lex_state = 42},
  [4127] = {.lex_state = 75, .external_lex_state = 20},
  [4128] = {.lex_state = 11, .external_lex_state = 41},
  [4129] = {.lex_state = 76, .external_lex_state = 39},
  [4130] = {.lex_state = 75, .external_lex_state = 20},
  [4131] = {.lex_state = 75, .external_lex_state = 20},
  [4132] = {.lex_state = 76, .external_lex_state = 39},
  [4133] = {.lex_state = 76, .external_lex_state = 39},
  [4134] = {.lex_state = 11, .external_lex_state = 41},
  [4135] = {.lex_state = 13, .external_lex_state = 30},
  [4136] = {.lex_state = 76, .external_lex_state = 39},
  [4137] = {.lex_state = 76, .external_lex_state = 39},
  [4138] = {.lex_state = 13, .external_lex_state = 31},
  [4139] = {.lex_state = 76, .external_lex_state = 20},
  [4140] = {.lex_state = 76, .external_lex_state = 20},
  [4141] = {.lex_state = 76, .external_lex_state = 20},
  [4142] = {.lex_state = 13, .external_lex_state = 31},
  [4143] = {.lex_state = 76, .external_lex_state = 20},
  [4144] = {.lex_state = 76, .external_lex_state = 20},
  [4145] = {.lex_state = 76, .external_lex_state = 20},
  [4146] = {.lex_state = 76, .external_lex_state = 20},
  [4147] = {.lex_state = 76, .external_lex_state = 20},
  [4148] = {.lex_state = 76, .external_lex_state = 20},
  [4149] = {.lex_state = 13, .external_lex_state = 43},
  [4150] = {.lex_state = 11, .external_lex_state = 32},
  [4151] = {.lex_state = 13, .external_lex_state = 33},
  [4152] = {.lex_state = 13, .external_lex_state = 20},
  [4153] = {.lex_state = 13, .external_lex_state = 31},
  [4154] = {.lex_state = 76, .external_lex_state = 20},
  [4155] = {.lex_state = 11, .external_lex_state = 32},
  [4156] = {.lex_state = 13, .external_lex_state = 30},
  [4157] = {.lex_state = 76, .external_lex_state = 20},
  [4158] = {.lex_state = 13, .external_lex_state = 43},
  [4159] = {.lex_state = 76, .external_lex_state = 20},
  [4160] = {.lex_state = 76, .external_lex_state = 20},
  [4161] = {.lex_state = 76, .external_lex_state = 20},
  [4162] = {.lex_state = 13, .external_lex_state = 31},
  [4163] = {.lex_state = 11, .external_lex_state = 32},
  [4164] = {.lex_state = 11, .external_lex_state = 32},
  [4165] = {.lex_state = 13, .external_lex_state = 32},
  [4166] = {.lex_state = 11, .external_lex_state = 18},
  [4167] = {.lex_state = 11, .external_lex_state = 32},
  [4168] = {.lex_state = 13, .external_lex_state = 40},
  [4169] = {.lex_state = 13, .external_lex_state = 31},
  [4170] = {.lex_state = 13, .external_lex_state = 31},
  [4171] = {.lex_state = 76, .external_lex_state = 20},
  [4172] = {.lex_state = 13, .external_lex_state = 31},
  [4173] = {.lex_state = 11, .external_lex_state = 32},
  [4174] = {.lex_state = 76, .external_lex_state = 20},
  [4175] = {.lex_state = 11, .external_lex_state = 32},
  [4176] = {.lex_state = 76, .external_lex_state = 20},
  [4177] = {.lex_state = 13, .external_lex_state = 20},
  [4178] = {.lex_state = 11, .external_lex_state = 32},
  [4179] = {.lex_state = 11, .external_lex_state = 32},
  [4180] = {.lex_state = 76, .external_lex_state = 20},
  [4181] = {.lex_state = 76, .external_lex_state = 20},
  [4182] = {.lex_state = 13, .external_lex_state = 20},
  [4183] = {.lex_state = 13, .external_lex_state = 32},
  [4184] = {.lex_state = 76, .external_lex_state = 20},
  [4185] = {.lex_state = 76, .external_lex_state = 20},
  [4186] = {.lex_state = 76, .external_lex_state = 20},
  [4187] = {.lex_state = 11, .external_lex_state = 32},
  [4188] = {.lex_state = 76, .external_lex_state = 20},
  [4189] = {.lex_state = 76, .external_lex_state = 20},
  [4190] = {.lex_state = 76, .external_lex_state = 20},
  [4191] = {.lex_state = 76, .external_lex_state = 20},
  [4192] = {.lex_state = 13, .external_lex_state = 32},
  [4193] = {.lex_state = 76, .external_lex_state = 20},
  [4194] = {.lex_state = 11, .external_lex_state = 32},
  [4195] = {.lex_state = 13, .external_lex_state = 20},
  [4196] = {.lex_state = 76, .external_lex_state = 20},
  [4197] = {.lex_state = 76, .external_lex_state = 20},
  [4198] = {.lex_state = 76, .external_lex_state = 20},
  [4199] = {.lex_state = 76, .external_lex_state = 20},
  [4200] = {.lex_state = 76, .external_lex_state = 20},
  [4201] = {.lex_state = 76, .external_lex_state = 20},
  [4202] = {.lex_state = 13, .external_lex_state = 31},
  [4203] = {.lex_state = 13, .external_lex_state = 31},
  [4204] = {.lex_state = 11, .external_lex_state = 32},
  [4205] = {.lex_state = 13, .external_lex_state = 20},
  [4206] = {.lex_state = 76, .external_lex_state = 20},
  [4207] = {.lex_state = 76, .external_lex_state = 20},
  [4208] = {.lex_state = 13, .external_lex_state = 20},
  [4209] = {.lex_state = 13, .external_lex_state = 31},
  [4210] = {.lex_state = 13, .external_lex_state = 40},
  [4211] = {.lex_state = 76, .external_lex_state = 20},
  [4212] = {.lex_state = 13, .external_lex_state = 31},
  [4213] = {.lex_state = 76, .external_lex_state = 20},
  [4214] = {.lex_state = 13, .external_lex_state = 43},
  [4215] = {.lex_state = 76, .external_lex_state = 20},
  [4216] = {.lex_state = 76, .external_lex_state = 20},
  [4217] = {.lex_state = 76, .external_lex_state = 20},
  [4218] = {.lex_state = 76, .external_lex_state = 20},
  [4219] = {.lex_state = 13, .external_lex_state = 43},
  [4220] = {.lex_state = 76, .external_lex_state = 20},
  [4221] = {.lex_state = 13, .external_lex_state = 20},
  [4222] = {.lex_state = 13, .external_lex_state = 32},
  [4223] = {.lex_state = 11, .external_lex_state = 32},
  [4224] = {.lex_state = 11, .external_lex_state = 32},
  [4225] = {.lex_state = 11, .external_lex_state = 32},
  [4226] = {.lex_state = 76, .external_lex_state = 20},
  [4227] = {.lex_state = 13, .external_lex_state = 30},
  [4228] = {.lex_state = 76, .external_lex_state = 20},
  [4229] = {.lex_state = 13, .external_lex_state = 30},
  [4230] = {.lex_state = 13, .external_lex_state = 20},
  [4231] = {.lex_state = 76, .external_lex_state = 20},
  [4232] = {.lex_state = 13, .external_lex_state = 20},
  [4233] = {.lex_state = 76, .external_lex_state = 20},
  [4234] = {.lex_state = 13, .external_lex_state = 20},
  [4235] = {.lex_state = 13, .external_lex_state = 20},
  [4236] = {.lex_state = 76, .external_lex_state = 20},
  [4237] = {.lex_state = 11, .external_lex_state = 32},
  [4238] = {.lex_state = 13, .external_lex_state = 20},
  [4239] = {.lex_state = 13, .external_lex_state = 20},
  [4240] = {.lex_state = 76, .external_lex_state = 20},
  [4241] = {.lex_state = 13, .external_lex_state = 30},
  [4242] = {.lex_state = 76, .external_lex_state = 20},
  [4243] = {.lex_state = 76, .external_lex_state = 20},
  [4244] = {.lex_state = 13, .external_lex_state = 32},
  [4245] = {.lex_state = 11, .external_lex_state = 32},
  [4246] = {.lex_state = 13, .external_lex_state = 31},
  [4247] = {.lex_state = 76, .external_lex_state = 20},
  [4248] = {.lex_state = 13, .external_lex_state = 20},
  [4249] = {.lex_state = 76, .external_lex_state = 20},
  [4250] = {.lex_state = 76, .external_lex_state = 20},
  [4251] = {.lex_state = 11, .external_lex_state = 18},
  [4252] = {.lex_state = 76, .external_lex_state = 20},
  [4253] = {.lex_state = 13, .external_lex_state = 20},
  [4254] = {.lex_state = 13, .external_lex_state = 30},
  [4255] = {.lex_state = 76, .external_lex_state = 20},
  [4256] = {.lex_state = 13, .external_lex_state = 30},
  [4257] = {.lex_state = 13, .external_lex_state = 30},
  [4258] = {.lex_state = 13, .external_lex_state = 30},
  [4259] = {.lex_state = 76, .external_lex_state = 20},
  [4260] = {.lex_state = 76, .external_lex_state = 20},
  [4261] = {.lex_state = 11, .external_lex_state = 32},
  [4262] = {.lex_state = 76, .external_lex_state = 20},
  [4263] = {.lex_state = 11, .external_lex_state = 32},
  [4264] = {.lex_state = 11, .external_lex_state = 32},
  [4265] = {.lex_state = 13, .external_lex_state = 43},
  [4266] = {.lex_state = 11, .external_lex_state = 32},
  [4267] = {.lex_state = 11, .external_lex_state = 32},
  [4268] = {.lex_state = 11, .external_lex_state = 32},
  [4269] = {.lex_state = 11, .external_lex_state = 32},
  [4270] = {.lex_state = 11, .external_lex_state = 32},
  [4271] = {.lex_state = 76, .external_lex_state = 20},
  [4272] = {.lex_state = 11, .external_lex_state = 32},
  [4273] = {.lex_state = 76, .external_lex_state = 20},
  [4274] = {.lex_state = 76, .external_lex_state = 20},
  [4275] = {.lex_state = 76, .external_lex_state = 20},
  [4276] = {.lex_state = 76, .external_lex_state = 20},
  [4277] = {.lex_state = 76, .external_lex_state = 20},
  [4278] = {.lex_state = 76, .external_lex_state = 20},
  [4279] = {.lex_state = 76, .external_lex_state = 20},
  [4280] = {.lex_state = 76, .external_lex_state = 20},
  [4281] = {.lex_state = 76, .external_lex_state = 20},
  [4282] = {.lex_state = 13, .external_lex_state = 30},
  [4283] = {.lex_state = 11, .external_lex_state = 32},
  [4284] = {.lex_state = 76, .external_lex_state = 20},
  [4285] = {.lex_state = 76, .external_lex_state = 20},
  [4286] = {.lex_state = 11, .external_lex_state = 32},
  [4287] = {.lex_state = 11, .external_lex_state = 32},
  [4288] = {.lex_state = 13, .external_lex_state = 32},
  [4289] = {.lex_state = 76, .external_lex_state = 20},
  [4290] = {.lex_state = 13, .external_lex_state = 20},
  [4291] = {.lex_state = 11, .external_lex_state = 32},
  [4292] = {.lex_state = 13, .external_lex_state = 32},
  [4293] = {.lex_state = 13, .external_lex_state = 32},
  [4294] = {.lex_state = 76, .external_lex_state = 20},
  [4295] = {.lex_state = 11, .external_lex_state = 32},
  [4296] = {.lex_state = 76, .external_lex_state = 20},
  [4297] = {.lex_state = 11, .external_lex_state = 32},
  [4298] = {.lex_state = 76, .external_lex_state = 20},
  [4299] = {.lex_state = 13, .external_lex_state = 20},
  [4300] = {.lex_state = 13, .external_lex_state = 30},
  [4301] = {.lex_state = 13, .external_lex_state = 30},
  [4302] = {.lex_state = 76, .external_lex_state = 20},
  [4303] = {.lex_state = 72, .external_lex_state = 28},
  [4304] = {.lex_state = 72, .external_lex_state = 44},
  [4305] = {.lex_state = 72, .external_lex_state = 44},
  [4306] = {.lex_state = 72, .external_lex_state = 44},
  [4307] = {.lex_state = 72, .external_lex_state = 44},
  [4308] = {.lex_state = 13, .external_lex_state = 32},
  [4309] = {.lex_state = 13, .external_lex_state = 32},
  [4310] = {.lex_state = 13, .external_lex_state = 32},
  [4311] = {.lex_state = 13, .external_lex_state = 32},
  [4312] = {.lex_state = 72, .external_lex_state = 44},
  [4313] = {.lex_state = 72, .external_lex_state = 44},
  [4314] = {.lex_state = 13, .external_lex_state = 32},
  [4315] = {.lex_state = 72, .external_lex_state = 28},
  [4316] = {.lex_state = 72, .external_lex_state = 44},
  [4317] = {.lex_state = 72, .external_lex_state = 44},
  [4318] = {.lex_state = 72, .external_lex_state = 28},
  [4319] = {.lex_state = 72, .external_lex_state = 44},
  [4320] = {.lex_state = 72, .external_lex_state = 44},
  [4321] = {.lex_state = 72, .external_lex_state = 44},
  [4322] = {.lex_state = 72, .external_lex_state = 44},
  [4323] = {.lex_state = 72, .external_lex_state = 44},
  [4324] = {.lex_state = 72, .external_lex_state = 44},
  [4325] = {.lex_state = 13, .external_lex_state = 33},
  [4326] = {.lex_state = 13, .external_lex_state = 33},
  [4327] = {.lex_state = 72, .external_lex_state = 44},
  [4328] = {.lex_state = 72, .external_lex_state = 44},
  [4329] = {.lex_state = 13, .external_lex_state = 33},
  [4330] = {.lex_state = 13, .external_lex_state = 33},
  [4331] = {.lex_state = 13, .external_lex_state = 33},
  [4332] = {.lex_state = 72, .external_lex_state = 44},
  [4333] = {.lex_state = 72, .external_lex_state = 44},
  [4334] = {.lex_state = 72, .external_lex_state = 44},
  [4335] = {.lex_state = 72, .external_lex_state = 44},
  [4336] = {.lex_state = 72, .external_lex_state = 44},
  [4337] = {.lex_state = 72, .external_lex_state = 44},
  [4338] = {.lex_state = 72, .external_lex_state = 44},
  [4339] = {.lex_state = 72, .external_lex_state = 44},
  [4340] = {.lex_state = 72, .external_lex_state = 44},
  [4341] = {.lex_state = 72, .external_lex_state = 44},
  [4342] = {.lex_state = 72, .external_lex_state = 44},
  [4343] = {.lex_state = 72, .external_lex_state = 44},
  [4344] = {.lex_state = 72, .external_lex_state = 44},
  [4345] = {.lex_state = 72, .external_lex_state = 44},
  [4346] = {.lex_state = 72, .external_lex_state = 44},
  [4347] = {.lex_state = 72, .external_lex_state = 44},
  [4348] = {.lex_state = 72, .external_lex_state = 44},
  [4349] = {.lex_state = 72, .external_lex_state = 44},
  [4350] = {.lex_state = 11, .external_lex_state = 41},
  [4351] = {.lex_state = 72, .external_lex_state = 28},
  [4352] = {.lex_state = 72, .external_lex_state = 28},
  [4353] = {.lex_state = 72, .external_lex_state = 28},
  [4354] = {.lex_state = 72, .external_lex_state = 28},
  [4355] = {.lex_state = 72, .external_lex_state = 28},
  [4356] = {.lex_state = 72, .external_lex_state = 44},
  [4357] = {.lex_state = 72, .external_lex_state = 44},
  [4358] = {.lex_state = 13, .external_lex_state = 33},
  [4359] = {.lex_state = 72, .external_lex_state = 44},
  [4360] = {.lex_state = 72, .external_lex_state = 44},
  [4361] = {.lex_state = 13, .external_lex_state = 33},
  [4362] = {.lex_state = 72, .external_lex_state = 28},
  [4363] = {.lex_state = 72, .external_lex_state = 44},
  [4364] = {.lex_state = 72, .external_lex_state = 44},
  [4365] = {.lex_state = 72, .external_lex_state = 44},
  [4366] = {.lex_state = 72, .external_lex_state = 44},
  [4367] = {.lex_state = 72, .external_lex_state = 28},
  [4368] = {.lex_state = 72, .external_lex_state = 44},
  [4369] = {.lex_state = 72, .external_lex_state = 44},
  [4370] = {.lex_state = 13, .external_lex_state = 33},
  [4371] = {.lex_state = 13, .external_lex_state = 43},
  [4372] = {.lex_state = 13, .external_lex_state = 43},
  [4373] = {.lex_state = 72, .external_lex_state = 44},
  [4374] = {.lex_state = 11, .external_lex_state = 32},
  [4375] = {.lex_state = 72, .external_lex_state = 44},
  [4376] = {.lex_state = 72, .external_lex_state = 44},
  [4377] = {.lex_state = 72, .external_lex_state = 44},
  [4378] = {.lex_state = 72, .external_lex_state = 44},
  [4379] = {.lex_state = 72, .external_lex_state = 44},
  [4380] = {.lex_state = 72, .external_lex_state = 44},
  [4381] = {.lex_state = 72, .external_lex_state = 44},
  [4382] = {.lex_state = 13, .external_lex_state = 32},
  [4383] = {.lex_state = 13, .external_lex_state = 32},
  [4384] = {.lex_state = 13, .external_lex_state = 32},
  [4385] = {.lex_state = 13, .external_lex_state = 32},
  [4386] = {.lex_state = 72, .external_lex_state = 44},
  [4387] = {.lex_state = 72, .external_lex_state = 28},
  [4388] = {.lex_state = 72, .external_lex_state = 44},
  [4389] = {.lex_state = 13, .external_lex_state = 20},
  [4390] = {.lex_state = 13, .external_lex_state = 32},
  [4391] = {.lex_state = 13, .external_lex_state = 32},
  [4392] = {.lex_state = 72, .external_lex_state = 28},
  [4393] = {.lex_state = 72, .external_lex_state = 44},
  [4394] = {.lex_state = 72, .external_lex_state = 44},
  [4395] = {.lex_state = 11, .external_lex_state = 31},
  [4396] = {.lex_state = 11, .external_lex_state = 31},
  [4397] = {.lex_state = 72, .external_lex_state = 44},
  [4398] = {.lex_state = 72, .external_lex_state = 44},
  [4399] = {.lex_state = 11, .external_lex_state = 30},
  [4400] = {.lex_state = 11, .external_lex_state = 31},
  [4401] = {.lex_state = 11, .external_lex_state = 30},
  [4402] = {.lex_state = 11, .external_lex_state = 31},
  [4403] = {.lex_state = 72, .external_lex_state = 44},
  [4404] = {.lex_state = 72, .external_lex_state = 44},
  [4405] = {.lex_state = 80, .external_lex_state = 28},
  [4406] = {.lex_state = 11, .external_lex_state = 31},
  [4407] = {.lex_state = 11, .external_lex_state = 31},
  [4408] = {.lex_state = 11, .external_lex_state = 31},
  [4409] = {.lex_state = 13, .external_lex_state = 33},
  [4410] = {.lex_state = 11, .external_lex_state = 31},
  [4411] = {.lex_state = 13, .external_lex_state = 33},
  [4412] = {.lex_state = 11, .external_lex_state = 30},
  [4413] = {.lex_state = 11, .external_lex_state = 30},
  [4414] = {.lex_state = 11, .external_lex_state = 30},
  [4415] = {.lex_state = 11, .external_lex_state = 30},
  [4416] = {.lex_state = 13, .external_lex_state = 33},
  [4417] = {.lex_state = 11, .external_lex_state = 31},
  [4418] = {.lex_state = 11, .external_lex_state = 31},
  [4419] = {.lex_state = 11, .external_lex_state = 30},
  [4420] = {.lex_state = 11, .external_lex_state = 30},
  [4421] = {.lex_state = 80, .external_lex_state = 28},
  [4422] = {.lex_state = 80, .external_lex_state = 28},
  [4423] = {.lex_state = 72, .external_lex_state = 45},
  [4424] = {.lex_state = 11, .external_lex_state = 30},
  [4425] = {.lex_state = 11, .external_lex_state = 30},
  [4426] = {.lex_state = 11, .external_lex_state = 30},
  [4427] = {.lex_state = 72, .external_lex_state = 44},
  [4428] = {.lex_state = 11, .external_lex_state = 31},
  [4429] = {.lex_state = 72, .external_lex_state = 44},
  [4430] = {.lex_state = 72, .external_lex_state = 44},
  [4431] = {.lex_state = 13, .external_lex_state = 33},
  [4432] = {.lex_state = 11, .external_lex_state = 31},
  [4433] = {.lex_state = 13, .external_lex_state = 33},
  [4434] = {.lex_state = 72, .external_lex_state = 44},
  [4435] = {.lex_state = 11, .external_lex_state = 31},
  [4436] = {.lex_state = 11, .external_lex_state = 31},
  [4437] = {.lex_state = 72, .external_lex_state = 44},
  [4438] = {.lex_state = 72, .external_lex_state = 44},
  [4439] = {.lex_state = 11, .external_lex_state = 30},
  [4440] = {.lex_state = 11, .external_lex_state = 30},
  [4441] = {.lex_state = 11, .external_lex_state = 31},
  [4442] = {.lex_state = 11, .external_lex_state = 31},
  [4443] = {.lex_state = 72, .external_lex_state = 45},
  [4444] = {.lex_state = 11, .external_lex_state = 30},
  [4445] = {.lex_state = 11, .external_lex_state = 30},
  [4446] = {.lex_state = 11, .external_lex_state = 30},
  [4447] = {.lex_state = 11, .external_lex_state = 30},
  [4448] = {.lex_state = 11, .external_lex_state = 31},
  [4449] = {.lex_state = 11, .external_lex_state = 31},
  [4450] = {.lex_state = 72, .external_lex_state = 44},
  [4451] = {.lex_state = 13, .external_lex_state = 33},
  [4452] = {.lex_state = 11, .external_lex_state = 31},
  [4453] = {.lex_state = 13, .external_lex_state = 33},
  [4454] = {.lex_state = 11, .external_lex_state = 30},
  [4455] = {.lex_state = 11, .external_lex_state = 30},
  [4456] = {.lex_state = 11, .external_lex_state = 31},
  [4457] = {.lex_state = 11, .external_lex_state = 31},
  [4458] = {.lex_state = 13, .external_lex_state = 33},
  [4459] = {.lex_state = 13, .external_lex_state = 33},
  [4460] = {.lex_state = 11, .external_lex_state = 31},
  [4461] = {.lex_state = 11, .external_lex_state = 31},
  [4462] = {.lex_state = 11, .external_lex_state = 30},
  [4463] = {.lex_state = 13, .external_lex_state = 33},
  [4464] = {.lex_state = 13, .external_lex_state = 33},
  [4465] = {.lex_state = 11, .external_lex_state = 30},
  [4466] = {.lex_state = 80, .external_lex_state = 28},
  [4467] = {.lex_state = 80, .external_lex_state = 28},
  [4468] = {.lex_state = 72, .external_lex_state = 45},
  [4469] = {.lex_state = 11, .external_lex_state = 31},
  [4470] = {.lex_state = 11, .external_lex_state = 30},
  [4471] = {.lex_state = 11, .external_lex_state = 31},
  [4472] = {.lex_state = 11, .external_lex_state = 30},
  [4473] = {.lex_state = 11, .external_lex_state = 31},
  [4474] = {.lex_state = 11, .external_lex_state = 30},
  [4475] = {.lex_state = 11, .external_lex_state = 30},
  [4476] = {.lex_state = 11, .external_lex_state = 31},
  [4477] = {.lex_state = 11, .external_lex_state = 31},
  [4478] = {.lex_state = 72, .external_lex_state = 44},
  [4479] = {.lex_state = 13, .external_lex_state = 20},
  [4480] = {.lex_state = 80, .external_lex_state = 28},
  [4481] = {.lex_state = 11, .external_lex_state = 30},
  [4482] = {.lex_state = 11, .external_lex_state = 31},
  [4483] = {.lex_state = 11, .external_lex_state = 30},
  [4484] = {.lex_state = 11, .external_lex_state = 30},
  [4485] = {.lex_state = 72, .external_lex_state = 28},
  [4486] = {.lex_state = 11, .external_lex_state = 32},
  [4487] = {.lex_state = 11, .external_lex_state = 32},
  [4488] = {.lex_state = 11, .external_lex_state = 18},
  [4489] = {.lex_state = 79, .external_lex_state = 42},
  [4490] = {.lex_state = 79, .external_lex_state = 42},
  [4491] = {.lex_state = 56, .external_lex_state = 20},
  [4492] = {.lex_state = 11, .external_lex_state = 32},
  [4493] = {.lex_state = 11, .external_lex_state = 32},
  [4494] = {.lex_state = 72, .external_lex_state = 28},
  [4495] = {.lex_state = 11, .external_lex_state = 32},
  [4496] = {.lex_state = 56, .external_lex_state = 20},
  [4497] = {.lex_state = 11, .external_lex_state = 32},
  [4498] = {.lex_state = 56, .external_lex_state = 20},
  [4499] = {.lex_state = 11, .external_lex_state = 41},
  [4500] = {.lex_state = 56, .external_lex_state = 20},
  [4501] = {.lex_state = 11, .external_lex_state = 41},
  [4502] = {.lex_state = 11, .external_lex_state = 41},
  [4503] = {.lex_state = 80, .external_lex_state = 20},
  [4504] = {.lex_state = 11, .external_lex_state = 32},
  [4505] = {.lex_state = 11, .external_lex_state = 32},
  [4506] = {.lex_state = 11, .external_lex_state = 32},
  [4507] = {.lex_state = 11, .external_lex_state = 32},
  [4508] = {.lex_state = 11, .external_lex_state = 32},
  [4509] = {.lex_state = 11, .external_lex_state = 32},
  [4510] = {.lex_state = 56, .external_lex_state = 20},
  [4511] = {.lex_state = 11, .external_lex_state = 18},
  [4512] = {.lex_state = 11, .external_lex_state = 32},
  [4513] = {.lex_state = 11, .external_lex_state = 32},
  [4514] = {.lex_state = 72, .external_lex_state = 44},
  [4515] = {.lex_state = 11, .external_lex_state = 18},
  [4516] = {.lex_state = 56, .external_lex_state = 20},
  [4517] = {.lex_state = 11, .external_lex_state = 41},
  [4518] = {.lex_state = 72, .external_lex_state = 28},
  [4519] = {.lex_state = 11, .external_lex_state = 32},
  [4520] = {.lex_state = 56, .external_lex_state = 20},
  [4521] = {.lex_state = 11, .external_lex_state = 32},
  [4522] = {.lex_state = 56, .external_lex_state = 20},
  [4523] = {.lex_state = 80, .external_lex_state = 28},
  [4524] = {.lex_state = 11, .external_lex_state = 32},
  [4525] = {.lex_state = 11, .external_lex_state = 32},
  [4526] = {.lex_state = 79, .external_lex_state = 42},
  [4527] = {.lex_state = 11, .external_lex_state = 32},
  [4528] = {.lex_state = 79, .external_lex_state = 42},
  [4529] = {.lex_state = 11, .external_lex_state = 32},
  [4530] = {.lex_state = 11, .external_lex_state = 32},
  [4531] = {.lex_state = 11, .external_lex_state = 32},
  [4532] = {.lex_state = 11, .external_lex_state = 20},
  [4533] = {.lex_state = 80, .external_lex_state = 28},
  [4534] = {.lex_state = 11, .external_lex_state = 32},
  [4535] = {.lex_state = 11, .external_lex_state = 32},
  [4536] = {.lex_state = 11, .external_lex_state = 32},
  [4537] = {.lex_state = 11, .external_lex_state = 32},
  [4538] = {.lex_state = 72, .external_lex_state = 44},
  [4539] = {.lex_state = 72, .external_lex_state = 44},
  [4540] = {.lex_state = 11, .external_lex_state = 20},
  [4541] = {.lex_state = 11, .external_lex_state = 32},
  [4542] = {.lex_state = 11, .external_lex_state = 20},
  [4543] = {.lex_state = 11, .external_lex_state = 20},
  [4544] = {.lex_state = 56, .external_lex_state = 20},
  [4545] = {.lex_state = 11, .external_lex_state = 32},
  [4546] = {.lex_state = 11, .external_lex_state = 18},
  [4547] = {.lex_state = 11, .external_lex_state = 32},
  [4548] = {.lex_state = 56, .external_lex_state = 20},
  [4549] = {.lex_state = 80, .external_lex_state = 20},
  [4550] = {.lex_state = 80, .external_lex_state = 28},
  [4551] = {.lex_state = 11, .external_lex_state = 32},
  [4552] = {.lex_state = 11, .external_lex_state = 32},
  [4553] = {.lex_state = 11, .external_lex_state = 32},
  [4554] = {.lex_state = 72, .external_lex_state = 45},
  [4555] = {.lex_state = 80, .external_lex_state = 22},
  [4556] = {.lex_state = 11, .external_lex_state = 33},
  [4557] = {.lex_state = 11, .external_lex_state = 33},
  [4558] = {.lex_state = 11, .external_lex_state = 33},
  [4559] = {.lex_state = 11, .external_lex_state = 33},
  [4560] = {.lex_state = 11, .external_lex_state = 33},
  [4561] = {.lex_state = 11, .external_lex_state = 33},
  [4562] = {.lex_state = 11, .external_lex_state = 33},
  [4563] = {.lex_state = 11, .external_lex_state = 33},
  [4564] = {.lex_state = 11, .external_lex_state = 33},
  [4565] = {.lex_state = 11, .external_lex_state = 33},
  [4566] = {.lex_state = 11, .external_lex_state = 33},
  [4567] = {.lex_state = 80, .external_lex_state = 21},
  [4568] = {.lex_state = 11, .external_lex_state = 33},
  [4569] = {.lex_state = 80, .external_lex_state = 23},
  [4570] = {.lex_state = 13, .external_lex_state = 18},
  [4571] = {.lex_state = 72, .external_lex_state = 45},
  [4572] = {.lex_state = 13, .external_lex_state = 20},
  [4573] = {.lex_state = 13, .external_lex_state = 20},
  [4574] = {.lex_state = 11, .external_lex_state = 33},
  [4575] = {.lex_state = 80, .external_lex_state = 21},
  [4576] = {.lex_state = 11, .external_lex_state = 33},
  [4577] = {.lex_state = 11, .external_lex_state = 33},
  [4578] = {.lex_state = 11, .external_lex_state = 33},
  [4579] = {.lex_state = 80, .external_lex_state = 22},
  [4580] = {.lex_state = 11, .external_lex_state = 33},
  [4581] = {.lex_state = 13, .external_lex_state = 18},
  [4582] = {.lex_state = 13, .external_lex_state = 18},
  [4583] = {.lex_state = 11, .external_lex_state = 33},
  [4584] = {.lex_state = 80, .external_lex_state = 23},
  [4585] = {.lex_state = 11, .external_lex_state = 33},
  [4586] = {.lex_state = 11, .external_lex_state = 33},
  [4587] = {.lex_state = 11, .external_lex_state = 33},
  [4588] = {.lex_state = 11, .external_lex_state = 20},
  [4589] = {.lex_state = 11, .external_lex_state = 20},
  [4590] = {.lex_state = 11, .external_lex_state = 33},
  [4591] = {.lex_state = 11, .external_lex_state = 33},
  [4592] = {.lex_state = 72, .external_lex_state = 45},
  [4593] = {.lex_state = 11, .external_lex_state = 33},
  [4594] = {.lex_state = 11, .external_lex_state = 33},
  [4595] = {.lex_state = 11, .external_lex_state = 33},
  [4596] = {.lex_state = 113, .external_lex_state = 46},
  [4597] = {.lex_state = 3, .external_lex_state = 44},
  [4598] = {.lex_state = 113, .external_lex_state = 46},
  [4599] = {.lex_state = 3, .external_lex_state = 44},
  [4600] = {.lex_state = 3, .external_lex_state = 44},
  [4601] = {.lex_state = 3, .external_lex_state = 44},
  [4602] = {.lex_state = 3, .external_lex_state = 44},
  [4603] = {.lex_state = 73, .external_lex_state = 20},
  [4604] = {.lex_state = 3, .external_lex_state = 44},
  [4605] = {.lex_state = 3, .external_lex_state = 44},
  [4606] = {.lex_state = 3, .external_lex_state = 28},
  [4607] = {.lex_state = 113, .external_lex_state = 46},
  [4608] = {.lex_state = 113, .external_lex_state = 46},
  [4609] = {.lex_state = 3, .external_lex_state = 44},
  [4610] = {.lex_state = 3, .external_lex_state = 44},
  [4611] = {.lex_state = 113, .external_lex_state = 46},
  [4612] = {.lex_state = 3, .external_lex_state = 44},
  [4613] = {.lex_state = 113, .external_lex_state = 46},
  [4614] = {.lex_state = 74, .external_lex_state = 36},
  [4615] = {.lex_state = 113, .external_lex_state = 46},
  [4616] = {.lex_state = 3, .external_lex_state = 44},
  [4617] = {.lex_state = 3, .external_lex_state = 44},
  [4618] = {.lex_state = 3, .external_lex_state = 44},
  [4619] = {.lex_state = 3, .external_lex_state = 44},
  [4620] = {.lex_state = 3, .external_lex_state = 44},
  [4621] = {.lex_state = 113, .external_lex_state = 46},
  [4622] = {.lex_state = 3, .external_lex_state = 44},
  [4623] = {.lex_state = 113, .external_lex_state = 46},
  [4624] = {.lex_state = 113, .external_lex_state = 46},
  [4625] = {.lex_state = 113, .external_lex_state = 46},
  [4626] = {.lex_state = 3, .external_lex_state = 44},
  [4627] = {.lex_state = 3, .external_lex_state = 44},
  [4628] = {.lex_state = 113, .external_lex_state = 46},
  [4629] = {.lex_state = 3, .external_lex_state = 44},
  [4630] = {.lex_state = 3, .external_lex_state = 44},
  [4631] = {.lex_state = 3, .external_lex_state = 44},
  [4632] = {.lex_state = 113, .external_lex_state = 46},
  [4633] = {.lex_state = 3, .external_lex_state = 44},
  [4634] = {.lex_state = 3, .external_lex_state = 44},
  [4635] = {.lex_state = 3, .external_lex_state = 44},
  [4636] = {.lex_state = 74, .external_lex_state = 36},
  [4637] = {.lex_state = 3, .external_lex_state = 44},
  [4638] = {.lex_state = 3, .external_lex_state = 44},
  [4639] = {.lex_state = 113, .external_lex_state = 46},
  [4640] = {.lex_state = 113, .external_lex_state = 46},
  [4641] = {.lex_state = 113, .external_lex_state = 46},
  [4642] = {.lex_state = 3, .external_lex_state = 44},
  [4643] = {.lex_state = 113, .external_lex_state = 46},
  [4644] = {.lex_state = 113, .external_lex_state = 46},
  [4645] = {.lex_state = 3, .external_lex_state = 28},
  [4646] = {.lex_state = 3, .external_lex_state = 44},
  [4647] = {.lex_state = 3, .external_lex_state = 28},
  [4648] = {.lex_state = 113, .external_lex_state = 46},
  [4649] = {.lex_state = 72, .external_lex_state = 47},
  [4650] = {.lex_state = 113, .external_lex_state = 46},
  [4651] = {.lex_state = 113, .external_lex_state = 46},
  [4652] = {.lex_state = 3, .external_lex_state = 44},
  [4653] = {.lex_state = 74, .external_lex_state = 36},
  [4654] = {.lex_state = 3, .external_lex_state = 44},
  [4655] = {.lex_state = 3, .external_lex_state = 44},
  [4656] = {.lex_state = 113, .external_lex_state = 46},
  [4657] = {.lex_state = 3, .external_lex_state = 44},
  [4658] = {.lex_state = 3, .external_lex_state = 44},
  [4659] = {.lex_state = 113, .external_lex_state = 46},
  [4660] = {.lex_state = 113, .external_lex_state = 46},
  [4661] = {.lex_state = 113, .external_lex_state = 46},
  [4662] = {.lex_state = 113, .external_lex_state = 46},
  [4663] = {.lex_state = 3, .external_lex_state = 44},
  [4664] = {.lex_state = 3, .external_lex_state = 44},
  [4665] = {.lex_state = 113, .external_lex_state = 46},
  [4666] = {.lex_state = 3, .external_lex_state = 44},
  [4667] = {.lex_state = 3, .external_lex_state = 44},
  [4668] = {.lex_state = 113, .external_lex_state = 46},
  [4669] = {.lex_state = 113, .external_lex_state = 46},
  [4670] = {.lex_state = 3, .external_lex_state = 28},
  [4671] = {.lex_state = 3, .external_lex_state = 28},
  [4672] = {.lex_state = 74, .external_lex_state = 36},
  [4673] = {.lex_state = 3, .external_lex_state = 44},
  [4674] = {.lex_state = 74, .external_lex_state = 36},
  [4675] = {.lex_state = 3, .external_lex_state = 44},
  [4676] = {.lex_state = 74, .external_lex_state = 36},
  [4677] = {.lex_state = 72, .external_lex_state = 29},
  [4678] = {.lex_state = 3, .external_lex_state = 44},
  [4679] = {.lex_state = 3, .external_lex_state = 44},
  [4680] = {.lex_state = 72, .external_lex_state = 29},
  [4681] = {.lex_state = 113, .external_lex_state = 46},
  [4682] = {.lex_state = 74, .external_lex_state = 36},
  [4683] = {.lex_state = 113, .external_lex_state = 46},
  [4684] = {.lex_state = 3, .external_lex_state = 44},
  [4685] = {.lex_state = 3, .external_lex_state = 44},
  [4686] = {.lex_state = 113, .external_lex_state = 46},
  [4687] = {.lex_state = 3, .external_lex_state = 44},
  [4688] = {.lex_state = 3, .external_lex_state = 44},
  [4689] = {.lex_state = 3, .external_lex_state = 44},
  [4690] = {.lex_state = 72, .external_lex_state = 47},
  [4691] = {.lex_state = 3, .external_lex_state = 44},
  [4692] = {.lex_state = 3, .external_lex_state = 44},
  [4693] = {.lex_state = 3, .external_lex_state = 44},
  [4694] = {.lex_state = 3, .external_lex_state = 44},
  [4695] = {.lex_state = 3, .external_lex_state = 44},
  [4696] = {.lex_state = 3, .external_lex_state = 44},
  [4697] = {.lex_state = 3, .external_lex_state = 44},
  [4698] = {.lex_state = 74, .external_lex_state = 36},
  [4699] = {.lex_state = 3, .external_lex_state = 44},
  [4700] = {.lex_state = 3, .external_lex_state = 44},
  [4701] = {.lex_state = 3, .external_lex_state = 44},
  [4702] = {.lex_state = 73, .external_lex_state = 20},
  [4703] = {.lex_state = 73, .external_lex_state = 20},
  [4704] = {.lex_state = 3, .external_lex_state = 28},
  [4705] = {.lex_state = 113, .external_lex_state = 48},
  [4706] = {.lex_state = 113, .external_lex_state = 48},
  [4707] = {.lex_state = 80, .external_lex_state = 20},
  [4708] = {.lex_state = 113, .external_lex_state = 44},
  [4709] = {.lex_state = 113, .external_lex_state = 48},
  [4710] = {.lex_state = 113, .external_lex_state = 44},
  [4711] = {.lex_state = 72, .external_lex_state = 44},
  [4712] = {.lex_state = 113, .external_lex_state = 48},
  [4713] = {.lex_state = 113, .external_lex_state = 48},
  [4714] = {.lex_state = 113, .external_lex_state = 48},
  [4715] = {.lex_state = 74, .external_lex_state = 20},
  [4716] = {.lex_state = 113, .external_lex_state = 48},
  [4717] = {.lex_state = 113, .external_lex_state = 48},
  [4718] = {.lex_state = 113, .external_lex_state = 44},
  [4719] = {.lex_state = 72, .external_lex_state = 28},
  [4720] = {.lex_state = 113, .external_lex_state = 48},
  [4721] = {.lex_state = 113, .external_lex_state = 48},
  [4722] = {.lex_state = 113, .external_lex_state = 48},
  [4723] = {.lex_state = 113, .external_lex_state = 48},
  [4724] = {.lex_state = 113, .external_lex_state = 48},
  [4725] = {.lex_state = 113, .external_lex_state = 48},
  [4726] = {.lex_state = 113, .external_lex_state = 48},
  [4727] = {.lex_state = 113, .external_lex_state = 48},
  [4728] = {.lex_state = 113, .external_lex_state = 48},
  [4729] = {.lex_state = 113, .external_lex_state = 48},
  [4730] = {.lex_state = 113, .external_lex_state = 44},
  [4731] = {.lex_state = 113, .external_lex_state = 48},
  [4732] = {.lex_state = 113, .external_lex_state = 44},
  [4733] = {.lex_state = 113, .external_lex_state = 48},
  [4734] = {.lex_state = 113, .external_lex_state = 44},
  [4735] = {.lex_state = 113, .external_lex_state = 44},
  [4736] = {.lex_state = 113, .external_lex_state = 48},
  [4737] = {.lex_state = 113, .external_lex_state = 44},
  [4738] = {.lex_state = 113, .external_lex_state = 48},
  [4739] = {.lex_state = 113, .external_lex_state = 48},
  [4740] = {.lex_state = 113, .external_lex_state = 48},
  [4741] = {.lex_state = 113, .external_lex_state = 48},
  [4742] = {.lex_state = 113, .external_lex_state = 48},
  [4743] = {.lex_state = 113, .external_lex_state = 48},
  [4744] = {.lex_state = 113, .external_lex_state = 48},
  [4745] = {.lex_state = 113, .external_lex_state = 48},
  [4746] = {.lex_state = 113, .external_lex_state = 48},
  [4747] = {.lex_state = 74, .external_lex_state = 20},
  [4748] = {.lex_state = 3, .external_lex_state = 45},
  [4749] = {.lex_state = 113, .external_lex_state = 48},
  [4750] = {.lex_state = 113, .external_lex_state = 48},
  [4751] = {.lex_state = 113, .external_lex_state = 48},
  [4752] = {.lex_state = 113, .external_lex_state = 28},
  [4753] = {.lex_state = 74, .external_lex_state = 20},
  [4754] = {.lex_state = 113, .external_lex_state = 48},
  [4755] = {.lex_state = 3, .external_lex_state = 45},
  [4756] = {.lex_state = 113, .external_lex_state = 48},
  [4757] = {.lex_state = 113, .external_lex_state = 44},
  [4758] = {.lex_state = 113, .external_lex_state = 48},
  [4759] = {.lex_state = 113, .external_lex_state = 44},
  [4760] = {.lex_state = 113, .external_lex_state = 48},
  [4761] = {.lex_state = 113, .external_lex_state = 48},
  [4762] = {.lex_state = 113, .external_lex_state = 48},
  [4763] = {.lex_state = 113, .external_lex_state = 48},
  [4764] = {.lex_state = 113, .external_lex_state = 48},
  [4765] = {.lex_state = 113, .external_lex_state = 48},
  [4766] = {.lex_state = 113, .external_lex_state = 48},
  [4767] = {.lex_state = 113, .external_lex_state = 48},
  [4768] = {.lex_state = 113, .external_lex_state = 48},
  [4769] = {.lex_state = 113, .external_lex_state = 48},
  [4770] = {.lex_state = 72, .external_lex_state = 28},
  [4771] = {.lex_state = 81, .external_lex_state = 20},
  [4772] = {.lex_state = 113, .external_lex_state = 48},
  [4773] = {.lex_state = 74, .external_lex_state = 20},
  [4774] = {.lex_state = 113, .external_lex_state = 48},
  [4775] = {.lex_state = 113, .external_lex_state = 48},
  [4776] = {.lex_state = 113, .external_lex_state = 48},
  [4777] = {.lex_state = 113, .external_lex_state = 48},
  [4778] = {.lex_state = 113, .external_lex_state = 48},
  [4779] = {.lex_state = 113, .external_lex_state = 48},
  [4780] = {.lex_state = 113, .external_lex_state = 48},
  [4781] = {.lex_state = 113, .external_lex_state = 48},
  [4782] = {.lex_state = 113, .external_lex_state = 28},
  [4783] = {.lex_state = 113, .external_lex_state = 48},
  [4784] = {.lex_state = 113, .external_lex_state = 48},
  [4785] = {.lex_state = 113, .external_lex_state = 48},
  [4786] = {.lex_state = 113, .external_lex_state = 44},
  [4787] = {.lex_state = 113, .external_lex_state = 48},
  [4788] = {.lex_state = 81, .external_lex_state = 20},
  [4789] = {.lex_state = 113, .external_lex_state = 48},
  [4790] = {.lex_state = 113, .external_lex_state = 44},
  [4791] = {.lex_state = 113, .external_lex_state = 28},
  [4792] = {.lex_state = 113, .external_lex_state = 48},
  [4793] = {.lex_state = 13, .external_lex_state = 20},
  [4794] = {.lex_state = 13, .external_lex_state = 18},
  [4795] = {.lex_state = 113, .external_lex_state = 44},
  [4796] = {.lex_state = 113, .external_lex_state = 48},
  [4797] = {.lex_state = 13, .external_lex_state = 18},
  [4798] = {.lex_state = 113, .external_lex_state = 48},
  [4799] = {.lex_state = 113, .external_lex_state = 48},
  [4800] = {.lex_state = 113, .external_lex_state = 48},
  [4801] = {.lex_state = 113, .external_lex_state = 48},
  [4802] = {.lex_state = 113, .external_lex_state = 48},
  [4803] = {.lex_state = 113, .external_lex_state = 48},
  [4804] = {.lex_state = 113, .external_lex_state = 48},
  [4805] = {.lex_state = 113, .external_lex_state = 28},
  [4806] = {.lex_state = 113, .external_lex_state = 44},
  [4807] = {.lex_state = 113, .external_lex_state = 44},
  [4808] = {.lex_state = 113, .external_lex_state = 48},
  [4809] = {.lex_state = 113, .external_lex_state = 44},
  [4810] = {.lex_state = 113, .external_lex_state = 44},
  [4811] = {.lex_state = 113, .external_lex_state = 48},
  [4812] = {.lex_state = 113, .external_lex_state = 44},
  [4813] = {.lex_state = 113, .external_lex_state = 48},
  [4814] = {.lex_state = 113, .external_lex_state = 48},
  [4815] = {.lex_state = 113, .external_lex_state = 48},
  [4816] = {.lex_state = 113, .external_lex_state = 44},
  [4817] = {.lex_state = 113, .external_lex_state = 48},
  [4818] = {.lex_state = 113, .external_lex_state = 48},
  [4819] = {.lex_state = 113, .external_lex_state = 48},
  [4820] = {.lex_state = 113, .external_lex_state = 28},
  [4821] = {.lex_state = 113, .external_lex_state = 48},
  [4822] = {.lex_state = 113, .external_lex_state = 48},
  [4823] = {.lex_state = 113, .external_lex_state = 44},
  [4824] = {.lex_state = 113, .external_lex_state = 48},
  [4825] = {.lex_state = 113, .external_lex_state = 44},
  [4826] = {.lex_state = 113, .external_lex_state = 48},
  [4827] = {.lex_state = 113, .external_lex_state = 48},
  [4828] = {.lex_state = 113, .external_lex_state = 44},
  [4829] = {.lex_state = 113, .external_lex_state = 48},
  [4830] = {.lex_state = 113, .external_lex_state = 44},
  [4831] = {.lex_state = 113, .external_lex_state = 48},
  [4832] = {.lex_state = 113, .external_lex_state = 48},
  [4833] = {.lex_state = 113, .external_lex_state = 48},
  [4834] = {.lex_state = 113, .external_lex_state = 48},
  [4835] = {.lex_state = 113, .external_lex_state = 48},
  [4836] = {.lex_state = 74, .external_lex_state = 20},
  [4837] = {.lex_state = 113, .external_lex_state = 48},
  [4838] = {.lex_state = 113, .external_lex_state = 48},
  [4839] = {.lex_state = 113, .external_lex_state = 48},
  [4840] = {.lex_state = 113, .external_lex_state = 48},
  [4841] = {.lex_state = 113, .external_lex_state = 48},
  [4842] = {.lex_state = 113, .external_lex_state = 44},
  [4843] = {.lex_state = 113, .external_lex_state = 44},
  [4844] = {.lex_state = 113, .external_lex_state = 48},
  [4845] = {.lex_state = 113, .external_lex_state = 48},
  [4846] = {.lex_state = 113, .external_lex_state = 48},
  [4847] = {.lex_state = 113, .external_lex_state = 48},
  [4848] = {.lex_state = 113, .external_lex_state = 44},
  [4849] = {.lex_state = 113, .external_lex_state = 44},
  [4850] = {.lex_state = 113, .external_lex_state = 44},
  [4851] = {.lex_state = 113, .external_lex_state = 44},
  [4852] = {.lex_state = 113, .external_lex_state = 44},
  [4853] = {.lex_state = 113, .external_lex_state = 48},
  [4854] = {.lex_state = 113, .external_lex_state = 48},
  [4855] = {.lex_state = 113, .external_lex_state = 48},
  [4856] = {.lex_state = 113, .external_lex_state = 44},
  [4857] = {.lex_state = 113, .external_lex_state = 44},
  [4858] = {.lex_state = 113, .external_lex_state = 48},
  [4859] = {.lex_state = 113, .external_lex_state = 48},
  [4860] = {.lex_state = 113, .external_lex_state = 44},
  [4861] = {.lex_state = 113, .external_lex_state = 48},
  [4862] = {.lex_state = 113, .external_lex_state = 44},
  [4863] = {.lex_state = 113, .external_lex_state = 44},
  [4864] = {.lex_state = 113, .external_lex_state = 44},
  [4865] = {.lex_state = 3, .external_lex_state = 45},
  [4866] = {.lex_state = 113, .external_lex_state = 48},
  [4867] = {.lex_state = 113, .external_lex_state = 48},
  [4868] = {.lex_state = 113, .external_lex_state = 48},
  [4869] = {.lex_state = 13, .external_lex_state = 18},
  [4870] = {.lex_state = 113, .external_lex_state = 48},
  [4871] = {.lex_state = 113, .external_lex_state = 48},
  [4872] = {.lex_state = 113, .external_lex_state = 44},
  [4873] = {.lex_state = 113, .external_lex_state = 48},
  [4874] = {.lex_state = 113, .external_lex_state = 28},
  [4875] = {.lex_state = 113, .external_lex_state = 44},
  [4876] = {.lex_state = 113, .external_lex_state = 44},
  [4877] = {.lex_state = 113, .external_lex_state = 44},
  [4878] = {.lex_state = 113, .external_lex_state = 48},
  [4879] = {.lex_state = 113, .external_lex_state = 44},
  [4880] = {.lex_state = 113, .external_lex_state = 48},
  [4881] = {.lex_state = 113, .external_lex_state = 48},
  [4882] = {.lex_state = 113, .external_lex_state = 48},
  [4883] = {.lex_state = 113, .external_lex_state = 48},
  [4884] = {.lex_state = 113, .external_lex_state = 28},
  [4885] = {.lex_state = 113, .external_lex_state = 48},
  [4886] = {.lex_state = 113, .external_lex_state = 28},
  [4887] = {.lex_state = 113, .external_lex_state = 48},
  [4888] = {.lex_state = 13, .external_lex_state = 20},
  [4889] = {.lex_state = 113, .external_lex_state = 28},
  [4890] = {.lex_state = 72, .external_lex_state = 44},
  [4891] = {.lex_state = 113, .external_lex_state = 48},
  [4892] = {.lex_state = 113, .external_lex_state = 44},
  [4893] = {.lex_state = 113, .external_lex_state = 44},
  [4894] = {.lex_state = 113, .external_lex_state = 48},
  [4895] = {.lex_state = 113, .external_lex_state = 48},
  [4896] = {.lex_state = 113, .external_lex_state = 48},
  [4897] = {.lex_state = 113, .external_lex_state = 48},
  [4898] = {.lex_state = 113, .external_lex_state = 44},
  [4899] = {.lex_state = 113, .external_lex_state = 44},
  [4900] = {.lex_state = 113, .external_lex_state = 28},
  [4901] = {.lex_state = 113, .external_lex_state = 48},
  [4902] = {.lex_state = 113, .external_lex_state = 44},
  [4903] = {.lex_state = 113, .external_lex_state = 48},
  [4904] = {.lex_state = 113, .external_lex_state = 48},
  [4905] = {.lex_state = 113, .external_lex_state = 44},
  [4906] = {.lex_state = 113, .external_lex_state = 48},
  [4907] = {.lex_state = 113, .external_lex_state = 48},
  [4908] = {.lex_state = 113, .external_lex_state = 48},
  [4909] = {.lex_state = 74, .external_lex_state = 20},
  [4910] = {.lex_state = 113, .external_lex_state = 48},
  [4911] = {.lex_state = 113, .external_lex_state = 48},
  [4912] = {.lex_state = 113, .external_lex_state = 44},
  [4913] = {.lex_state = 72, .external_lex_state = 44},
  [4914] = {.lex_state = 113, .external_lex_state = 44},
  [4915] = {.lex_state = 113, .external_lex_state = 48},
  [4916] = {.lex_state = 57, .external_lex_state = 20},
  [4917] = {.lex_state = 74, .external_lex_state = 20},
  [4918] = {.lex_state = 113, .external_lex_state = 48},
  [4919] = {.lex_state = 113, .external_lex_state = 48},
  [4920] = {.lex_state = 113, .external_lex_state = 48},
  [4921] = {.lex_state = 57, .external_lex_state = 20},
  [4922] = {.lex_state = 113, .external_lex_state = 48},
  [4923] = {.lex_state = 113, .external_lex_state = 48},
  [4924] = {.lex_state = 113, .external_lex_state = 48},
  [4925] = {.lex_state = 74, .external_lex_state = 20},
  [4926] = {.lex_state = 113, .external_lex_state = 48},
  [4927] = {.lex_state = 113, .external_lex_state = 48},
  [4928] = {.lex_state = 113, .external_lex_state = 48},
  [4929] = {.lex_state = 113, .external_lex_state = 48},
  [4930] = {.lex_state = 57, .external_lex_state = 20},
  [4931] = {.lex_state = 113, .external_lex_state = 28},
  [4932] = {.lex_state = 113, .external_lex_state = 48},
  [4933] = {.lex_state = 57, .external_lex_state = 20},
  [4934] = {.lex_state = 113, .external_lex_state = 48},
  [4935] = {.lex_state = 57, .external_lex_state = 20},
  [4936] = {.lex_state = 113, .external_lex_state = 48},
  [4937] = {.lex_state = 72, .external_lex_state = 49},
  [4938] = {.lex_state = 57, .external_lex_state = 20},
  [4939] = {.lex_state = 113, .external_lex_state = 48},
  [4940] = {.lex_state = 113, .external_lex_state = 48},
  [4941] = {.lex_state = 113, .external_lex_state = 48},
  [4942] = {.lex_state = 113, .external_lex_state = 48},
  [4943] = {.lex_state = 113, .external_lex_state = 48},
  [4944] = {.lex_state = 113, .external_lex_state = 44},
  [4945] = {.lex_state = 113, .external_lex_state = 48},
  [4946] = {.lex_state = 113, .external_lex_state = 44},
  [4947] = {.lex_state = 113, .external_lex_state = 48},
  [4948] = {.lex_state = 72, .external_lex_state = 49},
  [4949] = {.lex_state = 113, .external_lex_state = 44},
  [4950] = {.lex_state = 113, .external_lex_state = 48},
  [4951] = {.lex_state = 113, .external_lex_state = 28},
  [4952] = {.lex_state = 113, .external_lex_state = 44},
  [4953] = {.lex_state = 113, .external_lex_state = 44},
  [4954] = {.lex_state = 80, .external_lex_state = 28},
  [4955] = {.lex_state = 113, .external_lex_state = 48},
  [4956] = {.lex_state = 113, .external_lex_state = 48},
  [4957] = {.lex_state = 113, .external_lex_state = 48},
  [4958] = {.lex_state = 113, .external_lex_state = 28},
  [4959] = {.lex_state = 113, .external_lex_state = 48},
  [4960] = {.lex_state = 113, .external_lex_state = 28},
  [4961] = {.lex_state = 113, .external_lex_state = 48},
  [4962] = {.lex_state = 113, .external_lex_state = 44},
  [4963] = {.lex_state = 113, .external_lex_state = 28},
  [4964] = {.lex_state = 80, .external_lex_state = 20},
  [4965] = {.lex_state = 113, .external_lex_state = 48},
  [4966] = {.lex_state = 113, .external_lex_state = 48},
  [4967] = {.lex_state = 113, .external_lex_state = 48},
  [4968] = {.lex_state = 113, .external_lex_state = 48},
  [4969] = {.lex_state = 80, .external_lex_state = 20},
  [4970] = {.lex_state = 81, .external_lex_state = 20},
  [4971] = {.lex_state = 72, .external_lex_state = 28},
  [4972] = {.lex_state = 113, .external_lex_state = 44},
  [4973] = {.lex_state = 78, .external_lex_state = 20},
  [4974] = {.lex_state = 80, .external_lex_state = 23},
  [4975] = {.lex_state = 113, .external_lex_state = 20},
  [4976] = {.lex_state = 80, .external_lex_state = 23},
  [4977] = {.lex_state = 80, .external_lex_state = 21},
  [4978] = {.lex_state = 80, .external_lex_state = 21},
  [4979] = {.lex_state = 80, .external_lex_state = 21},
  [4980] = {.lex_state = 80, .external_lex_state = 21},
  [4981] = {.lex_state = 78, .external_lex_state = 20},
  [4982] = {.lex_state = 80, .external_lex_state = 23},
  [4983] = {.lex_state = 80, .external_lex_state = 21},
  [4984] = {.lex_state = 80, .external_lex_state = 23},
  [4985] = {.lex_state = 80, .external_lex_state = 21},
  [4986] = {.lex_state = 80, .external_lex_state = 23},
  [4987] = {.lex_state = 80, .external_lex_state = 21},
  [4988] = {.lex_state = 78, .external_lex_state = 20},
  [4989] = {.lex_state = 78, .external_lex_state = 20},
  [4990] = {.lex_state = 80, .external_lex_state = 23},
  [4991] = {.lex_state = 80, .external_lex_state = 23},
  [4992] = {.lex_state = 80, .external_lex_state = 21},
  [4993] = {.lex_state = 80, .external_lex_state = 21},
  [4994] = {.lex_state = 80, .external_lex_state = 21},
  [4995] = {.lex_state = 80, .external_lex_state = 21},
  [4996] = {.lex_state = 80, .external_lex_state = 23},
  [4997] = {.lex_state = 80, .external_lex_state = 21},
  [4998] = {.lex_state = 80, .external_lex_state = 23},
  [4999] = {.lex_state = 80, .external_lex_state = 23},
  [5000] = {.lex_state = 80, .external_lex_state = 21},
  [5001] = {.lex_state = 80, .external_lex_state = 23},
  [5002] = {.lex_state = 80, .external_lex_state = 23},
  [5003] = {.lex_state = 80, .external_lex_state = 21},
  [5004] = {.lex_state = 80, .external_lex_state = 21},
  [5005] = {.lex_state = 80, .external_lex_state = 23},
  [5006] = {.lex_state = 80, .external_lex_state = 23},
  [5007] = {.lex_state = 78, .external_lex_state = 20},
  [5008] = {.lex_state = 78, .external_lex_state = 20},
  [5009] = {.lex_state = 78, .external_lex_state = 20},
  [5010] = {.lex_state = 80, .external_lex_state = 21},
  [5011] = {.lex_state = 80, .external_lex_state = 21},
  [5012] = {.lex_state = 80, .external_lex_state = 23},
  [5013] = {.lex_state = 113, .external_lex_state = 20},
  [5014] = {.lex_state = 80, .external_lex_state = 23},
  [5015] = {.lex_state = 80, .external_lex_state = 21},
  [5016] = {.lex_state = 80, .external_lex_state = 21},
  [5017] = {.lex_state = 80, .external_lex_state = 23},
  [5018] = {.lex_state = 80, .external_lex_state = 23},
  [5019] = {.lex_state = 78, .external_lex_state = 20},
  [5020] = {.lex_state = 78, .external_lex_state = 20},
  [5021] = {.lex_state = 80, .external_lex_state = 23},
  [5022] = {.lex_state = 78, .external_lex_state = 20},
  [5023] = {.lex_state = 78, .external_lex_state = 20},
  [5024] = {.lex_state = 78, .external_lex_state = 20},
  [5025] = {.lex_state = 78, .external_lex_state = 20},
  [5026] = {.lex_state = 80, .external_lex_state = 23},
  [5027] = {.lex_state = 80, .external_lex_state = 21},
  [5028] = {.lex_state = 80, .external_lex_state = 23},
  [5029] = {.lex_state = 80, .external_lex_state = 21},
  [5030] = {.lex_state = 113, .external_lex_state = 46},
  [5031] = {.lex_state = 78, .external_lex_state = 20},
  [5032] = {.lex_state = 80, .external_lex_state = 23},
  [5033] = {.lex_state = 78, .external_lex_state = 20},
  [5034] = {.lex_state = 80, .external_lex_state = 21},
  [5035] = {.lex_state = 13, .external_lex_state = 20},
  [5036] = {.lex_state = 113, .external_lex_state = 48},
  [5037] = {.lex_state = 80, .external_lex_state = 23},
  [5038] = {.lex_state = 80, .external_lex_state = 23},
  [5039] = {.lex_state = 80, .external_lex_state = 21},
  [5040] = {.lex_state = 76, .external_lex_state = 20},
  [5041] = {.lex_state = 113, .external_lex_state = 28},
  [5042] = {.lex_state = 80, .external_lex_state = 21},
  [5043] = {.lex_state = 80, .external_lex_state = 21},
  [5044] = {.lex_state = 13, .external_lex_state = 20},
  [5045] = {.lex_state = 80, .external_lex_state = 23},
  [5046] = {.lex_state = 80, .external_lex_state = 21},
  [5047] = {.lex_state = 80, .external_lex_state = 21},
  [5048] = {.lex_state = 80, .external_lex_state = 23},
  [5049] = {.lex_state = 80, .external_lex_state = 22},
  [5050] = {.lex_state = 76, .external_lex_state = 20},
  [5051] = {.lex_state = 80, .external_lex_state = 22},
  [5052] = {.lex_state = 80, .external_lex_state = 23},
  [5053] = {.lex_state = 78, .external_lex_state = 20},
  [5054] = {.lex_state = 78, .external_lex_state = 20},
  [5055] = {.lex_state = 80, .external_lex_state = 22},
  [5056] = {.lex_state = 80, .external_lex_state = 21},
  [5057] = {.lex_state = 78, .external_lex_state = 20},
  [5058] = {.lex_state = 72, .external_lex_state = 45},
  [5059] = {.lex_state = 113, .external_lex_state = 20},
  [5060] = {.lex_state = 80, .external_lex_state = 21},
  [5061] = {.lex_state = 80, .external_lex_state = 23},
  [5062] = {.lex_state = 80, .external_lex_state = 21},
  [5063] = {.lex_state = 80, .external_lex_state = 21},
  [5064] = {.lex_state = 80, .external_lex_state = 21},
  [5065] = {.lex_state = 80, .external_lex_state = 21},
  [5066] = {.lex_state = 80, .external_lex_state = 21},
  [5067] = {.lex_state = 80, .external_lex_state = 21},
  [5068] = {.lex_state = 80, .external_lex_state = 23},
  [5069] = {.lex_state = 80, .external_lex_state = 23},
  [5070] = {.lex_state = 80, .external_lex_state = 23},
  [5071] = {.lex_state = 80, .external_lex_state = 21},
  [5072] = {.lex_state = 13, .external_lex_state = 20},
  [5073] = {.lex_state = 78, .external_lex_state = 20},
  [5074] = {.lex_state = 80, .external_lex_state = 21},
  [5075] = {.lex_state = 80, .external_lex_state = 23},
  [5076] = {.lex_state = 80, .external_lex_state = 21},
  [5077] = {.lex_state = 3, .external_lex_state = 20},
  [5078] = {.lex_state = 113, .external_lex_state = 45},
  [5079] = {.lex_state = 80, .external_lex_state = 21},
  [5080] = {.lex_state = 80, .external_lex_state = 23},
  [5081] = {.lex_state = 80, .external_lex_state = 23},
  [5082] = {.lex_state = 78, .external_lex_state = 20},
  [5083] = {.lex_state = 78, .external_lex_state = 20},
  [5084] = {.lex_state = 78, .external_lex_state = 20},
  [5085] = {.lex_state = 3, .external_lex_state = 20},
  [5086] = {.lex_state = 80, .external_lex_state = 21},
  [5087] = {.lex_state = 80, .external_lex_state = 23},
  [5088] = {.lex_state = 80, .external_lex_state = 21},
  [5089] = {.lex_state = 80, .external_lex_state = 21},
  [5090] = {.lex_state = 80, .external_lex_state = 21},
  [5091] = {.lex_state = 80, .external_lex_state = 21},
  [5092] = {.lex_state = 3, .external_lex_state = 20},
  [5093] = {.lex_state = 13, .external_lex_state = 20},
  [5094] = {.lex_state = 80, .external_lex_state = 21},
  [5095] = {.lex_state = 13, .external_lex_state = 20},
  [5096] = {.lex_state = 3, .external_lex_state = 20},
  [5097] = {.lex_state = 113, .external_lex_state = 45},
  [5098] = {.lex_state = 80, .external_lex_state = 23},
  [5099] = {.lex_state = 80, .external_lex_state = 23},
  [5100] = {.lex_state = 80, .external_lex_state = 21},
  [5101] = {.lex_state = 113, .external_lex_state = 28},
  [5102] = {.lex_state = 13, .external_lex_state = 20},
  [5103] = {.lex_state = 80, .external_lex_state = 21},
  [5104] = {.lex_state = 13, .external_lex_state = 20},
  [5105] = {.lex_state = 80, .external_lex_state = 21},
  [5106] = {.lex_state = 80, .external_lex_state = 23},
  [5107] = {.lex_state = 80, .external_lex_state = 21},
  [5108] = {.lex_state = 80, .external_lex_state = 21},
  [5109] = {.lex_state = 80, .external_lex_state = 23},
  [5110] = {.lex_state = 80, .external_lex_state = 21},
  [5111] = {.lex_state = 113, .external_lex_state = 20},
  [5112] = {.lex_state = 80, .external_lex_state = 21},
  [5113] = {.lex_state = 80, .external_lex_state = 23},
  [5114] = {.lex_state = 13, .external_lex_state = 20},
  [5115] = {.lex_state = 57, .external_lex_state = 20},
  [5116] = {.lex_state = 80, .external_lex_state = 21},
  [5117] = {.lex_state = 72, .external_lex_state = 45},
  [5118] = {.lex_state = 80, .external_lex_state = 21},
  [5119] = {.lex_state = 13, .external_lex_state = 20},
  [5120] = {.lex_state = 80, .external_lex_state = 23},
  [5121] = {.lex_state = 80, .external_lex_state = 21},
  [5122] = {.lex_state = 80, .external_lex_state = 23},
  [5123] = {.lex_state = 80, .external_lex_state = 23},
  [5124] = {.lex_state = 80, .external_lex_state = 23},
  [5125] = {.lex_state = 80, .external_lex_state = 21},
  [5126] = {.lex_state = 80, .external_lex_state = 21},
  [5127] = {.lex_state = 80, .external_lex_state = 21},
  [5128] = {.lex_state = 80, .external_lex_state = 21},
  [5129] = {.lex_state = 113, .external_lex_state = 20},
  [5130] = {.lex_state = 78, .external_lex_state = 20},
  [5131] = {.lex_state = 78, .external_lex_state = 20},
  [5132] = {.lex_state = 78, .external_lex_state = 20},
  [5133] = {.lex_state = 80, .external_lex_state = 21},
  [5134] = {.lex_state = 80, .external_lex_state = 23},
  [5135] = {.lex_state = 80, .external_lex_state = 23},
  [5136] = {.lex_state = 80, .external_lex_state = 21},
  [5137] = {.lex_state = 113, .external_lex_state = 45},
  [5138] = {.lex_state = 13, .external_lex_state = 20},
  [5139] = {.lex_state = 13, .external_lex_state = 20},
  [5140] = {.lex_state = 80, .external_lex_state = 23},
  [5141] = {.lex_state = 80, .external_lex_state = 21},
  [5142] = {.lex_state = 80, .external_lex_state = 23},
  [5143] = {.lex_state = 13, .external_lex_state = 20},
  [5144] = {.lex_state = 80, .external_lex_state = 21},
  [5145] = {.lex_state = 80, .external_lex_state = 21},
  [5146] = {.lex_state = 80, .external_lex_state = 21},
  [5147] = {.lex_state = 72, .external_lex_state = 45},
  [5148] = {.lex_state = 80, .external_lex_state = 21},
  [5149] = {.lex_state = 13, .external_lex_state = 20},
  [5150] = {.lex_state = 13, .external_lex_state = 20},
  [5151] = {.lex_state = 80, .external_lex_state = 23},
  [5152] = {.lex_state = 80, .external_lex_state = 22},
  [5153] = {.lex_state = 80, .external_lex_state = 21},
  [5154] = {.lex_state = 80, .external_lex_state = 23},
  [5155] = {.lex_state = 78, .external_lex_state = 20},
  [5156] = {.lex_state = 78, .external_lex_state = 20},
  [5157] = {.lex_state = 80, .external_lex_state = 23},
  [5158] = {.lex_state = 13, .external_lex_state = 20},
  [5159] = {.lex_state = 80, .external_lex_state = 20},
  [5160] = {.lex_state = 113, .external_lex_state = 20},
  [5161] = {.lex_state = 80, .external_lex_state = 20},
  [5162] = {.lex_state = 113, .external_lex_state = 50},
  [5163] = {.lex_state = 80, .external_lex_state = 20},
  [5164] = {.lex_state = 11, .external_lex_state = 20},
  [5165] = {.lex_state = 113, .external_lex_state = 20},
  [5166] = {.lex_state = 80, .external_lex_state = 20},
  [5167] = {.lex_state = 80, .external_lex_state = 20},
  [5168] = {.lex_state = 80, .external_lex_state = 20},
  [5169] = {.lex_state = 80, .external_lex_state = 20},
  [5170] = {.lex_state = 80, .external_lex_state = 20},
  [5171] = {.lex_state = 80, .external_lex_state = 20},
  [5172] = {.lex_state = 80, .external_lex_state = 20},
  [5173] = {.lex_state = 80, .external_lex_state = 20},
  [5174] = {.lex_state = 113, .external_lex_state = 20},
  [5175] = {.lex_state = 80, .external_lex_state = 20},
  [5176] = {.lex_state = 11, .external_lex_state = 20},
  [5177] = {.lex_state = 80, .external_lex_state = 20},
  [5178] = {.lex_state = 80, .external_lex_state = 20},
  [5179] = {.lex_state = 80, .external_lex_state = 20},
  [5180] = {.lex_state = 113, .external_lex_state = 50},
  [5181] = {.lex_state = 80, .external_lex_state = 20},
  [5182] = {.lex_state = 113, .external_lex_state = 50},
  [5183] = {.lex_state = 80, .external_lex_state = 20},
  [5184] = {.lex_state = 72, .external_lex_state = 20},
  [5185] = {.lex_state = 113, .external_lex_state = 50},
  [5186] = {.lex_state = 113, .external_lex_state = 50},
  [5187] = {.lex_state = 72, .external_lex_state = 20},
  [5188] = {.lex_state = 80, .external_lex_state = 20},
  [5189] = {.lex_state = 80, .external_lex_state = 20},
  [5190] = {.lex_state = 80, .external_lex_state = 20},
  [5191] = {.lex_state = 80, .external_lex_state = 20},
  [5192] = {.lex_state = 80, .external_lex_state = 20},
  [5193] = {.lex_state = 80, .external_lex_state = 20},
  [5194] = {.lex_state = 80, .external_lex_state = 20},
  [5195] = {.lex_state = 80, .external_lex_state = 20},
  [5196] = {.lex_state = 72, .external_lex_state = 20},
  [5197] = {.lex_state = 80, .external_lex_state = 20},
  [5198] = {.lex_state = 113, .external_lex_state = 50},
  [5199] = {.lex_state = 113, .external_lex_state = 50},
  [5200] = {.lex_state = 113, .external_lex_state = 50},
  [5201] = {.lex_state = 113, .external_lex_state = 50},
  [5202] = {.lex_state = 80, .external_lex_state = 20},
  [5203] = {.lex_state = 80, .external_lex_state = 20},
  [5204] = {.lex_state = 80, .external_lex_state = 20},
  [5205] = {.lex_state = 80, .external_lex_state = 20},
  [5206] = {.lex_state = 80, .external_lex_state = 20},
  [5207] = {.lex_state = 80, .external_lex_state = 20},
  [5208] = {.lex_state = 80, .external_lex_state = 20},
  [5209] = {.lex_state = 113, .external_lex_state = 20},
  [5210] = {.lex_state = 80, .external_lex_state = 20},
  [5211] = {.lex_state = 80, .external_lex_state = 20},
  [5212] = {.lex_state = 113, .external_lex_state = 20},
  [5213] = {.lex_state = 113, .external_lex_state = 20},
  [5214] = {.lex_state = 113, .external_lex_state = 20},
  [5215] = {.lex_state = 80, .external_lex_state = 20},
  [5216] = {.lex_state = 113, .external_lex_state = 50},
  [5217] = {.lex_state = 113, .external_lex_state = 50},
  [5218] = {.lex_state = 113, .external_lex_state = 50},
  [5219] = {.lex_state = 113, .external_lex_state = 50},
  [5220] = {.lex_state = 113, .external_lex_state = 50},
  [5221] = {.lex_state = 80, .external_lex_state = 20},
  [5222] = {.lex_state = 113, .external_lex_state = 20},
  [5223] = {.lex_state = 80, .external_lex_state = 20},
  [5224] = {.lex_state = 80, .external_lex_state = 20},
  [5225] = {.lex_state = 80, .external_lex_state = 20},
  [5226] = {.lex_state = 113, .external_lex_state = 20},
  [5227] = {.lex_state = 80, .external_lex_state = 20},
  [5228] = {.lex_state = 80, .external_lex_state = 20},
  [5229] = {.lex_state = 113, .external_lex_state = 20},
  [5230] = {.lex_state = 113, .external_lex_state = 20},
  [5231] = {.lex_state = 113, .external_lex_state = 50},
  [5232] = {.lex_state = 113, .external_lex_state = 50},
  [5233] = {.lex_state = 113, .external_lex_state = 50},
  [5234] = {.lex_state = 113, .external_lex_state = 50},
  [5235] = {.lex_state = 80, .external_lex_state = 20},
  [5236] = {.lex_state = 80, .external_lex_state = 20},
  [5237] = {.lex_state = 80, .external_lex_state = 20},
  [5238] = {.lex_state = 80, .external_lex_state = 20},
  [5239] = {.lex_state = 113, .external_lex_state = 20},
  [5240] = {.lex_state = 72, .external_lex_state = 20},
  [5241] = {.lex_state = 11, .external_lex_state = 20},
  [5242] = {.lex_state = 80, .external_lex_state = 20},
  [5243] = {.lex_state = 72, .external_lex_state = 20},
  [5244] = {.lex_state = 72, .external_lex_state = 20},
  [5245] = {.lex_state = 80, .external_lex_state = 20},
  [5246] = {.lex_state = 80, .external_lex_state = 20},
  [5247] = {.lex_state = 80, .external_lex_state = 20},
  [5248] = {.lex_state = 80, .external_lex_state = 20},
  [5249] = {.lex_state = 80, .external_lex_state = 20},
  [5250] = {.lex_state = 113, .external_lex_state = 46},
  [5251] = {.lex_state = 11, .external_lex_state = 20},
  [5252] = {.lex_state = 113, .external_lex_state = 50},
  [5253] = {.lex_state = 113, .external_lex_state = 50},
  [5254] = {.lex_state = 113, .external_lex_state = 50},
  [5255] = {.lex_state = 80, .external_lex_state = 20},
  [5256] = {.lex_state = 80, .external_lex_state = 20},
  [5257] = {.lex_state = 80, .external_lex_state = 20},
  [5258] = {.lex_state = 80, .external_lex_state = 20},
  [5259] = {.lex_state = 80, .external_lex_state = 20},
  [5260] = {.lex_state = 80, .external_lex_state = 20},
  [5261] = {.lex_state = 80, .external_lex_state = 20},
  [5262] = {.lex_state = 80, .external_lex_state = 20},
  [5263] = {.lex_state = 80, .external_lex_state = 20},
  [5264] = {.lex_state = 80, .external_lex_state = 20},
  [5265] = {.lex_state = 80, .external_lex_state = 20},
  [5266] = {.lex_state = 11, .external_lex_state = 20},
  [5267] = {.lex_state = 113, .external_lex_state = 50},
  [5268] = {.lex_state = 113, .external_lex_state = 50},
  [5269] = {.lex_state = 113, .external_lex_state = 50},
  [5270] = {.lex_state = 113, .external_lex_state = 50},
  [5271] = {.lex_state = 80, .external_lex_state = 20},
  [5272] = {.lex_state = 80, .external_lex_state = 20},
  [5273] = {.lex_state = 113, .external_lex_state = 20},
  [5274] = {.lex_state = 72, .external_lex_state = 20},
  [5275] = {.lex_state = 113, .external_lex_state = 50},
  [5276] = {.lex_state = 113, .external_lex_state = 50},
  [5277] = {.lex_state = 113, .external_lex_state = 48},
  [5278] = {.lex_state = 80, .external_lex_state = 20},
  [5279] = {.lex_state = 80, .external_lex_state = 20},
  [5280] = {.lex_state = 80, .external_lex_state = 20},
  [5281] = {.lex_state = 80, .external_lex_state = 20},
  [5282] = {.lex_state = 80, .external_lex_state = 20},
  [5283] = {.lex_state = 113, .external_lex_state = 48},
  [5284] = {.lex_state = 80, .external_lex_state = 20},
  [5285] = {.lex_state = 13, .external_lex_state = 20},
  [5286] = {.lex_state = 11, .external_lex_state = 20},
  [5287] = {.lex_state = 80, .external_lex_state = 20},
  [5288] = {.lex_state = 113, .external_lex_state = 20},
  [5289] = {.lex_state = 72, .external_lex_state = 20},
  [5290] = {.lex_state = 80, .external_lex_state = 20},
  [5291] = {.lex_state = 80, .external_lex_state = 20},
  [5292] = {.lex_state = 113, .external_lex_state = 20},
  [5293] = {.lex_state = 80, .external_lex_state = 20},
  [5294] = {.lex_state = 113, .external_lex_state = 20},
  [5295] = {.lex_state = 113, .external_lex_state = 20},
  [5296] = {.lex_state = 80, .external_lex_state = 20},
  [5297] = {.lex_state = 113, .external_lex_state = 48},
  [5298] = {.lex_state = 80, .external_lex_state = 20},
  [5299] = {.lex_state = 3, .external_lex_state = 44},
  [5300] = {.lex_state = 80, .external_lex_state = 20},
  [5301] = {.lex_state = 113, .external_lex_state = 50},
  [5302] = {.lex_state = 80, .external_lex_state = 20},
  [5303] = {.lex_state = 80, .external_lex_state = 20},
  [5304] = {.lex_state = 80, .external_lex_state = 20},
  [5305] = {.lex_state = 72, .external_lex_state = 20},
  [5306] = {.lex_state = 80, .external_lex_state = 20},
  [5307] = {.lex_state = 80, .external_lex_state = 20},
  [5308] = {.lex_state = 80, .external_lex_state = 20},
  [5309] = {.lex_state = 80, .external_lex_state = 20},
  [5310] = {.lex_state = 72, .external_lex_state = 20},
  [5311] = {.lex_state = 72, .external_lex_state = 20},
  [5312] = {.lex_state = 11, .external_lex_state = 20},
  [5313] = {.lex_state = 72, .external_lex_state = 20},
  [5314] = {.lex_state = 113, .external_lex_state = 50},
  [5315] = {.lex_state = 113, .external_lex_state = 50},
  [5316] = {.lex_state = 113, .external_lex_state = 48},
  [5317] = {.lex_state = 80, .external_lex_state = 20},
  [5318] = {.lex_state = 113, .external_lex_state = 46},
  [5319] = {.lex_state = 113, .external_lex_state = 50},
  [5320] = {.lex_state = 11, .external_lex_state = 20},
  [5321] = {.lex_state = 113, .external_lex_state = 50},
  [5322] = {.lex_state = 113, .external_lex_state = 20},
  [5323] = {.lex_state = 113, .external_lex_state = 50},
  [5324] = {.lex_state = 113, .external_lex_state = 50},
  [5325] = {.lex_state = 113, .external_lex_state = 50},
  [5326] = {.lex_state = 113, .external_lex_state = 50},
  [5327] = {.lex_state = 113, .external_lex_state = 50},
  [5328] = {.lex_state = 113, .external_lex_state = 50},
  [5329] = {.lex_state = 113, .external_lex_state = 50},
  [5330] = {.lex_state = 80, .external_lex_state = 20},
  [5331] = {.lex_state = 113, .external_lex_state = 50},
  [5332] = {.lex_state = 113, .external_lex_state = 20},
  [5333] = {.lex_state = 113, .external_lex_state = 20},
  [5334] = {.lex_state = 113, .external_lex_state = 20},
  [5335] = {.lex_state = 113, .external_lex_state = 20},
  [5336] = {.lex_state = 113, .external_lex_state = 20},
  [5337] = {.lex_state = 113, .external_lex_state = 20},
  [5338] = {.lex_state = 113, .external_lex_state = 20},
  [5339] = {.lex_state = 113, .external_lex_state = 20},
  [5340] = {.lex_state = 113, .external_lex_state = 20},
  [5341] = {.lex_state = 72, .external_lex_state = 20},
  [5342] = {.lex_state = 72, .external_lex_state = 20},
  [5343] = {.lex_state = 113, .external_lex_state = 20},
  [5344] = {.lex_state = 113, .external_lex_state = 20},
  [5345] = {.lex_state = 113, .external_lex_state = 20},
  [5346] = {.lex_state = 113, .external_lex_state = 20},
  [5347] = {.lex_state = 113, .external_lex_state = 28},
  [5348] = {.lex_state = 113, .external_lex_state = 20},
  [5349] = {.lex_state = 11, .external_lex_state = 20},
  [5350] = {.lex_state = 113, .external_lex_state = 20},
  [5351] = {.lex_state = 113, .external_lex_state = 20},
  [5352] = {.lex_state = 113, .external_lex_state = 20},
  [5353] = {.lex_state = 113, .external_lex_state = 20},
  [5354] = {.lex_state = 113, .external_lex_state = 20},
  [5355] = {.lex_state = 113, .external_lex_state = 20},
  [5356] = {.lex_state = 113, .external_lex_state = 20},
  [5357] = {.lex_state = 113, .external_lex_state = 20},
  [5358] = {.lex_state = 113, .external_lex_state = 20},
  [5359] = {.lex_state = 113, .external_lex_state = 20},
  [5360] = {.lex_state = 113, .external_lex_state = 28},
  [5361] = {.lex_state = 113, .external_lex_state = 20},
  [5362] = {.lex_state = 113, .external_lex_state = 20},
  [5363] = {.lex_state = 113, .external_lex_state = 20},
  [5364] = {.lex_state = 113, .external_lex_state = 20},
  [5365] = {.lex_state = 72, .external_lex_state = 20},
  [5366] = {.lex_state = 113, .external_lex_state = 20},
  [5367] = {.lex_state = 72, .external_lex_state = 20},
  [5368] = {.lex_state = 113, .external_lex_state = 20},
  [5369] = {.lex_state = 113, .external_lex_state = 20},
  [5370] = {.lex_state = 113, .external_lex_state = 20},
  [5371] = {.lex_state = 113, .external_lex_state = 20},
  [5372] = {.lex_state = 113, .external_lex_state = 20},
  [5373] = {.lex_state = 113, .external_lex_state = 20},
  [5374] = {.lex_state = 113, .external_lex_state = 20},
  [5375] = {.lex_state = 113, .external_lex_state = 20},
  [5376] = {.lex_state = 113, .external_lex_state = 20},
  [5377] = {.lex_state = 113, .external_lex_state = 20},
  [5378] = {.lex_state = 11, .external_lex_state = 20},
  [5379] = {.lex_state = 11, .external_lex_state = 20},
  [5380] = {.lex_state = 11, .external_lex_state = 20},
  [5381] = {.lex_state = 113, .external_lex_state = 20},
  [5382] = {.lex_state = 113, .external_lex_state = 20},
  [5383] = {.lex_state = 113, .external_lex_state = 28},
  [5384] = {.lex_state = 113, .external_lex_state = 20},
  [5385] = {.lex_state = 113, .external_lex_state = 20},
  [5386] = {.lex_state = 113, .external_lex_state = 20},
  [5387] = {.lex_state = 113, .external_lex_state = 20},
  [5388] = {.lex_state = 113, .external_lex_state = 20},
  [5389] = {.lex_state = 113, .external_lex_state = 20},
  [5390] = {.lex_state = 113, .external_lex_state = 20},
  [5391] = {.lex_state = 113, .external_lex_state = 28},
  [5392] = {.lex_state = 113, .external_lex_state = 20},
  [5393] = {.lex_state = 72, .external_lex_state = 20},
  [5394] = {.lex_state = 72, .external_lex_state = 20},
  [5395] = {.lex_state = 72, .external_lex_state = 20},
  [5396] = {.lex_state = 113, .external_lex_state = 51},
  [5397] = {.lex_state = 113, .external_lex_state = 20},
  [5398] = {.lex_state = 113, .external_lex_state = 20},
  [5399] = {.lex_state = 113, .external_lex_state = 20},
  [5400] = {.lex_state = 113, .external_lex_state = 20},
  [5401] = {.lex_state = 113, .external_lex_state = 20},
  [5402] = {.lex_state = 72, .external_lex_state = 20},
  [5403] = {.lex_state = 72, .external_lex_state = 20},
  [5404] = {.lex_state = 11, .external_lex_state = 20},
  [5405] = {.lex_state = 113, .external_lex_state = 20},
  [5406] = {.lex_state = 113, .external_lex_state = 20},
  [5407] = {.lex_state = 113, .external_lex_state = 20},
  [5408] = {.lex_state = 113, .external_lex_state = 20},
  [5409] = {.lex_state = 72, .external_lex_state = 20},
  [5410] = {.lex_state = 72, .external_lex_state = 20},
  [5411] = {.lex_state = 113, .external_lex_state = 20},
  [5412] = {.lex_state = 113, .external_lex_state = 20},
  [5413] = {.lex_state = 113, .external_lex_state = 20},
  [5414] = {.lex_state = 113, .external_lex_state = 20},
  [5415] = {.lex_state = 113, .external_lex_state = 20},
  [5416] = {.lex_state = 113, .external_lex_state = 20},
  [5417] = {.lex_state = 113, .external_lex_state = 20},
  [5418] = {.lex_state = 11, .external_lex_state = 20},
  [5419] = {.lex_state = 113, .external_lex_state = 20},
  [5420] = {.lex_state = 113, .external_lex_state = 20},
  [5421] = {.lex_state = 113, .external_lex_state = 20},
  [5422] = {.lex_state = 113, .external_lex_state = 20},
  [5423] = {.lex_state = 113, .external_lex_state = 20},
  [5424] = {.lex_state = 11, .external_lex_state = 20},
  [5425] = {.lex_state = 11, .external_lex_state = 20},
  [5426] = {.lex_state = 113, .external_lex_state = 51},
  [5427] = {.lex_state = 113, .external_lex_state = 20},
  [5428] = {.lex_state = 113, .external_lex_state = 20},
  [5429] = {.lex_state = 113, .external_lex_state = 20},
  [5430] = {.lex_state = 113, .external_lex_state = 20},
  [5431] = {.lex_state = 113, .external_lex_state = 20},
  [5432] = {.lex_state = 113, .external_lex_state = 20},
  [5433] = {.lex_state = 113, .external_lex_state = 20},
  [5434] = {.lex_state = 113, .external_lex_state = 20},
  [5435] = {.lex_state = 113, .external_lex_state = 20},
  [5436] = {.lex_state = 113, .external_lex_state = 20},
  [5437] = {.lex_state = 113, .external_lex_state = 20},
  [5438] = {.lex_state = 113, .external_lex_state = 20},
  [5439] = {.lex_state = 113, .external_lex_state = 20},
  [5440] = {.lex_state = 113, .external_lex_state = 20},
  [5441] = {.lex_state = 113, .external_lex_state = 20},
  [5442] = {.lex_state = 113, .external_lex_state = 28},
  [5443] = {.lex_state = 113, .external_lex_state = 20},
  [5444] = {.lex_state = 113, .external_lex_state = 20},
  [5445] = {.lex_state = 113, .external_lex_state = 20},
  [5446] = {.lex_state = 113, .external_lex_state = 20},
  [5447] = {.lex_state = 113, .external_lex_state = 20},
  [5448] = {.lex_state = 113, .external_lex_state = 20},
  [5449] = {.lex_state = 113, .external_lex_state = 20},
  [5450] = {.lex_state = 113, .external_lex_state = 20},
  [5451] = {.lex_state = 113, .external_lex_state = 20},
  [5452] = {.lex_state = 113, .external_lex_state = 20},
  [5453] = {.lex_state = 113, .external_lex_state = 20},
  [5454] = {.lex_state = 113, .external_lex_state = 20},
  [5455] = {.lex_state = 113, .external_lex_state = 20},
  [5456] = {.lex_state = 113, .external_lex_state = 20},
  [5457] = {.lex_state = 113, .external_lex_state = 20},
  [5458] = {.lex_state = 113, .external_lex_state = 20},
  [5459] = {.lex_state = 113, .external_lex_state = 20},
  [5460] = {.lex_state = 113, .external_lex_state = 20},
  [5461] = {.lex_state = 113, .external_lex_state = 20},
  [5462] = {.lex_state = 113, .external_lex_state = 20},
  [5463] = {.lex_state = 113, .external_lex_state = 20},
  [5464] = {.lex_state = 72, .external_lex_state = 20},
  [5465] = {.lex_state = 113, .external_lex_state = 20},
  [5466] = {.lex_state = 113, .external_lex_state = 20},
  [5467] = {.lex_state = 113, .external_lex_state = 20},
  [5468] = {.lex_state = 113, .external_lex_state = 20},
  [5469] = {.lex_state = 113, .external_lex_state = 20},
  [5470] = {.lex_state = 113, .external_lex_state = 20},
  [5471] = {.lex_state = 113, .external_lex_state = 20},
  [5472] = {.lex_state = 113, .external_lex_state = 20},
  [5473] = {.lex_state = 113, .external_lex_state = 20},
  [5474] = {.lex_state = 113, .external_lex_state = 20},
  [5475] = {.lex_state = 113, .external_lex_state = 20},
  [5476] = {.lex_state = 113, .external_lex_state = 20},
  [5477] = {.lex_state = 113, .external_lex_state = 20},
  [5478] = {.lex_state = 113, .external_lex_state = 20},
  [5479] = {.lex_state = 113, .external_lex_state = 20},
  [5480] = {.lex_state = 113, .external_lex_state = 28},
  [5481] = {.lex_state = 113, .external_lex_state = 20},
  [5482] = {.lex_state = 113, .external_lex_state = 20},
  [5483] = {.lex_state = 113, .external_lex_state = 20},
  [5484] = {.lex_state = 113, .external_lex_state = 20},
  [5485] = {.lex_state = 113, .external_lex_state = 20},
  [5486] = {.lex_state = 113, .external_lex_state = 20},
  [5487] = {.lex_state = 113, .external_lex_state = 20},
  [5488] = {.lex_state = 113, .external_lex_state = 20},
  [5489] = {.lex_state = 113, .external_lex_state = 20},
  [5490] = {.lex_state = 113, .external_lex_state = 20},
  [5491] = {.lex_state = 72, .external_lex_state = 20},
  [5492] = {.lex_state = 113, .external_lex_state = 20},
  [5493] = {.lex_state = 113, .external_lex_state = 20},
  [5494] = {.lex_state = 113, .external_lex_state = 28},
  [5495] = {.lex_state = 113, .external_lex_state = 20},
  [5496] = {.lex_state = 113, .external_lex_state = 20},
  [5497] = {.lex_state = 113, .external_lex_state = 20},
  [5498] = {.lex_state = 113, .external_lex_state = 20},
  [5499] = {.lex_state = 72, .external_lex_state = 20},
  [5500] = {.lex_state = 113, .external_lex_state = 20},
  [5501] = {.lex_state = 113, .external_lex_state = 20},
  [5502] = {.lex_state = 113, .external_lex_state = 20},
  [5503] = {.lex_state = 113, .external_lex_state = 20},
  [5504] = {.lex_state = 113, .external_lex_state = 20},
  [5505] = {.lex_state = 113, .external_lex_state = 20},
  [5506] = {.lex_state = 113, .external_lex_state = 20},
  [5507] = {.lex_state = 113, .external_lex_state = 20},
  [5508] = {.lex_state = 113, .external_lex_state = 20},
  [5509] = {.lex_state = 113, .external_lex_state = 20},
  [5510] = {.lex_state = 113, .external_lex_state = 20},
  [5511] = {.lex_state = 113, .external_lex_state = 20},
  [5512] = {.lex_state = 11, .external_lex_state = 20},
  [5513] = {.lex_state = 113, .external_lex_state = 20},
  [5514] = {.lex_state = 113, .external_lex_state = 20},
  [5515] = {.lex_state = 113, .external_lex_state = 20},
  [5516] = {.lex_state = 72, .external_lex_state = 20},
  [5517] = {.lex_state = 113, .external_lex_state = 20},
  [5518] = {.lex_state = 113, .external_lex_state = 20},
  [5519] = {.lex_state = 113, .external_lex_state = 20},
  [5520] = {.lex_state = 113, .external_lex_state = 20},
  [5521] = {.lex_state = 11, .external_lex_state = 20},
  [5522] = {.lex_state = 113, .external_lex_state = 20},
  [5523] = {.lex_state = 11, .external_lex_state = 20},
  [5524] = {.lex_state = 113, .external_lex_state = 20},
  [5525] = {.lex_state = 113, .external_lex_state = 20},
  [5526] = {.lex_state = 113, .external_lex_state = 20},
  [5527] = {.lex_state = 113, .external_lex_state = 20},
  [5528] = {.lex_state = 113, .external_lex_state = 20},
  [5529] = {.lex_state = 113, .external_lex_state = 20},
  [5530] = {.lex_state = 113, .external_lex_state = 20},
  [5531] = {.lex_state = 11, .external_lex_state = 20},
  [5532] = {.lex_state = 113, .external_lex_state = 20},
  [5533] = {.lex_state = 113, .external_lex_state = 20},
  [5534] = {.lex_state = 113, .external_lex_state = 20},
  [5535] = {.lex_state = 72, .external_lex_state = 20},
  [5536] = {.lex_state = 113, .external_lex_state = 20},
  [5537] = {.lex_state = 113, .external_lex_state = 20},
  [5538] = {.lex_state = 113, .external_lex_state = 20},
  [5539] = {.lex_state = 113, .external_lex_state = 20},
  [5540] = {.lex_state = 11, .external_lex_state = 20},
  [5541] = {.lex_state = 113, .external_lex_state = 20},
  [5542] = {.lex_state = 11, .external_lex_state = 20},
  [5543] = {.lex_state = 113, .external_lex_state = 20},
  [5544] = {.lex_state = 113, .external_lex_state = 20},
  [5545] = {.lex_state = 113, .external_lex_state = 20},
  [5546] = {.lex_state = 113, .external_lex_state = 20},
  [5547] = {.lex_state = 11, .external_lex_state = 20},
  [5548] = {.lex_state = 113, .external_lex_state = 20},
  [5549] = {.lex_state = 113, .external_lex_state = 20},
  [5550] = {.lex_state = 113, .external_lex_state = 20},
  [5551] = {.lex_state = 113, .external_lex_state = 20},
  [5552] = {.lex_state = 113, .external_lex_state = 20},
  [5553] = {.lex_state = 113, .external_lex_state = 20},
  [5554] = {.lex_state = 113, .external_lex_state = 20},
  [5555] = {.lex_state = 113, .external_lex_state = 28},
  [5556] = {.lex_state = 113, .external_lex_state = 20},
  [5557] = {.lex_state = 113, .external_lex_state = 20},
  [5558] = {.lex_state = 113, .external_lex_state = 20},
  [5559] = {.lex_state = 11, .external_lex_state = 20},
  [5560] = {.lex_state = 113, .external_lex_state = 20},
  [5561] = {.lex_state = 72, .external_lex_state = 20},
  [5562] = {.lex_state = 113, .external_lex_state = 20},
  [5563] = {.lex_state = 113, .external_lex_state = 20},
  [5564] = {.lex_state = 113, .external_lex_state = 28},
  [5565] = {.lex_state = 72, .external_lex_state = 20},
  [5566] = {.lex_state = 72, .external_lex_state = 20},
  [5567] = {.lex_state = 72, .external_lex_state = 20},
  [5568] = {.lex_state = 113, .external_lex_state = 20},
  [5569] = {.lex_state = 113, .external_lex_state = 20},
  [5570] = {.lex_state = 113, .external_lex_state = 20},
  [5571] = {.lex_state = 113, .external_lex_state = 20},
  [5572] = {.lex_state = 113, .external_lex_state = 20},
  [5573] = {.lex_state = 113, .external_lex_state = 20},
  [5574] = {.lex_state = 72, .external_lex_state = 20},
  [5575] = {.lex_state = 113, .external_lex_state = 20},
  [5576] = {.lex_state = 113, .external_lex_state = 20},
  [5577] = {.lex_state = 113, .external_lex_state = 22},
  [5578] = {.lex_state = 113, .external_lex_state = 52},
  [5579] = {.lex_state = 113, .external_lex_state = 22},
  [5580] = {.lex_state = 113, .external_lex_state = 52},
  [5581] = {.lex_state = 113, .external_lex_state = 20},
  [5582] = {.lex_state = 113, .external_lex_state = 52},
  [5583] = {.lex_state = 113, .external_lex_state = 20},
  [5584] = {.lex_state = 113, .external_lex_state = 53},
  [5585] = {.lex_state = 113, .external_lex_state = 53},
  [5586] = {.lex_state = 113, .external_lex_state = 53},
  [5587] = {.lex_state = 113, .external_lex_state = 22},
  [5588] = {.lex_state = 113, .external_lex_state = 20},
  [5589] = {.lex_state = 113, .external_lex_state = 22},
  [5590] = {.lex_state = 113, .external_lex_state = 22},
  [5591] = {.lex_state = 113, .external_lex_state = 22},
  [5592] = {.lex_state = 113, .external_lex_state = 22},
  [5593] = {.lex_state = 113, .external_lex_state = 22},
  [5594] = {.lex_state = 113, .external_lex_state = 52},
  [5595] = {.lex_state = 113, .external_lex_state = 22},
  [5596] = {.lex_state = 113, .external_lex_state = 20},
  [5597] = {.lex_state = 113, .external_lex_state = 22},
  [5598] = {.lex_state = 113, .external_lex_state = 20},
  [5599] = {.lex_state = 13, .external_lex_state = 20},
  [5600] = {.lex_state = 11, .external_lex_state = 20},
  [5601] = {.lex_state = 113, .external_lex_state = 22},
  [5602] = {.lex_state = 113, .external_lex_state = 22},
  [5603] = {.lex_state = 113, .external_lex_state = 20},
  [5604] = {.lex_state = 113, .external_lex_state = 22},
  [5605] = {.lex_state = 113, .external_lex_state = 22},
  [5606] = {.lex_state = 113, .external_lex_state = 20},
  [5607] = {.lex_state = 113, .external_lex_state = 20},
  [5608] = {.lex_state = 113, .external_lex_state = 22},
  [5609] = {.lex_state = 113, .external_lex_state = 20},
  [5610] = {.lex_state = 113, .external_lex_state = 53},
  [5611] = {.lex_state = 113, .external_lex_state = 53},
  [5612] = {.lex_state = 113, .external_lex_state = 53},
  [5613] = {.lex_state = 113, .external_lex_state = 53},
  [5614] = {.lex_state = 113, .external_lex_state = 20},
  [5615] = {.lex_state = 113, .external_lex_state = 52},
  [5616] = {.lex_state = 113, .external_lex_state = 22},
  [5617] = {.lex_state = 113, .external_lex_state = 22},
  [5618] = {.lex_state = 113, .external_lex_state = 20},
  [5619] = {.lex_state = 113, .external_lex_state = 53},
  [5620] = {.lex_state = 113, .external_lex_state = 53},
  [5621] = {.lex_state = 113, .external_lex_state = 53},
  [5622] = {.lex_state = 113, .external_lex_state = 53},
  [5623] = {.lex_state = 11, .external_lex_state = 20},
  [5624] = {.lex_state = 113, .external_lex_state = 22},
  [5625] = {.lex_state = 113, .external_lex_state = 22},
  [5626] = {.lex_state = 113, .external_lex_state = 22},
  [5627] = {.lex_state = 113, .external_lex_state = 22},
  [5628] = {.lex_state = 113, .external_lex_state = 22},
  [5629] = {.lex_state = 113, .external_lex_state = 22},
  [5630] = {.lex_state = 113, .external_lex_state = 22},
  [5631] = {.lex_state = 113, .external_lex_state = 22},
  [5632] = {.lex_state = 113, .external_lex_state = 22},
  [5633] = {.lex_state = 113, .external_lex_state = 53},
  [5634] = {.lex_state = 113, .external_lex_state = 53},
  [5635] = {.lex_state = 113, .external_lex_state = 53},
  [5636] = {.lex_state = 113, .external_lex_state = 53},
  [5637] = {.lex_state = 113, .external_lex_state = 22},
  [5638] = {.lex_state = 113, .external_lex_state = 53},
  [5639] = {.lex_state = 113, .external_lex_state = 53},
  [5640] = {.lex_state = 113, .external_lex_state = 20},
  [5641] = {.lex_state = 113, .external_lex_state = 52},
  [5642] = {.lex_state = 113, .external_lex_state = 22},
  [5643] = {.lex_state = 113, .external_lex_state = 20},
  [5644] = {.lex_state = 113, .external_lex_state = 20},
  [5645] = {.lex_state = 113, .external_lex_state = 22},
  [5646] = {.lex_state = 113, .external_lex_state = 22},
  [5647] = {.lex_state = 113, .external_lex_state = 20},
  [5648] = {.lex_state = 113, .external_lex_state = 53},
  [5649] = {.lex_state = 113, .external_lex_state = 53},
  [5650] = {.lex_state = 113, .external_lex_state = 53},
  [5651] = {.lex_state = 113, .external_lex_state = 52},
  [5652] = {.lex_state = 113, .external_lex_state = 22},
  [5653] = {.lex_state = 11, .external_lex_state = 20},
  [5654] = {.lex_state = 113, .external_lex_state = 20},
  [5655] = {.lex_state = 113, .external_lex_state = 22},
  [5656] = {.lex_state = 113, .external_lex_state = 22},
  [5657] = {.lex_state = 113, .external_lex_state = 22},
  [5658] = {.lex_state = 113, .external_lex_state = 22},
  [5659] = {.lex_state = 113, .external_lex_state = 22},
  [5660] = {.lex_state = 11, .external_lex_state = 20},
  [5661] = {.lex_state = 113, .external_lex_state = 20},
  [5662] = {.lex_state = 113, .external_lex_state = 22},
  [5663] = {.lex_state = 113, .external_lex_state = 22},
  [5664] = {.lex_state = 113, .external_lex_state = 22},
  [5665] = {.lex_state = 113, .external_lex_state = 51},
  [5666] = {.lex_state = 113, .external_lex_state = 20},
  [5667] = {.lex_state = 113, .external_lex_state = 22},
  [5668] = {.lex_state = 113, .external_lex_state = 20},
  [5669] = {.lex_state = 11, .external_lex_state = 20},
  [5670] = {.lex_state = 113, .external_lex_state = 22},
  [5671] = {.lex_state = 11, .external_lex_state = 20},
  [5672] = {.lex_state = 113, .external_lex_state = 20},
  [5673] = {.lex_state = 113, .external_lex_state = 22},
  [5674] = {.lex_state = 113, .external_lex_state = 52},
  [5675] = {.lex_state = 113, .external_lex_state = 22},
  [5676] = {.lex_state = 113, .external_lex_state = 22},
  [5677] = {.lex_state = 11, .external_lex_state = 20},
  [5678] = {.lex_state = 113, .external_lex_state = 22},
  [5679] = {.lex_state = 113, .external_lex_state = 22},
  [5680] = {.lex_state = 113, .external_lex_state = 20},
  [5681] = {.lex_state = 113, .external_lex_state = 20},
  [5682] = {.lex_state = 113, .external_lex_state = 20},
  [5683] = {.lex_state = 113, .external_lex_state = 20},
  [5684] = {.lex_state = 113, .external_lex_state = 20},
  [5685] = {.lex_state = 113, .external_lex_state = 20},
  [5686] = {.lex_state = 113, .external_lex_state = 22},
  [5687] = {.lex_state = 113, .external_lex_state = 20},
  [5688] = {.lex_state = 113, .external_lex_state = 20},
  [5689] = {.lex_state = 113, .external_lex_state = 22},
  [5690] = {.lex_state = 113, .external_lex_state = 53},
  [5691] = {.lex_state = 113, .external_lex_state = 22},
  [5692] = {.lex_state = 113, .external_lex_state = 22},
  [5693] = {.lex_state = 113, .external_lex_state = 22},
  [5694] = {.lex_state = 113, .external_lex_state = 20},
  [5695] = {.lex_state = 113, .external_lex_state = 20},
  [5696] = {.lex_state = 113, .external_lex_state = 20},
  [5697] = {.lex_state = 113, .external_lex_state = 22},
  [5698] = {.lex_state = 113, .external_lex_state = 22},
  [5699] = {.lex_state = 113, .external_lex_state = 22},
  [5700] = {.lex_state = 113, .external_lex_state = 22},
  [5701] = {.lex_state = 113, .external_lex_state = 20},
  [5702] = {.lex_state = 113, .external_lex_state = 22},
  [5703] = {.lex_state = 113, .external_lex_state = 20},
  [5704] = {.lex_state = 113, .external_lex_state = 20},
  [5705] = {.lex_state = 113, .external_lex_state = 20},
  [5706] = {.lex_state = 113, .external_lex_state = 22},
  [5707] = {.lex_state = 113, .external_lex_state = 20},
  [5708] = {.lex_state = 113, .external_lex_state = 20},
  [5709] = {.lex_state = 113, .external_lex_state = 20},
  [5710] = {.lex_state = 113, .external_lex_state = 20},
  [5711] = {.lex_state = 113, .external_lex_state = 53},
  [5712] = {.lex_state = 113, .external_lex_state = 53},
  [5713] = {.lex_state = 113, .external_lex_state = 20},
  [5714] = {.lex_state = 113, .external_lex_state = 20},
  [5715] = {.lex_state = 113, .external_lex_state = 20},
  [5716] = {.lex_state = 113, .external_lex_state = 53},
  [5717] = {.lex_state = 113, .external_lex_state = 20},
  [5718] = {.lex_state = 113, .external_lex_state = 20},
  [5719] = {.lex_state = 113, .external_lex_state = 20},
  [5720] = {.lex_state = 113, .external_lex_state = 22},
  [5721] = {.lex_state = 113, .external_lex_state = 20},
  [5722] = {.lex_state = 113, .external_lex_state = 20},
  [5723] = {.lex_state = 113, .external_lex_state = 20},
  [5724] = {.lex_state = 113, .external_lex_state = 20},
  [5725] = {.lex_state = 11, .external_lex_state = 20},
  [5726] = {.lex_state = 113, .external_lex_state = 20},
  [5727] = {.lex_state = 113, .external_lex_state = 20},
  [5728] = {.lex_state = 11, .external_lex_state = 20},
  [5729] = {.lex_state = 50, .external_lex_state = 20},
  [5730] = {.lex_state = 113, .external_lex_state = 54},
  [5731] = {.lex_state = 113, .external_lex_state = 42},
  [5732] = {.lex_state = 113, .external_lex_state = 20},
  [5733] = {.lex_state = 11, .external_lex_state = 20},
  [5734] = {.lex_state = 113, .external_lex_state = 20},
  [5735] = {.lex_state = 11, .external_lex_state = 20},
  [5736] = {.lex_state = 113, .external_lex_state = 20},
  [5737] = {.lex_state = 113, .external_lex_state = 55},
  [5738] = {.lex_state = 113, .external_lex_state = 20},
  [5739] = {.lex_state = 113, .external_lex_state = 20},
  [5740] = {.lex_state = 113, .external_lex_state = 39},
  [5741] = {.lex_state = 113, .external_lex_state = 20},
  [5742] = {.lex_state = 113, .external_lex_state = 20},
  [5743] = {.lex_state = 113, .external_lex_state = 20},
  [5744] = {.lex_state = 113, .external_lex_state = 20},
  [5745] = {.lex_state = 113, .external_lex_state = 20},
  [5746] = {.lex_state = 113, .external_lex_state = 20},
  [5747] = {.lex_state = 113, .external_lex_state = 20},
  [5748] = {.lex_state = 113, .external_lex_state = 20},
  [5749] = {.lex_state = 113, .external_lex_state = 20},
  [5750] = {.lex_state = 113, .external_lex_state = 20},
  [5751] = {.lex_state = 113, .external_lex_state = 20},
  [5752] = {.lex_state = 113, .external_lex_state = 39},
  [5753] = {.lex_state = 262, .external_lex_state = 20},
  [5754] = {.lex_state = 113, .external_lex_state = 20},
  [5755] = {.lex_state = 113, .external_lex_state = 56},
  [5756] = {.lex_state = 113, .external_lex_state = 20},
  [5757] = {.lex_state = 113, .external_lex_state = 54},
  [5758] = {.lex_state = 113, .external_lex_state = 20},
  [5759] = {.lex_state = 113, .external_lex_state = 20},
  [5760] = {.lex_state = 113, .external_lex_state = 54},
  [5761] = {.lex_state = 113, .external_lex_state = 55},
  [5762] = {.lex_state = 113, .external_lex_state = 20},
  [5763] = {.lex_state = 113, .external_lex_state = 20},
  [5764] = {.lex_state = 11, .external_lex_state = 20},
  [5765] = {.lex_state = 113, .external_lex_state = 20},
  [5766] = {.lex_state = 50, .external_lex_state = 20},
  [5767] = {.lex_state = 113, .external_lex_state = 54},
  [5768] = {.lex_state = 113, .external_lex_state = 57},
  [5769] = {.lex_state = 113, .external_lex_state = 20},
  [5770] = {.lex_state = 113, .external_lex_state = 20},
  [5771] = {.lex_state = 13, .external_lex_state = 20},
  [5772] = {.lex_state = 11, .external_lex_state = 20},
  [5773] = {.lex_state = 113, .external_lex_state = 20},
  [5774] = {.lex_state = 113, .external_lex_state = 20},
  [5775] = {.lex_state = 113, .external_lex_state = 20},
  [5776] = {.lex_state = 113, .external_lex_state = 20},
  [5777] = {.lex_state = 113, .external_lex_state = 39},
  [5778] = {.lex_state = 113, .external_lex_state = 20},
  [5779] = {.lex_state = 113, .external_lex_state = 58},
  [5780] = {.lex_state = 113, .external_lex_state = 55},
  [5781] = {.lex_state = 113, .external_lex_state = 20},
  [5782] = {.lex_state = 113, .external_lex_state = 57},
  [5783] = {.lex_state = 113, .external_lex_state = 56},
  [5784] = {.lex_state = 113, .external_lex_state = 20},
  [5785] = {.lex_state = 113, .external_lex_state = 20},
  [5786] = {.lex_state = 113, .external_lex_state = 20},
  [5787] = {.lex_state = 113, .external_lex_state = 20},
  [5788] = {.lex_state = 13, .external_lex_state = 20},
  [5789] = {.lex_state = 113, .external_lex_state = 58},
  [5790] = {.lex_state = 113, .external_lex_state = 20},
  [5791] = {.lex_state = 113, .external_lex_state = 20},
  [5792] = {.lex_state = 113, .external_lex_state = 20},
  [5793] = {.lex_state = 113, .external_lex_state = 55},
  [5794] = {.lex_state = 113, .external_lex_state = 57},
  [5795] = {.lex_state = 113, .external_lex_state = 20},
  [5796] = {.lex_state = 113, .external_lex_state = 20},
  [5797] = {.lex_state = 113, .external_lex_state = 20},
  [5798] = {.lex_state = 113, .external_lex_state = 56},
  [5799] = {.lex_state = 113, .external_lex_state = 20},
  [5800] = {.lex_state = 113, .external_lex_state = 20},
  [5801] = {.lex_state = 113, .external_lex_state = 20},
  [5802] = {.lex_state = 113, .external_lex_state = 39},
  [5803] = {.lex_state = 113, .external_lex_state = 20},
  [5804] = {.lex_state = 113, .external_lex_state = 20},
  [5805] = {.lex_state = 113, .external_lex_state = 20},
  [5806] = {.lex_state = 11, .external_lex_state = 20},
  [5807] = {.lex_state = 113, .external_lex_state = 20},
  [5808] = {.lex_state = 113, .external_lex_state = 20},
  [5809] = {.lex_state = 11, .external_lex_state = 20},
  [5810] = {.lex_state = 113, .external_lex_state = 20},
  [5811] = {.lex_state = 113, .external_lex_state = 57},
  [5812] = {.lex_state = 113, .external_lex_state = 20},
  [5813] = {.lex_state = 113, .external_lex_state = 20},
  [5814] = {.lex_state = 113, .external_lex_state = 20},
  [5815] = {.lex_state = 113, .external_lex_state = 20},
  [5816] = {.lex_state = 13, .external_lex_state = 20},
  [5817] = {.lex_state = 13, .external_lex_state = 20},
  [5818] = {.lex_state = 113, .external_lex_state = 20},
  [5819] = {.lex_state = 113, .external_lex_state = 20},
  [5820] = {.lex_state = 113, .external_lex_state = 54},
  [5821] = {.lex_state = 113, .external_lex_state = 20},
  [5822] = {.lex_state = 113, .external_lex_state = 20},
  [5823] = {.lex_state = 113, .external_lex_state = 55},
  [5824] = {.lex_state = 262, .external_lex_state = 20},
  [5825] = {.lex_state = 113, .external_lex_state = 20},
  [5826] = {.lex_state = 113, .external_lex_state = 20},
  [5827] = {.lex_state = 113, .external_lex_state = 20},
  [5828] = {.lex_state = 113, .external_lex_state = 20},
  [5829] = {.lex_state = 113, .external_lex_state = 20},
  [5830] = {.lex_state = 113, .external_lex_state = 20},
  [5831] = {.lex_state = 113, .external_lex_state = 56},
  [5832] = {.lex_state = 113, .external_lex_state = 57},
  [5833] = {.lex_state = 113, .external_lex_state = 55},
  [5834] = {.lex_state = 113, .external_lex_state = 20},
  [5835] = {.lex_state = 113, .external_lex_state = 58},
  [5836] = {.lex_state = 113, .external_lex_state = 54},
  [5837] = {.lex_state = 113, .external_lex_state = 20},
  [5838] = {.lex_state = 113, .external_lex_state = 20},
  [5839] = {.lex_state = 113, .external_lex_state = 20},
  [5840] = {.lex_state = 13, .external_lex_state = 20},
  [5841] = {.lex_state = 113, .external_lex_state = 56},
  [5842] = {.lex_state = 113, .external_lex_state = 39},
  [5843] = {.lex_state = 113, .external_lex_state = 20},
  [5844] = {.lex_state = 113, .external_lex_state = 55},
  [5845] = {.lex_state = 113, .external_lex_state = 20},
  [5846] = {.lex_state = 113, .external_lex_state = 20},
  [5847] = {.lex_state = 113, .external_lex_state = 56},
  [5848] = {.lex_state = 13, .external_lex_state = 20},
  [5849] = {.lex_state = 113, .external_lex_state = 55},
  [5850] = {.lex_state = 113, .external_lex_state = 58},
  [5851] = {.lex_state = 113, .external_lex_state = 20},
  [5852] = {.lex_state = 113, .external_lex_state = 56},
  [5853] = {.lex_state = 113, .external_lex_state = 20},
  [5854] = {.lex_state = 113, .external_lex_state = 55},
  [5855] = {.lex_state = 113, .external_lex_state = 20},
  [5856] = {.lex_state = 113, .external_lex_state = 20},
  [5857] = {.lex_state = 113, .external_lex_state = 56},
  [5858] = {.lex_state = 11, .external_lex_state = 20},
  [5859] = {.lex_state = 113, .external_lex_state = 55},
  [5860] = {.lex_state = 113, .external_lex_state = 54},
  [5861] = {.lex_state = 113, .external_lex_state = 20},
  [5862] = {.lex_state = 113, .external_lex_state = 56},
  [5863] = {.lex_state = 113, .external_lex_state = 20},
  [5864] = {.lex_state = 113, .external_lex_state = 55},
  [5865] = {.lex_state = 113, .external_lex_state = 20},
  [5866] = {.lex_state = 113, .external_lex_state = 54},
  [5867] = {.lex_state = 113, .external_lex_state = 20},
  [5868] = {.lex_state = 113, .external_lex_state = 55},
  [5869] = {.lex_state = 113, .external_lex_state = 54},
  [5870] = {.lex_state = 113, .external_lex_state = 20},
  [5871] = {.lex_state = 113, .external_lex_state = 55},
  [5872] = {.lex_state = 113, .external_lex_state = 20},
  [5873] = {.lex_state = 113, .external_lex_state = 55},
  [5874] = {.lex_state = 262, .external_lex_state = 20},
  [5875] = {.lex_state = 113, .external_lex_state = 20},
  [5876] = {.lex_state = 113, .external_lex_state = 55},
  [5877] = {.lex_state = 113, .external_lex_state = 20},
  [5878] = {.lex_state = 113, .external_lex_state = 20},
  [5879] = {.lex_state = 113, .external_lex_state = 55},
  [5880] = {.lex_state = 113, .external_lex_state = 58},
  [5881] = {.lex_state = 11, .external_lex_state = 20},
  [5882] = {.lex_state = 113, .external_lex_state = 55},
  [5883] = {.lex_state = 11, .external_lex_state = 20},
  [5884] = {.lex_state = 113, .external_lex_state = 20},
  [5885] = {.lex_state = 113, .external_lex_state = 55},
  [5886] = {.lex_state = 113, .external_lex_state = 39},
  [5887] = {.lex_state = 113, .external_lex_state = 20},
  [5888] = {.lex_state = 113, .external_lex_state = 55},
  [5889] = {.lex_state = 113, .external_lex_state = 54},
  [5890] = {.lex_state = 113, .external_lex_state = 20},
  [5891] = {.lex_state = 113, .external_lex_state = 55},
  [5892] = {.lex_state = 113, .external_lex_state = 20},
  [5893] = {.lex_state = 113, .external_lex_state = 20},
  [5894] = {.lex_state = 113, .external_lex_state = 55},
  [5895] = {.lex_state = 113, .external_lex_state = 20},
  [5896] = {.lex_state = 113, .external_lex_state = 20},
  [5897] = {.lex_state = 113, .external_lex_state = 55},
  [5898] = {.lex_state = 113, .external_lex_state = 56},
  [5899] = {.lex_state = 11, .external_lex_state = 20},
  [5900] = {.lex_state = 113, .external_lex_state = 55},
  [5901] = {.lex_state = 113, .external_lex_state = 20},
  [5902] = {.lex_state = 113, .external_lex_state = 20},
  [5903] = {.lex_state = 113, .external_lex_state = 55},
  [5904] = {.lex_state = 113, .external_lex_state = 58},
  [5905] = {.lex_state = 113, .external_lex_state = 20},
  [5906] = {.lex_state = 113, .external_lex_state = 55},
  [5907] = {.lex_state = 113, .external_lex_state = 20},
  [5908] = {.lex_state = 113, .external_lex_state = 20},
  [5909] = {.lex_state = 113, .external_lex_state = 55},
  [5910] = {.lex_state = 3, .external_lex_state = 20},
  [5911] = {.lex_state = 113, .external_lex_state = 20},
  [5912] = {.lex_state = 113, .external_lex_state = 55},
  [5913] = {.lex_state = 113, .external_lex_state = 20},
  [5914] = {.lex_state = 113, .external_lex_state = 20},
  [5915] = {.lex_state = 113, .external_lex_state = 55},
  [5916] = {.lex_state = 113, .external_lex_state = 20},
  [5917] = {.lex_state = 113, .external_lex_state = 20},
  [5918] = {.lex_state = 113, .external_lex_state = 55},
  [5919] = {.lex_state = 113, .external_lex_state = 20},
  [5920] = {.lex_state = 113, .external_lex_state = 20},
  [5921] = {.lex_state = 113, .external_lex_state = 55},
  [5922] = {.lex_state = 113, .external_lex_state = 20},
  [5923] = {.lex_state = 113, .external_lex_state = 20},
  [5924] = {.lex_state = 113, .external_lex_state = 55},
  [5925] = {.lex_state = 11, .external_lex_state = 20},
  [5926] = {.lex_state = 113, .external_lex_state = 20},
  [5927] = {.lex_state = 113, .external_lex_state = 55},
  [5928] = {.lex_state = 113, .external_lex_state = 20},
  [5929] = {.lex_state = 11, .external_lex_state = 20},
  [5930] = {.lex_state = 113, .external_lex_state = 55},
  [5931] = {.lex_state = 13, .external_lex_state = 20},
  [5932] = {.lex_state = 113, .external_lex_state = 20},
  [5933] = {.lex_state = 113, .external_lex_state = 55},
  [5934] = {.lex_state = 113, .external_lex_state = 20},
  [5935] = {.lex_state = 113, .external_lex_state = 20},
  [5936] = {.lex_state = 113, .external_lex_state = 55},
  [5937] = {.lex_state = 113, .external_lex_state = 20},
  [5938] = {.lex_state = 113, .external_lex_state = 20},
  [5939] = {.lex_state = 113, .external_lex_state = 20},
  [5940] = {.lex_state = 113, .external_lex_state = 20},
  [5941] = {.lex_state = 113, .external_lex_state = 54},
  [5942] = {.lex_state = 113, .external_lex_state = 57},
  [5943] = {.lex_state = 113, .external_lex_state = 20},
  [5944] = {.lex_state = 11, .external_lex_state = 20},
  [5945] = {.lex_state = 113, .external_lex_state = 20},
  [5946] = {.lex_state = 113, .external_lex_state = 54},
  [5947] = {.lex_state = 113, .external_lex_state = 54},
  [5948] = {.lex_state = 113, .external_lex_state = 20},
  [5949] = {.lex_state = 11, .external_lex_state = 20},
  [5950] = {.lex_state = 113, .external_lex_state = 54},
  [5951] = {.lex_state = 113, .external_lex_state = 20},
  [5952] = {.lex_state = 113, .external_lex_state = 42},
  [5953] = {.lex_state = 113, .external_lex_state = 20},
  [5954] = {.lex_state = 113, .external_lex_state = 54},
  [5955] = {.lex_state = 113, .external_lex_state = 20},
  [5956] = {.lex_state = 113, .external_lex_state = 55},
  [5957] = {.lex_state = 113, .external_lex_state = 56},
  [5958] = {.lex_state = 113, .external_lex_state = 20},
  [5959] = {.lex_state = 113, .external_lex_state = 54},
  [5960] = {.lex_state = 113, .external_lex_state = 54},
  [5961] = {.lex_state = 113, .external_lex_state = 20},
  [5962] = {.lex_state = 113, .external_lex_state = 54},
  [5963] = {.lex_state = 113, .external_lex_state = 20},
  [5964] = {.lex_state = 113, .external_lex_state = 20},
  [5965] = {.lex_state = 113, .external_lex_state = 58},
  [5966] = {.lex_state = 113, .external_lex_state = 20},
  [5967] = {.lex_state = 113, .external_lex_state = 20},
  [5968] = {.lex_state = 113, .external_lex_state = 20},
  [5969] = {.lex_state = 113, .external_lex_state = 20},
  [5970] = {.lex_state = 113, .external_lex_state = 20},
  [5971] = {.lex_state = 113, .external_lex_state = 20},
  [5972] = {.lex_state = 113, .external_lex_state = 55},
  [5973] = {.lex_state = 113, .external_lex_state = 20},
  [5974] = {.lex_state = 113, .external_lex_state = 54},
  [5975] = {.lex_state = 113, .external_lex_state = 20},
  [5976] = {.lex_state = 113, .external_lex_state = 20},
  [5977] = {.lex_state = 113, .external_lex_state = 20},
  [5978] = {.lex_state = 113, .external_lex_state = 55},
  [5979] = {.lex_state = 113, .external_lex_state = 39},
  [5980] = {.lex_state = 113, .external_lex_state = 55},
  [5981] = {.lex_state = 113, .external_lex_state = 20},
  [5982] = {.lex_state = 13, .external_lex_state = 20},
  [5983] = {.lex_state = 113, .external_lex_state = 20},
  [5984] = {.lex_state = 113, .external_lex_state = 20},
  [5985] = {.lex_state = 113, .external_lex_state = 54},
  [5986] = {.lex_state = 113, .external_lex_state = 54},
  [5987] = {.lex_state = 113, .external_lex_state = 20},
  [5988] = {.lex_state = 113, .external_lex_state = 20},
  [5989] = {.lex_state = 113, .external_lex_state = 20},
  [5990] = {.lex_state = 113, .external_lex_state = 54},
  [5991] = {.lex_state = 113, .external_lex_state = 20},
  [5992] = {.lex_state = 113, .external_lex_state = 39},
  [5993] = {.lex_state = 11, .external_lex_state = 20},
  [5994] = {.lex_state = 113, .external_lex_state = 20},
  [5995] = {.lex_state = 113, .external_lex_state = 20},
  [5996] = {.lex_state = 262, .external_lex_state = 20},
  [5997] = {.lex_state = 113, .external_lex_state = 20},
  [5998] = {.lex_state = 113, .external_lex_state = 20},
  [5999] = {.lex_state = 113, .external_lex_state = 20},
  [6000] = {.lex_state = 113, .external_lex_state = 57},
  [6001] = {.lex_state = 113, .external_lex_state = 20},
  [6002] = {.lex_state = 113, .external_lex_state = 20},
  [6003] = {.lex_state = 113, .external_lex_state = 56},
  [6004] = {.lex_state = 113, .external_lex_state = 20},
  [6005] = {.lex_state = 113, .external_lex_state = 42},
  [6006] = {.lex_state = 113, .external_lex_state = 42},
  [6007] = {.lex_state = 113, .external_lex_state = 58},
  [6008] = {.lex_state = 113, .external_lex_state = 20},
  [6009] = {.lex_state = 11, .external_lex_state = 20},
  [6010] = {.lex_state = 113, .external_lex_state = 20},
  [6011] = {.lex_state = 113, .external_lex_state = 20},
  [6012] = {.lex_state = 113, .external_lex_state = 20},
  [6013] = {.lex_state = 113, .external_lex_state = 20},
  [6014] = {.lex_state = 113, .external_lex_state = 20},
  [6015] = {.lex_state = 113, .external_lex_state = 20},
  [6016] = {.lex_state = 113, .external_lex_state = 20},
  [6017] = {.lex_state = 113, .external_lex_state = 20},
  [6018] = {.lex_state = 113, .external_lex_state = 20},
  [6019] = {.lex_state = 113, .external_lex_state = 20},
  [6020] = {.lex_state = 113, .external_lex_state = 20},
  [6021] = {.lex_state = 113, .external_lex_state = 20},
  [6022] = {.lex_state = 113, .external_lex_state = 20},
  [6023] = {.lex_state = 113, .external_lex_state = 20},
  [6024] = {.lex_state = 113, .external_lex_state = 20},
  [6025] = {.lex_state = 113, .external_lex_state = 20},
  [6026] = {.lex_state = 113, .external_lex_state = 20},
  [6027] = {.lex_state = 113, .external_lex_state = 20},
  [6028] = {.lex_state = 113, .external_lex_state = 20},
  [6029] = {.lex_state = 113, .external_lex_state = 20},
  [6030] = {.lex_state = 113, .external_lex_state = 20},
  [6031] = {.lex_state = 113, .external_lex_state = 20},
  [6032] = {.lex_state = 113, .external_lex_state = 20},
  [6033] = {.lex_state = 113, .external_lex_state = 20},
  [6034] = {.lex_state = 113, .external_lex_state = 20},
  [6035] = {.lex_state = 113, .external_lex_state = 20},
  [6036] = {.lex_state = 113, .external_lex_state = 20},
  [6037] = {.lex_state = 113, .external_lex_state = 20},
  [6038] = {.lex_state = 113, .external_lex_state = 57},
  [6039] = {.lex_state = 113, .external_lex_state = 20},
  [6040] = {.lex_state = 113, .external_lex_state = 20},
  [6041] = {.lex_state = 113, .external_lex_state = 54},
  [6042] = {.lex_state = 113, .external_lex_state = 54},
  [6043] = {.lex_state = 113, .external_lex_state = 20},
  [6044] = {.lex_state = 3, .external_lex_state = 20},
  [6045] = {.lex_state = 113, .external_lex_state = 39},
  [6046] = {.lex_state = 113, .external_lex_state = 20},
  [6047] = {(TSStateId)(-1),},
  [6048] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_heredoc_content] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__identifier] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_package] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_role] = ACTIONS(1),
    [anon_sym_BUILD] = ACTIONS(1),
    [anon_sym_ADJUST] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(1),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_my] = ACTIONS(1),
    [anon_sym_extended] = ACTIONS(1),
    [anon_sym_async] = ACTIONS(1),
    [anon_sym_sub] = ACTIONS(1),
    [anon_sym_method] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_state] = ACTIONS(1),
    [anon_sym_our] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_catch] = ACTIONS(1),
    [anon_sym_finally] = ACTIONS(1),
    [anon_sym_defer] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_elsif] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_xor] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_DOT_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_x_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_SLASH_SLASH] = ACTIONS(1),
    [anon_sym_CARET_CARET] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_GT_GT2] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_x] = ACTIONS(1),
    [anon_sym_EQ_TILDE] = ACTIONS(1),
    [anon_sym_BANG_TILDE] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_eq] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_equ] = ACTIONS(1),
    [anon_sym_eqr] = ACTIONS(1),
    [anon_sym_ne] = ACTIONS(1),
    [anon_sym_LT_EQ_GT] = ACTIONS(1),
    [anon_sym_cmp] = ACTIONS(1),
    [anon_sym_TILDE_TILDE] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_lt] = ACTIONS(1),
    [anon_sym_le] = ACTIONS(1),
    [anon_sym_ge] = ACTIONS(1),
    [anon_sym_gt] = ACTIONS(1),
    [anon_sym_isa] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_eval] = ACTIONS(1),
    [anon_sym_await] = ACTIONS(1),
    [anon_sym_field] = ACTIONS(1),
    [anon_sym_local] = ACTIONS(1),
    [anon_sym_dynamically] = ACTIONS(1),
    [anon_sym_DOLLAR_POUND] = ACTIONS(1),
    [anon_sym_require] = ACTIONS(1),
    [anon_sym_map] = ACTIONS(1),
    [anon_sym_grep] = ACTIONS(1),
    [anon_sym_sort] = ACTIONS(1),
    [anon_sym_goto] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_undef] = ACTIONS(1),
    [anon_sym_PERCENT2] = ACTIONS(1),
    [anon_sym_AMP2] = ACTIONS(1),
    [anon_sym_STAR2] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [anon_sym_use] = ACTIONS(1),
    [anon_sym_no] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_foreach] = ACTIONS(1),
    [anon_sym_last] = ACTIONS(1),
    [anon_sym_next] = ACTIONS(1),
    [anon_sym_redo] = ACTIONS(1),
    [anon_sym_BEGIN] = ACTIONS(1),
    [anon_sym_INIT] = ACTIONS(1),
    [anon_sym_CHECK] = ACTIONS(1),
    [anon_sym_UNITCHECK] = ACTIONS(1),
    [anon_sym_END] = ACTIONS(1),
    [anon_sym___FILE__] = ACTIONS(1),
    [anon_sym___LINE__] = ACTIONS(1),
    [anon_sym___PACKAGE__] = ACTIONS(1),
    [anon_sym___SUB__] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_fork] = ACTIONS(1),
    [anon_sym_getppid] = ACTIONS(1),
    [anon_sym_time] = ACTIONS(1),
    [anon_sym_times] = ACTIONS(1),
    [anon_sym_wait] = ACTIONS(1),
    [anon_sym_wantarray] = ACTIONS(1),
    [anon_sym_abs] = ACTIONS(1),
    [anon_sym_alarm] = ACTIONS(1),
    [anon_sym_chop] = ACTIONS(1),
    [anon_sym_chdir] = ACTIONS(1),
    [anon_sym_close] = ACTIONS(1),
    [anon_sym_closedir] = ACTIONS(1),
    [anon_sym_caller] = ACTIONS(1),
    [anon_sym_chomp] = ACTIONS(1),
    [anon_sym_chr] = ACTIONS(1),
    [anon_sym_cos] = ACTIONS(1),
    [anon_sym_chroot] = ACTIONS(1),
    [anon_sym_defined] = ACTIONS(1),
    [anon_sym_delete] = ACTIONS(1),
    [anon_sym_dbmclose] = ACTIONS(1),
    [anon_sym_exists] = ACTIONS(1),
    [anon_sym_exit] = ACTIONS(1),
    [anon_sym_eof] = ACTIONS(1),
    [anon_sym_exp] = ACTIONS(1),
    [anon_sym_each] = ACTIONS(1),
    [anon_sym_fc] = ACTIONS(1),
    [anon_sym_fileno] = ACTIONS(1),
    [anon_sym_gmtime] = ACTIONS(1),
    [anon_sym_getc] = ACTIONS(1),
    [anon_sym_getpgrp] = ACTIONS(1),
    [anon_sym_getprotobyname] = ACTIONS(1),
    [anon_sym_getpwname] = ACTIONS(1),
    [anon_sym_getpwuid] = ACTIONS(1),
    [anon_sym_getpeername] = ACTIONS(1),
    [anon_sym_getnetbyname] = ACTIONS(1),
    [anon_sym_getsockname] = ACTIONS(1),
    [anon_sym_getgrnam] = ACTIONS(1),
    [anon_sym_getgrgid] = ACTIONS(1),
    [anon_sym_hex] = ACTIONS(1),
    [anon_sym_int] = ACTIONS(1),
    [anon_sym_keys] = ACTIONS(1),
    [anon_sym_lc] = ACTIONS(1),
    [anon_sym_lcfirst] = ACTIONS(1),
    [anon_sym_length] = ACTIONS(1),
    [anon_sym_localtime] = ACTIONS(1),
    [anon_sym_log] = ACTIONS(1),
    [anon_sym_lock] = ACTIONS(1),
    [anon_sym_lstat] = ACTIONS(1),
    [anon_sym_oct] = ACTIONS(1),
    [anon_sym_ord] = ACTIONS(1),
    [anon_sym_prototype] = ACTIONS(1),
    [anon_sym_pop] = ACTIONS(1),
    [anon_sym_pos] = ACTIONS(1),
    [anon_sym_quotemeta] = ACTIONS(1),
    [anon_sym_reset] = ACTIONS(1),
    [anon_sym_rand] = ACTIONS(1),
    [anon_sym_rmdir] = ACTIONS(1),
    [anon_sym_readdir] = ACTIONS(1),
    [anon_sym_readline] = ACTIONS(1),
    [anon_sym_readpipe] = ACTIONS(1),
    [anon_sym_rewinddir] = ACTIONS(1),
    [anon_sym_readlink] = ACTIONS(1),
    [anon_sym_ref] = ACTIONS(1),
    [anon_sym_scalar] = ACTIONS(1),
    [anon_sym_shift] = ACTIONS(1),
    [anon_sym_sin] = ACTIONS(1),
    [anon_sym_sleep] = ACTIONS(1),
    [anon_sym_sqrt] = ACTIONS(1),
    [anon_sym_srand] = ACTIONS(1),
    [anon_sym_stat] = ACTIONS(1),
    [anon_sym_study] = ACTIONS(1),
    [anon_sym_tell] = ACTIONS(1),
    [anon_sym_telldir] = ACTIONS(1),
    [anon_sym_tied] = ACTIONS(1),
    [anon_sym_uc] = ACTIONS(1),
    [anon_sym_ucfirst] = ACTIONS(1),
    [anon_sym_untie] = ACTIONS(1),
    [anon_sym_umask] = ACTIONS(1),
    [anon_sym_values] = ACTIONS(1),
    [anon_sym_write] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(1),
    [anon_sym___END__] = ACTIONS(1),
    [anon_sym_EOT] = ACTIONS(1),
    [anon_sym_q] = ACTIONS(1),
    [anon_sym_qq] = ACTIONS(1),
    [anon_sym_LBRACK2] = ACTIONS(1),
    [anon_sym_LBRACE2] = ACTIONS(1),
    [anon_sym_DOLLAR_STAR] = ACTIONS(1),
    [anon_sym_AT_STAR] = ACTIONS(1),
    [anon_sym_qw] = ACTIONS(1),
    [anon_sym_qx] = ACTIONS(1),
    [anon_sym_qr] = ACTIONS(1),
    [anon_sym_m] = ACTIONS(1),
    [anon_sym_s] = ACTIONS(1),
    [sym_quoted_regexp_modifiers] = ACTIONS(1),
    [anon_sym_tr] = ACTIONS(1),
    [anon_sym_y] = ACTIONS(1),
    [sym_version] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_unless] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_until] = ACTIONS(1),
    [aux_sym__bareword_token1] = ACTIONS(1),
    [sym_number] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [sym__single_quote] = ACTIONS(1),
    [sym__double_quote] = ACTIONS(1),
    [sym__backtick_quote] = ACTIONS(1),
    [sym__search_slash_quote] = ACTIONS(1),
    [sym__no_search_slash_plz] = ACTIONS(1),
    [sym__open_readline_bracket] = ACTIONS(1),
    [sym__open_fileglob_bracket] = ACTIONS(1),
    [sym__PERLY_SEMICOLON] = ACTIONS(1),
    [sym__PERLY_HEREDOC] = ACTIONS(1),
    [sym__ctrl_z_hack] = ACTIONS(1),
    [sym__quotelike_begin_quote] = ACTIONS(1),
    [sym__quotelike_middle_close_quote] = ACTIONS(1),
    [sym__quotelike_middle_skip] = ACTIONS(1),
    [sym__quotelike_end_zw] = ACTIONS(1),
    [sym__quotelike_end_quote] = ACTIONS(1),
    [sym__q_string_content] = ACTIONS(1),
    [sym__qq_string_content] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [sym_escaped_delimiter] = ACTIONS(1),
    [sym__dollar_in_regexp] = ACTIONS(1),
    [sym_pod] = ACTIONS(5),
    [sym__gobbled_content] = ACTIONS(1),
    [sym__attribute_value_begin] = ACTIONS(1),
    [sym_attribute_value] = ACTIONS(1),
    [sym_prototype] = ACTIONS(1),
    [sym__signature_start] = ACTIONS(1),
    [sym__heredoc_delimiter] = ACTIONS(1),
    [sym__command_heredoc_delimiter] = ACTIONS(1),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__heredoc_middle] = ACTIONS(1),
    [sym_heredoc_end] = ACTIONS(1),
    [sym__fat_comma_autoquoted] = ACTIONS(1),
    [sym__filetest] = ACTIONS(1),
    [sym__brace_autoquoted_token] = ACTIONS(1),
    [sym__brace_end_zw] = ACTIONS(1),
    [sym__dollar_ident_zw] = ACTIONS(1),
    [sym__no_interp_whitespace_zw] = ACTIONS(1),
    [sym__NONASSOC] = ACTIONS(1),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(1),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(1),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(1),
    [sym__RECOVER_ARROW] = ACTIONS(1),
    [sym__ERROR] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(6021),
    [sym__PERLY_BRACE_OPEN] = STATE(117),
    [sym_block] = STATE(1437),
    [sym__fullstmt] = STATE(1448),
    [sym_statement_label] = STATE(1441),
    [sym__barestmt] = STATE(1441),
    [sym_package_statement] = STATE(1471),
    [sym_class_statement] = STATE(1471),
    [sym_role_statement] = STATE(1471),
    [sym_class_phaser_statement] = STATE(1471),
    [sym_use_version_statement] = STATE(1471),
    [sym_use_statement] = STATE(1471),
    [sym_subroutine_declaration_statement] = STATE(1471),
    [sym_method_declaration_statement] = STATE(1471),
    [sym_phaser_statement] = STATE(1471),
    [sym_conditional_statement] = STATE(1471),
    [sym_loop_statement] = STATE(1471),
    [sym_cstyle_for_statement] = STATE(1471),
    [sym_for_statement] = STATE(1471),
    [sym_try_statement] = STATE(1471),
    [sym_defer_statement] = STATE(1471),
    [sym_expression_statement] = STATE(4582),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym___DATA__] = STATE(5846),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1),
    [sym__conditionals] = STATE(5920),
    [sym__loops] = STATE(5932),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(5910),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(34),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5035),
    [ts_builtin_sym_end] = ACTIONS(9),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(15),
    [anon_sym_package] = ACTIONS(17),
    [anon_sym_class] = ACTIONS(19),
    [anon_sym_role] = ACTIONS(21),
    [anon_sym_BUILD] = ACTIONS(23),
    [anon_sym_ADJUST] = ACTIONS(23),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(29),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(33),
    [anon_sym_method] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(43),
    [anon_sym_defer] = ACTIONS(45),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(91),
    [anon_sym_no] = ACTIONS(91),
    [anon_sym_for] = ACTIONS(93),
    [anon_sym_foreach] = ACTIONS(93),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(97),
    [anon_sym_INIT] = ACTIONS(97),
    [anon_sym_CHECK] = ACTIONS(97),
    [anon_sym_UNITCHECK] = ACTIONS(97),
    [anon_sym_END] = ACTIONS(97),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(101),
    [anon_sym___END__] = ACTIONS(101),
    [anon_sym_EOT] = ACTIONS(103),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(145),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(2)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1881),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym__label_arg] = STATE(1877),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(2),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(1942),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(151),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(155),
    [anon_sym_SEMI] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_for] = ACTIONS(155),
    [anon_sym_foreach] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(155),
    [anon_sym___END__] = ACTIONS(155),
    [anon_sym_EOT] = ACTIONS(155),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(155),
    [anon_sym_unless] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(155),
    [anon_sym_until] = ACTIONS(155),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(161),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(161),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(3)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1946),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(3),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(163),
    [anon_sym_SEMI] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(168),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(174),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(177),
    [anon_sym_DASH_DASH] = ACTIONS(177),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_for] = ACTIONS(163),
    [anon_sym_foreach] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(163),
    [anon_sym___END__] = ACTIONS(163),
    [anon_sym_EOT] = ACTIONS(163),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(163),
    [anon_sym_unless] = ACTIONS(163),
    [anon_sym_while] = ACTIONS(163),
    [anon_sym_until] = ACTIONS(163),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(180),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(180),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(4)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1951),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(4),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(182),
    [anon_sym_SEMI] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_for] = ACTIONS(182),
    [anon_sym_foreach] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(182),
    [anon_sym___END__] = ACTIONS(182),
    [anon_sym_EOT] = ACTIONS(182),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(182),
    [anon_sym_unless] = ACTIONS(182),
    [anon_sym_while] = ACTIONS(182),
    [anon_sym_until] = ACTIONS(182),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(184),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(184),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(5)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1901),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(5),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(188),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_for] = ACTIONS(186),
    [anon_sym_foreach] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(186),
    [anon_sym___END__] = ACTIONS(186),
    [anon_sym_EOT] = ACTIONS(186),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(186),
    [anon_sym_unless] = ACTIONS(186),
    [anon_sym_while] = ACTIONS(186),
    [anon_sym_until] = ACTIONS(186),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(190),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(190),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(6)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1939),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(6),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(197),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(200),
    [anon_sym_DASH] = ACTIONS(203),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(206),
    [anon_sym_DASH_DASH] = ACTIONS(206),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_for] = ACTIONS(192),
    [anon_sym_foreach] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(192),
    [anon_sym___END__] = ACTIONS(192),
    [anon_sym_EOT] = ACTIONS(192),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(192),
    [anon_sym_unless] = ACTIONS(192),
    [anon_sym_while] = ACTIONS(192),
    [anon_sym_until] = ACTIONS(192),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(209),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(7)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2072),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym__label_arg] = STATE(1877),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(7),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(1942),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(155),
    [anon_sym___END__] = ACTIONS(155),
    [anon_sym_EOT] = ACTIONS(155),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(161),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(161),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(8)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1946),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(8),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(168),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(247),
    [anon_sym_DASH] = ACTIONS(250),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(253),
    [anon_sym_DASH_DASH] = ACTIONS(253),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(163),
    [anon_sym___END__] = ACTIONS(163),
    [anon_sym_EOT] = ACTIONS(163),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(180),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(180),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(9)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2062),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(9),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(188),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(186),
    [anon_sym___END__] = ACTIONS(186),
    [anon_sym_EOT] = ACTIONS(186),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(190),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(190),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(10)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2073),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(10),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(182),
    [anon_sym___END__] = ACTIONS(182),
    [anon_sym_EOT] = ACTIONS(182),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(184),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(184),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(11)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2012),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(11),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(197),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(256),
    [anon_sym_DASH] = ACTIONS(259),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(262),
    [anon_sym_DASH_DASH] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(192),
    [anon_sym___END__] = ACTIONS(192),
    [anon_sym_EOT] = ACTIONS(192),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(209),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(12)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2568),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(12),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(163),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(283),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(286),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(289),
    [anon_sym_DASH] = ACTIONS(292),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(297),
    [anon_sym_DASH_DASH] = ACTIONS(297),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(13)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2565),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym__label_arg] = STATE(2571),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(13),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_identifier] = STATE(2567),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(360),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(155),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(14)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2686),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(14),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(380),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(386),
    [anon_sym_DASH] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(394),
    [anon_sym_DASH_DASH] = ACTIONS(394),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(180),
  },
  [STATE(15)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3178),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(15),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(475),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(478),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(481),
    [anon_sym_DASH] = ACTIONS(484),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(489),
    [anon_sym_DASH_DASH] = ACTIONS(489),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(180),
  },
  [STATE(16)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(2831),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(16),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_and] = ACTIONS(163),
    [anon_sym_or] = ACTIONS(163),
    [anon_sym_xor] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(163),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(573),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(576),
    [anon_sym_DASH] = ACTIONS(579),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(584),
    [anon_sym_DASH_DASH] = ACTIONS(584),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(180),
  },
  [STATE(17)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3175),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym__label_arg] = STATE(3179),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(17),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_identifier] = STATE(3177),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(647),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(161),
  },
  [STATE(18)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2683),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym__label_arg] = STATE(2688),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(18),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_identifier] = STATE(2685),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(649),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(161),
  },
  [STATE(19)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2814),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym__label_arg] = STATE(2833),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(19),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_identifier] = STATE(2830),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(651),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_and] = ACTIONS(155),
    [anon_sym_or] = ACTIONS(155),
    [anon_sym_xor] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(155),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(161),
  },
  [STATE(20)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2493),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(20),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(192),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(653),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(656),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(659),
    [anon_sym_DASH] = ACTIONS(662),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(665),
    [anon_sym_DASH_DASH] = ACTIONS(665),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(21)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2569),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(21),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(182),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(22)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2352),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(22),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(668),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(186),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(23)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3156),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(23),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(670),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(190),
  },
  [STATE(24)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2956),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(24),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(192),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(672),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(675),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(678),
    [anon_sym_DASH] = ACTIONS(681),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(209),
  },
  [STATE(25)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2721),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(25),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(687),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(186),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(190),
  },
  [STATE(26)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2832),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(26),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(182),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(184),
  },
  [STATE(27)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2661),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(27),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(689),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_and] = ACTIONS(186),
    [anon_sym_or] = ACTIONS(186),
    [anon_sym_xor] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(190),
  },
  [STATE(28)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2687),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(28),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(184),
  },
  [STATE(29)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2866),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(29),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_and] = ACTIONS(182),
    [anon_sym_or] = ACTIONS(182),
    [anon_sym_xor] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(184),
  },
  [STATE(30)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2723),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(30),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(691),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(694),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(697),
    [anon_sym_DASH] = ACTIONS(700),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(703),
    [anon_sym_DASH_DASH] = ACTIONS(703),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(209),
  },
  [STATE(31)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2901),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(31),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(706),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(709),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(712),
    [anon_sym_DASH] = ACTIONS(715),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(718),
    [anon_sym_DASH_DASH] = ACTIONS(718),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(209),
  },
  [STATE(32)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3562),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym__label_arg] = STATE(3566),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(32),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_identifier] = STATE(3564),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(721),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(161),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(33)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3565),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(33),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(803),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(806),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(809),
    [anon_sym_DASH] = ACTIONS(812),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(815),
    [anon_sym_DASH_DASH] = ACTIONS(815),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(180),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(34)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(117),
    [sym_block] = STATE(1437),
    [sym__fullstmt] = STATE(1448),
    [sym_statement_label] = STATE(1441),
    [sym__barestmt] = STATE(1441),
    [sym_package_statement] = STATE(1471),
    [sym_class_statement] = STATE(1471),
    [sym_role_statement] = STATE(1471),
    [sym_class_phaser_statement] = STATE(1471),
    [sym_use_version_statement] = STATE(1471),
    [sym_use_statement] = STATE(1471),
    [sym_subroutine_declaration_statement] = STATE(1471),
    [sym_method_declaration_statement] = STATE(1471),
    [sym_phaser_statement] = STATE(1471),
    [sym_conditional_statement] = STATE(1471),
    [sym_loop_statement] = STATE(1471),
    [sym_cstyle_for_statement] = STATE(1471),
    [sym_for_statement] = STATE(1471),
    [sym_try_statement] = STATE(1471),
    [sym_defer_statement] = STATE(1471),
    [sym_expression_statement] = STATE(4582),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym___DATA__] = STATE(5843),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(34),
    [sym__conditionals] = STATE(5920),
    [sym__loops] = STATE(5932),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(5910),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(43),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5035),
    [ts_builtin_sym_end] = ACTIONS(818),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(15),
    [anon_sym_package] = ACTIONS(17),
    [anon_sym_class] = ACTIONS(19),
    [anon_sym_role] = ACTIONS(21),
    [anon_sym_BUILD] = ACTIONS(23),
    [anon_sym_ADJUST] = ACTIONS(23),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(29),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(33),
    [anon_sym_method] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(43),
    [anon_sym_defer] = ACTIONS(45),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(91),
    [anon_sym_no] = ACTIONS(91),
    [anon_sym_for] = ACTIONS(93),
    [anon_sym_foreach] = ACTIONS(93),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(97),
    [anon_sym_INIT] = ACTIONS(97),
    [anon_sym_CHECK] = ACTIONS(97),
    [anon_sym_UNITCHECK] = ACTIONS(97),
    [anon_sym_END] = ACTIONS(97),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(101),
    [anon_sym___END__] = ACTIONS(101),
    [anon_sym_EOT] = ACTIONS(103),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(145),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(35)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3461),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(35),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(838),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(841),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(844),
    [anon_sym_DASH] = ACTIONS(847),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(852),
    [anon_sym_DASH_DASH] = ACTIONS(852),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(36)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3524),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3463),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(36),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(37)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2865),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(37),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(919),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(922),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(937),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(940),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(946),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(951),
    [anon_sym_DASH_DASH] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(978),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(38)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3129),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2867),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(38),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(155),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(155),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(155),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(39)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(39),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5915),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(180),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1019),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1045),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1055),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1057),
  },
  [STATE(40)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2865),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(40),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(919),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(922),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(937),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(940),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(946),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(951),
    [anon_sym_DASH_DASH] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(978),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(41)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3258),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2867),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(41),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(155),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(155),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(155),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(42)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(42),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5978),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(146),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1089),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1091),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1093),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1095),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1097),
  },
  [STATE(43)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(117),
    [sym_block] = STATE(1437),
    [sym__fullstmt] = STATE(1448),
    [sym_statement_label] = STATE(1441),
    [sym__barestmt] = STATE(1441),
    [sym_package_statement] = STATE(1471),
    [sym_class_statement] = STATE(1471),
    [sym_role_statement] = STATE(1471),
    [sym_class_phaser_statement] = STATE(1471),
    [sym_use_version_statement] = STATE(1471),
    [sym_use_statement] = STATE(1471),
    [sym_subroutine_declaration_statement] = STATE(1471),
    [sym_method_declaration_statement] = STATE(1471),
    [sym_phaser_statement] = STATE(1471),
    [sym_conditional_statement] = STATE(1471),
    [sym_loop_statement] = STATE(1471),
    [sym_cstyle_for_statement] = STATE(1471),
    [sym_for_statement] = STATE(1471),
    [sym_try_statement] = STATE(1471),
    [sym_defer_statement] = STATE(1471),
    [sym_expression_statement] = STATE(4582),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(43),
    [sym__conditionals] = STATE(5920),
    [sym__loops] = STATE(5932),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(5910),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(43),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5035),
    [ts_builtin_sym_end] = ACTIONS(1099),
    [sym__identifier] = ACTIONS(1101),
    [anon_sym_LBRACE] = ACTIONS(1104),
    [anon_sym_SEMI] = ACTIONS(1107),
    [anon_sym_package] = ACTIONS(1110),
    [anon_sym_class] = ACTIONS(1113),
    [anon_sym_role] = ACTIONS(1116),
    [anon_sym_BUILD] = ACTIONS(1119),
    [anon_sym_ADJUST] = ACTIONS(1119),
    [anon_sym_DOLLAR] = ACTIONS(1122),
    [anon_sym_AT] = ACTIONS(1125),
    [anon_sym_my] = ACTIONS(1128),
    [anon_sym_extended] = ACTIONS(1131),
    [anon_sym_async] = ACTIONS(1131),
    [anon_sym_sub] = ACTIONS(1134),
    [anon_sym_method] = ACTIONS(1137),
    [anon_sym_LPAREN] = ACTIONS(1140),
    [anon_sym_continue] = ACTIONS(1143),
    [anon_sym_state] = ACTIONS(1146),
    [anon_sym_our] = ACTIONS(1146),
    [anon_sym_try] = ACTIONS(1149),
    [anon_sym_defer] = ACTIONS(1152),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(1158),
    [anon_sym_LT_LT] = ACTIONS(1161),
    [anon_sym_SLASH_SLASH] = ACTIONS(1164),
    [anon_sym_PLUS] = ACTIONS(1167),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1167),
    [anon_sym_BANG] = ACTIONS(1167),
    [anon_sym_PLUS_PLUS] = ACTIONS(1173),
    [anon_sym_DASH_DASH] = ACTIONS(1173),
    [anon_sym_BSLASH] = ACTIONS(1176),
    [anon_sym_do] = ACTIONS(1179),
    [anon_sym_eval] = ACTIONS(1182),
    [anon_sym_await] = ACTIONS(1185),
    [anon_sym_field] = ACTIONS(1146),
    [anon_sym_local] = ACTIONS(1188),
    [anon_sym_dynamically] = ACTIONS(1188),
    [anon_sym_DOLLAR_POUND] = ACTIONS(1191),
    [anon_sym_require] = ACTIONS(1194),
    [anon_sym_map] = ACTIONS(1197),
    [anon_sym_grep] = ACTIONS(1197),
    [anon_sym_sort] = ACTIONS(1200),
    [anon_sym_goto] = ACTIONS(1203),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_undef] = ACTIONS(1209),
    [anon_sym_PERCENT2] = ACTIONS(1212),
    [anon_sym_AMP2] = ACTIONS(1215),
    [anon_sym_STAR2] = ACTIONS(1218),
    [anon_sym_use] = ACTIONS(1221),
    [anon_sym_no] = ACTIONS(1221),
    [anon_sym_for] = ACTIONS(1224),
    [anon_sym_foreach] = ACTIONS(1224),
    [anon_sym_last] = ACTIONS(1227),
    [anon_sym_next] = ACTIONS(1227),
    [anon_sym_redo] = ACTIONS(1227),
    [anon_sym_BEGIN] = ACTIONS(1230),
    [anon_sym_INIT] = ACTIONS(1230),
    [anon_sym_CHECK] = ACTIONS(1230),
    [anon_sym_UNITCHECK] = ACTIONS(1230),
    [anon_sym_END] = ACTIONS(1230),
    [anon_sym___FILE__] = ACTIONS(1143),
    [anon_sym___LINE__] = ACTIONS(1143),
    [anon_sym___PACKAGE__] = ACTIONS(1143),
    [anon_sym___SUB__] = ACTIONS(1143),
    [anon_sym_break] = ACTIONS(1143),
    [anon_sym_fork] = ACTIONS(1143),
    [anon_sym_getppid] = ACTIONS(1143),
    [anon_sym_time] = ACTIONS(1143),
    [anon_sym_times] = ACTIONS(1143),
    [anon_sym_wait] = ACTIONS(1143),
    [anon_sym_wantarray] = ACTIONS(1143),
    [anon_sym_abs] = ACTIONS(1233),
    [anon_sym_alarm] = ACTIONS(1233),
    [anon_sym_chop] = ACTIONS(1233),
    [anon_sym_chdir] = ACTIONS(1233),
    [anon_sym_close] = ACTIONS(1233),
    [anon_sym_closedir] = ACTIONS(1233),
    [anon_sym_caller] = ACTIONS(1233),
    [anon_sym_chomp] = ACTIONS(1233),
    [anon_sym_chr] = ACTIONS(1233),
    [anon_sym_cos] = ACTIONS(1233),
    [anon_sym_chroot] = ACTIONS(1233),
    [anon_sym_defined] = ACTIONS(1233),
    [anon_sym_delete] = ACTIONS(1233),
    [anon_sym_dbmclose] = ACTIONS(1233),
    [anon_sym_exists] = ACTIONS(1233),
    [anon_sym_exit] = ACTIONS(1233),
    [anon_sym_eof] = ACTIONS(1233),
    [anon_sym_exp] = ACTIONS(1233),
    [anon_sym_each] = ACTIONS(1233),
    [anon_sym_fc] = ACTIONS(1233),
    [anon_sym_fileno] = ACTIONS(1233),
    [anon_sym_gmtime] = ACTIONS(1233),
    [anon_sym_getc] = ACTIONS(1233),
    [anon_sym_getpgrp] = ACTIONS(1233),
    [anon_sym_getprotobyname] = ACTIONS(1233),
    [anon_sym_getpwname] = ACTIONS(1233),
    [anon_sym_getpwuid] = ACTIONS(1233),
    [anon_sym_getpeername] = ACTIONS(1233),
    [anon_sym_getnetbyname] = ACTIONS(1233),
    [anon_sym_getsockname] = ACTIONS(1233),
    [anon_sym_getgrnam] = ACTIONS(1233),
    [anon_sym_getgrgid] = ACTIONS(1233),
    [anon_sym_hex] = ACTIONS(1233),
    [anon_sym_int] = ACTIONS(1233),
    [anon_sym_keys] = ACTIONS(1233),
    [anon_sym_lc] = ACTIONS(1233),
    [anon_sym_lcfirst] = ACTIONS(1233),
    [anon_sym_length] = ACTIONS(1233),
    [anon_sym_localtime] = ACTIONS(1233),
    [anon_sym_log] = ACTIONS(1233),
    [anon_sym_lock] = ACTIONS(1233),
    [anon_sym_lstat] = ACTIONS(1233),
    [anon_sym_oct] = ACTIONS(1233),
    [anon_sym_ord] = ACTIONS(1233),
    [anon_sym_prototype] = ACTIONS(1233),
    [anon_sym_pop] = ACTIONS(1233),
    [anon_sym_pos] = ACTIONS(1233),
    [anon_sym_quotemeta] = ACTIONS(1233),
    [anon_sym_reset] = ACTIONS(1233),
    [anon_sym_rand] = ACTIONS(1233),
    [anon_sym_rmdir] = ACTIONS(1233),
    [anon_sym_readdir] = ACTIONS(1233),
    [anon_sym_readline] = ACTIONS(1233),
    [anon_sym_readpipe] = ACTIONS(1233),
    [anon_sym_rewinddir] = ACTIONS(1233),
    [anon_sym_readlink] = ACTIONS(1233),
    [anon_sym_ref] = ACTIONS(1233),
    [anon_sym_scalar] = ACTIONS(1233),
    [anon_sym_shift] = ACTIONS(1233),
    [anon_sym_sin] = ACTIONS(1233),
    [anon_sym_sleep] = ACTIONS(1233),
    [anon_sym_sqrt] = ACTIONS(1233),
    [anon_sym_srand] = ACTIONS(1233),
    [anon_sym_stat] = ACTIONS(1233),
    [anon_sym_study] = ACTIONS(1233),
    [anon_sym_tell] = ACTIONS(1233),
    [anon_sym_telldir] = ACTIONS(1233),
    [anon_sym_tied] = ACTIONS(1233),
    [anon_sym_uc] = ACTIONS(1233),
    [anon_sym_ucfirst] = ACTIONS(1233),
    [anon_sym_untie] = ACTIONS(1233),
    [anon_sym_umask] = ACTIONS(1233),
    [anon_sym_values] = ACTIONS(1233),
    [anon_sym_write] = ACTIONS(1233),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(1236),
    [anon_sym___END__] = ACTIONS(1236),
    [anon_sym_EOT] = ACTIONS(1236),
    [anon_sym_q] = ACTIONS(1238),
    [anon_sym_qq] = ACTIONS(1241),
    [anon_sym_qw] = ACTIONS(1244),
    [anon_sym_qx] = ACTIONS(1247),
    [anon_sym_qr] = ACTIONS(1250),
    [anon_sym_m] = ACTIONS(1253),
    [anon_sym_s] = ACTIONS(1256),
    [anon_sym_tr] = ACTIONS(1259),
    [anon_sym_y] = ACTIONS(1259),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_unless] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1265),
    [anon_sym_until] = ACTIONS(1265),
    [aux_sym__bareword_token1] = ACTIONS(1268),
    [sym_number] = ACTIONS(1271),
    [anon_sym_true] = ACTIONS(1274),
    [anon_sym_false] = ACTIONS(1274),
    [sym__single_quote] = ACTIONS(1277),
    [sym__double_quote] = ACTIONS(1280),
    [sym__backtick_quote] = ACTIONS(1283),
    [sym__search_slash_quote] = ACTIONS(1286),
    [sym__open_readline_bracket] = ACTIONS(1289),
    [sym__open_fileglob_bracket] = ACTIONS(1292),
    [sym__PERLY_HEREDOC] = ACTIONS(1295),
    [sym__ctrl_z_hack] = ACTIONS(1099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1298),
    [sym__filetest] = ACTIONS(1301),
  },
  [STATE(44)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(44),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5793),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(138),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1308),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1310),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1312),
  },
  [STATE(45)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(45),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5833),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(141),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1318),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1320),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1322),
  },
  [STATE(46)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(46),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5761),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(145),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1326),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1328),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1330),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1332),
  },
  [STATE(47)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(47),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5972),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(148),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1334),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1338),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1340),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1342),
  },
  [STATE(48)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(48),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5737),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(154),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1348),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1350),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1352),
  },
  [STATE(49)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(49),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5780),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(135),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1354),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1358),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1360),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1362),
  },
  [STATE(50)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(50),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5823),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(204),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1368),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1370),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1372),
  },
  [STATE(51)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(51),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5844),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(167),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1378),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1380),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1382),
  },
  [STATE(52)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(52),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5854),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(171),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1384),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1388),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1390),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1392),
  },
  [STATE(53)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(53),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5859),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(172),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1398),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1400),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1402),
  },
  [STATE(54)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(54),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5864),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(151),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1406),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1408),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1410),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1412),
  },
  [STATE(55)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(55),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5868),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(173),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1416),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1418),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1420),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1422),
  },
  [STATE(56)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(56),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5871),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(156),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1424),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1426),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1428),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1430),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1432),
  },
  [STATE(57)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(57),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5873),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(175),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1436),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1438),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1440),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1442),
  },
  [STATE(58)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(58),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5876),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(161),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1444),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1446),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1448),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1450),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1452),
  },
  [STATE(59)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(59),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5879),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(176),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1454),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1456),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1458),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1460),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1462),
  },
  [STATE(60)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(60),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5882),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(177),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1466),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1468),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1470),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1472),
  },
  [STATE(61)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(61),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5885),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(179),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1474),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1476),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1478),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1480),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1482),
  },
  [STATE(62)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(62),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5888),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(181),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1486),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1488),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1490),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1492),
  },
  [STATE(63)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(63),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5891),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(183),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1494),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1496),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1498),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1500),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1502),
  },
  [STATE(64)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(64),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5894),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(185),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1504),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1506),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1508),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1510),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1512),
  },
  [STATE(65)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(65),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5897),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(186),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1514),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1516),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1518),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1520),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1522),
  },
  [STATE(66)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(66),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5900),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(187),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1524),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1526),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1528),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1530),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1532),
  },
  [STATE(67)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(67),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5903),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(178),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1534),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1536),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1538),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1540),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1542),
  },
  [STATE(68)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(68),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5906),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(188),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1544),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1548),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1550),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1552),
  },
  [STATE(69)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(69),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5909),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(190),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1556),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1558),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1560),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1562),
  },
  [STATE(70)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(70),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5912),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(191),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1564),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1568),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1570),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1572),
  },
  [STATE(71)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(71),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5918),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(193),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1576),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1578),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1580),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1582),
  },
  [STATE(72)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(72),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5921),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(182),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1584),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1586),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1588),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1590),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1592),
  },
  [STATE(73)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(73),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5924),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(194),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1596),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1598),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1600),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1602),
  },
  [STATE(74)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(74),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5927),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(184),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1608),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1610),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1612),
  },
  [STATE(75)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(75),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5930),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(195),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1618),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1620),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1622),
  },
  [STATE(76)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(76),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5933),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(197),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1624),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1628),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1630),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1632),
  },
  [STATE(77)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(77),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5936),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(198),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1638),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1640),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1642),
  },
  [STATE(78)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3461),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(78),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(838),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(841),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(1646),
    [anon_sym_DASH] = ACTIONS(1649),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(1654),
    [anon_sym_DASH_DASH] = ACTIONS(1654),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(163),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(79)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3720),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3463),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(79),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(155),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(80)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(80),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym__ident_special] = STATE(5849),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(169),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1685),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1687),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1031),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(1043),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(1689),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__ident_special_token1] = ACTIONS(1053),
    [aux_sym__bareword_token1] = ACTIONS(1691),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__brace_autoquoted_token] = ACTIONS(1693),
  },
  [STATE(81)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3030),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(81),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(1695),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(1701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(1704),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(1707),
    [anon_sym_DASH] = ACTIONS(1710),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(1713),
    [anon_sym_DASH_DASH] = ACTIONS(1713),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(1716),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(82)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3132),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(82),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(182),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(182),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(182),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(83)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3516),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(83),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(84)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3543),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(84),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(1721),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(190),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(85)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3587),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(85),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(1723),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(1726),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(1729),
    [anon_sym_DASH] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(1735),
    [anon_sym_DASH_DASH] = ACTIONS(1735),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(86)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3461),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(86),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(163),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(838),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(841),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(844),
    [anon_sym_DASH] = ACTIONS(847),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(852),
    [anon_sym_DASH_DASH] = ACTIONS(852),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(87)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3740),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3463),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(87),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(88)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2865),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(88),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(919),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(922),
    [anon_sym_RPAREN] = ACTIONS(163),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(937),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(940),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(946),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(951),
    [anon_sym_DASH_DASH] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(978),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(89)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3414),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2867),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(89),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(155),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(155),
    [anon_sym_RPAREN] = ACTIONS(155),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(155),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(90)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3688),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(90),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(184),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(91)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3527),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(91),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(92)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3672),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(92),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(1798),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(1801),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(1804),
    [anon_sym_DASH] = ACTIONS(1807),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(1810),
    [anon_sym_DASH_DASH] = ACTIONS(1810),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(93)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3121),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(93),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(186),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(186),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(1813),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(186),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(94)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3263),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(94),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(182),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(182),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(182),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(95)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3265),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(95),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(186),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(186),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(1813),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(186),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(96)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3461),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(96),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(163),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(163),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(163),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(163),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(163),
    [anon_sym_DASH_GT] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(163),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(163),
    [anon_sym_GT] = ACTIONS(163),
    [anon_sym_LT_LT] = ACTIONS(838),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(163),
    [anon_sym_PLUS_EQ] = ACTIONS(163),
    [anon_sym_DASH_EQ] = ACTIONS(163),
    [anon_sym_DOT_EQ] = ACTIONS(163),
    [anon_sym_STAR_EQ] = ACTIONS(163),
    [anon_sym_SLASH_EQ] = ACTIONS(163),
    [anon_sym_PERCENT_EQ] = ACTIONS(163),
    [anon_sym_x_EQ] = ACTIONS(163),
    [anon_sym_AMP_EQ] = ACTIONS(163),
    [anon_sym_PIPE_EQ] = ACTIONS(163),
    [anon_sym_CARET_EQ] = ACTIONS(163),
    [anon_sym_LT_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_GT_EQ] = ACTIONS(163),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(163),
    [anon_sym_DOT_DOT] = ACTIONS(163),
    [anon_sym_STAR_STAR] = ACTIONS(163),
    [anon_sym_PIPE_PIPE] = ACTIONS(163),
    [anon_sym_SLASH_SLASH] = ACTIONS(841),
    [anon_sym_CARET_CARET] = ACTIONS(163),
    [anon_sym_AMP_AMP] = ACTIONS(163),
    [anon_sym_PIPE] = ACTIONS(163),
    [anon_sym_CARET] = ACTIONS(163),
    [anon_sym_AMP] = ACTIONS(163),
    [anon_sym_GT_GT2] = ACTIONS(163),
    [anon_sym_PLUS] = ACTIONS(1646),
    [anon_sym_DASH] = ACTIONS(1649),
    [anon_sym_DOT] = ACTIONS(163),
    [anon_sym_SLASH] = ACTIONS(163),
    [anon_sym_x] = ACTIONS(163),
    [anon_sym_EQ_TILDE] = ACTIONS(163),
    [anon_sym_BANG_TILDE] = ACTIONS(163),
    [anon_sym_EQ_EQ] = ACTIONS(163),
    [anon_sym_BANG_EQ] = ACTIONS(163),
    [anon_sym_eq] = ACTIONS(163),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(163),
    [anon_sym_equ] = ACTIONS(163),
    [anon_sym_eqr] = ACTIONS(163),
    [anon_sym_ne] = ACTIONS(163),
    [anon_sym_LT_EQ_GT] = ACTIONS(163),
    [anon_sym_cmp] = ACTIONS(163),
    [anon_sym_TILDE_TILDE] = ACTIONS(163),
    [anon_sym_LT] = ACTIONS(163),
    [anon_sym_LT_EQ] = ACTIONS(163),
    [anon_sym_GT_EQ] = ACTIONS(163),
    [anon_sym_lt] = ACTIONS(163),
    [anon_sym_le] = ACTIONS(163),
    [anon_sym_ge] = ACTIONS(163),
    [anon_sym_gt] = ACTIONS(163),
    [anon_sym_isa] = ACTIONS(163),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(1654),
    [anon_sym_DASH_DASH] = ACTIONS(1654),
    [anon_sym_QMARK] = ACTIONS(163),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(97)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3712),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(97),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(186),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(98)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3772),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3463),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(98),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(155),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(155),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(155),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(155),
    [anon_sym_DASH_GT] = ACTIONS(155),
    [anon_sym_STAR] = ACTIONS(155),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(155),
    [anon_sym_GT] = ACTIONS(155),
    [anon_sym_LT_LT] = ACTIONS(155),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(155),
    [anon_sym_PLUS_EQ] = ACTIONS(155),
    [anon_sym_DASH_EQ] = ACTIONS(155),
    [anon_sym_DOT_EQ] = ACTIONS(155),
    [anon_sym_STAR_EQ] = ACTIONS(155),
    [anon_sym_SLASH_EQ] = ACTIONS(155),
    [anon_sym_PERCENT_EQ] = ACTIONS(155),
    [anon_sym_x_EQ] = ACTIONS(155),
    [anon_sym_AMP_EQ] = ACTIONS(155),
    [anon_sym_PIPE_EQ] = ACTIONS(155),
    [anon_sym_CARET_EQ] = ACTIONS(155),
    [anon_sym_LT_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_GT_EQ] = ACTIONS(155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(155),
    [anon_sym_DOT_DOT] = ACTIONS(155),
    [anon_sym_STAR_STAR] = ACTIONS(155),
    [anon_sym_PIPE_PIPE] = ACTIONS(155),
    [anon_sym_SLASH_SLASH] = ACTIONS(155),
    [anon_sym_CARET_CARET] = ACTIONS(155),
    [anon_sym_AMP_AMP] = ACTIONS(155),
    [anon_sym_PIPE] = ACTIONS(155),
    [anon_sym_CARET] = ACTIONS(155),
    [anon_sym_AMP] = ACTIONS(155),
    [anon_sym_GT_GT2] = ACTIONS(155),
    [anon_sym_PLUS] = ACTIONS(155),
    [anon_sym_DASH] = ACTIONS(155),
    [anon_sym_DOT] = ACTIONS(155),
    [anon_sym_SLASH] = ACTIONS(155),
    [anon_sym_x] = ACTIONS(155),
    [anon_sym_EQ_TILDE] = ACTIONS(155),
    [anon_sym_BANG_TILDE] = ACTIONS(155),
    [anon_sym_EQ_EQ] = ACTIONS(155),
    [anon_sym_BANG_EQ] = ACTIONS(155),
    [anon_sym_eq] = ACTIONS(155),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(155),
    [anon_sym_equ] = ACTIONS(155),
    [anon_sym_eqr] = ACTIONS(155),
    [anon_sym_ne] = ACTIONS(155),
    [anon_sym_LT_EQ_GT] = ACTIONS(155),
    [anon_sym_cmp] = ACTIONS(155),
    [anon_sym_TILDE_TILDE] = ACTIONS(155),
    [anon_sym_LT] = ACTIONS(155),
    [anon_sym_LT_EQ] = ACTIONS(155),
    [anon_sym_GT_EQ] = ACTIONS(155),
    [anon_sym_lt] = ACTIONS(155),
    [anon_sym_le] = ACTIONS(155),
    [anon_sym_ge] = ACTIONS(155),
    [anon_sym_gt] = ACTIONS(155),
    [anon_sym_isa] = ACTIONS(155),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(155),
    [anon_sym_DASH_DASH] = ACTIONS(155),
    [anon_sym_QMARK] = ACTIONS(155),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(99)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3721),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(99),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(182),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(100)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3697),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(100),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(192),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(192),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(192),
    [anon_sym_DASH_GT] = ACTIONS(192),
    [anon_sym_STAR] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(192),
    [anon_sym_GT] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(1798),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(192),
    [anon_sym_PLUS_EQ] = ACTIONS(192),
    [anon_sym_DASH_EQ] = ACTIONS(192),
    [anon_sym_DOT_EQ] = ACTIONS(192),
    [anon_sym_STAR_EQ] = ACTIONS(192),
    [anon_sym_SLASH_EQ] = ACTIONS(192),
    [anon_sym_PERCENT_EQ] = ACTIONS(192),
    [anon_sym_x_EQ] = ACTIONS(192),
    [anon_sym_AMP_EQ] = ACTIONS(192),
    [anon_sym_PIPE_EQ] = ACTIONS(192),
    [anon_sym_CARET_EQ] = ACTIONS(192),
    [anon_sym_LT_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_GT_EQ] = ACTIONS(192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(192),
    [anon_sym_DOT_DOT] = ACTIONS(192),
    [anon_sym_STAR_STAR] = ACTIONS(192),
    [anon_sym_PIPE_PIPE] = ACTIONS(192),
    [anon_sym_SLASH_SLASH] = ACTIONS(1801),
    [anon_sym_CARET_CARET] = ACTIONS(192),
    [anon_sym_AMP_AMP] = ACTIONS(192),
    [anon_sym_PIPE] = ACTIONS(192),
    [anon_sym_CARET] = ACTIONS(192),
    [anon_sym_AMP] = ACTIONS(192),
    [anon_sym_GT_GT2] = ACTIONS(192),
    [anon_sym_PLUS] = ACTIONS(1845),
    [anon_sym_DASH] = ACTIONS(1848),
    [anon_sym_DOT] = ACTIONS(192),
    [anon_sym_SLASH] = ACTIONS(192),
    [anon_sym_x] = ACTIONS(192),
    [anon_sym_EQ_TILDE] = ACTIONS(192),
    [anon_sym_BANG_TILDE] = ACTIONS(192),
    [anon_sym_EQ_EQ] = ACTIONS(192),
    [anon_sym_BANG_EQ] = ACTIONS(192),
    [anon_sym_eq] = ACTIONS(192),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(192),
    [anon_sym_equ] = ACTIONS(192),
    [anon_sym_eqr] = ACTIONS(192),
    [anon_sym_ne] = ACTIONS(192),
    [anon_sym_LT_EQ_GT] = ACTIONS(192),
    [anon_sym_cmp] = ACTIONS(192),
    [anon_sym_TILDE_TILDE] = ACTIONS(192),
    [anon_sym_LT] = ACTIONS(192),
    [anon_sym_LT_EQ] = ACTIONS(192),
    [anon_sym_GT_EQ] = ACTIONS(192),
    [anon_sym_lt] = ACTIONS(192),
    [anon_sym_le] = ACTIONS(192),
    [anon_sym_ge] = ACTIONS(192),
    [anon_sym_gt] = ACTIONS(192),
    [anon_sym_isa] = ACTIONS(192),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(1851),
    [anon_sym_DASH_DASH] = ACTIONS(1851),
    [anon_sym_QMARK] = ACTIONS(192),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(101)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4011),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(101),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(204),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(102)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(102),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(146),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(103)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3997),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5808),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(103),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(151),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(104)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4000),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5839),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(104),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(105)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4005),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5797),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(105),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(156),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(106)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3722),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(106),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(107)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3416),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(107),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(182),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(182),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(108)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(108),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(109)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3999),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5839),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(109),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(161),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(110)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(110),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(165),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(111)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4009),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5747),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(111),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(178),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(112)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4007),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5808),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(112),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(148),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(113)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4011),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(113),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(180),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1872),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(114)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4007),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5808),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(114),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(115)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4004),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5863),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(115),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(182),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(116)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4003),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5797),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(116),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(154),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(117)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(117),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(143),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(118)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4003),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5797),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(118),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(119)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4013),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5818),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(119),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(161),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(120)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4000),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5839),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(120),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(135),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(121)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4001),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5747),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(121),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(177),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1884),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(122)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4001),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5747),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(122),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(123)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(123),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(124)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3995),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5750),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(124),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1888),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(125)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(125),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1872),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(126)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4002),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5863),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(126),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(181),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(127)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4002),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5863),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(127),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(128)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3995),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5750),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(128),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(183),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(129)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3998),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5818),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(129),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(197),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(130)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3998),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5818),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(130),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(131)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4008),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(131),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(165),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(132)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3732),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(132),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(133)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3406),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(133),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(186),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(186),
    [anon_sym_RPAREN] = ACTIONS(186),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(1813),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(186),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(134)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4012),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym__tricky_list] = STATE(5750),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1968),
    [sym_interpolated_string_literal] = STATE(1968),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1968),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(134),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1966),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(184),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1888),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(135)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(135),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1896),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(136)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(136),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(137)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(137),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(165),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(138)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(138),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(139)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(139),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(146),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1091),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(140)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(140),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(164),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(141)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(141),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(142)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(142),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(1101),
    [anon_sym_LBRACE] = ACTIONS(1104),
    [anon_sym_RBRACE] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_package] = ACTIONS(1911),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_role] = ACTIONS(1917),
    [anon_sym_BUILD] = ACTIONS(1920),
    [anon_sym_ADJUST] = ACTIONS(1920),
    [anon_sym_DOLLAR] = ACTIONS(1122),
    [anon_sym_AT] = ACTIONS(1125),
    [anon_sym_my] = ACTIONS(1923),
    [anon_sym_extended] = ACTIONS(1131),
    [anon_sym_async] = ACTIONS(1131),
    [anon_sym_sub] = ACTIONS(1926),
    [anon_sym_method] = ACTIONS(1929),
    [anon_sym_LPAREN] = ACTIONS(1140),
    [anon_sym_continue] = ACTIONS(1143),
    [anon_sym_state] = ACTIONS(1146),
    [anon_sym_our] = ACTIONS(1146),
    [anon_sym_try] = ACTIONS(1932),
    [anon_sym_defer] = ACTIONS(1935),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(1158),
    [anon_sym_LT_LT] = ACTIONS(1161),
    [anon_sym_SLASH_SLASH] = ACTIONS(1164),
    [anon_sym_PLUS] = ACTIONS(1167),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1167),
    [anon_sym_BANG] = ACTIONS(1167),
    [anon_sym_PLUS_PLUS] = ACTIONS(1173),
    [anon_sym_DASH_DASH] = ACTIONS(1173),
    [anon_sym_BSLASH] = ACTIONS(1176),
    [anon_sym_do] = ACTIONS(1179),
    [anon_sym_eval] = ACTIONS(1182),
    [anon_sym_await] = ACTIONS(1185),
    [anon_sym_field] = ACTIONS(1146),
    [anon_sym_local] = ACTIONS(1188),
    [anon_sym_dynamically] = ACTIONS(1188),
    [anon_sym_DOLLAR_POUND] = ACTIONS(1191),
    [anon_sym_require] = ACTIONS(1194),
    [anon_sym_map] = ACTIONS(1197),
    [anon_sym_grep] = ACTIONS(1197),
    [anon_sym_sort] = ACTIONS(1200),
    [anon_sym_goto] = ACTIONS(1203),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_undef] = ACTIONS(1209),
    [anon_sym_PERCENT2] = ACTIONS(1212),
    [anon_sym_AMP2] = ACTIONS(1215),
    [anon_sym_STAR2] = ACTIONS(1218),
    [anon_sym_use] = ACTIONS(1938),
    [anon_sym_no] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1941),
    [anon_sym_foreach] = ACTIONS(1941),
    [anon_sym_last] = ACTIONS(1227),
    [anon_sym_next] = ACTIONS(1227),
    [anon_sym_redo] = ACTIONS(1227),
    [anon_sym_BEGIN] = ACTIONS(1944),
    [anon_sym_INIT] = ACTIONS(1944),
    [anon_sym_CHECK] = ACTIONS(1944),
    [anon_sym_UNITCHECK] = ACTIONS(1944),
    [anon_sym_END] = ACTIONS(1944),
    [anon_sym___FILE__] = ACTIONS(1143),
    [anon_sym___LINE__] = ACTIONS(1143),
    [anon_sym___PACKAGE__] = ACTIONS(1143),
    [anon_sym___SUB__] = ACTIONS(1143),
    [anon_sym_break] = ACTIONS(1143),
    [anon_sym_fork] = ACTIONS(1143),
    [anon_sym_getppid] = ACTIONS(1143),
    [anon_sym_time] = ACTIONS(1143),
    [anon_sym_times] = ACTIONS(1143),
    [anon_sym_wait] = ACTIONS(1143),
    [anon_sym_wantarray] = ACTIONS(1143),
    [anon_sym_abs] = ACTIONS(1233),
    [anon_sym_alarm] = ACTIONS(1233),
    [anon_sym_chop] = ACTIONS(1233),
    [anon_sym_chdir] = ACTIONS(1233),
    [anon_sym_close] = ACTIONS(1233),
    [anon_sym_closedir] = ACTIONS(1233),
    [anon_sym_caller] = ACTIONS(1233),
    [anon_sym_chomp] = ACTIONS(1233),
    [anon_sym_chr] = ACTIONS(1233),
    [anon_sym_cos] = ACTIONS(1233),
    [anon_sym_chroot] = ACTIONS(1233),
    [anon_sym_defined] = ACTIONS(1233),
    [anon_sym_delete] = ACTIONS(1233),
    [anon_sym_dbmclose] = ACTIONS(1233),
    [anon_sym_exists] = ACTIONS(1233),
    [anon_sym_exit] = ACTIONS(1233),
    [anon_sym_eof] = ACTIONS(1233),
    [anon_sym_exp] = ACTIONS(1233),
    [anon_sym_each] = ACTIONS(1233),
    [anon_sym_fc] = ACTIONS(1233),
    [anon_sym_fileno] = ACTIONS(1233),
    [anon_sym_gmtime] = ACTIONS(1233),
    [anon_sym_getc] = ACTIONS(1233),
    [anon_sym_getpgrp] = ACTIONS(1233),
    [anon_sym_getprotobyname] = ACTIONS(1233),
    [anon_sym_getpwname] = ACTIONS(1233),
    [anon_sym_getpwuid] = ACTIONS(1233),
    [anon_sym_getpeername] = ACTIONS(1233),
    [anon_sym_getnetbyname] = ACTIONS(1233),
    [anon_sym_getsockname] = ACTIONS(1233),
    [anon_sym_getgrnam] = ACTIONS(1233),
    [anon_sym_getgrgid] = ACTIONS(1233),
    [anon_sym_hex] = ACTIONS(1233),
    [anon_sym_int] = ACTIONS(1233),
    [anon_sym_keys] = ACTIONS(1233),
    [anon_sym_lc] = ACTIONS(1233),
    [anon_sym_lcfirst] = ACTIONS(1233),
    [anon_sym_length] = ACTIONS(1233),
    [anon_sym_localtime] = ACTIONS(1233),
    [anon_sym_log] = ACTIONS(1233),
    [anon_sym_lock] = ACTIONS(1233),
    [anon_sym_lstat] = ACTIONS(1233),
    [anon_sym_oct] = ACTIONS(1233),
    [anon_sym_ord] = ACTIONS(1233),
    [anon_sym_prototype] = ACTIONS(1233),
    [anon_sym_pop] = ACTIONS(1233),
    [anon_sym_pos] = ACTIONS(1233),
    [anon_sym_quotemeta] = ACTIONS(1233),
    [anon_sym_reset] = ACTIONS(1233),
    [anon_sym_rand] = ACTIONS(1233),
    [anon_sym_rmdir] = ACTIONS(1233),
    [anon_sym_readdir] = ACTIONS(1233),
    [anon_sym_readline] = ACTIONS(1233),
    [anon_sym_readpipe] = ACTIONS(1233),
    [anon_sym_rewinddir] = ACTIONS(1233),
    [anon_sym_readlink] = ACTIONS(1233),
    [anon_sym_ref] = ACTIONS(1233),
    [anon_sym_scalar] = ACTIONS(1233),
    [anon_sym_shift] = ACTIONS(1233),
    [anon_sym_sin] = ACTIONS(1233),
    [anon_sym_sleep] = ACTIONS(1233),
    [anon_sym_sqrt] = ACTIONS(1233),
    [anon_sym_srand] = ACTIONS(1233),
    [anon_sym_stat] = ACTIONS(1233),
    [anon_sym_study] = ACTIONS(1233),
    [anon_sym_tell] = ACTIONS(1233),
    [anon_sym_telldir] = ACTIONS(1233),
    [anon_sym_tied] = ACTIONS(1233),
    [anon_sym_uc] = ACTIONS(1233),
    [anon_sym_ucfirst] = ACTIONS(1233),
    [anon_sym_untie] = ACTIONS(1233),
    [anon_sym_umask] = ACTIONS(1233),
    [anon_sym_values] = ACTIONS(1233),
    [anon_sym_write] = ACTIONS(1233),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(1238),
    [anon_sym_qq] = ACTIONS(1241),
    [anon_sym_qw] = ACTIONS(1244),
    [anon_sym_qx] = ACTIONS(1247),
    [anon_sym_qr] = ACTIONS(1250),
    [anon_sym_m] = ACTIONS(1253),
    [anon_sym_s] = ACTIONS(1256),
    [anon_sym_tr] = ACTIONS(1259),
    [anon_sym_y] = ACTIONS(1259),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_unless] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1265),
    [anon_sym_until] = ACTIONS(1265),
    [aux_sym__bareword_token1] = ACTIONS(1268),
    [sym_number] = ACTIONS(1271),
    [anon_sym_true] = ACTIONS(1274),
    [anon_sym_false] = ACTIONS(1274),
    [sym__single_quote] = ACTIONS(1277),
    [sym__double_quote] = ACTIONS(1280),
    [sym__backtick_quote] = ACTIONS(1283),
    [sym__search_slash_quote] = ACTIONS(1286),
    [sym__open_readline_bracket] = ACTIONS(1289),
    [sym__open_fileglob_bracket] = ACTIONS(1292),
    [sym__PERLY_HEREDOC] = ACTIONS(1295),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1298),
    [sym__filetest] = ACTIONS(1301),
  },
  [STATE(143)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(143),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(144)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(144),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(174),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1949),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(145)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(145),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(146)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(146),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1953),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(147)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(147),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(177),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1466),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(148)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(148),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(149)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(149),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(148),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(150)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(150),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(181),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1486),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(151)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(151),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1957),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(152)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(152),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(153)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(153),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(183),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1496),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(154)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(154),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1961),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(155)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(155),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(152),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(156)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(156),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1965),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(157)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(157),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(189),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(158)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(158),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(143),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1969),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(159)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3800),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(159),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(182),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(182),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(182),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(182),
    [anon_sym_DASH_GT] = ACTIONS(182),
    [anon_sym_STAR] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(182),
    [anon_sym_GT] = ACTIONS(182),
    [anon_sym_LT_LT] = ACTIONS(182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_DOT_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_x_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(182),
    [anon_sym_DOT_DOT] = ACTIONS(182),
    [anon_sym_STAR_STAR] = ACTIONS(182),
    [anon_sym_PIPE_PIPE] = ACTIONS(182),
    [anon_sym_SLASH_SLASH] = ACTIONS(182),
    [anon_sym_CARET_CARET] = ACTIONS(182),
    [anon_sym_AMP_AMP] = ACTIONS(182),
    [anon_sym_PIPE] = ACTIONS(182),
    [anon_sym_CARET] = ACTIONS(182),
    [anon_sym_AMP] = ACTIONS(182),
    [anon_sym_GT_GT2] = ACTIONS(182),
    [anon_sym_PLUS] = ACTIONS(182),
    [anon_sym_DASH] = ACTIONS(182),
    [anon_sym_DOT] = ACTIONS(182),
    [anon_sym_SLASH] = ACTIONS(182),
    [anon_sym_x] = ACTIONS(182),
    [anon_sym_EQ_TILDE] = ACTIONS(182),
    [anon_sym_BANG_TILDE] = ACTIONS(182),
    [anon_sym_EQ_EQ] = ACTIONS(182),
    [anon_sym_BANG_EQ] = ACTIONS(182),
    [anon_sym_eq] = ACTIONS(182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(182),
    [anon_sym_equ] = ACTIONS(182),
    [anon_sym_eqr] = ACTIONS(182),
    [anon_sym_ne] = ACTIONS(182),
    [anon_sym_LT_EQ_GT] = ACTIONS(182),
    [anon_sym_cmp] = ACTIONS(182),
    [anon_sym_TILDE_TILDE] = ACTIONS(182),
    [anon_sym_LT] = ACTIONS(182),
    [anon_sym_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_EQ] = ACTIONS(182),
    [anon_sym_lt] = ACTIONS(182),
    [anon_sym_le] = ACTIONS(182),
    [anon_sym_ge] = ACTIONS(182),
    [anon_sym_gt] = ACTIONS(182),
    [anon_sym_isa] = ACTIONS(182),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(182),
    [anon_sym_DASH_DASH] = ACTIONS(182),
    [anon_sym_QMARK] = ACTIONS(182),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(160)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(160),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(192),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(161)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(161),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1973),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(162)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(162),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(154),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(163)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(163),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(196),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(164)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(164),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1977),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(165)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(165),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(166)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(166),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(197),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(167)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(167),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(168)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(168),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(199),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(169)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(169),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(170)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(170),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(200),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(171)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(171),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(172)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(172),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(173)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(173),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(174)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(174),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(175)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(175),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1997),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(176)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(176),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(177)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(177),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2001),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(178)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(178),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(179)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(179),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2005),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(180)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(180),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(181)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(181),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2009),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(182)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(182),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(183)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(183),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(184)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(184),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(185)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(185),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2017),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(186)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(186),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(187)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(187),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(188)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(188),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(189)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(189),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(190)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(190),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(191)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(191),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(192)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(192),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(193)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(193),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2033),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(194)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(194),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(195)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(195),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2037),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(196)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(196),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(197)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(197),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(198)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(198),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(199)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(199),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2045),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(200)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(200),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(201)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(201),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(202)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(202),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(135),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(203)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3775),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(203),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(186),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(186),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(186),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(186),
    [anon_sym_DASH_GT] = ACTIONS(186),
    [anon_sym_STAR] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_PERCENT] = ACTIONS(186),
    [anon_sym_GT] = ACTIONS(186),
    [anon_sym_LT_LT] = ACTIONS(186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(186),
    [anon_sym_PLUS_EQ] = ACTIONS(186),
    [anon_sym_DASH_EQ] = ACTIONS(186),
    [anon_sym_DOT_EQ] = ACTIONS(186),
    [anon_sym_STAR_EQ] = ACTIONS(186),
    [anon_sym_SLASH_EQ] = ACTIONS(186),
    [anon_sym_PERCENT_EQ] = ACTIONS(186),
    [anon_sym_x_EQ] = ACTIONS(186),
    [anon_sym_AMP_EQ] = ACTIONS(186),
    [anon_sym_PIPE_EQ] = ACTIONS(186),
    [anon_sym_CARET_EQ] = ACTIONS(186),
    [anon_sym_LT_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_GT_EQ] = ACTIONS(186),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(186),
    [anon_sym_DOT_DOT] = ACTIONS(186),
    [anon_sym_STAR_STAR] = ACTIONS(186),
    [anon_sym_PIPE_PIPE] = ACTIONS(186),
    [anon_sym_SLASH_SLASH] = ACTIONS(186),
    [anon_sym_CARET_CARET] = ACTIONS(186),
    [anon_sym_AMP_AMP] = ACTIONS(186),
    [anon_sym_PIPE] = ACTIONS(186),
    [anon_sym_CARET] = ACTIONS(186),
    [anon_sym_AMP] = ACTIONS(186),
    [anon_sym_GT_GT2] = ACTIONS(186),
    [anon_sym_PLUS] = ACTIONS(186),
    [anon_sym_DASH] = ACTIONS(186),
    [anon_sym_DOT] = ACTIONS(186),
    [anon_sym_SLASH] = ACTIONS(186),
    [anon_sym_x] = ACTIONS(186),
    [anon_sym_EQ_TILDE] = ACTIONS(186),
    [anon_sym_BANG_TILDE] = ACTIONS(186),
    [anon_sym_EQ_EQ] = ACTIONS(186),
    [anon_sym_BANG_EQ] = ACTIONS(186),
    [anon_sym_eq] = ACTIONS(186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(186),
    [anon_sym_equ] = ACTIONS(186),
    [anon_sym_eqr] = ACTIONS(186),
    [anon_sym_ne] = ACTIONS(186),
    [anon_sym_LT_EQ_GT] = ACTIONS(186),
    [anon_sym_cmp] = ACTIONS(186),
    [anon_sym_TILDE_TILDE] = ACTIONS(186),
    [anon_sym_LT] = ACTIONS(186),
    [anon_sym_LT_EQ] = ACTIONS(186),
    [anon_sym_GT_EQ] = ACTIONS(186),
    [anon_sym_lt] = ACTIONS(186),
    [anon_sym_le] = ACTIONS(186),
    [anon_sym_ge] = ACTIONS(186),
    [anon_sym_gt] = ACTIONS(186),
    [anon_sym_isa] = ACTIONS(186),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(186),
    [anon_sym_DASH_DASH] = ACTIONS(186),
    [anon_sym_QMARK] = ACTIONS(186),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(204)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(204),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(142),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(205)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(205),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(136),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(206)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1620),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(206),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_source_file_repeat1] = STATE(201),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(207)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(117),
    [sym_block] = STATE(1437),
    [sym__fullstmt] = STATE(1447),
    [sym_statement_label] = STATE(1441),
    [sym__barestmt] = STATE(1441),
    [sym_package_statement] = STATE(1471),
    [sym_class_statement] = STATE(1471),
    [sym_role_statement] = STATE(1471),
    [sym_class_phaser_statement] = STATE(1471),
    [sym_use_version_statement] = STATE(1471),
    [sym_use_statement] = STATE(1471),
    [sym_subroutine_declaration_statement] = STATE(1471),
    [sym_method_declaration_statement] = STATE(1471),
    [sym_phaser_statement] = STATE(1471),
    [sym_conditional_statement] = STATE(1471),
    [sym_loop_statement] = STATE(1471),
    [sym_cstyle_for_statement] = STATE(1471),
    [sym_for_statement] = STATE(1471),
    [sym_try_statement] = STATE(1471),
    [sym_defer_statement] = STATE(1471),
    [sym_expression_statement] = STATE(4582),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(207),
    [sym__conditionals] = STATE(5920),
    [sym__loops] = STATE(5932),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(5910),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5035),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(15),
    [anon_sym_package] = ACTIONS(17),
    [anon_sym_class] = ACTIONS(19),
    [anon_sym_role] = ACTIONS(21),
    [anon_sym_BUILD] = ACTIONS(23),
    [anon_sym_ADJUST] = ACTIONS(23),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(29),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(33),
    [anon_sym_method] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(43),
    [anon_sym_defer] = ACTIONS(45),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(91),
    [anon_sym_no] = ACTIONS(91),
    [anon_sym_for] = ACTIONS(93),
    [anon_sym_foreach] = ACTIONS(93),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(97),
    [anon_sym_INIT] = ACTIONS(97),
    [anon_sym_CHECK] = ACTIONS(97),
    [anon_sym_UNITCHECK] = ACTIONS(97),
    [anon_sym_END] = ACTIONS(97),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(208)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(110),
    [sym_block] = STATE(1619),
    [sym__fullstmt] = STATE(1545),
    [sym_statement_label] = STATE(1621),
    [sym__barestmt] = STATE(1621),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(208),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(6044),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(11),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(209)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(123),
    [sym_block] = STATE(1717),
    [sym__barestmt] = STATE(390),
    [sym_package_statement] = STATE(1710),
    [sym_class_statement] = STATE(1710),
    [sym_role_statement] = STATE(1710),
    [sym_class_phaser_statement] = STATE(1710),
    [sym_use_version_statement] = STATE(1710),
    [sym_use_statement] = STATE(1710),
    [sym_subroutine_declaration_statement] = STATE(1710),
    [sym_method_declaration_statement] = STATE(1710),
    [sym_phaser_statement] = STATE(1710),
    [sym_conditional_statement] = STATE(1710),
    [sym_loop_statement] = STATE(1710),
    [sym_cstyle_for_statement] = STATE(1710),
    [sym_for_statement] = STATE(1710),
    [sym_try_statement] = STATE(1710),
    [sym_defer_statement] = STATE(1710),
    [sym_expression_statement] = STATE(4570),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(209),
    [sym__conditionals] = STATE(6036),
    [sym__loops] = STATE(6037),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5149),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_package] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2061),
    [anon_sym_role] = ACTIONS(2063),
    [anon_sym_BUILD] = ACTIONS(2065),
    [anon_sym_ADJUST] = ACTIONS(2065),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(2067),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(2069),
    [anon_sym_method] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(2073),
    [anon_sym_defer] = ACTIONS(2075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(2077),
    [anon_sym_no] = ACTIONS(2077),
    [anon_sym_for] = ACTIONS(2079),
    [anon_sym_foreach] = ACTIONS(2079),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(2081),
    [anon_sym_INIT] = ACTIONS(2081),
    [anon_sym_CHECK] = ACTIONS(2081),
    [anon_sym_UNITCHECK] = ACTIONS(2081),
    [anon_sym_END] = ACTIONS(2081),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(210)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(123),
    [sym_block] = STATE(1717),
    [sym__barestmt] = STATE(418),
    [sym_package_statement] = STATE(1710),
    [sym_class_statement] = STATE(1710),
    [sym_role_statement] = STATE(1710),
    [sym_class_phaser_statement] = STATE(1710),
    [sym_use_version_statement] = STATE(1710),
    [sym_use_statement] = STATE(1710),
    [sym_subroutine_declaration_statement] = STATE(1710),
    [sym_method_declaration_statement] = STATE(1710),
    [sym_phaser_statement] = STATE(1710),
    [sym_conditional_statement] = STATE(1710),
    [sym_loop_statement] = STATE(1710),
    [sym_cstyle_for_statement] = STATE(1710),
    [sym_for_statement] = STATE(1710),
    [sym_try_statement] = STATE(1710),
    [sym_defer_statement] = STATE(1710),
    [sym_expression_statement] = STATE(4570),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(210),
    [sym__conditionals] = STATE(6036),
    [sym__loops] = STATE(6037),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5149),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_package] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2061),
    [anon_sym_role] = ACTIONS(2063),
    [anon_sym_BUILD] = ACTIONS(2065),
    [anon_sym_ADJUST] = ACTIONS(2065),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(2067),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(2069),
    [anon_sym_method] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(2073),
    [anon_sym_defer] = ACTIONS(2075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(2077),
    [anon_sym_no] = ACTIONS(2077),
    [anon_sym_for] = ACTIONS(2079),
    [anon_sym_foreach] = ACTIONS(2079),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(2081),
    [anon_sym_INIT] = ACTIONS(2081),
    [anon_sym_CHECK] = ACTIONS(2081),
    [anon_sym_UNITCHECK] = ACTIONS(2081),
    [anon_sym_END] = ACTIONS(2081),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(211)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__barestmt] = STATE(210),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4010),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(211),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(212)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__barestmt] = STATE(209),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4006),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(212),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(213)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(131),
    [sym_block] = STATE(1619),
    [sym__barestmt] = STATE(214),
    [sym_package_statement] = STATE(1616),
    [sym_class_statement] = STATE(1616),
    [sym_role_statement] = STATE(1616),
    [sym_class_phaser_statement] = STATE(1616),
    [sym_use_version_statement] = STATE(1616),
    [sym_use_statement] = STATE(1616),
    [sym_subroutine_declaration_statement] = STATE(1616),
    [sym_method_declaration_statement] = STATE(1616),
    [sym_phaser_statement] = STATE(1616),
    [sym_conditional_statement] = STATE(1616),
    [sym_loop_statement] = STATE(1616),
    [sym_cstyle_for_statement] = STATE(1616),
    [sym_for_statement] = STATE(1616),
    [sym_try_statement] = STATE(1616),
    [sym_defer_statement] = STATE(1616),
    [sym_expression_statement] = STATE(4581),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(3996),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(213),
    [sym__conditionals] = STATE(6029),
    [sym__loops] = STATE(6030),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5138),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(1021),
    [anon_sym_package] = ACTIONS(1023),
    [anon_sym_class] = ACTIONS(1025),
    [anon_sym_role] = ACTIONS(1027),
    [anon_sym_BUILD] = ACTIONS(1029),
    [anon_sym_ADJUST] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(1033),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(1035),
    [anon_sym_method] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(1039),
    [anon_sym_defer] = ACTIONS(1041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(1047),
    [anon_sym_no] = ACTIONS(1047),
    [anon_sym_for] = ACTIONS(1049),
    [anon_sym_foreach] = ACTIONS(1049),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(1051),
    [anon_sym_INIT] = ACTIONS(1051),
    [anon_sym_CHECK] = ACTIONS(1051),
    [anon_sym_UNITCHECK] = ACTIONS(1051),
    [anon_sym_END] = ACTIONS(1051),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(214)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(123),
    [sym_block] = STATE(1717),
    [sym__barestmt] = STATE(413),
    [sym_package_statement] = STATE(1710),
    [sym_class_statement] = STATE(1710),
    [sym_role_statement] = STATE(1710),
    [sym_class_phaser_statement] = STATE(1710),
    [sym_use_version_statement] = STATE(1710),
    [sym_use_statement] = STATE(1710),
    [sym_subroutine_declaration_statement] = STATE(1710),
    [sym_method_declaration_statement] = STATE(1710),
    [sym_phaser_statement] = STATE(1710),
    [sym_conditional_statement] = STATE(1710),
    [sym_loop_statement] = STATE(1710),
    [sym_cstyle_for_statement] = STATE(1710),
    [sym_for_statement] = STATE(1710),
    [sym_try_statement] = STATE(1710),
    [sym_defer_statement] = STATE(1710),
    [sym_expression_statement] = STATE(4570),
    [sym__postfix_expressions] = STATE(4794),
    [sym_postfix_conditional_expression] = STATE(4869),
    [sym_postfix_loop_expression] = STATE(4869),
    [sym_postfix_for_expression] = STATE(4869),
    [sym_yadayada] = STATE(4794),
    [sym__expr] = STATE(4018),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(214),
    [sym__conditionals] = STATE(6036),
    [sym__loops] = STATE(6037),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5149),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_package] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2061),
    [anon_sym_role] = ACTIONS(2063),
    [anon_sym_BUILD] = ACTIONS(2065),
    [anon_sym_ADJUST] = ACTIONS(2065),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(2067),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(2069),
    [anon_sym_method] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_try] = ACTIONS(2073),
    [anon_sym_defer] = ACTIONS(2075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(47),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_use] = ACTIONS(2077),
    [anon_sym_no] = ACTIONS(2077),
    [anon_sym_for] = ACTIONS(2079),
    [anon_sym_foreach] = ACTIONS(2079),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym_BEGIN] = ACTIONS(2081),
    [anon_sym_INIT] = ACTIONS(2081),
    [anon_sym_CHECK] = ACTIONS(2081),
    [anon_sym_UNITCHECK] = ACTIONS(2081),
    [anon_sym_END] = ACTIONS(2081),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(121),
    [anon_sym_unless] = ACTIONS(121),
    [anon_sym_while] = ACTIONS(123),
    [anon_sym_until] = ACTIONS(123),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(215)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1939),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(215),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_for] = ACTIONS(192),
    [anon_sym_foreach] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(192),
    [anon_sym___END__] = ACTIONS(192),
    [anon_sym_EOT] = ACTIONS(192),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(192),
    [anon_sym_unless] = ACTIONS(192),
    [anon_sym_while] = ACTIONS(192),
    [anon_sym_until] = ACTIONS(192),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(209),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(216)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2012),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(216),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(192),
    [anon_sym___END__] = ACTIONS(192),
    [anon_sym_EOT] = ACTIONS(192),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym__ctrl_z_hack] = ACTIONS(209),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(217)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(4984),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(531),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(217),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2089),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2099),
  },
  [STATE(218)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(5134),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(559),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(218),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2101),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2103),
  },
  [STATE(219)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(5026),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(518),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(219),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2105),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2107),
  },
  [STATE(220)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(5068),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(544),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(220),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2111),
  },
  [STATE(221)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(4998),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(572),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(221),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2113),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2115),
  },
  [STATE(222)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(5098),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(499),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(222),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2117),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2119),
  },
  [STATE(223)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2493),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(223),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_RBRACK] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(224)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(4990),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(478),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(224),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2129),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2131),
  },
  [STATE(225)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1732),
    [sym__expr] = STATE(5032),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(428),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(225),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2133),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2135),
  },
  [STATE(226)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5577),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(226),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(227)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5602),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(227),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(228)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5675),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(228),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(229)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2956),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(229),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(192),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(209),
  },
  [STATE(230)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5616),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(230),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(231)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5604),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(231),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(232)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5763),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(232),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(233)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5778),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(233),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(234)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1492),
    [sym__listexpr] = STATE(5555),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(234),
    [sym__version] = STATE(394),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [sym_version] = ACTIONS(2167),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2171),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(235)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5989),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(235),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(236)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5971),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(236),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(237)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5579),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(237),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(238)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5631),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(238),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(239)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5991),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(239),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(240)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5686),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(240),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(241)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5977),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(241),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(242)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5699),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(242),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(243)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1724),
    [sym__listexpr] = STATE(5494),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(243),
    [sym__version] = STATE(392),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [sym_version] = ACTIONS(2167),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2175),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(244)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2901),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(244),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(209),
  },
  [STATE(245)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1543),
    [sym__listexpr] = STATE(5564),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(245),
    [sym__version] = STATE(412),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2177),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [sym_version] = ACTIONS(2167),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2179),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(246)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5655),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(246),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(247)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5670),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(247),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(248)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5676),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(248),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(249)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5692),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(249),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(250)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5958),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(250),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(251)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5837),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(251),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(252)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(6020),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(252),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(253)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5746),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(253),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(254)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5628),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(254),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(255)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5739),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(255),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(256)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5663),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(256),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(257)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5679),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(257),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(258)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(6031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(258),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(259)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5945),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(259),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(260)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2723),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(260),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_and] = ACTIONS(192),
    [anon_sym_or] = ACTIONS(192),
    [anon_sym_xor] = ACTIONS(192),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(209),
  },
  [STATE(261)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5693),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(261),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(262)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5697),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(262),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(263)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5702),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(263),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(264)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5720),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(264),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(265)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5781),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(265),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(266)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5999),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(266),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(267)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(6012),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(267),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(268)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5738),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(268),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(269)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5601),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(269),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(270)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5964),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(270),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(271)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5624),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(271),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(272)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5734),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(272),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(273)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5627),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(273),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(274)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5784),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(274),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(275)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5629),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(275),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(276)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5632),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(276),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(277)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5642),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(277),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(278)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5645),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(278),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(279)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5803),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(279),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(280)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5890),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(280),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(281)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(6016),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(281),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(282)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5981),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(282),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(283)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5626),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(283),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(284)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5821),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(284),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(285)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5698),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(285),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(286)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5988),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(286),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(287)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5706),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(287),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(288)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5953),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(288),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(289)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5700),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(289),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(290)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5625),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(290),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(291)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5605),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(291),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(292)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5657),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(292),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(293)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5825),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(293),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(294)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5589),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(294),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(295)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5673),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(295),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(296)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(6011),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(296),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(297)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5592),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(297),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(298)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5591),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(298),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(299)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5593),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(299),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(300)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5595),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(300),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(301)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5597),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(301),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(302)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5810),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(302),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(303)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5617),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(303),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(304)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5656),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(304),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(305)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5829),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(305),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(306)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5659),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(306),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(307)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5658),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(307),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(308)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5662),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(308),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(309)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5664),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(309),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(310)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5667),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(310),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(311)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5884),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(311),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(312)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5678),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(312),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(313)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5961),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(313),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(314)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5689),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(314),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(315)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5630),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(315),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(316)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5637),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(316),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(317)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5646),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(317),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(318)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5652),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(318),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(319)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5917),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(319),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(320)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5587),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(320),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(321)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5055),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym__hash_key] = STATE(5590),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(321),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym__brace_autoquoted] = STATE(5691),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__brace_autoquoted_token] = ACTIONS(2145),
  },
  [STATE(322)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5742),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(322),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(323)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5248),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym__hash_key] = STATE(5911),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(323),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym__brace_autoquoted] = STATE(5754),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__brace_autoquoted_token] = ACTIONS(2155),
  },
  [STATE(324)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5258),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5743),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(324),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2181),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(325)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5221),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(325),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2185),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2187),
  },
  [STATE(326)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5028),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(326),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2189),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2191),
  },
  [STATE(327)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5045),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(327),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2193),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2195),
  },
  [STATE(328)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5052),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(328),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2197),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2199),
  },
  [STATE(329)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5103),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(329),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2201),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2203),
  },
  [STATE(330)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5122),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(330),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2205),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2207),
  },
  [STATE(331)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(331),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2211),
  },
  [STATE(332)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(332),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2213),
  },
  [STATE(333)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5264),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5747),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(333),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2215),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(334)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5037),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(334),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2217),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2219),
  },
  [STATE(335)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5145),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(335),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2221),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2223),
  },
  [STATE(336)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4974),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(336),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2225),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2227),
  },
  [STATE(337)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5265),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(337),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2229),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2231),
  },
  [STATE(338)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4976),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(338),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2233),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2235),
  },
  [STATE(339)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5120),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(339),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2237),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2239),
  },
  [STATE(340)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4986),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(340),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2241),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2243),
  },
  [STATE(341)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5012),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(341),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2245),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2247),
  },
  [STATE(342)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5018),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(342),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2249),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2251),
  },
  [STATE(343)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5046),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(343),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2253),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2255),
  },
  [STATE(344)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4991),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(344),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2257),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2259),
  },
  [STATE(345)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5099),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(345),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2261),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2263),
  },
  [STATE(346)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5206),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(346),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2265),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2267),
  },
  [STATE(347)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5039),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(347),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2269),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2271),
  },
  [STATE(348)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5272),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5863),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(348),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2273),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(349)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5067),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(349),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2275),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2277),
  },
  [STATE(350)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5061),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(350),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2279),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2281),
  },
  [STATE(351)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5034),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(351),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2283),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2285),
  },
  [STATE(352)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5203),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5808),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(352),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2287),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(353)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5307),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(353),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2289),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2291),
  },
  [STATE(354)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5070),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(354),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2293),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2295),
  },
  [STATE(355)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5069),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(355),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2297),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2299),
  },
  [STATE(356)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5080),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(356),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2301),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2303),
  },
  [STATE(357)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4996),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(357),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2305),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2307),
  },
  [STATE(358)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5081),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(358),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2309),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2311),
  },
  [STATE(359)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5105),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(359),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2313),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2315),
  },
  [STATE(360)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5168),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5797),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(360),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2317),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(361)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5225),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(361),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2319),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2321),
  },
  [STATE(362)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5259),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5750),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(362),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2323),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(363)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5192),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5839),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(363),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2325),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(364)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5169),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(364),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2327),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2329),
  },
  [STATE(365)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5157),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(365),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2331),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2333),
  },
  [STATE(366)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(366),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2335),
  },
  [STATE(367)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5124),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(367),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2337),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2339),
  },
  [STATE(368)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5017),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(368),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2341),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2343),
  },
  [STATE(369)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5135),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(369),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2345),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2347),
  },
  [STATE(370)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5151),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(370),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2349),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2351),
  },
  [STATE(371)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5154),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(371),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2353),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2355),
  },
  [STATE(372)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4983),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(372),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_RBRACK] = ACTIONS(2357),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2359),
  },
  [STATE(373)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5087),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(373),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2361),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2363),
  },
  [STATE(374)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5298),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym__tricky_list] = STATE(5818),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(3270),
    [sym_interpolated_string_literal] = STATE(3270),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(3270),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(374),
    [sym_autoquoted_bareword] = STATE(3271),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(2365),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(375)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5021),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(375),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2367),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2369),
  },
  [STATE(376)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5109),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(376),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2371),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2373),
  },
  [STATE(377)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5113),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(377),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2375),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2377),
  },
  [STATE(378)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(4999),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(378),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2379),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2381),
  },
  [STATE(379)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5005),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(379),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2383),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2385),
  },
  [STATE(380)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5006),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(380),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_RPAREN] = ACTIONS(2387),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2389),
  },
  [STATE(381)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(381),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2391),
  },
  [STATE(382)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(382),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2393),
  },
  [STATE(383)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(383),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2395),
  },
  [STATE(384)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(384),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2397),
  },
  [STATE(385)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(385),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2399),
  },
  [STATE(386)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(386),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2401),
  },
  [STATE(387)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5236),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(387),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2403),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2405),
  },
  [STATE(388)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(119),
    [sym_block] = STATE(1407),
    [sym__listexpr] = STATE(2875),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_indirect_object] = STATE(642),
    [sym_function_call_expression] = STATE(2840),
    [sym__tricky_indirob_hashref] = STATE(3130),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1377),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(388),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(389)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5257),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(389),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2421),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(390)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5261),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(390),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2423),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(391)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5197),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(391),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2425),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(392)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1656),
    [sym__listexpr] = STATE(5383),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(392),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2175),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(393)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5255),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(393),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2427),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(394)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1525),
    [sym__listexpr] = STATE(5480),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(394),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2171),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(395)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(109),
    [sym_block] = STATE(1386),
    [sym__listexpr] = STATE(3465),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_indirect_object] = STATE(608),
    [sym_function_call_expression] = STATE(3427),
    [sym__tricky_indirob_hashref] = STATE(3528),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1367),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(395),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(396)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(103),
    [sym_block] = STATE(1312),
    [sym__listexpr] = STATE(2600),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_indirect_object] = STATE(677),
    [sym_function_call_expression] = STATE(2443),
    [sym__tricky_indirob_hashref] = STATE(2602),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(1269),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(396),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2437),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(397)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(105),
    [sym_block] = STATE(1346),
    [sym__listexpr] = STATE(2854),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_indirect_object] = STATE(709),
    [sym_function_call_expression] = STATE(2698),
    [sym__tricky_indirob_hashref] = STATE(2861),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(1305),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(397),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2439),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(398)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5307),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(398),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2289),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(399)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5260),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(399),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(400)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(109),
    [sym_block] = STATE(1386),
    [sym__listexpr] = STATE(3465),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_indirect_object] = STATE(663),
    [sym_function_call_expression] = STATE(3427),
    [sym__tricky_indirob_hashref] = STATE(3692),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1367),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(400),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(401)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5265),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(401),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2229),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(402)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5169),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(402),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2327),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(403)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5263),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(403),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2441),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(404)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(134),
    [sym_block] = STATE(1329),
    [sym__listexpr] = STATE(2697),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_indirect_object] = STATE(710),
    [sym_function_call_expression] = STATE(2655),
    [sym__tricky_indirob_hashref] = STATE(2699),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1296),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(404),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2087),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(405)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5225),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(405),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2319),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(406)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(119),
    [sym_block] = STATE(1407),
    [sym__listexpr] = STATE(2875),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_indirect_object] = STATE(692),
    [sym_function_call_expression] = STATE(2840),
    [sym__tricky_indirob_hashref] = STATE(3142),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1377),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(406),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(407)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(115),
    [sym_block] = STATE(1361),
    [sym__listexpr] = STATE(3181),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_indirect_object] = STATE(625),
    [sym_function_call_expression] = STATE(3150),
    [sym__tricky_indirob_hashref] = STATE(3182),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(1301),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(407),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2443),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(408)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5296),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(408),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2445),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(409)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5167),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(409),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2447),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(410)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(111),
    [sym_block] = STATE(1401),
    [sym__listexpr] = STATE(3568),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_indirect_object] = STATE(632),
    [sym_function_call_expression] = STATE(3537),
    [sym__tricky_indirob_hashref] = STATE(3569),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(1397),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(410),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2449),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(411)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(109),
    [sym_block] = STATE(1386),
    [sym__listexpr] = STATE(3465),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_indirect_object] = STATE(688),
    [sym_function_call_expression] = STATE(3427),
    [sym__tricky_indirob_hashref] = STATE(3690),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1367),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(411),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(412)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__semicolon] = STATE(1559),
    [sym__listexpr] = STATE(5442),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3401),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(412),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(2177),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(2179),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(413)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5183),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(413),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2455),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(414)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(119),
    [sym_block] = STATE(1407),
    [sym__listexpr] = STATE(2875),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_indirect_object] = STATE(656),
    [sym_function_call_expression] = STATE(2840),
    [sym__tricky_indirob_hashref] = STATE(2876),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1377),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(414),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(415)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5206),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(415),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2265),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(416)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(113),
    [sym_block] = STATE(1259),
    [sym__listexpr] = STATE(1909),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_indirect_object] = STATE(671),
    [sym_function_call_expression] = STATE(1892),
    [sym__tricky_indirob_hashref] = STATE(2074),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1234),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(416),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2457),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(417)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(109),
    [sym_block] = STATE(1386),
    [sym__listexpr] = STATE(3465),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_indirect_object] = STATE(607),
    [sym_function_call_expression] = STATE(3427),
    [sym__tricky_indirob_hashref] = STATE(3695),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1367),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(417),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(418)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5290),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(418),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2459),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(419)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5236),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(419),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2403),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(420)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5191),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(420),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2461),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(421)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5221),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(421),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2185),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(422)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(101),
    [sym_block] = STATE(1154),
    [sym__listexpr] = STATE(1909),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_indirect_object] = STATE(595),
    [sym_function_call_expression] = STATE(1892),
    [sym__tricky_indirob_hashref] = STATE(1860),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1138),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(422),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2463),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(423)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5242),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(423),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_RPAREN] = ACTIONS(2465),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(424)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5126),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(424),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(425)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(130),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2862),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym__sort_routine] = STATE(635),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(425),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(2471),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(426)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5282),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(426),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(427)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4488),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2010),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(427),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(428)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5140),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(428),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(429)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5144),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(429),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(430)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3672),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(430),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_RBRACE] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_RPAREN] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(431)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5047),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(431),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(432)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5153),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(432),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(433)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5027),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(433),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(434)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5306),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(434),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(435)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5166),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(435),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(436)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5063),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(436),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(437)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3458),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(606),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(437),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2473),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(438)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(114),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2562),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym__sort_routine] = STATE(651),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(1308),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(438),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(2475),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2477),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(2479),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(2475),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(439)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5048),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(439),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(440)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5123),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(440),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(441)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(130),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2862),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym__sort_routine] = STATE(641),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(441),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(2471),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(442)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5086),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(442),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(443)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(4969),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(443),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(444)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(4707),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(444),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(445)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3458),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(661),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(445),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2473),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(446)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(130),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2862),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym__sort_routine] = STATE(685),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(446),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(2471),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(447)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(118),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2803),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym__sort_routine] = STATE(706),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(1345),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(447),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(2481),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2483),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(2485),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(2481),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(448)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5056),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(448),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(449)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5060),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(449),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(450)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3458),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(603),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(450),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2473),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(451)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(108),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(1847),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym__sort_routine] = STATE(631),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1164),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(451),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(2487),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2489),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(2491),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(2487),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(452)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4020),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2010),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(452),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(453)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4515),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2010),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(453),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(454)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5170),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(454),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(455)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5189),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(455),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(456)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(122),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3561),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym__sort_routine] = STATE(623),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(1414),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(456),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(2493),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2495),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(2497),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(2493),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(457)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5202),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(457),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(458)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5246),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(458),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(459)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5051),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(459),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(460)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__expr] = STATE(5049),
    [sym_lowprec_logical_expression] = STATE(5152),
    [sym__listexpr] = STATE(5152),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2358),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(460),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(461)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5280),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(461),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(462)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(125),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(1847),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym__sort_routine] = STATE(645),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1244),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(462),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(2499),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2501),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(2491),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(2499),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(463)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5287),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(463),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(464)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5284),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(464),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(465)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5121),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(465),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(466)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5133),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(466),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(467)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5141),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(467),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(468)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4985),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(468),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(469)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5207),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(469),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(470)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5238),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(470),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(471)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5302),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(471),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(472)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5188),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(472),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(473)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5279),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(473),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(474)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5828),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(668),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(474),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(475)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4997),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(475),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(476)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5303),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(476),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(477)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5317),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(477),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(478)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5014),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(478),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(479)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5015),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(479),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(480)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5016),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(480),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(481)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5205),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(481),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(482)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3458),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(687),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(482),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2473),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(483)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5966),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(615),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(483),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(484)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5088),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(484),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(485)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5089),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(485),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(486)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5090),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(486),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(487)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5091),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(487),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(488)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5190),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(488),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(489)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5193),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(489),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(490)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5194),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(490),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(491)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(127),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(3174),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym__sort_routine] = STATE(599),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(1356),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(491),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(2503),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2505),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(2507),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(2503),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(492)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5195),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(492),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(493)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5204),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(493),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(494)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5773),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(591),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(494),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(495)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5100),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(495),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(496)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5224),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(496),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(497)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5227),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(497),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(498)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5223),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(498),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(499)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5106),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(499),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(500)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4020),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(500),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(501)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5107),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(501),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(502)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5108),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(502),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(503)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5262),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(503),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(504)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4019),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1809),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(504),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(505)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5110),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(505),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(506)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5112),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(506),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(507)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5116),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(507),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(508)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5118),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(508),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(509)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5179),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(509),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(510)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5278),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(510),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(511)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5281),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(511),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(512)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5171),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(512),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(513)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5249),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(513),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(514)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(6004),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(619),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(514),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(515)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5029),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(515),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(516)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5330),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(516),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(517)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5163),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(517),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(518)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5038),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(518),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(519)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5042),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(519),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(520)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5043),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(520),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(521)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5210),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(521),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(522)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4977),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(522),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(523)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4978),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(523),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(524)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4979),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(524),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(525)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4980),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(525),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(526)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5215),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(526),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(527)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5247),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(527),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(528)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5907),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(639),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(528),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(529)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4987),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(529),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(530)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5177),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(530),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(531)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5001),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(531),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(532)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4546),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2010),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(532),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(533)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5010),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(533),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(534)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5011),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(534),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(535)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5062),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(535),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(536)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5064),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(536),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(537)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5065),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(537),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(538)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5066),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(538),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(539)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5304),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(539),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(540)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5309),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(540),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(541)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5970),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(655),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(541),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(542)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5071),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(542),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(543)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5161),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(543),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(544)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5075),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(544),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(545)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5076),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(545),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(546)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(124),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(2682),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym__sort_routine] = STATE(675),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(1322),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(546),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(2509),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2511),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(2513),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(2509),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(547)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5079),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(547),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(548)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5074),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(548),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(549)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5125),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(549),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(550)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__expr] = STATE(4511),
    [sym_lowprec_logical_expression] = STATE(4021),
    [sym__listexpr] = STATE(4021),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2010),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(550),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(551)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5228),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(551),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(552)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5127),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(552),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(553)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5128),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(553),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(554)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5291),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(554),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(555)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5293),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(555),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(556)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(5901),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(679),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(556),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(557)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5136),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(557),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(558)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5300),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(558),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(559)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5142),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(559),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(560)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5146),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(560),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(561)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3587),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(561),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_SEMI] = ACTIONS(192),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_COMMA] = ACTIONS(192),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_EQ_GT] = ACTIONS(192),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_SEMICOLON] = ACTIONS(209),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(562)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5148),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(562),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(563)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4992),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(563),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(564)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4993),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(564),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(565)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4994),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(565),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(566)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(4995),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(566),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(567)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5308),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(567),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(568)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5178),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(568),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(569)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(1733),
    [sym__listexpr] = STATE(6017),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym__sort_routine] = STATE(695),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(1385),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(569),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2467),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2469),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2467),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(570)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5000),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(570),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(571)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5208),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(571),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(572)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__expr] = STATE(5002),
    [sym_lowprec_logical_expression] = STATE(4982),
    [sym__listexpr] = STATE(4982),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2392),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(572),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(573)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5003),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(573),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(574)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__expr] = STATE(5004),
    [sym_lowprec_logical_expression] = STATE(5094),
    [sym__listexpr] = STATE(5094),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2540),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(574),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(575)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5235),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(575),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(576)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5237),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(576),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(577)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5245),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(577),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(578)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5256),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(578),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(579)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5271),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(579),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(580)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5172),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(580),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(581)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5173),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(581),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(582)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5175),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(582),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(583)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5181),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(583),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(584)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5159),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(584),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(585)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__expr] = STATE(5211),
    [sym_lowprec_logical_expression] = STATE(4964),
    [sym__listexpr] = STATE(4964),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2209),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(585),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(586)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3283),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(586),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(2515),
    [anon_sym_DOLLAR] = ACTIONS(2517),
    [anon_sym_AT] = ACTIONS(2520),
    [anon_sym_COMMA] = ACTIONS(2515),
    [anon_sym_RPAREN] = ACTIONS(2515),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2527),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(587)] = {
    [sym_heredoc_content] = STATE(587),
    [ts_builtin_sym_end] = ACTIONS(2530),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2534),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2534),
    [anon_sym___END__] = ACTIONS(2534),
    [anon_sym_EOT] = ACTIONS(2534),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_unless] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_until] = ACTIONS(2534),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(588)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3193),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(588),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(589)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(3045),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(589),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(590)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5997),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(590),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(591)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5998),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(591),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(592)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3523),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(592),
    [sym__version] = STATE(3457),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [sym_version] = ACTIONS(2538),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(2540),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(593)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5851),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(593),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(594)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3524),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3459),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(594),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(595)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1846),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(595),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(596)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2776),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(596),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(597)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5774),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(597),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(598)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3486),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(598),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(599)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3194),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(599),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(600)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2565),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym__label_arg] = STATE(2566),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(600),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_identifier] = STATE(2567),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(360),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(601)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5819),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(601),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(602)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3739),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(602),
    [sym__version] = STATE(3457),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [sym_version] = ACTIONS(2538),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(2540),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(603)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3487),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(603),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(604)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3413),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(604),
    [sym__version] = STATE(2860),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [sym_version] = ACTIONS(2552),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(2554),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(605)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3486),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(605),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(606)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3487),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(606),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(607)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3497),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(607),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(608)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3497),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(608),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(609)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3504),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(609),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(610)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3417),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3438),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(610),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(611)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3504),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(611),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(612)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3559),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(612),
    [sym__version] = STATE(3560),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [sym_version] = ACTIONS(2556),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(2558),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(613)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3417),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3525),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(613),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(614)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3562),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym__label_arg] = STATE(3563),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(614),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_identifier] = STATE(3564),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(721),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(615)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5856),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(615),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(616)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3724),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(616),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_COMMA] = ACTIONS(182),
    [anon_sym_RPAREN] = ACTIONS(182),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(617)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1868),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(617),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(618)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5867),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(618),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(619)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5870),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(619),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(620)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1881),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym__label_arg] = STATE(1882),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(620),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(1942),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(151),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(621)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3580),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(621),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(622)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3258),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2863),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(622),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(623)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3581),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(623),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(624)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5916),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(624),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(625)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3200),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(625),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(626)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3719),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(626),
    [sym__version] = STATE(3457),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [sym_version] = ACTIONS(2538),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(2540),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(627)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3211),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(627),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(628)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2892),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(628),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(629)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym__listexpr] = STATE(3228),
    [sym__term_rightward] = STATE(3146),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2605),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(629),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(630)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3129),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2863),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(630),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(631)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1871),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(631),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(632)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3602),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(632),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(633)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3613),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(633),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(634)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym__listexpr] = STATE(3633),
    [sym__term_rightward] = STATE(3532),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3398),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(634),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(635)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2893),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(635),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(636)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2071),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(636),
    [sym__version] = STATE(1859),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [sym_version] = ACTIONS(2560),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(2562),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(637)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2892),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(637),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(638)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(6023),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(638),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(639)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(6028),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(639),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(640)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2072),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym__label_arg] = STATE(1882),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(640),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_identifier] = STATE(1942),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(211),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(641)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2893),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(641),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(642)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2916),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(642),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(643)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1868),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(643),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(644)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5995),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(644),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(645)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1871),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(645),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(646)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1952),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(646),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(647)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2932),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(647),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(648)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2953),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2344),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(648),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(649)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1829),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(649),
    [sym__version] = STATE(1859),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [sym_version] = ACTIONS(2560),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(2562),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(650)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2526),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(650),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(651)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2351),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(651),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(652)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(3027),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(652),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(653)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5822),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(653),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(654)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5807),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(654),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(655)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5812),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(655),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(656)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2916),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(656),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(657)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2680),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(657),
    [sym__version] = STATE(2681),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [sym_version] = ACTIONS(2564),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(2566),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(658)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3740),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3459),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(658),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(659)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3486),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(659),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(660)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2683),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym__label_arg] = STATE(2684),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(660),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_identifier] = STATE(2685),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(649),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(661)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3487),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(661),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(662)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5922),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(662),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(663)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3497),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(663),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(664)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3362),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(664),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_COLON] = ACTIONS(2515),
    [anon_sym_DOLLAR] = ACTIONS(2517),
    [anon_sym_AT] = ACTIONS(2520),
    [anon_sym_RPAREN] = ACTIONS(2515),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2527),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(665)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1921),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1784),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(665),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(666)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3504),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(666),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(667)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5785),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(667),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(668)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5795),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(668),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(669)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2712),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(669),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(670)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3417),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3533),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(670),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(671)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1846),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(671),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(672)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2932),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(672),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(673)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1952),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(673),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(674)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym__listexpr] = STATE(1921),
    [sym__term_rightward] = STATE(1945),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1998),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(674),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(675)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2714),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(675),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(676)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3414),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym__label_arg] = STATE(2863),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(676),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_identifier] = STATE(2864),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(1015),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(677)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2330),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(677),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(678)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5938),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(678),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(679)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5939),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(679),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(680)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5973),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(680),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(681)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3796),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(681),
    [sym__version] = STATE(3457),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [sym_version] = ACTIONS(2538),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(2540),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(682)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2892),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(682),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(683)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3720),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3459),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(683),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(684)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(6014),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(684),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(685)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2893),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(685),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(686)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3486),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(686),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(687)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3487),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(687),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(688)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3497),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(688),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(689)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3504),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(689),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(690)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(3417),
    [sym__term_rightward] = STATE(3514),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3623),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(690),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(691)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3128),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(691),
    [sym__version] = STATE(2860),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [sym_version] = ACTIONS(2552),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(2554),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(692)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2916),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(692),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(693)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2932),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(693),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(694)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5902),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(694),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(695)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5940),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(695),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(696)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2953),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2355),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(696),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(697)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2452),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(697),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(698)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym__listexpr] = STATE(2611),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2250),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(698),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(699)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym__listexpr] = STATE(5727),
    [sym__term_rightward] = STATE(2376),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3386),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(699),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(700)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2793),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(700),
    [sym__version] = STATE(2794),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [sym_version] = ACTIONS(2572),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(2574),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(701)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3172),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(701),
    [sym__version] = STATE(3173),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [sym_version] = ACTIONS(2576),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(2578),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(702)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym__listexpr] = STATE(2953),
    [sym__term_rightward] = STATE(2836),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2324),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(702),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(703)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2814),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym__label_arg] = STATE(2829),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(703),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_identifier] = STATE(2830),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(651),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(704)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(2922),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(704),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(705)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3175),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym__label_arg] = STATE(3176),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(705),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_identifier] = STATE(3177),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(647),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(706)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(2924),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(706),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(707)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3772),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym__label_arg] = STATE(3459),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(707),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_identifier] = STATE(3460),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(915),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(708)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2557),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(708),
    [sym__version] = STATE(2558),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [sym_version] = ACTIONS(2580),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(709)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym__listexpr] = STATE(3011),
    [sym__term_rightward] = STATE(2692),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2396),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(709),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(710)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2738),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(710),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(711)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym__listexpr] = STATE(2750),
    [sym__term_rightward] = STATE(2651),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2270),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(711),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(712)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3240),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(712),
    [sym__version] = STATE(2860),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [sym_version] = ACTIONS(2552),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(2554),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(713)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3588),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(713),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
    [sym__NONASSOC] = ACTIONS(2584),
  },
  [STATE(714)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(605),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3726),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(714),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2586),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(715)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(659),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3731),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(715),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2586),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(716)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(628),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3744),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(716),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(717)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(637),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3747),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(717),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(718)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(682),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3743),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(718),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2588),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(719)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3673),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(719),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
    [sym__NONASSOC] = ACTIONS(2590),
  },
  [STATE(720)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3685),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(720),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
    [sym__NONASSOC] = ACTIONS(2592),
  },
  [STATE(721)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3686),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(721),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
    [sym__NONASSOC] = ACTIONS(2594),
  },
  [STATE(722)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(121),
    [sym_block] = STATE(3552),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3553),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(722),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(723)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(121),
    [sym_block] = STATE(3554),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3555),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(723),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(724)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3801),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(724),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
    [sym__NONASSOC] = ACTIONS(2590),
  },
  [STATE(725)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3792),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(725),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
    [sym__NONASSOC] = ACTIONS(2592),
  },
  [STATE(726)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3795),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(726),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
    [sym__NONASSOC] = ACTIONS(2594),
  },
  [STATE(727)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2915),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(727),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__NONASSOC] = ACTIONS(2596),
  },
  [STATE(728)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3600),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(728),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
    [sym__NONASSOC] = ACTIONS(2598),
  },
  [STATE(729)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3601),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(729),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
    [sym__NONASSOC] = ACTIONS(2600),
  },
  [STATE(730)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(102),
    [sym_block] = STATE(1918),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2067),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(730),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(731)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(102),
    [sym_block] = STATE(1912),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2068),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(731),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(732)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1948),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(732),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__NONASSOC] = ACTIONS(2602),
  },
  [STATE(733)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(593),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3745),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(733),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(734)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2855),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3264),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(734),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(735)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3284),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(735),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
    [sym__NONASSOC] = ACTIONS(2604),
  },
  [STATE(736)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3308),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(736),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
    [sym__NONASSOC] = ACTIONS(2606),
  },
  [STATE(737)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3276),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(737),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
    [sym__NONASSOC] = ACTIONS(2608),
  },
  [STATE(738)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3098),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(738),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
    [sym__NONASSOC] = ACTIONS(2604),
  },
  [STATE(739)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3110),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(739),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
    [sym__NONASSOC] = ACTIONS(2606),
  },
  [STATE(740)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3111),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(740),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
    [sym__NONASSOC] = ACTIONS(2608),
  },
  [STATE(741)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2041),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(741),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
    [sym__NONASSOC] = ACTIONS(2602),
  },
  [STATE(742)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2053),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(742),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
    [sym__NONASSOC] = ACTIONS(2610),
  },
  [STATE(743)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3451),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3783),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(743),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(744)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3452),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3786),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(744),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(745)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(704),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3760),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(745),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2612),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(746)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2054),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(746),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
    [sym__NONASSOC] = ACTIONS(2614),
  },
  [STATE(747)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2856),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3266),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(747),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(748)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3698),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(748),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
    [sym__NONASSOC] = ACTIONS(2590),
  },
  [STATE(749)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3710),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(749),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
    [sym__NONASSOC] = ACTIONS(2592),
  },
  [STATE(750)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3711),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(750),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
    [sym__NONASSOC] = ACTIONS(2594),
  },
  [STATE(751)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2855),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3124),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(751),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(752)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2856),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3125),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(752),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(753)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(686),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3752),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(753),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2586),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(754)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(126),
    [sym_block] = STATE(3165),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3166),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(754),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(755)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(126),
    [sym_block] = STATE(3167),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3168),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(755),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(756)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(102),
    [sym_block] = STATE(1918),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1884),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(756),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(757)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3451),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3735),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(757),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(758)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3452),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3736),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(758),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(759)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2855),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3409),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(759),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(760)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(129),
    [sym_block] = STATE(2856),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3410),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(760),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(761)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2914),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(761),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__NONASSOC] = ACTIONS(2616),
  },
  [STATE(762)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3751),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(762),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
    [sym__NONASSOC] = ACTIONS(2592),
  },
  [STATE(763)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3753),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(763),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
    [sym__NONASSOC] = ACTIONS(2594),
  },
  [STATE(764)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(102),
    [sym_block] = STATE(1912),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1913),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(764),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(765)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1844),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(765),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__NONASSOC] = ACTIONS(2610),
  },
  [STATE(766)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1845),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(766),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
    [sym__NONASSOC] = ACTIONS(2614),
  },
  [STATE(767)] = {
    [sym_heredoc_content] = STATE(767),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2534),
    [anon_sym___END__] = ACTIONS(2534),
    [anon_sym_EOT] = ACTIONS(2534),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_unless] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_until] = ACTIONS(2534),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(768)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3451),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3715),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(768),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(769)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3452),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3716),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(769),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(770)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2724),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(770),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__NONASSOC] = ACTIONS(2618),
  },
  [STATE(771)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2736),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(771),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__NONASSOC] = ACTIONS(2620),
  },
  [STATE(772)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2737),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(772),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
    [sym__NONASSOC] = ACTIONS(2622),
  },
  [STATE(773)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(650),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3756),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(773),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2624),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(774)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3454),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(774),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
    [sym__NONASSOC] = ACTIONS(2604),
  },
  [STATE(775)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3498),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(775),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
    [sym__NONASSOC] = ACTIONS(2606),
  },
  [STATE(776)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3499),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(776),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
    [sym__NONASSOC] = ACTIONS(2608),
  },
  [STATE(777)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(598),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3770),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(777),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2586),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(778)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2521),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(778),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2626),
  },
  [STATE(779)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2313),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(779),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2628),
  },
  [STATE(780)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2327),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(780),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
    [sym__NONASSOC] = ACTIONS(2630),
  },
  [STATE(781)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(128),
    [sym_block] = STATE(2671),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2672),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(781),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(782)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(128),
    [sym_block] = STATE(2673),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2674),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(782),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(783)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(116),
    [sym_block] = STATE(2775),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2777),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(783),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(784)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(116),
    [sym_block] = STATE(2779),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2781),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(784),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(785)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(621),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3759),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(785),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2632),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(786)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(643),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3741),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(786),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(787)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(588),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3750),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(787),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(788)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(112),
    [sym_block] = STATE(2549),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2550),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(788),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(789)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(112),
    [sym_block] = STATE(2551),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2552),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(789),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(790)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2958),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(790),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__NONASSOC] = ACTIONS(2638),
  },
  [STATE(791)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3009),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(791),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__NONASSOC] = ACTIONS(2640),
  },
  [STATE(792)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3010),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(792),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
    [sym__NONASSOC] = ACTIONS(2642),
  },
  [STATE(793)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(669),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3758),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(793),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2644),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(794)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(667),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3761),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(794),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(795)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3451),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3519),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(795),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(796)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(120),
    [sym_block] = STATE(3452),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3520),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(796),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(797)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(590),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3766),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(797),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(798)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(618),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3767),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(798),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(799)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(638),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3768),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(799),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(800)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(654),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3769),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(800),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(801)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(678),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3746),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(801),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(802)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(694),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3757),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(802),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(803)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(104),
    [sym_block] = STATE(617),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3749),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(803),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(2634),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(804)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2902),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(804),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
    [sym__NONASSOC] = ACTIONS(2646),
  },
  [STATE(805)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3762),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(805),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
    [sym__NONASSOC] = ACTIONS(2590),
  },
  [STATE(806)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2726),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(806),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(807)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1866),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(807),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(808)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3681),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(808),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(809)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3682),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(809),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(810)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1938),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(810),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(811)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2034),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(811),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(812)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3261),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2431),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(812),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(813)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2041),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(813),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(814)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2042),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(814),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(815)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2043),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(815),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(816)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2044),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(816),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(817)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3683),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(817),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(818)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2045),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(818),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(819)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3097),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(819),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(820)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2046),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(820),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(821)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3098),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(821),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(822)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2047),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(822),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(823)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3099),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(823),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(824)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3779),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(824),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(825)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2077),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(825),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(826)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3100),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(826),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(827)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3101),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(827),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(828)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2049),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(828),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(829)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3102),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(829),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(830)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2050),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(830),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(831)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3103),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(831),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(832)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2051),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(832),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(833)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3104),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(833),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(834)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3105),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(834),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(835)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2052),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(835),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(836)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3106),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(836),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(837)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2053),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(837),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(838)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3107),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(838),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(839)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3108),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(839),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(840)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3362),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(840),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(841)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3109),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(841),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(842)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3110),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(842),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(843)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3111),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(843),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(844)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2054),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(844),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(845)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3301),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(845),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(846)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3684),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(846),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(847)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3685),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(847),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(848)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3686),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(848),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(849)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3801),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(849),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(850)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3802),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(850),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(851)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3804),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(851),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(852)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3112),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(852),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(853)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3363),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(853),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(854)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3260),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(854),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(855)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3671),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(855),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(856)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3777),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(856),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(857)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3777),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(857),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2648),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2648),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2650),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(858)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3448),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(858),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(859)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3782),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3282),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(859),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(860)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3298),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(860),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(861)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3799),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(861),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(862)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3787),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(862),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(863)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3789),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(863),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(864)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1948),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(864),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(865)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3526),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(865),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(866)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3805),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(866),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(867)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2075),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(867),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(868)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1856),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1819),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(868),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(869)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3161),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(869),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(870)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3548),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(870),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(871)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3696),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(871),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(872)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3698),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(872),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(873)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3699),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(873),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(874)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3700),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(874),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(875)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3701),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(875),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(876)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3702),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(876),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(877)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3703),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(877),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(878)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3704),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(878),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(879)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3705),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(879),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(880)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3706),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(880),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(881)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3707),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(881),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(882)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3708),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(882),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(883)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3709),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(883),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(884)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3710),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(884),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(885)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3711),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(885),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(886)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2489),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(886),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(887)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1824),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(887),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(888)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2521),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(888),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(889)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3548),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(889),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(2652),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(2652),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2654),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(890)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3550),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(890),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(891)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3551),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3347),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(891),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(892)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3691),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(892),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(893)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3161),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(893),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(2656),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(2656),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2658),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(894)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3163),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(894),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(895)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3122),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(895),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(896)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3122),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(896),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2660),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2660),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2662),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(897)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2853),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(897),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(898)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3123),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2431),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(898),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(899)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3673),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(899),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(900)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3674),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(900),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(901)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3126),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(901),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(902)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3127),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(388),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(902),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(949),
    [anon_sym_DASH] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(949),
    [anon_sym_BANG] = ACTIONS(949),
    [anon_sym_PLUS_PLUS] = ACTIONS(2417),
    [anon_sym_DASH_DASH] = ACTIONS(2417),
    [anon_sym_BSLASH] = ACTIONS(954),
    [anon_sym_do] = ACTIONS(956),
    [anon_sym_eval] = ACTIONS(958),
    [anon_sym_await] = ACTIONS(960),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(962),
    [anon_sym_dynamically] = ACTIONS(962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(966),
    [anon_sym_map] = ACTIONS(968),
    [anon_sym_grep] = ACTIONS(968),
    [anon_sym_sort] = ACTIONS(970),
    [anon_sym_goto] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(974),
    [anon_sym_undef] = ACTIONS(976),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(981),
    [anon_sym_next] = ACTIONS(981),
    [anon_sym_redo] = ACTIONS(981),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(983),
    [anon_sym_alarm] = ACTIONS(983),
    [anon_sym_chop] = ACTIONS(983),
    [anon_sym_chdir] = ACTIONS(983),
    [anon_sym_close] = ACTIONS(983),
    [anon_sym_closedir] = ACTIONS(983),
    [anon_sym_caller] = ACTIONS(983),
    [anon_sym_chomp] = ACTIONS(983),
    [anon_sym_chr] = ACTIONS(983),
    [anon_sym_cos] = ACTIONS(983),
    [anon_sym_chroot] = ACTIONS(983),
    [anon_sym_defined] = ACTIONS(983),
    [anon_sym_delete] = ACTIONS(983),
    [anon_sym_dbmclose] = ACTIONS(983),
    [anon_sym_exists] = ACTIONS(983),
    [anon_sym_exit] = ACTIONS(983),
    [anon_sym_eof] = ACTIONS(983),
    [anon_sym_exp] = ACTIONS(983),
    [anon_sym_each] = ACTIONS(983),
    [anon_sym_fc] = ACTIONS(983),
    [anon_sym_fileno] = ACTIONS(983),
    [anon_sym_gmtime] = ACTIONS(983),
    [anon_sym_getc] = ACTIONS(983),
    [anon_sym_getpgrp] = ACTIONS(983),
    [anon_sym_getprotobyname] = ACTIONS(983),
    [anon_sym_getpwname] = ACTIONS(983),
    [anon_sym_getpwuid] = ACTIONS(983),
    [anon_sym_getpeername] = ACTIONS(983),
    [anon_sym_getnetbyname] = ACTIONS(983),
    [anon_sym_getsockname] = ACTIONS(983),
    [anon_sym_getgrnam] = ACTIONS(983),
    [anon_sym_getgrgid] = ACTIONS(983),
    [anon_sym_hex] = ACTIONS(983),
    [anon_sym_int] = ACTIONS(983),
    [anon_sym_keys] = ACTIONS(983),
    [anon_sym_lc] = ACTIONS(983),
    [anon_sym_lcfirst] = ACTIONS(983),
    [anon_sym_length] = ACTIONS(983),
    [anon_sym_localtime] = ACTIONS(983),
    [anon_sym_log] = ACTIONS(983),
    [anon_sym_lock] = ACTIONS(983),
    [anon_sym_lstat] = ACTIONS(983),
    [anon_sym_oct] = ACTIONS(983),
    [anon_sym_ord] = ACTIONS(983),
    [anon_sym_prototype] = ACTIONS(983),
    [anon_sym_pop] = ACTIONS(983),
    [anon_sym_pos] = ACTIONS(983),
    [anon_sym_quotemeta] = ACTIONS(983),
    [anon_sym_reset] = ACTIONS(983),
    [anon_sym_rand] = ACTIONS(983),
    [anon_sym_rmdir] = ACTIONS(983),
    [anon_sym_readdir] = ACTIONS(983),
    [anon_sym_readline] = ACTIONS(983),
    [anon_sym_readpipe] = ACTIONS(983),
    [anon_sym_rewinddir] = ACTIONS(983),
    [anon_sym_readlink] = ACTIONS(983),
    [anon_sym_ref] = ACTIONS(983),
    [anon_sym_scalar] = ACTIONS(983),
    [anon_sym_shift] = ACTIONS(983),
    [anon_sym_sin] = ACTIONS(983),
    [anon_sym_sleep] = ACTIONS(983),
    [anon_sym_sqrt] = ACTIONS(983),
    [anon_sym_srand] = ACTIONS(983),
    [anon_sym_stat] = ACTIONS(983),
    [anon_sym_study] = ACTIONS(983),
    [anon_sym_tell] = ACTIONS(983),
    [anon_sym_telldir] = ACTIONS(983),
    [anon_sym_tied] = ACTIONS(983),
    [anon_sym_uc] = ACTIONS(983),
    [anon_sym_ucfirst] = ACTIONS(983),
    [anon_sym_untie] = ACTIONS(983),
    [anon_sym_umask] = ACTIONS(983),
    [anon_sym_values] = ACTIONS(983),
    [anon_sym_write] = ACTIONS(983),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1013),
  },
  [STATE(903)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2528),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(903),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(904)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3556),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(904),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(905)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3164),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2450),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(905),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(906)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2529),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(906),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(907)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3557),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(907),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(908)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2530),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(908),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(909)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3810),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(909),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(910)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2576),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(910),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(911)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3169),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(911),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(912)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2587),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(912),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(913)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3794),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(913),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(914)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(3170),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(914),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(915)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2604),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(915),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(916)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2612),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(916),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(917)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3807),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(917),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(918)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2620),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(918),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(919)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2272),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(919),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(920)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2297),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(920),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(921)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2309),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(921),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(922)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3267),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(922),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(923)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3755),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(923),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(924)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1956),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(924),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(925)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3762),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(925),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(926)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3763),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(926),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(927)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3764),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(927),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(928)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3765),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(928),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(929)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3723),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(929),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(930)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3725),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(930),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(931)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3727),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(931),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(932)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3728),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(932),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(933)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3729),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(933),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(934)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3730),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(934),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(935)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3742),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(935),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(936)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3748),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(936),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(937)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3751),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(937),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(938)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3753),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(938),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(939)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2722),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(939),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(940)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3733),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(940),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(941)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3733),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(941),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2648),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2648),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2650),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(942)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3448),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(942),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(943)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3734),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3282),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(943),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(944)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3809),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(944),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(945)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3780),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(945),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(946)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3737),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(946),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(947)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3738),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(947),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(948)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3407),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(948),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(949)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3407),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(949),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2660),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2660),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2662),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(950)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2853),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(950),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(951)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3408),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2431),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(951),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(952)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3785),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(952),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(953)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3797),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(953),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(954)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3411),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(954),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(955)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3412),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(955),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(956)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2724),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(956),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(957)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3790),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(957),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(958)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2725),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(958),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(959)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3791),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(959),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(960)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2975),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(960),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(961)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3792),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(961),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(962)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2727),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(962),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(963)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3754),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(400),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(963),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1738),
    [anon_sym_DASH] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(1738),
    [anon_sym_BANG] = ACTIONS(1738),
    [anon_sym_PLUS_PLUS] = ACTIONS(2544),
    [anon_sym_DASH_DASH] = ACTIONS(2544),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1742),
    [anon_sym_eval] = ACTIONS(1744),
    [anon_sym_await] = ACTIONS(1746),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(1748),
    [anon_sym_dynamically] = ACTIONS(1748),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1750),
    [anon_sym_map] = ACTIONS(1752),
    [anon_sym_grep] = ACTIONS(1752),
    [anon_sym_sort] = ACTIONS(1754),
    [anon_sym_goto] = ACTIONS(1756),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_undef] = ACTIONS(1760),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1762),
    [anon_sym_next] = ACTIONS(1762),
    [anon_sym_redo] = ACTIONS(1762),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1764),
    [anon_sym_alarm] = ACTIONS(1764),
    [anon_sym_chop] = ACTIONS(1764),
    [anon_sym_chdir] = ACTIONS(1764),
    [anon_sym_close] = ACTIONS(1764),
    [anon_sym_closedir] = ACTIONS(1764),
    [anon_sym_caller] = ACTIONS(1764),
    [anon_sym_chomp] = ACTIONS(1764),
    [anon_sym_chr] = ACTIONS(1764),
    [anon_sym_cos] = ACTIONS(1764),
    [anon_sym_chroot] = ACTIONS(1764),
    [anon_sym_defined] = ACTIONS(1764),
    [anon_sym_delete] = ACTIONS(1764),
    [anon_sym_dbmclose] = ACTIONS(1764),
    [anon_sym_exists] = ACTIONS(1764),
    [anon_sym_exit] = ACTIONS(1764),
    [anon_sym_eof] = ACTIONS(1764),
    [anon_sym_exp] = ACTIONS(1764),
    [anon_sym_each] = ACTIONS(1764),
    [anon_sym_fc] = ACTIONS(1764),
    [anon_sym_fileno] = ACTIONS(1764),
    [anon_sym_gmtime] = ACTIONS(1764),
    [anon_sym_getc] = ACTIONS(1764),
    [anon_sym_getpgrp] = ACTIONS(1764),
    [anon_sym_getprotobyname] = ACTIONS(1764),
    [anon_sym_getpwname] = ACTIONS(1764),
    [anon_sym_getpwuid] = ACTIONS(1764),
    [anon_sym_getpeername] = ACTIONS(1764),
    [anon_sym_getnetbyname] = ACTIONS(1764),
    [anon_sym_getsockname] = ACTIONS(1764),
    [anon_sym_getgrnam] = ACTIONS(1764),
    [anon_sym_getgrgid] = ACTIONS(1764),
    [anon_sym_hex] = ACTIONS(1764),
    [anon_sym_int] = ACTIONS(1764),
    [anon_sym_keys] = ACTIONS(1764),
    [anon_sym_lc] = ACTIONS(1764),
    [anon_sym_lcfirst] = ACTIONS(1764),
    [anon_sym_length] = ACTIONS(1764),
    [anon_sym_localtime] = ACTIONS(1764),
    [anon_sym_log] = ACTIONS(1764),
    [anon_sym_lock] = ACTIONS(1764),
    [anon_sym_lstat] = ACTIONS(1764),
    [anon_sym_oct] = ACTIONS(1764),
    [anon_sym_ord] = ACTIONS(1764),
    [anon_sym_prototype] = ACTIONS(1764),
    [anon_sym_pop] = ACTIONS(1764),
    [anon_sym_pos] = ACTIONS(1764),
    [anon_sym_quotemeta] = ACTIONS(1764),
    [anon_sym_reset] = ACTIONS(1764),
    [anon_sym_rand] = ACTIONS(1764),
    [anon_sym_rmdir] = ACTIONS(1764),
    [anon_sym_readdir] = ACTIONS(1764),
    [anon_sym_readline] = ACTIONS(1764),
    [anon_sym_readpipe] = ACTIONS(1764),
    [anon_sym_rewinddir] = ACTIONS(1764),
    [anon_sym_readlink] = ACTIONS(1764),
    [anon_sym_ref] = ACTIONS(1764),
    [anon_sym_scalar] = ACTIONS(1764),
    [anon_sym_shift] = ACTIONS(1764),
    [anon_sym_sin] = ACTIONS(1764),
    [anon_sym_sleep] = ACTIONS(1764),
    [anon_sym_sqrt] = ACTIONS(1764),
    [anon_sym_srand] = ACTIONS(1764),
    [anon_sym_stat] = ACTIONS(1764),
    [anon_sym_study] = ACTIONS(1764),
    [anon_sym_tell] = ACTIONS(1764),
    [anon_sym_telldir] = ACTIONS(1764),
    [anon_sym_tied] = ACTIONS(1764),
    [anon_sym_uc] = ACTIONS(1764),
    [anon_sym_ucfirst] = ACTIONS(1764),
    [anon_sym_untie] = ACTIONS(1764),
    [anon_sym_umask] = ACTIONS(1764),
    [anon_sym_values] = ACTIONS(1764),
    [anon_sym_write] = ACTIONS(1764),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1766),
  },
  [STATE(964)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3773),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(964),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(965)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2728),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(965),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(966)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3795),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(966),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(967)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3302),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(967),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(968)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2729),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(968),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(969)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2313),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(969),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(970)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2730),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(970),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(971)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2327),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(971),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(972)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2731),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(972),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(973)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3303),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(973),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(974)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2732),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(974),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(975)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3586),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(975),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(976)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2733),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(976),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(977)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3450),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(977),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(978)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3454),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(978),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(979)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3462),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(979),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(980)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3466),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(980),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(981)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3467),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(981),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(982)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3297),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(982),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(983)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3471),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(983),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(984)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2735),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(984),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(985)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2736),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(985),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(986)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3474),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(986),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(987)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2737),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(987),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(988)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3483),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(988),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(989)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3490),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(989),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(990)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3492),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(990),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(991)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3713),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(991),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(992)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3713),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(992),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2648),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2648),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2650),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(993)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3448),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(993),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(994)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3714),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3282),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(994),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(995)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3306),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(995),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(996)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3588),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(996),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(997)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3717),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(997),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(998)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3718),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(411),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(998),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_PLUS_PLUS] = ACTIONS(2453),
    [anon_sym_DASH_DASH] = ACTIONS(2453),
    [anon_sym_BSLASH] = ACTIONS(1657),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_eval] = ACTIONS(1661),
    [anon_sym_await] = ACTIONS(1663),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1665),
    [anon_sym_dynamically] = ACTIONS(1665),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1667),
    [anon_sym_map] = ACTIONS(1669),
    [anon_sym_grep] = ACTIONS(1669),
    [anon_sym_sort] = ACTIONS(1671),
    [anon_sym_goto] = ACTIONS(1673),
    [anon_sym_return] = ACTIONS(1675),
    [anon_sym_undef] = ACTIONS(1677),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1679),
    [anon_sym_next] = ACTIONS(1679),
    [anon_sym_redo] = ACTIONS(1679),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1681),
    [anon_sym_alarm] = ACTIONS(1681),
    [anon_sym_chop] = ACTIONS(1681),
    [anon_sym_chdir] = ACTIONS(1681),
    [anon_sym_close] = ACTIONS(1681),
    [anon_sym_closedir] = ACTIONS(1681),
    [anon_sym_caller] = ACTIONS(1681),
    [anon_sym_chomp] = ACTIONS(1681),
    [anon_sym_chr] = ACTIONS(1681),
    [anon_sym_cos] = ACTIONS(1681),
    [anon_sym_chroot] = ACTIONS(1681),
    [anon_sym_defined] = ACTIONS(1681),
    [anon_sym_delete] = ACTIONS(1681),
    [anon_sym_dbmclose] = ACTIONS(1681),
    [anon_sym_exists] = ACTIONS(1681),
    [anon_sym_exit] = ACTIONS(1681),
    [anon_sym_eof] = ACTIONS(1681),
    [anon_sym_exp] = ACTIONS(1681),
    [anon_sym_each] = ACTIONS(1681),
    [anon_sym_fc] = ACTIONS(1681),
    [anon_sym_fileno] = ACTIONS(1681),
    [anon_sym_gmtime] = ACTIONS(1681),
    [anon_sym_getc] = ACTIONS(1681),
    [anon_sym_getpgrp] = ACTIONS(1681),
    [anon_sym_getprotobyname] = ACTIONS(1681),
    [anon_sym_getpwname] = ACTIONS(1681),
    [anon_sym_getpwuid] = ACTIONS(1681),
    [anon_sym_getpeername] = ACTIONS(1681),
    [anon_sym_getnetbyname] = ACTIONS(1681),
    [anon_sym_getsockname] = ACTIONS(1681),
    [anon_sym_getgrnam] = ACTIONS(1681),
    [anon_sym_getgrgid] = ACTIONS(1681),
    [anon_sym_hex] = ACTIONS(1681),
    [anon_sym_int] = ACTIONS(1681),
    [anon_sym_keys] = ACTIONS(1681),
    [anon_sym_lc] = ACTIONS(1681),
    [anon_sym_lcfirst] = ACTIONS(1681),
    [anon_sym_length] = ACTIONS(1681),
    [anon_sym_localtime] = ACTIONS(1681),
    [anon_sym_log] = ACTIONS(1681),
    [anon_sym_lock] = ACTIONS(1681),
    [anon_sym_lstat] = ACTIONS(1681),
    [anon_sym_oct] = ACTIONS(1681),
    [anon_sym_ord] = ACTIONS(1681),
    [anon_sym_prototype] = ACTIONS(1681),
    [anon_sym_pop] = ACTIONS(1681),
    [anon_sym_pos] = ACTIONS(1681),
    [anon_sym_quotemeta] = ACTIONS(1681),
    [anon_sym_reset] = ACTIONS(1681),
    [anon_sym_rand] = ACTIONS(1681),
    [anon_sym_rmdir] = ACTIONS(1681),
    [anon_sym_readdir] = ACTIONS(1681),
    [anon_sym_readline] = ACTIONS(1681),
    [anon_sym_readpipe] = ACTIONS(1681),
    [anon_sym_rewinddir] = ACTIONS(1681),
    [anon_sym_readlink] = ACTIONS(1681),
    [anon_sym_ref] = ACTIONS(1681),
    [anon_sym_scalar] = ACTIONS(1681),
    [anon_sym_shift] = ACTIONS(1681),
    [anon_sym_sin] = ACTIONS(1681),
    [anon_sym_sleep] = ACTIONS(1681),
    [anon_sym_sqrt] = ACTIONS(1681),
    [anon_sym_srand] = ACTIONS(1681),
    [anon_sym_stat] = ACTIONS(1681),
    [anon_sym_study] = ACTIONS(1681),
    [anon_sym_tell] = ACTIONS(1681),
    [anon_sym_telldir] = ACTIONS(1681),
    [anon_sym_tied] = ACTIONS(1681),
    [anon_sym_uc] = ACTIONS(1681),
    [anon_sym_ucfirst] = ACTIONS(1681),
    [anon_sym_untie] = ACTIONS(1681),
    [anon_sym_umask] = ACTIONS(1681),
    [anon_sym_values] = ACTIONS(1681),
    [anon_sym_write] = ACTIONS(1681),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1683),
  },
  [STATE(999)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3493),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(999),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1000)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3495),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1000),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1001)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3496),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1001),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1002)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3498),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1002),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1003)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3499),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1003),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1004)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3589),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1004),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1005)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1903),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1005),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1006)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3590),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1006),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1007)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3269),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1007),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1008)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3776),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1008),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1009)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3591),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1009),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1010)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3592),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1010),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1011)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3593),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1011),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1012)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3594),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1012),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1013)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2797),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(1013),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(1014)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3595),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1014),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1015)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3505),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(406),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1015),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_PLUS_PLUS] = ACTIONS(2550),
    [anon_sym_DASH_DASH] = ACTIONS(2550),
    [anon_sym_BSLASH] = ACTIONS(1770),
    [anon_sym_do] = ACTIONS(1772),
    [anon_sym_eval] = ACTIONS(1774),
    [anon_sym_await] = ACTIONS(1776),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1778),
    [anon_sym_dynamically] = ACTIONS(1778),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1780),
    [anon_sym_map] = ACTIONS(1782),
    [anon_sym_grep] = ACTIONS(1782),
    [anon_sym_sort] = ACTIONS(1784),
    [anon_sym_goto] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1788),
    [anon_sym_undef] = ACTIONS(1790),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1792),
    [anon_sym_next] = ACTIONS(1792),
    [anon_sym_redo] = ACTIONS(1792),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1794),
    [anon_sym_alarm] = ACTIONS(1794),
    [anon_sym_chop] = ACTIONS(1794),
    [anon_sym_chdir] = ACTIONS(1794),
    [anon_sym_close] = ACTIONS(1794),
    [anon_sym_closedir] = ACTIONS(1794),
    [anon_sym_caller] = ACTIONS(1794),
    [anon_sym_chomp] = ACTIONS(1794),
    [anon_sym_chr] = ACTIONS(1794),
    [anon_sym_cos] = ACTIONS(1794),
    [anon_sym_chroot] = ACTIONS(1794),
    [anon_sym_defined] = ACTIONS(1794),
    [anon_sym_delete] = ACTIONS(1794),
    [anon_sym_dbmclose] = ACTIONS(1794),
    [anon_sym_exists] = ACTIONS(1794),
    [anon_sym_exit] = ACTIONS(1794),
    [anon_sym_eof] = ACTIONS(1794),
    [anon_sym_exp] = ACTIONS(1794),
    [anon_sym_each] = ACTIONS(1794),
    [anon_sym_fc] = ACTIONS(1794),
    [anon_sym_fileno] = ACTIONS(1794),
    [anon_sym_gmtime] = ACTIONS(1794),
    [anon_sym_getc] = ACTIONS(1794),
    [anon_sym_getpgrp] = ACTIONS(1794),
    [anon_sym_getprotobyname] = ACTIONS(1794),
    [anon_sym_getpwname] = ACTIONS(1794),
    [anon_sym_getpwuid] = ACTIONS(1794),
    [anon_sym_getpeername] = ACTIONS(1794),
    [anon_sym_getnetbyname] = ACTIONS(1794),
    [anon_sym_getsockname] = ACTIONS(1794),
    [anon_sym_getgrnam] = ACTIONS(1794),
    [anon_sym_getgrgid] = ACTIONS(1794),
    [anon_sym_hex] = ACTIONS(1794),
    [anon_sym_int] = ACTIONS(1794),
    [anon_sym_keys] = ACTIONS(1794),
    [anon_sym_lc] = ACTIONS(1794),
    [anon_sym_lcfirst] = ACTIONS(1794),
    [anon_sym_length] = ACTIONS(1794),
    [anon_sym_localtime] = ACTIONS(1794),
    [anon_sym_log] = ACTIONS(1794),
    [anon_sym_lock] = ACTIONS(1794),
    [anon_sym_lstat] = ACTIONS(1794),
    [anon_sym_oct] = ACTIONS(1794),
    [anon_sym_ord] = ACTIONS(1794),
    [anon_sym_prototype] = ACTIONS(1794),
    [anon_sym_pop] = ACTIONS(1794),
    [anon_sym_pos] = ACTIONS(1794),
    [anon_sym_quotemeta] = ACTIONS(1794),
    [anon_sym_reset] = ACTIONS(1794),
    [anon_sym_rand] = ACTIONS(1794),
    [anon_sym_rmdir] = ACTIONS(1794),
    [anon_sym_readdir] = ACTIONS(1794),
    [anon_sym_readline] = ACTIONS(1794),
    [anon_sym_readpipe] = ACTIONS(1794),
    [anon_sym_rewinddir] = ACTIONS(1794),
    [anon_sym_readlink] = ACTIONS(1794),
    [anon_sym_ref] = ACTIONS(1794),
    [anon_sym_scalar] = ACTIONS(1794),
    [anon_sym_shift] = ACTIONS(1794),
    [anon_sym_sin] = ACTIONS(1794),
    [anon_sym_sleep] = ACTIONS(1794),
    [anon_sym_sqrt] = ACTIONS(1794),
    [anon_sym_srand] = ACTIONS(1794),
    [anon_sym_stat] = ACTIONS(1794),
    [anon_sym_study] = ACTIONS(1794),
    [anon_sym_tell] = ACTIONS(1794),
    [anon_sym_telldir] = ACTIONS(1794),
    [anon_sym_tied] = ACTIONS(1794),
    [anon_sym_uc] = ACTIONS(1794),
    [anon_sym_ucfirst] = ACTIONS(1794),
    [anon_sym_untie] = ACTIONS(1794),
    [anon_sym_umask] = ACTIONS(1794),
    [anon_sym_values] = ACTIONS(1794),
    [anon_sym_write] = ACTIONS(1794),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1796),
  },
  [STATE(1016)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1828),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1016),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1017)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1831),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1017),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1018)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3259),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1018),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1019)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3596),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1019),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1020)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1832),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1020),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1021)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2667),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1021),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1022)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2667),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1022),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(2664),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(2664),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2666),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1023)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3255),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1023),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1024)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3597),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1024),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1025)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2669),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1025),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1026)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3598),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1026),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1027)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2670),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2305),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1027),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1028)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3599),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1028),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1029)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3600),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1029),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1030)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2365),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1030),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1031)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3601),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1031),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1032)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1834),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1032),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1033)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3778),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1033),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1034)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2675),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1034),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1035)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2761),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1035),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1036)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2761),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1036),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(2668),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(2668),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2670),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1037)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2772),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1037),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1038)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2676),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1038),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1039)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2774),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2592),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1039),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1040)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3675),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1040),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1041)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3299),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1041),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1042)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2782),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1042),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1043)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3676),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1043),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1044)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2783),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1044),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1045)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3280),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1045),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1046)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1855),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1046),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1047)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3259),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1047),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(2660),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(2660),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2662),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1048)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1837),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1048),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1049)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1838),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1049),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1050)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1839),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1050),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1051)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1840),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1051),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1052)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3284),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1052),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1053)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1842),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1053),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1054)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2543),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1054),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1055)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2543),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1055),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(2672),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(2672),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2674),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1056)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2546),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1056),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1057)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2954),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1057),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1058)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3781),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1058),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1059)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3677),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1059),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1060)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1843),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1060),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1061)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2958),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1061),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1062)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2966),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1062),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1063)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2968),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1063),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1064)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2969),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1064),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1065)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2971),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1065),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1066)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2972),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1066),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1067)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2973),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1067),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1068)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2981),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1068),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1069)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2900),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1069),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1070)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2986),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1070),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1071)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2989),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1071),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1072)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2902),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1072),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1073)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(2994),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1073),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1074)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3008),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1074),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1075)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2903),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1075),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1076)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3009),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1076),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1077)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2904),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1077),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1078)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3678),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1078),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1079)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2905),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1079),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1080)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3010),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1080),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1081)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2906),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1081),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1082)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2547),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2232),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1082),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1083)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3803),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1083),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1084)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2907),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1084),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1085)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3679),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1085),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1086)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1844),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1086),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1087)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(333),
    [sym_subscripted] = STATE(3113),
    [sym_container_variable] = STATE(5708),
    [sym_glob_slot_expression] = STATE(3114),
    [sym_array_element_expression] = STATE(3114),
    [sym_hash_element_expression] = STATE(3114),
    [sym_coderef_call_expression] = STATE(3114),
    [sym_anonymous_slice_expression] = STATE(3114),
    [sym_slices] = STATE(3534),
    [sym_slice_container_variable] = STATE(5709),
    [sym_slice_expression] = STATE(3535),
    [sym_keyval_container_variable] = STATE(5710),
    [sym_keyval_expression] = STATE(3535),
    [sym__term] = STATE(3644),
    [sym_readline_expression] = STATE(3534),
    [sym_fileglob_expression] = STATE(3534),
    [sym_assignment_expression] = STATE(3534),
    [sym_binary_expression] = STATE(3534),
    [sym_equality_expression] = STATE(3534),
    [sym_relational_expression] = STATE(3534),
    [sym_unary_expression] = STATE(3534),
    [sym_preinc_expression] = STATE(3534),
    [sym_postinc_expression] = STATE(3534),
    [sym_conditional_expression] = STATE(3534),
    [sym_refgen_expression] = STATE(3534),
    [sym_anonymous_array_expression] = STATE(3534),
    [sym_anonymous_hash_expression] = STATE(3534),
    [sym_anonymous_subroutine_expression] = STATE(3534),
    [sym_anonymous_method_expression] = STATE(3534),
    [sym_do_expression] = STATE(3534),
    [sym_eval_expression] = STATE(3534),
    [sym_await_expression] = STATE(3534),
    [sym_variable_declaration] = STATE(3534),
    [sym_localization_expression] = STATE(3534),
    [sym_stub_expression] = STATE(3534),
    [sym_postfix_deref] = STATE(3534),
    [sym_scalar_deref_expression] = STATE(3536),
    [sym_array_deref_expression] = STATE(3536),
    [sym_arraylen_deref_expression] = STATE(3536),
    [sym_hash_deref_expression] = STATE(3536),
    [sym_amper_deref_expression] = STATE(3536),
    [sym_glob_deref_expression] = STATE(3536),
    [sym_require_expression] = STATE(3534),
    [sym_require_version_expression] = STATE(3534),
    [sym_func0op_call_expression] = STATE(3534),
    [sym_func1op_call_expression] = STATE(3534),
    [sym_map_grep_expression] = STATE(3534),
    [sym_sort_expression] = STATE(3534),
    [sym_loopex_expression] = STATE(3534),
    [sym_goto_expression] = STATE(3534),
    [sym_return_expression] = STATE(3534),
    [sym_undef_expression] = STATE(3534),
    [sym__listop] = STATE(3534),
    [sym_function_call_expression] = STATE(3537),
    [sym_ambiguous_function_call_expression] = STATE(3537),
    [sym_function] = STATE(410),
    [sym_method_call_expression] = STATE(3537),
    [sym__variables] = STATE(3534),
    [sym_scalar] = STATE(3538),
    [sym_array] = STATE(3538),
    [sym__SUB_AMPER] = STATE(4249),
    [sym__GLOB_STAR] = STATE(4250),
    [sym_hash] = STATE(3538),
    [sym_arraylen] = STATE(3538),
    [sym_glob] = STATE(3340),
    [sym_amper_sub] = STATE(3341),
    [sym__literal] = STATE(3534),
    [sym__apostrophe] = STATE(4751),
    [sym__quotation_mark] = STATE(3885),
    [sym__backtick] = STATE(3887),
    [sym__search_slash] = STATE(3831),
    [sym_string_literal] = STATE(3539),
    [sym_interpolated_string_literal] = STATE(3539),
    [sym_quoted_word_list] = STATE(3342),
    [sym_command_string] = STATE(3539),
    [sym_quoted_regexp] = STATE(3539),
    [sym_match_regexp] = STATE(3539),
    [sym_substitution_regexp] = STATE(3539),
    [sym_transliteration_expression] = STATE(3539),
    [sym_heredoc_token] = STATE(3534),
    [sym_command_heredoc_token] = STATE(3534),
    [sym_heredoc_content] = STATE(1087),
    [sym_autoquoted_bareword] = STATE(3534),
    [sym_bareword] = STATE(3534),
    [sym_primitive] = STATE(3534),
    [sym_boolean] = STATE(3530),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5102),
    [sym__identifier] = ACTIONS(787),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(723),
    [anon_sym_AT] = ACTIONS(725),
    [anon_sym_my] = ACTIONS(727),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(729),
    [anon_sym_method] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_continue] = ACTIONS(735),
    [anon_sym_state] = ACTIONS(727),
    [anon_sym_our] = ACTIONS(727),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_LT_LT] = ACTIONS(2159),
    [anon_sym_SLASH_SLASH] = ACTIONS(2161),
    [anon_sym_PLUS] = ACTIONS(739),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_BANG] = ACTIONS(739),
    [anon_sym_PLUS_PLUS] = ACTIONS(2165),
    [anon_sym_DASH_DASH] = ACTIONS(2165),
    [anon_sym_BSLASH] = ACTIONS(741),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_eval] = ACTIONS(745),
    [anon_sym_await] = ACTIONS(747),
    [anon_sym_field] = ACTIONS(727),
    [anon_sym_local] = ACTIONS(749),
    [anon_sym_dynamically] = ACTIONS(749),
    [anon_sym_DOLLAR_POUND] = ACTIONS(751),
    [anon_sym_require] = ACTIONS(753),
    [anon_sym_map] = ACTIONS(755),
    [anon_sym_grep] = ACTIONS(755),
    [anon_sym_sort] = ACTIONS(757),
    [anon_sym_goto] = ACTIONS(759),
    [anon_sym_return] = ACTIONS(761),
    [anon_sym_undef] = ACTIONS(763),
    [anon_sym_PERCENT2] = ACTIONS(765),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(767),
    [anon_sym_next] = ACTIONS(767),
    [anon_sym_redo] = ACTIONS(767),
    [anon_sym___FILE__] = ACTIONS(735),
    [anon_sym___LINE__] = ACTIONS(735),
    [anon_sym___PACKAGE__] = ACTIONS(735),
    [anon_sym___SUB__] = ACTIONS(735),
    [anon_sym_break] = ACTIONS(735),
    [anon_sym_fork] = ACTIONS(735),
    [anon_sym_getppid] = ACTIONS(735),
    [anon_sym_time] = ACTIONS(735),
    [anon_sym_times] = ACTIONS(735),
    [anon_sym_wait] = ACTIONS(735),
    [anon_sym_wantarray] = ACTIONS(735),
    [anon_sym_abs] = ACTIONS(769),
    [anon_sym_alarm] = ACTIONS(769),
    [anon_sym_chop] = ACTIONS(769),
    [anon_sym_chdir] = ACTIONS(769),
    [anon_sym_close] = ACTIONS(769),
    [anon_sym_closedir] = ACTIONS(769),
    [anon_sym_caller] = ACTIONS(769),
    [anon_sym_chomp] = ACTIONS(769),
    [anon_sym_chr] = ACTIONS(769),
    [anon_sym_cos] = ACTIONS(769),
    [anon_sym_chroot] = ACTIONS(769),
    [anon_sym_defined] = ACTIONS(769),
    [anon_sym_delete] = ACTIONS(769),
    [anon_sym_dbmclose] = ACTIONS(769),
    [anon_sym_exists] = ACTIONS(769),
    [anon_sym_exit] = ACTIONS(769),
    [anon_sym_eof] = ACTIONS(769),
    [anon_sym_exp] = ACTIONS(769),
    [anon_sym_each] = ACTIONS(769),
    [anon_sym_fc] = ACTIONS(769),
    [anon_sym_fileno] = ACTIONS(769),
    [anon_sym_gmtime] = ACTIONS(769),
    [anon_sym_getc] = ACTIONS(769),
    [anon_sym_getpgrp] = ACTIONS(769),
    [anon_sym_getprotobyname] = ACTIONS(769),
    [anon_sym_getpwname] = ACTIONS(769),
    [anon_sym_getpwuid] = ACTIONS(769),
    [anon_sym_getpeername] = ACTIONS(769),
    [anon_sym_getnetbyname] = ACTIONS(769),
    [anon_sym_getsockname] = ACTIONS(769),
    [anon_sym_getgrnam] = ACTIONS(769),
    [anon_sym_getgrgid] = ACTIONS(769),
    [anon_sym_hex] = ACTIONS(769),
    [anon_sym_int] = ACTIONS(769),
    [anon_sym_keys] = ACTIONS(769),
    [anon_sym_lc] = ACTIONS(769),
    [anon_sym_lcfirst] = ACTIONS(769),
    [anon_sym_length] = ACTIONS(769),
    [anon_sym_localtime] = ACTIONS(769),
    [anon_sym_log] = ACTIONS(769),
    [anon_sym_lock] = ACTIONS(769),
    [anon_sym_lstat] = ACTIONS(769),
    [anon_sym_oct] = ACTIONS(769),
    [anon_sym_ord] = ACTIONS(769),
    [anon_sym_prototype] = ACTIONS(769),
    [anon_sym_pop] = ACTIONS(769),
    [anon_sym_pos] = ACTIONS(769),
    [anon_sym_quotemeta] = ACTIONS(769),
    [anon_sym_reset] = ACTIONS(769),
    [anon_sym_rand] = ACTIONS(769),
    [anon_sym_rmdir] = ACTIONS(769),
    [anon_sym_readdir] = ACTIONS(769),
    [anon_sym_readline] = ACTIONS(769),
    [anon_sym_readpipe] = ACTIONS(769),
    [anon_sym_rewinddir] = ACTIONS(769),
    [anon_sym_readlink] = ACTIONS(769),
    [anon_sym_ref] = ACTIONS(769),
    [anon_sym_scalar] = ACTIONS(769),
    [anon_sym_shift] = ACTIONS(769),
    [anon_sym_sin] = ACTIONS(769),
    [anon_sym_sleep] = ACTIONS(769),
    [anon_sym_sqrt] = ACTIONS(769),
    [anon_sym_srand] = ACTIONS(769),
    [anon_sym_stat] = ACTIONS(769),
    [anon_sym_study] = ACTIONS(769),
    [anon_sym_tell] = ACTIONS(769),
    [anon_sym_telldir] = ACTIONS(769),
    [anon_sym_tied] = ACTIONS(769),
    [anon_sym_uc] = ACTIONS(769),
    [anon_sym_ucfirst] = ACTIONS(769),
    [anon_sym_untie] = ACTIONS(769),
    [anon_sym_umask] = ACTIONS(769),
    [anon_sym_values] = ACTIONS(769),
    [anon_sym_write] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(771),
    [anon_sym_qq] = ACTIONS(773),
    [anon_sym_qw] = ACTIONS(775),
    [anon_sym_qx] = ACTIONS(777),
    [anon_sym_qr] = ACTIONS(779),
    [anon_sym_m] = ACTIONS(781),
    [anon_sym_s] = ACTIONS(783),
    [anon_sym_tr] = ACTIONS(785),
    [anon_sym_y] = ACTIONS(785),
    [aux_sym__bareword_token1] = ACTIONS(787),
    [sym_number] = ACTIONS(789),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(793),
    [sym__open_fileglob_bracket] = ACTIONS(795),
    [sym__PERLY_HEREDOC] = ACTIONS(797),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(799),
    [sym__filetest] = ACTIONS(801),
  },
  [STATE(1088)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2908),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1088),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1089)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3798),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1089),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1090)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2065),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1090),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1091)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2065),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1091),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(2676),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(2676),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2678),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1092)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(360),
    [sym_subscripted] = STATE(2080),
    [sym_container_variable] = STATE(5694),
    [sym_glob_slot_expression] = STATE(2081),
    [sym_array_element_expression] = STATE(2081),
    [sym_hash_element_expression] = STATE(2081),
    [sym_coderef_call_expression] = STATE(2081),
    [sym_anonymous_slice_expression] = STATE(2081),
    [sym_slices] = STATE(2693),
    [sym_slice_container_variable] = STATE(5695),
    [sym_slice_expression] = STATE(2694),
    [sym_keyval_container_variable] = STATE(5696),
    [sym_keyval_expression] = STATE(2694),
    [sym__term] = STATE(3057),
    [sym_readline_expression] = STATE(2693),
    [sym_fileglob_expression] = STATE(2693),
    [sym_assignment_expression] = STATE(2693),
    [sym_binary_expression] = STATE(2693),
    [sym_equality_expression] = STATE(2693),
    [sym_relational_expression] = STATE(2693),
    [sym_unary_expression] = STATE(2693),
    [sym_preinc_expression] = STATE(2693),
    [sym_postinc_expression] = STATE(2693),
    [sym_conditional_expression] = STATE(2693),
    [sym_refgen_expression] = STATE(2693),
    [sym_anonymous_array_expression] = STATE(2693),
    [sym_anonymous_hash_expression] = STATE(2693),
    [sym_anonymous_subroutine_expression] = STATE(2693),
    [sym_anonymous_method_expression] = STATE(2693),
    [sym_do_expression] = STATE(2693),
    [sym_eval_expression] = STATE(2693),
    [sym_await_expression] = STATE(2693),
    [sym_variable_declaration] = STATE(2693),
    [sym_localization_expression] = STATE(2693),
    [sym_stub_expression] = STATE(2693),
    [sym_postfix_deref] = STATE(2693),
    [sym_scalar_deref_expression] = STATE(2696),
    [sym_array_deref_expression] = STATE(2696),
    [sym_arraylen_deref_expression] = STATE(2696),
    [sym_hash_deref_expression] = STATE(2696),
    [sym_amper_deref_expression] = STATE(2696),
    [sym_glob_deref_expression] = STATE(2696),
    [sym_require_expression] = STATE(2693),
    [sym_require_version_expression] = STATE(2693),
    [sym_func0op_call_expression] = STATE(2693),
    [sym_func1op_call_expression] = STATE(2693),
    [sym_map_grep_expression] = STATE(2693),
    [sym_sort_expression] = STATE(2693),
    [sym_loopex_expression] = STATE(2693),
    [sym_goto_expression] = STATE(2693),
    [sym_return_expression] = STATE(2693),
    [sym_undef_expression] = STATE(2693),
    [sym__listop] = STATE(2693),
    [sym_function_call_expression] = STATE(2698),
    [sym_ambiguous_function_call_expression] = STATE(2698),
    [sym_function] = STATE(397),
    [sym_method_call_expression] = STATE(2698),
    [sym__variables] = STATE(2693),
    [sym_scalar] = STATE(2700),
    [sym_array] = STATE(2700),
    [sym__SUB_AMPER] = STATE(4218),
    [sym__GLOB_STAR] = STATE(4220),
    [sym_hash] = STATE(2700),
    [sym_arraylen] = STATE(2700),
    [sym_glob] = STATE(2548),
    [sym_amper_sub] = STATE(2570),
    [sym__literal] = STATE(2693),
    [sym__apostrophe] = STATE(4756),
    [sym__quotation_mark] = STATE(3941),
    [sym__backtick] = STATE(3943),
    [sym__search_slash] = STATE(3861),
    [sym_string_literal] = STATE(2704),
    [sym_interpolated_string_literal] = STATE(2704),
    [sym_quoted_word_list] = STATE(2572),
    [sym_command_string] = STATE(2704),
    [sym_quoted_regexp] = STATE(2704),
    [sym_match_regexp] = STATE(2704),
    [sym_substitution_regexp] = STATE(2704),
    [sym_transliteration_expression] = STATE(2704),
    [sym_heredoc_token] = STATE(2693),
    [sym_command_heredoc_token] = STATE(2693),
    [sym_heredoc_content] = STATE(1092),
    [sym_autoquoted_bareword] = STATE(2693),
    [sym_bareword] = STATE(2693),
    [sym_primitive] = STATE(2693),
    [sym_boolean] = STATE(2690),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5072),
    [sym__identifier] = ACTIONS(552),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(554),
    [anon_sym_AT] = ACTIONS(556),
    [anon_sym_my] = ACTIONS(558),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(560),
    [anon_sym_method] = ACTIONS(562),
    [anon_sym_LPAREN] = ACTIONS(564),
    [anon_sym_continue] = ACTIONS(566),
    [anon_sym_state] = ACTIONS(558),
    [anon_sym_our] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(568),
    [anon_sym_LT_LT] = ACTIONS(2147),
    [anon_sym_SLASH_SLASH] = ACTIONS(2149),
    [anon_sym_PLUS] = ACTIONS(582),
    [anon_sym_DASH] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(582),
    [anon_sym_BANG] = ACTIONS(582),
    [anon_sym_PLUS_PLUS] = ACTIONS(2153),
    [anon_sym_DASH_DASH] = ACTIONS(2153),
    [anon_sym_BSLASH] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(589),
    [anon_sym_eval] = ACTIONS(591),
    [anon_sym_await] = ACTIONS(593),
    [anon_sym_field] = ACTIONS(558),
    [anon_sym_local] = ACTIONS(595),
    [anon_sym_dynamically] = ACTIONS(595),
    [anon_sym_DOLLAR_POUND] = ACTIONS(597),
    [anon_sym_require] = ACTIONS(599),
    [anon_sym_map] = ACTIONS(601),
    [anon_sym_grep] = ACTIONS(601),
    [anon_sym_sort] = ACTIONS(603),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym_return] = ACTIONS(607),
    [anon_sym_undef] = ACTIONS(609),
    [anon_sym_PERCENT2] = ACTIONS(611),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(613),
    [anon_sym_next] = ACTIONS(613),
    [anon_sym_redo] = ACTIONS(613),
    [anon_sym___FILE__] = ACTIONS(566),
    [anon_sym___LINE__] = ACTIONS(566),
    [anon_sym___PACKAGE__] = ACTIONS(566),
    [anon_sym___SUB__] = ACTIONS(566),
    [anon_sym_break] = ACTIONS(566),
    [anon_sym_fork] = ACTIONS(566),
    [anon_sym_getppid] = ACTIONS(566),
    [anon_sym_time] = ACTIONS(566),
    [anon_sym_times] = ACTIONS(566),
    [anon_sym_wait] = ACTIONS(566),
    [anon_sym_wantarray] = ACTIONS(566),
    [anon_sym_abs] = ACTIONS(615),
    [anon_sym_alarm] = ACTIONS(615),
    [anon_sym_chop] = ACTIONS(615),
    [anon_sym_chdir] = ACTIONS(615),
    [anon_sym_close] = ACTIONS(615),
    [anon_sym_closedir] = ACTIONS(615),
    [anon_sym_caller] = ACTIONS(615),
    [anon_sym_chomp] = ACTIONS(615),
    [anon_sym_chr] = ACTIONS(615),
    [anon_sym_cos] = ACTIONS(615),
    [anon_sym_chroot] = ACTIONS(615),
    [anon_sym_defined] = ACTIONS(615),
    [anon_sym_delete] = ACTIONS(615),
    [anon_sym_dbmclose] = ACTIONS(615),
    [anon_sym_exists] = ACTIONS(615),
    [anon_sym_exit] = ACTIONS(615),
    [anon_sym_eof] = ACTIONS(615),
    [anon_sym_exp] = ACTIONS(615),
    [anon_sym_each] = ACTIONS(615),
    [anon_sym_fc] = ACTIONS(615),
    [anon_sym_fileno] = ACTIONS(615),
    [anon_sym_gmtime] = ACTIONS(615),
    [anon_sym_getc] = ACTIONS(615),
    [anon_sym_getpgrp] = ACTIONS(615),
    [anon_sym_getprotobyname] = ACTIONS(615),
    [anon_sym_getpwname] = ACTIONS(615),
    [anon_sym_getpwuid] = ACTIONS(615),
    [anon_sym_getpeername] = ACTIONS(615),
    [anon_sym_getnetbyname] = ACTIONS(615),
    [anon_sym_getsockname] = ACTIONS(615),
    [anon_sym_getgrnam] = ACTIONS(615),
    [anon_sym_getgrgid] = ACTIONS(615),
    [anon_sym_hex] = ACTIONS(615),
    [anon_sym_int] = ACTIONS(615),
    [anon_sym_keys] = ACTIONS(615),
    [anon_sym_lc] = ACTIONS(615),
    [anon_sym_lcfirst] = ACTIONS(615),
    [anon_sym_length] = ACTIONS(615),
    [anon_sym_localtime] = ACTIONS(615),
    [anon_sym_log] = ACTIONS(615),
    [anon_sym_lock] = ACTIONS(615),
    [anon_sym_lstat] = ACTIONS(615),
    [anon_sym_oct] = ACTIONS(615),
    [anon_sym_ord] = ACTIONS(615),
    [anon_sym_prototype] = ACTIONS(615),
    [anon_sym_pop] = ACTIONS(615),
    [anon_sym_pos] = ACTIONS(615),
    [anon_sym_quotemeta] = ACTIONS(615),
    [anon_sym_reset] = ACTIONS(615),
    [anon_sym_rand] = ACTIONS(615),
    [anon_sym_rmdir] = ACTIONS(615),
    [anon_sym_readdir] = ACTIONS(615),
    [anon_sym_readline] = ACTIONS(615),
    [anon_sym_readpipe] = ACTIONS(615),
    [anon_sym_rewinddir] = ACTIONS(615),
    [anon_sym_readlink] = ACTIONS(615),
    [anon_sym_ref] = ACTIONS(615),
    [anon_sym_scalar] = ACTIONS(615),
    [anon_sym_shift] = ACTIONS(615),
    [anon_sym_sin] = ACTIONS(615),
    [anon_sym_sleep] = ACTIONS(615),
    [anon_sym_sqrt] = ACTIONS(615),
    [anon_sym_srand] = ACTIONS(615),
    [anon_sym_stat] = ACTIONS(615),
    [anon_sym_study] = ACTIONS(615),
    [anon_sym_tell] = ACTIONS(615),
    [anon_sym_telldir] = ACTIONS(615),
    [anon_sym_tied] = ACTIONS(615),
    [anon_sym_uc] = ACTIONS(615),
    [anon_sym_ucfirst] = ACTIONS(615),
    [anon_sym_untie] = ACTIONS(615),
    [anon_sym_umask] = ACTIONS(615),
    [anon_sym_values] = ACTIONS(615),
    [anon_sym_write] = ACTIONS(615),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(617),
    [anon_sym_qq] = ACTIONS(619),
    [anon_sym_qw] = ACTIONS(621),
    [anon_sym_qx] = ACTIONS(623),
    [anon_sym_qr] = ACTIONS(625),
    [anon_sym_m] = ACTIONS(627),
    [anon_sym_s] = ACTIONS(629),
    [anon_sym_tr] = ACTIONS(631),
    [anon_sym_y] = ACTIONS(631),
    [aux_sym__bareword_token1] = ACTIONS(552),
    [sym_number] = ACTIONS(633),
    [anon_sym_true] = ACTIONS(635),
    [anon_sym_false] = ACTIONS(635),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(637),
    [sym__open_fileglob_bracket] = ACTIONS(639),
    [sym__PERLY_HEREDOC] = ACTIONS(641),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(643),
    [sym__filetest] = ACTIONS(645),
  },
  [STATE(1093)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1845),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1093),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1094)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2909),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1094),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1095)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3517),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1095),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1096)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2910),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1096),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1097)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2911),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1097),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1098)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3517),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1098),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(2648),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(2648),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2650),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1099)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3448),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1099),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1100)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3774),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1100),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1101)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3806),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1101),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1102)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3784),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1102),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1103)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3808),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1103),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1104)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3788),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1104),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1105)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3771),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1105),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1106)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2912),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1106),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1107)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3518),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3282),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1107),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1108)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1956),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1108),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1109)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2066),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1819),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1109),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1110)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2913),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1110),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1111)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3300),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1111),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1112)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3680),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1112),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1113)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3521),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1113),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1114)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2914),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1114),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1115)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3522),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(395),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1115),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(826),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(826),
    [anon_sym_our] = ACTIONS(826),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(850),
    [anon_sym_DASH] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(850),
    [anon_sym_BANG] = ACTIONS(850),
    [anon_sym_PLUS_PLUS] = ACTIONS(2435),
    [anon_sym_DASH_DASH] = ACTIONS(2435),
    [anon_sym_BSLASH] = ACTIONS(855),
    [anon_sym_do] = ACTIONS(857),
    [anon_sym_eval] = ACTIONS(859),
    [anon_sym_await] = ACTIONS(861),
    [anon_sym_field] = ACTIONS(826),
    [anon_sym_local] = ACTIONS(863),
    [anon_sym_dynamically] = ACTIONS(863),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(867),
    [anon_sym_map] = ACTIONS(869),
    [anon_sym_grep] = ACTIONS(869),
    [anon_sym_sort] = ACTIONS(871),
    [anon_sym_goto] = ACTIONS(873),
    [anon_sym_return] = ACTIONS(875),
    [anon_sym_undef] = ACTIONS(877),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(881),
    [anon_sym_next] = ACTIONS(881),
    [anon_sym_redo] = ACTIONS(881),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(883),
    [anon_sym_alarm] = ACTIONS(883),
    [anon_sym_chop] = ACTIONS(883),
    [anon_sym_chdir] = ACTIONS(883),
    [anon_sym_close] = ACTIONS(883),
    [anon_sym_closedir] = ACTIONS(883),
    [anon_sym_caller] = ACTIONS(883),
    [anon_sym_chomp] = ACTIONS(883),
    [anon_sym_chr] = ACTIONS(883),
    [anon_sym_cos] = ACTIONS(883),
    [anon_sym_chroot] = ACTIONS(883),
    [anon_sym_defined] = ACTIONS(883),
    [anon_sym_delete] = ACTIONS(883),
    [anon_sym_dbmclose] = ACTIONS(883),
    [anon_sym_exists] = ACTIONS(883),
    [anon_sym_exit] = ACTIONS(883),
    [anon_sym_eof] = ACTIONS(883),
    [anon_sym_exp] = ACTIONS(883),
    [anon_sym_each] = ACTIONS(883),
    [anon_sym_fc] = ACTIONS(883),
    [anon_sym_fileno] = ACTIONS(883),
    [anon_sym_gmtime] = ACTIONS(883),
    [anon_sym_getc] = ACTIONS(883),
    [anon_sym_getpgrp] = ACTIONS(883),
    [anon_sym_getprotobyname] = ACTIONS(883),
    [anon_sym_getpwname] = ACTIONS(883),
    [anon_sym_getpwuid] = ACTIONS(883),
    [anon_sym_getpeername] = ACTIONS(883),
    [anon_sym_getnetbyname] = ACTIONS(883),
    [anon_sym_getsockname] = ACTIONS(883),
    [anon_sym_getgrnam] = ACTIONS(883),
    [anon_sym_getgrgid] = ACTIONS(883),
    [anon_sym_hex] = ACTIONS(883),
    [anon_sym_int] = ACTIONS(883),
    [anon_sym_keys] = ACTIONS(883),
    [anon_sym_lc] = ACTIONS(883),
    [anon_sym_lcfirst] = ACTIONS(883),
    [anon_sym_length] = ACTIONS(883),
    [anon_sym_localtime] = ACTIONS(883),
    [anon_sym_log] = ACTIONS(883),
    [anon_sym_lock] = ACTIONS(883),
    [anon_sym_lstat] = ACTIONS(883),
    [anon_sym_oct] = ACTIONS(883),
    [anon_sym_ord] = ACTIONS(883),
    [anon_sym_prototype] = ACTIONS(883),
    [anon_sym_pop] = ACTIONS(883),
    [anon_sym_pos] = ACTIONS(883),
    [anon_sym_quotemeta] = ACTIONS(883),
    [anon_sym_reset] = ACTIONS(883),
    [anon_sym_rand] = ACTIONS(883),
    [anon_sym_rmdir] = ACTIONS(883),
    [anon_sym_readdir] = ACTIONS(883),
    [anon_sym_readline] = ACTIONS(883),
    [anon_sym_readpipe] = ACTIONS(883),
    [anon_sym_rewinddir] = ACTIONS(883),
    [anon_sym_readlink] = ACTIONS(883),
    [anon_sym_ref] = ACTIONS(883),
    [anon_sym_scalar] = ACTIONS(883),
    [anon_sym_shift] = ACTIONS(883),
    [anon_sym_sin] = ACTIONS(883),
    [anon_sym_sleep] = ACTIONS(883),
    [anon_sym_sqrt] = ACTIONS(883),
    [anon_sym_srand] = ACTIONS(883),
    [anon_sym_stat] = ACTIONS(883),
    [anon_sym_study] = ACTIONS(883),
    [anon_sym_tell] = ACTIONS(883),
    [anon_sym_telldir] = ACTIONS(883),
    [anon_sym_tied] = ACTIONS(883),
    [anon_sym_uc] = ACTIONS(883),
    [anon_sym_ucfirst] = ACTIONS(883),
    [anon_sym_untie] = ACTIONS(883),
    [anon_sym_umask] = ACTIONS(883),
    [anon_sym_values] = ACTIONS(883),
    [anon_sym_write] = ACTIONS(883),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(913),
  },
  [STATE(1116)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2553),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1116),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1117)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2069),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1117),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1118)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(362),
    [sym_subscripted] = STATE(2128),
    [sym_container_variable] = STATE(5717),
    [sym_glob_slot_expression] = STATE(2129),
    [sym_array_element_expression] = STATE(2129),
    [sym_hash_element_expression] = STATE(2129),
    [sym_coderef_call_expression] = STATE(2129),
    [sym_anonymous_slice_expression] = STATE(2129),
    [sym_slices] = STATE(2652),
    [sym_slice_container_variable] = STATE(5718),
    [sym_slice_expression] = STATE(2653),
    [sym_keyval_container_variable] = STATE(5719),
    [sym_keyval_expression] = STATE(2653),
    [sym__term] = STATE(2915),
    [sym_readline_expression] = STATE(2652),
    [sym_fileglob_expression] = STATE(2652),
    [sym_assignment_expression] = STATE(2652),
    [sym_binary_expression] = STATE(2652),
    [sym_equality_expression] = STATE(2652),
    [sym_relational_expression] = STATE(2652),
    [sym_unary_expression] = STATE(2652),
    [sym_preinc_expression] = STATE(2652),
    [sym_postinc_expression] = STATE(2652),
    [sym_conditional_expression] = STATE(2652),
    [sym_refgen_expression] = STATE(2652),
    [sym_anonymous_array_expression] = STATE(2652),
    [sym_anonymous_hash_expression] = STATE(2652),
    [sym_anonymous_subroutine_expression] = STATE(2652),
    [sym_anonymous_method_expression] = STATE(2652),
    [sym_do_expression] = STATE(2652),
    [sym_eval_expression] = STATE(2652),
    [sym_await_expression] = STATE(2652),
    [sym_variable_declaration] = STATE(2652),
    [sym_localization_expression] = STATE(2652),
    [sym_stub_expression] = STATE(2652),
    [sym_postfix_deref] = STATE(2652),
    [sym_scalar_deref_expression] = STATE(2654),
    [sym_array_deref_expression] = STATE(2654),
    [sym_arraylen_deref_expression] = STATE(2654),
    [sym_hash_deref_expression] = STATE(2654),
    [sym_amper_deref_expression] = STATE(2654),
    [sym_glob_deref_expression] = STATE(2654),
    [sym_require_expression] = STATE(2652),
    [sym_require_version_expression] = STATE(2652),
    [sym_func0op_call_expression] = STATE(2652),
    [sym_func1op_call_expression] = STATE(2652),
    [sym_map_grep_expression] = STATE(2652),
    [sym_sort_expression] = STATE(2652),
    [sym_loopex_expression] = STATE(2652),
    [sym_goto_expression] = STATE(2652),
    [sym_return_expression] = STATE(2652),
    [sym_undef_expression] = STATE(2652),
    [sym__listop] = STATE(2652),
    [sym_function_call_expression] = STATE(2655),
    [sym_ambiguous_function_call_expression] = STATE(2655),
    [sym_function] = STATE(404),
    [sym_method_call_expression] = STATE(2655),
    [sym__variables] = STATE(2652),
    [sym_scalar] = STATE(2656),
    [sym_array] = STATE(2656),
    [sym__SUB_AMPER] = STATE(4196),
    [sym__GLOB_STAR] = STATE(4198),
    [sym_hash] = STATE(2656),
    [sym_arraylen] = STATE(2656),
    [sym_glob] = STATE(2455),
    [sym_amper_sub] = STATE(2537),
    [sym__literal] = STATE(2652),
    [sym__apostrophe] = STATE(4947),
    [sym__quotation_mark] = STATE(3905),
    [sym__backtick] = STATE(3906),
    [sym__search_slash] = STATE(3850),
    [sym_string_literal] = STATE(2657),
    [sym_interpolated_string_literal] = STATE(2657),
    [sym_quoted_word_list] = STATE(2597),
    [sym_command_string] = STATE(2657),
    [sym_quoted_regexp] = STATE(2657),
    [sym_match_regexp] = STATE(2657),
    [sym_substitution_regexp] = STATE(2657),
    [sym_transliteration_expression] = STATE(2657),
    [sym_heredoc_token] = STATE(2652),
    [sym_command_heredoc_token] = STATE(2652),
    [sym_heredoc_content] = STATE(1118),
    [sym_autoquoted_bareword] = STATE(2652),
    [sym_bareword] = STATE(2652),
    [sym_primitive] = STATE(2652),
    [sym_boolean] = STATE(2649),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5114),
    [sym__identifier] = ACTIONS(362),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(364),
    [anon_sym_AT] = ACTIONS(366),
    [anon_sym_my] = ACTIONS(368),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(370),
    [anon_sym_method] = ACTIONS(372),
    [anon_sym_LPAREN] = ACTIONS(374),
    [anon_sym_continue] = ACTIONS(376),
    [anon_sym_state] = ACTIONS(368),
    [anon_sym_our] = ACTIONS(368),
    [anon_sym_LBRACK] = ACTIONS(378),
    [anon_sym_LT_LT] = ACTIONS(2091),
    [anon_sym_SLASH_SLASH] = ACTIONS(2093),
    [anon_sym_PLUS] = ACTIONS(392),
    [anon_sym_DASH] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(392),
    [anon_sym_BANG] = ACTIONS(392),
    [anon_sym_PLUS_PLUS] = ACTIONS(2097),
    [anon_sym_DASH_DASH] = ACTIONS(2097),
    [anon_sym_BSLASH] = ACTIONS(397),
    [anon_sym_do] = ACTIONS(399),
    [anon_sym_eval] = ACTIONS(401),
    [anon_sym_await] = ACTIONS(403),
    [anon_sym_field] = ACTIONS(368),
    [anon_sym_local] = ACTIONS(405),
    [anon_sym_dynamically] = ACTIONS(405),
    [anon_sym_DOLLAR_POUND] = ACTIONS(407),
    [anon_sym_require] = ACTIONS(409),
    [anon_sym_map] = ACTIONS(411),
    [anon_sym_grep] = ACTIONS(411),
    [anon_sym_sort] = ACTIONS(413),
    [anon_sym_goto] = ACTIONS(415),
    [anon_sym_return] = ACTIONS(417),
    [anon_sym_undef] = ACTIONS(419),
    [anon_sym_PERCENT2] = ACTIONS(421),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(423),
    [anon_sym_next] = ACTIONS(423),
    [anon_sym_redo] = ACTIONS(423),
    [anon_sym___FILE__] = ACTIONS(376),
    [anon_sym___LINE__] = ACTIONS(376),
    [anon_sym___PACKAGE__] = ACTIONS(376),
    [anon_sym___SUB__] = ACTIONS(376),
    [anon_sym_break] = ACTIONS(376),
    [anon_sym_fork] = ACTIONS(376),
    [anon_sym_getppid] = ACTIONS(376),
    [anon_sym_time] = ACTIONS(376),
    [anon_sym_times] = ACTIONS(376),
    [anon_sym_wait] = ACTIONS(376),
    [anon_sym_wantarray] = ACTIONS(376),
    [anon_sym_abs] = ACTIONS(425),
    [anon_sym_alarm] = ACTIONS(425),
    [anon_sym_chop] = ACTIONS(425),
    [anon_sym_chdir] = ACTIONS(425),
    [anon_sym_close] = ACTIONS(425),
    [anon_sym_closedir] = ACTIONS(425),
    [anon_sym_caller] = ACTIONS(425),
    [anon_sym_chomp] = ACTIONS(425),
    [anon_sym_chr] = ACTIONS(425),
    [anon_sym_cos] = ACTIONS(425),
    [anon_sym_chroot] = ACTIONS(425),
    [anon_sym_defined] = ACTIONS(425),
    [anon_sym_delete] = ACTIONS(425),
    [anon_sym_dbmclose] = ACTIONS(425),
    [anon_sym_exists] = ACTIONS(425),
    [anon_sym_exit] = ACTIONS(425),
    [anon_sym_eof] = ACTIONS(425),
    [anon_sym_exp] = ACTIONS(425),
    [anon_sym_each] = ACTIONS(425),
    [anon_sym_fc] = ACTIONS(425),
    [anon_sym_fileno] = ACTIONS(425),
    [anon_sym_gmtime] = ACTIONS(425),
    [anon_sym_getc] = ACTIONS(425),
    [anon_sym_getpgrp] = ACTIONS(425),
    [anon_sym_getprotobyname] = ACTIONS(425),
    [anon_sym_getpwname] = ACTIONS(425),
    [anon_sym_getpwuid] = ACTIONS(425),
    [anon_sym_getpeername] = ACTIONS(425),
    [anon_sym_getnetbyname] = ACTIONS(425),
    [anon_sym_getsockname] = ACTIONS(425),
    [anon_sym_getgrnam] = ACTIONS(425),
    [anon_sym_getgrgid] = ACTIONS(425),
    [anon_sym_hex] = ACTIONS(425),
    [anon_sym_int] = ACTIONS(425),
    [anon_sym_keys] = ACTIONS(425),
    [anon_sym_lc] = ACTIONS(425),
    [anon_sym_lcfirst] = ACTIONS(425),
    [anon_sym_length] = ACTIONS(425),
    [anon_sym_localtime] = ACTIONS(425),
    [anon_sym_log] = ACTIONS(425),
    [anon_sym_lock] = ACTIONS(425),
    [anon_sym_lstat] = ACTIONS(425),
    [anon_sym_oct] = ACTIONS(425),
    [anon_sym_ord] = ACTIONS(425),
    [anon_sym_prototype] = ACTIONS(425),
    [anon_sym_pop] = ACTIONS(425),
    [anon_sym_pos] = ACTIONS(425),
    [anon_sym_quotemeta] = ACTIONS(425),
    [anon_sym_reset] = ACTIONS(425),
    [anon_sym_rand] = ACTIONS(425),
    [anon_sym_rmdir] = ACTIONS(425),
    [anon_sym_readdir] = ACTIONS(425),
    [anon_sym_readline] = ACTIONS(425),
    [anon_sym_readpipe] = ACTIONS(425),
    [anon_sym_rewinddir] = ACTIONS(425),
    [anon_sym_readlink] = ACTIONS(425),
    [anon_sym_ref] = ACTIONS(425),
    [anon_sym_scalar] = ACTIONS(425),
    [anon_sym_shift] = ACTIONS(425),
    [anon_sym_sin] = ACTIONS(425),
    [anon_sym_sleep] = ACTIONS(425),
    [anon_sym_sqrt] = ACTIONS(425),
    [anon_sym_srand] = ACTIONS(425),
    [anon_sym_stat] = ACTIONS(425),
    [anon_sym_study] = ACTIONS(425),
    [anon_sym_tell] = ACTIONS(425),
    [anon_sym_telldir] = ACTIONS(425),
    [anon_sym_tied] = ACTIONS(425),
    [anon_sym_uc] = ACTIONS(425),
    [anon_sym_ucfirst] = ACTIONS(425),
    [anon_sym_untie] = ACTIONS(425),
    [anon_sym_umask] = ACTIONS(425),
    [anon_sym_values] = ACTIONS(425),
    [anon_sym_write] = ACTIONS(425),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(427),
    [anon_sym_qq] = ACTIONS(429),
    [anon_sym_qw] = ACTIONS(431),
    [anon_sym_qx] = ACTIONS(433),
    [anon_sym_qr] = ACTIONS(435),
    [anon_sym_m] = ACTIONS(437),
    [anon_sym_s] = ACTIONS(439),
    [anon_sym_tr] = ACTIONS(441),
    [anon_sym_y] = ACTIONS(441),
    [aux_sym__bareword_token1] = ACTIONS(362),
    [sym_number] = ACTIONS(443),
    [anon_sym_true] = ACTIONS(445),
    [anon_sym_false] = ACTIONS(445),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(447),
    [sym__open_fileglob_bracket] = ACTIONS(449),
    [sym__PERLY_HEREDOC] = ACTIONS(451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(453),
    [sym__filetest] = ACTIONS(455),
  },
  [STATE(1119)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3295),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1119),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1120)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3296),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1120),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1121)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(363),
    [sym_subscripted] = STATE(2444),
    [sym_container_variable] = STATE(5703),
    [sym_glob_slot_expression] = STATE(2445),
    [sym_array_element_expression] = STATE(2445),
    [sym_hash_element_expression] = STATE(2445),
    [sym_coderef_call_expression] = STATE(2445),
    [sym_anonymous_slice_expression] = STATE(2445),
    [sym_slices] = STATE(3424),
    [sym_slice_container_variable] = STATE(5704),
    [sym_slice_expression] = STATE(3425),
    [sym_keyval_container_variable] = STATE(5705),
    [sym_keyval_expression] = STATE(3425),
    [sym__term] = STATE(3793),
    [sym_readline_expression] = STATE(3424),
    [sym_fileglob_expression] = STATE(3424),
    [sym_assignment_expression] = STATE(3424),
    [sym_binary_expression] = STATE(3424),
    [sym_equality_expression] = STATE(3424),
    [sym_relational_expression] = STATE(3424),
    [sym_unary_expression] = STATE(3424),
    [sym_preinc_expression] = STATE(3424),
    [sym_postinc_expression] = STATE(3424),
    [sym_conditional_expression] = STATE(3424),
    [sym_refgen_expression] = STATE(3424),
    [sym_anonymous_array_expression] = STATE(3424),
    [sym_anonymous_hash_expression] = STATE(3424),
    [sym_anonymous_subroutine_expression] = STATE(3424),
    [sym_anonymous_method_expression] = STATE(3424),
    [sym_do_expression] = STATE(3424),
    [sym_eval_expression] = STATE(3424),
    [sym_await_expression] = STATE(3424),
    [sym_variable_declaration] = STATE(3424),
    [sym_localization_expression] = STATE(3424),
    [sym_stub_expression] = STATE(3424),
    [sym_postfix_deref] = STATE(3424),
    [sym_scalar_deref_expression] = STATE(3426),
    [sym_array_deref_expression] = STATE(3426),
    [sym_arraylen_deref_expression] = STATE(3426),
    [sym_hash_deref_expression] = STATE(3426),
    [sym_amper_deref_expression] = STATE(3426),
    [sym_glob_deref_expression] = STATE(3426),
    [sym_require_expression] = STATE(3424),
    [sym_require_version_expression] = STATE(3424),
    [sym_func0op_call_expression] = STATE(3424),
    [sym_func1op_call_expression] = STATE(3424),
    [sym_map_grep_expression] = STATE(3424),
    [sym_sort_expression] = STATE(3424),
    [sym_loopex_expression] = STATE(3424),
    [sym_goto_expression] = STATE(3424),
    [sym_return_expression] = STATE(3424),
    [sym_undef_expression] = STATE(3424),
    [sym__listop] = STATE(3424),
    [sym_function_call_expression] = STATE(3427),
    [sym_ambiguous_function_call_expression] = STATE(3427),
    [sym_function] = STATE(417),
    [sym_method_call_expression] = STATE(3427),
    [sym__variables] = STATE(3424),
    [sym_scalar] = STATE(3428),
    [sym_array] = STATE(3428),
    [sym__SUB_AMPER] = STATE(4186),
    [sym__GLOB_STAR] = STATE(4207),
    [sym_hash] = STATE(3428),
    [sym_arraylen] = STATE(3428),
    [sym_glob] = STATE(3277),
    [sym_amper_sub] = STATE(3278),
    [sym__literal] = STATE(3424),
    [sym__apostrophe] = STATE(4731),
    [sym__quotation_mark] = STATE(3895),
    [sym__backtick] = STATE(3909),
    [sym__search_slash] = STATE(3856),
    [sym_string_literal] = STATE(3432),
    [sym_interpolated_string_literal] = STATE(3432),
    [sym_quoted_word_list] = STATE(3279),
    [sym_command_string] = STATE(3432),
    [sym_quoted_regexp] = STATE(3432),
    [sym_match_regexp] = STATE(3432),
    [sym_substitution_regexp] = STATE(3432),
    [sym_transliteration_expression] = STATE(3432),
    [sym_heredoc_token] = STATE(3424),
    [sym_command_heredoc_token] = STATE(3424),
    [sym_heredoc_content] = STATE(1121),
    [sym_autoquoted_bareword] = STATE(3424),
    [sym_bareword] = STATE(3424),
    [sym_primitive] = STATE(3424),
    [sym_boolean] = STATE(3419),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5093),
    [sym__identifier] = ACTIONS(820),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(822),
    [anon_sym_AT] = ACTIONS(824),
    [anon_sym_my] = ACTIONS(1644),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(828),
    [anon_sym_method] = ACTIONS(830),
    [anon_sym_LPAREN] = ACTIONS(832),
    [anon_sym_continue] = ACTIONS(834),
    [anon_sym_state] = ACTIONS(1644),
    [anon_sym_our] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(836),
    [anon_sym_LT_LT] = ACTIONS(2429),
    [anon_sym_SLASH_SLASH] = ACTIONS(2431),
    [anon_sym_PLUS] = ACTIONS(1815),
    [anon_sym_DASH] = ACTIONS(2568),
    [anon_sym_TILDE] = ACTIONS(1815),
    [anon_sym_BANG] = ACTIONS(1815),
    [anon_sym_PLUS_PLUS] = ACTIONS(2570),
    [anon_sym_DASH_DASH] = ACTIONS(2570),
    [anon_sym_BSLASH] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1819),
    [anon_sym_eval] = ACTIONS(1821),
    [anon_sym_await] = ACTIONS(1823),
    [anon_sym_field] = ACTIONS(1644),
    [anon_sym_local] = ACTIONS(1825),
    [anon_sym_dynamically] = ACTIONS(1825),
    [anon_sym_DOLLAR_POUND] = ACTIONS(865),
    [anon_sym_require] = ACTIONS(1827),
    [anon_sym_map] = ACTIONS(1829),
    [anon_sym_grep] = ACTIONS(1829),
    [anon_sym_sort] = ACTIONS(1831),
    [anon_sym_goto] = ACTIONS(1833),
    [anon_sym_return] = ACTIONS(1835),
    [anon_sym_undef] = ACTIONS(1837),
    [anon_sym_PERCENT2] = ACTIONS(879),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1839),
    [anon_sym_next] = ACTIONS(1839),
    [anon_sym_redo] = ACTIONS(1839),
    [anon_sym___FILE__] = ACTIONS(834),
    [anon_sym___LINE__] = ACTIONS(834),
    [anon_sym___PACKAGE__] = ACTIONS(834),
    [anon_sym___SUB__] = ACTIONS(834),
    [anon_sym_break] = ACTIONS(834),
    [anon_sym_fork] = ACTIONS(834),
    [anon_sym_getppid] = ACTIONS(834),
    [anon_sym_time] = ACTIONS(834),
    [anon_sym_times] = ACTIONS(834),
    [anon_sym_wait] = ACTIONS(834),
    [anon_sym_wantarray] = ACTIONS(834),
    [anon_sym_abs] = ACTIONS(1841),
    [anon_sym_alarm] = ACTIONS(1841),
    [anon_sym_chop] = ACTIONS(1841),
    [anon_sym_chdir] = ACTIONS(1841),
    [anon_sym_close] = ACTIONS(1841),
    [anon_sym_closedir] = ACTIONS(1841),
    [anon_sym_caller] = ACTIONS(1841),
    [anon_sym_chomp] = ACTIONS(1841),
    [anon_sym_chr] = ACTIONS(1841),
    [anon_sym_cos] = ACTIONS(1841),
    [anon_sym_chroot] = ACTIONS(1841),
    [anon_sym_defined] = ACTIONS(1841),
    [anon_sym_delete] = ACTIONS(1841),
    [anon_sym_dbmclose] = ACTIONS(1841),
    [anon_sym_exists] = ACTIONS(1841),
    [anon_sym_exit] = ACTIONS(1841),
    [anon_sym_eof] = ACTIONS(1841),
    [anon_sym_exp] = ACTIONS(1841),
    [anon_sym_each] = ACTIONS(1841),
    [anon_sym_fc] = ACTIONS(1841),
    [anon_sym_fileno] = ACTIONS(1841),
    [anon_sym_gmtime] = ACTIONS(1841),
    [anon_sym_getc] = ACTIONS(1841),
    [anon_sym_getpgrp] = ACTIONS(1841),
    [anon_sym_getprotobyname] = ACTIONS(1841),
    [anon_sym_getpwname] = ACTIONS(1841),
    [anon_sym_getpwuid] = ACTIONS(1841),
    [anon_sym_getpeername] = ACTIONS(1841),
    [anon_sym_getnetbyname] = ACTIONS(1841),
    [anon_sym_getsockname] = ACTIONS(1841),
    [anon_sym_getgrnam] = ACTIONS(1841),
    [anon_sym_getgrgid] = ACTIONS(1841),
    [anon_sym_hex] = ACTIONS(1841),
    [anon_sym_int] = ACTIONS(1841),
    [anon_sym_keys] = ACTIONS(1841),
    [anon_sym_lc] = ACTIONS(1841),
    [anon_sym_lcfirst] = ACTIONS(1841),
    [anon_sym_length] = ACTIONS(1841),
    [anon_sym_localtime] = ACTIONS(1841),
    [anon_sym_log] = ACTIONS(1841),
    [anon_sym_lock] = ACTIONS(1841),
    [anon_sym_lstat] = ACTIONS(1841),
    [anon_sym_oct] = ACTIONS(1841),
    [anon_sym_ord] = ACTIONS(1841),
    [anon_sym_prototype] = ACTIONS(1841),
    [anon_sym_pop] = ACTIONS(1841),
    [anon_sym_pos] = ACTIONS(1841),
    [anon_sym_quotemeta] = ACTIONS(1841),
    [anon_sym_reset] = ACTIONS(1841),
    [anon_sym_rand] = ACTIONS(1841),
    [anon_sym_rmdir] = ACTIONS(1841),
    [anon_sym_readdir] = ACTIONS(1841),
    [anon_sym_readline] = ACTIONS(1841),
    [anon_sym_readpipe] = ACTIONS(1841),
    [anon_sym_rewinddir] = ACTIONS(1841),
    [anon_sym_readlink] = ACTIONS(1841),
    [anon_sym_ref] = ACTIONS(1841),
    [anon_sym_scalar] = ACTIONS(1841),
    [anon_sym_shift] = ACTIONS(1841),
    [anon_sym_sin] = ACTIONS(1841),
    [anon_sym_sleep] = ACTIONS(1841),
    [anon_sym_sqrt] = ACTIONS(1841),
    [anon_sym_srand] = ACTIONS(1841),
    [anon_sym_stat] = ACTIONS(1841),
    [anon_sym_study] = ACTIONS(1841),
    [anon_sym_tell] = ACTIONS(1841),
    [anon_sym_telldir] = ACTIONS(1841),
    [anon_sym_tied] = ACTIONS(1841),
    [anon_sym_uc] = ACTIONS(1841),
    [anon_sym_ucfirst] = ACTIONS(1841),
    [anon_sym_untie] = ACTIONS(1841),
    [anon_sym_umask] = ACTIONS(1841),
    [anon_sym_values] = ACTIONS(1841),
    [anon_sym_write] = ACTIONS(1841),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(885),
    [anon_sym_qq] = ACTIONS(887),
    [anon_sym_qw] = ACTIONS(889),
    [anon_sym_qx] = ACTIONS(891),
    [anon_sym_qr] = ACTIONS(893),
    [anon_sym_m] = ACTIONS(895),
    [anon_sym_s] = ACTIONS(897),
    [anon_sym_tr] = ACTIONS(899),
    [anon_sym_y] = ACTIONS(899),
    [aux_sym__bareword_token1] = ACTIONS(820),
    [sym_number] = ACTIONS(901),
    [anon_sym_true] = ACTIONS(903),
    [anon_sym_false] = ACTIONS(903),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(905),
    [sym__open_fileglob_bracket] = ACTIONS(907),
    [sym__PERLY_HEREDOC] = ACTIONS(909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(911),
    [sym__filetest] = ACTIONS(1843),
  },
  [STATE(1122)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(2070),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(416),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1122),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(243),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(213),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(213),
    [anon_sym_our] = ACTIONS(213),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(215),
    [anon_sym_DASH] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(215),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_PLUS_PLUS] = ACTIONS(2085),
    [anon_sym_DASH_DASH] = ACTIONS(2085),
    [anon_sym_BSLASH] = ACTIONS(217),
    [anon_sym_do] = ACTIONS(219),
    [anon_sym_eval] = ACTIONS(221),
    [anon_sym_await] = ACTIONS(223),
    [anon_sym_field] = ACTIONS(213),
    [anon_sym_local] = ACTIONS(225),
    [anon_sym_dynamically] = ACTIONS(225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(227),
    [anon_sym_map] = ACTIONS(229),
    [anon_sym_grep] = ACTIONS(229),
    [anon_sym_sort] = ACTIONS(231),
    [anon_sym_goto] = ACTIONS(233),
    [anon_sym_return] = ACTIONS(235),
    [anon_sym_undef] = ACTIONS(237),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(239),
    [anon_sym_next] = ACTIONS(239),
    [anon_sym_redo] = ACTIONS(239),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(241),
    [anon_sym_alarm] = ACTIONS(241),
    [anon_sym_chop] = ACTIONS(241),
    [anon_sym_chdir] = ACTIONS(241),
    [anon_sym_close] = ACTIONS(241),
    [anon_sym_closedir] = ACTIONS(241),
    [anon_sym_caller] = ACTIONS(241),
    [anon_sym_chomp] = ACTIONS(241),
    [anon_sym_chr] = ACTIONS(241),
    [anon_sym_cos] = ACTIONS(241),
    [anon_sym_chroot] = ACTIONS(241),
    [anon_sym_defined] = ACTIONS(241),
    [anon_sym_delete] = ACTIONS(241),
    [anon_sym_dbmclose] = ACTIONS(241),
    [anon_sym_exists] = ACTIONS(241),
    [anon_sym_exit] = ACTIONS(241),
    [anon_sym_eof] = ACTIONS(241),
    [anon_sym_exp] = ACTIONS(241),
    [anon_sym_each] = ACTIONS(241),
    [anon_sym_fc] = ACTIONS(241),
    [anon_sym_fileno] = ACTIONS(241),
    [anon_sym_gmtime] = ACTIONS(241),
    [anon_sym_getc] = ACTIONS(241),
    [anon_sym_getpgrp] = ACTIONS(241),
    [anon_sym_getprotobyname] = ACTIONS(241),
    [anon_sym_getpwname] = ACTIONS(241),
    [anon_sym_getpwuid] = ACTIONS(241),
    [anon_sym_getpeername] = ACTIONS(241),
    [anon_sym_getnetbyname] = ACTIONS(241),
    [anon_sym_getsockname] = ACTIONS(241),
    [anon_sym_getgrnam] = ACTIONS(241),
    [anon_sym_getgrgid] = ACTIONS(241),
    [anon_sym_hex] = ACTIONS(241),
    [anon_sym_int] = ACTIONS(241),
    [anon_sym_keys] = ACTIONS(241),
    [anon_sym_lc] = ACTIONS(241),
    [anon_sym_lcfirst] = ACTIONS(241),
    [anon_sym_length] = ACTIONS(241),
    [anon_sym_localtime] = ACTIONS(241),
    [anon_sym_log] = ACTIONS(241),
    [anon_sym_lock] = ACTIONS(241),
    [anon_sym_lstat] = ACTIONS(241),
    [anon_sym_oct] = ACTIONS(241),
    [anon_sym_ord] = ACTIONS(241),
    [anon_sym_prototype] = ACTIONS(241),
    [anon_sym_pop] = ACTIONS(241),
    [anon_sym_pos] = ACTIONS(241),
    [anon_sym_quotemeta] = ACTIONS(241),
    [anon_sym_reset] = ACTIONS(241),
    [anon_sym_rand] = ACTIONS(241),
    [anon_sym_rmdir] = ACTIONS(241),
    [anon_sym_readdir] = ACTIONS(241),
    [anon_sym_readline] = ACTIONS(241),
    [anon_sym_readpipe] = ACTIONS(241),
    [anon_sym_rewinddir] = ACTIONS(241),
    [anon_sym_readlink] = ACTIONS(241),
    [anon_sym_ref] = ACTIONS(241),
    [anon_sym_scalar] = ACTIONS(241),
    [anon_sym_shift] = ACTIONS(241),
    [anon_sym_sin] = ACTIONS(241),
    [anon_sym_sleep] = ACTIONS(241),
    [anon_sym_sqrt] = ACTIONS(241),
    [anon_sym_srand] = ACTIONS(241),
    [anon_sym_stat] = ACTIONS(241),
    [anon_sym_study] = ACTIONS(241),
    [anon_sym_tell] = ACTIONS(241),
    [anon_sym_telldir] = ACTIONS(241),
    [anon_sym_tied] = ACTIONS(241),
    [anon_sym_uc] = ACTIONS(241),
    [anon_sym_ucfirst] = ACTIONS(241),
    [anon_sym_untie] = ACTIONS(241),
    [anon_sym_umask] = ACTIONS(241),
    [anon_sym_values] = ACTIONS(241),
    [anon_sym_write] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(243),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(245),
  },
  [STATE(1123)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3307),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1123),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1124)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3308),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1124),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1125)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3276),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1125),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1126)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1906),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1126),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(125),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(125),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(147),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1127)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(2853),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1127),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1128)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(352),
    [sym_subscripted] = STATE(2029),
    [sym_container_variable] = STATE(5681),
    [sym_glob_slot_expression] = STATE(2031),
    [sym_array_element_expression] = STATE(2031),
    [sym_hash_element_expression] = STATE(2031),
    [sym_coderef_call_expression] = STATE(2031),
    [sym_anonymous_slice_expression] = STATE(2031),
    [sym_slices] = STATE(2383),
    [sym_slice_container_variable] = STATE(5682),
    [sym_slice_expression] = STATE(2387),
    [sym_keyval_container_variable] = STATE(5683),
    [sym_keyval_expression] = STATE(2387),
    [sym__term] = STATE(2554),
    [sym_readline_expression] = STATE(2383),
    [sym_fileglob_expression] = STATE(2383),
    [sym_assignment_expression] = STATE(2383),
    [sym_binary_expression] = STATE(2383),
    [sym_equality_expression] = STATE(2383),
    [sym_relational_expression] = STATE(2383),
    [sym_unary_expression] = STATE(2383),
    [sym_preinc_expression] = STATE(2383),
    [sym_postinc_expression] = STATE(2383),
    [sym_conditional_expression] = STATE(2383),
    [sym_refgen_expression] = STATE(2383),
    [sym_anonymous_array_expression] = STATE(2383),
    [sym_anonymous_hash_expression] = STATE(2383),
    [sym_anonymous_subroutine_expression] = STATE(2383),
    [sym_anonymous_method_expression] = STATE(2383),
    [sym_do_expression] = STATE(2383),
    [sym_eval_expression] = STATE(2383),
    [sym_await_expression] = STATE(2383),
    [sym_variable_declaration] = STATE(2383),
    [sym_localization_expression] = STATE(2383),
    [sym_stub_expression] = STATE(2383),
    [sym_postfix_deref] = STATE(2383),
    [sym_scalar_deref_expression] = STATE(2398),
    [sym_array_deref_expression] = STATE(2398),
    [sym_arraylen_deref_expression] = STATE(2398),
    [sym_hash_deref_expression] = STATE(2398),
    [sym_amper_deref_expression] = STATE(2398),
    [sym_glob_deref_expression] = STATE(2398),
    [sym_require_expression] = STATE(2383),
    [sym_require_version_expression] = STATE(2383),
    [sym_func0op_call_expression] = STATE(2383),
    [sym_func1op_call_expression] = STATE(2383),
    [sym_map_grep_expression] = STATE(2383),
    [sym_sort_expression] = STATE(2383),
    [sym_loopex_expression] = STATE(2383),
    [sym_goto_expression] = STATE(2383),
    [sym_return_expression] = STATE(2383),
    [sym_undef_expression] = STATE(2383),
    [sym__listop] = STATE(2383),
    [sym_function_call_expression] = STATE(2443),
    [sym_ambiguous_function_call_expression] = STATE(2443),
    [sym_function] = STATE(396),
    [sym_method_call_expression] = STATE(2443),
    [sym__variables] = STATE(2383),
    [sym_scalar] = STATE(2467),
    [sym_array] = STATE(2467),
    [sym__SUB_AMPER] = STATE(4140),
    [sym__GLOB_STAR] = STATE(4255),
    [sym_hash] = STATE(2467),
    [sym_arraylen] = STATE(2467),
    [sym_glob] = STATE(2226),
    [sym_amper_sub] = STATE(2262),
    [sym__literal] = STATE(2383),
    [sym__apostrophe] = STATE(4799),
    [sym__quotation_mark] = STATE(3912),
    [sym__backtick] = STATE(3939),
    [sym__search_slash] = STATE(3838),
    [sym_string_literal] = STATE(2594),
    [sym_interpolated_string_literal] = STATE(2594),
    [sym_quoted_word_list] = STATE(2247),
    [sym_command_string] = STATE(2594),
    [sym_quoted_regexp] = STATE(2594),
    [sym_match_regexp] = STATE(2594),
    [sym_substitution_regexp] = STATE(2594),
    [sym_transliteration_expression] = STATE(2594),
    [sym_heredoc_token] = STATE(2383),
    [sym_command_heredoc_token] = STATE(2383),
    [sym_heredoc_content] = STATE(1128),
    [sym_autoquoted_bareword] = STATE(2383),
    [sym_bareword] = STATE(2383),
    [sym_primitive] = STATE(2383),
    [sym_boolean] = STATE(2303),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5044),
    [sym__identifier] = ACTIONS(265),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(267),
    [anon_sym_AT] = ACTIONS(269),
    [anon_sym_my] = ACTIONS(271),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(273),
    [anon_sym_method] = ACTIONS(275),
    [anon_sym_LPAREN] = ACTIONS(277),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_state] = ACTIONS(271),
    [anon_sym_our] = ACTIONS(271),
    [anon_sym_LBRACK] = ACTIONS(281),
    [anon_sym_LT_LT] = ACTIONS(2121),
    [anon_sym_SLASH_SLASH] = ACTIONS(2123),
    [anon_sym_PLUS] = ACTIONS(295),
    [anon_sym_DASH] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(295),
    [anon_sym_BANG] = ACTIONS(295),
    [anon_sym_PLUS_PLUS] = ACTIONS(2127),
    [anon_sym_DASH_DASH] = ACTIONS(2127),
    [anon_sym_BSLASH] = ACTIONS(300),
    [anon_sym_do] = ACTIONS(302),
    [anon_sym_eval] = ACTIONS(304),
    [anon_sym_await] = ACTIONS(306),
    [anon_sym_field] = ACTIONS(271),
    [anon_sym_local] = ACTIONS(308),
    [anon_sym_dynamically] = ACTIONS(308),
    [anon_sym_DOLLAR_POUND] = ACTIONS(310),
    [anon_sym_require] = ACTIONS(312),
    [anon_sym_map] = ACTIONS(314),
    [anon_sym_grep] = ACTIONS(314),
    [anon_sym_sort] = ACTIONS(316),
    [anon_sym_goto] = ACTIONS(318),
    [anon_sym_return] = ACTIONS(320),
    [anon_sym_undef] = ACTIONS(322),
    [anon_sym_PERCENT2] = ACTIONS(324),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(326),
    [anon_sym_next] = ACTIONS(326),
    [anon_sym_redo] = ACTIONS(326),
    [anon_sym___FILE__] = ACTIONS(279),
    [anon_sym___LINE__] = ACTIONS(279),
    [anon_sym___PACKAGE__] = ACTIONS(279),
    [anon_sym___SUB__] = ACTIONS(279),
    [anon_sym_break] = ACTIONS(279),
    [anon_sym_fork] = ACTIONS(279),
    [anon_sym_getppid] = ACTIONS(279),
    [anon_sym_time] = ACTIONS(279),
    [anon_sym_times] = ACTIONS(279),
    [anon_sym_wait] = ACTIONS(279),
    [anon_sym_wantarray] = ACTIONS(279),
    [anon_sym_abs] = ACTIONS(328),
    [anon_sym_alarm] = ACTIONS(328),
    [anon_sym_chop] = ACTIONS(328),
    [anon_sym_chdir] = ACTIONS(328),
    [anon_sym_close] = ACTIONS(328),
    [anon_sym_closedir] = ACTIONS(328),
    [anon_sym_caller] = ACTIONS(328),
    [anon_sym_chomp] = ACTIONS(328),
    [anon_sym_chr] = ACTIONS(328),
    [anon_sym_cos] = ACTIONS(328),
    [anon_sym_chroot] = ACTIONS(328),
    [anon_sym_defined] = ACTIONS(328),
    [anon_sym_delete] = ACTIONS(328),
    [anon_sym_dbmclose] = ACTIONS(328),
    [anon_sym_exists] = ACTIONS(328),
    [anon_sym_exit] = ACTIONS(328),
    [anon_sym_eof] = ACTIONS(328),
    [anon_sym_exp] = ACTIONS(328),
    [anon_sym_each] = ACTIONS(328),
    [anon_sym_fc] = ACTIONS(328),
    [anon_sym_fileno] = ACTIONS(328),
    [anon_sym_gmtime] = ACTIONS(328),
    [anon_sym_getc] = ACTIONS(328),
    [anon_sym_getpgrp] = ACTIONS(328),
    [anon_sym_getprotobyname] = ACTIONS(328),
    [anon_sym_getpwname] = ACTIONS(328),
    [anon_sym_getpwuid] = ACTIONS(328),
    [anon_sym_getpeername] = ACTIONS(328),
    [anon_sym_getnetbyname] = ACTIONS(328),
    [anon_sym_getsockname] = ACTIONS(328),
    [anon_sym_getgrnam] = ACTIONS(328),
    [anon_sym_getgrgid] = ACTIONS(328),
    [anon_sym_hex] = ACTIONS(328),
    [anon_sym_int] = ACTIONS(328),
    [anon_sym_keys] = ACTIONS(328),
    [anon_sym_lc] = ACTIONS(328),
    [anon_sym_lcfirst] = ACTIONS(328),
    [anon_sym_length] = ACTIONS(328),
    [anon_sym_localtime] = ACTIONS(328),
    [anon_sym_log] = ACTIONS(328),
    [anon_sym_lock] = ACTIONS(328),
    [anon_sym_lstat] = ACTIONS(328),
    [anon_sym_oct] = ACTIONS(328),
    [anon_sym_ord] = ACTIONS(328),
    [anon_sym_prototype] = ACTIONS(328),
    [anon_sym_pop] = ACTIONS(328),
    [anon_sym_pos] = ACTIONS(328),
    [anon_sym_quotemeta] = ACTIONS(328),
    [anon_sym_reset] = ACTIONS(328),
    [anon_sym_rand] = ACTIONS(328),
    [anon_sym_rmdir] = ACTIONS(328),
    [anon_sym_readdir] = ACTIONS(328),
    [anon_sym_readline] = ACTIONS(328),
    [anon_sym_readpipe] = ACTIONS(328),
    [anon_sym_rewinddir] = ACTIONS(328),
    [anon_sym_readlink] = ACTIONS(328),
    [anon_sym_ref] = ACTIONS(328),
    [anon_sym_scalar] = ACTIONS(328),
    [anon_sym_shift] = ACTIONS(328),
    [anon_sym_sin] = ACTIONS(328),
    [anon_sym_sleep] = ACTIONS(328),
    [anon_sym_sqrt] = ACTIONS(328),
    [anon_sym_srand] = ACTIONS(328),
    [anon_sym_stat] = ACTIONS(328),
    [anon_sym_study] = ACTIONS(328),
    [anon_sym_tell] = ACTIONS(328),
    [anon_sym_telldir] = ACTIONS(328),
    [anon_sym_tied] = ACTIONS(328),
    [anon_sym_uc] = ACTIONS(328),
    [anon_sym_ucfirst] = ACTIONS(328),
    [anon_sym_untie] = ACTIONS(328),
    [anon_sym_umask] = ACTIONS(328),
    [anon_sym_values] = ACTIONS(328),
    [anon_sym_write] = ACTIONS(328),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(330),
    [anon_sym_qq] = ACTIONS(332),
    [anon_sym_qw] = ACTIONS(334),
    [anon_sym_qx] = ACTIONS(336),
    [anon_sym_qr] = ACTIONS(338),
    [anon_sym_m] = ACTIONS(340),
    [anon_sym_s] = ACTIONS(342),
    [anon_sym_tr] = ACTIONS(344),
    [anon_sym_y] = ACTIONS(344),
    [aux_sym__bareword_token1] = ACTIONS(265),
    [sym_number] = ACTIONS(346),
    [anon_sym_true] = ACTIONS(348),
    [anon_sym_false] = ACTIONS(348),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(350),
    [sym__open_fileglob_bracket] = ACTIONS(352),
    [sym__PERLY_HEREDOC] = ACTIONS(354),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(356),
    [sym__filetest] = ACTIONS(358),
  },
  [STATE(1129)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(374),
    [sym_subscripted] = STATE(2192),
    [sym_container_variable] = STATE(5721),
    [sym_glob_slot_expression] = STATE(2131),
    [sym_array_element_expression] = STATE(2131),
    [sym_hash_element_expression] = STATE(2131),
    [sym_coderef_call_expression] = STATE(2131),
    [sym_anonymous_slice_expression] = STATE(2131),
    [sym_slices] = STATE(2837),
    [sym_slice_container_variable] = STATE(5722),
    [sym_slice_expression] = STATE(2838),
    [sym_keyval_container_variable] = STATE(5723),
    [sym_keyval_expression] = STATE(2838),
    [sym__term] = STATE(3283),
    [sym_readline_expression] = STATE(2837),
    [sym_fileglob_expression] = STATE(2837),
    [sym_assignment_expression] = STATE(2837),
    [sym_binary_expression] = STATE(2837),
    [sym_equality_expression] = STATE(2837),
    [sym_relational_expression] = STATE(2837),
    [sym_unary_expression] = STATE(2837),
    [sym_preinc_expression] = STATE(2837),
    [sym_postinc_expression] = STATE(2837),
    [sym_conditional_expression] = STATE(2837),
    [sym_refgen_expression] = STATE(2837),
    [sym_anonymous_array_expression] = STATE(2837),
    [sym_anonymous_hash_expression] = STATE(2837),
    [sym_anonymous_subroutine_expression] = STATE(2837),
    [sym_anonymous_method_expression] = STATE(2837),
    [sym_do_expression] = STATE(2837),
    [sym_eval_expression] = STATE(2837),
    [sym_await_expression] = STATE(2837),
    [sym_variable_declaration] = STATE(2837),
    [sym_localization_expression] = STATE(2837),
    [sym_stub_expression] = STATE(2837),
    [sym_postfix_deref] = STATE(2837),
    [sym_scalar_deref_expression] = STATE(2839),
    [sym_array_deref_expression] = STATE(2839),
    [sym_arraylen_deref_expression] = STATE(2839),
    [sym_hash_deref_expression] = STATE(2839),
    [sym_amper_deref_expression] = STATE(2839),
    [sym_glob_deref_expression] = STATE(2839),
    [sym_require_expression] = STATE(2837),
    [sym_require_version_expression] = STATE(2837),
    [sym_func0op_call_expression] = STATE(2837),
    [sym_func1op_call_expression] = STATE(2837),
    [sym_map_grep_expression] = STATE(2837),
    [sym_sort_expression] = STATE(2837),
    [sym_loopex_expression] = STATE(2837),
    [sym_goto_expression] = STATE(2837),
    [sym_return_expression] = STATE(2837),
    [sym_undef_expression] = STATE(2837),
    [sym__listop] = STATE(2837),
    [sym_function_call_expression] = STATE(2840),
    [sym_ambiguous_function_call_expression] = STATE(2840),
    [sym_function] = STATE(414),
    [sym_method_call_expression] = STATE(2840),
    [sym__variables] = STATE(2837),
    [sym_scalar] = STATE(2841),
    [sym_array] = STATE(2841),
    [sym__SUB_AMPER] = STATE(4280),
    [sym__GLOB_STAR] = STATE(4281),
    [sym_hash] = STATE(2841),
    [sym_arraylen] = STATE(2841),
    [sym_glob] = STATE(2419),
    [sym_amper_sub] = STATE(2420),
    [sym__literal] = STATE(2837),
    [sym__apostrophe] = STATE(4789),
    [sym__quotation_mark] = STATE(3928),
    [sym__backtick] = STATE(3980),
    [sym__search_slash] = STATE(3857),
    [sym_string_literal] = STATE(2842),
    [sym_interpolated_string_literal] = STATE(2842),
    [sym_quoted_word_list] = STATE(2421),
    [sym_command_string] = STATE(2842),
    [sym_quoted_regexp] = STATE(2842),
    [sym_match_regexp] = STATE(2842),
    [sym_substitution_regexp] = STATE(2842),
    [sym_transliteration_expression] = STATE(2842),
    [sym_heredoc_token] = STATE(2837),
    [sym_command_heredoc_token] = STATE(2837),
    [sym_heredoc_content] = STATE(1129),
    [sym_autoquoted_bareword] = STATE(2837),
    [sym_bareword] = STATE(2837),
    [sym_primitive] = STATE(2837),
    [sym_boolean] = STATE(2834),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5119),
    [sym__identifier] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(2546),
    [anon_sym_AT] = ACTIONS(2409),
    [anon_sym_my] = ACTIONS(925),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(927),
    [anon_sym_method] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(931),
    [anon_sym_continue] = ACTIONS(933),
    [anon_sym_state] = ACTIONS(925),
    [anon_sym_our] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(935),
    [anon_sym_LT_LT] = ACTIONS(2411),
    [anon_sym_SLASH_SLASH] = ACTIONS(2413),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(2523),
    [anon_sym_TILDE] = ACTIONS(1059),
    [anon_sym_BANG] = ACTIONS(1059),
    [anon_sym_PLUS_PLUS] = ACTIONS(2525),
    [anon_sym_DASH_DASH] = ACTIONS(2525),
    [anon_sym_BSLASH] = ACTIONS(1061),
    [anon_sym_do] = ACTIONS(1063),
    [anon_sym_eval] = ACTIONS(1065),
    [anon_sym_await] = ACTIONS(1067),
    [anon_sym_field] = ACTIONS(925),
    [anon_sym_local] = ACTIONS(1069),
    [anon_sym_dynamically] = ACTIONS(1069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(964),
    [anon_sym_require] = ACTIONS(1071),
    [anon_sym_map] = ACTIONS(1073),
    [anon_sym_grep] = ACTIONS(1073),
    [anon_sym_sort] = ACTIONS(1075),
    [anon_sym_goto] = ACTIONS(1077),
    [anon_sym_return] = ACTIONS(1079),
    [anon_sym_undef] = ACTIONS(1081),
    [anon_sym_PERCENT2] = ACTIONS(2419),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(1083),
    [anon_sym_next] = ACTIONS(1083),
    [anon_sym_redo] = ACTIONS(1083),
    [anon_sym___FILE__] = ACTIONS(933),
    [anon_sym___LINE__] = ACTIONS(933),
    [anon_sym___PACKAGE__] = ACTIONS(933),
    [anon_sym___SUB__] = ACTIONS(933),
    [anon_sym_break] = ACTIONS(933),
    [anon_sym_fork] = ACTIONS(933),
    [anon_sym_getppid] = ACTIONS(933),
    [anon_sym_time] = ACTIONS(933),
    [anon_sym_times] = ACTIONS(933),
    [anon_sym_wait] = ACTIONS(933),
    [anon_sym_wantarray] = ACTIONS(933),
    [anon_sym_abs] = ACTIONS(1085),
    [anon_sym_alarm] = ACTIONS(1085),
    [anon_sym_chop] = ACTIONS(1085),
    [anon_sym_chdir] = ACTIONS(1085),
    [anon_sym_close] = ACTIONS(1085),
    [anon_sym_closedir] = ACTIONS(1085),
    [anon_sym_caller] = ACTIONS(1085),
    [anon_sym_chomp] = ACTIONS(1085),
    [anon_sym_chr] = ACTIONS(1085),
    [anon_sym_cos] = ACTIONS(1085),
    [anon_sym_chroot] = ACTIONS(1085),
    [anon_sym_defined] = ACTIONS(1085),
    [anon_sym_delete] = ACTIONS(1085),
    [anon_sym_dbmclose] = ACTIONS(1085),
    [anon_sym_exists] = ACTIONS(1085),
    [anon_sym_exit] = ACTIONS(1085),
    [anon_sym_eof] = ACTIONS(1085),
    [anon_sym_exp] = ACTIONS(1085),
    [anon_sym_each] = ACTIONS(1085),
    [anon_sym_fc] = ACTIONS(1085),
    [anon_sym_fileno] = ACTIONS(1085),
    [anon_sym_gmtime] = ACTIONS(1085),
    [anon_sym_getc] = ACTIONS(1085),
    [anon_sym_getpgrp] = ACTIONS(1085),
    [anon_sym_getprotobyname] = ACTIONS(1085),
    [anon_sym_getpwname] = ACTIONS(1085),
    [anon_sym_getpwuid] = ACTIONS(1085),
    [anon_sym_getpeername] = ACTIONS(1085),
    [anon_sym_getnetbyname] = ACTIONS(1085),
    [anon_sym_getsockname] = ACTIONS(1085),
    [anon_sym_getgrnam] = ACTIONS(1085),
    [anon_sym_getgrgid] = ACTIONS(1085),
    [anon_sym_hex] = ACTIONS(1085),
    [anon_sym_int] = ACTIONS(1085),
    [anon_sym_keys] = ACTIONS(1085),
    [anon_sym_lc] = ACTIONS(1085),
    [anon_sym_lcfirst] = ACTIONS(1085),
    [anon_sym_length] = ACTIONS(1085),
    [anon_sym_localtime] = ACTIONS(1085),
    [anon_sym_log] = ACTIONS(1085),
    [anon_sym_lock] = ACTIONS(1085),
    [anon_sym_lstat] = ACTIONS(1085),
    [anon_sym_oct] = ACTIONS(1085),
    [anon_sym_ord] = ACTIONS(1085),
    [anon_sym_prototype] = ACTIONS(1085),
    [anon_sym_pop] = ACTIONS(1085),
    [anon_sym_pos] = ACTIONS(1085),
    [anon_sym_quotemeta] = ACTIONS(1085),
    [anon_sym_reset] = ACTIONS(1085),
    [anon_sym_rand] = ACTIONS(1085),
    [anon_sym_rmdir] = ACTIONS(1085),
    [anon_sym_readdir] = ACTIONS(1085),
    [anon_sym_readline] = ACTIONS(1085),
    [anon_sym_readpipe] = ACTIONS(1085),
    [anon_sym_rewinddir] = ACTIONS(1085),
    [anon_sym_readlink] = ACTIONS(1085),
    [anon_sym_ref] = ACTIONS(1085),
    [anon_sym_scalar] = ACTIONS(1085),
    [anon_sym_shift] = ACTIONS(1085),
    [anon_sym_sin] = ACTIONS(1085),
    [anon_sym_sleep] = ACTIONS(1085),
    [anon_sym_sqrt] = ACTIONS(1085),
    [anon_sym_srand] = ACTIONS(1085),
    [anon_sym_stat] = ACTIONS(1085),
    [anon_sym_study] = ACTIONS(1085),
    [anon_sym_tell] = ACTIONS(1085),
    [anon_sym_telldir] = ACTIONS(1085),
    [anon_sym_tied] = ACTIONS(1085),
    [anon_sym_uc] = ACTIONS(1085),
    [anon_sym_ucfirst] = ACTIONS(1085),
    [anon_sym_untie] = ACTIONS(1085),
    [anon_sym_umask] = ACTIONS(1085),
    [anon_sym_values] = ACTIONS(1085),
    [anon_sym_write] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(985),
    [anon_sym_qq] = ACTIONS(987),
    [anon_sym_qw] = ACTIONS(989),
    [anon_sym_qx] = ACTIONS(991),
    [anon_sym_qr] = ACTIONS(993),
    [anon_sym_m] = ACTIONS(995),
    [anon_sym_s] = ACTIONS(997),
    [anon_sym_tr] = ACTIONS(999),
    [anon_sym_y] = ACTIONS(999),
    [aux_sym__bareword_token1] = ACTIONS(917),
    [sym_number] = ACTIONS(1001),
    [anon_sym_true] = ACTIONS(1003),
    [anon_sym_false] = ACTIONS(1003),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(1005),
    [sym__open_fileglob_bracket] = ACTIONS(1007),
    [sym__PERLY_HEREDOC] = ACTIONS(1009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(1011),
    [sym__filetest] = ACTIONS(1087),
  },
  [STATE(1130)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(324),
    [sym_subscripted] = STATE(1770),
    [sym_container_variable] = STATE(5618),
    [sym_glob_slot_expression] = STATE(1746),
    [sym_array_element_expression] = STATE(1746),
    [sym_hash_element_expression] = STATE(1746),
    [sym_coderef_call_expression] = STATE(1746),
    [sym_anonymous_slice_expression] = STATE(1746),
    [sym_slices] = STATE(1920),
    [sym_slice_container_variable] = STATE(5684),
    [sym_slice_expression] = STATE(1929),
    [sym_keyval_container_variable] = STATE(5680),
    [sym_keyval_expression] = STATE(1929),
    [sym__term] = STATE(1906),
    [sym_readline_expression] = STATE(1920),
    [sym_fileglob_expression] = STATE(1920),
    [sym_assignment_expression] = STATE(1920),
    [sym_binary_expression] = STATE(1920),
    [sym_equality_expression] = STATE(1920),
    [sym_relational_expression] = STATE(1920),
    [sym_unary_expression] = STATE(1920),
    [sym_preinc_expression] = STATE(1920),
    [sym_postinc_expression] = STATE(1920),
    [sym_conditional_expression] = STATE(1920),
    [sym_refgen_expression] = STATE(1920),
    [sym_anonymous_array_expression] = STATE(1920),
    [sym_anonymous_hash_expression] = STATE(1920),
    [sym_anonymous_subroutine_expression] = STATE(1920),
    [sym_anonymous_method_expression] = STATE(1920),
    [sym_do_expression] = STATE(1920),
    [sym_eval_expression] = STATE(1920),
    [sym_await_expression] = STATE(1920),
    [sym_variable_declaration] = STATE(1920),
    [sym_localization_expression] = STATE(1920),
    [sym_stub_expression] = STATE(1920),
    [sym_postfix_deref] = STATE(1920),
    [sym_scalar_deref_expression] = STATE(1947),
    [sym_array_deref_expression] = STATE(1947),
    [sym_arraylen_deref_expression] = STATE(1947),
    [sym_hash_deref_expression] = STATE(1947),
    [sym_amper_deref_expression] = STATE(1947),
    [sym_glob_deref_expression] = STATE(1947),
    [sym_require_expression] = STATE(1920),
    [sym_require_version_expression] = STATE(1920),
    [sym_func0op_call_expression] = STATE(1920),
    [sym_func1op_call_expression] = STATE(1920),
    [sym_map_grep_expression] = STATE(1920),
    [sym_sort_expression] = STATE(1920),
    [sym_loopex_expression] = STATE(1920),
    [sym_goto_expression] = STATE(1920),
    [sym_return_expression] = STATE(1920),
    [sym_undef_expression] = STATE(1920),
    [sym__listop] = STATE(1920),
    [sym_function_call_expression] = STATE(1892),
    [sym_ambiguous_function_call_expression] = STATE(1892),
    [sym_function] = STATE(422),
    [sym_method_call_expression] = STATE(1892),
    [sym__variables] = STATE(1920),
    [sym_scalar] = STATE(1854),
    [sym_array] = STATE(1854),
    [sym__SUB_AMPER] = STATE(4201),
    [sym__GLOB_STAR] = STATE(4190),
    [sym_hash] = STATE(1854),
    [sym_arraylen] = STATE(1854),
    [sym_glob] = STATE(1798),
    [sym_amper_sub] = STATE(1801),
    [sym__literal] = STATE(1920),
    [sym__apostrophe] = STATE(4744),
    [sym__quotation_mark] = STATE(3951),
    [sym__backtick] = STATE(3934),
    [sym__search_slash] = STATE(3829),
    [sym_string_literal] = STATE(1886),
    [sym_interpolated_string_literal] = STATE(1886),
    [sym_quoted_word_list] = STATE(1803),
    [sym_command_string] = STATE(1886),
    [sym_quoted_regexp] = STATE(1886),
    [sym_match_regexp] = STATE(1886),
    [sym_substitution_regexp] = STATE(1886),
    [sym_transliteration_expression] = STATE(1886),
    [sym_heredoc_token] = STATE(1920),
    [sym_command_heredoc_token] = STATE(1920),
    [sym_heredoc_content] = STATE(1130),
    [sym_autoquoted_bareword] = STATE(1920),
    [sym_bareword] = STATE(1920),
    [sym_primitive] = STATE(1920),
    [sym_boolean] = STATE(1922),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5143),
    [sym__identifier] = ACTIONS(2680),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(25),
    [anon_sym_AT] = ACTIONS(27),
    [anon_sym_my] = ACTIONS(41),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(157),
    [anon_sym_method] = ACTIONS(159),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_continue] = ACTIONS(39),
    [anon_sym_state] = ACTIONS(41),
    [anon_sym_our] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(49),
    [anon_sym_LT_LT] = ACTIONS(51),
    [anon_sym_SLASH_SLASH] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(57),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_BSLASH] = ACTIONS(61),
    [anon_sym_do] = ACTIONS(63),
    [anon_sym_eval] = ACTIONS(65),
    [anon_sym_await] = ACTIONS(67),
    [anon_sym_field] = ACTIONS(41),
    [anon_sym_local] = ACTIONS(69),
    [anon_sym_dynamically] = ACTIONS(69),
    [anon_sym_DOLLAR_POUND] = ACTIONS(71),
    [anon_sym_require] = ACTIONS(73),
    [anon_sym_map] = ACTIONS(75),
    [anon_sym_grep] = ACTIONS(75),
    [anon_sym_sort] = ACTIONS(77),
    [anon_sym_goto] = ACTIONS(79),
    [anon_sym_return] = ACTIONS(81),
    [anon_sym_undef] = ACTIONS(83),
    [anon_sym_PERCENT2] = ACTIONS(85),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(95),
    [anon_sym_next] = ACTIONS(95),
    [anon_sym_redo] = ACTIONS(95),
    [anon_sym___FILE__] = ACTIONS(39),
    [anon_sym___LINE__] = ACTIONS(39),
    [anon_sym___PACKAGE__] = ACTIONS(39),
    [anon_sym___SUB__] = ACTIONS(39),
    [anon_sym_break] = ACTIONS(39),
    [anon_sym_fork] = ACTIONS(39),
    [anon_sym_getppid] = ACTIONS(39),
    [anon_sym_time] = ACTIONS(39),
    [anon_sym_times] = ACTIONS(39),
    [anon_sym_wait] = ACTIONS(39),
    [anon_sym_wantarray] = ACTIONS(39),
    [anon_sym_abs] = ACTIONS(99),
    [anon_sym_alarm] = ACTIONS(99),
    [anon_sym_chop] = ACTIONS(99),
    [anon_sym_chdir] = ACTIONS(99),
    [anon_sym_close] = ACTIONS(99),
    [anon_sym_closedir] = ACTIONS(99),
    [anon_sym_caller] = ACTIONS(99),
    [anon_sym_chomp] = ACTIONS(99),
    [anon_sym_chr] = ACTIONS(99),
    [anon_sym_cos] = ACTIONS(99),
    [anon_sym_chroot] = ACTIONS(99),
    [anon_sym_defined] = ACTIONS(99),
    [anon_sym_delete] = ACTIONS(99),
    [anon_sym_dbmclose] = ACTIONS(99),
    [anon_sym_exists] = ACTIONS(99),
    [anon_sym_exit] = ACTIONS(99),
    [anon_sym_eof] = ACTIONS(99),
    [anon_sym_exp] = ACTIONS(99),
    [anon_sym_each] = ACTIONS(99),
    [anon_sym_fc] = ACTIONS(99),
    [anon_sym_fileno] = ACTIONS(99),
    [anon_sym_gmtime] = ACTIONS(99),
    [anon_sym_getc] = ACTIONS(99),
    [anon_sym_getpgrp] = ACTIONS(99),
    [anon_sym_getprotobyname] = ACTIONS(99),
    [anon_sym_getpwname] = ACTIONS(99),
    [anon_sym_getpwuid] = ACTIONS(99),
    [anon_sym_getpeername] = ACTIONS(99),
    [anon_sym_getnetbyname] = ACTIONS(99),
    [anon_sym_getsockname] = ACTIONS(99),
    [anon_sym_getgrnam] = ACTIONS(99),
    [anon_sym_getgrgid] = ACTIONS(99),
    [anon_sym_hex] = ACTIONS(99),
    [anon_sym_int] = ACTIONS(99),
    [anon_sym_keys] = ACTIONS(99),
    [anon_sym_lc] = ACTIONS(99),
    [anon_sym_lcfirst] = ACTIONS(99),
    [anon_sym_length] = ACTIONS(99),
    [anon_sym_localtime] = ACTIONS(99),
    [anon_sym_log] = ACTIONS(99),
    [anon_sym_lock] = ACTIONS(99),
    [anon_sym_lstat] = ACTIONS(99),
    [anon_sym_oct] = ACTIONS(99),
    [anon_sym_ord] = ACTIONS(99),
    [anon_sym_prototype] = ACTIONS(99),
    [anon_sym_pop] = ACTIONS(99),
    [anon_sym_pos] = ACTIONS(99),
    [anon_sym_quotemeta] = ACTIONS(99),
    [anon_sym_reset] = ACTIONS(99),
    [anon_sym_rand] = ACTIONS(99),
    [anon_sym_rmdir] = ACTIONS(99),
    [anon_sym_readdir] = ACTIONS(99),
    [anon_sym_readline] = ACTIONS(99),
    [anon_sym_readpipe] = ACTIONS(99),
    [anon_sym_rewinddir] = ACTIONS(99),
    [anon_sym_readlink] = ACTIONS(99),
    [anon_sym_ref] = ACTIONS(99),
    [anon_sym_scalar] = ACTIONS(99),
    [anon_sym_shift] = ACTIONS(99),
    [anon_sym_sin] = ACTIONS(99),
    [anon_sym_sleep] = ACTIONS(99),
    [anon_sym_sqrt] = ACTIONS(99),
    [anon_sym_srand] = ACTIONS(99),
    [anon_sym_stat] = ACTIONS(99),
    [anon_sym_study] = ACTIONS(99),
    [anon_sym_tell] = ACTIONS(99),
    [anon_sym_telldir] = ACTIONS(99),
    [anon_sym_tied] = ACTIONS(99),
    [anon_sym_uc] = ACTIONS(99),
    [anon_sym_ucfirst] = ACTIONS(99),
    [anon_sym_untie] = ACTIONS(99),
    [anon_sym_umask] = ACTIONS(99),
    [anon_sym_values] = ACTIONS(99),
    [anon_sym_write] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(105),
    [anon_sym_qq] = ACTIONS(107),
    [anon_sym_qw] = ACTIONS(109),
    [anon_sym_qx] = ACTIONS(111),
    [anon_sym_qr] = ACTIONS(113),
    [anon_sym_m] = ACTIONS(115),
    [anon_sym_s] = ACTIONS(117),
    [anon_sym_tr] = ACTIONS(119),
    [anon_sym_y] = ACTIONS(119),
    [aux_sym__bareword_token1] = ACTIONS(2680),
    [sym_number] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(129),
    [anon_sym_false] = ACTIONS(129),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(139),
    [sym__open_fileglob_bracket] = ACTIONS(141),
    [sym__PERLY_HEREDOC] = ACTIONS(143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2678),
    [sym__filetest] = ACTIONS(149),
  },
  [STATE(1131)] = {
    [sym__PERLY_BRACE_OPEN] = STATE(348),
    [sym_subscripted] = STATE(2126),
    [sym_container_variable] = STATE(5713),
    [sym_glob_slot_expression] = STATE(2127),
    [sym_array_element_expression] = STATE(2127),
    [sym_hash_element_expression] = STATE(2127),
    [sym_coderef_call_expression] = STATE(2127),
    [sym_anonymous_slice_expression] = STATE(2127),
    [sym_slices] = STATE(3147),
    [sym_slice_container_variable] = STATE(5714),
    [sym_slice_expression] = STATE(3148),
    [sym_keyval_container_variable] = STATE(5715),
    [sym_keyval_expression] = STATE(3148),
    [sym__term] = STATE(2734),
    [sym_readline_expression] = STATE(3147),
    [sym_fileglob_expression] = STATE(3147),
    [sym_assignment_expression] = STATE(3147),
    [sym_binary_expression] = STATE(3147),
    [sym_equality_expression] = STATE(3147),
    [sym_relational_expression] = STATE(3147),
    [sym_unary_expression] = STATE(3147),
    [sym_preinc_expression] = STATE(3147),
    [sym_postinc_expression] = STATE(3147),
    [sym_conditional_expression] = STATE(3147),
    [sym_refgen_expression] = STATE(3147),
    [sym_anonymous_array_expression] = STATE(3147),
    [sym_anonymous_hash_expression] = STATE(3147),
    [sym_anonymous_subroutine_expression] = STATE(3147),
    [sym_anonymous_method_expression] = STATE(3147),
    [sym_do_expression] = STATE(3147),
    [sym_eval_expression] = STATE(3147),
    [sym_await_expression] = STATE(3147),
    [sym_variable_declaration] = STATE(3147),
    [sym_localization_expression] = STATE(3147),
    [sym_stub_expression] = STATE(3147),
    [sym_postfix_deref] = STATE(3147),
    [sym_scalar_deref_expression] = STATE(3149),
    [sym_array_deref_expression] = STATE(3149),
    [sym_arraylen_deref_expression] = STATE(3149),
    [sym_hash_deref_expression] = STATE(3149),
    [sym_amper_deref_expression] = STATE(3149),
    [sym_glob_deref_expression] = STATE(3149),
    [sym_require_expression] = STATE(3147),
    [sym_require_version_expression] = STATE(3147),
    [sym_func0op_call_expression] = STATE(3147),
    [sym_func1op_call_expression] = STATE(3147),
    [sym_map_grep_expression] = STATE(3147),
    [sym_sort_expression] = STATE(3147),
    [sym_loopex_expression] = STATE(3147),
    [sym_goto_expression] = STATE(3147),
    [sym_return_expression] = STATE(3147),
    [sym_undef_expression] = STATE(3147),
    [sym__listop] = STATE(3147),
    [sym_function_call_expression] = STATE(3150),
    [sym_ambiguous_function_call_expression] = STATE(3150),
    [sym_function] = STATE(407),
    [sym_method_call_expression] = STATE(3150),
    [sym__variables] = STATE(3147),
    [sym_scalar] = STATE(3151),
    [sym_array] = STATE(3151),
    [sym__SUB_AMPER] = STATE(4159),
    [sym__GLOB_STAR] = STATE(4160),
    [sym_hash] = STATE(3151),
    [sym_arraylen] = STATE(3151),
    [sym_glob] = STATE(2446),
    [sym_amper_sub] = STATE(2454),
    [sym__literal] = STATE(3147),
    [sym__apostrophe] = STATE(4774),
    [sym__quotation_mark] = STATE(3978),
    [sym__backtick] = STATE(3878),
    [sym__search_slash] = STATE(3839),
    [sym_string_literal] = STATE(3152),
    [sym_interpolated_string_literal] = STATE(3152),
    [sym_quoted_word_list] = STATE(2268),
    [sym_command_string] = STATE(3152),
    [sym_quoted_regexp] = STATE(3152),
    [sym_match_regexp] = STATE(3152),
    [sym_substitution_regexp] = STATE(3152),
    [sym_transliteration_expression] = STATE(3152),
    [sym_heredoc_token] = STATE(3147),
    [sym_command_heredoc_token] = STATE(3147),
    [sym_heredoc_content] = STATE(1131),
    [sym_autoquoted_bareword] = STATE(3147),
    [sym_bareword] = STATE(3147),
    [sym_primitive] = STATE(3147),
    [sym_boolean] = STATE(3144),
    [aux_sym_subroutine_declaration_statement_repeat1] = STATE(5104),
    [sym__identifier] = ACTIONS(457),
    [anon_sym_LBRACE] = ACTIONS(153),
    [anon_sym_DOLLAR] = ACTIONS(459),
    [anon_sym_AT] = ACTIONS(461),
    [anon_sym_my] = ACTIONS(463),
    [anon_sym_extended] = ACTIONS(31),
    [anon_sym_async] = ACTIONS(31),
    [anon_sym_sub] = ACTIONS(465),
    [anon_sym_method] = ACTIONS(467),
    [anon_sym_LPAREN] = ACTIONS(469),
    [anon_sym_continue] = ACTIONS(471),
    [anon_sym_state] = ACTIONS(463),
    [anon_sym_our] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(473),
    [anon_sym_LT_LT] = ACTIONS(2137),
    [anon_sym_SLASH_SLASH] = ACTIONS(2139),
    [anon_sym_PLUS] = ACTIONS(487),
    [anon_sym_DASH] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(487),
    [anon_sym_BANG] = ACTIONS(487),
    [anon_sym_PLUS_PLUS] = ACTIONS(2143),
    [anon_sym_DASH_DASH] = ACTIONS(2143),
    [anon_sym_BSLASH] = ACTIONS(492),
    [anon_sym_do] = ACTIONS(494),
    [anon_sym_eval] = ACTIONS(496),
    [anon_sym_await] = ACTIONS(498),
    [anon_sym_field] = ACTIONS(463),
    [anon_sym_local] = ACTIONS(500),
    [anon_sym_dynamically] = ACTIONS(500),
    [anon_sym_DOLLAR_POUND] = ACTIONS(502),
    [anon_sym_require] = ACTIONS(504),
    [anon_sym_map] = ACTIONS(506),
    [anon_sym_grep] = ACTIONS(506),
    [anon_sym_sort] = ACTIONS(508),
    [anon_sym_goto] = ACTIONS(510),
    [anon_sym_return] = ACTIONS(512),
    [anon_sym_undef] = ACTIONS(514),
    [anon_sym_PERCENT2] = ACTIONS(516),
    [anon_sym_AMP2] = ACTIONS(87),
    [anon_sym_STAR2] = ACTIONS(89),
    [anon_sym_last] = ACTIONS(518),
    [anon_sym_next] = ACTIONS(518),
    [anon_sym_redo] = ACTIONS(518),
    [anon_sym___FILE__] = ACTIONS(471),
    [anon_sym___LINE__] = ACTIONS(471),
    [anon_sym___PACKAGE__] = ACTIONS(471),
    [anon_sym___SUB__] = ACTIONS(471),
    [anon_sym_break] = ACTIONS(471),
    [anon_sym_fork] = ACTIONS(471),
    [anon_sym_getppid] = ACTIONS(471),
    [anon_sym_time] = ACTIONS(471),
    [anon_sym_times] = ACTIONS(471),
    [anon_sym_wait] = ACTIONS(471),
    [anon_sym_wantarray] = ACTIONS(471),
    [anon_sym_abs] = ACTIONS(520),
    [anon_sym_alarm] = ACTIONS(520),
    [anon_sym_chop] = ACTIONS(520),
    [anon_sym_chdir] = ACTIONS(520),
    [anon_sym_close] = ACTIONS(520),
    [anon_sym_closedir] = ACTIONS(520),
    [anon_sym_caller] = ACTIONS(520),
    [anon_sym_chomp] = ACTIONS(520),
    [anon_sym_chr] = ACTIONS(520),
    [anon_sym_cos] = ACTIONS(520),
    [anon_sym_chroot] = ACTIONS(520),
    [anon_sym_defined] = ACTIONS(520),
    [anon_sym_delete] = ACTIONS(520),
    [anon_sym_dbmclose] = ACTIONS(520),
    [anon_sym_exists] = ACTIONS(520),
    [anon_sym_exit] = ACTIONS(520),
    [anon_sym_eof] = ACTIONS(520),
    [anon_sym_exp] = ACTIONS(520),
    [anon_sym_each] = ACTIONS(520),
    [anon_sym_fc] = ACTIONS(520),
    [anon_sym_fileno] = ACTIONS(520),
    [anon_sym_gmtime] = ACTIONS(520),
    [anon_sym_getc] = ACTIONS(520),
    [anon_sym_getpgrp] = ACTIONS(520),
    [anon_sym_getprotobyname] = ACTIONS(520),
    [anon_sym_getpwname] = ACTIONS(520),
    [anon_sym_getpwuid] = ACTIONS(520),
    [anon_sym_getpeername] = ACTIONS(520),
    [anon_sym_getnetbyname] = ACTIONS(520),
    [anon_sym_getsockname] = ACTIONS(520),
    [anon_sym_getgrnam] = ACTIONS(520),
    [anon_sym_getgrgid] = ACTIONS(520),
    [anon_sym_hex] = ACTIONS(520),
    [anon_sym_int] = ACTIONS(520),
    [anon_sym_keys] = ACTIONS(520),
    [anon_sym_lc] = ACTIONS(520),
    [anon_sym_lcfirst] = ACTIONS(520),
    [anon_sym_length] = ACTIONS(520),
    [anon_sym_localtime] = ACTIONS(520),
    [anon_sym_log] = ACTIONS(520),
    [anon_sym_lock] = ACTIONS(520),
    [anon_sym_lstat] = ACTIONS(520),
    [anon_sym_oct] = ACTIONS(520),
    [anon_sym_ord] = ACTIONS(520),
    [anon_sym_prototype] = ACTIONS(520),
    [anon_sym_pop] = ACTIONS(520),
    [anon_sym_pos] = ACTIONS(520),
    [anon_sym_quotemeta] = ACTIONS(520),
    [anon_sym_reset] = ACTIONS(520),
    [anon_sym_rand] = ACTIONS(520),
    [anon_sym_rmdir] = ACTIONS(520),
    [anon_sym_readdir] = ACTIONS(520),
    [anon_sym_readline] = ACTIONS(520),
    [anon_sym_readpipe] = ACTIONS(520),
    [anon_sym_rewinddir] = ACTIONS(520),
    [anon_sym_readlink] = ACTIONS(520),
    [anon_sym_ref] = ACTIONS(520),
    [anon_sym_scalar] = ACTIONS(520),
    [anon_sym_shift] = ACTIONS(520),
    [anon_sym_sin] = ACTIONS(520),
    [anon_sym_sleep] = ACTIONS(520),
    [anon_sym_sqrt] = ACTIONS(520),
    [anon_sym_srand] = ACTIONS(520),
    [anon_sym_stat] = ACTIONS(520),
    [anon_sym_study] = ACTIONS(520),
    [anon_sym_tell] = ACTIONS(520),
    [anon_sym_telldir] = ACTIONS(520),
    [anon_sym_tied] = ACTIONS(520),
    [anon_sym_uc] = ACTIONS(520),
    [anon_sym_ucfirst] = ACTIONS(520),
    [anon_sym_untie] = ACTIONS(520),
    [anon_sym_umask] = ACTIONS(520),
    [anon_sym_values] = ACTIONS(520),
    [anon_sym_write] = ACTIONS(520),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(522),
    [anon_sym_qq] = ACTIONS(524),
    [anon_sym_qw] = ACTIONS(526),
    [anon_sym_qx] = ACTIONS(528),
    [anon_sym_qr] = ACTIONS(530),
    [anon_sym_m] = ACTIONS(532),
    [anon_sym_s] = ACTIONS(534),
    [anon_sym_tr] = ACTIONS(536),
    [anon_sym_y] = ACTIONS(536),
    [aux_sym__bareword_token1] = ACTIONS(457),
    [sym_number] = ACTIONS(538),
    [anon_sym_true] = ACTIONS(540),
    [anon_sym_false] = ACTIONS(540),
    [sym__single_quote] = ACTIONS(131),
    [sym__double_quote] = ACTIONS(133),
    [sym__backtick_quote] = ACTIONS(135),
    [sym__search_slash_quote] = ACTIONS(137),
    [sym__open_readline_bracket] = ACTIONS(542),
    [sym__open_fileglob_bracket] = ACTIONS(544),
    [sym__PERLY_HEREDOC] = ACTIONS(546),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(548),
    [sym__filetest] = ACTIONS(550),
  },
  [STATE(1132)] = {
    [sym_heredoc_content] = STATE(1132),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [anon_sym_if] = ACTIONS(2682),
    [anon_sym_unless] = ACTIONS(2682),
    [anon_sym_while] = ACTIONS(2682),
    [anon_sym_until] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1133)] = {
    [sym_heredoc_content] = STATE(1133),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1134)] = {
    [sym_heredoc_content] = STATE(1134),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1135)] = {
    [sym_heredoc_content] = STATE(1135),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_foreach] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2692),
    [anon_sym___END__] = ACTIONS(2692),
    [anon_sym_EOT] = ACTIONS(2692),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_unless] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_until] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym__ctrl_z_hack] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1136)] = {
    [sym_heredoc_content] = STATE(1136),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [anon_sym_if] = ACTIONS(2682),
    [anon_sym_unless] = ACTIONS(2682),
    [anon_sym_while] = ACTIONS(2682),
    [anon_sym_until] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1137)] = {
    [sym_heredoc_content] = STATE(1137),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_foreach] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2696),
    [anon_sym___END__] = ACTIONS(2696),
    [anon_sym_EOT] = ACTIONS(2696),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_unless] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_until] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym__ctrl_z_hack] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1138)] = {
    [sym_heredoc_content] = STATE(1138),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_for] = ACTIONS(2702),
    [anon_sym_foreach] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2702),
    [anon_sym___END__] = ACTIONS(2702),
    [anon_sym_EOT] = ACTIONS(2702),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2702),
    [anon_sym_unless] = ACTIONS(2702),
    [anon_sym_while] = ACTIONS(2702),
    [anon_sym_until] = ACTIONS(2702),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1139)] = {
    [sym_heredoc_content] = STATE(1139),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_for] = ACTIONS(2713),
    [anon_sym_foreach] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2713),
    [anon_sym___END__] = ACTIONS(2713),
    [anon_sym_EOT] = ACTIONS(2713),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [anon_sym_if] = ACTIONS(2713),
    [anon_sym_unless] = ACTIONS(2713),
    [anon_sym_while] = ACTIONS(2713),
    [anon_sym_until] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym__ctrl_z_hack] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1140)] = {
    [sym_heredoc_content] = STATE(1140),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_for] = ACTIONS(2717),
    [anon_sym_foreach] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2717),
    [anon_sym___END__] = ACTIONS(2717),
    [anon_sym_EOT] = ACTIONS(2717),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [anon_sym_if] = ACTIONS(2717),
    [anon_sym_unless] = ACTIONS(2717),
    [anon_sym_while] = ACTIONS(2717),
    [anon_sym_until] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym__ctrl_z_hack] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1141)] = {
    [sym_heredoc_content] = STATE(1141),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_for] = ACTIONS(2721),
    [anon_sym_foreach] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2721),
    [anon_sym___END__] = ACTIONS(2721),
    [anon_sym_EOT] = ACTIONS(2721),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [anon_sym_if] = ACTIONS(2721),
    [anon_sym_unless] = ACTIONS(2721),
    [anon_sym_while] = ACTIONS(2721),
    [anon_sym_until] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym__ctrl_z_hack] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1142)] = {
    [sym_heredoc_content] = STATE(1142),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [anon_sym_if] = ACTIONS(2682),
    [anon_sym_unless] = ACTIONS(2682),
    [anon_sym_while] = ACTIONS(2682),
    [anon_sym_until] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1143)] = {
    [sym_heredoc_content] = STATE(1143),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2741),
  },
  [STATE(1144)] = {
    [sym_heredoc_content] = STATE(1144),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1145)] = {
    [sym_heredoc_content] = STATE(1145),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_for] = ACTIONS(2721),
    [anon_sym_foreach] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2721),
    [anon_sym___END__] = ACTIONS(2721),
    [anon_sym_EOT] = ACTIONS(2721),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [anon_sym_if] = ACTIONS(2721),
    [anon_sym_unless] = ACTIONS(2721),
    [anon_sym_while] = ACTIONS(2721),
    [anon_sym_until] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym__ctrl_z_hack] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1146)] = {
    [sym_heredoc_content] = STATE(1146),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_foreach] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2696),
    [anon_sym___END__] = ACTIONS(2696),
    [anon_sym_EOT] = ACTIONS(2696),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [anon_sym_if] = ACTIONS(2696),
    [anon_sym_unless] = ACTIONS(2696),
    [anon_sym_while] = ACTIONS(2696),
    [anon_sym_until] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym__ctrl_z_hack] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1147)] = {
    [sym_heredoc_content] = STATE(1147),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_foreach] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2692),
    [anon_sym___END__] = ACTIONS(2692),
    [anon_sym_EOT] = ACTIONS(2692),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [anon_sym_if] = ACTIONS(2692),
    [anon_sym_unless] = ACTIONS(2692),
    [anon_sym_while] = ACTIONS(2692),
    [anon_sym_until] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym__ctrl_z_hack] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1148)] = {
    [sym_heredoc_content] = STATE(1148),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_for] = ACTIONS(2717),
    [anon_sym_foreach] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2717),
    [anon_sym___END__] = ACTIONS(2717),
    [anon_sym_EOT] = ACTIONS(2717),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [anon_sym_if] = ACTIONS(2717),
    [anon_sym_unless] = ACTIONS(2717),
    [anon_sym_while] = ACTIONS(2717),
    [anon_sym_until] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym__ctrl_z_hack] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1149)] = {
    [sym_heredoc_content] = STATE(1149),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1150)] = {
    [sym_heredoc_content] = STATE(1150),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1151)] = {
    [sym_heredoc_content] = STATE(1151),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_RPAREN] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_for] = ACTIONS(2727),
    [anon_sym_foreach] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2727),
    [anon_sym___END__] = ACTIONS(2727),
    [anon_sym_EOT] = ACTIONS(2727),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2727),
    [anon_sym_unless] = ACTIONS(2727),
    [anon_sym_while] = ACTIONS(2727),
    [anon_sym_until] = ACTIONS(2727),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2743),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2743),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1152)] = {
    [sym_heredoc_content] = STATE(1152),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_for] = ACTIONS(2713),
    [anon_sym_foreach] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2713),
    [anon_sym___END__] = ACTIONS(2713),
    [anon_sym_EOT] = ACTIONS(2713),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [anon_sym_if] = ACTIONS(2713),
    [anon_sym_unless] = ACTIONS(2713),
    [anon_sym_while] = ACTIONS(2713),
    [anon_sym_until] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym__ctrl_z_hack] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1153)] = {
    [sym_heredoc_content] = STATE(1153),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1154)] = {
    [sym_heredoc_content] = STATE(1154),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2745),
    [anon_sym_SEMI] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_RPAREN] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_for] = ACTIONS(2745),
    [anon_sym_foreach] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2745),
    [anon_sym___END__] = ACTIONS(2745),
    [anon_sym_EOT] = ACTIONS(2745),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [anon_sym_if] = ACTIONS(2745),
    [anon_sym_unless] = ACTIONS(2745),
    [anon_sym_while] = ACTIONS(2745),
    [anon_sym_until] = ACTIONS(2745),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2747),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym__ctrl_z_hack] = ACTIONS(2747),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1155)] = {
    [sym_heredoc_content] = STATE(1155),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2749),
    [anon_sym_SEMI] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_RPAREN] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_for] = ACTIONS(2749),
    [anon_sym_foreach] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2749),
    [anon_sym___END__] = ACTIONS(2749),
    [anon_sym_EOT] = ACTIONS(2749),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2749),
    [anon_sym_unless] = ACTIONS(2749),
    [anon_sym_while] = ACTIONS(2749),
    [anon_sym_until] = ACTIONS(2749),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2751),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2751),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1156)] = {
    [sym_heredoc_content] = STATE(1156),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2753),
  },
  [STATE(1157)] = {
    [sym_heredoc_content] = STATE(1157),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2755),
  },
  [STATE(1158)] = {
    [sym_heredoc_content] = STATE(1158),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2757),
  },
  [STATE(1159)] = {
    [sym_heredoc_content] = STATE(1159),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2759),
  },
  [STATE(1160)] = {
    [sym_heredoc_content] = STATE(1160),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2761),
  },
  [STATE(1161)] = {
    [sym_heredoc_content] = STATE(1161),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
  },
  [STATE(1162)] = {
    [sym_heredoc_content] = STATE(1162),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2534),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2534),
    [anon_sym___END__] = ACTIONS(2534),
    [anon_sym_EOT] = ACTIONS(2534),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_unless] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_until] = ACTIONS(2534),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1163)] = {
    [sym_heredoc_content] = STATE(1163),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2767),
  },
  [STATE(1164)] = {
    [sym_heredoc_content] = STATE(1164),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_for] = ACTIONS(2702),
    [anon_sym_foreach] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2702),
    [anon_sym___END__] = ACTIONS(2702),
    [anon_sym_EOT] = ACTIONS(2702),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [anon_sym_if] = ACTIONS(2702),
    [anon_sym_unless] = ACTIONS(2702),
    [anon_sym_while] = ACTIONS(2702),
    [anon_sym_until] = ACTIONS(2702),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1165)] = {
    [sym_heredoc_content] = STATE(1165),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2773),
  },
  [STATE(1166)] = {
    [sym_heredoc_content] = STATE(1166),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2775),
  },
  [STATE(1167)] = {
    [sym_heredoc_content] = STATE(1167),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2777),
  },
  [STATE(1168)] = {
    [sym_heredoc_content] = STATE(1168),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2779),
  },
  [STATE(1169)] = {
    [sym_heredoc_content] = STATE(1169),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2781),
  },
  [STATE(1170)] = {
    [sym_heredoc_content] = STATE(1170),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2783),
  },
  [STATE(1171)] = {
    [sym_heredoc_content] = STATE(1171),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2785),
  },
  [STATE(1172)] = {
    [sym_heredoc_content] = STATE(1172),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2787),
  },
  [STATE(1173)] = {
    [sym_heredoc_content] = STATE(1173),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2789),
  },
  [STATE(1174)] = {
    [sym_heredoc_content] = STATE(1174),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2791),
  },
  [STATE(1175)] = {
    [sym_heredoc_content] = STATE(1175),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2793),
  },
  [STATE(1176)] = {
    [sym_heredoc_content] = STATE(1176),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2795),
  },
  [STATE(1177)] = {
    [sym_heredoc_content] = STATE(1177),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2797),
  },
  [STATE(1178)] = {
    [sym_heredoc_content] = STATE(1178),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2799),
  },
  [STATE(1179)] = {
    [sym_heredoc_content] = STATE(1179),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2801),
  },
  [STATE(1180)] = {
    [sym_heredoc_content] = STATE(1180),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2803),
  },
  [STATE(1181)] = {
    [sym_heredoc_content] = STATE(1181),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2805),
  },
  [STATE(1182)] = {
    [sym_heredoc_content] = STATE(1182),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2807),
  },
  [STATE(1183)] = {
    [sym_heredoc_content] = STATE(1183),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2809),
  },
  [STATE(1184)] = {
    [sym_heredoc_content] = STATE(1184),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2811),
  },
  [STATE(1185)] = {
    [sym_heredoc_content] = STATE(1185),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2813),
  },
  [STATE(1186)] = {
    [sym_heredoc_content] = STATE(1186),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2815),
  },
  [STATE(1187)] = {
    [sym_heredoc_content] = STATE(1187),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2817),
  },
  [STATE(1188)] = {
    [sym_heredoc_content] = STATE(1188),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2819),
  },
  [STATE(1189)] = {
    [sym_heredoc_content] = STATE(1189),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2821),
  },
  [STATE(1190)] = {
    [sym_heredoc_content] = STATE(1190),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2823),
  },
  [STATE(1191)] = {
    [sym_heredoc_content] = STATE(1191),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2825),
  },
  [STATE(1192)] = {
    [sym_heredoc_content] = STATE(1192),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2827),
  },
  [STATE(1193)] = {
    [sym_heredoc_content] = STATE(1193),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2829),
  },
  [STATE(1194)] = {
    [sym_heredoc_content] = STATE(1194),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2831),
  },
  [STATE(1195)] = {
    [sym_heredoc_content] = STATE(1195),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [anon_sym_if] = ACTIONS(2682),
    [anon_sym_unless] = ACTIONS(2682),
    [anon_sym_while] = ACTIONS(2682),
    [anon_sym_until] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1196)] = {
    [sym_heredoc_content] = STATE(1196),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2791),
  },
  [STATE(1197)] = {
    [sym_heredoc_content] = STATE(1197),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2755),
  },
  [STATE(1198)] = {
    [sym_heredoc_content] = STATE(1198),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2757),
  },
  [STATE(1199)] = {
    [sym_heredoc_content] = STATE(1199),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2759),
  },
  [STATE(1200)] = {
    [sym_heredoc_content] = STATE(1200),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2761),
  },
  [STATE(1201)] = {
    [sym_heredoc_content] = STATE(1201),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2767),
  },
  [STATE(1202)] = {
    [sym_heredoc_content] = STATE(1202),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2773),
  },
  [STATE(1203)] = {
    [sym_heredoc_content] = STATE(1203),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2775),
  },
  [STATE(1204)] = {
    [sym_heredoc_content] = STATE(1204),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2777),
  },
  [STATE(1205)] = {
    [sym_heredoc_content] = STATE(1205),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2779),
  },
  [STATE(1206)] = {
    [sym_heredoc_content] = STATE(1206),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2781),
  },
  [STATE(1207)] = {
    [sym_heredoc_content] = STATE(1207),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2783),
  },
  [STATE(1208)] = {
    [sym_heredoc_content] = STATE(1208),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2785),
  },
  [STATE(1209)] = {
    [sym_heredoc_content] = STATE(1209),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2787),
  },
  [STATE(1210)] = {
    [sym_heredoc_content] = STATE(1210),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2789),
  },
  [STATE(1211)] = {
    [sym_heredoc_content] = STATE(1211),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2741),
  },
  [STATE(1212)] = {
    [sym_heredoc_content] = STATE(1212),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2753),
  },
  [STATE(1213)] = {
    [sym_heredoc_content] = STATE(1213),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2795),
  },
  [STATE(1214)] = {
    [sym_heredoc_content] = STATE(1214),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2797),
  },
  [STATE(1215)] = {
    [sym_heredoc_content] = STATE(1215),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2799),
  },
  [STATE(1216)] = {
    [sym_heredoc_content] = STATE(1216),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2801),
  },
  [STATE(1217)] = {
    [sym_heredoc_content] = STATE(1217),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2803),
  },
  [STATE(1218)] = {
    [sym_heredoc_content] = STATE(1218),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2805),
  },
  [STATE(1219)] = {
    [sym_heredoc_content] = STATE(1219),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2807),
  },
  [STATE(1220)] = {
    [sym_heredoc_content] = STATE(1220),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2809),
  },
  [STATE(1221)] = {
    [sym_heredoc_content] = STATE(1221),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2811),
  },
  [STATE(1222)] = {
    [sym_heredoc_content] = STATE(1222),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2813),
  },
  [STATE(1223)] = {
    [sym_heredoc_content] = STATE(1223),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2815),
  },
  [STATE(1224)] = {
    [sym_heredoc_content] = STATE(1224),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2817),
  },
  [STATE(1225)] = {
    [sym_heredoc_content] = STATE(1225),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2819),
  },
  [STATE(1226)] = {
    [sym_heredoc_content] = STATE(1226),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2821),
  },
  [STATE(1227)] = {
    [sym_heredoc_content] = STATE(1227),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2823),
  },
  [STATE(1228)] = {
    [sym_heredoc_content] = STATE(1228),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2825),
  },
  [STATE(1229)] = {
    [sym_heredoc_content] = STATE(1229),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2534),
    [anon_sym___END__] = ACTIONS(2534),
    [anon_sym_EOT] = ACTIONS(2534),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2534),
    [anon_sym_unless] = ACTIONS(2534),
    [anon_sym_while] = ACTIONS(2534),
    [anon_sym_until] = ACTIONS(2534),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1230)] = {
    [sym_heredoc_content] = STATE(1230),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2827),
  },
  [STATE(1231)] = {
    [sym_heredoc_content] = STATE(1231),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2829),
  },
  [STATE(1232)] = {
    [sym_heredoc_content] = STATE(1232),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2831),
  },
  [STATE(1233)] = {
    [sym_heredoc_content] = STATE(1233),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_for] = ACTIONS(2729),
    [anon_sym_foreach] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [anon_sym_if] = ACTIONS(2729),
    [anon_sym_unless] = ACTIONS(2729),
    [anon_sym_while] = ACTIONS(2729),
    [anon_sym_until] = ACTIONS(2729),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__brace_end_zw] = ACTIONS(2793),
  },
  [STATE(1234)] = {
    [sym_heredoc_content] = STATE(1234),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2702),
    [anon_sym___END__] = ACTIONS(2702),
    [anon_sym_EOT] = ACTIONS(2702),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1235)] = {
    [sym_heredoc_content] = STATE(1235),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1236)] = {
    [sym_heredoc_content] = STATE(1236),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2721),
    [anon_sym___END__] = ACTIONS(2721),
    [anon_sym_EOT] = ACTIONS(2721),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym__ctrl_z_hack] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1237)] = {
    [sym_heredoc_content] = STATE(1237),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2696),
    [anon_sym___END__] = ACTIONS(2696),
    [anon_sym_EOT] = ACTIONS(2696),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym__ctrl_z_hack] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1238)] = {
    [sym_heredoc_content] = STATE(1238),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2717),
    [anon_sym___END__] = ACTIONS(2717),
    [anon_sym_EOT] = ACTIONS(2717),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym__ctrl_z_hack] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1239)] = {
    [sym_heredoc_content] = STATE(1239),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2713),
    [anon_sym___END__] = ACTIONS(2713),
    [anon_sym_EOT] = ACTIONS(2713),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym__ctrl_z_hack] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1240)] = {
    [sym_heredoc_content] = STATE(1240),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1241)] = {
    [sym_heredoc_content] = STATE(1241),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2692),
    [anon_sym___END__] = ACTIONS(2692),
    [anon_sym_EOT] = ACTIONS(2692),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym__ctrl_z_hack] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1242)] = {
    [sym_heredoc_content] = STATE(1242),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1243)] = {
    [sym_heredoc_content] = STATE(1243),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1244)] = {
    [sym_heredoc_content] = STATE(1244),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2702),
    [anon_sym___END__] = ACTIONS(2702),
    [anon_sym_EOT] = ACTIONS(2702),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1245)] = {
    [sym_heredoc_content] = STATE(1245),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2696),
    [anon_sym___END__] = ACTIONS(2696),
    [anon_sym_EOT] = ACTIONS(2696),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym__ctrl_z_hack] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1246)] = {
    [sym_heredoc_content] = STATE(1246),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2713),
    [anon_sym___END__] = ACTIONS(2713),
    [anon_sym_EOT] = ACTIONS(2713),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym__ctrl_z_hack] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1247)] = {
    [sym_heredoc_content] = STATE(1247),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1248)] = {
    [sym_heredoc_content] = STATE(1248),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1249)] = {
    [sym_heredoc_content] = STATE(1249),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2749),
    [anon_sym___END__] = ACTIONS(2749),
    [anon_sym_EOT] = ACTIONS(2749),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2751),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2751),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1250)] = {
    [sym_heredoc_content] = STATE(1250),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2534),
    [anon_sym___END__] = ACTIONS(2534),
    [anon_sym_EOT] = ACTIONS(2534),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1251)] = {
    [sym_heredoc_content] = STATE(1251),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2727),
    [anon_sym___END__] = ACTIONS(2727),
    [anon_sym_EOT] = ACTIONS(2727),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2743),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2743),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1252)] = {
    [sym_heredoc_content] = STATE(1252),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2692),
    [anon_sym___END__] = ACTIONS(2692),
    [anon_sym_EOT] = ACTIONS(2692),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym__ctrl_z_hack] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1253)] = {
    [sym_heredoc_content] = STATE(1253),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
  },
  [STATE(1254)] = {
    [sym_heredoc_content] = STATE(1254),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2721),
    [anon_sym___END__] = ACTIONS(2721),
    [anon_sym_EOT] = ACTIONS(2721),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym__ctrl_z_hack] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1255)] = {
    [sym_heredoc_content] = STATE(1255),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2717),
    [anon_sym___END__] = ACTIONS(2717),
    [anon_sym_EOT] = ACTIONS(2717),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym__ctrl_z_hack] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1256)] = {
    [sym_heredoc_content] = STATE(1256),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1257)] = {
    [sym_heredoc_content] = STATE(1257),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2731),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2729),
    [anon_sym___END__] = ACTIONS(2729),
    [anon_sym_EOT] = ACTIONS(2729),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym__ctrl_z_hack] = ACTIONS(2739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1258)] = {
    [sym_heredoc_content] = STATE(1258),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2682),
    [anon_sym___END__] = ACTIONS(2682),
    [anon_sym_EOT] = ACTIONS(2682),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1259)] = {
    [sym_heredoc_content] = STATE(1259),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_SEMI] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2745),
    [anon_sym___END__] = ACTIONS(2745),
    [anon_sym_EOT] = ACTIONS(2745),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2747),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym__ctrl_z_hack] = ACTIONS(2747),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1260)] = {
    [sym_heredoc_content] = STATE(1260),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1261)] = {
    [sym_heredoc_content] = STATE(1261),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_RBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1262)] = {
    [sym_heredoc_content] = STATE(1262),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_RBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1263)] = {
    [sym_heredoc_content] = STATE(1263),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_RBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1264)] = {
    [sym_heredoc_content] = STATE(1264),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_RBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1265)] = {
    [sym_heredoc_content] = STATE(1265),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1266)] = {
    [sym_heredoc_content] = STATE(1266),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1267)] = {
    [sym_heredoc_content] = STATE(1267),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_RBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1268)] = {
    [sym_heredoc_content] = STATE(1268),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_RBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1269)] = {
    [sym_heredoc_content] = STATE(1269),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_RBRACK] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1270)] = {
    [sym_heredoc_content] = STATE(1270),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2694),
  },
  [STATE(1271)] = {
    [sym_heredoc_content] = STATE(1271),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(1272)] = {
    [sym_heredoc_content] = STATE(1272),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2723),
  },
  [STATE(1273)] = {
    [sym_heredoc_content] = STATE(1273),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1274)] = {
    [sym_heredoc_content] = STATE(1274),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2698),
  },
  [STATE(1275)] = {
    [sym_heredoc_content] = STATE(1275),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2723),
  },
  [STATE(1276)] = {
    [sym_heredoc_content] = STATE(1276),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_RPAREN] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2833),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2727),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1277)] = {
    [sym_heredoc_content] = STATE(1277),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_RBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1278)] = {
    [sym_heredoc_content] = STATE(1278),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_RBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2690),
  },
  [STATE(1279)] = {
    [sym_heredoc_content] = STATE(1279),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_RBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2723),
  },
  [STATE(1280)] = {
    [sym_heredoc_content] = STATE(1280),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2719),
  },
  [STATE(1281)] = {
    [sym_heredoc_content] = STATE(1281),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_RBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2698),
  },
  [STATE(1282)] = {
    [sym_heredoc_content] = STATE(1282),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2715),
  },
  [STATE(1283)] = {
    [sym_heredoc_content] = STATE(1283),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1284)] = {
    [sym_heredoc_content] = STATE(1284),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_RBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1285)] = {
    [sym_heredoc_content] = STATE(1285),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2719),
  },
  [STATE(1286)] = {
    [sym_heredoc_content] = STATE(1286),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(1287)] = {
    [sym_heredoc_content] = STATE(1287),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_RBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2719),
  },
  [STATE(1288)] = {
    [sym_heredoc_content] = STATE(1288),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_RBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1289)] = {
    [sym_heredoc_content] = STATE(1289),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(1290)] = {
    [sym_heredoc_content] = STATE(1290),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1291)] = {
    [sym_heredoc_content] = STATE(1291),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_RBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2715),
  },
  [STATE(1292)] = {
    [sym_heredoc_content] = STATE(1292),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2530),
  },
  [STATE(1293)] = {
    [sym_heredoc_content] = STATE(1293),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_RBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2694),
  },
  [STATE(1294)] = {
    [sym_heredoc_content] = STATE(1294),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_RBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1295)] = {
    [sym_heredoc_content] = STATE(1295),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(1296)] = {
    [sym_heredoc_content] = STATE(1296),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2711),
  },
  [STATE(1297)] = {
    [sym_heredoc_content] = STATE(1297),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1298)] = {
    [sym_heredoc_content] = STATE(1298),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2690),
  },
  [STATE(1299)] = {
    [sym_heredoc_content] = STATE(1299),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(1300)] = {
    [sym_heredoc_content] = STATE(1300),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2530),
  },
  [STATE(1301)] = {
    [sym_heredoc_content] = STATE(1301),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2711),
  },
  [STATE(1302)] = {
    [sym_heredoc_content] = STATE(1302),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2698),
  },
  [STATE(1303)] = {
    [sym_heredoc_content] = STATE(1303),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_RBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1304)] = {
    [sym_heredoc_content] = STATE(1304),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2690),
  },
  [STATE(1305)] = {
    [sym_heredoc_content] = STATE(1305),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_RBRACK] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2711),
  },
  [STATE(1306)] = {
    [sym_heredoc_content] = STATE(1306),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_RPAREN] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2833),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2749),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1307)] = {
    [sym_heredoc_content] = STATE(1307),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_RBRACK] = ACTIONS(2729),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
  },
  [STATE(1308)] = {
    [sym_heredoc_content] = STATE(1308),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_RBRACK] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1309)] = {
    [sym_heredoc_content] = STATE(1309),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2530),
  },
  [STATE(1310)] = {
    [sym_heredoc_content] = STATE(1310),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2833),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2729),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1311)] = {
    [sym_heredoc_content] = STATE(1311),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2715),
  },
  [STATE(1312)] = {
    [sym_heredoc_content] = STATE(1312),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_RPAREN] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_RBRACK] = ACTIONS(2745),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1313)] = {
    [sym_heredoc_content] = STATE(1313),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(1314)] = {
    [sym_heredoc_content] = STATE(1314),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_RBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1315)] = {
    [sym_heredoc_content] = STATE(1315),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2694),
  },
  [STATE(1316)] = {
    [sym_heredoc_content] = STATE(1316),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2690),
  },
  [STATE(1317)] = {
    [sym_heredoc_content] = STATE(1317),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(1318)] = {
    [sym_heredoc_content] = STATE(1318),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2743),
  },
  [STATE(1319)] = {
    [sym_heredoc_content] = STATE(1319),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2839),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2739),
  },
  [STATE(1320)] = {
    [sym_heredoc_content] = STATE(1320),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_RPAREN] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2839),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2751),
  },
  [STATE(1321)] = {
    [sym_heredoc_content] = STATE(1321),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2739),
  },
  [STATE(1322)] = {
    [sym_heredoc_content] = STATE(1322),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2711),
  },
  [STATE(1323)] = {
    [sym_heredoc_content] = STATE(1323),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2536),
  },
  [STATE(1324)] = {
    [sym_heredoc_content] = STATE(1324),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2723),
  },
  [STATE(1325)] = {
    [sym_heredoc_content] = STATE(1325),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2698),
  },
  [STATE(1326)] = {
    [sym_heredoc_content] = STATE(1326),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2719),
  },
  [STATE(1327)] = {
    [sym_heredoc_content] = STATE(1327),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2715),
  },
  [STATE(1328)] = {
    [sym_heredoc_content] = STATE(1328),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2723),
  },
  [STATE(1329)] = {
    [sym_heredoc_content] = STATE(1329),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_RPAREN] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2747),
  },
  [STATE(1330)] = {
    [sym_heredoc_content] = STATE(1330),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2698),
  },
  [STATE(1331)] = {
    [sym_heredoc_content] = STATE(1331),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(1332)] = {
    [sym_heredoc_content] = STATE(1332),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2719),
  },
  [STATE(1333)] = {
    [sym_heredoc_content] = STATE(1333),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2715),
  },
  [STATE(1334)] = {
    [sym_heredoc_content] = STATE(1334),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(1335)] = {
    [sym_heredoc_content] = STATE(1335),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2690),
  },
  [STATE(1336)] = {
    [sym_heredoc_content] = STATE(1336),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2694),
  },
  [STATE(1337)] = {
    [sym_heredoc_content] = STATE(1337),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2694),
  },
  [STATE(1338)] = {
    [sym_heredoc_content] = STATE(1338),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2530),
  },
  [STATE(1339)] = {
    [sym_heredoc_content] = STATE(1339),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2530),
  },
  [STATE(1340)] = {
    [sym_heredoc_content] = STATE(1340),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_RPAREN] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2839),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2743),
  },
  [STATE(1341)] = {
    [sym_heredoc_content] = STATE(1341),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(1342)] = {
    [sym_heredoc_content] = STATE(1342),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2842),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2729),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2739),
  },
  [STATE(1343)] = {
    [sym_heredoc_content] = STATE(1343),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(1344)] = {
    [sym_heredoc_content] = STATE(1344),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_RBRACK] = ACTIONS(2729),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2739),
  },
  [STATE(1345)] = {
    [sym_heredoc_content] = STATE(1345),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_RBRACK] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2711),
  },
  [STATE(1346)] = {
    [sym_heredoc_content] = STATE(1346),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_RBRACK] = ACTIONS(2745),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2747),
  },
  [STATE(1347)] = {
    [sym_heredoc_content] = STATE(1347),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2842),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_and] = ACTIONS(2727),
    [anon_sym_or] = ACTIONS(2727),
    [anon_sym_xor] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2727),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2743),
  },
  [STATE(1348)] = {
    [sym_heredoc_content] = STATE(1348),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2536),
  },
  [STATE(1349)] = {
    [sym_heredoc_content] = STATE(1349),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_and] = ACTIONS(2721),
    [anon_sym_or] = ACTIONS(2721),
    [anon_sym_xor] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_RBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2723),
  },
  [STATE(1350)] = {
    [sym_heredoc_content] = STATE(1350),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_and] = ACTIONS(2696),
    [anon_sym_or] = ACTIONS(2696),
    [anon_sym_xor] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_RBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2698),
  },
  [STATE(1351)] = {
    [sym_heredoc_content] = STATE(1351),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2739),
  },
  [STATE(1352)] = {
    [sym_heredoc_content] = STATE(1352),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_and] = ACTIONS(2688),
    [anon_sym_or] = ACTIONS(2688),
    [anon_sym_xor] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_RBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2690),
  },
  [STATE(1353)] = {
    [sym_heredoc_content] = STATE(1353),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2836),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2751),
  },
  [STATE(1354)] = {
    [sym_heredoc_content] = STATE(1354),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_and] = ACTIONS(2729),
    [anon_sym_or] = ACTIONS(2729),
    [anon_sym_xor] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2739),
  },
  [STATE(1355)] = {
    [sym_heredoc_content] = STATE(1355),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_xor] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_RBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2719),
  },
  [STATE(1356)] = {
    [sym_heredoc_content] = STATE(1356),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_and] = ACTIONS(2702),
    [anon_sym_or] = ACTIONS(2702),
    [anon_sym_xor] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2711),
  },
  [STATE(1357)] = {
    [sym_heredoc_content] = STATE(1357),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_and] = ACTIONS(2713),
    [anon_sym_or] = ACTIONS(2713),
    [anon_sym_xor] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_RBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2715),
  },
  [STATE(1358)] = {
    [sym_heredoc_content] = STATE(1358),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_and] = ACTIONS(2534),
    [anon_sym_or] = ACTIONS(2534),
    [anon_sym_xor] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2536),
  },
  [STATE(1359)] = {
    [sym_heredoc_content] = STATE(1359),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_and] = ACTIONS(2692),
    [anon_sym_or] = ACTIONS(2692),
    [anon_sym_xor] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_RBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2694),
  },
  [STATE(1360)] = {
    [sym_heredoc_content] = STATE(1360),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_and] = ACTIONS(2682),
    [anon_sym_or] = ACTIONS(2682),
    [anon_sym_xor] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_RBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(1361)] = {
    [sym_heredoc_content] = STATE(1361),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_and] = ACTIONS(2745),
    [anon_sym_or] = ACTIONS(2745),
    [anon_sym_xor] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2747),
  },
  [STATE(1362)] = {
    [sym_heredoc_content] = STATE(1362),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_and] = ACTIONS(2532),
    [anon_sym_or] = ACTIONS(2532),
    [anon_sym_xor] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_RBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2530),
  },
  [STATE(1363)] = {
    [sym_heredoc_content] = STATE(1363),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2842),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_and] = ACTIONS(2749),
    [anon_sym_or] = ACTIONS(2749),
    [anon_sym_xor] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_RBRACK] = ACTIONS(2749),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2751),
  },
  [STATE(1364)] = {
    [sym_heredoc_content] = STATE(1364),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_COLON] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1365)] = {
    [sym_heredoc_content] = STATE(1365),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_COLON] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1366)] = {
    [sym_heredoc_content] = STATE(1366),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1367)] = {
    [sym_heredoc_content] = STATE(1367),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_COLON] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1368)] = {
    [sym_heredoc_content] = STATE(1368),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_COLON] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1369)] = {
    [sym_heredoc_content] = STATE(1369),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_COLON] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1370)] = {
    [sym_heredoc_content] = STATE(1370),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_COLON] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1371)] = {
    [sym_heredoc_content] = STATE(1371),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_COLON] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1372)] = {
    [sym_heredoc_content] = STATE(1372),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_COLON] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1373)] = {
    [sym_heredoc_content] = STATE(1373),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_COLON] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1374)] = {
    [sym_heredoc_content] = STATE(1374),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_COLON] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2845),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1375)] = {
    [sym_heredoc_content] = STATE(1375),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2534),
    [anon_sym_COLON] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1376)] = {
    [sym_heredoc_content] = STATE(1376),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1377)] = {
    [sym_heredoc_content] = STATE(1377),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_COLON] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2704),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2704),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2704),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1378)] = {
    [sym_heredoc_content] = STATE(1378),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__no_search_slash_plz] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1379)] = {
    [sym_heredoc_content] = STATE(1379),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_COLON] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1380)] = {
    [sym_heredoc_content] = STATE(1380),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2749),
    [anon_sym_COLON] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_RPAREN] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2845),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1381)] = {
    [sym_heredoc_content] = STATE(1381),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_RBRACE] = ACTIONS(2721),
    [anon_sym_COLON] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_RPAREN] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1382)] = {
    [sym_heredoc_content] = STATE(1382),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_RBRACE] = ACTIONS(2696),
    [anon_sym_COLON] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_RPAREN] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1383)] = {
    [sym_heredoc_content] = STATE(1383),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_RBRACE] = ACTIONS(2729),
    [anon_sym_COLON] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
  },
  [STATE(1384)] = {
    [sym_heredoc_content] = STATE(1384),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_COLON] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1385)] = {
    [sym_heredoc_content] = STATE(1385),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_RBRACE] = ACTIONS(2702),
    [anon_sym_COLON] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_RPAREN] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1386)] = {
    [sym_heredoc_content] = STATE(1386),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_RBRACE] = ACTIONS(2745),
    [anon_sym_COLON] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_RPAREN] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1387)] = {
    [sym_heredoc_content] = STATE(1387),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_RBRACE] = ACTIONS(2717),
    [anon_sym_COLON] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1388)] = {
    [sym_heredoc_content] = STATE(1388),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__no_search_slash_plz] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1389)] = {
    [sym_heredoc_content] = STATE(1389),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_COLON] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1390)] = {
    [sym_heredoc_content] = STATE(1390),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_RBRACE] = ACTIONS(2713),
    [anon_sym_COLON] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_RPAREN] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1391)] = {
    [sym_heredoc_content] = STATE(1391),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__no_search_slash_plz] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1392)] = {
    [sym_heredoc_content] = STATE(1392),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__no_search_slash_plz] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1393)] = {
    [sym_heredoc_content] = STATE(1393),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__no_search_slash_plz] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1394)] = {
    [sym_heredoc_content] = STATE(1394),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__no_search_slash_plz] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1395)] = {
    [sym_heredoc_content] = STATE(1395),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_RBRACE] = ACTIONS(2692),
    [anon_sym_COLON] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_RPAREN] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1396)] = {
    [sym_heredoc_content] = STATE(1396),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_RBRACE] = ACTIONS(2682),
    [anon_sym_COLON] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_RPAREN] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1397)] = {
    [sym_heredoc_content] = STATE(1397),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2704),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2704),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2704),
    [anon_sym_DASH] = ACTIONS(2704),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2704),
    [anon_sym_DASH_DASH] = ACTIONS(2704),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__no_search_slash_plz] = ACTIONS(2709),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1398)] = {
    [sym_heredoc_content] = STATE(1398),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_RBRACE] = ACTIONS(2727),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_RPAREN] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2845),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1399)] = {
    [sym_heredoc_content] = STATE(1399),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__no_search_slash_plz] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1400)] = {
    [sym_heredoc_content] = STATE(1400),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__no_search_slash_plz] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1401)] = {
    [sym_heredoc_content] = STATE(1401),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_SEMI] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_SEMICOLON] = ACTIONS(2747),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1402)] = {
    [sym_heredoc_content] = STATE(1402),
    [sym__identifier] = ACTIONS(2763),
    [anon_sym_LBRACE] = ACTIONS(2763),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2763),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2763),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2763),
    [anon_sym_extended] = ACTIONS(2763),
    [anon_sym_async] = ACTIONS(2763),
    [anon_sym_sub] = ACTIONS(2763),
    [anon_sym_method] = ACTIONS(2763),
    [anon_sym_LPAREN] = ACTIONS(2763),
    [anon_sym_continue] = ACTIONS(2763),
    [anon_sym_state] = ACTIONS(2763),
    [anon_sym_our] = ACTIONS(2763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2763),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2763),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2763),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2763),
    [anon_sym_DASH] = ACTIONS(2763),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2763),
    [anon_sym_BANG] = ACTIONS(2763),
    [anon_sym_PLUS_PLUS] = ACTIONS(2763),
    [anon_sym_DASH_DASH] = ACTIONS(2763),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2765),
    [anon_sym_do] = ACTIONS(2763),
    [anon_sym_eval] = ACTIONS(2763),
    [anon_sym_await] = ACTIONS(2763),
    [anon_sym_field] = ACTIONS(2763),
    [anon_sym_local] = ACTIONS(2763),
    [anon_sym_dynamically] = ACTIONS(2763),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2763),
    [anon_sym_require] = ACTIONS(2763),
    [anon_sym_map] = ACTIONS(2763),
    [anon_sym_grep] = ACTIONS(2763),
    [anon_sym_sort] = ACTIONS(2763),
    [anon_sym_goto] = ACTIONS(2763),
    [anon_sym_return] = ACTIONS(2763),
    [anon_sym_undef] = ACTIONS(2763),
    [anon_sym_PERCENT2] = ACTIONS(2763),
    [anon_sym_AMP2] = ACTIONS(2763),
    [anon_sym_STAR2] = ACTIONS(2763),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2763),
    [anon_sym_next] = ACTIONS(2763),
    [anon_sym_redo] = ACTIONS(2763),
    [anon_sym___FILE__] = ACTIONS(2763),
    [anon_sym___LINE__] = ACTIONS(2763),
    [anon_sym___PACKAGE__] = ACTIONS(2763),
    [anon_sym___SUB__] = ACTIONS(2763),
    [anon_sym_break] = ACTIONS(2763),
    [anon_sym_fork] = ACTIONS(2763),
    [anon_sym_getppid] = ACTIONS(2763),
    [anon_sym_time] = ACTIONS(2763),
    [anon_sym_times] = ACTIONS(2763),
    [anon_sym_wait] = ACTIONS(2763),
    [anon_sym_wantarray] = ACTIONS(2763),
    [anon_sym_abs] = ACTIONS(2763),
    [anon_sym_alarm] = ACTIONS(2763),
    [anon_sym_chop] = ACTIONS(2763),
    [anon_sym_chdir] = ACTIONS(2763),
    [anon_sym_close] = ACTIONS(2763),
    [anon_sym_closedir] = ACTIONS(2763),
    [anon_sym_caller] = ACTIONS(2763),
    [anon_sym_chomp] = ACTIONS(2763),
    [anon_sym_chr] = ACTIONS(2763),
    [anon_sym_cos] = ACTIONS(2763),
    [anon_sym_chroot] = ACTIONS(2763),
    [anon_sym_defined] = ACTIONS(2763),
    [anon_sym_delete] = ACTIONS(2763),
    [anon_sym_dbmclose] = ACTIONS(2763),
    [anon_sym_exists] = ACTIONS(2763),
    [anon_sym_exit] = ACTIONS(2763),
    [anon_sym_eof] = ACTIONS(2763),
    [anon_sym_exp] = ACTIONS(2763),
    [anon_sym_each] = ACTIONS(2763),
    [anon_sym_fc] = ACTIONS(2763),
    [anon_sym_fileno] = ACTIONS(2763),
    [anon_sym_gmtime] = ACTIONS(2763),
    [anon_sym_getc] = ACTIONS(2763),
    [anon_sym_getpgrp] = ACTIONS(2763),
    [anon_sym_getprotobyname] = ACTIONS(2763),
    [anon_sym_getpwname] = ACTIONS(2763),
    [anon_sym_getpwuid] = ACTIONS(2763),
    [anon_sym_getpeername] = ACTIONS(2763),
    [anon_sym_getnetbyname] = ACTIONS(2763),
    [anon_sym_getsockname] = ACTIONS(2763),
    [anon_sym_getgrnam] = ACTIONS(2763),
    [anon_sym_getgrgid] = ACTIONS(2763),
    [anon_sym_hex] = ACTIONS(2763),
    [anon_sym_int] = ACTIONS(2763),
    [anon_sym_keys] = ACTIONS(2763),
    [anon_sym_lc] = ACTIONS(2763),
    [anon_sym_lcfirst] = ACTIONS(2763),
    [anon_sym_length] = ACTIONS(2763),
    [anon_sym_localtime] = ACTIONS(2763),
    [anon_sym_log] = ACTIONS(2763),
    [anon_sym_lock] = ACTIONS(2763),
    [anon_sym_lstat] = ACTIONS(2763),
    [anon_sym_oct] = ACTIONS(2763),
    [anon_sym_ord] = ACTIONS(2763),
    [anon_sym_prototype] = ACTIONS(2763),
    [anon_sym_pop] = ACTIONS(2763),
    [anon_sym_pos] = ACTIONS(2763),
    [anon_sym_quotemeta] = ACTIONS(2763),
    [anon_sym_reset] = ACTIONS(2763),
    [anon_sym_rand] = ACTIONS(2763),
    [anon_sym_rmdir] = ACTIONS(2763),
    [anon_sym_readdir] = ACTIONS(2763),
    [anon_sym_readline] = ACTIONS(2763),
    [anon_sym_readpipe] = ACTIONS(2763),
    [anon_sym_rewinddir] = ACTIONS(2763),
    [anon_sym_readlink] = ACTIONS(2763),
    [anon_sym_ref] = ACTIONS(2763),
    [anon_sym_scalar] = ACTIONS(2763),
    [anon_sym_shift] = ACTIONS(2763),
    [anon_sym_sin] = ACTIONS(2763),
    [anon_sym_sleep] = ACTIONS(2763),
    [anon_sym_sqrt] = ACTIONS(2763),
    [anon_sym_srand] = ACTIONS(2763),
    [anon_sym_stat] = ACTIONS(2763),
    [anon_sym_study] = ACTIONS(2763),
    [anon_sym_tell] = ACTIONS(2763),
    [anon_sym_telldir] = ACTIONS(2763),
    [anon_sym_tied] = ACTIONS(2763),
    [anon_sym_uc] = ACTIONS(2763),
    [anon_sym_ucfirst] = ACTIONS(2763),
    [anon_sym_untie] = ACTIONS(2763),
    [anon_sym_umask] = ACTIONS(2763),
    [anon_sym_values] = ACTIONS(2763),
    [anon_sym_write] = ACTIONS(2763),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2763),
    [anon_sym_qq] = ACTIONS(2763),
    [anon_sym_qw] = ACTIONS(2763),
    [anon_sym_qx] = ACTIONS(2763),
    [anon_sym_qr] = ACTIONS(2763),
    [anon_sym_m] = ACTIONS(2763),
    [anon_sym_s] = ACTIONS(2763),
    [anon_sym_tr] = ACTIONS(2763),
    [anon_sym_y] = ACTIONS(2763),
    [aux_sym__bareword_token1] = ACTIONS(2763),
    [sym_number] = ACTIONS(2763),
    [anon_sym_true] = ACTIONS(2763),
    [anon_sym_false] = ACTIONS(2763),
    [sym__single_quote] = ACTIONS(2765),
    [sym__double_quote] = ACTIONS(2765),
    [sym__backtick_quote] = ACTIONS(2765),
    [sym__search_slash_quote] = ACTIONS(2765),
    [sym__open_readline_bracket] = ACTIONS(2765),
    [sym__open_fileglob_bracket] = ACTIONS(2765),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2765),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2765),
    [sym__filetest] = ACTIONS(2765),
  },
  [STATE(1403)] = {
    [sym_heredoc_content] = STATE(1403),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2749),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2749),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_RPAREN] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2749),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1404)] = {
    [sym_heredoc_content] = STATE(1404),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_COMMA] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_DASH_GT] = ACTIONS(2688),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2688),
    [anon_sym_PLUS_EQ] = ACTIONS(2688),
    [anon_sym_DASH_EQ] = ACTIONS(2688),
    [anon_sym_DOT_EQ] = ACTIONS(2688),
    [anon_sym_STAR_EQ] = ACTIONS(2688),
    [anon_sym_SLASH_EQ] = ACTIONS(2688),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2688),
    [anon_sym_AMP_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_EQ] = ACTIONS(2688),
    [anon_sym_CARET_EQ] = ACTIONS(2688),
    [anon_sym_LT_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_GT_EQ] = ACTIONS(2688),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2688),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2688),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2688),
    [anon_sym_BANG_TILDE] = ACTIONS(2688),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2688),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2688),
    [anon_sym_equ] = ACTIONS(2688),
    [anon_sym_eqr] = ACTIONS(2688),
    [anon_sym_ne] = ACTIONS(2688),
    [anon_sym_LT_EQ_GT] = ACTIONS(2688),
    [anon_sym_cmp] = ACTIONS(2688),
    [anon_sym_TILDE_TILDE] = ACTIONS(2688),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2688),
    [anon_sym_lt] = ACTIONS(2688),
    [anon_sym_le] = ACTIONS(2688),
    [anon_sym_ge] = ACTIONS(2688),
    [anon_sym_gt] = ACTIONS(2688),
    [anon_sym_isa] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_QMARK] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_EQ_GT] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1405)] = {
    [sym_heredoc_content] = STATE(1405),
    [sym__identifier] = ACTIONS(2692),
    [anon_sym_LBRACE] = ACTIONS(2692),
    [anon_sym_SEMI] = ACTIONS(2692),
    [anon_sym_DOLLAR] = ACTIONS(2692),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_AT] = ACTIONS(2692),
    [anon_sym_COMMA] = ACTIONS(2692),
    [anon_sym_my] = ACTIONS(2692),
    [anon_sym_extended] = ACTIONS(2692),
    [anon_sym_async] = ACTIONS(2692),
    [anon_sym_sub] = ACTIONS(2692),
    [anon_sym_method] = ACTIONS(2692),
    [anon_sym_LPAREN] = ACTIONS(2692),
    [anon_sym_continue] = ACTIONS(2692),
    [anon_sym_state] = ACTIONS(2692),
    [anon_sym_our] = ACTIONS(2692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2692),
    [anon_sym_DASH_GT] = ACTIONS(2692),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2692),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2692),
    [anon_sym_PLUS_EQ] = ACTIONS(2692),
    [anon_sym_DASH_EQ] = ACTIONS(2692),
    [anon_sym_DOT_EQ] = ACTIONS(2692),
    [anon_sym_STAR_EQ] = ACTIONS(2692),
    [anon_sym_SLASH_EQ] = ACTIONS(2692),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2692),
    [anon_sym_AMP_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_EQ] = ACTIONS(2692),
    [anon_sym_CARET_EQ] = ACTIONS(2692),
    [anon_sym_LT_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_GT_EQ] = ACTIONS(2692),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2692),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2692),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2692),
    [anon_sym_BANG_TILDE] = ACTIONS(2692),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2692),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2692),
    [anon_sym_equ] = ACTIONS(2692),
    [anon_sym_eqr] = ACTIONS(2692),
    [anon_sym_ne] = ACTIONS(2692),
    [anon_sym_LT_EQ_GT] = ACTIONS(2692),
    [anon_sym_cmp] = ACTIONS(2692),
    [anon_sym_TILDE_TILDE] = ACTIONS(2692),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2692),
    [anon_sym_lt] = ACTIONS(2692),
    [anon_sym_le] = ACTIONS(2692),
    [anon_sym_ge] = ACTIONS(2692),
    [anon_sym_gt] = ACTIONS(2692),
    [anon_sym_isa] = ACTIONS(2692),
    [anon_sym_TILDE] = ACTIONS(2692),
    [anon_sym_BANG] = ACTIONS(2692),
    [anon_sym_PLUS_PLUS] = ACTIONS(2692),
    [anon_sym_DASH_DASH] = ACTIONS(2692),
    [anon_sym_QMARK] = ACTIONS(2692),
    [anon_sym_BSLASH] = ACTIONS(2694),
    [anon_sym_do] = ACTIONS(2692),
    [anon_sym_eval] = ACTIONS(2692),
    [anon_sym_await] = ACTIONS(2692),
    [anon_sym_field] = ACTIONS(2692),
    [anon_sym_local] = ACTIONS(2692),
    [anon_sym_dynamically] = ACTIONS(2692),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2692),
    [anon_sym_require] = ACTIONS(2692),
    [anon_sym_map] = ACTIONS(2692),
    [anon_sym_grep] = ACTIONS(2692),
    [anon_sym_sort] = ACTIONS(2692),
    [anon_sym_goto] = ACTIONS(2692),
    [anon_sym_return] = ACTIONS(2692),
    [anon_sym_undef] = ACTIONS(2692),
    [anon_sym_PERCENT2] = ACTIONS(2692),
    [anon_sym_AMP2] = ACTIONS(2692),
    [anon_sym_STAR2] = ACTIONS(2692),
    [anon_sym_EQ_GT] = ACTIONS(2692),
    [anon_sym_last] = ACTIONS(2692),
    [anon_sym_next] = ACTIONS(2692),
    [anon_sym_redo] = ACTIONS(2692),
    [anon_sym___FILE__] = ACTIONS(2692),
    [anon_sym___LINE__] = ACTIONS(2692),
    [anon_sym___PACKAGE__] = ACTIONS(2692),
    [anon_sym___SUB__] = ACTIONS(2692),
    [anon_sym_break] = ACTIONS(2692),
    [anon_sym_fork] = ACTIONS(2692),
    [anon_sym_getppid] = ACTIONS(2692),
    [anon_sym_time] = ACTIONS(2692),
    [anon_sym_times] = ACTIONS(2692),
    [anon_sym_wait] = ACTIONS(2692),
    [anon_sym_wantarray] = ACTIONS(2692),
    [anon_sym_abs] = ACTIONS(2692),
    [anon_sym_alarm] = ACTIONS(2692),
    [anon_sym_chop] = ACTIONS(2692),
    [anon_sym_chdir] = ACTIONS(2692),
    [anon_sym_close] = ACTIONS(2692),
    [anon_sym_closedir] = ACTIONS(2692),
    [anon_sym_caller] = ACTIONS(2692),
    [anon_sym_chomp] = ACTIONS(2692),
    [anon_sym_chr] = ACTIONS(2692),
    [anon_sym_cos] = ACTIONS(2692),
    [anon_sym_chroot] = ACTIONS(2692),
    [anon_sym_defined] = ACTIONS(2692),
    [anon_sym_delete] = ACTIONS(2692),
    [anon_sym_dbmclose] = ACTIONS(2692),
    [anon_sym_exists] = ACTIONS(2692),
    [anon_sym_exit] = ACTIONS(2692),
    [anon_sym_eof] = ACTIONS(2692),
    [anon_sym_exp] = ACTIONS(2692),
    [anon_sym_each] = ACTIONS(2692),
    [anon_sym_fc] = ACTIONS(2692),
    [anon_sym_fileno] = ACTIONS(2692),
    [anon_sym_gmtime] = ACTIONS(2692),
    [anon_sym_getc] = ACTIONS(2692),
    [anon_sym_getpgrp] = ACTIONS(2692),
    [anon_sym_getprotobyname] = ACTIONS(2692),
    [anon_sym_getpwname] = ACTIONS(2692),
    [anon_sym_getpwuid] = ACTIONS(2692),
    [anon_sym_getpeername] = ACTIONS(2692),
    [anon_sym_getnetbyname] = ACTIONS(2692),
    [anon_sym_getsockname] = ACTIONS(2692),
    [anon_sym_getgrnam] = ACTIONS(2692),
    [anon_sym_getgrgid] = ACTIONS(2692),
    [anon_sym_hex] = ACTIONS(2692),
    [anon_sym_int] = ACTIONS(2692),
    [anon_sym_keys] = ACTIONS(2692),
    [anon_sym_lc] = ACTIONS(2692),
    [anon_sym_lcfirst] = ACTIONS(2692),
    [anon_sym_length] = ACTIONS(2692),
    [anon_sym_localtime] = ACTIONS(2692),
    [anon_sym_log] = ACTIONS(2692),
    [anon_sym_lock] = ACTIONS(2692),
    [anon_sym_lstat] = ACTIONS(2692),
    [anon_sym_oct] = ACTIONS(2692),
    [anon_sym_ord] = ACTIONS(2692),
    [anon_sym_prototype] = ACTIONS(2692),
    [anon_sym_pop] = ACTIONS(2692),
    [anon_sym_pos] = ACTIONS(2692),
    [anon_sym_quotemeta] = ACTIONS(2692),
    [anon_sym_reset] = ACTIONS(2692),
    [anon_sym_rand] = ACTIONS(2692),
    [anon_sym_rmdir] = ACTIONS(2692),
    [anon_sym_readdir] = ACTIONS(2692),
    [anon_sym_readline] = ACTIONS(2692),
    [anon_sym_readpipe] = ACTIONS(2692),
    [anon_sym_rewinddir] = ACTIONS(2692),
    [anon_sym_readlink] = ACTIONS(2692),
    [anon_sym_ref] = ACTIONS(2692),
    [anon_sym_scalar] = ACTIONS(2692),
    [anon_sym_shift] = ACTIONS(2692),
    [anon_sym_sin] = ACTIONS(2692),
    [anon_sym_sleep] = ACTIONS(2692),
    [anon_sym_sqrt] = ACTIONS(2692),
    [anon_sym_srand] = ACTIONS(2692),
    [anon_sym_stat] = ACTIONS(2692),
    [anon_sym_study] = ACTIONS(2692),
    [anon_sym_tell] = ACTIONS(2692),
    [anon_sym_telldir] = ACTIONS(2692),
    [anon_sym_tied] = ACTIONS(2692),
    [anon_sym_uc] = ACTIONS(2692),
    [anon_sym_ucfirst] = ACTIONS(2692),
    [anon_sym_untie] = ACTIONS(2692),
    [anon_sym_umask] = ACTIONS(2692),
    [anon_sym_values] = ACTIONS(2692),
    [anon_sym_write] = ACTIONS(2692),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2692),
    [anon_sym_qq] = ACTIONS(2692),
    [anon_sym_qw] = ACTIONS(2692),
    [anon_sym_qx] = ACTIONS(2692),
    [anon_sym_qr] = ACTIONS(2692),
    [anon_sym_m] = ACTIONS(2692),
    [anon_sym_s] = ACTIONS(2692),
    [anon_sym_tr] = ACTIONS(2692),
    [anon_sym_y] = ACTIONS(2692),
    [aux_sym__bareword_token1] = ACTIONS(2692),
    [sym_number] = ACTIONS(2692),
    [anon_sym_true] = ACTIONS(2692),
    [anon_sym_false] = ACTIONS(2692),
    [sym__single_quote] = ACTIONS(2694),
    [sym__double_quote] = ACTIONS(2694),
    [sym__backtick_quote] = ACTIONS(2694),
    [sym__search_slash_quote] = ACTIONS(2694),
    [sym__open_readline_bracket] = ACTIONS(2694),
    [sym__open_fileglob_bracket] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__PERLY_HEREDOC] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2694),
    [sym__filetest] = ACTIONS(2694),
  },
  [STATE(1406)] = {
    [sym_heredoc_content] = STATE(1406),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2851),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2743),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1407)] = {
    [sym_heredoc_content] = STATE(1407),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_COLON] = ACTIONS(2745),
    [anon_sym_DOLLAR] = ACTIONS(2745),
    [anon_sym_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_AT] = ACTIONS(2745),
    [anon_sym_COMMA] = ACTIONS(2745),
    [anon_sym_RPAREN] = ACTIONS(2745),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2745),
    [anon_sym_DASH_GT] = ACTIONS(2745),
    [anon_sym_STAR] = ACTIONS(2745),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_PERCENT] = ACTIONS(2745),
    [anon_sym_GT] = ACTIONS(2745),
    [anon_sym_LT_LT] = ACTIONS(2745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2745),
    [anon_sym_PLUS_EQ] = ACTIONS(2745),
    [anon_sym_DASH_EQ] = ACTIONS(2745),
    [anon_sym_DOT_EQ] = ACTIONS(2745),
    [anon_sym_STAR_EQ] = ACTIONS(2745),
    [anon_sym_SLASH_EQ] = ACTIONS(2745),
    [anon_sym_PERCENT_EQ] = ACTIONS(2745),
    [anon_sym_x_EQ] = ACTIONS(2745),
    [anon_sym_AMP_EQ] = ACTIONS(2745),
    [anon_sym_PIPE_EQ] = ACTIONS(2745),
    [anon_sym_CARET_EQ] = ACTIONS(2745),
    [anon_sym_LT_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_GT_EQ] = ACTIONS(2745),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2745),
    [anon_sym_DOT_DOT] = ACTIONS(2745),
    [anon_sym_STAR_STAR] = ACTIONS(2745),
    [anon_sym_PIPE_PIPE] = ACTIONS(2745),
    [anon_sym_SLASH_SLASH] = ACTIONS(2745),
    [anon_sym_CARET_CARET] = ACTIONS(2745),
    [anon_sym_AMP_AMP] = ACTIONS(2745),
    [anon_sym_PIPE] = ACTIONS(2745),
    [anon_sym_CARET] = ACTIONS(2745),
    [anon_sym_AMP] = ACTIONS(2745),
    [anon_sym_GT_GT2] = ACTIONS(2745),
    [anon_sym_PLUS] = ACTIONS(2745),
    [anon_sym_DASH] = ACTIONS(2745),
    [anon_sym_DOT] = ACTIONS(2745),
    [anon_sym_SLASH] = ACTIONS(2745),
    [anon_sym_x] = ACTIONS(2745),
    [anon_sym_EQ_TILDE] = ACTIONS(2745),
    [anon_sym_BANG_TILDE] = ACTIONS(2745),
    [anon_sym_EQ_EQ] = ACTIONS(2745),
    [anon_sym_BANG_EQ] = ACTIONS(2745),
    [anon_sym_eq] = ACTIONS(2745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2745),
    [anon_sym_equ] = ACTIONS(2745),
    [anon_sym_eqr] = ACTIONS(2745),
    [anon_sym_ne] = ACTIONS(2745),
    [anon_sym_LT_EQ_GT] = ACTIONS(2745),
    [anon_sym_cmp] = ACTIONS(2745),
    [anon_sym_TILDE_TILDE] = ACTIONS(2745),
    [anon_sym_LT] = ACTIONS(2745),
    [anon_sym_LT_EQ] = ACTIONS(2745),
    [anon_sym_GT_EQ] = ACTIONS(2745),
    [anon_sym_lt] = ACTIONS(2745),
    [anon_sym_le] = ACTIONS(2745),
    [anon_sym_ge] = ACTIONS(2745),
    [anon_sym_gt] = ACTIONS(2745),
    [anon_sym_isa] = ACTIONS(2745),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2745),
    [anon_sym_DASH_DASH] = ACTIONS(2745),
    [anon_sym_QMARK] = ACTIONS(2745),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2745),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_EQ_GT] = ACTIONS(2745),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1408)] = {
    [sym_heredoc_content] = STATE(1408),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2727),
    [anon_sym_DOLLAR] = ACTIONS(2727),
    [anon_sym_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_AT] = ACTIONS(2727),
    [anon_sym_COMMA] = ACTIONS(2727),
    [anon_sym_RPAREN] = ACTIONS(2727),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2727),
    [anon_sym_DASH_GT] = ACTIONS(2727),
    [anon_sym_STAR] = ACTIONS(2727),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2727),
    [anon_sym_GT] = ACTIONS(2727),
    [anon_sym_LT_LT] = ACTIONS(2727),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2727),
    [anon_sym_PLUS_EQ] = ACTIONS(2727),
    [anon_sym_DASH_EQ] = ACTIONS(2727),
    [anon_sym_DOT_EQ] = ACTIONS(2727),
    [anon_sym_STAR_EQ] = ACTIONS(2727),
    [anon_sym_SLASH_EQ] = ACTIONS(2727),
    [anon_sym_PERCENT_EQ] = ACTIONS(2727),
    [anon_sym_x_EQ] = ACTIONS(2727),
    [anon_sym_AMP_EQ] = ACTIONS(2727),
    [anon_sym_PIPE_EQ] = ACTIONS(2727),
    [anon_sym_CARET_EQ] = ACTIONS(2727),
    [anon_sym_LT_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_GT_EQ] = ACTIONS(2727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2727),
    [anon_sym_DOT_DOT] = ACTIONS(2727),
    [anon_sym_STAR_STAR] = ACTIONS(2727),
    [anon_sym_PIPE_PIPE] = ACTIONS(2727),
    [anon_sym_SLASH_SLASH] = ACTIONS(2727),
    [anon_sym_CARET_CARET] = ACTIONS(2727),
    [anon_sym_AMP_AMP] = ACTIONS(2727),
    [anon_sym_PIPE] = ACTIONS(2727),
    [anon_sym_CARET] = ACTIONS(2727),
    [anon_sym_AMP] = ACTIONS(2727),
    [anon_sym_GT_GT2] = ACTIONS(2727),
    [anon_sym_PLUS] = ACTIONS(2727),
    [anon_sym_DASH] = ACTIONS(2727),
    [anon_sym_DOT] = ACTIONS(2727),
    [anon_sym_SLASH] = ACTIONS(2727),
    [anon_sym_x] = ACTIONS(2727),
    [anon_sym_EQ_TILDE] = ACTIONS(2727),
    [anon_sym_BANG_TILDE] = ACTIONS(2727),
    [anon_sym_EQ_EQ] = ACTIONS(2727),
    [anon_sym_BANG_EQ] = ACTIONS(2727),
    [anon_sym_eq] = ACTIONS(2727),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2727),
    [anon_sym_equ] = ACTIONS(2727),
    [anon_sym_eqr] = ACTIONS(2727),
    [anon_sym_ne] = ACTIONS(2727),
    [anon_sym_LT_EQ_GT] = ACTIONS(2727),
    [anon_sym_cmp] = ACTIONS(2727),
    [anon_sym_TILDE_TILDE] = ACTIONS(2727),
    [anon_sym_LT] = ACTIONS(2727),
    [anon_sym_LT_EQ] = ACTIONS(2727),
    [anon_sym_GT_EQ] = ACTIONS(2727),
    [anon_sym_lt] = ACTIONS(2727),
    [anon_sym_le] = ACTIONS(2727),
    [anon_sym_ge] = ACTIONS(2727),
    [anon_sym_gt] = ACTIONS(2727),
    [anon_sym_isa] = ACTIONS(2727),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2727),
    [anon_sym_DASH_DASH] = ACTIONS(2727),
    [anon_sym_QMARK] = ACTIONS(2727),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2727),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2727),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1409)] = {
    [sym_heredoc_content] = STATE(1409),
    [sym__identifier] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_SEMI] = ACTIONS(2717),
    [anon_sym_DOLLAR] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_COMMA] = ACTIONS(2717),
    [anon_sym_my] = ACTIONS(2717),
    [anon_sym_extended] = ACTIONS(2717),
    [anon_sym_async] = ACTIONS(2717),
    [anon_sym_sub] = ACTIONS(2717),
    [anon_sym_method] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_continue] = ACTIONS(2717),
    [anon_sym_state] = ACTIONS(2717),
    [anon_sym_our] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2717),
    [anon_sym_PLUS_EQ] = ACTIONS(2717),
    [anon_sym_DASH_EQ] = ACTIONS(2717),
    [anon_sym_DOT_EQ] = ACTIONS(2717),
    [anon_sym_STAR_EQ] = ACTIONS(2717),
    [anon_sym_SLASH_EQ] = ACTIONS(2717),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2717),
    [anon_sym_AMP_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_EQ] = ACTIONS(2717),
    [anon_sym_CARET_EQ] = ACTIONS(2717),
    [anon_sym_LT_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_GT_EQ] = ACTIONS(2717),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_BANG_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_equ] = ACTIONS(2717),
    [anon_sym_eqr] = ACTIONS(2717),
    [anon_sym_ne] = ACTIONS(2717),
    [anon_sym_LT_EQ_GT] = ACTIONS(2717),
    [anon_sym_cmp] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_lt] = ACTIONS(2717),
    [anon_sym_le] = ACTIONS(2717),
    [anon_sym_ge] = ACTIONS(2717),
    [anon_sym_gt] = ACTIONS(2717),
    [anon_sym_isa] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_BSLASH] = ACTIONS(2719),
    [anon_sym_do] = ACTIONS(2717),
    [anon_sym_eval] = ACTIONS(2717),
    [anon_sym_await] = ACTIONS(2717),
    [anon_sym_field] = ACTIONS(2717),
    [anon_sym_local] = ACTIONS(2717),
    [anon_sym_dynamically] = ACTIONS(2717),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2717),
    [anon_sym_require] = ACTIONS(2717),
    [anon_sym_map] = ACTIONS(2717),
    [anon_sym_grep] = ACTIONS(2717),
    [anon_sym_sort] = ACTIONS(2717),
    [anon_sym_goto] = ACTIONS(2717),
    [anon_sym_return] = ACTIONS(2717),
    [anon_sym_undef] = ACTIONS(2717),
    [anon_sym_PERCENT2] = ACTIONS(2717),
    [anon_sym_AMP2] = ACTIONS(2717),
    [anon_sym_STAR2] = ACTIONS(2717),
    [anon_sym_EQ_GT] = ACTIONS(2717),
    [anon_sym_last] = ACTIONS(2717),
    [anon_sym_next] = ACTIONS(2717),
    [anon_sym_redo] = ACTIONS(2717),
    [anon_sym___FILE__] = ACTIONS(2717),
    [anon_sym___LINE__] = ACTIONS(2717),
    [anon_sym___PACKAGE__] = ACTIONS(2717),
    [anon_sym___SUB__] = ACTIONS(2717),
    [anon_sym_break] = ACTIONS(2717),
    [anon_sym_fork] = ACTIONS(2717),
    [anon_sym_getppid] = ACTIONS(2717),
    [anon_sym_time] = ACTIONS(2717),
    [anon_sym_times] = ACTIONS(2717),
    [anon_sym_wait] = ACTIONS(2717),
    [anon_sym_wantarray] = ACTIONS(2717),
    [anon_sym_abs] = ACTIONS(2717),
    [anon_sym_alarm] = ACTIONS(2717),
    [anon_sym_chop] = ACTIONS(2717),
    [anon_sym_chdir] = ACTIONS(2717),
    [anon_sym_close] = ACTIONS(2717),
    [anon_sym_closedir] = ACTIONS(2717),
    [anon_sym_caller] = ACTIONS(2717),
    [anon_sym_chomp] = ACTIONS(2717),
    [anon_sym_chr] = ACTIONS(2717),
    [anon_sym_cos] = ACTIONS(2717),
    [anon_sym_chroot] = ACTIONS(2717),
    [anon_sym_defined] = ACTIONS(2717),
    [anon_sym_delete] = ACTIONS(2717),
    [anon_sym_dbmclose] = ACTIONS(2717),
    [anon_sym_exists] = ACTIONS(2717),
    [anon_sym_exit] = ACTIONS(2717),
    [anon_sym_eof] = ACTIONS(2717),
    [anon_sym_exp] = ACTIONS(2717),
    [anon_sym_each] = ACTIONS(2717),
    [anon_sym_fc] = ACTIONS(2717),
    [anon_sym_fileno] = ACTIONS(2717),
    [anon_sym_gmtime] = ACTIONS(2717),
    [anon_sym_getc] = ACTIONS(2717),
    [anon_sym_getpgrp] = ACTIONS(2717),
    [anon_sym_getprotobyname] = ACTIONS(2717),
    [anon_sym_getpwname] = ACTIONS(2717),
    [anon_sym_getpwuid] = ACTIONS(2717),
    [anon_sym_getpeername] = ACTIONS(2717),
    [anon_sym_getnetbyname] = ACTIONS(2717),
    [anon_sym_getsockname] = ACTIONS(2717),
    [anon_sym_getgrnam] = ACTIONS(2717),
    [anon_sym_getgrgid] = ACTIONS(2717),
    [anon_sym_hex] = ACTIONS(2717),
    [anon_sym_int] = ACTIONS(2717),
    [anon_sym_keys] = ACTIONS(2717),
    [anon_sym_lc] = ACTIONS(2717),
    [anon_sym_lcfirst] = ACTIONS(2717),
    [anon_sym_length] = ACTIONS(2717),
    [anon_sym_localtime] = ACTIONS(2717),
    [anon_sym_log] = ACTIONS(2717),
    [anon_sym_lock] = ACTIONS(2717),
    [anon_sym_lstat] = ACTIONS(2717),
    [anon_sym_oct] = ACTIONS(2717),
    [anon_sym_ord] = ACTIONS(2717),
    [anon_sym_prototype] = ACTIONS(2717),
    [anon_sym_pop] = ACTIONS(2717),
    [anon_sym_pos] = ACTIONS(2717),
    [anon_sym_quotemeta] = ACTIONS(2717),
    [anon_sym_reset] = ACTIONS(2717),
    [anon_sym_rand] = ACTIONS(2717),
    [anon_sym_rmdir] = ACTIONS(2717),
    [anon_sym_readdir] = ACTIONS(2717),
    [anon_sym_readline] = ACTIONS(2717),
    [anon_sym_readpipe] = ACTIONS(2717),
    [anon_sym_rewinddir] = ACTIONS(2717),
    [anon_sym_readlink] = ACTIONS(2717),
    [anon_sym_ref] = ACTIONS(2717),
    [anon_sym_scalar] = ACTIONS(2717),
    [anon_sym_shift] = ACTIONS(2717),
    [anon_sym_sin] = ACTIONS(2717),
    [anon_sym_sleep] = ACTIONS(2717),
    [anon_sym_sqrt] = ACTIONS(2717),
    [anon_sym_srand] = ACTIONS(2717),
    [anon_sym_stat] = ACTIONS(2717),
    [anon_sym_study] = ACTIONS(2717),
    [anon_sym_tell] = ACTIONS(2717),
    [anon_sym_telldir] = ACTIONS(2717),
    [anon_sym_tied] = ACTIONS(2717),
    [anon_sym_uc] = ACTIONS(2717),
    [anon_sym_ucfirst] = ACTIONS(2717),
    [anon_sym_untie] = ACTIONS(2717),
    [anon_sym_umask] = ACTIONS(2717),
    [anon_sym_values] = ACTIONS(2717),
    [anon_sym_write] = ACTIONS(2717),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2717),
    [anon_sym_qq] = ACTIONS(2717),
    [anon_sym_qw] = ACTIONS(2717),
    [anon_sym_qx] = ACTIONS(2717),
    [anon_sym_qr] = ACTIONS(2717),
    [anon_sym_m] = ACTIONS(2717),
    [anon_sym_s] = ACTIONS(2717),
    [anon_sym_tr] = ACTIONS(2717),
    [anon_sym_y] = ACTIONS(2717),
    [aux_sym__bareword_token1] = ACTIONS(2717),
    [sym_number] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [sym__single_quote] = ACTIONS(2719),
    [sym__double_quote] = ACTIONS(2719),
    [sym__backtick_quote] = ACTIONS(2719),
    [sym__search_slash_quote] = ACTIONS(2719),
    [sym__open_readline_bracket] = ACTIONS(2719),
    [sym__open_fileglob_bracket] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__PERLY_HEREDOC] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2719),
    [sym__filetest] = ACTIONS(2719),
  },
  [STATE(1410)] = {
    [sym_heredoc_content] = STATE(1410),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_DASH_GT] = ACTIONS(2532),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2532),
    [anon_sym_PLUS_EQ] = ACTIONS(2532),
    [anon_sym_DASH_EQ] = ACTIONS(2532),
    [anon_sym_DOT_EQ] = ACTIONS(2532),
    [anon_sym_STAR_EQ] = ACTIONS(2532),
    [anon_sym_SLASH_EQ] = ACTIONS(2532),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2532),
    [anon_sym_AMP_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_EQ] = ACTIONS(2532),
    [anon_sym_CARET_EQ] = ACTIONS(2532),
    [anon_sym_LT_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_GT_EQ] = ACTIONS(2532),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2532),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2532),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2532),
    [anon_sym_BANG_TILDE] = ACTIONS(2532),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2532),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2532),
    [anon_sym_equ] = ACTIONS(2532),
    [anon_sym_eqr] = ACTIONS(2532),
    [anon_sym_ne] = ACTIONS(2532),
    [anon_sym_LT_EQ_GT] = ACTIONS(2532),
    [anon_sym_cmp] = ACTIONS(2532),
    [anon_sym_TILDE_TILDE] = ACTIONS(2532),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2532),
    [anon_sym_lt] = ACTIONS(2532),
    [anon_sym_le] = ACTIONS(2532),
    [anon_sym_ge] = ACTIONS(2532),
    [anon_sym_gt] = ACTIONS(2532),
    [anon_sym_isa] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_QMARK] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1411)] = {
    [sym_heredoc_content] = STATE(1411),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1412)] = {
    [sym_heredoc_content] = STATE(1412),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2682),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2682),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1413)] = {
    [sym_heredoc_content] = STATE(1413),
    [sym__identifier] = ACTIONS(2713),
    [anon_sym_LBRACE] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2713),
    [anon_sym_DOLLAR] = ACTIONS(2713),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_AT] = ACTIONS(2713),
    [anon_sym_COMMA] = ACTIONS(2713),
    [anon_sym_my] = ACTIONS(2713),
    [anon_sym_extended] = ACTIONS(2713),
    [anon_sym_async] = ACTIONS(2713),
    [anon_sym_sub] = ACTIONS(2713),
    [anon_sym_method] = ACTIONS(2713),
    [anon_sym_LPAREN] = ACTIONS(2713),
    [anon_sym_continue] = ACTIONS(2713),
    [anon_sym_state] = ACTIONS(2713),
    [anon_sym_our] = ACTIONS(2713),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2713),
    [anon_sym_DASH_GT] = ACTIONS(2713),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2713),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2713),
    [anon_sym_PLUS_EQ] = ACTIONS(2713),
    [anon_sym_DASH_EQ] = ACTIONS(2713),
    [anon_sym_DOT_EQ] = ACTIONS(2713),
    [anon_sym_STAR_EQ] = ACTIONS(2713),
    [anon_sym_SLASH_EQ] = ACTIONS(2713),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2713),
    [anon_sym_AMP_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_EQ] = ACTIONS(2713),
    [anon_sym_CARET_EQ] = ACTIONS(2713),
    [anon_sym_LT_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_GT_EQ] = ACTIONS(2713),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2713),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2713),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2713),
    [anon_sym_BANG_TILDE] = ACTIONS(2713),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2713),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2713),
    [anon_sym_equ] = ACTIONS(2713),
    [anon_sym_eqr] = ACTIONS(2713),
    [anon_sym_ne] = ACTIONS(2713),
    [anon_sym_LT_EQ_GT] = ACTIONS(2713),
    [anon_sym_cmp] = ACTIONS(2713),
    [anon_sym_TILDE_TILDE] = ACTIONS(2713),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2713),
    [anon_sym_lt] = ACTIONS(2713),
    [anon_sym_le] = ACTIONS(2713),
    [anon_sym_ge] = ACTIONS(2713),
    [anon_sym_gt] = ACTIONS(2713),
    [anon_sym_isa] = ACTIONS(2713),
    [anon_sym_TILDE] = ACTIONS(2713),
    [anon_sym_BANG] = ACTIONS(2713),
    [anon_sym_PLUS_PLUS] = ACTIONS(2713),
    [anon_sym_DASH_DASH] = ACTIONS(2713),
    [anon_sym_QMARK] = ACTIONS(2713),
    [anon_sym_BSLASH] = ACTIONS(2715),
    [anon_sym_do] = ACTIONS(2713),
    [anon_sym_eval] = ACTIONS(2713),
    [anon_sym_await] = ACTIONS(2713),
    [anon_sym_field] = ACTIONS(2713),
    [anon_sym_local] = ACTIONS(2713),
    [anon_sym_dynamically] = ACTIONS(2713),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2713),
    [anon_sym_require] = ACTIONS(2713),
    [anon_sym_map] = ACTIONS(2713),
    [anon_sym_grep] = ACTIONS(2713),
    [anon_sym_sort] = ACTIONS(2713),
    [anon_sym_goto] = ACTIONS(2713),
    [anon_sym_return] = ACTIONS(2713),
    [anon_sym_undef] = ACTIONS(2713),
    [anon_sym_PERCENT2] = ACTIONS(2713),
    [anon_sym_AMP2] = ACTIONS(2713),
    [anon_sym_STAR2] = ACTIONS(2713),
    [anon_sym_EQ_GT] = ACTIONS(2713),
    [anon_sym_last] = ACTIONS(2713),
    [anon_sym_next] = ACTIONS(2713),
    [anon_sym_redo] = ACTIONS(2713),
    [anon_sym___FILE__] = ACTIONS(2713),
    [anon_sym___LINE__] = ACTIONS(2713),
    [anon_sym___PACKAGE__] = ACTIONS(2713),
    [anon_sym___SUB__] = ACTIONS(2713),
    [anon_sym_break] = ACTIONS(2713),
    [anon_sym_fork] = ACTIONS(2713),
    [anon_sym_getppid] = ACTIONS(2713),
    [anon_sym_time] = ACTIONS(2713),
    [anon_sym_times] = ACTIONS(2713),
    [anon_sym_wait] = ACTIONS(2713),
    [anon_sym_wantarray] = ACTIONS(2713),
    [anon_sym_abs] = ACTIONS(2713),
    [anon_sym_alarm] = ACTIONS(2713),
    [anon_sym_chop] = ACTIONS(2713),
    [anon_sym_chdir] = ACTIONS(2713),
    [anon_sym_close] = ACTIONS(2713),
    [anon_sym_closedir] = ACTIONS(2713),
    [anon_sym_caller] = ACTIONS(2713),
    [anon_sym_chomp] = ACTIONS(2713),
    [anon_sym_chr] = ACTIONS(2713),
    [anon_sym_cos] = ACTIONS(2713),
    [anon_sym_chroot] = ACTIONS(2713),
    [anon_sym_defined] = ACTIONS(2713),
    [anon_sym_delete] = ACTIONS(2713),
    [anon_sym_dbmclose] = ACTIONS(2713),
    [anon_sym_exists] = ACTIONS(2713),
    [anon_sym_exit] = ACTIONS(2713),
    [anon_sym_eof] = ACTIONS(2713),
    [anon_sym_exp] = ACTIONS(2713),
    [anon_sym_each] = ACTIONS(2713),
    [anon_sym_fc] = ACTIONS(2713),
    [anon_sym_fileno] = ACTIONS(2713),
    [anon_sym_gmtime] = ACTIONS(2713),
    [anon_sym_getc] = ACTIONS(2713),
    [anon_sym_getpgrp] = ACTIONS(2713),
    [anon_sym_getprotobyname] = ACTIONS(2713),
    [anon_sym_getpwname] = ACTIONS(2713),
    [anon_sym_getpwuid] = ACTIONS(2713),
    [anon_sym_getpeername] = ACTIONS(2713),
    [anon_sym_getnetbyname] = ACTIONS(2713),
    [anon_sym_getsockname] = ACTIONS(2713),
    [anon_sym_getgrnam] = ACTIONS(2713),
    [anon_sym_getgrgid] = ACTIONS(2713),
    [anon_sym_hex] = ACTIONS(2713),
    [anon_sym_int] = ACTIONS(2713),
    [anon_sym_keys] = ACTIONS(2713),
    [anon_sym_lc] = ACTIONS(2713),
    [anon_sym_lcfirst] = ACTIONS(2713),
    [anon_sym_length] = ACTIONS(2713),
    [anon_sym_localtime] = ACTIONS(2713),
    [anon_sym_log] = ACTIONS(2713),
    [anon_sym_lock] = ACTIONS(2713),
    [anon_sym_lstat] = ACTIONS(2713),
    [anon_sym_oct] = ACTIONS(2713),
    [anon_sym_ord] = ACTIONS(2713),
    [anon_sym_prototype] = ACTIONS(2713),
    [anon_sym_pop] = ACTIONS(2713),
    [anon_sym_pos] = ACTIONS(2713),
    [anon_sym_quotemeta] = ACTIONS(2713),
    [anon_sym_reset] = ACTIONS(2713),
    [anon_sym_rand] = ACTIONS(2713),
    [anon_sym_rmdir] = ACTIONS(2713),
    [anon_sym_readdir] = ACTIONS(2713),
    [anon_sym_readline] = ACTIONS(2713),
    [anon_sym_readpipe] = ACTIONS(2713),
    [anon_sym_rewinddir] = ACTIONS(2713),
    [anon_sym_readlink] = ACTIONS(2713),
    [anon_sym_ref] = ACTIONS(2713),
    [anon_sym_scalar] = ACTIONS(2713),
    [anon_sym_shift] = ACTIONS(2713),
    [anon_sym_sin] = ACTIONS(2713),
    [anon_sym_sleep] = ACTIONS(2713),
    [anon_sym_sqrt] = ACTIONS(2713),
    [anon_sym_srand] = ACTIONS(2713),
    [anon_sym_stat] = ACTIONS(2713),
    [anon_sym_study] = ACTIONS(2713),
    [anon_sym_tell] = ACTIONS(2713),
    [anon_sym_telldir] = ACTIONS(2713),
    [anon_sym_tied] = ACTIONS(2713),
    [anon_sym_uc] = ACTIONS(2713),
    [anon_sym_ucfirst] = ACTIONS(2713),
    [anon_sym_untie] = ACTIONS(2713),
    [anon_sym_umask] = ACTIONS(2713),
    [anon_sym_values] = ACTIONS(2713),
    [anon_sym_write] = ACTIONS(2713),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2713),
    [anon_sym_qq] = ACTIONS(2713),
    [anon_sym_qw] = ACTIONS(2713),
    [anon_sym_qx] = ACTIONS(2713),
    [anon_sym_qr] = ACTIONS(2713),
    [anon_sym_m] = ACTIONS(2713),
    [anon_sym_s] = ACTIONS(2713),
    [anon_sym_tr] = ACTIONS(2713),
    [anon_sym_y] = ACTIONS(2713),
    [aux_sym__bareword_token1] = ACTIONS(2713),
    [sym_number] = ACTIONS(2713),
    [anon_sym_true] = ACTIONS(2713),
    [anon_sym_false] = ACTIONS(2713),
    [sym__single_quote] = ACTIONS(2715),
    [sym__double_quote] = ACTIONS(2715),
    [sym__backtick_quote] = ACTIONS(2715),
    [sym__search_slash_quote] = ACTIONS(2715),
    [sym__open_readline_bracket] = ACTIONS(2715),
    [sym__open_fileglob_bracket] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__PERLY_HEREDOC] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2715),
    [sym__filetest] = ACTIONS(2715),
  },
  [STATE(1414)] = {
    [sym_heredoc_content] = STATE(1414),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_SEMI] = ACTIONS(2702),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_COMMA] = ACTIONS(2702),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2702),
    [anon_sym_DASH_GT] = ACTIONS(2702),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2702),
    [anon_sym_PLUS_EQ] = ACTIONS(2702),
    [anon_sym_DASH_EQ] = ACTIONS(2702),
    [anon_sym_DOT_EQ] = ACTIONS(2702),
    [anon_sym_STAR_EQ] = ACTIONS(2702),
    [anon_sym_SLASH_EQ] = ACTIONS(2702),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2702),
    [anon_sym_AMP_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_EQ] = ACTIONS(2702),
    [anon_sym_CARET_EQ] = ACTIONS(2702),
    [anon_sym_LT_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_GT_EQ] = ACTIONS(2702),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2702),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_CARET_CARET] = ACTIONS(2702),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2702),
    [anon_sym_BANG_TILDE] = ACTIONS(2702),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2702),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2702),
    [anon_sym_equ] = ACTIONS(2702),
    [anon_sym_eqr] = ACTIONS(2702),
    [anon_sym_ne] = ACTIONS(2702),
    [anon_sym_LT_EQ_GT] = ACTIONS(2702),
    [anon_sym_cmp] = ACTIONS(2702),
    [anon_sym_TILDE_TILDE] = ACTIONS(2702),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2702),
    [anon_sym_lt] = ACTIONS(2702),
    [anon_sym_le] = ACTIONS(2702),
    [anon_sym_ge] = ACTIONS(2702),
    [anon_sym_gt] = ACTIONS(2702),
    [anon_sym_isa] = ACTIONS(2702),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_QMARK] = ACTIONS(2702),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_EQ_GT] = ACTIONS(2702),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1415)] = {
    [sym_heredoc_content] = STATE(1415),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2851),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2739),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1416)] = {
    [sym_heredoc_content] = STATE(1416),
    [sym__identifier] = ACTIONS(2721),
    [anon_sym_LBRACE] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2721),
    [anon_sym_DOLLAR] = ACTIONS(2721),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_AT] = ACTIONS(2721),
    [anon_sym_COMMA] = ACTIONS(2721),
    [anon_sym_my] = ACTIONS(2721),
    [anon_sym_extended] = ACTIONS(2721),
    [anon_sym_async] = ACTIONS(2721),
    [anon_sym_sub] = ACTIONS(2721),
    [anon_sym_method] = ACTIONS(2721),
    [anon_sym_LPAREN] = ACTIONS(2721),
    [anon_sym_continue] = ACTIONS(2721),
    [anon_sym_state] = ACTIONS(2721),
    [anon_sym_our] = ACTIONS(2721),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2721),
    [anon_sym_DASH_GT] = ACTIONS(2721),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2721),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2721),
    [anon_sym_PLUS_EQ] = ACTIONS(2721),
    [anon_sym_DASH_EQ] = ACTIONS(2721),
    [anon_sym_DOT_EQ] = ACTIONS(2721),
    [anon_sym_STAR_EQ] = ACTIONS(2721),
    [anon_sym_SLASH_EQ] = ACTIONS(2721),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2721),
    [anon_sym_AMP_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_EQ] = ACTIONS(2721),
    [anon_sym_CARET_EQ] = ACTIONS(2721),
    [anon_sym_LT_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_GT_EQ] = ACTIONS(2721),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2721),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2721),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2721),
    [anon_sym_BANG_TILDE] = ACTIONS(2721),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2721),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2721),
    [anon_sym_equ] = ACTIONS(2721),
    [anon_sym_eqr] = ACTIONS(2721),
    [anon_sym_ne] = ACTIONS(2721),
    [anon_sym_LT_EQ_GT] = ACTIONS(2721),
    [anon_sym_cmp] = ACTIONS(2721),
    [anon_sym_TILDE_TILDE] = ACTIONS(2721),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2721),
    [anon_sym_lt] = ACTIONS(2721),
    [anon_sym_le] = ACTIONS(2721),
    [anon_sym_ge] = ACTIONS(2721),
    [anon_sym_gt] = ACTIONS(2721),
    [anon_sym_isa] = ACTIONS(2721),
    [anon_sym_TILDE] = ACTIONS(2721),
    [anon_sym_BANG] = ACTIONS(2721),
    [anon_sym_PLUS_PLUS] = ACTIONS(2721),
    [anon_sym_DASH_DASH] = ACTIONS(2721),
    [anon_sym_QMARK] = ACTIONS(2721),
    [anon_sym_BSLASH] = ACTIONS(2723),
    [anon_sym_do] = ACTIONS(2721),
    [anon_sym_eval] = ACTIONS(2721),
    [anon_sym_await] = ACTIONS(2721),
    [anon_sym_field] = ACTIONS(2721),
    [anon_sym_local] = ACTIONS(2721),
    [anon_sym_dynamically] = ACTIONS(2721),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2721),
    [anon_sym_require] = ACTIONS(2721),
    [anon_sym_map] = ACTIONS(2721),
    [anon_sym_grep] = ACTIONS(2721),
    [anon_sym_sort] = ACTIONS(2721),
    [anon_sym_goto] = ACTIONS(2721),
    [anon_sym_return] = ACTIONS(2721),
    [anon_sym_undef] = ACTIONS(2721),
    [anon_sym_PERCENT2] = ACTIONS(2721),
    [anon_sym_AMP2] = ACTIONS(2721),
    [anon_sym_STAR2] = ACTIONS(2721),
    [anon_sym_EQ_GT] = ACTIONS(2721),
    [anon_sym_last] = ACTIONS(2721),
    [anon_sym_next] = ACTIONS(2721),
    [anon_sym_redo] = ACTIONS(2721),
    [anon_sym___FILE__] = ACTIONS(2721),
    [anon_sym___LINE__] = ACTIONS(2721),
    [anon_sym___PACKAGE__] = ACTIONS(2721),
    [anon_sym___SUB__] = ACTIONS(2721),
    [anon_sym_break] = ACTIONS(2721),
    [anon_sym_fork] = ACTIONS(2721),
    [anon_sym_getppid] = ACTIONS(2721),
    [anon_sym_time] = ACTIONS(2721),
    [anon_sym_times] = ACTIONS(2721),
    [anon_sym_wait] = ACTIONS(2721),
    [anon_sym_wantarray] = ACTIONS(2721),
    [anon_sym_abs] = ACTIONS(2721),
    [anon_sym_alarm] = ACTIONS(2721),
    [anon_sym_chop] = ACTIONS(2721),
    [anon_sym_chdir] = ACTIONS(2721),
    [anon_sym_close] = ACTIONS(2721),
    [anon_sym_closedir] = ACTIONS(2721),
    [anon_sym_caller] = ACTIONS(2721),
    [anon_sym_chomp] = ACTIONS(2721),
    [anon_sym_chr] = ACTIONS(2721),
    [anon_sym_cos] = ACTIONS(2721),
    [anon_sym_chroot] = ACTIONS(2721),
    [anon_sym_defined] = ACTIONS(2721),
    [anon_sym_delete] = ACTIONS(2721),
    [anon_sym_dbmclose] = ACTIONS(2721),
    [anon_sym_exists] = ACTIONS(2721),
    [anon_sym_exit] = ACTIONS(2721),
    [anon_sym_eof] = ACTIONS(2721),
    [anon_sym_exp] = ACTIONS(2721),
    [anon_sym_each] = ACTIONS(2721),
    [anon_sym_fc] = ACTIONS(2721),
    [anon_sym_fileno] = ACTIONS(2721),
    [anon_sym_gmtime] = ACTIONS(2721),
    [anon_sym_getc] = ACTIONS(2721),
    [anon_sym_getpgrp] = ACTIONS(2721),
    [anon_sym_getprotobyname] = ACTIONS(2721),
    [anon_sym_getpwname] = ACTIONS(2721),
    [anon_sym_getpwuid] = ACTIONS(2721),
    [anon_sym_getpeername] = ACTIONS(2721),
    [anon_sym_getnetbyname] = ACTIONS(2721),
    [anon_sym_getsockname] = ACTIONS(2721),
    [anon_sym_getgrnam] = ACTIONS(2721),
    [anon_sym_getgrgid] = ACTIONS(2721),
    [anon_sym_hex] = ACTIONS(2721),
    [anon_sym_int] = ACTIONS(2721),
    [anon_sym_keys] = ACTIONS(2721),
    [anon_sym_lc] = ACTIONS(2721),
    [anon_sym_lcfirst] = ACTIONS(2721),
    [anon_sym_length] = ACTIONS(2721),
    [anon_sym_localtime] = ACTIONS(2721),
    [anon_sym_log] = ACTIONS(2721),
    [anon_sym_lock] = ACTIONS(2721),
    [anon_sym_lstat] = ACTIONS(2721),
    [anon_sym_oct] = ACTIONS(2721),
    [anon_sym_ord] = ACTIONS(2721),
    [anon_sym_prototype] = ACTIONS(2721),
    [anon_sym_pop] = ACTIONS(2721),
    [anon_sym_pos] = ACTIONS(2721),
    [anon_sym_quotemeta] = ACTIONS(2721),
    [anon_sym_reset] = ACTIONS(2721),
    [anon_sym_rand] = ACTIONS(2721),
    [anon_sym_rmdir] = ACTIONS(2721),
    [anon_sym_readdir] = ACTIONS(2721),
    [anon_sym_readline] = ACTIONS(2721),
    [anon_sym_readpipe] = ACTIONS(2721),
    [anon_sym_rewinddir] = ACTIONS(2721),
    [anon_sym_readlink] = ACTIONS(2721),
    [anon_sym_ref] = ACTIONS(2721),
    [anon_sym_scalar] = ACTIONS(2721),
    [anon_sym_shift] = ACTIONS(2721),
    [anon_sym_sin] = ACTIONS(2721),
    [anon_sym_sleep] = ACTIONS(2721),
    [anon_sym_sqrt] = ACTIONS(2721),
    [anon_sym_srand] = ACTIONS(2721),
    [anon_sym_stat] = ACTIONS(2721),
    [anon_sym_study] = ACTIONS(2721),
    [anon_sym_tell] = ACTIONS(2721),
    [anon_sym_telldir] = ACTIONS(2721),
    [anon_sym_tied] = ACTIONS(2721),
    [anon_sym_uc] = ACTIONS(2721),
    [anon_sym_ucfirst] = ACTIONS(2721),
    [anon_sym_untie] = ACTIONS(2721),
    [anon_sym_umask] = ACTIONS(2721),
    [anon_sym_values] = ACTIONS(2721),
    [anon_sym_write] = ACTIONS(2721),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2721),
    [anon_sym_qq] = ACTIONS(2721),
    [anon_sym_qw] = ACTIONS(2721),
    [anon_sym_qx] = ACTIONS(2721),
    [anon_sym_qr] = ACTIONS(2721),
    [anon_sym_m] = ACTIONS(2721),
    [anon_sym_s] = ACTIONS(2721),
    [anon_sym_tr] = ACTIONS(2721),
    [anon_sym_y] = ACTIONS(2721),
    [aux_sym__bareword_token1] = ACTIONS(2721),
    [sym_number] = ACTIONS(2721),
    [anon_sym_true] = ACTIONS(2721),
    [anon_sym_false] = ACTIONS(2721),
    [sym__single_quote] = ACTIONS(2723),
    [sym__double_quote] = ACTIONS(2723),
    [sym__backtick_quote] = ACTIONS(2723),
    [sym__search_slash_quote] = ACTIONS(2723),
    [sym__open_readline_bracket] = ACTIONS(2723),
    [sym__open_fileglob_bracket] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__PERLY_HEREDOC] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2723),
    [sym__filetest] = ACTIONS(2723),
  },
  [STATE(1417)] = {
    [sym_heredoc_content] = STATE(1417),
    [sym__identifier] = ACTIONS(2682),
    [anon_sym_LBRACE] = ACTIONS(2684),
    [anon_sym_SEMI] = ACTIONS(2682),
    [anon_sym_DOLLAR] = ACTIONS(2682),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_AT] = ACTIONS(2682),
    [anon_sym_COMMA] = ACTIONS(2682),
    [anon_sym_my] = ACTIONS(2682),
    [anon_sym_extended] = ACTIONS(2682),
    [anon_sym_async] = ACTIONS(2682),
    [anon_sym_sub] = ACTIONS(2682),
    [anon_sym_method] = ACTIONS(2682),
    [anon_sym_LPAREN] = ACTIONS(2682),
    [anon_sym_continue] = ACTIONS(2682),
    [anon_sym_state] = ACTIONS(2682),
    [anon_sym_our] = ACTIONS(2682),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2682),
    [anon_sym_DASH_GT] = ACTIONS(2682),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2684),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2682),
    [anon_sym_PLUS_EQ] = ACTIONS(2682),
    [anon_sym_DASH_EQ] = ACTIONS(2682),
    [anon_sym_DOT_EQ] = ACTIONS(2682),
    [anon_sym_STAR_EQ] = ACTIONS(2682),
    [anon_sym_SLASH_EQ] = ACTIONS(2682),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2682),
    [anon_sym_AMP_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_EQ] = ACTIONS(2682),
    [anon_sym_CARET_EQ] = ACTIONS(2682),
    [anon_sym_LT_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_GT_EQ] = ACTIONS(2682),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2682),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2682),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2682),
    [anon_sym_BANG_TILDE] = ACTIONS(2682),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2682),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2682),
    [anon_sym_equ] = ACTIONS(2682),
    [anon_sym_eqr] = ACTIONS(2682),
    [anon_sym_ne] = ACTIONS(2682),
    [anon_sym_LT_EQ_GT] = ACTIONS(2682),
    [anon_sym_cmp] = ACTIONS(2682),
    [anon_sym_TILDE_TILDE] = ACTIONS(2682),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2682),
    [anon_sym_lt] = ACTIONS(2682),
    [anon_sym_le] = ACTIONS(2682),
    [anon_sym_ge] = ACTIONS(2682),
    [anon_sym_gt] = ACTIONS(2682),
    [anon_sym_isa] = ACTIONS(2682),
    [anon_sym_TILDE] = ACTIONS(2682),
    [anon_sym_BANG] = ACTIONS(2682),
    [anon_sym_PLUS_PLUS] = ACTIONS(2682),
    [anon_sym_DASH_DASH] = ACTIONS(2682),
    [anon_sym_QMARK] = ACTIONS(2682),
    [anon_sym_BSLASH] = ACTIONS(2686),
    [anon_sym_do] = ACTIONS(2682),
    [anon_sym_eval] = ACTIONS(2682),
    [anon_sym_await] = ACTIONS(2682),
    [anon_sym_field] = ACTIONS(2682),
    [anon_sym_local] = ACTIONS(2682),
    [anon_sym_dynamically] = ACTIONS(2682),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2682),
    [anon_sym_require] = ACTIONS(2682),
    [anon_sym_map] = ACTIONS(2682),
    [anon_sym_grep] = ACTIONS(2682),
    [anon_sym_sort] = ACTIONS(2682),
    [anon_sym_goto] = ACTIONS(2682),
    [anon_sym_return] = ACTIONS(2682),
    [anon_sym_undef] = ACTIONS(2682),
    [anon_sym_PERCENT2] = ACTIONS(2682),
    [anon_sym_AMP2] = ACTIONS(2682),
    [anon_sym_STAR2] = ACTIONS(2682),
    [anon_sym_EQ_GT] = ACTIONS(2682),
    [anon_sym_last] = ACTIONS(2682),
    [anon_sym_next] = ACTIONS(2682),
    [anon_sym_redo] = ACTIONS(2682),
    [anon_sym___FILE__] = ACTIONS(2682),
    [anon_sym___LINE__] = ACTIONS(2682),
    [anon_sym___PACKAGE__] = ACTIONS(2682),
    [anon_sym___SUB__] = ACTIONS(2682),
    [anon_sym_break] = ACTIONS(2682),
    [anon_sym_fork] = ACTIONS(2682),
    [anon_sym_getppid] = ACTIONS(2682),
    [anon_sym_time] = ACTIONS(2682),
    [anon_sym_times] = ACTIONS(2682),
    [anon_sym_wait] = ACTIONS(2682),
    [anon_sym_wantarray] = ACTIONS(2682),
    [anon_sym_abs] = ACTIONS(2682),
    [anon_sym_alarm] = ACTIONS(2682),
    [anon_sym_chop] = ACTIONS(2682),
    [anon_sym_chdir] = ACTIONS(2682),
    [anon_sym_close] = ACTIONS(2682),
    [anon_sym_closedir] = ACTIONS(2682),
    [anon_sym_caller] = ACTIONS(2682),
    [anon_sym_chomp] = ACTIONS(2682),
    [anon_sym_chr] = ACTIONS(2682),
    [anon_sym_cos] = ACTIONS(2682),
    [anon_sym_chroot] = ACTIONS(2682),
    [anon_sym_defined] = ACTIONS(2682),
    [anon_sym_delete] = ACTIONS(2682),
    [anon_sym_dbmclose] = ACTIONS(2682),
    [anon_sym_exists] = ACTIONS(2682),
    [anon_sym_exit] = ACTIONS(2682),
    [anon_sym_eof] = ACTIONS(2682),
    [anon_sym_exp] = ACTIONS(2682),
    [anon_sym_each] = ACTIONS(2682),
    [anon_sym_fc] = ACTIONS(2682),
    [anon_sym_fileno] = ACTIONS(2682),
    [anon_sym_gmtime] = ACTIONS(2682),
    [anon_sym_getc] = ACTIONS(2682),
    [anon_sym_getpgrp] = ACTIONS(2682),
    [anon_sym_getprotobyname] = ACTIONS(2682),
    [anon_sym_getpwname] = ACTIONS(2682),
    [anon_sym_getpwuid] = ACTIONS(2682),
    [anon_sym_getpeername] = ACTIONS(2682),
    [anon_sym_getnetbyname] = ACTIONS(2682),
    [anon_sym_getsockname] = ACTIONS(2682),
    [anon_sym_getgrnam] = ACTIONS(2682),
    [anon_sym_getgrgid] = ACTIONS(2682),
    [anon_sym_hex] = ACTIONS(2682),
    [anon_sym_int] = ACTIONS(2682),
    [anon_sym_keys] = ACTIONS(2682),
    [anon_sym_lc] = ACTIONS(2682),
    [anon_sym_lcfirst] = ACTIONS(2682),
    [anon_sym_length] = ACTIONS(2682),
    [anon_sym_localtime] = ACTIONS(2682),
    [anon_sym_log] = ACTIONS(2682),
    [anon_sym_lock] = ACTIONS(2682),
    [anon_sym_lstat] = ACTIONS(2682),
    [anon_sym_oct] = ACTIONS(2682),
    [anon_sym_ord] = ACTIONS(2682),
    [anon_sym_prototype] = ACTIONS(2682),
    [anon_sym_pop] = ACTIONS(2682),
    [anon_sym_pos] = ACTIONS(2682),
    [anon_sym_quotemeta] = ACTIONS(2682),
    [anon_sym_reset] = ACTIONS(2682),
    [anon_sym_rand] = ACTIONS(2682),
    [anon_sym_rmdir] = ACTIONS(2682),
    [anon_sym_readdir] = ACTIONS(2682),
    [anon_sym_readline] = ACTIONS(2682),
    [anon_sym_readpipe] = ACTIONS(2682),
    [anon_sym_rewinddir] = ACTIONS(2682),
    [anon_sym_readlink] = ACTIONS(2682),
    [anon_sym_ref] = ACTIONS(2682),
    [anon_sym_scalar] = ACTIONS(2682),
    [anon_sym_shift] = ACTIONS(2682),
    [anon_sym_sin] = ACTIONS(2682),
    [anon_sym_sleep] = ACTIONS(2682),
    [anon_sym_sqrt] = ACTIONS(2682),
    [anon_sym_srand] = ACTIONS(2682),
    [anon_sym_stat] = ACTIONS(2682),
    [anon_sym_study] = ACTIONS(2682),
    [anon_sym_tell] = ACTIONS(2682),
    [anon_sym_telldir] = ACTIONS(2682),
    [anon_sym_tied] = ACTIONS(2682),
    [anon_sym_uc] = ACTIONS(2682),
    [anon_sym_ucfirst] = ACTIONS(2682),
    [anon_sym_untie] = ACTIONS(2682),
    [anon_sym_umask] = ACTIONS(2682),
    [anon_sym_values] = ACTIONS(2682),
    [anon_sym_write] = ACTIONS(2682),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2682),
    [anon_sym_qq] = ACTIONS(2682),
    [anon_sym_qw] = ACTIONS(2682),
    [anon_sym_qx] = ACTIONS(2682),
    [anon_sym_qr] = ACTIONS(2682),
    [anon_sym_m] = ACTIONS(2682),
    [anon_sym_s] = ACTIONS(2682),
    [anon_sym_tr] = ACTIONS(2682),
    [anon_sym_y] = ACTIONS(2682),
    [aux_sym__bareword_token1] = ACTIONS(2682),
    [sym_number] = ACTIONS(2682),
    [anon_sym_true] = ACTIONS(2682),
    [anon_sym_false] = ACTIONS(2682),
    [sym__single_quote] = ACTIONS(2686),
    [sym__double_quote] = ACTIONS(2686),
    [sym__backtick_quote] = ACTIONS(2686),
    [sym__search_slash_quote] = ACTIONS(2686),
    [sym__open_readline_bracket] = ACTIONS(2686),
    [sym__open_fileglob_bracket] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__PERLY_HEREDOC] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2686),
    [sym__filetest] = ACTIONS(2686),
  },
  [STATE(1418)] = {
    [sym_heredoc_content] = STATE(1418),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_COLON] = ACTIONS(2534),
    [anon_sym_DOLLAR] = ACTIONS(2534),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_AT] = ACTIONS(2534),
    [anon_sym_COMMA] = ACTIONS(2534),
    [anon_sym_RPAREN] = ACTIONS(2534),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2534),
    [anon_sym_DASH_GT] = ACTIONS(2534),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2534),
    [anon_sym_PLUS_EQ] = ACTIONS(2534),
    [anon_sym_DASH_EQ] = ACTIONS(2534),
    [anon_sym_DOT_EQ] = ACTIONS(2534),
    [anon_sym_STAR_EQ] = ACTIONS(2534),
    [anon_sym_SLASH_EQ] = ACTIONS(2534),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2534),
    [anon_sym_AMP_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_EQ] = ACTIONS(2534),
    [anon_sym_CARET_EQ] = ACTIONS(2534),
    [anon_sym_LT_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_GT_EQ] = ACTIONS(2534),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2534),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2534),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2534),
    [anon_sym_BANG_TILDE] = ACTIONS(2534),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2534),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2534),
    [anon_sym_equ] = ACTIONS(2534),
    [anon_sym_eqr] = ACTIONS(2534),
    [anon_sym_ne] = ACTIONS(2534),
    [anon_sym_LT_EQ_GT] = ACTIONS(2534),
    [anon_sym_cmp] = ACTIONS(2534),
    [anon_sym_TILDE_TILDE] = ACTIONS(2534),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2534),
    [anon_sym_lt] = ACTIONS(2534),
    [anon_sym_le] = ACTIONS(2534),
    [anon_sym_ge] = ACTIONS(2534),
    [anon_sym_gt] = ACTIONS(2534),
    [anon_sym_isa] = ACTIONS(2534),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2534),
    [anon_sym_DASH_DASH] = ACTIONS(2534),
    [anon_sym_QMARK] = ACTIONS(2534),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2534),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_EQ_GT] = ACTIONS(2534),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1419)] = {
    [sym_heredoc_content] = STATE(1419),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_COLON] = ACTIONS(2729),
    [anon_sym_DOLLAR] = ACTIONS(2734),
    [anon_sym_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_AT] = ACTIONS(2734),
    [anon_sym_COMMA] = ACTIONS(2729),
    [anon_sym_RPAREN] = ACTIONS(2729),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2848),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2729),
    [anon_sym_DASH_GT] = ACTIONS(2729),
    [anon_sym_STAR] = ACTIONS(2729),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2729),
    [anon_sym_GT] = ACTIONS(2729),
    [anon_sym_LT_LT] = ACTIONS(2734),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2729),
    [anon_sym_PLUS_EQ] = ACTIONS(2729),
    [anon_sym_DASH_EQ] = ACTIONS(2729),
    [anon_sym_DOT_EQ] = ACTIONS(2729),
    [anon_sym_STAR_EQ] = ACTIONS(2729),
    [anon_sym_SLASH_EQ] = ACTIONS(2729),
    [anon_sym_PERCENT_EQ] = ACTIONS(2729),
    [anon_sym_x_EQ] = ACTIONS(2729),
    [anon_sym_AMP_EQ] = ACTIONS(2729),
    [anon_sym_PIPE_EQ] = ACTIONS(2729),
    [anon_sym_CARET_EQ] = ACTIONS(2729),
    [anon_sym_LT_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_GT_EQ] = ACTIONS(2729),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2729),
    [anon_sym_DOT_DOT] = ACTIONS(2729),
    [anon_sym_STAR_STAR] = ACTIONS(2729),
    [anon_sym_PIPE_PIPE] = ACTIONS(2729),
    [anon_sym_SLASH_SLASH] = ACTIONS(2734),
    [anon_sym_CARET_CARET] = ACTIONS(2729),
    [anon_sym_AMP_AMP] = ACTIONS(2729),
    [anon_sym_PIPE] = ACTIONS(2729),
    [anon_sym_CARET] = ACTIONS(2729),
    [anon_sym_AMP] = ACTIONS(2729),
    [anon_sym_GT_GT2] = ACTIONS(2729),
    [anon_sym_PLUS] = ACTIONS(2734),
    [anon_sym_DASH] = ACTIONS(2734),
    [anon_sym_DOT] = ACTIONS(2729),
    [anon_sym_SLASH] = ACTIONS(2729),
    [anon_sym_x] = ACTIONS(2729),
    [anon_sym_EQ_TILDE] = ACTIONS(2729),
    [anon_sym_BANG_TILDE] = ACTIONS(2729),
    [anon_sym_EQ_EQ] = ACTIONS(2729),
    [anon_sym_BANG_EQ] = ACTIONS(2729),
    [anon_sym_eq] = ACTIONS(2729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2729),
    [anon_sym_equ] = ACTIONS(2729),
    [anon_sym_eqr] = ACTIONS(2729),
    [anon_sym_ne] = ACTIONS(2729),
    [anon_sym_LT_EQ_GT] = ACTIONS(2729),
    [anon_sym_cmp] = ACTIONS(2729),
    [anon_sym_TILDE_TILDE] = ACTIONS(2729),
    [anon_sym_LT] = ACTIONS(2729),
    [anon_sym_LT_EQ] = ACTIONS(2729),
    [anon_sym_GT_EQ] = ACTIONS(2729),
    [anon_sym_lt] = ACTIONS(2729),
    [anon_sym_le] = ACTIONS(2729),
    [anon_sym_ge] = ACTIONS(2729),
    [anon_sym_gt] = ACTIONS(2729),
    [anon_sym_isa] = ACTIONS(2729),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2734),
    [anon_sym_DASH_DASH] = ACTIONS(2734),
    [anon_sym_QMARK] = ACTIONS(2729),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2734),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2729),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1420)] = {
    [sym_heredoc_content] = STATE(1420),
    [sym__identifier] = ACTIONS(2725),
    [anon_sym_LBRACE] = ACTIONS(2725),
    [anon_sym_SEMI] = ACTIONS(2749),
    [anon_sym_DOLLAR] = ACTIONS(2725),
    [anon_sym_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_AT] = ACTIONS(2725),
    [anon_sym_COMMA] = ACTIONS(2749),
    [anon_sym_my] = ACTIONS(2725),
    [anon_sym_extended] = ACTIONS(2725),
    [anon_sym_async] = ACTIONS(2725),
    [anon_sym_sub] = ACTIONS(2725),
    [anon_sym_method] = ACTIONS(2725),
    [anon_sym_LPAREN] = ACTIONS(2851),
    [anon_sym_continue] = ACTIONS(2725),
    [anon_sym_state] = ACTIONS(2725),
    [anon_sym_our] = ACTIONS(2725),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2749),
    [anon_sym_DASH_GT] = ACTIONS(2749),
    [anon_sym_STAR] = ACTIONS(2749),
    [anon_sym_LBRACK] = ACTIONS(2725),
    [anon_sym_PERCENT] = ACTIONS(2749),
    [anon_sym_GT] = ACTIONS(2749),
    [anon_sym_LT_LT] = ACTIONS(2749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2749),
    [anon_sym_PLUS_EQ] = ACTIONS(2749),
    [anon_sym_DASH_EQ] = ACTIONS(2749),
    [anon_sym_DOT_EQ] = ACTIONS(2749),
    [anon_sym_STAR_EQ] = ACTIONS(2749),
    [anon_sym_SLASH_EQ] = ACTIONS(2749),
    [anon_sym_PERCENT_EQ] = ACTIONS(2749),
    [anon_sym_x_EQ] = ACTIONS(2749),
    [anon_sym_AMP_EQ] = ACTIONS(2749),
    [anon_sym_PIPE_EQ] = ACTIONS(2749),
    [anon_sym_CARET_EQ] = ACTIONS(2749),
    [anon_sym_LT_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_GT_EQ] = ACTIONS(2749),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2749),
    [anon_sym_DOT_DOT] = ACTIONS(2749),
    [anon_sym_STAR_STAR] = ACTIONS(2749),
    [anon_sym_PIPE_PIPE] = ACTIONS(2749),
    [anon_sym_SLASH_SLASH] = ACTIONS(2749),
    [anon_sym_CARET_CARET] = ACTIONS(2749),
    [anon_sym_AMP_AMP] = ACTIONS(2749),
    [anon_sym_PIPE] = ACTIONS(2749),
    [anon_sym_CARET] = ACTIONS(2749),
    [anon_sym_AMP] = ACTIONS(2749),
    [anon_sym_GT_GT2] = ACTIONS(2749),
    [anon_sym_PLUS] = ACTIONS(2749),
    [anon_sym_DASH] = ACTIONS(2749),
    [anon_sym_DOT] = ACTIONS(2749),
    [anon_sym_SLASH] = ACTIONS(2749),
    [anon_sym_x] = ACTIONS(2749),
    [anon_sym_EQ_TILDE] = ACTIONS(2749),
    [anon_sym_BANG_TILDE] = ACTIONS(2749),
    [anon_sym_EQ_EQ] = ACTIONS(2749),
    [anon_sym_BANG_EQ] = ACTIONS(2749),
    [anon_sym_eq] = ACTIONS(2749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2749),
    [anon_sym_equ] = ACTIONS(2749),
    [anon_sym_eqr] = ACTIONS(2749),
    [anon_sym_ne] = ACTIONS(2749),
    [anon_sym_LT_EQ_GT] = ACTIONS(2749),
    [anon_sym_cmp] = ACTIONS(2749),
    [anon_sym_TILDE_TILDE] = ACTIONS(2749),
    [anon_sym_LT] = ACTIONS(2749),
    [anon_sym_LT_EQ] = ACTIONS(2749),
    [anon_sym_GT_EQ] = ACTIONS(2749),
    [anon_sym_lt] = ACTIONS(2749),
    [anon_sym_le] = ACTIONS(2749),
    [anon_sym_ge] = ACTIONS(2749),
    [anon_sym_gt] = ACTIONS(2749),
    [anon_sym_isa] = ACTIONS(2749),
    [anon_sym_TILDE] = ACTIONS(2725),
    [anon_sym_BANG] = ACTIONS(2725),
    [anon_sym_PLUS_PLUS] = ACTIONS(2749),
    [anon_sym_DASH_DASH] = ACTIONS(2749),
    [anon_sym_QMARK] = ACTIONS(2749),
    [anon_sym_BSLASH] = ACTIONS(2737),
    [anon_sym_do] = ACTIONS(2725),
    [anon_sym_eval] = ACTIONS(2725),
    [anon_sym_await] = ACTIONS(2725),
    [anon_sym_field] = ACTIONS(2725),
    [anon_sym_local] = ACTIONS(2725),
    [anon_sym_dynamically] = ACTIONS(2725),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2725),
    [anon_sym_require] = ACTIONS(2725),
    [anon_sym_map] = ACTIONS(2725),
    [anon_sym_grep] = ACTIONS(2725),
    [anon_sym_sort] = ACTIONS(2725),
    [anon_sym_goto] = ACTIONS(2725),
    [anon_sym_return] = ACTIONS(2725),
    [anon_sym_undef] = ACTIONS(2725),
    [anon_sym_PERCENT2] = ACTIONS(2725),
    [anon_sym_AMP2] = ACTIONS(2725),
    [anon_sym_STAR2] = ACTIONS(2725),
    [anon_sym_EQ_GT] = ACTIONS(2749),
    [anon_sym_last] = ACTIONS(2725),
    [anon_sym_next] = ACTIONS(2725),
    [anon_sym_redo] = ACTIONS(2725),
    [anon_sym___FILE__] = ACTIONS(2725),
    [anon_sym___LINE__] = ACTIONS(2725),
    [anon_sym___PACKAGE__] = ACTIONS(2725),
    [anon_sym___SUB__] = ACTIONS(2725),
    [anon_sym_break] = ACTIONS(2725),
    [anon_sym_fork] = ACTIONS(2725),
    [anon_sym_getppid] = ACTIONS(2725),
    [anon_sym_time] = ACTIONS(2725),
    [anon_sym_times] = ACTIONS(2725),
    [anon_sym_wait] = ACTIONS(2725),
    [anon_sym_wantarray] = ACTIONS(2725),
    [anon_sym_abs] = ACTIONS(2725),
    [anon_sym_alarm] = ACTIONS(2725),
    [anon_sym_chop] = ACTIONS(2725),
    [anon_sym_chdir] = ACTIONS(2725),
    [anon_sym_close] = ACTIONS(2725),
    [anon_sym_closedir] = ACTIONS(2725),
    [anon_sym_caller] = ACTIONS(2725),
    [anon_sym_chomp] = ACTIONS(2725),
    [anon_sym_chr] = ACTIONS(2725),
    [anon_sym_cos] = ACTIONS(2725),
    [anon_sym_chroot] = ACTIONS(2725),
    [anon_sym_defined] = ACTIONS(2725),
    [anon_sym_delete] = ACTIONS(2725),
    [anon_sym_dbmclose] = ACTIONS(2725),
    [anon_sym_exists] = ACTIONS(2725),
    [anon_sym_exit] = ACTIONS(2725),
    [anon_sym_eof] = ACTIONS(2725),
    [anon_sym_exp] = ACTIONS(2725),
    [anon_sym_each] = ACTIONS(2725),
    [anon_sym_fc] = ACTIONS(2725),
    [anon_sym_fileno] = ACTIONS(2725),
    [anon_sym_gmtime] = ACTIONS(2725),
    [anon_sym_getc] = ACTIONS(2725),
    [anon_sym_getpgrp] = ACTIONS(2725),
    [anon_sym_getprotobyname] = ACTIONS(2725),
    [anon_sym_getpwname] = ACTIONS(2725),
    [anon_sym_getpwuid] = ACTIONS(2725),
    [anon_sym_getpeername] = ACTIONS(2725),
    [anon_sym_getnetbyname] = ACTIONS(2725),
    [anon_sym_getsockname] = ACTIONS(2725),
    [anon_sym_getgrnam] = ACTIONS(2725),
    [anon_sym_getgrgid] = ACTIONS(2725),
    [anon_sym_hex] = ACTIONS(2725),
    [anon_sym_int] = ACTIONS(2725),
    [anon_sym_keys] = ACTIONS(2725),
    [anon_sym_lc] = ACTIONS(2725),
    [anon_sym_lcfirst] = ACTIONS(2725),
    [anon_sym_length] = ACTIONS(2725),
    [anon_sym_localtime] = ACTIONS(2725),
    [anon_sym_log] = ACTIONS(2725),
    [anon_sym_lock] = ACTIONS(2725),
    [anon_sym_lstat] = ACTIONS(2725),
    [anon_sym_oct] = ACTIONS(2725),
    [anon_sym_ord] = ACTIONS(2725),
    [anon_sym_prototype] = ACTIONS(2725),
    [anon_sym_pop] = ACTIONS(2725),
    [anon_sym_pos] = ACTIONS(2725),
    [anon_sym_quotemeta] = ACTIONS(2725),
    [anon_sym_reset] = ACTIONS(2725),
    [anon_sym_rand] = ACTIONS(2725),
    [anon_sym_rmdir] = ACTIONS(2725),
    [anon_sym_readdir] = ACTIONS(2725),
    [anon_sym_readline] = ACTIONS(2725),
    [anon_sym_readpipe] = ACTIONS(2725),
    [anon_sym_rewinddir] = ACTIONS(2725),
    [anon_sym_readlink] = ACTIONS(2725),
    [anon_sym_ref] = ACTIONS(2725),
    [anon_sym_scalar] = ACTIONS(2725),
    [anon_sym_shift] = ACTIONS(2725),
    [anon_sym_sin] = ACTIONS(2725),
    [anon_sym_sleep] = ACTIONS(2725),
    [anon_sym_sqrt] = ACTIONS(2725),
    [anon_sym_srand] = ACTIONS(2725),
    [anon_sym_stat] = ACTIONS(2725),
    [anon_sym_study] = ACTIONS(2725),
    [anon_sym_tell] = ACTIONS(2725),
    [anon_sym_telldir] = ACTIONS(2725),
    [anon_sym_tied] = ACTIONS(2725),
    [anon_sym_uc] = ACTIONS(2725),
    [anon_sym_ucfirst] = ACTIONS(2725),
    [anon_sym_untie] = ACTIONS(2725),
    [anon_sym_umask] = ACTIONS(2725),
    [anon_sym_values] = ACTIONS(2725),
    [anon_sym_write] = ACTIONS(2725),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2725),
    [anon_sym_qq] = ACTIONS(2725),
    [anon_sym_qw] = ACTIONS(2725),
    [anon_sym_qx] = ACTIONS(2725),
    [anon_sym_qr] = ACTIONS(2725),
    [anon_sym_m] = ACTIONS(2725),
    [anon_sym_s] = ACTIONS(2725),
    [anon_sym_tr] = ACTIONS(2725),
    [anon_sym_y] = ACTIONS(2725),
    [aux_sym__bareword_token1] = ACTIONS(2725),
    [sym_number] = ACTIONS(2725),
    [anon_sym_true] = ACTIONS(2725),
    [anon_sym_false] = ACTIONS(2725),
    [sym__single_quote] = ACTIONS(2737),
    [sym__double_quote] = ACTIONS(2737),
    [sym__backtick_quote] = ACTIONS(2737),
    [sym__search_slash_quote] = ACTIONS(2737),
    [sym__open_readline_bracket] = ACTIONS(2737),
    [sym__open_fileglob_bracket] = ACTIONS(2737),
    [sym__PERLY_SEMICOLON] = ACTIONS(2751),
    [sym__PERLY_HEREDOC] = ACTIONS(2737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2737),
    [sym__filetest] = ACTIONS(2737),
  },
  [STATE(1421)] = {
    [sym_heredoc_content] = STATE(1421),
    [sym__identifier] = ACTIONS(2854),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym_DOLLAR] = ACTIONS(2854),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2856),
    [anon_sym_AT] = ACTIONS(2854),
    [anon_sym_COMMA] = ACTIONS(2856),
    [anon_sym_my] = ACTIONS(2854),
    [anon_sym_extended] = ACTIONS(2854),
    [anon_sym_async] = ACTIONS(2854),
    [anon_sym_sub] = ACTIONS(2854),
    [anon_sym_method] = ACTIONS(2854),
    [anon_sym_LPAREN] = ACTIONS(2854),
    [anon_sym_continue] = ACTIONS(2854),
    [anon_sym_state] = ACTIONS(2854),
    [anon_sym_our] = ACTIONS(2854),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2856),
    [anon_sym_DASH_GT] = ACTIONS(2856),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2856),
    [anon_sym_PLUS_EQ] = ACTIONS(2856),
    [anon_sym_DASH_EQ] = ACTIONS(2856),
    [anon_sym_DOT_EQ] = ACTIONS(2856),
    [anon_sym_STAR_EQ] = ACTIONS(2856),
    [anon_sym_SLASH_EQ] = ACTIONS(2856),
    [anon_sym_PERCENT_EQ] = ACTIONS(2856),
    [anon_sym_x_EQ] = ACTIONS(2856),
    [anon_sym_AMP_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_EQ] = ACTIONS(2856),
    [anon_sym_CARET_EQ] = ACTIONS(2856),
    [anon_sym_LT_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_GT_EQ] = ACTIONS(2856),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2856),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2856),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(2856),
    [anon_sym_BANG_TILDE] = ACTIONS(2856),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(2856),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2856),
    [anon_sym_equ] = ACTIONS(2856),
    [anon_sym_eqr] = ACTIONS(2856),
    [anon_sym_ne] = ACTIONS(2856),
    [anon_sym_LT_EQ_GT] = ACTIONS(2856),
    [anon_sym_cmp] = ACTIONS(2856),
    [anon_sym_TILDE_TILDE] = ACTIONS(2856),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(2856),
    [anon_sym_lt] = ACTIONS(2856),
    [anon_sym_le] = ACTIONS(2856),
    [anon_sym_ge] = ACTIONS(2856),
    [anon_sym_gt] = ACTIONS(2856),
    [anon_sym_isa] = ACTIONS(2856),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_QMARK] = ACTIONS(2856),
    [anon_sym_BSLASH] = ACTIONS(2858),
    [anon_sym_do] = ACTIONS(2854),
    [anon_sym_eval] = ACTIONS(2854),
    [anon_sym_await] = ACTIONS(2854),
    [anon_sym_field] = ACTIONS(2854),
    [anon_sym_local] = ACTIONS(2854),
    [anon_sym_dynamically] = ACTIONS(2854),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2854),
    [anon_sym_require] = ACTIONS(2854),
    [anon_sym_map] = ACTIONS(2854),
    [anon_sym_grep] = ACTIONS(2854),
    [anon_sym_sort] = ACTIONS(2854),
    [anon_sym_goto] = ACTIONS(2854),
    [anon_sym_return] = ACTIONS(2854),
    [anon_sym_undef] = ACTIONS(2854),
    [anon_sym_PERCENT2] = ACTIONS(2854),
    [anon_sym_AMP2] = ACTIONS(2854),
    [anon_sym_STAR2] = ACTIONS(2854),
    [anon_sym_EQ_GT] = ACTIONS(2856),
    [anon_sym_last] = ACTIONS(2854),
    [anon_sym_next] = ACTIONS(2854),
    [anon_sym_redo] = ACTIONS(2854),
    [anon_sym___FILE__] = ACTIONS(2854),
    [anon_sym___LINE__] = ACTIONS(2854),
    [anon_sym___PACKAGE__] = ACTIONS(2854),
    [anon_sym___SUB__] = ACTIONS(2854),
    [anon_sym_break] = ACTIONS(2854),
    [anon_sym_fork] = ACTIONS(2854),
    [anon_sym_getppid] = ACTIONS(2854),
    [anon_sym_time] = ACTIONS(2854),
    [anon_sym_times] = ACTIONS(2854),
    [anon_sym_wait] = ACTIONS(2854),
    [anon_sym_wantarray] = ACTIONS(2854),
    [anon_sym_abs] = ACTIONS(2854),
    [anon_sym_alarm] = ACTIONS(2854),
    [anon_sym_chop] = ACTIONS(2854),
    [anon_sym_chdir] = ACTIONS(2854),
    [anon_sym_close] = ACTIONS(2854),
    [anon_sym_closedir] = ACTIONS(2854),
    [anon_sym_caller] = ACTIONS(2854),
    [anon_sym_chomp] = ACTIONS(2854),
    [anon_sym_chr] = ACTIONS(2854),
    [anon_sym_cos] = ACTIONS(2854),
    [anon_sym_chroot] = ACTIONS(2854),
    [anon_sym_defined] = ACTIONS(2854),
    [anon_sym_delete] = ACTIONS(2854),
    [anon_sym_dbmclose] = ACTIONS(2854),
    [anon_sym_exists] = ACTIONS(2854),
    [anon_sym_exit] = ACTIONS(2854),
    [anon_sym_eof] = ACTIONS(2854),
    [anon_sym_exp] = ACTIONS(2854),
    [anon_sym_each] = ACTIONS(2854),
    [anon_sym_fc] = ACTIONS(2854),
    [anon_sym_fileno] = ACTIONS(2854),
    [anon_sym_gmtime] = ACTIONS(2854),
    [anon_sym_getc] = ACTIONS(2854),
    [anon_sym_getpgrp] = ACTIONS(2854),
    [anon_sym_getprotobyname] = ACTIONS(2854),
    [anon_sym_getpwname] = ACTIONS(2854),
    [anon_sym_getpwuid] = ACTIONS(2854),
    [anon_sym_getpeername] = ACTIONS(2854),
    [anon_sym_getnetbyname] = ACTIONS(2854),
    [anon_sym_getsockname] = ACTIONS(2854),
    [anon_sym_getgrnam] = ACTIONS(2854),
    [anon_sym_getgrgid] = ACTIONS(2854),
    [anon_sym_hex] = ACTIONS(2854),
    [anon_sym_int] = ACTIONS(2854),
    [anon_sym_keys] = ACTIONS(2854),
    [anon_sym_lc] = ACTIONS(2854),
    [anon_sym_lcfirst] = ACTIONS(2854),
    [anon_sym_length] = ACTIONS(2854),
    [anon_sym_localtime] = ACTIONS(2854),
    [anon_sym_log] = ACTIONS(2854),
    [anon_sym_lock] = ACTIONS(2854),
    [anon_sym_lstat] = ACTIONS(2854),
    [anon_sym_oct] = ACTIONS(2854),
    [anon_sym_ord] = ACTIONS(2854),
    [anon_sym_prototype] = ACTIONS(2854),
    [anon_sym_pop] = ACTIONS(2854),
    [anon_sym_pos] = ACTIONS(2854),
    [anon_sym_quotemeta] = ACTIONS(2854),
    [anon_sym_reset] = ACTIONS(2854),
    [anon_sym_rand] = ACTIONS(2854),
    [anon_sym_rmdir] = ACTIONS(2854),
    [anon_sym_readdir] = ACTIONS(2854),
    [anon_sym_readline] = ACTIONS(2854),
    [anon_sym_readpipe] = ACTIONS(2854),
    [anon_sym_rewinddir] = ACTIONS(2854),
    [anon_sym_readlink] = ACTIONS(2854),
    [anon_sym_ref] = ACTIONS(2854),
    [anon_sym_scalar] = ACTIONS(2854),
    [anon_sym_shift] = ACTIONS(2854),
    [anon_sym_sin] = ACTIONS(2854),
    [anon_sym_sleep] = ACTIONS(2854),
    [anon_sym_sqrt] = ACTIONS(2854),
    [anon_sym_srand] = ACTIONS(2854),
    [anon_sym_stat] = ACTIONS(2854),
    [anon_sym_study] = ACTIONS(2854),
    [anon_sym_tell] = ACTIONS(2854),
    [anon_sym_telldir] = ACTIONS(2854),
    [anon_sym_tied] = ACTIONS(2854),
    [anon_sym_uc] = ACTIONS(2854),
    [anon_sym_ucfirst] = ACTIONS(2854),
    [anon_sym_untie] = ACTIONS(2854),
    [anon_sym_umask] = ACTIONS(2854),
    [anon_sym_values] = ACTIONS(2854),
    [anon_sym_write] = ACTIONS(2854),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2854),
    [anon_sym_qq] = ACTIONS(2854),
    [anon_sym_qw] = ACTIONS(2854),
    [anon_sym_qx] = ACTIONS(2854),
    [anon_sym_qr] = ACTIONS(2854),
    [anon_sym_m] = ACTIONS(2854),
    [anon_sym_s] = ACTIONS(2854),
    [anon_sym_tr] = ACTIONS(2854),
    [anon_sym_y] = ACTIONS(2854),
    [aux_sym__bareword_token1] = ACTIONS(2854),
    [sym_number] = ACTIONS(2854),
    [anon_sym_true] = ACTIONS(2854),
    [anon_sym_false] = ACTIONS(2854),
    [sym__single_quote] = ACTIONS(2858),
    [sym__double_quote] = ACTIONS(2858),
    [sym__backtick_quote] = ACTIONS(2858),
    [sym__search_slash_quote] = ACTIONS(2858),
    [sym__open_readline_bracket] = ACTIONS(2858),
    [sym__open_fileglob_bracket] = ACTIONS(2858),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym__PERLY_HEREDOC] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2858),
    [sym__filetest] = ACTIONS(2858),
  },
  [STATE(1422)] = {
    [sym_heredoc_content] = STATE(1422),
    [sym__identifier] = ACTIONS(2696),
    [anon_sym_LBRACE] = ACTIONS(2696),
    [anon_sym_SEMI] = ACTIONS(2696),
    [anon_sym_DOLLAR] = ACTIONS(2696),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_AT] = ACTIONS(2696),
    [anon_sym_COMMA] = ACTIONS(2696),
    [anon_sym_my] = ACTIONS(2696),
    [anon_sym_extended] = ACTIONS(2696),
    [anon_sym_async] = ACTIONS(2696),
    [anon_sym_sub] = ACTIONS(2696),
    [anon_sym_method] = ACTIONS(2696),
    [anon_sym_LPAREN] = ACTIONS(2696),
    [anon_sym_continue] = ACTIONS(2696),
    [anon_sym_state] = ACTIONS(2696),
    [anon_sym_our] = ACTIONS(2696),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2696),
    [anon_sym_DASH_GT] = ACTIONS(2696),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2696),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2696),
    [anon_sym_PLUS_EQ] = ACTIONS(2696),
    [anon_sym_DASH_EQ] = ACTIONS(2696),
    [anon_sym_DOT_EQ] = ACTIONS(2696),
    [anon_sym_STAR_EQ] = ACTIONS(2696),
    [anon_sym_SLASH_EQ] = ACTIONS(2696),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2696),
    [anon_sym_AMP_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_EQ] = ACTIONS(2696),
    [anon_sym_CARET_EQ] = ACTIONS(2696),
    [anon_sym_LT_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_GT_EQ] = ACTIONS(2696),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2696),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2696),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2696),
    [anon_sym_BANG_TILDE] = ACTIONS(2696),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2696),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2696),
    [anon_sym_equ] = ACTIONS(2696),
    [anon_sym_eqr] = ACTIONS(2696),
    [anon_sym_ne] = ACTIONS(2696),
    [anon_sym_LT_EQ_GT] = ACTIONS(2696),
    [anon_sym_cmp] = ACTIONS(2696),
    [anon_sym_TILDE_TILDE] = ACTIONS(2696),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2696),
    [anon_sym_lt] = ACTIONS(2696),
    [anon_sym_le] = ACTIONS(2696),
    [anon_sym_ge] = ACTIONS(2696),
    [anon_sym_gt] = ACTIONS(2696),
    [anon_sym_isa] = ACTIONS(2696),
    [anon_sym_TILDE] = ACTIONS(2696),
    [anon_sym_BANG] = ACTIONS(2696),
    [anon_sym_PLUS_PLUS] = ACTIONS(2696),
    [anon_sym_DASH_DASH] = ACTIONS(2696),
    [anon_sym_QMARK] = ACTIONS(2696),
    [anon_sym_BSLASH] = ACTIONS(2698),
    [anon_sym_do] = ACTIONS(2696),
    [anon_sym_eval] = ACTIONS(2696),
    [anon_sym_await] = ACTIONS(2696),
    [anon_sym_field] = ACTIONS(2696),
    [anon_sym_local] = ACTIONS(2696),
    [anon_sym_dynamically] = ACTIONS(2696),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2696),
    [anon_sym_require] = ACTIONS(2696),
    [anon_sym_map] = ACTIONS(2696),
    [anon_sym_grep] = ACTIONS(2696),
    [anon_sym_sort] = ACTIONS(2696),
    [anon_sym_goto] = ACTIONS(2696),
    [anon_sym_return] = ACTIONS(2696),
    [anon_sym_undef] = ACTIONS(2696),
    [anon_sym_PERCENT2] = ACTIONS(2696),
    [anon_sym_AMP2] = ACTIONS(2696),
    [anon_sym_STAR2] = ACTIONS(2696),
    [anon_sym_EQ_GT] = ACTIONS(2696),
    [anon_sym_last] = ACTIONS(2696),
    [anon_sym_next] = ACTIONS(2696),
    [anon_sym_redo] = ACTIONS(2696),
    [anon_sym___FILE__] = ACTIONS(2696),
    [anon_sym___LINE__] = ACTIONS(2696),
    [anon_sym___PACKAGE__] = ACTIONS(2696),
    [anon_sym___SUB__] = ACTIONS(2696),
    [anon_sym_break] = ACTIONS(2696),
    [anon_sym_fork] = ACTIONS(2696),
    [anon_sym_getppid] = ACTIONS(2696),
    [anon_sym_time] = ACTIONS(2696),
    [anon_sym_times] = ACTIONS(2696),
    [anon_sym_wait] = ACTIONS(2696),
    [anon_sym_wantarray] = ACTIONS(2696),
    [anon_sym_abs] = ACTIONS(2696),
    [anon_sym_alarm] = ACTIONS(2696),
    [anon_sym_chop] = ACTIONS(2696),
    [anon_sym_chdir] = ACTIONS(2696),
    [anon_sym_close] = ACTIONS(2696),
    [anon_sym_closedir] = ACTIONS(2696),
    [anon_sym_caller] = ACTIONS(2696),
    [anon_sym_chomp] = ACTIONS(2696),
    [anon_sym_chr] = ACTIONS(2696),
    [anon_sym_cos] = ACTIONS(2696),
    [anon_sym_chroot] = ACTIONS(2696),
    [anon_sym_defined] = ACTIONS(2696),
    [anon_sym_delete] = ACTIONS(2696),
    [anon_sym_dbmclose] = ACTIONS(2696),
    [anon_sym_exists] = ACTIONS(2696),
    [anon_sym_exit] = ACTIONS(2696),
    [anon_sym_eof] = ACTIONS(2696),
    [anon_sym_exp] = ACTIONS(2696),
    [anon_sym_each] = ACTIONS(2696),
    [anon_sym_fc] = ACTIONS(2696),
    [anon_sym_fileno] = ACTIONS(2696),
    [anon_sym_gmtime] = ACTIONS(2696),
    [anon_sym_getc] = ACTIONS(2696),
    [anon_sym_getpgrp] = ACTIONS(2696),
    [anon_sym_getprotobyname] = ACTIONS(2696),
    [anon_sym_getpwname] = ACTIONS(2696),
    [anon_sym_getpwuid] = ACTIONS(2696),
    [anon_sym_getpeername] = ACTIONS(2696),
    [anon_sym_getnetbyname] = ACTIONS(2696),
    [anon_sym_getsockname] = ACTIONS(2696),
    [anon_sym_getgrnam] = ACTIONS(2696),
    [anon_sym_getgrgid] = ACTIONS(2696),
    [anon_sym_hex] = ACTIONS(2696),
    [anon_sym_int] = ACTIONS(2696),
    [anon_sym_keys] = ACTIONS(2696),
    [anon_sym_lc] = ACTIONS(2696),
    [anon_sym_lcfirst] = ACTIONS(2696),
    [anon_sym_length] = ACTIONS(2696),
    [anon_sym_localtime] = ACTIONS(2696),
    [anon_sym_log] = ACTIONS(2696),
    [anon_sym_lock] = ACTIONS(2696),
    [anon_sym_lstat] = ACTIONS(2696),
    [anon_sym_oct] = ACTIONS(2696),
    [anon_sym_ord] = ACTIONS(2696),
    [anon_sym_prototype] = ACTIONS(2696),
    [anon_sym_pop] = ACTIONS(2696),
    [anon_sym_pos] = ACTIONS(2696),
    [anon_sym_quotemeta] = ACTIONS(2696),
    [anon_sym_reset] = ACTIONS(2696),
    [anon_sym_rand] = ACTIONS(2696),
    [anon_sym_rmdir] = ACTIONS(2696),
    [anon_sym_readdir] = ACTIONS(2696),
    [anon_sym_readline] = ACTIONS(2696),
    [anon_sym_readpipe] = ACTIONS(2696),
    [anon_sym_rewinddir] = ACTIONS(2696),
    [anon_sym_readlink] = ACTIONS(2696),
    [anon_sym_ref] = ACTIONS(2696),
    [anon_sym_scalar] = ACTIONS(2696),
    [anon_sym_shift] = ACTIONS(2696),
    [anon_sym_sin] = ACTIONS(2696),
    [anon_sym_sleep] = ACTIONS(2696),
    [anon_sym_sqrt] = ACTIONS(2696),
    [anon_sym_srand] = ACTIONS(2696),
    [anon_sym_stat] = ACTIONS(2696),
    [anon_sym_study] = ACTIONS(2696),
    [anon_sym_tell] = ACTIONS(2696),
    [anon_sym_telldir] = ACTIONS(2696),
    [anon_sym_tied] = ACTIONS(2696),
    [anon_sym_uc] = ACTIONS(2696),
    [anon_sym_ucfirst] = ACTIONS(2696),
    [anon_sym_untie] = ACTIONS(2696),
    [anon_sym_umask] = ACTIONS(2696),
    [anon_sym_values] = ACTIONS(2696),
    [anon_sym_write] = ACTIONS(2696),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2696),
    [anon_sym_qq] = ACTIONS(2696),
    [anon_sym_qw] = ACTIONS(2696),
    [anon_sym_qx] = ACTIONS(2696),
    [anon_sym_qr] = ACTIONS(2696),
    [anon_sym_m] = ACTIONS(2696),
    [anon_sym_s] = ACTIONS(2696),
    [anon_sym_tr] = ACTIONS(2696),
    [anon_sym_y] = ACTIONS(2696),
    [aux_sym__bareword_token1] = ACTIONS(2696),
    [sym_number] = ACTIONS(2696),
    [anon_sym_true] = ACTIONS(2696),
    [anon_sym_false] = ACTIONS(2696),
    [sym__single_quote] = ACTIONS(2698),
    [sym__double_quote] = ACTIONS(2698),
    [sym__backtick_quote] = ACTIONS(2698),
    [sym__search_slash_quote] = ACTIONS(2698),
    [sym__open_readline_bracket] = ACTIONS(2698),
    [sym__open_fileglob_bracket] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__PERLY_HEREDOC] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2698),
    [sym__filetest] = ACTIONS(2698),
  },
  [STATE(1423)] = {
    [sym__else] = STATE(1473),
    [sym_else] = STATE(1478),
    [sym_elsif] = STATE(1478),
    [sym_heredoc_content] = STATE(1423),
    [ts_builtin_sym_end] = ACTIONS(2860),
    [sym__identifier] = ACTIONS(2862),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym_package] = ACTIONS(2862),
    [anon_sym_class] = ACTIONS(2862),
    [anon_sym_role] = ACTIONS(2862),
    [anon_sym_BUILD] = ACTIONS(2862),
    [anon_sym_ADJUST] = ACTIONS(2862),
    [anon_sym_DOLLAR] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2862),
    [anon_sym_my] = ACTIONS(2862),
    [anon_sym_extended] = ACTIONS(2862),
    [anon_sym_async] = ACTIONS(2862),
    [anon_sym_sub] = ACTIONS(2862),
    [anon_sym_method] = ACTIONS(2862),
    [anon_sym_LPAREN] = ACTIONS(2862),
    [anon_sym_continue] = ACTIONS(2862),
    [anon_sym_state] = ACTIONS(2862),
    [anon_sym_our] = ACTIONS(2862),
    [anon_sym_try] = ACTIONS(2862),
    [anon_sym_defer] = ACTIONS(2862),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2862),
    [anon_sym_else] = ACTIONS(2864),
    [anon_sym_elsif] = ACTIONS(2866),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym_LT_LT] = ACTIONS(2862),
    [anon_sym_SLASH_SLASH] = ACTIONS(2862),
    [anon_sym_PLUS] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_BSLASH] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2862),
    [anon_sym_eval] = ACTIONS(2862),
    [anon_sym_await] = ACTIONS(2862),
    [anon_sym_field] = ACTIONS(2862),
    [anon_sym_local] = ACTIONS(2862),
    [anon_sym_dynamically] = ACTIONS(2862),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2862),
    [anon_sym_require] = ACTIONS(2862),
    [anon_sym_map] = ACTIONS(2862),
    [anon_sym_grep] = ACTIONS(2862),
    [anon_sym_sort] = ACTIONS(2862),
    [anon_sym_goto] = ACTIONS(2862),
    [anon_sym_return] = ACTIONS(2862),
    [anon_sym_undef] = ACTIONS(2862),
    [anon_sym_PERCENT2] = ACTIONS(2862),
    [anon_sym_AMP2] = ACTIONS(2862),
    [anon_sym_STAR2] = ACTIONS(2862),
    [anon_sym_use] = ACTIONS(2862),
    [anon_sym_no] = ACTIONS(2862),
    [anon_sym_for] = ACTIONS(2862),
    [anon_sym_foreach] = ACTIONS(2862),
    [anon_sym_last] = ACTIONS(2862),
    [anon_sym_next] = ACTIONS(2862),
    [anon_sym_redo] = ACTIONS(2862),
    [anon_sym_BEGIN] = ACTIONS(2862),
    [anon_sym_INIT] = ACTIONS(2862),
    [anon_sym_CHECK] = ACTIONS(2862),
    [anon_sym_UNITCHECK] = ACTIONS(2862),
    [anon_sym_END] = ACTIONS(2862),
    [anon_sym___FILE__] = ACTIONS(2862),
    [anon_sym___LINE__] = ACTIONS(2862),
    [anon_sym___PACKAGE__] = ACTIONS(2862),
    [anon_sym___SUB__] = ACTIONS(2862),
    [anon_sym_break] = ACTIONS(2862),
    [anon_sym_fork] = ACTIONS(2862),
    [anon_sym_getppid] = ACTIONS(2862),
    [anon_sym_time] = ACTIONS(2862),
    [anon_sym_times] = ACTIONS(2862),
    [anon_sym_wait] = ACTIONS(2862),
    [anon_sym_wantarray] = ACTIONS(2862),
    [anon_sym_abs] = ACTIONS(2862),
    [anon_sym_alarm] = ACTIONS(2862),
    [anon_sym_chop] = ACTIONS(2862),
    [anon_sym_chdir] = ACTIONS(2862),
    [anon_sym_close] = ACTIONS(2862),
    [anon_sym_closedir] = ACTIONS(2862),
    [anon_sym_caller] = ACTIONS(2862),
    [anon_sym_chomp] = ACTIONS(2862),
    [anon_sym_chr] = ACTIONS(2862),
    [anon_sym_cos] = ACTIONS(2862),
    [anon_sym_chroot] = ACTIONS(2862),
    [anon_sym_defined] = ACTIONS(2862),
    [anon_sym_delete] = ACTIONS(2862),
    [anon_sym_dbmclose] = ACTIONS(2862),
    [anon_sym_exists] = ACTIONS(2862),
    [anon_sym_exit] = ACTIONS(2862),
    [anon_sym_eof] = ACTIONS(2862),
    [anon_sym_exp] = ACTIONS(2862),
    [anon_sym_each] = ACTIONS(2862),
    [anon_sym_fc] = ACTIONS(2862),
    [anon_sym_fileno] = ACTIONS(2862),
    [anon_sym_gmtime] = ACTIONS(2862),
    [anon_sym_getc] = ACTIONS(2862),
    [anon_sym_getpgrp] = ACTIONS(2862),
    [anon_sym_getprotobyname] = ACTIONS(2862),
    [anon_sym_getpwname] = ACTIONS(2862),
    [anon_sym_getpwuid] = ACTIONS(2862),
    [anon_sym_getpeername] = ACTIONS(2862),
    [anon_sym_getnetbyname] = ACTIONS(2862),
    [anon_sym_getsockname] = ACTIONS(2862),
    [anon_sym_getgrnam] = ACTIONS(2862),
    [anon_sym_getgrgid] = ACTIONS(2862),
    [anon_sym_hex] = ACTIONS(2862),
    [anon_sym_int] = ACTIONS(2862),
    [anon_sym_keys] = ACTIONS(2862),
    [anon_sym_lc] = ACTIONS(2862),
    [anon_sym_lcfirst] = ACTIONS(2862),
    [anon_sym_length] = ACTIONS(2862),
    [anon_sym_localtime] = ACTIONS(2862),
    [anon_sym_log] = ACTIONS(2862),
    [anon_sym_lock] = ACTIONS(2862),
    [anon_sym_lstat] = ACTIONS(2862),
    [anon_sym_oct] = ACTIONS(2862),
    [anon_sym_ord] = ACTIONS(2862),
    [anon_sym_prototype] = ACTIONS(2862),
    [anon_sym_pop] = ACTIONS(2862),
    [anon_sym_pos] = ACTIONS(2862),
    [anon_sym_quotemeta] = ACTIONS(2862),
    [anon_sym_reset] = ACTIONS(2862),
    [anon_sym_rand] = ACTIONS(2862),
    [anon_sym_rmdir] = ACTIONS(2862),
    [anon_sym_readdir] = ACTIONS(2862),
    [anon_sym_readline] = ACTIONS(2862),
    [anon_sym_readpipe] = ACTIONS(2862),
    [anon_sym_rewinddir] = ACTIONS(2862),
    [anon_sym_readlink] = ACTIONS(2862),
    [anon_sym_ref] = ACTIONS(2862),
    [anon_sym_scalar] = ACTIONS(2862),
    [anon_sym_shift] = ACTIONS(2862),
    [anon_sym_sin] = ACTIONS(2862),
    [anon_sym_sleep] = ACTIONS(2862),
    [anon_sym_sqrt] = ACTIONS(2862),
    [anon_sym_srand] = ACTIONS(2862),
    [anon_sym_stat] = ACTIONS(2862),
    [anon_sym_study] = ACTIONS(2862),
    [anon_sym_tell] = ACTIONS(2862),
    [anon_sym_telldir] = ACTIONS(2862),
    [anon_sym_tied] = ACTIONS(2862),
    [anon_sym_uc] = ACTIONS(2862),
    [anon_sym_ucfirst] = ACTIONS(2862),
    [anon_sym_untie] = ACTIONS(2862),
    [anon_sym_umask] = ACTIONS(2862),
    [anon_sym_values] = ACTIONS(2862),
    [anon_sym_write] = ACTIONS(2862),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2862),
    [anon_sym___END__] = ACTIONS(2862),
    [anon_sym_EOT] = ACTIONS(2862),
    [anon_sym_q] = ACTIONS(2862),
    [anon_sym_qq] = ACTIONS(2862),
    [anon_sym_qw] = ACTIONS(2862),
    [anon_sym_qx] = ACTIONS(2862),
    [anon_sym_qr] = ACTIONS(2862),
    [anon_sym_m] = ACTIONS(2862),
    [anon_sym_s] = ACTIONS(2862),
    [anon_sym_tr] = ACTIONS(2862),
    [anon_sym_y] = ACTIONS(2862),
    [anon_sym_if] = ACTIONS(2862),
    [anon_sym_unless] = ACTIONS(2862),
    [anon_sym_while] = ACTIONS(2862),
    [anon_sym_until] = ACTIONS(2862),
    [aux_sym__bareword_token1] = ACTIONS(2862),
    [sym_number] = ACTIONS(2862),
    [anon_sym_true] = ACTIONS(2862),
    [anon_sym_false] = ACTIONS(2862),
    [sym__single_quote] = ACTIONS(2860),
    [sym__double_quote] = ACTIONS(2860),
    [sym__backtick_quote] = ACTIONS(2860),
    [sym__search_slash_quote] = ACTIONS(2860),
    [sym__open_readline_bracket] = ACTIONS(2860),
    [sym__open_fileglob_bracket] = ACTIONS(2860),
    [sym__PERLY_HEREDOC] = ACTIONS(2860),
    [sym__ctrl_z_hack] = ACTIONS(2860),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2860),
    [sym__filetest] = ACTIONS(2860),
  },
  [STATE(1424)] = {
    [sym__else] = STATE(1507),
    [sym_else] = STATE(1478),
    [sym_elsif] = STATE(1478),
    [sym_heredoc_content] = STATE(1424),
    [ts_builtin_sym_end] = ACTIONS(2868),
    [sym__identifier] = ACTIONS(2870),
    [anon_sym_LBRACE] = ACTIONS(2870),
    [anon_sym_SEMI] = ACTIONS(2870),
    [anon_sym_package] = ACTIONS(2870),
    [anon_sym_class] = ACTIONS(2870),
    [anon_sym_role] = ACTIONS(2870),
    [anon_sym_BUILD] = ACTIONS(2870),
    [anon_sym_ADJUST] = ACTIONS(2870),
    [anon_sym_DOLLAR] = ACTIONS(2870),
    [anon_sym_AT] = ACTIONS(2870),
    [anon_sym_my] = ACTIONS(2870),
    [anon_sym_extended] = ACTIONS(2870),
    [anon_sym_async] = ACTIONS(2870),
    [anon_sym_sub] = ACTIONS(2870),
    [anon_sym_method] = ACTIONS(2870),
    [anon_sym_LPAREN] = ACTIONS(2870),
    [anon_sym_continue] = ACTIONS(2870),
    [anon_sym_state] = ACTIONS(2870),
    [anon_sym_our] = ACTIONS(2870),
    [anon_sym_try] = ACTIONS(2870),
    [anon_sym_defer] = ACTIONS(2870),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2870),
    [anon_sym_else] = ACTIONS(2864),
    [anon_sym_elsif] = ACTIONS(2866),
    [anon_sym_LBRACK] = ACTIONS(2870),
    [anon_sym_LT_LT] = ACTIONS(2870),
    [anon_sym_SLASH_SLASH] = ACTIONS(2870),
    [anon_sym_PLUS] = ACTIONS(2870),
    [anon_sym_DASH] = ACTIONS(2870),
    [anon_sym_TILDE] = ACTIONS(2870),
    [anon_sym_BANG] = ACTIONS(2870),
    [anon_sym_PLUS_PLUS] = ACTIONS(2870),
    [anon_sym_DASH_DASH] = ACTIONS(2870),
    [anon_sym_BSLASH] = ACTIONS(2868),
    [anon_sym_do] = ACTIONS(2870),
    [anon_sym_eval] = ACTIONS(2870),
    [anon_sym_await] = ACTIONS(2870),
    [anon_sym_field] = ACTIONS(2870),
    [anon_sym_local] = ACTIONS(2870),
    [anon_sym_dynamically] = ACTIONS(2870),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2870),
    [anon_sym_require] = ACTIONS(2870),
    [anon_sym_map] = ACTIONS(2870),
    [anon_sym_grep] = ACTIONS(2870),
    [anon_sym_sort] = ACTIONS(2870),
    [anon_sym_goto] = ACTIONS(2870),
    [anon_sym_return] = ACTIONS(2870),
    [anon_sym_undef] = ACTIONS(2870),
    [anon_sym_PERCENT2] = ACTIONS(2870),
    [anon_sym_AMP2] = ACTIONS(2870),
    [anon_sym_STAR2] = ACTIONS(2870),
    [anon_sym_use] = ACTIONS(2870),
    [anon_sym_no] = ACTIONS(2870),
    [anon_sym_for] = ACTIONS(2870),
    [anon_sym_foreach] = ACTIONS(2870),
    [anon_sym_last] = ACTIONS(2870),
    [anon_sym_next] = ACTIONS(2870),
    [anon_sym_redo] = ACTIONS(2870),
    [anon_sym_BEGIN] = ACTIONS(2870),
    [anon_sym_INIT] = ACTIONS(2870),
    [anon_sym_CHECK] = ACTIONS(2870),
    [anon_sym_UNITCHECK] = ACTIONS(2870),
    [anon_sym_END] = ACTIONS(2870),
    [anon_sym___FILE__] = ACTIONS(2870),
    [anon_sym___LINE__] = ACTIONS(2870),
    [anon_sym___PACKAGE__] = ACTIONS(2870),
    [anon_sym___SUB__] = ACTIONS(2870),
    [anon_sym_break] = ACTIONS(2870),
    [anon_sym_fork] = ACTIONS(2870),
    [anon_sym_getppid] = ACTIONS(2870),
    [anon_sym_time] = ACTIONS(2870),
    [anon_sym_times] = ACTIONS(2870),
    [anon_sym_wait] = ACTIONS(2870),
    [anon_sym_wantarray] = ACTIONS(2870),
    [anon_sym_abs] = ACTIONS(2870),
    [anon_sym_alarm] = ACTIONS(2870),
    [anon_sym_chop] = ACTIONS(2870),
    [anon_sym_chdir] = ACTIONS(2870),
    [anon_sym_close] = ACTIONS(2870),
    [anon_sym_closedir] = ACTIONS(2870),
    [anon_sym_caller] = ACTIONS(2870),
    [anon_sym_chomp] = ACTIONS(2870),
    [anon_sym_chr] = ACTIONS(2870),
    [anon_sym_cos] = ACTIONS(2870),
    [anon_sym_chroot] = ACTIONS(2870),
    [anon_sym_defined] = ACTIONS(2870),
    [anon_sym_delete] = ACTIONS(2870),
    [anon_sym_dbmclose] = ACTIONS(2870),
    [anon_sym_exists] = ACTIONS(2870),
    [anon_sym_exit] = ACTIONS(2870),
    [anon_sym_eof] = ACTIONS(2870),
    [anon_sym_exp] = ACTIONS(2870),
    [anon_sym_each] = ACTIONS(2870),
    [anon_sym_fc] = ACTIONS(2870),
    [anon_sym_fileno] = ACTIONS(2870),
    [anon_sym_gmtime] = ACTIONS(2870),
    [anon_sym_getc] = ACTIONS(2870),
    [anon_sym_getpgrp] = ACTIONS(2870),
    [anon_sym_getprotobyname] = ACTIONS(2870),
    [anon_sym_getpwname] = ACTIONS(2870),
    [anon_sym_getpwuid] = ACTIONS(2870),
    [anon_sym_getpeername] = ACTIONS(2870),
    [anon_sym_getnetbyname] = ACTIONS(2870),
    [anon_sym_getsockname] = ACTIONS(2870),
    [anon_sym_getgrnam] = ACTIONS(2870),
    [anon_sym_getgrgid] = ACTIONS(2870),
    [anon_sym_hex] = ACTIONS(2870),
    [anon_sym_int] = ACTIONS(2870),
    [anon_sym_keys] = ACTIONS(2870),
    [anon_sym_lc] = ACTIONS(2870),
    [anon_sym_lcfirst] = ACTIONS(2870),
    [anon_sym_length] = ACTIONS(2870),
    [anon_sym_localtime] = ACTIONS(2870),
    [anon_sym_log] = ACTIONS(2870),
    [anon_sym_lock] = ACTIONS(2870),
    [anon_sym_lstat] = ACTIONS(2870),
    [anon_sym_oct] = ACTIONS(2870),
    [anon_sym_ord] = ACTIONS(2870),
    [anon_sym_prototype] = ACTIONS(2870),
    [anon_sym_pop] = ACTIONS(2870),
    [anon_sym_pos] = ACTIONS(2870),
    [anon_sym_quotemeta] = ACTIONS(2870),
    [anon_sym_reset] = ACTIONS(2870),
    [anon_sym_rand] = ACTIONS(2870),
    [anon_sym_rmdir] = ACTIONS(2870),
    [anon_sym_readdir] = ACTIONS(2870),
    [anon_sym_readline] = ACTIONS(2870),
    [anon_sym_readpipe] = ACTIONS(2870),
    [anon_sym_rewinddir] = ACTIONS(2870),
    [anon_sym_readlink] = ACTIONS(2870),
    [anon_sym_ref] = ACTIONS(2870),
    [anon_sym_scalar] = ACTIONS(2870),
    [anon_sym_shift] = ACTIONS(2870),
    [anon_sym_sin] = ACTIONS(2870),
    [anon_sym_sleep] = ACTIONS(2870),
    [anon_sym_sqrt] = ACTIONS(2870),
    [anon_sym_srand] = ACTIONS(2870),
    [anon_sym_stat] = ACTIONS(2870),
    [anon_sym_study] = ACTIONS(2870),
    [anon_sym_tell] = ACTIONS(2870),
    [anon_sym_telldir] = ACTIONS(2870),
    [anon_sym_tied] = ACTIONS(2870),
    [anon_sym_uc] = ACTIONS(2870),
    [anon_sym_ucfirst] = ACTIONS(2870),
    [anon_sym_untie] = ACTIONS(2870),
    [anon_sym_umask] = ACTIONS(2870),
    [anon_sym_values] = ACTIONS(2870),
    [anon_sym_write] = ACTIONS(2870),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2870),
    [anon_sym___END__] = ACTIONS(2870),
    [anon_sym_EOT] = ACTIONS(2870),
    [anon_sym_q] = ACTIONS(2870),
    [anon_sym_qq] = ACTIONS(2870),
    [anon_sym_qw] = ACTIONS(2870),
    [anon_sym_qx] = ACTIONS(2870),
    [anon_sym_qr] = ACTIONS(2870),
    [anon_sym_m] = ACTIONS(2870),
    [anon_sym_s] = ACTIONS(2870),
    [anon_sym_tr] = ACTIONS(2870),
    [anon_sym_y] = ACTIONS(2870),
    [anon_sym_if] = ACTIONS(2870),
    [anon_sym_unless] = ACTIONS(2870),
    [anon_sym_while] = ACTIONS(2870),
    [anon_sym_until] = ACTIONS(2870),
    [aux_sym__bareword_token1] = ACTIONS(2870),
    [sym_number] = ACTIONS(2870),
    [anon_sym_true] = ACTIONS(2870),
    [anon_sym_false] = ACTIONS(2870),
    [sym__single_quote] = ACTIONS(2868),
    [sym__double_quote] = ACTIONS(2868),
    [sym__backtick_quote] = ACTIONS(2868),
    [sym__search_slash_quote] = ACTIONS(2868),
    [sym__open_readline_bracket] = ACTIONS(2868),
    [sym__open_fileglob_bracket] = ACTIONS(2868),
    [sym__PERLY_HEREDOC] = ACTIONS(2868),
    [sym__ctrl_z_hack] = ACTIONS(2868),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2868),
    [sym__filetest] = ACTIONS(2868),
  },
  [STATE(1425)] = {
    [sym_heredoc_content] = STATE(1425),
    [ts_builtin_sym_end] = ACTIONS(2530),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_else] = ACTIONS(2532),
    [anon_sym_elsif] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1426)] = {
    [sym_heredoc_content] = STATE(1426),
    [ts_builtin_sym_end] = ACTIONS(2690),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_catch] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1427)] = {
    [sym_heredoc_content] = STATE(1427),
    [ts_builtin_sym_end] = ACTIONS(2530),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_catch] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1428)] = {
    [sym_heredoc_content] = STATE(1428),
    [ts_builtin_sym_end] = ACTIONS(2872),
    [sym__identifier] = ACTIONS(2874),
    [anon_sym_LBRACE] = ACTIONS(2874),
    [anon_sym_SEMI] = ACTIONS(2874),
    [anon_sym_package] = ACTIONS(2874),
    [anon_sym_class] = ACTIONS(2874),
    [anon_sym_role] = ACTIONS(2874),
    [anon_sym_BUILD] = ACTIONS(2874),
    [anon_sym_ADJUST] = ACTIONS(2874),
    [anon_sym_DOLLAR] = ACTIONS(2874),
    [anon_sym_AT] = ACTIONS(2874),
    [anon_sym_my] = ACTIONS(2874),
    [anon_sym_extended] = ACTIONS(2874),
    [anon_sym_async] = ACTIONS(2874),
    [anon_sym_sub] = ACTIONS(2874),
    [anon_sym_method] = ACTIONS(2874),
    [anon_sym_LPAREN] = ACTIONS(2874),
    [anon_sym_continue] = ACTIONS(2874),
    [anon_sym_state] = ACTIONS(2874),
    [anon_sym_our] = ACTIONS(2874),
    [anon_sym_try] = ACTIONS(2874),
    [anon_sym_catch] = ACTIONS(2876),
    [anon_sym_finally] = ACTIONS(2878),
    [anon_sym_defer] = ACTIONS(2874),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2874),
    [anon_sym_LBRACK] = ACTIONS(2874),
    [anon_sym_LT_LT] = ACTIONS(2874),
    [anon_sym_SLASH_SLASH] = ACTIONS(2874),
    [anon_sym_PLUS] = ACTIONS(2874),
    [anon_sym_DASH] = ACTIONS(2874),
    [anon_sym_TILDE] = ACTIONS(2874),
    [anon_sym_BANG] = ACTIONS(2874),
    [anon_sym_PLUS_PLUS] = ACTIONS(2874),
    [anon_sym_DASH_DASH] = ACTIONS(2874),
    [anon_sym_BSLASH] = ACTIONS(2872),
    [anon_sym_do] = ACTIONS(2874),
    [anon_sym_eval] = ACTIONS(2874),
    [anon_sym_await] = ACTIONS(2874),
    [anon_sym_field] = ACTIONS(2874),
    [anon_sym_local] = ACTIONS(2874),
    [anon_sym_dynamically] = ACTIONS(2874),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2874),
    [anon_sym_require] = ACTIONS(2874),
    [anon_sym_map] = ACTIONS(2874),
    [anon_sym_grep] = ACTIONS(2874),
    [anon_sym_sort] = ACTIONS(2874),
    [anon_sym_goto] = ACTIONS(2874),
    [anon_sym_return] = ACTIONS(2874),
    [anon_sym_undef] = ACTIONS(2874),
    [anon_sym_PERCENT2] = ACTIONS(2874),
    [anon_sym_AMP2] = ACTIONS(2874),
    [anon_sym_STAR2] = ACTIONS(2874),
    [anon_sym_use] = ACTIONS(2874),
    [anon_sym_no] = ACTIONS(2874),
    [anon_sym_for] = ACTIONS(2874),
    [anon_sym_foreach] = ACTIONS(2874),
    [anon_sym_last] = ACTIONS(2874),
    [anon_sym_next] = ACTIONS(2874),
    [anon_sym_redo] = ACTIONS(2874),
    [anon_sym_BEGIN] = ACTIONS(2874),
    [anon_sym_INIT] = ACTIONS(2874),
    [anon_sym_CHECK] = ACTIONS(2874),
    [anon_sym_UNITCHECK] = ACTIONS(2874),
    [anon_sym_END] = ACTIONS(2874),
    [anon_sym___FILE__] = ACTIONS(2874),
    [anon_sym___LINE__] = ACTIONS(2874),
    [anon_sym___PACKAGE__] = ACTIONS(2874),
    [anon_sym___SUB__] = ACTIONS(2874),
    [anon_sym_break] = ACTIONS(2874),
    [anon_sym_fork] = ACTIONS(2874),
    [anon_sym_getppid] = ACTIONS(2874),
    [anon_sym_time] = ACTIONS(2874),
    [anon_sym_times] = ACTIONS(2874),
    [anon_sym_wait] = ACTIONS(2874),
    [anon_sym_wantarray] = ACTIONS(2874),
    [anon_sym_abs] = ACTIONS(2874),
    [anon_sym_alarm] = ACTIONS(2874),
    [anon_sym_chop] = ACTIONS(2874),
    [anon_sym_chdir] = ACTIONS(2874),
    [anon_sym_close] = ACTIONS(2874),
    [anon_sym_closedir] = ACTIONS(2874),
    [anon_sym_caller] = ACTIONS(2874),
    [anon_sym_chomp] = ACTIONS(2874),
    [anon_sym_chr] = ACTIONS(2874),
    [anon_sym_cos] = ACTIONS(2874),
    [anon_sym_chroot] = ACTIONS(2874),
    [anon_sym_defined] = ACTIONS(2874),
    [anon_sym_delete] = ACTIONS(2874),
    [anon_sym_dbmclose] = ACTIONS(2874),
    [anon_sym_exists] = ACTIONS(2874),
    [anon_sym_exit] = ACTIONS(2874),
    [anon_sym_eof] = ACTIONS(2874),
    [anon_sym_exp] = ACTIONS(2874),
    [anon_sym_each] = ACTIONS(2874),
    [anon_sym_fc] = ACTIONS(2874),
    [anon_sym_fileno] = ACTIONS(2874),
    [anon_sym_gmtime] = ACTIONS(2874),
    [anon_sym_getc] = ACTIONS(2874),
    [anon_sym_getpgrp] = ACTIONS(2874),
    [anon_sym_getprotobyname] = ACTIONS(2874),
    [anon_sym_getpwname] = ACTIONS(2874),
    [anon_sym_getpwuid] = ACTIONS(2874),
    [anon_sym_getpeername] = ACTIONS(2874),
    [anon_sym_getnetbyname] = ACTIONS(2874),
    [anon_sym_getsockname] = ACTIONS(2874),
    [anon_sym_getgrnam] = ACTIONS(2874),
    [anon_sym_getgrgid] = ACTIONS(2874),
    [anon_sym_hex] = ACTIONS(2874),
    [anon_sym_int] = ACTIONS(2874),
    [anon_sym_keys] = ACTIONS(2874),
    [anon_sym_lc] = ACTIONS(2874),
    [anon_sym_lcfirst] = ACTIONS(2874),
    [anon_sym_length] = ACTIONS(2874),
    [anon_sym_localtime] = ACTIONS(2874),
    [anon_sym_log] = ACTIONS(2874),
    [anon_sym_lock] = ACTIONS(2874),
    [anon_sym_lstat] = ACTIONS(2874),
    [anon_sym_oct] = ACTIONS(2874),
    [anon_sym_ord] = ACTIONS(2874),
    [anon_sym_prototype] = ACTIONS(2874),
    [anon_sym_pop] = ACTIONS(2874),
    [anon_sym_pos] = ACTIONS(2874),
    [anon_sym_quotemeta] = ACTIONS(2874),
    [anon_sym_reset] = ACTIONS(2874),
    [anon_sym_rand] = ACTIONS(2874),
    [anon_sym_rmdir] = ACTIONS(2874),
    [anon_sym_readdir] = ACTIONS(2874),
    [anon_sym_readline] = ACTIONS(2874),
    [anon_sym_readpipe] = ACTIONS(2874),
    [anon_sym_rewinddir] = ACTIONS(2874),
    [anon_sym_readlink] = ACTIONS(2874),
    [anon_sym_ref] = ACTIONS(2874),
    [anon_sym_scalar] = ACTIONS(2874),
    [anon_sym_shift] = ACTIONS(2874),
    [anon_sym_sin] = ACTIONS(2874),
    [anon_sym_sleep] = ACTIONS(2874),
    [anon_sym_sqrt] = ACTIONS(2874),
    [anon_sym_srand] = ACTIONS(2874),
    [anon_sym_stat] = ACTIONS(2874),
    [anon_sym_study] = ACTIONS(2874),
    [anon_sym_tell] = ACTIONS(2874),
    [anon_sym_telldir] = ACTIONS(2874),
    [anon_sym_tied] = ACTIONS(2874),
    [anon_sym_uc] = ACTIONS(2874),
    [anon_sym_ucfirst] = ACTIONS(2874),
    [anon_sym_untie] = ACTIONS(2874),
    [anon_sym_umask] = ACTIONS(2874),
    [anon_sym_values] = ACTIONS(2874),
    [anon_sym_write] = ACTIONS(2874),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2874),
    [anon_sym___END__] = ACTIONS(2874),
    [anon_sym_EOT] = ACTIONS(2874),
    [anon_sym_q] = ACTIONS(2874),
    [anon_sym_qq] = ACTIONS(2874),
    [anon_sym_qw] = ACTIONS(2874),
    [anon_sym_qx] = ACTIONS(2874),
    [anon_sym_qr] = ACTIONS(2874),
    [anon_sym_m] = ACTIONS(2874),
    [anon_sym_s] = ACTIONS(2874),
    [anon_sym_tr] = ACTIONS(2874),
    [anon_sym_y] = ACTIONS(2874),
    [anon_sym_if] = ACTIONS(2874),
    [anon_sym_unless] = ACTIONS(2874),
    [anon_sym_while] = ACTIONS(2874),
    [anon_sym_until] = ACTIONS(2874),
    [aux_sym__bareword_token1] = ACTIONS(2874),
    [sym_number] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2874),
    [anon_sym_false] = ACTIONS(2874),
    [sym__single_quote] = ACTIONS(2872),
    [sym__double_quote] = ACTIONS(2872),
    [sym__backtick_quote] = ACTIONS(2872),
    [sym__search_slash_quote] = ACTIONS(2872),
    [sym__open_readline_bracket] = ACTIONS(2872),
    [sym__open_fileglob_bracket] = ACTIONS(2872),
    [sym__PERLY_HEREDOC] = ACTIONS(2872),
    [sym__ctrl_z_hack] = ACTIONS(2872),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2872),
    [sym__filetest] = ACTIONS(2872),
  },
  [STATE(1429)] = {
    [sym_heredoc_content] = STATE(1429),
    [ts_builtin_sym_end] = ACTIONS(2690),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_else] = ACTIONS(2688),
    [anon_sym_elsif] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1430)] = {
    [sym_heredoc_content] = STATE(1430),
    [ts_builtin_sym_end] = ACTIONS(2530),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1431)] = {
    [sym__else] = STATE(1613),
    [sym_else] = STATE(1594),
    [sym_elsif] = STATE(1594),
    [sym_heredoc_content] = STATE(1431),
    [sym__identifier] = ACTIONS(2870),
    [anon_sym_LBRACE] = ACTIONS(2870),
    [anon_sym_RBRACE] = ACTIONS(2870),
    [anon_sym_SEMI] = ACTIONS(2870),
    [anon_sym_package] = ACTIONS(2870),
    [anon_sym_class] = ACTIONS(2870),
    [anon_sym_role] = ACTIONS(2870),
    [anon_sym_BUILD] = ACTIONS(2870),
    [anon_sym_ADJUST] = ACTIONS(2870),
    [anon_sym_DOLLAR] = ACTIONS(2870),
    [anon_sym_AT] = ACTIONS(2870),
    [anon_sym_my] = ACTIONS(2870),
    [anon_sym_extended] = ACTIONS(2870),
    [anon_sym_async] = ACTIONS(2870),
    [anon_sym_sub] = ACTIONS(2870),
    [anon_sym_method] = ACTIONS(2870),
    [anon_sym_LPAREN] = ACTIONS(2870),
    [anon_sym_continue] = ACTIONS(2870),
    [anon_sym_state] = ACTIONS(2870),
    [anon_sym_our] = ACTIONS(2870),
    [anon_sym_try] = ACTIONS(2870),
    [anon_sym_defer] = ACTIONS(2870),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2870),
    [anon_sym_else] = ACTIONS(2880),
    [anon_sym_elsif] = ACTIONS(2882),
    [anon_sym_LBRACK] = ACTIONS(2870),
    [anon_sym_LT_LT] = ACTIONS(2870),
    [anon_sym_SLASH_SLASH] = ACTIONS(2870),
    [anon_sym_PLUS] = ACTIONS(2870),
    [anon_sym_DASH] = ACTIONS(2870),
    [anon_sym_TILDE] = ACTIONS(2870),
    [anon_sym_BANG] = ACTIONS(2870),
    [anon_sym_PLUS_PLUS] = ACTIONS(2870),
    [anon_sym_DASH_DASH] = ACTIONS(2870),
    [anon_sym_BSLASH] = ACTIONS(2868),
    [anon_sym_do] = ACTIONS(2870),
    [anon_sym_eval] = ACTIONS(2870),
    [anon_sym_await] = ACTIONS(2870),
    [anon_sym_field] = ACTIONS(2870),
    [anon_sym_local] = ACTIONS(2870),
    [anon_sym_dynamically] = ACTIONS(2870),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2870),
    [anon_sym_require] = ACTIONS(2870),
    [anon_sym_map] = ACTIONS(2870),
    [anon_sym_grep] = ACTIONS(2870),
    [anon_sym_sort] = ACTIONS(2870),
    [anon_sym_goto] = ACTIONS(2870),
    [anon_sym_return] = ACTIONS(2870),
    [anon_sym_undef] = ACTIONS(2870),
    [anon_sym_PERCENT2] = ACTIONS(2870),
    [anon_sym_AMP2] = ACTIONS(2870),
    [anon_sym_STAR2] = ACTIONS(2870),
    [anon_sym_use] = ACTIONS(2870),
    [anon_sym_no] = ACTIONS(2870),
    [anon_sym_for] = ACTIONS(2870),
    [anon_sym_foreach] = ACTIONS(2870),
    [anon_sym_last] = ACTIONS(2870),
    [anon_sym_next] = ACTIONS(2870),
    [anon_sym_redo] = ACTIONS(2870),
    [anon_sym_BEGIN] = ACTIONS(2870),
    [anon_sym_INIT] = ACTIONS(2870),
    [anon_sym_CHECK] = ACTIONS(2870),
    [anon_sym_UNITCHECK] = ACTIONS(2870),
    [anon_sym_END] = ACTIONS(2870),
    [anon_sym___FILE__] = ACTIONS(2870),
    [anon_sym___LINE__] = ACTIONS(2870),
    [anon_sym___PACKAGE__] = ACTIONS(2870),
    [anon_sym___SUB__] = ACTIONS(2870),
    [anon_sym_break] = ACTIONS(2870),
    [anon_sym_fork] = ACTIONS(2870),
    [anon_sym_getppid] = ACTIONS(2870),
    [anon_sym_time] = ACTIONS(2870),
    [anon_sym_times] = ACTIONS(2870),
    [anon_sym_wait] = ACTIONS(2870),
    [anon_sym_wantarray] = ACTIONS(2870),
    [anon_sym_abs] = ACTIONS(2870),
    [anon_sym_alarm] = ACTIONS(2870),
    [anon_sym_chop] = ACTIONS(2870),
    [anon_sym_chdir] = ACTIONS(2870),
    [anon_sym_close] = ACTIONS(2870),
    [anon_sym_closedir] = ACTIONS(2870),
    [anon_sym_caller] = ACTIONS(2870),
    [anon_sym_chomp] = ACTIONS(2870),
    [anon_sym_chr] = ACTIONS(2870),
    [anon_sym_cos] = ACTIONS(2870),
    [anon_sym_chroot] = ACTIONS(2870),
    [anon_sym_defined] = ACTIONS(2870),
    [anon_sym_delete] = ACTIONS(2870),
    [anon_sym_dbmclose] = ACTIONS(2870),
    [anon_sym_exists] = ACTIONS(2870),
    [anon_sym_exit] = ACTIONS(2870),
    [anon_sym_eof] = ACTIONS(2870),
    [anon_sym_exp] = ACTIONS(2870),
    [anon_sym_each] = ACTIONS(2870),
    [anon_sym_fc] = ACTIONS(2870),
    [anon_sym_fileno] = ACTIONS(2870),
    [anon_sym_gmtime] = ACTIONS(2870),
    [anon_sym_getc] = ACTIONS(2870),
    [anon_sym_getpgrp] = ACTIONS(2870),
    [anon_sym_getprotobyname] = ACTIONS(2870),
    [anon_sym_getpwname] = ACTIONS(2870),
    [anon_sym_getpwuid] = ACTIONS(2870),
    [anon_sym_getpeername] = ACTIONS(2870),
    [anon_sym_getnetbyname] = ACTIONS(2870),
    [anon_sym_getsockname] = ACTIONS(2870),
    [anon_sym_getgrnam] = ACTIONS(2870),
    [anon_sym_getgrgid] = ACTIONS(2870),
    [anon_sym_hex] = ACTIONS(2870),
    [anon_sym_int] = ACTIONS(2870),
    [anon_sym_keys] = ACTIONS(2870),
    [anon_sym_lc] = ACTIONS(2870),
    [anon_sym_lcfirst] = ACTIONS(2870),
    [anon_sym_length] = ACTIONS(2870),
    [anon_sym_localtime] = ACTIONS(2870),
    [anon_sym_log] = ACTIONS(2870),
    [anon_sym_lock] = ACTIONS(2870),
    [anon_sym_lstat] = ACTIONS(2870),
    [anon_sym_oct] = ACTIONS(2870),
    [anon_sym_ord] = ACTIONS(2870),
    [anon_sym_prototype] = ACTIONS(2870),
    [anon_sym_pop] = ACTIONS(2870),
    [anon_sym_pos] = ACTIONS(2870),
    [anon_sym_quotemeta] = ACTIONS(2870),
    [anon_sym_reset] = ACTIONS(2870),
    [anon_sym_rand] = ACTIONS(2870),
    [anon_sym_rmdir] = ACTIONS(2870),
    [anon_sym_readdir] = ACTIONS(2870),
    [anon_sym_readline] = ACTIONS(2870),
    [anon_sym_readpipe] = ACTIONS(2870),
    [anon_sym_rewinddir] = ACTIONS(2870),
    [anon_sym_readlink] = ACTIONS(2870),
    [anon_sym_ref] = ACTIONS(2870),
    [anon_sym_scalar] = ACTIONS(2870),
    [anon_sym_shift] = ACTIONS(2870),
    [anon_sym_sin] = ACTIONS(2870),
    [anon_sym_sleep] = ACTIONS(2870),
    [anon_sym_sqrt] = ACTIONS(2870),
    [anon_sym_srand] = ACTIONS(2870),
    [anon_sym_stat] = ACTIONS(2870),
    [anon_sym_study] = ACTIONS(2870),
    [anon_sym_tell] = ACTIONS(2870),
    [anon_sym_telldir] = ACTIONS(2870),
    [anon_sym_tied] = ACTIONS(2870),
    [anon_sym_uc] = ACTIONS(2870),
    [anon_sym_ucfirst] = ACTIONS(2870),
    [anon_sym_untie] = ACTIONS(2870),
    [anon_sym_umask] = ACTIONS(2870),
    [anon_sym_values] = ACTIONS(2870),
    [anon_sym_write] = ACTIONS(2870),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2870),
    [anon_sym_qq] = ACTIONS(2870),
    [anon_sym_qw] = ACTIONS(2870),
    [anon_sym_qx] = ACTIONS(2870),
    [anon_sym_qr] = ACTIONS(2870),
    [anon_sym_m] = ACTIONS(2870),
    [anon_sym_s] = ACTIONS(2870),
    [anon_sym_tr] = ACTIONS(2870),
    [anon_sym_y] = ACTIONS(2870),
    [anon_sym_if] = ACTIONS(2870),
    [anon_sym_unless] = ACTIONS(2870),
    [anon_sym_while] = ACTIONS(2870),
    [anon_sym_until] = ACTIONS(2870),
    [aux_sym__bareword_token1] = ACTIONS(2870),
    [sym_number] = ACTIONS(2870),
    [anon_sym_true] = ACTIONS(2870),
    [anon_sym_false] = ACTIONS(2870),
    [sym__single_quote] = ACTIONS(2868),
    [sym__double_quote] = ACTIONS(2868),
    [sym__backtick_quote] = ACTIONS(2868),
    [sym__search_slash_quote] = ACTIONS(2868),
    [sym__open_readline_bracket] = ACTIONS(2868),
    [sym__open_fileglob_bracket] = ACTIONS(2868),
    [sym__PERLY_HEREDOC] = ACTIONS(2868),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2868),
    [sym__filetest] = ACTIONS(2868),
  },
  [STATE(1432)] = {
    [sym_heredoc_content] = STATE(1432),
    [ts_builtin_sym_end] = ACTIONS(2884),
    [sym__identifier] = ACTIONS(2886),
    [anon_sym_LBRACE] = ACTIONS(2886),
    [anon_sym_SEMI] = ACTIONS(2886),
    [anon_sym_package] = ACTIONS(2886),
    [anon_sym_class] = ACTIONS(2886),
    [anon_sym_role] = ACTIONS(2886),
    [anon_sym_BUILD] = ACTIONS(2886),
    [anon_sym_ADJUST] = ACTIONS(2886),
    [anon_sym_DOLLAR] = ACTIONS(2886),
    [anon_sym_AT] = ACTIONS(2886),
    [anon_sym_my] = ACTIONS(2886),
    [anon_sym_extended] = ACTIONS(2886),
    [anon_sym_async] = ACTIONS(2886),
    [anon_sym_sub] = ACTIONS(2886),
    [anon_sym_method] = ACTIONS(2886),
    [anon_sym_LPAREN] = ACTIONS(2886),
    [anon_sym_continue] = ACTIONS(2886),
    [anon_sym_state] = ACTIONS(2886),
    [anon_sym_our] = ACTIONS(2886),
    [anon_sym_try] = ACTIONS(2886),
    [anon_sym_finally] = ACTIONS(2888),
    [anon_sym_defer] = ACTIONS(2886),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2886),
    [anon_sym_LBRACK] = ACTIONS(2886),
    [anon_sym_LT_LT] = ACTIONS(2886),
    [anon_sym_SLASH_SLASH] = ACTIONS(2886),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_PLUS_PLUS] = ACTIONS(2886),
    [anon_sym_DASH_DASH] = ACTIONS(2886),
    [anon_sym_BSLASH] = ACTIONS(2884),
    [anon_sym_do] = ACTIONS(2886),
    [anon_sym_eval] = ACTIONS(2886),
    [anon_sym_await] = ACTIONS(2886),
    [anon_sym_field] = ACTIONS(2886),
    [anon_sym_local] = ACTIONS(2886),
    [anon_sym_dynamically] = ACTIONS(2886),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2886),
    [anon_sym_require] = ACTIONS(2886),
    [anon_sym_map] = ACTIONS(2886),
    [anon_sym_grep] = ACTIONS(2886),
    [anon_sym_sort] = ACTIONS(2886),
    [anon_sym_goto] = ACTIONS(2886),
    [anon_sym_return] = ACTIONS(2886),
    [anon_sym_undef] = ACTIONS(2886),
    [anon_sym_PERCENT2] = ACTIONS(2886),
    [anon_sym_AMP2] = ACTIONS(2886),
    [anon_sym_STAR2] = ACTIONS(2886),
    [anon_sym_use] = ACTIONS(2886),
    [anon_sym_no] = ACTIONS(2886),
    [anon_sym_for] = ACTIONS(2886),
    [anon_sym_foreach] = ACTIONS(2886),
    [anon_sym_last] = ACTIONS(2886),
    [anon_sym_next] = ACTIONS(2886),
    [anon_sym_redo] = ACTIONS(2886),
    [anon_sym_BEGIN] = ACTIONS(2886),
    [anon_sym_INIT] = ACTIONS(2886),
    [anon_sym_CHECK] = ACTIONS(2886),
    [anon_sym_UNITCHECK] = ACTIONS(2886),
    [anon_sym_END] = ACTIONS(2886),
    [anon_sym___FILE__] = ACTIONS(2886),
    [anon_sym___LINE__] = ACTIONS(2886),
    [anon_sym___PACKAGE__] = ACTIONS(2886),
    [anon_sym___SUB__] = ACTIONS(2886),
    [anon_sym_break] = ACTIONS(2886),
    [anon_sym_fork] = ACTIONS(2886),
    [anon_sym_getppid] = ACTIONS(2886),
    [anon_sym_time] = ACTIONS(2886),
    [anon_sym_times] = ACTIONS(2886),
    [anon_sym_wait] = ACTIONS(2886),
    [anon_sym_wantarray] = ACTIONS(2886),
    [anon_sym_abs] = ACTIONS(2886),
    [anon_sym_alarm] = ACTIONS(2886),
    [anon_sym_chop] = ACTIONS(2886),
    [anon_sym_chdir] = ACTIONS(2886),
    [anon_sym_close] = ACTIONS(2886),
    [anon_sym_closedir] = ACTIONS(2886),
    [anon_sym_caller] = ACTIONS(2886),
    [anon_sym_chomp] = ACTIONS(2886),
    [anon_sym_chr] = ACTIONS(2886),
    [anon_sym_cos] = ACTIONS(2886),
    [anon_sym_chroot] = ACTIONS(2886),
    [anon_sym_defined] = ACTIONS(2886),
    [anon_sym_delete] = ACTIONS(2886),
    [anon_sym_dbmclose] = ACTIONS(2886),
    [anon_sym_exists] = ACTIONS(2886),
    [anon_sym_exit] = ACTIONS(2886),
    [anon_sym_eof] = ACTIONS(2886),
    [anon_sym_exp] = ACTIONS(2886),
    [anon_sym_each] = ACTIONS(2886),
    [anon_sym_fc] = ACTIONS(2886),
    [anon_sym_fileno] = ACTIONS(2886),
    [anon_sym_gmtime] = ACTIONS(2886),
    [anon_sym_getc] = ACTIONS(2886),
    [anon_sym_getpgrp] = ACTIONS(2886),
    [anon_sym_getprotobyname] = ACTIONS(2886),
    [anon_sym_getpwname] = ACTIONS(2886),
    [anon_sym_getpwuid] = ACTIONS(2886),
    [anon_sym_getpeername] = ACTIONS(2886),
    [anon_sym_getnetbyname] = ACTIONS(2886),
    [anon_sym_getsockname] = ACTIONS(2886),
    [anon_sym_getgrnam] = ACTIONS(2886),
    [anon_sym_getgrgid] = ACTIONS(2886),
    [anon_sym_hex] = ACTIONS(2886),
    [anon_sym_int] = ACTIONS(2886),
    [anon_sym_keys] = ACTIONS(2886),
    [anon_sym_lc] = ACTIONS(2886),
    [anon_sym_lcfirst] = ACTIONS(2886),
    [anon_sym_length] = ACTIONS(2886),
    [anon_sym_localtime] = ACTIONS(2886),
    [anon_sym_log] = ACTIONS(2886),
    [anon_sym_lock] = ACTIONS(2886),
    [anon_sym_lstat] = ACTIONS(2886),
    [anon_sym_oct] = ACTIONS(2886),
    [anon_sym_ord] = ACTIONS(2886),
    [anon_sym_prototype] = ACTIONS(2886),
    [anon_sym_pop] = ACTIONS(2886),
    [anon_sym_pos] = ACTIONS(2886),
    [anon_sym_quotemeta] = ACTIONS(2886),
    [anon_sym_reset] = ACTIONS(2886),
    [anon_sym_rand] = ACTIONS(2886),
    [anon_sym_rmdir] = ACTIONS(2886),
    [anon_sym_readdir] = ACTIONS(2886),
    [anon_sym_readline] = ACTIONS(2886),
    [anon_sym_readpipe] = ACTIONS(2886),
    [anon_sym_rewinddir] = ACTIONS(2886),
    [anon_sym_readlink] = ACTIONS(2886),
    [anon_sym_ref] = ACTIONS(2886),
    [anon_sym_scalar] = ACTIONS(2886),
    [anon_sym_shift] = ACTIONS(2886),
    [anon_sym_sin] = ACTIONS(2886),
    [anon_sym_sleep] = ACTIONS(2886),
    [anon_sym_sqrt] = ACTIONS(2886),
    [anon_sym_srand] = ACTIONS(2886),
    [anon_sym_stat] = ACTIONS(2886),
    [anon_sym_study] = ACTIONS(2886),
    [anon_sym_tell] = ACTIONS(2886),
    [anon_sym_telldir] = ACTIONS(2886),
    [anon_sym_tied] = ACTIONS(2886),
    [anon_sym_uc] = ACTIONS(2886),
    [anon_sym_ucfirst] = ACTIONS(2886),
    [anon_sym_untie] = ACTIONS(2886),
    [anon_sym_umask] = ACTIONS(2886),
    [anon_sym_values] = ACTIONS(2886),
    [anon_sym_write] = ACTIONS(2886),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2886),
    [anon_sym___END__] = ACTIONS(2886),
    [anon_sym_EOT] = ACTIONS(2886),
    [anon_sym_q] = ACTIONS(2886),
    [anon_sym_qq] = ACTIONS(2886),
    [anon_sym_qw] = ACTIONS(2886),
    [anon_sym_qx] = ACTIONS(2886),
    [anon_sym_qr] = ACTIONS(2886),
    [anon_sym_m] = ACTIONS(2886),
    [anon_sym_s] = ACTIONS(2886),
    [anon_sym_tr] = ACTIONS(2886),
    [anon_sym_y] = ACTIONS(2886),
    [anon_sym_if] = ACTIONS(2886),
    [anon_sym_unless] = ACTIONS(2886),
    [anon_sym_while] = ACTIONS(2886),
    [anon_sym_until] = ACTIONS(2886),
    [aux_sym__bareword_token1] = ACTIONS(2886),
    [sym_number] = ACTIONS(2886),
    [anon_sym_true] = ACTIONS(2886),
    [anon_sym_false] = ACTIONS(2886),
    [sym__single_quote] = ACTIONS(2884),
    [sym__double_quote] = ACTIONS(2884),
    [sym__backtick_quote] = ACTIONS(2884),
    [sym__search_slash_quote] = ACTIONS(2884),
    [sym__open_readline_bracket] = ACTIONS(2884),
    [sym__open_fileglob_bracket] = ACTIONS(2884),
    [sym__PERLY_HEREDOC] = ACTIONS(2884),
    [sym__ctrl_z_hack] = ACTIONS(2884),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2884),
    [sym__filetest] = ACTIONS(2884),
  },
  [STATE(1433)] = {
    [sym_heredoc_content] = STATE(1433),
    [ts_builtin_sym_end] = ACTIONS(2690),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1434)] = {
    [sym_heredoc_content] = STATE(1434),
    [ts_builtin_sym_end] = ACTIONS(2890),
    [sym__identifier] = ACTIONS(2892),
    [anon_sym_LBRACE] = ACTIONS(2892),
    [anon_sym_SEMI] = ACTIONS(2892),
    [anon_sym_package] = ACTIONS(2892),
    [anon_sym_class] = ACTIONS(2892),
    [anon_sym_role] = ACTIONS(2892),
    [anon_sym_BUILD] = ACTIONS(2892),
    [anon_sym_ADJUST] = ACTIONS(2892),
    [anon_sym_DOLLAR] = ACTIONS(2892),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_my] = ACTIONS(2892),
    [anon_sym_extended] = ACTIONS(2892),
    [anon_sym_async] = ACTIONS(2892),
    [anon_sym_sub] = ACTIONS(2892),
    [anon_sym_method] = ACTIONS(2892),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_continue] = ACTIONS(2892),
    [anon_sym_state] = ACTIONS(2892),
    [anon_sym_our] = ACTIONS(2892),
    [anon_sym_try] = ACTIONS(2892),
    [anon_sym_finally] = ACTIONS(2894),
    [anon_sym_defer] = ACTIONS(2892),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2892),
    [anon_sym_LBRACK] = ACTIONS(2892),
    [anon_sym_LT_LT] = ACTIONS(2892),
    [anon_sym_SLASH_SLASH] = ACTIONS(2892),
    [anon_sym_PLUS] = ACTIONS(2892),
    [anon_sym_DASH] = ACTIONS(2892),
    [anon_sym_TILDE] = ACTIONS(2892),
    [anon_sym_BANG] = ACTIONS(2892),
    [anon_sym_PLUS_PLUS] = ACTIONS(2892),
    [anon_sym_DASH_DASH] = ACTIONS(2892),
    [anon_sym_BSLASH] = ACTIONS(2890),
    [anon_sym_do] = ACTIONS(2892),
    [anon_sym_eval] = ACTIONS(2892),
    [anon_sym_await] = ACTIONS(2892),
    [anon_sym_field] = ACTIONS(2892),
    [anon_sym_local] = ACTIONS(2892),
    [anon_sym_dynamically] = ACTIONS(2892),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2892),
    [anon_sym_require] = ACTIONS(2892),
    [anon_sym_map] = ACTIONS(2892),
    [anon_sym_grep] = ACTIONS(2892),
    [anon_sym_sort] = ACTIONS(2892),
    [anon_sym_goto] = ACTIONS(2892),
    [anon_sym_return] = ACTIONS(2892),
    [anon_sym_undef] = ACTIONS(2892),
    [anon_sym_PERCENT2] = ACTIONS(2892),
    [anon_sym_AMP2] = ACTIONS(2892),
    [anon_sym_STAR2] = ACTIONS(2892),
    [anon_sym_use] = ACTIONS(2892),
    [anon_sym_no] = ACTIONS(2892),
    [anon_sym_for] = ACTIONS(2892),
    [anon_sym_foreach] = ACTIONS(2892),
    [anon_sym_last] = ACTIONS(2892),
    [anon_sym_next] = ACTIONS(2892),
    [anon_sym_redo] = ACTIONS(2892),
    [anon_sym_BEGIN] = ACTIONS(2892),
    [anon_sym_INIT] = ACTIONS(2892),
    [anon_sym_CHECK] = ACTIONS(2892),
    [anon_sym_UNITCHECK] = ACTIONS(2892),
    [anon_sym_END] = ACTIONS(2892),
    [anon_sym___FILE__] = ACTIONS(2892),
    [anon_sym___LINE__] = ACTIONS(2892),
    [anon_sym___PACKAGE__] = ACTIONS(2892),
    [anon_sym___SUB__] = ACTIONS(2892),
    [anon_sym_break] = ACTIONS(2892),
    [anon_sym_fork] = ACTIONS(2892),
    [anon_sym_getppid] = ACTIONS(2892),
    [anon_sym_time] = ACTIONS(2892),
    [anon_sym_times] = ACTIONS(2892),
    [anon_sym_wait] = ACTIONS(2892),
    [anon_sym_wantarray] = ACTIONS(2892),
    [anon_sym_abs] = ACTIONS(2892),
    [anon_sym_alarm] = ACTIONS(2892),
    [anon_sym_chop] = ACTIONS(2892),
    [anon_sym_chdir] = ACTIONS(2892),
    [anon_sym_close] = ACTIONS(2892),
    [anon_sym_closedir] = ACTIONS(2892),
    [anon_sym_caller] = ACTIONS(2892),
    [anon_sym_chomp] = ACTIONS(2892),
    [anon_sym_chr] = ACTIONS(2892),
    [anon_sym_cos] = ACTIONS(2892),
    [anon_sym_chroot] = ACTIONS(2892),
    [anon_sym_defined] = ACTIONS(2892),
    [anon_sym_delete] = ACTIONS(2892),
    [anon_sym_dbmclose] = ACTIONS(2892),
    [anon_sym_exists] = ACTIONS(2892),
    [anon_sym_exit] = ACTIONS(2892),
    [anon_sym_eof] = ACTIONS(2892),
    [anon_sym_exp] = ACTIONS(2892),
    [anon_sym_each] = ACTIONS(2892),
    [anon_sym_fc] = ACTIONS(2892),
    [anon_sym_fileno] = ACTIONS(2892),
    [anon_sym_gmtime] = ACTIONS(2892),
    [anon_sym_getc] = ACTIONS(2892),
    [anon_sym_getpgrp] = ACTIONS(2892),
    [anon_sym_getprotobyname] = ACTIONS(2892),
    [anon_sym_getpwname] = ACTIONS(2892),
    [anon_sym_getpwuid] = ACTIONS(2892),
    [anon_sym_getpeername] = ACTIONS(2892),
    [anon_sym_getnetbyname] = ACTIONS(2892),
    [anon_sym_getsockname] = ACTIONS(2892),
    [anon_sym_getgrnam] = ACTIONS(2892),
    [anon_sym_getgrgid] = ACTIONS(2892),
    [anon_sym_hex] = ACTIONS(2892),
    [anon_sym_int] = ACTIONS(2892),
    [anon_sym_keys] = ACTIONS(2892),
    [anon_sym_lc] = ACTIONS(2892),
    [anon_sym_lcfirst] = ACTIONS(2892),
    [anon_sym_length] = ACTIONS(2892),
    [anon_sym_localtime] = ACTIONS(2892),
    [anon_sym_log] = ACTIONS(2892),
    [anon_sym_lock] = ACTIONS(2892),
    [anon_sym_lstat] = ACTIONS(2892),
    [anon_sym_oct] = ACTIONS(2892),
    [anon_sym_ord] = ACTIONS(2892),
    [anon_sym_prototype] = ACTIONS(2892),
    [anon_sym_pop] = ACTIONS(2892),
    [anon_sym_pos] = ACTIONS(2892),
    [anon_sym_quotemeta] = ACTIONS(2892),
    [anon_sym_reset] = ACTIONS(2892),
    [anon_sym_rand] = ACTIONS(2892),
    [anon_sym_rmdir] = ACTIONS(2892),
    [anon_sym_readdir] = ACTIONS(2892),
    [anon_sym_readline] = ACTIONS(2892),
    [anon_sym_readpipe] = ACTIONS(2892),
    [anon_sym_rewinddir] = ACTIONS(2892),
    [anon_sym_readlink] = ACTIONS(2892),
    [anon_sym_ref] = ACTIONS(2892),
    [anon_sym_scalar] = ACTIONS(2892),
    [anon_sym_shift] = ACTIONS(2892),
    [anon_sym_sin] = ACTIONS(2892),
    [anon_sym_sleep] = ACTIONS(2892),
    [anon_sym_sqrt] = ACTIONS(2892),
    [anon_sym_srand] = ACTIONS(2892),
    [anon_sym_stat] = ACTIONS(2892),
    [anon_sym_study] = ACTIONS(2892),
    [anon_sym_tell] = ACTIONS(2892),
    [anon_sym_telldir] = ACTIONS(2892),
    [anon_sym_tied] = ACTIONS(2892),
    [anon_sym_uc] = ACTIONS(2892),
    [anon_sym_ucfirst] = ACTIONS(2892),
    [anon_sym_untie] = ACTIONS(2892),
    [anon_sym_umask] = ACTIONS(2892),
    [anon_sym_values] = ACTIONS(2892),
    [anon_sym_write] = ACTIONS(2892),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2892),
    [anon_sym___END__] = ACTIONS(2892),
    [anon_sym_EOT] = ACTIONS(2892),
    [anon_sym_q] = ACTIONS(2892),
    [anon_sym_qq] = ACTIONS(2892),
    [anon_sym_qw] = ACTIONS(2892),
    [anon_sym_qx] = ACTIONS(2892),
    [anon_sym_qr] = ACTIONS(2892),
    [anon_sym_m] = ACTIONS(2892),
    [anon_sym_s] = ACTIONS(2892),
    [anon_sym_tr] = ACTIONS(2892),
    [anon_sym_y] = ACTIONS(2892),
    [anon_sym_if] = ACTIONS(2892),
    [anon_sym_unless] = ACTIONS(2892),
    [anon_sym_while] = ACTIONS(2892),
    [anon_sym_until] = ACTIONS(2892),
    [aux_sym__bareword_token1] = ACTIONS(2892),
    [sym_number] = ACTIONS(2892),
    [anon_sym_true] = ACTIONS(2892),
    [anon_sym_false] = ACTIONS(2892),
    [sym__single_quote] = ACTIONS(2890),
    [sym__double_quote] = ACTIONS(2890),
    [sym__backtick_quote] = ACTIONS(2890),
    [sym__search_slash_quote] = ACTIONS(2890),
    [sym__open_readline_bracket] = ACTIONS(2890),
    [sym__open_fileglob_bracket] = ACTIONS(2890),
    [sym__PERLY_HEREDOC] = ACTIONS(2890),
    [sym__ctrl_z_hack] = ACTIONS(2890),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2890),
    [sym__filetest] = ACTIONS(2890),
  },
  [STATE(1435)] = {
    [sym__else] = STATE(1593),
    [sym_else] = STATE(1594),
    [sym_elsif] = STATE(1594),
    [sym_heredoc_content] = STATE(1435),
    [sym__identifier] = ACTIONS(2862),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_RBRACE] = ACTIONS(2862),
    [anon_sym_SEMI] = ACTIONS(2862),
    [anon_sym_package] = ACTIONS(2862),
    [anon_sym_class] = ACTIONS(2862),
    [anon_sym_role] = ACTIONS(2862),
    [anon_sym_BUILD] = ACTIONS(2862),
    [anon_sym_ADJUST] = ACTIONS(2862),
    [anon_sym_DOLLAR] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2862),
    [anon_sym_my] = ACTIONS(2862),
    [anon_sym_extended] = ACTIONS(2862),
    [anon_sym_async] = ACTIONS(2862),
    [anon_sym_sub] = ACTIONS(2862),
    [anon_sym_method] = ACTIONS(2862),
    [anon_sym_LPAREN] = ACTIONS(2862),
    [anon_sym_continue] = ACTIONS(2862),
    [anon_sym_state] = ACTIONS(2862),
    [anon_sym_our] = ACTIONS(2862),
    [anon_sym_try] = ACTIONS(2862),
    [anon_sym_defer] = ACTIONS(2862),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2862),
    [anon_sym_else] = ACTIONS(2880),
    [anon_sym_elsif] = ACTIONS(2882),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym_LT_LT] = ACTIONS(2862),
    [anon_sym_SLASH_SLASH] = ACTIONS(2862),
    [anon_sym_PLUS] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_BSLASH] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2862),
    [anon_sym_eval] = ACTIONS(2862),
    [anon_sym_await] = ACTIONS(2862),
    [anon_sym_field] = ACTIONS(2862),
    [anon_sym_local] = ACTIONS(2862),
    [anon_sym_dynamically] = ACTIONS(2862),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2862),
    [anon_sym_require] = ACTIONS(2862),
    [anon_sym_map] = ACTIONS(2862),
    [anon_sym_grep] = ACTIONS(2862),
    [anon_sym_sort] = ACTIONS(2862),
    [anon_sym_goto] = ACTIONS(2862),
    [anon_sym_return] = ACTIONS(2862),
    [anon_sym_undef] = ACTIONS(2862),
    [anon_sym_PERCENT2] = ACTIONS(2862),
    [anon_sym_AMP2] = ACTIONS(2862),
    [anon_sym_STAR2] = ACTIONS(2862),
    [anon_sym_use] = ACTIONS(2862),
    [anon_sym_no] = ACTIONS(2862),
    [anon_sym_for] = ACTIONS(2862),
    [anon_sym_foreach] = ACTIONS(2862),
    [anon_sym_last] = ACTIONS(2862),
    [anon_sym_next] = ACTIONS(2862),
    [anon_sym_redo] = ACTIONS(2862),
    [anon_sym_BEGIN] = ACTIONS(2862),
    [anon_sym_INIT] = ACTIONS(2862),
    [anon_sym_CHECK] = ACTIONS(2862),
    [anon_sym_UNITCHECK] = ACTIONS(2862),
    [anon_sym_END] = ACTIONS(2862),
    [anon_sym___FILE__] = ACTIONS(2862),
    [anon_sym___LINE__] = ACTIONS(2862),
    [anon_sym___PACKAGE__] = ACTIONS(2862),
    [anon_sym___SUB__] = ACTIONS(2862),
    [anon_sym_break] = ACTIONS(2862),
    [anon_sym_fork] = ACTIONS(2862),
    [anon_sym_getppid] = ACTIONS(2862),
    [anon_sym_time] = ACTIONS(2862),
    [anon_sym_times] = ACTIONS(2862),
    [anon_sym_wait] = ACTIONS(2862),
    [anon_sym_wantarray] = ACTIONS(2862),
    [anon_sym_abs] = ACTIONS(2862),
    [anon_sym_alarm] = ACTIONS(2862),
    [anon_sym_chop] = ACTIONS(2862),
    [anon_sym_chdir] = ACTIONS(2862),
    [anon_sym_close] = ACTIONS(2862),
    [anon_sym_closedir] = ACTIONS(2862),
    [anon_sym_caller] = ACTIONS(2862),
    [anon_sym_chomp] = ACTIONS(2862),
    [anon_sym_chr] = ACTIONS(2862),
    [anon_sym_cos] = ACTIONS(2862),
    [anon_sym_chroot] = ACTIONS(2862),
    [anon_sym_defined] = ACTIONS(2862),
    [anon_sym_delete] = ACTIONS(2862),
    [anon_sym_dbmclose] = ACTIONS(2862),
    [anon_sym_exists] = ACTIONS(2862),
    [anon_sym_exit] = ACTIONS(2862),
    [anon_sym_eof] = ACTIONS(2862),
    [anon_sym_exp] = ACTIONS(2862),
    [anon_sym_each] = ACTIONS(2862),
    [anon_sym_fc] = ACTIONS(2862),
    [anon_sym_fileno] = ACTIONS(2862),
    [anon_sym_gmtime] = ACTIONS(2862),
    [anon_sym_getc] = ACTIONS(2862),
    [anon_sym_getpgrp] = ACTIONS(2862),
    [anon_sym_getprotobyname] = ACTIONS(2862),
    [anon_sym_getpwname] = ACTIONS(2862),
    [anon_sym_getpwuid] = ACTIONS(2862),
    [anon_sym_getpeername] = ACTIONS(2862),
    [anon_sym_getnetbyname] = ACTIONS(2862),
    [anon_sym_getsockname] = ACTIONS(2862),
    [anon_sym_getgrnam] = ACTIONS(2862),
    [anon_sym_getgrgid] = ACTIONS(2862),
    [anon_sym_hex] = ACTIONS(2862),
    [anon_sym_int] = ACTIONS(2862),
    [anon_sym_keys] = ACTIONS(2862),
    [anon_sym_lc] = ACTIONS(2862),
    [anon_sym_lcfirst] = ACTIONS(2862),
    [anon_sym_length] = ACTIONS(2862),
    [anon_sym_localtime] = ACTIONS(2862),
    [anon_sym_log] = ACTIONS(2862),
    [anon_sym_lock] = ACTIONS(2862),
    [anon_sym_lstat] = ACTIONS(2862),
    [anon_sym_oct] = ACTIONS(2862),
    [anon_sym_ord] = ACTIONS(2862),
    [anon_sym_prototype] = ACTIONS(2862),
    [anon_sym_pop] = ACTIONS(2862),
    [anon_sym_pos] = ACTIONS(2862),
    [anon_sym_quotemeta] = ACTIONS(2862),
    [anon_sym_reset] = ACTIONS(2862),
    [anon_sym_rand] = ACTIONS(2862),
    [anon_sym_rmdir] = ACTIONS(2862),
    [anon_sym_readdir] = ACTIONS(2862),
    [anon_sym_readline] = ACTIONS(2862),
    [anon_sym_readpipe] = ACTIONS(2862),
    [anon_sym_rewinddir] = ACTIONS(2862),
    [anon_sym_readlink] = ACTIONS(2862),
    [anon_sym_ref] = ACTIONS(2862),
    [anon_sym_scalar] = ACTIONS(2862),
    [anon_sym_shift] = ACTIONS(2862),
    [anon_sym_sin] = ACTIONS(2862),
    [anon_sym_sleep] = ACTIONS(2862),
    [anon_sym_sqrt] = ACTIONS(2862),
    [anon_sym_srand] = ACTIONS(2862),
    [anon_sym_stat] = ACTIONS(2862),
    [anon_sym_study] = ACTIONS(2862),
    [anon_sym_tell] = ACTIONS(2862),
    [anon_sym_telldir] = ACTIONS(2862),
    [anon_sym_tied] = ACTIONS(2862),
    [anon_sym_uc] = ACTIONS(2862),
    [anon_sym_ucfirst] = ACTIONS(2862),
    [anon_sym_untie] = ACTIONS(2862),
    [anon_sym_umask] = ACTIONS(2862),
    [anon_sym_values] = ACTIONS(2862),
    [anon_sym_write] = ACTIONS(2862),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2862),
    [anon_sym_qq] = ACTIONS(2862),
    [anon_sym_qw] = ACTIONS(2862),
    [anon_sym_qx] = ACTIONS(2862),
    [anon_sym_qr] = ACTIONS(2862),
    [anon_sym_m] = ACTIONS(2862),
    [anon_sym_s] = ACTIONS(2862),
    [anon_sym_tr] = ACTIONS(2862),
    [anon_sym_y] = ACTIONS(2862),
    [anon_sym_if] = ACTIONS(2862),
    [anon_sym_unless] = ACTIONS(2862),
    [anon_sym_while] = ACTIONS(2862),
    [anon_sym_until] = ACTIONS(2862),
    [aux_sym__bareword_token1] = ACTIONS(2862),
    [sym_number] = ACTIONS(2862),
    [anon_sym_true] = ACTIONS(2862),
    [anon_sym_false] = ACTIONS(2862),
    [sym__single_quote] = ACTIONS(2860),
    [sym__double_quote] = ACTIONS(2860),
    [sym__backtick_quote] = ACTIONS(2860),
    [sym__search_slash_quote] = ACTIONS(2860),
    [sym__open_readline_bracket] = ACTIONS(2860),
    [sym__open_fileglob_bracket] = ACTIONS(2860),
    [sym__PERLY_HEREDOC] = ACTIONS(2860),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2860),
    [sym__filetest] = ACTIONS(2860),
  },
  [STATE(1436)] = {
    [sym_heredoc_content] = STATE(1436),
    [ts_builtin_sym_end] = ACTIONS(2896),
    [sym__identifier] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2898),
    [anon_sym_SEMI] = ACTIONS(2898),
    [anon_sym_package] = ACTIONS(2898),
    [anon_sym_class] = ACTIONS(2898),
    [anon_sym_role] = ACTIONS(2898),
    [anon_sym_BUILD] = ACTIONS(2898),
    [anon_sym_ADJUST] = ACTIONS(2898),
    [anon_sym_DOLLAR] = ACTIONS(2898),
    [anon_sym_AT] = ACTIONS(2898),
    [anon_sym_my] = ACTIONS(2898),
    [anon_sym_extended] = ACTIONS(2898),
    [anon_sym_async] = ACTIONS(2898),
    [anon_sym_sub] = ACTIONS(2898),
    [anon_sym_method] = ACTIONS(2898),
    [anon_sym_LPAREN] = ACTIONS(2898),
    [anon_sym_continue] = ACTIONS(2898),
    [anon_sym_state] = ACTIONS(2898),
    [anon_sym_our] = ACTIONS(2898),
    [anon_sym_try] = ACTIONS(2898),
    [anon_sym_defer] = ACTIONS(2898),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2898),
    [anon_sym_LBRACK] = ACTIONS(2898),
    [anon_sym_LT_LT] = ACTIONS(2898),
    [anon_sym_SLASH_SLASH] = ACTIONS(2898),
    [anon_sym_PLUS] = ACTIONS(2898),
    [anon_sym_DASH] = ACTIONS(2898),
    [anon_sym_TILDE] = ACTIONS(2898),
    [anon_sym_BANG] = ACTIONS(2898),
    [anon_sym_PLUS_PLUS] = ACTIONS(2898),
    [anon_sym_DASH_DASH] = ACTIONS(2898),
    [anon_sym_BSLASH] = ACTIONS(2896),
    [anon_sym_do] = ACTIONS(2898),
    [anon_sym_eval] = ACTIONS(2898),
    [anon_sym_await] = ACTIONS(2898),
    [anon_sym_field] = ACTIONS(2898),
    [anon_sym_local] = ACTIONS(2898),
    [anon_sym_dynamically] = ACTIONS(2898),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2898),
    [anon_sym_require] = ACTIONS(2898),
    [anon_sym_map] = ACTIONS(2898),
    [anon_sym_grep] = ACTIONS(2898),
    [anon_sym_sort] = ACTIONS(2898),
    [anon_sym_goto] = ACTIONS(2898),
    [anon_sym_return] = ACTIONS(2898),
    [anon_sym_undef] = ACTIONS(2898),
    [anon_sym_PERCENT2] = ACTIONS(2898),
    [anon_sym_AMP2] = ACTIONS(2898),
    [anon_sym_STAR2] = ACTIONS(2898),
    [anon_sym_use] = ACTIONS(2898),
    [anon_sym_no] = ACTIONS(2898),
    [anon_sym_for] = ACTIONS(2898),
    [anon_sym_foreach] = ACTIONS(2898),
    [anon_sym_last] = ACTIONS(2898),
    [anon_sym_next] = ACTIONS(2898),
    [anon_sym_redo] = ACTIONS(2898),
    [anon_sym_BEGIN] = ACTIONS(2898),
    [anon_sym_INIT] = ACTIONS(2898),
    [anon_sym_CHECK] = ACTIONS(2898),
    [anon_sym_UNITCHECK] = ACTIONS(2898),
    [anon_sym_END] = ACTIONS(2898),
    [anon_sym___FILE__] = ACTIONS(2898),
    [anon_sym___LINE__] = ACTIONS(2898),
    [anon_sym___PACKAGE__] = ACTIONS(2898),
    [anon_sym___SUB__] = ACTIONS(2898),
    [anon_sym_break] = ACTIONS(2898),
    [anon_sym_fork] = ACTIONS(2898),
    [anon_sym_getppid] = ACTIONS(2898),
    [anon_sym_time] = ACTIONS(2898),
    [anon_sym_times] = ACTIONS(2898),
    [anon_sym_wait] = ACTIONS(2898),
    [anon_sym_wantarray] = ACTIONS(2898),
    [anon_sym_abs] = ACTIONS(2898),
    [anon_sym_alarm] = ACTIONS(2898),
    [anon_sym_chop] = ACTIONS(2898),
    [anon_sym_chdir] = ACTIONS(2898),
    [anon_sym_close] = ACTIONS(2898),
    [anon_sym_closedir] = ACTIONS(2898),
    [anon_sym_caller] = ACTIONS(2898),
    [anon_sym_chomp] = ACTIONS(2898),
    [anon_sym_chr] = ACTIONS(2898),
    [anon_sym_cos] = ACTIONS(2898),
    [anon_sym_chroot] = ACTIONS(2898),
    [anon_sym_defined] = ACTIONS(2898),
    [anon_sym_delete] = ACTIONS(2898),
    [anon_sym_dbmclose] = ACTIONS(2898),
    [anon_sym_exists] = ACTIONS(2898),
    [anon_sym_exit] = ACTIONS(2898),
    [anon_sym_eof] = ACTIONS(2898),
    [anon_sym_exp] = ACTIONS(2898),
    [anon_sym_each] = ACTIONS(2898),
    [anon_sym_fc] = ACTIONS(2898),
    [anon_sym_fileno] = ACTIONS(2898),
    [anon_sym_gmtime] = ACTIONS(2898),
    [anon_sym_getc] = ACTIONS(2898),
    [anon_sym_getpgrp] = ACTIONS(2898),
    [anon_sym_getprotobyname] = ACTIONS(2898),
    [anon_sym_getpwname] = ACTIONS(2898),
    [anon_sym_getpwuid] = ACTIONS(2898),
    [anon_sym_getpeername] = ACTIONS(2898),
    [anon_sym_getnetbyname] = ACTIONS(2898),
    [anon_sym_getsockname] = ACTIONS(2898),
    [anon_sym_getgrnam] = ACTIONS(2898),
    [anon_sym_getgrgid] = ACTIONS(2898),
    [anon_sym_hex] = ACTIONS(2898),
    [anon_sym_int] = ACTIONS(2898),
    [anon_sym_keys] = ACTIONS(2898),
    [anon_sym_lc] = ACTIONS(2898),
    [anon_sym_lcfirst] = ACTIONS(2898),
    [anon_sym_length] = ACTIONS(2898),
    [anon_sym_localtime] = ACTIONS(2898),
    [anon_sym_log] = ACTIONS(2898),
    [anon_sym_lock] = ACTIONS(2898),
    [anon_sym_lstat] = ACTIONS(2898),
    [anon_sym_oct] = ACTIONS(2898),
    [anon_sym_ord] = ACTIONS(2898),
    [anon_sym_prototype] = ACTIONS(2898),
    [anon_sym_pop] = ACTIONS(2898),
    [anon_sym_pos] = ACTIONS(2898),
    [anon_sym_quotemeta] = ACTIONS(2898),
    [anon_sym_reset] = ACTIONS(2898),
    [anon_sym_rand] = ACTIONS(2898),
    [anon_sym_rmdir] = ACTIONS(2898),
    [anon_sym_readdir] = ACTIONS(2898),
    [anon_sym_readline] = ACTIONS(2898),
    [anon_sym_readpipe] = ACTIONS(2898),
    [anon_sym_rewinddir] = ACTIONS(2898),
    [anon_sym_readlink] = ACTIONS(2898),
    [anon_sym_ref] = ACTIONS(2898),
    [anon_sym_scalar] = ACTIONS(2898),
    [anon_sym_shift] = ACTIONS(2898),
    [anon_sym_sin] = ACTIONS(2898),
    [anon_sym_sleep] = ACTIONS(2898),
    [anon_sym_sqrt] = ACTIONS(2898),
    [anon_sym_srand] = ACTIONS(2898),
    [anon_sym_stat] = ACTIONS(2898),
    [anon_sym_study] = ACTIONS(2898),
    [anon_sym_tell] = ACTIONS(2898),
    [anon_sym_telldir] = ACTIONS(2898),
    [anon_sym_tied] = ACTIONS(2898),
    [anon_sym_uc] = ACTIONS(2898),
    [anon_sym_ucfirst] = ACTIONS(2898),
    [anon_sym_untie] = ACTIONS(2898),
    [anon_sym_umask] = ACTIONS(2898),
    [anon_sym_values] = ACTIONS(2898),
    [anon_sym_write] = ACTIONS(2898),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2898),
    [anon_sym___END__] = ACTIONS(2898),
    [anon_sym_EOT] = ACTIONS(2898),
    [anon_sym_q] = ACTIONS(2898),
    [anon_sym_qq] = ACTIONS(2898),
    [anon_sym_qw] = ACTIONS(2898),
    [anon_sym_qx] = ACTIONS(2898),
    [anon_sym_qr] = ACTIONS(2898),
    [anon_sym_m] = ACTIONS(2898),
    [anon_sym_s] = ACTIONS(2898),
    [anon_sym_tr] = ACTIONS(2898),
    [anon_sym_y] = ACTIONS(2898),
    [anon_sym_if] = ACTIONS(2898),
    [anon_sym_unless] = ACTIONS(2898),
    [anon_sym_while] = ACTIONS(2898),
    [anon_sym_until] = ACTIONS(2898),
    [aux_sym__bareword_token1] = ACTIONS(2898),
    [sym_number] = ACTIONS(2898),
    [anon_sym_true] = ACTIONS(2898),
    [anon_sym_false] = ACTIONS(2898),
    [sym__single_quote] = ACTIONS(2896),
    [sym__double_quote] = ACTIONS(2896),
    [sym__backtick_quote] = ACTIONS(2896),
    [sym__search_slash_quote] = ACTIONS(2896),
    [sym__open_readline_bracket] = ACTIONS(2896),
    [sym__open_fileglob_bracket] = ACTIONS(2896),
    [sym__PERLY_HEREDOC] = ACTIONS(2896),
    [sym__ctrl_z_hack] = ACTIONS(2896),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2896),
    [sym__filetest] = ACTIONS(2896),
  },
  [STATE(1437)] = {
    [sym_heredoc_content] = STATE(1437),
    [ts_builtin_sym_end] = ACTIONS(2900),
    [sym__identifier] = ACTIONS(2902),
    [anon_sym_LBRACE] = ACTIONS(2902),
    [anon_sym_SEMI] = ACTIONS(2902),
    [anon_sym_package] = ACTIONS(2902),
    [anon_sym_class] = ACTIONS(2902),
    [anon_sym_role] = ACTIONS(2902),
    [anon_sym_BUILD] = ACTIONS(2902),
    [anon_sym_ADJUST] = ACTIONS(2902),
    [anon_sym_DOLLAR] = ACTIONS(2902),
    [anon_sym_AT] = ACTIONS(2902),
    [anon_sym_my] = ACTIONS(2902),
    [anon_sym_extended] = ACTIONS(2902),
    [anon_sym_async] = ACTIONS(2902),
    [anon_sym_sub] = ACTIONS(2902),
    [anon_sym_method] = ACTIONS(2902),
    [anon_sym_LPAREN] = ACTIONS(2902),
    [anon_sym_continue] = ACTIONS(2902),
    [anon_sym_state] = ACTIONS(2902),
    [anon_sym_our] = ACTIONS(2902),
    [anon_sym_try] = ACTIONS(2902),
    [anon_sym_defer] = ACTIONS(2902),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2902),
    [anon_sym_LBRACK] = ACTIONS(2902),
    [anon_sym_LT_LT] = ACTIONS(2902),
    [anon_sym_SLASH_SLASH] = ACTIONS(2902),
    [anon_sym_PLUS] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_PLUS_PLUS] = ACTIONS(2902),
    [anon_sym_DASH_DASH] = ACTIONS(2902),
    [anon_sym_BSLASH] = ACTIONS(2900),
    [anon_sym_do] = ACTIONS(2902),
    [anon_sym_eval] = ACTIONS(2902),
    [anon_sym_await] = ACTIONS(2902),
    [anon_sym_field] = ACTIONS(2902),
    [anon_sym_local] = ACTIONS(2902),
    [anon_sym_dynamically] = ACTIONS(2902),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2902),
    [anon_sym_require] = ACTIONS(2902),
    [anon_sym_map] = ACTIONS(2902),
    [anon_sym_grep] = ACTIONS(2902),
    [anon_sym_sort] = ACTIONS(2902),
    [anon_sym_goto] = ACTIONS(2902),
    [anon_sym_return] = ACTIONS(2902),
    [anon_sym_undef] = ACTIONS(2902),
    [anon_sym_PERCENT2] = ACTIONS(2902),
    [anon_sym_AMP2] = ACTIONS(2902),
    [anon_sym_STAR2] = ACTIONS(2902),
    [anon_sym_use] = ACTIONS(2902),
    [anon_sym_no] = ACTIONS(2902),
    [anon_sym_for] = ACTIONS(2902),
    [anon_sym_foreach] = ACTIONS(2902),
    [anon_sym_last] = ACTIONS(2902),
    [anon_sym_next] = ACTIONS(2902),
    [anon_sym_redo] = ACTIONS(2902),
    [anon_sym_BEGIN] = ACTIONS(2902),
    [anon_sym_INIT] = ACTIONS(2902),
    [anon_sym_CHECK] = ACTIONS(2902),
    [anon_sym_UNITCHECK] = ACTIONS(2902),
    [anon_sym_END] = ACTIONS(2902),
    [anon_sym___FILE__] = ACTIONS(2902),
    [anon_sym___LINE__] = ACTIONS(2902),
    [anon_sym___PACKAGE__] = ACTIONS(2902),
    [anon_sym___SUB__] = ACTIONS(2902),
    [anon_sym_break] = ACTIONS(2902),
    [anon_sym_fork] = ACTIONS(2902),
    [anon_sym_getppid] = ACTIONS(2902),
    [anon_sym_time] = ACTIONS(2902),
    [anon_sym_times] = ACTIONS(2902),
    [anon_sym_wait] = ACTIONS(2902),
    [anon_sym_wantarray] = ACTIONS(2902),
    [anon_sym_abs] = ACTIONS(2902),
    [anon_sym_alarm] = ACTIONS(2902),
    [anon_sym_chop] = ACTIONS(2902),
    [anon_sym_chdir] = ACTIONS(2902),
    [anon_sym_close] = ACTIONS(2902),
    [anon_sym_closedir] = ACTIONS(2902),
    [anon_sym_caller] = ACTIONS(2902),
    [anon_sym_chomp] = ACTIONS(2902),
    [anon_sym_chr] = ACTIONS(2902),
    [anon_sym_cos] = ACTIONS(2902),
    [anon_sym_chroot] = ACTIONS(2902),
    [anon_sym_defined] = ACTIONS(2902),
    [anon_sym_delete] = ACTIONS(2902),
    [anon_sym_dbmclose] = ACTIONS(2902),
    [anon_sym_exists] = ACTIONS(2902),
    [anon_sym_exit] = ACTIONS(2902),
    [anon_sym_eof] = ACTIONS(2902),
    [anon_sym_exp] = ACTIONS(2902),
    [anon_sym_each] = ACTIONS(2902),
    [anon_sym_fc] = ACTIONS(2902),
    [anon_sym_fileno] = ACTIONS(2902),
    [anon_sym_gmtime] = ACTIONS(2902),
    [anon_sym_getc] = ACTIONS(2902),
    [anon_sym_getpgrp] = ACTIONS(2902),
    [anon_sym_getprotobyname] = ACTIONS(2902),
    [anon_sym_getpwname] = ACTIONS(2902),
    [anon_sym_getpwuid] = ACTIONS(2902),
    [anon_sym_getpeername] = ACTIONS(2902),
    [anon_sym_getnetbyname] = ACTIONS(2902),
    [anon_sym_getsockname] = ACTIONS(2902),
    [anon_sym_getgrnam] = ACTIONS(2902),
    [anon_sym_getgrgid] = ACTIONS(2902),
    [anon_sym_hex] = ACTIONS(2902),
    [anon_sym_int] = ACTIONS(2902),
    [anon_sym_keys] = ACTIONS(2902),
    [anon_sym_lc] = ACTIONS(2902),
    [anon_sym_lcfirst] = ACTIONS(2902),
    [anon_sym_length] = ACTIONS(2902),
    [anon_sym_localtime] = ACTIONS(2902),
    [anon_sym_log] = ACTIONS(2902),
    [anon_sym_lock] = ACTIONS(2902),
    [anon_sym_lstat] = ACTIONS(2902),
    [anon_sym_oct] = ACTIONS(2902),
    [anon_sym_ord] = ACTIONS(2902),
    [anon_sym_prototype] = ACTIONS(2902),
    [anon_sym_pop] = ACTIONS(2902),
    [anon_sym_pos] = ACTIONS(2902),
    [anon_sym_quotemeta] = ACTIONS(2902),
    [anon_sym_reset] = ACTIONS(2902),
    [anon_sym_rand] = ACTIONS(2902),
    [anon_sym_rmdir] = ACTIONS(2902),
    [anon_sym_readdir] = ACTIONS(2902),
    [anon_sym_readline] = ACTIONS(2902),
    [anon_sym_readpipe] = ACTIONS(2902),
    [anon_sym_rewinddir] = ACTIONS(2902),
    [anon_sym_readlink] = ACTIONS(2902),
    [anon_sym_ref] = ACTIONS(2902),
    [anon_sym_scalar] = ACTIONS(2902),
    [anon_sym_shift] = ACTIONS(2902),
    [anon_sym_sin] = ACTIONS(2902),
    [anon_sym_sleep] = ACTIONS(2902),
    [anon_sym_sqrt] = ACTIONS(2902),
    [anon_sym_srand] = ACTIONS(2902),
    [anon_sym_stat] = ACTIONS(2902),
    [anon_sym_study] = ACTIONS(2902),
    [anon_sym_tell] = ACTIONS(2902),
    [anon_sym_telldir] = ACTIONS(2902),
    [anon_sym_tied] = ACTIONS(2902),
    [anon_sym_uc] = ACTIONS(2902),
    [anon_sym_ucfirst] = ACTIONS(2902),
    [anon_sym_untie] = ACTIONS(2902),
    [anon_sym_umask] = ACTIONS(2902),
    [anon_sym_values] = ACTIONS(2902),
    [anon_sym_write] = ACTIONS(2902),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2902),
    [anon_sym___END__] = ACTIONS(2902),
    [anon_sym_EOT] = ACTIONS(2902),
    [anon_sym_q] = ACTIONS(2902),
    [anon_sym_qq] = ACTIONS(2902),
    [anon_sym_qw] = ACTIONS(2902),
    [anon_sym_qx] = ACTIONS(2902),
    [anon_sym_qr] = ACTIONS(2902),
    [anon_sym_m] = ACTIONS(2902),
    [anon_sym_s] = ACTIONS(2902),
    [anon_sym_tr] = ACTIONS(2902),
    [anon_sym_y] = ACTIONS(2902),
    [anon_sym_if] = ACTIONS(2902),
    [anon_sym_unless] = ACTIONS(2902),
    [anon_sym_while] = ACTIONS(2902),
    [anon_sym_until] = ACTIONS(2902),
    [aux_sym__bareword_token1] = ACTIONS(2902),
    [sym_number] = ACTIONS(2902),
    [anon_sym_true] = ACTIONS(2902),
    [anon_sym_false] = ACTIONS(2902),
    [sym__single_quote] = ACTIONS(2900),
    [sym__double_quote] = ACTIONS(2900),
    [sym__backtick_quote] = ACTIONS(2900),
    [sym__search_slash_quote] = ACTIONS(2900),
    [sym__open_readline_bracket] = ACTIONS(2900),
    [sym__open_fileglob_bracket] = ACTIONS(2900),
    [sym__PERLY_HEREDOC] = ACTIONS(2900),
    [sym__ctrl_z_hack] = ACTIONS(2900),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2900),
    [sym__filetest] = ACTIONS(2900),
  },
  [STATE(1438)] = {
    [sym_heredoc_content] = STATE(1438),
    [ts_builtin_sym_end] = ACTIONS(2904),
    [sym__identifier] = ACTIONS(2906),
    [anon_sym_LBRACE] = ACTIONS(2906),
    [anon_sym_SEMI] = ACTIONS(2906),
    [anon_sym_package] = ACTIONS(2906),
    [anon_sym_class] = ACTIONS(2906),
    [anon_sym_role] = ACTIONS(2906),
    [anon_sym_BUILD] = ACTIONS(2906),
    [anon_sym_ADJUST] = ACTIONS(2906),
    [anon_sym_DOLLAR] = ACTIONS(2906),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_my] = ACTIONS(2906),
    [anon_sym_extended] = ACTIONS(2906),
    [anon_sym_async] = ACTIONS(2906),
    [anon_sym_sub] = ACTIONS(2906),
    [anon_sym_method] = ACTIONS(2906),
    [anon_sym_LPAREN] = ACTIONS(2906),
    [anon_sym_continue] = ACTIONS(2906),
    [anon_sym_state] = ACTIONS(2906),
    [anon_sym_our] = ACTIONS(2906),
    [anon_sym_try] = ACTIONS(2906),
    [anon_sym_defer] = ACTIONS(2906),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2906),
    [anon_sym_LBRACK] = ACTIONS(2906),
    [anon_sym_LT_LT] = ACTIONS(2906),
    [anon_sym_SLASH_SLASH] = ACTIONS(2906),
    [anon_sym_PLUS] = ACTIONS(2906),
    [anon_sym_DASH] = ACTIONS(2906),
    [anon_sym_TILDE] = ACTIONS(2906),
    [anon_sym_BANG] = ACTIONS(2906),
    [anon_sym_PLUS_PLUS] = ACTIONS(2906),
    [anon_sym_DASH_DASH] = ACTIONS(2906),
    [anon_sym_BSLASH] = ACTIONS(2904),
    [anon_sym_do] = ACTIONS(2906),
    [anon_sym_eval] = ACTIONS(2906),
    [anon_sym_await] = ACTIONS(2906),
    [anon_sym_field] = ACTIONS(2906),
    [anon_sym_local] = ACTIONS(2906),
    [anon_sym_dynamically] = ACTIONS(2906),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2906),
    [anon_sym_require] = ACTIONS(2906),
    [anon_sym_map] = ACTIONS(2906),
    [anon_sym_grep] = ACTIONS(2906),
    [anon_sym_sort] = ACTIONS(2906),
    [anon_sym_goto] = ACTIONS(2906),
    [anon_sym_return] = ACTIONS(2906),
    [anon_sym_undef] = ACTIONS(2906),
    [anon_sym_PERCENT2] = ACTIONS(2906),
    [anon_sym_AMP2] = ACTIONS(2906),
    [anon_sym_STAR2] = ACTIONS(2906),
    [anon_sym_use] = ACTIONS(2906),
    [anon_sym_no] = ACTIONS(2906),
    [anon_sym_for] = ACTIONS(2906),
    [anon_sym_foreach] = ACTIONS(2906),
    [anon_sym_last] = ACTIONS(2906),
    [anon_sym_next] = ACTIONS(2906),
    [anon_sym_redo] = ACTIONS(2906),
    [anon_sym_BEGIN] = ACTIONS(2906),
    [anon_sym_INIT] = ACTIONS(2906),
    [anon_sym_CHECK] = ACTIONS(2906),
    [anon_sym_UNITCHECK] = ACTIONS(2906),
    [anon_sym_END] = ACTIONS(2906),
    [anon_sym___FILE__] = ACTIONS(2906),
    [anon_sym___LINE__] = ACTIONS(2906),
    [anon_sym___PACKAGE__] = ACTIONS(2906),
    [anon_sym___SUB__] = ACTIONS(2906),
    [anon_sym_break] = ACTIONS(2906),
    [anon_sym_fork] = ACTIONS(2906),
    [anon_sym_getppid] = ACTIONS(2906),
    [anon_sym_time] = ACTIONS(2906),
    [anon_sym_times] = ACTIONS(2906),
    [anon_sym_wait] = ACTIONS(2906),
    [anon_sym_wantarray] = ACTIONS(2906),
    [anon_sym_abs] = ACTIONS(2906),
    [anon_sym_alarm] = ACTIONS(2906),
    [anon_sym_chop] = ACTIONS(2906),
    [anon_sym_chdir] = ACTIONS(2906),
    [anon_sym_close] = ACTIONS(2906),
    [anon_sym_closedir] = ACTIONS(2906),
    [anon_sym_caller] = ACTIONS(2906),
    [anon_sym_chomp] = ACTIONS(2906),
    [anon_sym_chr] = ACTIONS(2906),
    [anon_sym_cos] = ACTIONS(2906),
    [anon_sym_chroot] = ACTIONS(2906),
    [anon_sym_defined] = ACTIONS(2906),
    [anon_sym_delete] = ACTIONS(2906),
    [anon_sym_dbmclose] = ACTIONS(2906),
    [anon_sym_exists] = ACTIONS(2906),
    [anon_sym_exit] = ACTIONS(2906),
    [anon_sym_eof] = ACTIONS(2906),
    [anon_sym_exp] = ACTIONS(2906),
    [anon_sym_each] = ACTIONS(2906),
    [anon_sym_fc] = ACTIONS(2906),
    [anon_sym_fileno] = ACTIONS(2906),
    [anon_sym_gmtime] = ACTIONS(2906),
    [anon_sym_getc] = ACTIONS(2906),
    [anon_sym_getpgrp] = ACTIONS(2906),
    [anon_sym_getprotobyname] = ACTIONS(2906),
    [anon_sym_getpwname] = ACTIONS(2906),
    [anon_sym_getpwuid] = ACTIONS(2906),
    [anon_sym_getpeername] = ACTIONS(2906),
    [anon_sym_getnetbyname] = ACTIONS(2906),
    [anon_sym_getsockname] = ACTIONS(2906),
    [anon_sym_getgrnam] = ACTIONS(2906),
    [anon_sym_getgrgid] = ACTIONS(2906),
    [anon_sym_hex] = ACTIONS(2906),
    [anon_sym_int] = ACTIONS(2906),
    [anon_sym_keys] = ACTIONS(2906),
    [anon_sym_lc] = ACTIONS(2906),
    [anon_sym_lcfirst] = ACTIONS(2906),
    [anon_sym_length] = ACTIONS(2906),
    [anon_sym_localtime] = ACTIONS(2906),
    [anon_sym_log] = ACTIONS(2906),
    [anon_sym_lock] = ACTIONS(2906),
    [anon_sym_lstat] = ACTIONS(2906),
    [anon_sym_oct] = ACTIONS(2906),
    [anon_sym_ord] = ACTIONS(2906),
    [anon_sym_prototype] = ACTIONS(2906),
    [anon_sym_pop] = ACTIONS(2906),
    [anon_sym_pos] = ACTIONS(2906),
    [anon_sym_quotemeta] = ACTIONS(2906),
    [anon_sym_reset] = ACTIONS(2906),
    [anon_sym_rand] = ACTIONS(2906),
    [anon_sym_rmdir] = ACTIONS(2906),
    [anon_sym_readdir] = ACTIONS(2906),
    [anon_sym_readline] = ACTIONS(2906),
    [anon_sym_readpipe] = ACTIONS(2906),
    [anon_sym_rewinddir] = ACTIONS(2906),
    [anon_sym_readlink] = ACTIONS(2906),
    [anon_sym_ref] = ACTIONS(2906),
    [anon_sym_scalar] = ACTIONS(2906),
    [anon_sym_shift] = ACTIONS(2906),
    [anon_sym_sin] = ACTIONS(2906),
    [anon_sym_sleep] = ACTIONS(2906),
    [anon_sym_sqrt] = ACTIONS(2906),
    [anon_sym_srand] = ACTIONS(2906),
    [anon_sym_stat] = ACTIONS(2906),
    [anon_sym_study] = ACTIONS(2906),
    [anon_sym_tell] = ACTIONS(2906),
    [anon_sym_telldir] = ACTIONS(2906),
    [anon_sym_tied] = ACTIONS(2906),
    [anon_sym_uc] = ACTIONS(2906),
    [anon_sym_ucfirst] = ACTIONS(2906),
    [anon_sym_untie] = ACTIONS(2906),
    [anon_sym_umask] = ACTIONS(2906),
    [anon_sym_values] = ACTIONS(2906),
    [anon_sym_write] = ACTIONS(2906),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2906),
    [anon_sym___END__] = ACTIONS(2906),
    [anon_sym_EOT] = ACTIONS(2906),
    [anon_sym_q] = ACTIONS(2906),
    [anon_sym_qq] = ACTIONS(2906),
    [anon_sym_qw] = ACTIONS(2906),
    [anon_sym_qx] = ACTIONS(2906),
    [anon_sym_qr] = ACTIONS(2906),
    [anon_sym_m] = ACTIONS(2906),
    [anon_sym_s] = ACTIONS(2906),
    [anon_sym_tr] = ACTIONS(2906),
    [anon_sym_y] = ACTIONS(2906),
    [anon_sym_if] = ACTIONS(2906),
    [anon_sym_unless] = ACTIONS(2906),
    [anon_sym_while] = ACTIONS(2906),
    [anon_sym_until] = ACTIONS(2906),
    [aux_sym__bareword_token1] = ACTIONS(2906),
    [sym_number] = ACTIONS(2906),
    [anon_sym_true] = ACTIONS(2906),
    [anon_sym_false] = ACTIONS(2906),
    [sym__single_quote] = ACTIONS(2904),
    [sym__double_quote] = ACTIONS(2904),
    [sym__backtick_quote] = ACTIONS(2904),
    [sym__search_slash_quote] = ACTIONS(2904),
    [sym__open_readline_bracket] = ACTIONS(2904),
    [sym__open_fileglob_bracket] = ACTIONS(2904),
    [sym__PERLY_HEREDOC] = ACTIONS(2904),
    [sym__ctrl_z_hack] = ACTIONS(2904),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2904),
    [sym__filetest] = ACTIONS(2904),
  },
  [STATE(1439)] = {
    [sym_heredoc_content] = STATE(1439),
    [ts_builtin_sym_end] = ACTIONS(2908),
    [sym__identifier] = ACTIONS(2910),
    [anon_sym_LBRACE] = ACTIONS(2910),
    [anon_sym_SEMI] = ACTIONS(2910),
    [anon_sym_package] = ACTIONS(2910),
    [anon_sym_class] = ACTIONS(2910),
    [anon_sym_role] = ACTIONS(2910),
    [anon_sym_BUILD] = ACTIONS(2910),
    [anon_sym_ADJUST] = ACTIONS(2910),
    [anon_sym_DOLLAR] = ACTIONS(2910),
    [anon_sym_AT] = ACTIONS(2910),
    [anon_sym_my] = ACTIONS(2910),
    [anon_sym_extended] = ACTIONS(2910),
    [anon_sym_async] = ACTIONS(2910),
    [anon_sym_sub] = ACTIONS(2910),
    [anon_sym_method] = ACTIONS(2910),
    [anon_sym_LPAREN] = ACTIONS(2910),
    [anon_sym_continue] = ACTIONS(2910),
    [anon_sym_state] = ACTIONS(2910),
    [anon_sym_our] = ACTIONS(2910),
    [anon_sym_try] = ACTIONS(2910),
    [anon_sym_defer] = ACTIONS(2910),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2910),
    [anon_sym_LT_LT] = ACTIONS(2910),
    [anon_sym_SLASH_SLASH] = ACTIONS(2910),
    [anon_sym_PLUS] = ACTIONS(2910),
    [anon_sym_DASH] = ACTIONS(2910),
    [anon_sym_TILDE] = ACTIONS(2910),
    [anon_sym_BANG] = ACTIONS(2910),
    [anon_sym_PLUS_PLUS] = ACTIONS(2910),
    [anon_sym_DASH_DASH] = ACTIONS(2910),
    [anon_sym_BSLASH] = ACTIONS(2908),
    [anon_sym_do] = ACTIONS(2910),
    [anon_sym_eval] = ACTIONS(2910),
    [anon_sym_await] = ACTIONS(2910),
    [anon_sym_field] = ACTIONS(2910),
    [anon_sym_local] = ACTIONS(2910),
    [anon_sym_dynamically] = ACTIONS(2910),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2910),
    [anon_sym_require] = ACTIONS(2910),
    [anon_sym_map] = ACTIONS(2910),
    [anon_sym_grep] = ACTIONS(2910),
    [anon_sym_sort] = ACTIONS(2910),
    [anon_sym_goto] = ACTIONS(2910),
    [anon_sym_return] = ACTIONS(2910),
    [anon_sym_undef] = ACTIONS(2910),
    [anon_sym_PERCENT2] = ACTIONS(2910),
    [anon_sym_AMP2] = ACTIONS(2910),
    [anon_sym_STAR2] = ACTIONS(2910),
    [anon_sym_use] = ACTIONS(2910),
    [anon_sym_no] = ACTIONS(2910),
    [anon_sym_for] = ACTIONS(2910),
    [anon_sym_foreach] = ACTIONS(2910),
    [anon_sym_last] = ACTIONS(2910),
    [anon_sym_next] = ACTIONS(2910),
    [anon_sym_redo] = ACTIONS(2910),
    [anon_sym_BEGIN] = ACTIONS(2910),
    [anon_sym_INIT] = ACTIONS(2910),
    [anon_sym_CHECK] = ACTIONS(2910),
    [anon_sym_UNITCHECK] = ACTIONS(2910),
    [anon_sym_END] = ACTIONS(2910),
    [anon_sym___FILE__] = ACTIONS(2910),
    [anon_sym___LINE__] = ACTIONS(2910),
    [anon_sym___PACKAGE__] = ACTIONS(2910),
    [anon_sym___SUB__] = ACTIONS(2910),
    [anon_sym_break] = ACTIONS(2910),
    [anon_sym_fork] = ACTIONS(2910),
    [anon_sym_getppid] = ACTIONS(2910),
    [anon_sym_time] = ACTIONS(2910),
    [anon_sym_times] = ACTIONS(2910),
    [anon_sym_wait] = ACTIONS(2910),
    [anon_sym_wantarray] = ACTIONS(2910),
    [anon_sym_abs] = ACTIONS(2910),
    [anon_sym_alarm] = ACTIONS(2910),
    [anon_sym_chop] = ACTIONS(2910),
    [anon_sym_chdir] = ACTIONS(2910),
    [anon_sym_close] = ACTIONS(2910),
    [anon_sym_closedir] = ACTIONS(2910),
    [anon_sym_caller] = ACTIONS(2910),
    [anon_sym_chomp] = ACTIONS(2910),
    [anon_sym_chr] = ACTIONS(2910),
    [anon_sym_cos] = ACTIONS(2910),
    [anon_sym_chroot] = ACTIONS(2910),
    [anon_sym_defined] = ACTIONS(2910),
    [anon_sym_delete] = ACTIONS(2910),
    [anon_sym_dbmclose] = ACTIONS(2910),
    [anon_sym_exists] = ACTIONS(2910),
    [anon_sym_exit] = ACTIONS(2910),
    [anon_sym_eof] = ACTIONS(2910),
    [anon_sym_exp] = ACTIONS(2910),
    [anon_sym_each] = ACTIONS(2910),
    [anon_sym_fc] = ACTIONS(2910),
    [anon_sym_fileno] = ACTIONS(2910),
    [anon_sym_gmtime] = ACTIONS(2910),
    [anon_sym_getc] = ACTIONS(2910),
    [anon_sym_getpgrp] = ACTIONS(2910),
    [anon_sym_getprotobyname] = ACTIONS(2910),
    [anon_sym_getpwname] = ACTIONS(2910),
    [anon_sym_getpwuid] = ACTIONS(2910),
    [anon_sym_getpeername] = ACTIONS(2910),
    [anon_sym_getnetbyname] = ACTIONS(2910),
    [anon_sym_getsockname] = ACTIONS(2910),
    [anon_sym_getgrnam] = ACTIONS(2910),
    [anon_sym_getgrgid] = ACTIONS(2910),
    [anon_sym_hex] = ACTIONS(2910),
    [anon_sym_int] = ACTIONS(2910),
    [anon_sym_keys] = ACTIONS(2910),
    [anon_sym_lc] = ACTIONS(2910),
    [anon_sym_lcfirst] = ACTIONS(2910),
    [anon_sym_length] = ACTIONS(2910),
    [anon_sym_localtime] = ACTIONS(2910),
    [anon_sym_log] = ACTIONS(2910),
    [anon_sym_lock] = ACTIONS(2910),
    [anon_sym_lstat] = ACTIONS(2910),
    [anon_sym_oct] = ACTIONS(2910),
    [anon_sym_ord] = ACTIONS(2910),
    [anon_sym_prototype] = ACTIONS(2910),
    [anon_sym_pop] = ACTIONS(2910),
    [anon_sym_pos] = ACTIONS(2910),
    [anon_sym_quotemeta] = ACTIONS(2910),
    [anon_sym_reset] = ACTIONS(2910),
    [anon_sym_rand] = ACTIONS(2910),
    [anon_sym_rmdir] = ACTIONS(2910),
    [anon_sym_readdir] = ACTIONS(2910),
    [anon_sym_readline] = ACTIONS(2910),
    [anon_sym_readpipe] = ACTIONS(2910),
    [anon_sym_rewinddir] = ACTIONS(2910),
    [anon_sym_readlink] = ACTIONS(2910),
    [anon_sym_ref] = ACTIONS(2910),
    [anon_sym_scalar] = ACTIONS(2910),
    [anon_sym_shift] = ACTIONS(2910),
    [anon_sym_sin] = ACTIONS(2910),
    [anon_sym_sleep] = ACTIONS(2910),
    [anon_sym_sqrt] = ACTIONS(2910),
    [anon_sym_srand] = ACTIONS(2910),
    [anon_sym_stat] = ACTIONS(2910),
    [anon_sym_study] = ACTIONS(2910),
    [anon_sym_tell] = ACTIONS(2910),
    [anon_sym_telldir] = ACTIONS(2910),
    [anon_sym_tied] = ACTIONS(2910),
    [anon_sym_uc] = ACTIONS(2910),
    [anon_sym_ucfirst] = ACTIONS(2910),
    [anon_sym_untie] = ACTIONS(2910),
    [anon_sym_umask] = ACTIONS(2910),
    [anon_sym_values] = ACTIONS(2910),
    [anon_sym_write] = ACTIONS(2910),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2910),
    [anon_sym___END__] = ACTIONS(2910),
    [anon_sym_EOT] = ACTIONS(2910),
    [anon_sym_q] = ACTIONS(2910),
    [anon_sym_qq] = ACTIONS(2910),
    [anon_sym_qw] = ACTIONS(2910),
    [anon_sym_qx] = ACTIONS(2910),
    [anon_sym_qr] = ACTIONS(2910),
    [anon_sym_m] = ACTIONS(2910),
    [anon_sym_s] = ACTIONS(2910),
    [anon_sym_tr] = ACTIONS(2910),
    [anon_sym_y] = ACTIONS(2910),
    [anon_sym_if] = ACTIONS(2910),
    [anon_sym_unless] = ACTIONS(2910),
    [anon_sym_while] = ACTIONS(2910),
    [anon_sym_until] = ACTIONS(2910),
    [aux_sym__bareword_token1] = ACTIONS(2910),
    [sym_number] = ACTIONS(2910),
    [anon_sym_true] = ACTIONS(2910),
    [anon_sym_false] = ACTIONS(2910),
    [sym__single_quote] = ACTIONS(2908),
    [sym__double_quote] = ACTIONS(2908),
    [sym__backtick_quote] = ACTIONS(2908),
    [sym__search_slash_quote] = ACTIONS(2908),
    [sym__open_readline_bracket] = ACTIONS(2908),
    [sym__open_fileglob_bracket] = ACTIONS(2908),
    [sym__PERLY_HEREDOC] = ACTIONS(2908),
    [sym__ctrl_z_hack] = ACTIONS(2908),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2908),
    [sym__filetest] = ACTIONS(2908),
  },
  [STATE(1440)] = {
    [sym_heredoc_content] = STATE(1440),
    [ts_builtin_sym_end] = ACTIONS(2912),
    [sym__identifier] = ACTIONS(2914),
    [anon_sym_LBRACE] = ACTIONS(2914),
    [anon_sym_SEMI] = ACTIONS(2914),
    [anon_sym_package] = ACTIONS(2914),
    [anon_sym_class] = ACTIONS(2914),
    [anon_sym_role] = ACTIONS(2914),
    [anon_sym_BUILD] = ACTIONS(2914),
    [anon_sym_ADJUST] = ACTIONS(2914),
    [anon_sym_DOLLAR] = ACTIONS(2914),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_my] = ACTIONS(2914),
    [anon_sym_extended] = ACTIONS(2914),
    [anon_sym_async] = ACTIONS(2914),
    [anon_sym_sub] = ACTIONS(2914),
    [anon_sym_method] = ACTIONS(2914),
    [anon_sym_LPAREN] = ACTIONS(2914),
    [anon_sym_continue] = ACTIONS(2914),
    [anon_sym_state] = ACTIONS(2914),
    [anon_sym_our] = ACTIONS(2914),
    [anon_sym_try] = ACTIONS(2914),
    [anon_sym_defer] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2914),
    [anon_sym_LT_LT] = ACTIONS(2914),
    [anon_sym_SLASH_SLASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2914),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_PLUS_PLUS] = ACTIONS(2914),
    [anon_sym_DASH_DASH] = ACTIONS(2914),
    [anon_sym_BSLASH] = ACTIONS(2912),
    [anon_sym_do] = ACTIONS(2914),
    [anon_sym_eval] = ACTIONS(2914),
    [anon_sym_await] = ACTIONS(2914),
    [anon_sym_field] = ACTIONS(2914),
    [anon_sym_local] = ACTIONS(2914),
    [anon_sym_dynamically] = ACTIONS(2914),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2914),
    [anon_sym_require] = ACTIONS(2914),
    [anon_sym_map] = ACTIONS(2914),
    [anon_sym_grep] = ACTIONS(2914),
    [anon_sym_sort] = ACTIONS(2914),
    [anon_sym_goto] = ACTIONS(2914),
    [anon_sym_return] = ACTIONS(2914),
    [anon_sym_undef] = ACTIONS(2914),
    [anon_sym_PERCENT2] = ACTIONS(2914),
    [anon_sym_AMP2] = ACTIONS(2914),
    [anon_sym_STAR2] = ACTIONS(2914),
    [anon_sym_use] = ACTIONS(2914),
    [anon_sym_no] = ACTIONS(2914),
    [anon_sym_for] = ACTIONS(2914),
    [anon_sym_foreach] = ACTIONS(2914),
    [anon_sym_last] = ACTIONS(2914),
    [anon_sym_next] = ACTIONS(2914),
    [anon_sym_redo] = ACTIONS(2914),
    [anon_sym_BEGIN] = ACTIONS(2914),
    [anon_sym_INIT] = ACTIONS(2914),
    [anon_sym_CHECK] = ACTIONS(2914),
    [anon_sym_UNITCHECK] = ACTIONS(2914),
    [anon_sym_END] = ACTIONS(2914),
    [anon_sym___FILE__] = ACTIONS(2914),
    [anon_sym___LINE__] = ACTIONS(2914),
    [anon_sym___PACKAGE__] = ACTIONS(2914),
    [anon_sym___SUB__] = ACTIONS(2914),
    [anon_sym_break] = ACTIONS(2914),
    [anon_sym_fork] = ACTIONS(2914),
    [anon_sym_getppid] = ACTIONS(2914),
    [anon_sym_time] = ACTIONS(2914),
    [anon_sym_times] = ACTIONS(2914),
    [anon_sym_wait] = ACTIONS(2914),
    [anon_sym_wantarray] = ACTIONS(2914),
    [anon_sym_abs] = ACTIONS(2914),
    [anon_sym_alarm] = ACTIONS(2914),
    [anon_sym_chop] = ACTIONS(2914),
    [anon_sym_chdir] = ACTIONS(2914),
    [anon_sym_close] = ACTIONS(2914),
    [anon_sym_closedir] = ACTIONS(2914),
    [anon_sym_caller] = ACTIONS(2914),
    [anon_sym_chomp] = ACTIONS(2914),
    [anon_sym_chr] = ACTIONS(2914),
    [anon_sym_cos] = ACTIONS(2914),
    [anon_sym_chroot] = ACTIONS(2914),
    [anon_sym_defined] = ACTIONS(2914),
    [anon_sym_delete] = ACTIONS(2914),
    [anon_sym_dbmclose] = ACTIONS(2914),
    [anon_sym_exists] = ACTIONS(2914),
    [anon_sym_exit] = ACTIONS(2914),
    [anon_sym_eof] = ACTIONS(2914),
    [anon_sym_exp] = ACTIONS(2914),
    [anon_sym_each] = ACTIONS(2914),
    [anon_sym_fc] = ACTIONS(2914),
    [anon_sym_fileno] = ACTIONS(2914),
    [anon_sym_gmtime] = ACTIONS(2914),
    [anon_sym_getc] = ACTIONS(2914),
    [anon_sym_getpgrp] = ACTIONS(2914),
    [anon_sym_getprotobyname] = ACTIONS(2914),
    [anon_sym_getpwname] = ACTIONS(2914),
    [anon_sym_getpwuid] = ACTIONS(2914),
    [anon_sym_getpeername] = ACTIONS(2914),
    [anon_sym_getnetbyname] = ACTIONS(2914),
    [anon_sym_getsockname] = ACTIONS(2914),
    [anon_sym_getgrnam] = ACTIONS(2914),
    [anon_sym_getgrgid] = ACTIONS(2914),
    [anon_sym_hex] = ACTIONS(2914),
    [anon_sym_int] = ACTIONS(2914),
    [anon_sym_keys] = ACTIONS(2914),
    [anon_sym_lc] = ACTIONS(2914),
    [anon_sym_lcfirst] = ACTIONS(2914),
    [anon_sym_length] = ACTIONS(2914),
    [anon_sym_localtime] = ACTIONS(2914),
    [anon_sym_log] = ACTIONS(2914),
    [anon_sym_lock] = ACTIONS(2914),
    [anon_sym_lstat] = ACTIONS(2914),
    [anon_sym_oct] = ACTIONS(2914),
    [anon_sym_ord] = ACTIONS(2914),
    [anon_sym_prototype] = ACTIONS(2914),
    [anon_sym_pop] = ACTIONS(2914),
    [anon_sym_pos] = ACTIONS(2914),
    [anon_sym_quotemeta] = ACTIONS(2914),
    [anon_sym_reset] = ACTIONS(2914),
    [anon_sym_rand] = ACTIONS(2914),
    [anon_sym_rmdir] = ACTIONS(2914),
    [anon_sym_readdir] = ACTIONS(2914),
    [anon_sym_readline] = ACTIONS(2914),
    [anon_sym_readpipe] = ACTIONS(2914),
    [anon_sym_rewinddir] = ACTIONS(2914),
    [anon_sym_readlink] = ACTIONS(2914),
    [anon_sym_ref] = ACTIONS(2914),
    [anon_sym_scalar] = ACTIONS(2914),
    [anon_sym_shift] = ACTIONS(2914),
    [anon_sym_sin] = ACTIONS(2914),
    [anon_sym_sleep] = ACTIONS(2914),
    [anon_sym_sqrt] = ACTIONS(2914),
    [anon_sym_srand] = ACTIONS(2914),
    [anon_sym_stat] = ACTIONS(2914),
    [anon_sym_study] = ACTIONS(2914),
    [anon_sym_tell] = ACTIONS(2914),
    [anon_sym_telldir] = ACTIONS(2914),
    [anon_sym_tied] = ACTIONS(2914),
    [anon_sym_uc] = ACTIONS(2914),
    [anon_sym_ucfirst] = ACTIONS(2914),
    [anon_sym_untie] = ACTIONS(2914),
    [anon_sym_umask] = ACTIONS(2914),
    [anon_sym_values] = ACTIONS(2914),
    [anon_sym_write] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2914),
    [anon_sym___END__] = ACTIONS(2914),
    [anon_sym_EOT] = ACTIONS(2914),
    [anon_sym_q] = ACTIONS(2914),
    [anon_sym_qq] = ACTIONS(2914),
    [anon_sym_qw] = ACTIONS(2914),
    [anon_sym_qx] = ACTIONS(2914),
    [anon_sym_qr] = ACTIONS(2914),
    [anon_sym_m] = ACTIONS(2914),
    [anon_sym_s] = ACTIONS(2914),
    [anon_sym_tr] = ACTIONS(2914),
    [anon_sym_y] = ACTIONS(2914),
    [anon_sym_if] = ACTIONS(2914),
    [anon_sym_unless] = ACTIONS(2914),
    [anon_sym_while] = ACTIONS(2914),
    [anon_sym_until] = ACTIONS(2914),
    [aux_sym__bareword_token1] = ACTIONS(2914),
    [sym_number] = ACTIONS(2914),
    [anon_sym_true] = ACTIONS(2914),
    [anon_sym_false] = ACTIONS(2914),
    [sym__single_quote] = ACTIONS(2912),
    [sym__double_quote] = ACTIONS(2912),
    [sym__backtick_quote] = ACTIONS(2912),
    [sym__search_slash_quote] = ACTIONS(2912),
    [sym__open_readline_bracket] = ACTIONS(2912),
    [sym__open_fileglob_bracket] = ACTIONS(2912),
    [sym__PERLY_HEREDOC] = ACTIONS(2912),
    [sym__ctrl_z_hack] = ACTIONS(2912),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2912),
    [sym__filetest] = ACTIONS(2912),
  },
  [STATE(1441)] = {
    [sym_heredoc_content] = STATE(1441),
    [ts_builtin_sym_end] = ACTIONS(2916),
    [sym__identifier] = ACTIONS(2918),
    [anon_sym_LBRACE] = ACTIONS(2918),
    [anon_sym_SEMI] = ACTIONS(2918),
    [anon_sym_package] = ACTIONS(2918),
    [anon_sym_class] = ACTIONS(2918),
    [anon_sym_role] = ACTIONS(2918),
    [anon_sym_BUILD] = ACTIONS(2918),
    [anon_sym_ADJUST] = ACTIONS(2918),
    [anon_sym_DOLLAR] = ACTIONS(2918),
    [anon_sym_AT] = ACTIONS(2918),
    [anon_sym_my] = ACTIONS(2918),
    [anon_sym_extended] = ACTIONS(2918),
    [anon_sym_async] = ACTIONS(2918),
    [anon_sym_sub] = ACTIONS(2918),
    [anon_sym_method] = ACTIONS(2918),
    [anon_sym_LPAREN] = ACTIONS(2918),
    [anon_sym_continue] = ACTIONS(2918),
    [anon_sym_state] = ACTIONS(2918),
    [anon_sym_our] = ACTIONS(2918),
    [anon_sym_try] = ACTIONS(2918),
    [anon_sym_defer] = ACTIONS(2918),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2918),
    [anon_sym_LBRACK] = ACTIONS(2918),
    [anon_sym_LT_LT] = ACTIONS(2918),
    [anon_sym_SLASH_SLASH] = ACTIONS(2918),
    [anon_sym_PLUS] = ACTIONS(2918),
    [anon_sym_DASH] = ACTIONS(2918),
    [anon_sym_TILDE] = ACTIONS(2918),
    [anon_sym_BANG] = ACTIONS(2918),
    [anon_sym_PLUS_PLUS] = ACTIONS(2918),
    [anon_sym_DASH_DASH] = ACTIONS(2918),
    [anon_sym_BSLASH] = ACTIONS(2916),
    [anon_sym_do] = ACTIONS(2918),
    [anon_sym_eval] = ACTIONS(2918),
    [anon_sym_await] = ACTIONS(2918),
    [anon_sym_field] = ACTIONS(2918),
    [anon_sym_local] = ACTIONS(2918),
    [anon_sym_dynamically] = ACTIONS(2918),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2918),
    [anon_sym_require] = ACTIONS(2918),
    [anon_sym_map] = ACTIONS(2918),
    [anon_sym_grep] = ACTIONS(2918),
    [anon_sym_sort] = ACTIONS(2918),
    [anon_sym_goto] = ACTIONS(2918),
    [anon_sym_return] = ACTIONS(2918),
    [anon_sym_undef] = ACTIONS(2918),
    [anon_sym_PERCENT2] = ACTIONS(2918),
    [anon_sym_AMP2] = ACTIONS(2918),
    [anon_sym_STAR2] = ACTIONS(2918),
    [anon_sym_use] = ACTIONS(2918),
    [anon_sym_no] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2918),
    [anon_sym_foreach] = ACTIONS(2918),
    [anon_sym_last] = ACTIONS(2918),
    [anon_sym_next] = ACTIONS(2918),
    [anon_sym_redo] = ACTIONS(2918),
    [anon_sym_BEGIN] = ACTIONS(2918),
    [anon_sym_INIT] = ACTIONS(2918),
    [anon_sym_CHECK] = ACTIONS(2918),
    [anon_sym_UNITCHECK] = ACTIONS(2918),
    [anon_sym_END] = ACTIONS(2918),
    [anon_sym___FILE__] = ACTIONS(2918),
    [anon_sym___LINE__] = ACTIONS(2918),
    [anon_sym___PACKAGE__] = ACTIONS(2918),
    [anon_sym___SUB__] = ACTIONS(2918),
    [anon_sym_break] = ACTIONS(2918),
    [anon_sym_fork] = ACTIONS(2918),
    [anon_sym_getppid] = ACTIONS(2918),
    [anon_sym_time] = ACTIONS(2918),
    [anon_sym_times] = ACTIONS(2918),
    [anon_sym_wait] = ACTIONS(2918),
    [anon_sym_wantarray] = ACTIONS(2918),
    [anon_sym_abs] = ACTIONS(2918),
    [anon_sym_alarm] = ACTIONS(2918),
    [anon_sym_chop] = ACTIONS(2918),
    [anon_sym_chdir] = ACTIONS(2918),
    [anon_sym_close] = ACTIONS(2918),
    [anon_sym_closedir] = ACTIONS(2918),
    [anon_sym_caller] = ACTIONS(2918),
    [anon_sym_chomp] = ACTIONS(2918),
    [anon_sym_chr] = ACTIONS(2918),
    [anon_sym_cos] = ACTIONS(2918),
    [anon_sym_chroot] = ACTIONS(2918),
    [anon_sym_defined] = ACTIONS(2918),
    [anon_sym_delete] = ACTIONS(2918),
    [anon_sym_dbmclose] = ACTIONS(2918),
    [anon_sym_exists] = ACTIONS(2918),
    [anon_sym_exit] = ACTIONS(2918),
    [anon_sym_eof] = ACTIONS(2918),
    [anon_sym_exp] = ACTIONS(2918),
    [anon_sym_each] = ACTIONS(2918),
    [anon_sym_fc] = ACTIONS(2918),
    [anon_sym_fileno] = ACTIONS(2918),
    [anon_sym_gmtime] = ACTIONS(2918),
    [anon_sym_getc] = ACTIONS(2918),
    [anon_sym_getpgrp] = ACTIONS(2918),
    [anon_sym_getprotobyname] = ACTIONS(2918),
    [anon_sym_getpwname] = ACTIONS(2918),
    [anon_sym_getpwuid] = ACTIONS(2918),
    [anon_sym_getpeername] = ACTIONS(2918),
    [anon_sym_getnetbyname] = ACTIONS(2918),
    [anon_sym_getsockname] = ACTIONS(2918),
    [anon_sym_getgrnam] = ACTIONS(2918),
    [anon_sym_getgrgid] = ACTIONS(2918),
    [anon_sym_hex] = ACTIONS(2918),
    [anon_sym_int] = ACTIONS(2918),
    [anon_sym_keys] = ACTIONS(2918),
    [anon_sym_lc] = ACTIONS(2918),
    [anon_sym_lcfirst] = ACTIONS(2918),
    [anon_sym_length] = ACTIONS(2918),
    [anon_sym_localtime] = ACTIONS(2918),
    [anon_sym_log] = ACTIONS(2918),
    [anon_sym_lock] = ACTIONS(2918),
    [anon_sym_lstat] = ACTIONS(2918),
    [anon_sym_oct] = ACTIONS(2918),
    [anon_sym_ord] = ACTIONS(2918),
    [anon_sym_prototype] = ACTIONS(2918),
    [anon_sym_pop] = ACTIONS(2918),
    [anon_sym_pos] = ACTIONS(2918),
    [anon_sym_quotemeta] = ACTIONS(2918),
    [anon_sym_reset] = ACTIONS(2918),
    [anon_sym_rand] = ACTIONS(2918),
    [anon_sym_rmdir] = ACTIONS(2918),
    [anon_sym_readdir] = ACTIONS(2918),
    [anon_sym_readline] = ACTIONS(2918),
    [anon_sym_readpipe] = ACTIONS(2918),
    [anon_sym_rewinddir] = ACTIONS(2918),
    [anon_sym_readlink] = ACTIONS(2918),
    [anon_sym_ref] = ACTIONS(2918),
    [anon_sym_scalar] = ACTIONS(2918),
    [anon_sym_shift] = ACTIONS(2918),
    [anon_sym_sin] = ACTIONS(2918),
    [anon_sym_sleep] = ACTIONS(2918),
    [anon_sym_sqrt] = ACTIONS(2918),
    [anon_sym_srand] = ACTIONS(2918),
    [anon_sym_stat] = ACTIONS(2918),
    [anon_sym_study] = ACTIONS(2918),
    [anon_sym_tell] = ACTIONS(2918),
    [anon_sym_telldir] = ACTIONS(2918),
    [anon_sym_tied] = ACTIONS(2918),
    [anon_sym_uc] = ACTIONS(2918),
    [anon_sym_ucfirst] = ACTIONS(2918),
    [anon_sym_untie] = ACTIONS(2918),
    [anon_sym_umask] = ACTIONS(2918),
    [anon_sym_values] = ACTIONS(2918),
    [anon_sym_write] = ACTIONS(2918),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2918),
    [anon_sym___END__] = ACTIONS(2918),
    [anon_sym_EOT] = ACTIONS(2918),
    [anon_sym_q] = ACTIONS(2918),
    [anon_sym_qq] = ACTIONS(2918),
    [anon_sym_qw] = ACTIONS(2918),
    [anon_sym_qx] = ACTIONS(2918),
    [anon_sym_qr] = ACTIONS(2918),
    [anon_sym_m] = ACTIONS(2918),
    [anon_sym_s] = ACTIONS(2918),
    [anon_sym_tr] = ACTIONS(2918),
    [anon_sym_y] = ACTIONS(2918),
    [anon_sym_if] = ACTIONS(2918),
    [anon_sym_unless] = ACTIONS(2918),
    [anon_sym_while] = ACTIONS(2918),
    [anon_sym_until] = ACTIONS(2918),
    [aux_sym__bareword_token1] = ACTIONS(2918),
    [sym_number] = ACTIONS(2918),
    [anon_sym_true] = ACTIONS(2918),
    [anon_sym_false] = ACTIONS(2918),
    [sym__single_quote] = ACTIONS(2916),
    [sym__double_quote] = ACTIONS(2916),
    [sym__backtick_quote] = ACTIONS(2916),
    [sym__search_slash_quote] = ACTIONS(2916),
    [sym__open_readline_bracket] = ACTIONS(2916),
    [sym__open_fileglob_bracket] = ACTIONS(2916),
    [sym__PERLY_HEREDOC] = ACTIONS(2916),
    [sym__ctrl_z_hack] = ACTIONS(2916),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2916),
    [sym__filetest] = ACTIONS(2916),
  },
  [STATE(1442)] = {
    [sym_heredoc_content] = STATE(1442),
    [ts_builtin_sym_end] = ACTIONS(2920),
    [sym__identifier] = ACTIONS(2922),
    [anon_sym_LBRACE] = ACTIONS(2922),
    [anon_sym_SEMI] = ACTIONS(2922),
    [anon_sym_package] = ACTIONS(2922),
    [anon_sym_class] = ACTIONS(2922),
    [anon_sym_role] = ACTIONS(2922),
    [anon_sym_BUILD] = ACTIONS(2922),
    [anon_sym_ADJUST] = ACTIONS(2922),
    [anon_sym_DOLLAR] = ACTIONS(2922),
    [anon_sym_AT] = ACTIONS(2922),
    [anon_sym_my] = ACTIONS(2922),
    [anon_sym_extended] = ACTIONS(2922),
    [anon_sym_async] = ACTIONS(2922),
    [anon_sym_sub] = ACTIONS(2922),
    [anon_sym_method] = ACTIONS(2922),
    [anon_sym_LPAREN] = ACTIONS(2922),
    [anon_sym_continue] = ACTIONS(2922),
    [anon_sym_state] = ACTIONS(2922),
    [anon_sym_our] = ACTIONS(2922),
    [anon_sym_try] = ACTIONS(2922),
    [anon_sym_defer] = ACTIONS(2922),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2922),
    [anon_sym_LBRACK] = ACTIONS(2922),
    [anon_sym_LT_LT] = ACTIONS(2922),
    [anon_sym_SLASH_SLASH] = ACTIONS(2922),
    [anon_sym_PLUS] = ACTIONS(2922),
    [anon_sym_DASH] = ACTIONS(2922),
    [anon_sym_TILDE] = ACTIONS(2922),
    [anon_sym_BANG] = ACTIONS(2922),
    [anon_sym_PLUS_PLUS] = ACTIONS(2922),
    [anon_sym_DASH_DASH] = ACTIONS(2922),
    [anon_sym_BSLASH] = ACTIONS(2920),
    [anon_sym_do] = ACTIONS(2922),
    [anon_sym_eval] = ACTIONS(2922),
    [anon_sym_await] = ACTIONS(2922),
    [anon_sym_field] = ACTIONS(2922),
    [anon_sym_local] = ACTIONS(2922),
    [anon_sym_dynamically] = ACTIONS(2922),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2922),
    [anon_sym_require] = ACTIONS(2922),
    [anon_sym_map] = ACTIONS(2922),
    [anon_sym_grep] = ACTIONS(2922),
    [anon_sym_sort] = ACTIONS(2922),
    [anon_sym_goto] = ACTIONS(2922),
    [anon_sym_return] = ACTIONS(2922),
    [anon_sym_undef] = ACTIONS(2922),
    [anon_sym_PERCENT2] = ACTIONS(2922),
    [anon_sym_AMP2] = ACTIONS(2922),
    [anon_sym_STAR2] = ACTIONS(2922),
    [anon_sym_use] = ACTIONS(2922),
    [anon_sym_no] = ACTIONS(2922),
    [anon_sym_for] = ACTIONS(2922),
    [anon_sym_foreach] = ACTIONS(2922),
    [anon_sym_last] = ACTIONS(2922),
    [anon_sym_next] = ACTIONS(2922),
    [anon_sym_redo] = ACTIONS(2922),
    [anon_sym_BEGIN] = ACTIONS(2922),
    [anon_sym_INIT] = ACTIONS(2922),
    [anon_sym_CHECK] = ACTIONS(2922),
    [anon_sym_UNITCHECK] = ACTIONS(2922),
    [anon_sym_END] = ACTIONS(2922),
    [anon_sym___FILE__] = ACTIONS(2922),
    [anon_sym___LINE__] = ACTIONS(2922),
    [anon_sym___PACKAGE__] = ACTIONS(2922),
    [anon_sym___SUB__] = ACTIONS(2922),
    [anon_sym_break] = ACTIONS(2922),
    [anon_sym_fork] = ACTIONS(2922),
    [anon_sym_getppid] = ACTIONS(2922),
    [anon_sym_time] = ACTIONS(2922),
    [anon_sym_times] = ACTIONS(2922),
    [anon_sym_wait] = ACTIONS(2922),
    [anon_sym_wantarray] = ACTIONS(2922),
    [anon_sym_abs] = ACTIONS(2922),
    [anon_sym_alarm] = ACTIONS(2922),
    [anon_sym_chop] = ACTIONS(2922),
    [anon_sym_chdir] = ACTIONS(2922),
    [anon_sym_close] = ACTIONS(2922),
    [anon_sym_closedir] = ACTIONS(2922),
    [anon_sym_caller] = ACTIONS(2922),
    [anon_sym_chomp] = ACTIONS(2922),
    [anon_sym_chr] = ACTIONS(2922),
    [anon_sym_cos] = ACTIONS(2922),
    [anon_sym_chroot] = ACTIONS(2922),
    [anon_sym_defined] = ACTIONS(2922),
    [anon_sym_delete] = ACTIONS(2922),
    [anon_sym_dbmclose] = ACTIONS(2922),
    [anon_sym_exists] = ACTIONS(2922),
    [anon_sym_exit] = ACTIONS(2922),
    [anon_sym_eof] = ACTIONS(2922),
    [anon_sym_exp] = ACTIONS(2922),
    [anon_sym_each] = ACTIONS(2922),
    [anon_sym_fc] = ACTIONS(2922),
    [anon_sym_fileno] = ACTIONS(2922),
    [anon_sym_gmtime] = ACTIONS(2922),
    [anon_sym_getc] = ACTIONS(2922),
    [anon_sym_getpgrp] = ACTIONS(2922),
    [anon_sym_getprotobyname] = ACTIONS(2922),
    [anon_sym_getpwname] = ACTIONS(2922),
    [anon_sym_getpwuid] = ACTIONS(2922),
    [anon_sym_getpeername] = ACTIONS(2922),
    [anon_sym_getnetbyname] = ACTIONS(2922),
    [anon_sym_getsockname] = ACTIONS(2922),
    [anon_sym_getgrnam] = ACTIONS(2922),
    [anon_sym_getgrgid] = ACTIONS(2922),
    [anon_sym_hex] = ACTIONS(2922),
    [anon_sym_int] = ACTIONS(2922),
    [anon_sym_keys] = ACTIONS(2922),
    [anon_sym_lc] = ACTIONS(2922),
    [anon_sym_lcfirst] = ACTIONS(2922),
    [anon_sym_length] = ACTIONS(2922),
    [anon_sym_localtime] = ACTIONS(2922),
    [anon_sym_log] = ACTIONS(2922),
    [anon_sym_lock] = ACTIONS(2922),
    [anon_sym_lstat] = ACTIONS(2922),
    [anon_sym_oct] = ACTIONS(2922),
    [anon_sym_ord] = ACTIONS(2922),
    [anon_sym_prototype] = ACTIONS(2922),
    [anon_sym_pop] = ACTIONS(2922),
    [anon_sym_pos] = ACTIONS(2922),
    [anon_sym_quotemeta] = ACTIONS(2922),
    [anon_sym_reset] = ACTIONS(2922),
    [anon_sym_rand] = ACTIONS(2922),
    [anon_sym_rmdir] = ACTIONS(2922),
    [anon_sym_readdir] = ACTIONS(2922),
    [anon_sym_readline] = ACTIONS(2922),
    [anon_sym_readpipe] = ACTIONS(2922),
    [anon_sym_rewinddir] = ACTIONS(2922),
    [anon_sym_readlink] = ACTIONS(2922),
    [anon_sym_ref] = ACTIONS(2922),
    [anon_sym_scalar] = ACTIONS(2922),
    [anon_sym_shift] = ACTIONS(2922),
    [anon_sym_sin] = ACTIONS(2922),
    [anon_sym_sleep] = ACTIONS(2922),
    [anon_sym_sqrt] = ACTIONS(2922),
    [anon_sym_srand] = ACTIONS(2922),
    [anon_sym_stat] = ACTIONS(2922),
    [anon_sym_study] = ACTIONS(2922),
    [anon_sym_tell] = ACTIONS(2922),
    [anon_sym_telldir] = ACTIONS(2922),
    [anon_sym_tied] = ACTIONS(2922),
    [anon_sym_uc] = ACTIONS(2922),
    [anon_sym_ucfirst] = ACTIONS(2922),
    [anon_sym_untie] = ACTIONS(2922),
    [anon_sym_umask] = ACTIONS(2922),
    [anon_sym_values] = ACTIONS(2922),
    [anon_sym_write] = ACTIONS(2922),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2922),
    [anon_sym___END__] = ACTIONS(2922),
    [anon_sym_EOT] = ACTIONS(2922),
    [anon_sym_q] = ACTIONS(2922),
    [anon_sym_qq] = ACTIONS(2922),
    [anon_sym_qw] = ACTIONS(2922),
    [anon_sym_qx] = ACTIONS(2922),
    [anon_sym_qr] = ACTIONS(2922),
    [anon_sym_m] = ACTIONS(2922),
    [anon_sym_s] = ACTIONS(2922),
    [anon_sym_tr] = ACTIONS(2922),
    [anon_sym_y] = ACTIONS(2922),
    [anon_sym_if] = ACTIONS(2922),
    [anon_sym_unless] = ACTIONS(2922),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_until] = ACTIONS(2922),
    [aux_sym__bareword_token1] = ACTIONS(2922),
    [sym_number] = ACTIONS(2922),
    [anon_sym_true] = ACTIONS(2922),
    [anon_sym_false] = ACTIONS(2922),
    [sym__single_quote] = ACTIONS(2920),
    [sym__double_quote] = ACTIONS(2920),
    [sym__backtick_quote] = ACTIONS(2920),
    [sym__search_slash_quote] = ACTIONS(2920),
    [sym__open_readline_bracket] = ACTIONS(2920),
    [sym__open_fileglob_bracket] = ACTIONS(2920),
    [sym__PERLY_HEREDOC] = ACTIONS(2920),
    [sym__ctrl_z_hack] = ACTIONS(2920),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2920),
    [sym__filetest] = ACTIONS(2920),
  },
  [STATE(1443)] = {
    [sym_heredoc_content] = STATE(1443),
    [ts_builtin_sym_end] = ACTIONS(2924),
    [sym__identifier] = ACTIONS(2926),
    [anon_sym_LBRACE] = ACTIONS(2926),
    [anon_sym_SEMI] = ACTIONS(2926),
    [anon_sym_package] = ACTIONS(2926),
    [anon_sym_class] = ACTIONS(2926),
    [anon_sym_role] = ACTIONS(2926),
    [anon_sym_BUILD] = ACTIONS(2926),
    [anon_sym_ADJUST] = ACTIONS(2926),
    [anon_sym_DOLLAR] = ACTIONS(2926),
    [anon_sym_AT] = ACTIONS(2926),
    [anon_sym_my] = ACTIONS(2926),
    [anon_sym_extended] = ACTIONS(2926),
    [anon_sym_async] = ACTIONS(2926),
    [anon_sym_sub] = ACTIONS(2926),
    [anon_sym_method] = ACTIONS(2926),
    [anon_sym_LPAREN] = ACTIONS(2926),
    [anon_sym_continue] = ACTIONS(2926),
    [anon_sym_state] = ACTIONS(2926),
    [anon_sym_our] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2926),
    [anon_sym_defer] = ACTIONS(2926),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2926),
    [anon_sym_LBRACK] = ACTIONS(2926),
    [anon_sym_LT_LT] = ACTIONS(2926),
    [anon_sym_SLASH_SLASH] = ACTIONS(2926),
    [anon_sym_PLUS] = ACTIONS(2926),
    [anon_sym_DASH] = ACTIONS(2926),
    [anon_sym_TILDE] = ACTIONS(2926),
    [anon_sym_BANG] = ACTIONS(2926),
    [anon_sym_PLUS_PLUS] = ACTIONS(2926),
    [anon_sym_DASH_DASH] = ACTIONS(2926),
    [anon_sym_BSLASH] = ACTIONS(2924),
    [anon_sym_do] = ACTIONS(2926),
    [anon_sym_eval] = ACTIONS(2926),
    [anon_sym_await] = ACTIONS(2926),
    [anon_sym_field] = ACTIONS(2926),
    [anon_sym_local] = ACTIONS(2926),
    [anon_sym_dynamically] = ACTIONS(2926),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2926),
    [anon_sym_require] = ACTIONS(2926),
    [anon_sym_map] = ACTIONS(2926),
    [anon_sym_grep] = ACTIONS(2926),
    [anon_sym_sort] = ACTIONS(2926),
    [anon_sym_goto] = ACTIONS(2926),
    [anon_sym_return] = ACTIONS(2926),
    [anon_sym_undef] = ACTIONS(2926),
    [anon_sym_PERCENT2] = ACTIONS(2926),
    [anon_sym_AMP2] = ACTIONS(2926),
    [anon_sym_STAR2] = ACTIONS(2926),
    [anon_sym_use] = ACTIONS(2926),
    [anon_sym_no] = ACTIONS(2926),
    [anon_sym_for] = ACTIONS(2926),
    [anon_sym_foreach] = ACTIONS(2926),
    [anon_sym_last] = ACTIONS(2926),
    [anon_sym_next] = ACTIONS(2926),
    [anon_sym_redo] = ACTIONS(2926),
    [anon_sym_BEGIN] = ACTIONS(2926),
    [anon_sym_INIT] = ACTIONS(2926),
    [anon_sym_CHECK] = ACTIONS(2926),
    [anon_sym_UNITCHECK] = ACTIONS(2926),
    [anon_sym_END] = ACTIONS(2926),
    [anon_sym___FILE__] = ACTIONS(2926),
    [anon_sym___LINE__] = ACTIONS(2926),
    [anon_sym___PACKAGE__] = ACTIONS(2926),
    [anon_sym___SUB__] = ACTIONS(2926),
    [anon_sym_break] = ACTIONS(2926),
    [anon_sym_fork] = ACTIONS(2926),
    [anon_sym_getppid] = ACTIONS(2926),
    [anon_sym_time] = ACTIONS(2926),
    [anon_sym_times] = ACTIONS(2926),
    [anon_sym_wait] = ACTIONS(2926),
    [anon_sym_wantarray] = ACTIONS(2926),
    [anon_sym_abs] = ACTIONS(2926),
    [anon_sym_alarm] = ACTIONS(2926),
    [anon_sym_chop] = ACTIONS(2926),
    [anon_sym_chdir] = ACTIONS(2926),
    [anon_sym_close] = ACTIONS(2926),
    [anon_sym_closedir] = ACTIONS(2926),
    [anon_sym_caller] = ACTIONS(2926),
    [anon_sym_chomp] = ACTIONS(2926),
    [anon_sym_chr] = ACTIONS(2926),
    [anon_sym_cos] = ACTIONS(2926),
    [anon_sym_chroot] = ACTIONS(2926),
    [anon_sym_defined] = ACTIONS(2926),
    [anon_sym_delete] = ACTIONS(2926),
    [anon_sym_dbmclose] = ACTIONS(2926),
    [anon_sym_exists] = ACTIONS(2926),
    [anon_sym_exit] = ACTIONS(2926),
    [anon_sym_eof] = ACTIONS(2926),
    [anon_sym_exp] = ACTIONS(2926),
    [anon_sym_each] = ACTIONS(2926),
    [anon_sym_fc] = ACTIONS(2926),
    [anon_sym_fileno] = ACTIONS(2926),
    [anon_sym_gmtime] = ACTIONS(2926),
    [anon_sym_getc] = ACTIONS(2926),
    [anon_sym_getpgrp] = ACTIONS(2926),
    [anon_sym_getprotobyname] = ACTIONS(2926),
    [anon_sym_getpwname] = ACTIONS(2926),
    [anon_sym_getpwuid] = ACTIONS(2926),
    [anon_sym_getpeername] = ACTIONS(2926),
    [anon_sym_getnetbyname] = ACTIONS(2926),
    [anon_sym_getsockname] = ACTIONS(2926),
    [anon_sym_getgrnam] = ACTIONS(2926),
    [anon_sym_getgrgid] = ACTIONS(2926),
    [anon_sym_hex] = ACTIONS(2926),
    [anon_sym_int] = ACTIONS(2926),
    [anon_sym_keys] = ACTIONS(2926),
    [anon_sym_lc] = ACTIONS(2926),
    [anon_sym_lcfirst] = ACTIONS(2926),
    [anon_sym_length] = ACTIONS(2926),
    [anon_sym_localtime] = ACTIONS(2926),
    [anon_sym_log] = ACTIONS(2926),
    [anon_sym_lock] = ACTIONS(2926),
    [anon_sym_lstat] = ACTIONS(2926),
    [anon_sym_oct] = ACTIONS(2926),
    [anon_sym_ord] = ACTIONS(2926),
    [anon_sym_prototype] = ACTIONS(2926),
    [anon_sym_pop] = ACTIONS(2926),
    [anon_sym_pos] = ACTIONS(2926),
    [anon_sym_quotemeta] = ACTIONS(2926),
    [anon_sym_reset] = ACTIONS(2926),
    [anon_sym_rand] = ACTIONS(2926),
    [anon_sym_rmdir] = ACTIONS(2926),
    [anon_sym_readdir] = ACTIONS(2926),
    [anon_sym_readline] = ACTIONS(2926),
    [anon_sym_readpipe] = ACTIONS(2926),
    [anon_sym_rewinddir] = ACTIONS(2926),
    [anon_sym_readlink] = ACTIONS(2926),
    [anon_sym_ref] = ACTIONS(2926),
    [anon_sym_scalar] = ACTIONS(2926),
    [anon_sym_shift] = ACTIONS(2926),
    [anon_sym_sin] = ACTIONS(2926),
    [anon_sym_sleep] = ACTIONS(2926),
    [anon_sym_sqrt] = ACTIONS(2926),
    [anon_sym_srand] = ACTIONS(2926),
    [anon_sym_stat] = ACTIONS(2926),
    [anon_sym_study] = ACTIONS(2926),
    [anon_sym_tell] = ACTIONS(2926),
    [anon_sym_telldir] = ACTIONS(2926),
    [anon_sym_tied] = ACTIONS(2926),
    [anon_sym_uc] = ACTIONS(2926),
    [anon_sym_ucfirst] = ACTIONS(2926),
    [anon_sym_untie] = ACTIONS(2926),
    [anon_sym_umask] = ACTIONS(2926),
    [anon_sym_values] = ACTIONS(2926),
    [anon_sym_write] = ACTIONS(2926),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2926),
    [anon_sym___END__] = ACTIONS(2926),
    [anon_sym_EOT] = ACTIONS(2926),
    [anon_sym_q] = ACTIONS(2926),
    [anon_sym_qq] = ACTIONS(2926),
    [anon_sym_qw] = ACTIONS(2926),
    [anon_sym_qx] = ACTIONS(2926),
    [anon_sym_qr] = ACTIONS(2926),
    [anon_sym_m] = ACTIONS(2926),
    [anon_sym_s] = ACTIONS(2926),
    [anon_sym_tr] = ACTIONS(2926),
    [anon_sym_y] = ACTIONS(2926),
    [anon_sym_if] = ACTIONS(2926),
    [anon_sym_unless] = ACTIONS(2926),
    [anon_sym_while] = ACTIONS(2926),
    [anon_sym_until] = ACTIONS(2926),
    [aux_sym__bareword_token1] = ACTIONS(2926),
    [sym_number] = ACTIONS(2926),
    [anon_sym_true] = ACTIONS(2926),
    [anon_sym_false] = ACTIONS(2926),
    [sym__single_quote] = ACTIONS(2924),
    [sym__double_quote] = ACTIONS(2924),
    [sym__backtick_quote] = ACTIONS(2924),
    [sym__search_slash_quote] = ACTIONS(2924),
    [sym__open_readline_bracket] = ACTIONS(2924),
    [sym__open_fileglob_bracket] = ACTIONS(2924),
    [sym__PERLY_HEREDOC] = ACTIONS(2924),
    [sym__ctrl_z_hack] = ACTIONS(2924),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2924),
    [sym__filetest] = ACTIONS(2924),
  },
  [STATE(1444)] = {
    [sym_heredoc_content] = STATE(1444),
    [ts_builtin_sym_end] = ACTIONS(2928),
    [sym__identifier] = ACTIONS(2930),
    [anon_sym_LBRACE] = ACTIONS(2930),
    [anon_sym_SEMI] = ACTIONS(2930),
    [anon_sym_package] = ACTIONS(2930),
    [anon_sym_class] = ACTIONS(2930),
    [anon_sym_role] = ACTIONS(2930),
    [anon_sym_BUILD] = ACTIONS(2930),
    [anon_sym_ADJUST] = ACTIONS(2930),
    [anon_sym_DOLLAR] = ACTIONS(2930),
    [anon_sym_AT] = ACTIONS(2930),
    [anon_sym_my] = ACTIONS(2930),
    [anon_sym_extended] = ACTIONS(2930),
    [anon_sym_async] = ACTIONS(2930),
    [anon_sym_sub] = ACTIONS(2930),
    [anon_sym_method] = ACTIONS(2930),
    [anon_sym_LPAREN] = ACTIONS(2930),
    [anon_sym_continue] = ACTIONS(2930),
    [anon_sym_state] = ACTIONS(2930),
    [anon_sym_our] = ACTIONS(2930),
    [anon_sym_try] = ACTIONS(2930),
    [anon_sym_defer] = ACTIONS(2930),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2930),
    [anon_sym_LBRACK] = ACTIONS(2930),
    [anon_sym_LT_LT] = ACTIONS(2930),
    [anon_sym_SLASH_SLASH] = ACTIONS(2930),
    [anon_sym_PLUS] = ACTIONS(2930),
    [anon_sym_DASH] = ACTIONS(2930),
    [anon_sym_TILDE] = ACTIONS(2930),
    [anon_sym_BANG] = ACTIONS(2930),
    [anon_sym_PLUS_PLUS] = ACTIONS(2930),
    [anon_sym_DASH_DASH] = ACTIONS(2930),
    [anon_sym_BSLASH] = ACTIONS(2928),
    [anon_sym_do] = ACTIONS(2930),
    [anon_sym_eval] = ACTIONS(2930),
    [anon_sym_await] = ACTIONS(2930),
    [anon_sym_field] = ACTIONS(2930),
    [anon_sym_local] = ACTIONS(2930),
    [anon_sym_dynamically] = ACTIONS(2930),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2930),
    [anon_sym_require] = ACTIONS(2930),
    [anon_sym_map] = ACTIONS(2930),
    [anon_sym_grep] = ACTIONS(2930),
    [anon_sym_sort] = ACTIONS(2930),
    [anon_sym_goto] = ACTIONS(2930),
    [anon_sym_return] = ACTIONS(2930),
    [anon_sym_undef] = ACTIONS(2930),
    [anon_sym_PERCENT2] = ACTIONS(2930),
    [anon_sym_AMP2] = ACTIONS(2930),
    [anon_sym_STAR2] = ACTIONS(2930),
    [anon_sym_use] = ACTIONS(2930),
    [anon_sym_no] = ACTIONS(2930),
    [anon_sym_for] = ACTIONS(2930),
    [anon_sym_foreach] = ACTIONS(2930),
    [anon_sym_last] = ACTIONS(2930),
    [anon_sym_next] = ACTIONS(2930),
    [anon_sym_redo] = ACTIONS(2930),
    [anon_sym_BEGIN] = ACTIONS(2930),
    [anon_sym_INIT] = ACTIONS(2930),
    [anon_sym_CHECK] = ACTIONS(2930),
    [anon_sym_UNITCHECK] = ACTIONS(2930),
    [anon_sym_END] = ACTIONS(2930),
    [anon_sym___FILE__] = ACTIONS(2930),
    [anon_sym___LINE__] = ACTIONS(2930),
    [anon_sym___PACKAGE__] = ACTIONS(2930),
    [anon_sym___SUB__] = ACTIONS(2930),
    [anon_sym_break] = ACTIONS(2930),
    [anon_sym_fork] = ACTIONS(2930),
    [anon_sym_getppid] = ACTIONS(2930),
    [anon_sym_time] = ACTIONS(2930),
    [anon_sym_times] = ACTIONS(2930),
    [anon_sym_wait] = ACTIONS(2930),
    [anon_sym_wantarray] = ACTIONS(2930),
    [anon_sym_abs] = ACTIONS(2930),
    [anon_sym_alarm] = ACTIONS(2930),
    [anon_sym_chop] = ACTIONS(2930),
    [anon_sym_chdir] = ACTIONS(2930),
    [anon_sym_close] = ACTIONS(2930),
    [anon_sym_closedir] = ACTIONS(2930),
    [anon_sym_caller] = ACTIONS(2930),
    [anon_sym_chomp] = ACTIONS(2930),
    [anon_sym_chr] = ACTIONS(2930),
    [anon_sym_cos] = ACTIONS(2930),
    [anon_sym_chroot] = ACTIONS(2930),
    [anon_sym_defined] = ACTIONS(2930),
    [anon_sym_delete] = ACTIONS(2930),
    [anon_sym_dbmclose] = ACTIONS(2930),
    [anon_sym_exists] = ACTIONS(2930),
    [anon_sym_exit] = ACTIONS(2930),
    [anon_sym_eof] = ACTIONS(2930),
    [anon_sym_exp] = ACTIONS(2930),
    [anon_sym_each] = ACTIONS(2930),
    [anon_sym_fc] = ACTIONS(2930),
    [anon_sym_fileno] = ACTIONS(2930),
    [anon_sym_gmtime] = ACTIONS(2930),
    [anon_sym_getc] = ACTIONS(2930),
    [anon_sym_getpgrp] = ACTIONS(2930),
    [anon_sym_getprotobyname] = ACTIONS(2930),
    [anon_sym_getpwname] = ACTIONS(2930),
    [anon_sym_getpwuid] = ACTIONS(2930),
    [anon_sym_getpeername] = ACTIONS(2930),
    [anon_sym_getnetbyname] = ACTIONS(2930),
    [anon_sym_getsockname] = ACTIONS(2930),
    [anon_sym_getgrnam] = ACTIONS(2930),
    [anon_sym_getgrgid] = ACTIONS(2930),
    [anon_sym_hex] = ACTIONS(2930),
    [anon_sym_int] = ACTIONS(2930),
    [anon_sym_keys] = ACTIONS(2930),
    [anon_sym_lc] = ACTIONS(2930),
    [anon_sym_lcfirst] = ACTIONS(2930),
    [anon_sym_length] = ACTIONS(2930),
    [anon_sym_localtime] = ACTIONS(2930),
    [anon_sym_log] = ACTIONS(2930),
    [anon_sym_lock] = ACTIONS(2930),
    [anon_sym_lstat] = ACTIONS(2930),
    [anon_sym_oct] = ACTIONS(2930),
    [anon_sym_ord] = ACTIONS(2930),
    [anon_sym_prototype] = ACTIONS(2930),
    [anon_sym_pop] = ACTIONS(2930),
    [anon_sym_pos] = ACTIONS(2930),
    [anon_sym_quotemeta] = ACTIONS(2930),
    [anon_sym_reset] = ACTIONS(2930),
    [anon_sym_rand] = ACTIONS(2930),
    [anon_sym_rmdir] = ACTIONS(2930),
    [anon_sym_readdir] = ACTIONS(2930),
    [anon_sym_readline] = ACTIONS(2930),
    [anon_sym_readpipe] = ACTIONS(2930),
    [anon_sym_rewinddir] = ACTIONS(2930),
    [anon_sym_readlink] = ACTIONS(2930),
    [anon_sym_ref] = ACTIONS(2930),
    [anon_sym_scalar] = ACTIONS(2930),
    [anon_sym_shift] = ACTIONS(2930),
    [anon_sym_sin] = ACTIONS(2930),
    [anon_sym_sleep] = ACTIONS(2930),
    [anon_sym_sqrt] = ACTIONS(2930),
    [anon_sym_srand] = ACTIONS(2930),
    [anon_sym_stat] = ACTIONS(2930),
    [anon_sym_study] = ACTIONS(2930),
    [anon_sym_tell] = ACTIONS(2930),
    [anon_sym_telldir] = ACTIONS(2930),
    [anon_sym_tied] = ACTIONS(2930),
    [anon_sym_uc] = ACTIONS(2930),
    [anon_sym_ucfirst] = ACTIONS(2930),
    [anon_sym_untie] = ACTIONS(2930),
    [anon_sym_umask] = ACTIONS(2930),
    [anon_sym_values] = ACTIONS(2930),
    [anon_sym_write] = ACTIONS(2930),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2930),
    [anon_sym___END__] = ACTIONS(2930),
    [anon_sym_EOT] = ACTIONS(2930),
    [anon_sym_q] = ACTIONS(2930),
    [anon_sym_qq] = ACTIONS(2930),
    [anon_sym_qw] = ACTIONS(2930),
    [anon_sym_qx] = ACTIONS(2930),
    [anon_sym_qr] = ACTIONS(2930),
    [anon_sym_m] = ACTIONS(2930),
    [anon_sym_s] = ACTIONS(2930),
    [anon_sym_tr] = ACTIONS(2930),
    [anon_sym_y] = ACTIONS(2930),
    [anon_sym_if] = ACTIONS(2930),
    [anon_sym_unless] = ACTIONS(2930),
    [anon_sym_while] = ACTIONS(2930),
    [anon_sym_until] = ACTIONS(2930),
    [aux_sym__bareword_token1] = ACTIONS(2930),
    [sym_number] = ACTIONS(2930),
    [anon_sym_true] = ACTIONS(2930),
    [anon_sym_false] = ACTIONS(2930),
    [sym__single_quote] = ACTIONS(2928),
    [sym__double_quote] = ACTIONS(2928),
    [sym__backtick_quote] = ACTIONS(2928),
    [sym__search_slash_quote] = ACTIONS(2928),
    [sym__open_readline_bracket] = ACTIONS(2928),
    [sym__open_fileglob_bracket] = ACTIONS(2928),
    [sym__PERLY_HEREDOC] = ACTIONS(2928),
    [sym__ctrl_z_hack] = ACTIONS(2928),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2928),
    [sym__filetest] = ACTIONS(2928),
  },
  [STATE(1445)] = {
    [sym_heredoc_content] = STATE(1445),
    [ts_builtin_sym_end] = ACTIONS(2932),
    [sym__identifier] = ACTIONS(2934),
    [anon_sym_LBRACE] = ACTIONS(2934),
    [anon_sym_SEMI] = ACTIONS(2934),
    [anon_sym_package] = ACTIONS(2934),
    [anon_sym_class] = ACTIONS(2934),
    [anon_sym_role] = ACTIONS(2934),
    [anon_sym_BUILD] = ACTIONS(2934),
    [anon_sym_ADJUST] = ACTIONS(2934),
    [anon_sym_DOLLAR] = ACTIONS(2934),
    [anon_sym_AT] = ACTIONS(2934),
    [anon_sym_my] = ACTIONS(2934),
    [anon_sym_extended] = ACTIONS(2934),
    [anon_sym_async] = ACTIONS(2934),
    [anon_sym_sub] = ACTIONS(2934),
    [anon_sym_method] = ACTIONS(2934),
    [anon_sym_LPAREN] = ACTIONS(2934),
    [anon_sym_continue] = ACTIONS(2934),
    [anon_sym_state] = ACTIONS(2934),
    [anon_sym_our] = ACTIONS(2934),
    [anon_sym_try] = ACTIONS(2934),
    [anon_sym_defer] = ACTIONS(2934),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2934),
    [anon_sym_LBRACK] = ACTIONS(2934),
    [anon_sym_LT_LT] = ACTIONS(2934),
    [anon_sym_SLASH_SLASH] = ACTIONS(2934),
    [anon_sym_PLUS] = ACTIONS(2934),
    [anon_sym_DASH] = ACTIONS(2934),
    [anon_sym_TILDE] = ACTIONS(2934),
    [anon_sym_BANG] = ACTIONS(2934),
    [anon_sym_PLUS_PLUS] = ACTIONS(2934),
    [anon_sym_DASH_DASH] = ACTIONS(2934),
    [anon_sym_BSLASH] = ACTIONS(2932),
    [anon_sym_do] = ACTIONS(2934),
    [anon_sym_eval] = ACTIONS(2934),
    [anon_sym_await] = ACTIONS(2934),
    [anon_sym_field] = ACTIONS(2934),
    [anon_sym_local] = ACTIONS(2934),
    [anon_sym_dynamically] = ACTIONS(2934),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2934),
    [anon_sym_require] = ACTIONS(2934),
    [anon_sym_map] = ACTIONS(2934),
    [anon_sym_grep] = ACTIONS(2934),
    [anon_sym_sort] = ACTIONS(2934),
    [anon_sym_goto] = ACTIONS(2934),
    [anon_sym_return] = ACTIONS(2934),
    [anon_sym_undef] = ACTIONS(2934),
    [anon_sym_PERCENT2] = ACTIONS(2934),
    [anon_sym_AMP2] = ACTIONS(2934),
    [anon_sym_STAR2] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2934),
    [anon_sym_no] = ACTIONS(2934),
    [anon_sym_for] = ACTIONS(2934),
    [anon_sym_foreach] = ACTIONS(2934),
    [anon_sym_last] = ACTIONS(2934),
    [anon_sym_next] = ACTIONS(2934),
    [anon_sym_redo] = ACTIONS(2934),
    [anon_sym_BEGIN] = ACTIONS(2934),
    [anon_sym_INIT] = ACTIONS(2934),
    [anon_sym_CHECK] = ACTIONS(2934),
    [anon_sym_UNITCHECK] = ACTIONS(2934),
    [anon_sym_END] = ACTIONS(2934),
    [anon_sym___FILE__] = ACTIONS(2934),
    [anon_sym___LINE__] = ACTIONS(2934),
    [anon_sym___PACKAGE__] = ACTIONS(2934),
    [anon_sym___SUB__] = ACTIONS(2934),
    [anon_sym_break] = ACTIONS(2934),
    [anon_sym_fork] = ACTIONS(2934),
    [anon_sym_getppid] = ACTIONS(2934),
    [anon_sym_time] = ACTIONS(2934),
    [anon_sym_times] = ACTIONS(2934),
    [anon_sym_wait] = ACTIONS(2934),
    [anon_sym_wantarray] = ACTIONS(2934),
    [anon_sym_abs] = ACTIONS(2934),
    [anon_sym_alarm] = ACTIONS(2934),
    [anon_sym_chop] = ACTIONS(2934),
    [anon_sym_chdir] = ACTIONS(2934),
    [anon_sym_close] = ACTIONS(2934),
    [anon_sym_closedir] = ACTIONS(2934),
    [anon_sym_caller] = ACTIONS(2934),
    [anon_sym_chomp] = ACTIONS(2934),
    [anon_sym_chr] = ACTIONS(2934),
    [anon_sym_cos] = ACTIONS(2934),
    [anon_sym_chroot] = ACTIONS(2934),
    [anon_sym_defined] = ACTIONS(2934),
    [anon_sym_delete] = ACTIONS(2934),
    [anon_sym_dbmclose] = ACTIONS(2934),
    [anon_sym_exists] = ACTIONS(2934),
    [anon_sym_exit] = ACTIONS(2934),
    [anon_sym_eof] = ACTIONS(2934),
    [anon_sym_exp] = ACTIONS(2934),
    [anon_sym_each] = ACTIONS(2934),
    [anon_sym_fc] = ACTIONS(2934),
    [anon_sym_fileno] = ACTIONS(2934),
    [anon_sym_gmtime] = ACTIONS(2934),
    [anon_sym_getc] = ACTIONS(2934),
    [anon_sym_getpgrp] = ACTIONS(2934),
    [anon_sym_getprotobyname] = ACTIONS(2934),
    [anon_sym_getpwname] = ACTIONS(2934),
    [anon_sym_getpwuid] = ACTIONS(2934),
    [anon_sym_getpeername] = ACTIONS(2934),
    [anon_sym_getnetbyname] = ACTIONS(2934),
    [anon_sym_getsockname] = ACTIONS(2934),
    [anon_sym_getgrnam] = ACTIONS(2934),
    [anon_sym_getgrgid] = ACTIONS(2934),
    [anon_sym_hex] = ACTIONS(2934),
    [anon_sym_int] = ACTIONS(2934),
    [anon_sym_keys] = ACTIONS(2934),
    [anon_sym_lc] = ACTIONS(2934),
    [anon_sym_lcfirst] = ACTIONS(2934),
    [anon_sym_length] = ACTIONS(2934),
    [anon_sym_localtime] = ACTIONS(2934),
    [anon_sym_log] = ACTIONS(2934),
    [anon_sym_lock] = ACTIONS(2934),
    [anon_sym_lstat] = ACTIONS(2934),
    [anon_sym_oct] = ACTIONS(2934),
    [anon_sym_ord] = ACTIONS(2934),
    [anon_sym_prototype] = ACTIONS(2934),
    [anon_sym_pop] = ACTIONS(2934),
    [anon_sym_pos] = ACTIONS(2934),
    [anon_sym_quotemeta] = ACTIONS(2934),
    [anon_sym_reset] = ACTIONS(2934),
    [anon_sym_rand] = ACTIONS(2934),
    [anon_sym_rmdir] = ACTIONS(2934),
    [anon_sym_readdir] = ACTIONS(2934),
    [anon_sym_readline] = ACTIONS(2934),
    [anon_sym_readpipe] = ACTIONS(2934),
    [anon_sym_rewinddir] = ACTIONS(2934),
    [anon_sym_readlink] = ACTIONS(2934),
    [anon_sym_ref] = ACTIONS(2934),
    [anon_sym_scalar] = ACTIONS(2934),
    [anon_sym_shift] = ACTIONS(2934),
    [anon_sym_sin] = ACTIONS(2934),
    [anon_sym_sleep] = ACTIONS(2934),
    [anon_sym_sqrt] = ACTIONS(2934),
    [anon_sym_srand] = ACTIONS(2934),
    [anon_sym_stat] = ACTIONS(2934),
    [anon_sym_study] = ACTIONS(2934),
    [anon_sym_tell] = ACTIONS(2934),
    [anon_sym_telldir] = ACTIONS(2934),
    [anon_sym_tied] = ACTIONS(2934),
    [anon_sym_uc] = ACTIONS(2934),
    [anon_sym_ucfirst] = ACTIONS(2934),
    [anon_sym_untie] = ACTIONS(2934),
    [anon_sym_umask] = ACTIONS(2934),
    [anon_sym_values] = ACTIONS(2934),
    [anon_sym_write] = ACTIONS(2934),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2934),
    [anon_sym___END__] = ACTIONS(2934),
    [anon_sym_EOT] = ACTIONS(2934),
    [anon_sym_q] = ACTIONS(2934),
    [anon_sym_qq] = ACTIONS(2934),
    [anon_sym_qw] = ACTIONS(2934),
    [anon_sym_qx] = ACTIONS(2934),
    [anon_sym_qr] = ACTIONS(2934),
    [anon_sym_m] = ACTIONS(2934),
    [anon_sym_s] = ACTIONS(2934),
    [anon_sym_tr] = ACTIONS(2934),
    [anon_sym_y] = ACTIONS(2934),
    [anon_sym_if] = ACTIONS(2934),
    [anon_sym_unless] = ACTIONS(2934),
    [anon_sym_while] = ACTIONS(2934),
    [anon_sym_until] = ACTIONS(2934),
    [aux_sym__bareword_token1] = ACTIONS(2934),
    [sym_number] = ACTIONS(2934),
    [anon_sym_true] = ACTIONS(2934),
    [anon_sym_false] = ACTIONS(2934),
    [sym__single_quote] = ACTIONS(2932),
    [sym__double_quote] = ACTIONS(2932),
    [sym__backtick_quote] = ACTIONS(2932),
    [sym__search_slash_quote] = ACTIONS(2932),
    [sym__open_readline_bracket] = ACTIONS(2932),
    [sym__open_fileglob_bracket] = ACTIONS(2932),
    [sym__PERLY_HEREDOC] = ACTIONS(2932),
    [sym__ctrl_z_hack] = ACTIONS(2932),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2932),
    [sym__filetest] = ACTIONS(2932),
  },
  [STATE(1446)] = {
    [sym_heredoc_content] = STATE(1446),
    [ts_builtin_sym_end] = ACTIONS(2936),
    [sym__identifier] = ACTIONS(2938),
    [anon_sym_LBRACE] = ACTIONS(2938),
    [anon_sym_SEMI] = ACTIONS(2938),
    [anon_sym_package] = ACTIONS(2938),
    [anon_sym_class] = ACTIONS(2938),
    [anon_sym_role] = ACTIONS(2938),
    [anon_sym_BUILD] = ACTIONS(2938),
    [anon_sym_ADJUST] = ACTIONS(2938),
    [anon_sym_DOLLAR] = ACTIONS(2938),
    [anon_sym_AT] = ACTIONS(2938),
    [anon_sym_my] = ACTIONS(2938),
    [anon_sym_extended] = ACTIONS(2938),
    [anon_sym_async] = ACTIONS(2938),
    [anon_sym_sub] = ACTIONS(2938),
    [anon_sym_method] = ACTIONS(2938),
    [anon_sym_LPAREN] = ACTIONS(2938),
    [anon_sym_continue] = ACTIONS(2938),
    [anon_sym_state] = ACTIONS(2938),
    [anon_sym_our] = ACTIONS(2938),
    [anon_sym_try] = ACTIONS(2938),
    [anon_sym_defer] = ACTIONS(2938),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2938),
    [anon_sym_LBRACK] = ACTIONS(2938),
    [anon_sym_LT_LT] = ACTIONS(2938),
    [anon_sym_SLASH_SLASH] = ACTIONS(2938),
    [anon_sym_PLUS] = ACTIONS(2938),
    [anon_sym_DASH] = ACTIONS(2938),
    [anon_sym_TILDE] = ACTIONS(2938),
    [anon_sym_BANG] = ACTIONS(2938),
    [anon_sym_PLUS_PLUS] = ACTIONS(2938),
    [anon_sym_DASH_DASH] = ACTIONS(2938),
    [anon_sym_BSLASH] = ACTIONS(2936),
    [anon_sym_do] = ACTIONS(2938),
    [anon_sym_eval] = ACTIONS(2938),
    [anon_sym_await] = ACTIONS(2938),
    [anon_sym_field] = ACTIONS(2938),
    [anon_sym_local] = ACTIONS(2938),
    [anon_sym_dynamically] = ACTIONS(2938),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2938),
    [anon_sym_require] = ACTIONS(2938),
    [anon_sym_map] = ACTIONS(2938),
    [anon_sym_grep] = ACTIONS(2938),
    [anon_sym_sort] = ACTIONS(2938),
    [anon_sym_goto] = ACTIONS(2938),
    [anon_sym_return] = ACTIONS(2938),
    [anon_sym_undef] = ACTIONS(2938),
    [anon_sym_PERCENT2] = ACTIONS(2938),
    [anon_sym_AMP2] = ACTIONS(2938),
    [anon_sym_STAR2] = ACTIONS(2938),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_no] = ACTIONS(2938),
    [anon_sym_for] = ACTIONS(2938),
    [anon_sym_foreach] = ACTIONS(2938),
    [anon_sym_last] = ACTIONS(2938),
    [anon_sym_next] = ACTIONS(2938),
    [anon_sym_redo] = ACTIONS(2938),
    [anon_sym_BEGIN] = ACTIONS(2938),
    [anon_sym_INIT] = ACTIONS(2938),
    [anon_sym_CHECK] = ACTIONS(2938),
    [anon_sym_UNITCHECK] = ACTIONS(2938),
    [anon_sym_END] = ACTIONS(2938),
    [anon_sym___FILE__] = ACTIONS(2938),
    [anon_sym___LINE__] = ACTIONS(2938),
    [anon_sym___PACKAGE__] = ACTIONS(2938),
    [anon_sym___SUB__] = ACTIONS(2938),
    [anon_sym_break] = ACTIONS(2938),
    [anon_sym_fork] = ACTIONS(2938),
    [anon_sym_getppid] = ACTIONS(2938),
    [anon_sym_time] = ACTIONS(2938),
    [anon_sym_times] = ACTIONS(2938),
    [anon_sym_wait] = ACTIONS(2938),
    [anon_sym_wantarray] = ACTIONS(2938),
    [anon_sym_abs] = ACTIONS(2938),
    [anon_sym_alarm] = ACTIONS(2938),
    [anon_sym_chop] = ACTIONS(2938),
    [anon_sym_chdir] = ACTIONS(2938),
    [anon_sym_close] = ACTIONS(2938),
    [anon_sym_closedir] = ACTIONS(2938),
    [anon_sym_caller] = ACTIONS(2938),
    [anon_sym_chomp] = ACTIONS(2938),
    [anon_sym_chr] = ACTIONS(2938),
    [anon_sym_cos] = ACTIONS(2938),
    [anon_sym_chroot] = ACTIONS(2938),
    [anon_sym_defined] = ACTIONS(2938),
    [anon_sym_delete] = ACTIONS(2938),
    [anon_sym_dbmclose] = ACTIONS(2938),
    [anon_sym_exists] = ACTIONS(2938),
    [anon_sym_exit] = ACTIONS(2938),
    [anon_sym_eof] = ACTIONS(2938),
    [anon_sym_exp] = ACTIONS(2938),
    [anon_sym_each] = ACTIONS(2938),
    [anon_sym_fc] = ACTIONS(2938),
    [anon_sym_fileno] = ACTIONS(2938),
    [anon_sym_gmtime] = ACTIONS(2938),
    [anon_sym_getc] = ACTIONS(2938),
    [anon_sym_getpgrp] = ACTIONS(2938),
    [anon_sym_getprotobyname] = ACTIONS(2938),
    [anon_sym_getpwname] = ACTIONS(2938),
    [anon_sym_getpwuid] = ACTIONS(2938),
    [anon_sym_getpeername] = ACTIONS(2938),
    [anon_sym_getnetbyname] = ACTIONS(2938),
    [anon_sym_getsockname] = ACTIONS(2938),
    [anon_sym_getgrnam] = ACTIONS(2938),
    [anon_sym_getgrgid] = ACTIONS(2938),
    [anon_sym_hex] = ACTIONS(2938),
    [anon_sym_int] = ACTIONS(2938),
    [anon_sym_keys] = ACTIONS(2938),
    [anon_sym_lc] = ACTIONS(2938),
    [anon_sym_lcfirst] = ACTIONS(2938),
    [anon_sym_length] = ACTIONS(2938),
    [anon_sym_localtime] = ACTIONS(2938),
    [anon_sym_log] = ACTIONS(2938),
    [anon_sym_lock] = ACTIONS(2938),
    [anon_sym_lstat] = ACTIONS(2938),
    [anon_sym_oct] = ACTIONS(2938),
    [anon_sym_ord] = ACTIONS(2938),
    [anon_sym_prototype] = ACTIONS(2938),
    [anon_sym_pop] = ACTIONS(2938),
    [anon_sym_pos] = ACTIONS(2938),
    [anon_sym_quotemeta] = ACTIONS(2938),
    [anon_sym_reset] = ACTIONS(2938),
    [anon_sym_rand] = ACTIONS(2938),
    [anon_sym_rmdir] = ACTIONS(2938),
    [anon_sym_readdir] = ACTIONS(2938),
    [anon_sym_readline] = ACTIONS(2938),
    [anon_sym_readpipe] = ACTIONS(2938),
    [anon_sym_rewinddir] = ACTIONS(2938),
    [anon_sym_readlink] = ACTIONS(2938),
    [anon_sym_ref] = ACTIONS(2938),
    [anon_sym_scalar] = ACTIONS(2938),
    [anon_sym_shift] = ACTIONS(2938),
    [anon_sym_sin] = ACTIONS(2938),
    [anon_sym_sleep] = ACTIONS(2938),
    [anon_sym_sqrt] = ACTIONS(2938),
    [anon_sym_srand] = ACTIONS(2938),
    [anon_sym_stat] = ACTIONS(2938),
    [anon_sym_study] = ACTIONS(2938),
    [anon_sym_tell] = ACTIONS(2938),
    [anon_sym_telldir] = ACTIONS(2938),
    [anon_sym_tied] = ACTIONS(2938),
    [anon_sym_uc] = ACTIONS(2938),
    [anon_sym_ucfirst] = ACTIONS(2938),
    [anon_sym_untie] = ACTIONS(2938),
    [anon_sym_umask] = ACTIONS(2938),
    [anon_sym_values] = ACTIONS(2938),
    [anon_sym_write] = ACTIONS(2938),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2938),
    [anon_sym___END__] = ACTIONS(2938),
    [anon_sym_EOT] = ACTIONS(2938),
    [anon_sym_q] = ACTIONS(2938),
    [anon_sym_qq] = ACTIONS(2938),
    [anon_sym_qw] = ACTIONS(2938),
    [anon_sym_qx] = ACTIONS(2938),
    [anon_sym_qr] = ACTIONS(2938),
    [anon_sym_m] = ACTIONS(2938),
    [anon_sym_s] = ACTIONS(2938),
    [anon_sym_tr] = ACTIONS(2938),
    [anon_sym_y] = ACTIONS(2938),
    [anon_sym_if] = ACTIONS(2938),
    [anon_sym_unless] = ACTIONS(2938),
    [anon_sym_while] = ACTIONS(2938),
    [anon_sym_until] = ACTIONS(2938),
    [aux_sym__bareword_token1] = ACTIONS(2938),
    [sym_number] = ACTIONS(2938),
    [anon_sym_true] = ACTIONS(2938),
    [anon_sym_false] = ACTIONS(2938),
    [sym__single_quote] = ACTIONS(2936),
    [sym__double_quote] = ACTIONS(2936),
    [sym__backtick_quote] = ACTIONS(2936),
    [sym__search_slash_quote] = ACTIONS(2936),
    [sym__open_readline_bracket] = ACTIONS(2936),
    [sym__open_fileglob_bracket] = ACTIONS(2936),
    [sym__PERLY_HEREDOC] = ACTIONS(2936),
    [sym__ctrl_z_hack] = ACTIONS(2936),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2936),
    [sym__filetest] = ACTIONS(2936),
  },
  [STATE(1447)] = {
    [sym_heredoc_content] = STATE(1447),
    [ts_builtin_sym_end] = ACTIONS(2940),
    [sym__identifier] = ACTIONS(2942),
    [anon_sym_LBRACE] = ACTIONS(2942),
    [anon_sym_SEMI] = ACTIONS(2942),
    [anon_sym_package] = ACTIONS(2942),
    [anon_sym_class] = ACTIONS(2942),
    [anon_sym_role] = ACTIONS(2942),
    [anon_sym_BUILD] = ACTIONS(2942),
    [anon_sym_ADJUST] = ACTIONS(2942),
    [anon_sym_DOLLAR] = ACTIONS(2942),
    [anon_sym_AT] = ACTIONS(2942),
    [anon_sym_my] = ACTIONS(2942),
    [anon_sym_extended] = ACTIONS(2942),
    [anon_sym_async] = ACTIONS(2942),
    [anon_sym_sub] = ACTIONS(2942),
    [anon_sym_method] = ACTIONS(2942),
    [anon_sym_LPAREN] = ACTIONS(2942),
    [anon_sym_continue] = ACTIONS(2942),
    [anon_sym_state] = ACTIONS(2942),
    [anon_sym_our] = ACTIONS(2942),
    [anon_sym_try] = ACTIONS(2942),
    [anon_sym_defer] = ACTIONS(2942),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2942),
    [anon_sym_LBRACK] = ACTIONS(2942),
    [anon_sym_LT_LT] = ACTIONS(2942),
    [anon_sym_SLASH_SLASH] = ACTIONS(2942),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_TILDE] = ACTIONS(2942),
    [anon_sym_BANG] = ACTIONS(2942),
    [anon_sym_PLUS_PLUS] = ACTIONS(2942),
    [anon_sym_DASH_DASH] = ACTIONS(2942),
    [anon_sym_BSLASH] = ACTIONS(2940),
    [anon_sym_do] = ACTIONS(2942),
    [anon_sym_eval] = ACTIONS(2942),
    [anon_sym_await] = ACTIONS(2942),
    [anon_sym_field] = ACTIONS(2942),
    [anon_sym_local] = ACTIONS(2942),
    [anon_sym_dynamically] = ACTIONS(2942),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2942),
    [anon_sym_require] = ACTIONS(2942),
    [anon_sym_map] = ACTIONS(2942),
    [anon_sym_grep] = ACTIONS(2942),
    [anon_sym_sort] = ACTIONS(2942),
    [anon_sym_goto] = ACTIONS(2942),
    [anon_sym_return] = ACTIONS(2942),
    [anon_sym_undef] = ACTIONS(2942),
    [anon_sym_PERCENT2] = ACTIONS(2942),
    [anon_sym_AMP2] = ACTIONS(2942),
    [anon_sym_STAR2] = ACTIONS(2942),
    [anon_sym_use] = ACTIONS(2942),
    [anon_sym_no] = ACTIONS(2942),
    [anon_sym_for] = ACTIONS(2942),
    [anon_sym_foreach] = ACTIONS(2942),
    [anon_sym_last] = ACTIONS(2942),
    [anon_sym_next] = ACTIONS(2942),
    [anon_sym_redo] = ACTIONS(2942),
    [anon_sym_BEGIN] = ACTIONS(2942),
    [anon_sym_INIT] = ACTIONS(2942),
    [anon_sym_CHECK] = ACTIONS(2942),
    [anon_sym_UNITCHECK] = ACTIONS(2942),
    [anon_sym_END] = ACTIONS(2942),
    [anon_sym___FILE__] = ACTIONS(2942),
    [anon_sym___LINE__] = ACTIONS(2942),
    [anon_sym___PACKAGE__] = ACTIONS(2942),
    [anon_sym___SUB__] = ACTIONS(2942),
    [anon_sym_break] = ACTIONS(2942),
    [anon_sym_fork] = ACTIONS(2942),
    [anon_sym_getppid] = ACTIONS(2942),
    [anon_sym_time] = ACTIONS(2942),
    [anon_sym_times] = ACTIONS(2942),
    [anon_sym_wait] = ACTIONS(2942),
    [anon_sym_wantarray] = ACTIONS(2942),
    [anon_sym_abs] = ACTIONS(2942),
    [anon_sym_alarm] = ACTIONS(2942),
    [anon_sym_chop] = ACTIONS(2942),
    [anon_sym_chdir] = ACTIONS(2942),
    [anon_sym_close] = ACTIONS(2942),
    [anon_sym_closedir] = ACTIONS(2942),
    [anon_sym_caller] = ACTIONS(2942),
    [anon_sym_chomp] = ACTIONS(2942),
    [anon_sym_chr] = ACTIONS(2942),
    [anon_sym_cos] = ACTIONS(2942),
    [anon_sym_chroot] = ACTIONS(2942),
    [anon_sym_defined] = ACTIONS(2942),
    [anon_sym_delete] = ACTIONS(2942),
    [anon_sym_dbmclose] = ACTIONS(2942),
    [anon_sym_exists] = ACTIONS(2942),
    [anon_sym_exit] = ACTIONS(2942),
    [anon_sym_eof] = ACTIONS(2942),
    [anon_sym_exp] = ACTIONS(2942),
    [anon_sym_each] = ACTIONS(2942),
    [anon_sym_fc] = ACTIONS(2942),
    [anon_sym_fileno] = ACTIONS(2942),
    [anon_sym_gmtime] = ACTIONS(2942),
    [anon_sym_getc] = ACTIONS(2942),
    [anon_sym_getpgrp] = ACTIONS(2942),
    [anon_sym_getprotobyname] = ACTIONS(2942),
    [anon_sym_getpwname] = ACTIONS(2942),
    [anon_sym_getpwuid] = ACTIONS(2942),
    [anon_sym_getpeername] = ACTIONS(2942),
    [anon_sym_getnetbyname] = ACTIONS(2942),
    [anon_sym_getsockname] = ACTIONS(2942),
    [anon_sym_getgrnam] = ACTIONS(2942),
    [anon_sym_getgrgid] = ACTIONS(2942),
    [anon_sym_hex] = ACTIONS(2942),
    [anon_sym_int] = ACTIONS(2942),
    [anon_sym_keys] = ACTIONS(2942),
    [anon_sym_lc] = ACTIONS(2942),
    [anon_sym_lcfirst] = ACTIONS(2942),
    [anon_sym_length] = ACTIONS(2942),
    [anon_sym_localtime] = ACTIONS(2942),
    [anon_sym_log] = ACTIONS(2942),
    [anon_sym_lock] = ACTIONS(2942),
    [anon_sym_lstat] = ACTIONS(2942),
    [anon_sym_oct] = ACTIONS(2942),
    [anon_sym_ord] = ACTIONS(2942),
    [anon_sym_prototype] = ACTIONS(2942),
    [anon_sym_pop] = ACTIONS(2942),
    [anon_sym_pos] = ACTIONS(2942),
    [anon_sym_quotemeta] = ACTIONS(2942),
    [anon_sym_reset] = ACTIONS(2942),
    [anon_sym_rand] = ACTIONS(2942),
    [anon_sym_rmdir] = ACTIONS(2942),
    [anon_sym_readdir] = ACTIONS(2942),
    [anon_sym_readline] = ACTIONS(2942),
    [anon_sym_readpipe] = ACTIONS(2942),
    [anon_sym_rewinddir] = ACTIONS(2942),
    [anon_sym_readlink] = ACTIONS(2942),
    [anon_sym_ref] = ACTIONS(2942),
    [anon_sym_scalar] = ACTIONS(2942),
    [anon_sym_shift] = ACTIONS(2942),
    [anon_sym_sin] = ACTIONS(2942),
    [anon_sym_sleep] = ACTIONS(2942),
    [anon_sym_sqrt] = ACTIONS(2942),
    [anon_sym_srand] = ACTIONS(2942),
    [anon_sym_stat] = ACTIONS(2942),
    [anon_sym_study] = ACTIONS(2942),
    [anon_sym_tell] = ACTIONS(2942),
    [anon_sym_telldir] = ACTIONS(2942),
    [anon_sym_tied] = ACTIONS(2942),
    [anon_sym_uc] = ACTIONS(2942),
    [anon_sym_ucfirst] = ACTIONS(2942),
    [anon_sym_untie] = ACTIONS(2942),
    [anon_sym_umask] = ACTIONS(2942),
    [anon_sym_values] = ACTIONS(2942),
    [anon_sym_write] = ACTIONS(2942),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2942),
    [anon_sym___END__] = ACTIONS(2942),
    [anon_sym_EOT] = ACTIONS(2942),
    [anon_sym_q] = ACTIONS(2942),
    [anon_sym_qq] = ACTIONS(2942),
    [anon_sym_qw] = ACTIONS(2942),
    [anon_sym_qx] = ACTIONS(2942),
    [anon_sym_qr] = ACTIONS(2942),
    [anon_sym_m] = ACTIONS(2942),
    [anon_sym_s] = ACTIONS(2942),
    [anon_sym_tr] = ACTIONS(2942),
    [anon_sym_y] = ACTIONS(2942),
    [anon_sym_if] = ACTIONS(2942),
    [anon_sym_unless] = ACTIONS(2942),
    [anon_sym_while] = ACTIONS(2942),
    [anon_sym_until] = ACTIONS(2942),
    [aux_sym__bareword_token1] = ACTIONS(2942),
    [sym_number] = ACTIONS(2942),
    [anon_sym_true] = ACTIONS(2942),
    [anon_sym_false] = ACTIONS(2942),
    [sym__single_quote] = ACTIONS(2940),
    [sym__double_quote] = ACTIONS(2940),
    [sym__backtick_quote] = ACTIONS(2940),
    [sym__search_slash_quote] = ACTIONS(2940),
    [sym__open_readline_bracket] = ACTIONS(2940),
    [sym__open_fileglob_bracket] = ACTIONS(2940),
    [sym__PERLY_HEREDOC] = ACTIONS(2940),
    [sym__ctrl_z_hack] = ACTIONS(2940),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2940),
    [sym__filetest] = ACTIONS(2940),
  },
  [STATE(1448)] = {
    [sym_heredoc_content] = STATE(1448),
    [ts_builtin_sym_end] = ACTIONS(2944),
    [sym__identifier] = ACTIONS(2946),
    [anon_sym_LBRACE] = ACTIONS(2946),
    [anon_sym_SEMI] = ACTIONS(2946),
    [anon_sym_package] = ACTIONS(2946),
    [anon_sym_class] = ACTIONS(2946),
    [anon_sym_role] = ACTIONS(2946),
    [anon_sym_BUILD] = ACTIONS(2946),
    [anon_sym_ADJUST] = ACTIONS(2946),
    [anon_sym_DOLLAR] = ACTIONS(2946),
    [anon_sym_AT] = ACTIONS(2946),
    [anon_sym_my] = ACTIONS(2946),
    [anon_sym_extended] = ACTIONS(2946),
    [anon_sym_async] = ACTIONS(2946),
    [anon_sym_sub] = ACTIONS(2946),
    [anon_sym_method] = ACTIONS(2946),
    [anon_sym_LPAREN] = ACTIONS(2946),
    [anon_sym_continue] = ACTIONS(2946),
    [anon_sym_state] = ACTIONS(2946),
    [anon_sym_our] = ACTIONS(2946),
    [anon_sym_try] = ACTIONS(2946),
    [anon_sym_defer] = ACTIONS(2946),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2946),
    [anon_sym_LBRACK] = ACTIONS(2946),
    [anon_sym_LT_LT] = ACTIONS(2946),
    [anon_sym_SLASH_SLASH] = ACTIONS(2946),
    [anon_sym_PLUS] = ACTIONS(2946),
    [anon_sym_DASH] = ACTIONS(2946),
    [anon_sym_TILDE] = ACTIONS(2946),
    [anon_sym_BANG] = ACTIONS(2946),
    [anon_sym_PLUS_PLUS] = ACTIONS(2946),
    [anon_sym_DASH_DASH] = ACTIONS(2946),
    [anon_sym_BSLASH] = ACTIONS(2944),
    [anon_sym_do] = ACTIONS(2946),
    [anon_sym_eval] = ACTIONS(2946),
    [anon_sym_await] = ACTIONS(2946),
    [anon_sym_field] = ACTIONS(2946),
    [anon_sym_local] = ACTIONS(2946),
    [anon_sym_dynamically] = ACTIONS(2946),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2946),
    [anon_sym_require] = ACTIONS(2946),
    [anon_sym_map] = ACTIONS(2946),
    [anon_sym_grep] = ACTIONS(2946),
    [anon_sym_sort] = ACTIONS(2946),
    [anon_sym_goto] = ACTIONS(2946),
    [anon_sym_return] = ACTIONS(2946),
    [anon_sym_undef] = ACTIONS(2946),
    [anon_sym_PERCENT2] = ACTIONS(2946),
    [anon_sym_AMP2] = ACTIONS(2946),
    [anon_sym_STAR2] = ACTIONS(2946),
    [anon_sym_use] = ACTIONS(2946),
    [anon_sym_no] = ACTIONS(2946),
    [anon_sym_for] = ACTIONS(2946),
    [anon_sym_foreach] = ACTIONS(2946),
    [anon_sym_last] = ACTIONS(2946),
    [anon_sym_next] = ACTIONS(2946),
    [anon_sym_redo] = ACTIONS(2946),
    [anon_sym_BEGIN] = ACTIONS(2946),
    [anon_sym_INIT] = ACTIONS(2946),
    [anon_sym_CHECK] = ACTIONS(2946),
    [anon_sym_UNITCHECK] = ACTIONS(2946),
    [anon_sym_END] = ACTIONS(2946),
    [anon_sym___FILE__] = ACTIONS(2946),
    [anon_sym___LINE__] = ACTIONS(2946),
    [anon_sym___PACKAGE__] = ACTIONS(2946),
    [anon_sym___SUB__] = ACTIONS(2946),
    [anon_sym_break] = ACTIONS(2946),
    [anon_sym_fork] = ACTIONS(2946),
    [anon_sym_getppid] = ACTIONS(2946),
    [anon_sym_time] = ACTIONS(2946),
    [anon_sym_times] = ACTIONS(2946),
    [anon_sym_wait] = ACTIONS(2946),
    [anon_sym_wantarray] = ACTIONS(2946),
    [anon_sym_abs] = ACTIONS(2946),
    [anon_sym_alarm] = ACTIONS(2946),
    [anon_sym_chop] = ACTIONS(2946),
    [anon_sym_chdir] = ACTIONS(2946),
    [anon_sym_close] = ACTIONS(2946),
    [anon_sym_closedir] = ACTIONS(2946),
    [anon_sym_caller] = ACTIONS(2946),
    [anon_sym_chomp] = ACTIONS(2946),
    [anon_sym_chr] = ACTIONS(2946),
    [anon_sym_cos] = ACTIONS(2946),
    [anon_sym_chroot] = ACTIONS(2946),
    [anon_sym_defined] = ACTIONS(2946),
    [anon_sym_delete] = ACTIONS(2946),
    [anon_sym_dbmclose] = ACTIONS(2946),
    [anon_sym_exists] = ACTIONS(2946),
    [anon_sym_exit] = ACTIONS(2946),
    [anon_sym_eof] = ACTIONS(2946),
    [anon_sym_exp] = ACTIONS(2946),
    [anon_sym_each] = ACTIONS(2946),
    [anon_sym_fc] = ACTIONS(2946),
    [anon_sym_fileno] = ACTIONS(2946),
    [anon_sym_gmtime] = ACTIONS(2946),
    [anon_sym_getc] = ACTIONS(2946),
    [anon_sym_getpgrp] = ACTIONS(2946),
    [anon_sym_getprotobyname] = ACTIONS(2946),
    [anon_sym_getpwname] = ACTIONS(2946),
    [anon_sym_getpwuid] = ACTIONS(2946),
    [anon_sym_getpeername] = ACTIONS(2946),
    [anon_sym_getnetbyname] = ACTIONS(2946),
    [anon_sym_getsockname] = ACTIONS(2946),
    [anon_sym_getgrnam] = ACTIONS(2946),
    [anon_sym_getgrgid] = ACTIONS(2946),
    [anon_sym_hex] = ACTIONS(2946),
    [anon_sym_int] = ACTIONS(2946),
    [anon_sym_keys] = ACTIONS(2946),
    [anon_sym_lc] = ACTIONS(2946),
    [anon_sym_lcfirst] = ACTIONS(2946),
    [anon_sym_length] = ACTIONS(2946),
    [anon_sym_localtime] = ACTIONS(2946),
    [anon_sym_log] = ACTIONS(2946),
    [anon_sym_lock] = ACTIONS(2946),
    [anon_sym_lstat] = ACTIONS(2946),
    [anon_sym_oct] = ACTIONS(2946),
    [anon_sym_ord] = ACTIONS(2946),
    [anon_sym_prototype] = ACTIONS(2946),
    [anon_sym_pop] = ACTIONS(2946),
    [anon_sym_pos] = ACTIONS(2946),
    [anon_sym_quotemeta] = ACTIONS(2946),
    [anon_sym_reset] = ACTIONS(2946),
    [anon_sym_rand] = ACTIONS(2946),
    [anon_sym_rmdir] = ACTIONS(2946),
    [anon_sym_readdir] = ACTIONS(2946),
    [anon_sym_readline] = ACTIONS(2946),
    [anon_sym_readpipe] = ACTIONS(2946),
    [anon_sym_rewinddir] = ACTIONS(2946),
    [anon_sym_readlink] = ACTIONS(2946),
    [anon_sym_ref] = ACTIONS(2946),
    [anon_sym_scalar] = ACTIONS(2946),
    [anon_sym_shift] = ACTIONS(2946),
    [anon_sym_sin] = ACTIONS(2946),
    [anon_sym_sleep] = ACTIONS(2946),
    [anon_sym_sqrt] = ACTIONS(2946),
    [anon_sym_srand] = ACTIONS(2946),
    [anon_sym_stat] = ACTIONS(2946),
    [anon_sym_study] = ACTIONS(2946),
    [anon_sym_tell] = ACTIONS(2946),
    [anon_sym_telldir] = ACTIONS(2946),
    [anon_sym_tied] = ACTIONS(2946),
    [anon_sym_uc] = ACTIONS(2946),
    [anon_sym_ucfirst] = ACTIONS(2946),
    [anon_sym_untie] = ACTIONS(2946),
    [anon_sym_umask] = ACTIONS(2946),
    [anon_sym_values] = ACTIONS(2946),
    [anon_sym_write] = ACTIONS(2946),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2946),
    [anon_sym___END__] = ACTIONS(2946),
    [anon_sym_EOT] = ACTIONS(2946),
    [anon_sym_q] = ACTIONS(2946),
    [anon_sym_qq] = ACTIONS(2946),
    [anon_sym_qw] = ACTIONS(2946),
    [anon_sym_qx] = ACTIONS(2946),
    [anon_sym_qr] = ACTIONS(2946),
    [anon_sym_m] = ACTIONS(2946),
    [anon_sym_s] = ACTIONS(2946),
    [anon_sym_tr] = ACTIONS(2946),
    [anon_sym_y] = ACTIONS(2946),
    [anon_sym_if] = ACTIONS(2946),
    [anon_sym_unless] = ACTIONS(2946),
    [anon_sym_while] = ACTIONS(2946),
    [anon_sym_until] = ACTIONS(2946),
    [aux_sym__bareword_token1] = ACTIONS(2946),
    [sym_number] = ACTIONS(2946),
    [anon_sym_true] = ACTIONS(2946),
    [anon_sym_false] = ACTIONS(2946),
    [sym__single_quote] = ACTIONS(2944),
    [sym__double_quote] = ACTIONS(2944),
    [sym__backtick_quote] = ACTIONS(2944),
    [sym__search_slash_quote] = ACTIONS(2944),
    [sym__open_readline_bracket] = ACTIONS(2944),
    [sym__open_fileglob_bracket] = ACTIONS(2944),
    [sym__PERLY_HEREDOC] = ACTIONS(2944),
    [sym__ctrl_z_hack] = ACTIONS(2944),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2944),
    [sym__filetest] = ACTIONS(2944),
  },
  [STATE(1449)] = {
    [sym_heredoc_content] = STATE(1449),
    [ts_builtin_sym_end] = ACTIONS(2948),
    [sym__identifier] = ACTIONS(2950),
    [anon_sym_LBRACE] = ACTIONS(2950),
    [anon_sym_SEMI] = ACTIONS(2950),
    [anon_sym_package] = ACTIONS(2950),
    [anon_sym_class] = ACTIONS(2950),
    [anon_sym_role] = ACTIONS(2950),
    [anon_sym_BUILD] = ACTIONS(2950),
    [anon_sym_ADJUST] = ACTIONS(2950),
    [anon_sym_DOLLAR] = ACTIONS(2950),
    [anon_sym_AT] = ACTIONS(2950),
    [anon_sym_my] = ACTIONS(2950),
    [anon_sym_extended] = ACTIONS(2950),
    [anon_sym_async] = ACTIONS(2950),
    [anon_sym_sub] = ACTIONS(2950),
    [anon_sym_method] = ACTIONS(2950),
    [anon_sym_LPAREN] = ACTIONS(2950),
    [anon_sym_continue] = ACTIONS(2950),
    [anon_sym_state] = ACTIONS(2950),
    [anon_sym_our] = ACTIONS(2950),
    [anon_sym_try] = ACTIONS(2950),
    [anon_sym_defer] = ACTIONS(2950),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2950),
    [anon_sym_LBRACK] = ACTIONS(2950),
    [anon_sym_LT_LT] = ACTIONS(2950),
    [anon_sym_SLASH_SLASH] = ACTIONS(2950),
    [anon_sym_PLUS] = ACTIONS(2950),
    [anon_sym_DASH] = ACTIONS(2950),
    [anon_sym_TILDE] = ACTIONS(2950),
    [anon_sym_BANG] = ACTIONS(2950),
    [anon_sym_PLUS_PLUS] = ACTIONS(2950),
    [anon_sym_DASH_DASH] = ACTIONS(2950),
    [anon_sym_BSLASH] = ACTIONS(2948),
    [anon_sym_do] = ACTIONS(2950),
    [anon_sym_eval] = ACTIONS(2950),
    [anon_sym_await] = ACTIONS(2950),
    [anon_sym_field] = ACTIONS(2950),
    [anon_sym_local] = ACTIONS(2950),
    [anon_sym_dynamically] = ACTIONS(2950),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2950),
    [anon_sym_require] = ACTIONS(2950),
    [anon_sym_map] = ACTIONS(2950),
    [anon_sym_grep] = ACTIONS(2950),
    [anon_sym_sort] = ACTIONS(2950),
    [anon_sym_goto] = ACTIONS(2950),
    [anon_sym_return] = ACTIONS(2950),
    [anon_sym_undef] = ACTIONS(2950),
    [anon_sym_PERCENT2] = ACTIONS(2950),
    [anon_sym_AMP2] = ACTIONS(2950),
    [anon_sym_STAR2] = ACTIONS(2950),
    [anon_sym_use] = ACTIONS(2950),
    [anon_sym_no] = ACTIONS(2950),
    [anon_sym_for] = ACTIONS(2950),
    [anon_sym_foreach] = ACTIONS(2950),
    [anon_sym_last] = ACTIONS(2950),
    [anon_sym_next] = ACTIONS(2950),
    [anon_sym_redo] = ACTIONS(2950),
    [anon_sym_BEGIN] = ACTIONS(2950),
    [anon_sym_INIT] = ACTIONS(2950),
    [anon_sym_CHECK] = ACTIONS(2950),
    [anon_sym_UNITCHECK] = ACTIONS(2950),
    [anon_sym_END] = ACTIONS(2950),
    [anon_sym___FILE__] = ACTIONS(2950),
    [anon_sym___LINE__] = ACTIONS(2950),
    [anon_sym___PACKAGE__] = ACTIONS(2950),
    [anon_sym___SUB__] = ACTIONS(2950),
    [anon_sym_break] = ACTIONS(2950),
    [anon_sym_fork] = ACTIONS(2950),
    [anon_sym_getppid] = ACTIONS(2950),
    [anon_sym_time] = ACTIONS(2950),
    [anon_sym_times] = ACTIONS(2950),
    [anon_sym_wait] = ACTIONS(2950),
    [anon_sym_wantarray] = ACTIONS(2950),
    [anon_sym_abs] = ACTIONS(2950),
    [anon_sym_alarm] = ACTIONS(2950),
    [anon_sym_chop] = ACTIONS(2950),
    [anon_sym_chdir] = ACTIONS(2950),
    [anon_sym_close] = ACTIONS(2950),
    [anon_sym_closedir] = ACTIONS(2950),
    [anon_sym_caller] = ACTIONS(2950),
    [anon_sym_chomp] = ACTIONS(2950),
    [anon_sym_chr] = ACTIONS(2950),
    [anon_sym_cos] = ACTIONS(2950),
    [anon_sym_chroot] = ACTIONS(2950),
    [anon_sym_defined] = ACTIONS(2950),
    [anon_sym_delete] = ACTIONS(2950),
    [anon_sym_dbmclose] = ACTIONS(2950),
    [anon_sym_exists] = ACTIONS(2950),
    [anon_sym_exit] = ACTIONS(2950),
    [anon_sym_eof] = ACTIONS(2950),
    [anon_sym_exp] = ACTIONS(2950),
    [anon_sym_each] = ACTIONS(2950),
    [anon_sym_fc] = ACTIONS(2950),
    [anon_sym_fileno] = ACTIONS(2950),
    [anon_sym_gmtime] = ACTIONS(2950),
    [anon_sym_getc] = ACTIONS(2950),
    [anon_sym_getpgrp] = ACTIONS(2950),
    [anon_sym_getprotobyname] = ACTIONS(2950),
    [anon_sym_getpwname] = ACTIONS(2950),
    [anon_sym_getpwuid] = ACTIONS(2950),
    [anon_sym_getpeername] = ACTIONS(2950),
    [anon_sym_getnetbyname] = ACTIONS(2950),
    [anon_sym_getsockname] = ACTIONS(2950),
    [anon_sym_getgrnam] = ACTIONS(2950),
    [anon_sym_getgrgid] = ACTIONS(2950),
    [anon_sym_hex] = ACTIONS(2950),
    [anon_sym_int] = ACTIONS(2950),
    [anon_sym_keys] = ACTIONS(2950),
    [anon_sym_lc] = ACTIONS(2950),
    [anon_sym_lcfirst] = ACTIONS(2950),
    [anon_sym_length] = ACTIONS(2950),
    [anon_sym_localtime] = ACTIONS(2950),
    [anon_sym_log] = ACTIONS(2950),
    [anon_sym_lock] = ACTIONS(2950),
    [anon_sym_lstat] = ACTIONS(2950),
    [anon_sym_oct] = ACTIONS(2950),
    [anon_sym_ord] = ACTIONS(2950),
    [anon_sym_prototype] = ACTIONS(2950),
    [anon_sym_pop] = ACTIONS(2950),
    [anon_sym_pos] = ACTIONS(2950),
    [anon_sym_quotemeta] = ACTIONS(2950),
    [anon_sym_reset] = ACTIONS(2950),
    [anon_sym_rand] = ACTIONS(2950),
    [anon_sym_rmdir] = ACTIONS(2950),
    [anon_sym_readdir] = ACTIONS(2950),
    [anon_sym_readline] = ACTIONS(2950),
    [anon_sym_readpipe] = ACTIONS(2950),
    [anon_sym_rewinddir] = ACTIONS(2950),
    [anon_sym_readlink] = ACTIONS(2950),
    [anon_sym_ref] = ACTIONS(2950),
    [anon_sym_scalar] = ACTIONS(2950),
    [anon_sym_shift] = ACTIONS(2950),
    [anon_sym_sin] = ACTIONS(2950),
    [anon_sym_sleep] = ACTIONS(2950),
    [anon_sym_sqrt] = ACTIONS(2950),
    [anon_sym_srand] = ACTIONS(2950),
    [anon_sym_stat] = ACTIONS(2950),
    [anon_sym_study] = ACTIONS(2950),
    [anon_sym_tell] = ACTIONS(2950),
    [anon_sym_telldir] = ACTIONS(2950),
    [anon_sym_tied] = ACTIONS(2950),
    [anon_sym_uc] = ACTIONS(2950),
    [anon_sym_ucfirst] = ACTIONS(2950),
    [anon_sym_untie] = ACTIONS(2950),
    [anon_sym_umask] = ACTIONS(2950),
    [anon_sym_values] = ACTIONS(2950),
    [anon_sym_write] = ACTIONS(2950),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2950),
    [anon_sym___END__] = ACTIONS(2950),
    [anon_sym_EOT] = ACTIONS(2950),
    [anon_sym_q] = ACTIONS(2950),
    [anon_sym_qq] = ACTIONS(2950),
    [anon_sym_qw] = ACTIONS(2950),
    [anon_sym_qx] = ACTIONS(2950),
    [anon_sym_qr] = ACTIONS(2950),
    [anon_sym_m] = ACTIONS(2950),
    [anon_sym_s] = ACTIONS(2950),
    [anon_sym_tr] = ACTIONS(2950),
    [anon_sym_y] = ACTIONS(2950),
    [anon_sym_if] = ACTIONS(2950),
    [anon_sym_unless] = ACTIONS(2950),
    [anon_sym_while] = ACTIONS(2950),
    [anon_sym_until] = ACTIONS(2950),
    [aux_sym__bareword_token1] = ACTIONS(2950),
    [sym_number] = ACTIONS(2950),
    [anon_sym_true] = ACTIONS(2950),
    [anon_sym_false] = ACTIONS(2950),
    [sym__single_quote] = ACTIONS(2948),
    [sym__double_quote] = ACTIONS(2948),
    [sym__backtick_quote] = ACTIONS(2948),
    [sym__search_slash_quote] = ACTIONS(2948),
    [sym__open_readline_bracket] = ACTIONS(2948),
    [sym__open_fileglob_bracket] = ACTIONS(2948),
    [sym__PERLY_HEREDOC] = ACTIONS(2948),
    [sym__ctrl_z_hack] = ACTIONS(2948),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2948),
    [sym__filetest] = ACTIONS(2948),
  },
  [STATE(1450)] = {
    [sym_heredoc_content] = STATE(1450),
    [ts_builtin_sym_end] = ACTIONS(2952),
    [sym__identifier] = ACTIONS(2954),
    [anon_sym_LBRACE] = ACTIONS(2954),
    [anon_sym_SEMI] = ACTIONS(2954),
    [anon_sym_package] = ACTIONS(2954),
    [anon_sym_class] = ACTIONS(2954),
    [anon_sym_role] = ACTIONS(2954),
    [anon_sym_BUILD] = ACTIONS(2954),
    [anon_sym_ADJUST] = ACTIONS(2954),
    [anon_sym_DOLLAR] = ACTIONS(2954),
    [anon_sym_AT] = ACTIONS(2954),
    [anon_sym_my] = ACTIONS(2954),
    [anon_sym_extended] = ACTIONS(2954),
    [anon_sym_async] = ACTIONS(2954),
    [anon_sym_sub] = ACTIONS(2954),
    [anon_sym_method] = ACTIONS(2954),
    [anon_sym_LPAREN] = ACTIONS(2954),
    [anon_sym_continue] = ACTIONS(2954),
    [anon_sym_state] = ACTIONS(2954),
    [anon_sym_our] = ACTIONS(2954),
    [anon_sym_try] = ACTIONS(2954),
    [anon_sym_defer] = ACTIONS(2954),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2954),
    [anon_sym_LBRACK] = ACTIONS(2954),
    [anon_sym_LT_LT] = ACTIONS(2954),
    [anon_sym_SLASH_SLASH] = ACTIONS(2954),
    [anon_sym_PLUS] = ACTIONS(2954),
    [anon_sym_DASH] = ACTIONS(2954),
    [anon_sym_TILDE] = ACTIONS(2954),
    [anon_sym_BANG] = ACTIONS(2954),
    [anon_sym_PLUS_PLUS] = ACTIONS(2954),
    [anon_sym_DASH_DASH] = ACTIONS(2954),
    [anon_sym_BSLASH] = ACTIONS(2952),
    [anon_sym_do] = ACTIONS(2954),
    [anon_sym_eval] = ACTIONS(2954),
    [anon_sym_await] = ACTIONS(2954),
    [anon_sym_field] = ACTIONS(2954),
    [anon_sym_local] = ACTIONS(2954),
    [anon_sym_dynamically] = ACTIONS(2954),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2954),
    [anon_sym_require] = ACTIONS(2954),
    [anon_sym_map] = ACTIONS(2954),
    [anon_sym_grep] = ACTIONS(2954),
    [anon_sym_sort] = ACTIONS(2954),
    [anon_sym_goto] = ACTIONS(2954),
    [anon_sym_return] = ACTIONS(2954),
    [anon_sym_undef] = ACTIONS(2954),
    [anon_sym_PERCENT2] = ACTIONS(2954),
    [anon_sym_AMP2] = ACTIONS(2954),
    [anon_sym_STAR2] = ACTIONS(2954),
    [anon_sym_use] = ACTIONS(2954),
    [anon_sym_no] = ACTIONS(2954),
    [anon_sym_for] = ACTIONS(2954),
    [anon_sym_foreach] = ACTIONS(2954),
    [anon_sym_last] = ACTIONS(2954),
    [anon_sym_next] = ACTIONS(2954),
    [anon_sym_redo] = ACTIONS(2954),
    [anon_sym_BEGIN] = ACTIONS(2954),
    [anon_sym_INIT] = ACTIONS(2954),
    [anon_sym_CHECK] = ACTIONS(2954),
    [anon_sym_UNITCHECK] = ACTIONS(2954),
    [anon_sym_END] = ACTIONS(2954),
    [anon_sym___FILE__] = ACTIONS(2954),
    [anon_sym___LINE__] = ACTIONS(2954),
    [anon_sym___PACKAGE__] = ACTIONS(2954),
    [anon_sym___SUB__] = ACTIONS(2954),
    [anon_sym_break] = ACTIONS(2954),
    [anon_sym_fork] = ACTIONS(2954),
    [anon_sym_getppid] = ACTIONS(2954),
    [anon_sym_time] = ACTIONS(2954),
    [anon_sym_times] = ACTIONS(2954),
    [anon_sym_wait] = ACTIONS(2954),
    [anon_sym_wantarray] = ACTIONS(2954),
    [anon_sym_abs] = ACTIONS(2954),
    [anon_sym_alarm] = ACTIONS(2954),
    [anon_sym_chop] = ACTIONS(2954),
    [anon_sym_chdir] = ACTIONS(2954),
    [anon_sym_close] = ACTIONS(2954),
    [anon_sym_closedir] = ACTIONS(2954),
    [anon_sym_caller] = ACTIONS(2954),
    [anon_sym_chomp] = ACTIONS(2954),
    [anon_sym_chr] = ACTIONS(2954),
    [anon_sym_cos] = ACTIONS(2954),
    [anon_sym_chroot] = ACTIONS(2954),
    [anon_sym_defined] = ACTIONS(2954),
    [anon_sym_delete] = ACTIONS(2954),
    [anon_sym_dbmclose] = ACTIONS(2954),
    [anon_sym_exists] = ACTIONS(2954),
    [anon_sym_exit] = ACTIONS(2954),
    [anon_sym_eof] = ACTIONS(2954),
    [anon_sym_exp] = ACTIONS(2954),
    [anon_sym_each] = ACTIONS(2954),
    [anon_sym_fc] = ACTIONS(2954),
    [anon_sym_fileno] = ACTIONS(2954),
    [anon_sym_gmtime] = ACTIONS(2954),
    [anon_sym_getc] = ACTIONS(2954),
    [anon_sym_getpgrp] = ACTIONS(2954),
    [anon_sym_getprotobyname] = ACTIONS(2954),
    [anon_sym_getpwname] = ACTIONS(2954),
    [anon_sym_getpwuid] = ACTIONS(2954),
    [anon_sym_getpeername] = ACTIONS(2954),
    [anon_sym_getnetbyname] = ACTIONS(2954),
    [anon_sym_getsockname] = ACTIONS(2954),
    [anon_sym_getgrnam] = ACTIONS(2954),
    [anon_sym_getgrgid] = ACTIONS(2954),
    [anon_sym_hex] = ACTIONS(2954),
    [anon_sym_int] = ACTIONS(2954),
    [anon_sym_keys] = ACTIONS(2954),
    [anon_sym_lc] = ACTIONS(2954),
    [anon_sym_lcfirst] = ACTIONS(2954),
    [anon_sym_length] = ACTIONS(2954),
    [anon_sym_localtime] = ACTIONS(2954),
    [anon_sym_log] = ACTIONS(2954),
    [anon_sym_lock] = ACTIONS(2954),
    [anon_sym_lstat] = ACTIONS(2954),
    [anon_sym_oct] = ACTIONS(2954),
    [anon_sym_ord] = ACTIONS(2954),
    [anon_sym_prototype] = ACTIONS(2954),
    [anon_sym_pop] = ACTIONS(2954),
    [anon_sym_pos] = ACTIONS(2954),
    [anon_sym_quotemeta] = ACTIONS(2954),
    [anon_sym_reset] = ACTIONS(2954),
    [anon_sym_rand] = ACTIONS(2954),
    [anon_sym_rmdir] = ACTIONS(2954),
    [anon_sym_readdir] = ACTIONS(2954),
    [anon_sym_readline] = ACTIONS(2954),
    [anon_sym_readpipe] = ACTIONS(2954),
    [anon_sym_rewinddir] = ACTIONS(2954),
    [anon_sym_readlink] = ACTIONS(2954),
    [anon_sym_ref] = ACTIONS(2954),
    [anon_sym_scalar] = ACTIONS(2954),
    [anon_sym_shift] = ACTIONS(2954),
    [anon_sym_sin] = ACTIONS(2954),
    [anon_sym_sleep] = ACTIONS(2954),
    [anon_sym_sqrt] = ACTIONS(2954),
    [anon_sym_srand] = ACTIONS(2954),
    [anon_sym_stat] = ACTIONS(2954),
    [anon_sym_study] = ACTIONS(2954),
    [anon_sym_tell] = ACTIONS(2954),
    [anon_sym_telldir] = ACTIONS(2954),
    [anon_sym_tied] = ACTIONS(2954),
    [anon_sym_uc] = ACTIONS(2954),
    [anon_sym_ucfirst] = ACTIONS(2954),
    [anon_sym_untie] = ACTIONS(2954),
    [anon_sym_umask] = ACTIONS(2954),
    [anon_sym_values] = ACTIONS(2954),
    [anon_sym_write] = ACTIONS(2954),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2954),
    [anon_sym___END__] = ACTIONS(2954),
    [anon_sym_EOT] = ACTIONS(2954),
    [anon_sym_q] = ACTIONS(2954),
    [anon_sym_qq] = ACTIONS(2954),
    [anon_sym_qw] = ACTIONS(2954),
    [anon_sym_qx] = ACTIONS(2954),
    [anon_sym_qr] = ACTIONS(2954),
    [anon_sym_m] = ACTIONS(2954),
    [anon_sym_s] = ACTIONS(2954),
    [anon_sym_tr] = ACTIONS(2954),
    [anon_sym_y] = ACTIONS(2954),
    [anon_sym_if] = ACTIONS(2954),
    [anon_sym_unless] = ACTIONS(2954),
    [anon_sym_while] = ACTIONS(2954),
    [anon_sym_until] = ACTIONS(2954),
    [aux_sym__bareword_token1] = ACTIONS(2954),
    [sym_number] = ACTIONS(2954),
    [anon_sym_true] = ACTIONS(2954),
    [anon_sym_false] = ACTIONS(2954),
    [sym__single_quote] = ACTIONS(2952),
    [sym__double_quote] = ACTIONS(2952),
    [sym__backtick_quote] = ACTIONS(2952),
    [sym__search_slash_quote] = ACTIONS(2952),
    [sym__open_readline_bracket] = ACTIONS(2952),
    [sym__open_fileglob_bracket] = ACTIONS(2952),
    [sym__PERLY_HEREDOC] = ACTIONS(2952),
    [sym__ctrl_z_hack] = ACTIONS(2952),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2952),
    [sym__filetest] = ACTIONS(2952),
  },
  [STATE(1451)] = {
    [sym_heredoc_content] = STATE(1451),
    [ts_builtin_sym_end] = ACTIONS(2956),
    [sym__identifier] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2958),
    [anon_sym_SEMI] = ACTIONS(2958),
    [anon_sym_package] = ACTIONS(2958),
    [anon_sym_class] = ACTIONS(2958),
    [anon_sym_role] = ACTIONS(2958),
    [anon_sym_BUILD] = ACTIONS(2958),
    [anon_sym_ADJUST] = ACTIONS(2958),
    [anon_sym_DOLLAR] = ACTIONS(2958),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_my] = ACTIONS(2958),
    [anon_sym_extended] = ACTIONS(2958),
    [anon_sym_async] = ACTIONS(2958),
    [anon_sym_sub] = ACTIONS(2958),
    [anon_sym_method] = ACTIONS(2958),
    [anon_sym_LPAREN] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_state] = ACTIONS(2958),
    [anon_sym_our] = ACTIONS(2958),
    [anon_sym_try] = ACTIONS(2958),
    [anon_sym_defer] = ACTIONS(2958),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2958),
    [anon_sym_LT_LT] = ACTIONS(2958),
    [anon_sym_SLASH_SLASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_TILDE] = ACTIONS(2958),
    [anon_sym_BANG] = ACTIONS(2958),
    [anon_sym_PLUS_PLUS] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2958),
    [anon_sym_BSLASH] = ACTIONS(2956),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_eval] = ACTIONS(2958),
    [anon_sym_await] = ACTIONS(2958),
    [anon_sym_field] = ACTIONS(2958),
    [anon_sym_local] = ACTIONS(2958),
    [anon_sym_dynamically] = ACTIONS(2958),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2958),
    [anon_sym_require] = ACTIONS(2958),
    [anon_sym_map] = ACTIONS(2958),
    [anon_sym_grep] = ACTIONS(2958),
    [anon_sym_sort] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_undef] = ACTIONS(2958),
    [anon_sym_PERCENT2] = ACTIONS(2958),
    [anon_sym_AMP2] = ACTIONS(2958),
    [anon_sym_STAR2] = ACTIONS(2958),
    [anon_sym_use] = ACTIONS(2958),
    [anon_sym_no] = ACTIONS(2958),
    [anon_sym_for] = ACTIONS(2958),
    [anon_sym_foreach] = ACTIONS(2958),
    [anon_sym_last] = ACTIONS(2958),
    [anon_sym_next] = ACTIONS(2958),
    [anon_sym_redo] = ACTIONS(2958),
    [anon_sym_BEGIN] = ACTIONS(2958),
    [anon_sym_INIT] = ACTIONS(2958),
    [anon_sym_CHECK] = ACTIONS(2958),
    [anon_sym_UNITCHECK] = ACTIONS(2958),
    [anon_sym_END] = ACTIONS(2958),
    [anon_sym___FILE__] = ACTIONS(2958),
    [anon_sym___LINE__] = ACTIONS(2958),
    [anon_sym___PACKAGE__] = ACTIONS(2958),
    [anon_sym___SUB__] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_fork] = ACTIONS(2958),
    [anon_sym_getppid] = ACTIONS(2958),
    [anon_sym_time] = ACTIONS(2958),
    [anon_sym_times] = ACTIONS(2958),
    [anon_sym_wait] = ACTIONS(2958),
    [anon_sym_wantarray] = ACTIONS(2958),
    [anon_sym_abs] = ACTIONS(2958),
    [anon_sym_alarm] = ACTIONS(2958),
    [anon_sym_chop] = ACTIONS(2958),
    [anon_sym_chdir] = ACTIONS(2958),
    [anon_sym_close] = ACTIONS(2958),
    [anon_sym_closedir] = ACTIONS(2958),
    [anon_sym_caller] = ACTIONS(2958),
    [anon_sym_chomp] = ACTIONS(2958),
    [anon_sym_chr] = ACTIONS(2958),
    [anon_sym_cos] = ACTIONS(2958),
    [anon_sym_chroot] = ACTIONS(2958),
    [anon_sym_defined] = ACTIONS(2958),
    [anon_sym_delete] = ACTIONS(2958),
    [anon_sym_dbmclose] = ACTIONS(2958),
    [anon_sym_exists] = ACTIONS(2958),
    [anon_sym_exit] = ACTIONS(2958),
    [anon_sym_eof] = ACTIONS(2958),
    [anon_sym_exp] = ACTIONS(2958),
    [anon_sym_each] = ACTIONS(2958),
    [anon_sym_fc] = ACTIONS(2958),
    [anon_sym_fileno] = ACTIONS(2958),
    [anon_sym_gmtime] = ACTIONS(2958),
    [anon_sym_getc] = ACTIONS(2958),
    [anon_sym_getpgrp] = ACTIONS(2958),
    [anon_sym_getprotobyname] = ACTIONS(2958),
    [anon_sym_getpwname] = ACTIONS(2958),
    [anon_sym_getpwuid] = ACTIONS(2958),
    [anon_sym_getpeername] = ACTIONS(2958),
    [anon_sym_getnetbyname] = ACTIONS(2958),
    [anon_sym_getsockname] = ACTIONS(2958),
    [anon_sym_getgrnam] = ACTIONS(2958),
    [anon_sym_getgrgid] = ACTIONS(2958),
    [anon_sym_hex] = ACTIONS(2958),
    [anon_sym_int] = ACTIONS(2958),
    [anon_sym_keys] = ACTIONS(2958),
    [anon_sym_lc] = ACTIONS(2958),
    [anon_sym_lcfirst] = ACTIONS(2958),
    [anon_sym_length] = ACTIONS(2958),
    [anon_sym_localtime] = ACTIONS(2958),
    [anon_sym_log] = ACTIONS(2958),
    [anon_sym_lock] = ACTIONS(2958),
    [anon_sym_lstat] = ACTIONS(2958),
    [anon_sym_oct] = ACTIONS(2958),
    [anon_sym_ord] = ACTIONS(2958),
    [anon_sym_prototype] = ACTIONS(2958),
    [anon_sym_pop] = ACTIONS(2958),
    [anon_sym_pos] = ACTIONS(2958),
    [anon_sym_quotemeta] = ACTIONS(2958),
    [anon_sym_reset] = ACTIONS(2958),
    [anon_sym_rand] = ACTIONS(2958),
    [anon_sym_rmdir] = ACTIONS(2958),
    [anon_sym_readdir] = ACTIONS(2958),
    [anon_sym_readline] = ACTIONS(2958),
    [anon_sym_readpipe] = ACTIONS(2958),
    [anon_sym_rewinddir] = ACTIONS(2958),
    [anon_sym_readlink] = ACTIONS(2958),
    [anon_sym_ref] = ACTIONS(2958),
    [anon_sym_scalar] = ACTIONS(2958),
    [anon_sym_shift] = ACTIONS(2958),
    [anon_sym_sin] = ACTIONS(2958),
    [anon_sym_sleep] = ACTIONS(2958),
    [anon_sym_sqrt] = ACTIONS(2958),
    [anon_sym_srand] = ACTIONS(2958),
    [anon_sym_stat] = ACTIONS(2958),
    [anon_sym_study] = ACTIONS(2958),
    [anon_sym_tell] = ACTIONS(2958),
    [anon_sym_telldir] = ACTIONS(2958),
    [anon_sym_tied] = ACTIONS(2958),
    [anon_sym_uc] = ACTIONS(2958),
    [anon_sym_ucfirst] = ACTIONS(2958),
    [anon_sym_untie] = ACTIONS(2958),
    [anon_sym_umask] = ACTIONS(2958),
    [anon_sym_values] = ACTIONS(2958),
    [anon_sym_write] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2958),
    [anon_sym___END__] = ACTIONS(2958),
    [anon_sym_EOT] = ACTIONS(2958),
    [anon_sym_q] = ACTIONS(2958),
    [anon_sym_qq] = ACTIONS(2958),
    [anon_sym_qw] = ACTIONS(2958),
    [anon_sym_qx] = ACTIONS(2958),
    [anon_sym_qr] = ACTIONS(2958),
    [anon_sym_m] = ACTIONS(2958),
    [anon_sym_s] = ACTIONS(2958),
    [anon_sym_tr] = ACTIONS(2958),
    [anon_sym_y] = ACTIONS(2958),
    [anon_sym_if] = ACTIONS(2958),
    [anon_sym_unless] = ACTIONS(2958),
    [anon_sym_while] = ACTIONS(2958),
    [anon_sym_until] = ACTIONS(2958),
    [aux_sym__bareword_token1] = ACTIONS(2958),
    [sym_number] = ACTIONS(2958),
    [anon_sym_true] = ACTIONS(2958),
    [anon_sym_false] = ACTIONS(2958),
    [sym__single_quote] = ACTIONS(2956),
    [sym__double_quote] = ACTIONS(2956),
    [sym__backtick_quote] = ACTIONS(2956),
    [sym__search_slash_quote] = ACTIONS(2956),
    [sym__open_readline_bracket] = ACTIONS(2956),
    [sym__open_fileglob_bracket] = ACTIONS(2956),
    [sym__PERLY_HEREDOC] = ACTIONS(2956),
    [sym__ctrl_z_hack] = ACTIONS(2956),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2956),
    [sym__filetest] = ACTIONS(2956),
  },
  [STATE(1452)] = {
    [sym_heredoc_content] = STATE(1452),
    [ts_builtin_sym_end] = ACTIONS(2960),
    [sym__identifier] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2962),
    [anon_sym_SEMI] = ACTIONS(2962),
    [anon_sym_package] = ACTIONS(2962),
    [anon_sym_class] = ACTIONS(2962),
    [anon_sym_role] = ACTIONS(2962),
    [anon_sym_BUILD] = ACTIONS(2962),
    [anon_sym_ADJUST] = ACTIONS(2962),
    [anon_sym_DOLLAR] = ACTIONS(2962),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_my] = ACTIONS(2962),
    [anon_sym_extended] = ACTIONS(2962),
    [anon_sym_async] = ACTIONS(2962),
    [anon_sym_sub] = ACTIONS(2962),
    [anon_sym_method] = ACTIONS(2962),
    [anon_sym_LPAREN] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_state] = ACTIONS(2962),
    [anon_sym_our] = ACTIONS(2962),
    [anon_sym_try] = ACTIONS(2962),
    [anon_sym_defer] = ACTIONS(2962),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2962),
    [anon_sym_LT_LT] = ACTIONS(2962),
    [anon_sym_SLASH_SLASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_TILDE] = ACTIONS(2962),
    [anon_sym_BANG] = ACTIONS(2962),
    [anon_sym_PLUS_PLUS] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2962),
    [anon_sym_BSLASH] = ACTIONS(2960),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_eval] = ACTIONS(2962),
    [anon_sym_await] = ACTIONS(2962),
    [anon_sym_field] = ACTIONS(2962),
    [anon_sym_local] = ACTIONS(2962),
    [anon_sym_dynamically] = ACTIONS(2962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2962),
    [anon_sym_require] = ACTIONS(2962),
    [anon_sym_map] = ACTIONS(2962),
    [anon_sym_grep] = ACTIONS(2962),
    [anon_sym_sort] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_undef] = ACTIONS(2962),
    [anon_sym_PERCENT2] = ACTIONS(2962),
    [anon_sym_AMP2] = ACTIONS(2962),
    [anon_sym_STAR2] = ACTIONS(2962),
    [anon_sym_use] = ACTIONS(2962),
    [anon_sym_no] = ACTIONS(2962),
    [anon_sym_for] = ACTIONS(2962),
    [anon_sym_foreach] = ACTIONS(2962),
    [anon_sym_last] = ACTIONS(2962),
    [anon_sym_next] = ACTIONS(2962),
    [anon_sym_redo] = ACTIONS(2962),
    [anon_sym_BEGIN] = ACTIONS(2962),
    [anon_sym_INIT] = ACTIONS(2962),
    [anon_sym_CHECK] = ACTIONS(2962),
    [anon_sym_UNITCHECK] = ACTIONS(2962),
    [anon_sym_END] = ACTIONS(2962),
    [anon_sym___FILE__] = ACTIONS(2962),
    [anon_sym___LINE__] = ACTIONS(2962),
    [anon_sym___PACKAGE__] = ACTIONS(2962),
    [anon_sym___SUB__] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_fork] = ACTIONS(2962),
    [anon_sym_getppid] = ACTIONS(2962),
    [anon_sym_time] = ACTIONS(2962),
    [anon_sym_times] = ACTIONS(2962),
    [anon_sym_wait] = ACTIONS(2962),
    [anon_sym_wantarray] = ACTIONS(2962),
    [anon_sym_abs] = ACTIONS(2962),
    [anon_sym_alarm] = ACTIONS(2962),
    [anon_sym_chop] = ACTIONS(2962),
    [anon_sym_chdir] = ACTIONS(2962),
    [anon_sym_close] = ACTIONS(2962),
    [anon_sym_closedir] = ACTIONS(2962),
    [anon_sym_caller] = ACTIONS(2962),
    [anon_sym_chomp] = ACTIONS(2962),
    [anon_sym_chr] = ACTIONS(2962),
    [anon_sym_cos] = ACTIONS(2962),
    [anon_sym_chroot] = ACTIONS(2962),
    [anon_sym_defined] = ACTIONS(2962),
    [anon_sym_delete] = ACTIONS(2962),
    [anon_sym_dbmclose] = ACTIONS(2962),
    [anon_sym_exists] = ACTIONS(2962),
    [anon_sym_exit] = ACTIONS(2962),
    [anon_sym_eof] = ACTIONS(2962),
    [anon_sym_exp] = ACTIONS(2962),
    [anon_sym_each] = ACTIONS(2962),
    [anon_sym_fc] = ACTIONS(2962),
    [anon_sym_fileno] = ACTIONS(2962),
    [anon_sym_gmtime] = ACTIONS(2962),
    [anon_sym_getc] = ACTIONS(2962),
    [anon_sym_getpgrp] = ACTIONS(2962),
    [anon_sym_getprotobyname] = ACTIONS(2962),
    [anon_sym_getpwname] = ACTIONS(2962),
    [anon_sym_getpwuid] = ACTIONS(2962),
    [anon_sym_getpeername] = ACTIONS(2962),
    [anon_sym_getnetbyname] = ACTIONS(2962),
    [anon_sym_getsockname] = ACTIONS(2962),
    [anon_sym_getgrnam] = ACTIONS(2962),
    [anon_sym_getgrgid] = ACTIONS(2962),
    [anon_sym_hex] = ACTIONS(2962),
    [anon_sym_int] = ACTIONS(2962),
    [anon_sym_keys] = ACTIONS(2962),
    [anon_sym_lc] = ACTIONS(2962),
    [anon_sym_lcfirst] = ACTIONS(2962),
    [anon_sym_length] = ACTIONS(2962),
    [anon_sym_localtime] = ACTIONS(2962),
    [anon_sym_log] = ACTIONS(2962),
    [anon_sym_lock] = ACTIONS(2962),
    [anon_sym_lstat] = ACTIONS(2962),
    [anon_sym_oct] = ACTIONS(2962),
    [anon_sym_ord] = ACTIONS(2962),
    [anon_sym_prototype] = ACTIONS(2962),
    [anon_sym_pop] = ACTIONS(2962),
    [anon_sym_pos] = ACTIONS(2962),
    [anon_sym_quotemeta] = ACTIONS(2962),
    [anon_sym_reset] = ACTIONS(2962),
    [anon_sym_rand] = ACTIONS(2962),
    [anon_sym_rmdir] = ACTIONS(2962),
    [anon_sym_readdir] = ACTIONS(2962),
    [anon_sym_readline] = ACTIONS(2962),
    [anon_sym_readpipe] = ACTIONS(2962),
    [anon_sym_rewinddir] = ACTIONS(2962),
    [anon_sym_readlink] = ACTIONS(2962),
    [anon_sym_ref] = ACTIONS(2962),
    [anon_sym_scalar] = ACTIONS(2962),
    [anon_sym_shift] = ACTIONS(2962),
    [anon_sym_sin] = ACTIONS(2962),
    [anon_sym_sleep] = ACTIONS(2962),
    [anon_sym_sqrt] = ACTIONS(2962),
    [anon_sym_srand] = ACTIONS(2962),
    [anon_sym_stat] = ACTIONS(2962),
    [anon_sym_study] = ACTIONS(2962),
    [anon_sym_tell] = ACTIONS(2962),
    [anon_sym_telldir] = ACTIONS(2962),
    [anon_sym_tied] = ACTIONS(2962),
    [anon_sym_uc] = ACTIONS(2962),
    [anon_sym_ucfirst] = ACTIONS(2962),
    [anon_sym_untie] = ACTIONS(2962),
    [anon_sym_umask] = ACTIONS(2962),
    [anon_sym_values] = ACTIONS(2962),
    [anon_sym_write] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2962),
    [anon_sym___END__] = ACTIONS(2962),
    [anon_sym_EOT] = ACTIONS(2962),
    [anon_sym_q] = ACTIONS(2962),
    [anon_sym_qq] = ACTIONS(2962),
    [anon_sym_qw] = ACTIONS(2962),
    [anon_sym_qx] = ACTIONS(2962),
    [anon_sym_qr] = ACTIONS(2962),
    [anon_sym_m] = ACTIONS(2962),
    [anon_sym_s] = ACTIONS(2962),
    [anon_sym_tr] = ACTIONS(2962),
    [anon_sym_y] = ACTIONS(2962),
    [anon_sym_if] = ACTIONS(2962),
    [anon_sym_unless] = ACTIONS(2962),
    [anon_sym_while] = ACTIONS(2962),
    [anon_sym_until] = ACTIONS(2962),
    [aux_sym__bareword_token1] = ACTIONS(2962),
    [sym_number] = ACTIONS(2962),
    [anon_sym_true] = ACTIONS(2962),
    [anon_sym_false] = ACTIONS(2962),
    [sym__single_quote] = ACTIONS(2960),
    [sym__double_quote] = ACTIONS(2960),
    [sym__backtick_quote] = ACTIONS(2960),
    [sym__search_slash_quote] = ACTIONS(2960),
    [sym__open_readline_bracket] = ACTIONS(2960),
    [sym__open_fileglob_bracket] = ACTIONS(2960),
    [sym__PERLY_HEREDOC] = ACTIONS(2960),
    [sym__ctrl_z_hack] = ACTIONS(2960),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2960),
    [sym__filetest] = ACTIONS(2960),
  },
  [STATE(1453)] = {
    [sym_heredoc_content] = STATE(1453),
    [ts_builtin_sym_end] = ACTIONS(2964),
    [sym__identifier] = ACTIONS(2966),
    [anon_sym_LBRACE] = ACTIONS(2966),
    [anon_sym_SEMI] = ACTIONS(2966),
    [anon_sym_package] = ACTIONS(2966),
    [anon_sym_class] = ACTIONS(2966),
    [anon_sym_role] = ACTIONS(2966),
    [anon_sym_BUILD] = ACTIONS(2966),
    [anon_sym_ADJUST] = ACTIONS(2966),
    [anon_sym_DOLLAR] = ACTIONS(2966),
    [anon_sym_AT] = ACTIONS(2966),
    [anon_sym_my] = ACTIONS(2966),
    [anon_sym_extended] = ACTIONS(2966),
    [anon_sym_async] = ACTIONS(2966),
    [anon_sym_sub] = ACTIONS(2966),
    [anon_sym_method] = ACTIONS(2966),
    [anon_sym_LPAREN] = ACTIONS(2966),
    [anon_sym_continue] = ACTIONS(2966),
    [anon_sym_state] = ACTIONS(2966),
    [anon_sym_our] = ACTIONS(2966),
    [anon_sym_try] = ACTIONS(2966),
    [anon_sym_defer] = ACTIONS(2966),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2966),
    [anon_sym_LBRACK] = ACTIONS(2966),
    [anon_sym_LT_LT] = ACTIONS(2966),
    [anon_sym_SLASH_SLASH] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(2966),
    [anon_sym_DASH] = ACTIONS(2966),
    [anon_sym_TILDE] = ACTIONS(2966),
    [anon_sym_BANG] = ACTIONS(2966),
    [anon_sym_PLUS_PLUS] = ACTIONS(2966),
    [anon_sym_DASH_DASH] = ACTIONS(2966),
    [anon_sym_BSLASH] = ACTIONS(2964),
    [anon_sym_do] = ACTIONS(2966),
    [anon_sym_eval] = ACTIONS(2966),
    [anon_sym_await] = ACTIONS(2966),
    [anon_sym_field] = ACTIONS(2966),
    [anon_sym_local] = ACTIONS(2966),
    [anon_sym_dynamically] = ACTIONS(2966),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2966),
    [anon_sym_require] = ACTIONS(2966),
    [anon_sym_map] = ACTIONS(2966),
    [anon_sym_grep] = ACTIONS(2966),
    [anon_sym_sort] = ACTIONS(2966),
    [anon_sym_goto] = ACTIONS(2966),
    [anon_sym_return] = ACTIONS(2966),
    [anon_sym_undef] = ACTIONS(2966),
    [anon_sym_PERCENT2] = ACTIONS(2966),
    [anon_sym_AMP2] = ACTIONS(2966),
    [anon_sym_STAR2] = ACTIONS(2966),
    [anon_sym_use] = ACTIONS(2966),
    [anon_sym_no] = ACTIONS(2966),
    [anon_sym_for] = ACTIONS(2966),
    [anon_sym_foreach] = ACTIONS(2966),
    [anon_sym_last] = ACTIONS(2966),
    [anon_sym_next] = ACTIONS(2966),
    [anon_sym_redo] = ACTIONS(2966),
    [anon_sym_BEGIN] = ACTIONS(2966),
    [anon_sym_INIT] = ACTIONS(2966),
    [anon_sym_CHECK] = ACTIONS(2966),
    [anon_sym_UNITCHECK] = ACTIONS(2966),
    [anon_sym_END] = ACTIONS(2966),
    [anon_sym___FILE__] = ACTIONS(2966),
    [anon_sym___LINE__] = ACTIONS(2966),
    [anon_sym___PACKAGE__] = ACTIONS(2966),
    [anon_sym___SUB__] = ACTIONS(2966),
    [anon_sym_break] = ACTIONS(2966),
    [anon_sym_fork] = ACTIONS(2966),
    [anon_sym_getppid] = ACTIONS(2966),
    [anon_sym_time] = ACTIONS(2966),
    [anon_sym_times] = ACTIONS(2966),
    [anon_sym_wait] = ACTIONS(2966),
    [anon_sym_wantarray] = ACTIONS(2966),
    [anon_sym_abs] = ACTIONS(2966),
    [anon_sym_alarm] = ACTIONS(2966),
    [anon_sym_chop] = ACTIONS(2966),
    [anon_sym_chdir] = ACTIONS(2966),
    [anon_sym_close] = ACTIONS(2966),
    [anon_sym_closedir] = ACTIONS(2966),
    [anon_sym_caller] = ACTIONS(2966),
    [anon_sym_chomp] = ACTIONS(2966),
    [anon_sym_chr] = ACTIONS(2966),
    [anon_sym_cos] = ACTIONS(2966),
    [anon_sym_chroot] = ACTIONS(2966),
    [anon_sym_defined] = ACTIONS(2966),
    [anon_sym_delete] = ACTIONS(2966),
    [anon_sym_dbmclose] = ACTIONS(2966),
    [anon_sym_exists] = ACTIONS(2966),
    [anon_sym_exit] = ACTIONS(2966),
    [anon_sym_eof] = ACTIONS(2966),
    [anon_sym_exp] = ACTIONS(2966),
    [anon_sym_each] = ACTIONS(2966),
    [anon_sym_fc] = ACTIONS(2966),
    [anon_sym_fileno] = ACTIONS(2966),
    [anon_sym_gmtime] = ACTIONS(2966),
    [anon_sym_getc] = ACTIONS(2966),
    [anon_sym_getpgrp] = ACTIONS(2966),
    [anon_sym_getprotobyname] = ACTIONS(2966),
    [anon_sym_getpwname] = ACTIONS(2966),
    [anon_sym_getpwuid] = ACTIONS(2966),
    [anon_sym_getpeername] = ACTIONS(2966),
    [anon_sym_getnetbyname] = ACTIONS(2966),
    [anon_sym_getsockname] = ACTIONS(2966),
    [anon_sym_getgrnam] = ACTIONS(2966),
    [anon_sym_getgrgid] = ACTIONS(2966),
    [anon_sym_hex] = ACTIONS(2966),
    [anon_sym_int] = ACTIONS(2966),
    [anon_sym_keys] = ACTIONS(2966),
    [anon_sym_lc] = ACTIONS(2966),
    [anon_sym_lcfirst] = ACTIONS(2966),
    [anon_sym_length] = ACTIONS(2966),
    [anon_sym_localtime] = ACTIONS(2966),
    [anon_sym_log] = ACTIONS(2966),
    [anon_sym_lock] = ACTIONS(2966),
    [anon_sym_lstat] = ACTIONS(2966),
    [anon_sym_oct] = ACTIONS(2966),
    [anon_sym_ord] = ACTIONS(2966),
    [anon_sym_prototype] = ACTIONS(2966),
    [anon_sym_pop] = ACTIONS(2966),
    [anon_sym_pos] = ACTIONS(2966),
    [anon_sym_quotemeta] = ACTIONS(2966),
    [anon_sym_reset] = ACTIONS(2966),
    [anon_sym_rand] = ACTIONS(2966),
    [anon_sym_rmdir] = ACTIONS(2966),
    [anon_sym_readdir] = ACTIONS(2966),
    [anon_sym_readline] = ACTIONS(2966),
    [anon_sym_readpipe] = ACTIONS(2966),
    [anon_sym_rewinddir] = ACTIONS(2966),
    [anon_sym_readlink] = ACTIONS(2966),
    [anon_sym_ref] = ACTIONS(2966),
    [anon_sym_scalar] = ACTIONS(2966),
    [anon_sym_shift] = ACTIONS(2966),
    [anon_sym_sin] = ACTIONS(2966),
    [anon_sym_sleep] = ACTIONS(2966),
    [anon_sym_sqrt] = ACTIONS(2966),
    [anon_sym_srand] = ACTIONS(2966),
    [anon_sym_stat] = ACTIONS(2966),
    [anon_sym_study] = ACTIONS(2966),
    [anon_sym_tell] = ACTIONS(2966),
    [anon_sym_telldir] = ACTIONS(2966),
    [anon_sym_tied] = ACTIONS(2966),
    [anon_sym_uc] = ACTIONS(2966),
    [anon_sym_ucfirst] = ACTIONS(2966),
    [anon_sym_untie] = ACTIONS(2966),
    [anon_sym_umask] = ACTIONS(2966),
    [anon_sym_values] = ACTIONS(2966),
    [anon_sym_write] = ACTIONS(2966),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2966),
    [anon_sym___END__] = ACTIONS(2966),
    [anon_sym_EOT] = ACTIONS(2966),
    [anon_sym_q] = ACTIONS(2966),
    [anon_sym_qq] = ACTIONS(2966),
    [anon_sym_qw] = ACTIONS(2966),
    [anon_sym_qx] = ACTIONS(2966),
    [anon_sym_qr] = ACTIONS(2966),
    [anon_sym_m] = ACTIONS(2966),
    [anon_sym_s] = ACTIONS(2966),
    [anon_sym_tr] = ACTIONS(2966),
    [anon_sym_y] = ACTIONS(2966),
    [anon_sym_if] = ACTIONS(2966),
    [anon_sym_unless] = ACTIONS(2966),
    [anon_sym_while] = ACTIONS(2966),
    [anon_sym_until] = ACTIONS(2966),
    [aux_sym__bareword_token1] = ACTIONS(2966),
    [sym_number] = ACTIONS(2966),
    [anon_sym_true] = ACTIONS(2966),
    [anon_sym_false] = ACTIONS(2966),
    [sym__single_quote] = ACTIONS(2964),
    [sym__double_quote] = ACTIONS(2964),
    [sym__backtick_quote] = ACTIONS(2964),
    [sym__search_slash_quote] = ACTIONS(2964),
    [sym__open_readline_bracket] = ACTIONS(2964),
    [sym__open_fileglob_bracket] = ACTIONS(2964),
    [sym__PERLY_HEREDOC] = ACTIONS(2964),
    [sym__ctrl_z_hack] = ACTIONS(2964),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2964),
    [sym__filetest] = ACTIONS(2964),
  },
  [STATE(1454)] = {
    [sym_heredoc_content] = STATE(1454),
    [ts_builtin_sym_end] = ACTIONS(2968),
    [sym__identifier] = ACTIONS(2970),
    [anon_sym_LBRACE] = ACTIONS(2970),
    [anon_sym_SEMI] = ACTIONS(2970),
    [anon_sym_package] = ACTIONS(2970),
    [anon_sym_class] = ACTIONS(2970),
    [anon_sym_role] = ACTIONS(2970),
    [anon_sym_BUILD] = ACTIONS(2970),
    [anon_sym_ADJUST] = ACTIONS(2970),
    [anon_sym_DOLLAR] = ACTIONS(2970),
    [anon_sym_AT] = ACTIONS(2970),
    [anon_sym_my] = ACTIONS(2970),
    [anon_sym_extended] = ACTIONS(2970),
    [anon_sym_async] = ACTIONS(2970),
    [anon_sym_sub] = ACTIONS(2970),
    [anon_sym_method] = ACTIONS(2970),
    [anon_sym_LPAREN] = ACTIONS(2970),
    [anon_sym_continue] = ACTIONS(2970),
    [anon_sym_state] = ACTIONS(2970),
    [anon_sym_our] = ACTIONS(2970),
    [anon_sym_try] = ACTIONS(2970),
    [anon_sym_defer] = ACTIONS(2970),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2970),
    [anon_sym_LBRACK] = ACTIONS(2970),
    [anon_sym_LT_LT] = ACTIONS(2970),
    [anon_sym_SLASH_SLASH] = ACTIONS(2970),
    [anon_sym_PLUS] = ACTIONS(2970),
    [anon_sym_DASH] = ACTIONS(2970),
    [anon_sym_TILDE] = ACTIONS(2970),
    [anon_sym_BANG] = ACTIONS(2970),
    [anon_sym_PLUS_PLUS] = ACTIONS(2970),
    [anon_sym_DASH_DASH] = ACTIONS(2970),
    [anon_sym_BSLASH] = ACTIONS(2968),
    [anon_sym_do] = ACTIONS(2970),
    [anon_sym_eval] = ACTIONS(2970),
    [anon_sym_await] = ACTIONS(2970),
    [anon_sym_field] = ACTIONS(2970),
    [anon_sym_local] = ACTIONS(2970),
    [anon_sym_dynamically] = ACTIONS(2970),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2970),
    [anon_sym_require] = ACTIONS(2970),
    [anon_sym_map] = ACTIONS(2970),
    [anon_sym_grep] = ACTIONS(2970),
    [anon_sym_sort] = ACTIONS(2970),
    [anon_sym_goto] = ACTIONS(2970),
    [anon_sym_return] = ACTIONS(2970),
    [anon_sym_undef] = ACTIONS(2970),
    [anon_sym_PERCENT2] = ACTIONS(2970),
    [anon_sym_AMP2] = ACTIONS(2970),
    [anon_sym_STAR2] = ACTIONS(2970),
    [anon_sym_use] = ACTIONS(2970),
    [anon_sym_no] = ACTIONS(2970),
    [anon_sym_for] = ACTIONS(2970),
    [anon_sym_foreach] = ACTIONS(2970),
    [anon_sym_last] = ACTIONS(2970),
    [anon_sym_next] = ACTIONS(2970),
    [anon_sym_redo] = ACTIONS(2970),
    [anon_sym_BEGIN] = ACTIONS(2970),
    [anon_sym_INIT] = ACTIONS(2970),
    [anon_sym_CHECK] = ACTIONS(2970),
    [anon_sym_UNITCHECK] = ACTIONS(2970),
    [anon_sym_END] = ACTIONS(2970),
    [anon_sym___FILE__] = ACTIONS(2970),
    [anon_sym___LINE__] = ACTIONS(2970),
    [anon_sym___PACKAGE__] = ACTIONS(2970),
    [anon_sym___SUB__] = ACTIONS(2970),
    [anon_sym_break] = ACTIONS(2970),
    [anon_sym_fork] = ACTIONS(2970),
    [anon_sym_getppid] = ACTIONS(2970),
    [anon_sym_time] = ACTIONS(2970),
    [anon_sym_times] = ACTIONS(2970),
    [anon_sym_wait] = ACTIONS(2970),
    [anon_sym_wantarray] = ACTIONS(2970),
    [anon_sym_abs] = ACTIONS(2970),
    [anon_sym_alarm] = ACTIONS(2970),
    [anon_sym_chop] = ACTIONS(2970),
    [anon_sym_chdir] = ACTIONS(2970),
    [anon_sym_close] = ACTIONS(2970),
    [anon_sym_closedir] = ACTIONS(2970),
    [anon_sym_caller] = ACTIONS(2970),
    [anon_sym_chomp] = ACTIONS(2970),
    [anon_sym_chr] = ACTIONS(2970),
    [anon_sym_cos] = ACTIONS(2970),
    [anon_sym_chroot] = ACTIONS(2970),
    [anon_sym_defined] = ACTIONS(2970),
    [anon_sym_delete] = ACTIONS(2970),
    [anon_sym_dbmclose] = ACTIONS(2970),
    [anon_sym_exists] = ACTIONS(2970),
    [anon_sym_exit] = ACTIONS(2970),
    [anon_sym_eof] = ACTIONS(2970),
    [anon_sym_exp] = ACTIONS(2970),
    [anon_sym_each] = ACTIONS(2970),
    [anon_sym_fc] = ACTIONS(2970),
    [anon_sym_fileno] = ACTIONS(2970),
    [anon_sym_gmtime] = ACTIONS(2970),
    [anon_sym_getc] = ACTIONS(2970),
    [anon_sym_getpgrp] = ACTIONS(2970),
    [anon_sym_getprotobyname] = ACTIONS(2970),
    [anon_sym_getpwname] = ACTIONS(2970),
    [anon_sym_getpwuid] = ACTIONS(2970),
    [anon_sym_getpeername] = ACTIONS(2970),
    [anon_sym_getnetbyname] = ACTIONS(2970),
    [anon_sym_getsockname] = ACTIONS(2970),
    [anon_sym_getgrnam] = ACTIONS(2970),
    [anon_sym_getgrgid] = ACTIONS(2970),
    [anon_sym_hex] = ACTIONS(2970),
    [anon_sym_int] = ACTIONS(2970),
    [anon_sym_keys] = ACTIONS(2970),
    [anon_sym_lc] = ACTIONS(2970),
    [anon_sym_lcfirst] = ACTIONS(2970),
    [anon_sym_length] = ACTIONS(2970),
    [anon_sym_localtime] = ACTIONS(2970),
    [anon_sym_log] = ACTIONS(2970),
    [anon_sym_lock] = ACTIONS(2970),
    [anon_sym_lstat] = ACTIONS(2970),
    [anon_sym_oct] = ACTIONS(2970),
    [anon_sym_ord] = ACTIONS(2970),
    [anon_sym_prototype] = ACTIONS(2970),
    [anon_sym_pop] = ACTIONS(2970),
    [anon_sym_pos] = ACTIONS(2970),
    [anon_sym_quotemeta] = ACTIONS(2970),
    [anon_sym_reset] = ACTIONS(2970),
    [anon_sym_rand] = ACTIONS(2970),
    [anon_sym_rmdir] = ACTIONS(2970),
    [anon_sym_readdir] = ACTIONS(2970),
    [anon_sym_readline] = ACTIONS(2970),
    [anon_sym_readpipe] = ACTIONS(2970),
    [anon_sym_rewinddir] = ACTIONS(2970),
    [anon_sym_readlink] = ACTIONS(2970),
    [anon_sym_ref] = ACTIONS(2970),
    [anon_sym_scalar] = ACTIONS(2970),
    [anon_sym_shift] = ACTIONS(2970),
    [anon_sym_sin] = ACTIONS(2970),
    [anon_sym_sleep] = ACTIONS(2970),
    [anon_sym_sqrt] = ACTIONS(2970),
    [anon_sym_srand] = ACTIONS(2970),
    [anon_sym_stat] = ACTIONS(2970),
    [anon_sym_study] = ACTIONS(2970),
    [anon_sym_tell] = ACTIONS(2970),
    [anon_sym_telldir] = ACTIONS(2970),
    [anon_sym_tied] = ACTIONS(2970),
    [anon_sym_uc] = ACTIONS(2970),
    [anon_sym_ucfirst] = ACTIONS(2970),
    [anon_sym_untie] = ACTIONS(2970),
    [anon_sym_umask] = ACTIONS(2970),
    [anon_sym_values] = ACTIONS(2970),
    [anon_sym_write] = ACTIONS(2970),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2970),
    [anon_sym___END__] = ACTIONS(2970),
    [anon_sym_EOT] = ACTIONS(2970),
    [anon_sym_q] = ACTIONS(2970),
    [anon_sym_qq] = ACTIONS(2970),
    [anon_sym_qw] = ACTIONS(2970),
    [anon_sym_qx] = ACTIONS(2970),
    [anon_sym_qr] = ACTIONS(2970),
    [anon_sym_m] = ACTIONS(2970),
    [anon_sym_s] = ACTIONS(2970),
    [anon_sym_tr] = ACTIONS(2970),
    [anon_sym_y] = ACTIONS(2970),
    [anon_sym_if] = ACTIONS(2970),
    [anon_sym_unless] = ACTIONS(2970),
    [anon_sym_while] = ACTIONS(2970),
    [anon_sym_until] = ACTIONS(2970),
    [aux_sym__bareword_token1] = ACTIONS(2970),
    [sym_number] = ACTIONS(2970),
    [anon_sym_true] = ACTIONS(2970),
    [anon_sym_false] = ACTIONS(2970),
    [sym__single_quote] = ACTIONS(2968),
    [sym__double_quote] = ACTIONS(2968),
    [sym__backtick_quote] = ACTIONS(2968),
    [sym__search_slash_quote] = ACTIONS(2968),
    [sym__open_readline_bracket] = ACTIONS(2968),
    [sym__open_fileglob_bracket] = ACTIONS(2968),
    [sym__PERLY_HEREDOC] = ACTIONS(2968),
    [sym__ctrl_z_hack] = ACTIONS(2968),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2968),
    [sym__filetest] = ACTIONS(2968),
  },
  [STATE(1455)] = {
    [sym_heredoc_content] = STATE(1455),
    [ts_builtin_sym_end] = ACTIONS(2972),
    [sym__identifier] = ACTIONS(2974),
    [anon_sym_LBRACE] = ACTIONS(2974),
    [anon_sym_SEMI] = ACTIONS(2974),
    [anon_sym_package] = ACTIONS(2974),
    [anon_sym_class] = ACTIONS(2974),
    [anon_sym_role] = ACTIONS(2974),
    [anon_sym_BUILD] = ACTIONS(2974),
    [anon_sym_ADJUST] = ACTIONS(2974),
    [anon_sym_DOLLAR] = ACTIONS(2974),
    [anon_sym_AT] = ACTIONS(2974),
    [anon_sym_my] = ACTIONS(2974),
    [anon_sym_extended] = ACTIONS(2974),
    [anon_sym_async] = ACTIONS(2974),
    [anon_sym_sub] = ACTIONS(2974),
    [anon_sym_method] = ACTIONS(2974),
    [anon_sym_LPAREN] = ACTIONS(2974),
    [anon_sym_continue] = ACTIONS(2974),
    [anon_sym_state] = ACTIONS(2974),
    [anon_sym_our] = ACTIONS(2974),
    [anon_sym_try] = ACTIONS(2974),
    [anon_sym_defer] = ACTIONS(2974),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2974),
    [anon_sym_LBRACK] = ACTIONS(2974),
    [anon_sym_LT_LT] = ACTIONS(2974),
    [anon_sym_SLASH_SLASH] = ACTIONS(2974),
    [anon_sym_PLUS] = ACTIONS(2974),
    [anon_sym_DASH] = ACTIONS(2974),
    [anon_sym_TILDE] = ACTIONS(2974),
    [anon_sym_BANG] = ACTIONS(2974),
    [anon_sym_PLUS_PLUS] = ACTIONS(2974),
    [anon_sym_DASH_DASH] = ACTIONS(2974),
    [anon_sym_BSLASH] = ACTIONS(2972),
    [anon_sym_do] = ACTIONS(2974),
    [anon_sym_eval] = ACTIONS(2974),
    [anon_sym_await] = ACTIONS(2974),
    [anon_sym_field] = ACTIONS(2974),
    [anon_sym_local] = ACTIONS(2974),
    [anon_sym_dynamically] = ACTIONS(2974),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2974),
    [anon_sym_require] = ACTIONS(2974),
    [anon_sym_map] = ACTIONS(2974),
    [anon_sym_grep] = ACTIONS(2974),
    [anon_sym_sort] = ACTIONS(2974),
    [anon_sym_goto] = ACTIONS(2974),
    [anon_sym_return] = ACTIONS(2974),
    [anon_sym_undef] = ACTIONS(2974),
    [anon_sym_PERCENT2] = ACTIONS(2974),
    [anon_sym_AMP2] = ACTIONS(2974),
    [anon_sym_STAR2] = ACTIONS(2974),
    [anon_sym_use] = ACTIONS(2974),
    [anon_sym_no] = ACTIONS(2974),
    [anon_sym_for] = ACTIONS(2974),
    [anon_sym_foreach] = ACTIONS(2974),
    [anon_sym_last] = ACTIONS(2974),
    [anon_sym_next] = ACTIONS(2974),
    [anon_sym_redo] = ACTIONS(2974),
    [anon_sym_BEGIN] = ACTIONS(2974),
    [anon_sym_INIT] = ACTIONS(2974),
    [anon_sym_CHECK] = ACTIONS(2974),
    [anon_sym_UNITCHECK] = ACTIONS(2974),
    [anon_sym_END] = ACTIONS(2974),
    [anon_sym___FILE__] = ACTIONS(2974),
    [anon_sym___LINE__] = ACTIONS(2974),
    [anon_sym___PACKAGE__] = ACTIONS(2974),
    [anon_sym___SUB__] = ACTIONS(2974),
    [anon_sym_break] = ACTIONS(2974),
    [anon_sym_fork] = ACTIONS(2974),
    [anon_sym_getppid] = ACTIONS(2974),
    [anon_sym_time] = ACTIONS(2974),
    [anon_sym_times] = ACTIONS(2974),
    [anon_sym_wait] = ACTIONS(2974),
    [anon_sym_wantarray] = ACTIONS(2974),
    [anon_sym_abs] = ACTIONS(2974),
    [anon_sym_alarm] = ACTIONS(2974),
    [anon_sym_chop] = ACTIONS(2974),
    [anon_sym_chdir] = ACTIONS(2974),
    [anon_sym_close] = ACTIONS(2974),
    [anon_sym_closedir] = ACTIONS(2974),
    [anon_sym_caller] = ACTIONS(2974),
    [anon_sym_chomp] = ACTIONS(2974),
    [anon_sym_chr] = ACTIONS(2974),
    [anon_sym_cos] = ACTIONS(2974),
    [anon_sym_chroot] = ACTIONS(2974),
    [anon_sym_defined] = ACTIONS(2974),
    [anon_sym_delete] = ACTIONS(2974),
    [anon_sym_dbmclose] = ACTIONS(2974),
    [anon_sym_exists] = ACTIONS(2974),
    [anon_sym_exit] = ACTIONS(2974),
    [anon_sym_eof] = ACTIONS(2974),
    [anon_sym_exp] = ACTIONS(2974),
    [anon_sym_each] = ACTIONS(2974),
    [anon_sym_fc] = ACTIONS(2974),
    [anon_sym_fileno] = ACTIONS(2974),
    [anon_sym_gmtime] = ACTIONS(2974),
    [anon_sym_getc] = ACTIONS(2974),
    [anon_sym_getpgrp] = ACTIONS(2974),
    [anon_sym_getprotobyname] = ACTIONS(2974),
    [anon_sym_getpwname] = ACTIONS(2974),
    [anon_sym_getpwuid] = ACTIONS(2974),
    [anon_sym_getpeername] = ACTIONS(2974),
    [anon_sym_getnetbyname] = ACTIONS(2974),
    [anon_sym_getsockname] = ACTIONS(2974),
    [anon_sym_getgrnam] = ACTIONS(2974),
    [anon_sym_getgrgid] = ACTIONS(2974),
    [anon_sym_hex] = ACTIONS(2974),
    [anon_sym_int] = ACTIONS(2974),
    [anon_sym_keys] = ACTIONS(2974),
    [anon_sym_lc] = ACTIONS(2974),
    [anon_sym_lcfirst] = ACTIONS(2974),
    [anon_sym_length] = ACTIONS(2974),
    [anon_sym_localtime] = ACTIONS(2974),
    [anon_sym_log] = ACTIONS(2974),
    [anon_sym_lock] = ACTIONS(2974),
    [anon_sym_lstat] = ACTIONS(2974),
    [anon_sym_oct] = ACTIONS(2974),
    [anon_sym_ord] = ACTIONS(2974),
    [anon_sym_prototype] = ACTIONS(2974),
    [anon_sym_pop] = ACTIONS(2974),
    [anon_sym_pos] = ACTIONS(2974),
    [anon_sym_quotemeta] = ACTIONS(2974),
    [anon_sym_reset] = ACTIONS(2974),
    [anon_sym_rand] = ACTIONS(2974),
    [anon_sym_rmdir] = ACTIONS(2974),
    [anon_sym_readdir] = ACTIONS(2974),
    [anon_sym_readline] = ACTIONS(2974),
    [anon_sym_readpipe] = ACTIONS(2974),
    [anon_sym_rewinddir] = ACTIONS(2974),
    [anon_sym_readlink] = ACTIONS(2974),
    [anon_sym_ref] = ACTIONS(2974),
    [anon_sym_scalar] = ACTIONS(2974),
    [anon_sym_shift] = ACTIONS(2974),
    [anon_sym_sin] = ACTIONS(2974),
    [anon_sym_sleep] = ACTIONS(2974),
    [anon_sym_sqrt] = ACTIONS(2974),
    [anon_sym_srand] = ACTIONS(2974),
    [anon_sym_stat] = ACTIONS(2974),
    [anon_sym_study] = ACTIONS(2974),
    [anon_sym_tell] = ACTIONS(2974),
    [anon_sym_telldir] = ACTIONS(2974),
    [anon_sym_tied] = ACTIONS(2974),
    [anon_sym_uc] = ACTIONS(2974),
    [anon_sym_ucfirst] = ACTIONS(2974),
    [anon_sym_untie] = ACTIONS(2974),
    [anon_sym_umask] = ACTIONS(2974),
    [anon_sym_values] = ACTIONS(2974),
    [anon_sym_write] = ACTIONS(2974),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2974),
    [anon_sym___END__] = ACTIONS(2974),
    [anon_sym_EOT] = ACTIONS(2974),
    [anon_sym_q] = ACTIONS(2974),
    [anon_sym_qq] = ACTIONS(2974),
    [anon_sym_qw] = ACTIONS(2974),
    [anon_sym_qx] = ACTIONS(2974),
    [anon_sym_qr] = ACTIONS(2974),
    [anon_sym_m] = ACTIONS(2974),
    [anon_sym_s] = ACTIONS(2974),
    [anon_sym_tr] = ACTIONS(2974),
    [anon_sym_y] = ACTIONS(2974),
    [anon_sym_if] = ACTIONS(2974),
    [anon_sym_unless] = ACTIONS(2974),
    [anon_sym_while] = ACTIONS(2974),
    [anon_sym_until] = ACTIONS(2974),
    [aux_sym__bareword_token1] = ACTIONS(2974),
    [sym_number] = ACTIONS(2974),
    [anon_sym_true] = ACTIONS(2974),
    [anon_sym_false] = ACTIONS(2974),
    [sym__single_quote] = ACTIONS(2972),
    [sym__double_quote] = ACTIONS(2972),
    [sym__backtick_quote] = ACTIONS(2972),
    [sym__search_slash_quote] = ACTIONS(2972),
    [sym__open_readline_bracket] = ACTIONS(2972),
    [sym__open_fileglob_bracket] = ACTIONS(2972),
    [sym__PERLY_HEREDOC] = ACTIONS(2972),
    [sym__ctrl_z_hack] = ACTIONS(2972),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2972),
    [sym__filetest] = ACTIONS(2972),
  },
  [STATE(1456)] = {
    [sym_heredoc_content] = STATE(1456),
    [ts_builtin_sym_end] = ACTIONS(2976),
    [sym__identifier] = ACTIONS(2978),
    [anon_sym_LBRACE] = ACTIONS(2978),
    [anon_sym_SEMI] = ACTIONS(2978),
    [anon_sym_package] = ACTIONS(2978),
    [anon_sym_class] = ACTIONS(2978),
    [anon_sym_role] = ACTIONS(2978),
    [anon_sym_BUILD] = ACTIONS(2978),
    [anon_sym_ADJUST] = ACTIONS(2978),
    [anon_sym_DOLLAR] = ACTIONS(2978),
    [anon_sym_AT] = ACTIONS(2978),
    [anon_sym_my] = ACTIONS(2978),
    [anon_sym_extended] = ACTIONS(2978),
    [anon_sym_async] = ACTIONS(2978),
    [anon_sym_sub] = ACTIONS(2978),
    [anon_sym_method] = ACTIONS(2978),
    [anon_sym_LPAREN] = ACTIONS(2978),
    [anon_sym_continue] = ACTIONS(2978),
    [anon_sym_state] = ACTIONS(2978),
    [anon_sym_our] = ACTIONS(2978),
    [anon_sym_try] = ACTIONS(2978),
    [anon_sym_defer] = ACTIONS(2978),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2978),
    [anon_sym_LBRACK] = ACTIONS(2978),
    [anon_sym_LT_LT] = ACTIONS(2978),
    [anon_sym_SLASH_SLASH] = ACTIONS(2978),
    [anon_sym_PLUS] = ACTIONS(2978),
    [anon_sym_DASH] = ACTIONS(2978),
    [anon_sym_TILDE] = ACTIONS(2978),
    [anon_sym_BANG] = ACTIONS(2978),
    [anon_sym_PLUS_PLUS] = ACTIONS(2978),
    [anon_sym_DASH_DASH] = ACTIONS(2978),
    [anon_sym_BSLASH] = ACTIONS(2976),
    [anon_sym_do] = ACTIONS(2978),
    [anon_sym_eval] = ACTIONS(2978),
    [anon_sym_await] = ACTIONS(2978),
    [anon_sym_field] = ACTIONS(2978),
    [anon_sym_local] = ACTIONS(2978),
    [anon_sym_dynamically] = ACTIONS(2978),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2978),
    [anon_sym_require] = ACTIONS(2978),
    [anon_sym_map] = ACTIONS(2978),
    [anon_sym_grep] = ACTIONS(2978),
    [anon_sym_sort] = ACTIONS(2978),
    [anon_sym_goto] = ACTIONS(2978),
    [anon_sym_return] = ACTIONS(2978),
    [anon_sym_undef] = ACTIONS(2978),
    [anon_sym_PERCENT2] = ACTIONS(2978),
    [anon_sym_AMP2] = ACTIONS(2978),
    [anon_sym_STAR2] = ACTIONS(2978),
    [anon_sym_use] = ACTIONS(2978),
    [anon_sym_no] = ACTIONS(2978),
    [anon_sym_for] = ACTIONS(2978),
    [anon_sym_foreach] = ACTIONS(2978),
    [anon_sym_last] = ACTIONS(2978),
    [anon_sym_next] = ACTIONS(2978),
    [anon_sym_redo] = ACTIONS(2978),
    [anon_sym_BEGIN] = ACTIONS(2978),
    [anon_sym_INIT] = ACTIONS(2978),
    [anon_sym_CHECK] = ACTIONS(2978),
    [anon_sym_UNITCHECK] = ACTIONS(2978),
    [anon_sym_END] = ACTIONS(2978),
    [anon_sym___FILE__] = ACTIONS(2978),
    [anon_sym___LINE__] = ACTIONS(2978),
    [anon_sym___PACKAGE__] = ACTIONS(2978),
    [anon_sym___SUB__] = ACTIONS(2978),
    [anon_sym_break] = ACTIONS(2978),
    [anon_sym_fork] = ACTIONS(2978),
    [anon_sym_getppid] = ACTIONS(2978),
    [anon_sym_time] = ACTIONS(2978),
    [anon_sym_times] = ACTIONS(2978),
    [anon_sym_wait] = ACTIONS(2978),
    [anon_sym_wantarray] = ACTIONS(2978),
    [anon_sym_abs] = ACTIONS(2978),
    [anon_sym_alarm] = ACTIONS(2978),
    [anon_sym_chop] = ACTIONS(2978),
    [anon_sym_chdir] = ACTIONS(2978),
    [anon_sym_close] = ACTIONS(2978),
    [anon_sym_closedir] = ACTIONS(2978),
    [anon_sym_caller] = ACTIONS(2978),
    [anon_sym_chomp] = ACTIONS(2978),
    [anon_sym_chr] = ACTIONS(2978),
    [anon_sym_cos] = ACTIONS(2978),
    [anon_sym_chroot] = ACTIONS(2978),
    [anon_sym_defined] = ACTIONS(2978),
    [anon_sym_delete] = ACTIONS(2978),
    [anon_sym_dbmclose] = ACTIONS(2978),
    [anon_sym_exists] = ACTIONS(2978),
    [anon_sym_exit] = ACTIONS(2978),
    [anon_sym_eof] = ACTIONS(2978),
    [anon_sym_exp] = ACTIONS(2978),
    [anon_sym_each] = ACTIONS(2978),
    [anon_sym_fc] = ACTIONS(2978),
    [anon_sym_fileno] = ACTIONS(2978),
    [anon_sym_gmtime] = ACTIONS(2978),
    [anon_sym_getc] = ACTIONS(2978),
    [anon_sym_getpgrp] = ACTIONS(2978),
    [anon_sym_getprotobyname] = ACTIONS(2978),
    [anon_sym_getpwname] = ACTIONS(2978),
    [anon_sym_getpwuid] = ACTIONS(2978),
    [anon_sym_getpeername] = ACTIONS(2978),
    [anon_sym_getnetbyname] = ACTIONS(2978),
    [anon_sym_getsockname] = ACTIONS(2978),
    [anon_sym_getgrnam] = ACTIONS(2978),
    [anon_sym_getgrgid] = ACTIONS(2978),
    [anon_sym_hex] = ACTIONS(2978),
    [anon_sym_int] = ACTIONS(2978),
    [anon_sym_keys] = ACTIONS(2978),
    [anon_sym_lc] = ACTIONS(2978),
    [anon_sym_lcfirst] = ACTIONS(2978),
    [anon_sym_length] = ACTIONS(2978),
    [anon_sym_localtime] = ACTIONS(2978),
    [anon_sym_log] = ACTIONS(2978),
    [anon_sym_lock] = ACTIONS(2978),
    [anon_sym_lstat] = ACTIONS(2978),
    [anon_sym_oct] = ACTIONS(2978),
    [anon_sym_ord] = ACTIONS(2978),
    [anon_sym_prototype] = ACTIONS(2978),
    [anon_sym_pop] = ACTIONS(2978),
    [anon_sym_pos] = ACTIONS(2978),
    [anon_sym_quotemeta] = ACTIONS(2978),
    [anon_sym_reset] = ACTIONS(2978),
    [anon_sym_rand] = ACTIONS(2978),
    [anon_sym_rmdir] = ACTIONS(2978),
    [anon_sym_readdir] = ACTIONS(2978),
    [anon_sym_readline] = ACTIONS(2978),
    [anon_sym_readpipe] = ACTIONS(2978),
    [anon_sym_rewinddir] = ACTIONS(2978),
    [anon_sym_readlink] = ACTIONS(2978),
    [anon_sym_ref] = ACTIONS(2978),
    [anon_sym_scalar] = ACTIONS(2978),
    [anon_sym_shift] = ACTIONS(2978),
    [anon_sym_sin] = ACTIONS(2978),
    [anon_sym_sleep] = ACTIONS(2978),
    [anon_sym_sqrt] = ACTIONS(2978),
    [anon_sym_srand] = ACTIONS(2978),
    [anon_sym_stat] = ACTIONS(2978),
    [anon_sym_study] = ACTIONS(2978),
    [anon_sym_tell] = ACTIONS(2978),
    [anon_sym_telldir] = ACTIONS(2978),
    [anon_sym_tied] = ACTIONS(2978),
    [anon_sym_uc] = ACTIONS(2978),
    [anon_sym_ucfirst] = ACTIONS(2978),
    [anon_sym_untie] = ACTIONS(2978),
    [anon_sym_umask] = ACTIONS(2978),
    [anon_sym_values] = ACTIONS(2978),
    [anon_sym_write] = ACTIONS(2978),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2978),
    [anon_sym___END__] = ACTIONS(2978),
    [anon_sym_EOT] = ACTIONS(2978),
    [anon_sym_q] = ACTIONS(2978),
    [anon_sym_qq] = ACTIONS(2978),
    [anon_sym_qw] = ACTIONS(2978),
    [anon_sym_qx] = ACTIONS(2978),
    [anon_sym_qr] = ACTIONS(2978),
    [anon_sym_m] = ACTIONS(2978),
    [anon_sym_s] = ACTIONS(2978),
    [anon_sym_tr] = ACTIONS(2978),
    [anon_sym_y] = ACTIONS(2978),
    [anon_sym_if] = ACTIONS(2978),
    [anon_sym_unless] = ACTIONS(2978),
    [anon_sym_while] = ACTIONS(2978),
    [anon_sym_until] = ACTIONS(2978),
    [aux_sym__bareword_token1] = ACTIONS(2978),
    [sym_number] = ACTIONS(2978),
    [anon_sym_true] = ACTIONS(2978),
    [anon_sym_false] = ACTIONS(2978),
    [sym__single_quote] = ACTIONS(2976),
    [sym__double_quote] = ACTIONS(2976),
    [sym__backtick_quote] = ACTIONS(2976),
    [sym__search_slash_quote] = ACTIONS(2976),
    [sym__open_readline_bracket] = ACTIONS(2976),
    [sym__open_fileglob_bracket] = ACTIONS(2976),
    [sym__PERLY_HEREDOC] = ACTIONS(2976),
    [sym__ctrl_z_hack] = ACTIONS(2976),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2976),
    [sym__filetest] = ACTIONS(2976),
  },
  [STATE(1457)] = {
    [sym_heredoc_content] = STATE(1457),
    [ts_builtin_sym_end] = ACTIONS(2980),
    [sym__identifier] = ACTIONS(2982),
    [anon_sym_LBRACE] = ACTIONS(2982),
    [anon_sym_SEMI] = ACTIONS(2982),
    [anon_sym_package] = ACTIONS(2982),
    [anon_sym_class] = ACTIONS(2982),
    [anon_sym_role] = ACTIONS(2982),
    [anon_sym_BUILD] = ACTIONS(2982),
    [anon_sym_ADJUST] = ACTIONS(2982),
    [anon_sym_DOLLAR] = ACTIONS(2982),
    [anon_sym_AT] = ACTIONS(2982),
    [anon_sym_my] = ACTIONS(2982),
    [anon_sym_extended] = ACTIONS(2982),
    [anon_sym_async] = ACTIONS(2982),
    [anon_sym_sub] = ACTIONS(2982),
    [anon_sym_method] = ACTIONS(2982),
    [anon_sym_LPAREN] = ACTIONS(2982),
    [anon_sym_continue] = ACTIONS(2982),
    [anon_sym_state] = ACTIONS(2982),
    [anon_sym_our] = ACTIONS(2982),
    [anon_sym_try] = ACTIONS(2982),
    [anon_sym_defer] = ACTIONS(2982),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2982),
    [anon_sym_LBRACK] = ACTIONS(2982),
    [anon_sym_LT_LT] = ACTIONS(2982),
    [anon_sym_SLASH_SLASH] = ACTIONS(2982),
    [anon_sym_PLUS] = ACTIONS(2982),
    [anon_sym_DASH] = ACTIONS(2982),
    [anon_sym_TILDE] = ACTIONS(2982),
    [anon_sym_BANG] = ACTIONS(2982),
    [anon_sym_PLUS_PLUS] = ACTIONS(2982),
    [anon_sym_DASH_DASH] = ACTIONS(2982),
    [anon_sym_BSLASH] = ACTIONS(2980),
    [anon_sym_do] = ACTIONS(2982),
    [anon_sym_eval] = ACTIONS(2982),
    [anon_sym_await] = ACTIONS(2982),
    [anon_sym_field] = ACTIONS(2982),
    [anon_sym_local] = ACTIONS(2982),
    [anon_sym_dynamically] = ACTIONS(2982),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2982),
    [anon_sym_require] = ACTIONS(2982),
    [anon_sym_map] = ACTIONS(2982),
    [anon_sym_grep] = ACTIONS(2982),
    [anon_sym_sort] = ACTIONS(2982),
    [anon_sym_goto] = ACTIONS(2982),
    [anon_sym_return] = ACTIONS(2982),
    [anon_sym_undef] = ACTIONS(2982),
    [anon_sym_PERCENT2] = ACTIONS(2982),
    [anon_sym_AMP2] = ACTIONS(2982),
    [anon_sym_STAR2] = ACTIONS(2982),
    [anon_sym_use] = ACTIONS(2982),
    [anon_sym_no] = ACTIONS(2982),
    [anon_sym_for] = ACTIONS(2982),
    [anon_sym_foreach] = ACTIONS(2982),
    [anon_sym_last] = ACTIONS(2982),
    [anon_sym_next] = ACTIONS(2982),
    [anon_sym_redo] = ACTIONS(2982),
    [anon_sym_BEGIN] = ACTIONS(2982),
    [anon_sym_INIT] = ACTIONS(2982),
    [anon_sym_CHECK] = ACTIONS(2982),
    [anon_sym_UNITCHECK] = ACTIONS(2982),
    [anon_sym_END] = ACTIONS(2982),
    [anon_sym___FILE__] = ACTIONS(2982),
    [anon_sym___LINE__] = ACTIONS(2982),
    [anon_sym___PACKAGE__] = ACTIONS(2982),
    [anon_sym___SUB__] = ACTIONS(2982),
    [anon_sym_break] = ACTIONS(2982),
    [anon_sym_fork] = ACTIONS(2982),
    [anon_sym_getppid] = ACTIONS(2982),
    [anon_sym_time] = ACTIONS(2982),
    [anon_sym_times] = ACTIONS(2982),
    [anon_sym_wait] = ACTIONS(2982),
    [anon_sym_wantarray] = ACTIONS(2982),
    [anon_sym_abs] = ACTIONS(2982),
    [anon_sym_alarm] = ACTIONS(2982),
    [anon_sym_chop] = ACTIONS(2982),
    [anon_sym_chdir] = ACTIONS(2982),
    [anon_sym_close] = ACTIONS(2982),
    [anon_sym_closedir] = ACTIONS(2982),
    [anon_sym_caller] = ACTIONS(2982),
    [anon_sym_chomp] = ACTIONS(2982),
    [anon_sym_chr] = ACTIONS(2982),
    [anon_sym_cos] = ACTIONS(2982),
    [anon_sym_chroot] = ACTIONS(2982),
    [anon_sym_defined] = ACTIONS(2982),
    [anon_sym_delete] = ACTIONS(2982),
    [anon_sym_dbmclose] = ACTIONS(2982),
    [anon_sym_exists] = ACTIONS(2982),
    [anon_sym_exit] = ACTIONS(2982),
    [anon_sym_eof] = ACTIONS(2982),
    [anon_sym_exp] = ACTIONS(2982),
    [anon_sym_each] = ACTIONS(2982),
    [anon_sym_fc] = ACTIONS(2982),
    [anon_sym_fileno] = ACTIONS(2982),
    [anon_sym_gmtime] = ACTIONS(2982),
    [anon_sym_getc] = ACTIONS(2982),
    [anon_sym_getpgrp] = ACTIONS(2982),
    [anon_sym_getprotobyname] = ACTIONS(2982),
    [anon_sym_getpwname] = ACTIONS(2982),
    [anon_sym_getpwuid] = ACTIONS(2982),
    [anon_sym_getpeername] = ACTIONS(2982),
    [anon_sym_getnetbyname] = ACTIONS(2982),
    [anon_sym_getsockname] = ACTIONS(2982),
    [anon_sym_getgrnam] = ACTIONS(2982),
    [anon_sym_getgrgid] = ACTIONS(2982),
    [anon_sym_hex] = ACTIONS(2982),
    [anon_sym_int] = ACTIONS(2982),
    [anon_sym_keys] = ACTIONS(2982),
    [anon_sym_lc] = ACTIONS(2982),
    [anon_sym_lcfirst] = ACTIONS(2982),
    [anon_sym_length] = ACTIONS(2982),
    [anon_sym_localtime] = ACTIONS(2982),
    [anon_sym_log] = ACTIONS(2982),
    [anon_sym_lock] = ACTIONS(2982),
    [anon_sym_lstat] = ACTIONS(2982),
    [anon_sym_oct] = ACTIONS(2982),
    [anon_sym_ord] = ACTIONS(2982),
    [anon_sym_prototype] = ACTIONS(2982),
    [anon_sym_pop] = ACTIONS(2982),
    [anon_sym_pos] = ACTIONS(2982),
    [anon_sym_quotemeta] = ACTIONS(2982),
    [anon_sym_reset] = ACTIONS(2982),
    [anon_sym_rand] = ACTIONS(2982),
    [anon_sym_rmdir] = ACTIONS(2982),
    [anon_sym_readdir] = ACTIONS(2982),
    [anon_sym_readline] = ACTIONS(2982),
    [anon_sym_readpipe] = ACTIONS(2982),
    [anon_sym_rewinddir] = ACTIONS(2982),
    [anon_sym_readlink] = ACTIONS(2982),
    [anon_sym_ref] = ACTIONS(2982),
    [anon_sym_scalar] = ACTIONS(2982),
    [anon_sym_shift] = ACTIONS(2982),
    [anon_sym_sin] = ACTIONS(2982),
    [anon_sym_sleep] = ACTIONS(2982),
    [anon_sym_sqrt] = ACTIONS(2982),
    [anon_sym_srand] = ACTIONS(2982),
    [anon_sym_stat] = ACTIONS(2982),
    [anon_sym_study] = ACTIONS(2982),
    [anon_sym_tell] = ACTIONS(2982),
    [anon_sym_telldir] = ACTIONS(2982),
    [anon_sym_tied] = ACTIONS(2982),
    [anon_sym_uc] = ACTIONS(2982),
    [anon_sym_ucfirst] = ACTIONS(2982),
    [anon_sym_untie] = ACTIONS(2982),
    [anon_sym_umask] = ACTIONS(2982),
    [anon_sym_values] = ACTIONS(2982),
    [anon_sym_write] = ACTIONS(2982),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2982),
    [anon_sym___END__] = ACTIONS(2982),
    [anon_sym_EOT] = ACTIONS(2982),
    [anon_sym_q] = ACTIONS(2982),
    [anon_sym_qq] = ACTIONS(2982),
    [anon_sym_qw] = ACTIONS(2982),
    [anon_sym_qx] = ACTIONS(2982),
    [anon_sym_qr] = ACTIONS(2982),
    [anon_sym_m] = ACTIONS(2982),
    [anon_sym_s] = ACTIONS(2982),
    [anon_sym_tr] = ACTIONS(2982),
    [anon_sym_y] = ACTIONS(2982),
    [anon_sym_if] = ACTIONS(2982),
    [anon_sym_unless] = ACTIONS(2982),
    [anon_sym_while] = ACTIONS(2982),
    [anon_sym_until] = ACTIONS(2982),
    [aux_sym__bareword_token1] = ACTIONS(2982),
    [sym_number] = ACTIONS(2982),
    [anon_sym_true] = ACTIONS(2982),
    [anon_sym_false] = ACTIONS(2982),
    [sym__single_quote] = ACTIONS(2980),
    [sym__double_quote] = ACTIONS(2980),
    [sym__backtick_quote] = ACTIONS(2980),
    [sym__search_slash_quote] = ACTIONS(2980),
    [sym__open_readline_bracket] = ACTIONS(2980),
    [sym__open_fileglob_bracket] = ACTIONS(2980),
    [sym__PERLY_HEREDOC] = ACTIONS(2980),
    [sym__ctrl_z_hack] = ACTIONS(2980),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2980),
    [sym__filetest] = ACTIONS(2980),
  },
  [STATE(1458)] = {
    [sym_heredoc_content] = STATE(1458),
    [ts_builtin_sym_end] = ACTIONS(2984),
    [sym__identifier] = ACTIONS(2986),
    [anon_sym_LBRACE] = ACTIONS(2986),
    [anon_sym_SEMI] = ACTIONS(2986),
    [anon_sym_package] = ACTIONS(2986),
    [anon_sym_class] = ACTIONS(2986),
    [anon_sym_role] = ACTIONS(2986),
    [anon_sym_BUILD] = ACTIONS(2986),
    [anon_sym_ADJUST] = ACTIONS(2986),
    [anon_sym_DOLLAR] = ACTIONS(2986),
    [anon_sym_AT] = ACTIONS(2986),
    [anon_sym_my] = ACTIONS(2986),
    [anon_sym_extended] = ACTIONS(2986),
    [anon_sym_async] = ACTIONS(2986),
    [anon_sym_sub] = ACTIONS(2986),
    [anon_sym_method] = ACTIONS(2986),
    [anon_sym_LPAREN] = ACTIONS(2986),
    [anon_sym_continue] = ACTIONS(2986),
    [anon_sym_state] = ACTIONS(2986),
    [anon_sym_our] = ACTIONS(2986),
    [anon_sym_try] = ACTIONS(2986),
    [anon_sym_defer] = ACTIONS(2986),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2986),
    [anon_sym_LBRACK] = ACTIONS(2986),
    [anon_sym_LT_LT] = ACTIONS(2986),
    [anon_sym_SLASH_SLASH] = ACTIONS(2986),
    [anon_sym_PLUS] = ACTIONS(2986),
    [anon_sym_DASH] = ACTIONS(2986),
    [anon_sym_TILDE] = ACTIONS(2986),
    [anon_sym_BANG] = ACTIONS(2986),
    [anon_sym_PLUS_PLUS] = ACTIONS(2986),
    [anon_sym_DASH_DASH] = ACTIONS(2986),
    [anon_sym_BSLASH] = ACTIONS(2984),
    [anon_sym_do] = ACTIONS(2986),
    [anon_sym_eval] = ACTIONS(2986),
    [anon_sym_await] = ACTIONS(2986),
    [anon_sym_field] = ACTIONS(2986),
    [anon_sym_local] = ACTIONS(2986),
    [anon_sym_dynamically] = ACTIONS(2986),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2986),
    [anon_sym_require] = ACTIONS(2986),
    [anon_sym_map] = ACTIONS(2986),
    [anon_sym_grep] = ACTIONS(2986),
    [anon_sym_sort] = ACTIONS(2986),
    [anon_sym_goto] = ACTIONS(2986),
    [anon_sym_return] = ACTIONS(2986),
    [anon_sym_undef] = ACTIONS(2986),
    [anon_sym_PERCENT2] = ACTIONS(2986),
    [anon_sym_AMP2] = ACTIONS(2986),
    [anon_sym_STAR2] = ACTIONS(2986),
    [anon_sym_use] = ACTIONS(2986),
    [anon_sym_no] = ACTIONS(2986),
    [anon_sym_for] = ACTIONS(2986),
    [anon_sym_foreach] = ACTIONS(2986),
    [anon_sym_last] = ACTIONS(2986),
    [anon_sym_next] = ACTIONS(2986),
    [anon_sym_redo] = ACTIONS(2986),
    [anon_sym_BEGIN] = ACTIONS(2986),
    [anon_sym_INIT] = ACTIONS(2986),
    [anon_sym_CHECK] = ACTIONS(2986),
    [anon_sym_UNITCHECK] = ACTIONS(2986),
    [anon_sym_END] = ACTIONS(2986),
    [anon_sym___FILE__] = ACTIONS(2986),
    [anon_sym___LINE__] = ACTIONS(2986),
    [anon_sym___PACKAGE__] = ACTIONS(2986),
    [anon_sym___SUB__] = ACTIONS(2986),
    [anon_sym_break] = ACTIONS(2986),
    [anon_sym_fork] = ACTIONS(2986),
    [anon_sym_getppid] = ACTIONS(2986),
    [anon_sym_time] = ACTIONS(2986),
    [anon_sym_times] = ACTIONS(2986),
    [anon_sym_wait] = ACTIONS(2986),
    [anon_sym_wantarray] = ACTIONS(2986),
    [anon_sym_abs] = ACTIONS(2986),
    [anon_sym_alarm] = ACTIONS(2986),
    [anon_sym_chop] = ACTIONS(2986),
    [anon_sym_chdir] = ACTIONS(2986),
    [anon_sym_close] = ACTIONS(2986),
    [anon_sym_closedir] = ACTIONS(2986),
    [anon_sym_caller] = ACTIONS(2986),
    [anon_sym_chomp] = ACTIONS(2986),
    [anon_sym_chr] = ACTIONS(2986),
    [anon_sym_cos] = ACTIONS(2986),
    [anon_sym_chroot] = ACTIONS(2986),
    [anon_sym_defined] = ACTIONS(2986),
    [anon_sym_delete] = ACTIONS(2986),
    [anon_sym_dbmclose] = ACTIONS(2986),
    [anon_sym_exists] = ACTIONS(2986),
    [anon_sym_exit] = ACTIONS(2986),
    [anon_sym_eof] = ACTIONS(2986),
    [anon_sym_exp] = ACTIONS(2986),
    [anon_sym_each] = ACTIONS(2986),
    [anon_sym_fc] = ACTIONS(2986),
    [anon_sym_fileno] = ACTIONS(2986),
    [anon_sym_gmtime] = ACTIONS(2986),
    [anon_sym_getc] = ACTIONS(2986),
    [anon_sym_getpgrp] = ACTIONS(2986),
    [anon_sym_getprotobyname] = ACTIONS(2986),
    [anon_sym_getpwname] = ACTIONS(2986),
    [anon_sym_getpwuid] = ACTIONS(2986),
    [anon_sym_getpeername] = ACTIONS(2986),
    [anon_sym_getnetbyname] = ACTIONS(2986),
    [anon_sym_getsockname] = ACTIONS(2986),
    [anon_sym_getgrnam] = ACTIONS(2986),
    [anon_sym_getgrgid] = ACTIONS(2986),
    [anon_sym_hex] = ACTIONS(2986),
    [anon_sym_int] = ACTIONS(2986),
    [anon_sym_keys] = ACTIONS(2986),
    [anon_sym_lc] = ACTIONS(2986),
    [anon_sym_lcfirst] = ACTIONS(2986),
    [anon_sym_length] = ACTIONS(2986),
    [anon_sym_localtime] = ACTIONS(2986),
    [anon_sym_log] = ACTIONS(2986),
    [anon_sym_lock] = ACTIONS(2986),
    [anon_sym_lstat] = ACTIONS(2986),
    [anon_sym_oct] = ACTIONS(2986),
    [anon_sym_ord] = ACTIONS(2986),
    [anon_sym_prototype] = ACTIONS(2986),
    [anon_sym_pop] = ACTIONS(2986),
    [anon_sym_pos] = ACTIONS(2986),
    [anon_sym_quotemeta] = ACTIONS(2986),
    [anon_sym_reset] = ACTIONS(2986),
    [anon_sym_rand] = ACTIONS(2986),
    [anon_sym_rmdir] = ACTIONS(2986),
    [anon_sym_readdir] = ACTIONS(2986),
    [anon_sym_readline] = ACTIONS(2986),
    [anon_sym_readpipe] = ACTIONS(2986),
    [anon_sym_rewinddir] = ACTIONS(2986),
    [anon_sym_readlink] = ACTIONS(2986),
    [anon_sym_ref] = ACTIONS(2986),
    [anon_sym_scalar] = ACTIONS(2986),
    [anon_sym_shift] = ACTIONS(2986),
    [anon_sym_sin] = ACTIONS(2986),
    [anon_sym_sleep] = ACTIONS(2986),
    [anon_sym_sqrt] = ACTIONS(2986),
    [anon_sym_srand] = ACTIONS(2986),
    [anon_sym_stat] = ACTIONS(2986),
    [anon_sym_study] = ACTIONS(2986),
    [anon_sym_tell] = ACTIONS(2986),
    [anon_sym_telldir] = ACTIONS(2986),
    [anon_sym_tied] = ACTIONS(2986),
    [anon_sym_uc] = ACTIONS(2986),
    [anon_sym_ucfirst] = ACTIONS(2986),
    [anon_sym_untie] = ACTIONS(2986),
    [anon_sym_umask] = ACTIONS(2986),
    [anon_sym_values] = ACTIONS(2986),
    [anon_sym_write] = ACTIONS(2986),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2986),
    [anon_sym___END__] = ACTIONS(2986),
    [anon_sym_EOT] = ACTIONS(2986),
    [anon_sym_q] = ACTIONS(2986),
    [anon_sym_qq] = ACTIONS(2986),
    [anon_sym_qw] = ACTIONS(2986),
    [anon_sym_qx] = ACTIONS(2986),
    [anon_sym_qr] = ACTIONS(2986),
    [anon_sym_m] = ACTIONS(2986),
    [anon_sym_s] = ACTIONS(2986),
    [anon_sym_tr] = ACTIONS(2986),
    [anon_sym_y] = ACTIONS(2986),
    [anon_sym_if] = ACTIONS(2986),
    [anon_sym_unless] = ACTIONS(2986),
    [anon_sym_while] = ACTIONS(2986),
    [anon_sym_until] = ACTIONS(2986),
    [aux_sym__bareword_token1] = ACTIONS(2986),
    [sym_number] = ACTIONS(2986),
    [anon_sym_true] = ACTIONS(2986),
    [anon_sym_false] = ACTIONS(2986),
    [sym__single_quote] = ACTIONS(2984),
    [sym__double_quote] = ACTIONS(2984),
    [sym__backtick_quote] = ACTIONS(2984),
    [sym__search_slash_quote] = ACTIONS(2984),
    [sym__open_readline_bracket] = ACTIONS(2984),
    [sym__open_fileglob_bracket] = ACTIONS(2984),
    [sym__PERLY_HEREDOC] = ACTIONS(2984),
    [sym__ctrl_z_hack] = ACTIONS(2984),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2984),
    [sym__filetest] = ACTIONS(2984),
  },
  [STATE(1459)] = {
    [sym_heredoc_content] = STATE(1459),
    [ts_builtin_sym_end] = ACTIONS(2988),
    [sym__identifier] = ACTIONS(2990),
    [anon_sym_LBRACE] = ACTIONS(2990),
    [anon_sym_SEMI] = ACTIONS(2990),
    [anon_sym_package] = ACTIONS(2990),
    [anon_sym_class] = ACTIONS(2990),
    [anon_sym_role] = ACTIONS(2990),
    [anon_sym_BUILD] = ACTIONS(2990),
    [anon_sym_ADJUST] = ACTIONS(2990),
    [anon_sym_DOLLAR] = ACTIONS(2990),
    [anon_sym_AT] = ACTIONS(2990),
    [anon_sym_my] = ACTIONS(2990),
    [anon_sym_extended] = ACTIONS(2990),
    [anon_sym_async] = ACTIONS(2990),
    [anon_sym_sub] = ACTIONS(2990),
    [anon_sym_method] = ACTIONS(2990),
    [anon_sym_LPAREN] = ACTIONS(2990),
    [anon_sym_continue] = ACTIONS(2990),
    [anon_sym_state] = ACTIONS(2990),
    [anon_sym_our] = ACTIONS(2990),
    [anon_sym_try] = ACTIONS(2990),
    [anon_sym_defer] = ACTIONS(2990),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2990),
    [anon_sym_LBRACK] = ACTIONS(2990),
    [anon_sym_LT_LT] = ACTIONS(2990),
    [anon_sym_SLASH_SLASH] = ACTIONS(2990),
    [anon_sym_PLUS] = ACTIONS(2990),
    [anon_sym_DASH] = ACTIONS(2990),
    [anon_sym_TILDE] = ACTIONS(2990),
    [anon_sym_BANG] = ACTIONS(2990),
    [anon_sym_PLUS_PLUS] = ACTIONS(2990),
    [anon_sym_DASH_DASH] = ACTIONS(2990),
    [anon_sym_BSLASH] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2990),
    [anon_sym_eval] = ACTIONS(2990),
    [anon_sym_await] = ACTIONS(2990),
    [anon_sym_field] = ACTIONS(2990),
    [anon_sym_local] = ACTIONS(2990),
    [anon_sym_dynamically] = ACTIONS(2990),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2990),
    [anon_sym_require] = ACTIONS(2990),
    [anon_sym_map] = ACTIONS(2990),
    [anon_sym_grep] = ACTIONS(2990),
    [anon_sym_sort] = ACTIONS(2990),
    [anon_sym_goto] = ACTIONS(2990),
    [anon_sym_return] = ACTIONS(2990),
    [anon_sym_undef] = ACTIONS(2990),
    [anon_sym_PERCENT2] = ACTIONS(2990),
    [anon_sym_AMP2] = ACTIONS(2990),
    [anon_sym_STAR2] = ACTIONS(2990),
    [anon_sym_use] = ACTIONS(2990),
    [anon_sym_no] = ACTIONS(2990),
    [anon_sym_for] = ACTIONS(2990),
    [anon_sym_foreach] = ACTIONS(2990),
    [anon_sym_last] = ACTIONS(2990),
    [anon_sym_next] = ACTIONS(2990),
    [anon_sym_redo] = ACTIONS(2990),
    [anon_sym_BEGIN] = ACTIONS(2990),
    [anon_sym_INIT] = ACTIONS(2990),
    [anon_sym_CHECK] = ACTIONS(2990),
    [anon_sym_UNITCHECK] = ACTIONS(2990),
    [anon_sym_END] = ACTIONS(2990),
    [anon_sym___FILE__] = ACTIONS(2990),
    [anon_sym___LINE__] = ACTIONS(2990),
    [anon_sym___PACKAGE__] = ACTIONS(2990),
    [anon_sym___SUB__] = ACTIONS(2990),
    [anon_sym_break] = ACTIONS(2990),
    [anon_sym_fork] = ACTIONS(2990),
    [anon_sym_getppid] = ACTIONS(2990),
    [anon_sym_time] = ACTIONS(2990),
    [anon_sym_times] = ACTIONS(2990),
    [anon_sym_wait] = ACTIONS(2990),
    [anon_sym_wantarray] = ACTIONS(2990),
    [anon_sym_abs] = ACTIONS(2990),
    [anon_sym_alarm] = ACTIONS(2990),
    [anon_sym_chop] = ACTIONS(2990),
    [anon_sym_chdir] = ACTIONS(2990),
    [anon_sym_close] = ACTIONS(2990),
    [anon_sym_closedir] = ACTIONS(2990),
    [anon_sym_caller] = ACTIONS(2990),
    [anon_sym_chomp] = ACTIONS(2990),
    [anon_sym_chr] = ACTIONS(2990),
    [anon_sym_cos] = ACTIONS(2990),
    [anon_sym_chroot] = ACTIONS(2990),
    [anon_sym_defined] = ACTIONS(2990),
    [anon_sym_delete] = ACTIONS(2990),
    [anon_sym_dbmclose] = ACTIONS(2990),
    [anon_sym_exists] = ACTIONS(2990),
    [anon_sym_exit] = ACTIONS(2990),
    [anon_sym_eof] = ACTIONS(2990),
    [anon_sym_exp] = ACTIONS(2990),
    [anon_sym_each] = ACTIONS(2990),
    [anon_sym_fc] = ACTIONS(2990),
    [anon_sym_fileno] = ACTIONS(2990),
    [anon_sym_gmtime] = ACTIONS(2990),
    [anon_sym_getc] = ACTIONS(2990),
    [anon_sym_getpgrp] = ACTIONS(2990),
    [anon_sym_getprotobyname] = ACTIONS(2990),
    [anon_sym_getpwname] = ACTIONS(2990),
    [anon_sym_getpwuid] = ACTIONS(2990),
    [anon_sym_getpeername] = ACTIONS(2990),
    [anon_sym_getnetbyname] = ACTIONS(2990),
    [anon_sym_getsockname] = ACTIONS(2990),
    [anon_sym_getgrnam] = ACTIONS(2990),
    [anon_sym_getgrgid] = ACTIONS(2990),
    [anon_sym_hex] = ACTIONS(2990),
    [anon_sym_int] = ACTIONS(2990),
    [anon_sym_keys] = ACTIONS(2990),
    [anon_sym_lc] = ACTIONS(2990),
    [anon_sym_lcfirst] = ACTIONS(2990),
    [anon_sym_length] = ACTIONS(2990),
    [anon_sym_localtime] = ACTIONS(2990),
    [anon_sym_log] = ACTIONS(2990),
    [anon_sym_lock] = ACTIONS(2990),
    [anon_sym_lstat] = ACTIONS(2990),
    [anon_sym_oct] = ACTIONS(2990),
    [anon_sym_ord] = ACTIONS(2990),
    [anon_sym_prototype] = ACTIONS(2990),
    [anon_sym_pop] = ACTIONS(2990),
    [anon_sym_pos] = ACTIONS(2990),
    [anon_sym_quotemeta] = ACTIONS(2990),
    [anon_sym_reset] = ACTIONS(2990),
    [anon_sym_rand] = ACTIONS(2990),
    [anon_sym_rmdir] = ACTIONS(2990),
    [anon_sym_readdir] = ACTIONS(2990),
    [anon_sym_readline] = ACTIONS(2990),
    [anon_sym_readpipe] = ACTIONS(2990),
    [anon_sym_rewinddir] = ACTIONS(2990),
    [anon_sym_readlink] = ACTIONS(2990),
    [anon_sym_ref] = ACTIONS(2990),
    [anon_sym_scalar] = ACTIONS(2990),
    [anon_sym_shift] = ACTIONS(2990),
    [anon_sym_sin] = ACTIONS(2990),
    [anon_sym_sleep] = ACTIONS(2990),
    [anon_sym_sqrt] = ACTIONS(2990),
    [anon_sym_srand] = ACTIONS(2990),
    [anon_sym_stat] = ACTIONS(2990),
    [anon_sym_study] = ACTIONS(2990),
    [anon_sym_tell] = ACTIONS(2990),
    [anon_sym_telldir] = ACTIONS(2990),
    [anon_sym_tied] = ACTIONS(2990),
    [anon_sym_uc] = ACTIONS(2990),
    [anon_sym_ucfirst] = ACTIONS(2990),
    [anon_sym_untie] = ACTIONS(2990),
    [anon_sym_umask] = ACTIONS(2990),
    [anon_sym_values] = ACTIONS(2990),
    [anon_sym_write] = ACTIONS(2990),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2990),
    [anon_sym___END__] = ACTIONS(2990),
    [anon_sym_EOT] = ACTIONS(2990),
    [anon_sym_q] = ACTIONS(2990),
    [anon_sym_qq] = ACTIONS(2990),
    [anon_sym_qw] = ACTIONS(2990),
    [anon_sym_qx] = ACTIONS(2990),
    [anon_sym_qr] = ACTIONS(2990),
    [anon_sym_m] = ACTIONS(2990),
    [anon_sym_s] = ACTIONS(2990),
    [anon_sym_tr] = ACTIONS(2990),
    [anon_sym_y] = ACTIONS(2990),
    [anon_sym_if] = ACTIONS(2990),
    [anon_sym_unless] = ACTIONS(2990),
    [anon_sym_while] = ACTIONS(2990),
    [anon_sym_until] = ACTIONS(2990),
    [aux_sym__bareword_token1] = ACTIONS(2990),
    [sym_number] = ACTIONS(2990),
    [anon_sym_true] = ACTIONS(2990),
    [anon_sym_false] = ACTIONS(2990),
    [sym__single_quote] = ACTIONS(2988),
    [sym__double_quote] = ACTIONS(2988),
    [sym__backtick_quote] = ACTIONS(2988),
    [sym__search_slash_quote] = ACTIONS(2988),
    [sym__open_readline_bracket] = ACTIONS(2988),
    [sym__open_fileglob_bracket] = ACTIONS(2988),
    [sym__PERLY_HEREDOC] = ACTIONS(2988),
    [sym__ctrl_z_hack] = ACTIONS(2988),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2988),
    [sym__filetest] = ACTIONS(2988),
  },
  [STATE(1460)] = {
    [sym_heredoc_content] = STATE(1460),
    [ts_builtin_sym_end] = ACTIONS(2992),
    [sym__identifier] = ACTIONS(2994),
    [anon_sym_LBRACE] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(2994),
    [anon_sym_package] = ACTIONS(2994),
    [anon_sym_class] = ACTIONS(2994),
    [anon_sym_role] = ACTIONS(2994),
    [anon_sym_BUILD] = ACTIONS(2994),
    [anon_sym_ADJUST] = ACTIONS(2994),
    [anon_sym_DOLLAR] = ACTIONS(2994),
    [anon_sym_AT] = ACTIONS(2994),
    [anon_sym_my] = ACTIONS(2994),
    [anon_sym_extended] = ACTIONS(2994),
    [anon_sym_async] = ACTIONS(2994),
    [anon_sym_sub] = ACTIONS(2994),
    [anon_sym_method] = ACTIONS(2994),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_continue] = ACTIONS(2994),
    [anon_sym_state] = ACTIONS(2994),
    [anon_sym_our] = ACTIONS(2994),
    [anon_sym_try] = ACTIONS(2994),
    [anon_sym_defer] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2994),
    [anon_sym_LBRACK] = ACTIONS(2994),
    [anon_sym_LT_LT] = ACTIONS(2994),
    [anon_sym_SLASH_SLASH] = ACTIONS(2994),
    [anon_sym_PLUS] = ACTIONS(2994),
    [anon_sym_DASH] = ACTIONS(2994),
    [anon_sym_TILDE] = ACTIONS(2994),
    [anon_sym_BANG] = ACTIONS(2994),
    [anon_sym_PLUS_PLUS] = ACTIONS(2994),
    [anon_sym_DASH_DASH] = ACTIONS(2994),
    [anon_sym_BSLASH] = ACTIONS(2992),
    [anon_sym_do] = ACTIONS(2994),
    [anon_sym_eval] = ACTIONS(2994),
    [anon_sym_await] = ACTIONS(2994),
    [anon_sym_field] = ACTIONS(2994),
    [anon_sym_local] = ACTIONS(2994),
    [anon_sym_dynamically] = ACTIONS(2994),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2994),
    [anon_sym_require] = ACTIONS(2994),
    [anon_sym_map] = ACTIONS(2994),
    [anon_sym_grep] = ACTIONS(2994),
    [anon_sym_sort] = ACTIONS(2994),
    [anon_sym_goto] = ACTIONS(2994),
    [anon_sym_return] = ACTIONS(2994),
    [anon_sym_undef] = ACTIONS(2994),
    [anon_sym_PERCENT2] = ACTIONS(2994),
    [anon_sym_AMP2] = ACTIONS(2994),
    [anon_sym_STAR2] = ACTIONS(2994),
    [anon_sym_use] = ACTIONS(2994),
    [anon_sym_no] = ACTIONS(2994),
    [anon_sym_for] = ACTIONS(2994),
    [anon_sym_foreach] = ACTIONS(2994),
    [anon_sym_last] = ACTIONS(2994),
    [anon_sym_next] = ACTIONS(2994),
    [anon_sym_redo] = ACTIONS(2994),
    [anon_sym_BEGIN] = ACTIONS(2994),
    [anon_sym_INIT] = ACTIONS(2994),
    [anon_sym_CHECK] = ACTIONS(2994),
    [anon_sym_UNITCHECK] = ACTIONS(2994),
    [anon_sym_END] = ACTIONS(2994),
    [anon_sym___FILE__] = ACTIONS(2994),
    [anon_sym___LINE__] = ACTIONS(2994),
    [anon_sym___PACKAGE__] = ACTIONS(2994),
    [anon_sym___SUB__] = ACTIONS(2994),
    [anon_sym_break] = ACTIONS(2994),
    [anon_sym_fork] = ACTIONS(2994),
    [anon_sym_getppid] = ACTIONS(2994),
    [anon_sym_time] = ACTIONS(2994),
    [anon_sym_times] = ACTIONS(2994),
    [anon_sym_wait] = ACTIONS(2994),
    [anon_sym_wantarray] = ACTIONS(2994),
    [anon_sym_abs] = ACTIONS(2994),
    [anon_sym_alarm] = ACTIONS(2994),
    [anon_sym_chop] = ACTIONS(2994),
    [anon_sym_chdir] = ACTIONS(2994),
    [anon_sym_close] = ACTIONS(2994),
    [anon_sym_closedir] = ACTIONS(2994),
    [anon_sym_caller] = ACTIONS(2994),
    [anon_sym_chomp] = ACTIONS(2994),
    [anon_sym_chr] = ACTIONS(2994),
    [anon_sym_cos] = ACTIONS(2994),
    [anon_sym_chroot] = ACTIONS(2994),
    [anon_sym_defined] = ACTIONS(2994),
    [anon_sym_delete] = ACTIONS(2994),
    [anon_sym_dbmclose] = ACTIONS(2994),
    [anon_sym_exists] = ACTIONS(2994),
    [anon_sym_exit] = ACTIONS(2994),
    [anon_sym_eof] = ACTIONS(2994),
    [anon_sym_exp] = ACTIONS(2994),
    [anon_sym_each] = ACTIONS(2994),
    [anon_sym_fc] = ACTIONS(2994),
    [anon_sym_fileno] = ACTIONS(2994),
    [anon_sym_gmtime] = ACTIONS(2994),
    [anon_sym_getc] = ACTIONS(2994),
    [anon_sym_getpgrp] = ACTIONS(2994),
    [anon_sym_getprotobyname] = ACTIONS(2994),
    [anon_sym_getpwname] = ACTIONS(2994),
    [anon_sym_getpwuid] = ACTIONS(2994),
    [anon_sym_getpeername] = ACTIONS(2994),
    [anon_sym_getnetbyname] = ACTIONS(2994),
    [anon_sym_getsockname] = ACTIONS(2994),
    [anon_sym_getgrnam] = ACTIONS(2994),
    [anon_sym_getgrgid] = ACTIONS(2994),
    [anon_sym_hex] = ACTIONS(2994),
    [anon_sym_int] = ACTIONS(2994),
    [anon_sym_keys] = ACTIONS(2994),
    [anon_sym_lc] = ACTIONS(2994),
    [anon_sym_lcfirst] = ACTIONS(2994),
    [anon_sym_length] = ACTIONS(2994),
    [anon_sym_localtime] = ACTIONS(2994),
    [anon_sym_log] = ACTIONS(2994),
    [anon_sym_lock] = ACTIONS(2994),
    [anon_sym_lstat] = ACTIONS(2994),
    [anon_sym_oct] = ACTIONS(2994),
    [anon_sym_ord] = ACTIONS(2994),
    [anon_sym_prototype] = ACTIONS(2994),
    [anon_sym_pop] = ACTIONS(2994),
    [anon_sym_pos] = ACTIONS(2994),
    [anon_sym_quotemeta] = ACTIONS(2994),
    [anon_sym_reset] = ACTIONS(2994),
    [anon_sym_rand] = ACTIONS(2994),
    [anon_sym_rmdir] = ACTIONS(2994),
    [anon_sym_readdir] = ACTIONS(2994),
    [anon_sym_readline] = ACTIONS(2994),
    [anon_sym_readpipe] = ACTIONS(2994),
    [anon_sym_rewinddir] = ACTIONS(2994),
    [anon_sym_readlink] = ACTIONS(2994),
    [anon_sym_ref] = ACTIONS(2994),
    [anon_sym_scalar] = ACTIONS(2994),
    [anon_sym_shift] = ACTIONS(2994),
    [anon_sym_sin] = ACTIONS(2994),
    [anon_sym_sleep] = ACTIONS(2994),
    [anon_sym_sqrt] = ACTIONS(2994),
    [anon_sym_srand] = ACTIONS(2994),
    [anon_sym_stat] = ACTIONS(2994),
    [anon_sym_study] = ACTIONS(2994),
    [anon_sym_tell] = ACTIONS(2994),
    [anon_sym_telldir] = ACTIONS(2994),
    [anon_sym_tied] = ACTIONS(2994),
    [anon_sym_uc] = ACTIONS(2994),
    [anon_sym_ucfirst] = ACTIONS(2994),
    [anon_sym_untie] = ACTIONS(2994),
    [anon_sym_umask] = ACTIONS(2994),
    [anon_sym_values] = ACTIONS(2994),
    [anon_sym_write] = ACTIONS(2994),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2994),
    [anon_sym___END__] = ACTIONS(2994),
    [anon_sym_EOT] = ACTIONS(2994),
    [anon_sym_q] = ACTIONS(2994),
    [anon_sym_qq] = ACTIONS(2994),
    [anon_sym_qw] = ACTIONS(2994),
    [anon_sym_qx] = ACTIONS(2994),
    [anon_sym_qr] = ACTIONS(2994),
    [anon_sym_m] = ACTIONS(2994),
    [anon_sym_s] = ACTIONS(2994),
    [anon_sym_tr] = ACTIONS(2994),
    [anon_sym_y] = ACTIONS(2994),
    [anon_sym_if] = ACTIONS(2994),
    [anon_sym_unless] = ACTIONS(2994),
    [anon_sym_while] = ACTIONS(2994),
    [anon_sym_until] = ACTIONS(2994),
    [aux_sym__bareword_token1] = ACTIONS(2994),
    [sym_number] = ACTIONS(2994),
    [anon_sym_true] = ACTIONS(2994),
    [anon_sym_false] = ACTIONS(2994),
    [sym__single_quote] = ACTIONS(2992),
    [sym__double_quote] = ACTIONS(2992),
    [sym__backtick_quote] = ACTIONS(2992),
    [sym__search_slash_quote] = ACTIONS(2992),
    [sym__open_readline_bracket] = ACTIONS(2992),
    [sym__open_fileglob_bracket] = ACTIONS(2992),
    [sym__PERLY_HEREDOC] = ACTIONS(2992),
    [sym__ctrl_z_hack] = ACTIONS(2992),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2992),
    [sym__filetest] = ACTIONS(2992),
  },
  [STATE(1461)] = {
    [sym_heredoc_content] = STATE(1461),
    [ts_builtin_sym_end] = ACTIONS(2996),
    [sym__identifier] = ACTIONS(2998),
    [anon_sym_LBRACE] = ACTIONS(2998),
    [anon_sym_SEMI] = ACTIONS(2998),
    [anon_sym_package] = ACTIONS(2998),
    [anon_sym_class] = ACTIONS(2998),
    [anon_sym_role] = ACTIONS(2998),
    [anon_sym_BUILD] = ACTIONS(2998),
    [anon_sym_ADJUST] = ACTIONS(2998),
    [anon_sym_DOLLAR] = ACTIONS(2998),
    [anon_sym_AT] = ACTIONS(2998),
    [anon_sym_my] = ACTIONS(2998),
    [anon_sym_extended] = ACTIONS(2998),
    [anon_sym_async] = ACTIONS(2998),
    [anon_sym_sub] = ACTIONS(2998),
    [anon_sym_method] = ACTIONS(2998),
    [anon_sym_LPAREN] = ACTIONS(2998),
    [anon_sym_continue] = ACTIONS(2998),
    [anon_sym_state] = ACTIONS(2998),
    [anon_sym_our] = ACTIONS(2998),
    [anon_sym_try] = ACTIONS(2998),
    [anon_sym_defer] = ACTIONS(2998),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2998),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LT_LT] = ACTIONS(2998),
    [anon_sym_SLASH_SLASH] = ACTIONS(2998),
    [anon_sym_PLUS] = ACTIONS(2998),
    [anon_sym_DASH] = ACTIONS(2998),
    [anon_sym_TILDE] = ACTIONS(2998),
    [anon_sym_BANG] = ACTIONS(2998),
    [anon_sym_PLUS_PLUS] = ACTIONS(2998),
    [anon_sym_DASH_DASH] = ACTIONS(2998),
    [anon_sym_BSLASH] = ACTIONS(2996),
    [anon_sym_do] = ACTIONS(2998),
    [anon_sym_eval] = ACTIONS(2998),
    [anon_sym_await] = ACTIONS(2998),
    [anon_sym_field] = ACTIONS(2998),
    [anon_sym_local] = ACTIONS(2998),
    [anon_sym_dynamically] = ACTIONS(2998),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2998),
    [anon_sym_require] = ACTIONS(2998),
    [anon_sym_map] = ACTIONS(2998),
    [anon_sym_grep] = ACTIONS(2998),
    [anon_sym_sort] = ACTIONS(2998),
    [anon_sym_goto] = ACTIONS(2998),
    [anon_sym_return] = ACTIONS(2998),
    [anon_sym_undef] = ACTIONS(2998),
    [anon_sym_PERCENT2] = ACTIONS(2998),
    [anon_sym_AMP2] = ACTIONS(2998),
    [anon_sym_STAR2] = ACTIONS(2998),
    [anon_sym_use] = ACTIONS(2998),
    [anon_sym_no] = ACTIONS(2998),
    [anon_sym_for] = ACTIONS(2998),
    [anon_sym_foreach] = ACTIONS(2998),
    [anon_sym_last] = ACTIONS(2998),
    [anon_sym_next] = ACTIONS(2998),
    [anon_sym_redo] = ACTIONS(2998),
    [anon_sym_BEGIN] = ACTIONS(2998),
    [anon_sym_INIT] = ACTIONS(2998),
    [anon_sym_CHECK] = ACTIONS(2998),
    [anon_sym_UNITCHECK] = ACTIONS(2998),
    [anon_sym_END] = ACTIONS(2998),
    [anon_sym___FILE__] = ACTIONS(2998),
    [anon_sym___LINE__] = ACTIONS(2998),
    [anon_sym___PACKAGE__] = ACTIONS(2998),
    [anon_sym___SUB__] = ACTIONS(2998),
    [anon_sym_break] = ACTIONS(2998),
    [anon_sym_fork] = ACTIONS(2998),
    [anon_sym_getppid] = ACTIONS(2998),
    [anon_sym_time] = ACTIONS(2998),
    [anon_sym_times] = ACTIONS(2998),
    [anon_sym_wait] = ACTIONS(2998),
    [anon_sym_wantarray] = ACTIONS(2998),
    [anon_sym_abs] = ACTIONS(2998),
    [anon_sym_alarm] = ACTIONS(2998),
    [anon_sym_chop] = ACTIONS(2998),
    [anon_sym_chdir] = ACTIONS(2998),
    [anon_sym_close] = ACTIONS(2998),
    [anon_sym_closedir] = ACTIONS(2998),
    [anon_sym_caller] = ACTIONS(2998),
    [anon_sym_chomp] = ACTIONS(2998),
    [anon_sym_chr] = ACTIONS(2998),
    [anon_sym_cos] = ACTIONS(2998),
    [anon_sym_chroot] = ACTIONS(2998),
    [anon_sym_defined] = ACTIONS(2998),
    [anon_sym_delete] = ACTIONS(2998),
    [anon_sym_dbmclose] = ACTIONS(2998),
    [anon_sym_exists] = ACTIONS(2998),
    [anon_sym_exit] = ACTIONS(2998),
    [anon_sym_eof] = ACTIONS(2998),
    [anon_sym_exp] = ACTIONS(2998),
    [anon_sym_each] = ACTIONS(2998),
    [anon_sym_fc] = ACTIONS(2998),
    [anon_sym_fileno] = ACTIONS(2998),
    [anon_sym_gmtime] = ACTIONS(2998),
    [anon_sym_getc] = ACTIONS(2998),
    [anon_sym_getpgrp] = ACTIONS(2998),
    [anon_sym_getprotobyname] = ACTIONS(2998),
    [anon_sym_getpwname] = ACTIONS(2998),
    [anon_sym_getpwuid] = ACTIONS(2998),
    [anon_sym_getpeername] = ACTIONS(2998),
    [anon_sym_getnetbyname] = ACTIONS(2998),
    [anon_sym_getsockname] = ACTIONS(2998),
    [anon_sym_getgrnam] = ACTIONS(2998),
    [anon_sym_getgrgid] = ACTIONS(2998),
    [anon_sym_hex] = ACTIONS(2998),
    [anon_sym_int] = ACTIONS(2998),
    [anon_sym_keys] = ACTIONS(2998),
    [anon_sym_lc] = ACTIONS(2998),
    [anon_sym_lcfirst] = ACTIONS(2998),
    [anon_sym_length] = ACTIONS(2998),
    [anon_sym_localtime] = ACTIONS(2998),
    [anon_sym_log] = ACTIONS(2998),
    [anon_sym_lock] = ACTIONS(2998),
    [anon_sym_lstat] = ACTIONS(2998),
    [anon_sym_oct] = ACTIONS(2998),
    [anon_sym_ord] = ACTIONS(2998),
    [anon_sym_prototype] = ACTIONS(2998),
    [anon_sym_pop] = ACTIONS(2998),
    [anon_sym_pos] = ACTIONS(2998),
    [anon_sym_quotemeta] = ACTIONS(2998),
    [anon_sym_reset] = ACTIONS(2998),
    [anon_sym_rand] = ACTIONS(2998),
    [anon_sym_rmdir] = ACTIONS(2998),
    [anon_sym_readdir] = ACTIONS(2998),
    [anon_sym_readline] = ACTIONS(2998),
    [anon_sym_readpipe] = ACTIONS(2998),
    [anon_sym_rewinddir] = ACTIONS(2998),
    [anon_sym_readlink] = ACTIONS(2998),
    [anon_sym_ref] = ACTIONS(2998),
    [anon_sym_scalar] = ACTIONS(2998),
    [anon_sym_shift] = ACTIONS(2998),
    [anon_sym_sin] = ACTIONS(2998),
    [anon_sym_sleep] = ACTIONS(2998),
    [anon_sym_sqrt] = ACTIONS(2998),
    [anon_sym_srand] = ACTIONS(2998),
    [anon_sym_stat] = ACTIONS(2998),
    [anon_sym_study] = ACTIONS(2998),
    [anon_sym_tell] = ACTIONS(2998),
    [anon_sym_telldir] = ACTIONS(2998),
    [anon_sym_tied] = ACTIONS(2998),
    [anon_sym_uc] = ACTIONS(2998),
    [anon_sym_ucfirst] = ACTIONS(2998),
    [anon_sym_untie] = ACTIONS(2998),
    [anon_sym_umask] = ACTIONS(2998),
    [anon_sym_values] = ACTIONS(2998),
    [anon_sym_write] = ACTIONS(2998),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2998),
    [anon_sym___END__] = ACTIONS(2998),
    [anon_sym_EOT] = ACTIONS(2998),
    [anon_sym_q] = ACTIONS(2998),
    [anon_sym_qq] = ACTIONS(2998),
    [anon_sym_qw] = ACTIONS(2998),
    [anon_sym_qx] = ACTIONS(2998),
    [anon_sym_qr] = ACTIONS(2998),
    [anon_sym_m] = ACTIONS(2998),
    [anon_sym_s] = ACTIONS(2998),
    [anon_sym_tr] = ACTIONS(2998),
    [anon_sym_y] = ACTIONS(2998),
    [anon_sym_if] = ACTIONS(2998),
    [anon_sym_unless] = ACTIONS(2998),
    [anon_sym_while] = ACTIONS(2998),
    [anon_sym_until] = ACTIONS(2998),
    [aux_sym__bareword_token1] = ACTIONS(2998),
    [sym_number] = ACTIONS(2998),
    [anon_sym_true] = ACTIONS(2998),
    [anon_sym_false] = ACTIONS(2998),
    [sym__single_quote] = ACTIONS(2996),
    [sym__double_quote] = ACTIONS(2996),
    [sym__backtick_quote] = ACTIONS(2996),
    [sym__search_slash_quote] = ACTIONS(2996),
    [sym__open_readline_bracket] = ACTIONS(2996),
    [sym__open_fileglob_bracket] = ACTIONS(2996),
    [sym__PERLY_HEREDOC] = ACTIONS(2996),
    [sym__ctrl_z_hack] = ACTIONS(2996),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2996),
    [sym__filetest] = ACTIONS(2996),
  },
  [STATE(1462)] = {
    [sym_heredoc_content] = STATE(1462),
    [ts_builtin_sym_end] = ACTIONS(3000),
    [sym__identifier] = ACTIONS(3002),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_SEMI] = ACTIONS(3002),
    [anon_sym_package] = ACTIONS(3002),
    [anon_sym_class] = ACTIONS(3002),
    [anon_sym_role] = ACTIONS(3002),
    [anon_sym_BUILD] = ACTIONS(3002),
    [anon_sym_ADJUST] = ACTIONS(3002),
    [anon_sym_DOLLAR] = ACTIONS(3002),
    [anon_sym_AT] = ACTIONS(3002),
    [anon_sym_my] = ACTIONS(3002),
    [anon_sym_extended] = ACTIONS(3002),
    [anon_sym_async] = ACTIONS(3002),
    [anon_sym_sub] = ACTIONS(3002),
    [anon_sym_method] = ACTIONS(3002),
    [anon_sym_LPAREN] = ACTIONS(3002),
    [anon_sym_continue] = ACTIONS(3002),
    [anon_sym_state] = ACTIONS(3002),
    [anon_sym_our] = ACTIONS(3002),
    [anon_sym_try] = ACTIONS(3002),
    [anon_sym_defer] = ACTIONS(3002),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3002),
    [anon_sym_LBRACK] = ACTIONS(3002),
    [anon_sym_LT_LT] = ACTIONS(3002),
    [anon_sym_SLASH_SLASH] = ACTIONS(3002),
    [anon_sym_PLUS] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3002),
    [anon_sym_TILDE] = ACTIONS(3002),
    [anon_sym_BANG] = ACTIONS(3002),
    [anon_sym_PLUS_PLUS] = ACTIONS(3002),
    [anon_sym_DASH_DASH] = ACTIONS(3002),
    [anon_sym_BSLASH] = ACTIONS(3000),
    [anon_sym_do] = ACTIONS(3002),
    [anon_sym_eval] = ACTIONS(3002),
    [anon_sym_await] = ACTIONS(3002),
    [anon_sym_field] = ACTIONS(3002),
    [anon_sym_local] = ACTIONS(3002),
    [anon_sym_dynamically] = ACTIONS(3002),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3002),
    [anon_sym_require] = ACTIONS(3002),
    [anon_sym_map] = ACTIONS(3002),
    [anon_sym_grep] = ACTIONS(3002),
    [anon_sym_sort] = ACTIONS(3002),
    [anon_sym_goto] = ACTIONS(3002),
    [anon_sym_return] = ACTIONS(3002),
    [anon_sym_undef] = ACTIONS(3002),
    [anon_sym_PERCENT2] = ACTIONS(3002),
    [anon_sym_AMP2] = ACTIONS(3002),
    [anon_sym_STAR2] = ACTIONS(3002),
    [anon_sym_use] = ACTIONS(3002),
    [anon_sym_no] = ACTIONS(3002),
    [anon_sym_for] = ACTIONS(3002),
    [anon_sym_foreach] = ACTIONS(3002),
    [anon_sym_last] = ACTIONS(3002),
    [anon_sym_next] = ACTIONS(3002),
    [anon_sym_redo] = ACTIONS(3002),
    [anon_sym_BEGIN] = ACTIONS(3002),
    [anon_sym_INIT] = ACTIONS(3002),
    [anon_sym_CHECK] = ACTIONS(3002),
    [anon_sym_UNITCHECK] = ACTIONS(3002),
    [anon_sym_END] = ACTIONS(3002),
    [anon_sym___FILE__] = ACTIONS(3002),
    [anon_sym___LINE__] = ACTIONS(3002),
    [anon_sym___PACKAGE__] = ACTIONS(3002),
    [anon_sym___SUB__] = ACTIONS(3002),
    [anon_sym_break] = ACTIONS(3002),
    [anon_sym_fork] = ACTIONS(3002),
    [anon_sym_getppid] = ACTIONS(3002),
    [anon_sym_time] = ACTIONS(3002),
    [anon_sym_times] = ACTIONS(3002),
    [anon_sym_wait] = ACTIONS(3002),
    [anon_sym_wantarray] = ACTIONS(3002),
    [anon_sym_abs] = ACTIONS(3002),
    [anon_sym_alarm] = ACTIONS(3002),
    [anon_sym_chop] = ACTIONS(3002),
    [anon_sym_chdir] = ACTIONS(3002),
    [anon_sym_close] = ACTIONS(3002),
    [anon_sym_closedir] = ACTIONS(3002),
    [anon_sym_caller] = ACTIONS(3002),
    [anon_sym_chomp] = ACTIONS(3002),
    [anon_sym_chr] = ACTIONS(3002),
    [anon_sym_cos] = ACTIONS(3002),
    [anon_sym_chroot] = ACTIONS(3002),
    [anon_sym_defined] = ACTIONS(3002),
    [anon_sym_delete] = ACTIONS(3002),
    [anon_sym_dbmclose] = ACTIONS(3002),
    [anon_sym_exists] = ACTIONS(3002),
    [anon_sym_exit] = ACTIONS(3002),
    [anon_sym_eof] = ACTIONS(3002),
    [anon_sym_exp] = ACTIONS(3002),
    [anon_sym_each] = ACTIONS(3002),
    [anon_sym_fc] = ACTIONS(3002),
    [anon_sym_fileno] = ACTIONS(3002),
    [anon_sym_gmtime] = ACTIONS(3002),
    [anon_sym_getc] = ACTIONS(3002),
    [anon_sym_getpgrp] = ACTIONS(3002),
    [anon_sym_getprotobyname] = ACTIONS(3002),
    [anon_sym_getpwname] = ACTIONS(3002),
    [anon_sym_getpwuid] = ACTIONS(3002),
    [anon_sym_getpeername] = ACTIONS(3002),
    [anon_sym_getnetbyname] = ACTIONS(3002),
    [anon_sym_getsockname] = ACTIONS(3002),
    [anon_sym_getgrnam] = ACTIONS(3002),
    [anon_sym_getgrgid] = ACTIONS(3002),
    [anon_sym_hex] = ACTIONS(3002),
    [anon_sym_int] = ACTIONS(3002),
    [anon_sym_keys] = ACTIONS(3002),
    [anon_sym_lc] = ACTIONS(3002),
    [anon_sym_lcfirst] = ACTIONS(3002),
    [anon_sym_length] = ACTIONS(3002),
    [anon_sym_localtime] = ACTIONS(3002),
    [anon_sym_log] = ACTIONS(3002),
    [anon_sym_lock] = ACTIONS(3002),
    [anon_sym_lstat] = ACTIONS(3002),
    [anon_sym_oct] = ACTIONS(3002),
    [anon_sym_ord] = ACTIONS(3002),
    [anon_sym_prototype] = ACTIONS(3002),
    [anon_sym_pop] = ACTIONS(3002),
    [anon_sym_pos] = ACTIONS(3002),
    [anon_sym_quotemeta] = ACTIONS(3002),
    [anon_sym_reset] = ACTIONS(3002),
    [anon_sym_rand] = ACTIONS(3002),
    [anon_sym_rmdir] = ACTIONS(3002),
    [anon_sym_readdir] = ACTIONS(3002),
    [anon_sym_readline] = ACTIONS(3002),
    [anon_sym_readpipe] = ACTIONS(3002),
    [anon_sym_rewinddir] = ACTIONS(3002),
    [anon_sym_readlink] = ACTIONS(3002),
    [anon_sym_ref] = ACTIONS(3002),
    [anon_sym_scalar] = ACTIONS(3002),
    [anon_sym_shift] = ACTIONS(3002),
    [anon_sym_sin] = ACTIONS(3002),
    [anon_sym_sleep] = ACTIONS(3002),
    [anon_sym_sqrt] = ACTIONS(3002),
    [anon_sym_srand] = ACTIONS(3002),
    [anon_sym_stat] = ACTIONS(3002),
    [anon_sym_study] = ACTIONS(3002),
    [anon_sym_tell] = ACTIONS(3002),
    [anon_sym_telldir] = ACTIONS(3002),
    [anon_sym_tied] = ACTIONS(3002),
    [anon_sym_uc] = ACTIONS(3002),
    [anon_sym_ucfirst] = ACTIONS(3002),
    [anon_sym_untie] = ACTIONS(3002),
    [anon_sym_umask] = ACTIONS(3002),
    [anon_sym_values] = ACTIONS(3002),
    [anon_sym_write] = ACTIONS(3002),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3002),
    [anon_sym___END__] = ACTIONS(3002),
    [anon_sym_EOT] = ACTIONS(3002),
    [anon_sym_q] = ACTIONS(3002),
    [anon_sym_qq] = ACTIONS(3002),
    [anon_sym_qw] = ACTIONS(3002),
    [anon_sym_qx] = ACTIONS(3002),
    [anon_sym_qr] = ACTIONS(3002),
    [anon_sym_m] = ACTIONS(3002),
    [anon_sym_s] = ACTIONS(3002),
    [anon_sym_tr] = ACTIONS(3002),
    [anon_sym_y] = ACTIONS(3002),
    [anon_sym_if] = ACTIONS(3002),
    [anon_sym_unless] = ACTIONS(3002),
    [anon_sym_while] = ACTIONS(3002),
    [anon_sym_until] = ACTIONS(3002),
    [aux_sym__bareword_token1] = ACTIONS(3002),
    [sym_number] = ACTIONS(3002),
    [anon_sym_true] = ACTIONS(3002),
    [anon_sym_false] = ACTIONS(3002),
    [sym__single_quote] = ACTIONS(3000),
    [sym__double_quote] = ACTIONS(3000),
    [sym__backtick_quote] = ACTIONS(3000),
    [sym__search_slash_quote] = ACTIONS(3000),
    [sym__open_readline_bracket] = ACTIONS(3000),
    [sym__open_fileglob_bracket] = ACTIONS(3000),
    [sym__PERLY_HEREDOC] = ACTIONS(3000),
    [sym__ctrl_z_hack] = ACTIONS(3000),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3000),
    [sym__filetest] = ACTIONS(3000),
  },
  [STATE(1463)] = {
    [sym_heredoc_content] = STATE(1463),
    [ts_builtin_sym_end] = ACTIONS(3004),
    [sym__identifier] = ACTIONS(3006),
    [anon_sym_LBRACE] = ACTIONS(3006),
    [anon_sym_SEMI] = ACTIONS(3006),
    [anon_sym_package] = ACTIONS(3006),
    [anon_sym_class] = ACTIONS(3006),
    [anon_sym_role] = ACTIONS(3006),
    [anon_sym_BUILD] = ACTIONS(3006),
    [anon_sym_ADJUST] = ACTIONS(3006),
    [anon_sym_DOLLAR] = ACTIONS(3006),
    [anon_sym_AT] = ACTIONS(3006),
    [anon_sym_my] = ACTIONS(3006),
    [anon_sym_extended] = ACTIONS(3006),
    [anon_sym_async] = ACTIONS(3006),
    [anon_sym_sub] = ACTIONS(3006),
    [anon_sym_method] = ACTIONS(3006),
    [anon_sym_LPAREN] = ACTIONS(3006),
    [anon_sym_continue] = ACTIONS(3006),
    [anon_sym_state] = ACTIONS(3006),
    [anon_sym_our] = ACTIONS(3006),
    [anon_sym_try] = ACTIONS(3006),
    [anon_sym_defer] = ACTIONS(3006),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3006),
    [anon_sym_LBRACK] = ACTIONS(3006),
    [anon_sym_LT_LT] = ACTIONS(3006),
    [anon_sym_SLASH_SLASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_TILDE] = ACTIONS(3006),
    [anon_sym_BANG] = ACTIONS(3006),
    [anon_sym_PLUS_PLUS] = ACTIONS(3006),
    [anon_sym_DASH_DASH] = ACTIONS(3006),
    [anon_sym_BSLASH] = ACTIONS(3004),
    [anon_sym_do] = ACTIONS(3006),
    [anon_sym_eval] = ACTIONS(3006),
    [anon_sym_await] = ACTIONS(3006),
    [anon_sym_field] = ACTIONS(3006),
    [anon_sym_local] = ACTIONS(3006),
    [anon_sym_dynamically] = ACTIONS(3006),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3006),
    [anon_sym_require] = ACTIONS(3006),
    [anon_sym_map] = ACTIONS(3006),
    [anon_sym_grep] = ACTIONS(3006),
    [anon_sym_sort] = ACTIONS(3006),
    [anon_sym_goto] = ACTIONS(3006),
    [anon_sym_return] = ACTIONS(3006),
    [anon_sym_undef] = ACTIONS(3006),
    [anon_sym_PERCENT2] = ACTIONS(3006),
    [anon_sym_AMP2] = ACTIONS(3006),
    [anon_sym_STAR2] = ACTIONS(3006),
    [anon_sym_use] = ACTIONS(3006),
    [anon_sym_no] = ACTIONS(3006),
    [anon_sym_for] = ACTIONS(3006),
    [anon_sym_foreach] = ACTIONS(3006),
    [anon_sym_last] = ACTIONS(3006),
    [anon_sym_next] = ACTIONS(3006),
    [anon_sym_redo] = ACTIONS(3006),
    [anon_sym_BEGIN] = ACTIONS(3006),
    [anon_sym_INIT] = ACTIONS(3006),
    [anon_sym_CHECK] = ACTIONS(3006),
    [anon_sym_UNITCHECK] = ACTIONS(3006),
    [anon_sym_END] = ACTIONS(3006),
    [anon_sym___FILE__] = ACTIONS(3006),
    [anon_sym___LINE__] = ACTIONS(3006),
    [anon_sym___PACKAGE__] = ACTIONS(3006),
    [anon_sym___SUB__] = ACTIONS(3006),
    [anon_sym_break] = ACTIONS(3006),
    [anon_sym_fork] = ACTIONS(3006),
    [anon_sym_getppid] = ACTIONS(3006),
    [anon_sym_time] = ACTIONS(3006),
    [anon_sym_times] = ACTIONS(3006),
    [anon_sym_wait] = ACTIONS(3006),
    [anon_sym_wantarray] = ACTIONS(3006),
    [anon_sym_abs] = ACTIONS(3006),
    [anon_sym_alarm] = ACTIONS(3006),
    [anon_sym_chop] = ACTIONS(3006),
    [anon_sym_chdir] = ACTIONS(3006),
    [anon_sym_close] = ACTIONS(3006),
    [anon_sym_closedir] = ACTIONS(3006),
    [anon_sym_caller] = ACTIONS(3006),
    [anon_sym_chomp] = ACTIONS(3006),
    [anon_sym_chr] = ACTIONS(3006),
    [anon_sym_cos] = ACTIONS(3006),
    [anon_sym_chroot] = ACTIONS(3006),
    [anon_sym_defined] = ACTIONS(3006),
    [anon_sym_delete] = ACTIONS(3006),
    [anon_sym_dbmclose] = ACTIONS(3006),
    [anon_sym_exists] = ACTIONS(3006),
    [anon_sym_exit] = ACTIONS(3006),
    [anon_sym_eof] = ACTIONS(3006),
    [anon_sym_exp] = ACTIONS(3006),
    [anon_sym_each] = ACTIONS(3006),
    [anon_sym_fc] = ACTIONS(3006),
    [anon_sym_fileno] = ACTIONS(3006),
    [anon_sym_gmtime] = ACTIONS(3006),
    [anon_sym_getc] = ACTIONS(3006),
    [anon_sym_getpgrp] = ACTIONS(3006),
    [anon_sym_getprotobyname] = ACTIONS(3006),
    [anon_sym_getpwname] = ACTIONS(3006),
    [anon_sym_getpwuid] = ACTIONS(3006),
    [anon_sym_getpeername] = ACTIONS(3006),
    [anon_sym_getnetbyname] = ACTIONS(3006),
    [anon_sym_getsockname] = ACTIONS(3006),
    [anon_sym_getgrnam] = ACTIONS(3006),
    [anon_sym_getgrgid] = ACTIONS(3006),
    [anon_sym_hex] = ACTIONS(3006),
    [anon_sym_int] = ACTIONS(3006),
    [anon_sym_keys] = ACTIONS(3006),
    [anon_sym_lc] = ACTIONS(3006),
    [anon_sym_lcfirst] = ACTIONS(3006),
    [anon_sym_length] = ACTIONS(3006),
    [anon_sym_localtime] = ACTIONS(3006),
    [anon_sym_log] = ACTIONS(3006),
    [anon_sym_lock] = ACTIONS(3006),
    [anon_sym_lstat] = ACTIONS(3006),
    [anon_sym_oct] = ACTIONS(3006),
    [anon_sym_ord] = ACTIONS(3006),
    [anon_sym_prototype] = ACTIONS(3006),
    [anon_sym_pop] = ACTIONS(3006),
    [anon_sym_pos] = ACTIONS(3006),
    [anon_sym_quotemeta] = ACTIONS(3006),
    [anon_sym_reset] = ACTIONS(3006),
    [anon_sym_rand] = ACTIONS(3006),
    [anon_sym_rmdir] = ACTIONS(3006),
    [anon_sym_readdir] = ACTIONS(3006),
    [anon_sym_readline] = ACTIONS(3006),
    [anon_sym_readpipe] = ACTIONS(3006),
    [anon_sym_rewinddir] = ACTIONS(3006),
    [anon_sym_readlink] = ACTIONS(3006),
    [anon_sym_ref] = ACTIONS(3006),
    [anon_sym_scalar] = ACTIONS(3006),
    [anon_sym_shift] = ACTIONS(3006),
    [anon_sym_sin] = ACTIONS(3006),
    [anon_sym_sleep] = ACTIONS(3006),
    [anon_sym_sqrt] = ACTIONS(3006),
    [anon_sym_srand] = ACTIONS(3006),
    [anon_sym_stat] = ACTIONS(3006),
    [anon_sym_study] = ACTIONS(3006),
    [anon_sym_tell] = ACTIONS(3006),
    [anon_sym_telldir] = ACTIONS(3006),
    [anon_sym_tied] = ACTIONS(3006),
    [anon_sym_uc] = ACTIONS(3006),
    [anon_sym_ucfirst] = ACTIONS(3006),
    [anon_sym_untie] = ACTIONS(3006),
    [anon_sym_umask] = ACTIONS(3006),
    [anon_sym_values] = ACTIONS(3006),
    [anon_sym_write] = ACTIONS(3006),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3006),
    [anon_sym___END__] = ACTIONS(3006),
    [anon_sym_EOT] = ACTIONS(3006),
    [anon_sym_q] = ACTIONS(3006),
    [anon_sym_qq] = ACTIONS(3006),
    [anon_sym_qw] = ACTIONS(3006),
    [anon_sym_qx] = ACTIONS(3006),
    [anon_sym_qr] = ACTIONS(3006),
    [anon_sym_m] = ACTIONS(3006),
    [anon_sym_s] = ACTIONS(3006),
    [anon_sym_tr] = ACTIONS(3006),
    [anon_sym_y] = ACTIONS(3006),
    [anon_sym_if] = ACTIONS(3006),
    [anon_sym_unless] = ACTIONS(3006),
    [anon_sym_while] = ACTIONS(3006),
    [anon_sym_until] = ACTIONS(3006),
    [aux_sym__bareword_token1] = ACTIONS(3006),
    [sym_number] = ACTIONS(3006),
    [anon_sym_true] = ACTIONS(3006),
    [anon_sym_false] = ACTIONS(3006),
    [sym__single_quote] = ACTIONS(3004),
    [sym__double_quote] = ACTIONS(3004),
    [sym__backtick_quote] = ACTIONS(3004),
    [sym__search_slash_quote] = ACTIONS(3004),
    [sym__open_readline_bracket] = ACTIONS(3004),
    [sym__open_fileglob_bracket] = ACTIONS(3004),
    [sym__PERLY_HEREDOC] = ACTIONS(3004),
    [sym__ctrl_z_hack] = ACTIONS(3004),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3004),
    [sym__filetest] = ACTIONS(3004),
  },
  [STATE(1464)] = {
    [sym_heredoc_content] = STATE(1464),
    [ts_builtin_sym_end] = ACTIONS(3008),
    [sym__identifier] = ACTIONS(3010),
    [anon_sym_LBRACE] = ACTIONS(3010),
    [anon_sym_SEMI] = ACTIONS(3010),
    [anon_sym_package] = ACTIONS(3010),
    [anon_sym_class] = ACTIONS(3010),
    [anon_sym_role] = ACTIONS(3010),
    [anon_sym_BUILD] = ACTIONS(3010),
    [anon_sym_ADJUST] = ACTIONS(3010),
    [anon_sym_DOLLAR] = ACTIONS(3010),
    [anon_sym_AT] = ACTIONS(3010),
    [anon_sym_my] = ACTIONS(3010),
    [anon_sym_extended] = ACTIONS(3010),
    [anon_sym_async] = ACTIONS(3010),
    [anon_sym_sub] = ACTIONS(3010),
    [anon_sym_method] = ACTIONS(3010),
    [anon_sym_LPAREN] = ACTIONS(3010),
    [anon_sym_continue] = ACTIONS(3010),
    [anon_sym_state] = ACTIONS(3010),
    [anon_sym_our] = ACTIONS(3010),
    [anon_sym_try] = ACTIONS(3010),
    [anon_sym_defer] = ACTIONS(3010),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3010),
    [anon_sym_LBRACK] = ACTIONS(3010),
    [anon_sym_LT_LT] = ACTIONS(3010),
    [anon_sym_SLASH_SLASH] = ACTIONS(3010),
    [anon_sym_PLUS] = ACTIONS(3010),
    [anon_sym_DASH] = ACTIONS(3010),
    [anon_sym_TILDE] = ACTIONS(3010),
    [anon_sym_BANG] = ACTIONS(3010),
    [anon_sym_PLUS_PLUS] = ACTIONS(3010),
    [anon_sym_DASH_DASH] = ACTIONS(3010),
    [anon_sym_BSLASH] = ACTIONS(3008),
    [anon_sym_do] = ACTIONS(3010),
    [anon_sym_eval] = ACTIONS(3010),
    [anon_sym_await] = ACTIONS(3010),
    [anon_sym_field] = ACTIONS(3010),
    [anon_sym_local] = ACTIONS(3010),
    [anon_sym_dynamically] = ACTIONS(3010),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3010),
    [anon_sym_require] = ACTIONS(3010),
    [anon_sym_map] = ACTIONS(3010),
    [anon_sym_grep] = ACTIONS(3010),
    [anon_sym_sort] = ACTIONS(3010),
    [anon_sym_goto] = ACTIONS(3010),
    [anon_sym_return] = ACTIONS(3010),
    [anon_sym_undef] = ACTIONS(3010),
    [anon_sym_PERCENT2] = ACTIONS(3010),
    [anon_sym_AMP2] = ACTIONS(3010),
    [anon_sym_STAR2] = ACTIONS(3010),
    [anon_sym_use] = ACTIONS(3010),
    [anon_sym_no] = ACTIONS(3010),
    [anon_sym_for] = ACTIONS(3010),
    [anon_sym_foreach] = ACTIONS(3010),
    [anon_sym_last] = ACTIONS(3010),
    [anon_sym_next] = ACTIONS(3010),
    [anon_sym_redo] = ACTIONS(3010),
    [anon_sym_BEGIN] = ACTIONS(3010),
    [anon_sym_INIT] = ACTIONS(3010),
    [anon_sym_CHECK] = ACTIONS(3010),
    [anon_sym_UNITCHECK] = ACTIONS(3010),
    [anon_sym_END] = ACTIONS(3010),
    [anon_sym___FILE__] = ACTIONS(3010),
    [anon_sym___LINE__] = ACTIONS(3010),
    [anon_sym___PACKAGE__] = ACTIONS(3010),
    [anon_sym___SUB__] = ACTIONS(3010),
    [anon_sym_break] = ACTIONS(3010),
    [anon_sym_fork] = ACTIONS(3010),
    [anon_sym_getppid] = ACTIONS(3010),
    [anon_sym_time] = ACTIONS(3010),
    [anon_sym_times] = ACTIONS(3010),
    [anon_sym_wait] = ACTIONS(3010),
    [anon_sym_wantarray] = ACTIONS(3010),
    [anon_sym_abs] = ACTIONS(3010),
    [anon_sym_alarm] = ACTIONS(3010),
    [anon_sym_chop] = ACTIONS(3010),
    [anon_sym_chdir] = ACTIONS(3010),
    [anon_sym_close] = ACTIONS(3010),
    [anon_sym_closedir] = ACTIONS(3010),
    [anon_sym_caller] = ACTIONS(3010),
    [anon_sym_chomp] = ACTIONS(3010),
    [anon_sym_chr] = ACTIONS(3010),
    [anon_sym_cos] = ACTIONS(3010),
    [anon_sym_chroot] = ACTIONS(3010),
    [anon_sym_defined] = ACTIONS(3010),
    [anon_sym_delete] = ACTIONS(3010),
    [anon_sym_dbmclose] = ACTIONS(3010),
    [anon_sym_exists] = ACTIONS(3010),
    [anon_sym_exit] = ACTIONS(3010),
    [anon_sym_eof] = ACTIONS(3010),
    [anon_sym_exp] = ACTIONS(3010),
    [anon_sym_each] = ACTIONS(3010),
    [anon_sym_fc] = ACTIONS(3010),
    [anon_sym_fileno] = ACTIONS(3010),
    [anon_sym_gmtime] = ACTIONS(3010),
    [anon_sym_getc] = ACTIONS(3010),
    [anon_sym_getpgrp] = ACTIONS(3010),
    [anon_sym_getprotobyname] = ACTIONS(3010),
    [anon_sym_getpwname] = ACTIONS(3010),
    [anon_sym_getpwuid] = ACTIONS(3010),
    [anon_sym_getpeername] = ACTIONS(3010),
    [anon_sym_getnetbyname] = ACTIONS(3010),
    [anon_sym_getsockname] = ACTIONS(3010),
    [anon_sym_getgrnam] = ACTIONS(3010),
    [anon_sym_getgrgid] = ACTIONS(3010),
    [anon_sym_hex] = ACTIONS(3010),
    [anon_sym_int] = ACTIONS(3010),
    [anon_sym_keys] = ACTIONS(3010),
    [anon_sym_lc] = ACTIONS(3010),
    [anon_sym_lcfirst] = ACTIONS(3010),
    [anon_sym_length] = ACTIONS(3010),
    [anon_sym_localtime] = ACTIONS(3010),
    [anon_sym_log] = ACTIONS(3010),
    [anon_sym_lock] = ACTIONS(3010),
    [anon_sym_lstat] = ACTIONS(3010),
    [anon_sym_oct] = ACTIONS(3010),
    [anon_sym_ord] = ACTIONS(3010),
    [anon_sym_prototype] = ACTIONS(3010),
    [anon_sym_pop] = ACTIONS(3010),
    [anon_sym_pos] = ACTIONS(3010),
    [anon_sym_quotemeta] = ACTIONS(3010),
    [anon_sym_reset] = ACTIONS(3010),
    [anon_sym_rand] = ACTIONS(3010),
    [anon_sym_rmdir] = ACTIONS(3010),
    [anon_sym_readdir] = ACTIONS(3010),
    [anon_sym_readline] = ACTIONS(3010),
    [anon_sym_readpipe] = ACTIONS(3010),
    [anon_sym_rewinddir] = ACTIONS(3010),
    [anon_sym_readlink] = ACTIONS(3010),
    [anon_sym_ref] = ACTIONS(3010),
    [anon_sym_scalar] = ACTIONS(3010),
    [anon_sym_shift] = ACTIONS(3010),
    [anon_sym_sin] = ACTIONS(3010),
    [anon_sym_sleep] = ACTIONS(3010),
    [anon_sym_sqrt] = ACTIONS(3010),
    [anon_sym_srand] = ACTIONS(3010),
    [anon_sym_stat] = ACTIONS(3010),
    [anon_sym_study] = ACTIONS(3010),
    [anon_sym_tell] = ACTIONS(3010),
    [anon_sym_telldir] = ACTIONS(3010),
    [anon_sym_tied] = ACTIONS(3010),
    [anon_sym_uc] = ACTIONS(3010),
    [anon_sym_ucfirst] = ACTIONS(3010),
    [anon_sym_untie] = ACTIONS(3010),
    [anon_sym_umask] = ACTIONS(3010),
    [anon_sym_values] = ACTIONS(3010),
    [anon_sym_write] = ACTIONS(3010),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3010),
    [anon_sym___END__] = ACTIONS(3010),
    [anon_sym_EOT] = ACTIONS(3010),
    [anon_sym_q] = ACTIONS(3010),
    [anon_sym_qq] = ACTIONS(3010),
    [anon_sym_qw] = ACTIONS(3010),
    [anon_sym_qx] = ACTIONS(3010),
    [anon_sym_qr] = ACTIONS(3010),
    [anon_sym_m] = ACTIONS(3010),
    [anon_sym_s] = ACTIONS(3010),
    [anon_sym_tr] = ACTIONS(3010),
    [anon_sym_y] = ACTIONS(3010),
    [anon_sym_if] = ACTIONS(3010),
    [anon_sym_unless] = ACTIONS(3010),
    [anon_sym_while] = ACTIONS(3010),
    [anon_sym_until] = ACTIONS(3010),
    [aux_sym__bareword_token1] = ACTIONS(3010),
    [sym_number] = ACTIONS(3010),
    [anon_sym_true] = ACTIONS(3010),
    [anon_sym_false] = ACTIONS(3010),
    [sym__single_quote] = ACTIONS(3008),
    [sym__double_quote] = ACTIONS(3008),
    [sym__backtick_quote] = ACTIONS(3008),
    [sym__search_slash_quote] = ACTIONS(3008),
    [sym__open_readline_bracket] = ACTIONS(3008),
    [sym__open_fileglob_bracket] = ACTIONS(3008),
    [sym__PERLY_HEREDOC] = ACTIONS(3008),
    [sym__ctrl_z_hack] = ACTIONS(3008),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3008),
    [sym__filetest] = ACTIONS(3008),
  },
  [STATE(1465)] = {
    [sym_heredoc_content] = STATE(1465),
    [ts_builtin_sym_end] = ACTIONS(3012),
    [sym__identifier] = ACTIONS(3014),
    [anon_sym_LBRACE] = ACTIONS(3014),
    [anon_sym_SEMI] = ACTIONS(3014),
    [anon_sym_package] = ACTIONS(3014),
    [anon_sym_class] = ACTIONS(3014),
    [anon_sym_role] = ACTIONS(3014),
    [anon_sym_BUILD] = ACTIONS(3014),
    [anon_sym_ADJUST] = ACTIONS(3014),
    [anon_sym_DOLLAR] = ACTIONS(3014),
    [anon_sym_AT] = ACTIONS(3014),
    [anon_sym_my] = ACTIONS(3014),
    [anon_sym_extended] = ACTIONS(3014),
    [anon_sym_async] = ACTIONS(3014),
    [anon_sym_sub] = ACTIONS(3014),
    [anon_sym_method] = ACTIONS(3014),
    [anon_sym_LPAREN] = ACTIONS(3014),
    [anon_sym_continue] = ACTIONS(3014),
    [anon_sym_state] = ACTIONS(3014),
    [anon_sym_our] = ACTIONS(3014),
    [anon_sym_try] = ACTIONS(3014),
    [anon_sym_defer] = ACTIONS(3014),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3014),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym_LT_LT] = ACTIONS(3014),
    [anon_sym_SLASH_SLASH] = ACTIONS(3014),
    [anon_sym_PLUS] = ACTIONS(3014),
    [anon_sym_DASH] = ACTIONS(3014),
    [anon_sym_TILDE] = ACTIONS(3014),
    [anon_sym_BANG] = ACTIONS(3014),
    [anon_sym_PLUS_PLUS] = ACTIONS(3014),
    [anon_sym_DASH_DASH] = ACTIONS(3014),
    [anon_sym_BSLASH] = ACTIONS(3012),
    [anon_sym_do] = ACTIONS(3014),
    [anon_sym_eval] = ACTIONS(3014),
    [anon_sym_await] = ACTIONS(3014),
    [anon_sym_field] = ACTIONS(3014),
    [anon_sym_local] = ACTIONS(3014),
    [anon_sym_dynamically] = ACTIONS(3014),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3014),
    [anon_sym_require] = ACTIONS(3014),
    [anon_sym_map] = ACTIONS(3014),
    [anon_sym_grep] = ACTIONS(3014),
    [anon_sym_sort] = ACTIONS(3014),
    [anon_sym_goto] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_undef] = ACTIONS(3014),
    [anon_sym_PERCENT2] = ACTIONS(3014),
    [anon_sym_AMP2] = ACTIONS(3014),
    [anon_sym_STAR2] = ACTIONS(3014),
    [anon_sym_use] = ACTIONS(3014),
    [anon_sym_no] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3014),
    [anon_sym_foreach] = ACTIONS(3014),
    [anon_sym_last] = ACTIONS(3014),
    [anon_sym_next] = ACTIONS(3014),
    [anon_sym_redo] = ACTIONS(3014),
    [anon_sym_BEGIN] = ACTIONS(3014),
    [anon_sym_INIT] = ACTIONS(3014),
    [anon_sym_CHECK] = ACTIONS(3014),
    [anon_sym_UNITCHECK] = ACTIONS(3014),
    [anon_sym_END] = ACTIONS(3014),
    [anon_sym___FILE__] = ACTIONS(3014),
    [anon_sym___LINE__] = ACTIONS(3014),
    [anon_sym___PACKAGE__] = ACTIONS(3014),
    [anon_sym___SUB__] = ACTIONS(3014),
    [anon_sym_break] = ACTIONS(3014),
    [anon_sym_fork] = ACTIONS(3014),
    [anon_sym_getppid] = ACTIONS(3014),
    [anon_sym_time] = ACTIONS(3014),
    [anon_sym_times] = ACTIONS(3014),
    [anon_sym_wait] = ACTIONS(3014),
    [anon_sym_wantarray] = ACTIONS(3014),
    [anon_sym_abs] = ACTIONS(3014),
    [anon_sym_alarm] = ACTIONS(3014),
    [anon_sym_chop] = ACTIONS(3014),
    [anon_sym_chdir] = ACTIONS(3014),
    [anon_sym_close] = ACTIONS(3014),
    [anon_sym_closedir] = ACTIONS(3014),
    [anon_sym_caller] = ACTIONS(3014),
    [anon_sym_chomp] = ACTIONS(3014),
    [anon_sym_chr] = ACTIONS(3014),
    [anon_sym_cos] = ACTIONS(3014),
    [anon_sym_chroot] = ACTIONS(3014),
    [anon_sym_defined] = ACTIONS(3014),
    [anon_sym_delete] = ACTIONS(3014),
    [anon_sym_dbmclose] = ACTIONS(3014),
    [anon_sym_exists] = ACTIONS(3014),
    [anon_sym_exit] = ACTIONS(3014),
    [anon_sym_eof] = ACTIONS(3014),
    [anon_sym_exp] = ACTIONS(3014),
    [anon_sym_each] = ACTIONS(3014),
    [anon_sym_fc] = ACTIONS(3014),
    [anon_sym_fileno] = ACTIONS(3014),
    [anon_sym_gmtime] = ACTIONS(3014),
    [anon_sym_getc] = ACTIONS(3014),
    [anon_sym_getpgrp] = ACTIONS(3014),
    [anon_sym_getprotobyname] = ACTIONS(3014),
    [anon_sym_getpwname] = ACTIONS(3014),
    [anon_sym_getpwuid] = ACTIONS(3014),
    [anon_sym_getpeername] = ACTIONS(3014),
    [anon_sym_getnetbyname] = ACTIONS(3014),
    [anon_sym_getsockname] = ACTIONS(3014),
    [anon_sym_getgrnam] = ACTIONS(3014),
    [anon_sym_getgrgid] = ACTIONS(3014),
    [anon_sym_hex] = ACTIONS(3014),
    [anon_sym_int] = ACTIONS(3014),
    [anon_sym_keys] = ACTIONS(3014),
    [anon_sym_lc] = ACTIONS(3014),
    [anon_sym_lcfirst] = ACTIONS(3014),
    [anon_sym_length] = ACTIONS(3014),
    [anon_sym_localtime] = ACTIONS(3014),
    [anon_sym_log] = ACTIONS(3014),
    [anon_sym_lock] = ACTIONS(3014),
    [anon_sym_lstat] = ACTIONS(3014),
    [anon_sym_oct] = ACTIONS(3014),
    [anon_sym_ord] = ACTIONS(3014),
    [anon_sym_prototype] = ACTIONS(3014),
    [anon_sym_pop] = ACTIONS(3014),
    [anon_sym_pos] = ACTIONS(3014),
    [anon_sym_quotemeta] = ACTIONS(3014),
    [anon_sym_reset] = ACTIONS(3014),
    [anon_sym_rand] = ACTIONS(3014),
    [anon_sym_rmdir] = ACTIONS(3014),
    [anon_sym_readdir] = ACTIONS(3014),
    [anon_sym_readline] = ACTIONS(3014),
    [anon_sym_readpipe] = ACTIONS(3014),
    [anon_sym_rewinddir] = ACTIONS(3014),
    [anon_sym_readlink] = ACTIONS(3014),
    [anon_sym_ref] = ACTIONS(3014),
    [anon_sym_scalar] = ACTIONS(3014),
    [anon_sym_shift] = ACTIONS(3014),
    [anon_sym_sin] = ACTIONS(3014),
    [anon_sym_sleep] = ACTIONS(3014),
    [anon_sym_sqrt] = ACTIONS(3014),
    [anon_sym_srand] = ACTIONS(3014),
    [anon_sym_stat] = ACTIONS(3014),
    [anon_sym_study] = ACTIONS(3014),
    [anon_sym_tell] = ACTIONS(3014),
    [anon_sym_telldir] = ACTIONS(3014),
    [anon_sym_tied] = ACTIONS(3014),
    [anon_sym_uc] = ACTIONS(3014),
    [anon_sym_ucfirst] = ACTIONS(3014),
    [anon_sym_untie] = ACTIONS(3014),
    [anon_sym_umask] = ACTIONS(3014),
    [anon_sym_values] = ACTIONS(3014),
    [anon_sym_write] = ACTIONS(3014),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3014),
    [anon_sym___END__] = ACTIONS(3014),
    [anon_sym_EOT] = ACTIONS(3014),
    [anon_sym_q] = ACTIONS(3014),
    [anon_sym_qq] = ACTIONS(3014),
    [anon_sym_qw] = ACTIONS(3014),
    [anon_sym_qx] = ACTIONS(3014),
    [anon_sym_qr] = ACTIONS(3014),
    [anon_sym_m] = ACTIONS(3014),
    [anon_sym_s] = ACTIONS(3014),
    [anon_sym_tr] = ACTIONS(3014),
    [anon_sym_y] = ACTIONS(3014),
    [anon_sym_if] = ACTIONS(3014),
    [anon_sym_unless] = ACTIONS(3014),
    [anon_sym_while] = ACTIONS(3014),
    [anon_sym_until] = ACTIONS(3014),
    [aux_sym__bareword_token1] = ACTIONS(3014),
    [sym_number] = ACTIONS(3014),
    [anon_sym_true] = ACTIONS(3014),
    [anon_sym_false] = ACTIONS(3014),
    [sym__single_quote] = ACTIONS(3012),
    [sym__double_quote] = ACTIONS(3012),
    [sym__backtick_quote] = ACTIONS(3012),
    [sym__search_slash_quote] = ACTIONS(3012),
    [sym__open_readline_bracket] = ACTIONS(3012),
    [sym__open_fileglob_bracket] = ACTIONS(3012),
    [sym__PERLY_HEREDOC] = ACTIONS(3012),
    [sym__ctrl_z_hack] = ACTIONS(3012),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3012),
    [sym__filetest] = ACTIONS(3012),
  },
  [STATE(1466)] = {
    [sym_heredoc_content] = STATE(1466),
    [ts_builtin_sym_end] = ACTIONS(3016),
    [sym__identifier] = ACTIONS(3018),
    [anon_sym_LBRACE] = ACTIONS(3018),
    [anon_sym_SEMI] = ACTIONS(3018),
    [anon_sym_package] = ACTIONS(3018),
    [anon_sym_class] = ACTIONS(3018),
    [anon_sym_role] = ACTIONS(3018),
    [anon_sym_BUILD] = ACTIONS(3018),
    [anon_sym_ADJUST] = ACTIONS(3018),
    [anon_sym_DOLLAR] = ACTIONS(3018),
    [anon_sym_AT] = ACTIONS(3018),
    [anon_sym_my] = ACTIONS(3018),
    [anon_sym_extended] = ACTIONS(3018),
    [anon_sym_async] = ACTIONS(3018),
    [anon_sym_sub] = ACTIONS(3018),
    [anon_sym_method] = ACTIONS(3018),
    [anon_sym_LPAREN] = ACTIONS(3018),
    [anon_sym_continue] = ACTIONS(3020),
    [anon_sym_state] = ACTIONS(3018),
    [anon_sym_our] = ACTIONS(3018),
    [anon_sym_try] = ACTIONS(3018),
    [anon_sym_defer] = ACTIONS(3018),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3018),
    [anon_sym_LBRACK] = ACTIONS(3018),
    [anon_sym_LT_LT] = ACTIONS(3018),
    [anon_sym_SLASH_SLASH] = ACTIONS(3018),
    [anon_sym_PLUS] = ACTIONS(3018),
    [anon_sym_DASH] = ACTIONS(3018),
    [anon_sym_TILDE] = ACTIONS(3018),
    [anon_sym_BANG] = ACTIONS(3018),
    [anon_sym_PLUS_PLUS] = ACTIONS(3018),
    [anon_sym_DASH_DASH] = ACTIONS(3018),
    [anon_sym_BSLASH] = ACTIONS(3016),
    [anon_sym_do] = ACTIONS(3018),
    [anon_sym_eval] = ACTIONS(3018),
    [anon_sym_await] = ACTIONS(3018),
    [anon_sym_field] = ACTIONS(3018),
    [anon_sym_local] = ACTIONS(3018),
    [anon_sym_dynamically] = ACTIONS(3018),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3018),
    [anon_sym_require] = ACTIONS(3018),
    [anon_sym_map] = ACTIONS(3018),
    [anon_sym_grep] = ACTIONS(3018),
    [anon_sym_sort] = ACTIONS(3018),
    [anon_sym_goto] = ACTIONS(3018),
    [anon_sym_return] = ACTIONS(3018),
    [anon_sym_undef] = ACTIONS(3018),
    [anon_sym_PERCENT2] = ACTIONS(3018),
    [anon_sym_AMP2] = ACTIONS(3018),
    [anon_sym_STAR2] = ACTIONS(3018),
    [anon_sym_use] = ACTIONS(3018),
    [anon_sym_no] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3018),
    [anon_sym_foreach] = ACTIONS(3018),
    [anon_sym_last] = ACTIONS(3018),
    [anon_sym_next] = ACTIONS(3018),
    [anon_sym_redo] = ACTIONS(3018),
    [anon_sym_BEGIN] = ACTIONS(3018),
    [anon_sym_INIT] = ACTIONS(3018),
    [anon_sym_CHECK] = ACTIONS(3018),
    [anon_sym_UNITCHECK] = ACTIONS(3018),
    [anon_sym_END] = ACTIONS(3018),
    [anon_sym___FILE__] = ACTIONS(3018),
    [anon_sym___LINE__] = ACTIONS(3018),
    [anon_sym___PACKAGE__] = ACTIONS(3018),
    [anon_sym___SUB__] = ACTIONS(3018),
    [anon_sym_break] = ACTIONS(3018),
    [anon_sym_fork] = ACTIONS(3018),
    [anon_sym_getppid] = ACTIONS(3018),
    [anon_sym_time] = ACTIONS(3018),
    [anon_sym_times] = ACTIONS(3018),
    [anon_sym_wait] = ACTIONS(3018),
    [anon_sym_wantarray] = ACTIONS(3018),
    [anon_sym_abs] = ACTIONS(3018),
    [anon_sym_alarm] = ACTIONS(3018),
    [anon_sym_chop] = ACTIONS(3018),
    [anon_sym_chdir] = ACTIONS(3018),
    [anon_sym_close] = ACTIONS(3018),
    [anon_sym_closedir] = ACTIONS(3018),
    [anon_sym_caller] = ACTIONS(3018),
    [anon_sym_chomp] = ACTIONS(3018),
    [anon_sym_chr] = ACTIONS(3018),
    [anon_sym_cos] = ACTIONS(3018),
    [anon_sym_chroot] = ACTIONS(3018),
    [anon_sym_defined] = ACTIONS(3018),
    [anon_sym_delete] = ACTIONS(3018),
    [anon_sym_dbmclose] = ACTIONS(3018),
    [anon_sym_exists] = ACTIONS(3018),
    [anon_sym_exit] = ACTIONS(3018),
    [anon_sym_eof] = ACTIONS(3018),
    [anon_sym_exp] = ACTIONS(3018),
    [anon_sym_each] = ACTIONS(3018),
    [anon_sym_fc] = ACTIONS(3018),
    [anon_sym_fileno] = ACTIONS(3018),
    [anon_sym_gmtime] = ACTIONS(3018),
    [anon_sym_getc] = ACTIONS(3018),
    [anon_sym_getpgrp] = ACTIONS(3018),
    [anon_sym_getprotobyname] = ACTIONS(3018),
    [anon_sym_getpwname] = ACTIONS(3018),
    [anon_sym_getpwuid] = ACTIONS(3018),
    [anon_sym_getpeername] = ACTIONS(3018),
    [anon_sym_getnetbyname] = ACTIONS(3018),
    [anon_sym_getsockname] = ACTIONS(3018),
    [anon_sym_getgrnam] = ACTIONS(3018),
    [anon_sym_getgrgid] = ACTIONS(3018),
    [anon_sym_hex] = ACTIONS(3018),
    [anon_sym_int] = ACTIONS(3018),
    [anon_sym_keys] = ACTIONS(3018),
    [anon_sym_lc] = ACTIONS(3018),
    [anon_sym_lcfirst] = ACTIONS(3018),
    [anon_sym_length] = ACTIONS(3018),
    [anon_sym_localtime] = ACTIONS(3018),
    [anon_sym_log] = ACTIONS(3018),
    [anon_sym_lock] = ACTIONS(3018),
    [anon_sym_lstat] = ACTIONS(3018),
    [anon_sym_oct] = ACTIONS(3018),
    [anon_sym_ord] = ACTIONS(3018),
    [anon_sym_prototype] = ACTIONS(3018),
    [anon_sym_pop] = ACTIONS(3018),
    [anon_sym_pos] = ACTIONS(3018),
    [anon_sym_quotemeta] = ACTIONS(3018),
    [anon_sym_reset] = ACTIONS(3018),
    [anon_sym_rand] = ACTIONS(3018),
    [anon_sym_rmdir] = ACTIONS(3018),
    [anon_sym_readdir] = ACTIONS(3018),
    [anon_sym_readline] = ACTIONS(3018),
    [anon_sym_readpipe] = ACTIONS(3018),
    [anon_sym_rewinddir] = ACTIONS(3018),
    [anon_sym_readlink] = ACTIONS(3018),
    [anon_sym_ref] = ACTIONS(3018),
    [anon_sym_scalar] = ACTIONS(3018),
    [anon_sym_shift] = ACTIONS(3018),
    [anon_sym_sin] = ACTIONS(3018),
    [anon_sym_sleep] = ACTIONS(3018),
    [anon_sym_sqrt] = ACTIONS(3018),
    [anon_sym_srand] = ACTIONS(3018),
    [anon_sym_stat] = ACTIONS(3018),
    [anon_sym_study] = ACTIONS(3018),
    [anon_sym_tell] = ACTIONS(3018),
    [anon_sym_telldir] = ACTIONS(3018),
    [anon_sym_tied] = ACTIONS(3018),
    [anon_sym_uc] = ACTIONS(3018),
    [anon_sym_ucfirst] = ACTIONS(3018),
    [anon_sym_untie] = ACTIONS(3018),
    [anon_sym_umask] = ACTIONS(3018),
    [anon_sym_values] = ACTIONS(3018),
    [anon_sym_write] = ACTIONS(3018),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3018),
    [anon_sym___END__] = ACTIONS(3018),
    [anon_sym_EOT] = ACTIONS(3018),
    [anon_sym_q] = ACTIONS(3018),
    [anon_sym_qq] = ACTIONS(3018),
    [anon_sym_qw] = ACTIONS(3018),
    [anon_sym_qx] = ACTIONS(3018),
    [anon_sym_qr] = ACTIONS(3018),
    [anon_sym_m] = ACTIONS(3018),
    [anon_sym_s] = ACTIONS(3018),
    [anon_sym_tr] = ACTIONS(3018),
    [anon_sym_y] = ACTIONS(3018),
    [anon_sym_if] = ACTIONS(3018),
    [anon_sym_unless] = ACTIONS(3018),
    [anon_sym_while] = ACTIONS(3018),
    [anon_sym_until] = ACTIONS(3018),
    [aux_sym__bareword_token1] = ACTIONS(3018),
    [sym_number] = ACTIONS(3018),
    [anon_sym_true] = ACTIONS(3018),
    [anon_sym_false] = ACTIONS(3018),
    [sym__single_quote] = ACTIONS(3016),
    [sym__double_quote] = ACTIONS(3016),
    [sym__backtick_quote] = ACTIONS(3016),
    [sym__search_slash_quote] = ACTIONS(3016),
    [sym__open_readline_bracket] = ACTIONS(3016),
    [sym__open_fileglob_bracket] = ACTIONS(3016),
    [sym__PERLY_HEREDOC] = ACTIONS(3016),
    [sym__ctrl_z_hack] = ACTIONS(3016),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3016),
    [sym__filetest] = ACTIONS(3016),
  },
  [STATE(1467)] = {
    [sym_heredoc_content] = STATE(1467),
    [ts_builtin_sym_end] = ACTIONS(3023),
    [sym__identifier] = ACTIONS(3025),
    [anon_sym_LBRACE] = ACTIONS(3025),
    [anon_sym_SEMI] = ACTIONS(3025),
    [anon_sym_package] = ACTIONS(3025),
    [anon_sym_class] = ACTIONS(3025),
    [anon_sym_role] = ACTIONS(3025),
    [anon_sym_BUILD] = ACTIONS(3025),
    [anon_sym_ADJUST] = ACTIONS(3025),
    [anon_sym_DOLLAR] = ACTIONS(3025),
    [anon_sym_AT] = ACTIONS(3025),
    [anon_sym_my] = ACTIONS(3025),
    [anon_sym_extended] = ACTIONS(3025),
    [anon_sym_async] = ACTIONS(3025),
    [anon_sym_sub] = ACTIONS(3025),
    [anon_sym_method] = ACTIONS(3025),
    [anon_sym_LPAREN] = ACTIONS(3025),
    [anon_sym_continue] = ACTIONS(3025),
    [anon_sym_state] = ACTIONS(3025),
    [anon_sym_our] = ACTIONS(3025),
    [anon_sym_try] = ACTIONS(3025),
    [anon_sym_defer] = ACTIONS(3025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3025),
    [anon_sym_LBRACK] = ACTIONS(3025),
    [anon_sym_LT_LT] = ACTIONS(3025),
    [anon_sym_SLASH_SLASH] = ACTIONS(3025),
    [anon_sym_PLUS] = ACTIONS(3025),
    [anon_sym_DASH] = ACTIONS(3025),
    [anon_sym_TILDE] = ACTIONS(3025),
    [anon_sym_BANG] = ACTIONS(3025),
    [anon_sym_PLUS_PLUS] = ACTIONS(3025),
    [anon_sym_DASH_DASH] = ACTIONS(3025),
    [anon_sym_BSLASH] = ACTIONS(3023),
    [anon_sym_do] = ACTIONS(3025),
    [anon_sym_eval] = ACTIONS(3025),
    [anon_sym_await] = ACTIONS(3025),
    [anon_sym_field] = ACTIONS(3025),
    [anon_sym_local] = ACTIONS(3025),
    [anon_sym_dynamically] = ACTIONS(3025),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3025),
    [anon_sym_require] = ACTIONS(3025),
    [anon_sym_map] = ACTIONS(3025),
    [anon_sym_grep] = ACTIONS(3025),
    [anon_sym_sort] = ACTIONS(3025),
    [anon_sym_goto] = ACTIONS(3025),
    [anon_sym_return] = ACTIONS(3025),
    [anon_sym_undef] = ACTIONS(3025),
    [anon_sym_PERCENT2] = ACTIONS(3025),
    [anon_sym_AMP2] = ACTIONS(3025),
    [anon_sym_STAR2] = ACTIONS(3025),
    [anon_sym_use] = ACTIONS(3025),
    [anon_sym_no] = ACTIONS(3025),
    [anon_sym_for] = ACTIONS(3025),
    [anon_sym_foreach] = ACTIONS(3025),
    [anon_sym_last] = ACTIONS(3025),
    [anon_sym_next] = ACTIONS(3025),
    [anon_sym_redo] = ACTIONS(3025),
    [anon_sym_BEGIN] = ACTIONS(3025),
    [anon_sym_INIT] = ACTIONS(3025),
    [anon_sym_CHECK] = ACTIONS(3025),
    [anon_sym_UNITCHECK] = ACTIONS(3025),
    [anon_sym_END] = ACTIONS(3025),
    [anon_sym___FILE__] = ACTIONS(3025),
    [anon_sym___LINE__] = ACTIONS(3025),
    [anon_sym___PACKAGE__] = ACTIONS(3025),
    [anon_sym___SUB__] = ACTIONS(3025),
    [anon_sym_break] = ACTIONS(3025),
    [anon_sym_fork] = ACTIONS(3025),
    [anon_sym_getppid] = ACTIONS(3025),
    [anon_sym_time] = ACTIONS(3025),
    [anon_sym_times] = ACTIONS(3025),
    [anon_sym_wait] = ACTIONS(3025),
    [anon_sym_wantarray] = ACTIONS(3025),
    [anon_sym_abs] = ACTIONS(3025),
    [anon_sym_alarm] = ACTIONS(3025),
    [anon_sym_chop] = ACTIONS(3025),
    [anon_sym_chdir] = ACTIONS(3025),
    [anon_sym_close] = ACTIONS(3025),
    [anon_sym_closedir] = ACTIONS(3025),
    [anon_sym_caller] = ACTIONS(3025),
    [anon_sym_chomp] = ACTIONS(3025),
    [anon_sym_chr] = ACTIONS(3025),
    [anon_sym_cos] = ACTIONS(3025),
    [anon_sym_chroot] = ACTIONS(3025),
    [anon_sym_defined] = ACTIONS(3025),
    [anon_sym_delete] = ACTIONS(3025),
    [anon_sym_dbmclose] = ACTIONS(3025),
    [anon_sym_exists] = ACTIONS(3025),
    [anon_sym_exit] = ACTIONS(3025),
    [anon_sym_eof] = ACTIONS(3025),
    [anon_sym_exp] = ACTIONS(3025),
    [anon_sym_each] = ACTIONS(3025),
    [anon_sym_fc] = ACTIONS(3025),
    [anon_sym_fileno] = ACTIONS(3025),
    [anon_sym_gmtime] = ACTIONS(3025),
    [anon_sym_getc] = ACTIONS(3025),
    [anon_sym_getpgrp] = ACTIONS(3025),
    [anon_sym_getprotobyname] = ACTIONS(3025),
    [anon_sym_getpwname] = ACTIONS(3025),
    [anon_sym_getpwuid] = ACTIONS(3025),
    [anon_sym_getpeername] = ACTIONS(3025),
    [anon_sym_getnetbyname] = ACTIONS(3025),
    [anon_sym_getsockname] = ACTIONS(3025),
    [anon_sym_getgrnam] = ACTIONS(3025),
    [anon_sym_getgrgid] = ACTIONS(3025),
    [anon_sym_hex] = ACTIONS(3025),
    [anon_sym_int] = ACTIONS(3025),
    [anon_sym_keys] = ACTIONS(3025),
    [anon_sym_lc] = ACTIONS(3025),
    [anon_sym_lcfirst] = ACTIONS(3025),
    [anon_sym_length] = ACTIONS(3025),
    [anon_sym_localtime] = ACTIONS(3025),
    [anon_sym_log] = ACTIONS(3025),
    [anon_sym_lock] = ACTIONS(3025),
    [anon_sym_lstat] = ACTIONS(3025),
    [anon_sym_oct] = ACTIONS(3025),
    [anon_sym_ord] = ACTIONS(3025),
    [anon_sym_prototype] = ACTIONS(3025),
    [anon_sym_pop] = ACTIONS(3025),
    [anon_sym_pos] = ACTIONS(3025),
    [anon_sym_quotemeta] = ACTIONS(3025),
    [anon_sym_reset] = ACTIONS(3025),
    [anon_sym_rand] = ACTIONS(3025),
    [anon_sym_rmdir] = ACTIONS(3025),
    [anon_sym_readdir] = ACTIONS(3025),
    [anon_sym_readline] = ACTIONS(3025),
    [anon_sym_readpipe] = ACTIONS(3025),
    [anon_sym_rewinddir] = ACTIONS(3025),
    [anon_sym_readlink] = ACTIONS(3025),
    [anon_sym_ref] = ACTIONS(3025),
    [anon_sym_scalar] = ACTIONS(3025),
    [anon_sym_shift] = ACTIONS(3025),
    [anon_sym_sin] = ACTIONS(3025),
    [anon_sym_sleep] = ACTIONS(3025),
    [anon_sym_sqrt] = ACTIONS(3025),
    [anon_sym_srand] = ACTIONS(3025),
    [anon_sym_stat] = ACTIONS(3025),
    [anon_sym_study] = ACTIONS(3025),
    [anon_sym_tell] = ACTIONS(3025),
    [anon_sym_telldir] = ACTIONS(3025),
    [anon_sym_tied] = ACTIONS(3025),
    [anon_sym_uc] = ACTIONS(3025),
    [anon_sym_ucfirst] = ACTIONS(3025),
    [anon_sym_untie] = ACTIONS(3025),
    [anon_sym_umask] = ACTIONS(3025),
    [anon_sym_values] = ACTIONS(3025),
    [anon_sym_write] = ACTIONS(3025),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3025),
    [anon_sym___END__] = ACTIONS(3025),
    [anon_sym_EOT] = ACTIONS(3025),
    [anon_sym_q] = ACTIONS(3025),
    [anon_sym_qq] = ACTIONS(3025),
    [anon_sym_qw] = ACTIONS(3025),
    [anon_sym_qx] = ACTIONS(3025),
    [anon_sym_qr] = ACTIONS(3025),
    [anon_sym_m] = ACTIONS(3025),
    [anon_sym_s] = ACTIONS(3025),
    [anon_sym_tr] = ACTIONS(3025),
    [anon_sym_y] = ACTIONS(3025),
    [anon_sym_if] = ACTIONS(3025),
    [anon_sym_unless] = ACTIONS(3025),
    [anon_sym_while] = ACTIONS(3025),
    [anon_sym_until] = ACTIONS(3025),
    [aux_sym__bareword_token1] = ACTIONS(3025),
    [sym_number] = ACTIONS(3025),
    [anon_sym_true] = ACTIONS(3025),
    [anon_sym_false] = ACTIONS(3025),
    [sym__single_quote] = ACTIONS(3023),
    [sym__double_quote] = ACTIONS(3023),
    [sym__backtick_quote] = ACTIONS(3023),
    [sym__search_slash_quote] = ACTIONS(3023),
    [sym__open_readline_bracket] = ACTIONS(3023),
    [sym__open_fileglob_bracket] = ACTIONS(3023),
    [sym__PERLY_HEREDOC] = ACTIONS(3023),
    [sym__ctrl_z_hack] = ACTIONS(3023),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3023),
    [sym__filetest] = ACTIONS(3023),
  },
  [STATE(1468)] = {
    [sym_heredoc_content] = STATE(1468),
    [ts_builtin_sym_end] = ACTIONS(3027),
    [sym__identifier] = ACTIONS(3029),
    [anon_sym_LBRACE] = ACTIONS(3029),
    [anon_sym_SEMI] = ACTIONS(3029),
    [anon_sym_package] = ACTIONS(3029),
    [anon_sym_class] = ACTIONS(3029),
    [anon_sym_role] = ACTIONS(3029),
    [anon_sym_BUILD] = ACTIONS(3029),
    [anon_sym_ADJUST] = ACTIONS(3029),
    [anon_sym_DOLLAR] = ACTIONS(3029),
    [anon_sym_AT] = ACTIONS(3029),
    [anon_sym_my] = ACTIONS(3029),
    [anon_sym_extended] = ACTIONS(3029),
    [anon_sym_async] = ACTIONS(3029),
    [anon_sym_sub] = ACTIONS(3029),
    [anon_sym_method] = ACTIONS(3029),
    [anon_sym_LPAREN] = ACTIONS(3029),
    [anon_sym_continue] = ACTIONS(3029),
    [anon_sym_state] = ACTIONS(3029),
    [anon_sym_our] = ACTIONS(3029),
    [anon_sym_try] = ACTIONS(3029),
    [anon_sym_defer] = ACTIONS(3029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3029),
    [anon_sym_LBRACK] = ACTIONS(3029),
    [anon_sym_LT_LT] = ACTIONS(3029),
    [anon_sym_SLASH_SLASH] = ACTIONS(3029),
    [anon_sym_PLUS] = ACTIONS(3029),
    [anon_sym_DASH] = ACTIONS(3029),
    [anon_sym_TILDE] = ACTIONS(3029),
    [anon_sym_BANG] = ACTIONS(3029),
    [anon_sym_PLUS_PLUS] = ACTIONS(3029),
    [anon_sym_DASH_DASH] = ACTIONS(3029),
    [anon_sym_BSLASH] = ACTIONS(3027),
    [anon_sym_do] = ACTIONS(3029),
    [anon_sym_eval] = ACTIONS(3029),
    [anon_sym_await] = ACTIONS(3029),
    [anon_sym_field] = ACTIONS(3029),
    [anon_sym_local] = ACTIONS(3029),
    [anon_sym_dynamically] = ACTIONS(3029),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3029),
    [anon_sym_require] = ACTIONS(3029),
    [anon_sym_map] = ACTIONS(3029),
    [anon_sym_grep] = ACTIONS(3029),
    [anon_sym_sort] = ACTIONS(3029),
    [anon_sym_goto] = ACTIONS(3029),
    [anon_sym_return] = ACTIONS(3029),
    [anon_sym_undef] = ACTIONS(3029),
    [anon_sym_PERCENT2] = ACTIONS(3029),
    [anon_sym_AMP2] = ACTIONS(3029),
    [anon_sym_STAR2] = ACTIONS(3029),
    [anon_sym_use] = ACTIONS(3029),
    [anon_sym_no] = ACTIONS(3029),
    [anon_sym_for] = ACTIONS(3029),
    [anon_sym_foreach] = ACTIONS(3029),
    [anon_sym_last] = ACTIONS(3029),
    [anon_sym_next] = ACTIONS(3029),
    [anon_sym_redo] = ACTIONS(3029),
    [anon_sym_BEGIN] = ACTIONS(3029),
    [anon_sym_INIT] = ACTIONS(3029),
    [anon_sym_CHECK] = ACTIONS(3029),
    [anon_sym_UNITCHECK] = ACTIONS(3029),
    [anon_sym_END] = ACTIONS(3029),
    [anon_sym___FILE__] = ACTIONS(3029),
    [anon_sym___LINE__] = ACTIONS(3029),
    [anon_sym___PACKAGE__] = ACTIONS(3029),
    [anon_sym___SUB__] = ACTIONS(3029),
    [anon_sym_break] = ACTIONS(3029),
    [anon_sym_fork] = ACTIONS(3029),
    [anon_sym_getppid] = ACTIONS(3029),
    [anon_sym_time] = ACTIONS(3029),
    [anon_sym_times] = ACTIONS(3029),
    [anon_sym_wait] = ACTIONS(3029),
    [anon_sym_wantarray] = ACTIONS(3029),
    [anon_sym_abs] = ACTIONS(3029),
    [anon_sym_alarm] = ACTIONS(3029),
    [anon_sym_chop] = ACTIONS(3029),
    [anon_sym_chdir] = ACTIONS(3029),
    [anon_sym_close] = ACTIONS(3029),
    [anon_sym_closedir] = ACTIONS(3029),
    [anon_sym_caller] = ACTIONS(3029),
    [anon_sym_chomp] = ACTIONS(3029),
    [anon_sym_chr] = ACTIONS(3029),
    [anon_sym_cos] = ACTIONS(3029),
    [anon_sym_chroot] = ACTIONS(3029),
    [anon_sym_defined] = ACTIONS(3029),
    [anon_sym_delete] = ACTIONS(3029),
    [anon_sym_dbmclose] = ACTIONS(3029),
    [anon_sym_exists] = ACTIONS(3029),
    [anon_sym_exit] = ACTIONS(3029),
    [anon_sym_eof] = ACTIONS(3029),
    [anon_sym_exp] = ACTIONS(3029),
    [anon_sym_each] = ACTIONS(3029),
    [anon_sym_fc] = ACTIONS(3029),
    [anon_sym_fileno] = ACTIONS(3029),
    [anon_sym_gmtime] = ACTIONS(3029),
    [anon_sym_getc] = ACTIONS(3029),
    [anon_sym_getpgrp] = ACTIONS(3029),
    [anon_sym_getprotobyname] = ACTIONS(3029),
    [anon_sym_getpwname] = ACTIONS(3029),
    [anon_sym_getpwuid] = ACTIONS(3029),
    [anon_sym_getpeername] = ACTIONS(3029),
    [anon_sym_getnetbyname] = ACTIONS(3029),
    [anon_sym_getsockname] = ACTIONS(3029),
    [anon_sym_getgrnam] = ACTIONS(3029),
    [anon_sym_getgrgid] = ACTIONS(3029),
    [anon_sym_hex] = ACTIONS(3029),
    [anon_sym_int] = ACTIONS(3029),
    [anon_sym_keys] = ACTIONS(3029),
    [anon_sym_lc] = ACTIONS(3029),
    [anon_sym_lcfirst] = ACTIONS(3029),
    [anon_sym_length] = ACTIONS(3029),
    [anon_sym_localtime] = ACTIONS(3029),
    [anon_sym_log] = ACTIONS(3029),
    [anon_sym_lock] = ACTIONS(3029),
    [anon_sym_lstat] = ACTIONS(3029),
    [anon_sym_oct] = ACTIONS(3029),
    [anon_sym_ord] = ACTIONS(3029),
    [anon_sym_prototype] = ACTIONS(3029),
    [anon_sym_pop] = ACTIONS(3029),
    [anon_sym_pos] = ACTIONS(3029),
    [anon_sym_quotemeta] = ACTIONS(3029),
    [anon_sym_reset] = ACTIONS(3029),
    [anon_sym_rand] = ACTIONS(3029),
    [anon_sym_rmdir] = ACTIONS(3029),
    [anon_sym_readdir] = ACTIONS(3029),
    [anon_sym_readline] = ACTIONS(3029),
    [anon_sym_readpipe] = ACTIONS(3029),
    [anon_sym_rewinddir] = ACTIONS(3029),
    [anon_sym_readlink] = ACTIONS(3029),
    [anon_sym_ref] = ACTIONS(3029),
    [anon_sym_scalar] = ACTIONS(3029),
    [anon_sym_shift] = ACTIONS(3029),
    [anon_sym_sin] = ACTIONS(3029),
    [anon_sym_sleep] = ACTIONS(3029),
    [anon_sym_sqrt] = ACTIONS(3029),
    [anon_sym_srand] = ACTIONS(3029),
    [anon_sym_stat] = ACTIONS(3029),
    [anon_sym_study] = ACTIONS(3029),
    [anon_sym_tell] = ACTIONS(3029),
    [anon_sym_telldir] = ACTIONS(3029),
    [anon_sym_tied] = ACTIONS(3029),
    [anon_sym_uc] = ACTIONS(3029),
    [anon_sym_ucfirst] = ACTIONS(3029),
    [anon_sym_untie] = ACTIONS(3029),
    [anon_sym_umask] = ACTIONS(3029),
    [anon_sym_values] = ACTIONS(3029),
    [anon_sym_write] = ACTIONS(3029),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3029),
    [anon_sym___END__] = ACTIONS(3029),
    [anon_sym_EOT] = ACTIONS(3029),
    [anon_sym_q] = ACTIONS(3029),
    [anon_sym_qq] = ACTIONS(3029),
    [anon_sym_qw] = ACTIONS(3029),
    [anon_sym_qx] = ACTIONS(3029),
    [anon_sym_qr] = ACTIONS(3029),
    [anon_sym_m] = ACTIONS(3029),
    [anon_sym_s] = ACTIONS(3029),
    [anon_sym_tr] = ACTIONS(3029),
    [anon_sym_y] = ACTIONS(3029),
    [anon_sym_if] = ACTIONS(3029),
    [anon_sym_unless] = ACTIONS(3029),
    [anon_sym_while] = ACTIONS(3029),
    [anon_sym_until] = ACTIONS(3029),
    [aux_sym__bareword_token1] = ACTIONS(3029),
    [sym_number] = ACTIONS(3029),
    [anon_sym_true] = ACTIONS(3029),
    [anon_sym_false] = ACTIONS(3029),
    [sym__single_quote] = ACTIONS(3027),
    [sym__double_quote] = ACTIONS(3027),
    [sym__backtick_quote] = ACTIONS(3027),
    [sym__search_slash_quote] = ACTIONS(3027),
    [sym__open_readline_bracket] = ACTIONS(3027),
    [sym__open_fileglob_bracket] = ACTIONS(3027),
    [sym__PERLY_HEREDOC] = ACTIONS(3027),
    [sym__ctrl_z_hack] = ACTIONS(3027),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3027),
    [sym__filetest] = ACTIONS(3027),
  },
  [STATE(1469)] = {
    [sym_heredoc_content] = STATE(1469),
    [ts_builtin_sym_end] = ACTIONS(3031),
    [sym__identifier] = ACTIONS(3033),
    [anon_sym_LBRACE] = ACTIONS(3033),
    [anon_sym_SEMI] = ACTIONS(3033),
    [anon_sym_package] = ACTIONS(3033),
    [anon_sym_class] = ACTIONS(3033),
    [anon_sym_role] = ACTIONS(3033),
    [anon_sym_BUILD] = ACTIONS(3033),
    [anon_sym_ADJUST] = ACTIONS(3033),
    [anon_sym_DOLLAR] = ACTIONS(3033),
    [anon_sym_AT] = ACTIONS(3033),
    [anon_sym_my] = ACTIONS(3033),
    [anon_sym_extended] = ACTIONS(3033),
    [anon_sym_async] = ACTIONS(3033),
    [anon_sym_sub] = ACTIONS(3033),
    [anon_sym_method] = ACTIONS(3033),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_continue] = ACTIONS(3033),
    [anon_sym_state] = ACTIONS(3033),
    [anon_sym_our] = ACTIONS(3033),
    [anon_sym_try] = ACTIONS(3033),
    [anon_sym_defer] = ACTIONS(3033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3033),
    [anon_sym_LBRACK] = ACTIONS(3033),
    [anon_sym_LT_LT] = ACTIONS(3033),
    [anon_sym_SLASH_SLASH] = ACTIONS(3033),
    [anon_sym_PLUS] = ACTIONS(3033),
    [anon_sym_DASH] = ACTIONS(3033),
    [anon_sym_TILDE] = ACTIONS(3033),
    [anon_sym_BANG] = ACTIONS(3033),
    [anon_sym_PLUS_PLUS] = ACTIONS(3033),
    [anon_sym_DASH_DASH] = ACTIONS(3033),
    [anon_sym_BSLASH] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3033),
    [anon_sym_eval] = ACTIONS(3033),
    [anon_sym_await] = ACTIONS(3033),
    [anon_sym_field] = ACTIONS(3033),
    [anon_sym_local] = ACTIONS(3033),
    [anon_sym_dynamically] = ACTIONS(3033),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3033),
    [anon_sym_require] = ACTIONS(3033),
    [anon_sym_map] = ACTIONS(3033),
    [anon_sym_grep] = ACTIONS(3033),
    [anon_sym_sort] = ACTIONS(3033),
    [anon_sym_goto] = ACTIONS(3033),
    [anon_sym_return] = ACTIONS(3033),
    [anon_sym_undef] = ACTIONS(3033),
    [anon_sym_PERCENT2] = ACTIONS(3033),
    [anon_sym_AMP2] = ACTIONS(3033),
    [anon_sym_STAR2] = ACTIONS(3033),
    [anon_sym_use] = ACTIONS(3033),
    [anon_sym_no] = ACTIONS(3033),
    [anon_sym_for] = ACTIONS(3033),
    [anon_sym_foreach] = ACTIONS(3033),
    [anon_sym_last] = ACTIONS(3033),
    [anon_sym_next] = ACTIONS(3033),
    [anon_sym_redo] = ACTIONS(3033),
    [anon_sym_BEGIN] = ACTIONS(3033),
    [anon_sym_INIT] = ACTIONS(3033),
    [anon_sym_CHECK] = ACTIONS(3033),
    [anon_sym_UNITCHECK] = ACTIONS(3033),
    [anon_sym_END] = ACTIONS(3033),
    [anon_sym___FILE__] = ACTIONS(3033),
    [anon_sym___LINE__] = ACTIONS(3033),
    [anon_sym___PACKAGE__] = ACTIONS(3033),
    [anon_sym___SUB__] = ACTIONS(3033),
    [anon_sym_break] = ACTIONS(3033),
    [anon_sym_fork] = ACTIONS(3033),
    [anon_sym_getppid] = ACTIONS(3033),
    [anon_sym_time] = ACTIONS(3033),
    [anon_sym_times] = ACTIONS(3033),
    [anon_sym_wait] = ACTIONS(3033),
    [anon_sym_wantarray] = ACTIONS(3033),
    [anon_sym_abs] = ACTIONS(3033),
    [anon_sym_alarm] = ACTIONS(3033),
    [anon_sym_chop] = ACTIONS(3033),
    [anon_sym_chdir] = ACTIONS(3033),
    [anon_sym_close] = ACTIONS(3033),
    [anon_sym_closedir] = ACTIONS(3033),
    [anon_sym_caller] = ACTIONS(3033),
    [anon_sym_chomp] = ACTIONS(3033),
    [anon_sym_chr] = ACTIONS(3033),
    [anon_sym_cos] = ACTIONS(3033),
    [anon_sym_chroot] = ACTIONS(3033),
    [anon_sym_defined] = ACTIONS(3033),
    [anon_sym_delete] = ACTIONS(3033),
    [anon_sym_dbmclose] = ACTIONS(3033),
    [anon_sym_exists] = ACTIONS(3033),
    [anon_sym_exit] = ACTIONS(3033),
    [anon_sym_eof] = ACTIONS(3033),
    [anon_sym_exp] = ACTIONS(3033),
    [anon_sym_each] = ACTIONS(3033),
    [anon_sym_fc] = ACTIONS(3033),
    [anon_sym_fileno] = ACTIONS(3033),
    [anon_sym_gmtime] = ACTIONS(3033),
    [anon_sym_getc] = ACTIONS(3033),
    [anon_sym_getpgrp] = ACTIONS(3033),
    [anon_sym_getprotobyname] = ACTIONS(3033),
    [anon_sym_getpwname] = ACTIONS(3033),
    [anon_sym_getpwuid] = ACTIONS(3033),
    [anon_sym_getpeername] = ACTIONS(3033),
    [anon_sym_getnetbyname] = ACTIONS(3033),
    [anon_sym_getsockname] = ACTIONS(3033),
    [anon_sym_getgrnam] = ACTIONS(3033),
    [anon_sym_getgrgid] = ACTIONS(3033),
    [anon_sym_hex] = ACTIONS(3033),
    [anon_sym_int] = ACTIONS(3033),
    [anon_sym_keys] = ACTIONS(3033),
    [anon_sym_lc] = ACTIONS(3033),
    [anon_sym_lcfirst] = ACTIONS(3033),
    [anon_sym_length] = ACTIONS(3033),
    [anon_sym_localtime] = ACTIONS(3033),
    [anon_sym_log] = ACTIONS(3033),
    [anon_sym_lock] = ACTIONS(3033),
    [anon_sym_lstat] = ACTIONS(3033),
    [anon_sym_oct] = ACTIONS(3033),
    [anon_sym_ord] = ACTIONS(3033),
    [anon_sym_prototype] = ACTIONS(3033),
    [anon_sym_pop] = ACTIONS(3033),
    [anon_sym_pos] = ACTIONS(3033),
    [anon_sym_quotemeta] = ACTIONS(3033),
    [anon_sym_reset] = ACTIONS(3033),
    [anon_sym_rand] = ACTIONS(3033),
    [anon_sym_rmdir] = ACTIONS(3033),
    [anon_sym_readdir] = ACTIONS(3033),
    [anon_sym_readline] = ACTIONS(3033),
    [anon_sym_readpipe] = ACTIONS(3033),
    [anon_sym_rewinddir] = ACTIONS(3033),
    [anon_sym_readlink] = ACTIONS(3033),
    [anon_sym_ref] = ACTIONS(3033),
    [anon_sym_scalar] = ACTIONS(3033),
    [anon_sym_shift] = ACTIONS(3033),
    [anon_sym_sin] = ACTIONS(3033),
    [anon_sym_sleep] = ACTIONS(3033),
    [anon_sym_sqrt] = ACTIONS(3033),
    [anon_sym_srand] = ACTIONS(3033),
    [anon_sym_stat] = ACTIONS(3033),
    [anon_sym_study] = ACTIONS(3033),
    [anon_sym_tell] = ACTIONS(3033),
    [anon_sym_telldir] = ACTIONS(3033),
    [anon_sym_tied] = ACTIONS(3033),
    [anon_sym_uc] = ACTIONS(3033),
    [anon_sym_ucfirst] = ACTIONS(3033),
    [anon_sym_untie] = ACTIONS(3033),
    [anon_sym_umask] = ACTIONS(3033),
    [anon_sym_values] = ACTIONS(3033),
    [anon_sym_write] = ACTIONS(3033),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3033),
    [anon_sym___END__] = ACTIONS(3033),
    [anon_sym_EOT] = ACTIONS(3033),
    [anon_sym_q] = ACTIONS(3033),
    [anon_sym_qq] = ACTIONS(3033),
    [anon_sym_qw] = ACTIONS(3033),
    [anon_sym_qx] = ACTIONS(3033),
    [anon_sym_qr] = ACTIONS(3033),
    [anon_sym_m] = ACTIONS(3033),
    [anon_sym_s] = ACTIONS(3033),
    [anon_sym_tr] = ACTIONS(3033),
    [anon_sym_y] = ACTIONS(3033),
    [anon_sym_if] = ACTIONS(3033),
    [anon_sym_unless] = ACTIONS(3033),
    [anon_sym_while] = ACTIONS(3033),
    [anon_sym_until] = ACTIONS(3033),
    [aux_sym__bareword_token1] = ACTIONS(3033),
    [sym_number] = ACTIONS(3033),
    [anon_sym_true] = ACTIONS(3033),
    [anon_sym_false] = ACTIONS(3033),
    [sym__single_quote] = ACTIONS(3031),
    [sym__double_quote] = ACTIONS(3031),
    [sym__backtick_quote] = ACTIONS(3031),
    [sym__search_slash_quote] = ACTIONS(3031),
    [sym__open_readline_bracket] = ACTIONS(3031),
    [sym__open_fileglob_bracket] = ACTIONS(3031),
    [sym__PERLY_HEREDOC] = ACTIONS(3031),
    [sym__ctrl_z_hack] = ACTIONS(3031),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3031),
    [sym__filetest] = ACTIONS(3031),
  },
  [STATE(1470)] = {
    [sym_heredoc_content] = STATE(1470),
    [ts_builtin_sym_end] = ACTIONS(3035),
    [sym__identifier] = ACTIONS(3037),
    [anon_sym_LBRACE] = ACTIONS(3037),
    [anon_sym_SEMI] = ACTIONS(3037),
    [anon_sym_package] = ACTIONS(3037),
    [anon_sym_class] = ACTIONS(3037),
    [anon_sym_role] = ACTIONS(3037),
    [anon_sym_BUILD] = ACTIONS(3037),
    [anon_sym_ADJUST] = ACTIONS(3037),
    [anon_sym_DOLLAR] = ACTIONS(3037),
    [anon_sym_AT] = ACTIONS(3037),
    [anon_sym_my] = ACTIONS(3037),
    [anon_sym_extended] = ACTIONS(3037),
    [anon_sym_async] = ACTIONS(3037),
    [anon_sym_sub] = ACTIONS(3037),
    [anon_sym_method] = ACTIONS(3037),
    [anon_sym_LPAREN] = ACTIONS(3037),
    [anon_sym_continue] = ACTIONS(3037),
    [anon_sym_state] = ACTIONS(3037),
    [anon_sym_our] = ACTIONS(3037),
    [anon_sym_try] = ACTIONS(3037),
    [anon_sym_defer] = ACTIONS(3037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3037),
    [anon_sym_LBRACK] = ACTIONS(3037),
    [anon_sym_LT_LT] = ACTIONS(3037),
    [anon_sym_SLASH_SLASH] = ACTIONS(3037),
    [anon_sym_PLUS] = ACTIONS(3037),
    [anon_sym_DASH] = ACTIONS(3037),
    [anon_sym_TILDE] = ACTIONS(3037),
    [anon_sym_BANG] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH] = ACTIONS(3037),
    [anon_sym_BSLASH] = ACTIONS(3035),
    [anon_sym_do] = ACTIONS(3037),
    [anon_sym_eval] = ACTIONS(3037),
    [anon_sym_await] = ACTIONS(3037),
    [anon_sym_field] = ACTIONS(3037),
    [anon_sym_local] = ACTIONS(3037),
    [anon_sym_dynamically] = ACTIONS(3037),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3037),
    [anon_sym_require] = ACTIONS(3037),
    [anon_sym_map] = ACTIONS(3037),
    [anon_sym_grep] = ACTIONS(3037),
    [anon_sym_sort] = ACTIONS(3037),
    [anon_sym_goto] = ACTIONS(3037),
    [anon_sym_return] = ACTIONS(3037),
    [anon_sym_undef] = ACTIONS(3037),
    [anon_sym_PERCENT2] = ACTIONS(3037),
    [anon_sym_AMP2] = ACTIONS(3037),
    [anon_sym_STAR2] = ACTIONS(3037),
    [anon_sym_use] = ACTIONS(3037),
    [anon_sym_no] = ACTIONS(3037),
    [anon_sym_for] = ACTIONS(3037),
    [anon_sym_foreach] = ACTIONS(3037),
    [anon_sym_last] = ACTIONS(3037),
    [anon_sym_next] = ACTIONS(3037),
    [anon_sym_redo] = ACTIONS(3037),
    [anon_sym_BEGIN] = ACTIONS(3037),
    [anon_sym_INIT] = ACTIONS(3037),
    [anon_sym_CHECK] = ACTIONS(3037),
    [anon_sym_UNITCHECK] = ACTIONS(3037),
    [anon_sym_END] = ACTIONS(3037),
    [anon_sym___FILE__] = ACTIONS(3037),
    [anon_sym___LINE__] = ACTIONS(3037),
    [anon_sym___PACKAGE__] = ACTIONS(3037),
    [anon_sym___SUB__] = ACTIONS(3037),
    [anon_sym_break] = ACTIONS(3037),
    [anon_sym_fork] = ACTIONS(3037),
    [anon_sym_getppid] = ACTIONS(3037),
    [anon_sym_time] = ACTIONS(3037),
    [anon_sym_times] = ACTIONS(3037),
    [anon_sym_wait] = ACTIONS(3037),
    [anon_sym_wantarray] = ACTIONS(3037),
    [anon_sym_abs] = ACTIONS(3037),
    [anon_sym_alarm] = ACTIONS(3037),
    [anon_sym_chop] = ACTIONS(3037),
    [anon_sym_chdir] = ACTIONS(3037),
    [anon_sym_close] = ACTIONS(3037),
    [anon_sym_closedir] = ACTIONS(3037),
    [anon_sym_caller] = ACTIONS(3037),
    [anon_sym_chomp] = ACTIONS(3037),
    [anon_sym_chr] = ACTIONS(3037),
    [anon_sym_cos] = ACTIONS(3037),
    [anon_sym_chroot] = ACTIONS(3037),
    [anon_sym_defined] = ACTIONS(3037),
    [anon_sym_delete] = ACTIONS(3037),
    [anon_sym_dbmclose] = ACTIONS(3037),
    [anon_sym_exists] = ACTIONS(3037),
    [anon_sym_exit] = ACTIONS(3037),
    [anon_sym_eof] = ACTIONS(3037),
    [anon_sym_exp] = ACTIONS(3037),
    [anon_sym_each] = ACTIONS(3037),
    [anon_sym_fc] = ACTIONS(3037),
    [anon_sym_fileno] = ACTIONS(3037),
    [anon_sym_gmtime] = ACTIONS(3037),
    [anon_sym_getc] = ACTIONS(3037),
    [anon_sym_getpgrp] = ACTIONS(3037),
    [anon_sym_getprotobyname] = ACTIONS(3037),
    [anon_sym_getpwname] = ACTIONS(3037),
    [anon_sym_getpwuid] = ACTIONS(3037),
    [anon_sym_getpeername] = ACTIONS(3037),
    [anon_sym_getnetbyname] = ACTIONS(3037),
    [anon_sym_getsockname] = ACTIONS(3037),
    [anon_sym_getgrnam] = ACTIONS(3037),
    [anon_sym_getgrgid] = ACTIONS(3037),
    [anon_sym_hex] = ACTIONS(3037),
    [anon_sym_int] = ACTIONS(3037),
    [anon_sym_keys] = ACTIONS(3037),
    [anon_sym_lc] = ACTIONS(3037),
    [anon_sym_lcfirst] = ACTIONS(3037),
    [anon_sym_length] = ACTIONS(3037),
    [anon_sym_localtime] = ACTIONS(3037),
    [anon_sym_log] = ACTIONS(3037),
    [anon_sym_lock] = ACTIONS(3037),
    [anon_sym_lstat] = ACTIONS(3037),
    [anon_sym_oct] = ACTIONS(3037),
    [anon_sym_ord] = ACTIONS(3037),
    [anon_sym_prototype] = ACTIONS(3037),
    [anon_sym_pop] = ACTIONS(3037),
    [anon_sym_pos] = ACTIONS(3037),
    [anon_sym_quotemeta] = ACTIONS(3037),
    [anon_sym_reset] = ACTIONS(3037),
    [anon_sym_rand] = ACTIONS(3037),
    [anon_sym_rmdir] = ACTIONS(3037),
    [anon_sym_readdir] = ACTIONS(3037),
    [anon_sym_readline] = ACTIONS(3037),
    [anon_sym_readpipe] = ACTIONS(3037),
    [anon_sym_rewinddir] = ACTIONS(3037),
    [anon_sym_readlink] = ACTIONS(3037),
    [anon_sym_ref] = ACTIONS(3037),
    [anon_sym_scalar] = ACTIONS(3037),
    [anon_sym_shift] = ACTIONS(3037),
    [anon_sym_sin] = ACTIONS(3037),
    [anon_sym_sleep] = ACTIONS(3037),
    [anon_sym_sqrt] = ACTIONS(3037),
    [anon_sym_srand] = ACTIONS(3037),
    [anon_sym_stat] = ACTIONS(3037),
    [anon_sym_study] = ACTIONS(3037),
    [anon_sym_tell] = ACTIONS(3037),
    [anon_sym_telldir] = ACTIONS(3037),
    [anon_sym_tied] = ACTIONS(3037),
    [anon_sym_uc] = ACTIONS(3037),
    [anon_sym_ucfirst] = ACTIONS(3037),
    [anon_sym_untie] = ACTIONS(3037),
    [anon_sym_umask] = ACTIONS(3037),
    [anon_sym_values] = ACTIONS(3037),
    [anon_sym_write] = ACTIONS(3037),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3037),
    [anon_sym___END__] = ACTIONS(3037),
    [anon_sym_EOT] = ACTIONS(3037),
    [anon_sym_q] = ACTIONS(3037),
    [anon_sym_qq] = ACTIONS(3037),
    [anon_sym_qw] = ACTIONS(3037),
    [anon_sym_qx] = ACTIONS(3037),
    [anon_sym_qr] = ACTIONS(3037),
    [anon_sym_m] = ACTIONS(3037),
    [anon_sym_s] = ACTIONS(3037),
    [anon_sym_tr] = ACTIONS(3037),
    [anon_sym_y] = ACTIONS(3037),
    [anon_sym_if] = ACTIONS(3037),
    [anon_sym_unless] = ACTIONS(3037),
    [anon_sym_while] = ACTIONS(3037),
    [anon_sym_until] = ACTIONS(3037),
    [aux_sym__bareword_token1] = ACTIONS(3037),
    [sym_number] = ACTIONS(3037),
    [anon_sym_true] = ACTIONS(3037),
    [anon_sym_false] = ACTIONS(3037),
    [sym__single_quote] = ACTIONS(3035),
    [sym__double_quote] = ACTIONS(3035),
    [sym__backtick_quote] = ACTIONS(3035),
    [sym__search_slash_quote] = ACTIONS(3035),
    [sym__open_readline_bracket] = ACTIONS(3035),
    [sym__open_fileglob_bracket] = ACTIONS(3035),
    [sym__PERLY_HEREDOC] = ACTIONS(3035),
    [sym__ctrl_z_hack] = ACTIONS(3035),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3035),
    [sym__filetest] = ACTIONS(3035),
  },
  [STATE(1471)] = {
    [sym_heredoc_content] = STATE(1471),
    [ts_builtin_sym_end] = ACTIONS(3039),
    [sym__identifier] = ACTIONS(3041),
    [anon_sym_LBRACE] = ACTIONS(3041),
    [anon_sym_SEMI] = ACTIONS(3041),
    [anon_sym_package] = ACTIONS(3041),
    [anon_sym_class] = ACTIONS(3041),
    [anon_sym_role] = ACTIONS(3041),
    [anon_sym_BUILD] = ACTIONS(3041),
    [anon_sym_ADJUST] = ACTIONS(3041),
    [anon_sym_DOLLAR] = ACTIONS(3041),
    [anon_sym_AT] = ACTIONS(3041),
    [anon_sym_my] = ACTIONS(3041),
    [anon_sym_extended] = ACTIONS(3041),
    [anon_sym_async] = ACTIONS(3041),
    [anon_sym_sub] = ACTIONS(3041),
    [anon_sym_method] = ACTIONS(3041),
    [anon_sym_LPAREN] = ACTIONS(3041),
    [anon_sym_continue] = ACTIONS(3041),
    [anon_sym_state] = ACTIONS(3041),
    [anon_sym_our] = ACTIONS(3041),
    [anon_sym_try] = ACTIONS(3041),
    [anon_sym_defer] = ACTIONS(3041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3041),
    [anon_sym_LBRACK] = ACTIONS(3041),
    [anon_sym_LT_LT] = ACTIONS(3041),
    [anon_sym_SLASH_SLASH] = ACTIONS(3041),
    [anon_sym_PLUS] = ACTIONS(3041),
    [anon_sym_DASH] = ACTIONS(3041),
    [anon_sym_TILDE] = ACTIONS(3041),
    [anon_sym_BANG] = ACTIONS(3041),
    [anon_sym_PLUS_PLUS] = ACTIONS(3041),
    [anon_sym_DASH_DASH] = ACTIONS(3041),
    [anon_sym_BSLASH] = ACTIONS(3039),
    [anon_sym_do] = ACTIONS(3041),
    [anon_sym_eval] = ACTIONS(3041),
    [anon_sym_await] = ACTIONS(3041),
    [anon_sym_field] = ACTIONS(3041),
    [anon_sym_local] = ACTIONS(3041),
    [anon_sym_dynamically] = ACTIONS(3041),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3041),
    [anon_sym_require] = ACTIONS(3041),
    [anon_sym_map] = ACTIONS(3041),
    [anon_sym_grep] = ACTIONS(3041),
    [anon_sym_sort] = ACTIONS(3041),
    [anon_sym_goto] = ACTIONS(3041),
    [anon_sym_return] = ACTIONS(3041),
    [anon_sym_undef] = ACTIONS(3041),
    [anon_sym_PERCENT2] = ACTIONS(3041),
    [anon_sym_AMP2] = ACTIONS(3041),
    [anon_sym_STAR2] = ACTIONS(3041),
    [anon_sym_use] = ACTIONS(3041),
    [anon_sym_no] = ACTIONS(3041),
    [anon_sym_for] = ACTIONS(3041),
    [anon_sym_foreach] = ACTIONS(3041),
    [anon_sym_last] = ACTIONS(3041),
    [anon_sym_next] = ACTIONS(3041),
    [anon_sym_redo] = ACTIONS(3041),
    [anon_sym_BEGIN] = ACTIONS(3041),
    [anon_sym_INIT] = ACTIONS(3041),
    [anon_sym_CHECK] = ACTIONS(3041),
    [anon_sym_UNITCHECK] = ACTIONS(3041),
    [anon_sym_END] = ACTIONS(3041),
    [anon_sym___FILE__] = ACTIONS(3041),
    [anon_sym___LINE__] = ACTIONS(3041),
    [anon_sym___PACKAGE__] = ACTIONS(3041),
    [anon_sym___SUB__] = ACTIONS(3041),
    [anon_sym_break] = ACTIONS(3041),
    [anon_sym_fork] = ACTIONS(3041),
    [anon_sym_getppid] = ACTIONS(3041),
    [anon_sym_time] = ACTIONS(3041),
    [anon_sym_times] = ACTIONS(3041),
    [anon_sym_wait] = ACTIONS(3041),
    [anon_sym_wantarray] = ACTIONS(3041),
    [anon_sym_abs] = ACTIONS(3041),
    [anon_sym_alarm] = ACTIONS(3041),
    [anon_sym_chop] = ACTIONS(3041),
    [anon_sym_chdir] = ACTIONS(3041),
    [anon_sym_close] = ACTIONS(3041),
    [anon_sym_closedir] = ACTIONS(3041),
    [anon_sym_caller] = ACTIONS(3041),
    [anon_sym_chomp] = ACTIONS(3041),
    [anon_sym_chr] = ACTIONS(3041),
    [anon_sym_cos] = ACTIONS(3041),
    [anon_sym_chroot] = ACTIONS(3041),
    [anon_sym_defined] = ACTIONS(3041),
    [anon_sym_delete] = ACTIONS(3041),
    [anon_sym_dbmclose] = ACTIONS(3041),
    [anon_sym_exists] = ACTIONS(3041),
    [anon_sym_exit] = ACTIONS(3041),
    [anon_sym_eof] = ACTIONS(3041),
    [anon_sym_exp] = ACTIONS(3041),
    [anon_sym_each] = ACTIONS(3041),
    [anon_sym_fc] = ACTIONS(3041),
    [anon_sym_fileno] = ACTIONS(3041),
    [anon_sym_gmtime] = ACTIONS(3041),
    [anon_sym_getc] = ACTIONS(3041),
    [anon_sym_getpgrp] = ACTIONS(3041),
    [anon_sym_getprotobyname] = ACTIONS(3041),
    [anon_sym_getpwname] = ACTIONS(3041),
    [anon_sym_getpwuid] = ACTIONS(3041),
    [anon_sym_getpeername] = ACTIONS(3041),
    [anon_sym_getnetbyname] = ACTIONS(3041),
    [anon_sym_getsockname] = ACTIONS(3041),
    [anon_sym_getgrnam] = ACTIONS(3041),
    [anon_sym_getgrgid] = ACTIONS(3041),
    [anon_sym_hex] = ACTIONS(3041),
    [anon_sym_int] = ACTIONS(3041),
    [anon_sym_keys] = ACTIONS(3041),
    [anon_sym_lc] = ACTIONS(3041),
    [anon_sym_lcfirst] = ACTIONS(3041),
    [anon_sym_length] = ACTIONS(3041),
    [anon_sym_localtime] = ACTIONS(3041),
    [anon_sym_log] = ACTIONS(3041),
    [anon_sym_lock] = ACTIONS(3041),
    [anon_sym_lstat] = ACTIONS(3041),
    [anon_sym_oct] = ACTIONS(3041),
    [anon_sym_ord] = ACTIONS(3041),
    [anon_sym_prototype] = ACTIONS(3041),
    [anon_sym_pop] = ACTIONS(3041),
    [anon_sym_pos] = ACTIONS(3041),
    [anon_sym_quotemeta] = ACTIONS(3041),
    [anon_sym_reset] = ACTIONS(3041),
    [anon_sym_rand] = ACTIONS(3041),
    [anon_sym_rmdir] = ACTIONS(3041),
    [anon_sym_readdir] = ACTIONS(3041),
    [anon_sym_readline] = ACTIONS(3041),
    [anon_sym_readpipe] = ACTIONS(3041),
    [anon_sym_rewinddir] = ACTIONS(3041),
    [anon_sym_readlink] = ACTIONS(3041),
    [anon_sym_ref] = ACTIONS(3041),
    [anon_sym_scalar] = ACTIONS(3041),
    [anon_sym_shift] = ACTIONS(3041),
    [anon_sym_sin] = ACTIONS(3041),
    [anon_sym_sleep] = ACTIONS(3041),
    [anon_sym_sqrt] = ACTIONS(3041),
    [anon_sym_srand] = ACTIONS(3041),
    [anon_sym_stat] = ACTIONS(3041),
    [anon_sym_study] = ACTIONS(3041),
    [anon_sym_tell] = ACTIONS(3041),
    [anon_sym_telldir] = ACTIONS(3041),
    [anon_sym_tied] = ACTIONS(3041),
    [anon_sym_uc] = ACTIONS(3041),
    [anon_sym_ucfirst] = ACTIONS(3041),
    [anon_sym_untie] = ACTIONS(3041),
    [anon_sym_umask] = ACTIONS(3041),
    [anon_sym_values] = ACTIONS(3041),
    [anon_sym_write] = ACTIONS(3041),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3041),
    [anon_sym___END__] = ACTIONS(3041),
    [anon_sym_EOT] = ACTIONS(3041),
    [anon_sym_q] = ACTIONS(3041),
    [anon_sym_qq] = ACTIONS(3041),
    [anon_sym_qw] = ACTIONS(3041),
    [anon_sym_qx] = ACTIONS(3041),
    [anon_sym_qr] = ACTIONS(3041),
    [anon_sym_m] = ACTIONS(3041),
    [anon_sym_s] = ACTIONS(3041),
    [anon_sym_tr] = ACTIONS(3041),
    [anon_sym_y] = ACTIONS(3041),
    [anon_sym_if] = ACTIONS(3041),
    [anon_sym_unless] = ACTIONS(3041),
    [anon_sym_while] = ACTIONS(3041),
    [anon_sym_until] = ACTIONS(3041),
    [aux_sym__bareword_token1] = ACTIONS(3041),
    [sym_number] = ACTIONS(3041),
    [anon_sym_true] = ACTIONS(3041),
    [anon_sym_false] = ACTIONS(3041),
    [sym__single_quote] = ACTIONS(3039),
    [sym__double_quote] = ACTIONS(3039),
    [sym__backtick_quote] = ACTIONS(3039),
    [sym__search_slash_quote] = ACTIONS(3039),
    [sym__open_readline_bracket] = ACTIONS(3039),
    [sym__open_fileglob_bracket] = ACTIONS(3039),
    [sym__PERLY_HEREDOC] = ACTIONS(3039),
    [sym__ctrl_z_hack] = ACTIONS(3039),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3039),
    [sym__filetest] = ACTIONS(3039),
  },
  [STATE(1472)] = {
    [sym_heredoc_content] = STATE(1472),
    [ts_builtin_sym_end] = ACTIONS(3043),
    [sym__identifier] = ACTIONS(3045),
    [anon_sym_LBRACE] = ACTIONS(3045),
    [anon_sym_SEMI] = ACTIONS(3045),
    [anon_sym_package] = ACTIONS(3045),
    [anon_sym_class] = ACTIONS(3045),
    [anon_sym_role] = ACTIONS(3045),
    [anon_sym_BUILD] = ACTIONS(3045),
    [anon_sym_ADJUST] = ACTIONS(3045),
    [anon_sym_DOLLAR] = ACTIONS(3045),
    [anon_sym_AT] = ACTIONS(3045),
    [anon_sym_my] = ACTIONS(3045),
    [anon_sym_extended] = ACTIONS(3045),
    [anon_sym_async] = ACTIONS(3045),
    [anon_sym_sub] = ACTIONS(3045),
    [anon_sym_method] = ACTIONS(3045),
    [anon_sym_LPAREN] = ACTIONS(3045),
    [anon_sym_continue] = ACTIONS(3045),
    [anon_sym_state] = ACTIONS(3045),
    [anon_sym_our] = ACTIONS(3045),
    [anon_sym_try] = ACTIONS(3045),
    [anon_sym_defer] = ACTIONS(3045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3045),
    [anon_sym_LBRACK] = ACTIONS(3045),
    [anon_sym_LT_LT] = ACTIONS(3045),
    [anon_sym_SLASH_SLASH] = ACTIONS(3045),
    [anon_sym_PLUS] = ACTIONS(3045),
    [anon_sym_DASH] = ACTIONS(3045),
    [anon_sym_TILDE] = ACTIONS(3045),
    [anon_sym_BANG] = ACTIONS(3045),
    [anon_sym_PLUS_PLUS] = ACTIONS(3045),
    [anon_sym_DASH_DASH] = ACTIONS(3045),
    [anon_sym_BSLASH] = ACTIONS(3043),
    [anon_sym_do] = ACTIONS(3045),
    [anon_sym_eval] = ACTIONS(3045),
    [anon_sym_await] = ACTIONS(3045),
    [anon_sym_field] = ACTIONS(3045),
    [anon_sym_local] = ACTIONS(3045),
    [anon_sym_dynamically] = ACTIONS(3045),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3045),
    [anon_sym_require] = ACTIONS(3045),
    [anon_sym_map] = ACTIONS(3045),
    [anon_sym_grep] = ACTIONS(3045),
    [anon_sym_sort] = ACTIONS(3045),
    [anon_sym_goto] = ACTIONS(3045),
    [anon_sym_return] = ACTIONS(3045),
    [anon_sym_undef] = ACTIONS(3045),
    [anon_sym_PERCENT2] = ACTIONS(3045),
    [anon_sym_AMP2] = ACTIONS(3045),
    [anon_sym_STAR2] = ACTIONS(3045),
    [anon_sym_use] = ACTIONS(3045),
    [anon_sym_no] = ACTIONS(3045),
    [anon_sym_for] = ACTIONS(3045),
    [anon_sym_foreach] = ACTIONS(3045),
    [anon_sym_last] = ACTIONS(3045),
    [anon_sym_next] = ACTIONS(3045),
    [anon_sym_redo] = ACTIONS(3045),
    [anon_sym_BEGIN] = ACTIONS(3045),
    [anon_sym_INIT] = ACTIONS(3045),
    [anon_sym_CHECK] = ACTIONS(3045),
    [anon_sym_UNITCHECK] = ACTIONS(3045),
    [anon_sym_END] = ACTIONS(3045),
    [anon_sym___FILE__] = ACTIONS(3045),
    [anon_sym___LINE__] = ACTIONS(3045),
    [anon_sym___PACKAGE__] = ACTIONS(3045),
    [anon_sym___SUB__] = ACTIONS(3045),
    [anon_sym_break] = ACTIONS(3045),
    [anon_sym_fork] = ACTIONS(3045),
    [anon_sym_getppid] = ACTIONS(3045),
    [anon_sym_time] = ACTIONS(3045),
    [anon_sym_times] = ACTIONS(3045),
    [anon_sym_wait] = ACTIONS(3045),
    [anon_sym_wantarray] = ACTIONS(3045),
    [anon_sym_abs] = ACTIONS(3045),
    [anon_sym_alarm] = ACTIONS(3045),
    [anon_sym_chop] = ACTIONS(3045),
    [anon_sym_chdir] = ACTIONS(3045),
    [anon_sym_close] = ACTIONS(3045),
    [anon_sym_closedir] = ACTIONS(3045),
    [anon_sym_caller] = ACTIONS(3045),
    [anon_sym_chomp] = ACTIONS(3045),
    [anon_sym_chr] = ACTIONS(3045),
    [anon_sym_cos] = ACTIONS(3045),
    [anon_sym_chroot] = ACTIONS(3045),
    [anon_sym_defined] = ACTIONS(3045),
    [anon_sym_delete] = ACTIONS(3045),
    [anon_sym_dbmclose] = ACTIONS(3045),
    [anon_sym_exists] = ACTIONS(3045),
    [anon_sym_exit] = ACTIONS(3045),
    [anon_sym_eof] = ACTIONS(3045),
    [anon_sym_exp] = ACTIONS(3045),
    [anon_sym_each] = ACTIONS(3045),
    [anon_sym_fc] = ACTIONS(3045),
    [anon_sym_fileno] = ACTIONS(3045),
    [anon_sym_gmtime] = ACTIONS(3045),
    [anon_sym_getc] = ACTIONS(3045),
    [anon_sym_getpgrp] = ACTIONS(3045),
    [anon_sym_getprotobyname] = ACTIONS(3045),
    [anon_sym_getpwname] = ACTIONS(3045),
    [anon_sym_getpwuid] = ACTIONS(3045),
    [anon_sym_getpeername] = ACTIONS(3045),
    [anon_sym_getnetbyname] = ACTIONS(3045),
    [anon_sym_getsockname] = ACTIONS(3045),
    [anon_sym_getgrnam] = ACTIONS(3045),
    [anon_sym_getgrgid] = ACTIONS(3045),
    [anon_sym_hex] = ACTIONS(3045),
    [anon_sym_int] = ACTIONS(3045),
    [anon_sym_keys] = ACTIONS(3045),
    [anon_sym_lc] = ACTIONS(3045),
    [anon_sym_lcfirst] = ACTIONS(3045),
    [anon_sym_length] = ACTIONS(3045),
    [anon_sym_localtime] = ACTIONS(3045),
    [anon_sym_log] = ACTIONS(3045),
    [anon_sym_lock] = ACTIONS(3045),
    [anon_sym_lstat] = ACTIONS(3045),
    [anon_sym_oct] = ACTIONS(3045),
    [anon_sym_ord] = ACTIONS(3045),
    [anon_sym_prototype] = ACTIONS(3045),
    [anon_sym_pop] = ACTIONS(3045),
    [anon_sym_pos] = ACTIONS(3045),
    [anon_sym_quotemeta] = ACTIONS(3045),
    [anon_sym_reset] = ACTIONS(3045),
    [anon_sym_rand] = ACTIONS(3045),
    [anon_sym_rmdir] = ACTIONS(3045),
    [anon_sym_readdir] = ACTIONS(3045),
    [anon_sym_readline] = ACTIONS(3045),
    [anon_sym_readpipe] = ACTIONS(3045),
    [anon_sym_rewinddir] = ACTIONS(3045),
    [anon_sym_readlink] = ACTIONS(3045),
    [anon_sym_ref] = ACTIONS(3045),
    [anon_sym_scalar] = ACTIONS(3045),
    [anon_sym_shift] = ACTIONS(3045),
    [anon_sym_sin] = ACTIONS(3045),
    [anon_sym_sleep] = ACTIONS(3045),
    [anon_sym_sqrt] = ACTIONS(3045),
    [anon_sym_srand] = ACTIONS(3045),
    [anon_sym_stat] = ACTIONS(3045),
    [anon_sym_study] = ACTIONS(3045),
    [anon_sym_tell] = ACTIONS(3045),
    [anon_sym_telldir] = ACTIONS(3045),
    [anon_sym_tied] = ACTIONS(3045),
    [anon_sym_uc] = ACTIONS(3045),
    [anon_sym_ucfirst] = ACTIONS(3045),
    [anon_sym_untie] = ACTIONS(3045),
    [anon_sym_umask] = ACTIONS(3045),
    [anon_sym_values] = ACTIONS(3045),
    [anon_sym_write] = ACTIONS(3045),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3045),
    [anon_sym___END__] = ACTIONS(3045),
    [anon_sym_EOT] = ACTIONS(3045),
    [anon_sym_q] = ACTIONS(3045),
    [anon_sym_qq] = ACTIONS(3045),
    [anon_sym_qw] = ACTIONS(3045),
    [anon_sym_qx] = ACTIONS(3045),
    [anon_sym_qr] = ACTIONS(3045),
    [anon_sym_m] = ACTIONS(3045),
    [anon_sym_s] = ACTIONS(3045),
    [anon_sym_tr] = ACTIONS(3045),
    [anon_sym_y] = ACTIONS(3045),
    [anon_sym_if] = ACTIONS(3045),
    [anon_sym_unless] = ACTIONS(3045),
    [anon_sym_while] = ACTIONS(3045),
    [anon_sym_until] = ACTIONS(3045),
    [aux_sym__bareword_token1] = ACTIONS(3045),
    [sym_number] = ACTIONS(3045),
    [anon_sym_true] = ACTIONS(3045),
    [anon_sym_false] = ACTIONS(3045),
    [sym__single_quote] = ACTIONS(3043),
    [sym__double_quote] = ACTIONS(3043),
    [sym__backtick_quote] = ACTIONS(3043),
    [sym__search_slash_quote] = ACTIONS(3043),
    [sym__open_readline_bracket] = ACTIONS(3043),
    [sym__open_fileglob_bracket] = ACTIONS(3043),
    [sym__PERLY_HEREDOC] = ACTIONS(3043),
    [sym__ctrl_z_hack] = ACTIONS(3043),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3043),
    [sym__filetest] = ACTIONS(3043),
  },
  [STATE(1473)] = {
    [sym_heredoc_content] = STATE(1473),
    [ts_builtin_sym_end] = ACTIONS(3047),
    [sym__identifier] = ACTIONS(3049),
    [anon_sym_LBRACE] = ACTIONS(3049),
    [anon_sym_SEMI] = ACTIONS(3049),
    [anon_sym_package] = ACTIONS(3049),
    [anon_sym_class] = ACTIONS(3049),
    [anon_sym_role] = ACTIONS(3049),
    [anon_sym_BUILD] = ACTIONS(3049),
    [anon_sym_ADJUST] = ACTIONS(3049),
    [anon_sym_DOLLAR] = ACTIONS(3049),
    [anon_sym_AT] = ACTIONS(3049),
    [anon_sym_my] = ACTIONS(3049),
    [anon_sym_extended] = ACTIONS(3049),
    [anon_sym_async] = ACTIONS(3049),
    [anon_sym_sub] = ACTIONS(3049),
    [anon_sym_method] = ACTIONS(3049),
    [anon_sym_LPAREN] = ACTIONS(3049),
    [anon_sym_continue] = ACTIONS(3049),
    [anon_sym_state] = ACTIONS(3049),
    [anon_sym_our] = ACTIONS(3049),
    [anon_sym_try] = ACTIONS(3049),
    [anon_sym_defer] = ACTIONS(3049),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3049),
    [anon_sym_LBRACK] = ACTIONS(3049),
    [anon_sym_LT_LT] = ACTIONS(3049),
    [anon_sym_SLASH_SLASH] = ACTIONS(3049),
    [anon_sym_PLUS] = ACTIONS(3049),
    [anon_sym_DASH] = ACTIONS(3049),
    [anon_sym_TILDE] = ACTIONS(3049),
    [anon_sym_BANG] = ACTIONS(3049),
    [anon_sym_PLUS_PLUS] = ACTIONS(3049),
    [anon_sym_DASH_DASH] = ACTIONS(3049),
    [anon_sym_BSLASH] = ACTIONS(3047),
    [anon_sym_do] = ACTIONS(3049),
    [anon_sym_eval] = ACTIONS(3049),
    [anon_sym_await] = ACTIONS(3049),
    [anon_sym_field] = ACTIONS(3049),
    [anon_sym_local] = ACTIONS(3049),
    [anon_sym_dynamically] = ACTIONS(3049),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3049),
    [anon_sym_require] = ACTIONS(3049),
    [anon_sym_map] = ACTIONS(3049),
    [anon_sym_grep] = ACTIONS(3049),
    [anon_sym_sort] = ACTIONS(3049),
    [anon_sym_goto] = ACTIONS(3049),
    [anon_sym_return] = ACTIONS(3049),
    [anon_sym_undef] = ACTIONS(3049),
    [anon_sym_PERCENT2] = ACTIONS(3049),
    [anon_sym_AMP2] = ACTIONS(3049),
    [anon_sym_STAR2] = ACTIONS(3049),
    [anon_sym_use] = ACTIONS(3049),
    [anon_sym_no] = ACTIONS(3049),
    [anon_sym_for] = ACTIONS(3049),
    [anon_sym_foreach] = ACTIONS(3049),
    [anon_sym_last] = ACTIONS(3049),
    [anon_sym_next] = ACTIONS(3049),
    [anon_sym_redo] = ACTIONS(3049),
    [anon_sym_BEGIN] = ACTIONS(3049),
    [anon_sym_INIT] = ACTIONS(3049),
    [anon_sym_CHECK] = ACTIONS(3049),
    [anon_sym_UNITCHECK] = ACTIONS(3049),
    [anon_sym_END] = ACTIONS(3049),
    [anon_sym___FILE__] = ACTIONS(3049),
    [anon_sym___LINE__] = ACTIONS(3049),
    [anon_sym___PACKAGE__] = ACTIONS(3049),
    [anon_sym___SUB__] = ACTIONS(3049),
    [anon_sym_break] = ACTIONS(3049),
    [anon_sym_fork] = ACTIONS(3049),
    [anon_sym_getppid] = ACTIONS(3049),
    [anon_sym_time] = ACTIONS(3049),
    [anon_sym_times] = ACTIONS(3049),
    [anon_sym_wait] = ACTIONS(3049),
    [anon_sym_wantarray] = ACTIONS(3049),
    [anon_sym_abs] = ACTIONS(3049),
    [anon_sym_alarm] = ACTIONS(3049),
    [anon_sym_chop] = ACTIONS(3049),
    [anon_sym_chdir] = ACTIONS(3049),
    [anon_sym_close] = ACTIONS(3049),
    [anon_sym_closedir] = ACTIONS(3049),
    [anon_sym_caller] = ACTIONS(3049),
    [anon_sym_chomp] = ACTIONS(3049),
    [anon_sym_chr] = ACTIONS(3049),
    [anon_sym_cos] = ACTIONS(3049),
    [anon_sym_chroot] = ACTIONS(3049),
    [anon_sym_defined] = ACTIONS(3049),
    [anon_sym_delete] = ACTIONS(3049),
    [anon_sym_dbmclose] = ACTIONS(3049),
    [anon_sym_exists] = ACTIONS(3049),
    [anon_sym_exit] = ACTIONS(3049),
    [anon_sym_eof] = ACTIONS(3049),
    [anon_sym_exp] = ACTIONS(3049),
    [anon_sym_each] = ACTIONS(3049),
    [anon_sym_fc] = ACTIONS(3049),
    [anon_sym_fileno] = ACTIONS(3049),
    [anon_sym_gmtime] = ACTIONS(3049),
    [anon_sym_getc] = ACTIONS(3049),
    [anon_sym_getpgrp] = ACTIONS(3049),
    [anon_sym_getprotobyname] = ACTIONS(3049),
    [anon_sym_getpwname] = ACTIONS(3049),
    [anon_sym_getpwuid] = ACTIONS(3049),
    [anon_sym_getpeername] = ACTIONS(3049),
    [anon_sym_getnetbyname] = ACTIONS(3049),
    [anon_sym_getsockname] = ACTIONS(3049),
    [anon_sym_getgrnam] = ACTIONS(3049),
    [anon_sym_getgrgid] = ACTIONS(3049),
    [anon_sym_hex] = ACTIONS(3049),
    [anon_sym_int] = ACTIONS(3049),
    [anon_sym_keys] = ACTIONS(3049),
    [anon_sym_lc] = ACTIONS(3049),
    [anon_sym_lcfirst] = ACTIONS(3049),
    [anon_sym_length] = ACTIONS(3049),
    [anon_sym_localtime] = ACTIONS(3049),
    [anon_sym_log] = ACTIONS(3049),
    [anon_sym_lock] = ACTIONS(3049),
    [anon_sym_lstat] = ACTIONS(3049),
    [anon_sym_oct] = ACTIONS(3049),
    [anon_sym_ord] = ACTIONS(3049),
    [anon_sym_prototype] = ACTIONS(3049),
    [anon_sym_pop] = ACTIONS(3049),
    [anon_sym_pos] = ACTIONS(3049),
    [anon_sym_quotemeta] = ACTIONS(3049),
    [anon_sym_reset] = ACTIONS(3049),
    [anon_sym_rand] = ACTIONS(3049),
    [anon_sym_rmdir] = ACTIONS(3049),
    [anon_sym_readdir] = ACTIONS(3049),
    [anon_sym_readline] = ACTIONS(3049),
    [anon_sym_readpipe] = ACTIONS(3049),
    [anon_sym_rewinddir] = ACTIONS(3049),
    [anon_sym_readlink] = ACTIONS(3049),
    [anon_sym_ref] = ACTIONS(3049),
    [anon_sym_scalar] = ACTIONS(3049),
    [anon_sym_shift] = ACTIONS(3049),
    [anon_sym_sin] = ACTIONS(3049),
    [anon_sym_sleep] = ACTIONS(3049),
    [anon_sym_sqrt] = ACTIONS(3049),
    [anon_sym_srand] = ACTIONS(3049),
    [anon_sym_stat] = ACTIONS(3049),
    [anon_sym_study] = ACTIONS(3049),
    [anon_sym_tell] = ACTIONS(3049),
    [anon_sym_telldir] = ACTIONS(3049),
    [anon_sym_tied] = ACTIONS(3049),
    [anon_sym_uc] = ACTIONS(3049),
    [anon_sym_ucfirst] = ACTIONS(3049),
    [anon_sym_untie] = ACTIONS(3049),
    [anon_sym_umask] = ACTIONS(3049),
    [anon_sym_values] = ACTIONS(3049),
    [anon_sym_write] = ACTIONS(3049),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3049),
    [anon_sym___END__] = ACTIONS(3049),
    [anon_sym_EOT] = ACTIONS(3049),
    [anon_sym_q] = ACTIONS(3049),
    [anon_sym_qq] = ACTIONS(3049),
    [anon_sym_qw] = ACTIONS(3049),
    [anon_sym_qx] = ACTIONS(3049),
    [anon_sym_qr] = ACTIONS(3049),
    [anon_sym_m] = ACTIONS(3049),
    [anon_sym_s] = ACTIONS(3049),
    [anon_sym_tr] = ACTIONS(3049),
    [anon_sym_y] = ACTIONS(3049),
    [anon_sym_if] = ACTIONS(3049),
    [anon_sym_unless] = ACTIONS(3049),
    [anon_sym_while] = ACTIONS(3049),
    [anon_sym_until] = ACTIONS(3049),
    [aux_sym__bareword_token1] = ACTIONS(3049),
    [sym_number] = ACTIONS(3049),
    [anon_sym_true] = ACTIONS(3049),
    [anon_sym_false] = ACTIONS(3049),
    [sym__single_quote] = ACTIONS(3047),
    [sym__double_quote] = ACTIONS(3047),
    [sym__backtick_quote] = ACTIONS(3047),
    [sym__search_slash_quote] = ACTIONS(3047),
    [sym__open_readline_bracket] = ACTIONS(3047),
    [sym__open_fileglob_bracket] = ACTIONS(3047),
    [sym__PERLY_HEREDOC] = ACTIONS(3047),
    [sym__ctrl_z_hack] = ACTIONS(3047),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3047),
    [sym__filetest] = ACTIONS(3047),
  },
  [STATE(1474)] = {
    [sym_heredoc_content] = STATE(1474),
    [ts_builtin_sym_end] = ACTIONS(3051),
    [sym__identifier] = ACTIONS(3053),
    [anon_sym_LBRACE] = ACTIONS(3053),
    [anon_sym_SEMI] = ACTIONS(3053),
    [anon_sym_package] = ACTIONS(3053),
    [anon_sym_class] = ACTIONS(3053),
    [anon_sym_role] = ACTIONS(3053),
    [anon_sym_BUILD] = ACTIONS(3053),
    [anon_sym_ADJUST] = ACTIONS(3053),
    [anon_sym_DOLLAR] = ACTIONS(3053),
    [anon_sym_AT] = ACTIONS(3053),
    [anon_sym_my] = ACTIONS(3053),
    [anon_sym_extended] = ACTIONS(3053),
    [anon_sym_async] = ACTIONS(3053),
    [anon_sym_sub] = ACTIONS(3053),
    [anon_sym_method] = ACTIONS(3053),
    [anon_sym_LPAREN] = ACTIONS(3053),
    [anon_sym_continue] = ACTIONS(3053),
    [anon_sym_state] = ACTIONS(3053),
    [anon_sym_our] = ACTIONS(3053),
    [anon_sym_try] = ACTIONS(3053),
    [anon_sym_defer] = ACTIONS(3053),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3053),
    [anon_sym_LBRACK] = ACTIONS(3053),
    [anon_sym_LT_LT] = ACTIONS(3053),
    [anon_sym_SLASH_SLASH] = ACTIONS(3053),
    [anon_sym_PLUS] = ACTIONS(3053),
    [anon_sym_DASH] = ACTIONS(3053),
    [anon_sym_TILDE] = ACTIONS(3053),
    [anon_sym_BANG] = ACTIONS(3053),
    [anon_sym_PLUS_PLUS] = ACTIONS(3053),
    [anon_sym_DASH_DASH] = ACTIONS(3053),
    [anon_sym_BSLASH] = ACTIONS(3051),
    [anon_sym_do] = ACTIONS(3053),
    [anon_sym_eval] = ACTIONS(3053),
    [anon_sym_await] = ACTIONS(3053),
    [anon_sym_field] = ACTIONS(3053),
    [anon_sym_local] = ACTIONS(3053),
    [anon_sym_dynamically] = ACTIONS(3053),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3053),
    [anon_sym_require] = ACTIONS(3053),
    [anon_sym_map] = ACTIONS(3053),
    [anon_sym_grep] = ACTIONS(3053),
    [anon_sym_sort] = ACTIONS(3053),
    [anon_sym_goto] = ACTIONS(3053),
    [anon_sym_return] = ACTIONS(3053),
    [anon_sym_undef] = ACTIONS(3053),
    [anon_sym_PERCENT2] = ACTIONS(3053),
    [anon_sym_AMP2] = ACTIONS(3053),
    [anon_sym_STAR2] = ACTIONS(3053),
    [anon_sym_use] = ACTIONS(3053),
    [anon_sym_no] = ACTIONS(3053),
    [anon_sym_for] = ACTIONS(3053),
    [anon_sym_foreach] = ACTIONS(3053),
    [anon_sym_last] = ACTIONS(3053),
    [anon_sym_next] = ACTIONS(3053),
    [anon_sym_redo] = ACTIONS(3053),
    [anon_sym_BEGIN] = ACTIONS(3053),
    [anon_sym_INIT] = ACTIONS(3053),
    [anon_sym_CHECK] = ACTIONS(3053),
    [anon_sym_UNITCHECK] = ACTIONS(3053),
    [anon_sym_END] = ACTIONS(3053),
    [anon_sym___FILE__] = ACTIONS(3053),
    [anon_sym___LINE__] = ACTIONS(3053),
    [anon_sym___PACKAGE__] = ACTIONS(3053),
    [anon_sym___SUB__] = ACTIONS(3053),
    [anon_sym_break] = ACTIONS(3053),
    [anon_sym_fork] = ACTIONS(3053),
    [anon_sym_getppid] = ACTIONS(3053),
    [anon_sym_time] = ACTIONS(3053),
    [anon_sym_times] = ACTIONS(3053),
    [anon_sym_wait] = ACTIONS(3053),
    [anon_sym_wantarray] = ACTIONS(3053),
    [anon_sym_abs] = ACTIONS(3053),
    [anon_sym_alarm] = ACTIONS(3053),
    [anon_sym_chop] = ACTIONS(3053),
    [anon_sym_chdir] = ACTIONS(3053),
    [anon_sym_close] = ACTIONS(3053),
    [anon_sym_closedir] = ACTIONS(3053),
    [anon_sym_caller] = ACTIONS(3053),
    [anon_sym_chomp] = ACTIONS(3053),
    [anon_sym_chr] = ACTIONS(3053),
    [anon_sym_cos] = ACTIONS(3053),
    [anon_sym_chroot] = ACTIONS(3053),
    [anon_sym_defined] = ACTIONS(3053),
    [anon_sym_delete] = ACTIONS(3053),
    [anon_sym_dbmclose] = ACTIONS(3053),
    [anon_sym_exists] = ACTIONS(3053),
    [anon_sym_exit] = ACTIONS(3053),
    [anon_sym_eof] = ACTIONS(3053),
    [anon_sym_exp] = ACTIONS(3053),
    [anon_sym_each] = ACTIONS(3053),
    [anon_sym_fc] = ACTIONS(3053),
    [anon_sym_fileno] = ACTIONS(3053),
    [anon_sym_gmtime] = ACTIONS(3053),
    [anon_sym_getc] = ACTIONS(3053),
    [anon_sym_getpgrp] = ACTIONS(3053),
    [anon_sym_getprotobyname] = ACTIONS(3053),
    [anon_sym_getpwname] = ACTIONS(3053),
    [anon_sym_getpwuid] = ACTIONS(3053),
    [anon_sym_getpeername] = ACTIONS(3053),
    [anon_sym_getnetbyname] = ACTIONS(3053),
    [anon_sym_getsockname] = ACTIONS(3053),
    [anon_sym_getgrnam] = ACTIONS(3053),
    [anon_sym_getgrgid] = ACTIONS(3053),
    [anon_sym_hex] = ACTIONS(3053),
    [anon_sym_int] = ACTIONS(3053),
    [anon_sym_keys] = ACTIONS(3053),
    [anon_sym_lc] = ACTIONS(3053),
    [anon_sym_lcfirst] = ACTIONS(3053),
    [anon_sym_length] = ACTIONS(3053),
    [anon_sym_localtime] = ACTIONS(3053),
    [anon_sym_log] = ACTIONS(3053),
    [anon_sym_lock] = ACTIONS(3053),
    [anon_sym_lstat] = ACTIONS(3053),
    [anon_sym_oct] = ACTIONS(3053),
    [anon_sym_ord] = ACTIONS(3053),
    [anon_sym_prototype] = ACTIONS(3053),
    [anon_sym_pop] = ACTIONS(3053),
    [anon_sym_pos] = ACTIONS(3053),
    [anon_sym_quotemeta] = ACTIONS(3053),
    [anon_sym_reset] = ACTIONS(3053),
    [anon_sym_rand] = ACTIONS(3053),
    [anon_sym_rmdir] = ACTIONS(3053),
    [anon_sym_readdir] = ACTIONS(3053),
    [anon_sym_readline] = ACTIONS(3053),
    [anon_sym_readpipe] = ACTIONS(3053),
    [anon_sym_rewinddir] = ACTIONS(3053),
    [anon_sym_readlink] = ACTIONS(3053),
    [anon_sym_ref] = ACTIONS(3053),
    [anon_sym_scalar] = ACTIONS(3053),
    [anon_sym_shift] = ACTIONS(3053),
    [anon_sym_sin] = ACTIONS(3053),
    [anon_sym_sleep] = ACTIONS(3053),
    [anon_sym_sqrt] = ACTIONS(3053),
    [anon_sym_srand] = ACTIONS(3053),
    [anon_sym_stat] = ACTIONS(3053),
    [anon_sym_study] = ACTIONS(3053),
    [anon_sym_tell] = ACTIONS(3053),
    [anon_sym_telldir] = ACTIONS(3053),
    [anon_sym_tied] = ACTIONS(3053),
    [anon_sym_uc] = ACTIONS(3053),
    [anon_sym_ucfirst] = ACTIONS(3053),
    [anon_sym_untie] = ACTIONS(3053),
    [anon_sym_umask] = ACTIONS(3053),
    [anon_sym_values] = ACTIONS(3053),
    [anon_sym_write] = ACTIONS(3053),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3053),
    [anon_sym___END__] = ACTIONS(3053),
    [anon_sym_EOT] = ACTIONS(3053),
    [anon_sym_q] = ACTIONS(3053),
    [anon_sym_qq] = ACTIONS(3053),
    [anon_sym_qw] = ACTIONS(3053),
    [anon_sym_qx] = ACTIONS(3053),
    [anon_sym_qr] = ACTIONS(3053),
    [anon_sym_m] = ACTIONS(3053),
    [anon_sym_s] = ACTIONS(3053),
    [anon_sym_tr] = ACTIONS(3053),
    [anon_sym_y] = ACTIONS(3053),
    [anon_sym_if] = ACTIONS(3053),
    [anon_sym_unless] = ACTIONS(3053),
    [anon_sym_while] = ACTIONS(3053),
    [anon_sym_until] = ACTIONS(3053),
    [aux_sym__bareword_token1] = ACTIONS(3053),
    [sym_number] = ACTIONS(3053),
    [anon_sym_true] = ACTIONS(3053),
    [anon_sym_false] = ACTIONS(3053),
    [sym__single_quote] = ACTIONS(3051),
    [sym__double_quote] = ACTIONS(3051),
    [sym__backtick_quote] = ACTIONS(3051),
    [sym__search_slash_quote] = ACTIONS(3051),
    [sym__open_readline_bracket] = ACTIONS(3051),
    [sym__open_fileglob_bracket] = ACTIONS(3051),
    [sym__PERLY_HEREDOC] = ACTIONS(3051),
    [sym__ctrl_z_hack] = ACTIONS(3051),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3051),
    [sym__filetest] = ACTIONS(3051),
  },
  [STATE(1475)] = {
    [sym_heredoc_content] = STATE(1475),
    [ts_builtin_sym_end] = ACTIONS(3055),
    [sym__identifier] = ACTIONS(3057),
    [anon_sym_LBRACE] = ACTIONS(3057),
    [anon_sym_SEMI] = ACTIONS(3057),
    [anon_sym_package] = ACTIONS(3057),
    [anon_sym_class] = ACTIONS(3057),
    [anon_sym_role] = ACTIONS(3057),
    [anon_sym_BUILD] = ACTIONS(3057),
    [anon_sym_ADJUST] = ACTIONS(3057),
    [anon_sym_DOLLAR] = ACTIONS(3057),
    [anon_sym_AT] = ACTIONS(3057),
    [anon_sym_my] = ACTIONS(3057),
    [anon_sym_extended] = ACTIONS(3057),
    [anon_sym_async] = ACTIONS(3057),
    [anon_sym_sub] = ACTIONS(3057),
    [anon_sym_method] = ACTIONS(3057),
    [anon_sym_LPAREN] = ACTIONS(3057),
    [anon_sym_continue] = ACTIONS(3057),
    [anon_sym_state] = ACTIONS(3057),
    [anon_sym_our] = ACTIONS(3057),
    [anon_sym_try] = ACTIONS(3057),
    [anon_sym_defer] = ACTIONS(3057),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3057),
    [anon_sym_LBRACK] = ACTIONS(3057),
    [anon_sym_LT_LT] = ACTIONS(3057),
    [anon_sym_SLASH_SLASH] = ACTIONS(3057),
    [anon_sym_PLUS] = ACTIONS(3057),
    [anon_sym_DASH] = ACTIONS(3057),
    [anon_sym_TILDE] = ACTIONS(3057),
    [anon_sym_BANG] = ACTIONS(3057),
    [anon_sym_PLUS_PLUS] = ACTIONS(3057),
    [anon_sym_DASH_DASH] = ACTIONS(3057),
    [anon_sym_BSLASH] = ACTIONS(3055),
    [anon_sym_do] = ACTIONS(3057),
    [anon_sym_eval] = ACTIONS(3057),
    [anon_sym_await] = ACTIONS(3057),
    [anon_sym_field] = ACTIONS(3057),
    [anon_sym_local] = ACTIONS(3057),
    [anon_sym_dynamically] = ACTIONS(3057),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3057),
    [anon_sym_require] = ACTIONS(3057),
    [anon_sym_map] = ACTIONS(3057),
    [anon_sym_grep] = ACTIONS(3057),
    [anon_sym_sort] = ACTIONS(3057),
    [anon_sym_goto] = ACTIONS(3057),
    [anon_sym_return] = ACTIONS(3057),
    [anon_sym_undef] = ACTIONS(3057),
    [anon_sym_PERCENT2] = ACTIONS(3057),
    [anon_sym_AMP2] = ACTIONS(3057),
    [anon_sym_STAR2] = ACTIONS(3057),
    [anon_sym_use] = ACTIONS(3057),
    [anon_sym_no] = ACTIONS(3057),
    [anon_sym_for] = ACTIONS(3057),
    [anon_sym_foreach] = ACTIONS(3057),
    [anon_sym_last] = ACTIONS(3057),
    [anon_sym_next] = ACTIONS(3057),
    [anon_sym_redo] = ACTIONS(3057),
    [anon_sym_BEGIN] = ACTIONS(3057),
    [anon_sym_INIT] = ACTIONS(3057),
    [anon_sym_CHECK] = ACTIONS(3057),
    [anon_sym_UNITCHECK] = ACTIONS(3057),
    [anon_sym_END] = ACTIONS(3057),
    [anon_sym___FILE__] = ACTIONS(3057),
    [anon_sym___LINE__] = ACTIONS(3057),
    [anon_sym___PACKAGE__] = ACTIONS(3057),
    [anon_sym___SUB__] = ACTIONS(3057),
    [anon_sym_break] = ACTIONS(3057),
    [anon_sym_fork] = ACTIONS(3057),
    [anon_sym_getppid] = ACTIONS(3057),
    [anon_sym_time] = ACTIONS(3057),
    [anon_sym_times] = ACTIONS(3057),
    [anon_sym_wait] = ACTIONS(3057),
    [anon_sym_wantarray] = ACTIONS(3057),
    [anon_sym_abs] = ACTIONS(3057),
    [anon_sym_alarm] = ACTIONS(3057),
    [anon_sym_chop] = ACTIONS(3057),
    [anon_sym_chdir] = ACTIONS(3057),
    [anon_sym_close] = ACTIONS(3057),
    [anon_sym_closedir] = ACTIONS(3057),
    [anon_sym_caller] = ACTIONS(3057),
    [anon_sym_chomp] = ACTIONS(3057),
    [anon_sym_chr] = ACTIONS(3057),
    [anon_sym_cos] = ACTIONS(3057),
    [anon_sym_chroot] = ACTIONS(3057),
    [anon_sym_defined] = ACTIONS(3057),
    [anon_sym_delete] = ACTIONS(3057),
    [anon_sym_dbmclose] = ACTIONS(3057),
    [anon_sym_exists] = ACTIONS(3057),
    [anon_sym_exit] = ACTIONS(3057),
    [anon_sym_eof] = ACTIONS(3057),
    [anon_sym_exp] = ACTIONS(3057),
    [anon_sym_each] = ACTIONS(3057),
    [anon_sym_fc] = ACTIONS(3057),
    [anon_sym_fileno] = ACTIONS(3057),
    [anon_sym_gmtime] = ACTIONS(3057),
    [anon_sym_getc] = ACTIONS(3057),
    [anon_sym_getpgrp] = ACTIONS(3057),
    [anon_sym_getprotobyname] = ACTIONS(3057),
    [anon_sym_getpwname] = ACTIONS(3057),
    [anon_sym_getpwuid] = ACTIONS(3057),
    [anon_sym_getpeername] = ACTIONS(3057),
    [anon_sym_getnetbyname] = ACTIONS(3057),
    [anon_sym_getsockname] = ACTIONS(3057),
    [anon_sym_getgrnam] = ACTIONS(3057),
    [anon_sym_getgrgid] = ACTIONS(3057),
    [anon_sym_hex] = ACTIONS(3057),
    [anon_sym_int] = ACTIONS(3057),
    [anon_sym_keys] = ACTIONS(3057),
    [anon_sym_lc] = ACTIONS(3057),
    [anon_sym_lcfirst] = ACTIONS(3057),
    [anon_sym_length] = ACTIONS(3057),
    [anon_sym_localtime] = ACTIONS(3057),
    [anon_sym_log] = ACTIONS(3057),
    [anon_sym_lock] = ACTIONS(3057),
    [anon_sym_lstat] = ACTIONS(3057),
    [anon_sym_oct] = ACTIONS(3057),
    [anon_sym_ord] = ACTIONS(3057),
    [anon_sym_prototype] = ACTIONS(3057),
    [anon_sym_pop] = ACTIONS(3057),
    [anon_sym_pos] = ACTIONS(3057),
    [anon_sym_quotemeta] = ACTIONS(3057),
    [anon_sym_reset] = ACTIONS(3057),
    [anon_sym_rand] = ACTIONS(3057),
    [anon_sym_rmdir] = ACTIONS(3057),
    [anon_sym_readdir] = ACTIONS(3057),
    [anon_sym_readline] = ACTIONS(3057),
    [anon_sym_readpipe] = ACTIONS(3057),
    [anon_sym_rewinddir] = ACTIONS(3057),
    [anon_sym_readlink] = ACTIONS(3057),
    [anon_sym_ref] = ACTIONS(3057),
    [anon_sym_scalar] = ACTIONS(3057),
    [anon_sym_shift] = ACTIONS(3057),
    [anon_sym_sin] = ACTIONS(3057),
    [anon_sym_sleep] = ACTIONS(3057),
    [anon_sym_sqrt] = ACTIONS(3057),
    [anon_sym_srand] = ACTIONS(3057),
    [anon_sym_stat] = ACTIONS(3057),
    [anon_sym_study] = ACTIONS(3057),
    [anon_sym_tell] = ACTIONS(3057),
    [anon_sym_telldir] = ACTIONS(3057),
    [anon_sym_tied] = ACTIONS(3057),
    [anon_sym_uc] = ACTIONS(3057),
    [anon_sym_ucfirst] = ACTIONS(3057),
    [anon_sym_untie] = ACTIONS(3057),
    [anon_sym_umask] = ACTIONS(3057),
    [anon_sym_values] = ACTIONS(3057),
    [anon_sym_write] = ACTIONS(3057),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3057),
    [anon_sym___END__] = ACTIONS(3057),
    [anon_sym_EOT] = ACTIONS(3057),
    [anon_sym_q] = ACTIONS(3057),
    [anon_sym_qq] = ACTIONS(3057),
    [anon_sym_qw] = ACTIONS(3057),
    [anon_sym_qx] = ACTIONS(3057),
    [anon_sym_qr] = ACTIONS(3057),
    [anon_sym_m] = ACTIONS(3057),
    [anon_sym_s] = ACTIONS(3057),
    [anon_sym_tr] = ACTIONS(3057),
    [anon_sym_y] = ACTIONS(3057),
    [anon_sym_if] = ACTIONS(3057),
    [anon_sym_unless] = ACTIONS(3057),
    [anon_sym_while] = ACTIONS(3057),
    [anon_sym_until] = ACTIONS(3057),
    [aux_sym__bareword_token1] = ACTIONS(3057),
    [sym_number] = ACTIONS(3057),
    [anon_sym_true] = ACTIONS(3057),
    [anon_sym_false] = ACTIONS(3057),
    [sym__single_quote] = ACTIONS(3055),
    [sym__double_quote] = ACTIONS(3055),
    [sym__backtick_quote] = ACTIONS(3055),
    [sym__search_slash_quote] = ACTIONS(3055),
    [sym__open_readline_bracket] = ACTIONS(3055),
    [sym__open_fileglob_bracket] = ACTIONS(3055),
    [sym__PERLY_HEREDOC] = ACTIONS(3055),
    [sym__ctrl_z_hack] = ACTIONS(3055),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3055),
    [sym__filetest] = ACTIONS(3055),
  },
  [STATE(1476)] = {
    [sym_heredoc_content] = STATE(1476),
    [ts_builtin_sym_end] = ACTIONS(3059),
    [sym__identifier] = ACTIONS(3061),
    [anon_sym_LBRACE] = ACTIONS(3061),
    [anon_sym_SEMI] = ACTIONS(3061),
    [anon_sym_package] = ACTIONS(3061),
    [anon_sym_class] = ACTIONS(3061),
    [anon_sym_role] = ACTIONS(3061),
    [anon_sym_BUILD] = ACTIONS(3061),
    [anon_sym_ADJUST] = ACTIONS(3061),
    [anon_sym_DOLLAR] = ACTIONS(3061),
    [anon_sym_AT] = ACTIONS(3061),
    [anon_sym_my] = ACTIONS(3061),
    [anon_sym_extended] = ACTIONS(3061),
    [anon_sym_async] = ACTIONS(3061),
    [anon_sym_sub] = ACTIONS(3061),
    [anon_sym_method] = ACTIONS(3061),
    [anon_sym_LPAREN] = ACTIONS(3061),
    [anon_sym_continue] = ACTIONS(3061),
    [anon_sym_state] = ACTIONS(3061),
    [anon_sym_our] = ACTIONS(3061),
    [anon_sym_try] = ACTIONS(3061),
    [anon_sym_defer] = ACTIONS(3061),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3061),
    [anon_sym_LBRACK] = ACTIONS(3061),
    [anon_sym_LT_LT] = ACTIONS(3061),
    [anon_sym_SLASH_SLASH] = ACTIONS(3061),
    [anon_sym_PLUS] = ACTIONS(3061),
    [anon_sym_DASH] = ACTIONS(3061),
    [anon_sym_TILDE] = ACTIONS(3061),
    [anon_sym_BANG] = ACTIONS(3061),
    [anon_sym_PLUS_PLUS] = ACTIONS(3061),
    [anon_sym_DASH_DASH] = ACTIONS(3061),
    [anon_sym_BSLASH] = ACTIONS(3059),
    [anon_sym_do] = ACTIONS(3061),
    [anon_sym_eval] = ACTIONS(3061),
    [anon_sym_await] = ACTIONS(3061),
    [anon_sym_field] = ACTIONS(3061),
    [anon_sym_local] = ACTIONS(3061),
    [anon_sym_dynamically] = ACTIONS(3061),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3061),
    [anon_sym_require] = ACTIONS(3061),
    [anon_sym_map] = ACTIONS(3061),
    [anon_sym_grep] = ACTIONS(3061),
    [anon_sym_sort] = ACTIONS(3061),
    [anon_sym_goto] = ACTIONS(3061),
    [anon_sym_return] = ACTIONS(3061),
    [anon_sym_undef] = ACTIONS(3061),
    [anon_sym_PERCENT2] = ACTIONS(3061),
    [anon_sym_AMP2] = ACTIONS(3061),
    [anon_sym_STAR2] = ACTIONS(3061),
    [anon_sym_use] = ACTIONS(3061),
    [anon_sym_no] = ACTIONS(3061),
    [anon_sym_for] = ACTIONS(3061),
    [anon_sym_foreach] = ACTIONS(3061),
    [anon_sym_last] = ACTIONS(3061),
    [anon_sym_next] = ACTIONS(3061),
    [anon_sym_redo] = ACTIONS(3061),
    [anon_sym_BEGIN] = ACTIONS(3061),
    [anon_sym_INIT] = ACTIONS(3061),
    [anon_sym_CHECK] = ACTIONS(3061),
    [anon_sym_UNITCHECK] = ACTIONS(3061),
    [anon_sym_END] = ACTIONS(3061),
    [anon_sym___FILE__] = ACTIONS(3061),
    [anon_sym___LINE__] = ACTIONS(3061),
    [anon_sym___PACKAGE__] = ACTIONS(3061),
    [anon_sym___SUB__] = ACTIONS(3061),
    [anon_sym_break] = ACTIONS(3061),
    [anon_sym_fork] = ACTIONS(3061),
    [anon_sym_getppid] = ACTIONS(3061),
    [anon_sym_time] = ACTIONS(3061),
    [anon_sym_times] = ACTIONS(3061),
    [anon_sym_wait] = ACTIONS(3061),
    [anon_sym_wantarray] = ACTIONS(3061),
    [anon_sym_abs] = ACTIONS(3061),
    [anon_sym_alarm] = ACTIONS(3061),
    [anon_sym_chop] = ACTIONS(3061),
    [anon_sym_chdir] = ACTIONS(3061),
    [anon_sym_close] = ACTIONS(3061),
    [anon_sym_closedir] = ACTIONS(3061),
    [anon_sym_caller] = ACTIONS(3061),
    [anon_sym_chomp] = ACTIONS(3061),
    [anon_sym_chr] = ACTIONS(3061),
    [anon_sym_cos] = ACTIONS(3061),
    [anon_sym_chroot] = ACTIONS(3061),
    [anon_sym_defined] = ACTIONS(3061),
    [anon_sym_delete] = ACTIONS(3061),
    [anon_sym_dbmclose] = ACTIONS(3061),
    [anon_sym_exists] = ACTIONS(3061),
    [anon_sym_exit] = ACTIONS(3061),
    [anon_sym_eof] = ACTIONS(3061),
    [anon_sym_exp] = ACTIONS(3061),
    [anon_sym_each] = ACTIONS(3061),
    [anon_sym_fc] = ACTIONS(3061),
    [anon_sym_fileno] = ACTIONS(3061),
    [anon_sym_gmtime] = ACTIONS(3061),
    [anon_sym_getc] = ACTIONS(3061),
    [anon_sym_getpgrp] = ACTIONS(3061),
    [anon_sym_getprotobyname] = ACTIONS(3061),
    [anon_sym_getpwname] = ACTIONS(3061),
    [anon_sym_getpwuid] = ACTIONS(3061),
    [anon_sym_getpeername] = ACTIONS(3061),
    [anon_sym_getnetbyname] = ACTIONS(3061),
    [anon_sym_getsockname] = ACTIONS(3061),
    [anon_sym_getgrnam] = ACTIONS(3061),
    [anon_sym_getgrgid] = ACTIONS(3061),
    [anon_sym_hex] = ACTIONS(3061),
    [anon_sym_int] = ACTIONS(3061),
    [anon_sym_keys] = ACTIONS(3061),
    [anon_sym_lc] = ACTIONS(3061),
    [anon_sym_lcfirst] = ACTIONS(3061),
    [anon_sym_length] = ACTIONS(3061),
    [anon_sym_localtime] = ACTIONS(3061),
    [anon_sym_log] = ACTIONS(3061),
    [anon_sym_lock] = ACTIONS(3061),
    [anon_sym_lstat] = ACTIONS(3061),
    [anon_sym_oct] = ACTIONS(3061),
    [anon_sym_ord] = ACTIONS(3061),
    [anon_sym_prototype] = ACTIONS(3061),
    [anon_sym_pop] = ACTIONS(3061),
    [anon_sym_pos] = ACTIONS(3061),
    [anon_sym_quotemeta] = ACTIONS(3061),
    [anon_sym_reset] = ACTIONS(3061),
    [anon_sym_rand] = ACTIONS(3061),
    [anon_sym_rmdir] = ACTIONS(3061),
    [anon_sym_readdir] = ACTIONS(3061),
    [anon_sym_readline] = ACTIONS(3061),
    [anon_sym_readpipe] = ACTIONS(3061),
    [anon_sym_rewinddir] = ACTIONS(3061),
    [anon_sym_readlink] = ACTIONS(3061),
    [anon_sym_ref] = ACTIONS(3061),
    [anon_sym_scalar] = ACTIONS(3061),
    [anon_sym_shift] = ACTIONS(3061),
    [anon_sym_sin] = ACTIONS(3061),
    [anon_sym_sleep] = ACTIONS(3061),
    [anon_sym_sqrt] = ACTIONS(3061),
    [anon_sym_srand] = ACTIONS(3061),
    [anon_sym_stat] = ACTIONS(3061),
    [anon_sym_study] = ACTIONS(3061),
    [anon_sym_tell] = ACTIONS(3061),
    [anon_sym_telldir] = ACTIONS(3061),
    [anon_sym_tied] = ACTIONS(3061),
    [anon_sym_uc] = ACTIONS(3061),
    [anon_sym_ucfirst] = ACTIONS(3061),
    [anon_sym_untie] = ACTIONS(3061),
    [anon_sym_umask] = ACTIONS(3061),
    [anon_sym_values] = ACTIONS(3061),
    [anon_sym_write] = ACTIONS(3061),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3061),
    [anon_sym___END__] = ACTIONS(3061),
    [anon_sym_EOT] = ACTIONS(3061),
    [anon_sym_q] = ACTIONS(3061),
    [anon_sym_qq] = ACTIONS(3061),
    [anon_sym_qw] = ACTIONS(3061),
    [anon_sym_qx] = ACTIONS(3061),
    [anon_sym_qr] = ACTIONS(3061),
    [anon_sym_m] = ACTIONS(3061),
    [anon_sym_s] = ACTIONS(3061),
    [anon_sym_tr] = ACTIONS(3061),
    [anon_sym_y] = ACTIONS(3061),
    [anon_sym_if] = ACTIONS(3061),
    [anon_sym_unless] = ACTIONS(3061),
    [anon_sym_while] = ACTIONS(3061),
    [anon_sym_until] = ACTIONS(3061),
    [aux_sym__bareword_token1] = ACTIONS(3061),
    [sym_number] = ACTIONS(3061),
    [anon_sym_true] = ACTIONS(3061),
    [anon_sym_false] = ACTIONS(3061),
    [sym__single_quote] = ACTIONS(3059),
    [sym__double_quote] = ACTIONS(3059),
    [sym__backtick_quote] = ACTIONS(3059),
    [sym__search_slash_quote] = ACTIONS(3059),
    [sym__open_readline_bracket] = ACTIONS(3059),
    [sym__open_fileglob_bracket] = ACTIONS(3059),
    [sym__PERLY_HEREDOC] = ACTIONS(3059),
    [sym__ctrl_z_hack] = ACTIONS(3059),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3059),
    [sym__filetest] = ACTIONS(3059),
  },
  [STATE(1477)] = {
    [sym_heredoc_content] = STATE(1477),
    [ts_builtin_sym_end] = ACTIONS(3063),
    [sym__identifier] = ACTIONS(3065),
    [anon_sym_LBRACE] = ACTIONS(3065),
    [anon_sym_SEMI] = ACTIONS(3065),
    [anon_sym_package] = ACTIONS(3065),
    [anon_sym_class] = ACTIONS(3065),
    [anon_sym_role] = ACTIONS(3065),
    [anon_sym_BUILD] = ACTIONS(3065),
    [anon_sym_ADJUST] = ACTIONS(3065),
    [anon_sym_DOLLAR] = ACTIONS(3065),
    [anon_sym_AT] = ACTIONS(3065),
    [anon_sym_my] = ACTIONS(3065),
    [anon_sym_extended] = ACTIONS(3065),
    [anon_sym_async] = ACTIONS(3065),
    [anon_sym_sub] = ACTIONS(3065),
    [anon_sym_method] = ACTIONS(3065),
    [anon_sym_LPAREN] = ACTIONS(3065),
    [anon_sym_continue] = ACTIONS(3065),
    [anon_sym_state] = ACTIONS(3065),
    [anon_sym_our] = ACTIONS(3065),
    [anon_sym_try] = ACTIONS(3065),
    [anon_sym_defer] = ACTIONS(3065),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3065),
    [anon_sym_LBRACK] = ACTIONS(3065),
    [anon_sym_LT_LT] = ACTIONS(3065),
    [anon_sym_SLASH_SLASH] = ACTIONS(3065),
    [anon_sym_PLUS] = ACTIONS(3065),
    [anon_sym_DASH] = ACTIONS(3065),
    [anon_sym_TILDE] = ACTIONS(3065),
    [anon_sym_BANG] = ACTIONS(3065),
    [anon_sym_PLUS_PLUS] = ACTIONS(3065),
    [anon_sym_DASH_DASH] = ACTIONS(3065),
    [anon_sym_BSLASH] = ACTIONS(3063),
    [anon_sym_do] = ACTIONS(3065),
    [anon_sym_eval] = ACTIONS(3065),
    [anon_sym_await] = ACTIONS(3065),
    [anon_sym_field] = ACTIONS(3065),
    [anon_sym_local] = ACTIONS(3065),
    [anon_sym_dynamically] = ACTIONS(3065),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3065),
    [anon_sym_require] = ACTIONS(3065),
    [anon_sym_map] = ACTIONS(3065),
    [anon_sym_grep] = ACTIONS(3065),
    [anon_sym_sort] = ACTIONS(3065),
    [anon_sym_goto] = ACTIONS(3065),
    [anon_sym_return] = ACTIONS(3065),
    [anon_sym_undef] = ACTIONS(3065),
    [anon_sym_PERCENT2] = ACTIONS(3065),
    [anon_sym_AMP2] = ACTIONS(3065),
    [anon_sym_STAR2] = ACTIONS(3065),
    [anon_sym_use] = ACTIONS(3065),
    [anon_sym_no] = ACTIONS(3065),
    [anon_sym_for] = ACTIONS(3065),
    [anon_sym_foreach] = ACTIONS(3065),
    [anon_sym_last] = ACTIONS(3065),
    [anon_sym_next] = ACTIONS(3065),
    [anon_sym_redo] = ACTIONS(3065),
    [anon_sym_BEGIN] = ACTIONS(3065),
    [anon_sym_INIT] = ACTIONS(3065),
    [anon_sym_CHECK] = ACTIONS(3065),
    [anon_sym_UNITCHECK] = ACTIONS(3065),
    [anon_sym_END] = ACTIONS(3065),
    [anon_sym___FILE__] = ACTIONS(3065),
    [anon_sym___LINE__] = ACTIONS(3065),
    [anon_sym___PACKAGE__] = ACTIONS(3065),
    [anon_sym___SUB__] = ACTIONS(3065),
    [anon_sym_break] = ACTIONS(3065),
    [anon_sym_fork] = ACTIONS(3065),
    [anon_sym_getppid] = ACTIONS(3065),
    [anon_sym_time] = ACTIONS(3065),
    [anon_sym_times] = ACTIONS(3065),
    [anon_sym_wait] = ACTIONS(3065),
    [anon_sym_wantarray] = ACTIONS(3065),
    [anon_sym_abs] = ACTIONS(3065),
    [anon_sym_alarm] = ACTIONS(3065),
    [anon_sym_chop] = ACTIONS(3065),
    [anon_sym_chdir] = ACTIONS(3065),
    [anon_sym_close] = ACTIONS(3065),
    [anon_sym_closedir] = ACTIONS(3065),
    [anon_sym_caller] = ACTIONS(3065),
    [anon_sym_chomp] = ACTIONS(3065),
    [anon_sym_chr] = ACTIONS(3065),
    [anon_sym_cos] = ACTIONS(3065),
    [anon_sym_chroot] = ACTIONS(3065),
    [anon_sym_defined] = ACTIONS(3065),
    [anon_sym_delete] = ACTIONS(3065),
    [anon_sym_dbmclose] = ACTIONS(3065),
    [anon_sym_exists] = ACTIONS(3065),
    [anon_sym_exit] = ACTIONS(3065),
    [anon_sym_eof] = ACTIONS(3065),
    [anon_sym_exp] = ACTIONS(3065),
    [anon_sym_each] = ACTIONS(3065),
    [anon_sym_fc] = ACTIONS(3065),
    [anon_sym_fileno] = ACTIONS(3065),
    [anon_sym_gmtime] = ACTIONS(3065),
    [anon_sym_getc] = ACTIONS(3065),
    [anon_sym_getpgrp] = ACTIONS(3065),
    [anon_sym_getprotobyname] = ACTIONS(3065),
    [anon_sym_getpwname] = ACTIONS(3065),
    [anon_sym_getpwuid] = ACTIONS(3065),
    [anon_sym_getpeername] = ACTIONS(3065),
    [anon_sym_getnetbyname] = ACTIONS(3065),
    [anon_sym_getsockname] = ACTIONS(3065),
    [anon_sym_getgrnam] = ACTIONS(3065),
    [anon_sym_getgrgid] = ACTIONS(3065),
    [anon_sym_hex] = ACTIONS(3065),
    [anon_sym_int] = ACTIONS(3065),
    [anon_sym_keys] = ACTIONS(3065),
    [anon_sym_lc] = ACTIONS(3065),
    [anon_sym_lcfirst] = ACTIONS(3065),
    [anon_sym_length] = ACTIONS(3065),
    [anon_sym_localtime] = ACTIONS(3065),
    [anon_sym_log] = ACTIONS(3065),
    [anon_sym_lock] = ACTIONS(3065),
    [anon_sym_lstat] = ACTIONS(3065),
    [anon_sym_oct] = ACTIONS(3065),
    [anon_sym_ord] = ACTIONS(3065),
    [anon_sym_prototype] = ACTIONS(3065),
    [anon_sym_pop] = ACTIONS(3065),
    [anon_sym_pos] = ACTIONS(3065),
    [anon_sym_quotemeta] = ACTIONS(3065),
    [anon_sym_reset] = ACTIONS(3065),
    [anon_sym_rand] = ACTIONS(3065),
    [anon_sym_rmdir] = ACTIONS(3065),
    [anon_sym_readdir] = ACTIONS(3065),
    [anon_sym_readline] = ACTIONS(3065),
    [anon_sym_readpipe] = ACTIONS(3065),
    [anon_sym_rewinddir] = ACTIONS(3065),
    [anon_sym_readlink] = ACTIONS(3065),
    [anon_sym_ref] = ACTIONS(3065),
    [anon_sym_scalar] = ACTIONS(3065),
    [anon_sym_shift] = ACTIONS(3065),
    [anon_sym_sin] = ACTIONS(3065),
    [anon_sym_sleep] = ACTIONS(3065),
    [anon_sym_sqrt] = ACTIONS(3065),
    [anon_sym_srand] = ACTIONS(3065),
    [anon_sym_stat] = ACTIONS(3065),
    [anon_sym_study] = ACTIONS(3065),
    [anon_sym_tell] = ACTIONS(3065),
    [anon_sym_telldir] = ACTIONS(3065),
    [anon_sym_tied] = ACTIONS(3065),
    [anon_sym_uc] = ACTIONS(3065),
    [anon_sym_ucfirst] = ACTIONS(3065),
    [anon_sym_untie] = ACTIONS(3065),
    [anon_sym_umask] = ACTIONS(3065),
    [anon_sym_values] = ACTIONS(3065),
    [anon_sym_write] = ACTIONS(3065),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3065),
    [anon_sym___END__] = ACTIONS(3065),
    [anon_sym_EOT] = ACTIONS(3065),
    [anon_sym_q] = ACTIONS(3065),
    [anon_sym_qq] = ACTIONS(3065),
    [anon_sym_qw] = ACTIONS(3065),
    [anon_sym_qx] = ACTIONS(3065),
    [anon_sym_qr] = ACTIONS(3065),
    [anon_sym_m] = ACTIONS(3065),
    [anon_sym_s] = ACTIONS(3065),
    [anon_sym_tr] = ACTIONS(3065),
    [anon_sym_y] = ACTIONS(3065),
    [anon_sym_if] = ACTIONS(3065),
    [anon_sym_unless] = ACTIONS(3065),
    [anon_sym_while] = ACTIONS(3065),
    [anon_sym_until] = ACTIONS(3065),
    [aux_sym__bareword_token1] = ACTIONS(3065),
    [sym_number] = ACTIONS(3065),
    [anon_sym_true] = ACTIONS(3065),
    [anon_sym_false] = ACTIONS(3065),
    [sym__single_quote] = ACTIONS(3063),
    [sym__double_quote] = ACTIONS(3063),
    [sym__backtick_quote] = ACTIONS(3063),
    [sym__search_slash_quote] = ACTIONS(3063),
    [sym__open_readline_bracket] = ACTIONS(3063),
    [sym__open_fileglob_bracket] = ACTIONS(3063),
    [sym__PERLY_HEREDOC] = ACTIONS(3063),
    [sym__ctrl_z_hack] = ACTIONS(3063),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3063),
    [sym__filetest] = ACTIONS(3063),
  },
  [STATE(1478)] = {
    [sym_heredoc_content] = STATE(1478),
    [ts_builtin_sym_end] = ACTIONS(3067),
    [sym__identifier] = ACTIONS(3069),
    [anon_sym_LBRACE] = ACTIONS(3069),
    [anon_sym_SEMI] = ACTIONS(3069),
    [anon_sym_package] = ACTIONS(3069),
    [anon_sym_class] = ACTIONS(3069),
    [anon_sym_role] = ACTIONS(3069),
    [anon_sym_BUILD] = ACTIONS(3069),
    [anon_sym_ADJUST] = ACTIONS(3069),
    [anon_sym_DOLLAR] = ACTIONS(3069),
    [anon_sym_AT] = ACTIONS(3069),
    [anon_sym_my] = ACTIONS(3069),
    [anon_sym_extended] = ACTIONS(3069),
    [anon_sym_async] = ACTIONS(3069),
    [anon_sym_sub] = ACTIONS(3069),
    [anon_sym_method] = ACTIONS(3069),
    [anon_sym_LPAREN] = ACTIONS(3069),
    [anon_sym_continue] = ACTIONS(3069),
    [anon_sym_state] = ACTIONS(3069),
    [anon_sym_our] = ACTIONS(3069),
    [anon_sym_try] = ACTIONS(3069),
    [anon_sym_defer] = ACTIONS(3069),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3069),
    [anon_sym_LBRACK] = ACTIONS(3069),
    [anon_sym_LT_LT] = ACTIONS(3069),
    [anon_sym_SLASH_SLASH] = ACTIONS(3069),
    [anon_sym_PLUS] = ACTIONS(3069),
    [anon_sym_DASH] = ACTIONS(3069),
    [anon_sym_TILDE] = ACTIONS(3069),
    [anon_sym_BANG] = ACTIONS(3069),
    [anon_sym_PLUS_PLUS] = ACTIONS(3069),
    [anon_sym_DASH_DASH] = ACTIONS(3069),
    [anon_sym_BSLASH] = ACTIONS(3067),
    [anon_sym_do] = ACTIONS(3069),
    [anon_sym_eval] = ACTIONS(3069),
    [anon_sym_await] = ACTIONS(3069),
    [anon_sym_field] = ACTIONS(3069),
    [anon_sym_local] = ACTIONS(3069),
    [anon_sym_dynamically] = ACTIONS(3069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3069),
    [anon_sym_require] = ACTIONS(3069),
    [anon_sym_map] = ACTIONS(3069),
    [anon_sym_grep] = ACTIONS(3069),
    [anon_sym_sort] = ACTIONS(3069),
    [anon_sym_goto] = ACTIONS(3069),
    [anon_sym_return] = ACTIONS(3069),
    [anon_sym_undef] = ACTIONS(3069),
    [anon_sym_PERCENT2] = ACTIONS(3069),
    [anon_sym_AMP2] = ACTIONS(3069),
    [anon_sym_STAR2] = ACTIONS(3069),
    [anon_sym_use] = ACTIONS(3069),
    [anon_sym_no] = ACTIONS(3069),
    [anon_sym_for] = ACTIONS(3069),
    [anon_sym_foreach] = ACTIONS(3069),
    [anon_sym_last] = ACTIONS(3069),
    [anon_sym_next] = ACTIONS(3069),
    [anon_sym_redo] = ACTIONS(3069),
    [anon_sym_BEGIN] = ACTIONS(3069),
    [anon_sym_INIT] = ACTIONS(3069),
    [anon_sym_CHECK] = ACTIONS(3069),
    [anon_sym_UNITCHECK] = ACTIONS(3069),
    [anon_sym_END] = ACTIONS(3069),
    [anon_sym___FILE__] = ACTIONS(3069),
    [anon_sym___LINE__] = ACTIONS(3069),
    [anon_sym___PACKAGE__] = ACTIONS(3069),
    [anon_sym___SUB__] = ACTIONS(3069),
    [anon_sym_break] = ACTIONS(3069),
    [anon_sym_fork] = ACTIONS(3069),
    [anon_sym_getppid] = ACTIONS(3069),
    [anon_sym_time] = ACTIONS(3069),
    [anon_sym_times] = ACTIONS(3069),
    [anon_sym_wait] = ACTIONS(3069),
    [anon_sym_wantarray] = ACTIONS(3069),
    [anon_sym_abs] = ACTIONS(3069),
    [anon_sym_alarm] = ACTIONS(3069),
    [anon_sym_chop] = ACTIONS(3069),
    [anon_sym_chdir] = ACTIONS(3069),
    [anon_sym_close] = ACTIONS(3069),
    [anon_sym_closedir] = ACTIONS(3069),
    [anon_sym_caller] = ACTIONS(3069),
    [anon_sym_chomp] = ACTIONS(3069),
    [anon_sym_chr] = ACTIONS(3069),
    [anon_sym_cos] = ACTIONS(3069),
    [anon_sym_chroot] = ACTIONS(3069),
    [anon_sym_defined] = ACTIONS(3069),
    [anon_sym_delete] = ACTIONS(3069),
    [anon_sym_dbmclose] = ACTIONS(3069),
    [anon_sym_exists] = ACTIONS(3069),
    [anon_sym_exit] = ACTIONS(3069),
    [anon_sym_eof] = ACTIONS(3069),
    [anon_sym_exp] = ACTIONS(3069),
    [anon_sym_each] = ACTIONS(3069),
    [anon_sym_fc] = ACTIONS(3069),
    [anon_sym_fileno] = ACTIONS(3069),
    [anon_sym_gmtime] = ACTIONS(3069),
    [anon_sym_getc] = ACTIONS(3069),
    [anon_sym_getpgrp] = ACTIONS(3069),
    [anon_sym_getprotobyname] = ACTIONS(3069),
    [anon_sym_getpwname] = ACTIONS(3069),
    [anon_sym_getpwuid] = ACTIONS(3069),
    [anon_sym_getpeername] = ACTIONS(3069),
    [anon_sym_getnetbyname] = ACTIONS(3069),
    [anon_sym_getsockname] = ACTIONS(3069),
    [anon_sym_getgrnam] = ACTIONS(3069),
    [anon_sym_getgrgid] = ACTIONS(3069),
    [anon_sym_hex] = ACTIONS(3069),
    [anon_sym_int] = ACTIONS(3069),
    [anon_sym_keys] = ACTIONS(3069),
    [anon_sym_lc] = ACTIONS(3069),
    [anon_sym_lcfirst] = ACTIONS(3069),
    [anon_sym_length] = ACTIONS(3069),
    [anon_sym_localtime] = ACTIONS(3069),
    [anon_sym_log] = ACTIONS(3069),
    [anon_sym_lock] = ACTIONS(3069),
    [anon_sym_lstat] = ACTIONS(3069),
    [anon_sym_oct] = ACTIONS(3069),
    [anon_sym_ord] = ACTIONS(3069),
    [anon_sym_prototype] = ACTIONS(3069),
    [anon_sym_pop] = ACTIONS(3069),
    [anon_sym_pos] = ACTIONS(3069),
    [anon_sym_quotemeta] = ACTIONS(3069),
    [anon_sym_reset] = ACTIONS(3069),
    [anon_sym_rand] = ACTIONS(3069),
    [anon_sym_rmdir] = ACTIONS(3069),
    [anon_sym_readdir] = ACTIONS(3069),
    [anon_sym_readline] = ACTIONS(3069),
    [anon_sym_readpipe] = ACTIONS(3069),
    [anon_sym_rewinddir] = ACTIONS(3069),
    [anon_sym_readlink] = ACTIONS(3069),
    [anon_sym_ref] = ACTIONS(3069),
    [anon_sym_scalar] = ACTIONS(3069),
    [anon_sym_shift] = ACTIONS(3069),
    [anon_sym_sin] = ACTIONS(3069),
    [anon_sym_sleep] = ACTIONS(3069),
    [anon_sym_sqrt] = ACTIONS(3069),
    [anon_sym_srand] = ACTIONS(3069),
    [anon_sym_stat] = ACTIONS(3069),
    [anon_sym_study] = ACTIONS(3069),
    [anon_sym_tell] = ACTIONS(3069),
    [anon_sym_telldir] = ACTIONS(3069),
    [anon_sym_tied] = ACTIONS(3069),
    [anon_sym_uc] = ACTIONS(3069),
    [anon_sym_ucfirst] = ACTIONS(3069),
    [anon_sym_untie] = ACTIONS(3069),
    [anon_sym_umask] = ACTIONS(3069),
    [anon_sym_values] = ACTIONS(3069),
    [anon_sym_write] = ACTIONS(3069),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3069),
    [anon_sym___END__] = ACTIONS(3069),
    [anon_sym_EOT] = ACTIONS(3069),
    [anon_sym_q] = ACTIONS(3069),
    [anon_sym_qq] = ACTIONS(3069),
    [anon_sym_qw] = ACTIONS(3069),
    [anon_sym_qx] = ACTIONS(3069),
    [anon_sym_qr] = ACTIONS(3069),
    [anon_sym_m] = ACTIONS(3069),
    [anon_sym_s] = ACTIONS(3069),
    [anon_sym_tr] = ACTIONS(3069),
    [anon_sym_y] = ACTIONS(3069),
    [anon_sym_if] = ACTIONS(3069),
    [anon_sym_unless] = ACTIONS(3069),
    [anon_sym_while] = ACTIONS(3069),
    [anon_sym_until] = ACTIONS(3069),
    [aux_sym__bareword_token1] = ACTIONS(3069),
    [sym_number] = ACTIONS(3069),
    [anon_sym_true] = ACTIONS(3069),
    [anon_sym_false] = ACTIONS(3069),
    [sym__single_quote] = ACTIONS(3067),
    [sym__double_quote] = ACTIONS(3067),
    [sym__backtick_quote] = ACTIONS(3067),
    [sym__search_slash_quote] = ACTIONS(3067),
    [sym__open_readline_bracket] = ACTIONS(3067),
    [sym__open_fileglob_bracket] = ACTIONS(3067),
    [sym__PERLY_HEREDOC] = ACTIONS(3067),
    [sym__ctrl_z_hack] = ACTIONS(3067),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3067),
    [sym__filetest] = ACTIONS(3067),
  },
  [STATE(1479)] = {
    [sym_heredoc_content] = STATE(1479),
    [ts_builtin_sym_end] = ACTIONS(3071),
    [sym__identifier] = ACTIONS(3073),
    [anon_sym_LBRACE] = ACTIONS(3073),
    [anon_sym_SEMI] = ACTIONS(3073),
    [anon_sym_package] = ACTIONS(3073),
    [anon_sym_class] = ACTIONS(3073),
    [anon_sym_role] = ACTIONS(3073),
    [anon_sym_BUILD] = ACTIONS(3073),
    [anon_sym_ADJUST] = ACTIONS(3073),
    [anon_sym_DOLLAR] = ACTIONS(3073),
    [anon_sym_AT] = ACTIONS(3073),
    [anon_sym_my] = ACTIONS(3073),
    [anon_sym_extended] = ACTIONS(3073),
    [anon_sym_async] = ACTIONS(3073),
    [anon_sym_sub] = ACTIONS(3073),
    [anon_sym_method] = ACTIONS(3073),
    [anon_sym_LPAREN] = ACTIONS(3073),
    [anon_sym_continue] = ACTIONS(3073),
    [anon_sym_state] = ACTIONS(3073),
    [anon_sym_our] = ACTIONS(3073),
    [anon_sym_try] = ACTIONS(3073),
    [anon_sym_defer] = ACTIONS(3073),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3073),
    [anon_sym_LBRACK] = ACTIONS(3073),
    [anon_sym_LT_LT] = ACTIONS(3073),
    [anon_sym_SLASH_SLASH] = ACTIONS(3073),
    [anon_sym_PLUS] = ACTIONS(3073),
    [anon_sym_DASH] = ACTIONS(3073),
    [anon_sym_TILDE] = ACTIONS(3073),
    [anon_sym_BANG] = ACTIONS(3073),
    [anon_sym_PLUS_PLUS] = ACTIONS(3073),
    [anon_sym_DASH_DASH] = ACTIONS(3073),
    [anon_sym_BSLASH] = ACTIONS(3071),
    [anon_sym_do] = ACTIONS(3073),
    [anon_sym_eval] = ACTIONS(3073),
    [anon_sym_await] = ACTIONS(3073),
    [anon_sym_field] = ACTIONS(3073),
    [anon_sym_local] = ACTIONS(3073),
    [anon_sym_dynamically] = ACTIONS(3073),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3073),
    [anon_sym_require] = ACTIONS(3073),
    [anon_sym_map] = ACTIONS(3073),
    [anon_sym_grep] = ACTIONS(3073),
    [anon_sym_sort] = ACTIONS(3073),
    [anon_sym_goto] = ACTIONS(3073),
    [anon_sym_return] = ACTIONS(3073),
    [anon_sym_undef] = ACTIONS(3073),
    [anon_sym_PERCENT2] = ACTIONS(3073),
    [anon_sym_AMP2] = ACTIONS(3073),
    [anon_sym_STAR2] = ACTIONS(3073),
    [anon_sym_use] = ACTIONS(3073),
    [anon_sym_no] = ACTIONS(3073),
    [anon_sym_for] = ACTIONS(3073),
    [anon_sym_foreach] = ACTIONS(3073),
    [anon_sym_last] = ACTIONS(3073),
    [anon_sym_next] = ACTIONS(3073),
    [anon_sym_redo] = ACTIONS(3073),
    [anon_sym_BEGIN] = ACTIONS(3073),
    [anon_sym_INIT] = ACTIONS(3073),
    [anon_sym_CHECK] = ACTIONS(3073),
    [anon_sym_UNITCHECK] = ACTIONS(3073),
    [anon_sym_END] = ACTIONS(3073),
    [anon_sym___FILE__] = ACTIONS(3073),
    [anon_sym___LINE__] = ACTIONS(3073),
    [anon_sym___PACKAGE__] = ACTIONS(3073),
    [anon_sym___SUB__] = ACTIONS(3073),
    [anon_sym_break] = ACTIONS(3073),
    [anon_sym_fork] = ACTIONS(3073),
    [anon_sym_getppid] = ACTIONS(3073),
    [anon_sym_time] = ACTIONS(3073),
    [anon_sym_times] = ACTIONS(3073),
    [anon_sym_wait] = ACTIONS(3073),
    [anon_sym_wantarray] = ACTIONS(3073),
    [anon_sym_abs] = ACTIONS(3073),
    [anon_sym_alarm] = ACTIONS(3073),
    [anon_sym_chop] = ACTIONS(3073),
    [anon_sym_chdir] = ACTIONS(3073),
    [anon_sym_close] = ACTIONS(3073),
    [anon_sym_closedir] = ACTIONS(3073),
    [anon_sym_caller] = ACTIONS(3073),
    [anon_sym_chomp] = ACTIONS(3073),
    [anon_sym_chr] = ACTIONS(3073),
    [anon_sym_cos] = ACTIONS(3073),
    [anon_sym_chroot] = ACTIONS(3073),
    [anon_sym_defined] = ACTIONS(3073),
    [anon_sym_delete] = ACTIONS(3073),
    [anon_sym_dbmclose] = ACTIONS(3073),
    [anon_sym_exists] = ACTIONS(3073),
    [anon_sym_exit] = ACTIONS(3073),
    [anon_sym_eof] = ACTIONS(3073),
    [anon_sym_exp] = ACTIONS(3073),
    [anon_sym_each] = ACTIONS(3073),
    [anon_sym_fc] = ACTIONS(3073),
    [anon_sym_fileno] = ACTIONS(3073),
    [anon_sym_gmtime] = ACTIONS(3073),
    [anon_sym_getc] = ACTIONS(3073),
    [anon_sym_getpgrp] = ACTIONS(3073),
    [anon_sym_getprotobyname] = ACTIONS(3073),
    [anon_sym_getpwname] = ACTIONS(3073),
    [anon_sym_getpwuid] = ACTIONS(3073),
    [anon_sym_getpeername] = ACTIONS(3073),
    [anon_sym_getnetbyname] = ACTIONS(3073),
    [anon_sym_getsockname] = ACTIONS(3073),
    [anon_sym_getgrnam] = ACTIONS(3073),
    [anon_sym_getgrgid] = ACTIONS(3073),
    [anon_sym_hex] = ACTIONS(3073),
    [anon_sym_int] = ACTIONS(3073),
    [anon_sym_keys] = ACTIONS(3073),
    [anon_sym_lc] = ACTIONS(3073),
    [anon_sym_lcfirst] = ACTIONS(3073),
    [anon_sym_length] = ACTIONS(3073),
    [anon_sym_localtime] = ACTIONS(3073),
    [anon_sym_log] = ACTIONS(3073),
    [anon_sym_lock] = ACTIONS(3073),
    [anon_sym_lstat] = ACTIONS(3073),
    [anon_sym_oct] = ACTIONS(3073),
    [anon_sym_ord] = ACTIONS(3073),
    [anon_sym_prototype] = ACTIONS(3073),
    [anon_sym_pop] = ACTIONS(3073),
    [anon_sym_pos] = ACTIONS(3073),
    [anon_sym_quotemeta] = ACTIONS(3073),
    [anon_sym_reset] = ACTIONS(3073),
    [anon_sym_rand] = ACTIONS(3073),
    [anon_sym_rmdir] = ACTIONS(3073),
    [anon_sym_readdir] = ACTIONS(3073),
    [anon_sym_readline] = ACTIONS(3073),
    [anon_sym_readpipe] = ACTIONS(3073),
    [anon_sym_rewinddir] = ACTIONS(3073),
    [anon_sym_readlink] = ACTIONS(3073),
    [anon_sym_ref] = ACTIONS(3073),
    [anon_sym_scalar] = ACTIONS(3073),
    [anon_sym_shift] = ACTIONS(3073),
    [anon_sym_sin] = ACTIONS(3073),
    [anon_sym_sleep] = ACTIONS(3073),
    [anon_sym_sqrt] = ACTIONS(3073),
    [anon_sym_srand] = ACTIONS(3073),
    [anon_sym_stat] = ACTIONS(3073),
    [anon_sym_study] = ACTIONS(3073),
    [anon_sym_tell] = ACTIONS(3073),
    [anon_sym_telldir] = ACTIONS(3073),
    [anon_sym_tied] = ACTIONS(3073),
    [anon_sym_uc] = ACTIONS(3073),
    [anon_sym_ucfirst] = ACTIONS(3073),
    [anon_sym_untie] = ACTIONS(3073),
    [anon_sym_umask] = ACTIONS(3073),
    [anon_sym_values] = ACTIONS(3073),
    [anon_sym_write] = ACTIONS(3073),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3073),
    [anon_sym___END__] = ACTIONS(3073),
    [anon_sym_EOT] = ACTIONS(3073),
    [anon_sym_q] = ACTIONS(3073),
    [anon_sym_qq] = ACTIONS(3073),
    [anon_sym_qw] = ACTIONS(3073),
    [anon_sym_qx] = ACTIONS(3073),
    [anon_sym_qr] = ACTIONS(3073),
    [anon_sym_m] = ACTIONS(3073),
    [anon_sym_s] = ACTIONS(3073),
    [anon_sym_tr] = ACTIONS(3073),
    [anon_sym_y] = ACTIONS(3073),
    [anon_sym_if] = ACTIONS(3073),
    [anon_sym_unless] = ACTIONS(3073),
    [anon_sym_while] = ACTIONS(3073),
    [anon_sym_until] = ACTIONS(3073),
    [aux_sym__bareword_token1] = ACTIONS(3073),
    [sym_number] = ACTIONS(3073),
    [anon_sym_true] = ACTIONS(3073),
    [anon_sym_false] = ACTIONS(3073),
    [sym__single_quote] = ACTIONS(3071),
    [sym__double_quote] = ACTIONS(3071),
    [sym__backtick_quote] = ACTIONS(3071),
    [sym__search_slash_quote] = ACTIONS(3071),
    [sym__open_readline_bracket] = ACTIONS(3071),
    [sym__open_fileglob_bracket] = ACTIONS(3071),
    [sym__PERLY_HEREDOC] = ACTIONS(3071),
    [sym__ctrl_z_hack] = ACTIONS(3071),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3071),
    [sym__filetest] = ACTIONS(3071),
  },
  [STATE(1480)] = {
    [sym_heredoc_content] = STATE(1480),
    [ts_builtin_sym_end] = ACTIONS(3075),
    [sym__identifier] = ACTIONS(3077),
    [anon_sym_LBRACE] = ACTIONS(3077),
    [anon_sym_SEMI] = ACTIONS(3077),
    [anon_sym_package] = ACTIONS(3077),
    [anon_sym_class] = ACTIONS(3077),
    [anon_sym_role] = ACTIONS(3077),
    [anon_sym_BUILD] = ACTIONS(3077),
    [anon_sym_ADJUST] = ACTIONS(3077),
    [anon_sym_DOLLAR] = ACTIONS(3077),
    [anon_sym_AT] = ACTIONS(3077),
    [anon_sym_my] = ACTIONS(3077),
    [anon_sym_extended] = ACTIONS(3077),
    [anon_sym_async] = ACTIONS(3077),
    [anon_sym_sub] = ACTIONS(3077),
    [anon_sym_method] = ACTIONS(3077),
    [anon_sym_LPAREN] = ACTIONS(3077),
    [anon_sym_continue] = ACTIONS(3077),
    [anon_sym_state] = ACTIONS(3077),
    [anon_sym_our] = ACTIONS(3077),
    [anon_sym_try] = ACTIONS(3077),
    [anon_sym_defer] = ACTIONS(3077),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3077),
    [anon_sym_LBRACK] = ACTIONS(3077),
    [anon_sym_LT_LT] = ACTIONS(3077),
    [anon_sym_SLASH_SLASH] = ACTIONS(3077),
    [anon_sym_PLUS] = ACTIONS(3077),
    [anon_sym_DASH] = ACTIONS(3077),
    [anon_sym_TILDE] = ACTIONS(3077),
    [anon_sym_BANG] = ACTIONS(3077),
    [anon_sym_PLUS_PLUS] = ACTIONS(3077),
    [anon_sym_DASH_DASH] = ACTIONS(3077),
    [anon_sym_BSLASH] = ACTIONS(3075),
    [anon_sym_do] = ACTIONS(3077),
    [anon_sym_eval] = ACTIONS(3077),
    [anon_sym_await] = ACTIONS(3077),
    [anon_sym_field] = ACTIONS(3077),
    [anon_sym_local] = ACTIONS(3077),
    [anon_sym_dynamically] = ACTIONS(3077),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3077),
    [anon_sym_require] = ACTIONS(3077),
    [anon_sym_map] = ACTIONS(3077),
    [anon_sym_grep] = ACTIONS(3077),
    [anon_sym_sort] = ACTIONS(3077),
    [anon_sym_goto] = ACTIONS(3077),
    [anon_sym_return] = ACTIONS(3077),
    [anon_sym_undef] = ACTIONS(3077),
    [anon_sym_PERCENT2] = ACTIONS(3077),
    [anon_sym_AMP2] = ACTIONS(3077),
    [anon_sym_STAR2] = ACTIONS(3077),
    [anon_sym_use] = ACTIONS(3077),
    [anon_sym_no] = ACTIONS(3077),
    [anon_sym_for] = ACTIONS(3077),
    [anon_sym_foreach] = ACTIONS(3077),
    [anon_sym_last] = ACTIONS(3077),
    [anon_sym_next] = ACTIONS(3077),
    [anon_sym_redo] = ACTIONS(3077),
    [anon_sym_BEGIN] = ACTIONS(3077),
    [anon_sym_INIT] = ACTIONS(3077),
    [anon_sym_CHECK] = ACTIONS(3077),
    [anon_sym_UNITCHECK] = ACTIONS(3077),
    [anon_sym_END] = ACTIONS(3077),
    [anon_sym___FILE__] = ACTIONS(3077),
    [anon_sym___LINE__] = ACTIONS(3077),
    [anon_sym___PACKAGE__] = ACTIONS(3077),
    [anon_sym___SUB__] = ACTIONS(3077),
    [anon_sym_break] = ACTIONS(3077),
    [anon_sym_fork] = ACTIONS(3077),
    [anon_sym_getppid] = ACTIONS(3077),
    [anon_sym_time] = ACTIONS(3077),
    [anon_sym_times] = ACTIONS(3077),
    [anon_sym_wait] = ACTIONS(3077),
    [anon_sym_wantarray] = ACTIONS(3077),
    [anon_sym_abs] = ACTIONS(3077),
    [anon_sym_alarm] = ACTIONS(3077),
    [anon_sym_chop] = ACTIONS(3077),
    [anon_sym_chdir] = ACTIONS(3077),
    [anon_sym_close] = ACTIONS(3077),
    [anon_sym_closedir] = ACTIONS(3077),
    [anon_sym_caller] = ACTIONS(3077),
    [anon_sym_chomp] = ACTIONS(3077),
    [anon_sym_chr] = ACTIONS(3077),
    [anon_sym_cos] = ACTIONS(3077),
    [anon_sym_chroot] = ACTIONS(3077),
    [anon_sym_defined] = ACTIONS(3077),
    [anon_sym_delete] = ACTIONS(3077),
    [anon_sym_dbmclose] = ACTIONS(3077),
    [anon_sym_exists] = ACTIONS(3077),
    [anon_sym_exit] = ACTIONS(3077),
    [anon_sym_eof] = ACTIONS(3077),
    [anon_sym_exp] = ACTIONS(3077),
    [anon_sym_each] = ACTIONS(3077),
    [anon_sym_fc] = ACTIONS(3077),
    [anon_sym_fileno] = ACTIONS(3077),
    [anon_sym_gmtime] = ACTIONS(3077),
    [anon_sym_getc] = ACTIONS(3077),
    [anon_sym_getpgrp] = ACTIONS(3077),
    [anon_sym_getprotobyname] = ACTIONS(3077),
    [anon_sym_getpwname] = ACTIONS(3077),
    [anon_sym_getpwuid] = ACTIONS(3077),
    [anon_sym_getpeername] = ACTIONS(3077),
    [anon_sym_getnetbyname] = ACTIONS(3077),
    [anon_sym_getsockname] = ACTIONS(3077),
    [anon_sym_getgrnam] = ACTIONS(3077),
    [anon_sym_getgrgid] = ACTIONS(3077),
    [anon_sym_hex] = ACTIONS(3077),
    [anon_sym_int] = ACTIONS(3077),
    [anon_sym_keys] = ACTIONS(3077),
    [anon_sym_lc] = ACTIONS(3077),
    [anon_sym_lcfirst] = ACTIONS(3077),
    [anon_sym_length] = ACTIONS(3077),
    [anon_sym_localtime] = ACTIONS(3077),
    [anon_sym_log] = ACTIONS(3077),
    [anon_sym_lock] = ACTIONS(3077),
    [anon_sym_lstat] = ACTIONS(3077),
    [anon_sym_oct] = ACTIONS(3077),
    [anon_sym_ord] = ACTIONS(3077),
    [anon_sym_prototype] = ACTIONS(3077),
    [anon_sym_pop] = ACTIONS(3077),
    [anon_sym_pos] = ACTIONS(3077),
    [anon_sym_quotemeta] = ACTIONS(3077),
    [anon_sym_reset] = ACTIONS(3077),
    [anon_sym_rand] = ACTIONS(3077),
    [anon_sym_rmdir] = ACTIONS(3077),
    [anon_sym_readdir] = ACTIONS(3077),
    [anon_sym_readline] = ACTIONS(3077),
    [anon_sym_readpipe] = ACTIONS(3077),
    [anon_sym_rewinddir] = ACTIONS(3077),
    [anon_sym_readlink] = ACTIONS(3077),
    [anon_sym_ref] = ACTIONS(3077),
    [anon_sym_scalar] = ACTIONS(3077),
    [anon_sym_shift] = ACTIONS(3077),
    [anon_sym_sin] = ACTIONS(3077),
    [anon_sym_sleep] = ACTIONS(3077),
    [anon_sym_sqrt] = ACTIONS(3077),
    [anon_sym_srand] = ACTIONS(3077),
    [anon_sym_stat] = ACTIONS(3077),
    [anon_sym_study] = ACTIONS(3077),
    [anon_sym_tell] = ACTIONS(3077),
    [anon_sym_telldir] = ACTIONS(3077),
    [anon_sym_tied] = ACTIONS(3077),
    [anon_sym_uc] = ACTIONS(3077),
    [anon_sym_ucfirst] = ACTIONS(3077),
    [anon_sym_untie] = ACTIONS(3077),
    [anon_sym_umask] = ACTIONS(3077),
    [anon_sym_values] = ACTIONS(3077),
    [anon_sym_write] = ACTIONS(3077),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3077),
    [anon_sym___END__] = ACTIONS(3077),
    [anon_sym_EOT] = ACTIONS(3077),
    [anon_sym_q] = ACTIONS(3077),
    [anon_sym_qq] = ACTIONS(3077),
    [anon_sym_qw] = ACTIONS(3077),
    [anon_sym_qx] = ACTIONS(3077),
    [anon_sym_qr] = ACTIONS(3077),
    [anon_sym_m] = ACTIONS(3077),
    [anon_sym_s] = ACTIONS(3077),
    [anon_sym_tr] = ACTIONS(3077),
    [anon_sym_y] = ACTIONS(3077),
    [anon_sym_if] = ACTIONS(3077),
    [anon_sym_unless] = ACTIONS(3077),
    [anon_sym_while] = ACTIONS(3077),
    [anon_sym_until] = ACTIONS(3077),
    [aux_sym__bareword_token1] = ACTIONS(3077),
    [sym_number] = ACTIONS(3077),
    [anon_sym_true] = ACTIONS(3077),
    [anon_sym_false] = ACTIONS(3077),
    [sym__single_quote] = ACTIONS(3075),
    [sym__double_quote] = ACTIONS(3075),
    [sym__backtick_quote] = ACTIONS(3075),
    [sym__search_slash_quote] = ACTIONS(3075),
    [sym__open_readline_bracket] = ACTIONS(3075),
    [sym__open_fileglob_bracket] = ACTIONS(3075),
    [sym__PERLY_HEREDOC] = ACTIONS(3075),
    [sym__ctrl_z_hack] = ACTIONS(3075),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3075),
    [sym__filetest] = ACTIONS(3075),
  },
  [STATE(1481)] = {
    [sym_heredoc_content] = STATE(1481),
    [ts_builtin_sym_end] = ACTIONS(3079),
    [sym__identifier] = ACTIONS(3081),
    [anon_sym_LBRACE] = ACTIONS(3081),
    [anon_sym_SEMI] = ACTIONS(3081),
    [anon_sym_package] = ACTIONS(3081),
    [anon_sym_class] = ACTIONS(3081),
    [anon_sym_role] = ACTIONS(3081),
    [anon_sym_BUILD] = ACTIONS(3081),
    [anon_sym_ADJUST] = ACTIONS(3081),
    [anon_sym_DOLLAR] = ACTIONS(3081),
    [anon_sym_AT] = ACTIONS(3081),
    [anon_sym_my] = ACTIONS(3081),
    [anon_sym_extended] = ACTIONS(3081),
    [anon_sym_async] = ACTIONS(3081),
    [anon_sym_sub] = ACTIONS(3081),
    [anon_sym_method] = ACTIONS(3081),
    [anon_sym_LPAREN] = ACTIONS(3081),
    [anon_sym_continue] = ACTIONS(3081),
    [anon_sym_state] = ACTIONS(3081),
    [anon_sym_our] = ACTIONS(3081),
    [anon_sym_try] = ACTIONS(3081),
    [anon_sym_defer] = ACTIONS(3081),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3081),
    [anon_sym_LBRACK] = ACTIONS(3081),
    [anon_sym_LT_LT] = ACTIONS(3081),
    [anon_sym_SLASH_SLASH] = ACTIONS(3081),
    [anon_sym_PLUS] = ACTIONS(3081),
    [anon_sym_DASH] = ACTIONS(3081),
    [anon_sym_TILDE] = ACTIONS(3081),
    [anon_sym_BANG] = ACTIONS(3081),
    [anon_sym_PLUS_PLUS] = ACTIONS(3081),
    [anon_sym_DASH_DASH] = ACTIONS(3081),
    [anon_sym_BSLASH] = ACTIONS(3079),
    [anon_sym_do] = ACTIONS(3081),
    [anon_sym_eval] = ACTIONS(3081),
    [anon_sym_await] = ACTIONS(3081),
    [anon_sym_field] = ACTIONS(3081),
    [anon_sym_local] = ACTIONS(3081),
    [anon_sym_dynamically] = ACTIONS(3081),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3081),
    [anon_sym_require] = ACTIONS(3081),
    [anon_sym_map] = ACTIONS(3081),
    [anon_sym_grep] = ACTIONS(3081),
    [anon_sym_sort] = ACTIONS(3081),
    [anon_sym_goto] = ACTIONS(3081),
    [anon_sym_return] = ACTIONS(3081),
    [anon_sym_undef] = ACTIONS(3081),
    [anon_sym_PERCENT2] = ACTIONS(3081),
    [anon_sym_AMP2] = ACTIONS(3081),
    [anon_sym_STAR2] = ACTIONS(3081),
    [anon_sym_use] = ACTIONS(3081),
    [anon_sym_no] = ACTIONS(3081),
    [anon_sym_for] = ACTIONS(3081),
    [anon_sym_foreach] = ACTIONS(3081),
    [anon_sym_last] = ACTIONS(3081),
    [anon_sym_next] = ACTIONS(3081),
    [anon_sym_redo] = ACTIONS(3081),
    [anon_sym_BEGIN] = ACTIONS(3081),
    [anon_sym_INIT] = ACTIONS(3081),
    [anon_sym_CHECK] = ACTIONS(3081),
    [anon_sym_UNITCHECK] = ACTIONS(3081),
    [anon_sym_END] = ACTIONS(3081),
    [anon_sym___FILE__] = ACTIONS(3081),
    [anon_sym___LINE__] = ACTIONS(3081),
    [anon_sym___PACKAGE__] = ACTIONS(3081),
    [anon_sym___SUB__] = ACTIONS(3081),
    [anon_sym_break] = ACTIONS(3081),
    [anon_sym_fork] = ACTIONS(3081),
    [anon_sym_getppid] = ACTIONS(3081),
    [anon_sym_time] = ACTIONS(3081),
    [anon_sym_times] = ACTIONS(3081),
    [anon_sym_wait] = ACTIONS(3081),
    [anon_sym_wantarray] = ACTIONS(3081),
    [anon_sym_abs] = ACTIONS(3081),
    [anon_sym_alarm] = ACTIONS(3081),
    [anon_sym_chop] = ACTIONS(3081),
    [anon_sym_chdir] = ACTIONS(3081),
    [anon_sym_close] = ACTIONS(3081),
    [anon_sym_closedir] = ACTIONS(3081),
    [anon_sym_caller] = ACTIONS(3081),
    [anon_sym_chomp] = ACTIONS(3081),
    [anon_sym_chr] = ACTIONS(3081),
    [anon_sym_cos] = ACTIONS(3081),
    [anon_sym_chroot] = ACTIONS(3081),
    [anon_sym_defined] = ACTIONS(3081),
    [anon_sym_delete] = ACTIONS(3081),
    [anon_sym_dbmclose] = ACTIONS(3081),
    [anon_sym_exists] = ACTIONS(3081),
    [anon_sym_exit] = ACTIONS(3081),
    [anon_sym_eof] = ACTIONS(3081),
    [anon_sym_exp] = ACTIONS(3081),
    [anon_sym_each] = ACTIONS(3081),
    [anon_sym_fc] = ACTIONS(3081),
    [anon_sym_fileno] = ACTIONS(3081),
    [anon_sym_gmtime] = ACTIONS(3081),
    [anon_sym_getc] = ACTIONS(3081),
    [anon_sym_getpgrp] = ACTIONS(3081),
    [anon_sym_getprotobyname] = ACTIONS(3081),
    [anon_sym_getpwname] = ACTIONS(3081),
    [anon_sym_getpwuid] = ACTIONS(3081),
    [anon_sym_getpeername] = ACTIONS(3081),
    [anon_sym_getnetbyname] = ACTIONS(3081),
    [anon_sym_getsockname] = ACTIONS(3081),
    [anon_sym_getgrnam] = ACTIONS(3081),
    [anon_sym_getgrgid] = ACTIONS(3081),
    [anon_sym_hex] = ACTIONS(3081),
    [anon_sym_int] = ACTIONS(3081),
    [anon_sym_keys] = ACTIONS(3081),
    [anon_sym_lc] = ACTIONS(3081),
    [anon_sym_lcfirst] = ACTIONS(3081),
    [anon_sym_length] = ACTIONS(3081),
    [anon_sym_localtime] = ACTIONS(3081),
    [anon_sym_log] = ACTIONS(3081),
    [anon_sym_lock] = ACTIONS(3081),
    [anon_sym_lstat] = ACTIONS(3081),
    [anon_sym_oct] = ACTIONS(3081),
    [anon_sym_ord] = ACTIONS(3081),
    [anon_sym_prototype] = ACTIONS(3081),
    [anon_sym_pop] = ACTIONS(3081),
    [anon_sym_pos] = ACTIONS(3081),
    [anon_sym_quotemeta] = ACTIONS(3081),
    [anon_sym_reset] = ACTIONS(3081),
    [anon_sym_rand] = ACTIONS(3081),
    [anon_sym_rmdir] = ACTIONS(3081),
    [anon_sym_readdir] = ACTIONS(3081),
    [anon_sym_readline] = ACTIONS(3081),
    [anon_sym_readpipe] = ACTIONS(3081),
    [anon_sym_rewinddir] = ACTIONS(3081),
    [anon_sym_readlink] = ACTIONS(3081),
    [anon_sym_ref] = ACTIONS(3081),
    [anon_sym_scalar] = ACTIONS(3081),
    [anon_sym_shift] = ACTIONS(3081),
    [anon_sym_sin] = ACTIONS(3081),
    [anon_sym_sleep] = ACTIONS(3081),
    [anon_sym_sqrt] = ACTIONS(3081),
    [anon_sym_srand] = ACTIONS(3081),
    [anon_sym_stat] = ACTIONS(3081),
    [anon_sym_study] = ACTIONS(3081),
    [anon_sym_tell] = ACTIONS(3081),
    [anon_sym_telldir] = ACTIONS(3081),
    [anon_sym_tied] = ACTIONS(3081),
    [anon_sym_uc] = ACTIONS(3081),
    [anon_sym_ucfirst] = ACTIONS(3081),
    [anon_sym_untie] = ACTIONS(3081),
    [anon_sym_umask] = ACTIONS(3081),
    [anon_sym_values] = ACTIONS(3081),
    [anon_sym_write] = ACTIONS(3081),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3081),
    [anon_sym___END__] = ACTIONS(3081),
    [anon_sym_EOT] = ACTIONS(3081),
    [anon_sym_q] = ACTIONS(3081),
    [anon_sym_qq] = ACTIONS(3081),
    [anon_sym_qw] = ACTIONS(3081),
    [anon_sym_qx] = ACTIONS(3081),
    [anon_sym_qr] = ACTIONS(3081),
    [anon_sym_m] = ACTIONS(3081),
    [anon_sym_s] = ACTIONS(3081),
    [anon_sym_tr] = ACTIONS(3081),
    [anon_sym_y] = ACTIONS(3081),
    [anon_sym_if] = ACTIONS(3081),
    [anon_sym_unless] = ACTIONS(3081),
    [anon_sym_while] = ACTIONS(3081),
    [anon_sym_until] = ACTIONS(3081),
    [aux_sym__bareword_token1] = ACTIONS(3081),
    [sym_number] = ACTIONS(3081),
    [anon_sym_true] = ACTIONS(3081),
    [anon_sym_false] = ACTIONS(3081),
    [sym__single_quote] = ACTIONS(3079),
    [sym__double_quote] = ACTIONS(3079),
    [sym__backtick_quote] = ACTIONS(3079),
    [sym__search_slash_quote] = ACTIONS(3079),
    [sym__open_readline_bracket] = ACTIONS(3079),
    [sym__open_fileglob_bracket] = ACTIONS(3079),
    [sym__PERLY_HEREDOC] = ACTIONS(3079),
    [sym__ctrl_z_hack] = ACTIONS(3079),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3079),
    [sym__filetest] = ACTIONS(3079),
  },
  [STATE(1482)] = {
    [sym_heredoc_content] = STATE(1482),
    [ts_builtin_sym_end] = ACTIONS(3083),
    [sym__identifier] = ACTIONS(3085),
    [anon_sym_LBRACE] = ACTIONS(3085),
    [anon_sym_SEMI] = ACTIONS(3085),
    [anon_sym_package] = ACTIONS(3085),
    [anon_sym_class] = ACTIONS(3085),
    [anon_sym_role] = ACTIONS(3085),
    [anon_sym_BUILD] = ACTIONS(3085),
    [anon_sym_ADJUST] = ACTIONS(3085),
    [anon_sym_DOLLAR] = ACTIONS(3085),
    [anon_sym_AT] = ACTIONS(3085),
    [anon_sym_my] = ACTIONS(3085),
    [anon_sym_extended] = ACTIONS(3085),
    [anon_sym_async] = ACTIONS(3085),
    [anon_sym_sub] = ACTIONS(3085),
    [anon_sym_method] = ACTIONS(3085),
    [anon_sym_LPAREN] = ACTIONS(3085),
    [anon_sym_continue] = ACTIONS(3085),
    [anon_sym_state] = ACTIONS(3085),
    [anon_sym_our] = ACTIONS(3085),
    [anon_sym_try] = ACTIONS(3085),
    [anon_sym_defer] = ACTIONS(3085),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3085),
    [anon_sym_LBRACK] = ACTIONS(3085),
    [anon_sym_LT_LT] = ACTIONS(3085),
    [anon_sym_SLASH_SLASH] = ACTIONS(3085),
    [anon_sym_PLUS] = ACTIONS(3085),
    [anon_sym_DASH] = ACTIONS(3085),
    [anon_sym_TILDE] = ACTIONS(3085),
    [anon_sym_BANG] = ACTIONS(3085),
    [anon_sym_PLUS_PLUS] = ACTIONS(3085),
    [anon_sym_DASH_DASH] = ACTIONS(3085),
    [anon_sym_BSLASH] = ACTIONS(3083),
    [anon_sym_do] = ACTIONS(3085),
    [anon_sym_eval] = ACTIONS(3085),
    [anon_sym_await] = ACTIONS(3085),
    [anon_sym_field] = ACTIONS(3085),
    [anon_sym_local] = ACTIONS(3085),
    [anon_sym_dynamically] = ACTIONS(3085),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3085),
    [anon_sym_require] = ACTIONS(3085),
    [anon_sym_map] = ACTIONS(3085),
    [anon_sym_grep] = ACTIONS(3085),
    [anon_sym_sort] = ACTIONS(3085),
    [anon_sym_goto] = ACTIONS(3085),
    [anon_sym_return] = ACTIONS(3085),
    [anon_sym_undef] = ACTIONS(3085),
    [anon_sym_PERCENT2] = ACTIONS(3085),
    [anon_sym_AMP2] = ACTIONS(3085),
    [anon_sym_STAR2] = ACTIONS(3085),
    [anon_sym_use] = ACTIONS(3085),
    [anon_sym_no] = ACTIONS(3085),
    [anon_sym_for] = ACTIONS(3085),
    [anon_sym_foreach] = ACTIONS(3085),
    [anon_sym_last] = ACTIONS(3085),
    [anon_sym_next] = ACTIONS(3085),
    [anon_sym_redo] = ACTIONS(3085),
    [anon_sym_BEGIN] = ACTIONS(3085),
    [anon_sym_INIT] = ACTIONS(3085),
    [anon_sym_CHECK] = ACTIONS(3085),
    [anon_sym_UNITCHECK] = ACTIONS(3085),
    [anon_sym_END] = ACTIONS(3085),
    [anon_sym___FILE__] = ACTIONS(3085),
    [anon_sym___LINE__] = ACTIONS(3085),
    [anon_sym___PACKAGE__] = ACTIONS(3085),
    [anon_sym___SUB__] = ACTIONS(3085),
    [anon_sym_break] = ACTIONS(3085),
    [anon_sym_fork] = ACTIONS(3085),
    [anon_sym_getppid] = ACTIONS(3085),
    [anon_sym_time] = ACTIONS(3085),
    [anon_sym_times] = ACTIONS(3085),
    [anon_sym_wait] = ACTIONS(3085),
    [anon_sym_wantarray] = ACTIONS(3085),
    [anon_sym_abs] = ACTIONS(3085),
    [anon_sym_alarm] = ACTIONS(3085),
    [anon_sym_chop] = ACTIONS(3085),
    [anon_sym_chdir] = ACTIONS(3085),
    [anon_sym_close] = ACTIONS(3085),
    [anon_sym_closedir] = ACTIONS(3085),
    [anon_sym_caller] = ACTIONS(3085),
    [anon_sym_chomp] = ACTIONS(3085),
    [anon_sym_chr] = ACTIONS(3085),
    [anon_sym_cos] = ACTIONS(3085),
    [anon_sym_chroot] = ACTIONS(3085),
    [anon_sym_defined] = ACTIONS(3085),
    [anon_sym_delete] = ACTIONS(3085),
    [anon_sym_dbmclose] = ACTIONS(3085),
    [anon_sym_exists] = ACTIONS(3085),
    [anon_sym_exit] = ACTIONS(3085),
    [anon_sym_eof] = ACTIONS(3085),
    [anon_sym_exp] = ACTIONS(3085),
    [anon_sym_each] = ACTIONS(3085),
    [anon_sym_fc] = ACTIONS(3085),
    [anon_sym_fileno] = ACTIONS(3085),
    [anon_sym_gmtime] = ACTIONS(3085),
    [anon_sym_getc] = ACTIONS(3085),
    [anon_sym_getpgrp] = ACTIONS(3085),
    [anon_sym_getprotobyname] = ACTIONS(3085),
    [anon_sym_getpwname] = ACTIONS(3085),
    [anon_sym_getpwuid] = ACTIONS(3085),
    [anon_sym_getpeername] = ACTIONS(3085),
    [anon_sym_getnetbyname] = ACTIONS(3085),
    [anon_sym_getsockname] = ACTIONS(3085),
    [anon_sym_getgrnam] = ACTIONS(3085),
    [anon_sym_getgrgid] = ACTIONS(3085),
    [anon_sym_hex] = ACTIONS(3085),
    [anon_sym_int] = ACTIONS(3085),
    [anon_sym_keys] = ACTIONS(3085),
    [anon_sym_lc] = ACTIONS(3085),
    [anon_sym_lcfirst] = ACTIONS(3085),
    [anon_sym_length] = ACTIONS(3085),
    [anon_sym_localtime] = ACTIONS(3085),
    [anon_sym_log] = ACTIONS(3085),
    [anon_sym_lock] = ACTIONS(3085),
    [anon_sym_lstat] = ACTIONS(3085),
    [anon_sym_oct] = ACTIONS(3085),
    [anon_sym_ord] = ACTIONS(3085),
    [anon_sym_prototype] = ACTIONS(3085),
    [anon_sym_pop] = ACTIONS(3085),
    [anon_sym_pos] = ACTIONS(3085),
    [anon_sym_quotemeta] = ACTIONS(3085),
    [anon_sym_reset] = ACTIONS(3085),
    [anon_sym_rand] = ACTIONS(3085),
    [anon_sym_rmdir] = ACTIONS(3085),
    [anon_sym_readdir] = ACTIONS(3085),
    [anon_sym_readline] = ACTIONS(3085),
    [anon_sym_readpipe] = ACTIONS(3085),
    [anon_sym_rewinddir] = ACTIONS(3085),
    [anon_sym_readlink] = ACTIONS(3085),
    [anon_sym_ref] = ACTIONS(3085),
    [anon_sym_scalar] = ACTIONS(3085),
    [anon_sym_shift] = ACTIONS(3085),
    [anon_sym_sin] = ACTIONS(3085),
    [anon_sym_sleep] = ACTIONS(3085),
    [anon_sym_sqrt] = ACTIONS(3085),
    [anon_sym_srand] = ACTIONS(3085),
    [anon_sym_stat] = ACTIONS(3085),
    [anon_sym_study] = ACTIONS(3085),
    [anon_sym_tell] = ACTIONS(3085),
    [anon_sym_telldir] = ACTIONS(3085),
    [anon_sym_tied] = ACTIONS(3085),
    [anon_sym_uc] = ACTIONS(3085),
    [anon_sym_ucfirst] = ACTIONS(3085),
    [anon_sym_untie] = ACTIONS(3085),
    [anon_sym_umask] = ACTIONS(3085),
    [anon_sym_values] = ACTIONS(3085),
    [anon_sym_write] = ACTIONS(3085),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3085),
    [anon_sym___END__] = ACTIONS(3085),
    [anon_sym_EOT] = ACTIONS(3085),
    [anon_sym_q] = ACTIONS(3085),
    [anon_sym_qq] = ACTIONS(3085),
    [anon_sym_qw] = ACTIONS(3085),
    [anon_sym_qx] = ACTIONS(3085),
    [anon_sym_qr] = ACTIONS(3085),
    [anon_sym_m] = ACTIONS(3085),
    [anon_sym_s] = ACTIONS(3085),
    [anon_sym_tr] = ACTIONS(3085),
    [anon_sym_y] = ACTIONS(3085),
    [anon_sym_if] = ACTIONS(3085),
    [anon_sym_unless] = ACTIONS(3085),
    [anon_sym_while] = ACTIONS(3085),
    [anon_sym_until] = ACTIONS(3085),
    [aux_sym__bareword_token1] = ACTIONS(3085),
    [sym_number] = ACTIONS(3085),
    [anon_sym_true] = ACTIONS(3085),
    [anon_sym_false] = ACTIONS(3085),
    [sym__single_quote] = ACTIONS(3083),
    [sym__double_quote] = ACTIONS(3083),
    [sym__backtick_quote] = ACTIONS(3083),
    [sym__search_slash_quote] = ACTIONS(3083),
    [sym__open_readline_bracket] = ACTIONS(3083),
    [sym__open_fileglob_bracket] = ACTIONS(3083),
    [sym__PERLY_HEREDOC] = ACTIONS(3083),
    [sym__ctrl_z_hack] = ACTIONS(3083),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3083),
    [sym__filetest] = ACTIONS(3083),
  },
  [STATE(1483)] = {
    [sym_heredoc_content] = STATE(1483),
    [ts_builtin_sym_end] = ACTIONS(3087),
    [sym__identifier] = ACTIONS(3089),
    [anon_sym_LBRACE] = ACTIONS(3089),
    [anon_sym_SEMI] = ACTIONS(3089),
    [anon_sym_package] = ACTIONS(3089),
    [anon_sym_class] = ACTIONS(3089),
    [anon_sym_role] = ACTIONS(3089),
    [anon_sym_BUILD] = ACTIONS(3089),
    [anon_sym_ADJUST] = ACTIONS(3089),
    [anon_sym_DOLLAR] = ACTIONS(3089),
    [anon_sym_AT] = ACTIONS(3089),
    [anon_sym_my] = ACTIONS(3089),
    [anon_sym_extended] = ACTIONS(3089),
    [anon_sym_async] = ACTIONS(3089),
    [anon_sym_sub] = ACTIONS(3089),
    [anon_sym_method] = ACTIONS(3089),
    [anon_sym_LPAREN] = ACTIONS(3089),
    [anon_sym_continue] = ACTIONS(3089),
    [anon_sym_state] = ACTIONS(3089),
    [anon_sym_our] = ACTIONS(3089),
    [anon_sym_try] = ACTIONS(3089),
    [anon_sym_defer] = ACTIONS(3089),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3089),
    [anon_sym_LBRACK] = ACTIONS(3089),
    [anon_sym_LT_LT] = ACTIONS(3089),
    [anon_sym_SLASH_SLASH] = ACTIONS(3089),
    [anon_sym_PLUS] = ACTIONS(3089),
    [anon_sym_DASH] = ACTIONS(3089),
    [anon_sym_TILDE] = ACTIONS(3089),
    [anon_sym_BANG] = ACTIONS(3089),
    [anon_sym_PLUS_PLUS] = ACTIONS(3089),
    [anon_sym_DASH_DASH] = ACTIONS(3089),
    [anon_sym_BSLASH] = ACTIONS(3087),
    [anon_sym_do] = ACTIONS(3089),
    [anon_sym_eval] = ACTIONS(3089),
    [anon_sym_await] = ACTIONS(3089),
    [anon_sym_field] = ACTIONS(3089),
    [anon_sym_local] = ACTIONS(3089),
    [anon_sym_dynamically] = ACTIONS(3089),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3089),
    [anon_sym_require] = ACTIONS(3089),
    [anon_sym_map] = ACTIONS(3089),
    [anon_sym_grep] = ACTIONS(3089),
    [anon_sym_sort] = ACTIONS(3089),
    [anon_sym_goto] = ACTIONS(3089),
    [anon_sym_return] = ACTIONS(3089),
    [anon_sym_undef] = ACTIONS(3089),
    [anon_sym_PERCENT2] = ACTIONS(3089),
    [anon_sym_AMP2] = ACTIONS(3089),
    [anon_sym_STAR2] = ACTIONS(3089),
    [anon_sym_use] = ACTIONS(3089),
    [anon_sym_no] = ACTIONS(3089),
    [anon_sym_for] = ACTIONS(3089),
    [anon_sym_foreach] = ACTIONS(3089),
    [anon_sym_last] = ACTIONS(3089),
    [anon_sym_next] = ACTIONS(3089),
    [anon_sym_redo] = ACTIONS(3089),
    [anon_sym_BEGIN] = ACTIONS(3089),
    [anon_sym_INIT] = ACTIONS(3089),
    [anon_sym_CHECK] = ACTIONS(3089),
    [anon_sym_UNITCHECK] = ACTIONS(3089),
    [anon_sym_END] = ACTIONS(3089),
    [anon_sym___FILE__] = ACTIONS(3089),
    [anon_sym___LINE__] = ACTIONS(3089),
    [anon_sym___PACKAGE__] = ACTIONS(3089),
    [anon_sym___SUB__] = ACTIONS(3089),
    [anon_sym_break] = ACTIONS(3089),
    [anon_sym_fork] = ACTIONS(3089),
    [anon_sym_getppid] = ACTIONS(3089),
    [anon_sym_time] = ACTIONS(3089),
    [anon_sym_times] = ACTIONS(3089),
    [anon_sym_wait] = ACTIONS(3089),
    [anon_sym_wantarray] = ACTIONS(3089),
    [anon_sym_abs] = ACTIONS(3089),
    [anon_sym_alarm] = ACTIONS(3089),
    [anon_sym_chop] = ACTIONS(3089),
    [anon_sym_chdir] = ACTIONS(3089),
    [anon_sym_close] = ACTIONS(3089),
    [anon_sym_closedir] = ACTIONS(3089),
    [anon_sym_caller] = ACTIONS(3089),
    [anon_sym_chomp] = ACTIONS(3089),
    [anon_sym_chr] = ACTIONS(3089),
    [anon_sym_cos] = ACTIONS(3089),
    [anon_sym_chroot] = ACTIONS(3089),
    [anon_sym_defined] = ACTIONS(3089),
    [anon_sym_delete] = ACTIONS(3089),
    [anon_sym_dbmclose] = ACTIONS(3089),
    [anon_sym_exists] = ACTIONS(3089),
    [anon_sym_exit] = ACTIONS(3089),
    [anon_sym_eof] = ACTIONS(3089),
    [anon_sym_exp] = ACTIONS(3089),
    [anon_sym_each] = ACTIONS(3089),
    [anon_sym_fc] = ACTIONS(3089),
    [anon_sym_fileno] = ACTIONS(3089),
    [anon_sym_gmtime] = ACTIONS(3089),
    [anon_sym_getc] = ACTIONS(3089),
    [anon_sym_getpgrp] = ACTIONS(3089),
    [anon_sym_getprotobyname] = ACTIONS(3089),
    [anon_sym_getpwname] = ACTIONS(3089),
    [anon_sym_getpwuid] = ACTIONS(3089),
    [anon_sym_getpeername] = ACTIONS(3089),
    [anon_sym_getnetbyname] = ACTIONS(3089),
    [anon_sym_getsockname] = ACTIONS(3089),
    [anon_sym_getgrnam] = ACTIONS(3089),
    [anon_sym_getgrgid] = ACTIONS(3089),
    [anon_sym_hex] = ACTIONS(3089),
    [anon_sym_int] = ACTIONS(3089),
    [anon_sym_keys] = ACTIONS(3089),
    [anon_sym_lc] = ACTIONS(3089),
    [anon_sym_lcfirst] = ACTIONS(3089),
    [anon_sym_length] = ACTIONS(3089),
    [anon_sym_localtime] = ACTIONS(3089),
    [anon_sym_log] = ACTIONS(3089),
    [anon_sym_lock] = ACTIONS(3089),
    [anon_sym_lstat] = ACTIONS(3089),
    [anon_sym_oct] = ACTIONS(3089),
    [anon_sym_ord] = ACTIONS(3089),
    [anon_sym_prototype] = ACTIONS(3089),
    [anon_sym_pop] = ACTIONS(3089),
    [anon_sym_pos] = ACTIONS(3089),
    [anon_sym_quotemeta] = ACTIONS(3089),
    [anon_sym_reset] = ACTIONS(3089),
    [anon_sym_rand] = ACTIONS(3089),
    [anon_sym_rmdir] = ACTIONS(3089),
    [anon_sym_readdir] = ACTIONS(3089),
    [anon_sym_readline] = ACTIONS(3089),
    [anon_sym_readpipe] = ACTIONS(3089),
    [anon_sym_rewinddir] = ACTIONS(3089),
    [anon_sym_readlink] = ACTIONS(3089),
    [anon_sym_ref] = ACTIONS(3089),
    [anon_sym_scalar] = ACTIONS(3089),
    [anon_sym_shift] = ACTIONS(3089),
    [anon_sym_sin] = ACTIONS(3089),
    [anon_sym_sleep] = ACTIONS(3089),
    [anon_sym_sqrt] = ACTIONS(3089),
    [anon_sym_srand] = ACTIONS(3089),
    [anon_sym_stat] = ACTIONS(3089),
    [anon_sym_study] = ACTIONS(3089),
    [anon_sym_tell] = ACTIONS(3089),
    [anon_sym_telldir] = ACTIONS(3089),
    [anon_sym_tied] = ACTIONS(3089),
    [anon_sym_uc] = ACTIONS(3089),
    [anon_sym_ucfirst] = ACTIONS(3089),
    [anon_sym_untie] = ACTIONS(3089),
    [anon_sym_umask] = ACTIONS(3089),
    [anon_sym_values] = ACTIONS(3089),
    [anon_sym_write] = ACTIONS(3089),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3089),
    [anon_sym___END__] = ACTIONS(3089),
    [anon_sym_EOT] = ACTIONS(3089),
    [anon_sym_q] = ACTIONS(3089),
    [anon_sym_qq] = ACTIONS(3089),
    [anon_sym_qw] = ACTIONS(3089),
    [anon_sym_qx] = ACTIONS(3089),
    [anon_sym_qr] = ACTIONS(3089),
    [anon_sym_m] = ACTIONS(3089),
    [anon_sym_s] = ACTIONS(3089),
    [anon_sym_tr] = ACTIONS(3089),
    [anon_sym_y] = ACTIONS(3089),
    [anon_sym_if] = ACTIONS(3089),
    [anon_sym_unless] = ACTIONS(3089),
    [anon_sym_while] = ACTIONS(3089),
    [anon_sym_until] = ACTIONS(3089),
    [aux_sym__bareword_token1] = ACTIONS(3089),
    [sym_number] = ACTIONS(3089),
    [anon_sym_true] = ACTIONS(3089),
    [anon_sym_false] = ACTIONS(3089),
    [sym__single_quote] = ACTIONS(3087),
    [sym__double_quote] = ACTIONS(3087),
    [sym__backtick_quote] = ACTIONS(3087),
    [sym__search_slash_quote] = ACTIONS(3087),
    [sym__open_readline_bracket] = ACTIONS(3087),
    [sym__open_fileglob_bracket] = ACTIONS(3087),
    [sym__PERLY_HEREDOC] = ACTIONS(3087),
    [sym__ctrl_z_hack] = ACTIONS(3087),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3087),
    [sym__filetest] = ACTIONS(3087),
  },
  [STATE(1484)] = {
    [sym_heredoc_content] = STATE(1484),
    [ts_builtin_sym_end] = ACTIONS(3091),
    [sym__identifier] = ACTIONS(3093),
    [anon_sym_LBRACE] = ACTIONS(3093),
    [anon_sym_SEMI] = ACTIONS(3093),
    [anon_sym_package] = ACTIONS(3093),
    [anon_sym_class] = ACTIONS(3093),
    [anon_sym_role] = ACTIONS(3093),
    [anon_sym_BUILD] = ACTIONS(3093),
    [anon_sym_ADJUST] = ACTIONS(3093),
    [anon_sym_DOLLAR] = ACTIONS(3093),
    [anon_sym_AT] = ACTIONS(3093),
    [anon_sym_my] = ACTIONS(3093),
    [anon_sym_extended] = ACTIONS(3093),
    [anon_sym_async] = ACTIONS(3093),
    [anon_sym_sub] = ACTIONS(3093),
    [anon_sym_method] = ACTIONS(3093),
    [anon_sym_LPAREN] = ACTIONS(3093),
    [anon_sym_continue] = ACTIONS(3093),
    [anon_sym_state] = ACTIONS(3093),
    [anon_sym_our] = ACTIONS(3093),
    [anon_sym_try] = ACTIONS(3093),
    [anon_sym_defer] = ACTIONS(3093),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3093),
    [anon_sym_LBRACK] = ACTIONS(3093),
    [anon_sym_LT_LT] = ACTIONS(3093),
    [anon_sym_SLASH_SLASH] = ACTIONS(3093),
    [anon_sym_PLUS] = ACTIONS(3093),
    [anon_sym_DASH] = ACTIONS(3093),
    [anon_sym_TILDE] = ACTIONS(3093),
    [anon_sym_BANG] = ACTIONS(3093),
    [anon_sym_PLUS_PLUS] = ACTIONS(3093),
    [anon_sym_DASH_DASH] = ACTIONS(3093),
    [anon_sym_BSLASH] = ACTIONS(3091),
    [anon_sym_do] = ACTIONS(3093),
    [anon_sym_eval] = ACTIONS(3093),
    [anon_sym_await] = ACTIONS(3093),
    [anon_sym_field] = ACTIONS(3093),
    [anon_sym_local] = ACTIONS(3093),
    [anon_sym_dynamically] = ACTIONS(3093),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3093),
    [anon_sym_require] = ACTIONS(3093),
    [anon_sym_map] = ACTIONS(3093),
    [anon_sym_grep] = ACTIONS(3093),
    [anon_sym_sort] = ACTIONS(3093),
    [anon_sym_goto] = ACTIONS(3093),
    [anon_sym_return] = ACTIONS(3093),
    [anon_sym_undef] = ACTIONS(3093),
    [anon_sym_PERCENT2] = ACTIONS(3093),
    [anon_sym_AMP2] = ACTIONS(3093),
    [anon_sym_STAR2] = ACTIONS(3093),
    [anon_sym_use] = ACTIONS(3093),
    [anon_sym_no] = ACTIONS(3093),
    [anon_sym_for] = ACTIONS(3093),
    [anon_sym_foreach] = ACTIONS(3093),
    [anon_sym_last] = ACTIONS(3093),
    [anon_sym_next] = ACTIONS(3093),
    [anon_sym_redo] = ACTIONS(3093),
    [anon_sym_BEGIN] = ACTIONS(3093),
    [anon_sym_INIT] = ACTIONS(3093),
    [anon_sym_CHECK] = ACTIONS(3093),
    [anon_sym_UNITCHECK] = ACTIONS(3093),
    [anon_sym_END] = ACTIONS(3093),
    [anon_sym___FILE__] = ACTIONS(3093),
    [anon_sym___LINE__] = ACTIONS(3093),
    [anon_sym___PACKAGE__] = ACTIONS(3093),
    [anon_sym___SUB__] = ACTIONS(3093),
    [anon_sym_break] = ACTIONS(3093),
    [anon_sym_fork] = ACTIONS(3093),
    [anon_sym_getppid] = ACTIONS(3093),
    [anon_sym_time] = ACTIONS(3093),
    [anon_sym_times] = ACTIONS(3093),
    [anon_sym_wait] = ACTIONS(3093),
    [anon_sym_wantarray] = ACTIONS(3093),
    [anon_sym_abs] = ACTIONS(3093),
    [anon_sym_alarm] = ACTIONS(3093),
    [anon_sym_chop] = ACTIONS(3093),
    [anon_sym_chdir] = ACTIONS(3093),
    [anon_sym_close] = ACTIONS(3093),
    [anon_sym_closedir] = ACTIONS(3093),
    [anon_sym_caller] = ACTIONS(3093),
    [anon_sym_chomp] = ACTIONS(3093),
    [anon_sym_chr] = ACTIONS(3093),
    [anon_sym_cos] = ACTIONS(3093),
    [anon_sym_chroot] = ACTIONS(3093),
    [anon_sym_defined] = ACTIONS(3093),
    [anon_sym_delete] = ACTIONS(3093),
    [anon_sym_dbmclose] = ACTIONS(3093),
    [anon_sym_exists] = ACTIONS(3093),
    [anon_sym_exit] = ACTIONS(3093),
    [anon_sym_eof] = ACTIONS(3093),
    [anon_sym_exp] = ACTIONS(3093),
    [anon_sym_each] = ACTIONS(3093),
    [anon_sym_fc] = ACTIONS(3093),
    [anon_sym_fileno] = ACTIONS(3093),
    [anon_sym_gmtime] = ACTIONS(3093),
    [anon_sym_getc] = ACTIONS(3093),
    [anon_sym_getpgrp] = ACTIONS(3093),
    [anon_sym_getprotobyname] = ACTIONS(3093),
    [anon_sym_getpwname] = ACTIONS(3093),
    [anon_sym_getpwuid] = ACTIONS(3093),
    [anon_sym_getpeername] = ACTIONS(3093),
    [anon_sym_getnetbyname] = ACTIONS(3093),
    [anon_sym_getsockname] = ACTIONS(3093),
    [anon_sym_getgrnam] = ACTIONS(3093),
    [anon_sym_getgrgid] = ACTIONS(3093),
    [anon_sym_hex] = ACTIONS(3093),
    [anon_sym_int] = ACTIONS(3093),
    [anon_sym_keys] = ACTIONS(3093),
    [anon_sym_lc] = ACTIONS(3093),
    [anon_sym_lcfirst] = ACTIONS(3093),
    [anon_sym_length] = ACTIONS(3093),
    [anon_sym_localtime] = ACTIONS(3093),
    [anon_sym_log] = ACTIONS(3093),
    [anon_sym_lock] = ACTIONS(3093),
    [anon_sym_lstat] = ACTIONS(3093),
    [anon_sym_oct] = ACTIONS(3093),
    [anon_sym_ord] = ACTIONS(3093),
    [anon_sym_prototype] = ACTIONS(3093),
    [anon_sym_pop] = ACTIONS(3093),
    [anon_sym_pos] = ACTIONS(3093),
    [anon_sym_quotemeta] = ACTIONS(3093),
    [anon_sym_reset] = ACTIONS(3093),
    [anon_sym_rand] = ACTIONS(3093),
    [anon_sym_rmdir] = ACTIONS(3093),
    [anon_sym_readdir] = ACTIONS(3093),
    [anon_sym_readline] = ACTIONS(3093),
    [anon_sym_readpipe] = ACTIONS(3093),
    [anon_sym_rewinddir] = ACTIONS(3093),
    [anon_sym_readlink] = ACTIONS(3093),
    [anon_sym_ref] = ACTIONS(3093),
    [anon_sym_scalar] = ACTIONS(3093),
    [anon_sym_shift] = ACTIONS(3093),
    [anon_sym_sin] = ACTIONS(3093),
    [anon_sym_sleep] = ACTIONS(3093),
    [anon_sym_sqrt] = ACTIONS(3093),
    [anon_sym_srand] = ACTIONS(3093),
    [anon_sym_stat] = ACTIONS(3093),
    [anon_sym_study] = ACTIONS(3093),
    [anon_sym_tell] = ACTIONS(3093),
    [anon_sym_telldir] = ACTIONS(3093),
    [anon_sym_tied] = ACTIONS(3093),
    [anon_sym_uc] = ACTIONS(3093),
    [anon_sym_ucfirst] = ACTIONS(3093),
    [anon_sym_untie] = ACTIONS(3093),
    [anon_sym_umask] = ACTIONS(3093),
    [anon_sym_values] = ACTIONS(3093),
    [anon_sym_write] = ACTIONS(3093),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3093),
    [anon_sym___END__] = ACTIONS(3093),
    [anon_sym_EOT] = ACTIONS(3093),
    [anon_sym_q] = ACTIONS(3093),
    [anon_sym_qq] = ACTIONS(3093),
    [anon_sym_qw] = ACTIONS(3093),
    [anon_sym_qx] = ACTIONS(3093),
    [anon_sym_qr] = ACTIONS(3093),
    [anon_sym_m] = ACTIONS(3093),
    [anon_sym_s] = ACTIONS(3093),
    [anon_sym_tr] = ACTIONS(3093),
    [anon_sym_y] = ACTIONS(3093),
    [anon_sym_if] = ACTIONS(3093),
    [anon_sym_unless] = ACTIONS(3093),
    [anon_sym_while] = ACTIONS(3093),
    [anon_sym_until] = ACTIONS(3093),
    [aux_sym__bareword_token1] = ACTIONS(3093),
    [sym_number] = ACTIONS(3093),
    [anon_sym_true] = ACTIONS(3093),
    [anon_sym_false] = ACTIONS(3093),
    [sym__single_quote] = ACTIONS(3091),
    [sym__double_quote] = ACTIONS(3091),
    [sym__backtick_quote] = ACTIONS(3091),
    [sym__search_slash_quote] = ACTIONS(3091),
    [sym__open_readline_bracket] = ACTIONS(3091),
    [sym__open_fileglob_bracket] = ACTIONS(3091),
    [sym__PERLY_HEREDOC] = ACTIONS(3091),
    [sym__ctrl_z_hack] = ACTIONS(3091),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3091),
    [sym__filetest] = ACTIONS(3091),
  },
  [STATE(1485)] = {
    [sym_heredoc_content] = STATE(1485),
    [ts_builtin_sym_end] = ACTIONS(3095),
    [sym__identifier] = ACTIONS(3097),
    [anon_sym_LBRACE] = ACTIONS(3097),
    [anon_sym_SEMI] = ACTIONS(3097),
    [anon_sym_package] = ACTIONS(3097),
    [anon_sym_class] = ACTIONS(3097),
    [anon_sym_role] = ACTIONS(3097),
    [anon_sym_BUILD] = ACTIONS(3097),
    [anon_sym_ADJUST] = ACTIONS(3097),
    [anon_sym_DOLLAR] = ACTIONS(3097),
    [anon_sym_AT] = ACTIONS(3097),
    [anon_sym_my] = ACTIONS(3097),
    [anon_sym_extended] = ACTIONS(3097),
    [anon_sym_async] = ACTIONS(3097),
    [anon_sym_sub] = ACTIONS(3097),
    [anon_sym_method] = ACTIONS(3097),
    [anon_sym_LPAREN] = ACTIONS(3097),
    [anon_sym_continue] = ACTIONS(3097),
    [anon_sym_state] = ACTIONS(3097),
    [anon_sym_our] = ACTIONS(3097),
    [anon_sym_try] = ACTIONS(3097),
    [anon_sym_defer] = ACTIONS(3097),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3097),
    [anon_sym_LBRACK] = ACTIONS(3097),
    [anon_sym_LT_LT] = ACTIONS(3097),
    [anon_sym_SLASH_SLASH] = ACTIONS(3097),
    [anon_sym_PLUS] = ACTIONS(3097),
    [anon_sym_DASH] = ACTIONS(3097),
    [anon_sym_TILDE] = ACTIONS(3097),
    [anon_sym_BANG] = ACTIONS(3097),
    [anon_sym_PLUS_PLUS] = ACTIONS(3097),
    [anon_sym_DASH_DASH] = ACTIONS(3097),
    [anon_sym_BSLASH] = ACTIONS(3095),
    [anon_sym_do] = ACTIONS(3097),
    [anon_sym_eval] = ACTIONS(3097),
    [anon_sym_await] = ACTIONS(3097),
    [anon_sym_field] = ACTIONS(3097),
    [anon_sym_local] = ACTIONS(3097),
    [anon_sym_dynamically] = ACTIONS(3097),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3097),
    [anon_sym_require] = ACTIONS(3097),
    [anon_sym_map] = ACTIONS(3097),
    [anon_sym_grep] = ACTIONS(3097),
    [anon_sym_sort] = ACTIONS(3097),
    [anon_sym_goto] = ACTIONS(3097),
    [anon_sym_return] = ACTIONS(3097),
    [anon_sym_undef] = ACTIONS(3097),
    [anon_sym_PERCENT2] = ACTIONS(3097),
    [anon_sym_AMP2] = ACTIONS(3097),
    [anon_sym_STAR2] = ACTIONS(3097),
    [anon_sym_use] = ACTIONS(3097),
    [anon_sym_no] = ACTIONS(3097),
    [anon_sym_for] = ACTIONS(3097),
    [anon_sym_foreach] = ACTIONS(3097),
    [anon_sym_last] = ACTIONS(3097),
    [anon_sym_next] = ACTIONS(3097),
    [anon_sym_redo] = ACTIONS(3097),
    [anon_sym_BEGIN] = ACTIONS(3097),
    [anon_sym_INIT] = ACTIONS(3097),
    [anon_sym_CHECK] = ACTIONS(3097),
    [anon_sym_UNITCHECK] = ACTIONS(3097),
    [anon_sym_END] = ACTIONS(3097),
    [anon_sym___FILE__] = ACTIONS(3097),
    [anon_sym___LINE__] = ACTIONS(3097),
    [anon_sym___PACKAGE__] = ACTIONS(3097),
    [anon_sym___SUB__] = ACTIONS(3097),
    [anon_sym_break] = ACTIONS(3097),
    [anon_sym_fork] = ACTIONS(3097),
    [anon_sym_getppid] = ACTIONS(3097),
    [anon_sym_time] = ACTIONS(3097),
    [anon_sym_times] = ACTIONS(3097),
    [anon_sym_wait] = ACTIONS(3097),
    [anon_sym_wantarray] = ACTIONS(3097),
    [anon_sym_abs] = ACTIONS(3097),
    [anon_sym_alarm] = ACTIONS(3097),
    [anon_sym_chop] = ACTIONS(3097),
    [anon_sym_chdir] = ACTIONS(3097),
    [anon_sym_close] = ACTIONS(3097),
    [anon_sym_closedir] = ACTIONS(3097),
    [anon_sym_caller] = ACTIONS(3097),
    [anon_sym_chomp] = ACTIONS(3097),
    [anon_sym_chr] = ACTIONS(3097),
    [anon_sym_cos] = ACTIONS(3097),
    [anon_sym_chroot] = ACTIONS(3097),
    [anon_sym_defined] = ACTIONS(3097),
    [anon_sym_delete] = ACTIONS(3097),
    [anon_sym_dbmclose] = ACTIONS(3097),
    [anon_sym_exists] = ACTIONS(3097),
    [anon_sym_exit] = ACTIONS(3097),
    [anon_sym_eof] = ACTIONS(3097),
    [anon_sym_exp] = ACTIONS(3097),
    [anon_sym_each] = ACTIONS(3097),
    [anon_sym_fc] = ACTIONS(3097),
    [anon_sym_fileno] = ACTIONS(3097),
    [anon_sym_gmtime] = ACTIONS(3097),
    [anon_sym_getc] = ACTIONS(3097),
    [anon_sym_getpgrp] = ACTIONS(3097),
    [anon_sym_getprotobyname] = ACTIONS(3097),
    [anon_sym_getpwname] = ACTIONS(3097),
    [anon_sym_getpwuid] = ACTIONS(3097),
    [anon_sym_getpeername] = ACTIONS(3097),
    [anon_sym_getnetbyname] = ACTIONS(3097),
    [anon_sym_getsockname] = ACTIONS(3097),
    [anon_sym_getgrnam] = ACTIONS(3097),
    [anon_sym_getgrgid] = ACTIONS(3097),
    [anon_sym_hex] = ACTIONS(3097),
    [anon_sym_int] = ACTIONS(3097),
    [anon_sym_keys] = ACTIONS(3097),
    [anon_sym_lc] = ACTIONS(3097),
    [anon_sym_lcfirst] = ACTIONS(3097),
    [anon_sym_length] = ACTIONS(3097),
    [anon_sym_localtime] = ACTIONS(3097),
    [anon_sym_log] = ACTIONS(3097),
    [anon_sym_lock] = ACTIONS(3097),
    [anon_sym_lstat] = ACTIONS(3097),
    [anon_sym_oct] = ACTIONS(3097),
    [anon_sym_ord] = ACTIONS(3097),
    [anon_sym_prototype] = ACTIONS(3097),
    [anon_sym_pop] = ACTIONS(3097),
    [anon_sym_pos] = ACTIONS(3097),
    [anon_sym_quotemeta] = ACTIONS(3097),
    [anon_sym_reset] = ACTIONS(3097),
    [anon_sym_rand] = ACTIONS(3097),
    [anon_sym_rmdir] = ACTIONS(3097),
    [anon_sym_readdir] = ACTIONS(3097),
    [anon_sym_readline] = ACTIONS(3097),
    [anon_sym_readpipe] = ACTIONS(3097),
    [anon_sym_rewinddir] = ACTIONS(3097),
    [anon_sym_readlink] = ACTIONS(3097),
    [anon_sym_ref] = ACTIONS(3097),
    [anon_sym_scalar] = ACTIONS(3097),
    [anon_sym_shift] = ACTIONS(3097),
    [anon_sym_sin] = ACTIONS(3097),
    [anon_sym_sleep] = ACTIONS(3097),
    [anon_sym_sqrt] = ACTIONS(3097),
    [anon_sym_srand] = ACTIONS(3097),
    [anon_sym_stat] = ACTIONS(3097),
    [anon_sym_study] = ACTIONS(3097),
    [anon_sym_tell] = ACTIONS(3097),
    [anon_sym_telldir] = ACTIONS(3097),
    [anon_sym_tied] = ACTIONS(3097),
    [anon_sym_uc] = ACTIONS(3097),
    [anon_sym_ucfirst] = ACTIONS(3097),
    [anon_sym_untie] = ACTIONS(3097),
    [anon_sym_umask] = ACTIONS(3097),
    [anon_sym_values] = ACTIONS(3097),
    [anon_sym_write] = ACTIONS(3097),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3097),
    [anon_sym___END__] = ACTIONS(3097),
    [anon_sym_EOT] = ACTIONS(3097),
    [anon_sym_q] = ACTIONS(3097),
    [anon_sym_qq] = ACTIONS(3097),
    [anon_sym_qw] = ACTIONS(3097),
    [anon_sym_qx] = ACTIONS(3097),
    [anon_sym_qr] = ACTIONS(3097),
    [anon_sym_m] = ACTIONS(3097),
    [anon_sym_s] = ACTIONS(3097),
    [anon_sym_tr] = ACTIONS(3097),
    [anon_sym_y] = ACTIONS(3097),
    [anon_sym_if] = ACTIONS(3097),
    [anon_sym_unless] = ACTIONS(3097),
    [anon_sym_while] = ACTIONS(3097),
    [anon_sym_until] = ACTIONS(3097),
    [aux_sym__bareword_token1] = ACTIONS(3097),
    [sym_number] = ACTIONS(3097),
    [anon_sym_true] = ACTIONS(3097),
    [anon_sym_false] = ACTIONS(3097),
    [sym__single_quote] = ACTIONS(3095),
    [sym__double_quote] = ACTIONS(3095),
    [sym__backtick_quote] = ACTIONS(3095),
    [sym__search_slash_quote] = ACTIONS(3095),
    [sym__open_readline_bracket] = ACTIONS(3095),
    [sym__open_fileglob_bracket] = ACTIONS(3095),
    [sym__PERLY_HEREDOC] = ACTIONS(3095),
    [sym__ctrl_z_hack] = ACTIONS(3095),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3095),
    [sym__filetest] = ACTIONS(3095),
  },
  [STATE(1486)] = {
    [sym_heredoc_content] = STATE(1486),
    [ts_builtin_sym_end] = ACTIONS(3099),
    [sym__identifier] = ACTIONS(3101),
    [anon_sym_LBRACE] = ACTIONS(3101),
    [anon_sym_SEMI] = ACTIONS(3101),
    [anon_sym_package] = ACTIONS(3101),
    [anon_sym_class] = ACTIONS(3101),
    [anon_sym_role] = ACTIONS(3101),
    [anon_sym_BUILD] = ACTIONS(3101),
    [anon_sym_ADJUST] = ACTIONS(3101),
    [anon_sym_DOLLAR] = ACTIONS(3101),
    [anon_sym_AT] = ACTIONS(3101),
    [anon_sym_my] = ACTIONS(3101),
    [anon_sym_extended] = ACTIONS(3101),
    [anon_sym_async] = ACTIONS(3101),
    [anon_sym_sub] = ACTIONS(3101),
    [anon_sym_method] = ACTIONS(3101),
    [anon_sym_LPAREN] = ACTIONS(3101),
    [anon_sym_continue] = ACTIONS(3101),
    [anon_sym_state] = ACTIONS(3101),
    [anon_sym_our] = ACTIONS(3101),
    [anon_sym_try] = ACTIONS(3101),
    [anon_sym_defer] = ACTIONS(3101),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3101),
    [anon_sym_LBRACK] = ACTIONS(3101),
    [anon_sym_LT_LT] = ACTIONS(3101),
    [anon_sym_SLASH_SLASH] = ACTIONS(3101),
    [anon_sym_PLUS] = ACTIONS(3101),
    [anon_sym_DASH] = ACTIONS(3101),
    [anon_sym_TILDE] = ACTIONS(3101),
    [anon_sym_BANG] = ACTIONS(3101),
    [anon_sym_PLUS_PLUS] = ACTIONS(3101),
    [anon_sym_DASH_DASH] = ACTIONS(3101),
    [anon_sym_BSLASH] = ACTIONS(3099),
    [anon_sym_do] = ACTIONS(3101),
    [anon_sym_eval] = ACTIONS(3101),
    [anon_sym_await] = ACTIONS(3101),
    [anon_sym_field] = ACTIONS(3101),
    [anon_sym_local] = ACTIONS(3101),
    [anon_sym_dynamically] = ACTIONS(3101),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3101),
    [anon_sym_require] = ACTIONS(3101),
    [anon_sym_map] = ACTIONS(3101),
    [anon_sym_grep] = ACTIONS(3101),
    [anon_sym_sort] = ACTIONS(3101),
    [anon_sym_goto] = ACTIONS(3101),
    [anon_sym_return] = ACTIONS(3101),
    [anon_sym_undef] = ACTIONS(3101),
    [anon_sym_PERCENT2] = ACTIONS(3101),
    [anon_sym_AMP2] = ACTIONS(3101),
    [anon_sym_STAR2] = ACTIONS(3101),
    [anon_sym_use] = ACTIONS(3101),
    [anon_sym_no] = ACTIONS(3101),
    [anon_sym_for] = ACTIONS(3101),
    [anon_sym_foreach] = ACTIONS(3101),
    [anon_sym_last] = ACTIONS(3101),
    [anon_sym_next] = ACTIONS(3101),
    [anon_sym_redo] = ACTIONS(3101),
    [anon_sym_BEGIN] = ACTIONS(3101),
    [anon_sym_INIT] = ACTIONS(3101),
    [anon_sym_CHECK] = ACTIONS(3101),
    [anon_sym_UNITCHECK] = ACTIONS(3101),
    [anon_sym_END] = ACTIONS(3101),
    [anon_sym___FILE__] = ACTIONS(3101),
    [anon_sym___LINE__] = ACTIONS(3101),
    [anon_sym___PACKAGE__] = ACTIONS(3101),
    [anon_sym___SUB__] = ACTIONS(3101),
    [anon_sym_break] = ACTIONS(3101),
    [anon_sym_fork] = ACTIONS(3101),
    [anon_sym_getppid] = ACTIONS(3101),
    [anon_sym_time] = ACTIONS(3101),
    [anon_sym_times] = ACTIONS(3101),
    [anon_sym_wait] = ACTIONS(3101),
    [anon_sym_wantarray] = ACTIONS(3101),
    [anon_sym_abs] = ACTIONS(3101),
    [anon_sym_alarm] = ACTIONS(3101),
    [anon_sym_chop] = ACTIONS(3101),
    [anon_sym_chdir] = ACTIONS(3101),
    [anon_sym_close] = ACTIONS(3101),
    [anon_sym_closedir] = ACTIONS(3101),
    [anon_sym_caller] = ACTIONS(3101),
    [anon_sym_chomp] = ACTIONS(3101),
    [anon_sym_chr] = ACTIONS(3101),
    [anon_sym_cos] = ACTIONS(3101),
    [anon_sym_chroot] = ACTIONS(3101),
    [anon_sym_defined] = ACTIONS(3101),
    [anon_sym_delete] = ACTIONS(3101),
    [anon_sym_dbmclose] = ACTIONS(3101),
    [anon_sym_exists] = ACTIONS(3101),
    [anon_sym_exit] = ACTIONS(3101),
    [anon_sym_eof] = ACTIONS(3101),
    [anon_sym_exp] = ACTIONS(3101),
    [anon_sym_each] = ACTIONS(3101),
    [anon_sym_fc] = ACTIONS(3101),
    [anon_sym_fileno] = ACTIONS(3101),
    [anon_sym_gmtime] = ACTIONS(3101),
    [anon_sym_getc] = ACTIONS(3101),
    [anon_sym_getpgrp] = ACTIONS(3101),
    [anon_sym_getprotobyname] = ACTIONS(3101),
    [anon_sym_getpwname] = ACTIONS(3101),
    [anon_sym_getpwuid] = ACTIONS(3101),
    [anon_sym_getpeername] = ACTIONS(3101),
    [anon_sym_getnetbyname] = ACTIONS(3101),
    [anon_sym_getsockname] = ACTIONS(3101),
    [anon_sym_getgrnam] = ACTIONS(3101),
    [anon_sym_getgrgid] = ACTIONS(3101),
    [anon_sym_hex] = ACTIONS(3101),
    [anon_sym_int] = ACTIONS(3101),
    [anon_sym_keys] = ACTIONS(3101),
    [anon_sym_lc] = ACTIONS(3101),
    [anon_sym_lcfirst] = ACTIONS(3101),
    [anon_sym_length] = ACTIONS(3101),
    [anon_sym_localtime] = ACTIONS(3101),
    [anon_sym_log] = ACTIONS(3101),
    [anon_sym_lock] = ACTIONS(3101),
    [anon_sym_lstat] = ACTIONS(3101),
    [anon_sym_oct] = ACTIONS(3101),
    [anon_sym_ord] = ACTIONS(3101),
    [anon_sym_prototype] = ACTIONS(3101),
    [anon_sym_pop] = ACTIONS(3101),
    [anon_sym_pos] = ACTIONS(3101),
    [anon_sym_quotemeta] = ACTIONS(3101),
    [anon_sym_reset] = ACTIONS(3101),
    [anon_sym_rand] = ACTIONS(3101),
    [anon_sym_rmdir] = ACTIONS(3101),
    [anon_sym_readdir] = ACTIONS(3101),
    [anon_sym_readline] = ACTIONS(3101),
    [anon_sym_readpipe] = ACTIONS(3101),
    [anon_sym_rewinddir] = ACTIONS(3101),
    [anon_sym_readlink] = ACTIONS(3101),
    [anon_sym_ref] = ACTIONS(3101),
    [anon_sym_scalar] = ACTIONS(3101),
    [anon_sym_shift] = ACTIONS(3101),
    [anon_sym_sin] = ACTIONS(3101),
    [anon_sym_sleep] = ACTIONS(3101),
    [anon_sym_sqrt] = ACTIONS(3101),
    [anon_sym_srand] = ACTIONS(3101),
    [anon_sym_stat] = ACTIONS(3101),
    [anon_sym_study] = ACTIONS(3101),
    [anon_sym_tell] = ACTIONS(3101),
    [anon_sym_telldir] = ACTIONS(3101),
    [anon_sym_tied] = ACTIONS(3101),
    [anon_sym_uc] = ACTIONS(3101),
    [anon_sym_ucfirst] = ACTIONS(3101),
    [anon_sym_untie] = ACTIONS(3101),
    [anon_sym_umask] = ACTIONS(3101),
    [anon_sym_values] = ACTIONS(3101),
    [anon_sym_write] = ACTIONS(3101),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3101),
    [anon_sym___END__] = ACTIONS(3101),
    [anon_sym_EOT] = ACTIONS(3101),
    [anon_sym_q] = ACTIONS(3101),
    [anon_sym_qq] = ACTIONS(3101),
    [anon_sym_qw] = ACTIONS(3101),
    [anon_sym_qx] = ACTIONS(3101),
    [anon_sym_qr] = ACTIONS(3101),
    [anon_sym_m] = ACTIONS(3101),
    [anon_sym_s] = ACTIONS(3101),
    [anon_sym_tr] = ACTIONS(3101),
    [anon_sym_y] = ACTIONS(3101),
    [anon_sym_if] = ACTIONS(3101),
    [anon_sym_unless] = ACTIONS(3101),
    [anon_sym_while] = ACTIONS(3101),
    [anon_sym_until] = ACTIONS(3101),
    [aux_sym__bareword_token1] = ACTIONS(3101),
    [sym_number] = ACTIONS(3101),
    [anon_sym_true] = ACTIONS(3101),
    [anon_sym_false] = ACTIONS(3101),
    [sym__single_quote] = ACTIONS(3099),
    [sym__double_quote] = ACTIONS(3099),
    [sym__backtick_quote] = ACTIONS(3099),
    [sym__search_slash_quote] = ACTIONS(3099),
    [sym__open_readline_bracket] = ACTIONS(3099),
    [sym__open_fileglob_bracket] = ACTIONS(3099),
    [sym__PERLY_HEREDOC] = ACTIONS(3099),
    [sym__ctrl_z_hack] = ACTIONS(3099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3099),
    [sym__filetest] = ACTIONS(3099),
  },
  [STATE(1487)] = {
    [sym_heredoc_content] = STATE(1487),
    [ts_builtin_sym_end] = ACTIONS(3103),
    [sym__identifier] = ACTIONS(3105),
    [anon_sym_LBRACE] = ACTIONS(3105),
    [anon_sym_SEMI] = ACTIONS(3105),
    [anon_sym_package] = ACTIONS(3105),
    [anon_sym_class] = ACTIONS(3105),
    [anon_sym_role] = ACTIONS(3105),
    [anon_sym_BUILD] = ACTIONS(3105),
    [anon_sym_ADJUST] = ACTIONS(3105),
    [anon_sym_DOLLAR] = ACTIONS(3105),
    [anon_sym_AT] = ACTIONS(3105),
    [anon_sym_my] = ACTIONS(3105),
    [anon_sym_extended] = ACTIONS(3105),
    [anon_sym_async] = ACTIONS(3105),
    [anon_sym_sub] = ACTIONS(3105),
    [anon_sym_method] = ACTIONS(3105),
    [anon_sym_LPAREN] = ACTIONS(3105),
    [anon_sym_continue] = ACTIONS(3105),
    [anon_sym_state] = ACTIONS(3105),
    [anon_sym_our] = ACTIONS(3105),
    [anon_sym_try] = ACTIONS(3105),
    [anon_sym_defer] = ACTIONS(3105),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3105),
    [anon_sym_LBRACK] = ACTIONS(3105),
    [anon_sym_LT_LT] = ACTIONS(3105),
    [anon_sym_SLASH_SLASH] = ACTIONS(3105),
    [anon_sym_PLUS] = ACTIONS(3105),
    [anon_sym_DASH] = ACTIONS(3105),
    [anon_sym_TILDE] = ACTIONS(3105),
    [anon_sym_BANG] = ACTIONS(3105),
    [anon_sym_PLUS_PLUS] = ACTIONS(3105),
    [anon_sym_DASH_DASH] = ACTIONS(3105),
    [anon_sym_BSLASH] = ACTIONS(3103),
    [anon_sym_do] = ACTIONS(3105),
    [anon_sym_eval] = ACTIONS(3105),
    [anon_sym_await] = ACTIONS(3105),
    [anon_sym_field] = ACTIONS(3105),
    [anon_sym_local] = ACTIONS(3105),
    [anon_sym_dynamically] = ACTIONS(3105),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3105),
    [anon_sym_require] = ACTIONS(3105),
    [anon_sym_map] = ACTIONS(3105),
    [anon_sym_grep] = ACTIONS(3105),
    [anon_sym_sort] = ACTIONS(3105),
    [anon_sym_goto] = ACTIONS(3105),
    [anon_sym_return] = ACTIONS(3105),
    [anon_sym_undef] = ACTIONS(3105),
    [anon_sym_PERCENT2] = ACTIONS(3105),
    [anon_sym_AMP2] = ACTIONS(3105),
    [anon_sym_STAR2] = ACTIONS(3105),
    [anon_sym_use] = ACTIONS(3105),
    [anon_sym_no] = ACTIONS(3105),
    [anon_sym_for] = ACTIONS(3105),
    [anon_sym_foreach] = ACTIONS(3105),
    [anon_sym_last] = ACTIONS(3105),
    [anon_sym_next] = ACTIONS(3105),
    [anon_sym_redo] = ACTIONS(3105),
    [anon_sym_BEGIN] = ACTIONS(3105),
    [anon_sym_INIT] = ACTIONS(3105),
    [anon_sym_CHECK] = ACTIONS(3105),
    [anon_sym_UNITCHECK] = ACTIONS(3105),
    [anon_sym_END] = ACTIONS(3105),
    [anon_sym___FILE__] = ACTIONS(3105),
    [anon_sym___LINE__] = ACTIONS(3105),
    [anon_sym___PACKAGE__] = ACTIONS(3105),
    [anon_sym___SUB__] = ACTIONS(3105),
    [anon_sym_break] = ACTIONS(3105),
    [anon_sym_fork] = ACTIONS(3105),
    [anon_sym_getppid] = ACTIONS(3105),
    [anon_sym_time] = ACTIONS(3105),
    [anon_sym_times] = ACTIONS(3105),
    [anon_sym_wait] = ACTIONS(3105),
    [anon_sym_wantarray] = ACTIONS(3105),
    [anon_sym_abs] = ACTIONS(3105),
    [anon_sym_alarm] = ACTIONS(3105),
    [anon_sym_chop] = ACTIONS(3105),
    [anon_sym_chdir] = ACTIONS(3105),
    [anon_sym_close] = ACTIONS(3105),
    [anon_sym_closedir] = ACTIONS(3105),
    [anon_sym_caller] = ACTIONS(3105),
    [anon_sym_chomp] = ACTIONS(3105),
    [anon_sym_chr] = ACTIONS(3105),
    [anon_sym_cos] = ACTIONS(3105),
    [anon_sym_chroot] = ACTIONS(3105),
    [anon_sym_defined] = ACTIONS(3105),
    [anon_sym_delete] = ACTIONS(3105),
    [anon_sym_dbmclose] = ACTIONS(3105),
    [anon_sym_exists] = ACTIONS(3105),
    [anon_sym_exit] = ACTIONS(3105),
    [anon_sym_eof] = ACTIONS(3105),
    [anon_sym_exp] = ACTIONS(3105),
    [anon_sym_each] = ACTIONS(3105),
    [anon_sym_fc] = ACTIONS(3105),
    [anon_sym_fileno] = ACTIONS(3105),
    [anon_sym_gmtime] = ACTIONS(3105),
    [anon_sym_getc] = ACTIONS(3105),
    [anon_sym_getpgrp] = ACTIONS(3105),
    [anon_sym_getprotobyname] = ACTIONS(3105),
    [anon_sym_getpwname] = ACTIONS(3105),
    [anon_sym_getpwuid] = ACTIONS(3105),
    [anon_sym_getpeername] = ACTIONS(3105),
    [anon_sym_getnetbyname] = ACTIONS(3105),
    [anon_sym_getsockname] = ACTIONS(3105),
    [anon_sym_getgrnam] = ACTIONS(3105),
    [anon_sym_getgrgid] = ACTIONS(3105),
    [anon_sym_hex] = ACTIONS(3105),
    [anon_sym_int] = ACTIONS(3105),
    [anon_sym_keys] = ACTIONS(3105),
    [anon_sym_lc] = ACTIONS(3105),
    [anon_sym_lcfirst] = ACTIONS(3105),
    [anon_sym_length] = ACTIONS(3105),
    [anon_sym_localtime] = ACTIONS(3105),
    [anon_sym_log] = ACTIONS(3105),
    [anon_sym_lock] = ACTIONS(3105),
    [anon_sym_lstat] = ACTIONS(3105),
    [anon_sym_oct] = ACTIONS(3105),
    [anon_sym_ord] = ACTIONS(3105),
    [anon_sym_prototype] = ACTIONS(3105),
    [anon_sym_pop] = ACTIONS(3105),
    [anon_sym_pos] = ACTIONS(3105),
    [anon_sym_quotemeta] = ACTIONS(3105),
    [anon_sym_reset] = ACTIONS(3105),
    [anon_sym_rand] = ACTIONS(3105),
    [anon_sym_rmdir] = ACTIONS(3105),
    [anon_sym_readdir] = ACTIONS(3105),
    [anon_sym_readline] = ACTIONS(3105),
    [anon_sym_readpipe] = ACTIONS(3105),
    [anon_sym_rewinddir] = ACTIONS(3105),
    [anon_sym_readlink] = ACTIONS(3105),
    [anon_sym_ref] = ACTIONS(3105),
    [anon_sym_scalar] = ACTIONS(3105),
    [anon_sym_shift] = ACTIONS(3105),
    [anon_sym_sin] = ACTIONS(3105),
    [anon_sym_sleep] = ACTIONS(3105),
    [anon_sym_sqrt] = ACTIONS(3105),
    [anon_sym_srand] = ACTIONS(3105),
    [anon_sym_stat] = ACTIONS(3105),
    [anon_sym_study] = ACTIONS(3105),
    [anon_sym_tell] = ACTIONS(3105),
    [anon_sym_telldir] = ACTIONS(3105),
    [anon_sym_tied] = ACTIONS(3105),
    [anon_sym_uc] = ACTIONS(3105),
    [anon_sym_ucfirst] = ACTIONS(3105),
    [anon_sym_untie] = ACTIONS(3105),
    [anon_sym_umask] = ACTIONS(3105),
    [anon_sym_values] = ACTIONS(3105),
    [anon_sym_write] = ACTIONS(3105),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3105),
    [anon_sym___END__] = ACTIONS(3105),
    [anon_sym_EOT] = ACTIONS(3105),
    [anon_sym_q] = ACTIONS(3105),
    [anon_sym_qq] = ACTIONS(3105),
    [anon_sym_qw] = ACTIONS(3105),
    [anon_sym_qx] = ACTIONS(3105),
    [anon_sym_qr] = ACTIONS(3105),
    [anon_sym_m] = ACTIONS(3105),
    [anon_sym_s] = ACTIONS(3105),
    [anon_sym_tr] = ACTIONS(3105),
    [anon_sym_y] = ACTIONS(3105),
    [anon_sym_if] = ACTIONS(3105),
    [anon_sym_unless] = ACTIONS(3105),
    [anon_sym_while] = ACTIONS(3105),
    [anon_sym_until] = ACTIONS(3105),
    [aux_sym__bareword_token1] = ACTIONS(3105),
    [sym_number] = ACTIONS(3105),
    [anon_sym_true] = ACTIONS(3105),
    [anon_sym_false] = ACTIONS(3105),
    [sym__single_quote] = ACTIONS(3103),
    [sym__double_quote] = ACTIONS(3103),
    [sym__backtick_quote] = ACTIONS(3103),
    [sym__search_slash_quote] = ACTIONS(3103),
    [sym__open_readline_bracket] = ACTIONS(3103),
    [sym__open_fileglob_bracket] = ACTIONS(3103),
    [sym__PERLY_HEREDOC] = ACTIONS(3103),
    [sym__ctrl_z_hack] = ACTIONS(3103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3103),
    [sym__filetest] = ACTIONS(3103),
  },
  [STATE(1488)] = {
    [sym_heredoc_content] = STATE(1488),
    [ts_builtin_sym_end] = ACTIONS(3107),
    [sym__identifier] = ACTIONS(3109),
    [anon_sym_LBRACE] = ACTIONS(3109),
    [anon_sym_SEMI] = ACTIONS(3109),
    [anon_sym_package] = ACTIONS(3109),
    [anon_sym_class] = ACTIONS(3109),
    [anon_sym_role] = ACTIONS(3109),
    [anon_sym_BUILD] = ACTIONS(3109),
    [anon_sym_ADJUST] = ACTIONS(3109),
    [anon_sym_DOLLAR] = ACTIONS(3109),
    [anon_sym_AT] = ACTIONS(3109),
    [anon_sym_my] = ACTIONS(3109),
    [anon_sym_extended] = ACTIONS(3109),
    [anon_sym_async] = ACTIONS(3109),
    [anon_sym_sub] = ACTIONS(3109),
    [anon_sym_method] = ACTIONS(3109),
    [anon_sym_LPAREN] = ACTIONS(3109),
    [anon_sym_continue] = ACTIONS(3109),
    [anon_sym_state] = ACTIONS(3109),
    [anon_sym_our] = ACTIONS(3109),
    [anon_sym_try] = ACTIONS(3109),
    [anon_sym_defer] = ACTIONS(3109),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3109),
    [anon_sym_LBRACK] = ACTIONS(3109),
    [anon_sym_LT_LT] = ACTIONS(3109),
    [anon_sym_SLASH_SLASH] = ACTIONS(3109),
    [anon_sym_PLUS] = ACTIONS(3109),
    [anon_sym_DASH] = ACTIONS(3109),
    [anon_sym_TILDE] = ACTIONS(3109),
    [anon_sym_BANG] = ACTIONS(3109),
    [anon_sym_PLUS_PLUS] = ACTIONS(3109),
    [anon_sym_DASH_DASH] = ACTIONS(3109),
    [anon_sym_BSLASH] = ACTIONS(3107),
    [anon_sym_do] = ACTIONS(3109),
    [anon_sym_eval] = ACTIONS(3109),
    [anon_sym_await] = ACTIONS(3109),
    [anon_sym_field] = ACTIONS(3109),
    [anon_sym_local] = ACTIONS(3109),
    [anon_sym_dynamically] = ACTIONS(3109),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3109),
    [anon_sym_require] = ACTIONS(3109),
    [anon_sym_map] = ACTIONS(3109),
    [anon_sym_grep] = ACTIONS(3109),
    [anon_sym_sort] = ACTIONS(3109),
    [anon_sym_goto] = ACTIONS(3109),
    [anon_sym_return] = ACTIONS(3109),
    [anon_sym_undef] = ACTIONS(3109),
    [anon_sym_PERCENT2] = ACTIONS(3109),
    [anon_sym_AMP2] = ACTIONS(3109),
    [anon_sym_STAR2] = ACTIONS(3109),
    [anon_sym_use] = ACTIONS(3109),
    [anon_sym_no] = ACTIONS(3109),
    [anon_sym_for] = ACTIONS(3109),
    [anon_sym_foreach] = ACTIONS(3109),
    [anon_sym_last] = ACTIONS(3109),
    [anon_sym_next] = ACTIONS(3109),
    [anon_sym_redo] = ACTIONS(3109),
    [anon_sym_BEGIN] = ACTIONS(3109),
    [anon_sym_INIT] = ACTIONS(3109),
    [anon_sym_CHECK] = ACTIONS(3109),
    [anon_sym_UNITCHECK] = ACTIONS(3109),
    [anon_sym_END] = ACTIONS(3109),
    [anon_sym___FILE__] = ACTIONS(3109),
    [anon_sym___LINE__] = ACTIONS(3109),
    [anon_sym___PACKAGE__] = ACTIONS(3109),
    [anon_sym___SUB__] = ACTIONS(3109),
    [anon_sym_break] = ACTIONS(3109),
    [anon_sym_fork] = ACTIONS(3109),
    [anon_sym_getppid] = ACTIONS(3109),
    [anon_sym_time] = ACTIONS(3109),
    [anon_sym_times] = ACTIONS(3109),
    [anon_sym_wait] = ACTIONS(3109),
    [anon_sym_wantarray] = ACTIONS(3109),
    [anon_sym_abs] = ACTIONS(3109),
    [anon_sym_alarm] = ACTIONS(3109),
    [anon_sym_chop] = ACTIONS(3109),
    [anon_sym_chdir] = ACTIONS(3109),
    [anon_sym_close] = ACTIONS(3109),
    [anon_sym_closedir] = ACTIONS(3109),
    [anon_sym_caller] = ACTIONS(3109),
    [anon_sym_chomp] = ACTIONS(3109),
    [anon_sym_chr] = ACTIONS(3109),
    [anon_sym_cos] = ACTIONS(3109),
    [anon_sym_chroot] = ACTIONS(3109),
    [anon_sym_defined] = ACTIONS(3109),
    [anon_sym_delete] = ACTIONS(3109),
    [anon_sym_dbmclose] = ACTIONS(3109),
    [anon_sym_exists] = ACTIONS(3109),
    [anon_sym_exit] = ACTIONS(3109),
    [anon_sym_eof] = ACTIONS(3109),
    [anon_sym_exp] = ACTIONS(3109),
    [anon_sym_each] = ACTIONS(3109),
    [anon_sym_fc] = ACTIONS(3109),
    [anon_sym_fileno] = ACTIONS(3109),
    [anon_sym_gmtime] = ACTIONS(3109),
    [anon_sym_getc] = ACTIONS(3109),
    [anon_sym_getpgrp] = ACTIONS(3109),
    [anon_sym_getprotobyname] = ACTIONS(3109),
    [anon_sym_getpwname] = ACTIONS(3109),
    [anon_sym_getpwuid] = ACTIONS(3109),
    [anon_sym_getpeername] = ACTIONS(3109),
    [anon_sym_getnetbyname] = ACTIONS(3109),
    [anon_sym_getsockname] = ACTIONS(3109),
    [anon_sym_getgrnam] = ACTIONS(3109),
    [anon_sym_getgrgid] = ACTIONS(3109),
    [anon_sym_hex] = ACTIONS(3109),
    [anon_sym_int] = ACTIONS(3109),
    [anon_sym_keys] = ACTIONS(3109),
    [anon_sym_lc] = ACTIONS(3109),
    [anon_sym_lcfirst] = ACTIONS(3109),
    [anon_sym_length] = ACTIONS(3109),
    [anon_sym_localtime] = ACTIONS(3109),
    [anon_sym_log] = ACTIONS(3109),
    [anon_sym_lock] = ACTIONS(3109),
    [anon_sym_lstat] = ACTIONS(3109),
    [anon_sym_oct] = ACTIONS(3109),
    [anon_sym_ord] = ACTIONS(3109),
    [anon_sym_prototype] = ACTIONS(3109),
    [anon_sym_pop] = ACTIONS(3109),
    [anon_sym_pos] = ACTIONS(3109),
    [anon_sym_quotemeta] = ACTIONS(3109),
    [anon_sym_reset] = ACTIONS(3109),
    [anon_sym_rand] = ACTIONS(3109),
    [anon_sym_rmdir] = ACTIONS(3109),
    [anon_sym_readdir] = ACTIONS(3109),
    [anon_sym_readline] = ACTIONS(3109),
    [anon_sym_readpipe] = ACTIONS(3109),
    [anon_sym_rewinddir] = ACTIONS(3109),
    [anon_sym_readlink] = ACTIONS(3109),
    [anon_sym_ref] = ACTIONS(3109),
    [anon_sym_scalar] = ACTIONS(3109),
    [anon_sym_shift] = ACTIONS(3109),
    [anon_sym_sin] = ACTIONS(3109),
    [anon_sym_sleep] = ACTIONS(3109),
    [anon_sym_sqrt] = ACTIONS(3109),
    [anon_sym_srand] = ACTIONS(3109),
    [anon_sym_stat] = ACTIONS(3109),
    [anon_sym_study] = ACTIONS(3109),
    [anon_sym_tell] = ACTIONS(3109),
    [anon_sym_telldir] = ACTIONS(3109),
    [anon_sym_tied] = ACTIONS(3109),
    [anon_sym_uc] = ACTIONS(3109),
    [anon_sym_ucfirst] = ACTIONS(3109),
    [anon_sym_untie] = ACTIONS(3109),
    [anon_sym_umask] = ACTIONS(3109),
    [anon_sym_values] = ACTIONS(3109),
    [anon_sym_write] = ACTIONS(3109),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3109),
    [anon_sym___END__] = ACTIONS(3109),
    [anon_sym_EOT] = ACTIONS(3109),
    [anon_sym_q] = ACTIONS(3109),
    [anon_sym_qq] = ACTIONS(3109),
    [anon_sym_qw] = ACTIONS(3109),
    [anon_sym_qx] = ACTIONS(3109),
    [anon_sym_qr] = ACTIONS(3109),
    [anon_sym_m] = ACTIONS(3109),
    [anon_sym_s] = ACTIONS(3109),
    [anon_sym_tr] = ACTIONS(3109),
    [anon_sym_y] = ACTIONS(3109),
    [anon_sym_if] = ACTIONS(3109),
    [anon_sym_unless] = ACTIONS(3109),
    [anon_sym_while] = ACTIONS(3109),
    [anon_sym_until] = ACTIONS(3109),
    [aux_sym__bareword_token1] = ACTIONS(3109),
    [sym_number] = ACTIONS(3109),
    [anon_sym_true] = ACTIONS(3109),
    [anon_sym_false] = ACTIONS(3109),
    [sym__single_quote] = ACTIONS(3107),
    [sym__double_quote] = ACTIONS(3107),
    [sym__backtick_quote] = ACTIONS(3107),
    [sym__search_slash_quote] = ACTIONS(3107),
    [sym__open_readline_bracket] = ACTIONS(3107),
    [sym__open_fileglob_bracket] = ACTIONS(3107),
    [sym__PERLY_HEREDOC] = ACTIONS(3107),
    [sym__ctrl_z_hack] = ACTIONS(3107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3107),
    [sym__filetest] = ACTIONS(3107),
  },
  [STATE(1489)] = {
    [sym_heredoc_content] = STATE(1489),
    [ts_builtin_sym_end] = ACTIONS(3111),
    [sym__identifier] = ACTIONS(3113),
    [anon_sym_LBRACE] = ACTIONS(3113),
    [anon_sym_SEMI] = ACTIONS(3113),
    [anon_sym_package] = ACTIONS(3113),
    [anon_sym_class] = ACTIONS(3113),
    [anon_sym_role] = ACTIONS(3113),
    [anon_sym_BUILD] = ACTIONS(3113),
    [anon_sym_ADJUST] = ACTIONS(3113),
    [anon_sym_DOLLAR] = ACTIONS(3113),
    [anon_sym_AT] = ACTIONS(3113),
    [anon_sym_my] = ACTIONS(3113),
    [anon_sym_extended] = ACTIONS(3113),
    [anon_sym_async] = ACTIONS(3113),
    [anon_sym_sub] = ACTIONS(3113),
    [anon_sym_method] = ACTIONS(3113),
    [anon_sym_LPAREN] = ACTIONS(3113),
    [anon_sym_continue] = ACTIONS(3113),
    [anon_sym_state] = ACTIONS(3113),
    [anon_sym_our] = ACTIONS(3113),
    [anon_sym_try] = ACTIONS(3113),
    [anon_sym_defer] = ACTIONS(3113),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3113),
    [anon_sym_LBRACK] = ACTIONS(3113),
    [anon_sym_LT_LT] = ACTIONS(3113),
    [anon_sym_SLASH_SLASH] = ACTIONS(3113),
    [anon_sym_PLUS] = ACTIONS(3113),
    [anon_sym_DASH] = ACTIONS(3113),
    [anon_sym_TILDE] = ACTIONS(3113),
    [anon_sym_BANG] = ACTIONS(3113),
    [anon_sym_PLUS_PLUS] = ACTIONS(3113),
    [anon_sym_DASH_DASH] = ACTIONS(3113),
    [anon_sym_BSLASH] = ACTIONS(3111),
    [anon_sym_do] = ACTIONS(3113),
    [anon_sym_eval] = ACTIONS(3113),
    [anon_sym_await] = ACTIONS(3113),
    [anon_sym_field] = ACTIONS(3113),
    [anon_sym_local] = ACTIONS(3113),
    [anon_sym_dynamically] = ACTIONS(3113),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3113),
    [anon_sym_require] = ACTIONS(3113),
    [anon_sym_map] = ACTIONS(3113),
    [anon_sym_grep] = ACTIONS(3113),
    [anon_sym_sort] = ACTIONS(3113),
    [anon_sym_goto] = ACTIONS(3113),
    [anon_sym_return] = ACTIONS(3113),
    [anon_sym_undef] = ACTIONS(3113),
    [anon_sym_PERCENT2] = ACTIONS(3113),
    [anon_sym_AMP2] = ACTIONS(3113),
    [anon_sym_STAR2] = ACTIONS(3113),
    [anon_sym_use] = ACTIONS(3113),
    [anon_sym_no] = ACTIONS(3113),
    [anon_sym_for] = ACTIONS(3113),
    [anon_sym_foreach] = ACTIONS(3113),
    [anon_sym_last] = ACTIONS(3113),
    [anon_sym_next] = ACTIONS(3113),
    [anon_sym_redo] = ACTIONS(3113),
    [anon_sym_BEGIN] = ACTIONS(3113),
    [anon_sym_INIT] = ACTIONS(3113),
    [anon_sym_CHECK] = ACTIONS(3113),
    [anon_sym_UNITCHECK] = ACTIONS(3113),
    [anon_sym_END] = ACTIONS(3113),
    [anon_sym___FILE__] = ACTIONS(3113),
    [anon_sym___LINE__] = ACTIONS(3113),
    [anon_sym___PACKAGE__] = ACTIONS(3113),
    [anon_sym___SUB__] = ACTIONS(3113),
    [anon_sym_break] = ACTIONS(3113),
    [anon_sym_fork] = ACTIONS(3113),
    [anon_sym_getppid] = ACTIONS(3113),
    [anon_sym_time] = ACTIONS(3113),
    [anon_sym_times] = ACTIONS(3113),
    [anon_sym_wait] = ACTIONS(3113),
    [anon_sym_wantarray] = ACTIONS(3113),
    [anon_sym_abs] = ACTIONS(3113),
    [anon_sym_alarm] = ACTIONS(3113),
    [anon_sym_chop] = ACTIONS(3113),
    [anon_sym_chdir] = ACTIONS(3113),
    [anon_sym_close] = ACTIONS(3113),
    [anon_sym_closedir] = ACTIONS(3113),
    [anon_sym_caller] = ACTIONS(3113),
    [anon_sym_chomp] = ACTIONS(3113),
    [anon_sym_chr] = ACTIONS(3113),
    [anon_sym_cos] = ACTIONS(3113),
    [anon_sym_chroot] = ACTIONS(3113),
    [anon_sym_defined] = ACTIONS(3113),
    [anon_sym_delete] = ACTIONS(3113),
    [anon_sym_dbmclose] = ACTIONS(3113),
    [anon_sym_exists] = ACTIONS(3113),
    [anon_sym_exit] = ACTIONS(3113),
    [anon_sym_eof] = ACTIONS(3113),
    [anon_sym_exp] = ACTIONS(3113),
    [anon_sym_each] = ACTIONS(3113),
    [anon_sym_fc] = ACTIONS(3113),
    [anon_sym_fileno] = ACTIONS(3113),
    [anon_sym_gmtime] = ACTIONS(3113),
    [anon_sym_getc] = ACTIONS(3113),
    [anon_sym_getpgrp] = ACTIONS(3113),
    [anon_sym_getprotobyname] = ACTIONS(3113),
    [anon_sym_getpwname] = ACTIONS(3113),
    [anon_sym_getpwuid] = ACTIONS(3113),
    [anon_sym_getpeername] = ACTIONS(3113),
    [anon_sym_getnetbyname] = ACTIONS(3113),
    [anon_sym_getsockname] = ACTIONS(3113),
    [anon_sym_getgrnam] = ACTIONS(3113),
    [anon_sym_getgrgid] = ACTIONS(3113),
    [anon_sym_hex] = ACTIONS(3113),
    [anon_sym_int] = ACTIONS(3113),
    [anon_sym_keys] = ACTIONS(3113),
    [anon_sym_lc] = ACTIONS(3113),
    [anon_sym_lcfirst] = ACTIONS(3113),
    [anon_sym_length] = ACTIONS(3113),
    [anon_sym_localtime] = ACTIONS(3113),
    [anon_sym_log] = ACTIONS(3113),
    [anon_sym_lock] = ACTIONS(3113),
    [anon_sym_lstat] = ACTIONS(3113),
    [anon_sym_oct] = ACTIONS(3113),
    [anon_sym_ord] = ACTIONS(3113),
    [anon_sym_prototype] = ACTIONS(3113),
    [anon_sym_pop] = ACTIONS(3113),
    [anon_sym_pos] = ACTIONS(3113),
    [anon_sym_quotemeta] = ACTIONS(3113),
    [anon_sym_reset] = ACTIONS(3113),
    [anon_sym_rand] = ACTIONS(3113),
    [anon_sym_rmdir] = ACTIONS(3113),
    [anon_sym_readdir] = ACTIONS(3113),
    [anon_sym_readline] = ACTIONS(3113),
    [anon_sym_readpipe] = ACTIONS(3113),
    [anon_sym_rewinddir] = ACTIONS(3113),
    [anon_sym_readlink] = ACTIONS(3113),
    [anon_sym_ref] = ACTIONS(3113),
    [anon_sym_scalar] = ACTIONS(3113),
    [anon_sym_shift] = ACTIONS(3113),
    [anon_sym_sin] = ACTIONS(3113),
    [anon_sym_sleep] = ACTIONS(3113),
    [anon_sym_sqrt] = ACTIONS(3113),
    [anon_sym_srand] = ACTIONS(3113),
    [anon_sym_stat] = ACTIONS(3113),
    [anon_sym_study] = ACTIONS(3113),
    [anon_sym_tell] = ACTIONS(3113),
    [anon_sym_telldir] = ACTIONS(3113),
    [anon_sym_tied] = ACTIONS(3113),
    [anon_sym_uc] = ACTIONS(3113),
    [anon_sym_ucfirst] = ACTIONS(3113),
    [anon_sym_untie] = ACTIONS(3113),
    [anon_sym_umask] = ACTIONS(3113),
    [anon_sym_values] = ACTIONS(3113),
    [anon_sym_write] = ACTIONS(3113),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3113),
    [anon_sym___END__] = ACTIONS(3113),
    [anon_sym_EOT] = ACTIONS(3113),
    [anon_sym_q] = ACTIONS(3113),
    [anon_sym_qq] = ACTIONS(3113),
    [anon_sym_qw] = ACTIONS(3113),
    [anon_sym_qx] = ACTIONS(3113),
    [anon_sym_qr] = ACTIONS(3113),
    [anon_sym_m] = ACTIONS(3113),
    [anon_sym_s] = ACTIONS(3113),
    [anon_sym_tr] = ACTIONS(3113),
    [anon_sym_y] = ACTIONS(3113),
    [anon_sym_if] = ACTIONS(3113),
    [anon_sym_unless] = ACTIONS(3113),
    [anon_sym_while] = ACTIONS(3113),
    [anon_sym_until] = ACTIONS(3113),
    [aux_sym__bareword_token1] = ACTIONS(3113),
    [sym_number] = ACTIONS(3113),
    [anon_sym_true] = ACTIONS(3113),
    [anon_sym_false] = ACTIONS(3113),
    [sym__single_quote] = ACTIONS(3111),
    [sym__double_quote] = ACTIONS(3111),
    [sym__backtick_quote] = ACTIONS(3111),
    [sym__search_slash_quote] = ACTIONS(3111),
    [sym__open_readline_bracket] = ACTIONS(3111),
    [sym__open_fileglob_bracket] = ACTIONS(3111),
    [sym__PERLY_HEREDOC] = ACTIONS(3111),
    [sym__ctrl_z_hack] = ACTIONS(3111),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3111),
    [sym__filetest] = ACTIONS(3111),
  },
  [STATE(1490)] = {
    [sym_heredoc_content] = STATE(1490),
    [ts_builtin_sym_end] = ACTIONS(3115),
    [sym__identifier] = ACTIONS(3117),
    [anon_sym_LBRACE] = ACTIONS(3117),
    [anon_sym_SEMI] = ACTIONS(3117),
    [anon_sym_package] = ACTIONS(3117),
    [anon_sym_class] = ACTIONS(3117),
    [anon_sym_role] = ACTIONS(3117),
    [anon_sym_BUILD] = ACTIONS(3117),
    [anon_sym_ADJUST] = ACTIONS(3117),
    [anon_sym_DOLLAR] = ACTIONS(3117),
    [anon_sym_AT] = ACTIONS(3117),
    [anon_sym_my] = ACTIONS(3117),
    [anon_sym_extended] = ACTIONS(3117),
    [anon_sym_async] = ACTIONS(3117),
    [anon_sym_sub] = ACTIONS(3117),
    [anon_sym_method] = ACTIONS(3117),
    [anon_sym_LPAREN] = ACTIONS(3117),
    [anon_sym_continue] = ACTIONS(3117),
    [anon_sym_state] = ACTIONS(3117),
    [anon_sym_our] = ACTIONS(3117),
    [anon_sym_try] = ACTIONS(3117),
    [anon_sym_defer] = ACTIONS(3117),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3117),
    [anon_sym_LBRACK] = ACTIONS(3117),
    [anon_sym_LT_LT] = ACTIONS(3117),
    [anon_sym_SLASH_SLASH] = ACTIONS(3117),
    [anon_sym_PLUS] = ACTIONS(3117),
    [anon_sym_DASH] = ACTIONS(3117),
    [anon_sym_TILDE] = ACTIONS(3117),
    [anon_sym_BANG] = ACTIONS(3117),
    [anon_sym_PLUS_PLUS] = ACTIONS(3117),
    [anon_sym_DASH_DASH] = ACTIONS(3117),
    [anon_sym_BSLASH] = ACTIONS(3115),
    [anon_sym_do] = ACTIONS(3117),
    [anon_sym_eval] = ACTIONS(3117),
    [anon_sym_await] = ACTIONS(3117),
    [anon_sym_field] = ACTIONS(3117),
    [anon_sym_local] = ACTIONS(3117),
    [anon_sym_dynamically] = ACTIONS(3117),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3117),
    [anon_sym_require] = ACTIONS(3117),
    [anon_sym_map] = ACTIONS(3117),
    [anon_sym_grep] = ACTIONS(3117),
    [anon_sym_sort] = ACTIONS(3117),
    [anon_sym_goto] = ACTIONS(3117),
    [anon_sym_return] = ACTIONS(3117),
    [anon_sym_undef] = ACTIONS(3117),
    [anon_sym_PERCENT2] = ACTIONS(3117),
    [anon_sym_AMP2] = ACTIONS(3117),
    [anon_sym_STAR2] = ACTIONS(3117),
    [anon_sym_use] = ACTIONS(3117),
    [anon_sym_no] = ACTIONS(3117),
    [anon_sym_for] = ACTIONS(3117),
    [anon_sym_foreach] = ACTIONS(3117),
    [anon_sym_last] = ACTIONS(3117),
    [anon_sym_next] = ACTIONS(3117),
    [anon_sym_redo] = ACTIONS(3117),
    [anon_sym_BEGIN] = ACTIONS(3117),
    [anon_sym_INIT] = ACTIONS(3117),
    [anon_sym_CHECK] = ACTIONS(3117),
    [anon_sym_UNITCHECK] = ACTIONS(3117),
    [anon_sym_END] = ACTIONS(3117),
    [anon_sym___FILE__] = ACTIONS(3117),
    [anon_sym___LINE__] = ACTIONS(3117),
    [anon_sym___PACKAGE__] = ACTIONS(3117),
    [anon_sym___SUB__] = ACTIONS(3117),
    [anon_sym_break] = ACTIONS(3117),
    [anon_sym_fork] = ACTIONS(3117),
    [anon_sym_getppid] = ACTIONS(3117),
    [anon_sym_time] = ACTIONS(3117),
    [anon_sym_times] = ACTIONS(3117),
    [anon_sym_wait] = ACTIONS(3117),
    [anon_sym_wantarray] = ACTIONS(3117),
    [anon_sym_abs] = ACTIONS(3117),
    [anon_sym_alarm] = ACTIONS(3117),
    [anon_sym_chop] = ACTIONS(3117),
    [anon_sym_chdir] = ACTIONS(3117),
    [anon_sym_close] = ACTIONS(3117),
    [anon_sym_closedir] = ACTIONS(3117),
    [anon_sym_caller] = ACTIONS(3117),
    [anon_sym_chomp] = ACTIONS(3117),
    [anon_sym_chr] = ACTIONS(3117),
    [anon_sym_cos] = ACTIONS(3117),
    [anon_sym_chroot] = ACTIONS(3117),
    [anon_sym_defined] = ACTIONS(3117),
    [anon_sym_delete] = ACTIONS(3117),
    [anon_sym_dbmclose] = ACTIONS(3117),
    [anon_sym_exists] = ACTIONS(3117),
    [anon_sym_exit] = ACTIONS(3117),
    [anon_sym_eof] = ACTIONS(3117),
    [anon_sym_exp] = ACTIONS(3117),
    [anon_sym_each] = ACTIONS(3117),
    [anon_sym_fc] = ACTIONS(3117),
    [anon_sym_fileno] = ACTIONS(3117),
    [anon_sym_gmtime] = ACTIONS(3117),
    [anon_sym_getc] = ACTIONS(3117),
    [anon_sym_getpgrp] = ACTIONS(3117),
    [anon_sym_getprotobyname] = ACTIONS(3117),
    [anon_sym_getpwname] = ACTIONS(3117),
    [anon_sym_getpwuid] = ACTIONS(3117),
    [anon_sym_getpeername] = ACTIONS(3117),
    [anon_sym_getnetbyname] = ACTIONS(3117),
    [anon_sym_getsockname] = ACTIONS(3117),
    [anon_sym_getgrnam] = ACTIONS(3117),
    [anon_sym_getgrgid] = ACTIONS(3117),
    [anon_sym_hex] = ACTIONS(3117),
    [anon_sym_int] = ACTIONS(3117),
    [anon_sym_keys] = ACTIONS(3117),
    [anon_sym_lc] = ACTIONS(3117),
    [anon_sym_lcfirst] = ACTIONS(3117),
    [anon_sym_length] = ACTIONS(3117),
    [anon_sym_localtime] = ACTIONS(3117),
    [anon_sym_log] = ACTIONS(3117),
    [anon_sym_lock] = ACTIONS(3117),
    [anon_sym_lstat] = ACTIONS(3117),
    [anon_sym_oct] = ACTIONS(3117),
    [anon_sym_ord] = ACTIONS(3117),
    [anon_sym_prototype] = ACTIONS(3117),
    [anon_sym_pop] = ACTIONS(3117),
    [anon_sym_pos] = ACTIONS(3117),
    [anon_sym_quotemeta] = ACTIONS(3117),
    [anon_sym_reset] = ACTIONS(3117),
    [anon_sym_rand] = ACTIONS(3117),
    [anon_sym_rmdir] = ACTIONS(3117),
    [anon_sym_readdir] = ACTIONS(3117),
    [anon_sym_readline] = ACTIONS(3117),
    [anon_sym_readpipe] = ACTIONS(3117),
    [anon_sym_rewinddir] = ACTIONS(3117),
    [anon_sym_readlink] = ACTIONS(3117),
    [anon_sym_ref] = ACTIONS(3117),
    [anon_sym_scalar] = ACTIONS(3117),
    [anon_sym_shift] = ACTIONS(3117),
    [anon_sym_sin] = ACTIONS(3117),
    [anon_sym_sleep] = ACTIONS(3117),
    [anon_sym_sqrt] = ACTIONS(3117),
    [anon_sym_srand] = ACTIONS(3117),
    [anon_sym_stat] = ACTIONS(3117),
    [anon_sym_study] = ACTIONS(3117),
    [anon_sym_tell] = ACTIONS(3117),
    [anon_sym_telldir] = ACTIONS(3117),
    [anon_sym_tied] = ACTIONS(3117),
    [anon_sym_uc] = ACTIONS(3117),
    [anon_sym_ucfirst] = ACTIONS(3117),
    [anon_sym_untie] = ACTIONS(3117),
    [anon_sym_umask] = ACTIONS(3117),
    [anon_sym_values] = ACTIONS(3117),
    [anon_sym_write] = ACTIONS(3117),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3117),
    [anon_sym___END__] = ACTIONS(3117),
    [anon_sym_EOT] = ACTIONS(3117),
    [anon_sym_q] = ACTIONS(3117),
    [anon_sym_qq] = ACTIONS(3117),
    [anon_sym_qw] = ACTIONS(3117),
    [anon_sym_qx] = ACTIONS(3117),
    [anon_sym_qr] = ACTIONS(3117),
    [anon_sym_m] = ACTIONS(3117),
    [anon_sym_s] = ACTIONS(3117),
    [anon_sym_tr] = ACTIONS(3117),
    [anon_sym_y] = ACTIONS(3117),
    [anon_sym_if] = ACTIONS(3117),
    [anon_sym_unless] = ACTIONS(3117),
    [anon_sym_while] = ACTIONS(3117),
    [anon_sym_until] = ACTIONS(3117),
    [aux_sym__bareword_token1] = ACTIONS(3117),
    [sym_number] = ACTIONS(3117),
    [anon_sym_true] = ACTIONS(3117),
    [anon_sym_false] = ACTIONS(3117),
    [sym__single_quote] = ACTIONS(3115),
    [sym__double_quote] = ACTIONS(3115),
    [sym__backtick_quote] = ACTIONS(3115),
    [sym__search_slash_quote] = ACTIONS(3115),
    [sym__open_readline_bracket] = ACTIONS(3115),
    [sym__open_fileglob_bracket] = ACTIONS(3115),
    [sym__PERLY_HEREDOC] = ACTIONS(3115),
    [sym__ctrl_z_hack] = ACTIONS(3115),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3115),
    [sym__filetest] = ACTIONS(3115),
  },
  [STATE(1491)] = {
    [sym_heredoc_content] = STATE(1491),
    [ts_builtin_sym_end] = ACTIONS(3119),
    [sym__identifier] = ACTIONS(3121),
    [anon_sym_LBRACE] = ACTIONS(3121),
    [anon_sym_SEMI] = ACTIONS(3121),
    [anon_sym_package] = ACTIONS(3121),
    [anon_sym_class] = ACTIONS(3121),
    [anon_sym_role] = ACTIONS(3121),
    [anon_sym_BUILD] = ACTIONS(3121),
    [anon_sym_ADJUST] = ACTIONS(3121),
    [anon_sym_DOLLAR] = ACTIONS(3121),
    [anon_sym_AT] = ACTIONS(3121),
    [anon_sym_my] = ACTIONS(3121),
    [anon_sym_extended] = ACTIONS(3121),
    [anon_sym_async] = ACTIONS(3121),
    [anon_sym_sub] = ACTIONS(3121),
    [anon_sym_method] = ACTIONS(3121),
    [anon_sym_LPAREN] = ACTIONS(3121),
    [anon_sym_continue] = ACTIONS(3121),
    [anon_sym_state] = ACTIONS(3121),
    [anon_sym_our] = ACTIONS(3121),
    [anon_sym_try] = ACTIONS(3121),
    [anon_sym_defer] = ACTIONS(3121),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3121),
    [anon_sym_LBRACK] = ACTIONS(3121),
    [anon_sym_LT_LT] = ACTIONS(3121),
    [anon_sym_SLASH_SLASH] = ACTIONS(3121),
    [anon_sym_PLUS] = ACTIONS(3121),
    [anon_sym_DASH] = ACTIONS(3121),
    [anon_sym_TILDE] = ACTIONS(3121),
    [anon_sym_BANG] = ACTIONS(3121),
    [anon_sym_PLUS_PLUS] = ACTIONS(3121),
    [anon_sym_DASH_DASH] = ACTIONS(3121),
    [anon_sym_BSLASH] = ACTIONS(3119),
    [anon_sym_do] = ACTIONS(3121),
    [anon_sym_eval] = ACTIONS(3121),
    [anon_sym_await] = ACTIONS(3121),
    [anon_sym_field] = ACTIONS(3121),
    [anon_sym_local] = ACTIONS(3121),
    [anon_sym_dynamically] = ACTIONS(3121),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3121),
    [anon_sym_require] = ACTIONS(3121),
    [anon_sym_map] = ACTIONS(3121),
    [anon_sym_grep] = ACTIONS(3121),
    [anon_sym_sort] = ACTIONS(3121),
    [anon_sym_goto] = ACTIONS(3121),
    [anon_sym_return] = ACTIONS(3121),
    [anon_sym_undef] = ACTIONS(3121),
    [anon_sym_PERCENT2] = ACTIONS(3121),
    [anon_sym_AMP2] = ACTIONS(3121),
    [anon_sym_STAR2] = ACTIONS(3121),
    [anon_sym_use] = ACTIONS(3121),
    [anon_sym_no] = ACTIONS(3121),
    [anon_sym_for] = ACTIONS(3121),
    [anon_sym_foreach] = ACTIONS(3121),
    [anon_sym_last] = ACTIONS(3121),
    [anon_sym_next] = ACTIONS(3121),
    [anon_sym_redo] = ACTIONS(3121),
    [anon_sym_BEGIN] = ACTIONS(3121),
    [anon_sym_INIT] = ACTIONS(3121),
    [anon_sym_CHECK] = ACTIONS(3121),
    [anon_sym_UNITCHECK] = ACTIONS(3121),
    [anon_sym_END] = ACTIONS(3121),
    [anon_sym___FILE__] = ACTIONS(3121),
    [anon_sym___LINE__] = ACTIONS(3121),
    [anon_sym___PACKAGE__] = ACTIONS(3121),
    [anon_sym___SUB__] = ACTIONS(3121),
    [anon_sym_break] = ACTIONS(3121),
    [anon_sym_fork] = ACTIONS(3121),
    [anon_sym_getppid] = ACTIONS(3121),
    [anon_sym_time] = ACTIONS(3121),
    [anon_sym_times] = ACTIONS(3121),
    [anon_sym_wait] = ACTIONS(3121),
    [anon_sym_wantarray] = ACTIONS(3121),
    [anon_sym_abs] = ACTIONS(3121),
    [anon_sym_alarm] = ACTIONS(3121),
    [anon_sym_chop] = ACTIONS(3121),
    [anon_sym_chdir] = ACTIONS(3121),
    [anon_sym_close] = ACTIONS(3121),
    [anon_sym_closedir] = ACTIONS(3121),
    [anon_sym_caller] = ACTIONS(3121),
    [anon_sym_chomp] = ACTIONS(3121),
    [anon_sym_chr] = ACTIONS(3121),
    [anon_sym_cos] = ACTIONS(3121),
    [anon_sym_chroot] = ACTIONS(3121),
    [anon_sym_defined] = ACTIONS(3121),
    [anon_sym_delete] = ACTIONS(3121),
    [anon_sym_dbmclose] = ACTIONS(3121),
    [anon_sym_exists] = ACTIONS(3121),
    [anon_sym_exit] = ACTIONS(3121),
    [anon_sym_eof] = ACTIONS(3121),
    [anon_sym_exp] = ACTIONS(3121),
    [anon_sym_each] = ACTIONS(3121),
    [anon_sym_fc] = ACTIONS(3121),
    [anon_sym_fileno] = ACTIONS(3121),
    [anon_sym_gmtime] = ACTIONS(3121),
    [anon_sym_getc] = ACTIONS(3121),
    [anon_sym_getpgrp] = ACTIONS(3121),
    [anon_sym_getprotobyname] = ACTIONS(3121),
    [anon_sym_getpwname] = ACTIONS(3121),
    [anon_sym_getpwuid] = ACTIONS(3121),
    [anon_sym_getpeername] = ACTIONS(3121),
    [anon_sym_getnetbyname] = ACTIONS(3121),
    [anon_sym_getsockname] = ACTIONS(3121),
    [anon_sym_getgrnam] = ACTIONS(3121),
    [anon_sym_getgrgid] = ACTIONS(3121),
    [anon_sym_hex] = ACTIONS(3121),
    [anon_sym_int] = ACTIONS(3121),
    [anon_sym_keys] = ACTIONS(3121),
    [anon_sym_lc] = ACTIONS(3121),
    [anon_sym_lcfirst] = ACTIONS(3121),
    [anon_sym_length] = ACTIONS(3121),
    [anon_sym_localtime] = ACTIONS(3121),
    [anon_sym_log] = ACTIONS(3121),
    [anon_sym_lock] = ACTIONS(3121),
    [anon_sym_lstat] = ACTIONS(3121),
    [anon_sym_oct] = ACTIONS(3121),
    [anon_sym_ord] = ACTIONS(3121),
    [anon_sym_prototype] = ACTIONS(3121),
    [anon_sym_pop] = ACTIONS(3121),
    [anon_sym_pos] = ACTIONS(3121),
    [anon_sym_quotemeta] = ACTIONS(3121),
    [anon_sym_reset] = ACTIONS(3121),
    [anon_sym_rand] = ACTIONS(3121),
    [anon_sym_rmdir] = ACTIONS(3121),
    [anon_sym_readdir] = ACTIONS(3121),
    [anon_sym_readline] = ACTIONS(3121),
    [anon_sym_readpipe] = ACTIONS(3121),
    [anon_sym_rewinddir] = ACTIONS(3121),
    [anon_sym_readlink] = ACTIONS(3121),
    [anon_sym_ref] = ACTIONS(3121),
    [anon_sym_scalar] = ACTIONS(3121),
    [anon_sym_shift] = ACTIONS(3121),
    [anon_sym_sin] = ACTIONS(3121),
    [anon_sym_sleep] = ACTIONS(3121),
    [anon_sym_sqrt] = ACTIONS(3121),
    [anon_sym_srand] = ACTIONS(3121),
    [anon_sym_stat] = ACTIONS(3121),
    [anon_sym_study] = ACTIONS(3121),
    [anon_sym_tell] = ACTIONS(3121),
    [anon_sym_telldir] = ACTIONS(3121),
    [anon_sym_tied] = ACTIONS(3121),
    [anon_sym_uc] = ACTIONS(3121),
    [anon_sym_ucfirst] = ACTIONS(3121),
    [anon_sym_untie] = ACTIONS(3121),
    [anon_sym_umask] = ACTIONS(3121),
    [anon_sym_values] = ACTIONS(3121),
    [anon_sym_write] = ACTIONS(3121),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3121),
    [anon_sym___END__] = ACTIONS(3121),
    [anon_sym_EOT] = ACTIONS(3121),
    [anon_sym_q] = ACTIONS(3121),
    [anon_sym_qq] = ACTIONS(3121),
    [anon_sym_qw] = ACTIONS(3121),
    [anon_sym_qx] = ACTIONS(3121),
    [anon_sym_qr] = ACTIONS(3121),
    [anon_sym_m] = ACTIONS(3121),
    [anon_sym_s] = ACTIONS(3121),
    [anon_sym_tr] = ACTIONS(3121),
    [anon_sym_y] = ACTIONS(3121),
    [anon_sym_if] = ACTIONS(3121),
    [anon_sym_unless] = ACTIONS(3121),
    [anon_sym_while] = ACTIONS(3121),
    [anon_sym_until] = ACTIONS(3121),
    [aux_sym__bareword_token1] = ACTIONS(3121),
    [sym_number] = ACTIONS(3121),
    [anon_sym_true] = ACTIONS(3121),
    [anon_sym_false] = ACTIONS(3121),
    [sym__single_quote] = ACTIONS(3119),
    [sym__double_quote] = ACTIONS(3119),
    [sym__backtick_quote] = ACTIONS(3119),
    [sym__search_slash_quote] = ACTIONS(3119),
    [sym__open_readline_bracket] = ACTIONS(3119),
    [sym__open_fileglob_bracket] = ACTIONS(3119),
    [sym__PERLY_HEREDOC] = ACTIONS(3119),
    [sym__ctrl_z_hack] = ACTIONS(3119),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3119),
    [sym__filetest] = ACTIONS(3119),
  },
  [STATE(1492)] = {
    [sym_heredoc_content] = STATE(1492),
    [ts_builtin_sym_end] = ACTIONS(3123),
    [sym__identifier] = ACTIONS(3125),
    [anon_sym_LBRACE] = ACTIONS(3125),
    [anon_sym_SEMI] = ACTIONS(3125),
    [anon_sym_package] = ACTIONS(3125),
    [anon_sym_class] = ACTIONS(3125),
    [anon_sym_role] = ACTIONS(3125),
    [anon_sym_BUILD] = ACTIONS(3125),
    [anon_sym_ADJUST] = ACTIONS(3125),
    [anon_sym_DOLLAR] = ACTIONS(3125),
    [anon_sym_AT] = ACTIONS(3125),
    [anon_sym_my] = ACTIONS(3125),
    [anon_sym_extended] = ACTIONS(3125),
    [anon_sym_async] = ACTIONS(3125),
    [anon_sym_sub] = ACTIONS(3125),
    [anon_sym_method] = ACTIONS(3125),
    [anon_sym_LPAREN] = ACTIONS(3125),
    [anon_sym_continue] = ACTIONS(3125),
    [anon_sym_state] = ACTIONS(3125),
    [anon_sym_our] = ACTIONS(3125),
    [anon_sym_try] = ACTIONS(3125),
    [anon_sym_defer] = ACTIONS(3125),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3125),
    [anon_sym_LBRACK] = ACTIONS(3125),
    [anon_sym_LT_LT] = ACTIONS(3125),
    [anon_sym_SLASH_SLASH] = ACTIONS(3125),
    [anon_sym_PLUS] = ACTIONS(3125),
    [anon_sym_DASH] = ACTIONS(3125),
    [anon_sym_TILDE] = ACTIONS(3125),
    [anon_sym_BANG] = ACTIONS(3125),
    [anon_sym_PLUS_PLUS] = ACTIONS(3125),
    [anon_sym_DASH_DASH] = ACTIONS(3125),
    [anon_sym_BSLASH] = ACTIONS(3123),
    [anon_sym_do] = ACTIONS(3125),
    [anon_sym_eval] = ACTIONS(3125),
    [anon_sym_await] = ACTIONS(3125),
    [anon_sym_field] = ACTIONS(3125),
    [anon_sym_local] = ACTIONS(3125),
    [anon_sym_dynamically] = ACTIONS(3125),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3125),
    [anon_sym_require] = ACTIONS(3125),
    [anon_sym_map] = ACTIONS(3125),
    [anon_sym_grep] = ACTIONS(3125),
    [anon_sym_sort] = ACTIONS(3125),
    [anon_sym_goto] = ACTIONS(3125),
    [anon_sym_return] = ACTIONS(3125),
    [anon_sym_undef] = ACTIONS(3125),
    [anon_sym_PERCENT2] = ACTIONS(3125),
    [anon_sym_AMP2] = ACTIONS(3125),
    [anon_sym_STAR2] = ACTIONS(3125),
    [anon_sym_use] = ACTIONS(3125),
    [anon_sym_no] = ACTIONS(3125),
    [anon_sym_for] = ACTIONS(3125),
    [anon_sym_foreach] = ACTIONS(3125),
    [anon_sym_last] = ACTIONS(3125),
    [anon_sym_next] = ACTIONS(3125),
    [anon_sym_redo] = ACTIONS(3125),
    [anon_sym_BEGIN] = ACTIONS(3125),
    [anon_sym_INIT] = ACTIONS(3125),
    [anon_sym_CHECK] = ACTIONS(3125),
    [anon_sym_UNITCHECK] = ACTIONS(3125),
    [anon_sym_END] = ACTIONS(3125),
    [anon_sym___FILE__] = ACTIONS(3125),
    [anon_sym___LINE__] = ACTIONS(3125),
    [anon_sym___PACKAGE__] = ACTIONS(3125),
    [anon_sym___SUB__] = ACTIONS(3125),
    [anon_sym_break] = ACTIONS(3125),
    [anon_sym_fork] = ACTIONS(3125),
    [anon_sym_getppid] = ACTIONS(3125),
    [anon_sym_time] = ACTIONS(3125),
    [anon_sym_times] = ACTIONS(3125),
    [anon_sym_wait] = ACTIONS(3125),
    [anon_sym_wantarray] = ACTIONS(3125),
    [anon_sym_abs] = ACTIONS(3125),
    [anon_sym_alarm] = ACTIONS(3125),
    [anon_sym_chop] = ACTIONS(3125),
    [anon_sym_chdir] = ACTIONS(3125),
    [anon_sym_close] = ACTIONS(3125),
    [anon_sym_closedir] = ACTIONS(3125),
    [anon_sym_caller] = ACTIONS(3125),
    [anon_sym_chomp] = ACTIONS(3125),
    [anon_sym_chr] = ACTIONS(3125),
    [anon_sym_cos] = ACTIONS(3125),
    [anon_sym_chroot] = ACTIONS(3125),
    [anon_sym_defined] = ACTIONS(3125),
    [anon_sym_delete] = ACTIONS(3125),
    [anon_sym_dbmclose] = ACTIONS(3125),
    [anon_sym_exists] = ACTIONS(3125),
    [anon_sym_exit] = ACTIONS(3125),
    [anon_sym_eof] = ACTIONS(3125),
    [anon_sym_exp] = ACTIONS(3125),
    [anon_sym_each] = ACTIONS(3125),
    [anon_sym_fc] = ACTIONS(3125),
    [anon_sym_fileno] = ACTIONS(3125),
    [anon_sym_gmtime] = ACTIONS(3125),
    [anon_sym_getc] = ACTIONS(3125),
    [anon_sym_getpgrp] = ACTIONS(3125),
    [anon_sym_getprotobyname] = ACTIONS(3125),
    [anon_sym_getpwname] = ACTIONS(3125),
    [anon_sym_getpwuid] = ACTIONS(3125),
    [anon_sym_getpeername] = ACTIONS(3125),
    [anon_sym_getnetbyname] = ACTIONS(3125),
    [anon_sym_getsockname] = ACTIONS(3125),
    [anon_sym_getgrnam] = ACTIONS(3125),
    [anon_sym_getgrgid] = ACTIONS(3125),
    [anon_sym_hex] = ACTIONS(3125),
    [anon_sym_int] = ACTIONS(3125),
    [anon_sym_keys] = ACTIONS(3125),
    [anon_sym_lc] = ACTIONS(3125),
    [anon_sym_lcfirst] = ACTIONS(3125),
    [anon_sym_length] = ACTIONS(3125),
    [anon_sym_localtime] = ACTIONS(3125),
    [anon_sym_log] = ACTIONS(3125),
    [anon_sym_lock] = ACTIONS(3125),
    [anon_sym_lstat] = ACTIONS(3125),
    [anon_sym_oct] = ACTIONS(3125),
    [anon_sym_ord] = ACTIONS(3125),
    [anon_sym_prototype] = ACTIONS(3125),
    [anon_sym_pop] = ACTIONS(3125),
    [anon_sym_pos] = ACTIONS(3125),
    [anon_sym_quotemeta] = ACTIONS(3125),
    [anon_sym_reset] = ACTIONS(3125),
    [anon_sym_rand] = ACTIONS(3125),
    [anon_sym_rmdir] = ACTIONS(3125),
    [anon_sym_readdir] = ACTIONS(3125),
    [anon_sym_readline] = ACTIONS(3125),
    [anon_sym_readpipe] = ACTIONS(3125),
    [anon_sym_rewinddir] = ACTIONS(3125),
    [anon_sym_readlink] = ACTIONS(3125),
    [anon_sym_ref] = ACTIONS(3125),
    [anon_sym_scalar] = ACTIONS(3125),
    [anon_sym_shift] = ACTIONS(3125),
    [anon_sym_sin] = ACTIONS(3125),
    [anon_sym_sleep] = ACTIONS(3125),
    [anon_sym_sqrt] = ACTIONS(3125),
    [anon_sym_srand] = ACTIONS(3125),
    [anon_sym_stat] = ACTIONS(3125),
    [anon_sym_study] = ACTIONS(3125),
    [anon_sym_tell] = ACTIONS(3125),
    [anon_sym_telldir] = ACTIONS(3125),
    [anon_sym_tied] = ACTIONS(3125),
    [anon_sym_uc] = ACTIONS(3125),
    [anon_sym_ucfirst] = ACTIONS(3125),
    [anon_sym_untie] = ACTIONS(3125),
    [anon_sym_umask] = ACTIONS(3125),
    [anon_sym_values] = ACTIONS(3125),
    [anon_sym_write] = ACTIONS(3125),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3125),
    [anon_sym___END__] = ACTIONS(3125),
    [anon_sym_EOT] = ACTIONS(3125),
    [anon_sym_q] = ACTIONS(3125),
    [anon_sym_qq] = ACTIONS(3125),
    [anon_sym_qw] = ACTIONS(3125),
    [anon_sym_qx] = ACTIONS(3125),
    [anon_sym_qr] = ACTIONS(3125),
    [anon_sym_m] = ACTIONS(3125),
    [anon_sym_s] = ACTIONS(3125),
    [anon_sym_tr] = ACTIONS(3125),
    [anon_sym_y] = ACTIONS(3125),
    [anon_sym_if] = ACTIONS(3125),
    [anon_sym_unless] = ACTIONS(3125),
    [anon_sym_while] = ACTIONS(3125),
    [anon_sym_until] = ACTIONS(3125),
    [aux_sym__bareword_token1] = ACTIONS(3125),
    [sym_number] = ACTIONS(3125),
    [anon_sym_true] = ACTIONS(3125),
    [anon_sym_false] = ACTIONS(3125),
    [sym__single_quote] = ACTIONS(3123),
    [sym__double_quote] = ACTIONS(3123),
    [sym__backtick_quote] = ACTIONS(3123),
    [sym__search_slash_quote] = ACTIONS(3123),
    [sym__open_readline_bracket] = ACTIONS(3123),
    [sym__open_fileglob_bracket] = ACTIONS(3123),
    [sym__PERLY_HEREDOC] = ACTIONS(3123),
    [sym__ctrl_z_hack] = ACTIONS(3123),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3123),
    [sym__filetest] = ACTIONS(3123),
  },
  [STATE(1493)] = {
    [sym_heredoc_content] = STATE(1493),
    [ts_builtin_sym_end] = ACTIONS(3127),
    [sym__identifier] = ACTIONS(3129),
    [anon_sym_LBRACE] = ACTIONS(3129),
    [anon_sym_SEMI] = ACTIONS(3129),
    [anon_sym_package] = ACTIONS(3129),
    [anon_sym_class] = ACTIONS(3129),
    [anon_sym_role] = ACTIONS(3129),
    [anon_sym_BUILD] = ACTIONS(3129),
    [anon_sym_ADJUST] = ACTIONS(3129),
    [anon_sym_DOLLAR] = ACTIONS(3129),
    [anon_sym_AT] = ACTIONS(3129),
    [anon_sym_my] = ACTIONS(3129),
    [anon_sym_extended] = ACTIONS(3129),
    [anon_sym_async] = ACTIONS(3129),
    [anon_sym_sub] = ACTIONS(3129),
    [anon_sym_method] = ACTIONS(3129),
    [anon_sym_LPAREN] = ACTIONS(3129),
    [anon_sym_continue] = ACTIONS(3129),
    [anon_sym_state] = ACTIONS(3129),
    [anon_sym_our] = ACTIONS(3129),
    [anon_sym_try] = ACTIONS(3129),
    [anon_sym_defer] = ACTIONS(3129),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3129),
    [anon_sym_LBRACK] = ACTIONS(3129),
    [anon_sym_LT_LT] = ACTIONS(3129),
    [anon_sym_SLASH_SLASH] = ACTIONS(3129),
    [anon_sym_PLUS] = ACTIONS(3129),
    [anon_sym_DASH] = ACTIONS(3129),
    [anon_sym_TILDE] = ACTIONS(3129),
    [anon_sym_BANG] = ACTIONS(3129),
    [anon_sym_PLUS_PLUS] = ACTIONS(3129),
    [anon_sym_DASH_DASH] = ACTIONS(3129),
    [anon_sym_BSLASH] = ACTIONS(3127),
    [anon_sym_do] = ACTIONS(3129),
    [anon_sym_eval] = ACTIONS(3129),
    [anon_sym_await] = ACTIONS(3129),
    [anon_sym_field] = ACTIONS(3129),
    [anon_sym_local] = ACTIONS(3129),
    [anon_sym_dynamically] = ACTIONS(3129),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3129),
    [anon_sym_require] = ACTIONS(3129),
    [anon_sym_map] = ACTIONS(3129),
    [anon_sym_grep] = ACTIONS(3129),
    [anon_sym_sort] = ACTIONS(3129),
    [anon_sym_goto] = ACTIONS(3129),
    [anon_sym_return] = ACTIONS(3129),
    [anon_sym_undef] = ACTIONS(3129),
    [anon_sym_PERCENT2] = ACTIONS(3129),
    [anon_sym_AMP2] = ACTIONS(3129),
    [anon_sym_STAR2] = ACTIONS(3129),
    [anon_sym_use] = ACTIONS(3129),
    [anon_sym_no] = ACTIONS(3129),
    [anon_sym_for] = ACTIONS(3129),
    [anon_sym_foreach] = ACTIONS(3129),
    [anon_sym_last] = ACTIONS(3129),
    [anon_sym_next] = ACTIONS(3129),
    [anon_sym_redo] = ACTIONS(3129),
    [anon_sym_BEGIN] = ACTIONS(3129),
    [anon_sym_INIT] = ACTIONS(3129),
    [anon_sym_CHECK] = ACTIONS(3129),
    [anon_sym_UNITCHECK] = ACTIONS(3129),
    [anon_sym_END] = ACTIONS(3129),
    [anon_sym___FILE__] = ACTIONS(3129),
    [anon_sym___LINE__] = ACTIONS(3129),
    [anon_sym___PACKAGE__] = ACTIONS(3129),
    [anon_sym___SUB__] = ACTIONS(3129),
    [anon_sym_break] = ACTIONS(3129),
    [anon_sym_fork] = ACTIONS(3129),
    [anon_sym_getppid] = ACTIONS(3129),
    [anon_sym_time] = ACTIONS(3129),
    [anon_sym_times] = ACTIONS(3129),
    [anon_sym_wait] = ACTIONS(3129),
    [anon_sym_wantarray] = ACTIONS(3129),
    [anon_sym_abs] = ACTIONS(3129),
    [anon_sym_alarm] = ACTIONS(3129),
    [anon_sym_chop] = ACTIONS(3129),
    [anon_sym_chdir] = ACTIONS(3129),
    [anon_sym_close] = ACTIONS(3129),
    [anon_sym_closedir] = ACTIONS(3129),
    [anon_sym_caller] = ACTIONS(3129),
    [anon_sym_chomp] = ACTIONS(3129),
    [anon_sym_chr] = ACTIONS(3129),
    [anon_sym_cos] = ACTIONS(3129),
    [anon_sym_chroot] = ACTIONS(3129),
    [anon_sym_defined] = ACTIONS(3129),
    [anon_sym_delete] = ACTIONS(3129),
    [anon_sym_dbmclose] = ACTIONS(3129),
    [anon_sym_exists] = ACTIONS(3129),
    [anon_sym_exit] = ACTIONS(3129),
    [anon_sym_eof] = ACTIONS(3129),
    [anon_sym_exp] = ACTIONS(3129),
    [anon_sym_each] = ACTIONS(3129),
    [anon_sym_fc] = ACTIONS(3129),
    [anon_sym_fileno] = ACTIONS(3129),
    [anon_sym_gmtime] = ACTIONS(3129),
    [anon_sym_getc] = ACTIONS(3129),
    [anon_sym_getpgrp] = ACTIONS(3129),
    [anon_sym_getprotobyname] = ACTIONS(3129),
    [anon_sym_getpwname] = ACTIONS(3129),
    [anon_sym_getpwuid] = ACTIONS(3129),
    [anon_sym_getpeername] = ACTIONS(3129),
    [anon_sym_getnetbyname] = ACTIONS(3129),
    [anon_sym_getsockname] = ACTIONS(3129),
    [anon_sym_getgrnam] = ACTIONS(3129),
    [anon_sym_getgrgid] = ACTIONS(3129),
    [anon_sym_hex] = ACTIONS(3129),
    [anon_sym_int] = ACTIONS(3129),
    [anon_sym_keys] = ACTIONS(3129),
    [anon_sym_lc] = ACTIONS(3129),
    [anon_sym_lcfirst] = ACTIONS(3129),
    [anon_sym_length] = ACTIONS(3129),
    [anon_sym_localtime] = ACTIONS(3129),
    [anon_sym_log] = ACTIONS(3129),
    [anon_sym_lock] = ACTIONS(3129),
    [anon_sym_lstat] = ACTIONS(3129),
    [anon_sym_oct] = ACTIONS(3129),
    [anon_sym_ord] = ACTIONS(3129),
    [anon_sym_prototype] = ACTIONS(3129),
    [anon_sym_pop] = ACTIONS(3129),
    [anon_sym_pos] = ACTIONS(3129),
    [anon_sym_quotemeta] = ACTIONS(3129),
    [anon_sym_reset] = ACTIONS(3129),
    [anon_sym_rand] = ACTIONS(3129),
    [anon_sym_rmdir] = ACTIONS(3129),
    [anon_sym_readdir] = ACTIONS(3129),
    [anon_sym_readline] = ACTIONS(3129),
    [anon_sym_readpipe] = ACTIONS(3129),
    [anon_sym_rewinddir] = ACTIONS(3129),
    [anon_sym_readlink] = ACTIONS(3129),
    [anon_sym_ref] = ACTIONS(3129),
    [anon_sym_scalar] = ACTIONS(3129),
    [anon_sym_shift] = ACTIONS(3129),
    [anon_sym_sin] = ACTIONS(3129),
    [anon_sym_sleep] = ACTIONS(3129),
    [anon_sym_sqrt] = ACTIONS(3129),
    [anon_sym_srand] = ACTIONS(3129),
    [anon_sym_stat] = ACTIONS(3129),
    [anon_sym_study] = ACTIONS(3129),
    [anon_sym_tell] = ACTIONS(3129),
    [anon_sym_telldir] = ACTIONS(3129),
    [anon_sym_tied] = ACTIONS(3129),
    [anon_sym_uc] = ACTIONS(3129),
    [anon_sym_ucfirst] = ACTIONS(3129),
    [anon_sym_untie] = ACTIONS(3129),
    [anon_sym_umask] = ACTIONS(3129),
    [anon_sym_values] = ACTIONS(3129),
    [anon_sym_write] = ACTIONS(3129),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3129),
    [anon_sym___END__] = ACTIONS(3129),
    [anon_sym_EOT] = ACTIONS(3129),
    [anon_sym_q] = ACTIONS(3129),
    [anon_sym_qq] = ACTIONS(3129),
    [anon_sym_qw] = ACTIONS(3129),
    [anon_sym_qx] = ACTIONS(3129),
    [anon_sym_qr] = ACTIONS(3129),
    [anon_sym_m] = ACTIONS(3129),
    [anon_sym_s] = ACTIONS(3129),
    [anon_sym_tr] = ACTIONS(3129),
    [anon_sym_y] = ACTIONS(3129),
    [anon_sym_if] = ACTIONS(3129),
    [anon_sym_unless] = ACTIONS(3129),
    [anon_sym_while] = ACTIONS(3129),
    [anon_sym_until] = ACTIONS(3129),
    [aux_sym__bareword_token1] = ACTIONS(3129),
    [sym_number] = ACTIONS(3129),
    [anon_sym_true] = ACTIONS(3129),
    [anon_sym_false] = ACTIONS(3129),
    [sym__single_quote] = ACTIONS(3127),
    [sym__double_quote] = ACTIONS(3127),
    [sym__backtick_quote] = ACTIONS(3127),
    [sym__search_slash_quote] = ACTIONS(3127),
    [sym__open_readline_bracket] = ACTIONS(3127),
    [sym__open_fileglob_bracket] = ACTIONS(3127),
    [sym__PERLY_HEREDOC] = ACTIONS(3127),
    [sym__ctrl_z_hack] = ACTIONS(3127),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3127),
    [sym__filetest] = ACTIONS(3127),
  },
  [STATE(1494)] = {
    [sym_heredoc_content] = STATE(1494),
    [ts_builtin_sym_end] = ACTIONS(3131),
    [sym__identifier] = ACTIONS(3133),
    [anon_sym_LBRACE] = ACTIONS(3133),
    [anon_sym_SEMI] = ACTIONS(3133),
    [anon_sym_package] = ACTIONS(3133),
    [anon_sym_class] = ACTIONS(3133),
    [anon_sym_role] = ACTIONS(3133),
    [anon_sym_BUILD] = ACTIONS(3133),
    [anon_sym_ADJUST] = ACTIONS(3133),
    [anon_sym_DOLLAR] = ACTIONS(3133),
    [anon_sym_AT] = ACTIONS(3133),
    [anon_sym_my] = ACTIONS(3133),
    [anon_sym_extended] = ACTIONS(3133),
    [anon_sym_async] = ACTIONS(3133),
    [anon_sym_sub] = ACTIONS(3133),
    [anon_sym_method] = ACTIONS(3133),
    [anon_sym_LPAREN] = ACTIONS(3133),
    [anon_sym_continue] = ACTIONS(3133),
    [anon_sym_state] = ACTIONS(3133),
    [anon_sym_our] = ACTIONS(3133),
    [anon_sym_try] = ACTIONS(3133),
    [anon_sym_defer] = ACTIONS(3133),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3133),
    [anon_sym_LBRACK] = ACTIONS(3133),
    [anon_sym_LT_LT] = ACTIONS(3133),
    [anon_sym_SLASH_SLASH] = ACTIONS(3133),
    [anon_sym_PLUS] = ACTIONS(3133),
    [anon_sym_DASH] = ACTIONS(3133),
    [anon_sym_TILDE] = ACTIONS(3133),
    [anon_sym_BANG] = ACTIONS(3133),
    [anon_sym_PLUS_PLUS] = ACTIONS(3133),
    [anon_sym_DASH_DASH] = ACTIONS(3133),
    [anon_sym_BSLASH] = ACTIONS(3131),
    [anon_sym_do] = ACTIONS(3133),
    [anon_sym_eval] = ACTIONS(3133),
    [anon_sym_await] = ACTIONS(3133),
    [anon_sym_field] = ACTIONS(3133),
    [anon_sym_local] = ACTIONS(3133),
    [anon_sym_dynamically] = ACTIONS(3133),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3133),
    [anon_sym_require] = ACTIONS(3133),
    [anon_sym_map] = ACTIONS(3133),
    [anon_sym_grep] = ACTIONS(3133),
    [anon_sym_sort] = ACTIONS(3133),
    [anon_sym_goto] = ACTIONS(3133),
    [anon_sym_return] = ACTIONS(3133),
    [anon_sym_undef] = ACTIONS(3133),
    [anon_sym_PERCENT2] = ACTIONS(3133),
    [anon_sym_AMP2] = ACTIONS(3133),
    [anon_sym_STAR2] = ACTIONS(3133),
    [anon_sym_use] = ACTIONS(3133),
    [anon_sym_no] = ACTIONS(3133),
    [anon_sym_for] = ACTIONS(3133),
    [anon_sym_foreach] = ACTIONS(3133),
    [anon_sym_last] = ACTIONS(3133),
    [anon_sym_next] = ACTIONS(3133),
    [anon_sym_redo] = ACTIONS(3133),
    [anon_sym_BEGIN] = ACTIONS(3133),
    [anon_sym_INIT] = ACTIONS(3133),
    [anon_sym_CHECK] = ACTIONS(3133),
    [anon_sym_UNITCHECK] = ACTIONS(3133),
    [anon_sym_END] = ACTIONS(3133),
    [anon_sym___FILE__] = ACTIONS(3133),
    [anon_sym___LINE__] = ACTIONS(3133),
    [anon_sym___PACKAGE__] = ACTIONS(3133),
    [anon_sym___SUB__] = ACTIONS(3133),
    [anon_sym_break] = ACTIONS(3133),
    [anon_sym_fork] = ACTIONS(3133),
    [anon_sym_getppid] = ACTIONS(3133),
    [anon_sym_time] = ACTIONS(3133),
    [anon_sym_times] = ACTIONS(3133),
    [anon_sym_wait] = ACTIONS(3133),
    [anon_sym_wantarray] = ACTIONS(3133),
    [anon_sym_abs] = ACTIONS(3133),
    [anon_sym_alarm] = ACTIONS(3133),
    [anon_sym_chop] = ACTIONS(3133),
    [anon_sym_chdir] = ACTIONS(3133),
    [anon_sym_close] = ACTIONS(3133),
    [anon_sym_closedir] = ACTIONS(3133),
    [anon_sym_caller] = ACTIONS(3133),
    [anon_sym_chomp] = ACTIONS(3133),
    [anon_sym_chr] = ACTIONS(3133),
    [anon_sym_cos] = ACTIONS(3133),
    [anon_sym_chroot] = ACTIONS(3133),
    [anon_sym_defined] = ACTIONS(3133),
    [anon_sym_delete] = ACTIONS(3133),
    [anon_sym_dbmclose] = ACTIONS(3133),
    [anon_sym_exists] = ACTIONS(3133),
    [anon_sym_exit] = ACTIONS(3133),
    [anon_sym_eof] = ACTIONS(3133),
    [anon_sym_exp] = ACTIONS(3133),
    [anon_sym_each] = ACTIONS(3133),
    [anon_sym_fc] = ACTIONS(3133),
    [anon_sym_fileno] = ACTIONS(3133),
    [anon_sym_gmtime] = ACTIONS(3133),
    [anon_sym_getc] = ACTIONS(3133),
    [anon_sym_getpgrp] = ACTIONS(3133),
    [anon_sym_getprotobyname] = ACTIONS(3133),
    [anon_sym_getpwname] = ACTIONS(3133),
    [anon_sym_getpwuid] = ACTIONS(3133),
    [anon_sym_getpeername] = ACTIONS(3133),
    [anon_sym_getnetbyname] = ACTIONS(3133),
    [anon_sym_getsockname] = ACTIONS(3133),
    [anon_sym_getgrnam] = ACTIONS(3133),
    [anon_sym_getgrgid] = ACTIONS(3133),
    [anon_sym_hex] = ACTIONS(3133),
    [anon_sym_int] = ACTIONS(3133),
    [anon_sym_keys] = ACTIONS(3133),
    [anon_sym_lc] = ACTIONS(3133),
    [anon_sym_lcfirst] = ACTIONS(3133),
    [anon_sym_length] = ACTIONS(3133),
    [anon_sym_localtime] = ACTIONS(3133),
    [anon_sym_log] = ACTIONS(3133),
    [anon_sym_lock] = ACTIONS(3133),
    [anon_sym_lstat] = ACTIONS(3133),
    [anon_sym_oct] = ACTIONS(3133),
    [anon_sym_ord] = ACTIONS(3133),
    [anon_sym_prototype] = ACTIONS(3133),
    [anon_sym_pop] = ACTIONS(3133),
    [anon_sym_pos] = ACTIONS(3133),
    [anon_sym_quotemeta] = ACTIONS(3133),
    [anon_sym_reset] = ACTIONS(3133),
    [anon_sym_rand] = ACTIONS(3133),
    [anon_sym_rmdir] = ACTIONS(3133),
    [anon_sym_readdir] = ACTIONS(3133),
    [anon_sym_readline] = ACTIONS(3133),
    [anon_sym_readpipe] = ACTIONS(3133),
    [anon_sym_rewinddir] = ACTIONS(3133),
    [anon_sym_readlink] = ACTIONS(3133),
    [anon_sym_ref] = ACTIONS(3133),
    [anon_sym_scalar] = ACTIONS(3133),
    [anon_sym_shift] = ACTIONS(3133),
    [anon_sym_sin] = ACTIONS(3133),
    [anon_sym_sleep] = ACTIONS(3133),
    [anon_sym_sqrt] = ACTIONS(3133),
    [anon_sym_srand] = ACTIONS(3133),
    [anon_sym_stat] = ACTIONS(3133),
    [anon_sym_study] = ACTIONS(3133),
    [anon_sym_tell] = ACTIONS(3133),
    [anon_sym_telldir] = ACTIONS(3133),
    [anon_sym_tied] = ACTIONS(3133),
    [anon_sym_uc] = ACTIONS(3133),
    [anon_sym_ucfirst] = ACTIONS(3133),
    [anon_sym_untie] = ACTIONS(3133),
    [anon_sym_umask] = ACTIONS(3133),
    [anon_sym_values] = ACTIONS(3133),
    [anon_sym_write] = ACTIONS(3133),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3133),
    [anon_sym___END__] = ACTIONS(3133),
    [anon_sym_EOT] = ACTIONS(3133),
    [anon_sym_q] = ACTIONS(3133),
    [anon_sym_qq] = ACTIONS(3133),
    [anon_sym_qw] = ACTIONS(3133),
    [anon_sym_qx] = ACTIONS(3133),
    [anon_sym_qr] = ACTIONS(3133),
    [anon_sym_m] = ACTIONS(3133),
    [anon_sym_s] = ACTIONS(3133),
    [anon_sym_tr] = ACTIONS(3133),
    [anon_sym_y] = ACTIONS(3133),
    [anon_sym_if] = ACTIONS(3133),
    [anon_sym_unless] = ACTIONS(3133),
    [anon_sym_while] = ACTIONS(3133),
    [anon_sym_until] = ACTIONS(3133),
    [aux_sym__bareword_token1] = ACTIONS(3133),
    [sym_number] = ACTIONS(3133),
    [anon_sym_true] = ACTIONS(3133),
    [anon_sym_false] = ACTIONS(3133),
    [sym__single_quote] = ACTIONS(3131),
    [sym__double_quote] = ACTIONS(3131),
    [sym__backtick_quote] = ACTIONS(3131),
    [sym__search_slash_quote] = ACTIONS(3131),
    [sym__open_readline_bracket] = ACTIONS(3131),
    [sym__open_fileglob_bracket] = ACTIONS(3131),
    [sym__PERLY_HEREDOC] = ACTIONS(3131),
    [sym__ctrl_z_hack] = ACTIONS(3131),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3131),
    [sym__filetest] = ACTIONS(3131),
  },
  [STATE(1495)] = {
    [sym_heredoc_content] = STATE(1495),
    [ts_builtin_sym_end] = ACTIONS(3135),
    [sym__identifier] = ACTIONS(3137),
    [anon_sym_LBRACE] = ACTIONS(3137),
    [anon_sym_SEMI] = ACTIONS(3137),
    [anon_sym_package] = ACTIONS(3137),
    [anon_sym_class] = ACTIONS(3137),
    [anon_sym_role] = ACTIONS(3137),
    [anon_sym_BUILD] = ACTIONS(3137),
    [anon_sym_ADJUST] = ACTIONS(3137),
    [anon_sym_DOLLAR] = ACTIONS(3137),
    [anon_sym_AT] = ACTIONS(3137),
    [anon_sym_my] = ACTIONS(3137),
    [anon_sym_extended] = ACTIONS(3137),
    [anon_sym_async] = ACTIONS(3137),
    [anon_sym_sub] = ACTIONS(3137),
    [anon_sym_method] = ACTIONS(3137),
    [anon_sym_LPAREN] = ACTIONS(3137),
    [anon_sym_continue] = ACTIONS(3137),
    [anon_sym_state] = ACTIONS(3137),
    [anon_sym_our] = ACTIONS(3137),
    [anon_sym_try] = ACTIONS(3137),
    [anon_sym_defer] = ACTIONS(3137),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3137),
    [anon_sym_LBRACK] = ACTIONS(3137),
    [anon_sym_LT_LT] = ACTIONS(3137),
    [anon_sym_SLASH_SLASH] = ACTIONS(3137),
    [anon_sym_PLUS] = ACTIONS(3137),
    [anon_sym_DASH] = ACTIONS(3137),
    [anon_sym_TILDE] = ACTIONS(3137),
    [anon_sym_BANG] = ACTIONS(3137),
    [anon_sym_PLUS_PLUS] = ACTIONS(3137),
    [anon_sym_DASH_DASH] = ACTIONS(3137),
    [anon_sym_BSLASH] = ACTIONS(3135),
    [anon_sym_do] = ACTIONS(3137),
    [anon_sym_eval] = ACTIONS(3137),
    [anon_sym_await] = ACTIONS(3137),
    [anon_sym_field] = ACTIONS(3137),
    [anon_sym_local] = ACTIONS(3137),
    [anon_sym_dynamically] = ACTIONS(3137),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3137),
    [anon_sym_require] = ACTIONS(3137),
    [anon_sym_map] = ACTIONS(3137),
    [anon_sym_grep] = ACTIONS(3137),
    [anon_sym_sort] = ACTIONS(3137),
    [anon_sym_goto] = ACTIONS(3137),
    [anon_sym_return] = ACTIONS(3137),
    [anon_sym_undef] = ACTIONS(3137),
    [anon_sym_PERCENT2] = ACTIONS(3137),
    [anon_sym_AMP2] = ACTIONS(3137),
    [anon_sym_STAR2] = ACTIONS(3137),
    [anon_sym_use] = ACTIONS(3137),
    [anon_sym_no] = ACTIONS(3137),
    [anon_sym_for] = ACTIONS(3137),
    [anon_sym_foreach] = ACTIONS(3137),
    [anon_sym_last] = ACTIONS(3137),
    [anon_sym_next] = ACTIONS(3137),
    [anon_sym_redo] = ACTIONS(3137),
    [anon_sym_BEGIN] = ACTIONS(3137),
    [anon_sym_INIT] = ACTIONS(3137),
    [anon_sym_CHECK] = ACTIONS(3137),
    [anon_sym_UNITCHECK] = ACTIONS(3137),
    [anon_sym_END] = ACTIONS(3137),
    [anon_sym___FILE__] = ACTIONS(3137),
    [anon_sym___LINE__] = ACTIONS(3137),
    [anon_sym___PACKAGE__] = ACTIONS(3137),
    [anon_sym___SUB__] = ACTIONS(3137),
    [anon_sym_break] = ACTIONS(3137),
    [anon_sym_fork] = ACTIONS(3137),
    [anon_sym_getppid] = ACTIONS(3137),
    [anon_sym_time] = ACTIONS(3137),
    [anon_sym_times] = ACTIONS(3137),
    [anon_sym_wait] = ACTIONS(3137),
    [anon_sym_wantarray] = ACTIONS(3137),
    [anon_sym_abs] = ACTIONS(3137),
    [anon_sym_alarm] = ACTIONS(3137),
    [anon_sym_chop] = ACTIONS(3137),
    [anon_sym_chdir] = ACTIONS(3137),
    [anon_sym_close] = ACTIONS(3137),
    [anon_sym_closedir] = ACTIONS(3137),
    [anon_sym_caller] = ACTIONS(3137),
    [anon_sym_chomp] = ACTIONS(3137),
    [anon_sym_chr] = ACTIONS(3137),
    [anon_sym_cos] = ACTIONS(3137),
    [anon_sym_chroot] = ACTIONS(3137),
    [anon_sym_defined] = ACTIONS(3137),
    [anon_sym_delete] = ACTIONS(3137),
    [anon_sym_dbmclose] = ACTIONS(3137),
    [anon_sym_exists] = ACTIONS(3137),
    [anon_sym_exit] = ACTIONS(3137),
    [anon_sym_eof] = ACTIONS(3137),
    [anon_sym_exp] = ACTIONS(3137),
    [anon_sym_each] = ACTIONS(3137),
    [anon_sym_fc] = ACTIONS(3137),
    [anon_sym_fileno] = ACTIONS(3137),
    [anon_sym_gmtime] = ACTIONS(3137),
    [anon_sym_getc] = ACTIONS(3137),
    [anon_sym_getpgrp] = ACTIONS(3137),
    [anon_sym_getprotobyname] = ACTIONS(3137),
    [anon_sym_getpwname] = ACTIONS(3137),
    [anon_sym_getpwuid] = ACTIONS(3137),
    [anon_sym_getpeername] = ACTIONS(3137),
    [anon_sym_getnetbyname] = ACTIONS(3137),
    [anon_sym_getsockname] = ACTIONS(3137),
    [anon_sym_getgrnam] = ACTIONS(3137),
    [anon_sym_getgrgid] = ACTIONS(3137),
    [anon_sym_hex] = ACTIONS(3137),
    [anon_sym_int] = ACTIONS(3137),
    [anon_sym_keys] = ACTIONS(3137),
    [anon_sym_lc] = ACTIONS(3137),
    [anon_sym_lcfirst] = ACTIONS(3137),
    [anon_sym_length] = ACTIONS(3137),
    [anon_sym_localtime] = ACTIONS(3137),
    [anon_sym_log] = ACTIONS(3137),
    [anon_sym_lock] = ACTIONS(3137),
    [anon_sym_lstat] = ACTIONS(3137),
    [anon_sym_oct] = ACTIONS(3137),
    [anon_sym_ord] = ACTIONS(3137),
    [anon_sym_prototype] = ACTIONS(3137),
    [anon_sym_pop] = ACTIONS(3137),
    [anon_sym_pos] = ACTIONS(3137),
    [anon_sym_quotemeta] = ACTIONS(3137),
    [anon_sym_reset] = ACTIONS(3137),
    [anon_sym_rand] = ACTIONS(3137),
    [anon_sym_rmdir] = ACTIONS(3137),
    [anon_sym_readdir] = ACTIONS(3137),
    [anon_sym_readline] = ACTIONS(3137),
    [anon_sym_readpipe] = ACTIONS(3137),
    [anon_sym_rewinddir] = ACTIONS(3137),
    [anon_sym_readlink] = ACTIONS(3137),
    [anon_sym_ref] = ACTIONS(3137),
    [anon_sym_scalar] = ACTIONS(3137),
    [anon_sym_shift] = ACTIONS(3137),
    [anon_sym_sin] = ACTIONS(3137),
    [anon_sym_sleep] = ACTIONS(3137),
    [anon_sym_sqrt] = ACTIONS(3137),
    [anon_sym_srand] = ACTIONS(3137),
    [anon_sym_stat] = ACTIONS(3137),
    [anon_sym_study] = ACTIONS(3137),
    [anon_sym_tell] = ACTIONS(3137),
    [anon_sym_telldir] = ACTIONS(3137),
    [anon_sym_tied] = ACTIONS(3137),
    [anon_sym_uc] = ACTIONS(3137),
    [anon_sym_ucfirst] = ACTIONS(3137),
    [anon_sym_untie] = ACTIONS(3137),
    [anon_sym_umask] = ACTIONS(3137),
    [anon_sym_values] = ACTIONS(3137),
    [anon_sym_write] = ACTIONS(3137),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3137),
    [anon_sym___END__] = ACTIONS(3137),
    [anon_sym_EOT] = ACTIONS(3137),
    [anon_sym_q] = ACTIONS(3137),
    [anon_sym_qq] = ACTIONS(3137),
    [anon_sym_qw] = ACTIONS(3137),
    [anon_sym_qx] = ACTIONS(3137),
    [anon_sym_qr] = ACTIONS(3137),
    [anon_sym_m] = ACTIONS(3137),
    [anon_sym_s] = ACTIONS(3137),
    [anon_sym_tr] = ACTIONS(3137),
    [anon_sym_y] = ACTIONS(3137),
    [anon_sym_if] = ACTIONS(3137),
    [anon_sym_unless] = ACTIONS(3137),
    [anon_sym_while] = ACTIONS(3137),
    [anon_sym_until] = ACTIONS(3137),
    [aux_sym__bareword_token1] = ACTIONS(3137),
    [sym_number] = ACTIONS(3137),
    [anon_sym_true] = ACTIONS(3137),
    [anon_sym_false] = ACTIONS(3137),
    [sym__single_quote] = ACTIONS(3135),
    [sym__double_quote] = ACTIONS(3135),
    [sym__backtick_quote] = ACTIONS(3135),
    [sym__search_slash_quote] = ACTIONS(3135),
    [sym__open_readline_bracket] = ACTIONS(3135),
    [sym__open_fileglob_bracket] = ACTIONS(3135),
    [sym__PERLY_HEREDOC] = ACTIONS(3135),
    [sym__ctrl_z_hack] = ACTIONS(3135),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3135),
    [sym__filetest] = ACTIONS(3135),
  },
  [STATE(1496)] = {
    [sym_heredoc_content] = STATE(1496),
    [ts_builtin_sym_end] = ACTIONS(3139),
    [sym__identifier] = ACTIONS(3141),
    [anon_sym_LBRACE] = ACTIONS(3141),
    [anon_sym_SEMI] = ACTIONS(3141),
    [anon_sym_package] = ACTIONS(3141),
    [anon_sym_class] = ACTIONS(3141),
    [anon_sym_role] = ACTIONS(3141),
    [anon_sym_BUILD] = ACTIONS(3141),
    [anon_sym_ADJUST] = ACTIONS(3141),
    [anon_sym_DOLLAR] = ACTIONS(3141),
    [anon_sym_AT] = ACTIONS(3141),
    [anon_sym_my] = ACTIONS(3141),
    [anon_sym_extended] = ACTIONS(3141),
    [anon_sym_async] = ACTIONS(3141),
    [anon_sym_sub] = ACTIONS(3141),
    [anon_sym_method] = ACTIONS(3141),
    [anon_sym_LPAREN] = ACTIONS(3141),
    [anon_sym_continue] = ACTIONS(3141),
    [anon_sym_state] = ACTIONS(3141),
    [anon_sym_our] = ACTIONS(3141),
    [anon_sym_try] = ACTIONS(3141),
    [anon_sym_defer] = ACTIONS(3141),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3141),
    [anon_sym_LBRACK] = ACTIONS(3141),
    [anon_sym_LT_LT] = ACTIONS(3141),
    [anon_sym_SLASH_SLASH] = ACTIONS(3141),
    [anon_sym_PLUS] = ACTIONS(3141),
    [anon_sym_DASH] = ACTIONS(3141),
    [anon_sym_TILDE] = ACTIONS(3141),
    [anon_sym_BANG] = ACTIONS(3141),
    [anon_sym_PLUS_PLUS] = ACTIONS(3141),
    [anon_sym_DASH_DASH] = ACTIONS(3141),
    [anon_sym_BSLASH] = ACTIONS(3139),
    [anon_sym_do] = ACTIONS(3141),
    [anon_sym_eval] = ACTIONS(3141),
    [anon_sym_await] = ACTIONS(3141),
    [anon_sym_field] = ACTIONS(3141),
    [anon_sym_local] = ACTIONS(3141),
    [anon_sym_dynamically] = ACTIONS(3141),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3141),
    [anon_sym_require] = ACTIONS(3141),
    [anon_sym_map] = ACTIONS(3141),
    [anon_sym_grep] = ACTIONS(3141),
    [anon_sym_sort] = ACTIONS(3141),
    [anon_sym_goto] = ACTIONS(3141),
    [anon_sym_return] = ACTIONS(3141),
    [anon_sym_undef] = ACTIONS(3141),
    [anon_sym_PERCENT2] = ACTIONS(3141),
    [anon_sym_AMP2] = ACTIONS(3141),
    [anon_sym_STAR2] = ACTIONS(3141),
    [anon_sym_use] = ACTIONS(3141),
    [anon_sym_no] = ACTIONS(3141),
    [anon_sym_for] = ACTIONS(3141),
    [anon_sym_foreach] = ACTIONS(3141),
    [anon_sym_last] = ACTIONS(3141),
    [anon_sym_next] = ACTIONS(3141),
    [anon_sym_redo] = ACTIONS(3141),
    [anon_sym_BEGIN] = ACTIONS(3141),
    [anon_sym_INIT] = ACTIONS(3141),
    [anon_sym_CHECK] = ACTIONS(3141),
    [anon_sym_UNITCHECK] = ACTIONS(3141),
    [anon_sym_END] = ACTIONS(3141),
    [anon_sym___FILE__] = ACTIONS(3141),
    [anon_sym___LINE__] = ACTIONS(3141),
    [anon_sym___PACKAGE__] = ACTIONS(3141),
    [anon_sym___SUB__] = ACTIONS(3141),
    [anon_sym_break] = ACTIONS(3141),
    [anon_sym_fork] = ACTIONS(3141),
    [anon_sym_getppid] = ACTIONS(3141),
    [anon_sym_time] = ACTIONS(3141),
    [anon_sym_times] = ACTIONS(3141),
    [anon_sym_wait] = ACTIONS(3141),
    [anon_sym_wantarray] = ACTIONS(3141),
    [anon_sym_abs] = ACTIONS(3141),
    [anon_sym_alarm] = ACTIONS(3141),
    [anon_sym_chop] = ACTIONS(3141),
    [anon_sym_chdir] = ACTIONS(3141),
    [anon_sym_close] = ACTIONS(3141),
    [anon_sym_closedir] = ACTIONS(3141),
    [anon_sym_caller] = ACTIONS(3141),
    [anon_sym_chomp] = ACTIONS(3141),
    [anon_sym_chr] = ACTIONS(3141),
    [anon_sym_cos] = ACTIONS(3141),
    [anon_sym_chroot] = ACTIONS(3141),
    [anon_sym_defined] = ACTIONS(3141),
    [anon_sym_delete] = ACTIONS(3141),
    [anon_sym_dbmclose] = ACTIONS(3141),
    [anon_sym_exists] = ACTIONS(3141),
    [anon_sym_exit] = ACTIONS(3141),
    [anon_sym_eof] = ACTIONS(3141),
    [anon_sym_exp] = ACTIONS(3141),
    [anon_sym_each] = ACTIONS(3141),
    [anon_sym_fc] = ACTIONS(3141),
    [anon_sym_fileno] = ACTIONS(3141),
    [anon_sym_gmtime] = ACTIONS(3141),
    [anon_sym_getc] = ACTIONS(3141),
    [anon_sym_getpgrp] = ACTIONS(3141),
    [anon_sym_getprotobyname] = ACTIONS(3141),
    [anon_sym_getpwname] = ACTIONS(3141),
    [anon_sym_getpwuid] = ACTIONS(3141),
    [anon_sym_getpeername] = ACTIONS(3141),
    [anon_sym_getnetbyname] = ACTIONS(3141),
    [anon_sym_getsockname] = ACTIONS(3141),
    [anon_sym_getgrnam] = ACTIONS(3141),
    [anon_sym_getgrgid] = ACTIONS(3141),
    [anon_sym_hex] = ACTIONS(3141),
    [anon_sym_int] = ACTIONS(3141),
    [anon_sym_keys] = ACTIONS(3141),
    [anon_sym_lc] = ACTIONS(3141),
    [anon_sym_lcfirst] = ACTIONS(3141),
    [anon_sym_length] = ACTIONS(3141),
    [anon_sym_localtime] = ACTIONS(3141),
    [anon_sym_log] = ACTIONS(3141),
    [anon_sym_lock] = ACTIONS(3141),
    [anon_sym_lstat] = ACTIONS(3141),
    [anon_sym_oct] = ACTIONS(3141),
    [anon_sym_ord] = ACTIONS(3141),
    [anon_sym_prototype] = ACTIONS(3141),
    [anon_sym_pop] = ACTIONS(3141),
    [anon_sym_pos] = ACTIONS(3141),
    [anon_sym_quotemeta] = ACTIONS(3141),
    [anon_sym_reset] = ACTIONS(3141),
    [anon_sym_rand] = ACTIONS(3141),
    [anon_sym_rmdir] = ACTIONS(3141),
    [anon_sym_readdir] = ACTIONS(3141),
    [anon_sym_readline] = ACTIONS(3141),
    [anon_sym_readpipe] = ACTIONS(3141),
    [anon_sym_rewinddir] = ACTIONS(3141),
    [anon_sym_readlink] = ACTIONS(3141),
    [anon_sym_ref] = ACTIONS(3141),
    [anon_sym_scalar] = ACTIONS(3141),
    [anon_sym_shift] = ACTIONS(3141),
    [anon_sym_sin] = ACTIONS(3141),
    [anon_sym_sleep] = ACTIONS(3141),
    [anon_sym_sqrt] = ACTIONS(3141),
    [anon_sym_srand] = ACTIONS(3141),
    [anon_sym_stat] = ACTIONS(3141),
    [anon_sym_study] = ACTIONS(3141),
    [anon_sym_tell] = ACTIONS(3141),
    [anon_sym_telldir] = ACTIONS(3141),
    [anon_sym_tied] = ACTIONS(3141),
    [anon_sym_uc] = ACTIONS(3141),
    [anon_sym_ucfirst] = ACTIONS(3141),
    [anon_sym_untie] = ACTIONS(3141),
    [anon_sym_umask] = ACTIONS(3141),
    [anon_sym_values] = ACTIONS(3141),
    [anon_sym_write] = ACTIONS(3141),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3141),
    [anon_sym___END__] = ACTIONS(3141),
    [anon_sym_EOT] = ACTIONS(3141),
    [anon_sym_q] = ACTIONS(3141),
    [anon_sym_qq] = ACTIONS(3141),
    [anon_sym_qw] = ACTIONS(3141),
    [anon_sym_qx] = ACTIONS(3141),
    [anon_sym_qr] = ACTIONS(3141),
    [anon_sym_m] = ACTIONS(3141),
    [anon_sym_s] = ACTIONS(3141),
    [anon_sym_tr] = ACTIONS(3141),
    [anon_sym_y] = ACTIONS(3141),
    [anon_sym_if] = ACTIONS(3141),
    [anon_sym_unless] = ACTIONS(3141),
    [anon_sym_while] = ACTIONS(3141),
    [anon_sym_until] = ACTIONS(3141),
    [aux_sym__bareword_token1] = ACTIONS(3141),
    [sym_number] = ACTIONS(3141),
    [anon_sym_true] = ACTIONS(3141),
    [anon_sym_false] = ACTIONS(3141),
    [sym__single_quote] = ACTIONS(3139),
    [sym__double_quote] = ACTIONS(3139),
    [sym__backtick_quote] = ACTIONS(3139),
    [sym__search_slash_quote] = ACTIONS(3139),
    [sym__open_readline_bracket] = ACTIONS(3139),
    [sym__open_fileglob_bracket] = ACTIONS(3139),
    [sym__PERLY_HEREDOC] = ACTIONS(3139),
    [sym__ctrl_z_hack] = ACTIONS(3139),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3139),
    [sym__filetest] = ACTIONS(3139),
  },
  [STATE(1497)] = {
    [sym_heredoc_content] = STATE(1497),
    [ts_builtin_sym_end] = ACTIONS(3143),
    [sym__identifier] = ACTIONS(3145),
    [anon_sym_LBRACE] = ACTIONS(3145),
    [anon_sym_SEMI] = ACTIONS(3145),
    [anon_sym_package] = ACTIONS(3145),
    [anon_sym_class] = ACTIONS(3145),
    [anon_sym_role] = ACTIONS(3145),
    [anon_sym_BUILD] = ACTIONS(3145),
    [anon_sym_ADJUST] = ACTIONS(3145),
    [anon_sym_DOLLAR] = ACTIONS(3145),
    [anon_sym_AT] = ACTIONS(3145),
    [anon_sym_my] = ACTIONS(3145),
    [anon_sym_extended] = ACTIONS(3145),
    [anon_sym_async] = ACTIONS(3145),
    [anon_sym_sub] = ACTIONS(3145),
    [anon_sym_method] = ACTIONS(3145),
    [anon_sym_LPAREN] = ACTIONS(3145),
    [anon_sym_continue] = ACTIONS(3145),
    [anon_sym_state] = ACTIONS(3145),
    [anon_sym_our] = ACTIONS(3145),
    [anon_sym_try] = ACTIONS(3145),
    [anon_sym_defer] = ACTIONS(3145),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3145),
    [anon_sym_LBRACK] = ACTIONS(3145),
    [anon_sym_LT_LT] = ACTIONS(3145),
    [anon_sym_SLASH_SLASH] = ACTIONS(3145),
    [anon_sym_PLUS] = ACTIONS(3145),
    [anon_sym_DASH] = ACTIONS(3145),
    [anon_sym_TILDE] = ACTIONS(3145),
    [anon_sym_BANG] = ACTIONS(3145),
    [anon_sym_PLUS_PLUS] = ACTIONS(3145),
    [anon_sym_DASH_DASH] = ACTIONS(3145),
    [anon_sym_BSLASH] = ACTIONS(3143),
    [anon_sym_do] = ACTIONS(3145),
    [anon_sym_eval] = ACTIONS(3145),
    [anon_sym_await] = ACTIONS(3145),
    [anon_sym_field] = ACTIONS(3145),
    [anon_sym_local] = ACTIONS(3145),
    [anon_sym_dynamically] = ACTIONS(3145),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3145),
    [anon_sym_require] = ACTIONS(3145),
    [anon_sym_map] = ACTIONS(3145),
    [anon_sym_grep] = ACTIONS(3145),
    [anon_sym_sort] = ACTIONS(3145),
    [anon_sym_goto] = ACTIONS(3145),
    [anon_sym_return] = ACTIONS(3145),
    [anon_sym_undef] = ACTIONS(3145),
    [anon_sym_PERCENT2] = ACTIONS(3145),
    [anon_sym_AMP2] = ACTIONS(3145),
    [anon_sym_STAR2] = ACTIONS(3145),
    [anon_sym_use] = ACTIONS(3145),
    [anon_sym_no] = ACTIONS(3145),
    [anon_sym_for] = ACTIONS(3145),
    [anon_sym_foreach] = ACTIONS(3145),
    [anon_sym_last] = ACTIONS(3145),
    [anon_sym_next] = ACTIONS(3145),
    [anon_sym_redo] = ACTIONS(3145),
    [anon_sym_BEGIN] = ACTIONS(3145),
    [anon_sym_INIT] = ACTIONS(3145),
    [anon_sym_CHECK] = ACTIONS(3145),
    [anon_sym_UNITCHECK] = ACTIONS(3145),
    [anon_sym_END] = ACTIONS(3145),
    [anon_sym___FILE__] = ACTIONS(3145),
    [anon_sym___LINE__] = ACTIONS(3145),
    [anon_sym___PACKAGE__] = ACTIONS(3145),
    [anon_sym___SUB__] = ACTIONS(3145),
    [anon_sym_break] = ACTIONS(3145),
    [anon_sym_fork] = ACTIONS(3145),
    [anon_sym_getppid] = ACTIONS(3145),
    [anon_sym_time] = ACTIONS(3145),
    [anon_sym_times] = ACTIONS(3145),
    [anon_sym_wait] = ACTIONS(3145),
    [anon_sym_wantarray] = ACTIONS(3145),
    [anon_sym_abs] = ACTIONS(3145),
    [anon_sym_alarm] = ACTIONS(3145),
    [anon_sym_chop] = ACTIONS(3145),
    [anon_sym_chdir] = ACTIONS(3145),
    [anon_sym_close] = ACTIONS(3145),
    [anon_sym_closedir] = ACTIONS(3145),
    [anon_sym_caller] = ACTIONS(3145),
    [anon_sym_chomp] = ACTIONS(3145),
    [anon_sym_chr] = ACTIONS(3145),
    [anon_sym_cos] = ACTIONS(3145),
    [anon_sym_chroot] = ACTIONS(3145),
    [anon_sym_defined] = ACTIONS(3145),
    [anon_sym_delete] = ACTIONS(3145),
    [anon_sym_dbmclose] = ACTIONS(3145),
    [anon_sym_exists] = ACTIONS(3145),
    [anon_sym_exit] = ACTIONS(3145),
    [anon_sym_eof] = ACTIONS(3145),
    [anon_sym_exp] = ACTIONS(3145),
    [anon_sym_each] = ACTIONS(3145),
    [anon_sym_fc] = ACTIONS(3145),
    [anon_sym_fileno] = ACTIONS(3145),
    [anon_sym_gmtime] = ACTIONS(3145),
    [anon_sym_getc] = ACTIONS(3145),
    [anon_sym_getpgrp] = ACTIONS(3145),
    [anon_sym_getprotobyname] = ACTIONS(3145),
    [anon_sym_getpwname] = ACTIONS(3145),
    [anon_sym_getpwuid] = ACTIONS(3145),
    [anon_sym_getpeername] = ACTIONS(3145),
    [anon_sym_getnetbyname] = ACTIONS(3145),
    [anon_sym_getsockname] = ACTIONS(3145),
    [anon_sym_getgrnam] = ACTIONS(3145),
    [anon_sym_getgrgid] = ACTIONS(3145),
    [anon_sym_hex] = ACTIONS(3145),
    [anon_sym_int] = ACTIONS(3145),
    [anon_sym_keys] = ACTIONS(3145),
    [anon_sym_lc] = ACTIONS(3145),
    [anon_sym_lcfirst] = ACTIONS(3145),
    [anon_sym_length] = ACTIONS(3145),
    [anon_sym_localtime] = ACTIONS(3145),
    [anon_sym_log] = ACTIONS(3145),
    [anon_sym_lock] = ACTIONS(3145),
    [anon_sym_lstat] = ACTIONS(3145),
    [anon_sym_oct] = ACTIONS(3145),
    [anon_sym_ord] = ACTIONS(3145),
    [anon_sym_prototype] = ACTIONS(3145),
    [anon_sym_pop] = ACTIONS(3145),
    [anon_sym_pos] = ACTIONS(3145),
    [anon_sym_quotemeta] = ACTIONS(3145),
    [anon_sym_reset] = ACTIONS(3145),
    [anon_sym_rand] = ACTIONS(3145),
    [anon_sym_rmdir] = ACTIONS(3145),
    [anon_sym_readdir] = ACTIONS(3145),
    [anon_sym_readline] = ACTIONS(3145),
    [anon_sym_readpipe] = ACTIONS(3145),
    [anon_sym_rewinddir] = ACTIONS(3145),
    [anon_sym_readlink] = ACTIONS(3145),
    [anon_sym_ref] = ACTIONS(3145),
    [anon_sym_scalar] = ACTIONS(3145),
    [anon_sym_shift] = ACTIONS(3145),
    [anon_sym_sin] = ACTIONS(3145),
    [anon_sym_sleep] = ACTIONS(3145),
    [anon_sym_sqrt] = ACTIONS(3145),
    [anon_sym_srand] = ACTIONS(3145),
    [anon_sym_stat] = ACTIONS(3145),
    [anon_sym_study] = ACTIONS(3145),
    [anon_sym_tell] = ACTIONS(3145),
    [anon_sym_telldir] = ACTIONS(3145),
    [anon_sym_tied] = ACTIONS(3145),
    [anon_sym_uc] = ACTIONS(3145),
    [anon_sym_ucfirst] = ACTIONS(3145),
    [anon_sym_untie] = ACTIONS(3145),
    [anon_sym_umask] = ACTIONS(3145),
    [anon_sym_values] = ACTIONS(3145),
    [anon_sym_write] = ACTIONS(3145),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3145),
    [anon_sym___END__] = ACTIONS(3145),
    [anon_sym_EOT] = ACTIONS(3145),
    [anon_sym_q] = ACTIONS(3145),
    [anon_sym_qq] = ACTIONS(3145),
    [anon_sym_qw] = ACTIONS(3145),
    [anon_sym_qx] = ACTIONS(3145),
    [anon_sym_qr] = ACTIONS(3145),
    [anon_sym_m] = ACTIONS(3145),
    [anon_sym_s] = ACTIONS(3145),
    [anon_sym_tr] = ACTIONS(3145),
    [anon_sym_y] = ACTIONS(3145),
    [anon_sym_if] = ACTIONS(3145),
    [anon_sym_unless] = ACTIONS(3145),
    [anon_sym_while] = ACTIONS(3145),
    [anon_sym_until] = ACTIONS(3145),
    [aux_sym__bareword_token1] = ACTIONS(3145),
    [sym_number] = ACTIONS(3145),
    [anon_sym_true] = ACTIONS(3145),
    [anon_sym_false] = ACTIONS(3145),
    [sym__single_quote] = ACTIONS(3143),
    [sym__double_quote] = ACTIONS(3143),
    [sym__backtick_quote] = ACTIONS(3143),
    [sym__search_slash_quote] = ACTIONS(3143),
    [sym__open_readline_bracket] = ACTIONS(3143),
    [sym__open_fileglob_bracket] = ACTIONS(3143),
    [sym__PERLY_HEREDOC] = ACTIONS(3143),
    [sym__ctrl_z_hack] = ACTIONS(3143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3143),
    [sym__filetest] = ACTIONS(3143),
  },
  [STATE(1498)] = {
    [sym_heredoc_content] = STATE(1498),
    [ts_builtin_sym_end] = ACTIONS(3147),
    [sym__identifier] = ACTIONS(3149),
    [anon_sym_LBRACE] = ACTIONS(3149),
    [anon_sym_SEMI] = ACTIONS(3149),
    [anon_sym_package] = ACTIONS(3149),
    [anon_sym_class] = ACTIONS(3149),
    [anon_sym_role] = ACTIONS(3149),
    [anon_sym_BUILD] = ACTIONS(3149),
    [anon_sym_ADJUST] = ACTIONS(3149),
    [anon_sym_DOLLAR] = ACTIONS(3149),
    [anon_sym_AT] = ACTIONS(3149),
    [anon_sym_my] = ACTIONS(3149),
    [anon_sym_extended] = ACTIONS(3149),
    [anon_sym_async] = ACTIONS(3149),
    [anon_sym_sub] = ACTIONS(3149),
    [anon_sym_method] = ACTIONS(3149),
    [anon_sym_LPAREN] = ACTIONS(3149),
    [anon_sym_continue] = ACTIONS(3149),
    [anon_sym_state] = ACTIONS(3149),
    [anon_sym_our] = ACTIONS(3149),
    [anon_sym_try] = ACTIONS(3149),
    [anon_sym_defer] = ACTIONS(3149),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3149),
    [anon_sym_LBRACK] = ACTIONS(3149),
    [anon_sym_LT_LT] = ACTIONS(3149),
    [anon_sym_SLASH_SLASH] = ACTIONS(3149),
    [anon_sym_PLUS] = ACTIONS(3149),
    [anon_sym_DASH] = ACTIONS(3149),
    [anon_sym_TILDE] = ACTIONS(3149),
    [anon_sym_BANG] = ACTIONS(3149),
    [anon_sym_PLUS_PLUS] = ACTIONS(3149),
    [anon_sym_DASH_DASH] = ACTIONS(3149),
    [anon_sym_BSLASH] = ACTIONS(3147),
    [anon_sym_do] = ACTIONS(3149),
    [anon_sym_eval] = ACTIONS(3149),
    [anon_sym_await] = ACTIONS(3149),
    [anon_sym_field] = ACTIONS(3149),
    [anon_sym_local] = ACTIONS(3149),
    [anon_sym_dynamically] = ACTIONS(3149),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3149),
    [anon_sym_require] = ACTIONS(3149),
    [anon_sym_map] = ACTIONS(3149),
    [anon_sym_grep] = ACTIONS(3149),
    [anon_sym_sort] = ACTIONS(3149),
    [anon_sym_goto] = ACTIONS(3149),
    [anon_sym_return] = ACTIONS(3149),
    [anon_sym_undef] = ACTIONS(3149),
    [anon_sym_PERCENT2] = ACTIONS(3149),
    [anon_sym_AMP2] = ACTIONS(3149),
    [anon_sym_STAR2] = ACTIONS(3149),
    [anon_sym_use] = ACTIONS(3149),
    [anon_sym_no] = ACTIONS(3149),
    [anon_sym_for] = ACTIONS(3149),
    [anon_sym_foreach] = ACTIONS(3149),
    [anon_sym_last] = ACTIONS(3149),
    [anon_sym_next] = ACTIONS(3149),
    [anon_sym_redo] = ACTIONS(3149),
    [anon_sym_BEGIN] = ACTIONS(3149),
    [anon_sym_INIT] = ACTIONS(3149),
    [anon_sym_CHECK] = ACTIONS(3149),
    [anon_sym_UNITCHECK] = ACTIONS(3149),
    [anon_sym_END] = ACTIONS(3149),
    [anon_sym___FILE__] = ACTIONS(3149),
    [anon_sym___LINE__] = ACTIONS(3149),
    [anon_sym___PACKAGE__] = ACTIONS(3149),
    [anon_sym___SUB__] = ACTIONS(3149),
    [anon_sym_break] = ACTIONS(3149),
    [anon_sym_fork] = ACTIONS(3149),
    [anon_sym_getppid] = ACTIONS(3149),
    [anon_sym_time] = ACTIONS(3149),
    [anon_sym_times] = ACTIONS(3149),
    [anon_sym_wait] = ACTIONS(3149),
    [anon_sym_wantarray] = ACTIONS(3149),
    [anon_sym_abs] = ACTIONS(3149),
    [anon_sym_alarm] = ACTIONS(3149),
    [anon_sym_chop] = ACTIONS(3149),
    [anon_sym_chdir] = ACTIONS(3149),
    [anon_sym_close] = ACTIONS(3149),
    [anon_sym_closedir] = ACTIONS(3149),
    [anon_sym_caller] = ACTIONS(3149),
    [anon_sym_chomp] = ACTIONS(3149),
    [anon_sym_chr] = ACTIONS(3149),
    [anon_sym_cos] = ACTIONS(3149),
    [anon_sym_chroot] = ACTIONS(3149),
    [anon_sym_defined] = ACTIONS(3149),
    [anon_sym_delete] = ACTIONS(3149),
    [anon_sym_dbmclose] = ACTIONS(3149),
    [anon_sym_exists] = ACTIONS(3149),
    [anon_sym_exit] = ACTIONS(3149),
    [anon_sym_eof] = ACTIONS(3149),
    [anon_sym_exp] = ACTIONS(3149),
    [anon_sym_each] = ACTIONS(3149),
    [anon_sym_fc] = ACTIONS(3149),
    [anon_sym_fileno] = ACTIONS(3149),
    [anon_sym_gmtime] = ACTIONS(3149),
    [anon_sym_getc] = ACTIONS(3149),
    [anon_sym_getpgrp] = ACTIONS(3149),
    [anon_sym_getprotobyname] = ACTIONS(3149),
    [anon_sym_getpwname] = ACTIONS(3149),
    [anon_sym_getpwuid] = ACTIONS(3149),
    [anon_sym_getpeername] = ACTIONS(3149),
    [anon_sym_getnetbyname] = ACTIONS(3149),
    [anon_sym_getsockname] = ACTIONS(3149),
    [anon_sym_getgrnam] = ACTIONS(3149),
    [anon_sym_getgrgid] = ACTIONS(3149),
    [anon_sym_hex] = ACTIONS(3149),
    [anon_sym_int] = ACTIONS(3149),
    [anon_sym_keys] = ACTIONS(3149),
    [anon_sym_lc] = ACTIONS(3149),
    [anon_sym_lcfirst] = ACTIONS(3149),
    [anon_sym_length] = ACTIONS(3149),
    [anon_sym_localtime] = ACTIONS(3149),
    [anon_sym_log] = ACTIONS(3149),
    [anon_sym_lock] = ACTIONS(3149),
    [anon_sym_lstat] = ACTIONS(3149),
    [anon_sym_oct] = ACTIONS(3149),
    [anon_sym_ord] = ACTIONS(3149),
    [anon_sym_prototype] = ACTIONS(3149),
    [anon_sym_pop] = ACTIONS(3149),
    [anon_sym_pos] = ACTIONS(3149),
    [anon_sym_quotemeta] = ACTIONS(3149),
    [anon_sym_reset] = ACTIONS(3149),
    [anon_sym_rand] = ACTIONS(3149),
    [anon_sym_rmdir] = ACTIONS(3149),
    [anon_sym_readdir] = ACTIONS(3149),
    [anon_sym_readline] = ACTIONS(3149),
    [anon_sym_readpipe] = ACTIONS(3149),
    [anon_sym_rewinddir] = ACTIONS(3149),
    [anon_sym_readlink] = ACTIONS(3149),
    [anon_sym_ref] = ACTIONS(3149),
    [anon_sym_scalar] = ACTIONS(3149),
    [anon_sym_shift] = ACTIONS(3149),
    [anon_sym_sin] = ACTIONS(3149),
    [anon_sym_sleep] = ACTIONS(3149),
    [anon_sym_sqrt] = ACTIONS(3149),
    [anon_sym_srand] = ACTIONS(3149),
    [anon_sym_stat] = ACTIONS(3149),
    [anon_sym_study] = ACTIONS(3149),
    [anon_sym_tell] = ACTIONS(3149),
    [anon_sym_telldir] = ACTIONS(3149),
    [anon_sym_tied] = ACTIONS(3149),
    [anon_sym_uc] = ACTIONS(3149),
    [anon_sym_ucfirst] = ACTIONS(3149),
    [anon_sym_untie] = ACTIONS(3149),
    [anon_sym_umask] = ACTIONS(3149),
    [anon_sym_values] = ACTIONS(3149),
    [anon_sym_write] = ACTIONS(3149),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3149),
    [anon_sym___END__] = ACTIONS(3149),
    [anon_sym_EOT] = ACTIONS(3149),
    [anon_sym_q] = ACTIONS(3149),
    [anon_sym_qq] = ACTIONS(3149),
    [anon_sym_qw] = ACTIONS(3149),
    [anon_sym_qx] = ACTIONS(3149),
    [anon_sym_qr] = ACTIONS(3149),
    [anon_sym_m] = ACTIONS(3149),
    [anon_sym_s] = ACTIONS(3149),
    [anon_sym_tr] = ACTIONS(3149),
    [anon_sym_y] = ACTIONS(3149),
    [anon_sym_if] = ACTIONS(3149),
    [anon_sym_unless] = ACTIONS(3149),
    [anon_sym_while] = ACTIONS(3149),
    [anon_sym_until] = ACTIONS(3149),
    [aux_sym__bareword_token1] = ACTIONS(3149),
    [sym_number] = ACTIONS(3149),
    [anon_sym_true] = ACTIONS(3149),
    [anon_sym_false] = ACTIONS(3149),
    [sym__single_quote] = ACTIONS(3147),
    [sym__double_quote] = ACTIONS(3147),
    [sym__backtick_quote] = ACTIONS(3147),
    [sym__search_slash_quote] = ACTIONS(3147),
    [sym__open_readline_bracket] = ACTIONS(3147),
    [sym__open_fileglob_bracket] = ACTIONS(3147),
    [sym__PERLY_HEREDOC] = ACTIONS(3147),
    [sym__ctrl_z_hack] = ACTIONS(3147),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3147),
    [sym__filetest] = ACTIONS(3147),
  },
  [STATE(1499)] = {
    [sym_heredoc_content] = STATE(1499),
    [ts_builtin_sym_end] = ACTIONS(3151),
    [sym__identifier] = ACTIONS(3153),
    [anon_sym_LBRACE] = ACTIONS(3153),
    [anon_sym_SEMI] = ACTIONS(3153),
    [anon_sym_package] = ACTIONS(3153),
    [anon_sym_class] = ACTIONS(3153),
    [anon_sym_role] = ACTIONS(3153),
    [anon_sym_BUILD] = ACTIONS(3153),
    [anon_sym_ADJUST] = ACTIONS(3153),
    [anon_sym_DOLLAR] = ACTIONS(3153),
    [anon_sym_AT] = ACTIONS(3153),
    [anon_sym_my] = ACTIONS(3153),
    [anon_sym_extended] = ACTIONS(3153),
    [anon_sym_async] = ACTIONS(3153),
    [anon_sym_sub] = ACTIONS(3153),
    [anon_sym_method] = ACTIONS(3153),
    [anon_sym_LPAREN] = ACTIONS(3153),
    [anon_sym_continue] = ACTIONS(3153),
    [anon_sym_state] = ACTIONS(3153),
    [anon_sym_our] = ACTIONS(3153),
    [anon_sym_try] = ACTIONS(3153),
    [anon_sym_defer] = ACTIONS(3153),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3153),
    [anon_sym_LBRACK] = ACTIONS(3153),
    [anon_sym_LT_LT] = ACTIONS(3153),
    [anon_sym_SLASH_SLASH] = ACTIONS(3153),
    [anon_sym_PLUS] = ACTIONS(3153),
    [anon_sym_DASH] = ACTIONS(3153),
    [anon_sym_TILDE] = ACTIONS(3153),
    [anon_sym_BANG] = ACTIONS(3153),
    [anon_sym_PLUS_PLUS] = ACTIONS(3153),
    [anon_sym_DASH_DASH] = ACTIONS(3153),
    [anon_sym_BSLASH] = ACTIONS(3151),
    [anon_sym_do] = ACTIONS(3153),
    [anon_sym_eval] = ACTIONS(3153),
    [anon_sym_await] = ACTIONS(3153),
    [anon_sym_field] = ACTIONS(3153),
    [anon_sym_local] = ACTIONS(3153),
    [anon_sym_dynamically] = ACTIONS(3153),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3153),
    [anon_sym_require] = ACTIONS(3153),
    [anon_sym_map] = ACTIONS(3153),
    [anon_sym_grep] = ACTIONS(3153),
    [anon_sym_sort] = ACTIONS(3153),
    [anon_sym_goto] = ACTIONS(3153),
    [anon_sym_return] = ACTIONS(3153),
    [anon_sym_undef] = ACTIONS(3153),
    [anon_sym_PERCENT2] = ACTIONS(3153),
    [anon_sym_AMP2] = ACTIONS(3153),
    [anon_sym_STAR2] = ACTIONS(3153),
    [anon_sym_use] = ACTIONS(3153),
    [anon_sym_no] = ACTIONS(3153),
    [anon_sym_for] = ACTIONS(3153),
    [anon_sym_foreach] = ACTIONS(3153),
    [anon_sym_last] = ACTIONS(3153),
    [anon_sym_next] = ACTIONS(3153),
    [anon_sym_redo] = ACTIONS(3153),
    [anon_sym_BEGIN] = ACTIONS(3153),
    [anon_sym_INIT] = ACTIONS(3153),
    [anon_sym_CHECK] = ACTIONS(3153),
    [anon_sym_UNITCHECK] = ACTIONS(3153),
    [anon_sym_END] = ACTIONS(3153),
    [anon_sym___FILE__] = ACTIONS(3153),
    [anon_sym___LINE__] = ACTIONS(3153),
    [anon_sym___PACKAGE__] = ACTIONS(3153),
    [anon_sym___SUB__] = ACTIONS(3153),
    [anon_sym_break] = ACTIONS(3153),
    [anon_sym_fork] = ACTIONS(3153),
    [anon_sym_getppid] = ACTIONS(3153),
    [anon_sym_time] = ACTIONS(3153),
    [anon_sym_times] = ACTIONS(3153),
    [anon_sym_wait] = ACTIONS(3153),
    [anon_sym_wantarray] = ACTIONS(3153),
    [anon_sym_abs] = ACTIONS(3153),
    [anon_sym_alarm] = ACTIONS(3153),
    [anon_sym_chop] = ACTIONS(3153),
    [anon_sym_chdir] = ACTIONS(3153),
    [anon_sym_close] = ACTIONS(3153),
    [anon_sym_closedir] = ACTIONS(3153),
    [anon_sym_caller] = ACTIONS(3153),
    [anon_sym_chomp] = ACTIONS(3153),
    [anon_sym_chr] = ACTIONS(3153),
    [anon_sym_cos] = ACTIONS(3153),
    [anon_sym_chroot] = ACTIONS(3153),
    [anon_sym_defined] = ACTIONS(3153),
    [anon_sym_delete] = ACTIONS(3153),
    [anon_sym_dbmclose] = ACTIONS(3153),
    [anon_sym_exists] = ACTIONS(3153),
    [anon_sym_exit] = ACTIONS(3153),
    [anon_sym_eof] = ACTIONS(3153),
    [anon_sym_exp] = ACTIONS(3153),
    [anon_sym_each] = ACTIONS(3153),
    [anon_sym_fc] = ACTIONS(3153),
    [anon_sym_fileno] = ACTIONS(3153),
    [anon_sym_gmtime] = ACTIONS(3153),
    [anon_sym_getc] = ACTIONS(3153),
    [anon_sym_getpgrp] = ACTIONS(3153),
    [anon_sym_getprotobyname] = ACTIONS(3153),
    [anon_sym_getpwname] = ACTIONS(3153),
    [anon_sym_getpwuid] = ACTIONS(3153),
    [anon_sym_getpeername] = ACTIONS(3153),
    [anon_sym_getnetbyname] = ACTIONS(3153),
    [anon_sym_getsockname] = ACTIONS(3153),
    [anon_sym_getgrnam] = ACTIONS(3153),
    [anon_sym_getgrgid] = ACTIONS(3153),
    [anon_sym_hex] = ACTIONS(3153),
    [anon_sym_int] = ACTIONS(3153),
    [anon_sym_keys] = ACTIONS(3153),
    [anon_sym_lc] = ACTIONS(3153),
    [anon_sym_lcfirst] = ACTIONS(3153),
    [anon_sym_length] = ACTIONS(3153),
    [anon_sym_localtime] = ACTIONS(3153),
    [anon_sym_log] = ACTIONS(3153),
    [anon_sym_lock] = ACTIONS(3153),
    [anon_sym_lstat] = ACTIONS(3153),
    [anon_sym_oct] = ACTIONS(3153),
    [anon_sym_ord] = ACTIONS(3153),
    [anon_sym_prototype] = ACTIONS(3153),
    [anon_sym_pop] = ACTIONS(3153),
    [anon_sym_pos] = ACTIONS(3153),
    [anon_sym_quotemeta] = ACTIONS(3153),
    [anon_sym_reset] = ACTIONS(3153),
    [anon_sym_rand] = ACTIONS(3153),
    [anon_sym_rmdir] = ACTIONS(3153),
    [anon_sym_readdir] = ACTIONS(3153),
    [anon_sym_readline] = ACTIONS(3153),
    [anon_sym_readpipe] = ACTIONS(3153),
    [anon_sym_rewinddir] = ACTIONS(3153),
    [anon_sym_readlink] = ACTIONS(3153),
    [anon_sym_ref] = ACTIONS(3153),
    [anon_sym_scalar] = ACTIONS(3153),
    [anon_sym_shift] = ACTIONS(3153),
    [anon_sym_sin] = ACTIONS(3153),
    [anon_sym_sleep] = ACTIONS(3153),
    [anon_sym_sqrt] = ACTIONS(3153),
    [anon_sym_srand] = ACTIONS(3153),
    [anon_sym_stat] = ACTIONS(3153),
    [anon_sym_study] = ACTIONS(3153),
    [anon_sym_tell] = ACTIONS(3153),
    [anon_sym_telldir] = ACTIONS(3153),
    [anon_sym_tied] = ACTIONS(3153),
    [anon_sym_uc] = ACTIONS(3153),
    [anon_sym_ucfirst] = ACTIONS(3153),
    [anon_sym_untie] = ACTIONS(3153),
    [anon_sym_umask] = ACTIONS(3153),
    [anon_sym_values] = ACTIONS(3153),
    [anon_sym_write] = ACTIONS(3153),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3153),
    [anon_sym___END__] = ACTIONS(3153),
    [anon_sym_EOT] = ACTIONS(3153),
    [anon_sym_q] = ACTIONS(3153),
    [anon_sym_qq] = ACTIONS(3153),
    [anon_sym_qw] = ACTIONS(3153),
    [anon_sym_qx] = ACTIONS(3153),
    [anon_sym_qr] = ACTIONS(3153),
    [anon_sym_m] = ACTIONS(3153),
    [anon_sym_s] = ACTIONS(3153),
    [anon_sym_tr] = ACTIONS(3153),
    [anon_sym_y] = ACTIONS(3153),
    [anon_sym_if] = ACTIONS(3153),
    [anon_sym_unless] = ACTIONS(3153),
    [anon_sym_while] = ACTIONS(3153),
    [anon_sym_until] = ACTIONS(3153),
    [aux_sym__bareword_token1] = ACTIONS(3153),
    [sym_number] = ACTIONS(3153),
    [anon_sym_true] = ACTIONS(3153),
    [anon_sym_false] = ACTIONS(3153),
    [sym__single_quote] = ACTIONS(3151),
    [sym__double_quote] = ACTIONS(3151),
    [sym__backtick_quote] = ACTIONS(3151),
    [sym__search_slash_quote] = ACTIONS(3151),
    [sym__open_readline_bracket] = ACTIONS(3151),
    [sym__open_fileglob_bracket] = ACTIONS(3151),
    [sym__PERLY_HEREDOC] = ACTIONS(3151),
    [sym__ctrl_z_hack] = ACTIONS(3151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3151),
    [sym__filetest] = ACTIONS(3151),
  },
  [STATE(1500)] = {
    [sym_heredoc_content] = STATE(1500),
    [ts_builtin_sym_end] = ACTIONS(3155),
    [sym__identifier] = ACTIONS(3157),
    [anon_sym_LBRACE] = ACTIONS(3157),
    [anon_sym_SEMI] = ACTIONS(3157),
    [anon_sym_package] = ACTIONS(3157),
    [anon_sym_class] = ACTIONS(3157),
    [anon_sym_role] = ACTIONS(3157),
    [anon_sym_BUILD] = ACTIONS(3157),
    [anon_sym_ADJUST] = ACTIONS(3157),
    [anon_sym_DOLLAR] = ACTIONS(3157),
    [anon_sym_AT] = ACTIONS(3157),
    [anon_sym_my] = ACTIONS(3157),
    [anon_sym_extended] = ACTIONS(3157),
    [anon_sym_async] = ACTIONS(3157),
    [anon_sym_sub] = ACTIONS(3157),
    [anon_sym_method] = ACTIONS(3157),
    [anon_sym_LPAREN] = ACTIONS(3157),
    [anon_sym_continue] = ACTIONS(3157),
    [anon_sym_state] = ACTIONS(3157),
    [anon_sym_our] = ACTIONS(3157),
    [anon_sym_try] = ACTIONS(3157),
    [anon_sym_defer] = ACTIONS(3157),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3157),
    [anon_sym_LBRACK] = ACTIONS(3157),
    [anon_sym_LT_LT] = ACTIONS(3157),
    [anon_sym_SLASH_SLASH] = ACTIONS(3157),
    [anon_sym_PLUS] = ACTIONS(3157),
    [anon_sym_DASH] = ACTIONS(3157),
    [anon_sym_TILDE] = ACTIONS(3157),
    [anon_sym_BANG] = ACTIONS(3157),
    [anon_sym_PLUS_PLUS] = ACTIONS(3157),
    [anon_sym_DASH_DASH] = ACTIONS(3157),
    [anon_sym_BSLASH] = ACTIONS(3155),
    [anon_sym_do] = ACTIONS(3157),
    [anon_sym_eval] = ACTIONS(3157),
    [anon_sym_await] = ACTIONS(3157),
    [anon_sym_field] = ACTIONS(3157),
    [anon_sym_local] = ACTIONS(3157),
    [anon_sym_dynamically] = ACTIONS(3157),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3157),
    [anon_sym_require] = ACTIONS(3157),
    [anon_sym_map] = ACTIONS(3157),
    [anon_sym_grep] = ACTIONS(3157),
    [anon_sym_sort] = ACTIONS(3157),
    [anon_sym_goto] = ACTIONS(3157),
    [anon_sym_return] = ACTIONS(3157),
    [anon_sym_undef] = ACTIONS(3157),
    [anon_sym_PERCENT2] = ACTIONS(3157),
    [anon_sym_AMP2] = ACTIONS(3157),
    [anon_sym_STAR2] = ACTIONS(3157),
    [anon_sym_use] = ACTIONS(3157),
    [anon_sym_no] = ACTIONS(3157),
    [anon_sym_for] = ACTIONS(3157),
    [anon_sym_foreach] = ACTIONS(3157),
    [anon_sym_last] = ACTIONS(3157),
    [anon_sym_next] = ACTIONS(3157),
    [anon_sym_redo] = ACTIONS(3157),
    [anon_sym_BEGIN] = ACTIONS(3157),
    [anon_sym_INIT] = ACTIONS(3157),
    [anon_sym_CHECK] = ACTIONS(3157),
    [anon_sym_UNITCHECK] = ACTIONS(3157),
    [anon_sym_END] = ACTIONS(3157),
    [anon_sym___FILE__] = ACTIONS(3157),
    [anon_sym___LINE__] = ACTIONS(3157),
    [anon_sym___PACKAGE__] = ACTIONS(3157),
    [anon_sym___SUB__] = ACTIONS(3157),
    [anon_sym_break] = ACTIONS(3157),
    [anon_sym_fork] = ACTIONS(3157),
    [anon_sym_getppid] = ACTIONS(3157),
    [anon_sym_time] = ACTIONS(3157),
    [anon_sym_times] = ACTIONS(3157),
    [anon_sym_wait] = ACTIONS(3157),
    [anon_sym_wantarray] = ACTIONS(3157),
    [anon_sym_abs] = ACTIONS(3157),
    [anon_sym_alarm] = ACTIONS(3157),
    [anon_sym_chop] = ACTIONS(3157),
    [anon_sym_chdir] = ACTIONS(3157),
    [anon_sym_close] = ACTIONS(3157),
    [anon_sym_closedir] = ACTIONS(3157),
    [anon_sym_caller] = ACTIONS(3157),
    [anon_sym_chomp] = ACTIONS(3157),
    [anon_sym_chr] = ACTIONS(3157),
    [anon_sym_cos] = ACTIONS(3157),
    [anon_sym_chroot] = ACTIONS(3157),
    [anon_sym_defined] = ACTIONS(3157),
    [anon_sym_delete] = ACTIONS(3157),
    [anon_sym_dbmclose] = ACTIONS(3157),
    [anon_sym_exists] = ACTIONS(3157),
    [anon_sym_exit] = ACTIONS(3157),
    [anon_sym_eof] = ACTIONS(3157),
    [anon_sym_exp] = ACTIONS(3157),
    [anon_sym_each] = ACTIONS(3157),
    [anon_sym_fc] = ACTIONS(3157),
    [anon_sym_fileno] = ACTIONS(3157),
    [anon_sym_gmtime] = ACTIONS(3157),
    [anon_sym_getc] = ACTIONS(3157),
    [anon_sym_getpgrp] = ACTIONS(3157),
    [anon_sym_getprotobyname] = ACTIONS(3157),
    [anon_sym_getpwname] = ACTIONS(3157),
    [anon_sym_getpwuid] = ACTIONS(3157),
    [anon_sym_getpeername] = ACTIONS(3157),
    [anon_sym_getnetbyname] = ACTIONS(3157),
    [anon_sym_getsockname] = ACTIONS(3157),
    [anon_sym_getgrnam] = ACTIONS(3157),
    [anon_sym_getgrgid] = ACTIONS(3157),
    [anon_sym_hex] = ACTIONS(3157),
    [anon_sym_int] = ACTIONS(3157),
    [anon_sym_keys] = ACTIONS(3157),
    [anon_sym_lc] = ACTIONS(3157),
    [anon_sym_lcfirst] = ACTIONS(3157),
    [anon_sym_length] = ACTIONS(3157),
    [anon_sym_localtime] = ACTIONS(3157),
    [anon_sym_log] = ACTIONS(3157),
    [anon_sym_lock] = ACTIONS(3157),
    [anon_sym_lstat] = ACTIONS(3157),
    [anon_sym_oct] = ACTIONS(3157),
    [anon_sym_ord] = ACTIONS(3157),
    [anon_sym_prototype] = ACTIONS(3157),
    [anon_sym_pop] = ACTIONS(3157),
    [anon_sym_pos] = ACTIONS(3157),
    [anon_sym_quotemeta] = ACTIONS(3157),
    [anon_sym_reset] = ACTIONS(3157),
    [anon_sym_rand] = ACTIONS(3157),
    [anon_sym_rmdir] = ACTIONS(3157),
    [anon_sym_readdir] = ACTIONS(3157),
    [anon_sym_readline] = ACTIONS(3157),
    [anon_sym_readpipe] = ACTIONS(3157),
    [anon_sym_rewinddir] = ACTIONS(3157),
    [anon_sym_readlink] = ACTIONS(3157),
    [anon_sym_ref] = ACTIONS(3157),
    [anon_sym_scalar] = ACTIONS(3157),
    [anon_sym_shift] = ACTIONS(3157),
    [anon_sym_sin] = ACTIONS(3157),
    [anon_sym_sleep] = ACTIONS(3157),
    [anon_sym_sqrt] = ACTIONS(3157),
    [anon_sym_srand] = ACTIONS(3157),
    [anon_sym_stat] = ACTIONS(3157),
    [anon_sym_study] = ACTIONS(3157),
    [anon_sym_tell] = ACTIONS(3157),
    [anon_sym_telldir] = ACTIONS(3157),
    [anon_sym_tied] = ACTIONS(3157),
    [anon_sym_uc] = ACTIONS(3157),
    [anon_sym_ucfirst] = ACTIONS(3157),
    [anon_sym_untie] = ACTIONS(3157),
    [anon_sym_umask] = ACTIONS(3157),
    [anon_sym_values] = ACTIONS(3157),
    [anon_sym_write] = ACTIONS(3157),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3157),
    [anon_sym___END__] = ACTIONS(3157),
    [anon_sym_EOT] = ACTIONS(3157),
    [anon_sym_q] = ACTIONS(3157),
    [anon_sym_qq] = ACTIONS(3157),
    [anon_sym_qw] = ACTIONS(3157),
    [anon_sym_qx] = ACTIONS(3157),
    [anon_sym_qr] = ACTIONS(3157),
    [anon_sym_m] = ACTIONS(3157),
    [anon_sym_s] = ACTIONS(3157),
    [anon_sym_tr] = ACTIONS(3157),
    [anon_sym_y] = ACTIONS(3157),
    [anon_sym_if] = ACTIONS(3157),
    [anon_sym_unless] = ACTIONS(3157),
    [anon_sym_while] = ACTIONS(3157),
    [anon_sym_until] = ACTIONS(3157),
    [aux_sym__bareword_token1] = ACTIONS(3157),
    [sym_number] = ACTIONS(3157),
    [anon_sym_true] = ACTIONS(3157),
    [anon_sym_false] = ACTIONS(3157),
    [sym__single_quote] = ACTIONS(3155),
    [sym__double_quote] = ACTIONS(3155),
    [sym__backtick_quote] = ACTIONS(3155),
    [sym__search_slash_quote] = ACTIONS(3155),
    [sym__open_readline_bracket] = ACTIONS(3155),
    [sym__open_fileglob_bracket] = ACTIONS(3155),
    [sym__PERLY_HEREDOC] = ACTIONS(3155),
    [sym__ctrl_z_hack] = ACTIONS(3155),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3155),
    [sym__filetest] = ACTIONS(3155),
  },
  [STATE(1501)] = {
    [sym_heredoc_content] = STATE(1501),
    [ts_builtin_sym_end] = ACTIONS(3159),
    [sym__identifier] = ACTIONS(3161),
    [anon_sym_LBRACE] = ACTIONS(3161),
    [anon_sym_SEMI] = ACTIONS(3161),
    [anon_sym_package] = ACTIONS(3161),
    [anon_sym_class] = ACTIONS(3161),
    [anon_sym_role] = ACTIONS(3161),
    [anon_sym_BUILD] = ACTIONS(3161),
    [anon_sym_ADJUST] = ACTIONS(3161),
    [anon_sym_DOLLAR] = ACTIONS(3161),
    [anon_sym_AT] = ACTIONS(3161),
    [anon_sym_my] = ACTIONS(3161),
    [anon_sym_extended] = ACTIONS(3161),
    [anon_sym_async] = ACTIONS(3161),
    [anon_sym_sub] = ACTIONS(3161),
    [anon_sym_method] = ACTIONS(3161),
    [anon_sym_LPAREN] = ACTIONS(3161),
    [anon_sym_continue] = ACTIONS(3161),
    [anon_sym_state] = ACTIONS(3161),
    [anon_sym_our] = ACTIONS(3161),
    [anon_sym_try] = ACTIONS(3161),
    [anon_sym_defer] = ACTIONS(3161),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3161),
    [anon_sym_LBRACK] = ACTIONS(3161),
    [anon_sym_LT_LT] = ACTIONS(3161),
    [anon_sym_SLASH_SLASH] = ACTIONS(3161),
    [anon_sym_PLUS] = ACTIONS(3161),
    [anon_sym_DASH] = ACTIONS(3161),
    [anon_sym_TILDE] = ACTIONS(3161),
    [anon_sym_BANG] = ACTIONS(3161),
    [anon_sym_PLUS_PLUS] = ACTIONS(3161),
    [anon_sym_DASH_DASH] = ACTIONS(3161),
    [anon_sym_BSLASH] = ACTIONS(3159),
    [anon_sym_do] = ACTIONS(3161),
    [anon_sym_eval] = ACTIONS(3161),
    [anon_sym_await] = ACTIONS(3161),
    [anon_sym_field] = ACTIONS(3161),
    [anon_sym_local] = ACTIONS(3161),
    [anon_sym_dynamically] = ACTIONS(3161),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3161),
    [anon_sym_require] = ACTIONS(3161),
    [anon_sym_map] = ACTIONS(3161),
    [anon_sym_grep] = ACTIONS(3161),
    [anon_sym_sort] = ACTIONS(3161),
    [anon_sym_goto] = ACTIONS(3161),
    [anon_sym_return] = ACTIONS(3161),
    [anon_sym_undef] = ACTIONS(3161),
    [anon_sym_PERCENT2] = ACTIONS(3161),
    [anon_sym_AMP2] = ACTIONS(3161),
    [anon_sym_STAR2] = ACTIONS(3161),
    [anon_sym_use] = ACTIONS(3161),
    [anon_sym_no] = ACTIONS(3161),
    [anon_sym_for] = ACTIONS(3161),
    [anon_sym_foreach] = ACTIONS(3161),
    [anon_sym_last] = ACTIONS(3161),
    [anon_sym_next] = ACTIONS(3161),
    [anon_sym_redo] = ACTIONS(3161),
    [anon_sym_BEGIN] = ACTIONS(3161),
    [anon_sym_INIT] = ACTIONS(3161),
    [anon_sym_CHECK] = ACTIONS(3161),
    [anon_sym_UNITCHECK] = ACTIONS(3161),
    [anon_sym_END] = ACTIONS(3161),
    [anon_sym___FILE__] = ACTIONS(3161),
    [anon_sym___LINE__] = ACTIONS(3161),
    [anon_sym___PACKAGE__] = ACTIONS(3161),
    [anon_sym___SUB__] = ACTIONS(3161),
    [anon_sym_break] = ACTIONS(3161),
    [anon_sym_fork] = ACTIONS(3161),
    [anon_sym_getppid] = ACTIONS(3161),
    [anon_sym_time] = ACTIONS(3161),
    [anon_sym_times] = ACTIONS(3161),
    [anon_sym_wait] = ACTIONS(3161),
    [anon_sym_wantarray] = ACTIONS(3161),
    [anon_sym_abs] = ACTIONS(3161),
    [anon_sym_alarm] = ACTIONS(3161),
    [anon_sym_chop] = ACTIONS(3161),
    [anon_sym_chdir] = ACTIONS(3161),
    [anon_sym_close] = ACTIONS(3161),
    [anon_sym_closedir] = ACTIONS(3161),
    [anon_sym_caller] = ACTIONS(3161),
    [anon_sym_chomp] = ACTIONS(3161),
    [anon_sym_chr] = ACTIONS(3161),
    [anon_sym_cos] = ACTIONS(3161),
    [anon_sym_chroot] = ACTIONS(3161),
    [anon_sym_defined] = ACTIONS(3161),
    [anon_sym_delete] = ACTIONS(3161),
    [anon_sym_dbmclose] = ACTIONS(3161),
    [anon_sym_exists] = ACTIONS(3161),
    [anon_sym_exit] = ACTIONS(3161),
    [anon_sym_eof] = ACTIONS(3161),
    [anon_sym_exp] = ACTIONS(3161),
    [anon_sym_each] = ACTIONS(3161),
    [anon_sym_fc] = ACTIONS(3161),
    [anon_sym_fileno] = ACTIONS(3161),
    [anon_sym_gmtime] = ACTIONS(3161),
    [anon_sym_getc] = ACTIONS(3161),
    [anon_sym_getpgrp] = ACTIONS(3161),
    [anon_sym_getprotobyname] = ACTIONS(3161),
    [anon_sym_getpwname] = ACTIONS(3161),
    [anon_sym_getpwuid] = ACTIONS(3161),
    [anon_sym_getpeername] = ACTIONS(3161),
    [anon_sym_getnetbyname] = ACTIONS(3161),
    [anon_sym_getsockname] = ACTIONS(3161),
    [anon_sym_getgrnam] = ACTIONS(3161),
    [anon_sym_getgrgid] = ACTIONS(3161),
    [anon_sym_hex] = ACTIONS(3161),
    [anon_sym_int] = ACTIONS(3161),
    [anon_sym_keys] = ACTIONS(3161),
    [anon_sym_lc] = ACTIONS(3161),
    [anon_sym_lcfirst] = ACTIONS(3161),
    [anon_sym_length] = ACTIONS(3161),
    [anon_sym_localtime] = ACTIONS(3161),
    [anon_sym_log] = ACTIONS(3161),
    [anon_sym_lock] = ACTIONS(3161),
    [anon_sym_lstat] = ACTIONS(3161),
    [anon_sym_oct] = ACTIONS(3161),
    [anon_sym_ord] = ACTIONS(3161),
    [anon_sym_prototype] = ACTIONS(3161),
    [anon_sym_pop] = ACTIONS(3161),
    [anon_sym_pos] = ACTIONS(3161),
    [anon_sym_quotemeta] = ACTIONS(3161),
    [anon_sym_reset] = ACTIONS(3161),
    [anon_sym_rand] = ACTIONS(3161),
    [anon_sym_rmdir] = ACTIONS(3161),
    [anon_sym_readdir] = ACTIONS(3161),
    [anon_sym_readline] = ACTIONS(3161),
    [anon_sym_readpipe] = ACTIONS(3161),
    [anon_sym_rewinddir] = ACTIONS(3161),
    [anon_sym_readlink] = ACTIONS(3161),
    [anon_sym_ref] = ACTIONS(3161),
    [anon_sym_scalar] = ACTIONS(3161),
    [anon_sym_shift] = ACTIONS(3161),
    [anon_sym_sin] = ACTIONS(3161),
    [anon_sym_sleep] = ACTIONS(3161),
    [anon_sym_sqrt] = ACTIONS(3161),
    [anon_sym_srand] = ACTIONS(3161),
    [anon_sym_stat] = ACTIONS(3161),
    [anon_sym_study] = ACTIONS(3161),
    [anon_sym_tell] = ACTIONS(3161),
    [anon_sym_telldir] = ACTIONS(3161),
    [anon_sym_tied] = ACTIONS(3161),
    [anon_sym_uc] = ACTIONS(3161),
    [anon_sym_ucfirst] = ACTIONS(3161),
    [anon_sym_untie] = ACTIONS(3161),
    [anon_sym_umask] = ACTIONS(3161),
    [anon_sym_values] = ACTIONS(3161),
    [anon_sym_write] = ACTIONS(3161),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3161),
    [anon_sym___END__] = ACTIONS(3161),
    [anon_sym_EOT] = ACTIONS(3161),
    [anon_sym_q] = ACTIONS(3161),
    [anon_sym_qq] = ACTIONS(3161),
    [anon_sym_qw] = ACTIONS(3161),
    [anon_sym_qx] = ACTIONS(3161),
    [anon_sym_qr] = ACTIONS(3161),
    [anon_sym_m] = ACTIONS(3161),
    [anon_sym_s] = ACTIONS(3161),
    [anon_sym_tr] = ACTIONS(3161),
    [anon_sym_y] = ACTIONS(3161),
    [anon_sym_if] = ACTIONS(3161),
    [anon_sym_unless] = ACTIONS(3161),
    [anon_sym_while] = ACTIONS(3161),
    [anon_sym_until] = ACTIONS(3161),
    [aux_sym__bareword_token1] = ACTIONS(3161),
    [sym_number] = ACTIONS(3161),
    [anon_sym_true] = ACTIONS(3161),
    [anon_sym_false] = ACTIONS(3161),
    [sym__single_quote] = ACTIONS(3159),
    [sym__double_quote] = ACTIONS(3159),
    [sym__backtick_quote] = ACTIONS(3159),
    [sym__search_slash_quote] = ACTIONS(3159),
    [sym__open_readline_bracket] = ACTIONS(3159),
    [sym__open_fileglob_bracket] = ACTIONS(3159),
    [sym__PERLY_HEREDOC] = ACTIONS(3159),
    [sym__ctrl_z_hack] = ACTIONS(3159),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3159),
    [sym__filetest] = ACTIONS(3159),
  },
  [STATE(1502)] = {
    [sym_heredoc_content] = STATE(1502),
    [ts_builtin_sym_end] = ACTIONS(3163),
    [sym__identifier] = ACTIONS(3165),
    [anon_sym_LBRACE] = ACTIONS(3165),
    [anon_sym_SEMI] = ACTIONS(3165),
    [anon_sym_package] = ACTIONS(3165),
    [anon_sym_class] = ACTIONS(3165),
    [anon_sym_role] = ACTIONS(3165),
    [anon_sym_BUILD] = ACTIONS(3165),
    [anon_sym_ADJUST] = ACTIONS(3165),
    [anon_sym_DOLLAR] = ACTIONS(3165),
    [anon_sym_AT] = ACTIONS(3165),
    [anon_sym_my] = ACTIONS(3165),
    [anon_sym_extended] = ACTIONS(3165),
    [anon_sym_async] = ACTIONS(3165),
    [anon_sym_sub] = ACTIONS(3165),
    [anon_sym_method] = ACTIONS(3165),
    [anon_sym_LPAREN] = ACTIONS(3165),
    [anon_sym_continue] = ACTIONS(3165),
    [anon_sym_state] = ACTIONS(3165),
    [anon_sym_our] = ACTIONS(3165),
    [anon_sym_try] = ACTIONS(3165),
    [anon_sym_defer] = ACTIONS(3165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3165),
    [anon_sym_LBRACK] = ACTIONS(3165),
    [anon_sym_LT_LT] = ACTIONS(3165),
    [anon_sym_SLASH_SLASH] = ACTIONS(3165),
    [anon_sym_PLUS] = ACTIONS(3165),
    [anon_sym_DASH] = ACTIONS(3165),
    [anon_sym_TILDE] = ACTIONS(3165),
    [anon_sym_BANG] = ACTIONS(3165),
    [anon_sym_PLUS_PLUS] = ACTIONS(3165),
    [anon_sym_DASH_DASH] = ACTIONS(3165),
    [anon_sym_BSLASH] = ACTIONS(3163),
    [anon_sym_do] = ACTIONS(3165),
    [anon_sym_eval] = ACTIONS(3165),
    [anon_sym_await] = ACTIONS(3165),
    [anon_sym_field] = ACTIONS(3165),
    [anon_sym_local] = ACTIONS(3165),
    [anon_sym_dynamically] = ACTIONS(3165),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3165),
    [anon_sym_require] = ACTIONS(3165),
    [anon_sym_map] = ACTIONS(3165),
    [anon_sym_grep] = ACTIONS(3165),
    [anon_sym_sort] = ACTIONS(3165),
    [anon_sym_goto] = ACTIONS(3165),
    [anon_sym_return] = ACTIONS(3165),
    [anon_sym_undef] = ACTIONS(3165),
    [anon_sym_PERCENT2] = ACTIONS(3165),
    [anon_sym_AMP2] = ACTIONS(3165),
    [anon_sym_STAR2] = ACTIONS(3165),
    [anon_sym_use] = ACTIONS(3165),
    [anon_sym_no] = ACTIONS(3165),
    [anon_sym_for] = ACTIONS(3165),
    [anon_sym_foreach] = ACTIONS(3165),
    [anon_sym_last] = ACTIONS(3165),
    [anon_sym_next] = ACTIONS(3165),
    [anon_sym_redo] = ACTIONS(3165),
    [anon_sym_BEGIN] = ACTIONS(3165),
    [anon_sym_INIT] = ACTIONS(3165),
    [anon_sym_CHECK] = ACTIONS(3165),
    [anon_sym_UNITCHECK] = ACTIONS(3165),
    [anon_sym_END] = ACTIONS(3165),
    [anon_sym___FILE__] = ACTIONS(3165),
    [anon_sym___LINE__] = ACTIONS(3165),
    [anon_sym___PACKAGE__] = ACTIONS(3165),
    [anon_sym___SUB__] = ACTIONS(3165),
    [anon_sym_break] = ACTIONS(3165),
    [anon_sym_fork] = ACTIONS(3165),
    [anon_sym_getppid] = ACTIONS(3165),
    [anon_sym_time] = ACTIONS(3165),
    [anon_sym_times] = ACTIONS(3165),
    [anon_sym_wait] = ACTIONS(3165),
    [anon_sym_wantarray] = ACTIONS(3165),
    [anon_sym_abs] = ACTIONS(3165),
    [anon_sym_alarm] = ACTIONS(3165),
    [anon_sym_chop] = ACTIONS(3165),
    [anon_sym_chdir] = ACTIONS(3165),
    [anon_sym_close] = ACTIONS(3165),
    [anon_sym_closedir] = ACTIONS(3165),
    [anon_sym_caller] = ACTIONS(3165),
    [anon_sym_chomp] = ACTIONS(3165),
    [anon_sym_chr] = ACTIONS(3165),
    [anon_sym_cos] = ACTIONS(3165),
    [anon_sym_chroot] = ACTIONS(3165),
    [anon_sym_defined] = ACTIONS(3165),
    [anon_sym_delete] = ACTIONS(3165),
    [anon_sym_dbmclose] = ACTIONS(3165),
    [anon_sym_exists] = ACTIONS(3165),
    [anon_sym_exit] = ACTIONS(3165),
    [anon_sym_eof] = ACTIONS(3165),
    [anon_sym_exp] = ACTIONS(3165),
    [anon_sym_each] = ACTIONS(3165),
    [anon_sym_fc] = ACTIONS(3165),
    [anon_sym_fileno] = ACTIONS(3165),
    [anon_sym_gmtime] = ACTIONS(3165),
    [anon_sym_getc] = ACTIONS(3165),
    [anon_sym_getpgrp] = ACTIONS(3165),
    [anon_sym_getprotobyname] = ACTIONS(3165),
    [anon_sym_getpwname] = ACTIONS(3165),
    [anon_sym_getpwuid] = ACTIONS(3165),
    [anon_sym_getpeername] = ACTIONS(3165),
    [anon_sym_getnetbyname] = ACTIONS(3165),
    [anon_sym_getsockname] = ACTIONS(3165),
    [anon_sym_getgrnam] = ACTIONS(3165),
    [anon_sym_getgrgid] = ACTIONS(3165),
    [anon_sym_hex] = ACTIONS(3165),
    [anon_sym_int] = ACTIONS(3165),
    [anon_sym_keys] = ACTIONS(3165),
    [anon_sym_lc] = ACTIONS(3165),
    [anon_sym_lcfirst] = ACTIONS(3165),
    [anon_sym_length] = ACTIONS(3165),
    [anon_sym_localtime] = ACTIONS(3165),
    [anon_sym_log] = ACTIONS(3165),
    [anon_sym_lock] = ACTIONS(3165),
    [anon_sym_lstat] = ACTIONS(3165),
    [anon_sym_oct] = ACTIONS(3165),
    [anon_sym_ord] = ACTIONS(3165),
    [anon_sym_prototype] = ACTIONS(3165),
    [anon_sym_pop] = ACTIONS(3165),
    [anon_sym_pos] = ACTIONS(3165),
    [anon_sym_quotemeta] = ACTIONS(3165),
    [anon_sym_reset] = ACTIONS(3165),
    [anon_sym_rand] = ACTIONS(3165),
    [anon_sym_rmdir] = ACTIONS(3165),
    [anon_sym_readdir] = ACTIONS(3165),
    [anon_sym_readline] = ACTIONS(3165),
    [anon_sym_readpipe] = ACTIONS(3165),
    [anon_sym_rewinddir] = ACTIONS(3165),
    [anon_sym_readlink] = ACTIONS(3165),
    [anon_sym_ref] = ACTIONS(3165),
    [anon_sym_scalar] = ACTIONS(3165),
    [anon_sym_shift] = ACTIONS(3165),
    [anon_sym_sin] = ACTIONS(3165),
    [anon_sym_sleep] = ACTIONS(3165),
    [anon_sym_sqrt] = ACTIONS(3165),
    [anon_sym_srand] = ACTIONS(3165),
    [anon_sym_stat] = ACTIONS(3165),
    [anon_sym_study] = ACTIONS(3165),
    [anon_sym_tell] = ACTIONS(3165),
    [anon_sym_telldir] = ACTIONS(3165),
    [anon_sym_tied] = ACTIONS(3165),
    [anon_sym_uc] = ACTIONS(3165),
    [anon_sym_ucfirst] = ACTIONS(3165),
    [anon_sym_untie] = ACTIONS(3165),
    [anon_sym_umask] = ACTIONS(3165),
    [anon_sym_values] = ACTIONS(3165),
    [anon_sym_write] = ACTIONS(3165),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3165),
    [anon_sym___END__] = ACTIONS(3165),
    [anon_sym_EOT] = ACTIONS(3165),
    [anon_sym_q] = ACTIONS(3165),
    [anon_sym_qq] = ACTIONS(3165),
    [anon_sym_qw] = ACTIONS(3165),
    [anon_sym_qx] = ACTIONS(3165),
    [anon_sym_qr] = ACTIONS(3165),
    [anon_sym_m] = ACTIONS(3165),
    [anon_sym_s] = ACTIONS(3165),
    [anon_sym_tr] = ACTIONS(3165),
    [anon_sym_y] = ACTIONS(3165),
    [anon_sym_if] = ACTIONS(3165),
    [anon_sym_unless] = ACTIONS(3165),
    [anon_sym_while] = ACTIONS(3165),
    [anon_sym_until] = ACTIONS(3165),
    [aux_sym__bareword_token1] = ACTIONS(3165),
    [sym_number] = ACTIONS(3165),
    [anon_sym_true] = ACTIONS(3165),
    [anon_sym_false] = ACTIONS(3165),
    [sym__single_quote] = ACTIONS(3163),
    [sym__double_quote] = ACTIONS(3163),
    [sym__backtick_quote] = ACTIONS(3163),
    [sym__search_slash_quote] = ACTIONS(3163),
    [sym__open_readline_bracket] = ACTIONS(3163),
    [sym__open_fileglob_bracket] = ACTIONS(3163),
    [sym__PERLY_HEREDOC] = ACTIONS(3163),
    [sym__ctrl_z_hack] = ACTIONS(3163),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3163),
    [sym__filetest] = ACTIONS(3163),
  },
  [STATE(1503)] = {
    [sym_heredoc_content] = STATE(1503),
    [ts_builtin_sym_end] = ACTIONS(3167),
    [sym__identifier] = ACTIONS(3169),
    [anon_sym_LBRACE] = ACTIONS(3169),
    [anon_sym_SEMI] = ACTIONS(3169),
    [anon_sym_package] = ACTIONS(3169),
    [anon_sym_class] = ACTIONS(3169),
    [anon_sym_role] = ACTIONS(3169),
    [anon_sym_BUILD] = ACTIONS(3169),
    [anon_sym_ADJUST] = ACTIONS(3169),
    [anon_sym_DOLLAR] = ACTIONS(3169),
    [anon_sym_AT] = ACTIONS(3169),
    [anon_sym_my] = ACTIONS(3169),
    [anon_sym_extended] = ACTIONS(3169),
    [anon_sym_async] = ACTIONS(3169),
    [anon_sym_sub] = ACTIONS(3169),
    [anon_sym_method] = ACTIONS(3169),
    [anon_sym_LPAREN] = ACTIONS(3169),
    [anon_sym_continue] = ACTIONS(3169),
    [anon_sym_state] = ACTIONS(3169),
    [anon_sym_our] = ACTIONS(3169),
    [anon_sym_try] = ACTIONS(3169),
    [anon_sym_defer] = ACTIONS(3169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3169),
    [anon_sym_LBRACK] = ACTIONS(3169),
    [anon_sym_LT_LT] = ACTIONS(3169),
    [anon_sym_SLASH_SLASH] = ACTIONS(3169),
    [anon_sym_PLUS] = ACTIONS(3169),
    [anon_sym_DASH] = ACTIONS(3169),
    [anon_sym_TILDE] = ACTIONS(3169),
    [anon_sym_BANG] = ACTIONS(3169),
    [anon_sym_PLUS_PLUS] = ACTIONS(3169),
    [anon_sym_DASH_DASH] = ACTIONS(3169),
    [anon_sym_BSLASH] = ACTIONS(3167),
    [anon_sym_do] = ACTIONS(3169),
    [anon_sym_eval] = ACTIONS(3169),
    [anon_sym_await] = ACTIONS(3169),
    [anon_sym_field] = ACTIONS(3169),
    [anon_sym_local] = ACTIONS(3169),
    [anon_sym_dynamically] = ACTIONS(3169),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3169),
    [anon_sym_require] = ACTIONS(3169),
    [anon_sym_map] = ACTIONS(3169),
    [anon_sym_grep] = ACTIONS(3169),
    [anon_sym_sort] = ACTIONS(3169),
    [anon_sym_goto] = ACTIONS(3169),
    [anon_sym_return] = ACTIONS(3169),
    [anon_sym_undef] = ACTIONS(3169),
    [anon_sym_PERCENT2] = ACTIONS(3169),
    [anon_sym_AMP2] = ACTIONS(3169),
    [anon_sym_STAR2] = ACTIONS(3169),
    [anon_sym_use] = ACTIONS(3169),
    [anon_sym_no] = ACTIONS(3169),
    [anon_sym_for] = ACTIONS(3169),
    [anon_sym_foreach] = ACTIONS(3169),
    [anon_sym_last] = ACTIONS(3169),
    [anon_sym_next] = ACTIONS(3169),
    [anon_sym_redo] = ACTIONS(3169),
    [anon_sym_BEGIN] = ACTIONS(3169),
    [anon_sym_INIT] = ACTIONS(3169),
    [anon_sym_CHECK] = ACTIONS(3169),
    [anon_sym_UNITCHECK] = ACTIONS(3169),
    [anon_sym_END] = ACTIONS(3169),
    [anon_sym___FILE__] = ACTIONS(3169),
    [anon_sym___LINE__] = ACTIONS(3169),
    [anon_sym___PACKAGE__] = ACTIONS(3169),
    [anon_sym___SUB__] = ACTIONS(3169),
    [anon_sym_break] = ACTIONS(3169),
    [anon_sym_fork] = ACTIONS(3169),
    [anon_sym_getppid] = ACTIONS(3169),
    [anon_sym_time] = ACTIONS(3169),
    [anon_sym_times] = ACTIONS(3169),
    [anon_sym_wait] = ACTIONS(3169),
    [anon_sym_wantarray] = ACTIONS(3169),
    [anon_sym_abs] = ACTIONS(3169),
    [anon_sym_alarm] = ACTIONS(3169),
    [anon_sym_chop] = ACTIONS(3169),
    [anon_sym_chdir] = ACTIONS(3169),
    [anon_sym_close] = ACTIONS(3169),
    [anon_sym_closedir] = ACTIONS(3169),
    [anon_sym_caller] = ACTIONS(3169),
    [anon_sym_chomp] = ACTIONS(3169),
    [anon_sym_chr] = ACTIONS(3169),
    [anon_sym_cos] = ACTIONS(3169),
    [anon_sym_chroot] = ACTIONS(3169),
    [anon_sym_defined] = ACTIONS(3169),
    [anon_sym_delete] = ACTIONS(3169),
    [anon_sym_dbmclose] = ACTIONS(3169),
    [anon_sym_exists] = ACTIONS(3169),
    [anon_sym_exit] = ACTIONS(3169),
    [anon_sym_eof] = ACTIONS(3169),
    [anon_sym_exp] = ACTIONS(3169),
    [anon_sym_each] = ACTIONS(3169),
    [anon_sym_fc] = ACTIONS(3169),
    [anon_sym_fileno] = ACTIONS(3169),
    [anon_sym_gmtime] = ACTIONS(3169),
    [anon_sym_getc] = ACTIONS(3169),
    [anon_sym_getpgrp] = ACTIONS(3169),
    [anon_sym_getprotobyname] = ACTIONS(3169),
    [anon_sym_getpwname] = ACTIONS(3169),
    [anon_sym_getpwuid] = ACTIONS(3169),
    [anon_sym_getpeername] = ACTIONS(3169),
    [anon_sym_getnetbyname] = ACTIONS(3169),
    [anon_sym_getsockname] = ACTIONS(3169),
    [anon_sym_getgrnam] = ACTIONS(3169),
    [anon_sym_getgrgid] = ACTIONS(3169),
    [anon_sym_hex] = ACTIONS(3169),
    [anon_sym_int] = ACTIONS(3169),
    [anon_sym_keys] = ACTIONS(3169),
    [anon_sym_lc] = ACTIONS(3169),
    [anon_sym_lcfirst] = ACTIONS(3169),
    [anon_sym_length] = ACTIONS(3169),
    [anon_sym_localtime] = ACTIONS(3169),
    [anon_sym_log] = ACTIONS(3169),
    [anon_sym_lock] = ACTIONS(3169),
    [anon_sym_lstat] = ACTIONS(3169),
    [anon_sym_oct] = ACTIONS(3169),
    [anon_sym_ord] = ACTIONS(3169),
    [anon_sym_prototype] = ACTIONS(3169),
    [anon_sym_pop] = ACTIONS(3169),
    [anon_sym_pos] = ACTIONS(3169),
    [anon_sym_quotemeta] = ACTIONS(3169),
    [anon_sym_reset] = ACTIONS(3169),
    [anon_sym_rand] = ACTIONS(3169),
    [anon_sym_rmdir] = ACTIONS(3169),
    [anon_sym_readdir] = ACTIONS(3169),
    [anon_sym_readline] = ACTIONS(3169),
    [anon_sym_readpipe] = ACTIONS(3169),
    [anon_sym_rewinddir] = ACTIONS(3169),
    [anon_sym_readlink] = ACTIONS(3169),
    [anon_sym_ref] = ACTIONS(3169),
    [anon_sym_scalar] = ACTIONS(3169),
    [anon_sym_shift] = ACTIONS(3169),
    [anon_sym_sin] = ACTIONS(3169),
    [anon_sym_sleep] = ACTIONS(3169),
    [anon_sym_sqrt] = ACTIONS(3169),
    [anon_sym_srand] = ACTIONS(3169),
    [anon_sym_stat] = ACTIONS(3169),
    [anon_sym_study] = ACTIONS(3169),
    [anon_sym_tell] = ACTIONS(3169),
    [anon_sym_telldir] = ACTIONS(3169),
    [anon_sym_tied] = ACTIONS(3169),
    [anon_sym_uc] = ACTIONS(3169),
    [anon_sym_ucfirst] = ACTIONS(3169),
    [anon_sym_untie] = ACTIONS(3169),
    [anon_sym_umask] = ACTIONS(3169),
    [anon_sym_values] = ACTIONS(3169),
    [anon_sym_write] = ACTIONS(3169),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3169),
    [anon_sym___END__] = ACTIONS(3169),
    [anon_sym_EOT] = ACTIONS(3169),
    [anon_sym_q] = ACTIONS(3169),
    [anon_sym_qq] = ACTIONS(3169),
    [anon_sym_qw] = ACTIONS(3169),
    [anon_sym_qx] = ACTIONS(3169),
    [anon_sym_qr] = ACTIONS(3169),
    [anon_sym_m] = ACTIONS(3169),
    [anon_sym_s] = ACTIONS(3169),
    [anon_sym_tr] = ACTIONS(3169),
    [anon_sym_y] = ACTIONS(3169),
    [anon_sym_if] = ACTIONS(3169),
    [anon_sym_unless] = ACTIONS(3169),
    [anon_sym_while] = ACTIONS(3169),
    [anon_sym_until] = ACTIONS(3169),
    [aux_sym__bareword_token1] = ACTIONS(3169),
    [sym_number] = ACTIONS(3169),
    [anon_sym_true] = ACTIONS(3169),
    [anon_sym_false] = ACTIONS(3169),
    [sym__single_quote] = ACTIONS(3167),
    [sym__double_quote] = ACTIONS(3167),
    [sym__backtick_quote] = ACTIONS(3167),
    [sym__search_slash_quote] = ACTIONS(3167),
    [sym__open_readline_bracket] = ACTIONS(3167),
    [sym__open_fileglob_bracket] = ACTIONS(3167),
    [sym__PERLY_HEREDOC] = ACTIONS(3167),
    [sym__ctrl_z_hack] = ACTIONS(3167),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3167),
    [sym__filetest] = ACTIONS(3167),
  },
  [STATE(1504)] = {
    [sym_heredoc_content] = STATE(1504),
    [ts_builtin_sym_end] = ACTIONS(3171),
    [sym__identifier] = ACTIONS(3173),
    [anon_sym_LBRACE] = ACTIONS(3173),
    [anon_sym_SEMI] = ACTIONS(3173),
    [anon_sym_package] = ACTIONS(3173),
    [anon_sym_class] = ACTIONS(3173),
    [anon_sym_role] = ACTIONS(3173),
    [anon_sym_BUILD] = ACTIONS(3173),
    [anon_sym_ADJUST] = ACTIONS(3173),
    [anon_sym_DOLLAR] = ACTIONS(3173),
    [anon_sym_AT] = ACTIONS(3173),
    [anon_sym_my] = ACTIONS(3173),
    [anon_sym_extended] = ACTIONS(3173),
    [anon_sym_async] = ACTIONS(3173),
    [anon_sym_sub] = ACTIONS(3173),
    [anon_sym_method] = ACTIONS(3173),
    [anon_sym_LPAREN] = ACTIONS(3173),
    [anon_sym_continue] = ACTIONS(3173),
    [anon_sym_state] = ACTIONS(3173),
    [anon_sym_our] = ACTIONS(3173),
    [anon_sym_try] = ACTIONS(3173),
    [anon_sym_defer] = ACTIONS(3173),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3173),
    [anon_sym_LBRACK] = ACTIONS(3173),
    [anon_sym_LT_LT] = ACTIONS(3173),
    [anon_sym_SLASH_SLASH] = ACTIONS(3173),
    [anon_sym_PLUS] = ACTIONS(3173),
    [anon_sym_DASH] = ACTIONS(3173),
    [anon_sym_TILDE] = ACTIONS(3173),
    [anon_sym_BANG] = ACTIONS(3173),
    [anon_sym_PLUS_PLUS] = ACTIONS(3173),
    [anon_sym_DASH_DASH] = ACTIONS(3173),
    [anon_sym_BSLASH] = ACTIONS(3171),
    [anon_sym_do] = ACTIONS(3173),
    [anon_sym_eval] = ACTIONS(3173),
    [anon_sym_await] = ACTIONS(3173),
    [anon_sym_field] = ACTIONS(3173),
    [anon_sym_local] = ACTIONS(3173),
    [anon_sym_dynamically] = ACTIONS(3173),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3173),
    [anon_sym_require] = ACTIONS(3173),
    [anon_sym_map] = ACTIONS(3173),
    [anon_sym_grep] = ACTIONS(3173),
    [anon_sym_sort] = ACTIONS(3173),
    [anon_sym_goto] = ACTIONS(3173),
    [anon_sym_return] = ACTIONS(3173),
    [anon_sym_undef] = ACTIONS(3173),
    [anon_sym_PERCENT2] = ACTIONS(3173),
    [anon_sym_AMP2] = ACTIONS(3173),
    [anon_sym_STAR2] = ACTIONS(3173),
    [anon_sym_use] = ACTIONS(3173),
    [anon_sym_no] = ACTIONS(3173),
    [anon_sym_for] = ACTIONS(3173),
    [anon_sym_foreach] = ACTIONS(3173),
    [anon_sym_last] = ACTIONS(3173),
    [anon_sym_next] = ACTIONS(3173),
    [anon_sym_redo] = ACTIONS(3173),
    [anon_sym_BEGIN] = ACTIONS(3173),
    [anon_sym_INIT] = ACTIONS(3173),
    [anon_sym_CHECK] = ACTIONS(3173),
    [anon_sym_UNITCHECK] = ACTIONS(3173),
    [anon_sym_END] = ACTIONS(3173),
    [anon_sym___FILE__] = ACTIONS(3173),
    [anon_sym___LINE__] = ACTIONS(3173),
    [anon_sym___PACKAGE__] = ACTIONS(3173),
    [anon_sym___SUB__] = ACTIONS(3173),
    [anon_sym_break] = ACTIONS(3173),
    [anon_sym_fork] = ACTIONS(3173),
    [anon_sym_getppid] = ACTIONS(3173),
    [anon_sym_time] = ACTIONS(3173),
    [anon_sym_times] = ACTIONS(3173),
    [anon_sym_wait] = ACTIONS(3173),
    [anon_sym_wantarray] = ACTIONS(3173),
    [anon_sym_abs] = ACTIONS(3173),
    [anon_sym_alarm] = ACTIONS(3173),
    [anon_sym_chop] = ACTIONS(3173),
    [anon_sym_chdir] = ACTIONS(3173),
    [anon_sym_close] = ACTIONS(3173),
    [anon_sym_closedir] = ACTIONS(3173),
    [anon_sym_caller] = ACTIONS(3173),
    [anon_sym_chomp] = ACTIONS(3173),
    [anon_sym_chr] = ACTIONS(3173),
    [anon_sym_cos] = ACTIONS(3173),
    [anon_sym_chroot] = ACTIONS(3173),
    [anon_sym_defined] = ACTIONS(3173),
    [anon_sym_delete] = ACTIONS(3173),
    [anon_sym_dbmclose] = ACTIONS(3173),
    [anon_sym_exists] = ACTIONS(3173),
    [anon_sym_exit] = ACTIONS(3173),
    [anon_sym_eof] = ACTIONS(3173),
    [anon_sym_exp] = ACTIONS(3173),
    [anon_sym_each] = ACTIONS(3173),
    [anon_sym_fc] = ACTIONS(3173),
    [anon_sym_fileno] = ACTIONS(3173),
    [anon_sym_gmtime] = ACTIONS(3173),
    [anon_sym_getc] = ACTIONS(3173),
    [anon_sym_getpgrp] = ACTIONS(3173),
    [anon_sym_getprotobyname] = ACTIONS(3173),
    [anon_sym_getpwname] = ACTIONS(3173),
    [anon_sym_getpwuid] = ACTIONS(3173),
    [anon_sym_getpeername] = ACTIONS(3173),
    [anon_sym_getnetbyname] = ACTIONS(3173),
    [anon_sym_getsockname] = ACTIONS(3173),
    [anon_sym_getgrnam] = ACTIONS(3173),
    [anon_sym_getgrgid] = ACTIONS(3173),
    [anon_sym_hex] = ACTIONS(3173),
    [anon_sym_int] = ACTIONS(3173),
    [anon_sym_keys] = ACTIONS(3173),
    [anon_sym_lc] = ACTIONS(3173),
    [anon_sym_lcfirst] = ACTIONS(3173),
    [anon_sym_length] = ACTIONS(3173),
    [anon_sym_localtime] = ACTIONS(3173),
    [anon_sym_log] = ACTIONS(3173),
    [anon_sym_lock] = ACTIONS(3173),
    [anon_sym_lstat] = ACTIONS(3173),
    [anon_sym_oct] = ACTIONS(3173),
    [anon_sym_ord] = ACTIONS(3173),
    [anon_sym_prototype] = ACTIONS(3173),
    [anon_sym_pop] = ACTIONS(3173),
    [anon_sym_pos] = ACTIONS(3173),
    [anon_sym_quotemeta] = ACTIONS(3173),
    [anon_sym_reset] = ACTIONS(3173),
    [anon_sym_rand] = ACTIONS(3173),
    [anon_sym_rmdir] = ACTIONS(3173),
    [anon_sym_readdir] = ACTIONS(3173),
    [anon_sym_readline] = ACTIONS(3173),
    [anon_sym_readpipe] = ACTIONS(3173),
    [anon_sym_rewinddir] = ACTIONS(3173),
    [anon_sym_readlink] = ACTIONS(3173),
    [anon_sym_ref] = ACTIONS(3173),
    [anon_sym_scalar] = ACTIONS(3173),
    [anon_sym_shift] = ACTIONS(3173),
    [anon_sym_sin] = ACTIONS(3173),
    [anon_sym_sleep] = ACTIONS(3173),
    [anon_sym_sqrt] = ACTIONS(3173),
    [anon_sym_srand] = ACTIONS(3173),
    [anon_sym_stat] = ACTIONS(3173),
    [anon_sym_study] = ACTIONS(3173),
    [anon_sym_tell] = ACTIONS(3173),
    [anon_sym_telldir] = ACTIONS(3173),
    [anon_sym_tied] = ACTIONS(3173),
    [anon_sym_uc] = ACTIONS(3173),
    [anon_sym_ucfirst] = ACTIONS(3173),
    [anon_sym_untie] = ACTIONS(3173),
    [anon_sym_umask] = ACTIONS(3173),
    [anon_sym_values] = ACTIONS(3173),
    [anon_sym_write] = ACTIONS(3173),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3173),
    [anon_sym___END__] = ACTIONS(3173),
    [anon_sym_EOT] = ACTIONS(3173),
    [anon_sym_q] = ACTIONS(3173),
    [anon_sym_qq] = ACTIONS(3173),
    [anon_sym_qw] = ACTIONS(3173),
    [anon_sym_qx] = ACTIONS(3173),
    [anon_sym_qr] = ACTIONS(3173),
    [anon_sym_m] = ACTIONS(3173),
    [anon_sym_s] = ACTIONS(3173),
    [anon_sym_tr] = ACTIONS(3173),
    [anon_sym_y] = ACTIONS(3173),
    [anon_sym_if] = ACTIONS(3173),
    [anon_sym_unless] = ACTIONS(3173),
    [anon_sym_while] = ACTIONS(3173),
    [anon_sym_until] = ACTIONS(3173),
    [aux_sym__bareword_token1] = ACTIONS(3173),
    [sym_number] = ACTIONS(3173),
    [anon_sym_true] = ACTIONS(3173),
    [anon_sym_false] = ACTIONS(3173),
    [sym__single_quote] = ACTIONS(3171),
    [sym__double_quote] = ACTIONS(3171),
    [sym__backtick_quote] = ACTIONS(3171),
    [sym__search_slash_quote] = ACTIONS(3171),
    [sym__open_readline_bracket] = ACTIONS(3171),
    [sym__open_fileglob_bracket] = ACTIONS(3171),
    [sym__PERLY_HEREDOC] = ACTIONS(3171),
    [sym__ctrl_z_hack] = ACTIONS(3171),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3171),
    [sym__filetest] = ACTIONS(3171),
  },
  [STATE(1505)] = {
    [sym_heredoc_content] = STATE(1505),
    [ts_builtin_sym_end] = ACTIONS(3175),
    [sym__identifier] = ACTIONS(3177),
    [anon_sym_LBRACE] = ACTIONS(3177),
    [anon_sym_SEMI] = ACTIONS(3177),
    [anon_sym_package] = ACTIONS(3177),
    [anon_sym_class] = ACTIONS(3177),
    [anon_sym_role] = ACTIONS(3177),
    [anon_sym_BUILD] = ACTIONS(3177),
    [anon_sym_ADJUST] = ACTIONS(3177),
    [anon_sym_DOLLAR] = ACTIONS(3177),
    [anon_sym_AT] = ACTIONS(3177),
    [anon_sym_my] = ACTIONS(3177),
    [anon_sym_extended] = ACTIONS(3177),
    [anon_sym_async] = ACTIONS(3177),
    [anon_sym_sub] = ACTIONS(3177),
    [anon_sym_method] = ACTIONS(3177),
    [anon_sym_LPAREN] = ACTIONS(3177),
    [anon_sym_continue] = ACTIONS(3177),
    [anon_sym_state] = ACTIONS(3177),
    [anon_sym_our] = ACTIONS(3177),
    [anon_sym_try] = ACTIONS(3177),
    [anon_sym_defer] = ACTIONS(3177),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3177),
    [anon_sym_LBRACK] = ACTIONS(3177),
    [anon_sym_LT_LT] = ACTIONS(3177),
    [anon_sym_SLASH_SLASH] = ACTIONS(3177),
    [anon_sym_PLUS] = ACTIONS(3177),
    [anon_sym_DASH] = ACTIONS(3177),
    [anon_sym_TILDE] = ACTIONS(3177),
    [anon_sym_BANG] = ACTIONS(3177),
    [anon_sym_PLUS_PLUS] = ACTIONS(3177),
    [anon_sym_DASH_DASH] = ACTIONS(3177),
    [anon_sym_BSLASH] = ACTIONS(3175),
    [anon_sym_do] = ACTIONS(3177),
    [anon_sym_eval] = ACTIONS(3177),
    [anon_sym_await] = ACTIONS(3177),
    [anon_sym_field] = ACTIONS(3177),
    [anon_sym_local] = ACTIONS(3177),
    [anon_sym_dynamically] = ACTIONS(3177),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3177),
    [anon_sym_require] = ACTIONS(3177),
    [anon_sym_map] = ACTIONS(3177),
    [anon_sym_grep] = ACTIONS(3177),
    [anon_sym_sort] = ACTIONS(3177),
    [anon_sym_goto] = ACTIONS(3177),
    [anon_sym_return] = ACTIONS(3177),
    [anon_sym_undef] = ACTIONS(3177),
    [anon_sym_PERCENT2] = ACTIONS(3177),
    [anon_sym_AMP2] = ACTIONS(3177),
    [anon_sym_STAR2] = ACTIONS(3177),
    [anon_sym_use] = ACTIONS(3177),
    [anon_sym_no] = ACTIONS(3177),
    [anon_sym_for] = ACTIONS(3177),
    [anon_sym_foreach] = ACTIONS(3177),
    [anon_sym_last] = ACTIONS(3177),
    [anon_sym_next] = ACTIONS(3177),
    [anon_sym_redo] = ACTIONS(3177),
    [anon_sym_BEGIN] = ACTIONS(3177),
    [anon_sym_INIT] = ACTIONS(3177),
    [anon_sym_CHECK] = ACTIONS(3177),
    [anon_sym_UNITCHECK] = ACTIONS(3177),
    [anon_sym_END] = ACTIONS(3177),
    [anon_sym___FILE__] = ACTIONS(3177),
    [anon_sym___LINE__] = ACTIONS(3177),
    [anon_sym___PACKAGE__] = ACTIONS(3177),
    [anon_sym___SUB__] = ACTIONS(3177),
    [anon_sym_break] = ACTIONS(3177),
    [anon_sym_fork] = ACTIONS(3177),
    [anon_sym_getppid] = ACTIONS(3177),
    [anon_sym_time] = ACTIONS(3177),
    [anon_sym_times] = ACTIONS(3177),
    [anon_sym_wait] = ACTIONS(3177),
    [anon_sym_wantarray] = ACTIONS(3177),
    [anon_sym_abs] = ACTIONS(3177),
    [anon_sym_alarm] = ACTIONS(3177),
    [anon_sym_chop] = ACTIONS(3177),
    [anon_sym_chdir] = ACTIONS(3177),
    [anon_sym_close] = ACTIONS(3177),
    [anon_sym_closedir] = ACTIONS(3177),
    [anon_sym_caller] = ACTIONS(3177),
    [anon_sym_chomp] = ACTIONS(3177),
    [anon_sym_chr] = ACTIONS(3177),
    [anon_sym_cos] = ACTIONS(3177),
    [anon_sym_chroot] = ACTIONS(3177),
    [anon_sym_defined] = ACTIONS(3177),
    [anon_sym_delete] = ACTIONS(3177),
    [anon_sym_dbmclose] = ACTIONS(3177),
    [anon_sym_exists] = ACTIONS(3177),
    [anon_sym_exit] = ACTIONS(3177),
    [anon_sym_eof] = ACTIONS(3177),
    [anon_sym_exp] = ACTIONS(3177),
    [anon_sym_each] = ACTIONS(3177),
    [anon_sym_fc] = ACTIONS(3177),
    [anon_sym_fileno] = ACTIONS(3177),
    [anon_sym_gmtime] = ACTIONS(3177),
    [anon_sym_getc] = ACTIONS(3177),
    [anon_sym_getpgrp] = ACTIONS(3177),
    [anon_sym_getprotobyname] = ACTIONS(3177),
    [anon_sym_getpwname] = ACTIONS(3177),
    [anon_sym_getpwuid] = ACTIONS(3177),
    [anon_sym_getpeername] = ACTIONS(3177),
    [anon_sym_getnetbyname] = ACTIONS(3177),
    [anon_sym_getsockname] = ACTIONS(3177),
    [anon_sym_getgrnam] = ACTIONS(3177),
    [anon_sym_getgrgid] = ACTIONS(3177),
    [anon_sym_hex] = ACTIONS(3177),
    [anon_sym_int] = ACTIONS(3177),
    [anon_sym_keys] = ACTIONS(3177),
    [anon_sym_lc] = ACTIONS(3177),
    [anon_sym_lcfirst] = ACTIONS(3177),
    [anon_sym_length] = ACTIONS(3177),
    [anon_sym_localtime] = ACTIONS(3177),
    [anon_sym_log] = ACTIONS(3177),
    [anon_sym_lock] = ACTIONS(3177),
    [anon_sym_lstat] = ACTIONS(3177),
    [anon_sym_oct] = ACTIONS(3177),
    [anon_sym_ord] = ACTIONS(3177),
    [anon_sym_prototype] = ACTIONS(3177),
    [anon_sym_pop] = ACTIONS(3177),
    [anon_sym_pos] = ACTIONS(3177),
    [anon_sym_quotemeta] = ACTIONS(3177),
    [anon_sym_reset] = ACTIONS(3177),
    [anon_sym_rand] = ACTIONS(3177),
    [anon_sym_rmdir] = ACTIONS(3177),
    [anon_sym_readdir] = ACTIONS(3177),
    [anon_sym_readline] = ACTIONS(3177),
    [anon_sym_readpipe] = ACTIONS(3177),
    [anon_sym_rewinddir] = ACTIONS(3177),
    [anon_sym_readlink] = ACTIONS(3177),
    [anon_sym_ref] = ACTIONS(3177),
    [anon_sym_scalar] = ACTIONS(3177),
    [anon_sym_shift] = ACTIONS(3177),
    [anon_sym_sin] = ACTIONS(3177),
    [anon_sym_sleep] = ACTIONS(3177),
    [anon_sym_sqrt] = ACTIONS(3177),
    [anon_sym_srand] = ACTIONS(3177),
    [anon_sym_stat] = ACTIONS(3177),
    [anon_sym_study] = ACTIONS(3177),
    [anon_sym_tell] = ACTIONS(3177),
    [anon_sym_telldir] = ACTIONS(3177),
    [anon_sym_tied] = ACTIONS(3177),
    [anon_sym_uc] = ACTIONS(3177),
    [anon_sym_ucfirst] = ACTIONS(3177),
    [anon_sym_untie] = ACTIONS(3177),
    [anon_sym_umask] = ACTIONS(3177),
    [anon_sym_values] = ACTIONS(3177),
    [anon_sym_write] = ACTIONS(3177),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3177),
    [anon_sym___END__] = ACTIONS(3177),
    [anon_sym_EOT] = ACTIONS(3177),
    [anon_sym_q] = ACTIONS(3177),
    [anon_sym_qq] = ACTIONS(3177),
    [anon_sym_qw] = ACTIONS(3177),
    [anon_sym_qx] = ACTIONS(3177),
    [anon_sym_qr] = ACTIONS(3177),
    [anon_sym_m] = ACTIONS(3177),
    [anon_sym_s] = ACTIONS(3177),
    [anon_sym_tr] = ACTIONS(3177),
    [anon_sym_y] = ACTIONS(3177),
    [anon_sym_if] = ACTIONS(3177),
    [anon_sym_unless] = ACTIONS(3177),
    [anon_sym_while] = ACTIONS(3177),
    [anon_sym_until] = ACTIONS(3177),
    [aux_sym__bareword_token1] = ACTIONS(3177),
    [sym_number] = ACTIONS(3177),
    [anon_sym_true] = ACTIONS(3177),
    [anon_sym_false] = ACTIONS(3177),
    [sym__single_quote] = ACTIONS(3175),
    [sym__double_quote] = ACTIONS(3175),
    [sym__backtick_quote] = ACTIONS(3175),
    [sym__search_slash_quote] = ACTIONS(3175),
    [sym__open_readline_bracket] = ACTIONS(3175),
    [sym__open_fileglob_bracket] = ACTIONS(3175),
    [sym__PERLY_HEREDOC] = ACTIONS(3175),
    [sym__ctrl_z_hack] = ACTIONS(3175),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3175),
    [sym__filetest] = ACTIONS(3175),
  },
  [STATE(1506)] = {
    [sym_heredoc_content] = STATE(1506),
    [ts_builtin_sym_end] = ACTIONS(3179),
    [sym__identifier] = ACTIONS(3181),
    [anon_sym_LBRACE] = ACTIONS(3181),
    [anon_sym_SEMI] = ACTIONS(3181),
    [anon_sym_package] = ACTIONS(3181),
    [anon_sym_class] = ACTIONS(3181),
    [anon_sym_role] = ACTIONS(3181),
    [anon_sym_BUILD] = ACTIONS(3181),
    [anon_sym_ADJUST] = ACTIONS(3181),
    [anon_sym_DOLLAR] = ACTIONS(3181),
    [anon_sym_AT] = ACTIONS(3181),
    [anon_sym_my] = ACTIONS(3181),
    [anon_sym_extended] = ACTIONS(3181),
    [anon_sym_async] = ACTIONS(3181),
    [anon_sym_sub] = ACTIONS(3181),
    [anon_sym_method] = ACTIONS(3181),
    [anon_sym_LPAREN] = ACTIONS(3181),
    [anon_sym_continue] = ACTIONS(3181),
    [anon_sym_state] = ACTIONS(3181),
    [anon_sym_our] = ACTIONS(3181),
    [anon_sym_try] = ACTIONS(3181),
    [anon_sym_defer] = ACTIONS(3181),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3181),
    [anon_sym_LBRACK] = ACTIONS(3181),
    [anon_sym_LT_LT] = ACTIONS(3181),
    [anon_sym_SLASH_SLASH] = ACTIONS(3181),
    [anon_sym_PLUS] = ACTIONS(3181),
    [anon_sym_DASH] = ACTIONS(3181),
    [anon_sym_TILDE] = ACTIONS(3181),
    [anon_sym_BANG] = ACTIONS(3181),
    [anon_sym_PLUS_PLUS] = ACTIONS(3181),
    [anon_sym_DASH_DASH] = ACTIONS(3181),
    [anon_sym_BSLASH] = ACTIONS(3179),
    [anon_sym_do] = ACTIONS(3181),
    [anon_sym_eval] = ACTIONS(3181),
    [anon_sym_await] = ACTIONS(3181),
    [anon_sym_field] = ACTIONS(3181),
    [anon_sym_local] = ACTIONS(3181),
    [anon_sym_dynamically] = ACTIONS(3181),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3181),
    [anon_sym_require] = ACTIONS(3181),
    [anon_sym_map] = ACTIONS(3181),
    [anon_sym_grep] = ACTIONS(3181),
    [anon_sym_sort] = ACTIONS(3181),
    [anon_sym_goto] = ACTIONS(3181),
    [anon_sym_return] = ACTIONS(3181),
    [anon_sym_undef] = ACTIONS(3181),
    [anon_sym_PERCENT2] = ACTIONS(3181),
    [anon_sym_AMP2] = ACTIONS(3181),
    [anon_sym_STAR2] = ACTIONS(3181),
    [anon_sym_use] = ACTIONS(3181),
    [anon_sym_no] = ACTIONS(3181),
    [anon_sym_for] = ACTIONS(3181),
    [anon_sym_foreach] = ACTIONS(3181),
    [anon_sym_last] = ACTIONS(3181),
    [anon_sym_next] = ACTIONS(3181),
    [anon_sym_redo] = ACTIONS(3181),
    [anon_sym_BEGIN] = ACTIONS(3181),
    [anon_sym_INIT] = ACTIONS(3181),
    [anon_sym_CHECK] = ACTIONS(3181),
    [anon_sym_UNITCHECK] = ACTIONS(3181),
    [anon_sym_END] = ACTIONS(3181),
    [anon_sym___FILE__] = ACTIONS(3181),
    [anon_sym___LINE__] = ACTIONS(3181),
    [anon_sym___PACKAGE__] = ACTIONS(3181),
    [anon_sym___SUB__] = ACTIONS(3181),
    [anon_sym_break] = ACTIONS(3181),
    [anon_sym_fork] = ACTIONS(3181),
    [anon_sym_getppid] = ACTIONS(3181),
    [anon_sym_time] = ACTIONS(3181),
    [anon_sym_times] = ACTIONS(3181),
    [anon_sym_wait] = ACTIONS(3181),
    [anon_sym_wantarray] = ACTIONS(3181),
    [anon_sym_abs] = ACTIONS(3181),
    [anon_sym_alarm] = ACTIONS(3181),
    [anon_sym_chop] = ACTIONS(3181),
    [anon_sym_chdir] = ACTIONS(3181),
    [anon_sym_close] = ACTIONS(3181),
    [anon_sym_closedir] = ACTIONS(3181),
    [anon_sym_caller] = ACTIONS(3181),
    [anon_sym_chomp] = ACTIONS(3181),
    [anon_sym_chr] = ACTIONS(3181),
    [anon_sym_cos] = ACTIONS(3181),
    [anon_sym_chroot] = ACTIONS(3181),
    [anon_sym_defined] = ACTIONS(3181),
    [anon_sym_delete] = ACTIONS(3181),
    [anon_sym_dbmclose] = ACTIONS(3181),
    [anon_sym_exists] = ACTIONS(3181),
    [anon_sym_exit] = ACTIONS(3181),
    [anon_sym_eof] = ACTIONS(3181),
    [anon_sym_exp] = ACTIONS(3181),
    [anon_sym_each] = ACTIONS(3181),
    [anon_sym_fc] = ACTIONS(3181),
    [anon_sym_fileno] = ACTIONS(3181),
    [anon_sym_gmtime] = ACTIONS(3181),
    [anon_sym_getc] = ACTIONS(3181),
    [anon_sym_getpgrp] = ACTIONS(3181),
    [anon_sym_getprotobyname] = ACTIONS(3181),
    [anon_sym_getpwname] = ACTIONS(3181),
    [anon_sym_getpwuid] = ACTIONS(3181),
    [anon_sym_getpeername] = ACTIONS(3181),
    [anon_sym_getnetbyname] = ACTIONS(3181),
    [anon_sym_getsockname] = ACTIONS(3181),
    [anon_sym_getgrnam] = ACTIONS(3181),
    [anon_sym_getgrgid] = ACTIONS(3181),
    [anon_sym_hex] = ACTIONS(3181),
    [anon_sym_int] = ACTIONS(3181),
    [anon_sym_keys] = ACTIONS(3181),
    [anon_sym_lc] = ACTIONS(3181),
    [anon_sym_lcfirst] = ACTIONS(3181),
    [anon_sym_length] = ACTIONS(3181),
    [anon_sym_localtime] = ACTIONS(3181),
    [anon_sym_log] = ACTIONS(3181),
    [anon_sym_lock] = ACTIONS(3181),
    [anon_sym_lstat] = ACTIONS(3181),
    [anon_sym_oct] = ACTIONS(3181),
    [anon_sym_ord] = ACTIONS(3181),
    [anon_sym_prototype] = ACTIONS(3181),
    [anon_sym_pop] = ACTIONS(3181),
    [anon_sym_pos] = ACTIONS(3181),
    [anon_sym_quotemeta] = ACTIONS(3181),
    [anon_sym_reset] = ACTIONS(3181),
    [anon_sym_rand] = ACTIONS(3181),
    [anon_sym_rmdir] = ACTIONS(3181),
    [anon_sym_readdir] = ACTIONS(3181),
    [anon_sym_readline] = ACTIONS(3181),
    [anon_sym_readpipe] = ACTIONS(3181),
    [anon_sym_rewinddir] = ACTIONS(3181),
    [anon_sym_readlink] = ACTIONS(3181),
    [anon_sym_ref] = ACTIONS(3181),
    [anon_sym_scalar] = ACTIONS(3181),
    [anon_sym_shift] = ACTIONS(3181),
    [anon_sym_sin] = ACTIONS(3181),
    [anon_sym_sleep] = ACTIONS(3181),
    [anon_sym_sqrt] = ACTIONS(3181),
    [anon_sym_srand] = ACTIONS(3181),
    [anon_sym_stat] = ACTIONS(3181),
    [anon_sym_study] = ACTIONS(3181),
    [anon_sym_tell] = ACTIONS(3181),
    [anon_sym_telldir] = ACTIONS(3181),
    [anon_sym_tied] = ACTIONS(3181),
    [anon_sym_uc] = ACTIONS(3181),
    [anon_sym_ucfirst] = ACTIONS(3181),
    [anon_sym_untie] = ACTIONS(3181),
    [anon_sym_umask] = ACTIONS(3181),
    [anon_sym_values] = ACTIONS(3181),
    [anon_sym_write] = ACTIONS(3181),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3181),
    [anon_sym___END__] = ACTIONS(3181),
    [anon_sym_EOT] = ACTIONS(3181),
    [anon_sym_q] = ACTIONS(3181),
    [anon_sym_qq] = ACTIONS(3181),
    [anon_sym_qw] = ACTIONS(3181),
    [anon_sym_qx] = ACTIONS(3181),
    [anon_sym_qr] = ACTIONS(3181),
    [anon_sym_m] = ACTIONS(3181),
    [anon_sym_s] = ACTIONS(3181),
    [anon_sym_tr] = ACTIONS(3181),
    [anon_sym_y] = ACTIONS(3181),
    [anon_sym_if] = ACTIONS(3181),
    [anon_sym_unless] = ACTIONS(3181),
    [anon_sym_while] = ACTIONS(3181),
    [anon_sym_until] = ACTIONS(3181),
    [aux_sym__bareword_token1] = ACTIONS(3181),
    [sym_number] = ACTIONS(3181),
    [anon_sym_true] = ACTIONS(3181),
    [anon_sym_false] = ACTIONS(3181),
    [sym__single_quote] = ACTIONS(3179),
    [sym__double_quote] = ACTIONS(3179),
    [sym__backtick_quote] = ACTIONS(3179),
    [sym__search_slash_quote] = ACTIONS(3179),
    [sym__open_readline_bracket] = ACTIONS(3179),
    [sym__open_fileglob_bracket] = ACTIONS(3179),
    [sym__PERLY_HEREDOC] = ACTIONS(3179),
    [sym__ctrl_z_hack] = ACTIONS(3179),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3179),
    [sym__filetest] = ACTIONS(3179),
  },
  [STATE(1507)] = {
    [sym_heredoc_content] = STATE(1507),
    [ts_builtin_sym_end] = ACTIONS(3183),
    [sym__identifier] = ACTIONS(3185),
    [anon_sym_LBRACE] = ACTIONS(3185),
    [anon_sym_SEMI] = ACTIONS(3185),
    [anon_sym_package] = ACTIONS(3185),
    [anon_sym_class] = ACTIONS(3185),
    [anon_sym_role] = ACTIONS(3185),
    [anon_sym_BUILD] = ACTIONS(3185),
    [anon_sym_ADJUST] = ACTIONS(3185),
    [anon_sym_DOLLAR] = ACTIONS(3185),
    [anon_sym_AT] = ACTIONS(3185),
    [anon_sym_my] = ACTIONS(3185),
    [anon_sym_extended] = ACTIONS(3185),
    [anon_sym_async] = ACTIONS(3185),
    [anon_sym_sub] = ACTIONS(3185),
    [anon_sym_method] = ACTIONS(3185),
    [anon_sym_LPAREN] = ACTIONS(3185),
    [anon_sym_continue] = ACTIONS(3185),
    [anon_sym_state] = ACTIONS(3185),
    [anon_sym_our] = ACTIONS(3185),
    [anon_sym_try] = ACTIONS(3185),
    [anon_sym_defer] = ACTIONS(3185),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3185),
    [anon_sym_LBRACK] = ACTIONS(3185),
    [anon_sym_LT_LT] = ACTIONS(3185),
    [anon_sym_SLASH_SLASH] = ACTIONS(3185),
    [anon_sym_PLUS] = ACTIONS(3185),
    [anon_sym_DASH] = ACTIONS(3185),
    [anon_sym_TILDE] = ACTIONS(3185),
    [anon_sym_BANG] = ACTIONS(3185),
    [anon_sym_PLUS_PLUS] = ACTIONS(3185),
    [anon_sym_DASH_DASH] = ACTIONS(3185),
    [anon_sym_BSLASH] = ACTIONS(3183),
    [anon_sym_do] = ACTIONS(3185),
    [anon_sym_eval] = ACTIONS(3185),
    [anon_sym_await] = ACTIONS(3185),
    [anon_sym_field] = ACTIONS(3185),
    [anon_sym_local] = ACTIONS(3185),
    [anon_sym_dynamically] = ACTIONS(3185),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3185),
    [anon_sym_require] = ACTIONS(3185),
    [anon_sym_map] = ACTIONS(3185),
    [anon_sym_grep] = ACTIONS(3185),
    [anon_sym_sort] = ACTIONS(3185),
    [anon_sym_goto] = ACTIONS(3185),
    [anon_sym_return] = ACTIONS(3185),
    [anon_sym_undef] = ACTIONS(3185),
    [anon_sym_PERCENT2] = ACTIONS(3185),
    [anon_sym_AMP2] = ACTIONS(3185),
    [anon_sym_STAR2] = ACTIONS(3185),
    [anon_sym_use] = ACTIONS(3185),
    [anon_sym_no] = ACTIONS(3185),
    [anon_sym_for] = ACTIONS(3185),
    [anon_sym_foreach] = ACTIONS(3185),
    [anon_sym_last] = ACTIONS(3185),
    [anon_sym_next] = ACTIONS(3185),
    [anon_sym_redo] = ACTIONS(3185),
    [anon_sym_BEGIN] = ACTIONS(3185),
    [anon_sym_INIT] = ACTIONS(3185),
    [anon_sym_CHECK] = ACTIONS(3185),
    [anon_sym_UNITCHECK] = ACTIONS(3185),
    [anon_sym_END] = ACTIONS(3185),
    [anon_sym___FILE__] = ACTIONS(3185),
    [anon_sym___LINE__] = ACTIONS(3185),
    [anon_sym___PACKAGE__] = ACTIONS(3185),
    [anon_sym___SUB__] = ACTIONS(3185),
    [anon_sym_break] = ACTIONS(3185),
    [anon_sym_fork] = ACTIONS(3185),
    [anon_sym_getppid] = ACTIONS(3185),
    [anon_sym_time] = ACTIONS(3185),
    [anon_sym_times] = ACTIONS(3185),
    [anon_sym_wait] = ACTIONS(3185),
    [anon_sym_wantarray] = ACTIONS(3185),
    [anon_sym_abs] = ACTIONS(3185),
    [anon_sym_alarm] = ACTIONS(3185),
    [anon_sym_chop] = ACTIONS(3185),
    [anon_sym_chdir] = ACTIONS(3185),
    [anon_sym_close] = ACTIONS(3185),
    [anon_sym_closedir] = ACTIONS(3185),
    [anon_sym_caller] = ACTIONS(3185),
    [anon_sym_chomp] = ACTIONS(3185),
    [anon_sym_chr] = ACTIONS(3185),
    [anon_sym_cos] = ACTIONS(3185),
    [anon_sym_chroot] = ACTIONS(3185),
    [anon_sym_defined] = ACTIONS(3185),
    [anon_sym_delete] = ACTIONS(3185),
    [anon_sym_dbmclose] = ACTIONS(3185),
    [anon_sym_exists] = ACTIONS(3185),
    [anon_sym_exit] = ACTIONS(3185),
    [anon_sym_eof] = ACTIONS(3185),
    [anon_sym_exp] = ACTIONS(3185),
    [anon_sym_each] = ACTIONS(3185),
    [anon_sym_fc] = ACTIONS(3185),
    [anon_sym_fileno] = ACTIONS(3185),
    [anon_sym_gmtime] = ACTIONS(3185),
    [anon_sym_getc] = ACTIONS(3185),
    [anon_sym_getpgrp] = ACTIONS(3185),
    [anon_sym_getprotobyname] = ACTIONS(3185),
    [anon_sym_getpwname] = ACTIONS(3185),
    [anon_sym_getpwuid] = ACTIONS(3185),
    [anon_sym_getpeername] = ACTIONS(3185),
    [anon_sym_getnetbyname] = ACTIONS(3185),
    [anon_sym_getsockname] = ACTIONS(3185),
    [anon_sym_getgrnam] = ACTIONS(3185),
    [anon_sym_getgrgid] = ACTIONS(3185),
    [anon_sym_hex] = ACTIONS(3185),
    [anon_sym_int] = ACTIONS(3185),
    [anon_sym_keys] = ACTIONS(3185),
    [anon_sym_lc] = ACTIONS(3185),
    [anon_sym_lcfirst] = ACTIONS(3185),
    [anon_sym_length] = ACTIONS(3185),
    [anon_sym_localtime] = ACTIONS(3185),
    [anon_sym_log] = ACTIONS(3185),
    [anon_sym_lock] = ACTIONS(3185),
    [anon_sym_lstat] = ACTIONS(3185),
    [anon_sym_oct] = ACTIONS(3185),
    [anon_sym_ord] = ACTIONS(3185),
    [anon_sym_prototype] = ACTIONS(3185),
    [anon_sym_pop] = ACTIONS(3185),
    [anon_sym_pos] = ACTIONS(3185),
    [anon_sym_quotemeta] = ACTIONS(3185),
    [anon_sym_reset] = ACTIONS(3185),
    [anon_sym_rand] = ACTIONS(3185),
    [anon_sym_rmdir] = ACTIONS(3185),
    [anon_sym_readdir] = ACTIONS(3185),
    [anon_sym_readline] = ACTIONS(3185),
    [anon_sym_readpipe] = ACTIONS(3185),
    [anon_sym_rewinddir] = ACTIONS(3185),
    [anon_sym_readlink] = ACTIONS(3185),
    [anon_sym_ref] = ACTIONS(3185),
    [anon_sym_scalar] = ACTIONS(3185),
    [anon_sym_shift] = ACTIONS(3185),
    [anon_sym_sin] = ACTIONS(3185),
    [anon_sym_sleep] = ACTIONS(3185),
    [anon_sym_sqrt] = ACTIONS(3185),
    [anon_sym_srand] = ACTIONS(3185),
    [anon_sym_stat] = ACTIONS(3185),
    [anon_sym_study] = ACTIONS(3185),
    [anon_sym_tell] = ACTIONS(3185),
    [anon_sym_telldir] = ACTIONS(3185),
    [anon_sym_tied] = ACTIONS(3185),
    [anon_sym_uc] = ACTIONS(3185),
    [anon_sym_ucfirst] = ACTIONS(3185),
    [anon_sym_untie] = ACTIONS(3185),
    [anon_sym_umask] = ACTIONS(3185),
    [anon_sym_values] = ACTIONS(3185),
    [anon_sym_write] = ACTIONS(3185),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3185),
    [anon_sym___END__] = ACTIONS(3185),
    [anon_sym_EOT] = ACTIONS(3185),
    [anon_sym_q] = ACTIONS(3185),
    [anon_sym_qq] = ACTIONS(3185),
    [anon_sym_qw] = ACTIONS(3185),
    [anon_sym_qx] = ACTIONS(3185),
    [anon_sym_qr] = ACTIONS(3185),
    [anon_sym_m] = ACTIONS(3185),
    [anon_sym_s] = ACTIONS(3185),
    [anon_sym_tr] = ACTIONS(3185),
    [anon_sym_y] = ACTIONS(3185),
    [anon_sym_if] = ACTIONS(3185),
    [anon_sym_unless] = ACTIONS(3185),
    [anon_sym_while] = ACTIONS(3185),
    [anon_sym_until] = ACTIONS(3185),
    [aux_sym__bareword_token1] = ACTIONS(3185),
    [sym_number] = ACTIONS(3185),
    [anon_sym_true] = ACTIONS(3185),
    [anon_sym_false] = ACTIONS(3185),
    [sym__single_quote] = ACTIONS(3183),
    [sym__double_quote] = ACTIONS(3183),
    [sym__backtick_quote] = ACTIONS(3183),
    [sym__search_slash_quote] = ACTIONS(3183),
    [sym__open_readline_bracket] = ACTIONS(3183),
    [sym__open_fileglob_bracket] = ACTIONS(3183),
    [sym__PERLY_HEREDOC] = ACTIONS(3183),
    [sym__ctrl_z_hack] = ACTIONS(3183),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3183),
    [sym__filetest] = ACTIONS(3183),
  },
  [STATE(1508)] = {
    [sym_heredoc_content] = STATE(1508),
    [ts_builtin_sym_end] = ACTIONS(3187),
    [sym__identifier] = ACTIONS(3189),
    [anon_sym_LBRACE] = ACTIONS(3189),
    [anon_sym_SEMI] = ACTIONS(3189),
    [anon_sym_package] = ACTIONS(3189),
    [anon_sym_class] = ACTIONS(3189),
    [anon_sym_role] = ACTIONS(3189),
    [anon_sym_BUILD] = ACTIONS(3189),
    [anon_sym_ADJUST] = ACTIONS(3189),
    [anon_sym_DOLLAR] = ACTIONS(3189),
    [anon_sym_AT] = ACTIONS(3189),
    [anon_sym_my] = ACTIONS(3189),
    [anon_sym_extended] = ACTIONS(3189),
    [anon_sym_async] = ACTIONS(3189),
    [anon_sym_sub] = ACTIONS(3189),
    [anon_sym_method] = ACTIONS(3189),
    [anon_sym_LPAREN] = ACTIONS(3189),
    [anon_sym_continue] = ACTIONS(3189),
    [anon_sym_state] = ACTIONS(3189),
    [anon_sym_our] = ACTIONS(3189),
    [anon_sym_try] = ACTIONS(3189),
    [anon_sym_defer] = ACTIONS(3189),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3189),
    [anon_sym_LBRACK] = ACTIONS(3189),
    [anon_sym_LT_LT] = ACTIONS(3189),
    [anon_sym_SLASH_SLASH] = ACTIONS(3189),
    [anon_sym_PLUS] = ACTIONS(3189),
    [anon_sym_DASH] = ACTIONS(3189),
    [anon_sym_TILDE] = ACTIONS(3189),
    [anon_sym_BANG] = ACTIONS(3189),
    [anon_sym_PLUS_PLUS] = ACTIONS(3189),
    [anon_sym_DASH_DASH] = ACTIONS(3189),
    [anon_sym_BSLASH] = ACTIONS(3187),
    [anon_sym_do] = ACTIONS(3189),
    [anon_sym_eval] = ACTIONS(3189),
    [anon_sym_await] = ACTIONS(3189),
    [anon_sym_field] = ACTIONS(3189),
    [anon_sym_local] = ACTIONS(3189),
    [anon_sym_dynamically] = ACTIONS(3189),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3189),
    [anon_sym_require] = ACTIONS(3189),
    [anon_sym_map] = ACTIONS(3189),
    [anon_sym_grep] = ACTIONS(3189),
    [anon_sym_sort] = ACTIONS(3189),
    [anon_sym_goto] = ACTIONS(3189),
    [anon_sym_return] = ACTIONS(3189),
    [anon_sym_undef] = ACTIONS(3189),
    [anon_sym_PERCENT2] = ACTIONS(3189),
    [anon_sym_AMP2] = ACTIONS(3189),
    [anon_sym_STAR2] = ACTIONS(3189),
    [anon_sym_use] = ACTIONS(3189),
    [anon_sym_no] = ACTIONS(3189),
    [anon_sym_for] = ACTIONS(3189),
    [anon_sym_foreach] = ACTIONS(3189),
    [anon_sym_last] = ACTIONS(3189),
    [anon_sym_next] = ACTIONS(3189),
    [anon_sym_redo] = ACTIONS(3189),
    [anon_sym_BEGIN] = ACTIONS(3189),
    [anon_sym_INIT] = ACTIONS(3189),
    [anon_sym_CHECK] = ACTIONS(3189),
    [anon_sym_UNITCHECK] = ACTIONS(3189),
    [anon_sym_END] = ACTIONS(3189),
    [anon_sym___FILE__] = ACTIONS(3189),
    [anon_sym___LINE__] = ACTIONS(3189),
    [anon_sym___PACKAGE__] = ACTIONS(3189),
    [anon_sym___SUB__] = ACTIONS(3189),
    [anon_sym_break] = ACTIONS(3189),
    [anon_sym_fork] = ACTIONS(3189),
    [anon_sym_getppid] = ACTIONS(3189),
    [anon_sym_time] = ACTIONS(3189),
    [anon_sym_times] = ACTIONS(3189),
    [anon_sym_wait] = ACTIONS(3189),
    [anon_sym_wantarray] = ACTIONS(3189),
    [anon_sym_abs] = ACTIONS(3189),
    [anon_sym_alarm] = ACTIONS(3189),
    [anon_sym_chop] = ACTIONS(3189),
    [anon_sym_chdir] = ACTIONS(3189),
    [anon_sym_close] = ACTIONS(3189),
    [anon_sym_closedir] = ACTIONS(3189),
    [anon_sym_caller] = ACTIONS(3189),
    [anon_sym_chomp] = ACTIONS(3189),
    [anon_sym_chr] = ACTIONS(3189),
    [anon_sym_cos] = ACTIONS(3189),
    [anon_sym_chroot] = ACTIONS(3189),
    [anon_sym_defined] = ACTIONS(3189),
    [anon_sym_delete] = ACTIONS(3189),
    [anon_sym_dbmclose] = ACTIONS(3189),
    [anon_sym_exists] = ACTIONS(3189),
    [anon_sym_exit] = ACTIONS(3189),
    [anon_sym_eof] = ACTIONS(3189),
    [anon_sym_exp] = ACTIONS(3189),
    [anon_sym_each] = ACTIONS(3189),
    [anon_sym_fc] = ACTIONS(3189),
    [anon_sym_fileno] = ACTIONS(3189),
    [anon_sym_gmtime] = ACTIONS(3189),
    [anon_sym_getc] = ACTIONS(3189),
    [anon_sym_getpgrp] = ACTIONS(3189),
    [anon_sym_getprotobyname] = ACTIONS(3189),
    [anon_sym_getpwname] = ACTIONS(3189),
    [anon_sym_getpwuid] = ACTIONS(3189),
    [anon_sym_getpeername] = ACTIONS(3189),
    [anon_sym_getnetbyname] = ACTIONS(3189),
    [anon_sym_getsockname] = ACTIONS(3189),
    [anon_sym_getgrnam] = ACTIONS(3189),
    [anon_sym_getgrgid] = ACTIONS(3189),
    [anon_sym_hex] = ACTIONS(3189),
    [anon_sym_int] = ACTIONS(3189),
    [anon_sym_keys] = ACTIONS(3189),
    [anon_sym_lc] = ACTIONS(3189),
    [anon_sym_lcfirst] = ACTIONS(3189),
    [anon_sym_length] = ACTIONS(3189),
    [anon_sym_localtime] = ACTIONS(3189),
    [anon_sym_log] = ACTIONS(3189),
    [anon_sym_lock] = ACTIONS(3189),
    [anon_sym_lstat] = ACTIONS(3189),
    [anon_sym_oct] = ACTIONS(3189),
    [anon_sym_ord] = ACTIONS(3189),
    [anon_sym_prototype] = ACTIONS(3189),
    [anon_sym_pop] = ACTIONS(3189),
    [anon_sym_pos] = ACTIONS(3189),
    [anon_sym_quotemeta] = ACTIONS(3189),
    [anon_sym_reset] = ACTIONS(3189),
    [anon_sym_rand] = ACTIONS(3189),
    [anon_sym_rmdir] = ACTIONS(3189),
    [anon_sym_readdir] = ACTIONS(3189),
    [anon_sym_readline] = ACTIONS(3189),
    [anon_sym_readpipe] = ACTIONS(3189),
    [anon_sym_rewinddir] = ACTIONS(3189),
    [anon_sym_readlink] = ACTIONS(3189),
    [anon_sym_ref] = ACTIONS(3189),
    [anon_sym_scalar] = ACTIONS(3189),
    [anon_sym_shift] = ACTIONS(3189),
    [anon_sym_sin] = ACTIONS(3189),
    [anon_sym_sleep] = ACTIONS(3189),
    [anon_sym_sqrt] = ACTIONS(3189),
    [anon_sym_srand] = ACTIONS(3189),
    [anon_sym_stat] = ACTIONS(3189),
    [anon_sym_study] = ACTIONS(3189),
    [anon_sym_tell] = ACTIONS(3189),
    [anon_sym_telldir] = ACTIONS(3189),
    [anon_sym_tied] = ACTIONS(3189),
    [anon_sym_uc] = ACTIONS(3189),
    [anon_sym_ucfirst] = ACTIONS(3189),
    [anon_sym_untie] = ACTIONS(3189),
    [anon_sym_umask] = ACTIONS(3189),
    [anon_sym_values] = ACTIONS(3189),
    [anon_sym_write] = ACTIONS(3189),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3189),
    [anon_sym___END__] = ACTIONS(3189),
    [anon_sym_EOT] = ACTIONS(3189),
    [anon_sym_q] = ACTIONS(3189),
    [anon_sym_qq] = ACTIONS(3189),
    [anon_sym_qw] = ACTIONS(3189),
    [anon_sym_qx] = ACTIONS(3189),
    [anon_sym_qr] = ACTIONS(3189),
    [anon_sym_m] = ACTIONS(3189),
    [anon_sym_s] = ACTIONS(3189),
    [anon_sym_tr] = ACTIONS(3189),
    [anon_sym_y] = ACTIONS(3189),
    [anon_sym_if] = ACTIONS(3189),
    [anon_sym_unless] = ACTIONS(3189),
    [anon_sym_while] = ACTIONS(3189),
    [anon_sym_until] = ACTIONS(3189),
    [aux_sym__bareword_token1] = ACTIONS(3189),
    [sym_number] = ACTIONS(3189),
    [anon_sym_true] = ACTIONS(3189),
    [anon_sym_false] = ACTIONS(3189),
    [sym__single_quote] = ACTIONS(3187),
    [sym__double_quote] = ACTIONS(3187),
    [sym__backtick_quote] = ACTIONS(3187),
    [sym__search_slash_quote] = ACTIONS(3187),
    [sym__open_readline_bracket] = ACTIONS(3187),
    [sym__open_fileglob_bracket] = ACTIONS(3187),
    [sym__PERLY_HEREDOC] = ACTIONS(3187),
    [sym__ctrl_z_hack] = ACTIONS(3187),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3187),
    [sym__filetest] = ACTIONS(3187),
  },
  [STATE(1509)] = {
    [sym_heredoc_content] = STATE(1509),
    [ts_builtin_sym_end] = ACTIONS(3191),
    [sym__identifier] = ACTIONS(3193),
    [anon_sym_LBRACE] = ACTIONS(3193),
    [anon_sym_SEMI] = ACTIONS(3193),
    [anon_sym_package] = ACTIONS(3193),
    [anon_sym_class] = ACTIONS(3193),
    [anon_sym_role] = ACTIONS(3193),
    [anon_sym_BUILD] = ACTIONS(3193),
    [anon_sym_ADJUST] = ACTIONS(3193),
    [anon_sym_DOLLAR] = ACTIONS(3193),
    [anon_sym_AT] = ACTIONS(3193),
    [anon_sym_my] = ACTIONS(3193),
    [anon_sym_extended] = ACTIONS(3193),
    [anon_sym_async] = ACTIONS(3193),
    [anon_sym_sub] = ACTIONS(3193),
    [anon_sym_method] = ACTIONS(3193),
    [anon_sym_LPAREN] = ACTIONS(3193),
    [anon_sym_continue] = ACTIONS(3193),
    [anon_sym_state] = ACTIONS(3193),
    [anon_sym_our] = ACTIONS(3193),
    [anon_sym_try] = ACTIONS(3193),
    [anon_sym_defer] = ACTIONS(3193),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3193),
    [anon_sym_LBRACK] = ACTIONS(3193),
    [anon_sym_LT_LT] = ACTIONS(3193),
    [anon_sym_SLASH_SLASH] = ACTIONS(3193),
    [anon_sym_PLUS] = ACTIONS(3193),
    [anon_sym_DASH] = ACTIONS(3193),
    [anon_sym_TILDE] = ACTIONS(3193),
    [anon_sym_BANG] = ACTIONS(3193),
    [anon_sym_PLUS_PLUS] = ACTIONS(3193),
    [anon_sym_DASH_DASH] = ACTIONS(3193),
    [anon_sym_BSLASH] = ACTIONS(3191),
    [anon_sym_do] = ACTIONS(3193),
    [anon_sym_eval] = ACTIONS(3193),
    [anon_sym_await] = ACTIONS(3193),
    [anon_sym_field] = ACTIONS(3193),
    [anon_sym_local] = ACTIONS(3193),
    [anon_sym_dynamically] = ACTIONS(3193),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3193),
    [anon_sym_require] = ACTIONS(3193),
    [anon_sym_map] = ACTIONS(3193),
    [anon_sym_grep] = ACTIONS(3193),
    [anon_sym_sort] = ACTIONS(3193),
    [anon_sym_goto] = ACTIONS(3193),
    [anon_sym_return] = ACTIONS(3193),
    [anon_sym_undef] = ACTIONS(3193),
    [anon_sym_PERCENT2] = ACTIONS(3193),
    [anon_sym_AMP2] = ACTIONS(3193),
    [anon_sym_STAR2] = ACTIONS(3193),
    [anon_sym_use] = ACTIONS(3193),
    [anon_sym_no] = ACTIONS(3193),
    [anon_sym_for] = ACTIONS(3193),
    [anon_sym_foreach] = ACTIONS(3193),
    [anon_sym_last] = ACTIONS(3193),
    [anon_sym_next] = ACTIONS(3193),
    [anon_sym_redo] = ACTIONS(3193),
    [anon_sym_BEGIN] = ACTIONS(3193),
    [anon_sym_INIT] = ACTIONS(3193),
    [anon_sym_CHECK] = ACTIONS(3193),
    [anon_sym_UNITCHECK] = ACTIONS(3193),
    [anon_sym_END] = ACTIONS(3193),
    [anon_sym___FILE__] = ACTIONS(3193),
    [anon_sym___LINE__] = ACTIONS(3193),
    [anon_sym___PACKAGE__] = ACTIONS(3193),
    [anon_sym___SUB__] = ACTIONS(3193),
    [anon_sym_break] = ACTIONS(3193),
    [anon_sym_fork] = ACTIONS(3193),
    [anon_sym_getppid] = ACTIONS(3193),
    [anon_sym_time] = ACTIONS(3193),
    [anon_sym_times] = ACTIONS(3193),
    [anon_sym_wait] = ACTIONS(3193),
    [anon_sym_wantarray] = ACTIONS(3193),
    [anon_sym_abs] = ACTIONS(3193),
    [anon_sym_alarm] = ACTIONS(3193),
    [anon_sym_chop] = ACTIONS(3193),
    [anon_sym_chdir] = ACTIONS(3193),
    [anon_sym_close] = ACTIONS(3193),
    [anon_sym_closedir] = ACTIONS(3193),
    [anon_sym_caller] = ACTIONS(3193),
    [anon_sym_chomp] = ACTIONS(3193),
    [anon_sym_chr] = ACTIONS(3193),
    [anon_sym_cos] = ACTIONS(3193),
    [anon_sym_chroot] = ACTIONS(3193),
    [anon_sym_defined] = ACTIONS(3193),
    [anon_sym_delete] = ACTIONS(3193),
    [anon_sym_dbmclose] = ACTIONS(3193),
    [anon_sym_exists] = ACTIONS(3193),
    [anon_sym_exit] = ACTIONS(3193),
    [anon_sym_eof] = ACTIONS(3193),
    [anon_sym_exp] = ACTIONS(3193),
    [anon_sym_each] = ACTIONS(3193),
    [anon_sym_fc] = ACTIONS(3193),
    [anon_sym_fileno] = ACTIONS(3193),
    [anon_sym_gmtime] = ACTIONS(3193),
    [anon_sym_getc] = ACTIONS(3193),
    [anon_sym_getpgrp] = ACTIONS(3193),
    [anon_sym_getprotobyname] = ACTIONS(3193),
    [anon_sym_getpwname] = ACTIONS(3193),
    [anon_sym_getpwuid] = ACTIONS(3193),
    [anon_sym_getpeername] = ACTIONS(3193),
    [anon_sym_getnetbyname] = ACTIONS(3193),
    [anon_sym_getsockname] = ACTIONS(3193),
    [anon_sym_getgrnam] = ACTIONS(3193),
    [anon_sym_getgrgid] = ACTIONS(3193),
    [anon_sym_hex] = ACTIONS(3193),
    [anon_sym_int] = ACTIONS(3193),
    [anon_sym_keys] = ACTIONS(3193),
    [anon_sym_lc] = ACTIONS(3193),
    [anon_sym_lcfirst] = ACTIONS(3193),
    [anon_sym_length] = ACTIONS(3193),
    [anon_sym_localtime] = ACTIONS(3193),
    [anon_sym_log] = ACTIONS(3193),
    [anon_sym_lock] = ACTIONS(3193),
    [anon_sym_lstat] = ACTIONS(3193),
    [anon_sym_oct] = ACTIONS(3193),
    [anon_sym_ord] = ACTIONS(3193),
    [anon_sym_prototype] = ACTIONS(3193),
    [anon_sym_pop] = ACTIONS(3193),
    [anon_sym_pos] = ACTIONS(3193),
    [anon_sym_quotemeta] = ACTIONS(3193),
    [anon_sym_reset] = ACTIONS(3193),
    [anon_sym_rand] = ACTIONS(3193),
    [anon_sym_rmdir] = ACTIONS(3193),
    [anon_sym_readdir] = ACTIONS(3193),
    [anon_sym_readline] = ACTIONS(3193),
    [anon_sym_readpipe] = ACTIONS(3193),
    [anon_sym_rewinddir] = ACTIONS(3193),
    [anon_sym_readlink] = ACTIONS(3193),
    [anon_sym_ref] = ACTIONS(3193),
    [anon_sym_scalar] = ACTIONS(3193),
    [anon_sym_shift] = ACTIONS(3193),
    [anon_sym_sin] = ACTIONS(3193),
    [anon_sym_sleep] = ACTIONS(3193),
    [anon_sym_sqrt] = ACTIONS(3193),
    [anon_sym_srand] = ACTIONS(3193),
    [anon_sym_stat] = ACTIONS(3193),
    [anon_sym_study] = ACTIONS(3193),
    [anon_sym_tell] = ACTIONS(3193),
    [anon_sym_telldir] = ACTIONS(3193),
    [anon_sym_tied] = ACTIONS(3193),
    [anon_sym_uc] = ACTIONS(3193),
    [anon_sym_ucfirst] = ACTIONS(3193),
    [anon_sym_untie] = ACTIONS(3193),
    [anon_sym_umask] = ACTIONS(3193),
    [anon_sym_values] = ACTIONS(3193),
    [anon_sym_write] = ACTIONS(3193),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3193),
    [anon_sym___END__] = ACTIONS(3193),
    [anon_sym_EOT] = ACTIONS(3193),
    [anon_sym_q] = ACTIONS(3193),
    [anon_sym_qq] = ACTIONS(3193),
    [anon_sym_qw] = ACTIONS(3193),
    [anon_sym_qx] = ACTIONS(3193),
    [anon_sym_qr] = ACTIONS(3193),
    [anon_sym_m] = ACTIONS(3193),
    [anon_sym_s] = ACTIONS(3193),
    [anon_sym_tr] = ACTIONS(3193),
    [anon_sym_y] = ACTIONS(3193),
    [anon_sym_if] = ACTIONS(3193),
    [anon_sym_unless] = ACTIONS(3193),
    [anon_sym_while] = ACTIONS(3193),
    [anon_sym_until] = ACTIONS(3193),
    [aux_sym__bareword_token1] = ACTIONS(3193),
    [sym_number] = ACTIONS(3193),
    [anon_sym_true] = ACTIONS(3193),
    [anon_sym_false] = ACTIONS(3193),
    [sym__single_quote] = ACTIONS(3191),
    [sym__double_quote] = ACTIONS(3191),
    [sym__backtick_quote] = ACTIONS(3191),
    [sym__search_slash_quote] = ACTIONS(3191),
    [sym__open_readline_bracket] = ACTIONS(3191),
    [sym__open_fileglob_bracket] = ACTIONS(3191),
    [sym__PERLY_HEREDOC] = ACTIONS(3191),
    [sym__ctrl_z_hack] = ACTIONS(3191),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3191),
    [sym__filetest] = ACTIONS(3191),
  },
  [STATE(1510)] = {
    [sym_heredoc_content] = STATE(1510),
    [ts_builtin_sym_end] = ACTIONS(3195),
    [sym__identifier] = ACTIONS(3197),
    [anon_sym_LBRACE] = ACTIONS(3197),
    [anon_sym_SEMI] = ACTIONS(3197),
    [anon_sym_package] = ACTIONS(3197),
    [anon_sym_class] = ACTIONS(3197),
    [anon_sym_role] = ACTIONS(3197),
    [anon_sym_BUILD] = ACTIONS(3197),
    [anon_sym_ADJUST] = ACTIONS(3197),
    [anon_sym_DOLLAR] = ACTIONS(3197),
    [anon_sym_AT] = ACTIONS(3197),
    [anon_sym_my] = ACTIONS(3197),
    [anon_sym_extended] = ACTIONS(3197),
    [anon_sym_async] = ACTIONS(3197),
    [anon_sym_sub] = ACTIONS(3197),
    [anon_sym_method] = ACTIONS(3197),
    [anon_sym_LPAREN] = ACTIONS(3197),
    [anon_sym_continue] = ACTIONS(3197),
    [anon_sym_state] = ACTIONS(3197),
    [anon_sym_our] = ACTIONS(3197),
    [anon_sym_try] = ACTIONS(3197),
    [anon_sym_defer] = ACTIONS(3197),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3197),
    [anon_sym_LBRACK] = ACTIONS(3197),
    [anon_sym_LT_LT] = ACTIONS(3197),
    [anon_sym_SLASH_SLASH] = ACTIONS(3197),
    [anon_sym_PLUS] = ACTIONS(3197),
    [anon_sym_DASH] = ACTIONS(3197),
    [anon_sym_TILDE] = ACTIONS(3197),
    [anon_sym_BANG] = ACTIONS(3197),
    [anon_sym_PLUS_PLUS] = ACTIONS(3197),
    [anon_sym_DASH_DASH] = ACTIONS(3197),
    [anon_sym_BSLASH] = ACTIONS(3195),
    [anon_sym_do] = ACTIONS(3197),
    [anon_sym_eval] = ACTIONS(3197),
    [anon_sym_await] = ACTIONS(3197),
    [anon_sym_field] = ACTIONS(3197),
    [anon_sym_local] = ACTIONS(3197),
    [anon_sym_dynamically] = ACTIONS(3197),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3197),
    [anon_sym_require] = ACTIONS(3197),
    [anon_sym_map] = ACTIONS(3197),
    [anon_sym_grep] = ACTIONS(3197),
    [anon_sym_sort] = ACTIONS(3197),
    [anon_sym_goto] = ACTIONS(3197),
    [anon_sym_return] = ACTIONS(3197),
    [anon_sym_undef] = ACTIONS(3197),
    [anon_sym_PERCENT2] = ACTIONS(3197),
    [anon_sym_AMP2] = ACTIONS(3197),
    [anon_sym_STAR2] = ACTIONS(3197),
    [anon_sym_use] = ACTIONS(3197),
    [anon_sym_no] = ACTIONS(3197),
    [anon_sym_for] = ACTIONS(3197),
    [anon_sym_foreach] = ACTIONS(3197),
    [anon_sym_last] = ACTIONS(3197),
    [anon_sym_next] = ACTIONS(3197),
    [anon_sym_redo] = ACTIONS(3197),
    [anon_sym_BEGIN] = ACTIONS(3197),
    [anon_sym_INIT] = ACTIONS(3197),
    [anon_sym_CHECK] = ACTIONS(3197),
    [anon_sym_UNITCHECK] = ACTIONS(3197),
    [anon_sym_END] = ACTIONS(3197),
    [anon_sym___FILE__] = ACTIONS(3197),
    [anon_sym___LINE__] = ACTIONS(3197),
    [anon_sym___PACKAGE__] = ACTIONS(3197),
    [anon_sym___SUB__] = ACTIONS(3197),
    [anon_sym_break] = ACTIONS(3197),
    [anon_sym_fork] = ACTIONS(3197),
    [anon_sym_getppid] = ACTIONS(3197),
    [anon_sym_time] = ACTIONS(3197),
    [anon_sym_times] = ACTIONS(3197),
    [anon_sym_wait] = ACTIONS(3197),
    [anon_sym_wantarray] = ACTIONS(3197),
    [anon_sym_abs] = ACTIONS(3197),
    [anon_sym_alarm] = ACTIONS(3197),
    [anon_sym_chop] = ACTIONS(3197),
    [anon_sym_chdir] = ACTIONS(3197),
    [anon_sym_close] = ACTIONS(3197),
    [anon_sym_closedir] = ACTIONS(3197),
    [anon_sym_caller] = ACTIONS(3197),
    [anon_sym_chomp] = ACTIONS(3197),
    [anon_sym_chr] = ACTIONS(3197),
    [anon_sym_cos] = ACTIONS(3197),
    [anon_sym_chroot] = ACTIONS(3197),
    [anon_sym_defined] = ACTIONS(3197),
    [anon_sym_delete] = ACTIONS(3197),
    [anon_sym_dbmclose] = ACTIONS(3197),
    [anon_sym_exists] = ACTIONS(3197),
    [anon_sym_exit] = ACTIONS(3197),
    [anon_sym_eof] = ACTIONS(3197),
    [anon_sym_exp] = ACTIONS(3197),
    [anon_sym_each] = ACTIONS(3197),
    [anon_sym_fc] = ACTIONS(3197),
    [anon_sym_fileno] = ACTIONS(3197),
    [anon_sym_gmtime] = ACTIONS(3197),
    [anon_sym_getc] = ACTIONS(3197),
    [anon_sym_getpgrp] = ACTIONS(3197),
    [anon_sym_getprotobyname] = ACTIONS(3197),
    [anon_sym_getpwname] = ACTIONS(3197),
    [anon_sym_getpwuid] = ACTIONS(3197),
    [anon_sym_getpeername] = ACTIONS(3197),
    [anon_sym_getnetbyname] = ACTIONS(3197),
    [anon_sym_getsockname] = ACTIONS(3197),
    [anon_sym_getgrnam] = ACTIONS(3197),
    [anon_sym_getgrgid] = ACTIONS(3197),
    [anon_sym_hex] = ACTIONS(3197),
    [anon_sym_int] = ACTIONS(3197),
    [anon_sym_keys] = ACTIONS(3197),
    [anon_sym_lc] = ACTIONS(3197),
    [anon_sym_lcfirst] = ACTIONS(3197),
    [anon_sym_length] = ACTIONS(3197),
    [anon_sym_localtime] = ACTIONS(3197),
    [anon_sym_log] = ACTIONS(3197),
    [anon_sym_lock] = ACTIONS(3197),
    [anon_sym_lstat] = ACTIONS(3197),
    [anon_sym_oct] = ACTIONS(3197),
    [anon_sym_ord] = ACTIONS(3197),
    [anon_sym_prototype] = ACTIONS(3197),
    [anon_sym_pop] = ACTIONS(3197),
    [anon_sym_pos] = ACTIONS(3197),
    [anon_sym_quotemeta] = ACTIONS(3197),
    [anon_sym_reset] = ACTIONS(3197),
    [anon_sym_rand] = ACTIONS(3197),
    [anon_sym_rmdir] = ACTIONS(3197),
    [anon_sym_readdir] = ACTIONS(3197),
    [anon_sym_readline] = ACTIONS(3197),
    [anon_sym_readpipe] = ACTIONS(3197),
    [anon_sym_rewinddir] = ACTIONS(3197),
    [anon_sym_readlink] = ACTIONS(3197),
    [anon_sym_ref] = ACTIONS(3197),
    [anon_sym_scalar] = ACTIONS(3197),
    [anon_sym_shift] = ACTIONS(3197),
    [anon_sym_sin] = ACTIONS(3197),
    [anon_sym_sleep] = ACTIONS(3197),
    [anon_sym_sqrt] = ACTIONS(3197),
    [anon_sym_srand] = ACTIONS(3197),
    [anon_sym_stat] = ACTIONS(3197),
    [anon_sym_study] = ACTIONS(3197),
    [anon_sym_tell] = ACTIONS(3197),
    [anon_sym_telldir] = ACTIONS(3197),
    [anon_sym_tied] = ACTIONS(3197),
    [anon_sym_uc] = ACTIONS(3197),
    [anon_sym_ucfirst] = ACTIONS(3197),
    [anon_sym_untie] = ACTIONS(3197),
    [anon_sym_umask] = ACTIONS(3197),
    [anon_sym_values] = ACTIONS(3197),
    [anon_sym_write] = ACTIONS(3197),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3197),
    [anon_sym___END__] = ACTIONS(3197),
    [anon_sym_EOT] = ACTIONS(3197),
    [anon_sym_q] = ACTIONS(3197),
    [anon_sym_qq] = ACTIONS(3197),
    [anon_sym_qw] = ACTIONS(3197),
    [anon_sym_qx] = ACTIONS(3197),
    [anon_sym_qr] = ACTIONS(3197),
    [anon_sym_m] = ACTIONS(3197),
    [anon_sym_s] = ACTIONS(3197),
    [anon_sym_tr] = ACTIONS(3197),
    [anon_sym_y] = ACTIONS(3197),
    [anon_sym_if] = ACTIONS(3197),
    [anon_sym_unless] = ACTIONS(3197),
    [anon_sym_while] = ACTIONS(3197),
    [anon_sym_until] = ACTIONS(3197),
    [aux_sym__bareword_token1] = ACTIONS(3197),
    [sym_number] = ACTIONS(3197),
    [anon_sym_true] = ACTIONS(3197),
    [anon_sym_false] = ACTIONS(3197),
    [sym__single_quote] = ACTIONS(3195),
    [sym__double_quote] = ACTIONS(3195),
    [sym__backtick_quote] = ACTIONS(3195),
    [sym__search_slash_quote] = ACTIONS(3195),
    [sym__open_readline_bracket] = ACTIONS(3195),
    [sym__open_fileglob_bracket] = ACTIONS(3195),
    [sym__PERLY_HEREDOC] = ACTIONS(3195),
    [sym__ctrl_z_hack] = ACTIONS(3195),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3195),
    [sym__filetest] = ACTIONS(3195),
  },
  [STATE(1511)] = {
    [sym_heredoc_content] = STATE(1511),
    [ts_builtin_sym_end] = ACTIONS(3199),
    [sym__identifier] = ACTIONS(3201),
    [anon_sym_LBRACE] = ACTIONS(3201),
    [anon_sym_SEMI] = ACTIONS(3201),
    [anon_sym_package] = ACTIONS(3201),
    [anon_sym_class] = ACTIONS(3201),
    [anon_sym_role] = ACTIONS(3201),
    [anon_sym_BUILD] = ACTIONS(3201),
    [anon_sym_ADJUST] = ACTIONS(3201),
    [anon_sym_DOLLAR] = ACTIONS(3201),
    [anon_sym_AT] = ACTIONS(3201),
    [anon_sym_my] = ACTIONS(3201),
    [anon_sym_extended] = ACTIONS(3201),
    [anon_sym_async] = ACTIONS(3201),
    [anon_sym_sub] = ACTIONS(3201),
    [anon_sym_method] = ACTIONS(3201),
    [anon_sym_LPAREN] = ACTIONS(3201),
    [anon_sym_continue] = ACTIONS(3201),
    [anon_sym_state] = ACTIONS(3201),
    [anon_sym_our] = ACTIONS(3201),
    [anon_sym_try] = ACTIONS(3201),
    [anon_sym_defer] = ACTIONS(3201),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3201),
    [anon_sym_LBRACK] = ACTIONS(3201),
    [anon_sym_LT_LT] = ACTIONS(3201),
    [anon_sym_SLASH_SLASH] = ACTIONS(3201),
    [anon_sym_PLUS] = ACTIONS(3201),
    [anon_sym_DASH] = ACTIONS(3201),
    [anon_sym_TILDE] = ACTIONS(3201),
    [anon_sym_BANG] = ACTIONS(3201),
    [anon_sym_PLUS_PLUS] = ACTIONS(3201),
    [anon_sym_DASH_DASH] = ACTIONS(3201),
    [anon_sym_BSLASH] = ACTIONS(3199),
    [anon_sym_do] = ACTIONS(3201),
    [anon_sym_eval] = ACTIONS(3201),
    [anon_sym_await] = ACTIONS(3201),
    [anon_sym_field] = ACTIONS(3201),
    [anon_sym_local] = ACTIONS(3201),
    [anon_sym_dynamically] = ACTIONS(3201),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3201),
    [anon_sym_require] = ACTIONS(3201),
    [anon_sym_map] = ACTIONS(3201),
    [anon_sym_grep] = ACTIONS(3201),
    [anon_sym_sort] = ACTIONS(3201),
    [anon_sym_goto] = ACTIONS(3201),
    [anon_sym_return] = ACTIONS(3201),
    [anon_sym_undef] = ACTIONS(3201),
    [anon_sym_PERCENT2] = ACTIONS(3201),
    [anon_sym_AMP2] = ACTIONS(3201),
    [anon_sym_STAR2] = ACTIONS(3201),
    [anon_sym_use] = ACTIONS(3201),
    [anon_sym_no] = ACTIONS(3201),
    [anon_sym_for] = ACTIONS(3201),
    [anon_sym_foreach] = ACTIONS(3201),
    [anon_sym_last] = ACTIONS(3201),
    [anon_sym_next] = ACTIONS(3201),
    [anon_sym_redo] = ACTIONS(3201),
    [anon_sym_BEGIN] = ACTIONS(3201),
    [anon_sym_INIT] = ACTIONS(3201),
    [anon_sym_CHECK] = ACTIONS(3201),
    [anon_sym_UNITCHECK] = ACTIONS(3201),
    [anon_sym_END] = ACTIONS(3201),
    [anon_sym___FILE__] = ACTIONS(3201),
    [anon_sym___LINE__] = ACTIONS(3201),
    [anon_sym___PACKAGE__] = ACTIONS(3201),
    [anon_sym___SUB__] = ACTIONS(3201),
    [anon_sym_break] = ACTIONS(3201),
    [anon_sym_fork] = ACTIONS(3201),
    [anon_sym_getppid] = ACTIONS(3201),
    [anon_sym_time] = ACTIONS(3201),
    [anon_sym_times] = ACTIONS(3201),
    [anon_sym_wait] = ACTIONS(3201),
    [anon_sym_wantarray] = ACTIONS(3201),
    [anon_sym_abs] = ACTIONS(3201),
    [anon_sym_alarm] = ACTIONS(3201),
    [anon_sym_chop] = ACTIONS(3201),
    [anon_sym_chdir] = ACTIONS(3201),
    [anon_sym_close] = ACTIONS(3201),
    [anon_sym_closedir] = ACTIONS(3201),
    [anon_sym_caller] = ACTIONS(3201),
    [anon_sym_chomp] = ACTIONS(3201),
    [anon_sym_chr] = ACTIONS(3201),
    [anon_sym_cos] = ACTIONS(3201),
    [anon_sym_chroot] = ACTIONS(3201),
    [anon_sym_defined] = ACTIONS(3201),
    [anon_sym_delete] = ACTIONS(3201),
    [anon_sym_dbmclose] = ACTIONS(3201),
    [anon_sym_exists] = ACTIONS(3201),
    [anon_sym_exit] = ACTIONS(3201),
    [anon_sym_eof] = ACTIONS(3201),
    [anon_sym_exp] = ACTIONS(3201),
    [anon_sym_each] = ACTIONS(3201),
    [anon_sym_fc] = ACTIONS(3201),
    [anon_sym_fileno] = ACTIONS(3201),
    [anon_sym_gmtime] = ACTIONS(3201),
    [anon_sym_getc] = ACTIONS(3201),
    [anon_sym_getpgrp] = ACTIONS(3201),
    [anon_sym_getprotobyname] = ACTIONS(3201),
    [anon_sym_getpwname] = ACTIONS(3201),
    [anon_sym_getpwuid] = ACTIONS(3201),
    [anon_sym_getpeername] = ACTIONS(3201),
    [anon_sym_getnetbyname] = ACTIONS(3201),
    [anon_sym_getsockname] = ACTIONS(3201),
    [anon_sym_getgrnam] = ACTIONS(3201),
    [anon_sym_getgrgid] = ACTIONS(3201),
    [anon_sym_hex] = ACTIONS(3201),
    [anon_sym_int] = ACTIONS(3201),
    [anon_sym_keys] = ACTIONS(3201),
    [anon_sym_lc] = ACTIONS(3201),
    [anon_sym_lcfirst] = ACTIONS(3201),
    [anon_sym_length] = ACTIONS(3201),
    [anon_sym_localtime] = ACTIONS(3201),
    [anon_sym_log] = ACTIONS(3201),
    [anon_sym_lock] = ACTIONS(3201),
    [anon_sym_lstat] = ACTIONS(3201),
    [anon_sym_oct] = ACTIONS(3201),
    [anon_sym_ord] = ACTIONS(3201),
    [anon_sym_prototype] = ACTIONS(3201),
    [anon_sym_pop] = ACTIONS(3201),
    [anon_sym_pos] = ACTIONS(3201),
    [anon_sym_quotemeta] = ACTIONS(3201),
    [anon_sym_reset] = ACTIONS(3201),
    [anon_sym_rand] = ACTIONS(3201),
    [anon_sym_rmdir] = ACTIONS(3201),
    [anon_sym_readdir] = ACTIONS(3201),
    [anon_sym_readline] = ACTIONS(3201),
    [anon_sym_readpipe] = ACTIONS(3201),
    [anon_sym_rewinddir] = ACTIONS(3201),
    [anon_sym_readlink] = ACTIONS(3201),
    [anon_sym_ref] = ACTIONS(3201),
    [anon_sym_scalar] = ACTIONS(3201),
    [anon_sym_shift] = ACTIONS(3201),
    [anon_sym_sin] = ACTIONS(3201),
    [anon_sym_sleep] = ACTIONS(3201),
    [anon_sym_sqrt] = ACTIONS(3201),
    [anon_sym_srand] = ACTIONS(3201),
    [anon_sym_stat] = ACTIONS(3201),
    [anon_sym_study] = ACTIONS(3201),
    [anon_sym_tell] = ACTIONS(3201),
    [anon_sym_telldir] = ACTIONS(3201),
    [anon_sym_tied] = ACTIONS(3201),
    [anon_sym_uc] = ACTIONS(3201),
    [anon_sym_ucfirst] = ACTIONS(3201),
    [anon_sym_untie] = ACTIONS(3201),
    [anon_sym_umask] = ACTIONS(3201),
    [anon_sym_values] = ACTIONS(3201),
    [anon_sym_write] = ACTIONS(3201),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3201),
    [anon_sym___END__] = ACTIONS(3201),
    [anon_sym_EOT] = ACTIONS(3201),
    [anon_sym_q] = ACTIONS(3201),
    [anon_sym_qq] = ACTIONS(3201),
    [anon_sym_qw] = ACTIONS(3201),
    [anon_sym_qx] = ACTIONS(3201),
    [anon_sym_qr] = ACTIONS(3201),
    [anon_sym_m] = ACTIONS(3201),
    [anon_sym_s] = ACTIONS(3201),
    [anon_sym_tr] = ACTIONS(3201),
    [anon_sym_y] = ACTIONS(3201),
    [anon_sym_if] = ACTIONS(3201),
    [anon_sym_unless] = ACTIONS(3201),
    [anon_sym_while] = ACTIONS(3201),
    [anon_sym_until] = ACTIONS(3201),
    [aux_sym__bareword_token1] = ACTIONS(3201),
    [sym_number] = ACTIONS(3201),
    [anon_sym_true] = ACTIONS(3201),
    [anon_sym_false] = ACTIONS(3201),
    [sym__single_quote] = ACTIONS(3199),
    [sym__double_quote] = ACTIONS(3199),
    [sym__backtick_quote] = ACTIONS(3199),
    [sym__search_slash_quote] = ACTIONS(3199),
    [sym__open_readline_bracket] = ACTIONS(3199),
    [sym__open_fileglob_bracket] = ACTIONS(3199),
    [sym__PERLY_HEREDOC] = ACTIONS(3199),
    [sym__ctrl_z_hack] = ACTIONS(3199),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3199),
    [sym__filetest] = ACTIONS(3199),
  },
  [STATE(1512)] = {
    [sym_heredoc_content] = STATE(1512),
    [ts_builtin_sym_end] = ACTIONS(3203),
    [sym__identifier] = ACTIONS(3205),
    [anon_sym_LBRACE] = ACTIONS(3205),
    [anon_sym_SEMI] = ACTIONS(3205),
    [anon_sym_package] = ACTIONS(3205),
    [anon_sym_class] = ACTIONS(3205),
    [anon_sym_role] = ACTIONS(3205),
    [anon_sym_BUILD] = ACTIONS(3205),
    [anon_sym_ADJUST] = ACTIONS(3205),
    [anon_sym_DOLLAR] = ACTIONS(3205),
    [anon_sym_AT] = ACTIONS(3205),
    [anon_sym_my] = ACTIONS(3205),
    [anon_sym_extended] = ACTIONS(3205),
    [anon_sym_async] = ACTIONS(3205),
    [anon_sym_sub] = ACTIONS(3205),
    [anon_sym_method] = ACTIONS(3205),
    [anon_sym_LPAREN] = ACTIONS(3205),
    [anon_sym_continue] = ACTIONS(3205),
    [anon_sym_state] = ACTIONS(3205),
    [anon_sym_our] = ACTIONS(3205),
    [anon_sym_try] = ACTIONS(3205),
    [anon_sym_defer] = ACTIONS(3205),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3205),
    [anon_sym_LBRACK] = ACTIONS(3205),
    [anon_sym_LT_LT] = ACTIONS(3205),
    [anon_sym_SLASH_SLASH] = ACTIONS(3205),
    [anon_sym_PLUS] = ACTIONS(3205),
    [anon_sym_DASH] = ACTIONS(3205),
    [anon_sym_TILDE] = ACTIONS(3205),
    [anon_sym_BANG] = ACTIONS(3205),
    [anon_sym_PLUS_PLUS] = ACTIONS(3205),
    [anon_sym_DASH_DASH] = ACTIONS(3205),
    [anon_sym_BSLASH] = ACTIONS(3203),
    [anon_sym_do] = ACTIONS(3205),
    [anon_sym_eval] = ACTIONS(3205),
    [anon_sym_await] = ACTIONS(3205),
    [anon_sym_field] = ACTIONS(3205),
    [anon_sym_local] = ACTIONS(3205),
    [anon_sym_dynamically] = ACTIONS(3205),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3205),
    [anon_sym_require] = ACTIONS(3205),
    [anon_sym_map] = ACTIONS(3205),
    [anon_sym_grep] = ACTIONS(3205),
    [anon_sym_sort] = ACTIONS(3205),
    [anon_sym_goto] = ACTIONS(3205),
    [anon_sym_return] = ACTIONS(3205),
    [anon_sym_undef] = ACTIONS(3205),
    [anon_sym_PERCENT2] = ACTIONS(3205),
    [anon_sym_AMP2] = ACTIONS(3205),
    [anon_sym_STAR2] = ACTIONS(3205),
    [anon_sym_use] = ACTIONS(3205),
    [anon_sym_no] = ACTIONS(3205),
    [anon_sym_for] = ACTIONS(3205),
    [anon_sym_foreach] = ACTIONS(3205),
    [anon_sym_last] = ACTIONS(3205),
    [anon_sym_next] = ACTIONS(3205),
    [anon_sym_redo] = ACTIONS(3205),
    [anon_sym_BEGIN] = ACTIONS(3205),
    [anon_sym_INIT] = ACTIONS(3205),
    [anon_sym_CHECK] = ACTIONS(3205),
    [anon_sym_UNITCHECK] = ACTIONS(3205),
    [anon_sym_END] = ACTIONS(3205),
    [anon_sym___FILE__] = ACTIONS(3205),
    [anon_sym___LINE__] = ACTIONS(3205),
    [anon_sym___PACKAGE__] = ACTIONS(3205),
    [anon_sym___SUB__] = ACTIONS(3205),
    [anon_sym_break] = ACTIONS(3205),
    [anon_sym_fork] = ACTIONS(3205),
    [anon_sym_getppid] = ACTIONS(3205),
    [anon_sym_time] = ACTIONS(3205),
    [anon_sym_times] = ACTIONS(3205),
    [anon_sym_wait] = ACTIONS(3205),
    [anon_sym_wantarray] = ACTIONS(3205),
    [anon_sym_abs] = ACTIONS(3205),
    [anon_sym_alarm] = ACTIONS(3205),
    [anon_sym_chop] = ACTIONS(3205),
    [anon_sym_chdir] = ACTIONS(3205),
    [anon_sym_close] = ACTIONS(3205),
    [anon_sym_closedir] = ACTIONS(3205),
    [anon_sym_caller] = ACTIONS(3205),
    [anon_sym_chomp] = ACTIONS(3205),
    [anon_sym_chr] = ACTIONS(3205),
    [anon_sym_cos] = ACTIONS(3205),
    [anon_sym_chroot] = ACTIONS(3205),
    [anon_sym_defined] = ACTIONS(3205),
    [anon_sym_delete] = ACTIONS(3205),
    [anon_sym_dbmclose] = ACTIONS(3205),
    [anon_sym_exists] = ACTIONS(3205),
    [anon_sym_exit] = ACTIONS(3205),
    [anon_sym_eof] = ACTIONS(3205),
    [anon_sym_exp] = ACTIONS(3205),
    [anon_sym_each] = ACTIONS(3205),
    [anon_sym_fc] = ACTIONS(3205),
    [anon_sym_fileno] = ACTIONS(3205),
    [anon_sym_gmtime] = ACTIONS(3205),
    [anon_sym_getc] = ACTIONS(3205),
    [anon_sym_getpgrp] = ACTIONS(3205),
    [anon_sym_getprotobyname] = ACTIONS(3205),
    [anon_sym_getpwname] = ACTIONS(3205),
    [anon_sym_getpwuid] = ACTIONS(3205),
    [anon_sym_getpeername] = ACTIONS(3205),
    [anon_sym_getnetbyname] = ACTIONS(3205),
    [anon_sym_getsockname] = ACTIONS(3205),
    [anon_sym_getgrnam] = ACTIONS(3205),
    [anon_sym_getgrgid] = ACTIONS(3205),
    [anon_sym_hex] = ACTIONS(3205),
    [anon_sym_int] = ACTIONS(3205),
    [anon_sym_keys] = ACTIONS(3205),
    [anon_sym_lc] = ACTIONS(3205),
    [anon_sym_lcfirst] = ACTIONS(3205),
    [anon_sym_length] = ACTIONS(3205),
    [anon_sym_localtime] = ACTIONS(3205),
    [anon_sym_log] = ACTIONS(3205),
    [anon_sym_lock] = ACTIONS(3205),
    [anon_sym_lstat] = ACTIONS(3205),
    [anon_sym_oct] = ACTIONS(3205),
    [anon_sym_ord] = ACTIONS(3205),
    [anon_sym_prototype] = ACTIONS(3205),
    [anon_sym_pop] = ACTIONS(3205),
    [anon_sym_pos] = ACTIONS(3205),
    [anon_sym_quotemeta] = ACTIONS(3205),
    [anon_sym_reset] = ACTIONS(3205),
    [anon_sym_rand] = ACTIONS(3205),
    [anon_sym_rmdir] = ACTIONS(3205),
    [anon_sym_readdir] = ACTIONS(3205),
    [anon_sym_readline] = ACTIONS(3205),
    [anon_sym_readpipe] = ACTIONS(3205),
    [anon_sym_rewinddir] = ACTIONS(3205),
    [anon_sym_readlink] = ACTIONS(3205),
    [anon_sym_ref] = ACTIONS(3205),
    [anon_sym_scalar] = ACTIONS(3205),
    [anon_sym_shift] = ACTIONS(3205),
    [anon_sym_sin] = ACTIONS(3205),
    [anon_sym_sleep] = ACTIONS(3205),
    [anon_sym_sqrt] = ACTIONS(3205),
    [anon_sym_srand] = ACTIONS(3205),
    [anon_sym_stat] = ACTIONS(3205),
    [anon_sym_study] = ACTIONS(3205),
    [anon_sym_tell] = ACTIONS(3205),
    [anon_sym_telldir] = ACTIONS(3205),
    [anon_sym_tied] = ACTIONS(3205),
    [anon_sym_uc] = ACTIONS(3205),
    [anon_sym_ucfirst] = ACTIONS(3205),
    [anon_sym_untie] = ACTIONS(3205),
    [anon_sym_umask] = ACTIONS(3205),
    [anon_sym_values] = ACTIONS(3205),
    [anon_sym_write] = ACTIONS(3205),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3205),
    [anon_sym___END__] = ACTIONS(3205),
    [anon_sym_EOT] = ACTIONS(3205),
    [anon_sym_q] = ACTIONS(3205),
    [anon_sym_qq] = ACTIONS(3205),
    [anon_sym_qw] = ACTIONS(3205),
    [anon_sym_qx] = ACTIONS(3205),
    [anon_sym_qr] = ACTIONS(3205),
    [anon_sym_m] = ACTIONS(3205),
    [anon_sym_s] = ACTIONS(3205),
    [anon_sym_tr] = ACTIONS(3205),
    [anon_sym_y] = ACTIONS(3205),
    [anon_sym_if] = ACTIONS(3205),
    [anon_sym_unless] = ACTIONS(3205),
    [anon_sym_while] = ACTIONS(3205),
    [anon_sym_until] = ACTIONS(3205),
    [aux_sym__bareword_token1] = ACTIONS(3205),
    [sym_number] = ACTIONS(3205),
    [anon_sym_true] = ACTIONS(3205),
    [anon_sym_false] = ACTIONS(3205),
    [sym__single_quote] = ACTIONS(3203),
    [sym__double_quote] = ACTIONS(3203),
    [sym__backtick_quote] = ACTIONS(3203),
    [sym__search_slash_quote] = ACTIONS(3203),
    [sym__open_readline_bracket] = ACTIONS(3203),
    [sym__open_fileglob_bracket] = ACTIONS(3203),
    [sym__PERLY_HEREDOC] = ACTIONS(3203),
    [sym__ctrl_z_hack] = ACTIONS(3203),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3203),
    [sym__filetest] = ACTIONS(3203),
  },
  [STATE(1513)] = {
    [sym_heredoc_content] = STATE(1513),
    [ts_builtin_sym_end] = ACTIONS(3207),
    [sym__identifier] = ACTIONS(3209),
    [anon_sym_LBRACE] = ACTIONS(3209),
    [anon_sym_SEMI] = ACTIONS(3209),
    [anon_sym_package] = ACTIONS(3209),
    [anon_sym_class] = ACTIONS(3209),
    [anon_sym_role] = ACTIONS(3209),
    [anon_sym_BUILD] = ACTIONS(3209),
    [anon_sym_ADJUST] = ACTIONS(3209),
    [anon_sym_DOLLAR] = ACTIONS(3209),
    [anon_sym_AT] = ACTIONS(3209),
    [anon_sym_my] = ACTIONS(3209),
    [anon_sym_extended] = ACTIONS(3209),
    [anon_sym_async] = ACTIONS(3209),
    [anon_sym_sub] = ACTIONS(3209),
    [anon_sym_method] = ACTIONS(3209),
    [anon_sym_LPAREN] = ACTIONS(3209),
    [anon_sym_continue] = ACTIONS(3209),
    [anon_sym_state] = ACTIONS(3209),
    [anon_sym_our] = ACTIONS(3209),
    [anon_sym_try] = ACTIONS(3209),
    [anon_sym_defer] = ACTIONS(3209),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3209),
    [anon_sym_LBRACK] = ACTIONS(3209),
    [anon_sym_LT_LT] = ACTIONS(3209),
    [anon_sym_SLASH_SLASH] = ACTIONS(3209),
    [anon_sym_PLUS] = ACTIONS(3209),
    [anon_sym_DASH] = ACTIONS(3209),
    [anon_sym_TILDE] = ACTIONS(3209),
    [anon_sym_BANG] = ACTIONS(3209),
    [anon_sym_PLUS_PLUS] = ACTIONS(3209),
    [anon_sym_DASH_DASH] = ACTIONS(3209),
    [anon_sym_BSLASH] = ACTIONS(3207),
    [anon_sym_do] = ACTIONS(3209),
    [anon_sym_eval] = ACTIONS(3209),
    [anon_sym_await] = ACTIONS(3209),
    [anon_sym_field] = ACTIONS(3209),
    [anon_sym_local] = ACTIONS(3209),
    [anon_sym_dynamically] = ACTIONS(3209),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3209),
    [anon_sym_require] = ACTIONS(3209),
    [anon_sym_map] = ACTIONS(3209),
    [anon_sym_grep] = ACTIONS(3209),
    [anon_sym_sort] = ACTIONS(3209),
    [anon_sym_goto] = ACTIONS(3209),
    [anon_sym_return] = ACTIONS(3209),
    [anon_sym_undef] = ACTIONS(3209),
    [anon_sym_PERCENT2] = ACTIONS(3209),
    [anon_sym_AMP2] = ACTIONS(3209),
    [anon_sym_STAR2] = ACTIONS(3209),
    [anon_sym_use] = ACTIONS(3209),
    [anon_sym_no] = ACTIONS(3209),
    [anon_sym_for] = ACTIONS(3209),
    [anon_sym_foreach] = ACTIONS(3209),
    [anon_sym_last] = ACTIONS(3209),
    [anon_sym_next] = ACTIONS(3209),
    [anon_sym_redo] = ACTIONS(3209),
    [anon_sym_BEGIN] = ACTIONS(3209),
    [anon_sym_INIT] = ACTIONS(3209),
    [anon_sym_CHECK] = ACTIONS(3209),
    [anon_sym_UNITCHECK] = ACTIONS(3209),
    [anon_sym_END] = ACTIONS(3209),
    [anon_sym___FILE__] = ACTIONS(3209),
    [anon_sym___LINE__] = ACTIONS(3209),
    [anon_sym___PACKAGE__] = ACTIONS(3209),
    [anon_sym___SUB__] = ACTIONS(3209),
    [anon_sym_break] = ACTIONS(3209),
    [anon_sym_fork] = ACTIONS(3209),
    [anon_sym_getppid] = ACTIONS(3209),
    [anon_sym_time] = ACTIONS(3209),
    [anon_sym_times] = ACTIONS(3209),
    [anon_sym_wait] = ACTIONS(3209),
    [anon_sym_wantarray] = ACTIONS(3209),
    [anon_sym_abs] = ACTIONS(3209),
    [anon_sym_alarm] = ACTIONS(3209),
    [anon_sym_chop] = ACTIONS(3209),
    [anon_sym_chdir] = ACTIONS(3209),
    [anon_sym_close] = ACTIONS(3209),
    [anon_sym_closedir] = ACTIONS(3209),
    [anon_sym_caller] = ACTIONS(3209),
    [anon_sym_chomp] = ACTIONS(3209),
    [anon_sym_chr] = ACTIONS(3209),
    [anon_sym_cos] = ACTIONS(3209),
    [anon_sym_chroot] = ACTIONS(3209),
    [anon_sym_defined] = ACTIONS(3209),
    [anon_sym_delete] = ACTIONS(3209),
    [anon_sym_dbmclose] = ACTIONS(3209),
    [anon_sym_exists] = ACTIONS(3209),
    [anon_sym_exit] = ACTIONS(3209),
    [anon_sym_eof] = ACTIONS(3209),
    [anon_sym_exp] = ACTIONS(3209),
    [anon_sym_each] = ACTIONS(3209),
    [anon_sym_fc] = ACTIONS(3209),
    [anon_sym_fileno] = ACTIONS(3209),
    [anon_sym_gmtime] = ACTIONS(3209),
    [anon_sym_getc] = ACTIONS(3209),
    [anon_sym_getpgrp] = ACTIONS(3209),
    [anon_sym_getprotobyname] = ACTIONS(3209),
    [anon_sym_getpwname] = ACTIONS(3209),
    [anon_sym_getpwuid] = ACTIONS(3209),
    [anon_sym_getpeername] = ACTIONS(3209),
    [anon_sym_getnetbyname] = ACTIONS(3209),
    [anon_sym_getsockname] = ACTIONS(3209),
    [anon_sym_getgrnam] = ACTIONS(3209),
    [anon_sym_getgrgid] = ACTIONS(3209),
    [anon_sym_hex] = ACTIONS(3209),
    [anon_sym_int] = ACTIONS(3209),
    [anon_sym_keys] = ACTIONS(3209),
    [anon_sym_lc] = ACTIONS(3209),
    [anon_sym_lcfirst] = ACTIONS(3209),
    [anon_sym_length] = ACTIONS(3209),
    [anon_sym_localtime] = ACTIONS(3209),
    [anon_sym_log] = ACTIONS(3209),
    [anon_sym_lock] = ACTIONS(3209),
    [anon_sym_lstat] = ACTIONS(3209),
    [anon_sym_oct] = ACTIONS(3209),
    [anon_sym_ord] = ACTIONS(3209),
    [anon_sym_prototype] = ACTIONS(3209),
    [anon_sym_pop] = ACTIONS(3209),
    [anon_sym_pos] = ACTIONS(3209),
    [anon_sym_quotemeta] = ACTIONS(3209),
    [anon_sym_reset] = ACTIONS(3209),
    [anon_sym_rand] = ACTIONS(3209),
    [anon_sym_rmdir] = ACTIONS(3209),
    [anon_sym_readdir] = ACTIONS(3209),
    [anon_sym_readline] = ACTIONS(3209),
    [anon_sym_readpipe] = ACTIONS(3209),
    [anon_sym_rewinddir] = ACTIONS(3209),
    [anon_sym_readlink] = ACTIONS(3209),
    [anon_sym_ref] = ACTIONS(3209),
    [anon_sym_scalar] = ACTIONS(3209),
    [anon_sym_shift] = ACTIONS(3209),
    [anon_sym_sin] = ACTIONS(3209),
    [anon_sym_sleep] = ACTIONS(3209),
    [anon_sym_sqrt] = ACTIONS(3209),
    [anon_sym_srand] = ACTIONS(3209),
    [anon_sym_stat] = ACTIONS(3209),
    [anon_sym_study] = ACTIONS(3209),
    [anon_sym_tell] = ACTIONS(3209),
    [anon_sym_telldir] = ACTIONS(3209),
    [anon_sym_tied] = ACTIONS(3209),
    [anon_sym_uc] = ACTIONS(3209),
    [anon_sym_ucfirst] = ACTIONS(3209),
    [anon_sym_untie] = ACTIONS(3209),
    [anon_sym_umask] = ACTIONS(3209),
    [anon_sym_values] = ACTIONS(3209),
    [anon_sym_write] = ACTIONS(3209),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3209),
    [anon_sym___END__] = ACTIONS(3209),
    [anon_sym_EOT] = ACTIONS(3209),
    [anon_sym_q] = ACTIONS(3209),
    [anon_sym_qq] = ACTIONS(3209),
    [anon_sym_qw] = ACTIONS(3209),
    [anon_sym_qx] = ACTIONS(3209),
    [anon_sym_qr] = ACTIONS(3209),
    [anon_sym_m] = ACTIONS(3209),
    [anon_sym_s] = ACTIONS(3209),
    [anon_sym_tr] = ACTIONS(3209),
    [anon_sym_y] = ACTIONS(3209),
    [anon_sym_if] = ACTIONS(3209),
    [anon_sym_unless] = ACTIONS(3209),
    [anon_sym_while] = ACTIONS(3209),
    [anon_sym_until] = ACTIONS(3209),
    [aux_sym__bareword_token1] = ACTIONS(3209),
    [sym_number] = ACTIONS(3209),
    [anon_sym_true] = ACTIONS(3209),
    [anon_sym_false] = ACTIONS(3209),
    [sym__single_quote] = ACTIONS(3207),
    [sym__double_quote] = ACTIONS(3207),
    [sym__backtick_quote] = ACTIONS(3207),
    [sym__search_slash_quote] = ACTIONS(3207),
    [sym__open_readline_bracket] = ACTIONS(3207),
    [sym__open_fileglob_bracket] = ACTIONS(3207),
    [sym__PERLY_HEREDOC] = ACTIONS(3207),
    [sym__ctrl_z_hack] = ACTIONS(3207),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3207),
    [sym__filetest] = ACTIONS(3207),
  },
  [STATE(1514)] = {
    [sym_heredoc_content] = STATE(1514),
    [ts_builtin_sym_end] = ACTIONS(3211),
    [sym__identifier] = ACTIONS(3213),
    [anon_sym_LBRACE] = ACTIONS(3213),
    [anon_sym_SEMI] = ACTIONS(3213),
    [anon_sym_package] = ACTIONS(3213),
    [anon_sym_class] = ACTIONS(3213),
    [anon_sym_role] = ACTIONS(3213),
    [anon_sym_BUILD] = ACTIONS(3213),
    [anon_sym_ADJUST] = ACTIONS(3213),
    [anon_sym_DOLLAR] = ACTIONS(3213),
    [anon_sym_AT] = ACTIONS(3213),
    [anon_sym_my] = ACTIONS(3213),
    [anon_sym_extended] = ACTIONS(3213),
    [anon_sym_async] = ACTIONS(3213),
    [anon_sym_sub] = ACTIONS(3213),
    [anon_sym_method] = ACTIONS(3213),
    [anon_sym_LPAREN] = ACTIONS(3213),
    [anon_sym_continue] = ACTIONS(3213),
    [anon_sym_state] = ACTIONS(3213),
    [anon_sym_our] = ACTIONS(3213),
    [anon_sym_try] = ACTIONS(3213),
    [anon_sym_defer] = ACTIONS(3213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3213),
    [anon_sym_LBRACK] = ACTIONS(3213),
    [anon_sym_LT_LT] = ACTIONS(3213),
    [anon_sym_SLASH_SLASH] = ACTIONS(3213),
    [anon_sym_PLUS] = ACTIONS(3213),
    [anon_sym_DASH] = ACTIONS(3213),
    [anon_sym_TILDE] = ACTIONS(3213),
    [anon_sym_BANG] = ACTIONS(3213),
    [anon_sym_PLUS_PLUS] = ACTIONS(3213),
    [anon_sym_DASH_DASH] = ACTIONS(3213),
    [anon_sym_BSLASH] = ACTIONS(3211),
    [anon_sym_do] = ACTIONS(3213),
    [anon_sym_eval] = ACTIONS(3213),
    [anon_sym_await] = ACTIONS(3213),
    [anon_sym_field] = ACTIONS(3213),
    [anon_sym_local] = ACTIONS(3213),
    [anon_sym_dynamically] = ACTIONS(3213),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3213),
    [anon_sym_require] = ACTIONS(3213),
    [anon_sym_map] = ACTIONS(3213),
    [anon_sym_grep] = ACTIONS(3213),
    [anon_sym_sort] = ACTIONS(3213),
    [anon_sym_goto] = ACTIONS(3213),
    [anon_sym_return] = ACTIONS(3213),
    [anon_sym_undef] = ACTIONS(3213),
    [anon_sym_PERCENT2] = ACTIONS(3213),
    [anon_sym_AMP2] = ACTIONS(3213),
    [anon_sym_STAR2] = ACTIONS(3213),
    [anon_sym_use] = ACTIONS(3213),
    [anon_sym_no] = ACTIONS(3213),
    [anon_sym_for] = ACTIONS(3213),
    [anon_sym_foreach] = ACTIONS(3213),
    [anon_sym_last] = ACTIONS(3213),
    [anon_sym_next] = ACTIONS(3213),
    [anon_sym_redo] = ACTIONS(3213),
    [anon_sym_BEGIN] = ACTIONS(3213),
    [anon_sym_INIT] = ACTIONS(3213),
    [anon_sym_CHECK] = ACTIONS(3213),
    [anon_sym_UNITCHECK] = ACTIONS(3213),
    [anon_sym_END] = ACTIONS(3213),
    [anon_sym___FILE__] = ACTIONS(3213),
    [anon_sym___LINE__] = ACTIONS(3213),
    [anon_sym___PACKAGE__] = ACTIONS(3213),
    [anon_sym___SUB__] = ACTIONS(3213),
    [anon_sym_break] = ACTIONS(3213),
    [anon_sym_fork] = ACTIONS(3213),
    [anon_sym_getppid] = ACTIONS(3213),
    [anon_sym_time] = ACTIONS(3213),
    [anon_sym_times] = ACTIONS(3213),
    [anon_sym_wait] = ACTIONS(3213),
    [anon_sym_wantarray] = ACTIONS(3213),
    [anon_sym_abs] = ACTIONS(3213),
    [anon_sym_alarm] = ACTIONS(3213),
    [anon_sym_chop] = ACTIONS(3213),
    [anon_sym_chdir] = ACTIONS(3213),
    [anon_sym_close] = ACTIONS(3213),
    [anon_sym_closedir] = ACTIONS(3213),
    [anon_sym_caller] = ACTIONS(3213),
    [anon_sym_chomp] = ACTIONS(3213),
    [anon_sym_chr] = ACTIONS(3213),
    [anon_sym_cos] = ACTIONS(3213),
    [anon_sym_chroot] = ACTIONS(3213),
    [anon_sym_defined] = ACTIONS(3213),
    [anon_sym_delete] = ACTIONS(3213),
    [anon_sym_dbmclose] = ACTIONS(3213),
    [anon_sym_exists] = ACTIONS(3213),
    [anon_sym_exit] = ACTIONS(3213),
    [anon_sym_eof] = ACTIONS(3213),
    [anon_sym_exp] = ACTIONS(3213),
    [anon_sym_each] = ACTIONS(3213),
    [anon_sym_fc] = ACTIONS(3213),
    [anon_sym_fileno] = ACTIONS(3213),
    [anon_sym_gmtime] = ACTIONS(3213),
    [anon_sym_getc] = ACTIONS(3213),
    [anon_sym_getpgrp] = ACTIONS(3213),
    [anon_sym_getprotobyname] = ACTIONS(3213),
    [anon_sym_getpwname] = ACTIONS(3213),
    [anon_sym_getpwuid] = ACTIONS(3213),
    [anon_sym_getpeername] = ACTIONS(3213),
    [anon_sym_getnetbyname] = ACTIONS(3213),
    [anon_sym_getsockname] = ACTIONS(3213),
    [anon_sym_getgrnam] = ACTIONS(3213),
    [anon_sym_getgrgid] = ACTIONS(3213),
    [anon_sym_hex] = ACTIONS(3213),
    [anon_sym_int] = ACTIONS(3213),
    [anon_sym_keys] = ACTIONS(3213),
    [anon_sym_lc] = ACTIONS(3213),
    [anon_sym_lcfirst] = ACTIONS(3213),
    [anon_sym_length] = ACTIONS(3213),
    [anon_sym_localtime] = ACTIONS(3213),
    [anon_sym_log] = ACTIONS(3213),
    [anon_sym_lock] = ACTIONS(3213),
    [anon_sym_lstat] = ACTIONS(3213),
    [anon_sym_oct] = ACTIONS(3213),
    [anon_sym_ord] = ACTIONS(3213),
    [anon_sym_prototype] = ACTIONS(3213),
    [anon_sym_pop] = ACTIONS(3213),
    [anon_sym_pos] = ACTIONS(3213),
    [anon_sym_quotemeta] = ACTIONS(3213),
    [anon_sym_reset] = ACTIONS(3213),
    [anon_sym_rand] = ACTIONS(3213),
    [anon_sym_rmdir] = ACTIONS(3213),
    [anon_sym_readdir] = ACTIONS(3213),
    [anon_sym_readline] = ACTIONS(3213),
    [anon_sym_readpipe] = ACTIONS(3213),
    [anon_sym_rewinddir] = ACTIONS(3213),
    [anon_sym_readlink] = ACTIONS(3213),
    [anon_sym_ref] = ACTIONS(3213),
    [anon_sym_scalar] = ACTIONS(3213),
    [anon_sym_shift] = ACTIONS(3213),
    [anon_sym_sin] = ACTIONS(3213),
    [anon_sym_sleep] = ACTIONS(3213),
    [anon_sym_sqrt] = ACTIONS(3213),
    [anon_sym_srand] = ACTIONS(3213),
    [anon_sym_stat] = ACTIONS(3213),
    [anon_sym_study] = ACTIONS(3213),
    [anon_sym_tell] = ACTIONS(3213),
    [anon_sym_telldir] = ACTIONS(3213),
    [anon_sym_tied] = ACTIONS(3213),
    [anon_sym_uc] = ACTIONS(3213),
    [anon_sym_ucfirst] = ACTIONS(3213),
    [anon_sym_untie] = ACTIONS(3213),
    [anon_sym_umask] = ACTIONS(3213),
    [anon_sym_values] = ACTIONS(3213),
    [anon_sym_write] = ACTIONS(3213),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3213),
    [anon_sym___END__] = ACTIONS(3213),
    [anon_sym_EOT] = ACTIONS(3213),
    [anon_sym_q] = ACTIONS(3213),
    [anon_sym_qq] = ACTIONS(3213),
    [anon_sym_qw] = ACTIONS(3213),
    [anon_sym_qx] = ACTIONS(3213),
    [anon_sym_qr] = ACTIONS(3213),
    [anon_sym_m] = ACTIONS(3213),
    [anon_sym_s] = ACTIONS(3213),
    [anon_sym_tr] = ACTIONS(3213),
    [anon_sym_y] = ACTIONS(3213),
    [anon_sym_if] = ACTIONS(3213),
    [anon_sym_unless] = ACTIONS(3213),
    [anon_sym_while] = ACTIONS(3213),
    [anon_sym_until] = ACTIONS(3213),
    [aux_sym__bareword_token1] = ACTIONS(3213),
    [sym_number] = ACTIONS(3213),
    [anon_sym_true] = ACTIONS(3213),
    [anon_sym_false] = ACTIONS(3213),
    [sym__single_quote] = ACTIONS(3211),
    [sym__double_quote] = ACTIONS(3211),
    [sym__backtick_quote] = ACTIONS(3211),
    [sym__search_slash_quote] = ACTIONS(3211),
    [sym__open_readline_bracket] = ACTIONS(3211),
    [sym__open_fileglob_bracket] = ACTIONS(3211),
    [sym__PERLY_HEREDOC] = ACTIONS(3211),
    [sym__ctrl_z_hack] = ACTIONS(3211),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3211),
    [sym__filetest] = ACTIONS(3211),
  },
  [STATE(1515)] = {
    [sym_heredoc_content] = STATE(1515),
    [ts_builtin_sym_end] = ACTIONS(3215),
    [sym__identifier] = ACTIONS(3217),
    [anon_sym_LBRACE] = ACTIONS(3217),
    [anon_sym_SEMI] = ACTIONS(3217),
    [anon_sym_package] = ACTIONS(3217),
    [anon_sym_class] = ACTIONS(3217),
    [anon_sym_role] = ACTIONS(3217),
    [anon_sym_BUILD] = ACTIONS(3217),
    [anon_sym_ADJUST] = ACTIONS(3217),
    [anon_sym_DOLLAR] = ACTIONS(3217),
    [anon_sym_AT] = ACTIONS(3217),
    [anon_sym_my] = ACTIONS(3217),
    [anon_sym_extended] = ACTIONS(3217),
    [anon_sym_async] = ACTIONS(3217),
    [anon_sym_sub] = ACTIONS(3217),
    [anon_sym_method] = ACTIONS(3217),
    [anon_sym_LPAREN] = ACTIONS(3217),
    [anon_sym_continue] = ACTIONS(3217),
    [anon_sym_state] = ACTIONS(3217),
    [anon_sym_our] = ACTIONS(3217),
    [anon_sym_try] = ACTIONS(3217),
    [anon_sym_defer] = ACTIONS(3217),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3217),
    [anon_sym_LBRACK] = ACTIONS(3217),
    [anon_sym_LT_LT] = ACTIONS(3217),
    [anon_sym_SLASH_SLASH] = ACTIONS(3217),
    [anon_sym_PLUS] = ACTIONS(3217),
    [anon_sym_DASH] = ACTIONS(3217),
    [anon_sym_TILDE] = ACTIONS(3217),
    [anon_sym_BANG] = ACTIONS(3217),
    [anon_sym_PLUS_PLUS] = ACTIONS(3217),
    [anon_sym_DASH_DASH] = ACTIONS(3217),
    [anon_sym_BSLASH] = ACTIONS(3215),
    [anon_sym_do] = ACTIONS(3217),
    [anon_sym_eval] = ACTIONS(3217),
    [anon_sym_await] = ACTIONS(3217),
    [anon_sym_field] = ACTIONS(3217),
    [anon_sym_local] = ACTIONS(3217),
    [anon_sym_dynamically] = ACTIONS(3217),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3217),
    [anon_sym_require] = ACTIONS(3217),
    [anon_sym_map] = ACTIONS(3217),
    [anon_sym_grep] = ACTIONS(3217),
    [anon_sym_sort] = ACTIONS(3217),
    [anon_sym_goto] = ACTIONS(3217),
    [anon_sym_return] = ACTIONS(3217),
    [anon_sym_undef] = ACTIONS(3217),
    [anon_sym_PERCENT2] = ACTIONS(3217),
    [anon_sym_AMP2] = ACTIONS(3217),
    [anon_sym_STAR2] = ACTIONS(3217),
    [anon_sym_use] = ACTIONS(3217),
    [anon_sym_no] = ACTIONS(3217),
    [anon_sym_for] = ACTIONS(3217),
    [anon_sym_foreach] = ACTIONS(3217),
    [anon_sym_last] = ACTIONS(3217),
    [anon_sym_next] = ACTIONS(3217),
    [anon_sym_redo] = ACTIONS(3217),
    [anon_sym_BEGIN] = ACTIONS(3217),
    [anon_sym_INIT] = ACTIONS(3217),
    [anon_sym_CHECK] = ACTIONS(3217),
    [anon_sym_UNITCHECK] = ACTIONS(3217),
    [anon_sym_END] = ACTIONS(3217),
    [anon_sym___FILE__] = ACTIONS(3217),
    [anon_sym___LINE__] = ACTIONS(3217),
    [anon_sym___PACKAGE__] = ACTIONS(3217),
    [anon_sym___SUB__] = ACTIONS(3217),
    [anon_sym_break] = ACTIONS(3217),
    [anon_sym_fork] = ACTIONS(3217),
    [anon_sym_getppid] = ACTIONS(3217),
    [anon_sym_time] = ACTIONS(3217),
    [anon_sym_times] = ACTIONS(3217),
    [anon_sym_wait] = ACTIONS(3217),
    [anon_sym_wantarray] = ACTIONS(3217),
    [anon_sym_abs] = ACTIONS(3217),
    [anon_sym_alarm] = ACTIONS(3217),
    [anon_sym_chop] = ACTIONS(3217),
    [anon_sym_chdir] = ACTIONS(3217),
    [anon_sym_close] = ACTIONS(3217),
    [anon_sym_closedir] = ACTIONS(3217),
    [anon_sym_caller] = ACTIONS(3217),
    [anon_sym_chomp] = ACTIONS(3217),
    [anon_sym_chr] = ACTIONS(3217),
    [anon_sym_cos] = ACTIONS(3217),
    [anon_sym_chroot] = ACTIONS(3217),
    [anon_sym_defined] = ACTIONS(3217),
    [anon_sym_delete] = ACTIONS(3217),
    [anon_sym_dbmclose] = ACTIONS(3217),
    [anon_sym_exists] = ACTIONS(3217),
    [anon_sym_exit] = ACTIONS(3217),
    [anon_sym_eof] = ACTIONS(3217),
    [anon_sym_exp] = ACTIONS(3217),
    [anon_sym_each] = ACTIONS(3217),
    [anon_sym_fc] = ACTIONS(3217),
    [anon_sym_fileno] = ACTIONS(3217),
    [anon_sym_gmtime] = ACTIONS(3217),
    [anon_sym_getc] = ACTIONS(3217),
    [anon_sym_getpgrp] = ACTIONS(3217),
    [anon_sym_getprotobyname] = ACTIONS(3217),
    [anon_sym_getpwname] = ACTIONS(3217),
    [anon_sym_getpwuid] = ACTIONS(3217),
    [anon_sym_getpeername] = ACTIONS(3217),
    [anon_sym_getnetbyname] = ACTIONS(3217),
    [anon_sym_getsockname] = ACTIONS(3217),
    [anon_sym_getgrnam] = ACTIONS(3217),
    [anon_sym_getgrgid] = ACTIONS(3217),
    [anon_sym_hex] = ACTIONS(3217),
    [anon_sym_int] = ACTIONS(3217),
    [anon_sym_keys] = ACTIONS(3217),
    [anon_sym_lc] = ACTIONS(3217),
    [anon_sym_lcfirst] = ACTIONS(3217),
    [anon_sym_length] = ACTIONS(3217),
    [anon_sym_localtime] = ACTIONS(3217),
    [anon_sym_log] = ACTIONS(3217),
    [anon_sym_lock] = ACTIONS(3217),
    [anon_sym_lstat] = ACTIONS(3217),
    [anon_sym_oct] = ACTIONS(3217),
    [anon_sym_ord] = ACTIONS(3217),
    [anon_sym_prototype] = ACTIONS(3217),
    [anon_sym_pop] = ACTIONS(3217),
    [anon_sym_pos] = ACTIONS(3217),
    [anon_sym_quotemeta] = ACTIONS(3217),
    [anon_sym_reset] = ACTIONS(3217),
    [anon_sym_rand] = ACTIONS(3217),
    [anon_sym_rmdir] = ACTIONS(3217),
    [anon_sym_readdir] = ACTIONS(3217),
    [anon_sym_readline] = ACTIONS(3217),
    [anon_sym_readpipe] = ACTIONS(3217),
    [anon_sym_rewinddir] = ACTIONS(3217),
    [anon_sym_readlink] = ACTIONS(3217),
    [anon_sym_ref] = ACTIONS(3217),
    [anon_sym_scalar] = ACTIONS(3217),
    [anon_sym_shift] = ACTIONS(3217),
    [anon_sym_sin] = ACTIONS(3217),
    [anon_sym_sleep] = ACTIONS(3217),
    [anon_sym_sqrt] = ACTIONS(3217),
    [anon_sym_srand] = ACTIONS(3217),
    [anon_sym_stat] = ACTIONS(3217),
    [anon_sym_study] = ACTIONS(3217),
    [anon_sym_tell] = ACTIONS(3217),
    [anon_sym_telldir] = ACTIONS(3217),
    [anon_sym_tied] = ACTIONS(3217),
    [anon_sym_uc] = ACTIONS(3217),
    [anon_sym_ucfirst] = ACTIONS(3217),
    [anon_sym_untie] = ACTIONS(3217),
    [anon_sym_umask] = ACTIONS(3217),
    [anon_sym_values] = ACTIONS(3217),
    [anon_sym_write] = ACTIONS(3217),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3217),
    [anon_sym___END__] = ACTIONS(3217),
    [anon_sym_EOT] = ACTIONS(3217),
    [anon_sym_q] = ACTIONS(3217),
    [anon_sym_qq] = ACTIONS(3217),
    [anon_sym_qw] = ACTIONS(3217),
    [anon_sym_qx] = ACTIONS(3217),
    [anon_sym_qr] = ACTIONS(3217),
    [anon_sym_m] = ACTIONS(3217),
    [anon_sym_s] = ACTIONS(3217),
    [anon_sym_tr] = ACTIONS(3217),
    [anon_sym_y] = ACTIONS(3217),
    [anon_sym_if] = ACTIONS(3217),
    [anon_sym_unless] = ACTIONS(3217),
    [anon_sym_while] = ACTIONS(3217),
    [anon_sym_until] = ACTIONS(3217),
    [aux_sym__bareword_token1] = ACTIONS(3217),
    [sym_number] = ACTIONS(3217),
    [anon_sym_true] = ACTIONS(3217),
    [anon_sym_false] = ACTIONS(3217),
    [sym__single_quote] = ACTIONS(3215),
    [sym__double_quote] = ACTIONS(3215),
    [sym__backtick_quote] = ACTIONS(3215),
    [sym__search_slash_quote] = ACTIONS(3215),
    [sym__open_readline_bracket] = ACTIONS(3215),
    [sym__open_fileglob_bracket] = ACTIONS(3215),
    [sym__PERLY_HEREDOC] = ACTIONS(3215),
    [sym__ctrl_z_hack] = ACTIONS(3215),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3215),
    [sym__filetest] = ACTIONS(3215),
  },
  [STATE(1516)] = {
    [sym_heredoc_content] = STATE(1516),
    [ts_builtin_sym_end] = ACTIONS(3219),
    [sym__identifier] = ACTIONS(3221),
    [anon_sym_LBRACE] = ACTIONS(3221),
    [anon_sym_SEMI] = ACTIONS(3221),
    [anon_sym_package] = ACTIONS(3221),
    [anon_sym_class] = ACTIONS(3221),
    [anon_sym_role] = ACTIONS(3221),
    [anon_sym_BUILD] = ACTIONS(3221),
    [anon_sym_ADJUST] = ACTIONS(3221),
    [anon_sym_DOLLAR] = ACTIONS(3221),
    [anon_sym_AT] = ACTIONS(3221),
    [anon_sym_my] = ACTIONS(3221),
    [anon_sym_extended] = ACTIONS(3221),
    [anon_sym_async] = ACTIONS(3221),
    [anon_sym_sub] = ACTIONS(3221),
    [anon_sym_method] = ACTIONS(3221),
    [anon_sym_LPAREN] = ACTIONS(3221),
    [anon_sym_continue] = ACTIONS(3221),
    [anon_sym_state] = ACTIONS(3221),
    [anon_sym_our] = ACTIONS(3221),
    [anon_sym_try] = ACTIONS(3221),
    [anon_sym_defer] = ACTIONS(3221),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3221),
    [anon_sym_LBRACK] = ACTIONS(3221),
    [anon_sym_LT_LT] = ACTIONS(3221),
    [anon_sym_SLASH_SLASH] = ACTIONS(3221),
    [anon_sym_PLUS] = ACTIONS(3221),
    [anon_sym_DASH] = ACTIONS(3221),
    [anon_sym_TILDE] = ACTIONS(3221),
    [anon_sym_BANG] = ACTIONS(3221),
    [anon_sym_PLUS_PLUS] = ACTIONS(3221),
    [anon_sym_DASH_DASH] = ACTIONS(3221),
    [anon_sym_BSLASH] = ACTIONS(3219),
    [anon_sym_do] = ACTIONS(3221),
    [anon_sym_eval] = ACTIONS(3221),
    [anon_sym_await] = ACTIONS(3221),
    [anon_sym_field] = ACTIONS(3221),
    [anon_sym_local] = ACTIONS(3221),
    [anon_sym_dynamically] = ACTIONS(3221),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3221),
    [anon_sym_require] = ACTIONS(3221),
    [anon_sym_map] = ACTIONS(3221),
    [anon_sym_grep] = ACTIONS(3221),
    [anon_sym_sort] = ACTIONS(3221),
    [anon_sym_goto] = ACTIONS(3221),
    [anon_sym_return] = ACTIONS(3221),
    [anon_sym_undef] = ACTIONS(3221),
    [anon_sym_PERCENT2] = ACTIONS(3221),
    [anon_sym_AMP2] = ACTIONS(3221),
    [anon_sym_STAR2] = ACTIONS(3221),
    [anon_sym_use] = ACTIONS(3221),
    [anon_sym_no] = ACTIONS(3221),
    [anon_sym_for] = ACTIONS(3221),
    [anon_sym_foreach] = ACTIONS(3221),
    [anon_sym_last] = ACTIONS(3221),
    [anon_sym_next] = ACTIONS(3221),
    [anon_sym_redo] = ACTIONS(3221),
    [anon_sym_BEGIN] = ACTIONS(3221),
    [anon_sym_INIT] = ACTIONS(3221),
    [anon_sym_CHECK] = ACTIONS(3221),
    [anon_sym_UNITCHECK] = ACTIONS(3221),
    [anon_sym_END] = ACTIONS(3221),
    [anon_sym___FILE__] = ACTIONS(3221),
    [anon_sym___LINE__] = ACTIONS(3221),
    [anon_sym___PACKAGE__] = ACTIONS(3221),
    [anon_sym___SUB__] = ACTIONS(3221),
    [anon_sym_break] = ACTIONS(3221),
    [anon_sym_fork] = ACTIONS(3221),
    [anon_sym_getppid] = ACTIONS(3221),
    [anon_sym_time] = ACTIONS(3221),
    [anon_sym_times] = ACTIONS(3221),
    [anon_sym_wait] = ACTIONS(3221),
    [anon_sym_wantarray] = ACTIONS(3221),
    [anon_sym_abs] = ACTIONS(3221),
    [anon_sym_alarm] = ACTIONS(3221),
    [anon_sym_chop] = ACTIONS(3221),
    [anon_sym_chdir] = ACTIONS(3221),
    [anon_sym_close] = ACTIONS(3221),
    [anon_sym_closedir] = ACTIONS(3221),
    [anon_sym_caller] = ACTIONS(3221),
    [anon_sym_chomp] = ACTIONS(3221),
    [anon_sym_chr] = ACTIONS(3221),
    [anon_sym_cos] = ACTIONS(3221),
    [anon_sym_chroot] = ACTIONS(3221),
    [anon_sym_defined] = ACTIONS(3221),
    [anon_sym_delete] = ACTIONS(3221),
    [anon_sym_dbmclose] = ACTIONS(3221),
    [anon_sym_exists] = ACTIONS(3221),
    [anon_sym_exit] = ACTIONS(3221),
    [anon_sym_eof] = ACTIONS(3221),
    [anon_sym_exp] = ACTIONS(3221),
    [anon_sym_each] = ACTIONS(3221),
    [anon_sym_fc] = ACTIONS(3221),
    [anon_sym_fileno] = ACTIONS(3221),
    [anon_sym_gmtime] = ACTIONS(3221),
    [anon_sym_getc] = ACTIONS(3221),
    [anon_sym_getpgrp] = ACTIONS(3221),
    [anon_sym_getprotobyname] = ACTIONS(3221),
    [anon_sym_getpwname] = ACTIONS(3221),
    [anon_sym_getpwuid] = ACTIONS(3221),
    [anon_sym_getpeername] = ACTIONS(3221),
    [anon_sym_getnetbyname] = ACTIONS(3221),
    [anon_sym_getsockname] = ACTIONS(3221),
    [anon_sym_getgrnam] = ACTIONS(3221),
    [anon_sym_getgrgid] = ACTIONS(3221),
    [anon_sym_hex] = ACTIONS(3221),
    [anon_sym_int] = ACTIONS(3221),
    [anon_sym_keys] = ACTIONS(3221),
    [anon_sym_lc] = ACTIONS(3221),
    [anon_sym_lcfirst] = ACTIONS(3221),
    [anon_sym_length] = ACTIONS(3221),
    [anon_sym_localtime] = ACTIONS(3221),
    [anon_sym_log] = ACTIONS(3221),
    [anon_sym_lock] = ACTIONS(3221),
    [anon_sym_lstat] = ACTIONS(3221),
    [anon_sym_oct] = ACTIONS(3221),
    [anon_sym_ord] = ACTIONS(3221),
    [anon_sym_prototype] = ACTIONS(3221),
    [anon_sym_pop] = ACTIONS(3221),
    [anon_sym_pos] = ACTIONS(3221),
    [anon_sym_quotemeta] = ACTIONS(3221),
    [anon_sym_reset] = ACTIONS(3221),
    [anon_sym_rand] = ACTIONS(3221),
    [anon_sym_rmdir] = ACTIONS(3221),
    [anon_sym_readdir] = ACTIONS(3221),
    [anon_sym_readline] = ACTIONS(3221),
    [anon_sym_readpipe] = ACTIONS(3221),
    [anon_sym_rewinddir] = ACTIONS(3221),
    [anon_sym_readlink] = ACTIONS(3221),
    [anon_sym_ref] = ACTIONS(3221),
    [anon_sym_scalar] = ACTIONS(3221),
    [anon_sym_shift] = ACTIONS(3221),
    [anon_sym_sin] = ACTIONS(3221),
    [anon_sym_sleep] = ACTIONS(3221),
    [anon_sym_sqrt] = ACTIONS(3221),
    [anon_sym_srand] = ACTIONS(3221),
    [anon_sym_stat] = ACTIONS(3221),
    [anon_sym_study] = ACTIONS(3221),
    [anon_sym_tell] = ACTIONS(3221),
    [anon_sym_telldir] = ACTIONS(3221),
    [anon_sym_tied] = ACTIONS(3221),
    [anon_sym_uc] = ACTIONS(3221),
    [anon_sym_ucfirst] = ACTIONS(3221),
    [anon_sym_untie] = ACTIONS(3221),
    [anon_sym_umask] = ACTIONS(3221),
    [anon_sym_values] = ACTIONS(3221),
    [anon_sym_write] = ACTIONS(3221),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3221),
    [anon_sym___END__] = ACTIONS(3221),
    [anon_sym_EOT] = ACTIONS(3221),
    [anon_sym_q] = ACTIONS(3221),
    [anon_sym_qq] = ACTIONS(3221),
    [anon_sym_qw] = ACTIONS(3221),
    [anon_sym_qx] = ACTIONS(3221),
    [anon_sym_qr] = ACTIONS(3221),
    [anon_sym_m] = ACTIONS(3221),
    [anon_sym_s] = ACTIONS(3221),
    [anon_sym_tr] = ACTIONS(3221),
    [anon_sym_y] = ACTIONS(3221),
    [anon_sym_if] = ACTIONS(3221),
    [anon_sym_unless] = ACTIONS(3221),
    [anon_sym_while] = ACTIONS(3221),
    [anon_sym_until] = ACTIONS(3221),
    [aux_sym__bareword_token1] = ACTIONS(3221),
    [sym_number] = ACTIONS(3221),
    [anon_sym_true] = ACTIONS(3221),
    [anon_sym_false] = ACTIONS(3221),
    [sym__single_quote] = ACTIONS(3219),
    [sym__double_quote] = ACTIONS(3219),
    [sym__backtick_quote] = ACTIONS(3219),
    [sym__search_slash_quote] = ACTIONS(3219),
    [sym__open_readline_bracket] = ACTIONS(3219),
    [sym__open_fileglob_bracket] = ACTIONS(3219),
    [sym__PERLY_HEREDOC] = ACTIONS(3219),
    [sym__ctrl_z_hack] = ACTIONS(3219),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3219),
    [sym__filetest] = ACTIONS(3219),
  },
  [STATE(1517)] = {
    [sym_heredoc_content] = STATE(1517),
    [ts_builtin_sym_end] = ACTIONS(3223),
    [sym__identifier] = ACTIONS(3225),
    [anon_sym_LBRACE] = ACTIONS(3225),
    [anon_sym_SEMI] = ACTIONS(3225),
    [anon_sym_package] = ACTIONS(3225),
    [anon_sym_class] = ACTIONS(3225),
    [anon_sym_role] = ACTIONS(3225),
    [anon_sym_BUILD] = ACTIONS(3225),
    [anon_sym_ADJUST] = ACTIONS(3225),
    [anon_sym_DOLLAR] = ACTIONS(3225),
    [anon_sym_AT] = ACTIONS(3225),
    [anon_sym_my] = ACTIONS(3225),
    [anon_sym_extended] = ACTIONS(3225),
    [anon_sym_async] = ACTIONS(3225),
    [anon_sym_sub] = ACTIONS(3225),
    [anon_sym_method] = ACTIONS(3225),
    [anon_sym_LPAREN] = ACTIONS(3225),
    [anon_sym_continue] = ACTIONS(3225),
    [anon_sym_state] = ACTIONS(3225),
    [anon_sym_our] = ACTIONS(3225),
    [anon_sym_try] = ACTIONS(3225),
    [anon_sym_defer] = ACTIONS(3225),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3225),
    [anon_sym_LBRACK] = ACTIONS(3225),
    [anon_sym_LT_LT] = ACTIONS(3225),
    [anon_sym_SLASH_SLASH] = ACTIONS(3225),
    [anon_sym_PLUS] = ACTIONS(3225),
    [anon_sym_DASH] = ACTIONS(3225),
    [anon_sym_TILDE] = ACTIONS(3225),
    [anon_sym_BANG] = ACTIONS(3225),
    [anon_sym_PLUS_PLUS] = ACTIONS(3225),
    [anon_sym_DASH_DASH] = ACTIONS(3225),
    [anon_sym_BSLASH] = ACTIONS(3223),
    [anon_sym_do] = ACTIONS(3225),
    [anon_sym_eval] = ACTIONS(3225),
    [anon_sym_await] = ACTIONS(3225),
    [anon_sym_field] = ACTIONS(3225),
    [anon_sym_local] = ACTIONS(3225),
    [anon_sym_dynamically] = ACTIONS(3225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3225),
    [anon_sym_require] = ACTIONS(3225),
    [anon_sym_map] = ACTIONS(3225),
    [anon_sym_grep] = ACTIONS(3225),
    [anon_sym_sort] = ACTIONS(3225),
    [anon_sym_goto] = ACTIONS(3225),
    [anon_sym_return] = ACTIONS(3225),
    [anon_sym_undef] = ACTIONS(3225),
    [anon_sym_PERCENT2] = ACTIONS(3225),
    [anon_sym_AMP2] = ACTIONS(3225),
    [anon_sym_STAR2] = ACTIONS(3225),
    [anon_sym_use] = ACTIONS(3225),
    [anon_sym_no] = ACTIONS(3225),
    [anon_sym_for] = ACTIONS(3225),
    [anon_sym_foreach] = ACTIONS(3225),
    [anon_sym_last] = ACTIONS(3225),
    [anon_sym_next] = ACTIONS(3225),
    [anon_sym_redo] = ACTIONS(3225),
    [anon_sym_BEGIN] = ACTIONS(3225),
    [anon_sym_INIT] = ACTIONS(3225),
    [anon_sym_CHECK] = ACTIONS(3225),
    [anon_sym_UNITCHECK] = ACTIONS(3225),
    [anon_sym_END] = ACTIONS(3225),
    [anon_sym___FILE__] = ACTIONS(3225),
    [anon_sym___LINE__] = ACTIONS(3225),
    [anon_sym___PACKAGE__] = ACTIONS(3225),
    [anon_sym___SUB__] = ACTIONS(3225),
    [anon_sym_break] = ACTIONS(3225),
    [anon_sym_fork] = ACTIONS(3225),
    [anon_sym_getppid] = ACTIONS(3225),
    [anon_sym_time] = ACTIONS(3225),
    [anon_sym_times] = ACTIONS(3225),
    [anon_sym_wait] = ACTIONS(3225),
    [anon_sym_wantarray] = ACTIONS(3225),
    [anon_sym_abs] = ACTIONS(3225),
    [anon_sym_alarm] = ACTIONS(3225),
    [anon_sym_chop] = ACTIONS(3225),
    [anon_sym_chdir] = ACTIONS(3225),
    [anon_sym_close] = ACTIONS(3225),
    [anon_sym_closedir] = ACTIONS(3225),
    [anon_sym_caller] = ACTIONS(3225),
    [anon_sym_chomp] = ACTIONS(3225),
    [anon_sym_chr] = ACTIONS(3225),
    [anon_sym_cos] = ACTIONS(3225),
    [anon_sym_chroot] = ACTIONS(3225),
    [anon_sym_defined] = ACTIONS(3225),
    [anon_sym_delete] = ACTIONS(3225),
    [anon_sym_dbmclose] = ACTIONS(3225),
    [anon_sym_exists] = ACTIONS(3225),
    [anon_sym_exit] = ACTIONS(3225),
    [anon_sym_eof] = ACTIONS(3225),
    [anon_sym_exp] = ACTIONS(3225),
    [anon_sym_each] = ACTIONS(3225),
    [anon_sym_fc] = ACTIONS(3225),
    [anon_sym_fileno] = ACTIONS(3225),
    [anon_sym_gmtime] = ACTIONS(3225),
    [anon_sym_getc] = ACTIONS(3225),
    [anon_sym_getpgrp] = ACTIONS(3225),
    [anon_sym_getprotobyname] = ACTIONS(3225),
    [anon_sym_getpwname] = ACTIONS(3225),
    [anon_sym_getpwuid] = ACTIONS(3225),
    [anon_sym_getpeername] = ACTIONS(3225),
    [anon_sym_getnetbyname] = ACTIONS(3225),
    [anon_sym_getsockname] = ACTIONS(3225),
    [anon_sym_getgrnam] = ACTIONS(3225),
    [anon_sym_getgrgid] = ACTIONS(3225),
    [anon_sym_hex] = ACTIONS(3225),
    [anon_sym_int] = ACTIONS(3225),
    [anon_sym_keys] = ACTIONS(3225),
    [anon_sym_lc] = ACTIONS(3225),
    [anon_sym_lcfirst] = ACTIONS(3225),
    [anon_sym_length] = ACTIONS(3225),
    [anon_sym_localtime] = ACTIONS(3225),
    [anon_sym_log] = ACTIONS(3225),
    [anon_sym_lock] = ACTIONS(3225),
    [anon_sym_lstat] = ACTIONS(3225),
    [anon_sym_oct] = ACTIONS(3225),
    [anon_sym_ord] = ACTIONS(3225),
    [anon_sym_prototype] = ACTIONS(3225),
    [anon_sym_pop] = ACTIONS(3225),
    [anon_sym_pos] = ACTIONS(3225),
    [anon_sym_quotemeta] = ACTIONS(3225),
    [anon_sym_reset] = ACTIONS(3225),
    [anon_sym_rand] = ACTIONS(3225),
    [anon_sym_rmdir] = ACTIONS(3225),
    [anon_sym_readdir] = ACTIONS(3225),
    [anon_sym_readline] = ACTIONS(3225),
    [anon_sym_readpipe] = ACTIONS(3225),
    [anon_sym_rewinddir] = ACTIONS(3225),
    [anon_sym_readlink] = ACTIONS(3225),
    [anon_sym_ref] = ACTIONS(3225),
    [anon_sym_scalar] = ACTIONS(3225),
    [anon_sym_shift] = ACTIONS(3225),
    [anon_sym_sin] = ACTIONS(3225),
    [anon_sym_sleep] = ACTIONS(3225),
    [anon_sym_sqrt] = ACTIONS(3225),
    [anon_sym_srand] = ACTIONS(3225),
    [anon_sym_stat] = ACTIONS(3225),
    [anon_sym_study] = ACTIONS(3225),
    [anon_sym_tell] = ACTIONS(3225),
    [anon_sym_telldir] = ACTIONS(3225),
    [anon_sym_tied] = ACTIONS(3225),
    [anon_sym_uc] = ACTIONS(3225),
    [anon_sym_ucfirst] = ACTIONS(3225),
    [anon_sym_untie] = ACTIONS(3225),
    [anon_sym_umask] = ACTIONS(3225),
    [anon_sym_values] = ACTIONS(3225),
    [anon_sym_write] = ACTIONS(3225),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3225),
    [anon_sym___END__] = ACTIONS(3225),
    [anon_sym_EOT] = ACTIONS(3225),
    [anon_sym_q] = ACTIONS(3225),
    [anon_sym_qq] = ACTIONS(3225),
    [anon_sym_qw] = ACTIONS(3225),
    [anon_sym_qx] = ACTIONS(3225),
    [anon_sym_qr] = ACTIONS(3225),
    [anon_sym_m] = ACTIONS(3225),
    [anon_sym_s] = ACTIONS(3225),
    [anon_sym_tr] = ACTIONS(3225),
    [anon_sym_y] = ACTIONS(3225),
    [anon_sym_if] = ACTIONS(3225),
    [anon_sym_unless] = ACTIONS(3225),
    [anon_sym_while] = ACTIONS(3225),
    [anon_sym_until] = ACTIONS(3225),
    [aux_sym__bareword_token1] = ACTIONS(3225),
    [sym_number] = ACTIONS(3225),
    [anon_sym_true] = ACTIONS(3225),
    [anon_sym_false] = ACTIONS(3225),
    [sym__single_quote] = ACTIONS(3223),
    [sym__double_quote] = ACTIONS(3223),
    [sym__backtick_quote] = ACTIONS(3223),
    [sym__search_slash_quote] = ACTIONS(3223),
    [sym__open_readline_bracket] = ACTIONS(3223),
    [sym__open_fileglob_bracket] = ACTIONS(3223),
    [sym__PERLY_HEREDOC] = ACTIONS(3223),
    [sym__ctrl_z_hack] = ACTIONS(3223),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3223),
    [sym__filetest] = ACTIONS(3223),
  },
  [STATE(1518)] = {
    [sym_heredoc_content] = STATE(1518),
    [ts_builtin_sym_end] = ACTIONS(3227),
    [sym__identifier] = ACTIONS(3229),
    [anon_sym_LBRACE] = ACTIONS(3229),
    [anon_sym_SEMI] = ACTIONS(3229),
    [anon_sym_package] = ACTIONS(3229),
    [anon_sym_class] = ACTIONS(3229),
    [anon_sym_role] = ACTIONS(3229),
    [anon_sym_BUILD] = ACTIONS(3229),
    [anon_sym_ADJUST] = ACTIONS(3229),
    [anon_sym_DOLLAR] = ACTIONS(3229),
    [anon_sym_AT] = ACTIONS(3229),
    [anon_sym_my] = ACTIONS(3229),
    [anon_sym_extended] = ACTIONS(3229),
    [anon_sym_async] = ACTIONS(3229),
    [anon_sym_sub] = ACTIONS(3229),
    [anon_sym_method] = ACTIONS(3229),
    [anon_sym_LPAREN] = ACTIONS(3229),
    [anon_sym_continue] = ACTIONS(3229),
    [anon_sym_state] = ACTIONS(3229),
    [anon_sym_our] = ACTIONS(3229),
    [anon_sym_try] = ACTIONS(3229),
    [anon_sym_defer] = ACTIONS(3229),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3229),
    [anon_sym_LBRACK] = ACTIONS(3229),
    [anon_sym_LT_LT] = ACTIONS(3229),
    [anon_sym_SLASH_SLASH] = ACTIONS(3229),
    [anon_sym_PLUS] = ACTIONS(3229),
    [anon_sym_DASH] = ACTIONS(3229),
    [anon_sym_TILDE] = ACTIONS(3229),
    [anon_sym_BANG] = ACTIONS(3229),
    [anon_sym_PLUS_PLUS] = ACTIONS(3229),
    [anon_sym_DASH_DASH] = ACTIONS(3229),
    [anon_sym_BSLASH] = ACTIONS(3227),
    [anon_sym_do] = ACTIONS(3229),
    [anon_sym_eval] = ACTIONS(3229),
    [anon_sym_await] = ACTIONS(3229),
    [anon_sym_field] = ACTIONS(3229),
    [anon_sym_local] = ACTIONS(3229),
    [anon_sym_dynamically] = ACTIONS(3229),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3229),
    [anon_sym_require] = ACTIONS(3229),
    [anon_sym_map] = ACTIONS(3229),
    [anon_sym_grep] = ACTIONS(3229),
    [anon_sym_sort] = ACTIONS(3229),
    [anon_sym_goto] = ACTIONS(3229),
    [anon_sym_return] = ACTIONS(3229),
    [anon_sym_undef] = ACTIONS(3229),
    [anon_sym_PERCENT2] = ACTIONS(3229),
    [anon_sym_AMP2] = ACTIONS(3229),
    [anon_sym_STAR2] = ACTIONS(3229),
    [anon_sym_use] = ACTIONS(3229),
    [anon_sym_no] = ACTIONS(3229),
    [anon_sym_for] = ACTIONS(3229),
    [anon_sym_foreach] = ACTIONS(3229),
    [anon_sym_last] = ACTIONS(3229),
    [anon_sym_next] = ACTIONS(3229),
    [anon_sym_redo] = ACTIONS(3229),
    [anon_sym_BEGIN] = ACTIONS(3229),
    [anon_sym_INIT] = ACTIONS(3229),
    [anon_sym_CHECK] = ACTIONS(3229),
    [anon_sym_UNITCHECK] = ACTIONS(3229),
    [anon_sym_END] = ACTIONS(3229),
    [anon_sym___FILE__] = ACTIONS(3229),
    [anon_sym___LINE__] = ACTIONS(3229),
    [anon_sym___PACKAGE__] = ACTIONS(3229),
    [anon_sym___SUB__] = ACTIONS(3229),
    [anon_sym_break] = ACTIONS(3229),
    [anon_sym_fork] = ACTIONS(3229),
    [anon_sym_getppid] = ACTIONS(3229),
    [anon_sym_time] = ACTIONS(3229),
    [anon_sym_times] = ACTIONS(3229),
    [anon_sym_wait] = ACTIONS(3229),
    [anon_sym_wantarray] = ACTIONS(3229),
    [anon_sym_abs] = ACTIONS(3229),
    [anon_sym_alarm] = ACTIONS(3229),
    [anon_sym_chop] = ACTIONS(3229),
    [anon_sym_chdir] = ACTIONS(3229),
    [anon_sym_close] = ACTIONS(3229),
    [anon_sym_closedir] = ACTIONS(3229),
    [anon_sym_caller] = ACTIONS(3229),
    [anon_sym_chomp] = ACTIONS(3229),
    [anon_sym_chr] = ACTIONS(3229),
    [anon_sym_cos] = ACTIONS(3229),
    [anon_sym_chroot] = ACTIONS(3229),
    [anon_sym_defined] = ACTIONS(3229),
    [anon_sym_delete] = ACTIONS(3229),
    [anon_sym_dbmclose] = ACTIONS(3229),
    [anon_sym_exists] = ACTIONS(3229),
    [anon_sym_exit] = ACTIONS(3229),
    [anon_sym_eof] = ACTIONS(3229),
    [anon_sym_exp] = ACTIONS(3229),
    [anon_sym_each] = ACTIONS(3229),
    [anon_sym_fc] = ACTIONS(3229),
    [anon_sym_fileno] = ACTIONS(3229),
    [anon_sym_gmtime] = ACTIONS(3229),
    [anon_sym_getc] = ACTIONS(3229),
    [anon_sym_getpgrp] = ACTIONS(3229),
    [anon_sym_getprotobyname] = ACTIONS(3229),
    [anon_sym_getpwname] = ACTIONS(3229),
    [anon_sym_getpwuid] = ACTIONS(3229),
    [anon_sym_getpeername] = ACTIONS(3229),
    [anon_sym_getnetbyname] = ACTIONS(3229),
    [anon_sym_getsockname] = ACTIONS(3229),
    [anon_sym_getgrnam] = ACTIONS(3229),
    [anon_sym_getgrgid] = ACTIONS(3229),
    [anon_sym_hex] = ACTIONS(3229),
    [anon_sym_int] = ACTIONS(3229),
    [anon_sym_keys] = ACTIONS(3229),
    [anon_sym_lc] = ACTIONS(3229),
    [anon_sym_lcfirst] = ACTIONS(3229),
    [anon_sym_length] = ACTIONS(3229),
    [anon_sym_localtime] = ACTIONS(3229),
    [anon_sym_log] = ACTIONS(3229),
    [anon_sym_lock] = ACTIONS(3229),
    [anon_sym_lstat] = ACTIONS(3229),
    [anon_sym_oct] = ACTIONS(3229),
    [anon_sym_ord] = ACTIONS(3229),
    [anon_sym_prototype] = ACTIONS(3229),
    [anon_sym_pop] = ACTIONS(3229),
    [anon_sym_pos] = ACTIONS(3229),
    [anon_sym_quotemeta] = ACTIONS(3229),
    [anon_sym_reset] = ACTIONS(3229),
    [anon_sym_rand] = ACTIONS(3229),
    [anon_sym_rmdir] = ACTIONS(3229),
    [anon_sym_readdir] = ACTIONS(3229),
    [anon_sym_readline] = ACTIONS(3229),
    [anon_sym_readpipe] = ACTIONS(3229),
    [anon_sym_rewinddir] = ACTIONS(3229),
    [anon_sym_readlink] = ACTIONS(3229),
    [anon_sym_ref] = ACTIONS(3229),
    [anon_sym_scalar] = ACTIONS(3229),
    [anon_sym_shift] = ACTIONS(3229),
    [anon_sym_sin] = ACTIONS(3229),
    [anon_sym_sleep] = ACTIONS(3229),
    [anon_sym_sqrt] = ACTIONS(3229),
    [anon_sym_srand] = ACTIONS(3229),
    [anon_sym_stat] = ACTIONS(3229),
    [anon_sym_study] = ACTIONS(3229),
    [anon_sym_tell] = ACTIONS(3229),
    [anon_sym_telldir] = ACTIONS(3229),
    [anon_sym_tied] = ACTIONS(3229),
    [anon_sym_uc] = ACTIONS(3229),
    [anon_sym_ucfirst] = ACTIONS(3229),
    [anon_sym_untie] = ACTIONS(3229),
    [anon_sym_umask] = ACTIONS(3229),
    [anon_sym_values] = ACTIONS(3229),
    [anon_sym_write] = ACTIONS(3229),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3229),
    [anon_sym___END__] = ACTIONS(3229),
    [anon_sym_EOT] = ACTIONS(3229),
    [anon_sym_q] = ACTIONS(3229),
    [anon_sym_qq] = ACTIONS(3229),
    [anon_sym_qw] = ACTIONS(3229),
    [anon_sym_qx] = ACTIONS(3229),
    [anon_sym_qr] = ACTIONS(3229),
    [anon_sym_m] = ACTIONS(3229),
    [anon_sym_s] = ACTIONS(3229),
    [anon_sym_tr] = ACTIONS(3229),
    [anon_sym_y] = ACTIONS(3229),
    [anon_sym_if] = ACTIONS(3229),
    [anon_sym_unless] = ACTIONS(3229),
    [anon_sym_while] = ACTIONS(3229),
    [anon_sym_until] = ACTIONS(3229),
    [aux_sym__bareword_token1] = ACTIONS(3229),
    [sym_number] = ACTIONS(3229),
    [anon_sym_true] = ACTIONS(3229),
    [anon_sym_false] = ACTIONS(3229),
    [sym__single_quote] = ACTIONS(3227),
    [sym__double_quote] = ACTIONS(3227),
    [sym__backtick_quote] = ACTIONS(3227),
    [sym__search_slash_quote] = ACTIONS(3227),
    [sym__open_readline_bracket] = ACTIONS(3227),
    [sym__open_fileglob_bracket] = ACTIONS(3227),
    [sym__PERLY_HEREDOC] = ACTIONS(3227),
    [sym__ctrl_z_hack] = ACTIONS(3227),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3227),
    [sym__filetest] = ACTIONS(3227),
  },
  [STATE(1519)] = {
    [sym_heredoc_content] = STATE(1519),
    [ts_builtin_sym_end] = ACTIONS(2530),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2532),
    [anon_sym___END__] = ACTIONS(2532),
    [anon_sym_EOT] = ACTIONS(2532),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1520)] = {
    [sym_heredoc_content] = STATE(1520),
    [ts_builtin_sym_end] = ACTIONS(2690),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(2688),
    [anon_sym___END__] = ACTIONS(2688),
    [anon_sym_EOT] = ACTIONS(2688),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1521)] = {
    [sym_heredoc_content] = STATE(1521),
    [ts_builtin_sym_end] = ACTIONS(3231),
    [sym__identifier] = ACTIONS(3233),
    [anon_sym_LBRACE] = ACTIONS(3233),
    [anon_sym_SEMI] = ACTIONS(3233),
    [anon_sym_package] = ACTIONS(3233),
    [anon_sym_class] = ACTIONS(3233),
    [anon_sym_role] = ACTIONS(3233),
    [anon_sym_BUILD] = ACTIONS(3233),
    [anon_sym_ADJUST] = ACTIONS(3233),
    [anon_sym_DOLLAR] = ACTIONS(3233),
    [anon_sym_AT] = ACTIONS(3233),
    [anon_sym_my] = ACTIONS(3233),
    [anon_sym_extended] = ACTIONS(3233),
    [anon_sym_async] = ACTIONS(3233),
    [anon_sym_sub] = ACTIONS(3233),
    [anon_sym_method] = ACTIONS(3233),
    [anon_sym_LPAREN] = ACTIONS(3233),
    [anon_sym_continue] = ACTIONS(3233),
    [anon_sym_state] = ACTIONS(3233),
    [anon_sym_our] = ACTIONS(3233),
    [anon_sym_try] = ACTIONS(3233),
    [anon_sym_defer] = ACTIONS(3233),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3233),
    [anon_sym_LBRACK] = ACTIONS(3233),
    [anon_sym_LT_LT] = ACTIONS(3233),
    [anon_sym_SLASH_SLASH] = ACTIONS(3233),
    [anon_sym_PLUS] = ACTIONS(3233),
    [anon_sym_DASH] = ACTIONS(3233),
    [anon_sym_TILDE] = ACTIONS(3233),
    [anon_sym_BANG] = ACTIONS(3233),
    [anon_sym_PLUS_PLUS] = ACTIONS(3233),
    [anon_sym_DASH_DASH] = ACTIONS(3233),
    [anon_sym_BSLASH] = ACTIONS(3231),
    [anon_sym_do] = ACTIONS(3233),
    [anon_sym_eval] = ACTIONS(3233),
    [anon_sym_await] = ACTIONS(3233),
    [anon_sym_field] = ACTIONS(3233),
    [anon_sym_local] = ACTIONS(3233),
    [anon_sym_dynamically] = ACTIONS(3233),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3233),
    [anon_sym_require] = ACTIONS(3233),
    [anon_sym_map] = ACTIONS(3233),
    [anon_sym_grep] = ACTIONS(3233),
    [anon_sym_sort] = ACTIONS(3233),
    [anon_sym_goto] = ACTIONS(3233),
    [anon_sym_return] = ACTIONS(3233),
    [anon_sym_undef] = ACTIONS(3233),
    [anon_sym_PERCENT2] = ACTIONS(3233),
    [anon_sym_AMP2] = ACTIONS(3233),
    [anon_sym_STAR2] = ACTIONS(3233),
    [anon_sym_use] = ACTIONS(3233),
    [anon_sym_no] = ACTIONS(3233),
    [anon_sym_for] = ACTIONS(3233),
    [anon_sym_foreach] = ACTIONS(3233),
    [anon_sym_last] = ACTIONS(3233),
    [anon_sym_next] = ACTIONS(3233),
    [anon_sym_redo] = ACTIONS(3233),
    [anon_sym_BEGIN] = ACTIONS(3233),
    [anon_sym_INIT] = ACTIONS(3233),
    [anon_sym_CHECK] = ACTIONS(3233),
    [anon_sym_UNITCHECK] = ACTIONS(3233),
    [anon_sym_END] = ACTIONS(3233),
    [anon_sym___FILE__] = ACTIONS(3233),
    [anon_sym___LINE__] = ACTIONS(3233),
    [anon_sym___PACKAGE__] = ACTIONS(3233),
    [anon_sym___SUB__] = ACTIONS(3233),
    [anon_sym_break] = ACTIONS(3233),
    [anon_sym_fork] = ACTIONS(3233),
    [anon_sym_getppid] = ACTIONS(3233),
    [anon_sym_time] = ACTIONS(3233),
    [anon_sym_times] = ACTIONS(3233),
    [anon_sym_wait] = ACTIONS(3233),
    [anon_sym_wantarray] = ACTIONS(3233),
    [anon_sym_abs] = ACTIONS(3233),
    [anon_sym_alarm] = ACTIONS(3233),
    [anon_sym_chop] = ACTIONS(3233),
    [anon_sym_chdir] = ACTIONS(3233),
    [anon_sym_close] = ACTIONS(3233),
    [anon_sym_closedir] = ACTIONS(3233),
    [anon_sym_caller] = ACTIONS(3233),
    [anon_sym_chomp] = ACTIONS(3233),
    [anon_sym_chr] = ACTIONS(3233),
    [anon_sym_cos] = ACTIONS(3233),
    [anon_sym_chroot] = ACTIONS(3233),
    [anon_sym_defined] = ACTIONS(3233),
    [anon_sym_delete] = ACTIONS(3233),
    [anon_sym_dbmclose] = ACTIONS(3233),
    [anon_sym_exists] = ACTIONS(3233),
    [anon_sym_exit] = ACTIONS(3233),
    [anon_sym_eof] = ACTIONS(3233),
    [anon_sym_exp] = ACTIONS(3233),
    [anon_sym_each] = ACTIONS(3233),
    [anon_sym_fc] = ACTIONS(3233),
    [anon_sym_fileno] = ACTIONS(3233),
    [anon_sym_gmtime] = ACTIONS(3233),
    [anon_sym_getc] = ACTIONS(3233),
    [anon_sym_getpgrp] = ACTIONS(3233),
    [anon_sym_getprotobyname] = ACTIONS(3233),
    [anon_sym_getpwname] = ACTIONS(3233),
    [anon_sym_getpwuid] = ACTIONS(3233),
    [anon_sym_getpeername] = ACTIONS(3233),
    [anon_sym_getnetbyname] = ACTIONS(3233),
    [anon_sym_getsockname] = ACTIONS(3233),
    [anon_sym_getgrnam] = ACTIONS(3233),
    [anon_sym_getgrgid] = ACTIONS(3233),
    [anon_sym_hex] = ACTIONS(3233),
    [anon_sym_int] = ACTIONS(3233),
    [anon_sym_keys] = ACTIONS(3233),
    [anon_sym_lc] = ACTIONS(3233),
    [anon_sym_lcfirst] = ACTIONS(3233),
    [anon_sym_length] = ACTIONS(3233),
    [anon_sym_localtime] = ACTIONS(3233),
    [anon_sym_log] = ACTIONS(3233),
    [anon_sym_lock] = ACTIONS(3233),
    [anon_sym_lstat] = ACTIONS(3233),
    [anon_sym_oct] = ACTIONS(3233),
    [anon_sym_ord] = ACTIONS(3233),
    [anon_sym_prototype] = ACTIONS(3233),
    [anon_sym_pop] = ACTIONS(3233),
    [anon_sym_pos] = ACTIONS(3233),
    [anon_sym_quotemeta] = ACTIONS(3233),
    [anon_sym_reset] = ACTIONS(3233),
    [anon_sym_rand] = ACTIONS(3233),
    [anon_sym_rmdir] = ACTIONS(3233),
    [anon_sym_readdir] = ACTIONS(3233),
    [anon_sym_readline] = ACTIONS(3233),
    [anon_sym_readpipe] = ACTIONS(3233),
    [anon_sym_rewinddir] = ACTIONS(3233),
    [anon_sym_readlink] = ACTIONS(3233),
    [anon_sym_ref] = ACTIONS(3233),
    [anon_sym_scalar] = ACTIONS(3233),
    [anon_sym_shift] = ACTIONS(3233),
    [anon_sym_sin] = ACTIONS(3233),
    [anon_sym_sleep] = ACTIONS(3233),
    [anon_sym_sqrt] = ACTIONS(3233),
    [anon_sym_srand] = ACTIONS(3233),
    [anon_sym_stat] = ACTIONS(3233),
    [anon_sym_study] = ACTIONS(3233),
    [anon_sym_tell] = ACTIONS(3233),
    [anon_sym_telldir] = ACTIONS(3233),
    [anon_sym_tied] = ACTIONS(3233),
    [anon_sym_uc] = ACTIONS(3233),
    [anon_sym_ucfirst] = ACTIONS(3233),
    [anon_sym_untie] = ACTIONS(3233),
    [anon_sym_umask] = ACTIONS(3233),
    [anon_sym_values] = ACTIONS(3233),
    [anon_sym_write] = ACTIONS(3233),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3233),
    [anon_sym___END__] = ACTIONS(3233),
    [anon_sym_EOT] = ACTIONS(3233),
    [anon_sym_q] = ACTIONS(3233),
    [anon_sym_qq] = ACTIONS(3233),
    [anon_sym_qw] = ACTIONS(3233),
    [anon_sym_qx] = ACTIONS(3233),
    [anon_sym_qr] = ACTIONS(3233),
    [anon_sym_m] = ACTIONS(3233),
    [anon_sym_s] = ACTIONS(3233),
    [anon_sym_tr] = ACTIONS(3233),
    [anon_sym_y] = ACTIONS(3233),
    [anon_sym_if] = ACTIONS(3233),
    [anon_sym_unless] = ACTIONS(3233),
    [anon_sym_while] = ACTIONS(3233),
    [anon_sym_until] = ACTIONS(3233),
    [aux_sym__bareword_token1] = ACTIONS(3233),
    [sym_number] = ACTIONS(3233),
    [anon_sym_true] = ACTIONS(3233),
    [anon_sym_false] = ACTIONS(3233),
    [sym__single_quote] = ACTIONS(3231),
    [sym__double_quote] = ACTIONS(3231),
    [sym__backtick_quote] = ACTIONS(3231),
    [sym__search_slash_quote] = ACTIONS(3231),
    [sym__open_readline_bracket] = ACTIONS(3231),
    [sym__open_fileglob_bracket] = ACTIONS(3231),
    [sym__PERLY_HEREDOC] = ACTIONS(3231),
    [sym__ctrl_z_hack] = ACTIONS(3231),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3231),
    [sym__filetest] = ACTIONS(3231),
  },
  [STATE(1522)] = {
    [sym_heredoc_content] = STATE(1522),
    [ts_builtin_sym_end] = ACTIONS(3235),
    [sym__identifier] = ACTIONS(3237),
    [anon_sym_LBRACE] = ACTIONS(3237),
    [anon_sym_SEMI] = ACTIONS(3237),
    [anon_sym_package] = ACTIONS(3237),
    [anon_sym_class] = ACTIONS(3237),
    [anon_sym_role] = ACTIONS(3237),
    [anon_sym_BUILD] = ACTIONS(3237),
    [anon_sym_ADJUST] = ACTIONS(3237),
    [anon_sym_DOLLAR] = ACTIONS(3237),
    [anon_sym_AT] = ACTIONS(3237),
    [anon_sym_my] = ACTIONS(3237),
    [anon_sym_extended] = ACTIONS(3237),
    [anon_sym_async] = ACTIONS(3237),
    [anon_sym_sub] = ACTIONS(3237),
    [anon_sym_method] = ACTIONS(3237),
    [anon_sym_LPAREN] = ACTIONS(3237),
    [anon_sym_continue] = ACTIONS(3237),
    [anon_sym_state] = ACTIONS(3237),
    [anon_sym_our] = ACTIONS(3237),
    [anon_sym_try] = ACTIONS(3237),
    [anon_sym_defer] = ACTIONS(3237),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3237),
    [anon_sym_LBRACK] = ACTIONS(3237),
    [anon_sym_LT_LT] = ACTIONS(3237),
    [anon_sym_SLASH_SLASH] = ACTIONS(3237),
    [anon_sym_PLUS] = ACTIONS(3237),
    [anon_sym_DASH] = ACTIONS(3237),
    [anon_sym_TILDE] = ACTIONS(3237),
    [anon_sym_BANG] = ACTIONS(3237),
    [anon_sym_PLUS_PLUS] = ACTIONS(3237),
    [anon_sym_DASH_DASH] = ACTIONS(3237),
    [anon_sym_BSLASH] = ACTIONS(3235),
    [anon_sym_do] = ACTIONS(3237),
    [anon_sym_eval] = ACTIONS(3237),
    [anon_sym_await] = ACTIONS(3237),
    [anon_sym_field] = ACTIONS(3237),
    [anon_sym_local] = ACTIONS(3237),
    [anon_sym_dynamically] = ACTIONS(3237),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3237),
    [anon_sym_require] = ACTIONS(3237),
    [anon_sym_map] = ACTIONS(3237),
    [anon_sym_grep] = ACTIONS(3237),
    [anon_sym_sort] = ACTIONS(3237),
    [anon_sym_goto] = ACTIONS(3237),
    [anon_sym_return] = ACTIONS(3237),
    [anon_sym_undef] = ACTIONS(3237),
    [anon_sym_PERCENT2] = ACTIONS(3237),
    [anon_sym_AMP2] = ACTIONS(3237),
    [anon_sym_STAR2] = ACTIONS(3237),
    [anon_sym_use] = ACTIONS(3237),
    [anon_sym_no] = ACTIONS(3237),
    [anon_sym_for] = ACTIONS(3237),
    [anon_sym_foreach] = ACTIONS(3237),
    [anon_sym_last] = ACTIONS(3237),
    [anon_sym_next] = ACTIONS(3237),
    [anon_sym_redo] = ACTIONS(3237),
    [anon_sym_BEGIN] = ACTIONS(3237),
    [anon_sym_INIT] = ACTIONS(3237),
    [anon_sym_CHECK] = ACTIONS(3237),
    [anon_sym_UNITCHECK] = ACTIONS(3237),
    [anon_sym_END] = ACTIONS(3237),
    [anon_sym___FILE__] = ACTIONS(3237),
    [anon_sym___LINE__] = ACTIONS(3237),
    [anon_sym___PACKAGE__] = ACTIONS(3237),
    [anon_sym___SUB__] = ACTIONS(3237),
    [anon_sym_break] = ACTIONS(3237),
    [anon_sym_fork] = ACTIONS(3237),
    [anon_sym_getppid] = ACTIONS(3237),
    [anon_sym_time] = ACTIONS(3237),
    [anon_sym_times] = ACTIONS(3237),
    [anon_sym_wait] = ACTIONS(3237),
    [anon_sym_wantarray] = ACTIONS(3237),
    [anon_sym_abs] = ACTIONS(3237),
    [anon_sym_alarm] = ACTIONS(3237),
    [anon_sym_chop] = ACTIONS(3237),
    [anon_sym_chdir] = ACTIONS(3237),
    [anon_sym_close] = ACTIONS(3237),
    [anon_sym_closedir] = ACTIONS(3237),
    [anon_sym_caller] = ACTIONS(3237),
    [anon_sym_chomp] = ACTIONS(3237),
    [anon_sym_chr] = ACTIONS(3237),
    [anon_sym_cos] = ACTIONS(3237),
    [anon_sym_chroot] = ACTIONS(3237),
    [anon_sym_defined] = ACTIONS(3237),
    [anon_sym_delete] = ACTIONS(3237),
    [anon_sym_dbmclose] = ACTIONS(3237),
    [anon_sym_exists] = ACTIONS(3237),
    [anon_sym_exit] = ACTIONS(3237),
    [anon_sym_eof] = ACTIONS(3237),
    [anon_sym_exp] = ACTIONS(3237),
    [anon_sym_each] = ACTIONS(3237),
    [anon_sym_fc] = ACTIONS(3237),
    [anon_sym_fileno] = ACTIONS(3237),
    [anon_sym_gmtime] = ACTIONS(3237),
    [anon_sym_getc] = ACTIONS(3237),
    [anon_sym_getpgrp] = ACTIONS(3237),
    [anon_sym_getprotobyname] = ACTIONS(3237),
    [anon_sym_getpwname] = ACTIONS(3237),
    [anon_sym_getpwuid] = ACTIONS(3237),
    [anon_sym_getpeername] = ACTIONS(3237),
    [anon_sym_getnetbyname] = ACTIONS(3237),
    [anon_sym_getsockname] = ACTIONS(3237),
    [anon_sym_getgrnam] = ACTIONS(3237),
    [anon_sym_getgrgid] = ACTIONS(3237),
    [anon_sym_hex] = ACTIONS(3237),
    [anon_sym_int] = ACTIONS(3237),
    [anon_sym_keys] = ACTIONS(3237),
    [anon_sym_lc] = ACTIONS(3237),
    [anon_sym_lcfirst] = ACTIONS(3237),
    [anon_sym_length] = ACTIONS(3237),
    [anon_sym_localtime] = ACTIONS(3237),
    [anon_sym_log] = ACTIONS(3237),
    [anon_sym_lock] = ACTIONS(3237),
    [anon_sym_lstat] = ACTIONS(3237),
    [anon_sym_oct] = ACTIONS(3237),
    [anon_sym_ord] = ACTIONS(3237),
    [anon_sym_prototype] = ACTIONS(3237),
    [anon_sym_pop] = ACTIONS(3237),
    [anon_sym_pos] = ACTIONS(3237),
    [anon_sym_quotemeta] = ACTIONS(3237),
    [anon_sym_reset] = ACTIONS(3237),
    [anon_sym_rand] = ACTIONS(3237),
    [anon_sym_rmdir] = ACTIONS(3237),
    [anon_sym_readdir] = ACTIONS(3237),
    [anon_sym_readline] = ACTIONS(3237),
    [anon_sym_readpipe] = ACTIONS(3237),
    [anon_sym_rewinddir] = ACTIONS(3237),
    [anon_sym_readlink] = ACTIONS(3237),
    [anon_sym_ref] = ACTIONS(3237),
    [anon_sym_scalar] = ACTIONS(3237),
    [anon_sym_shift] = ACTIONS(3237),
    [anon_sym_sin] = ACTIONS(3237),
    [anon_sym_sleep] = ACTIONS(3237),
    [anon_sym_sqrt] = ACTIONS(3237),
    [anon_sym_srand] = ACTIONS(3237),
    [anon_sym_stat] = ACTIONS(3237),
    [anon_sym_study] = ACTIONS(3237),
    [anon_sym_tell] = ACTIONS(3237),
    [anon_sym_telldir] = ACTIONS(3237),
    [anon_sym_tied] = ACTIONS(3237),
    [anon_sym_uc] = ACTIONS(3237),
    [anon_sym_ucfirst] = ACTIONS(3237),
    [anon_sym_untie] = ACTIONS(3237),
    [anon_sym_umask] = ACTIONS(3237),
    [anon_sym_values] = ACTIONS(3237),
    [anon_sym_write] = ACTIONS(3237),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3237),
    [anon_sym___END__] = ACTIONS(3237),
    [anon_sym_EOT] = ACTIONS(3237),
    [anon_sym_q] = ACTIONS(3237),
    [anon_sym_qq] = ACTIONS(3237),
    [anon_sym_qw] = ACTIONS(3237),
    [anon_sym_qx] = ACTIONS(3237),
    [anon_sym_qr] = ACTIONS(3237),
    [anon_sym_m] = ACTIONS(3237),
    [anon_sym_s] = ACTIONS(3237),
    [anon_sym_tr] = ACTIONS(3237),
    [anon_sym_y] = ACTIONS(3237),
    [anon_sym_if] = ACTIONS(3237),
    [anon_sym_unless] = ACTIONS(3237),
    [anon_sym_while] = ACTIONS(3237),
    [anon_sym_until] = ACTIONS(3237),
    [aux_sym__bareword_token1] = ACTIONS(3237),
    [sym_number] = ACTIONS(3237),
    [anon_sym_true] = ACTIONS(3237),
    [anon_sym_false] = ACTIONS(3237),
    [sym__single_quote] = ACTIONS(3235),
    [sym__double_quote] = ACTIONS(3235),
    [sym__backtick_quote] = ACTIONS(3235),
    [sym__search_slash_quote] = ACTIONS(3235),
    [sym__open_readline_bracket] = ACTIONS(3235),
    [sym__open_fileglob_bracket] = ACTIONS(3235),
    [sym__PERLY_HEREDOC] = ACTIONS(3235),
    [sym__ctrl_z_hack] = ACTIONS(3235),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3235),
    [sym__filetest] = ACTIONS(3235),
  },
  [STATE(1523)] = {
    [sym_heredoc_content] = STATE(1523),
    [ts_builtin_sym_end] = ACTIONS(3239),
    [sym__identifier] = ACTIONS(3241),
    [anon_sym_LBRACE] = ACTIONS(3241),
    [anon_sym_SEMI] = ACTIONS(3241),
    [anon_sym_package] = ACTIONS(3241),
    [anon_sym_class] = ACTIONS(3241),
    [anon_sym_role] = ACTIONS(3241),
    [anon_sym_BUILD] = ACTIONS(3241),
    [anon_sym_ADJUST] = ACTIONS(3241),
    [anon_sym_DOLLAR] = ACTIONS(3241),
    [anon_sym_AT] = ACTIONS(3241),
    [anon_sym_my] = ACTIONS(3241),
    [anon_sym_extended] = ACTIONS(3241),
    [anon_sym_async] = ACTIONS(3241),
    [anon_sym_sub] = ACTIONS(3241),
    [anon_sym_method] = ACTIONS(3241),
    [anon_sym_LPAREN] = ACTIONS(3241),
    [anon_sym_continue] = ACTIONS(3241),
    [anon_sym_state] = ACTIONS(3241),
    [anon_sym_our] = ACTIONS(3241),
    [anon_sym_try] = ACTIONS(3241),
    [anon_sym_defer] = ACTIONS(3241),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3241),
    [anon_sym_LBRACK] = ACTIONS(3241),
    [anon_sym_LT_LT] = ACTIONS(3241),
    [anon_sym_SLASH_SLASH] = ACTIONS(3241),
    [anon_sym_PLUS] = ACTIONS(3241),
    [anon_sym_DASH] = ACTIONS(3241),
    [anon_sym_TILDE] = ACTIONS(3241),
    [anon_sym_BANG] = ACTIONS(3241),
    [anon_sym_PLUS_PLUS] = ACTIONS(3241),
    [anon_sym_DASH_DASH] = ACTIONS(3241),
    [anon_sym_BSLASH] = ACTIONS(3239),
    [anon_sym_do] = ACTIONS(3241),
    [anon_sym_eval] = ACTIONS(3241),
    [anon_sym_await] = ACTIONS(3241),
    [anon_sym_field] = ACTIONS(3241),
    [anon_sym_local] = ACTIONS(3241),
    [anon_sym_dynamically] = ACTIONS(3241),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3241),
    [anon_sym_require] = ACTIONS(3241),
    [anon_sym_map] = ACTIONS(3241),
    [anon_sym_grep] = ACTIONS(3241),
    [anon_sym_sort] = ACTIONS(3241),
    [anon_sym_goto] = ACTIONS(3241),
    [anon_sym_return] = ACTIONS(3241),
    [anon_sym_undef] = ACTIONS(3241),
    [anon_sym_PERCENT2] = ACTIONS(3241),
    [anon_sym_AMP2] = ACTIONS(3241),
    [anon_sym_STAR2] = ACTIONS(3241),
    [anon_sym_use] = ACTIONS(3241),
    [anon_sym_no] = ACTIONS(3241),
    [anon_sym_for] = ACTIONS(3241),
    [anon_sym_foreach] = ACTIONS(3241),
    [anon_sym_last] = ACTIONS(3241),
    [anon_sym_next] = ACTIONS(3241),
    [anon_sym_redo] = ACTIONS(3241),
    [anon_sym_BEGIN] = ACTIONS(3241),
    [anon_sym_INIT] = ACTIONS(3241),
    [anon_sym_CHECK] = ACTIONS(3241),
    [anon_sym_UNITCHECK] = ACTIONS(3241),
    [anon_sym_END] = ACTIONS(3241),
    [anon_sym___FILE__] = ACTIONS(3241),
    [anon_sym___LINE__] = ACTIONS(3241),
    [anon_sym___PACKAGE__] = ACTIONS(3241),
    [anon_sym___SUB__] = ACTIONS(3241),
    [anon_sym_break] = ACTIONS(3241),
    [anon_sym_fork] = ACTIONS(3241),
    [anon_sym_getppid] = ACTIONS(3241),
    [anon_sym_time] = ACTIONS(3241),
    [anon_sym_times] = ACTIONS(3241),
    [anon_sym_wait] = ACTIONS(3241),
    [anon_sym_wantarray] = ACTIONS(3241),
    [anon_sym_abs] = ACTIONS(3241),
    [anon_sym_alarm] = ACTIONS(3241),
    [anon_sym_chop] = ACTIONS(3241),
    [anon_sym_chdir] = ACTIONS(3241),
    [anon_sym_close] = ACTIONS(3241),
    [anon_sym_closedir] = ACTIONS(3241),
    [anon_sym_caller] = ACTIONS(3241),
    [anon_sym_chomp] = ACTIONS(3241),
    [anon_sym_chr] = ACTIONS(3241),
    [anon_sym_cos] = ACTIONS(3241),
    [anon_sym_chroot] = ACTIONS(3241),
    [anon_sym_defined] = ACTIONS(3241),
    [anon_sym_delete] = ACTIONS(3241),
    [anon_sym_dbmclose] = ACTIONS(3241),
    [anon_sym_exists] = ACTIONS(3241),
    [anon_sym_exit] = ACTIONS(3241),
    [anon_sym_eof] = ACTIONS(3241),
    [anon_sym_exp] = ACTIONS(3241),
    [anon_sym_each] = ACTIONS(3241),
    [anon_sym_fc] = ACTIONS(3241),
    [anon_sym_fileno] = ACTIONS(3241),
    [anon_sym_gmtime] = ACTIONS(3241),
    [anon_sym_getc] = ACTIONS(3241),
    [anon_sym_getpgrp] = ACTIONS(3241),
    [anon_sym_getprotobyname] = ACTIONS(3241),
    [anon_sym_getpwname] = ACTIONS(3241),
    [anon_sym_getpwuid] = ACTIONS(3241),
    [anon_sym_getpeername] = ACTIONS(3241),
    [anon_sym_getnetbyname] = ACTIONS(3241),
    [anon_sym_getsockname] = ACTIONS(3241),
    [anon_sym_getgrnam] = ACTIONS(3241),
    [anon_sym_getgrgid] = ACTIONS(3241),
    [anon_sym_hex] = ACTIONS(3241),
    [anon_sym_int] = ACTIONS(3241),
    [anon_sym_keys] = ACTIONS(3241),
    [anon_sym_lc] = ACTIONS(3241),
    [anon_sym_lcfirst] = ACTIONS(3241),
    [anon_sym_length] = ACTIONS(3241),
    [anon_sym_localtime] = ACTIONS(3241),
    [anon_sym_log] = ACTIONS(3241),
    [anon_sym_lock] = ACTIONS(3241),
    [anon_sym_lstat] = ACTIONS(3241),
    [anon_sym_oct] = ACTIONS(3241),
    [anon_sym_ord] = ACTIONS(3241),
    [anon_sym_prototype] = ACTIONS(3241),
    [anon_sym_pop] = ACTIONS(3241),
    [anon_sym_pos] = ACTIONS(3241),
    [anon_sym_quotemeta] = ACTIONS(3241),
    [anon_sym_reset] = ACTIONS(3241),
    [anon_sym_rand] = ACTIONS(3241),
    [anon_sym_rmdir] = ACTIONS(3241),
    [anon_sym_readdir] = ACTIONS(3241),
    [anon_sym_readline] = ACTIONS(3241),
    [anon_sym_readpipe] = ACTIONS(3241),
    [anon_sym_rewinddir] = ACTIONS(3241),
    [anon_sym_readlink] = ACTIONS(3241),
    [anon_sym_ref] = ACTIONS(3241),
    [anon_sym_scalar] = ACTIONS(3241),
    [anon_sym_shift] = ACTIONS(3241),
    [anon_sym_sin] = ACTIONS(3241),
    [anon_sym_sleep] = ACTIONS(3241),
    [anon_sym_sqrt] = ACTIONS(3241),
    [anon_sym_srand] = ACTIONS(3241),
    [anon_sym_stat] = ACTIONS(3241),
    [anon_sym_study] = ACTIONS(3241),
    [anon_sym_tell] = ACTIONS(3241),
    [anon_sym_telldir] = ACTIONS(3241),
    [anon_sym_tied] = ACTIONS(3241),
    [anon_sym_uc] = ACTIONS(3241),
    [anon_sym_ucfirst] = ACTIONS(3241),
    [anon_sym_untie] = ACTIONS(3241),
    [anon_sym_umask] = ACTIONS(3241),
    [anon_sym_values] = ACTIONS(3241),
    [anon_sym_write] = ACTIONS(3241),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3241),
    [anon_sym___END__] = ACTIONS(3241),
    [anon_sym_EOT] = ACTIONS(3241),
    [anon_sym_q] = ACTIONS(3241),
    [anon_sym_qq] = ACTIONS(3241),
    [anon_sym_qw] = ACTIONS(3241),
    [anon_sym_qx] = ACTIONS(3241),
    [anon_sym_qr] = ACTIONS(3241),
    [anon_sym_m] = ACTIONS(3241),
    [anon_sym_s] = ACTIONS(3241),
    [anon_sym_tr] = ACTIONS(3241),
    [anon_sym_y] = ACTIONS(3241),
    [anon_sym_if] = ACTIONS(3241),
    [anon_sym_unless] = ACTIONS(3241),
    [anon_sym_while] = ACTIONS(3241),
    [anon_sym_until] = ACTIONS(3241),
    [aux_sym__bareword_token1] = ACTIONS(3241),
    [sym_number] = ACTIONS(3241),
    [anon_sym_true] = ACTIONS(3241),
    [anon_sym_false] = ACTIONS(3241),
    [sym__single_quote] = ACTIONS(3239),
    [sym__double_quote] = ACTIONS(3239),
    [sym__backtick_quote] = ACTIONS(3239),
    [sym__search_slash_quote] = ACTIONS(3239),
    [sym__open_readline_bracket] = ACTIONS(3239),
    [sym__open_fileglob_bracket] = ACTIONS(3239),
    [sym__PERLY_HEREDOC] = ACTIONS(3239),
    [sym__ctrl_z_hack] = ACTIONS(3239),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3239),
    [sym__filetest] = ACTIONS(3239),
  },
  [STATE(1524)] = {
    [sym_heredoc_content] = STATE(1524),
    [ts_builtin_sym_end] = ACTIONS(3243),
    [sym__identifier] = ACTIONS(3245),
    [anon_sym_LBRACE] = ACTIONS(3245),
    [anon_sym_SEMI] = ACTIONS(3245),
    [anon_sym_package] = ACTIONS(3245),
    [anon_sym_class] = ACTIONS(3245),
    [anon_sym_role] = ACTIONS(3245),
    [anon_sym_BUILD] = ACTIONS(3245),
    [anon_sym_ADJUST] = ACTIONS(3245),
    [anon_sym_DOLLAR] = ACTIONS(3245),
    [anon_sym_AT] = ACTIONS(3245),
    [anon_sym_my] = ACTIONS(3245),
    [anon_sym_extended] = ACTIONS(3245),
    [anon_sym_async] = ACTIONS(3245),
    [anon_sym_sub] = ACTIONS(3245),
    [anon_sym_method] = ACTIONS(3245),
    [anon_sym_LPAREN] = ACTIONS(3245),
    [anon_sym_continue] = ACTIONS(3245),
    [anon_sym_state] = ACTIONS(3245),
    [anon_sym_our] = ACTIONS(3245),
    [anon_sym_try] = ACTIONS(3245),
    [anon_sym_defer] = ACTIONS(3245),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3245),
    [anon_sym_LBRACK] = ACTIONS(3245),
    [anon_sym_LT_LT] = ACTIONS(3245),
    [anon_sym_SLASH_SLASH] = ACTIONS(3245),
    [anon_sym_PLUS] = ACTIONS(3245),
    [anon_sym_DASH] = ACTIONS(3245),
    [anon_sym_TILDE] = ACTIONS(3245),
    [anon_sym_BANG] = ACTIONS(3245),
    [anon_sym_PLUS_PLUS] = ACTIONS(3245),
    [anon_sym_DASH_DASH] = ACTIONS(3245),
    [anon_sym_BSLASH] = ACTIONS(3243),
    [anon_sym_do] = ACTIONS(3245),
    [anon_sym_eval] = ACTIONS(3245),
    [anon_sym_await] = ACTIONS(3245),
    [anon_sym_field] = ACTIONS(3245),
    [anon_sym_local] = ACTIONS(3245),
    [anon_sym_dynamically] = ACTIONS(3245),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3245),
    [anon_sym_require] = ACTIONS(3245),
    [anon_sym_map] = ACTIONS(3245),
    [anon_sym_grep] = ACTIONS(3245),
    [anon_sym_sort] = ACTIONS(3245),
    [anon_sym_goto] = ACTIONS(3245),
    [anon_sym_return] = ACTIONS(3245),
    [anon_sym_undef] = ACTIONS(3245),
    [anon_sym_PERCENT2] = ACTIONS(3245),
    [anon_sym_AMP2] = ACTIONS(3245),
    [anon_sym_STAR2] = ACTIONS(3245),
    [anon_sym_use] = ACTIONS(3245),
    [anon_sym_no] = ACTIONS(3245),
    [anon_sym_for] = ACTIONS(3245),
    [anon_sym_foreach] = ACTIONS(3245),
    [anon_sym_last] = ACTIONS(3245),
    [anon_sym_next] = ACTIONS(3245),
    [anon_sym_redo] = ACTIONS(3245),
    [anon_sym_BEGIN] = ACTIONS(3245),
    [anon_sym_INIT] = ACTIONS(3245),
    [anon_sym_CHECK] = ACTIONS(3245),
    [anon_sym_UNITCHECK] = ACTIONS(3245),
    [anon_sym_END] = ACTIONS(3245),
    [anon_sym___FILE__] = ACTIONS(3245),
    [anon_sym___LINE__] = ACTIONS(3245),
    [anon_sym___PACKAGE__] = ACTIONS(3245),
    [anon_sym___SUB__] = ACTIONS(3245),
    [anon_sym_break] = ACTIONS(3245),
    [anon_sym_fork] = ACTIONS(3245),
    [anon_sym_getppid] = ACTIONS(3245),
    [anon_sym_time] = ACTIONS(3245),
    [anon_sym_times] = ACTIONS(3245),
    [anon_sym_wait] = ACTIONS(3245),
    [anon_sym_wantarray] = ACTIONS(3245),
    [anon_sym_abs] = ACTIONS(3245),
    [anon_sym_alarm] = ACTIONS(3245),
    [anon_sym_chop] = ACTIONS(3245),
    [anon_sym_chdir] = ACTIONS(3245),
    [anon_sym_close] = ACTIONS(3245),
    [anon_sym_closedir] = ACTIONS(3245),
    [anon_sym_caller] = ACTIONS(3245),
    [anon_sym_chomp] = ACTIONS(3245),
    [anon_sym_chr] = ACTIONS(3245),
    [anon_sym_cos] = ACTIONS(3245),
    [anon_sym_chroot] = ACTIONS(3245),
    [anon_sym_defined] = ACTIONS(3245),
    [anon_sym_delete] = ACTIONS(3245),
    [anon_sym_dbmclose] = ACTIONS(3245),
    [anon_sym_exists] = ACTIONS(3245),
    [anon_sym_exit] = ACTIONS(3245),
    [anon_sym_eof] = ACTIONS(3245),
    [anon_sym_exp] = ACTIONS(3245),
    [anon_sym_each] = ACTIONS(3245),
    [anon_sym_fc] = ACTIONS(3245),
    [anon_sym_fileno] = ACTIONS(3245),
    [anon_sym_gmtime] = ACTIONS(3245),
    [anon_sym_getc] = ACTIONS(3245),
    [anon_sym_getpgrp] = ACTIONS(3245),
    [anon_sym_getprotobyname] = ACTIONS(3245),
    [anon_sym_getpwname] = ACTIONS(3245),
    [anon_sym_getpwuid] = ACTIONS(3245),
    [anon_sym_getpeername] = ACTIONS(3245),
    [anon_sym_getnetbyname] = ACTIONS(3245),
    [anon_sym_getsockname] = ACTIONS(3245),
    [anon_sym_getgrnam] = ACTIONS(3245),
    [anon_sym_getgrgid] = ACTIONS(3245),
    [anon_sym_hex] = ACTIONS(3245),
    [anon_sym_int] = ACTIONS(3245),
    [anon_sym_keys] = ACTIONS(3245),
    [anon_sym_lc] = ACTIONS(3245),
    [anon_sym_lcfirst] = ACTIONS(3245),
    [anon_sym_length] = ACTIONS(3245),
    [anon_sym_localtime] = ACTIONS(3245),
    [anon_sym_log] = ACTIONS(3245),
    [anon_sym_lock] = ACTIONS(3245),
    [anon_sym_lstat] = ACTIONS(3245),
    [anon_sym_oct] = ACTIONS(3245),
    [anon_sym_ord] = ACTIONS(3245),
    [anon_sym_prototype] = ACTIONS(3245),
    [anon_sym_pop] = ACTIONS(3245),
    [anon_sym_pos] = ACTIONS(3245),
    [anon_sym_quotemeta] = ACTIONS(3245),
    [anon_sym_reset] = ACTIONS(3245),
    [anon_sym_rand] = ACTIONS(3245),
    [anon_sym_rmdir] = ACTIONS(3245),
    [anon_sym_readdir] = ACTIONS(3245),
    [anon_sym_readline] = ACTIONS(3245),
    [anon_sym_readpipe] = ACTIONS(3245),
    [anon_sym_rewinddir] = ACTIONS(3245),
    [anon_sym_readlink] = ACTIONS(3245),
    [anon_sym_ref] = ACTIONS(3245),
    [anon_sym_scalar] = ACTIONS(3245),
    [anon_sym_shift] = ACTIONS(3245),
    [anon_sym_sin] = ACTIONS(3245),
    [anon_sym_sleep] = ACTIONS(3245),
    [anon_sym_sqrt] = ACTIONS(3245),
    [anon_sym_srand] = ACTIONS(3245),
    [anon_sym_stat] = ACTIONS(3245),
    [anon_sym_study] = ACTIONS(3245),
    [anon_sym_tell] = ACTIONS(3245),
    [anon_sym_telldir] = ACTIONS(3245),
    [anon_sym_tied] = ACTIONS(3245),
    [anon_sym_uc] = ACTIONS(3245),
    [anon_sym_ucfirst] = ACTIONS(3245),
    [anon_sym_untie] = ACTIONS(3245),
    [anon_sym_umask] = ACTIONS(3245),
    [anon_sym_values] = ACTIONS(3245),
    [anon_sym_write] = ACTIONS(3245),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3245),
    [anon_sym___END__] = ACTIONS(3245),
    [anon_sym_EOT] = ACTIONS(3245),
    [anon_sym_q] = ACTIONS(3245),
    [anon_sym_qq] = ACTIONS(3245),
    [anon_sym_qw] = ACTIONS(3245),
    [anon_sym_qx] = ACTIONS(3245),
    [anon_sym_qr] = ACTIONS(3245),
    [anon_sym_m] = ACTIONS(3245),
    [anon_sym_s] = ACTIONS(3245),
    [anon_sym_tr] = ACTIONS(3245),
    [anon_sym_y] = ACTIONS(3245),
    [anon_sym_if] = ACTIONS(3245),
    [anon_sym_unless] = ACTIONS(3245),
    [anon_sym_while] = ACTIONS(3245),
    [anon_sym_until] = ACTIONS(3245),
    [aux_sym__bareword_token1] = ACTIONS(3245),
    [sym_number] = ACTIONS(3245),
    [anon_sym_true] = ACTIONS(3245),
    [anon_sym_false] = ACTIONS(3245),
    [sym__single_quote] = ACTIONS(3243),
    [sym__double_quote] = ACTIONS(3243),
    [sym__backtick_quote] = ACTIONS(3243),
    [sym__search_slash_quote] = ACTIONS(3243),
    [sym__open_readline_bracket] = ACTIONS(3243),
    [sym__open_fileglob_bracket] = ACTIONS(3243),
    [sym__PERLY_HEREDOC] = ACTIONS(3243),
    [sym__ctrl_z_hack] = ACTIONS(3243),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3243),
    [sym__filetest] = ACTIONS(3243),
  },
  [STATE(1525)] = {
    [sym_heredoc_content] = STATE(1525),
    [ts_builtin_sym_end] = ACTIONS(3247),
    [sym__identifier] = ACTIONS(3249),
    [anon_sym_LBRACE] = ACTIONS(3249),
    [anon_sym_SEMI] = ACTIONS(3249),
    [anon_sym_package] = ACTIONS(3249),
    [anon_sym_class] = ACTIONS(3249),
    [anon_sym_role] = ACTIONS(3249),
    [anon_sym_BUILD] = ACTIONS(3249),
    [anon_sym_ADJUST] = ACTIONS(3249),
    [anon_sym_DOLLAR] = ACTIONS(3249),
    [anon_sym_AT] = ACTIONS(3249),
    [anon_sym_my] = ACTIONS(3249),
    [anon_sym_extended] = ACTIONS(3249),
    [anon_sym_async] = ACTIONS(3249),
    [anon_sym_sub] = ACTIONS(3249),
    [anon_sym_method] = ACTIONS(3249),
    [anon_sym_LPAREN] = ACTIONS(3249),
    [anon_sym_continue] = ACTIONS(3249),
    [anon_sym_state] = ACTIONS(3249),
    [anon_sym_our] = ACTIONS(3249),
    [anon_sym_try] = ACTIONS(3249),
    [anon_sym_defer] = ACTIONS(3249),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3249),
    [anon_sym_LBRACK] = ACTIONS(3249),
    [anon_sym_LT_LT] = ACTIONS(3249),
    [anon_sym_SLASH_SLASH] = ACTIONS(3249),
    [anon_sym_PLUS] = ACTIONS(3249),
    [anon_sym_DASH] = ACTIONS(3249),
    [anon_sym_TILDE] = ACTIONS(3249),
    [anon_sym_BANG] = ACTIONS(3249),
    [anon_sym_PLUS_PLUS] = ACTIONS(3249),
    [anon_sym_DASH_DASH] = ACTIONS(3249),
    [anon_sym_BSLASH] = ACTIONS(3247),
    [anon_sym_do] = ACTIONS(3249),
    [anon_sym_eval] = ACTIONS(3249),
    [anon_sym_await] = ACTIONS(3249),
    [anon_sym_field] = ACTIONS(3249),
    [anon_sym_local] = ACTIONS(3249),
    [anon_sym_dynamically] = ACTIONS(3249),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3249),
    [anon_sym_require] = ACTIONS(3249),
    [anon_sym_map] = ACTIONS(3249),
    [anon_sym_grep] = ACTIONS(3249),
    [anon_sym_sort] = ACTIONS(3249),
    [anon_sym_goto] = ACTIONS(3249),
    [anon_sym_return] = ACTIONS(3249),
    [anon_sym_undef] = ACTIONS(3249),
    [anon_sym_PERCENT2] = ACTIONS(3249),
    [anon_sym_AMP2] = ACTIONS(3249),
    [anon_sym_STAR2] = ACTIONS(3249),
    [anon_sym_use] = ACTIONS(3249),
    [anon_sym_no] = ACTIONS(3249),
    [anon_sym_for] = ACTIONS(3249),
    [anon_sym_foreach] = ACTIONS(3249),
    [anon_sym_last] = ACTIONS(3249),
    [anon_sym_next] = ACTIONS(3249),
    [anon_sym_redo] = ACTIONS(3249),
    [anon_sym_BEGIN] = ACTIONS(3249),
    [anon_sym_INIT] = ACTIONS(3249),
    [anon_sym_CHECK] = ACTIONS(3249),
    [anon_sym_UNITCHECK] = ACTIONS(3249),
    [anon_sym_END] = ACTIONS(3249),
    [anon_sym___FILE__] = ACTIONS(3249),
    [anon_sym___LINE__] = ACTIONS(3249),
    [anon_sym___PACKAGE__] = ACTIONS(3249),
    [anon_sym___SUB__] = ACTIONS(3249),
    [anon_sym_break] = ACTIONS(3249),
    [anon_sym_fork] = ACTIONS(3249),
    [anon_sym_getppid] = ACTIONS(3249),
    [anon_sym_time] = ACTIONS(3249),
    [anon_sym_times] = ACTIONS(3249),
    [anon_sym_wait] = ACTIONS(3249),
    [anon_sym_wantarray] = ACTIONS(3249),
    [anon_sym_abs] = ACTIONS(3249),
    [anon_sym_alarm] = ACTIONS(3249),
    [anon_sym_chop] = ACTIONS(3249),
    [anon_sym_chdir] = ACTIONS(3249),
    [anon_sym_close] = ACTIONS(3249),
    [anon_sym_closedir] = ACTIONS(3249),
    [anon_sym_caller] = ACTIONS(3249),
    [anon_sym_chomp] = ACTIONS(3249),
    [anon_sym_chr] = ACTIONS(3249),
    [anon_sym_cos] = ACTIONS(3249),
    [anon_sym_chroot] = ACTIONS(3249),
    [anon_sym_defined] = ACTIONS(3249),
    [anon_sym_delete] = ACTIONS(3249),
    [anon_sym_dbmclose] = ACTIONS(3249),
    [anon_sym_exists] = ACTIONS(3249),
    [anon_sym_exit] = ACTIONS(3249),
    [anon_sym_eof] = ACTIONS(3249),
    [anon_sym_exp] = ACTIONS(3249),
    [anon_sym_each] = ACTIONS(3249),
    [anon_sym_fc] = ACTIONS(3249),
    [anon_sym_fileno] = ACTIONS(3249),
    [anon_sym_gmtime] = ACTIONS(3249),
    [anon_sym_getc] = ACTIONS(3249),
    [anon_sym_getpgrp] = ACTIONS(3249),
    [anon_sym_getprotobyname] = ACTIONS(3249),
    [anon_sym_getpwname] = ACTIONS(3249),
    [anon_sym_getpwuid] = ACTIONS(3249),
    [anon_sym_getpeername] = ACTIONS(3249),
    [anon_sym_getnetbyname] = ACTIONS(3249),
    [anon_sym_getsockname] = ACTIONS(3249),
    [anon_sym_getgrnam] = ACTIONS(3249),
    [anon_sym_getgrgid] = ACTIONS(3249),
    [anon_sym_hex] = ACTIONS(3249),
    [anon_sym_int] = ACTIONS(3249),
    [anon_sym_keys] = ACTIONS(3249),
    [anon_sym_lc] = ACTIONS(3249),
    [anon_sym_lcfirst] = ACTIONS(3249),
    [anon_sym_length] = ACTIONS(3249),
    [anon_sym_localtime] = ACTIONS(3249),
    [anon_sym_log] = ACTIONS(3249),
    [anon_sym_lock] = ACTIONS(3249),
    [anon_sym_lstat] = ACTIONS(3249),
    [anon_sym_oct] = ACTIONS(3249),
    [anon_sym_ord] = ACTIONS(3249),
    [anon_sym_prototype] = ACTIONS(3249),
    [anon_sym_pop] = ACTIONS(3249),
    [anon_sym_pos] = ACTIONS(3249),
    [anon_sym_quotemeta] = ACTIONS(3249),
    [anon_sym_reset] = ACTIONS(3249),
    [anon_sym_rand] = ACTIONS(3249),
    [anon_sym_rmdir] = ACTIONS(3249),
    [anon_sym_readdir] = ACTIONS(3249),
    [anon_sym_readline] = ACTIONS(3249),
    [anon_sym_readpipe] = ACTIONS(3249),
    [anon_sym_rewinddir] = ACTIONS(3249),
    [anon_sym_readlink] = ACTIONS(3249),
    [anon_sym_ref] = ACTIONS(3249),
    [anon_sym_scalar] = ACTIONS(3249),
    [anon_sym_shift] = ACTIONS(3249),
    [anon_sym_sin] = ACTIONS(3249),
    [anon_sym_sleep] = ACTIONS(3249),
    [anon_sym_sqrt] = ACTIONS(3249),
    [anon_sym_srand] = ACTIONS(3249),
    [anon_sym_stat] = ACTIONS(3249),
    [anon_sym_study] = ACTIONS(3249),
    [anon_sym_tell] = ACTIONS(3249),
    [anon_sym_telldir] = ACTIONS(3249),
    [anon_sym_tied] = ACTIONS(3249),
    [anon_sym_uc] = ACTIONS(3249),
    [anon_sym_ucfirst] = ACTIONS(3249),
    [anon_sym_untie] = ACTIONS(3249),
    [anon_sym_umask] = ACTIONS(3249),
    [anon_sym_values] = ACTIONS(3249),
    [anon_sym_write] = ACTIONS(3249),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3249),
    [anon_sym___END__] = ACTIONS(3249),
    [anon_sym_EOT] = ACTIONS(3249),
    [anon_sym_q] = ACTIONS(3249),
    [anon_sym_qq] = ACTIONS(3249),
    [anon_sym_qw] = ACTIONS(3249),
    [anon_sym_qx] = ACTIONS(3249),
    [anon_sym_qr] = ACTIONS(3249),
    [anon_sym_m] = ACTIONS(3249),
    [anon_sym_s] = ACTIONS(3249),
    [anon_sym_tr] = ACTIONS(3249),
    [anon_sym_y] = ACTIONS(3249),
    [anon_sym_if] = ACTIONS(3249),
    [anon_sym_unless] = ACTIONS(3249),
    [anon_sym_while] = ACTIONS(3249),
    [anon_sym_until] = ACTIONS(3249),
    [aux_sym__bareword_token1] = ACTIONS(3249),
    [sym_number] = ACTIONS(3249),
    [anon_sym_true] = ACTIONS(3249),
    [anon_sym_false] = ACTIONS(3249),
    [sym__single_quote] = ACTIONS(3247),
    [sym__double_quote] = ACTIONS(3247),
    [sym__backtick_quote] = ACTIONS(3247),
    [sym__search_slash_quote] = ACTIONS(3247),
    [sym__open_readline_bracket] = ACTIONS(3247),
    [sym__open_fileglob_bracket] = ACTIONS(3247),
    [sym__PERLY_HEREDOC] = ACTIONS(3247),
    [sym__ctrl_z_hack] = ACTIONS(3247),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3247),
    [sym__filetest] = ACTIONS(3247),
  },
  [STATE(1526)] = {
    [sym_heredoc_content] = STATE(1526),
    [ts_builtin_sym_end] = ACTIONS(3251),
    [sym__identifier] = ACTIONS(3253),
    [anon_sym_LBRACE] = ACTIONS(3253),
    [anon_sym_SEMI] = ACTIONS(3253),
    [anon_sym_package] = ACTIONS(3253),
    [anon_sym_class] = ACTIONS(3253),
    [anon_sym_role] = ACTIONS(3253),
    [anon_sym_BUILD] = ACTIONS(3253),
    [anon_sym_ADJUST] = ACTIONS(3253),
    [anon_sym_DOLLAR] = ACTIONS(3253),
    [anon_sym_AT] = ACTIONS(3253),
    [anon_sym_my] = ACTIONS(3253),
    [anon_sym_extended] = ACTIONS(3253),
    [anon_sym_async] = ACTIONS(3253),
    [anon_sym_sub] = ACTIONS(3253),
    [anon_sym_method] = ACTIONS(3253),
    [anon_sym_LPAREN] = ACTIONS(3253),
    [anon_sym_continue] = ACTIONS(3253),
    [anon_sym_state] = ACTIONS(3253),
    [anon_sym_our] = ACTIONS(3253),
    [anon_sym_try] = ACTIONS(3253),
    [anon_sym_defer] = ACTIONS(3253),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3253),
    [anon_sym_LBRACK] = ACTIONS(3253),
    [anon_sym_LT_LT] = ACTIONS(3253),
    [anon_sym_SLASH_SLASH] = ACTIONS(3253),
    [anon_sym_PLUS] = ACTIONS(3253),
    [anon_sym_DASH] = ACTIONS(3253),
    [anon_sym_TILDE] = ACTIONS(3253),
    [anon_sym_BANG] = ACTIONS(3253),
    [anon_sym_PLUS_PLUS] = ACTIONS(3253),
    [anon_sym_DASH_DASH] = ACTIONS(3253),
    [anon_sym_BSLASH] = ACTIONS(3251),
    [anon_sym_do] = ACTIONS(3253),
    [anon_sym_eval] = ACTIONS(3253),
    [anon_sym_await] = ACTIONS(3253),
    [anon_sym_field] = ACTIONS(3253),
    [anon_sym_local] = ACTIONS(3253),
    [anon_sym_dynamically] = ACTIONS(3253),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3253),
    [anon_sym_require] = ACTIONS(3253),
    [anon_sym_map] = ACTIONS(3253),
    [anon_sym_grep] = ACTIONS(3253),
    [anon_sym_sort] = ACTIONS(3253),
    [anon_sym_goto] = ACTIONS(3253),
    [anon_sym_return] = ACTIONS(3253),
    [anon_sym_undef] = ACTIONS(3253),
    [anon_sym_PERCENT2] = ACTIONS(3253),
    [anon_sym_AMP2] = ACTIONS(3253),
    [anon_sym_STAR2] = ACTIONS(3253),
    [anon_sym_use] = ACTIONS(3253),
    [anon_sym_no] = ACTIONS(3253),
    [anon_sym_for] = ACTIONS(3253),
    [anon_sym_foreach] = ACTIONS(3253),
    [anon_sym_last] = ACTIONS(3253),
    [anon_sym_next] = ACTIONS(3253),
    [anon_sym_redo] = ACTIONS(3253),
    [anon_sym_BEGIN] = ACTIONS(3253),
    [anon_sym_INIT] = ACTIONS(3253),
    [anon_sym_CHECK] = ACTIONS(3253),
    [anon_sym_UNITCHECK] = ACTIONS(3253),
    [anon_sym_END] = ACTIONS(3253),
    [anon_sym___FILE__] = ACTIONS(3253),
    [anon_sym___LINE__] = ACTIONS(3253),
    [anon_sym___PACKAGE__] = ACTIONS(3253),
    [anon_sym___SUB__] = ACTIONS(3253),
    [anon_sym_break] = ACTIONS(3253),
    [anon_sym_fork] = ACTIONS(3253),
    [anon_sym_getppid] = ACTIONS(3253),
    [anon_sym_time] = ACTIONS(3253),
    [anon_sym_times] = ACTIONS(3253),
    [anon_sym_wait] = ACTIONS(3253),
    [anon_sym_wantarray] = ACTIONS(3253),
    [anon_sym_abs] = ACTIONS(3253),
    [anon_sym_alarm] = ACTIONS(3253),
    [anon_sym_chop] = ACTIONS(3253),
    [anon_sym_chdir] = ACTIONS(3253),
    [anon_sym_close] = ACTIONS(3253),
    [anon_sym_closedir] = ACTIONS(3253),
    [anon_sym_caller] = ACTIONS(3253),
    [anon_sym_chomp] = ACTIONS(3253),
    [anon_sym_chr] = ACTIONS(3253),
    [anon_sym_cos] = ACTIONS(3253),
    [anon_sym_chroot] = ACTIONS(3253),
    [anon_sym_defined] = ACTIONS(3253),
    [anon_sym_delete] = ACTIONS(3253),
    [anon_sym_dbmclose] = ACTIONS(3253),
    [anon_sym_exists] = ACTIONS(3253),
    [anon_sym_exit] = ACTIONS(3253),
    [anon_sym_eof] = ACTIONS(3253),
    [anon_sym_exp] = ACTIONS(3253),
    [anon_sym_each] = ACTIONS(3253),
    [anon_sym_fc] = ACTIONS(3253),
    [anon_sym_fileno] = ACTIONS(3253),
    [anon_sym_gmtime] = ACTIONS(3253),
    [anon_sym_getc] = ACTIONS(3253),
    [anon_sym_getpgrp] = ACTIONS(3253),
    [anon_sym_getprotobyname] = ACTIONS(3253),
    [anon_sym_getpwname] = ACTIONS(3253),
    [anon_sym_getpwuid] = ACTIONS(3253),
    [anon_sym_getpeername] = ACTIONS(3253),
    [anon_sym_getnetbyname] = ACTIONS(3253),
    [anon_sym_getsockname] = ACTIONS(3253),
    [anon_sym_getgrnam] = ACTIONS(3253),
    [anon_sym_getgrgid] = ACTIONS(3253),
    [anon_sym_hex] = ACTIONS(3253),
    [anon_sym_int] = ACTIONS(3253),
    [anon_sym_keys] = ACTIONS(3253),
    [anon_sym_lc] = ACTIONS(3253),
    [anon_sym_lcfirst] = ACTIONS(3253),
    [anon_sym_length] = ACTIONS(3253),
    [anon_sym_localtime] = ACTIONS(3253),
    [anon_sym_log] = ACTIONS(3253),
    [anon_sym_lock] = ACTIONS(3253),
    [anon_sym_lstat] = ACTIONS(3253),
    [anon_sym_oct] = ACTIONS(3253),
    [anon_sym_ord] = ACTIONS(3253),
    [anon_sym_prototype] = ACTIONS(3253),
    [anon_sym_pop] = ACTIONS(3253),
    [anon_sym_pos] = ACTIONS(3253),
    [anon_sym_quotemeta] = ACTIONS(3253),
    [anon_sym_reset] = ACTIONS(3253),
    [anon_sym_rand] = ACTIONS(3253),
    [anon_sym_rmdir] = ACTIONS(3253),
    [anon_sym_readdir] = ACTIONS(3253),
    [anon_sym_readline] = ACTIONS(3253),
    [anon_sym_readpipe] = ACTIONS(3253),
    [anon_sym_rewinddir] = ACTIONS(3253),
    [anon_sym_readlink] = ACTIONS(3253),
    [anon_sym_ref] = ACTIONS(3253),
    [anon_sym_scalar] = ACTIONS(3253),
    [anon_sym_shift] = ACTIONS(3253),
    [anon_sym_sin] = ACTIONS(3253),
    [anon_sym_sleep] = ACTIONS(3253),
    [anon_sym_sqrt] = ACTIONS(3253),
    [anon_sym_srand] = ACTIONS(3253),
    [anon_sym_stat] = ACTIONS(3253),
    [anon_sym_study] = ACTIONS(3253),
    [anon_sym_tell] = ACTIONS(3253),
    [anon_sym_telldir] = ACTIONS(3253),
    [anon_sym_tied] = ACTIONS(3253),
    [anon_sym_uc] = ACTIONS(3253),
    [anon_sym_ucfirst] = ACTIONS(3253),
    [anon_sym_untie] = ACTIONS(3253),
    [anon_sym_umask] = ACTIONS(3253),
    [anon_sym_values] = ACTIONS(3253),
    [anon_sym_write] = ACTIONS(3253),
    [sym_comment] = ACTIONS(3),
    [anon_sym___DATA__] = ACTIONS(3253),
    [anon_sym___END__] = ACTIONS(3253),
    [anon_sym_EOT] = ACTIONS(3253),
    [anon_sym_q] = ACTIONS(3253),
    [anon_sym_qq] = ACTIONS(3253),
    [anon_sym_qw] = ACTIONS(3253),
    [anon_sym_qx] = ACTIONS(3253),
    [anon_sym_qr] = ACTIONS(3253),
    [anon_sym_m] = ACTIONS(3253),
    [anon_sym_s] = ACTIONS(3253),
    [anon_sym_tr] = ACTIONS(3253),
    [anon_sym_y] = ACTIONS(3253),
    [anon_sym_if] = ACTIONS(3253),
    [anon_sym_unless] = ACTIONS(3253),
    [anon_sym_while] = ACTIONS(3253),
    [anon_sym_until] = ACTIONS(3253),
    [aux_sym__bareword_token1] = ACTIONS(3253),
    [sym_number] = ACTIONS(3253),
    [anon_sym_true] = ACTIONS(3253),
    [anon_sym_false] = ACTIONS(3253),
    [sym__single_quote] = ACTIONS(3251),
    [sym__double_quote] = ACTIONS(3251),
    [sym__backtick_quote] = ACTIONS(3251),
    [sym__search_slash_quote] = ACTIONS(3251),
    [sym__open_readline_bracket] = ACTIONS(3251),
    [sym__open_fileglob_bracket] = ACTIONS(3251),
    [sym__PERLY_HEREDOC] = ACTIONS(3251),
    [sym__ctrl_z_hack] = ACTIONS(3251),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3251),
    [sym__filetest] = ACTIONS(3251),
  },
  [STATE(1527)] = {
    [sym_heredoc_content] = STATE(1527),
    [sym__identifier] = ACTIONS(3255),
    [anon_sym_LBRACE] = ACTIONS(3255),
    [anon_sym_RBRACE] = ACTIONS(3255),
    [anon_sym_SEMI] = ACTIONS(3255),
    [anon_sym_package] = ACTIONS(3255),
    [anon_sym_class] = ACTIONS(3255),
    [anon_sym_role] = ACTIONS(3255),
    [anon_sym_BUILD] = ACTIONS(3255),
    [anon_sym_ADJUST] = ACTIONS(3255),
    [anon_sym_DOLLAR] = ACTIONS(3255),
    [anon_sym_AT] = ACTIONS(3255),
    [anon_sym_my] = ACTIONS(3255),
    [anon_sym_extended] = ACTIONS(3255),
    [anon_sym_async] = ACTIONS(3255),
    [anon_sym_sub] = ACTIONS(3255),
    [anon_sym_method] = ACTIONS(3255),
    [anon_sym_LPAREN] = ACTIONS(3255),
    [anon_sym_continue] = ACTIONS(3255),
    [anon_sym_state] = ACTIONS(3255),
    [anon_sym_our] = ACTIONS(3255),
    [anon_sym_try] = ACTIONS(3255),
    [anon_sym_defer] = ACTIONS(3255),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3255),
    [anon_sym_LBRACK] = ACTIONS(3255),
    [anon_sym_LT_LT] = ACTIONS(3255),
    [anon_sym_SLASH_SLASH] = ACTIONS(3255),
    [anon_sym_PLUS] = ACTIONS(3255),
    [anon_sym_DASH] = ACTIONS(3255),
    [anon_sym_TILDE] = ACTIONS(3255),
    [anon_sym_BANG] = ACTIONS(3255),
    [anon_sym_PLUS_PLUS] = ACTIONS(3255),
    [anon_sym_DASH_DASH] = ACTIONS(3255),
    [anon_sym_BSLASH] = ACTIONS(3255),
    [anon_sym_do] = ACTIONS(3255),
    [anon_sym_eval] = ACTIONS(3255),
    [anon_sym_await] = ACTIONS(3255),
    [anon_sym_field] = ACTIONS(3255),
    [anon_sym_local] = ACTIONS(3255),
    [anon_sym_dynamically] = ACTIONS(3255),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3255),
    [anon_sym_require] = ACTIONS(3255),
    [anon_sym_map] = ACTIONS(3255),
    [anon_sym_grep] = ACTIONS(3255),
    [anon_sym_sort] = ACTIONS(3255),
    [anon_sym_goto] = ACTIONS(3255),
    [anon_sym_return] = ACTIONS(3255),
    [anon_sym_undef] = ACTIONS(3255),
    [anon_sym_PERCENT2] = ACTIONS(3255),
    [anon_sym_AMP2] = ACTIONS(3255),
    [anon_sym_STAR2] = ACTIONS(3255),
    [aux_sym__var_indirob_autoquote_token1] = ACTIONS(3255),
    [anon_sym_use] = ACTIONS(3255),
    [anon_sym_no] = ACTIONS(3255),
    [anon_sym_for] = ACTIONS(3255),
    [anon_sym_foreach] = ACTIONS(3255),
    [anon_sym_last] = ACTIONS(3255),
    [anon_sym_next] = ACTIONS(3255),
    [anon_sym_redo] = ACTIONS(3255),
    [anon_sym_BEGIN] = ACTIONS(3255),
    [anon_sym_INIT] = ACTIONS(3255),
    [anon_sym_CHECK] = ACTIONS(3255),
    [anon_sym_UNITCHECK] = ACTIONS(3255),
    [anon_sym_END] = ACTIONS(3255),
    [anon_sym___FILE__] = ACTIONS(3255),
    [anon_sym___LINE__] = ACTIONS(3255),
    [anon_sym___PACKAGE__] = ACTIONS(3255),
    [anon_sym___SUB__] = ACTIONS(3255),
    [anon_sym_break] = ACTIONS(3255),
    [anon_sym_fork] = ACTIONS(3255),
    [anon_sym_getppid] = ACTIONS(3255),
    [anon_sym_time] = ACTIONS(3255),
    [anon_sym_times] = ACTIONS(3255),
    [anon_sym_wait] = ACTIONS(3255),
    [anon_sym_wantarray] = ACTIONS(3255),
    [anon_sym_abs] = ACTIONS(3255),
    [anon_sym_alarm] = ACTIONS(3255),
    [anon_sym_chop] = ACTIONS(3255),
    [anon_sym_chdir] = ACTIONS(3255),
    [anon_sym_close] = ACTIONS(3255),
    [anon_sym_closedir] = ACTIONS(3255),
    [anon_sym_caller] = ACTIONS(3255),
    [anon_sym_chomp] = ACTIONS(3255),
    [anon_sym_chr] = ACTIONS(3255),
    [anon_sym_cos] = ACTIONS(3255),
    [anon_sym_chroot] = ACTIONS(3255),
    [anon_sym_defined] = ACTIONS(3255),
    [anon_sym_delete] = ACTIONS(3255),
    [anon_sym_dbmclose] = ACTIONS(3255),
    [anon_sym_exists] = ACTIONS(3255),
    [anon_sym_exit] = ACTIONS(3255),
    [anon_sym_eof] = ACTIONS(3255),
    [anon_sym_exp] = ACTIONS(3255),
    [anon_sym_each] = ACTIONS(3255),
    [anon_sym_fc] = ACTIONS(3255),
    [anon_sym_fileno] = ACTIONS(3255),
    [anon_sym_gmtime] = ACTIONS(3255),
    [anon_sym_getc] = ACTIONS(3255),
    [anon_sym_getpgrp] = ACTIONS(3255),
    [anon_sym_getprotobyname] = ACTIONS(3255),
    [anon_sym_getpwname] = ACTIONS(3255),
    [anon_sym_getpwuid] = ACTIONS(3255),
    [anon_sym_getpeername] = ACTIONS(3255),
    [anon_sym_getnetbyname] = ACTIONS(3255),
    [anon_sym_getsockname] = ACTIONS(3255),
    [anon_sym_getgrnam] = ACTIONS(3255),
    [anon_sym_getgrgid] = ACTIONS(3255),
    [anon_sym_hex] = ACTIONS(3255),
    [anon_sym_int] = ACTIONS(3255),
    [anon_sym_keys] = ACTIONS(3255),
    [anon_sym_lc] = ACTIONS(3255),
    [anon_sym_lcfirst] = ACTIONS(3255),
    [anon_sym_length] = ACTIONS(3255),
    [anon_sym_localtime] = ACTIONS(3255),
    [anon_sym_log] = ACTIONS(3255),
    [anon_sym_lock] = ACTIONS(3255),
    [anon_sym_lstat] = ACTIONS(3255),
    [anon_sym_oct] = ACTIONS(3255),
    [anon_sym_ord] = ACTIONS(3255),
    [anon_sym_prototype] = ACTIONS(3255),
    [anon_sym_pop] = ACTIONS(3255),
    [anon_sym_pos] = ACTIONS(3255),
    [anon_sym_quotemeta] = ACTIONS(3255),
    [anon_sym_reset] = ACTIONS(3255),
    [anon_sym_rand] = ACTIONS(3255),
    [anon_sym_rmdir] = ACTIONS(3255),
    [anon_sym_readdir] = ACTIONS(3255),
    [anon_sym_readline] = ACTIONS(3255),
    [anon_sym_readpipe] = ACTIONS(3255),
    [anon_sym_rewinddir] = ACTIONS(3255),
    [anon_sym_readlink] = ACTIONS(3255),
    [anon_sym_ref] = ACTIONS(3255),
    [anon_sym_scalar] = ACTIONS(3255),
    [anon_sym_shift] = ACTIONS(3255),
    [anon_sym_sin] = ACTIONS(3255),
    [anon_sym_sleep] = ACTIONS(3255),
    [anon_sym_sqrt] = ACTIONS(3255),
    [anon_sym_srand] = ACTIONS(3255),
    [anon_sym_stat] = ACTIONS(3255),
    [anon_sym_study] = ACTIONS(3255),
    [anon_sym_tell] = ACTIONS(3255),
    [anon_sym_telldir] = ACTIONS(3255),
    [anon_sym_tied] = ACTIONS(3255),
    [anon_sym_uc] = ACTIONS(3255),
    [anon_sym_ucfirst] = ACTIONS(3255),
    [anon_sym_untie] = ACTIONS(3255),
    [anon_sym_umask] = ACTIONS(3255),
    [anon_sym_values] = ACTIONS(3255),
    [anon_sym_write] = ACTIONS(3255),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3255),
    [anon_sym_qq] = ACTIONS(3255),
    [anon_sym_qw] = ACTIONS(3255),
    [anon_sym_qx] = ACTIONS(3255),
    [anon_sym_qr] = ACTIONS(3255),
    [anon_sym_m] = ACTIONS(3255),
    [anon_sym_s] = ACTIONS(3255),
    [anon_sym_tr] = ACTIONS(3255),
    [anon_sym_y] = ACTIONS(3255),
    [anon_sym_if] = ACTIONS(3255),
    [anon_sym_unless] = ACTIONS(3255),
    [anon_sym_while] = ACTIONS(3255),
    [anon_sym_until] = ACTIONS(3255),
    [aux_sym__ident_special_token1] = ACTIONS(3255),
    [aux_sym__bareword_token1] = ACTIONS(3255),
    [sym_number] = ACTIONS(3255),
    [anon_sym_true] = ACTIONS(3255),
    [anon_sym_false] = ACTIONS(3255),
    [sym__single_quote] = ACTIONS(3257),
    [sym__double_quote] = ACTIONS(3257),
    [sym__backtick_quote] = ACTIONS(3257),
    [sym__search_slash_quote] = ACTIONS(3257),
    [sym__open_readline_bracket] = ACTIONS(3257),
    [sym__open_fileglob_bracket] = ACTIONS(3257),
    [sym__PERLY_HEREDOC] = ACTIONS(3257),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3257),
    [sym__filetest] = ACTIONS(3257),
    [sym__brace_autoquoted_token] = ACTIONS(3257),
  },
  [STATE(1528)] = {
    [sym_heredoc_content] = STATE(1528),
    [sym__identifier] = ACTIONS(2874),
    [anon_sym_LBRACE] = ACTIONS(2874),
    [anon_sym_RBRACE] = ACTIONS(2874),
    [anon_sym_SEMI] = ACTIONS(2874),
    [anon_sym_package] = ACTIONS(2874),
    [anon_sym_class] = ACTIONS(2874),
    [anon_sym_role] = ACTIONS(2874),
    [anon_sym_BUILD] = ACTIONS(2874),
    [anon_sym_ADJUST] = ACTIONS(2874),
    [anon_sym_DOLLAR] = ACTIONS(2874),
    [anon_sym_AT] = ACTIONS(2874),
    [anon_sym_my] = ACTIONS(2874),
    [anon_sym_extended] = ACTIONS(2874),
    [anon_sym_async] = ACTIONS(2874),
    [anon_sym_sub] = ACTIONS(2874),
    [anon_sym_method] = ACTIONS(2874),
    [anon_sym_LPAREN] = ACTIONS(2874),
    [anon_sym_continue] = ACTIONS(2874),
    [anon_sym_state] = ACTIONS(2874),
    [anon_sym_our] = ACTIONS(2874),
    [anon_sym_try] = ACTIONS(2874),
    [anon_sym_catch] = ACTIONS(3259),
    [anon_sym_finally] = ACTIONS(3261),
    [anon_sym_defer] = ACTIONS(2874),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2874),
    [anon_sym_LBRACK] = ACTIONS(2874),
    [anon_sym_LT_LT] = ACTIONS(2874),
    [anon_sym_SLASH_SLASH] = ACTIONS(2874),
    [anon_sym_PLUS] = ACTIONS(2874),
    [anon_sym_DASH] = ACTIONS(2874),
    [anon_sym_TILDE] = ACTIONS(2874),
    [anon_sym_BANG] = ACTIONS(2874),
    [anon_sym_PLUS_PLUS] = ACTIONS(2874),
    [anon_sym_DASH_DASH] = ACTIONS(2874),
    [anon_sym_BSLASH] = ACTIONS(2872),
    [anon_sym_do] = ACTIONS(2874),
    [anon_sym_eval] = ACTIONS(2874),
    [anon_sym_await] = ACTIONS(2874),
    [anon_sym_field] = ACTIONS(2874),
    [anon_sym_local] = ACTIONS(2874),
    [anon_sym_dynamically] = ACTIONS(2874),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2874),
    [anon_sym_require] = ACTIONS(2874),
    [anon_sym_map] = ACTIONS(2874),
    [anon_sym_grep] = ACTIONS(2874),
    [anon_sym_sort] = ACTIONS(2874),
    [anon_sym_goto] = ACTIONS(2874),
    [anon_sym_return] = ACTIONS(2874),
    [anon_sym_undef] = ACTIONS(2874),
    [anon_sym_PERCENT2] = ACTIONS(2874),
    [anon_sym_AMP2] = ACTIONS(2874),
    [anon_sym_STAR2] = ACTIONS(2874),
    [anon_sym_use] = ACTIONS(2874),
    [anon_sym_no] = ACTIONS(2874),
    [anon_sym_for] = ACTIONS(2874),
    [anon_sym_foreach] = ACTIONS(2874),
    [anon_sym_last] = ACTIONS(2874),
    [anon_sym_next] = ACTIONS(2874),
    [anon_sym_redo] = ACTIONS(2874),
    [anon_sym_BEGIN] = ACTIONS(2874),
    [anon_sym_INIT] = ACTIONS(2874),
    [anon_sym_CHECK] = ACTIONS(2874),
    [anon_sym_UNITCHECK] = ACTIONS(2874),
    [anon_sym_END] = ACTIONS(2874),
    [anon_sym___FILE__] = ACTIONS(2874),
    [anon_sym___LINE__] = ACTIONS(2874),
    [anon_sym___PACKAGE__] = ACTIONS(2874),
    [anon_sym___SUB__] = ACTIONS(2874),
    [anon_sym_break] = ACTIONS(2874),
    [anon_sym_fork] = ACTIONS(2874),
    [anon_sym_getppid] = ACTIONS(2874),
    [anon_sym_time] = ACTIONS(2874),
    [anon_sym_times] = ACTIONS(2874),
    [anon_sym_wait] = ACTIONS(2874),
    [anon_sym_wantarray] = ACTIONS(2874),
    [anon_sym_abs] = ACTIONS(2874),
    [anon_sym_alarm] = ACTIONS(2874),
    [anon_sym_chop] = ACTIONS(2874),
    [anon_sym_chdir] = ACTIONS(2874),
    [anon_sym_close] = ACTIONS(2874),
    [anon_sym_closedir] = ACTIONS(2874),
    [anon_sym_caller] = ACTIONS(2874),
    [anon_sym_chomp] = ACTIONS(2874),
    [anon_sym_chr] = ACTIONS(2874),
    [anon_sym_cos] = ACTIONS(2874),
    [anon_sym_chroot] = ACTIONS(2874),
    [anon_sym_defined] = ACTIONS(2874),
    [anon_sym_delete] = ACTIONS(2874),
    [anon_sym_dbmclose] = ACTIONS(2874),
    [anon_sym_exists] = ACTIONS(2874),
    [anon_sym_exit] = ACTIONS(2874),
    [anon_sym_eof] = ACTIONS(2874),
    [anon_sym_exp] = ACTIONS(2874),
    [anon_sym_each] = ACTIONS(2874),
    [anon_sym_fc] = ACTIONS(2874),
    [anon_sym_fileno] = ACTIONS(2874),
    [anon_sym_gmtime] = ACTIONS(2874),
    [anon_sym_getc] = ACTIONS(2874),
    [anon_sym_getpgrp] = ACTIONS(2874),
    [anon_sym_getprotobyname] = ACTIONS(2874),
    [anon_sym_getpwname] = ACTIONS(2874),
    [anon_sym_getpwuid] = ACTIONS(2874),
    [anon_sym_getpeername] = ACTIONS(2874),
    [anon_sym_getnetbyname] = ACTIONS(2874),
    [anon_sym_getsockname] = ACTIONS(2874),
    [anon_sym_getgrnam] = ACTIONS(2874),
    [anon_sym_getgrgid] = ACTIONS(2874),
    [anon_sym_hex] = ACTIONS(2874),
    [anon_sym_int] = ACTIONS(2874),
    [anon_sym_keys] = ACTIONS(2874),
    [anon_sym_lc] = ACTIONS(2874),
    [anon_sym_lcfirst] = ACTIONS(2874),
    [anon_sym_length] = ACTIONS(2874),
    [anon_sym_localtime] = ACTIONS(2874),
    [anon_sym_log] = ACTIONS(2874),
    [anon_sym_lock] = ACTIONS(2874),
    [anon_sym_lstat] = ACTIONS(2874),
    [anon_sym_oct] = ACTIONS(2874),
    [anon_sym_ord] = ACTIONS(2874),
    [anon_sym_prototype] = ACTIONS(2874),
    [anon_sym_pop] = ACTIONS(2874),
    [anon_sym_pos] = ACTIONS(2874),
    [anon_sym_quotemeta] = ACTIONS(2874),
    [anon_sym_reset] = ACTIONS(2874),
    [anon_sym_rand] = ACTIONS(2874),
    [anon_sym_rmdir] = ACTIONS(2874),
    [anon_sym_readdir] = ACTIONS(2874),
    [anon_sym_readline] = ACTIONS(2874),
    [anon_sym_readpipe] = ACTIONS(2874),
    [anon_sym_rewinddir] = ACTIONS(2874),
    [anon_sym_readlink] = ACTIONS(2874),
    [anon_sym_ref] = ACTIONS(2874),
    [anon_sym_scalar] = ACTIONS(2874),
    [anon_sym_shift] = ACTIONS(2874),
    [anon_sym_sin] = ACTIONS(2874),
    [anon_sym_sleep] = ACTIONS(2874),
    [anon_sym_sqrt] = ACTIONS(2874),
    [anon_sym_srand] = ACTIONS(2874),
    [anon_sym_stat] = ACTIONS(2874),
    [anon_sym_study] = ACTIONS(2874),
    [anon_sym_tell] = ACTIONS(2874),
    [anon_sym_telldir] = ACTIONS(2874),
    [anon_sym_tied] = ACTIONS(2874),
    [anon_sym_uc] = ACTIONS(2874),
    [anon_sym_ucfirst] = ACTIONS(2874),
    [anon_sym_untie] = ACTIONS(2874),
    [anon_sym_umask] = ACTIONS(2874),
    [anon_sym_values] = ACTIONS(2874),
    [anon_sym_write] = ACTIONS(2874),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2874),
    [anon_sym_qq] = ACTIONS(2874),
    [anon_sym_qw] = ACTIONS(2874),
    [anon_sym_qx] = ACTIONS(2874),
    [anon_sym_qr] = ACTIONS(2874),
    [anon_sym_m] = ACTIONS(2874),
    [anon_sym_s] = ACTIONS(2874),
    [anon_sym_tr] = ACTIONS(2874),
    [anon_sym_y] = ACTIONS(2874),
    [anon_sym_if] = ACTIONS(2874),
    [anon_sym_unless] = ACTIONS(2874),
    [anon_sym_while] = ACTIONS(2874),
    [anon_sym_until] = ACTIONS(2874),
    [aux_sym__bareword_token1] = ACTIONS(2874),
    [sym_number] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2874),
    [anon_sym_false] = ACTIONS(2874),
    [sym__single_quote] = ACTIONS(2872),
    [sym__double_quote] = ACTIONS(2872),
    [sym__backtick_quote] = ACTIONS(2872),
    [sym__search_slash_quote] = ACTIONS(2872),
    [sym__open_readline_bracket] = ACTIONS(2872),
    [sym__open_fileglob_bracket] = ACTIONS(2872),
    [sym__PERLY_HEREDOC] = ACTIONS(2872),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2872),
    [sym__filetest] = ACTIONS(2872),
  },
  [STATE(1529)] = {
    [sym_heredoc_content] = STATE(1529),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_catch] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1530)] = {
    [sym_heredoc_content] = STATE(1530),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_else] = ACTIONS(2688),
    [anon_sym_elsif] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1531)] = {
    [sym_heredoc_content] = STATE(1531),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_catch] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1532)] = {
    [sym_heredoc_content] = STATE(1532),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_else] = ACTIONS(2532),
    [anon_sym_elsif] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1533)] = {
    [sym_heredoc_content] = STATE(1533),
    [sym__identifier] = ACTIONS(2886),
    [anon_sym_LBRACE] = ACTIONS(2886),
    [anon_sym_RBRACE] = ACTIONS(2886),
    [anon_sym_SEMI] = ACTIONS(2886),
    [anon_sym_package] = ACTIONS(2886),
    [anon_sym_class] = ACTIONS(2886),
    [anon_sym_role] = ACTIONS(2886),
    [anon_sym_BUILD] = ACTIONS(2886),
    [anon_sym_ADJUST] = ACTIONS(2886),
    [anon_sym_DOLLAR] = ACTIONS(2886),
    [anon_sym_AT] = ACTIONS(2886),
    [anon_sym_my] = ACTIONS(2886),
    [anon_sym_extended] = ACTIONS(2886),
    [anon_sym_async] = ACTIONS(2886),
    [anon_sym_sub] = ACTIONS(2886),
    [anon_sym_method] = ACTIONS(2886),
    [anon_sym_LPAREN] = ACTIONS(2886),
    [anon_sym_continue] = ACTIONS(2886),
    [anon_sym_state] = ACTIONS(2886),
    [anon_sym_our] = ACTIONS(2886),
    [anon_sym_try] = ACTIONS(2886),
    [anon_sym_finally] = ACTIONS(3263),
    [anon_sym_defer] = ACTIONS(2886),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2886),
    [anon_sym_LBRACK] = ACTIONS(2886),
    [anon_sym_LT_LT] = ACTIONS(2886),
    [anon_sym_SLASH_SLASH] = ACTIONS(2886),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_PLUS_PLUS] = ACTIONS(2886),
    [anon_sym_DASH_DASH] = ACTIONS(2886),
    [anon_sym_BSLASH] = ACTIONS(2884),
    [anon_sym_do] = ACTIONS(2886),
    [anon_sym_eval] = ACTIONS(2886),
    [anon_sym_await] = ACTIONS(2886),
    [anon_sym_field] = ACTIONS(2886),
    [anon_sym_local] = ACTIONS(2886),
    [anon_sym_dynamically] = ACTIONS(2886),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2886),
    [anon_sym_require] = ACTIONS(2886),
    [anon_sym_map] = ACTIONS(2886),
    [anon_sym_grep] = ACTIONS(2886),
    [anon_sym_sort] = ACTIONS(2886),
    [anon_sym_goto] = ACTIONS(2886),
    [anon_sym_return] = ACTIONS(2886),
    [anon_sym_undef] = ACTIONS(2886),
    [anon_sym_PERCENT2] = ACTIONS(2886),
    [anon_sym_AMP2] = ACTIONS(2886),
    [anon_sym_STAR2] = ACTIONS(2886),
    [anon_sym_use] = ACTIONS(2886),
    [anon_sym_no] = ACTIONS(2886),
    [anon_sym_for] = ACTIONS(2886),
    [anon_sym_foreach] = ACTIONS(2886),
    [anon_sym_last] = ACTIONS(2886),
    [anon_sym_next] = ACTIONS(2886),
    [anon_sym_redo] = ACTIONS(2886),
    [anon_sym_BEGIN] = ACTIONS(2886),
    [anon_sym_INIT] = ACTIONS(2886),
    [anon_sym_CHECK] = ACTIONS(2886),
    [anon_sym_UNITCHECK] = ACTIONS(2886),
    [anon_sym_END] = ACTIONS(2886),
    [anon_sym___FILE__] = ACTIONS(2886),
    [anon_sym___LINE__] = ACTIONS(2886),
    [anon_sym___PACKAGE__] = ACTIONS(2886),
    [anon_sym___SUB__] = ACTIONS(2886),
    [anon_sym_break] = ACTIONS(2886),
    [anon_sym_fork] = ACTIONS(2886),
    [anon_sym_getppid] = ACTIONS(2886),
    [anon_sym_time] = ACTIONS(2886),
    [anon_sym_times] = ACTIONS(2886),
    [anon_sym_wait] = ACTIONS(2886),
    [anon_sym_wantarray] = ACTIONS(2886),
    [anon_sym_abs] = ACTIONS(2886),
    [anon_sym_alarm] = ACTIONS(2886),
    [anon_sym_chop] = ACTIONS(2886),
    [anon_sym_chdir] = ACTIONS(2886),
    [anon_sym_close] = ACTIONS(2886),
    [anon_sym_closedir] = ACTIONS(2886),
    [anon_sym_caller] = ACTIONS(2886),
    [anon_sym_chomp] = ACTIONS(2886),
    [anon_sym_chr] = ACTIONS(2886),
    [anon_sym_cos] = ACTIONS(2886),
    [anon_sym_chroot] = ACTIONS(2886),
    [anon_sym_defined] = ACTIONS(2886),
    [anon_sym_delete] = ACTIONS(2886),
    [anon_sym_dbmclose] = ACTIONS(2886),
    [anon_sym_exists] = ACTIONS(2886),
    [anon_sym_exit] = ACTIONS(2886),
    [anon_sym_eof] = ACTIONS(2886),
    [anon_sym_exp] = ACTIONS(2886),
    [anon_sym_each] = ACTIONS(2886),
    [anon_sym_fc] = ACTIONS(2886),
    [anon_sym_fileno] = ACTIONS(2886),
    [anon_sym_gmtime] = ACTIONS(2886),
    [anon_sym_getc] = ACTIONS(2886),
    [anon_sym_getpgrp] = ACTIONS(2886),
    [anon_sym_getprotobyname] = ACTIONS(2886),
    [anon_sym_getpwname] = ACTIONS(2886),
    [anon_sym_getpwuid] = ACTIONS(2886),
    [anon_sym_getpeername] = ACTIONS(2886),
    [anon_sym_getnetbyname] = ACTIONS(2886),
    [anon_sym_getsockname] = ACTIONS(2886),
    [anon_sym_getgrnam] = ACTIONS(2886),
    [anon_sym_getgrgid] = ACTIONS(2886),
    [anon_sym_hex] = ACTIONS(2886),
    [anon_sym_int] = ACTIONS(2886),
    [anon_sym_keys] = ACTIONS(2886),
    [anon_sym_lc] = ACTIONS(2886),
    [anon_sym_lcfirst] = ACTIONS(2886),
    [anon_sym_length] = ACTIONS(2886),
    [anon_sym_localtime] = ACTIONS(2886),
    [anon_sym_log] = ACTIONS(2886),
    [anon_sym_lock] = ACTIONS(2886),
    [anon_sym_lstat] = ACTIONS(2886),
    [anon_sym_oct] = ACTIONS(2886),
    [anon_sym_ord] = ACTIONS(2886),
    [anon_sym_prototype] = ACTIONS(2886),
    [anon_sym_pop] = ACTIONS(2886),
    [anon_sym_pos] = ACTIONS(2886),
    [anon_sym_quotemeta] = ACTIONS(2886),
    [anon_sym_reset] = ACTIONS(2886),
    [anon_sym_rand] = ACTIONS(2886),
    [anon_sym_rmdir] = ACTIONS(2886),
    [anon_sym_readdir] = ACTIONS(2886),
    [anon_sym_readline] = ACTIONS(2886),
    [anon_sym_readpipe] = ACTIONS(2886),
    [anon_sym_rewinddir] = ACTIONS(2886),
    [anon_sym_readlink] = ACTIONS(2886),
    [anon_sym_ref] = ACTIONS(2886),
    [anon_sym_scalar] = ACTIONS(2886),
    [anon_sym_shift] = ACTIONS(2886),
    [anon_sym_sin] = ACTIONS(2886),
    [anon_sym_sleep] = ACTIONS(2886),
    [anon_sym_sqrt] = ACTIONS(2886),
    [anon_sym_srand] = ACTIONS(2886),
    [anon_sym_stat] = ACTIONS(2886),
    [anon_sym_study] = ACTIONS(2886),
    [anon_sym_tell] = ACTIONS(2886),
    [anon_sym_telldir] = ACTIONS(2886),
    [anon_sym_tied] = ACTIONS(2886),
    [anon_sym_uc] = ACTIONS(2886),
    [anon_sym_ucfirst] = ACTIONS(2886),
    [anon_sym_untie] = ACTIONS(2886),
    [anon_sym_umask] = ACTIONS(2886),
    [anon_sym_values] = ACTIONS(2886),
    [anon_sym_write] = ACTIONS(2886),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2886),
    [anon_sym_qq] = ACTIONS(2886),
    [anon_sym_qw] = ACTIONS(2886),
    [anon_sym_qx] = ACTIONS(2886),
    [anon_sym_qr] = ACTIONS(2886),
    [anon_sym_m] = ACTIONS(2886),
    [anon_sym_s] = ACTIONS(2886),
    [anon_sym_tr] = ACTIONS(2886),
    [anon_sym_y] = ACTIONS(2886),
    [anon_sym_if] = ACTIONS(2886),
    [anon_sym_unless] = ACTIONS(2886),
    [anon_sym_while] = ACTIONS(2886),
    [anon_sym_until] = ACTIONS(2886),
    [aux_sym__bareword_token1] = ACTIONS(2886),
    [sym_number] = ACTIONS(2886),
    [anon_sym_true] = ACTIONS(2886),
    [anon_sym_false] = ACTIONS(2886),
    [sym__single_quote] = ACTIONS(2884),
    [sym__double_quote] = ACTIONS(2884),
    [sym__backtick_quote] = ACTIONS(2884),
    [sym__search_slash_quote] = ACTIONS(2884),
    [sym__open_readline_bracket] = ACTIONS(2884),
    [sym__open_fileglob_bracket] = ACTIONS(2884),
    [sym__PERLY_HEREDOC] = ACTIONS(2884),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2884),
    [sym__filetest] = ACTIONS(2884),
  },
  [STATE(1534)] = {
    [sym_heredoc_content] = STATE(1534),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1535)] = {
    [sym_heredoc_content] = STATE(1535),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1536)] = {
    [sym_heredoc_content] = STATE(1536),
    [sym__identifier] = ACTIONS(2892),
    [anon_sym_LBRACE] = ACTIONS(2892),
    [anon_sym_RBRACE] = ACTIONS(2892),
    [anon_sym_SEMI] = ACTIONS(2892),
    [anon_sym_package] = ACTIONS(2892),
    [anon_sym_class] = ACTIONS(2892),
    [anon_sym_role] = ACTIONS(2892),
    [anon_sym_BUILD] = ACTIONS(2892),
    [anon_sym_ADJUST] = ACTIONS(2892),
    [anon_sym_DOLLAR] = ACTIONS(2892),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_my] = ACTIONS(2892),
    [anon_sym_extended] = ACTIONS(2892),
    [anon_sym_async] = ACTIONS(2892),
    [anon_sym_sub] = ACTIONS(2892),
    [anon_sym_method] = ACTIONS(2892),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_continue] = ACTIONS(2892),
    [anon_sym_state] = ACTIONS(2892),
    [anon_sym_our] = ACTIONS(2892),
    [anon_sym_try] = ACTIONS(2892),
    [anon_sym_finally] = ACTIONS(3265),
    [anon_sym_defer] = ACTIONS(2892),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2892),
    [anon_sym_LBRACK] = ACTIONS(2892),
    [anon_sym_LT_LT] = ACTIONS(2892),
    [anon_sym_SLASH_SLASH] = ACTIONS(2892),
    [anon_sym_PLUS] = ACTIONS(2892),
    [anon_sym_DASH] = ACTIONS(2892),
    [anon_sym_TILDE] = ACTIONS(2892),
    [anon_sym_BANG] = ACTIONS(2892),
    [anon_sym_PLUS_PLUS] = ACTIONS(2892),
    [anon_sym_DASH_DASH] = ACTIONS(2892),
    [anon_sym_BSLASH] = ACTIONS(2890),
    [anon_sym_do] = ACTIONS(2892),
    [anon_sym_eval] = ACTIONS(2892),
    [anon_sym_await] = ACTIONS(2892),
    [anon_sym_field] = ACTIONS(2892),
    [anon_sym_local] = ACTIONS(2892),
    [anon_sym_dynamically] = ACTIONS(2892),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2892),
    [anon_sym_require] = ACTIONS(2892),
    [anon_sym_map] = ACTIONS(2892),
    [anon_sym_grep] = ACTIONS(2892),
    [anon_sym_sort] = ACTIONS(2892),
    [anon_sym_goto] = ACTIONS(2892),
    [anon_sym_return] = ACTIONS(2892),
    [anon_sym_undef] = ACTIONS(2892),
    [anon_sym_PERCENT2] = ACTIONS(2892),
    [anon_sym_AMP2] = ACTIONS(2892),
    [anon_sym_STAR2] = ACTIONS(2892),
    [anon_sym_use] = ACTIONS(2892),
    [anon_sym_no] = ACTIONS(2892),
    [anon_sym_for] = ACTIONS(2892),
    [anon_sym_foreach] = ACTIONS(2892),
    [anon_sym_last] = ACTIONS(2892),
    [anon_sym_next] = ACTIONS(2892),
    [anon_sym_redo] = ACTIONS(2892),
    [anon_sym_BEGIN] = ACTIONS(2892),
    [anon_sym_INIT] = ACTIONS(2892),
    [anon_sym_CHECK] = ACTIONS(2892),
    [anon_sym_UNITCHECK] = ACTIONS(2892),
    [anon_sym_END] = ACTIONS(2892),
    [anon_sym___FILE__] = ACTIONS(2892),
    [anon_sym___LINE__] = ACTIONS(2892),
    [anon_sym___PACKAGE__] = ACTIONS(2892),
    [anon_sym___SUB__] = ACTIONS(2892),
    [anon_sym_break] = ACTIONS(2892),
    [anon_sym_fork] = ACTIONS(2892),
    [anon_sym_getppid] = ACTIONS(2892),
    [anon_sym_time] = ACTIONS(2892),
    [anon_sym_times] = ACTIONS(2892),
    [anon_sym_wait] = ACTIONS(2892),
    [anon_sym_wantarray] = ACTIONS(2892),
    [anon_sym_abs] = ACTIONS(2892),
    [anon_sym_alarm] = ACTIONS(2892),
    [anon_sym_chop] = ACTIONS(2892),
    [anon_sym_chdir] = ACTIONS(2892),
    [anon_sym_close] = ACTIONS(2892),
    [anon_sym_closedir] = ACTIONS(2892),
    [anon_sym_caller] = ACTIONS(2892),
    [anon_sym_chomp] = ACTIONS(2892),
    [anon_sym_chr] = ACTIONS(2892),
    [anon_sym_cos] = ACTIONS(2892),
    [anon_sym_chroot] = ACTIONS(2892),
    [anon_sym_defined] = ACTIONS(2892),
    [anon_sym_delete] = ACTIONS(2892),
    [anon_sym_dbmclose] = ACTIONS(2892),
    [anon_sym_exists] = ACTIONS(2892),
    [anon_sym_exit] = ACTIONS(2892),
    [anon_sym_eof] = ACTIONS(2892),
    [anon_sym_exp] = ACTIONS(2892),
    [anon_sym_each] = ACTIONS(2892),
    [anon_sym_fc] = ACTIONS(2892),
    [anon_sym_fileno] = ACTIONS(2892),
    [anon_sym_gmtime] = ACTIONS(2892),
    [anon_sym_getc] = ACTIONS(2892),
    [anon_sym_getpgrp] = ACTIONS(2892),
    [anon_sym_getprotobyname] = ACTIONS(2892),
    [anon_sym_getpwname] = ACTIONS(2892),
    [anon_sym_getpwuid] = ACTIONS(2892),
    [anon_sym_getpeername] = ACTIONS(2892),
    [anon_sym_getnetbyname] = ACTIONS(2892),
    [anon_sym_getsockname] = ACTIONS(2892),
    [anon_sym_getgrnam] = ACTIONS(2892),
    [anon_sym_getgrgid] = ACTIONS(2892),
    [anon_sym_hex] = ACTIONS(2892),
    [anon_sym_int] = ACTIONS(2892),
    [anon_sym_keys] = ACTIONS(2892),
    [anon_sym_lc] = ACTIONS(2892),
    [anon_sym_lcfirst] = ACTIONS(2892),
    [anon_sym_length] = ACTIONS(2892),
    [anon_sym_localtime] = ACTIONS(2892),
    [anon_sym_log] = ACTIONS(2892),
    [anon_sym_lock] = ACTIONS(2892),
    [anon_sym_lstat] = ACTIONS(2892),
    [anon_sym_oct] = ACTIONS(2892),
    [anon_sym_ord] = ACTIONS(2892),
    [anon_sym_prototype] = ACTIONS(2892),
    [anon_sym_pop] = ACTIONS(2892),
    [anon_sym_pos] = ACTIONS(2892),
    [anon_sym_quotemeta] = ACTIONS(2892),
    [anon_sym_reset] = ACTIONS(2892),
    [anon_sym_rand] = ACTIONS(2892),
    [anon_sym_rmdir] = ACTIONS(2892),
    [anon_sym_readdir] = ACTIONS(2892),
    [anon_sym_readline] = ACTIONS(2892),
    [anon_sym_readpipe] = ACTIONS(2892),
    [anon_sym_rewinddir] = ACTIONS(2892),
    [anon_sym_readlink] = ACTIONS(2892),
    [anon_sym_ref] = ACTIONS(2892),
    [anon_sym_scalar] = ACTIONS(2892),
    [anon_sym_shift] = ACTIONS(2892),
    [anon_sym_sin] = ACTIONS(2892),
    [anon_sym_sleep] = ACTIONS(2892),
    [anon_sym_sqrt] = ACTIONS(2892),
    [anon_sym_srand] = ACTIONS(2892),
    [anon_sym_stat] = ACTIONS(2892),
    [anon_sym_study] = ACTIONS(2892),
    [anon_sym_tell] = ACTIONS(2892),
    [anon_sym_telldir] = ACTIONS(2892),
    [anon_sym_tied] = ACTIONS(2892),
    [anon_sym_uc] = ACTIONS(2892),
    [anon_sym_ucfirst] = ACTIONS(2892),
    [anon_sym_untie] = ACTIONS(2892),
    [anon_sym_umask] = ACTIONS(2892),
    [anon_sym_values] = ACTIONS(2892),
    [anon_sym_write] = ACTIONS(2892),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2892),
    [anon_sym_qq] = ACTIONS(2892),
    [anon_sym_qw] = ACTIONS(2892),
    [anon_sym_qx] = ACTIONS(2892),
    [anon_sym_qr] = ACTIONS(2892),
    [anon_sym_m] = ACTIONS(2892),
    [anon_sym_s] = ACTIONS(2892),
    [anon_sym_tr] = ACTIONS(2892),
    [anon_sym_y] = ACTIONS(2892),
    [anon_sym_if] = ACTIONS(2892),
    [anon_sym_unless] = ACTIONS(2892),
    [anon_sym_while] = ACTIONS(2892),
    [anon_sym_until] = ACTIONS(2892),
    [aux_sym__bareword_token1] = ACTIONS(2892),
    [sym_number] = ACTIONS(2892),
    [anon_sym_true] = ACTIONS(2892),
    [anon_sym_false] = ACTIONS(2892),
    [sym__single_quote] = ACTIONS(2890),
    [sym__double_quote] = ACTIONS(2890),
    [sym__backtick_quote] = ACTIONS(2890),
    [sym__search_slash_quote] = ACTIONS(2890),
    [sym__open_readline_bracket] = ACTIONS(2890),
    [sym__open_fileglob_bracket] = ACTIONS(2890),
    [sym__PERLY_HEREDOC] = ACTIONS(2890),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2890),
    [sym__filetest] = ACTIONS(2890),
  },
  [STATE(1537)] = {
    [sym_heredoc_content] = STATE(1537),
    [sym__identifier] = ACTIONS(3033),
    [anon_sym_LBRACE] = ACTIONS(3033),
    [anon_sym_RBRACE] = ACTIONS(3033),
    [anon_sym_SEMI] = ACTIONS(3033),
    [anon_sym_package] = ACTIONS(3033),
    [anon_sym_class] = ACTIONS(3033),
    [anon_sym_role] = ACTIONS(3033),
    [anon_sym_BUILD] = ACTIONS(3033),
    [anon_sym_ADJUST] = ACTIONS(3033),
    [anon_sym_DOLLAR] = ACTIONS(3033),
    [anon_sym_AT] = ACTIONS(3033),
    [anon_sym_my] = ACTIONS(3033),
    [anon_sym_extended] = ACTIONS(3033),
    [anon_sym_async] = ACTIONS(3033),
    [anon_sym_sub] = ACTIONS(3033),
    [anon_sym_method] = ACTIONS(3033),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_continue] = ACTIONS(3033),
    [anon_sym_state] = ACTIONS(3033),
    [anon_sym_our] = ACTIONS(3033),
    [anon_sym_try] = ACTIONS(3033),
    [anon_sym_defer] = ACTIONS(3033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3033),
    [anon_sym_LBRACK] = ACTIONS(3033),
    [anon_sym_LT_LT] = ACTIONS(3033),
    [anon_sym_SLASH_SLASH] = ACTIONS(3033),
    [anon_sym_PLUS] = ACTIONS(3033),
    [anon_sym_DASH] = ACTIONS(3033),
    [anon_sym_TILDE] = ACTIONS(3033),
    [anon_sym_BANG] = ACTIONS(3033),
    [anon_sym_PLUS_PLUS] = ACTIONS(3033),
    [anon_sym_DASH_DASH] = ACTIONS(3033),
    [anon_sym_BSLASH] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3033),
    [anon_sym_eval] = ACTIONS(3033),
    [anon_sym_await] = ACTIONS(3033),
    [anon_sym_field] = ACTIONS(3033),
    [anon_sym_local] = ACTIONS(3033),
    [anon_sym_dynamically] = ACTIONS(3033),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3033),
    [anon_sym_require] = ACTIONS(3033),
    [anon_sym_map] = ACTIONS(3033),
    [anon_sym_grep] = ACTIONS(3033),
    [anon_sym_sort] = ACTIONS(3033),
    [anon_sym_goto] = ACTIONS(3033),
    [anon_sym_return] = ACTIONS(3033),
    [anon_sym_undef] = ACTIONS(3033),
    [anon_sym_PERCENT2] = ACTIONS(3033),
    [anon_sym_AMP2] = ACTIONS(3033),
    [anon_sym_STAR2] = ACTIONS(3033),
    [anon_sym_use] = ACTIONS(3033),
    [anon_sym_no] = ACTIONS(3033),
    [anon_sym_for] = ACTIONS(3033),
    [anon_sym_foreach] = ACTIONS(3033),
    [anon_sym_last] = ACTIONS(3033),
    [anon_sym_next] = ACTIONS(3033),
    [anon_sym_redo] = ACTIONS(3033),
    [anon_sym_BEGIN] = ACTIONS(3033),
    [anon_sym_INIT] = ACTIONS(3033),
    [anon_sym_CHECK] = ACTIONS(3033),
    [anon_sym_UNITCHECK] = ACTIONS(3033),
    [anon_sym_END] = ACTIONS(3033),
    [anon_sym___FILE__] = ACTIONS(3033),
    [anon_sym___LINE__] = ACTIONS(3033),
    [anon_sym___PACKAGE__] = ACTIONS(3033),
    [anon_sym___SUB__] = ACTIONS(3033),
    [anon_sym_break] = ACTIONS(3033),
    [anon_sym_fork] = ACTIONS(3033),
    [anon_sym_getppid] = ACTIONS(3033),
    [anon_sym_time] = ACTIONS(3033),
    [anon_sym_times] = ACTIONS(3033),
    [anon_sym_wait] = ACTIONS(3033),
    [anon_sym_wantarray] = ACTIONS(3033),
    [anon_sym_abs] = ACTIONS(3033),
    [anon_sym_alarm] = ACTIONS(3033),
    [anon_sym_chop] = ACTIONS(3033),
    [anon_sym_chdir] = ACTIONS(3033),
    [anon_sym_close] = ACTIONS(3033),
    [anon_sym_closedir] = ACTIONS(3033),
    [anon_sym_caller] = ACTIONS(3033),
    [anon_sym_chomp] = ACTIONS(3033),
    [anon_sym_chr] = ACTIONS(3033),
    [anon_sym_cos] = ACTIONS(3033),
    [anon_sym_chroot] = ACTIONS(3033),
    [anon_sym_defined] = ACTIONS(3033),
    [anon_sym_delete] = ACTIONS(3033),
    [anon_sym_dbmclose] = ACTIONS(3033),
    [anon_sym_exists] = ACTIONS(3033),
    [anon_sym_exit] = ACTIONS(3033),
    [anon_sym_eof] = ACTIONS(3033),
    [anon_sym_exp] = ACTIONS(3033),
    [anon_sym_each] = ACTIONS(3033),
    [anon_sym_fc] = ACTIONS(3033),
    [anon_sym_fileno] = ACTIONS(3033),
    [anon_sym_gmtime] = ACTIONS(3033),
    [anon_sym_getc] = ACTIONS(3033),
    [anon_sym_getpgrp] = ACTIONS(3033),
    [anon_sym_getprotobyname] = ACTIONS(3033),
    [anon_sym_getpwname] = ACTIONS(3033),
    [anon_sym_getpwuid] = ACTIONS(3033),
    [anon_sym_getpeername] = ACTIONS(3033),
    [anon_sym_getnetbyname] = ACTIONS(3033),
    [anon_sym_getsockname] = ACTIONS(3033),
    [anon_sym_getgrnam] = ACTIONS(3033),
    [anon_sym_getgrgid] = ACTIONS(3033),
    [anon_sym_hex] = ACTIONS(3033),
    [anon_sym_int] = ACTIONS(3033),
    [anon_sym_keys] = ACTIONS(3033),
    [anon_sym_lc] = ACTIONS(3033),
    [anon_sym_lcfirst] = ACTIONS(3033),
    [anon_sym_length] = ACTIONS(3033),
    [anon_sym_localtime] = ACTIONS(3033),
    [anon_sym_log] = ACTIONS(3033),
    [anon_sym_lock] = ACTIONS(3033),
    [anon_sym_lstat] = ACTIONS(3033),
    [anon_sym_oct] = ACTIONS(3033),
    [anon_sym_ord] = ACTIONS(3033),
    [anon_sym_prototype] = ACTIONS(3033),
    [anon_sym_pop] = ACTIONS(3033),
    [anon_sym_pos] = ACTIONS(3033),
    [anon_sym_quotemeta] = ACTIONS(3033),
    [anon_sym_reset] = ACTIONS(3033),
    [anon_sym_rand] = ACTIONS(3033),
    [anon_sym_rmdir] = ACTIONS(3033),
    [anon_sym_readdir] = ACTIONS(3033),
    [anon_sym_readline] = ACTIONS(3033),
    [anon_sym_readpipe] = ACTIONS(3033),
    [anon_sym_rewinddir] = ACTIONS(3033),
    [anon_sym_readlink] = ACTIONS(3033),
    [anon_sym_ref] = ACTIONS(3033),
    [anon_sym_scalar] = ACTIONS(3033),
    [anon_sym_shift] = ACTIONS(3033),
    [anon_sym_sin] = ACTIONS(3033),
    [anon_sym_sleep] = ACTIONS(3033),
    [anon_sym_sqrt] = ACTIONS(3033),
    [anon_sym_srand] = ACTIONS(3033),
    [anon_sym_stat] = ACTIONS(3033),
    [anon_sym_study] = ACTIONS(3033),
    [anon_sym_tell] = ACTIONS(3033),
    [anon_sym_telldir] = ACTIONS(3033),
    [anon_sym_tied] = ACTIONS(3033),
    [anon_sym_uc] = ACTIONS(3033),
    [anon_sym_ucfirst] = ACTIONS(3033),
    [anon_sym_untie] = ACTIONS(3033),
    [anon_sym_umask] = ACTIONS(3033),
    [anon_sym_values] = ACTIONS(3033),
    [anon_sym_write] = ACTIONS(3033),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3033),
    [anon_sym_qq] = ACTIONS(3033),
    [anon_sym_qw] = ACTIONS(3033),
    [anon_sym_qx] = ACTIONS(3033),
    [anon_sym_qr] = ACTIONS(3033),
    [anon_sym_m] = ACTIONS(3033),
    [anon_sym_s] = ACTIONS(3033),
    [anon_sym_tr] = ACTIONS(3033),
    [anon_sym_y] = ACTIONS(3033),
    [anon_sym_if] = ACTIONS(3033),
    [anon_sym_unless] = ACTIONS(3033),
    [anon_sym_while] = ACTIONS(3033),
    [anon_sym_until] = ACTIONS(3033),
    [aux_sym__bareword_token1] = ACTIONS(3033),
    [sym_number] = ACTIONS(3033),
    [anon_sym_true] = ACTIONS(3033),
    [anon_sym_false] = ACTIONS(3033),
    [sym__single_quote] = ACTIONS(3031),
    [sym__double_quote] = ACTIONS(3031),
    [sym__backtick_quote] = ACTIONS(3031),
    [sym__search_slash_quote] = ACTIONS(3031),
    [sym__open_readline_bracket] = ACTIONS(3031),
    [sym__open_fileglob_bracket] = ACTIONS(3031),
    [sym__PERLY_HEREDOC] = ACTIONS(3031),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3031),
    [sym__filetest] = ACTIONS(3031),
  },
  [STATE(1538)] = {
    [sym_heredoc_content] = STATE(1538),
    [sym__identifier] = ACTIONS(3169),
    [anon_sym_LBRACE] = ACTIONS(3169),
    [anon_sym_RBRACE] = ACTIONS(3169),
    [anon_sym_SEMI] = ACTIONS(3169),
    [anon_sym_package] = ACTIONS(3169),
    [anon_sym_class] = ACTIONS(3169),
    [anon_sym_role] = ACTIONS(3169),
    [anon_sym_BUILD] = ACTIONS(3169),
    [anon_sym_ADJUST] = ACTIONS(3169),
    [anon_sym_DOLLAR] = ACTIONS(3169),
    [anon_sym_AT] = ACTIONS(3169),
    [anon_sym_my] = ACTIONS(3169),
    [anon_sym_extended] = ACTIONS(3169),
    [anon_sym_async] = ACTIONS(3169),
    [anon_sym_sub] = ACTIONS(3169),
    [anon_sym_method] = ACTIONS(3169),
    [anon_sym_LPAREN] = ACTIONS(3169),
    [anon_sym_continue] = ACTIONS(3169),
    [anon_sym_state] = ACTIONS(3169),
    [anon_sym_our] = ACTIONS(3169),
    [anon_sym_try] = ACTIONS(3169),
    [anon_sym_defer] = ACTIONS(3169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3169),
    [anon_sym_LBRACK] = ACTIONS(3169),
    [anon_sym_LT_LT] = ACTIONS(3169),
    [anon_sym_SLASH_SLASH] = ACTIONS(3169),
    [anon_sym_PLUS] = ACTIONS(3169),
    [anon_sym_DASH] = ACTIONS(3169),
    [anon_sym_TILDE] = ACTIONS(3169),
    [anon_sym_BANG] = ACTIONS(3169),
    [anon_sym_PLUS_PLUS] = ACTIONS(3169),
    [anon_sym_DASH_DASH] = ACTIONS(3169),
    [anon_sym_BSLASH] = ACTIONS(3167),
    [anon_sym_do] = ACTIONS(3169),
    [anon_sym_eval] = ACTIONS(3169),
    [anon_sym_await] = ACTIONS(3169),
    [anon_sym_field] = ACTIONS(3169),
    [anon_sym_local] = ACTIONS(3169),
    [anon_sym_dynamically] = ACTIONS(3169),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3169),
    [anon_sym_require] = ACTIONS(3169),
    [anon_sym_map] = ACTIONS(3169),
    [anon_sym_grep] = ACTIONS(3169),
    [anon_sym_sort] = ACTIONS(3169),
    [anon_sym_goto] = ACTIONS(3169),
    [anon_sym_return] = ACTIONS(3169),
    [anon_sym_undef] = ACTIONS(3169),
    [anon_sym_PERCENT2] = ACTIONS(3169),
    [anon_sym_AMP2] = ACTIONS(3169),
    [anon_sym_STAR2] = ACTIONS(3169),
    [anon_sym_use] = ACTIONS(3169),
    [anon_sym_no] = ACTIONS(3169),
    [anon_sym_for] = ACTIONS(3169),
    [anon_sym_foreach] = ACTIONS(3169),
    [anon_sym_last] = ACTIONS(3169),
    [anon_sym_next] = ACTIONS(3169),
    [anon_sym_redo] = ACTIONS(3169),
    [anon_sym_BEGIN] = ACTIONS(3169),
    [anon_sym_INIT] = ACTIONS(3169),
    [anon_sym_CHECK] = ACTIONS(3169),
    [anon_sym_UNITCHECK] = ACTIONS(3169),
    [anon_sym_END] = ACTIONS(3169),
    [anon_sym___FILE__] = ACTIONS(3169),
    [anon_sym___LINE__] = ACTIONS(3169),
    [anon_sym___PACKAGE__] = ACTIONS(3169),
    [anon_sym___SUB__] = ACTIONS(3169),
    [anon_sym_break] = ACTIONS(3169),
    [anon_sym_fork] = ACTIONS(3169),
    [anon_sym_getppid] = ACTIONS(3169),
    [anon_sym_time] = ACTIONS(3169),
    [anon_sym_times] = ACTIONS(3169),
    [anon_sym_wait] = ACTIONS(3169),
    [anon_sym_wantarray] = ACTIONS(3169),
    [anon_sym_abs] = ACTIONS(3169),
    [anon_sym_alarm] = ACTIONS(3169),
    [anon_sym_chop] = ACTIONS(3169),
    [anon_sym_chdir] = ACTIONS(3169),
    [anon_sym_close] = ACTIONS(3169),
    [anon_sym_closedir] = ACTIONS(3169),
    [anon_sym_caller] = ACTIONS(3169),
    [anon_sym_chomp] = ACTIONS(3169),
    [anon_sym_chr] = ACTIONS(3169),
    [anon_sym_cos] = ACTIONS(3169),
    [anon_sym_chroot] = ACTIONS(3169),
    [anon_sym_defined] = ACTIONS(3169),
    [anon_sym_delete] = ACTIONS(3169),
    [anon_sym_dbmclose] = ACTIONS(3169),
    [anon_sym_exists] = ACTIONS(3169),
    [anon_sym_exit] = ACTIONS(3169),
    [anon_sym_eof] = ACTIONS(3169),
    [anon_sym_exp] = ACTIONS(3169),
    [anon_sym_each] = ACTIONS(3169),
    [anon_sym_fc] = ACTIONS(3169),
    [anon_sym_fileno] = ACTIONS(3169),
    [anon_sym_gmtime] = ACTIONS(3169),
    [anon_sym_getc] = ACTIONS(3169),
    [anon_sym_getpgrp] = ACTIONS(3169),
    [anon_sym_getprotobyname] = ACTIONS(3169),
    [anon_sym_getpwname] = ACTIONS(3169),
    [anon_sym_getpwuid] = ACTIONS(3169),
    [anon_sym_getpeername] = ACTIONS(3169),
    [anon_sym_getnetbyname] = ACTIONS(3169),
    [anon_sym_getsockname] = ACTIONS(3169),
    [anon_sym_getgrnam] = ACTIONS(3169),
    [anon_sym_getgrgid] = ACTIONS(3169),
    [anon_sym_hex] = ACTIONS(3169),
    [anon_sym_int] = ACTIONS(3169),
    [anon_sym_keys] = ACTIONS(3169),
    [anon_sym_lc] = ACTIONS(3169),
    [anon_sym_lcfirst] = ACTIONS(3169),
    [anon_sym_length] = ACTIONS(3169),
    [anon_sym_localtime] = ACTIONS(3169),
    [anon_sym_log] = ACTIONS(3169),
    [anon_sym_lock] = ACTIONS(3169),
    [anon_sym_lstat] = ACTIONS(3169),
    [anon_sym_oct] = ACTIONS(3169),
    [anon_sym_ord] = ACTIONS(3169),
    [anon_sym_prototype] = ACTIONS(3169),
    [anon_sym_pop] = ACTIONS(3169),
    [anon_sym_pos] = ACTIONS(3169),
    [anon_sym_quotemeta] = ACTIONS(3169),
    [anon_sym_reset] = ACTIONS(3169),
    [anon_sym_rand] = ACTIONS(3169),
    [anon_sym_rmdir] = ACTIONS(3169),
    [anon_sym_readdir] = ACTIONS(3169),
    [anon_sym_readline] = ACTIONS(3169),
    [anon_sym_readpipe] = ACTIONS(3169),
    [anon_sym_rewinddir] = ACTIONS(3169),
    [anon_sym_readlink] = ACTIONS(3169),
    [anon_sym_ref] = ACTIONS(3169),
    [anon_sym_scalar] = ACTIONS(3169),
    [anon_sym_shift] = ACTIONS(3169),
    [anon_sym_sin] = ACTIONS(3169),
    [anon_sym_sleep] = ACTIONS(3169),
    [anon_sym_sqrt] = ACTIONS(3169),
    [anon_sym_srand] = ACTIONS(3169),
    [anon_sym_stat] = ACTIONS(3169),
    [anon_sym_study] = ACTIONS(3169),
    [anon_sym_tell] = ACTIONS(3169),
    [anon_sym_telldir] = ACTIONS(3169),
    [anon_sym_tied] = ACTIONS(3169),
    [anon_sym_uc] = ACTIONS(3169),
    [anon_sym_ucfirst] = ACTIONS(3169),
    [anon_sym_untie] = ACTIONS(3169),
    [anon_sym_umask] = ACTIONS(3169),
    [anon_sym_values] = ACTIONS(3169),
    [anon_sym_write] = ACTIONS(3169),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3169),
    [anon_sym_qq] = ACTIONS(3169),
    [anon_sym_qw] = ACTIONS(3169),
    [anon_sym_qx] = ACTIONS(3169),
    [anon_sym_qr] = ACTIONS(3169),
    [anon_sym_m] = ACTIONS(3169),
    [anon_sym_s] = ACTIONS(3169),
    [anon_sym_tr] = ACTIONS(3169),
    [anon_sym_y] = ACTIONS(3169),
    [anon_sym_if] = ACTIONS(3169),
    [anon_sym_unless] = ACTIONS(3169),
    [anon_sym_while] = ACTIONS(3169),
    [anon_sym_until] = ACTIONS(3169),
    [aux_sym__bareword_token1] = ACTIONS(3169),
    [sym_number] = ACTIONS(3169),
    [anon_sym_true] = ACTIONS(3169),
    [anon_sym_false] = ACTIONS(3169),
    [sym__single_quote] = ACTIONS(3167),
    [sym__double_quote] = ACTIONS(3167),
    [sym__backtick_quote] = ACTIONS(3167),
    [sym__search_slash_quote] = ACTIONS(3167),
    [sym__open_readline_bracket] = ACTIONS(3167),
    [sym__open_fileglob_bracket] = ACTIONS(3167),
    [sym__PERLY_HEREDOC] = ACTIONS(3167),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3167),
    [sym__filetest] = ACTIONS(3167),
  },
  [STATE(1539)] = {
    [sym_heredoc_content] = STATE(1539),
    [sym__identifier] = ACTIONS(3189),
    [anon_sym_LBRACE] = ACTIONS(3189),
    [anon_sym_RBRACE] = ACTIONS(3189),
    [anon_sym_SEMI] = ACTIONS(3189),
    [anon_sym_package] = ACTIONS(3189),
    [anon_sym_class] = ACTIONS(3189),
    [anon_sym_role] = ACTIONS(3189),
    [anon_sym_BUILD] = ACTIONS(3189),
    [anon_sym_ADJUST] = ACTIONS(3189),
    [anon_sym_DOLLAR] = ACTIONS(3189),
    [anon_sym_AT] = ACTIONS(3189),
    [anon_sym_my] = ACTIONS(3189),
    [anon_sym_extended] = ACTIONS(3189),
    [anon_sym_async] = ACTIONS(3189),
    [anon_sym_sub] = ACTIONS(3189),
    [anon_sym_method] = ACTIONS(3189),
    [anon_sym_LPAREN] = ACTIONS(3189),
    [anon_sym_continue] = ACTIONS(3189),
    [anon_sym_state] = ACTIONS(3189),
    [anon_sym_our] = ACTIONS(3189),
    [anon_sym_try] = ACTIONS(3189),
    [anon_sym_defer] = ACTIONS(3189),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3189),
    [anon_sym_LBRACK] = ACTIONS(3189),
    [anon_sym_LT_LT] = ACTIONS(3189),
    [anon_sym_SLASH_SLASH] = ACTIONS(3189),
    [anon_sym_PLUS] = ACTIONS(3189),
    [anon_sym_DASH] = ACTIONS(3189),
    [anon_sym_TILDE] = ACTIONS(3189),
    [anon_sym_BANG] = ACTIONS(3189),
    [anon_sym_PLUS_PLUS] = ACTIONS(3189),
    [anon_sym_DASH_DASH] = ACTIONS(3189),
    [anon_sym_BSLASH] = ACTIONS(3187),
    [anon_sym_do] = ACTIONS(3189),
    [anon_sym_eval] = ACTIONS(3189),
    [anon_sym_await] = ACTIONS(3189),
    [anon_sym_field] = ACTIONS(3189),
    [anon_sym_local] = ACTIONS(3189),
    [anon_sym_dynamically] = ACTIONS(3189),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3189),
    [anon_sym_require] = ACTIONS(3189),
    [anon_sym_map] = ACTIONS(3189),
    [anon_sym_grep] = ACTIONS(3189),
    [anon_sym_sort] = ACTIONS(3189),
    [anon_sym_goto] = ACTIONS(3189),
    [anon_sym_return] = ACTIONS(3189),
    [anon_sym_undef] = ACTIONS(3189),
    [anon_sym_PERCENT2] = ACTIONS(3189),
    [anon_sym_AMP2] = ACTIONS(3189),
    [anon_sym_STAR2] = ACTIONS(3189),
    [anon_sym_use] = ACTIONS(3189),
    [anon_sym_no] = ACTIONS(3189),
    [anon_sym_for] = ACTIONS(3189),
    [anon_sym_foreach] = ACTIONS(3189),
    [anon_sym_last] = ACTIONS(3189),
    [anon_sym_next] = ACTIONS(3189),
    [anon_sym_redo] = ACTIONS(3189),
    [anon_sym_BEGIN] = ACTIONS(3189),
    [anon_sym_INIT] = ACTIONS(3189),
    [anon_sym_CHECK] = ACTIONS(3189),
    [anon_sym_UNITCHECK] = ACTIONS(3189),
    [anon_sym_END] = ACTIONS(3189),
    [anon_sym___FILE__] = ACTIONS(3189),
    [anon_sym___LINE__] = ACTIONS(3189),
    [anon_sym___PACKAGE__] = ACTIONS(3189),
    [anon_sym___SUB__] = ACTIONS(3189),
    [anon_sym_break] = ACTIONS(3189),
    [anon_sym_fork] = ACTIONS(3189),
    [anon_sym_getppid] = ACTIONS(3189),
    [anon_sym_time] = ACTIONS(3189),
    [anon_sym_times] = ACTIONS(3189),
    [anon_sym_wait] = ACTIONS(3189),
    [anon_sym_wantarray] = ACTIONS(3189),
    [anon_sym_abs] = ACTIONS(3189),
    [anon_sym_alarm] = ACTIONS(3189),
    [anon_sym_chop] = ACTIONS(3189),
    [anon_sym_chdir] = ACTIONS(3189),
    [anon_sym_close] = ACTIONS(3189),
    [anon_sym_closedir] = ACTIONS(3189),
    [anon_sym_caller] = ACTIONS(3189),
    [anon_sym_chomp] = ACTIONS(3189),
    [anon_sym_chr] = ACTIONS(3189),
    [anon_sym_cos] = ACTIONS(3189),
    [anon_sym_chroot] = ACTIONS(3189),
    [anon_sym_defined] = ACTIONS(3189),
    [anon_sym_delete] = ACTIONS(3189),
    [anon_sym_dbmclose] = ACTIONS(3189),
    [anon_sym_exists] = ACTIONS(3189),
    [anon_sym_exit] = ACTIONS(3189),
    [anon_sym_eof] = ACTIONS(3189),
    [anon_sym_exp] = ACTIONS(3189),
    [anon_sym_each] = ACTIONS(3189),
    [anon_sym_fc] = ACTIONS(3189),
    [anon_sym_fileno] = ACTIONS(3189),
    [anon_sym_gmtime] = ACTIONS(3189),
    [anon_sym_getc] = ACTIONS(3189),
    [anon_sym_getpgrp] = ACTIONS(3189),
    [anon_sym_getprotobyname] = ACTIONS(3189),
    [anon_sym_getpwname] = ACTIONS(3189),
    [anon_sym_getpwuid] = ACTIONS(3189),
    [anon_sym_getpeername] = ACTIONS(3189),
    [anon_sym_getnetbyname] = ACTIONS(3189),
    [anon_sym_getsockname] = ACTIONS(3189),
    [anon_sym_getgrnam] = ACTIONS(3189),
    [anon_sym_getgrgid] = ACTIONS(3189),
    [anon_sym_hex] = ACTIONS(3189),
    [anon_sym_int] = ACTIONS(3189),
    [anon_sym_keys] = ACTIONS(3189),
    [anon_sym_lc] = ACTIONS(3189),
    [anon_sym_lcfirst] = ACTIONS(3189),
    [anon_sym_length] = ACTIONS(3189),
    [anon_sym_localtime] = ACTIONS(3189),
    [anon_sym_log] = ACTIONS(3189),
    [anon_sym_lock] = ACTIONS(3189),
    [anon_sym_lstat] = ACTIONS(3189),
    [anon_sym_oct] = ACTIONS(3189),
    [anon_sym_ord] = ACTIONS(3189),
    [anon_sym_prototype] = ACTIONS(3189),
    [anon_sym_pop] = ACTIONS(3189),
    [anon_sym_pos] = ACTIONS(3189),
    [anon_sym_quotemeta] = ACTIONS(3189),
    [anon_sym_reset] = ACTIONS(3189),
    [anon_sym_rand] = ACTIONS(3189),
    [anon_sym_rmdir] = ACTIONS(3189),
    [anon_sym_readdir] = ACTIONS(3189),
    [anon_sym_readline] = ACTIONS(3189),
    [anon_sym_readpipe] = ACTIONS(3189),
    [anon_sym_rewinddir] = ACTIONS(3189),
    [anon_sym_readlink] = ACTIONS(3189),
    [anon_sym_ref] = ACTIONS(3189),
    [anon_sym_scalar] = ACTIONS(3189),
    [anon_sym_shift] = ACTIONS(3189),
    [anon_sym_sin] = ACTIONS(3189),
    [anon_sym_sleep] = ACTIONS(3189),
    [anon_sym_sqrt] = ACTIONS(3189),
    [anon_sym_srand] = ACTIONS(3189),
    [anon_sym_stat] = ACTIONS(3189),
    [anon_sym_study] = ACTIONS(3189),
    [anon_sym_tell] = ACTIONS(3189),
    [anon_sym_telldir] = ACTIONS(3189),
    [anon_sym_tied] = ACTIONS(3189),
    [anon_sym_uc] = ACTIONS(3189),
    [anon_sym_ucfirst] = ACTIONS(3189),
    [anon_sym_untie] = ACTIONS(3189),
    [anon_sym_umask] = ACTIONS(3189),
    [anon_sym_values] = ACTIONS(3189),
    [anon_sym_write] = ACTIONS(3189),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3189),
    [anon_sym_qq] = ACTIONS(3189),
    [anon_sym_qw] = ACTIONS(3189),
    [anon_sym_qx] = ACTIONS(3189),
    [anon_sym_qr] = ACTIONS(3189),
    [anon_sym_m] = ACTIONS(3189),
    [anon_sym_s] = ACTIONS(3189),
    [anon_sym_tr] = ACTIONS(3189),
    [anon_sym_y] = ACTIONS(3189),
    [anon_sym_if] = ACTIONS(3189),
    [anon_sym_unless] = ACTIONS(3189),
    [anon_sym_while] = ACTIONS(3189),
    [anon_sym_until] = ACTIONS(3189),
    [aux_sym__bareword_token1] = ACTIONS(3189),
    [sym_number] = ACTIONS(3189),
    [anon_sym_true] = ACTIONS(3189),
    [anon_sym_false] = ACTIONS(3189),
    [sym__single_quote] = ACTIONS(3187),
    [sym__double_quote] = ACTIONS(3187),
    [sym__backtick_quote] = ACTIONS(3187),
    [sym__search_slash_quote] = ACTIONS(3187),
    [sym__open_readline_bracket] = ACTIONS(3187),
    [sym__open_fileglob_bracket] = ACTIONS(3187),
    [sym__PERLY_HEREDOC] = ACTIONS(3187),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3187),
    [sym__filetest] = ACTIONS(3187),
  },
  [STATE(1540)] = {
    [sym_heredoc_content] = STATE(1540),
    [sym__identifier] = ACTIONS(3225),
    [anon_sym_LBRACE] = ACTIONS(3225),
    [anon_sym_RBRACE] = ACTIONS(3225),
    [anon_sym_SEMI] = ACTIONS(3225),
    [anon_sym_package] = ACTIONS(3225),
    [anon_sym_class] = ACTIONS(3225),
    [anon_sym_role] = ACTIONS(3225),
    [anon_sym_BUILD] = ACTIONS(3225),
    [anon_sym_ADJUST] = ACTIONS(3225),
    [anon_sym_DOLLAR] = ACTIONS(3225),
    [anon_sym_AT] = ACTIONS(3225),
    [anon_sym_my] = ACTIONS(3225),
    [anon_sym_extended] = ACTIONS(3225),
    [anon_sym_async] = ACTIONS(3225),
    [anon_sym_sub] = ACTIONS(3225),
    [anon_sym_method] = ACTIONS(3225),
    [anon_sym_LPAREN] = ACTIONS(3225),
    [anon_sym_continue] = ACTIONS(3225),
    [anon_sym_state] = ACTIONS(3225),
    [anon_sym_our] = ACTIONS(3225),
    [anon_sym_try] = ACTIONS(3225),
    [anon_sym_defer] = ACTIONS(3225),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3225),
    [anon_sym_LBRACK] = ACTIONS(3225),
    [anon_sym_LT_LT] = ACTIONS(3225),
    [anon_sym_SLASH_SLASH] = ACTIONS(3225),
    [anon_sym_PLUS] = ACTIONS(3225),
    [anon_sym_DASH] = ACTIONS(3225),
    [anon_sym_TILDE] = ACTIONS(3225),
    [anon_sym_BANG] = ACTIONS(3225),
    [anon_sym_PLUS_PLUS] = ACTIONS(3225),
    [anon_sym_DASH_DASH] = ACTIONS(3225),
    [anon_sym_BSLASH] = ACTIONS(3223),
    [anon_sym_do] = ACTIONS(3225),
    [anon_sym_eval] = ACTIONS(3225),
    [anon_sym_await] = ACTIONS(3225),
    [anon_sym_field] = ACTIONS(3225),
    [anon_sym_local] = ACTIONS(3225),
    [anon_sym_dynamically] = ACTIONS(3225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3225),
    [anon_sym_require] = ACTIONS(3225),
    [anon_sym_map] = ACTIONS(3225),
    [anon_sym_grep] = ACTIONS(3225),
    [anon_sym_sort] = ACTIONS(3225),
    [anon_sym_goto] = ACTIONS(3225),
    [anon_sym_return] = ACTIONS(3225),
    [anon_sym_undef] = ACTIONS(3225),
    [anon_sym_PERCENT2] = ACTIONS(3225),
    [anon_sym_AMP2] = ACTIONS(3225),
    [anon_sym_STAR2] = ACTIONS(3225),
    [anon_sym_use] = ACTIONS(3225),
    [anon_sym_no] = ACTIONS(3225),
    [anon_sym_for] = ACTIONS(3225),
    [anon_sym_foreach] = ACTIONS(3225),
    [anon_sym_last] = ACTIONS(3225),
    [anon_sym_next] = ACTIONS(3225),
    [anon_sym_redo] = ACTIONS(3225),
    [anon_sym_BEGIN] = ACTIONS(3225),
    [anon_sym_INIT] = ACTIONS(3225),
    [anon_sym_CHECK] = ACTIONS(3225),
    [anon_sym_UNITCHECK] = ACTIONS(3225),
    [anon_sym_END] = ACTIONS(3225),
    [anon_sym___FILE__] = ACTIONS(3225),
    [anon_sym___LINE__] = ACTIONS(3225),
    [anon_sym___PACKAGE__] = ACTIONS(3225),
    [anon_sym___SUB__] = ACTIONS(3225),
    [anon_sym_break] = ACTIONS(3225),
    [anon_sym_fork] = ACTIONS(3225),
    [anon_sym_getppid] = ACTIONS(3225),
    [anon_sym_time] = ACTIONS(3225),
    [anon_sym_times] = ACTIONS(3225),
    [anon_sym_wait] = ACTIONS(3225),
    [anon_sym_wantarray] = ACTIONS(3225),
    [anon_sym_abs] = ACTIONS(3225),
    [anon_sym_alarm] = ACTIONS(3225),
    [anon_sym_chop] = ACTIONS(3225),
    [anon_sym_chdir] = ACTIONS(3225),
    [anon_sym_close] = ACTIONS(3225),
    [anon_sym_closedir] = ACTIONS(3225),
    [anon_sym_caller] = ACTIONS(3225),
    [anon_sym_chomp] = ACTIONS(3225),
    [anon_sym_chr] = ACTIONS(3225),
    [anon_sym_cos] = ACTIONS(3225),
    [anon_sym_chroot] = ACTIONS(3225),
    [anon_sym_defined] = ACTIONS(3225),
    [anon_sym_delete] = ACTIONS(3225),
    [anon_sym_dbmclose] = ACTIONS(3225),
    [anon_sym_exists] = ACTIONS(3225),
    [anon_sym_exit] = ACTIONS(3225),
    [anon_sym_eof] = ACTIONS(3225),
    [anon_sym_exp] = ACTIONS(3225),
    [anon_sym_each] = ACTIONS(3225),
    [anon_sym_fc] = ACTIONS(3225),
    [anon_sym_fileno] = ACTIONS(3225),
    [anon_sym_gmtime] = ACTIONS(3225),
    [anon_sym_getc] = ACTIONS(3225),
    [anon_sym_getpgrp] = ACTIONS(3225),
    [anon_sym_getprotobyname] = ACTIONS(3225),
    [anon_sym_getpwname] = ACTIONS(3225),
    [anon_sym_getpwuid] = ACTIONS(3225),
    [anon_sym_getpeername] = ACTIONS(3225),
    [anon_sym_getnetbyname] = ACTIONS(3225),
    [anon_sym_getsockname] = ACTIONS(3225),
    [anon_sym_getgrnam] = ACTIONS(3225),
    [anon_sym_getgrgid] = ACTIONS(3225),
    [anon_sym_hex] = ACTIONS(3225),
    [anon_sym_int] = ACTIONS(3225),
    [anon_sym_keys] = ACTIONS(3225),
    [anon_sym_lc] = ACTIONS(3225),
    [anon_sym_lcfirst] = ACTIONS(3225),
    [anon_sym_length] = ACTIONS(3225),
    [anon_sym_localtime] = ACTIONS(3225),
    [anon_sym_log] = ACTIONS(3225),
    [anon_sym_lock] = ACTIONS(3225),
    [anon_sym_lstat] = ACTIONS(3225),
    [anon_sym_oct] = ACTIONS(3225),
    [anon_sym_ord] = ACTIONS(3225),
    [anon_sym_prototype] = ACTIONS(3225),
    [anon_sym_pop] = ACTIONS(3225),
    [anon_sym_pos] = ACTIONS(3225),
    [anon_sym_quotemeta] = ACTIONS(3225),
    [anon_sym_reset] = ACTIONS(3225),
    [anon_sym_rand] = ACTIONS(3225),
    [anon_sym_rmdir] = ACTIONS(3225),
    [anon_sym_readdir] = ACTIONS(3225),
    [anon_sym_readline] = ACTIONS(3225),
    [anon_sym_readpipe] = ACTIONS(3225),
    [anon_sym_rewinddir] = ACTIONS(3225),
    [anon_sym_readlink] = ACTIONS(3225),
    [anon_sym_ref] = ACTIONS(3225),
    [anon_sym_scalar] = ACTIONS(3225),
    [anon_sym_shift] = ACTIONS(3225),
    [anon_sym_sin] = ACTIONS(3225),
    [anon_sym_sleep] = ACTIONS(3225),
    [anon_sym_sqrt] = ACTIONS(3225),
    [anon_sym_srand] = ACTIONS(3225),
    [anon_sym_stat] = ACTIONS(3225),
    [anon_sym_study] = ACTIONS(3225),
    [anon_sym_tell] = ACTIONS(3225),
    [anon_sym_telldir] = ACTIONS(3225),
    [anon_sym_tied] = ACTIONS(3225),
    [anon_sym_uc] = ACTIONS(3225),
    [anon_sym_ucfirst] = ACTIONS(3225),
    [anon_sym_untie] = ACTIONS(3225),
    [anon_sym_umask] = ACTIONS(3225),
    [anon_sym_values] = ACTIONS(3225),
    [anon_sym_write] = ACTIONS(3225),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3225),
    [anon_sym_qq] = ACTIONS(3225),
    [anon_sym_qw] = ACTIONS(3225),
    [anon_sym_qx] = ACTIONS(3225),
    [anon_sym_qr] = ACTIONS(3225),
    [anon_sym_m] = ACTIONS(3225),
    [anon_sym_s] = ACTIONS(3225),
    [anon_sym_tr] = ACTIONS(3225),
    [anon_sym_y] = ACTIONS(3225),
    [anon_sym_if] = ACTIONS(3225),
    [anon_sym_unless] = ACTIONS(3225),
    [anon_sym_while] = ACTIONS(3225),
    [anon_sym_until] = ACTIONS(3225),
    [aux_sym__bareword_token1] = ACTIONS(3225),
    [sym_number] = ACTIONS(3225),
    [anon_sym_true] = ACTIONS(3225),
    [anon_sym_false] = ACTIONS(3225),
    [sym__single_quote] = ACTIONS(3223),
    [sym__double_quote] = ACTIONS(3223),
    [sym__backtick_quote] = ACTIONS(3223),
    [sym__search_slash_quote] = ACTIONS(3223),
    [sym__open_readline_bracket] = ACTIONS(3223),
    [sym__open_fileglob_bracket] = ACTIONS(3223),
    [sym__PERLY_HEREDOC] = ACTIONS(3223),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3223),
    [sym__filetest] = ACTIONS(3223),
  },
  [STATE(1541)] = {
    [sym_heredoc_content] = STATE(1541),
    [sym__identifier] = ACTIONS(3081),
    [anon_sym_LBRACE] = ACTIONS(3081),
    [anon_sym_RBRACE] = ACTIONS(3081),
    [anon_sym_SEMI] = ACTIONS(3081),
    [anon_sym_package] = ACTIONS(3081),
    [anon_sym_class] = ACTIONS(3081),
    [anon_sym_role] = ACTIONS(3081),
    [anon_sym_BUILD] = ACTIONS(3081),
    [anon_sym_ADJUST] = ACTIONS(3081),
    [anon_sym_DOLLAR] = ACTIONS(3081),
    [anon_sym_AT] = ACTIONS(3081),
    [anon_sym_my] = ACTIONS(3081),
    [anon_sym_extended] = ACTIONS(3081),
    [anon_sym_async] = ACTIONS(3081),
    [anon_sym_sub] = ACTIONS(3081),
    [anon_sym_method] = ACTIONS(3081),
    [anon_sym_LPAREN] = ACTIONS(3081),
    [anon_sym_continue] = ACTIONS(3081),
    [anon_sym_state] = ACTIONS(3081),
    [anon_sym_our] = ACTIONS(3081),
    [anon_sym_try] = ACTIONS(3081),
    [anon_sym_defer] = ACTIONS(3081),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3081),
    [anon_sym_LBRACK] = ACTIONS(3081),
    [anon_sym_LT_LT] = ACTIONS(3081),
    [anon_sym_SLASH_SLASH] = ACTIONS(3081),
    [anon_sym_PLUS] = ACTIONS(3081),
    [anon_sym_DASH] = ACTIONS(3081),
    [anon_sym_TILDE] = ACTIONS(3081),
    [anon_sym_BANG] = ACTIONS(3081),
    [anon_sym_PLUS_PLUS] = ACTIONS(3081),
    [anon_sym_DASH_DASH] = ACTIONS(3081),
    [anon_sym_BSLASH] = ACTIONS(3079),
    [anon_sym_do] = ACTIONS(3081),
    [anon_sym_eval] = ACTIONS(3081),
    [anon_sym_await] = ACTIONS(3081),
    [anon_sym_field] = ACTIONS(3081),
    [anon_sym_local] = ACTIONS(3081),
    [anon_sym_dynamically] = ACTIONS(3081),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3081),
    [anon_sym_require] = ACTIONS(3081),
    [anon_sym_map] = ACTIONS(3081),
    [anon_sym_grep] = ACTIONS(3081),
    [anon_sym_sort] = ACTIONS(3081),
    [anon_sym_goto] = ACTIONS(3081),
    [anon_sym_return] = ACTIONS(3081),
    [anon_sym_undef] = ACTIONS(3081),
    [anon_sym_PERCENT2] = ACTIONS(3081),
    [anon_sym_AMP2] = ACTIONS(3081),
    [anon_sym_STAR2] = ACTIONS(3081),
    [anon_sym_use] = ACTIONS(3081),
    [anon_sym_no] = ACTIONS(3081),
    [anon_sym_for] = ACTIONS(3081),
    [anon_sym_foreach] = ACTIONS(3081),
    [anon_sym_last] = ACTIONS(3081),
    [anon_sym_next] = ACTIONS(3081),
    [anon_sym_redo] = ACTIONS(3081),
    [anon_sym_BEGIN] = ACTIONS(3081),
    [anon_sym_INIT] = ACTIONS(3081),
    [anon_sym_CHECK] = ACTIONS(3081),
    [anon_sym_UNITCHECK] = ACTIONS(3081),
    [anon_sym_END] = ACTIONS(3081),
    [anon_sym___FILE__] = ACTIONS(3081),
    [anon_sym___LINE__] = ACTIONS(3081),
    [anon_sym___PACKAGE__] = ACTIONS(3081),
    [anon_sym___SUB__] = ACTIONS(3081),
    [anon_sym_break] = ACTIONS(3081),
    [anon_sym_fork] = ACTIONS(3081),
    [anon_sym_getppid] = ACTIONS(3081),
    [anon_sym_time] = ACTIONS(3081),
    [anon_sym_times] = ACTIONS(3081),
    [anon_sym_wait] = ACTIONS(3081),
    [anon_sym_wantarray] = ACTIONS(3081),
    [anon_sym_abs] = ACTIONS(3081),
    [anon_sym_alarm] = ACTIONS(3081),
    [anon_sym_chop] = ACTIONS(3081),
    [anon_sym_chdir] = ACTIONS(3081),
    [anon_sym_close] = ACTIONS(3081),
    [anon_sym_closedir] = ACTIONS(3081),
    [anon_sym_caller] = ACTIONS(3081),
    [anon_sym_chomp] = ACTIONS(3081),
    [anon_sym_chr] = ACTIONS(3081),
    [anon_sym_cos] = ACTIONS(3081),
    [anon_sym_chroot] = ACTIONS(3081),
    [anon_sym_defined] = ACTIONS(3081),
    [anon_sym_delete] = ACTIONS(3081),
    [anon_sym_dbmclose] = ACTIONS(3081),
    [anon_sym_exists] = ACTIONS(3081),
    [anon_sym_exit] = ACTIONS(3081),
    [anon_sym_eof] = ACTIONS(3081),
    [anon_sym_exp] = ACTIONS(3081),
    [anon_sym_each] = ACTIONS(3081),
    [anon_sym_fc] = ACTIONS(3081),
    [anon_sym_fileno] = ACTIONS(3081),
    [anon_sym_gmtime] = ACTIONS(3081),
    [anon_sym_getc] = ACTIONS(3081),
    [anon_sym_getpgrp] = ACTIONS(3081),
    [anon_sym_getprotobyname] = ACTIONS(3081),
    [anon_sym_getpwname] = ACTIONS(3081),
    [anon_sym_getpwuid] = ACTIONS(3081),
    [anon_sym_getpeername] = ACTIONS(3081),
    [anon_sym_getnetbyname] = ACTIONS(3081),
    [anon_sym_getsockname] = ACTIONS(3081),
    [anon_sym_getgrnam] = ACTIONS(3081),
    [anon_sym_getgrgid] = ACTIONS(3081),
    [anon_sym_hex] = ACTIONS(3081),
    [anon_sym_int] = ACTIONS(3081),
    [anon_sym_keys] = ACTIONS(3081),
    [anon_sym_lc] = ACTIONS(3081),
    [anon_sym_lcfirst] = ACTIONS(3081),
    [anon_sym_length] = ACTIONS(3081),
    [anon_sym_localtime] = ACTIONS(3081),
    [anon_sym_log] = ACTIONS(3081),
    [anon_sym_lock] = ACTIONS(3081),
    [anon_sym_lstat] = ACTIONS(3081),
    [anon_sym_oct] = ACTIONS(3081),
    [anon_sym_ord] = ACTIONS(3081),
    [anon_sym_prototype] = ACTIONS(3081),
    [anon_sym_pop] = ACTIONS(3081),
    [anon_sym_pos] = ACTIONS(3081),
    [anon_sym_quotemeta] = ACTIONS(3081),
    [anon_sym_reset] = ACTIONS(3081),
    [anon_sym_rand] = ACTIONS(3081),
    [anon_sym_rmdir] = ACTIONS(3081),
    [anon_sym_readdir] = ACTIONS(3081),
    [anon_sym_readline] = ACTIONS(3081),
    [anon_sym_readpipe] = ACTIONS(3081),
    [anon_sym_rewinddir] = ACTIONS(3081),
    [anon_sym_readlink] = ACTIONS(3081),
    [anon_sym_ref] = ACTIONS(3081),
    [anon_sym_scalar] = ACTIONS(3081),
    [anon_sym_shift] = ACTIONS(3081),
    [anon_sym_sin] = ACTIONS(3081),
    [anon_sym_sleep] = ACTIONS(3081),
    [anon_sym_sqrt] = ACTIONS(3081),
    [anon_sym_srand] = ACTIONS(3081),
    [anon_sym_stat] = ACTIONS(3081),
    [anon_sym_study] = ACTIONS(3081),
    [anon_sym_tell] = ACTIONS(3081),
    [anon_sym_telldir] = ACTIONS(3081),
    [anon_sym_tied] = ACTIONS(3081),
    [anon_sym_uc] = ACTIONS(3081),
    [anon_sym_ucfirst] = ACTIONS(3081),
    [anon_sym_untie] = ACTIONS(3081),
    [anon_sym_umask] = ACTIONS(3081),
    [anon_sym_values] = ACTIONS(3081),
    [anon_sym_write] = ACTIONS(3081),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3081),
    [anon_sym_qq] = ACTIONS(3081),
    [anon_sym_qw] = ACTIONS(3081),
    [anon_sym_qx] = ACTIONS(3081),
    [anon_sym_qr] = ACTIONS(3081),
    [anon_sym_m] = ACTIONS(3081),
    [anon_sym_s] = ACTIONS(3081),
    [anon_sym_tr] = ACTIONS(3081),
    [anon_sym_y] = ACTIONS(3081),
    [anon_sym_if] = ACTIONS(3081),
    [anon_sym_unless] = ACTIONS(3081),
    [anon_sym_while] = ACTIONS(3081),
    [anon_sym_until] = ACTIONS(3081),
    [aux_sym__bareword_token1] = ACTIONS(3081),
    [sym_number] = ACTIONS(3081),
    [anon_sym_true] = ACTIONS(3081),
    [anon_sym_false] = ACTIONS(3081),
    [sym__single_quote] = ACTIONS(3079),
    [sym__double_quote] = ACTIONS(3079),
    [sym__backtick_quote] = ACTIONS(3079),
    [sym__search_slash_quote] = ACTIONS(3079),
    [sym__open_readline_bracket] = ACTIONS(3079),
    [sym__open_fileglob_bracket] = ACTIONS(3079),
    [sym__PERLY_HEREDOC] = ACTIONS(3079),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3079),
    [sym__filetest] = ACTIONS(3079),
  },
  [STATE(1542)] = {
    [sym_heredoc_content] = STATE(1542),
    [sym__identifier] = ACTIONS(3121),
    [anon_sym_LBRACE] = ACTIONS(3121),
    [anon_sym_RBRACE] = ACTIONS(3121),
    [anon_sym_SEMI] = ACTIONS(3121),
    [anon_sym_package] = ACTIONS(3121),
    [anon_sym_class] = ACTIONS(3121),
    [anon_sym_role] = ACTIONS(3121),
    [anon_sym_BUILD] = ACTIONS(3121),
    [anon_sym_ADJUST] = ACTIONS(3121),
    [anon_sym_DOLLAR] = ACTIONS(3121),
    [anon_sym_AT] = ACTIONS(3121),
    [anon_sym_my] = ACTIONS(3121),
    [anon_sym_extended] = ACTIONS(3121),
    [anon_sym_async] = ACTIONS(3121),
    [anon_sym_sub] = ACTIONS(3121),
    [anon_sym_method] = ACTIONS(3121),
    [anon_sym_LPAREN] = ACTIONS(3121),
    [anon_sym_continue] = ACTIONS(3121),
    [anon_sym_state] = ACTIONS(3121),
    [anon_sym_our] = ACTIONS(3121),
    [anon_sym_try] = ACTIONS(3121),
    [anon_sym_defer] = ACTIONS(3121),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3121),
    [anon_sym_LBRACK] = ACTIONS(3121),
    [anon_sym_LT_LT] = ACTIONS(3121),
    [anon_sym_SLASH_SLASH] = ACTIONS(3121),
    [anon_sym_PLUS] = ACTIONS(3121),
    [anon_sym_DASH] = ACTIONS(3121),
    [anon_sym_TILDE] = ACTIONS(3121),
    [anon_sym_BANG] = ACTIONS(3121),
    [anon_sym_PLUS_PLUS] = ACTIONS(3121),
    [anon_sym_DASH_DASH] = ACTIONS(3121),
    [anon_sym_BSLASH] = ACTIONS(3119),
    [anon_sym_do] = ACTIONS(3121),
    [anon_sym_eval] = ACTIONS(3121),
    [anon_sym_await] = ACTIONS(3121),
    [anon_sym_field] = ACTIONS(3121),
    [anon_sym_local] = ACTIONS(3121),
    [anon_sym_dynamically] = ACTIONS(3121),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3121),
    [anon_sym_require] = ACTIONS(3121),
    [anon_sym_map] = ACTIONS(3121),
    [anon_sym_grep] = ACTIONS(3121),
    [anon_sym_sort] = ACTIONS(3121),
    [anon_sym_goto] = ACTIONS(3121),
    [anon_sym_return] = ACTIONS(3121),
    [anon_sym_undef] = ACTIONS(3121),
    [anon_sym_PERCENT2] = ACTIONS(3121),
    [anon_sym_AMP2] = ACTIONS(3121),
    [anon_sym_STAR2] = ACTIONS(3121),
    [anon_sym_use] = ACTIONS(3121),
    [anon_sym_no] = ACTIONS(3121),
    [anon_sym_for] = ACTIONS(3121),
    [anon_sym_foreach] = ACTIONS(3121),
    [anon_sym_last] = ACTIONS(3121),
    [anon_sym_next] = ACTIONS(3121),
    [anon_sym_redo] = ACTIONS(3121),
    [anon_sym_BEGIN] = ACTIONS(3121),
    [anon_sym_INIT] = ACTIONS(3121),
    [anon_sym_CHECK] = ACTIONS(3121),
    [anon_sym_UNITCHECK] = ACTIONS(3121),
    [anon_sym_END] = ACTIONS(3121),
    [anon_sym___FILE__] = ACTIONS(3121),
    [anon_sym___LINE__] = ACTIONS(3121),
    [anon_sym___PACKAGE__] = ACTIONS(3121),
    [anon_sym___SUB__] = ACTIONS(3121),
    [anon_sym_break] = ACTIONS(3121),
    [anon_sym_fork] = ACTIONS(3121),
    [anon_sym_getppid] = ACTIONS(3121),
    [anon_sym_time] = ACTIONS(3121),
    [anon_sym_times] = ACTIONS(3121),
    [anon_sym_wait] = ACTIONS(3121),
    [anon_sym_wantarray] = ACTIONS(3121),
    [anon_sym_abs] = ACTIONS(3121),
    [anon_sym_alarm] = ACTIONS(3121),
    [anon_sym_chop] = ACTIONS(3121),
    [anon_sym_chdir] = ACTIONS(3121),
    [anon_sym_close] = ACTIONS(3121),
    [anon_sym_closedir] = ACTIONS(3121),
    [anon_sym_caller] = ACTIONS(3121),
    [anon_sym_chomp] = ACTIONS(3121),
    [anon_sym_chr] = ACTIONS(3121),
    [anon_sym_cos] = ACTIONS(3121),
    [anon_sym_chroot] = ACTIONS(3121),
    [anon_sym_defined] = ACTIONS(3121),
    [anon_sym_delete] = ACTIONS(3121),
    [anon_sym_dbmclose] = ACTIONS(3121),
    [anon_sym_exists] = ACTIONS(3121),
    [anon_sym_exit] = ACTIONS(3121),
    [anon_sym_eof] = ACTIONS(3121),
    [anon_sym_exp] = ACTIONS(3121),
    [anon_sym_each] = ACTIONS(3121),
    [anon_sym_fc] = ACTIONS(3121),
    [anon_sym_fileno] = ACTIONS(3121),
    [anon_sym_gmtime] = ACTIONS(3121),
    [anon_sym_getc] = ACTIONS(3121),
    [anon_sym_getpgrp] = ACTIONS(3121),
    [anon_sym_getprotobyname] = ACTIONS(3121),
    [anon_sym_getpwname] = ACTIONS(3121),
    [anon_sym_getpwuid] = ACTIONS(3121),
    [anon_sym_getpeername] = ACTIONS(3121),
    [anon_sym_getnetbyname] = ACTIONS(3121),
    [anon_sym_getsockname] = ACTIONS(3121),
    [anon_sym_getgrnam] = ACTIONS(3121),
    [anon_sym_getgrgid] = ACTIONS(3121),
    [anon_sym_hex] = ACTIONS(3121),
    [anon_sym_int] = ACTIONS(3121),
    [anon_sym_keys] = ACTIONS(3121),
    [anon_sym_lc] = ACTIONS(3121),
    [anon_sym_lcfirst] = ACTIONS(3121),
    [anon_sym_length] = ACTIONS(3121),
    [anon_sym_localtime] = ACTIONS(3121),
    [anon_sym_log] = ACTIONS(3121),
    [anon_sym_lock] = ACTIONS(3121),
    [anon_sym_lstat] = ACTIONS(3121),
    [anon_sym_oct] = ACTIONS(3121),
    [anon_sym_ord] = ACTIONS(3121),
    [anon_sym_prototype] = ACTIONS(3121),
    [anon_sym_pop] = ACTIONS(3121),
    [anon_sym_pos] = ACTIONS(3121),
    [anon_sym_quotemeta] = ACTIONS(3121),
    [anon_sym_reset] = ACTIONS(3121),
    [anon_sym_rand] = ACTIONS(3121),
    [anon_sym_rmdir] = ACTIONS(3121),
    [anon_sym_readdir] = ACTIONS(3121),
    [anon_sym_readline] = ACTIONS(3121),
    [anon_sym_readpipe] = ACTIONS(3121),
    [anon_sym_rewinddir] = ACTIONS(3121),
    [anon_sym_readlink] = ACTIONS(3121),
    [anon_sym_ref] = ACTIONS(3121),
    [anon_sym_scalar] = ACTIONS(3121),
    [anon_sym_shift] = ACTIONS(3121),
    [anon_sym_sin] = ACTIONS(3121),
    [anon_sym_sleep] = ACTIONS(3121),
    [anon_sym_sqrt] = ACTIONS(3121),
    [anon_sym_srand] = ACTIONS(3121),
    [anon_sym_stat] = ACTIONS(3121),
    [anon_sym_study] = ACTIONS(3121),
    [anon_sym_tell] = ACTIONS(3121),
    [anon_sym_telldir] = ACTIONS(3121),
    [anon_sym_tied] = ACTIONS(3121),
    [anon_sym_uc] = ACTIONS(3121),
    [anon_sym_ucfirst] = ACTIONS(3121),
    [anon_sym_untie] = ACTIONS(3121),
    [anon_sym_umask] = ACTIONS(3121),
    [anon_sym_values] = ACTIONS(3121),
    [anon_sym_write] = ACTIONS(3121),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3121),
    [anon_sym_qq] = ACTIONS(3121),
    [anon_sym_qw] = ACTIONS(3121),
    [anon_sym_qx] = ACTIONS(3121),
    [anon_sym_qr] = ACTIONS(3121),
    [anon_sym_m] = ACTIONS(3121),
    [anon_sym_s] = ACTIONS(3121),
    [anon_sym_tr] = ACTIONS(3121),
    [anon_sym_y] = ACTIONS(3121),
    [anon_sym_if] = ACTIONS(3121),
    [anon_sym_unless] = ACTIONS(3121),
    [anon_sym_while] = ACTIONS(3121),
    [anon_sym_until] = ACTIONS(3121),
    [aux_sym__bareword_token1] = ACTIONS(3121),
    [sym_number] = ACTIONS(3121),
    [anon_sym_true] = ACTIONS(3121),
    [anon_sym_false] = ACTIONS(3121),
    [sym__single_quote] = ACTIONS(3119),
    [sym__double_quote] = ACTIONS(3119),
    [sym__backtick_quote] = ACTIONS(3119),
    [sym__search_slash_quote] = ACTIONS(3119),
    [sym__open_readline_bracket] = ACTIONS(3119),
    [sym__open_fileglob_bracket] = ACTIONS(3119),
    [sym__PERLY_HEREDOC] = ACTIONS(3119),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3119),
    [sym__filetest] = ACTIONS(3119),
  },
  [STATE(1543)] = {
    [sym_heredoc_content] = STATE(1543),
    [sym__identifier] = ACTIONS(3125),
    [anon_sym_LBRACE] = ACTIONS(3125),
    [anon_sym_RBRACE] = ACTIONS(3125),
    [anon_sym_SEMI] = ACTIONS(3125),
    [anon_sym_package] = ACTIONS(3125),
    [anon_sym_class] = ACTIONS(3125),
    [anon_sym_role] = ACTIONS(3125),
    [anon_sym_BUILD] = ACTIONS(3125),
    [anon_sym_ADJUST] = ACTIONS(3125),
    [anon_sym_DOLLAR] = ACTIONS(3125),
    [anon_sym_AT] = ACTIONS(3125),
    [anon_sym_my] = ACTIONS(3125),
    [anon_sym_extended] = ACTIONS(3125),
    [anon_sym_async] = ACTIONS(3125),
    [anon_sym_sub] = ACTIONS(3125),
    [anon_sym_method] = ACTIONS(3125),
    [anon_sym_LPAREN] = ACTIONS(3125),
    [anon_sym_continue] = ACTIONS(3125),
    [anon_sym_state] = ACTIONS(3125),
    [anon_sym_our] = ACTIONS(3125),
    [anon_sym_try] = ACTIONS(3125),
    [anon_sym_defer] = ACTIONS(3125),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3125),
    [anon_sym_LBRACK] = ACTIONS(3125),
    [anon_sym_LT_LT] = ACTIONS(3125),
    [anon_sym_SLASH_SLASH] = ACTIONS(3125),
    [anon_sym_PLUS] = ACTIONS(3125),
    [anon_sym_DASH] = ACTIONS(3125),
    [anon_sym_TILDE] = ACTIONS(3125),
    [anon_sym_BANG] = ACTIONS(3125),
    [anon_sym_PLUS_PLUS] = ACTIONS(3125),
    [anon_sym_DASH_DASH] = ACTIONS(3125),
    [anon_sym_BSLASH] = ACTIONS(3123),
    [anon_sym_do] = ACTIONS(3125),
    [anon_sym_eval] = ACTIONS(3125),
    [anon_sym_await] = ACTIONS(3125),
    [anon_sym_field] = ACTIONS(3125),
    [anon_sym_local] = ACTIONS(3125),
    [anon_sym_dynamically] = ACTIONS(3125),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3125),
    [anon_sym_require] = ACTIONS(3125),
    [anon_sym_map] = ACTIONS(3125),
    [anon_sym_grep] = ACTIONS(3125),
    [anon_sym_sort] = ACTIONS(3125),
    [anon_sym_goto] = ACTIONS(3125),
    [anon_sym_return] = ACTIONS(3125),
    [anon_sym_undef] = ACTIONS(3125),
    [anon_sym_PERCENT2] = ACTIONS(3125),
    [anon_sym_AMP2] = ACTIONS(3125),
    [anon_sym_STAR2] = ACTIONS(3125),
    [anon_sym_use] = ACTIONS(3125),
    [anon_sym_no] = ACTIONS(3125),
    [anon_sym_for] = ACTIONS(3125),
    [anon_sym_foreach] = ACTIONS(3125),
    [anon_sym_last] = ACTIONS(3125),
    [anon_sym_next] = ACTIONS(3125),
    [anon_sym_redo] = ACTIONS(3125),
    [anon_sym_BEGIN] = ACTIONS(3125),
    [anon_sym_INIT] = ACTIONS(3125),
    [anon_sym_CHECK] = ACTIONS(3125),
    [anon_sym_UNITCHECK] = ACTIONS(3125),
    [anon_sym_END] = ACTIONS(3125),
    [anon_sym___FILE__] = ACTIONS(3125),
    [anon_sym___LINE__] = ACTIONS(3125),
    [anon_sym___PACKAGE__] = ACTIONS(3125),
    [anon_sym___SUB__] = ACTIONS(3125),
    [anon_sym_break] = ACTIONS(3125),
    [anon_sym_fork] = ACTIONS(3125),
    [anon_sym_getppid] = ACTIONS(3125),
    [anon_sym_time] = ACTIONS(3125),
    [anon_sym_times] = ACTIONS(3125),
    [anon_sym_wait] = ACTIONS(3125),
    [anon_sym_wantarray] = ACTIONS(3125),
    [anon_sym_abs] = ACTIONS(3125),
    [anon_sym_alarm] = ACTIONS(3125),
    [anon_sym_chop] = ACTIONS(3125),
    [anon_sym_chdir] = ACTIONS(3125),
    [anon_sym_close] = ACTIONS(3125),
    [anon_sym_closedir] = ACTIONS(3125),
    [anon_sym_caller] = ACTIONS(3125),
    [anon_sym_chomp] = ACTIONS(3125),
    [anon_sym_chr] = ACTIONS(3125),
    [anon_sym_cos] = ACTIONS(3125),
    [anon_sym_chroot] = ACTIONS(3125),
    [anon_sym_defined] = ACTIONS(3125),
    [anon_sym_delete] = ACTIONS(3125),
    [anon_sym_dbmclose] = ACTIONS(3125),
    [anon_sym_exists] = ACTIONS(3125),
    [anon_sym_exit] = ACTIONS(3125),
    [anon_sym_eof] = ACTIONS(3125),
    [anon_sym_exp] = ACTIONS(3125),
    [anon_sym_each] = ACTIONS(3125),
    [anon_sym_fc] = ACTIONS(3125),
    [anon_sym_fileno] = ACTIONS(3125),
    [anon_sym_gmtime] = ACTIONS(3125),
    [anon_sym_getc] = ACTIONS(3125),
    [anon_sym_getpgrp] = ACTIONS(3125),
    [anon_sym_getprotobyname] = ACTIONS(3125),
    [anon_sym_getpwname] = ACTIONS(3125),
    [anon_sym_getpwuid] = ACTIONS(3125),
    [anon_sym_getpeername] = ACTIONS(3125),
    [anon_sym_getnetbyname] = ACTIONS(3125),
    [anon_sym_getsockname] = ACTIONS(3125),
    [anon_sym_getgrnam] = ACTIONS(3125),
    [anon_sym_getgrgid] = ACTIONS(3125),
    [anon_sym_hex] = ACTIONS(3125),
    [anon_sym_int] = ACTIONS(3125),
    [anon_sym_keys] = ACTIONS(3125),
    [anon_sym_lc] = ACTIONS(3125),
    [anon_sym_lcfirst] = ACTIONS(3125),
    [anon_sym_length] = ACTIONS(3125),
    [anon_sym_localtime] = ACTIONS(3125),
    [anon_sym_log] = ACTIONS(3125),
    [anon_sym_lock] = ACTIONS(3125),
    [anon_sym_lstat] = ACTIONS(3125),
    [anon_sym_oct] = ACTIONS(3125),
    [anon_sym_ord] = ACTIONS(3125),
    [anon_sym_prototype] = ACTIONS(3125),
    [anon_sym_pop] = ACTIONS(3125),
    [anon_sym_pos] = ACTIONS(3125),
    [anon_sym_quotemeta] = ACTIONS(3125),
    [anon_sym_reset] = ACTIONS(3125),
    [anon_sym_rand] = ACTIONS(3125),
    [anon_sym_rmdir] = ACTIONS(3125),
    [anon_sym_readdir] = ACTIONS(3125),
    [anon_sym_readline] = ACTIONS(3125),
    [anon_sym_readpipe] = ACTIONS(3125),
    [anon_sym_rewinddir] = ACTIONS(3125),
    [anon_sym_readlink] = ACTIONS(3125),
    [anon_sym_ref] = ACTIONS(3125),
    [anon_sym_scalar] = ACTIONS(3125),
    [anon_sym_shift] = ACTIONS(3125),
    [anon_sym_sin] = ACTIONS(3125),
    [anon_sym_sleep] = ACTIONS(3125),
    [anon_sym_sqrt] = ACTIONS(3125),
    [anon_sym_srand] = ACTIONS(3125),
    [anon_sym_stat] = ACTIONS(3125),
    [anon_sym_study] = ACTIONS(3125),
    [anon_sym_tell] = ACTIONS(3125),
    [anon_sym_telldir] = ACTIONS(3125),
    [anon_sym_tied] = ACTIONS(3125),
    [anon_sym_uc] = ACTIONS(3125),
    [anon_sym_ucfirst] = ACTIONS(3125),
    [anon_sym_untie] = ACTIONS(3125),
    [anon_sym_umask] = ACTIONS(3125),
    [anon_sym_values] = ACTIONS(3125),
    [anon_sym_write] = ACTIONS(3125),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3125),
    [anon_sym_qq] = ACTIONS(3125),
    [anon_sym_qw] = ACTIONS(3125),
    [anon_sym_qx] = ACTIONS(3125),
    [anon_sym_qr] = ACTIONS(3125),
    [anon_sym_m] = ACTIONS(3125),
    [anon_sym_s] = ACTIONS(3125),
    [anon_sym_tr] = ACTIONS(3125),
    [anon_sym_y] = ACTIONS(3125),
    [anon_sym_if] = ACTIONS(3125),
    [anon_sym_unless] = ACTIONS(3125),
    [anon_sym_while] = ACTIONS(3125),
    [anon_sym_until] = ACTIONS(3125),
    [aux_sym__bareword_token1] = ACTIONS(3125),
    [sym_number] = ACTIONS(3125),
    [anon_sym_true] = ACTIONS(3125),
    [anon_sym_false] = ACTIONS(3125),
    [sym__single_quote] = ACTIONS(3123),
    [sym__double_quote] = ACTIONS(3123),
    [sym__backtick_quote] = ACTIONS(3123),
    [sym__search_slash_quote] = ACTIONS(3123),
    [sym__open_readline_bracket] = ACTIONS(3123),
    [sym__open_fileglob_bracket] = ACTIONS(3123),
    [sym__PERLY_HEREDOC] = ACTIONS(3123),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3123),
    [sym__filetest] = ACTIONS(3123),
  },
  [STATE(1544)] = {
    [sym_heredoc_content] = STATE(1544),
    [sym__identifier] = ACTIONS(3157),
    [anon_sym_LBRACE] = ACTIONS(3157),
    [anon_sym_RBRACE] = ACTIONS(3157),
    [anon_sym_SEMI] = ACTIONS(3157),
    [anon_sym_package] = ACTIONS(3157),
    [anon_sym_class] = ACTIONS(3157),
    [anon_sym_role] = ACTIONS(3157),
    [anon_sym_BUILD] = ACTIONS(3157),
    [anon_sym_ADJUST] = ACTIONS(3157),
    [anon_sym_DOLLAR] = ACTIONS(3157),
    [anon_sym_AT] = ACTIONS(3157),
    [anon_sym_my] = ACTIONS(3157),
    [anon_sym_extended] = ACTIONS(3157),
    [anon_sym_async] = ACTIONS(3157),
    [anon_sym_sub] = ACTIONS(3157),
    [anon_sym_method] = ACTIONS(3157),
    [anon_sym_LPAREN] = ACTIONS(3157),
    [anon_sym_continue] = ACTIONS(3157),
    [anon_sym_state] = ACTIONS(3157),
    [anon_sym_our] = ACTIONS(3157),
    [anon_sym_try] = ACTIONS(3157),
    [anon_sym_defer] = ACTIONS(3157),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3157),
    [anon_sym_LBRACK] = ACTIONS(3157),
    [anon_sym_LT_LT] = ACTIONS(3157),
    [anon_sym_SLASH_SLASH] = ACTIONS(3157),
    [anon_sym_PLUS] = ACTIONS(3157),
    [anon_sym_DASH] = ACTIONS(3157),
    [anon_sym_TILDE] = ACTIONS(3157),
    [anon_sym_BANG] = ACTIONS(3157),
    [anon_sym_PLUS_PLUS] = ACTIONS(3157),
    [anon_sym_DASH_DASH] = ACTIONS(3157),
    [anon_sym_BSLASH] = ACTIONS(3155),
    [anon_sym_do] = ACTIONS(3157),
    [anon_sym_eval] = ACTIONS(3157),
    [anon_sym_await] = ACTIONS(3157),
    [anon_sym_field] = ACTIONS(3157),
    [anon_sym_local] = ACTIONS(3157),
    [anon_sym_dynamically] = ACTIONS(3157),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3157),
    [anon_sym_require] = ACTIONS(3157),
    [anon_sym_map] = ACTIONS(3157),
    [anon_sym_grep] = ACTIONS(3157),
    [anon_sym_sort] = ACTIONS(3157),
    [anon_sym_goto] = ACTIONS(3157),
    [anon_sym_return] = ACTIONS(3157),
    [anon_sym_undef] = ACTIONS(3157),
    [anon_sym_PERCENT2] = ACTIONS(3157),
    [anon_sym_AMP2] = ACTIONS(3157),
    [anon_sym_STAR2] = ACTIONS(3157),
    [anon_sym_use] = ACTIONS(3157),
    [anon_sym_no] = ACTIONS(3157),
    [anon_sym_for] = ACTIONS(3157),
    [anon_sym_foreach] = ACTIONS(3157),
    [anon_sym_last] = ACTIONS(3157),
    [anon_sym_next] = ACTIONS(3157),
    [anon_sym_redo] = ACTIONS(3157),
    [anon_sym_BEGIN] = ACTIONS(3157),
    [anon_sym_INIT] = ACTIONS(3157),
    [anon_sym_CHECK] = ACTIONS(3157),
    [anon_sym_UNITCHECK] = ACTIONS(3157),
    [anon_sym_END] = ACTIONS(3157),
    [anon_sym___FILE__] = ACTIONS(3157),
    [anon_sym___LINE__] = ACTIONS(3157),
    [anon_sym___PACKAGE__] = ACTIONS(3157),
    [anon_sym___SUB__] = ACTIONS(3157),
    [anon_sym_break] = ACTIONS(3157),
    [anon_sym_fork] = ACTIONS(3157),
    [anon_sym_getppid] = ACTIONS(3157),
    [anon_sym_time] = ACTIONS(3157),
    [anon_sym_times] = ACTIONS(3157),
    [anon_sym_wait] = ACTIONS(3157),
    [anon_sym_wantarray] = ACTIONS(3157),
    [anon_sym_abs] = ACTIONS(3157),
    [anon_sym_alarm] = ACTIONS(3157),
    [anon_sym_chop] = ACTIONS(3157),
    [anon_sym_chdir] = ACTIONS(3157),
    [anon_sym_close] = ACTIONS(3157),
    [anon_sym_closedir] = ACTIONS(3157),
    [anon_sym_caller] = ACTIONS(3157),
    [anon_sym_chomp] = ACTIONS(3157),
    [anon_sym_chr] = ACTIONS(3157),
    [anon_sym_cos] = ACTIONS(3157),
    [anon_sym_chroot] = ACTIONS(3157),
    [anon_sym_defined] = ACTIONS(3157),
    [anon_sym_delete] = ACTIONS(3157),
    [anon_sym_dbmclose] = ACTIONS(3157),
    [anon_sym_exists] = ACTIONS(3157),
    [anon_sym_exit] = ACTIONS(3157),
    [anon_sym_eof] = ACTIONS(3157),
    [anon_sym_exp] = ACTIONS(3157),
    [anon_sym_each] = ACTIONS(3157),
    [anon_sym_fc] = ACTIONS(3157),
    [anon_sym_fileno] = ACTIONS(3157),
    [anon_sym_gmtime] = ACTIONS(3157),
    [anon_sym_getc] = ACTIONS(3157),
    [anon_sym_getpgrp] = ACTIONS(3157),
    [anon_sym_getprotobyname] = ACTIONS(3157),
    [anon_sym_getpwname] = ACTIONS(3157),
    [anon_sym_getpwuid] = ACTIONS(3157),
    [anon_sym_getpeername] = ACTIONS(3157),
    [anon_sym_getnetbyname] = ACTIONS(3157),
    [anon_sym_getsockname] = ACTIONS(3157),
    [anon_sym_getgrnam] = ACTIONS(3157),
    [anon_sym_getgrgid] = ACTIONS(3157),
    [anon_sym_hex] = ACTIONS(3157),
    [anon_sym_int] = ACTIONS(3157),
    [anon_sym_keys] = ACTIONS(3157),
    [anon_sym_lc] = ACTIONS(3157),
    [anon_sym_lcfirst] = ACTIONS(3157),
    [anon_sym_length] = ACTIONS(3157),
    [anon_sym_localtime] = ACTIONS(3157),
    [anon_sym_log] = ACTIONS(3157),
    [anon_sym_lock] = ACTIONS(3157),
    [anon_sym_lstat] = ACTIONS(3157),
    [anon_sym_oct] = ACTIONS(3157),
    [anon_sym_ord] = ACTIONS(3157),
    [anon_sym_prototype] = ACTIONS(3157),
    [anon_sym_pop] = ACTIONS(3157),
    [anon_sym_pos] = ACTIONS(3157),
    [anon_sym_quotemeta] = ACTIONS(3157),
    [anon_sym_reset] = ACTIONS(3157),
    [anon_sym_rand] = ACTIONS(3157),
    [anon_sym_rmdir] = ACTIONS(3157),
    [anon_sym_readdir] = ACTIONS(3157),
    [anon_sym_readline] = ACTIONS(3157),
    [anon_sym_readpipe] = ACTIONS(3157),
    [anon_sym_rewinddir] = ACTIONS(3157),
    [anon_sym_readlink] = ACTIONS(3157),
    [anon_sym_ref] = ACTIONS(3157),
    [anon_sym_scalar] = ACTIONS(3157),
    [anon_sym_shift] = ACTIONS(3157),
    [anon_sym_sin] = ACTIONS(3157),
    [anon_sym_sleep] = ACTIONS(3157),
    [anon_sym_sqrt] = ACTIONS(3157),
    [anon_sym_srand] = ACTIONS(3157),
    [anon_sym_stat] = ACTIONS(3157),
    [anon_sym_study] = ACTIONS(3157),
    [anon_sym_tell] = ACTIONS(3157),
    [anon_sym_telldir] = ACTIONS(3157),
    [anon_sym_tied] = ACTIONS(3157),
    [anon_sym_uc] = ACTIONS(3157),
    [anon_sym_ucfirst] = ACTIONS(3157),
    [anon_sym_untie] = ACTIONS(3157),
    [anon_sym_umask] = ACTIONS(3157),
    [anon_sym_values] = ACTIONS(3157),
    [anon_sym_write] = ACTIONS(3157),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3157),
    [anon_sym_qq] = ACTIONS(3157),
    [anon_sym_qw] = ACTIONS(3157),
    [anon_sym_qx] = ACTIONS(3157),
    [anon_sym_qr] = ACTIONS(3157),
    [anon_sym_m] = ACTIONS(3157),
    [anon_sym_s] = ACTIONS(3157),
    [anon_sym_tr] = ACTIONS(3157),
    [anon_sym_y] = ACTIONS(3157),
    [anon_sym_if] = ACTIONS(3157),
    [anon_sym_unless] = ACTIONS(3157),
    [anon_sym_while] = ACTIONS(3157),
    [anon_sym_until] = ACTIONS(3157),
    [aux_sym__bareword_token1] = ACTIONS(3157),
    [sym_number] = ACTIONS(3157),
    [anon_sym_true] = ACTIONS(3157),
    [anon_sym_false] = ACTIONS(3157),
    [sym__single_quote] = ACTIONS(3155),
    [sym__double_quote] = ACTIONS(3155),
    [sym__backtick_quote] = ACTIONS(3155),
    [sym__search_slash_quote] = ACTIONS(3155),
    [sym__open_readline_bracket] = ACTIONS(3155),
    [sym__open_fileglob_bracket] = ACTIONS(3155),
    [sym__PERLY_HEREDOC] = ACTIONS(3155),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3155),
    [sym__filetest] = ACTIONS(3155),
  },
  [STATE(1545)] = {
    [sym_heredoc_content] = STATE(1545),
    [sym__identifier] = ACTIONS(2942),
    [anon_sym_LBRACE] = ACTIONS(2942),
    [anon_sym_RBRACE] = ACTIONS(2942),
    [anon_sym_SEMI] = ACTIONS(2942),
    [anon_sym_package] = ACTIONS(2942),
    [anon_sym_class] = ACTIONS(2942),
    [anon_sym_role] = ACTIONS(2942),
    [anon_sym_BUILD] = ACTIONS(2942),
    [anon_sym_ADJUST] = ACTIONS(2942),
    [anon_sym_DOLLAR] = ACTIONS(2942),
    [anon_sym_AT] = ACTIONS(2942),
    [anon_sym_my] = ACTIONS(2942),
    [anon_sym_extended] = ACTIONS(2942),
    [anon_sym_async] = ACTIONS(2942),
    [anon_sym_sub] = ACTIONS(2942),
    [anon_sym_method] = ACTIONS(2942),
    [anon_sym_LPAREN] = ACTIONS(2942),
    [anon_sym_continue] = ACTIONS(2942),
    [anon_sym_state] = ACTIONS(2942),
    [anon_sym_our] = ACTIONS(2942),
    [anon_sym_try] = ACTIONS(2942),
    [anon_sym_defer] = ACTIONS(2942),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2942),
    [anon_sym_LBRACK] = ACTIONS(2942),
    [anon_sym_LT_LT] = ACTIONS(2942),
    [anon_sym_SLASH_SLASH] = ACTIONS(2942),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_TILDE] = ACTIONS(2942),
    [anon_sym_BANG] = ACTIONS(2942),
    [anon_sym_PLUS_PLUS] = ACTIONS(2942),
    [anon_sym_DASH_DASH] = ACTIONS(2942),
    [anon_sym_BSLASH] = ACTIONS(2940),
    [anon_sym_do] = ACTIONS(2942),
    [anon_sym_eval] = ACTIONS(2942),
    [anon_sym_await] = ACTIONS(2942),
    [anon_sym_field] = ACTIONS(2942),
    [anon_sym_local] = ACTIONS(2942),
    [anon_sym_dynamically] = ACTIONS(2942),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2942),
    [anon_sym_require] = ACTIONS(2942),
    [anon_sym_map] = ACTIONS(2942),
    [anon_sym_grep] = ACTIONS(2942),
    [anon_sym_sort] = ACTIONS(2942),
    [anon_sym_goto] = ACTIONS(2942),
    [anon_sym_return] = ACTIONS(2942),
    [anon_sym_undef] = ACTIONS(2942),
    [anon_sym_PERCENT2] = ACTIONS(2942),
    [anon_sym_AMP2] = ACTIONS(2942),
    [anon_sym_STAR2] = ACTIONS(2942),
    [anon_sym_use] = ACTIONS(2942),
    [anon_sym_no] = ACTIONS(2942),
    [anon_sym_for] = ACTIONS(2942),
    [anon_sym_foreach] = ACTIONS(2942),
    [anon_sym_last] = ACTIONS(2942),
    [anon_sym_next] = ACTIONS(2942),
    [anon_sym_redo] = ACTIONS(2942),
    [anon_sym_BEGIN] = ACTIONS(2942),
    [anon_sym_INIT] = ACTIONS(2942),
    [anon_sym_CHECK] = ACTIONS(2942),
    [anon_sym_UNITCHECK] = ACTIONS(2942),
    [anon_sym_END] = ACTIONS(2942),
    [anon_sym___FILE__] = ACTIONS(2942),
    [anon_sym___LINE__] = ACTIONS(2942),
    [anon_sym___PACKAGE__] = ACTIONS(2942),
    [anon_sym___SUB__] = ACTIONS(2942),
    [anon_sym_break] = ACTIONS(2942),
    [anon_sym_fork] = ACTIONS(2942),
    [anon_sym_getppid] = ACTIONS(2942),
    [anon_sym_time] = ACTIONS(2942),
    [anon_sym_times] = ACTIONS(2942),
    [anon_sym_wait] = ACTIONS(2942),
    [anon_sym_wantarray] = ACTIONS(2942),
    [anon_sym_abs] = ACTIONS(2942),
    [anon_sym_alarm] = ACTIONS(2942),
    [anon_sym_chop] = ACTIONS(2942),
    [anon_sym_chdir] = ACTIONS(2942),
    [anon_sym_close] = ACTIONS(2942),
    [anon_sym_closedir] = ACTIONS(2942),
    [anon_sym_caller] = ACTIONS(2942),
    [anon_sym_chomp] = ACTIONS(2942),
    [anon_sym_chr] = ACTIONS(2942),
    [anon_sym_cos] = ACTIONS(2942),
    [anon_sym_chroot] = ACTIONS(2942),
    [anon_sym_defined] = ACTIONS(2942),
    [anon_sym_delete] = ACTIONS(2942),
    [anon_sym_dbmclose] = ACTIONS(2942),
    [anon_sym_exists] = ACTIONS(2942),
    [anon_sym_exit] = ACTIONS(2942),
    [anon_sym_eof] = ACTIONS(2942),
    [anon_sym_exp] = ACTIONS(2942),
    [anon_sym_each] = ACTIONS(2942),
    [anon_sym_fc] = ACTIONS(2942),
    [anon_sym_fileno] = ACTIONS(2942),
    [anon_sym_gmtime] = ACTIONS(2942),
    [anon_sym_getc] = ACTIONS(2942),
    [anon_sym_getpgrp] = ACTIONS(2942),
    [anon_sym_getprotobyname] = ACTIONS(2942),
    [anon_sym_getpwname] = ACTIONS(2942),
    [anon_sym_getpwuid] = ACTIONS(2942),
    [anon_sym_getpeername] = ACTIONS(2942),
    [anon_sym_getnetbyname] = ACTIONS(2942),
    [anon_sym_getsockname] = ACTIONS(2942),
    [anon_sym_getgrnam] = ACTIONS(2942),
    [anon_sym_getgrgid] = ACTIONS(2942),
    [anon_sym_hex] = ACTIONS(2942),
    [anon_sym_int] = ACTIONS(2942),
    [anon_sym_keys] = ACTIONS(2942),
    [anon_sym_lc] = ACTIONS(2942),
    [anon_sym_lcfirst] = ACTIONS(2942),
    [anon_sym_length] = ACTIONS(2942),
    [anon_sym_localtime] = ACTIONS(2942),
    [anon_sym_log] = ACTIONS(2942),
    [anon_sym_lock] = ACTIONS(2942),
    [anon_sym_lstat] = ACTIONS(2942),
    [anon_sym_oct] = ACTIONS(2942),
    [anon_sym_ord] = ACTIONS(2942),
    [anon_sym_prototype] = ACTIONS(2942),
    [anon_sym_pop] = ACTIONS(2942),
    [anon_sym_pos] = ACTIONS(2942),
    [anon_sym_quotemeta] = ACTIONS(2942),
    [anon_sym_reset] = ACTIONS(2942),
    [anon_sym_rand] = ACTIONS(2942),
    [anon_sym_rmdir] = ACTIONS(2942),
    [anon_sym_readdir] = ACTIONS(2942),
    [anon_sym_readline] = ACTIONS(2942),
    [anon_sym_readpipe] = ACTIONS(2942),
    [anon_sym_rewinddir] = ACTIONS(2942),
    [anon_sym_readlink] = ACTIONS(2942),
    [anon_sym_ref] = ACTIONS(2942),
    [anon_sym_scalar] = ACTIONS(2942),
    [anon_sym_shift] = ACTIONS(2942),
    [anon_sym_sin] = ACTIONS(2942),
    [anon_sym_sleep] = ACTIONS(2942),
    [anon_sym_sqrt] = ACTIONS(2942),
    [anon_sym_srand] = ACTIONS(2942),
    [anon_sym_stat] = ACTIONS(2942),
    [anon_sym_study] = ACTIONS(2942),
    [anon_sym_tell] = ACTIONS(2942),
    [anon_sym_telldir] = ACTIONS(2942),
    [anon_sym_tied] = ACTIONS(2942),
    [anon_sym_uc] = ACTIONS(2942),
    [anon_sym_ucfirst] = ACTIONS(2942),
    [anon_sym_untie] = ACTIONS(2942),
    [anon_sym_umask] = ACTIONS(2942),
    [anon_sym_values] = ACTIONS(2942),
    [anon_sym_write] = ACTIONS(2942),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2942),
    [anon_sym_qq] = ACTIONS(2942),
    [anon_sym_qw] = ACTIONS(2942),
    [anon_sym_qx] = ACTIONS(2942),
    [anon_sym_qr] = ACTIONS(2942),
    [anon_sym_m] = ACTIONS(2942),
    [anon_sym_s] = ACTIONS(2942),
    [anon_sym_tr] = ACTIONS(2942),
    [anon_sym_y] = ACTIONS(2942),
    [anon_sym_if] = ACTIONS(2942),
    [anon_sym_unless] = ACTIONS(2942),
    [anon_sym_while] = ACTIONS(2942),
    [anon_sym_until] = ACTIONS(2942),
    [aux_sym__bareword_token1] = ACTIONS(2942),
    [sym_number] = ACTIONS(2942),
    [anon_sym_true] = ACTIONS(2942),
    [anon_sym_false] = ACTIONS(2942),
    [sym__single_quote] = ACTIONS(2940),
    [sym__double_quote] = ACTIONS(2940),
    [sym__backtick_quote] = ACTIONS(2940),
    [sym__search_slash_quote] = ACTIONS(2940),
    [sym__open_readline_bracket] = ACTIONS(2940),
    [sym__open_fileglob_bracket] = ACTIONS(2940),
    [sym__PERLY_HEREDOC] = ACTIONS(2940),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2940),
    [sym__filetest] = ACTIONS(2940),
  },
  [STATE(1546)] = {
    [sym_heredoc_content] = STATE(1546),
    [sym__identifier] = ACTIONS(2966),
    [anon_sym_LBRACE] = ACTIONS(2966),
    [anon_sym_RBRACE] = ACTIONS(2966),
    [anon_sym_SEMI] = ACTIONS(2966),
    [anon_sym_package] = ACTIONS(2966),
    [anon_sym_class] = ACTIONS(2966),
    [anon_sym_role] = ACTIONS(2966),
    [anon_sym_BUILD] = ACTIONS(2966),
    [anon_sym_ADJUST] = ACTIONS(2966),
    [anon_sym_DOLLAR] = ACTIONS(2966),
    [anon_sym_AT] = ACTIONS(2966),
    [anon_sym_my] = ACTIONS(2966),
    [anon_sym_extended] = ACTIONS(2966),
    [anon_sym_async] = ACTIONS(2966),
    [anon_sym_sub] = ACTIONS(2966),
    [anon_sym_method] = ACTIONS(2966),
    [anon_sym_LPAREN] = ACTIONS(2966),
    [anon_sym_continue] = ACTIONS(2966),
    [anon_sym_state] = ACTIONS(2966),
    [anon_sym_our] = ACTIONS(2966),
    [anon_sym_try] = ACTIONS(2966),
    [anon_sym_defer] = ACTIONS(2966),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2966),
    [anon_sym_LBRACK] = ACTIONS(2966),
    [anon_sym_LT_LT] = ACTIONS(2966),
    [anon_sym_SLASH_SLASH] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(2966),
    [anon_sym_DASH] = ACTIONS(2966),
    [anon_sym_TILDE] = ACTIONS(2966),
    [anon_sym_BANG] = ACTIONS(2966),
    [anon_sym_PLUS_PLUS] = ACTIONS(2966),
    [anon_sym_DASH_DASH] = ACTIONS(2966),
    [anon_sym_BSLASH] = ACTIONS(2964),
    [anon_sym_do] = ACTIONS(2966),
    [anon_sym_eval] = ACTIONS(2966),
    [anon_sym_await] = ACTIONS(2966),
    [anon_sym_field] = ACTIONS(2966),
    [anon_sym_local] = ACTIONS(2966),
    [anon_sym_dynamically] = ACTIONS(2966),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2966),
    [anon_sym_require] = ACTIONS(2966),
    [anon_sym_map] = ACTIONS(2966),
    [anon_sym_grep] = ACTIONS(2966),
    [anon_sym_sort] = ACTIONS(2966),
    [anon_sym_goto] = ACTIONS(2966),
    [anon_sym_return] = ACTIONS(2966),
    [anon_sym_undef] = ACTIONS(2966),
    [anon_sym_PERCENT2] = ACTIONS(2966),
    [anon_sym_AMP2] = ACTIONS(2966),
    [anon_sym_STAR2] = ACTIONS(2966),
    [anon_sym_use] = ACTIONS(2966),
    [anon_sym_no] = ACTIONS(2966),
    [anon_sym_for] = ACTIONS(2966),
    [anon_sym_foreach] = ACTIONS(2966),
    [anon_sym_last] = ACTIONS(2966),
    [anon_sym_next] = ACTIONS(2966),
    [anon_sym_redo] = ACTIONS(2966),
    [anon_sym_BEGIN] = ACTIONS(2966),
    [anon_sym_INIT] = ACTIONS(2966),
    [anon_sym_CHECK] = ACTIONS(2966),
    [anon_sym_UNITCHECK] = ACTIONS(2966),
    [anon_sym_END] = ACTIONS(2966),
    [anon_sym___FILE__] = ACTIONS(2966),
    [anon_sym___LINE__] = ACTIONS(2966),
    [anon_sym___PACKAGE__] = ACTIONS(2966),
    [anon_sym___SUB__] = ACTIONS(2966),
    [anon_sym_break] = ACTIONS(2966),
    [anon_sym_fork] = ACTIONS(2966),
    [anon_sym_getppid] = ACTIONS(2966),
    [anon_sym_time] = ACTIONS(2966),
    [anon_sym_times] = ACTIONS(2966),
    [anon_sym_wait] = ACTIONS(2966),
    [anon_sym_wantarray] = ACTIONS(2966),
    [anon_sym_abs] = ACTIONS(2966),
    [anon_sym_alarm] = ACTIONS(2966),
    [anon_sym_chop] = ACTIONS(2966),
    [anon_sym_chdir] = ACTIONS(2966),
    [anon_sym_close] = ACTIONS(2966),
    [anon_sym_closedir] = ACTIONS(2966),
    [anon_sym_caller] = ACTIONS(2966),
    [anon_sym_chomp] = ACTIONS(2966),
    [anon_sym_chr] = ACTIONS(2966),
    [anon_sym_cos] = ACTIONS(2966),
    [anon_sym_chroot] = ACTIONS(2966),
    [anon_sym_defined] = ACTIONS(2966),
    [anon_sym_delete] = ACTIONS(2966),
    [anon_sym_dbmclose] = ACTIONS(2966),
    [anon_sym_exists] = ACTIONS(2966),
    [anon_sym_exit] = ACTIONS(2966),
    [anon_sym_eof] = ACTIONS(2966),
    [anon_sym_exp] = ACTIONS(2966),
    [anon_sym_each] = ACTIONS(2966),
    [anon_sym_fc] = ACTIONS(2966),
    [anon_sym_fileno] = ACTIONS(2966),
    [anon_sym_gmtime] = ACTIONS(2966),
    [anon_sym_getc] = ACTIONS(2966),
    [anon_sym_getpgrp] = ACTIONS(2966),
    [anon_sym_getprotobyname] = ACTIONS(2966),
    [anon_sym_getpwname] = ACTIONS(2966),
    [anon_sym_getpwuid] = ACTIONS(2966),
    [anon_sym_getpeername] = ACTIONS(2966),
    [anon_sym_getnetbyname] = ACTIONS(2966),
    [anon_sym_getsockname] = ACTIONS(2966),
    [anon_sym_getgrnam] = ACTIONS(2966),
    [anon_sym_getgrgid] = ACTIONS(2966),
    [anon_sym_hex] = ACTIONS(2966),
    [anon_sym_int] = ACTIONS(2966),
    [anon_sym_keys] = ACTIONS(2966),
    [anon_sym_lc] = ACTIONS(2966),
    [anon_sym_lcfirst] = ACTIONS(2966),
    [anon_sym_length] = ACTIONS(2966),
    [anon_sym_localtime] = ACTIONS(2966),
    [anon_sym_log] = ACTIONS(2966),
    [anon_sym_lock] = ACTIONS(2966),
    [anon_sym_lstat] = ACTIONS(2966),
    [anon_sym_oct] = ACTIONS(2966),
    [anon_sym_ord] = ACTIONS(2966),
    [anon_sym_prototype] = ACTIONS(2966),
    [anon_sym_pop] = ACTIONS(2966),
    [anon_sym_pos] = ACTIONS(2966),
    [anon_sym_quotemeta] = ACTIONS(2966),
    [anon_sym_reset] = ACTIONS(2966),
    [anon_sym_rand] = ACTIONS(2966),
    [anon_sym_rmdir] = ACTIONS(2966),
    [anon_sym_readdir] = ACTIONS(2966),
    [anon_sym_readline] = ACTIONS(2966),
    [anon_sym_readpipe] = ACTIONS(2966),
    [anon_sym_rewinddir] = ACTIONS(2966),
    [anon_sym_readlink] = ACTIONS(2966),
    [anon_sym_ref] = ACTIONS(2966),
    [anon_sym_scalar] = ACTIONS(2966),
    [anon_sym_shift] = ACTIONS(2966),
    [anon_sym_sin] = ACTIONS(2966),
    [anon_sym_sleep] = ACTIONS(2966),
    [anon_sym_sqrt] = ACTIONS(2966),
    [anon_sym_srand] = ACTIONS(2966),
    [anon_sym_stat] = ACTIONS(2966),
    [anon_sym_study] = ACTIONS(2966),
    [anon_sym_tell] = ACTIONS(2966),
    [anon_sym_telldir] = ACTIONS(2966),
    [anon_sym_tied] = ACTIONS(2966),
    [anon_sym_uc] = ACTIONS(2966),
    [anon_sym_ucfirst] = ACTIONS(2966),
    [anon_sym_untie] = ACTIONS(2966),
    [anon_sym_umask] = ACTIONS(2966),
    [anon_sym_values] = ACTIONS(2966),
    [anon_sym_write] = ACTIONS(2966),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2966),
    [anon_sym_qq] = ACTIONS(2966),
    [anon_sym_qw] = ACTIONS(2966),
    [anon_sym_qx] = ACTIONS(2966),
    [anon_sym_qr] = ACTIONS(2966),
    [anon_sym_m] = ACTIONS(2966),
    [anon_sym_s] = ACTIONS(2966),
    [anon_sym_tr] = ACTIONS(2966),
    [anon_sym_y] = ACTIONS(2966),
    [anon_sym_if] = ACTIONS(2966),
    [anon_sym_unless] = ACTIONS(2966),
    [anon_sym_while] = ACTIONS(2966),
    [anon_sym_until] = ACTIONS(2966),
    [aux_sym__bareword_token1] = ACTIONS(2966),
    [sym_number] = ACTIONS(2966),
    [anon_sym_true] = ACTIONS(2966),
    [anon_sym_false] = ACTIONS(2966),
    [sym__single_quote] = ACTIONS(2964),
    [sym__double_quote] = ACTIONS(2964),
    [sym__backtick_quote] = ACTIONS(2964),
    [sym__search_slash_quote] = ACTIONS(2964),
    [sym__open_readline_bracket] = ACTIONS(2964),
    [sym__open_fileglob_bracket] = ACTIONS(2964),
    [sym__PERLY_HEREDOC] = ACTIONS(2964),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2964),
    [sym__filetest] = ACTIONS(2964),
  },
  [STATE(1547)] = {
    [sym_heredoc_content] = STATE(1547),
    [sym__identifier] = ACTIONS(2970),
    [anon_sym_LBRACE] = ACTIONS(2970),
    [anon_sym_RBRACE] = ACTIONS(2970),
    [anon_sym_SEMI] = ACTIONS(2970),
    [anon_sym_package] = ACTIONS(2970),
    [anon_sym_class] = ACTIONS(2970),
    [anon_sym_role] = ACTIONS(2970),
    [anon_sym_BUILD] = ACTIONS(2970),
    [anon_sym_ADJUST] = ACTIONS(2970),
    [anon_sym_DOLLAR] = ACTIONS(2970),
    [anon_sym_AT] = ACTIONS(2970),
    [anon_sym_my] = ACTIONS(2970),
    [anon_sym_extended] = ACTIONS(2970),
    [anon_sym_async] = ACTIONS(2970),
    [anon_sym_sub] = ACTIONS(2970),
    [anon_sym_method] = ACTIONS(2970),
    [anon_sym_LPAREN] = ACTIONS(2970),
    [anon_sym_continue] = ACTIONS(2970),
    [anon_sym_state] = ACTIONS(2970),
    [anon_sym_our] = ACTIONS(2970),
    [anon_sym_try] = ACTIONS(2970),
    [anon_sym_defer] = ACTIONS(2970),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2970),
    [anon_sym_LBRACK] = ACTIONS(2970),
    [anon_sym_LT_LT] = ACTIONS(2970),
    [anon_sym_SLASH_SLASH] = ACTIONS(2970),
    [anon_sym_PLUS] = ACTIONS(2970),
    [anon_sym_DASH] = ACTIONS(2970),
    [anon_sym_TILDE] = ACTIONS(2970),
    [anon_sym_BANG] = ACTIONS(2970),
    [anon_sym_PLUS_PLUS] = ACTIONS(2970),
    [anon_sym_DASH_DASH] = ACTIONS(2970),
    [anon_sym_BSLASH] = ACTIONS(2968),
    [anon_sym_do] = ACTIONS(2970),
    [anon_sym_eval] = ACTIONS(2970),
    [anon_sym_await] = ACTIONS(2970),
    [anon_sym_field] = ACTIONS(2970),
    [anon_sym_local] = ACTIONS(2970),
    [anon_sym_dynamically] = ACTIONS(2970),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2970),
    [anon_sym_require] = ACTIONS(2970),
    [anon_sym_map] = ACTIONS(2970),
    [anon_sym_grep] = ACTIONS(2970),
    [anon_sym_sort] = ACTIONS(2970),
    [anon_sym_goto] = ACTIONS(2970),
    [anon_sym_return] = ACTIONS(2970),
    [anon_sym_undef] = ACTIONS(2970),
    [anon_sym_PERCENT2] = ACTIONS(2970),
    [anon_sym_AMP2] = ACTIONS(2970),
    [anon_sym_STAR2] = ACTIONS(2970),
    [anon_sym_use] = ACTIONS(2970),
    [anon_sym_no] = ACTIONS(2970),
    [anon_sym_for] = ACTIONS(2970),
    [anon_sym_foreach] = ACTIONS(2970),
    [anon_sym_last] = ACTIONS(2970),
    [anon_sym_next] = ACTIONS(2970),
    [anon_sym_redo] = ACTIONS(2970),
    [anon_sym_BEGIN] = ACTIONS(2970),
    [anon_sym_INIT] = ACTIONS(2970),
    [anon_sym_CHECK] = ACTIONS(2970),
    [anon_sym_UNITCHECK] = ACTIONS(2970),
    [anon_sym_END] = ACTIONS(2970),
    [anon_sym___FILE__] = ACTIONS(2970),
    [anon_sym___LINE__] = ACTIONS(2970),
    [anon_sym___PACKAGE__] = ACTIONS(2970),
    [anon_sym___SUB__] = ACTIONS(2970),
    [anon_sym_break] = ACTIONS(2970),
    [anon_sym_fork] = ACTIONS(2970),
    [anon_sym_getppid] = ACTIONS(2970),
    [anon_sym_time] = ACTIONS(2970),
    [anon_sym_times] = ACTIONS(2970),
    [anon_sym_wait] = ACTIONS(2970),
    [anon_sym_wantarray] = ACTIONS(2970),
    [anon_sym_abs] = ACTIONS(2970),
    [anon_sym_alarm] = ACTIONS(2970),
    [anon_sym_chop] = ACTIONS(2970),
    [anon_sym_chdir] = ACTIONS(2970),
    [anon_sym_close] = ACTIONS(2970),
    [anon_sym_closedir] = ACTIONS(2970),
    [anon_sym_caller] = ACTIONS(2970),
    [anon_sym_chomp] = ACTIONS(2970),
    [anon_sym_chr] = ACTIONS(2970),
    [anon_sym_cos] = ACTIONS(2970),
    [anon_sym_chroot] = ACTIONS(2970),
    [anon_sym_defined] = ACTIONS(2970),
    [anon_sym_delete] = ACTIONS(2970),
    [anon_sym_dbmclose] = ACTIONS(2970),
    [anon_sym_exists] = ACTIONS(2970),
    [anon_sym_exit] = ACTIONS(2970),
    [anon_sym_eof] = ACTIONS(2970),
    [anon_sym_exp] = ACTIONS(2970),
    [anon_sym_each] = ACTIONS(2970),
    [anon_sym_fc] = ACTIONS(2970),
    [anon_sym_fileno] = ACTIONS(2970),
    [anon_sym_gmtime] = ACTIONS(2970),
    [anon_sym_getc] = ACTIONS(2970),
    [anon_sym_getpgrp] = ACTIONS(2970),
    [anon_sym_getprotobyname] = ACTIONS(2970),
    [anon_sym_getpwname] = ACTIONS(2970),
    [anon_sym_getpwuid] = ACTIONS(2970),
    [anon_sym_getpeername] = ACTIONS(2970),
    [anon_sym_getnetbyname] = ACTIONS(2970),
    [anon_sym_getsockname] = ACTIONS(2970),
    [anon_sym_getgrnam] = ACTIONS(2970),
    [anon_sym_getgrgid] = ACTIONS(2970),
    [anon_sym_hex] = ACTIONS(2970),
    [anon_sym_int] = ACTIONS(2970),
    [anon_sym_keys] = ACTIONS(2970),
    [anon_sym_lc] = ACTIONS(2970),
    [anon_sym_lcfirst] = ACTIONS(2970),
    [anon_sym_length] = ACTIONS(2970),
    [anon_sym_localtime] = ACTIONS(2970),
    [anon_sym_log] = ACTIONS(2970),
    [anon_sym_lock] = ACTIONS(2970),
    [anon_sym_lstat] = ACTIONS(2970),
    [anon_sym_oct] = ACTIONS(2970),
    [anon_sym_ord] = ACTIONS(2970),
    [anon_sym_prototype] = ACTIONS(2970),
    [anon_sym_pop] = ACTIONS(2970),
    [anon_sym_pos] = ACTIONS(2970),
    [anon_sym_quotemeta] = ACTIONS(2970),
    [anon_sym_reset] = ACTIONS(2970),
    [anon_sym_rand] = ACTIONS(2970),
    [anon_sym_rmdir] = ACTIONS(2970),
    [anon_sym_readdir] = ACTIONS(2970),
    [anon_sym_readline] = ACTIONS(2970),
    [anon_sym_readpipe] = ACTIONS(2970),
    [anon_sym_rewinddir] = ACTIONS(2970),
    [anon_sym_readlink] = ACTIONS(2970),
    [anon_sym_ref] = ACTIONS(2970),
    [anon_sym_scalar] = ACTIONS(2970),
    [anon_sym_shift] = ACTIONS(2970),
    [anon_sym_sin] = ACTIONS(2970),
    [anon_sym_sleep] = ACTIONS(2970),
    [anon_sym_sqrt] = ACTIONS(2970),
    [anon_sym_srand] = ACTIONS(2970),
    [anon_sym_stat] = ACTIONS(2970),
    [anon_sym_study] = ACTIONS(2970),
    [anon_sym_tell] = ACTIONS(2970),
    [anon_sym_telldir] = ACTIONS(2970),
    [anon_sym_tied] = ACTIONS(2970),
    [anon_sym_uc] = ACTIONS(2970),
    [anon_sym_ucfirst] = ACTIONS(2970),
    [anon_sym_untie] = ACTIONS(2970),
    [anon_sym_umask] = ACTIONS(2970),
    [anon_sym_values] = ACTIONS(2970),
    [anon_sym_write] = ACTIONS(2970),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2970),
    [anon_sym_qq] = ACTIONS(2970),
    [anon_sym_qw] = ACTIONS(2970),
    [anon_sym_qx] = ACTIONS(2970),
    [anon_sym_qr] = ACTIONS(2970),
    [anon_sym_m] = ACTIONS(2970),
    [anon_sym_s] = ACTIONS(2970),
    [anon_sym_tr] = ACTIONS(2970),
    [anon_sym_y] = ACTIONS(2970),
    [anon_sym_if] = ACTIONS(2970),
    [anon_sym_unless] = ACTIONS(2970),
    [anon_sym_while] = ACTIONS(2970),
    [anon_sym_until] = ACTIONS(2970),
    [aux_sym__bareword_token1] = ACTIONS(2970),
    [sym_number] = ACTIONS(2970),
    [anon_sym_true] = ACTIONS(2970),
    [anon_sym_false] = ACTIONS(2970),
    [sym__single_quote] = ACTIONS(2968),
    [sym__double_quote] = ACTIONS(2968),
    [sym__backtick_quote] = ACTIONS(2968),
    [sym__search_slash_quote] = ACTIONS(2968),
    [sym__open_readline_bracket] = ACTIONS(2968),
    [sym__open_fileglob_bracket] = ACTIONS(2968),
    [sym__PERLY_HEREDOC] = ACTIONS(2968),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2968),
    [sym__filetest] = ACTIONS(2968),
  },
  [STATE(1548)] = {
    [sym_heredoc_content] = STATE(1548),
    [sym__identifier] = ACTIONS(2974),
    [anon_sym_LBRACE] = ACTIONS(2974),
    [anon_sym_RBRACE] = ACTIONS(2974),
    [anon_sym_SEMI] = ACTIONS(2974),
    [anon_sym_package] = ACTIONS(2974),
    [anon_sym_class] = ACTIONS(2974),
    [anon_sym_role] = ACTIONS(2974),
    [anon_sym_BUILD] = ACTIONS(2974),
    [anon_sym_ADJUST] = ACTIONS(2974),
    [anon_sym_DOLLAR] = ACTIONS(2974),
    [anon_sym_AT] = ACTIONS(2974),
    [anon_sym_my] = ACTIONS(2974),
    [anon_sym_extended] = ACTIONS(2974),
    [anon_sym_async] = ACTIONS(2974),
    [anon_sym_sub] = ACTIONS(2974),
    [anon_sym_method] = ACTIONS(2974),
    [anon_sym_LPAREN] = ACTIONS(2974),
    [anon_sym_continue] = ACTIONS(2974),
    [anon_sym_state] = ACTIONS(2974),
    [anon_sym_our] = ACTIONS(2974),
    [anon_sym_try] = ACTIONS(2974),
    [anon_sym_defer] = ACTIONS(2974),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2974),
    [anon_sym_LBRACK] = ACTIONS(2974),
    [anon_sym_LT_LT] = ACTIONS(2974),
    [anon_sym_SLASH_SLASH] = ACTIONS(2974),
    [anon_sym_PLUS] = ACTIONS(2974),
    [anon_sym_DASH] = ACTIONS(2974),
    [anon_sym_TILDE] = ACTIONS(2974),
    [anon_sym_BANG] = ACTIONS(2974),
    [anon_sym_PLUS_PLUS] = ACTIONS(2974),
    [anon_sym_DASH_DASH] = ACTIONS(2974),
    [anon_sym_BSLASH] = ACTIONS(2972),
    [anon_sym_do] = ACTIONS(2974),
    [anon_sym_eval] = ACTIONS(2974),
    [anon_sym_await] = ACTIONS(2974),
    [anon_sym_field] = ACTIONS(2974),
    [anon_sym_local] = ACTIONS(2974),
    [anon_sym_dynamically] = ACTIONS(2974),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2974),
    [anon_sym_require] = ACTIONS(2974),
    [anon_sym_map] = ACTIONS(2974),
    [anon_sym_grep] = ACTIONS(2974),
    [anon_sym_sort] = ACTIONS(2974),
    [anon_sym_goto] = ACTIONS(2974),
    [anon_sym_return] = ACTIONS(2974),
    [anon_sym_undef] = ACTIONS(2974),
    [anon_sym_PERCENT2] = ACTIONS(2974),
    [anon_sym_AMP2] = ACTIONS(2974),
    [anon_sym_STAR2] = ACTIONS(2974),
    [anon_sym_use] = ACTIONS(2974),
    [anon_sym_no] = ACTIONS(2974),
    [anon_sym_for] = ACTIONS(2974),
    [anon_sym_foreach] = ACTIONS(2974),
    [anon_sym_last] = ACTIONS(2974),
    [anon_sym_next] = ACTIONS(2974),
    [anon_sym_redo] = ACTIONS(2974),
    [anon_sym_BEGIN] = ACTIONS(2974),
    [anon_sym_INIT] = ACTIONS(2974),
    [anon_sym_CHECK] = ACTIONS(2974),
    [anon_sym_UNITCHECK] = ACTIONS(2974),
    [anon_sym_END] = ACTIONS(2974),
    [anon_sym___FILE__] = ACTIONS(2974),
    [anon_sym___LINE__] = ACTIONS(2974),
    [anon_sym___PACKAGE__] = ACTIONS(2974),
    [anon_sym___SUB__] = ACTIONS(2974),
    [anon_sym_break] = ACTIONS(2974),
    [anon_sym_fork] = ACTIONS(2974),
    [anon_sym_getppid] = ACTIONS(2974),
    [anon_sym_time] = ACTIONS(2974),
    [anon_sym_times] = ACTIONS(2974),
    [anon_sym_wait] = ACTIONS(2974),
    [anon_sym_wantarray] = ACTIONS(2974),
    [anon_sym_abs] = ACTIONS(2974),
    [anon_sym_alarm] = ACTIONS(2974),
    [anon_sym_chop] = ACTIONS(2974),
    [anon_sym_chdir] = ACTIONS(2974),
    [anon_sym_close] = ACTIONS(2974),
    [anon_sym_closedir] = ACTIONS(2974),
    [anon_sym_caller] = ACTIONS(2974),
    [anon_sym_chomp] = ACTIONS(2974),
    [anon_sym_chr] = ACTIONS(2974),
    [anon_sym_cos] = ACTIONS(2974),
    [anon_sym_chroot] = ACTIONS(2974),
    [anon_sym_defined] = ACTIONS(2974),
    [anon_sym_delete] = ACTIONS(2974),
    [anon_sym_dbmclose] = ACTIONS(2974),
    [anon_sym_exists] = ACTIONS(2974),
    [anon_sym_exit] = ACTIONS(2974),
    [anon_sym_eof] = ACTIONS(2974),
    [anon_sym_exp] = ACTIONS(2974),
    [anon_sym_each] = ACTIONS(2974),
    [anon_sym_fc] = ACTIONS(2974),
    [anon_sym_fileno] = ACTIONS(2974),
    [anon_sym_gmtime] = ACTIONS(2974),
    [anon_sym_getc] = ACTIONS(2974),
    [anon_sym_getpgrp] = ACTIONS(2974),
    [anon_sym_getprotobyname] = ACTIONS(2974),
    [anon_sym_getpwname] = ACTIONS(2974),
    [anon_sym_getpwuid] = ACTIONS(2974),
    [anon_sym_getpeername] = ACTIONS(2974),
    [anon_sym_getnetbyname] = ACTIONS(2974),
    [anon_sym_getsockname] = ACTIONS(2974),
    [anon_sym_getgrnam] = ACTIONS(2974),
    [anon_sym_getgrgid] = ACTIONS(2974),
    [anon_sym_hex] = ACTIONS(2974),
    [anon_sym_int] = ACTIONS(2974),
    [anon_sym_keys] = ACTIONS(2974),
    [anon_sym_lc] = ACTIONS(2974),
    [anon_sym_lcfirst] = ACTIONS(2974),
    [anon_sym_length] = ACTIONS(2974),
    [anon_sym_localtime] = ACTIONS(2974),
    [anon_sym_log] = ACTIONS(2974),
    [anon_sym_lock] = ACTIONS(2974),
    [anon_sym_lstat] = ACTIONS(2974),
    [anon_sym_oct] = ACTIONS(2974),
    [anon_sym_ord] = ACTIONS(2974),
    [anon_sym_prototype] = ACTIONS(2974),
    [anon_sym_pop] = ACTIONS(2974),
    [anon_sym_pos] = ACTIONS(2974),
    [anon_sym_quotemeta] = ACTIONS(2974),
    [anon_sym_reset] = ACTIONS(2974),
    [anon_sym_rand] = ACTIONS(2974),
    [anon_sym_rmdir] = ACTIONS(2974),
    [anon_sym_readdir] = ACTIONS(2974),
    [anon_sym_readline] = ACTIONS(2974),
    [anon_sym_readpipe] = ACTIONS(2974),
    [anon_sym_rewinddir] = ACTIONS(2974),
    [anon_sym_readlink] = ACTIONS(2974),
    [anon_sym_ref] = ACTIONS(2974),
    [anon_sym_scalar] = ACTIONS(2974),
    [anon_sym_shift] = ACTIONS(2974),
    [anon_sym_sin] = ACTIONS(2974),
    [anon_sym_sleep] = ACTIONS(2974),
    [anon_sym_sqrt] = ACTIONS(2974),
    [anon_sym_srand] = ACTIONS(2974),
    [anon_sym_stat] = ACTIONS(2974),
    [anon_sym_study] = ACTIONS(2974),
    [anon_sym_tell] = ACTIONS(2974),
    [anon_sym_telldir] = ACTIONS(2974),
    [anon_sym_tied] = ACTIONS(2974),
    [anon_sym_uc] = ACTIONS(2974),
    [anon_sym_ucfirst] = ACTIONS(2974),
    [anon_sym_untie] = ACTIONS(2974),
    [anon_sym_umask] = ACTIONS(2974),
    [anon_sym_values] = ACTIONS(2974),
    [anon_sym_write] = ACTIONS(2974),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2974),
    [anon_sym_qq] = ACTIONS(2974),
    [anon_sym_qw] = ACTIONS(2974),
    [anon_sym_qx] = ACTIONS(2974),
    [anon_sym_qr] = ACTIONS(2974),
    [anon_sym_m] = ACTIONS(2974),
    [anon_sym_s] = ACTIONS(2974),
    [anon_sym_tr] = ACTIONS(2974),
    [anon_sym_y] = ACTIONS(2974),
    [anon_sym_if] = ACTIONS(2974),
    [anon_sym_unless] = ACTIONS(2974),
    [anon_sym_while] = ACTIONS(2974),
    [anon_sym_until] = ACTIONS(2974),
    [aux_sym__bareword_token1] = ACTIONS(2974),
    [sym_number] = ACTIONS(2974),
    [anon_sym_true] = ACTIONS(2974),
    [anon_sym_false] = ACTIONS(2974),
    [sym__single_quote] = ACTIONS(2972),
    [sym__double_quote] = ACTIONS(2972),
    [sym__backtick_quote] = ACTIONS(2972),
    [sym__search_slash_quote] = ACTIONS(2972),
    [sym__open_readline_bracket] = ACTIONS(2972),
    [sym__open_fileglob_bracket] = ACTIONS(2972),
    [sym__PERLY_HEREDOC] = ACTIONS(2972),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2972),
    [sym__filetest] = ACTIONS(2972),
  },
  [STATE(1549)] = {
    [sym_heredoc_content] = STATE(1549),
    [sym__identifier] = ACTIONS(2978),
    [anon_sym_LBRACE] = ACTIONS(2978),
    [anon_sym_RBRACE] = ACTIONS(2978),
    [anon_sym_SEMI] = ACTIONS(2978),
    [anon_sym_package] = ACTIONS(2978),
    [anon_sym_class] = ACTIONS(2978),
    [anon_sym_role] = ACTIONS(2978),
    [anon_sym_BUILD] = ACTIONS(2978),
    [anon_sym_ADJUST] = ACTIONS(2978),
    [anon_sym_DOLLAR] = ACTIONS(2978),
    [anon_sym_AT] = ACTIONS(2978),
    [anon_sym_my] = ACTIONS(2978),
    [anon_sym_extended] = ACTIONS(2978),
    [anon_sym_async] = ACTIONS(2978),
    [anon_sym_sub] = ACTIONS(2978),
    [anon_sym_method] = ACTIONS(2978),
    [anon_sym_LPAREN] = ACTIONS(2978),
    [anon_sym_continue] = ACTIONS(2978),
    [anon_sym_state] = ACTIONS(2978),
    [anon_sym_our] = ACTIONS(2978),
    [anon_sym_try] = ACTIONS(2978),
    [anon_sym_defer] = ACTIONS(2978),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2978),
    [anon_sym_LBRACK] = ACTIONS(2978),
    [anon_sym_LT_LT] = ACTIONS(2978),
    [anon_sym_SLASH_SLASH] = ACTIONS(2978),
    [anon_sym_PLUS] = ACTIONS(2978),
    [anon_sym_DASH] = ACTIONS(2978),
    [anon_sym_TILDE] = ACTIONS(2978),
    [anon_sym_BANG] = ACTIONS(2978),
    [anon_sym_PLUS_PLUS] = ACTIONS(2978),
    [anon_sym_DASH_DASH] = ACTIONS(2978),
    [anon_sym_BSLASH] = ACTIONS(2976),
    [anon_sym_do] = ACTIONS(2978),
    [anon_sym_eval] = ACTIONS(2978),
    [anon_sym_await] = ACTIONS(2978),
    [anon_sym_field] = ACTIONS(2978),
    [anon_sym_local] = ACTIONS(2978),
    [anon_sym_dynamically] = ACTIONS(2978),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2978),
    [anon_sym_require] = ACTIONS(2978),
    [anon_sym_map] = ACTIONS(2978),
    [anon_sym_grep] = ACTIONS(2978),
    [anon_sym_sort] = ACTIONS(2978),
    [anon_sym_goto] = ACTIONS(2978),
    [anon_sym_return] = ACTIONS(2978),
    [anon_sym_undef] = ACTIONS(2978),
    [anon_sym_PERCENT2] = ACTIONS(2978),
    [anon_sym_AMP2] = ACTIONS(2978),
    [anon_sym_STAR2] = ACTIONS(2978),
    [anon_sym_use] = ACTIONS(2978),
    [anon_sym_no] = ACTIONS(2978),
    [anon_sym_for] = ACTIONS(2978),
    [anon_sym_foreach] = ACTIONS(2978),
    [anon_sym_last] = ACTIONS(2978),
    [anon_sym_next] = ACTIONS(2978),
    [anon_sym_redo] = ACTIONS(2978),
    [anon_sym_BEGIN] = ACTIONS(2978),
    [anon_sym_INIT] = ACTIONS(2978),
    [anon_sym_CHECK] = ACTIONS(2978),
    [anon_sym_UNITCHECK] = ACTIONS(2978),
    [anon_sym_END] = ACTIONS(2978),
    [anon_sym___FILE__] = ACTIONS(2978),
    [anon_sym___LINE__] = ACTIONS(2978),
    [anon_sym___PACKAGE__] = ACTIONS(2978),
    [anon_sym___SUB__] = ACTIONS(2978),
    [anon_sym_break] = ACTIONS(2978),
    [anon_sym_fork] = ACTIONS(2978),
    [anon_sym_getppid] = ACTIONS(2978),
    [anon_sym_time] = ACTIONS(2978),
    [anon_sym_times] = ACTIONS(2978),
    [anon_sym_wait] = ACTIONS(2978),
    [anon_sym_wantarray] = ACTIONS(2978),
    [anon_sym_abs] = ACTIONS(2978),
    [anon_sym_alarm] = ACTIONS(2978),
    [anon_sym_chop] = ACTIONS(2978),
    [anon_sym_chdir] = ACTIONS(2978),
    [anon_sym_close] = ACTIONS(2978),
    [anon_sym_closedir] = ACTIONS(2978),
    [anon_sym_caller] = ACTIONS(2978),
    [anon_sym_chomp] = ACTIONS(2978),
    [anon_sym_chr] = ACTIONS(2978),
    [anon_sym_cos] = ACTIONS(2978),
    [anon_sym_chroot] = ACTIONS(2978),
    [anon_sym_defined] = ACTIONS(2978),
    [anon_sym_delete] = ACTIONS(2978),
    [anon_sym_dbmclose] = ACTIONS(2978),
    [anon_sym_exists] = ACTIONS(2978),
    [anon_sym_exit] = ACTIONS(2978),
    [anon_sym_eof] = ACTIONS(2978),
    [anon_sym_exp] = ACTIONS(2978),
    [anon_sym_each] = ACTIONS(2978),
    [anon_sym_fc] = ACTIONS(2978),
    [anon_sym_fileno] = ACTIONS(2978),
    [anon_sym_gmtime] = ACTIONS(2978),
    [anon_sym_getc] = ACTIONS(2978),
    [anon_sym_getpgrp] = ACTIONS(2978),
    [anon_sym_getprotobyname] = ACTIONS(2978),
    [anon_sym_getpwname] = ACTIONS(2978),
    [anon_sym_getpwuid] = ACTIONS(2978),
    [anon_sym_getpeername] = ACTIONS(2978),
    [anon_sym_getnetbyname] = ACTIONS(2978),
    [anon_sym_getsockname] = ACTIONS(2978),
    [anon_sym_getgrnam] = ACTIONS(2978),
    [anon_sym_getgrgid] = ACTIONS(2978),
    [anon_sym_hex] = ACTIONS(2978),
    [anon_sym_int] = ACTIONS(2978),
    [anon_sym_keys] = ACTIONS(2978),
    [anon_sym_lc] = ACTIONS(2978),
    [anon_sym_lcfirst] = ACTIONS(2978),
    [anon_sym_length] = ACTIONS(2978),
    [anon_sym_localtime] = ACTIONS(2978),
    [anon_sym_log] = ACTIONS(2978),
    [anon_sym_lock] = ACTIONS(2978),
    [anon_sym_lstat] = ACTIONS(2978),
    [anon_sym_oct] = ACTIONS(2978),
    [anon_sym_ord] = ACTIONS(2978),
    [anon_sym_prototype] = ACTIONS(2978),
    [anon_sym_pop] = ACTIONS(2978),
    [anon_sym_pos] = ACTIONS(2978),
    [anon_sym_quotemeta] = ACTIONS(2978),
    [anon_sym_reset] = ACTIONS(2978),
    [anon_sym_rand] = ACTIONS(2978),
    [anon_sym_rmdir] = ACTIONS(2978),
    [anon_sym_readdir] = ACTIONS(2978),
    [anon_sym_readline] = ACTIONS(2978),
    [anon_sym_readpipe] = ACTIONS(2978),
    [anon_sym_rewinddir] = ACTIONS(2978),
    [anon_sym_readlink] = ACTIONS(2978),
    [anon_sym_ref] = ACTIONS(2978),
    [anon_sym_scalar] = ACTIONS(2978),
    [anon_sym_shift] = ACTIONS(2978),
    [anon_sym_sin] = ACTIONS(2978),
    [anon_sym_sleep] = ACTIONS(2978),
    [anon_sym_sqrt] = ACTIONS(2978),
    [anon_sym_srand] = ACTIONS(2978),
    [anon_sym_stat] = ACTIONS(2978),
    [anon_sym_study] = ACTIONS(2978),
    [anon_sym_tell] = ACTIONS(2978),
    [anon_sym_telldir] = ACTIONS(2978),
    [anon_sym_tied] = ACTIONS(2978),
    [anon_sym_uc] = ACTIONS(2978),
    [anon_sym_ucfirst] = ACTIONS(2978),
    [anon_sym_untie] = ACTIONS(2978),
    [anon_sym_umask] = ACTIONS(2978),
    [anon_sym_values] = ACTIONS(2978),
    [anon_sym_write] = ACTIONS(2978),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2978),
    [anon_sym_qq] = ACTIONS(2978),
    [anon_sym_qw] = ACTIONS(2978),
    [anon_sym_qx] = ACTIONS(2978),
    [anon_sym_qr] = ACTIONS(2978),
    [anon_sym_m] = ACTIONS(2978),
    [anon_sym_s] = ACTIONS(2978),
    [anon_sym_tr] = ACTIONS(2978),
    [anon_sym_y] = ACTIONS(2978),
    [anon_sym_if] = ACTIONS(2978),
    [anon_sym_unless] = ACTIONS(2978),
    [anon_sym_while] = ACTIONS(2978),
    [anon_sym_until] = ACTIONS(2978),
    [aux_sym__bareword_token1] = ACTIONS(2978),
    [sym_number] = ACTIONS(2978),
    [anon_sym_true] = ACTIONS(2978),
    [anon_sym_false] = ACTIONS(2978),
    [sym__single_quote] = ACTIONS(2976),
    [sym__double_quote] = ACTIONS(2976),
    [sym__backtick_quote] = ACTIONS(2976),
    [sym__search_slash_quote] = ACTIONS(2976),
    [sym__open_readline_bracket] = ACTIONS(2976),
    [sym__open_fileglob_bracket] = ACTIONS(2976),
    [sym__PERLY_HEREDOC] = ACTIONS(2976),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2976),
    [sym__filetest] = ACTIONS(2976),
  },
  [STATE(1550)] = {
    [sym_heredoc_content] = STATE(1550),
    [sym__identifier] = ACTIONS(2982),
    [anon_sym_LBRACE] = ACTIONS(2982),
    [anon_sym_RBRACE] = ACTIONS(2982),
    [anon_sym_SEMI] = ACTIONS(2982),
    [anon_sym_package] = ACTIONS(2982),
    [anon_sym_class] = ACTIONS(2982),
    [anon_sym_role] = ACTIONS(2982),
    [anon_sym_BUILD] = ACTIONS(2982),
    [anon_sym_ADJUST] = ACTIONS(2982),
    [anon_sym_DOLLAR] = ACTIONS(2982),
    [anon_sym_AT] = ACTIONS(2982),
    [anon_sym_my] = ACTIONS(2982),
    [anon_sym_extended] = ACTIONS(2982),
    [anon_sym_async] = ACTIONS(2982),
    [anon_sym_sub] = ACTIONS(2982),
    [anon_sym_method] = ACTIONS(2982),
    [anon_sym_LPAREN] = ACTIONS(2982),
    [anon_sym_continue] = ACTIONS(2982),
    [anon_sym_state] = ACTIONS(2982),
    [anon_sym_our] = ACTIONS(2982),
    [anon_sym_try] = ACTIONS(2982),
    [anon_sym_defer] = ACTIONS(2982),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2982),
    [anon_sym_LBRACK] = ACTIONS(2982),
    [anon_sym_LT_LT] = ACTIONS(2982),
    [anon_sym_SLASH_SLASH] = ACTIONS(2982),
    [anon_sym_PLUS] = ACTIONS(2982),
    [anon_sym_DASH] = ACTIONS(2982),
    [anon_sym_TILDE] = ACTIONS(2982),
    [anon_sym_BANG] = ACTIONS(2982),
    [anon_sym_PLUS_PLUS] = ACTIONS(2982),
    [anon_sym_DASH_DASH] = ACTIONS(2982),
    [anon_sym_BSLASH] = ACTIONS(2980),
    [anon_sym_do] = ACTIONS(2982),
    [anon_sym_eval] = ACTIONS(2982),
    [anon_sym_await] = ACTIONS(2982),
    [anon_sym_field] = ACTIONS(2982),
    [anon_sym_local] = ACTIONS(2982),
    [anon_sym_dynamically] = ACTIONS(2982),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2982),
    [anon_sym_require] = ACTIONS(2982),
    [anon_sym_map] = ACTIONS(2982),
    [anon_sym_grep] = ACTIONS(2982),
    [anon_sym_sort] = ACTIONS(2982),
    [anon_sym_goto] = ACTIONS(2982),
    [anon_sym_return] = ACTIONS(2982),
    [anon_sym_undef] = ACTIONS(2982),
    [anon_sym_PERCENT2] = ACTIONS(2982),
    [anon_sym_AMP2] = ACTIONS(2982),
    [anon_sym_STAR2] = ACTIONS(2982),
    [anon_sym_use] = ACTIONS(2982),
    [anon_sym_no] = ACTIONS(2982),
    [anon_sym_for] = ACTIONS(2982),
    [anon_sym_foreach] = ACTIONS(2982),
    [anon_sym_last] = ACTIONS(2982),
    [anon_sym_next] = ACTIONS(2982),
    [anon_sym_redo] = ACTIONS(2982),
    [anon_sym_BEGIN] = ACTIONS(2982),
    [anon_sym_INIT] = ACTIONS(2982),
    [anon_sym_CHECK] = ACTIONS(2982),
    [anon_sym_UNITCHECK] = ACTIONS(2982),
    [anon_sym_END] = ACTIONS(2982),
    [anon_sym___FILE__] = ACTIONS(2982),
    [anon_sym___LINE__] = ACTIONS(2982),
    [anon_sym___PACKAGE__] = ACTIONS(2982),
    [anon_sym___SUB__] = ACTIONS(2982),
    [anon_sym_break] = ACTIONS(2982),
    [anon_sym_fork] = ACTIONS(2982),
    [anon_sym_getppid] = ACTIONS(2982),
    [anon_sym_time] = ACTIONS(2982),
    [anon_sym_times] = ACTIONS(2982),
    [anon_sym_wait] = ACTIONS(2982),
    [anon_sym_wantarray] = ACTIONS(2982),
    [anon_sym_abs] = ACTIONS(2982),
    [anon_sym_alarm] = ACTIONS(2982),
    [anon_sym_chop] = ACTIONS(2982),
    [anon_sym_chdir] = ACTIONS(2982),
    [anon_sym_close] = ACTIONS(2982),
    [anon_sym_closedir] = ACTIONS(2982),
    [anon_sym_caller] = ACTIONS(2982),
    [anon_sym_chomp] = ACTIONS(2982),
    [anon_sym_chr] = ACTIONS(2982),
    [anon_sym_cos] = ACTIONS(2982),
    [anon_sym_chroot] = ACTIONS(2982),
    [anon_sym_defined] = ACTIONS(2982),
    [anon_sym_delete] = ACTIONS(2982),
    [anon_sym_dbmclose] = ACTIONS(2982),
    [anon_sym_exists] = ACTIONS(2982),
    [anon_sym_exit] = ACTIONS(2982),
    [anon_sym_eof] = ACTIONS(2982),
    [anon_sym_exp] = ACTIONS(2982),
    [anon_sym_each] = ACTIONS(2982),
    [anon_sym_fc] = ACTIONS(2982),
    [anon_sym_fileno] = ACTIONS(2982),
    [anon_sym_gmtime] = ACTIONS(2982),
    [anon_sym_getc] = ACTIONS(2982),
    [anon_sym_getpgrp] = ACTIONS(2982),
    [anon_sym_getprotobyname] = ACTIONS(2982),
    [anon_sym_getpwname] = ACTIONS(2982),
    [anon_sym_getpwuid] = ACTIONS(2982),
    [anon_sym_getpeername] = ACTIONS(2982),
    [anon_sym_getnetbyname] = ACTIONS(2982),
    [anon_sym_getsockname] = ACTIONS(2982),
    [anon_sym_getgrnam] = ACTIONS(2982),
    [anon_sym_getgrgid] = ACTIONS(2982),
    [anon_sym_hex] = ACTIONS(2982),
    [anon_sym_int] = ACTIONS(2982),
    [anon_sym_keys] = ACTIONS(2982),
    [anon_sym_lc] = ACTIONS(2982),
    [anon_sym_lcfirst] = ACTIONS(2982),
    [anon_sym_length] = ACTIONS(2982),
    [anon_sym_localtime] = ACTIONS(2982),
    [anon_sym_log] = ACTIONS(2982),
    [anon_sym_lock] = ACTIONS(2982),
    [anon_sym_lstat] = ACTIONS(2982),
    [anon_sym_oct] = ACTIONS(2982),
    [anon_sym_ord] = ACTIONS(2982),
    [anon_sym_prototype] = ACTIONS(2982),
    [anon_sym_pop] = ACTIONS(2982),
    [anon_sym_pos] = ACTIONS(2982),
    [anon_sym_quotemeta] = ACTIONS(2982),
    [anon_sym_reset] = ACTIONS(2982),
    [anon_sym_rand] = ACTIONS(2982),
    [anon_sym_rmdir] = ACTIONS(2982),
    [anon_sym_readdir] = ACTIONS(2982),
    [anon_sym_readline] = ACTIONS(2982),
    [anon_sym_readpipe] = ACTIONS(2982),
    [anon_sym_rewinddir] = ACTIONS(2982),
    [anon_sym_readlink] = ACTIONS(2982),
    [anon_sym_ref] = ACTIONS(2982),
    [anon_sym_scalar] = ACTIONS(2982),
    [anon_sym_shift] = ACTIONS(2982),
    [anon_sym_sin] = ACTIONS(2982),
    [anon_sym_sleep] = ACTIONS(2982),
    [anon_sym_sqrt] = ACTIONS(2982),
    [anon_sym_srand] = ACTIONS(2982),
    [anon_sym_stat] = ACTIONS(2982),
    [anon_sym_study] = ACTIONS(2982),
    [anon_sym_tell] = ACTIONS(2982),
    [anon_sym_telldir] = ACTIONS(2982),
    [anon_sym_tied] = ACTIONS(2982),
    [anon_sym_uc] = ACTIONS(2982),
    [anon_sym_ucfirst] = ACTIONS(2982),
    [anon_sym_untie] = ACTIONS(2982),
    [anon_sym_umask] = ACTIONS(2982),
    [anon_sym_values] = ACTIONS(2982),
    [anon_sym_write] = ACTIONS(2982),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2982),
    [anon_sym_qq] = ACTIONS(2982),
    [anon_sym_qw] = ACTIONS(2982),
    [anon_sym_qx] = ACTIONS(2982),
    [anon_sym_qr] = ACTIONS(2982),
    [anon_sym_m] = ACTIONS(2982),
    [anon_sym_s] = ACTIONS(2982),
    [anon_sym_tr] = ACTIONS(2982),
    [anon_sym_y] = ACTIONS(2982),
    [anon_sym_if] = ACTIONS(2982),
    [anon_sym_unless] = ACTIONS(2982),
    [anon_sym_while] = ACTIONS(2982),
    [anon_sym_until] = ACTIONS(2982),
    [aux_sym__bareword_token1] = ACTIONS(2982),
    [sym_number] = ACTIONS(2982),
    [anon_sym_true] = ACTIONS(2982),
    [anon_sym_false] = ACTIONS(2982),
    [sym__single_quote] = ACTIONS(2980),
    [sym__double_quote] = ACTIONS(2980),
    [sym__backtick_quote] = ACTIONS(2980),
    [sym__search_slash_quote] = ACTIONS(2980),
    [sym__open_readline_bracket] = ACTIONS(2980),
    [sym__open_fileglob_bracket] = ACTIONS(2980),
    [sym__PERLY_HEREDOC] = ACTIONS(2980),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2980),
    [sym__filetest] = ACTIONS(2980),
  },
  [STATE(1551)] = {
    [sym_heredoc_content] = STATE(1551),
    [sym__identifier] = ACTIONS(2994),
    [anon_sym_LBRACE] = ACTIONS(2994),
    [anon_sym_RBRACE] = ACTIONS(2994),
    [anon_sym_SEMI] = ACTIONS(2994),
    [anon_sym_package] = ACTIONS(2994),
    [anon_sym_class] = ACTIONS(2994),
    [anon_sym_role] = ACTIONS(2994),
    [anon_sym_BUILD] = ACTIONS(2994),
    [anon_sym_ADJUST] = ACTIONS(2994),
    [anon_sym_DOLLAR] = ACTIONS(2994),
    [anon_sym_AT] = ACTIONS(2994),
    [anon_sym_my] = ACTIONS(2994),
    [anon_sym_extended] = ACTIONS(2994),
    [anon_sym_async] = ACTIONS(2994),
    [anon_sym_sub] = ACTIONS(2994),
    [anon_sym_method] = ACTIONS(2994),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_continue] = ACTIONS(2994),
    [anon_sym_state] = ACTIONS(2994),
    [anon_sym_our] = ACTIONS(2994),
    [anon_sym_try] = ACTIONS(2994),
    [anon_sym_defer] = ACTIONS(2994),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2994),
    [anon_sym_LBRACK] = ACTIONS(2994),
    [anon_sym_LT_LT] = ACTIONS(2994),
    [anon_sym_SLASH_SLASH] = ACTIONS(2994),
    [anon_sym_PLUS] = ACTIONS(2994),
    [anon_sym_DASH] = ACTIONS(2994),
    [anon_sym_TILDE] = ACTIONS(2994),
    [anon_sym_BANG] = ACTIONS(2994),
    [anon_sym_PLUS_PLUS] = ACTIONS(2994),
    [anon_sym_DASH_DASH] = ACTIONS(2994),
    [anon_sym_BSLASH] = ACTIONS(2992),
    [anon_sym_do] = ACTIONS(2994),
    [anon_sym_eval] = ACTIONS(2994),
    [anon_sym_await] = ACTIONS(2994),
    [anon_sym_field] = ACTIONS(2994),
    [anon_sym_local] = ACTIONS(2994),
    [anon_sym_dynamically] = ACTIONS(2994),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2994),
    [anon_sym_require] = ACTIONS(2994),
    [anon_sym_map] = ACTIONS(2994),
    [anon_sym_grep] = ACTIONS(2994),
    [anon_sym_sort] = ACTIONS(2994),
    [anon_sym_goto] = ACTIONS(2994),
    [anon_sym_return] = ACTIONS(2994),
    [anon_sym_undef] = ACTIONS(2994),
    [anon_sym_PERCENT2] = ACTIONS(2994),
    [anon_sym_AMP2] = ACTIONS(2994),
    [anon_sym_STAR2] = ACTIONS(2994),
    [anon_sym_use] = ACTIONS(2994),
    [anon_sym_no] = ACTIONS(2994),
    [anon_sym_for] = ACTIONS(2994),
    [anon_sym_foreach] = ACTIONS(2994),
    [anon_sym_last] = ACTIONS(2994),
    [anon_sym_next] = ACTIONS(2994),
    [anon_sym_redo] = ACTIONS(2994),
    [anon_sym_BEGIN] = ACTIONS(2994),
    [anon_sym_INIT] = ACTIONS(2994),
    [anon_sym_CHECK] = ACTIONS(2994),
    [anon_sym_UNITCHECK] = ACTIONS(2994),
    [anon_sym_END] = ACTIONS(2994),
    [anon_sym___FILE__] = ACTIONS(2994),
    [anon_sym___LINE__] = ACTIONS(2994),
    [anon_sym___PACKAGE__] = ACTIONS(2994),
    [anon_sym___SUB__] = ACTIONS(2994),
    [anon_sym_break] = ACTIONS(2994),
    [anon_sym_fork] = ACTIONS(2994),
    [anon_sym_getppid] = ACTIONS(2994),
    [anon_sym_time] = ACTIONS(2994),
    [anon_sym_times] = ACTIONS(2994),
    [anon_sym_wait] = ACTIONS(2994),
    [anon_sym_wantarray] = ACTIONS(2994),
    [anon_sym_abs] = ACTIONS(2994),
    [anon_sym_alarm] = ACTIONS(2994),
    [anon_sym_chop] = ACTIONS(2994),
    [anon_sym_chdir] = ACTIONS(2994),
    [anon_sym_close] = ACTIONS(2994),
    [anon_sym_closedir] = ACTIONS(2994),
    [anon_sym_caller] = ACTIONS(2994),
    [anon_sym_chomp] = ACTIONS(2994),
    [anon_sym_chr] = ACTIONS(2994),
    [anon_sym_cos] = ACTIONS(2994),
    [anon_sym_chroot] = ACTIONS(2994),
    [anon_sym_defined] = ACTIONS(2994),
    [anon_sym_delete] = ACTIONS(2994),
    [anon_sym_dbmclose] = ACTIONS(2994),
    [anon_sym_exists] = ACTIONS(2994),
    [anon_sym_exit] = ACTIONS(2994),
    [anon_sym_eof] = ACTIONS(2994),
    [anon_sym_exp] = ACTIONS(2994),
    [anon_sym_each] = ACTIONS(2994),
    [anon_sym_fc] = ACTIONS(2994),
    [anon_sym_fileno] = ACTIONS(2994),
    [anon_sym_gmtime] = ACTIONS(2994),
    [anon_sym_getc] = ACTIONS(2994),
    [anon_sym_getpgrp] = ACTIONS(2994),
    [anon_sym_getprotobyname] = ACTIONS(2994),
    [anon_sym_getpwname] = ACTIONS(2994),
    [anon_sym_getpwuid] = ACTIONS(2994),
    [anon_sym_getpeername] = ACTIONS(2994),
    [anon_sym_getnetbyname] = ACTIONS(2994),
    [anon_sym_getsockname] = ACTIONS(2994),
    [anon_sym_getgrnam] = ACTIONS(2994),
    [anon_sym_getgrgid] = ACTIONS(2994),
    [anon_sym_hex] = ACTIONS(2994),
    [anon_sym_int] = ACTIONS(2994),
    [anon_sym_keys] = ACTIONS(2994),
    [anon_sym_lc] = ACTIONS(2994),
    [anon_sym_lcfirst] = ACTIONS(2994),
    [anon_sym_length] = ACTIONS(2994),
    [anon_sym_localtime] = ACTIONS(2994),
    [anon_sym_log] = ACTIONS(2994),
    [anon_sym_lock] = ACTIONS(2994),
    [anon_sym_lstat] = ACTIONS(2994),
    [anon_sym_oct] = ACTIONS(2994),
    [anon_sym_ord] = ACTIONS(2994),
    [anon_sym_prototype] = ACTIONS(2994),
    [anon_sym_pop] = ACTIONS(2994),
    [anon_sym_pos] = ACTIONS(2994),
    [anon_sym_quotemeta] = ACTIONS(2994),
    [anon_sym_reset] = ACTIONS(2994),
    [anon_sym_rand] = ACTIONS(2994),
    [anon_sym_rmdir] = ACTIONS(2994),
    [anon_sym_readdir] = ACTIONS(2994),
    [anon_sym_readline] = ACTIONS(2994),
    [anon_sym_readpipe] = ACTIONS(2994),
    [anon_sym_rewinddir] = ACTIONS(2994),
    [anon_sym_readlink] = ACTIONS(2994),
    [anon_sym_ref] = ACTIONS(2994),
    [anon_sym_scalar] = ACTIONS(2994),
    [anon_sym_shift] = ACTIONS(2994),
    [anon_sym_sin] = ACTIONS(2994),
    [anon_sym_sleep] = ACTIONS(2994),
    [anon_sym_sqrt] = ACTIONS(2994),
    [anon_sym_srand] = ACTIONS(2994),
    [anon_sym_stat] = ACTIONS(2994),
    [anon_sym_study] = ACTIONS(2994),
    [anon_sym_tell] = ACTIONS(2994),
    [anon_sym_telldir] = ACTIONS(2994),
    [anon_sym_tied] = ACTIONS(2994),
    [anon_sym_uc] = ACTIONS(2994),
    [anon_sym_ucfirst] = ACTIONS(2994),
    [anon_sym_untie] = ACTIONS(2994),
    [anon_sym_umask] = ACTIONS(2994),
    [anon_sym_values] = ACTIONS(2994),
    [anon_sym_write] = ACTIONS(2994),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2994),
    [anon_sym_qq] = ACTIONS(2994),
    [anon_sym_qw] = ACTIONS(2994),
    [anon_sym_qx] = ACTIONS(2994),
    [anon_sym_qr] = ACTIONS(2994),
    [anon_sym_m] = ACTIONS(2994),
    [anon_sym_s] = ACTIONS(2994),
    [anon_sym_tr] = ACTIONS(2994),
    [anon_sym_y] = ACTIONS(2994),
    [anon_sym_if] = ACTIONS(2994),
    [anon_sym_unless] = ACTIONS(2994),
    [anon_sym_while] = ACTIONS(2994),
    [anon_sym_until] = ACTIONS(2994),
    [aux_sym__bareword_token1] = ACTIONS(2994),
    [sym_number] = ACTIONS(2994),
    [anon_sym_true] = ACTIONS(2994),
    [anon_sym_false] = ACTIONS(2994),
    [sym__single_quote] = ACTIONS(2992),
    [sym__double_quote] = ACTIONS(2992),
    [sym__backtick_quote] = ACTIONS(2992),
    [sym__search_slash_quote] = ACTIONS(2992),
    [sym__open_readline_bracket] = ACTIONS(2992),
    [sym__open_fileglob_bracket] = ACTIONS(2992),
    [sym__PERLY_HEREDOC] = ACTIONS(2992),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2992),
    [sym__filetest] = ACTIONS(2992),
  },
  [STATE(1552)] = {
    [sym_heredoc_content] = STATE(1552),
    [sym__identifier] = ACTIONS(2998),
    [anon_sym_LBRACE] = ACTIONS(2998),
    [anon_sym_RBRACE] = ACTIONS(2998),
    [anon_sym_SEMI] = ACTIONS(2998),
    [anon_sym_package] = ACTIONS(2998),
    [anon_sym_class] = ACTIONS(2998),
    [anon_sym_role] = ACTIONS(2998),
    [anon_sym_BUILD] = ACTIONS(2998),
    [anon_sym_ADJUST] = ACTIONS(2998),
    [anon_sym_DOLLAR] = ACTIONS(2998),
    [anon_sym_AT] = ACTIONS(2998),
    [anon_sym_my] = ACTIONS(2998),
    [anon_sym_extended] = ACTIONS(2998),
    [anon_sym_async] = ACTIONS(2998),
    [anon_sym_sub] = ACTIONS(2998),
    [anon_sym_method] = ACTIONS(2998),
    [anon_sym_LPAREN] = ACTIONS(2998),
    [anon_sym_continue] = ACTIONS(2998),
    [anon_sym_state] = ACTIONS(2998),
    [anon_sym_our] = ACTIONS(2998),
    [anon_sym_try] = ACTIONS(2998),
    [anon_sym_defer] = ACTIONS(2998),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2998),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LT_LT] = ACTIONS(2998),
    [anon_sym_SLASH_SLASH] = ACTIONS(2998),
    [anon_sym_PLUS] = ACTIONS(2998),
    [anon_sym_DASH] = ACTIONS(2998),
    [anon_sym_TILDE] = ACTIONS(2998),
    [anon_sym_BANG] = ACTIONS(2998),
    [anon_sym_PLUS_PLUS] = ACTIONS(2998),
    [anon_sym_DASH_DASH] = ACTIONS(2998),
    [anon_sym_BSLASH] = ACTIONS(2996),
    [anon_sym_do] = ACTIONS(2998),
    [anon_sym_eval] = ACTIONS(2998),
    [anon_sym_await] = ACTIONS(2998),
    [anon_sym_field] = ACTIONS(2998),
    [anon_sym_local] = ACTIONS(2998),
    [anon_sym_dynamically] = ACTIONS(2998),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2998),
    [anon_sym_require] = ACTIONS(2998),
    [anon_sym_map] = ACTIONS(2998),
    [anon_sym_grep] = ACTIONS(2998),
    [anon_sym_sort] = ACTIONS(2998),
    [anon_sym_goto] = ACTIONS(2998),
    [anon_sym_return] = ACTIONS(2998),
    [anon_sym_undef] = ACTIONS(2998),
    [anon_sym_PERCENT2] = ACTIONS(2998),
    [anon_sym_AMP2] = ACTIONS(2998),
    [anon_sym_STAR2] = ACTIONS(2998),
    [anon_sym_use] = ACTIONS(2998),
    [anon_sym_no] = ACTIONS(2998),
    [anon_sym_for] = ACTIONS(2998),
    [anon_sym_foreach] = ACTIONS(2998),
    [anon_sym_last] = ACTIONS(2998),
    [anon_sym_next] = ACTIONS(2998),
    [anon_sym_redo] = ACTIONS(2998),
    [anon_sym_BEGIN] = ACTIONS(2998),
    [anon_sym_INIT] = ACTIONS(2998),
    [anon_sym_CHECK] = ACTIONS(2998),
    [anon_sym_UNITCHECK] = ACTIONS(2998),
    [anon_sym_END] = ACTIONS(2998),
    [anon_sym___FILE__] = ACTIONS(2998),
    [anon_sym___LINE__] = ACTIONS(2998),
    [anon_sym___PACKAGE__] = ACTIONS(2998),
    [anon_sym___SUB__] = ACTIONS(2998),
    [anon_sym_break] = ACTIONS(2998),
    [anon_sym_fork] = ACTIONS(2998),
    [anon_sym_getppid] = ACTIONS(2998),
    [anon_sym_time] = ACTIONS(2998),
    [anon_sym_times] = ACTIONS(2998),
    [anon_sym_wait] = ACTIONS(2998),
    [anon_sym_wantarray] = ACTIONS(2998),
    [anon_sym_abs] = ACTIONS(2998),
    [anon_sym_alarm] = ACTIONS(2998),
    [anon_sym_chop] = ACTIONS(2998),
    [anon_sym_chdir] = ACTIONS(2998),
    [anon_sym_close] = ACTIONS(2998),
    [anon_sym_closedir] = ACTIONS(2998),
    [anon_sym_caller] = ACTIONS(2998),
    [anon_sym_chomp] = ACTIONS(2998),
    [anon_sym_chr] = ACTIONS(2998),
    [anon_sym_cos] = ACTIONS(2998),
    [anon_sym_chroot] = ACTIONS(2998),
    [anon_sym_defined] = ACTIONS(2998),
    [anon_sym_delete] = ACTIONS(2998),
    [anon_sym_dbmclose] = ACTIONS(2998),
    [anon_sym_exists] = ACTIONS(2998),
    [anon_sym_exit] = ACTIONS(2998),
    [anon_sym_eof] = ACTIONS(2998),
    [anon_sym_exp] = ACTIONS(2998),
    [anon_sym_each] = ACTIONS(2998),
    [anon_sym_fc] = ACTIONS(2998),
    [anon_sym_fileno] = ACTIONS(2998),
    [anon_sym_gmtime] = ACTIONS(2998),
    [anon_sym_getc] = ACTIONS(2998),
    [anon_sym_getpgrp] = ACTIONS(2998),
    [anon_sym_getprotobyname] = ACTIONS(2998),
    [anon_sym_getpwname] = ACTIONS(2998),
    [anon_sym_getpwuid] = ACTIONS(2998),
    [anon_sym_getpeername] = ACTIONS(2998),
    [anon_sym_getnetbyname] = ACTIONS(2998),
    [anon_sym_getsockname] = ACTIONS(2998),
    [anon_sym_getgrnam] = ACTIONS(2998),
    [anon_sym_getgrgid] = ACTIONS(2998),
    [anon_sym_hex] = ACTIONS(2998),
    [anon_sym_int] = ACTIONS(2998),
    [anon_sym_keys] = ACTIONS(2998),
    [anon_sym_lc] = ACTIONS(2998),
    [anon_sym_lcfirst] = ACTIONS(2998),
    [anon_sym_length] = ACTIONS(2998),
    [anon_sym_localtime] = ACTIONS(2998),
    [anon_sym_log] = ACTIONS(2998),
    [anon_sym_lock] = ACTIONS(2998),
    [anon_sym_lstat] = ACTIONS(2998),
    [anon_sym_oct] = ACTIONS(2998),
    [anon_sym_ord] = ACTIONS(2998),
    [anon_sym_prototype] = ACTIONS(2998),
    [anon_sym_pop] = ACTIONS(2998),
    [anon_sym_pos] = ACTIONS(2998),
    [anon_sym_quotemeta] = ACTIONS(2998),
    [anon_sym_reset] = ACTIONS(2998),
    [anon_sym_rand] = ACTIONS(2998),
    [anon_sym_rmdir] = ACTIONS(2998),
    [anon_sym_readdir] = ACTIONS(2998),
    [anon_sym_readline] = ACTIONS(2998),
    [anon_sym_readpipe] = ACTIONS(2998),
    [anon_sym_rewinddir] = ACTIONS(2998),
    [anon_sym_readlink] = ACTIONS(2998),
    [anon_sym_ref] = ACTIONS(2998),
    [anon_sym_scalar] = ACTIONS(2998),
    [anon_sym_shift] = ACTIONS(2998),
    [anon_sym_sin] = ACTIONS(2998),
    [anon_sym_sleep] = ACTIONS(2998),
    [anon_sym_sqrt] = ACTIONS(2998),
    [anon_sym_srand] = ACTIONS(2998),
    [anon_sym_stat] = ACTIONS(2998),
    [anon_sym_study] = ACTIONS(2998),
    [anon_sym_tell] = ACTIONS(2998),
    [anon_sym_telldir] = ACTIONS(2998),
    [anon_sym_tied] = ACTIONS(2998),
    [anon_sym_uc] = ACTIONS(2998),
    [anon_sym_ucfirst] = ACTIONS(2998),
    [anon_sym_untie] = ACTIONS(2998),
    [anon_sym_umask] = ACTIONS(2998),
    [anon_sym_values] = ACTIONS(2998),
    [anon_sym_write] = ACTIONS(2998),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2998),
    [anon_sym_qq] = ACTIONS(2998),
    [anon_sym_qw] = ACTIONS(2998),
    [anon_sym_qx] = ACTIONS(2998),
    [anon_sym_qr] = ACTIONS(2998),
    [anon_sym_m] = ACTIONS(2998),
    [anon_sym_s] = ACTIONS(2998),
    [anon_sym_tr] = ACTIONS(2998),
    [anon_sym_y] = ACTIONS(2998),
    [anon_sym_if] = ACTIONS(2998),
    [anon_sym_unless] = ACTIONS(2998),
    [anon_sym_while] = ACTIONS(2998),
    [anon_sym_until] = ACTIONS(2998),
    [aux_sym__bareword_token1] = ACTIONS(2998),
    [sym_number] = ACTIONS(2998),
    [anon_sym_true] = ACTIONS(2998),
    [anon_sym_false] = ACTIONS(2998),
    [sym__single_quote] = ACTIONS(2996),
    [sym__double_quote] = ACTIONS(2996),
    [sym__backtick_quote] = ACTIONS(2996),
    [sym__search_slash_quote] = ACTIONS(2996),
    [sym__open_readline_bracket] = ACTIONS(2996),
    [sym__open_fileglob_bracket] = ACTIONS(2996),
    [sym__PERLY_HEREDOC] = ACTIONS(2996),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2996),
    [sym__filetest] = ACTIONS(2996),
  },
  [STATE(1553)] = {
    [sym_heredoc_content] = STATE(1553),
    [sym__identifier] = ACTIONS(3002),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_RBRACE] = ACTIONS(3002),
    [anon_sym_SEMI] = ACTIONS(3002),
    [anon_sym_package] = ACTIONS(3002),
    [anon_sym_class] = ACTIONS(3002),
    [anon_sym_role] = ACTIONS(3002),
    [anon_sym_BUILD] = ACTIONS(3002),
    [anon_sym_ADJUST] = ACTIONS(3002),
    [anon_sym_DOLLAR] = ACTIONS(3002),
    [anon_sym_AT] = ACTIONS(3002),
    [anon_sym_my] = ACTIONS(3002),
    [anon_sym_extended] = ACTIONS(3002),
    [anon_sym_async] = ACTIONS(3002),
    [anon_sym_sub] = ACTIONS(3002),
    [anon_sym_method] = ACTIONS(3002),
    [anon_sym_LPAREN] = ACTIONS(3002),
    [anon_sym_continue] = ACTIONS(3002),
    [anon_sym_state] = ACTIONS(3002),
    [anon_sym_our] = ACTIONS(3002),
    [anon_sym_try] = ACTIONS(3002),
    [anon_sym_defer] = ACTIONS(3002),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3002),
    [anon_sym_LBRACK] = ACTIONS(3002),
    [anon_sym_LT_LT] = ACTIONS(3002),
    [anon_sym_SLASH_SLASH] = ACTIONS(3002),
    [anon_sym_PLUS] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3002),
    [anon_sym_TILDE] = ACTIONS(3002),
    [anon_sym_BANG] = ACTIONS(3002),
    [anon_sym_PLUS_PLUS] = ACTIONS(3002),
    [anon_sym_DASH_DASH] = ACTIONS(3002),
    [anon_sym_BSLASH] = ACTIONS(3000),
    [anon_sym_do] = ACTIONS(3002),
    [anon_sym_eval] = ACTIONS(3002),
    [anon_sym_await] = ACTIONS(3002),
    [anon_sym_field] = ACTIONS(3002),
    [anon_sym_local] = ACTIONS(3002),
    [anon_sym_dynamically] = ACTIONS(3002),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3002),
    [anon_sym_require] = ACTIONS(3002),
    [anon_sym_map] = ACTIONS(3002),
    [anon_sym_grep] = ACTIONS(3002),
    [anon_sym_sort] = ACTIONS(3002),
    [anon_sym_goto] = ACTIONS(3002),
    [anon_sym_return] = ACTIONS(3002),
    [anon_sym_undef] = ACTIONS(3002),
    [anon_sym_PERCENT2] = ACTIONS(3002),
    [anon_sym_AMP2] = ACTIONS(3002),
    [anon_sym_STAR2] = ACTIONS(3002),
    [anon_sym_use] = ACTIONS(3002),
    [anon_sym_no] = ACTIONS(3002),
    [anon_sym_for] = ACTIONS(3002),
    [anon_sym_foreach] = ACTIONS(3002),
    [anon_sym_last] = ACTIONS(3002),
    [anon_sym_next] = ACTIONS(3002),
    [anon_sym_redo] = ACTIONS(3002),
    [anon_sym_BEGIN] = ACTIONS(3002),
    [anon_sym_INIT] = ACTIONS(3002),
    [anon_sym_CHECK] = ACTIONS(3002),
    [anon_sym_UNITCHECK] = ACTIONS(3002),
    [anon_sym_END] = ACTIONS(3002),
    [anon_sym___FILE__] = ACTIONS(3002),
    [anon_sym___LINE__] = ACTIONS(3002),
    [anon_sym___PACKAGE__] = ACTIONS(3002),
    [anon_sym___SUB__] = ACTIONS(3002),
    [anon_sym_break] = ACTIONS(3002),
    [anon_sym_fork] = ACTIONS(3002),
    [anon_sym_getppid] = ACTIONS(3002),
    [anon_sym_time] = ACTIONS(3002),
    [anon_sym_times] = ACTIONS(3002),
    [anon_sym_wait] = ACTIONS(3002),
    [anon_sym_wantarray] = ACTIONS(3002),
    [anon_sym_abs] = ACTIONS(3002),
    [anon_sym_alarm] = ACTIONS(3002),
    [anon_sym_chop] = ACTIONS(3002),
    [anon_sym_chdir] = ACTIONS(3002),
    [anon_sym_close] = ACTIONS(3002),
    [anon_sym_closedir] = ACTIONS(3002),
    [anon_sym_caller] = ACTIONS(3002),
    [anon_sym_chomp] = ACTIONS(3002),
    [anon_sym_chr] = ACTIONS(3002),
    [anon_sym_cos] = ACTIONS(3002),
    [anon_sym_chroot] = ACTIONS(3002),
    [anon_sym_defined] = ACTIONS(3002),
    [anon_sym_delete] = ACTIONS(3002),
    [anon_sym_dbmclose] = ACTIONS(3002),
    [anon_sym_exists] = ACTIONS(3002),
    [anon_sym_exit] = ACTIONS(3002),
    [anon_sym_eof] = ACTIONS(3002),
    [anon_sym_exp] = ACTIONS(3002),
    [anon_sym_each] = ACTIONS(3002),
    [anon_sym_fc] = ACTIONS(3002),
    [anon_sym_fileno] = ACTIONS(3002),
    [anon_sym_gmtime] = ACTIONS(3002),
    [anon_sym_getc] = ACTIONS(3002),
    [anon_sym_getpgrp] = ACTIONS(3002),
    [anon_sym_getprotobyname] = ACTIONS(3002),
    [anon_sym_getpwname] = ACTIONS(3002),
    [anon_sym_getpwuid] = ACTIONS(3002),
    [anon_sym_getpeername] = ACTIONS(3002),
    [anon_sym_getnetbyname] = ACTIONS(3002),
    [anon_sym_getsockname] = ACTIONS(3002),
    [anon_sym_getgrnam] = ACTIONS(3002),
    [anon_sym_getgrgid] = ACTIONS(3002),
    [anon_sym_hex] = ACTIONS(3002),
    [anon_sym_int] = ACTIONS(3002),
    [anon_sym_keys] = ACTIONS(3002),
    [anon_sym_lc] = ACTIONS(3002),
    [anon_sym_lcfirst] = ACTIONS(3002),
    [anon_sym_length] = ACTIONS(3002),
    [anon_sym_localtime] = ACTIONS(3002),
    [anon_sym_log] = ACTIONS(3002),
    [anon_sym_lock] = ACTIONS(3002),
    [anon_sym_lstat] = ACTIONS(3002),
    [anon_sym_oct] = ACTIONS(3002),
    [anon_sym_ord] = ACTIONS(3002),
    [anon_sym_prototype] = ACTIONS(3002),
    [anon_sym_pop] = ACTIONS(3002),
    [anon_sym_pos] = ACTIONS(3002),
    [anon_sym_quotemeta] = ACTIONS(3002),
    [anon_sym_reset] = ACTIONS(3002),
    [anon_sym_rand] = ACTIONS(3002),
    [anon_sym_rmdir] = ACTIONS(3002),
    [anon_sym_readdir] = ACTIONS(3002),
    [anon_sym_readline] = ACTIONS(3002),
    [anon_sym_readpipe] = ACTIONS(3002),
    [anon_sym_rewinddir] = ACTIONS(3002),
    [anon_sym_readlink] = ACTIONS(3002),
    [anon_sym_ref] = ACTIONS(3002),
    [anon_sym_scalar] = ACTIONS(3002),
    [anon_sym_shift] = ACTIONS(3002),
    [anon_sym_sin] = ACTIONS(3002),
    [anon_sym_sleep] = ACTIONS(3002),
    [anon_sym_sqrt] = ACTIONS(3002),
    [anon_sym_srand] = ACTIONS(3002),
    [anon_sym_stat] = ACTIONS(3002),
    [anon_sym_study] = ACTIONS(3002),
    [anon_sym_tell] = ACTIONS(3002),
    [anon_sym_telldir] = ACTIONS(3002),
    [anon_sym_tied] = ACTIONS(3002),
    [anon_sym_uc] = ACTIONS(3002),
    [anon_sym_ucfirst] = ACTIONS(3002),
    [anon_sym_untie] = ACTIONS(3002),
    [anon_sym_umask] = ACTIONS(3002),
    [anon_sym_values] = ACTIONS(3002),
    [anon_sym_write] = ACTIONS(3002),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3002),
    [anon_sym_qq] = ACTIONS(3002),
    [anon_sym_qw] = ACTIONS(3002),
    [anon_sym_qx] = ACTIONS(3002),
    [anon_sym_qr] = ACTIONS(3002),
    [anon_sym_m] = ACTIONS(3002),
    [anon_sym_s] = ACTIONS(3002),
    [anon_sym_tr] = ACTIONS(3002),
    [anon_sym_y] = ACTIONS(3002),
    [anon_sym_if] = ACTIONS(3002),
    [anon_sym_unless] = ACTIONS(3002),
    [anon_sym_while] = ACTIONS(3002),
    [anon_sym_until] = ACTIONS(3002),
    [aux_sym__bareword_token1] = ACTIONS(3002),
    [sym_number] = ACTIONS(3002),
    [anon_sym_true] = ACTIONS(3002),
    [anon_sym_false] = ACTIONS(3002),
    [sym__single_quote] = ACTIONS(3000),
    [sym__double_quote] = ACTIONS(3000),
    [sym__backtick_quote] = ACTIONS(3000),
    [sym__search_slash_quote] = ACTIONS(3000),
    [sym__open_readline_bracket] = ACTIONS(3000),
    [sym__open_fileglob_bracket] = ACTIONS(3000),
    [sym__PERLY_HEREDOC] = ACTIONS(3000),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3000),
    [sym__filetest] = ACTIONS(3000),
  },
  [STATE(1554)] = {
    [sym_heredoc_content] = STATE(1554),
    [sym__identifier] = ACTIONS(3006),
    [anon_sym_LBRACE] = ACTIONS(3006),
    [anon_sym_RBRACE] = ACTIONS(3006),
    [anon_sym_SEMI] = ACTIONS(3006),
    [anon_sym_package] = ACTIONS(3006),
    [anon_sym_class] = ACTIONS(3006),
    [anon_sym_role] = ACTIONS(3006),
    [anon_sym_BUILD] = ACTIONS(3006),
    [anon_sym_ADJUST] = ACTIONS(3006),
    [anon_sym_DOLLAR] = ACTIONS(3006),
    [anon_sym_AT] = ACTIONS(3006),
    [anon_sym_my] = ACTIONS(3006),
    [anon_sym_extended] = ACTIONS(3006),
    [anon_sym_async] = ACTIONS(3006),
    [anon_sym_sub] = ACTIONS(3006),
    [anon_sym_method] = ACTIONS(3006),
    [anon_sym_LPAREN] = ACTIONS(3006),
    [anon_sym_continue] = ACTIONS(3006),
    [anon_sym_state] = ACTIONS(3006),
    [anon_sym_our] = ACTIONS(3006),
    [anon_sym_try] = ACTIONS(3006),
    [anon_sym_defer] = ACTIONS(3006),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3006),
    [anon_sym_LBRACK] = ACTIONS(3006),
    [anon_sym_LT_LT] = ACTIONS(3006),
    [anon_sym_SLASH_SLASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_TILDE] = ACTIONS(3006),
    [anon_sym_BANG] = ACTIONS(3006),
    [anon_sym_PLUS_PLUS] = ACTIONS(3006),
    [anon_sym_DASH_DASH] = ACTIONS(3006),
    [anon_sym_BSLASH] = ACTIONS(3004),
    [anon_sym_do] = ACTIONS(3006),
    [anon_sym_eval] = ACTIONS(3006),
    [anon_sym_await] = ACTIONS(3006),
    [anon_sym_field] = ACTIONS(3006),
    [anon_sym_local] = ACTIONS(3006),
    [anon_sym_dynamically] = ACTIONS(3006),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3006),
    [anon_sym_require] = ACTIONS(3006),
    [anon_sym_map] = ACTIONS(3006),
    [anon_sym_grep] = ACTIONS(3006),
    [anon_sym_sort] = ACTIONS(3006),
    [anon_sym_goto] = ACTIONS(3006),
    [anon_sym_return] = ACTIONS(3006),
    [anon_sym_undef] = ACTIONS(3006),
    [anon_sym_PERCENT2] = ACTIONS(3006),
    [anon_sym_AMP2] = ACTIONS(3006),
    [anon_sym_STAR2] = ACTIONS(3006),
    [anon_sym_use] = ACTIONS(3006),
    [anon_sym_no] = ACTIONS(3006),
    [anon_sym_for] = ACTIONS(3006),
    [anon_sym_foreach] = ACTIONS(3006),
    [anon_sym_last] = ACTIONS(3006),
    [anon_sym_next] = ACTIONS(3006),
    [anon_sym_redo] = ACTIONS(3006),
    [anon_sym_BEGIN] = ACTIONS(3006),
    [anon_sym_INIT] = ACTIONS(3006),
    [anon_sym_CHECK] = ACTIONS(3006),
    [anon_sym_UNITCHECK] = ACTIONS(3006),
    [anon_sym_END] = ACTIONS(3006),
    [anon_sym___FILE__] = ACTIONS(3006),
    [anon_sym___LINE__] = ACTIONS(3006),
    [anon_sym___PACKAGE__] = ACTIONS(3006),
    [anon_sym___SUB__] = ACTIONS(3006),
    [anon_sym_break] = ACTIONS(3006),
    [anon_sym_fork] = ACTIONS(3006),
    [anon_sym_getppid] = ACTIONS(3006),
    [anon_sym_time] = ACTIONS(3006),
    [anon_sym_times] = ACTIONS(3006),
    [anon_sym_wait] = ACTIONS(3006),
    [anon_sym_wantarray] = ACTIONS(3006),
    [anon_sym_abs] = ACTIONS(3006),
    [anon_sym_alarm] = ACTIONS(3006),
    [anon_sym_chop] = ACTIONS(3006),
    [anon_sym_chdir] = ACTIONS(3006),
    [anon_sym_close] = ACTIONS(3006),
    [anon_sym_closedir] = ACTIONS(3006),
    [anon_sym_caller] = ACTIONS(3006),
    [anon_sym_chomp] = ACTIONS(3006),
    [anon_sym_chr] = ACTIONS(3006),
    [anon_sym_cos] = ACTIONS(3006),
    [anon_sym_chroot] = ACTIONS(3006),
    [anon_sym_defined] = ACTIONS(3006),
    [anon_sym_delete] = ACTIONS(3006),
    [anon_sym_dbmclose] = ACTIONS(3006),
    [anon_sym_exists] = ACTIONS(3006),
    [anon_sym_exit] = ACTIONS(3006),
    [anon_sym_eof] = ACTIONS(3006),
    [anon_sym_exp] = ACTIONS(3006),
    [anon_sym_each] = ACTIONS(3006),
    [anon_sym_fc] = ACTIONS(3006),
    [anon_sym_fileno] = ACTIONS(3006),
    [anon_sym_gmtime] = ACTIONS(3006),
    [anon_sym_getc] = ACTIONS(3006),
    [anon_sym_getpgrp] = ACTIONS(3006),
    [anon_sym_getprotobyname] = ACTIONS(3006),
    [anon_sym_getpwname] = ACTIONS(3006),
    [anon_sym_getpwuid] = ACTIONS(3006),
    [anon_sym_getpeername] = ACTIONS(3006),
    [anon_sym_getnetbyname] = ACTIONS(3006),
    [anon_sym_getsockname] = ACTIONS(3006),
    [anon_sym_getgrnam] = ACTIONS(3006),
    [anon_sym_getgrgid] = ACTIONS(3006),
    [anon_sym_hex] = ACTIONS(3006),
    [anon_sym_int] = ACTIONS(3006),
    [anon_sym_keys] = ACTIONS(3006),
    [anon_sym_lc] = ACTIONS(3006),
    [anon_sym_lcfirst] = ACTIONS(3006),
    [anon_sym_length] = ACTIONS(3006),
    [anon_sym_localtime] = ACTIONS(3006),
    [anon_sym_log] = ACTIONS(3006),
    [anon_sym_lock] = ACTIONS(3006),
    [anon_sym_lstat] = ACTIONS(3006),
    [anon_sym_oct] = ACTIONS(3006),
    [anon_sym_ord] = ACTIONS(3006),
    [anon_sym_prototype] = ACTIONS(3006),
    [anon_sym_pop] = ACTIONS(3006),
    [anon_sym_pos] = ACTIONS(3006),
    [anon_sym_quotemeta] = ACTIONS(3006),
    [anon_sym_reset] = ACTIONS(3006),
    [anon_sym_rand] = ACTIONS(3006),
    [anon_sym_rmdir] = ACTIONS(3006),
    [anon_sym_readdir] = ACTIONS(3006),
    [anon_sym_readline] = ACTIONS(3006),
    [anon_sym_readpipe] = ACTIONS(3006),
    [anon_sym_rewinddir] = ACTIONS(3006),
    [anon_sym_readlink] = ACTIONS(3006),
    [anon_sym_ref] = ACTIONS(3006),
    [anon_sym_scalar] = ACTIONS(3006),
    [anon_sym_shift] = ACTIONS(3006),
    [anon_sym_sin] = ACTIONS(3006),
    [anon_sym_sleep] = ACTIONS(3006),
    [anon_sym_sqrt] = ACTIONS(3006),
    [anon_sym_srand] = ACTIONS(3006),
    [anon_sym_stat] = ACTIONS(3006),
    [anon_sym_study] = ACTIONS(3006),
    [anon_sym_tell] = ACTIONS(3006),
    [anon_sym_telldir] = ACTIONS(3006),
    [anon_sym_tied] = ACTIONS(3006),
    [anon_sym_uc] = ACTIONS(3006),
    [anon_sym_ucfirst] = ACTIONS(3006),
    [anon_sym_untie] = ACTIONS(3006),
    [anon_sym_umask] = ACTIONS(3006),
    [anon_sym_values] = ACTIONS(3006),
    [anon_sym_write] = ACTIONS(3006),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3006),
    [anon_sym_qq] = ACTIONS(3006),
    [anon_sym_qw] = ACTIONS(3006),
    [anon_sym_qx] = ACTIONS(3006),
    [anon_sym_qr] = ACTIONS(3006),
    [anon_sym_m] = ACTIONS(3006),
    [anon_sym_s] = ACTIONS(3006),
    [anon_sym_tr] = ACTIONS(3006),
    [anon_sym_y] = ACTIONS(3006),
    [anon_sym_if] = ACTIONS(3006),
    [anon_sym_unless] = ACTIONS(3006),
    [anon_sym_while] = ACTIONS(3006),
    [anon_sym_until] = ACTIONS(3006),
    [aux_sym__bareword_token1] = ACTIONS(3006),
    [sym_number] = ACTIONS(3006),
    [anon_sym_true] = ACTIONS(3006),
    [anon_sym_false] = ACTIONS(3006),
    [sym__single_quote] = ACTIONS(3004),
    [sym__double_quote] = ACTIONS(3004),
    [sym__backtick_quote] = ACTIONS(3004),
    [sym__search_slash_quote] = ACTIONS(3004),
    [sym__open_readline_bracket] = ACTIONS(3004),
    [sym__open_fileglob_bracket] = ACTIONS(3004),
    [sym__PERLY_HEREDOC] = ACTIONS(3004),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3004),
    [sym__filetest] = ACTIONS(3004),
  },
  [STATE(1555)] = {
    [sym_heredoc_content] = STATE(1555),
    [sym__identifier] = ACTIONS(3010),
    [anon_sym_LBRACE] = ACTIONS(3010),
    [anon_sym_RBRACE] = ACTIONS(3010),
    [anon_sym_SEMI] = ACTIONS(3010),
    [anon_sym_package] = ACTIONS(3010),
    [anon_sym_class] = ACTIONS(3010),
    [anon_sym_role] = ACTIONS(3010),
    [anon_sym_BUILD] = ACTIONS(3010),
    [anon_sym_ADJUST] = ACTIONS(3010),
    [anon_sym_DOLLAR] = ACTIONS(3010),
    [anon_sym_AT] = ACTIONS(3010),
    [anon_sym_my] = ACTIONS(3010),
    [anon_sym_extended] = ACTIONS(3010),
    [anon_sym_async] = ACTIONS(3010),
    [anon_sym_sub] = ACTIONS(3010),
    [anon_sym_method] = ACTIONS(3010),
    [anon_sym_LPAREN] = ACTIONS(3010),
    [anon_sym_continue] = ACTIONS(3010),
    [anon_sym_state] = ACTIONS(3010),
    [anon_sym_our] = ACTIONS(3010),
    [anon_sym_try] = ACTIONS(3010),
    [anon_sym_defer] = ACTIONS(3010),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3010),
    [anon_sym_LBRACK] = ACTIONS(3010),
    [anon_sym_LT_LT] = ACTIONS(3010),
    [anon_sym_SLASH_SLASH] = ACTIONS(3010),
    [anon_sym_PLUS] = ACTIONS(3010),
    [anon_sym_DASH] = ACTIONS(3010),
    [anon_sym_TILDE] = ACTIONS(3010),
    [anon_sym_BANG] = ACTIONS(3010),
    [anon_sym_PLUS_PLUS] = ACTIONS(3010),
    [anon_sym_DASH_DASH] = ACTIONS(3010),
    [anon_sym_BSLASH] = ACTIONS(3008),
    [anon_sym_do] = ACTIONS(3010),
    [anon_sym_eval] = ACTIONS(3010),
    [anon_sym_await] = ACTIONS(3010),
    [anon_sym_field] = ACTIONS(3010),
    [anon_sym_local] = ACTIONS(3010),
    [anon_sym_dynamically] = ACTIONS(3010),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3010),
    [anon_sym_require] = ACTIONS(3010),
    [anon_sym_map] = ACTIONS(3010),
    [anon_sym_grep] = ACTIONS(3010),
    [anon_sym_sort] = ACTIONS(3010),
    [anon_sym_goto] = ACTIONS(3010),
    [anon_sym_return] = ACTIONS(3010),
    [anon_sym_undef] = ACTIONS(3010),
    [anon_sym_PERCENT2] = ACTIONS(3010),
    [anon_sym_AMP2] = ACTIONS(3010),
    [anon_sym_STAR2] = ACTIONS(3010),
    [anon_sym_use] = ACTIONS(3010),
    [anon_sym_no] = ACTIONS(3010),
    [anon_sym_for] = ACTIONS(3010),
    [anon_sym_foreach] = ACTIONS(3010),
    [anon_sym_last] = ACTIONS(3010),
    [anon_sym_next] = ACTIONS(3010),
    [anon_sym_redo] = ACTIONS(3010),
    [anon_sym_BEGIN] = ACTIONS(3010),
    [anon_sym_INIT] = ACTIONS(3010),
    [anon_sym_CHECK] = ACTIONS(3010),
    [anon_sym_UNITCHECK] = ACTIONS(3010),
    [anon_sym_END] = ACTIONS(3010),
    [anon_sym___FILE__] = ACTIONS(3010),
    [anon_sym___LINE__] = ACTIONS(3010),
    [anon_sym___PACKAGE__] = ACTIONS(3010),
    [anon_sym___SUB__] = ACTIONS(3010),
    [anon_sym_break] = ACTIONS(3010),
    [anon_sym_fork] = ACTIONS(3010),
    [anon_sym_getppid] = ACTIONS(3010),
    [anon_sym_time] = ACTIONS(3010),
    [anon_sym_times] = ACTIONS(3010),
    [anon_sym_wait] = ACTIONS(3010),
    [anon_sym_wantarray] = ACTIONS(3010),
    [anon_sym_abs] = ACTIONS(3010),
    [anon_sym_alarm] = ACTIONS(3010),
    [anon_sym_chop] = ACTIONS(3010),
    [anon_sym_chdir] = ACTIONS(3010),
    [anon_sym_close] = ACTIONS(3010),
    [anon_sym_closedir] = ACTIONS(3010),
    [anon_sym_caller] = ACTIONS(3010),
    [anon_sym_chomp] = ACTIONS(3010),
    [anon_sym_chr] = ACTIONS(3010),
    [anon_sym_cos] = ACTIONS(3010),
    [anon_sym_chroot] = ACTIONS(3010),
    [anon_sym_defined] = ACTIONS(3010),
    [anon_sym_delete] = ACTIONS(3010),
    [anon_sym_dbmclose] = ACTIONS(3010),
    [anon_sym_exists] = ACTIONS(3010),
    [anon_sym_exit] = ACTIONS(3010),
    [anon_sym_eof] = ACTIONS(3010),
    [anon_sym_exp] = ACTIONS(3010),
    [anon_sym_each] = ACTIONS(3010),
    [anon_sym_fc] = ACTIONS(3010),
    [anon_sym_fileno] = ACTIONS(3010),
    [anon_sym_gmtime] = ACTIONS(3010),
    [anon_sym_getc] = ACTIONS(3010),
    [anon_sym_getpgrp] = ACTIONS(3010),
    [anon_sym_getprotobyname] = ACTIONS(3010),
    [anon_sym_getpwname] = ACTIONS(3010),
    [anon_sym_getpwuid] = ACTIONS(3010),
    [anon_sym_getpeername] = ACTIONS(3010),
    [anon_sym_getnetbyname] = ACTIONS(3010),
    [anon_sym_getsockname] = ACTIONS(3010),
    [anon_sym_getgrnam] = ACTIONS(3010),
    [anon_sym_getgrgid] = ACTIONS(3010),
    [anon_sym_hex] = ACTIONS(3010),
    [anon_sym_int] = ACTIONS(3010),
    [anon_sym_keys] = ACTIONS(3010),
    [anon_sym_lc] = ACTIONS(3010),
    [anon_sym_lcfirst] = ACTIONS(3010),
    [anon_sym_length] = ACTIONS(3010),
    [anon_sym_localtime] = ACTIONS(3010),
    [anon_sym_log] = ACTIONS(3010),
    [anon_sym_lock] = ACTIONS(3010),
    [anon_sym_lstat] = ACTIONS(3010),
    [anon_sym_oct] = ACTIONS(3010),
    [anon_sym_ord] = ACTIONS(3010),
    [anon_sym_prototype] = ACTIONS(3010),
    [anon_sym_pop] = ACTIONS(3010),
    [anon_sym_pos] = ACTIONS(3010),
    [anon_sym_quotemeta] = ACTIONS(3010),
    [anon_sym_reset] = ACTIONS(3010),
    [anon_sym_rand] = ACTIONS(3010),
    [anon_sym_rmdir] = ACTIONS(3010),
    [anon_sym_readdir] = ACTIONS(3010),
    [anon_sym_readline] = ACTIONS(3010),
    [anon_sym_readpipe] = ACTIONS(3010),
    [anon_sym_rewinddir] = ACTIONS(3010),
    [anon_sym_readlink] = ACTIONS(3010),
    [anon_sym_ref] = ACTIONS(3010),
    [anon_sym_scalar] = ACTIONS(3010),
    [anon_sym_shift] = ACTIONS(3010),
    [anon_sym_sin] = ACTIONS(3010),
    [anon_sym_sleep] = ACTIONS(3010),
    [anon_sym_sqrt] = ACTIONS(3010),
    [anon_sym_srand] = ACTIONS(3010),
    [anon_sym_stat] = ACTIONS(3010),
    [anon_sym_study] = ACTIONS(3010),
    [anon_sym_tell] = ACTIONS(3010),
    [anon_sym_telldir] = ACTIONS(3010),
    [anon_sym_tied] = ACTIONS(3010),
    [anon_sym_uc] = ACTIONS(3010),
    [anon_sym_ucfirst] = ACTIONS(3010),
    [anon_sym_untie] = ACTIONS(3010),
    [anon_sym_umask] = ACTIONS(3010),
    [anon_sym_values] = ACTIONS(3010),
    [anon_sym_write] = ACTIONS(3010),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3010),
    [anon_sym_qq] = ACTIONS(3010),
    [anon_sym_qw] = ACTIONS(3010),
    [anon_sym_qx] = ACTIONS(3010),
    [anon_sym_qr] = ACTIONS(3010),
    [anon_sym_m] = ACTIONS(3010),
    [anon_sym_s] = ACTIONS(3010),
    [anon_sym_tr] = ACTIONS(3010),
    [anon_sym_y] = ACTIONS(3010),
    [anon_sym_if] = ACTIONS(3010),
    [anon_sym_unless] = ACTIONS(3010),
    [anon_sym_while] = ACTIONS(3010),
    [anon_sym_until] = ACTIONS(3010),
    [aux_sym__bareword_token1] = ACTIONS(3010),
    [sym_number] = ACTIONS(3010),
    [anon_sym_true] = ACTIONS(3010),
    [anon_sym_false] = ACTIONS(3010),
    [sym__single_quote] = ACTIONS(3008),
    [sym__double_quote] = ACTIONS(3008),
    [sym__backtick_quote] = ACTIONS(3008),
    [sym__search_slash_quote] = ACTIONS(3008),
    [sym__open_readline_bracket] = ACTIONS(3008),
    [sym__open_fileglob_bracket] = ACTIONS(3008),
    [sym__PERLY_HEREDOC] = ACTIONS(3008),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3008),
    [sym__filetest] = ACTIONS(3008),
  },
  [STATE(1556)] = {
    [sym_heredoc_content] = STATE(1556),
    [sym__identifier] = ACTIONS(3037),
    [anon_sym_LBRACE] = ACTIONS(3037),
    [anon_sym_RBRACE] = ACTIONS(3037),
    [anon_sym_SEMI] = ACTIONS(3037),
    [anon_sym_package] = ACTIONS(3037),
    [anon_sym_class] = ACTIONS(3037),
    [anon_sym_role] = ACTIONS(3037),
    [anon_sym_BUILD] = ACTIONS(3037),
    [anon_sym_ADJUST] = ACTIONS(3037),
    [anon_sym_DOLLAR] = ACTIONS(3037),
    [anon_sym_AT] = ACTIONS(3037),
    [anon_sym_my] = ACTIONS(3037),
    [anon_sym_extended] = ACTIONS(3037),
    [anon_sym_async] = ACTIONS(3037),
    [anon_sym_sub] = ACTIONS(3037),
    [anon_sym_method] = ACTIONS(3037),
    [anon_sym_LPAREN] = ACTIONS(3037),
    [anon_sym_continue] = ACTIONS(3037),
    [anon_sym_state] = ACTIONS(3037),
    [anon_sym_our] = ACTIONS(3037),
    [anon_sym_try] = ACTIONS(3037),
    [anon_sym_defer] = ACTIONS(3037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3037),
    [anon_sym_LBRACK] = ACTIONS(3037),
    [anon_sym_LT_LT] = ACTIONS(3037),
    [anon_sym_SLASH_SLASH] = ACTIONS(3037),
    [anon_sym_PLUS] = ACTIONS(3037),
    [anon_sym_DASH] = ACTIONS(3037),
    [anon_sym_TILDE] = ACTIONS(3037),
    [anon_sym_BANG] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH] = ACTIONS(3037),
    [anon_sym_BSLASH] = ACTIONS(3035),
    [anon_sym_do] = ACTIONS(3037),
    [anon_sym_eval] = ACTIONS(3037),
    [anon_sym_await] = ACTIONS(3037),
    [anon_sym_field] = ACTIONS(3037),
    [anon_sym_local] = ACTIONS(3037),
    [anon_sym_dynamically] = ACTIONS(3037),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3037),
    [anon_sym_require] = ACTIONS(3037),
    [anon_sym_map] = ACTIONS(3037),
    [anon_sym_grep] = ACTIONS(3037),
    [anon_sym_sort] = ACTIONS(3037),
    [anon_sym_goto] = ACTIONS(3037),
    [anon_sym_return] = ACTIONS(3037),
    [anon_sym_undef] = ACTIONS(3037),
    [anon_sym_PERCENT2] = ACTIONS(3037),
    [anon_sym_AMP2] = ACTIONS(3037),
    [anon_sym_STAR2] = ACTIONS(3037),
    [anon_sym_use] = ACTIONS(3037),
    [anon_sym_no] = ACTIONS(3037),
    [anon_sym_for] = ACTIONS(3037),
    [anon_sym_foreach] = ACTIONS(3037),
    [anon_sym_last] = ACTIONS(3037),
    [anon_sym_next] = ACTIONS(3037),
    [anon_sym_redo] = ACTIONS(3037),
    [anon_sym_BEGIN] = ACTIONS(3037),
    [anon_sym_INIT] = ACTIONS(3037),
    [anon_sym_CHECK] = ACTIONS(3037),
    [anon_sym_UNITCHECK] = ACTIONS(3037),
    [anon_sym_END] = ACTIONS(3037),
    [anon_sym___FILE__] = ACTIONS(3037),
    [anon_sym___LINE__] = ACTIONS(3037),
    [anon_sym___PACKAGE__] = ACTIONS(3037),
    [anon_sym___SUB__] = ACTIONS(3037),
    [anon_sym_break] = ACTIONS(3037),
    [anon_sym_fork] = ACTIONS(3037),
    [anon_sym_getppid] = ACTIONS(3037),
    [anon_sym_time] = ACTIONS(3037),
    [anon_sym_times] = ACTIONS(3037),
    [anon_sym_wait] = ACTIONS(3037),
    [anon_sym_wantarray] = ACTIONS(3037),
    [anon_sym_abs] = ACTIONS(3037),
    [anon_sym_alarm] = ACTIONS(3037),
    [anon_sym_chop] = ACTIONS(3037),
    [anon_sym_chdir] = ACTIONS(3037),
    [anon_sym_close] = ACTIONS(3037),
    [anon_sym_closedir] = ACTIONS(3037),
    [anon_sym_caller] = ACTIONS(3037),
    [anon_sym_chomp] = ACTIONS(3037),
    [anon_sym_chr] = ACTIONS(3037),
    [anon_sym_cos] = ACTIONS(3037),
    [anon_sym_chroot] = ACTIONS(3037),
    [anon_sym_defined] = ACTIONS(3037),
    [anon_sym_delete] = ACTIONS(3037),
    [anon_sym_dbmclose] = ACTIONS(3037),
    [anon_sym_exists] = ACTIONS(3037),
    [anon_sym_exit] = ACTIONS(3037),
    [anon_sym_eof] = ACTIONS(3037),
    [anon_sym_exp] = ACTIONS(3037),
    [anon_sym_each] = ACTIONS(3037),
    [anon_sym_fc] = ACTIONS(3037),
    [anon_sym_fileno] = ACTIONS(3037),
    [anon_sym_gmtime] = ACTIONS(3037),
    [anon_sym_getc] = ACTIONS(3037),
    [anon_sym_getpgrp] = ACTIONS(3037),
    [anon_sym_getprotobyname] = ACTIONS(3037),
    [anon_sym_getpwname] = ACTIONS(3037),
    [anon_sym_getpwuid] = ACTIONS(3037),
    [anon_sym_getpeername] = ACTIONS(3037),
    [anon_sym_getnetbyname] = ACTIONS(3037),
    [anon_sym_getsockname] = ACTIONS(3037),
    [anon_sym_getgrnam] = ACTIONS(3037),
    [anon_sym_getgrgid] = ACTIONS(3037),
    [anon_sym_hex] = ACTIONS(3037),
    [anon_sym_int] = ACTIONS(3037),
    [anon_sym_keys] = ACTIONS(3037),
    [anon_sym_lc] = ACTIONS(3037),
    [anon_sym_lcfirst] = ACTIONS(3037),
    [anon_sym_length] = ACTIONS(3037),
    [anon_sym_localtime] = ACTIONS(3037),
    [anon_sym_log] = ACTIONS(3037),
    [anon_sym_lock] = ACTIONS(3037),
    [anon_sym_lstat] = ACTIONS(3037),
    [anon_sym_oct] = ACTIONS(3037),
    [anon_sym_ord] = ACTIONS(3037),
    [anon_sym_prototype] = ACTIONS(3037),
    [anon_sym_pop] = ACTIONS(3037),
    [anon_sym_pos] = ACTIONS(3037),
    [anon_sym_quotemeta] = ACTIONS(3037),
    [anon_sym_reset] = ACTIONS(3037),
    [anon_sym_rand] = ACTIONS(3037),
    [anon_sym_rmdir] = ACTIONS(3037),
    [anon_sym_readdir] = ACTIONS(3037),
    [anon_sym_readline] = ACTIONS(3037),
    [anon_sym_readpipe] = ACTIONS(3037),
    [anon_sym_rewinddir] = ACTIONS(3037),
    [anon_sym_readlink] = ACTIONS(3037),
    [anon_sym_ref] = ACTIONS(3037),
    [anon_sym_scalar] = ACTIONS(3037),
    [anon_sym_shift] = ACTIONS(3037),
    [anon_sym_sin] = ACTIONS(3037),
    [anon_sym_sleep] = ACTIONS(3037),
    [anon_sym_sqrt] = ACTIONS(3037),
    [anon_sym_srand] = ACTIONS(3037),
    [anon_sym_stat] = ACTIONS(3037),
    [anon_sym_study] = ACTIONS(3037),
    [anon_sym_tell] = ACTIONS(3037),
    [anon_sym_telldir] = ACTIONS(3037),
    [anon_sym_tied] = ACTIONS(3037),
    [anon_sym_uc] = ACTIONS(3037),
    [anon_sym_ucfirst] = ACTIONS(3037),
    [anon_sym_untie] = ACTIONS(3037),
    [anon_sym_umask] = ACTIONS(3037),
    [anon_sym_values] = ACTIONS(3037),
    [anon_sym_write] = ACTIONS(3037),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3037),
    [anon_sym_qq] = ACTIONS(3037),
    [anon_sym_qw] = ACTIONS(3037),
    [anon_sym_qx] = ACTIONS(3037),
    [anon_sym_qr] = ACTIONS(3037),
    [anon_sym_m] = ACTIONS(3037),
    [anon_sym_s] = ACTIONS(3037),
    [anon_sym_tr] = ACTIONS(3037),
    [anon_sym_y] = ACTIONS(3037),
    [anon_sym_if] = ACTIONS(3037),
    [anon_sym_unless] = ACTIONS(3037),
    [anon_sym_while] = ACTIONS(3037),
    [anon_sym_until] = ACTIONS(3037),
    [aux_sym__bareword_token1] = ACTIONS(3037),
    [sym_number] = ACTIONS(3037),
    [anon_sym_true] = ACTIONS(3037),
    [anon_sym_false] = ACTIONS(3037),
    [sym__single_quote] = ACTIONS(3035),
    [sym__double_quote] = ACTIONS(3035),
    [sym__backtick_quote] = ACTIONS(3035),
    [sym__search_slash_quote] = ACTIONS(3035),
    [sym__open_readline_bracket] = ACTIONS(3035),
    [sym__open_fileglob_bracket] = ACTIONS(3035),
    [sym__PERLY_HEREDOC] = ACTIONS(3035),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3035),
    [sym__filetest] = ACTIONS(3035),
  },
  [STATE(1557)] = {
    [sym_heredoc_content] = STATE(1557),
    [sym__identifier] = ACTIONS(3237),
    [anon_sym_LBRACE] = ACTIONS(3237),
    [anon_sym_RBRACE] = ACTIONS(3237),
    [anon_sym_SEMI] = ACTIONS(3237),
    [anon_sym_package] = ACTIONS(3237),
    [anon_sym_class] = ACTIONS(3237),
    [anon_sym_role] = ACTIONS(3237),
    [anon_sym_BUILD] = ACTIONS(3237),
    [anon_sym_ADJUST] = ACTIONS(3237),
    [anon_sym_DOLLAR] = ACTIONS(3237),
    [anon_sym_AT] = ACTIONS(3237),
    [anon_sym_my] = ACTIONS(3237),
    [anon_sym_extended] = ACTIONS(3237),
    [anon_sym_async] = ACTIONS(3237),
    [anon_sym_sub] = ACTIONS(3237),
    [anon_sym_method] = ACTIONS(3237),
    [anon_sym_LPAREN] = ACTIONS(3237),
    [anon_sym_continue] = ACTIONS(3237),
    [anon_sym_state] = ACTIONS(3237),
    [anon_sym_our] = ACTIONS(3237),
    [anon_sym_try] = ACTIONS(3237),
    [anon_sym_defer] = ACTIONS(3237),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3237),
    [anon_sym_LBRACK] = ACTIONS(3237),
    [anon_sym_LT_LT] = ACTIONS(3237),
    [anon_sym_SLASH_SLASH] = ACTIONS(3237),
    [anon_sym_PLUS] = ACTIONS(3237),
    [anon_sym_DASH] = ACTIONS(3237),
    [anon_sym_TILDE] = ACTIONS(3237),
    [anon_sym_BANG] = ACTIONS(3237),
    [anon_sym_PLUS_PLUS] = ACTIONS(3237),
    [anon_sym_DASH_DASH] = ACTIONS(3237),
    [anon_sym_BSLASH] = ACTIONS(3235),
    [anon_sym_do] = ACTIONS(3237),
    [anon_sym_eval] = ACTIONS(3237),
    [anon_sym_await] = ACTIONS(3237),
    [anon_sym_field] = ACTIONS(3237),
    [anon_sym_local] = ACTIONS(3237),
    [anon_sym_dynamically] = ACTIONS(3237),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3237),
    [anon_sym_require] = ACTIONS(3237),
    [anon_sym_map] = ACTIONS(3237),
    [anon_sym_grep] = ACTIONS(3237),
    [anon_sym_sort] = ACTIONS(3237),
    [anon_sym_goto] = ACTIONS(3237),
    [anon_sym_return] = ACTIONS(3237),
    [anon_sym_undef] = ACTIONS(3237),
    [anon_sym_PERCENT2] = ACTIONS(3237),
    [anon_sym_AMP2] = ACTIONS(3237),
    [anon_sym_STAR2] = ACTIONS(3237),
    [anon_sym_use] = ACTIONS(3237),
    [anon_sym_no] = ACTIONS(3237),
    [anon_sym_for] = ACTIONS(3237),
    [anon_sym_foreach] = ACTIONS(3237),
    [anon_sym_last] = ACTIONS(3237),
    [anon_sym_next] = ACTIONS(3237),
    [anon_sym_redo] = ACTIONS(3237),
    [anon_sym_BEGIN] = ACTIONS(3237),
    [anon_sym_INIT] = ACTIONS(3237),
    [anon_sym_CHECK] = ACTIONS(3237),
    [anon_sym_UNITCHECK] = ACTIONS(3237),
    [anon_sym_END] = ACTIONS(3237),
    [anon_sym___FILE__] = ACTIONS(3237),
    [anon_sym___LINE__] = ACTIONS(3237),
    [anon_sym___PACKAGE__] = ACTIONS(3237),
    [anon_sym___SUB__] = ACTIONS(3237),
    [anon_sym_break] = ACTIONS(3237),
    [anon_sym_fork] = ACTIONS(3237),
    [anon_sym_getppid] = ACTIONS(3237),
    [anon_sym_time] = ACTIONS(3237),
    [anon_sym_times] = ACTIONS(3237),
    [anon_sym_wait] = ACTIONS(3237),
    [anon_sym_wantarray] = ACTIONS(3237),
    [anon_sym_abs] = ACTIONS(3237),
    [anon_sym_alarm] = ACTIONS(3237),
    [anon_sym_chop] = ACTIONS(3237),
    [anon_sym_chdir] = ACTIONS(3237),
    [anon_sym_close] = ACTIONS(3237),
    [anon_sym_closedir] = ACTIONS(3237),
    [anon_sym_caller] = ACTIONS(3237),
    [anon_sym_chomp] = ACTIONS(3237),
    [anon_sym_chr] = ACTIONS(3237),
    [anon_sym_cos] = ACTIONS(3237),
    [anon_sym_chroot] = ACTIONS(3237),
    [anon_sym_defined] = ACTIONS(3237),
    [anon_sym_delete] = ACTIONS(3237),
    [anon_sym_dbmclose] = ACTIONS(3237),
    [anon_sym_exists] = ACTIONS(3237),
    [anon_sym_exit] = ACTIONS(3237),
    [anon_sym_eof] = ACTIONS(3237),
    [anon_sym_exp] = ACTIONS(3237),
    [anon_sym_each] = ACTIONS(3237),
    [anon_sym_fc] = ACTIONS(3237),
    [anon_sym_fileno] = ACTIONS(3237),
    [anon_sym_gmtime] = ACTIONS(3237),
    [anon_sym_getc] = ACTIONS(3237),
    [anon_sym_getpgrp] = ACTIONS(3237),
    [anon_sym_getprotobyname] = ACTIONS(3237),
    [anon_sym_getpwname] = ACTIONS(3237),
    [anon_sym_getpwuid] = ACTIONS(3237),
    [anon_sym_getpeername] = ACTIONS(3237),
    [anon_sym_getnetbyname] = ACTIONS(3237),
    [anon_sym_getsockname] = ACTIONS(3237),
    [anon_sym_getgrnam] = ACTIONS(3237),
    [anon_sym_getgrgid] = ACTIONS(3237),
    [anon_sym_hex] = ACTIONS(3237),
    [anon_sym_int] = ACTIONS(3237),
    [anon_sym_keys] = ACTIONS(3237),
    [anon_sym_lc] = ACTIONS(3237),
    [anon_sym_lcfirst] = ACTIONS(3237),
    [anon_sym_length] = ACTIONS(3237),
    [anon_sym_localtime] = ACTIONS(3237),
    [anon_sym_log] = ACTIONS(3237),
    [anon_sym_lock] = ACTIONS(3237),
    [anon_sym_lstat] = ACTIONS(3237),
    [anon_sym_oct] = ACTIONS(3237),
    [anon_sym_ord] = ACTIONS(3237),
    [anon_sym_prototype] = ACTIONS(3237),
    [anon_sym_pop] = ACTIONS(3237),
    [anon_sym_pos] = ACTIONS(3237),
    [anon_sym_quotemeta] = ACTIONS(3237),
    [anon_sym_reset] = ACTIONS(3237),
    [anon_sym_rand] = ACTIONS(3237),
    [anon_sym_rmdir] = ACTIONS(3237),
    [anon_sym_readdir] = ACTIONS(3237),
    [anon_sym_readline] = ACTIONS(3237),
    [anon_sym_readpipe] = ACTIONS(3237),
    [anon_sym_rewinddir] = ACTIONS(3237),
    [anon_sym_readlink] = ACTIONS(3237),
    [anon_sym_ref] = ACTIONS(3237),
    [anon_sym_scalar] = ACTIONS(3237),
    [anon_sym_shift] = ACTIONS(3237),
    [anon_sym_sin] = ACTIONS(3237),
    [anon_sym_sleep] = ACTIONS(3237),
    [anon_sym_sqrt] = ACTIONS(3237),
    [anon_sym_srand] = ACTIONS(3237),
    [anon_sym_stat] = ACTIONS(3237),
    [anon_sym_study] = ACTIONS(3237),
    [anon_sym_tell] = ACTIONS(3237),
    [anon_sym_telldir] = ACTIONS(3237),
    [anon_sym_tied] = ACTIONS(3237),
    [anon_sym_uc] = ACTIONS(3237),
    [anon_sym_ucfirst] = ACTIONS(3237),
    [anon_sym_untie] = ACTIONS(3237),
    [anon_sym_umask] = ACTIONS(3237),
    [anon_sym_values] = ACTIONS(3237),
    [anon_sym_write] = ACTIONS(3237),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3237),
    [anon_sym_qq] = ACTIONS(3237),
    [anon_sym_qw] = ACTIONS(3237),
    [anon_sym_qx] = ACTIONS(3237),
    [anon_sym_qr] = ACTIONS(3237),
    [anon_sym_m] = ACTIONS(3237),
    [anon_sym_s] = ACTIONS(3237),
    [anon_sym_tr] = ACTIONS(3237),
    [anon_sym_y] = ACTIONS(3237),
    [anon_sym_if] = ACTIONS(3237),
    [anon_sym_unless] = ACTIONS(3237),
    [anon_sym_while] = ACTIONS(3237),
    [anon_sym_until] = ACTIONS(3237),
    [aux_sym__bareword_token1] = ACTIONS(3237),
    [sym_number] = ACTIONS(3237),
    [anon_sym_true] = ACTIONS(3237),
    [anon_sym_false] = ACTIONS(3237),
    [sym__single_quote] = ACTIONS(3235),
    [sym__double_quote] = ACTIONS(3235),
    [sym__backtick_quote] = ACTIONS(3235),
    [sym__search_slash_quote] = ACTIONS(3235),
    [sym__open_readline_bracket] = ACTIONS(3235),
    [sym__open_fileglob_bracket] = ACTIONS(3235),
    [sym__PERLY_HEREDOC] = ACTIONS(3235),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3235),
    [sym__filetest] = ACTIONS(3235),
  },
  [STATE(1558)] = {
    [sym_heredoc_content] = STATE(1558),
    [sym__identifier] = ACTIONS(3245),
    [anon_sym_LBRACE] = ACTIONS(3245),
    [anon_sym_RBRACE] = ACTIONS(3245),
    [anon_sym_SEMI] = ACTIONS(3245),
    [anon_sym_package] = ACTIONS(3245),
    [anon_sym_class] = ACTIONS(3245),
    [anon_sym_role] = ACTIONS(3245),
    [anon_sym_BUILD] = ACTIONS(3245),
    [anon_sym_ADJUST] = ACTIONS(3245),
    [anon_sym_DOLLAR] = ACTIONS(3245),
    [anon_sym_AT] = ACTIONS(3245),
    [anon_sym_my] = ACTIONS(3245),
    [anon_sym_extended] = ACTIONS(3245),
    [anon_sym_async] = ACTIONS(3245),
    [anon_sym_sub] = ACTIONS(3245),
    [anon_sym_method] = ACTIONS(3245),
    [anon_sym_LPAREN] = ACTIONS(3245),
    [anon_sym_continue] = ACTIONS(3245),
    [anon_sym_state] = ACTIONS(3245),
    [anon_sym_our] = ACTIONS(3245),
    [anon_sym_try] = ACTIONS(3245),
    [anon_sym_defer] = ACTIONS(3245),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3245),
    [anon_sym_LBRACK] = ACTIONS(3245),
    [anon_sym_LT_LT] = ACTIONS(3245),
    [anon_sym_SLASH_SLASH] = ACTIONS(3245),
    [anon_sym_PLUS] = ACTIONS(3245),
    [anon_sym_DASH] = ACTIONS(3245),
    [anon_sym_TILDE] = ACTIONS(3245),
    [anon_sym_BANG] = ACTIONS(3245),
    [anon_sym_PLUS_PLUS] = ACTIONS(3245),
    [anon_sym_DASH_DASH] = ACTIONS(3245),
    [anon_sym_BSLASH] = ACTIONS(3243),
    [anon_sym_do] = ACTIONS(3245),
    [anon_sym_eval] = ACTIONS(3245),
    [anon_sym_await] = ACTIONS(3245),
    [anon_sym_field] = ACTIONS(3245),
    [anon_sym_local] = ACTIONS(3245),
    [anon_sym_dynamically] = ACTIONS(3245),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3245),
    [anon_sym_require] = ACTIONS(3245),
    [anon_sym_map] = ACTIONS(3245),
    [anon_sym_grep] = ACTIONS(3245),
    [anon_sym_sort] = ACTIONS(3245),
    [anon_sym_goto] = ACTIONS(3245),
    [anon_sym_return] = ACTIONS(3245),
    [anon_sym_undef] = ACTIONS(3245),
    [anon_sym_PERCENT2] = ACTIONS(3245),
    [anon_sym_AMP2] = ACTIONS(3245),
    [anon_sym_STAR2] = ACTIONS(3245),
    [anon_sym_use] = ACTIONS(3245),
    [anon_sym_no] = ACTIONS(3245),
    [anon_sym_for] = ACTIONS(3245),
    [anon_sym_foreach] = ACTIONS(3245),
    [anon_sym_last] = ACTIONS(3245),
    [anon_sym_next] = ACTIONS(3245),
    [anon_sym_redo] = ACTIONS(3245),
    [anon_sym_BEGIN] = ACTIONS(3245),
    [anon_sym_INIT] = ACTIONS(3245),
    [anon_sym_CHECK] = ACTIONS(3245),
    [anon_sym_UNITCHECK] = ACTIONS(3245),
    [anon_sym_END] = ACTIONS(3245),
    [anon_sym___FILE__] = ACTIONS(3245),
    [anon_sym___LINE__] = ACTIONS(3245),
    [anon_sym___PACKAGE__] = ACTIONS(3245),
    [anon_sym___SUB__] = ACTIONS(3245),
    [anon_sym_break] = ACTIONS(3245),
    [anon_sym_fork] = ACTIONS(3245),
    [anon_sym_getppid] = ACTIONS(3245),
    [anon_sym_time] = ACTIONS(3245),
    [anon_sym_times] = ACTIONS(3245),
    [anon_sym_wait] = ACTIONS(3245),
    [anon_sym_wantarray] = ACTIONS(3245),
    [anon_sym_abs] = ACTIONS(3245),
    [anon_sym_alarm] = ACTIONS(3245),
    [anon_sym_chop] = ACTIONS(3245),
    [anon_sym_chdir] = ACTIONS(3245),
    [anon_sym_close] = ACTIONS(3245),
    [anon_sym_closedir] = ACTIONS(3245),
    [anon_sym_caller] = ACTIONS(3245),
    [anon_sym_chomp] = ACTIONS(3245),
    [anon_sym_chr] = ACTIONS(3245),
    [anon_sym_cos] = ACTIONS(3245),
    [anon_sym_chroot] = ACTIONS(3245),
    [anon_sym_defined] = ACTIONS(3245),
    [anon_sym_delete] = ACTIONS(3245),
    [anon_sym_dbmclose] = ACTIONS(3245),
    [anon_sym_exists] = ACTIONS(3245),
    [anon_sym_exit] = ACTIONS(3245),
    [anon_sym_eof] = ACTIONS(3245),
    [anon_sym_exp] = ACTIONS(3245),
    [anon_sym_each] = ACTIONS(3245),
    [anon_sym_fc] = ACTIONS(3245),
    [anon_sym_fileno] = ACTIONS(3245),
    [anon_sym_gmtime] = ACTIONS(3245),
    [anon_sym_getc] = ACTIONS(3245),
    [anon_sym_getpgrp] = ACTIONS(3245),
    [anon_sym_getprotobyname] = ACTIONS(3245),
    [anon_sym_getpwname] = ACTIONS(3245),
    [anon_sym_getpwuid] = ACTIONS(3245),
    [anon_sym_getpeername] = ACTIONS(3245),
    [anon_sym_getnetbyname] = ACTIONS(3245),
    [anon_sym_getsockname] = ACTIONS(3245),
    [anon_sym_getgrnam] = ACTIONS(3245),
    [anon_sym_getgrgid] = ACTIONS(3245),
    [anon_sym_hex] = ACTIONS(3245),
    [anon_sym_int] = ACTIONS(3245),
    [anon_sym_keys] = ACTIONS(3245),
    [anon_sym_lc] = ACTIONS(3245),
    [anon_sym_lcfirst] = ACTIONS(3245),
    [anon_sym_length] = ACTIONS(3245),
    [anon_sym_localtime] = ACTIONS(3245),
    [anon_sym_log] = ACTIONS(3245),
    [anon_sym_lock] = ACTIONS(3245),
    [anon_sym_lstat] = ACTIONS(3245),
    [anon_sym_oct] = ACTIONS(3245),
    [anon_sym_ord] = ACTIONS(3245),
    [anon_sym_prototype] = ACTIONS(3245),
    [anon_sym_pop] = ACTIONS(3245),
    [anon_sym_pos] = ACTIONS(3245),
    [anon_sym_quotemeta] = ACTIONS(3245),
    [anon_sym_reset] = ACTIONS(3245),
    [anon_sym_rand] = ACTIONS(3245),
    [anon_sym_rmdir] = ACTIONS(3245),
    [anon_sym_readdir] = ACTIONS(3245),
    [anon_sym_readline] = ACTIONS(3245),
    [anon_sym_readpipe] = ACTIONS(3245),
    [anon_sym_rewinddir] = ACTIONS(3245),
    [anon_sym_readlink] = ACTIONS(3245),
    [anon_sym_ref] = ACTIONS(3245),
    [anon_sym_scalar] = ACTIONS(3245),
    [anon_sym_shift] = ACTIONS(3245),
    [anon_sym_sin] = ACTIONS(3245),
    [anon_sym_sleep] = ACTIONS(3245),
    [anon_sym_sqrt] = ACTIONS(3245),
    [anon_sym_srand] = ACTIONS(3245),
    [anon_sym_stat] = ACTIONS(3245),
    [anon_sym_study] = ACTIONS(3245),
    [anon_sym_tell] = ACTIONS(3245),
    [anon_sym_telldir] = ACTIONS(3245),
    [anon_sym_tied] = ACTIONS(3245),
    [anon_sym_uc] = ACTIONS(3245),
    [anon_sym_ucfirst] = ACTIONS(3245),
    [anon_sym_untie] = ACTIONS(3245),
    [anon_sym_umask] = ACTIONS(3245),
    [anon_sym_values] = ACTIONS(3245),
    [anon_sym_write] = ACTIONS(3245),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3245),
    [anon_sym_qq] = ACTIONS(3245),
    [anon_sym_qw] = ACTIONS(3245),
    [anon_sym_qx] = ACTIONS(3245),
    [anon_sym_qr] = ACTIONS(3245),
    [anon_sym_m] = ACTIONS(3245),
    [anon_sym_s] = ACTIONS(3245),
    [anon_sym_tr] = ACTIONS(3245),
    [anon_sym_y] = ACTIONS(3245),
    [anon_sym_if] = ACTIONS(3245),
    [anon_sym_unless] = ACTIONS(3245),
    [anon_sym_while] = ACTIONS(3245),
    [anon_sym_until] = ACTIONS(3245),
    [aux_sym__bareword_token1] = ACTIONS(3245),
    [sym_number] = ACTIONS(3245),
    [anon_sym_true] = ACTIONS(3245),
    [anon_sym_false] = ACTIONS(3245),
    [sym__single_quote] = ACTIONS(3243),
    [sym__double_quote] = ACTIONS(3243),
    [sym__backtick_quote] = ACTIONS(3243),
    [sym__search_slash_quote] = ACTIONS(3243),
    [sym__open_readline_bracket] = ACTIONS(3243),
    [sym__open_fileglob_bracket] = ACTIONS(3243),
    [sym__PERLY_HEREDOC] = ACTIONS(3243),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3243),
    [sym__filetest] = ACTIONS(3243),
  },
  [STATE(1559)] = {
    [sym_heredoc_content] = STATE(1559),
    [sym__identifier] = ACTIONS(3249),
    [anon_sym_LBRACE] = ACTIONS(3249),
    [anon_sym_RBRACE] = ACTIONS(3249),
    [anon_sym_SEMI] = ACTIONS(3249),
    [anon_sym_package] = ACTIONS(3249),
    [anon_sym_class] = ACTIONS(3249),
    [anon_sym_role] = ACTIONS(3249),
    [anon_sym_BUILD] = ACTIONS(3249),
    [anon_sym_ADJUST] = ACTIONS(3249),
    [anon_sym_DOLLAR] = ACTIONS(3249),
    [anon_sym_AT] = ACTIONS(3249),
    [anon_sym_my] = ACTIONS(3249),
    [anon_sym_extended] = ACTIONS(3249),
    [anon_sym_async] = ACTIONS(3249),
    [anon_sym_sub] = ACTIONS(3249),
    [anon_sym_method] = ACTIONS(3249),
    [anon_sym_LPAREN] = ACTIONS(3249),
    [anon_sym_continue] = ACTIONS(3249),
    [anon_sym_state] = ACTIONS(3249),
    [anon_sym_our] = ACTIONS(3249),
    [anon_sym_try] = ACTIONS(3249),
    [anon_sym_defer] = ACTIONS(3249),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3249),
    [anon_sym_LBRACK] = ACTIONS(3249),
    [anon_sym_LT_LT] = ACTIONS(3249),
    [anon_sym_SLASH_SLASH] = ACTIONS(3249),
    [anon_sym_PLUS] = ACTIONS(3249),
    [anon_sym_DASH] = ACTIONS(3249),
    [anon_sym_TILDE] = ACTIONS(3249),
    [anon_sym_BANG] = ACTIONS(3249),
    [anon_sym_PLUS_PLUS] = ACTIONS(3249),
    [anon_sym_DASH_DASH] = ACTIONS(3249),
    [anon_sym_BSLASH] = ACTIONS(3247),
    [anon_sym_do] = ACTIONS(3249),
    [anon_sym_eval] = ACTIONS(3249),
    [anon_sym_await] = ACTIONS(3249),
    [anon_sym_field] = ACTIONS(3249),
    [anon_sym_local] = ACTIONS(3249),
    [anon_sym_dynamically] = ACTIONS(3249),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3249),
    [anon_sym_require] = ACTIONS(3249),
    [anon_sym_map] = ACTIONS(3249),
    [anon_sym_grep] = ACTIONS(3249),
    [anon_sym_sort] = ACTIONS(3249),
    [anon_sym_goto] = ACTIONS(3249),
    [anon_sym_return] = ACTIONS(3249),
    [anon_sym_undef] = ACTIONS(3249),
    [anon_sym_PERCENT2] = ACTIONS(3249),
    [anon_sym_AMP2] = ACTIONS(3249),
    [anon_sym_STAR2] = ACTIONS(3249),
    [anon_sym_use] = ACTIONS(3249),
    [anon_sym_no] = ACTIONS(3249),
    [anon_sym_for] = ACTIONS(3249),
    [anon_sym_foreach] = ACTIONS(3249),
    [anon_sym_last] = ACTIONS(3249),
    [anon_sym_next] = ACTIONS(3249),
    [anon_sym_redo] = ACTIONS(3249),
    [anon_sym_BEGIN] = ACTIONS(3249),
    [anon_sym_INIT] = ACTIONS(3249),
    [anon_sym_CHECK] = ACTIONS(3249),
    [anon_sym_UNITCHECK] = ACTIONS(3249),
    [anon_sym_END] = ACTIONS(3249),
    [anon_sym___FILE__] = ACTIONS(3249),
    [anon_sym___LINE__] = ACTIONS(3249),
    [anon_sym___PACKAGE__] = ACTIONS(3249),
    [anon_sym___SUB__] = ACTIONS(3249),
    [anon_sym_break] = ACTIONS(3249),
    [anon_sym_fork] = ACTIONS(3249),
    [anon_sym_getppid] = ACTIONS(3249),
    [anon_sym_time] = ACTIONS(3249),
    [anon_sym_times] = ACTIONS(3249),
    [anon_sym_wait] = ACTIONS(3249),
    [anon_sym_wantarray] = ACTIONS(3249),
    [anon_sym_abs] = ACTIONS(3249),
    [anon_sym_alarm] = ACTIONS(3249),
    [anon_sym_chop] = ACTIONS(3249),
    [anon_sym_chdir] = ACTIONS(3249),
    [anon_sym_close] = ACTIONS(3249),
    [anon_sym_closedir] = ACTIONS(3249),
    [anon_sym_caller] = ACTIONS(3249),
    [anon_sym_chomp] = ACTIONS(3249),
    [anon_sym_chr] = ACTIONS(3249),
    [anon_sym_cos] = ACTIONS(3249),
    [anon_sym_chroot] = ACTIONS(3249),
    [anon_sym_defined] = ACTIONS(3249),
    [anon_sym_delete] = ACTIONS(3249),
    [anon_sym_dbmclose] = ACTIONS(3249),
    [anon_sym_exists] = ACTIONS(3249),
    [anon_sym_exit] = ACTIONS(3249),
    [anon_sym_eof] = ACTIONS(3249),
    [anon_sym_exp] = ACTIONS(3249),
    [anon_sym_each] = ACTIONS(3249),
    [anon_sym_fc] = ACTIONS(3249),
    [anon_sym_fileno] = ACTIONS(3249),
    [anon_sym_gmtime] = ACTIONS(3249),
    [anon_sym_getc] = ACTIONS(3249),
    [anon_sym_getpgrp] = ACTIONS(3249),
    [anon_sym_getprotobyname] = ACTIONS(3249),
    [anon_sym_getpwname] = ACTIONS(3249),
    [anon_sym_getpwuid] = ACTIONS(3249),
    [anon_sym_getpeername] = ACTIONS(3249),
    [anon_sym_getnetbyname] = ACTIONS(3249),
    [anon_sym_getsockname] = ACTIONS(3249),
    [anon_sym_getgrnam] = ACTIONS(3249),
    [anon_sym_getgrgid] = ACTIONS(3249),
    [anon_sym_hex] = ACTIONS(3249),
    [anon_sym_int] = ACTIONS(3249),
    [anon_sym_keys] = ACTIONS(3249),
    [anon_sym_lc] = ACTIONS(3249),
    [anon_sym_lcfirst] = ACTIONS(3249),
    [anon_sym_length] = ACTIONS(3249),
    [anon_sym_localtime] = ACTIONS(3249),
    [anon_sym_log] = ACTIONS(3249),
    [anon_sym_lock] = ACTIONS(3249),
    [anon_sym_lstat] = ACTIONS(3249),
    [anon_sym_oct] = ACTIONS(3249),
    [anon_sym_ord] = ACTIONS(3249),
    [anon_sym_prototype] = ACTIONS(3249),
    [anon_sym_pop] = ACTIONS(3249),
    [anon_sym_pos] = ACTIONS(3249),
    [anon_sym_quotemeta] = ACTIONS(3249),
    [anon_sym_reset] = ACTIONS(3249),
    [anon_sym_rand] = ACTIONS(3249),
    [anon_sym_rmdir] = ACTIONS(3249),
    [anon_sym_readdir] = ACTIONS(3249),
    [anon_sym_readline] = ACTIONS(3249),
    [anon_sym_readpipe] = ACTIONS(3249),
    [anon_sym_rewinddir] = ACTIONS(3249),
    [anon_sym_readlink] = ACTIONS(3249),
    [anon_sym_ref] = ACTIONS(3249),
    [anon_sym_scalar] = ACTIONS(3249),
    [anon_sym_shift] = ACTIONS(3249),
    [anon_sym_sin] = ACTIONS(3249),
    [anon_sym_sleep] = ACTIONS(3249),
    [anon_sym_sqrt] = ACTIONS(3249),
    [anon_sym_srand] = ACTIONS(3249),
    [anon_sym_stat] = ACTIONS(3249),
    [anon_sym_study] = ACTIONS(3249),
    [anon_sym_tell] = ACTIONS(3249),
    [anon_sym_telldir] = ACTIONS(3249),
    [anon_sym_tied] = ACTIONS(3249),
    [anon_sym_uc] = ACTIONS(3249),
    [anon_sym_ucfirst] = ACTIONS(3249),
    [anon_sym_untie] = ACTIONS(3249),
    [anon_sym_umask] = ACTIONS(3249),
    [anon_sym_values] = ACTIONS(3249),
    [anon_sym_write] = ACTIONS(3249),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3249),
    [anon_sym_qq] = ACTIONS(3249),
    [anon_sym_qw] = ACTIONS(3249),
    [anon_sym_qx] = ACTIONS(3249),
    [anon_sym_qr] = ACTIONS(3249),
    [anon_sym_m] = ACTIONS(3249),
    [anon_sym_s] = ACTIONS(3249),
    [anon_sym_tr] = ACTIONS(3249),
    [anon_sym_y] = ACTIONS(3249),
    [anon_sym_if] = ACTIONS(3249),
    [anon_sym_unless] = ACTIONS(3249),
    [anon_sym_while] = ACTIONS(3249),
    [anon_sym_until] = ACTIONS(3249),
    [aux_sym__bareword_token1] = ACTIONS(3249),
    [sym_number] = ACTIONS(3249),
    [anon_sym_true] = ACTIONS(3249),
    [anon_sym_false] = ACTIONS(3249),
    [sym__single_quote] = ACTIONS(3247),
    [sym__double_quote] = ACTIONS(3247),
    [sym__backtick_quote] = ACTIONS(3247),
    [sym__search_slash_quote] = ACTIONS(3247),
    [sym__open_readline_bracket] = ACTIONS(3247),
    [sym__open_fileglob_bracket] = ACTIONS(3247),
    [sym__PERLY_HEREDOC] = ACTIONS(3247),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3247),
    [sym__filetest] = ACTIONS(3247),
  },
  [STATE(1560)] = {
    [sym_heredoc_content] = STATE(1560),
    [sym__identifier] = ACTIONS(3145),
    [anon_sym_LBRACE] = ACTIONS(3145),
    [anon_sym_RBRACE] = ACTIONS(3145),
    [anon_sym_SEMI] = ACTIONS(3145),
    [anon_sym_package] = ACTIONS(3145),
    [anon_sym_class] = ACTIONS(3145),
    [anon_sym_role] = ACTIONS(3145),
    [anon_sym_BUILD] = ACTIONS(3145),
    [anon_sym_ADJUST] = ACTIONS(3145),
    [anon_sym_DOLLAR] = ACTIONS(3145),
    [anon_sym_AT] = ACTIONS(3145),
    [anon_sym_my] = ACTIONS(3145),
    [anon_sym_extended] = ACTIONS(3145),
    [anon_sym_async] = ACTIONS(3145),
    [anon_sym_sub] = ACTIONS(3145),
    [anon_sym_method] = ACTIONS(3145),
    [anon_sym_LPAREN] = ACTIONS(3145),
    [anon_sym_continue] = ACTIONS(3145),
    [anon_sym_state] = ACTIONS(3145),
    [anon_sym_our] = ACTIONS(3145),
    [anon_sym_try] = ACTIONS(3145),
    [anon_sym_defer] = ACTIONS(3145),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3145),
    [anon_sym_LBRACK] = ACTIONS(3145),
    [anon_sym_LT_LT] = ACTIONS(3145),
    [anon_sym_SLASH_SLASH] = ACTIONS(3145),
    [anon_sym_PLUS] = ACTIONS(3145),
    [anon_sym_DASH] = ACTIONS(3145),
    [anon_sym_TILDE] = ACTIONS(3145),
    [anon_sym_BANG] = ACTIONS(3145),
    [anon_sym_PLUS_PLUS] = ACTIONS(3145),
    [anon_sym_DASH_DASH] = ACTIONS(3145),
    [anon_sym_BSLASH] = ACTIONS(3143),
    [anon_sym_do] = ACTIONS(3145),
    [anon_sym_eval] = ACTIONS(3145),
    [anon_sym_await] = ACTIONS(3145),
    [anon_sym_field] = ACTIONS(3145),
    [anon_sym_local] = ACTIONS(3145),
    [anon_sym_dynamically] = ACTIONS(3145),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3145),
    [anon_sym_require] = ACTIONS(3145),
    [anon_sym_map] = ACTIONS(3145),
    [anon_sym_grep] = ACTIONS(3145),
    [anon_sym_sort] = ACTIONS(3145),
    [anon_sym_goto] = ACTIONS(3145),
    [anon_sym_return] = ACTIONS(3145),
    [anon_sym_undef] = ACTIONS(3145),
    [anon_sym_PERCENT2] = ACTIONS(3145),
    [anon_sym_AMP2] = ACTIONS(3145),
    [anon_sym_STAR2] = ACTIONS(3145),
    [anon_sym_use] = ACTIONS(3145),
    [anon_sym_no] = ACTIONS(3145),
    [anon_sym_for] = ACTIONS(3145),
    [anon_sym_foreach] = ACTIONS(3145),
    [anon_sym_last] = ACTIONS(3145),
    [anon_sym_next] = ACTIONS(3145),
    [anon_sym_redo] = ACTIONS(3145),
    [anon_sym_BEGIN] = ACTIONS(3145),
    [anon_sym_INIT] = ACTIONS(3145),
    [anon_sym_CHECK] = ACTIONS(3145),
    [anon_sym_UNITCHECK] = ACTIONS(3145),
    [anon_sym_END] = ACTIONS(3145),
    [anon_sym___FILE__] = ACTIONS(3145),
    [anon_sym___LINE__] = ACTIONS(3145),
    [anon_sym___PACKAGE__] = ACTIONS(3145),
    [anon_sym___SUB__] = ACTIONS(3145),
    [anon_sym_break] = ACTIONS(3145),
    [anon_sym_fork] = ACTIONS(3145),
    [anon_sym_getppid] = ACTIONS(3145),
    [anon_sym_time] = ACTIONS(3145),
    [anon_sym_times] = ACTIONS(3145),
    [anon_sym_wait] = ACTIONS(3145),
    [anon_sym_wantarray] = ACTIONS(3145),
    [anon_sym_abs] = ACTIONS(3145),
    [anon_sym_alarm] = ACTIONS(3145),
    [anon_sym_chop] = ACTIONS(3145),
    [anon_sym_chdir] = ACTIONS(3145),
    [anon_sym_close] = ACTIONS(3145),
    [anon_sym_closedir] = ACTIONS(3145),
    [anon_sym_caller] = ACTIONS(3145),
    [anon_sym_chomp] = ACTIONS(3145),
    [anon_sym_chr] = ACTIONS(3145),
    [anon_sym_cos] = ACTIONS(3145),
    [anon_sym_chroot] = ACTIONS(3145),
    [anon_sym_defined] = ACTIONS(3145),
    [anon_sym_delete] = ACTIONS(3145),
    [anon_sym_dbmclose] = ACTIONS(3145),
    [anon_sym_exists] = ACTIONS(3145),
    [anon_sym_exit] = ACTIONS(3145),
    [anon_sym_eof] = ACTIONS(3145),
    [anon_sym_exp] = ACTIONS(3145),
    [anon_sym_each] = ACTIONS(3145),
    [anon_sym_fc] = ACTIONS(3145),
    [anon_sym_fileno] = ACTIONS(3145),
    [anon_sym_gmtime] = ACTIONS(3145),
    [anon_sym_getc] = ACTIONS(3145),
    [anon_sym_getpgrp] = ACTIONS(3145),
    [anon_sym_getprotobyname] = ACTIONS(3145),
    [anon_sym_getpwname] = ACTIONS(3145),
    [anon_sym_getpwuid] = ACTIONS(3145),
    [anon_sym_getpeername] = ACTIONS(3145),
    [anon_sym_getnetbyname] = ACTIONS(3145),
    [anon_sym_getsockname] = ACTIONS(3145),
    [anon_sym_getgrnam] = ACTIONS(3145),
    [anon_sym_getgrgid] = ACTIONS(3145),
    [anon_sym_hex] = ACTIONS(3145),
    [anon_sym_int] = ACTIONS(3145),
    [anon_sym_keys] = ACTIONS(3145),
    [anon_sym_lc] = ACTIONS(3145),
    [anon_sym_lcfirst] = ACTIONS(3145),
    [anon_sym_length] = ACTIONS(3145),
    [anon_sym_localtime] = ACTIONS(3145),
    [anon_sym_log] = ACTIONS(3145),
    [anon_sym_lock] = ACTIONS(3145),
    [anon_sym_lstat] = ACTIONS(3145),
    [anon_sym_oct] = ACTIONS(3145),
    [anon_sym_ord] = ACTIONS(3145),
    [anon_sym_prototype] = ACTIONS(3145),
    [anon_sym_pop] = ACTIONS(3145),
    [anon_sym_pos] = ACTIONS(3145),
    [anon_sym_quotemeta] = ACTIONS(3145),
    [anon_sym_reset] = ACTIONS(3145),
    [anon_sym_rand] = ACTIONS(3145),
    [anon_sym_rmdir] = ACTIONS(3145),
    [anon_sym_readdir] = ACTIONS(3145),
    [anon_sym_readline] = ACTIONS(3145),
    [anon_sym_readpipe] = ACTIONS(3145),
    [anon_sym_rewinddir] = ACTIONS(3145),
    [anon_sym_readlink] = ACTIONS(3145),
    [anon_sym_ref] = ACTIONS(3145),
    [anon_sym_scalar] = ACTIONS(3145),
    [anon_sym_shift] = ACTIONS(3145),
    [anon_sym_sin] = ACTIONS(3145),
    [anon_sym_sleep] = ACTIONS(3145),
    [anon_sym_sqrt] = ACTIONS(3145),
    [anon_sym_srand] = ACTIONS(3145),
    [anon_sym_stat] = ACTIONS(3145),
    [anon_sym_study] = ACTIONS(3145),
    [anon_sym_tell] = ACTIONS(3145),
    [anon_sym_telldir] = ACTIONS(3145),
    [anon_sym_tied] = ACTIONS(3145),
    [anon_sym_uc] = ACTIONS(3145),
    [anon_sym_ucfirst] = ACTIONS(3145),
    [anon_sym_untie] = ACTIONS(3145),
    [anon_sym_umask] = ACTIONS(3145),
    [anon_sym_values] = ACTIONS(3145),
    [anon_sym_write] = ACTIONS(3145),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3145),
    [anon_sym_qq] = ACTIONS(3145),
    [anon_sym_qw] = ACTIONS(3145),
    [anon_sym_qx] = ACTIONS(3145),
    [anon_sym_qr] = ACTIONS(3145),
    [anon_sym_m] = ACTIONS(3145),
    [anon_sym_s] = ACTIONS(3145),
    [anon_sym_tr] = ACTIONS(3145),
    [anon_sym_y] = ACTIONS(3145),
    [anon_sym_if] = ACTIONS(3145),
    [anon_sym_unless] = ACTIONS(3145),
    [anon_sym_while] = ACTIONS(3145),
    [anon_sym_until] = ACTIONS(3145),
    [aux_sym__bareword_token1] = ACTIONS(3145),
    [sym_number] = ACTIONS(3145),
    [anon_sym_true] = ACTIONS(3145),
    [anon_sym_false] = ACTIONS(3145),
    [sym__single_quote] = ACTIONS(3143),
    [sym__double_quote] = ACTIONS(3143),
    [sym__backtick_quote] = ACTIONS(3143),
    [sym__search_slash_quote] = ACTIONS(3143),
    [sym__open_readline_bracket] = ACTIONS(3143),
    [sym__open_fileglob_bracket] = ACTIONS(3143),
    [sym__PERLY_HEREDOC] = ACTIONS(3143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3143),
    [sym__filetest] = ACTIONS(3143),
  },
  [STATE(1561)] = {
    [sym_heredoc_content] = STATE(1561),
    [sym__identifier] = ACTIONS(2926),
    [anon_sym_LBRACE] = ACTIONS(2926),
    [anon_sym_RBRACE] = ACTIONS(2926),
    [anon_sym_SEMI] = ACTIONS(2926),
    [anon_sym_package] = ACTIONS(2926),
    [anon_sym_class] = ACTIONS(2926),
    [anon_sym_role] = ACTIONS(2926),
    [anon_sym_BUILD] = ACTIONS(2926),
    [anon_sym_ADJUST] = ACTIONS(2926),
    [anon_sym_DOLLAR] = ACTIONS(2926),
    [anon_sym_AT] = ACTIONS(2926),
    [anon_sym_my] = ACTIONS(2926),
    [anon_sym_extended] = ACTIONS(2926),
    [anon_sym_async] = ACTIONS(2926),
    [anon_sym_sub] = ACTIONS(2926),
    [anon_sym_method] = ACTIONS(2926),
    [anon_sym_LPAREN] = ACTIONS(2926),
    [anon_sym_continue] = ACTIONS(2926),
    [anon_sym_state] = ACTIONS(2926),
    [anon_sym_our] = ACTIONS(2926),
    [anon_sym_try] = ACTIONS(2926),
    [anon_sym_defer] = ACTIONS(2926),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2926),
    [anon_sym_LBRACK] = ACTIONS(2926),
    [anon_sym_LT_LT] = ACTIONS(2926),
    [anon_sym_SLASH_SLASH] = ACTIONS(2926),
    [anon_sym_PLUS] = ACTIONS(2926),
    [anon_sym_DASH] = ACTIONS(2926),
    [anon_sym_TILDE] = ACTIONS(2926),
    [anon_sym_BANG] = ACTIONS(2926),
    [anon_sym_PLUS_PLUS] = ACTIONS(2926),
    [anon_sym_DASH_DASH] = ACTIONS(2926),
    [anon_sym_BSLASH] = ACTIONS(2924),
    [anon_sym_do] = ACTIONS(2926),
    [anon_sym_eval] = ACTIONS(2926),
    [anon_sym_await] = ACTIONS(2926),
    [anon_sym_field] = ACTIONS(2926),
    [anon_sym_local] = ACTIONS(2926),
    [anon_sym_dynamically] = ACTIONS(2926),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2926),
    [anon_sym_require] = ACTIONS(2926),
    [anon_sym_map] = ACTIONS(2926),
    [anon_sym_grep] = ACTIONS(2926),
    [anon_sym_sort] = ACTIONS(2926),
    [anon_sym_goto] = ACTIONS(2926),
    [anon_sym_return] = ACTIONS(2926),
    [anon_sym_undef] = ACTIONS(2926),
    [anon_sym_PERCENT2] = ACTIONS(2926),
    [anon_sym_AMP2] = ACTIONS(2926),
    [anon_sym_STAR2] = ACTIONS(2926),
    [anon_sym_use] = ACTIONS(2926),
    [anon_sym_no] = ACTIONS(2926),
    [anon_sym_for] = ACTIONS(2926),
    [anon_sym_foreach] = ACTIONS(2926),
    [anon_sym_last] = ACTIONS(2926),
    [anon_sym_next] = ACTIONS(2926),
    [anon_sym_redo] = ACTIONS(2926),
    [anon_sym_BEGIN] = ACTIONS(2926),
    [anon_sym_INIT] = ACTIONS(2926),
    [anon_sym_CHECK] = ACTIONS(2926),
    [anon_sym_UNITCHECK] = ACTIONS(2926),
    [anon_sym_END] = ACTIONS(2926),
    [anon_sym___FILE__] = ACTIONS(2926),
    [anon_sym___LINE__] = ACTIONS(2926),
    [anon_sym___PACKAGE__] = ACTIONS(2926),
    [anon_sym___SUB__] = ACTIONS(2926),
    [anon_sym_break] = ACTIONS(2926),
    [anon_sym_fork] = ACTIONS(2926),
    [anon_sym_getppid] = ACTIONS(2926),
    [anon_sym_time] = ACTIONS(2926),
    [anon_sym_times] = ACTIONS(2926),
    [anon_sym_wait] = ACTIONS(2926),
    [anon_sym_wantarray] = ACTIONS(2926),
    [anon_sym_abs] = ACTIONS(2926),
    [anon_sym_alarm] = ACTIONS(2926),
    [anon_sym_chop] = ACTIONS(2926),
    [anon_sym_chdir] = ACTIONS(2926),
    [anon_sym_close] = ACTIONS(2926),
    [anon_sym_closedir] = ACTIONS(2926),
    [anon_sym_caller] = ACTIONS(2926),
    [anon_sym_chomp] = ACTIONS(2926),
    [anon_sym_chr] = ACTIONS(2926),
    [anon_sym_cos] = ACTIONS(2926),
    [anon_sym_chroot] = ACTIONS(2926),
    [anon_sym_defined] = ACTIONS(2926),
    [anon_sym_delete] = ACTIONS(2926),
    [anon_sym_dbmclose] = ACTIONS(2926),
    [anon_sym_exists] = ACTIONS(2926),
    [anon_sym_exit] = ACTIONS(2926),
    [anon_sym_eof] = ACTIONS(2926),
    [anon_sym_exp] = ACTIONS(2926),
    [anon_sym_each] = ACTIONS(2926),
    [anon_sym_fc] = ACTIONS(2926),
    [anon_sym_fileno] = ACTIONS(2926),
    [anon_sym_gmtime] = ACTIONS(2926),
    [anon_sym_getc] = ACTIONS(2926),
    [anon_sym_getpgrp] = ACTIONS(2926),
    [anon_sym_getprotobyname] = ACTIONS(2926),
    [anon_sym_getpwname] = ACTIONS(2926),
    [anon_sym_getpwuid] = ACTIONS(2926),
    [anon_sym_getpeername] = ACTIONS(2926),
    [anon_sym_getnetbyname] = ACTIONS(2926),
    [anon_sym_getsockname] = ACTIONS(2926),
    [anon_sym_getgrnam] = ACTIONS(2926),
    [anon_sym_getgrgid] = ACTIONS(2926),
    [anon_sym_hex] = ACTIONS(2926),
    [anon_sym_int] = ACTIONS(2926),
    [anon_sym_keys] = ACTIONS(2926),
    [anon_sym_lc] = ACTIONS(2926),
    [anon_sym_lcfirst] = ACTIONS(2926),
    [anon_sym_length] = ACTIONS(2926),
    [anon_sym_localtime] = ACTIONS(2926),
    [anon_sym_log] = ACTIONS(2926),
    [anon_sym_lock] = ACTIONS(2926),
    [anon_sym_lstat] = ACTIONS(2926),
    [anon_sym_oct] = ACTIONS(2926),
    [anon_sym_ord] = ACTIONS(2926),
    [anon_sym_prototype] = ACTIONS(2926),
    [anon_sym_pop] = ACTIONS(2926),
    [anon_sym_pos] = ACTIONS(2926),
    [anon_sym_quotemeta] = ACTIONS(2926),
    [anon_sym_reset] = ACTIONS(2926),
    [anon_sym_rand] = ACTIONS(2926),
    [anon_sym_rmdir] = ACTIONS(2926),
    [anon_sym_readdir] = ACTIONS(2926),
    [anon_sym_readline] = ACTIONS(2926),
    [anon_sym_readpipe] = ACTIONS(2926),
    [anon_sym_rewinddir] = ACTIONS(2926),
    [anon_sym_readlink] = ACTIONS(2926),
    [anon_sym_ref] = ACTIONS(2926),
    [anon_sym_scalar] = ACTIONS(2926),
    [anon_sym_shift] = ACTIONS(2926),
    [anon_sym_sin] = ACTIONS(2926),
    [anon_sym_sleep] = ACTIONS(2926),
    [anon_sym_sqrt] = ACTIONS(2926),
    [anon_sym_srand] = ACTIONS(2926),
    [anon_sym_stat] = ACTIONS(2926),
    [anon_sym_study] = ACTIONS(2926),
    [anon_sym_tell] = ACTIONS(2926),
    [anon_sym_telldir] = ACTIONS(2926),
    [anon_sym_tied] = ACTIONS(2926),
    [anon_sym_uc] = ACTIONS(2926),
    [anon_sym_ucfirst] = ACTIONS(2926),
    [anon_sym_untie] = ACTIONS(2926),
    [anon_sym_umask] = ACTIONS(2926),
    [anon_sym_values] = ACTIONS(2926),
    [anon_sym_write] = ACTIONS(2926),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2926),
    [anon_sym_qq] = ACTIONS(2926),
    [anon_sym_qw] = ACTIONS(2926),
    [anon_sym_qx] = ACTIONS(2926),
    [anon_sym_qr] = ACTIONS(2926),
    [anon_sym_m] = ACTIONS(2926),
    [anon_sym_s] = ACTIONS(2926),
    [anon_sym_tr] = ACTIONS(2926),
    [anon_sym_y] = ACTIONS(2926),
    [anon_sym_if] = ACTIONS(2926),
    [anon_sym_unless] = ACTIONS(2926),
    [anon_sym_while] = ACTIONS(2926),
    [anon_sym_until] = ACTIONS(2926),
    [aux_sym__bareword_token1] = ACTIONS(2926),
    [sym_number] = ACTIONS(2926),
    [anon_sym_true] = ACTIONS(2926),
    [anon_sym_false] = ACTIONS(2926),
    [sym__single_quote] = ACTIONS(2924),
    [sym__double_quote] = ACTIONS(2924),
    [sym__backtick_quote] = ACTIONS(2924),
    [sym__search_slash_quote] = ACTIONS(2924),
    [sym__open_readline_bracket] = ACTIONS(2924),
    [sym__open_fileglob_bracket] = ACTIONS(2924),
    [sym__PERLY_HEREDOC] = ACTIONS(2924),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2924),
    [sym__filetest] = ACTIONS(2924),
  },
  [STATE(1562)] = {
    [sym_heredoc_content] = STATE(1562),
    [sym__identifier] = ACTIONS(3061),
    [anon_sym_LBRACE] = ACTIONS(3061),
    [anon_sym_RBRACE] = ACTIONS(3061),
    [anon_sym_SEMI] = ACTIONS(3061),
    [anon_sym_package] = ACTIONS(3061),
    [anon_sym_class] = ACTIONS(3061),
    [anon_sym_role] = ACTIONS(3061),
    [anon_sym_BUILD] = ACTIONS(3061),
    [anon_sym_ADJUST] = ACTIONS(3061),
    [anon_sym_DOLLAR] = ACTIONS(3061),
    [anon_sym_AT] = ACTIONS(3061),
    [anon_sym_my] = ACTIONS(3061),
    [anon_sym_extended] = ACTIONS(3061),
    [anon_sym_async] = ACTIONS(3061),
    [anon_sym_sub] = ACTIONS(3061),
    [anon_sym_method] = ACTIONS(3061),
    [anon_sym_LPAREN] = ACTIONS(3061),
    [anon_sym_continue] = ACTIONS(3061),
    [anon_sym_state] = ACTIONS(3061),
    [anon_sym_our] = ACTIONS(3061),
    [anon_sym_try] = ACTIONS(3061),
    [anon_sym_defer] = ACTIONS(3061),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3061),
    [anon_sym_LBRACK] = ACTIONS(3061),
    [anon_sym_LT_LT] = ACTIONS(3061),
    [anon_sym_SLASH_SLASH] = ACTIONS(3061),
    [anon_sym_PLUS] = ACTIONS(3061),
    [anon_sym_DASH] = ACTIONS(3061),
    [anon_sym_TILDE] = ACTIONS(3061),
    [anon_sym_BANG] = ACTIONS(3061),
    [anon_sym_PLUS_PLUS] = ACTIONS(3061),
    [anon_sym_DASH_DASH] = ACTIONS(3061),
    [anon_sym_BSLASH] = ACTIONS(3059),
    [anon_sym_do] = ACTIONS(3061),
    [anon_sym_eval] = ACTIONS(3061),
    [anon_sym_await] = ACTIONS(3061),
    [anon_sym_field] = ACTIONS(3061),
    [anon_sym_local] = ACTIONS(3061),
    [anon_sym_dynamically] = ACTIONS(3061),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3061),
    [anon_sym_require] = ACTIONS(3061),
    [anon_sym_map] = ACTIONS(3061),
    [anon_sym_grep] = ACTIONS(3061),
    [anon_sym_sort] = ACTIONS(3061),
    [anon_sym_goto] = ACTIONS(3061),
    [anon_sym_return] = ACTIONS(3061),
    [anon_sym_undef] = ACTIONS(3061),
    [anon_sym_PERCENT2] = ACTIONS(3061),
    [anon_sym_AMP2] = ACTIONS(3061),
    [anon_sym_STAR2] = ACTIONS(3061),
    [anon_sym_use] = ACTIONS(3061),
    [anon_sym_no] = ACTIONS(3061),
    [anon_sym_for] = ACTIONS(3061),
    [anon_sym_foreach] = ACTIONS(3061),
    [anon_sym_last] = ACTIONS(3061),
    [anon_sym_next] = ACTIONS(3061),
    [anon_sym_redo] = ACTIONS(3061),
    [anon_sym_BEGIN] = ACTIONS(3061),
    [anon_sym_INIT] = ACTIONS(3061),
    [anon_sym_CHECK] = ACTIONS(3061),
    [anon_sym_UNITCHECK] = ACTIONS(3061),
    [anon_sym_END] = ACTIONS(3061),
    [anon_sym___FILE__] = ACTIONS(3061),
    [anon_sym___LINE__] = ACTIONS(3061),
    [anon_sym___PACKAGE__] = ACTIONS(3061),
    [anon_sym___SUB__] = ACTIONS(3061),
    [anon_sym_break] = ACTIONS(3061),
    [anon_sym_fork] = ACTIONS(3061),
    [anon_sym_getppid] = ACTIONS(3061),
    [anon_sym_time] = ACTIONS(3061),
    [anon_sym_times] = ACTIONS(3061),
    [anon_sym_wait] = ACTIONS(3061),
    [anon_sym_wantarray] = ACTIONS(3061),
    [anon_sym_abs] = ACTIONS(3061),
    [anon_sym_alarm] = ACTIONS(3061),
    [anon_sym_chop] = ACTIONS(3061),
    [anon_sym_chdir] = ACTIONS(3061),
    [anon_sym_close] = ACTIONS(3061),
    [anon_sym_closedir] = ACTIONS(3061),
    [anon_sym_caller] = ACTIONS(3061),
    [anon_sym_chomp] = ACTIONS(3061),
    [anon_sym_chr] = ACTIONS(3061),
    [anon_sym_cos] = ACTIONS(3061),
    [anon_sym_chroot] = ACTIONS(3061),
    [anon_sym_defined] = ACTIONS(3061),
    [anon_sym_delete] = ACTIONS(3061),
    [anon_sym_dbmclose] = ACTIONS(3061),
    [anon_sym_exists] = ACTIONS(3061),
    [anon_sym_exit] = ACTIONS(3061),
    [anon_sym_eof] = ACTIONS(3061),
    [anon_sym_exp] = ACTIONS(3061),
    [anon_sym_each] = ACTIONS(3061),
    [anon_sym_fc] = ACTIONS(3061),
    [anon_sym_fileno] = ACTIONS(3061),
    [anon_sym_gmtime] = ACTIONS(3061),
    [anon_sym_getc] = ACTIONS(3061),
    [anon_sym_getpgrp] = ACTIONS(3061),
    [anon_sym_getprotobyname] = ACTIONS(3061),
    [anon_sym_getpwname] = ACTIONS(3061),
    [anon_sym_getpwuid] = ACTIONS(3061),
    [anon_sym_getpeername] = ACTIONS(3061),
    [anon_sym_getnetbyname] = ACTIONS(3061),
    [anon_sym_getsockname] = ACTIONS(3061),
    [anon_sym_getgrnam] = ACTIONS(3061),
    [anon_sym_getgrgid] = ACTIONS(3061),
    [anon_sym_hex] = ACTIONS(3061),
    [anon_sym_int] = ACTIONS(3061),
    [anon_sym_keys] = ACTIONS(3061),
    [anon_sym_lc] = ACTIONS(3061),
    [anon_sym_lcfirst] = ACTIONS(3061),
    [anon_sym_length] = ACTIONS(3061),
    [anon_sym_localtime] = ACTIONS(3061),
    [anon_sym_log] = ACTIONS(3061),
    [anon_sym_lock] = ACTIONS(3061),
    [anon_sym_lstat] = ACTIONS(3061),
    [anon_sym_oct] = ACTIONS(3061),
    [anon_sym_ord] = ACTIONS(3061),
    [anon_sym_prototype] = ACTIONS(3061),
    [anon_sym_pop] = ACTIONS(3061),
    [anon_sym_pos] = ACTIONS(3061),
    [anon_sym_quotemeta] = ACTIONS(3061),
    [anon_sym_reset] = ACTIONS(3061),
    [anon_sym_rand] = ACTIONS(3061),
    [anon_sym_rmdir] = ACTIONS(3061),
    [anon_sym_readdir] = ACTIONS(3061),
    [anon_sym_readline] = ACTIONS(3061),
    [anon_sym_readpipe] = ACTIONS(3061),
    [anon_sym_rewinddir] = ACTIONS(3061),
    [anon_sym_readlink] = ACTIONS(3061),
    [anon_sym_ref] = ACTIONS(3061),
    [anon_sym_scalar] = ACTIONS(3061),
    [anon_sym_shift] = ACTIONS(3061),
    [anon_sym_sin] = ACTIONS(3061),
    [anon_sym_sleep] = ACTIONS(3061),
    [anon_sym_sqrt] = ACTIONS(3061),
    [anon_sym_srand] = ACTIONS(3061),
    [anon_sym_stat] = ACTIONS(3061),
    [anon_sym_study] = ACTIONS(3061),
    [anon_sym_tell] = ACTIONS(3061),
    [anon_sym_telldir] = ACTIONS(3061),
    [anon_sym_tied] = ACTIONS(3061),
    [anon_sym_uc] = ACTIONS(3061),
    [anon_sym_ucfirst] = ACTIONS(3061),
    [anon_sym_untie] = ACTIONS(3061),
    [anon_sym_umask] = ACTIONS(3061),
    [anon_sym_values] = ACTIONS(3061),
    [anon_sym_write] = ACTIONS(3061),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3061),
    [anon_sym_qq] = ACTIONS(3061),
    [anon_sym_qw] = ACTIONS(3061),
    [anon_sym_qx] = ACTIONS(3061),
    [anon_sym_qr] = ACTIONS(3061),
    [anon_sym_m] = ACTIONS(3061),
    [anon_sym_s] = ACTIONS(3061),
    [anon_sym_tr] = ACTIONS(3061),
    [anon_sym_y] = ACTIONS(3061),
    [anon_sym_if] = ACTIONS(3061),
    [anon_sym_unless] = ACTIONS(3061),
    [anon_sym_while] = ACTIONS(3061),
    [anon_sym_until] = ACTIONS(3061),
    [aux_sym__bareword_token1] = ACTIONS(3061),
    [sym_number] = ACTIONS(3061),
    [anon_sym_true] = ACTIONS(3061),
    [anon_sym_false] = ACTIONS(3061),
    [sym__single_quote] = ACTIONS(3059),
    [sym__double_quote] = ACTIONS(3059),
    [sym__backtick_quote] = ACTIONS(3059),
    [sym__search_slash_quote] = ACTIONS(3059),
    [sym__open_readline_bracket] = ACTIONS(3059),
    [sym__open_fileglob_bracket] = ACTIONS(3059),
    [sym__PERLY_HEREDOC] = ACTIONS(3059),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3059),
    [sym__filetest] = ACTIONS(3059),
  },
  [STATE(1563)] = {
    [sym_heredoc_content] = STATE(1563),
    [sym__identifier] = ACTIONS(3233),
    [anon_sym_LBRACE] = ACTIONS(3233),
    [anon_sym_RBRACE] = ACTIONS(3233),
    [anon_sym_SEMI] = ACTIONS(3233),
    [anon_sym_package] = ACTIONS(3233),
    [anon_sym_class] = ACTIONS(3233),
    [anon_sym_role] = ACTIONS(3233),
    [anon_sym_BUILD] = ACTIONS(3233),
    [anon_sym_ADJUST] = ACTIONS(3233),
    [anon_sym_DOLLAR] = ACTIONS(3233),
    [anon_sym_AT] = ACTIONS(3233),
    [anon_sym_my] = ACTIONS(3233),
    [anon_sym_extended] = ACTIONS(3233),
    [anon_sym_async] = ACTIONS(3233),
    [anon_sym_sub] = ACTIONS(3233),
    [anon_sym_method] = ACTIONS(3233),
    [anon_sym_LPAREN] = ACTIONS(3233),
    [anon_sym_continue] = ACTIONS(3233),
    [anon_sym_state] = ACTIONS(3233),
    [anon_sym_our] = ACTIONS(3233),
    [anon_sym_try] = ACTIONS(3233),
    [anon_sym_defer] = ACTIONS(3233),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3233),
    [anon_sym_LBRACK] = ACTIONS(3233),
    [anon_sym_LT_LT] = ACTIONS(3233),
    [anon_sym_SLASH_SLASH] = ACTIONS(3233),
    [anon_sym_PLUS] = ACTIONS(3233),
    [anon_sym_DASH] = ACTIONS(3233),
    [anon_sym_TILDE] = ACTIONS(3233),
    [anon_sym_BANG] = ACTIONS(3233),
    [anon_sym_PLUS_PLUS] = ACTIONS(3233),
    [anon_sym_DASH_DASH] = ACTIONS(3233),
    [anon_sym_BSLASH] = ACTIONS(3231),
    [anon_sym_do] = ACTIONS(3233),
    [anon_sym_eval] = ACTIONS(3233),
    [anon_sym_await] = ACTIONS(3233),
    [anon_sym_field] = ACTIONS(3233),
    [anon_sym_local] = ACTIONS(3233),
    [anon_sym_dynamically] = ACTIONS(3233),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3233),
    [anon_sym_require] = ACTIONS(3233),
    [anon_sym_map] = ACTIONS(3233),
    [anon_sym_grep] = ACTIONS(3233),
    [anon_sym_sort] = ACTIONS(3233),
    [anon_sym_goto] = ACTIONS(3233),
    [anon_sym_return] = ACTIONS(3233),
    [anon_sym_undef] = ACTIONS(3233),
    [anon_sym_PERCENT2] = ACTIONS(3233),
    [anon_sym_AMP2] = ACTIONS(3233),
    [anon_sym_STAR2] = ACTIONS(3233),
    [anon_sym_use] = ACTIONS(3233),
    [anon_sym_no] = ACTIONS(3233),
    [anon_sym_for] = ACTIONS(3233),
    [anon_sym_foreach] = ACTIONS(3233),
    [anon_sym_last] = ACTIONS(3233),
    [anon_sym_next] = ACTIONS(3233),
    [anon_sym_redo] = ACTIONS(3233),
    [anon_sym_BEGIN] = ACTIONS(3233),
    [anon_sym_INIT] = ACTIONS(3233),
    [anon_sym_CHECK] = ACTIONS(3233),
    [anon_sym_UNITCHECK] = ACTIONS(3233),
    [anon_sym_END] = ACTIONS(3233),
    [anon_sym___FILE__] = ACTIONS(3233),
    [anon_sym___LINE__] = ACTIONS(3233),
    [anon_sym___PACKAGE__] = ACTIONS(3233),
    [anon_sym___SUB__] = ACTIONS(3233),
    [anon_sym_break] = ACTIONS(3233),
    [anon_sym_fork] = ACTIONS(3233),
    [anon_sym_getppid] = ACTIONS(3233),
    [anon_sym_time] = ACTIONS(3233),
    [anon_sym_times] = ACTIONS(3233),
    [anon_sym_wait] = ACTIONS(3233),
    [anon_sym_wantarray] = ACTIONS(3233),
    [anon_sym_abs] = ACTIONS(3233),
    [anon_sym_alarm] = ACTIONS(3233),
    [anon_sym_chop] = ACTIONS(3233),
    [anon_sym_chdir] = ACTIONS(3233),
    [anon_sym_close] = ACTIONS(3233),
    [anon_sym_closedir] = ACTIONS(3233),
    [anon_sym_caller] = ACTIONS(3233),
    [anon_sym_chomp] = ACTIONS(3233),
    [anon_sym_chr] = ACTIONS(3233),
    [anon_sym_cos] = ACTIONS(3233),
    [anon_sym_chroot] = ACTIONS(3233),
    [anon_sym_defined] = ACTIONS(3233),
    [anon_sym_delete] = ACTIONS(3233),
    [anon_sym_dbmclose] = ACTIONS(3233),
    [anon_sym_exists] = ACTIONS(3233),
    [anon_sym_exit] = ACTIONS(3233),
    [anon_sym_eof] = ACTIONS(3233),
    [anon_sym_exp] = ACTIONS(3233),
    [anon_sym_each] = ACTIONS(3233),
    [anon_sym_fc] = ACTIONS(3233),
    [anon_sym_fileno] = ACTIONS(3233),
    [anon_sym_gmtime] = ACTIONS(3233),
    [anon_sym_getc] = ACTIONS(3233),
    [anon_sym_getpgrp] = ACTIONS(3233),
    [anon_sym_getprotobyname] = ACTIONS(3233),
    [anon_sym_getpwname] = ACTIONS(3233),
    [anon_sym_getpwuid] = ACTIONS(3233),
    [anon_sym_getpeername] = ACTIONS(3233),
    [anon_sym_getnetbyname] = ACTIONS(3233),
    [anon_sym_getsockname] = ACTIONS(3233),
    [anon_sym_getgrnam] = ACTIONS(3233),
    [anon_sym_getgrgid] = ACTIONS(3233),
    [anon_sym_hex] = ACTIONS(3233),
    [anon_sym_int] = ACTIONS(3233),
    [anon_sym_keys] = ACTIONS(3233),
    [anon_sym_lc] = ACTIONS(3233),
    [anon_sym_lcfirst] = ACTIONS(3233),
    [anon_sym_length] = ACTIONS(3233),
    [anon_sym_localtime] = ACTIONS(3233),
    [anon_sym_log] = ACTIONS(3233),
    [anon_sym_lock] = ACTIONS(3233),
    [anon_sym_lstat] = ACTIONS(3233),
    [anon_sym_oct] = ACTIONS(3233),
    [anon_sym_ord] = ACTIONS(3233),
    [anon_sym_prototype] = ACTIONS(3233),
    [anon_sym_pop] = ACTIONS(3233),
    [anon_sym_pos] = ACTIONS(3233),
    [anon_sym_quotemeta] = ACTIONS(3233),
    [anon_sym_reset] = ACTIONS(3233),
    [anon_sym_rand] = ACTIONS(3233),
    [anon_sym_rmdir] = ACTIONS(3233),
    [anon_sym_readdir] = ACTIONS(3233),
    [anon_sym_readline] = ACTIONS(3233),
    [anon_sym_readpipe] = ACTIONS(3233),
    [anon_sym_rewinddir] = ACTIONS(3233),
    [anon_sym_readlink] = ACTIONS(3233),
    [anon_sym_ref] = ACTIONS(3233),
    [anon_sym_scalar] = ACTIONS(3233),
    [anon_sym_shift] = ACTIONS(3233),
    [anon_sym_sin] = ACTIONS(3233),
    [anon_sym_sleep] = ACTIONS(3233),
    [anon_sym_sqrt] = ACTIONS(3233),
    [anon_sym_srand] = ACTIONS(3233),
    [anon_sym_stat] = ACTIONS(3233),
    [anon_sym_study] = ACTIONS(3233),
    [anon_sym_tell] = ACTIONS(3233),
    [anon_sym_telldir] = ACTIONS(3233),
    [anon_sym_tied] = ACTIONS(3233),
    [anon_sym_uc] = ACTIONS(3233),
    [anon_sym_ucfirst] = ACTIONS(3233),
    [anon_sym_untie] = ACTIONS(3233),
    [anon_sym_umask] = ACTIONS(3233),
    [anon_sym_values] = ACTIONS(3233),
    [anon_sym_write] = ACTIONS(3233),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3233),
    [anon_sym_qq] = ACTIONS(3233),
    [anon_sym_qw] = ACTIONS(3233),
    [anon_sym_qx] = ACTIONS(3233),
    [anon_sym_qr] = ACTIONS(3233),
    [anon_sym_m] = ACTIONS(3233),
    [anon_sym_s] = ACTIONS(3233),
    [anon_sym_tr] = ACTIONS(3233),
    [anon_sym_y] = ACTIONS(3233),
    [anon_sym_if] = ACTIONS(3233),
    [anon_sym_unless] = ACTIONS(3233),
    [anon_sym_while] = ACTIONS(3233),
    [anon_sym_until] = ACTIONS(3233),
    [aux_sym__bareword_token1] = ACTIONS(3233),
    [sym_number] = ACTIONS(3233),
    [anon_sym_true] = ACTIONS(3233),
    [anon_sym_false] = ACTIONS(3233),
    [sym__single_quote] = ACTIONS(3231),
    [sym__double_quote] = ACTIONS(3231),
    [sym__backtick_quote] = ACTIONS(3231),
    [sym__search_slash_quote] = ACTIONS(3231),
    [sym__open_readline_bracket] = ACTIONS(3231),
    [sym__open_fileglob_bracket] = ACTIONS(3231),
    [sym__PERLY_HEREDOC] = ACTIONS(3231),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3231),
    [sym__filetest] = ACTIONS(3231),
  },
  [STATE(1564)] = {
    [sym_heredoc_content] = STATE(1564),
    [sym__identifier] = ACTIONS(2922),
    [anon_sym_LBRACE] = ACTIONS(2922),
    [anon_sym_RBRACE] = ACTIONS(2922),
    [anon_sym_SEMI] = ACTIONS(2922),
    [anon_sym_package] = ACTIONS(2922),
    [anon_sym_class] = ACTIONS(2922),
    [anon_sym_role] = ACTIONS(2922),
    [anon_sym_BUILD] = ACTIONS(2922),
    [anon_sym_ADJUST] = ACTIONS(2922),
    [anon_sym_DOLLAR] = ACTIONS(2922),
    [anon_sym_AT] = ACTIONS(2922),
    [anon_sym_my] = ACTIONS(2922),
    [anon_sym_extended] = ACTIONS(2922),
    [anon_sym_async] = ACTIONS(2922),
    [anon_sym_sub] = ACTIONS(2922),
    [anon_sym_method] = ACTIONS(2922),
    [anon_sym_LPAREN] = ACTIONS(2922),
    [anon_sym_continue] = ACTIONS(2922),
    [anon_sym_state] = ACTIONS(2922),
    [anon_sym_our] = ACTIONS(2922),
    [anon_sym_try] = ACTIONS(2922),
    [anon_sym_defer] = ACTIONS(2922),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2922),
    [anon_sym_LBRACK] = ACTIONS(2922),
    [anon_sym_LT_LT] = ACTIONS(2922),
    [anon_sym_SLASH_SLASH] = ACTIONS(2922),
    [anon_sym_PLUS] = ACTIONS(2922),
    [anon_sym_DASH] = ACTIONS(2922),
    [anon_sym_TILDE] = ACTIONS(2922),
    [anon_sym_BANG] = ACTIONS(2922),
    [anon_sym_PLUS_PLUS] = ACTIONS(2922),
    [anon_sym_DASH_DASH] = ACTIONS(2922),
    [anon_sym_BSLASH] = ACTIONS(2920),
    [anon_sym_do] = ACTIONS(2922),
    [anon_sym_eval] = ACTIONS(2922),
    [anon_sym_await] = ACTIONS(2922),
    [anon_sym_field] = ACTIONS(2922),
    [anon_sym_local] = ACTIONS(2922),
    [anon_sym_dynamically] = ACTIONS(2922),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2922),
    [anon_sym_require] = ACTIONS(2922),
    [anon_sym_map] = ACTIONS(2922),
    [anon_sym_grep] = ACTIONS(2922),
    [anon_sym_sort] = ACTIONS(2922),
    [anon_sym_goto] = ACTIONS(2922),
    [anon_sym_return] = ACTIONS(2922),
    [anon_sym_undef] = ACTIONS(2922),
    [anon_sym_PERCENT2] = ACTIONS(2922),
    [anon_sym_AMP2] = ACTIONS(2922),
    [anon_sym_STAR2] = ACTIONS(2922),
    [anon_sym_use] = ACTIONS(2922),
    [anon_sym_no] = ACTIONS(2922),
    [anon_sym_for] = ACTIONS(2922),
    [anon_sym_foreach] = ACTIONS(2922),
    [anon_sym_last] = ACTIONS(2922),
    [anon_sym_next] = ACTIONS(2922),
    [anon_sym_redo] = ACTIONS(2922),
    [anon_sym_BEGIN] = ACTIONS(2922),
    [anon_sym_INIT] = ACTIONS(2922),
    [anon_sym_CHECK] = ACTIONS(2922),
    [anon_sym_UNITCHECK] = ACTIONS(2922),
    [anon_sym_END] = ACTIONS(2922),
    [anon_sym___FILE__] = ACTIONS(2922),
    [anon_sym___LINE__] = ACTIONS(2922),
    [anon_sym___PACKAGE__] = ACTIONS(2922),
    [anon_sym___SUB__] = ACTIONS(2922),
    [anon_sym_break] = ACTIONS(2922),
    [anon_sym_fork] = ACTIONS(2922),
    [anon_sym_getppid] = ACTIONS(2922),
    [anon_sym_time] = ACTIONS(2922),
    [anon_sym_times] = ACTIONS(2922),
    [anon_sym_wait] = ACTIONS(2922),
    [anon_sym_wantarray] = ACTIONS(2922),
    [anon_sym_abs] = ACTIONS(2922),
    [anon_sym_alarm] = ACTIONS(2922),
    [anon_sym_chop] = ACTIONS(2922),
    [anon_sym_chdir] = ACTIONS(2922),
    [anon_sym_close] = ACTIONS(2922),
    [anon_sym_closedir] = ACTIONS(2922),
    [anon_sym_caller] = ACTIONS(2922),
    [anon_sym_chomp] = ACTIONS(2922),
    [anon_sym_chr] = ACTIONS(2922),
    [anon_sym_cos] = ACTIONS(2922),
    [anon_sym_chroot] = ACTIONS(2922),
    [anon_sym_defined] = ACTIONS(2922),
    [anon_sym_delete] = ACTIONS(2922),
    [anon_sym_dbmclose] = ACTIONS(2922),
    [anon_sym_exists] = ACTIONS(2922),
    [anon_sym_exit] = ACTIONS(2922),
    [anon_sym_eof] = ACTIONS(2922),
    [anon_sym_exp] = ACTIONS(2922),
    [anon_sym_each] = ACTIONS(2922),
    [anon_sym_fc] = ACTIONS(2922),
    [anon_sym_fileno] = ACTIONS(2922),
    [anon_sym_gmtime] = ACTIONS(2922),
    [anon_sym_getc] = ACTIONS(2922),
    [anon_sym_getpgrp] = ACTIONS(2922),
    [anon_sym_getprotobyname] = ACTIONS(2922),
    [anon_sym_getpwname] = ACTIONS(2922),
    [anon_sym_getpwuid] = ACTIONS(2922),
    [anon_sym_getpeername] = ACTIONS(2922),
    [anon_sym_getnetbyname] = ACTIONS(2922),
    [anon_sym_getsockname] = ACTIONS(2922),
    [anon_sym_getgrnam] = ACTIONS(2922),
    [anon_sym_getgrgid] = ACTIONS(2922),
    [anon_sym_hex] = ACTIONS(2922),
    [anon_sym_int] = ACTIONS(2922),
    [anon_sym_keys] = ACTIONS(2922),
    [anon_sym_lc] = ACTIONS(2922),
    [anon_sym_lcfirst] = ACTIONS(2922),
    [anon_sym_length] = ACTIONS(2922),
    [anon_sym_localtime] = ACTIONS(2922),
    [anon_sym_log] = ACTIONS(2922),
    [anon_sym_lock] = ACTIONS(2922),
    [anon_sym_lstat] = ACTIONS(2922),
    [anon_sym_oct] = ACTIONS(2922),
    [anon_sym_ord] = ACTIONS(2922),
    [anon_sym_prototype] = ACTIONS(2922),
    [anon_sym_pop] = ACTIONS(2922),
    [anon_sym_pos] = ACTIONS(2922),
    [anon_sym_quotemeta] = ACTIONS(2922),
    [anon_sym_reset] = ACTIONS(2922),
    [anon_sym_rand] = ACTIONS(2922),
    [anon_sym_rmdir] = ACTIONS(2922),
    [anon_sym_readdir] = ACTIONS(2922),
    [anon_sym_readline] = ACTIONS(2922),
    [anon_sym_readpipe] = ACTIONS(2922),
    [anon_sym_rewinddir] = ACTIONS(2922),
    [anon_sym_readlink] = ACTIONS(2922),
    [anon_sym_ref] = ACTIONS(2922),
    [anon_sym_scalar] = ACTIONS(2922),
    [anon_sym_shift] = ACTIONS(2922),
    [anon_sym_sin] = ACTIONS(2922),
    [anon_sym_sleep] = ACTIONS(2922),
    [anon_sym_sqrt] = ACTIONS(2922),
    [anon_sym_srand] = ACTIONS(2922),
    [anon_sym_stat] = ACTIONS(2922),
    [anon_sym_study] = ACTIONS(2922),
    [anon_sym_tell] = ACTIONS(2922),
    [anon_sym_telldir] = ACTIONS(2922),
    [anon_sym_tied] = ACTIONS(2922),
    [anon_sym_uc] = ACTIONS(2922),
    [anon_sym_ucfirst] = ACTIONS(2922),
    [anon_sym_untie] = ACTIONS(2922),
    [anon_sym_umask] = ACTIONS(2922),
    [anon_sym_values] = ACTIONS(2922),
    [anon_sym_write] = ACTIONS(2922),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2922),
    [anon_sym_qq] = ACTIONS(2922),
    [anon_sym_qw] = ACTIONS(2922),
    [anon_sym_qx] = ACTIONS(2922),
    [anon_sym_qr] = ACTIONS(2922),
    [anon_sym_m] = ACTIONS(2922),
    [anon_sym_s] = ACTIONS(2922),
    [anon_sym_tr] = ACTIONS(2922),
    [anon_sym_y] = ACTIONS(2922),
    [anon_sym_if] = ACTIONS(2922),
    [anon_sym_unless] = ACTIONS(2922),
    [anon_sym_while] = ACTIONS(2922),
    [anon_sym_until] = ACTIONS(2922),
    [aux_sym__bareword_token1] = ACTIONS(2922),
    [sym_number] = ACTIONS(2922),
    [anon_sym_true] = ACTIONS(2922),
    [anon_sym_false] = ACTIONS(2922),
    [sym__single_quote] = ACTIONS(2920),
    [sym__double_quote] = ACTIONS(2920),
    [sym__backtick_quote] = ACTIONS(2920),
    [sym__search_slash_quote] = ACTIONS(2920),
    [sym__open_readline_bracket] = ACTIONS(2920),
    [sym__open_fileglob_bracket] = ACTIONS(2920),
    [sym__PERLY_HEREDOC] = ACTIONS(2920),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2920),
    [sym__filetest] = ACTIONS(2920),
  },
  [STATE(1565)] = {
    [sym_heredoc_content] = STATE(1565),
    [sym__identifier] = ACTIONS(3085),
    [anon_sym_LBRACE] = ACTIONS(3085),
    [anon_sym_RBRACE] = ACTIONS(3085),
    [anon_sym_SEMI] = ACTIONS(3085),
    [anon_sym_package] = ACTIONS(3085),
    [anon_sym_class] = ACTIONS(3085),
    [anon_sym_role] = ACTIONS(3085),
    [anon_sym_BUILD] = ACTIONS(3085),
    [anon_sym_ADJUST] = ACTIONS(3085),
    [anon_sym_DOLLAR] = ACTIONS(3085),
    [anon_sym_AT] = ACTIONS(3085),
    [anon_sym_my] = ACTIONS(3085),
    [anon_sym_extended] = ACTIONS(3085),
    [anon_sym_async] = ACTIONS(3085),
    [anon_sym_sub] = ACTIONS(3085),
    [anon_sym_method] = ACTIONS(3085),
    [anon_sym_LPAREN] = ACTIONS(3085),
    [anon_sym_continue] = ACTIONS(3085),
    [anon_sym_state] = ACTIONS(3085),
    [anon_sym_our] = ACTIONS(3085),
    [anon_sym_try] = ACTIONS(3085),
    [anon_sym_defer] = ACTIONS(3085),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3085),
    [anon_sym_LBRACK] = ACTIONS(3085),
    [anon_sym_LT_LT] = ACTIONS(3085),
    [anon_sym_SLASH_SLASH] = ACTIONS(3085),
    [anon_sym_PLUS] = ACTIONS(3085),
    [anon_sym_DASH] = ACTIONS(3085),
    [anon_sym_TILDE] = ACTIONS(3085),
    [anon_sym_BANG] = ACTIONS(3085),
    [anon_sym_PLUS_PLUS] = ACTIONS(3085),
    [anon_sym_DASH_DASH] = ACTIONS(3085),
    [anon_sym_BSLASH] = ACTIONS(3083),
    [anon_sym_do] = ACTIONS(3085),
    [anon_sym_eval] = ACTIONS(3085),
    [anon_sym_await] = ACTIONS(3085),
    [anon_sym_field] = ACTIONS(3085),
    [anon_sym_local] = ACTIONS(3085),
    [anon_sym_dynamically] = ACTIONS(3085),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3085),
    [anon_sym_require] = ACTIONS(3085),
    [anon_sym_map] = ACTIONS(3085),
    [anon_sym_grep] = ACTIONS(3085),
    [anon_sym_sort] = ACTIONS(3085),
    [anon_sym_goto] = ACTIONS(3085),
    [anon_sym_return] = ACTIONS(3085),
    [anon_sym_undef] = ACTIONS(3085),
    [anon_sym_PERCENT2] = ACTIONS(3085),
    [anon_sym_AMP2] = ACTIONS(3085),
    [anon_sym_STAR2] = ACTIONS(3085),
    [anon_sym_use] = ACTIONS(3085),
    [anon_sym_no] = ACTIONS(3085),
    [anon_sym_for] = ACTIONS(3085),
    [anon_sym_foreach] = ACTIONS(3085),
    [anon_sym_last] = ACTIONS(3085),
    [anon_sym_next] = ACTIONS(3085),
    [anon_sym_redo] = ACTIONS(3085),
    [anon_sym_BEGIN] = ACTIONS(3085),
    [anon_sym_INIT] = ACTIONS(3085),
    [anon_sym_CHECK] = ACTIONS(3085),
    [anon_sym_UNITCHECK] = ACTIONS(3085),
    [anon_sym_END] = ACTIONS(3085),
    [anon_sym___FILE__] = ACTIONS(3085),
    [anon_sym___LINE__] = ACTIONS(3085),
    [anon_sym___PACKAGE__] = ACTIONS(3085),
    [anon_sym___SUB__] = ACTIONS(3085),
    [anon_sym_break] = ACTIONS(3085),
    [anon_sym_fork] = ACTIONS(3085),
    [anon_sym_getppid] = ACTIONS(3085),
    [anon_sym_time] = ACTIONS(3085),
    [anon_sym_times] = ACTIONS(3085),
    [anon_sym_wait] = ACTIONS(3085),
    [anon_sym_wantarray] = ACTIONS(3085),
    [anon_sym_abs] = ACTIONS(3085),
    [anon_sym_alarm] = ACTIONS(3085),
    [anon_sym_chop] = ACTIONS(3085),
    [anon_sym_chdir] = ACTIONS(3085),
    [anon_sym_close] = ACTIONS(3085),
    [anon_sym_closedir] = ACTIONS(3085),
    [anon_sym_caller] = ACTIONS(3085),
    [anon_sym_chomp] = ACTIONS(3085),
    [anon_sym_chr] = ACTIONS(3085),
    [anon_sym_cos] = ACTIONS(3085),
    [anon_sym_chroot] = ACTIONS(3085),
    [anon_sym_defined] = ACTIONS(3085),
    [anon_sym_delete] = ACTIONS(3085),
    [anon_sym_dbmclose] = ACTIONS(3085),
    [anon_sym_exists] = ACTIONS(3085),
    [anon_sym_exit] = ACTIONS(3085),
    [anon_sym_eof] = ACTIONS(3085),
    [anon_sym_exp] = ACTIONS(3085),
    [anon_sym_each] = ACTIONS(3085),
    [anon_sym_fc] = ACTIONS(3085),
    [anon_sym_fileno] = ACTIONS(3085),
    [anon_sym_gmtime] = ACTIONS(3085),
    [anon_sym_getc] = ACTIONS(3085),
    [anon_sym_getpgrp] = ACTIONS(3085),
    [anon_sym_getprotobyname] = ACTIONS(3085),
    [anon_sym_getpwname] = ACTIONS(3085),
    [anon_sym_getpwuid] = ACTIONS(3085),
    [anon_sym_getpeername] = ACTIONS(3085),
    [anon_sym_getnetbyname] = ACTIONS(3085),
    [anon_sym_getsockname] = ACTIONS(3085),
    [anon_sym_getgrnam] = ACTIONS(3085),
    [anon_sym_getgrgid] = ACTIONS(3085),
    [anon_sym_hex] = ACTIONS(3085),
    [anon_sym_int] = ACTIONS(3085),
    [anon_sym_keys] = ACTIONS(3085),
    [anon_sym_lc] = ACTIONS(3085),
    [anon_sym_lcfirst] = ACTIONS(3085),
    [anon_sym_length] = ACTIONS(3085),
    [anon_sym_localtime] = ACTIONS(3085),
    [anon_sym_log] = ACTIONS(3085),
    [anon_sym_lock] = ACTIONS(3085),
    [anon_sym_lstat] = ACTIONS(3085),
    [anon_sym_oct] = ACTIONS(3085),
    [anon_sym_ord] = ACTIONS(3085),
    [anon_sym_prototype] = ACTIONS(3085),
    [anon_sym_pop] = ACTIONS(3085),
    [anon_sym_pos] = ACTIONS(3085),
    [anon_sym_quotemeta] = ACTIONS(3085),
    [anon_sym_reset] = ACTIONS(3085),
    [anon_sym_rand] = ACTIONS(3085),
    [anon_sym_rmdir] = ACTIONS(3085),
    [anon_sym_readdir] = ACTIONS(3085),
    [anon_sym_readline] = ACTIONS(3085),
    [anon_sym_readpipe] = ACTIONS(3085),
    [anon_sym_rewinddir] = ACTIONS(3085),
    [anon_sym_readlink] = ACTIONS(3085),
    [anon_sym_ref] = ACTIONS(3085),
    [anon_sym_scalar] = ACTIONS(3085),
    [anon_sym_shift] = ACTIONS(3085),
    [anon_sym_sin] = ACTIONS(3085),
    [anon_sym_sleep] = ACTIONS(3085),
    [anon_sym_sqrt] = ACTIONS(3085),
    [anon_sym_srand] = ACTIONS(3085),
    [anon_sym_stat] = ACTIONS(3085),
    [anon_sym_study] = ACTIONS(3085),
    [anon_sym_tell] = ACTIONS(3085),
    [anon_sym_telldir] = ACTIONS(3085),
    [anon_sym_tied] = ACTIONS(3085),
    [anon_sym_uc] = ACTIONS(3085),
    [anon_sym_ucfirst] = ACTIONS(3085),
    [anon_sym_untie] = ACTIONS(3085),
    [anon_sym_umask] = ACTIONS(3085),
    [anon_sym_values] = ACTIONS(3085),
    [anon_sym_write] = ACTIONS(3085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3085),
    [anon_sym_qq] = ACTIONS(3085),
    [anon_sym_qw] = ACTIONS(3085),
    [anon_sym_qx] = ACTIONS(3085),
    [anon_sym_qr] = ACTIONS(3085),
    [anon_sym_m] = ACTIONS(3085),
    [anon_sym_s] = ACTIONS(3085),
    [anon_sym_tr] = ACTIONS(3085),
    [anon_sym_y] = ACTIONS(3085),
    [anon_sym_if] = ACTIONS(3085),
    [anon_sym_unless] = ACTIONS(3085),
    [anon_sym_while] = ACTIONS(3085),
    [anon_sym_until] = ACTIONS(3085),
    [aux_sym__bareword_token1] = ACTIONS(3085),
    [sym_number] = ACTIONS(3085),
    [anon_sym_true] = ACTIONS(3085),
    [anon_sym_false] = ACTIONS(3085),
    [sym__single_quote] = ACTIONS(3083),
    [sym__double_quote] = ACTIONS(3083),
    [sym__backtick_quote] = ACTIONS(3083),
    [sym__search_slash_quote] = ACTIONS(3083),
    [sym__open_readline_bracket] = ACTIONS(3083),
    [sym__open_fileglob_bracket] = ACTIONS(3083),
    [sym__PERLY_HEREDOC] = ACTIONS(3083),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3083),
    [sym__filetest] = ACTIONS(3083),
  },
  [STATE(1566)] = {
    [sym_heredoc_content] = STATE(1566),
    [sym__identifier] = ACTIONS(2906),
    [anon_sym_LBRACE] = ACTIONS(2906),
    [anon_sym_RBRACE] = ACTIONS(2906),
    [anon_sym_SEMI] = ACTIONS(2906),
    [anon_sym_package] = ACTIONS(2906),
    [anon_sym_class] = ACTIONS(2906),
    [anon_sym_role] = ACTIONS(2906),
    [anon_sym_BUILD] = ACTIONS(2906),
    [anon_sym_ADJUST] = ACTIONS(2906),
    [anon_sym_DOLLAR] = ACTIONS(2906),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_my] = ACTIONS(2906),
    [anon_sym_extended] = ACTIONS(2906),
    [anon_sym_async] = ACTIONS(2906),
    [anon_sym_sub] = ACTIONS(2906),
    [anon_sym_method] = ACTIONS(2906),
    [anon_sym_LPAREN] = ACTIONS(2906),
    [anon_sym_continue] = ACTIONS(2906),
    [anon_sym_state] = ACTIONS(2906),
    [anon_sym_our] = ACTIONS(2906),
    [anon_sym_try] = ACTIONS(2906),
    [anon_sym_defer] = ACTIONS(2906),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2906),
    [anon_sym_LBRACK] = ACTIONS(2906),
    [anon_sym_LT_LT] = ACTIONS(2906),
    [anon_sym_SLASH_SLASH] = ACTIONS(2906),
    [anon_sym_PLUS] = ACTIONS(2906),
    [anon_sym_DASH] = ACTIONS(2906),
    [anon_sym_TILDE] = ACTIONS(2906),
    [anon_sym_BANG] = ACTIONS(2906),
    [anon_sym_PLUS_PLUS] = ACTIONS(2906),
    [anon_sym_DASH_DASH] = ACTIONS(2906),
    [anon_sym_BSLASH] = ACTIONS(2904),
    [anon_sym_do] = ACTIONS(2906),
    [anon_sym_eval] = ACTIONS(2906),
    [anon_sym_await] = ACTIONS(2906),
    [anon_sym_field] = ACTIONS(2906),
    [anon_sym_local] = ACTIONS(2906),
    [anon_sym_dynamically] = ACTIONS(2906),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2906),
    [anon_sym_require] = ACTIONS(2906),
    [anon_sym_map] = ACTIONS(2906),
    [anon_sym_grep] = ACTIONS(2906),
    [anon_sym_sort] = ACTIONS(2906),
    [anon_sym_goto] = ACTIONS(2906),
    [anon_sym_return] = ACTIONS(2906),
    [anon_sym_undef] = ACTIONS(2906),
    [anon_sym_PERCENT2] = ACTIONS(2906),
    [anon_sym_AMP2] = ACTIONS(2906),
    [anon_sym_STAR2] = ACTIONS(2906),
    [anon_sym_use] = ACTIONS(2906),
    [anon_sym_no] = ACTIONS(2906),
    [anon_sym_for] = ACTIONS(2906),
    [anon_sym_foreach] = ACTIONS(2906),
    [anon_sym_last] = ACTIONS(2906),
    [anon_sym_next] = ACTIONS(2906),
    [anon_sym_redo] = ACTIONS(2906),
    [anon_sym_BEGIN] = ACTIONS(2906),
    [anon_sym_INIT] = ACTIONS(2906),
    [anon_sym_CHECK] = ACTIONS(2906),
    [anon_sym_UNITCHECK] = ACTIONS(2906),
    [anon_sym_END] = ACTIONS(2906),
    [anon_sym___FILE__] = ACTIONS(2906),
    [anon_sym___LINE__] = ACTIONS(2906),
    [anon_sym___PACKAGE__] = ACTIONS(2906),
    [anon_sym___SUB__] = ACTIONS(2906),
    [anon_sym_break] = ACTIONS(2906),
    [anon_sym_fork] = ACTIONS(2906),
    [anon_sym_getppid] = ACTIONS(2906),
    [anon_sym_time] = ACTIONS(2906),
    [anon_sym_times] = ACTIONS(2906),
    [anon_sym_wait] = ACTIONS(2906),
    [anon_sym_wantarray] = ACTIONS(2906),
    [anon_sym_abs] = ACTIONS(2906),
    [anon_sym_alarm] = ACTIONS(2906),
    [anon_sym_chop] = ACTIONS(2906),
    [anon_sym_chdir] = ACTIONS(2906),
    [anon_sym_close] = ACTIONS(2906),
    [anon_sym_closedir] = ACTIONS(2906),
    [anon_sym_caller] = ACTIONS(2906),
    [anon_sym_chomp] = ACTIONS(2906),
    [anon_sym_chr] = ACTIONS(2906),
    [anon_sym_cos] = ACTIONS(2906),
    [anon_sym_chroot] = ACTIONS(2906),
    [anon_sym_defined] = ACTIONS(2906),
    [anon_sym_delete] = ACTIONS(2906),
    [anon_sym_dbmclose] = ACTIONS(2906),
    [anon_sym_exists] = ACTIONS(2906),
    [anon_sym_exit] = ACTIONS(2906),
    [anon_sym_eof] = ACTIONS(2906),
    [anon_sym_exp] = ACTIONS(2906),
    [anon_sym_each] = ACTIONS(2906),
    [anon_sym_fc] = ACTIONS(2906),
    [anon_sym_fileno] = ACTIONS(2906),
    [anon_sym_gmtime] = ACTIONS(2906),
    [anon_sym_getc] = ACTIONS(2906),
    [anon_sym_getpgrp] = ACTIONS(2906),
    [anon_sym_getprotobyname] = ACTIONS(2906),
    [anon_sym_getpwname] = ACTIONS(2906),
    [anon_sym_getpwuid] = ACTIONS(2906),
    [anon_sym_getpeername] = ACTIONS(2906),
    [anon_sym_getnetbyname] = ACTIONS(2906),
    [anon_sym_getsockname] = ACTIONS(2906),
    [anon_sym_getgrnam] = ACTIONS(2906),
    [anon_sym_getgrgid] = ACTIONS(2906),
    [anon_sym_hex] = ACTIONS(2906),
    [anon_sym_int] = ACTIONS(2906),
    [anon_sym_keys] = ACTIONS(2906),
    [anon_sym_lc] = ACTIONS(2906),
    [anon_sym_lcfirst] = ACTIONS(2906),
    [anon_sym_length] = ACTIONS(2906),
    [anon_sym_localtime] = ACTIONS(2906),
    [anon_sym_log] = ACTIONS(2906),
    [anon_sym_lock] = ACTIONS(2906),
    [anon_sym_lstat] = ACTIONS(2906),
    [anon_sym_oct] = ACTIONS(2906),
    [anon_sym_ord] = ACTIONS(2906),
    [anon_sym_prototype] = ACTIONS(2906),
    [anon_sym_pop] = ACTIONS(2906),
    [anon_sym_pos] = ACTIONS(2906),
    [anon_sym_quotemeta] = ACTIONS(2906),
    [anon_sym_reset] = ACTIONS(2906),
    [anon_sym_rand] = ACTIONS(2906),
    [anon_sym_rmdir] = ACTIONS(2906),
    [anon_sym_readdir] = ACTIONS(2906),
    [anon_sym_readline] = ACTIONS(2906),
    [anon_sym_readpipe] = ACTIONS(2906),
    [anon_sym_rewinddir] = ACTIONS(2906),
    [anon_sym_readlink] = ACTIONS(2906),
    [anon_sym_ref] = ACTIONS(2906),
    [anon_sym_scalar] = ACTIONS(2906),
    [anon_sym_shift] = ACTIONS(2906),
    [anon_sym_sin] = ACTIONS(2906),
    [anon_sym_sleep] = ACTIONS(2906),
    [anon_sym_sqrt] = ACTIONS(2906),
    [anon_sym_srand] = ACTIONS(2906),
    [anon_sym_stat] = ACTIONS(2906),
    [anon_sym_study] = ACTIONS(2906),
    [anon_sym_tell] = ACTIONS(2906),
    [anon_sym_telldir] = ACTIONS(2906),
    [anon_sym_tied] = ACTIONS(2906),
    [anon_sym_uc] = ACTIONS(2906),
    [anon_sym_ucfirst] = ACTIONS(2906),
    [anon_sym_untie] = ACTIONS(2906),
    [anon_sym_umask] = ACTIONS(2906),
    [anon_sym_values] = ACTIONS(2906),
    [anon_sym_write] = ACTIONS(2906),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2906),
    [anon_sym_qq] = ACTIONS(2906),
    [anon_sym_qw] = ACTIONS(2906),
    [anon_sym_qx] = ACTIONS(2906),
    [anon_sym_qr] = ACTIONS(2906),
    [anon_sym_m] = ACTIONS(2906),
    [anon_sym_s] = ACTIONS(2906),
    [anon_sym_tr] = ACTIONS(2906),
    [anon_sym_y] = ACTIONS(2906),
    [anon_sym_if] = ACTIONS(2906),
    [anon_sym_unless] = ACTIONS(2906),
    [anon_sym_while] = ACTIONS(2906),
    [anon_sym_until] = ACTIONS(2906),
    [aux_sym__bareword_token1] = ACTIONS(2906),
    [sym_number] = ACTIONS(2906),
    [anon_sym_true] = ACTIONS(2906),
    [anon_sym_false] = ACTIONS(2906),
    [sym__single_quote] = ACTIONS(2904),
    [sym__double_quote] = ACTIONS(2904),
    [sym__backtick_quote] = ACTIONS(2904),
    [sym__search_slash_quote] = ACTIONS(2904),
    [sym__open_readline_bracket] = ACTIONS(2904),
    [sym__open_fileglob_bracket] = ACTIONS(2904),
    [sym__PERLY_HEREDOC] = ACTIONS(2904),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2904),
    [sym__filetest] = ACTIONS(2904),
  },
  [STATE(1567)] = {
    [sym_heredoc_content] = STATE(1567),
    [sym__identifier] = ACTIONS(3165),
    [anon_sym_LBRACE] = ACTIONS(3165),
    [anon_sym_RBRACE] = ACTIONS(3165),
    [anon_sym_SEMI] = ACTIONS(3165),
    [anon_sym_package] = ACTIONS(3165),
    [anon_sym_class] = ACTIONS(3165),
    [anon_sym_role] = ACTIONS(3165),
    [anon_sym_BUILD] = ACTIONS(3165),
    [anon_sym_ADJUST] = ACTIONS(3165),
    [anon_sym_DOLLAR] = ACTIONS(3165),
    [anon_sym_AT] = ACTIONS(3165),
    [anon_sym_my] = ACTIONS(3165),
    [anon_sym_extended] = ACTIONS(3165),
    [anon_sym_async] = ACTIONS(3165),
    [anon_sym_sub] = ACTIONS(3165),
    [anon_sym_method] = ACTIONS(3165),
    [anon_sym_LPAREN] = ACTIONS(3165),
    [anon_sym_continue] = ACTIONS(3165),
    [anon_sym_state] = ACTIONS(3165),
    [anon_sym_our] = ACTIONS(3165),
    [anon_sym_try] = ACTIONS(3165),
    [anon_sym_defer] = ACTIONS(3165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3165),
    [anon_sym_LBRACK] = ACTIONS(3165),
    [anon_sym_LT_LT] = ACTIONS(3165),
    [anon_sym_SLASH_SLASH] = ACTIONS(3165),
    [anon_sym_PLUS] = ACTIONS(3165),
    [anon_sym_DASH] = ACTIONS(3165),
    [anon_sym_TILDE] = ACTIONS(3165),
    [anon_sym_BANG] = ACTIONS(3165),
    [anon_sym_PLUS_PLUS] = ACTIONS(3165),
    [anon_sym_DASH_DASH] = ACTIONS(3165),
    [anon_sym_BSLASH] = ACTIONS(3163),
    [anon_sym_do] = ACTIONS(3165),
    [anon_sym_eval] = ACTIONS(3165),
    [anon_sym_await] = ACTIONS(3165),
    [anon_sym_field] = ACTIONS(3165),
    [anon_sym_local] = ACTIONS(3165),
    [anon_sym_dynamically] = ACTIONS(3165),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3165),
    [anon_sym_require] = ACTIONS(3165),
    [anon_sym_map] = ACTIONS(3165),
    [anon_sym_grep] = ACTIONS(3165),
    [anon_sym_sort] = ACTIONS(3165),
    [anon_sym_goto] = ACTIONS(3165),
    [anon_sym_return] = ACTIONS(3165),
    [anon_sym_undef] = ACTIONS(3165),
    [anon_sym_PERCENT2] = ACTIONS(3165),
    [anon_sym_AMP2] = ACTIONS(3165),
    [anon_sym_STAR2] = ACTIONS(3165),
    [anon_sym_use] = ACTIONS(3165),
    [anon_sym_no] = ACTIONS(3165),
    [anon_sym_for] = ACTIONS(3165),
    [anon_sym_foreach] = ACTIONS(3165),
    [anon_sym_last] = ACTIONS(3165),
    [anon_sym_next] = ACTIONS(3165),
    [anon_sym_redo] = ACTIONS(3165),
    [anon_sym_BEGIN] = ACTIONS(3165),
    [anon_sym_INIT] = ACTIONS(3165),
    [anon_sym_CHECK] = ACTIONS(3165),
    [anon_sym_UNITCHECK] = ACTIONS(3165),
    [anon_sym_END] = ACTIONS(3165),
    [anon_sym___FILE__] = ACTIONS(3165),
    [anon_sym___LINE__] = ACTIONS(3165),
    [anon_sym___PACKAGE__] = ACTIONS(3165),
    [anon_sym___SUB__] = ACTIONS(3165),
    [anon_sym_break] = ACTIONS(3165),
    [anon_sym_fork] = ACTIONS(3165),
    [anon_sym_getppid] = ACTIONS(3165),
    [anon_sym_time] = ACTIONS(3165),
    [anon_sym_times] = ACTIONS(3165),
    [anon_sym_wait] = ACTIONS(3165),
    [anon_sym_wantarray] = ACTIONS(3165),
    [anon_sym_abs] = ACTIONS(3165),
    [anon_sym_alarm] = ACTIONS(3165),
    [anon_sym_chop] = ACTIONS(3165),
    [anon_sym_chdir] = ACTIONS(3165),
    [anon_sym_close] = ACTIONS(3165),
    [anon_sym_closedir] = ACTIONS(3165),
    [anon_sym_caller] = ACTIONS(3165),
    [anon_sym_chomp] = ACTIONS(3165),
    [anon_sym_chr] = ACTIONS(3165),
    [anon_sym_cos] = ACTIONS(3165),
    [anon_sym_chroot] = ACTIONS(3165),
    [anon_sym_defined] = ACTIONS(3165),
    [anon_sym_delete] = ACTIONS(3165),
    [anon_sym_dbmclose] = ACTIONS(3165),
    [anon_sym_exists] = ACTIONS(3165),
    [anon_sym_exit] = ACTIONS(3165),
    [anon_sym_eof] = ACTIONS(3165),
    [anon_sym_exp] = ACTIONS(3165),
    [anon_sym_each] = ACTIONS(3165),
    [anon_sym_fc] = ACTIONS(3165),
    [anon_sym_fileno] = ACTIONS(3165),
    [anon_sym_gmtime] = ACTIONS(3165),
    [anon_sym_getc] = ACTIONS(3165),
    [anon_sym_getpgrp] = ACTIONS(3165),
    [anon_sym_getprotobyname] = ACTIONS(3165),
    [anon_sym_getpwname] = ACTIONS(3165),
    [anon_sym_getpwuid] = ACTIONS(3165),
    [anon_sym_getpeername] = ACTIONS(3165),
    [anon_sym_getnetbyname] = ACTIONS(3165),
    [anon_sym_getsockname] = ACTIONS(3165),
    [anon_sym_getgrnam] = ACTIONS(3165),
    [anon_sym_getgrgid] = ACTIONS(3165),
    [anon_sym_hex] = ACTIONS(3165),
    [anon_sym_int] = ACTIONS(3165),
    [anon_sym_keys] = ACTIONS(3165),
    [anon_sym_lc] = ACTIONS(3165),
    [anon_sym_lcfirst] = ACTIONS(3165),
    [anon_sym_length] = ACTIONS(3165),
    [anon_sym_localtime] = ACTIONS(3165),
    [anon_sym_log] = ACTIONS(3165),
    [anon_sym_lock] = ACTIONS(3165),
    [anon_sym_lstat] = ACTIONS(3165),
    [anon_sym_oct] = ACTIONS(3165),
    [anon_sym_ord] = ACTIONS(3165),
    [anon_sym_prototype] = ACTIONS(3165),
    [anon_sym_pop] = ACTIONS(3165),
    [anon_sym_pos] = ACTIONS(3165),
    [anon_sym_quotemeta] = ACTIONS(3165),
    [anon_sym_reset] = ACTIONS(3165),
    [anon_sym_rand] = ACTIONS(3165),
    [anon_sym_rmdir] = ACTIONS(3165),
    [anon_sym_readdir] = ACTIONS(3165),
    [anon_sym_readline] = ACTIONS(3165),
    [anon_sym_readpipe] = ACTIONS(3165),
    [anon_sym_rewinddir] = ACTIONS(3165),
    [anon_sym_readlink] = ACTIONS(3165),
    [anon_sym_ref] = ACTIONS(3165),
    [anon_sym_scalar] = ACTIONS(3165),
    [anon_sym_shift] = ACTIONS(3165),
    [anon_sym_sin] = ACTIONS(3165),
    [anon_sym_sleep] = ACTIONS(3165),
    [anon_sym_sqrt] = ACTIONS(3165),
    [anon_sym_srand] = ACTIONS(3165),
    [anon_sym_stat] = ACTIONS(3165),
    [anon_sym_study] = ACTIONS(3165),
    [anon_sym_tell] = ACTIONS(3165),
    [anon_sym_telldir] = ACTIONS(3165),
    [anon_sym_tied] = ACTIONS(3165),
    [anon_sym_uc] = ACTIONS(3165),
    [anon_sym_ucfirst] = ACTIONS(3165),
    [anon_sym_untie] = ACTIONS(3165),
    [anon_sym_umask] = ACTIONS(3165),
    [anon_sym_values] = ACTIONS(3165),
    [anon_sym_write] = ACTIONS(3165),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3165),
    [anon_sym_qq] = ACTIONS(3165),
    [anon_sym_qw] = ACTIONS(3165),
    [anon_sym_qx] = ACTIONS(3165),
    [anon_sym_qr] = ACTIONS(3165),
    [anon_sym_m] = ACTIONS(3165),
    [anon_sym_s] = ACTIONS(3165),
    [anon_sym_tr] = ACTIONS(3165),
    [anon_sym_y] = ACTIONS(3165),
    [anon_sym_if] = ACTIONS(3165),
    [anon_sym_unless] = ACTIONS(3165),
    [anon_sym_while] = ACTIONS(3165),
    [anon_sym_until] = ACTIONS(3165),
    [aux_sym__bareword_token1] = ACTIONS(3165),
    [sym_number] = ACTIONS(3165),
    [anon_sym_true] = ACTIONS(3165),
    [anon_sym_false] = ACTIONS(3165),
    [sym__single_quote] = ACTIONS(3163),
    [sym__double_quote] = ACTIONS(3163),
    [sym__backtick_quote] = ACTIONS(3163),
    [sym__search_slash_quote] = ACTIONS(3163),
    [sym__open_readline_bracket] = ACTIONS(3163),
    [sym__open_fileglob_bracket] = ACTIONS(3163),
    [sym__PERLY_HEREDOC] = ACTIONS(3163),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3163),
    [sym__filetest] = ACTIONS(3163),
  },
  [STATE(1568)] = {
    [sym_heredoc_content] = STATE(1568),
    [sym__identifier] = ACTIONS(3229),
    [anon_sym_LBRACE] = ACTIONS(3229),
    [anon_sym_RBRACE] = ACTIONS(3229),
    [anon_sym_SEMI] = ACTIONS(3229),
    [anon_sym_package] = ACTIONS(3229),
    [anon_sym_class] = ACTIONS(3229),
    [anon_sym_role] = ACTIONS(3229),
    [anon_sym_BUILD] = ACTIONS(3229),
    [anon_sym_ADJUST] = ACTIONS(3229),
    [anon_sym_DOLLAR] = ACTIONS(3229),
    [anon_sym_AT] = ACTIONS(3229),
    [anon_sym_my] = ACTIONS(3229),
    [anon_sym_extended] = ACTIONS(3229),
    [anon_sym_async] = ACTIONS(3229),
    [anon_sym_sub] = ACTIONS(3229),
    [anon_sym_method] = ACTIONS(3229),
    [anon_sym_LPAREN] = ACTIONS(3229),
    [anon_sym_continue] = ACTIONS(3229),
    [anon_sym_state] = ACTIONS(3229),
    [anon_sym_our] = ACTIONS(3229),
    [anon_sym_try] = ACTIONS(3229),
    [anon_sym_defer] = ACTIONS(3229),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3229),
    [anon_sym_LBRACK] = ACTIONS(3229),
    [anon_sym_LT_LT] = ACTIONS(3229),
    [anon_sym_SLASH_SLASH] = ACTIONS(3229),
    [anon_sym_PLUS] = ACTIONS(3229),
    [anon_sym_DASH] = ACTIONS(3229),
    [anon_sym_TILDE] = ACTIONS(3229),
    [anon_sym_BANG] = ACTIONS(3229),
    [anon_sym_PLUS_PLUS] = ACTIONS(3229),
    [anon_sym_DASH_DASH] = ACTIONS(3229),
    [anon_sym_BSLASH] = ACTIONS(3227),
    [anon_sym_do] = ACTIONS(3229),
    [anon_sym_eval] = ACTIONS(3229),
    [anon_sym_await] = ACTIONS(3229),
    [anon_sym_field] = ACTIONS(3229),
    [anon_sym_local] = ACTIONS(3229),
    [anon_sym_dynamically] = ACTIONS(3229),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3229),
    [anon_sym_require] = ACTIONS(3229),
    [anon_sym_map] = ACTIONS(3229),
    [anon_sym_grep] = ACTIONS(3229),
    [anon_sym_sort] = ACTIONS(3229),
    [anon_sym_goto] = ACTIONS(3229),
    [anon_sym_return] = ACTIONS(3229),
    [anon_sym_undef] = ACTIONS(3229),
    [anon_sym_PERCENT2] = ACTIONS(3229),
    [anon_sym_AMP2] = ACTIONS(3229),
    [anon_sym_STAR2] = ACTIONS(3229),
    [anon_sym_use] = ACTIONS(3229),
    [anon_sym_no] = ACTIONS(3229),
    [anon_sym_for] = ACTIONS(3229),
    [anon_sym_foreach] = ACTIONS(3229),
    [anon_sym_last] = ACTIONS(3229),
    [anon_sym_next] = ACTIONS(3229),
    [anon_sym_redo] = ACTIONS(3229),
    [anon_sym_BEGIN] = ACTIONS(3229),
    [anon_sym_INIT] = ACTIONS(3229),
    [anon_sym_CHECK] = ACTIONS(3229),
    [anon_sym_UNITCHECK] = ACTIONS(3229),
    [anon_sym_END] = ACTIONS(3229),
    [anon_sym___FILE__] = ACTIONS(3229),
    [anon_sym___LINE__] = ACTIONS(3229),
    [anon_sym___PACKAGE__] = ACTIONS(3229),
    [anon_sym___SUB__] = ACTIONS(3229),
    [anon_sym_break] = ACTIONS(3229),
    [anon_sym_fork] = ACTIONS(3229),
    [anon_sym_getppid] = ACTIONS(3229),
    [anon_sym_time] = ACTIONS(3229),
    [anon_sym_times] = ACTIONS(3229),
    [anon_sym_wait] = ACTIONS(3229),
    [anon_sym_wantarray] = ACTIONS(3229),
    [anon_sym_abs] = ACTIONS(3229),
    [anon_sym_alarm] = ACTIONS(3229),
    [anon_sym_chop] = ACTIONS(3229),
    [anon_sym_chdir] = ACTIONS(3229),
    [anon_sym_close] = ACTIONS(3229),
    [anon_sym_closedir] = ACTIONS(3229),
    [anon_sym_caller] = ACTIONS(3229),
    [anon_sym_chomp] = ACTIONS(3229),
    [anon_sym_chr] = ACTIONS(3229),
    [anon_sym_cos] = ACTIONS(3229),
    [anon_sym_chroot] = ACTIONS(3229),
    [anon_sym_defined] = ACTIONS(3229),
    [anon_sym_delete] = ACTIONS(3229),
    [anon_sym_dbmclose] = ACTIONS(3229),
    [anon_sym_exists] = ACTIONS(3229),
    [anon_sym_exit] = ACTIONS(3229),
    [anon_sym_eof] = ACTIONS(3229),
    [anon_sym_exp] = ACTIONS(3229),
    [anon_sym_each] = ACTIONS(3229),
    [anon_sym_fc] = ACTIONS(3229),
    [anon_sym_fileno] = ACTIONS(3229),
    [anon_sym_gmtime] = ACTIONS(3229),
    [anon_sym_getc] = ACTIONS(3229),
    [anon_sym_getpgrp] = ACTIONS(3229),
    [anon_sym_getprotobyname] = ACTIONS(3229),
    [anon_sym_getpwname] = ACTIONS(3229),
    [anon_sym_getpwuid] = ACTIONS(3229),
    [anon_sym_getpeername] = ACTIONS(3229),
    [anon_sym_getnetbyname] = ACTIONS(3229),
    [anon_sym_getsockname] = ACTIONS(3229),
    [anon_sym_getgrnam] = ACTIONS(3229),
    [anon_sym_getgrgid] = ACTIONS(3229),
    [anon_sym_hex] = ACTIONS(3229),
    [anon_sym_int] = ACTIONS(3229),
    [anon_sym_keys] = ACTIONS(3229),
    [anon_sym_lc] = ACTIONS(3229),
    [anon_sym_lcfirst] = ACTIONS(3229),
    [anon_sym_length] = ACTIONS(3229),
    [anon_sym_localtime] = ACTIONS(3229),
    [anon_sym_log] = ACTIONS(3229),
    [anon_sym_lock] = ACTIONS(3229),
    [anon_sym_lstat] = ACTIONS(3229),
    [anon_sym_oct] = ACTIONS(3229),
    [anon_sym_ord] = ACTIONS(3229),
    [anon_sym_prototype] = ACTIONS(3229),
    [anon_sym_pop] = ACTIONS(3229),
    [anon_sym_pos] = ACTIONS(3229),
    [anon_sym_quotemeta] = ACTIONS(3229),
    [anon_sym_reset] = ACTIONS(3229),
    [anon_sym_rand] = ACTIONS(3229),
    [anon_sym_rmdir] = ACTIONS(3229),
    [anon_sym_readdir] = ACTIONS(3229),
    [anon_sym_readline] = ACTIONS(3229),
    [anon_sym_readpipe] = ACTIONS(3229),
    [anon_sym_rewinddir] = ACTIONS(3229),
    [anon_sym_readlink] = ACTIONS(3229),
    [anon_sym_ref] = ACTIONS(3229),
    [anon_sym_scalar] = ACTIONS(3229),
    [anon_sym_shift] = ACTIONS(3229),
    [anon_sym_sin] = ACTIONS(3229),
    [anon_sym_sleep] = ACTIONS(3229),
    [anon_sym_sqrt] = ACTIONS(3229),
    [anon_sym_srand] = ACTIONS(3229),
    [anon_sym_stat] = ACTIONS(3229),
    [anon_sym_study] = ACTIONS(3229),
    [anon_sym_tell] = ACTIONS(3229),
    [anon_sym_telldir] = ACTIONS(3229),
    [anon_sym_tied] = ACTIONS(3229),
    [anon_sym_uc] = ACTIONS(3229),
    [anon_sym_ucfirst] = ACTIONS(3229),
    [anon_sym_untie] = ACTIONS(3229),
    [anon_sym_umask] = ACTIONS(3229),
    [anon_sym_values] = ACTIONS(3229),
    [anon_sym_write] = ACTIONS(3229),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3229),
    [anon_sym_qq] = ACTIONS(3229),
    [anon_sym_qw] = ACTIONS(3229),
    [anon_sym_qx] = ACTIONS(3229),
    [anon_sym_qr] = ACTIONS(3229),
    [anon_sym_m] = ACTIONS(3229),
    [anon_sym_s] = ACTIONS(3229),
    [anon_sym_tr] = ACTIONS(3229),
    [anon_sym_y] = ACTIONS(3229),
    [anon_sym_if] = ACTIONS(3229),
    [anon_sym_unless] = ACTIONS(3229),
    [anon_sym_while] = ACTIONS(3229),
    [anon_sym_until] = ACTIONS(3229),
    [aux_sym__bareword_token1] = ACTIONS(3229),
    [sym_number] = ACTIONS(3229),
    [anon_sym_true] = ACTIONS(3229),
    [anon_sym_false] = ACTIONS(3229),
    [sym__single_quote] = ACTIONS(3227),
    [sym__double_quote] = ACTIONS(3227),
    [sym__backtick_quote] = ACTIONS(3227),
    [sym__search_slash_quote] = ACTIONS(3227),
    [sym__open_readline_bracket] = ACTIONS(3227),
    [sym__open_fileglob_bracket] = ACTIONS(3227),
    [sym__PERLY_HEREDOC] = ACTIONS(3227),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3227),
    [sym__filetest] = ACTIONS(3227),
  },
  [STATE(1569)] = {
    [sym_heredoc_content] = STATE(1569),
    [sym__identifier] = ACTIONS(2930),
    [anon_sym_LBRACE] = ACTIONS(2930),
    [anon_sym_RBRACE] = ACTIONS(2930),
    [anon_sym_SEMI] = ACTIONS(2930),
    [anon_sym_package] = ACTIONS(2930),
    [anon_sym_class] = ACTIONS(2930),
    [anon_sym_role] = ACTIONS(2930),
    [anon_sym_BUILD] = ACTIONS(2930),
    [anon_sym_ADJUST] = ACTIONS(2930),
    [anon_sym_DOLLAR] = ACTIONS(2930),
    [anon_sym_AT] = ACTIONS(2930),
    [anon_sym_my] = ACTIONS(2930),
    [anon_sym_extended] = ACTIONS(2930),
    [anon_sym_async] = ACTIONS(2930),
    [anon_sym_sub] = ACTIONS(2930),
    [anon_sym_method] = ACTIONS(2930),
    [anon_sym_LPAREN] = ACTIONS(2930),
    [anon_sym_continue] = ACTIONS(2930),
    [anon_sym_state] = ACTIONS(2930),
    [anon_sym_our] = ACTIONS(2930),
    [anon_sym_try] = ACTIONS(2930),
    [anon_sym_defer] = ACTIONS(2930),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2930),
    [anon_sym_LBRACK] = ACTIONS(2930),
    [anon_sym_LT_LT] = ACTIONS(2930),
    [anon_sym_SLASH_SLASH] = ACTIONS(2930),
    [anon_sym_PLUS] = ACTIONS(2930),
    [anon_sym_DASH] = ACTIONS(2930),
    [anon_sym_TILDE] = ACTIONS(2930),
    [anon_sym_BANG] = ACTIONS(2930),
    [anon_sym_PLUS_PLUS] = ACTIONS(2930),
    [anon_sym_DASH_DASH] = ACTIONS(2930),
    [anon_sym_BSLASH] = ACTIONS(2928),
    [anon_sym_do] = ACTIONS(2930),
    [anon_sym_eval] = ACTIONS(2930),
    [anon_sym_await] = ACTIONS(2930),
    [anon_sym_field] = ACTIONS(2930),
    [anon_sym_local] = ACTIONS(2930),
    [anon_sym_dynamically] = ACTIONS(2930),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2930),
    [anon_sym_require] = ACTIONS(2930),
    [anon_sym_map] = ACTIONS(2930),
    [anon_sym_grep] = ACTIONS(2930),
    [anon_sym_sort] = ACTIONS(2930),
    [anon_sym_goto] = ACTIONS(2930),
    [anon_sym_return] = ACTIONS(2930),
    [anon_sym_undef] = ACTIONS(2930),
    [anon_sym_PERCENT2] = ACTIONS(2930),
    [anon_sym_AMP2] = ACTIONS(2930),
    [anon_sym_STAR2] = ACTIONS(2930),
    [anon_sym_use] = ACTIONS(2930),
    [anon_sym_no] = ACTIONS(2930),
    [anon_sym_for] = ACTIONS(2930),
    [anon_sym_foreach] = ACTIONS(2930),
    [anon_sym_last] = ACTIONS(2930),
    [anon_sym_next] = ACTIONS(2930),
    [anon_sym_redo] = ACTIONS(2930),
    [anon_sym_BEGIN] = ACTIONS(2930),
    [anon_sym_INIT] = ACTIONS(2930),
    [anon_sym_CHECK] = ACTIONS(2930),
    [anon_sym_UNITCHECK] = ACTIONS(2930),
    [anon_sym_END] = ACTIONS(2930),
    [anon_sym___FILE__] = ACTIONS(2930),
    [anon_sym___LINE__] = ACTIONS(2930),
    [anon_sym___PACKAGE__] = ACTIONS(2930),
    [anon_sym___SUB__] = ACTIONS(2930),
    [anon_sym_break] = ACTIONS(2930),
    [anon_sym_fork] = ACTIONS(2930),
    [anon_sym_getppid] = ACTIONS(2930),
    [anon_sym_time] = ACTIONS(2930),
    [anon_sym_times] = ACTIONS(2930),
    [anon_sym_wait] = ACTIONS(2930),
    [anon_sym_wantarray] = ACTIONS(2930),
    [anon_sym_abs] = ACTIONS(2930),
    [anon_sym_alarm] = ACTIONS(2930),
    [anon_sym_chop] = ACTIONS(2930),
    [anon_sym_chdir] = ACTIONS(2930),
    [anon_sym_close] = ACTIONS(2930),
    [anon_sym_closedir] = ACTIONS(2930),
    [anon_sym_caller] = ACTIONS(2930),
    [anon_sym_chomp] = ACTIONS(2930),
    [anon_sym_chr] = ACTIONS(2930),
    [anon_sym_cos] = ACTIONS(2930),
    [anon_sym_chroot] = ACTIONS(2930),
    [anon_sym_defined] = ACTIONS(2930),
    [anon_sym_delete] = ACTIONS(2930),
    [anon_sym_dbmclose] = ACTIONS(2930),
    [anon_sym_exists] = ACTIONS(2930),
    [anon_sym_exit] = ACTIONS(2930),
    [anon_sym_eof] = ACTIONS(2930),
    [anon_sym_exp] = ACTIONS(2930),
    [anon_sym_each] = ACTIONS(2930),
    [anon_sym_fc] = ACTIONS(2930),
    [anon_sym_fileno] = ACTIONS(2930),
    [anon_sym_gmtime] = ACTIONS(2930),
    [anon_sym_getc] = ACTIONS(2930),
    [anon_sym_getpgrp] = ACTIONS(2930),
    [anon_sym_getprotobyname] = ACTIONS(2930),
    [anon_sym_getpwname] = ACTIONS(2930),
    [anon_sym_getpwuid] = ACTIONS(2930),
    [anon_sym_getpeername] = ACTIONS(2930),
    [anon_sym_getnetbyname] = ACTIONS(2930),
    [anon_sym_getsockname] = ACTIONS(2930),
    [anon_sym_getgrnam] = ACTIONS(2930),
    [anon_sym_getgrgid] = ACTIONS(2930),
    [anon_sym_hex] = ACTIONS(2930),
    [anon_sym_int] = ACTIONS(2930),
    [anon_sym_keys] = ACTIONS(2930),
    [anon_sym_lc] = ACTIONS(2930),
    [anon_sym_lcfirst] = ACTIONS(2930),
    [anon_sym_length] = ACTIONS(2930),
    [anon_sym_localtime] = ACTIONS(2930),
    [anon_sym_log] = ACTIONS(2930),
    [anon_sym_lock] = ACTIONS(2930),
    [anon_sym_lstat] = ACTIONS(2930),
    [anon_sym_oct] = ACTIONS(2930),
    [anon_sym_ord] = ACTIONS(2930),
    [anon_sym_prototype] = ACTIONS(2930),
    [anon_sym_pop] = ACTIONS(2930),
    [anon_sym_pos] = ACTIONS(2930),
    [anon_sym_quotemeta] = ACTIONS(2930),
    [anon_sym_reset] = ACTIONS(2930),
    [anon_sym_rand] = ACTIONS(2930),
    [anon_sym_rmdir] = ACTIONS(2930),
    [anon_sym_readdir] = ACTIONS(2930),
    [anon_sym_readline] = ACTIONS(2930),
    [anon_sym_readpipe] = ACTIONS(2930),
    [anon_sym_rewinddir] = ACTIONS(2930),
    [anon_sym_readlink] = ACTIONS(2930),
    [anon_sym_ref] = ACTIONS(2930),
    [anon_sym_scalar] = ACTIONS(2930),
    [anon_sym_shift] = ACTIONS(2930),
    [anon_sym_sin] = ACTIONS(2930),
    [anon_sym_sleep] = ACTIONS(2930),
    [anon_sym_sqrt] = ACTIONS(2930),
    [anon_sym_srand] = ACTIONS(2930),
    [anon_sym_stat] = ACTIONS(2930),
    [anon_sym_study] = ACTIONS(2930),
    [anon_sym_tell] = ACTIONS(2930),
    [anon_sym_telldir] = ACTIONS(2930),
    [anon_sym_tied] = ACTIONS(2930),
    [anon_sym_uc] = ACTIONS(2930),
    [anon_sym_ucfirst] = ACTIONS(2930),
    [anon_sym_untie] = ACTIONS(2930),
    [anon_sym_umask] = ACTIONS(2930),
    [anon_sym_values] = ACTIONS(2930),
    [anon_sym_write] = ACTIONS(2930),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2930),
    [anon_sym_qq] = ACTIONS(2930),
    [anon_sym_qw] = ACTIONS(2930),
    [anon_sym_qx] = ACTIONS(2930),
    [anon_sym_qr] = ACTIONS(2930),
    [anon_sym_m] = ACTIONS(2930),
    [anon_sym_s] = ACTIONS(2930),
    [anon_sym_tr] = ACTIONS(2930),
    [anon_sym_y] = ACTIONS(2930),
    [anon_sym_if] = ACTIONS(2930),
    [anon_sym_unless] = ACTIONS(2930),
    [anon_sym_while] = ACTIONS(2930),
    [anon_sym_until] = ACTIONS(2930),
    [aux_sym__bareword_token1] = ACTIONS(2930),
    [sym_number] = ACTIONS(2930),
    [anon_sym_true] = ACTIONS(2930),
    [anon_sym_false] = ACTIONS(2930),
    [sym__single_quote] = ACTIONS(2928),
    [sym__double_quote] = ACTIONS(2928),
    [sym__backtick_quote] = ACTIONS(2928),
    [sym__search_slash_quote] = ACTIONS(2928),
    [sym__open_readline_bracket] = ACTIONS(2928),
    [sym__open_fileglob_bracket] = ACTIONS(2928),
    [sym__PERLY_HEREDOC] = ACTIONS(2928),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2928),
    [sym__filetest] = ACTIONS(2928),
  },
  [STATE(1570)] = {
    [sym_heredoc_content] = STATE(1570),
    [sym__identifier] = ACTIONS(2934),
    [anon_sym_LBRACE] = ACTIONS(2934),
    [anon_sym_RBRACE] = ACTIONS(2934),
    [anon_sym_SEMI] = ACTIONS(2934),
    [anon_sym_package] = ACTIONS(2934),
    [anon_sym_class] = ACTIONS(2934),
    [anon_sym_role] = ACTIONS(2934),
    [anon_sym_BUILD] = ACTIONS(2934),
    [anon_sym_ADJUST] = ACTIONS(2934),
    [anon_sym_DOLLAR] = ACTIONS(2934),
    [anon_sym_AT] = ACTIONS(2934),
    [anon_sym_my] = ACTIONS(2934),
    [anon_sym_extended] = ACTIONS(2934),
    [anon_sym_async] = ACTIONS(2934),
    [anon_sym_sub] = ACTIONS(2934),
    [anon_sym_method] = ACTIONS(2934),
    [anon_sym_LPAREN] = ACTIONS(2934),
    [anon_sym_continue] = ACTIONS(2934),
    [anon_sym_state] = ACTIONS(2934),
    [anon_sym_our] = ACTIONS(2934),
    [anon_sym_try] = ACTIONS(2934),
    [anon_sym_defer] = ACTIONS(2934),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2934),
    [anon_sym_LBRACK] = ACTIONS(2934),
    [anon_sym_LT_LT] = ACTIONS(2934),
    [anon_sym_SLASH_SLASH] = ACTIONS(2934),
    [anon_sym_PLUS] = ACTIONS(2934),
    [anon_sym_DASH] = ACTIONS(2934),
    [anon_sym_TILDE] = ACTIONS(2934),
    [anon_sym_BANG] = ACTIONS(2934),
    [anon_sym_PLUS_PLUS] = ACTIONS(2934),
    [anon_sym_DASH_DASH] = ACTIONS(2934),
    [anon_sym_BSLASH] = ACTIONS(2932),
    [anon_sym_do] = ACTIONS(2934),
    [anon_sym_eval] = ACTIONS(2934),
    [anon_sym_await] = ACTIONS(2934),
    [anon_sym_field] = ACTIONS(2934),
    [anon_sym_local] = ACTIONS(2934),
    [anon_sym_dynamically] = ACTIONS(2934),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2934),
    [anon_sym_require] = ACTIONS(2934),
    [anon_sym_map] = ACTIONS(2934),
    [anon_sym_grep] = ACTIONS(2934),
    [anon_sym_sort] = ACTIONS(2934),
    [anon_sym_goto] = ACTIONS(2934),
    [anon_sym_return] = ACTIONS(2934),
    [anon_sym_undef] = ACTIONS(2934),
    [anon_sym_PERCENT2] = ACTIONS(2934),
    [anon_sym_AMP2] = ACTIONS(2934),
    [anon_sym_STAR2] = ACTIONS(2934),
    [anon_sym_use] = ACTIONS(2934),
    [anon_sym_no] = ACTIONS(2934),
    [anon_sym_for] = ACTIONS(2934),
    [anon_sym_foreach] = ACTIONS(2934),
    [anon_sym_last] = ACTIONS(2934),
    [anon_sym_next] = ACTIONS(2934),
    [anon_sym_redo] = ACTIONS(2934),
    [anon_sym_BEGIN] = ACTIONS(2934),
    [anon_sym_INIT] = ACTIONS(2934),
    [anon_sym_CHECK] = ACTIONS(2934),
    [anon_sym_UNITCHECK] = ACTIONS(2934),
    [anon_sym_END] = ACTIONS(2934),
    [anon_sym___FILE__] = ACTIONS(2934),
    [anon_sym___LINE__] = ACTIONS(2934),
    [anon_sym___PACKAGE__] = ACTIONS(2934),
    [anon_sym___SUB__] = ACTIONS(2934),
    [anon_sym_break] = ACTIONS(2934),
    [anon_sym_fork] = ACTIONS(2934),
    [anon_sym_getppid] = ACTIONS(2934),
    [anon_sym_time] = ACTIONS(2934),
    [anon_sym_times] = ACTIONS(2934),
    [anon_sym_wait] = ACTIONS(2934),
    [anon_sym_wantarray] = ACTIONS(2934),
    [anon_sym_abs] = ACTIONS(2934),
    [anon_sym_alarm] = ACTIONS(2934),
    [anon_sym_chop] = ACTIONS(2934),
    [anon_sym_chdir] = ACTIONS(2934),
    [anon_sym_close] = ACTIONS(2934),
    [anon_sym_closedir] = ACTIONS(2934),
    [anon_sym_caller] = ACTIONS(2934),
    [anon_sym_chomp] = ACTIONS(2934),
    [anon_sym_chr] = ACTIONS(2934),
    [anon_sym_cos] = ACTIONS(2934),
    [anon_sym_chroot] = ACTIONS(2934),
    [anon_sym_defined] = ACTIONS(2934),
    [anon_sym_delete] = ACTIONS(2934),
    [anon_sym_dbmclose] = ACTIONS(2934),
    [anon_sym_exists] = ACTIONS(2934),
    [anon_sym_exit] = ACTIONS(2934),
    [anon_sym_eof] = ACTIONS(2934),
    [anon_sym_exp] = ACTIONS(2934),
    [anon_sym_each] = ACTIONS(2934),
    [anon_sym_fc] = ACTIONS(2934),
    [anon_sym_fileno] = ACTIONS(2934),
    [anon_sym_gmtime] = ACTIONS(2934),
    [anon_sym_getc] = ACTIONS(2934),
    [anon_sym_getpgrp] = ACTIONS(2934),
    [anon_sym_getprotobyname] = ACTIONS(2934),
    [anon_sym_getpwname] = ACTIONS(2934),
    [anon_sym_getpwuid] = ACTIONS(2934),
    [anon_sym_getpeername] = ACTIONS(2934),
    [anon_sym_getnetbyname] = ACTIONS(2934),
    [anon_sym_getsockname] = ACTIONS(2934),
    [anon_sym_getgrnam] = ACTIONS(2934),
    [anon_sym_getgrgid] = ACTIONS(2934),
    [anon_sym_hex] = ACTIONS(2934),
    [anon_sym_int] = ACTIONS(2934),
    [anon_sym_keys] = ACTIONS(2934),
    [anon_sym_lc] = ACTIONS(2934),
    [anon_sym_lcfirst] = ACTIONS(2934),
    [anon_sym_length] = ACTIONS(2934),
    [anon_sym_localtime] = ACTIONS(2934),
    [anon_sym_log] = ACTIONS(2934),
    [anon_sym_lock] = ACTIONS(2934),
    [anon_sym_lstat] = ACTIONS(2934),
    [anon_sym_oct] = ACTIONS(2934),
    [anon_sym_ord] = ACTIONS(2934),
    [anon_sym_prototype] = ACTIONS(2934),
    [anon_sym_pop] = ACTIONS(2934),
    [anon_sym_pos] = ACTIONS(2934),
    [anon_sym_quotemeta] = ACTIONS(2934),
    [anon_sym_reset] = ACTIONS(2934),
    [anon_sym_rand] = ACTIONS(2934),
    [anon_sym_rmdir] = ACTIONS(2934),
    [anon_sym_readdir] = ACTIONS(2934),
    [anon_sym_readline] = ACTIONS(2934),
    [anon_sym_readpipe] = ACTIONS(2934),
    [anon_sym_rewinddir] = ACTIONS(2934),
    [anon_sym_readlink] = ACTIONS(2934),
    [anon_sym_ref] = ACTIONS(2934),
    [anon_sym_scalar] = ACTIONS(2934),
    [anon_sym_shift] = ACTIONS(2934),
    [anon_sym_sin] = ACTIONS(2934),
    [anon_sym_sleep] = ACTIONS(2934),
    [anon_sym_sqrt] = ACTIONS(2934),
    [anon_sym_srand] = ACTIONS(2934),
    [anon_sym_stat] = ACTIONS(2934),
    [anon_sym_study] = ACTIONS(2934),
    [anon_sym_tell] = ACTIONS(2934),
    [anon_sym_telldir] = ACTIONS(2934),
    [anon_sym_tied] = ACTIONS(2934),
    [anon_sym_uc] = ACTIONS(2934),
    [anon_sym_ucfirst] = ACTIONS(2934),
    [anon_sym_untie] = ACTIONS(2934),
    [anon_sym_umask] = ACTIONS(2934),
    [anon_sym_values] = ACTIONS(2934),
    [anon_sym_write] = ACTIONS(2934),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2934),
    [anon_sym_qq] = ACTIONS(2934),
    [anon_sym_qw] = ACTIONS(2934),
    [anon_sym_qx] = ACTIONS(2934),
    [anon_sym_qr] = ACTIONS(2934),
    [anon_sym_m] = ACTIONS(2934),
    [anon_sym_s] = ACTIONS(2934),
    [anon_sym_tr] = ACTIONS(2934),
    [anon_sym_y] = ACTIONS(2934),
    [anon_sym_if] = ACTIONS(2934),
    [anon_sym_unless] = ACTIONS(2934),
    [anon_sym_while] = ACTIONS(2934),
    [anon_sym_until] = ACTIONS(2934),
    [aux_sym__bareword_token1] = ACTIONS(2934),
    [sym_number] = ACTIONS(2934),
    [anon_sym_true] = ACTIONS(2934),
    [anon_sym_false] = ACTIONS(2934),
    [sym__single_quote] = ACTIONS(2932),
    [sym__double_quote] = ACTIONS(2932),
    [sym__backtick_quote] = ACTIONS(2932),
    [sym__search_slash_quote] = ACTIONS(2932),
    [sym__open_readline_bracket] = ACTIONS(2932),
    [sym__open_fileglob_bracket] = ACTIONS(2932),
    [sym__PERLY_HEREDOC] = ACTIONS(2932),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2932),
    [sym__filetest] = ACTIONS(2932),
  },
  [STATE(1571)] = {
    [sym_heredoc_content] = STATE(1571),
    [sym__identifier] = ACTIONS(2938),
    [anon_sym_LBRACE] = ACTIONS(2938),
    [anon_sym_RBRACE] = ACTIONS(2938),
    [anon_sym_SEMI] = ACTIONS(2938),
    [anon_sym_package] = ACTIONS(2938),
    [anon_sym_class] = ACTIONS(2938),
    [anon_sym_role] = ACTIONS(2938),
    [anon_sym_BUILD] = ACTIONS(2938),
    [anon_sym_ADJUST] = ACTIONS(2938),
    [anon_sym_DOLLAR] = ACTIONS(2938),
    [anon_sym_AT] = ACTIONS(2938),
    [anon_sym_my] = ACTIONS(2938),
    [anon_sym_extended] = ACTIONS(2938),
    [anon_sym_async] = ACTIONS(2938),
    [anon_sym_sub] = ACTIONS(2938),
    [anon_sym_method] = ACTIONS(2938),
    [anon_sym_LPAREN] = ACTIONS(2938),
    [anon_sym_continue] = ACTIONS(2938),
    [anon_sym_state] = ACTIONS(2938),
    [anon_sym_our] = ACTIONS(2938),
    [anon_sym_try] = ACTIONS(2938),
    [anon_sym_defer] = ACTIONS(2938),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2938),
    [anon_sym_LBRACK] = ACTIONS(2938),
    [anon_sym_LT_LT] = ACTIONS(2938),
    [anon_sym_SLASH_SLASH] = ACTIONS(2938),
    [anon_sym_PLUS] = ACTIONS(2938),
    [anon_sym_DASH] = ACTIONS(2938),
    [anon_sym_TILDE] = ACTIONS(2938),
    [anon_sym_BANG] = ACTIONS(2938),
    [anon_sym_PLUS_PLUS] = ACTIONS(2938),
    [anon_sym_DASH_DASH] = ACTIONS(2938),
    [anon_sym_BSLASH] = ACTIONS(2936),
    [anon_sym_do] = ACTIONS(2938),
    [anon_sym_eval] = ACTIONS(2938),
    [anon_sym_await] = ACTIONS(2938),
    [anon_sym_field] = ACTIONS(2938),
    [anon_sym_local] = ACTIONS(2938),
    [anon_sym_dynamically] = ACTIONS(2938),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2938),
    [anon_sym_require] = ACTIONS(2938),
    [anon_sym_map] = ACTIONS(2938),
    [anon_sym_grep] = ACTIONS(2938),
    [anon_sym_sort] = ACTIONS(2938),
    [anon_sym_goto] = ACTIONS(2938),
    [anon_sym_return] = ACTIONS(2938),
    [anon_sym_undef] = ACTIONS(2938),
    [anon_sym_PERCENT2] = ACTIONS(2938),
    [anon_sym_AMP2] = ACTIONS(2938),
    [anon_sym_STAR2] = ACTIONS(2938),
    [anon_sym_use] = ACTIONS(2938),
    [anon_sym_no] = ACTIONS(2938),
    [anon_sym_for] = ACTIONS(2938),
    [anon_sym_foreach] = ACTIONS(2938),
    [anon_sym_last] = ACTIONS(2938),
    [anon_sym_next] = ACTIONS(2938),
    [anon_sym_redo] = ACTIONS(2938),
    [anon_sym_BEGIN] = ACTIONS(2938),
    [anon_sym_INIT] = ACTIONS(2938),
    [anon_sym_CHECK] = ACTIONS(2938),
    [anon_sym_UNITCHECK] = ACTIONS(2938),
    [anon_sym_END] = ACTIONS(2938),
    [anon_sym___FILE__] = ACTIONS(2938),
    [anon_sym___LINE__] = ACTIONS(2938),
    [anon_sym___PACKAGE__] = ACTIONS(2938),
    [anon_sym___SUB__] = ACTIONS(2938),
    [anon_sym_break] = ACTIONS(2938),
    [anon_sym_fork] = ACTIONS(2938),
    [anon_sym_getppid] = ACTIONS(2938),
    [anon_sym_time] = ACTIONS(2938),
    [anon_sym_times] = ACTIONS(2938),
    [anon_sym_wait] = ACTIONS(2938),
    [anon_sym_wantarray] = ACTIONS(2938),
    [anon_sym_abs] = ACTIONS(2938),
    [anon_sym_alarm] = ACTIONS(2938),
    [anon_sym_chop] = ACTIONS(2938),
    [anon_sym_chdir] = ACTIONS(2938),
    [anon_sym_close] = ACTIONS(2938),
    [anon_sym_closedir] = ACTIONS(2938),
    [anon_sym_caller] = ACTIONS(2938),
    [anon_sym_chomp] = ACTIONS(2938),
    [anon_sym_chr] = ACTIONS(2938),
    [anon_sym_cos] = ACTIONS(2938),
    [anon_sym_chroot] = ACTIONS(2938),
    [anon_sym_defined] = ACTIONS(2938),
    [anon_sym_delete] = ACTIONS(2938),
    [anon_sym_dbmclose] = ACTIONS(2938),
    [anon_sym_exists] = ACTIONS(2938),
    [anon_sym_exit] = ACTIONS(2938),
    [anon_sym_eof] = ACTIONS(2938),
    [anon_sym_exp] = ACTIONS(2938),
    [anon_sym_each] = ACTIONS(2938),
    [anon_sym_fc] = ACTIONS(2938),
    [anon_sym_fileno] = ACTIONS(2938),
    [anon_sym_gmtime] = ACTIONS(2938),
    [anon_sym_getc] = ACTIONS(2938),
    [anon_sym_getpgrp] = ACTIONS(2938),
    [anon_sym_getprotobyname] = ACTIONS(2938),
    [anon_sym_getpwname] = ACTIONS(2938),
    [anon_sym_getpwuid] = ACTIONS(2938),
    [anon_sym_getpeername] = ACTIONS(2938),
    [anon_sym_getnetbyname] = ACTIONS(2938),
    [anon_sym_getsockname] = ACTIONS(2938),
    [anon_sym_getgrnam] = ACTIONS(2938),
    [anon_sym_getgrgid] = ACTIONS(2938),
    [anon_sym_hex] = ACTIONS(2938),
    [anon_sym_int] = ACTIONS(2938),
    [anon_sym_keys] = ACTIONS(2938),
    [anon_sym_lc] = ACTIONS(2938),
    [anon_sym_lcfirst] = ACTIONS(2938),
    [anon_sym_length] = ACTIONS(2938),
    [anon_sym_localtime] = ACTIONS(2938),
    [anon_sym_log] = ACTIONS(2938),
    [anon_sym_lock] = ACTIONS(2938),
    [anon_sym_lstat] = ACTIONS(2938),
    [anon_sym_oct] = ACTIONS(2938),
    [anon_sym_ord] = ACTIONS(2938),
    [anon_sym_prototype] = ACTIONS(2938),
    [anon_sym_pop] = ACTIONS(2938),
    [anon_sym_pos] = ACTIONS(2938),
    [anon_sym_quotemeta] = ACTIONS(2938),
    [anon_sym_reset] = ACTIONS(2938),
    [anon_sym_rand] = ACTIONS(2938),
    [anon_sym_rmdir] = ACTIONS(2938),
    [anon_sym_readdir] = ACTIONS(2938),
    [anon_sym_readline] = ACTIONS(2938),
    [anon_sym_readpipe] = ACTIONS(2938),
    [anon_sym_rewinddir] = ACTIONS(2938),
    [anon_sym_readlink] = ACTIONS(2938),
    [anon_sym_ref] = ACTIONS(2938),
    [anon_sym_scalar] = ACTIONS(2938),
    [anon_sym_shift] = ACTIONS(2938),
    [anon_sym_sin] = ACTIONS(2938),
    [anon_sym_sleep] = ACTIONS(2938),
    [anon_sym_sqrt] = ACTIONS(2938),
    [anon_sym_srand] = ACTIONS(2938),
    [anon_sym_stat] = ACTIONS(2938),
    [anon_sym_study] = ACTIONS(2938),
    [anon_sym_tell] = ACTIONS(2938),
    [anon_sym_telldir] = ACTIONS(2938),
    [anon_sym_tied] = ACTIONS(2938),
    [anon_sym_uc] = ACTIONS(2938),
    [anon_sym_ucfirst] = ACTIONS(2938),
    [anon_sym_untie] = ACTIONS(2938),
    [anon_sym_umask] = ACTIONS(2938),
    [anon_sym_values] = ACTIONS(2938),
    [anon_sym_write] = ACTIONS(2938),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2938),
    [anon_sym_qq] = ACTIONS(2938),
    [anon_sym_qw] = ACTIONS(2938),
    [anon_sym_qx] = ACTIONS(2938),
    [anon_sym_qr] = ACTIONS(2938),
    [anon_sym_m] = ACTIONS(2938),
    [anon_sym_s] = ACTIONS(2938),
    [anon_sym_tr] = ACTIONS(2938),
    [anon_sym_y] = ACTIONS(2938),
    [anon_sym_if] = ACTIONS(2938),
    [anon_sym_unless] = ACTIONS(2938),
    [anon_sym_while] = ACTIONS(2938),
    [anon_sym_until] = ACTIONS(2938),
    [aux_sym__bareword_token1] = ACTIONS(2938),
    [sym_number] = ACTIONS(2938),
    [anon_sym_true] = ACTIONS(2938),
    [anon_sym_false] = ACTIONS(2938),
    [sym__single_quote] = ACTIONS(2936),
    [sym__double_quote] = ACTIONS(2936),
    [sym__backtick_quote] = ACTIONS(2936),
    [sym__search_slash_quote] = ACTIONS(2936),
    [sym__open_readline_bracket] = ACTIONS(2936),
    [sym__open_fileglob_bracket] = ACTIONS(2936),
    [sym__PERLY_HEREDOC] = ACTIONS(2936),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2936),
    [sym__filetest] = ACTIONS(2936),
  },
  [STATE(1572)] = {
    [sym_heredoc_content] = STATE(1572),
    [sym__identifier] = ACTIONS(2954),
    [anon_sym_LBRACE] = ACTIONS(2954),
    [anon_sym_RBRACE] = ACTIONS(2954),
    [anon_sym_SEMI] = ACTIONS(2954),
    [anon_sym_package] = ACTIONS(2954),
    [anon_sym_class] = ACTIONS(2954),
    [anon_sym_role] = ACTIONS(2954),
    [anon_sym_BUILD] = ACTIONS(2954),
    [anon_sym_ADJUST] = ACTIONS(2954),
    [anon_sym_DOLLAR] = ACTIONS(2954),
    [anon_sym_AT] = ACTIONS(2954),
    [anon_sym_my] = ACTIONS(2954),
    [anon_sym_extended] = ACTIONS(2954),
    [anon_sym_async] = ACTIONS(2954),
    [anon_sym_sub] = ACTIONS(2954),
    [anon_sym_method] = ACTIONS(2954),
    [anon_sym_LPAREN] = ACTIONS(2954),
    [anon_sym_continue] = ACTIONS(2954),
    [anon_sym_state] = ACTIONS(2954),
    [anon_sym_our] = ACTIONS(2954),
    [anon_sym_try] = ACTIONS(2954),
    [anon_sym_defer] = ACTIONS(2954),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2954),
    [anon_sym_LBRACK] = ACTIONS(2954),
    [anon_sym_LT_LT] = ACTIONS(2954),
    [anon_sym_SLASH_SLASH] = ACTIONS(2954),
    [anon_sym_PLUS] = ACTIONS(2954),
    [anon_sym_DASH] = ACTIONS(2954),
    [anon_sym_TILDE] = ACTIONS(2954),
    [anon_sym_BANG] = ACTIONS(2954),
    [anon_sym_PLUS_PLUS] = ACTIONS(2954),
    [anon_sym_DASH_DASH] = ACTIONS(2954),
    [anon_sym_BSLASH] = ACTIONS(2952),
    [anon_sym_do] = ACTIONS(2954),
    [anon_sym_eval] = ACTIONS(2954),
    [anon_sym_await] = ACTIONS(2954),
    [anon_sym_field] = ACTIONS(2954),
    [anon_sym_local] = ACTIONS(2954),
    [anon_sym_dynamically] = ACTIONS(2954),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2954),
    [anon_sym_require] = ACTIONS(2954),
    [anon_sym_map] = ACTIONS(2954),
    [anon_sym_grep] = ACTIONS(2954),
    [anon_sym_sort] = ACTIONS(2954),
    [anon_sym_goto] = ACTIONS(2954),
    [anon_sym_return] = ACTIONS(2954),
    [anon_sym_undef] = ACTIONS(2954),
    [anon_sym_PERCENT2] = ACTIONS(2954),
    [anon_sym_AMP2] = ACTIONS(2954),
    [anon_sym_STAR2] = ACTIONS(2954),
    [anon_sym_use] = ACTIONS(2954),
    [anon_sym_no] = ACTIONS(2954),
    [anon_sym_for] = ACTIONS(2954),
    [anon_sym_foreach] = ACTIONS(2954),
    [anon_sym_last] = ACTIONS(2954),
    [anon_sym_next] = ACTIONS(2954),
    [anon_sym_redo] = ACTIONS(2954),
    [anon_sym_BEGIN] = ACTIONS(2954),
    [anon_sym_INIT] = ACTIONS(2954),
    [anon_sym_CHECK] = ACTIONS(2954),
    [anon_sym_UNITCHECK] = ACTIONS(2954),
    [anon_sym_END] = ACTIONS(2954),
    [anon_sym___FILE__] = ACTIONS(2954),
    [anon_sym___LINE__] = ACTIONS(2954),
    [anon_sym___PACKAGE__] = ACTIONS(2954),
    [anon_sym___SUB__] = ACTIONS(2954),
    [anon_sym_break] = ACTIONS(2954),
    [anon_sym_fork] = ACTIONS(2954),
    [anon_sym_getppid] = ACTIONS(2954),
    [anon_sym_time] = ACTIONS(2954),
    [anon_sym_times] = ACTIONS(2954),
    [anon_sym_wait] = ACTIONS(2954),
    [anon_sym_wantarray] = ACTIONS(2954),
    [anon_sym_abs] = ACTIONS(2954),
    [anon_sym_alarm] = ACTIONS(2954),
    [anon_sym_chop] = ACTIONS(2954),
    [anon_sym_chdir] = ACTIONS(2954),
    [anon_sym_close] = ACTIONS(2954),
    [anon_sym_closedir] = ACTIONS(2954),
    [anon_sym_caller] = ACTIONS(2954),
    [anon_sym_chomp] = ACTIONS(2954),
    [anon_sym_chr] = ACTIONS(2954),
    [anon_sym_cos] = ACTIONS(2954),
    [anon_sym_chroot] = ACTIONS(2954),
    [anon_sym_defined] = ACTIONS(2954),
    [anon_sym_delete] = ACTIONS(2954),
    [anon_sym_dbmclose] = ACTIONS(2954),
    [anon_sym_exists] = ACTIONS(2954),
    [anon_sym_exit] = ACTIONS(2954),
    [anon_sym_eof] = ACTIONS(2954),
    [anon_sym_exp] = ACTIONS(2954),
    [anon_sym_each] = ACTIONS(2954),
    [anon_sym_fc] = ACTIONS(2954),
    [anon_sym_fileno] = ACTIONS(2954),
    [anon_sym_gmtime] = ACTIONS(2954),
    [anon_sym_getc] = ACTIONS(2954),
    [anon_sym_getpgrp] = ACTIONS(2954),
    [anon_sym_getprotobyname] = ACTIONS(2954),
    [anon_sym_getpwname] = ACTIONS(2954),
    [anon_sym_getpwuid] = ACTIONS(2954),
    [anon_sym_getpeername] = ACTIONS(2954),
    [anon_sym_getnetbyname] = ACTIONS(2954),
    [anon_sym_getsockname] = ACTIONS(2954),
    [anon_sym_getgrnam] = ACTIONS(2954),
    [anon_sym_getgrgid] = ACTIONS(2954),
    [anon_sym_hex] = ACTIONS(2954),
    [anon_sym_int] = ACTIONS(2954),
    [anon_sym_keys] = ACTIONS(2954),
    [anon_sym_lc] = ACTIONS(2954),
    [anon_sym_lcfirst] = ACTIONS(2954),
    [anon_sym_length] = ACTIONS(2954),
    [anon_sym_localtime] = ACTIONS(2954),
    [anon_sym_log] = ACTIONS(2954),
    [anon_sym_lock] = ACTIONS(2954),
    [anon_sym_lstat] = ACTIONS(2954),
    [anon_sym_oct] = ACTIONS(2954),
    [anon_sym_ord] = ACTIONS(2954),
    [anon_sym_prototype] = ACTIONS(2954),
    [anon_sym_pop] = ACTIONS(2954),
    [anon_sym_pos] = ACTIONS(2954),
    [anon_sym_quotemeta] = ACTIONS(2954),
    [anon_sym_reset] = ACTIONS(2954),
    [anon_sym_rand] = ACTIONS(2954),
    [anon_sym_rmdir] = ACTIONS(2954),
    [anon_sym_readdir] = ACTIONS(2954),
    [anon_sym_readline] = ACTIONS(2954),
    [anon_sym_readpipe] = ACTIONS(2954),
    [anon_sym_rewinddir] = ACTIONS(2954),
    [anon_sym_readlink] = ACTIONS(2954),
    [anon_sym_ref] = ACTIONS(2954),
    [anon_sym_scalar] = ACTIONS(2954),
    [anon_sym_shift] = ACTIONS(2954),
    [anon_sym_sin] = ACTIONS(2954),
    [anon_sym_sleep] = ACTIONS(2954),
    [anon_sym_sqrt] = ACTIONS(2954),
    [anon_sym_srand] = ACTIONS(2954),
    [anon_sym_stat] = ACTIONS(2954),
    [anon_sym_study] = ACTIONS(2954),
    [anon_sym_tell] = ACTIONS(2954),
    [anon_sym_telldir] = ACTIONS(2954),
    [anon_sym_tied] = ACTIONS(2954),
    [anon_sym_uc] = ACTIONS(2954),
    [anon_sym_ucfirst] = ACTIONS(2954),
    [anon_sym_untie] = ACTIONS(2954),
    [anon_sym_umask] = ACTIONS(2954),
    [anon_sym_values] = ACTIONS(2954),
    [anon_sym_write] = ACTIONS(2954),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2954),
    [anon_sym_qq] = ACTIONS(2954),
    [anon_sym_qw] = ACTIONS(2954),
    [anon_sym_qx] = ACTIONS(2954),
    [anon_sym_qr] = ACTIONS(2954),
    [anon_sym_m] = ACTIONS(2954),
    [anon_sym_s] = ACTIONS(2954),
    [anon_sym_tr] = ACTIONS(2954),
    [anon_sym_y] = ACTIONS(2954),
    [anon_sym_if] = ACTIONS(2954),
    [anon_sym_unless] = ACTIONS(2954),
    [anon_sym_while] = ACTIONS(2954),
    [anon_sym_until] = ACTIONS(2954),
    [aux_sym__bareword_token1] = ACTIONS(2954),
    [sym_number] = ACTIONS(2954),
    [anon_sym_true] = ACTIONS(2954),
    [anon_sym_false] = ACTIONS(2954),
    [sym__single_quote] = ACTIONS(2952),
    [sym__double_quote] = ACTIONS(2952),
    [sym__backtick_quote] = ACTIONS(2952),
    [sym__search_slash_quote] = ACTIONS(2952),
    [sym__open_readline_bracket] = ACTIONS(2952),
    [sym__open_fileglob_bracket] = ACTIONS(2952),
    [sym__PERLY_HEREDOC] = ACTIONS(2952),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2952),
    [sym__filetest] = ACTIONS(2952),
  },
  [STATE(1573)] = {
    [sym_heredoc_content] = STATE(1573),
    [sym__identifier] = ACTIONS(2986),
    [anon_sym_LBRACE] = ACTIONS(2986),
    [anon_sym_RBRACE] = ACTIONS(2986),
    [anon_sym_SEMI] = ACTIONS(2986),
    [anon_sym_package] = ACTIONS(2986),
    [anon_sym_class] = ACTIONS(2986),
    [anon_sym_role] = ACTIONS(2986),
    [anon_sym_BUILD] = ACTIONS(2986),
    [anon_sym_ADJUST] = ACTIONS(2986),
    [anon_sym_DOLLAR] = ACTIONS(2986),
    [anon_sym_AT] = ACTIONS(2986),
    [anon_sym_my] = ACTIONS(2986),
    [anon_sym_extended] = ACTIONS(2986),
    [anon_sym_async] = ACTIONS(2986),
    [anon_sym_sub] = ACTIONS(2986),
    [anon_sym_method] = ACTIONS(2986),
    [anon_sym_LPAREN] = ACTIONS(2986),
    [anon_sym_continue] = ACTIONS(2986),
    [anon_sym_state] = ACTIONS(2986),
    [anon_sym_our] = ACTIONS(2986),
    [anon_sym_try] = ACTIONS(2986),
    [anon_sym_defer] = ACTIONS(2986),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2986),
    [anon_sym_LBRACK] = ACTIONS(2986),
    [anon_sym_LT_LT] = ACTIONS(2986),
    [anon_sym_SLASH_SLASH] = ACTIONS(2986),
    [anon_sym_PLUS] = ACTIONS(2986),
    [anon_sym_DASH] = ACTIONS(2986),
    [anon_sym_TILDE] = ACTIONS(2986),
    [anon_sym_BANG] = ACTIONS(2986),
    [anon_sym_PLUS_PLUS] = ACTIONS(2986),
    [anon_sym_DASH_DASH] = ACTIONS(2986),
    [anon_sym_BSLASH] = ACTIONS(2984),
    [anon_sym_do] = ACTIONS(2986),
    [anon_sym_eval] = ACTIONS(2986),
    [anon_sym_await] = ACTIONS(2986),
    [anon_sym_field] = ACTIONS(2986),
    [anon_sym_local] = ACTIONS(2986),
    [anon_sym_dynamically] = ACTIONS(2986),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2986),
    [anon_sym_require] = ACTIONS(2986),
    [anon_sym_map] = ACTIONS(2986),
    [anon_sym_grep] = ACTIONS(2986),
    [anon_sym_sort] = ACTIONS(2986),
    [anon_sym_goto] = ACTIONS(2986),
    [anon_sym_return] = ACTIONS(2986),
    [anon_sym_undef] = ACTIONS(2986),
    [anon_sym_PERCENT2] = ACTIONS(2986),
    [anon_sym_AMP2] = ACTIONS(2986),
    [anon_sym_STAR2] = ACTIONS(2986),
    [anon_sym_use] = ACTIONS(2986),
    [anon_sym_no] = ACTIONS(2986),
    [anon_sym_for] = ACTIONS(2986),
    [anon_sym_foreach] = ACTIONS(2986),
    [anon_sym_last] = ACTIONS(2986),
    [anon_sym_next] = ACTIONS(2986),
    [anon_sym_redo] = ACTIONS(2986),
    [anon_sym_BEGIN] = ACTIONS(2986),
    [anon_sym_INIT] = ACTIONS(2986),
    [anon_sym_CHECK] = ACTIONS(2986),
    [anon_sym_UNITCHECK] = ACTIONS(2986),
    [anon_sym_END] = ACTIONS(2986),
    [anon_sym___FILE__] = ACTIONS(2986),
    [anon_sym___LINE__] = ACTIONS(2986),
    [anon_sym___PACKAGE__] = ACTIONS(2986),
    [anon_sym___SUB__] = ACTIONS(2986),
    [anon_sym_break] = ACTIONS(2986),
    [anon_sym_fork] = ACTIONS(2986),
    [anon_sym_getppid] = ACTIONS(2986),
    [anon_sym_time] = ACTIONS(2986),
    [anon_sym_times] = ACTIONS(2986),
    [anon_sym_wait] = ACTIONS(2986),
    [anon_sym_wantarray] = ACTIONS(2986),
    [anon_sym_abs] = ACTIONS(2986),
    [anon_sym_alarm] = ACTIONS(2986),
    [anon_sym_chop] = ACTIONS(2986),
    [anon_sym_chdir] = ACTIONS(2986),
    [anon_sym_close] = ACTIONS(2986),
    [anon_sym_closedir] = ACTIONS(2986),
    [anon_sym_caller] = ACTIONS(2986),
    [anon_sym_chomp] = ACTIONS(2986),
    [anon_sym_chr] = ACTIONS(2986),
    [anon_sym_cos] = ACTIONS(2986),
    [anon_sym_chroot] = ACTIONS(2986),
    [anon_sym_defined] = ACTIONS(2986),
    [anon_sym_delete] = ACTIONS(2986),
    [anon_sym_dbmclose] = ACTIONS(2986),
    [anon_sym_exists] = ACTIONS(2986),
    [anon_sym_exit] = ACTIONS(2986),
    [anon_sym_eof] = ACTIONS(2986),
    [anon_sym_exp] = ACTIONS(2986),
    [anon_sym_each] = ACTIONS(2986),
    [anon_sym_fc] = ACTIONS(2986),
    [anon_sym_fileno] = ACTIONS(2986),
    [anon_sym_gmtime] = ACTIONS(2986),
    [anon_sym_getc] = ACTIONS(2986),
    [anon_sym_getpgrp] = ACTIONS(2986),
    [anon_sym_getprotobyname] = ACTIONS(2986),
    [anon_sym_getpwname] = ACTIONS(2986),
    [anon_sym_getpwuid] = ACTIONS(2986),
    [anon_sym_getpeername] = ACTIONS(2986),
    [anon_sym_getnetbyname] = ACTIONS(2986),
    [anon_sym_getsockname] = ACTIONS(2986),
    [anon_sym_getgrnam] = ACTIONS(2986),
    [anon_sym_getgrgid] = ACTIONS(2986),
    [anon_sym_hex] = ACTIONS(2986),
    [anon_sym_int] = ACTIONS(2986),
    [anon_sym_keys] = ACTIONS(2986),
    [anon_sym_lc] = ACTIONS(2986),
    [anon_sym_lcfirst] = ACTIONS(2986),
    [anon_sym_length] = ACTIONS(2986),
    [anon_sym_localtime] = ACTIONS(2986),
    [anon_sym_log] = ACTIONS(2986),
    [anon_sym_lock] = ACTIONS(2986),
    [anon_sym_lstat] = ACTIONS(2986),
    [anon_sym_oct] = ACTIONS(2986),
    [anon_sym_ord] = ACTIONS(2986),
    [anon_sym_prototype] = ACTIONS(2986),
    [anon_sym_pop] = ACTIONS(2986),
    [anon_sym_pos] = ACTIONS(2986),
    [anon_sym_quotemeta] = ACTIONS(2986),
    [anon_sym_reset] = ACTIONS(2986),
    [anon_sym_rand] = ACTIONS(2986),
    [anon_sym_rmdir] = ACTIONS(2986),
    [anon_sym_readdir] = ACTIONS(2986),
    [anon_sym_readline] = ACTIONS(2986),
    [anon_sym_readpipe] = ACTIONS(2986),
    [anon_sym_rewinddir] = ACTIONS(2986),
    [anon_sym_readlink] = ACTIONS(2986),
    [anon_sym_ref] = ACTIONS(2986),
    [anon_sym_scalar] = ACTIONS(2986),
    [anon_sym_shift] = ACTIONS(2986),
    [anon_sym_sin] = ACTIONS(2986),
    [anon_sym_sleep] = ACTIONS(2986),
    [anon_sym_sqrt] = ACTIONS(2986),
    [anon_sym_srand] = ACTIONS(2986),
    [anon_sym_stat] = ACTIONS(2986),
    [anon_sym_study] = ACTIONS(2986),
    [anon_sym_tell] = ACTIONS(2986),
    [anon_sym_telldir] = ACTIONS(2986),
    [anon_sym_tied] = ACTIONS(2986),
    [anon_sym_uc] = ACTIONS(2986),
    [anon_sym_ucfirst] = ACTIONS(2986),
    [anon_sym_untie] = ACTIONS(2986),
    [anon_sym_umask] = ACTIONS(2986),
    [anon_sym_values] = ACTIONS(2986),
    [anon_sym_write] = ACTIONS(2986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2986),
    [anon_sym_qq] = ACTIONS(2986),
    [anon_sym_qw] = ACTIONS(2986),
    [anon_sym_qx] = ACTIONS(2986),
    [anon_sym_qr] = ACTIONS(2986),
    [anon_sym_m] = ACTIONS(2986),
    [anon_sym_s] = ACTIONS(2986),
    [anon_sym_tr] = ACTIONS(2986),
    [anon_sym_y] = ACTIONS(2986),
    [anon_sym_if] = ACTIONS(2986),
    [anon_sym_unless] = ACTIONS(2986),
    [anon_sym_while] = ACTIONS(2986),
    [anon_sym_until] = ACTIONS(2986),
    [aux_sym__bareword_token1] = ACTIONS(2986),
    [sym_number] = ACTIONS(2986),
    [anon_sym_true] = ACTIONS(2986),
    [anon_sym_false] = ACTIONS(2986),
    [sym__single_quote] = ACTIONS(2984),
    [sym__double_quote] = ACTIONS(2984),
    [sym__backtick_quote] = ACTIONS(2984),
    [sym__search_slash_quote] = ACTIONS(2984),
    [sym__open_readline_bracket] = ACTIONS(2984),
    [sym__open_fileglob_bracket] = ACTIONS(2984),
    [sym__PERLY_HEREDOC] = ACTIONS(2984),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2984),
    [sym__filetest] = ACTIONS(2984),
  },
  [STATE(1574)] = {
    [sym_heredoc_content] = STATE(1574),
    [sym__identifier] = ACTIONS(2990),
    [anon_sym_LBRACE] = ACTIONS(2990),
    [anon_sym_RBRACE] = ACTIONS(2990),
    [anon_sym_SEMI] = ACTIONS(2990),
    [anon_sym_package] = ACTIONS(2990),
    [anon_sym_class] = ACTIONS(2990),
    [anon_sym_role] = ACTIONS(2990),
    [anon_sym_BUILD] = ACTIONS(2990),
    [anon_sym_ADJUST] = ACTIONS(2990),
    [anon_sym_DOLLAR] = ACTIONS(2990),
    [anon_sym_AT] = ACTIONS(2990),
    [anon_sym_my] = ACTIONS(2990),
    [anon_sym_extended] = ACTIONS(2990),
    [anon_sym_async] = ACTIONS(2990),
    [anon_sym_sub] = ACTIONS(2990),
    [anon_sym_method] = ACTIONS(2990),
    [anon_sym_LPAREN] = ACTIONS(2990),
    [anon_sym_continue] = ACTIONS(2990),
    [anon_sym_state] = ACTIONS(2990),
    [anon_sym_our] = ACTIONS(2990),
    [anon_sym_try] = ACTIONS(2990),
    [anon_sym_defer] = ACTIONS(2990),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2990),
    [anon_sym_LBRACK] = ACTIONS(2990),
    [anon_sym_LT_LT] = ACTIONS(2990),
    [anon_sym_SLASH_SLASH] = ACTIONS(2990),
    [anon_sym_PLUS] = ACTIONS(2990),
    [anon_sym_DASH] = ACTIONS(2990),
    [anon_sym_TILDE] = ACTIONS(2990),
    [anon_sym_BANG] = ACTIONS(2990),
    [anon_sym_PLUS_PLUS] = ACTIONS(2990),
    [anon_sym_DASH_DASH] = ACTIONS(2990),
    [anon_sym_BSLASH] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2990),
    [anon_sym_eval] = ACTIONS(2990),
    [anon_sym_await] = ACTIONS(2990),
    [anon_sym_field] = ACTIONS(2990),
    [anon_sym_local] = ACTIONS(2990),
    [anon_sym_dynamically] = ACTIONS(2990),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2990),
    [anon_sym_require] = ACTIONS(2990),
    [anon_sym_map] = ACTIONS(2990),
    [anon_sym_grep] = ACTIONS(2990),
    [anon_sym_sort] = ACTIONS(2990),
    [anon_sym_goto] = ACTIONS(2990),
    [anon_sym_return] = ACTIONS(2990),
    [anon_sym_undef] = ACTIONS(2990),
    [anon_sym_PERCENT2] = ACTIONS(2990),
    [anon_sym_AMP2] = ACTIONS(2990),
    [anon_sym_STAR2] = ACTIONS(2990),
    [anon_sym_use] = ACTIONS(2990),
    [anon_sym_no] = ACTIONS(2990),
    [anon_sym_for] = ACTIONS(2990),
    [anon_sym_foreach] = ACTIONS(2990),
    [anon_sym_last] = ACTIONS(2990),
    [anon_sym_next] = ACTIONS(2990),
    [anon_sym_redo] = ACTIONS(2990),
    [anon_sym_BEGIN] = ACTIONS(2990),
    [anon_sym_INIT] = ACTIONS(2990),
    [anon_sym_CHECK] = ACTIONS(2990),
    [anon_sym_UNITCHECK] = ACTIONS(2990),
    [anon_sym_END] = ACTIONS(2990),
    [anon_sym___FILE__] = ACTIONS(2990),
    [anon_sym___LINE__] = ACTIONS(2990),
    [anon_sym___PACKAGE__] = ACTIONS(2990),
    [anon_sym___SUB__] = ACTIONS(2990),
    [anon_sym_break] = ACTIONS(2990),
    [anon_sym_fork] = ACTIONS(2990),
    [anon_sym_getppid] = ACTIONS(2990),
    [anon_sym_time] = ACTIONS(2990),
    [anon_sym_times] = ACTIONS(2990),
    [anon_sym_wait] = ACTIONS(2990),
    [anon_sym_wantarray] = ACTIONS(2990),
    [anon_sym_abs] = ACTIONS(2990),
    [anon_sym_alarm] = ACTIONS(2990),
    [anon_sym_chop] = ACTIONS(2990),
    [anon_sym_chdir] = ACTIONS(2990),
    [anon_sym_close] = ACTIONS(2990),
    [anon_sym_closedir] = ACTIONS(2990),
    [anon_sym_caller] = ACTIONS(2990),
    [anon_sym_chomp] = ACTIONS(2990),
    [anon_sym_chr] = ACTIONS(2990),
    [anon_sym_cos] = ACTIONS(2990),
    [anon_sym_chroot] = ACTIONS(2990),
    [anon_sym_defined] = ACTIONS(2990),
    [anon_sym_delete] = ACTIONS(2990),
    [anon_sym_dbmclose] = ACTIONS(2990),
    [anon_sym_exists] = ACTIONS(2990),
    [anon_sym_exit] = ACTIONS(2990),
    [anon_sym_eof] = ACTIONS(2990),
    [anon_sym_exp] = ACTIONS(2990),
    [anon_sym_each] = ACTIONS(2990),
    [anon_sym_fc] = ACTIONS(2990),
    [anon_sym_fileno] = ACTIONS(2990),
    [anon_sym_gmtime] = ACTIONS(2990),
    [anon_sym_getc] = ACTIONS(2990),
    [anon_sym_getpgrp] = ACTIONS(2990),
    [anon_sym_getprotobyname] = ACTIONS(2990),
    [anon_sym_getpwname] = ACTIONS(2990),
    [anon_sym_getpwuid] = ACTIONS(2990),
    [anon_sym_getpeername] = ACTIONS(2990),
    [anon_sym_getnetbyname] = ACTIONS(2990),
    [anon_sym_getsockname] = ACTIONS(2990),
    [anon_sym_getgrnam] = ACTIONS(2990),
    [anon_sym_getgrgid] = ACTIONS(2990),
    [anon_sym_hex] = ACTIONS(2990),
    [anon_sym_int] = ACTIONS(2990),
    [anon_sym_keys] = ACTIONS(2990),
    [anon_sym_lc] = ACTIONS(2990),
    [anon_sym_lcfirst] = ACTIONS(2990),
    [anon_sym_length] = ACTIONS(2990),
    [anon_sym_localtime] = ACTIONS(2990),
    [anon_sym_log] = ACTIONS(2990),
    [anon_sym_lock] = ACTIONS(2990),
    [anon_sym_lstat] = ACTIONS(2990),
    [anon_sym_oct] = ACTIONS(2990),
    [anon_sym_ord] = ACTIONS(2990),
    [anon_sym_prototype] = ACTIONS(2990),
    [anon_sym_pop] = ACTIONS(2990),
    [anon_sym_pos] = ACTIONS(2990),
    [anon_sym_quotemeta] = ACTIONS(2990),
    [anon_sym_reset] = ACTIONS(2990),
    [anon_sym_rand] = ACTIONS(2990),
    [anon_sym_rmdir] = ACTIONS(2990),
    [anon_sym_readdir] = ACTIONS(2990),
    [anon_sym_readline] = ACTIONS(2990),
    [anon_sym_readpipe] = ACTIONS(2990),
    [anon_sym_rewinddir] = ACTIONS(2990),
    [anon_sym_readlink] = ACTIONS(2990),
    [anon_sym_ref] = ACTIONS(2990),
    [anon_sym_scalar] = ACTIONS(2990),
    [anon_sym_shift] = ACTIONS(2990),
    [anon_sym_sin] = ACTIONS(2990),
    [anon_sym_sleep] = ACTIONS(2990),
    [anon_sym_sqrt] = ACTIONS(2990),
    [anon_sym_srand] = ACTIONS(2990),
    [anon_sym_stat] = ACTIONS(2990),
    [anon_sym_study] = ACTIONS(2990),
    [anon_sym_tell] = ACTIONS(2990),
    [anon_sym_telldir] = ACTIONS(2990),
    [anon_sym_tied] = ACTIONS(2990),
    [anon_sym_uc] = ACTIONS(2990),
    [anon_sym_ucfirst] = ACTIONS(2990),
    [anon_sym_untie] = ACTIONS(2990),
    [anon_sym_umask] = ACTIONS(2990),
    [anon_sym_values] = ACTIONS(2990),
    [anon_sym_write] = ACTIONS(2990),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2990),
    [anon_sym_qq] = ACTIONS(2990),
    [anon_sym_qw] = ACTIONS(2990),
    [anon_sym_qx] = ACTIONS(2990),
    [anon_sym_qr] = ACTIONS(2990),
    [anon_sym_m] = ACTIONS(2990),
    [anon_sym_s] = ACTIONS(2990),
    [anon_sym_tr] = ACTIONS(2990),
    [anon_sym_y] = ACTIONS(2990),
    [anon_sym_if] = ACTIONS(2990),
    [anon_sym_unless] = ACTIONS(2990),
    [anon_sym_while] = ACTIONS(2990),
    [anon_sym_until] = ACTIONS(2990),
    [aux_sym__bareword_token1] = ACTIONS(2990),
    [sym_number] = ACTIONS(2990),
    [anon_sym_true] = ACTIONS(2990),
    [anon_sym_false] = ACTIONS(2990),
    [sym__single_quote] = ACTIONS(2988),
    [sym__double_quote] = ACTIONS(2988),
    [sym__backtick_quote] = ACTIONS(2988),
    [sym__search_slash_quote] = ACTIONS(2988),
    [sym__open_readline_bracket] = ACTIONS(2988),
    [sym__open_fileglob_bracket] = ACTIONS(2988),
    [sym__PERLY_HEREDOC] = ACTIONS(2988),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2988),
    [sym__filetest] = ACTIONS(2988),
  },
  [STATE(1575)] = {
    [sym_heredoc_content] = STATE(1575),
    [sym__identifier] = ACTIONS(3014),
    [anon_sym_LBRACE] = ACTIONS(3014),
    [anon_sym_RBRACE] = ACTIONS(3014),
    [anon_sym_SEMI] = ACTIONS(3014),
    [anon_sym_package] = ACTIONS(3014),
    [anon_sym_class] = ACTIONS(3014),
    [anon_sym_role] = ACTIONS(3014),
    [anon_sym_BUILD] = ACTIONS(3014),
    [anon_sym_ADJUST] = ACTIONS(3014),
    [anon_sym_DOLLAR] = ACTIONS(3014),
    [anon_sym_AT] = ACTIONS(3014),
    [anon_sym_my] = ACTIONS(3014),
    [anon_sym_extended] = ACTIONS(3014),
    [anon_sym_async] = ACTIONS(3014),
    [anon_sym_sub] = ACTIONS(3014),
    [anon_sym_method] = ACTIONS(3014),
    [anon_sym_LPAREN] = ACTIONS(3014),
    [anon_sym_continue] = ACTIONS(3014),
    [anon_sym_state] = ACTIONS(3014),
    [anon_sym_our] = ACTIONS(3014),
    [anon_sym_try] = ACTIONS(3014),
    [anon_sym_defer] = ACTIONS(3014),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3014),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym_LT_LT] = ACTIONS(3014),
    [anon_sym_SLASH_SLASH] = ACTIONS(3014),
    [anon_sym_PLUS] = ACTIONS(3014),
    [anon_sym_DASH] = ACTIONS(3014),
    [anon_sym_TILDE] = ACTIONS(3014),
    [anon_sym_BANG] = ACTIONS(3014),
    [anon_sym_PLUS_PLUS] = ACTIONS(3014),
    [anon_sym_DASH_DASH] = ACTIONS(3014),
    [anon_sym_BSLASH] = ACTIONS(3012),
    [anon_sym_do] = ACTIONS(3014),
    [anon_sym_eval] = ACTIONS(3014),
    [anon_sym_await] = ACTIONS(3014),
    [anon_sym_field] = ACTIONS(3014),
    [anon_sym_local] = ACTIONS(3014),
    [anon_sym_dynamically] = ACTIONS(3014),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3014),
    [anon_sym_require] = ACTIONS(3014),
    [anon_sym_map] = ACTIONS(3014),
    [anon_sym_grep] = ACTIONS(3014),
    [anon_sym_sort] = ACTIONS(3014),
    [anon_sym_goto] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_undef] = ACTIONS(3014),
    [anon_sym_PERCENT2] = ACTIONS(3014),
    [anon_sym_AMP2] = ACTIONS(3014),
    [anon_sym_STAR2] = ACTIONS(3014),
    [anon_sym_use] = ACTIONS(3014),
    [anon_sym_no] = ACTIONS(3014),
    [anon_sym_for] = ACTIONS(3014),
    [anon_sym_foreach] = ACTIONS(3014),
    [anon_sym_last] = ACTIONS(3014),
    [anon_sym_next] = ACTIONS(3014),
    [anon_sym_redo] = ACTIONS(3014),
    [anon_sym_BEGIN] = ACTIONS(3014),
    [anon_sym_INIT] = ACTIONS(3014),
    [anon_sym_CHECK] = ACTIONS(3014),
    [anon_sym_UNITCHECK] = ACTIONS(3014),
    [anon_sym_END] = ACTIONS(3014),
    [anon_sym___FILE__] = ACTIONS(3014),
    [anon_sym___LINE__] = ACTIONS(3014),
    [anon_sym___PACKAGE__] = ACTIONS(3014),
    [anon_sym___SUB__] = ACTIONS(3014),
    [anon_sym_break] = ACTIONS(3014),
    [anon_sym_fork] = ACTIONS(3014),
    [anon_sym_getppid] = ACTIONS(3014),
    [anon_sym_time] = ACTIONS(3014),
    [anon_sym_times] = ACTIONS(3014),
    [anon_sym_wait] = ACTIONS(3014),
    [anon_sym_wantarray] = ACTIONS(3014),
    [anon_sym_abs] = ACTIONS(3014),
    [anon_sym_alarm] = ACTIONS(3014),
    [anon_sym_chop] = ACTIONS(3014),
    [anon_sym_chdir] = ACTIONS(3014),
    [anon_sym_close] = ACTIONS(3014),
    [anon_sym_closedir] = ACTIONS(3014),
    [anon_sym_caller] = ACTIONS(3014),
    [anon_sym_chomp] = ACTIONS(3014),
    [anon_sym_chr] = ACTIONS(3014),
    [anon_sym_cos] = ACTIONS(3014),
    [anon_sym_chroot] = ACTIONS(3014),
    [anon_sym_defined] = ACTIONS(3014),
    [anon_sym_delete] = ACTIONS(3014),
    [anon_sym_dbmclose] = ACTIONS(3014),
    [anon_sym_exists] = ACTIONS(3014),
    [anon_sym_exit] = ACTIONS(3014),
    [anon_sym_eof] = ACTIONS(3014),
    [anon_sym_exp] = ACTIONS(3014),
    [anon_sym_each] = ACTIONS(3014),
    [anon_sym_fc] = ACTIONS(3014),
    [anon_sym_fileno] = ACTIONS(3014),
    [anon_sym_gmtime] = ACTIONS(3014),
    [anon_sym_getc] = ACTIONS(3014),
    [anon_sym_getpgrp] = ACTIONS(3014),
    [anon_sym_getprotobyname] = ACTIONS(3014),
    [anon_sym_getpwname] = ACTIONS(3014),
    [anon_sym_getpwuid] = ACTIONS(3014),
    [anon_sym_getpeername] = ACTIONS(3014),
    [anon_sym_getnetbyname] = ACTIONS(3014),
    [anon_sym_getsockname] = ACTIONS(3014),
    [anon_sym_getgrnam] = ACTIONS(3014),
    [anon_sym_getgrgid] = ACTIONS(3014),
    [anon_sym_hex] = ACTIONS(3014),
    [anon_sym_int] = ACTIONS(3014),
    [anon_sym_keys] = ACTIONS(3014),
    [anon_sym_lc] = ACTIONS(3014),
    [anon_sym_lcfirst] = ACTIONS(3014),
    [anon_sym_length] = ACTIONS(3014),
    [anon_sym_localtime] = ACTIONS(3014),
    [anon_sym_log] = ACTIONS(3014),
    [anon_sym_lock] = ACTIONS(3014),
    [anon_sym_lstat] = ACTIONS(3014),
    [anon_sym_oct] = ACTIONS(3014),
    [anon_sym_ord] = ACTIONS(3014),
    [anon_sym_prototype] = ACTIONS(3014),
    [anon_sym_pop] = ACTIONS(3014),
    [anon_sym_pos] = ACTIONS(3014),
    [anon_sym_quotemeta] = ACTIONS(3014),
    [anon_sym_reset] = ACTIONS(3014),
    [anon_sym_rand] = ACTIONS(3014),
    [anon_sym_rmdir] = ACTIONS(3014),
    [anon_sym_readdir] = ACTIONS(3014),
    [anon_sym_readline] = ACTIONS(3014),
    [anon_sym_readpipe] = ACTIONS(3014),
    [anon_sym_rewinddir] = ACTIONS(3014),
    [anon_sym_readlink] = ACTIONS(3014),
    [anon_sym_ref] = ACTIONS(3014),
    [anon_sym_scalar] = ACTIONS(3014),
    [anon_sym_shift] = ACTIONS(3014),
    [anon_sym_sin] = ACTIONS(3014),
    [anon_sym_sleep] = ACTIONS(3014),
    [anon_sym_sqrt] = ACTIONS(3014),
    [anon_sym_srand] = ACTIONS(3014),
    [anon_sym_stat] = ACTIONS(3014),
    [anon_sym_study] = ACTIONS(3014),
    [anon_sym_tell] = ACTIONS(3014),
    [anon_sym_telldir] = ACTIONS(3014),
    [anon_sym_tied] = ACTIONS(3014),
    [anon_sym_uc] = ACTIONS(3014),
    [anon_sym_ucfirst] = ACTIONS(3014),
    [anon_sym_untie] = ACTIONS(3014),
    [anon_sym_umask] = ACTIONS(3014),
    [anon_sym_values] = ACTIONS(3014),
    [anon_sym_write] = ACTIONS(3014),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3014),
    [anon_sym_qq] = ACTIONS(3014),
    [anon_sym_qw] = ACTIONS(3014),
    [anon_sym_qx] = ACTIONS(3014),
    [anon_sym_qr] = ACTIONS(3014),
    [anon_sym_m] = ACTIONS(3014),
    [anon_sym_s] = ACTIONS(3014),
    [anon_sym_tr] = ACTIONS(3014),
    [anon_sym_y] = ACTIONS(3014),
    [anon_sym_if] = ACTIONS(3014),
    [anon_sym_unless] = ACTIONS(3014),
    [anon_sym_while] = ACTIONS(3014),
    [anon_sym_until] = ACTIONS(3014),
    [aux_sym__bareword_token1] = ACTIONS(3014),
    [sym_number] = ACTIONS(3014),
    [anon_sym_true] = ACTIONS(3014),
    [anon_sym_false] = ACTIONS(3014),
    [sym__single_quote] = ACTIONS(3012),
    [sym__double_quote] = ACTIONS(3012),
    [sym__backtick_quote] = ACTIONS(3012),
    [sym__search_slash_quote] = ACTIONS(3012),
    [sym__open_readline_bracket] = ACTIONS(3012),
    [sym__open_fileglob_bracket] = ACTIONS(3012),
    [sym__PERLY_HEREDOC] = ACTIONS(3012),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3012),
    [sym__filetest] = ACTIONS(3012),
  },
  [STATE(1576)] = {
    [sym_heredoc_content] = STATE(1576),
    [sym__identifier] = ACTIONS(3018),
    [anon_sym_LBRACE] = ACTIONS(3018),
    [anon_sym_RBRACE] = ACTIONS(3018),
    [anon_sym_SEMI] = ACTIONS(3018),
    [anon_sym_package] = ACTIONS(3018),
    [anon_sym_class] = ACTIONS(3018),
    [anon_sym_role] = ACTIONS(3018),
    [anon_sym_BUILD] = ACTIONS(3018),
    [anon_sym_ADJUST] = ACTIONS(3018),
    [anon_sym_DOLLAR] = ACTIONS(3018),
    [anon_sym_AT] = ACTIONS(3018),
    [anon_sym_my] = ACTIONS(3018),
    [anon_sym_extended] = ACTIONS(3018),
    [anon_sym_async] = ACTIONS(3018),
    [anon_sym_sub] = ACTIONS(3018),
    [anon_sym_method] = ACTIONS(3018),
    [anon_sym_LPAREN] = ACTIONS(3018),
    [anon_sym_continue] = ACTIONS(3267),
    [anon_sym_state] = ACTIONS(3018),
    [anon_sym_our] = ACTIONS(3018),
    [anon_sym_try] = ACTIONS(3018),
    [anon_sym_defer] = ACTIONS(3018),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3018),
    [anon_sym_LBRACK] = ACTIONS(3018),
    [anon_sym_LT_LT] = ACTIONS(3018),
    [anon_sym_SLASH_SLASH] = ACTIONS(3018),
    [anon_sym_PLUS] = ACTIONS(3018),
    [anon_sym_DASH] = ACTIONS(3018),
    [anon_sym_TILDE] = ACTIONS(3018),
    [anon_sym_BANG] = ACTIONS(3018),
    [anon_sym_PLUS_PLUS] = ACTIONS(3018),
    [anon_sym_DASH_DASH] = ACTIONS(3018),
    [anon_sym_BSLASH] = ACTIONS(3016),
    [anon_sym_do] = ACTIONS(3018),
    [anon_sym_eval] = ACTIONS(3018),
    [anon_sym_await] = ACTIONS(3018),
    [anon_sym_field] = ACTIONS(3018),
    [anon_sym_local] = ACTIONS(3018),
    [anon_sym_dynamically] = ACTIONS(3018),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3018),
    [anon_sym_require] = ACTIONS(3018),
    [anon_sym_map] = ACTIONS(3018),
    [anon_sym_grep] = ACTIONS(3018),
    [anon_sym_sort] = ACTIONS(3018),
    [anon_sym_goto] = ACTIONS(3018),
    [anon_sym_return] = ACTIONS(3018),
    [anon_sym_undef] = ACTIONS(3018),
    [anon_sym_PERCENT2] = ACTIONS(3018),
    [anon_sym_AMP2] = ACTIONS(3018),
    [anon_sym_STAR2] = ACTIONS(3018),
    [anon_sym_use] = ACTIONS(3018),
    [anon_sym_no] = ACTIONS(3018),
    [anon_sym_for] = ACTIONS(3018),
    [anon_sym_foreach] = ACTIONS(3018),
    [anon_sym_last] = ACTIONS(3018),
    [anon_sym_next] = ACTIONS(3018),
    [anon_sym_redo] = ACTIONS(3018),
    [anon_sym_BEGIN] = ACTIONS(3018),
    [anon_sym_INIT] = ACTIONS(3018),
    [anon_sym_CHECK] = ACTIONS(3018),
    [anon_sym_UNITCHECK] = ACTIONS(3018),
    [anon_sym_END] = ACTIONS(3018),
    [anon_sym___FILE__] = ACTIONS(3018),
    [anon_sym___LINE__] = ACTIONS(3018),
    [anon_sym___PACKAGE__] = ACTIONS(3018),
    [anon_sym___SUB__] = ACTIONS(3018),
    [anon_sym_break] = ACTIONS(3018),
    [anon_sym_fork] = ACTIONS(3018),
    [anon_sym_getppid] = ACTIONS(3018),
    [anon_sym_time] = ACTIONS(3018),
    [anon_sym_times] = ACTIONS(3018),
    [anon_sym_wait] = ACTIONS(3018),
    [anon_sym_wantarray] = ACTIONS(3018),
    [anon_sym_abs] = ACTIONS(3018),
    [anon_sym_alarm] = ACTIONS(3018),
    [anon_sym_chop] = ACTIONS(3018),
    [anon_sym_chdir] = ACTIONS(3018),
    [anon_sym_close] = ACTIONS(3018),
    [anon_sym_closedir] = ACTIONS(3018),
    [anon_sym_caller] = ACTIONS(3018),
    [anon_sym_chomp] = ACTIONS(3018),
    [anon_sym_chr] = ACTIONS(3018),
    [anon_sym_cos] = ACTIONS(3018),
    [anon_sym_chroot] = ACTIONS(3018),
    [anon_sym_defined] = ACTIONS(3018),
    [anon_sym_delete] = ACTIONS(3018),
    [anon_sym_dbmclose] = ACTIONS(3018),
    [anon_sym_exists] = ACTIONS(3018),
    [anon_sym_exit] = ACTIONS(3018),
    [anon_sym_eof] = ACTIONS(3018),
    [anon_sym_exp] = ACTIONS(3018),
    [anon_sym_each] = ACTIONS(3018),
    [anon_sym_fc] = ACTIONS(3018),
    [anon_sym_fileno] = ACTIONS(3018),
    [anon_sym_gmtime] = ACTIONS(3018),
    [anon_sym_getc] = ACTIONS(3018),
    [anon_sym_getpgrp] = ACTIONS(3018),
    [anon_sym_getprotobyname] = ACTIONS(3018),
    [anon_sym_getpwname] = ACTIONS(3018),
    [anon_sym_getpwuid] = ACTIONS(3018),
    [anon_sym_getpeername] = ACTIONS(3018),
    [anon_sym_getnetbyname] = ACTIONS(3018),
    [anon_sym_getsockname] = ACTIONS(3018),
    [anon_sym_getgrnam] = ACTIONS(3018),
    [anon_sym_getgrgid] = ACTIONS(3018),
    [anon_sym_hex] = ACTIONS(3018),
    [anon_sym_int] = ACTIONS(3018),
    [anon_sym_keys] = ACTIONS(3018),
    [anon_sym_lc] = ACTIONS(3018),
    [anon_sym_lcfirst] = ACTIONS(3018),
    [anon_sym_length] = ACTIONS(3018),
    [anon_sym_localtime] = ACTIONS(3018),
    [anon_sym_log] = ACTIONS(3018),
    [anon_sym_lock] = ACTIONS(3018),
    [anon_sym_lstat] = ACTIONS(3018),
    [anon_sym_oct] = ACTIONS(3018),
    [anon_sym_ord] = ACTIONS(3018),
    [anon_sym_prototype] = ACTIONS(3018),
    [anon_sym_pop] = ACTIONS(3018),
    [anon_sym_pos] = ACTIONS(3018),
    [anon_sym_quotemeta] = ACTIONS(3018),
    [anon_sym_reset] = ACTIONS(3018),
    [anon_sym_rand] = ACTIONS(3018),
    [anon_sym_rmdir] = ACTIONS(3018),
    [anon_sym_readdir] = ACTIONS(3018),
    [anon_sym_readline] = ACTIONS(3018),
    [anon_sym_readpipe] = ACTIONS(3018),
    [anon_sym_rewinddir] = ACTIONS(3018),
    [anon_sym_readlink] = ACTIONS(3018),
    [anon_sym_ref] = ACTIONS(3018),
    [anon_sym_scalar] = ACTIONS(3018),
    [anon_sym_shift] = ACTIONS(3018),
    [anon_sym_sin] = ACTIONS(3018),
    [anon_sym_sleep] = ACTIONS(3018),
    [anon_sym_sqrt] = ACTIONS(3018),
    [anon_sym_srand] = ACTIONS(3018),
    [anon_sym_stat] = ACTIONS(3018),
    [anon_sym_study] = ACTIONS(3018),
    [anon_sym_tell] = ACTIONS(3018),
    [anon_sym_telldir] = ACTIONS(3018),
    [anon_sym_tied] = ACTIONS(3018),
    [anon_sym_uc] = ACTIONS(3018),
    [anon_sym_ucfirst] = ACTIONS(3018),
    [anon_sym_untie] = ACTIONS(3018),
    [anon_sym_umask] = ACTIONS(3018),
    [anon_sym_values] = ACTIONS(3018),
    [anon_sym_write] = ACTIONS(3018),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3018),
    [anon_sym_qq] = ACTIONS(3018),
    [anon_sym_qw] = ACTIONS(3018),
    [anon_sym_qx] = ACTIONS(3018),
    [anon_sym_qr] = ACTIONS(3018),
    [anon_sym_m] = ACTIONS(3018),
    [anon_sym_s] = ACTIONS(3018),
    [anon_sym_tr] = ACTIONS(3018),
    [anon_sym_y] = ACTIONS(3018),
    [anon_sym_if] = ACTIONS(3018),
    [anon_sym_unless] = ACTIONS(3018),
    [anon_sym_while] = ACTIONS(3018),
    [anon_sym_until] = ACTIONS(3018),
    [aux_sym__bareword_token1] = ACTIONS(3018),
    [sym_number] = ACTIONS(3018),
    [anon_sym_true] = ACTIONS(3018),
    [anon_sym_false] = ACTIONS(3018),
    [sym__single_quote] = ACTIONS(3016),
    [sym__double_quote] = ACTIONS(3016),
    [sym__backtick_quote] = ACTIONS(3016),
    [sym__search_slash_quote] = ACTIONS(3016),
    [sym__open_readline_bracket] = ACTIONS(3016),
    [sym__open_fileglob_bracket] = ACTIONS(3016),
    [sym__PERLY_HEREDOC] = ACTIONS(3016),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3016),
    [sym__filetest] = ACTIONS(3016),
  },
  [STATE(1577)] = {
    [sym_heredoc_content] = STATE(1577),
    [sym__identifier] = ACTIONS(3089),
    [anon_sym_LBRACE] = ACTIONS(3089),
    [anon_sym_RBRACE] = ACTIONS(3089),
    [anon_sym_SEMI] = ACTIONS(3089),
    [anon_sym_package] = ACTIONS(3089),
    [anon_sym_class] = ACTIONS(3089),
    [anon_sym_role] = ACTIONS(3089),
    [anon_sym_BUILD] = ACTIONS(3089),
    [anon_sym_ADJUST] = ACTIONS(3089),
    [anon_sym_DOLLAR] = ACTIONS(3089),
    [anon_sym_AT] = ACTIONS(3089),
    [anon_sym_my] = ACTIONS(3089),
    [anon_sym_extended] = ACTIONS(3089),
    [anon_sym_async] = ACTIONS(3089),
    [anon_sym_sub] = ACTIONS(3089),
    [anon_sym_method] = ACTIONS(3089),
    [anon_sym_LPAREN] = ACTIONS(3089),
    [anon_sym_continue] = ACTIONS(3089),
    [anon_sym_state] = ACTIONS(3089),
    [anon_sym_our] = ACTIONS(3089),
    [anon_sym_try] = ACTIONS(3089),
    [anon_sym_defer] = ACTIONS(3089),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3089),
    [anon_sym_LBRACK] = ACTIONS(3089),
    [anon_sym_LT_LT] = ACTIONS(3089),
    [anon_sym_SLASH_SLASH] = ACTIONS(3089),
    [anon_sym_PLUS] = ACTIONS(3089),
    [anon_sym_DASH] = ACTIONS(3089),
    [anon_sym_TILDE] = ACTIONS(3089),
    [anon_sym_BANG] = ACTIONS(3089),
    [anon_sym_PLUS_PLUS] = ACTIONS(3089),
    [anon_sym_DASH_DASH] = ACTIONS(3089),
    [anon_sym_BSLASH] = ACTIONS(3087),
    [anon_sym_do] = ACTIONS(3089),
    [anon_sym_eval] = ACTIONS(3089),
    [anon_sym_await] = ACTIONS(3089),
    [anon_sym_field] = ACTIONS(3089),
    [anon_sym_local] = ACTIONS(3089),
    [anon_sym_dynamically] = ACTIONS(3089),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3089),
    [anon_sym_require] = ACTIONS(3089),
    [anon_sym_map] = ACTIONS(3089),
    [anon_sym_grep] = ACTIONS(3089),
    [anon_sym_sort] = ACTIONS(3089),
    [anon_sym_goto] = ACTIONS(3089),
    [anon_sym_return] = ACTIONS(3089),
    [anon_sym_undef] = ACTIONS(3089),
    [anon_sym_PERCENT2] = ACTIONS(3089),
    [anon_sym_AMP2] = ACTIONS(3089),
    [anon_sym_STAR2] = ACTIONS(3089),
    [anon_sym_use] = ACTIONS(3089),
    [anon_sym_no] = ACTIONS(3089),
    [anon_sym_for] = ACTIONS(3089),
    [anon_sym_foreach] = ACTIONS(3089),
    [anon_sym_last] = ACTIONS(3089),
    [anon_sym_next] = ACTIONS(3089),
    [anon_sym_redo] = ACTIONS(3089),
    [anon_sym_BEGIN] = ACTIONS(3089),
    [anon_sym_INIT] = ACTIONS(3089),
    [anon_sym_CHECK] = ACTIONS(3089),
    [anon_sym_UNITCHECK] = ACTIONS(3089),
    [anon_sym_END] = ACTIONS(3089),
    [anon_sym___FILE__] = ACTIONS(3089),
    [anon_sym___LINE__] = ACTIONS(3089),
    [anon_sym___PACKAGE__] = ACTIONS(3089),
    [anon_sym___SUB__] = ACTIONS(3089),
    [anon_sym_break] = ACTIONS(3089),
    [anon_sym_fork] = ACTIONS(3089),
    [anon_sym_getppid] = ACTIONS(3089),
    [anon_sym_time] = ACTIONS(3089),
    [anon_sym_times] = ACTIONS(3089),
    [anon_sym_wait] = ACTIONS(3089),
    [anon_sym_wantarray] = ACTIONS(3089),
    [anon_sym_abs] = ACTIONS(3089),
    [anon_sym_alarm] = ACTIONS(3089),
    [anon_sym_chop] = ACTIONS(3089),
    [anon_sym_chdir] = ACTIONS(3089),
    [anon_sym_close] = ACTIONS(3089),
    [anon_sym_closedir] = ACTIONS(3089),
    [anon_sym_caller] = ACTIONS(3089),
    [anon_sym_chomp] = ACTIONS(3089),
    [anon_sym_chr] = ACTIONS(3089),
    [anon_sym_cos] = ACTIONS(3089),
    [anon_sym_chroot] = ACTIONS(3089),
    [anon_sym_defined] = ACTIONS(3089),
    [anon_sym_delete] = ACTIONS(3089),
    [anon_sym_dbmclose] = ACTIONS(3089),
    [anon_sym_exists] = ACTIONS(3089),
    [anon_sym_exit] = ACTIONS(3089),
    [anon_sym_eof] = ACTIONS(3089),
    [anon_sym_exp] = ACTIONS(3089),
    [anon_sym_each] = ACTIONS(3089),
    [anon_sym_fc] = ACTIONS(3089),
    [anon_sym_fileno] = ACTIONS(3089),
    [anon_sym_gmtime] = ACTIONS(3089),
    [anon_sym_getc] = ACTIONS(3089),
    [anon_sym_getpgrp] = ACTIONS(3089),
    [anon_sym_getprotobyname] = ACTIONS(3089),
    [anon_sym_getpwname] = ACTIONS(3089),
    [anon_sym_getpwuid] = ACTIONS(3089),
    [anon_sym_getpeername] = ACTIONS(3089),
    [anon_sym_getnetbyname] = ACTIONS(3089),
    [anon_sym_getsockname] = ACTIONS(3089),
    [anon_sym_getgrnam] = ACTIONS(3089),
    [anon_sym_getgrgid] = ACTIONS(3089),
    [anon_sym_hex] = ACTIONS(3089),
    [anon_sym_int] = ACTIONS(3089),
    [anon_sym_keys] = ACTIONS(3089),
    [anon_sym_lc] = ACTIONS(3089),
    [anon_sym_lcfirst] = ACTIONS(3089),
    [anon_sym_length] = ACTIONS(3089),
    [anon_sym_localtime] = ACTIONS(3089),
    [anon_sym_log] = ACTIONS(3089),
    [anon_sym_lock] = ACTIONS(3089),
    [anon_sym_lstat] = ACTIONS(3089),
    [anon_sym_oct] = ACTIONS(3089),
    [anon_sym_ord] = ACTIONS(3089),
    [anon_sym_prototype] = ACTIONS(3089),
    [anon_sym_pop] = ACTIONS(3089),
    [anon_sym_pos] = ACTIONS(3089),
    [anon_sym_quotemeta] = ACTIONS(3089),
    [anon_sym_reset] = ACTIONS(3089),
    [anon_sym_rand] = ACTIONS(3089),
    [anon_sym_rmdir] = ACTIONS(3089),
    [anon_sym_readdir] = ACTIONS(3089),
    [anon_sym_readline] = ACTIONS(3089),
    [anon_sym_readpipe] = ACTIONS(3089),
    [anon_sym_rewinddir] = ACTIONS(3089),
    [anon_sym_readlink] = ACTIONS(3089),
    [anon_sym_ref] = ACTIONS(3089),
    [anon_sym_scalar] = ACTIONS(3089),
    [anon_sym_shift] = ACTIONS(3089),
    [anon_sym_sin] = ACTIONS(3089),
    [anon_sym_sleep] = ACTIONS(3089),
    [anon_sym_sqrt] = ACTIONS(3089),
    [anon_sym_srand] = ACTIONS(3089),
    [anon_sym_stat] = ACTIONS(3089),
    [anon_sym_study] = ACTIONS(3089),
    [anon_sym_tell] = ACTIONS(3089),
    [anon_sym_telldir] = ACTIONS(3089),
    [anon_sym_tied] = ACTIONS(3089),
    [anon_sym_uc] = ACTIONS(3089),
    [anon_sym_ucfirst] = ACTIONS(3089),
    [anon_sym_untie] = ACTIONS(3089),
    [anon_sym_umask] = ACTIONS(3089),
    [anon_sym_values] = ACTIONS(3089),
    [anon_sym_write] = ACTIONS(3089),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3089),
    [anon_sym_qq] = ACTIONS(3089),
    [anon_sym_qw] = ACTIONS(3089),
    [anon_sym_qx] = ACTIONS(3089),
    [anon_sym_qr] = ACTIONS(3089),
    [anon_sym_m] = ACTIONS(3089),
    [anon_sym_s] = ACTIONS(3089),
    [anon_sym_tr] = ACTIONS(3089),
    [anon_sym_y] = ACTIONS(3089),
    [anon_sym_if] = ACTIONS(3089),
    [anon_sym_unless] = ACTIONS(3089),
    [anon_sym_while] = ACTIONS(3089),
    [anon_sym_until] = ACTIONS(3089),
    [aux_sym__bareword_token1] = ACTIONS(3089),
    [sym_number] = ACTIONS(3089),
    [anon_sym_true] = ACTIONS(3089),
    [anon_sym_false] = ACTIONS(3089),
    [sym__single_quote] = ACTIONS(3087),
    [sym__double_quote] = ACTIONS(3087),
    [sym__backtick_quote] = ACTIONS(3087),
    [sym__search_slash_quote] = ACTIONS(3087),
    [sym__open_readline_bracket] = ACTIONS(3087),
    [sym__open_fileglob_bracket] = ACTIONS(3087),
    [sym__PERLY_HEREDOC] = ACTIONS(3087),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3087),
    [sym__filetest] = ACTIONS(3087),
  },
  [STATE(1578)] = {
    [sym_heredoc_content] = STATE(1578),
    [sym__identifier] = ACTIONS(3093),
    [anon_sym_LBRACE] = ACTIONS(3093),
    [anon_sym_RBRACE] = ACTIONS(3093),
    [anon_sym_SEMI] = ACTIONS(3093),
    [anon_sym_package] = ACTIONS(3093),
    [anon_sym_class] = ACTIONS(3093),
    [anon_sym_role] = ACTIONS(3093),
    [anon_sym_BUILD] = ACTIONS(3093),
    [anon_sym_ADJUST] = ACTIONS(3093),
    [anon_sym_DOLLAR] = ACTIONS(3093),
    [anon_sym_AT] = ACTIONS(3093),
    [anon_sym_my] = ACTIONS(3093),
    [anon_sym_extended] = ACTIONS(3093),
    [anon_sym_async] = ACTIONS(3093),
    [anon_sym_sub] = ACTIONS(3093),
    [anon_sym_method] = ACTIONS(3093),
    [anon_sym_LPAREN] = ACTIONS(3093),
    [anon_sym_continue] = ACTIONS(3093),
    [anon_sym_state] = ACTIONS(3093),
    [anon_sym_our] = ACTIONS(3093),
    [anon_sym_try] = ACTIONS(3093),
    [anon_sym_defer] = ACTIONS(3093),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3093),
    [anon_sym_LBRACK] = ACTIONS(3093),
    [anon_sym_LT_LT] = ACTIONS(3093),
    [anon_sym_SLASH_SLASH] = ACTIONS(3093),
    [anon_sym_PLUS] = ACTIONS(3093),
    [anon_sym_DASH] = ACTIONS(3093),
    [anon_sym_TILDE] = ACTIONS(3093),
    [anon_sym_BANG] = ACTIONS(3093),
    [anon_sym_PLUS_PLUS] = ACTIONS(3093),
    [anon_sym_DASH_DASH] = ACTIONS(3093),
    [anon_sym_BSLASH] = ACTIONS(3091),
    [anon_sym_do] = ACTIONS(3093),
    [anon_sym_eval] = ACTIONS(3093),
    [anon_sym_await] = ACTIONS(3093),
    [anon_sym_field] = ACTIONS(3093),
    [anon_sym_local] = ACTIONS(3093),
    [anon_sym_dynamically] = ACTIONS(3093),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3093),
    [anon_sym_require] = ACTIONS(3093),
    [anon_sym_map] = ACTIONS(3093),
    [anon_sym_grep] = ACTIONS(3093),
    [anon_sym_sort] = ACTIONS(3093),
    [anon_sym_goto] = ACTIONS(3093),
    [anon_sym_return] = ACTIONS(3093),
    [anon_sym_undef] = ACTIONS(3093),
    [anon_sym_PERCENT2] = ACTIONS(3093),
    [anon_sym_AMP2] = ACTIONS(3093),
    [anon_sym_STAR2] = ACTIONS(3093),
    [anon_sym_use] = ACTIONS(3093),
    [anon_sym_no] = ACTIONS(3093),
    [anon_sym_for] = ACTIONS(3093),
    [anon_sym_foreach] = ACTIONS(3093),
    [anon_sym_last] = ACTIONS(3093),
    [anon_sym_next] = ACTIONS(3093),
    [anon_sym_redo] = ACTIONS(3093),
    [anon_sym_BEGIN] = ACTIONS(3093),
    [anon_sym_INIT] = ACTIONS(3093),
    [anon_sym_CHECK] = ACTIONS(3093),
    [anon_sym_UNITCHECK] = ACTIONS(3093),
    [anon_sym_END] = ACTIONS(3093),
    [anon_sym___FILE__] = ACTIONS(3093),
    [anon_sym___LINE__] = ACTIONS(3093),
    [anon_sym___PACKAGE__] = ACTIONS(3093),
    [anon_sym___SUB__] = ACTIONS(3093),
    [anon_sym_break] = ACTIONS(3093),
    [anon_sym_fork] = ACTIONS(3093),
    [anon_sym_getppid] = ACTIONS(3093),
    [anon_sym_time] = ACTIONS(3093),
    [anon_sym_times] = ACTIONS(3093),
    [anon_sym_wait] = ACTIONS(3093),
    [anon_sym_wantarray] = ACTIONS(3093),
    [anon_sym_abs] = ACTIONS(3093),
    [anon_sym_alarm] = ACTIONS(3093),
    [anon_sym_chop] = ACTIONS(3093),
    [anon_sym_chdir] = ACTIONS(3093),
    [anon_sym_close] = ACTIONS(3093),
    [anon_sym_closedir] = ACTIONS(3093),
    [anon_sym_caller] = ACTIONS(3093),
    [anon_sym_chomp] = ACTIONS(3093),
    [anon_sym_chr] = ACTIONS(3093),
    [anon_sym_cos] = ACTIONS(3093),
    [anon_sym_chroot] = ACTIONS(3093),
    [anon_sym_defined] = ACTIONS(3093),
    [anon_sym_delete] = ACTIONS(3093),
    [anon_sym_dbmclose] = ACTIONS(3093),
    [anon_sym_exists] = ACTIONS(3093),
    [anon_sym_exit] = ACTIONS(3093),
    [anon_sym_eof] = ACTIONS(3093),
    [anon_sym_exp] = ACTIONS(3093),
    [anon_sym_each] = ACTIONS(3093),
    [anon_sym_fc] = ACTIONS(3093),
    [anon_sym_fileno] = ACTIONS(3093),
    [anon_sym_gmtime] = ACTIONS(3093),
    [anon_sym_getc] = ACTIONS(3093),
    [anon_sym_getpgrp] = ACTIONS(3093),
    [anon_sym_getprotobyname] = ACTIONS(3093),
    [anon_sym_getpwname] = ACTIONS(3093),
    [anon_sym_getpwuid] = ACTIONS(3093),
    [anon_sym_getpeername] = ACTIONS(3093),
    [anon_sym_getnetbyname] = ACTIONS(3093),
    [anon_sym_getsockname] = ACTIONS(3093),
    [anon_sym_getgrnam] = ACTIONS(3093),
    [anon_sym_getgrgid] = ACTIONS(3093),
    [anon_sym_hex] = ACTIONS(3093),
    [anon_sym_int] = ACTIONS(3093),
    [anon_sym_keys] = ACTIONS(3093),
    [anon_sym_lc] = ACTIONS(3093),
    [anon_sym_lcfirst] = ACTIONS(3093),
    [anon_sym_length] = ACTIONS(3093),
    [anon_sym_localtime] = ACTIONS(3093),
    [anon_sym_log] = ACTIONS(3093),
    [anon_sym_lock] = ACTIONS(3093),
    [anon_sym_lstat] = ACTIONS(3093),
    [anon_sym_oct] = ACTIONS(3093),
    [anon_sym_ord] = ACTIONS(3093),
    [anon_sym_prototype] = ACTIONS(3093),
    [anon_sym_pop] = ACTIONS(3093),
    [anon_sym_pos] = ACTIONS(3093),
    [anon_sym_quotemeta] = ACTIONS(3093),
    [anon_sym_reset] = ACTIONS(3093),
    [anon_sym_rand] = ACTIONS(3093),
    [anon_sym_rmdir] = ACTIONS(3093),
    [anon_sym_readdir] = ACTIONS(3093),
    [anon_sym_readline] = ACTIONS(3093),
    [anon_sym_readpipe] = ACTIONS(3093),
    [anon_sym_rewinddir] = ACTIONS(3093),
    [anon_sym_readlink] = ACTIONS(3093),
    [anon_sym_ref] = ACTIONS(3093),
    [anon_sym_scalar] = ACTIONS(3093),
    [anon_sym_shift] = ACTIONS(3093),
    [anon_sym_sin] = ACTIONS(3093),
    [anon_sym_sleep] = ACTIONS(3093),
    [anon_sym_sqrt] = ACTIONS(3093),
    [anon_sym_srand] = ACTIONS(3093),
    [anon_sym_stat] = ACTIONS(3093),
    [anon_sym_study] = ACTIONS(3093),
    [anon_sym_tell] = ACTIONS(3093),
    [anon_sym_telldir] = ACTIONS(3093),
    [anon_sym_tied] = ACTIONS(3093),
    [anon_sym_uc] = ACTIONS(3093),
    [anon_sym_ucfirst] = ACTIONS(3093),
    [anon_sym_untie] = ACTIONS(3093),
    [anon_sym_umask] = ACTIONS(3093),
    [anon_sym_values] = ACTIONS(3093),
    [anon_sym_write] = ACTIONS(3093),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3093),
    [anon_sym_qq] = ACTIONS(3093),
    [anon_sym_qw] = ACTIONS(3093),
    [anon_sym_qx] = ACTIONS(3093),
    [anon_sym_qr] = ACTIONS(3093),
    [anon_sym_m] = ACTIONS(3093),
    [anon_sym_s] = ACTIONS(3093),
    [anon_sym_tr] = ACTIONS(3093),
    [anon_sym_y] = ACTIONS(3093),
    [anon_sym_if] = ACTIONS(3093),
    [anon_sym_unless] = ACTIONS(3093),
    [anon_sym_while] = ACTIONS(3093),
    [anon_sym_until] = ACTIONS(3093),
    [aux_sym__bareword_token1] = ACTIONS(3093),
    [sym_number] = ACTIONS(3093),
    [anon_sym_true] = ACTIONS(3093),
    [anon_sym_false] = ACTIONS(3093),
    [sym__single_quote] = ACTIONS(3091),
    [sym__double_quote] = ACTIONS(3091),
    [sym__backtick_quote] = ACTIONS(3091),
    [sym__search_slash_quote] = ACTIONS(3091),
    [sym__open_readline_bracket] = ACTIONS(3091),
    [sym__open_fileglob_bracket] = ACTIONS(3091),
    [sym__PERLY_HEREDOC] = ACTIONS(3091),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3091),
    [sym__filetest] = ACTIONS(3091),
  },
  [STATE(1579)] = {
    [sym_heredoc_content] = STATE(1579),
    [sym__identifier] = ACTIONS(3101),
    [anon_sym_LBRACE] = ACTIONS(3101),
    [anon_sym_RBRACE] = ACTIONS(3101),
    [anon_sym_SEMI] = ACTIONS(3101),
    [anon_sym_package] = ACTIONS(3101),
    [anon_sym_class] = ACTIONS(3101),
    [anon_sym_role] = ACTIONS(3101),
    [anon_sym_BUILD] = ACTIONS(3101),
    [anon_sym_ADJUST] = ACTIONS(3101),
    [anon_sym_DOLLAR] = ACTIONS(3101),
    [anon_sym_AT] = ACTIONS(3101),
    [anon_sym_my] = ACTIONS(3101),
    [anon_sym_extended] = ACTIONS(3101),
    [anon_sym_async] = ACTIONS(3101),
    [anon_sym_sub] = ACTIONS(3101),
    [anon_sym_method] = ACTIONS(3101),
    [anon_sym_LPAREN] = ACTIONS(3101),
    [anon_sym_continue] = ACTIONS(3101),
    [anon_sym_state] = ACTIONS(3101),
    [anon_sym_our] = ACTIONS(3101),
    [anon_sym_try] = ACTIONS(3101),
    [anon_sym_defer] = ACTIONS(3101),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3101),
    [anon_sym_LBRACK] = ACTIONS(3101),
    [anon_sym_LT_LT] = ACTIONS(3101),
    [anon_sym_SLASH_SLASH] = ACTIONS(3101),
    [anon_sym_PLUS] = ACTIONS(3101),
    [anon_sym_DASH] = ACTIONS(3101),
    [anon_sym_TILDE] = ACTIONS(3101),
    [anon_sym_BANG] = ACTIONS(3101),
    [anon_sym_PLUS_PLUS] = ACTIONS(3101),
    [anon_sym_DASH_DASH] = ACTIONS(3101),
    [anon_sym_BSLASH] = ACTIONS(3099),
    [anon_sym_do] = ACTIONS(3101),
    [anon_sym_eval] = ACTIONS(3101),
    [anon_sym_await] = ACTIONS(3101),
    [anon_sym_field] = ACTIONS(3101),
    [anon_sym_local] = ACTIONS(3101),
    [anon_sym_dynamically] = ACTIONS(3101),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3101),
    [anon_sym_require] = ACTIONS(3101),
    [anon_sym_map] = ACTIONS(3101),
    [anon_sym_grep] = ACTIONS(3101),
    [anon_sym_sort] = ACTIONS(3101),
    [anon_sym_goto] = ACTIONS(3101),
    [anon_sym_return] = ACTIONS(3101),
    [anon_sym_undef] = ACTIONS(3101),
    [anon_sym_PERCENT2] = ACTIONS(3101),
    [anon_sym_AMP2] = ACTIONS(3101),
    [anon_sym_STAR2] = ACTIONS(3101),
    [anon_sym_use] = ACTIONS(3101),
    [anon_sym_no] = ACTIONS(3101),
    [anon_sym_for] = ACTIONS(3101),
    [anon_sym_foreach] = ACTIONS(3101),
    [anon_sym_last] = ACTIONS(3101),
    [anon_sym_next] = ACTIONS(3101),
    [anon_sym_redo] = ACTIONS(3101),
    [anon_sym_BEGIN] = ACTIONS(3101),
    [anon_sym_INIT] = ACTIONS(3101),
    [anon_sym_CHECK] = ACTIONS(3101),
    [anon_sym_UNITCHECK] = ACTIONS(3101),
    [anon_sym_END] = ACTIONS(3101),
    [anon_sym___FILE__] = ACTIONS(3101),
    [anon_sym___LINE__] = ACTIONS(3101),
    [anon_sym___PACKAGE__] = ACTIONS(3101),
    [anon_sym___SUB__] = ACTIONS(3101),
    [anon_sym_break] = ACTIONS(3101),
    [anon_sym_fork] = ACTIONS(3101),
    [anon_sym_getppid] = ACTIONS(3101),
    [anon_sym_time] = ACTIONS(3101),
    [anon_sym_times] = ACTIONS(3101),
    [anon_sym_wait] = ACTIONS(3101),
    [anon_sym_wantarray] = ACTIONS(3101),
    [anon_sym_abs] = ACTIONS(3101),
    [anon_sym_alarm] = ACTIONS(3101),
    [anon_sym_chop] = ACTIONS(3101),
    [anon_sym_chdir] = ACTIONS(3101),
    [anon_sym_close] = ACTIONS(3101),
    [anon_sym_closedir] = ACTIONS(3101),
    [anon_sym_caller] = ACTIONS(3101),
    [anon_sym_chomp] = ACTIONS(3101),
    [anon_sym_chr] = ACTIONS(3101),
    [anon_sym_cos] = ACTIONS(3101),
    [anon_sym_chroot] = ACTIONS(3101),
    [anon_sym_defined] = ACTIONS(3101),
    [anon_sym_delete] = ACTIONS(3101),
    [anon_sym_dbmclose] = ACTIONS(3101),
    [anon_sym_exists] = ACTIONS(3101),
    [anon_sym_exit] = ACTIONS(3101),
    [anon_sym_eof] = ACTIONS(3101),
    [anon_sym_exp] = ACTIONS(3101),
    [anon_sym_each] = ACTIONS(3101),
    [anon_sym_fc] = ACTIONS(3101),
    [anon_sym_fileno] = ACTIONS(3101),
    [anon_sym_gmtime] = ACTIONS(3101),
    [anon_sym_getc] = ACTIONS(3101),
    [anon_sym_getpgrp] = ACTIONS(3101),
    [anon_sym_getprotobyname] = ACTIONS(3101),
    [anon_sym_getpwname] = ACTIONS(3101),
    [anon_sym_getpwuid] = ACTIONS(3101),
    [anon_sym_getpeername] = ACTIONS(3101),
    [anon_sym_getnetbyname] = ACTIONS(3101),
    [anon_sym_getsockname] = ACTIONS(3101),
    [anon_sym_getgrnam] = ACTIONS(3101),
    [anon_sym_getgrgid] = ACTIONS(3101),
    [anon_sym_hex] = ACTIONS(3101),
    [anon_sym_int] = ACTIONS(3101),
    [anon_sym_keys] = ACTIONS(3101),
    [anon_sym_lc] = ACTIONS(3101),
    [anon_sym_lcfirst] = ACTIONS(3101),
    [anon_sym_length] = ACTIONS(3101),
    [anon_sym_localtime] = ACTIONS(3101),
    [anon_sym_log] = ACTIONS(3101),
    [anon_sym_lock] = ACTIONS(3101),
    [anon_sym_lstat] = ACTIONS(3101),
    [anon_sym_oct] = ACTIONS(3101),
    [anon_sym_ord] = ACTIONS(3101),
    [anon_sym_prototype] = ACTIONS(3101),
    [anon_sym_pop] = ACTIONS(3101),
    [anon_sym_pos] = ACTIONS(3101),
    [anon_sym_quotemeta] = ACTIONS(3101),
    [anon_sym_reset] = ACTIONS(3101),
    [anon_sym_rand] = ACTIONS(3101),
    [anon_sym_rmdir] = ACTIONS(3101),
    [anon_sym_readdir] = ACTIONS(3101),
    [anon_sym_readline] = ACTIONS(3101),
    [anon_sym_readpipe] = ACTIONS(3101),
    [anon_sym_rewinddir] = ACTIONS(3101),
    [anon_sym_readlink] = ACTIONS(3101),
    [anon_sym_ref] = ACTIONS(3101),
    [anon_sym_scalar] = ACTIONS(3101),
    [anon_sym_shift] = ACTIONS(3101),
    [anon_sym_sin] = ACTIONS(3101),
    [anon_sym_sleep] = ACTIONS(3101),
    [anon_sym_sqrt] = ACTIONS(3101),
    [anon_sym_srand] = ACTIONS(3101),
    [anon_sym_stat] = ACTIONS(3101),
    [anon_sym_study] = ACTIONS(3101),
    [anon_sym_tell] = ACTIONS(3101),
    [anon_sym_telldir] = ACTIONS(3101),
    [anon_sym_tied] = ACTIONS(3101),
    [anon_sym_uc] = ACTIONS(3101),
    [anon_sym_ucfirst] = ACTIONS(3101),
    [anon_sym_untie] = ACTIONS(3101),
    [anon_sym_umask] = ACTIONS(3101),
    [anon_sym_values] = ACTIONS(3101),
    [anon_sym_write] = ACTIONS(3101),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3101),
    [anon_sym_qq] = ACTIONS(3101),
    [anon_sym_qw] = ACTIONS(3101),
    [anon_sym_qx] = ACTIONS(3101),
    [anon_sym_qr] = ACTIONS(3101),
    [anon_sym_m] = ACTIONS(3101),
    [anon_sym_s] = ACTIONS(3101),
    [anon_sym_tr] = ACTIONS(3101),
    [anon_sym_y] = ACTIONS(3101),
    [anon_sym_if] = ACTIONS(3101),
    [anon_sym_unless] = ACTIONS(3101),
    [anon_sym_while] = ACTIONS(3101),
    [anon_sym_until] = ACTIONS(3101),
    [aux_sym__bareword_token1] = ACTIONS(3101),
    [sym_number] = ACTIONS(3101),
    [anon_sym_true] = ACTIONS(3101),
    [anon_sym_false] = ACTIONS(3101),
    [sym__single_quote] = ACTIONS(3099),
    [sym__double_quote] = ACTIONS(3099),
    [sym__backtick_quote] = ACTIONS(3099),
    [sym__search_slash_quote] = ACTIONS(3099),
    [sym__open_readline_bracket] = ACTIONS(3099),
    [sym__open_fileglob_bracket] = ACTIONS(3099),
    [sym__PERLY_HEREDOC] = ACTIONS(3099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3099),
    [sym__filetest] = ACTIONS(3099),
  },
  [STATE(1580)] = {
    [sym_heredoc_content] = STATE(1580),
    [sym__identifier] = ACTIONS(3105),
    [anon_sym_LBRACE] = ACTIONS(3105),
    [anon_sym_RBRACE] = ACTIONS(3105),
    [anon_sym_SEMI] = ACTIONS(3105),
    [anon_sym_package] = ACTIONS(3105),
    [anon_sym_class] = ACTIONS(3105),
    [anon_sym_role] = ACTIONS(3105),
    [anon_sym_BUILD] = ACTIONS(3105),
    [anon_sym_ADJUST] = ACTIONS(3105),
    [anon_sym_DOLLAR] = ACTIONS(3105),
    [anon_sym_AT] = ACTIONS(3105),
    [anon_sym_my] = ACTIONS(3105),
    [anon_sym_extended] = ACTIONS(3105),
    [anon_sym_async] = ACTIONS(3105),
    [anon_sym_sub] = ACTIONS(3105),
    [anon_sym_method] = ACTIONS(3105),
    [anon_sym_LPAREN] = ACTIONS(3105),
    [anon_sym_continue] = ACTIONS(3105),
    [anon_sym_state] = ACTIONS(3105),
    [anon_sym_our] = ACTIONS(3105),
    [anon_sym_try] = ACTIONS(3105),
    [anon_sym_defer] = ACTIONS(3105),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3105),
    [anon_sym_LBRACK] = ACTIONS(3105),
    [anon_sym_LT_LT] = ACTIONS(3105),
    [anon_sym_SLASH_SLASH] = ACTIONS(3105),
    [anon_sym_PLUS] = ACTIONS(3105),
    [anon_sym_DASH] = ACTIONS(3105),
    [anon_sym_TILDE] = ACTIONS(3105),
    [anon_sym_BANG] = ACTIONS(3105),
    [anon_sym_PLUS_PLUS] = ACTIONS(3105),
    [anon_sym_DASH_DASH] = ACTIONS(3105),
    [anon_sym_BSLASH] = ACTIONS(3103),
    [anon_sym_do] = ACTIONS(3105),
    [anon_sym_eval] = ACTIONS(3105),
    [anon_sym_await] = ACTIONS(3105),
    [anon_sym_field] = ACTIONS(3105),
    [anon_sym_local] = ACTIONS(3105),
    [anon_sym_dynamically] = ACTIONS(3105),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3105),
    [anon_sym_require] = ACTIONS(3105),
    [anon_sym_map] = ACTIONS(3105),
    [anon_sym_grep] = ACTIONS(3105),
    [anon_sym_sort] = ACTIONS(3105),
    [anon_sym_goto] = ACTIONS(3105),
    [anon_sym_return] = ACTIONS(3105),
    [anon_sym_undef] = ACTIONS(3105),
    [anon_sym_PERCENT2] = ACTIONS(3105),
    [anon_sym_AMP2] = ACTIONS(3105),
    [anon_sym_STAR2] = ACTIONS(3105),
    [anon_sym_use] = ACTIONS(3105),
    [anon_sym_no] = ACTIONS(3105),
    [anon_sym_for] = ACTIONS(3105),
    [anon_sym_foreach] = ACTIONS(3105),
    [anon_sym_last] = ACTIONS(3105),
    [anon_sym_next] = ACTIONS(3105),
    [anon_sym_redo] = ACTIONS(3105),
    [anon_sym_BEGIN] = ACTIONS(3105),
    [anon_sym_INIT] = ACTIONS(3105),
    [anon_sym_CHECK] = ACTIONS(3105),
    [anon_sym_UNITCHECK] = ACTIONS(3105),
    [anon_sym_END] = ACTIONS(3105),
    [anon_sym___FILE__] = ACTIONS(3105),
    [anon_sym___LINE__] = ACTIONS(3105),
    [anon_sym___PACKAGE__] = ACTIONS(3105),
    [anon_sym___SUB__] = ACTIONS(3105),
    [anon_sym_break] = ACTIONS(3105),
    [anon_sym_fork] = ACTIONS(3105),
    [anon_sym_getppid] = ACTIONS(3105),
    [anon_sym_time] = ACTIONS(3105),
    [anon_sym_times] = ACTIONS(3105),
    [anon_sym_wait] = ACTIONS(3105),
    [anon_sym_wantarray] = ACTIONS(3105),
    [anon_sym_abs] = ACTIONS(3105),
    [anon_sym_alarm] = ACTIONS(3105),
    [anon_sym_chop] = ACTIONS(3105),
    [anon_sym_chdir] = ACTIONS(3105),
    [anon_sym_close] = ACTIONS(3105),
    [anon_sym_closedir] = ACTIONS(3105),
    [anon_sym_caller] = ACTIONS(3105),
    [anon_sym_chomp] = ACTIONS(3105),
    [anon_sym_chr] = ACTIONS(3105),
    [anon_sym_cos] = ACTIONS(3105),
    [anon_sym_chroot] = ACTIONS(3105),
    [anon_sym_defined] = ACTIONS(3105),
    [anon_sym_delete] = ACTIONS(3105),
    [anon_sym_dbmclose] = ACTIONS(3105),
    [anon_sym_exists] = ACTIONS(3105),
    [anon_sym_exit] = ACTIONS(3105),
    [anon_sym_eof] = ACTIONS(3105),
    [anon_sym_exp] = ACTIONS(3105),
    [anon_sym_each] = ACTIONS(3105),
    [anon_sym_fc] = ACTIONS(3105),
    [anon_sym_fileno] = ACTIONS(3105),
    [anon_sym_gmtime] = ACTIONS(3105),
    [anon_sym_getc] = ACTIONS(3105),
    [anon_sym_getpgrp] = ACTIONS(3105),
    [anon_sym_getprotobyname] = ACTIONS(3105),
    [anon_sym_getpwname] = ACTIONS(3105),
    [anon_sym_getpwuid] = ACTIONS(3105),
    [anon_sym_getpeername] = ACTIONS(3105),
    [anon_sym_getnetbyname] = ACTIONS(3105),
    [anon_sym_getsockname] = ACTIONS(3105),
    [anon_sym_getgrnam] = ACTIONS(3105),
    [anon_sym_getgrgid] = ACTIONS(3105),
    [anon_sym_hex] = ACTIONS(3105),
    [anon_sym_int] = ACTIONS(3105),
    [anon_sym_keys] = ACTIONS(3105),
    [anon_sym_lc] = ACTIONS(3105),
    [anon_sym_lcfirst] = ACTIONS(3105),
    [anon_sym_length] = ACTIONS(3105),
    [anon_sym_localtime] = ACTIONS(3105),
    [anon_sym_log] = ACTIONS(3105),
    [anon_sym_lock] = ACTIONS(3105),
    [anon_sym_lstat] = ACTIONS(3105),
    [anon_sym_oct] = ACTIONS(3105),
    [anon_sym_ord] = ACTIONS(3105),
    [anon_sym_prototype] = ACTIONS(3105),
    [anon_sym_pop] = ACTIONS(3105),
    [anon_sym_pos] = ACTIONS(3105),
    [anon_sym_quotemeta] = ACTIONS(3105),
    [anon_sym_reset] = ACTIONS(3105),
    [anon_sym_rand] = ACTIONS(3105),
    [anon_sym_rmdir] = ACTIONS(3105),
    [anon_sym_readdir] = ACTIONS(3105),
    [anon_sym_readline] = ACTIONS(3105),
    [anon_sym_readpipe] = ACTIONS(3105),
    [anon_sym_rewinddir] = ACTIONS(3105),
    [anon_sym_readlink] = ACTIONS(3105),
    [anon_sym_ref] = ACTIONS(3105),
    [anon_sym_scalar] = ACTIONS(3105),
    [anon_sym_shift] = ACTIONS(3105),
    [anon_sym_sin] = ACTIONS(3105),
    [anon_sym_sleep] = ACTIONS(3105),
    [anon_sym_sqrt] = ACTIONS(3105),
    [anon_sym_srand] = ACTIONS(3105),
    [anon_sym_stat] = ACTIONS(3105),
    [anon_sym_study] = ACTIONS(3105),
    [anon_sym_tell] = ACTIONS(3105),
    [anon_sym_telldir] = ACTIONS(3105),
    [anon_sym_tied] = ACTIONS(3105),
    [anon_sym_uc] = ACTIONS(3105),
    [anon_sym_ucfirst] = ACTIONS(3105),
    [anon_sym_untie] = ACTIONS(3105),
    [anon_sym_umask] = ACTIONS(3105),
    [anon_sym_values] = ACTIONS(3105),
    [anon_sym_write] = ACTIONS(3105),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3105),
    [anon_sym_qq] = ACTIONS(3105),
    [anon_sym_qw] = ACTIONS(3105),
    [anon_sym_qx] = ACTIONS(3105),
    [anon_sym_qr] = ACTIONS(3105),
    [anon_sym_m] = ACTIONS(3105),
    [anon_sym_s] = ACTIONS(3105),
    [anon_sym_tr] = ACTIONS(3105),
    [anon_sym_y] = ACTIONS(3105),
    [anon_sym_if] = ACTIONS(3105),
    [anon_sym_unless] = ACTIONS(3105),
    [anon_sym_while] = ACTIONS(3105),
    [anon_sym_until] = ACTIONS(3105),
    [aux_sym__bareword_token1] = ACTIONS(3105),
    [sym_number] = ACTIONS(3105),
    [anon_sym_true] = ACTIONS(3105),
    [anon_sym_false] = ACTIONS(3105),
    [sym__single_quote] = ACTIONS(3103),
    [sym__double_quote] = ACTIONS(3103),
    [sym__backtick_quote] = ACTIONS(3103),
    [sym__search_slash_quote] = ACTIONS(3103),
    [sym__open_readline_bracket] = ACTIONS(3103),
    [sym__open_fileglob_bracket] = ACTIONS(3103),
    [sym__PERLY_HEREDOC] = ACTIONS(3103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3103),
    [sym__filetest] = ACTIONS(3103),
  },
  [STATE(1581)] = {
    [sym_heredoc_content] = STATE(1581),
    [sym__identifier] = ACTIONS(3141),
    [anon_sym_LBRACE] = ACTIONS(3141),
    [anon_sym_RBRACE] = ACTIONS(3141),
    [anon_sym_SEMI] = ACTIONS(3141),
    [anon_sym_package] = ACTIONS(3141),
    [anon_sym_class] = ACTIONS(3141),
    [anon_sym_role] = ACTIONS(3141),
    [anon_sym_BUILD] = ACTIONS(3141),
    [anon_sym_ADJUST] = ACTIONS(3141),
    [anon_sym_DOLLAR] = ACTIONS(3141),
    [anon_sym_AT] = ACTIONS(3141),
    [anon_sym_my] = ACTIONS(3141),
    [anon_sym_extended] = ACTIONS(3141),
    [anon_sym_async] = ACTIONS(3141),
    [anon_sym_sub] = ACTIONS(3141),
    [anon_sym_method] = ACTIONS(3141),
    [anon_sym_LPAREN] = ACTIONS(3141),
    [anon_sym_continue] = ACTIONS(3141),
    [anon_sym_state] = ACTIONS(3141),
    [anon_sym_our] = ACTIONS(3141),
    [anon_sym_try] = ACTIONS(3141),
    [anon_sym_defer] = ACTIONS(3141),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3141),
    [anon_sym_LBRACK] = ACTIONS(3141),
    [anon_sym_LT_LT] = ACTIONS(3141),
    [anon_sym_SLASH_SLASH] = ACTIONS(3141),
    [anon_sym_PLUS] = ACTIONS(3141),
    [anon_sym_DASH] = ACTIONS(3141),
    [anon_sym_TILDE] = ACTIONS(3141),
    [anon_sym_BANG] = ACTIONS(3141),
    [anon_sym_PLUS_PLUS] = ACTIONS(3141),
    [anon_sym_DASH_DASH] = ACTIONS(3141),
    [anon_sym_BSLASH] = ACTIONS(3139),
    [anon_sym_do] = ACTIONS(3141),
    [anon_sym_eval] = ACTIONS(3141),
    [anon_sym_await] = ACTIONS(3141),
    [anon_sym_field] = ACTIONS(3141),
    [anon_sym_local] = ACTIONS(3141),
    [anon_sym_dynamically] = ACTIONS(3141),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3141),
    [anon_sym_require] = ACTIONS(3141),
    [anon_sym_map] = ACTIONS(3141),
    [anon_sym_grep] = ACTIONS(3141),
    [anon_sym_sort] = ACTIONS(3141),
    [anon_sym_goto] = ACTIONS(3141),
    [anon_sym_return] = ACTIONS(3141),
    [anon_sym_undef] = ACTIONS(3141),
    [anon_sym_PERCENT2] = ACTIONS(3141),
    [anon_sym_AMP2] = ACTIONS(3141),
    [anon_sym_STAR2] = ACTIONS(3141),
    [anon_sym_use] = ACTIONS(3141),
    [anon_sym_no] = ACTIONS(3141),
    [anon_sym_for] = ACTIONS(3141),
    [anon_sym_foreach] = ACTIONS(3141),
    [anon_sym_last] = ACTIONS(3141),
    [anon_sym_next] = ACTIONS(3141),
    [anon_sym_redo] = ACTIONS(3141),
    [anon_sym_BEGIN] = ACTIONS(3141),
    [anon_sym_INIT] = ACTIONS(3141),
    [anon_sym_CHECK] = ACTIONS(3141),
    [anon_sym_UNITCHECK] = ACTIONS(3141),
    [anon_sym_END] = ACTIONS(3141),
    [anon_sym___FILE__] = ACTIONS(3141),
    [anon_sym___LINE__] = ACTIONS(3141),
    [anon_sym___PACKAGE__] = ACTIONS(3141),
    [anon_sym___SUB__] = ACTIONS(3141),
    [anon_sym_break] = ACTIONS(3141),
    [anon_sym_fork] = ACTIONS(3141),
    [anon_sym_getppid] = ACTIONS(3141),
    [anon_sym_time] = ACTIONS(3141),
    [anon_sym_times] = ACTIONS(3141),
    [anon_sym_wait] = ACTIONS(3141),
    [anon_sym_wantarray] = ACTIONS(3141),
    [anon_sym_abs] = ACTIONS(3141),
    [anon_sym_alarm] = ACTIONS(3141),
    [anon_sym_chop] = ACTIONS(3141),
    [anon_sym_chdir] = ACTIONS(3141),
    [anon_sym_close] = ACTIONS(3141),
    [anon_sym_closedir] = ACTIONS(3141),
    [anon_sym_caller] = ACTIONS(3141),
    [anon_sym_chomp] = ACTIONS(3141),
    [anon_sym_chr] = ACTIONS(3141),
    [anon_sym_cos] = ACTIONS(3141),
    [anon_sym_chroot] = ACTIONS(3141),
    [anon_sym_defined] = ACTIONS(3141),
    [anon_sym_delete] = ACTIONS(3141),
    [anon_sym_dbmclose] = ACTIONS(3141),
    [anon_sym_exists] = ACTIONS(3141),
    [anon_sym_exit] = ACTIONS(3141),
    [anon_sym_eof] = ACTIONS(3141),
    [anon_sym_exp] = ACTIONS(3141),
    [anon_sym_each] = ACTIONS(3141),
    [anon_sym_fc] = ACTIONS(3141),
    [anon_sym_fileno] = ACTIONS(3141),
    [anon_sym_gmtime] = ACTIONS(3141),
    [anon_sym_getc] = ACTIONS(3141),
    [anon_sym_getpgrp] = ACTIONS(3141),
    [anon_sym_getprotobyname] = ACTIONS(3141),
    [anon_sym_getpwname] = ACTIONS(3141),
    [anon_sym_getpwuid] = ACTIONS(3141),
    [anon_sym_getpeername] = ACTIONS(3141),
    [anon_sym_getnetbyname] = ACTIONS(3141),
    [anon_sym_getsockname] = ACTIONS(3141),
    [anon_sym_getgrnam] = ACTIONS(3141),
    [anon_sym_getgrgid] = ACTIONS(3141),
    [anon_sym_hex] = ACTIONS(3141),
    [anon_sym_int] = ACTIONS(3141),
    [anon_sym_keys] = ACTIONS(3141),
    [anon_sym_lc] = ACTIONS(3141),
    [anon_sym_lcfirst] = ACTIONS(3141),
    [anon_sym_length] = ACTIONS(3141),
    [anon_sym_localtime] = ACTIONS(3141),
    [anon_sym_log] = ACTIONS(3141),
    [anon_sym_lock] = ACTIONS(3141),
    [anon_sym_lstat] = ACTIONS(3141),
    [anon_sym_oct] = ACTIONS(3141),
    [anon_sym_ord] = ACTIONS(3141),
    [anon_sym_prototype] = ACTIONS(3141),
    [anon_sym_pop] = ACTIONS(3141),
    [anon_sym_pos] = ACTIONS(3141),
    [anon_sym_quotemeta] = ACTIONS(3141),
    [anon_sym_reset] = ACTIONS(3141),
    [anon_sym_rand] = ACTIONS(3141),
    [anon_sym_rmdir] = ACTIONS(3141),
    [anon_sym_readdir] = ACTIONS(3141),
    [anon_sym_readline] = ACTIONS(3141),
    [anon_sym_readpipe] = ACTIONS(3141),
    [anon_sym_rewinddir] = ACTIONS(3141),
    [anon_sym_readlink] = ACTIONS(3141),
    [anon_sym_ref] = ACTIONS(3141),
    [anon_sym_scalar] = ACTIONS(3141),
    [anon_sym_shift] = ACTIONS(3141),
    [anon_sym_sin] = ACTIONS(3141),
    [anon_sym_sleep] = ACTIONS(3141),
    [anon_sym_sqrt] = ACTIONS(3141),
    [anon_sym_srand] = ACTIONS(3141),
    [anon_sym_stat] = ACTIONS(3141),
    [anon_sym_study] = ACTIONS(3141),
    [anon_sym_tell] = ACTIONS(3141),
    [anon_sym_telldir] = ACTIONS(3141),
    [anon_sym_tied] = ACTIONS(3141),
    [anon_sym_uc] = ACTIONS(3141),
    [anon_sym_ucfirst] = ACTIONS(3141),
    [anon_sym_untie] = ACTIONS(3141),
    [anon_sym_umask] = ACTIONS(3141),
    [anon_sym_values] = ACTIONS(3141),
    [anon_sym_write] = ACTIONS(3141),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3141),
    [anon_sym_qq] = ACTIONS(3141),
    [anon_sym_qw] = ACTIONS(3141),
    [anon_sym_qx] = ACTIONS(3141),
    [anon_sym_qr] = ACTIONS(3141),
    [anon_sym_m] = ACTIONS(3141),
    [anon_sym_s] = ACTIONS(3141),
    [anon_sym_tr] = ACTIONS(3141),
    [anon_sym_y] = ACTIONS(3141),
    [anon_sym_if] = ACTIONS(3141),
    [anon_sym_unless] = ACTIONS(3141),
    [anon_sym_while] = ACTIONS(3141),
    [anon_sym_until] = ACTIONS(3141),
    [aux_sym__bareword_token1] = ACTIONS(3141),
    [sym_number] = ACTIONS(3141),
    [anon_sym_true] = ACTIONS(3141),
    [anon_sym_false] = ACTIONS(3141),
    [sym__single_quote] = ACTIONS(3139),
    [sym__double_quote] = ACTIONS(3139),
    [sym__backtick_quote] = ACTIONS(3139),
    [sym__search_slash_quote] = ACTIONS(3139),
    [sym__open_readline_bracket] = ACTIONS(3139),
    [sym__open_fileglob_bracket] = ACTIONS(3139),
    [sym__PERLY_HEREDOC] = ACTIONS(3139),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3139),
    [sym__filetest] = ACTIONS(3139),
  },
  [STATE(1582)] = {
    [sym_heredoc_content] = STATE(1582),
    [sym__identifier] = ACTIONS(3161),
    [anon_sym_LBRACE] = ACTIONS(3161),
    [anon_sym_RBRACE] = ACTIONS(3161),
    [anon_sym_SEMI] = ACTIONS(3161),
    [anon_sym_package] = ACTIONS(3161),
    [anon_sym_class] = ACTIONS(3161),
    [anon_sym_role] = ACTIONS(3161),
    [anon_sym_BUILD] = ACTIONS(3161),
    [anon_sym_ADJUST] = ACTIONS(3161),
    [anon_sym_DOLLAR] = ACTIONS(3161),
    [anon_sym_AT] = ACTIONS(3161),
    [anon_sym_my] = ACTIONS(3161),
    [anon_sym_extended] = ACTIONS(3161),
    [anon_sym_async] = ACTIONS(3161),
    [anon_sym_sub] = ACTIONS(3161),
    [anon_sym_method] = ACTIONS(3161),
    [anon_sym_LPAREN] = ACTIONS(3161),
    [anon_sym_continue] = ACTIONS(3161),
    [anon_sym_state] = ACTIONS(3161),
    [anon_sym_our] = ACTIONS(3161),
    [anon_sym_try] = ACTIONS(3161),
    [anon_sym_defer] = ACTIONS(3161),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3161),
    [anon_sym_LBRACK] = ACTIONS(3161),
    [anon_sym_LT_LT] = ACTIONS(3161),
    [anon_sym_SLASH_SLASH] = ACTIONS(3161),
    [anon_sym_PLUS] = ACTIONS(3161),
    [anon_sym_DASH] = ACTIONS(3161),
    [anon_sym_TILDE] = ACTIONS(3161),
    [anon_sym_BANG] = ACTIONS(3161),
    [anon_sym_PLUS_PLUS] = ACTIONS(3161),
    [anon_sym_DASH_DASH] = ACTIONS(3161),
    [anon_sym_BSLASH] = ACTIONS(3159),
    [anon_sym_do] = ACTIONS(3161),
    [anon_sym_eval] = ACTIONS(3161),
    [anon_sym_await] = ACTIONS(3161),
    [anon_sym_field] = ACTIONS(3161),
    [anon_sym_local] = ACTIONS(3161),
    [anon_sym_dynamically] = ACTIONS(3161),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3161),
    [anon_sym_require] = ACTIONS(3161),
    [anon_sym_map] = ACTIONS(3161),
    [anon_sym_grep] = ACTIONS(3161),
    [anon_sym_sort] = ACTIONS(3161),
    [anon_sym_goto] = ACTIONS(3161),
    [anon_sym_return] = ACTIONS(3161),
    [anon_sym_undef] = ACTIONS(3161),
    [anon_sym_PERCENT2] = ACTIONS(3161),
    [anon_sym_AMP2] = ACTIONS(3161),
    [anon_sym_STAR2] = ACTIONS(3161),
    [anon_sym_use] = ACTIONS(3161),
    [anon_sym_no] = ACTIONS(3161),
    [anon_sym_for] = ACTIONS(3161),
    [anon_sym_foreach] = ACTIONS(3161),
    [anon_sym_last] = ACTIONS(3161),
    [anon_sym_next] = ACTIONS(3161),
    [anon_sym_redo] = ACTIONS(3161),
    [anon_sym_BEGIN] = ACTIONS(3161),
    [anon_sym_INIT] = ACTIONS(3161),
    [anon_sym_CHECK] = ACTIONS(3161),
    [anon_sym_UNITCHECK] = ACTIONS(3161),
    [anon_sym_END] = ACTIONS(3161),
    [anon_sym___FILE__] = ACTIONS(3161),
    [anon_sym___LINE__] = ACTIONS(3161),
    [anon_sym___PACKAGE__] = ACTIONS(3161),
    [anon_sym___SUB__] = ACTIONS(3161),
    [anon_sym_break] = ACTIONS(3161),
    [anon_sym_fork] = ACTIONS(3161),
    [anon_sym_getppid] = ACTIONS(3161),
    [anon_sym_time] = ACTIONS(3161),
    [anon_sym_times] = ACTIONS(3161),
    [anon_sym_wait] = ACTIONS(3161),
    [anon_sym_wantarray] = ACTIONS(3161),
    [anon_sym_abs] = ACTIONS(3161),
    [anon_sym_alarm] = ACTIONS(3161),
    [anon_sym_chop] = ACTIONS(3161),
    [anon_sym_chdir] = ACTIONS(3161),
    [anon_sym_close] = ACTIONS(3161),
    [anon_sym_closedir] = ACTIONS(3161),
    [anon_sym_caller] = ACTIONS(3161),
    [anon_sym_chomp] = ACTIONS(3161),
    [anon_sym_chr] = ACTIONS(3161),
    [anon_sym_cos] = ACTIONS(3161),
    [anon_sym_chroot] = ACTIONS(3161),
    [anon_sym_defined] = ACTIONS(3161),
    [anon_sym_delete] = ACTIONS(3161),
    [anon_sym_dbmclose] = ACTIONS(3161),
    [anon_sym_exists] = ACTIONS(3161),
    [anon_sym_exit] = ACTIONS(3161),
    [anon_sym_eof] = ACTIONS(3161),
    [anon_sym_exp] = ACTIONS(3161),
    [anon_sym_each] = ACTIONS(3161),
    [anon_sym_fc] = ACTIONS(3161),
    [anon_sym_fileno] = ACTIONS(3161),
    [anon_sym_gmtime] = ACTIONS(3161),
    [anon_sym_getc] = ACTIONS(3161),
    [anon_sym_getpgrp] = ACTIONS(3161),
    [anon_sym_getprotobyname] = ACTIONS(3161),
    [anon_sym_getpwname] = ACTIONS(3161),
    [anon_sym_getpwuid] = ACTIONS(3161),
    [anon_sym_getpeername] = ACTIONS(3161),
    [anon_sym_getnetbyname] = ACTIONS(3161),
    [anon_sym_getsockname] = ACTIONS(3161),
    [anon_sym_getgrnam] = ACTIONS(3161),
    [anon_sym_getgrgid] = ACTIONS(3161),
    [anon_sym_hex] = ACTIONS(3161),
    [anon_sym_int] = ACTIONS(3161),
    [anon_sym_keys] = ACTIONS(3161),
    [anon_sym_lc] = ACTIONS(3161),
    [anon_sym_lcfirst] = ACTIONS(3161),
    [anon_sym_length] = ACTIONS(3161),
    [anon_sym_localtime] = ACTIONS(3161),
    [anon_sym_log] = ACTIONS(3161),
    [anon_sym_lock] = ACTIONS(3161),
    [anon_sym_lstat] = ACTIONS(3161),
    [anon_sym_oct] = ACTIONS(3161),
    [anon_sym_ord] = ACTIONS(3161),
    [anon_sym_prototype] = ACTIONS(3161),
    [anon_sym_pop] = ACTIONS(3161),
    [anon_sym_pos] = ACTIONS(3161),
    [anon_sym_quotemeta] = ACTIONS(3161),
    [anon_sym_reset] = ACTIONS(3161),
    [anon_sym_rand] = ACTIONS(3161),
    [anon_sym_rmdir] = ACTIONS(3161),
    [anon_sym_readdir] = ACTIONS(3161),
    [anon_sym_readline] = ACTIONS(3161),
    [anon_sym_readpipe] = ACTIONS(3161),
    [anon_sym_rewinddir] = ACTIONS(3161),
    [anon_sym_readlink] = ACTIONS(3161),
    [anon_sym_ref] = ACTIONS(3161),
    [anon_sym_scalar] = ACTIONS(3161),
    [anon_sym_shift] = ACTIONS(3161),
    [anon_sym_sin] = ACTIONS(3161),
    [anon_sym_sleep] = ACTIONS(3161),
    [anon_sym_sqrt] = ACTIONS(3161),
    [anon_sym_srand] = ACTIONS(3161),
    [anon_sym_stat] = ACTIONS(3161),
    [anon_sym_study] = ACTIONS(3161),
    [anon_sym_tell] = ACTIONS(3161),
    [anon_sym_telldir] = ACTIONS(3161),
    [anon_sym_tied] = ACTIONS(3161),
    [anon_sym_uc] = ACTIONS(3161),
    [anon_sym_ucfirst] = ACTIONS(3161),
    [anon_sym_untie] = ACTIONS(3161),
    [anon_sym_umask] = ACTIONS(3161),
    [anon_sym_values] = ACTIONS(3161),
    [anon_sym_write] = ACTIONS(3161),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3161),
    [anon_sym_qq] = ACTIONS(3161),
    [anon_sym_qw] = ACTIONS(3161),
    [anon_sym_qx] = ACTIONS(3161),
    [anon_sym_qr] = ACTIONS(3161),
    [anon_sym_m] = ACTIONS(3161),
    [anon_sym_s] = ACTIONS(3161),
    [anon_sym_tr] = ACTIONS(3161),
    [anon_sym_y] = ACTIONS(3161),
    [anon_sym_if] = ACTIONS(3161),
    [anon_sym_unless] = ACTIONS(3161),
    [anon_sym_while] = ACTIONS(3161),
    [anon_sym_until] = ACTIONS(3161),
    [aux_sym__bareword_token1] = ACTIONS(3161),
    [sym_number] = ACTIONS(3161),
    [anon_sym_true] = ACTIONS(3161),
    [anon_sym_false] = ACTIONS(3161),
    [sym__single_quote] = ACTIONS(3159),
    [sym__double_quote] = ACTIONS(3159),
    [sym__backtick_quote] = ACTIONS(3159),
    [sym__search_slash_quote] = ACTIONS(3159),
    [sym__open_readline_bracket] = ACTIONS(3159),
    [sym__open_fileglob_bracket] = ACTIONS(3159),
    [sym__PERLY_HEREDOC] = ACTIONS(3159),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3159),
    [sym__filetest] = ACTIONS(3159),
  },
  [STATE(1583)] = {
    [sym_heredoc_content] = STATE(1583),
    [sym__identifier] = ACTIONS(3173),
    [anon_sym_LBRACE] = ACTIONS(3173),
    [anon_sym_RBRACE] = ACTIONS(3173),
    [anon_sym_SEMI] = ACTIONS(3173),
    [anon_sym_package] = ACTIONS(3173),
    [anon_sym_class] = ACTIONS(3173),
    [anon_sym_role] = ACTIONS(3173),
    [anon_sym_BUILD] = ACTIONS(3173),
    [anon_sym_ADJUST] = ACTIONS(3173),
    [anon_sym_DOLLAR] = ACTIONS(3173),
    [anon_sym_AT] = ACTIONS(3173),
    [anon_sym_my] = ACTIONS(3173),
    [anon_sym_extended] = ACTIONS(3173),
    [anon_sym_async] = ACTIONS(3173),
    [anon_sym_sub] = ACTIONS(3173),
    [anon_sym_method] = ACTIONS(3173),
    [anon_sym_LPAREN] = ACTIONS(3173),
    [anon_sym_continue] = ACTIONS(3173),
    [anon_sym_state] = ACTIONS(3173),
    [anon_sym_our] = ACTIONS(3173),
    [anon_sym_try] = ACTIONS(3173),
    [anon_sym_defer] = ACTIONS(3173),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3173),
    [anon_sym_LBRACK] = ACTIONS(3173),
    [anon_sym_LT_LT] = ACTIONS(3173),
    [anon_sym_SLASH_SLASH] = ACTIONS(3173),
    [anon_sym_PLUS] = ACTIONS(3173),
    [anon_sym_DASH] = ACTIONS(3173),
    [anon_sym_TILDE] = ACTIONS(3173),
    [anon_sym_BANG] = ACTIONS(3173),
    [anon_sym_PLUS_PLUS] = ACTIONS(3173),
    [anon_sym_DASH_DASH] = ACTIONS(3173),
    [anon_sym_BSLASH] = ACTIONS(3171),
    [anon_sym_do] = ACTIONS(3173),
    [anon_sym_eval] = ACTIONS(3173),
    [anon_sym_await] = ACTIONS(3173),
    [anon_sym_field] = ACTIONS(3173),
    [anon_sym_local] = ACTIONS(3173),
    [anon_sym_dynamically] = ACTIONS(3173),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3173),
    [anon_sym_require] = ACTIONS(3173),
    [anon_sym_map] = ACTIONS(3173),
    [anon_sym_grep] = ACTIONS(3173),
    [anon_sym_sort] = ACTIONS(3173),
    [anon_sym_goto] = ACTIONS(3173),
    [anon_sym_return] = ACTIONS(3173),
    [anon_sym_undef] = ACTIONS(3173),
    [anon_sym_PERCENT2] = ACTIONS(3173),
    [anon_sym_AMP2] = ACTIONS(3173),
    [anon_sym_STAR2] = ACTIONS(3173),
    [anon_sym_use] = ACTIONS(3173),
    [anon_sym_no] = ACTIONS(3173),
    [anon_sym_for] = ACTIONS(3173),
    [anon_sym_foreach] = ACTIONS(3173),
    [anon_sym_last] = ACTIONS(3173),
    [anon_sym_next] = ACTIONS(3173),
    [anon_sym_redo] = ACTIONS(3173),
    [anon_sym_BEGIN] = ACTIONS(3173),
    [anon_sym_INIT] = ACTIONS(3173),
    [anon_sym_CHECK] = ACTIONS(3173),
    [anon_sym_UNITCHECK] = ACTIONS(3173),
    [anon_sym_END] = ACTIONS(3173),
    [anon_sym___FILE__] = ACTIONS(3173),
    [anon_sym___LINE__] = ACTIONS(3173),
    [anon_sym___PACKAGE__] = ACTIONS(3173),
    [anon_sym___SUB__] = ACTIONS(3173),
    [anon_sym_break] = ACTIONS(3173),
    [anon_sym_fork] = ACTIONS(3173),
    [anon_sym_getppid] = ACTIONS(3173),
    [anon_sym_time] = ACTIONS(3173),
    [anon_sym_times] = ACTIONS(3173),
    [anon_sym_wait] = ACTIONS(3173),
    [anon_sym_wantarray] = ACTIONS(3173),
    [anon_sym_abs] = ACTIONS(3173),
    [anon_sym_alarm] = ACTIONS(3173),
    [anon_sym_chop] = ACTIONS(3173),
    [anon_sym_chdir] = ACTIONS(3173),
    [anon_sym_close] = ACTIONS(3173),
    [anon_sym_closedir] = ACTIONS(3173),
    [anon_sym_caller] = ACTIONS(3173),
    [anon_sym_chomp] = ACTIONS(3173),
    [anon_sym_chr] = ACTIONS(3173),
    [anon_sym_cos] = ACTIONS(3173),
    [anon_sym_chroot] = ACTIONS(3173),
    [anon_sym_defined] = ACTIONS(3173),
    [anon_sym_delete] = ACTIONS(3173),
    [anon_sym_dbmclose] = ACTIONS(3173),
    [anon_sym_exists] = ACTIONS(3173),
    [anon_sym_exit] = ACTIONS(3173),
    [anon_sym_eof] = ACTIONS(3173),
    [anon_sym_exp] = ACTIONS(3173),
    [anon_sym_each] = ACTIONS(3173),
    [anon_sym_fc] = ACTIONS(3173),
    [anon_sym_fileno] = ACTIONS(3173),
    [anon_sym_gmtime] = ACTIONS(3173),
    [anon_sym_getc] = ACTIONS(3173),
    [anon_sym_getpgrp] = ACTIONS(3173),
    [anon_sym_getprotobyname] = ACTIONS(3173),
    [anon_sym_getpwname] = ACTIONS(3173),
    [anon_sym_getpwuid] = ACTIONS(3173),
    [anon_sym_getpeername] = ACTIONS(3173),
    [anon_sym_getnetbyname] = ACTIONS(3173),
    [anon_sym_getsockname] = ACTIONS(3173),
    [anon_sym_getgrnam] = ACTIONS(3173),
    [anon_sym_getgrgid] = ACTIONS(3173),
    [anon_sym_hex] = ACTIONS(3173),
    [anon_sym_int] = ACTIONS(3173),
    [anon_sym_keys] = ACTIONS(3173),
    [anon_sym_lc] = ACTIONS(3173),
    [anon_sym_lcfirst] = ACTIONS(3173),
    [anon_sym_length] = ACTIONS(3173),
    [anon_sym_localtime] = ACTIONS(3173),
    [anon_sym_log] = ACTIONS(3173),
    [anon_sym_lock] = ACTIONS(3173),
    [anon_sym_lstat] = ACTIONS(3173),
    [anon_sym_oct] = ACTIONS(3173),
    [anon_sym_ord] = ACTIONS(3173),
    [anon_sym_prototype] = ACTIONS(3173),
    [anon_sym_pop] = ACTIONS(3173),
    [anon_sym_pos] = ACTIONS(3173),
    [anon_sym_quotemeta] = ACTIONS(3173),
    [anon_sym_reset] = ACTIONS(3173),
    [anon_sym_rand] = ACTIONS(3173),
    [anon_sym_rmdir] = ACTIONS(3173),
    [anon_sym_readdir] = ACTIONS(3173),
    [anon_sym_readline] = ACTIONS(3173),
    [anon_sym_readpipe] = ACTIONS(3173),
    [anon_sym_rewinddir] = ACTIONS(3173),
    [anon_sym_readlink] = ACTIONS(3173),
    [anon_sym_ref] = ACTIONS(3173),
    [anon_sym_scalar] = ACTIONS(3173),
    [anon_sym_shift] = ACTIONS(3173),
    [anon_sym_sin] = ACTIONS(3173),
    [anon_sym_sleep] = ACTIONS(3173),
    [anon_sym_sqrt] = ACTIONS(3173),
    [anon_sym_srand] = ACTIONS(3173),
    [anon_sym_stat] = ACTIONS(3173),
    [anon_sym_study] = ACTIONS(3173),
    [anon_sym_tell] = ACTIONS(3173),
    [anon_sym_telldir] = ACTIONS(3173),
    [anon_sym_tied] = ACTIONS(3173),
    [anon_sym_uc] = ACTIONS(3173),
    [anon_sym_ucfirst] = ACTIONS(3173),
    [anon_sym_untie] = ACTIONS(3173),
    [anon_sym_umask] = ACTIONS(3173),
    [anon_sym_values] = ACTIONS(3173),
    [anon_sym_write] = ACTIONS(3173),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3173),
    [anon_sym_qq] = ACTIONS(3173),
    [anon_sym_qw] = ACTIONS(3173),
    [anon_sym_qx] = ACTIONS(3173),
    [anon_sym_qr] = ACTIONS(3173),
    [anon_sym_m] = ACTIONS(3173),
    [anon_sym_s] = ACTIONS(3173),
    [anon_sym_tr] = ACTIONS(3173),
    [anon_sym_y] = ACTIONS(3173),
    [anon_sym_if] = ACTIONS(3173),
    [anon_sym_unless] = ACTIONS(3173),
    [anon_sym_while] = ACTIONS(3173),
    [anon_sym_until] = ACTIONS(3173),
    [aux_sym__bareword_token1] = ACTIONS(3173),
    [sym_number] = ACTIONS(3173),
    [anon_sym_true] = ACTIONS(3173),
    [anon_sym_false] = ACTIONS(3173),
    [sym__single_quote] = ACTIONS(3171),
    [sym__double_quote] = ACTIONS(3171),
    [sym__backtick_quote] = ACTIONS(3171),
    [sym__search_slash_quote] = ACTIONS(3171),
    [sym__open_readline_bracket] = ACTIONS(3171),
    [sym__open_fileglob_bracket] = ACTIONS(3171),
    [sym__PERLY_HEREDOC] = ACTIONS(3171),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3171),
    [sym__filetest] = ACTIONS(3171),
  },
  [STATE(1584)] = {
    [sym_heredoc_content] = STATE(1584),
    [sym__identifier] = ACTIONS(3181),
    [anon_sym_LBRACE] = ACTIONS(3181),
    [anon_sym_RBRACE] = ACTIONS(3181),
    [anon_sym_SEMI] = ACTIONS(3181),
    [anon_sym_package] = ACTIONS(3181),
    [anon_sym_class] = ACTIONS(3181),
    [anon_sym_role] = ACTIONS(3181),
    [anon_sym_BUILD] = ACTIONS(3181),
    [anon_sym_ADJUST] = ACTIONS(3181),
    [anon_sym_DOLLAR] = ACTIONS(3181),
    [anon_sym_AT] = ACTIONS(3181),
    [anon_sym_my] = ACTIONS(3181),
    [anon_sym_extended] = ACTIONS(3181),
    [anon_sym_async] = ACTIONS(3181),
    [anon_sym_sub] = ACTIONS(3181),
    [anon_sym_method] = ACTIONS(3181),
    [anon_sym_LPAREN] = ACTIONS(3181),
    [anon_sym_continue] = ACTIONS(3181),
    [anon_sym_state] = ACTIONS(3181),
    [anon_sym_our] = ACTIONS(3181),
    [anon_sym_try] = ACTIONS(3181),
    [anon_sym_defer] = ACTIONS(3181),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3181),
    [anon_sym_LBRACK] = ACTIONS(3181),
    [anon_sym_LT_LT] = ACTIONS(3181),
    [anon_sym_SLASH_SLASH] = ACTIONS(3181),
    [anon_sym_PLUS] = ACTIONS(3181),
    [anon_sym_DASH] = ACTIONS(3181),
    [anon_sym_TILDE] = ACTIONS(3181),
    [anon_sym_BANG] = ACTIONS(3181),
    [anon_sym_PLUS_PLUS] = ACTIONS(3181),
    [anon_sym_DASH_DASH] = ACTIONS(3181),
    [anon_sym_BSLASH] = ACTIONS(3179),
    [anon_sym_do] = ACTIONS(3181),
    [anon_sym_eval] = ACTIONS(3181),
    [anon_sym_await] = ACTIONS(3181),
    [anon_sym_field] = ACTIONS(3181),
    [anon_sym_local] = ACTIONS(3181),
    [anon_sym_dynamically] = ACTIONS(3181),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3181),
    [anon_sym_require] = ACTIONS(3181),
    [anon_sym_map] = ACTIONS(3181),
    [anon_sym_grep] = ACTIONS(3181),
    [anon_sym_sort] = ACTIONS(3181),
    [anon_sym_goto] = ACTIONS(3181),
    [anon_sym_return] = ACTIONS(3181),
    [anon_sym_undef] = ACTIONS(3181),
    [anon_sym_PERCENT2] = ACTIONS(3181),
    [anon_sym_AMP2] = ACTIONS(3181),
    [anon_sym_STAR2] = ACTIONS(3181),
    [anon_sym_use] = ACTIONS(3181),
    [anon_sym_no] = ACTIONS(3181),
    [anon_sym_for] = ACTIONS(3181),
    [anon_sym_foreach] = ACTIONS(3181),
    [anon_sym_last] = ACTIONS(3181),
    [anon_sym_next] = ACTIONS(3181),
    [anon_sym_redo] = ACTIONS(3181),
    [anon_sym_BEGIN] = ACTIONS(3181),
    [anon_sym_INIT] = ACTIONS(3181),
    [anon_sym_CHECK] = ACTIONS(3181),
    [anon_sym_UNITCHECK] = ACTIONS(3181),
    [anon_sym_END] = ACTIONS(3181),
    [anon_sym___FILE__] = ACTIONS(3181),
    [anon_sym___LINE__] = ACTIONS(3181),
    [anon_sym___PACKAGE__] = ACTIONS(3181),
    [anon_sym___SUB__] = ACTIONS(3181),
    [anon_sym_break] = ACTIONS(3181),
    [anon_sym_fork] = ACTIONS(3181),
    [anon_sym_getppid] = ACTIONS(3181),
    [anon_sym_time] = ACTIONS(3181),
    [anon_sym_times] = ACTIONS(3181),
    [anon_sym_wait] = ACTIONS(3181),
    [anon_sym_wantarray] = ACTIONS(3181),
    [anon_sym_abs] = ACTIONS(3181),
    [anon_sym_alarm] = ACTIONS(3181),
    [anon_sym_chop] = ACTIONS(3181),
    [anon_sym_chdir] = ACTIONS(3181),
    [anon_sym_close] = ACTIONS(3181),
    [anon_sym_closedir] = ACTIONS(3181),
    [anon_sym_caller] = ACTIONS(3181),
    [anon_sym_chomp] = ACTIONS(3181),
    [anon_sym_chr] = ACTIONS(3181),
    [anon_sym_cos] = ACTIONS(3181),
    [anon_sym_chroot] = ACTIONS(3181),
    [anon_sym_defined] = ACTIONS(3181),
    [anon_sym_delete] = ACTIONS(3181),
    [anon_sym_dbmclose] = ACTIONS(3181),
    [anon_sym_exists] = ACTIONS(3181),
    [anon_sym_exit] = ACTIONS(3181),
    [anon_sym_eof] = ACTIONS(3181),
    [anon_sym_exp] = ACTIONS(3181),
    [anon_sym_each] = ACTIONS(3181),
    [anon_sym_fc] = ACTIONS(3181),
    [anon_sym_fileno] = ACTIONS(3181),
    [anon_sym_gmtime] = ACTIONS(3181),
    [anon_sym_getc] = ACTIONS(3181),
    [anon_sym_getpgrp] = ACTIONS(3181),
    [anon_sym_getprotobyname] = ACTIONS(3181),
    [anon_sym_getpwname] = ACTIONS(3181),
    [anon_sym_getpwuid] = ACTIONS(3181),
    [anon_sym_getpeername] = ACTIONS(3181),
    [anon_sym_getnetbyname] = ACTIONS(3181),
    [anon_sym_getsockname] = ACTIONS(3181),
    [anon_sym_getgrnam] = ACTIONS(3181),
    [anon_sym_getgrgid] = ACTIONS(3181),
    [anon_sym_hex] = ACTIONS(3181),
    [anon_sym_int] = ACTIONS(3181),
    [anon_sym_keys] = ACTIONS(3181),
    [anon_sym_lc] = ACTIONS(3181),
    [anon_sym_lcfirst] = ACTIONS(3181),
    [anon_sym_length] = ACTIONS(3181),
    [anon_sym_localtime] = ACTIONS(3181),
    [anon_sym_log] = ACTIONS(3181),
    [anon_sym_lock] = ACTIONS(3181),
    [anon_sym_lstat] = ACTIONS(3181),
    [anon_sym_oct] = ACTIONS(3181),
    [anon_sym_ord] = ACTIONS(3181),
    [anon_sym_prototype] = ACTIONS(3181),
    [anon_sym_pop] = ACTIONS(3181),
    [anon_sym_pos] = ACTIONS(3181),
    [anon_sym_quotemeta] = ACTIONS(3181),
    [anon_sym_reset] = ACTIONS(3181),
    [anon_sym_rand] = ACTIONS(3181),
    [anon_sym_rmdir] = ACTIONS(3181),
    [anon_sym_readdir] = ACTIONS(3181),
    [anon_sym_readline] = ACTIONS(3181),
    [anon_sym_readpipe] = ACTIONS(3181),
    [anon_sym_rewinddir] = ACTIONS(3181),
    [anon_sym_readlink] = ACTIONS(3181),
    [anon_sym_ref] = ACTIONS(3181),
    [anon_sym_scalar] = ACTIONS(3181),
    [anon_sym_shift] = ACTIONS(3181),
    [anon_sym_sin] = ACTIONS(3181),
    [anon_sym_sleep] = ACTIONS(3181),
    [anon_sym_sqrt] = ACTIONS(3181),
    [anon_sym_srand] = ACTIONS(3181),
    [anon_sym_stat] = ACTIONS(3181),
    [anon_sym_study] = ACTIONS(3181),
    [anon_sym_tell] = ACTIONS(3181),
    [anon_sym_telldir] = ACTIONS(3181),
    [anon_sym_tied] = ACTIONS(3181),
    [anon_sym_uc] = ACTIONS(3181),
    [anon_sym_ucfirst] = ACTIONS(3181),
    [anon_sym_untie] = ACTIONS(3181),
    [anon_sym_umask] = ACTIONS(3181),
    [anon_sym_values] = ACTIONS(3181),
    [anon_sym_write] = ACTIONS(3181),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3181),
    [anon_sym_qq] = ACTIONS(3181),
    [anon_sym_qw] = ACTIONS(3181),
    [anon_sym_qx] = ACTIONS(3181),
    [anon_sym_qr] = ACTIONS(3181),
    [anon_sym_m] = ACTIONS(3181),
    [anon_sym_s] = ACTIONS(3181),
    [anon_sym_tr] = ACTIONS(3181),
    [anon_sym_y] = ACTIONS(3181),
    [anon_sym_if] = ACTIONS(3181),
    [anon_sym_unless] = ACTIONS(3181),
    [anon_sym_while] = ACTIONS(3181),
    [anon_sym_until] = ACTIONS(3181),
    [aux_sym__bareword_token1] = ACTIONS(3181),
    [sym_number] = ACTIONS(3181),
    [anon_sym_true] = ACTIONS(3181),
    [anon_sym_false] = ACTIONS(3181),
    [sym__single_quote] = ACTIONS(3179),
    [sym__double_quote] = ACTIONS(3179),
    [sym__backtick_quote] = ACTIONS(3179),
    [sym__search_slash_quote] = ACTIONS(3179),
    [sym__open_readline_bracket] = ACTIONS(3179),
    [sym__open_fileglob_bracket] = ACTIONS(3179),
    [sym__PERLY_HEREDOC] = ACTIONS(3179),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3179),
    [sym__filetest] = ACTIONS(3179),
  },
  [STATE(1585)] = {
    [sym_heredoc_content] = STATE(1585),
    [sym__identifier] = ACTIONS(3129),
    [anon_sym_LBRACE] = ACTIONS(3129),
    [anon_sym_RBRACE] = ACTIONS(3129),
    [anon_sym_SEMI] = ACTIONS(3129),
    [anon_sym_package] = ACTIONS(3129),
    [anon_sym_class] = ACTIONS(3129),
    [anon_sym_role] = ACTIONS(3129),
    [anon_sym_BUILD] = ACTIONS(3129),
    [anon_sym_ADJUST] = ACTIONS(3129),
    [anon_sym_DOLLAR] = ACTIONS(3129),
    [anon_sym_AT] = ACTIONS(3129),
    [anon_sym_my] = ACTIONS(3129),
    [anon_sym_extended] = ACTIONS(3129),
    [anon_sym_async] = ACTIONS(3129),
    [anon_sym_sub] = ACTIONS(3129),
    [anon_sym_method] = ACTIONS(3129),
    [anon_sym_LPAREN] = ACTIONS(3129),
    [anon_sym_continue] = ACTIONS(3129),
    [anon_sym_state] = ACTIONS(3129),
    [anon_sym_our] = ACTIONS(3129),
    [anon_sym_try] = ACTIONS(3129),
    [anon_sym_defer] = ACTIONS(3129),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3129),
    [anon_sym_LBRACK] = ACTIONS(3129),
    [anon_sym_LT_LT] = ACTIONS(3129),
    [anon_sym_SLASH_SLASH] = ACTIONS(3129),
    [anon_sym_PLUS] = ACTIONS(3129),
    [anon_sym_DASH] = ACTIONS(3129),
    [anon_sym_TILDE] = ACTIONS(3129),
    [anon_sym_BANG] = ACTIONS(3129),
    [anon_sym_PLUS_PLUS] = ACTIONS(3129),
    [anon_sym_DASH_DASH] = ACTIONS(3129),
    [anon_sym_BSLASH] = ACTIONS(3127),
    [anon_sym_do] = ACTIONS(3129),
    [anon_sym_eval] = ACTIONS(3129),
    [anon_sym_await] = ACTIONS(3129),
    [anon_sym_field] = ACTIONS(3129),
    [anon_sym_local] = ACTIONS(3129),
    [anon_sym_dynamically] = ACTIONS(3129),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3129),
    [anon_sym_require] = ACTIONS(3129),
    [anon_sym_map] = ACTIONS(3129),
    [anon_sym_grep] = ACTIONS(3129),
    [anon_sym_sort] = ACTIONS(3129),
    [anon_sym_goto] = ACTIONS(3129),
    [anon_sym_return] = ACTIONS(3129),
    [anon_sym_undef] = ACTIONS(3129),
    [anon_sym_PERCENT2] = ACTIONS(3129),
    [anon_sym_AMP2] = ACTIONS(3129),
    [anon_sym_STAR2] = ACTIONS(3129),
    [anon_sym_use] = ACTIONS(3129),
    [anon_sym_no] = ACTIONS(3129),
    [anon_sym_for] = ACTIONS(3129),
    [anon_sym_foreach] = ACTIONS(3129),
    [anon_sym_last] = ACTIONS(3129),
    [anon_sym_next] = ACTIONS(3129),
    [anon_sym_redo] = ACTIONS(3129),
    [anon_sym_BEGIN] = ACTIONS(3129),
    [anon_sym_INIT] = ACTIONS(3129),
    [anon_sym_CHECK] = ACTIONS(3129),
    [anon_sym_UNITCHECK] = ACTIONS(3129),
    [anon_sym_END] = ACTIONS(3129),
    [anon_sym___FILE__] = ACTIONS(3129),
    [anon_sym___LINE__] = ACTIONS(3129),
    [anon_sym___PACKAGE__] = ACTIONS(3129),
    [anon_sym___SUB__] = ACTIONS(3129),
    [anon_sym_break] = ACTIONS(3129),
    [anon_sym_fork] = ACTIONS(3129),
    [anon_sym_getppid] = ACTIONS(3129),
    [anon_sym_time] = ACTIONS(3129),
    [anon_sym_times] = ACTIONS(3129),
    [anon_sym_wait] = ACTIONS(3129),
    [anon_sym_wantarray] = ACTIONS(3129),
    [anon_sym_abs] = ACTIONS(3129),
    [anon_sym_alarm] = ACTIONS(3129),
    [anon_sym_chop] = ACTIONS(3129),
    [anon_sym_chdir] = ACTIONS(3129),
    [anon_sym_close] = ACTIONS(3129),
    [anon_sym_closedir] = ACTIONS(3129),
    [anon_sym_caller] = ACTIONS(3129),
    [anon_sym_chomp] = ACTIONS(3129),
    [anon_sym_chr] = ACTIONS(3129),
    [anon_sym_cos] = ACTIONS(3129),
    [anon_sym_chroot] = ACTIONS(3129),
    [anon_sym_defined] = ACTIONS(3129),
    [anon_sym_delete] = ACTIONS(3129),
    [anon_sym_dbmclose] = ACTIONS(3129),
    [anon_sym_exists] = ACTIONS(3129),
    [anon_sym_exit] = ACTIONS(3129),
    [anon_sym_eof] = ACTIONS(3129),
    [anon_sym_exp] = ACTIONS(3129),
    [anon_sym_each] = ACTIONS(3129),
    [anon_sym_fc] = ACTIONS(3129),
    [anon_sym_fileno] = ACTIONS(3129),
    [anon_sym_gmtime] = ACTIONS(3129),
    [anon_sym_getc] = ACTIONS(3129),
    [anon_sym_getpgrp] = ACTIONS(3129),
    [anon_sym_getprotobyname] = ACTIONS(3129),
    [anon_sym_getpwname] = ACTIONS(3129),
    [anon_sym_getpwuid] = ACTIONS(3129),
    [anon_sym_getpeername] = ACTIONS(3129),
    [anon_sym_getnetbyname] = ACTIONS(3129),
    [anon_sym_getsockname] = ACTIONS(3129),
    [anon_sym_getgrnam] = ACTIONS(3129),
    [anon_sym_getgrgid] = ACTIONS(3129),
    [anon_sym_hex] = ACTIONS(3129),
    [anon_sym_int] = ACTIONS(3129),
    [anon_sym_keys] = ACTIONS(3129),
    [anon_sym_lc] = ACTIONS(3129),
    [anon_sym_lcfirst] = ACTIONS(3129),
    [anon_sym_length] = ACTIONS(3129),
    [anon_sym_localtime] = ACTIONS(3129),
    [anon_sym_log] = ACTIONS(3129),
    [anon_sym_lock] = ACTIONS(3129),
    [anon_sym_lstat] = ACTIONS(3129),
    [anon_sym_oct] = ACTIONS(3129),
    [anon_sym_ord] = ACTIONS(3129),
    [anon_sym_prototype] = ACTIONS(3129),
    [anon_sym_pop] = ACTIONS(3129),
    [anon_sym_pos] = ACTIONS(3129),
    [anon_sym_quotemeta] = ACTIONS(3129),
    [anon_sym_reset] = ACTIONS(3129),
    [anon_sym_rand] = ACTIONS(3129),
    [anon_sym_rmdir] = ACTIONS(3129),
    [anon_sym_readdir] = ACTIONS(3129),
    [anon_sym_readline] = ACTIONS(3129),
    [anon_sym_readpipe] = ACTIONS(3129),
    [anon_sym_rewinddir] = ACTIONS(3129),
    [anon_sym_readlink] = ACTIONS(3129),
    [anon_sym_ref] = ACTIONS(3129),
    [anon_sym_scalar] = ACTIONS(3129),
    [anon_sym_shift] = ACTIONS(3129),
    [anon_sym_sin] = ACTIONS(3129),
    [anon_sym_sleep] = ACTIONS(3129),
    [anon_sym_sqrt] = ACTIONS(3129),
    [anon_sym_srand] = ACTIONS(3129),
    [anon_sym_stat] = ACTIONS(3129),
    [anon_sym_study] = ACTIONS(3129),
    [anon_sym_tell] = ACTIONS(3129),
    [anon_sym_telldir] = ACTIONS(3129),
    [anon_sym_tied] = ACTIONS(3129),
    [anon_sym_uc] = ACTIONS(3129),
    [anon_sym_ucfirst] = ACTIONS(3129),
    [anon_sym_untie] = ACTIONS(3129),
    [anon_sym_umask] = ACTIONS(3129),
    [anon_sym_values] = ACTIONS(3129),
    [anon_sym_write] = ACTIONS(3129),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3129),
    [anon_sym_qq] = ACTIONS(3129),
    [anon_sym_qw] = ACTIONS(3129),
    [anon_sym_qx] = ACTIONS(3129),
    [anon_sym_qr] = ACTIONS(3129),
    [anon_sym_m] = ACTIONS(3129),
    [anon_sym_s] = ACTIONS(3129),
    [anon_sym_tr] = ACTIONS(3129),
    [anon_sym_y] = ACTIONS(3129),
    [anon_sym_if] = ACTIONS(3129),
    [anon_sym_unless] = ACTIONS(3129),
    [anon_sym_while] = ACTIONS(3129),
    [anon_sym_until] = ACTIONS(3129),
    [aux_sym__bareword_token1] = ACTIONS(3129),
    [sym_number] = ACTIONS(3129),
    [anon_sym_true] = ACTIONS(3129),
    [anon_sym_false] = ACTIONS(3129),
    [sym__single_quote] = ACTIONS(3127),
    [sym__double_quote] = ACTIONS(3127),
    [sym__backtick_quote] = ACTIONS(3127),
    [sym__search_slash_quote] = ACTIONS(3127),
    [sym__open_readline_bracket] = ACTIONS(3127),
    [sym__open_fileglob_bracket] = ACTIONS(3127),
    [sym__PERLY_HEREDOC] = ACTIONS(3127),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3127),
    [sym__filetest] = ACTIONS(3127),
  },
  [STATE(1586)] = {
    [sym_heredoc_content] = STATE(1586),
    [sym__identifier] = ACTIONS(3197),
    [anon_sym_LBRACE] = ACTIONS(3197),
    [anon_sym_RBRACE] = ACTIONS(3197),
    [anon_sym_SEMI] = ACTIONS(3197),
    [anon_sym_package] = ACTIONS(3197),
    [anon_sym_class] = ACTIONS(3197),
    [anon_sym_role] = ACTIONS(3197),
    [anon_sym_BUILD] = ACTIONS(3197),
    [anon_sym_ADJUST] = ACTIONS(3197),
    [anon_sym_DOLLAR] = ACTIONS(3197),
    [anon_sym_AT] = ACTIONS(3197),
    [anon_sym_my] = ACTIONS(3197),
    [anon_sym_extended] = ACTIONS(3197),
    [anon_sym_async] = ACTIONS(3197),
    [anon_sym_sub] = ACTIONS(3197),
    [anon_sym_method] = ACTIONS(3197),
    [anon_sym_LPAREN] = ACTIONS(3197),
    [anon_sym_continue] = ACTIONS(3197),
    [anon_sym_state] = ACTIONS(3197),
    [anon_sym_our] = ACTIONS(3197),
    [anon_sym_try] = ACTIONS(3197),
    [anon_sym_defer] = ACTIONS(3197),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3197),
    [anon_sym_LBRACK] = ACTIONS(3197),
    [anon_sym_LT_LT] = ACTIONS(3197),
    [anon_sym_SLASH_SLASH] = ACTIONS(3197),
    [anon_sym_PLUS] = ACTIONS(3197),
    [anon_sym_DASH] = ACTIONS(3197),
    [anon_sym_TILDE] = ACTIONS(3197),
    [anon_sym_BANG] = ACTIONS(3197),
    [anon_sym_PLUS_PLUS] = ACTIONS(3197),
    [anon_sym_DASH_DASH] = ACTIONS(3197),
    [anon_sym_BSLASH] = ACTIONS(3195),
    [anon_sym_do] = ACTIONS(3197),
    [anon_sym_eval] = ACTIONS(3197),
    [anon_sym_await] = ACTIONS(3197),
    [anon_sym_field] = ACTIONS(3197),
    [anon_sym_local] = ACTIONS(3197),
    [anon_sym_dynamically] = ACTIONS(3197),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3197),
    [anon_sym_require] = ACTIONS(3197),
    [anon_sym_map] = ACTIONS(3197),
    [anon_sym_grep] = ACTIONS(3197),
    [anon_sym_sort] = ACTIONS(3197),
    [anon_sym_goto] = ACTIONS(3197),
    [anon_sym_return] = ACTIONS(3197),
    [anon_sym_undef] = ACTIONS(3197),
    [anon_sym_PERCENT2] = ACTIONS(3197),
    [anon_sym_AMP2] = ACTIONS(3197),
    [anon_sym_STAR2] = ACTIONS(3197),
    [anon_sym_use] = ACTIONS(3197),
    [anon_sym_no] = ACTIONS(3197),
    [anon_sym_for] = ACTIONS(3197),
    [anon_sym_foreach] = ACTIONS(3197),
    [anon_sym_last] = ACTIONS(3197),
    [anon_sym_next] = ACTIONS(3197),
    [anon_sym_redo] = ACTIONS(3197),
    [anon_sym_BEGIN] = ACTIONS(3197),
    [anon_sym_INIT] = ACTIONS(3197),
    [anon_sym_CHECK] = ACTIONS(3197),
    [anon_sym_UNITCHECK] = ACTIONS(3197),
    [anon_sym_END] = ACTIONS(3197),
    [anon_sym___FILE__] = ACTIONS(3197),
    [anon_sym___LINE__] = ACTIONS(3197),
    [anon_sym___PACKAGE__] = ACTIONS(3197),
    [anon_sym___SUB__] = ACTIONS(3197),
    [anon_sym_break] = ACTIONS(3197),
    [anon_sym_fork] = ACTIONS(3197),
    [anon_sym_getppid] = ACTIONS(3197),
    [anon_sym_time] = ACTIONS(3197),
    [anon_sym_times] = ACTIONS(3197),
    [anon_sym_wait] = ACTIONS(3197),
    [anon_sym_wantarray] = ACTIONS(3197),
    [anon_sym_abs] = ACTIONS(3197),
    [anon_sym_alarm] = ACTIONS(3197),
    [anon_sym_chop] = ACTIONS(3197),
    [anon_sym_chdir] = ACTIONS(3197),
    [anon_sym_close] = ACTIONS(3197),
    [anon_sym_closedir] = ACTIONS(3197),
    [anon_sym_caller] = ACTIONS(3197),
    [anon_sym_chomp] = ACTIONS(3197),
    [anon_sym_chr] = ACTIONS(3197),
    [anon_sym_cos] = ACTIONS(3197),
    [anon_sym_chroot] = ACTIONS(3197),
    [anon_sym_defined] = ACTIONS(3197),
    [anon_sym_delete] = ACTIONS(3197),
    [anon_sym_dbmclose] = ACTIONS(3197),
    [anon_sym_exists] = ACTIONS(3197),
    [anon_sym_exit] = ACTIONS(3197),
    [anon_sym_eof] = ACTIONS(3197),
    [anon_sym_exp] = ACTIONS(3197),
    [anon_sym_each] = ACTIONS(3197),
    [anon_sym_fc] = ACTIONS(3197),
    [anon_sym_fileno] = ACTIONS(3197),
    [anon_sym_gmtime] = ACTIONS(3197),
    [anon_sym_getc] = ACTIONS(3197),
    [anon_sym_getpgrp] = ACTIONS(3197),
    [anon_sym_getprotobyname] = ACTIONS(3197),
    [anon_sym_getpwname] = ACTIONS(3197),
    [anon_sym_getpwuid] = ACTIONS(3197),
    [anon_sym_getpeername] = ACTIONS(3197),
    [anon_sym_getnetbyname] = ACTIONS(3197),
    [anon_sym_getsockname] = ACTIONS(3197),
    [anon_sym_getgrnam] = ACTIONS(3197),
    [anon_sym_getgrgid] = ACTIONS(3197),
    [anon_sym_hex] = ACTIONS(3197),
    [anon_sym_int] = ACTIONS(3197),
    [anon_sym_keys] = ACTIONS(3197),
    [anon_sym_lc] = ACTIONS(3197),
    [anon_sym_lcfirst] = ACTIONS(3197),
    [anon_sym_length] = ACTIONS(3197),
    [anon_sym_localtime] = ACTIONS(3197),
    [anon_sym_log] = ACTIONS(3197),
    [anon_sym_lock] = ACTIONS(3197),
    [anon_sym_lstat] = ACTIONS(3197),
    [anon_sym_oct] = ACTIONS(3197),
    [anon_sym_ord] = ACTIONS(3197),
    [anon_sym_prototype] = ACTIONS(3197),
    [anon_sym_pop] = ACTIONS(3197),
    [anon_sym_pos] = ACTIONS(3197),
    [anon_sym_quotemeta] = ACTIONS(3197),
    [anon_sym_reset] = ACTIONS(3197),
    [anon_sym_rand] = ACTIONS(3197),
    [anon_sym_rmdir] = ACTIONS(3197),
    [anon_sym_readdir] = ACTIONS(3197),
    [anon_sym_readline] = ACTIONS(3197),
    [anon_sym_readpipe] = ACTIONS(3197),
    [anon_sym_rewinddir] = ACTIONS(3197),
    [anon_sym_readlink] = ACTIONS(3197),
    [anon_sym_ref] = ACTIONS(3197),
    [anon_sym_scalar] = ACTIONS(3197),
    [anon_sym_shift] = ACTIONS(3197),
    [anon_sym_sin] = ACTIONS(3197),
    [anon_sym_sleep] = ACTIONS(3197),
    [anon_sym_sqrt] = ACTIONS(3197),
    [anon_sym_srand] = ACTIONS(3197),
    [anon_sym_stat] = ACTIONS(3197),
    [anon_sym_study] = ACTIONS(3197),
    [anon_sym_tell] = ACTIONS(3197),
    [anon_sym_telldir] = ACTIONS(3197),
    [anon_sym_tied] = ACTIONS(3197),
    [anon_sym_uc] = ACTIONS(3197),
    [anon_sym_ucfirst] = ACTIONS(3197),
    [anon_sym_untie] = ACTIONS(3197),
    [anon_sym_umask] = ACTIONS(3197),
    [anon_sym_values] = ACTIONS(3197),
    [anon_sym_write] = ACTIONS(3197),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3197),
    [anon_sym_qq] = ACTIONS(3197),
    [anon_sym_qw] = ACTIONS(3197),
    [anon_sym_qx] = ACTIONS(3197),
    [anon_sym_qr] = ACTIONS(3197),
    [anon_sym_m] = ACTIONS(3197),
    [anon_sym_s] = ACTIONS(3197),
    [anon_sym_tr] = ACTIONS(3197),
    [anon_sym_y] = ACTIONS(3197),
    [anon_sym_if] = ACTIONS(3197),
    [anon_sym_unless] = ACTIONS(3197),
    [anon_sym_while] = ACTIONS(3197),
    [anon_sym_until] = ACTIONS(3197),
    [aux_sym__bareword_token1] = ACTIONS(3197),
    [sym_number] = ACTIONS(3197),
    [anon_sym_true] = ACTIONS(3197),
    [anon_sym_false] = ACTIONS(3197),
    [sym__single_quote] = ACTIONS(3195),
    [sym__double_quote] = ACTIONS(3195),
    [sym__backtick_quote] = ACTIONS(3195),
    [sym__search_slash_quote] = ACTIONS(3195),
    [sym__open_readline_bracket] = ACTIONS(3195),
    [sym__open_fileglob_bracket] = ACTIONS(3195),
    [sym__PERLY_HEREDOC] = ACTIONS(3195),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3195),
    [sym__filetest] = ACTIONS(3195),
  },
  [STATE(1587)] = {
    [sym_heredoc_content] = STATE(1587),
    [sym__identifier] = ACTIONS(3209),
    [anon_sym_LBRACE] = ACTIONS(3209),
    [anon_sym_RBRACE] = ACTIONS(3209),
    [anon_sym_SEMI] = ACTIONS(3209),
    [anon_sym_package] = ACTIONS(3209),
    [anon_sym_class] = ACTIONS(3209),
    [anon_sym_role] = ACTIONS(3209),
    [anon_sym_BUILD] = ACTIONS(3209),
    [anon_sym_ADJUST] = ACTIONS(3209),
    [anon_sym_DOLLAR] = ACTIONS(3209),
    [anon_sym_AT] = ACTIONS(3209),
    [anon_sym_my] = ACTIONS(3209),
    [anon_sym_extended] = ACTIONS(3209),
    [anon_sym_async] = ACTIONS(3209),
    [anon_sym_sub] = ACTIONS(3209),
    [anon_sym_method] = ACTIONS(3209),
    [anon_sym_LPAREN] = ACTIONS(3209),
    [anon_sym_continue] = ACTIONS(3209),
    [anon_sym_state] = ACTIONS(3209),
    [anon_sym_our] = ACTIONS(3209),
    [anon_sym_try] = ACTIONS(3209),
    [anon_sym_defer] = ACTIONS(3209),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3209),
    [anon_sym_LBRACK] = ACTIONS(3209),
    [anon_sym_LT_LT] = ACTIONS(3209),
    [anon_sym_SLASH_SLASH] = ACTIONS(3209),
    [anon_sym_PLUS] = ACTIONS(3209),
    [anon_sym_DASH] = ACTIONS(3209),
    [anon_sym_TILDE] = ACTIONS(3209),
    [anon_sym_BANG] = ACTIONS(3209),
    [anon_sym_PLUS_PLUS] = ACTIONS(3209),
    [anon_sym_DASH_DASH] = ACTIONS(3209),
    [anon_sym_BSLASH] = ACTIONS(3207),
    [anon_sym_do] = ACTIONS(3209),
    [anon_sym_eval] = ACTIONS(3209),
    [anon_sym_await] = ACTIONS(3209),
    [anon_sym_field] = ACTIONS(3209),
    [anon_sym_local] = ACTIONS(3209),
    [anon_sym_dynamically] = ACTIONS(3209),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3209),
    [anon_sym_require] = ACTIONS(3209),
    [anon_sym_map] = ACTIONS(3209),
    [anon_sym_grep] = ACTIONS(3209),
    [anon_sym_sort] = ACTIONS(3209),
    [anon_sym_goto] = ACTIONS(3209),
    [anon_sym_return] = ACTIONS(3209),
    [anon_sym_undef] = ACTIONS(3209),
    [anon_sym_PERCENT2] = ACTIONS(3209),
    [anon_sym_AMP2] = ACTIONS(3209),
    [anon_sym_STAR2] = ACTIONS(3209),
    [anon_sym_use] = ACTIONS(3209),
    [anon_sym_no] = ACTIONS(3209),
    [anon_sym_for] = ACTIONS(3209),
    [anon_sym_foreach] = ACTIONS(3209),
    [anon_sym_last] = ACTIONS(3209),
    [anon_sym_next] = ACTIONS(3209),
    [anon_sym_redo] = ACTIONS(3209),
    [anon_sym_BEGIN] = ACTIONS(3209),
    [anon_sym_INIT] = ACTIONS(3209),
    [anon_sym_CHECK] = ACTIONS(3209),
    [anon_sym_UNITCHECK] = ACTIONS(3209),
    [anon_sym_END] = ACTIONS(3209),
    [anon_sym___FILE__] = ACTIONS(3209),
    [anon_sym___LINE__] = ACTIONS(3209),
    [anon_sym___PACKAGE__] = ACTIONS(3209),
    [anon_sym___SUB__] = ACTIONS(3209),
    [anon_sym_break] = ACTIONS(3209),
    [anon_sym_fork] = ACTIONS(3209),
    [anon_sym_getppid] = ACTIONS(3209),
    [anon_sym_time] = ACTIONS(3209),
    [anon_sym_times] = ACTIONS(3209),
    [anon_sym_wait] = ACTIONS(3209),
    [anon_sym_wantarray] = ACTIONS(3209),
    [anon_sym_abs] = ACTIONS(3209),
    [anon_sym_alarm] = ACTIONS(3209),
    [anon_sym_chop] = ACTIONS(3209),
    [anon_sym_chdir] = ACTIONS(3209),
    [anon_sym_close] = ACTIONS(3209),
    [anon_sym_closedir] = ACTIONS(3209),
    [anon_sym_caller] = ACTIONS(3209),
    [anon_sym_chomp] = ACTIONS(3209),
    [anon_sym_chr] = ACTIONS(3209),
    [anon_sym_cos] = ACTIONS(3209),
    [anon_sym_chroot] = ACTIONS(3209),
    [anon_sym_defined] = ACTIONS(3209),
    [anon_sym_delete] = ACTIONS(3209),
    [anon_sym_dbmclose] = ACTIONS(3209),
    [anon_sym_exists] = ACTIONS(3209),
    [anon_sym_exit] = ACTIONS(3209),
    [anon_sym_eof] = ACTIONS(3209),
    [anon_sym_exp] = ACTIONS(3209),
    [anon_sym_each] = ACTIONS(3209),
    [anon_sym_fc] = ACTIONS(3209),
    [anon_sym_fileno] = ACTIONS(3209),
    [anon_sym_gmtime] = ACTIONS(3209),
    [anon_sym_getc] = ACTIONS(3209),
    [anon_sym_getpgrp] = ACTIONS(3209),
    [anon_sym_getprotobyname] = ACTIONS(3209),
    [anon_sym_getpwname] = ACTIONS(3209),
    [anon_sym_getpwuid] = ACTIONS(3209),
    [anon_sym_getpeername] = ACTIONS(3209),
    [anon_sym_getnetbyname] = ACTIONS(3209),
    [anon_sym_getsockname] = ACTIONS(3209),
    [anon_sym_getgrnam] = ACTIONS(3209),
    [anon_sym_getgrgid] = ACTIONS(3209),
    [anon_sym_hex] = ACTIONS(3209),
    [anon_sym_int] = ACTIONS(3209),
    [anon_sym_keys] = ACTIONS(3209),
    [anon_sym_lc] = ACTIONS(3209),
    [anon_sym_lcfirst] = ACTIONS(3209),
    [anon_sym_length] = ACTIONS(3209),
    [anon_sym_localtime] = ACTIONS(3209),
    [anon_sym_log] = ACTIONS(3209),
    [anon_sym_lock] = ACTIONS(3209),
    [anon_sym_lstat] = ACTIONS(3209),
    [anon_sym_oct] = ACTIONS(3209),
    [anon_sym_ord] = ACTIONS(3209),
    [anon_sym_prototype] = ACTIONS(3209),
    [anon_sym_pop] = ACTIONS(3209),
    [anon_sym_pos] = ACTIONS(3209),
    [anon_sym_quotemeta] = ACTIONS(3209),
    [anon_sym_reset] = ACTIONS(3209),
    [anon_sym_rand] = ACTIONS(3209),
    [anon_sym_rmdir] = ACTIONS(3209),
    [anon_sym_readdir] = ACTIONS(3209),
    [anon_sym_readline] = ACTIONS(3209),
    [anon_sym_readpipe] = ACTIONS(3209),
    [anon_sym_rewinddir] = ACTIONS(3209),
    [anon_sym_readlink] = ACTIONS(3209),
    [anon_sym_ref] = ACTIONS(3209),
    [anon_sym_scalar] = ACTIONS(3209),
    [anon_sym_shift] = ACTIONS(3209),
    [anon_sym_sin] = ACTIONS(3209),
    [anon_sym_sleep] = ACTIONS(3209),
    [anon_sym_sqrt] = ACTIONS(3209),
    [anon_sym_srand] = ACTIONS(3209),
    [anon_sym_stat] = ACTIONS(3209),
    [anon_sym_study] = ACTIONS(3209),
    [anon_sym_tell] = ACTIONS(3209),
    [anon_sym_telldir] = ACTIONS(3209),
    [anon_sym_tied] = ACTIONS(3209),
    [anon_sym_uc] = ACTIONS(3209),
    [anon_sym_ucfirst] = ACTIONS(3209),
    [anon_sym_untie] = ACTIONS(3209),
    [anon_sym_umask] = ACTIONS(3209),
    [anon_sym_values] = ACTIONS(3209),
    [anon_sym_write] = ACTIONS(3209),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3209),
    [anon_sym_qq] = ACTIONS(3209),
    [anon_sym_qw] = ACTIONS(3209),
    [anon_sym_qx] = ACTIONS(3209),
    [anon_sym_qr] = ACTIONS(3209),
    [anon_sym_m] = ACTIONS(3209),
    [anon_sym_s] = ACTIONS(3209),
    [anon_sym_tr] = ACTIONS(3209),
    [anon_sym_y] = ACTIONS(3209),
    [anon_sym_if] = ACTIONS(3209),
    [anon_sym_unless] = ACTIONS(3209),
    [anon_sym_while] = ACTIONS(3209),
    [anon_sym_until] = ACTIONS(3209),
    [aux_sym__bareword_token1] = ACTIONS(3209),
    [sym_number] = ACTIONS(3209),
    [anon_sym_true] = ACTIONS(3209),
    [anon_sym_false] = ACTIONS(3209),
    [sym__single_quote] = ACTIONS(3207),
    [sym__double_quote] = ACTIONS(3207),
    [sym__backtick_quote] = ACTIONS(3207),
    [sym__search_slash_quote] = ACTIONS(3207),
    [sym__open_readline_bracket] = ACTIONS(3207),
    [sym__open_fileglob_bracket] = ACTIONS(3207),
    [sym__PERLY_HEREDOC] = ACTIONS(3207),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3207),
    [sym__filetest] = ACTIONS(3207),
  },
  [STATE(1588)] = {
    [sym_heredoc_content] = STATE(1588),
    [sym__identifier] = ACTIONS(3213),
    [anon_sym_LBRACE] = ACTIONS(3213),
    [anon_sym_RBRACE] = ACTIONS(3213),
    [anon_sym_SEMI] = ACTIONS(3213),
    [anon_sym_package] = ACTIONS(3213),
    [anon_sym_class] = ACTIONS(3213),
    [anon_sym_role] = ACTIONS(3213),
    [anon_sym_BUILD] = ACTIONS(3213),
    [anon_sym_ADJUST] = ACTIONS(3213),
    [anon_sym_DOLLAR] = ACTIONS(3213),
    [anon_sym_AT] = ACTIONS(3213),
    [anon_sym_my] = ACTIONS(3213),
    [anon_sym_extended] = ACTIONS(3213),
    [anon_sym_async] = ACTIONS(3213),
    [anon_sym_sub] = ACTIONS(3213),
    [anon_sym_method] = ACTIONS(3213),
    [anon_sym_LPAREN] = ACTIONS(3213),
    [anon_sym_continue] = ACTIONS(3213),
    [anon_sym_state] = ACTIONS(3213),
    [anon_sym_our] = ACTIONS(3213),
    [anon_sym_try] = ACTIONS(3213),
    [anon_sym_defer] = ACTIONS(3213),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3213),
    [anon_sym_LBRACK] = ACTIONS(3213),
    [anon_sym_LT_LT] = ACTIONS(3213),
    [anon_sym_SLASH_SLASH] = ACTIONS(3213),
    [anon_sym_PLUS] = ACTIONS(3213),
    [anon_sym_DASH] = ACTIONS(3213),
    [anon_sym_TILDE] = ACTIONS(3213),
    [anon_sym_BANG] = ACTIONS(3213),
    [anon_sym_PLUS_PLUS] = ACTIONS(3213),
    [anon_sym_DASH_DASH] = ACTIONS(3213),
    [anon_sym_BSLASH] = ACTIONS(3211),
    [anon_sym_do] = ACTIONS(3213),
    [anon_sym_eval] = ACTIONS(3213),
    [anon_sym_await] = ACTIONS(3213),
    [anon_sym_field] = ACTIONS(3213),
    [anon_sym_local] = ACTIONS(3213),
    [anon_sym_dynamically] = ACTIONS(3213),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3213),
    [anon_sym_require] = ACTIONS(3213),
    [anon_sym_map] = ACTIONS(3213),
    [anon_sym_grep] = ACTIONS(3213),
    [anon_sym_sort] = ACTIONS(3213),
    [anon_sym_goto] = ACTIONS(3213),
    [anon_sym_return] = ACTIONS(3213),
    [anon_sym_undef] = ACTIONS(3213),
    [anon_sym_PERCENT2] = ACTIONS(3213),
    [anon_sym_AMP2] = ACTIONS(3213),
    [anon_sym_STAR2] = ACTIONS(3213),
    [anon_sym_use] = ACTIONS(3213),
    [anon_sym_no] = ACTIONS(3213),
    [anon_sym_for] = ACTIONS(3213),
    [anon_sym_foreach] = ACTIONS(3213),
    [anon_sym_last] = ACTIONS(3213),
    [anon_sym_next] = ACTIONS(3213),
    [anon_sym_redo] = ACTIONS(3213),
    [anon_sym_BEGIN] = ACTIONS(3213),
    [anon_sym_INIT] = ACTIONS(3213),
    [anon_sym_CHECK] = ACTIONS(3213),
    [anon_sym_UNITCHECK] = ACTIONS(3213),
    [anon_sym_END] = ACTIONS(3213),
    [anon_sym___FILE__] = ACTIONS(3213),
    [anon_sym___LINE__] = ACTIONS(3213),
    [anon_sym___PACKAGE__] = ACTIONS(3213),
    [anon_sym___SUB__] = ACTIONS(3213),
    [anon_sym_break] = ACTIONS(3213),
    [anon_sym_fork] = ACTIONS(3213),
    [anon_sym_getppid] = ACTIONS(3213),
    [anon_sym_time] = ACTIONS(3213),
    [anon_sym_times] = ACTIONS(3213),
    [anon_sym_wait] = ACTIONS(3213),
    [anon_sym_wantarray] = ACTIONS(3213),
    [anon_sym_abs] = ACTIONS(3213),
    [anon_sym_alarm] = ACTIONS(3213),
    [anon_sym_chop] = ACTIONS(3213),
    [anon_sym_chdir] = ACTIONS(3213),
    [anon_sym_close] = ACTIONS(3213),
    [anon_sym_closedir] = ACTIONS(3213),
    [anon_sym_caller] = ACTIONS(3213),
    [anon_sym_chomp] = ACTIONS(3213),
    [anon_sym_chr] = ACTIONS(3213),
    [anon_sym_cos] = ACTIONS(3213),
    [anon_sym_chroot] = ACTIONS(3213),
    [anon_sym_defined] = ACTIONS(3213),
    [anon_sym_delete] = ACTIONS(3213),
    [anon_sym_dbmclose] = ACTIONS(3213),
    [anon_sym_exists] = ACTIONS(3213),
    [anon_sym_exit] = ACTIONS(3213),
    [anon_sym_eof] = ACTIONS(3213),
    [anon_sym_exp] = ACTIONS(3213),
    [anon_sym_each] = ACTIONS(3213),
    [anon_sym_fc] = ACTIONS(3213),
    [anon_sym_fileno] = ACTIONS(3213),
    [anon_sym_gmtime] = ACTIONS(3213),
    [anon_sym_getc] = ACTIONS(3213),
    [anon_sym_getpgrp] = ACTIONS(3213),
    [anon_sym_getprotobyname] = ACTIONS(3213),
    [anon_sym_getpwname] = ACTIONS(3213),
    [anon_sym_getpwuid] = ACTIONS(3213),
    [anon_sym_getpeername] = ACTIONS(3213),
    [anon_sym_getnetbyname] = ACTIONS(3213),
    [anon_sym_getsockname] = ACTIONS(3213),
    [anon_sym_getgrnam] = ACTIONS(3213),
    [anon_sym_getgrgid] = ACTIONS(3213),
    [anon_sym_hex] = ACTIONS(3213),
    [anon_sym_int] = ACTIONS(3213),
    [anon_sym_keys] = ACTIONS(3213),
    [anon_sym_lc] = ACTIONS(3213),
    [anon_sym_lcfirst] = ACTIONS(3213),
    [anon_sym_length] = ACTIONS(3213),
    [anon_sym_localtime] = ACTIONS(3213),
    [anon_sym_log] = ACTIONS(3213),
    [anon_sym_lock] = ACTIONS(3213),
    [anon_sym_lstat] = ACTIONS(3213),
    [anon_sym_oct] = ACTIONS(3213),
    [anon_sym_ord] = ACTIONS(3213),
    [anon_sym_prototype] = ACTIONS(3213),
    [anon_sym_pop] = ACTIONS(3213),
    [anon_sym_pos] = ACTIONS(3213),
    [anon_sym_quotemeta] = ACTIONS(3213),
    [anon_sym_reset] = ACTIONS(3213),
    [anon_sym_rand] = ACTIONS(3213),
    [anon_sym_rmdir] = ACTIONS(3213),
    [anon_sym_readdir] = ACTIONS(3213),
    [anon_sym_readline] = ACTIONS(3213),
    [anon_sym_readpipe] = ACTIONS(3213),
    [anon_sym_rewinddir] = ACTIONS(3213),
    [anon_sym_readlink] = ACTIONS(3213),
    [anon_sym_ref] = ACTIONS(3213),
    [anon_sym_scalar] = ACTIONS(3213),
    [anon_sym_shift] = ACTIONS(3213),
    [anon_sym_sin] = ACTIONS(3213),
    [anon_sym_sleep] = ACTIONS(3213),
    [anon_sym_sqrt] = ACTIONS(3213),
    [anon_sym_srand] = ACTIONS(3213),
    [anon_sym_stat] = ACTIONS(3213),
    [anon_sym_study] = ACTIONS(3213),
    [anon_sym_tell] = ACTIONS(3213),
    [anon_sym_telldir] = ACTIONS(3213),
    [anon_sym_tied] = ACTIONS(3213),
    [anon_sym_uc] = ACTIONS(3213),
    [anon_sym_ucfirst] = ACTIONS(3213),
    [anon_sym_untie] = ACTIONS(3213),
    [anon_sym_umask] = ACTIONS(3213),
    [anon_sym_values] = ACTIONS(3213),
    [anon_sym_write] = ACTIONS(3213),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3213),
    [anon_sym_qq] = ACTIONS(3213),
    [anon_sym_qw] = ACTIONS(3213),
    [anon_sym_qx] = ACTIONS(3213),
    [anon_sym_qr] = ACTIONS(3213),
    [anon_sym_m] = ACTIONS(3213),
    [anon_sym_s] = ACTIONS(3213),
    [anon_sym_tr] = ACTIONS(3213),
    [anon_sym_y] = ACTIONS(3213),
    [anon_sym_if] = ACTIONS(3213),
    [anon_sym_unless] = ACTIONS(3213),
    [anon_sym_while] = ACTIONS(3213),
    [anon_sym_until] = ACTIONS(3213),
    [aux_sym__bareword_token1] = ACTIONS(3213),
    [sym_number] = ACTIONS(3213),
    [anon_sym_true] = ACTIONS(3213),
    [anon_sym_false] = ACTIONS(3213),
    [sym__single_quote] = ACTIONS(3211),
    [sym__double_quote] = ACTIONS(3211),
    [sym__backtick_quote] = ACTIONS(3211),
    [sym__search_slash_quote] = ACTIONS(3211),
    [sym__open_readline_bracket] = ACTIONS(3211),
    [sym__open_fileglob_bracket] = ACTIONS(3211),
    [sym__PERLY_HEREDOC] = ACTIONS(3211),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3211),
    [sym__filetest] = ACTIONS(3211),
  },
  [STATE(1589)] = {
    [sym_heredoc_content] = STATE(1589),
    [sym__identifier] = ACTIONS(3217),
    [anon_sym_LBRACE] = ACTIONS(3217),
    [anon_sym_RBRACE] = ACTIONS(3217),
    [anon_sym_SEMI] = ACTIONS(3217),
    [anon_sym_package] = ACTIONS(3217),
    [anon_sym_class] = ACTIONS(3217),
    [anon_sym_role] = ACTIONS(3217),
    [anon_sym_BUILD] = ACTIONS(3217),
    [anon_sym_ADJUST] = ACTIONS(3217),
    [anon_sym_DOLLAR] = ACTIONS(3217),
    [anon_sym_AT] = ACTIONS(3217),
    [anon_sym_my] = ACTIONS(3217),
    [anon_sym_extended] = ACTIONS(3217),
    [anon_sym_async] = ACTIONS(3217),
    [anon_sym_sub] = ACTIONS(3217),
    [anon_sym_method] = ACTIONS(3217),
    [anon_sym_LPAREN] = ACTIONS(3217),
    [anon_sym_continue] = ACTIONS(3217),
    [anon_sym_state] = ACTIONS(3217),
    [anon_sym_our] = ACTIONS(3217),
    [anon_sym_try] = ACTIONS(3217),
    [anon_sym_defer] = ACTIONS(3217),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3217),
    [anon_sym_LBRACK] = ACTIONS(3217),
    [anon_sym_LT_LT] = ACTIONS(3217),
    [anon_sym_SLASH_SLASH] = ACTIONS(3217),
    [anon_sym_PLUS] = ACTIONS(3217),
    [anon_sym_DASH] = ACTIONS(3217),
    [anon_sym_TILDE] = ACTIONS(3217),
    [anon_sym_BANG] = ACTIONS(3217),
    [anon_sym_PLUS_PLUS] = ACTIONS(3217),
    [anon_sym_DASH_DASH] = ACTIONS(3217),
    [anon_sym_BSLASH] = ACTIONS(3215),
    [anon_sym_do] = ACTIONS(3217),
    [anon_sym_eval] = ACTIONS(3217),
    [anon_sym_await] = ACTIONS(3217),
    [anon_sym_field] = ACTIONS(3217),
    [anon_sym_local] = ACTIONS(3217),
    [anon_sym_dynamically] = ACTIONS(3217),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3217),
    [anon_sym_require] = ACTIONS(3217),
    [anon_sym_map] = ACTIONS(3217),
    [anon_sym_grep] = ACTIONS(3217),
    [anon_sym_sort] = ACTIONS(3217),
    [anon_sym_goto] = ACTIONS(3217),
    [anon_sym_return] = ACTIONS(3217),
    [anon_sym_undef] = ACTIONS(3217),
    [anon_sym_PERCENT2] = ACTIONS(3217),
    [anon_sym_AMP2] = ACTIONS(3217),
    [anon_sym_STAR2] = ACTIONS(3217),
    [anon_sym_use] = ACTIONS(3217),
    [anon_sym_no] = ACTIONS(3217),
    [anon_sym_for] = ACTIONS(3217),
    [anon_sym_foreach] = ACTIONS(3217),
    [anon_sym_last] = ACTIONS(3217),
    [anon_sym_next] = ACTIONS(3217),
    [anon_sym_redo] = ACTIONS(3217),
    [anon_sym_BEGIN] = ACTIONS(3217),
    [anon_sym_INIT] = ACTIONS(3217),
    [anon_sym_CHECK] = ACTIONS(3217),
    [anon_sym_UNITCHECK] = ACTIONS(3217),
    [anon_sym_END] = ACTIONS(3217),
    [anon_sym___FILE__] = ACTIONS(3217),
    [anon_sym___LINE__] = ACTIONS(3217),
    [anon_sym___PACKAGE__] = ACTIONS(3217),
    [anon_sym___SUB__] = ACTIONS(3217),
    [anon_sym_break] = ACTIONS(3217),
    [anon_sym_fork] = ACTIONS(3217),
    [anon_sym_getppid] = ACTIONS(3217),
    [anon_sym_time] = ACTIONS(3217),
    [anon_sym_times] = ACTIONS(3217),
    [anon_sym_wait] = ACTIONS(3217),
    [anon_sym_wantarray] = ACTIONS(3217),
    [anon_sym_abs] = ACTIONS(3217),
    [anon_sym_alarm] = ACTIONS(3217),
    [anon_sym_chop] = ACTIONS(3217),
    [anon_sym_chdir] = ACTIONS(3217),
    [anon_sym_close] = ACTIONS(3217),
    [anon_sym_closedir] = ACTIONS(3217),
    [anon_sym_caller] = ACTIONS(3217),
    [anon_sym_chomp] = ACTIONS(3217),
    [anon_sym_chr] = ACTIONS(3217),
    [anon_sym_cos] = ACTIONS(3217),
    [anon_sym_chroot] = ACTIONS(3217),
    [anon_sym_defined] = ACTIONS(3217),
    [anon_sym_delete] = ACTIONS(3217),
    [anon_sym_dbmclose] = ACTIONS(3217),
    [anon_sym_exists] = ACTIONS(3217),
    [anon_sym_exit] = ACTIONS(3217),
    [anon_sym_eof] = ACTIONS(3217),
    [anon_sym_exp] = ACTIONS(3217),
    [anon_sym_each] = ACTIONS(3217),
    [anon_sym_fc] = ACTIONS(3217),
    [anon_sym_fileno] = ACTIONS(3217),
    [anon_sym_gmtime] = ACTIONS(3217),
    [anon_sym_getc] = ACTIONS(3217),
    [anon_sym_getpgrp] = ACTIONS(3217),
    [anon_sym_getprotobyname] = ACTIONS(3217),
    [anon_sym_getpwname] = ACTIONS(3217),
    [anon_sym_getpwuid] = ACTIONS(3217),
    [anon_sym_getpeername] = ACTIONS(3217),
    [anon_sym_getnetbyname] = ACTIONS(3217),
    [anon_sym_getsockname] = ACTIONS(3217),
    [anon_sym_getgrnam] = ACTIONS(3217),
    [anon_sym_getgrgid] = ACTIONS(3217),
    [anon_sym_hex] = ACTIONS(3217),
    [anon_sym_int] = ACTIONS(3217),
    [anon_sym_keys] = ACTIONS(3217),
    [anon_sym_lc] = ACTIONS(3217),
    [anon_sym_lcfirst] = ACTIONS(3217),
    [anon_sym_length] = ACTIONS(3217),
    [anon_sym_localtime] = ACTIONS(3217),
    [anon_sym_log] = ACTIONS(3217),
    [anon_sym_lock] = ACTIONS(3217),
    [anon_sym_lstat] = ACTIONS(3217),
    [anon_sym_oct] = ACTIONS(3217),
    [anon_sym_ord] = ACTIONS(3217),
    [anon_sym_prototype] = ACTIONS(3217),
    [anon_sym_pop] = ACTIONS(3217),
    [anon_sym_pos] = ACTIONS(3217),
    [anon_sym_quotemeta] = ACTIONS(3217),
    [anon_sym_reset] = ACTIONS(3217),
    [anon_sym_rand] = ACTIONS(3217),
    [anon_sym_rmdir] = ACTIONS(3217),
    [anon_sym_readdir] = ACTIONS(3217),
    [anon_sym_readline] = ACTIONS(3217),
    [anon_sym_readpipe] = ACTIONS(3217),
    [anon_sym_rewinddir] = ACTIONS(3217),
    [anon_sym_readlink] = ACTIONS(3217),
    [anon_sym_ref] = ACTIONS(3217),
    [anon_sym_scalar] = ACTIONS(3217),
    [anon_sym_shift] = ACTIONS(3217),
    [anon_sym_sin] = ACTIONS(3217),
    [anon_sym_sleep] = ACTIONS(3217),
    [anon_sym_sqrt] = ACTIONS(3217),
    [anon_sym_srand] = ACTIONS(3217),
    [anon_sym_stat] = ACTIONS(3217),
    [anon_sym_study] = ACTIONS(3217),
    [anon_sym_tell] = ACTIONS(3217),
    [anon_sym_telldir] = ACTIONS(3217),
    [anon_sym_tied] = ACTIONS(3217),
    [anon_sym_uc] = ACTIONS(3217),
    [anon_sym_ucfirst] = ACTIONS(3217),
    [anon_sym_untie] = ACTIONS(3217),
    [anon_sym_umask] = ACTIONS(3217),
    [anon_sym_values] = ACTIONS(3217),
    [anon_sym_write] = ACTIONS(3217),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3217),
    [anon_sym_qq] = ACTIONS(3217),
    [anon_sym_qw] = ACTIONS(3217),
    [anon_sym_qx] = ACTIONS(3217),
    [anon_sym_qr] = ACTIONS(3217),
    [anon_sym_m] = ACTIONS(3217),
    [anon_sym_s] = ACTIONS(3217),
    [anon_sym_tr] = ACTIONS(3217),
    [anon_sym_y] = ACTIONS(3217),
    [anon_sym_if] = ACTIONS(3217),
    [anon_sym_unless] = ACTIONS(3217),
    [anon_sym_while] = ACTIONS(3217),
    [anon_sym_until] = ACTIONS(3217),
    [aux_sym__bareword_token1] = ACTIONS(3217),
    [sym_number] = ACTIONS(3217),
    [anon_sym_true] = ACTIONS(3217),
    [anon_sym_false] = ACTIONS(3217),
    [sym__single_quote] = ACTIONS(3215),
    [sym__double_quote] = ACTIONS(3215),
    [sym__backtick_quote] = ACTIONS(3215),
    [sym__search_slash_quote] = ACTIONS(3215),
    [sym__open_readline_bracket] = ACTIONS(3215),
    [sym__open_fileglob_bracket] = ACTIONS(3215),
    [sym__PERLY_HEREDOC] = ACTIONS(3215),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3215),
    [sym__filetest] = ACTIONS(3215),
  },
  [STATE(1590)] = {
    [sym_heredoc_content] = STATE(1590),
    [sym__identifier] = ACTIONS(3029),
    [anon_sym_LBRACE] = ACTIONS(3029),
    [anon_sym_RBRACE] = ACTIONS(3029),
    [anon_sym_SEMI] = ACTIONS(3029),
    [anon_sym_package] = ACTIONS(3029),
    [anon_sym_class] = ACTIONS(3029),
    [anon_sym_role] = ACTIONS(3029),
    [anon_sym_BUILD] = ACTIONS(3029),
    [anon_sym_ADJUST] = ACTIONS(3029),
    [anon_sym_DOLLAR] = ACTIONS(3029),
    [anon_sym_AT] = ACTIONS(3029),
    [anon_sym_my] = ACTIONS(3029),
    [anon_sym_extended] = ACTIONS(3029),
    [anon_sym_async] = ACTIONS(3029),
    [anon_sym_sub] = ACTIONS(3029),
    [anon_sym_method] = ACTIONS(3029),
    [anon_sym_LPAREN] = ACTIONS(3029),
    [anon_sym_continue] = ACTIONS(3029),
    [anon_sym_state] = ACTIONS(3029),
    [anon_sym_our] = ACTIONS(3029),
    [anon_sym_try] = ACTIONS(3029),
    [anon_sym_defer] = ACTIONS(3029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3029),
    [anon_sym_LBRACK] = ACTIONS(3029),
    [anon_sym_LT_LT] = ACTIONS(3029),
    [anon_sym_SLASH_SLASH] = ACTIONS(3029),
    [anon_sym_PLUS] = ACTIONS(3029),
    [anon_sym_DASH] = ACTIONS(3029),
    [anon_sym_TILDE] = ACTIONS(3029),
    [anon_sym_BANG] = ACTIONS(3029),
    [anon_sym_PLUS_PLUS] = ACTIONS(3029),
    [anon_sym_DASH_DASH] = ACTIONS(3029),
    [anon_sym_BSLASH] = ACTIONS(3027),
    [anon_sym_do] = ACTIONS(3029),
    [anon_sym_eval] = ACTIONS(3029),
    [anon_sym_await] = ACTIONS(3029),
    [anon_sym_field] = ACTIONS(3029),
    [anon_sym_local] = ACTIONS(3029),
    [anon_sym_dynamically] = ACTIONS(3029),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3029),
    [anon_sym_require] = ACTIONS(3029),
    [anon_sym_map] = ACTIONS(3029),
    [anon_sym_grep] = ACTIONS(3029),
    [anon_sym_sort] = ACTIONS(3029),
    [anon_sym_goto] = ACTIONS(3029),
    [anon_sym_return] = ACTIONS(3029),
    [anon_sym_undef] = ACTIONS(3029),
    [anon_sym_PERCENT2] = ACTIONS(3029),
    [anon_sym_AMP2] = ACTIONS(3029),
    [anon_sym_STAR2] = ACTIONS(3029),
    [anon_sym_use] = ACTIONS(3029),
    [anon_sym_no] = ACTIONS(3029),
    [anon_sym_for] = ACTIONS(3029),
    [anon_sym_foreach] = ACTIONS(3029),
    [anon_sym_last] = ACTIONS(3029),
    [anon_sym_next] = ACTIONS(3029),
    [anon_sym_redo] = ACTIONS(3029),
    [anon_sym_BEGIN] = ACTIONS(3029),
    [anon_sym_INIT] = ACTIONS(3029),
    [anon_sym_CHECK] = ACTIONS(3029),
    [anon_sym_UNITCHECK] = ACTIONS(3029),
    [anon_sym_END] = ACTIONS(3029),
    [anon_sym___FILE__] = ACTIONS(3029),
    [anon_sym___LINE__] = ACTIONS(3029),
    [anon_sym___PACKAGE__] = ACTIONS(3029),
    [anon_sym___SUB__] = ACTIONS(3029),
    [anon_sym_break] = ACTIONS(3029),
    [anon_sym_fork] = ACTIONS(3029),
    [anon_sym_getppid] = ACTIONS(3029),
    [anon_sym_time] = ACTIONS(3029),
    [anon_sym_times] = ACTIONS(3029),
    [anon_sym_wait] = ACTIONS(3029),
    [anon_sym_wantarray] = ACTIONS(3029),
    [anon_sym_abs] = ACTIONS(3029),
    [anon_sym_alarm] = ACTIONS(3029),
    [anon_sym_chop] = ACTIONS(3029),
    [anon_sym_chdir] = ACTIONS(3029),
    [anon_sym_close] = ACTIONS(3029),
    [anon_sym_closedir] = ACTIONS(3029),
    [anon_sym_caller] = ACTIONS(3029),
    [anon_sym_chomp] = ACTIONS(3029),
    [anon_sym_chr] = ACTIONS(3029),
    [anon_sym_cos] = ACTIONS(3029),
    [anon_sym_chroot] = ACTIONS(3029),
    [anon_sym_defined] = ACTIONS(3029),
    [anon_sym_delete] = ACTIONS(3029),
    [anon_sym_dbmclose] = ACTIONS(3029),
    [anon_sym_exists] = ACTIONS(3029),
    [anon_sym_exit] = ACTIONS(3029),
    [anon_sym_eof] = ACTIONS(3029),
    [anon_sym_exp] = ACTIONS(3029),
    [anon_sym_each] = ACTIONS(3029),
    [anon_sym_fc] = ACTIONS(3029),
    [anon_sym_fileno] = ACTIONS(3029),
    [anon_sym_gmtime] = ACTIONS(3029),
    [anon_sym_getc] = ACTIONS(3029),
    [anon_sym_getpgrp] = ACTIONS(3029),
    [anon_sym_getprotobyname] = ACTIONS(3029),
    [anon_sym_getpwname] = ACTIONS(3029),
    [anon_sym_getpwuid] = ACTIONS(3029),
    [anon_sym_getpeername] = ACTIONS(3029),
    [anon_sym_getnetbyname] = ACTIONS(3029),
    [anon_sym_getsockname] = ACTIONS(3029),
    [anon_sym_getgrnam] = ACTIONS(3029),
    [anon_sym_getgrgid] = ACTIONS(3029),
    [anon_sym_hex] = ACTIONS(3029),
    [anon_sym_int] = ACTIONS(3029),
    [anon_sym_keys] = ACTIONS(3029),
    [anon_sym_lc] = ACTIONS(3029),
    [anon_sym_lcfirst] = ACTIONS(3029),
    [anon_sym_length] = ACTIONS(3029),
    [anon_sym_localtime] = ACTIONS(3029),
    [anon_sym_log] = ACTIONS(3029),
    [anon_sym_lock] = ACTIONS(3029),
    [anon_sym_lstat] = ACTIONS(3029),
    [anon_sym_oct] = ACTIONS(3029),
    [anon_sym_ord] = ACTIONS(3029),
    [anon_sym_prototype] = ACTIONS(3029),
    [anon_sym_pop] = ACTIONS(3029),
    [anon_sym_pos] = ACTIONS(3029),
    [anon_sym_quotemeta] = ACTIONS(3029),
    [anon_sym_reset] = ACTIONS(3029),
    [anon_sym_rand] = ACTIONS(3029),
    [anon_sym_rmdir] = ACTIONS(3029),
    [anon_sym_readdir] = ACTIONS(3029),
    [anon_sym_readline] = ACTIONS(3029),
    [anon_sym_readpipe] = ACTIONS(3029),
    [anon_sym_rewinddir] = ACTIONS(3029),
    [anon_sym_readlink] = ACTIONS(3029),
    [anon_sym_ref] = ACTIONS(3029),
    [anon_sym_scalar] = ACTIONS(3029),
    [anon_sym_shift] = ACTIONS(3029),
    [anon_sym_sin] = ACTIONS(3029),
    [anon_sym_sleep] = ACTIONS(3029),
    [anon_sym_sqrt] = ACTIONS(3029),
    [anon_sym_srand] = ACTIONS(3029),
    [anon_sym_stat] = ACTIONS(3029),
    [anon_sym_study] = ACTIONS(3029),
    [anon_sym_tell] = ACTIONS(3029),
    [anon_sym_telldir] = ACTIONS(3029),
    [anon_sym_tied] = ACTIONS(3029),
    [anon_sym_uc] = ACTIONS(3029),
    [anon_sym_ucfirst] = ACTIONS(3029),
    [anon_sym_untie] = ACTIONS(3029),
    [anon_sym_umask] = ACTIONS(3029),
    [anon_sym_values] = ACTIONS(3029),
    [anon_sym_write] = ACTIONS(3029),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3029),
    [anon_sym_qq] = ACTIONS(3029),
    [anon_sym_qw] = ACTIONS(3029),
    [anon_sym_qx] = ACTIONS(3029),
    [anon_sym_qr] = ACTIONS(3029),
    [anon_sym_m] = ACTIONS(3029),
    [anon_sym_s] = ACTIONS(3029),
    [anon_sym_tr] = ACTIONS(3029),
    [anon_sym_y] = ACTIONS(3029),
    [anon_sym_if] = ACTIONS(3029),
    [anon_sym_unless] = ACTIONS(3029),
    [anon_sym_while] = ACTIONS(3029),
    [anon_sym_until] = ACTIONS(3029),
    [aux_sym__bareword_token1] = ACTIONS(3029),
    [sym_number] = ACTIONS(3029),
    [anon_sym_true] = ACTIONS(3029),
    [anon_sym_false] = ACTIONS(3029),
    [sym__single_quote] = ACTIONS(3027),
    [sym__double_quote] = ACTIONS(3027),
    [sym__backtick_quote] = ACTIONS(3027),
    [sym__search_slash_quote] = ACTIONS(3027),
    [sym__open_readline_bracket] = ACTIONS(3027),
    [sym__open_fileglob_bracket] = ACTIONS(3027),
    [sym__PERLY_HEREDOC] = ACTIONS(3027),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3027),
    [sym__filetest] = ACTIONS(3027),
  },
  [STATE(1591)] = {
    [sym_heredoc_content] = STATE(1591),
    [sym__identifier] = ACTIONS(2910),
    [anon_sym_LBRACE] = ACTIONS(2910),
    [anon_sym_RBRACE] = ACTIONS(2910),
    [anon_sym_SEMI] = ACTIONS(2910),
    [anon_sym_package] = ACTIONS(2910),
    [anon_sym_class] = ACTIONS(2910),
    [anon_sym_role] = ACTIONS(2910),
    [anon_sym_BUILD] = ACTIONS(2910),
    [anon_sym_ADJUST] = ACTIONS(2910),
    [anon_sym_DOLLAR] = ACTIONS(2910),
    [anon_sym_AT] = ACTIONS(2910),
    [anon_sym_my] = ACTIONS(2910),
    [anon_sym_extended] = ACTIONS(2910),
    [anon_sym_async] = ACTIONS(2910),
    [anon_sym_sub] = ACTIONS(2910),
    [anon_sym_method] = ACTIONS(2910),
    [anon_sym_LPAREN] = ACTIONS(2910),
    [anon_sym_continue] = ACTIONS(2910),
    [anon_sym_state] = ACTIONS(2910),
    [anon_sym_our] = ACTIONS(2910),
    [anon_sym_try] = ACTIONS(2910),
    [anon_sym_defer] = ACTIONS(2910),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2910),
    [anon_sym_LT_LT] = ACTIONS(2910),
    [anon_sym_SLASH_SLASH] = ACTIONS(2910),
    [anon_sym_PLUS] = ACTIONS(2910),
    [anon_sym_DASH] = ACTIONS(2910),
    [anon_sym_TILDE] = ACTIONS(2910),
    [anon_sym_BANG] = ACTIONS(2910),
    [anon_sym_PLUS_PLUS] = ACTIONS(2910),
    [anon_sym_DASH_DASH] = ACTIONS(2910),
    [anon_sym_BSLASH] = ACTIONS(2908),
    [anon_sym_do] = ACTIONS(2910),
    [anon_sym_eval] = ACTIONS(2910),
    [anon_sym_await] = ACTIONS(2910),
    [anon_sym_field] = ACTIONS(2910),
    [anon_sym_local] = ACTIONS(2910),
    [anon_sym_dynamically] = ACTIONS(2910),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2910),
    [anon_sym_require] = ACTIONS(2910),
    [anon_sym_map] = ACTIONS(2910),
    [anon_sym_grep] = ACTIONS(2910),
    [anon_sym_sort] = ACTIONS(2910),
    [anon_sym_goto] = ACTIONS(2910),
    [anon_sym_return] = ACTIONS(2910),
    [anon_sym_undef] = ACTIONS(2910),
    [anon_sym_PERCENT2] = ACTIONS(2910),
    [anon_sym_AMP2] = ACTIONS(2910),
    [anon_sym_STAR2] = ACTIONS(2910),
    [anon_sym_use] = ACTIONS(2910),
    [anon_sym_no] = ACTIONS(2910),
    [anon_sym_for] = ACTIONS(2910),
    [anon_sym_foreach] = ACTIONS(2910),
    [anon_sym_last] = ACTIONS(2910),
    [anon_sym_next] = ACTIONS(2910),
    [anon_sym_redo] = ACTIONS(2910),
    [anon_sym_BEGIN] = ACTIONS(2910),
    [anon_sym_INIT] = ACTIONS(2910),
    [anon_sym_CHECK] = ACTIONS(2910),
    [anon_sym_UNITCHECK] = ACTIONS(2910),
    [anon_sym_END] = ACTIONS(2910),
    [anon_sym___FILE__] = ACTIONS(2910),
    [anon_sym___LINE__] = ACTIONS(2910),
    [anon_sym___PACKAGE__] = ACTIONS(2910),
    [anon_sym___SUB__] = ACTIONS(2910),
    [anon_sym_break] = ACTIONS(2910),
    [anon_sym_fork] = ACTIONS(2910),
    [anon_sym_getppid] = ACTIONS(2910),
    [anon_sym_time] = ACTIONS(2910),
    [anon_sym_times] = ACTIONS(2910),
    [anon_sym_wait] = ACTIONS(2910),
    [anon_sym_wantarray] = ACTIONS(2910),
    [anon_sym_abs] = ACTIONS(2910),
    [anon_sym_alarm] = ACTIONS(2910),
    [anon_sym_chop] = ACTIONS(2910),
    [anon_sym_chdir] = ACTIONS(2910),
    [anon_sym_close] = ACTIONS(2910),
    [anon_sym_closedir] = ACTIONS(2910),
    [anon_sym_caller] = ACTIONS(2910),
    [anon_sym_chomp] = ACTIONS(2910),
    [anon_sym_chr] = ACTIONS(2910),
    [anon_sym_cos] = ACTIONS(2910),
    [anon_sym_chroot] = ACTIONS(2910),
    [anon_sym_defined] = ACTIONS(2910),
    [anon_sym_delete] = ACTIONS(2910),
    [anon_sym_dbmclose] = ACTIONS(2910),
    [anon_sym_exists] = ACTIONS(2910),
    [anon_sym_exit] = ACTIONS(2910),
    [anon_sym_eof] = ACTIONS(2910),
    [anon_sym_exp] = ACTIONS(2910),
    [anon_sym_each] = ACTIONS(2910),
    [anon_sym_fc] = ACTIONS(2910),
    [anon_sym_fileno] = ACTIONS(2910),
    [anon_sym_gmtime] = ACTIONS(2910),
    [anon_sym_getc] = ACTIONS(2910),
    [anon_sym_getpgrp] = ACTIONS(2910),
    [anon_sym_getprotobyname] = ACTIONS(2910),
    [anon_sym_getpwname] = ACTIONS(2910),
    [anon_sym_getpwuid] = ACTIONS(2910),
    [anon_sym_getpeername] = ACTIONS(2910),
    [anon_sym_getnetbyname] = ACTIONS(2910),
    [anon_sym_getsockname] = ACTIONS(2910),
    [anon_sym_getgrnam] = ACTIONS(2910),
    [anon_sym_getgrgid] = ACTIONS(2910),
    [anon_sym_hex] = ACTIONS(2910),
    [anon_sym_int] = ACTIONS(2910),
    [anon_sym_keys] = ACTIONS(2910),
    [anon_sym_lc] = ACTIONS(2910),
    [anon_sym_lcfirst] = ACTIONS(2910),
    [anon_sym_length] = ACTIONS(2910),
    [anon_sym_localtime] = ACTIONS(2910),
    [anon_sym_log] = ACTIONS(2910),
    [anon_sym_lock] = ACTIONS(2910),
    [anon_sym_lstat] = ACTIONS(2910),
    [anon_sym_oct] = ACTIONS(2910),
    [anon_sym_ord] = ACTIONS(2910),
    [anon_sym_prototype] = ACTIONS(2910),
    [anon_sym_pop] = ACTIONS(2910),
    [anon_sym_pos] = ACTIONS(2910),
    [anon_sym_quotemeta] = ACTIONS(2910),
    [anon_sym_reset] = ACTIONS(2910),
    [anon_sym_rand] = ACTIONS(2910),
    [anon_sym_rmdir] = ACTIONS(2910),
    [anon_sym_readdir] = ACTIONS(2910),
    [anon_sym_readline] = ACTIONS(2910),
    [anon_sym_readpipe] = ACTIONS(2910),
    [anon_sym_rewinddir] = ACTIONS(2910),
    [anon_sym_readlink] = ACTIONS(2910),
    [anon_sym_ref] = ACTIONS(2910),
    [anon_sym_scalar] = ACTIONS(2910),
    [anon_sym_shift] = ACTIONS(2910),
    [anon_sym_sin] = ACTIONS(2910),
    [anon_sym_sleep] = ACTIONS(2910),
    [anon_sym_sqrt] = ACTIONS(2910),
    [anon_sym_srand] = ACTIONS(2910),
    [anon_sym_stat] = ACTIONS(2910),
    [anon_sym_study] = ACTIONS(2910),
    [anon_sym_tell] = ACTIONS(2910),
    [anon_sym_telldir] = ACTIONS(2910),
    [anon_sym_tied] = ACTIONS(2910),
    [anon_sym_uc] = ACTIONS(2910),
    [anon_sym_ucfirst] = ACTIONS(2910),
    [anon_sym_untie] = ACTIONS(2910),
    [anon_sym_umask] = ACTIONS(2910),
    [anon_sym_values] = ACTIONS(2910),
    [anon_sym_write] = ACTIONS(2910),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2910),
    [anon_sym_qq] = ACTIONS(2910),
    [anon_sym_qw] = ACTIONS(2910),
    [anon_sym_qx] = ACTIONS(2910),
    [anon_sym_qr] = ACTIONS(2910),
    [anon_sym_m] = ACTIONS(2910),
    [anon_sym_s] = ACTIONS(2910),
    [anon_sym_tr] = ACTIONS(2910),
    [anon_sym_y] = ACTIONS(2910),
    [anon_sym_if] = ACTIONS(2910),
    [anon_sym_unless] = ACTIONS(2910),
    [anon_sym_while] = ACTIONS(2910),
    [anon_sym_until] = ACTIONS(2910),
    [aux_sym__bareword_token1] = ACTIONS(2910),
    [sym_number] = ACTIONS(2910),
    [anon_sym_true] = ACTIONS(2910),
    [anon_sym_false] = ACTIONS(2910),
    [sym__single_quote] = ACTIONS(2908),
    [sym__double_quote] = ACTIONS(2908),
    [sym__backtick_quote] = ACTIONS(2908),
    [sym__search_slash_quote] = ACTIONS(2908),
    [sym__open_readline_bracket] = ACTIONS(2908),
    [sym__open_fileglob_bracket] = ACTIONS(2908),
    [sym__PERLY_HEREDOC] = ACTIONS(2908),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2908),
    [sym__filetest] = ACTIONS(2908),
  },
  [STATE(1592)] = {
    [sym_heredoc_content] = STATE(1592),
    [sym__identifier] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2958),
    [anon_sym_RBRACE] = ACTIONS(2958),
    [anon_sym_SEMI] = ACTIONS(2958),
    [anon_sym_package] = ACTIONS(2958),
    [anon_sym_class] = ACTIONS(2958),
    [anon_sym_role] = ACTIONS(2958),
    [anon_sym_BUILD] = ACTIONS(2958),
    [anon_sym_ADJUST] = ACTIONS(2958),
    [anon_sym_DOLLAR] = ACTIONS(2958),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_my] = ACTIONS(2958),
    [anon_sym_extended] = ACTIONS(2958),
    [anon_sym_async] = ACTIONS(2958),
    [anon_sym_sub] = ACTIONS(2958),
    [anon_sym_method] = ACTIONS(2958),
    [anon_sym_LPAREN] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_state] = ACTIONS(2958),
    [anon_sym_our] = ACTIONS(2958),
    [anon_sym_try] = ACTIONS(2958),
    [anon_sym_defer] = ACTIONS(2958),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2958),
    [anon_sym_LT_LT] = ACTIONS(2958),
    [anon_sym_SLASH_SLASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_TILDE] = ACTIONS(2958),
    [anon_sym_BANG] = ACTIONS(2958),
    [anon_sym_PLUS_PLUS] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2958),
    [anon_sym_BSLASH] = ACTIONS(2956),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_eval] = ACTIONS(2958),
    [anon_sym_await] = ACTIONS(2958),
    [anon_sym_field] = ACTIONS(2958),
    [anon_sym_local] = ACTIONS(2958),
    [anon_sym_dynamically] = ACTIONS(2958),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2958),
    [anon_sym_require] = ACTIONS(2958),
    [anon_sym_map] = ACTIONS(2958),
    [anon_sym_grep] = ACTIONS(2958),
    [anon_sym_sort] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_undef] = ACTIONS(2958),
    [anon_sym_PERCENT2] = ACTIONS(2958),
    [anon_sym_AMP2] = ACTIONS(2958),
    [anon_sym_STAR2] = ACTIONS(2958),
    [anon_sym_use] = ACTIONS(2958),
    [anon_sym_no] = ACTIONS(2958),
    [anon_sym_for] = ACTIONS(2958),
    [anon_sym_foreach] = ACTIONS(2958),
    [anon_sym_last] = ACTIONS(2958),
    [anon_sym_next] = ACTIONS(2958),
    [anon_sym_redo] = ACTIONS(2958),
    [anon_sym_BEGIN] = ACTIONS(2958),
    [anon_sym_INIT] = ACTIONS(2958),
    [anon_sym_CHECK] = ACTIONS(2958),
    [anon_sym_UNITCHECK] = ACTIONS(2958),
    [anon_sym_END] = ACTIONS(2958),
    [anon_sym___FILE__] = ACTIONS(2958),
    [anon_sym___LINE__] = ACTIONS(2958),
    [anon_sym___PACKAGE__] = ACTIONS(2958),
    [anon_sym___SUB__] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_fork] = ACTIONS(2958),
    [anon_sym_getppid] = ACTIONS(2958),
    [anon_sym_time] = ACTIONS(2958),
    [anon_sym_times] = ACTIONS(2958),
    [anon_sym_wait] = ACTIONS(2958),
    [anon_sym_wantarray] = ACTIONS(2958),
    [anon_sym_abs] = ACTIONS(2958),
    [anon_sym_alarm] = ACTIONS(2958),
    [anon_sym_chop] = ACTIONS(2958),
    [anon_sym_chdir] = ACTIONS(2958),
    [anon_sym_close] = ACTIONS(2958),
    [anon_sym_closedir] = ACTIONS(2958),
    [anon_sym_caller] = ACTIONS(2958),
    [anon_sym_chomp] = ACTIONS(2958),
    [anon_sym_chr] = ACTIONS(2958),
    [anon_sym_cos] = ACTIONS(2958),
    [anon_sym_chroot] = ACTIONS(2958),
    [anon_sym_defined] = ACTIONS(2958),
    [anon_sym_delete] = ACTIONS(2958),
    [anon_sym_dbmclose] = ACTIONS(2958),
    [anon_sym_exists] = ACTIONS(2958),
    [anon_sym_exit] = ACTIONS(2958),
    [anon_sym_eof] = ACTIONS(2958),
    [anon_sym_exp] = ACTIONS(2958),
    [anon_sym_each] = ACTIONS(2958),
    [anon_sym_fc] = ACTIONS(2958),
    [anon_sym_fileno] = ACTIONS(2958),
    [anon_sym_gmtime] = ACTIONS(2958),
    [anon_sym_getc] = ACTIONS(2958),
    [anon_sym_getpgrp] = ACTIONS(2958),
    [anon_sym_getprotobyname] = ACTIONS(2958),
    [anon_sym_getpwname] = ACTIONS(2958),
    [anon_sym_getpwuid] = ACTIONS(2958),
    [anon_sym_getpeername] = ACTIONS(2958),
    [anon_sym_getnetbyname] = ACTIONS(2958),
    [anon_sym_getsockname] = ACTIONS(2958),
    [anon_sym_getgrnam] = ACTIONS(2958),
    [anon_sym_getgrgid] = ACTIONS(2958),
    [anon_sym_hex] = ACTIONS(2958),
    [anon_sym_int] = ACTIONS(2958),
    [anon_sym_keys] = ACTIONS(2958),
    [anon_sym_lc] = ACTIONS(2958),
    [anon_sym_lcfirst] = ACTIONS(2958),
    [anon_sym_length] = ACTIONS(2958),
    [anon_sym_localtime] = ACTIONS(2958),
    [anon_sym_log] = ACTIONS(2958),
    [anon_sym_lock] = ACTIONS(2958),
    [anon_sym_lstat] = ACTIONS(2958),
    [anon_sym_oct] = ACTIONS(2958),
    [anon_sym_ord] = ACTIONS(2958),
    [anon_sym_prototype] = ACTIONS(2958),
    [anon_sym_pop] = ACTIONS(2958),
    [anon_sym_pos] = ACTIONS(2958),
    [anon_sym_quotemeta] = ACTIONS(2958),
    [anon_sym_reset] = ACTIONS(2958),
    [anon_sym_rand] = ACTIONS(2958),
    [anon_sym_rmdir] = ACTIONS(2958),
    [anon_sym_readdir] = ACTIONS(2958),
    [anon_sym_readline] = ACTIONS(2958),
    [anon_sym_readpipe] = ACTIONS(2958),
    [anon_sym_rewinddir] = ACTIONS(2958),
    [anon_sym_readlink] = ACTIONS(2958),
    [anon_sym_ref] = ACTIONS(2958),
    [anon_sym_scalar] = ACTIONS(2958),
    [anon_sym_shift] = ACTIONS(2958),
    [anon_sym_sin] = ACTIONS(2958),
    [anon_sym_sleep] = ACTIONS(2958),
    [anon_sym_sqrt] = ACTIONS(2958),
    [anon_sym_srand] = ACTIONS(2958),
    [anon_sym_stat] = ACTIONS(2958),
    [anon_sym_study] = ACTIONS(2958),
    [anon_sym_tell] = ACTIONS(2958),
    [anon_sym_telldir] = ACTIONS(2958),
    [anon_sym_tied] = ACTIONS(2958),
    [anon_sym_uc] = ACTIONS(2958),
    [anon_sym_ucfirst] = ACTIONS(2958),
    [anon_sym_untie] = ACTIONS(2958),
    [anon_sym_umask] = ACTIONS(2958),
    [anon_sym_values] = ACTIONS(2958),
    [anon_sym_write] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2958),
    [anon_sym_qq] = ACTIONS(2958),
    [anon_sym_qw] = ACTIONS(2958),
    [anon_sym_qx] = ACTIONS(2958),
    [anon_sym_qr] = ACTIONS(2958),
    [anon_sym_m] = ACTIONS(2958),
    [anon_sym_s] = ACTIONS(2958),
    [anon_sym_tr] = ACTIONS(2958),
    [anon_sym_y] = ACTIONS(2958),
    [anon_sym_if] = ACTIONS(2958),
    [anon_sym_unless] = ACTIONS(2958),
    [anon_sym_while] = ACTIONS(2958),
    [anon_sym_until] = ACTIONS(2958),
    [aux_sym__bareword_token1] = ACTIONS(2958),
    [sym_number] = ACTIONS(2958),
    [anon_sym_true] = ACTIONS(2958),
    [anon_sym_false] = ACTIONS(2958),
    [sym__single_quote] = ACTIONS(2956),
    [sym__double_quote] = ACTIONS(2956),
    [sym__backtick_quote] = ACTIONS(2956),
    [sym__search_slash_quote] = ACTIONS(2956),
    [sym__open_readline_bracket] = ACTIONS(2956),
    [sym__open_fileglob_bracket] = ACTIONS(2956),
    [sym__PERLY_HEREDOC] = ACTIONS(2956),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2956),
    [sym__filetest] = ACTIONS(2956),
  },
  [STATE(1593)] = {
    [sym_heredoc_content] = STATE(1593),
    [sym__identifier] = ACTIONS(3049),
    [anon_sym_LBRACE] = ACTIONS(3049),
    [anon_sym_RBRACE] = ACTIONS(3049),
    [anon_sym_SEMI] = ACTIONS(3049),
    [anon_sym_package] = ACTIONS(3049),
    [anon_sym_class] = ACTIONS(3049),
    [anon_sym_role] = ACTIONS(3049),
    [anon_sym_BUILD] = ACTIONS(3049),
    [anon_sym_ADJUST] = ACTIONS(3049),
    [anon_sym_DOLLAR] = ACTIONS(3049),
    [anon_sym_AT] = ACTIONS(3049),
    [anon_sym_my] = ACTIONS(3049),
    [anon_sym_extended] = ACTIONS(3049),
    [anon_sym_async] = ACTIONS(3049),
    [anon_sym_sub] = ACTIONS(3049),
    [anon_sym_method] = ACTIONS(3049),
    [anon_sym_LPAREN] = ACTIONS(3049),
    [anon_sym_continue] = ACTIONS(3049),
    [anon_sym_state] = ACTIONS(3049),
    [anon_sym_our] = ACTIONS(3049),
    [anon_sym_try] = ACTIONS(3049),
    [anon_sym_defer] = ACTIONS(3049),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3049),
    [anon_sym_LBRACK] = ACTIONS(3049),
    [anon_sym_LT_LT] = ACTIONS(3049),
    [anon_sym_SLASH_SLASH] = ACTIONS(3049),
    [anon_sym_PLUS] = ACTIONS(3049),
    [anon_sym_DASH] = ACTIONS(3049),
    [anon_sym_TILDE] = ACTIONS(3049),
    [anon_sym_BANG] = ACTIONS(3049),
    [anon_sym_PLUS_PLUS] = ACTIONS(3049),
    [anon_sym_DASH_DASH] = ACTIONS(3049),
    [anon_sym_BSLASH] = ACTIONS(3047),
    [anon_sym_do] = ACTIONS(3049),
    [anon_sym_eval] = ACTIONS(3049),
    [anon_sym_await] = ACTIONS(3049),
    [anon_sym_field] = ACTIONS(3049),
    [anon_sym_local] = ACTIONS(3049),
    [anon_sym_dynamically] = ACTIONS(3049),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3049),
    [anon_sym_require] = ACTIONS(3049),
    [anon_sym_map] = ACTIONS(3049),
    [anon_sym_grep] = ACTIONS(3049),
    [anon_sym_sort] = ACTIONS(3049),
    [anon_sym_goto] = ACTIONS(3049),
    [anon_sym_return] = ACTIONS(3049),
    [anon_sym_undef] = ACTIONS(3049),
    [anon_sym_PERCENT2] = ACTIONS(3049),
    [anon_sym_AMP2] = ACTIONS(3049),
    [anon_sym_STAR2] = ACTIONS(3049),
    [anon_sym_use] = ACTIONS(3049),
    [anon_sym_no] = ACTIONS(3049),
    [anon_sym_for] = ACTIONS(3049),
    [anon_sym_foreach] = ACTIONS(3049),
    [anon_sym_last] = ACTIONS(3049),
    [anon_sym_next] = ACTIONS(3049),
    [anon_sym_redo] = ACTIONS(3049),
    [anon_sym_BEGIN] = ACTIONS(3049),
    [anon_sym_INIT] = ACTIONS(3049),
    [anon_sym_CHECK] = ACTIONS(3049),
    [anon_sym_UNITCHECK] = ACTIONS(3049),
    [anon_sym_END] = ACTIONS(3049),
    [anon_sym___FILE__] = ACTIONS(3049),
    [anon_sym___LINE__] = ACTIONS(3049),
    [anon_sym___PACKAGE__] = ACTIONS(3049),
    [anon_sym___SUB__] = ACTIONS(3049),
    [anon_sym_break] = ACTIONS(3049),
    [anon_sym_fork] = ACTIONS(3049),
    [anon_sym_getppid] = ACTIONS(3049),
    [anon_sym_time] = ACTIONS(3049),
    [anon_sym_times] = ACTIONS(3049),
    [anon_sym_wait] = ACTIONS(3049),
    [anon_sym_wantarray] = ACTIONS(3049),
    [anon_sym_abs] = ACTIONS(3049),
    [anon_sym_alarm] = ACTIONS(3049),
    [anon_sym_chop] = ACTIONS(3049),
    [anon_sym_chdir] = ACTIONS(3049),
    [anon_sym_close] = ACTIONS(3049),
    [anon_sym_closedir] = ACTIONS(3049),
    [anon_sym_caller] = ACTIONS(3049),
    [anon_sym_chomp] = ACTIONS(3049),
    [anon_sym_chr] = ACTIONS(3049),
    [anon_sym_cos] = ACTIONS(3049),
    [anon_sym_chroot] = ACTIONS(3049),
    [anon_sym_defined] = ACTIONS(3049),
    [anon_sym_delete] = ACTIONS(3049),
    [anon_sym_dbmclose] = ACTIONS(3049),
    [anon_sym_exists] = ACTIONS(3049),
    [anon_sym_exit] = ACTIONS(3049),
    [anon_sym_eof] = ACTIONS(3049),
    [anon_sym_exp] = ACTIONS(3049),
    [anon_sym_each] = ACTIONS(3049),
    [anon_sym_fc] = ACTIONS(3049),
    [anon_sym_fileno] = ACTIONS(3049),
    [anon_sym_gmtime] = ACTIONS(3049),
    [anon_sym_getc] = ACTIONS(3049),
    [anon_sym_getpgrp] = ACTIONS(3049),
    [anon_sym_getprotobyname] = ACTIONS(3049),
    [anon_sym_getpwname] = ACTIONS(3049),
    [anon_sym_getpwuid] = ACTIONS(3049),
    [anon_sym_getpeername] = ACTIONS(3049),
    [anon_sym_getnetbyname] = ACTIONS(3049),
    [anon_sym_getsockname] = ACTIONS(3049),
    [anon_sym_getgrnam] = ACTIONS(3049),
    [anon_sym_getgrgid] = ACTIONS(3049),
    [anon_sym_hex] = ACTIONS(3049),
    [anon_sym_int] = ACTIONS(3049),
    [anon_sym_keys] = ACTIONS(3049),
    [anon_sym_lc] = ACTIONS(3049),
    [anon_sym_lcfirst] = ACTIONS(3049),
    [anon_sym_length] = ACTIONS(3049),
    [anon_sym_localtime] = ACTIONS(3049),
    [anon_sym_log] = ACTIONS(3049),
    [anon_sym_lock] = ACTIONS(3049),
    [anon_sym_lstat] = ACTIONS(3049),
    [anon_sym_oct] = ACTIONS(3049),
    [anon_sym_ord] = ACTIONS(3049),
    [anon_sym_prototype] = ACTIONS(3049),
    [anon_sym_pop] = ACTIONS(3049),
    [anon_sym_pos] = ACTIONS(3049),
    [anon_sym_quotemeta] = ACTIONS(3049),
    [anon_sym_reset] = ACTIONS(3049),
    [anon_sym_rand] = ACTIONS(3049),
    [anon_sym_rmdir] = ACTIONS(3049),
    [anon_sym_readdir] = ACTIONS(3049),
    [anon_sym_readline] = ACTIONS(3049),
    [anon_sym_readpipe] = ACTIONS(3049),
    [anon_sym_rewinddir] = ACTIONS(3049),
    [anon_sym_readlink] = ACTIONS(3049),
    [anon_sym_ref] = ACTIONS(3049),
    [anon_sym_scalar] = ACTIONS(3049),
    [anon_sym_shift] = ACTIONS(3049),
    [anon_sym_sin] = ACTIONS(3049),
    [anon_sym_sleep] = ACTIONS(3049),
    [anon_sym_sqrt] = ACTIONS(3049),
    [anon_sym_srand] = ACTIONS(3049),
    [anon_sym_stat] = ACTIONS(3049),
    [anon_sym_study] = ACTIONS(3049),
    [anon_sym_tell] = ACTIONS(3049),
    [anon_sym_telldir] = ACTIONS(3049),
    [anon_sym_tied] = ACTIONS(3049),
    [anon_sym_uc] = ACTIONS(3049),
    [anon_sym_ucfirst] = ACTIONS(3049),
    [anon_sym_untie] = ACTIONS(3049),
    [anon_sym_umask] = ACTIONS(3049),
    [anon_sym_values] = ACTIONS(3049),
    [anon_sym_write] = ACTIONS(3049),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3049),
    [anon_sym_qq] = ACTIONS(3049),
    [anon_sym_qw] = ACTIONS(3049),
    [anon_sym_qx] = ACTIONS(3049),
    [anon_sym_qr] = ACTIONS(3049),
    [anon_sym_m] = ACTIONS(3049),
    [anon_sym_s] = ACTIONS(3049),
    [anon_sym_tr] = ACTIONS(3049),
    [anon_sym_y] = ACTIONS(3049),
    [anon_sym_if] = ACTIONS(3049),
    [anon_sym_unless] = ACTIONS(3049),
    [anon_sym_while] = ACTIONS(3049),
    [anon_sym_until] = ACTIONS(3049),
    [aux_sym__bareword_token1] = ACTIONS(3049),
    [sym_number] = ACTIONS(3049),
    [anon_sym_true] = ACTIONS(3049),
    [anon_sym_false] = ACTIONS(3049),
    [sym__single_quote] = ACTIONS(3047),
    [sym__double_quote] = ACTIONS(3047),
    [sym__backtick_quote] = ACTIONS(3047),
    [sym__search_slash_quote] = ACTIONS(3047),
    [sym__open_readline_bracket] = ACTIONS(3047),
    [sym__open_fileglob_bracket] = ACTIONS(3047),
    [sym__PERLY_HEREDOC] = ACTIONS(3047),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3047),
    [sym__filetest] = ACTIONS(3047),
  },
  [STATE(1594)] = {
    [sym_heredoc_content] = STATE(1594),
    [sym__identifier] = ACTIONS(3069),
    [anon_sym_LBRACE] = ACTIONS(3069),
    [anon_sym_RBRACE] = ACTIONS(3069),
    [anon_sym_SEMI] = ACTIONS(3069),
    [anon_sym_package] = ACTIONS(3069),
    [anon_sym_class] = ACTIONS(3069),
    [anon_sym_role] = ACTIONS(3069),
    [anon_sym_BUILD] = ACTIONS(3069),
    [anon_sym_ADJUST] = ACTIONS(3069),
    [anon_sym_DOLLAR] = ACTIONS(3069),
    [anon_sym_AT] = ACTIONS(3069),
    [anon_sym_my] = ACTIONS(3069),
    [anon_sym_extended] = ACTIONS(3069),
    [anon_sym_async] = ACTIONS(3069),
    [anon_sym_sub] = ACTIONS(3069),
    [anon_sym_method] = ACTIONS(3069),
    [anon_sym_LPAREN] = ACTIONS(3069),
    [anon_sym_continue] = ACTIONS(3069),
    [anon_sym_state] = ACTIONS(3069),
    [anon_sym_our] = ACTIONS(3069),
    [anon_sym_try] = ACTIONS(3069),
    [anon_sym_defer] = ACTIONS(3069),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3069),
    [anon_sym_LBRACK] = ACTIONS(3069),
    [anon_sym_LT_LT] = ACTIONS(3069),
    [anon_sym_SLASH_SLASH] = ACTIONS(3069),
    [anon_sym_PLUS] = ACTIONS(3069),
    [anon_sym_DASH] = ACTIONS(3069),
    [anon_sym_TILDE] = ACTIONS(3069),
    [anon_sym_BANG] = ACTIONS(3069),
    [anon_sym_PLUS_PLUS] = ACTIONS(3069),
    [anon_sym_DASH_DASH] = ACTIONS(3069),
    [anon_sym_BSLASH] = ACTIONS(3067),
    [anon_sym_do] = ACTIONS(3069),
    [anon_sym_eval] = ACTIONS(3069),
    [anon_sym_await] = ACTIONS(3069),
    [anon_sym_field] = ACTIONS(3069),
    [anon_sym_local] = ACTIONS(3069),
    [anon_sym_dynamically] = ACTIONS(3069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3069),
    [anon_sym_require] = ACTIONS(3069),
    [anon_sym_map] = ACTIONS(3069),
    [anon_sym_grep] = ACTIONS(3069),
    [anon_sym_sort] = ACTIONS(3069),
    [anon_sym_goto] = ACTIONS(3069),
    [anon_sym_return] = ACTIONS(3069),
    [anon_sym_undef] = ACTIONS(3069),
    [anon_sym_PERCENT2] = ACTIONS(3069),
    [anon_sym_AMP2] = ACTIONS(3069),
    [anon_sym_STAR2] = ACTIONS(3069),
    [anon_sym_use] = ACTIONS(3069),
    [anon_sym_no] = ACTIONS(3069),
    [anon_sym_for] = ACTIONS(3069),
    [anon_sym_foreach] = ACTIONS(3069),
    [anon_sym_last] = ACTIONS(3069),
    [anon_sym_next] = ACTIONS(3069),
    [anon_sym_redo] = ACTIONS(3069),
    [anon_sym_BEGIN] = ACTIONS(3069),
    [anon_sym_INIT] = ACTIONS(3069),
    [anon_sym_CHECK] = ACTIONS(3069),
    [anon_sym_UNITCHECK] = ACTIONS(3069),
    [anon_sym_END] = ACTIONS(3069),
    [anon_sym___FILE__] = ACTIONS(3069),
    [anon_sym___LINE__] = ACTIONS(3069),
    [anon_sym___PACKAGE__] = ACTIONS(3069),
    [anon_sym___SUB__] = ACTIONS(3069),
    [anon_sym_break] = ACTIONS(3069),
    [anon_sym_fork] = ACTIONS(3069),
    [anon_sym_getppid] = ACTIONS(3069),
    [anon_sym_time] = ACTIONS(3069),
    [anon_sym_times] = ACTIONS(3069),
    [anon_sym_wait] = ACTIONS(3069),
    [anon_sym_wantarray] = ACTIONS(3069),
    [anon_sym_abs] = ACTIONS(3069),
    [anon_sym_alarm] = ACTIONS(3069),
    [anon_sym_chop] = ACTIONS(3069),
    [anon_sym_chdir] = ACTIONS(3069),
    [anon_sym_close] = ACTIONS(3069),
    [anon_sym_closedir] = ACTIONS(3069),
    [anon_sym_caller] = ACTIONS(3069),
    [anon_sym_chomp] = ACTIONS(3069),
    [anon_sym_chr] = ACTIONS(3069),
    [anon_sym_cos] = ACTIONS(3069),
    [anon_sym_chroot] = ACTIONS(3069),
    [anon_sym_defined] = ACTIONS(3069),
    [anon_sym_delete] = ACTIONS(3069),
    [anon_sym_dbmclose] = ACTIONS(3069),
    [anon_sym_exists] = ACTIONS(3069),
    [anon_sym_exit] = ACTIONS(3069),
    [anon_sym_eof] = ACTIONS(3069),
    [anon_sym_exp] = ACTIONS(3069),
    [anon_sym_each] = ACTIONS(3069),
    [anon_sym_fc] = ACTIONS(3069),
    [anon_sym_fileno] = ACTIONS(3069),
    [anon_sym_gmtime] = ACTIONS(3069),
    [anon_sym_getc] = ACTIONS(3069),
    [anon_sym_getpgrp] = ACTIONS(3069),
    [anon_sym_getprotobyname] = ACTIONS(3069),
    [anon_sym_getpwname] = ACTIONS(3069),
    [anon_sym_getpwuid] = ACTIONS(3069),
    [anon_sym_getpeername] = ACTIONS(3069),
    [anon_sym_getnetbyname] = ACTIONS(3069),
    [anon_sym_getsockname] = ACTIONS(3069),
    [anon_sym_getgrnam] = ACTIONS(3069),
    [anon_sym_getgrgid] = ACTIONS(3069),
    [anon_sym_hex] = ACTIONS(3069),
    [anon_sym_int] = ACTIONS(3069),
    [anon_sym_keys] = ACTIONS(3069),
    [anon_sym_lc] = ACTIONS(3069),
    [anon_sym_lcfirst] = ACTIONS(3069),
    [anon_sym_length] = ACTIONS(3069),
    [anon_sym_localtime] = ACTIONS(3069),
    [anon_sym_log] = ACTIONS(3069),
    [anon_sym_lock] = ACTIONS(3069),
    [anon_sym_lstat] = ACTIONS(3069),
    [anon_sym_oct] = ACTIONS(3069),
    [anon_sym_ord] = ACTIONS(3069),
    [anon_sym_prototype] = ACTIONS(3069),
    [anon_sym_pop] = ACTIONS(3069),
    [anon_sym_pos] = ACTIONS(3069),
    [anon_sym_quotemeta] = ACTIONS(3069),
    [anon_sym_reset] = ACTIONS(3069),
    [anon_sym_rand] = ACTIONS(3069),
    [anon_sym_rmdir] = ACTIONS(3069),
    [anon_sym_readdir] = ACTIONS(3069),
    [anon_sym_readline] = ACTIONS(3069),
    [anon_sym_readpipe] = ACTIONS(3069),
    [anon_sym_rewinddir] = ACTIONS(3069),
    [anon_sym_readlink] = ACTIONS(3069),
    [anon_sym_ref] = ACTIONS(3069),
    [anon_sym_scalar] = ACTIONS(3069),
    [anon_sym_shift] = ACTIONS(3069),
    [anon_sym_sin] = ACTIONS(3069),
    [anon_sym_sleep] = ACTIONS(3069),
    [anon_sym_sqrt] = ACTIONS(3069),
    [anon_sym_srand] = ACTIONS(3069),
    [anon_sym_stat] = ACTIONS(3069),
    [anon_sym_study] = ACTIONS(3069),
    [anon_sym_tell] = ACTIONS(3069),
    [anon_sym_telldir] = ACTIONS(3069),
    [anon_sym_tied] = ACTIONS(3069),
    [anon_sym_uc] = ACTIONS(3069),
    [anon_sym_ucfirst] = ACTIONS(3069),
    [anon_sym_untie] = ACTIONS(3069),
    [anon_sym_umask] = ACTIONS(3069),
    [anon_sym_values] = ACTIONS(3069),
    [anon_sym_write] = ACTIONS(3069),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3069),
    [anon_sym_qq] = ACTIONS(3069),
    [anon_sym_qw] = ACTIONS(3069),
    [anon_sym_qx] = ACTIONS(3069),
    [anon_sym_qr] = ACTIONS(3069),
    [anon_sym_m] = ACTIONS(3069),
    [anon_sym_s] = ACTIONS(3069),
    [anon_sym_tr] = ACTIONS(3069),
    [anon_sym_y] = ACTIONS(3069),
    [anon_sym_if] = ACTIONS(3069),
    [anon_sym_unless] = ACTIONS(3069),
    [anon_sym_while] = ACTIONS(3069),
    [anon_sym_until] = ACTIONS(3069),
    [aux_sym__bareword_token1] = ACTIONS(3069),
    [sym_number] = ACTIONS(3069),
    [anon_sym_true] = ACTIONS(3069),
    [anon_sym_false] = ACTIONS(3069),
    [sym__single_quote] = ACTIONS(3067),
    [sym__double_quote] = ACTIONS(3067),
    [sym__backtick_quote] = ACTIONS(3067),
    [sym__search_slash_quote] = ACTIONS(3067),
    [sym__open_readline_bracket] = ACTIONS(3067),
    [sym__open_fileglob_bracket] = ACTIONS(3067),
    [sym__PERLY_HEREDOC] = ACTIONS(3067),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3067),
    [sym__filetest] = ACTIONS(3067),
  },
  [STATE(1595)] = {
    [sym_heredoc_content] = STATE(1595),
    [sym__identifier] = ACTIONS(3253),
    [anon_sym_LBRACE] = ACTIONS(3253),
    [anon_sym_RBRACE] = ACTIONS(3253),
    [anon_sym_SEMI] = ACTIONS(3253),
    [anon_sym_package] = ACTIONS(3253),
    [anon_sym_class] = ACTIONS(3253),
    [anon_sym_role] = ACTIONS(3253),
    [anon_sym_BUILD] = ACTIONS(3253),
    [anon_sym_ADJUST] = ACTIONS(3253),
    [anon_sym_DOLLAR] = ACTIONS(3253),
    [anon_sym_AT] = ACTIONS(3253),
    [anon_sym_my] = ACTIONS(3253),
    [anon_sym_extended] = ACTIONS(3253),
    [anon_sym_async] = ACTIONS(3253),
    [anon_sym_sub] = ACTIONS(3253),
    [anon_sym_method] = ACTIONS(3253),
    [anon_sym_LPAREN] = ACTIONS(3253),
    [anon_sym_continue] = ACTIONS(3253),
    [anon_sym_state] = ACTIONS(3253),
    [anon_sym_our] = ACTIONS(3253),
    [anon_sym_try] = ACTIONS(3253),
    [anon_sym_defer] = ACTIONS(3253),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3253),
    [anon_sym_LBRACK] = ACTIONS(3253),
    [anon_sym_LT_LT] = ACTIONS(3253),
    [anon_sym_SLASH_SLASH] = ACTIONS(3253),
    [anon_sym_PLUS] = ACTIONS(3253),
    [anon_sym_DASH] = ACTIONS(3253),
    [anon_sym_TILDE] = ACTIONS(3253),
    [anon_sym_BANG] = ACTIONS(3253),
    [anon_sym_PLUS_PLUS] = ACTIONS(3253),
    [anon_sym_DASH_DASH] = ACTIONS(3253),
    [anon_sym_BSLASH] = ACTIONS(3251),
    [anon_sym_do] = ACTIONS(3253),
    [anon_sym_eval] = ACTIONS(3253),
    [anon_sym_await] = ACTIONS(3253),
    [anon_sym_field] = ACTIONS(3253),
    [anon_sym_local] = ACTIONS(3253),
    [anon_sym_dynamically] = ACTIONS(3253),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3253),
    [anon_sym_require] = ACTIONS(3253),
    [anon_sym_map] = ACTIONS(3253),
    [anon_sym_grep] = ACTIONS(3253),
    [anon_sym_sort] = ACTIONS(3253),
    [anon_sym_goto] = ACTIONS(3253),
    [anon_sym_return] = ACTIONS(3253),
    [anon_sym_undef] = ACTIONS(3253),
    [anon_sym_PERCENT2] = ACTIONS(3253),
    [anon_sym_AMP2] = ACTIONS(3253),
    [anon_sym_STAR2] = ACTIONS(3253),
    [anon_sym_use] = ACTIONS(3253),
    [anon_sym_no] = ACTIONS(3253),
    [anon_sym_for] = ACTIONS(3253),
    [anon_sym_foreach] = ACTIONS(3253),
    [anon_sym_last] = ACTIONS(3253),
    [anon_sym_next] = ACTIONS(3253),
    [anon_sym_redo] = ACTIONS(3253),
    [anon_sym_BEGIN] = ACTIONS(3253),
    [anon_sym_INIT] = ACTIONS(3253),
    [anon_sym_CHECK] = ACTIONS(3253),
    [anon_sym_UNITCHECK] = ACTIONS(3253),
    [anon_sym_END] = ACTIONS(3253),
    [anon_sym___FILE__] = ACTIONS(3253),
    [anon_sym___LINE__] = ACTIONS(3253),
    [anon_sym___PACKAGE__] = ACTIONS(3253),
    [anon_sym___SUB__] = ACTIONS(3253),
    [anon_sym_break] = ACTIONS(3253),
    [anon_sym_fork] = ACTIONS(3253),
    [anon_sym_getppid] = ACTIONS(3253),
    [anon_sym_time] = ACTIONS(3253),
    [anon_sym_times] = ACTIONS(3253),
    [anon_sym_wait] = ACTIONS(3253),
    [anon_sym_wantarray] = ACTIONS(3253),
    [anon_sym_abs] = ACTIONS(3253),
    [anon_sym_alarm] = ACTIONS(3253),
    [anon_sym_chop] = ACTIONS(3253),
    [anon_sym_chdir] = ACTIONS(3253),
    [anon_sym_close] = ACTIONS(3253),
    [anon_sym_closedir] = ACTIONS(3253),
    [anon_sym_caller] = ACTIONS(3253),
    [anon_sym_chomp] = ACTIONS(3253),
    [anon_sym_chr] = ACTIONS(3253),
    [anon_sym_cos] = ACTIONS(3253),
    [anon_sym_chroot] = ACTIONS(3253),
    [anon_sym_defined] = ACTIONS(3253),
    [anon_sym_delete] = ACTIONS(3253),
    [anon_sym_dbmclose] = ACTIONS(3253),
    [anon_sym_exists] = ACTIONS(3253),
    [anon_sym_exit] = ACTIONS(3253),
    [anon_sym_eof] = ACTIONS(3253),
    [anon_sym_exp] = ACTIONS(3253),
    [anon_sym_each] = ACTIONS(3253),
    [anon_sym_fc] = ACTIONS(3253),
    [anon_sym_fileno] = ACTIONS(3253),
    [anon_sym_gmtime] = ACTIONS(3253),
    [anon_sym_getc] = ACTIONS(3253),
    [anon_sym_getpgrp] = ACTIONS(3253),
    [anon_sym_getprotobyname] = ACTIONS(3253),
    [anon_sym_getpwname] = ACTIONS(3253),
    [anon_sym_getpwuid] = ACTIONS(3253),
    [anon_sym_getpeername] = ACTIONS(3253),
    [anon_sym_getnetbyname] = ACTIONS(3253),
    [anon_sym_getsockname] = ACTIONS(3253),
    [anon_sym_getgrnam] = ACTIONS(3253),
    [anon_sym_getgrgid] = ACTIONS(3253),
    [anon_sym_hex] = ACTIONS(3253),
    [anon_sym_int] = ACTIONS(3253),
    [anon_sym_keys] = ACTIONS(3253),
    [anon_sym_lc] = ACTIONS(3253),
    [anon_sym_lcfirst] = ACTIONS(3253),
    [anon_sym_length] = ACTIONS(3253),
    [anon_sym_localtime] = ACTIONS(3253),
    [anon_sym_log] = ACTIONS(3253),
    [anon_sym_lock] = ACTIONS(3253),
    [anon_sym_lstat] = ACTIONS(3253),
    [anon_sym_oct] = ACTIONS(3253),
    [anon_sym_ord] = ACTIONS(3253),
    [anon_sym_prototype] = ACTIONS(3253),
    [anon_sym_pop] = ACTIONS(3253),
    [anon_sym_pos] = ACTIONS(3253),
    [anon_sym_quotemeta] = ACTIONS(3253),
    [anon_sym_reset] = ACTIONS(3253),
    [anon_sym_rand] = ACTIONS(3253),
    [anon_sym_rmdir] = ACTIONS(3253),
    [anon_sym_readdir] = ACTIONS(3253),
    [anon_sym_readline] = ACTIONS(3253),
    [anon_sym_readpipe] = ACTIONS(3253),
    [anon_sym_rewinddir] = ACTIONS(3253),
    [anon_sym_readlink] = ACTIONS(3253),
    [anon_sym_ref] = ACTIONS(3253),
    [anon_sym_scalar] = ACTIONS(3253),
    [anon_sym_shift] = ACTIONS(3253),
    [anon_sym_sin] = ACTIONS(3253),
    [anon_sym_sleep] = ACTIONS(3253),
    [anon_sym_sqrt] = ACTIONS(3253),
    [anon_sym_srand] = ACTIONS(3253),
    [anon_sym_stat] = ACTIONS(3253),
    [anon_sym_study] = ACTIONS(3253),
    [anon_sym_tell] = ACTIONS(3253),
    [anon_sym_telldir] = ACTIONS(3253),
    [anon_sym_tied] = ACTIONS(3253),
    [anon_sym_uc] = ACTIONS(3253),
    [anon_sym_ucfirst] = ACTIONS(3253),
    [anon_sym_untie] = ACTIONS(3253),
    [anon_sym_umask] = ACTIONS(3253),
    [anon_sym_values] = ACTIONS(3253),
    [anon_sym_write] = ACTIONS(3253),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3253),
    [anon_sym_qq] = ACTIONS(3253),
    [anon_sym_qw] = ACTIONS(3253),
    [anon_sym_qx] = ACTIONS(3253),
    [anon_sym_qr] = ACTIONS(3253),
    [anon_sym_m] = ACTIONS(3253),
    [anon_sym_s] = ACTIONS(3253),
    [anon_sym_tr] = ACTIONS(3253),
    [anon_sym_y] = ACTIONS(3253),
    [anon_sym_if] = ACTIONS(3253),
    [anon_sym_unless] = ACTIONS(3253),
    [anon_sym_while] = ACTIONS(3253),
    [anon_sym_until] = ACTIONS(3253),
    [aux_sym__bareword_token1] = ACTIONS(3253),
    [sym_number] = ACTIONS(3253),
    [anon_sym_true] = ACTIONS(3253),
    [anon_sym_false] = ACTIONS(3253),
    [sym__single_quote] = ACTIONS(3251),
    [sym__double_quote] = ACTIONS(3251),
    [sym__backtick_quote] = ACTIONS(3251),
    [sym__search_slash_quote] = ACTIONS(3251),
    [sym__open_readline_bracket] = ACTIONS(3251),
    [sym__open_fileglob_bracket] = ACTIONS(3251),
    [sym__PERLY_HEREDOC] = ACTIONS(3251),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3251),
    [sym__filetest] = ACTIONS(3251),
  },
  [STATE(1596)] = {
    [sym_heredoc_content] = STATE(1596),
    [sym__identifier] = ACTIONS(3113),
    [anon_sym_LBRACE] = ACTIONS(3113),
    [anon_sym_RBRACE] = ACTIONS(3113),
    [anon_sym_SEMI] = ACTIONS(3113),
    [anon_sym_package] = ACTIONS(3113),
    [anon_sym_class] = ACTIONS(3113),
    [anon_sym_role] = ACTIONS(3113),
    [anon_sym_BUILD] = ACTIONS(3113),
    [anon_sym_ADJUST] = ACTIONS(3113),
    [anon_sym_DOLLAR] = ACTIONS(3113),
    [anon_sym_AT] = ACTIONS(3113),
    [anon_sym_my] = ACTIONS(3113),
    [anon_sym_extended] = ACTIONS(3113),
    [anon_sym_async] = ACTIONS(3113),
    [anon_sym_sub] = ACTIONS(3113),
    [anon_sym_method] = ACTIONS(3113),
    [anon_sym_LPAREN] = ACTIONS(3113),
    [anon_sym_continue] = ACTIONS(3113),
    [anon_sym_state] = ACTIONS(3113),
    [anon_sym_our] = ACTIONS(3113),
    [anon_sym_try] = ACTIONS(3113),
    [anon_sym_defer] = ACTIONS(3113),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3113),
    [anon_sym_LBRACK] = ACTIONS(3113),
    [anon_sym_LT_LT] = ACTIONS(3113),
    [anon_sym_SLASH_SLASH] = ACTIONS(3113),
    [anon_sym_PLUS] = ACTIONS(3113),
    [anon_sym_DASH] = ACTIONS(3113),
    [anon_sym_TILDE] = ACTIONS(3113),
    [anon_sym_BANG] = ACTIONS(3113),
    [anon_sym_PLUS_PLUS] = ACTIONS(3113),
    [anon_sym_DASH_DASH] = ACTIONS(3113),
    [anon_sym_BSLASH] = ACTIONS(3111),
    [anon_sym_do] = ACTIONS(3113),
    [anon_sym_eval] = ACTIONS(3113),
    [anon_sym_await] = ACTIONS(3113),
    [anon_sym_field] = ACTIONS(3113),
    [anon_sym_local] = ACTIONS(3113),
    [anon_sym_dynamically] = ACTIONS(3113),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3113),
    [anon_sym_require] = ACTIONS(3113),
    [anon_sym_map] = ACTIONS(3113),
    [anon_sym_grep] = ACTIONS(3113),
    [anon_sym_sort] = ACTIONS(3113),
    [anon_sym_goto] = ACTIONS(3113),
    [anon_sym_return] = ACTIONS(3113),
    [anon_sym_undef] = ACTIONS(3113),
    [anon_sym_PERCENT2] = ACTIONS(3113),
    [anon_sym_AMP2] = ACTIONS(3113),
    [anon_sym_STAR2] = ACTIONS(3113),
    [anon_sym_use] = ACTIONS(3113),
    [anon_sym_no] = ACTIONS(3113),
    [anon_sym_for] = ACTIONS(3113),
    [anon_sym_foreach] = ACTIONS(3113),
    [anon_sym_last] = ACTIONS(3113),
    [anon_sym_next] = ACTIONS(3113),
    [anon_sym_redo] = ACTIONS(3113),
    [anon_sym_BEGIN] = ACTIONS(3113),
    [anon_sym_INIT] = ACTIONS(3113),
    [anon_sym_CHECK] = ACTIONS(3113),
    [anon_sym_UNITCHECK] = ACTIONS(3113),
    [anon_sym_END] = ACTIONS(3113),
    [anon_sym___FILE__] = ACTIONS(3113),
    [anon_sym___LINE__] = ACTIONS(3113),
    [anon_sym___PACKAGE__] = ACTIONS(3113),
    [anon_sym___SUB__] = ACTIONS(3113),
    [anon_sym_break] = ACTIONS(3113),
    [anon_sym_fork] = ACTIONS(3113),
    [anon_sym_getppid] = ACTIONS(3113),
    [anon_sym_time] = ACTIONS(3113),
    [anon_sym_times] = ACTIONS(3113),
    [anon_sym_wait] = ACTIONS(3113),
    [anon_sym_wantarray] = ACTIONS(3113),
    [anon_sym_abs] = ACTIONS(3113),
    [anon_sym_alarm] = ACTIONS(3113),
    [anon_sym_chop] = ACTIONS(3113),
    [anon_sym_chdir] = ACTIONS(3113),
    [anon_sym_close] = ACTIONS(3113),
    [anon_sym_closedir] = ACTIONS(3113),
    [anon_sym_caller] = ACTIONS(3113),
    [anon_sym_chomp] = ACTIONS(3113),
    [anon_sym_chr] = ACTIONS(3113),
    [anon_sym_cos] = ACTIONS(3113),
    [anon_sym_chroot] = ACTIONS(3113),
    [anon_sym_defined] = ACTIONS(3113),
    [anon_sym_delete] = ACTIONS(3113),
    [anon_sym_dbmclose] = ACTIONS(3113),
    [anon_sym_exists] = ACTIONS(3113),
    [anon_sym_exit] = ACTIONS(3113),
    [anon_sym_eof] = ACTIONS(3113),
    [anon_sym_exp] = ACTIONS(3113),
    [anon_sym_each] = ACTIONS(3113),
    [anon_sym_fc] = ACTIONS(3113),
    [anon_sym_fileno] = ACTIONS(3113),
    [anon_sym_gmtime] = ACTIONS(3113),
    [anon_sym_getc] = ACTIONS(3113),
    [anon_sym_getpgrp] = ACTIONS(3113),
    [anon_sym_getprotobyname] = ACTIONS(3113),
    [anon_sym_getpwname] = ACTIONS(3113),
    [anon_sym_getpwuid] = ACTIONS(3113),
    [anon_sym_getpeername] = ACTIONS(3113),
    [anon_sym_getnetbyname] = ACTIONS(3113),
    [anon_sym_getsockname] = ACTIONS(3113),
    [anon_sym_getgrnam] = ACTIONS(3113),
    [anon_sym_getgrgid] = ACTIONS(3113),
    [anon_sym_hex] = ACTIONS(3113),
    [anon_sym_int] = ACTIONS(3113),
    [anon_sym_keys] = ACTIONS(3113),
    [anon_sym_lc] = ACTIONS(3113),
    [anon_sym_lcfirst] = ACTIONS(3113),
    [anon_sym_length] = ACTIONS(3113),
    [anon_sym_localtime] = ACTIONS(3113),
    [anon_sym_log] = ACTIONS(3113),
    [anon_sym_lock] = ACTIONS(3113),
    [anon_sym_lstat] = ACTIONS(3113),
    [anon_sym_oct] = ACTIONS(3113),
    [anon_sym_ord] = ACTIONS(3113),
    [anon_sym_prototype] = ACTIONS(3113),
    [anon_sym_pop] = ACTIONS(3113),
    [anon_sym_pos] = ACTIONS(3113),
    [anon_sym_quotemeta] = ACTIONS(3113),
    [anon_sym_reset] = ACTIONS(3113),
    [anon_sym_rand] = ACTIONS(3113),
    [anon_sym_rmdir] = ACTIONS(3113),
    [anon_sym_readdir] = ACTIONS(3113),
    [anon_sym_readline] = ACTIONS(3113),
    [anon_sym_readpipe] = ACTIONS(3113),
    [anon_sym_rewinddir] = ACTIONS(3113),
    [anon_sym_readlink] = ACTIONS(3113),
    [anon_sym_ref] = ACTIONS(3113),
    [anon_sym_scalar] = ACTIONS(3113),
    [anon_sym_shift] = ACTIONS(3113),
    [anon_sym_sin] = ACTIONS(3113),
    [anon_sym_sleep] = ACTIONS(3113),
    [anon_sym_sqrt] = ACTIONS(3113),
    [anon_sym_srand] = ACTIONS(3113),
    [anon_sym_stat] = ACTIONS(3113),
    [anon_sym_study] = ACTIONS(3113),
    [anon_sym_tell] = ACTIONS(3113),
    [anon_sym_telldir] = ACTIONS(3113),
    [anon_sym_tied] = ACTIONS(3113),
    [anon_sym_uc] = ACTIONS(3113),
    [anon_sym_ucfirst] = ACTIONS(3113),
    [anon_sym_untie] = ACTIONS(3113),
    [anon_sym_umask] = ACTIONS(3113),
    [anon_sym_values] = ACTIONS(3113),
    [anon_sym_write] = ACTIONS(3113),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3113),
    [anon_sym_qq] = ACTIONS(3113),
    [anon_sym_qw] = ACTIONS(3113),
    [anon_sym_qx] = ACTIONS(3113),
    [anon_sym_qr] = ACTIONS(3113),
    [anon_sym_m] = ACTIONS(3113),
    [anon_sym_s] = ACTIONS(3113),
    [anon_sym_tr] = ACTIONS(3113),
    [anon_sym_y] = ACTIONS(3113),
    [anon_sym_if] = ACTIONS(3113),
    [anon_sym_unless] = ACTIONS(3113),
    [anon_sym_while] = ACTIONS(3113),
    [anon_sym_until] = ACTIONS(3113),
    [aux_sym__bareword_token1] = ACTIONS(3113),
    [sym_number] = ACTIONS(3113),
    [anon_sym_true] = ACTIONS(3113),
    [anon_sym_false] = ACTIONS(3113),
    [sym__single_quote] = ACTIONS(3111),
    [sym__double_quote] = ACTIONS(3111),
    [sym__backtick_quote] = ACTIONS(3111),
    [sym__search_slash_quote] = ACTIONS(3111),
    [sym__open_readline_bracket] = ACTIONS(3111),
    [sym__open_fileglob_bracket] = ACTIONS(3111),
    [sym__PERLY_HEREDOC] = ACTIONS(3111),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3111),
    [sym__filetest] = ACTIONS(3111),
  },
  [STATE(1597)] = {
    [sym_heredoc_content] = STATE(1597),
    [sym__identifier] = ACTIONS(3201),
    [anon_sym_LBRACE] = ACTIONS(3201),
    [anon_sym_RBRACE] = ACTIONS(3201),
    [anon_sym_SEMI] = ACTIONS(3201),
    [anon_sym_package] = ACTIONS(3201),
    [anon_sym_class] = ACTIONS(3201),
    [anon_sym_role] = ACTIONS(3201),
    [anon_sym_BUILD] = ACTIONS(3201),
    [anon_sym_ADJUST] = ACTIONS(3201),
    [anon_sym_DOLLAR] = ACTIONS(3201),
    [anon_sym_AT] = ACTIONS(3201),
    [anon_sym_my] = ACTIONS(3201),
    [anon_sym_extended] = ACTIONS(3201),
    [anon_sym_async] = ACTIONS(3201),
    [anon_sym_sub] = ACTIONS(3201),
    [anon_sym_method] = ACTIONS(3201),
    [anon_sym_LPAREN] = ACTIONS(3201),
    [anon_sym_continue] = ACTIONS(3201),
    [anon_sym_state] = ACTIONS(3201),
    [anon_sym_our] = ACTIONS(3201),
    [anon_sym_try] = ACTIONS(3201),
    [anon_sym_defer] = ACTIONS(3201),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3201),
    [anon_sym_LBRACK] = ACTIONS(3201),
    [anon_sym_LT_LT] = ACTIONS(3201),
    [anon_sym_SLASH_SLASH] = ACTIONS(3201),
    [anon_sym_PLUS] = ACTIONS(3201),
    [anon_sym_DASH] = ACTIONS(3201),
    [anon_sym_TILDE] = ACTIONS(3201),
    [anon_sym_BANG] = ACTIONS(3201),
    [anon_sym_PLUS_PLUS] = ACTIONS(3201),
    [anon_sym_DASH_DASH] = ACTIONS(3201),
    [anon_sym_BSLASH] = ACTIONS(3199),
    [anon_sym_do] = ACTIONS(3201),
    [anon_sym_eval] = ACTIONS(3201),
    [anon_sym_await] = ACTIONS(3201),
    [anon_sym_field] = ACTIONS(3201),
    [anon_sym_local] = ACTIONS(3201),
    [anon_sym_dynamically] = ACTIONS(3201),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3201),
    [anon_sym_require] = ACTIONS(3201),
    [anon_sym_map] = ACTIONS(3201),
    [anon_sym_grep] = ACTIONS(3201),
    [anon_sym_sort] = ACTIONS(3201),
    [anon_sym_goto] = ACTIONS(3201),
    [anon_sym_return] = ACTIONS(3201),
    [anon_sym_undef] = ACTIONS(3201),
    [anon_sym_PERCENT2] = ACTIONS(3201),
    [anon_sym_AMP2] = ACTIONS(3201),
    [anon_sym_STAR2] = ACTIONS(3201),
    [anon_sym_use] = ACTIONS(3201),
    [anon_sym_no] = ACTIONS(3201),
    [anon_sym_for] = ACTIONS(3201),
    [anon_sym_foreach] = ACTIONS(3201),
    [anon_sym_last] = ACTIONS(3201),
    [anon_sym_next] = ACTIONS(3201),
    [anon_sym_redo] = ACTIONS(3201),
    [anon_sym_BEGIN] = ACTIONS(3201),
    [anon_sym_INIT] = ACTIONS(3201),
    [anon_sym_CHECK] = ACTIONS(3201),
    [anon_sym_UNITCHECK] = ACTIONS(3201),
    [anon_sym_END] = ACTIONS(3201),
    [anon_sym___FILE__] = ACTIONS(3201),
    [anon_sym___LINE__] = ACTIONS(3201),
    [anon_sym___PACKAGE__] = ACTIONS(3201),
    [anon_sym___SUB__] = ACTIONS(3201),
    [anon_sym_break] = ACTIONS(3201),
    [anon_sym_fork] = ACTIONS(3201),
    [anon_sym_getppid] = ACTIONS(3201),
    [anon_sym_time] = ACTIONS(3201),
    [anon_sym_times] = ACTIONS(3201),
    [anon_sym_wait] = ACTIONS(3201),
    [anon_sym_wantarray] = ACTIONS(3201),
    [anon_sym_abs] = ACTIONS(3201),
    [anon_sym_alarm] = ACTIONS(3201),
    [anon_sym_chop] = ACTIONS(3201),
    [anon_sym_chdir] = ACTIONS(3201),
    [anon_sym_close] = ACTIONS(3201),
    [anon_sym_closedir] = ACTIONS(3201),
    [anon_sym_caller] = ACTIONS(3201),
    [anon_sym_chomp] = ACTIONS(3201),
    [anon_sym_chr] = ACTIONS(3201),
    [anon_sym_cos] = ACTIONS(3201),
    [anon_sym_chroot] = ACTIONS(3201),
    [anon_sym_defined] = ACTIONS(3201),
    [anon_sym_delete] = ACTIONS(3201),
    [anon_sym_dbmclose] = ACTIONS(3201),
    [anon_sym_exists] = ACTIONS(3201),
    [anon_sym_exit] = ACTIONS(3201),
    [anon_sym_eof] = ACTIONS(3201),
    [anon_sym_exp] = ACTIONS(3201),
    [anon_sym_each] = ACTIONS(3201),
    [anon_sym_fc] = ACTIONS(3201),
    [anon_sym_fileno] = ACTIONS(3201),
    [anon_sym_gmtime] = ACTIONS(3201),
    [anon_sym_getc] = ACTIONS(3201),
    [anon_sym_getpgrp] = ACTIONS(3201),
    [anon_sym_getprotobyname] = ACTIONS(3201),
    [anon_sym_getpwname] = ACTIONS(3201),
    [anon_sym_getpwuid] = ACTIONS(3201),
    [anon_sym_getpeername] = ACTIONS(3201),
    [anon_sym_getnetbyname] = ACTIONS(3201),
    [anon_sym_getsockname] = ACTIONS(3201),
    [anon_sym_getgrnam] = ACTIONS(3201),
    [anon_sym_getgrgid] = ACTIONS(3201),
    [anon_sym_hex] = ACTIONS(3201),
    [anon_sym_int] = ACTIONS(3201),
    [anon_sym_keys] = ACTIONS(3201),
    [anon_sym_lc] = ACTIONS(3201),
    [anon_sym_lcfirst] = ACTIONS(3201),
    [anon_sym_length] = ACTIONS(3201),
    [anon_sym_localtime] = ACTIONS(3201),
    [anon_sym_log] = ACTIONS(3201),
    [anon_sym_lock] = ACTIONS(3201),
    [anon_sym_lstat] = ACTIONS(3201),
    [anon_sym_oct] = ACTIONS(3201),
    [anon_sym_ord] = ACTIONS(3201),
    [anon_sym_prototype] = ACTIONS(3201),
    [anon_sym_pop] = ACTIONS(3201),
    [anon_sym_pos] = ACTIONS(3201),
    [anon_sym_quotemeta] = ACTIONS(3201),
    [anon_sym_reset] = ACTIONS(3201),
    [anon_sym_rand] = ACTIONS(3201),
    [anon_sym_rmdir] = ACTIONS(3201),
    [anon_sym_readdir] = ACTIONS(3201),
    [anon_sym_readline] = ACTIONS(3201),
    [anon_sym_readpipe] = ACTIONS(3201),
    [anon_sym_rewinddir] = ACTIONS(3201),
    [anon_sym_readlink] = ACTIONS(3201),
    [anon_sym_ref] = ACTIONS(3201),
    [anon_sym_scalar] = ACTIONS(3201),
    [anon_sym_shift] = ACTIONS(3201),
    [anon_sym_sin] = ACTIONS(3201),
    [anon_sym_sleep] = ACTIONS(3201),
    [anon_sym_sqrt] = ACTIONS(3201),
    [anon_sym_srand] = ACTIONS(3201),
    [anon_sym_stat] = ACTIONS(3201),
    [anon_sym_study] = ACTIONS(3201),
    [anon_sym_tell] = ACTIONS(3201),
    [anon_sym_telldir] = ACTIONS(3201),
    [anon_sym_tied] = ACTIONS(3201),
    [anon_sym_uc] = ACTIONS(3201),
    [anon_sym_ucfirst] = ACTIONS(3201),
    [anon_sym_untie] = ACTIONS(3201),
    [anon_sym_umask] = ACTIONS(3201),
    [anon_sym_values] = ACTIONS(3201),
    [anon_sym_write] = ACTIONS(3201),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3201),
    [anon_sym_qq] = ACTIONS(3201),
    [anon_sym_qw] = ACTIONS(3201),
    [anon_sym_qx] = ACTIONS(3201),
    [anon_sym_qr] = ACTIONS(3201),
    [anon_sym_m] = ACTIONS(3201),
    [anon_sym_s] = ACTIONS(3201),
    [anon_sym_tr] = ACTIONS(3201),
    [anon_sym_y] = ACTIONS(3201),
    [anon_sym_if] = ACTIONS(3201),
    [anon_sym_unless] = ACTIONS(3201),
    [anon_sym_while] = ACTIONS(3201),
    [anon_sym_until] = ACTIONS(3201),
    [aux_sym__bareword_token1] = ACTIONS(3201),
    [sym_number] = ACTIONS(3201),
    [anon_sym_true] = ACTIONS(3201),
    [anon_sym_false] = ACTIONS(3201),
    [sym__single_quote] = ACTIONS(3199),
    [sym__double_quote] = ACTIONS(3199),
    [sym__backtick_quote] = ACTIONS(3199),
    [sym__search_slash_quote] = ACTIONS(3199),
    [sym__open_readline_bracket] = ACTIONS(3199),
    [sym__open_fileglob_bracket] = ACTIONS(3199),
    [sym__PERLY_HEREDOC] = ACTIONS(3199),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3199),
    [sym__filetest] = ACTIONS(3199),
  },
  [STATE(1598)] = {
    [sym_heredoc_content] = STATE(1598),
    [sym__identifier] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2962),
    [anon_sym_RBRACE] = ACTIONS(2962),
    [anon_sym_SEMI] = ACTIONS(2962),
    [anon_sym_package] = ACTIONS(2962),
    [anon_sym_class] = ACTIONS(2962),
    [anon_sym_role] = ACTIONS(2962),
    [anon_sym_BUILD] = ACTIONS(2962),
    [anon_sym_ADJUST] = ACTIONS(2962),
    [anon_sym_DOLLAR] = ACTIONS(2962),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_my] = ACTIONS(2962),
    [anon_sym_extended] = ACTIONS(2962),
    [anon_sym_async] = ACTIONS(2962),
    [anon_sym_sub] = ACTIONS(2962),
    [anon_sym_method] = ACTIONS(2962),
    [anon_sym_LPAREN] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_state] = ACTIONS(2962),
    [anon_sym_our] = ACTIONS(2962),
    [anon_sym_try] = ACTIONS(2962),
    [anon_sym_defer] = ACTIONS(2962),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2962),
    [anon_sym_LT_LT] = ACTIONS(2962),
    [anon_sym_SLASH_SLASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_TILDE] = ACTIONS(2962),
    [anon_sym_BANG] = ACTIONS(2962),
    [anon_sym_PLUS_PLUS] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2962),
    [anon_sym_BSLASH] = ACTIONS(2960),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_eval] = ACTIONS(2962),
    [anon_sym_await] = ACTIONS(2962),
    [anon_sym_field] = ACTIONS(2962),
    [anon_sym_local] = ACTIONS(2962),
    [anon_sym_dynamically] = ACTIONS(2962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2962),
    [anon_sym_require] = ACTIONS(2962),
    [anon_sym_map] = ACTIONS(2962),
    [anon_sym_grep] = ACTIONS(2962),
    [anon_sym_sort] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_undef] = ACTIONS(2962),
    [anon_sym_PERCENT2] = ACTIONS(2962),
    [anon_sym_AMP2] = ACTIONS(2962),
    [anon_sym_STAR2] = ACTIONS(2962),
    [anon_sym_use] = ACTIONS(2962),
    [anon_sym_no] = ACTIONS(2962),
    [anon_sym_for] = ACTIONS(2962),
    [anon_sym_foreach] = ACTIONS(2962),
    [anon_sym_last] = ACTIONS(2962),
    [anon_sym_next] = ACTIONS(2962),
    [anon_sym_redo] = ACTIONS(2962),
    [anon_sym_BEGIN] = ACTIONS(2962),
    [anon_sym_INIT] = ACTIONS(2962),
    [anon_sym_CHECK] = ACTIONS(2962),
    [anon_sym_UNITCHECK] = ACTIONS(2962),
    [anon_sym_END] = ACTIONS(2962),
    [anon_sym___FILE__] = ACTIONS(2962),
    [anon_sym___LINE__] = ACTIONS(2962),
    [anon_sym___PACKAGE__] = ACTIONS(2962),
    [anon_sym___SUB__] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_fork] = ACTIONS(2962),
    [anon_sym_getppid] = ACTIONS(2962),
    [anon_sym_time] = ACTIONS(2962),
    [anon_sym_times] = ACTIONS(2962),
    [anon_sym_wait] = ACTIONS(2962),
    [anon_sym_wantarray] = ACTIONS(2962),
    [anon_sym_abs] = ACTIONS(2962),
    [anon_sym_alarm] = ACTIONS(2962),
    [anon_sym_chop] = ACTIONS(2962),
    [anon_sym_chdir] = ACTIONS(2962),
    [anon_sym_close] = ACTIONS(2962),
    [anon_sym_closedir] = ACTIONS(2962),
    [anon_sym_caller] = ACTIONS(2962),
    [anon_sym_chomp] = ACTIONS(2962),
    [anon_sym_chr] = ACTIONS(2962),
    [anon_sym_cos] = ACTIONS(2962),
    [anon_sym_chroot] = ACTIONS(2962),
    [anon_sym_defined] = ACTIONS(2962),
    [anon_sym_delete] = ACTIONS(2962),
    [anon_sym_dbmclose] = ACTIONS(2962),
    [anon_sym_exists] = ACTIONS(2962),
    [anon_sym_exit] = ACTIONS(2962),
    [anon_sym_eof] = ACTIONS(2962),
    [anon_sym_exp] = ACTIONS(2962),
    [anon_sym_each] = ACTIONS(2962),
    [anon_sym_fc] = ACTIONS(2962),
    [anon_sym_fileno] = ACTIONS(2962),
    [anon_sym_gmtime] = ACTIONS(2962),
    [anon_sym_getc] = ACTIONS(2962),
    [anon_sym_getpgrp] = ACTIONS(2962),
    [anon_sym_getprotobyname] = ACTIONS(2962),
    [anon_sym_getpwname] = ACTIONS(2962),
    [anon_sym_getpwuid] = ACTIONS(2962),
    [anon_sym_getpeername] = ACTIONS(2962),
    [anon_sym_getnetbyname] = ACTIONS(2962),
    [anon_sym_getsockname] = ACTIONS(2962),
    [anon_sym_getgrnam] = ACTIONS(2962),
    [anon_sym_getgrgid] = ACTIONS(2962),
    [anon_sym_hex] = ACTIONS(2962),
    [anon_sym_int] = ACTIONS(2962),
    [anon_sym_keys] = ACTIONS(2962),
    [anon_sym_lc] = ACTIONS(2962),
    [anon_sym_lcfirst] = ACTIONS(2962),
    [anon_sym_length] = ACTIONS(2962),
    [anon_sym_localtime] = ACTIONS(2962),
    [anon_sym_log] = ACTIONS(2962),
    [anon_sym_lock] = ACTIONS(2962),
    [anon_sym_lstat] = ACTIONS(2962),
    [anon_sym_oct] = ACTIONS(2962),
    [anon_sym_ord] = ACTIONS(2962),
    [anon_sym_prototype] = ACTIONS(2962),
    [anon_sym_pop] = ACTIONS(2962),
    [anon_sym_pos] = ACTIONS(2962),
    [anon_sym_quotemeta] = ACTIONS(2962),
    [anon_sym_reset] = ACTIONS(2962),
    [anon_sym_rand] = ACTIONS(2962),
    [anon_sym_rmdir] = ACTIONS(2962),
    [anon_sym_readdir] = ACTIONS(2962),
    [anon_sym_readline] = ACTIONS(2962),
    [anon_sym_readpipe] = ACTIONS(2962),
    [anon_sym_rewinddir] = ACTIONS(2962),
    [anon_sym_readlink] = ACTIONS(2962),
    [anon_sym_ref] = ACTIONS(2962),
    [anon_sym_scalar] = ACTIONS(2962),
    [anon_sym_shift] = ACTIONS(2962),
    [anon_sym_sin] = ACTIONS(2962),
    [anon_sym_sleep] = ACTIONS(2962),
    [anon_sym_sqrt] = ACTIONS(2962),
    [anon_sym_srand] = ACTIONS(2962),
    [anon_sym_stat] = ACTIONS(2962),
    [anon_sym_study] = ACTIONS(2962),
    [anon_sym_tell] = ACTIONS(2962),
    [anon_sym_telldir] = ACTIONS(2962),
    [anon_sym_tied] = ACTIONS(2962),
    [anon_sym_uc] = ACTIONS(2962),
    [anon_sym_ucfirst] = ACTIONS(2962),
    [anon_sym_untie] = ACTIONS(2962),
    [anon_sym_umask] = ACTIONS(2962),
    [anon_sym_values] = ACTIONS(2962),
    [anon_sym_write] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2962),
    [anon_sym_qq] = ACTIONS(2962),
    [anon_sym_qw] = ACTIONS(2962),
    [anon_sym_qx] = ACTIONS(2962),
    [anon_sym_qr] = ACTIONS(2962),
    [anon_sym_m] = ACTIONS(2962),
    [anon_sym_s] = ACTIONS(2962),
    [anon_sym_tr] = ACTIONS(2962),
    [anon_sym_y] = ACTIONS(2962),
    [anon_sym_if] = ACTIONS(2962),
    [anon_sym_unless] = ACTIONS(2962),
    [anon_sym_while] = ACTIONS(2962),
    [anon_sym_until] = ACTIONS(2962),
    [aux_sym__bareword_token1] = ACTIONS(2962),
    [sym_number] = ACTIONS(2962),
    [anon_sym_true] = ACTIONS(2962),
    [anon_sym_false] = ACTIONS(2962),
    [sym__single_quote] = ACTIONS(2960),
    [sym__double_quote] = ACTIONS(2960),
    [sym__backtick_quote] = ACTIONS(2960),
    [sym__search_slash_quote] = ACTIONS(2960),
    [sym__open_readline_bracket] = ACTIONS(2960),
    [sym__open_fileglob_bracket] = ACTIONS(2960),
    [sym__PERLY_HEREDOC] = ACTIONS(2960),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2960),
    [sym__filetest] = ACTIONS(2960),
  },
  [STATE(1599)] = {
    [sym_heredoc_content] = STATE(1599),
    [sym__identifier] = ACTIONS(3057),
    [anon_sym_LBRACE] = ACTIONS(3057),
    [anon_sym_RBRACE] = ACTIONS(3057),
    [anon_sym_SEMI] = ACTIONS(3057),
    [anon_sym_package] = ACTIONS(3057),
    [anon_sym_class] = ACTIONS(3057),
    [anon_sym_role] = ACTIONS(3057),
    [anon_sym_BUILD] = ACTIONS(3057),
    [anon_sym_ADJUST] = ACTIONS(3057),
    [anon_sym_DOLLAR] = ACTIONS(3057),
    [anon_sym_AT] = ACTIONS(3057),
    [anon_sym_my] = ACTIONS(3057),
    [anon_sym_extended] = ACTIONS(3057),
    [anon_sym_async] = ACTIONS(3057),
    [anon_sym_sub] = ACTIONS(3057),
    [anon_sym_method] = ACTIONS(3057),
    [anon_sym_LPAREN] = ACTIONS(3057),
    [anon_sym_continue] = ACTIONS(3057),
    [anon_sym_state] = ACTIONS(3057),
    [anon_sym_our] = ACTIONS(3057),
    [anon_sym_try] = ACTIONS(3057),
    [anon_sym_defer] = ACTIONS(3057),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3057),
    [anon_sym_LBRACK] = ACTIONS(3057),
    [anon_sym_LT_LT] = ACTIONS(3057),
    [anon_sym_SLASH_SLASH] = ACTIONS(3057),
    [anon_sym_PLUS] = ACTIONS(3057),
    [anon_sym_DASH] = ACTIONS(3057),
    [anon_sym_TILDE] = ACTIONS(3057),
    [anon_sym_BANG] = ACTIONS(3057),
    [anon_sym_PLUS_PLUS] = ACTIONS(3057),
    [anon_sym_DASH_DASH] = ACTIONS(3057),
    [anon_sym_BSLASH] = ACTIONS(3055),
    [anon_sym_do] = ACTIONS(3057),
    [anon_sym_eval] = ACTIONS(3057),
    [anon_sym_await] = ACTIONS(3057),
    [anon_sym_field] = ACTIONS(3057),
    [anon_sym_local] = ACTIONS(3057),
    [anon_sym_dynamically] = ACTIONS(3057),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3057),
    [anon_sym_require] = ACTIONS(3057),
    [anon_sym_map] = ACTIONS(3057),
    [anon_sym_grep] = ACTIONS(3057),
    [anon_sym_sort] = ACTIONS(3057),
    [anon_sym_goto] = ACTIONS(3057),
    [anon_sym_return] = ACTIONS(3057),
    [anon_sym_undef] = ACTIONS(3057),
    [anon_sym_PERCENT2] = ACTIONS(3057),
    [anon_sym_AMP2] = ACTIONS(3057),
    [anon_sym_STAR2] = ACTIONS(3057),
    [anon_sym_use] = ACTIONS(3057),
    [anon_sym_no] = ACTIONS(3057),
    [anon_sym_for] = ACTIONS(3057),
    [anon_sym_foreach] = ACTIONS(3057),
    [anon_sym_last] = ACTIONS(3057),
    [anon_sym_next] = ACTIONS(3057),
    [anon_sym_redo] = ACTIONS(3057),
    [anon_sym_BEGIN] = ACTIONS(3057),
    [anon_sym_INIT] = ACTIONS(3057),
    [anon_sym_CHECK] = ACTIONS(3057),
    [anon_sym_UNITCHECK] = ACTIONS(3057),
    [anon_sym_END] = ACTIONS(3057),
    [anon_sym___FILE__] = ACTIONS(3057),
    [anon_sym___LINE__] = ACTIONS(3057),
    [anon_sym___PACKAGE__] = ACTIONS(3057),
    [anon_sym___SUB__] = ACTIONS(3057),
    [anon_sym_break] = ACTIONS(3057),
    [anon_sym_fork] = ACTIONS(3057),
    [anon_sym_getppid] = ACTIONS(3057),
    [anon_sym_time] = ACTIONS(3057),
    [anon_sym_times] = ACTIONS(3057),
    [anon_sym_wait] = ACTIONS(3057),
    [anon_sym_wantarray] = ACTIONS(3057),
    [anon_sym_abs] = ACTIONS(3057),
    [anon_sym_alarm] = ACTIONS(3057),
    [anon_sym_chop] = ACTIONS(3057),
    [anon_sym_chdir] = ACTIONS(3057),
    [anon_sym_close] = ACTIONS(3057),
    [anon_sym_closedir] = ACTIONS(3057),
    [anon_sym_caller] = ACTIONS(3057),
    [anon_sym_chomp] = ACTIONS(3057),
    [anon_sym_chr] = ACTIONS(3057),
    [anon_sym_cos] = ACTIONS(3057),
    [anon_sym_chroot] = ACTIONS(3057),
    [anon_sym_defined] = ACTIONS(3057),
    [anon_sym_delete] = ACTIONS(3057),
    [anon_sym_dbmclose] = ACTIONS(3057),
    [anon_sym_exists] = ACTIONS(3057),
    [anon_sym_exit] = ACTIONS(3057),
    [anon_sym_eof] = ACTIONS(3057),
    [anon_sym_exp] = ACTIONS(3057),
    [anon_sym_each] = ACTIONS(3057),
    [anon_sym_fc] = ACTIONS(3057),
    [anon_sym_fileno] = ACTIONS(3057),
    [anon_sym_gmtime] = ACTIONS(3057),
    [anon_sym_getc] = ACTIONS(3057),
    [anon_sym_getpgrp] = ACTIONS(3057),
    [anon_sym_getprotobyname] = ACTIONS(3057),
    [anon_sym_getpwname] = ACTIONS(3057),
    [anon_sym_getpwuid] = ACTIONS(3057),
    [anon_sym_getpeername] = ACTIONS(3057),
    [anon_sym_getnetbyname] = ACTIONS(3057),
    [anon_sym_getsockname] = ACTIONS(3057),
    [anon_sym_getgrnam] = ACTIONS(3057),
    [anon_sym_getgrgid] = ACTIONS(3057),
    [anon_sym_hex] = ACTIONS(3057),
    [anon_sym_int] = ACTIONS(3057),
    [anon_sym_keys] = ACTIONS(3057),
    [anon_sym_lc] = ACTIONS(3057),
    [anon_sym_lcfirst] = ACTIONS(3057),
    [anon_sym_length] = ACTIONS(3057),
    [anon_sym_localtime] = ACTIONS(3057),
    [anon_sym_log] = ACTIONS(3057),
    [anon_sym_lock] = ACTIONS(3057),
    [anon_sym_lstat] = ACTIONS(3057),
    [anon_sym_oct] = ACTIONS(3057),
    [anon_sym_ord] = ACTIONS(3057),
    [anon_sym_prototype] = ACTIONS(3057),
    [anon_sym_pop] = ACTIONS(3057),
    [anon_sym_pos] = ACTIONS(3057),
    [anon_sym_quotemeta] = ACTIONS(3057),
    [anon_sym_reset] = ACTIONS(3057),
    [anon_sym_rand] = ACTIONS(3057),
    [anon_sym_rmdir] = ACTIONS(3057),
    [anon_sym_readdir] = ACTIONS(3057),
    [anon_sym_readline] = ACTIONS(3057),
    [anon_sym_readpipe] = ACTIONS(3057),
    [anon_sym_rewinddir] = ACTIONS(3057),
    [anon_sym_readlink] = ACTIONS(3057),
    [anon_sym_ref] = ACTIONS(3057),
    [anon_sym_scalar] = ACTIONS(3057),
    [anon_sym_shift] = ACTIONS(3057),
    [anon_sym_sin] = ACTIONS(3057),
    [anon_sym_sleep] = ACTIONS(3057),
    [anon_sym_sqrt] = ACTIONS(3057),
    [anon_sym_srand] = ACTIONS(3057),
    [anon_sym_stat] = ACTIONS(3057),
    [anon_sym_study] = ACTIONS(3057),
    [anon_sym_tell] = ACTIONS(3057),
    [anon_sym_telldir] = ACTIONS(3057),
    [anon_sym_tied] = ACTIONS(3057),
    [anon_sym_uc] = ACTIONS(3057),
    [anon_sym_ucfirst] = ACTIONS(3057),
    [anon_sym_untie] = ACTIONS(3057),
    [anon_sym_umask] = ACTIONS(3057),
    [anon_sym_values] = ACTIONS(3057),
    [anon_sym_write] = ACTIONS(3057),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3057),
    [anon_sym_qq] = ACTIONS(3057),
    [anon_sym_qw] = ACTIONS(3057),
    [anon_sym_qx] = ACTIONS(3057),
    [anon_sym_qr] = ACTIONS(3057),
    [anon_sym_m] = ACTIONS(3057),
    [anon_sym_s] = ACTIONS(3057),
    [anon_sym_tr] = ACTIONS(3057),
    [anon_sym_y] = ACTIONS(3057),
    [anon_sym_if] = ACTIONS(3057),
    [anon_sym_unless] = ACTIONS(3057),
    [anon_sym_while] = ACTIONS(3057),
    [anon_sym_until] = ACTIONS(3057),
    [aux_sym__bareword_token1] = ACTIONS(3057),
    [sym_number] = ACTIONS(3057),
    [anon_sym_true] = ACTIONS(3057),
    [anon_sym_false] = ACTIONS(3057),
    [sym__single_quote] = ACTIONS(3055),
    [sym__double_quote] = ACTIONS(3055),
    [sym__backtick_quote] = ACTIONS(3055),
    [sym__search_slash_quote] = ACTIONS(3055),
    [sym__open_readline_bracket] = ACTIONS(3055),
    [sym__open_fileglob_bracket] = ACTIONS(3055),
    [sym__PERLY_HEREDOC] = ACTIONS(3055),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3055),
    [sym__filetest] = ACTIONS(3055),
  },
  [STATE(1600)] = {
    [sym_heredoc_content] = STATE(1600),
    [sym__identifier] = ACTIONS(3065),
    [anon_sym_LBRACE] = ACTIONS(3065),
    [anon_sym_RBRACE] = ACTIONS(3065),
    [anon_sym_SEMI] = ACTIONS(3065),
    [anon_sym_package] = ACTIONS(3065),
    [anon_sym_class] = ACTIONS(3065),
    [anon_sym_role] = ACTIONS(3065),
    [anon_sym_BUILD] = ACTIONS(3065),
    [anon_sym_ADJUST] = ACTIONS(3065),
    [anon_sym_DOLLAR] = ACTIONS(3065),
    [anon_sym_AT] = ACTIONS(3065),
    [anon_sym_my] = ACTIONS(3065),
    [anon_sym_extended] = ACTIONS(3065),
    [anon_sym_async] = ACTIONS(3065),
    [anon_sym_sub] = ACTIONS(3065),
    [anon_sym_method] = ACTIONS(3065),
    [anon_sym_LPAREN] = ACTIONS(3065),
    [anon_sym_continue] = ACTIONS(3065),
    [anon_sym_state] = ACTIONS(3065),
    [anon_sym_our] = ACTIONS(3065),
    [anon_sym_try] = ACTIONS(3065),
    [anon_sym_defer] = ACTIONS(3065),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3065),
    [anon_sym_LBRACK] = ACTIONS(3065),
    [anon_sym_LT_LT] = ACTIONS(3065),
    [anon_sym_SLASH_SLASH] = ACTIONS(3065),
    [anon_sym_PLUS] = ACTIONS(3065),
    [anon_sym_DASH] = ACTIONS(3065),
    [anon_sym_TILDE] = ACTIONS(3065),
    [anon_sym_BANG] = ACTIONS(3065),
    [anon_sym_PLUS_PLUS] = ACTIONS(3065),
    [anon_sym_DASH_DASH] = ACTIONS(3065),
    [anon_sym_BSLASH] = ACTIONS(3063),
    [anon_sym_do] = ACTIONS(3065),
    [anon_sym_eval] = ACTIONS(3065),
    [anon_sym_await] = ACTIONS(3065),
    [anon_sym_field] = ACTIONS(3065),
    [anon_sym_local] = ACTIONS(3065),
    [anon_sym_dynamically] = ACTIONS(3065),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3065),
    [anon_sym_require] = ACTIONS(3065),
    [anon_sym_map] = ACTIONS(3065),
    [anon_sym_grep] = ACTIONS(3065),
    [anon_sym_sort] = ACTIONS(3065),
    [anon_sym_goto] = ACTIONS(3065),
    [anon_sym_return] = ACTIONS(3065),
    [anon_sym_undef] = ACTIONS(3065),
    [anon_sym_PERCENT2] = ACTIONS(3065),
    [anon_sym_AMP2] = ACTIONS(3065),
    [anon_sym_STAR2] = ACTIONS(3065),
    [anon_sym_use] = ACTIONS(3065),
    [anon_sym_no] = ACTIONS(3065),
    [anon_sym_for] = ACTIONS(3065),
    [anon_sym_foreach] = ACTIONS(3065),
    [anon_sym_last] = ACTIONS(3065),
    [anon_sym_next] = ACTIONS(3065),
    [anon_sym_redo] = ACTIONS(3065),
    [anon_sym_BEGIN] = ACTIONS(3065),
    [anon_sym_INIT] = ACTIONS(3065),
    [anon_sym_CHECK] = ACTIONS(3065),
    [anon_sym_UNITCHECK] = ACTIONS(3065),
    [anon_sym_END] = ACTIONS(3065),
    [anon_sym___FILE__] = ACTIONS(3065),
    [anon_sym___LINE__] = ACTIONS(3065),
    [anon_sym___PACKAGE__] = ACTIONS(3065),
    [anon_sym___SUB__] = ACTIONS(3065),
    [anon_sym_break] = ACTIONS(3065),
    [anon_sym_fork] = ACTIONS(3065),
    [anon_sym_getppid] = ACTIONS(3065),
    [anon_sym_time] = ACTIONS(3065),
    [anon_sym_times] = ACTIONS(3065),
    [anon_sym_wait] = ACTIONS(3065),
    [anon_sym_wantarray] = ACTIONS(3065),
    [anon_sym_abs] = ACTIONS(3065),
    [anon_sym_alarm] = ACTIONS(3065),
    [anon_sym_chop] = ACTIONS(3065),
    [anon_sym_chdir] = ACTIONS(3065),
    [anon_sym_close] = ACTIONS(3065),
    [anon_sym_closedir] = ACTIONS(3065),
    [anon_sym_caller] = ACTIONS(3065),
    [anon_sym_chomp] = ACTIONS(3065),
    [anon_sym_chr] = ACTIONS(3065),
    [anon_sym_cos] = ACTIONS(3065),
    [anon_sym_chroot] = ACTIONS(3065),
    [anon_sym_defined] = ACTIONS(3065),
    [anon_sym_delete] = ACTIONS(3065),
    [anon_sym_dbmclose] = ACTIONS(3065),
    [anon_sym_exists] = ACTIONS(3065),
    [anon_sym_exit] = ACTIONS(3065),
    [anon_sym_eof] = ACTIONS(3065),
    [anon_sym_exp] = ACTIONS(3065),
    [anon_sym_each] = ACTIONS(3065),
    [anon_sym_fc] = ACTIONS(3065),
    [anon_sym_fileno] = ACTIONS(3065),
    [anon_sym_gmtime] = ACTIONS(3065),
    [anon_sym_getc] = ACTIONS(3065),
    [anon_sym_getpgrp] = ACTIONS(3065),
    [anon_sym_getprotobyname] = ACTIONS(3065),
    [anon_sym_getpwname] = ACTIONS(3065),
    [anon_sym_getpwuid] = ACTIONS(3065),
    [anon_sym_getpeername] = ACTIONS(3065),
    [anon_sym_getnetbyname] = ACTIONS(3065),
    [anon_sym_getsockname] = ACTIONS(3065),
    [anon_sym_getgrnam] = ACTIONS(3065),
    [anon_sym_getgrgid] = ACTIONS(3065),
    [anon_sym_hex] = ACTIONS(3065),
    [anon_sym_int] = ACTIONS(3065),
    [anon_sym_keys] = ACTIONS(3065),
    [anon_sym_lc] = ACTIONS(3065),
    [anon_sym_lcfirst] = ACTIONS(3065),
    [anon_sym_length] = ACTIONS(3065),
    [anon_sym_localtime] = ACTIONS(3065),
    [anon_sym_log] = ACTIONS(3065),
    [anon_sym_lock] = ACTIONS(3065),
    [anon_sym_lstat] = ACTIONS(3065),
    [anon_sym_oct] = ACTIONS(3065),
    [anon_sym_ord] = ACTIONS(3065),
    [anon_sym_prototype] = ACTIONS(3065),
    [anon_sym_pop] = ACTIONS(3065),
    [anon_sym_pos] = ACTIONS(3065),
    [anon_sym_quotemeta] = ACTIONS(3065),
    [anon_sym_reset] = ACTIONS(3065),
    [anon_sym_rand] = ACTIONS(3065),
    [anon_sym_rmdir] = ACTIONS(3065),
    [anon_sym_readdir] = ACTIONS(3065),
    [anon_sym_readline] = ACTIONS(3065),
    [anon_sym_readpipe] = ACTIONS(3065),
    [anon_sym_rewinddir] = ACTIONS(3065),
    [anon_sym_readlink] = ACTIONS(3065),
    [anon_sym_ref] = ACTIONS(3065),
    [anon_sym_scalar] = ACTIONS(3065),
    [anon_sym_shift] = ACTIONS(3065),
    [anon_sym_sin] = ACTIONS(3065),
    [anon_sym_sleep] = ACTIONS(3065),
    [anon_sym_sqrt] = ACTIONS(3065),
    [anon_sym_srand] = ACTIONS(3065),
    [anon_sym_stat] = ACTIONS(3065),
    [anon_sym_study] = ACTIONS(3065),
    [anon_sym_tell] = ACTIONS(3065),
    [anon_sym_telldir] = ACTIONS(3065),
    [anon_sym_tied] = ACTIONS(3065),
    [anon_sym_uc] = ACTIONS(3065),
    [anon_sym_ucfirst] = ACTIONS(3065),
    [anon_sym_untie] = ACTIONS(3065),
    [anon_sym_umask] = ACTIONS(3065),
    [anon_sym_values] = ACTIONS(3065),
    [anon_sym_write] = ACTIONS(3065),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3065),
    [anon_sym_qq] = ACTIONS(3065),
    [anon_sym_qw] = ACTIONS(3065),
    [anon_sym_qx] = ACTIONS(3065),
    [anon_sym_qr] = ACTIONS(3065),
    [anon_sym_m] = ACTIONS(3065),
    [anon_sym_s] = ACTIONS(3065),
    [anon_sym_tr] = ACTIONS(3065),
    [anon_sym_y] = ACTIONS(3065),
    [anon_sym_if] = ACTIONS(3065),
    [anon_sym_unless] = ACTIONS(3065),
    [anon_sym_while] = ACTIONS(3065),
    [anon_sym_until] = ACTIONS(3065),
    [aux_sym__bareword_token1] = ACTIONS(3065),
    [sym_number] = ACTIONS(3065),
    [anon_sym_true] = ACTIONS(3065),
    [anon_sym_false] = ACTIONS(3065),
    [sym__single_quote] = ACTIONS(3063),
    [sym__double_quote] = ACTIONS(3063),
    [sym__backtick_quote] = ACTIONS(3063),
    [sym__search_slash_quote] = ACTIONS(3063),
    [sym__open_readline_bracket] = ACTIONS(3063),
    [sym__open_fileglob_bracket] = ACTIONS(3063),
    [sym__PERLY_HEREDOC] = ACTIONS(3063),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3063),
    [sym__filetest] = ACTIONS(3063),
  },
  [STATE(1601)] = {
    [sym_heredoc_content] = STATE(1601),
    [sym__identifier] = ACTIONS(2914),
    [anon_sym_LBRACE] = ACTIONS(2914),
    [anon_sym_RBRACE] = ACTIONS(2914),
    [anon_sym_SEMI] = ACTIONS(2914),
    [anon_sym_package] = ACTIONS(2914),
    [anon_sym_class] = ACTIONS(2914),
    [anon_sym_role] = ACTIONS(2914),
    [anon_sym_BUILD] = ACTIONS(2914),
    [anon_sym_ADJUST] = ACTIONS(2914),
    [anon_sym_DOLLAR] = ACTIONS(2914),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_my] = ACTIONS(2914),
    [anon_sym_extended] = ACTIONS(2914),
    [anon_sym_async] = ACTIONS(2914),
    [anon_sym_sub] = ACTIONS(2914),
    [anon_sym_method] = ACTIONS(2914),
    [anon_sym_LPAREN] = ACTIONS(2914),
    [anon_sym_continue] = ACTIONS(2914),
    [anon_sym_state] = ACTIONS(2914),
    [anon_sym_our] = ACTIONS(2914),
    [anon_sym_try] = ACTIONS(2914),
    [anon_sym_defer] = ACTIONS(2914),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2914),
    [anon_sym_LT_LT] = ACTIONS(2914),
    [anon_sym_SLASH_SLASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2914),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_PLUS_PLUS] = ACTIONS(2914),
    [anon_sym_DASH_DASH] = ACTIONS(2914),
    [anon_sym_BSLASH] = ACTIONS(2912),
    [anon_sym_do] = ACTIONS(2914),
    [anon_sym_eval] = ACTIONS(2914),
    [anon_sym_await] = ACTIONS(2914),
    [anon_sym_field] = ACTIONS(2914),
    [anon_sym_local] = ACTIONS(2914),
    [anon_sym_dynamically] = ACTIONS(2914),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2914),
    [anon_sym_require] = ACTIONS(2914),
    [anon_sym_map] = ACTIONS(2914),
    [anon_sym_grep] = ACTIONS(2914),
    [anon_sym_sort] = ACTIONS(2914),
    [anon_sym_goto] = ACTIONS(2914),
    [anon_sym_return] = ACTIONS(2914),
    [anon_sym_undef] = ACTIONS(2914),
    [anon_sym_PERCENT2] = ACTIONS(2914),
    [anon_sym_AMP2] = ACTIONS(2914),
    [anon_sym_STAR2] = ACTIONS(2914),
    [anon_sym_use] = ACTIONS(2914),
    [anon_sym_no] = ACTIONS(2914),
    [anon_sym_for] = ACTIONS(2914),
    [anon_sym_foreach] = ACTIONS(2914),
    [anon_sym_last] = ACTIONS(2914),
    [anon_sym_next] = ACTIONS(2914),
    [anon_sym_redo] = ACTIONS(2914),
    [anon_sym_BEGIN] = ACTIONS(2914),
    [anon_sym_INIT] = ACTIONS(2914),
    [anon_sym_CHECK] = ACTIONS(2914),
    [anon_sym_UNITCHECK] = ACTIONS(2914),
    [anon_sym_END] = ACTIONS(2914),
    [anon_sym___FILE__] = ACTIONS(2914),
    [anon_sym___LINE__] = ACTIONS(2914),
    [anon_sym___PACKAGE__] = ACTIONS(2914),
    [anon_sym___SUB__] = ACTIONS(2914),
    [anon_sym_break] = ACTIONS(2914),
    [anon_sym_fork] = ACTIONS(2914),
    [anon_sym_getppid] = ACTIONS(2914),
    [anon_sym_time] = ACTIONS(2914),
    [anon_sym_times] = ACTIONS(2914),
    [anon_sym_wait] = ACTIONS(2914),
    [anon_sym_wantarray] = ACTIONS(2914),
    [anon_sym_abs] = ACTIONS(2914),
    [anon_sym_alarm] = ACTIONS(2914),
    [anon_sym_chop] = ACTIONS(2914),
    [anon_sym_chdir] = ACTIONS(2914),
    [anon_sym_close] = ACTIONS(2914),
    [anon_sym_closedir] = ACTIONS(2914),
    [anon_sym_caller] = ACTIONS(2914),
    [anon_sym_chomp] = ACTIONS(2914),
    [anon_sym_chr] = ACTIONS(2914),
    [anon_sym_cos] = ACTIONS(2914),
    [anon_sym_chroot] = ACTIONS(2914),
    [anon_sym_defined] = ACTIONS(2914),
    [anon_sym_delete] = ACTIONS(2914),
    [anon_sym_dbmclose] = ACTIONS(2914),
    [anon_sym_exists] = ACTIONS(2914),
    [anon_sym_exit] = ACTIONS(2914),
    [anon_sym_eof] = ACTIONS(2914),
    [anon_sym_exp] = ACTIONS(2914),
    [anon_sym_each] = ACTIONS(2914),
    [anon_sym_fc] = ACTIONS(2914),
    [anon_sym_fileno] = ACTIONS(2914),
    [anon_sym_gmtime] = ACTIONS(2914),
    [anon_sym_getc] = ACTIONS(2914),
    [anon_sym_getpgrp] = ACTIONS(2914),
    [anon_sym_getprotobyname] = ACTIONS(2914),
    [anon_sym_getpwname] = ACTIONS(2914),
    [anon_sym_getpwuid] = ACTIONS(2914),
    [anon_sym_getpeername] = ACTIONS(2914),
    [anon_sym_getnetbyname] = ACTIONS(2914),
    [anon_sym_getsockname] = ACTIONS(2914),
    [anon_sym_getgrnam] = ACTIONS(2914),
    [anon_sym_getgrgid] = ACTIONS(2914),
    [anon_sym_hex] = ACTIONS(2914),
    [anon_sym_int] = ACTIONS(2914),
    [anon_sym_keys] = ACTIONS(2914),
    [anon_sym_lc] = ACTIONS(2914),
    [anon_sym_lcfirst] = ACTIONS(2914),
    [anon_sym_length] = ACTIONS(2914),
    [anon_sym_localtime] = ACTIONS(2914),
    [anon_sym_log] = ACTIONS(2914),
    [anon_sym_lock] = ACTIONS(2914),
    [anon_sym_lstat] = ACTIONS(2914),
    [anon_sym_oct] = ACTIONS(2914),
    [anon_sym_ord] = ACTIONS(2914),
    [anon_sym_prototype] = ACTIONS(2914),
    [anon_sym_pop] = ACTIONS(2914),
    [anon_sym_pos] = ACTIONS(2914),
    [anon_sym_quotemeta] = ACTIONS(2914),
    [anon_sym_reset] = ACTIONS(2914),
    [anon_sym_rand] = ACTIONS(2914),
    [anon_sym_rmdir] = ACTIONS(2914),
    [anon_sym_readdir] = ACTIONS(2914),
    [anon_sym_readline] = ACTIONS(2914),
    [anon_sym_readpipe] = ACTIONS(2914),
    [anon_sym_rewinddir] = ACTIONS(2914),
    [anon_sym_readlink] = ACTIONS(2914),
    [anon_sym_ref] = ACTIONS(2914),
    [anon_sym_scalar] = ACTIONS(2914),
    [anon_sym_shift] = ACTIONS(2914),
    [anon_sym_sin] = ACTIONS(2914),
    [anon_sym_sleep] = ACTIONS(2914),
    [anon_sym_sqrt] = ACTIONS(2914),
    [anon_sym_srand] = ACTIONS(2914),
    [anon_sym_stat] = ACTIONS(2914),
    [anon_sym_study] = ACTIONS(2914),
    [anon_sym_tell] = ACTIONS(2914),
    [anon_sym_telldir] = ACTIONS(2914),
    [anon_sym_tied] = ACTIONS(2914),
    [anon_sym_uc] = ACTIONS(2914),
    [anon_sym_ucfirst] = ACTIONS(2914),
    [anon_sym_untie] = ACTIONS(2914),
    [anon_sym_umask] = ACTIONS(2914),
    [anon_sym_values] = ACTIONS(2914),
    [anon_sym_write] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2914),
    [anon_sym_qq] = ACTIONS(2914),
    [anon_sym_qw] = ACTIONS(2914),
    [anon_sym_qx] = ACTIONS(2914),
    [anon_sym_qr] = ACTIONS(2914),
    [anon_sym_m] = ACTIONS(2914),
    [anon_sym_s] = ACTIONS(2914),
    [anon_sym_tr] = ACTIONS(2914),
    [anon_sym_y] = ACTIONS(2914),
    [anon_sym_if] = ACTIONS(2914),
    [anon_sym_unless] = ACTIONS(2914),
    [anon_sym_while] = ACTIONS(2914),
    [anon_sym_until] = ACTIONS(2914),
    [aux_sym__bareword_token1] = ACTIONS(2914),
    [sym_number] = ACTIONS(2914),
    [anon_sym_true] = ACTIONS(2914),
    [anon_sym_false] = ACTIONS(2914),
    [sym__single_quote] = ACTIONS(2912),
    [sym__double_quote] = ACTIONS(2912),
    [sym__backtick_quote] = ACTIONS(2912),
    [sym__search_slash_quote] = ACTIONS(2912),
    [sym__open_readline_bracket] = ACTIONS(2912),
    [sym__open_fileglob_bracket] = ACTIONS(2912),
    [sym__PERLY_HEREDOC] = ACTIONS(2912),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2912),
    [sym__filetest] = ACTIONS(2912),
  },
  [STATE(1602)] = {
    [sym_heredoc_content] = STATE(1602),
    [sym__identifier] = ACTIONS(2950),
    [anon_sym_LBRACE] = ACTIONS(2950),
    [anon_sym_RBRACE] = ACTIONS(2950),
    [anon_sym_SEMI] = ACTIONS(2950),
    [anon_sym_package] = ACTIONS(2950),
    [anon_sym_class] = ACTIONS(2950),
    [anon_sym_role] = ACTIONS(2950),
    [anon_sym_BUILD] = ACTIONS(2950),
    [anon_sym_ADJUST] = ACTIONS(2950),
    [anon_sym_DOLLAR] = ACTIONS(2950),
    [anon_sym_AT] = ACTIONS(2950),
    [anon_sym_my] = ACTIONS(2950),
    [anon_sym_extended] = ACTIONS(2950),
    [anon_sym_async] = ACTIONS(2950),
    [anon_sym_sub] = ACTIONS(2950),
    [anon_sym_method] = ACTIONS(2950),
    [anon_sym_LPAREN] = ACTIONS(2950),
    [anon_sym_continue] = ACTIONS(2950),
    [anon_sym_state] = ACTIONS(2950),
    [anon_sym_our] = ACTIONS(2950),
    [anon_sym_try] = ACTIONS(2950),
    [anon_sym_defer] = ACTIONS(2950),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2950),
    [anon_sym_LBRACK] = ACTIONS(2950),
    [anon_sym_LT_LT] = ACTIONS(2950),
    [anon_sym_SLASH_SLASH] = ACTIONS(2950),
    [anon_sym_PLUS] = ACTIONS(2950),
    [anon_sym_DASH] = ACTIONS(2950),
    [anon_sym_TILDE] = ACTIONS(2950),
    [anon_sym_BANG] = ACTIONS(2950),
    [anon_sym_PLUS_PLUS] = ACTIONS(2950),
    [anon_sym_DASH_DASH] = ACTIONS(2950),
    [anon_sym_BSLASH] = ACTIONS(2948),
    [anon_sym_do] = ACTIONS(2950),
    [anon_sym_eval] = ACTIONS(2950),
    [anon_sym_await] = ACTIONS(2950),
    [anon_sym_field] = ACTIONS(2950),
    [anon_sym_local] = ACTIONS(2950),
    [anon_sym_dynamically] = ACTIONS(2950),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2950),
    [anon_sym_require] = ACTIONS(2950),
    [anon_sym_map] = ACTIONS(2950),
    [anon_sym_grep] = ACTIONS(2950),
    [anon_sym_sort] = ACTIONS(2950),
    [anon_sym_goto] = ACTIONS(2950),
    [anon_sym_return] = ACTIONS(2950),
    [anon_sym_undef] = ACTIONS(2950),
    [anon_sym_PERCENT2] = ACTIONS(2950),
    [anon_sym_AMP2] = ACTIONS(2950),
    [anon_sym_STAR2] = ACTIONS(2950),
    [anon_sym_use] = ACTIONS(2950),
    [anon_sym_no] = ACTIONS(2950),
    [anon_sym_for] = ACTIONS(2950),
    [anon_sym_foreach] = ACTIONS(2950),
    [anon_sym_last] = ACTIONS(2950),
    [anon_sym_next] = ACTIONS(2950),
    [anon_sym_redo] = ACTIONS(2950),
    [anon_sym_BEGIN] = ACTIONS(2950),
    [anon_sym_INIT] = ACTIONS(2950),
    [anon_sym_CHECK] = ACTIONS(2950),
    [anon_sym_UNITCHECK] = ACTIONS(2950),
    [anon_sym_END] = ACTIONS(2950),
    [anon_sym___FILE__] = ACTIONS(2950),
    [anon_sym___LINE__] = ACTIONS(2950),
    [anon_sym___PACKAGE__] = ACTIONS(2950),
    [anon_sym___SUB__] = ACTIONS(2950),
    [anon_sym_break] = ACTIONS(2950),
    [anon_sym_fork] = ACTIONS(2950),
    [anon_sym_getppid] = ACTIONS(2950),
    [anon_sym_time] = ACTIONS(2950),
    [anon_sym_times] = ACTIONS(2950),
    [anon_sym_wait] = ACTIONS(2950),
    [anon_sym_wantarray] = ACTIONS(2950),
    [anon_sym_abs] = ACTIONS(2950),
    [anon_sym_alarm] = ACTIONS(2950),
    [anon_sym_chop] = ACTIONS(2950),
    [anon_sym_chdir] = ACTIONS(2950),
    [anon_sym_close] = ACTIONS(2950),
    [anon_sym_closedir] = ACTIONS(2950),
    [anon_sym_caller] = ACTIONS(2950),
    [anon_sym_chomp] = ACTIONS(2950),
    [anon_sym_chr] = ACTIONS(2950),
    [anon_sym_cos] = ACTIONS(2950),
    [anon_sym_chroot] = ACTIONS(2950),
    [anon_sym_defined] = ACTIONS(2950),
    [anon_sym_delete] = ACTIONS(2950),
    [anon_sym_dbmclose] = ACTIONS(2950),
    [anon_sym_exists] = ACTIONS(2950),
    [anon_sym_exit] = ACTIONS(2950),
    [anon_sym_eof] = ACTIONS(2950),
    [anon_sym_exp] = ACTIONS(2950),
    [anon_sym_each] = ACTIONS(2950),
    [anon_sym_fc] = ACTIONS(2950),
    [anon_sym_fileno] = ACTIONS(2950),
    [anon_sym_gmtime] = ACTIONS(2950),
    [anon_sym_getc] = ACTIONS(2950),
    [anon_sym_getpgrp] = ACTIONS(2950),
    [anon_sym_getprotobyname] = ACTIONS(2950),
    [anon_sym_getpwname] = ACTIONS(2950),
    [anon_sym_getpwuid] = ACTIONS(2950),
    [anon_sym_getpeername] = ACTIONS(2950),
    [anon_sym_getnetbyname] = ACTIONS(2950),
    [anon_sym_getsockname] = ACTIONS(2950),
    [anon_sym_getgrnam] = ACTIONS(2950),
    [anon_sym_getgrgid] = ACTIONS(2950),
    [anon_sym_hex] = ACTIONS(2950),
    [anon_sym_int] = ACTIONS(2950),
    [anon_sym_keys] = ACTIONS(2950),
    [anon_sym_lc] = ACTIONS(2950),
    [anon_sym_lcfirst] = ACTIONS(2950),
    [anon_sym_length] = ACTIONS(2950),
    [anon_sym_localtime] = ACTIONS(2950),
    [anon_sym_log] = ACTIONS(2950),
    [anon_sym_lock] = ACTIONS(2950),
    [anon_sym_lstat] = ACTIONS(2950),
    [anon_sym_oct] = ACTIONS(2950),
    [anon_sym_ord] = ACTIONS(2950),
    [anon_sym_prototype] = ACTIONS(2950),
    [anon_sym_pop] = ACTIONS(2950),
    [anon_sym_pos] = ACTIONS(2950),
    [anon_sym_quotemeta] = ACTIONS(2950),
    [anon_sym_reset] = ACTIONS(2950),
    [anon_sym_rand] = ACTIONS(2950),
    [anon_sym_rmdir] = ACTIONS(2950),
    [anon_sym_readdir] = ACTIONS(2950),
    [anon_sym_readline] = ACTIONS(2950),
    [anon_sym_readpipe] = ACTIONS(2950),
    [anon_sym_rewinddir] = ACTIONS(2950),
    [anon_sym_readlink] = ACTIONS(2950),
    [anon_sym_ref] = ACTIONS(2950),
    [anon_sym_scalar] = ACTIONS(2950),
    [anon_sym_shift] = ACTIONS(2950),
    [anon_sym_sin] = ACTIONS(2950),
    [anon_sym_sleep] = ACTIONS(2950),
    [anon_sym_sqrt] = ACTIONS(2950),
    [anon_sym_srand] = ACTIONS(2950),
    [anon_sym_stat] = ACTIONS(2950),
    [anon_sym_study] = ACTIONS(2950),
    [anon_sym_tell] = ACTIONS(2950),
    [anon_sym_telldir] = ACTIONS(2950),
    [anon_sym_tied] = ACTIONS(2950),
    [anon_sym_uc] = ACTIONS(2950),
    [anon_sym_ucfirst] = ACTIONS(2950),
    [anon_sym_untie] = ACTIONS(2950),
    [anon_sym_umask] = ACTIONS(2950),
    [anon_sym_values] = ACTIONS(2950),
    [anon_sym_write] = ACTIONS(2950),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2950),
    [anon_sym_qq] = ACTIONS(2950),
    [anon_sym_qw] = ACTIONS(2950),
    [anon_sym_qx] = ACTIONS(2950),
    [anon_sym_qr] = ACTIONS(2950),
    [anon_sym_m] = ACTIONS(2950),
    [anon_sym_s] = ACTIONS(2950),
    [anon_sym_tr] = ACTIONS(2950),
    [anon_sym_y] = ACTIONS(2950),
    [anon_sym_if] = ACTIONS(2950),
    [anon_sym_unless] = ACTIONS(2950),
    [anon_sym_while] = ACTIONS(2950),
    [anon_sym_until] = ACTIONS(2950),
    [aux_sym__bareword_token1] = ACTIONS(2950),
    [sym_number] = ACTIONS(2950),
    [anon_sym_true] = ACTIONS(2950),
    [anon_sym_false] = ACTIONS(2950),
    [sym__single_quote] = ACTIONS(2948),
    [sym__double_quote] = ACTIONS(2948),
    [sym__backtick_quote] = ACTIONS(2948),
    [sym__search_slash_quote] = ACTIONS(2948),
    [sym__open_readline_bracket] = ACTIONS(2948),
    [sym__open_fileglob_bracket] = ACTIONS(2948),
    [sym__PERLY_HEREDOC] = ACTIONS(2948),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2948),
    [sym__filetest] = ACTIONS(2948),
  },
  [STATE(1603)] = {
    [sym_heredoc_content] = STATE(1603),
    [sym__identifier] = ACTIONS(3025),
    [anon_sym_LBRACE] = ACTIONS(3025),
    [anon_sym_RBRACE] = ACTIONS(3025),
    [anon_sym_SEMI] = ACTIONS(3025),
    [anon_sym_package] = ACTIONS(3025),
    [anon_sym_class] = ACTIONS(3025),
    [anon_sym_role] = ACTIONS(3025),
    [anon_sym_BUILD] = ACTIONS(3025),
    [anon_sym_ADJUST] = ACTIONS(3025),
    [anon_sym_DOLLAR] = ACTIONS(3025),
    [anon_sym_AT] = ACTIONS(3025),
    [anon_sym_my] = ACTIONS(3025),
    [anon_sym_extended] = ACTIONS(3025),
    [anon_sym_async] = ACTIONS(3025),
    [anon_sym_sub] = ACTIONS(3025),
    [anon_sym_method] = ACTIONS(3025),
    [anon_sym_LPAREN] = ACTIONS(3025),
    [anon_sym_continue] = ACTIONS(3025),
    [anon_sym_state] = ACTIONS(3025),
    [anon_sym_our] = ACTIONS(3025),
    [anon_sym_try] = ACTIONS(3025),
    [anon_sym_defer] = ACTIONS(3025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3025),
    [anon_sym_LBRACK] = ACTIONS(3025),
    [anon_sym_LT_LT] = ACTIONS(3025),
    [anon_sym_SLASH_SLASH] = ACTIONS(3025),
    [anon_sym_PLUS] = ACTIONS(3025),
    [anon_sym_DASH] = ACTIONS(3025),
    [anon_sym_TILDE] = ACTIONS(3025),
    [anon_sym_BANG] = ACTIONS(3025),
    [anon_sym_PLUS_PLUS] = ACTIONS(3025),
    [anon_sym_DASH_DASH] = ACTIONS(3025),
    [anon_sym_BSLASH] = ACTIONS(3023),
    [anon_sym_do] = ACTIONS(3025),
    [anon_sym_eval] = ACTIONS(3025),
    [anon_sym_await] = ACTIONS(3025),
    [anon_sym_field] = ACTIONS(3025),
    [anon_sym_local] = ACTIONS(3025),
    [anon_sym_dynamically] = ACTIONS(3025),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3025),
    [anon_sym_require] = ACTIONS(3025),
    [anon_sym_map] = ACTIONS(3025),
    [anon_sym_grep] = ACTIONS(3025),
    [anon_sym_sort] = ACTIONS(3025),
    [anon_sym_goto] = ACTIONS(3025),
    [anon_sym_return] = ACTIONS(3025),
    [anon_sym_undef] = ACTIONS(3025),
    [anon_sym_PERCENT2] = ACTIONS(3025),
    [anon_sym_AMP2] = ACTIONS(3025),
    [anon_sym_STAR2] = ACTIONS(3025),
    [anon_sym_use] = ACTIONS(3025),
    [anon_sym_no] = ACTIONS(3025),
    [anon_sym_for] = ACTIONS(3025),
    [anon_sym_foreach] = ACTIONS(3025),
    [anon_sym_last] = ACTIONS(3025),
    [anon_sym_next] = ACTIONS(3025),
    [anon_sym_redo] = ACTIONS(3025),
    [anon_sym_BEGIN] = ACTIONS(3025),
    [anon_sym_INIT] = ACTIONS(3025),
    [anon_sym_CHECK] = ACTIONS(3025),
    [anon_sym_UNITCHECK] = ACTIONS(3025),
    [anon_sym_END] = ACTIONS(3025),
    [anon_sym___FILE__] = ACTIONS(3025),
    [anon_sym___LINE__] = ACTIONS(3025),
    [anon_sym___PACKAGE__] = ACTIONS(3025),
    [anon_sym___SUB__] = ACTIONS(3025),
    [anon_sym_break] = ACTIONS(3025),
    [anon_sym_fork] = ACTIONS(3025),
    [anon_sym_getppid] = ACTIONS(3025),
    [anon_sym_time] = ACTIONS(3025),
    [anon_sym_times] = ACTIONS(3025),
    [anon_sym_wait] = ACTIONS(3025),
    [anon_sym_wantarray] = ACTIONS(3025),
    [anon_sym_abs] = ACTIONS(3025),
    [anon_sym_alarm] = ACTIONS(3025),
    [anon_sym_chop] = ACTIONS(3025),
    [anon_sym_chdir] = ACTIONS(3025),
    [anon_sym_close] = ACTIONS(3025),
    [anon_sym_closedir] = ACTIONS(3025),
    [anon_sym_caller] = ACTIONS(3025),
    [anon_sym_chomp] = ACTIONS(3025),
    [anon_sym_chr] = ACTIONS(3025),
    [anon_sym_cos] = ACTIONS(3025),
    [anon_sym_chroot] = ACTIONS(3025),
    [anon_sym_defined] = ACTIONS(3025),
    [anon_sym_delete] = ACTIONS(3025),
    [anon_sym_dbmclose] = ACTIONS(3025),
    [anon_sym_exists] = ACTIONS(3025),
    [anon_sym_exit] = ACTIONS(3025),
    [anon_sym_eof] = ACTIONS(3025),
    [anon_sym_exp] = ACTIONS(3025),
    [anon_sym_each] = ACTIONS(3025),
    [anon_sym_fc] = ACTIONS(3025),
    [anon_sym_fileno] = ACTIONS(3025),
    [anon_sym_gmtime] = ACTIONS(3025),
    [anon_sym_getc] = ACTIONS(3025),
    [anon_sym_getpgrp] = ACTIONS(3025),
    [anon_sym_getprotobyname] = ACTIONS(3025),
    [anon_sym_getpwname] = ACTIONS(3025),
    [anon_sym_getpwuid] = ACTIONS(3025),
    [anon_sym_getpeername] = ACTIONS(3025),
    [anon_sym_getnetbyname] = ACTIONS(3025),
    [anon_sym_getsockname] = ACTIONS(3025),
    [anon_sym_getgrnam] = ACTIONS(3025),
    [anon_sym_getgrgid] = ACTIONS(3025),
    [anon_sym_hex] = ACTIONS(3025),
    [anon_sym_int] = ACTIONS(3025),
    [anon_sym_keys] = ACTIONS(3025),
    [anon_sym_lc] = ACTIONS(3025),
    [anon_sym_lcfirst] = ACTIONS(3025),
    [anon_sym_length] = ACTIONS(3025),
    [anon_sym_localtime] = ACTIONS(3025),
    [anon_sym_log] = ACTIONS(3025),
    [anon_sym_lock] = ACTIONS(3025),
    [anon_sym_lstat] = ACTIONS(3025),
    [anon_sym_oct] = ACTIONS(3025),
    [anon_sym_ord] = ACTIONS(3025),
    [anon_sym_prototype] = ACTIONS(3025),
    [anon_sym_pop] = ACTIONS(3025),
    [anon_sym_pos] = ACTIONS(3025),
    [anon_sym_quotemeta] = ACTIONS(3025),
    [anon_sym_reset] = ACTIONS(3025),
    [anon_sym_rand] = ACTIONS(3025),
    [anon_sym_rmdir] = ACTIONS(3025),
    [anon_sym_readdir] = ACTIONS(3025),
    [anon_sym_readline] = ACTIONS(3025),
    [anon_sym_readpipe] = ACTIONS(3025),
    [anon_sym_rewinddir] = ACTIONS(3025),
    [anon_sym_readlink] = ACTIONS(3025),
    [anon_sym_ref] = ACTIONS(3025),
    [anon_sym_scalar] = ACTIONS(3025),
    [anon_sym_shift] = ACTIONS(3025),
    [anon_sym_sin] = ACTIONS(3025),
    [anon_sym_sleep] = ACTIONS(3025),
    [anon_sym_sqrt] = ACTIONS(3025),
    [anon_sym_srand] = ACTIONS(3025),
    [anon_sym_stat] = ACTIONS(3025),
    [anon_sym_study] = ACTIONS(3025),
    [anon_sym_tell] = ACTIONS(3025),
    [anon_sym_telldir] = ACTIONS(3025),
    [anon_sym_tied] = ACTIONS(3025),
    [anon_sym_uc] = ACTIONS(3025),
    [anon_sym_ucfirst] = ACTIONS(3025),
    [anon_sym_untie] = ACTIONS(3025),
    [anon_sym_umask] = ACTIONS(3025),
    [anon_sym_values] = ACTIONS(3025),
    [anon_sym_write] = ACTIONS(3025),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3025),
    [anon_sym_qq] = ACTIONS(3025),
    [anon_sym_qw] = ACTIONS(3025),
    [anon_sym_qx] = ACTIONS(3025),
    [anon_sym_qr] = ACTIONS(3025),
    [anon_sym_m] = ACTIONS(3025),
    [anon_sym_s] = ACTIONS(3025),
    [anon_sym_tr] = ACTIONS(3025),
    [anon_sym_y] = ACTIONS(3025),
    [anon_sym_if] = ACTIONS(3025),
    [anon_sym_unless] = ACTIONS(3025),
    [anon_sym_while] = ACTIONS(3025),
    [anon_sym_until] = ACTIONS(3025),
    [aux_sym__bareword_token1] = ACTIONS(3025),
    [sym_number] = ACTIONS(3025),
    [anon_sym_true] = ACTIONS(3025),
    [anon_sym_false] = ACTIONS(3025),
    [sym__single_quote] = ACTIONS(3023),
    [sym__double_quote] = ACTIONS(3023),
    [sym__backtick_quote] = ACTIONS(3023),
    [sym__search_slash_quote] = ACTIONS(3023),
    [sym__open_readline_bracket] = ACTIONS(3023),
    [sym__open_fileglob_bracket] = ACTIONS(3023),
    [sym__PERLY_HEREDOC] = ACTIONS(3023),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3023),
    [sym__filetest] = ACTIONS(3023),
  },
  [STATE(1604)] = {
    [sym_heredoc_content] = STATE(1604),
    [sym__identifier] = ACTIONS(3053),
    [anon_sym_LBRACE] = ACTIONS(3053),
    [anon_sym_RBRACE] = ACTIONS(3053),
    [anon_sym_SEMI] = ACTIONS(3053),
    [anon_sym_package] = ACTIONS(3053),
    [anon_sym_class] = ACTIONS(3053),
    [anon_sym_role] = ACTIONS(3053),
    [anon_sym_BUILD] = ACTIONS(3053),
    [anon_sym_ADJUST] = ACTIONS(3053),
    [anon_sym_DOLLAR] = ACTIONS(3053),
    [anon_sym_AT] = ACTIONS(3053),
    [anon_sym_my] = ACTIONS(3053),
    [anon_sym_extended] = ACTIONS(3053),
    [anon_sym_async] = ACTIONS(3053),
    [anon_sym_sub] = ACTIONS(3053),
    [anon_sym_method] = ACTIONS(3053),
    [anon_sym_LPAREN] = ACTIONS(3053),
    [anon_sym_continue] = ACTIONS(3053),
    [anon_sym_state] = ACTIONS(3053),
    [anon_sym_our] = ACTIONS(3053),
    [anon_sym_try] = ACTIONS(3053),
    [anon_sym_defer] = ACTIONS(3053),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3053),
    [anon_sym_LBRACK] = ACTIONS(3053),
    [anon_sym_LT_LT] = ACTIONS(3053),
    [anon_sym_SLASH_SLASH] = ACTIONS(3053),
    [anon_sym_PLUS] = ACTIONS(3053),
    [anon_sym_DASH] = ACTIONS(3053),
    [anon_sym_TILDE] = ACTIONS(3053),
    [anon_sym_BANG] = ACTIONS(3053),
    [anon_sym_PLUS_PLUS] = ACTIONS(3053),
    [anon_sym_DASH_DASH] = ACTIONS(3053),
    [anon_sym_BSLASH] = ACTIONS(3051),
    [anon_sym_do] = ACTIONS(3053),
    [anon_sym_eval] = ACTIONS(3053),
    [anon_sym_await] = ACTIONS(3053),
    [anon_sym_field] = ACTIONS(3053),
    [anon_sym_local] = ACTIONS(3053),
    [anon_sym_dynamically] = ACTIONS(3053),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3053),
    [anon_sym_require] = ACTIONS(3053),
    [anon_sym_map] = ACTIONS(3053),
    [anon_sym_grep] = ACTIONS(3053),
    [anon_sym_sort] = ACTIONS(3053),
    [anon_sym_goto] = ACTIONS(3053),
    [anon_sym_return] = ACTIONS(3053),
    [anon_sym_undef] = ACTIONS(3053),
    [anon_sym_PERCENT2] = ACTIONS(3053),
    [anon_sym_AMP2] = ACTIONS(3053),
    [anon_sym_STAR2] = ACTIONS(3053),
    [anon_sym_use] = ACTIONS(3053),
    [anon_sym_no] = ACTIONS(3053),
    [anon_sym_for] = ACTIONS(3053),
    [anon_sym_foreach] = ACTIONS(3053),
    [anon_sym_last] = ACTIONS(3053),
    [anon_sym_next] = ACTIONS(3053),
    [anon_sym_redo] = ACTIONS(3053),
    [anon_sym_BEGIN] = ACTIONS(3053),
    [anon_sym_INIT] = ACTIONS(3053),
    [anon_sym_CHECK] = ACTIONS(3053),
    [anon_sym_UNITCHECK] = ACTIONS(3053),
    [anon_sym_END] = ACTIONS(3053),
    [anon_sym___FILE__] = ACTIONS(3053),
    [anon_sym___LINE__] = ACTIONS(3053),
    [anon_sym___PACKAGE__] = ACTIONS(3053),
    [anon_sym___SUB__] = ACTIONS(3053),
    [anon_sym_break] = ACTIONS(3053),
    [anon_sym_fork] = ACTIONS(3053),
    [anon_sym_getppid] = ACTIONS(3053),
    [anon_sym_time] = ACTIONS(3053),
    [anon_sym_times] = ACTIONS(3053),
    [anon_sym_wait] = ACTIONS(3053),
    [anon_sym_wantarray] = ACTIONS(3053),
    [anon_sym_abs] = ACTIONS(3053),
    [anon_sym_alarm] = ACTIONS(3053),
    [anon_sym_chop] = ACTIONS(3053),
    [anon_sym_chdir] = ACTIONS(3053),
    [anon_sym_close] = ACTIONS(3053),
    [anon_sym_closedir] = ACTIONS(3053),
    [anon_sym_caller] = ACTIONS(3053),
    [anon_sym_chomp] = ACTIONS(3053),
    [anon_sym_chr] = ACTIONS(3053),
    [anon_sym_cos] = ACTIONS(3053),
    [anon_sym_chroot] = ACTIONS(3053),
    [anon_sym_defined] = ACTIONS(3053),
    [anon_sym_delete] = ACTIONS(3053),
    [anon_sym_dbmclose] = ACTIONS(3053),
    [anon_sym_exists] = ACTIONS(3053),
    [anon_sym_exit] = ACTIONS(3053),
    [anon_sym_eof] = ACTIONS(3053),
    [anon_sym_exp] = ACTIONS(3053),
    [anon_sym_each] = ACTIONS(3053),
    [anon_sym_fc] = ACTIONS(3053),
    [anon_sym_fileno] = ACTIONS(3053),
    [anon_sym_gmtime] = ACTIONS(3053),
    [anon_sym_getc] = ACTIONS(3053),
    [anon_sym_getpgrp] = ACTIONS(3053),
    [anon_sym_getprotobyname] = ACTIONS(3053),
    [anon_sym_getpwname] = ACTIONS(3053),
    [anon_sym_getpwuid] = ACTIONS(3053),
    [anon_sym_getpeername] = ACTIONS(3053),
    [anon_sym_getnetbyname] = ACTIONS(3053),
    [anon_sym_getsockname] = ACTIONS(3053),
    [anon_sym_getgrnam] = ACTIONS(3053),
    [anon_sym_getgrgid] = ACTIONS(3053),
    [anon_sym_hex] = ACTIONS(3053),
    [anon_sym_int] = ACTIONS(3053),
    [anon_sym_keys] = ACTIONS(3053),
    [anon_sym_lc] = ACTIONS(3053),
    [anon_sym_lcfirst] = ACTIONS(3053),
    [anon_sym_length] = ACTIONS(3053),
    [anon_sym_localtime] = ACTIONS(3053),
    [anon_sym_log] = ACTIONS(3053),
    [anon_sym_lock] = ACTIONS(3053),
    [anon_sym_lstat] = ACTIONS(3053),
    [anon_sym_oct] = ACTIONS(3053),
    [anon_sym_ord] = ACTIONS(3053),
    [anon_sym_prototype] = ACTIONS(3053),
    [anon_sym_pop] = ACTIONS(3053),
    [anon_sym_pos] = ACTIONS(3053),
    [anon_sym_quotemeta] = ACTIONS(3053),
    [anon_sym_reset] = ACTIONS(3053),
    [anon_sym_rand] = ACTIONS(3053),
    [anon_sym_rmdir] = ACTIONS(3053),
    [anon_sym_readdir] = ACTIONS(3053),
    [anon_sym_readline] = ACTIONS(3053),
    [anon_sym_readpipe] = ACTIONS(3053),
    [anon_sym_rewinddir] = ACTIONS(3053),
    [anon_sym_readlink] = ACTIONS(3053),
    [anon_sym_ref] = ACTIONS(3053),
    [anon_sym_scalar] = ACTIONS(3053),
    [anon_sym_shift] = ACTIONS(3053),
    [anon_sym_sin] = ACTIONS(3053),
    [anon_sym_sleep] = ACTIONS(3053),
    [anon_sym_sqrt] = ACTIONS(3053),
    [anon_sym_srand] = ACTIONS(3053),
    [anon_sym_stat] = ACTIONS(3053),
    [anon_sym_study] = ACTIONS(3053),
    [anon_sym_tell] = ACTIONS(3053),
    [anon_sym_telldir] = ACTIONS(3053),
    [anon_sym_tied] = ACTIONS(3053),
    [anon_sym_uc] = ACTIONS(3053),
    [anon_sym_ucfirst] = ACTIONS(3053),
    [anon_sym_untie] = ACTIONS(3053),
    [anon_sym_umask] = ACTIONS(3053),
    [anon_sym_values] = ACTIONS(3053),
    [anon_sym_write] = ACTIONS(3053),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3053),
    [anon_sym_qq] = ACTIONS(3053),
    [anon_sym_qw] = ACTIONS(3053),
    [anon_sym_qx] = ACTIONS(3053),
    [anon_sym_qr] = ACTIONS(3053),
    [anon_sym_m] = ACTIONS(3053),
    [anon_sym_s] = ACTIONS(3053),
    [anon_sym_tr] = ACTIONS(3053),
    [anon_sym_y] = ACTIONS(3053),
    [anon_sym_if] = ACTIONS(3053),
    [anon_sym_unless] = ACTIONS(3053),
    [anon_sym_while] = ACTIONS(3053),
    [anon_sym_until] = ACTIONS(3053),
    [aux_sym__bareword_token1] = ACTIONS(3053),
    [sym_number] = ACTIONS(3053),
    [anon_sym_true] = ACTIONS(3053),
    [anon_sym_false] = ACTIONS(3053),
    [sym__single_quote] = ACTIONS(3051),
    [sym__double_quote] = ACTIONS(3051),
    [sym__backtick_quote] = ACTIONS(3051),
    [sym__search_slash_quote] = ACTIONS(3051),
    [sym__open_readline_bracket] = ACTIONS(3051),
    [sym__open_fileglob_bracket] = ACTIONS(3051),
    [sym__PERLY_HEREDOC] = ACTIONS(3051),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3051),
    [sym__filetest] = ACTIONS(3051),
  },
  [STATE(1605)] = {
    [sym_heredoc_content] = STATE(1605),
    [sym__identifier] = ACTIONS(3073),
    [anon_sym_LBRACE] = ACTIONS(3073),
    [anon_sym_RBRACE] = ACTIONS(3073),
    [anon_sym_SEMI] = ACTIONS(3073),
    [anon_sym_package] = ACTIONS(3073),
    [anon_sym_class] = ACTIONS(3073),
    [anon_sym_role] = ACTIONS(3073),
    [anon_sym_BUILD] = ACTIONS(3073),
    [anon_sym_ADJUST] = ACTIONS(3073),
    [anon_sym_DOLLAR] = ACTIONS(3073),
    [anon_sym_AT] = ACTIONS(3073),
    [anon_sym_my] = ACTIONS(3073),
    [anon_sym_extended] = ACTIONS(3073),
    [anon_sym_async] = ACTIONS(3073),
    [anon_sym_sub] = ACTIONS(3073),
    [anon_sym_method] = ACTIONS(3073),
    [anon_sym_LPAREN] = ACTIONS(3073),
    [anon_sym_continue] = ACTIONS(3073),
    [anon_sym_state] = ACTIONS(3073),
    [anon_sym_our] = ACTIONS(3073),
    [anon_sym_try] = ACTIONS(3073),
    [anon_sym_defer] = ACTIONS(3073),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3073),
    [anon_sym_LBRACK] = ACTIONS(3073),
    [anon_sym_LT_LT] = ACTIONS(3073),
    [anon_sym_SLASH_SLASH] = ACTIONS(3073),
    [anon_sym_PLUS] = ACTIONS(3073),
    [anon_sym_DASH] = ACTIONS(3073),
    [anon_sym_TILDE] = ACTIONS(3073),
    [anon_sym_BANG] = ACTIONS(3073),
    [anon_sym_PLUS_PLUS] = ACTIONS(3073),
    [anon_sym_DASH_DASH] = ACTIONS(3073),
    [anon_sym_BSLASH] = ACTIONS(3071),
    [anon_sym_do] = ACTIONS(3073),
    [anon_sym_eval] = ACTIONS(3073),
    [anon_sym_await] = ACTIONS(3073),
    [anon_sym_field] = ACTIONS(3073),
    [anon_sym_local] = ACTIONS(3073),
    [anon_sym_dynamically] = ACTIONS(3073),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3073),
    [anon_sym_require] = ACTIONS(3073),
    [anon_sym_map] = ACTIONS(3073),
    [anon_sym_grep] = ACTIONS(3073),
    [anon_sym_sort] = ACTIONS(3073),
    [anon_sym_goto] = ACTIONS(3073),
    [anon_sym_return] = ACTIONS(3073),
    [anon_sym_undef] = ACTIONS(3073),
    [anon_sym_PERCENT2] = ACTIONS(3073),
    [anon_sym_AMP2] = ACTIONS(3073),
    [anon_sym_STAR2] = ACTIONS(3073),
    [anon_sym_use] = ACTIONS(3073),
    [anon_sym_no] = ACTIONS(3073),
    [anon_sym_for] = ACTIONS(3073),
    [anon_sym_foreach] = ACTIONS(3073),
    [anon_sym_last] = ACTIONS(3073),
    [anon_sym_next] = ACTIONS(3073),
    [anon_sym_redo] = ACTIONS(3073),
    [anon_sym_BEGIN] = ACTIONS(3073),
    [anon_sym_INIT] = ACTIONS(3073),
    [anon_sym_CHECK] = ACTIONS(3073),
    [anon_sym_UNITCHECK] = ACTIONS(3073),
    [anon_sym_END] = ACTIONS(3073),
    [anon_sym___FILE__] = ACTIONS(3073),
    [anon_sym___LINE__] = ACTIONS(3073),
    [anon_sym___PACKAGE__] = ACTIONS(3073),
    [anon_sym___SUB__] = ACTIONS(3073),
    [anon_sym_break] = ACTIONS(3073),
    [anon_sym_fork] = ACTIONS(3073),
    [anon_sym_getppid] = ACTIONS(3073),
    [anon_sym_time] = ACTIONS(3073),
    [anon_sym_times] = ACTIONS(3073),
    [anon_sym_wait] = ACTIONS(3073),
    [anon_sym_wantarray] = ACTIONS(3073),
    [anon_sym_abs] = ACTIONS(3073),
    [anon_sym_alarm] = ACTIONS(3073),
    [anon_sym_chop] = ACTIONS(3073),
    [anon_sym_chdir] = ACTIONS(3073),
    [anon_sym_close] = ACTIONS(3073),
    [anon_sym_closedir] = ACTIONS(3073),
    [anon_sym_caller] = ACTIONS(3073),
    [anon_sym_chomp] = ACTIONS(3073),
    [anon_sym_chr] = ACTIONS(3073),
    [anon_sym_cos] = ACTIONS(3073),
    [anon_sym_chroot] = ACTIONS(3073),
    [anon_sym_defined] = ACTIONS(3073),
    [anon_sym_delete] = ACTIONS(3073),
    [anon_sym_dbmclose] = ACTIONS(3073),
    [anon_sym_exists] = ACTIONS(3073),
    [anon_sym_exit] = ACTIONS(3073),
    [anon_sym_eof] = ACTIONS(3073),
    [anon_sym_exp] = ACTIONS(3073),
    [anon_sym_each] = ACTIONS(3073),
    [anon_sym_fc] = ACTIONS(3073),
    [anon_sym_fileno] = ACTIONS(3073),
    [anon_sym_gmtime] = ACTIONS(3073),
    [anon_sym_getc] = ACTIONS(3073),
    [anon_sym_getpgrp] = ACTIONS(3073),
    [anon_sym_getprotobyname] = ACTIONS(3073),
    [anon_sym_getpwname] = ACTIONS(3073),
    [anon_sym_getpwuid] = ACTIONS(3073),
    [anon_sym_getpeername] = ACTIONS(3073),
    [anon_sym_getnetbyname] = ACTIONS(3073),
    [anon_sym_getsockname] = ACTIONS(3073),
    [anon_sym_getgrnam] = ACTIONS(3073),
    [anon_sym_getgrgid] = ACTIONS(3073),
    [anon_sym_hex] = ACTIONS(3073),
    [anon_sym_int] = ACTIONS(3073),
    [anon_sym_keys] = ACTIONS(3073),
    [anon_sym_lc] = ACTIONS(3073),
    [anon_sym_lcfirst] = ACTIONS(3073),
    [anon_sym_length] = ACTIONS(3073),
    [anon_sym_localtime] = ACTIONS(3073),
    [anon_sym_log] = ACTIONS(3073),
    [anon_sym_lock] = ACTIONS(3073),
    [anon_sym_lstat] = ACTIONS(3073),
    [anon_sym_oct] = ACTIONS(3073),
    [anon_sym_ord] = ACTIONS(3073),
    [anon_sym_prototype] = ACTIONS(3073),
    [anon_sym_pop] = ACTIONS(3073),
    [anon_sym_pos] = ACTIONS(3073),
    [anon_sym_quotemeta] = ACTIONS(3073),
    [anon_sym_reset] = ACTIONS(3073),
    [anon_sym_rand] = ACTIONS(3073),
    [anon_sym_rmdir] = ACTIONS(3073),
    [anon_sym_readdir] = ACTIONS(3073),
    [anon_sym_readline] = ACTIONS(3073),
    [anon_sym_readpipe] = ACTIONS(3073),
    [anon_sym_rewinddir] = ACTIONS(3073),
    [anon_sym_readlink] = ACTIONS(3073),
    [anon_sym_ref] = ACTIONS(3073),
    [anon_sym_scalar] = ACTIONS(3073),
    [anon_sym_shift] = ACTIONS(3073),
    [anon_sym_sin] = ACTIONS(3073),
    [anon_sym_sleep] = ACTIONS(3073),
    [anon_sym_sqrt] = ACTIONS(3073),
    [anon_sym_srand] = ACTIONS(3073),
    [anon_sym_stat] = ACTIONS(3073),
    [anon_sym_study] = ACTIONS(3073),
    [anon_sym_tell] = ACTIONS(3073),
    [anon_sym_telldir] = ACTIONS(3073),
    [anon_sym_tied] = ACTIONS(3073),
    [anon_sym_uc] = ACTIONS(3073),
    [anon_sym_ucfirst] = ACTIONS(3073),
    [anon_sym_untie] = ACTIONS(3073),
    [anon_sym_umask] = ACTIONS(3073),
    [anon_sym_values] = ACTIONS(3073),
    [anon_sym_write] = ACTIONS(3073),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3073),
    [anon_sym_qq] = ACTIONS(3073),
    [anon_sym_qw] = ACTIONS(3073),
    [anon_sym_qx] = ACTIONS(3073),
    [anon_sym_qr] = ACTIONS(3073),
    [anon_sym_m] = ACTIONS(3073),
    [anon_sym_s] = ACTIONS(3073),
    [anon_sym_tr] = ACTIONS(3073),
    [anon_sym_y] = ACTIONS(3073),
    [anon_sym_if] = ACTIONS(3073),
    [anon_sym_unless] = ACTIONS(3073),
    [anon_sym_while] = ACTIONS(3073),
    [anon_sym_until] = ACTIONS(3073),
    [aux_sym__bareword_token1] = ACTIONS(3073),
    [sym_number] = ACTIONS(3073),
    [anon_sym_true] = ACTIONS(3073),
    [anon_sym_false] = ACTIONS(3073),
    [sym__single_quote] = ACTIONS(3071),
    [sym__double_quote] = ACTIONS(3071),
    [sym__backtick_quote] = ACTIONS(3071),
    [sym__search_slash_quote] = ACTIONS(3071),
    [sym__open_readline_bracket] = ACTIONS(3071),
    [sym__open_fileglob_bracket] = ACTIONS(3071),
    [sym__PERLY_HEREDOC] = ACTIONS(3071),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3071),
    [sym__filetest] = ACTIONS(3071),
  },
  [STATE(1606)] = {
    [sym_heredoc_content] = STATE(1606),
    [sym__identifier] = ACTIONS(3077),
    [anon_sym_LBRACE] = ACTIONS(3077),
    [anon_sym_RBRACE] = ACTIONS(3077),
    [anon_sym_SEMI] = ACTIONS(3077),
    [anon_sym_package] = ACTIONS(3077),
    [anon_sym_class] = ACTIONS(3077),
    [anon_sym_role] = ACTIONS(3077),
    [anon_sym_BUILD] = ACTIONS(3077),
    [anon_sym_ADJUST] = ACTIONS(3077),
    [anon_sym_DOLLAR] = ACTIONS(3077),
    [anon_sym_AT] = ACTIONS(3077),
    [anon_sym_my] = ACTIONS(3077),
    [anon_sym_extended] = ACTIONS(3077),
    [anon_sym_async] = ACTIONS(3077),
    [anon_sym_sub] = ACTIONS(3077),
    [anon_sym_method] = ACTIONS(3077),
    [anon_sym_LPAREN] = ACTIONS(3077),
    [anon_sym_continue] = ACTIONS(3077),
    [anon_sym_state] = ACTIONS(3077),
    [anon_sym_our] = ACTIONS(3077),
    [anon_sym_try] = ACTIONS(3077),
    [anon_sym_defer] = ACTIONS(3077),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3077),
    [anon_sym_LBRACK] = ACTIONS(3077),
    [anon_sym_LT_LT] = ACTIONS(3077),
    [anon_sym_SLASH_SLASH] = ACTIONS(3077),
    [anon_sym_PLUS] = ACTIONS(3077),
    [anon_sym_DASH] = ACTIONS(3077),
    [anon_sym_TILDE] = ACTIONS(3077),
    [anon_sym_BANG] = ACTIONS(3077),
    [anon_sym_PLUS_PLUS] = ACTIONS(3077),
    [anon_sym_DASH_DASH] = ACTIONS(3077),
    [anon_sym_BSLASH] = ACTIONS(3075),
    [anon_sym_do] = ACTIONS(3077),
    [anon_sym_eval] = ACTIONS(3077),
    [anon_sym_await] = ACTIONS(3077),
    [anon_sym_field] = ACTIONS(3077),
    [anon_sym_local] = ACTIONS(3077),
    [anon_sym_dynamically] = ACTIONS(3077),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3077),
    [anon_sym_require] = ACTIONS(3077),
    [anon_sym_map] = ACTIONS(3077),
    [anon_sym_grep] = ACTIONS(3077),
    [anon_sym_sort] = ACTIONS(3077),
    [anon_sym_goto] = ACTIONS(3077),
    [anon_sym_return] = ACTIONS(3077),
    [anon_sym_undef] = ACTIONS(3077),
    [anon_sym_PERCENT2] = ACTIONS(3077),
    [anon_sym_AMP2] = ACTIONS(3077),
    [anon_sym_STAR2] = ACTIONS(3077),
    [anon_sym_use] = ACTIONS(3077),
    [anon_sym_no] = ACTIONS(3077),
    [anon_sym_for] = ACTIONS(3077),
    [anon_sym_foreach] = ACTIONS(3077),
    [anon_sym_last] = ACTIONS(3077),
    [anon_sym_next] = ACTIONS(3077),
    [anon_sym_redo] = ACTIONS(3077),
    [anon_sym_BEGIN] = ACTIONS(3077),
    [anon_sym_INIT] = ACTIONS(3077),
    [anon_sym_CHECK] = ACTIONS(3077),
    [anon_sym_UNITCHECK] = ACTIONS(3077),
    [anon_sym_END] = ACTIONS(3077),
    [anon_sym___FILE__] = ACTIONS(3077),
    [anon_sym___LINE__] = ACTIONS(3077),
    [anon_sym___PACKAGE__] = ACTIONS(3077),
    [anon_sym___SUB__] = ACTIONS(3077),
    [anon_sym_break] = ACTIONS(3077),
    [anon_sym_fork] = ACTIONS(3077),
    [anon_sym_getppid] = ACTIONS(3077),
    [anon_sym_time] = ACTIONS(3077),
    [anon_sym_times] = ACTIONS(3077),
    [anon_sym_wait] = ACTIONS(3077),
    [anon_sym_wantarray] = ACTIONS(3077),
    [anon_sym_abs] = ACTIONS(3077),
    [anon_sym_alarm] = ACTIONS(3077),
    [anon_sym_chop] = ACTIONS(3077),
    [anon_sym_chdir] = ACTIONS(3077),
    [anon_sym_close] = ACTIONS(3077),
    [anon_sym_closedir] = ACTIONS(3077),
    [anon_sym_caller] = ACTIONS(3077),
    [anon_sym_chomp] = ACTIONS(3077),
    [anon_sym_chr] = ACTIONS(3077),
    [anon_sym_cos] = ACTIONS(3077),
    [anon_sym_chroot] = ACTIONS(3077),
    [anon_sym_defined] = ACTIONS(3077),
    [anon_sym_delete] = ACTIONS(3077),
    [anon_sym_dbmclose] = ACTIONS(3077),
    [anon_sym_exists] = ACTIONS(3077),
    [anon_sym_exit] = ACTIONS(3077),
    [anon_sym_eof] = ACTIONS(3077),
    [anon_sym_exp] = ACTIONS(3077),
    [anon_sym_each] = ACTIONS(3077),
    [anon_sym_fc] = ACTIONS(3077),
    [anon_sym_fileno] = ACTIONS(3077),
    [anon_sym_gmtime] = ACTIONS(3077),
    [anon_sym_getc] = ACTIONS(3077),
    [anon_sym_getpgrp] = ACTIONS(3077),
    [anon_sym_getprotobyname] = ACTIONS(3077),
    [anon_sym_getpwname] = ACTIONS(3077),
    [anon_sym_getpwuid] = ACTIONS(3077),
    [anon_sym_getpeername] = ACTIONS(3077),
    [anon_sym_getnetbyname] = ACTIONS(3077),
    [anon_sym_getsockname] = ACTIONS(3077),
    [anon_sym_getgrnam] = ACTIONS(3077),
    [anon_sym_getgrgid] = ACTIONS(3077),
    [anon_sym_hex] = ACTIONS(3077),
    [anon_sym_int] = ACTIONS(3077),
    [anon_sym_keys] = ACTIONS(3077),
    [anon_sym_lc] = ACTIONS(3077),
    [anon_sym_lcfirst] = ACTIONS(3077),
    [anon_sym_length] = ACTIONS(3077),
    [anon_sym_localtime] = ACTIONS(3077),
    [anon_sym_log] = ACTIONS(3077),
    [anon_sym_lock] = ACTIONS(3077),
    [anon_sym_lstat] = ACTIONS(3077),
    [anon_sym_oct] = ACTIONS(3077),
    [anon_sym_ord] = ACTIONS(3077),
    [anon_sym_prototype] = ACTIONS(3077),
    [anon_sym_pop] = ACTIONS(3077),
    [anon_sym_pos] = ACTIONS(3077),
    [anon_sym_quotemeta] = ACTIONS(3077),
    [anon_sym_reset] = ACTIONS(3077),
    [anon_sym_rand] = ACTIONS(3077),
    [anon_sym_rmdir] = ACTIONS(3077),
    [anon_sym_readdir] = ACTIONS(3077),
    [anon_sym_readline] = ACTIONS(3077),
    [anon_sym_readpipe] = ACTIONS(3077),
    [anon_sym_rewinddir] = ACTIONS(3077),
    [anon_sym_readlink] = ACTIONS(3077),
    [anon_sym_ref] = ACTIONS(3077),
    [anon_sym_scalar] = ACTIONS(3077),
    [anon_sym_shift] = ACTIONS(3077),
    [anon_sym_sin] = ACTIONS(3077),
    [anon_sym_sleep] = ACTIONS(3077),
    [anon_sym_sqrt] = ACTIONS(3077),
    [anon_sym_srand] = ACTIONS(3077),
    [anon_sym_stat] = ACTIONS(3077),
    [anon_sym_study] = ACTIONS(3077),
    [anon_sym_tell] = ACTIONS(3077),
    [anon_sym_telldir] = ACTIONS(3077),
    [anon_sym_tied] = ACTIONS(3077),
    [anon_sym_uc] = ACTIONS(3077),
    [anon_sym_ucfirst] = ACTIONS(3077),
    [anon_sym_untie] = ACTIONS(3077),
    [anon_sym_umask] = ACTIONS(3077),
    [anon_sym_values] = ACTIONS(3077),
    [anon_sym_write] = ACTIONS(3077),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3077),
    [anon_sym_qq] = ACTIONS(3077),
    [anon_sym_qw] = ACTIONS(3077),
    [anon_sym_qx] = ACTIONS(3077),
    [anon_sym_qr] = ACTIONS(3077),
    [anon_sym_m] = ACTIONS(3077),
    [anon_sym_s] = ACTIONS(3077),
    [anon_sym_tr] = ACTIONS(3077),
    [anon_sym_y] = ACTIONS(3077),
    [anon_sym_if] = ACTIONS(3077),
    [anon_sym_unless] = ACTIONS(3077),
    [anon_sym_while] = ACTIONS(3077),
    [anon_sym_until] = ACTIONS(3077),
    [aux_sym__bareword_token1] = ACTIONS(3077),
    [sym_number] = ACTIONS(3077),
    [anon_sym_true] = ACTIONS(3077),
    [anon_sym_false] = ACTIONS(3077),
    [sym__single_quote] = ACTIONS(3075),
    [sym__double_quote] = ACTIONS(3075),
    [sym__backtick_quote] = ACTIONS(3075),
    [sym__search_slash_quote] = ACTIONS(3075),
    [sym__open_readline_bracket] = ACTIONS(3075),
    [sym__open_fileglob_bracket] = ACTIONS(3075),
    [sym__PERLY_HEREDOC] = ACTIONS(3075),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3075),
    [sym__filetest] = ACTIONS(3075),
  },
  [STATE(1607)] = {
    [sym_heredoc_content] = STATE(1607),
    [sym__identifier] = ACTIONS(3097),
    [anon_sym_LBRACE] = ACTIONS(3097),
    [anon_sym_RBRACE] = ACTIONS(3097),
    [anon_sym_SEMI] = ACTIONS(3097),
    [anon_sym_package] = ACTIONS(3097),
    [anon_sym_class] = ACTIONS(3097),
    [anon_sym_role] = ACTIONS(3097),
    [anon_sym_BUILD] = ACTIONS(3097),
    [anon_sym_ADJUST] = ACTIONS(3097),
    [anon_sym_DOLLAR] = ACTIONS(3097),
    [anon_sym_AT] = ACTIONS(3097),
    [anon_sym_my] = ACTIONS(3097),
    [anon_sym_extended] = ACTIONS(3097),
    [anon_sym_async] = ACTIONS(3097),
    [anon_sym_sub] = ACTIONS(3097),
    [anon_sym_method] = ACTIONS(3097),
    [anon_sym_LPAREN] = ACTIONS(3097),
    [anon_sym_continue] = ACTIONS(3097),
    [anon_sym_state] = ACTIONS(3097),
    [anon_sym_our] = ACTIONS(3097),
    [anon_sym_try] = ACTIONS(3097),
    [anon_sym_defer] = ACTIONS(3097),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3097),
    [anon_sym_LBRACK] = ACTIONS(3097),
    [anon_sym_LT_LT] = ACTIONS(3097),
    [anon_sym_SLASH_SLASH] = ACTIONS(3097),
    [anon_sym_PLUS] = ACTIONS(3097),
    [anon_sym_DASH] = ACTIONS(3097),
    [anon_sym_TILDE] = ACTIONS(3097),
    [anon_sym_BANG] = ACTIONS(3097),
    [anon_sym_PLUS_PLUS] = ACTIONS(3097),
    [anon_sym_DASH_DASH] = ACTIONS(3097),
    [anon_sym_BSLASH] = ACTIONS(3095),
    [anon_sym_do] = ACTIONS(3097),
    [anon_sym_eval] = ACTIONS(3097),
    [anon_sym_await] = ACTIONS(3097),
    [anon_sym_field] = ACTIONS(3097),
    [anon_sym_local] = ACTIONS(3097),
    [anon_sym_dynamically] = ACTIONS(3097),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3097),
    [anon_sym_require] = ACTIONS(3097),
    [anon_sym_map] = ACTIONS(3097),
    [anon_sym_grep] = ACTIONS(3097),
    [anon_sym_sort] = ACTIONS(3097),
    [anon_sym_goto] = ACTIONS(3097),
    [anon_sym_return] = ACTIONS(3097),
    [anon_sym_undef] = ACTIONS(3097),
    [anon_sym_PERCENT2] = ACTIONS(3097),
    [anon_sym_AMP2] = ACTIONS(3097),
    [anon_sym_STAR2] = ACTIONS(3097),
    [anon_sym_use] = ACTIONS(3097),
    [anon_sym_no] = ACTIONS(3097),
    [anon_sym_for] = ACTIONS(3097),
    [anon_sym_foreach] = ACTIONS(3097),
    [anon_sym_last] = ACTIONS(3097),
    [anon_sym_next] = ACTIONS(3097),
    [anon_sym_redo] = ACTIONS(3097),
    [anon_sym_BEGIN] = ACTIONS(3097),
    [anon_sym_INIT] = ACTIONS(3097),
    [anon_sym_CHECK] = ACTIONS(3097),
    [anon_sym_UNITCHECK] = ACTIONS(3097),
    [anon_sym_END] = ACTIONS(3097),
    [anon_sym___FILE__] = ACTIONS(3097),
    [anon_sym___LINE__] = ACTIONS(3097),
    [anon_sym___PACKAGE__] = ACTIONS(3097),
    [anon_sym___SUB__] = ACTIONS(3097),
    [anon_sym_break] = ACTIONS(3097),
    [anon_sym_fork] = ACTIONS(3097),
    [anon_sym_getppid] = ACTIONS(3097),
    [anon_sym_time] = ACTIONS(3097),
    [anon_sym_times] = ACTIONS(3097),
    [anon_sym_wait] = ACTIONS(3097),
    [anon_sym_wantarray] = ACTIONS(3097),
    [anon_sym_abs] = ACTIONS(3097),
    [anon_sym_alarm] = ACTIONS(3097),
    [anon_sym_chop] = ACTIONS(3097),
    [anon_sym_chdir] = ACTIONS(3097),
    [anon_sym_close] = ACTIONS(3097),
    [anon_sym_closedir] = ACTIONS(3097),
    [anon_sym_caller] = ACTIONS(3097),
    [anon_sym_chomp] = ACTIONS(3097),
    [anon_sym_chr] = ACTIONS(3097),
    [anon_sym_cos] = ACTIONS(3097),
    [anon_sym_chroot] = ACTIONS(3097),
    [anon_sym_defined] = ACTIONS(3097),
    [anon_sym_delete] = ACTIONS(3097),
    [anon_sym_dbmclose] = ACTIONS(3097),
    [anon_sym_exists] = ACTIONS(3097),
    [anon_sym_exit] = ACTIONS(3097),
    [anon_sym_eof] = ACTIONS(3097),
    [anon_sym_exp] = ACTIONS(3097),
    [anon_sym_each] = ACTIONS(3097),
    [anon_sym_fc] = ACTIONS(3097),
    [anon_sym_fileno] = ACTIONS(3097),
    [anon_sym_gmtime] = ACTIONS(3097),
    [anon_sym_getc] = ACTIONS(3097),
    [anon_sym_getpgrp] = ACTIONS(3097),
    [anon_sym_getprotobyname] = ACTIONS(3097),
    [anon_sym_getpwname] = ACTIONS(3097),
    [anon_sym_getpwuid] = ACTIONS(3097),
    [anon_sym_getpeername] = ACTIONS(3097),
    [anon_sym_getnetbyname] = ACTIONS(3097),
    [anon_sym_getsockname] = ACTIONS(3097),
    [anon_sym_getgrnam] = ACTIONS(3097),
    [anon_sym_getgrgid] = ACTIONS(3097),
    [anon_sym_hex] = ACTIONS(3097),
    [anon_sym_int] = ACTIONS(3097),
    [anon_sym_keys] = ACTIONS(3097),
    [anon_sym_lc] = ACTIONS(3097),
    [anon_sym_lcfirst] = ACTIONS(3097),
    [anon_sym_length] = ACTIONS(3097),
    [anon_sym_localtime] = ACTIONS(3097),
    [anon_sym_log] = ACTIONS(3097),
    [anon_sym_lock] = ACTIONS(3097),
    [anon_sym_lstat] = ACTIONS(3097),
    [anon_sym_oct] = ACTIONS(3097),
    [anon_sym_ord] = ACTIONS(3097),
    [anon_sym_prototype] = ACTIONS(3097),
    [anon_sym_pop] = ACTIONS(3097),
    [anon_sym_pos] = ACTIONS(3097),
    [anon_sym_quotemeta] = ACTIONS(3097),
    [anon_sym_reset] = ACTIONS(3097),
    [anon_sym_rand] = ACTIONS(3097),
    [anon_sym_rmdir] = ACTIONS(3097),
    [anon_sym_readdir] = ACTIONS(3097),
    [anon_sym_readline] = ACTIONS(3097),
    [anon_sym_readpipe] = ACTIONS(3097),
    [anon_sym_rewinddir] = ACTIONS(3097),
    [anon_sym_readlink] = ACTIONS(3097),
    [anon_sym_ref] = ACTIONS(3097),
    [anon_sym_scalar] = ACTIONS(3097),
    [anon_sym_shift] = ACTIONS(3097),
    [anon_sym_sin] = ACTIONS(3097),
    [anon_sym_sleep] = ACTIONS(3097),
    [anon_sym_sqrt] = ACTIONS(3097),
    [anon_sym_srand] = ACTIONS(3097),
    [anon_sym_stat] = ACTIONS(3097),
    [anon_sym_study] = ACTIONS(3097),
    [anon_sym_tell] = ACTIONS(3097),
    [anon_sym_telldir] = ACTIONS(3097),
    [anon_sym_tied] = ACTIONS(3097),
    [anon_sym_uc] = ACTIONS(3097),
    [anon_sym_ucfirst] = ACTIONS(3097),
    [anon_sym_untie] = ACTIONS(3097),
    [anon_sym_umask] = ACTIONS(3097),
    [anon_sym_values] = ACTIONS(3097),
    [anon_sym_write] = ACTIONS(3097),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3097),
    [anon_sym_qq] = ACTIONS(3097),
    [anon_sym_qw] = ACTIONS(3097),
    [anon_sym_qx] = ACTIONS(3097),
    [anon_sym_qr] = ACTIONS(3097),
    [anon_sym_m] = ACTIONS(3097),
    [anon_sym_s] = ACTIONS(3097),
    [anon_sym_tr] = ACTIONS(3097),
    [anon_sym_y] = ACTIONS(3097),
    [anon_sym_if] = ACTIONS(3097),
    [anon_sym_unless] = ACTIONS(3097),
    [anon_sym_while] = ACTIONS(3097),
    [anon_sym_until] = ACTIONS(3097),
    [aux_sym__bareword_token1] = ACTIONS(3097),
    [sym_number] = ACTIONS(3097),
    [anon_sym_true] = ACTIONS(3097),
    [anon_sym_false] = ACTIONS(3097),
    [sym__single_quote] = ACTIONS(3095),
    [sym__double_quote] = ACTIONS(3095),
    [sym__backtick_quote] = ACTIONS(3095),
    [sym__search_slash_quote] = ACTIONS(3095),
    [sym__open_readline_bracket] = ACTIONS(3095),
    [sym__open_fileglob_bracket] = ACTIONS(3095),
    [sym__PERLY_HEREDOC] = ACTIONS(3095),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3095),
    [sym__filetest] = ACTIONS(3095),
  },
  [STATE(1608)] = {
    [sym_heredoc_content] = STATE(1608),
    [sym__identifier] = ACTIONS(3109),
    [anon_sym_LBRACE] = ACTIONS(3109),
    [anon_sym_RBRACE] = ACTIONS(3109),
    [anon_sym_SEMI] = ACTIONS(3109),
    [anon_sym_package] = ACTIONS(3109),
    [anon_sym_class] = ACTIONS(3109),
    [anon_sym_role] = ACTIONS(3109),
    [anon_sym_BUILD] = ACTIONS(3109),
    [anon_sym_ADJUST] = ACTIONS(3109),
    [anon_sym_DOLLAR] = ACTIONS(3109),
    [anon_sym_AT] = ACTIONS(3109),
    [anon_sym_my] = ACTIONS(3109),
    [anon_sym_extended] = ACTIONS(3109),
    [anon_sym_async] = ACTIONS(3109),
    [anon_sym_sub] = ACTIONS(3109),
    [anon_sym_method] = ACTIONS(3109),
    [anon_sym_LPAREN] = ACTIONS(3109),
    [anon_sym_continue] = ACTIONS(3109),
    [anon_sym_state] = ACTIONS(3109),
    [anon_sym_our] = ACTIONS(3109),
    [anon_sym_try] = ACTIONS(3109),
    [anon_sym_defer] = ACTIONS(3109),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3109),
    [anon_sym_LBRACK] = ACTIONS(3109),
    [anon_sym_LT_LT] = ACTIONS(3109),
    [anon_sym_SLASH_SLASH] = ACTIONS(3109),
    [anon_sym_PLUS] = ACTIONS(3109),
    [anon_sym_DASH] = ACTIONS(3109),
    [anon_sym_TILDE] = ACTIONS(3109),
    [anon_sym_BANG] = ACTIONS(3109),
    [anon_sym_PLUS_PLUS] = ACTIONS(3109),
    [anon_sym_DASH_DASH] = ACTIONS(3109),
    [anon_sym_BSLASH] = ACTIONS(3107),
    [anon_sym_do] = ACTIONS(3109),
    [anon_sym_eval] = ACTIONS(3109),
    [anon_sym_await] = ACTIONS(3109),
    [anon_sym_field] = ACTIONS(3109),
    [anon_sym_local] = ACTIONS(3109),
    [anon_sym_dynamically] = ACTIONS(3109),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3109),
    [anon_sym_require] = ACTIONS(3109),
    [anon_sym_map] = ACTIONS(3109),
    [anon_sym_grep] = ACTIONS(3109),
    [anon_sym_sort] = ACTIONS(3109),
    [anon_sym_goto] = ACTIONS(3109),
    [anon_sym_return] = ACTIONS(3109),
    [anon_sym_undef] = ACTIONS(3109),
    [anon_sym_PERCENT2] = ACTIONS(3109),
    [anon_sym_AMP2] = ACTIONS(3109),
    [anon_sym_STAR2] = ACTIONS(3109),
    [anon_sym_use] = ACTIONS(3109),
    [anon_sym_no] = ACTIONS(3109),
    [anon_sym_for] = ACTIONS(3109),
    [anon_sym_foreach] = ACTIONS(3109),
    [anon_sym_last] = ACTIONS(3109),
    [anon_sym_next] = ACTIONS(3109),
    [anon_sym_redo] = ACTIONS(3109),
    [anon_sym_BEGIN] = ACTIONS(3109),
    [anon_sym_INIT] = ACTIONS(3109),
    [anon_sym_CHECK] = ACTIONS(3109),
    [anon_sym_UNITCHECK] = ACTIONS(3109),
    [anon_sym_END] = ACTIONS(3109),
    [anon_sym___FILE__] = ACTIONS(3109),
    [anon_sym___LINE__] = ACTIONS(3109),
    [anon_sym___PACKAGE__] = ACTIONS(3109),
    [anon_sym___SUB__] = ACTIONS(3109),
    [anon_sym_break] = ACTIONS(3109),
    [anon_sym_fork] = ACTIONS(3109),
    [anon_sym_getppid] = ACTIONS(3109),
    [anon_sym_time] = ACTIONS(3109),
    [anon_sym_times] = ACTIONS(3109),
    [anon_sym_wait] = ACTIONS(3109),
    [anon_sym_wantarray] = ACTIONS(3109),
    [anon_sym_abs] = ACTIONS(3109),
    [anon_sym_alarm] = ACTIONS(3109),
    [anon_sym_chop] = ACTIONS(3109),
    [anon_sym_chdir] = ACTIONS(3109),
    [anon_sym_close] = ACTIONS(3109),
    [anon_sym_closedir] = ACTIONS(3109),
    [anon_sym_caller] = ACTIONS(3109),
    [anon_sym_chomp] = ACTIONS(3109),
    [anon_sym_chr] = ACTIONS(3109),
    [anon_sym_cos] = ACTIONS(3109),
    [anon_sym_chroot] = ACTIONS(3109),
    [anon_sym_defined] = ACTIONS(3109),
    [anon_sym_delete] = ACTIONS(3109),
    [anon_sym_dbmclose] = ACTIONS(3109),
    [anon_sym_exists] = ACTIONS(3109),
    [anon_sym_exit] = ACTIONS(3109),
    [anon_sym_eof] = ACTIONS(3109),
    [anon_sym_exp] = ACTIONS(3109),
    [anon_sym_each] = ACTIONS(3109),
    [anon_sym_fc] = ACTIONS(3109),
    [anon_sym_fileno] = ACTIONS(3109),
    [anon_sym_gmtime] = ACTIONS(3109),
    [anon_sym_getc] = ACTIONS(3109),
    [anon_sym_getpgrp] = ACTIONS(3109),
    [anon_sym_getprotobyname] = ACTIONS(3109),
    [anon_sym_getpwname] = ACTIONS(3109),
    [anon_sym_getpwuid] = ACTIONS(3109),
    [anon_sym_getpeername] = ACTIONS(3109),
    [anon_sym_getnetbyname] = ACTIONS(3109),
    [anon_sym_getsockname] = ACTIONS(3109),
    [anon_sym_getgrnam] = ACTIONS(3109),
    [anon_sym_getgrgid] = ACTIONS(3109),
    [anon_sym_hex] = ACTIONS(3109),
    [anon_sym_int] = ACTIONS(3109),
    [anon_sym_keys] = ACTIONS(3109),
    [anon_sym_lc] = ACTIONS(3109),
    [anon_sym_lcfirst] = ACTIONS(3109),
    [anon_sym_length] = ACTIONS(3109),
    [anon_sym_localtime] = ACTIONS(3109),
    [anon_sym_log] = ACTIONS(3109),
    [anon_sym_lock] = ACTIONS(3109),
    [anon_sym_lstat] = ACTIONS(3109),
    [anon_sym_oct] = ACTIONS(3109),
    [anon_sym_ord] = ACTIONS(3109),
    [anon_sym_prototype] = ACTIONS(3109),
    [anon_sym_pop] = ACTIONS(3109),
    [anon_sym_pos] = ACTIONS(3109),
    [anon_sym_quotemeta] = ACTIONS(3109),
    [anon_sym_reset] = ACTIONS(3109),
    [anon_sym_rand] = ACTIONS(3109),
    [anon_sym_rmdir] = ACTIONS(3109),
    [anon_sym_readdir] = ACTIONS(3109),
    [anon_sym_readline] = ACTIONS(3109),
    [anon_sym_readpipe] = ACTIONS(3109),
    [anon_sym_rewinddir] = ACTIONS(3109),
    [anon_sym_readlink] = ACTIONS(3109),
    [anon_sym_ref] = ACTIONS(3109),
    [anon_sym_scalar] = ACTIONS(3109),
    [anon_sym_shift] = ACTIONS(3109),
    [anon_sym_sin] = ACTIONS(3109),
    [anon_sym_sleep] = ACTIONS(3109),
    [anon_sym_sqrt] = ACTIONS(3109),
    [anon_sym_srand] = ACTIONS(3109),
    [anon_sym_stat] = ACTIONS(3109),
    [anon_sym_study] = ACTIONS(3109),
    [anon_sym_tell] = ACTIONS(3109),
    [anon_sym_telldir] = ACTIONS(3109),
    [anon_sym_tied] = ACTIONS(3109),
    [anon_sym_uc] = ACTIONS(3109),
    [anon_sym_ucfirst] = ACTIONS(3109),
    [anon_sym_untie] = ACTIONS(3109),
    [anon_sym_umask] = ACTIONS(3109),
    [anon_sym_values] = ACTIONS(3109),
    [anon_sym_write] = ACTIONS(3109),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3109),
    [anon_sym_qq] = ACTIONS(3109),
    [anon_sym_qw] = ACTIONS(3109),
    [anon_sym_qx] = ACTIONS(3109),
    [anon_sym_qr] = ACTIONS(3109),
    [anon_sym_m] = ACTIONS(3109),
    [anon_sym_s] = ACTIONS(3109),
    [anon_sym_tr] = ACTIONS(3109),
    [anon_sym_y] = ACTIONS(3109),
    [anon_sym_if] = ACTIONS(3109),
    [anon_sym_unless] = ACTIONS(3109),
    [anon_sym_while] = ACTIONS(3109),
    [anon_sym_until] = ACTIONS(3109),
    [aux_sym__bareword_token1] = ACTIONS(3109),
    [sym_number] = ACTIONS(3109),
    [anon_sym_true] = ACTIONS(3109),
    [anon_sym_false] = ACTIONS(3109),
    [sym__single_quote] = ACTIONS(3107),
    [sym__double_quote] = ACTIONS(3107),
    [sym__backtick_quote] = ACTIONS(3107),
    [sym__search_slash_quote] = ACTIONS(3107),
    [sym__open_readline_bracket] = ACTIONS(3107),
    [sym__open_fileglob_bracket] = ACTIONS(3107),
    [sym__PERLY_HEREDOC] = ACTIONS(3107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3107),
    [sym__filetest] = ACTIONS(3107),
  },
  [STATE(1609)] = {
    [sym_heredoc_content] = STATE(1609),
    [sym__identifier] = ACTIONS(3117),
    [anon_sym_LBRACE] = ACTIONS(3117),
    [anon_sym_RBRACE] = ACTIONS(3117),
    [anon_sym_SEMI] = ACTIONS(3117),
    [anon_sym_package] = ACTIONS(3117),
    [anon_sym_class] = ACTIONS(3117),
    [anon_sym_role] = ACTIONS(3117),
    [anon_sym_BUILD] = ACTIONS(3117),
    [anon_sym_ADJUST] = ACTIONS(3117),
    [anon_sym_DOLLAR] = ACTIONS(3117),
    [anon_sym_AT] = ACTIONS(3117),
    [anon_sym_my] = ACTIONS(3117),
    [anon_sym_extended] = ACTIONS(3117),
    [anon_sym_async] = ACTIONS(3117),
    [anon_sym_sub] = ACTIONS(3117),
    [anon_sym_method] = ACTIONS(3117),
    [anon_sym_LPAREN] = ACTIONS(3117),
    [anon_sym_continue] = ACTIONS(3117),
    [anon_sym_state] = ACTIONS(3117),
    [anon_sym_our] = ACTIONS(3117),
    [anon_sym_try] = ACTIONS(3117),
    [anon_sym_defer] = ACTIONS(3117),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3117),
    [anon_sym_LBRACK] = ACTIONS(3117),
    [anon_sym_LT_LT] = ACTIONS(3117),
    [anon_sym_SLASH_SLASH] = ACTIONS(3117),
    [anon_sym_PLUS] = ACTIONS(3117),
    [anon_sym_DASH] = ACTIONS(3117),
    [anon_sym_TILDE] = ACTIONS(3117),
    [anon_sym_BANG] = ACTIONS(3117),
    [anon_sym_PLUS_PLUS] = ACTIONS(3117),
    [anon_sym_DASH_DASH] = ACTIONS(3117),
    [anon_sym_BSLASH] = ACTIONS(3115),
    [anon_sym_do] = ACTIONS(3117),
    [anon_sym_eval] = ACTIONS(3117),
    [anon_sym_await] = ACTIONS(3117),
    [anon_sym_field] = ACTIONS(3117),
    [anon_sym_local] = ACTIONS(3117),
    [anon_sym_dynamically] = ACTIONS(3117),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3117),
    [anon_sym_require] = ACTIONS(3117),
    [anon_sym_map] = ACTIONS(3117),
    [anon_sym_grep] = ACTIONS(3117),
    [anon_sym_sort] = ACTIONS(3117),
    [anon_sym_goto] = ACTIONS(3117),
    [anon_sym_return] = ACTIONS(3117),
    [anon_sym_undef] = ACTIONS(3117),
    [anon_sym_PERCENT2] = ACTIONS(3117),
    [anon_sym_AMP2] = ACTIONS(3117),
    [anon_sym_STAR2] = ACTIONS(3117),
    [anon_sym_use] = ACTIONS(3117),
    [anon_sym_no] = ACTIONS(3117),
    [anon_sym_for] = ACTIONS(3117),
    [anon_sym_foreach] = ACTIONS(3117),
    [anon_sym_last] = ACTIONS(3117),
    [anon_sym_next] = ACTIONS(3117),
    [anon_sym_redo] = ACTIONS(3117),
    [anon_sym_BEGIN] = ACTIONS(3117),
    [anon_sym_INIT] = ACTIONS(3117),
    [anon_sym_CHECK] = ACTIONS(3117),
    [anon_sym_UNITCHECK] = ACTIONS(3117),
    [anon_sym_END] = ACTIONS(3117),
    [anon_sym___FILE__] = ACTIONS(3117),
    [anon_sym___LINE__] = ACTIONS(3117),
    [anon_sym___PACKAGE__] = ACTIONS(3117),
    [anon_sym___SUB__] = ACTIONS(3117),
    [anon_sym_break] = ACTIONS(3117),
    [anon_sym_fork] = ACTIONS(3117),
    [anon_sym_getppid] = ACTIONS(3117),
    [anon_sym_time] = ACTIONS(3117),
    [anon_sym_times] = ACTIONS(3117),
    [anon_sym_wait] = ACTIONS(3117),
    [anon_sym_wantarray] = ACTIONS(3117),
    [anon_sym_abs] = ACTIONS(3117),
    [anon_sym_alarm] = ACTIONS(3117),
    [anon_sym_chop] = ACTIONS(3117),
    [anon_sym_chdir] = ACTIONS(3117),
    [anon_sym_close] = ACTIONS(3117),
    [anon_sym_closedir] = ACTIONS(3117),
    [anon_sym_caller] = ACTIONS(3117),
    [anon_sym_chomp] = ACTIONS(3117),
    [anon_sym_chr] = ACTIONS(3117),
    [anon_sym_cos] = ACTIONS(3117),
    [anon_sym_chroot] = ACTIONS(3117),
    [anon_sym_defined] = ACTIONS(3117),
    [anon_sym_delete] = ACTIONS(3117),
    [anon_sym_dbmclose] = ACTIONS(3117),
    [anon_sym_exists] = ACTIONS(3117),
    [anon_sym_exit] = ACTIONS(3117),
    [anon_sym_eof] = ACTIONS(3117),
    [anon_sym_exp] = ACTIONS(3117),
    [anon_sym_each] = ACTIONS(3117),
    [anon_sym_fc] = ACTIONS(3117),
    [anon_sym_fileno] = ACTIONS(3117),
    [anon_sym_gmtime] = ACTIONS(3117),
    [anon_sym_getc] = ACTIONS(3117),
    [anon_sym_getpgrp] = ACTIONS(3117),
    [anon_sym_getprotobyname] = ACTIONS(3117),
    [anon_sym_getpwname] = ACTIONS(3117),
    [anon_sym_getpwuid] = ACTIONS(3117),
    [anon_sym_getpeername] = ACTIONS(3117),
    [anon_sym_getnetbyname] = ACTIONS(3117),
    [anon_sym_getsockname] = ACTIONS(3117),
    [anon_sym_getgrnam] = ACTIONS(3117),
    [anon_sym_getgrgid] = ACTIONS(3117),
    [anon_sym_hex] = ACTIONS(3117),
    [anon_sym_int] = ACTIONS(3117),
    [anon_sym_keys] = ACTIONS(3117),
    [anon_sym_lc] = ACTIONS(3117),
    [anon_sym_lcfirst] = ACTIONS(3117),
    [anon_sym_length] = ACTIONS(3117),
    [anon_sym_localtime] = ACTIONS(3117),
    [anon_sym_log] = ACTIONS(3117),
    [anon_sym_lock] = ACTIONS(3117),
    [anon_sym_lstat] = ACTIONS(3117),
    [anon_sym_oct] = ACTIONS(3117),
    [anon_sym_ord] = ACTIONS(3117),
    [anon_sym_prototype] = ACTIONS(3117),
    [anon_sym_pop] = ACTIONS(3117),
    [anon_sym_pos] = ACTIONS(3117),
    [anon_sym_quotemeta] = ACTIONS(3117),
    [anon_sym_reset] = ACTIONS(3117),
    [anon_sym_rand] = ACTIONS(3117),
    [anon_sym_rmdir] = ACTIONS(3117),
    [anon_sym_readdir] = ACTIONS(3117),
    [anon_sym_readline] = ACTIONS(3117),
    [anon_sym_readpipe] = ACTIONS(3117),
    [anon_sym_rewinddir] = ACTIONS(3117),
    [anon_sym_readlink] = ACTIONS(3117),
    [anon_sym_ref] = ACTIONS(3117),
    [anon_sym_scalar] = ACTIONS(3117),
    [anon_sym_shift] = ACTIONS(3117),
    [anon_sym_sin] = ACTIONS(3117),
    [anon_sym_sleep] = ACTIONS(3117),
    [anon_sym_sqrt] = ACTIONS(3117),
    [anon_sym_srand] = ACTIONS(3117),
    [anon_sym_stat] = ACTIONS(3117),
    [anon_sym_study] = ACTIONS(3117),
    [anon_sym_tell] = ACTIONS(3117),
    [anon_sym_telldir] = ACTIONS(3117),
    [anon_sym_tied] = ACTIONS(3117),
    [anon_sym_uc] = ACTIONS(3117),
    [anon_sym_ucfirst] = ACTIONS(3117),
    [anon_sym_untie] = ACTIONS(3117),
    [anon_sym_umask] = ACTIONS(3117),
    [anon_sym_values] = ACTIONS(3117),
    [anon_sym_write] = ACTIONS(3117),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3117),
    [anon_sym_qq] = ACTIONS(3117),
    [anon_sym_qw] = ACTIONS(3117),
    [anon_sym_qx] = ACTIONS(3117),
    [anon_sym_qr] = ACTIONS(3117),
    [anon_sym_m] = ACTIONS(3117),
    [anon_sym_s] = ACTIONS(3117),
    [anon_sym_tr] = ACTIONS(3117),
    [anon_sym_y] = ACTIONS(3117),
    [anon_sym_if] = ACTIONS(3117),
    [anon_sym_unless] = ACTIONS(3117),
    [anon_sym_while] = ACTIONS(3117),
    [anon_sym_until] = ACTIONS(3117),
    [aux_sym__bareword_token1] = ACTIONS(3117),
    [sym_number] = ACTIONS(3117),
    [anon_sym_true] = ACTIONS(3117),
    [anon_sym_false] = ACTIONS(3117),
    [sym__single_quote] = ACTIONS(3115),
    [sym__double_quote] = ACTIONS(3115),
    [sym__backtick_quote] = ACTIONS(3115),
    [sym__search_slash_quote] = ACTIONS(3115),
    [sym__open_readline_bracket] = ACTIONS(3115),
    [sym__open_fileglob_bracket] = ACTIONS(3115),
    [sym__PERLY_HEREDOC] = ACTIONS(3115),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3115),
    [sym__filetest] = ACTIONS(3115),
  },
  [STATE(1610)] = {
    [sym_heredoc_content] = STATE(1610),
    [sym__identifier] = ACTIONS(3133),
    [anon_sym_LBRACE] = ACTIONS(3133),
    [anon_sym_RBRACE] = ACTIONS(3133),
    [anon_sym_SEMI] = ACTIONS(3133),
    [anon_sym_package] = ACTIONS(3133),
    [anon_sym_class] = ACTIONS(3133),
    [anon_sym_role] = ACTIONS(3133),
    [anon_sym_BUILD] = ACTIONS(3133),
    [anon_sym_ADJUST] = ACTIONS(3133),
    [anon_sym_DOLLAR] = ACTIONS(3133),
    [anon_sym_AT] = ACTIONS(3133),
    [anon_sym_my] = ACTIONS(3133),
    [anon_sym_extended] = ACTIONS(3133),
    [anon_sym_async] = ACTIONS(3133),
    [anon_sym_sub] = ACTIONS(3133),
    [anon_sym_method] = ACTIONS(3133),
    [anon_sym_LPAREN] = ACTIONS(3133),
    [anon_sym_continue] = ACTIONS(3133),
    [anon_sym_state] = ACTIONS(3133),
    [anon_sym_our] = ACTIONS(3133),
    [anon_sym_try] = ACTIONS(3133),
    [anon_sym_defer] = ACTIONS(3133),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3133),
    [anon_sym_LBRACK] = ACTIONS(3133),
    [anon_sym_LT_LT] = ACTIONS(3133),
    [anon_sym_SLASH_SLASH] = ACTIONS(3133),
    [anon_sym_PLUS] = ACTIONS(3133),
    [anon_sym_DASH] = ACTIONS(3133),
    [anon_sym_TILDE] = ACTIONS(3133),
    [anon_sym_BANG] = ACTIONS(3133),
    [anon_sym_PLUS_PLUS] = ACTIONS(3133),
    [anon_sym_DASH_DASH] = ACTIONS(3133),
    [anon_sym_BSLASH] = ACTIONS(3131),
    [anon_sym_do] = ACTIONS(3133),
    [anon_sym_eval] = ACTIONS(3133),
    [anon_sym_await] = ACTIONS(3133),
    [anon_sym_field] = ACTIONS(3133),
    [anon_sym_local] = ACTIONS(3133),
    [anon_sym_dynamically] = ACTIONS(3133),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3133),
    [anon_sym_require] = ACTIONS(3133),
    [anon_sym_map] = ACTIONS(3133),
    [anon_sym_grep] = ACTIONS(3133),
    [anon_sym_sort] = ACTIONS(3133),
    [anon_sym_goto] = ACTIONS(3133),
    [anon_sym_return] = ACTIONS(3133),
    [anon_sym_undef] = ACTIONS(3133),
    [anon_sym_PERCENT2] = ACTIONS(3133),
    [anon_sym_AMP2] = ACTIONS(3133),
    [anon_sym_STAR2] = ACTIONS(3133),
    [anon_sym_use] = ACTIONS(3133),
    [anon_sym_no] = ACTIONS(3133),
    [anon_sym_for] = ACTIONS(3133),
    [anon_sym_foreach] = ACTIONS(3133),
    [anon_sym_last] = ACTIONS(3133),
    [anon_sym_next] = ACTIONS(3133),
    [anon_sym_redo] = ACTIONS(3133),
    [anon_sym_BEGIN] = ACTIONS(3133),
    [anon_sym_INIT] = ACTIONS(3133),
    [anon_sym_CHECK] = ACTIONS(3133),
    [anon_sym_UNITCHECK] = ACTIONS(3133),
    [anon_sym_END] = ACTIONS(3133),
    [anon_sym___FILE__] = ACTIONS(3133),
    [anon_sym___LINE__] = ACTIONS(3133),
    [anon_sym___PACKAGE__] = ACTIONS(3133),
    [anon_sym___SUB__] = ACTIONS(3133),
    [anon_sym_break] = ACTIONS(3133),
    [anon_sym_fork] = ACTIONS(3133),
    [anon_sym_getppid] = ACTIONS(3133),
    [anon_sym_time] = ACTIONS(3133),
    [anon_sym_times] = ACTIONS(3133),
    [anon_sym_wait] = ACTIONS(3133),
    [anon_sym_wantarray] = ACTIONS(3133),
    [anon_sym_abs] = ACTIONS(3133),
    [anon_sym_alarm] = ACTIONS(3133),
    [anon_sym_chop] = ACTIONS(3133),
    [anon_sym_chdir] = ACTIONS(3133),
    [anon_sym_close] = ACTIONS(3133),
    [anon_sym_closedir] = ACTIONS(3133),
    [anon_sym_caller] = ACTIONS(3133),
    [anon_sym_chomp] = ACTIONS(3133),
    [anon_sym_chr] = ACTIONS(3133),
    [anon_sym_cos] = ACTIONS(3133),
    [anon_sym_chroot] = ACTIONS(3133),
    [anon_sym_defined] = ACTIONS(3133),
    [anon_sym_delete] = ACTIONS(3133),
    [anon_sym_dbmclose] = ACTIONS(3133),
    [anon_sym_exists] = ACTIONS(3133),
    [anon_sym_exit] = ACTIONS(3133),
    [anon_sym_eof] = ACTIONS(3133),
    [anon_sym_exp] = ACTIONS(3133),
    [anon_sym_each] = ACTIONS(3133),
    [anon_sym_fc] = ACTIONS(3133),
    [anon_sym_fileno] = ACTIONS(3133),
    [anon_sym_gmtime] = ACTIONS(3133),
    [anon_sym_getc] = ACTIONS(3133),
    [anon_sym_getpgrp] = ACTIONS(3133),
    [anon_sym_getprotobyname] = ACTIONS(3133),
    [anon_sym_getpwname] = ACTIONS(3133),
    [anon_sym_getpwuid] = ACTIONS(3133),
    [anon_sym_getpeername] = ACTIONS(3133),
    [anon_sym_getnetbyname] = ACTIONS(3133),
    [anon_sym_getsockname] = ACTIONS(3133),
    [anon_sym_getgrnam] = ACTIONS(3133),
    [anon_sym_getgrgid] = ACTIONS(3133),
    [anon_sym_hex] = ACTIONS(3133),
    [anon_sym_int] = ACTIONS(3133),
    [anon_sym_keys] = ACTIONS(3133),
    [anon_sym_lc] = ACTIONS(3133),
    [anon_sym_lcfirst] = ACTIONS(3133),
    [anon_sym_length] = ACTIONS(3133),
    [anon_sym_localtime] = ACTIONS(3133),
    [anon_sym_log] = ACTIONS(3133),
    [anon_sym_lock] = ACTIONS(3133),
    [anon_sym_lstat] = ACTIONS(3133),
    [anon_sym_oct] = ACTIONS(3133),
    [anon_sym_ord] = ACTIONS(3133),
    [anon_sym_prototype] = ACTIONS(3133),
    [anon_sym_pop] = ACTIONS(3133),
    [anon_sym_pos] = ACTIONS(3133),
    [anon_sym_quotemeta] = ACTIONS(3133),
    [anon_sym_reset] = ACTIONS(3133),
    [anon_sym_rand] = ACTIONS(3133),
    [anon_sym_rmdir] = ACTIONS(3133),
    [anon_sym_readdir] = ACTIONS(3133),
    [anon_sym_readline] = ACTIONS(3133),
    [anon_sym_readpipe] = ACTIONS(3133),
    [anon_sym_rewinddir] = ACTIONS(3133),
    [anon_sym_readlink] = ACTIONS(3133),
    [anon_sym_ref] = ACTIONS(3133),
    [anon_sym_scalar] = ACTIONS(3133),
    [anon_sym_shift] = ACTIONS(3133),
    [anon_sym_sin] = ACTIONS(3133),
    [anon_sym_sleep] = ACTIONS(3133),
    [anon_sym_sqrt] = ACTIONS(3133),
    [anon_sym_srand] = ACTIONS(3133),
    [anon_sym_stat] = ACTIONS(3133),
    [anon_sym_study] = ACTIONS(3133),
    [anon_sym_tell] = ACTIONS(3133),
    [anon_sym_telldir] = ACTIONS(3133),
    [anon_sym_tied] = ACTIONS(3133),
    [anon_sym_uc] = ACTIONS(3133),
    [anon_sym_ucfirst] = ACTIONS(3133),
    [anon_sym_untie] = ACTIONS(3133),
    [anon_sym_umask] = ACTIONS(3133),
    [anon_sym_values] = ACTIONS(3133),
    [anon_sym_write] = ACTIONS(3133),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3133),
    [anon_sym_qq] = ACTIONS(3133),
    [anon_sym_qw] = ACTIONS(3133),
    [anon_sym_qx] = ACTIONS(3133),
    [anon_sym_qr] = ACTIONS(3133),
    [anon_sym_m] = ACTIONS(3133),
    [anon_sym_s] = ACTIONS(3133),
    [anon_sym_tr] = ACTIONS(3133),
    [anon_sym_y] = ACTIONS(3133),
    [anon_sym_if] = ACTIONS(3133),
    [anon_sym_unless] = ACTIONS(3133),
    [anon_sym_while] = ACTIONS(3133),
    [anon_sym_until] = ACTIONS(3133),
    [aux_sym__bareword_token1] = ACTIONS(3133),
    [sym_number] = ACTIONS(3133),
    [anon_sym_true] = ACTIONS(3133),
    [anon_sym_false] = ACTIONS(3133),
    [sym__single_quote] = ACTIONS(3131),
    [sym__double_quote] = ACTIONS(3131),
    [sym__backtick_quote] = ACTIONS(3131),
    [sym__search_slash_quote] = ACTIONS(3131),
    [sym__open_readline_bracket] = ACTIONS(3131),
    [sym__open_fileglob_bracket] = ACTIONS(3131),
    [sym__PERLY_HEREDOC] = ACTIONS(3131),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3131),
    [sym__filetest] = ACTIONS(3131),
  },
  [STATE(1611)] = {
    [sym_heredoc_content] = STATE(1611),
    [sym__identifier] = ACTIONS(3137),
    [anon_sym_LBRACE] = ACTIONS(3137),
    [anon_sym_RBRACE] = ACTIONS(3137),
    [anon_sym_SEMI] = ACTIONS(3137),
    [anon_sym_package] = ACTIONS(3137),
    [anon_sym_class] = ACTIONS(3137),
    [anon_sym_role] = ACTIONS(3137),
    [anon_sym_BUILD] = ACTIONS(3137),
    [anon_sym_ADJUST] = ACTIONS(3137),
    [anon_sym_DOLLAR] = ACTIONS(3137),
    [anon_sym_AT] = ACTIONS(3137),
    [anon_sym_my] = ACTIONS(3137),
    [anon_sym_extended] = ACTIONS(3137),
    [anon_sym_async] = ACTIONS(3137),
    [anon_sym_sub] = ACTIONS(3137),
    [anon_sym_method] = ACTIONS(3137),
    [anon_sym_LPAREN] = ACTIONS(3137),
    [anon_sym_continue] = ACTIONS(3137),
    [anon_sym_state] = ACTIONS(3137),
    [anon_sym_our] = ACTIONS(3137),
    [anon_sym_try] = ACTIONS(3137),
    [anon_sym_defer] = ACTIONS(3137),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3137),
    [anon_sym_LBRACK] = ACTIONS(3137),
    [anon_sym_LT_LT] = ACTIONS(3137),
    [anon_sym_SLASH_SLASH] = ACTIONS(3137),
    [anon_sym_PLUS] = ACTIONS(3137),
    [anon_sym_DASH] = ACTIONS(3137),
    [anon_sym_TILDE] = ACTIONS(3137),
    [anon_sym_BANG] = ACTIONS(3137),
    [anon_sym_PLUS_PLUS] = ACTIONS(3137),
    [anon_sym_DASH_DASH] = ACTIONS(3137),
    [anon_sym_BSLASH] = ACTIONS(3135),
    [anon_sym_do] = ACTIONS(3137),
    [anon_sym_eval] = ACTIONS(3137),
    [anon_sym_await] = ACTIONS(3137),
    [anon_sym_field] = ACTIONS(3137),
    [anon_sym_local] = ACTIONS(3137),
    [anon_sym_dynamically] = ACTIONS(3137),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3137),
    [anon_sym_require] = ACTIONS(3137),
    [anon_sym_map] = ACTIONS(3137),
    [anon_sym_grep] = ACTIONS(3137),
    [anon_sym_sort] = ACTIONS(3137),
    [anon_sym_goto] = ACTIONS(3137),
    [anon_sym_return] = ACTIONS(3137),
    [anon_sym_undef] = ACTIONS(3137),
    [anon_sym_PERCENT2] = ACTIONS(3137),
    [anon_sym_AMP2] = ACTIONS(3137),
    [anon_sym_STAR2] = ACTIONS(3137),
    [anon_sym_use] = ACTIONS(3137),
    [anon_sym_no] = ACTIONS(3137),
    [anon_sym_for] = ACTIONS(3137),
    [anon_sym_foreach] = ACTIONS(3137),
    [anon_sym_last] = ACTIONS(3137),
    [anon_sym_next] = ACTIONS(3137),
    [anon_sym_redo] = ACTIONS(3137),
    [anon_sym_BEGIN] = ACTIONS(3137),
    [anon_sym_INIT] = ACTIONS(3137),
    [anon_sym_CHECK] = ACTIONS(3137),
    [anon_sym_UNITCHECK] = ACTIONS(3137),
    [anon_sym_END] = ACTIONS(3137),
    [anon_sym___FILE__] = ACTIONS(3137),
    [anon_sym___LINE__] = ACTIONS(3137),
    [anon_sym___PACKAGE__] = ACTIONS(3137),
    [anon_sym___SUB__] = ACTIONS(3137),
    [anon_sym_break] = ACTIONS(3137),
    [anon_sym_fork] = ACTIONS(3137),
    [anon_sym_getppid] = ACTIONS(3137),
    [anon_sym_time] = ACTIONS(3137),
    [anon_sym_times] = ACTIONS(3137),
    [anon_sym_wait] = ACTIONS(3137),
    [anon_sym_wantarray] = ACTIONS(3137),
    [anon_sym_abs] = ACTIONS(3137),
    [anon_sym_alarm] = ACTIONS(3137),
    [anon_sym_chop] = ACTIONS(3137),
    [anon_sym_chdir] = ACTIONS(3137),
    [anon_sym_close] = ACTIONS(3137),
    [anon_sym_closedir] = ACTIONS(3137),
    [anon_sym_caller] = ACTIONS(3137),
    [anon_sym_chomp] = ACTIONS(3137),
    [anon_sym_chr] = ACTIONS(3137),
    [anon_sym_cos] = ACTIONS(3137),
    [anon_sym_chroot] = ACTIONS(3137),
    [anon_sym_defined] = ACTIONS(3137),
    [anon_sym_delete] = ACTIONS(3137),
    [anon_sym_dbmclose] = ACTIONS(3137),
    [anon_sym_exists] = ACTIONS(3137),
    [anon_sym_exit] = ACTIONS(3137),
    [anon_sym_eof] = ACTIONS(3137),
    [anon_sym_exp] = ACTIONS(3137),
    [anon_sym_each] = ACTIONS(3137),
    [anon_sym_fc] = ACTIONS(3137),
    [anon_sym_fileno] = ACTIONS(3137),
    [anon_sym_gmtime] = ACTIONS(3137),
    [anon_sym_getc] = ACTIONS(3137),
    [anon_sym_getpgrp] = ACTIONS(3137),
    [anon_sym_getprotobyname] = ACTIONS(3137),
    [anon_sym_getpwname] = ACTIONS(3137),
    [anon_sym_getpwuid] = ACTIONS(3137),
    [anon_sym_getpeername] = ACTIONS(3137),
    [anon_sym_getnetbyname] = ACTIONS(3137),
    [anon_sym_getsockname] = ACTIONS(3137),
    [anon_sym_getgrnam] = ACTIONS(3137),
    [anon_sym_getgrgid] = ACTIONS(3137),
    [anon_sym_hex] = ACTIONS(3137),
    [anon_sym_int] = ACTIONS(3137),
    [anon_sym_keys] = ACTIONS(3137),
    [anon_sym_lc] = ACTIONS(3137),
    [anon_sym_lcfirst] = ACTIONS(3137),
    [anon_sym_length] = ACTIONS(3137),
    [anon_sym_localtime] = ACTIONS(3137),
    [anon_sym_log] = ACTIONS(3137),
    [anon_sym_lock] = ACTIONS(3137),
    [anon_sym_lstat] = ACTIONS(3137),
    [anon_sym_oct] = ACTIONS(3137),
    [anon_sym_ord] = ACTIONS(3137),
    [anon_sym_prototype] = ACTIONS(3137),
    [anon_sym_pop] = ACTIONS(3137),
    [anon_sym_pos] = ACTIONS(3137),
    [anon_sym_quotemeta] = ACTIONS(3137),
    [anon_sym_reset] = ACTIONS(3137),
    [anon_sym_rand] = ACTIONS(3137),
    [anon_sym_rmdir] = ACTIONS(3137),
    [anon_sym_readdir] = ACTIONS(3137),
    [anon_sym_readline] = ACTIONS(3137),
    [anon_sym_readpipe] = ACTIONS(3137),
    [anon_sym_rewinddir] = ACTIONS(3137),
    [anon_sym_readlink] = ACTIONS(3137),
    [anon_sym_ref] = ACTIONS(3137),
    [anon_sym_scalar] = ACTIONS(3137),
    [anon_sym_shift] = ACTIONS(3137),
    [anon_sym_sin] = ACTIONS(3137),
    [anon_sym_sleep] = ACTIONS(3137),
    [anon_sym_sqrt] = ACTIONS(3137),
    [anon_sym_srand] = ACTIONS(3137),
    [anon_sym_stat] = ACTIONS(3137),
    [anon_sym_study] = ACTIONS(3137),
    [anon_sym_tell] = ACTIONS(3137),
    [anon_sym_telldir] = ACTIONS(3137),
    [anon_sym_tied] = ACTIONS(3137),
    [anon_sym_uc] = ACTIONS(3137),
    [anon_sym_ucfirst] = ACTIONS(3137),
    [anon_sym_untie] = ACTIONS(3137),
    [anon_sym_umask] = ACTIONS(3137),
    [anon_sym_values] = ACTIONS(3137),
    [anon_sym_write] = ACTIONS(3137),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3137),
    [anon_sym_qq] = ACTIONS(3137),
    [anon_sym_qw] = ACTIONS(3137),
    [anon_sym_qx] = ACTIONS(3137),
    [anon_sym_qr] = ACTIONS(3137),
    [anon_sym_m] = ACTIONS(3137),
    [anon_sym_s] = ACTIONS(3137),
    [anon_sym_tr] = ACTIONS(3137),
    [anon_sym_y] = ACTIONS(3137),
    [anon_sym_if] = ACTIONS(3137),
    [anon_sym_unless] = ACTIONS(3137),
    [anon_sym_while] = ACTIONS(3137),
    [anon_sym_until] = ACTIONS(3137),
    [aux_sym__bareword_token1] = ACTIONS(3137),
    [sym_number] = ACTIONS(3137),
    [anon_sym_true] = ACTIONS(3137),
    [anon_sym_false] = ACTIONS(3137),
    [sym__single_quote] = ACTIONS(3135),
    [sym__double_quote] = ACTIONS(3135),
    [sym__backtick_quote] = ACTIONS(3135),
    [sym__search_slash_quote] = ACTIONS(3135),
    [sym__open_readline_bracket] = ACTIONS(3135),
    [sym__open_fileglob_bracket] = ACTIONS(3135),
    [sym__PERLY_HEREDOC] = ACTIONS(3135),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3135),
    [sym__filetest] = ACTIONS(3135),
  },
  [STATE(1612)] = {
    [sym_heredoc_content] = STATE(1612),
    [sym__identifier] = ACTIONS(3149),
    [anon_sym_LBRACE] = ACTIONS(3149),
    [anon_sym_RBRACE] = ACTIONS(3149),
    [anon_sym_SEMI] = ACTIONS(3149),
    [anon_sym_package] = ACTIONS(3149),
    [anon_sym_class] = ACTIONS(3149),
    [anon_sym_role] = ACTIONS(3149),
    [anon_sym_BUILD] = ACTIONS(3149),
    [anon_sym_ADJUST] = ACTIONS(3149),
    [anon_sym_DOLLAR] = ACTIONS(3149),
    [anon_sym_AT] = ACTIONS(3149),
    [anon_sym_my] = ACTIONS(3149),
    [anon_sym_extended] = ACTIONS(3149),
    [anon_sym_async] = ACTIONS(3149),
    [anon_sym_sub] = ACTIONS(3149),
    [anon_sym_method] = ACTIONS(3149),
    [anon_sym_LPAREN] = ACTIONS(3149),
    [anon_sym_continue] = ACTIONS(3149),
    [anon_sym_state] = ACTIONS(3149),
    [anon_sym_our] = ACTIONS(3149),
    [anon_sym_try] = ACTIONS(3149),
    [anon_sym_defer] = ACTIONS(3149),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3149),
    [anon_sym_LBRACK] = ACTIONS(3149),
    [anon_sym_LT_LT] = ACTIONS(3149),
    [anon_sym_SLASH_SLASH] = ACTIONS(3149),
    [anon_sym_PLUS] = ACTIONS(3149),
    [anon_sym_DASH] = ACTIONS(3149),
    [anon_sym_TILDE] = ACTIONS(3149),
    [anon_sym_BANG] = ACTIONS(3149),
    [anon_sym_PLUS_PLUS] = ACTIONS(3149),
    [anon_sym_DASH_DASH] = ACTIONS(3149),
    [anon_sym_BSLASH] = ACTIONS(3147),
    [anon_sym_do] = ACTIONS(3149),
    [anon_sym_eval] = ACTIONS(3149),
    [anon_sym_await] = ACTIONS(3149),
    [anon_sym_field] = ACTIONS(3149),
    [anon_sym_local] = ACTIONS(3149),
    [anon_sym_dynamically] = ACTIONS(3149),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3149),
    [anon_sym_require] = ACTIONS(3149),
    [anon_sym_map] = ACTIONS(3149),
    [anon_sym_grep] = ACTIONS(3149),
    [anon_sym_sort] = ACTIONS(3149),
    [anon_sym_goto] = ACTIONS(3149),
    [anon_sym_return] = ACTIONS(3149),
    [anon_sym_undef] = ACTIONS(3149),
    [anon_sym_PERCENT2] = ACTIONS(3149),
    [anon_sym_AMP2] = ACTIONS(3149),
    [anon_sym_STAR2] = ACTIONS(3149),
    [anon_sym_use] = ACTIONS(3149),
    [anon_sym_no] = ACTIONS(3149),
    [anon_sym_for] = ACTIONS(3149),
    [anon_sym_foreach] = ACTIONS(3149),
    [anon_sym_last] = ACTIONS(3149),
    [anon_sym_next] = ACTIONS(3149),
    [anon_sym_redo] = ACTIONS(3149),
    [anon_sym_BEGIN] = ACTIONS(3149),
    [anon_sym_INIT] = ACTIONS(3149),
    [anon_sym_CHECK] = ACTIONS(3149),
    [anon_sym_UNITCHECK] = ACTIONS(3149),
    [anon_sym_END] = ACTIONS(3149),
    [anon_sym___FILE__] = ACTIONS(3149),
    [anon_sym___LINE__] = ACTIONS(3149),
    [anon_sym___PACKAGE__] = ACTIONS(3149),
    [anon_sym___SUB__] = ACTIONS(3149),
    [anon_sym_break] = ACTIONS(3149),
    [anon_sym_fork] = ACTIONS(3149),
    [anon_sym_getppid] = ACTIONS(3149),
    [anon_sym_time] = ACTIONS(3149),
    [anon_sym_times] = ACTIONS(3149),
    [anon_sym_wait] = ACTIONS(3149),
    [anon_sym_wantarray] = ACTIONS(3149),
    [anon_sym_abs] = ACTIONS(3149),
    [anon_sym_alarm] = ACTIONS(3149),
    [anon_sym_chop] = ACTIONS(3149),
    [anon_sym_chdir] = ACTIONS(3149),
    [anon_sym_close] = ACTIONS(3149),
    [anon_sym_closedir] = ACTIONS(3149),
    [anon_sym_caller] = ACTIONS(3149),
    [anon_sym_chomp] = ACTIONS(3149),
    [anon_sym_chr] = ACTIONS(3149),
    [anon_sym_cos] = ACTIONS(3149),
    [anon_sym_chroot] = ACTIONS(3149),
    [anon_sym_defined] = ACTIONS(3149),
    [anon_sym_delete] = ACTIONS(3149),
    [anon_sym_dbmclose] = ACTIONS(3149),
    [anon_sym_exists] = ACTIONS(3149),
    [anon_sym_exit] = ACTIONS(3149),
    [anon_sym_eof] = ACTIONS(3149),
    [anon_sym_exp] = ACTIONS(3149),
    [anon_sym_each] = ACTIONS(3149),
    [anon_sym_fc] = ACTIONS(3149),
    [anon_sym_fileno] = ACTIONS(3149),
    [anon_sym_gmtime] = ACTIONS(3149),
    [anon_sym_getc] = ACTIONS(3149),
    [anon_sym_getpgrp] = ACTIONS(3149),
    [anon_sym_getprotobyname] = ACTIONS(3149),
    [anon_sym_getpwname] = ACTIONS(3149),
    [anon_sym_getpwuid] = ACTIONS(3149),
    [anon_sym_getpeername] = ACTIONS(3149),
    [anon_sym_getnetbyname] = ACTIONS(3149),
    [anon_sym_getsockname] = ACTIONS(3149),
    [anon_sym_getgrnam] = ACTIONS(3149),
    [anon_sym_getgrgid] = ACTIONS(3149),
    [anon_sym_hex] = ACTIONS(3149),
    [anon_sym_int] = ACTIONS(3149),
    [anon_sym_keys] = ACTIONS(3149),
    [anon_sym_lc] = ACTIONS(3149),
    [anon_sym_lcfirst] = ACTIONS(3149),
    [anon_sym_length] = ACTIONS(3149),
    [anon_sym_localtime] = ACTIONS(3149),
    [anon_sym_log] = ACTIONS(3149),
    [anon_sym_lock] = ACTIONS(3149),
    [anon_sym_lstat] = ACTIONS(3149),
    [anon_sym_oct] = ACTIONS(3149),
    [anon_sym_ord] = ACTIONS(3149),
    [anon_sym_prototype] = ACTIONS(3149),
    [anon_sym_pop] = ACTIONS(3149),
    [anon_sym_pos] = ACTIONS(3149),
    [anon_sym_quotemeta] = ACTIONS(3149),
    [anon_sym_reset] = ACTIONS(3149),
    [anon_sym_rand] = ACTIONS(3149),
    [anon_sym_rmdir] = ACTIONS(3149),
    [anon_sym_readdir] = ACTIONS(3149),
    [anon_sym_readline] = ACTIONS(3149),
    [anon_sym_readpipe] = ACTIONS(3149),
    [anon_sym_rewinddir] = ACTIONS(3149),
    [anon_sym_readlink] = ACTIONS(3149),
    [anon_sym_ref] = ACTIONS(3149),
    [anon_sym_scalar] = ACTIONS(3149),
    [anon_sym_shift] = ACTIONS(3149),
    [anon_sym_sin] = ACTIONS(3149),
    [anon_sym_sleep] = ACTIONS(3149),
    [anon_sym_sqrt] = ACTIONS(3149),
    [anon_sym_srand] = ACTIONS(3149),
    [anon_sym_stat] = ACTIONS(3149),
    [anon_sym_study] = ACTIONS(3149),
    [anon_sym_tell] = ACTIONS(3149),
    [anon_sym_telldir] = ACTIONS(3149),
    [anon_sym_tied] = ACTIONS(3149),
    [anon_sym_uc] = ACTIONS(3149),
    [anon_sym_ucfirst] = ACTIONS(3149),
    [anon_sym_untie] = ACTIONS(3149),
    [anon_sym_umask] = ACTIONS(3149),
    [anon_sym_values] = ACTIONS(3149),
    [anon_sym_write] = ACTIONS(3149),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3149),
    [anon_sym_qq] = ACTIONS(3149),
    [anon_sym_qw] = ACTIONS(3149),
    [anon_sym_qx] = ACTIONS(3149),
    [anon_sym_qr] = ACTIONS(3149),
    [anon_sym_m] = ACTIONS(3149),
    [anon_sym_s] = ACTIONS(3149),
    [anon_sym_tr] = ACTIONS(3149),
    [anon_sym_y] = ACTIONS(3149),
    [anon_sym_if] = ACTIONS(3149),
    [anon_sym_unless] = ACTIONS(3149),
    [anon_sym_while] = ACTIONS(3149),
    [anon_sym_until] = ACTIONS(3149),
    [aux_sym__bareword_token1] = ACTIONS(3149),
    [sym_number] = ACTIONS(3149),
    [anon_sym_true] = ACTIONS(3149),
    [anon_sym_false] = ACTIONS(3149),
    [sym__single_quote] = ACTIONS(3147),
    [sym__double_quote] = ACTIONS(3147),
    [sym__backtick_quote] = ACTIONS(3147),
    [sym__search_slash_quote] = ACTIONS(3147),
    [sym__open_readline_bracket] = ACTIONS(3147),
    [sym__open_fileglob_bracket] = ACTIONS(3147),
    [sym__PERLY_HEREDOC] = ACTIONS(3147),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3147),
    [sym__filetest] = ACTIONS(3147),
  },
  [STATE(1613)] = {
    [sym_heredoc_content] = STATE(1613),
    [sym__identifier] = ACTIONS(3185),
    [anon_sym_LBRACE] = ACTIONS(3185),
    [anon_sym_RBRACE] = ACTIONS(3185),
    [anon_sym_SEMI] = ACTIONS(3185),
    [anon_sym_package] = ACTIONS(3185),
    [anon_sym_class] = ACTIONS(3185),
    [anon_sym_role] = ACTIONS(3185),
    [anon_sym_BUILD] = ACTIONS(3185),
    [anon_sym_ADJUST] = ACTIONS(3185),
    [anon_sym_DOLLAR] = ACTIONS(3185),
    [anon_sym_AT] = ACTIONS(3185),
    [anon_sym_my] = ACTIONS(3185),
    [anon_sym_extended] = ACTIONS(3185),
    [anon_sym_async] = ACTIONS(3185),
    [anon_sym_sub] = ACTIONS(3185),
    [anon_sym_method] = ACTIONS(3185),
    [anon_sym_LPAREN] = ACTIONS(3185),
    [anon_sym_continue] = ACTIONS(3185),
    [anon_sym_state] = ACTIONS(3185),
    [anon_sym_our] = ACTIONS(3185),
    [anon_sym_try] = ACTIONS(3185),
    [anon_sym_defer] = ACTIONS(3185),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3185),
    [anon_sym_LBRACK] = ACTIONS(3185),
    [anon_sym_LT_LT] = ACTIONS(3185),
    [anon_sym_SLASH_SLASH] = ACTIONS(3185),
    [anon_sym_PLUS] = ACTIONS(3185),
    [anon_sym_DASH] = ACTIONS(3185),
    [anon_sym_TILDE] = ACTIONS(3185),
    [anon_sym_BANG] = ACTIONS(3185),
    [anon_sym_PLUS_PLUS] = ACTIONS(3185),
    [anon_sym_DASH_DASH] = ACTIONS(3185),
    [anon_sym_BSLASH] = ACTIONS(3183),
    [anon_sym_do] = ACTIONS(3185),
    [anon_sym_eval] = ACTIONS(3185),
    [anon_sym_await] = ACTIONS(3185),
    [anon_sym_field] = ACTIONS(3185),
    [anon_sym_local] = ACTIONS(3185),
    [anon_sym_dynamically] = ACTIONS(3185),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3185),
    [anon_sym_require] = ACTIONS(3185),
    [anon_sym_map] = ACTIONS(3185),
    [anon_sym_grep] = ACTIONS(3185),
    [anon_sym_sort] = ACTIONS(3185),
    [anon_sym_goto] = ACTIONS(3185),
    [anon_sym_return] = ACTIONS(3185),
    [anon_sym_undef] = ACTIONS(3185),
    [anon_sym_PERCENT2] = ACTIONS(3185),
    [anon_sym_AMP2] = ACTIONS(3185),
    [anon_sym_STAR2] = ACTIONS(3185),
    [anon_sym_use] = ACTIONS(3185),
    [anon_sym_no] = ACTIONS(3185),
    [anon_sym_for] = ACTIONS(3185),
    [anon_sym_foreach] = ACTIONS(3185),
    [anon_sym_last] = ACTIONS(3185),
    [anon_sym_next] = ACTIONS(3185),
    [anon_sym_redo] = ACTIONS(3185),
    [anon_sym_BEGIN] = ACTIONS(3185),
    [anon_sym_INIT] = ACTIONS(3185),
    [anon_sym_CHECK] = ACTIONS(3185),
    [anon_sym_UNITCHECK] = ACTIONS(3185),
    [anon_sym_END] = ACTIONS(3185),
    [anon_sym___FILE__] = ACTIONS(3185),
    [anon_sym___LINE__] = ACTIONS(3185),
    [anon_sym___PACKAGE__] = ACTIONS(3185),
    [anon_sym___SUB__] = ACTIONS(3185),
    [anon_sym_break] = ACTIONS(3185),
    [anon_sym_fork] = ACTIONS(3185),
    [anon_sym_getppid] = ACTIONS(3185),
    [anon_sym_time] = ACTIONS(3185),
    [anon_sym_times] = ACTIONS(3185),
    [anon_sym_wait] = ACTIONS(3185),
    [anon_sym_wantarray] = ACTIONS(3185),
    [anon_sym_abs] = ACTIONS(3185),
    [anon_sym_alarm] = ACTIONS(3185),
    [anon_sym_chop] = ACTIONS(3185),
    [anon_sym_chdir] = ACTIONS(3185),
    [anon_sym_close] = ACTIONS(3185),
    [anon_sym_closedir] = ACTIONS(3185),
    [anon_sym_caller] = ACTIONS(3185),
    [anon_sym_chomp] = ACTIONS(3185),
    [anon_sym_chr] = ACTIONS(3185),
    [anon_sym_cos] = ACTIONS(3185),
    [anon_sym_chroot] = ACTIONS(3185),
    [anon_sym_defined] = ACTIONS(3185),
    [anon_sym_delete] = ACTIONS(3185),
    [anon_sym_dbmclose] = ACTIONS(3185),
    [anon_sym_exists] = ACTIONS(3185),
    [anon_sym_exit] = ACTIONS(3185),
    [anon_sym_eof] = ACTIONS(3185),
    [anon_sym_exp] = ACTIONS(3185),
    [anon_sym_each] = ACTIONS(3185),
    [anon_sym_fc] = ACTIONS(3185),
    [anon_sym_fileno] = ACTIONS(3185),
    [anon_sym_gmtime] = ACTIONS(3185),
    [anon_sym_getc] = ACTIONS(3185),
    [anon_sym_getpgrp] = ACTIONS(3185),
    [anon_sym_getprotobyname] = ACTIONS(3185),
    [anon_sym_getpwname] = ACTIONS(3185),
    [anon_sym_getpwuid] = ACTIONS(3185),
    [anon_sym_getpeername] = ACTIONS(3185),
    [anon_sym_getnetbyname] = ACTIONS(3185),
    [anon_sym_getsockname] = ACTIONS(3185),
    [anon_sym_getgrnam] = ACTIONS(3185),
    [anon_sym_getgrgid] = ACTIONS(3185),
    [anon_sym_hex] = ACTIONS(3185),
    [anon_sym_int] = ACTIONS(3185),
    [anon_sym_keys] = ACTIONS(3185),
    [anon_sym_lc] = ACTIONS(3185),
    [anon_sym_lcfirst] = ACTIONS(3185),
    [anon_sym_length] = ACTIONS(3185),
    [anon_sym_localtime] = ACTIONS(3185),
    [anon_sym_log] = ACTIONS(3185),
    [anon_sym_lock] = ACTIONS(3185),
    [anon_sym_lstat] = ACTIONS(3185),
    [anon_sym_oct] = ACTIONS(3185),
    [anon_sym_ord] = ACTIONS(3185),
    [anon_sym_prototype] = ACTIONS(3185),
    [anon_sym_pop] = ACTIONS(3185),
    [anon_sym_pos] = ACTIONS(3185),
    [anon_sym_quotemeta] = ACTIONS(3185),
    [anon_sym_reset] = ACTIONS(3185),
    [anon_sym_rand] = ACTIONS(3185),
    [anon_sym_rmdir] = ACTIONS(3185),
    [anon_sym_readdir] = ACTIONS(3185),
    [anon_sym_readline] = ACTIONS(3185),
    [anon_sym_readpipe] = ACTIONS(3185),
    [anon_sym_rewinddir] = ACTIONS(3185),
    [anon_sym_readlink] = ACTIONS(3185),
    [anon_sym_ref] = ACTIONS(3185),
    [anon_sym_scalar] = ACTIONS(3185),
    [anon_sym_shift] = ACTIONS(3185),
    [anon_sym_sin] = ACTIONS(3185),
    [anon_sym_sleep] = ACTIONS(3185),
    [anon_sym_sqrt] = ACTIONS(3185),
    [anon_sym_srand] = ACTIONS(3185),
    [anon_sym_stat] = ACTIONS(3185),
    [anon_sym_study] = ACTIONS(3185),
    [anon_sym_tell] = ACTIONS(3185),
    [anon_sym_telldir] = ACTIONS(3185),
    [anon_sym_tied] = ACTIONS(3185),
    [anon_sym_uc] = ACTIONS(3185),
    [anon_sym_ucfirst] = ACTIONS(3185),
    [anon_sym_untie] = ACTIONS(3185),
    [anon_sym_umask] = ACTIONS(3185),
    [anon_sym_values] = ACTIONS(3185),
    [anon_sym_write] = ACTIONS(3185),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3185),
    [anon_sym_qq] = ACTIONS(3185),
    [anon_sym_qw] = ACTIONS(3185),
    [anon_sym_qx] = ACTIONS(3185),
    [anon_sym_qr] = ACTIONS(3185),
    [anon_sym_m] = ACTIONS(3185),
    [anon_sym_s] = ACTIONS(3185),
    [anon_sym_tr] = ACTIONS(3185),
    [anon_sym_y] = ACTIONS(3185),
    [anon_sym_if] = ACTIONS(3185),
    [anon_sym_unless] = ACTIONS(3185),
    [anon_sym_while] = ACTIONS(3185),
    [anon_sym_until] = ACTIONS(3185),
    [aux_sym__bareword_token1] = ACTIONS(3185),
    [sym_number] = ACTIONS(3185),
    [anon_sym_true] = ACTIONS(3185),
    [anon_sym_false] = ACTIONS(3185),
    [sym__single_quote] = ACTIONS(3183),
    [sym__double_quote] = ACTIONS(3183),
    [sym__backtick_quote] = ACTIONS(3183),
    [sym__search_slash_quote] = ACTIONS(3183),
    [sym__open_readline_bracket] = ACTIONS(3183),
    [sym__open_fileglob_bracket] = ACTIONS(3183),
    [sym__PERLY_HEREDOC] = ACTIONS(3183),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3183),
    [sym__filetest] = ACTIONS(3183),
  },
  [STATE(1614)] = {
    [sym_heredoc_content] = STATE(1614),
    [sym__identifier] = ACTIONS(3153),
    [anon_sym_LBRACE] = ACTIONS(3153),
    [anon_sym_RBRACE] = ACTIONS(3153),
    [anon_sym_SEMI] = ACTIONS(3153),
    [anon_sym_package] = ACTIONS(3153),
    [anon_sym_class] = ACTIONS(3153),
    [anon_sym_role] = ACTIONS(3153),
    [anon_sym_BUILD] = ACTIONS(3153),
    [anon_sym_ADJUST] = ACTIONS(3153),
    [anon_sym_DOLLAR] = ACTIONS(3153),
    [anon_sym_AT] = ACTIONS(3153),
    [anon_sym_my] = ACTIONS(3153),
    [anon_sym_extended] = ACTIONS(3153),
    [anon_sym_async] = ACTIONS(3153),
    [anon_sym_sub] = ACTIONS(3153),
    [anon_sym_method] = ACTIONS(3153),
    [anon_sym_LPAREN] = ACTIONS(3153),
    [anon_sym_continue] = ACTIONS(3153),
    [anon_sym_state] = ACTIONS(3153),
    [anon_sym_our] = ACTIONS(3153),
    [anon_sym_try] = ACTIONS(3153),
    [anon_sym_defer] = ACTIONS(3153),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3153),
    [anon_sym_LBRACK] = ACTIONS(3153),
    [anon_sym_LT_LT] = ACTIONS(3153),
    [anon_sym_SLASH_SLASH] = ACTIONS(3153),
    [anon_sym_PLUS] = ACTIONS(3153),
    [anon_sym_DASH] = ACTIONS(3153),
    [anon_sym_TILDE] = ACTIONS(3153),
    [anon_sym_BANG] = ACTIONS(3153),
    [anon_sym_PLUS_PLUS] = ACTIONS(3153),
    [anon_sym_DASH_DASH] = ACTIONS(3153),
    [anon_sym_BSLASH] = ACTIONS(3151),
    [anon_sym_do] = ACTIONS(3153),
    [anon_sym_eval] = ACTIONS(3153),
    [anon_sym_await] = ACTIONS(3153),
    [anon_sym_field] = ACTIONS(3153),
    [anon_sym_local] = ACTIONS(3153),
    [anon_sym_dynamically] = ACTIONS(3153),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3153),
    [anon_sym_require] = ACTIONS(3153),
    [anon_sym_map] = ACTIONS(3153),
    [anon_sym_grep] = ACTIONS(3153),
    [anon_sym_sort] = ACTIONS(3153),
    [anon_sym_goto] = ACTIONS(3153),
    [anon_sym_return] = ACTIONS(3153),
    [anon_sym_undef] = ACTIONS(3153),
    [anon_sym_PERCENT2] = ACTIONS(3153),
    [anon_sym_AMP2] = ACTIONS(3153),
    [anon_sym_STAR2] = ACTIONS(3153),
    [anon_sym_use] = ACTIONS(3153),
    [anon_sym_no] = ACTIONS(3153),
    [anon_sym_for] = ACTIONS(3153),
    [anon_sym_foreach] = ACTIONS(3153),
    [anon_sym_last] = ACTIONS(3153),
    [anon_sym_next] = ACTIONS(3153),
    [anon_sym_redo] = ACTIONS(3153),
    [anon_sym_BEGIN] = ACTIONS(3153),
    [anon_sym_INIT] = ACTIONS(3153),
    [anon_sym_CHECK] = ACTIONS(3153),
    [anon_sym_UNITCHECK] = ACTIONS(3153),
    [anon_sym_END] = ACTIONS(3153),
    [anon_sym___FILE__] = ACTIONS(3153),
    [anon_sym___LINE__] = ACTIONS(3153),
    [anon_sym___PACKAGE__] = ACTIONS(3153),
    [anon_sym___SUB__] = ACTIONS(3153),
    [anon_sym_break] = ACTIONS(3153),
    [anon_sym_fork] = ACTIONS(3153),
    [anon_sym_getppid] = ACTIONS(3153),
    [anon_sym_time] = ACTIONS(3153),
    [anon_sym_times] = ACTIONS(3153),
    [anon_sym_wait] = ACTIONS(3153),
    [anon_sym_wantarray] = ACTIONS(3153),
    [anon_sym_abs] = ACTIONS(3153),
    [anon_sym_alarm] = ACTIONS(3153),
    [anon_sym_chop] = ACTIONS(3153),
    [anon_sym_chdir] = ACTIONS(3153),
    [anon_sym_close] = ACTIONS(3153),
    [anon_sym_closedir] = ACTIONS(3153),
    [anon_sym_caller] = ACTIONS(3153),
    [anon_sym_chomp] = ACTIONS(3153),
    [anon_sym_chr] = ACTIONS(3153),
    [anon_sym_cos] = ACTIONS(3153),
    [anon_sym_chroot] = ACTIONS(3153),
    [anon_sym_defined] = ACTIONS(3153),
    [anon_sym_delete] = ACTIONS(3153),
    [anon_sym_dbmclose] = ACTIONS(3153),
    [anon_sym_exists] = ACTIONS(3153),
    [anon_sym_exit] = ACTIONS(3153),
    [anon_sym_eof] = ACTIONS(3153),
    [anon_sym_exp] = ACTIONS(3153),
    [anon_sym_each] = ACTIONS(3153),
    [anon_sym_fc] = ACTIONS(3153),
    [anon_sym_fileno] = ACTIONS(3153),
    [anon_sym_gmtime] = ACTIONS(3153),
    [anon_sym_getc] = ACTIONS(3153),
    [anon_sym_getpgrp] = ACTIONS(3153),
    [anon_sym_getprotobyname] = ACTIONS(3153),
    [anon_sym_getpwname] = ACTIONS(3153),
    [anon_sym_getpwuid] = ACTIONS(3153),
    [anon_sym_getpeername] = ACTIONS(3153),
    [anon_sym_getnetbyname] = ACTIONS(3153),
    [anon_sym_getsockname] = ACTIONS(3153),
    [anon_sym_getgrnam] = ACTIONS(3153),
    [anon_sym_getgrgid] = ACTIONS(3153),
    [anon_sym_hex] = ACTIONS(3153),
    [anon_sym_int] = ACTIONS(3153),
    [anon_sym_keys] = ACTIONS(3153),
    [anon_sym_lc] = ACTIONS(3153),
    [anon_sym_lcfirst] = ACTIONS(3153),
    [anon_sym_length] = ACTIONS(3153),
    [anon_sym_localtime] = ACTIONS(3153),
    [anon_sym_log] = ACTIONS(3153),
    [anon_sym_lock] = ACTIONS(3153),
    [anon_sym_lstat] = ACTIONS(3153),
    [anon_sym_oct] = ACTIONS(3153),
    [anon_sym_ord] = ACTIONS(3153),
    [anon_sym_prototype] = ACTIONS(3153),
    [anon_sym_pop] = ACTIONS(3153),
    [anon_sym_pos] = ACTIONS(3153),
    [anon_sym_quotemeta] = ACTIONS(3153),
    [anon_sym_reset] = ACTIONS(3153),
    [anon_sym_rand] = ACTIONS(3153),
    [anon_sym_rmdir] = ACTIONS(3153),
    [anon_sym_readdir] = ACTIONS(3153),
    [anon_sym_readline] = ACTIONS(3153),
    [anon_sym_readpipe] = ACTIONS(3153),
    [anon_sym_rewinddir] = ACTIONS(3153),
    [anon_sym_readlink] = ACTIONS(3153),
    [anon_sym_ref] = ACTIONS(3153),
    [anon_sym_scalar] = ACTIONS(3153),
    [anon_sym_shift] = ACTIONS(3153),
    [anon_sym_sin] = ACTIONS(3153),
    [anon_sym_sleep] = ACTIONS(3153),
    [anon_sym_sqrt] = ACTIONS(3153),
    [anon_sym_srand] = ACTIONS(3153),
    [anon_sym_stat] = ACTIONS(3153),
    [anon_sym_study] = ACTIONS(3153),
    [anon_sym_tell] = ACTIONS(3153),
    [anon_sym_telldir] = ACTIONS(3153),
    [anon_sym_tied] = ACTIONS(3153),
    [anon_sym_uc] = ACTIONS(3153),
    [anon_sym_ucfirst] = ACTIONS(3153),
    [anon_sym_untie] = ACTIONS(3153),
    [anon_sym_umask] = ACTIONS(3153),
    [anon_sym_values] = ACTIONS(3153),
    [anon_sym_write] = ACTIONS(3153),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3153),
    [anon_sym_qq] = ACTIONS(3153),
    [anon_sym_qw] = ACTIONS(3153),
    [anon_sym_qx] = ACTIONS(3153),
    [anon_sym_qr] = ACTIONS(3153),
    [anon_sym_m] = ACTIONS(3153),
    [anon_sym_s] = ACTIONS(3153),
    [anon_sym_tr] = ACTIONS(3153),
    [anon_sym_y] = ACTIONS(3153),
    [anon_sym_if] = ACTIONS(3153),
    [anon_sym_unless] = ACTIONS(3153),
    [anon_sym_while] = ACTIONS(3153),
    [anon_sym_until] = ACTIONS(3153),
    [aux_sym__bareword_token1] = ACTIONS(3153),
    [sym_number] = ACTIONS(3153),
    [anon_sym_true] = ACTIONS(3153),
    [anon_sym_false] = ACTIONS(3153),
    [sym__single_quote] = ACTIONS(3151),
    [sym__double_quote] = ACTIONS(3151),
    [sym__backtick_quote] = ACTIONS(3151),
    [sym__search_slash_quote] = ACTIONS(3151),
    [sym__open_readline_bracket] = ACTIONS(3151),
    [sym__open_fileglob_bracket] = ACTIONS(3151),
    [sym__PERLY_HEREDOC] = ACTIONS(3151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3151),
    [sym__filetest] = ACTIONS(3151),
  },
  [STATE(1615)] = {
    [sym_heredoc_content] = STATE(1615),
    [sym__identifier] = ACTIONS(3205),
    [anon_sym_LBRACE] = ACTIONS(3205),
    [anon_sym_RBRACE] = ACTIONS(3205),
    [anon_sym_SEMI] = ACTIONS(3205),
    [anon_sym_package] = ACTIONS(3205),
    [anon_sym_class] = ACTIONS(3205),
    [anon_sym_role] = ACTIONS(3205),
    [anon_sym_BUILD] = ACTIONS(3205),
    [anon_sym_ADJUST] = ACTIONS(3205),
    [anon_sym_DOLLAR] = ACTIONS(3205),
    [anon_sym_AT] = ACTIONS(3205),
    [anon_sym_my] = ACTIONS(3205),
    [anon_sym_extended] = ACTIONS(3205),
    [anon_sym_async] = ACTIONS(3205),
    [anon_sym_sub] = ACTIONS(3205),
    [anon_sym_method] = ACTIONS(3205),
    [anon_sym_LPAREN] = ACTIONS(3205),
    [anon_sym_continue] = ACTIONS(3205),
    [anon_sym_state] = ACTIONS(3205),
    [anon_sym_our] = ACTIONS(3205),
    [anon_sym_try] = ACTIONS(3205),
    [anon_sym_defer] = ACTIONS(3205),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3205),
    [anon_sym_LBRACK] = ACTIONS(3205),
    [anon_sym_LT_LT] = ACTIONS(3205),
    [anon_sym_SLASH_SLASH] = ACTIONS(3205),
    [anon_sym_PLUS] = ACTIONS(3205),
    [anon_sym_DASH] = ACTIONS(3205),
    [anon_sym_TILDE] = ACTIONS(3205),
    [anon_sym_BANG] = ACTIONS(3205),
    [anon_sym_PLUS_PLUS] = ACTIONS(3205),
    [anon_sym_DASH_DASH] = ACTIONS(3205),
    [anon_sym_BSLASH] = ACTIONS(3203),
    [anon_sym_do] = ACTIONS(3205),
    [anon_sym_eval] = ACTIONS(3205),
    [anon_sym_await] = ACTIONS(3205),
    [anon_sym_field] = ACTIONS(3205),
    [anon_sym_local] = ACTIONS(3205),
    [anon_sym_dynamically] = ACTIONS(3205),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3205),
    [anon_sym_require] = ACTIONS(3205),
    [anon_sym_map] = ACTIONS(3205),
    [anon_sym_grep] = ACTIONS(3205),
    [anon_sym_sort] = ACTIONS(3205),
    [anon_sym_goto] = ACTIONS(3205),
    [anon_sym_return] = ACTIONS(3205),
    [anon_sym_undef] = ACTIONS(3205),
    [anon_sym_PERCENT2] = ACTIONS(3205),
    [anon_sym_AMP2] = ACTIONS(3205),
    [anon_sym_STAR2] = ACTIONS(3205),
    [anon_sym_use] = ACTIONS(3205),
    [anon_sym_no] = ACTIONS(3205),
    [anon_sym_for] = ACTIONS(3205),
    [anon_sym_foreach] = ACTIONS(3205),
    [anon_sym_last] = ACTIONS(3205),
    [anon_sym_next] = ACTIONS(3205),
    [anon_sym_redo] = ACTIONS(3205),
    [anon_sym_BEGIN] = ACTIONS(3205),
    [anon_sym_INIT] = ACTIONS(3205),
    [anon_sym_CHECK] = ACTIONS(3205),
    [anon_sym_UNITCHECK] = ACTIONS(3205),
    [anon_sym_END] = ACTIONS(3205),
    [anon_sym___FILE__] = ACTIONS(3205),
    [anon_sym___LINE__] = ACTIONS(3205),
    [anon_sym___PACKAGE__] = ACTIONS(3205),
    [anon_sym___SUB__] = ACTIONS(3205),
    [anon_sym_break] = ACTIONS(3205),
    [anon_sym_fork] = ACTIONS(3205),
    [anon_sym_getppid] = ACTIONS(3205),
    [anon_sym_time] = ACTIONS(3205),
    [anon_sym_times] = ACTIONS(3205),
    [anon_sym_wait] = ACTIONS(3205),
    [anon_sym_wantarray] = ACTIONS(3205),
    [anon_sym_abs] = ACTIONS(3205),
    [anon_sym_alarm] = ACTIONS(3205),
    [anon_sym_chop] = ACTIONS(3205),
    [anon_sym_chdir] = ACTIONS(3205),
    [anon_sym_close] = ACTIONS(3205),
    [anon_sym_closedir] = ACTIONS(3205),
    [anon_sym_caller] = ACTIONS(3205),
    [anon_sym_chomp] = ACTIONS(3205),
    [anon_sym_chr] = ACTIONS(3205),
    [anon_sym_cos] = ACTIONS(3205),
    [anon_sym_chroot] = ACTIONS(3205),
    [anon_sym_defined] = ACTIONS(3205),
    [anon_sym_delete] = ACTIONS(3205),
    [anon_sym_dbmclose] = ACTIONS(3205),
    [anon_sym_exists] = ACTIONS(3205),
    [anon_sym_exit] = ACTIONS(3205),
    [anon_sym_eof] = ACTIONS(3205),
    [anon_sym_exp] = ACTIONS(3205),
    [anon_sym_each] = ACTIONS(3205),
    [anon_sym_fc] = ACTIONS(3205),
    [anon_sym_fileno] = ACTIONS(3205),
    [anon_sym_gmtime] = ACTIONS(3205),
    [anon_sym_getc] = ACTIONS(3205),
    [anon_sym_getpgrp] = ACTIONS(3205),
    [anon_sym_getprotobyname] = ACTIONS(3205),
    [anon_sym_getpwname] = ACTIONS(3205),
    [anon_sym_getpwuid] = ACTIONS(3205),
    [anon_sym_getpeername] = ACTIONS(3205),
    [anon_sym_getnetbyname] = ACTIONS(3205),
    [anon_sym_getsockname] = ACTIONS(3205),
    [anon_sym_getgrnam] = ACTIONS(3205),
    [anon_sym_getgrgid] = ACTIONS(3205),
    [anon_sym_hex] = ACTIONS(3205),
    [anon_sym_int] = ACTIONS(3205),
    [anon_sym_keys] = ACTIONS(3205),
    [anon_sym_lc] = ACTIONS(3205),
    [anon_sym_lcfirst] = ACTIONS(3205),
    [anon_sym_length] = ACTIONS(3205),
    [anon_sym_localtime] = ACTIONS(3205),
    [anon_sym_log] = ACTIONS(3205),
    [anon_sym_lock] = ACTIONS(3205),
    [anon_sym_lstat] = ACTIONS(3205),
    [anon_sym_oct] = ACTIONS(3205),
    [anon_sym_ord] = ACTIONS(3205),
    [anon_sym_prototype] = ACTIONS(3205),
    [anon_sym_pop] = ACTIONS(3205),
    [anon_sym_pos] = ACTIONS(3205),
    [anon_sym_quotemeta] = ACTIONS(3205),
    [anon_sym_reset] = ACTIONS(3205),
    [anon_sym_rand] = ACTIONS(3205),
    [anon_sym_rmdir] = ACTIONS(3205),
    [anon_sym_readdir] = ACTIONS(3205),
    [anon_sym_readline] = ACTIONS(3205),
    [anon_sym_readpipe] = ACTIONS(3205),
    [anon_sym_rewinddir] = ACTIONS(3205),
    [anon_sym_readlink] = ACTIONS(3205),
    [anon_sym_ref] = ACTIONS(3205),
    [anon_sym_scalar] = ACTIONS(3205),
    [anon_sym_shift] = ACTIONS(3205),
    [anon_sym_sin] = ACTIONS(3205),
    [anon_sym_sleep] = ACTIONS(3205),
    [anon_sym_sqrt] = ACTIONS(3205),
    [anon_sym_srand] = ACTIONS(3205),
    [anon_sym_stat] = ACTIONS(3205),
    [anon_sym_study] = ACTIONS(3205),
    [anon_sym_tell] = ACTIONS(3205),
    [anon_sym_telldir] = ACTIONS(3205),
    [anon_sym_tied] = ACTIONS(3205),
    [anon_sym_uc] = ACTIONS(3205),
    [anon_sym_ucfirst] = ACTIONS(3205),
    [anon_sym_untie] = ACTIONS(3205),
    [anon_sym_umask] = ACTIONS(3205),
    [anon_sym_values] = ACTIONS(3205),
    [anon_sym_write] = ACTIONS(3205),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3205),
    [anon_sym_qq] = ACTIONS(3205),
    [anon_sym_qw] = ACTIONS(3205),
    [anon_sym_qx] = ACTIONS(3205),
    [anon_sym_qr] = ACTIONS(3205),
    [anon_sym_m] = ACTIONS(3205),
    [anon_sym_s] = ACTIONS(3205),
    [anon_sym_tr] = ACTIONS(3205),
    [anon_sym_y] = ACTIONS(3205),
    [anon_sym_if] = ACTIONS(3205),
    [anon_sym_unless] = ACTIONS(3205),
    [anon_sym_while] = ACTIONS(3205),
    [anon_sym_until] = ACTIONS(3205),
    [aux_sym__bareword_token1] = ACTIONS(3205),
    [sym_number] = ACTIONS(3205),
    [anon_sym_true] = ACTIONS(3205),
    [anon_sym_false] = ACTIONS(3205),
    [sym__single_quote] = ACTIONS(3203),
    [sym__double_quote] = ACTIONS(3203),
    [sym__backtick_quote] = ACTIONS(3203),
    [sym__search_slash_quote] = ACTIONS(3203),
    [sym__open_readline_bracket] = ACTIONS(3203),
    [sym__open_fileglob_bracket] = ACTIONS(3203),
    [sym__PERLY_HEREDOC] = ACTIONS(3203),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3203),
    [sym__filetest] = ACTIONS(3203),
  },
  [STATE(1616)] = {
    [sym_heredoc_content] = STATE(1616),
    [sym__identifier] = ACTIONS(3041),
    [anon_sym_LBRACE] = ACTIONS(3041),
    [anon_sym_RBRACE] = ACTIONS(3041),
    [anon_sym_SEMI] = ACTIONS(3041),
    [anon_sym_package] = ACTIONS(3041),
    [anon_sym_class] = ACTIONS(3041),
    [anon_sym_role] = ACTIONS(3041),
    [anon_sym_BUILD] = ACTIONS(3041),
    [anon_sym_ADJUST] = ACTIONS(3041),
    [anon_sym_DOLLAR] = ACTIONS(3041),
    [anon_sym_AT] = ACTIONS(3041),
    [anon_sym_my] = ACTIONS(3041),
    [anon_sym_extended] = ACTIONS(3041),
    [anon_sym_async] = ACTIONS(3041),
    [anon_sym_sub] = ACTIONS(3041),
    [anon_sym_method] = ACTIONS(3041),
    [anon_sym_LPAREN] = ACTIONS(3041),
    [anon_sym_continue] = ACTIONS(3041),
    [anon_sym_state] = ACTIONS(3041),
    [anon_sym_our] = ACTIONS(3041),
    [anon_sym_try] = ACTIONS(3041),
    [anon_sym_defer] = ACTIONS(3041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3041),
    [anon_sym_LBRACK] = ACTIONS(3041),
    [anon_sym_LT_LT] = ACTIONS(3041),
    [anon_sym_SLASH_SLASH] = ACTIONS(3041),
    [anon_sym_PLUS] = ACTIONS(3041),
    [anon_sym_DASH] = ACTIONS(3041),
    [anon_sym_TILDE] = ACTIONS(3041),
    [anon_sym_BANG] = ACTIONS(3041),
    [anon_sym_PLUS_PLUS] = ACTIONS(3041),
    [anon_sym_DASH_DASH] = ACTIONS(3041),
    [anon_sym_BSLASH] = ACTIONS(3039),
    [anon_sym_do] = ACTIONS(3041),
    [anon_sym_eval] = ACTIONS(3041),
    [anon_sym_await] = ACTIONS(3041),
    [anon_sym_field] = ACTIONS(3041),
    [anon_sym_local] = ACTIONS(3041),
    [anon_sym_dynamically] = ACTIONS(3041),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3041),
    [anon_sym_require] = ACTIONS(3041),
    [anon_sym_map] = ACTIONS(3041),
    [anon_sym_grep] = ACTIONS(3041),
    [anon_sym_sort] = ACTIONS(3041),
    [anon_sym_goto] = ACTIONS(3041),
    [anon_sym_return] = ACTIONS(3041),
    [anon_sym_undef] = ACTIONS(3041),
    [anon_sym_PERCENT2] = ACTIONS(3041),
    [anon_sym_AMP2] = ACTIONS(3041),
    [anon_sym_STAR2] = ACTIONS(3041),
    [anon_sym_use] = ACTIONS(3041),
    [anon_sym_no] = ACTIONS(3041),
    [anon_sym_for] = ACTIONS(3041),
    [anon_sym_foreach] = ACTIONS(3041),
    [anon_sym_last] = ACTIONS(3041),
    [anon_sym_next] = ACTIONS(3041),
    [anon_sym_redo] = ACTIONS(3041),
    [anon_sym_BEGIN] = ACTIONS(3041),
    [anon_sym_INIT] = ACTIONS(3041),
    [anon_sym_CHECK] = ACTIONS(3041),
    [anon_sym_UNITCHECK] = ACTIONS(3041),
    [anon_sym_END] = ACTIONS(3041),
    [anon_sym___FILE__] = ACTIONS(3041),
    [anon_sym___LINE__] = ACTIONS(3041),
    [anon_sym___PACKAGE__] = ACTIONS(3041),
    [anon_sym___SUB__] = ACTIONS(3041),
    [anon_sym_break] = ACTIONS(3041),
    [anon_sym_fork] = ACTIONS(3041),
    [anon_sym_getppid] = ACTIONS(3041),
    [anon_sym_time] = ACTIONS(3041),
    [anon_sym_times] = ACTIONS(3041),
    [anon_sym_wait] = ACTIONS(3041),
    [anon_sym_wantarray] = ACTIONS(3041),
    [anon_sym_abs] = ACTIONS(3041),
    [anon_sym_alarm] = ACTIONS(3041),
    [anon_sym_chop] = ACTIONS(3041),
    [anon_sym_chdir] = ACTIONS(3041),
    [anon_sym_close] = ACTIONS(3041),
    [anon_sym_closedir] = ACTIONS(3041),
    [anon_sym_caller] = ACTIONS(3041),
    [anon_sym_chomp] = ACTIONS(3041),
    [anon_sym_chr] = ACTIONS(3041),
    [anon_sym_cos] = ACTIONS(3041),
    [anon_sym_chroot] = ACTIONS(3041),
    [anon_sym_defined] = ACTIONS(3041),
    [anon_sym_delete] = ACTIONS(3041),
    [anon_sym_dbmclose] = ACTIONS(3041),
    [anon_sym_exists] = ACTIONS(3041),
    [anon_sym_exit] = ACTIONS(3041),
    [anon_sym_eof] = ACTIONS(3041),
    [anon_sym_exp] = ACTIONS(3041),
    [anon_sym_each] = ACTIONS(3041),
    [anon_sym_fc] = ACTIONS(3041),
    [anon_sym_fileno] = ACTIONS(3041),
    [anon_sym_gmtime] = ACTIONS(3041),
    [anon_sym_getc] = ACTIONS(3041),
    [anon_sym_getpgrp] = ACTIONS(3041),
    [anon_sym_getprotobyname] = ACTIONS(3041),
    [anon_sym_getpwname] = ACTIONS(3041),
    [anon_sym_getpwuid] = ACTIONS(3041),
    [anon_sym_getpeername] = ACTIONS(3041),
    [anon_sym_getnetbyname] = ACTIONS(3041),
    [anon_sym_getsockname] = ACTIONS(3041),
    [anon_sym_getgrnam] = ACTIONS(3041),
    [anon_sym_getgrgid] = ACTIONS(3041),
    [anon_sym_hex] = ACTIONS(3041),
    [anon_sym_int] = ACTIONS(3041),
    [anon_sym_keys] = ACTIONS(3041),
    [anon_sym_lc] = ACTIONS(3041),
    [anon_sym_lcfirst] = ACTIONS(3041),
    [anon_sym_length] = ACTIONS(3041),
    [anon_sym_localtime] = ACTIONS(3041),
    [anon_sym_log] = ACTIONS(3041),
    [anon_sym_lock] = ACTIONS(3041),
    [anon_sym_lstat] = ACTIONS(3041),
    [anon_sym_oct] = ACTIONS(3041),
    [anon_sym_ord] = ACTIONS(3041),
    [anon_sym_prototype] = ACTIONS(3041),
    [anon_sym_pop] = ACTIONS(3041),
    [anon_sym_pos] = ACTIONS(3041),
    [anon_sym_quotemeta] = ACTIONS(3041),
    [anon_sym_reset] = ACTIONS(3041),
    [anon_sym_rand] = ACTIONS(3041),
    [anon_sym_rmdir] = ACTIONS(3041),
    [anon_sym_readdir] = ACTIONS(3041),
    [anon_sym_readline] = ACTIONS(3041),
    [anon_sym_readpipe] = ACTIONS(3041),
    [anon_sym_rewinddir] = ACTIONS(3041),
    [anon_sym_readlink] = ACTIONS(3041),
    [anon_sym_ref] = ACTIONS(3041),
    [anon_sym_scalar] = ACTIONS(3041),
    [anon_sym_shift] = ACTIONS(3041),
    [anon_sym_sin] = ACTIONS(3041),
    [anon_sym_sleep] = ACTIONS(3041),
    [anon_sym_sqrt] = ACTIONS(3041),
    [anon_sym_srand] = ACTIONS(3041),
    [anon_sym_stat] = ACTIONS(3041),
    [anon_sym_study] = ACTIONS(3041),
    [anon_sym_tell] = ACTIONS(3041),
    [anon_sym_telldir] = ACTIONS(3041),
    [anon_sym_tied] = ACTIONS(3041),
    [anon_sym_uc] = ACTIONS(3041),
    [anon_sym_ucfirst] = ACTIONS(3041),
    [anon_sym_untie] = ACTIONS(3041),
    [anon_sym_umask] = ACTIONS(3041),
    [anon_sym_values] = ACTIONS(3041),
    [anon_sym_write] = ACTIONS(3041),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3041),
    [anon_sym_qq] = ACTIONS(3041),
    [anon_sym_qw] = ACTIONS(3041),
    [anon_sym_qx] = ACTIONS(3041),
    [anon_sym_qr] = ACTIONS(3041),
    [anon_sym_m] = ACTIONS(3041),
    [anon_sym_s] = ACTIONS(3041),
    [anon_sym_tr] = ACTIONS(3041),
    [anon_sym_y] = ACTIONS(3041),
    [anon_sym_if] = ACTIONS(3041),
    [anon_sym_unless] = ACTIONS(3041),
    [anon_sym_while] = ACTIONS(3041),
    [anon_sym_until] = ACTIONS(3041),
    [aux_sym__bareword_token1] = ACTIONS(3041),
    [sym_number] = ACTIONS(3041),
    [anon_sym_true] = ACTIONS(3041),
    [anon_sym_false] = ACTIONS(3041),
    [sym__single_quote] = ACTIONS(3039),
    [sym__double_quote] = ACTIONS(3039),
    [sym__backtick_quote] = ACTIONS(3039),
    [sym__search_slash_quote] = ACTIONS(3039),
    [sym__open_readline_bracket] = ACTIONS(3039),
    [sym__open_fileglob_bracket] = ACTIONS(3039),
    [sym__PERLY_HEREDOC] = ACTIONS(3039),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3039),
    [sym__filetest] = ACTIONS(3039),
  },
  [STATE(1617)] = {
    [sym_heredoc_content] = STATE(1617),
    [sym__identifier] = ACTIONS(3255),
    [anon_sym_LBRACE] = ACTIONS(3255),
    [anon_sym_RBRACE] = ACTIONS(3255),
    [anon_sym_SEMI] = ACTIONS(3255),
    [anon_sym_package] = ACTIONS(3255),
    [anon_sym_class] = ACTIONS(3255),
    [anon_sym_role] = ACTIONS(3255),
    [anon_sym_BUILD] = ACTIONS(3255),
    [anon_sym_ADJUST] = ACTIONS(3255),
    [anon_sym_DOLLAR] = ACTIONS(3255),
    [anon_sym_AT] = ACTIONS(3255),
    [anon_sym_my] = ACTIONS(3255),
    [anon_sym_extended] = ACTIONS(3255),
    [anon_sym_async] = ACTIONS(3255),
    [anon_sym_sub] = ACTIONS(3255),
    [anon_sym_method] = ACTIONS(3255),
    [anon_sym_LPAREN] = ACTIONS(3255),
    [anon_sym_continue] = ACTIONS(3255),
    [anon_sym_state] = ACTIONS(3255),
    [anon_sym_our] = ACTIONS(3255),
    [anon_sym_try] = ACTIONS(3255),
    [anon_sym_defer] = ACTIONS(3255),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3255),
    [anon_sym_LBRACK] = ACTIONS(3255),
    [anon_sym_LT_LT] = ACTIONS(3255),
    [anon_sym_SLASH_SLASH] = ACTIONS(3255),
    [anon_sym_PLUS] = ACTIONS(3255),
    [anon_sym_DASH] = ACTIONS(3255),
    [anon_sym_TILDE] = ACTIONS(3255),
    [anon_sym_BANG] = ACTIONS(3255),
    [anon_sym_PLUS_PLUS] = ACTIONS(3255),
    [anon_sym_DASH_DASH] = ACTIONS(3255),
    [anon_sym_BSLASH] = ACTIONS(3257),
    [anon_sym_do] = ACTIONS(3255),
    [anon_sym_eval] = ACTIONS(3255),
    [anon_sym_await] = ACTIONS(3255),
    [anon_sym_field] = ACTIONS(3255),
    [anon_sym_local] = ACTIONS(3255),
    [anon_sym_dynamically] = ACTIONS(3255),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3255),
    [anon_sym_require] = ACTIONS(3255),
    [anon_sym_map] = ACTIONS(3255),
    [anon_sym_grep] = ACTIONS(3255),
    [anon_sym_sort] = ACTIONS(3255),
    [anon_sym_goto] = ACTIONS(3255),
    [anon_sym_return] = ACTIONS(3255),
    [anon_sym_undef] = ACTIONS(3255),
    [anon_sym_PERCENT2] = ACTIONS(3255),
    [anon_sym_AMP2] = ACTIONS(3255),
    [anon_sym_STAR2] = ACTIONS(3255),
    [anon_sym_use] = ACTIONS(3255),
    [anon_sym_no] = ACTIONS(3255),
    [anon_sym_for] = ACTIONS(3255),
    [anon_sym_foreach] = ACTIONS(3255),
    [anon_sym_last] = ACTIONS(3255),
    [anon_sym_next] = ACTIONS(3255),
    [anon_sym_redo] = ACTIONS(3255),
    [anon_sym_BEGIN] = ACTIONS(3255),
    [anon_sym_INIT] = ACTIONS(3255),
    [anon_sym_CHECK] = ACTIONS(3255),
    [anon_sym_UNITCHECK] = ACTIONS(3255),
    [anon_sym_END] = ACTIONS(3255),
    [anon_sym___FILE__] = ACTIONS(3255),
    [anon_sym___LINE__] = ACTIONS(3255),
    [anon_sym___PACKAGE__] = ACTIONS(3255),
    [anon_sym___SUB__] = ACTIONS(3255),
    [anon_sym_break] = ACTIONS(3255),
    [anon_sym_fork] = ACTIONS(3255),
    [anon_sym_getppid] = ACTIONS(3255),
    [anon_sym_time] = ACTIONS(3255),
    [anon_sym_times] = ACTIONS(3255),
    [anon_sym_wait] = ACTIONS(3255),
    [anon_sym_wantarray] = ACTIONS(3255),
    [anon_sym_abs] = ACTIONS(3255),
    [anon_sym_alarm] = ACTIONS(3255),
    [anon_sym_chop] = ACTIONS(3255),
    [anon_sym_chdir] = ACTIONS(3255),
    [anon_sym_close] = ACTIONS(3255),
    [anon_sym_closedir] = ACTIONS(3255),
    [anon_sym_caller] = ACTIONS(3255),
    [anon_sym_chomp] = ACTIONS(3255),
    [anon_sym_chr] = ACTIONS(3255),
    [anon_sym_cos] = ACTIONS(3255),
    [anon_sym_chroot] = ACTIONS(3255),
    [anon_sym_defined] = ACTIONS(3255),
    [anon_sym_delete] = ACTIONS(3255),
    [anon_sym_dbmclose] = ACTIONS(3255),
    [anon_sym_exists] = ACTIONS(3255),
    [anon_sym_exit] = ACTIONS(3255),
    [anon_sym_eof] = ACTIONS(3255),
    [anon_sym_exp] = ACTIONS(3255),
    [anon_sym_each] = ACTIONS(3255),
    [anon_sym_fc] = ACTIONS(3255),
    [anon_sym_fileno] = ACTIONS(3255),
    [anon_sym_gmtime] = ACTIONS(3255),
    [anon_sym_getc] = ACTIONS(3255),
    [anon_sym_getpgrp] = ACTIONS(3255),
    [anon_sym_getprotobyname] = ACTIONS(3255),
    [anon_sym_getpwname] = ACTIONS(3255),
    [anon_sym_getpwuid] = ACTIONS(3255),
    [anon_sym_getpeername] = ACTIONS(3255),
    [anon_sym_getnetbyname] = ACTIONS(3255),
    [anon_sym_getsockname] = ACTIONS(3255),
    [anon_sym_getgrnam] = ACTIONS(3255),
    [anon_sym_getgrgid] = ACTIONS(3255),
    [anon_sym_hex] = ACTIONS(3255),
    [anon_sym_int] = ACTIONS(3255),
    [anon_sym_keys] = ACTIONS(3255),
    [anon_sym_lc] = ACTIONS(3255),
    [anon_sym_lcfirst] = ACTIONS(3255),
    [anon_sym_length] = ACTIONS(3255),
    [anon_sym_localtime] = ACTIONS(3255),
    [anon_sym_log] = ACTIONS(3255),
    [anon_sym_lock] = ACTIONS(3255),
    [anon_sym_lstat] = ACTIONS(3255),
    [anon_sym_oct] = ACTIONS(3255),
    [anon_sym_ord] = ACTIONS(3255),
    [anon_sym_prototype] = ACTIONS(3255),
    [anon_sym_pop] = ACTIONS(3255),
    [anon_sym_pos] = ACTIONS(3255),
    [anon_sym_quotemeta] = ACTIONS(3255),
    [anon_sym_reset] = ACTIONS(3255),
    [anon_sym_rand] = ACTIONS(3255),
    [anon_sym_rmdir] = ACTIONS(3255),
    [anon_sym_readdir] = ACTIONS(3255),
    [anon_sym_readline] = ACTIONS(3255),
    [anon_sym_readpipe] = ACTIONS(3255),
    [anon_sym_rewinddir] = ACTIONS(3255),
    [anon_sym_readlink] = ACTIONS(3255),
    [anon_sym_ref] = ACTIONS(3255),
    [anon_sym_scalar] = ACTIONS(3255),
    [anon_sym_shift] = ACTIONS(3255),
    [anon_sym_sin] = ACTIONS(3255),
    [anon_sym_sleep] = ACTIONS(3255),
    [anon_sym_sqrt] = ACTIONS(3255),
    [anon_sym_srand] = ACTIONS(3255),
    [anon_sym_stat] = ACTIONS(3255),
    [anon_sym_study] = ACTIONS(3255),
    [anon_sym_tell] = ACTIONS(3255),
    [anon_sym_telldir] = ACTIONS(3255),
    [anon_sym_tied] = ACTIONS(3255),
    [anon_sym_uc] = ACTIONS(3255),
    [anon_sym_ucfirst] = ACTIONS(3255),
    [anon_sym_untie] = ACTIONS(3255),
    [anon_sym_umask] = ACTIONS(3255),
    [anon_sym_values] = ACTIONS(3255),
    [anon_sym_write] = ACTIONS(3255),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3255),
    [anon_sym_qq] = ACTIONS(3255),
    [anon_sym_qw] = ACTIONS(3255),
    [anon_sym_qx] = ACTIONS(3255),
    [anon_sym_qr] = ACTIONS(3255),
    [anon_sym_m] = ACTIONS(3255),
    [anon_sym_s] = ACTIONS(3255),
    [anon_sym_tr] = ACTIONS(3255),
    [anon_sym_y] = ACTIONS(3255),
    [anon_sym_if] = ACTIONS(3255),
    [anon_sym_unless] = ACTIONS(3255),
    [anon_sym_while] = ACTIONS(3255),
    [anon_sym_until] = ACTIONS(3255),
    [aux_sym__bareword_token1] = ACTIONS(3255),
    [sym_number] = ACTIONS(3255),
    [anon_sym_true] = ACTIONS(3255),
    [anon_sym_false] = ACTIONS(3255),
    [sym__single_quote] = ACTIONS(3257),
    [sym__double_quote] = ACTIONS(3257),
    [sym__backtick_quote] = ACTIONS(3257),
    [sym__search_slash_quote] = ACTIONS(3257),
    [sym__open_readline_bracket] = ACTIONS(3257),
    [sym__open_fileglob_bracket] = ACTIONS(3257),
    [sym__PERLY_HEREDOC] = ACTIONS(3257),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3257),
    [sym__filetest] = ACTIONS(3257),
  },
  [STATE(1618)] = {
    [sym_heredoc_content] = STATE(1618),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_RBRACE] = ACTIONS(2688),
    [anon_sym_SEMI] = ACTIONS(2688),
    [anon_sym_package] = ACTIONS(2688),
    [anon_sym_class] = ACTIONS(2688),
    [anon_sym_role] = ACTIONS(2688),
    [anon_sym_BUILD] = ACTIONS(2688),
    [anon_sym_ADJUST] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_try] = ACTIONS(2688),
    [anon_sym_defer] = ACTIONS(2688),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_use] = ACTIONS(2688),
    [anon_sym_no] = ACTIONS(2688),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym_BEGIN] = ACTIONS(2688),
    [anon_sym_INIT] = ACTIONS(2688),
    [anon_sym_CHECK] = ACTIONS(2688),
    [anon_sym_UNITCHECK] = ACTIONS(2688),
    [anon_sym_END] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [anon_sym_if] = ACTIONS(2688),
    [anon_sym_unless] = ACTIONS(2688),
    [anon_sym_while] = ACTIONS(2688),
    [anon_sym_until] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1619)] = {
    [sym_heredoc_content] = STATE(1619),
    [sym__identifier] = ACTIONS(2902),
    [anon_sym_LBRACE] = ACTIONS(2902),
    [anon_sym_RBRACE] = ACTIONS(2902),
    [anon_sym_SEMI] = ACTIONS(2902),
    [anon_sym_package] = ACTIONS(2902),
    [anon_sym_class] = ACTIONS(2902),
    [anon_sym_role] = ACTIONS(2902),
    [anon_sym_BUILD] = ACTIONS(2902),
    [anon_sym_ADJUST] = ACTIONS(2902),
    [anon_sym_DOLLAR] = ACTIONS(2902),
    [anon_sym_AT] = ACTIONS(2902),
    [anon_sym_my] = ACTIONS(2902),
    [anon_sym_extended] = ACTIONS(2902),
    [anon_sym_async] = ACTIONS(2902),
    [anon_sym_sub] = ACTIONS(2902),
    [anon_sym_method] = ACTIONS(2902),
    [anon_sym_LPAREN] = ACTIONS(2902),
    [anon_sym_continue] = ACTIONS(2902),
    [anon_sym_state] = ACTIONS(2902),
    [anon_sym_our] = ACTIONS(2902),
    [anon_sym_try] = ACTIONS(2902),
    [anon_sym_defer] = ACTIONS(2902),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2902),
    [anon_sym_LBRACK] = ACTIONS(2902),
    [anon_sym_LT_LT] = ACTIONS(2902),
    [anon_sym_SLASH_SLASH] = ACTIONS(2902),
    [anon_sym_PLUS] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_PLUS_PLUS] = ACTIONS(2902),
    [anon_sym_DASH_DASH] = ACTIONS(2902),
    [anon_sym_BSLASH] = ACTIONS(2900),
    [anon_sym_do] = ACTIONS(2902),
    [anon_sym_eval] = ACTIONS(2902),
    [anon_sym_await] = ACTIONS(2902),
    [anon_sym_field] = ACTIONS(2902),
    [anon_sym_local] = ACTIONS(2902),
    [anon_sym_dynamically] = ACTIONS(2902),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2902),
    [anon_sym_require] = ACTIONS(2902),
    [anon_sym_map] = ACTIONS(2902),
    [anon_sym_grep] = ACTIONS(2902),
    [anon_sym_sort] = ACTIONS(2902),
    [anon_sym_goto] = ACTIONS(2902),
    [anon_sym_return] = ACTIONS(2902),
    [anon_sym_undef] = ACTIONS(2902),
    [anon_sym_PERCENT2] = ACTIONS(2902),
    [anon_sym_AMP2] = ACTIONS(2902),
    [anon_sym_STAR2] = ACTIONS(2902),
    [anon_sym_use] = ACTIONS(2902),
    [anon_sym_no] = ACTIONS(2902),
    [anon_sym_for] = ACTIONS(2902),
    [anon_sym_foreach] = ACTIONS(2902),
    [anon_sym_last] = ACTIONS(2902),
    [anon_sym_next] = ACTIONS(2902),
    [anon_sym_redo] = ACTIONS(2902),
    [anon_sym_BEGIN] = ACTIONS(2902),
    [anon_sym_INIT] = ACTIONS(2902),
    [anon_sym_CHECK] = ACTIONS(2902),
    [anon_sym_UNITCHECK] = ACTIONS(2902),
    [anon_sym_END] = ACTIONS(2902),
    [anon_sym___FILE__] = ACTIONS(2902),
    [anon_sym___LINE__] = ACTIONS(2902),
    [anon_sym___PACKAGE__] = ACTIONS(2902),
    [anon_sym___SUB__] = ACTIONS(2902),
    [anon_sym_break] = ACTIONS(2902),
    [anon_sym_fork] = ACTIONS(2902),
    [anon_sym_getppid] = ACTIONS(2902),
    [anon_sym_time] = ACTIONS(2902),
    [anon_sym_times] = ACTIONS(2902),
    [anon_sym_wait] = ACTIONS(2902),
    [anon_sym_wantarray] = ACTIONS(2902),
    [anon_sym_abs] = ACTIONS(2902),
    [anon_sym_alarm] = ACTIONS(2902),
    [anon_sym_chop] = ACTIONS(2902),
    [anon_sym_chdir] = ACTIONS(2902),
    [anon_sym_close] = ACTIONS(2902),
    [anon_sym_closedir] = ACTIONS(2902),
    [anon_sym_caller] = ACTIONS(2902),
    [anon_sym_chomp] = ACTIONS(2902),
    [anon_sym_chr] = ACTIONS(2902),
    [anon_sym_cos] = ACTIONS(2902),
    [anon_sym_chroot] = ACTIONS(2902),
    [anon_sym_defined] = ACTIONS(2902),
    [anon_sym_delete] = ACTIONS(2902),
    [anon_sym_dbmclose] = ACTIONS(2902),
    [anon_sym_exists] = ACTIONS(2902),
    [anon_sym_exit] = ACTIONS(2902),
    [anon_sym_eof] = ACTIONS(2902),
    [anon_sym_exp] = ACTIONS(2902),
    [anon_sym_each] = ACTIONS(2902),
    [anon_sym_fc] = ACTIONS(2902),
    [anon_sym_fileno] = ACTIONS(2902),
    [anon_sym_gmtime] = ACTIONS(2902),
    [anon_sym_getc] = ACTIONS(2902),
    [anon_sym_getpgrp] = ACTIONS(2902),
    [anon_sym_getprotobyname] = ACTIONS(2902),
    [anon_sym_getpwname] = ACTIONS(2902),
    [anon_sym_getpwuid] = ACTIONS(2902),
    [anon_sym_getpeername] = ACTIONS(2902),
    [anon_sym_getnetbyname] = ACTIONS(2902),
    [anon_sym_getsockname] = ACTIONS(2902),
    [anon_sym_getgrnam] = ACTIONS(2902),
    [anon_sym_getgrgid] = ACTIONS(2902),
    [anon_sym_hex] = ACTIONS(2902),
    [anon_sym_int] = ACTIONS(2902),
    [anon_sym_keys] = ACTIONS(2902),
    [anon_sym_lc] = ACTIONS(2902),
    [anon_sym_lcfirst] = ACTIONS(2902),
    [anon_sym_length] = ACTIONS(2902),
    [anon_sym_localtime] = ACTIONS(2902),
    [anon_sym_log] = ACTIONS(2902),
    [anon_sym_lock] = ACTIONS(2902),
    [anon_sym_lstat] = ACTIONS(2902),
    [anon_sym_oct] = ACTIONS(2902),
    [anon_sym_ord] = ACTIONS(2902),
    [anon_sym_prototype] = ACTIONS(2902),
    [anon_sym_pop] = ACTIONS(2902),
    [anon_sym_pos] = ACTIONS(2902),
    [anon_sym_quotemeta] = ACTIONS(2902),
    [anon_sym_reset] = ACTIONS(2902),
    [anon_sym_rand] = ACTIONS(2902),
    [anon_sym_rmdir] = ACTIONS(2902),
    [anon_sym_readdir] = ACTIONS(2902),
    [anon_sym_readline] = ACTIONS(2902),
    [anon_sym_readpipe] = ACTIONS(2902),
    [anon_sym_rewinddir] = ACTIONS(2902),
    [anon_sym_readlink] = ACTIONS(2902),
    [anon_sym_ref] = ACTIONS(2902),
    [anon_sym_scalar] = ACTIONS(2902),
    [anon_sym_shift] = ACTIONS(2902),
    [anon_sym_sin] = ACTIONS(2902),
    [anon_sym_sleep] = ACTIONS(2902),
    [anon_sym_sqrt] = ACTIONS(2902),
    [anon_sym_srand] = ACTIONS(2902),
    [anon_sym_stat] = ACTIONS(2902),
    [anon_sym_study] = ACTIONS(2902),
    [anon_sym_tell] = ACTIONS(2902),
    [anon_sym_telldir] = ACTIONS(2902),
    [anon_sym_tied] = ACTIONS(2902),
    [anon_sym_uc] = ACTIONS(2902),
    [anon_sym_ucfirst] = ACTIONS(2902),
    [anon_sym_untie] = ACTIONS(2902),
    [anon_sym_umask] = ACTIONS(2902),
    [anon_sym_values] = ACTIONS(2902),
    [anon_sym_write] = ACTIONS(2902),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2902),
    [anon_sym_qq] = ACTIONS(2902),
    [anon_sym_qw] = ACTIONS(2902),
    [anon_sym_qx] = ACTIONS(2902),
    [anon_sym_qr] = ACTIONS(2902),
    [anon_sym_m] = ACTIONS(2902),
    [anon_sym_s] = ACTIONS(2902),
    [anon_sym_tr] = ACTIONS(2902),
    [anon_sym_y] = ACTIONS(2902),
    [anon_sym_if] = ACTIONS(2902),
    [anon_sym_unless] = ACTIONS(2902),
    [anon_sym_while] = ACTIONS(2902),
    [anon_sym_until] = ACTIONS(2902),
    [aux_sym__bareword_token1] = ACTIONS(2902),
    [sym_number] = ACTIONS(2902),
    [anon_sym_true] = ACTIONS(2902),
    [anon_sym_false] = ACTIONS(2902),
    [sym__single_quote] = ACTIONS(2900),
    [sym__double_quote] = ACTIONS(2900),
    [sym__backtick_quote] = ACTIONS(2900),
    [sym__search_slash_quote] = ACTIONS(2900),
    [sym__open_readline_bracket] = ACTIONS(2900),
    [sym__open_fileglob_bracket] = ACTIONS(2900),
    [sym__PERLY_HEREDOC] = ACTIONS(2900),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2900),
    [sym__filetest] = ACTIONS(2900),
  },
  [STATE(1620)] = {
    [sym_heredoc_content] = STATE(1620),
    [sym__identifier] = ACTIONS(2946),
    [anon_sym_LBRACE] = ACTIONS(2946),
    [anon_sym_RBRACE] = ACTIONS(2946),
    [anon_sym_SEMI] = ACTIONS(2946),
    [anon_sym_package] = ACTIONS(2946),
    [anon_sym_class] = ACTIONS(2946),
    [anon_sym_role] = ACTIONS(2946),
    [anon_sym_BUILD] = ACTIONS(2946),
    [anon_sym_ADJUST] = ACTIONS(2946),
    [anon_sym_DOLLAR] = ACTIONS(2946),
    [anon_sym_AT] = ACTIONS(2946),
    [anon_sym_my] = ACTIONS(2946),
    [anon_sym_extended] = ACTIONS(2946),
    [anon_sym_async] = ACTIONS(2946),
    [anon_sym_sub] = ACTIONS(2946),
    [anon_sym_method] = ACTIONS(2946),
    [anon_sym_LPAREN] = ACTIONS(2946),
    [anon_sym_continue] = ACTIONS(2946),
    [anon_sym_state] = ACTIONS(2946),
    [anon_sym_our] = ACTIONS(2946),
    [anon_sym_try] = ACTIONS(2946),
    [anon_sym_defer] = ACTIONS(2946),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2946),
    [anon_sym_LBRACK] = ACTIONS(2946),
    [anon_sym_LT_LT] = ACTIONS(2946),
    [anon_sym_SLASH_SLASH] = ACTIONS(2946),
    [anon_sym_PLUS] = ACTIONS(2946),
    [anon_sym_DASH] = ACTIONS(2946),
    [anon_sym_TILDE] = ACTIONS(2946),
    [anon_sym_BANG] = ACTIONS(2946),
    [anon_sym_PLUS_PLUS] = ACTIONS(2946),
    [anon_sym_DASH_DASH] = ACTIONS(2946),
    [anon_sym_BSLASH] = ACTIONS(2944),
    [anon_sym_do] = ACTIONS(2946),
    [anon_sym_eval] = ACTIONS(2946),
    [anon_sym_await] = ACTIONS(2946),
    [anon_sym_field] = ACTIONS(2946),
    [anon_sym_local] = ACTIONS(2946),
    [anon_sym_dynamically] = ACTIONS(2946),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2946),
    [anon_sym_require] = ACTIONS(2946),
    [anon_sym_map] = ACTIONS(2946),
    [anon_sym_grep] = ACTIONS(2946),
    [anon_sym_sort] = ACTIONS(2946),
    [anon_sym_goto] = ACTIONS(2946),
    [anon_sym_return] = ACTIONS(2946),
    [anon_sym_undef] = ACTIONS(2946),
    [anon_sym_PERCENT2] = ACTIONS(2946),
    [anon_sym_AMP2] = ACTIONS(2946),
    [anon_sym_STAR2] = ACTIONS(2946),
    [anon_sym_use] = ACTIONS(2946),
    [anon_sym_no] = ACTIONS(2946),
    [anon_sym_for] = ACTIONS(2946),
    [anon_sym_foreach] = ACTIONS(2946),
    [anon_sym_last] = ACTIONS(2946),
    [anon_sym_next] = ACTIONS(2946),
    [anon_sym_redo] = ACTIONS(2946),
    [anon_sym_BEGIN] = ACTIONS(2946),
    [anon_sym_INIT] = ACTIONS(2946),
    [anon_sym_CHECK] = ACTIONS(2946),
    [anon_sym_UNITCHECK] = ACTIONS(2946),
    [anon_sym_END] = ACTIONS(2946),
    [anon_sym___FILE__] = ACTIONS(2946),
    [anon_sym___LINE__] = ACTIONS(2946),
    [anon_sym___PACKAGE__] = ACTIONS(2946),
    [anon_sym___SUB__] = ACTIONS(2946),
    [anon_sym_break] = ACTIONS(2946),
    [anon_sym_fork] = ACTIONS(2946),
    [anon_sym_getppid] = ACTIONS(2946),
    [anon_sym_time] = ACTIONS(2946),
    [anon_sym_times] = ACTIONS(2946),
    [anon_sym_wait] = ACTIONS(2946),
    [anon_sym_wantarray] = ACTIONS(2946),
    [anon_sym_abs] = ACTIONS(2946),
    [anon_sym_alarm] = ACTIONS(2946),
    [anon_sym_chop] = ACTIONS(2946),
    [anon_sym_chdir] = ACTIONS(2946),
    [anon_sym_close] = ACTIONS(2946),
    [anon_sym_closedir] = ACTIONS(2946),
    [anon_sym_caller] = ACTIONS(2946),
    [anon_sym_chomp] = ACTIONS(2946),
    [anon_sym_chr] = ACTIONS(2946),
    [anon_sym_cos] = ACTIONS(2946),
    [anon_sym_chroot] = ACTIONS(2946),
    [anon_sym_defined] = ACTIONS(2946),
    [anon_sym_delete] = ACTIONS(2946),
    [anon_sym_dbmclose] = ACTIONS(2946),
    [anon_sym_exists] = ACTIONS(2946),
    [anon_sym_exit] = ACTIONS(2946),
    [anon_sym_eof] = ACTIONS(2946),
    [anon_sym_exp] = ACTIONS(2946),
    [anon_sym_each] = ACTIONS(2946),
    [anon_sym_fc] = ACTIONS(2946),
    [anon_sym_fileno] = ACTIONS(2946),
    [anon_sym_gmtime] = ACTIONS(2946),
    [anon_sym_getc] = ACTIONS(2946),
    [anon_sym_getpgrp] = ACTIONS(2946),
    [anon_sym_getprotobyname] = ACTIONS(2946),
    [anon_sym_getpwname] = ACTIONS(2946),
    [anon_sym_getpwuid] = ACTIONS(2946),
    [anon_sym_getpeername] = ACTIONS(2946),
    [anon_sym_getnetbyname] = ACTIONS(2946),
    [anon_sym_getsockname] = ACTIONS(2946),
    [anon_sym_getgrnam] = ACTIONS(2946),
    [anon_sym_getgrgid] = ACTIONS(2946),
    [anon_sym_hex] = ACTIONS(2946),
    [anon_sym_int] = ACTIONS(2946),
    [anon_sym_keys] = ACTIONS(2946),
    [anon_sym_lc] = ACTIONS(2946),
    [anon_sym_lcfirst] = ACTIONS(2946),
    [anon_sym_length] = ACTIONS(2946),
    [anon_sym_localtime] = ACTIONS(2946),
    [anon_sym_log] = ACTIONS(2946),
    [anon_sym_lock] = ACTIONS(2946),
    [anon_sym_lstat] = ACTIONS(2946),
    [anon_sym_oct] = ACTIONS(2946),
    [anon_sym_ord] = ACTIONS(2946),
    [anon_sym_prototype] = ACTIONS(2946),
    [anon_sym_pop] = ACTIONS(2946),
    [anon_sym_pos] = ACTIONS(2946),
    [anon_sym_quotemeta] = ACTIONS(2946),
    [anon_sym_reset] = ACTIONS(2946),
    [anon_sym_rand] = ACTIONS(2946),
    [anon_sym_rmdir] = ACTIONS(2946),
    [anon_sym_readdir] = ACTIONS(2946),
    [anon_sym_readline] = ACTIONS(2946),
    [anon_sym_readpipe] = ACTIONS(2946),
    [anon_sym_rewinddir] = ACTIONS(2946),
    [anon_sym_readlink] = ACTIONS(2946),
    [anon_sym_ref] = ACTIONS(2946),
    [anon_sym_scalar] = ACTIONS(2946),
    [anon_sym_shift] = ACTIONS(2946),
    [anon_sym_sin] = ACTIONS(2946),
    [anon_sym_sleep] = ACTIONS(2946),
    [anon_sym_sqrt] = ACTIONS(2946),
    [anon_sym_srand] = ACTIONS(2946),
    [anon_sym_stat] = ACTIONS(2946),
    [anon_sym_study] = ACTIONS(2946),
    [anon_sym_tell] = ACTIONS(2946),
    [anon_sym_telldir] = ACTIONS(2946),
    [anon_sym_tied] = ACTIONS(2946),
    [anon_sym_uc] = ACTIONS(2946),
    [anon_sym_ucfirst] = ACTIONS(2946),
    [anon_sym_untie] = ACTIONS(2946),
    [anon_sym_umask] = ACTIONS(2946),
    [anon_sym_values] = ACTIONS(2946),
    [anon_sym_write] = ACTIONS(2946),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2946),
    [anon_sym_qq] = ACTIONS(2946),
    [anon_sym_qw] = ACTIONS(2946),
    [anon_sym_qx] = ACTIONS(2946),
    [anon_sym_qr] = ACTIONS(2946),
    [anon_sym_m] = ACTIONS(2946),
    [anon_sym_s] = ACTIONS(2946),
    [anon_sym_tr] = ACTIONS(2946),
    [anon_sym_y] = ACTIONS(2946),
    [anon_sym_if] = ACTIONS(2946),
    [anon_sym_unless] = ACTIONS(2946),
    [anon_sym_while] = ACTIONS(2946),
    [anon_sym_until] = ACTIONS(2946),
    [aux_sym__bareword_token1] = ACTIONS(2946),
    [sym_number] = ACTIONS(2946),
    [anon_sym_true] = ACTIONS(2946),
    [anon_sym_false] = ACTIONS(2946),
    [sym__single_quote] = ACTIONS(2944),
    [sym__double_quote] = ACTIONS(2944),
    [sym__backtick_quote] = ACTIONS(2944),
    [sym__search_slash_quote] = ACTIONS(2944),
    [sym__open_readline_bracket] = ACTIONS(2944),
    [sym__open_fileglob_bracket] = ACTIONS(2944),
    [sym__PERLY_HEREDOC] = ACTIONS(2944),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2944),
    [sym__filetest] = ACTIONS(2944),
  },
  [STATE(1621)] = {
    [sym_heredoc_content] = STATE(1621),
    [sym__identifier] = ACTIONS(2918),
    [anon_sym_LBRACE] = ACTIONS(2918),
    [anon_sym_RBRACE] = ACTIONS(2918),
    [anon_sym_SEMI] = ACTIONS(2918),
    [anon_sym_package] = ACTIONS(2918),
    [anon_sym_class] = ACTIONS(2918),
    [anon_sym_role] = ACTIONS(2918),
    [anon_sym_BUILD] = ACTIONS(2918),
    [anon_sym_ADJUST] = ACTIONS(2918),
    [anon_sym_DOLLAR] = ACTIONS(2918),
    [anon_sym_AT] = ACTIONS(2918),
    [anon_sym_my] = ACTIONS(2918),
    [anon_sym_extended] = ACTIONS(2918),
    [anon_sym_async] = ACTIONS(2918),
    [anon_sym_sub] = ACTIONS(2918),
    [anon_sym_method] = ACTIONS(2918),
    [anon_sym_LPAREN] = ACTIONS(2918),
    [anon_sym_continue] = ACTIONS(2918),
    [anon_sym_state] = ACTIONS(2918),
    [anon_sym_our] = ACTIONS(2918),
    [anon_sym_try] = ACTIONS(2918),
    [anon_sym_defer] = ACTIONS(2918),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2918),
    [anon_sym_LBRACK] = ACTIONS(2918),
    [anon_sym_LT_LT] = ACTIONS(2918),
    [anon_sym_SLASH_SLASH] = ACTIONS(2918),
    [anon_sym_PLUS] = ACTIONS(2918),
    [anon_sym_DASH] = ACTIONS(2918),
    [anon_sym_TILDE] = ACTIONS(2918),
    [anon_sym_BANG] = ACTIONS(2918),
    [anon_sym_PLUS_PLUS] = ACTIONS(2918),
    [anon_sym_DASH_DASH] = ACTIONS(2918),
    [anon_sym_BSLASH] = ACTIONS(2916),
    [anon_sym_do] = ACTIONS(2918),
    [anon_sym_eval] = ACTIONS(2918),
    [anon_sym_await] = ACTIONS(2918),
    [anon_sym_field] = ACTIONS(2918),
    [anon_sym_local] = ACTIONS(2918),
    [anon_sym_dynamically] = ACTIONS(2918),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2918),
    [anon_sym_require] = ACTIONS(2918),
    [anon_sym_map] = ACTIONS(2918),
    [anon_sym_grep] = ACTIONS(2918),
    [anon_sym_sort] = ACTIONS(2918),
    [anon_sym_goto] = ACTIONS(2918),
    [anon_sym_return] = ACTIONS(2918),
    [anon_sym_undef] = ACTIONS(2918),
    [anon_sym_PERCENT2] = ACTIONS(2918),
    [anon_sym_AMP2] = ACTIONS(2918),
    [anon_sym_STAR2] = ACTIONS(2918),
    [anon_sym_use] = ACTIONS(2918),
    [anon_sym_no] = ACTIONS(2918),
    [anon_sym_for] = ACTIONS(2918),
    [anon_sym_foreach] = ACTIONS(2918),
    [anon_sym_last] = ACTIONS(2918),
    [anon_sym_next] = ACTIONS(2918),
    [anon_sym_redo] = ACTIONS(2918),
    [anon_sym_BEGIN] = ACTIONS(2918),
    [anon_sym_INIT] = ACTIONS(2918),
    [anon_sym_CHECK] = ACTIONS(2918),
    [anon_sym_UNITCHECK] = ACTIONS(2918),
    [anon_sym_END] = ACTIONS(2918),
    [anon_sym___FILE__] = ACTIONS(2918),
    [anon_sym___LINE__] = ACTIONS(2918),
    [anon_sym___PACKAGE__] = ACTIONS(2918),
    [anon_sym___SUB__] = ACTIONS(2918),
    [anon_sym_break] = ACTIONS(2918),
    [anon_sym_fork] = ACTIONS(2918),
    [anon_sym_getppid] = ACTIONS(2918),
    [anon_sym_time] = ACTIONS(2918),
    [anon_sym_times] = ACTIONS(2918),
    [anon_sym_wait] = ACTIONS(2918),
    [anon_sym_wantarray] = ACTIONS(2918),
    [anon_sym_abs] = ACTIONS(2918),
    [anon_sym_alarm] = ACTIONS(2918),
    [anon_sym_chop] = ACTIONS(2918),
    [anon_sym_chdir] = ACTIONS(2918),
    [anon_sym_close] = ACTIONS(2918),
    [anon_sym_closedir] = ACTIONS(2918),
    [anon_sym_caller] = ACTIONS(2918),
    [anon_sym_chomp] = ACTIONS(2918),
    [anon_sym_chr] = ACTIONS(2918),
    [anon_sym_cos] = ACTIONS(2918),
    [anon_sym_chroot] = ACTIONS(2918),
    [anon_sym_defined] = ACTIONS(2918),
    [anon_sym_delete] = ACTIONS(2918),
    [anon_sym_dbmclose] = ACTIONS(2918),
    [anon_sym_exists] = ACTIONS(2918),
    [anon_sym_exit] = ACTIONS(2918),
    [anon_sym_eof] = ACTIONS(2918),
    [anon_sym_exp] = ACTIONS(2918),
    [anon_sym_each] = ACTIONS(2918),
    [anon_sym_fc] = ACTIONS(2918),
    [anon_sym_fileno] = ACTIONS(2918),
    [anon_sym_gmtime] = ACTIONS(2918),
    [anon_sym_getc] = ACTIONS(2918),
    [anon_sym_getpgrp] = ACTIONS(2918),
    [anon_sym_getprotobyname] = ACTIONS(2918),
    [anon_sym_getpwname] = ACTIONS(2918),
    [anon_sym_getpwuid] = ACTIONS(2918),
    [anon_sym_getpeername] = ACTIONS(2918),
    [anon_sym_getnetbyname] = ACTIONS(2918),
    [anon_sym_getsockname] = ACTIONS(2918),
    [anon_sym_getgrnam] = ACTIONS(2918),
    [anon_sym_getgrgid] = ACTIONS(2918),
    [anon_sym_hex] = ACTIONS(2918),
    [anon_sym_int] = ACTIONS(2918),
    [anon_sym_keys] = ACTIONS(2918),
    [anon_sym_lc] = ACTIONS(2918),
    [anon_sym_lcfirst] = ACTIONS(2918),
    [anon_sym_length] = ACTIONS(2918),
    [anon_sym_localtime] = ACTIONS(2918),
    [anon_sym_log] = ACTIONS(2918),
    [anon_sym_lock] = ACTIONS(2918),
    [anon_sym_lstat] = ACTIONS(2918),
    [anon_sym_oct] = ACTIONS(2918),
    [anon_sym_ord] = ACTIONS(2918),
    [anon_sym_prototype] = ACTIONS(2918),
    [anon_sym_pop] = ACTIONS(2918),
    [anon_sym_pos] = ACTIONS(2918),
    [anon_sym_quotemeta] = ACTIONS(2918),
    [anon_sym_reset] = ACTIONS(2918),
    [anon_sym_rand] = ACTIONS(2918),
    [anon_sym_rmdir] = ACTIONS(2918),
    [anon_sym_readdir] = ACTIONS(2918),
    [anon_sym_readline] = ACTIONS(2918),
    [anon_sym_readpipe] = ACTIONS(2918),
    [anon_sym_rewinddir] = ACTIONS(2918),
    [anon_sym_readlink] = ACTIONS(2918),
    [anon_sym_ref] = ACTIONS(2918),
    [anon_sym_scalar] = ACTIONS(2918),
    [anon_sym_shift] = ACTIONS(2918),
    [anon_sym_sin] = ACTIONS(2918),
    [anon_sym_sleep] = ACTIONS(2918),
    [anon_sym_sqrt] = ACTIONS(2918),
    [anon_sym_srand] = ACTIONS(2918),
    [anon_sym_stat] = ACTIONS(2918),
    [anon_sym_study] = ACTIONS(2918),
    [anon_sym_tell] = ACTIONS(2918),
    [anon_sym_telldir] = ACTIONS(2918),
    [anon_sym_tied] = ACTIONS(2918),
    [anon_sym_uc] = ACTIONS(2918),
    [anon_sym_ucfirst] = ACTIONS(2918),
    [anon_sym_untie] = ACTIONS(2918),
    [anon_sym_umask] = ACTIONS(2918),
    [anon_sym_values] = ACTIONS(2918),
    [anon_sym_write] = ACTIONS(2918),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2918),
    [anon_sym_qq] = ACTIONS(2918),
    [anon_sym_qw] = ACTIONS(2918),
    [anon_sym_qx] = ACTIONS(2918),
    [anon_sym_qr] = ACTIONS(2918),
    [anon_sym_m] = ACTIONS(2918),
    [anon_sym_s] = ACTIONS(2918),
    [anon_sym_tr] = ACTIONS(2918),
    [anon_sym_y] = ACTIONS(2918),
    [anon_sym_if] = ACTIONS(2918),
    [anon_sym_unless] = ACTIONS(2918),
    [anon_sym_while] = ACTIONS(2918),
    [anon_sym_until] = ACTIONS(2918),
    [aux_sym__bareword_token1] = ACTIONS(2918),
    [sym_number] = ACTIONS(2918),
    [anon_sym_true] = ACTIONS(2918),
    [anon_sym_false] = ACTIONS(2918),
    [sym__single_quote] = ACTIONS(2916),
    [sym__double_quote] = ACTIONS(2916),
    [sym__backtick_quote] = ACTIONS(2916),
    [sym__search_slash_quote] = ACTIONS(2916),
    [sym__open_readline_bracket] = ACTIONS(2916),
    [sym__open_fileglob_bracket] = ACTIONS(2916),
    [sym__PERLY_HEREDOC] = ACTIONS(2916),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2916),
    [sym__filetest] = ACTIONS(2916),
  },
  [STATE(1622)] = {
    [sym_heredoc_content] = STATE(1622),
    [sym__identifier] = ACTIONS(3241),
    [anon_sym_LBRACE] = ACTIONS(3241),
    [anon_sym_RBRACE] = ACTIONS(3241),
    [anon_sym_SEMI] = ACTIONS(3241),
    [anon_sym_package] = ACTIONS(3241),
    [anon_sym_class] = ACTIONS(3241),
    [anon_sym_role] = ACTIONS(3241),
    [anon_sym_BUILD] = ACTIONS(3241),
    [anon_sym_ADJUST] = ACTIONS(3241),
    [anon_sym_DOLLAR] = ACTIONS(3241),
    [anon_sym_AT] = ACTIONS(3241),
    [anon_sym_my] = ACTIONS(3241),
    [anon_sym_extended] = ACTIONS(3241),
    [anon_sym_async] = ACTIONS(3241),
    [anon_sym_sub] = ACTIONS(3241),
    [anon_sym_method] = ACTIONS(3241),
    [anon_sym_LPAREN] = ACTIONS(3241),
    [anon_sym_continue] = ACTIONS(3241),
    [anon_sym_state] = ACTIONS(3241),
    [anon_sym_our] = ACTIONS(3241),
    [anon_sym_try] = ACTIONS(3241),
    [anon_sym_defer] = ACTIONS(3241),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3241),
    [anon_sym_LBRACK] = ACTIONS(3241),
    [anon_sym_LT_LT] = ACTIONS(3241),
    [anon_sym_SLASH_SLASH] = ACTIONS(3241),
    [anon_sym_PLUS] = ACTIONS(3241),
    [anon_sym_DASH] = ACTIONS(3241),
    [anon_sym_TILDE] = ACTIONS(3241),
    [anon_sym_BANG] = ACTIONS(3241),
    [anon_sym_PLUS_PLUS] = ACTIONS(3241),
    [anon_sym_DASH_DASH] = ACTIONS(3241),
    [anon_sym_BSLASH] = ACTIONS(3239),
    [anon_sym_do] = ACTIONS(3241),
    [anon_sym_eval] = ACTIONS(3241),
    [anon_sym_await] = ACTIONS(3241),
    [anon_sym_field] = ACTIONS(3241),
    [anon_sym_local] = ACTIONS(3241),
    [anon_sym_dynamically] = ACTIONS(3241),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3241),
    [anon_sym_require] = ACTIONS(3241),
    [anon_sym_map] = ACTIONS(3241),
    [anon_sym_grep] = ACTIONS(3241),
    [anon_sym_sort] = ACTIONS(3241),
    [anon_sym_goto] = ACTIONS(3241),
    [anon_sym_return] = ACTIONS(3241),
    [anon_sym_undef] = ACTIONS(3241),
    [anon_sym_PERCENT2] = ACTIONS(3241),
    [anon_sym_AMP2] = ACTIONS(3241),
    [anon_sym_STAR2] = ACTIONS(3241),
    [anon_sym_use] = ACTIONS(3241),
    [anon_sym_no] = ACTIONS(3241),
    [anon_sym_for] = ACTIONS(3241),
    [anon_sym_foreach] = ACTIONS(3241),
    [anon_sym_last] = ACTIONS(3241),
    [anon_sym_next] = ACTIONS(3241),
    [anon_sym_redo] = ACTIONS(3241),
    [anon_sym_BEGIN] = ACTIONS(3241),
    [anon_sym_INIT] = ACTIONS(3241),
    [anon_sym_CHECK] = ACTIONS(3241),
    [anon_sym_UNITCHECK] = ACTIONS(3241),
    [anon_sym_END] = ACTIONS(3241),
    [anon_sym___FILE__] = ACTIONS(3241),
    [anon_sym___LINE__] = ACTIONS(3241),
    [anon_sym___PACKAGE__] = ACTIONS(3241),
    [anon_sym___SUB__] = ACTIONS(3241),
    [anon_sym_break] = ACTIONS(3241),
    [anon_sym_fork] = ACTIONS(3241),
    [anon_sym_getppid] = ACTIONS(3241),
    [anon_sym_time] = ACTIONS(3241),
    [anon_sym_times] = ACTIONS(3241),
    [anon_sym_wait] = ACTIONS(3241),
    [anon_sym_wantarray] = ACTIONS(3241),
    [anon_sym_abs] = ACTIONS(3241),
    [anon_sym_alarm] = ACTIONS(3241),
    [anon_sym_chop] = ACTIONS(3241),
    [anon_sym_chdir] = ACTIONS(3241),
    [anon_sym_close] = ACTIONS(3241),
    [anon_sym_closedir] = ACTIONS(3241),
    [anon_sym_caller] = ACTIONS(3241),
    [anon_sym_chomp] = ACTIONS(3241),
    [anon_sym_chr] = ACTIONS(3241),
    [anon_sym_cos] = ACTIONS(3241),
    [anon_sym_chroot] = ACTIONS(3241),
    [anon_sym_defined] = ACTIONS(3241),
    [anon_sym_delete] = ACTIONS(3241),
    [anon_sym_dbmclose] = ACTIONS(3241),
    [anon_sym_exists] = ACTIONS(3241),
    [anon_sym_exit] = ACTIONS(3241),
    [anon_sym_eof] = ACTIONS(3241),
    [anon_sym_exp] = ACTIONS(3241),
    [anon_sym_each] = ACTIONS(3241),
    [anon_sym_fc] = ACTIONS(3241),
    [anon_sym_fileno] = ACTIONS(3241),
    [anon_sym_gmtime] = ACTIONS(3241),
    [anon_sym_getc] = ACTIONS(3241),
    [anon_sym_getpgrp] = ACTIONS(3241),
    [anon_sym_getprotobyname] = ACTIONS(3241),
    [anon_sym_getpwname] = ACTIONS(3241),
    [anon_sym_getpwuid] = ACTIONS(3241),
    [anon_sym_getpeername] = ACTIONS(3241),
    [anon_sym_getnetbyname] = ACTIONS(3241),
    [anon_sym_getsockname] = ACTIONS(3241),
    [anon_sym_getgrnam] = ACTIONS(3241),
    [anon_sym_getgrgid] = ACTIONS(3241),
    [anon_sym_hex] = ACTIONS(3241),
    [anon_sym_int] = ACTIONS(3241),
    [anon_sym_keys] = ACTIONS(3241),
    [anon_sym_lc] = ACTIONS(3241),
    [anon_sym_lcfirst] = ACTIONS(3241),
    [anon_sym_length] = ACTIONS(3241),
    [anon_sym_localtime] = ACTIONS(3241),
    [anon_sym_log] = ACTIONS(3241),
    [anon_sym_lock] = ACTIONS(3241),
    [anon_sym_lstat] = ACTIONS(3241),
    [anon_sym_oct] = ACTIONS(3241),
    [anon_sym_ord] = ACTIONS(3241),
    [anon_sym_prototype] = ACTIONS(3241),
    [anon_sym_pop] = ACTIONS(3241),
    [anon_sym_pos] = ACTIONS(3241),
    [anon_sym_quotemeta] = ACTIONS(3241),
    [anon_sym_reset] = ACTIONS(3241),
    [anon_sym_rand] = ACTIONS(3241),
    [anon_sym_rmdir] = ACTIONS(3241),
    [anon_sym_readdir] = ACTIONS(3241),
    [anon_sym_readline] = ACTIONS(3241),
    [anon_sym_readpipe] = ACTIONS(3241),
    [anon_sym_rewinddir] = ACTIONS(3241),
    [anon_sym_readlink] = ACTIONS(3241),
    [anon_sym_ref] = ACTIONS(3241),
    [anon_sym_scalar] = ACTIONS(3241),
    [anon_sym_shift] = ACTIONS(3241),
    [anon_sym_sin] = ACTIONS(3241),
    [anon_sym_sleep] = ACTIONS(3241),
    [anon_sym_sqrt] = ACTIONS(3241),
    [anon_sym_srand] = ACTIONS(3241),
    [anon_sym_stat] = ACTIONS(3241),
    [anon_sym_study] = ACTIONS(3241),
    [anon_sym_tell] = ACTIONS(3241),
    [anon_sym_telldir] = ACTIONS(3241),
    [anon_sym_tied] = ACTIONS(3241),
    [anon_sym_uc] = ACTIONS(3241),
    [anon_sym_ucfirst] = ACTIONS(3241),
    [anon_sym_untie] = ACTIONS(3241),
    [anon_sym_umask] = ACTIONS(3241),
    [anon_sym_values] = ACTIONS(3241),
    [anon_sym_write] = ACTIONS(3241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3241),
    [anon_sym_qq] = ACTIONS(3241),
    [anon_sym_qw] = ACTIONS(3241),
    [anon_sym_qx] = ACTIONS(3241),
    [anon_sym_qr] = ACTIONS(3241),
    [anon_sym_m] = ACTIONS(3241),
    [anon_sym_s] = ACTIONS(3241),
    [anon_sym_tr] = ACTIONS(3241),
    [anon_sym_y] = ACTIONS(3241),
    [anon_sym_if] = ACTIONS(3241),
    [anon_sym_unless] = ACTIONS(3241),
    [anon_sym_while] = ACTIONS(3241),
    [anon_sym_until] = ACTIONS(3241),
    [aux_sym__bareword_token1] = ACTIONS(3241),
    [sym_number] = ACTIONS(3241),
    [anon_sym_true] = ACTIONS(3241),
    [anon_sym_false] = ACTIONS(3241),
    [sym__single_quote] = ACTIONS(3239),
    [sym__double_quote] = ACTIONS(3239),
    [sym__backtick_quote] = ACTIONS(3239),
    [sym__search_slash_quote] = ACTIONS(3239),
    [sym__open_readline_bracket] = ACTIONS(3239),
    [sym__open_fileglob_bracket] = ACTIONS(3239),
    [sym__PERLY_HEREDOC] = ACTIONS(3239),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3239),
    [sym__filetest] = ACTIONS(3239),
  },
  [STATE(1623)] = {
    [sym_heredoc_content] = STATE(1623),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_RBRACE] = ACTIONS(2532),
    [anon_sym_SEMI] = ACTIONS(2532),
    [anon_sym_package] = ACTIONS(2532),
    [anon_sym_class] = ACTIONS(2532),
    [anon_sym_role] = ACTIONS(2532),
    [anon_sym_BUILD] = ACTIONS(2532),
    [anon_sym_ADJUST] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_try] = ACTIONS(2532),
    [anon_sym_defer] = ACTIONS(2532),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_use] = ACTIONS(2532),
    [anon_sym_no] = ACTIONS(2532),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym_BEGIN] = ACTIONS(2532),
    [anon_sym_INIT] = ACTIONS(2532),
    [anon_sym_CHECK] = ACTIONS(2532),
    [anon_sym_UNITCHECK] = ACTIONS(2532),
    [anon_sym_END] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [anon_sym_if] = ACTIONS(2532),
    [anon_sym_unless] = ACTIONS(2532),
    [anon_sym_while] = ACTIONS(2532),
    [anon_sym_until] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1624)] = {
    [sym_heredoc_content] = STATE(1624),
    [sym__identifier] = ACTIONS(3221),
    [anon_sym_LBRACE] = ACTIONS(3221),
    [anon_sym_RBRACE] = ACTIONS(3221),
    [anon_sym_SEMI] = ACTIONS(3221),
    [anon_sym_package] = ACTIONS(3221),
    [anon_sym_class] = ACTIONS(3221),
    [anon_sym_role] = ACTIONS(3221),
    [anon_sym_BUILD] = ACTIONS(3221),
    [anon_sym_ADJUST] = ACTIONS(3221),
    [anon_sym_DOLLAR] = ACTIONS(3221),
    [anon_sym_AT] = ACTIONS(3221),
    [anon_sym_my] = ACTIONS(3221),
    [anon_sym_extended] = ACTIONS(3221),
    [anon_sym_async] = ACTIONS(3221),
    [anon_sym_sub] = ACTIONS(3221),
    [anon_sym_method] = ACTIONS(3221),
    [anon_sym_LPAREN] = ACTIONS(3221),
    [anon_sym_continue] = ACTIONS(3221),
    [anon_sym_state] = ACTIONS(3221),
    [anon_sym_our] = ACTIONS(3221),
    [anon_sym_try] = ACTIONS(3221),
    [anon_sym_defer] = ACTIONS(3221),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3221),
    [anon_sym_LBRACK] = ACTIONS(3221),
    [anon_sym_LT_LT] = ACTIONS(3221),
    [anon_sym_SLASH_SLASH] = ACTIONS(3221),
    [anon_sym_PLUS] = ACTIONS(3221),
    [anon_sym_DASH] = ACTIONS(3221),
    [anon_sym_TILDE] = ACTIONS(3221),
    [anon_sym_BANG] = ACTIONS(3221),
    [anon_sym_PLUS_PLUS] = ACTIONS(3221),
    [anon_sym_DASH_DASH] = ACTIONS(3221),
    [anon_sym_BSLASH] = ACTIONS(3219),
    [anon_sym_do] = ACTIONS(3221),
    [anon_sym_eval] = ACTIONS(3221),
    [anon_sym_await] = ACTIONS(3221),
    [anon_sym_field] = ACTIONS(3221),
    [anon_sym_local] = ACTIONS(3221),
    [anon_sym_dynamically] = ACTIONS(3221),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3221),
    [anon_sym_require] = ACTIONS(3221),
    [anon_sym_map] = ACTIONS(3221),
    [anon_sym_grep] = ACTIONS(3221),
    [anon_sym_sort] = ACTIONS(3221),
    [anon_sym_goto] = ACTIONS(3221),
    [anon_sym_return] = ACTIONS(3221),
    [anon_sym_undef] = ACTIONS(3221),
    [anon_sym_PERCENT2] = ACTIONS(3221),
    [anon_sym_AMP2] = ACTIONS(3221),
    [anon_sym_STAR2] = ACTIONS(3221),
    [anon_sym_use] = ACTIONS(3221),
    [anon_sym_no] = ACTIONS(3221),
    [anon_sym_for] = ACTIONS(3221),
    [anon_sym_foreach] = ACTIONS(3221),
    [anon_sym_last] = ACTIONS(3221),
    [anon_sym_next] = ACTIONS(3221),
    [anon_sym_redo] = ACTIONS(3221),
    [anon_sym_BEGIN] = ACTIONS(3221),
    [anon_sym_INIT] = ACTIONS(3221),
    [anon_sym_CHECK] = ACTIONS(3221),
    [anon_sym_UNITCHECK] = ACTIONS(3221),
    [anon_sym_END] = ACTIONS(3221),
    [anon_sym___FILE__] = ACTIONS(3221),
    [anon_sym___LINE__] = ACTIONS(3221),
    [anon_sym___PACKAGE__] = ACTIONS(3221),
    [anon_sym___SUB__] = ACTIONS(3221),
    [anon_sym_break] = ACTIONS(3221),
    [anon_sym_fork] = ACTIONS(3221),
    [anon_sym_getppid] = ACTIONS(3221),
    [anon_sym_time] = ACTIONS(3221),
    [anon_sym_times] = ACTIONS(3221),
    [anon_sym_wait] = ACTIONS(3221),
    [anon_sym_wantarray] = ACTIONS(3221),
    [anon_sym_abs] = ACTIONS(3221),
    [anon_sym_alarm] = ACTIONS(3221),
    [anon_sym_chop] = ACTIONS(3221),
    [anon_sym_chdir] = ACTIONS(3221),
    [anon_sym_close] = ACTIONS(3221),
    [anon_sym_closedir] = ACTIONS(3221),
    [anon_sym_caller] = ACTIONS(3221),
    [anon_sym_chomp] = ACTIONS(3221),
    [anon_sym_chr] = ACTIONS(3221),
    [anon_sym_cos] = ACTIONS(3221),
    [anon_sym_chroot] = ACTIONS(3221),
    [anon_sym_defined] = ACTIONS(3221),
    [anon_sym_delete] = ACTIONS(3221),
    [anon_sym_dbmclose] = ACTIONS(3221),
    [anon_sym_exists] = ACTIONS(3221),
    [anon_sym_exit] = ACTIONS(3221),
    [anon_sym_eof] = ACTIONS(3221),
    [anon_sym_exp] = ACTIONS(3221),
    [anon_sym_each] = ACTIONS(3221),
    [anon_sym_fc] = ACTIONS(3221),
    [anon_sym_fileno] = ACTIONS(3221),
    [anon_sym_gmtime] = ACTIONS(3221),
    [anon_sym_getc] = ACTIONS(3221),
    [anon_sym_getpgrp] = ACTIONS(3221),
    [anon_sym_getprotobyname] = ACTIONS(3221),
    [anon_sym_getpwname] = ACTIONS(3221),
    [anon_sym_getpwuid] = ACTIONS(3221),
    [anon_sym_getpeername] = ACTIONS(3221),
    [anon_sym_getnetbyname] = ACTIONS(3221),
    [anon_sym_getsockname] = ACTIONS(3221),
    [anon_sym_getgrnam] = ACTIONS(3221),
    [anon_sym_getgrgid] = ACTIONS(3221),
    [anon_sym_hex] = ACTIONS(3221),
    [anon_sym_int] = ACTIONS(3221),
    [anon_sym_keys] = ACTIONS(3221),
    [anon_sym_lc] = ACTIONS(3221),
    [anon_sym_lcfirst] = ACTIONS(3221),
    [anon_sym_length] = ACTIONS(3221),
    [anon_sym_localtime] = ACTIONS(3221),
    [anon_sym_log] = ACTIONS(3221),
    [anon_sym_lock] = ACTIONS(3221),
    [anon_sym_lstat] = ACTIONS(3221),
    [anon_sym_oct] = ACTIONS(3221),
    [anon_sym_ord] = ACTIONS(3221),
    [anon_sym_prototype] = ACTIONS(3221),
    [anon_sym_pop] = ACTIONS(3221),
    [anon_sym_pos] = ACTIONS(3221),
    [anon_sym_quotemeta] = ACTIONS(3221),
    [anon_sym_reset] = ACTIONS(3221),
    [anon_sym_rand] = ACTIONS(3221),
    [anon_sym_rmdir] = ACTIONS(3221),
    [anon_sym_readdir] = ACTIONS(3221),
    [anon_sym_readline] = ACTIONS(3221),
    [anon_sym_readpipe] = ACTIONS(3221),
    [anon_sym_rewinddir] = ACTIONS(3221),
    [anon_sym_readlink] = ACTIONS(3221),
    [anon_sym_ref] = ACTIONS(3221),
    [anon_sym_scalar] = ACTIONS(3221),
    [anon_sym_shift] = ACTIONS(3221),
    [anon_sym_sin] = ACTIONS(3221),
    [anon_sym_sleep] = ACTIONS(3221),
    [anon_sym_sqrt] = ACTIONS(3221),
    [anon_sym_srand] = ACTIONS(3221),
    [anon_sym_stat] = ACTIONS(3221),
    [anon_sym_study] = ACTIONS(3221),
    [anon_sym_tell] = ACTIONS(3221),
    [anon_sym_telldir] = ACTIONS(3221),
    [anon_sym_tied] = ACTIONS(3221),
    [anon_sym_uc] = ACTIONS(3221),
    [anon_sym_ucfirst] = ACTIONS(3221),
    [anon_sym_untie] = ACTIONS(3221),
    [anon_sym_umask] = ACTIONS(3221),
    [anon_sym_values] = ACTIONS(3221),
    [anon_sym_write] = ACTIONS(3221),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3221),
    [anon_sym_qq] = ACTIONS(3221),
    [anon_sym_qw] = ACTIONS(3221),
    [anon_sym_qx] = ACTIONS(3221),
    [anon_sym_qr] = ACTIONS(3221),
    [anon_sym_m] = ACTIONS(3221),
    [anon_sym_s] = ACTIONS(3221),
    [anon_sym_tr] = ACTIONS(3221),
    [anon_sym_y] = ACTIONS(3221),
    [anon_sym_if] = ACTIONS(3221),
    [anon_sym_unless] = ACTIONS(3221),
    [anon_sym_while] = ACTIONS(3221),
    [anon_sym_until] = ACTIONS(3221),
    [aux_sym__bareword_token1] = ACTIONS(3221),
    [sym_number] = ACTIONS(3221),
    [anon_sym_true] = ACTIONS(3221),
    [anon_sym_false] = ACTIONS(3221),
    [sym__single_quote] = ACTIONS(3219),
    [sym__double_quote] = ACTIONS(3219),
    [sym__backtick_quote] = ACTIONS(3219),
    [sym__search_slash_quote] = ACTIONS(3219),
    [sym__open_readline_bracket] = ACTIONS(3219),
    [sym__open_fileglob_bracket] = ACTIONS(3219),
    [sym__PERLY_HEREDOC] = ACTIONS(3219),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3219),
    [sym__filetest] = ACTIONS(3219),
  },
  [STATE(1625)] = {
    [sym_heredoc_content] = STATE(1625),
    [sym__identifier] = ACTIONS(3177),
    [anon_sym_LBRACE] = ACTIONS(3177),
    [anon_sym_RBRACE] = ACTIONS(3177),
    [anon_sym_SEMI] = ACTIONS(3177),
    [anon_sym_package] = ACTIONS(3177),
    [anon_sym_class] = ACTIONS(3177),
    [anon_sym_role] = ACTIONS(3177),
    [anon_sym_BUILD] = ACTIONS(3177),
    [anon_sym_ADJUST] = ACTIONS(3177),
    [anon_sym_DOLLAR] = ACTIONS(3177),
    [anon_sym_AT] = ACTIONS(3177),
    [anon_sym_my] = ACTIONS(3177),
    [anon_sym_extended] = ACTIONS(3177),
    [anon_sym_async] = ACTIONS(3177),
    [anon_sym_sub] = ACTIONS(3177),
    [anon_sym_method] = ACTIONS(3177),
    [anon_sym_LPAREN] = ACTIONS(3177),
    [anon_sym_continue] = ACTIONS(3177),
    [anon_sym_state] = ACTIONS(3177),
    [anon_sym_our] = ACTIONS(3177),
    [anon_sym_try] = ACTIONS(3177),
    [anon_sym_defer] = ACTIONS(3177),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3177),
    [anon_sym_LBRACK] = ACTIONS(3177),
    [anon_sym_LT_LT] = ACTIONS(3177),
    [anon_sym_SLASH_SLASH] = ACTIONS(3177),
    [anon_sym_PLUS] = ACTIONS(3177),
    [anon_sym_DASH] = ACTIONS(3177),
    [anon_sym_TILDE] = ACTIONS(3177),
    [anon_sym_BANG] = ACTIONS(3177),
    [anon_sym_PLUS_PLUS] = ACTIONS(3177),
    [anon_sym_DASH_DASH] = ACTIONS(3177),
    [anon_sym_BSLASH] = ACTIONS(3175),
    [anon_sym_do] = ACTIONS(3177),
    [anon_sym_eval] = ACTIONS(3177),
    [anon_sym_await] = ACTIONS(3177),
    [anon_sym_field] = ACTIONS(3177),
    [anon_sym_local] = ACTIONS(3177),
    [anon_sym_dynamically] = ACTIONS(3177),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3177),
    [anon_sym_require] = ACTIONS(3177),
    [anon_sym_map] = ACTIONS(3177),
    [anon_sym_grep] = ACTIONS(3177),
    [anon_sym_sort] = ACTIONS(3177),
    [anon_sym_goto] = ACTIONS(3177),
    [anon_sym_return] = ACTIONS(3177),
    [anon_sym_undef] = ACTIONS(3177),
    [anon_sym_PERCENT2] = ACTIONS(3177),
    [anon_sym_AMP2] = ACTIONS(3177),
    [anon_sym_STAR2] = ACTIONS(3177),
    [anon_sym_use] = ACTIONS(3177),
    [anon_sym_no] = ACTIONS(3177),
    [anon_sym_for] = ACTIONS(3177),
    [anon_sym_foreach] = ACTIONS(3177),
    [anon_sym_last] = ACTIONS(3177),
    [anon_sym_next] = ACTIONS(3177),
    [anon_sym_redo] = ACTIONS(3177),
    [anon_sym_BEGIN] = ACTIONS(3177),
    [anon_sym_INIT] = ACTIONS(3177),
    [anon_sym_CHECK] = ACTIONS(3177),
    [anon_sym_UNITCHECK] = ACTIONS(3177),
    [anon_sym_END] = ACTIONS(3177),
    [anon_sym___FILE__] = ACTIONS(3177),
    [anon_sym___LINE__] = ACTIONS(3177),
    [anon_sym___PACKAGE__] = ACTIONS(3177),
    [anon_sym___SUB__] = ACTIONS(3177),
    [anon_sym_break] = ACTIONS(3177),
    [anon_sym_fork] = ACTIONS(3177),
    [anon_sym_getppid] = ACTIONS(3177),
    [anon_sym_time] = ACTIONS(3177),
    [anon_sym_times] = ACTIONS(3177),
    [anon_sym_wait] = ACTIONS(3177),
    [anon_sym_wantarray] = ACTIONS(3177),
    [anon_sym_abs] = ACTIONS(3177),
    [anon_sym_alarm] = ACTIONS(3177),
    [anon_sym_chop] = ACTIONS(3177),
    [anon_sym_chdir] = ACTIONS(3177),
    [anon_sym_close] = ACTIONS(3177),
    [anon_sym_closedir] = ACTIONS(3177),
    [anon_sym_caller] = ACTIONS(3177),
    [anon_sym_chomp] = ACTIONS(3177),
    [anon_sym_chr] = ACTIONS(3177),
    [anon_sym_cos] = ACTIONS(3177),
    [anon_sym_chroot] = ACTIONS(3177),
    [anon_sym_defined] = ACTIONS(3177),
    [anon_sym_delete] = ACTIONS(3177),
    [anon_sym_dbmclose] = ACTIONS(3177),
    [anon_sym_exists] = ACTIONS(3177),
    [anon_sym_exit] = ACTIONS(3177),
    [anon_sym_eof] = ACTIONS(3177),
    [anon_sym_exp] = ACTIONS(3177),
    [anon_sym_each] = ACTIONS(3177),
    [anon_sym_fc] = ACTIONS(3177),
    [anon_sym_fileno] = ACTIONS(3177),
    [anon_sym_gmtime] = ACTIONS(3177),
    [anon_sym_getc] = ACTIONS(3177),
    [anon_sym_getpgrp] = ACTIONS(3177),
    [anon_sym_getprotobyname] = ACTIONS(3177),
    [anon_sym_getpwname] = ACTIONS(3177),
    [anon_sym_getpwuid] = ACTIONS(3177),
    [anon_sym_getpeername] = ACTIONS(3177),
    [anon_sym_getnetbyname] = ACTIONS(3177),
    [anon_sym_getsockname] = ACTIONS(3177),
    [anon_sym_getgrnam] = ACTIONS(3177),
    [anon_sym_getgrgid] = ACTIONS(3177),
    [anon_sym_hex] = ACTIONS(3177),
    [anon_sym_int] = ACTIONS(3177),
    [anon_sym_keys] = ACTIONS(3177),
    [anon_sym_lc] = ACTIONS(3177),
    [anon_sym_lcfirst] = ACTIONS(3177),
    [anon_sym_length] = ACTIONS(3177),
    [anon_sym_localtime] = ACTIONS(3177),
    [anon_sym_log] = ACTIONS(3177),
    [anon_sym_lock] = ACTIONS(3177),
    [anon_sym_lstat] = ACTIONS(3177),
    [anon_sym_oct] = ACTIONS(3177),
    [anon_sym_ord] = ACTIONS(3177),
    [anon_sym_prototype] = ACTIONS(3177),
    [anon_sym_pop] = ACTIONS(3177),
    [anon_sym_pos] = ACTIONS(3177),
    [anon_sym_quotemeta] = ACTIONS(3177),
    [anon_sym_reset] = ACTIONS(3177),
    [anon_sym_rand] = ACTIONS(3177),
    [anon_sym_rmdir] = ACTIONS(3177),
    [anon_sym_readdir] = ACTIONS(3177),
    [anon_sym_readline] = ACTIONS(3177),
    [anon_sym_readpipe] = ACTIONS(3177),
    [anon_sym_rewinddir] = ACTIONS(3177),
    [anon_sym_readlink] = ACTIONS(3177),
    [anon_sym_ref] = ACTIONS(3177),
    [anon_sym_scalar] = ACTIONS(3177),
    [anon_sym_shift] = ACTIONS(3177),
    [anon_sym_sin] = ACTIONS(3177),
    [anon_sym_sleep] = ACTIONS(3177),
    [anon_sym_sqrt] = ACTIONS(3177),
    [anon_sym_srand] = ACTIONS(3177),
    [anon_sym_stat] = ACTIONS(3177),
    [anon_sym_study] = ACTIONS(3177),
    [anon_sym_tell] = ACTIONS(3177),
    [anon_sym_telldir] = ACTIONS(3177),
    [anon_sym_tied] = ACTIONS(3177),
    [anon_sym_uc] = ACTIONS(3177),
    [anon_sym_ucfirst] = ACTIONS(3177),
    [anon_sym_untie] = ACTIONS(3177),
    [anon_sym_umask] = ACTIONS(3177),
    [anon_sym_values] = ACTIONS(3177),
    [anon_sym_write] = ACTIONS(3177),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3177),
    [anon_sym_qq] = ACTIONS(3177),
    [anon_sym_qw] = ACTIONS(3177),
    [anon_sym_qx] = ACTIONS(3177),
    [anon_sym_qr] = ACTIONS(3177),
    [anon_sym_m] = ACTIONS(3177),
    [anon_sym_s] = ACTIONS(3177),
    [anon_sym_tr] = ACTIONS(3177),
    [anon_sym_y] = ACTIONS(3177),
    [anon_sym_if] = ACTIONS(3177),
    [anon_sym_unless] = ACTIONS(3177),
    [anon_sym_while] = ACTIONS(3177),
    [anon_sym_until] = ACTIONS(3177),
    [aux_sym__bareword_token1] = ACTIONS(3177),
    [sym_number] = ACTIONS(3177),
    [anon_sym_true] = ACTIONS(3177),
    [anon_sym_false] = ACTIONS(3177),
    [sym__single_quote] = ACTIONS(3175),
    [sym__double_quote] = ACTIONS(3175),
    [sym__backtick_quote] = ACTIONS(3175),
    [sym__search_slash_quote] = ACTIONS(3175),
    [sym__open_readline_bracket] = ACTIONS(3175),
    [sym__open_fileglob_bracket] = ACTIONS(3175),
    [sym__PERLY_HEREDOC] = ACTIONS(3175),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3175),
    [sym__filetest] = ACTIONS(3175),
  },
  [STATE(1626)] = {
    [sym_heredoc_content] = STATE(1626),
    [sym__identifier] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2898),
    [anon_sym_RBRACE] = ACTIONS(2898),
    [anon_sym_SEMI] = ACTIONS(2898),
    [anon_sym_package] = ACTIONS(2898),
    [anon_sym_class] = ACTIONS(2898),
    [anon_sym_role] = ACTIONS(2898),
    [anon_sym_BUILD] = ACTIONS(2898),
    [anon_sym_ADJUST] = ACTIONS(2898),
    [anon_sym_DOLLAR] = ACTIONS(2898),
    [anon_sym_AT] = ACTIONS(2898),
    [anon_sym_my] = ACTIONS(2898),
    [anon_sym_extended] = ACTIONS(2898),
    [anon_sym_async] = ACTIONS(2898),
    [anon_sym_sub] = ACTIONS(2898),
    [anon_sym_method] = ACTIONS(2898),
    [anon_sym_LPAREN] = ACTIONS(2898),
    [anon_sym_continue] = ACTIONS(2898),
    [anon_sym_state] = ACTIONS(2898),
    [anon_sym_our] = ACTIONS(2898),
    [anon_sym_try] = ACTIONS(2898),
    [anon_sym_defer] = ACTIONS(2898),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2898),
    [anon_sym_LBRACK] = ACTIONS(2898),
    [anon_sym_LT_LT] = ACTIONS(2898),
    [anon_sym_SLASH_SLASH] = ACTIONS(2898),
    [anon_sym_PLUS] = ACTIONS(2898),
    [anon_sym_DASH] = ACTIONS(2898),
    [anon_sym_TILDE] = ACTIONS(2898),
    [anon_sym_BANG] = ACTIONS(2898),
    [anon_sym_PLUS_PLUS] = ACTIONS(2898),
    [anon_sym_DASH_DASH] = ACTIONS(2898),
    [anon_sym_BSLASH] = ACTIONS(2896),
    [anon_sym_do] = ACTIONS(2898),
    [anon_sym_eval] = ACTIONS(2898),
    [anon_sym_await] = ACTIONS(2898),
    [anon_sym_field] = ACTIONS(2898),
    [anon_sym_local] = ACTIONS(2898),
    [anon_sym_dynamically] = ACTIONS(2898),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2898),
    [anon_sym_require] = ACTIONS(2898),
    [anon_sym_map] = ACTIONS(2898),
    [anon_sym_grep] = ACTIONS(2898),
    [anon_sym_sort] = ACTIONS(2898),
    [anon_sym_goto] = ACTIONS(2898),
    [anon_sym_return] = ACTIONS(2898),
    [anon_sym_undef] = ACTIONS(2898),
    [anon_sym_PERCENT2] = ACTIONS(2898),
    [anon_sym_AMP2] = ACTIONS(2898),
    [anon_sym_STAR2] = ACTIONS(2898),
    [anon_sym_use] = ACTIONS(2898),
    [anon_sym_no] = ACTIONS(2898),
    [anon_sym_for] = ACTIONS(2898),
    [anon_sym_foreach] = ACTIONS(2898),
    [anon_sym_last] = ACTIONS(2898),
    [anon_sym_next] = ACTIONS(2898),
    [anon_sym_redo] = ACTIONS(2898),
    [anon_sym_BEGIN] = ACTIONS(2898),
    [anon_sym_INIT] = ACTIONS(2898),
    [anon_sym_CHECK] = ACTIONS(2898),
    [anon_sym_UNITCHECK] = ACTIONS(2898),
    [anon_sym_END] = ACTIONS(2898),
    [anon_sym___FILE__] = ACTIONS(2898),
    [anon_sym___LINE__] = ACTIONS(2898),
    [anon_sym___PACKAGE__] = ACTIONS(2898),
    [anon_sym___SUB__] = ACTIONS(2898),
    [anon_sym_break] = ACTIONS(2898),
    [anon_sym_fork] = ACTIONS(2898),
    [anon_sym_getppid] = ACTIONS(2898),
    [anon_sym_time] = ACTIONS(2898),
    [anon_sym_times] = ACTIONS(2898),
    [anon_sym_wait] = ACTIONS(2898),
    [anon_sym_wantarray] = ACTIONS(2898),
    [anon_sym_abs] = ACTIONS(2898),
    [anon_sym_alarm] = ACTIONS(2898),
    [anon_sym_chop] = ACTIONS(2898),
    [anon_sym_chdir] = ACTIONS(2898),
    [anon_sym_close] = ACTIONS(2898),
    [anon_sym_closedir] = ACTIONS(2898),
    [anon_sym_caller] = ACTIONS(2898),
    [anon_sym_chomp] = ACTIONS(2898),
    [anon_sym_chr] = ACTIONS(2898),
    [anon_sym_cos] = ACTIONS(2898),
    [anon_sym_chroot] = ACTIONS(2898),
    [anon_sym_defined] = ACTIONS(2898),
    [anon_sym_delete] = ACTIONS(2898),
    [anon_sym_dbmclose] = ACTIONS(2898),
    [anon_sym_exists] = ACTIONS(2898),
    [anon_sym_exit] = ACTIONS(2898),
    [anon_sym_eof] = ACTIONS(2898),
    [anon_sym_exp] = ACTIONS(2898),
    [anon_sym_each] = ACTIONS(2898),
    [anon_sym_fc] = ACTIONS(2898),
    [anon_sym_fileno] = ACTIONS(2898),
    [anon_sym_gmtime] = ACTIONS(2898),
    [anon_sym_getc] = ACTIONS(2898),
    [anon_sym_getpgrp] = ACTIONS(2898),
    [anon_sym_getprotobyname] = ACTIONS(2898),
    [anon_sym_getpwname] = ACTIONS(2898),
    [anon_sym_getpwuid] = ACTIONS(2898),
    [anon_sym_getpeername] = ACTIONS(2898),
    [anon_sym_getnetbyname] = ACTIONS(2898),
    [anon_sym_getsockname] = ACTIONS(2898),
    [anon_sym_getgrnam] = ACTIONS(2898),
    [anon_sym_getgrgid] = ACTIONS(2898),
    [anon_sym_hex] = ACTIONS(2898),
    [anon_sym_int] = ACTIONS(2898),
    [anon_sym_keys] = ACTIONS(2898),
    [anon_sym_lc] = ACTIONS(2898),
    [anon_sym_lcfirst] = ACTIONS(2898),
    [anon_sym_length] = ACTIONS(2898),
    [anon_sym_localtime] = ACTIONS(2898),
    [anon_sym_log] = ACTIONS(2898),
    [anon_sym_lock] = ACTIONS(2898),
    [anon_sym_lstat] = ACTIONS(2898),
    [anon_sym_oct] = ACTIONS(2898),
    [anon_sym_ord] = ACTIONS(2898),
    [anon_sym_prototype] = ACTIONS(2898),
    [anon_sym_pop] = ACTIONS(2898),
    [anon_sym_pos] = ACTIONS(2898),
    [anon_sym_quotemeta] = ACTIONS(2898),
    [anon_sym_reset] = ACTIONS(2898),
    [anon_sym_rand] = ACTIONS(2898),
    [anon_sym_rmdir] = ACTIONS(2898),
    [anon_sym_readdir] = ACTIONS(2898),
    [anon_sym_readline] = ACTIONS(2898),
    [anon_sym_readpipe] = ACTIONS(2898),
    [anon_sym_rewinddir] = ACTIONS(2898),
    [anon_sym_readlink] = ACTIONS(2898),
    [anon_sym_ref] = ACTIONS(2898),
    [anon_sym_scalar] = ACTIONS(2898),
    [anon_sym_shift] = ACTIONS(2898),
    [anon_sym_sin] = ACTIONS(2898),
    [anon_sym_sleep] = ACTIONS(2898),
    [anon_sym_sqrt] = ACTIONS(2898),
    [anon_sym_srand] = ACTIONS(2898),
    [anon_sym_stat] = ACTIONS(2898),
    [anon_sym_study] = ACTIONS(2898),
    [anon_sym_tell] = ACTIONS(2898),
    [anon_sym_telldir] = ACTIONS(2898),
    [anon_sym_tied] = ACTIONS(2898),
    [anon_sym_uc] = ACTIONS(2898),
    [anon_sym_ucfirst] = ACTIONS(2898),
    [anon_sym_untie] = ACTIONS(2898),
    [anon_sym_umask] = ACTIONS(2898),
    [anon_sym_values] = ACTIONS(2898),
    [anon_sym_write] = ACTIONS(2898),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2898),
    [anon_sym_qq] = ACTIONS(2898),
    [anon_sym_qw] = ACTIONS(2898),
    [anon_sym_qx] = ACTIONS(2898),
    [anon_sym_qr] = ACTIONS(2898),
    [anon_sym_m] = ACTIONS(2898),
    [anon_sym_s] = ACTIONS(2898),
    [anon_sym_tr] = ACTIONS(2898),
    [anon_sym_y] = ACTIONS(2898),
    [anon_sym_if] = ACTIONS(2898),
    [anon_sym_unless] = ACTIONS(2898),
    [anon_sym_while] = ACTIONS(2898),
    [anon_sym_until] = ACTIONS(2898),
    [aux_sym__bareword_token1] = ACTIONS(2898),
    [sym_number] = ACTIONS(2898),
    [anon_sym_true] = ACTIONS(2898),
    [anon_sym_false] = ACTIONS(2898),
    [sym__single_quote] = ACTIONS(2896),
    [sym__double_quote] = ACTIONS(2896),
    [sym__backtick_quote] = ACTIONS(2896),
    [sym__search_slash_quote] = ACTIONS(2896),
    [sym__open_readline_bracket] = ACTIONS(2896),
    [sym__open_fileglob_bracket] = ACTIONS(2896),
    [sym__PERLY_HEREDOC] = ACTIONS(2896),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2896),
    [sym__filetest] = ACTIONS(2896),
  },
  [STATE(1627)] = {
    [sym_heredoc_content] = STATE(1627),
    [sym__identifier] = ACTIONS(3045),
    [anon_sym_LBRACE] = ACTIONS(3045),
    [anon_sym_RBRACE] = ACTIONS(3045),
    [anon_sym_SEMI] = ACTIONS(3045),
    [anon_sym_package] = ACTIONS(3045),
    [anon_sym_class] = ACTIONS(3045),
    [anon_sym_role] = ACTIONS(3045),
    [anon_sym_BUILD] = ACTIONS(3045),
    [anon_sym_ADJUST] = ACTIONS(3045),
    [anon_sym_DOLLAR] = ACTIONS(3045),
    [anon_sym_AT] = ACTIONS(3045),
    [anon_sym_my] = ACTIONS(3045),
    [anon_sym_extended] = ACTIONS(3045),
    [anon_sym_async] = ACTIONS(3045),
    [anon_sym_sub] = ACTIONS(3045),
    [anon_sym_method] = ACTIONS(3045),
    [anon_sym_LPAREN] = ACTIONS(3045),
    [anon_sym_continue] = ACTIONS(3045),
    [anon_sym_state] = ACTIONS(3045),
    [anon_sym_our] = ACTIONS(3045),
    [anon_sym_try] = ACTIONS(3045),
    [anon_sym_defer] = ACTIONS(3045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3045),
    [anon_sym_LBRACK] = ACTIONS(3045),
    [anon_sym_LT_LT] = ACTIONS(3045),
    [anon_sym_SLASH_SLASH] = ACTIONS(3045),
    [anon_sym_PLUS] = ACTIONS(3045),
    [anon_sym_DASH] = ACTIONS(3045),
    [anon_sym_TILDE] = ACTIONS(3045),
    [anon_sym_BANG] = ACTIONS(3045),
    [anon_sym_PLUS_PLUS] = ACTIONS(3045),
    [anon_sym_DASH_DASH] = ACTIONS(3045),
    [anon_sym_BSLASH] = ACTIONS(3043),
    [anon_sym_do] = ACTIONS(3045),
    [anon_sym_eval] = ACTIONS(3045),
    [anon_sym_await] = ACTIONS(3045),
    [anon_sym_field] = ACTIONS(3045),
    [anon_sym_local] = ACTIONS(3045),
    [anon_sym_dynamically] = ACTIONS(3045),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3045),
    [anon_sym_require] = ACTIONS(3045),
    [anon_sym_map] = ACTIONS(3045),
    [anon_sym_grep] = ACTIONS(3045),
    [anon_sym_sort] = ACTIONS(3045),
    [anon_sym_goto] = ACTIONS(3045),
    [anon_sym_return] = ACTIONS(3045),
    [anon_sym_undef] = ACTIONS(3045),
    [anon_sym_PERCENT2] = ACTIONS(3045),
    [anon_sym_AMP2] = ACTIONS(3045),
    [anon_sym_STAR2] = ACTIONS(3045),
    [anon_sym_use] = ACTIONS(3045),
    [anon_sym_no] = ACTIONS(3045),
    [anon_sym_for] = ACTIONS(3045),
    [anon_sym_foreach] = ACTIONS(3045),
    [anon_sym_last] = ACTIONS(3045),
    [anon_sym_next] = ACTIONS(3045),
    [anon_sym_redo] = ACTIONS(3045),
    [anon_sym_BEGIN] = ACTIONS(3045),
    [anon_sym_INIT] = ACTIONS(3045),
    [anon_sym_CHECK] = ACTIONS(3045),
    [anon_sym_UNITCHECK] = ACTIONS(3045),
    [anon_sym_END] = ACTIONS(3045),
    [anon_sym___FILE__] = ACTIONS(3045),
    [anon_sym___LINE__] = ACTIONS(3045),
    [anon_sym___PACKAGE__] = ACTIONS(3045),
    [anon_sym___SUB__] = ACTIONS(3045),
    [anon_sym_break] = ACTIONS(3045),
    [anon_sym_fork] = ACTIONS(3045),
    [anon_sym_getppid] = ACTIONS(3045),
    [anon_sym_time] = ACTIONS(3045),
    [anon_sym_times] = ACTIONS(3045),
    [anon_sym_wait] = ACTIONS(3045),
    [anon_sym_wantarray] = ACTIONS(3045),
    [anon_sym_abs] = ACTIONS(3045),
    [anon_sym_alarm] = ACTIONS(3045),
    [anon_sym_chop] = ACTIONS(3045),
    [anon_sym_chdir] = ACTIONS(3045),
    [anon_sym_close] = ACTIONS(3045),
    [anon_sym_closedir] = ACTIONS(3045),
    [anon_sym_caller] = ACTIONS(3045),
    [anon_sym_chomp] = ACTIONS(3045),
    [anon_sym_chr] = ACTIONS(3045),
    [anon_sym_cos] = ACTIONS(3045),
    [anon_sym_chroot] = ACTIONS(3045),
    [anon_sym_defined] = ACTIONS(3045),
    [anon_sym_delete] = ACTIONS(3045),
    [anon_sym_dbmclose] = ACTIONS(3045),
    [anon_sym_exists] = ACTIONS(3045),
    [anon_sym_exit] = ACTIONS(3045),
    [anon_sym_eof] = ACTIONS(3045),
    [anon_sym_exp] = ACTIONS(3045),
    [anon_sym_each] = ACTIONS(3045),
    [anon_sym_fc] = ACTIONS(3045),
    [anon_sym_fileno] = ACTIONS(3045),
    [anon_sym_gmtime] = ACTIONS(3045),
    [anon_sym_getc] = ACTIONS(3045),
    [anon_sym_getpgrp] = ACTIONS(3045),
    [anon_sym_getprotobyname] = ACTIONS(3045),
    [anon_sym_getpwname] = ACTIONS(3045),
    [anon_sym_getpwuid] = ACTIONS(3045),
    [anon_sym_getpeername] = ACTIONS(3045),
    [anon_sym_getnetbyname] = ACTIONS(3045),
    [anon_sym_getsockname] = ACTIONS(3045),
    [anon_sym_getgrnam] = ACTIONS(3045),
    [anon_sym_getgrgid] = ACTIONS(3045),
    [anon_sym_hex] = ACTIONS(3045),
    [anon_sym_int] = ACTIONS(3045),
    [anon_sym_keys] = ACTIONS(3045),
    [anon_sym_lc] = ACTIONS(3045),
    [anon_sym_lcfirst] = ACTIONS(3045),
    [anon_sym_length] = ACTIONS(3045),
    [anon_sym_localtime] = ACTIONS(3045),
    [anon_sym_log] = ACTIONS(3045),
    [anon_sym_lock] = ACTIONS(3045),
    [anon_sym_lstat] = ACTIONS(3045),
    [anon_sym_oct] = ACTIONS(3045),
    [anon_sym_ord] = ACTIONS(3045),
    [anon_sym_prototype] = ACTIONS(3045),
    [anon_sym_pop] = ACTIONS(3045),
    [anon_sym_pos] = ACTIONS(3045),
    [anon_sym_quotemeta] = ACTIONS(3045),
    [anon_sym_reset] = ACTIONS(3045),
    [anon_sym_rand] = ACTIONS(3045),
    [anon_sym_rmdir] = ACTIONS(3045),
    [anon_sym_readdir] = ACTIONS(3045),
    [anon_sym_readline] = ACTIONS(3045),
    [anon_sym_readpipe] = ACTIONS(3045),
    [anon_sym_rewinddir] = ACTIONS(3045),
    [anon_sym_readlink] = ACTIONS(3045),
    [anon_sym_ref] = ACTIONS(3045),
    [anon_sym_scalar] = ACTIONS(3045),
    [anon_sym_shift] = ACTIONS(3045),
    [anon_sym_sin] = ACTIONS(3045),
    [anon_sym_sleep] = ACTIONS(3045),
    [anon_sym_sqrt] = ACTIONS(3045),
    [anon_sym_srand] = ACTIONS(3045),
    [anon_sym_stat] = ACTIONS(3045),
    [anon_sym_study] = ACTIONS(3045),
    [anon_sym_tell] = ACTIONS(3045),
    [anon_sym_telldir] = ACTIONS(3045),
    [anon_sym_tied] = ACTIONS(3045),
    [anon_sym_uc] = ACTIONS(3045),
    [anon_sym_ucfirst] = ACTIONS(3045),
    [anon_sym_untie] = ACTIONS(3045),
    [anon_sym_umask] = ACTIONS(3045),
    [anon_sym_values] = ACTIONS(3045),
    [anon_sym_write] = ACTIONS(3045),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3045),
    [anon_sym_qq] = ACTIONS(3045),
    [anon_sym_qw] = ACTIONS(3045),
    [anon_sym_qx] = ACTIONS(3045),
    [anon_sym_qr] = ACTIONS(3045),
    [anon_sym_m] = ACTIONS(3045),
    [anon_sym_s] = ACTIONS(3045),
    [anon_sym_tr] = ACTIONS(3045),
    [anon_sym_y] = ACTIONS(3045),
    [anon_sym_if] = ACTIONS(3045),
    [anon_sym_unless] = ACTIONS(3045),
    [anon_sym_while] = ACTIONS(3045),
    [anon_sym_until] = ACTIONS(3045),
    [aux_sym__bareword_token1] = ACTIONS(3045),
    [sym_number] = ACTIONS(3045),
    [anon_sym_true] = ACTIONS(3045),
    [anon_sym_false] = ACTIONS(3045),
    [sym__single_quote] = ACTIONS(3043),
    [sym__double_quote] = ACTIONS(3043),
    [sym__backtick_quote] = ACTIONS(3043),
    [sym__search_slash_quote] = ACTIONS(3043),
    [sym__open_readline_bracket] = ACTIONS(3043),
    [sym__open_fileglob_bracket] = ACTIONS(3043),
    [sym__PERLY_HEREDOC] = ACTIONS(3043),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3043),
    [sym__filetest] = ACTIONS(3043),
  },
  [STATE(1628)] = {
    [sym_heredoc_content] = STATE(1628),
    [sym__identifier] = ACTIONS(3193),
    [anon_sym_LBRACE] = ACTIONS(3193),
    [anon_sym_RBRACE] = ACTIONS(3193),
    [anon_sym_SEMI] = ACTIONS(3193),
    [anon_sym_package] = ACTIONS(3193),
    [anon_sym_class] = ACTIONS(3193),
    [anon_sym_role] = ACTIONS(3193),
    [anon_sym_BUILD] = ACTIONS(3193),
    [anon_sym_ADJUST] = ACTIONS(3193),
    [anon_sym_DOLLAR] = ACTIONS(3193),
    [anon_sym_AT] = ACTIONS(3193),
    [anon_sym_my] = ACTIONS(3193),
    [anon_sym_extended] = ACTIONS(3193),
    [anon_sym_async] = ACTIONS(3193),
    [anon_sym_sub] = ACTIONS(3193),
    [anon_sym_method] = ACTIONS(3193),
    [anon_sym_LPAREN] = ACTIONS(3193),
    [anon_sym_continue] = ACTIONS(3193),
    [anon_sym_state] = ACTIONS(3193),
    [anon_sym_our] = ACTIONS(3193),
    [anon_sym_try] = ACTIONS(3193),
    [anon_sym_defer] = ACTIONS(3193),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3193),
    [anon_sym_LBRACK] = ACTIONS(3193),
    [anon_sym_LT_LT] = ACTIONS(3193),
    [anon_sym_SLASH_SLASH] = ACTIONS(3193),
    [anon_sym_PLUS] = ACTIONS(3193),
    [anon_sym_DASH] = ACTIONS(3193),
    [anon_sym_TILDE] = ACTIONS(3193),
    [anon_sym_BANG] = ACTIONS(3193),
    [anon_sym_PLUS_PLUS] = ACTIONS(3193),
    [anon_sym_DASH_DASH] = ACTIONS(3193),
    [anon_sym_BSLASH] = ACTIONS(3191),
    [anon_sym_do] = ACTIONS(3193),
    [anon_sym_eval] = ACTIONS(3193),
    [anon_sym_await] = ACTIONS(3193),
    [anon_sym_field] = ACTIONS(3193),
    [anon_sym_local] = ACTIONS(3193),
    [anon_sym_dynamically] = ACTIONS(3193),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3193),
    [anon_sym_require] = ACTIONS(3193),
    [anon_sym_map] = ACTIONS(3193),
    [anon_sym_grep] = ACTIONS(3193),
    [anon_sym_sort] = ACTIONS(3193),
    [anon_sym_goto] = ACTIONS(3193),
    [anon_sym_return] = ACTIONS(3193),
    [anon_sym_undef] = ACTIONS(3193),
    [anon_sym_PERCENT2] = ACTIONS(3193),
    [anon_sym_AMP2] = ACTIONS(3193),
    [anon_sym_STAR2] = ACTIONS(3193),
    [anon_sym_use] = ACTIONS(3193),
    [anon_sym_no] = ACTIONS(3193),
    [anon_sym_for] = ACTIONS(3193),
    [anon_sym_foreach] = ACTIONS(3193),
    [anon_sym_last] = ACTIONS(3193),
    [anon_sym_next] = ACTIONS(3193),
    [anon_sym_redo] = ACTIONS(3193),
    [anon_sym_BEGIN] = ACTIONS(3193),
    [anon_sym_INIT] = ACTIONS(3193),
    [anon_sym_CHECK] = ACTIONS(3193),
    [anon_sym_UNITCHECK] = ACTIONS(3193),
    [anon_sym_END] = ACTIONS(3193),
    [anon_sym___FILE__] = ACTIONS(3193),
    [anon_sym___LINE__] = ACTIONS(3193),
    [anon_sym___PACKAGE__] = ACTIONS(3193),
    [anon_sym___SUB__] = ACTIONS(3193),
    [anon_sym_break] = ACTIONS(3193),
    [anon_sym_fork] = ACTIONS(3193),
    [anon_sym_getppid] = ACTIONS(3193),
    [anon_sym_time] = ACTIONS(3193),
    [anon_sym_times] = ACTIONS(3193),
    [anon_sym_wait] = ACTIONS(3193),
    [anon_sym_wantarray] = ACTIONS(3193),
    [anon_sym_abs] = ACTIONS(3193),
    [anon_sym_alarm] = ACTIONS(3193),
    [anon_sym_chop] = ACTIONS(3193),
    [anon_sym_chdir] = ACTIONS(3193),
    [anon_sym_close] = ACTIONS(3193),
    [anon_sym_closedir] = ACTIONS(3193),
    [anon_sym_caller] = ACTIONS(3193),
    [anon_sym_chomp] = ACTIONS(3193),
    [anon_sym_chr] = ACTIONS(3193),
    [anon_sym_cos] = ACTIONS(3193),
    [anon_sym_chroot] = ACTIONS(3193),
    [anon_sym_defined] = ACTIONS(3193),
    [anon_sym_delete] = ACTIONS(3193),
    [anon_sym_dbmclose] = ACTIONS(3193),
    [anon_sym_exists] = ACTIONS(3193),
    [anon_sym_exit] = ACTIONS(3193),
    [anon_sym_eof] = ACTIONS(3193),
    [anon_sym_exp] = ACTIONS(3193),
    [anon_sym_each] = ACTIONS(3193),
    [anon_sym_fc] = ACTIONS(3193),
    [anon_sym_fileno] = ACTIONS(3193),
    [anon_sym_gmtime] = ACTIONS(3193),
    [anon_sym_getc] = ACTIONS(3193),
    [anon_sym_getpgrp] = ACTIONS(3193),
    [anon_sym_getprotobyname] = ACTIONS(3193),
    [anon_sym_getpwname] = ACTIONS(3193),
    [anon_sym_getpwuid] = ACTIONS(3193),
    [anon_sym_getpeername] = ACTIONS(3193),
    [anon_sym_getnetbyname] = ACTIONS(3193),
    [anon_sym_getsockname] = ACTIONS(3193),
    [anon_sym_getgrnam] = ACTIONS(3193),
    [anon_sym_getgrgid] = ACTIONS(3193),
    [anon_sym_hex] = ACTIONS(3193),
    [anon_sym_int] = ACTIONS(3193),
    [anon_sym_keys] = ACTIONS(3193),
    [anon_sym_lc] = ACTIONS(3193),
    [anon_sym_lcfirst] = ACTIONS(3193),
    [anon_sym_length] = ACTIONS(3193),
    [anon_sym_localtime] = ACTIONS(3193),
    [anon_sym_log] = ACTIONS(3193),
    [anon_sym_lock] = ACTIONS(3193),
    [anon_sym_lstat] = ACTIONS(3193),
    [anon_sym_oct] = ACTIONS(3193),
    [anon_sym_ord] = ACTIONS(3193),
    [anon_sym_prototype] = ACTIONS(3193),
    [anon_sym_pop] = ACTIONS(3193),
    [anon_sym_pos] = ACTIONS(3193),
    [anon_sym_quotemeta] = ACTIONS(3193),
    [anon_sym_reset] = ACTIONS(3193),
    [anon_sym_rand] = ACTIONS(3193),
    [anon_sym_rmdir] = ACTIONS(3193),
    [anon_sym_readdir] = ACTIONS(3193),
    [anon_sym_readline] = ACTIONS(3193),
    [anon_sym_readpipe] = ACTIONS(3193),
    [anon_sym_rewinddir] = ACTIONS(3193),
    [anon_sym_readlink] = ACTIONS(3193),
    [anon_sym_ref] = ACTIONS(3193),
    [anon_sym_scalar] = ACTIONS(3193),
    [anon_sym_shift] = ACTIONS(3193),
    [anon_sym_sin] = ACTIONS(3193),
    [anon_sym_sleep] = ACTIONS(3193),
    [anon_sym_sqrt] = ACTIONS(3193),
    [anon_sym_srand] = ACTIONS(3193),
    [anon_sym_stat] = ACTIONS(3193),
    [anon_sym_study] = ACTIONS(3193),
    [anon_sym_tell] = ACTIONS(3193),
    [anon_sym_telldir] = ACTIONS(3193),
    [anon_sym_tied] = ACTIONS(3193),
    [anon_sym_uc] = ACTIONS(3193),
    [anon_sym_ucfirst] = ACTIONS(3193),
    [anon_sym_untie] = ACTIONS(3193),
    [anon_sym_umask] = ACTIONS(3193),
    [anon_sym_values] = ACTIONS(3193),
    [anon_sym_write] = ACTIONS(3193),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3193),
    [anon_sym_qq] = ACTIONS(3193),
    [anon_sym_qw] = ACTIONS(3193),
    [anon_sym_qx] = ACTIONS(3193),
    [anon_sym_qr] = ACTIONS(3193),
    [anon_sym_m] = ACTIONS(3193),
    [anon_sym_s] = ACTIONS(3193),
    [anon_sym_tr] = ACTIONS(3193),
    [anon_sym_y] = ACTIONS(3193),
    [anon_sym_if] = ACTIONS(3193),
    [anon_sym_unless] = ACTIONS(3193),
    [anon_sym_while] = ACTIONS(3193),
    [anon_sym_until] = ACTIONS(3193),
    [aux_sym__bareword_token1] = ACTIONS(3193),
    [sym_number] = ACTIONS(3193),
    [anon_sym_true] = ACTIONS(3193),
    [anon_sym_false] = ACTIONS(3193),
    [sym__single_quote] = ACTIONS(3191),
    [sym__double_quote] = ACTIONS(3191),
    [sym__backtick_quote] = ACTIONS(3191),
    [sym__search_slash_quote] = ACTIONS(3191),
    [sym__open_readline_bracket] = ACTIONS(3191),
    [sym__open_fileglob_bracket] = ACTIONS(3191),
    [sym__PERLY_HEREDOC] = ACTIONS(3191),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3191),
    [sym__filetest] = ACTIONS(3191),
  },
  [STATE(1629)] = {
    [sym__else] = STATE(1691),
    [sym_else] = STATE(1692),
    [sym_elsif] = STATE(1692),
    [sym_heredoc_content] = STATE(1629),
    [sym__identifier] = ACTIONS(2862),
    [anon_sym_LBRACE] = ACTIONS(2862),
    [anon_sym_DOLLAR] = ACTIONS(2862),
    [anon_sym_AT] = ACTIONS(2862),
    [anon_sym_RPAREN] = ACTIONS(2862),
    [anon_sym_my] = ACTIONS(2862),
    [anon_sym_extended] = ACTIONS(2862),
    [anon_sym_async] = ACTIONS(2862),
    [anon_sym_sub] = ACTIONS(2862),
    [anon_sym_method] = ACTIONS(2862),
    [anon_sym_LPAREN] = ACTIONS(2862),
    [anon_sym_continue] = ACTIONS(2862),
    [anon_sym_state] = ACTIONS(2862),
    [anon_sym_our] = ACTIONS(2862),
    [anon_sym_else] = ACTIONS(3270),
    [anon_sym_elsif] = ACTIONS(3272),
    [anon_sym_LBRACK] = ACTIONS(2862),
    [anon_sym_LT_LT] = ACTIONS(2862),
    [anon_sym_SLASH_SLASH] = ACTIONS(2862),
    [anon_sym_PLUS] = ACTIONS(2862),
    [anon_sym_DASH] = ACTIONS(2862),
    [anon_sym_TILDE] = ACTIONS(2862),
    [anon_sym_BANG] = ACTIONS(2862),
    [anon_sym_PLUS_PLUS] = ACTIONS(2862),
    [anon_sym_DASH_DASH] = ACTIONS(2862),
    [anon_sym_BSLASH] = ACTIONS(2860),
    [anon_sym_do] = ACTIONS(2862),
    [anon_sym_eval] = ACTIONS(2862),
    [anon_sym_await] = ACTIONS(2862),
    [anon_sym_field] = ACTIONS(2862),
    [anon_sym_local] = ACTIONS(2862),
    [anon_sym_dynamically] = ACTIONS(2862),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2862),
    [anon_sym_require] = ACTIONS(2862),
    [anon_sym_map] = ACTIONS(2862),
    [anon_sym_grep] = ACTIONS(2862),
    [anon_sym_sort] = ACTIONS(2862),
    [anon_sym_goto] = ACTIONS(2862),
    [anon_sym_return] = ACTIONS(2862),
    [anon_sym_undef] = ACTIONS(2862),
    [anon_sym_PERCENT2] = ACTIONS(2862),
    [anon_sym_AMP2] = ACTIONS(2862),
    [anon_sym_STAR2] = ACTIONS(2862),
    [anon_sym_last] = ACTIONS(2862),
    [anon_sym_next] = ACTIONS(2862),
    [anon_sym_redo] = ACTIONS(2862),
    [anon_sym___FILE__] = ACTIONS(2862),
    [anon_sym___LINE__] = ACTIONS(2862),
    [anon_sym___PACKAGE__] = ACTIONS(2862),
    [anon_sym___SUB__] = ACTIONS(2862),
    [anon_sym_break] = ACTIONS(2862),
    [anon_sym_fork] = ACTIONS(2862),
    [anon_sym_getppid] = ACTIONS(2862),
    [anon_sym_time] = ACTIONS(2862),
    [anon_sym_times] = ACTIONS(2862),
    [anon_sym_wait] = ACTIONS(2862),
    [anon_sym_wantarray] = ACTIONS(2862),
    [anon_sym_abs] = ACTIONS(2862),
    [anon_sym_alarm] = ACTIONS(2862),
    [anon_sym_chop] = ACTIONS(2862),
    [anon_sym_chdir] = ACTIONS(2862),
    [anon_sym_close] = ACTIONS(2862),
    [anon_sym_closedir] = ACTIONS(2862),
    [anon_sym_caller] = ACTIONS(2862),
    [anon_sym_chomp] = ACTIONS(2862),
    [anon_sym_chr] = ACTIONS(2862),
    [anon_sym_cos] = ACTIONS(2862),
    [anon_sym_chroot] = ACTIONS(2862),
    [anon_sym_defined] = ACTIONS(2862),
    [anon_sym_delete] = ACTIONS(2862),
    [anon_sym_dbmclose] = ACTIONS(2862),
    [anon_sym_exists] = ACTIONS(2862),
    [anon_sym_exit] = ACTIONS(2862),
    [anon_sym_eof] = ACTIONS(2862),
    [anon_sym_exp] = ACTIONS(2862),
    [anon_sym_each] = ACTIONS(2862),
    [anon_sym_fc] = ACTIONS(2862),
    [anon_sym_fileno] = ACTIONS(2862),
    [anon_sym_gmtime] = ACTIONS(2862),
    [anon_sym_getc] = ACTIONS(2862),
    [anon_sym_getpgrp] = ACTIONS(2862),
    [anon_sym_getprotobyname] = ACTIONS(2862),
    [anon_sym_getpwname] = ACTIONS(2862),
    [anon_sym_getpwuid] = ACTIONS(2862),
    [anon_sym_getpeername] = ACTIONS(2862),
    [anon_sym_getnetbyname] = ACTIONS(2862),
    [anon_sym_getsockname] = ACTIONS(2862),
    [anon_sym_getgrnam] = ACTIONS(2862),
    [anon_sym_getgrgid] = ACTIONS(2862),
    [anon_sym_hex] = ACTIONS(2862),
    [anon_sym_int] = ACTIONS(2862),
    [anon_sym_keys] = ACTIONS(2862),
    [anon_sym_lc] = ACTIONS(2862),
    [anon_sym_lcfirst] = ACTIONS(2862),
    [anon_sym_length] = ACTIONS(2862),
    [anon_sym_localtime] = ACTIONS(2862),
    [anon_sym_log] = ACTIONS(2862),
    [anon_sym_lock] = ACTIONS(2862),
    [anon_sym_lstat] = ACTIONS(2862),
    [anon_sym_oct] = ACTIONS(2862),
    [anon_sym_ord] = ACTIONS(2862),
    [anon_sym_prototype] = ACTIONS(2862),
    [anon_sym_pop] = ACTIONS(2862),
    [anon_sym_pos] = ACTIONS(2862),
    [anon_sym_quotemeta] = ACTIONS(2862),
    [anon_sym_reset] = ACTIONS(2862),
    [anon_sym_rand] = ACTIONS(2862),
    [anon_sym_rmdir] = ACTIONS(2862),
    [anon_sym_readdir] = ACTIONS(2862),
    [anon_sym_readline] = ACTIONS(2862),
    [anon_sym_readpipe] = ACTIONS(2862),
    [anon_sym_rewinddir] = ACTIONS(2862),
    [anon_sym_readlink] = ACTIONS(2862),
    [anon_sym_ref] = ACTIONS(2862),
    [anon_sym_scalar] = ACTIONS(2862),
    [anon_sym_shift] = ACTIONS(2862),
    [anon_sym_sin] = ACTIONS(2862),
    [anon_sym_sleep] = ACTIONS(2862),
    [anon_sym_sqrt] = ACTIONS(2862),
    [anon_sym_srand] = ACTIONS(2862),
    [anon_sym_stat] = ACTIONS(2862),
    [anon_sym_study] = ACTIONS(2862),
    [anon_sym_tell] = ACTIONS(2862),
    [anon_sym_telldir] = ACTIONS(2862),
    [anon_sym_tied] = ACTIONS(2862),
    [anon_sym_uc] = ACTIONS(2862),
    [anon_sym_ucfirst] = ACTIONS(2862),
    [anon_sym_untie] = ACTIONS(2862),
    [anon_sym_umask] = ACTIONS(2862),
    [anon_sym_values] = ACTIONS(2862),
    [anon_sym_write] = ACTIONS(2862),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2862),
    [anon_sym_qq] = ACTIONS(2862),
    [anon_sym_qw] = ACTIONS(2862),
    [anon_sym_qx] = ACTIONS(2862),
    [anon_sym_qr] = ACTIONS(2862),
    [anon_sym_m] = ACTIONS(2862),
    [anon_sym_s] = ACTIONS(2862),
    [anon_sym_tr] = ACTIONS(2862),
    [anon_sym_y] = ACTIONS(2862),
    [aux_sym__bareword_token1] = ACTIONS(2862),
    [sym_number] = ACTIONS(2862),
    [anon_sym_true] = ACTIONS(2862),
    [anon_sym_false] = ACTIONS(2862),
    [sym__single_quote] = ACTIONS(2860),
    [sym__double_quote] = ACTIONS(2860),
    [sym__backtick_quote] = ACTIONS(2860),
    [sym__search_slash_quote] = ACTIONS(2860),
    [sym__open_readline_bracket] = ACTIONS(2860),
    [sym__open_fileglob_bracket] = ACTIONS(2860),
    [sym__PERLY_HEREDOC] = ACTIONS(2860),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2860),
    [sym__filetest] = ACTIONS(2860),
  },
  [STATE(1630)] = {
    [sym__else] = STATE(1711),
    [sym_else] = STATE(1692),
    [sym_elsif] = STATE(1692),
    [sym_heredoc_content] = STATE(1630),
    [sym__identifier] = ACTIONS(2870),
    [anon_sym_LBRACE] = ACTIONS(2870),
    [anon_sym_DOLLAR] = ACTIONS(2870),
    [anon_sym_AT] = ACTIONS(2870),
    [anon_sym_RPAREN] = ACTIONS(2870),
    [anon_sym_my] = ACTIONS(2870),
    [anon_sym_extended] = ACTIONS(2870),
    [anon_sym_async] = ACTIONS(2870),
    [anon_sym_sub] = ACTIONS(2870),
    [anon_sym_method] = ACTIONS(2870),
    [anon_sym_LPAREN] = ACTIONS(2870),
    [anon_sym_continue] = ACTIONS(2870),
    [anon_sym_state] = ACTIONS(2870),
    [anon_sym_our] = ACTIONS(2870),
    [anon_sym_else] = ACTIONS(3270),
    [anon_sym_elsif] = ACTIONS(3272),
    [anon_sym_LBRACK] = ACTIONS(2870),
    [anon_sym_LT_LT] = ACTIONS(2870),
    [anon_sym_SLASH_SLASH] = ACTIONS(2870),
    [anon_sym_PLUS] = ACTIONS(2870),
    [anon_sym_DASH] = ACTIONS(2870),
    [anon_sym_TILDE] = ACTIONS(2870),
    [anon_sym_BANG] = ACTIONS(2870),
    [anon_sym_PLUS_PLUS] = ACTIONS(2870),
    [anon_sym_DASH_DASH] = ACTIONS(2870),
    [anon_sym_BSLASH] = ACTIONS(2868),
    [anon_sym_do] = ACTIONS(2870),
    [anon_sym_eval] = ACTIONS(2870),
    [anon_sym_await] = ACTIONS(2870),
    [anon_sym_field] = ACTIONS(2870),
    [anon_sym_local] = ACTIONS(2870),
    [anon_sym_dynamically] = ACTIONS(2870),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2870),
    [anon_sym_require] = ACTIONS(2870),
    [anon_sym_map] = ACTIONS(2870),
    [anon_sym_grep] = ACTIONS(2870),
    [anon_sym_sort] = ACTIONS(2870),
    [anon_sym_goto] = ACTIONS(2870),
    [anon_sym_return] = ACTIONS(2870),
    [anon_sym_undef] = ACTIONS(2870),
    [anon_sym_PERCENT2] = ACTIONS(2870),
    [anon_sym_AMP2] = ACTIONS(2870),
    [anon_sym_STAR2] = ACTIONS(2870),
    [anon_sym_last] = ACTIONS(2870),
    [anon_sym_next] = ACTIONS(2870),
    [anon_sym_redo] = ACTIONS(2870),
    [anon_sym___FILE__] = ACTIONS(2870),
    [anon_sym___LINE__] = ACTIONS(2870),
    [anon_sym___PACKAGE__] = ACTIONS(2870),
    [anon_sym___SUB__] = ACTIONS(2870),
    [anon_sym_break] = ACTIONS(2870),
    [anon_sym_fork] = ACTIONS(2870),
    [anon_sym_getppid] = ACTIONS(2870),
    [anon_sym_time] = ACTIONS(2870),
    [anon_sym_times] = ACTIONS(2870),
    [anon_sym_wait] = ACTIONS(2870),
    [anon_sym_wantarray] = ACTIONS(2870),
    [anon_sym_abs] = ACTIONS(2870),
    [anon_sym_alarm] = ACTIONS(2870),
    [anon_sym_chop] = ACTIONS(2870),
    [anon_sym_chdir] = ACTIONS(2870),
    [anon_sym_close] = ACTIONS(2870),
    [anon_sym_closedir] = ACTIONS(2870),
    [anon_sym_caller] = ACTIONS(2870),
    [anon_sym_chomp] = ACTIONS(2870),
    [anon_sym_chr] = ACTIONS(2870),
    [anon_sym_cos] = ACTIONS(2870),
    [anon_sym_chroot] = ACTIONS(2870),
    [anon_sym_defined] = ACTIONS(2870),
    [anon_sym_delete] = ACTIONS(2870),
    [anon_sym_dbmclose] = ACTIONS(2870),
    [anon_sym_exists] = ACTIONS(2870),
    [anon_sym_exit] = ACTIONS(2870),
    [anon_sym_eof] = ACTIONS(2870),
    [anon_sym_exp] = ACTIONS(2870),
    [anon_sym_each] = ACTIONS(2870),
    [anon_sym_fc] = ACTIONS(2870),
    [anon_sym_fileno] = ACTIONS(2870),
    [anon_sym_gmtime] = ACTIONS(2870),
    [anon_sym_getc] = ACTIONS(2870),
    [anon_sym_getpgrp] = ACTIONS(2870),
    [anon_sym_getprotobyname] = ACTIONS(2870),
    [anon_sym_getpwname] = ACTIONS(2870),
    [anon_sym_getpwuid] = ACTIONS(2870),
    [anon_sym_getpeername] = ACTIONS(2870),
    [anon_sym_getnetbyname] = ACTIONS(2870),
    [anon_sym_getsockname] = ACTIONS(2870),
    [anon_sym_getgrnam] = ACTIONS(2870),
    [anon_sym_getgrgid] = ACTIONS(2870),
    [anon_sym_hex] = ACTIONS(2870),
    [anon_sym_int] = ACTIONS(2870),
    [anon_sym_keys] = ACTIONS(2870),
    [anon_sym_lc] = ACTIONS(2870),
    [anon_sym_lcfirst] = ACTIONS(2870),
    [anon_sym_length] = ACTIONS(2870),
    [anon_sym_localtime] = ACTIONS(2870),
    [anon_sym_log] = ACTIONS(2870),
    [anon_sym_lock] = ACTIONS(2870),
    [anon_sym_lstat] = ACTIONS(2870),
    [anon_sym_oct] = ACTIONS(2870),
    [anon_sym_ord] = ACTIONS(2870),
    [anon_sym_prototype] = ACTIONS(2870),
    [anon_sym_pop] = ACTIONS(2870),
    [anon_sym_pos] = ACTIONS(2870),
    [anon_sym_quotemeta] = ACTIONS(2870),
    [anon_sym_reset] = ACTIONS(2870),
    [anon_sym_rand] = ACTIONS(2870),
    [anon_sym_rmdir] = ACTIONS(2870),
    [anon_sym_readdir] = ACTIONS(2870),
    [anon_sym_readline] = ACTIONS(2870),
    [anon_sym_readpipe] = ACTIONS(2870),
    [anon_sym_rewinddir] = ACTIONS(2870),
    [anon_sym_readlink] = ACTIONS(2870),
    [anon_sym_ref] = ACTIONS(2870),
    [anon_sym_scalar] = ACTIONS(2870),
    [anon_sym_shift] = ACTIONS(2870),
    [anon_sym_sin] = ACTIONS(2870),
    [anon_sym_sleep] = ACTIONS(2870),
    [anon_sym_sqrt] = ACTIONS(2870),
    [anon_sym_srand] = ACTIONS(2870),
    [anon_sym_stat] = ACTIONS(2870),
    [anon_sym_study] = ACTIONS(2870),
    [anon_sym_tell] = ACTIONS(2870),
    [anon_sym_telldir] = ACTIONS(2870),
    [anon_sym_tied] = ACTIONS(2870),
    [anon_sym_uc] = ACTIONS(2870),
    [anon_sym_ucfirst] = ACTIONS(2870),
    [anon_sym_untie] = ACTIONS(2870),
    [anon_sym_umask] = ACTIONS(2870),
    [anon_sym_values] = ACTIONS(2870),
    [anon_sym_write] = ACTIONS(2870),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2870),
    [anon_sym_qq] = ACTIONS(2870),
    [anon_sym_qw] = ACTIONS(2870),
    [anon_sym_qx] = ACTIONS(2870),
    [anon_sym_qr] = ACTIONS(2870),
    [anon_sym_m] = ACTIONS(2870),
    [anon_sym_s] = ACTIONS(2870),
    [anon_sym_tr] = ACTIONS(2870),
    [anon_sym_y] = ACTIONS(2870),
    [aux_sym__bareword_token1] = ACTIONS(2870),
    [sym_number] = ACTIONS(2870),
    [anon_sym_true] = ACTIONS(2870),
    [anon_sym_false] = ACTIONS(2870),
    [sym__single_quote] = ACTIONS(2868),
    [sym__double_quote] = ACTIONS(2868),
    [sym__backtick_quote] = ACTIONS(2868),
    [sym__search_slash_quote] = ACTIONS(2868),
    [sym__open_readline_bracket] = ACTIONS(2868),
    [sym__open_fileglob_bracket] = ACTIONS(2868),
    [sym__PERLY_HEREDOC] = ACTIONS(2868),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2868),
    [sym__filetest] = ACTIONS(2868),
  },
  [STATE(1631)] = {
    [sym_heredoc_content] = STATE(1631),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_catch] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1632)] = {
    [sym_heredoc_content] = STATE(1632),
    [sym__identifier] = ACTIONS(3274),
    [anon_sym_LBRACE] = ACTIONS(3274),
    [anon_sym_SEMI] = ACTIONS(3274),
    [anon_sym_DOLLAR] = ACTIONS(3274),
    [anon_sym_AT] = ACTIONS(3274),
    [anon_sym_my] = ACTIONS(3274),
    [anon_sym_extended] = ACTIONS(3274),
    [anon_sym_async] = ACTIONS(3274),
    [anon_sym_sub] = ACTIONS(3274),
    [anon_sym_method] = ACTIONS(3274),
    [anon_sym_LPAREN] = ACTIONS(3274),
    [anon_sym_continue] = ACTIONS(3274),
    [anon_sym_state] = ACTIONS(3274),
    [anon_sym_our] = ACTIONS(3274),
    [anon_sym_LBRACK] = ACTIONS(3274),
    [anon_sym_LT_LT] = ACTIONS(3274),
    [anon_sym_SLASH_SLASH] = ACTIONS(3274),
    [anon_sym_PLUS] = ACTIONS(3274),
    [anon_sym_DASH] = ACTIONS(3274),
    [anon_sym_TILDE] = ACTIONS(3274),
    [anon_sym_BANG] = ACTIONS(3274),
    [anon_sym_PLUS_PLUS] = ACTIONS(3274),
    [anon_sym_DASH_DASH] = ACTIONS(3274),
    [anon_sym_BSLASH] = ACTIONS(3276),
    [anon_sym_do] = ACTIONS(3274),
    [anon_sym_eval] = ACTIONS(3274),
    [anon_sym_await] = ACTIONS(3274),
    [anon_sym_field] = ACTIONS(3274),
    [anon_sym_local] = ACTIONS(3274),
    [anon_sym_dynamically] = ACTIONS(3274),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3274),
    [anon_sym_require] = ACTIONS(3274),
    [anon_sym_map] = ACTIONS(3274),
    [anon_sym_grep] = ACTIONS(3274),
    [anon_sym_sort] = ACTIONS(3274),
    [anon_sym_goto] = ACTIONS(3274),
    [anon_sym_return] = ACTIONS(3274),
    [anon_sym_undef] = ACTIONS(3274),
    [anon_sym_PERCENT2] = ACTIONS(3274),
    [anon_sym_AMP2] = ACTIONS(3274),
    [anon_sym_STAR2] = ACTIONS(3274),
    [anon_sym_last] = ACTIONS(3274),
    [anon_sym_next] = ACTIONS(3274),
    [anon_sym_redo] = ACTIONS(3274),
    [anon_sym___FILE__] = ACTIONS(3274),
    [anon_sym___LINE__] = ACTIONS(3274),
    [anon_sym___PACKAGE__] = ACTIONS(3274),
    [anon_sym___SUB__] = ACTIONS(3274),
    [anon_sym_break] = ACTIONS(3274),
    [anon_sym_fork] = ACTIONS(3274),
    [anon_sym_getppid] = ACTIONS(3274),
    [anon_sym_time] = ACTIONS(3274),
    [anon_sym_times] = ACTIONS(3274),
    [anon_sym_wait] = ACTIONS(3274),
    [anon_sym_wantarray] = ACTIONS(3274),
    [anon_sym_abs] = ACTIONS(3274),
    [anon_sym_alarm] = ACTIONS(3274),
    [anon_sym_chop] = ACTIONS(3274),
    [anon_sym_chdir] = ACTIONS(3274),
    [anon_sym_close] = ACTIONS(3274),
    [anon_sym_closedir] = ACTIONS(3274),
    [anon_sym_caller] = ACTIONS(3274),
    [anon_sym_chomp] = ACTIONS(3274),
    [anon_sym_chr] = ACTIONS(3274),
    [anon_sym_cos] = ACTIONS(3274),
    [anon_sym_chroot] = ACTIONS(3274),
    [anon_sym_defined] = ACTIONS(3274),
    [anon_sym_delete] = ACTIONS(3274),
    [anon_sym_dbmclose] = ACTIONS(3274),
    [anon_sym_exists] = ACTIONS(3274),
    [anon_sym_exit] = ACTIONS(3274),
    [anon_sym_eof] = ACTIONS(3274),
    [anon_sym_exp] = ACTIONS(3274),
    [anon_sym_each] = ACTIONS(3274),
    [anon_sym_fc] = ACTIONS(3274),
    [anon_sym_fileno] = ACTIONS(3274),
    [anon_sym_gmtime] = ACTIONS(3274),
    [anon_sym_getc] = ACTIONS(3274),
    [anon_sym_getpgrp] = ACTIONS(3274),
    [anon_sym_getprotobyname] = ACTIONS(3274),
    [anon_sym_getpwname] = ACTIONS(3274),
    [anon_sym_getpwuid] = ACTIONS(3274),
    [anon_sym_getpeername] = ACTIONS(3274),
    [anon_sym_getnetbyname] = ACTIONS(3274),
    [anon_sym_getsockname] = ACTIONS(3274),
    [anon_sym_getgrnam] = ACTIONS(3274),
    [anon_sym_getgrgid] = ACTIONS(3274),
    [anon_sym_hex] = ACTIONS(3274),
    [anon_sym_int] = ACTIONS(3274),
    [anon_sym_keys] = ACTIONS(3274),
    [anon_sym_lc] = ACTIONS(3274),
    [anon_sym_lcfirst] = ACTIONS(3274),
    [anon_sym_length] = ACTIONS(3274),
    [anon_sym_localtime] = ACTIONS(3274),
    [anon_sym_log] = ACTIONS(3274),
    [anon_sym_lock] = ACTIONS(3274),
    [anon_sym_lstat] = ACTIONS(3274),
    [anon_sym_oct] = ACTIONS(3274),
    [anon_sym_ord] = ACTIONS(3274),
    [anon_sym_prototype] = ACTIONS(3274),
    [anon_sym_pop] = ACTIONS(3274),
    [anon_sym_pos] = ACTIONS(3274),
    [anon_sym_quotemeta] = ACTIONS(3274),
    [anon_sym_reset] = ACTIONS(3274),
    [anon_sym_rand] = ACTIONS(3274),
    [anon_sym_rmdir] = ACTIONS(3274),
    [anon_sym_readdir] = ACTIONS(3274),
    [anon_sym_readline] = ACTIONS(3274),
    [anon_sym_readpipe] = ACTIONS(3274),
    [anon_sym_rewinddir] = ACTIONS(3274),
    [anon_sym_readlink] = ACTIONS(3274),
    [anon_sym_ref] = ACTIONS(3274),
    [anon_sym_scalar] = ACTIONS(3274),
    [anon_sym_shift] = ACTIONS(3274),
    [anon_sym_sin] = ACTIONS(3274),
    [anon_sym_sleep] = ACTIONS(3274),
    [anon_sym_sqrt] = ACTIONS(3274),
    [anon_sym_srand] = ACTIONS(3274),
    [anon_sym_stat] = ACTIONS(3274),
    [anon_sym_study] = ACTIONS(3274),
    [anon_sym_tell] = ACTIONS(3274),
    [anon_sym_telldir] = ACTIONS(3274),
    [anon_sym_tied] = ACTIONS(3274),
    [anon_sym_uc] = ACTIONS(3274),
    [anon_sym_ucfirst] = ACTIONS(3274),
    [anon_sym_untie] = ACTIONS(3274),
    [anon_sym_umask] = ACTIONS(3274),
    [anon_sym_values] = ACTIONS(3274),
    [anon_sym_write] = ACTIONS(3274),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3274),
    [anon_sym_qq] = ACTIONS(3274),
    [anon_sym_qw] = ACTIONS(3274),
    [anon_sym_qx] = ACTIONS(3274),
    [anon_sym_qr] = ACTIONS(3274),
    [anon_sym_m] = ACTIONS(3274),
    [anon_sym_s] = ACTIONS(3274),
    [anon_sym_tr] = ACTIONS(3274),
    [anon_sym_y] = ACTIONS(3274),
    [sym_version] = ACTIONS(3274),
    [aux_sym__bareword_token1] = ACTIONS(3274),
    [sym_number] = ACTIONS(3274),
    [anon_sym_true] = ACTIONS(3274),
    [anon_sym_false] = ACTIONS(3274),
    [sym__single_quote] = ACTIONS(3276),
    [sym__double_quote] = ACTIONS(3276),
    [sym__backtick_quote] = ACTIONS(3276),
    [sym__search_slash_quote] = ACTIONS(3276),
    [sym__open_readline_bracket] = ACTIONS(3276),
    [sym__open_fileglob_bracket] = ACTIONS(3276),
    [sym__PERLY_SEMICOLON] = ACTIONS(3276),
    [sym__PERLY_HEREDOC] = ACTIONS(3276),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3276),
    [sym__filetest] = ACTIONS(3276),
  },
  [STATE(1633)] = {
    [sym_heredoc_content] = STATE(1633),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_else] = ACTIONS(2532),
    [anon_sym_elsif] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1634)] = {
    [sym_heredoc_content] = STATE(1634),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_catch] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1635)] = {
    [sym_heredoc_content] = STATE(1635),
    [sym__identifier] = ACTIONS(2874),
    [anon_sym_LBRACE] = ACTIONS(2874),
    [anon_sym_DOLLAR] = ACTIONS(2874),
    [anon_sym_AT] = ACTIONS(2874),
    [anon_sym_RPAREN] = ACTIONS(2874),
    [anon_sym_my] = ACTIONS(2874),
    [anon_sym_extended] = ACTIONS(2874),
    [anon_sym_async] = ACTIONS(2874),
    [anon_sym_sub] = ACTIONS(2874),
    [anon_sym_method] = ACTIONS(2874),
    [anon_sym_LPAREN] = ACTIONS(2874),
    [anon_sym_continue] = ACTIONS(2874),
    [anon_sym_state] = ACTIONS(2874),
    [anon_sym_our] = ACTIONS(2874),
    [anon_sym_catch] = ACTIONS(3278),
    [anon_sym_finally] = ACTIONS(3280),
    [anon_sym_LBRACK] = ACTIONS(2874),
    [anon_sym_LT_LT] = ACTIONS(2874),
    [anon_sym_SLASH_SLASH] = ACTIONS(2874),
    [anon_sym_PLUS] = ACTIONS(2874),
    [anon_sym_DASH] = ACTIONS(2874),
    [anon_sym_TILDE] = ACTIONS(2874),
    [anon_sym_BANG] = ACTIONS(2874),
    [anon_sym_PLUS_PLUS] = ACTIONS(2874),
    [anon_sym_DASH_DASH] = ACTIONS(2874),
    [anon_sym_BSLASH] = ACTIONS(2872),
    [anon_sym_do] = ACTIONS(2874),
    [anon_sym_eval] = ACTIONS(2874),
    [anon_sym_await] = ACTIONS(2874),
    [anon_sym_field] = ACTIONS(2874),
    [anon_sym_local] = ACTIONS(2874),
    [anon_sym_dynamically] = ACTIONS(2874),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2874),
    [anon_sym_require] = ACTIONS(2874),
    [anon_sym_map] = ACTIONS(2874),
    [anon_sym_grep] = ACTIONS(2874),
    [anon_sym_sort] = ACTIONS(2874),
    [anon_sym_goto] = ACTIONS(2874),
    [anon_sym_return] = ACTIONS(2874),
    [anon_sym_undef] = ACTIONS(2874),
    [anon_sym_PERCENT2] = ACTIONS(2874),
    [anon_sym_AMP2] = ACTIONS(2874),
    [anon_sym_STAR2] = ACTIONS(2874),
    [anon_sym_last] = ACTIONS(2874),
    [anon_sym_next] = ACTIONS(2874),
    [anon_sym_redo] = ACTIONS(2874),
    [anon_sym___FILE__] = ACTIONS(2874),
    [anon_sym___LINE__] = ACTIONS(2874),
    [anon_sym___PACKAGE__] = ACTIONS(2874),
    [anon_sym___SUB__] = ACTIONS(2874),
    [anon_sym_break] = ACTIONS(2874),
    [anon_sym_fork] = ACTIONS(2874),
    [anon_sym_getppid] = ACTIONS(2874),
    [anon_sym_time] = ACTIONS(2874),
    [anon_sym_times] = ACTIONS(2874),
    [anon_sym_wait] = ACTIONS(2874),
    [anon_sym_wantarray] = ACTIONS(2874),
    [anon_sym_abs] = ACTIONS(2874),
    [anon_sym_alarm] = ACTIONS(2874),
    [anon_sym_chop] = ACTIONS(2874),
    [anon_sym_chdir] = ACTIONS(2874),
    [anon_sym_close] = ACTIONS(2874),
    [anon_sym_closedir] = ACTIONS(2874),
    [anon_sym_caller] = ACTIONS(2874),
    [anon_sym_chomp] = ACTIONS(2874),
    [anon_sym_chr] = ACTIONS(2874),
    [anon_sym_cos] = ACTIONS(2874),
    [anon_sym_chroot] = ACTIONS(2874),
    [anon_sym_defined] = ACTIONS(2874),
    [anon_sym_delete] = ACTIONS(2874),
    [anon_sym_dbmclose] = ACTIONS(2874),
    [anon_sym_exists] = ACTIONS(2874),
    [anon_sym_exit] = ACTIONS(2874),
    [anon_sym_eof] = ACTIONS(2874),
    [anon_sym_exp] = ACTIONS(2874),
    [anon_sym_each] = ACTIONS(2874),
    [anon_sym_fc] = ACTIONS(2874),
    [anon_sym_fileno] = ACTIONS(2874),
    [anon_sym_gmtime] = ACTIONS(2874),
    [anon_sym_getc] = ACTIONS(2874),
    [anon_sym_getpgrp] = ACTIONS(2874),
    [anon_sym_getprotobyname] = ACTIONS(2874),
    [anon_sym_getpwname] = ACTIONS(2874),
    [anon_sym_getpwuid] = ACTIONS(2874),
    [anon_sym_getpeername] = ACTIONS(2874),
    [anon_sym_getnetbyname] = ACTIONS(2874),
    [anon_sym_getsockname] = ACTIONS(2874),
    [anon_sym_getgrnam] = ACTIONS(2874),
    [anon_sym_getgrgid] = ACTIONS(2874),
    [anon_sym_hex] = ACTIONS(2874),
    [anon_sym_int] = ACTIONS(2874),
    [anon_sym_keys] = ACTIONS(2874),
    [anon_sym_lc] = ACTIONS(2874),
    [anon_sym_lcfirst] = ACTIONS(2874),
    [anon_sym_length] = ACTIONS(2874),
    [anon_sym_localtime] = ACTIONS(2874),
    [anon_sym_log] = ACTIONS(2874),
    [anon_sym_lock] = ACTIONS(2874),
    [anon_sym_lstat] = ACTIONS(2874),
    [anon_sym_oct] = ACTIONS(2874),
    [anon_sym_ord] = ACTIONS(2874),
    [anon_sym_prototype] = ACTIONS(2874),
    [anon_sym_pop] = ACTIONS(2874),
    [anon_sym_pos] = ACTIONS(2874),
    [anon_sym_quotemeta] = ACTIONS(2874),
    [anon_sym_reset] = ACTIONS(2874),
    [anon_sym_rand] = ACTIONS(2874),
    [anon_sym_rmdir] = ACTIONS(2874),
    [anon_sym_readdir] = ACTIONS(2874),
    [anon_sym_readline] = ACTIONS(2874),
    [anon_sym_readpipe] = ACTIONS(2874),
    [anon_sym_rewinddir] = ACTIONS(2874),
    [anon_sym_readlink] = ACTIONS(2874),
    [anon_sym_ref] = ACTIONS(2874),
    [anon_sym_scalar] = ACTIONS(2874),
    [anon_sym_shift] = ACTIONS(2874),
    [anon_sym_sin] = ACTIONS(2874),
    [anon_sym_sleep] = ACTIONS(2874),
    [anon_sym_sqrt] = ACTIONS(2874),
    [anon_sym_srand] = ACTIONS(2874),
    [anon_sym_stat] = ACTIONS(2874),
    [anon_sym_study] = ACTIONS(2874),
    [anon_sym_tell] = ACTIONS(2874),
    [anon_sym_telldir] = ACTIONS(2874),
    [anon_sym_tied] = ACTIONS(2874),
    [anon_sym_uc] = ACTIONS(2874),
    [anon_sym_ucfirst] = ACTIONS(2874),
    [anon_sym_untie] = ACTIONS(2874),
    [anon_sym_umask] = ACTIONS(2874),
    [anon_sym_values] = ACTIONS(2874),
    [anon_sym_write] = ACTIONS(2874),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2874),
    [anon_sym_qq] = ACTIONS(2874),
    [anon_sym_qw] = ACTIONS(2874),
    [anon_sym_qx] = ACTIONS(2874),
    [anon_sym_qr] = ACTIONS(2874),
    [anon_sym_m] = ACTIONS(2874),
    [anon_sym_s] = ACTIONS(2874),
    [anon_sym_tr] = ACTIONS(2874),
    [anon_sym_y] = ACTIONS(2874),
    [aux_sym__bareword_token1] = ACTIONS(2874),
    [sym_number] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2874),
    [anon_sym_false] = ACTIONS(2874),
    [sym__single_quote] = ACTIONS(2872),
    [sym__double_quote] = ACTIONS(2872),
    [sym__backtick_quote] = ACTIONS(2872),
    [sym__search_slash_quote] = ACTIONS(2872),
    [sym__open_readline_bracket] = ACTIONS(2872),
    [sym__open_fileglob_bracket] = ACTIONS(2872),
    [sym__PERLY_HEREDOC] = ACTIONS(2872),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2872),
    [sym__filetest] = ACTIONS(2872),
  },
  [STATE(1636)] = {
    [sym_heredoc_content] = STATE(1636),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_else] = ACTIONS(2688),
    [anon_sym_elsif] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1637)] = {
    [sym_heredoc_content] = STATE(1637),
    [sym__identifier] = ACTIONS(2892),
    [anon_sym_LBRACE] = ACTIONS(2892),
    [anon_sym_DOLLAR] = ACTIONS(2892),
    [anon_sym_AT] = ACTIONS(2892),
    [anon_sym_RPAREN] = ACTIONS(2892),
    [anon_sym_my] = ACTIONS(2892),
    [anon_sym_extended] = ACTIONS(2892),
    [anon_sym_async] = ACTIONS(2892),
    [anon_sym_sub] = ACTIONS(2892),
    [anon_sym_method] = ACTIONS(2892),
    [anon_sym_LPAREN] = ACTIONS(2892),
    [anon_sym_continue] = ACTIONS(2892),
    [anon_sym_state] = ACTIONS(2892),
    [anon_sym_our] = ACTIONS(2892),
    [anon_sym_finally] = ACTIONS(3282),
    [anon_sym_LBRACK] = ACTIONS(2892),
    [anon_sym_LT_LT] = ACTIONS(2892),
    [anon_sym_SLASH_SLASH] = ACTIONS(2892),
    [anon_sym_PLUS] = ACTIONS(2892),
    [anon_sym_DASH] = ACTIONS(2892),
    [anon_sym_TILDE] = ACTIONS(2892),
    [anon_sym_BANG] = ACTIONS(2892),
    [anon_sym_PLUS_PLUS] = ACTIONS(2892),
    [anon_sym_DASH_DASH] = ACTIONS(2892),
    [anon_sym_BSLASH] = ACTIONS(2890),
    [anon_sym_do] = ACTIONS(2892),
    [anon_sym_eval] = ACTIONS(2892),
    [anon_sym_await] = ACTIONS(2892),
    [anon_sym_field] = ACTIONS(2892),
    [anon_sym_local] = ACTIONS(2892),
    [anon_sym_dynamically] = ACTIONS(2892),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2892),
    [anon_sym_require] = ACTIONS(2892),
    [anon_sym_map] = ACTIONS(2892),
    [anon_sym_grep] = ACTIONS(2892),
    [anon_sym_sort] = ACTIONS(2892),
    [anon_sym_goto] = ACTIONS(2892),
    [anon_sym_return] = ACTIONS(2892),
    [anon_sym_undef] = ACTIONS(2892),
    [anon_sym_PERCENT2] = ACTIONS(2892),
    [anon_sym_AMP2] = ACTIONS(2892),
    [anon_sym_STAR2] = ACTIONS(2892),
    [anon_sym_last] = ACTIONS(2892),
    [anon_sym_next] = ACTIONS(2892),
    [anon_sym_redo] = ACTIONS(2892),
    [anon_sym___FILE__] = ACTIONS(2892),
    [anon_sym___LINE__] = ACTIONS(2892),
    [anon_sym___PACKAGE__] = ACTIONS(2892),
    [anon_sym___SUB__] = ACTIONS(2892),
    [anon_sym_break] = ACTIONS(2892),
    [anon_sym_fork] = ACTIONS(2892),
    [anon_sym_getppid] = ACTIONS(2892),
    [anon_sym_time] = ACTIONS(2892),
    [anon_sym_times] = ACTIONS(2892),
    [anon_sym_wait] = ACTIONS(2892),
    [anon_sym_wantarray] = ACTIONS(2892),
    [anon_sym_abs] = ACTIONS(2892),
    [anon_sym_alarm] = ACTIONS(2892),
    [anon_sym_chop] = ACTIONS(2892),
    [anon_sym_chdir] = ACTIONS(2892),
    [anon_sym_close] = ACTIONS(2892),
    [anon_sym_closedir] = ACTIONS(2892),
    [anon_sym_caller] = ACTIONS(2892),
    [anon_sym_chomp] = ACTIONS(2892),
    [anon_sym_chr] = ACTIONS(2892),
    [anon_sym_cos] = ACTIONS(2892),
    [anon_sym_chroot] = ACTIONS(2892),
    [anon_sym_defined] = ACTIONS(2892),
    [anon_sym_delete] = ACTIONS(2892),
    [anon_sym_dbmclose] = ACTIONS(2892),
    [anon_sym_exists] = ACTIONS(2892),
    [anon_sym_exit] = ACTIONS(2892),
    [anon_sym_eof] = ACTIONS(2892),
    [anon_sym_exp] = ACTIONS(2892),
    [anon_sym_each] = ACTIONS(2892),
    [anon_sym_fc] = ACTIONS(2892),
    [anon_sym_fileno] = ACTIONS(2892),
    [anon_sym_gmtime] = ACTIONS(2892),
    [anon_sym_getc] = ACTIONS(2892),
    [anon_sym_getpgrp] = ACTIONS(2892),
    [anon_sym_getprotobyname] = ACTIONS(2892),
    [anon_sym_getpwname] = ACTIONS(2892),
    [anon_sym_getpwuid] = ACTIONS(2892),
    [anon_sym_getpeername] = ACTIONS(2892),
    [anon_sym_getnetbyname] = ACTIONS(2892),
    [anon_sym_getsockname] = ACTIONS(2892),
    [anon_sym_getgrnam] = ACTIONS(2892),
    [anon_sym_getgrgid] = ACTIONS(2892),
    [anon_sym_hex] = ACTIONS(2892),
    [anon_sym_int] = ACTIONS(2892),
    [anon_sym_keys] = ACTIONS(2892),
    [anon_sym_lc] = ACTIONS(2892),
    [anon_sym_lcfirst] = ACTIONS(2892),
    [anon_sym_length] = ACTIONS(2892),
    [anon_sym_localtime] = ACTIONS(2892),
    [anon_sym_log] = ACTIONS(2892),
    [anon_sym_lock] = ACTIONS(2892),
    [anon_sym_lstat] = ACTIONS(2892),
    [anon_sym_oct] = ACTIONS(2892),
    [anon_sym_ord] = ACTIONS(2892),
    [anon_sym_prototype] = ACTIONS(2892),
    [anon_sym_pop] = ACTIONS(2892),
    [anon_sym_pos] = ACTIONS(2892),
    [anon_sym_quotemeta] = ACTIONS(2892),
    [anon_sym_reset] = ACTIONS(2892),
    [anon_sym_rand] = ACTIONS(2892),
    [anon_sym_rmdir] = ACTIONS(2892),
    [anon_sym_readdir] = ACTIONS(2892),
    [anon_sym_readline] = ACTIONS(2892),
    [anon_sym_readpipe] = ACTIONS(2892),
    [anon_sym_rewinddir] = ACTIONS(2892),
    [anon_sym_readlink] = ACTIONS(2892),
    [anon_sym_ref] = ACTIONS(2892),
    [anon_sym_scalar] = ACTIONS(2892),
    [anon_sym_shift] = ACTIONS(2892),
    [anon_sym_sin] = ACTIONS(2892),
    [anon_sym_sleep] = ACTIONS(2892),
    [anon_sym_sqrt] = ACTIONS(2892),
    [anon_sym_srand] = ACTIONS(2892),
    [anon_sym_stat] = ACTIONS(2892),
    [anon_sym_study] = ACTIONS(2892),
    [anon_sym_tell] = ACTIONS(2892),
    [anon_sym_telldir] = ACTIONS(2892),
    [anon_sym_tied] = ACTIONS(2892),
    [anon_sym_uc] = ACTIONS(2892),
    [anon_sym_ucfirst] = ACTIONS(2892),
    [anon_sym_untie] = ACTIONS(2892),
    [anon_sym_umask] = ACTIONS(2892),
    [anon_sym_values] = ACTIONS(2892),
    [anon_sym_write] = ACTIONS(2892),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2892),
    [anon_sym_qq] = ACTIONS(2892),
    [anon_sym_qw] = ACTIONS(2892),
    [anon_sym_qx] = ACTIONS(2892),
    [anon_sym_qr] = ACTIONS(2892),
    [anon_sym_m] = ACTIONS(2892),
    [anon_sym_s] = ACTIONS(2892),
    [anon_sym_tr] = ACTIONS(2892),
    [anon_sym_y] = ACTIONS(2892),
    [aux_sym__bareword_token1] = ACTIONS(2892),
    [sym_number] = ACTIONS(2892),
    [anon_sym_true] = ACTIONS(2892),
    [anon_sym_false] = ACTIONS(2892),
    [sym__single_quote] = ACTIONS(2890),
    [sym__double_quote] = ACTIONS(2890),
    [sym__backtick_quote] = ACTIONS(2890),
    [sym__search_slash_quote] = ACTIONS(2890),
    [sym__open_readline_bracket] = ACTIONS(2890),
    [sym__open_fileglob_bracket] = ACTIONS(2890),
    [sym__PERLY_HEREDOC] = ACTIONS(2890),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2890),
    [sym__filetest] = ACTIONS(2890),
  },
  [STATE(1638)] = {
    [sym_heredoc_content] = STATE(1638),
    [sym__identifier] = ACTIONS(2886),
    [anon_sym_LBRACE] = ACTIONS(2886),
    [anon_sym_DOLLAR] = ACTIONS(2886),
    [anon_sym_AT] = ACTIONS(2886),
    [anon_sym_RPAREN] = ACTIONS(2886),
    [anon_sym_my] = ACTIONS(2886),
    [anon_sym_extended] = ACTIONS(2886),
    [anon_sym_async] = ACTIONS(2886),
    [anon_sym_sub] = ACTIONS(2886),
    [anon_sym_method] = ACTIONS(2886),
    [anon_sym_LPAREN] = ACTIONS(2886),
    [anon_sym_continue] = ACTIONS(2886),
    [anon_sym_state] = ACTIONS(2886),
    [anon_sym_our] = ACTIONS(2886),
    [anon_sym_finally] = ACTIONS(3284),
    [anon_sym_LBRACK] = ACTIONS(2886),
    [anon_sym_LT_LT] = ACTIONS(2886),
    [anon_sym_SLASH_SLASH] = ACTIONS(2886),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_TILDE] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_PLUS_PLUS] = ACTIONS(2886),
    [anon_sym_DASH_DASH] = ACTIONS(2886),
    [anon_sym_BSLASH] = ACTIONS(2884),
    [anon_sym_do] = ACTIONS(2886),
    [anon_sym_eval] = ACTIONS(2886),
    [anon_sym_await] = ACTIONS(2886),
    [anon_sym_field] = ACTIONS(2886),
    [anon_sym_local] = ACTIONS(2886),
    [anon_sym_dynamically] = ACTIONS(2886),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2886),
    [anon_sym_require] = ACTIONS(2886),
    [anon_sym_map] = ACTIONS(2886),
    [anon_sym_grep] = ACTIONS(2886),
    [anon_sym_sort] = ACTIONS(2886),
    [anon_sym_goto] = ACTIONS(2886),
    [anon_sym_return] = ACTIONS(2886),
    [anon_sym_undef] = ACTIONS(2886),
    [anon_sym_PERCENT2] = ACTIONS(2886),
    [anon_sym_AMP2] = ACTIONS(2886),
    [anon_sym_STAR2] = ACTIONS(2886),
    [anon_sym_last] = ACTIONS(2886),
    [anon_sym_next] = ACTIONS(2886),
    [anon_sym_redo] = ACTIONS(2886),
    [anon_sym___FILE__] = ACTIONS(2886),
    [anon_sym___LINE__] = ACTIONS(2886),
    [anon_sym___PACKAGE__] = ACTIONS(2886),
    [anon_sym___SUB__] = ACTIONS(2886),
    [anon_sym_break] = ACTIONS(2886),
    [anon_sym_fork] = ACTIONS(2886),
    [anon_sym_getppid] = ACTIONS(2886),
    [anon_sym_time] = ACTIONS(2886),
    [anon_sym_times] = ACTIONS(2886),
    [anon_sym_wait] = ACTIONS(2886),
    [anon_sym_wantarray] = ACTIONS(2886),
    [anon_sym_abs] = ACTIONS(2886),
    [anon_sym_alarm] = ACTIONS(2886),
    [anon_sym_chop] = ACTIONS(2886),
    [anon_sym_chdir] = ACTIONS(2886),
    [anon_sym_close] = ACTIONS(2886),
    [anon_sym_closedir] = ACTIONS(2886),
    [anon_sym_caller] = ACTIONS(2886),
    [anon_sym_chomp] = ACTIONS(2886),
    [anon_sym_chr] = ACTIONS(2886),
    [anon_sym_cos] = ACTIONS(2886),
    [anon_sym_chroot] = ACTIONS(2886),
    [anon_sym_defined] = ACTIONS(2886),
    [anon_sym_delete] = ACTIONS(2886),
    [anon_sym_dbmclose] = ACTIONS(2886),
    [anon_sym_exists] = ACTIONS(2886),
    [anon_sym_exit] = ACTIONS(2886),
    [anon_sym_eof] = ACTIONS(2886),
    [anon_sym_exp] = ACTIONS(2886),
    [anon_sym_each] = ACTIONS(2886),
    [anon_sym_fc] = ACTIONS(2886),
    [anon_sym_fileno] = ACTIONS(2886),
    [anon_sym_gmtime] = ACTIONS(2886),
    [anon_sym_getc] = ACTIONS(2886),
    [anon_sym_getpgrp] = ACTIONS(2886),
    [anon_sym_getprotobyname] = ACTIONS(2886),
    [anon_sym_getpwname] = ACTIONS(2886),
    [anon_sym_getpwuid] = ACTIONS(2886),
    [anon_sym_getpeername] = ACTIONS(2886),
    [anon_sym_getnetbyname] = ACTIONS(2886),
    [anon_sym_getsockname] = ACTIONS(2886),
    [anon_sym_getgrnam] = ACTIONS(2886),
    [anon_sym_getgrgid] = ACTIONS(2886),
    [anon_sym_hex] = ACTIONS(2886),
    [anon_sym_int] = ACTIONS(2886),
    [anon_sym_keys] = ACTIONS(2886),
    [anon_sym_lc] = ACTIONS(2886),
    [anon_sym_lcfirst] = ACTIONS(2886),
    [anon_sym_length] = ACTIONS(2886),
    [anon_sym_localtime] = ACTIONS(2886),
    [anon_sym_log] = ACTIONS(2886),
    [anon_sym_lock] = ACTIONS(2886),
    [anon_sym_lstat] = ACTIONS(2886),
    [anon_sym_oct] = ACTIONS(2886),
    [anon_sym_ord] = ACTIONS(2886),
    [anon_sym_prototype] = ACTIONS(2886),
    [anon_sym_pop] = ACTIONS(2886),
    [anon_sym_pos] = ACTIONS(2886),
    [anon_sym_quotemeta] = ACTIONS(2886),
    [anon_sym_reset] = ACTIONS(2886),
    [anon_sym_rand] = ACTIONS(2886),
    [anon_sym_rmdir] = ACTIONS(2886),
    [anon_sym_readdir] = ACTIONS(2886),
    [anon_sym_readline] = ACTIONS(2886),
    [anon_sym_readpipe] = ACTIONS(2886),
    [anon_sym_rewinddir] = ACTIONS(2886),
    [anon_sym_readlink] = ACTIONS(2886),
    [anon_sym_ref] = ACTIONS(2886),
    [anon_sym_scalar] = ACTIONS(2886),
    [anon_sym_shift] = ACTIONS(2886),
    [anon_sym_sin] = ACTIONS(2886),
    [anon_sym_sleep] = ACTIONS(2886),
    [anon_sym_sqrt] = ACTIONS(2886),
    [anon_sym_srand] = ACTIONS(2886),
    [anon_sym_stat] = ACTIONS(2886),
    [anon_sym_study] = ACTIONS(2886),
    [anon_sym_tell] = ACTIONS(2886),
    [anon_sym_telldir] = ACTIONS(2886),
    [anon_sym_tied] = ACTIONS(2886),
    [anon_sym_uc] = ACTIONS(2886),
    [anon_sym_ucfirst] = ACTIONS(2886),
    [anon_sym_untie] = ACTIONS(2886),
    [anon_sym_umask] = ACTIONS(2886),
    [anon_sym_values] = ACTIONS(2886),
    [anon_sym_write] = ACTIONS(2886),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2886),
    [anon_sym_qq] = ACTIONS(2886),
    [anon_sym_qw] = ACTIONS(2886),
    [anon_sym_qx] = ACTIONS(2886),
    [anon_sym_qr] = ACTIONS(2886),
    [anon_sym_m] = ACTIONS(2886),
    [anon_sym_s] = ACTIONS(2886),
    [anon_sym_tr] = ACTIONS(2886),
    [anon_sym_y] = ACTIONS(2886),
    [aux_sym__bareword_token1] = ACTIONS(2886),
    [sym_number] = ACTIONS(2886),
    [anon_sym_true] = ACTIONS(2886),
    [anon_sym_false] = ACTIONS(2886),
    [sym__single_quote] = ACTIONS(2884),
    [sym__double_quote] = ACTIONS(2884),
    [sym__backtick_quote] = ACTIONS(2884),
    [sym__search_slash_quote] = ACTIONS(2884),
    [sym__open_readline_bracket] = ACTIONS(2884),
    [sym__open_fileglob_bracket] = ACTIONS(2884),
    [sym__PERLY_HEREDOC] = ACTIONS(2884),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2884),
    [sym__filetest] = ACTIONS(2884),
  },
  [STATE(1639)] = {
    [sym_heredoc_content] = STATE(1639),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_finally] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1640)] = {
    [sym_heredoc_content] = STATE(1640),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_finally] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1641)] = {
    [sym_heredoc_content] = STATE(1641),
    [sym__identifier] = ACTIONS(2854),
    [anon_sym_LBRACE] = ACTIONS(2854),
    [anon_sym_SEMI] = ACTIONS(2854),
    [anon_sym_DOLLAR] = ACTIONS(2854),
    [anon_sym_AT] = ACTIONS(2854),
    [anon_sym_my] = ACTIONS(2854),
    [anon_sym_extended] = ACTIONS(2854),
    [anon_sym_async] = ACTIONS(2854),
    [anon_sym_sub] = ACTIONS(2854),
    [anon_sym_method] = ACTIONS(2854),
    [anon_sym_LPAREN] = ACTIONS(2854),
    [anon_sym_continue] = ACTIONS(2854),
    [anon_sym_state] = ACTIONS(2854),
    [anon_sym_our] = ACTIONS(2854),
    [anon_sym_LBRACK] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_TILDE] = ACTIONS(2854),
    [anon_sym_BANG] = ACTIONS(2854),
    [anon_sym_PLUS_PLUS] = ACTIONS(2854),
    [anon_sym_DASH_DASH] = ACTIONS(2854),
    [anon_sym_BSLASH] = ACTIONS(2858),
    [anon_sym_do] = ACTIONS(2854),
    [anon_sym_eval] = ACTIONS(2854),
    [anon_sym_await] = ACTIONS(2854),
    [anon_sym_field] = ACTIONS(2854),
    [anon_sym_local] = ACTIONS(2854),
    [anon_sym_dynamically] = ACTIONS(2854),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2854),
    [anon_sym_require] = ACTIONS(2854),
    [anon_sym_map] = ACTIONS(2854),
    [anon_sym_grep] = ACTIONS(2854),
    [anon_sym_sort] = ACTIONS(2854),
    [anon_sym_goto] = ACTIONS(2854),
    [anon_sym_return] = ACTIONS(2854),
    [anon_sym_undef] = ACTIONS(2854),
    [anon_sym_PERCENT2] = ACTIONS(2854),
    [anon_sym_AMP2] = ACTIONS(2854),
    [anon_sym_STAR2] = ACTIONS(2854),
    [anon_sym_last] = ACTIONS(2854),
    [anon_sym_next] = ACTIONS(2854),
    [anon_sym_redo] = ACTIONS(2854),
    [anon_sym___FILE__] = ACTIONS(2854),
    [anon_sym___LINE__] = ACTIONS(2854),
    [anon_sym___PACKAGE__] = ACTIONS(2854),
    [anon_sym___SUB__] = ACTIONS(2854),
    [anon_sym_break] = ACTIONS(2854),
    [anon_sym_fork] = ACTIONS(2854),
    [anon_sym_getppid] = ACTIONS(2854),
    [anon_sym_time] = ACTIONS(2854),
    [anon_sym_times] = ACTIONS(2854),
    [anon_sym_wait] = ACTIONS(2854),
    [anon_sym_wantarray] = ACTIONS(2854),
    [anon_sym_abs] = ACTIONS(2854),
    [anon_sym_alarm] = ACTIONS(2854),
    [anon_sym_chop] = ACTIONS(2854),
    [anon_sym_chdir] = ACTIONS(2854),
    [anon_sym_close] = ACTIONS(2854),
    [anon_sym_closedir] = ACTIONS(2854),
    [anon_sym_caller] = ACTIONS(2854),
    [anon_sym_chomp] = ACTIONS(2854),
    [anon_sym_chr] = ACTIONS(2854),
    [anon_sym_cos] = ACTIONS(2854),
    [anon_sym_chroot] = ACTIONS(2854),
    [anon_sym_defined] = ACTIONS(2854),
    [anon_sym_delete] = ACTIONS(2854),
    [anon_sym_dbmclose] = ACTIONS(2854),
    [anon_sym_exists] = ACTIONS(2854),
    [anon_sym_exit] = ACTIONS(2854),
    [anon_sym_eof] = ACTIONS(2854),
    [anon_sym_exp] = ACTIONS(2854),
    [anon_sym_each] = ACTIONS(2854),
    [anon_sym_fc] = ACTIONS(2854),
    [anon_sym_fileno] = ACTIONS(2854),
    [anon_sym_gmtime] = ACTIONS(2854),
    [anon_sym_getc] = ACTIONS(2854),
    [anon_sym_getpgrp] = ACTIONS(2854),
    [anon_sym_getprotobyname] = ACTIONS(2854),
    [anon_sym_getpwname] = ACTIONS(2854),
    [anon_sym_getpwuid] = ACTIONS(2854),
    [anon_sym_getpeername] = ACTIONS(2854),
    [anon_sym_getnetbyname] = ACTIONS(2854),
    [anon_sym_getsockname] = ACTIONS(2854),
    [anon_sym_getgrnam] = ACTIONS(2854),
    [anon_sym_getgrgid] = ACTIONS(2854),
    [anon_sym_hex] = ACTIONS(2854),
    [anon_sym_int] = ACTIONS(2854),
    [anon_sym_keys] = ACTIONS(2854),
    [anon_sym_lc] = ACTIONS(2854),
    [anon_sym_lcfirst] = ACTIONS(2854),
    [anon_sym_length] = ACTIONS(2854),
    [anon_sym_localtime] = ACTIONS(2854),
    [anon_sym_log] = ACTIONS(2854),
    [anon_sym_lock] = ACTIONS(2854),
    [anon_sym_lstat] = ACTIONS(2854),
    [anon_sym_oct] = ACTIONS(2854),
    [anon_sym_ord] = ACTIONS(2854),
    [anon_sym_prototype] = ACTIONS(2854),
    [anon_sym_pop] = ACTIONS(2854),
    [anon_sym_pos] = ACTIONS(2854),
    [anon_sym_quotemeta] = ACTIONS(2854),
    [anon_sym_reset] = ACTIONS(2854),
    [anon_sym_rand] = ACTIONS(2854),
    [anon_sym_rmdir] = ACTIONS(2854),
    [anon_sym_readdir] = ACTIONS(2854),
    [anon_sym_readline] = ACTIONS(2854),
    [anon_sym_readpipe] = ACTIONS(2854),
    [anon_sym_rewinddir] = ACTIONS(2854),
    [anon_sym_readlink] = ACTIONS(2854),
    [anon_sym_ref] = ACTIONS(2854),
    [anon_sym_scalar] = ACTIONS(2854),
    [anon_sym_shift] = ACTIONS(2854),
    [anon_sym_sin] = ACTIONS(2854),
    [anon_sym_sleep] = ACTIONS(2854),
    [anon_sym_sqrt] = ACTIONS(2854),
    [anon_sym_srand] = ACTIONS(2854),
    [anon_sym_stat] = ACTIONS(2854),
    [anon_sym_study] = ACTIONS(2854),
    [anon_sym_tell] = ACTIONS(2854),
    [anon_sym_telldir] = ACTIONS(2854),
    [anon_sym_tied] = ACTIONS(2854),
    [anon_sym_uc] = ACTIONS(2854),
    [anon_sym_ucfirst] = ACTIONS(2854),
    [anon_sym_untie] = ACTIONS(2854),
    [anon_sym_umask] = ACTIONS(2854),
    [anon_sym_values] = ACTIONS(2854),
    [anon_sym_write] = ACTIONS(2854),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2854),
    [anon_sym_qq] = ACTIONS(2854),
    [anon_sym_qw] = ACTIONS(2854),
    [anon_sym_qx] = ACTIONS(2854),
    [anon_sym_qr] = ACTIONS(2854),
    [anon_sym_m] = ACTIONS(2854),
    [anon_sym_s] = ACTIONS(2854),
    [anon_sym_tr] = ACTIONS(2854),
    [anon_sym_y] = ACTIONS(2854),
    [aux_sym__bareword_token1] = ACTIONS(2854),
    [sym_number] = ACTIONS(2854),
    [anon_sym_true] = ACTIONS(2854),
    [anon_sym_false] = ACTIONS(2854),
    [sym__single_quote] = ACTIONS(2858),
    [sym__double_quote] = ACTIONS(2858),
    [sym__backtick_quote] = ACTIONS(2858),
    [sym__search_slash_quote] = ACTIONS(2858),
    [sym__open_readline_bracket] = ACTIONS(2858),
    [sym__open_fileglob_bracket] = ACTIONS(2858),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym__PERLY_HEREDOC] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2858),
    [sym__filetest] = ACTIONS(2858),
  },
  [STATE(1642)] = {
    [sym_heredoc_content] = STATE(1642),
    [sym__identifier] = ACTIONS(3077),
    [anon_sym_LBRACE] = ACTIONS(3077),
    [anon_sym_DOLLAR] = ACTIONS(3077),
    [anon_sym_AT] = ACTIONS(3077),
    [anon_sym_RPAREN] = ACTIONS(3077),
    [anon_sym_my] = ACTIONS(3077),
    [anon_sym_extended] = ACTIONS(3077),
    [anon_sym_async] = ACTIONS(3077),
    [anon_sym_sub] = ACTIONS(3077),
    [anon_sym_method] = ACTIONS(3077),
    [anon_sym_LPAREN] = ACTIONS(3077),
    [anon_sym_continue] = ACTIONS(3077),
    [anon_sym_state] = ACTIONS(3077),
    [anon_sym_our] = ACTIONS(3077),
    [anon_sym_LBRACK] = ACTIONS(3077),
    [anon_sym_LT_LT] = ACTIONS(3077),
    [anon_sym_SLASH_SLASH] = ACTIONS(3077),
    [anon_sym_PLUS] = ACTIONS(3077),
    [anon_sym_DASH] = ACTIONS(3077),
    [anon_sym_TILDE] = ACTIONS(3077),
    [anon_sym_BANG] = ACTIONS(3077),
    [anon_sym_PLUS_PLUS] = ACTIONS(3077),
    [anon_sym_DASH_DASH] = ACTIONS(3077),
    [anon_sym_BSLASH] = ACTIONS(3075),
    [anon_sym_do] = ACTIONS(3077),
    [anon_sym_eval] = ACTIONS(3077),
    [anon_sym_await] = ACTIONS(3077),
    [anon_sym_field] = ACTIONS(3077),
    [anon_sym_local] = ACTIONS(3077),
    [anon_sym_dynamically] = ACTIONS(3077),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3077),
    [anon_sym_require] = ACTIONS(3077),
    [anon_sym_map] = ACTIONS(3077),
    [anon_sym_grep] = ACTIONS(3077),
    [anon_sym_sort] = ACTIONS(3077),
    [anon_sym_goto] = ACTIONS(3077),
    [anon_sym_return] = ACTIONS(3077),
    [anon_sym_undef] = ACTIONS(3077),
    [anon_sym_PERCENT2] = ACTIONS(3077),
    [anon_sym_AMP2] = ACTIONS(3077),
    [anon_sym_STAR2] = ACTIONS(3077),
    [anon_sym_last] = ACTIONS(3077),
    [anon_sym_next] = ACTIONS(3077),
    [anon_sym_redo] = ACTIONS(3077),
    [anon_sym___FILE__] = ACTIONS(3077),
    [anon_sym___LINE__] = ACTIONS(3077),
    [anon_sym___PACKAGE__] = ACTIONS(3077),
    [anon_sym___SUB__] = ACTIONS(3077),
    [anon_sym_break] = ACTIONS(3077),
    [anon_sym_fork] = ACTIONS(3077),
    [anon_sym_getppid] = ACTIONS(3077),
    [anon_sym_time] = ACTIONS(3077),
    [anon_sym_times] = ACTIONS(3077),
    [anon_sym_wait] = ACTIONS(3077),
    [anon_sym_wantarray] = ACTIONS(3077),
    [anon_sym_abs] = ACTIONS(3077),
    [anon_sym_alarm] = ACTIONS(3077),
    [anon_sym_chop] = ACTIONS(3077),
    [anon_sym_chdir] = ACTIONS(3077),
    [anon_sym_close] = ACTIONS(3077),
    [anon_sym_closedir] = ACTIONS(3077),
    [anon_sym_caller] = ACTIONS(3077),
    [anon_sym_chomp] = ACTIONS(3077),
    [anon_sym_chr] = ACTIONS(3077),
    [anon_sym_cos] = ACTIONS(3077),
    [anon_sym_chroot] = ACTIONS(3077),
    [anon_sym_defined] = ACTIONS(3077),
    [anon_sym_delete] = ACTIONS(3077),
    [anon_sym_dbmclose] = ACTIONS(3077),
    [anon_sym_exists] = ACTIONS(3077),
    [anon_sym_exit] = ACTIONS(3077),
    [anon_sym_eof] = ACTIONS(3077),
    [anon_sym_exp] = ACTIONS(3077),
    [anon_sym_each] = ACTIONS(3077),
    [anon_sym_fc] = ACTIONS(3077),
    [anon_sym_fileno] = ACTIONS(3077),
    [anon_sym_gmtime] = ACTIONS(3077),
    [anon_sym_getc] = ACTIONS(3077),
    [anon_sym_getpgrp] = ACTIONS(3077),
    [anon_sym_getprotobyname] = ACTIONS(3077),
    [anon_sym_getpwname] = ACTIONS(3077),
    [anon_sym_getpwuid] = ACTIONS(3077),
    [anon_sym_getpeername] = ACTIONS(3077),
    [anon_sym_getnetbyname] = ACTIONS(3077),
    [anon_sym_getsockname] = ACTIONS(3077),
    [anon_sym_getgrnam] = ACTIONS(3077),
    [anon_sym_getgrgid] = ACTIONS(3077),
    [anon_sym_hex] = ACTIONS(3077),
    [anon_sym_int] = ACTIONS(3077),
    [anon_sym_keys] = ACTIONS(3077),
    [anon_sym_lc] = ACTIONS(3077),
    [anon_sym_lcfirst] = ACTIONS(3077),
    [anon_sym_length] = ACTIONS(3077),
    [anon_sym_localtime] = ACTIONS(3077),
    [anon_sym_log] = ACTIONS(3077),
    [anon_sym_lock] = ACTIONS(3077),
    [anon_sym_lstat] = ACTIONS(3077),
    [anon_sym_oct] = ACTIONS(3077),
    [anon_sym_ord] = ACTIONS(3077),
    [anon_sym_prototype] = ACTIONS(3077),
    [anon_sym_pop] = ACTIONS(3077),
    [anon_sym_pos] = ACTIONS(3077),
    [anon_sym_quotemeta] = ACTIONS(3077),
    [anon_sym_reset] = ACTIONS(3077),
    [anon_sym_rand] = ACTIONS(3077),
    [anon_sym_rmdir] = ACTIONS(3077),
    [anon_sym_readdir] = ACTIONS(3077),
    [anon_sym_readline] = ACTIONS(3077),
    [anon_sym_readpipe] = ACTIONS(3077),
    [anon_sym_rewinddir] = ACTIONS(3077),
    [anon_sym_readlink] = ACTIONS(3077),
    [anon_sym_ref] = ACTIONS(3077),
    [anon_sym_scalar] = ACTIONS(3077),
    [anon_sym_shift] = ACTIONS(3077),
    [anon_sym_sin] = ACTIONS(3077),
    [anon_sym_sleep] = ACTIONS(3077),
    [anon_sym_sqrt] = ACTIONS(3077),
    [anon_sym_srand] = ACTIONS(3077),
    [anon_sym_stat] = ACTIONS(3077),
    [anon_sym_study] = ACTIONS(3077),
    [anon_sym_tell] = ACTIONS(3077),
    [anon_sym_telldir] = ACTIONS(3077),
    [anon_sym_tied] = ACTIONS(3077),
    [anon_sym_uc] = ACTIONS(3077),
    [anon_sym_ucfirst] = ACTIONS(3077),
    [anon_sym_untie] = ACTIONS(3077),
    [anon_sym_umask] = ACTIONS(3077),
    [anon_sym_values] = ACTIONS(3077),
    [anon_sym_write] = ACTIONS(3077),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3077),
    [anon_sym_qq] = ACTIONS(3077),
    [anon_sym_qw] = ACTIONS(3077),
    [anon_sym_qx] = ACTIONS(3077),
    [anon_sym_qr] = ACTIONS(3077),
    [anon_sym_m] = ACTIONS(3077),
    [anon_sym_s] = ACTIONS(3077),
    [anon_sym_tr] = ACTIONS(3077),
    [anon_sym_y] = ACTIONS(3077),
    [aux_sym__bareword_token1] = ACTIONS(3077),
    [sym_number] = ACTIONS(3077),
    [anon_sym_true] = ACTIONS(3077),
    [anon_sym_false] = ACTIONS(3077),
    [sym__single_quote] = ACTIONS(3075),
    [sym__double_quote] = ACTIONS(3075),
    [sym__backtick_quote] = ACTIONS(3075),
    [sym__search_slash_quote] = ACTIONS(3075),
    [sym__open_readline_bracket] = ACTIONS(3075),
    [sym__open_fileglob_bracket] = ACTIONS(3075),
    [sym__PERLY_HEREDOC] = ACTIONS(3075),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3075),
    [sym__filetest] = ACTIONS(3075),
  },
  [STATE(1643)] = {
    [sym_heredoc_content] = STATE(1643),
    [sym__identifier] = ACTIONS(2970),
    [anon_sym_LBRACE] = ACTIONS(2970),
    [anon_sym_DOLLAR] = ACTIONS(2970),
    [anon_sym_AT] = ACTIONS(2970),
    [anon_sym_RPAREN] = ACTIONS(2970),
    [anon_sym_my] = ACTIONS(2970),
    [anon_sym_extended] = ACTIONS(2970),
    [anon_sym_async] = ACTIONS(2970),
    [anon_sym_sub] = ACTIONS(2970),
    [anon_sym_method] = ACTIONS(2970),
    [anon_sym_LPAREN] = ACTIONS(2970),
    [anon_sym_continue] = ACTIONS(2970),
    [anon_sym_state] = ACTIONS(2970),
    [anon_sym_our] = ACTIONS(2970),
    [anon_sym_LBRACK] = ACTIONS(2970),
    [anon_sym_LT_LT] = ACTIONS(2970),
    [anon_sym_SLASH_SLASH] = ACTIONS(2970),
    [anon_sym_PLUS] = ACTIONS(2970),
    [anon_sym_DASH] = ACTIONS(2970),
    [anon_sym_TILDE] = ACTIONS(2970),
    [anon_sym_BANG] = ACTIONS(2970),
    [anon_sym_PLUS_PLUS] = ACTIONS(2970),
    [anon_sym_DASH_DASH] = ACTIONS(2970),
    [anon_sym_BSLASH] = ACTIONS(2968),
    [anon_sym_do] = ACTIONS(2970),
    [anon_sym_eval] = ACTIONS(2970),
    [anon_sym_await] = ACTIONS(2970),
    [anon_sym_field] = ACTIONS(2970),
    [anon_sym_local] = ACTIONS(2970),
    [anon_sym_dynamically] = ACTIONS(2970),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2970),
    [anon_sym_require] = ACTIONS(2970),
    [anon_sym_map] = ACTIONS(2970),
    [anon_sym_grep] = ACTIONS(2970),
    [anon_sym_sort] = ACTIONS(2970),
    [anon_sym_goto] = ACTIONS(2970),
    [anon_sym_return] = ACTIONS(2970),
    [anon_sym_undef] = ACTIONS(2970),
    [anon_sym_PERCENT2] = ACTIONS(2970),
    [anon_sym_AMP2] = ACTIONS(2970),
    [anon_sym_STAR2] = ACTIONS(2970),
    [anon_sym_last] = ACTIONS(2970),
    [anon_sym_next] = ACTIONS(2970),
    [anon_sym_redo] = ACTIONS(2970),
    [anon_sym___FILE__] = ACTIONS(2970),
    [anon_sym___LINE__] = ACTIONS(2970),
    [anon_sym___PACKAGE__] = ACTIONS(2970),
    [anon_sym___SUB__] = ACTIONS(2970),
    [anon_sym_break] = ACTIONS(2970),
    [anon_sym_fork] = ACTIONS(2970),
    [anon_sym_getppid] = ACTIONS(2970),
    [anon_sym_time] = ACTIONS(2970),
    [anon_sym_times] = ACTIONS(2970),
    [anon_sym_wait] = ACTIONS(2970),
    [anon_sym_wantarray] = ACTIONS(2970),
    [anon_sym_abs] = ACTIONS(2970),
    [anon_sym_alarm] = ACTIONS(2970),
    [anon_sym_chop] = ACTIONS(2970),
    [anon_sym_chdir] = ACTIONS(2970),
    [anon_sym_close] = ACTIONS(2970),
    [anon_sym_closedir] = ACTIONS(2970),
    [anon_sym_caller] = ACTIONS(2970),
    [anon_sym_chomp] = ACTIONS(2970),
    [anon_sym_chr] = ACTIONS(2970),
    [anon_sym_cos] = ACTIONS(2970),
    [anon_sym_chroot] = ACTIONS(2970),
    [anon_sym_defined] = ACTIONS(2970),
    [anon_sym_delete] = ACTIONS(2970),
    [anon_sym_dbmclose] = ACTIONS(2970),
    [anon_sym_exists] = ACTIONS(2970),
    [anon_sym_exit] = ACTIONS(2970),
    [anon_sym_eof] = ACTIONS(2970),
    [anon_sym_exp] = ACTIONS(2970),
    [anon_sym_each] = ACTIONS(2970),
    [anon_sym_fc] = ACTIONS(2970),
    [anon_sym_fileno] = ACTIONS(2970),
    [anon_sym_gmtime] = ACTIONS(2970),
    [anon_sym_getc] = ACTIONS(2970),
    [anon_sym_getpgrp] = ACTIONS(2970),
    [anon_sym_getprotobyname] = ACTIONS(2970),
    [anon_sym_getpwname] = ACTIONS(2970),
    [anon_sym_getpwuid] = ACTIONS(2970),
    [anon_sym_getpeername] = ACTIONS(2970),
    [anon_sym_getnetbyname] = ACTIONS(2970),
    [anon_sym_getsockname] = ACTIONS(2970),
    [anon_sym_getgrnam] = ACTIONS(2970),
    [anon_sym_getgrgid] = ACTIONS(2970),
    [anon_sym_hex] = ACTIONS(2970),
    [anon_sym_int] = ACTIONS(2970),
    [anon_sym_keys] = ACTIONS(2970),
    [anon_sym_lc] = ACTIONS(2970),
    [anon_sym_lcfirst] = ACTIONS(2970),
    [anon_sym_length] = ACTIONS(2970),
    [anon_sym_localtime] = ACTIONS(2970),
    [anon_sym_log] = ACTIONS(2970),
    [anon_sym_lock] = ACTIONS(2970),
    [anon_sym_lstat] = ACTIONS(2970),
    [anon_sym_oct] = ACTIONS(2970),
    [anon_sym_ord] = ACTIONS(2970),
    [anon_sym_prototype] = ACTIONS(2970),
    [anon_sym_pop] = ACTIONS(2970),
    [anon_sym_pos] = ACTIONS(2970),
    [anon_sym_quotemeta] = ACTIONS(2970),
    [anon_sym_reset] = ACTIONS(2970),
    [anon_sym_rand] = ACTIONS(2970),
    [anon_sym_rmdir] = ACTIONS(2970),
    [anon_sym_readdir] = ACTIONS(2970),
    [anon_sym_readline] = ACTIONS(2970),
    [anon_sym_readpipe] = ACTIONS(2970),
    [anon_sym_rewinddir] = ACTIONS(2970),
    [anon_sym_readlink] = ACTIONS(2970),
    [anon_sym_ref] = ACTIONS(2970),
    [anon_sym_scalar] = ACTIONS(2970),
    [anon_sym_shift] = ACTIONS(2970),
    [anon_sym_sin] = ACTIONS(2970),
    [anon_sym_sleep] = ACTIONS(2970),
    [anon_sym_sqrt] = ACTIONS(2970),
    [anon_sym_srand] = ACTIONS(2970),
    [anon_sym_stat] = ACTIONS(2970),
    [anon_sym_study] = ACTIONS(2970),
    [anon_sym_tell] = ACTIONS(2970),
    [anon_sym_telldir] = ACTIONS(2970),
    [anon_sym_tied] = ACTIONS(2970),
    [anon_sym_uc] = ACTIONS(2970),
    [anon_sym_ucfirst] = ACTIONS(2970),
    [anon_sym_untie] = ACTIONS(2970),
    [anon_sym_umask] = ACTIONS(2970),
    [anon_sym_values] = ACTIONS(2970),
    [anon_sym_write] = ACTIONS(2970),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2970),
    [anon_sym_qq] = ACTIONS(2970),
    [anon_sym_qw] = ACTIONS(2970),
    [anon_sym_qx] = ACTIONS(2970),
    [anon_sym_qr] = ACTIONS(2970),
    [anon_sym_m] = ACTIONS(2970),
    [anon_sym_s] = ACTIONS(2970),
    [anon_sym_tr] = ACTIONS(2970),
    [anon_sym_y] = ACTIONS(2970),
    [aux_sym__bareword_token1] = ACTIONS(2970),
    [sym_number] = ACTIONS(2970),
    [anon_sym_true] = ACTIONS(2970),
    [anon_sym_false] = ACTIONS(2970),
    [sym__single_quote] = ACTIONS(2968),
    [sym__double_quote] = ACTIONS(2968),
    [sym__backtick_quote] = ACTIONS(2968),
    [sym__search_slash_quote] = ACTIONS(2968),
    [sym__open_readline_bracket] = ACTIONS(2968),
    [sym__open_fileglob_bracket] = ACTIONS(2968),
    [sym__PERLY_HEREDOC] = ACTIONS(2968),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2968),
    [sym__filetest] = ACTIONS(2968),
  },
  [STATE(1644)] = {
    [sym_heredoc_content] = STATE(1644),
    [sym__identifier] = ACTIONS(2974),
    [anon_sym_LBRACE] = ACTIONS(2974),
    [anon_sym_DOLLAR] = ACTIONS(2974),
    [anon_sym_AT] = ACTIONS(2974),
    [anon_sym_RPAREN] = ACTIONS(2974),
    [anon_sym_my] = ACTIONS(2974),
    [anon_sym_extended] = ACTIONS(2974),
    [anon_sym_async] = ACTIONS(2974),
    [anon_sym_sub] = ACTIONS(2974),
    [anon_sym_method] = ACTIONS(2974),
    [anon_sym_LPAREN] = ACTIONS(2974),
    [anon_sym_continue] = ACTIONS(2974),
    [anon_sym_state] = ACTIONS(2974),
    [anon_sym_our] = ACTIONS(2974),
    [anon_sym_LBRACK] = ACTIONS(2974),
    [anon_sym_LT_LT] = ACTIONS(2974),
    [anon_sym_SLASH_SLASH] = ACTIONS(2974),
    [anon_sym_PLUS] = ACTIONS(2974),
    [anon_sym_DASH] = ACTIONS(2974),
    [anon_sym_TILDE] = ACTIONS(2974),
    [anon_sym_BANG] = ACTIONS(2974),
    [anon_sym_PLUS_PLUS] = ACTIONS(2974),
    [anon_sym_DASH_DASH] = ACTIONS(2974),
    [anon_sym_BSLASH] = ACTIONS(2972),
    [anon_sym_do] = ACTIONS(2974),
    [anon_sym_eval] = ACTIONS(2974),
    [anon_sym_await] = ACTIONS(2974),
    [anon_sym_field] = ACTIONS(2974),
    [anon_sym_local] = ACTIONS(2974),
    [anon_sym_dynamically] = ACTIONS(2974),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2974),
    [anon_sym_require] = ACTIONS(2974),
    [anon_sym_map] = ACTIONS(2974),
    [anon_sym_grep] = ACTIONS(2974),
    [anon_sym_sort] = ACTIONS(2974),
    [anon_sym_goto] = ACTIONS(2974),
    [anon_sym_return] = ACTIONS(2974),
    [anon_sym_undef] = ACTIONS(2974),
    [anon_sym_PERCENT2] = ACTIONS(2974),
    [anon_sym_AMP2] = ACTIONS(2974),
    [anon_sym_STAR2] = ACTIONS(2974),
    [anon_sym_last] = ACTIONS(2974),
    [anon_sym_next] = ACTIONS(2974),
    [anon_sym_redo] = ACTIONS(2974),
    [anon_sym___FILE__] = ACTIONS(2974),
    [anon_sym___LINE__] = ACTIONS(2974),
    [anon_sym___PACKAGE__] = ACTIONS(2974),
    [anon_sym___SUB__] = ACTIONS(2974),
    [anon_sym_break] = ACTIONS(2974),
    [anon_sym_fork] = ACTIONS(2974),
    [anon_sym_getppid] = ACTIONS(2974),
    [anon_sym_time] = ACTIONS(2974),
    [anon_sym_times] = ACTIONS(2974),
    [anon_sym_wait] = ACTIONS(2974),
    [anon_sym_wantarray] = ACTIONS(2974),
    [anon_sym_abs] = ACTIONS(2974),
    [anon_sym_alarm] = ACTIONS(2974),
    [anon_sym_chop] = ACTIONS(2974),
    [anon_sym_chdir] = ACTIONS(2974),
    [anon_sym_close] = ACTIONS(2974),
    [anon_sym_closedir] = ACTIONS(2974),
    [anon_sym_caller] = ACTIONS(2974),
    [anon_sym_chomp] = ACTIONS(2974),
    [anon_sym_chr] = ACTIONS(2974),
    [anon_sym_cos] = ACTIONS(2974),
    [anon_sym_chroot] = ACTIONS(2974),
    [anon_sym_defined] = ACTIONS(2974),
    [anon_sym_delete] = ACTIONS(2974),
    [anon_sym_dbmclose] = ACTIONS(2974),
    [anon_sym_exists] = ACTIONS(2974),
    [anon_sym_exit] = ACTIONS(2974),
    [anon_sym_eof] = ACTIONS(2974),
    [anon_sym_exp] = ACTIONS(2974),
    [anon_sym_each] = ACTIONS(2974),
    [anon_sym_fc] = ACTIONS(2974),
    [anon_sym_fileno] = ACTIONS(2974),
    [anon_sym_gmtime] = ACTIONS(2974),
    [anon_sym_getc] = ACTIONS(2974),
    [anon_sym_getpgrp] = ACTIONS(2974),
    [anon_sym_getprotobyname] = ACTIONS(2974),
    [anon_sym_getpwname] = ACTIONS(2974),
    [anon_sym_getpwuid] = ACTIONS(2974),
    [anon_sym_getpeername] = ACTIONS(2974),
    [anon_sym_getnetbyname] = ACTIONS(2974),
    [anon_sym_getsockname] = ACTIONS(2974),
    [anon_sym_getgrnam] = ACTIONS(2974),
    [anon_sym_getgrgid] = ACTIONS(2974),
    [anon_sym_hex] = ACTIONS(2974),
    [anon_sym_int] = ACTIONS(2974),
    [anon_sym_keys] = ACTIONS(2974),
    [anon_sym_lc] = ACTIONS(2974),
    [anon_sym_lcfirst] = ACTIONS(2974),
    [anon_sym_length] = ACTIONS(2974),
    [anon_sym_localtime] = ACTIONS(2974),
    [anon_sym_log] = ACTIONS(2974),
    [anon_sym_lock] = ACTIONS(2974),
    [anon_sym_lstat] = ACTIONS(2974),
    [anon_sym_oct] = ACTIONS(2974),
    [anon_sym_ord] = ACTIONS(2974),
    [anon_sym_prototype] = ACTIONS(2974),
    [anon_sym_pop] = ACTIONS(2974),
    [anon_sym_pos] = ACTIONS(2974),
    [anon_sym_quotemeta] = ACTIONS(2974),
    [anon_sym_reset] = ACTIONS(2974),
    [anon_sym_rand] = ACTIONS(2974),
    [anon_sym_rmdir] = ACTIONS(2974),
    [anon_sym_readdir] = ACTIONS(2974),
    [anon_sym_readline] = ACTIONS(2974),
    [anon_sym_readpipe] = ACTIONS(2974),
    [anon_sym_rewinddir] = ACTIONS(2974),
    [anon_sym_readlink] = ACTIONS(2974),
    [anon_sym_ref] = ACTIONS(2974),
    [anon_sym_scalar] = ACTIONS(2974),
    [anon_sym_shift] = ACTIONS(2974),
    [anon_sym_sin] = ACTIONS(2974),
    [anon_sym_sleep] = ACTIONS(2974),
    [anon_sym_sqrt] = ACTIONS(2974),
    [anon_sym_srand] = ACTIONS(2974),
    [anon_sym_stat] = ACTIONS(2974),
    [anon_sym_study] = ACTIONS(2974),
    [anon_sym_tell] = ACTIONS(2974),
    [anon_sym_telldir] = ACTIONS(2974),
    [anon_sym_tied] = ACTIONS(2974),
    [anon_sym_uc] = ACTIONS(2974),
    [anon_sym_ucfirst] = ACTIONS(2974),
    [anon_sym_untie] = ACTIONS(2974),
    [anon_sym_umask] = ACTIONS(2974),
    [anon_sym_values] = ACTIONS(2974),
    [anon_sym_write] = ACTIONS(2974),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2974),
    [anon_sym_qq] = ACTIONS(2974),
    [anon_sym_qw] = ACTIONS(2974),
    [anon_sym_qx] = ACTIONS(2974),
    [anon_sym_qr] = ACTIONS(2974),
    [anon_sym_m] = ACTIONS(2974),
    [anon_sym_s] = ACTIONS(2974),
    [anon_sym_tr] = ACTIONS(2974),
    [anon_sym_y] = ACTIONS(2974),
    [aux_sym__bareword_token1] = ACTIONS(2974),
    [sym_number] = ACTIONS(2974),
    [anon_sym_true] = ACTIONS(2974),
    [anon_sym_false] = ACTIONS(2974),
    [sym__single_quote] = ACTIONS(2972),
    [sym__double_quote] = ACTIONS(2972),
    [sym__backtick_quote] = ACTIONS(2972),
    [sym__search_slash_quote] = ACTIONS(2972),
    [sym__open_readline_bracket] = ACTIONS(2972),
    [sym__open_fileglob_bracket] = ACTIONS(2972),
    [sym__PERLY_HEREDOC] = ACTIONS(2972),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2972),
    [sym__filetest] = ACTIONS(2972),
  },
  [STATE(1645)] = {
    [sym_heredoc_content] = STATE(1645),
    [sym__identifier] = ACTIONS(2978),
    [anon_sym_LBRACE] = ACTIONS(2978),
    [anon_sym_DOLLAR] = ACTIONS(2978),
    [anon_sym_AT] = ACTIONS(2978),
    [anon_sym_RPAREN] = ACTIONS(2978),
    [anon_sym_my] = ACTIONS(2978),
    [anon_sym_extended] = ACTIONS(2978),
    [anon_sym_async] = ACTIONS(2978),
    [anon_sym_sub] = ACTIONS(2978),
    [anon_sym_method] = ACTIONS(2978),
    [anon_sym_LPAREN] = ACTIONS(2978),
    [anon_sym_continue] = ACTIONS(2978),
    [anon_sym_state] = ACTIONS(2978),
    [anon_sym_our] = ACTIONS(2978),
    [anon_sym_LBRACK] = ACTIONS(2978),
    [anon_sym_LT_LT] = ACTIONS(2978),
    [anon_sym_SLASH_SLASH] = ACTIONS(2978),
    [anon_sym_PLUS] = ACTIONS(2978),
    [anon_sym_DASH] = ACTIONS(2978),
    [anon_sym_TILDE] = ACTIONS(2978),
    [anon_sym_BANG] = ACTIONS(2978),
    [anon_sym_PLUS_PLUS] = ACTIONS(2978),
    [anon_sym_DASH_DASH] = ACTIONS(2978),
    [anon_sym_BSLASH] = ACTIONS(2976),
    [anon_sym_do] = ACTIONS(2978),
    [anon_sym_eval] = ACTIONS(2978),
    [anon_sym_await] = ACTIONS(2978),
    [anon_sym_field] = ACTIONS(2978),
    [anon_sym_local] = ACTIONS(2978),
    [anon_sym_dynamically] = ACTIONS(2978),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2978),
    [anon_sym_require] = ACTIONS(2978),
    [anon_sym_map] = ACTIONS(2978),
    [anon_sym_grep] = ACTIONS(2978),
    [anon_sym_sort] = ACTIONS(2978),
    [anon_sym_goto] = ACTIONS(2978),
    [anon_sym_return] = ACTIONS(2978),
    [anon_sym_undef] = ACTIONS(2978),
    [anon_sym_PERCENT2] = ACTIONS(2978),
    [anon_sym_AMP2] = ACTIONS(2978),
    [anon_sym_STAR2] = ACTIONS(2978),
    [anon_sym_last] = ACTIONS(2978),
    [anon_sym_next] = ACTIONS(2978),
    [anon_sym_redo] = ACTIONS(2978),
    [anon_sym___FILE__] = ACTIONS(2978),
    [anon_sym___LINE__] = ACTIONS(2978),
    [anon_sym___PACKAGE__] = ACTIONS(2978),
    [anon_sym___SUB__] = ACTIONS(2978),
    [anon_sym_break] = ACTIONS(2978),
    [anon_sym_fork] = ACTIONS(2978),
    [anon_sym_getppid] = ACTIONS(2978),
    [anon_sym_time] = ACTIONS(2978),
    [anon_sym_times] = ACTIONS(2978),
    [anon_sym_wait] = ACTIONS(2978),
    [anon_sym_wantarray] = ACTIONS(2978),
    [anon_sym_abs] = ACTIONS(2978),
    [anon_sym_alarm] = ACTIONS(2978),
    [anon_sym_chop] = ACTIONS(2978),
    [anon_sym_chdir] = ACTIONS(2978),
    [anon_sym_close] = ACTIONS(2978),
    [anon_sym_closedir] = ACTIONS(2978),
    [anon_sym_caller] = ACTIONS(2978),
    [anon_sym_chomp] = ACTIONS(2978),
    [anon_sym_chr] = ACTIONS(2978),
    [anon_sym_cos] = ACTIONS(2978),
    [anon_sym_chroot] = ACTIONS(2978),
    [anon_sym_defined] = ACTIONS(2978),
    [anon_sym_delete] = ACTIONS(2978),
    [anon_sym_dbmclose] = ACTIONS(2978),
    [anon_sym_exists] = ACTIONS(2978),
    [anon_sym_exit] = ACTIONS(2978),
    [anon_sym_eof] = ACTIONS(2978),
    [anon_sym_exp] = ACTIONS(2978),
    [anon_sym_each] = ACTIONS(2978),
    [anon_sym_fc] = ACTIONS(2978),
    [anon_sym_fileno] = ACTIONS(2978),
    [anon_sym_gmtime] = ACTIONS(2978),
    [anon_sym_getc] = ACTIONS(2978),
    [anon_sym_getpgrp] = ACTIONS(2978),
    [anon_sym_getprotobyname] = ACTIONS(2978),
    [anon_sym_getpwname] = ACTIONS(2978),
    [anon_sym_getpwuid] = ACTIONS(2978),
    [anon_sym_getpeername] = ACTIONS(2978),
    [anon_sym_getnetbyname] = ACTIONS(2978),
    [anon_sym_getsockname] = ACTIONS(2978),
    [anon_sym_getgrnam] = ACTIONS(2978),
    [anon_sym_getgrgid] = ACTIONS(2978),
    [anon_sym_hex] = ACTIONS(2978),
    [anon_sym_int] = ACTIONS(2978),
    [anon_sym_keys] = ACTIONS(2978),
    [anon_sym_lc] = ACTIONS(2978),
    [anon_sym_lcfirst] = ACTIONS(2978),
    [anon_sym_length] = ACTIONS(2978),
    [anon_sym_localtime] = ACTIONS(2978),
    [anon_sym_log] = ACTIONS(2978),
    [anon_sym_lock] = ACTIONS(2978),
    [anon_sym_lstat] = ACTIONS(2978),
    [anon_sym_oct] = ACTIONS(2978),
    [anon_sym_ord] = ACTIONS(2978),
    [anon_sym_prototype] = ACTIONS(2978),
    [anon_sym_pop] = ACTIONS(2978),
    [anon_sym_pos] = ACTIONS(2978),
    [anon_sym_quotemeta] = ACTIONS(2978),
    [anon_sym_reset] = ACTIONS(2978),
    [anon_sym_rand] = ACTIONS(2978),
    [anon_sym_rmdir] = ACTIONS(2978),
    [anon_sym_readdir] = ACTIONS(2978),
    [anon_sym_readline] = ACTIONS(2978),
    [anon_sym_readpipe] = ACTIONS(2978),
    [anon_sym_rewinddir] = ACTIONS(2978),
    [anon_sym_readlink] = ACTIONS(2978),
    [anon_sym_ref] = ACTIONS(2978),
    [anon_sym_scalar] = ACTIONS(2978),
    [anon_sym_shift] = ACTIONS(2978),
    [anon_sym_sin] = ACTIONS(2978),
    [anon_sym_sleep] = ACTIONS(2978),
    [anon_sym_sqrt] = ACTIONS(2978),
    [anon_sym_srand] = ACTIONS(2978),
    [anon_sym_stat] = ACTIONS(2978),
    [anon_sym_study] = ACTIONS(2978),
    [anon_sym_tell] = ACTIONS(2978),
    [anon_sym_telldir] = ACTIONS(2978),
    [anon_sym_tied] = ACTIONS(2978),
    [anon_sym_uc] = ACTIONS(2978),
    [anon_sym_ucfirst] = ACTIONS(2978),
    [anon_sym_untie] = ACTIONS(2978),
    [anon_sym_umask] = ACTIONS(2978),
    [anon_sym_values] = ACTIONS(2978),
    [anon_sym_write] = ACTIONS(2978),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2978),
    [anon_sym_qq] = ACTIONS(2978),
    [anon_sym_qw] = ACTIONS(2978),
    [anon_sym_qx] = ACTIONS(2978),
    [anon_sym_qr] = ACTIONS(2978),
    [anon_sym_m] = ACTIONS(2978),
    [anon_sym_s] = ACTIONS(2978),
    [anon_sym_tr] = ACTIONS(2978),
    [anon_sym_y] = ACTIONS(2978),
    [aux_sym__bareword_token1] = ACTIONS(2978),
    [sym_number] = ACTIONS(2978),
    [anon_sym_true] = ACTIONS(2978),
    [anon_sym_false] = ACTIONS(2978),
    [sym__single_quote] = ACTIONS(2976),
    [sym__double_quote] = ACTIONS(2976),
    [sym__backtick_quote] = ACTIONS(2976),
    [sym__search_slash_quote] = ACTIONS(2976),
    [sym__open_readline_bracket] = ACTIONS(2976),
    [sym__open_fileglob_bracket] = ACTIONS(2976),
    [sym__PERLY_HEREDOC] = ACTIONS(2976),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2976),
    [sym__filetest] = ACTIONS(2976),
  },
  [STATE(1646)] = {
    [sym_heredoc_content] = STATE(1646),
    [sym__identifier] = ACTIONS(2982),
    [anon_sym_LBRACE] = ACTIONS(2982),
    [anon_sym_DOLLAR] = ACTIONS(2982),
    [anon_sym_AT] = ACTIONS(2982),
    [anon_sym_RPAREN] = ACTIONS(2982),
    [anon_sym_my] = ACTIONS(2982),
    [anon_sym_extended] = ACTIONS(2982),
    [anon_sym_async] = ACTIONS(2982),
    [anon_sym_sub] = ACTIONS(2982),
    [anon_sym_method] = ACTIONS(2982),
    [anon_sym_LPAREN] = ACTIONS(2982),
    [anon_sym_continue] = ACTIONS(2982),
    [anon_sym_state] = ACTIONS(2982),
    [anon_sym_our] = ACTIONS(2982),
    [anon_sym_LBRACK] = ACTIONS(2982),
    [anon_sym_LT_LT] = ACTIONS(2982),
    [anon_sym_SLASH_SLASH] = ACTIONS(2982),
    [anon_sym_PLUS] = ACTIONS(2982),
    [anon_sym_DASH] = ACTIONS(2982),
    [anon_sym_TILDE] = ACTIONS(2982),
    [anon_sym_BANG] = ACTIONS(2982),
    [anon_sym_PLUS_PLUS] = ACTIONS(2982),
    [anon_sym_DASH_DASH] = ACTIONS(2982),
    [anon_sym_BSLASH] = ACTIONS(2980),
    [anon_sym_do] = ACTIONS(2982),
    [anon_sym_eval] = ACTIONS(2982),
    [anon_sym_await] = ACTIONS(2982),
    [anon_sym_field] = ACTIONS(2982),
    [anon_sym_local] = ACTIONS(2982),
    [anon_sym_dynamically] = ACTIONS(2982),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2982),
    [anon_sym_require] = ACTIONS(2982),
    [anon_sym_map] = ACTIONS(2982),
    [anon_sym_grep] = ACTIONS(2982),
    [anon_sym_sort] = ACTIONS(2982),
    [anon_sym_goto] = ACTIONS(2982),
    [anon_sym_return] = ACTIONS(2982),
    [anon_sym_undef] = ACTIONS(2982),
    [anon_sym_PERCENT2] = ACTIONS(2982),
    [anon_sym_AMP2] = ACTIONS(2982),
    [anon_sym_STAR2] = ACTIONS(2982),
    [anon_sym_last] = ACTIONS(2982),
    [anon_sym_next] = ACTIONS(2982),
    [anon_sym_redo] = ACTIONS(2982),
    [anon_sym___FILE__] = ACTIONS(2982),
    [anon_sym___LINE__] = ACTIONS(2982),
    [anon_sym___PACKAGE__] = ACTIONS(2982),
    [anon_sym___SUB__] = ACTIONS(2982),
    [anon_sym_break] = ACTIONS(2982),
    [anon_sym_fork] = ACTIONS(2982),
    [anon_sym_getppid] = ACTIONS(2982),
    [anon_sym_time] = ACTIONS(2982),
    [anon_sym_times] = ACTIONS(2982),
    [anon_sym_wait] = ACTIONS(2982),
    [anon_sym_wantarray] = ACTIONS(2982),
    [anon_sym_abs] = ACTIONS(2982),
    [anon_sym_alarm] = ACTIONS(2982),
    [anon_sym_chop] = ACTIONS(2982),
    [anon_sym_chdir] = ACTIONS(2982),
    [anon_sym_close] = ACTIONS(2982),
    [anon_sym_closedir] = ACTIONS(2982),
    [anon_sym_caller] = ACTIONS(2982),
    [anon_sym_chomp] = ACTIONS(2982),
    [anon_sym_chr] = ACTIONS(2982),
    [anon_sym_cos] = ACTIONS(2982),
    [anon_sym_chroot] = ACTIONS(2982),
    [anon_sym_defined] = ACTIONS(2982),
    [anon_sym_delete] = ACTIONS(2982),
    [anon_sym_dbmclose] = ACTIONS(2982),
    [anon_sym_exists] = ACTIONS(2982),
    [anon_sym_exit] = ACTIONS(2982),
    [anon_sym_eof] = ACTIONS(2982),
    [anon_sym_exp] = ACTIONS(2982),
    [anon_sym_each] = ACTIONS(2982),
    [anon_sym_fc] = ACTIONS(2982),
    [anon_sym_fileno] = ACTIONS(2982),
    [anon_sym_gmtime] = ACTIONS(2982),
    [anon_sym_getc] = ACTIONS(2982),
    [anon_sym_getpgrp] = ACTIONS(2982),
    [anon_sym_getprotobyname] = ACTIONS(2982),
    [anon_sym_getpwname] = ACTIONS(2982),
    [anon_sym_getpwuid] = ACTIONS(2982),
    [anon_sym_getpeername] = ACTIONS(2982),
    [anon_sym_getnetbyname] = ACTIONS(2982),
    [anon_sym_getsockname] = ACTIONS(2982),
    [anon_sym_getgrnam] = ACTIONS(2982),
    [anon_sym_getgrgid] = ACTIONS(2982),
    [anon_sym_hex] = ACTIONS(2982),
    [anon_sym_int] = ACTIONS(2982),
    [anon_sym_keys] = ACTIONS(2982),
    [anon_sym_lc] = ACTIONS(2982),
    [anon_sym_lcfirst] = ACTIONS(2982),
    [anon_sym_length] = ACTIONS(2982),
    [anon_sym_localtime] = ACTIONS(2982),
    [anon_sym_log] = ACTIONS(2982),
    [anon_sym_lock] = ACTIONS(2982),
    [anon_sym_lstat] = ACTIONS(2982),
    [anon_sym_oct] = ACTIONS(2982),
    [anon_sym_ord] = ACTIONS(2982),
    [anon_sym_prototype] = ACTIONS(2982),
    [anon_sym_pop] = ACTIONS(2982),
    [anon_sym_pos] = ACTIONS(2982),
    [anon_sym_quotemeta] = ACTIONS(2982),
    [anon_sym_reset] = ACTIONS(2982),
    [anon_sym_rand] = ACTIONS(2982),
    [anon_sym_rmdir] = ACTIONS(2982),
    [anon_sym_readdir] = ACTIONS(2982),
    [anon_sym_readline] = ACTIONS(2982),
    [anon_sym_readpipe] = ACTIONS(2982),
    [anon_sym_rewinddir] = ACTIONS(2982),
    [anon_sym_readlink] = ACTIONS(2982),
    [anon_sym_ref] = ACTIONS(2982),
    [anon_sym_scalar] = ACTIONS(2982),
    [anon_sym_shift] = ACTIONS(2982),
    [anon_sym_sin] = ACTIONS(2982),
    [anon_sym_sleep] = ACTIONS(2982),
    [anon_sym_sqrt] = ACTIONS(2982),
    [anon_sym_srand] = ACTIONS(2982),
    [anon_sym_stat] = ACTIONS(2982),
    [anon_sym_study] = ACTIONS(2982),
    [anon_sym_tell] = ACTIONS(2982),
    [anon_sym_telldir] = ACTIONS(2982),
    [anon_sym_tied] = ACTIONS(2982),
    [anon_sym_uc] = ACTIONS(2982),
    [anon_sym_ucfirst] = ACTIONS(2982),
    [anon_sym_untie] = ACTIONS(2982),
    [anon_sym_umask] = ACTIONS(2982),
    [anon_sym_values] = ACTIONS(2982),
    [anon_sym_write] = ACTIONS(2982),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2982),
    [anon_sym_qq] = ACTIONS(2982),
    [anon_sym_qw] = ACTIONS(2982),
    [anon_sym_qx] = ACTIONS(2982),
    [anon_sym_qr] = ACTIONS(2982),
    [anon_sym_m] = ACTIONS(2982),
    [anon_sym_s] = ACTIONS(2982),
    [anon_sym_tr] = ACTIONS(2982),
    [anon_sym_y] = ACTIONS(2982),
    [aux_sym__bareword_token1] = ACTIONS(2982),
    [sym_number] = ACTIONS(2982),
    [anon_sym_true] = ACTIONS(2982),
    [anon_sym_false] = ACTIONS(2982),
    [sym__single_quote] = ACTIONS(2980),
    [sym__double_quote] = ACTIONS(2980),
    [sym__backtick_quote] = ACTIONS(2980),
    [sym__search_slash_quote] = ACTIONS(2980),
    [sym__open_readline_bracket] = ACTIONS(2980),
    [sym__open_fileglob_bracket] = ACTIONS(2980),
    [sym__PERLY_HEREDOC] = ACTIONS(2980),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2980),
    [sym__filetest] = ACTIONS(2980),
  },
  [STATE(1647)] = {
    [sym_heredoc_content] = STATE(1647),
    [sym__identifier] = ACTIONS(2994),
    [anon_sym_LBRACE] = ACTIONS(2994),
    [anon_sym_DOLLAR] = ACTIONS(2994),
    [anon_sym_AT] = ACTIONS(2994),
    [anon_sym_RPAREN] = ACTIONS(2994),
    [anon_sym_my] = ACTIONS(2994),
    [anon_sym_extended] = ACTIONS(2994),
    [anon_sym_async] = ACTIONS(2994),
    [anon_sym_sub] = ACTIONS(2994),
    [anon_sym_method] = ACTIONS(2994),
    [anon_sym_LPAREN] = ACTIONS(2994),
    [anon_sym_continue] = ACTIONS(2994),
    [anon_sym_state] = ACTIONS(2994),
    [anon_sym_our] = ACTIONS(2994),
    [anon_sym_LBRACK] = ACTIONS(2994),
    [anon_sym_LT_LT] = ACTIONS(2994),
    [anon_sym_SLASH_SLASH] = ACTIONS(2994),
    [anon_sym_PLUS] = ACTIONS(2994),
    [anon_sym_DASH] = ACTIONS(2994),
    [anon_sym_TILDE] = ACTIONS(2994),
    [anon_sym_BANG] = ACTIONS(2994),
    [anon_sym_PLUS_PLUS] = ACTIONS(2994),
    [anon_sym_DASH_DASH] = ACTIONS(2994),
    [anon_sym_BSLASH] = ACTIONS(2992),
    [anon_sym_do] = ACTIONS(2994),
    [anon_sym_eval] = ACTIONS(2994),
    [anon_sym_await] = ACTIONS(2994),
    [anon_sym_field] = ACTIONS(2994),
    [anon_sym_local] = ACTIONS(2994),
    [anon_sym_dynamically] = ACTIONS(2994),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2994),
    [anon_sym_require] = ACTIONS(2994),
    [anon_sym_map] = ACTIONS(2994),
    [anon_sym_grep] = ACTIONS(2994),
    [anon_sym_sort] = ACTIONS(2994),
    [anon_sym_goto] = ACTIONS(2994),
    [anon_sym_return] = ACTIONS(2994),
    [anon_sym_undef] = ACTIONS(2994),
    [anon_sym_PERCENT2] = ACTIONS(2994),
    [anon_sym_AMP2] = ACTIONS(2994),
    [anon_sym_STAR2] = ACTIONS(2994),
    [anon_sym_last] = ACTIONS(2994),
    [anon_sym_next] = ACTIONS(2994),
    [anon_sym_redo] = ACTIONS(2994),
    [anon_sym___FILE__] = ACTIONS(2994),
    [anon_sym___LINE__] = ACTIONS(2994),
    [anon_sym___PACKAGE__] = ACTIONS(2994),
    [anon_sym___SUB__] = ACTIONS(2994),
    [anon_sym_break] = ACTIONS(2994),
    [anon_sym_fork] = ACTIONS(2994),
    [anon_sym_getppid] = ACTIONS(2994),
    [anon_sym_time] = ACTIONS(2994),
    [anon_sym_times] = ACTIONS(2994),
    [anon_sym_wait] = ACTIONS(2994),
    [anon_sym_wantarray] = ACTIONS(2994),
    [anon_sym_abs] = ACTIONS(2994),
    [anon_sym_alarm] = ACTIONS(2994),
    [anon_sym_chop] = ACTIONS(2994),
    [anon_sym_chdir] = ACTIONS(2994),
    [anon_sym_close] = ACTIONS(2994),
    [anon_sym_closedir] = ACTIONS(2994),
    [anon_sym_caller] = ACTIONS(2994),
    [anon_sym_chomp] = ACTIONS(2994),
    [anon_sym_chr] = ACTIONS(2994),
    [anon_sym_cos] = ACTIONS(2994),
    [anon_sym_chroot] = ACTIONS(2994),
    [anon_sym_defined] = ACTIONS(2994),
    [anon_sym_delete] = ACTIONS(2994),
    [anon_sym_dbmclose] = ACTIONS(2994),
    [anon_sym_exists] = ACTIONS(2994),
    [anon_sym_exit] = ACTIONS(2994),
    [anon_sym_eof] = ACTIONS(2994),
    [anon_sym_exp] = ACTIONS(2994),
    [anon_sym_each] = ACTIONS(2994),
    [anon_sym_fc] = ACTIONS(2994),
    [anon_sym_fileno] = ACTIONS(2994),
    [anon_sym_gmtime] = ACTIONS(2994),
    [anon_sym_getc] = ACTIONS(2994),
    [anon_sym_getpgrp] = ACTIONS(2994),
    [anon_sym_getprotobyname] = ACTIONS(2994),
    [anon_sym_getpwname] = ACTIONS(2994),
    [anon_sym_getpwuid] = ACTIONS(2994),
    [anon_sym_getpeername] = ACTIONS(2994),
    [anon_sym_getnetbyname] = ACTIONS(2994),
    [anon_sym_getsockname] = ACTIONS(2994),
    [anon_sym_getgrnam] = ACTIONS(2994),
    [anon_sym_getgrgid] = ACTIONS(2994),
    [anon_sym_hex] = ACTIONS(2994),
    [anon_sym_int] = ACTIONS(2994),
    [anon_sym_keys] = ACTIONS(2994),
    [anon_sym_lc] = ACTIONS(2994),
    [anon_sym_lcfirst] = ACTIONS(2994),
    [anon_sym_length] = ACTIONS(2994),
    [anon_sym_localtime] = ACTIONS(2994),
    [anon_sym_log] = ACTIONS(2994),
    [anon_sym_lock] = ACTIONS(2994),
    [anon_sym_lstat] = ACTIONS(2994),
    [anon_sym_oct] = ACTIONS(2994),
    [anon_sym_ord] = ACTIONS(2994),
    [anon_sym_prototype] = ACTIONS(2994),
    [anon_sym_pop] = ACTIONS(2994),
    [anon_sym_pos] = ACTIONS(2994),
    [anon_sym_quotemeta] = ACTIONS(2994),
    [anon_sym_reset] = ACTIONS(2994),
    [anon_sym_rand] = ACTIONS(2994),
    [anon_sym_rmdir] = ACTIONS(2994),
    [anon_sym_readdir] = ACTIONS(2994),
    [anon_sym_readline] = ACTIONS(2994),
    [anon_sym_readpipe] = ACTIONS(2994),
    [anon_sym_rewinddir] = ACTIONS(2994),
    [anon_sym_readlink] = ACTIONS(2994),
    [anon_sym_ref] = ACTIONS(2994),
    [anon_sym_scalar] = ACTIONS(2994),
    [anon_sym_shift] = ACTIONS(2994),
    [anon_sym_sin] = ACTIONS(2994),
    [anon_sym_sleep] = ACTIONS(2994),
    [anon_sym_sqrt] = ACTIONS(2994),
    [anon_sym_srand] = ACTIONS(2994),
    [anon_sym_stat] = ACTIONS(2994),
    [anon_sym_study] = ACTIONS(2994),
    [anon_sym_tell] = ACTIONS(2994),
    [anon_sym_telldir] = ACTIONS(2994),
    [anon_sym_tied] = ACTIONS(2994),
    [anon_sym_uc] = ACTIONS(2994),
    [anon_sym_ucfirst] = ACTIONS(2994),
    [anon_sym_untie] = ACTIONS(2994),
    [anon_sym_umask] = ACTIONS(2994),
    [anon_sym_values] = ACTIONS(2994),
    [anon_sym_write] = ACTIONS(2994),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2994),
    [anon_sym_qq] = ACTIONS(2994),
    [anon_sym_qw] = ACTIONS(2994),
    [anon_sym_qx] = ACTIONS(2994),
    [anon_sym_qr] = ACTIONS(2994),
    [anon_sym_m] = ACTIONS(2994),
    [anon_sym_s] = ACTIONS(2994),
    [anon_sym_tr] = ACTIONS(2994),
    [anon_sym_y] = ACTIONS(2994),
    [aux_sym__bareword_token1] = ACTIONS(2994),
    [sym_number] = ACTIONS(2994),
    [anon_sym_true] = ACTIONS(2994),
    [anon_sym_false] = ACTIONS(2994),
    [sym__single_quote] = ACTIONS(2992),
    [sym__double_quote] = ACTIONS(2992),
    [sym__backtick_quote] = ACTIONS(2992),
    [sym__search_slash_quote] = ACTIONS(2992),
    [sym__open_readline_bracket] = ACTIONS(2992),
    [sym__open_fileglob_bracket] = ACTIONS(2992),
    [sym__PERLY_HEREDOC] = ACTIONS(2992),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2992),
    [sym__filetest] = ACTIONS(2992),
  },
  [STATE(1648)] = {
    [sym_heredoc_content] = STATE(1648),
    [sym__identifier] = ACTIONS(2998),
    [anon_sym_LBRACE] = ACTIONS(2998),
    [anon_sym_DOLLAR] = ACTIONS(2998),
    [anon_sym_AT] = ACTIONS(2998),
    [anon_sym_RPAREN] = ACTIONS(2998),
    [anon_sym_my] = ACTIONS(2998),
    [anon_sym_extended] = ACTIONS(2998),
    [anon_sym_async] = ACTIONS(2998),
    [anon_sym_sub] = ACTIONS(2998),
    [anon_sym_method] = ACTIONS(2998),
    [anon_sym_LPAREN] = ACTIONS(2998),
    [anon_sym_continue] = ACTIONS(2998),
    [anon_sym_state] = ACTIONS(2998),
    [anon_sym_our] = ACTIONS(2998),
    [anon_sym_LBRACK] = ACTIONS(2998),
    [anon_sym_LT_LT] = ACTIONS(2998),
    [anon_sym_SLASH_SLASH] = ACTIONS(2998),
    [anon_sym_PLUS] = ACTIONS(2998),
    [anon_sym_DASH] = ACTIONS(2998),
    [anon_sym_TILDE] = ACTIONS(2998),
    [anon_sym_BANG] = ACTIONS(2998),
    [anon_sym_PLUS_PLUS] = ACTIONS(2998),
    [anon_sym_DASH_DASH] = ACTIONS(2998),
    [anon_sym_BSLASH] = ACTIONS(2996),
    [anon_sym_do] = ACTIONS(2998),
    [anon_sym_eval] = ACTIONS(2998),
    [anon_sym_await] = ACTIONS(2998),
    [anon_sym_field] = ACTIONS(2998),
    [anon_sym_local] = ACTIONS(2998),
    [anon_sym_dynamically] = ACTIONS(2998),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2998),
    [anon_sym_require] = ACTIONS(2998),
    [anon_sym_map] = ACTIONS(2998),
    [anon_sym_grep] = ACTIONS(2998),
    [anon_sym_sort] = ACTIONS(2998),
    [anon_sym_goto] = ACTIONS(2998),
    [anon_sym_return] = ACTIONS(2998),
    [anon_sym_undef] = ACTIONS(2998),
    [anon_sym_PERCENT2] = ACTIONS(2998),
    [anon_sym_AMP2] = ACTIONS(2998),
    [anon_sym_STAR2] = ACTIONS(2998),
    [anon_sym_last] = ACTIONS(2998),
    [anon_sym_next] = ACTIONS(2998),
    [anon_sym_redo] = ACTIONS(2998),
    [anon_sym___FILE__] = ACTIONS(2998),
    [anon_sym___LINE__] = ACTIONS(2998),
    [anon_sym___PACKAGE__] = ACTIONS(2998),
    [anon_sym___SUB__] = ACTIONS(2998),
    [anon_sym_break] = ACTIONS(2998),
    [anon_sym_fork] = ACTIONS(2998),
    [anon_sym_getppid] = ACTIONS(2998),
    [anon_sym_time] = ACTIONS(2998),
    [anon_sym_times] = ACTIONS(2998),
    [anon_sym_wait] = ACTIONS(2998),
    [anon_sym_wantarray] = ACTIONS(2998),
    [anon_sym_abs] = ACTIONS(2998),
    [anon_sym_alarm] = ACTIONS(2998),
    [anon_sym_chop] = ACTIONS(2998),
    [anon_sym_chdir] = ACTIONS(2998),
    [anon_sym_close] = ACTIONS(2998),
    [anon_sym_closedir] = ACTIONS(2998),
    [anon_sym_caller] = ACTIONS(2998),
    [anon_sym_chomp] = ACTIONS(2998),
    [anon_sym_chr] = ACTIONS(2998),
    [anon_sym_cos] = ACTIONS(2998),
    [anon_sym_chroot] = ACTIONS(2998),
    [anon_sym_defined] = ACTIONS(2998),
    [anon_sym_delete] = ACTIONS(2998),
    [anon_sym_dbmclose] = ACTIONS(2998),
    [anon_sym_exists] = ACTIONS(2998),
    [anon_sym_exit] = ACTIONS(2998),
    [anon_sym_eof] = ACTIONS(2998),
    [anon_sym_exp] = ACTIONS(2998),
    [anon_sym_each] = ACTIONS(2998),
    [anon_sym_fc] = ACTIONS(2998),
    [anon_sym_fileno] = ACTIONS(2998),
    [anon_sym_gmtime] = ACTIONS(2998),
    [anon_sym_getc] = ACTIONS(2998),
    [anon_sym_getpgrp] = ACTIONS(2998),
    [anon_sym_getprotobyname] = ACTIONS(2998),
    [anon_sym_getpwname] = ACTIONS(2998),
    [anon_sym_getpwuid] = ACTIONS(2998),
    [anon_sym_getpeername] = ACTIONS(2998),
    [anon_sym_getnetbyname] = ACTIONS(2998),
    [anon_sym_getsockname] = ACTIONS(2998),
    [anon_sym_getgrnam] = ACTIONS(2998),
    [anon_sym_getgrgid] = ACTIONS(2998),
    [anon_sym_hex] = ACTIONS(2998),
    [anon_sym_int] = ACTIONS(2998),
    [anon_sym_keys] = ACTIONS(2998),
    [anon_sym_lc] = ACTIONS(2998),
    [anon_sym_lcfirst] = ACTIONS(2998),
    [anon_sym_length] = ACTIONS(2998),
    [anon_sym_localtime] = ACTIONS(2998),
    [anon_sym_log] = ACTIONS(2998),
    [anon_sym_lock] = ACTIONS(2998),
    [anon_sym_lstat] = ACTIONS(2998),
    [anon_sym_oct] = ACTIONS(2998),
    [anon_sym_ord] = ACTIONS(2998),
    [anon_sym_prototype] = ACTIONS(2998),
    [anon_sym_pop] = ACTIONS(2998),
    [anon_sym_pos] = ACTIONS(2998),
    [anon_sym_quotemeta] = ACTIONS(2998),
    [anon_sym_reset] = ACTIONS(2998),
    [anon_sym_rand] = ACTIONS(2998),
    [anon_sym_rmdir] = ACTIONS(2998),
    [anon_sym_readdir] = ACTIONS(2998),
    [anon_sym_readline] = ACTIONS(2998),
    [anon_sym_readpipe] = ACTIONS(2998),
    [anon_sym_rewinddir] = ACTIONS(2998),
    [anon_sym_readlink] = ACTIONS(2998),
    [anon_sym_ref] = ACTIONS(2998),
    [anon_sym_scalar] = ACTIONS(2998),
    [anon_sym_shift] = ACTIONS(2998),
    [anon_sym_sin] = ACTIONS(2998),
    [anon_sym_sleep] = ACTIONS(2998),
    [anon_sym_sqrt] = ACTIONS(2998),
    [anon_sym_srand] = ACTIONS(2998),
    [anon_sym_stat] = ACTIONS(2998),
    [anon_sym_study] = ACTIONS(2998),
    [anon_sym_tell] = ACTIONS(2998),
    [anon_sym_telldir] = ACTIONS(2998),
    [anon_sym_tied] = ACTIONS(2998),
    [anon_sym_uc] = ACTIONS(2998),
    [anon_sym_ucfirst] = ACTIONS(2998),
    [anon_sym_untie] = ACTIONS(2998),
    [anon_sym_umask] = ACTIONS(2998),
    [anon_sym_values] = ACTIONS(2998),
    [anon_sym_write] = ACTIONS(2998),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2998),
    [anon_sym_qq] = ACTIONS(2998),
    [anon_sym_qw] = ACTIONS(2998),
    [anon_sym_qx] = ACTIONS(2998),
    [anon_sym_qr] = ACTIONS(2998),
    [anon_sym_m] = ACTIONS(2998),
    [anon_sym_s] = ACTIONS(2998),
    [anon_sym_tr] = ACTIONS(2998),
    [anon_sym_y] = ACTIONS(2998),
    [aux_sym__bareword_token1] = ACTIONS(2998),
    [sym_number] = ACTIONS(2998),
    [anon_sym_true] = ACTIONS(2998),
    [anon_sym_false] = ACTIONS(2998),
    [sym__single_quote] = ACTIONS(2996),
    [sym__double_quote] = ACTIONS(2996),
    [sym__backtick_quote] = ACTIONS(2996),
    [sym__search_slash_quote] = ACTIONS(2996),
    [sym__open_readline_bracket] = ACTIONS(2996),
    [sym__open_fileglob_bracket] = ACTIONS(2996),
    [sym__PERLY_HEREDOC] = ACTIONS(2996),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2996),
    [sym__filetest] = ACTIONS(2996),
  },
  [STATE(1649)] = {
    [sym_heredoc_content] = STATE(1649),
    [sym__identifier] = ACTIONS(3002),
    [anon_sym_LBRACE] = ACTIONS(3002),
    [anon_sym_DOLLAR] = ACTIONS(3002),
    [anon_sym_AT] = ACTIONS(3002),
    [anon_sym_RPAREN] = ACTIONS(3002),
    [anon_sym_my] = ACTIONS(3002),
    [anon_sym_extended] = ACTIONS(3002),
    [anon_sym_async] = ACTIONS(3002),
    [anon_sym_sub] = ACTIONS(3002),
    [anon_sym_method] = ACTIONS(3002),
    [anon_sym_LPAREN] = ACTIONS(3002),
    [anon_sym_continue] = ACTIONS(3002),
    [anon_sym_state] = ACTIONS(3002),
    [anon_sym_our] = ACTIONS(3002),
    [anon_sym_LBRACK] = ACTIONS(3002),
    [anon_sym_LT_LT] = ACTIONS(3002),
    [anon_sym_SLASH_SLASH] = ACTIONS(3002),
    [anon_sym_PLUS] = ACTIONS(3002),
    [anon_sym_DASH] = ACTIONS(3002),
    [anon_sym_TILDE] = ACTIONS(3002),
    [anon_sym_BANG] = ACTIONS(3002),
    [anon_sym_PLUS_PLUS] = ACTIONS(3002),
    [anon_sym_DASH_DASH] = ACTIONS(3002),
    [anon_sym_BSLASH] = ACTIONS(3000),
    [anon_sym_do] = ACTIONS(3002),
    [anon_sym_eval] = ACTIONS(3002),
    [anon_sym_await] = ACTIONS(3002),
    [anon_sym_field] = ACTIONS(3002),
    [anon_sym_local] = ACTIONS(3002),
    [anon_sym_dynamically] = ACTIONS(3002),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3002),
    [anon_sym_require] = ACTIONS(3002),
    [anon_sym_map] = ACTIONS(3002),
    [anon_sym_grep] = ACTIONS(3002),
    [anon_sym_sort] = ACTIONS(3002),
    [anon_sym_goto] = ACTIONS(3002),
    [anon_sym_return] = ACTIONS(3002),
    [anon_sym_undef] = ACTIONS(3002),
    [anon_sym_PERCENT2] = ACTIONS(3002),
    [anon_sym_AMP2] = ACTIONS(3002),
    [anon_sym_STAR2] = ACTIONS(3002),
    [anon_sym_last] = ACTIONS(3002),
    [anon_sym_next] = ACTIONS(3002),
    [anon_sym_redo] = ACTIONS(3002),
    [anon_sym___FILE__] = ACTIONS(3002),
    [anon_sym___LINE__] = ACTIONS(3002),
    [anon_sym___PACKAGE__] = ACTIONS(3002),
    [anon_sym___SUB__] = ACTIONS(3002),
    [anon_sym_break] = ACTIONS(3002),
    [anon_sym_fork] = ACTIONS(3002),
    [anon_sym_getppid] = ACTIONS(3002),
    [anon_sym_time] = ACTIONS(3002),
    [anon_sym_times] = ACTIONS(3002),
    [anon_sym_wait] = ACTIONS(3002),
    [anon_sym_wantarray] = ACTIONS(3002),
    [anon_sym_abs] = ACTIONS(3002),
    [anon_sym_alarm] = ACTIONS(3002),
    [anon_sym_chop] = ACTIONS(3002),
    [anon_sym_chdir] = ACTIONS(3002),
    [anon_sym_close] = ACTIONS(3002),
    [anon_sym_closedir] = ACTIONS(3002),
    [anon_sym_caller] = ACTIONS(3002),
    [anon_sym_chomp] = ACTIONS(3002),
    [anon_sym_chr] = ACTIONS(3002),
    [anon_sym_cos] = ACTIONS(3002),
    [anon_sym_chroot] = ACTIONS(3002),
    [anon_sym_defined] = ACTIONS(3002),
    [anon_sym_delete] = ACTIONS(3002),
    [anon_sym_dbmclose] = ACTIONS(3002),
    [anon_sym_exists] = ACTIONS(3002),
    [anon_sym_exit] = ACTIONS(3002),
    [anon_sym_eof] = ACTIONS(3002),
    [anon_sym_exp] = ACTIONS(3002),
    [anon_sym_each] = ACTIONS(3002),
    [anon_sym_fc] = ACTIONS(3002),
    [anon_sym_fileno] = ACTIONS(3002),
    [anon_sym_gmtime] = ACTIONS(3002),
    [anon_sym_getc] = ACTIONS(3002),
    [anon_sym_getpgrp] = ACTIONS(3002),
    [anon_sym_getprotobyname] = ACTIONS(3002),
    [anon_sym_getpwname] = ACTIONS(3002),
    [anon_sym_getpwuid] = ACTIONS(3002),
    [anon_sym_getpeername] = ACTIONS(3002),
    [anon_sym_getnetbyname] = ACTIONS(3002),
    [anon_sym_getsockname] = ACTIONS(3002),
    [anon_sym_getgrnam] = ACTIONS(3002),
    [anon_sym_getgrgid] = ACTIONS(3002),
    [anon_sym_hex] = ACTIONS(3002),
    [anon_sym_int] = ACTIONS(3002),
    [anon_sym_keys] = ACTIONS(3002),
    [anon_sym_lc] = ACTIONS(3002),
    [anon_sym_lcfirst] = ACTIONS(3002),
    [anon_sym_length] = ACTIONS(3002),
    [anon_sym_localtime] = ACTIONS(3002),
    [anon_sym_log] = ACTIONS(3002),
    [anon_sym_lock] = ACTIONS(3002),
    [anon_sym_lstat] = ACTIONS(3002),
    [anon_sym_oct] = ACTIONS(3002),
    [anon_sym_ord] = ACTIONS(3002),
    [anon_sym_prototype] = ACTIONS(3002),
    [anon_sym_pop] = ACTIONS(3002),
    [anon_sym_pos] = ACTIONS(3002),
    [anon_sym_quotemeta] = ACTIONS(3002),
    [anon_sym_reset] = ACTIONS(3002),
    [anon_sym_rand] = ACTIONS(3002),
    [anon_sym_rmdir] = ACTIONS(3002),
    [anon_sym_readdir] = ACTIONS(3002),
    [anon_sym_readline] = ACTIONS(3002),
    [anon_sym_readpipe] = ACTIONS(3002),
    [anon_sym_rewinddir] = ACTIONS(3002),
    [anon_sym_readlink] = ACTIONS(3002),
    [anon_sym_ref] = ACTIONS(3002),
    [anon_sym_scalar] = ACTIONS(3002),
    [anon_sym_shift] = ACTIONS(3002),
    [anon_sym_sin] = ACTIONS(3002),
    [anon_sym_sleep] = ACTIONS(3002),
    [anon_sym_sqrt] = ACTIONS(3002),
    [anon_sym_srand] = ACTIONS(3002),
    [anon_sym_stat] = ACTIONS(3002),
    [anon_sym_study] = ACTIONS(3002),
    [anon_sym_tell] = ACTIONS(3002),
    [anon_sym_telldir] = ACTIONS(3002),
    [anon_sym_tied] = ACTIONS(3002),
    [anon_sym_uc] = ACTIONS(3002),
    [anon_sym_ucfirst] = ACTIONS(3002),
    [anon_sym_untie] = ACTIONS(3002),
    [anon_sym_umask] = ACTIONS(3002),
    [anon_sym_values] = ACTIONS(3002),
    [anon_sym_write] = ACTIONS(3002),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3002),
    [anon_sym_qq] = ACTIONS(3002),
    [anon_sym_qw] = ACTIONS(3002),
    [anon_sym_qx] = ACTIONS(3002),
    [anon_sym_qr] = ACTIONS(3002),
    [anon_sym_m] = ACTIONS(3002),
    [anon_sym_s] = ACTIONS(3002),
    [anon_sym_tr] = ACTIONS(3002),
    [anon_sym_y] = ACTIONS(3002),
    [aux_sym__bareword_token1] = ACTIONS(3002),
    [sym_number] = ACTIONS(3002),
    [anon_sym_true] = ACTIONS(3002),
    [anon_sym_false] = ACTIONS(3002),
    [sym__single_quote] = ACTIONS(3000),
    [sym__double_quote] = ACTIONS(3000),
    [sym__backtick_quote] = ACTIONS(3000),
    [sym__search_slash_quote] = ACTIONS(3000),
    [sym__open_readline_bracket] = ACTIONS(3000),
    [sym__open_fileglob_bracket] = ACTIONS(3000),
    [sym__PERLY_HEREDOC] = ACTIONS(3000),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3000),
    [sym__filetest] = ACTIONS(3000),
  },
  [STATE(1650)] = {
    [sym_heredoc_content] = STATE(1650),
    [sym__identifier] = ACTIONS(3006),
    [anon_sym_LBRACE] = ACTIONS(3006),
    [anon_sym_DOLLAR] = ACTIONS(3006),
    [anon_sym_AT] = ACTIONS(3006),
    [anon_sym_RPAREN] = ACTIONS(3006),
    [anon_sym_my] = ACTIONS(3006),
    [anon_sym_extended] = ACTIONS(3006),
    [anon_sym_async] = ACTIONS(3006),
    [anon_sym_sub] = ACTIONS(3006),
    [anon_sym_method] = ACTIONS(3006),
    [anon_sym_LPAREN] = ACTIONS(3006),
    [anon_sym_continue] = ACTIONS(3006),
    [anon_sym_state] = ACTIONS(3006),
    [anon_sym_our] = ACTIONS(3006),
    [anon_sym_LBRACK] = ACTIONS(3006),
    [anon_sym_LT_LT] = ACTIONS(3006),
    [anon_sym_SLASH_SLASH] = ACTIONS(3006),
    [anon_sym_PLUS] = ACTIONS(3006),
    [anon_sym_DASH] = ACTIONS(3006),
    [anon_sym_TILDE] = ACTIONS(3006),
    [anon_sym_BANG] = ACTIONS(3006),
    [anon_sym_PLUS_PLUS] = ACTIONS(3006),
    [anon_sym_DASH_DASH] = ACTIONS(3006),
    [anon_sym_BSLASH] = ACTIONS(3004),
    [anon_sym_do] = ACTIONS(3006),
    [anon_sym_eval] = ACTIONS(3006),
    [anon_sym_await] = ACTIONS(3006),
    [anon_sym_field] = ACTIONS(3006),
    [anon_sym_local] = ACTIONS(3006),
    [anon_sym_dynamically] = ACTIONS(3006),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3006),
    [anon_sym_require] = ACTIONS(3006),
    [anon_sym_map] = ACTIONS(3006),
    [anon_sym_grep] = ACTIONS(3006),
    [anon_sym_sort] = ACTIONS(3006),
    [anon_sym_goto] = ACTIONS(3006),
    [anon_sym_return] = ACTIONS(3006),
    [anon_sym_undef] = ACTIONS(3006),
    [anon_sym_PERCENT2] = ACTIONS(3006),
    [anon_sym_AMP2] = ACTIONS(3006),
    [anon_sym_STAR2] = ACTIONS(3006),
    [anon_sym_last] = ACTIONS(3006),
    [anon_sym_next] = ACTIONS(3006),
    [anon_sym_redo] = ACTIONS(3006),
    [anon_sym___FILE__] = ACTIONS(3006),
    [anon_sym___LINE__] = ACTIONS(3006),
    [anon_sym___PACKAGE__] = ACTIONS(3006),
    [anon_sym___SUB__] = ACTIONS(3006),
    [anon_sym_break] = ACTIONS(3006),
    [anon_sym_fork] = ACTIONS(3006),
    [anon_sym_getppid] = ACTIONS(3006),
    [anon_sym_time] = ACTIONS(3006),
    [anon_sym_times] = ACTIONS(3006),
    [anon_sym_wait] = ACTIONS(3006),
    [anon_sym_wantarray] = ACTIONS(3006),
    [anon_sym_abs] = ACTIONS(3006),
    [anon_sym_alarm] = ACTIONS(3006),
    [anon_sym_chop] = ACTIONS(3006),
    [anon_sym_chdir] = ACTIONS(3006),
    [anon_sym_close] = ACTIONS(3006),
    [anon_sym_closedir] = ACTIONS(3006),
    [anon_sym_caller] = ACTIONS(3006),
    [anon_sym_chomp] = ACTIONS(3006),
    [anon_sym_chr] = ACTIONS(3006),
    [anon_sym_cos] = ACTIONS(3006),
    [anon_sym_chroot] = ACTIONS(3006),
    [anon_sym_defined] = ACTIONS(3006),
    [anon_sym_delete] = ACTIONS(3006),
    [anon_sym_dbmclose] = ACTIONS(3006),
    [anon_sym_exists] = ACTIONS(3006),
    [anon_sym_exit] = ACTIONS(3006),
    [anon_sym_eof] = ACTIONS(3006),
    [anon_sym_exp] = ACTIONS(3006),
    [anon_sym_each] = ACTIONS(3006),
    [anon_sym_fc] = ACTIONS(3006),
    [anon_sym_fileno] = ACTIONS(3006),
    [anon_sym_gmtime] = ACTIONS(3006),
    [anon_sym_getc] = ACTIONS(3006),
    [anon_sym_getpgrp] = ACTIONS(3006),
    [anon_sym_getprotobyname] = ACTIONS(3006),
    [anon_sym_getpwname] = ACTIONS(3006),
    [anon_sym_getpwuid] = ACTIONS(3006),
    [anon_sym_getpeername] = ACTIONS(3006),
    [anon_sym_getnetbyname] = ACTIONS(3006),
    [anon_sym_getsockname] = ACTIONS(3006),
    [anon_sym_getgrnam] = ACTIONS(3006),
    [anon_sym_getgrgid] = ACTIONS(3006),
    [anon_sym_hex] = ACTIONS(3006),
    [anon_sym_int] = ACTIONS(3006),
    [anon_sym_keys] = ACTIONS(3006),
    [anon_sym_lc] = ACTIONS(3006),
    [anon_sym_lcfirst] = ACTIONS(3006),
    [anon_sym_length] = ACTIONS(3006),
    [anon_sym_localtime] = ACTIONS(3006),
    [anon_sym_log] = ACTIONS(3006),
    [anon_sym_lock] = ACTIONS(3006),
    [anon_sym_lstat] = ACTIONS(3006),
    [anon_sym_oct] = ACTIONS(3006),
    [anon_sym_ord] = ACTIONS(3006),
    [anon_sym_prototype] = ACTIONS(3006),
    [anon_sym_pop] = ACTIONS(3006),
    [anon_sym_pos] = ACTIONS(3006),
    [anon_sym_quotemeta] = ACTIONS(3006),
    [anon_sym_reset] = ACTIONS(3006),
    [anon_sym_rand] = ACTIONS(3006),
    [anon_sym_rmdir] = ACTIONS(3006),
    [anon_sym_readdir] = ACTIONS(3006),
    [anon_sym_readline] = ACTIONS(3006),
    [anon_sym_readpipe] = ACTIONS(3006),
    [anon_sym_rewinddir] = ACTIONS(3006),
    [anon_sym_readlink] = ACTIONS(3006),
    [anon_sym_ref] = ACTIONS(3006),
    [anon_sym_scalar] = ACTIONS(3006),
    [anon_sym_shift] = ACTIONS(3006),
    [anon_sym_sin] = ACTIONS(3006),
    [anon_sym_sleep] = ACTIONS(3006),
    [anon_sym_sqrt] = ACTIONS(3006),
    [anon_sym_srand] = ACTIONS(3006),
    [anon_sym_stat] = ACTIONS(3006),
    [anon_sym_study] = ACTIONS(3006),
    [anon_sym_tell] = ACTIONS(3006),
    [anon_sym_telldir] = ACTIONS(3006),
    [anon_sym_tied] = ACTIONS(3006),
    [anon_sym_uc] = ACTIONS(3006),
    [anon_sym_ucfirst] = ACTIONS(3006),
    [anon_sym_untie] = ACTIONS(3006),
    [anon_sym_umask] = ACTIONS(3006),
    [anon_sym_values] = ACTIONS(3006),
    [anon_sym_write] = ACTIONS(3006),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3006),
    [anon_sym_qq] = ACTIONS(3006),
    [anon_sym_qw] = ACTIONS(3006),
    [anon_sym_qx] = ACTIONS(3006),
    [anon_sym_qr] = ACTIONS(3006),
    [anon_sym_m] = ACTIONS(3006),
    [anon_sym_s] = ACTIONS(3006),
    [anon_sym_tr] = ACTIONS(3006),
    [anon_sym_y] = ACTIONS(3006),
    [aux_sym__bareword_token1] = ACTIONS(3006),
    [sym_number] = ACTIONS(3006),
    [anon_sym_true] = ACTIONS(3006),
    [anon_sym_false] = ACTIONS(3006),
    [sym__single_quote] = ACTIONS(3004),
    [sym__double_quote] = ACTIONS(3004),
    [sym__backtick_quote] = ACTIONS(3004),
    [sym__search_slash_quote] = ACTIONS(3004),
    [sym__open_readline_bracket] = ACTIONS(3004),
    [sym__open_fileglob_bracket] = ACTIONS(3004),
    [sym__PERLY_HEREDOC] = ACTIONS(3004),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3004),
    [sym__filetest] = ACTIONS(3004),
  },
  [STATE(1651)] = {
    [sym_heredoc_content] = STATE(1651),
    [sym__identifier] = ACTIONS(3010),
    [anon_sym_LBRACE] = ACTIONS(3010),
    [anon_sym_DOLLAR] = ACTIONS(3010),
    [anon_sym_AT] = ACTIONS(3010),
    [anon_sym_RPAREN] = ACTIONS(3010),
    [anon_sym_my] = ACTIONS(3010),
    [anon_sym_extended] = ACTIONS(3010),
    [anon_sym_async] = ACTIONS(3010),
    [anon_sym_sub] = ACTIONS(3010),
    [anon_sym_method] = ACTIONS(3010),
    [anon_sym_LPAREN] = ACTIONS(3010),
    [anon_sym_continue] = ACTIONS(3010),
    [anon_sym_state] = ACTIONS(3010),
    [anon_sym_our] = ACTIONS(3010),
    [anon_sym_LBRACK] = ACTIONS(3010),
    [anon_sym_LT_LT] = ACTIONS(3010),
    [anon_sym_SLASH_SLASH] = ACTIONS(3010),
    [anon_sym_PLUS] = ACTIONS(3010),
    [anon_sym_DASH] = ACTIONS(3010),
    [anon_sym_TILDE] = ACTIONS(3010),
    [anon_sym_BANG] = ACTIONS(3010),
    [anon_sym_PLUS_PLUS] = ACTIONS(3010),
    [anon_sym_DASH_DASH] = ACTIONS(3010),
    [anon_sym_BSLASH] = ACTIONS(3008),
    [anon_sym_do] = ACTIONS(3010),
    [anon_sym_eval] = ACTIONS(3010),
    [anon_sym_await] = ACTIONS(3010),
    [anon_sym_field] = ACTIONS(3010),
    [anon_sym_local] = ACTIONS(3010),
    [anon_sym_dynamically] = ACTIONS(3010),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3010),
    [anon_sym_require] = ACTIONS(3010),
    [anon_sym_map] = ACTIONS(3010),
    [anon_sym_grep] = ACTIONS(3010),
    [anon_sym_sort] = ACTIONS(3010),
    [anon_sym_goto] = ACTIONS(3010),
    [anon_sym_return] = ACTIONS(3010),
    [anon_sym_undef] = ACTIONS(3010),
    [anon_sym_PERCENT2] = ACTIONS(3010),
    [anon_sym_AMP2] = ACTIONS(3010),
    [anon_sym_STAR2] = ACTIONS(3010),
    [anon_sym_last] = ACTIONS(3010),
    [anon_sym_next] = ACTIONS(3010),
    [anon_sym_redo] = ACTIONS(3010),
    [anon_sym___FILE__] = ACTIONS(3010),
    [anon_sym___LINE__] = ACTIONS(3010),
    [anon_sym___PACKAGE__] = ACTIONS(3010),
    [anon_sym___SUB__] = ACTIONS(3010),
    [anon_sym_break] = ACTIONS(3010),
    [anon_sym_fork] = ACTIONS(3010),
    [anon_sym_getppid] = ACTIONS(3010),
    [anon_sym_time] = ACTIONS(3010),
    [anon_sym_times] = ACTIONS(3010),
    [anon_sym_wait] = ACTIONS(3010),
    [anon_sym_wantarray] = ACTIONS(3010),
    [anon_sym_abs] = ACTIONS(3010),
    [anon_sym_alarm] = ACTIONS(3010),
    [anon_sym_chop] = ACTIONS(3010),
    [anon_sym_chdir] = ACTIONS(3010),
    [anon_sym_close] = ACTIONS(3010),
    [anon_sym_closedir] = ACTIONS(3010),
    [anon_sym_caller] = ACTIONS(3010),
    [anon_sym_chomp] = ACTIONS(3010),
    [anon_sym_chr] = ACTIONS(3010),
    [anon_sym_cos] = ACTIONS(3010),
    [anon_sym_chroot] = ACTIONS(3010),
    [anon_sym_defined] = ACTIONS(3010),
    [anon_sym_delete] = ACTIONS(3010),
    [anon_sym_dbmclose] = ACTIONS(3010),
    [anon_sym_exists] = ACTIONS(3010),
    [anon_sym_exit] = ACTIONS(3010),
    [anon_sym_eof] = ACTIONS(3010),
    [anon_sym_exp] = ACTIONS(3010),
    [anon_sym_each] = ACTIONS(3010),
    [anon_sym_fc] = ACTIONS(3010),
    [anon_sym_fileno] = ACTIONS(3010),
    [anon_sym_gmtime] = ACTIONS(3010),
    [anon_sym_getc] = ACTIONS(3010),
    [anon_sym_getpgrp] = ACTIONS(3010),
    [anon_sym_getprotobyname] = ACTIONS(3010),
    [anon_sym_getpwname] = ACTIONS(3010),
    [anon_sym_getpwuid] = ACTIONS(3010),
    [anon_sym_getpeername] = ACTIONS(3010),
    [anon_sym_getnetbyname] = ACTIONS(3010),
    [anon_sym_getsockname] = ACTIONS(3010),
    [anon_sym_getgrnam] = ACTIONS(3010),
    [anon_sym_getgrgid] = ACTIONS(3010),
    [anon_sym_hex] = ACTIONS(3010),
    [anon_sym_int] = ACTIONS(3010),
    [anon_sym_keys] = ACTIONS(3010),
    [anon_sym_lc] = ACTIONS(3010),
    [anon_sym_lcfirst] = ACTIONS(3010),
    [anon_sym_length] = ACTIONS(3010),
    [anon_sym_localtime] = ACTIONS(3010),
    [anon_sym_log] = ACTIONS(3010),
    [anon_sym_lock] = ACTIONS(3010),
    [anon_sym_lstat] = ACTIONS(3010),
    [anon_sym_oct] = ACTIONS(3010),
    [anon_sym_ord] = ACTIONS(3010),
    [anon_sym_prototype] = ACTIONS(3010),
    [anon_sym_pop] = ACTIONS(3010),
    [anon_sym_pos] = ACTIONS(3010),
    [anon_sym_quotemeta] = ACTIONS(3010),
    [anon_sym_reset] = ACTIONS(3010),
    [anon_sym_rand] = ACTIONS(3010),
    [anon_sym_rmdir] = ACTIONS(3010),
    [anon_sym_readdir] = ACTIONS(3010),
    [anon_sym_readline] = ACTIONS(3010),
    [anon_sym_readpipe] = ACTIONS(3010),
    [anon_sym_rewinddir] = ACTIONS(3010),
    [anon_sym_readlink] = ACTIONS(3010),
    [anon_sym_ref] = ACTIONS(3010),
    [anon_sym_scalar] = ACTIONS(3010),
    [anon_sym_shift] = ACTIONS(3010),
    [anon_sym_sin] = ACTIONS(3010),
    [anon_sym_sleep] = ACTIONS(3010),
    [anon_sym_sqrt] = ACTIONS(3010),
    [anon_sym_srand] = ACTIONS(3010),
    [anon_sym_stat] = ACTIONS(3010),
    [anon_sym_study] = ACTIONS(3010),
    [anon_sym_tell] = ACTIONS(3010),
    [anon_sym_telldir] = ACTIONS(3010),
    [anon_sym_tied] = ACTIONS(3010),
    [anon_sym_uc] = ACTIONS(3010),
    [anon_sym_ucfirst] = ACTIONS(3010),
    [anon_sym_untie] = ACTIONS(3010),
    [anon_sym_umask] = ACTIONS(3010),
    [anon_sym_values] = ACTIONS(3010),
    [anon_sym_write] = ACTIONS(3010),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3010),
    [anon_sym_qq] = ACTIONS(3010),
    [anon_sym_qw] = ACTIONS(3010),
    [anon_sym_qx] = ACTIONS(3010),
    [anon_sym_qr] = ACTIONS(3010),
    [anon_sym_m] = ACTIONS(3010),
    [anon_sym_s] = ACTIONS(3010),
    [anon_sym_tr] = ACTIONS(3010),
    [anon_sym_y] = ACTIONS(3010),
    [aux_sym__bareword_token1] = ACTIONS(3010),
    [sym_number] = ACTIONS(3010),
    [anon_sym_true] = ACTIONS(3010),
    [anon_sym_false] = ACTIONS(3010),
    [sym__single_quote] = ACTIONS(3008),
    [sym__double_quote] = ACTIONS(3008),
    [sym__backtick_quote] = ACTIONS(3008),
    [sym__search_slash_quote] = ACTIONS(3008),
    [sym__open_readline_bracket] = ACTIONS(3008),
    [sym__open_fileglob_bracket] = ACTIONS(3008),
    [sym__PERLY_HEREDOC] = ACTIONS(3008),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3008),
    [sym__filetest] = ACTIONS(3008),
  },
  [STATE(1652)] = {
    [sym_heredoc_content] = STATE(1652),
    [sym__identifier] = ACTIONS(3037),
    [anon_sym_LBRACE] = ACTIONS(3037),
    [anon_sym_DOLLAR] = ACTIONS(3037),
    [anon_sym_AT] = ACTIONS(3037),
    [anon_sym_RPAREN] = ACTIONS(3037),
    [anon_sym_my] = ACTIONS(3037),
    [anon_sym_extended] = ACTIONS(3037),
    [anon_sym_async] = ACTIONS(3037),
    [anon_sym_sub] = ACTIONS(3037),
    [anon_sym_method] = ACTIONS(3037),
    [anon_sym_LPAREN] = ACTIONS(3037),
    [anon_sym_continue] = ACTIONS(3037),
    [anon_sym_state] = ACTIONS(3037),
    [anon_sym_our] = ACTIONS(3037),
    [anon_sym_LBRACK] = ACTIONS(3037),
    [anon_sym_LT_LT] = ACTIONS(3037),
    [anon_sym_SLASH_SLASH] = ACTIONS(3037),
    [anon_sym_PLUS] = ACTIONS(3037),
    [anon_sym_DASH] = ACTIONS(3037),
    [anon_sym_TILDE] = ACTIONS(3037),
    [anon_sym_BANG] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH] = ACTIONS(3037),
    [anon_sym_BSLASH] = ACTIONS(3035),
    [anon_sym_do] = ACTIONS(3037),
    [anon_sym_eval] = ACTIONS(3037),
    [anon_sym_await] = ACTIONS(3037),
    [anon_sym_field] = ACTIONS(3037),
    [anon_sym_local] = ACTIONS(3037),
    [anon_sym_dynamically] = ACTIONS(3037),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3037),
    [anon_sym_require] = ACTIONS(3037),
    [anon_sym_map] = ACTIONS(3037),
    [anon_sym_grep] = ACTIONS(3037),
    [anon_sym_sort] = ACTIONS(3037),
    [anon_sym_goto] = ACTIONS(3037),
    [anon_sym_return] = ACTIONS(3037),
    [anon_sym_undef] = ACTIONS(3037),
    [anon_sym_PERCENT2] = ACTIONS(3037),
    [anon_sym_AMP2] = ACTIONS(3037),
    [anon_sym_STAR2] = ACTIONS(3037),
    [anon_sym_last] = ACTIONS(3037),
    [anon_sym_next] = ACTIONS(3037),
    [anon_sym_redo] = ACTIONS(3037),
    [anon_sym___FILE__] = ACTIONS(3037),
    [anon_sym___LINE__] = ACTIONS(3037),
    [anon_sym___PACKAGE__] = ACTIONS(3037),
    [anon_sym___SUB__] = ACTIONS(3037),
    [anon_sym_break] = ACTIONS(3037),
    [anon_sym_fork] = ACTIONS(3037),
    [anon_sym_getppid] = ACTIONS(3037),
    [anon_sym_time] = ACTIONS(3037),
    [anon_sym_times] = ACTIONS(3037),
    [anon_sym_wait] = ACTIONS(3037),
    [anon_sym_wantarray] = ACTIONS(3037),
    [anon_sym_abs] = ACTIONS(3037),
    [anon_sym_alarm] = ACTIONS(3037),
    [anon_sym_chop] = ACTIONS(3037),
    [anon_sym_chdir] = ACTIONS(3037),
    [anon_sym_close] = ACTIONS(3037),
    [anon_sym_closedir] = ACTIONS(3037),
    [anon_sym_caller] = ACTIONS(3037),
    [anon_sym_chomp] = ACTIONS(3037),
    [anon_sym_chr] = ACTIONS(3037),
    [anon_sym_cos] = ACTIONS(3037),
    [anon_sym_chroot] = ACTIONS(3037),
    [anon_sym_defined] = ACTIONS(3037),
    [anon_sym_delete] = ACTIONS(3037),
    [anon_sym_dbmclose] = ACTIONS(3037),
    [anon_sym_exists] = ACTIONS(3037),
    [anon_sym_exit] = ACTIONS(3037),
    [anon_sym_eof] = ACTIONS(3037),
    [anon_sym_exp] = ACTIONS(3037),
    [anon_sym_each] = ACTIONS(3037),
    [anon_sym_fc] = ACTIONS(3037),
    [anon_sym_fileno] = ACTIONS(3037),
    [anon_sym_gmtime] = ACTIONS(3037),
    [anon_sym_getc] = ACTIONS(3037),
    [anon_sym_getpgrp] = ACTIONS(3037),
    [anon_sym_getprotobyname] = ACTIONS(3037),
    [anon_sym_getpwname] = ACTIONS(3037),
    [anon_sym_getpwuid] = ACTIONS(3037),
    [anon_sym_getpeername] = ACTIONS(3037),
    [anon_sym_getnetbyname] = ACTIONS(3037),
    [anon_sym_getsockname] = ACTIONS(3037),
    [anon_sym_getgrnam] = ACTIONS(3037),
    [anon_sym_getgrgid] = ACTIONS(3037),
    [anon_sym_hex] = ACTIONS(3037),
    [anon_sym_int] = ACTIONS(3037),
    [anon_sym_keys] = ACTIONS(3037),
    [anon_sym_lc] = ACTIONS(3037),
    [anon_sym_lcfirst] = ACTIONS(3037),
    [anon_sym_length] = ACTIONS(3037),
    [anon_sym_localtime] = ACTIONS(3037),
    [anon_sym_log] = ACTIONS(3037),
    [anon_sym_lock] = ACTIONS(3037),
    [anon_sym_lstat] = ACTIONS(3037),
    [anon_sym_oct] = ACTIONS(3037),
    [anon_sym_ord] = ACTIONS(3037),
    [anon_sym_prototype] = ACTIONS(3037),
    [anon_sym_pop] = ACTIONS(3037),
    [anon_sym_pos] = ACTIONS(3037),
    [anon_sym_quotemeta] = ACTIONS(3037),
    [anon_sym_reset] = ACTIONS(3037),
    [anon_sym_rand] = ACTIONS(3037),
    [anon_sym_rmdir] = ACTIONS(3037),
    [anon_sym_readdir] = ACTIONS(3037),
    [anon_sym_readline] = ACTIONS(3037),
    [anon_sym_readpipe] = ACTIONS(3037),
    [anon_sym_rewinddir] = ACTIONS(3037),
    [anon_sym_readlink] = ACTIONS(3037),
    [anon_sym_ref] = ACTIONS(3037),
    [anon_sym_scalar] = ACTIONS(3037),
    [anon_sym_shift] = ACTIONS(3037),
    [anon_sym_sin] = ACTIONS(3037),
    [anon_sym_sleep] = ACTIONS(3037),
    [anon_sym_sqrt] = ACTIONS(3037),
    [anon_sym_srand] = ACTIONS(3037),
    [anon_sym_stat] = ACTIONS(3037),
    [anon_sym_study] = ACTIONS(3037),
    [anon_sym_tell] = ACTIONS(3037),
    [anon_sym_telldir] = ACTIONS(3037),
    [anon_sym_tied] = ACTIONS(3037),
    [anon_sym_uc] = ACTIONS(3037),
    [anon_sym_ucfirst] = ACTIONS(3037),
    [anon_sym_untie] = ACTIONS(3037),
    [anon_sym_umask] = ACTIONS(3037),
    [anon_sym_values] = ACTIONS(3037),
    [anon_sym_write] = ACTIONS(3037),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3037),
    [anon_sym_qq] = ACTIONS(3037),
    [anon_sym_qw] = ACTIONS(3037),
    [anon_sym_qx] = ACTIONS(3037),
    [anon_sym_qr] = ACTIONS(3037),
    [anon_sym_m] = ACTIONS(3037),
    [anon_sym_s] = ACTIONS(3037),
    [anon_sym_tr] = ACTIONS(3037),
    [anon_sym_y] = ACTIONS(3037),
    [aux_sym__bareword_token1] = ACTIONS(3037),
    [sym_number] = ACTIONS(3037),
    [anon_sym_true] = ACTIONS(3037),
    [anon_sym_false] = ACTIONS(3037),
    [sym__single_quote] = ACTIONS(3035),
    [sym__double_quote] = ACTIONS(3035),
    [sym__backtick_quote] = ACTIONS(3035),
    [sym__search_slash_quote] = ACTIONS(3035),
    [sym__open_readline_bracket] = ACTIONS(3035),
    [sym__open_fileglob_bracket] = ACTIONS(3035),
    [sym__PERLY_HEREDOC] = ACTIONS(3035),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3035),
    [sym__filetest] = ACTIONS(3035),
  },
  [STATE(1653)] = {
    [sym_heredoc_content] = STATE(1653),
    [sym__identifier] = ACTIONS(3129),
    [anon_sym_LBRACE] = ACTIONS(3129),
    [anon_sym_DOLLAR] = ACTIONS(3129),
    [anon_sym_AT] = ACTIONS(3129),
    [anon_sym_RPAREN] = ACTIONS(3129),
    [anon_sym_my] = ACTIONS(3129),
    [anon_sym_extended] = ACTIONS(3129),
    [anon_sym_async] = ACTIONS(3129),
    [anon_sym_sub] = ACTIONS(3129),
    [anon_sym_method] = ACTIONS(3129),
    [anon_sym_LPAREN] = ACTIONS(3129),
    [anon_sym_continue] = ACTIONS(3129),
    [anon_sym_state] = ACTIONS(3129),
    [anon_sym_our] = ACTIONS(3129),
    [anon_sym_LBRACK] = ACTIONS(3129),
    [anon_sym_LT_LT] = ACTIONS(3129),
    [anon_sym_SLASH_SLASH] = ACTIONS(3129),
    [anon_sym_PLUS] = ACTIONS(3129),
    [anon_sym_DASH] = ACTIONS(3129),
    [anon_sym_TILDE] = ACTIONS(3129),
    [anon_sym_BANG] = ACTIONS(3129),
    [anon_sym_PLUS_PLUS] = ACTIONS(3129),
    [anon_sym_DASH_DASH] = ACTIONS(3129),
    [anon_sym_BSLASH] = ACTIONS(3127),
    [anon_sym_do] = ACTIONS(3129),
    [anon_sym_eval] = ACTIONS(3129),
    [anon_sym_await] = ACTIONS(3129),
    [anon_sym_field] = ACTIONS(3129),
    [anon_sym_local] = ACTIONS(3129),
    [anon_sym_dynamically] = ACTIONS(3129),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3129),
    [anon_sym_require] = ACTIONS(3129),
    [anon_sym_map] = ACTIONS(3129),
    [anon_sym_grep] = ACTIONS(3129),
    [anon_sym_sort] = ACTIONS(3129),
    [anon_sym_goto] = ACTIONS(3129),
    [anon_sym_return] = ACTIONS(3129),
    [anon_sym_undef] = ACTIONS(3129),
    [anon_sym_PERCENT2] = ACTIONS(3129),
    [anon_sym_AMP2] = ACTIONS(3129),
    [anon_sym_STAR2] = ACTIONS(3129),
    [anon_sym_last] = ACTIONS(3129),
    [anon_sym_next] = ACTIONS(3129),
    [anon_sym_redo] = ACTIONS(3129),
    [anon_sym___FILE__] = ACTIONS(3129),
    [anon_sym___LINE__] = ACTIONS(3129),
    [anon_sym___PACKAGE__] = ACTIONS(3129),
    [anon_sym___SUB__] = ACTIONS(3129),
    [anon_sym_break] = ACTIONS(3129),
    [anon_sym_fork] = ACTIONS(3129),
    [anon_sym_getppid] = ACTIONS(3129),
    [anon_sym_time] = ACTIONS(3129),
    [anon_sym_times] = ACTIONS(3129),
    [anon_sym_wait] = ACTIONS(3129),
    [anon_sym_wantarray] = ACTIONS(3129),
    [anon_sym_abs] = ACTIONS(3129),
    [anon_sym_alarm] = ACTIONS(3129),
    [anon_sym_chop] = ACTIONS(3129),
    [anon_sym_chdir] = ACTIONS(3129),
    [anon_sym_close] = ACTIONS(3129),
    [anon_sym_closedir] = ACTIONS(3129),
    [anon_sym_caller] = ACTIONS(3129),
    [anon_sym_chomp] = ACTIONS(3129),
    [anon_sym_chr] = ACTIONS(3129),
    [anon_sym_cos] = ACTIONS(3129),
    [anon_sym_chroot] = ACTIONS(3129),
    [anon_sym_defined] = ACTIONS(3129),
    [anon_sym_delete] = ACTIONS(3129),
    [anon_sym_dbmclose] = ACTIONS(3129),
    [anon_sym_exists] = ACTIONS(3129),
    [anon_sym_exit] = ACTIONS(3129),
    [anon_sym_eof] = ACTIONS(3129),
    [anon_sym_exp] = ACTIONS(3129),
    [anon_sym_each] = ACTIONS(3129),
    [anon_sym_fc] = ACTIONS(3129),
    [anon_sym_fileno] = ACTIONS(3129),
    [anon_sym_gmtime] = ACTIONS(3129),
    [anon_sym_getc] = ACTIONS(3129),
    [anon_sym_getpgrp] = ACTIONS(3129),
    [anon_sym_getprotobyname] = ACTIONS(3129),
    [anon_sym_getpwname] = ACTIONS(3129),
    [anon_sym_getpwuid] = ACTIONS(3129),
    [anon_sym_getpeername] = ACTIONS(3129),
    [anon_sym_getnetbyname] = ACTIONS(3129),
    [anon_sym_getsockname] = ACTIONS(3129),
    [anon_sym_getgrnam] = ACTIONS(3129),
    [anon_sym_getgrgid] = ACTIONS(3129),
    [anon_sym_hex] = ACTIONS(3129),
    [anon_sym_int] = ACTIONS(3129),
    [anon_sym_keys] = ACTIONS(3129),
    [anon_sym_lc] = ACTIONS(3129),
    [anon_sym_lcfirst] = ACTIONS(3129),
    [anon_sym_length] = ACTIONS(3129),
    [anon_sym_localtime] = ACTIONS(3129),
    [anon_sym_log] = ACTIONS(3129),
    [anon_sym_lock] = ACTIONS(3129),
    [anon_sym_lstat] = ACTIONS(3129),
    [anon_sym_oct] = ACTIONS(3129),
    [anon_sym_ord] = ACTIONS(3129),
    [anon_sym_prototype] = ACTIONS(3129),
    [anon_sym_pop] = ACTIONS(3129),
    [anon_sym_pos] = ACTIONS(3129),
    [anon_sym_quotemeta] = ACTIONS(3129),
    [anon_sym_reset] = ACTIONS(3129),
    [anon_sym_rand] = ACTIONS(3129),
    [anon_sym_rmdir] = ACTIONS(3129),
    [anon_sym_readdir] = ACTIONS(3129),
    [anon_sym_readline] = ACTIONS(3129),
    [anon_sym_readpipe] = ACTIONS(3129),
    [anon_sym_rewinddir] = ACTIONS(3129),
    [anon_sym_readlink] = ACTIONS(3129),
    [anon_sym_ref] = ACTIONS(3129),
    [anon_sym_scalar] = ACTIONS(3129),
    [anon_sym_shift] = ACTIONS(3129),
    [anon_sym_sin] = ACTIONS(3129),
    [anon_sym_sleep] = ACTIONS(3129),
    [anon_sym_sqrt] = ACTIONS(3129),
    [anon_sym_srand] = ACTIONS(3129),
    [anon_sym_stat] = ACTIONS(3129),
    [anon_sym_study] = ACTIONS(3129),
    [anon_sym_tell] = ACTIONS(3129),
    [anon_sym_telldir] = ACTIONS(3129),
    [anon_sym_tied] = ACTIONS(3129),
    [anon_sym_uc] = ACTIONS(3129),
    [anon_sym_ucfirst] = ACTIONS(3129),
    [anon_sym_untie] = ACTIONS(3129),
    [anon_sym_umask] = ACTIONS(3129),
    [anon_sym_values] = ACTIONS(3129),
    [anon_sym_write] = ACTIONS(3129),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3129),
    [anon_sym_qq] = ACTIONS(3129),
    [anon_sym_qw] = ACTIONS(3129),
    [anon_sym_qx] = ACTIONS(3129),
    [anon_sym_qr] = ACTIONS(3129),
    [anon_sym_m] = ACTIONS(3129),
    [anon_sym_s] = ACTIONS(3129),
    [anon_sym_tr] = ACTIONS(3129),
    [anon_sym_y] = ACTIONS(3129),
    [aux_sym__bareword_token1] = ACTIONS(3129),
    [sym_number] = ACTIONS(3129),
    [anon_sym_true] = ACTIONS(3129),
    [anon_sym_false] = ACTIONS(3129),
    [sym__single_quote] = ACTIONS(3127),
    [sym__double_quote] = ACTIONS(3127),
    [sym__backtick_quote] = ACTIONS(3127),
    [sym__search_slash_quote] = ACTIONS(3127),
    [sym__open_readline_bracket] = ACTIONS(3127),
    [sym__open_fileglob_bracket] = ACTIONS(3127),
    [sym__PERLY_HEREDOC] = ACTIONS(3127),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3127),
    [sym__filetest] = ACTIONS(3127),
  },
  [STATE(1654)] = {
    [sym_heredoc_content] = STATE(1654),
    [sym__identifier] = ACTIONS(3237),
    [anon_sym_LBRACE] = ACTIONS(3237),
    [anon_sym_DOLLAR] = ACTIONS(3237),
    [anon_sym_AT] = ACTIONS(3237),
    [anon_sym_RPAREN] = ACTIONS(3237),
    [anon_sym_my] = ACTIONS(3237),
    [anon_sym_extended] = ACTIONS(3237),
    [anon_sym_async] = ACTIONS(3237),
    [anon_sym_sub] = ACTIONS(3237),
    [anon_sym_method] = ACTIONS(3237),
    [anon_sym_LPAREN] = ACTIONS(3237),
    [anon_sym_continue] = ACTIONS(3237),
    [anon_sym_state] = ACTIONS(3237),
    [anon_sym_our] = ACTIONS(3237),
    [anon_sym_LBRACK] = ACTIONS(3237),
    [anon_sym_LT_LT] = ACTIONS(3237),
    [anon_sym_SLASH_SLASH] = ACTIONS(3237),
    [anon_sym_PLUS] = ACTIONS(3237),
    [anon_sym_DASH] = ACTIONS(3237),
    [anon_sym_TILDE] = ACTIONS(3237),
    [anon_sym_BANG] = ACTIONS(3237),
    [anon_sym_PLUS_PLUS] = ACTIONS(3237),
    [anon_sym_DASH_DASH] = ACTIONS(3237),
    [anon_sym_BSLASH] = ACTIONS(3235),
    [anon_sym_do] = ACTIONS(3237),
    [anon_sym_eval] = ACTIONS(3237),
    [anon_sym_await] = ACTIONS(3237),
    [anon_sym_field] = ACTIONS(3237),
    [anon_sym_local] = ACTIONS(3237),
    [anon_sym_dynamically] = ACTIONS(3237),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3237),
    [anon_sym_require] = ACTIONS(3237),
    [anon_sym_map] = ACTIONS(3237),
    [anon_sym_grep] = ACTIONS(3237),
    [anon_sym_sort] = ACTIONS(3237),
    [anon_sym_goto] = ACTIONS(3237),
    [anon_sym_return] = ACTIONS(3237),
    [anon_sym_undef] = ACTIONS(3237),
    [anon_sym_PERCENT2] = ACTIONS(3237),
    [anon_sym_AMP2] = ACTIONS(3237),
    [anon_sym_STAR2] = ACTIONS(3237),
    [anon_sym_last] = ACTIONS(3237),
    [anon_sym_next] = ACTIONS(3237),
    [anon_sym_redo] = ACTIONS(3237),
    [anon_sym___FILE__] = ACTIONS(3237),
    [anon_sym___LINE__] = ACTIONS(3237),
    [anon_sym___PACKAGE__] = ACTIONS(3237),
    [anon_sym___SUB__] = ACTIONS(3237),
    [anon_sym_break] = ACTIONS(3237),
    [anon_sym_fork] = ACTIONS(3237),
    [anon_sym_getppid] = ACTIONS(3237),
    [anon_sym_time] = ACTIONS(3237),
    [anon_sym_times] = ACTIONS(3237),
    [anon_sym_wait] = ACTIONS(3237),
    [anon_sym_wantarray] = ACTIONS(3237),
    [anon_sym_abs] = ACTIONS(3237),
    [anon_sym_alarm] = ACTIONS(3237),
    [anon_sym_chop] = ACTIONS(3237),
    [anon_sym_chdir] = ACTIONS(3237),
    [anon_sym_close] = ACTIONS(3237),
    [anon_sym_closedir] = ACTIONS(3237),
    [anon_sym_caller] = ACTIONS(3237),
    [anon_sym_chomp] = ACTIONS(3237),
    [anon_sym_chr] = ACTIONS(3237),
    [anon_sym_cos] = ACTIONS(3237),
    [anon_sym_chroot] = ACTIONS(3237),
    [anon_sym_defined] = ACTIONS(3237),
    [anon_sym_delete] = ACTIONS(3237),
    [anon_sym_dbmclose] = ACTIONS(3237),
    [anon_sym_exists] = ACTIONS(3237),
    [anon_sym_exit] = ACTIONS(3237),
    [anon_sym_eof] = ACTIONS(3237),
    [anon_sym_exp] = ACTIONS(3237),
    [anon_sym_each] = ACTIONS(3237),
    [anon_sym_fc] = ACTIONS(3237),
    [anon_sym_fileno] = ACTIONS(3237),
    [anon_sym_gmtime] = ACTIONS(3237),
    [anon_sym_getc] = ACTIONS(3237),
    [anon_sym_getpgrp] = ACTIONS(3237),
    [anon_sym_getprotobyname] = ACTIONS(3237),
    [anon_sym_getpwname] = ACTIONS(3237),
    [anon_sym_getpwuid] = ACTIONS(3237),
    [anon_sym_getpeername] = ACTIONS(3237),
    [anon_sym_getnetbyname] = ACTIONS(3237),
    [anon_sym_getsockname] = ACTIONS(3237),
    [anon_sym_getgrnam] = ACTIONS(3237),
    [anon_sym_getgrgid] = ACTIONS(3237),
    [anon_sym_hex] = ACTIONS(3237),
    [anon_sym_int] = ACTIONS(3237),
    [anon_sym_keys] = ACTIONS(3237),
    [anon_sym_lc] = ACTIONS(3237),
    [anon_sym_lcfirst] = ACTIONS(3237),
    [anon_sym_length] = ACTIONS(3237),
    [anon_sym_localtime] = ACTIONS(3237),
    [anon_sym_log] = ACTIONS(3237),
    [anon_sym_lock] = ACTIONS(3237),
    [anon_sym_lstat] = ACTIONS(3237),
    [anon_sym_oct] = ACTIONS(3237),
    [anon_sym_ord] = ACTIONS(3237),
    [anon_sym_prototype] = ACTIONS(3237),
    [anon_sym_pop] = ACTIONS(3237),
    [anon_sym_pos] = ACTIONS(3237),
    [anon_sym_quotemeta] = ACTIONS(3237),
    [anon_sym_reset] = ACTIONS(3237),
    [anon_sym_rand] = ACTIONS(3237),
    [anon_sym_rmdir] = ACTIONS(3237),
    [anon_sym_readdir] = ACTIONS(3237),
    [anon_sym_readline] = ACTIONS(3237),
    [anon_sym_readpipe] = ACTIONS(3237),
    [anon_sym_rewinddir] = ACTIONS(3237),
    [anon_sym_readlink] = ACTIONS(3237),
    [anon_sym_ref] = ACTIONS(3237),
    [anon_sym_scalar] = ACTIONS(3237),
    [anon_sym_shift] = ACTIONS(3237),
    [anon_sym_sin] = ACTIONS(3237),
    [anon_sym_sleep] = ACTIONS(3237),
    [anon_sym_sqrt] = ACTIONS(3237),
    [anon_sym_srand] = ACTIONS(3237),
    [anon_sym_stat] = ACTIONS(3237),
    [anon_sym_study] = ACTIONS(3237),
    [anon_sym_tell] = ACTIONS(3237),
    [anon_sym_telldir] = ACTIONS(3237),
    [anon_sym_tied] = ACTIONS(3237),
    [anon_sym_uc] = ACTIONS(3237),
    [anon_sym_ucfirst] = ACTIONS(3237),
    [anon_sym_untie] = ACTIONS(3237),
    [anon_sym_umask] = ACTIONS(3237),
    [anon_sym_values] = ACTIONS(3237),
    [anon_sym_write] = ACTIONS(3237),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3237),
    [anon_sym_qq] = ACTIONS(3237),
    [anon_sym_qw] = ACTIONS(3237),
    [anon_sym_qx] = ACTIONS(3237),
    [anon_sym_qr] = ACTIONS(3237),
    [anon_sym_m] = ACTIONS(3237),
    [anon_sym_s] = ACTIONS(3237),
    [anon_sym_tr] = ACTIONS(3237),
    [anon_sym_y] = ACTIONS(3237),
    [aux_sym__bareword_token1] = ACTIONS(3237),
    [sym_number] = ACTIONS(3237),
    [anon_sym_true] = ACTIONS(3237),
    [anon_sym_false] = ACTIONS(3237),
    [sym__single_quote] = ACTIONS(3235),
    [sym__double_quote] = ACTIONS(3235),
    [sym__backtick_quote] = ACTIONS(3235),
    [sym__search_slash_quote] = ACTIONS(3235),
    [sym__open_readline_bracket] = ACTIONS(3235),
    [sym__open_fileglob_bracket] = ACTIONS(3235),
    [sym__PERLY_HEREDOC] = ACTIONS(3235),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3235),
    [sym__filetest] = ACTIONS(3235),
  },
  [STATE(1655)] = {
    [sym_heredoc_content] = STATE(1655),
    [sym__identifier] = ACTIONS(3245),
    [anon_sym_LBRACE] = ACTIONS(3245),
    [anon_sym_DOLLAR] = ACTIONS(3245),
    [anon_sym_AT] = ACTIONS(3245),
    [anon_sym_RPAREN] = ACTIONS(3245),
    [anon_sym_my] = ACTIONS(3245),
    [anon_sym_extended] = ACTIONS(3245),
    [anon_sym_async] = ACTIONS(3245),
    [anon_sym_sub] = ACTIONS(3245),
    [anon_sym_method] = ACTIONS(3245),
    [anon_sym_LPAREN] = ACTIONS(3245),
    [anon_sym_continue] = ACTIONS(3245),
    [anon_sym_state] = ACTIONS(3245),
    [anon_sym_our] = ACTIONS(3245),
    [anon_sym_LBRACK] = ACTIONS(3245),
    [anon_sym_LT_LT] = ACTIONS(3245),
    [anon_sym_SLASH_SLASH] = ACTIONS(3245),
    [anon_sym_PLUS] = ACTIONS(3245),
    [anon_sym_DASH] = ACTIONS(3245),
    [anon_sym_TILDE] = ACTIONS(3245),
    [anon_sym_BANG] = ACTIONS(3245),
    [anon_sym_PLUS_PLUS] = ACTIONS(3245),
    [anon_sym_DASH_DASH] = ACTIONS(3245),
    [anon_sym_BSLASH] = ACTIONS(3243),
    [anon_sym_do] = ACTIONS(3245),
    [anon_sym_eval] = ACTIONS(3245),
    [anon_sym_await] = ACTIONS(3245),
    [anon_sym_field] = ACTIONS(3245),
    [anon_sym_local] = ACTIONS(3245),
    [anon_sym_dynamically] = ACTIONS(3245),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3245),
    [anon_sym_require] = ACTIONS(3245),
    [anon_sym_map] = ACTIONS(3245),
    [anon_sym_grep] = ACTIONS(3245),
    [anon_sym_sort] = ACTIONS(3245),
    [anon_sym_goto] = ACTIONS(3245),
    [anon_sym_return] = ACTIONS(3245),
    [anon_sym_undef] = ACTIONS(3245),
    [anon_sym_PERCENT2] = ACTIONS(3245),
    [anon_sym_AMP2] = ACTIONS(3245),
    [anon_sym_STAR2] = ACTIONS(3245),
    [anon_sym_last] = ACTIONS(3245),
    [anon_sym_next] = ACTIONS(3245),
    [anon_sym_redo] = ACTIONS(3245),
    [anon_sym___FILE__] = ACTIONS(3245),
    [anon_sym___LINE__] = ACTIONS(3245),
    [anon_sym___PACKAGE__] = ACTIONS(3245),
    [anon_sym___SUB__] = ACTIONS(3245),
    [anon_sym_break] = ACTIONS(3245),
    [anon_sym_fork] = ACTIONS(3245),
    [anon_sym_getppid] = ACTIONS(3245),
    [anon_sym_time] = ACTIONS(3245),
    [anon_sym_times] = ACTIONS(3245),
    [anon_sym_wait] = ACTIONS(3245),
    [anon_sym_wantarray] = ACTIONS(3245),
    [anon_sym_abs] = ACTIONS(3245),
    [anon_sym_alarm] = ACTIONS(3245),
    [anon_sym_chop] = ACTIONS(3245),
    [anon_sym_chdir] = ACTIONS(3245),
    [anon_sym_close] = ACTIONS(3245),
    [anon_sym_closedir] = ACTIONS(3245),
    [anon_sym_caller] = ACTIONS(3245),
    [anon_sym_chomp] = ACTIONS(3245),
    [anon_sym_chr] = ACTIONS(3245),
    [anon_sym_cos] = ACTIONS(3245),
    [anon_sym_chroot] = ACTIONS(3245),
    [anon_sym_defined] = ACTIONS(3245),
    [anon_sym_delete] = ACTIONS(3245),
    [anon_sym_dbmclose] = ACTIONS(3245),
    [anon_sym_exists] = ACTIONS(3245),
    [anon_sym_exit] = ACTIONS(3245),
    [anon_sym_eof] = ACTIONS(3245),
    [anon_sym_exp] = ACTIONS(3245),
    [anon_sym_each] = ACTIONS(3245),
    [anon_sym_fc] = ACTIONS(3245),
    [anon_sym_fileno] = ACTIONS(3245),
    [anon_sym_gmtime] = ACTIONS(3245),
    [anon_sym_getc] = ACTIONS(3245),
    [anon_sym_getpgrp] = ACTIONS(3245),
    [anon_sym_getprotobyname] = ACTIONS(3245),
    [anon_sym_getpwname] = ACTIONS(3245),
    [anon_sym_getpwuid] = ACTIONS(3245),
    [anon_sym_getpeername] = ACTIONS(3245),
    [anon_sym_getnetbyname] = ACTIONS(3245),
    [anon_sym_getsockname] = ACTIONS(3245),
    [anon_sym_getgrnam] = ACTIONS(3245),
    [anon_sym_getgrgid] = ACTIONS(3245),
    [anon_sym_hex] = ACTIONS(3245),
    [anon_sym_int] = ACTIONS(3245),
    [anon_sym_keys] = ACTIONS(3245),
    [anon_sym_lc] = ACTIONS(3245),
    [anon_sym_lcfirst] = ACTIONS(3245),
    [anon_sym_length] = ACTIONS(3245),
    [anon_sym_localtime] = ACTIONS(3245),
    [anon_sym_log] = ACTIONS(3245),
    [anon_sym_lock] = ACTIONS(3245),
    [anon_sym_lstat] = ACTIONS(3245),
    [anon_sym_oct] = ACTIONS(3245),
    [anon_sym_ord] = ACTIONS(3245),
    [anon_sym_prototype] = ACTIONS(3245),
    [anon_sym_pop] = ACTIONS(3245),
    [anon_sym_pos] = ACTIONS(3245),
    [anon_sym_quotemeta] = ACTIONS(3245),
    [anon_sym_reset] = ACTIONS(3245),
    [anon_sym_rand] = ACTIONS(3245),
    [anon_sym_rmdir] = ACTIONS(3245),
    [anon_sym_readdir] = ACTIONS(3245),
    [anon_sym_readline] = ACTIONS(3245),
    [anon_sym_readpipe] = ACTIONS(3245),
    [anon_sym_rewinddir] = ACTIONS(3245),
    [anon_sym_readlink] = ACTIONS(3245),
    [anon_sym_ref] = ACTIONS(3245),
    [anon_sym_scalar] = ACTIONS(3245),
    [anon_sym_shift] = ACTIONS(3245),
    [anon_sym_sin] = ACTIONS(3245),
    [anon_sym_sleep] = ACTIONS(3245),
    [anon_sym_sqrt] = ACTIONS(3245),
    [anon_sym_srand] = ACTIONS(3245),
    [anon_sym_stat] = ACTIONS(3245),
    [anon_sym_study] = ACTIONS(3245),
    [anon_sym_tell] = ACTIONS(3245),
    [anon_sym_telldir] = ACTIONS(3245),
    [anon_sym_tied] = ACTIONS(3245),
    [anon_sym_uc] = ACTIONS(3245),
    [anon_sym_ucfirst] = ACTIONS(3245),
    [anon_sym_untie] = ACTIONS(3245),
    [anon_sym_umask] = ACTIONS(3245),
    [anon_sym_values] = ACTIONS(3245),
    [anon_sym_write] = ACTIONS(3245),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3245),
    [anon_sym_qq] = ACTIONS(3245),
    [anon_sym_qw] = ACTIONS(3245),
    [anon_sym_qx] = ACTIONS(3245),
    [anon_sym_qr] = ACTIONS(3245),
    [anon_sym_m] = ACTIONS(3245),
    [anon_sym_s] = ACTIONS(3245),
    [anon_sym_tr] = ACTIONS(3245),
    [anon_sym_y] = ACTIONS(3245),
    [aux_sym__bareword_token1] = ACTIONS(3245),
    [sym_number] = ACTIONS(3245),
    [anon_sym_true] = ACTIONS(3245),
    [anon_sym_false] = ACTIONS(3245),
    [sym__single_quote] = ACTIONS(3243),
    [sym__double_quote] = ACTIONS(3243),
    [sym__backtick_quote] = ACTIONS(3243),
    [sym__search_slash_quote] = ACTIONS(3243),
    [sym__open_readline_bracket] = ACTIONS(3243),
    [sym__open_fileglob_bracket] = ACTIONS(3243),
    [sym__PERLY_HEREDOC] = ACTIONS(3243),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3243),
    [sym__filetest] = ACTIONS(3243),
  },
  [STATE(1656)] = {
    [sym_heredoc_content] = STATE(1656),
    [sym__identifier] = ACTIONS(3249),
    [anon_sym_LBRACE] = ACTIONS(3249),
    [anon_sym_DOLLAR] = ACTIONS(3249),
    [anon_sym_AT] = ACTIONS(3249),
    [anon_sym_RPAREN] = ACTIONS(3249),
    [anon_sym_my] = ACTIONS(3249),
    [anon_sym_extended] = ACTIONS(3249),
    [anon_sym_async] = ACTIONS(3249),
    [anon_sym_sub] = ACTIONS(3249),
    [anon_sym_method] = ACTIONS(3249),
    [anon_sym_LPAREN] = ACTIONS(3249),
    [anon_sym_continue] = ACTIONS(3249),
    [anon_sym_state] = ACTIONS(3249),
    [anon_sym_our] = ACTIONS(3249),
    [anon_sym_LBRACK] = ACTIONS(3249),
    [anon_sym_LT_LT] = ACTIONS(3249),
    [anon_sym_SLASH_SLASH] = ACTIONS(3249),
    [anon_sym_PLUS] = ACTIONS(3249),
    [anon_sym_DASH] = ACTIONS(3249),
    [anon_sym_TILDE] = ACTIONS(3249),
    [anon_sym_BANG] = ACTIONS(3249),
    [anon_sym_PLUS_PLUS] = ACTIONS(3249),
    [anon_sym_DASH_DASH] = ACTIONS(3249),
    [anon_sym_BSLASH] = ACTIONS(3247),
    [anon_sym_do] = ACTIONS(3249),
    [anon_sym_eval] = ACTIONS(3249),
    [anon_sym_await] = ACTIONS(3249),
    [anon_sym_field] = ACTIONS(3249),
    [anon_sym_local] = ACTIONS(3249),
    [anon_sym_dynamically] = ACTIONS(3249),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3249),
    [anon_sym_require] = ACTIONS(3249),
    [anon_sym_map] = ACTIONS(3249),
    [anon_sym_grep] = ACTIONS(3249),
    [anon_sym_sort] = ACTIONS(3249),
    [anon_sym_goto] = ACTIONS(3249),
    [anon_sym_return] = ACTIONS(3249),
    [anon_sym_undef] = ACTIONS(3249),
    [anon_sym_PERCENT2] = ACTIONS(3249),
    [anon_sym_AMP2] = ACTIONS(3249),
    [anon_sym_STAR2] = ACTIONS(3249),
    [anon_sym_last] = ACTIONS(3249),
    [anon_sym_next] = ACTIONS(3249),
    [anon_sym_redo] = ACTIONS(3249),
    [anon_sym___FILE__] = ACTIONS(3249),
    [anon_sym___LINE__] = ACTIONS(3249),
    [anon_sym___PACKAGE__] = ACTIONS(3249),
    [anon_sym___SUB__] = ACTIONS(3249),
    [anon_sym_break] = ACTIONS(3249),
    [anon_sym_fork] = ACTIONS(3249),
    [anon_sym_getppid] = ACTIONS(3249),
    [anon_sym_time] = ACTIONS(3249),
    [anon_sym_times] = ACTIONS(3249),
    [anon_sym_wait] = ACTIONS(3249),
    [anon_sym_wantarray] = ACTIONS(3249),
    [anon_sym_abs] = ACTIONS(3249),
    [anon_sym_alarm] = ACTIONS(3249),
    [anon_sym_chop] = ACTIONS(3249),
    [anon_sym_chdir] = ACTIONS(3249),
    [anon_sym_close] = ACTIONS(3249),
    [anon_sym_closedir] = ACTIONS(3249),
    [anon_sym_caller] = ACTIONS(3249),
    [anon_sym_chomp] = ACTIONS(3249),
    [anon_sym_chr] = ACTIONS(3249),
    [anon_sym_cos] = ACTIONS(3249),
    [anon_sym_chroot] = ACTIONS(3249),
    [anon_sym_defined] = ACTIONS(3249),
    [anon_sym_delete] = ACTIONS(3249),
    [anon_sym_dbmclose] = ACTIONS(3249),
    [anon_sym_exists] = ACTIONS(3249),
    [anon_sym_exit] = ACTIONS(3249),
    [anon_sym_eof] = ACTIONS(3249),
    [anon_sym_exp] = ACTIONS(3249),
    [anon_sym_each] = ACTIONS(3249),
    [anon_sym_fc] = ACTIONS(3249),
    [anon_sym_fileno] = ACTIONS(3249),
    [anon_sym_gmtime] = ACTIONS(3249),
    [anon_sym_getc] = ACTIONS(3249),
    [anon_sym_getpgrp] = ACTIONS(3249),
    [anon_sym_getprotobyname] = ACTIONS(3249),
    [anon_sym_getpwname] = ACTIONS(3249),
    [anon_sym_getpwuid] = ACTIONS(3249),
    [anon_sym_getpeername] = ACTIONS(3249),
    [anon_sym_getnetbyname] = ACTIONS(3249),
    [anon_sym_getsockname] = ACTIONS(3249),
    [anon_sym_getgrnam] = ACTIONS(3249),
    [anon_sym_getgrgid] = ACTIONS(3249),
    [anon_sym_hex] = ACTIONS(3249),
    [anon_sym_int] = ACTIONS(3249),
    [anon_sym_keys] = ACTIONS(3249),
    [anon_sym_lc] = ACTIONS(3249),
    [anon_sym_lcfirst] = ACTIONS(3249),
    [anon_sym_length] = ACTIONS(3249),
    [anon_sym_localtime] = ACTIONS(3249),
    [anon_sym_log] = ACTIONS(3249),
    [anon_sym_lock] = ACTIONS(3249),
    [anon_sym_lstat] = ACTIONS(3249),
    [anon_sym_oct] = ACTIONS(3249),
    [anon_sym_ord] = ACTIONS(3249),
    [anon_sym_prototype] = ACTIONS(3249),
    [anon_sym_pop] = ACTIONS(3249),
    [anon_sym_pos] = ACTIONS(3249),
    [anon_sym_quotemeta] = ACTIONS(3249),
    [anon_sym_reset] = ACTIONS(3249),
    [anon_sym_rand] = ACTIONS(3249),
    [anon_sym_rmdir] = ACTIONS(3249),
    [anon_sym_readdir] = ACTIONS(3249),
    [anon_sym_readline] = ACTIONS(3249),
    [anon_sym_readpipe] = ACTIONS(3249),
    [anon_sym_rewinddir] = ACTIONS(3249),
    [anon_sym_readlink] = ACTIONS(3249),
    [anon_sym_ref] = ACTIONS(3249),
    [anon_sym_scalar] = ACTIONS(3249),
    [anon_sym_shift] = ACTIONS(3249),
    [anon_sym_sin] = ACTIONS(3249),
    [anon_sym_sleep] = ACTIONS(3249),
    [anon_sym_sqrt] = ACTIONS(3249),
    [anon_sym_srand] = ACTIONS(3249),
    [anon_sym_stat] = ACTIONS(3249),
    [anon_sym_study] = ACTIONS(3249),
    [anon_sym_tell] = ACTIONS(3249),
    [anon_sym_telldir] = ACTIONS(3249),
    [anon_sym_tied] = ACTIONS(3249),
    [anon_sym_uc] = ACTIONS(3249),
    [anon_sym_ucfirst] = ACTIONS(3249),
    [anon_sym_untie] = ACTIONS(3249),
    [anon_sym_umask] = ACTIONS(3249),
    [anon_sym_values] = ACTIONS(3249),
    [anon_sym_write] = ACTIONS(3249),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3249),
    [anon_sym_qq] = ACTIONS(3249),
    [anon_sym_qw] = ACTIONS(3249),
    [anon_sym_qx] = ACTIONS(3249),
    [anon_sym_qr] = ACTIONS(3249),
    [anon_sym_m] = ACTIONS(3249),
    [anon_sym_s] = ACTIONS(3249),
    [anon_sym_tr] = ACTIONS(3249),
    [anon_sym_y] = ACTIONS(3249),
    [aux_sym__bareword_token1] = ACTIONS(3249),
    [sym_number] = ACTIONS(3249),
    [anon_sym_true] = ACTIONS(3249),
    [anon_sym_false] = ACTIONS(3249),
    [sym__single_quote] = ACTIONS(3247),
    [sym__double_quote] = ACTIONS(3247),
    [sym__backtick_quote] = ACTIONS(3247),
    [sym__search_slash_quote] = ACTIONS(3247),
    [sym__open_readline_bracket] = ACTIONS(3247),
    [sym__open_fileglob_bracket] = ACTIONS(3247),
    [sym__PERLY_HEREDOC] = ACTIONS(3247),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3247),
    [sym__filetest] = ACTIONS(3247),
  },
  [STATE(1657)] = {
    [sym_heredoc_content] = STATE(1657),
    [sym__identifier] = ACTIONS(3145),
    [anon_sym_LBRACE] = ACTIONS(3145),
    [anon_sym_DOLLAR] = ACTIONS(3145),
    [anon_sym_AT] = ACTIONS(3145),
    [anon_sym_RPAREN] = ACTIONS(3145),
    [anon_sym_my] = ACTIONS(3145),
    [anon_sym_extended] = ACTIONS(3145),
    [anon_sym_async] = ACTIONS(3145),
    [anon_sym_sub] = ACTIONS(3145),
    [anon_sym_method] = ACTIONS(3145),
    [anon_sym_LPAREN] = ACTIONS(3145),
    [anon_sym_continue] = ACTIONS(3145),
    [anon_sym_state] = ACTIONS(3145),
    [anon_sym_our] = ACTIONS(3145),
    [anon_sym_LBRACK] = ACTIONS(3145),
    [anon_sym_LT_LT] = ACTIONS(3145),
    [anon_sym_SLASH_SLASH] = ACTIONS(3145),
    [anon_sym_PLUS] = ACTIONS(3145),
    [anon_sym_DASH] = ACTIONS(3145),
    [anon_sym_TILDE] = ACTIONS(3145),
    [anon_sym_BANG] = ACTIONS(3145),
    [anon_sym_PLUS_PLUS] = ACTIONS(3145),
    [anon_sym_DASH_DASH] = ACTIONS(3145),
    [anon_sym_BSLASH] = ACTIONS(3143),
    [anon_sym_do] = ACTIONS(3145),
    [anon_sym_eval] = ACTIONS(3145),
    [anon_sym_await] = ACTIONS(3145),
    [anon_sym_field] = ACTIONS(3145),
    [anon_sym_local] = ACTIONS(3145),
    [anon_sym_dynamically] = ACTIONS(3145),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3145),
    [anon_sym_require] = ACTIONS(3145),
    [anon_sym_map] = ACTIONS(3145),
    [anon_sym_grep] = ACTIONS(3145),
    [anon_sym_sort] = ACTIONS(3145),
    [anon_sym_goto] = ACTIONS(3145),
    [anon_sym_return] = ACTIONS(3145),
    [anon_sym_undef] = ACTIONS(3145),
    [anon_sym_PERCENT2] = ACTIONS(3145),
    [anon_sym_AMP2] = ACTIONS(3145),
    [anon_sym_STAR2] = ACTIONS(3145),
    [anon_sym_last] = ACTIONS(3145),
    [anon_sym_next] = ACTIONS(3145),
    [anon_sym_redo] = ACTIONS(3145),
    [anon_sym___FILE__] = ACTIONS(3145),
    [anon_sym___LINE__] = ACTIONS(3145),
    [anon_sym___PACKAGE__] = ACTIONS(3145),
    [anon_sym___SUB__] = ACTIONS(3145),
    [anon_sym_break] = ACTIONS(3145),
    [anon_sym_fork] = ACTIONS(3145),
    [anon_sym_getppid] = ACTIONS(3145),
    [anon_sym_time] = ACTIONS(3145),
    [anon_sym_times] = ACTIONS(3145),
    [anon_sym_wait] = ACTIONS(3145),
    [anon_sym_wantarray] = ACTIONS(3145),
    [anon_sym_abs] = ACTIONS(3145),
    [anon_sym_alarm] = ACTIONS(3145),
    [anon_sym_chop] = ACTIONS(3145),
    [anon_sym_chdir] = ACTIONS(3145),
    [anon_sym_close] = ACTIONS(3145),
    [anon_sym_closedir] = ACTIONS(3145),
    [anon_sym_caller] = ACTIONS(3145),
    [anon_sym_chomp] = ACTIONS(3145),
    [anon_sym_chr] = ACTIONS(3145),
    [anon_sym_cos] = ACTIONS(3145),
    [anon_sym_chroot] = ACTIONS(3145),
    [anon_sym_defined] = ACTIONS(3145),
    [anon_sym_delete] = ACTIONS(3145),
    [anon_sym_dbmclose] = ACTIONS(3145),
    [anon_sym_exists] = ACTIONS(3145),
    [anon_sym_exit] = ACTIONS(3145),
    [anon_sym_eof] = ACTIONS(3145),
    [anon_sym_exp] = ACTIONS(3145),
    [anon_sym_each] = ACTIONS(3145),
    [anon_sym_fc] = ACTIONS(3145),
    [anon_sym_fileno] = ACTIONS(3145),
    [anon_sym_gmtime] = ACTIONS(3145),
    [anon_sym_getc] = ACTIONS(3145),
    [anon_sym_getpgrp] = ACTIONS(3145),
    [anon_sym_getprotobyname] = ACTIONS(3145),
    [anon_sym_getpwname] = ACTIONS(3145),
    [anon_sym_getpwuid] = ACTIONS(3145),
    [anon_sym_getpeername] = ACTIONS(3145),
    [anon_sym_getnetbyname] = ACTIONS(3145),
    [anon_sym_getsockname] = ACTIONS(3145),
    [anon_sym_getgrnam] = ACTIONS(3145),
    [anon_sym_getgrgid] = ACTIONS(3145),
    [anon_sym_hex] = ACTIONS(3145),
    [anon_sym_int] = ACTIONS(3145),
    [anon_sym_keys] = ACTIONS(3145),
    [anon_sym_lc] = ACTIONS(3145),
    [anon_sym_lcfirst] = ACTIONS(3145),
    [anon_sym_length] = ACTIONS(3145),
    [anon_sym_localtime] = ACTIONS(3145),
    [anon_sym_log] = ACTIONS(3145),
    [anon_sym_lock] = ACTIONS(3145),
    [anon_sym_lstat] = ACTIONS(3145),
    [anon_sym_oct] = ACTIONS(3145),
    [anon_sym_ord] = ACTIONS(3145),
    [anon_sym_prototype] = ACTIONS(3145),
    [anon_sym_pop] = ACTIONS(3145),
    [anon_sym_pos] = ACTIONS(3145),
    [anon_sym_quotemeta] = ACTIONS(3145),
    [anon_sym_reset] = ACTIONS(3145),
    [anon_sym_rand] = ACTIONS(3145),
    [anon_sym_rmdir] = ACTIONS(3145),
    [anon_sym_readdir] = ACTIONS(3145),
    [anon_sym_readline] = ACTIONS(3145),
    [anon_sym_readpipe] = ACTIONS(3145),
    [anon_sym_rewinddir] = ACTIONS(3145),
    [anon_sym_readlink] = ACTIONS(3145),
    [anon_sym_ref] = ACTIONS(3145),
    [anon_sym_scalar] = ACTIONS(3145),
    [anon_sym_shift] = ACTIONS(3145),
    [anon_sym_sin] = ACTIONS(3145),
    [anon_sym_sleep] = ACTIONS(3145),
    [anon_sym_sqrt] = ACTIONS(3145),
    [anon_sym_srand] = ACTIONS(3145),
    [anon_sym_stat] = ACTIONS(3145),
    [anon_sym_study] = ACTIONS(3145),
    [anon_sym_tell] = ACTIONS(3145),
    [anon_sym_telldir] = ACTIONS(3145),
    [anon_sym_tied] = ACTIONS(3145),
    [anon_sym_uc] = ACTIONS(3145),
    [anon_sym_ucfirst] = ACTIONS(3145),
    [anon_sym_untie] = ACTIONS(3145),
    [anon_sym_umask] = ACTIONS(3145),
    [anon_sym_values] = ACTIONS(3145),
    [anon_sym_write] = ACTIONS(3145),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3145),
    [anon_sym_qq] = ACTIONS(3145),
    [anon_sym_qw] = ACTIONS(3145),
    [anon_sym_qx] = ACTIONS(3145),
    [anon_sym_qr] = ACTIONS(3145),
    [anon_sym_m] = ACTIONS(3145),
    [anon_sym_s] = ACTIONS(3145),
    [anon_sym_tr] = ACTIONS(3145),
    [anon_sym_y] = ACTIONS(3145),
    [aux_sym__bareword_token1] = ACTIONS(3145),
    [sym_number] = ACTIONS(3145),
    [anon_sym_true] = ACTIONS(3145),
    [anon_sym_false] = ACTIONS(3145),
    [sym__single_quote] = ACTIONS(3143),
    [sym__double_quote] = ACTIONS(3143),
    [sym__backtick_quote] = ACTIONS(3143),
    [sym__search_slash_quote] = ACTIONS(3143),
    [sym__open_readline_bracket] = ACTIONS(3143),
    [sym__open_fileglob_bracket] = ACTIONS(3143),
    [sym__PERLY_HEREDOC] = ACTIONS(3143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3143),
    [sym__filetest] = ACTIONS(3143),
  },
  [STATE(1658)] = {
    [sym_heredoc_content] = STATE(1658),
    [sym__identifier] = ACTIONS(2926),
    [anon_sym_LBRACE] = ACTIONS(2926),
    [anon_sym_DOLLAR] = ACTIONS(2926),
    [anon_sym_AT] = ACTIONS(2926),
    [anon_sym_RPAREN] = ACTIONS(2926),
    [anon_sym_my] = ACTIONS(2926),
    [anon_sym_extended] = ACTIONS(2926),
    [anon_sym_async] = ACTIONS(2926),
    [anon_sym_sub] = ACTIONS(2926),
    [anon_sym_method] = ACTIONS(2926),
    [anon_sym_LPAREN] = ACTIONS(2926),
    [anon_sym_continue] = ACTIONS(2926),
    [anon_sym_state] = ACTIONS(2926),
    [anon_sym_our] = ACTIONS(2926),
    [anon_sym_LBRACK] = ACTIONS(2926),
    [anon_sym_LT_LT] = ACTIONS(2926),
    [anon_sym_SLASH_SLASH] = ACTIONS(2926),
    [anon_sym_PLUS] = ACTIONS(2926),
    [anon_sym_DASH] = ACTIONS(2926),
    [anon_sym_TILDE] = ACTIONS(2926),
    [anon_sym_BANG] = ACTIONS(2926),
    [anon_sym_PLUS_PLUS] = ACTIONS(2926),
    [anon_sym_DASH_DASH] = ACTIONS(2926),
    [anon_sym_BSLASH] = ACTIONS(2924),
    [anon_sym_do] = ACTIONS(2926),
    [anon_sym_eval] = ACTIONS(2926),
    [anon_sym_await] = ACTIONS(2926),
    [anon_sym_field] = ACTIONS(2926),
    [anon_sym_local] = ACTIONS(2926),
    [anon_sym_dynamically] = ACTIONS(2926),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2926),
    [anon_sym_require] = ACTIONS(2926),
    [anon_sym_map] = ACTIONS(2926),
    [anon_sym_grep] = ACTIONS(2926),
    [anon_sym_sort] = ACTIONS(2926),
    [anon_sym_goto] = ACTIONS(2926),
    [anon_sym_return] = ACTIONS(2926),
    [anon_sym_undef] = ACTIONS(2926),
    [anon_sym_PERCENT2] = ACTIONS(2926),
    [anon_sym_AMP2] = ACTIONS(2926),
    [anon_sym_STAR2] = ACTIONS(2926),
    [anon_sym_last] = ACTIONS(2926),
    [anon_sym_next] = ACTIONS(2926),
    [anon_sym_redo] = ACTIONS(2926),
    [anon_sym___FILE__] = ACTIONS(2926),
    [anon_sym___LINE__] = ACTIONS(2926),
    [anon_sym___PACKAGE__] = ACTIONS(2926),
    [anon_sym___SUB__] = ACTIONS(2926),
    [anon_sym_break] = ACTIONS(2926),
    [anon_sym_fork] = ACTIONS(2926),
    [anon_sym_getppid] = ACTIONS(2926),
    [anon_sym_time] = ACTIONS(2926),
    [anon_sym_times] = ACTIONS(2926),
    [anon_sym_wait] = ACTIONS(2926),
    [anon_sym_wantarray] = ACTIONS(2926),
    [anon_sym_abs] = ACTIONS(2926),
    [anon_sym_alarm] = ACTIONS(2926),
    [anon_sym_chop] = ACTIONS(2926),
    [anon_sym_chdir] = ACTIONS(2926),
    [anon_sym_close] = ACTIONS(2926),
    [anon_sym_closedir] = ACTIONS(2926),
    [anon_sym_caller] = ACTIONS(2926),
    [anon_sym_chomp] = ACTIONS(2926),
    [anon_sym_chr] = ACTIONS(2926),
    [anon_sym_cos] = ACTIONS(2926),
    [anon_sym_chroot] = ACTIONS(2926),
    [anon_sym_defined] = ACTIONS(2926),
    [anon_sym_delete] = ACTIONS(2926),
    [anon_sym_dbmclose] = ACTIONS(2926),
    [anon_sym_exists] = ACTIONS(2926),
    [anon_sym_exit] = ACTIONS(2926),
    [anon_sym_eof] = ACTIONS(2926),
    [anon_sym_exp] = ACTIONS(2926),
    [anon_sym_each] = ACTIONS(2926),
    [anon_sym_fc] = ACTIONS(2926),
    [anon_sym_fileno] = ACTIONS(2926),
    [anon_sym_gmtime] = ACTIONS(2926),
    [anon_sym_getc] = ACTIONS(2926),
    [anon_sym_getpgrp] = ACTIONS(2926),
    [anon_sym_getprotobyname] = ACTIONS(2926),
    [anon_sym_getpwname] = ACTIONS(2926),
    [anon_sym_getpwuid] = ACTIONS(2926),
    [anon_sym_getpeername] = ACTIONS(2926),
    [anon_sym_getnetbyname] = ACTIONS(2926),
    [anon_sym_getsockname] = ACTIONS(2926),
    [anon_sym_getgrnam] = ACTIONS(2926),
    [anon_sym_getgrgid] = ACTIONS(2926),
    [anon_sym_hex] = ACTIONS(2926),
    [anon_sym_int] = ACTIONS(2926),
    [anon_sym_keys] = ACTIONS(2926),
    [anon_sym_lc] = ACTIONS(2926),
    [anon_sym_lcfirst] = ACTIONS(2926),
    [anon_sym_length] = ACTIONS(2926),
    [anon_sym_localtime] = ACTIONS(2926),
    [anon_sym_log] = ACTIONS(2926),
    [anon_sym_lock] = ACTIONS(2926),
    [anon_sym_lstat] = ACTIONS(2926),
    [anon_sym_oct] = ACTIONS(2926),
    [anon_sym_ord] = ACTIONS(2926),
    [anon_sym_prototype] = ACTIONS(2926),
    [anon_sym_pop] = ACTIONS(2926),
    [anon_sym_pos] = ACTIONS(2926),
    [anon_sym_quotemeta] = ACTIONS(2926),
    [anon_sym_reset] = ACTIONS(2926),
    [anon_sym_rand] = ACTIONS(2926),
    [anon_sym_rmdir] = ACTIONS(2926),
    [anon_sym_readdir] = ACTIONS(2926),
    [anon_sym_readline] = ACTIONS(2926),
    [anon_sym_readpipe] = ACTIONS(2926),
    [anon_sym_rewinddir] = ACTIONS(2926),
    [anon_sym_readlink] = ACTIONS(2926),
    [anon_sym_ref] = ACTIONS(2926),
    [anon_sym_scalar] = ACTIONS(2926),
    [anon_sym_shift] = ACTIONS(2926),
    [anon_sym_sin] = ACTIONS(2926),
    [anon_sym_sleep] = ACTIONS(2926),
    [anon_sym_sqrt] = ACTIONS(2926),
    [anon_sym_srand] = ACTIONS(2926),
    [anon_sym_stat] = ACTIONS(2926),
    [anon_sym_study] = ACTIONS(2926),
    [anon_sym_tell] = ACTIONS(2926),
    [anon_sym_telldir] = ACTIONS(2926),
    [anon_sym_tied] = ACTIONS(2926),
    [anon_sym_uc] = ACTIONS(2926),
    [anon_sym_ucfirst] = ACTIONS(2926),
    [anon_sym_untie] = ACTIONS(2926),
    [anon_sym_umask] = ACTIONS(2926),
    [anon_sym_values] = ACTIONS(2926),
    [anon_sym_write] = ACTIONS(2926),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2926),
    [anon_sym_qq] = ACTIONS(2926),
    [anon_sym_qw] = ACTIONS(2926),
    [anon_sym_qx] = ACTIONS(2926),
    [anon_sym_qr] = ACTIONS(2926),
    [anon_sym_m] = ACTIONS(2926),
    [anon_sym_s] = ACTIONS(2926),
    [anon_sym_tr] = ACTIONS(2926),
    [anon_sym_y] = ACTIONS(2926),
    [aux_sym__bareword_token1] = ACTIONS(2926),
    [sym_number] = ACTIONS(2926),
    [anon_sym_true] = ACTIONS(2926),
    [anon_sym_false] = ACTIONS(2926),
    [sym__single_quote] = ACTIONS(2924),
    [sym__double_quote] = ACTIONS(2924),
    [sym__backtick_quote] = ACTIONS(2924),
    [sym__search_slash_quote] = ACTIONS(2924),
    [sym__open_readline_bracket] = ACTIONS(2924),
    [sym__open_fileglob_bracket] = ACTIONS(2924),
    [sym__PERLY_HEREDOC] = ACTIONS(2924),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2924),
    [sym__filetest] = ACTIONS(2924),
  },
  [STATE(1659)] = {
    [sym_heredoc_content] = STATE(1659),
    [sym__identifier] = ACTIONS(3061),
    [anon_sym_LBRACE] = ACTIONS(3061),
    [anon_sym_DOLLAR] = ACTIONS(3061),
    [anon_sym_AT] = ACTIONS(3061),
    [anon_sym_RPAREN] = ACTIONS(3061),
    [anon_sym_my] = ACTIONS(3061),
    [anon_sym_extended] = ACTIONS(3061),
    [anon_sym_async] = ACTIONS(3061),
    [anon_sym_sub] = ACTIONS(3061),
    [anon_sym_method] = ACTIONS(3061),
    [anon_sym_LPAREN] = ACTIONS(3061),
    [anon_sym_continue] = ACTIONS(3061),
    [anon_sym_state] = ACTIONS(3061),
    [anon_sym_our] = ACTIONS(3061),
    [anon_sym_LBRACK] = ACTIONS(3061),
    [anon_sym_LT_LT] = ACTIONS(3061),
    [anon_sym_SLASH_SLASH] = ACTIONS(3061),
    [anon_sym_PLUS] = ACTIONS(3061),
    [anon_sym_DASH] = ACTIONS(3061),
    [anon_sym_TILDE] = ACTIONS(3061),
    [anon_sym_BANG] = ACTIONS(3061),
    [anon_sym_PLUS_PLUS] = ACTIONS(3061),
    [anon_sym_DASH_DASH] = ACTIONS(3061),
    [anon_sym_BSLASH] = ACTIONS(3059),
    [anon_sym_do] = ACTIONS(3061),
    [anon_sym_eval] = ACTIONS(3061),
    [anon_sym_await] = ACTIONS(3061),
    [anon_sym_field] = ACTIONS(3061),
    [anon_sym_local] = ACTIONS(3061),
    [anon_sym_dynamically] = ACTIONS(3061),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3061),
    [anon_sym_require] = ACTIONS(3061),
    [anon_sym_map] = ACTIONS(3061),
    [anon_sym_grep] = ACTIONS(3061),
    [anon_sym_sort] = ACTIONS(3061),
    [anon_sym_goto] = ACTIONS(3061),
    [anon_sym_return] = ACTIONS(3061),
    [anon_sym_undef] = ACTIONS(3061),
    [anon_sym_PERCENT2] = ACTIONS(3061),
    [anon_sym_AMP2] = ACTIONS(3061),
    [anon_sym_STAR2] = ACTIONS(3061),
    [anon_sym_last] = ACTIONS(3061),
    [anon_sym_next] = ACTIONS(3061),
    [anon_sym_redo] = ACTIONS(3061),
    [anon_sym___FILE__] = ACTIONS(3061),
    [anon_sym___LINE__] = ACTIONS(3061),
    [anon_sym___PACKAGE__] = ACTIONS(3061),
    [anon_sym___SUB__] = ACTIONS(3061),
    [anon_sym_break] = ACTIONS(3061),
    [anon_sym_fork] = ACTIONS(3061),
    [anon_sym_getppid] = ACTIONS(3061),
    [anon_sym_time] = ACTIONS(3061),
    [anon_sym_times] = ACTIONS(3061),
    [anon_sym_wait] = ACTIONS(3061),
    [anon_sym_wantarray] = ACTIONS(3061),
    [anon_sym_abs] = ACTIONS(3061),
    [anon_sym_alarm] = ACTIONS(3061),
    [anon_sym_chop] = ACTIONS(3061),
    [anon_sym_chdir] = ACTIONS(3061),
    [anon_sym_close] = ACTIONS(3061),
    [anon_sym_closedir] = ACTIONS(3061),
    [anon_sym_caller] = ACTIONS(3061),
    [anon_sym_chomp] = ACTIONS(3061),
    [anon_sym_chr] = ACTIONS(3061),
    [anon_sym_cos] = ACTIONS(3061),
    [anon_sym_chroot] = ACTIONS(3061),
    [anon_sym_defined] = ACTIONS(3061),
    [anon_sym_delete] = ACTIONS(3061),
    [anon_sym_dbmclose] = ACTIONS(3061),
    [anon_sym_exists] = ACTIONS(3061),
    [anon_sym_exit] = ACTIONS(3061),
    [anon_sym_eof] = ACTIONS(3061),
    [anon_sym_exp] = ACTIONS(3061),
    [anon_sym_each] = ACTIONS(3061),
    [anon_sym_fc] = ACTIONS(3061),
    [anon_sym_fileno] = ACTIONS(3061),
    [anon_sym_gmtime] = ACTIONS(3061),
    [anon_sym_getc] = ACTIONS(3061),
    [anon_sym_getpgrp] = ACTIONS(3061),
    [anon_sym_getprotobyname] = ACTIONS(3061),
    [anon_sym_getpwname] = ACTIONS(3061),
    [anon_sym_getpwuid] = ACTIONS(3061),
    [anon_sym_getpeername] = ACTIONS(3061),
    [anon_sym_getnetbyname] = ACTIONS(3061),
    [anon_sym_getsockname] = ACTIONS(3061),
    [anon_sym_getgrnam] = ACTIONS(3061),
    [anon_sym_getgrgid] = ACTIONS(3061),
    [anon_sym_hex] = ACTIONS(3061),
    [anon_sym_int] = ACTIONS(3061),
    [anon_sym_keys] = ACTIONS(3061),
    [anon_sym_lc] = ACTIONS(3061),
    [anon_sym_lcfirst] = ACTIONS(3061),
    [anon_sym_length] = ACTIONS(3061),
    [anon_sym_localtime] = ACTIONS(3061),
    [anon_sym_log] = ACTIONS(3061),
    [anon_sym_lock] = ACTIONS(3061),
    [anon_sym_lstat] = ACTIONS(3061),
    [anon_sym_oct] = ACTIONS(3061),
    [anon_sym_ord] = ACTIONS(3061),
    [anon_sym_prototype] = ACTIONS(3061),
    [anon_sym_pop] = ACTIONS(3061),
    [anon_sym_pos] = ACTIONS(3061),
    [anon_sym_quotemeta] = ACTIONS(3061),
    [anon_sym_reset] = ACTIONS(3061),
    [anon_sym_rand] = ACTIONS(3061),
    [anon_sym_rmdir] = ACTIONS(3061),
    [anon_sym_readdir] = ACTIONS(3061),
    [anon_sym_readline] = ACTIONS(3061),
    [anon_sym_readpipe] = ACTIONS(3061),
    [anon_sym_rewinddir] = ACTIONS(3061),
    [anon_sym_readlink] = ACTIONS(3061),
    [anon_sym_ref] = ACTIONS(3061),
    [anon_sym_scalar] = ACTIONS(3061),
    [anon_sym_shift] = ACTIONS(3061),
    [anon_sym_sin] = ACTIONS(3061),
    [anon_sym_sleep] = ACTIONS(3061),
    [anon_sym_sqrt] = ACTIONS(3061),
    [anon_sym_srand] = ACTIONS(3061),
    [anon_sym_stat] = ACTIONS(3061),
    [anon_sym_study] = ACTIONS(3061),
    [anon_sym_tell] = ACTIONS(3061),
    [anon_sym_telldir] = ACTIONS(3061),
    [anon_sym_tied] = ACTIONS(3061),
    [anon_sym_uc] = ACTIONS(3061),
    [anon_sym_ucfirst] = ACTIONS(3061),
    [anon_sym_untie] = ACTIONS(3061),
    [anon_sym_umask] = ACTIONS(3061),
    [anon_sym_values] = ACTIONS(3061),
    [anon_sym_write] = ACTIONS(3061),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3061),
    [anon_sym_qq] = ACTIONS(3061),
    [anon_sym_qw] = ACTIONS(3061),
    [anon_sym_qx] = ACTIONS(3061),
    [anon_sym_qr] = ACTIONS(3061),
    [anon_sym_m] = ACTIONS(3061),
    [anon_sym_s] = ACTIONS(3061),
    [anon_sym_tr] = ACTIONS(3061),
    [anon_sym_y] = ACTIONS(3061),
    [aux_sym__bareword_token1] = ACTIONS(3061),
    [sym_number] = ACTIONS(3061),
    [anon_sym_true] = ACTIONS(3061),
    [anon_sym_false] = ACTIONS(3061),
    [sym__single_quote] = ACTIONS(3059),
    [sym__double_quote] = ACTIONS(3059),
    [sym__backtick_quote] = ACTIONS(3059),
    [sym__search_slash_quote] = ACTIONS(3059),
    [sym__open_readline_bracket] = ACTIONS(3059),
    [sym__open_fileglob_bracket] = ACTIONS(3059),
    [sym__PERLY_HEREDOC] = ACTIONS(3059),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3059),
    [sym__filetest] = ACTIONS(3059),
  },
  [STATE(1660)] = {
    [sym_heredoc_content] = STATE(1660),
    [sym__identifier] = ACTIONS(3233),
    [anon_sym_LBRACE] = ACTIONS(3233),
    [anon_sym_DOLLAR] = ACTIONS(3233),
    [anon_sym_AT] = ACTIONS(3233),
    [anon_sym_RPAREN] = ACTIONS(3233),
    [anon_sym_my] = ACTIONS(3233),
    [anon_sym_extended] = ACTIONS(3233),
    [anon_sym_async] = ACTIONS(3233),
    [anon_sym_sub] = ACTIONS(3233),
    [anon_sym_method] = ACTIONS(3233),
    [anon_sym_LPAREN] = ACTIONS(3233),
    [anon_sym_continue] = ACTIONS(3233),
    [anon_sym_state] = ACTIONS(3233),
    [anon_sym_our] = ACTIONS(3233),
    [anon_sym_LBRACK] = ACTIONS(3233),
    [anon_sym_LT_LT] = ACTIONS(3233),
    [anon_sym_SLASH_SLASH] = ACTIONS(3233),
    [anon_sym_PLUS] = ACTIONS(3233),
    [anon_sym_DASH] = ACTIONS(3233),
    [anon_sym_TILDE] = ACTIONS(3233),
    [anon_sym_BANG] = ACTIONS(3233),
    [anon_sym_PLUS_PLUS] = ACTIONS(3233),
    [anon_sym_DASH_DASH] = ACTIONS(3233),
    [anon_sym_BSLASH] = ACTIONS(3231),
    [anon_sym_do] = ACTIONS(3233),
    [anon_sym_eval] = ACTIONS(3233),
    [anon_sym_await] = ACTIONS(3233),
    [anon_sym_field] = ACTIONS(3233),
    [anon_sym_local] = ACTIONS(3233),
    [anon_sym_dynamically] = ACTIONS(3233),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3233),
    [anon_sym_require] = ACTIONS(3233),
    [anon_sym_map] = ACTIONS(3233),
    [anon_sym_grep] = ACTIONS(3233),
    [anon_sym_sort] = ACTIONS(3233),
    [anon_sym_goto] = ACTIONS(3233),
    [anon_sym_return] = ACTIONS(3233),
    [anon_sym_undef] = ACTIONS(3233),
    [anon_sym_PERCENT2] = ACTIONS(3233),
    [anon_sym_AMP2] = ACTIONS(3233),
    [anon_sym_STAR2] = ACTIONS(3233),
    [anon_sym_last] = ACTIONS(3233),
    [anon_sym_next] = ACTIONS(3233),
    [anon_sym_redo] = ACTIONS(3233),
    [anon_sym___FILE__] = ACTIONS(3233),
    [anon_sym___LINE__] = ACTIONS(3233),
    [anon_sym___PACKAGE__] = ACTIONS(3233),
    [anon_sym___SUB__] = ACTIONS(3233),
    [anon_sym_break] = ACTIONS(3233),
    [anon_sym_fork] = ACTIONS(3233),
    [anon_sym_getppid] = ACTIONS(3233),
    [anon_sym_time] = ACTIONS(3233),
    [anon_sym_times] = ACTIONS(3233),
    [anon_sym_wait] = ACTIONS(3233),
    [anon_sym_wantarray] = ACTIONS(3233),
    [anon_sym_abs] = ACTIONS(3233),
    [anon_sym_alarm] = ACTIONS(3233),
    [anon_sym_chop] = ACTIONS(3233),
    [anon_sym_chdir] = ACTIONS(3233),
    [anon_sym_close] = ACTIONS(3233),
    [anon_sym_closedir] = ACTIONS(3233),
    [anon_sym_caller] = ACTIONS(3233),
    [anon_sym_chomp] = ACTIONS(3233),
    [anon_sym_chr] = ACTIONS(3233),
    [anon_sym_cos] = ACTIONS(3233),
    [anon_sym_chroot] = ACTIONS(3233),
    [anon_sym_defined] = ACTIONS(3233),
    [anon_sym_delete] = ACTIONS(3233),
    [anon_sym_dbmclose] = ACTIONS(3233),
    [anon_sym_exists] = ACTIONS(3233),
    [anon_sym_exit] = ACTIONS(3233),
    [anon_sym_eof] = ACTIONS(3233),
    [anon_sym_exp] = ACTIONS(3233),
    [anon_sym_each] = ACTIONS(3233),
    [anon_sym_fc] = ACTIONS(3233),
    [anon_sym_fileno] = ACTIONS(3233),
    [anon_sym_gmtime] = ACTIONS(3233),
    [anon_sym_getc] = ACTIONS(3233),
    [anon_sym_getpgrp] = ACTIONS(3233),
    [anon_sym_getprotobyname] = ACTIONS(3233),
    [anon_sym_getpwname] = ACTIONS(3233),
    [anon_sym_getpwuid] = ACTIONS(3233),
    [anon_sym_getpeername] = ACTIONS(3233),
    [anon_sym_getnetbyname] = ACTIONS(3233),
    [anon_sym_getsockname] = ACTIONS(3233),
    [anon_sym_getgrnam] = ACTIONS(3233),
    [anon_sym_getgrgid] = ACTIONS(3233),
    [anon_sym_hex] = ACTIONS(3233),
    [anon_sym_int] = ACTIONS(3233),
    [anon_sym_keys] = ACTIONS(3233),
    [anon_sym_lc] = ACTIONS(3233),
    [anon_sym_lcfirst] = ACTIONS(3233),
    [anon_sym_length] = ACTIONS(3233),
    [anon_sym_localtime] = ACTIONS(3233),
    [anon_sym_log] = ACTIONS(3233),
    [anon_sym_lock] = ACTIONS(3233),
    [anon_sym_lstat] = ACTIONS(3233),
    [anon_sym_oct] = ACTIONS(3233),
    [anon_sym_ord] = ACTIONS(3233),
    [anon_sym_prototype] = ACTIONS(3233),
    [anon_sym_pop] = ACTIONS(3233),
    [anon_sym_pos] = ACTIONS(3233),
    [anon_sym_quotemeta] = ACTIONS(3233),
    [anon_sym_reset] = ACTIONS(3233),
    [anon_sym_rand] = ACTIONS(3233),
    [anon_sym_rmdir] = ACTIONS(3233),
    [anon_sym_readdir] = ACTIONS(3233),
    [anon_sym_readline] = ACTIONS(3233),
    [anon_sym_readpipe] = ACTIONS(3233),
    [anon_sym_rewinddir] = ACTIONS(3233),
    [anon_sym_readlink] = ACTIONS(3233),
    [anon_sym_ref] = ACTIONS(3233),
    [anon_sym_scalar] = ACTIONS(3233),
    [anon_sym_shift] = ACTIONS(3233),
    [anon_sym_sin] = ACTIONS(3233),
    [anon_sym_sleep] = ACTIONS(3233),
    [anon_sym_sqrt] = ACTIONS(3233),
    [anon_sym_srand] = ACTIONS(3233),
    [anon_sym_stat] = ACTIONS(3233),
    [anon_sym_study] = ACTIONS(3233),
    [anon_sym_tell] = ACTIONS(3233),
    [anon_sym_telldir] = ACTIONS(3233),
    [anon_sym_tied] = ACTIONS(3233),
    [anon_sym_uc] = ACTIONS(3233),
    [anon_sym_ucfirst] = ACTIONS(3233),
    [anon_sym_untie] = ACTIONS(3233),
    [anon_sym_umask] = ACTIONS(3233),
    [anon_sym_values] = ACTIONS(3233),
    [anon_sym_write] = ACTIONS(3233),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3233),
    [anon_sym_qq] = ACTIONS(3233),
    [anon_sym_qw] = ACTIONS(3233),
    [anon_sym_qx] = ACTIONS(3233),
    [anon_sym_qr] = ACTIONS(3233),
    [anon_sym_m] = ACTIONS(3233),
    [anon_sym_s] = ACTIONS(3233),
    [anon_sym_tr] = ACTIONS(3233),
    [anon_sym_y] = ACTIONS(3233),
    [aux_sym__bareword_token1] = ACTIONS(3233),
    [sym_number] = ACTIONS(3233),
    [anon_sym_true] = ACTIONS(3233),
    [anon_sym_false] = ACTIONS(3233),
    [sym__single_quote] = ACTIONS(3231),
    [sym__double_quote] = ACTIONS(3231),
    [sym__backtick_quote] = ACTIONS(3231),
    [sym__search_slash_quote] = ACTIONS(3231),
    [sym__open_readline_bracket] = ACTIONS(3231),
    [sym__open_fileglob_bracket] = ACTIONS(3231),
    [sym__PERLY_HEREDOC] = ACTIONS(3231),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3231),
    [sym__filetest] = ACTIONS(3231),
  },
  [STATE(1661)] = {
    [sym_heredoc_content] = STATE(1661),
    [sym__identifier] = ACTIONS(2922),
    [anon_sym_LBRACE] = ACTIONS(2922),
    [anon_sym_DOLLAR] = ACTIONS(2922),
    [anon_sym_AT] = ACTIONS(2922),
    [anon_sym_RPAREN] = ACTIONS(2922),
    [anon_sym_my] = ACTIONS(2922),
    [anon_sym_extended] = ACTIONS(2922),
    [anon_sym_async] = ACTIONS(2922),
    [anon_sym_sub] = ACTIONS(2922),
    [anon_sym_method] = ACTIONS(2922),
    [anon_sym_LPAREN] = ACTIONS(2922),
    [anon_sym_continue] = ACTIONS(2922),
    [anon_sym_state] = ACTIONS(2922),
    [anon_sym_our] = ACTIONS(2922),
    [anon_sym_LBRACK] = ACTIONS(2922),
    [anon_sym_LT_LT] = ACTIONS(2922),
    [anon_sym_SLASH_SLASH] = ACTIONS(2922),
    [anon_sym_PLUS] = ACTIONS(2922),
    [anon_sym_DASH] = ACTIONS(2922),
    [anon_sym_TILDE] = ACTIONS(2922),
    [anon_sym_BANG] = ACTIONS(2922),
    [anon_sym_PLUS_PLUS] = ACTIONS(2922),
    [anon_sym_DASH_DASH] = ACTIONS(2922),
    [anon_sym_BSLASH] = ACTIONS(2920),
    [anon_sym_do] = ACTIONS(2922),
    [anon_sym_eval] = ACTIONS(2922),
    [anon_sym_await] = ACTIONS(2922),
    [anon_sym_field] = ACTIONS(2922),
    [anon_sym_local] = ACTIONS(2922),
    [anon_sym_dynamically] = ACTIONS(2922),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2922),
    [anon_sym_require] = ACTIONS(2922),
    [anon_sym_map] = ACTIONS(2922),
    [anon_sym_grep] = ACTIONS(2922),
    [anon_sym_sort] = ACTIONS(2922),
    [anon_sym_goto] = ACTIONS(2922),
    [anon_sym_return] = ACTIONS(2922),
    [anon_sym_undef] = ACTIONS(2922),
    [anon_sym_PERCENT2] = ACTIONS(2922),
    [anon_sym_AMP2] = ACTIONS(2922),
    [anon_sym_STAR2] = ACTIONS(2922),
    [anon_sym_last] = ACTIONS(2922),
    [anon_sym_next] = ACTIONS(2922),
    [anon_sym_redo] = ACTIONS(2922),
    [anon_sym___FILE__] = ACTIONS(2922),
    [anon_sym___LINE__] = ACTIONS(2922),
    [anon_sym___PACKAGE__] = ACTIONS(2922),
    [anon_sym___SUB__] = ACTIONS(2922),
    [anon_sym_break] = ACTIONS(2922),
    [anon_sym_fork] = ACTIONS(2922),
    [anon_sym_getppid] = ACTIONS(2922),
    [anon_sym_time] = ACTIONS(2922),
    [anon_sym_times] = ACTIONS(2922),
    [anon_sym_wait] = ACTIONS(2922),
    [anon_sym_wantarray] = ACTIONS(2922),
    [anon_sym_abs] = ACTIONS(2922),
    [anon_sym_alarm] = ACTIONS(2922),
    [anon_sym_chop] = ACTIONS(2922),
    [anon_sym_chdir] = ACTIONS(2922),
    [anon_sym_close] = ACTIONS(2922),
    [anon_sym_closedir] = ACTIONS(2922),
    [anon_sym_caller] = ACTIONS(2922),
    [anon_sym_chomp] = ACTIONS(2922),
    [anon_sym_chr] = ACTIONS(2922),
    [anon_sym_cos] = ACTIONS(2922),
    [anon_sym_chroot] = ACTIONS(2922),
    [anon_sym_defined] = ACTIONS(2922),
    [anon_sym_delete] = ACTIONS(2922),
    [anon_sym_dbmclose] = ACTIONS(2922),
    [anon_sym_exists] = ACTIONS(2922),
    [anon_sym_exit] = ACTIONS(2922),
    [anon_sym_eof] = ACTIONS(2922),
    [anon_sym_exp] = ACTIONS(2922),
    [anon_sym_each] = ACTIONS(2922),
    [anon_sym_fc] = ACTIONS(2922),
    [anon_sym_fileno] = ACTIONS(2922),
    [anon_sym_gmtime] = ACTIONS(2922),
    [anon_sym_getc] = ACTIONS(2922),
    [anon_sym_getpgrp] = ACTIONS(2922),
    [anon_sym_getprotobyname] = ACTIONS(2922),
    [anon_sym_getpwname] = ACTIONS(2922),
    [anon_sym_getpwuid] = ACTIONS(2922),
    [anon_sym_getpeername] = ACTIONS(2922),
    [anon_sym_getnetbyname] = ACTIONS(2922),
    [anon_sym_getsockname] = ACTIONS(2922),
    [anon_sym_getgrnam] = ACTIONS(2922),
    [anon_sym_getgrgid] = ACTIONS(2922),
    [anon_sym_hex] = ACTIONS(2922),
    [anon_sym_int] = ACTIONS(2922),
    [anon_sym_keys] = ACTIONS(2922),
    [anon_sym_lc] = ACTIONS(2922),
    [anon_sym_lcfirst] = ACTIONS(2922),
    [anon_sym_length] = ACTIONS(2922),
    [anon_sym_localtime] = ACTIONS(2922),
    [anon_sym_log] = ACTIONS(2922),
    [anon_sym_lock] = ACTIONS(2922),
    [anon_sym_lstat] = ACTIONS(2922),
    [anon_sym_oct] = ACTIONS(2922),
    [anon_sym_ord] = ACTIONS(2922),
    [anon_sym_prototype] = ACTIONS(2922),
    [anon_sym_pop] = ACTIONS(2922),
    [anon_sym_pos] = ACTIONS(2922),
    [anon_sym_quotemeta] = ACTIONS(2922),
    [anon_sym_reset] = ACTIONS(2922),
    [anon_sym_rand] = ACTIONS(2922),
    [anon_sym_rmdir] = ACTIONS(2922),
    [anon_sym_readdir] = ACTIONS(2922),
    [anon_sym_readline] = ACTIONS(2922),
    [anon_sym_readpipe] = ACTIONS(2922),
    [anon_sym_rewinddir] = ACTIONS(2922),
    [anon_sym_readlink] = ACTIONS(2922),
    [anon_sym_ref] = ACTIONS(2922),
    [anon_sym_scalar] = ACTIONS(2922),
    [anon_sym_shift] = ACTIONS(2922),
    [anon_sym_sin] = ACTIONS(2922),
    [anon_sym_sleep] = ACTIONS(2922),
    [anon_sym_sqrt] = ACTIONS(2922),
    [anon_sym_srand] = ACTIONS(2922),
    [anon_sym_stat] = ACTIONS(2922),
    [anon_sym_study] = ACTIONS(2922),
    [anon_sym_tell] = ACTIONS(2922),
    [anon_sym_telldir] = ACTIONS(2922),
    [anon_sym_tied] = ACTIONS(2922),
    [anon_sym_uc] = ACTIONS(2922),
    [anon_sym_ucfirst] = ACTIONS(2922),
    [anon_sym_untie] = ACTIONS(2922),
    [anon_sym_umask] = ACTIONS(2922),
    [anon_sym_values] = ACTIONS(2922),
    [anon_sym_write] = ACTIONS(2922),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2922),
    [anon_sym_qq] = ACTIONS(2922),
    [anon_sym_qw] = ACTIONS(2922),
    [anon_sym_qx] = ACTIONS(2922),
    [anon_sym_qr] = ACTIONS(2922),
    [anon_sym_m] = ACTIONS(2922),
    [anon_sym_s] = ACTIONS(2922),
    [anon_sym_tr] = ACTIONS(2922),
    [anon_sym_y] = ACTIONS(2922),
    [aux_sym__bareword_token1] = ACTIONS(2922),
    [sym_number] = ACTIONS(2922),
    [anon_sym_true] = ACTIONS(2922),
    [anon_sym_false] = ACTIONS(2922),
    [sym__single_quote] = ACTIONS(2920),
    [sym__double_quote] = ACTIONS(2920),
    [sym__backtick_quote] = ACTIONS(2920),
    [sym__search_slash_quote] = ACTIONS(2920),
    [sym__open_readline_bracket] = ACTIONS(2920),
    [sym__open_fileglob_bracket] = ACTIONS(2920),
    [sym__PERLY_HEREDOC] = ACTIONS(2920),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2920),
    [sym__filetest] = ACTIONS(2920),
  },
  [STATE(1662)] = {
    [sym_heredoc_content] = STATE(1662),
    [sym__identifier] = ACTIONS(3085),
    [anon_sym_LBRACE] = ACTIONS(3085),
    [anon_sym_DOLLAR] = ACTIONS(3085),
    [anon_sym_AT] = ACTIONS(3085),
    [anon_sym_RPAREN] = ACTIONS(3085),
    [anon_sym_my] = ACTIONS(3085),
    [anon_sym_extended] = ACTIONS(3085),
    [anon_sym_async] = ACTIONS(3085),
    [anon_sym_sub] = ACTIONS(3085),
    [anon_sym_method] = ACTIONS(3085),
    [anon_sym_LPAREN] = ACTIONS(3085),
    [anon_sym_continue] = ACTIONS(3085),
    [anon_sym_state] = ACTIONS(3085),
    [anon_sym_our] = ACTIONS(3085),
    [anon_sym_LBRACK] = ACTIONS(3085),
    [anon_sym_LT_LT] = ACTIONS(3085),
    [anon_sym_SLASH_SLASH] = ACTIONS(3085),
    [anon_sym_PLUS] = ACTIONS(3085),
    [anon_sym_DASH] = ACTIONS(3085),
    [anon_sym_TILDE] = ACTIONS(3085),
    [anon_sym_BANG] = ACTIONS(3085),
    [anon_sym_PLUS_PLUS] = ACTIONS(3085),
    [anon_sym_DASH_DASH] = ACTIONS(3085),
    [anon_sym_BSLASH] = ACTIONS(3083),
    [anon_sym_do] = ACTIONS(3085),
    [anon_sym_eval] = ACTIONS(3085),
    [anon_sym_await] = ACTIONS(3085),
    [anon_sym_field] = ACTIONS(3085),
    [anon_sym_local] = ACTIONS(3085),
    [anon_sym_dynamically] = ACTIONS(3085),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3085),
    [anon_sym_require] = ACTIONS(3085),
    [anon_sym_map] = ACTIONS(3085),
    [anon_sym_grep] = ACTIONS(3085),
    [anon_sym_sort] = ACTIONS(3085),
    [anon_sym_goto] = ACTIONS(3085),
    [anon_sym_return] = ACTIONS(3085),
    [anon_sym_undef] = ACTIONS(3085),
    [anon_sym_PERCENT2] = ACTIONS(3085),
    [anon_sym_AMP2] = ACTIONS(3085),
    [anon_sym_STAR2] = ACTIONS(3085),
    [anon_sym_last] = ACTIONS(3085),
    [anon_sym_next] = ACTIONS(3085),
    [anon_sym_redo] = ACTIONS(3085),
    [anon_sym___FILE__] = ACTIONS(3085),
    [anon_sym___LINE__] = ACTIONS(3085),
    [anon_sym___PACKAGE__] = ACTIONS(3085),
    [anon_sym___SUB__] = ACTIONS(3085),
    [anon_sym_break] = ACTIONS(3085),
    [anon_sym_fork] = ACTIONS(3085),
    [anon_sym_getppid] = ACTIONS(3085),
    [anon_sym_time] = ACTIONS(3085),
    [anon_sym_times] = ACTIONS(3085),
    [anon_sym_wait] = ACTIONS(3085),
    [anon_sym_wantarray] = ACTIONS(3085),
    [anon_sym_abs] = ACTIONS(3085),
    [anon_sym_alarm] = ACTIONS(3085),
    [anon_sym_chop] = ACTIONS(3085),
    [anon_sym_chdir] = ACTIONS(3085),
    [anon_sym_close] = ACTIONS(3085),
    [anon_sym_closedir] = ACTIONS(3085),
    [anon_sym_caller] = ACTIONS(3085),
    [anon_sym_chomp] = ACTIONS(3085),
    [anon_sym_chr] = ACTIONS(3085),
    [anon_sym_cos] = ACTIONS(3085),
    [anon_sym_chroot] = ACTIONS(3085),
    [anon_sym_defined] = ACTIONS(3085),
    [anon_sym_delete] = ACTIONS(3085),
    [anon_sym_dbmclose] = ACTIONS(3085),
    [anon_sym_exists] = ACTIONS(3085),
    [anon_sym_exit] = ACTIONS(3085),
    [anon_sym_eof] = ACTIONS(3085),
    [anon_sym_exp] = ACTIONS(3085),
    [anon_sym_each] = ACTIONS(3085),
    [anon_sym_fc] = ACTIONS(3085),
    [anon_sym_fileno] = ACTIONS(3085),
    [anon_sym_gmtime] = ACTIONS(3085),
    [anon_sym_getc] = ACTIONS(3085),
    [anon_sym_getpgrp] = ACTIONS(3085),
    [anon_sym_getprotobyname] = ACTIONS(3085),
    [anon_sym_getpwname] = ACTIONS(3085),
    [anon_sym_getpwuid] = ACTIONS(3085),
    [anon_sym_getpeername] = ACTIONS(3085),
    [anon_sym_getnetbyname] = ACTIONS(3085),
    [anon_sym_getsockname] = ACTIONS(3085),
    [anon_sym_getgrnam] = ACTIONS(3085),
    [anon_sym_getgrgid] = ACTIONS(3085),
    [anon_sym_hex] = ACTIONS(3085),
    [anon_sym_int] = ACTIONS(3085),
    [anon_sym_keys] = ACTIONS(3085),
    [anon_sym_lc] = ACTIONS(3085),
    [anon_sym_lcfirst] = ACTIONS(3085),
    [anon_sym_length] = ACTIONS(3085),
    [anon_sym_localtime] = ACTIONS(3085),
    [anon_sym_log] = ACTIONS(3085),
    [anon_sym_lock] = ACTIONS(3085),
    [anon_sym_lstat] = ACTIONS(3085),
    [anon_sym_oct] = ACTIONS(3085),
    [anon_sym_ord] = ACTIONS(3085),
    [anon_sym_prototype] = ACTIONS(3085),
    [anon_sym_pop] = ACTIONS(3085),
    [anon_sym_pos] = ACTIONS(3085),
    [anon_sym_quotemeta] = ACTIONS(3085),
    [anon_sym_reset] = ACTIONS(3085),
    [anon_sym_rand] = ACTIONS(3085),
    [anon_sym_rmdir] = ACTIONS(3085),
    [anon_sym_readdir] = ACTIONS(3085),
    [anon_sym_readline] = ACTIONS(3085),
    [anon_sym_readpipe] = ACTIONS(3085),
    [anon_sym_rewinddir] = ACTIONS(3085),
    [anon_sym_readlink] = ACTIONS(3085),
    [anon_sym_ref] = ACTIONS(3085),
    [anon_sym_scalar] = ACTIONS(3085),
    [anon_sym_shift] = ACTIONS(3085),
    [anon_sym_sin] = ACTIONS(3085),
    [anon_sym_sleep] = ACTIONS(3085),
    [anon_sym_sqrt] = ACTIONS(3085),
    [anon_sym_srand] = ACTIONS(3085),
    [anon_sym_stat] = ACTIONS(3085),
    [anon_sym_study] = ACTIONS(3085),
    [anon_sym_tell] = ACTIONS(3085),
    [anon_sym_telldir] = ACTIONS(3085),
    [anon_sym_tied] = ACTIONS(3085),
    [anon_sym_uc] = ACTIONS(3085),
    [anon_sym_ucfirst] = ACTIONS(3085),
    [anon_sym_untie] = ACTIONS(3085),
    [anon_sym_umask] = ACTIONS(3085),
    [anon_sym_values] = ACTIONS(3085),
    [anon_sym_write] = ACTIONS(3085),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3085),
    [anon_sym_qq] = ACTIONS(3085),
    [anon_sym_qw] = ACTIONS(3085),
    [anon_sym_qx] = ACTIONS(3085),
    [anon_sym_qr] = ACTIONS(3085),
    [anon_sym_m] = ACTIONS(3085),
    [anon_sym_s] = ACTIONS(3085),
    [anon_sym_tr] = ACTIONS(3085),
    [anon_sym_y] = ACTIONS(3085),
    [aux_sym__bareword_token1] = ACTIONS(3085),
    [sym_number] = ACTIONS(3085),
    [anon_sym_true] = ACTIONS(3085),
    [anon_sym_false] = ACTIONS(3085),
    [sym__single_quote] = ACTIONS(3083),
    [sym__double_quote] = ACTIONS(3083),
    [sym__backtick_quote] = ACTIONS(3083),
    [sym__search_slash_quote] = ACTIONS(3083),
    [sym__open_readline_bracket] = ACTIONS(3083),
    [sym__open_fileglob_bracket] = ACTIONS(3083),
    [sym__PERLY_HEREDOC] = ACTIONS(3083),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3083),
    [sym__filetest] = ACTIONS(3083),
  },
  [STATE(1663)] = {
    [sym_heredoc_content] = STATE(1663),
    [sym__identifier] = ACTIONS(2906),
    [anon_sym_LBRACE] = ACTIONS(2906),
    [anon_sym_DOLLAR] = ACTIONS(2906),
    [anon_sym_AT] = ACTIONS(2906),
    [anon_sym_RPAREN] = ACTIONS(2906),
    [anon_sym_my] = ACTIONS(2906),
    [anon_sym_extended] = ACTIONS(2906),
    [anon_sym_async] = ACTIONS(2906),
    [anon_sym_sub] = ACTIONS(2906),
    [anon_sym_method] = ACTIONS(2906),
    [anon_sym_LPAREN] = ACTIONS(2906),
    [anon_sym_continue] = ACTIONS(2906),
    [anon_sym_state] = ACTIONS(2906),
    [anon_sym_our] = ACTIONS(2906),
    [anon_sym_LBRACK] = ACTIONS(2906),
    [anon_sym_LT_LT] = ACTIONS(2906),
    [anon_sym_SLASH_SLASH] = ACTIONS(2906),
    [anon_sym_PLUS] = ACTIONS(2906),
    [anon_sym_DASH] = ACTIONS(2906),
    [anon_sym_TILDE] = ACTIONS(2906),
    [anon_sym_BANG] = ACTIONS(2906),
    [anon_sym_PLUS_PLUS] = ACTIONS(2906),
    [anon_sym_DASH_DASH] = ACTIONS(2906),
    [anon_sym_BSLASH] = ACTIONS(2904),
    [anon_sym_do] = ACTIONS(2906),
    [anon_sym_eval] = ACTIONS(2906),
    [anon_sym_await] = ACTIONS(2906),
    [anon_sym_field] = ACTIONS(2906),
    [anon_sym_local] = ACTIONS(2906),
    [anon_sym_dynamically] = ACTIONS(2906),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2906),
    [anon_sym_require] = ACTIONS(2906),
    [anon_sym_map] = ACTIONS(2906),
    [anon_sym_grep] = ACTIONS(2906),
    [anon_sym_sort] = ACTIONS(2906),
    [anon_sym_goto] = ACTIONS(2906),
    [anon_sym_return] = ACTIONS(2906),
    [anon_sym_undef] = ACTIONS(2906),
    [anon_sym_PERCENT2] = ACTIONS(2906),
    [anon_sym_AMP2] = ACTIONS(2906),
    [anon_sym_STAR2] = ACTIONS(2906),
    [anon_sym_last] = ACTIONS(2906),
    [anon_sym_next] = ACTIONS(2906),
    [anon_sym_redo] = ACTIONS(2906),
    [anon_sym___FILE__] = ACTIONS(2906),
    [anon_sym___LINE__] = ACTIONS(2906),
    [anon_sym___PACKAGE__] = ACTIONS(2906),
    [anon_sym___SUB__] = ACTIONS(2906),
    [anon_sym_break] = ACTIONS(2906),
    [anon_sym_fork] = ACTIONS(2906),
    [anon_sym_getppid] = ACTIONS(2906),
    [anon_sym_time] = ACTIONS(2906),
    [anon_sym_times] = ACTIONS(2906),
    [anon_sym_wait] = ACTIONS(2906),
    [anon_sym_wantarray] = ACTIONS(2906),
    [anon_sym_abs] = ACTIONS(2906),
    [anon_sym_alarm] = ACTIONS(2906),
    [anon_sym_chop] = ACTIONS(2906),
    [anon_sym_chdir] = ACTIONS(2906),
    [anon_sym_close] = ACTIONS(2906),
    [anon_sym_closedir] = ACTIONS(2906),
    [anon_sym_caller] = ACTIONS(2906),
    [anon_sym_chomp] = ACTIONS(2906),
    [anon_sym_chr] = ACTIONS(2906),
    [anon_sym_cos] = ACTIONS(2906),
    [anon_sym_chroot] = ACTIONS(2906),
    [anon_sym_defined] = ACTIONS(2906),
    [anon_sym_delete] = ACTIONS(2906),
    [anon_sym_dbmclose] = ACTIONS(2906),
    [anon_sym_exists] = ACTIONS(2906),
    [anon_sym_exit] = ACTIONS(2906),
    [anon_sym_eof] = ACTIONS(2906),
    [anon_sym_exp] = ACTIONS(2906),
    [anon_sym_each] = ACTIONS(2906),
    [anon_sym_fc] = ACTIONS(2906),
    [anon_sym_fileno] = ACTIONS(2906),
    [anon_sym_gmtime] = ACTIONS(2906),
    [anon_sym_getc] = ACTIONS(2906),
    [anon_sym_getpgrp] = ACTIONS(2906),
    [anon_sym_getprotobyname] = ACTIONS(2906),
    [anon_sym_getpwname] = ACTIONS(2906),
    [anon_sym_getpwuid] = ACTIONS(2906),
    [anon_sym_getpeername] = ACTIONS(2906),
    [anon_sym_getnetbyname] = ACTIONS(2906),
    [anon_sym_getsockname] = ACTIONS(2906),
    [anon_sym_getgrnam] = ACTIONS(2906),
    [anon_sym_getgrgid] = ACTIONS(2906),
    [anon_sym_hex] = ACTIONS(2906),
    [anon_sym_int] = ACTIONS(2906),
    [anon_sym_keys] = ACTIONS(2906),
    [anon_sym_lc] = ACTIONS(2906),
    [anon_sym_lcfirst] = ACTIONS(2906),
    [anon_sym_length] = ACTIONS(2906),
    [anon_sym_localtime] = ACTIONS(2906),
    [anon_sym_log] = ACTIONS(2906),
    [anon_sym_lock] = ACTIONS(2906),
    [anon_sym_lstat] = ACTIONS(2906),
    [anon_sym_oct] = ACTIONS(2906),
    [anon_sym_ord] = ACTIONS(2906),
    [anon_sym_prototype] = ACTIONS(2906),
    [anon_sym_pop] = ACTIONS(2906),
    [anon_sym_pos] = ACTIONS(2906),
    [anon_sym_quotemeta] = ACTIONS(2906),
    [anon_sym_reset] = ACTIONS(2906),
    [anon_sym_rand] = ACTIONS(2906),
    [anon_sym_rmdir] = ACTIONS(2906),
    [anon_sym_readdir] = ACTIONS(2906),
    [anon_sym_readline] = ACTIONS(2906),
    [anon_sym_readpipe] = ACTIONS(2906),
    [anon_sym_rewinddir] = ACTIONS(2906),
    [anon_sym_readlink] = ACTIONS(2906),
    [anon_sym_ref] = ACTIONS(2906),
    [anon_sym_scalar] = ACTIONS(2906),
    [anon_sym_shift] = ACTIONS(2906),
    [anon_sym_sin] = ACTIONS(2906),
    [anon_sym_sleep] = ACTIONS(2906),
    [anon_sym_sqrt] = ACTIONS(2906),
    [anon_sym_srand] = ACTIONS(2906),
    [anon_sym_stat] = ACTIONS(2906),
    [anon_sym_study] = ACTIONS(2906),
    [anon_sym_tell] = ACTIONS(2906),
    [anon_sym_telldir] = ACTIONS(2906),
    [anon_sym_tied] = ACTIONS(2906),
    [anon_sym_uc] = ACTIONS(2906),
    [anon_sym_ucfirst] = ACTIONS(2906),
    [anon_sym_untie] = ACTIONS(2906),
    [anon_sym_umask] = ACTIONS(2906),
    [anon_sym_values] = ACTIONS(2906),
    [anon_sym_write] = ACTIONS(2906),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2906),
    [anon_sym_qq] = ACTIONS(2906),
    [anon_sym_qw] = ACTIONS(2906),
    [anon_sym_qx] = ACTIONS(2906),
    [anon_sym_qr] = ACTIONS(2906),
    [anon_sym_m] = ACTIONS(2906),
    [anon_sym_s] = ACTIONS(2906),
    [anon_sym_tr] = ACTIONS(2906),
    [anon_sym_y] = ACTIONS(2906),
    [aux_sym__bareword_token1] = ACTIONS(2906),
    [sym_number] = ACTIONS(2906),
    [anon_sym_true] = ACTIONS(2906),
    [anon_sym_false] = ACTIONS(2906),
    [sym__single_quote] = ACTIONS(2904),
    [sym__double_quote] = ACTIONS(2904),
    [sym__backtick_quote] = ACTIONS(2904),
    [sym__search_slash_quote] = ACTIONS(2904),
    [sym__open_readline_bracket] = ACTIONS(2904),
    [sym__open_fileglob_bracket] = ACTIONS(2904),
    [sym__PERLY_HEREDOC] = ACTIONS(2904),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2904),
    [sym__filetest] = ACTIONS(2904),
  },
  [STATE(1664)] = {
    [sym_heredoc_content] = STATE(1664),
    [sym__identifier] = ACTIONS(3165),
    [anon_sym_LBRACE] = ACTIONS(3165),
    [anon_sym_DOLLAR] = ACTIONS(3165),
    [anon_sym_AT] = ACTIONS(3165),
    [anon_sym_RPAREN] = ACTIONS(3165),
    [anon_sym_my] = ACTIONS(3165),
    [anon_sym_extended] = ACTIONS(3165),
    [anon_sym_async] = ACTIONS(3165),
    [anon_sym_sub] = ACTIONS(3165),
    [anon_sym_method] = ACTIONS(3165),
    [anon_sym_LPAREN] = ACTIONS(3165),
    [anon_sym_continue] = ACTIONS(3165),
    [anon_sym_state] = ACTIONS(3165),
    [anon_sym_our] = ACTIONS(3165),
    [anon_sym_LBRACK] = ACTIONS(3165),
    [anon_sym_LT_LT] = ACTIONS(3165),
    [anon_sym_SLASH_SLASH] = ACTIONS(3165),
    [anon_sym_PLUS] = ACTIONS(3165),
    [anon_sym_DASH] = ACTIONS(3165),
    [anon_sym_TILDE] = ACTIONS(3165),
    [anon_sym_BANG] = ACTIONS(3165),
    [anon_sym_PLUS_PLUS] = ACTIONS(3165),
    [anon_sym_DASH_DASH] = ACTIONS(3165),
    [anon_sym_BSLASH] = ACTIONS(3163),
    [anon_sym_do] = ACTIONS(3165),
    [anon_sym_eval] = ACTIONS(3165),
    [anon_sym_await] = ACTIONS(3165),
    [anon_sym_field] = ACTIONS(3165),
    [anon_sym_local] = ACTIONS(3165),
    [anon_sym_dynamically] = ACTIONS(3165),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3165),
    [anon_sym_require] = ACTIONS(3165),
    [anon_sym_map] = ACTIONS(3165),
    [anon_sym_grep] = ACTIONS(3165),
    [anon_sym_sort] = ACTIONS(3165),
    [anon_sym_goto] = ACTIONS(3165),
    [anon_sym_return] = ACTIONS(3165),
    [anon_sym_undef] = ACTIONS(3165),
    [anon_sym_PERCENT2] = ACTIONS(3165),
    [anon_sym_AMP2] = ACTIONS(3165),
    [anon_sym_STAR2] = ACTIONS(3165),
    [anon_sym_last] = ACTIONS(3165),
    [anon_sym_next] = ACTIONS(3165),
    [anon_sym_redo] = ACTIONS(3165),
    [anon_sym___FILE__] = ACTIONS(3165),
    [anon_sym___LINE__] = ACTIONS(3165),
    [anon_sym___PACKAGE__] = ACTIONS(3165),
    [anon_sym___SUB__] = ACTIONS(3165),
    [anon_sym_break] = ACTIONS(3165),
    [anon_sym_fork] = ACTIONS(3165),
    [anon_sym_getppid] = ACTIONS(3165),
    [anon_sym_time] = ACTIONS(3165),
    [anon_sym_times] = ACTIONS(3165),
    [anon_sym_wait] = ACTIONS(3165),
    [anon_sym_wantarray] = ACTIONS(3165),
    [anon_sym_abs] = ACTIONS(3165),
    [anon_sym_alarm] = ACTIONS(3165),
    [anon_sym_chop] = ACTIONS(3165),
    [anon_sym_chdir] = ACTIONS(3165),
    [anon_sym_close] = ACTIONS(3165),
    [anon_sym_closedir] = ACTIONS(3165),
    [anon_sym_caller] = ACTIONS(3165),
    [anon_sym_chomp] = ACTIONS(3165),
    [anon_sym_chr] = ACTIONS(3165),
    [anon_sym_cos] = ACTIONS(3165),
    [anon_sym_chroot] = ACTIONS(3165),
    [anon_sym_defined] = ACTIONS(3165),
    [anon_sym_delete] = ACTIONS(3165),
    [anon_sym_dbmclose] = ACTIONS(3165),
    [anon_sym_exists] = ACTIONS(3165),
    [anon_sym_exit] = ACTIONS(3165),
    [anon_sym_eof] = ACTIONS(3165),
    [anon_sym_exp] = ACTIONS(3165),
    [anon_sym_each] = ACTIONS(3165),
    [anon_sym_fc] = ACTIONS(3165),
    [anon_sym_fileno] = ACTIONS(3165),
    [anon_sym_gmtime] = ACTIONS(3165),
    [anon_sym_getc] = ACTIONS(3165),
    [anon_sym_getpgrp] = ACTIONS(3165),
    [anon_sym_getprotobyname] = ACTIONS(3165),
    [anon_sym_getpwname] = ACTIONS(3165),
    [anon_sym_getpwuid] = ACTIONS(3165),
    [anon_sym_getpeername] = ACTIONS(3165),
    [anon_sym_getnetbyname] = ACTIONS(3165),
    [anon_sym_getsockname] = ACTIONS(3165),
    [anon_sym_getgrnam] = ACTIONS(3165),
    [anon_sym_getgrgid] = ACTIONS(3165),
    [anon_sym_hex] = ACTIONS(3165),
    [anon_sym_int] = ACTIONS(3165),
    [anon_sym_keys] = ACTIONS(3165),
    [anon_sym_lc] = ACTIONS(3165),
    [anon_sym_lcfirst] = ACTIONS(3165),
    [anon_sym_length] = ACTIONS(3165),
    [anon_sym_localtime] = ACTIONS(3165),
    [anon_sym_log] = ACTIONS(3165),
    [anon_sym_lock] = ACTIONS(3165),
    [anon_sym_lstat] = ACTIONS(3165),
    [anon_sym_oct] = ACTIONS(3165),
    [anon_sym_ord] = ACTIONS(3165),
    [anon_sym_prototype] = ACTIONS(3165),
    [anon_sym_pop] = ACTIONS(3165),
    [anon_sym_pos] = ACTIONS(3165),
    [anon_sym_quotemeta] = ACTIONS(3165),
    [anon_sym_reset] = ACTIONS(3165),
    [anon_sym_rand] = ACTIONS(3165),
    [anon_sym_rmdir] = ACTIONS(3165),
    [anon_sym_readdir] = ACTIONS(3165),
    [anon_sym_readline] = ACTIONS(3165),
    [anon_sym_readpipe] = ACTIONS(3165),
    [anon_sym_rewinddir] = ACTIONS(3165),
    [anon_sym_readlink] = ACTIONS(3165),
    [anon_sym_ref] = ACTIONS(3165),
    [anon_sym_scalar] = ACTIONS(3165),
    [anon_sym_shift] = ACTIONS(3165),
    [anon_sym_sin] = ACTIONS(3165),
    [anon_sym_sleep] = ACTIONS(3165),
    [anon_sym_sqrt] = ACTIONS(3165),
    [anon_sym_srand] = ACTIONS(3165),
    [anon_sym_stat] = ACTIONS(3165),
    [anon_sym_study] = ACTIONS(3165),
    [anon_sym_tell] = ACTIONS(3165),
    [anon_sym_telldir] = ACTIONS(3165),
    [anon_sym_tied] = ACTIONS(3165),
    [anon_sym_uc] = ACTIONS(3165),
    [anon_sym_ucfirst] = ACTIONS(3165),
    [anon_sym_untie] = ACTIONS(3165),
    [anon_sym_umask] = ACTIONS(3165),
    [anon_sym_values] = ACTIONS(3165),
    [anon_sym_write] = ACTIONS(3165),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3165),
    [anon_sym_qq] = ACTIONS(3165),
    [anon_sym_qw] = ACTIONS(3165),
    [anon_sym_qx] = ACTIONS(3165),
    [anon_sym_qr] = ACTIONS(3165),
    [anon_sym_m] = ACTIONS(3165),
    [anon_sym_s] = ACTIONS(3165),
    [anon_sym_tr] = ACTIONS(3165),
    [anon_sym_y] = ACTIONS(3165),
    [aux_sym__bareword_token1] = ACTIONS(3165),
    [sym_number] = ACTIONS(3165),
    [anon_sym_true] = ACTIONS(3165),
    [anon_sym_false] = ACTIONS(3165),
    [sym__single_quote] = ACTIONS(3163),
    [sym__double_quote] = ACTIONS(3163),
    [sym__backtick_quote] = ACTIONS(3163),
    [sym__search_slash_quote] = ACTIONS(3163),
    [sym__open_readline_bracket] = ACTIONS(3163),
    [sym__open_fileglob_bracket] = ACTIONS(3163),
    [sym__PERLY_HEREDOC] = ACTIONS(3163),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3163),
    [sym__filetest] = ACTIONS(3163),
  },
  [STATE(1665)] = {
    [sym_heredoc_content] = STATE(1665),
    [sym__identifier] = ACTIONS(3229),
    [anon_sym_LBRACE] = ACTIONS(3229),
    [anon_sym_DOLLAR] = ACTIONS(3229),
    [anon_sym_AT] = ACTIONS(3229),
    [anon_sym_RPAREN] = ACTIONS(3229),
    [anon_sym_my] = ACTIONS(3229),
    [anon_sym_extended] = ACTIONS(3229),
    [anon_sym_async] = ACTIONS(3229),
    [anon_sym_sub] = ACTIONS(3229),
    [anon_sym_method] = ACTIONS(3229),
    [anon_sym_LPAREN] = ACTIONS(3229),
    [anon_sym_continue] = ACTIONS(3229),
    [anon_sym_state] = ACTIONS(3229),
    [anon_sym_our] = ACTIONS(3229),
    [anon_sym_LBRACK] = ACTIONS(3229),
    [anon_sym_LT_LT] = ACTIONS(3229),
    [anon_sym_SLASH_SLASH] = ACTIONS(3229),
    [anon_sym_PLUS] = ACTIONS(3229),
    [anon_sym_DASH] = ACTIONS(3229),
    [anon_sym_TILDE] = ACTIONS(3229),
    [anon_sym_BANG] = ACTIONS(3229),
    [anon_sym_PLUS_PLUS] = ACTIONS(3229),
    [anon_sym_DASH_DASH] = ACTIONS(3229),
    [anon_sym_BSLASH] = ACTIONS(3227),
    [anon_sym_do] = ACTIONS(3229),
    [anon_sym_eval] = ACTIONS(3229),
    [anon_sym_await] = ACTIONS(3229),
    [anon_sym_field] = ACTIONS(3229),
    [anon_sym_local] = ACTIONS(3229),
    [anon_sym_dynamically] = ACTIONS(3229),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3229),
    [anon_sym_require] = ACTIONS(3229),
    [anon_sym_map] = ACTIONS(3229),
    [anon_sym_grep] = ACTIONS(3229),
    [anon_sym_sort] = ACTIONS(3229),
    [anon_sym_goto] = ACTIONS(3229),
    [anon_sym_return] = ACTIONS(3229),
    [anon_sym_undef] = ACTIONS(3229),
    [anon_sym_PERCENT2] = ACTIONS(3229),
    [anon_sym_AMP2] = ACTIONS(3229),
    [anon_sym_STAR2] = ACTIONS(3229),
    [anon_sym_last] = ACTIONS(3229),
    [anon_sym_next] = ACTIONS(3229),
    [anon_sym_redo] = ACTIONS(3229),
    [anon_sym___FILE__] = ACTIONS(3229),
    [anon_sym___LINE__] = ACTIONS(3229),
    [anon_sym___PACKAGE__] = ACTIONS(3229),
    [anon_sym___SUB__] = ACTIONS(3229),
    [anon_sym_break] = ACTIONS(3229),
    [anon_sym_fork] = ACTIONS(3229),
    [anon_sym_getppid] = ACTIONS(3229),
    [anon_sym_time] = ACTIONS(3229),
    [anon_sym_times] = ACTIONS(3229),
    [anon_sym_wait] = ACTIONS(3229),
    [anon_sym_wantarray] = ACTIONS(3229),
    [anon_sym_abs] = ACTIONS(3229),
    [anon_sym_alarm] = ACTIONS(3229),
    [anon_sym_chop] = ACTIONS(3229),
    [anon_sym_chdir] = ACTIONS(3229),
    [anon_sym_close] = ACTIONS(3229),
    [anon_sym_closedir] = ACTIONS(3229),
    [anon_sym_caller] = ACTIONS(3229),
    [anon_sym_chomp] = ACTIONS(3229),
    [anon_sym_chr] = ACTIONS(3229),
    [anon_sym_cos] = ACTIONS(3229),
    [anon_sym_chroot] = ACTIONS(3229),
    [anon_sym_defined] = ACTIONS(3229),
    [anon_sym_delete] = ACTIONS(3229),
    [anon_sym_dbmclose] = ACTIONS(3229),
    [anon_sym_exists] = ACTIONS(3229),
    [anon_sym_exit] = ACTIONS(3229),
    [anon_sym_eof] = ACTIONS(3229),
    [anon_sym_exp] = ACTIONS(3229),
    [anon_sym_each] = ACTIONS(3229),
    [anon_sym_fc] = ACTIONS(3229),
    [anon_sym_fileno] = ACTIONS(3229),
    [anon_sym_gmtime] = ACTIONS(3229),
    [anon_sym_getc] = ACTIONS(3229),
    [anon_sym_getpgrp] = ACTIONS(3229),
    [anon_sym_getprotobyname] = ACTIONS(3229),
    [anon_sym_getpwname] = ACTIONS(3229),
    [anon_sym_getpwuid] = ACTIONS(3229),
    [anon_sym_getpeername] = ACTIONS(3229),
    [anon_sym_getnetbyname] = ACTIONS(3229),
    [anon_sym_getsockname] = ACTIONS(3229),
    [anon_sym_getgrnam] = ACTIONS(3229),
    [anon_sym_getgrgid] = ACTIONS(3229),
    [anon_sym_hex] = ACTIONS(3229),
    [anon_sym_int] = ACTIONS(3229),
    [anon_sym_keys] = ACTIONS(3229),
    [anon_sym_lc] = ACTIONS(3229),
    [anon_sym_lcfirst] = ACTIONS(3229),
    [anon_sym_length] = ACTIONS(3229),
    [anon_sym_localtime] = ACTIONS(3229),
    [anon_sym_log] = ACTIONS(3229),
    [anon_sym_lock] = ACTIONS(3229),
    [anon_sym_lstat] = ACTIONS(3229),
    [anon_sym_oct] = ACTIONS(3229),
    [anon_sym_ord] = ACTIONS(3229),
    [anon_sym_prototype] = ACTIONS(3229),
    [anon_sym_pop] = ACTIONS(3229),
    [anon_sym_pos] = ACTIONS(3229),
    [anon_sym_quotemeta] = ACTIONS(3229),
    [anon_sym_reset] = ACTIONS(3229),
    [anon_sym_rand] = ACTIONS(3229),
    [anon_sym_rmdir] = ACTIONS(3229),
    [anon_sym_readdir] = ACTIONS(3229),
    [anon_sym_readline] = ACTIONS(3229),
    [anon_sym_readpipe] = ACTIONS(3229),
    [anon_sym_rewinddir] = ACTIONS(3229),
    [anon_sym_readlink] = ACTIONS(3229),
    [anon_sym_ref] = ACTIONS(3229),
    [anon_sym_scalar] = ACTIONS(3229),
    [anon_sym_shift] = ACTIONS(3229),
    [anon_sym_sin] = ACTIONS(3229),
    [anon_sym_sleep] = ACTIONS(3229),
    [anon_sym_sqrt] = ACTIONS(3229),
    [anon_sym_srand] = ACTIONS(3229),
    [anon_sym_stat] = ACTIONS(3229),
    [anon_sym_study] = ACTIONS(3229),
    [anon_sym_tell] = ACTIONS(3229),
    [anon_sym_telldir] = ACTIONS(3229),
    [anon_sym_tied] = ACTIONS(3229),
    [anon_sym_uc] = ACTIONS(3229),
    [anon_sym_ucfirst] = ACTIONS(3229),
    [anon_sym_untie] = ACTIONS(3229),
    [anon_sym_umask] = ACTIONS(3229),
    [anon_sym_values] = ACTIONS(3229),
    [anon_sym_write] = ACTIONS(3229),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3229),
    [anon_sym_qq] = ACTIONS(3229),
    [anon_sym_qw] = ACTIONS(3229),
    [anon_sym_qx] = ACTIONS(3229),
    [anon_sym_qr] = ACTIONS(3229),
    [anon_sym_m] = ACTIONS(3229),
    [anon_sym_s] = ACTIONS(3229),
    [anon_sym_tr] = ACTIONS(3229),
    [anon_sym_y] = ACTIONS(3229),
    [aux_sym__bareword_token1] = ACTIONS(3229),
    [sym_number] = ACTIONS(3229),
    [anon_sym_true] = ACTIONS(3229),
    [anon_sym_false] = ACTIONS(3229),
    [sym__single_quote] = ACTIONS(3227),
    [sym__double_quote] = ACTIONS(3227),
    [sym__backtick_quote] = ACTIONS(3227),
    [sym__search_slash_quote] = ACTIONS(3227),
    [sym__open_readline_bracket] = ACTIONS(3227),
    [sym__open_fileglob_bracket] = ACTIONS(3227),
    [sym__PERLY_HEREDOC] = ACTIONS(3227),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3227),
    [sym__filetest] = ACTIONS(3227),
  },
  [STATE(1666)] = {
    [sym_heredoc_content] = STATE(1666),
    [sym__identifier] = ACTIONS(2930),
    [anon_sym_LBRACE] = ACTIONS(2930),
    [anon_sym_DOLLAR] = ACTIONS(2930),
    [anon_sym_AT] = ACTIONS(2930),
    [anon_sym_RPAREN] = ACTIONS(2930),
    [anon_sym_my] = ACTIONS(2930),
    [anon_sym_extended] = ACTIONS(2930),
    [anon_sym_async] = ACTIONS(2930),
    [anon_sym_sub] = ACTIONS(2930),
    [anon_sym_method] = ACTIONS(2930),
    [anon_sym_LPAREN] = ACTIONS(2930),
    [anon_sym_continue] = ACTIONS(2930),
    [anon_sym_state] = ACTIONS(2930),
    [anon_sym_our] = ACTIONS(2930),
    [anon_sym_LBRACK] = ACTIONS(2930),
    [anon_sym_LT_LT] = ACTIONS(2930),
    [anon_sym_SLASH_SLASH] = ACTIONS(2930),
    [anon_sym_PLUS] = ACTIONS(2930),
    [anon_sym_DASH] = ACTIONS(2930),
    [anon_sym_TILDE] = ACTIONS(2930),
    [anon_sym_BANG] = ACTIONS(2930),
    [anon_sym_PLUS_PLUS] = ACTIONS(2930),
    [anon_sym_DASH_DASH] = ACTIONS(2930),
    [anon_sym_BSLASH] = ACTIONS(2928),
    [anon_sym_do] = ACTIONS(2930),
    [anon_sym_eval] = ACTIONS(2930),
    [anon_sym_await] = ACTIONS(2930),
    [anon_sym_field] = ACTIONS(2930),
    [anon_sym_local] = ACTIONS(2930),
    [anon_sym_dynamically] = ACTIONS(2930),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2930),
    [anon_sym_require] = ACTIONS(2930),
    [anon_sym_map] = ACTIONS(2930),
    [anon_sym_grep] = ACTIONS(2930),
    [anon_sym_sort] = ACTIONS(2930),
    [anon_sym_goto] = ACTIONS(2930),
    [anon_sym_return] = ACTIONS(2930),
    [anon_sym_undef] = ACTIONS(2930),
    [anon_sym_PERCENT2] = ACTIONS(2930),
    [anon_sym_AMP2] = ACTIONS(2930),
    [anon_sym_STAR2] = ACTIONS(2930),
    [anon_sym_last] = ACTIONS(2930),
    [anon_sym_next] = ACTIONS(2930),
    [anon_sym_redo] = ACTIONS(2930),
    [anon_sym___FILE__] = ACTIONS(2930),
    [anon_sym___LINE__] = ACTIONS(2930),
    [anon_sym___PACKAGE__] = ACTIONS(2930),
    [anon_sym___SUB__] = ACTIONS(2930),
    [anon_sym_break] = ACTIONS(2930),
    [anon_sym_fork] = ACTIONS(2930),
    [anon_sym_getppid] = ACTIONS(2930),
    [anon_sym_time] = ACTIONS(2930),
    [anon_sym_times] = ACTIONS(2930),
    [anon_sym_wait] = ACTIONS(2930),
    [anon_sym_wantarray] = ACTIONS(2930),
    [anon_sym_abs] = ACTIONS(2930),
    [anon_sym_alarm] = ACTIONS(2930),
    [anon_sym_chop] = ACTIONS(2930),
    [anon_sym_chdir] = ACTIONS(2930),
    [anon_sym_close] = ACTIONS(2930),
    [anon_sym_closedir] = ACTIONS(2930),
    [anon_sym_caller] = ACTIONS(2930),
    [anon_sym_chomp] = ACTIONS(2930),
    [anon_sym_chr] = ACTIONS(2930),
    [anon_sym_cos] = ACTIONS(2930),
    [anon_sym_chroot] = ACTIONS(2930),
    [anon_sym_defined] = ACTIONS(2930),
    [anon_sym_delete] = ACTIONS(2930),
    [anon_sym_dbmclose] = ACTIONS(2930),
    [anon_sym_exists] = ACTIONS(2930),
    [anon_sym_exit] = ACTIONS(2930),
    [anon_sym_eof] = ACTIONS(2930),
    [anon_sym_exp] = ACTIONS(2930),
    [anon_sym_each] = ACTIONS(2930),
    [anon_sym_fc] = ACTIONS(2930),
    [anon_sym_fileno] = ACTIONS(2930),
    [anon_sym_gmtime] = ACTIONS(2930),
    [anon_sym_getc] = ACTIONS(2930),
    [anon_sym_getpgrp] = ACTIONS(2930),
    [anon_sym_getprotobyname] = ACTIONS(2930),
    [anon_sym_getpwname] = ACTIONS(2930),
    [anon_sym_getpwuid] = ACTIONS(2930),
    [anon_sym_getpeername] = ACTIONS(2930),
    [anon_sym_getnetbyname] = ACTIONS(2930),
    [anon_sym_getsockname] = ACTIONS(2930),
    [anon_sym_getgrnam] = ACTIONS(2930),
    [anon_sym_getgrgid] = ACTIONS(2930),
    [anon_sym_hex] = ACTIONS(2930),
    [anon_sym_int] = ACTIONS(2930),
    [anon_sym_keys] = ACTIONS(2930),
    [anon_sym_lc] = ACTIONS(2930),
    [anon_sym_lcfirst] = ACTIONS(2930),
    [anon_sym_length] = ACTIONS(2930),
    [anon_sym_localtime] = ACTIONS(2930),
    [anon_sym_log] = ACTIONS(2930),
    [anon_sym_lock] = ACTIONS(2930),
    [anon_sym_lstat] = ACTIONS(2930),
    [anon_sym_oct] = ACTIONS(2930),
    [anon_sym_ord] = ACTIONS(2930),
    [anon_sym_prototype] = ACTIONS(2930),
    [anon_sym_pop] = ACTIONS(2930),
    [anon_sym_pos] = ACTIONS(2930),
    [anon_sym_quotemeta] = ACTIONS(2930),
    [anon_sym_reset] = ACTIONS(2930),
    [anon_sym_rand] = ACTIONS(2930),
    [anon_sym_rmdir] = ACTIONS(2930),
    [anon_sym_readdir] = ACTIONS(2930),
    [anon_sym_readline] = ACTIONS(2930),
    [anon_sym_readpipe] = ACTIONS(2930),
    [anon_sym_rewinddir] = ACTIONS(2930),
    [anon_sym_readlink] = ACTIONS(2930),
    [anon_sym_ref] = ACTIONS(2930),
    [anon_sym_scalar] = ACTIONS(2930),
    [anon_sym_shift] = ACTIONS(2930),
    [anon_sym_sin] = ACTIONS(2930),
    [anon_sym_sleep] = ACTIONS(2930),
    [anon_sym_sqrt] = ACTIONS(2930),
    [anon_sym_srand] = ACTIONS(2930),
    [anon_sym_stat] = ACTIONS(2930),
    [anon_sym_study] = ACTIONS(2930),
    [anon_sym_tell] = ACTIONS(2930),
    [anon_sym_telldir] = ACTIONS(2930),
    [anon_sym_tied] = ACTIONS(2930),
    [anon_sym_uc] = ACTIONS(2930),
    [anon_sym_ucfirst] = ACTIONS(2930),
    [anon_sym_untie] = ACTIONS(2930),
    [anon_sym_umask] = ACTIONS(2930),
    [anon_sym_values] = ACTIONS(2930),
    [anon_sym_write] = ACTIONS(2930),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2930),
    [anon_sym_qq] = ACTIONS(2930),
    [anon_sym_qw] = ACTIONS(2930),
    [anon_sym_qx] = ACTIONS(2930),
    [anon_sym_qr] = ACTIONS(2930),
    [anon_sym_m] = ACTIONS(2930),
    [anon_sym_s] = ACTIONS(2930),
    [anon_sym_tr] = ACTIONS(2930),
    [anon_sym_y] = ACTIONS(2930),
    [aux_sym__bareword_token1] = ACTIONS(2930),
    [sym_number] = ACTIONS(2930),
    [anon_sym_true] = ACTIONS(2930),
    [anon_sym_false] = ACTIONS(2930),
    [sym__single_quote] = ACTIONS(2928),
    [sym__double_quote] = ACTIONS(2928),
    [sym__backtick_quote] = ACTIONS(2928),
    [sym__search_slash_quote] = ACTIONS(2928),
    [sym__open_readline_bracket] = ACTIONS(2928),
    [sym__open_fileglob_bracket] = ACTIONS(2928),
    [sym__PERLY_HEREDOC] = ACTIONS(2928),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2928),
    [sym__filetest] = ACTIONS(2928),
  },
  [STATE(1667)] = {
    [sym_heredoc_content] = STATE(1667),
    [sym__identifier] = ACTIONS(2934),
    [anon_sym_LBRACE] = ACTIONS(2934),
    [anon_sym_DOLLAR] = ACTIONS(2934),
    [anon_sym_AT] = ACTIONS(2934),
    [anon_sym_RPAREN] = ACTIONS(2934),
    [anon_sym_my] = ACTIONS(2934),
    [anon_sym_extended] = ACTIONS(2934),
    [anon_sym_async] = ACTIONS(2934),
    [anon_sym_sub] = ACTIONS(2934),
    [anon_sym_method] = ACTIONS(2934),
    [anon_sym_LPAREN] = ACTIONS(2934),
    [anon_sym_continue] = ACTIONS(2934),
    [anon_sym_state] = ACTIONS(2934),
    [anon_sym_our] = ACTIONS(2934),
    [anon_sym_LBRACK] = ACTIONS(2934),
    [anon_sym_LT_LT] = ACTIONS(2934),
    [anon_sym_SLASH_SLASH] = ACTIONS(2934),
    [anon_sym_PLUS] = ACTIONS(2934),
    [anon_sym_DASH] = ACTIONS(2934),
    [anon_sym_TILDE] = ACTIONS(2934),
    [anon_sym_BANG] = ACTIONS(2934),
    [anon_sym_PLUS_PLUS] = ACTIONS(2934),
    [anon_sym_DASH_DASH] = ACTIONS(2934),
    [anon_sym_BSLASH] = ACTIONS(2932),
    [anon_sym_do] = ACTIONS(2934),
    [anon_sym_eval] = ACTIONS(2934),
    [anon_sym_await] = ACTIONS(2934),
    [anon_sym_field] = ACTIONS(2934),
    [anon_sym_local] = ACTIONS(2934),
    [anon_sym_dynamically] = ACTIONS(2934),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2934),
    [anon_sym_require] = ACTIONS(2934),
    [anon_sym_map] = ACTIONS(2934),
    [anon_sym_grep] = ACTIONS(2934),
    [anon_sym_sort] = ACTIONS(2934),
    [anon_sym_goto] = ACTIONS(2934),
    [anon_sym_return] = ACTIONS(2934),
    [anon_sym_undef] = ACTIONS(2934),
    [anon_sym_PERCENT2] = ACTIONS(2934),
    [anon_sym_AMP2] = ACTIONS(2934),
    [anon_sym_STAR2] = ACTIONS(2934),
    [anon_sym_last] = ACTIONS(2934),
    [anon_sym_next] = ACTIONS(2934),
    [anon_sym_redo] = ACTIONS(2934),
    [anon_sym___FILE__] = ACTIONS(2934),
    [anon_sym___LINE__] = ACTIONS(2934),
    [anon_sym___PACKAGE__] = ACTIONS(2934),
    [anon_sym___SUB__] = ACTIONS(2934),
    [anon_sym_break] = ACTIONS(2934),
    [anon_sym_fork] = ACTIONS(2934),
    [anon_sym_getppid] = ACTIONS(2934),
    [anon_sym_time] = ACTIONS(2934),
    [anon_sym_times] = ACTIONS(2934),
    [anon_sym_wait] = ACTIONS(2934),
    [anon_sym_wantarray] = ACTIONS(2934),
    [anon_sym_abs] = ACTIONS(2934),
    [anon_sym_alarm] = ACTIONS(2934),
    [anon_sym_chop] = ACTIONS(2934),
    [anon_sym_chdir] = ACTIONS(2934),
    [anon_sym_close] = ACTIONS(2934),
    [anon_sym_closedir] = ACTIONS(2934),
    [anon_sym_caller] = ACTIONS(2934),
    [anon_sym_chomp] = ACTIONS(2934),
    [anon_sym_chr] = ACTIONS(2934),
    [anon_sym_cos] = ACTIONS(2934),
    [anon_sym_chroot] = ACTIONS(2934),
    [anon_sym_defined] = ACTIONS(2934),
    [anon_sym_delete] = ACTIONS(2934),
    [anon_sym_dbmclose] = ACTIONS(2934),
    [anon_sym_exists] = ACTIONS(2934),
    [anon_sym_exit] = ACTIONS(2934),
    [anon_sym_eof] = ACTIONS(2934),
    [anon_sym_exp] = ACTIONS(2934),
    [anon_sym_each] = ACTIONS(2934),
    [anon_sym_fc] = ACTIONS(2934),
    [anon_sym_fileno] = ACTIONS(2934),
    [anon_sym_gmtime] = ACTIONS(2934),
    [anon_sym_getc] = ACTIONS(2934),
    [anon_sym_getpgrp] = ACTIONS(2934),
    [anon_sym_getprotobyname] = ACTIONS(2934),
    [anon_sym_getpwname] = ACTIONS(2934),
    [anon_sym_getpwuid] = ACTIONS(2934),
    [anon_sym_getpeername] = ACTIONS(2934),
    [anon_sym_getnetbyname] = ACTIONS(2934),
    [anon_sym_getsockname] = ACTIONS(2934),
    [anon_sym_getgrnam] = ACTIONS(2934),
    [anon_sym_getgrgid] = ACTIONS(2934),
    [anon_sym_hex] = ACTIONS(2934),
    [anon_sym_int] = ACTIONS(2934),
    [anon_sym_keys] = ACTIONS(2934),
    [anon_sym_lc] = ACTIONS(2934),
    [anon_sym_lcfirst] = ACTIONS(2934),
    [anon_sym_length] = ACTIONS(2934),
    [anon_sym_localtime] = ACTIONS(2934),
    [anon_sym_log] = ACTIONS(2934),
    [anon_sym_lock] = ACTIONS(2934),
    [anon_sym_lstat] = ACTIONS(2934),
    [anon_sym_oct] = ACTIONS(2934),
    [anon_sym_ord] = ACTIONS(2934),
    [anon_sym_prototype] = ACTIONS(2934),
    [anon_sym_pop] = ACTIONS(2934),
    [anon_sym_pos] = ACTIONS(2934),
    [anon_sym_quotemeta] = ACTIONS(2934),
    [anon_sym_reset] = ACTIONS(2934),
    [anon_sym_rand] = ACTIONS(2934),
    [anon_sym_rmdir] = ACTIONS(2934),
    [anon_sym_readdir] = ACTIONS(2934),
    [anon_sym_readline] = ACTIONS(2934),
    [anon_sym_readpipe] = ACTIONS(2934),
    [anon_sym_rewinddir] = ACTIONS(2934),
    [anon_sym_readlink] = ACTIONS(2934),
    [anon_sym_ref] = ACTIONS(2934),
    [anon_sym_scalar] = ACTIONS(2934),
    [anon_sym_shift] = ACTIONS(2934),
    [anon_sym_sin] = ACTIONS(2934),
    [anon_sym_sleep] = ACTIONS(2934),
    [anon_sym_sqrt] = ACTIONS(2934),
    [anon_sym_srand] = ACTIONS(2934),
    [anon_sym_stat] = ACTIONS(2934),
    [anon_sym_study] = ACTIONS(2934),
    [anon_sym_tell] = ACTIONS(2934),
    [anon_sym_telldir] = ACTIONS(2934),
    [anon_sym_tied] = ACTIONS(2934),
    [anon_sym_uc] = ACTIONS(2934),
    [anon_sym_ucfirst] = ACTIONS(2934),
    [anon_sym_untie] = ACTIONS(2934),
    [anon_sym_umask] = ACTIONS(2934),
    [anon_sym_values] = ACTIONS(2934),
    [anon_sym_write] = ACTIONS(2934),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2934),
    [anon_sym_qq] = ACTIONS(2934),
    [anon_sym_qw] = ACTIONS(2934),
    [anon_sym_qx] = ACTIONS(2934),
    [anon_sym_qr] = ACTIONS(2934),
    [anon_sym_m] = ACTIONS(2934),
    [anon_sym_s] = ACTIONS(2934),
    [anon_sym_tr] = ACTIONS(2934),
    [anon_sym_y] = ACTIONS(2934),
    [aux_sym__bareword_token1] = ACTIONS(2934),
    [sym_number] = ACTIONS(2934),
    [anon_sym_true] = ACTIONS(2934),
    [anon_sym_false] = ACTIONS(2934),
    [sym__single_quote] = ACTIONS(2932),
    [sym__double_quote] = ACTIONS(2932),
    [sym__backtick_quote] = ACTIONS(2932),
    [sym__search_slash_quote] = ACTIONS(2932),
    [sym__open_readline_bracket] = ACTIONS(2932),
    [sym__open_fileglob_bracket] = ACTIONS(2932),
    [sym__PERLY_HEREDOC] = ACTIONS(2932),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2932),
    [sym__filetest] = ACTIONS(2932),
  },
  [STATE(1668)] = {
    [sym_heredoc_content] = STATE(1668),
    [sym__identifier] = ACTIONS(2938),
    [anon_sym_LBRACE] = ACTIONS(2938),
    [anon_sym_DOLLAR] = ACTIONS(2938),
    [anon_sym_AT] = ACTIONS(2938),
    [anon_sym_RPAREN] = ACTIONS(2938),
    [anon_sym_my] = ACTIONS(2938),
    [anon_sym_extended] = ACTIONS(2938),
    [anon_sym_async] = ACTIONS(2938),
    [anon_sym_sub] = ACTIONS(2938),
    [anon_sym_method] = ACTIONS(2938),
    [anon_sym_LPAREN] = ACTIONS(2938),
    [anon_sym_continue] = ACTIONS(2938),
    [anon_sym_state] = ACTIONS(2938),
    [anon_sym_our] = ACTIONS(2938),
    [anon_sym_LBRACK] = ACTIONS(2938),
    [anon_sym_LT_LT] = ACTIONS(2938),
    [anon_sym_SLASH_SLASH] = ACTIONS(2938),
    [anon_sym_PLUS] = ACTIONS(2938),
    [anon_sym_DASH] = ACTIONS(2938),
    [anon_sym_TILDE] = ACTIONS(2938),
    [anon_sym_BANG] = ACTIONS(2938),
    [anon_sym_PLUS_PLUS] = ACTIONS(2938),
    [anon_sym_DASH_DASH] = ACTIONS(2938),
    [anon_sym_BSLASH] = ACTIONS(2936),
    [anon_sym_do] = ACTIONS(2938),
    [anon_sym_eval] = ACTIONS(2938),
    [anon_sym_await] = ACTIONS(2938),
    [anon_sym_field] = ACTIONS(2938),
    [anon_sym_local] = ACTIONS(2938),
    [anon_sym_dynamically] = ACTIONS(2938),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2938),
    [anon_sym_require] = ACTIONS(2938),
    [anon_sym_map] = ACTIONS(2938),
    [anon_sym_grep] = ACTIONS(2938),
    [anon_sym_sort] = ACTIONS(2938),
    [anon_sym_goto] = ACTIONS(2938),
    [anon_sym_return] = ACTIONS(2938),
    [anon_sym_undef] = ACTIONS(2938),
    [anon_sym_PERCENT2] = ACTIONS(2938),
    [anon_sym_AMP2] = ACTIONS(2938),
    [anon_sym_STAR2] = ACTIONS(2938),
    [anon_sym_last] = ACTIONS(2938),
    [anon_sym_next] = ACTIONS(2938),
    [anon_sym_redo] = ACTIONS(2938),
    [anon_sym___FILE__] = ACTIONS(2938),
    [anon_sym___LINE__] = ACTIONS(2938),
    [anon_sym___PACKAGE__] = ACTIONS(2938),
    [anon_sym___SUB__] = ACTIONS(2938),
    [anon_sym_break] = ACTIONS(2938),
    [anon_sym_fork] = ACTIONS(2938),
    [anon_sym_getppid] = ACTIONS(2938),
    [anon_sym_time] = ACTIONS(2938),
    [anon_sym_times] = ACTIONS(2938),
    [anon_sym_wait] = ACTIONS(2938),
    [anon_sym_wantarray] = ACTIONS(2938),
    [anon_sym_abs] = ACTIONS(2938),
    [anon_sym_alarm] = ACTIONS(2938),
    [anon_sym_chop] = ACTIONS(2938),
    [anon_sym_chdir] = ACTIONS(2938),
    [anon_sym_close] = ACTIONS(2938),
    [anon_sym_closedir] = ACTIONS(2938),
    [anon_sym_caller] = ACTIONS(2938),
    [anon_sym_chomp] = ACTIONS(2938),
    [anon_sym_chr] = ACTIONS(2938),
    [anon_sym_cos] = ACTIONS(2938),
    [anon_sym_chroot] = ACTIONS(2938),
    [anon_sym_defined] = ACTIONS(2938),
    [anon_sym_delete] = ACTIONS(2938),
    [anon_sym_dbmclose] = ACTIONS(2938),
    [anon_sym_exists] = ACTIONS(2938),
    [anon_sym_exit] = ACTIONS(2938),
    [anon_sym_eof] = ACTIONS(2938),
    [anon_sym_exp] = ACTIONS(2938),
    [anon_sym_each] = ACTIONS(2938),
    [anon_sym_fc] = ACTIONS(2938),
    [anon_sym_fileno] = ACTIONS(2938),
    [anon_sym_gmtime] = ACTIONS(2938),
    [anon_sym_getc] = ACTIONS(2938),
    [anon_sym_getpgrp] = ACTIONS(2938),
    [anon_sym_getprotobyname] = ACTIONS(2938),
    [anon_sym_getpwname] = ACTIONS(2938),
    [anon_sym_getpwuid] = ACTIONS(2938),
    [anon_sym_getpeername] = ACTIONS(2938),
    [anon_sym_getnetbyname] = ACTIONS(2938),
    [anon_sym_getsockname] = ACTIONS(2938),
    [anon_sym_getgrnam] = ACTIONS(2938),
    [anon_sym_getgrgid] = ACTIONS(2938),
    [anon_sym_hex] = ACTIONS(2938),
    [anon_sym_int] = ACTIONS(2938),
    [anon_sym_keys] = ACTIONS(2938),
    [anon_sym_lc] = ACTIONS(2938),
    [anon_sym_lcfirst] = ACTIONS(2938),
    [anon_sym_length] = ACTIONS(2938),
    [anon_sym_localtime] = ACTIONS(2938),
    [anon_sym_log] = ACTIONS(2938),
    [anon_sym_lock] = ACTIONS(2938),
    [anon_sym_lstat] = ACTIONS(2938),
    [anon_sym_oct] = ACTIONS(2938),
    [anon_sym_ord] = ACTIONS(2938),
    [anon_sym_prototype] = ACTIONS(2938),
    [anon_sym_pop] = ACTIONS(2938),
    [anon_sym_pos] = ACTIONS(2938),
    [anon_sym_quotemeta] = ACTIONS(2938),
    [anon_sym_reset] = ACTIONS(2938),
    [anon_sym_rand] = ACTIONS(2938),
    [anon_sym_rmdir] = ACTIONS(2938),
    [anon_sym_readdir] = ACTIONS(2938),
    [anon_sym_readline] = ACTIONS(2938),
    [anon_sym_readpipe] = ACTIONS(2938),
    [anon_sym_rewinddir] = ACTIONS(2938),
    [anon_sym_readlink] = ACTIONS(2938),
    [anon_sym_ref] = ACTIONS(2938),
    [anon_sym_scalar] = ACTIONS(2938),
    [anon_sym_shift] = ACTIONS(2938),
    [anon_sym_sin] = ACTIONS(2938),
    [anon_sym_sleep] = ACTIONS(2938),
    [anon_sym_sqrt] = ACTIONS(2938),
    [anon_sym_srand] = ACTIONS(2938),
    [anon_sym_stat] = ACTIONS(2938),
    [anon_sym_study] = ACTIONS(2938),
    [anon_sym_tell] = ACTIONS(2938),
    [anon_sym_telldir] = ACTIONS(2938),
    [anon_sym_tied] = ACTIONS(2938),
    [anon_sym_uc] = ACTIONS(2938),
    [anon_sym_ucfirst] = ACTIONS(2938),
    [anon_sym_untie] = ACTIONS(2938),
    [anon_sym_umask] = ACTIONS(2938),
    [anon_sym_values] = ACTIONS(2938),
    [anon_sym_write] = ACTIONS(2938),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2938),
    [anon_sym_qq] = ACTIONS(2938),
    [anon_sym_qw] = ACTIONS(2938),
    [anon_sym_qx] = ACTIONS(2938),
    [anon_sym_qr] = ACTIONS(2938),
    [anon_sym_m] = ACTIONS(2938),
    [anon_sym_s] = ACTIONS(2938),
    [anon_sym_tr] = ACTIONS(2938),
    [anon_sym_y] = ACTIONS(2938),
    [aux_sym__bareword_token1] = ACTIONS(2938),
    [sym_number] = ACTIONS(2938),
    [anon_sym_true] = ACTIONS(2938),
    [anon_sym_false] = ACTIONS(2938),
    [sym__single_quote] = ACTIONS(2936),
    [sym__double_quote] = ACTIONS(2936),
    [sym__backtick_quote] = ACTIONS(2936),
    [sym__search_slash_quote] = ACTIONS(2936),
    [sym__open_readline_bracket] = ACTIONS(2936),
    [sym__open_fileglob_bracket] = ACTIONS(2936),
    [sym__PERLY_HEREDOC] = ACTIONS(2936),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2936),
    [sym__filetest] = ACTIONS(2936),
  },
  [STATE(1669)] = {
    [sym_heredoc_content] = STATE(1669),
    [sym__identifier] = ACTIONS(2954),
    [anon_sym_LBRACE] = ACTIONS(2954),
    [anon_sym_DOLLAR] = ACTIONS(2954),
    [anon_sym_AT] = ACTIONS(2954),
    [anon_sym_RPAREN] = ACTIONS(2954),
    [anon_sym_my] = ACTIONS(2954),
    [anon_sym_extended] = ACTIONS(2954),
    [anon_sym_async] = ACTIONS(2954),
    [anon_sym_sub] = ACTIONS(2954),
    [anon_sym_method] = ACTIONS(2954),
    [anon_sym_LPAREN] = ACTIONS(2954),
    [anon_sym_continue] = ACTIONS(2954),
    [anon_sym_state] = ACTIONS(2954),
    [anon_sym_our] = ACTIONS(2954),
    [anon_sym_LBRACK] = ACTIONS(2954),
    [anon_sym_LT_LT] = ACTIONS(2954),
    [anon_sym_SLASH_SLASH] = ACTIONS(2954),
    [anon_sym_PLUS] = ACTIONS(2954),
    [anon_sym_DASH] = ACTIONS(2954),
    [anon_sym_TILDE] = ACTIONS(2954),
    [anon_sym_BANG] = ACTIONS(2954),
    [anon_sym_PLUS_PLUS] = ACTIONS(2954),
    [anon_sym_DASH_DASH] = ACTIONS(2954),
    [anon_sym_BSLASH] = ACTIONS(2952),
    [anon_sym_do] = ACTIONS(2954),
    [anon_sym_eval] = ACTIONS(2954),
    [anon_sym_await] = ACTIONS(2954),
    [anon_sym_field] = ACTIONS(2954),
    [anon_sym_local] = ACTIONS(2954),
    [anon_sym_dynamically] = ACTIONS(2954),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2954),
    [anon_sym_require] = ACTIONS(2954),
    [anon_sym_map] = ACTIONS(2954),
    [anon_sym_grep] = ACTIONS(2954),
    [anon_sym_sort] = ACTIONS(2954),
    [anon_sym_goto] = ACTIONS(2954),
    [anon_sym_return] = ACTIONS(2954),
    [anon_sym_undef] = ACTIONS(2954),
    [anon_sym_PERCENT2] = ACTIONS(2954),
    [anon_sym_AMP2] = ACTIONS(2954),
    [anon_sym_STAR2] = ACTIONS(2954),
    [anon_sym_last] = ACTIONS(2954),
    [anon_sym_next] = ACTIONS(2954),
    [anon_sym_redo] = ACTIONS(2954),
    [anon_sym___FILE__] = ACTIONS(2954),
    [anon_sym___LINE__] = ACTIONS(2954),
    [anon_sym___PACKAGE__] = ACTIONS(2954),
    [anon_sym___SUB__] = ACTIONS(2954),
    [anon_sym_break] = ACTIONS(2954),
    [anon_sym_fork] = ACTIONS(2954),
    [anon_sym_getppid] = ACTIONS(2954),
    [anon_sym_time] = ACTIONS(2954),
    [anon_sym_times] = ACTIONS(2954),
    [anon_sym_wait] = ACTIONS(2954),
    [anon_sym_wantarray] = ACTIONS(2954),
    [anon_sym_abs] = ACTIONS(2954),
    [anon_sym_alarm] = ACTIONS(2954),
    [anon_sym_chop] = ACTIONS(2954),
    [anon_sym_chdir] = ACTIONS(2954),
    [anon_sym_close] = ACTIONS(2954),
    [anon_sym_closedir] = ACTIONS(2954),
    [anon_sym_caller] = ACTIONS(2954),
    [anon_sym_chomp] = ACTIONS(2954),
    [anon_sym_chr] = ACTIONS(2954),
    [anon_sym_cos] = ACTIONS(2954),
    [anon_sym_chroot] = ACTIONS(2954),
    [anon_sym_defined] = ACTIONS(2954),
    [anon_sym_delete] = ACTIONS(2954),
    [anon_sym_dbmclose] = ACTIONS(2954),
    [anon_sym_exists] = ACTIONS(2954),
    [anon_sym_exit] = ACTIONS(2954),
    [anon_sym_eof] = ACTIONS(2954),
    [anon_sym_exp] = ACTIONS(2954),
    [anon_sym_each] = ACTIONS(2954),
    [anon_sym_fc] = ACTIONS(2954),
    [anon_sym_fileno] = ACTIONS(2954),
    [anon_sym_gmtime] = ACTIONS(2954),
    [anon_sym_getc] = ACTIONS(2954),
    [anon_sym_getpgrp] = ACTIONS(2954),
    [anon_sym_getprotobyname] = ACTIONS(2954),
    [anon_sym_getpwname] = ACTIONS(2954),
    [anon_sym_getpwuid] = ACTIONS(2954),
    [anon_sym_getpeername] = ACTIONS(2954),
    [anon_sym_getnetbyname] = ACTIONS(2954),
    [anon_sym_getsockname] = ACTIONS(2954),
    [anon_sym_getgrnam] = ACTIONS(2954),
    [anon_sym_getgrgid] = ACTIONS(2954),
    [anon_sym_hex] = ACTIONS(2954),
    [anon_sym_int] = ACTIONS(2954),
    [anon_sym_keys] = ACTIONS(2954),
    [anon_sym_lc] = ACTIONS(2954),
    [anon_sym_lcfirst] = ACTIONS(2954),
    [anon_sym_length] = ACTIONS(2954),
    [anon_sym_localtime] = ACTIONS(2954),
    [anon_sym_log] = ACTIONS(2954),
    [anon_sym_lock] = ACTIONS(2954),
    [anon_sym_lstat] = ACTIONS(2954),
    [anon_sym_oct] = ACTIONS(2954),
    [anon_sym_ord] = ACTIONS(2954),
    [anon_sym_prototype] = ACTIONS(2954),
    [anon_sym_pop] = ACTIONS(2954),
    [anon_sym_pos] = ACTIONS(2954),
    [anon_sym_quotemeta] = ACTIONS(2954),
    [anon_sym_reset] = ACTIONS(2954),
    [anon_sym_rand] = ACTIONS(2954),
    [anon_sym_rmdir] = ACTIONS(2954),
    [anon_sym_readdir] = ACTIONS(2954),
    [anon_sym_readline] = ACTIONS(2954),
    [anon_sym_readpipe] = ACTIONS(2954),
    [anon_sym_rewinddir] = ACTIONS(2954),
    [anon_sym_readlink] = ACTIONS(2954),
    [anon_sym_ref] = ACTIONS(2954),
    [anon_sym_scalar] = ACTIONS(2954),
    [anon_sym_shift] = ACTIONS(2954),
    [anon_sym_sin] = ACTIONS(2954),
    [anon_sym_sleep] = ACTIONS(2954),
    [anon_sym_sqrt] = ACTIONS(2954),
    [anon_sym_srand] = ACTIONS(2954),
    [anon_sym_stat] = ACTIONS(2954),
    [anon_sym_study] = ACTIONS(2954),
    [anon_sym_tell] = ACTIONS(2954),
    [anon_sym_telldir] = ACTIONS(2954),
    [anon_sym_tied] = ACTIONS(2954),
    [anon_sym_uc] = ACTIONS(2954),
    [anon_sym_ucfirst] = ACTIONS(2954),
    [anon_sym_untie] = ACTIONS(2954),
    [anon_sym_umask] = ACTIONS(2954),
    [anon_sym_values] = ACTIONS(2954),
    [anon_sym_write] = ACTIONS(2954),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2954),
    [anon_sym_qq] = ACTIONS(2954),
    [anon_sym_qw] = ACTIONS(2954),
    [anon_sym_qx] = ACTIONS(2954),
    [anon_sym_qr] = ACTIONS(2954),
    [anon_sym_m] = ACTIONS(2954),
    [anon_sym_s] = ACTIONS(2954),
    [anon_sym_tr] = ACTIONS(2954),
    [anon_sym_y] = ACTIONS(2954),
    [aux_sym__bareword_token1] = ACTIONS(2954),
    [sym_number] = ACTIONS(2954),
    [anon_sym_true] = ACTIONS(2954),
    [anon_sym_false] = ACTIONS(2954),
    [sym__single_quote] = ACTIONS(2952),
    [sym__double_quote] = ACTIONS(2952),
    [sym__backtick_quote] = ACTIONS(2952),
    [sym__search_slash_quote] = ACTIONS(2952),
    [sym__open_readline_bracket] = ACTIONS(2952),
    [sym__open_fileglob_bracket] = ACTIONS(2952),
    [sym__PERLY_HEREDOC] = ACTIONS(2952),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2952),
    [sym__filetest] = ACTIONS(2952),
  },
  [STATE(1670)] = {
    [sym_heredoc_content] = STATE(1670),
    [sym__identifier] = ACTIONS(2986),
    [anon_sym_LBRACE] = ACTIONS(2986),
    [anon_sym_DOLLAR] = ACTIONS(2986),
    [anon_sym_AT] = ACTIONS(2986),
    [anon_sym_RPAREN] = ACTIONS(2986),
    [anon_sym_my] = ACTIONS(2986),
    [anon_sym_extended] = ACTIONS(2986),
    [anon_sym_async] = ACTIONS(2986),
    [anon_sym_sub] = ACTIONS(2986),
    [anon_sym_method] = ACTIONS(2986),
    [anon_sym_LPAREN] = ACTIONS(2986),
    [anon_sym_continue] = ACTIONS(2986),
    [anon_sym_state] = ACTIONS(2986),
    [anon_sym_our] = ACTIONS(2986),
    [anon_sym_LBRACK] = ACTIONS(2986),
    [anon_sym_LT_LT] = ACTIONS(2986),
    [anon_sym_SLASH_SLASH] = ACTIONS(2986),
    [anon_sym_PLUS] = ACTIONS(2986),
    [anon_sym_DASH] = ACTIONS(2986),
    [anon_sym_TILDE] = ACTIONS(2986),
    [anon_sym_BANG] = ACTIONS(2986),
    [anon_sym_PLUS_PLUS] = ACTIONS(2986),
    [anon_sym_DASH_DASH] = ACTIONS(2986),
    [anon_sym_BSLASH] = ACTIONS(2984),
    [anon_sym_do] = ACTIONS(2986),
    [anon_sym_eval] = ACTIONS(2986),
    [anon_sym_await] = ACTIONS(2986),
    [anon_sym_field] = ACTIONS(2986),
    [anon_sym_local] = ACTIONS(2986),
    [anon_sym_dynamically] = ACTIONS(2986),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2986),
    [anon_sym_require] = ACTIONS(2986),
    [anon_sym_map] = ACTIONS(2986),
    [anon_sym_grep] = ACTIONS(2986),
    [anon_sym_sort] = ACTIONS(2986),
    [anon_sym_goto] = ACTIONS(2986),
    [anon_sym_return] = ACTIONS(2986),
    [anon_sym_undef] = ACTIONS(2986),
    [anon_sym_PERCENT2] = ACTIONS(2986),
    [anon_sym_AMP2] = ACTIONS(2986),
    [anon_sym_STAR2] = ACTIONS(2986),
    [anon_sym_last] = ACTIONS(2986),
    [anon_sym_next] = ACTIONS(2986),
    [anon_sym_redo] = ACTIONS(2986),
    [anon_sym___FILE__] = ACTIONS(2986),
    [anon_sym___LINE__] = ACTIONS(2986),
    [anon_sym___PACKAGE__] = ACTIONS(2986),
    [anon_sym___SUB__] = ACTIONS(2986),
    [anon_sym_break] = ACTIONS(2986),
    [anon_sym_fork] = ACTIONS(2986),
    [anon_sym_getppid] = ACTIONS(2986),
    [anon_sym_time] = ACTIONS(2986),
    [anon_sym_times] = ACTIONS(2986),
    [anon_sym_wait] = ACTIONS(2986),
    [anon_sym_wantarray] = ACTIONS(2986),
    [anon_sym_abs] = ACTIONS(2986),
    [anon_sym_alarm] = ACTIONS(2986),
    [anon_sym_chop] = ACTIONS(2986),
    [anon_sym_chdir] = ACTIONS(2986),
    [anon_sym_close] = ACTIONS(2986),
    [anon_sym_closedir] = ACTIONS(2986),
    [anon_sym_caller] = ACTIONS(2986),
    [anon_sym_chomp] = ACTIONS(2986),
    [anon_sym_chr] = ACTIONS(2986),
    [anon_sym_cos] = ACTIONS(2986),
    [anon_sym_chroot] = ACTIONS(2986),
    [anon_sym_defined] = ACTIONS(2986),
    [anon_sym_delete] = ACTIONS(2986),
    [anon_sym_dbmclose] = ACTIONS(2986),
    [anon_sym_exists] = ACTIONS(2986),
    [anon_sym_exit] = ACTIONS(2986),
    [anon_sym_eof] = ACTIONS(2986),
    [anon_sym_exp] = ACTIONS(2986),
    [anon_sym_each] = ACTIONS(2986),
    [anon_sym_fc] = ACTIONS(2986),
    [anon_sym_fileno] = ACTIONS(2986),
    [anon_sym_gmtime] = ACTIONS(2986),
    [anon_sym_getc] = ACTIONS(2986),
    [anon_sym_getpgrp] = ACTIONS(2986),
    [anon_sym_getprotobyname] = ACTIONS(2986),
    [anon_sym_getpwname] = ACTIONS(2986),
    [anon_sym_getpwuid] = ACTIONS(2986),
    [anon_sym_getpeername] = ACTIONS(2986),
    [anon_sym_getnetbyname] = ACTIONS(2986),
    [anon_sym_getsockname] = ACTIONS(2986),
    [anon_sym_getgrnam] = ACTIONS(2986),
    [anon_sym_getgrgid] = ACTIONS(2986),
    [anon_sym_hex] = ACTIONS(2986),
    [anon_sym_int] = ACTIONS(2986),
    [anon_sym_keys] = ACTIONS(2986),
    [anon_sym_lc] = ACTIONS(2986),
    [anon_sym_lcfirst] = ACTIONS(2986),
    [anon_sym_length] = ACTIONS(2986),
    [anon_sym_localtime] = ACTIONS(2986),
    [anon_sym_log] = ACTIONS(2986),
    [anon_sym_lock] = ACTIONS(2986),
    [anon_sym_lstat] = ACTIONS(2986),
    [anon_sym_oct] = ACTIONS(2986),
    [anon_sym_ord] = ACTIONS(2986),
    [anon_sym_prototype] = ACTIONS(2986),
    [anon_sym_pop] = ACTIONS(2986),
    [anon_sym_pos] = ACTIONS(2986),
    [anon_sym_quotemeta] = ACTIONS(2986),
    [anon_sym_reset] = ACTIONS(2986),
    [anon_sym_rand] = ACTIONS(2986),
    [anon_sym_rmdir] = ACTIONS(2986),
    [anon_sym_readdir] = ACTIONS(2986),
    [anon_sym_readline] = ACTIONS(2986),
    [anon_sym_readpipe] = ACTIONS(2986),
    [anon_sym_rewinddir] = ACTIONS(2986),
    [anon_sym_readlink] = ACTIONS(2986),
    [anon_sym_ref] = ACTIONS(2986),
    [anon_sym_scalar] = ACTIONS(2986),
    [anon_sym_shift] = ACTIONS(2986),
    [anon_sym_sin] = ACTIONS(2986),
    [anon_sym_sleep] = ACTIONS(2986),
    [anon_sym_sqrt] = ACTIONS(2986),
    [anon_sym_srand] = ACTIONS(2986),
    [anon_sym_stat] = ACTIONS(2986),
    [anon_sym_study] = ACTIONS(2986),
    [anon_sym_tell] = ACTIONS(2986),
    [anon_sym_telldir] = ACTIONS(2986),
    [anon_sym_tied] = ACTIONS(2986),
    [anon_sym_uc] = ACTIONS(2986),
    [anon_sym_ucfirst] = ACTIONS(2986),
    [anon_sym_untie] = ACTIONS(2986),
    [anon_sym_umask] = ACTIONS(2986),
    [anon_sym_values] = ACTIONS(2986),
    [anon_sym_write] = ACTIONS(2986),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2986),
    [anon_sym_qq] = ACTIONS(2986),
    [anon_sym_qw] = ACTIONS(2986),
    [anon_sym_qx] = ACTIONS(2986),
    [anon_sym_qr] = ACTIONS(2986),
    [anon_sym_m] = ACTIONS(2986),
    [anon_sym_s] = ACTIONS(2986),
    [anon_sym_tr] = ACTIONS(2986),
    [anon_sym_y] = ACTIONS(2986),
    [aux_sym__bareword_token1] = ACTIONS(2986),
    [sym_number] = ACTIONS(2986),
    [anon_sym_true] = ACTIONS(2986),
    [anon_sym_false] = ACTIONS(2986),
    [sym__single_quote] = ACTIONS(2984),
    [sym__double_quote] = ACTIONS(2984),
    [sym__backtick_quote] = ACTIONS(2984),
    [sym__search_slash_quote] = ACTIONS(2984),
    [sym__open_readline_bracket] = ACTIONS(2984),
    [sym__open_fileglob_bracket] = ACTIONS(2984),
    [sym__PERLY_HEREDOC] = ACTIONS(2984),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2984),
    [sym__filetest] = ACTIONS(2984),
  },
  [STATE(1671)] = {
    [sym_heredoc_content] = STATE(1671),
    [sym__identifier] = ACTIONS(2990),
    [anon_sym_LBRACE] = ACTIONS(2990),
    [anon_sym_DOLLAR] = ACTIONS(2990),
    [anon_sym_AT] = ACTIONS(2990),
    [anon_sym_RPAREN] = ACTIONS(2990),
    [anon_sym_my] = ACTIONS(2990),
    [anon_sym_extended] = ACTIONS(2990),
    [anon_sym_async] = ACTIONS(2990),
    [anon_sym_sub] = ACTIONS(2990),
    [anon_sym_method] = ACTIONS(2990),
    [anon_sym_LPAREN] = ACTIONS(2990),
    [anon_sym_continue] = ACTIONS(2990),
    [anon_sym_state] = ACTIONS(2990),
    [anon_sym_our] = ACTIONS(2990),
    [anon_sym_LBRACK] = ACTIONS(2990),
    [anon_sym_LT_LT] = ACTIONS(2990),
    [anon_sym_SLASH_SLASH] = ACTIONS(2990),
    [anon_sym_PLUS] = ACTIONS(2990),
    [anon_sym_DASH] = ACTIONS(2990),
    [anon_sym_TILDE] = ACTIONS(2990),
    [anon_sym_BANG] = ACTIONS(2990),
    [anon_sym_PLUS_PLUS] = ACTIONS(2990),
    [anon_sym_DASH_DASH] = ACTIONS(2990),
    [anon_sym_BSLASH] = ACTIONS(2988),
    [anon_sym_do] = ACTIONS(2990),
    [anon_sym_eval] = ACTIONS(2990),
    [anon_sym_await] = ACTIONS(2990),
    [anon_sym_field] = ACTIONS(2990),
    [anon_sym_local] = ACTIONS(2990),
    [anon_sym_dynamically] = ACTIONS(2990),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2990),
    [anon_sym_require] = ACTIONS(2990),
    [anon_sym_map] = ACTIONS(2990),
    [anon_sym_grep] = ACTIONS(2990),
    [anon_sym_sort] = ACTIONS(2990),
    [anon_sym_goto] = ACTIONS(2990),
    [anon_sym_return] = ACTIONS(2990),
    [anon_sym_undef] = ACTIONS(2990),
    [anon_sym_PERCENT2] = ACTIONS(2990),
    [anon_sym_AMP2] = ACTIONS(2990),
    [anon_sym_STAR2] = ACTIONS(2990),
    [anon_sym_last] = ACTIONS(2990),
    [anon_sym_next] = ACTIONS(2990),
    [anon_sym_redo] = ACTIONS(2990),
    [anon_sym___FILE__] = ACTIONS(2990),
    [anon_sym___LINE__] = ACTIONS(2990),
    [anon_sym___PACKAGE__] = ACTIONS(2990),
    [anon_sym___SUB__] = ACTIONS(2990),
    [anon_sym_break] = ACTIONS(2990),
    [anon_sym_fork] = ACTIONS(2990),
    [anon_sym_getppid] = ACTIONS(2990),
    [anon_sym_time] = ACTIONS(2990),
    [anon_sym_times] = ACTIONS(2990),
    [anon_sym_wait] = ACTIONS(2990),
    [anon_sym_wantarray] = ACTIONS(2990),
    [anon_sym_abs] = ACTIONS(2990),
    [anon_sym_alarm] = ACTIONS(2990),
    [anon_sym_chop] = ACTIONS(2990),
    [anon_sym_chdir] = ACTIONS(2990),
    [anon_sym_close] = ACTIONS(2990),
    [anon_sym_closedir] = ACTIONS(2990),
    [anon_sym_caller] = ACTIONS(2990),
    [anon_sym_chomp] = ACTIONS(2990),
    [anon_sym_chr] = ACTIONS(2990),
    [anon_sym_cos] = ACTIONS(2990),
    [anon_sym_chroot] = ACTIONS(2990),
    [anon_sym_defined] = ACTIONS(2990),
    [anon_sym_delete] = ACTIONS(2990),
    [anon_sym_dbmclose] = ACTIONS(2990),
    [anon_sym_exists] = ACTIONS(2990),
    [anon_sym_exit] = ACTIONS(2990),
    [anon_sym_eof] = ACTIONS(2990),
    [anon_sym_exp] = ACTIONS(2990),
    [anon_sym_each] = ACTIONS(2990),
    [anon_sym_fc] = ACTIONS(2990),
    [anon_sym_fileno] = ACTIONS(2990),
    [anon_sym_gmtime] = ACTIONS(2990),
    [anon_sym_getc] = ACTIONS(2990),
    [anon_sym_getpgrp] = ACTIONS(2990),
    [anon_sym_getprotobyname] = ACTIONS(2990),
    [anon_sym_getpwname] = ACTIONS(2990),
    [anon_sym_getpwuid] = ACTIONS(2990),
    [anon_sym_getpeername] = ACTIONS(2990),
    [anon_sym_getnetbyname] = ACTIONS(2990),
    [anon_sym_getsockname] = ACTIONS(2990),
    [anon_sym_getgrnam] = ACTIONS(2990),
    [anon_sym_getgrgid] = ACTIONS(2990),
    [anon_sym_hex] = ACTIONS(2990),
    [anon_sym_int] = ACTIONS(2990),
    [anon_sym_keys] = ACTIONS(2990),
    [anon_sym_lc] = ACTIONS(2990),
    [anon_sym_lcfirst] = ACTIONS(2990),
    [anon_sym_length] = ACTIONS(2990),
    [anon_sym_localtime] = ACTIONS(2990),
    [anon_sym_log] = ACTIONS(2990),
    [anon_sym_lock] = ACTIONS(2990),
    [anon_sym_lstat] = ACTIONS(2990),
    [anon_sym_oct] = ACTIONS(2990),
    [anon_sym_ord] = ACTIONS(2990),
    [anon_sym_prototype] = ACTIONS(2990),
    [anon_sym_pop] = ACTIONS(2990),
    [anon_sym_pos] = ACTIONS(2990),
    [anon_sym_quotemeta] = ACTIONS(2990),
    [anon_sym_reset] = ACTIONS(2990),
    [anon_sym_rand] = ACTIONS(2990),
    [anon_sym_rmdir] = ACTIONS(2990),
    [anon_sym_readdir] = ACTIONS(2990),
    [anon_sym_readline] = ACTIONS(2990),
    [anon_sym_readpipe] = ACTIONS(2990),
    [anon_sym_rewinddir] = ACTIONS(2990),
    [anon_sym_readlink] = ACTIONS(2990),
    [anon_sym_ref] = ACTIONS(2990),
    [anon_sym_scalar] = ACTIONS(2990),
    [anon_sym_shift] = ACTIONS(2990),
    [anon_sym_sin] = ACTIONS(2990),
    [anon_sym_sleep] = ACTIONS(2990),
    [anon_sym_sqrt] = ACTIONS(2990),
    [anon_sym_srand] = ACTIONS(2990),
    [anon_sym_stat] = ACTIONS(2990),
    [anon_sym_study] = ACTIONS(2990),
    [anon_sym_tell] = ACTIONS(2990),
    [anon_sym_telldir] = ACTIONS(2990),
    [anon_sym_tied] = ACTIONS(2990),
    [anon_sym_uc] = ACTIONS(2990),
    [anon_sym_ucfirst] = ACTIONS(2990),
    [anon_sym_untie] = ACTIONS(2990),
    [anon_sym_umask] = ACTIONS(2990),
    [anon_sym_values] = ACTIONS(2990),
    [anon_sym_write] = ACTIONS(2990),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2990),
    [anon_sym_qq] = ACTIONS(2990),
    [anon_sym_qw] = ACTIONS(2990),
    [anon_sym_qx] = ACTIONS(2990),
    [anon_sym_qr] = ACTIONS(2990),
    [anon_sym_m] = ACTIONS(2990),
    [anon_sym_s] = ACTIONS(2990),
    [anon_sym_tr] = ACTIONS(2990),
    [anon_sym_y] = ACTIONS(2990),
    [aux_sym__bareword_token1] = ACTIONS(2990),
    [sym_number] = ACTIONS(2990),
    [anon_sym_true] = ACTIONS(2990),
    [anon_sym_false] = ACTIONS(2990),
    [sym__single_quote] = ACTIONS(2988),
    [sym__double_quote] = ACTIONS(2988),
    [sym__backtick_quote] = ACTIONS(2988),
    [sym__search_slash_quote] = ACTIONS(2988),
    [sym__open_readline_bracket] = ACTIONS(2988),
    [sym__open_fileglob_bracket] = ACTIONS(2988),
    [sym__PERLY_HEREDOC] = ACTIONS(2988),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2988),
    [sym__filetest] = ACTIONS(2988),
  },
  [STATE(1672)] = {
    [sym_heredoc_content] = STATE(1672),
    [sym__identifier] = ACTIONS(3014),
    [anon_sym_LBRACE] = ACTIONS(3014),
    [anon_sym_DOLLAR] = ACTIONS(3014),
    [anon_sym_AT] = ACTIONS(3014),
    [anon_sym_RPAREN] = ACTIONS(3014),
    [anon_sym_my] = ACTIONS(3014),
    [anon_sym_extended] = ACTIONS(3014),
    [anon_sym_async] = ACTIONS(3014),
    [anon_sym_sub] = ACTIONS(3014),
    [anon_sym_method] = ACTIONS(3014),
    [anon_sym_LPAREN] = ACTIONS(3014),
    [anon_sym_continue] = ACTIONS(3014),
    [anon_sym_state] = ACTIONS(3014),
    [anon_sym_our] = ACTIONS(3014),
    [anon_sym_LBRACK] = ACTIONS(3014),
    [anon_sym_LT_LT] = ACTIONS(3014),
    [anon_sym_SLASH_SLASH] = ACTIONS(3014),
    [anon_sym_PLUS] = ACTIONS(3014),
    [anon_sym_DASH] = ACTIONS(3014),
    [anon_sym_TILDE] = ACTIONS(3014),
    [anon_sym_BANG] = ACTIONS(3014),
    [anon_sym_PLUS_PLUS] = ACTIONS(3014),
    [anon_sym_DASH_DASH] = ACTIONS(3014),
    [anon_sym_BSLASH] = ACTIONS(3012),
    [anon_sym_do] = ACTIONS(3014),
    [anon_sym_eval] = ACTIONS(3014),
    [anon_sym_await] = ACTIONS(3014),
    [anon_sym_field] = ACTIONS(3014),
    [anon_sym_local] = ACTIONS(3014),
    [anon_sym_dynamically] = ACTIONS(3014),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3014),
    [anon_sym_require] = ACTIONS(3014),
    [anon_sym_map] = ACTIONS(3014),
    [anon_sym_grep] = ACTIONS(3014),
    [anon_sym_sort] = ACTIONS(3014),
    [anon_sym_goto] = ACTIONS(3014),
    [anon_sym_return] = ACTIONS(3014),
    [anon_sym_undef] = ACTIONS(3014),
    [anon_sym_PERCENT2] = ACTIONS(3014),
    [anon_sym_AMP2] = ACTIONS(3014),
    [anon_sym_STAR2] = ACTIONS(3014),
    [anon_sym_last] = ACTIONS(3014),
    [anon_sym_next] = ACTIONS(3014),
    [anon_sym_redo] = ACTIONS(3014),
    [anon_sym___FILE__] = ACTIONS(3014),
    [anon_sym___LINE__] = ACTIONS(3014),
    [anon_sym___PACKAGE__] = ACTIONS(3014),
    [anon_sym___SUB__] = ACTIONS(3014),
    [anon_sym_break] = ACTIONS(3014),
    [anon_sym_fork] = ACTIONS(3014),
    [anon_sym_getppid] = ACTIONS(3014),
    [anon_sym_time] = ACTIONS(3014),
    [anon_sym_times] = ACTIONS(3014),
    [anon_sym_wait] = ACTIONS(3014),
    [anon_sym_wantarray] = ACTIONS(3014),
    [anon_sym_abs] = ACTIONS(3014),
    [anon_sym_alarm] = ACTIONS(3014),
    [anon_sym_chop] = ACTIONS(3014),
    [anon_sym_chdir] = ACTIONS(3014),
    [anon_sym_close] = ACTIONS(3014),
    [anon_sym_closedir] = ACTIONS(3014),
    [anon_sym_caller] = ACTIONS(3014),
    [anon_sym_chomp] = ACTIONS(3014),
    [anon_sym_chr] = ACTIONS(3014),
    [anon_sym_cos] = ACTIONS(3014),
    [anon_sym_chroot] = ACTIONS(3014),
    [anon_sym_defined] = ACTIONS(3014),
    [anon_sym_delete] = ACTIONS(3014),
    [anon_sym_dbmclose] = ACTIONS(3014),
    [anon_sym_exists] = ACTIONS(3014),
    [anon_sym_exit] = ACTIONS(3014),
    [anon_sym_eof] = ACTIONS(3014),
    [anon_sym_exp] = ACTIONS(3014),
    [anon_sym_each] = ACTIONS(3014),
    [anon_sym_fc] = ACTIONS(3014),
    [anon_sym_fileno] = ACTIONS(3014),
    [anon_sym_gmtime] = ACTIONS(3014),
    [anon_sym_getc] = ACTIONS(3014),
    [anon_sym_getpgrp] = ACTIONS(3014),
    [anon_sym_getprotobyname] = ACTIONS(3014),
    [anon_sym_getpwname] = ACTIONS(3014),
    [anon_sym_getpwuid] = ACTIONS(3014),
    [anon_sym_getpeername] = ACTIONS(3014),
    [anon_sym_getnetbyname] = ACTIONS(3014),
    [anon_sym_getsockname] = ACTIONS(3014),
    [anon_sym_getgrnam] = ACTIONS(3014),
    [anon_sym_getgrgid] = ACTIONS(3014),
    [anon_sym_hex] = ACTIONS(3014),
    [anon_sym_int] = ACTIONS(3014),
    [anon_sym_keys] = ACTIONS(3014),
    [anon_sym_lc] = ACTIONS(3014),
    [anon_sym_lcfirst] = ACTIONS(3014),
    [anon_sym_length] = ACTIONS(3014),
    [anon_sym_localtime] = ACTIONS(3014),
    [anon_sym_log] = ACTIONS(3014),
    [anon_sym_lock] = ACTIONS(3014),
    [anon_sym_lstat] = ACTIONS(3014),
    [anon_sym_oct] = ACTIONS(3014),
    [anon_sym_ord] = ACTIONS(3014),
    [anon_sym_prototype] = ACTIONS(3014),
    [anon_sym_pop] = ACTIONS(3014),
    [anon_sym_pos] = ACTIONS(3014),
    [anon_sym_quotemeta] = ACTIONS(3014),
    [anon_sym_reset] = ACTIONS(3014),
    [anon_sym_rand] = ACTIONS(3014),
    [anon_sym_rmdir] = ACTIONS(3014),
    [anon_sym_readdir] = ACTIONS(3014),
    [anon_sym_readline] = ACTIONS(3014),
    [anon_sym_readpipe] = ACTIONS(3014),
    [anon_sym_rewinddir] = ACTIONS(3014),
    [anon_sym_readlink] = ACTIONS(3014),
    [anon_sym_ref] = ACTIONS(3014),
    [anon_sym_scalar] = ACTIONS(3014),
    [anon_sym_shift] = ACTIONS(3014),
    [anon_sym_sin] = ACTIONS(3014),
    [anon_sym_sleep] = ACTIONS(3014),
    [anon_sym_sqrt] = ACTIONS(3014),
    [anon_sym_srand] = ACTIONS(3014),
    [anon_sym_stat] = ACTIONS(3014),
    [anon_sym_study] = ACTIONS(3014),
    [anon_sym_tell] = ACTIONS(3014),
    [anon_sym_telldir] = ACTIONS(3014),
    [anon_sym_tied] = ACTIONS(3014),
    [anon_sym_uc] = ACTIONS(3014),
    [anon_sym_ucfirst] = ACTIONS(3014),
    [anon_sym_untie] = ACTIONS(3014),
    [anon_sym_umask] = ACTIONS(3014),
    [anon_sym_values] = ACTIONS(3014),
    [anon_sym_write] = ACTIONS(3014),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3014),
    [anon_sym_qq] = ACTIONS(3014),
    [anon_sym_qw] = ACTIONS(3014),
    [anon_sym_qx] = ACTIONS(3014),
    [anon_sym_qr] = ACTIONS(3014),
    [anon_sym_m] = ACTIONS(3014),
    [anon_sym_s] = ACTIONS(3014),
    [anon_sym_tr] = ACTIONS(3014),
    [anon_sym_y] = ACTIONS(3014),
    [aux_sym__bareword_token1] = ACTIONS(3014),
    [sym_number] = ACTIONS(3014),
    [anon_sym_true] = ACTIONS(3014),
    [anon_sym_false] = ACTIONS(3014),
    [sym__single_quote] = ACTIONS(3012),
    [sym__double_quote] = ACTIONS(3012),
    [sym__backtick_quote] = ACTIONS(3012),
    [sym__search_slash_quote] = ACTIONS(3012),
    [sym__open_readline_bracket] = ACTIONS(3012),
    [sym__open_fileglob_bracket] = ACTIONS(3012),
    [sym__PERLY_HEREDOC] = ACTIONS(3012),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3012),
    [sym__filetest] = ACTIONS(3012),
  },
  [STATE(1673)] = {
    [sym_heredoc_content] = STATE(1673),
    [sym__identifier] = ACTIONS(3018),
    [anon_sym_LBRACE] = ACTIONS(3018),
    [anon_sym_DOLLAR] = ACTIONS(3018),
    [anon_sym_AT] = ACTIONS(3018),
    [anon_sym_RPAREN] = ACTIONS(3018),
    [anon_sym_my] = ACTIONS(3018),
    [anon_sym_extended] = ACTIONS(3018),
    [anon_sym_async] = ACTIONS(3018),
    [anon_sym_sub] = ACTIONS(3018),
    [anon_sym_method] = ACTIONS(3018),
    [anon_sym_LPAREN] = ACTIONS(3018),
    [anon_sym_continue] = ACTIONS(3286),
    [anon_sym_state] = ACTIONS(3018),
    [anon_sym_our] = ACTIONS(3018),
    [anon_sym_LBRACK] = ACTIONS(3018),
    [anon_sym_LT_LT] = ACTIONS(3018),
    [anon_sym_SLASH_SLASH] = ACTIONS(3018),
    [anon_sym_PLUS] = ACTIONS(3018),
    [anon_sym_DASH] = ACTIONS(3018),
    [anon_sym_TILDE] = ACTIONS(3018),
    [anon_sym_BANG] = ACTIONS(3018),
    [anon_sym_PLUS_PLUS] = ACTIONS(3018),
    [anon_sym_DASH_DASH] = ACTIONS(3018),
    [anon_sym_BSLASH] = ACTIONS(3016),
    [anon_sym_do] = ACTIONS(3018),
    [anon_sym_eval] = ACTIONS(3018),
    [anon_sym_await] = ACTIONS(3018),
    [anon_sym_field] = ACTIONS(3018),
    [anon_sym_local] = ACTIONS(3018),
    [anon_sym_dynamically] = ACTIONS(3018),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3018),
    [anon_sym_require] = ACTIONS(3018),
    [anon_sym_map] = ACTIONS(3018),
    [anon_sym_grep] = ACTIONS(3018),
    [anon_sym_sort] = ACTIONS(3018),
    [anon_sym_goto] = ACTIONS(3018),
    [anon_sym_return] = ACTIONS(3018),
    [anon_sym_undef] = ACTIONS(3018),
    [anon_sym_PERCENT2] = ACTIONS(3018),
    [anon_sym_AMP2] = ACTIONS(3018),
    [anon_sym_STAR2] = ACTIONS(3018),
    [anon_sym_last] = ACTIONS(3018),
    [anon_sym_next] = ACTIONS(3018),
    [anon_sym_redo] = ACTIONS(3018),
    [anon_sym___FILE__] = ACTIONS(3018),
    [anon_sym___LINE__] = ACTIONS(3018),
    [anon_sym___PACKAGE__] = ACTIONS(3018),
    [anon_sym___SUB__] = ACTIONS(3018),
    [anon_sym_break] = ACTIONS(3018),
    [anon_sym_fork] = ACTIONS(3018),
    [anon_sym_getppid] = ACTIONS(3018),
    [anon_sym_time] = ACTIONS(3018),
    [anon_sym_times] = ACTIONS(3018),
    [anon_sym_wait] = ACTIONS(3018),
    [anon_sym_wantarray] = ACTIONS(3018),
    [anon_sym_abs] = ACTIONS(3018),
    [anon_sym_alarm] = ACTIONS(3018),
    [anon_sym_chop] = ACTIONS(3018),
    [anon_sym_chdir] = ACTIONS(3018),
    [anon_sym_close] = ACTIONS(3018),
    [anon_sym_closedir] = ACTIONS(3018),
    [anon_sym_caller] = ACTIONS(3018),
    [anon_sym_chomp] = ACTIONS(3018),
    [anon_sym_chr] = ACTIONS(3018),
    [anon_sym_cos] = ACTIONS(3018),
    [anon_sym_chroot] = ACTIONS(3018),
    [anon_sym_defined] = ACTIONS(3018),
    [anon_sym_delete] = ACTIONS(3018),
    [anon_sym_dbmclose] = ACTIONS(3018),
    [anon_sym_exists] = ACTIONS(3018),
    [anon_sym_exit] = ACTIONS(3018),
    [anon_sym_eof] = ACTIONS(3018),
    [anon_sym_exp] = ACTIONS(3018),
    [anon_sym_each] = ACTIONS(3018),
    [anon_sym_fc] = ACTIONS(3018),
    [anon_sym_fileno] = ACTIONS(3018),
    [anon_sym_gmtime] = ACTIONS(3018),
    [anon_sym_getc] = ACTIONS(3018),
    [anon_sym_getpgrp] = ACTIONS(3018),
    [anon_sym_getprotobyname] = ACTIONS(3018),
    [anon_sym_getpwname] = ACTIONS(3018),
    [anon_sym_getpwuid] = ACTIONS(3018),
    [anon_sym_getpeername] = ACTIONS(3018),
    [anon_sym_getnetbyname] = ACTIONS(3018),
    [anon_sym_getsockname] = ACTIONS(3018),
    [anon_sym_getgrnam] = ACTIONS(3018),
    [anon_sym_getgrgid] = ACTIONS(3018),
    [anon_sym_hex] = ACTIONS(3018),
    [anon_sym_int] = ACTIONS(3018),
    [anon_sym_keys] = ACTIONS(3018),
    [anon_sym_lc] = ACTIONS(3018),
    [anon_sym_lcfirst] = ACTIONS(3018),
    [anon_sym_length] = ACTIONS(3018),
    [anon_sym_localtime] = ACTIONS(3018),
    [anon_sym_log] = ACTIONS(3018),
    [anon_sym_lock] = ACTIONS(3018),
    [anon_sym_lstat] = ACTIONS(3018),
    [anon_sym_oct] = ACTIONS(3018),
    [anon_sym_ord] = ACTIONS(3018),
    [anon_sym_prototype] = ACTIONS(3018),
    [anon_sym_pop] = ACTIONS(3018),
    [anon_sym_pos] = ACTIONS(3018),
    [anon_sym_quotemeta] = ACTIONS(3018),
    [anon_sym_reset] = ACTIONS(3018),
    [anon_sym_rand] = ACTIONS(3018),
    [anon_sym_rmdir] = ACTIONS(3018),
    [anon_sym_readdir] = ACTIONS(3018),
    [anon_sym_readline] = ACTIONS(3018),
    [anon_sym_readpipe] = ACTIONS(3018),
    [anon_sym_rewinddir] = ACTIONS(3018),
    [anon_sym_readlink] = ACTIONS(3018),
    [anon_sym_ref] = ACTIONS(3018),
    [anon_sym_scalar] = ACTIONS(3018),
    [anon_sym_shift] = ACTIONS(3018),
    [anon_sym_sin] = ACTIONS(3018),
    [anon_sym_sleep] = ACTIONS(3018),
    [anon_sym_sqrt] = ACTIONS(3018),
    [anon_sym_srand] = ACTIONS(3018),
    [anon_sym_stat] = ACTIONS(3018),
    [anon_sym_study] = ACTIONS(3018),
    [anon_sym_tell] = ACTIONS(3018),
    [anon_sym_telldir] = ACTIONS(3018),
    [anon_sym_tied] = ACTIONS(3018),
    [anon_sym_uc] = ACTIONS(3018),
    [anon_sym_ucfirst] = ACTIONS(3018),
    [anon_sym_untie] = ACTIONS(3018),
    [anon_sym_umask] = ACTIONS(3018),
    [anon_sym_values] = ACTIONS(3018),
    [anon_sym_write] = ACTIONS(3018),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3018),
    [anon_sym_qq] = ACTIONS(3018),
    [anon_sym_qw] = ACTIONS(3018),
    [anon_sym_qx] = ACTIONS(3018),
    [anon_sym_qr] = ACTIONS(3018),
    [anon_sym_m] = ACTIONS(3018),
    [anon_sym_s] = ACTIONS(3018),
    [anon_sym_tr] = ACTIONS(3018),
    [anon_sym_y] = ACTIONS(3018),
    [aux_sym__bareword_token1] = ACTIONS(3018),
    [sym_number] = ACTIONS(3018),
    [anon_sym_true] = ACTIONS(3018),
    [anon_sym_false] = ACTIONS(3018),
    [sym__single_quote] = ACTIONS(3016),
    [sym__double_quote] = ACTIONS(3016),
    [sym__backtick_quote] = ACTIONS(3016),
    [sym__search_slash_quote] = ACTIONS(3016),
    [sym__open_readline_bracket] = ACTIONS(3016),
    [sym__open_fileglob_bracket] = ACTIONS(3016),
    [sym__PERLY_HEREDOC] = ACTIONS(3016),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3016),
    [sym__filetest] = ACTIONS(3016),
  },
  [STATE(1674)] = {
    [sym_heredoc_content] = STATE(1674),
    [sym__identifier] = ACTIONS(3033),
    [anon_sym_LBRACE] = ACTIONS(3033),
    [anon_sym_DOLLAR] = ACTIONS(3033),
    [anon_sym_AT] = ACTIONS(3033),
    [anon_sym_RPAREN] = ACTIONS(3033),
    [anon_sym_my] = ACTIONS(3033),
    [anon_sym_extended] = ACTIONS(3033),
    [anon_sym_async] = ACTIONS(3033),
    [anon_sym_sub] = ACTIONS(3033),
    [anon_sym_method] = ACTIONS(3033),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_continue] = ACTIONS(3033),
    [anon_sym_state] = ACTIONS(3033),
    [anon_sym_our] = ACTIONS(3033),
    [anon_sym_LBRACK] = ACTIONS(3033),
    [anon_sym_LT_LT] = ACTIONS(3033),
    [anon_sym_SLASH_SLASH] = ACTIONS(3033),
    [anon_sym_PLUS] = ACTIONS(3033),
    [anon_sym_DASH] = ACTIONS(3033),
    [anon_sym_TILDE] = ACTIONS(3033),
    [anon_sym_BANG] = ACTIONS(3033),
    [anon_sym_PLUS_PLUS] = ACTIONS(3033),
    [anon_sym_DASH_DASH] = ACTIONS(3033),
    [anon_sym_BSLASH] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3033),
    [anon_sym_eval] = ACTIONS(3033),
    [anon_sym_await] = ACTIONS(3033),
    [anon_sym_field] = ACTIONS(3033),
    [anon_sym_local] = ACTIONS(3033),
    [anon_sym_dynamically] = ACTIONS(3033),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3033),
    [anon_sym_require] = ACTIONS(3033),
    [anon_sym_map] = ACTIONS(3033),
    [anon_sym_grep] = ACTIONS(3033),
    [anon_sym_sort] = ACTIONS(3033),
    [anon_sym_goto] = ACTIONS(3033),
    [anon_sym_return] = ACTIONS(3033),
    [anon_sym_undef] = ACTIONS(3033),
    [anon_sym_PERCENT2] = ACTIONS(3033),
    [anon_sym_AMP2] = ACTIONS(3033),
    [anon_sym_STAR2] = ACTIONS(3033),
    [anon_sym_last] = ACTIONS(3033),
    [anon_sym_next] = ACTIONS(3033),
    [anon_sym_redo] = ACTIONS(3033),
    [anon_sym___FILE__] = ACTIONS(3033),
    [anon_sym___LINE__] = ACTIONS(3033),
    [anon_sym___PACKAGE__] = ACTIONS(3033),
    [anon_sym___SUB__] = ACTIONS(3033),
    [anon_sym_break] = ACTIONS(3033),
    [anon_sym_fork] = ACTIONS(3033),
    [anon_sym_getppid] = ACTIONS(3033),
    [anon_sym_time] = ACTIONS(3033),
    [anon_sym_times] = ACTIONS(3033),
    [anon_sym_wait] = ACTIONS(3033),
    [anon_sym_wantarray] = ACTIONS(3033),
    [anon_sym_abs] = ACTIONS(3033),
    [anon_sym_alarm] = ACTIONS(3033),
    [anon_sym_chop] = ACTIONS(3033),
    [anon_sym_chdir] = ACTIONS(3033),
    [anon_sym_close] = ACTIONS(3033),
    [anon_sym_closedir] = ACTIONS(3033),
    [anon_sym_caller] = ACTIONS(3033),
    [anon_sym_chomp] = ACTIONS(3033),
    [anon_sym_chr] = ACTIONS(3033),
    [anon_sym_cos] = ACTIONS(3033),
    [anon_sym_chroot] = ACTIONS(3033),
    [anon_sym_defined] = ACTIONS(3033),
    [anon_sym_delete] = ACTIONS(3033),
    [anon_sym_dbmclose] = ACTIONS(3033),
    [anon_sym_exists] = ACTIONS(3033),
    [anon_sym_exit] = ACTIONS(3033),
    [anon_sym_eof] = ACTIONS(3033),
    [anon_sym_exp] = ACTIONS(3033),
    [anon_sym_each] = ACTIONS(3033),
    [anon_sym_fc] = ACTIONS(3033),
    [anon_sym_fileno] = ACTIONS(3033),
    [anon_sym_gmtime] = ACTIONS(3033),
    [anon_sym_getc] = ACTIONS(3033),
    [anon_sym_getpgrp] = ACTIONS(3033),
    [anon_sym_getprotobyname] = ACTIONS(3033),
    [anon_sym_getpwname] = ACTIONS(3033),
    [anon_sym_getpwuid] = ACTIONS(3033),
    [anon_sym_getpeername] = ACTIONS(3033),
    [anon_sym_getnetbyname] = ACTIONS(3033),
    [anon_sym_getsockname] = ACTIONS(3033),
    [anon_sym_getgrnam] = ACTIONS(3033),
    [anon_sym_getgrgid] = ACTIONS(3033),
    [anon_sym_hex] = ACTIONS(3033),
    [anon_sym_int] = ACTIONS(3033),
    [anon_sym_keys] = ACTIONS(3033),
    [anon_sym_lc] = ACTIONS(3033),
    [anon_sym_lcfirst] = ACTIONS(3033),
    [anon_sym_length] = ACTIONS(3033),
    [anon_sym_localtime] = ACTIONS(3033),
    [anon_sym_log] = ACTIONS(3033),
    [anon_sym_lock] = ACTIONS(3033),
    [anon_sym_lstat] = ACTIONS(3033),
    [anon_sym_oct] = ACTIONS(3033),
    [anon_sym_ord] = ACTIONS(3033),
    [anon_sym_prototype] = ACTIONS(3033),
    [anon_sym_pop] = ACTIONS(3033),
    [anon_sym_pos] = ACTIONS(3033),
    [anon_sym_quotemeta] = ACTIONS(3033),
    [anon_sym_reset] = ACTIONS(3033),
    [anon_sym_rand] = ACTIONS(3033),
    [anon_sym_rmdir] = ACTIONS(3033),
    [anon_sym_readdir] = ACTIONS(3033),
    [anon_sym_readline] = ACTIONS(3033),
    [anon_sym_readpipe] = ACTIONS(3033),
    [anon_sym_rewinddir] = ACTIONS(3033),
    [anon_sym_readlink] = ACTIONS(3033),
    [anon_sym_ref] = ACTIONS(3033),
    [anon_sym_scalar] = ACTIONS(3033),
    [anon_sym_shift] = ACTIONS(3033),
    [anon_sym_sin] = ACTIONS(3033),
    [anon_sym_sleep] = ACTIONS(3033),
    [anon_sym_sqrt] = ACTIONS(3033),
    [anon_sym_srand] = ACTIONS(3033),
    [anon_sym_stat] = ACTIONS(3033),
    [anon_sym_study] = ACTIONS(3033),
    [anon_sym_tell] = ACTIONS(3033),
    [anon_sym_telldir] = ACTIONS(3033),
    [anon_sym_tied] = ACTIONS(3033),
    [anon_sym_uc] = ACTIONS(3033),
    [anon_sym_ucfirst] = ACTIONS(3033),
    [anon_sym_untie] = ACTIONS(3033),
    [anon_sym_umask] = ACTIONS(3033),
    [anon_sym_values] = ACTIONS(3033),
    [anon_sym_write] = ACTIONS(3033),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3033),
    [anon_sym_qq] = ACTIONS(3033),
    [anon_sym_qw] = ACTIONS(3033),
    [anon_sym_qx] = ACTIONS(3033),
    [anon_sym_qr] = ACTIONS(3033),
    [anon_sym_m] = ACTIONS(3033),
    [anon_sym_s] = ACTIONS(3033),
    [anon_sym_tr] = ACTIONS(3033),
    [anon_sym_y] = ACTIONS(3033),
    [aux_sym__bareword_token1] = ACTIONS(3033),
    [sym_number] = ACTIONS(3033),
    [anon_sym_true] = ACTIONS(3033),
    [anon_sym_false] = ACTIONS(3033),
    [sym__single_quote] = ACTIONS(3031),
    [sym__double_quote] = ACTIONS(3031),
    [sym__backtick_quote] = ACTIONS(3031),
    [sym__search_slash_quote] = ACTIONS(3031),
    [sym__open_readline_bracket] = ACTIONS(3031),
    [sym__open_fileglob_bracket] = ACTIONS(3031),
    [sym__PERLY_HEREDOC] = ACTIONS(3031),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3031),
    [sym__filetest] = ACTIONS(3031),
  },
  [STATE(1675)] = {
    [sym_heredoc_content] = STATE(1675),
    [sym__identifier] = ACTIONS(3089),
    [anon_sym_LBRACE] = ACTIONS(3089),
    [anon_sym_DOLLAR] = ACTIONS(3089),
    [anon_sym_AT] = ACTIONS(3089),
    [anon_sym_RPAREN] = ACTIONS(3089),
    [anon_sym_my] = ACTIONS(3089),
    [anon_sym_extended] = ACTIONS(3089),
    [anon_sym_async] = ACTIONS(3089),
    [anon_sym_sub] = ACTIONS(3089),
    [anon_sym_method] = ACTIONS(3089),
    [anon_sym_LPAREN] = ACTIONS(3089),
    [anon_sym_continue] = ACTIONS(3089),
    [anon_sym_state] = ACTIONS(3089),
    [anon_sym_our] = ACTIONS(3089),
    [anon_sym_LBRACK] = ACTIONS(3089),
    [anon_sym_LT_LT] = ACTIONS(3089),
    [anon_sym_SLASH_SLASH] = ACTIONS(3089),
    [anon_sym_PLUS] = ACTIONS(3089),
    [anon_sym_DASH] = ACTIONS(3089),
    [anon_sym_TILDE] = ACTIONS(3089),
    [anon_sym_BANG] = ACTIONS(3089),
    [anon_sym_PLUS_PLUS] = ACTIONS(3089),
    [anon_sym_DASH_DASH] = ACTIONS(3089),
    [anon_sym_BSLASH] = ACTIONS(3087),
    [anon_sym_do] = ACTIONS(3089),
    [anon_sym_eval] = ACTIONS(3089),
    [anon_sym_await] = ACTIONS(3089),
    [anon_sym_field] = ACTIONS(3089),
    [anon_sym_local] = ACTIONS(3089),
    [anon_sym_dynamically] = ACTIONS(3089),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3089),
    [anon_sym_require] = ACTIONS(3089),
    [anon_sym_map] = ACTIONS(3089),
    [anon_sym_grep] = ACTIONS(3089),
    [anon_sym_sort] = ACTIONS(3089),
    [anon_sym_goto] = ACTIONS(3089),
    [anon_sym_return] = ACTIONS(3089),
    [anon_sym_undef] = ACTIONS(3089),
    [anon_sym_PERCENT2] = ACTIONS(3089),
    [anon_sym_AMP2] = ACTIONS(3089),
    [anon_sym_STAR2] = ACTIONS(3089),
    [anon_sym_last] = ACTIONS(3089),
    [anon_sym_next] = ACTIONS(3089),
    [anon_sym_redo] = ACTIONS(3089),
    [anon_sym___FILE__] = ACTIONS(3089),
    [anon_sym___LINE__] = ACTIONS(3089),
    [anon_sym___PACKAGE__] = ACTIONS(3089),
    [anon_sym___SUB__] = ACTIONS(3089),
    [anon_sym_break] = ACTIONS(3089),
    [anon_sym_fork] = ACTIONS(3089),
    [anon_sym_getppid] = ACTIONS(3089),
    [anon_sym_time] = ACTIONS(3089),
    [anon_sym_times] = ACTIONS(3089),
    [anon_sym_wait] = ACTIONS(3089),
    [anon_sym_wantarray] = ACTIONS(3089),
    [anon_sym_abs] = ACTIONS(3089),
    [anon_sym_alarm] = ACTIONS(3089),
    [anon_sym_chop] = ACTIONS(3089),
    [anon_sym_chdir] = ACTIONS(3089),
    [anon_sym_close] = ACTIONS(3089),
    [anon_sym_closedir] = ACTIONS(3089),
    [anon_sym_caller] = ACTIONS(3089),
    [anon_sym_chomp] = ACTIONS(3089),
    [anon_sym_chr] = ACTIONS(3089),
    [anon_sym_cos] = ACTIONS(3089),
    [anon_sym_chroot] = ACTIONS(3089),
    [anon_sym_defined] = ACTIONS(3089),
    [anon_sym_delete] = ACTIONS(3089),
    [anon_sym_dbmclose] = ACTIONS(3089),
    [anon_sym_exists] = ACTIONS(3089),
    [anon_sym_exit] = ACTIONS(3089),
    [anon_sym_eof] = ACTIONS(3089),
    [anon_sym_exp] = ACTIONS(3089),
    [anon_sym_each] = ACTIONS(3089),
    [anon_sym_fc] = ACTIONS(3089),
    [anon_sym_fileno] = ACTIONS(3089),
    [anon_sym_gmtime] = ACTIONS(3089),
    [anon_sym_getc] = ACTIONS(3089),
    [anon_sym_getpgrp] = ACTIONS(3089),
    [anon_sym_getprotobyname] = ACTIONS(3089),
    [anon_sym_getpwname] = ACTIONS(3089),
    [anon_sym_getpwuid] = ACTIONS(3089),
    [anon_sym_getpeername] = ACTIONS(3089),
    [anon_sym_getnetbyname] = ACTIONS(3089),
    [anon_sym_getsockname] = ACTIONS(3089),
    [anon_sym_getgrnam] = ACTIONS(3089),
    [anon_sym_getgrgid] = ACTIONS(3089),
    [anon_sym_hex] = ACTIONS(3089),
    [anon_sym_int] = ACTIONS(3089),
    [anon_sym_keys] = ACTIONS(3089),
    [anon_sym_lc] = ACTIONS(3089),
    [anon_sym_lcfirst] = ACTIONS(3089),
    [anon_sym_length] = ACTIONS(3089),
    [anon_sym_localtime] = ACTIONS(3089),
    [anon_sym_log] = ACTIONS(3089),
    [anon_sym_lock] = ACTIONS(3089),
    [anon_sym_lstat] = ACTIONS(3089),
    [anon_sym_oct] = ACTIONS(3089),
    [anon_sym_ord] = ACTIONS(3089),
    [anon_sym_prototype] = ACTIONS(3089),
    [anon_sym_pop] = ACTIONS(3089),
    [anon_sym_pos] = ACTIONS(3089),
    [anon_sym_quotemeta] = ACTIONS(3089),
    [anon_sym_reset] = ACTIONS(3089),
    [anon_sym_rand] = ACTIONS(3089),
    [anon_sym_rmdir] = ACTIONS(3089),
    [anon_sym_readdir] = ACTIONS(3089),
    [anon_sym_readline] = ACTIONS(3089),
    [anon_sym_readpipe] = ACTIONS(3089),
    [anon_sym_rewinddir] = ACTIONS(3089),
    [anon_sym_readlink] = ACTIONS(3089),
    [anon_sym_ref] = ACTIONS(3089),
    [anon_sym_scalar] = ACTIONS(3089),
    [anon_sym_shift] = ACTIONS(3089),
    [anon_sym_sin] = ACTIONS(3089),
    [anon_sym_sleep] = ACTIONS(3089),
    [anon_sym_sqrt] = ACTIONS(3089),
    [anon_sym_srand] = ACTIONS(3089),
    [anon_sym_stat] = ACTIONS(3089),
    [anon_sym_study] = ACTIONS(3089),
    [anon_sym_tell] = ACTIONS(3089),
    [anon_sym_telldir] = ACTIONS(3089),
    [anon_sym_tied] = ACTIONS(3089),
    [anon_sym_uc] = ACTIONS(3089),
    [anon_sym_ucfirst] = ACTIONS(3089),
    [anon_sym_untie] = ACTIONS(3089),
    [anon_sym_umask] = ACTIONS(3089),
    [anon_sym_values] = ACTIONS(3089),
    [anon_sym_write] = ACTIONS(3089),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3089),
    [anon_sym_qq] = ACTIONS(3089),
    [anon_sym_qw] = ACTIONS(3089),
    [anon_sym_qx] = ACTIONS(3089),
    [anon_sym_qr] = ACTIONS(3089),
    [anon_sym_m] = ACTIONS(3089),
    [anon_sym_s] = ACTIONS(3089),
    [anon_sym_tr] = ACTIONS(3089),
    [anon_sym_y] = ACTIONS(3089),
    [aux_sym__bareword_token1] = ACTIONS(3089),
    [sym_number] = ACTIONS(3089),
    [anon_sym_true] = ACTIONS(3089),
    [anon_sym_false] = ACTIONS(3089),
    [sym__single_quote] = ACTIONS(3087),
    [sym__double_quote] = ACTIONS(3087),
    [sym__backtick_quote] = ACTIONS(3087),
    [sym__search_slash_quote] = ACTIONS(3087),
    [sym__open_readline_bracket] = ACTIONS(3087),
    [sym__open_fileglob_bracket] = ACTIONS(3087),
    [sym__PERLY_HEREDOC] = ACTIONS(3087),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3087),
    [sym__filetest] = ACTIONS(3087),
  },
  [STATE(1676)] = {
    [sym_heredoc_content] = STATE(1676),
    [sym__identifier] = ACTIONS(3093),
    [anon_sym_LBRACE] = ACTIONS(3093),
    [anon_sym_DOLLAR] = ACTIONS(3093),
    [anon_sym_AT] = ACTIONS(3093),
    [anon_sym_RPAREN] = ACTIONS(3093),
    [anon_sym_my] = ACTIONS(3093),
    [anon_sym_extended] = ACTIONS(3093),
    [anon_sym_async] = ACTIONS(3093),
    [anon_sym_sub] = ACTIONS(3093),
    [anon_sym_method] = ACTIONS(3093),
    [anon_sym_LPAREN] = ACTIONS(3093),
    [anon_sym_continue] = ACTIONS(3093),
    [anon_sym_state] = ACTIONS(3093),
    [anon_sym_our] = ACTIONS(3093),
    [anon_sym_LBRACK] = ACTIONS(3093),
    [anon_sym_LT_LT] = ACTIONS(3093),
    [anon_sym_SLASH_SLASH] = ACTIONS(3093),
    [anon_sym_PLUS] = ACTIONS(3093),
    [anon_sym_DASH] = ACTIONS(3093),
    [anon_sym_TILDE] = ACTIONS(3093),
    [anon_sym_BANG] = ACTIONS(3093),
    [anon_sym_PLUS_PLUS] = ACTIONS(3093),
    [anon_sym_DASH_DASH] = ACTIONS(3093),
    [anon_sym_BSLASH] = ACTIONS(3091),
    [anon_sym_do] = ACTIONS(3093),
    [anon_sym_eval] = ACTIONS(3093),
    [anon_sym_await] = ACTIONS(3093),
    [anon_sym_field] = ACTIONS(3093),
    [anon_sym_local] = ACTIONS(3093),
    [anon_sym_dynamically] = ACTIONS(3093),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3093),
    [anon_sym_require] = ACTIONS(3093),
    [anon_sym_map] = ACTIONS(3093),
    [anon_sym_grep] = ACTIONS(3093),
    [anon_sym_sort] = ACTIONS(3093),
    [anon_sym_goto] = ACTIONS(3093),
    [anon_sym_return] = ACTIONS(3093),
    [anon_sym_undef] = ACTIONS(3093),
    [anon_sym_PERCENT2] = ACTIONS(3093),
    [anon_sym_AMP2] = ACTIONS(3093),
    [anon_sym_STAR2] = ACTIONS(3093),
    [anon_sym_last] = ACTIONS(3093),
    [anon_sym_next] = ACTIONS(3093),
    [anon_sym_redo] = ACTIONS(3093),
    [anon_sym___FILE__] = ACTIONS(3093),
    [anon_sym___LINE__] = ACTIONS(3093),
    [anon_sym___PACKAGE__] = ACTIONS(3093),
    [anon_sym___SUB__] = ACTIONS(3093),
    [anon_sym_break] = ACTIONS(3093),
    [anon_sym_fork] = ACTIONS(3093),
    [anon_sym_getppid] = ACTIONS(3093),
    [anon_sym_time] = ACTIONS(3093),
    [anon_sym_times] = ACTIONS(3093),
    [anon_sym_wait] = ACTIONS(3093),
    [anon_sym_wantarray] = ACTIONS(3093),
    [anon_sym_abs] = ACTIONS(3093),
    [anon_sym_alarm] = ACTIONS(3093),
    [anon_sym_chop] = ACTIONS(3093),
    [anon_sym_chdir] = ACTIONS(3093),
    [anon_sym_close] = ACTIONS(3093),
    [anon_sym_closedir] = ACTIONS(3093),
    [anon_sym_caller] = ACTIONS(3093),
    [anon_sym_chomp] = ACTIONS(3093),
    [anon_sym_chr] = ACTIONS(3093),
    [anon_sym_cos] = ACTIONS(3093),
    [anon_sym_chroot] = ACTIONS(3093),
    [anon_sym_defined] = ACTIONS(3093),
    [anon_sym_delete] = ACTIONS(3093),
    [anon_sym_dbmclose] = ACTIONS(3093),
    [anon_sym_exists] = ACTIONS(3093),
    [anon_sym_exit] = ACTIONS(3093),
    [anon_sym_eof] = ACTIONS(3093),
    [anon_sym_exp] = ACTIONS(3093),
    [anon_sym_each] = ACTIONS(3093),
    [anon_sym_fc] = ACTIONS(3093),
    [anon_sym_fileno] = ACTIONS(3093),
    [anon_sym_gmtime] = ACTIONS(3093),
    [anon_sym_getc] = ACTIONS(3093),
    [anon_sym_getpgrp] = ACTIONS(3093),
    [anon_sym_getprotobyname] = ACTIONS(3093),
    [anon_sym_getpwname] = ACTIONS(3093),
    [anon_sym_getpwuid] = ACTIONS(3093),
    [anon_sym_getpeername] = ACTIONS(3093),
    [anon_sym_getnetbyname] = ACTIONS(3093),
    [anon_sym_getsockname] = ACTIONS(3093),
    [anon_sym_getgrnam] = ACTIONS(3093),
    [anon_sym_getgrgid] = ACTIONS(3093),
    [anon_sym_hex] = ACTIONS(3093),
    [anon_sym_int] = ACTIONS(3093),
    [anon_sym_keys] = ACTIONS(3093),
    [anon_sym_lc] = ACTIONS(3093),
    [anon_sym_lcfirst] = ACTIONS(3093),
    [anon_sym_length] = ACTIONS(3093),
    [anon_sym_localtime] = ACTIONS(3093),
    [anon_sym_log] = ACTIONS(3093),
    [anon_sym_lock] = ACTIONS(3093),
    [anon_sym_lstat] = ACTIONS(3093),
    [anon_sym_oct] = ACTIONS(3093),
    [anon_sym_ord] = ACTIONS(3093),
    [anon_sym_prototype] = ACTIONS(3093),
    [anon_sym_pop] = ACTIONS(3093),
    [anon_sym_pos] = ACTIONS(3093),
    [anon_sym_quotemeta] = ACTIONS(3093),
    [anon_sym_reset] = ACTIONS(3093),
    [anon_sym_rand] = ACTIONS(3093),
    [anon_sym_rmdir] = ACTIONS(3093),
    [anon_sym_readdir] = ACTIONS(3093),
    [anon_sym_readline] = ACTIONS(3093),
    [anon_sym_readpipe] = ACTIONS(3093),
    [anon_sym_rewinddir] = ACTIONS(3093),
    [anon_sym_readlink] = ACTIONS(3093),
    [anon_sym_ref] = ACTIONS(3093),
    [anon_sym_scalar] = ACTIONS(3093),
    [anon_sym_shift] = ACTIONS(3093),
    [anon_sym_sin] = ACTIONS(3093),
    [anon_sym_sleep] = ACTIONS(3093),
    [anon_sym_sqrt] = ACTIONS(3093),
    [anon_sym_srand] = ACTIONS(3093),
    [anon_sym_stat] = ACTIONS(3093),
    [anon_sym_study] = ACTIONS(3093),
    [anon_sym_tell] = ACTIONS(3093),
    [anon_sym_telldir] = ACTIONS(3093),
    [anon_sym_tied] = ACTIONS(3093),
    [anon_sym_uc] = ACTIONS(3093),
    [anon_sym_ucfirst] = ACTIONS(3093),
    [anon_sym_untie] = ACTIONS(3093),
    [anon_sym_umask] = ACTIONS(3093),
    [anon_sym_values] = ACTIONS(3093),
    [anon_sym_write] = ACTIONS(3093),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3093),
    [anon_sym_qq] = ACTIONS(3093),
    [anon_sym_qw] = ACTIONS(3093),
    [anon_sym_qx] = ACTIONS(3093),
    [anon_sym_qr] = ACTIONS(3093),
    [anon_sym_m] = ACTIONS(3093),
    [anon_sym_s] = ACTIONS(3093),
    [anon_sym_tr] = ACTIONS(3093),
    [anon_sym_y] = ACTIONS(3093),
    [aux_sym__bareword_token1] = ACTIONS(3093),
    [sym_number] = ACTIONS(3093),
    [anon_sym_true] = ACTIONS(3093),
    [anon_sym_false] = ACTIONS(3093),
    [sym__single_quote] = ACTIONS(3091),
    [sym__double_quote] = ACTIONS(3091),
    [sym__backtick_quote] = ACTIONS(3091),
    [sym__search_slash_quote] = ACTIONS(3091),
    [sym__open_readline_bracket] = ACTIONS(3091),
    [sym__open_fileglob_bracket] = ACTIONS(3091),
    [sym__PERLY_HEREDOC] = ACTIONS(3091),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3091),
    [sym__filetest] = ACTIONS(3091),
  },
  [STATE(1677)] = {
    [sym_heredoc_content] = STATE(1677),
    [sym__identifier] = ACTIONS(3101),
    [anon_sym_LBRACE] = ACTIONS(3101),
    [anon_sym_DOLLAR] = ACTIONS(3101),
    [anon_sym_AT] = ACTIONS(3101),
    [anon_sym_RPAREN] = ACTIONS(3101),
    [anon_sym_my] = ACTIONS(3101),
    [anon_sym_extended] = ACTIONS(3101),
    [anon_sym_async] = ACTIONS(3101),
    [anon_sym_sub] = ACTIONS(3101),
    [anon_sym_method] = ACTIONS(3101),
    [anon_sym_LPAREN] = ACTIONS(3101),
    [anon_sym_continue] = ACTIONS(3101),
    [anon_sym_state] = ACTIONS(3101),
    [anon_sym_our] = ACTIONS(3101),
    [anon_sym_LBRACK] = ACTIONS(3101),
    [anon_sym_LT_LT] = ACTIONS(3101),
    [anon_sym_SLASH_SLASH] = ACTIONS(3101),
    [anon_sym_PLUS] = ACTIONS(3101),
    [anon_sym_DASH] = ACTIONS(3101),
    [anon_sym_TILDE] = ACTIONS(3101),
    [anon_sym_BANG] = ACTIONS(3101),
    [anon_sym_PLUS_PLUS] = ACTIONS(3101),
    [anon_sym_DASH_DASH] = ACTIONS(3101),
    [anon_sym_BSLASH] = ACTIONS(3099),
    [anon_sym_do] = ACTIONS(3101),
    [anon_sym_eval] = ACTIONS(3101),
    [anon_sym_await] = ACTIONS(3101),
    [anon_sym_field] = ACTIONS(3101),
    [anon_sym_local] = ACTIONS(3101),
    [anon_sym_dynamically] = ACTIONS(3101),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3101),
    [anon_sym_require] = ACTIONS(3101),
    [anon_sym_map] = ACTIONS(3101),
    [anon_sym_grep] = ACTIONS(3101),
    [anon_sym_sort] = ACTIONS(3101),
    [anon_sym_goto] = ACTIONS(3101),
    [anon_sym_return] = ACTIONS(3101),
    [anon_sym_undef] = ACTIONS(3101),
    [anon_sym_PERCENT2] = ACTIONS(3101),
    [anon_sym_AMP2] = ACTIONS(3101),
    [anon_sym_STAR2] = ACTIONS(3101),
    [anon_sym_last] = ACTIONS(3101),
    [anon_sym_next] = ACTIONS(3101),
    [anon_sym_redo] = ACTIONS(3101),
    [anon_sym___FILE__] = ACTIONS(3101),
    [anon_sym___LINE__] = ACTIONS(3101),
    [anon_sym___PACKAGE__] = ACTIONS(3101),
    [anon_sym___SUB__] = ACTIONS(3101),
    [anon_sym_break] = ACTIONS(3101),
    [anon_sym_fork] = ACTIONS(3101),
    [anon_sym_getppid] = ACTIONS(3101),
    [anon_sym_time] = ACTIONS(3101),
    [anon_sym_times] = ACTIONS(3101),
    [anon_sym_wait] = ACTIONS(3101),
    [anon_sym_wantarray] = ACTIONS(3101),
    [anon_sym_abs] = ACTIONS(3101),
    [anon_sym_alarm] = ACTIONS(3101),
    [anon_sym_chop] = ACTIONS(3101),
    [anon_sym_chdir] = ACTIONS(3101),
    [anon_sym_close] = ACTIONS(3101),
    [anon_sym_closedir] = ACTIONS(3101),
    [anon_sym_caller] = ACTIONS(3101),
    [anon_sym_chomp] = ACTIONS(3101),
    [anon_sym_chr] = ACTIONS(3101),
    [anon_sym_cos] = ACTIONS(3101),
    [anon_sym_chroot] = ACTIONS(3101),
    [anon_sym_defined] = ACTIONS(3101),
    [anon_sym_delete] = ACTIONS(3101),
    [anon_sym_dbmclose] = ACTIONS(3101),
    [anon_sym_exists] = ACTIONS(3101),
    [anon_sym_exit] = ACTIONS(3101),
    [anon_sym_eof] = ACTIONS(3101),
    [anon_sym_exp] = ACTIONS(3101),
    [anon_sym_each] = ACTIONS(3101),
    [anon_sym_fc] = ACTIONS(3101),
    [anon_sym_fileno] = ACTIONS(3101),
    [anon_sym_gmtime] = ACTIONS(3101),
    [anon_sym_getc] = ACTIONS(3101),
    [anon_sym_getpgrp] = ACTIONS(3101),
    [anon_sym_getprotobyname] = ACTIONS(3101),
    [anon_sym_getpwname] = ACTIONS(3101),
    [anon_sym_getpwuid] = ACTIONS(3101),
    [anon_sym_getpeername] = ACTIONS(3101),
    [anon_sym_getnetbyname] = ACTIONS(3101),
    [anon_sym_getsockname] = ACTIONS(3101),
    [anon_sym_getgrnam] = ACTIONS(3101),
    [anon_sym_getgrgid] = ACTIONS(3101),
    [anon_sym_hex] = ACTIONS(3101),
    [anon_sym_int] = ACTIONS(3101),
    [anon_sym_keys] = ACTIONS(3101),
    [anon_sym_lc] = ACTIONS(3101),
    [anon_sym_lcfirst] = ACTIONS(3101),
    [anon_sym_length] = ACTIONS(3101),
    [anon_sym_localtime] = ACTIONS(3101),
    [anon_sym_log] = ACTIONS(3101),
    [anon_sym_lock] = ACTIONS(3101),
    [anon_sym_lstat] = ACTIONS(3101),
    [anon_sym_oct] = ACTIONS(3101),
    [anon_sym_ord] = ACTIONS(3101),
    [anon_sym_prototype] = ACTIONS(3101),
    [anon_sym_pop] = ACTIONS(3101),
    [anon_sym_pos] = ACTIONS(3101),
    [anon_sym_quotemeta] = ACTIONS(3101),
    [anon_sym_reset] = ACTIONS(3101),
    [anon_sym_rand] = ACTIONS(3101),
    [anon_sym_rmdir] = ACTIONS(3101),
    [anon_sym_readdir] = ACTIONS(3101),
    [anon_sym_readline] = ACTIONS(3101),
    [anon_sym_readpipe] = ACTIONS(3101),
    [anon_sym_rewinddir] = ACTIONS(3101),
    [anon_sym_readlink] = ACTIONS(3101),
    [anon_sym_ref] = ACTIONS(3101),
    [anon_sym_scalar] = ACTIONS(3101),
    [anon_sym_shift] = ACTIONS(3101),
    [anon_sym_sin] = ACTIONS(3101),
    [anon_sym_sleep] = ACTIONS(3101),
    [anon_sym_sqrt] = ACTIONS(3101),
    [anon_sym_srand] = ACTIONS(3101),
    [anon_sym_stat] = ACTIONS(3101),
    [anon_sym_study] = ACTIONS(3101),
    [anon_sym_tell] = ACTIONS(3101),
    [anon_sym_telldir] = ACTIONS(3101),
    [anon_sym_tied] = ACTIONS(3101),
    [anon_sym_uc] = ACTIONS(3101),
    [anon_sym_ucfirst] = ACTIONS(3101),
    [anon_sym_untie] = ACTIONS(3101),
    [anon_sym_umask] = ACTIONS(3101),
    [anon_sym_values] = ACTIONS(3101),
    [anon_sym_write] = ACTIONS(3101),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3101),
    [anon_sym_qq] = ACTIONS(3101),
    [anon_sym_qw] = ACTIONS(3101),
    [anon_sym_qx] = ACTIONS(3101),
    [anon_sym_qr] = ACTIONS(3101),
    [anon_sym_m] = ACTIONS(3101),
    [anon_sym_s] = ACTIONS(3101),
    [anon_sym_tr] = ACTIONS(3101),
    [anon_sym_y] = ACTIONS(3101),
    [aux_sym__bareword_token1] = ACTIONS(3101),
    [sym_number] = ACTIONS(3101),
    [anon_sym_true] = ACTIONS(3101),
    [anon_sym_false] = ACTIONS(3101),
    [sym__single_quote] = ACTIONS(3099),
    [sym__double_quote] = ACTIONS(3099),
    [sym__backtick_quote] = ACTIONS(3099),
    [sym__search_slash_quote] = ACTIONS(3099),
    [sym__open_readline_bracket] = ACTIONS(3099),
    [sym__open_fileglob_bracket] = ACTIONS(3099),
    [sym__PERLY_HEREDOC] = ACTIONS(3099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3099),
    [sym__filetest] = ACTIONS(3099),
  },
  [STATE(1678)] = {
    [sym_heredoc_content] = STATE(1678),
    [sym__identifier] = ACTIONS(3105),
    [anon_sym_LBRACE] = ACTIONS(3105),
    [anon_sym_DOLLAR] = ACTIONS(3105),
    [anon_sym_AT] = ACTIONS(3105),
    [anon_sym_RPAREN] = ACTIONS(3105),
    [anon_sym_my] = ACTIONS(3105),
    [anon_sym_extended] = ACTIONS(3105),
    [anon_sym_async] = ACTIONS(3105),
    [anon_sym_sub] = ACTIONS(3105),
    [anon_sym_method] = ACTIONS(3105),
    [anon_sym_LPAREN] = ACTIONS(3105),
    [anon_sym_continue] = ACTIONS(3105),
    [anon_sym_state] = ACTIONS(3105),
    [anon_sym_our] = ACTIONS(3105),
    [anon_sym_LBRACK] = ACTIONS(3105),
    [anon_sym_LT_LT] = ACTIONS(3105),
    [anon_sym_SLASH_SLASH] = ACTIONS(3105),
    [anon_sym_PLUS] = ACTIONS(3105),
    [anon_sym_DASH] = ACTIONS(3105),
    [anon_sym_TILDE] = ACTIONS(3105),
    [anon_sym_BANG] = ACTIONS(3105),
    [anon_sym_PLUS_PLUS] = ACTIONS(3105),
    [anon_sym_DASH_DASH] = ACTIONS(3105),
    [anon_sym_BSLASH] = ACTIONS(3103),
    [anon_sym_do] = ACTIONS(3105),
    [anon_sym_eval] = ACTIONS(3105),
    [anon_sym_await] = ACTIONS(3105),
    [anon_sym_field] = ACTIONS(3105),
    [anon_sym_local] = ACTIONS(3105),
    [anon_sym_dynamically] = ACTIONS(3105),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3105),
    [anon_sym_require] = ACTIONS(3105),
    [anon_sym_map] = ACTIONS(3105),
    [anon_sym_grep] = ACTIONS(3105),
    [anon_sym_sort] = ACTIONS(3105),
    [anon_sym_goto] = ACTIONS(3105),
    [anon_sym_return] = ACTIONS(3105),
    [anon_sym_undef] = ACTIONS(3105),
    [anon_sym_PERCENT2] = ACTIONS(3105),
    [anon_sym_AMP2] = ACTIONS(3105),
    [anon_sym_STAR2] = ACTIONS(3105),
    [anon_sym_last] = ACTIONS(3105),
    [anon_sym_next] = ACTIONS(3105),
    [anon_sym_redo] = ACTIONS(3105),
    [anon_sym___FILE__] = ACTIONS(3105),
    [anon_sym___LINE__] = ACTIONS(3105),
    [anon_sym___PACKAGE__] = ACTIONS(3105),
    [anon_sym___SUB__] = ACTIONS(3105),
    [anon_sym_break] = ACTIONS(3105),
    [anon_sym_fork] = ACTIONS(3105),
    [anon_sym_getppid] = ACTIONS(3105),
    [anon_sym_time] = ACTIONS(3105),
    [anon_sym_times] = ACTIONS(3105),
    [anon_sym_wait] = ACTIONS(3105),
    [anon_sym_wantarray] = ACTIONS(3105),
    [anon_sym_abs] = ACTIONS(3105),
    [anon_sym_alarm] = ACTIONS(3105),
    [anon_sym_chop] = ACTIONS(3105),
    [anon_sym_chdir] = ACTIONS(3105),
    [anon_sym_close] = ACTIONS(3105),
    [anon_sym_closedir] = ACTIONS(3105),
    [anon_sym_caller] = ACTIONS(3105),
    [anon_sym_chomp] = ACTIONS(3105),
    [anon_sym_chr] = ACTIONS(3105),
    [anon_sym_cos] = ACTIONS(3105),
    [anon_sym_chroot] = ACTIONS(3105),
    [anon_sym_defined] = ACTIONS(3105),
    [anon_sym_delete] = ACTIONS(3105),
    [anon_sym_dbmclose] = ACTIONS(3105),
    [anon_sym_exists] = ACTIONS(3105),
    [anon_sym_exit] = ACTIONS(3105),
    [anon_sym_eof] = ACTIONS(3105),
    [anon_sym_exp] = ACTIONS(3105),
    [anon_sym_each] = ACTIONS(3105),
    [anon_sym_fc] = ACTIONS(3105),
    [anon_sym_fileno] = ACTIONS(3105),
    [anon_sym_gmtime] = ACTIONS(3105),
    [anon_sym_getc] = ACTIONS(3105),
    [anon_sym_getpgrp] = ACTIONS(3105),
    [anon_sym_getprotobyname] = ACTIONS(3105),
    [anon_sym_getpwname] = ACTIONS(3105),
    [anon_sym_getpwuid] = ACTIONS(3105),
    [anon_sym_getpeername] = ACTIONS(3105),
    [anon_sym_getnetbyname] = ACTIONS(3105),
    [anon_sym_getsockname] = ACTIONS(3105),
    [anon_sym_getgrnam] = ACTIONS(3105),
    [anon_sym_getgrgid] = ACTIONS(3105),
    [anon_sym_hex] = ACTIONS(3105),
    [anon_sym_int] = ACTIONS(3105),
    [anon_sym_keys] = ACTIONS(3105),
    [anon_sym_lc] = ACTIONS(3105),
    [anon_sym_lcfirst] = ACTIONS(3105),
    [anon_sym_length] = ACTIONS(3105),
    [anon_sym_localtime] = ACTIONS(3105),
    [anon_sym_log] = ACTIONS(3105),
    [anon_sym_lock] = ACTIONS(3105),
    [anon_sym_lstat] = ACTIONS(3105),
    [anon_sym_oct] = ACTIONS(3105),
    [anon_sym_ord] = ACTIONS(3105),
    [anon_sym_prototype] = ACTIONS(3105),
    [anon_sym_pop] = ACTIONS(3105),
    [anon_sym_pos] = ACTIONS(3105),
    [anon_sym_quotemeta] = ACTIONS(3105),
    [anon_sym_reset] = ACTIONS(3105),
    [anon_sym_rand] = ACTIONS(3105),
    [anon_sym_rmdir] = ACTIONS(3105),
    [anon_sym_readdir] = ACTIONS(3105),
    [anon_sym_readline] = ACTIONS(3105),
    [anon_sym_readpipe] = ACTIONS(3105),
    [anon_sym_rewinddir] = ACTIONS(3105),
    [anon_sym_readlink] = ACTIONS(3105),
    [anon_sym_ref] = ACTIONS(3105),
    [anon_sym_scalar] = ACTIONS(3105),
    [anon_sym_shift] = ACTIONS(3105),
    [anon_sym_sin] = ACTIONS(3105),
    [anon_sym_sleep] = ACTIONS(3105),
    [anon_sym_sqrt] = ACTIONS(3105),
    [anon_sym_srand] = ACTIONS(3105),
    [anon_sym_stat] = ACTIONS(3105),
    [anon_sym_study] = ACTIONS(3105),
    [anon_sym_tell] = ACTIONS(3105),
    [anon_sym_telldir] = ACTIONS(3105),
    [anon_sym_tied] = ACTIONS(3105),
    [anon_sym_uc] = ACTIONS(3105),
    [anon_sym_ucfirst] = ACTIONS(3105),
    [anon_sym_untie] = ACTIONS(3105),
    [anon_sym_umask] = ACTIONS(3105),
    [anon_sym_values] = ACTIONS(3105),
    [anon_sym_write] = ACTIONS(3105),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3105),
    [anon_sym_qq] = ACTIONS(3105),
    [anon_sym_qw] = ACTIONS(3105),
    [anon_sym_qx] = ACTIONS(3105),
    [anon_sym_qr] = ACTIONS(3105),
    [anon_sym_m] = ACTIONS(3105),
    [anon_sym_s] = ACTIONS(3105),
    [anon_sym_tr] = ACTIONS(3105),
    [anon_sym_y] = ACTIONS(3105),
    [aux_sym__bareword_token1] = ACTIONS(3105),
    [sym_number] = ACTIONS(3105),
    [anon_sym_true] = ACTIONS(3105),
    [anon_sym_false] = ACTIONS(3105),
    [sym__single_quote] = ACTIONS(3103),
    [sym__double_quote] = ACTIONS(3103),
    [sym__backtick_quote] = ACTIONS(3103),
    [sym__search_slash_quote] = ACTIONS(3103),
    [sym__open_readline_bracket] = ACTIONS(3103),
    [sym__open_fileglob_bracket] = ACTIONS(3103),
    [sym__PERLY_HEREDOC] = ACTIONS(3103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3103),
    [sym__filetest] = ACTIONS(3103),
  },
  [STATE(1679)] = {
    [sym_heredoc_content] = STATE(1679),
    [sym__identifier] = ACTIONS(3141),
    [anon_sym_LBRACE] = ACTIONS(3141),
    [anon_sym_DOLLAR] = ACTIONS(3141),
    [anon_sym_AT] = ACTIONS(3141),
    [anon_sym_RPAREN] = ACTIONS(3141),
    [anon_sym_my] = ACTIONS(3141),
    [anon_sym_extended] = ACTIONS(3141),
    [anon_sym_async] = ACTIONS(3141),
    [anon_sym_sub] = ACTIONS(3141),
    [anon_sym_method] = ACTIONS(3141),
    [anon_sym_LPAREN] = ACTIONS(3141),
    [anon_sym_continue] = ACTIONS(3141),
    [anon_sym_state] = ACTIONS(3141),
    [anon_sym_our] = ACTIONS(3141),
    [anon_sym_LBRACK] = ACTIONS(3141),
    [anon_sym_LT_LT] = ACTIONS(3141),
    [anon_sym_SLASH_SLASH] = ACTIONS(3141),
    [anon_sym_PLUS] = ACTIONS(3141),
    [anon_sym_DASH] = ACTIONS(3141),
    [anon_sym_TILDE] = ACTIONS(3141),
    [anon_sym_BANG] = ACTIONS(3141),
    [anon_sym_PLUS_PLUS] = ACTIONS(3141),
    [anon_sym_DASH_DASH] = ACTIONS(3141),
    [anon_sym_BSLASH] = ACTIONS(3139),
    [anon_sym_do] = ACTIONS(3141),
    [anon_sym_eval] = ACTIONS(3141),
    [anon_sym_await] = ACTIONS(3141),
    [anon_sym_field] = ACTIONS(3141),
    [anon_sym_local] = ACTIONS(3141),
    [anon_sym_dynamically] = ACTIONS(3141),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3141),
    [anon_sym_require] = ACTIONS(3141),
    [anon_sym_map] = ACTIONS(3141),
    [anon_sym_grep] = ACTIONS(3141),
    [anon_sym_sort] = ACTIONS(3141),
    [anon_sym_goto] = ACTIONS(3141),
    [anon_sym_return] = ACTIONS(3141),
    [anon_sym_undef] = ACTIONS(3141),
    [anon_sym_PERCENT2] = ACTIONS(3141),
    [anon_sym_AMP2] = ACTIONS(3141),
    [anon_sym_STAR2] = ACTIONS(3141),
    [anon_sym_last] = ACTIONS(3141),
    [anon_sym_next] = ACTIONS(3141),
    [anon_sym_redo] = ACTIONS(3141),
    [anon_sym___FILE__] = ACTIONS(3141),
    [anon_sym___LINE__] = ACTIONS(3141),
    [anon_sym___PACKAGE__] = ACTIONS(3141),
    [anon_sym___SUB__] = ACTIONS(3141),
    [anon_sym_break] = ACTIONS(3141),
    [anon_sym_fork] = ACTIONS(3141),
    [anon_sym_getppid] = ACTIONS(3141),
    [anon_sym_time] = ACTIONS(3141),
    [anon_sym_times] = ACTIONS(3141),
    [anon_sym_wait] = ACTIONS(3141),
    [anon_sym_wantarray] = ACTIONS(3141),
    [anon_sym_abs] = ACTIONS(3141),
    [anon_sym_alarm] = ACTIONS(3141),
    [anon_sym_chop] = ACTIONS(3141),
    [anon_sym_chdir] = ACTIONS(3141),
    [anon_sym_close] = ACTIONS(3141),
    [anon_sym_closedir] = ACTIONS(3141),
    [anon_sym_caller] = ACTIONS(3141),
    [anon_sym_chomp] = ACTIONS(3141),
    [anon_sym_chr] = ACTIONS(3141),
    [anon_sym_cos] = ACTIONS(3141),
    [anon_sym_chroot] = ACTIONS(3141),
    [anon_sym_defined] = ACTIONS(3141),
    [anon_sym_delete] = ACTIONS(3141),
    [anon_sym_dbmclose] = ACTIONS(3141),
    [anon_sym_exists] = ACTIONS(3141),
    [anon_sym_exit] = ACTIONS(3141),
    [anon_sym_eof] = ACTIONS(3141),
    [anon_sym_exp] = ACTIONS(3141),
    [anon_sym_each] = ACTIONS(3141),
    [anon_sym_fc] = ACTIONS(3141),
    [anon_sym_fileno] = ACTIONS(3141),
    [anon_sym_gmtime] = ACTIONS(3141),
    [anon_sym_getc] = ACTIONS(3141),
    [anon_sym_getpgrp] = ACTIONS(3141),
    [anon_sym_getprotobyname] = ACTIONS(3141),
    [anon_sym_getpwname] = ACTIONS(3141),
    [anon_sym_getpwuid] = ACTIONS(3141),
    [anon_sym_getpeername] = ACTIONS(3141),
    [anon_sym_getnetbyname] = ACTIONS(3141),
    [anon_sym_getsockname] = ACTIONS(3141),
    [anon_sym_getgrnam] = ACTIONS(3141),
    [anon_sym_getgrgid] = ACTIONS(3141),
    [anon_sym_hex] = ACTIONS(3141),
    [anon_sym_int] = ACTIONS(3141),
    [anon_sym_keys] = ACTIONS(3141),
    [anon_sym_lc] = ACTIONS(3141),
    [anon_sym_lcfirst] = ACTIONS(3141),
    [anon_sym_length] = ACTIONS(3141),
    [anon_sym_localtime] = ACTIONS(3141),
    [anon_sym_log] = ACTIONS(3141),
    [anon_sym_lock] = ACTIONS(3141),
    [anon_sym_lstat] = ACTIONS(3141),
    [anon_sym_oct] = ACTIONS(3141),
    [anon_sym_ord] = ACTIONS(3141),
    [anon_sym_prototype] = ACTIONS(3141),
    [anon_sym_pop] = ACTIONS(3141),
    [anon_sym_pos] = ACTIONS(3141),
    [anon_sym_quotemeta] = ACTIONS(3141),
    [anon_sym_reset] = ACTIONS(3141),
    [anon_sym_rand] = ACTIONS(3141),
    [anon_sym_rmdir] = ACTIONS(3141),
    [anon_sym_readdir] = ACTIONS(3141),
    [anon_sym_readline] = ACTIONS(3141),
    [anon_sym_readpipe] = ACTIONS(3141),
    [anon_sym_rewinddir] = ACTIONS(3141),
    [anon_sym_readlink] = ACTIONS(3141),
    [anon_sym_ref] = ACTIONS(3141),
    [anon_sym_scalar] = ACTIONS(3141),
    [anon_sym_shift] = ACTIONS(3141),
    [anon_sym_sin] = ACTIONS(3141),
    [anon_sym_sleep] = ACTIONS(3141),
    [anon_sym_sqrt] = ACTIONS(3141),
    [anon_sym_srand] = ACTIONS(3141),
    [anon_sym_stat] = ACTIONS(3141),
    [anon_sym_study] = ACTIONS(3141),
    [anon_sym_tell] = ACTIONS(3141),
    [anon_sym_telldir] = ACTIONS(3141),
    [anon_sym_tied] = ACTIONS(3141),
    [anon_sym_uc] = ACTIONS(3141),
    [anon_sym_ucfirst] = ACTIONS(3141),
    [anon_sym_untie] = ACTIONS(3141),
    [anon_sym_umask] = ACTIONS(3141),
    [anon_sym_values] = ACTIONS(3141),
    [anon_sym_write] = ACTIONS(3141),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3141),
    [anon_sym_qq] = ACTIONS(3141),
    [anon_sym_qw] = ACTIONS(3141),
    [anon_sym_qx] = ACTIONS(3141),
    [anon_sym_qr] = ACTIONS(3141),
    [anon_sym_m] = ACTIONS(3141),
    [anon_sym_s] = ACTIONS(3141),
    [anon_sym_tr] = ACTIONS(3141),
    [anon_sym_y] = ACTIONS(3141),
    [aux_sym__bareword_token1] = ACTIONS(3141),
    [sym_number] = ACTIONS(3141),
    [anon_sym_true] = ACTIONS(3141),
    [anon_sym_false] = ACTIONS(3141),
    [sym__single_quote] = ACTIONS(3139),
    [sym__double_quote] = ACTIONS(3139),
    [sym__backtick_quote] = ACTIONS(3139),
    [sym__search_slash_quote] = ACTIONS(3139),
    [sym__open_readline_bracket] = ACTIONS(3139),
    [sym__open_fileglob_bracket] = ACTIONS(3139),
    [sym__PERLY_HEREDOC] = ACTIONS(3139),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3139),
    [sym__filetest] = ACTIONS(3139),
  },
  [STATE(1680)] = {
    [sym_heredoc_content] = STATE(1680),
    [sym__identifier] = ACTIONS(3161),
    [anon_sym_LBRACE] = ACTIONS(3161),
    [anon_sym_DOLLAR] = ACTIONS(3161),
    [anon_sym_AT] = ACTIONS(3161),
    [anon_sym_RPAREN] = ACTIONS(3161),
    [anon_sym_my] = ACTIONS(3161),
    [anon_sym_extended] = ACTIONS(3161),
    [anon_sym_async] = ACTIONS(3161),
    [anon_sym_sub] = ACTIONS(3161),
    [anon_sym_method] = ACTIONS(3161),
    [anon_sym_LPAREN] = ACTIONS(3161),
    [anon_sym_continue] = ACTIONS(3161),
    [anon_sym_state] = ACTIONS(3161),
    [anon_sym_our] = ACTIONS(3161),
    [anon_sym_LBRACK] = ACTIONS(3161),
    [anon_sym_LT_LT] = ACTIONS(3161),
    [anon_sym_SLASH_SLASH] = ACTIONS(3161),
    [anon_sym_PLUS] = ACTIONS(3161),
    [anon_sym_DASH] = ACTIONS(3161),
    [anon_sym_TILDE] = ACTIONS(3161),
    [anon_sym_BANG] = ACTIONS(3161),
    [anon_sym_PLUS_PLUS] = ACTIONS(3161),
    [anon_sym_DASH_DASH] = ACTIONS(3161),
    [anon_sym_BSLASH] = ACTIONS(3159),
    [anon_sym_do] = ACTIONS(3161),
    [anon_sym_eval] = ACTIONS(3161),
    [anon_sym_await] = ACTIONS(3161),
    [anon_sym_field] = ACTIONS(3161),
    [anon_sym_local] = ACTIONS(3161),
    [anon_sym_dynamically] = ACTIONS(3161),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3161),
    [anon_sym_require] = ACTIONS(3161),
    [anon_sym_map] = ACTIONS(3161),
    [anon_sym_grep] = ACTIONS(3161),
    [anon_sym_sort] = ACTIONS(3161),
    [anon_sym_goto] = ACTIONS(3161),
    [anon_sym_return] = ACTIONS(3161),
    [anon_sym_undef] = ACTIONS(3161),
    [anon_sym_PERCENT2] = ACTIONS(3161),
    [anon_sym_AMP2] = ACTIONS(3161),
    [anon_sym_STAR2] = ACTIONS(3161),
    [anon_sym_last] = ACTIONS(3161),
    [anon_sym_next] = ACTIONS(3161),
    [anon_sym_redo] = ACTIONS(3161),
    [anon_sym___FILE__] = ACTIONS(3161),
    [anon_sym___LINE__] = ACTIONS(3161),
    [anon_sym___PACKAGE__] = ACTIONS(3161),
    [anon_sym___SUB__] = ACTIONS(3161),
    [anon_sym_break] = ACTIONS(3161),
    [anon_sym_fork] = ACTIONS(3161),
    [anon_sym_getppid] = ACTIONS(3161),
    [anon_sym_time] = ACTIONS(3161),
    [anon_sym_times] = ACTIONS(3161),
    [anon_sym_wait] = ACTIONS(3161),
    [anon_sym_wantarray] = ACTIONS(3161),
    [anon_sym_abs] = ACTIONS(3161),
    [anon_sym_alarm] = ACTIONS(3161),
    [anon_sym_chop] = ACTIONS(3161),
    [anon_sym_chdir] = ACTIONS(3161),
    [anon_sym_close] = ACTIONS(3161),
    [anon_sym_closedir] = ACTIONS(3161),
    [anon_sym_caller] = ACTIONS(3161),
    [anon_sym_chomp] = ACTIONS(3161),
    [anon_sym_chr] = ACTIONS(3161),
    [anon_sym_cos] = ACTIONS(3161),
    [anon_sym_chroot] = ACTIONS(3161),
    [anon_sym_defined] = ACTIONS(3161),
    [anon_sym_delete] = ACTIONS(3161),
    [anon_sym_dbmclose] = ACTIONS(3161),
    [anon_sym_exists] = ACTIONS(3161),
    [anon_sym_exit] = ACTIONS(3161),
    [anon_sym_eof] = ACTIONS(3161),
    [anon_sym_exp] = ACTIONS(3161),
    [anon_sym_each] = ACTIONS(3161),
    [anon_sym_fc] = ACTIONS(3161),
    [anon_sym_fileno] = ACTIONS(3161),
    [anon_sym_gmtime] = ACTIONS(3161),
    [anon_sym_getc] = ACTIONS(3161),
    [anon_sym_getpgrp] = ACTIONS(3161),
    [anon_sym_getprotobyname] = ACTIONS(3161),
    [anon_sym_getpwname] = ACTIONS(3161),
    [anon_sym_getpwuid] = ACTIONS(3161),
    [anon_sym_getpeername] = ACTIONS(3161),
    [anon_sym_getnetbyname] = ACTIONS(3161),
    [anon_sym_getsockname] = ACTIONS(3161),
    [anon_sym_getgrnam] = ACTIONS(3161),
    [anon_sym_getgrgid] = ACTIONS(3161),
    [anon_sym_hex] = ACTIONS(3161),
    [anon_sym_int] = ACTIONS(3161),
    [anon_sym_keys] = ACTIONS(3161),
    [anon_sym_lc] = ACTIONS(3161),
    [anon_sym_lcfirst] = ACTIONS(3161),
    [anon_sym_length] = ACTIONS(3161),
    [anon_sym_localtime] = ACTIONS(3161),
    [anon_sym_log] = ACTIONS(3161),
    [anon_sym_lock] = ACTIONS(3161),
    [anon_sym_lstat] = ACTIONS(3161),
    [anon_sym_oct] = ACTIONS(3161),
    [anon_sym_ord] = ACTIONS(3161),
    [anon_sym_prototype] = ACTIONS(3161),
    [anon_sym_pop] = ACTIONS(3161),
    [anon_sym_pos] = ACTIONS(3161),
    [anon_sym_quotemeta] = ACTIONS(3161),
    [anon_sym_reset] = ACTIONS(3161),
    [anon_sym_rand] = ACTIONS(3161),
    [anon_sym_rmdir] = ACTIONS(3161),
    [anon_sym_readdir] = ACTIONS(3161),
    [anon_sym_readline] = ACTIONS(3161),
    [anon_sym_readpipe] = ACTIONS(3161),
    [anon_sym_rewinddir] = ACTIONS(3161),
    [anon_sym_readlink] = ACTIONS(3161),
    [anon_sym_ref] = ACTIONS(3161),
    [anon_sym_scalar] = ACTIONS(3161),
    [anon_sym_shift] = ACTIONS(3161),
    [anon_sym_sin] = ACTIONS(3161),
    [anon_sym_sleep] = ACTIONS(3161),
    [anon_sym_sqrt] = ACTIONS(3161),
    [anon_sym_srand] = ACTIONS(3161),
    [anon_sym_stat] = ACTIONS(3161),
    [anon_sym_study] = ACTIONS(3161),
    [anon_sym_tell] = ACTIONS(3161),
    [anon_sym_telldir] = ACTIONS(3161),
    [anon_sym_tied] = ACTIONS(3161),
    [anon_sym_uc] = ACTIONS(3161),
    [anon_sym_ucfirst] = ACTIONS(3161),
    [anon_sym_untie] = ACTIONS(3161),
    [anon_sym_umask] = ACTIONS(3161),
    [anon_sym_values] = ACTIONS(3161),
    [anon_sym_write] = ACTIONS(3161),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3161),
    [anon_sym_qq] = ACTIONS(3161),
    [anon_sym_qw] = ACTIONS(3161),
    [anon_sym_qx] = ACTIONS(3161),
    [anon_sym_qr] = ACTIONS(3161),
    [anon_sym_m] = ACTIONS(3161),
    [anon_sym_s] = ACTIONS(3161),
    [anon_sym_tr] = ACTIONS(3161),
    [anon_sym_y] = ACTIONS(3161),
    [aux_sym__bareword_token1] = ACTIONS(3161),
    [sym_number] = ACTIONS(3161),
    [anon_sym_true] = ACTIONS(3161),
    [anon_sym_false] = ACTIONS(3161),
    [sym__single_quote] = ACTIONS(3159),
    [sym__double_quote] = ACTIONS(3159),
    [sym__backtick_quote] = ACTIONS(3159),
    [sym__search_slash_quote] = ACTIONS(3159),
    [sym__open_readline_bracket] = ACTIONS(3159),
    [sym__open_fileglob_bracket] = ACTIONS(3159),
    [sym__PERLY_HEREDOC] = ACTIONS(3159),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3159),
    [sym__filetest] = ACTIONS(3159),
  },
  [STATE(1681)] = {
    [sym_heredoc_content] = STATE(1681),
    [sym__identifier] = ACTIONS(3173),
    [anon_sym_LBRACE] = ACTIONS(3173),
    [anon_sym_DOLLAR] = ACTIONS(3173),
    [anon_sym_AT] = ACTIONS(3173),
    [anon_sym_RPAREN] = ACTIONS(3173),
    [anon_sym_my] = ACTIONS(3173),
    [anon_sym_extended] = ACTIONS(3173),
    [anon_sym_async] = ACTIONS(3173),
    [anon_sym_sub] = ACTIONS(3173),
    [anon_sym_method] = ACTIONS(3173),
    [anon_sym_LPAREN] = ACTIONS(3173),
    [anon_sym_continue] = ACTIONS(3173),
    [anon_sym_state] = ACTIONS(3173),
    [anon_sym_our] = ACTIONS(3173),
    [anon_sym_LBRACK] = ACTIONS(3173),
    [anon_sym_LT_LT] = ACTIONS(3173),
    [anon_sym_SLASH_SLASH] = ACTIONS(3173),
    [anon_sym_PLUS] = ACTIONS(3173),
    [anon_sym_DASH] = ACTIONS(3173),
    [anon_sym_TILDE] = ACTIONS(3173),
    [anon_sym_BANG] = ACTIONS(3173),
    [anon_sym_PLUS_PLUS] = ACTIONS(3173),
    [anon_sym_DASH_DASH] = ACTIONS(3173),
    [anon_sym_BSLASH] = ACTIONS(3171),
    [anon_sym_do] = ACTIONS(3173),
    [anon_sym_eval] = ACTIONS(3173),
    [anon_sym_await] = ACTIONS(3173),
    [anon_sym_field] = ACTIONS(3173),
    [anon_sym_local] = ACTIONS(3173),
    [anon_sym_dynamically] = ACTIONS(3173),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3173),
    [anon_sym_require] = ACTIONS(3173),
    [anon_sym_map] = ACTIONS(3173),
    [anon_sym_grep] = ACTIONS(3173),
    [anon_sym_sort] = ACTIONS(3173),
    [anon_sym_goto] = ACTIONS(3173),
    [anon_sym_return] = ACTIONS(3173),
    [anon_sym_undef] = ACTIONS(3173),
    [anon_sym_PERCENT2] = ACTIONS(3173),
    [anon_sym_AMP2] = ACTIONS(3173),
    [anon_sym_STAR2] = ACTIONS(3173),
    [anon_sym_last] = ACTIONS(3173),
    [anon_sym_next] = ACTIONS(3173),
    [anon_sym_redo] = ACTIONS(3173),
    [anon_sym___FILE__] = ACTIONS(3173),
    [anon_sym___LINE__] = ACTIONS(3173),
    [anon_sym___PACKAGE__] = ACTIONS(3173),
    [anon_sym___SUB__] = ACTIONS(3173),
    [anon_sym_break] = ACTIONS(3173),
    [anon_sym_fork] = ACTIONS(3173),
    [anon_sym_getppid] = ACTIONS(3173),
    [anon_sym_time] = ACTIONS(3173),
    [anon_sym_times] = ACTIONS(3173),
    [anon_sym_wait] = ACTIONS(3173),
    [anon_sym_wantarray] = ACTIONS(3173),
    [anon_sym_abs] = ACTIONS(3173),
    [anon_sym_alarm] = ACTIONS(3173),
    [anon_sym_chop] = ACTIONS(3173),
    [anon_sym_chdir] = ACTIONS(3173),
    [anon_sym_close] = ACTIONS(3173),
    [anon_sym_closedir] = ACTIONS(3173),
    [anon_sym_caller] = ACTIONS(3173),
    [anon_sym_chomp] = ACTIONS(3173),
    [anon_sym_chr] = ACTIONS(3173),
    [anon_sym_cos] = ACTIONS(3173),
    [anon_sym_chroot] = ACTIONS(3173),
    [anon_sym_defined] = ACTIONS(3173),
    [anon_sym_delete] = ACTIONS(3173),
    [anon_sym_dbmclose] = ACTIONS(3173),
    [anon_sym_exists] = ACTIONS(3173),
    [anon_sym_exit] = ACTIONS(3173),
    [anon_sym_eof] = ACTIONS(3173),
    [anon_sym_exp] = ACTIONS(3173),
    [anon_sym_each] = ACTIONS(3173),
    [anon_sym_fc] = ACTIONS(3173),
    [anon_sym_fileno] = ACTIONS(3173),
    [anon_sym_gmtime] = ACTIONS(3173),
    [anon_sym_getc] = ACTIONS(3173),
    [anon_sym_getpgrp] = ACTIONS(3173),
    [anon_sym_getprotobyname] = ACTIONS(3173),
    [anon_sym_getpwname] = ACTIONS(3173),
    [anon_sym_getpwuid] = ACTIONS(3173),
    [anon_sym_getpeername] = ACTIONS(3173),
    [anon_sym_getnetbyname] = ACTIONS(3173),
    [anon_sym_getsockname] = ACTIONS(3173),
    [anon_sym_getgrnam] = ACTIONS(3173),
    [anon_sym_getgrgid] = ACTIONS(3173),
    [anon_sym_hex] = ACTIONS(3173),
    [anon_sym_int] = ACTIONS(3173),
    [anon_sym_keys] = ACTIONS(3173),
    [anon_sym_lc] = ACTIONS(3173),
    [anon_sym_lcfirst] = ACTIONS(3173),
    [anon_sym_length] = ACTIONS(3173),
    [anon_sym_localtime] = ACTIONS(3173),
    [anon_sym_log] = ACTIONS(3173),
    [anon_sym_lock] = ACTIONS(3173),
    [anon_sym_lstat] = ACTIONS(3173),
    [anon_sym_oct] = ACTIONS(3173),
    [anon_sym_ord] = ACTIONS(3173),
    [anon_sym_prototype] = ACTIONS(3173),
    [anon_sym_pop] = ACTIONS(3173),
    [anon_sym_pos] = ACTIONS(3173),
    [anon_sym_quotemeta] = ACTIONS(3173),
    [anon_sym_reset] = ACTIONS(3173),
    [anon_sym_rand] = ACTIONS(3173),
    [anon_sym_rmdir] = ACTIONS(3173),
    [anon_sym_readdir] = ACTIONS(3173),
    [anon_sym_readline] = ACTIONS(3173),
    [anon_sym_readpipe] = ACTIONS(3173),
    [anon_sym_rewinddir] = ACTIONS(3173),
    [anon_sym_readlink] = ACTIONS(3173),
    [anon_sym_ref] = ACTIONS(3173),
    [anon_sym_scalar] = ACTIONS(3173),
    [anon_sym_shift] = ACTIONS(3173),
    [anon_sym_sin] = ACTIONS(3173),
    [anon_sym_sleep] = ACTIONS(3173),
    [anon_sym_sqrt] = ACTIONS(3173),
    [anon_sym_srand] = ACTIONS(3173),
    [anon_sym_stat] = ACTIONS(3173),
    [anon_sym_study] = ACTIONS(3173),
    [anon_sym_tell] = ACTIONS(3173),
    [anon_sym_telldir] = ACTIONS(3173),
    [anon_sym_tied] = ACTIONS(3173),
    [anon_sym_uc] = ACTIONS(3173),
    [anon_sym_ucfirst] = ACTIONS(3173),
    [anon_sym_untie] = ACTIONS(3173),
    [anon_sym_umask] = ACTIONS(3173),
    [anon_sym_values] = ACTIONS(3173),
    [anon_sym_write] = ACTIONS(3173),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3173),
    [anon_sym_qq] = ACTIONS(3173),
    [anon_sym_qw] = ACTIONS(3173),
    [anon_sym_qx] = ACTIONS(3173),
    [anon_sym_qr] = ACTIONS(3173),
    [anon_sym_m] = ACTIONS(3173),
    [anon_sym_s] = ACTIONS(3173),
    [anon_sym_tr] = ACTIONS(3173),
    [anon_sym_y] = ACTIONS(3173),
    [aux_sym__bareword_token1] = ACTIONS(3173),
    [sym_number] = ACTIONS(3173),
    [anon_sym_true] = ACTIONS(3173),
    [anon_sym_false] = ACTIONS(3173),
    [sym__single_quote] = ACTIONS(3171),
    [sym__double_quote] = ACTIONS(3171),
    [sym__backtick_quote] = ACTIONS(3171),
    [sym__search_slash_quote] = ACTIONS(3171),
    [sym__open_readline_bracket] = ACTIONS(3171),
    [sym__open_fileglob_bracket] = ACTIONS(3171),
    [sym__PERLY_HEREDOC] = ACTIONS(3171),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3171),
    [sym__filetest] = ACTIONS(3171),
  },
  [STATE(1682)] = {
    [sym_heredoc_content] = STATE(1682),
    [sym__identifier] = ACTIONS(3181),
    [anon_sym_LBRACE] = ACTIONS(3181),
    [anon_sym_DOLLAR] = ACTIONS(3181),
    [anon_sym_AT] = ACTIONS(3181),
    [anon_sym_RPAREN] = ACTIONS(3181),
    [anon_sym_my] = ACTIONS(3181),
    [anon_sym_extended] = ACTIONS(3181),
    [anon_sym_async] = ACTIONS(3181),
    [anon_sym_sub] = ACTIONS(3181),
    [anon_sym_method] = ACTIONS(3181),
    [anon_sym_LPAREN] = ACTIONS(3181),
    [anon_sym_continue] = ACTIONS(3181),
    [anon_sym_state] = ACTIONS(3181),
    [anon_sym_our] = ACTIONS(3181),
    [anon_sym_LBRACK] = ACTIONS(3181),
    [anon_sym_LT_LT] = ACTIONS(3181),
    [anon_sym_SLASH_SLASH] = ACTIONS(3181),
    [anon_sym_PLUS] = ACTIONS(3181),
    [anon_sym_DASH] = ACTIONS(3181),
    [anon_sym_TILDE] = ACTIONS(3181),
    [anon_sym_BANG] = ACTIONS(3181),
    [anon_sym_PLUS_PLUS] = ACTIONS(3181),
    [anon_sym_DASH_DASH] = ACTIONS(3181),
    [anon_sym_BSLASH] = ACTIONS(3179),
    [anon_sym_do] = ACTIONS(3181),
    [anon_sym_eval] = ACTIONS(3181),
    [anon_sym_await] = ACTIONS(3181),
    [anon_sym_field] = ACTIONS(3181),
    [anon_sym_local] = ACTIONS(3181),
    [anon_sym_dynamically] = ACTIONS(3181),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3181),
    [anon_sym_require] = ACTIONS(3181),
    [anon_sym_map] = ACTIONS(3181),
    [anon_sym_grep] = ACTIONS(3181),
    [anon_sym_sort] = ACTIONS(3181),
    [anon_sym_goto] = ACTIONS(3181),
    [anon_sym_return] = ACTIONS(3181),
    [anon_sym_undef] = ACTIONS(3181),
    [anon_sym_PERCENT2] = ACTIONS(3181),
    [anon_sym_AMP2] = ACTIONS(3181),
    [anon_sym_STAR2] = ACTIONS(3181),
    [anon_sym_last] = ACTIONS(3181),
    [anon_sym_next] = ACTIONS(3181),
    [anon_sym_redo] = ACTIONS(3181),
    [anon_sym___FILE__] = ACTIONS(3181),
    [anon_sym___LINE__] = ACTIONS(3181),
    [anon_sym___PACKAGE__] = ACTIONS(3181),
    [anon_sym___SUB__] = ACTIONS(3181),
    [anon_sym_break] = ACTIONS(3181),
    [anon_sym_fork] = ACTIONS(3181),
    [anon_sym_getppid] = ACTIONS(3181),
    [anon_sym_time] = ACTIONS(3181),
    [anon_sym_times] = ACTIONS(3181),
    [anon_sym_wait] = ACTIONS(3181),
    [anon_sym_wantarray] = ACTIONS(3181),
    [anon_sym_abs] = ACTIONS(3181),
    [anon_sym_alarm] = ACTIONS(3181),
    [anon_sym_chop] = ACTIONS(3181),
    [anon_sym_chdir] = ACTIONS(3181),
    [anon_sym_close] = ACTIONS(3181),
    [anon_sym_closedir] = ACTIONS(3181),
    [anon_sym_caller] = ACTIONS(3181),
    [anon_sym_chomp] = ACTIONS(3181),
    [anon_sym_chr] = ACTIONS(3181),
    [anon_sym_cos] = ACTIONS(3181),
    [anon_sym_chroot] = ACTIONS(3181),
    [anon_sym_defined] = ACTIONS(3181),
    [anon_sym_delete] = ACTIONS(3181),
    [anon_sym_dbmclose] = ACTIONS(3181),
    [anon_sym_exists] = ACTIONS(3181),
    [anon_sym_exit] = ACTIONS(3181),
    [anon_sym_eof] = ACTIONS(3181),
    [anon_sym_exp] = ACTIONS(3181),
    [anon_sym_each] = ACTIONS(3181),
    [anon_sym_fc] = ACTIONS(3181),
    [anon_sym_fileno] = ACTIONS(3181),
    [anon_sym_gmtime] = ACTIONS(3181),
    [anon_sym_getc] = ACTIONS(3181),
    [anon_sym_getpgrp] = ACTIONS(3181),
    [anon_sym_getprotobyname] = ACTIONS(3181),
    [anon_sym_getpwname] = ACTIONS(3181),
    [anon_sym_getpwuid] = ACTIONS(3181),
    [anon_sym_getpeername] = ACTIONS(3181),
    [anon_sym_getnetbyname] = ACTIONS(3181),
    [anon_sym_getsockname] = ACTIONS(3181),
    [anon_sym_getgrnam] = ACTIONS(3181),
    [anon_sym_getgrgid] = ACTIONS(3181),
    [anon_sym_hex] = ACTIONS(3181),
    [anon_sym_int] = ACTIONS(3181),
    [anon_sym_keys] = ACTIONS(3181),
    [anon_sym_lc] = ACTIONS(3181),
    [anon_sym_lcfirst] = ACTIONS(3181),
    [anon_sym_length] = ACTIONS(3181),
    [anon_sym_localtime] = ACTIONS(3181),
    [anon_sym_log] = ACTIONS(3181),
    [anon_sym_lock] = ACTIONS(3181),
    [anon_sym_lstat] = ACTIONS(3181),
    [anon_sym_oct] = ACTIONS(3181),
    [anon_sym_ord] = ACTIONS(3181),
    [anon_sym_prototype] = ACTIONS(3181),
    [anon_sym_pop] = ACTIONS(3181),
    [anon_sym_pos] = ACTIONS(3181),
    [anon_sym_quotemeta] = ACTIONS(3181),
    [anon_sym_reset] = ACTIONS(3181),
    [anon_sym_rand] = ACTIONS(3181),
    [anon_sym_rmdir] = ACTIONS(3181),
    [anon_sym_readdir] = ACTIONS(3181),
    [anon_sym_readline] = ACTIONS(3181),
    [anon_sym_readpipe] = ACTIONS(3181),
    [anon_sym_rewinddir] = ACTIONS(3181),
    [anon_sym_readlink] = ACTIONS(3181),
    [anon_sym_ref] = ACTIONS(3181),
    [anon_sym_scalar] = ACTIONS(3181),
    [anon_sym_shift] = ACTIONS(3181),
    [anon_sym_sin] = ACTIONS(3181),
    [anon_sym_sleep] = ACTIONS(3181),
    [anon_sym_sqrt] = ACTIONS(3181),
    [anon_sym_srand] = ACTIONS(3181),
    [anon_sym_stat] = ACTIONS(3181),
    [anon_sym_study] = ACTIONS(3181),
    [anon_sym_tell] = ACTIONS(3181),
    [anon_sym_telldir] = ACTIONS(3181),
    [anon_sym_tied] = ACTIONS(3181),
    [anon_sym_uc] = ACTIONS(3181),
    [anon_sym_ucfirst] = ACTIONS(3181),
    [anon_sym_untie] = ACTIONS(3181),
    [anon_sym_umask] = ACTIONS(3181),
    [anon_sym_values] = ACTIONS(3181),
    [anon_sym_write] = ACTIONS(3181),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3181),
    [anon_sym_qq] = ACTIONS(3181),
    [anon_sym_qw] = ACTIONS(3181),
    [anon_sym_qx] = ACTIONS(3181),
    [anon_sym_qr] = ACTIONS(3181),
    [anon_sym_m] = ACTIONS(3181),
    [anon_sym_s] = ACTIONS(3181),
    [anon_sym_tr] = ACTIONS(3181),
    [anon_sym_y] = ACTIONS(3181),
    [aux_sym__bareword_token1] = ACTIONS(3181),
    [sym_number] = ACTIONS(3181),
    [anon_sym_true] = ACTIONS(3181),
    [anon_sym_false] = ACTIONS(3181),
    [sym__single_quote] = ACTIONS(3179),
    [sym__double_quote] = ACTIONS(3179),
    [sym__backtick_quote] = ACTIONS(3179),
    [sym__search_slash_quote] = ACTIONS(3179),
    [sym__open_readline_bracket] = ACTIONS(3179),
    [sym__open_fileglob_bracket] = ACTIONS(3179),
    [sym__PERLY_HEREDOC] = ACTIONS(3179),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3179),
    [sym__filetest] = ACTIONS(3179),
  },
  [STATE(1683)] = {
    [sym_heredoc_content] = STATE(1683),
    [sym__identifier] = ACTIONS(3193),
    [anon_sym_LBRACE] = ACTIONS(3193),
    [anon_sym_DOLLAR] = ACTIONS(3193),
    [anon_sym_AT] = ACTIONS(3193),
    [anon_sym_RPAREN] = ACTIONS(3193),
    [anon_sym_my] = ACTIONS(3193),
    [anon_sym_extended] = ACTIONS(3193),
    [anon_sym_async] = ACTIONS(3193),
    [anon_sym_sub] = ACTIONS(3193),
    [anon_sym_method] = ACTIONS(3193),
    [anon_sym_LPAREN] = ACTIONS(3193),
    [anon_sym_continue] = ACTIONS(3193),
    [anon_sym_state] = ACTIONS(3193),
    [anon_sym_our] = ACTIONS(3193),
    [anon_sym_LBRACK] = ACTIONS(3193),
    [anon_sym_LT_LT] = ACTIONS(3193),
    [anon_sym_SLASH_SLASH] = ACTIONS(3193),
    [anon_sym_PLUS] = ACTIONS(3193),
    [anon_sym_DASH] = ACTIONS(3193),
    [anon_sym_TILDE] = ACTIONS(3193),
    [anon_sym_BANG] = ACTIONS(3193),
    [anon_sym_PLUS_PLUS] = ACTIONS(3193),
    [anon_sym_DASH_DASH] = ACTIONS(3193),
    [anon_sym_BSLASH] = ACTIONS(3191),
    [anon_sym_do] = ACTIONS(3193),
    [anon_sym_eval] = ACTIONS(3193),
    [anon_sym_await] = ACTIONS(3193),
    [anon_sym_field] = ACTIONS(3193),
    [anon_sym_local] = ACTIONS(3193),
    [anon_sym_dynamically] = ACTIONS(3193),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3193),
    [anon_sym_require] = ACTIONS(3193),
    [anon_sym_map] = ACTIONS(3193),
    [anon_sym_grep] = ACTIONS(3193),
    [anon_sym_sort] = ACTIONS(3193),
    [anon_sym_goto] = ACTIONS(3193),
    [anon_sym_return] = ACTIONS(3193),
    [anon_sym_undef] = ACTIONS(3193),
    [anon_sym_PERCENT2] = ACTIONS(3193),
    [anon_sym_AMP2] = ACTIONS(3193),
    [anon_sym_STAR2] = ACTIONS(3193),
    [anon_sym_last] = ACTIONS(3193),
    [anon_sym_next] = ACTIONS(3193),
    [anon_sym_redo] = ACTIONS(3193),
    [anon_sym___FILE__] = ACTIONS(3193),
    [anon_sym___LINE__] = ACTIONS(3193),
    [anon_sym___PACKAGE__] = ACTIONS(3193),
    [anon_sym___SUB__] = ACTIONS(3193),
    [anon_sym_break] = ACTIONS(3193),
    [anon_sym_fork] = ACTIONS(3193),
    [anon_sym_getppid] = ACTIONS(3193),
    [anon_sym_time] = ACTIONS(3193),
    [anon_sym_times] = ACTIONS(3193),
    [anon_sym_wait] = ACTIONS(3193),
    [anon_sym_wantarray] = ACTIONS(3193),
    [anon_sym_abs] = ACTIONS(3193),
    [anon_sym_alarm] = ACTIONS(3193),
    [anon_sym_chop] = ACTIONS(3193),
    [anon_sym_chdir] = ACTIONS(3193),
    [anon_sym_close] = ACTIONS(3193),
    [anon_sym_closedir] = ACTIONS(3193),
    [anon_sym_caller] = ACTIONS(3193),
    [anon_sym_chomp] = ACTIONS(3193),
    [anon_sym_chr] = ACTIONS(3193),
    [anon_sym_cos] = ACTIONS(3193),
    [anon_sym_chroot] = ACTIONS(3193),
    [anon_sym_defined] = ACTIONS(3193),
    [anon_sym_delete] = ACTIONS(3193),
    [anon_sym_dbmclose] = ACTIONS(3193),
    [anon_sym_exists] = ACTIONS(3193),
    [anon_sym_exit] = ACTIONS(3193),
    [anon_sym_eof] = ACTIONS(3193),
    [anon_sym_exp] = ACTIONS(3193),
    [anon_sym_each] = ACTIONS(3193),
    [anon_sym_fc] = ACTIONS(3193),
    [anon_sym_fileno] = ACTIONS(3193),
    [anon_sym_gmtime] = ACTIONS(3193),
    [anon_sym_getc] = ACTIONS(3193),
    [anon_sym_getpgrp] = ACTIONS(3193),
    [anon_sym_getprotobyname] = ACTIONS(3193),
    [anon_sym_getpwname] = ACTIONS(3193),
    [anon_sym_getpwuid] = ACTIONS(3193),
    [anon_sym_getpeername] = ACTIONS(3193),
    [anon_sym_getnetbyname] = ACTIONS(3193),
    [anon_sym_getsockname] = ACTIONS(3193),
    [anon_sym_getgrnam] = ACTIONS(3193),
    [anon_sym_getgrgid] = ACTIONS(3193),
    [anon_sym_hex] = ACTIONS(3193),
    [anon_sym_int] = ACTIONS(3193),
    [anon_sym_keys] = ACTIONS(3193),
    [anon_sym_lc] = ACTIONS(3193),
    [anon_sym_lcfirst] = ACTIONS(3193),
    [anon_sym_length] = ACTIONS(3193),
    [anon_sym_localtime] = ACTIONS(3193),
    [anon_sym_log] = ACTIONS(3193),
    [anon_sym_lock] = ACTIONS(3193),
    [anon_sym_lstat] = ACTIONS(3193),
    [anon_sym_oct] = ACTIONS(3193),
    [anon_sym_ord] = ACTIONS(3193),
    [anon_sym_prototype] = ACTIONS(3193),
    [anon_sym_pop] = ACTIONS(3193),
    [anon_sym_pos] = ACTIONS(3193),
    [anon_sym_quotemeta] = ACTIONS(3193),
    [anon_sym_reset] = ACTIONS(3193),
    [anon_sym_rand] = ACTIONS(3193),
    [anon_sym_rmdir] = ACTIONS(3193),
    [anon_sym_readdir] = ACTIONS(3193),
    [anon_sym_readline] = ACTIONS(3193),
    [anon_sym_readpipe] = ACTIONS(3193),
    [anon_sym_rewinddir] = ACTIONS(3193),
    [anon_sym_readlink] = ACTIONS(3193),
    [anon_sym_ref] = ACTIONS(3193),
    [anon_sym_scalar] = ACTIONS(3193),
    [anon_sym_shift] = ACTIONS(3193),
    [anon_sym_sin] = ACTIONS(3193),
    [anon_sym_sleep] = ACTIONS(3193),
    [anon_sym_sqrt] = ACTIONS(3193),
    [anon_sym_srand] = ACTIONS(3193),
    [anon_sym_stat] = ACTIONS(3193),
    [anon_sym_study] = ACTIONS(3193),
    [anon_sym_tell] = ACTIONS(3193),
    [anon_sym_telldir] = ACTIONS(3193),
    [anon_sym_tied] = ACTIONS(3193),
    [anon_sym_uc] = ACTIONS(3193),
    [anon_sym_ucfirst] = ACTIONS(3193),
    [anon_sym_untie] = ACTIONS(3193),
    [anon_sym_umask] = ACTIONS(3193),
    [anon_sym_values] = ACTIONS(3193),
    [anon_sym_write] = ACTIONS(3193),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3193),
    [anon_sym_qq] = ACTIONS(3193),
    [anon_sym_qw] = ACTIONS(3193),
    [anon_sym_qx] = ACTIONS(3193),
    [anon_sym_qr] = ACTIONS(3193),
    [anon_sym_m] = ACTIONS(3193),
    [anon_sym_s] = ACTIONS(3193),
    [anon_sym_tr] = ACTIONS(3193),
    [anon_sym_y] = ACTIONS(3193),
    [aux_sym__bareword_token1] = ACTIONS(3193),
    [sym_number] = ACTIONS(3193),
    [anon_sym_true] = ACTIONS(3193),
    [anon_sym_false] = ACTIONS(3193),
    [sym__single_quote] = ACTIONS(3191),
    [sym__double_quote] = ACTIONS(3191),
    [sym__backtick_quote] = ACTIONS(3191),
    [sym__search_slash_quote] = ACTIONS(3191),
    [sym__open_readline_bracket] = ACTIONS(3191),
    [sym__open_fileglob_bracket] = ACTIONS(3191),
    [sym__PERLY_HEREDOC] = ACTIONS(3191),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3191),
    [sym__filetest] = ACTIONS(3191),
  },
  [STATE(1684)] = {
    [sym_heredoc_content] = STATE(1684),
    [sym__identifier] = ACTIONS(3197),
    [anon_sym_LBRACE] = ACTIONS(3197),
    [anon_sym_DOLLAR] = ACTIONS(3197),
    [anon_sym_AT] = ACTIONS(3197),
    [anon_sym_RPAREN] = ACTIONS(3197),
    [anon_sym_my] = ACTIONS(3197),
    [anon_sym_extended] = ACTIONS(3197),
    [anon_sym_async] = ACTIONS(3197),
    [anon_sym_sub] = ACTIONS(3197),
    [anon_sym_method] = ACTIONS(3197),
    [anon_sym_LPAREN] = ACTIONS(3197),
    [anon_sym_continue] = ACTIONS(3197),
    [anon_sym_state] = ACTIONS(3197),
    [anon_sym_our] = ACTIONS(3197),
    [anon_sym_LBRACK] = ACTIONS(3197),
    [anon_sym_LT_LT] = ACTIONS(3197),
    [anon_sym_SLASH_SLASH] = ACTIONS(3197),
    [anon_sym_PLUS] = ACTIONS(3197),
    [anon_sym_DASH] = ACTIONS(3197),
    [anon_sym_TILDE] = ACTIONS(3197),
    [anon_sym_BANG] = ACTIONS(3197),
    [anon_sym_PLUS_PLUS] = ACTIONS(3197),
    [anon_sym_DASH_DASH] = ACTIONS(3197),
    [anon_sym_BSLASH] = ACTIONS(3195),
    [anon_sym_do] = ACTIONS(3197),
    [anon_sym_eval] = ACTIONS(3197),
    [anon_sym_await] = ACTIONS(3197),
    [anon_sym_field] = ACTIONS(3197),
    [anon_sym_local] = ACTIONS(3197),
    [anon_sym_dynamically] = ACTIONS(3197),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3197),
    [anon_sym_require] = ACTIONS(3197),
    [anon_sym_map] = ACTIONS(3197),
    [anon_sym_grep] = ACTIONS(3197),
    [anon_sym_sort] = ACTIONS(3197),
    [anon_sym_goto] = ACTIONS(3197),
    [anon_sym_return] = ACTIONS(3197),
    [anon_sym_undef] = ACTIONS(3197),
    [anon_sym_PERCENT2] = ACTIONS(3197),
    [anon_sym_AMP2] = ACTIONS(3197),
    [anon_sym_STAR2] = ACTIONS(3197),
    [anon_sym_last] = ACTIONS(3197),
    [anon_sym_next] = ACTIONS(3197),
    [anon_sym_redo] = ACTIONS(3197),
    [anon_sym___FILE__] = ACTIONS(3197),
    [anon_sym___LINE__] = ACTIONS(3197),
    [anon_sym___PACKAGE__] = ACTIONS(3197),
    [anon_sym___SUB__] = ACTIONS(3197),
    [anon_sym_break] = ACTIONS(3197),
    [anon_sym_fork] = ACTIONS(3197),
    [anon_sym_getppid] = ACTIONS(3197),
    [anon_sym_time] = ACTIONS(3197),
    [anon_sym_times] = ACTIONS(3197),
    [anon_sym_wait] = ACTIONS(3197),
    [anon_sym_wantarray] = ACTIONS(3197),
    [anon_sym_abs] = ACTIONS(3197),
    [anon_sym_alarm] = ACTIONS(3197),
    [anon_sym_chop] = ACTIONS(3197),
    [anon_sym_chdir] = ACTIONS(3197),
    [anon_sym_close] = ACTIONS(3197),
    [anon_sym_closedir] = ACTIONS(3197),
    [anon_sym_caller] = ACTIONS(3197),
    [anon_sym_chomp] = ACTIONS(3197),
    [anon_sym_chr] = ACTIONS(3197),
    [anon_sym_cos] = ACTIONS(3197),
    [anon_sym_chroot] = ACTIONS(3197),
    [anon_sym_defined] = ACTIONS(3197),
    [anon_sym_delete] = ACTIONS(3197),
    [anon_sym_dbmclose] = ACTIONS(3197),
    [anon_sym_exists] = ACTIONS(3197),
    [anon_sym_exit] = ACTIONS(3197),
    [anon_sym_eof] = ACTIONS(3197),
    [anon_sym_exp] = ACTIONS(3197),
    [anon_sym_each] = ACTIONS(3197),
    [anon_sym_fc] = ACTIONS(3197),
    [anon_sym_fileno] = ACTIONS(3197),
    [anon_sym_gmtime] = ACTIONS(3197),
    [anon_sym_getc] = ACTIONS(3197),
    [anon_sym_getpgrp] = ACTIONS(3197),
    [anon_sym_getprotobyname] = ACTIONS(3197),
    [anon_sym_getpwname] = ACTIONS(3197),
    [anon_sym_getpwuid] = ACTIONS(3197),
    [anon_sym_getpeername] = ACTIONS(3197),
    [anon_sym_getnetbyname] = ACTIONS(3197),
    [anon_sym_getsockname] = ACTIONS(3197),
    [anon_sym_getgrnam] = ACTIONS(3197),
    [anon_sym_getgrgid] = ACTIONS(3197),
    [anon_sym_hex] = ACTIONS(3197),
    [anon_sym_int] = ACTIONS(3197),
    [anon_sym_keys] = ACTIONS(3197),
    [anon_sym_lc] = ACTIONS(3197),
    [anon_sym_lcfirst] = ACTIONS(3197),
    [anon_sym_length] = ACTIONS(3197),
    [anon_sym_localtime] = ACTIONS(3197),
    [anon_sym_log] = ACTIONS(3197),
    [anon_sym_lock] = ACTIONS(3197),
    [anon_sym_lstat] = ACTIONS(3197),
    [anon_sym_oct] = ACTIONS(3197),
    [anon_sym_ord] = ACTIONS(3197),
    [anon_sym_prototype] = ACTIONS(3197),
    [anon_sym_pop] = ACTIONS(3197),
    [anon_sym_pos] = ACTIONS(3197),
    [anon_sym_quotemeta] = ACTIONS(3197),
    [anon_sym_reset] = ACTIONS(3197),
    [anon_sym_rand] = ACTIONS(3197),
    [anon_sym_rmdir] = ACTIONS(3197),
    [anon_sym_readdir] = ACTIONS(3197),
    [anon_sym_readline] = ACTIONS(3197),
    [anon_sym_readpipe] = ACTIONS(3197),
    [anon_sym_rewinddir] = ACTIONS(3197),
    [anon_sym_readlink] = ACTIONS(3197),
    [anon_sym_ref] = ACTIONS(3197),
    [anon_sym_scalar] = ACTIONS(3197),
    [anon_sym_shift] = ACTIONS(3197),
    [anon_sym_sin] = ACTIONS(3197),
    [anon_sym_sleep] = ACTIONS(3197),
    [anon_sym_sqrt] = ACTIONS(3197),
    [anon_sym_srand] = ACTIONS(3197),
    [anon_sym_stat] = ACTIONS(3197),
    [anon_sym_study] = ACTIONS(3197),
    [anon_sym_tell] = ACTIONS(3197),
    [anon_sym_telldir] = ACTIONS(3197),
    [anon_sym_tied] = ACTIONS(3197),
    [anon_sym_uc] = ACTIONS(3197),
    [anon_sym_ucfirst] = ACTIONS(3197),
    [anon_sym_untie] = ACTIONS(3197),
    [anon_sym_umask] = ACTIONS(3197),
    [anon_sym_values] = ACTIONS(3197),
    [anon_sym_write] = ACTIONS(3197),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3197),
    [anon_sym_qq] = ACTIONS(3197),
    [anon_sym_qw] = ACTIONS(3197),
    [anon_sym_qx] = ACTIONS(3197),
    [anon_sym_qr] = ACTIONS(3197),
    [anon_sym_m] = ACTIONS(3197),
    [anon_sym_s] = ACTIONS(3197),
    [anon_sym_tr] = ACTIONS(3197),
    [anon_sym_y] = ACTIONS(3197),
    [aux_sym__bareword_token1] = ACTIONS(3197),
    [sym_number] = ACTIONS(3197),
    [anon_sym_true] = ACTIONS(3197),
    [anon_sym_false] = ACTIONS(3197),
    [sym__single_quote] = ACTIONS(3195),
    [sym__double_quote] = ACTIONS(3195),
    [sym__backtick_quote] = ACTIONS(3195),
    [sym__search_slash_quote] = ACTIONS(3195),
    [sym__open_readline_bracket] = ACTIONS(3195),
    [sym__open_fileglob_bracket] = ACTIONS(3195),
    [sym__PERLY_HEREDOC] = ACTIONS(3195),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3195),
    [sym__filetest] = ACTIONS(3195),
  },
  [STATE(1685)] = {
    [sym_heredoc_content] = STATE(1685),
    [sym__identifier] = ACTIONS(3209),
    [anon_sym_LBRACE] = ACTIONS(3209),
    [anon_sym_DOLLAR] = ACTIONS(3209),
    [anon_sym_AT] = ACTIONS(3209),
    [anon_sym_RPAREN] = ACTIONS(3209),
    [anon_sym_my] = ACTIONS(3209),
    [anon_sym_extended] = ACTIONS(3209),
    [anon_sym_async] = ACTIONS(3209),
    [anon_sym_sub] = ACTIONS(3209),
    [anon_sym_method] = ACTIONS(3209),
    [anon_sym_LPAREN] = ACTIONS(3209),
    [anon_sym_continue] = ACTIONS(3209),
    [anon_sym_state] = ACTIONS(3209),
    [anon_sym_our] = ACTIONS(3209),
    [anon_sym_LBRACK] = ACTIONS(3209),
    [anon_sym_LT_LT] = ACTIONS(3209),
    [anon_sym_SLASH_SLASH] = ACTIONS(3209),
    [anon_sym_PLUS] = ACTIONS(3209),
    [anon_sym_DASH] = ACTIONS(3209),
    [anon_sym_TILDE] = ACTIONS(3209),
    [anon_sym_BANG] = ACTIONS(3209),
    [anon_sym_PLUS_PLUS] = ACTIONS(3209),
    [anon_sym_DASH_DASH] = ACTIONS(3209),
    [anon_sym_BSLASH] = ACTIONS(3207),
    [anon_sym_do] = ACTIONS(3209),
    [anon_sym_eval] = ACTIONS(3209),
    [anon_sym_await] = ACTIONS(3209),
    [anon_sym_field] = ACTIONS(3209),
    [anon_sym_local] = ACTIONS(3209),
    [anon_sym_dynamically] = ACTIONS(3209),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3209),
    [anon_sym_require] = ACTIONS(3209),
    [anon_sym_map] = ACTIONS(3209),
    [anon_sym_grep] = ACTIONS(3209),
    [anon_sym_sort] = ACTIONS(3209),
    [anon_sym_goto] = ACTIONS(3209),
    [anon_sym_return] = ACTIONS(3209),
    [anon_sym_undef] = ACTIONS(3209),
    [anon_sym_PERCENT2] = ACTIONS(3209),
    [anon_sym_AMP2] = ACTIONS(3209),
    [anon_sym_STAR2] = ACTIONS(3209),
    [anon_sym_last] = ACTIONS(3209),
    [anon_sym_next] = ACTIONS(3209),
    [anon_sym_redo] = ACTIONS(3209),
    [anon_sym___FILE__] = ACTIONS(3209),
    [anon_sym___LINE__] = ACTIONS(3209),
    [anon_sym___PACKAGE__] = ACTIONS(3209),
    [anon_sym___SUB__] = ACTIONS(3209),
    [anon_sym_break] = ACTIONS(3209),
    [anon_sym_fork] = ACTIONS(3209),
    [anon_sym_getppid] = ACTIONS(3209),
    [anon_sym_time] = ACTIONS(3209),
    [anon_sym_times] = ACTIONS(3209),
    [anon_sym_wait] = ACTIONS(3209),
    [anon_sym_wantarray] = ACTIONS(3209),
    [anon_sym_abs] = ACTIONS(3209),
    [anon_sym_alarm] = ACTIONS(3209),
    [anon_sym_chop] = ACTIONS(3209),
    [anon_sym_chdir] = ACTIONS(3209),
    [anon_sym_close] = ACTIONS(3209),
    [anon_sym_closedir] = ACTIONS(3209),
    [anon_sym_caller] = ACTIONS(3209),
    [anon_sym_chomp] = ACTIONS(3209),
    [anon_sym_chr] = ACTIONS(3209),
    [anon_sym_cos] = ACTIONS(3209),
    [anon_sym_chroot] = ACTIONS(3209),
    [anon_sym_defined] = ACTIONS(3209),
    [anon_sym_delete] = ACTIONS(3209),
    [anon_sym_dbmclose] = ACTIONS(3209),
    [anon_sym_exists] = ACTIONS(3209),
    [anon_sym_exit] = ACTIONS(3209),
    [anon_sym_eof] = ACTIONS(3209),
    [anon_sym_exp] = ACTIONS(3209),
    [anon_sym_each] = ACTIONS(3209),
    [anon_sym_fc] = ACTIONS(3209),
    [anon_sym_fileno] = ACTIONS(3209),
    [anon_sym_gmtime] = ACTIONS(3209),
    [anon_sym_getc] = ACTIONS(3209),
    [anon_sym_getpgrp] = ACTIONS(3209),
    [anon_sym_getprotobyname] = ACTIONS(3209),
    [anon_sym_getpwname] = ACTIONS(3209),
    [anon_sym_getpwuid] = ACTIONS(3209),
    [anon_sym_getpeername] = ACTIONS(3209),
    [anon_sym_getnetbyname] = ACTIONS(3209),
    [anon_sym_getsockname] = ACTIONS(3209),
    [anon_sym_getgrnam] = ACTIONS(3209),
    [anon_sym_getgrgid] = ACTIONS(3209),
    [anon_sym_hex] = ACTIONS(3209),
    [anon_sym_int] = ACTIONS(3209),
    [anon_sym_keys] = ACTIONS(3209),
    [anon_sym_lc] = ACTIONS(3209),
    [anon_sym_lcfirst] = ACTIONS(3209),
    [anon_sym_length] = ACTIONS(3209),
    [anon_sym_localtime] = ACTIONS(3209),
    [anon_sym_log] = ACTIONS(3209),
    [anon_sym_lock] = ACTIONS(3209),
    [anon_sym_lstat] = ACTIONS(3209),
    [anon_sym_oct] = ACTIONS(3209),
    [anon_sym_ord] = ACTIONS(3209),
    [anon_sym_prototype] = ACTIONS(3209),
    [anon_sym_pop] = ACTIONS(3209),
    [anon_sym_pos] = ACTIONS(3209),
    [anon_sym_quotemeta] = ACTIONS(3209),
    [anon_sym_reset] = ACTIONS(3209),
    [anon_sym_rand] = ACTIONS(3209),
    [anon_sym_rmdir] = ACTIONS(3209),
    [anon_sym_readdir] = ACTIONS(3209),
    [anon_sym_readline] = ACTIONS(3209),
    [anon_sym_readpipe] = ACTIONS(3209),
    [anon_sym_rewinddir] = ACTIONS(3209),
    [anon_sym_readlink] = ACTIONS(3209),
    [anon_sym_ref] = ACTIONS(3209),
    [anon_sym_scalar] = ACTIONS(3209),
    [anon_sym_shift] = ACTIONS(3209),
    [anon_sym_sin] = ACTIONS(3209),
    [anon_sym_sleep] = ACTIONS(3209),
    [anon_sym_sqrt] = ACTIONS(3209),
    [anon_sym_srand] = ACTIONS(3209),
    [anon_sym_stat] = ACTIONS(3209),
    [anon_sym_study] = ACTIONS(3209),
    [anon_sym_tell] = ACTIONS(3209),
    [anon_sym_telldir] = ACTIONS(3209),
    [anon_sym_tied] = ACTIONS(3209),
    [anon_sym_uc] = ACTIONS(3209),
    [anon_sym_ucfirst] = ACTIONS(3209),
    [anon_sym_untie] = ACTIONS(3209),
    [anon_sym_umask] = ACTIONS(3209),
    [anon_sym_values] = ACTIONS(3209),
    [anon_sym_write] = ACTIONS(3209),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3209),
    [anon_sym_qq] = ACTIONS(3209),
    [anon_sym_qw] = ACTIONS(3209),
    [anon_sym_qx] = ACTIONS(3209),
    [anon_sym_qr] = ACTIONS(3209),
    [anon_sym_m] = ACTIONS(3209),
    [anon_sym_s] = ACTIONS(3209),
    [anon_sym_tr] = ACTIONS(3209),
    [anon_sym_y] = ACTIONS(3209),
    [aux_sym__bareword_token1] = ACTIONS(3209),
    [sym_number] = ACTIONS(3209),
    [anon_sym_true] = ACTIONS(3209),
    [anon_sym_false] = ACTIONS(3209),
    [sym__single_quote] = ACTIONS(3207),
    [sym__double_quote] = ACTIONS(3207),
    [sym__backtick_quote] = ACTIONS(3207),
    [sym__search_slash_quote] = ACTIONS(3207),
    [sym__open_readline_bracket] = ACTIONS(3207),
    [sym__open_fileglob_bracket] = ACTIONS(3207),
    [sym__PERLY_HEREDOC] = ACTIONS(3207),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3207),
    [sym__filetest] = ACTIONS(3207),
  },
  [STATE(1686)] = {
    [sym_heredoc_content] = STATE(1686),
    [sym__identifier] = ACTIONS(3213),
    [anon_sym_LBRACE] = ACTIONS(3213),
    [anon_sym_DOLLAR] = ACTIONS(3213),
    [anon_sym_AT] = ACTIONS(3213),
    [anon_sym_RPAREN] = ACTIONS(3213),
    [anon_sym_my] = ACTIONS(3213),
    [anon_sym_extended] = ACTIONS(3213),
    [anon_sym_async] = ACTIONS(3213),
    [anon_sym_sub] = ACTIONS(3213),
    [anon_sym_method] = ACTIONS(3213),
    [anon_sym_LPAREN] = ACTIONS(3213),
    [anon_sym_continue] = ACTIONS(3213),
    [anon_sym_state] = ACTIONS(3213),
    [anon_sym_our] = ACTIONS(3213),
    [anon_sym_LBRACK] = ACTIONS(3213),
    [anon_sym_LT_LT] = ACTIONS(3213),
    [anon_sym_SLASH_SLASH] = ACTIONS(3213),
    [anon_sym_PLUS] = ACTIONS(3213),
    [anon_sym_DASH] = ACTIONS(3213),
    [anon_sym_TILDE] = ACTIONS(3213),
    [anon_sym_BANG] = ACTIONS(3213),
    [anon_sym_PLUS_PLUS] = ACTIONS(3213),
    [anon_sym_DASH_DASH] = ACTIONS(3213),
    [anon_sym_BSLASH] = ACTIONS(3211),
    [anon_sym_do] = ACTIONS(3213),
    [anon_sym_eval] = ACTIONS(3213),
    [anon_sym_await] = ACTIONS(3213),
    [anon_sym_field] = ACTIONS(3213),
    [anon_sym_local] = ACTIONS(3213),
    [anon_sym_dynamically] = ACTIONS(3213),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3213),
    [anon_sym_require] = ACTIONS(3213),
    [anon_sym_map] = ACTIONS(3213),
    [anon_sym_grep] = ACTIONS(3213),
    [anon_sym_sort] = ACTIONS(3213),
    [anon_sym_goto] = ACTIONS(3213),
    [anon_sym_return] = ACTIONS(3213),
    [anon_sym_undef] = ACTIONS(3213),
    [anon_sym_PERCENT2] = ACTIONS(3213),
    [anon_sym_AMP2] = ACTIONS(3213),
    [anon_sym_STAR2] = ACTIONS(3213),
    [anon_sym_last] = ACTIONS(3213),
    [anon_sym_next] = ACTIONS(3213),
    [anon_sym_redo] = ACTIONS(3213),
    [anon_sym___FILE__] = ACTIONS(3213),
    [anon_sym___LINE__] = ACTIONS(3213),
    [anon_sym___PACKAGE__] = ACTIONS(3213),
    [anon_sym___SUB__] = ACTIONS(3213),
    [anon_sym_break] = ACTIONS(3213),
    [anon_sym_fork] = ACTIONS(3213),
    [anon_sym_getppid] = ACTIONS(3213),
    [anon_sym_time] = ACTIONS(3213),
    [anon_sym_times] = ACTIONS(3213),
    [anon_sym_wait] = ACTIONS(3213),
    [anon_sym_wantarray] = ACTIONS(3213),
    [anon_sym_abs] = ACTIONS(3213),
    [anon_sym_alarm] = ACTIONS(3213),
    [anon_sym_chop] = ACTIONS(3213),
    [anon_sym_chdir] = ACTIONS(3213),
    [anon_sym_close] = ACTIONS(3213),
    [anon_sym_closedir] = ACTIONS(3213),
    [anon_sym_caller] = ACTIONS(3213),
    [anon_sym_chomp] = ACTIONS(3213),
    [anon_sym_chr] = ACTIONS(3213),
    [anon_sym_cos] = ACTIONS(3213),
    [anon_sym_chroot] = ACTIONS(3213),
    [anon_sym_defined] = ACTIONS(3213),
    [anon_sym_delete] = ACTIONS(3213),
    [anon_sym_dbmclose] = ACTIONS(3213),
    [anon_sym_exists] = ACTIONS(3213),
    [anon_sym_exit] = ACTIONS(3213),
    [anon_sym_eof] = ACTIONS(3213),
    [anon_sym_exp] = ACTIONS(3213),
    [anon_sym_each] = ACTIONS(3213),
    [anon_sym_fc] = ACTIONS(3213),
    [anon_sym_fileno] = ACTIONS(3213),
    [anon_sym_gmtime] = ACTIONS(3213),
    [anon_sym_getc] = ACTIONS(3213),
    [anon_sym_getpgrp] = ACTIONS(3213),
    [anon_sym_getprotobyname] = ACTIONS(3213),
    [anon_sym_getpwname] = ACTIONS(3213),
    [anon_sym_getpwuid] = ACTIONS(3213),
    [anon_sym_getpeername] = ACTIONS(3213),
    [anon_sym_getnetbyname] = ACTIONS(3213),
    [anon_sym_getsockname] = ACTIONS(3213),
    [anon_sym_getgrnam] = ACTIONS(3213),
    [anon_sym_getgrgid] = ACTIONS(3213),
    [anon_sym_hex] = ACTIONS(3213),
    [anon_sym_int] = ACTIONS(3213),
    [anon_sym_keys] = ACTIONS(3213),
    [anon_sym_lc] = ACTIONS(3213),
    [anon_sym_lcfirst] = ACTIONS(3213),
    [anon_sym_length] = ACTIONS(3213),
    [anon_sym_localtime] = ACTIONS(3213),
    [anon_sym_log] = ACTIONS(3213),
    [anon_sym_lock] = ACTIONS(3213),
    [anon_sym_lstat] = ACTIONS(3213),
    [anon_sym_oct] = ACTIONS(3213),
    [anon_sym_ord] = ACTIONS(3213),
    [anon_sym_prototype] = ACTIONS(3213),
    [anon_sym_pop] = ACTIONS(3213),
    [anon_sym_pos] = ACTIONS(3213),
    [anon_sym_quotemeta] = ACTIONS(3213),
    [anon_sym_reset] = ACTIONS(3213),
    [anon_sym_rand] = ACTIONS(3213),
    [anon_sym_rmdir] = ACTIONS(3213),
    [anon_sym_readdir] = ACTIONS(3213),
    [anon_sym_readline] = ACTIONS(3213),
    [anon_sym_readpipe] = ACTIONS(3213),
    [anon_sym_rewinddir] = ACTIONS(3213),
    [anon_sym_readlink] = ACTIONS(3213),
    [anon_sym_ref] = ACTIONS(3213),
    [anon_sym_scalar] = ACTIONS(3213),
    [anon_sym_shift] = ACTIONS(3213),
    [anon_sym_sin] = ACTIONS(3213),
    [anon_sym_sleep] = ACTIONS(3213),
    [anon_sym_sqrt] = ACTIONS(3213),
    [anon_sym_srand] = ACTIONS(3213),
    [anon_sym_stat] = ACTIONS(3213),
    [anon_sym_study] = ACTIONS(3213),
    [anon_sym_tell] = ACTIONS(3213),
    [anon_sym_telldir] = ACTIONS(3213),
    [anon_sym_tied] = ACTIONS(3213),
    [anon_sym_uc] = ACTIONS(3213),
    [anon_sym_ucfirst] = ACTIONS(3213),
    [anon_sym_untie] = ACTIONS(3213),
    [anon_sym_umask] = ACTIONS(3213),
    [anon_sym_values] = ACTIONS(3213),
    [anon_sym_write] = ACTIONS(3213),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3213),
    [anon_sym_qq] = ACTIONS(3213),
    [anon_sym_qw] = ACTIONS(3213),
    [anon_sym_qx] = ACTIONS(3213),
    [anon_sym_qr] = ACTIONS(3213),
    [anon_sym_m] = ACTIONS(3213),
    [anon_sym_s] = ACTIONS(3213),
    [anon_sym_tr] = ACTIONS(3213),
    [anon_sym_y] = ACTIONS(3213),
    [aux_sym__bareword_token1] = ACTIONS(3213),
    [sym_number] = ACTIONS(3213),
    [anon_sym_true] = ACTIONS(3213),
    [anon_sym_false] = ACTIONS(3213),
    [sym__single_quote] = ACTIONS(3211),
    [sym__double_quote] = ACTIONS(3211),
    [sym__backtick_quote] = ACTIONS(3211),
    [sym__search_slash_quote] = ACTIONS(3211),
    [sym__open_readline_bracket] = ACTIONS(3211),
    [sym__open_fileglob_bracket] = ACTIONS(3211),
    [sym__PERLY_HEREDOC] = ACTIONS(3211),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3211),
    [sym__filetest] = ACTIONS(3211),
  },
  [STATE(1687)] = {
    [sym_heredoc_content] = STATE(1687),
    [sym__identifier] = ACTIONS(3217),
    [anon_sym_LBRACE] = ACTIONS(3217),
    [anon_sym_DOLLAR] = ACTIONS(3217),
    [anon_sym_AT] = ACTIONS(3217),
    [anon_sym_RPAREN] = ACTIONS(3217),
    [anon_sym_my] = ACTIONS(3217),
    [anon_sym_extended] = ACTIONS(3217),
    [anon_sym_async] = ACTIONS(3217),
    [anon_sym_sub] = ACTIONS(3217),
    [anon_sym_method] = ACTIONS(3217),
    [anon_sym_LPAREN] = ACTIONS(3217),
    [anon_sym_continue] = ACTIONS(3217),
    [anon_sym_state] = ACTIONS(3217),
    [anon_sym_our] = ACTIONS(3217),
    [anon_sym_LBRACK] = ACTIONS(3217),
    [anon_sym_LT_LT] = ACTIONS(3217),
    [anon_sym_SLASH_SLASH] = ACTIONS(3217),
    [anon_sym_PLUS] = ACTIONS(3217),
    [anon_sym_DASH] = ACTIONS(3217),
    [anon_sym_TILDE] = ACTIONS(3217),
    [anon_sym_BANG] = ACTIONS(3217),
    [anon_sym_PLUS_PLUS] = ACTIONS(3217),
    [anon_sym_DASH_DASH] = ACTIONS(3217),
    [anon_sym_BSLASH] = ACTIONS(3215),
    [anon_sym_do] = ACTIONS(3217),
    [anon_sym_eval] = ACTIONS(3217),
    [anon_sym_await] = ACTIONS(3217),
    [anon_sym_field] = ACTIONS(3217),
    [anon_sym_local] = ACTIONS(3217),
    [anon_sym_dynamically] = ACTIONS(3217),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3217),
    [anon_sym_require] = ACTIONS(3217),
    [anon_sym_map] = ACTIONS(3217),
    [anon_sym_grep] = ACTIONS(3217),
    [anon_sym_sort] = ACTIONS(3217),
    [anon_sym_goto] = ACTIONS(3217),
    [anon_sym_return] = ACTIONS(3217),
    [anon_sym_undef] = ACTIONS(3217),
    [anon_sym_PERCENT2] = ACTIONS(3217),
    [anon_sym_AMP2] = ACTIONS(3217),
    [anon_sym_STAR2] = ACTIONS(3217),
    [anon_sym_last] = ACTIONS(3217),
    [anon_sym_next] = ACTIONS(3217),
    [anon_sym_redo] = ACTIONS(3217),
    [anon_sym___FILE__] = ACTIONS(3217),
    [anon_sym___LINE__] = ACTIONS(3217),
    [anon_sym___PACKAGE__] = ACTIONS(3217),
    [anon_sym___SUB__] = ACTIONS(3217),
    [anon_sym_break] = ACTIONS(3217),
    [anon_sym_fork] = ACTIONS(3217),
    [anon_sym_getppid] = ACTIONS(3217),
    [anon_sym_time] = ACTIONS(3217),
    [anon_sym_times] = ACTIONS(3217),
    [anon_sym_wait] = ACTIONS(3217),
    [anon_sym_wantarray] = ACTIONS(3217),
    [anon_sym_abs] = ACTIONS(3217),
    [anon_sym_alarm] = ACTIONS(3217),
    [anon_sym_chop] = ACTIONS(3217),
    [anon_sym_chdir] = ACTIONS(3217),
    [anon_sym_close] = ACTIONS(3217),
    [anon_sym_closedir] = ACTIONS(3217),
    [anon_sym_caller] = ACTIONS(3217),
    [anon_sym_chomp] = ACTIONS(3217),
    [anon_sym_chr] = ACTIONS(3217),
    [anon_sym_cos] = ACTIONS(3217),
    [anon_sym_chroot] = ACTIONS(3217),
    [anon_sym_defined] = ACTIONS(3217),
    [anon_sym_delete] = ACTIONS(3217),
    [anon_sym_dbmclose] = ACTIONS(3217),
    [anon_sym_exists] = ACTIONS(3217),
    [anon_sym_exit] = ACTIONS(3217),
    [anon_sym_eof] = ACTIONS(3217),
    [anon_sym_exp] = ACTIONS(3217),
    [anon_sym_each] = ACTIONS(3217),
    [anon_sym_fc] = ACTIONS(3217),
    [anon_sym_fileno] = ACTIONS(3217),
    [anon_sym_gmtime] = ACTIONS(3217),
    [anon_sym_getc] = ACTIONS(3217),
    [anon_sym_getpgrp] = ACTIONS(3217),
    [anon_sym_getprotobyname] = ACTIONS(3217),
    [anon_sym_getpwname] = ACTIONS(3217),
    [anon_sym_getpwuid] = ACTIONS(3217),
    [anon_sym_getpeername] = ACTIONS(3217),
    [anon_sym_getnetbyname] = ACTIONS(3217),
    [anon_sym_getsockname] = ACTIONS(3217),
    [anon_sym_getgrnam] = ACTIONS(3217),
    [anon_sym_getgrgid] = ACTIONS(3217),
    [anon_sym_hex] = ACTIONS(3217),
    [anon_sym_int] = ACTIONS(3217),
    [anon_sym_keys] = ACTIONS(3217),
    [anon_sym_lc] = ACTIONS(3217),
    [anon_sym_lcfirst] = ACTIONS(3217),
    [anon_sym_length] = ACTIONS(3217),
    [anon_sym_localtime] = ACTIONS(3217),
    [anon_sym_log] = ACTIONS(3217),
    [anon_sym_lock] = ACTIONS(3217),
    [anon_sym_lstat] = ACTIONS(3217),
    [anon_sym_oct] = ACTIONS(3217),
    [anon_sym_ord] = ACTIONS(3217),
    [anon_sym_prototype] = ACTIONS(3217),
    [anon_sym_pop] = ACTIONS(3217),
    [anon_sym_pos] = ACTIONS(3217),
    [anon_sym_quotemeta] = ACTIONS(3217),
    [anon_sym_reset] = ACTIONS(3217),
    [anon_sym_rand] = ACTIONS(3217),
    [anon_sym_rmdir] = ACTIONS(3217),
    [anon_sym_readdir] = ACTIONS(3217),
    [anon_sym_readline] = ACTIONS(3217),
    [anon_sym_readpipe] = ACTIONS(3217),
    [anon_sym_rewinddir] = ACTIONS(3217),
    [anon_sym_readlink] = ACTIONS(3217),
    [anon_sym_ref] = ACTIONS(3217),
    [anon_sym_scalar] = ACTIONS(3217),
    [anon_sym_shift] = ACTIONS(3217),
    [anon_sym_sin] = ACTIONS(3217),
    [anon_sym_sleep] = ACTIONS(3217),
    [anon_sym_sqrt] = ACTIONS(3217),
    [anon_sym_srand] = ACTIONS(3217),
    [anon_sym_stat] = ACTIONS(3217),
    [anon_sym_study] = ACTIONS(3217),
    [anon_sym_tell] = ACTIONS(3217),
    [anon_sym_telldir] = ACTIONS(3217),
    [anon_sym_tied] = ACTIONS(3217),
    [anon_sym_uc] = ACTIONS(3217),
    [anon_sym_ucfirst] = ACTIONS(3217),
    [anon_sym_untie] = ACTIONS(3217),
    [anon_sym_umask] = ACTIONS(3217),
    [anon_sym_values] = ACTIONS(3217),
    [anon_sym_write] = ACTIONS(3217),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3217),
    [anon_sym_qq] = ACTIONS(3217),
    [anon_sym_qw] = ACTIONS(3217),
    [anon_sym_qx] = ACTIONS(3217),
    [anon_sym_qr] = ACTIONS(3217),
    [anon_sym_m] = ACTIONS(3217),
    [anon_sym_s] = ACTIONS(3217),
    [anon_sym_tr] = ACTIONS(3217),
    [anon_sym_y] = ACTIONS(3217),
    [aux_sym__bareword_token1] = ACTIONS(3217),
    [sym_number] = ACTIONS(3217),
    [anon_sym_true] = ACTIONS(3217),
    [anon_sym_false] = ACTIONS(3217),
    [sym__single_quote] = ACTIONS(3215),
    [sym__double_quote] = ACTIONS(3215),
    [sym__backtick_quote] = ACTIONS(3215),
    [sym__search_slash_quote] = ACTIONS(3215),
    [sym__open_readline_bracket] = ACTIONS(3215),
    [sym__open_fileglob_bracket] = ACTIONS(3215),
    [sym__PERLY_HEREDOC] = ACTIONS(3215),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3215),
    [sym__filetest] = ACTIONS(3215),
  },
  [STATE(1688)] = {
    [sym_heredoc_content] = STATE(1688),
    [sym__identifier] = ACTIONS(3029),
    [anon_sym_LBRACE] = ACTIONS(3029),
    [anon_sym_DOLLAR] = ACTIONS(3029),
    [anon_sym_AT] = ACTIONS(3029),
    [anon_sym_RPAREN] = ACTIONS(3029),
    [anon_sym_my] = ACTIONS(3029),
    [anon_sym_extended] = ACTIONS(3029),
    [anon_sym_async] = ACTIONS(3029),
    [anon_sym_sub] = ACTIONS(3029),
    [anon_sym_method] = ACTIONS(3029),
    [anon_sym_LPAREN] = ACTIONS(3029),
    [anon_sym_continue] = ACTIONS(3029),
    [anon_sym_state] = ACTIONS(3029),
    [anon_sym_our] = ACTIONS(3029),
    [anon_sym_LBRACK] = ACTIONS(3029),
    [anon_sym_LT_LT] = ACTIONS(3029),
    [anon_sym_SLASH_SLASH] = ACTIONS(3029),
    [anon_sym_PLUS] = ACTIONS(3029),
    [anon_sym_DASH] = ACTIONS(3029),
    [anon_sym_TILDE] = ACTIONS(3029),
    [anon_sym_BANG] = ACTIONS(3029),
    [anon_sym_PLUS_PLUS] = ACTIONS(3029),
    [anon_sym_DASH_DASH] = ACTIONS(3029),
    [anon_sym_BSLASH] = ACTIONS(3027),
    [anon_sym_do] = ACTIONS(3029),
    [anon_sym_eval] = ACTIONS(3029),
    [anon_sym_await] = ACTIONS(3029),
    [anon_sym_field] = ACTIONS(3029),
    [anon_sym_local] = ACTIONS(3029),
    [anon_sym_dynamically] = ACTIONS(3029),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3029),
    [anon_sym_require] = ACTIONS(3029),
    [anon_sym_map] = ACTIONS(3029),
    [anon_sym_grep] = ACTIONS(3029),
    [anon_sym_sort] = ACTIONS(3029),
    [anon_sym_goto] = ACTIONS(3029),
    [anon_sym_return] = ACTIONS(3029),
    [anon_sym_undef] = ACTIONS(3029),
    [anon_sym_PERCENT2] = ACTIONS(3029),
    [anon_sym_AMP2] = ACTIONS(3029),
    [anon_sym_STAR2] = ACTIONS(3029),
    [anon_sym_last] = ACTIONS(3029),
    [anon_sym_next] = ACTIONS(3029),
    [anon_sym_redo] = ACTIONS(3029),
    [anon_sym___FILE__] = ACTIONS(3029),
    [anon_sym___LINE__] = ACTIONS(3029),
    [anon_sym___PACKAGE__] = ACTIONS(3029),
    [anon_sym___SUB__] = ACTIONS(3029),
    [anon_sym_break] = ACTIONS(3029),
    [anon_sym_fork] = ACTIONS(3029),
    [anon_sym_getppid] = ACTIONS(3029),
    [anon_sym_time] = ACTIONS(3029),
    [anon_sym_times] = ACTIONS(3029),
    [anon_sym_wait] = ACTIONS(3029),
    [anon_sym_wantarray] = ACTIONS(3029),
    [anon_sym_abs] = ACTIONS(3029),
    [anon_sym_alarm] = ACTIONS(3029),
    [anon_sym_chop] = ACTIONS(3029),
    [anon_sym_chdir] = ACTIONS(3029),
    [anon_sym_close] = ACTIONS(3029),
    [anon_sym_closedir] = ACTIONS(3029),
    [anon_sym_caller] = ACTIONS(3029),
    [anon_sym_chomp] = ACTIONS(3029),
    [anon_sym_chr] = ACTIONS(3029),
    [anon_sym_cos] = ACTIONS(3029),
    [anon_sym_chroot] = ACTIONS(3029),
    [anon_sym_defined] = ACTIONS(3029),
    [anon_sym_delete] = ACTIONS(3029),
    [anon_sym_dbmclose] = ACTIONS(3029),
    [anon_sym_exists] = ACTIONS(3029),
    [anon_sym_exit] = ACTIONS(3029),
    [anon_sym_eof] = ACTIONS(3029),
    [anon_sym_exp] = ACTIONS(3029),
    [anon_sym_each] = ACTIONS(3029),
    [anon_sym_fc] = ACTIONS(3029),
    [anon_sym_fileno] = ACTIONS(3029),
    [anon_sym_gmtime] = ACTIONS(3029),
    [anon_sym_getc] = ACTIONS(3029),
    [anon_sym_getpgrp] = ACTIONS(3029),
    [anon_sym_getprotobyname] = ACTIONS(3029),
    [anon_sym_getpwname] = ACTIONS(3029),
    [anon_sym_getpwuid] = ACTIONS(3029),
    [anon_sym_getpeername] = ACTIONS(3029),
    [anon_sym_getnetbyname] = ACTIONS(3029),
    [anon_sym_getsockname] = ACTIONS(3029),
    [anon_sym_getgrnam] = ACTIONS(3029),
    [anon_sym_getgrgid] = ACTIONS(3029),
    [anon_sym_hex] = ACTIONS(3029),
    [anon_sym_int] = ACTIONS(3029),
    [anon_sym_keys] = ACTIONS(3029),
    [anon_sym_lc] = ACTIONS(3029),
    [anon_sym_lcfirst] = ACTIONS(3029),
    [anon_sym_length] = ACTIONS(3029),
    [anon_sym_localtime] = ACTIONS(3029),
    [anon_sym_log] = ACTIONS(3029),
    [anon_sym_lock] = ACTIONS(3029),
    [anon_sym_lstat] = ACTIONS(3029),
    [anon_sym_oct] = ACTIONS(3029),
    [anon_sym_ord] = ACTIONS(3029),
    [anon_sym_prototype] = ACTIONS(3029),
    [anon_sym_pop] = ACTIONS(3029),
    [anon_sym_pos] = ACTIONS(3029),
    [anon_sym_quotemeta] = ACTIONS(3029),
    [anon_sym_reset] = ACTIONS(3029),
    [anon_sym_rand] = ACTIONS(3029),
    [anon_sym_rmdir] = ACTIONS(3029),
    [anon_sym_readdir] = ACTIONS(3029),
    [anon_sym_readline] = ACTIONS(3029),
    [anon_sym_readpipe] = ACTIONS(3029),
    [anon_sym_rewinddir] = ACTIONS(3029),
    [anon_sym_readlink] = ACTIONS(3029),
    [anon_sym_ref] = ACTIONS(3029),
    [anon_sym_scalar] = ACTIONS(3029),
    [anon_sym_shift] = ACTIONS(3029),
    [anon_sym_sin] = ACTIONS(3029),
    [anon_sym_sleep] = ACTIONS(3029),
    [anon_sym_sqrt] = ACTIONS(3029),
    [anon_sym_srand] = ACTIONS(3029),
    [anon_sym_stat] = ACTIONS(3029),
    [anon_sym_study] = ACTIONS(3029),
    [anon_sym_tell] = ACTIONS(3029),
    [anon_sym_telldir] = ACTIONS(3029),
    [anon_sym_tied] = ACTIONS(3029),
    [anon_sym_uc] = ACTIONS(3029),
    [anon_sym_ucfirst] = ACTIONS(3029),
    [anon_sym_untie] = ACTIONS(3029),
    [anon_sym_umask] = ACTIONS(3029),
    [anon_sym_values] = ACTIONS(3029),
    [anon_sym_write] = ACTIONS(3029),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3029),
    [anon_sym_qq] = ACTIONS(3029),
    [anon_sym_qw] = ACTIONS(3029),
    [anon_sym_qx] = ACTIONS(3029),
    [anon_sym_qr] = ACTIONS(3029),
    [anon_sym_m] = ACTIONS(3029),
    [anon_sym_s] = ACTIONS(3029),
    [anon_sym_tr] = ACTIONS(3029),
    [anon_sym_y] = ACTIONS(3029),
    [aux_sym__bareword_token1] = ACTIONS(3029),
    [sym_number] = ACTIONS(3029),
    [anon_sym_true] = ACTIONS(3029),
    [anon_sym_false] = ACTIONS(3029),
    [sym__single_quote] = ACTIONS(3027),
    [sym__double_quote] = ACTIONS(3027),
    [sym__backtick_quote] = ACTIONS(3027),
    [sym__search_slash_quote] = ACTIONS(3027),
    [sym__open_readline_bracket] = ACTIONS(3027),
    [sym__open_fileglob_bracket] = ACTIONS(3027),
    [sym__PERLY_HEREDOC] = ACTIONS(3027),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3027),
    [sym__filetest] = ACTIONS(3027),
  },
  [STATE(1689)] = {
    [sym_heredoc_content] = STATE(1689),
    [sym__identifier] = ACTIONS(2910),
    [anon_sym_LBRACE] = ACTIONS(2910),
    [anon_sym_DOLLAR] = ACTIONS(2910),
    [anon_sym_AT] = ACTIONS(2910),
    [anon_sym_RPAREN] = ACTIONS(2910),
    [anon_sym_my] = ACTIONS(2910),
    [anon_sym_extended] = ACTIONS(2910),
    [anon_sym_async] = ACTIONS(2910),
    [anon_sym_sub] = ACTIONS(2910),
    [anon_sym_method] = ACTIONS(2910),
    [anon_sym_LPAREN] = ACTIONS(2910),
    [anon_sym_continue] = ACTIONS(2910),
    [anon_sym_state] = ACTIONS(2910),
    [anon_sym_our] = ACTIONS(2910),
    [anon_sym_LBRACK] = ACTIONS(2910),
    [anon_sym_LT_LT] = ACTIONS(2910),
    [anon_sym_SLASH_SLASH] = ACTIONS(2910),
    [anon_sym_PLUS] = ACTIONS(2910),
    [anon_sym_DASH] = ACTIONS(2910),
    [anon_sym_TILDE] = ACTIONS(2910),
    [anon_sym_BANG] = ACTIONS(2910),
    [anon_sym_PLUS_PLUS] = ACTIONS(2910),
    [anon_sym_DASH_DASH] = ACTIONS(2910),
    [anon_sym_BSLASH] = ACTIONS(2908),
    [anon_sym_do] = ACTIONS(2910),
    [anon_sym_eval] = ACTIONS(2910),
    [anon_sym_await] = ACTIONS(2910),
    [anon_sym_field] = ACTIONS(2910),
    [anon_sym_local] = ACTIONS(2910),
    [anon_sym_dynamically] = ACTIONS(2910),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2910),
    [anon_sym_require] = ACTIONS(2910),
    [anon_sym_map] = ACTIONS(2910),
    [anon_sym_grep] = ACTIONS(2910),
    [anon_sym_sort] = ACTIONS(2910),
    [anon_sym_goto] = ACTIONS(2910),
    [anon_sym_return] = ACTIONS(2910),
    [anon_sym_undef] = ACTIONS(2910),
    [anon_sym_PERCENT2] = ACTIONS(2910),
    [anon_sym_AMP2] = ACTIONS(2910),
    [anon_sym_STAR2] = ACTIONS(2910),
    [anon_sym_last] = ACTIONS(2910),
    [anon_sym_next] = ACTIONS(2910),
    [anon_sym_redo] = ACTIONS(2910),
    [anon_sym___FILE__] = ACTIONS(2910),
    [anon_sym___LINE__] = ACTIONS(2910),
    [anon_sym___PACKAGE__] = ACTIONS(2910),
    [anon_sym___SUB__] = ACTIONS(2910),
    [anon_sym_break] = ACTIONS(2910),
    [anon_sym_fork] = ACTIONS(2910),
    [anon_sym_getppid] = ACTIONS(2910),
    [anon_sym_time] = ACTIONS(2910),
    [anon_sym_times] = ACTIONS(2910),
    [anon_sym_wait] = ACTIONS(2910),
    [anon_sym_wantarray] = ACTIONS(2910),
    [anon_sym_abs] = ACTIONS(2910),
    [anon_sym_alarm] = ACTIONS(2910),
    [anon_sym_chop] = ACTIONS(2910),
    [anon_sym_chdir] = ACTIONS(2910),
    [anon_sym_close] = ACTIONS(2910),
    [anon_sym_closedir] = ACTIONS(2910),
    [anon_sym_caller] = ACTIONS(2910),
    [anon_sym_chomp] = ACTIONS(2910),
    [anon_sym_chr] = ACTIONS(2910),
    [anon_sym_cos] = ACTIONS(2910),
    [anon_sym_chroot] = ACTIONS(2910),
    [anon_sym_defined] = ACTIONS(2910),
    [anon_sym_delete] = ACTIONS(2910),
    [anon_sym_dbmclose] = ACTIONS(2910),
    [anon_sym_exists] = ACTIONS(2910),
    [anon_sym_exit] = ACTIONS(2910),
    [anon_sym_eof] = ACTIONS(2910),
    [anon_sym_exp] = ACTIONS(2910),
    [anon_sym_each] = ACTIONS(2910),
    [anon_sym_fc] = ACTIONS(2910),
    [anon_sym_fileno] = ACTIONS(2910),
    [anon_sym_gmtime] = ACTIONS(2910),
    [anon_sym_getc] = ACTIONS(2910),
    [anon_sym_getpgrp] = ACTIONS(2910),
    [anon_sym_getprotobyname] = ACTIONS(2910),
    [anon_sym_getpwname] = ACTIONS(2910),
    [anon_sym_getpwuid] = ACTIONS(2910),
    [anon_sym_getpeername] = ACTIONS(2910),
    [anon_sym_getnetbyname] = ACTIONS(2910),
    [anon_sym_getsockname] = ACTIONS(2910),
    [anon_sym_getgrnam] = ACTIONS(2910),
    [anon_sym_getgrgid] = ACTIONS(2910),
    [anon_sym_hex] = ACTIONS(2910),
    [anon_sym_int] = ACTIONS(2910),
    [anon_sym_keys] = ACTIONS(2910),
    [anon_sym_lc] = ACTIONS(2910),
    [anon_sym_lcfirst] = ACTIONS(2910),
    [anon_sym_length] = ACTIONS(2910),
    [anon_sym_localtime] = ACTIONS(2910),
    [anon_sym_log] = ACTIONS(2910),
    [anon_sym_lock] = ACTIONS(2910),
    [anon_sym_lstat] = ACTIONS(2910),
    [anon_sym_oct] = ACTIONS(2910),
    [anon_sym_ord] = ACTIONS(2910),
    [anon_sym_prototype] = ACTIONS(2910),
    [anon_sym_pop] = ACTIONS(2910),
    [anon_sym_pos] = ACTIONS(2910),
    [anon_sym_quotemeta] = ACTIONS(2910),
    [anon_sym_reset] = ACTIONS(2910),
    [anon_sym_rand] = ACTIONS(2910),
    [anon_sym_rmdir] = ACTIONS(2910),
    [anon_sym_readdir] = ACTIONS(2910),
    [anon_sym_readline] = ACTIONS(2910),
    [anon_sym_readpipe] = ACTIONS(2910),
    [anon_sym_rewinddir] = ACTIONS(2910),
    [anon_sym_readlink] = ACTIONS(2910),
    [anon_sym_ref] = ACTIONS(2910),
    [anon_sym_scalar] = ACTIONS(2910),
    [anon_sym_shift] = ACTIONS(2910),
    [anon_sym_sin] = ACTIONS(2910),
    [anon_sym_sleep] = ACTIONS(2910),
    [anon_sym_sqrt] = ACTIONS(2910),
    [anon_sym_srand] = ACTIONS(2910),
    [anon_sym_stat] = ACTIONS(2910),
    [anon_sym_study] = ACTIONS(2910),
    [anon_sym_tell] = ACTIONS(2910),
    [anon_sym_telldir] = ACTIONS(2910),
    [anon_sym_tied] = ACTIONS(2910),
    [anon_sym_uc] = ACTIONS(2910),
    [anon_sym_ucfirst] = ACTIONS(2910),
    [anon_sym_untie] = ACTIONS(2910),
    [anon_sym_umask] = ACTIONS(2910),
    [anon_sym_values] = ACTIONS(2910),
    [anon_sym_write] = ACTIONS(2910),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2910),
    [anon_sym_qq] = ACTIONS(2910),
    [anon_sym_qw] = ACTIONS(2910),
    [anon_sym_qx] = ACTIONS(2910),
    [anon_sym_qr] = ACTIONS(2910),
    [anon_sym_m] = ACTIONS(2910),
    [anon_sym_s] = ACTIONS(2910),
    [anon_sym_tr] = ACTIONS(2910),
    [anon_sym_y] = ACTIONS(2910),
    [aux_sym__bareword_token1] = ACTIONS(2910),
    [sym_number] = ACTIONS(2910),
    [anon_sym_true] = ACTIONS(2910),
    [anon_sym_false] = ACTIONS(2910),
    [sym__single_quote] = ACTIONS(2908),
    [sym__double_quote] = ACTIONS(2908),
    [sym__backtick_quote] = ACTIONS(2908),
    [sym__search_slash_quote] = ACTIONS(2908),
    [sym__open_readline_bracket] = ACTIONS(2908),
    [sym__open_fileglob_bracket] = ACTIONS(2908),
    [sym__PERLY_HEREDOC] = ACTIONS(2908),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2908),
    [sym__filetest] = ACTIONS(2908),
  },
  [STATE(1690)] = {
    [sym_heredoc_content] = STATE(1690),
    [sym__identifier] = ACTIONS(2958),
    [anon_sym_LBRACE] = ACTIONS(2958),
    [anon_sym_DOLLAR] = ACTIONS(2958),
    [anon_sym_AT] = ACTIONS(2958),
    [anon_sym_RPAREN] = ACTIONS(2958),
    [anon_sym_my] = ACTIONS(2958),
    [anon_sym_extended] = ACTIONS(2958),
    [anon_sym_async] = ACTIONS(2958),
    [anon_sym_sub] = ACTIONS(2958),
    [anon_sym_method] = ACTIONS(2958),
    [anon_sym_LPAREN] = ACTIONS(2958),
    [anon_sym_continue] = ACTIONS(2958),
    [anon_sym_state] = ACTIONS(2958),
    [anon_sym_our] = ACTIONS(2958),
    [anon_sym_LBRACK] = ACTIONS(2958),
    [anon_sym_LT_LT] = ACTIONS(2958),
    [anon_sym_SLASH_SLASH] = ACTIONS(2958),
    [anon_sym_PLUS] = ACTIONS(2958),
    [anon_sym_DASH] = ACTIONS(2958),
    [anon_sym_TILDE] = ACTIONS(2958),
    [anon_sym_BANG] = ACTIONS(2958),
    [anon_sym_PLUS_PLUS] = ACTIONS(2958),
    [anon_sym_DASH_DASH] = ACTIONS(2958),
    [anon_sym_BSLASH] = ACTIONS(2956),
    [anon_sym_do] = ACTIONS(2958),
    [anon_sym_eval] = ACTIONS(2958),
    [anon_sym_await] = ACTIONS(2958),
    [anon_sym_field] = ACTIONS(2958),
    [anon_sym_local] = ACTIONS(2958),
    [anon_sym_dynamically] = ACTIONS(2958),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2958),
    [anon_sym_require] = ACTIONS(2958),
    [anon_sym_map] = ACTIONS(2958),
    [anon_sym_grep] = ACTIONS(2958),
    [anon_sym_sort] = ACTIONS(2958),
    [anon_sym_goto] = ACTIONS(2958),
    [anon_sym_return] = ACTIONS(2958),
    [anon_sym_undef] = ACTIONS(2958),
    [anon_sym_PERCENT2] = ACTIONS(2958),
    [anon_sym_AMP2] = ACTIONS(2958),
    [anon_sym_STAR2] = ACTIONS(2958),
    [anon_sym_last] = ACTIONS(2958),
    [anon_sym_next] = ACTIONS(2958),
    [anon_sym_redo] = ACTIONS(2958),
    [anon_sym___FILE__] = ACTIONS(2958),
    [anon_sym___LINE__] = ACTIONS(2958),
    [anon_sym___PACKAGE__] = ACTIONS(2958),
    [anon_sym___SUB__] = ACTIONS(2958),
    [anon_sym_break] = ACTIONS(2958),
    [anon_sym_fork] = ACTIONS(2958),
    [anon_sym_getppid] = ACTIONS(2958),
    [anon_sym_time] = ACTIONS(2958),
    [anon_sym_times] = ACTIONS(2958),
    [anon_sym_wait] = ACTIONS(2958),
    [anon_sym_wantarray] = ACTIONS(2958),
    [anon_sym_abs] = ACTIONS(2958),
    [anon_sym_alarm] = ACTIONS(2958),
    [anon_sym_chop] = ACTIONS(2958),
    [anon_sym_chdir] = ACTIONS(2958),
    [anon_sym_close] = ACTIONS(2958),
    [anon_sym_closedir] = ACTIONS(2958),
    [anon_sym_caller] = ACTIONS(2958),
    [anon_sym_chomp] = ACTIONS(2958),
    [anon_sym_chr] = ACTIONS(2958),
    [anon_sym_cos] = ACTIONS(2958),
    [anon_sym_chroot] = ACTIONS(2958),
    [anon_sym_defined] = ACTIONS(2958),
    [anon_sym_delete] = ACTIONS(2958),
    [anon_sym_dbmclose] = ACTIONS(2958),
    [anon_sym_exists] = ACTIONS(2958),
    [anon_sym_exit] = ACTIONS(2958),
    [anon_sym_eof] = ACTIONS(2958),
    [anon_sym_exp] = ACTIONS(2958),
    [anon_sym_each] = ACTIONS(2958),
    [anon_sym_fc] = ACTIONS(2958),
    [anon_sym_fileno] = ACTIONS(2958),
    [anon_sym_gmtime] = ACTIONS(2958),
    [anon_sym_getc] = ACTIONS(2958),
    [anon_sym_getpgrp] = ACTIONS(2958),
    [anon_sym_getprotobyname] = ACTIONS(2958),
    [anon_sym_getpwname] = ACTIONS(2958),
    [anon_sym_getpwuid] = ACTIONS(2958),
    [anon_sym_getpeername] = ACTIONS(2958),
    [anon_sym_getnetbyname] = ACTIONS(2958),
    [anon_sym_getsockname] = ACTIONS(2958),
    [anon_sym_getgrnam] = ACTIONS(2958),
    [anon_sym_getgrgid] = ACTIONS(2958),
    [anon_sym_hex] = ACTIONS(2958),
    [anon_sym_int] = ACTIONS(2958),
    [anon_sym_keys] = ACTIONS(2958),
    [anon_sym_lc] = ACTIONS(2958),
    [anon_sym_lcfirst] = ACTIONS(2958),
    [anon_sym_length] = ACTIONS(2958),
    [anon_sym_localtime] = ACTIONS(2958),
    [anon_sym_log] = ACTIONS(2958),
    [anon_sym_lock] = ACTIONS(2958),
    [anon_sym_lstat] = ACTIONS(2958),
    [anon_sym_oct] = ACTIONS(2958),
    [anon_sym_ord] = ACTIONS(2958),
    [anon_sym_prototype] = ACTIONS(2958),
    [anon_sym_pop] = ACTIONS(2958),
    [anon_sym_pos] = ACTIONS(2958),
    [anon_sym_quotemeta] = ACTIONS(2958),
    [anon_sym_reset] = ACTIONS(2958),
    [anon_sym_rand] = ACTIONS(2958),
    [anon_sym_rmdir] = ACTIONS(2958),
    [anon_sym_readdir] = ACTIONS(2958),
    [anon_sym_readline] = ACTIONS(2958),
    [anon_sym_readpipe] = ACTIONS(2958),
    [anon_sym_rewinddir] = ACTIONS(2958),
    [anon_sym_readlink] = ACTIONS(2958),
    [anon_sym_ref] = ACTIONS(2958),
    [anon_sym_scalar] = ACTIONS(2958),
    [anon_sym_shift] = ACTIONS(2958),
    [anon_sym_sin] = ACTIONS(2958),
    [anon_sym_sleep] = ACTIONS(2958),
    [anon_sym_sqrt] = ACTIONS(2958),
    [anon_sym_srand] = ACTIONS(2958),
    [anon_sym_stat] = ACTIONS(2958),
    [anon_sym_study] = ACTIONS(2958),
    [anon_sym_tell] = ACTIONS(2958),
    [anon_sym_telldir] = ACTIONS(2958),
    [anon_sym_tied] = ACTIONS(2958),
    [anon_sym_uc] = ACTIONS(2958),
    [anon_sym_ucfirst] = ACTIONS(2958),
    [anon_sym_untie] = ACTIONS(2958),
    [anon_sym_umask] = ACTIONS(2958),
    [anon_sym_values] = ACTIONS(2958),
    [anon_sym_write] = ACTIONS(2958),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2958),
    [anon_sym_qq] = ACTIONS(2958),
    [anon_sym_qw] = ACTIONS(2958),
    [anon_sym_qx] = ACTIONS(2958),
    [anon_sym_qr] = ACTIONS(2958),
    [anon_sym_m] = ACTIONS(2958),
    [anon_sym_s] = ACTIONS(2958),
    [anon_sym_tr] = ACTIONS(2958),
    [anon_sym_y] = ACTIONS(2958),
    [aux_sym__bareword_token1] = ACTIONS(2958),
    [sym_number] = ACTIONS(2958),
    [anon_sym_true] = ACTIONS(2958),
    [anon_sym_false] = ACTIONS(2958),
    [sym__single_quote] = ACTIONS(2956),
    [sym__double_quote] = ACTIONS(2956),
    [sym__backtick_quote] = ACTIONS(2956),
    [sym__search_slash_quote] = ACTIONS(2956),
    [sym__open_readline_bracket] = ACTIONS(2956),
    [sym__open_fileglob_bracket] = ACTIONS(2956),
    [sym__PERLY_HEREDOC] = ACTIONS(2956),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2956),
    [sym__filetest] = ACTIONS(2956),
  },
  [STATE(1691)] = {
    [sym_heredoc_content] = STATE(1691),
    [sym__identifier] = ACTIONS(3049),
    [anon_sym_LBRACE] = ACTIONS(3049),
    [anon_sym_DOLLAR] = ACTIONS(3049),
    [anon_sym_AT] = ACTIONS(3049),
    [anon_sym_RPAREN] = ACTIONS(3049),
    [anon_sym_my] = ACTIONS(3049),
    [anon_sym_extended] = ACTIONS(3049),
    [anon_sym_async] = ACTIONS(3049),
    [anon_sym_sub] = ACTIONS(3049),
    [anon_sym_method] = ACTIONS(3049),
    [anon_sym_LPAREN] = ACTIONS(3049),
    [anon_sym_continue] = ACTIONS(3049),
    [anon_sym_state] = ACTIONS(3049),
    [anon_sym_our] = ACTIONS(3049),
    [anon_sym_LBRACK] = ACTIONS(3049),
    [anon_sym_LT_LT] = ACTIONS(3049),
    [anon_sym_SLASH_SLASH] = ACTIONS(3049),
    [anon_sym_PLUS] = ACTIONS(3049),
    [anon_sym_DASH] = ACTIONS(3049),
    [anon_sym_TILDE] = ACTIONS(3049),
    [anon_sym_BANG] = ACTIONS(3049),
    [anon_sym_PLUS_PLUS] = ACTIONS(3049),
    [anon_sym_DASH_DASH] = ACTIONS(3049),
    [anon_sym_BSLASH] = ACTIONS(3047),
    [anon_sym_do] = ACTIONS(3049),
    [anon_sym_eval] = ACTIONS(3049),
    [anon_sym_await] = ACTIONS(3049),
    [anon_sym_field] = ACTIONS(3049),
    [anon_sym_local] = ACTIONS(3049),
    [anon_sym_dynamically] = ACTIONS(3049),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3049),
    [anon_sym_require] = ACTIONS(3049),
    [anon_sym_map] = ACTIONS(3049),
    [anon_sym_grep] = ACTIONS(3049),
    [anon_sym_sort] = ACTIONS(3049),
    [anon_sym_goto] = ACTIONS(3049),
    [anon_sym_return] = ACTIONS(3049),
    [anon_sym_undef] = ACTIONS(3049),
    [anon_sym_PERCENT2] = ACTIONS(3049),
    [anon_sym_AMP2] = ACTIONS(3049),
    [anon_sym_STAR2] = ACTIONS(3049),
    [anon_sym_last] = ACTIONS(3049),
    [anon_sym_next] = ACTIONS(3049),
    [anon_sym_redo] = ACTIONS(3049),
    [anon_sym___FILE__] = ACTIONS(3049),
    [anon_sym___LINE__] = ACTIONS(3049),
    [anon_sym___PACKAGE__] = ACTIONS(3049),
    [anon_sym___SUB__] = ACTIONS(3049),
    [anon_sym_break] = ACTIONS(3049),
    [anon_sym_fork] = ACTIONS(3049),
    [anon_sym_getppid] = ACTIONS(3049),
    [anon_sym_time] = ACTIONS(3049),
    [anon_sym_times] = ACTIONS(3049),
    [anon_sym_wait] = ACTIONS(3049),
    [anon_sym_wantarray] = ACTIONS(3049),
    [anon_sym_abs] = ACTIONS(3049),
    [anon_sym_alarm] = ACTIONS(3049),
    [anon_sym_chop] = ACTIONS(3049),
    [anon_sym_chdir] = ACTIONS(3049),
    [anon_sym_close] = ACTIONS(3049),
    [anon_sym_closedir] = ACTIONS(3049),
    [anon_sym_caller] = ACTIONS(3049),
    [anon_sym_chomp] = ACTIONS(3049),
    [anon_sym_chr] = ACTIONS(3049),
    [anon_sym_cos] = ACTIONS(3049),
    [anon_sym_chroot] = ACTIONS(3049),
    [anon_sym_defined] = ACTIONS(3049),
    [anon_sym_delete] = ACTIONS(3049),
    [anon_sym_dbmclose] = ACTIONS(3049),
    [anon_sym_exists] = ACTIONS(3049),
    [anon_sym_exit] = ACTIONS(3049),
    [anon_sym_eof] = ACTIONS(3049),
    [anon_sym_exp] = ACTIONS(3049),
    [anon_sym_each] = ACTIONS(3049),
    [anon_sym_fc] = ACTIONS(3049),
    [anon_sym_fileno] = ACTIONS(3049),
    [anon_sym_gmtime] = ACTIONS(3049),
    [anon_sym_getc] = ACTIONS(3049),
    [anon_sym_getpgrp] = ACTIONS(3049),
    [anon_sym_getprotobyname] = ACTIONS(3049),
    [anon_sym_getpwname] = ACTIONS(3049),
    [anon_sym_getpwuid] = ACTIONS(3049),
    [anon_sym_getpeername] = ACTIONS(3049),
    [anon_sym_getnetbyname] = ACTIONS(3049),
    [anon_sym_getsockname] = ACTIONS(3049),
    [anon_sym_getgrnam] = ACTIONS(3049),
    [anon_sym_getgrgid] = ACTIONS(3049),
    [anon_sym_hex] = ACTIONS(3049),
    [anon_sym_int] = ACTIONS(3049),
    [anon_sym_keys] = ACTIONS(3049),
    [anon_sym_lc] = ACTIONS(3049),
    [anon_sym_lcfirst] = ACTIONS(3049),
    [anon_sym_length] = ACTIONS(3049),
    [anon_sym_localtime] = ACTIONS(3049),
    [anon_sym_log] = ACTIONS(3049),
    [anon_sym_lock] = ACTIONS(3049),
    [anon_sym_lstat] = ACTIONS(3049),
    [anon_sym_oct] = ACTIONS(3049),
    [anon_sym_ord] = ACTIONS(3049),
    [anon_sym_prototype] = ACTIONS(3049),
    [anon_sym_pop] = ACTIONS(3049),
    [anon_sym_pos] = ACTIONS(3049),
    [anon_sym_quotemeta] = ACTIONS(3049),
    [anon_sym_reset] = ACTIONS(3049),
    [anon_sym_rand] = ACTIONS(3049),
    [anon_sym_rmdir] = ACTIONS(3049),
    [anon_sym_readdir] = ACTIONS(3049),
    [anon_sym_readline] = ACTIONS(3049),
    [anon_sym_readpipe] = ACTIONS(3049),
    [anon_sym_rewinddir] = ACTIONS(3049),
    [anon_sym_readlink] = ACTIONS(3049),
    [anon_sym_ref] = ACTIONS(3049),
    [anon_sym_scalar] = ACTIONS(3049),
    [anon_sym_shift] = ACTIONS(3049),
    [anon_sym_sin] = ACTIONS(3049),
    [anon_sym_sleep] = ACTIONS(3049),
    [anon_sym_sqrt] = ACTIONS(3049),
    [anon_sym_srand] = ACTIONS(3049),
    [anon_sym_stat] = ACTIONS(3049),
    [anon_sym_study] = ACTIONS(3049),
    [anon_sym_tell] = ACTIONS(3049),
    [anon_sym_telldir] = ACTIONS(3049),
    [anon_sym_tied] = ACTIONS(3049),
    [anon_sym_uc] = ACTIONS(3049),
    [anon_sym_ucfirst] = ACTIONS(3049),
    [anon_sym_untie] = ACTIONS(3049),
    [anon_sym_umask] = ACTIONS(3049),
    [anon_sym_values] = ACTIONS(3049),
    [anon_sym_write] = ACTIONS(3049),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3049),
    [anon_sym_qq] = ACTIONS(3049),
    [anon_sym_qw] = ACTIONS(3049),
    [anon_sym_qx] = ACTIONS(3049),
    [anon_sym_qr] = ACTIONS(3049),
    [anon_sym_m] = ACTIONS(3049),
    [anon_sym_s] = ACTIONS(3049),
    [anon_sym_tr] = ACTIONS(3049),
    [anon_sym_y] = ACTIONS(3049),
    [aux_sym__bareword_token1] = ACTIONS(3049),
    [sym_number] = ACTIONS(3049),
    [anon_sym_true] = ACTIONS(3049),
    [anon_sym_false] = ACTIONS(3049),
    [sym__single_quote] = ACTIONS(3047),
    [sym__double_quote] = ACTIONS(3047),
    [sym__backtick_quote] = ACTIONS(3047),
    [sym__search_slash_quote] = ACTIONS(3047),
    [sym__open_readline_bracket] = ACTIONS(3047),
    [sym__open_fileglob_bracket] = ACTIONS(3047),
    [sym__PERLY_HEREDOC] = ACTIONS(3047),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3047),
    [sym__filetest] = ACTIONS(3047),
  },
  [STATE(1692)] = {
    [sym_heredoc_content] = STATE(1692),
    [sym__identifier] = ACTIONS(3069),
    [anon_sym_LBRACE] = ACTIONS(3069),
    [anon_sym_DOLLAR] = ACTIONS(3069),
    [anon_sym_AT] = ACTIONS(3069),
    [anon_sym_RPAREN] = ACTIONS(3069),
    [anon_sym_my] = ACTIONS(3069),
    [anon_sym_extended] = ACTIONS(3069),
    [anon_sym_async] = ACTIONS(3069),
    [anon_sym_sub] = ACTIONS(3069),
    [anon_sym_method] = ACTIONS(3069),
    [anon_sym_LPAREN] = ACTIONS(3069),
    [anon_sym_continue] = ACTIONS(3069),
    [anon_sym_state] = ACTIONS(3069),
    [anon_sym_our] = ACTIONS(3069),
    [anon_sym_LBRACK] = ACTIONS(3069),
    [anon_sym_LT_LT] = ACTIONS(3069),
    [anon_sym_SLASH_SLASH] = ACTIONS(3069),
    [anon_sym_PLUS] = ACTIONS(3069),
    [anon_sym_DASH] = ACTIONS(3069),
    [anon_sym_TILDE] = ACTIONS(3069),
    [anon_sym_BANG] = ACTIONS(3069),
    [anon_sym_PLUS_PLUS] = ACTIONS(3069),
    [anon_sym_DASH_DASH] = ACTIONS(3069),
    [anon_sym_BSLASH] = ACTIONS(3067),
    [anon_sym_do] = ACTIONS(3069),
    [anon_sym_eval] = ACTIONS(3069),
    [anon_sym_await] = ACTIONS(3069),
    [anon_sym_field] = ACTIONS(3069),
    [anon_sym_local] = ACTIONS(3069),
    [anon_sym_dynamically] = ACTIONS(3069),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3069),
    [anon_sym_require] = ACTIONS(3069),
    [anon_sym_map] = ACTIONS(3069),
    [anon_sym_grep] = ACTIONS(3069),
    [anon_sym_sort] = ACTIONS(3069),
    [anon_sym_goto] = ACTIONS(3069),
    [anon_sym_return] = ACTIONS(3069),
    [anon_sym_undef] = ACTIONS(3069),
    [anon_sym_PERCENT2] = ACTIONS(3069),
    [anon_sym_AMP2] = ACTIONS(3069),
    [anon_sym_STAR2] = ACTIONS(3069),
    [anon_sym_last] = ACTIONS(3069),
    [anon_sym_next] = ACTIONS(3069),
    [anon_sym_redo] = ACTIONS(3069),
    [anon_sym___FILE__] = ACTIONS(3069),
    [anon_sym___LINE__] = ACTIONS(3069),
    [anon_sym___PACKAGE__] = ACTIONS(3069),
    [anon_sym___SUB__] = ACTIONS(3069),
    [anon_sym_break] = ACTIONS(3069),
    [anon_sym_fork] = ACTIONS(3069),
    [anon_sym_getppid] = ACTIONS(3069),
    [anon_sym_time] = ACTIONS(3069),
    [anon_sym_times] = ACTIONS(3069),
    [anon_sym_wait] = ACTIONS(3069),
    [anon_sym_wantarray] = ACTIONS(3069),
    [anon_sym_abs] = ACTIONS(3069),
    [anon_sym_alarm] = ACTIONS(3069),
    [anon_sym_chop] = ACTIONS(3069),
    [anon_sym_chdir] = ACTIONS(3069),
    [anon_sym_close] = ACTIONS(3069),
    [anon_sym_closedir] = ACTIONS(3069),
    [anon_sym_caller] = ACTIONS(3069),
    [anon_sym_chomp] = ACTIONS(3069),
    [anon_sym_chr] = ACTIONS(3069),
    [anon_sym_cos] = ACTIONS(3069),
    [anon_sym_chroot] = ACTIONS(3069),
    [anon_sym_defined] = ACTIONS(3069),
    [anon_sym_delete] = ACTIONS(3069),
    [anon_sym_dbmclose] = ACTIONS(3069),
    [anon_sym_exists] = ACTIONS(3069),
    [anon_sym_exit] = ACTIONS(3069),
    [anon_sym_eof] = ACTIONS(3069),
    [anon_sym_exp] = ACTIONS(3069),
    [anon_sym_each] = ACTIONS(3069),
    [anon_sym_fc] = ACTIONS(3069),
    [anon_sym_fileno] = ACTIONS(3069),
    [anon_sym_gmtime] = ACTIONS(3069),
    [anon_sym_getc] = ACTIONS(3069),
    [anon_sym_getpgrp] = ACTIONS(3069),
    [anon_sym_getprotobyname] = ACTIONS(3069),
    [anon_sym_getpwname] = ACTIONS(3069),
    [anon_sym_getpwuid] = ACTIONS(3069),
    [anon_sym_getpeername] = ACTIONS(3069),
    [anon_sym_getnetbyname] = ACTIONS(3069),
    [anon_sym_getsockname] = ACTIONS(3069),
    [anon_sym_getgrnam] = ACTIONS(3069),
    [anon_sym_getgrgid] = ACTIONS(3069),
    [anon_sym_hex] = ACTIONS(3069),
    [anon_sym_int] = ACTIONS(3069),
    [anon_sym_keys] = ACTIONS(3069),
    [anon_sym_lc] = ACTIONS(3069),
    [anon_sym_lcfirst] = ACTIONS(3069),
    [anon_sym_length] = ACTIONS(3069),
    [anon_sym_localtime] = ACTIONS(3069),
    [anon_sym_log] = ACTIONS(3069),
    [anon_sym_lock] = ACTIONS(3069),
    [anon_sym_lstat] = ACTIONS(3069),
    [anon_sym_oct] = ACTIONS(3069),
    [anon_sym_ord] = ACTIONS(3069),
    [anon_sym_prototype] = ACTIONS(3069),
    [anon_sym_pop] = ACTIONS(3069),
    [anon_sym_pos] = ACTIONS(3069),
    [anon_sym_quotemeta] = ACTIONS(3069),
    [anon_sym_reset] = ACTIONS(3069),
    [anon_sym_rand] = ACTIONS(3069),
    [anon_sym_rmdir] = ACTIONS(3069),
    [anon_sym_readdir] = ACTIONS(3069),
    [anon_sym_readline] = ACTIONS(3069),
    [anon_sym_readpipe] = ACTIONS(3069),
    [anon_sym_rewinddir] = ACTIONS(3069),
    [anon_sym_readlink] = ACTIONS(3069),
    [anon_sym_ref] = ACTIONS(3069),
    [anon_sym_scalar] = ACTIONS(3069),
    [anon_sym_shift] = ACTIONS(3069),
    [anon_sym_sin] = ACTIONS(3069),
    [anon_sym_sleep] = ACTIONS(3069),
    [anon_sym_sqrt] = ACTIONS(3069),
    [anon_sym_srand] = ACTIONS(3069),
    [anon_sym_stat] = ACTIONS(3069),
    [anon_sym_study] = ACTIONS(3069),
    [anon_sym_tell] = ACTIONS(3069),
    [anon_sym_telldir] = ACTIONS(3069),
    [anon_sym_tied] = ACTIONS(3069),
    [anon_sym_uc] = ACTIONS(3069),
    [anon_sym_ucfirst] = ACTIONS(3069),
    [anon_sym_untie] = ACTIONS(3069),
    [anon_sym_umask] = ACTIONS(3069),
    [anon_sym_values] = ACTIONS(3069),
    [anon_sym_write] = ACTIONS(3069),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3069),
    [anon_sym_qq] = ACTIONS(3069),
    [anon_sym_qw] = ACTIONS(3069),
    [anon_sym_qx] = ACTIONS(3069),
    [anon_sym_qr] = ACTIONS(3069),
    [anon_sym_m] = ACTIONS(3069),
    [anon_sym_s] = ACTIONS(3069),
    [anon_sym_tr] = ACTIONS(3069),
    [anon_sym_y] = ACTIONS(3069),
    [aux_sym__bareword_token1] = ACTIONS(3069),
    [sym_number] = ACTIONS(3069),
    [anon_sym_true] = ACTIONS(3069),
    [anon_sym_false] = ACTIONS(3069),
    [sym__single_quote] = ACTIONS(3067),
    [sym__double_quote] = ACTIONS(3067),
    [sym__backtick_quote] = ACTIONS(3067),
    [sym__search_slash_quote] = ACTIONS(3067),
    [sym__open_readline_bracket] = ACTIONS(3067),
    [sym__open_fileglob_bracket] = ACTIONS(3067),
    [sym__PERLY_HEREDOC] = ACTIONS(3067),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3067),
    [sym__filetest] = ACTIONS(3067),
  },
  [STATE(1693)] = {
    [sym_heredoc_content] = STATE(1693),
    [sym__identifier] = ACTIONS(3253),
    [anon_sym_LBRACE] = ACTIONS(3253),
    [anon_sym_DOLLAR] = ACTIONS(3253),
    [anon_sym_AT] = ACTIONS(3253),
    [anon_sym_RPAREN] = ACTIONS(3253),
    [anon_sym_my] = ACTIONS(3253),
    [anon_sym_extended] = ACTIONS(3253),
    [anon_sym_async] = ACTIONS(3253),
    [anon_sym_sub] = ACTIONS(3253),
    [anon_sym_method] = ACTIONS(3253),
    [anon_sym_LPAREN] = ACTIONS(3253),
    [anon_sym_continue] = ACTIONS(3253),
    [anon_sym_state] = ACTIONS(3253),
    [anon_sym_our] = ACTIONS(3253),
    [anon_sym_LBRACK] = ACTIONS(3253),
    [anon_sym_LT_LT] = ACTIONS(3253),
    [anon_sym_SLASH_SLASH] = ACTIONS(3253),
    [anon_sym_PLUS] = ACTIONS(3253),
    [anon_sym_DASH] = ACTIONS(3253),
    [anon_sym_TILDE] = ACTIONS(3253),
    [anon_sym_BANG] = ACTIONS(3253),
    [anon_sym_PLUS_PLUS] = ACTIONS(3253),
    [anon_sym_DASH_DASH] = ACTIONS(3253),
    [anon_sym_BSLASH] = ACTIONS(3251),
    [anon_sym_do] = ACTIONS(3253),
    [anon_sym_eval] = ACTIONS(3253),
    [anon_sym_await] = ACTIONS(3253),
    [anon_sym_field] = ACTIONS(3253),
    [anon_sym_local] = ACTIONS(3253),
    [anon_sym_dynamically] = ACTIONS(3253),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3253),
    [anon_sym_require] = ACTIONS(3253),
    [anon_sym_map] = ACTIONS(3253),
    [anon_sym_grep] = ACTIONS(3253),
    [anon_sym_sort] = ACTIONS(3253),
    [anon_sym_goto] = ACTIONS(3253),
    [anon_sym_return] = ACTIONS(3253),
    [anon_sym_undef] = ACTIONS(3253),
    [anon_sym_PERCENT2] = ACTIONS(3253),
    [anon_sym_AMP2] = ACTIONS(3253),
    [anon_sym_STAR2] = ACTIONS(3253),
    [anon_sym_last] = ACTIONS(3253),
    [anon_sym_next] = ACTIONS(3253),
    [anon_sym_redo] = ACTIONS(3253),
    [anon_sym___FILE__] = ACTIONS(3253),
    [anon_sym___LINE__] = ACTIONS(3253),
    [anon_sym___PACKAGE__] = ACTIONS(3253),
    [anon_sym___SUB__] = ACTIONS(3253),
    [anon_sym_break] = ACTIONS(3253),
    [anon_sym_fork] = ACTIONS(3253),
    [anon_sym_getppid] = ACTIONS(3253),
    [anon_sym_time] = ACTIONS(3253),
    [anon_sym_times] = ACTIONS(3253),
    [anon_sym_wait] = ACTIONS(3253),
    [anon_sym_wantarray] = ACTIONS(3253),
    [anon_sym_abs] = ACTIONS(3253),
    [anon_sym_alarm] = ACTIONS(3253),
    [anon_sym_chop] = ACTIONS(3253),
    [anon_sym_chdir] = ACTIONS(3253),
    [anon_sym_close] = ACTIONS(3253),
    [anon_sym_closedir] = ACTIONS(3253),
    [anon_sym_caller] = ACTIONS(3253),
    [anon_sym_chomp] = ACTIONS(3253),
    [anon_sym_chr] = ACTIONS(3253),
    [anon_sym_cos] = ACTIONS(3253),
    [anon_sym_chroot] = ACTIONS(3253),
    [anon_sym_defined] = ACTIONS(3253),
    [anon_sym_delete] = ACTIONS(3253),
    [anon_sym_dbmclose] = ACTIONS(3253),
    [anon_sym_exists] = ACTIONS(3253),
    [anon_sym_exit] = ACTIONS(3253),
    [anon_sym_eof] = ACTIONS(3253),
    [anon_sym_exp] = ACTIONS(3253),
    [anon_sym_each] = ACTIONS(3253),
    [anon_sym_fc] = ACTIONS(3253),
    [anon_sym_fileno] = ACTIONS(3253),
    [anon_sym_gmtime] = ACTIONS(3253),
    [anon_sym_getc] = ACTIONS(3253),
    [anon_sym_getpgrp] = ACTIONS(3253),
    [anon_sym_getprotobyname] = ACTIONS(3253),
    [anon_sym_getpwname] = ACTIONS(3253),
    [anon_sym_getpwuid] = ACTIONS(3253),
    [anon_sym_getpeername] = ACTIONS(3253),
    [anon_sym_getnetbyname] = ACTIONS(3253),
    [anon_sym_getsockname] = ACTIONS(3253),
    [anon_sym_getgrnam] = ACTIONS(3253),
    [anon_sym_getgrgid] = ACTIONS(3253),
    [anon_sym_hex] = ACTIONS(3253),
    [anon_sym_int] = ACTIONS(3253),
    [anon_sym_keys] = ACTIONS(3253),
    [anon_sym_lc] = ACTIONS(3253),
    [anon_sym_lcfirst] = ACTIONS(3253),
    [anon_sym_length] = ACTIONS(3253),
    [anon_sym_localtime] = ACTIONS(3253),
    [anon_sym_log] = ACTIONS(3253),
    [anon_sym_lock] = ACTIONS(3253),
    [anon_sym_lstat] = ACTIONS(3253),
    [anon_sym_oct] = ACTIONS(3253),
    [anon_sym_ord] = ACTIONS(3253),
    [anon_sym_prototype] = ACTIONS(3253),
    [anon_sym_pop] = ACTIONS(3253),
    [anon_sym_pos] = ACTIONS(3253),
    [anon_sym_quotemeta] = ACTIONS(3253),
    [anon_sym_reset] = ACTIONS(3253),
    [anon_sym_rand] = ACTIONS(3253),
    [anon_sym_rmdir] = ACTIONS(3253),
    [anon_sym_readdir] = ACTIONS(3253),
    [anon_sym_readline] = ACTIONS(3253),
    [anon_sym_readpipe] = ACTIONS(3253),
    [anon_sym_rewinddir] = ACTIONS(3253),
    [anon_sym_readlink] = ACTIONS(3253),
    [anon_sym_ref] = ACTIONS(3253),
    [anon_sym_scalar] = ACTIONS(3253),
    [anon_sym_shift] = ACTIONS(3253),
    [anon_sym_sin] = ACTIONS(3253),
    [anon_sym_sleep] = ACTIONS(3253),
    [anon_sym_sqrt] = ACTIONS(3253),
    [anon_sym_srand] = ACTIONS(3253),
    [anon_sym_stat] = ACTIONS(3253),
    [anon_sym_study] = ACTIONS(3253),
    [anon_sym_tell] = ACTIONS(3253),
    [anon_sym_telldir] = ACTIONS(3253),
    [anon_sym_tied] = ACTIONS(3253),
    [anon_sym_uc] = ACTIONS(3253),
    [anon_sym_ucfirst] = ACTIONS(3253),
    [anon_sym_untie] = ACTIONS(3253),
    [anon_sym_umask] = ACTIONS(3253),
    [anon_sym_values] = ACTIONS(3253),
    [anon_sym_write] = ACTIONS(3253),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3253),
    [anon_sym_qq] = ACTIONS(3253),
    [anon_sym_qw] = ACTIONS(3253),
    [anon_sym_qx] = ACTIONS(3253),
    [anon_sym_qr] = ACTIONS(3253),
    [anon_sym_m] = ACTIONS(3253),
    [anon_sym_s] = ACTIONS(3253),
    [anon_sym_tr] = ACTIONS(3253),
    [anon_sym_y] = ACTIONS(3253),
    [aux_sym__bareword_token1] = ACTIONS(3253),
    [sym_number] = ACTIONS(3253),
    [anon_sym_true] = ACTIONS(3253),
    [anon_sym_false] = ACTIONS(3253),
    [sym__single_quote] = ACTIONS(3251),
    [sym__double_quote] = ACTIONS(3251),
    [sym__backtick_quote] = ACTIONS(3251),
    [sym__search_slash_quote] = ACTIONS(3251),
    [sym__open_readline_bracket] = ACTIONS(3251),
    [sym__open_fileglob_bracket] = ACTIONS(3251),
    [sym__PERLY_HEREDOC] = ACTIONS(3251),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3251),
    [sym__filetest] = ACTIONS(3251),
  },
  [STATE(1694)] = {
    [sym_heredoc_content] = STATE(1694),
    [sym__identifier] = ACTIONS(3113),
    [anon_sym_LBRACE] = ACTIONS(3113),
    [anon_sym_DOLLAR] = ACTIONS(3113),
    [anon_sym_AT] = ACTIONS(3113),
    [anon_sym_RPAREN] = ACTIONS(3113),
    [anon_sym_my] = ACTIONS(3113),
    [anon_sym_extended] = ACTIONS(3113),
    [anon_sym_async] = ACTIONS(3113),
    [anon_sym_sub] = ACTIONS(3113),
    [anon_sym_method] = ACTIONS(3113),
    [anon_sym_LPAREN] = ACTIONS(3113),
    [anon_sym_continue] = ACTIONS(3113),
    [anon_sym_state] = ACTIONS(3113),
    [anon_sym_our] = ACTIONS(3113),
    [anon_sym_LBRACK] = ACTIONS(3113),
    [anon_sym_LT_LT] = ACTIONS(3113),
    [anon_sym_SLASH_SLASH] = ACTIONS(3113),
    [anon_sym_PLUS] = ACTIONS(3113),
    [anon_sym_DASH] = ACTIONS(3113),
    [anon_sym_TILDE] = ACTIONS(3113),
    [anon_sym_BANG] = ACTIONS(3113),
    [anon_sym_PLUS_PLUS] = ACTIONS(3113),
    [anon_sym_DASH_DASH] = ACTIONS(3113),
    [anon_sym_BSLASH] = ACTIONS(3111),
    [anon_sym_do] = ACTIONS(3113),
    [anon_sym_eval] = ACTIONS(3113),
    [anon_sym_await] = ACTIONS(3113),
    [anon_sym_field] = ACTIONS(3113),
    [anon_sym_local] = ACTIONS(3113),
    [anon_sym_dynamically] = ACTIONS(3113),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3113),
    [anon_sym_require] = ACTIONS(3113),
    [anon_sym_map] = ACTIONS(3113),
    [anon_sym_grep] = ACTIONS(3113),
    [anon_sym_sort] = ACTIONS(3113),
    [anon_sym_goto] = ACTIONS(3113),
    [anon_sym_return] = ACTIONS(3113),
    [anon_sym_undef] = ACTIONS(3113),
    [anon_sym_PERCENT2] = ACTIONS(3113),
    [anon_sym_AMP2] = ACTIONS(3113),
    [anon_sym_STAR2] = ACTIONS(3113),
    [anon_sym_last] = ACTIONS(3113),
    [anon_sym_next] = ACTIONS(3113),
    [anon_sym_redo] = ACTIONS(3113),
    [anon_sym___FILE__] = ACTIONS(3113),
    [anon_sym___LINE__] = ACTIONS(3113),
    [anon_sym___PACKAGE__] = ACTIONS(3113),
    [anon_sym___SUB__] = ACTIONS(3113),
    [anon_sym_break] = ACTIONS(3113),
    [anon_sym_fork] = ACTIONS(3113),
    [anon_sym_getppid] = ACTIONS(3113),
    [anon_sym_time] = ACTIONS(3113),
    [anon_sym_times] = ACTIONS(3113),
    [anon_sym_wait] = ACTIONS(3113),
    [anon_sym_wantarray] = ACTIONS(3113),
    [anon_sym_abs] = ACTIONS(3113),
    [anon_sym_alarm] = ACTIONS(3113),
    [anon_sym_chop] = ACTIONS(3113),
    [anon_sym_chdir] = ACTIONS(3113),
    [anon_sym_close] = ACTIONS(3113),
    [anon_sym_closedir] = ACTIONS(3113),
    [anon_sym_caller] = ACTIONS(3113),
    [anon_sym_chomp] = ACTIONS(3113),
    [anon_sym_chr] = ACTIONS(3113),
    [anon_sym_cos] = ACTIONS(3113),
    [anon_sym_chroot] = ACTIONS(3113),
    [anon_sym_defined] = ACTIONS(3113),
    [anon_sym_delete] = ACTIONS(3113),
    [anon_sym_dbmclose] = ACTIONS(3113),
    [anon_sym_exists] = ACTIONS(3113),
    [anon_sym_exit] = ACTIONS(3113),
    [anon_sym_eof] = ACTIONS(3113),
    [anon_sym_exp] = ACTIONS(3113),
    [anon_sym_each] = ACTIONS(3113),
    [anon_sym_fc] = ACTIONS(3113),
    [anon_sym_fileno] = ACTIONS(3113),
    [anon_sym_gmtime] = ACTIONS(3113),
    [anon_sym_getc] = ACTIONS(3113),
    [anon_sym_getpgrp] = ACTIONS(3113),
    [anon_sym_getprotobyname] = ACTIONS(3113),
    [anon_sym_getpwname] = ACTIONS(3113),
    [anon_sym_getpwuid] = ACTIONS(3113),
    [anon_sym_getpeername] = ACTIONS(3113),
    [anon_sym_getnetbyname] = ACTIONS(3113),
    [anon_sym_getsockname] = ACTIONS(3113),
    [anon_sym_getgrnam] = ACTIONS(3113),
    [anon_sym_getgrgid] = ACTIONS(3113),
    [anon_sym_hex] = ACTIONS(3113),
    [anon_sym_int] = ACTIONS(3113),
    [anon_sym_keys] = ACTIONS(3113),
    [anon_sym_lc] = ACTIONS(3113),
    [anon_sym_lcfirst] = ACTIONS(3113),
    [anon_sym_length] = ACTIONS(3113),
    [anon_sym_localtime] = ACTIONS(3113),
    [anon_sym_log] = ACTIONS(3113),
    [anon_sym_lock] = ACTIONS(3113),
    [anon_sym_lstat] = ACTIONS(3113),
    [anon_sym_oct] = ACTIONS(3113),
    [anon_sym_ord] = ACTIONS(3113),
    [anon_sym_prototype] = ACTIONS(3113),
    [anon_sym_pop] = ACTIONS(3113),
    [anon_sym_pos] = ACTIONS(3113),
    [anon_sym_quotemeta] = ACTIONS(3113),
    [anon_sym_reset] = ACTIONS(3113),
    [anon_sym_rand] = ACTIONS(3113),
    [anon_sym_rmdir] = ACTIONS(3113),
    [anon_sym_readdir] = ACTIONS(3113),
    [anon_sym_readline] = ACTIONS(3113),
    [anon_sym_readpipe] = ACTIONS(3113),
    [anon_sym_rewinddir] = ACTIONS(3113),
    [anon_sym_readlink] = ACTIONS(3113),
    [anon_sym_ref] = ACTIONS(3113),
    [anon_sym_scalar] = ACTIONS(3113),
    [anon_sym_shift] = ACTIONS(3113),
    [anon_sym_sin] = ACTIONS(3113),
    [anon_sym_sleep] = ACTIONS(3113),
    [anon_sym_sqrt] = ACTIONS(3113),
    [anon_sym_srand] = ACTIONS(3113),
    [anon_sym_stat] = ACTIONS(3113),
    [anon_sym_study] = ACTIONS(3113),
    [anon_sym_tell] = ACTIONS(3113),
    [anon_sym_telldir] = ACTIONS(3113),
    [anon_sym_tied] = ACTIONS(3113),
    [anon_sym_uc] = ACTIONS(3113),
    [anon_sym_ucfirst] = ACTIONS(3113),
    [anon_sym_untie] = ACTIONS(3113),
    [anon_sym_umask] = ACTIONS(3113),
    [anon_sym_values] = ACTIONS(3113),
    [anon_sym_write] = ACTIONS(3113),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3113),
    [anon_sym_qq] = ACTIONS(3113),
    [anon_sym_qw] = ACTIONS(3113),
    [anon_sym_qx] = ACTIONS(3113),
    [anon_sym_qr] = ACTIONS(3113),
    [anon_sym_m] = ACTIONS(3113),
    [anon_sym_s] = ACTIONS(3113),
    [anon_sym_tr] = ACTIONS(3113),
    [anon_sym_y] = ACTIONS(3113),
    [aux_sym__bareword_token1] = ACTIONS(3113),
    [sym_number] = ACTIONS(3113),
    [anon_sym_true] = ACTIONS(3113),
    [anon_sym_false] = ACTIONS(3113),
    [sym__single_quote] = ACTIONS(3111),
    [sym__double_quote] = ACTIONS(3111),
    [sym__backtick_quote] = ACTIONS(3111),
    [sym__search_slash_quote] = ACTIONS(3111),
    [sym__open_readline_bracket] = ACTIONS(3111),
    [sym__open_fileglob_bracket] = ACTIONS(3111),
    [sym__PERLY_HEREDOC] = ACTIONS(3111),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3111),
    [sym__filetest] = ACTIONS(3111),
  },
  [STATE(1695)] = {
    [sym_heredoc_content] = STATE(1695),
    [sym__identifier] = ACTIONS(3201),
    [anon_sym_LBRACE] = ACTIONS(3201),
    [anon_sym_DOLLAR] = ACTIONS(3201),
    [anon_sym_AT] = ACTIONS(3201),
    [anon_sym_RPAREN] = ACTIONS(3201),
    [anon_sym_my] = ACTIONS(3201),
    [anon_sym_extended] = ACTIONS(3201),
    [anon_sym_async] = ACTIONS(3201),
    [anon_sym_sub] = ACTIONS(3201),
    [anon_sym_method] = ACTIONS(3201),
    [anon_sym_LPAREN] = ACTIONS(3201),
    [anon_sym_continue] = ACTIONS(3201),
    [anon_sym_state] = ACTIONS(3201),
    [anon_sym_our] = ACTIONS(3201),
    [anon_sym_LBRACK] = ACTIONS(3201),
    [anon_sym_LT_LT] = ACTIONS(3201),
    [anon_sym_SLASH_SLASH] = ACTIONS(3201),
    [anon_sym_PLUS] = ACTIONS(3201),
    [anon_sym_DASH] = ACTIONS(3201),
    [anon_sym_TILDE] = ACTIONS(3201),
    [anon_sym_BANG] = ACTIONS(3201),
    [anon_sym_PLUS_PLUS] = ACTIONS(3201),
    [anon_sym_DASH_DASH] = ACTIONS(3201),
    [anon_sym_BSLASH] = ACTIONS(3199),
    [anon_sym_do] = ACTIONS(3201),
    [anon_sym_eval] = ACTIONS(3201),
    [anon_sym_await] = ACTIONS(3201),
    [anon_sym_field] = ACTIONS(3201),
    [anon_sym_local] = ACTIONS(3201),
    [anon_sym_dynamically] = ACTIONS(3201),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3201),
    [anon_sym_require] = ACTIONS(3201),
    [anon_sym_map] = ACTIONS(3201),
    [anon_sym_grep] = ACTIONS(3201),
    [anon_sym_sort] = ACTIONS(3201),
    [anon_sym_goto] = ACTIONS(3201),
    [anon_sym_return] = ACTIONS(3201),
    [anon_sym_undef] = ACTIONS(3201),
    [anon_sym_PERCENT2] = ACTIONS(3201),
    [anon_sym_AMP2] = ACTIONS(3201),
    [anon_sym_STAR2] = ACTIONS(3201),
    [anon_sym_last] = ACTIONS(3201),
    [anon_sym_next] = ACTIONS(3201),
    [anon_sym_redo] = ACTIONS(3201),
    [anon_sym___FILE__] = ACTIONS(3201),
    [anon_sym___LINE__] = ACTIONS(3201),
    [anon_sym___PACKAGE__] = ACTIONS(3201),
    [anon_sym___SUB__] = ACTIONS(3201),
    [anon_sym_break] = ACTIONS(3201),
    [anon_sym_fork] = ACTIONS(3201),
    [anon_sym_getppid] = ACTIONS(3201),
    [anon_sym_time] = ACTIONS(3201),
    [anon_sym_times] = ACTIONS(3201),
    [anon_sym_wait] = ACTIONS(3201),
    [anon_sym_wantarray] = ACTIONS(3201),
    [anon_sym_abs] = ACTIONS(3201),
    [anon_sym_alarm] = ACTIONS(3201),
    [anon_sym_chop] = ACTIONS(3201),
    [anon_sym_chdir] = ACTIONS(3201),
    [anon_sym_close] = ACTIONS(3201),
    [anon_sym_closedir] = ACTIONS(3201),
    [anon_sym_caller] = ACTIONS(3201),
    [anon_sym_chomp] = ACTIONS(3201),
    [anon_sym_chr] = ACTIONS(3201),
    [anon_sym_cos] = ACTIONS(3201),
    [anon_sym_chroot] = ACTIONS(3201),
    [anon_sym_defined] = ACTIONS(3201),
    [anon_sym_delete] = ACTIONS(3201),
    [anon_sym_dbmclose] = ACTIONS(3201),
    [anon_sym_exists] = ACTIONS(3201),
    [anon_sym_exit] = ACTIONS(3201),
    [anon_sym_eof] = ACTIONS(3201),
    [anon_sym_exp] = ACTIONS(3201),
    [anon_sym_each] = ACTIONS(3201),
    [anon_sym_fc] = ACTIONS(3201),
    [anon_sym_fileno] = ACTIONS(3201),
    [anon_sym_gmtime] = ACTIONS(3201),
    [anon_sym_getc] = ACTIONS(3201),
    [anon_sym_getpgrp] = ACTIONS(3201),
    [anon_sym_getprotobyname] = ACTIONS(3201),
    [anon_sym_getpwname] = ACTIONS(3201),
    [anon_sym_getpwuid] = ACTIONS(3201),
    [anon_sym_getpeername] = ACTIONS(3201),
    [anon_sym_getnetbyname] = ACTIONS(3201),
    [anon_sym_getsockname] = ACTIONS(3201),
    [anon_sym_getgrnam] = ACTIONS(3201),
    [anon_sym_getgrgid] = ACTIONS(3201),
    [anon_sym_hex] = ACTIONS(3201),
    [anon_sym_int] = ACTIONS(3201),
    [anon_sym_keys] = ACTIONS(3201),
    [anon_sym_lc] = ACTIONS(3201),
    [anon_sym_lcfirst] = ACTIONS(3201),
    [anon_sym_length] = ACTIONS(3201),
    [anon_sym_localtime] = ACTIONS(3201),
    [anon_sym_log] = ACTIONS(3201),
    [anon_sym_lock] = ACTIONS(3201),
    [anon_sym_lstat] = ACTIONS(3201),
    [anon_sym_oct] = ACTIONS(3201),
    [anon_sym_ord] = ACTIONS(3201),
    [anon_sym_prototype] = ACTIONS(3201),
    [anon_sym_pop] = ACTIONS(3201),
    [anon_sym_pos] = ACTIONS(3201),
    [anon_sym_quotemeta] = ACTIONS(3201),
    [anon_sym_reset] = ACTIONS(3201),
    [anon_sym_rand] = ACTIONS(3201),
    [anon_sym_rmdir] = ACTIONS(3201),
    [anon_sym_readdir] = ACTIONS(3201),
    [anon_sym_readline] = ACTIONS(3201),
    [anon_sym_readpipe] = ACTIONS(3201),
    [anon_sym_rewinddir] = ACTIONS(3201),
    [anon_sym_readlink] = ACTIONS(3201),
    [anon_sym_ref] = ACTIONS(3201),
    [anon_sym_scalar] = ACTIONS(3201),
    [anon_sym_shift] = ACTIONS(3201),
    [anon_sym_sin] = ACTIONS(3201),
    [anon_sym_sleep] = ACTIONS(3201),
    [anon_sym_sqrt] = ACTIONS(3201),
    [anon_sym_srand] = ACTIONS(3201),
    [anon_sym_stat] = ACTIONS(3201),
    [anon_sym_study] = ACTIONS(3201),
    [anon_sym_tell] = ACTIONS(3201),
    [anon_sym_telldir] = ACTIONS(3201),
    [anon_sym_tied] = ACTIONS(3201),
    [anon_sym_uc] = ACTIONS(3201),
    [anon_sym_ucfirst] = ACTIONS(3201),
    [anon_sym_untie] = ACTIONS(3201),
    [anon_sym_umask] = ACTIONS(3201),
    [anon_sym_values] = ACTIONS(3201),
    [anon_sym_write] = ACTIONS(3201),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3201),
    [anon_sym_qq] = ACTIONS(3201),
    [anon_sym_qw] = ACTIONS(3201),
    [anon_sym_qx] = ACTIONS(3201),
    [anon_sym_qr] = ACTIONS(3201),
    [anon_sym_m] = ACTIONS(3201),
    [anon_sym_s] = ACTIONS(3201),
    [anon_sym_tr] = ACTIONS(3201),
    [anon_sym_y] = ACTIONS(3201),
    [aux_sym__bareword_token1] = ACTIONS(3201),
    [sym_number] = ACTIONS(3201),
    [anon_sym_true] = ACTIONS(3201),
    [anon_sym_false] = ACTIONS(3201),
    [sym__single_quote] = ACTIONS(3199),
    [sym__double_quote] = ACTIONS(3199),
    [sym__backtick_quote] = ACTIONS(3199),
    [sym__search_slash_quote] = ACTIONS(3199),
    [sym__open_readline_bracket] = ACTIONS(3199),
    [sym__open_fileglob_bracket] = ACTIONS(3199),
    [sym__PERLY_HEREDOC] = ACTIONS(3199),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3199),
    [sym__filetest] = ACTIONS(3199),
  },
  [STATE(1696)] = {
    [sym_heredoc_content] = STATE(1696),
    [sym__identifier] = ACTIONS(2962),
    [anon_sym_LBRACE] = ACTIONS(2962),
    [anon_sym_DOLLAR] = ACTIONS(2962),
    [anon_sym_AT] = ACTIONS(2962),
    [anon_sym_RPAREN] = ACTIONS(2962),
    [anon_sym_my] = ACTIONS(2962),
    [anon_sym_extended] = ACTIONS(2962),
    [anon_sym_async] = ACTIONS(2962),
    [anon_sym_sub] = ACTIONS(2962),
    [anon_sym_method] = ACTIONS(2962),
    [anon_sym_LPAREN] = ACTIONS(2962),
    [anon_sym_continue] = ACTIONS(2962),
    [anon_sym_state] = ACTIONS(2962),
    [anon_sym_our] = ACTIONS(2962),
    [anon_sym_LBRACK] = ACTIONS(2962),
    [anon_sym_LT_LT] = ACTIONS(2962),
    [anon_sym_SLASH_SLASH] = ACTIONS(2962),
    [anon_sym_PLUS] = ACTIONS(2962),
    [anon_sym_DASH] = ACTIONS(2962),
    [anon_sym_TILDE] = ACTIONS(2962),
    [anon_sym_BANG] = ACTIONS(2962),
    [anon_sym_PLUS_PLUS] = ACTIONS(2962),
    [anon_sym_DASH_DASH] = ACTIONS(2962),
    [anon_sym_BSLASH] = ACTIONS(2960),
    [anon_sym_do] = ACTIONS(2962),
    [anon_sym_eval] = ACTIONS(2962),
    [anon_sym_await] = ACTIONS(2962),
    [anon_sym_field] = ACTIONS(2962),
    [anon_sym_local] = ACTIONS(2962),
    [anon_sym_dynamically] = ACTIONS(2962),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2962),
    [anon_sym_require] = ACTIONS(2962),
    [anon_sym_map] = ACTIONS(2962),
    [anon_sym_grep] = ACTIONS(2962),
    [anon_sym_sort] = ACTIONS(2962),
    [anon_sym_goto] = ACTIONS(2962),
    [anon_sym_return] = ACTIONS(2962),
    [anon_sym_undef] = ACTIONS(2962),
    [anon_sym_PERCENT2] = ACTIONS(2962),
    [anon_sym_AMP2] = ACTIONS(2962),
    [anon_sym_STAR2] = ACTIONS(2962),
    [anon_sym_last] = ACTIONS(2962),
    [anon_sym_next] = ACTIONS(2962),
    [anon_sym_redo] = ACTIONS(2962),
    [anon_sym___FILE__] = ACTIONS(2962),
    [anon_sym___LINE__] = ACTIONS(2962),
    [anon_sym___PACKAGE__] = ACTIONS(2962),
    [anon_sym___SUB__] = ACTIONS(2962),
    [anon_sym_break] = ACTIONS(2962),
    [anon_sym_fork] = ACTIONS(2962),
    [anon_sym_getppid] = ACTIONS(2962),
    [anon_sym_time] = ACTIONS(2962),
    [anon_sym_times] = ACTIONS(2962),
    [anon_sym_wait] = ACTIONS(2962),
    [anon_sym_wantarray] = ACTIONS(2962),
    [anon_sym_abs] = ACTIONS(2962),
    [anon_sym_alarm] = ACTIONS(2962),
    [anon_sym_chop] = ACTIONS(2962),
    [anon_sym_chdir] = ACTIONS(2962),
    [anon_sym_close] = ACTIONS(2962),
    [anon_sym_closedir] = ACTIONS(2962),
    [anon_sym_caller] = ACTIONS(2962),
    [anon_sym_chomp] = ACTIONS(2962),
    [anon_sym_chr] = ACTIONS(2962),
    [anon_sym_cos] = ACTIONS(2962),
    [anon_sym_chroot] = ACTIONS(2962),
    [anon_sym_defined] = ACTIONS(2962),
    [anon_sym_delete] = ACTIONS(2962),
    [anon_sym_dbmclose] = ACTIONS(2962),
    [anon_sym_exists] = ACTIONS(2962),
    [anon_sym_exit] = ACTIONS(2962),
    [anon_sym_eof] = ACTIONS(2962),
    [anon_sym_exp] = ACTIONS(2962),
    [anon_sym_each] = ACTIONS(2962),
    [anon_sym_fc] = ACTIONS(2962),
    [anon_sym_fileno] = ACTIONS(2962),
    [anon_sym_gmtime] = ACTIONS(2962),
    [anon_sym_getc] = ACTIONS(2962),
    [anon_sym_getpgrp] = ACTIONS(2962),
    [anon_sym_getprotobyname] = ACTIONS(2962),
    [anon_sym_getpwname] = ACTIONS(2962),
    [anon_sym_getpwuid] = ACTIONS(2962),
    [anon_sym_getpeername] = ACTIONS(2962),
    [anon_sym_getnetbyname] = ACTIONS(2962),
    [anon_sym_getsockname] = ACTIONS(2962),
    [anon_sym_getgrnam] = ACTIONS(2962),
    [anon_sym_getgrgid] = ACTIONS(2962),
    [anon_sym_hex] = ACTIONS(2962),
    [anon_sym_int] = ACTIONS(2962),
    [anon_sym_keys] = ACTIONS(2962),
    [anon_sym_lc] = ACTIONS(2962),
    [anon_sym_lcfirst] = ACTIONS(2962),
    [anon_sym_length] = ACTIONS(2962),
    [anon_sym_localtime] = ACTIONS(2962),
    [anon_sym_log] = ACTIONS(2962),
    [anon_sym_lock] = ACTIONS(2962),
    [anon_sym_lstat] = ACTIONS(2962),
    [anon_sym_oct] = ACTIONS(2962),
    [anon_sym_ord] = ACTIONS(2962),
    [anon_sym_prototype] = ACTIONS(2962),
    [anon_sym_pop] = ACTIONS(2962),
    [anon_sym_pos] = ACTIONS(2962),
    [anon_sym_quotemeta] = ACTIONS(2962),
    [anon_sym_reset] = ACTIONS(2962),
    [anon_sym_rand] = ACTIONS(2962),
    [anon_sym_rmdir] = ACTIONS(2962),
    [anon_sym_readdir] = ACTIONS(2962),
    [anon_sym_readline] = ACTIONS(2962),
    [anon_sym_readpipe] = ACTIONS(2962),
    [anon_sym_rewinddir] = ACTIONS(2962),
    [anon_sym_readlink] = ACTIONS(2962),
    [anon_sym_ref] = ACTIONS(2962),
    [anon_sym_scalar] = ACTIONS(2962),
    [anon_sym_shift] = ACTIONS(2962),
    [anon_sym_sin] = ACTIONS(2962),
    [anon_sym_sleep] = ACTIONS(2962),
    [anon_sym_sqrt] = ACTIONS(2962),
    [anon_sym_srand] = ACTIONS(2962),
    [anon_sym_stat] = ACTIONS(2962),
    [anon_sym_study] = ACTIONS(2962),
    [anon_sym_tell] = ACTIONS(2962),
    [anon_sym_telldir] = ACTIONS(2962),
    [anon_sym_tied] = ACTIONS(2962),
    [anon_sym_uc] = ACTIONS(2962),
    [anon_sym_ucfirst] = ACTIONS(2962),
    [anon_sym_untie] = ACTIONS(2962),
    [anon_sym_umask] = ACTIONS(2962),
    [anon_sym_values] = ACTIONS(2962),
    [anon_sym_write] = ACTIONS(2962),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2962),
    [anon_sym_qq] = ACTIONS(2962),
    [anon_sym_qw] = ACTIONS(2962),
    [anon_sym_qx] = ACTIONS(2962),
    [anon_sym_qr] = ACTIONS(2962),
    [anon_sym_m] = ACTIONS(2962),
    [anon_sym_s] = ACTIONS(2962),
    [anon_sym_tr] = ACTIONS(2962),
    [anon_sym_y] = ACTIONS(2962),
    [aux_sym__bareword_token1] = ACTIONS(2962),
    [sym_number] = ACTIONS(2962),
    [anon_sym_true] = ACTIONS(2962),
    [anon_sym_false] = ACTIONS(2962),
    [sym__single_quote] = ACTIONS(2960),
    [sym__double_quote] = ACTIONS(2960),
    [sym__backtick_quote] = ACTIONS(2960),
    [sym__search_slash_quote] = ACTIONS(2960),
    [sym__open_readline_bracket] = ACTIONS(2960),
    [sym__open_fileglob_bracket] = ACTIONS(2960),
    [sym__PERLY_HEREDOC] = ACTIONS(2960),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2960),
    [sym__filetest] = ACTIONS(2960),
  },
  [STATE(1697)] = {
    [sym_heredoc_content] = STATE(1697),
    [sym__identifier] = ACTIONS(3057),
    [anon_sym_LBRACE] = ACTIONS(3057),
    [anon_sym_DOLLAR] = ACTIONS(3057),
    [anon_sym_AT] = ACTIONS(3057),
    [anon_sym_RPAREN] = ACTIONS(3057),
    [anon_sym_my] = ACTIONS(3057),
    [anon_sym_extended] = ACTIONS(3057),
    [anon_sym_async] = ACTIONS(3057),
    [anon_sym_sub] = ACTIONS(3057),
    [anon_sym_method] = ACTIONS(3057),
    [anon_sym_LPAREN] = ACTIONS(3057),
    [anon_sym_continue] = ACTIONS(3057),
    [anon_sym_state] = ACTIONS(3057),
    [anon_sym_our] = ACTIONS(3057),
    [anon_sym_LBRACK] = ACTIONS(3057),
    [anon_sym_LT_LT] = ACTIONS(3057),
    [anon_sym_SLASH_SLASH] = ACTIONS(3057),
    [anon_sym_PLUS] = ACTIONS(3057),
    [anon_sym_DASH] = ACTIONS(3057),
    [anon_sym_TILDE] = ACTIONS(3057),
    [anon_sym_BANG] = ACTIONS(3057),
    [anon_sym_PLUS_PLUS] = ACTIONS(3057),
    [anon_sym_DASH_DASH] = ACTIONS(3057),
    [anon_sym_BSLASH] = ACTIONS(3055),
    [anon_sym_do] = ACTIONS(3057),
    [anon_sym_eval] = ACTIONS(3057),
    [anon_sym_await] = ACTIONS(3057),
    [anon_sym_field] = ACTIONS(3057),
    [anon_sym_local] = ACTIONS(3057),
    [anon_sym_dynamically] = ACTIONS(3057),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3057),
    [anon_sym_require] = ACTIONS(3057),
    [anon_sym_map] = ACTIONS(3057),
    [anon_sym_grep] = ACTIONS(3057),
    [anon_sym_sort] = ACTIONS(3057),
    [anon_sym_goto] = ACTIONS(3057),
    [anon_sym_return] = ACTIONS(3057),
    [anon_sym_undef] = ACTIONS(3057),
    [anon_sym_PERCENT2] = ACTIONS(3057),
    [anon_sym_AMP2] = ACTIONS(3057),
    [anon_sym_STAR2] = ACTIONS(3057),
    [anon_sym_last] = ACTIONS(3057),
    [anon_sym_next] = ACTIONS(3057),
    [anon_sym_redo] = ACTIONS(3057),
    [anon_sym___FILE__] = ACTIONS(3057),
    [anon_sym___LINE__] = ACTIONS(3057),
    [anon_sym___PACKAGE__] = ACTIONS(3057),
    [anon_sym___SUB__] = ACTIONS(3057),
    [anon_sym_break] = ACTIONS(3057),
    [anon_sym_fork] = ACTIONS(3057),
    [anon_sym_getppid] = ACTIONS(3057),
    [anon_sym_time] = ACTIONS(3057),
    [anon_sym_times] = ACTIONS(3057),
    [anon_sym_wait] = ACTIONS(3057),
    [anon_sym_wantarray] = ACTIONS(3057),
    [anon_sym_abs] = ACTIONS(3057),
    [anon_sym_alarm] = ACTIONS(3057),
    [anon_sym_chop] = ACTIONS(3057),
    [anon_sym_chdir] = ACTIONS(3057),
    [anon_sym_close] = ACTIONS(3057),
    [anon_sym_closedir] = ACTIONS(3057),
    [anon_sym_caller] = ACTIONS(3057),
    [anon_sym_chomp] = ACTIONS(3057),
    [anon_sym_chr] = ACTIONS(3057),
    [anon_sym_cos] = ACTIONS(3057),
    [anon_sym_chroot] = ACTIONS(3057),
    [anon_sym_defined] = ACTIONS(3057),
    [anon_sym_delete] = ACTIONS(3057),
    [anon_sym_dbmclose] = ACTIONS(3057),
    [anon_sym_exists] = ACTIONS(3057),
    [anon_sym_exit] = ACTIONS(3057),
    [anon_sym_eof] = ACTIONS(3057),
    [anon_sym_exp] = ACTIONS(3057),
    [anon_sym_each] = ACTIONS(3057),
    [anon_sym_fc] = ACTIONS(3057),
    [anon_sym_fileno] = ACTIONS(3057),
    [anon_sym_gmtime] = ACTIONS(3057),
    [anon_sym_getc] = ACTIONS(3057),
    [anon_sym_getpgrp] = ACTIONS(3057),
    [anon_sym_getprotobyname] = ACTIONS(3057),
    [anon_sym_getpwname] = ACTIONS(3057),
    [anon_sym_getpwuid] = ACTIONS(3057),
    [anon_sym_getpeername] = ACTIONS(3057),
    [anon_sym_getnetbyname] = ACTIONS(3057),
    [anon_sym_getsockname] = ACTIONS(3057),
    [anon_sym_getgrnam] = ACTIONS(3057),
    [anon_sym_getgrgid] = ACTIONS(3057),
    [anon_sym_hex] = ACTIONS(3057),
    [anon_sym_int] = ACTIONS(3057),
    [anon_sym_keys] = ACTIONS(3057),
    [anon_sym_lc] = ACTIONS(3057),
    [anon_sym_lcfirst] = ACTIONS(3057),
    [anon_sym_length] = ACTIONS(3057),
    [anon_sym_localtime] = ACTIONS(3057),
    [anon_sym_log] = ACTIONS(3057),
    [anon_sym_lock] = ACTIONS(3057),
    [anon_sym_lstat] = ACTIONS(3057),
    [anon_sym_oct] = ACTIONS(3057),
    [anon_sym_ord] = ACTIONS(3057),
    [anon_sym_prototype] = ACTIONS(3057),
    [anon_sym_pop] = ACTIONS(3057),
    [anon_sym_pos] = ACTIONS(3057),
    [anon_sym_quotemeta] = ACTIONS(3057),
    [anon_sym_reset] = ACTIONS(3057),
    [anon_sym_rand] = ACTIONS(3057),
    [anon_sym_rmdir] = ACTIONS(3057),
    [anon_sym_readdir] = ACTIONS(3057),
    [anon_sym_readline] = ACTIONS(3057),
    [anon_sym_readpipe] = ACTIONS(3057),
    [anon_sym_rewinddir] = ACTIONS(3057),
    [anon_sym_readlink] = ACTIONS(3057),
    [anon_sym_ref] = ACTIONS(3057),
    [anon_sym_scalar] = ACTIONS(3057),
    [anon_sym_shift] = ACTIONS(3057),
    [anon_sym_sin] = ACTIONS(3057),
    [anon_sym_sleep] = ACTIONS(3057),
    [anon_sym_sqrt] = ACTIONS(3057),
    [anon_sym_srand] = ACTIONS(3057),
    [anon_sym_stat] = ACTIONS(3057),
    [anon_sym_study] = ACTIONS(3057),
    [anon_sym_tell] = ACTIONS(3057),
    [anon_sym_telldir] = ACTIONS(3057),
    [anon_sym_tied] = ACTIONS(3057),
    [anon_sym_uc] = ACTIONS(3057),
    [anon_sym_ucfirst] = ACTIONS(3057),
    [anon_sym_untie] = ACTIONS(3057),
    [anon_sym_umask] = ACTIONS(3057),
    [anon_sym_values] = ACTIONS(3057),
    [anon_sym_write] = ACTIONS(3057),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3057),
    [anon_sym_qq] = ACTIONS(3057),
    [anon_sym_qw] = ACTIONS(3057),
    [anon_sym_qx] = ACTIONS(3057),
    [anon_sym_qr] = ACTIONS(3057),
    [anon_sym_m] = ACTIONS(3057),
    [anon_sym_s] = ACTIONS(3057),
    [anon_sym_tr] = ACTIONS(3057),
    [anon_sym_y] = ACTIONS(3057),
    [aux_sym__bareword_token1] = ACTIONS(3057),
    [sym_number] = ACTIONS(3057),
    [anon_sym_true] = ACTIONS(3057),
    [anon_sym_false] = ACTIONS(3057),
    [sym__single_quote] = ACTIONS(3055),
    [sym__double_quote] = ACTIONS(3055),
    [sym__backtick_quote] = ACTIONS(3055),
    [sym__search_slash_quote] = ACTIONS(3055),
    [sym__open_readline_bracket] = ACTIONS(3055),
    [sym__open_fileglob_bracket] = ACTIONS(3055),
    [sym__PERLY_HEREDOC] = ACTIONS(3055),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3055),
    [sym__filetest] = ACTIONS(3055),
  },
  [STATE(1698)] = {
    [sym_heredoc_content] = STATE(1698),
    [sym__identifier] = ACTIONS(3065),
    [anon_sym_LBRACE] = ACTIONS(3065),
    [anon_sym_DOLLAR] = ACTIONS(3065),
    [anon_sym_AT] = ACTIONS(3065),
    [anon_sym_RPAREN] = ACTIONS(3065),
    [anon_sym_my] = ACTIONS(3065),
    [anon_sym_extended] = ACTIONS(3065),
    [anon_sym_async] = ACTIONS(3065),
    [anon_sym_sub] = ACTIONS(3065),
    [anon_sym_method] = ACTIONS(3065),
    [anon_sym_LPAREN] = ACTIONS(3065),
    [anon_sym_continue] = ACTIONS(3065),
    [anon_sym_state] = ACTIONS(3065),
    [anon_sym_our] = ACTIONS(3065),
    [anon_sym_LBRACK] = ACTIONS(3065),
    [anon_sym_LT_LT] = ACTIONS(3065),
    [anon_sym_SLASH_SLASH] = ACTIONS(3065),
    [anon_sym_PLUS] = ACTIONS(3065),
    [anon_sym_DASH] = ACTIONS(3065),
    [anon_sym_TILDE] = ACTIONS(3065),
    [anon_sym_BANG] = ACTIONS(3065),
    [anon_sym_PLUS_PLUS] = ACTIONS(3065),
    [anon_sym_DASH_DASH] = ACTIONS(3065),
    [anon_sym_BSLASH] = ACTIONS(3063),
    [anon_sym_do] = ACTIONS(3065),
    [anon_sym_eval] = ACTIONS(3065),
    [anon_sym_await] = ACTIONS(3065),
    [anon_sym_field] = ACTIONS(3065),
    [anon_sym_local] = ACTIONS(3065),
    [anon_sym_dynamically] = ACTIONS(3065),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3065),
    [anon_sym_require] = ACTIONS(3065),
    [anon_sym_map] = ACTIONS(3065),
    [anon_sym_grep] = ACTIONS(3065),
    [anon_sym_sort] = ACTIONS(3065),
    [anon_sym_goto] = ACTIONS(3065),
    [anon_sym_return] = ACTIONS(3065),
    [anon_sym_undef] = ACTIONS(3065),
    [anon_sym_PERCENT2] = ACTIONS(3065),
    [anon_sym_AMP2] = ACTIONS(3065),
    [anon_sym_STAR2] = ACTIONS(3065),
    [anon_sym_last] = ACTIONS(3065),
    [anon_sym_next] = ACTIONS(3065),
    [anon_sym_redo] = ACTIONS(3065),
    [anon_sym___FILE__] = ACTIONS(3065),
    [anon_sym___LINE__] = ACTIONS(3065),
    [anon_sym___PACKAGE__] = ACTIONS(3065),
    [anon_sym___SUB__] = ACTIONS(3065),
    [anon_sym_break] = ACTIONS(3065),
    [anon_sym_fork] = ACTIONS(3065),
    [anon_sym_getppid] = ACTIONS(3065),
    [anon_sym_time] = ACTIONS(3065),
    [anon_sym_times] = ACTIONS(3065),
    [anon_sym_wait] = ACTIONS(3065),
    [anon_sym_wantarray] = ACTIONS(3065),
    [anon_sym_abs] = ACTIONS(3065),
    [anon_sym_alarm] = ACTIONS(3065),
    [anon_sym_chop] = ACTIONS(3065),
    [anon_sym_chdir] = ACTIONS(3065),
    [anon_sym_close] = ACTIONS(3065),
    [anon_sym_closedir] = ACTIONS(3065),
    [anon_sym_caller] = ACTIONS(3065),
    [anon_sym_chomp] = ACTIONS(3065),
    [anon_sym_chr] = ACTIONS(3065),
    [anon_sym_cos] = ACTIONS(3065),
    [anon_sym_chroot] = ACTIONS(3065),
    [anon_sym_defined] = ACTIONS(3065),
    [anon_sym_delete] = ACTIONS(3065),
    [anon_sym_dbmclose] = ACTIONS(3065),
    [anon_sym_exists] = ACTIONS(3065),
    [anon_sym_exit] = ACTIONS(3065),
    [anon_sym_eof] = ACTIONS(3065),
    [anon_sym_exp] = ACTIONS(3065),
    [anon_sym_each] = ACTIONS(3065),
    [anon_sym_fc] = ACTIONS(3065),
    [anon_sym_fileno] = ACTIONS(3065),
    [anon_sym_gmtime] = ACTIONS(3065),
    [anon_sym_getc] = ACTIONS(3065),
    [anon_sym_getpgrp] = ACTIONS(3065),
    [anon_sym_getprotobyname] = ACTIONS(3065),
    [anon_sym_getpwname] = ACTIONS(3065),
    [anon_sym_getpwuid] = ACTIONS(3065),
    [anon_sym_getpeername] = ACTIONS(3065),
    [anon_sym_getnetbyname] = ACTIONS(3065),
    [anon_sym_getsockname] = ACTIONS(3065),
    [anon_sym_getgrnam] = ACTIONS(3065),
    [anon_sym_getgrgid] = ACTIONS(3065),
    [anon_sym_hex] = ACTIONS(3065),
    [anon_sym_int] = ACTIONS(3065),
    [anon_sym_keys] = ACTIONS(3065),
    [anon_sym_lc] = ACTIONS(3065),
    [anon_sym_lcfirst] = ACTIONS(3065),
    [anon_sym_length] = ACTIONS(3065),
    [anon_sym_localtime] = ACTIONS(3065),
    [anon_sym_log] = ACTIONS(3065),
    [anon_sym_lock] = ACTIONS(3065),
    [anon_sym_lstat] = ACTIONS(3065),
    [anon_sym_oct] = ACTIONS(3065),
    [anon_sym_ord] = ACTIONS(3065),
    [anon_sym_prototype] = ACTIONS(3065),
    [anon_sym_pop] = ACTIONS(3065),
    [anon_sym_pos] = ACTIONS(3065),
    [anon_sym_quotemeta] = ACTIONS(3065),
    [anon_sym_reset] = ACTIONS(3065),
    [anon_sym_rand] = ACTIONS(3065),
    [anon_sym_rmdir] = ACTIONS(3065),
    [anon_sym_readdir] = ACTIONS(3065),
    [anon_sym_readline] = ACTIONS(3065),
    [anon_sym_readpipe] = ACTIONS(3065),
    [anon_sym_rewinddir] = ACTIONS(3065),
    [anon_sym_readlink] = ACTIONS(3065),
    [anon_sym_ref] = ACTIONS(3065),
    [anon_sym_scalar] = ACTIONS(3065),
    [anon_sym_shift] = ACTIONS(3065),
    [anon_sym_sin] = ACTIONS(3065),
    [anon_sym_sleep] = ACTIONS(3065),
    [anon_sym_sqrt] = ACTIONS(3065),
    [anon_sym_srand] = ACTIONS(3065),
    [anon_sym_stat] = ACTIONS(3065),
    [anon_sym_study] = ACTIONS(3065),
    [anon_sym_tell] = ACTIONS(3065),
    [anon_sym_telldir] = ACTIONS(3065),
    [anon_sym_tied] = ACTIONS(3065),
    [anon_sym_uc] = ACTIONS(3065),
    [anon_sym_ucfirst] = ACTIONS(3065),
    [anon_sym_untie] = ACTIONS(3065),
    [anon_sym_umask] = ACTIONS(3065),
    [anon_sym_values] = ACTIONS(3065),
    [anon_sym_write] = ACTIONS(3065),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3065),
    [anon_sym_qq] = ACTIONS(3065),
    [anon_sym_qw] = ACTIONS(3065),
    [anon_sym_qx] = ACTIONS(3065),
    [anon_sym_qr] = ACTIONS(3065),
    [anon_sym_m] = ACTIONS(3065),
    [anon_sym_s] = ACTIONS(3065),
    [anon_sym_tr] = ACTIONS(3065),
    [anon_sym_y] = ACTIONS(3065),
    [aux_sym__bareword_token1] = ACTIONS(3065),
    [sym_number] = ACTIONS(3065),
    [anon_sym_true] = ACTIONS(3065),
    [anon_sym_false] = ACTIONS(3065),
    [sym__single_quote] = ACTIONS(3063),
    [sym__double_quote] = ACTIONS(3063),
    [sym__backtick_quote] = ACTIONS(3063),
    [sym__search_slash_quote] = ACTIONS(3063),
    [sym__open_readline_bracket] = ACTIONS(3063),
    [sym__open_fileglob_bracket] = ACTIONS(3063),
    [sym__PERLY_HEREDOC] = ACTIONS(3063),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3063),
    [sym__filetest] = ACTIONS(3063),
  },
  [STATE(1699)] = {
    [sym_heredoc_content] = STATE(1699),
    [sym__identifier] = ACTIONS(2914),
    [anon_sym_LBRACE] = ACTIONS(2914),
    [anon_sym_DOLLAR] = ACTIONS(2914),
    [anon_sym_AT] = ACTIONS(2914),
    [anon_sym_RPAREN] = ACTIONS(2914),
    [anon_sym_my] = ACTIONS(2914),
    [anon_sym_extended] = ACTIONS(2914),
    [anon_sym_async] = ACTIONS(2914),
    [anon_sym_sub] = ACTIONS(2914),
    [anon_sym_method] = ACTIONS(2914),
    [anon_sym_LPAREN] = ACTIONS(2914),
    [anon_sym_continue] = ACTIONS(2914),
    [anon_sym_state] = ACTIONS(2914),
    [anon_sym_our] = ACTIONS(2914),
    [anon_sym_LBRACK] = ACTIONS(2914),
    [anon_sym_LT_LT] = ACTIONS(2914),
    [anon_sym_SLASH_SLASH] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2914),
    [anon_sym_DASH] = ACTIONS(2914),
    [anon_sym_TILDE] = ACTIONS(2914),
    [anon_sym_BANG] = ACTIONS(2914),
    [anon_sym_PLUS_PLUS] = ACTIONS(2914),
    [anon_sym_DASH_DASH] = ACTIONS(2914),
    [anon_sym_BSLASH] = ACTIONS(2912),
    [anon_sym_do] = ACTIONS(2914),
    [anon_sym_eval] = ACTIONS(2914),
    [anon_sym_await] = ACTIONS(2914),
    [anon_sym_field] = ACTIONS(2914),
    [anon_sym_local] = ACTIONS(2914),
    [anon_sym_dynamically] = ACTIONS(2914),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2914),
    [anon_sym_require] = ACTIONS(2914),
    [anon_sym_map] = ACTIONS(2914),
    [anon_sym_grep] = ACTIONS(2914),
    [anon_sym_sort] = ACTIONS(2914),
    [anon_sym_goto] = ACTIONS(2914),
    [anon_sym_return] = ACTIONS(2914),
    [anon_sym_undef] = ACTIONS(2914),
    [anon_sym_PERCENT2] = ACTIONS(2914),
    [anon_sym_AMP2] = ACTIONS(2914),
    [anon_sym_STAR2] = ACTIONS(2914),
    [anon_sym_last] = ACTIONS(2914),
    [anon_sym_next] = ACTIONS(2914),
    [anon_sym_redo] = ACTIONS(2914),
    [anon_sym___FILE__] = ACTIONS(2914),
    [anon_sym___LINE__] = ACTIONS(2914),
    [anon_sym___PACKAGE__] = ACTIONS(2914),
    [anon_sym___SUB__] = ACTIONS(2914),
    [anon_sym_break] = ACTIONS(2914),
    [anon_sym_fork] = ACTIONS(2914),
    [anon_sym_getppid] = ACTIONS(2914),
    [anon_sym_time] = ACTIONS(2914),
    [anon_sym_times] = ACTIONS(2914),
    [anon_sym_wait] = ACTIONS(2914),
    [anon_sym_wantarray] = ACTIONS(2914),
    [anon_sym_abs] = ACTIONS(2914),
    [anon_sym_alarm] = ACTIONS(2914),
    [anon_sym_chop] = ACTIONS(2914),
    [anon_sym_chdir] = ACTIONS(2914),
    [anon_sym_close] = ACTIONS(2914),
    [anon_sym_closedir] = ACTIONS(2914),
    [anon_sym_caller] = ACTIONS(2914),
    [anon_sym_chomp] = ACTIONS(2914),
    [anon_sym_chr] = ACTIONS(2914),
    [anon_sym_cos] = ACTIONS(2914),
    [anon_sym_chroot] = ACTIONS(2914),
    [anon_sym_defined] = ACTIONS(2914),
    [anon_sym_delete] = ACTIONS(2914),
    [anon_sym_dbmclose] = ACTIONS(2914),
    [anon_sym_exists] = ACTIONS(2914),
    [anon_sym_exit] = ACTIONS(2914),
    [anon_sym_eof] = ACTIONS(2914),
    [anon_sym_exp] = ACTIONS(2914),
    [anon_sym_each] = ACTIONS(2914),
    [anon_sym_fc] = ACTIONS(2914),
    [anon_sym_fileno] = ACTIONS(2914),
    [anon_sym_gmtime] = ACTIONS(2914),
    [anon_sym_getc] = ACTIONS(2914),
    [anon_sym_getpgrp] = ACTIONS(2914),
    [anon_sym_getprotobyname] = ACTIONS(2914),
    [anon_sym_getpwname] = ACTIONS(2914),
    [anon_sym_getpwuid] = ACTIONS(2914),
    [anon_sym_getpeername] = ACTIONS(2914),
    [anon_sym_getnetbyname] = ACTIONS(2914),
    [anon_sym_getsockname] = ACTIONS(2914),
    [anon_sym_getgrnam] = ACTIONS(2914),
    [anon_sym_getgrgid] = ACTIONS(2914),
    [anon_sym_hex] = ACTIONS(2914),
    [anon_sym_int] = ACTIONS(2914),
    [anon_sym_keys] = ACTIONS(2914),
    [anon_sym_lc] = ACTIONS(2914),
    [anon_sym_lcfirst] = ACTIONS(2914),
    [anon_sym_length] = ACTIONS(2914),
    [anon_sym_localtime] = ACTIONS(2914),
    [anon_sym_log] = ACTIONS(2914),
    [anon_sym_lock] = ACTIONS(2914),
    [anon_sym_lstat] = ACTIONS(2914),
    [anon_sym_oct] = ACTIONS(2914),
    [anon_sym_ord] = ACTIONS(2914),
    [anon_sym_prototype] = ACTIONS(2914),
    [anon_sym_pop] = ACTIONS(2914),
    [anon_sym_pos] = ACTIONS(2914),
    [anon_sym_quotemeta] = ACTIONS(2914),
    [anon_sym_reset] = ACTIONS(2914),
    [anon_sym_rand] = ACTIONS(2914),
    [anon_sym_rmdir] = ACTIONS(2914),
    [anon_sym_readdir] = ACTIONS(2914),
    [anon_sym_readline] = ACTIONS(2914),
    [anon_sym_readpipe] = ACTIONS(2914),
    [anon_sym_rewinddir] = ACTIONS(2914),
    [anon_sym_readlink] = ACTIONS(2914),
    [anon_sym_ref] = ACTIONS(2914),
    [anon_sym_scalar] = ACTIONS(2914),
    [anon_sym_shift] = ACTIONS(2914),
    [anon_sym_sin] = ACTIONS(2914),
    [anon_sym_sleep] = ACTIONS(2914),
    [anon_sym_sqrt] = ACTIONS(2914),
    [anon_sym_srand] = ACTIONS(2914),
    [anon_sym_stat] = ACTIONS(2914),
    [anon_sym_study] = ACTIONS(2914),
    [anon_sym_tell] = ACTIONS(2914),
    [anon_sym_telldir] = ACTIONS(2914),
    [anon_sym_tied] = ACTIONS(2914),
    [anon_sym_uc] = ACTIONS(2914),
    [anon_sym_ucfirst] = ACTIONS(2914),
    [anon_sym_untie] = ACTIONS(2914),
    [anon_sym_umask] = ACTIONS(2914),
    [anon_sym_values] = ACTIONS(2914),
    [anon_sym_write] = ACTIONS(2914),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2914),
    [anon_sym_qq] = ACTIONS(2914),
    [anon_sym_qw] = ACTIONS(2914),
    [anon_sym_qx] = ACTIONS(2914),
    [anon_sym_qr] = ACTIONS(2914),
    [anon_sym_m] = ACTIONS(2914),
    [anon_sym_s] = ACTIONS(2914),
    [anon_sym_tr] = ACTIONS(2914),
    [anon_sym_y] = ACTIONS(2914),
    [aux_sym__bareword_token1] = ACTIONS(2914),
    [sym_number] = ACTIONS(2914),
    [anon_sym_true] = ACTIONS(2914),
    [anon_sym_false] = ACTIONS(2914),
    [sym__single_quote] = ACTIONS(2912),
    [sym__double_quote] = ACTIONS(2912),
    [sym__backtick_quote] = ACTIONS(2912),
    [sym__search_slash_quote] = ACTIONS(2912),
    [sym__open_readline_bracket] = ACTIONS(2912),
    [sym__open_fileglob_bracket] = ACTIONS(2912),
    [sym__PERLY_HEREDOC] = ACTIONS(2912),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2912),
    [sym__filetest] = ACTIONS(2912),
  },
  [STATE(1700)] = {
    [sym_heredoc_content] = STATE(1700),
    [sym__identifier] = ACTIONS(2950),
    [anon_sym_LBRACE] = ACTIONS(2950),
    [anon_sym_DOLLAR] = ACTIONS(2950),
    [anon_sym_AT] = ACTIONS(2950),
    [anon_sym_RPAREN] = ACTIONS(2950),
    [anon_sym_my] = ACTIONS(2950),
    [anon_sym_extended] = ACTIONS(2950),
    [anon_sym_async] = ACTIONS(2950),
    [anon_sym_sub] = ACTIONS(2950),
    [anon_sym_method] = ACTIONS(2950),
    [anon_sym_LPAREN] = ACTIONS(2950),
    [anon_sym_continue] = ACTIONS(2950),
    [anon_sym_state] = ACTIONS(2950),
    [anon_sym_our] = ACTIONS(2950),
    [anon_sym_LBRACK] = ACTIONS(2950),
    [anon_sym_LT_LT] = ACTIONS(2950),
    [anon_sym_SLASH_SLASH] = ACTIONS(2950),
    [anon_sym_PLUS] = ACTIONS(2950),
    [anon_sym_DASH] = ACTIONS(2950),
    [anon_sym_TILDE] = ACTIONS(2950),
    [anon_sym_BANG] = ACTIONS(2950),
    [anon_sym_PLUS_PLUS] = ACTIONS(2950),
    [anon_sym_DASH_DASH] = ACTIONS(2950),
    [anon_sym_BSLASH] = ACTIONS(2948),
    [anon_sym_do] = ACTIONS(2950),
    [anon_sym_eval] = ACTIONS(2950),
    [anon_sym_await] = ACTIONS(2950),
    [anon_sym_field] = ACTIONS(2950),
    [anon_sym_local] = ACTIONS(2950),
    [anon_sym_dynamically] = ACTIONS(2950),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2950),
    [anon_sym_require] = ACTIONS(2950),
    [anon_sym_map] = ACTIONS(2950),
    [anon_sym_grep] = ACTIONS(2950),
    [anon_sym_sort] = ACTIONS(2950),
    [anon_sym_goto] = ACTIONS(2950),
    [anon_sym_return] = ACTIONS(2950),
    [anon_sym_undef] = ACTIONS(2950),
    [anon_sym_PERCENT2] = ACTIONS(2950),
    [anon_sym_AMP2] = ACTIONS(2950),
    [anon_sym_STAR2] = ACTIONS(2950),
    [anon_sym_last] = ACTIONS(2950),
    [anon_sym_next] = ACTIONS(2950),
    [anon_sym_redo] = ACTIONS(2950),
    [anon_sym___FILE__] = ACTIONS(2950),
    [anon_sym___LINE__] = ACTIONS(2950),
    [anon_sym___PACKAGE__] = ACTIONS(2950),
    [anon_sym___SUB__] = ACTIONS(2950),
    [anon_sym_break] = ACTIONS(2950),
    [anon_sym_fork] = ACTIONS(2950),
    [anon_sym_getppid] = ACTIONS(2950),
    [anon_sym_time] = ACTIONS(2950),
    [anon_sym_times] = ACTIONS(2950),
    [anon_sym_wait] = ACTIONS(2950),
    [anon_sym_wantarray] = ACTIONS(2950),
    [anon_sym_abs] = ACTIONS(2950),
    [anon_sym_alarm] = ACTIONS(2950),
    [anon_sym_chop] = ACTIONS(2950),
    [anon_sym_chdir] = ACTIONS(2950),
    [anon_sym_close] = ACTIONS(2950),
    [anon_sym_closedir] = ACTIONS(2950),
    [anon_sym_caller] = ACTIONS(2950),
    [anon_sym_chomp] = ACTIONS(2950),
    [anon_sym_chr] = ACTIONS(2950),
    [anon_sym_cos] = ACTIONS(2950),
    [anon_sym_chroot] = ACTIONS(2950),
    [anon_sym_defined] = ACTIONS(2950),
    [anon_sym_delete] = ACTIONS(2950),
    [anon_sym_dbmclose] = ACTIONS(2950),
    [anon_sym_exists] = ACTIONS(2950),
    [anon_sym_exit] = ACTIONS(2950),
    [anon_sym_eof] = ACTIONS(2950),
    [anon_sym_exp] = ACTIONS(2950),
    [anon_sym_each] = ACTIONS(2950),
    [anon_sym_fc] = ACTIONS(2950),
    [anon_sym_fileno] = ACTIONS(2950),
    [anon_sym_gmtime] = ACTIONS(2950),
    [anon_sym_getc] = ACTIONS(2950),
    [anon_sym_getpgrp] = ACTIONS(2950),
    [anon_sym_getprotobyname] = ACTIONS(2950),
    [anon_sym_getpwname] = ACTIONS(2950),
    [anon_sym_getpwuid] = ACTIONS(2950),
    [anon_sym_getpeername] = ACTIONS(2950),
    [anon_sym_getnetbyname] = ACTIONS(2950),
    [anon_sym_getsockname] = ACTIONS(2950),
    [anon_sym_getgrnam] = ACTIONS(2950),
    [anon_sym_getgrgid] = ACTIONS(2950),
    [anon_sym_hex] = ACTIONS(2950),
    [anon_sym_int] = ACTIONS(2950),
    [anon_sym_keys] = ACTIONS(2950),
    [anon_sym_lc] = ACTIONS(2950),
    [anon_sym_lcfirst] = ACTIONS(2950),
    [anon_sym_length] = ACTIONS(2950),
    [anon_sym_localtime] = ACTIONS(2950),
    [anon_sym_log] = ACTIONS(2950),
    [anon_sym_lock] = ACTIONS(2950),
    [anon_sym_lstat] = ACTIONS(2950),
    [anon_sym_oct] = ACTIONS(2950),
    [anon_sym_ord] = ACTIONS(2950),
    [anon_sym_prototype] = ACTIONS(2950),
    [anon_sym_pop] = ACTIONS(2950),
    [anon_sym_pos] = ACTIONS(2950),
    [anon_sym_quotemeta] = ACTIONS(2950),
    [anon_sym_reset] = ACTIONS(2950),
    [anon_sym_rand] = ACTIONS(2950),
    [anon_sym_rmdir] = ACTIONS(2950),
    [anon_sym_readdir] = ACTIONS(2950),
    [anon_sym_readline] = ACTIONS(2950),
    [anon_sym_readpipe] = ACTIONS(2950),
    [anon_sym_rewinddir] = ACTIONS(2950),
    [anon_sym_readlink] = ACTIONS(2950),
    [anon_sym_ref] = ACTIONS(2950),
    [anon_sym_scalar] = ACTIONS(2950),
    [anon_sym_shift] = ACTIONS(2950),
    [anon_sym_sin] = ACTIONS(2950),
    [anon_sym_sleep] = ACTIONS(2950),
    [anon_sym_sqrt] = ACTIONS(2950),
    [anon_sym_srand] = ACTIONS(2950),
    [anon_sym_stat] = ACTIONS(2950),
    [anon_sym_study] = ACTIONS(2950),
    [anon_sym_tell] = ACTIONS(2950),
    [anon_sym_telldir] = ACTIONS(2950),
    [anon_sym_tied] = ACTIONS(2950),
    [anon_sym_uc] = ACTIONS(2950),
    [anon_sym_ucfirst] = ACTIONS(2950),
    [anon_sym_untie] = ACTIONS(2950),
    [anon_sym_umask] = ACTIONS(2950),
    [anon_sym_values] = ACTIONS(2950),
    [anon_sym_write] = ACTIONS(2950),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2950),
    [anon_sym_qq] = ACTIONS(2950),
    [anon_sym_qw] = ACTIONS(2950),
    [anon_sym_qx] = ACTIONS(2950),
    [anon_sym_qr] = ACTIONS(2950),
    [anon_sym_m] = ACTIONS(2950),
    [anon_sym_s] = ACTIONS(2950),
    [anon_sym_tr] = ACTIONS(2950),
    [anon_sym_y] = ACTIONS(2950),
    [aux_sym__bareword_token1] = ACTIONS(2950),
    [sym_number] = ACTIONS(2950),
    [anon_sym_true] = ACTIONS(2950),
    [anon_sym_false] = ACTIONS(2950),
    [sym__single_quote] = ACTIONS(2948),
    [sym__double_quote] = ACTIONS(2948),
    [sym__backtick_quote] = ACTIONS(2948),
    [sym__search_slash_quote] = ACTIONS(2948),
    [sym__open_readline_bracket] = ACTIONS(2948),
    [sym__open_fileglob_bracket] = ACTIONS(2948),
    [sym__PERLY_HEREDOC] = ACTIONS(2948),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2948),
    [sym__filetest] = ACTIONS(2948),
  },
  [STATE(1701)] = {
    [sym_heredoc_content] = STATE(1701),
    [sym__identifier] = ACTIONS(3025),
    [anon_sym_LBRACE] = ACTIONS(3025),
    [anon_sym_DOLLAR] = ACTIONS(3025),
    [anon_sym_AT] = ACTIONS(3025),
    [anon_sym_RPAREN] = ACTIONS(3025),
    [anon_sym_my] = ACTIONS(3025),
    [anon_sym_extended] = ACTIONS(3025),
    [anon_sym_async] = ACTIONS(3025),
    [anon_sym_sub] = ACTIONS(3025),
    [anon_sym_method] = ACTIONS(3025),
    [anon_sym_LPAREN] = ACTIONS(3025),
    [anon_sym_continue] = ACTIONS(3025),
    [anon_sym_state] = ACTIONS(3025),
    [anon_sym_our] = ACTIONS(3025),
    [anon_sym_LBRACK] = ACTIONS(3025),
    [anon_sym_LT_LT] = ACTIONS(3025),
    [anon_sym_SLASH_SLASH] = ACTIONS(3025),
    [anon_sym_PLUS] = ACTIONS(3025),
    [anon_sym_DASH] = ACTIONS(3025),
    [anon_sym_TILDE] = ACTIONS(3025),
    [anon_sym_BANG] = ACTIONS(3025),
    [anon_sym_PLUS_PLUS] = ACTIONS(3025),
    [anon_sym_DASH_DASH] = ACTIONS(3025),
    [anon_sym_BSLASH] = ACTIONS(3023),
    [anon_sym_do] = ACTIONS(3025),
    [anon_sym_eval] = ACTIONS(3025),
    [anon_sym_await] = ACTIONS(3025),
    [anon_sym_field] = ACTIONS(3025),
    [anon_sym_local] = ACTIONS(3025),
    [anon_sym_dynamically] = ACTIONS(3025),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3025),
    [anon_sym_require] = ACTIONS(3025),
    [anon_sym_map] = ACTIONS(3025),
    [anon_sym_grep] = ACTIONS(3025),
    [anon_sym_sort] = ACTIONS(3025),
    [anon_sym_goto] = ACTIONS(3025),
    [anon_sym_return] = ACTIONS(3025),
    [anon_sym_undef] = ACTIONS(3025),
    [anon_sym_PERCENT2] = ACTIONS(3025),
    [anon_sym_AMP2] = ACTIONS(3025),
    [anon_sym_STAR2] = ACTIONS(3025),
    [anon_sym_last] = ACTIONS(3025),
    [anon_sym_next] = ACTIONS(3025),
    [anon_sym_redo] = ACTIONS(3025),
    [anon_sym___FILE__] = ACTIONS(3025),
    [anon_sym___LINE__] = ACTIONS(3025),
    [anon_sym___PACKAGE__] = ACTIONS(3025),
    [anon_sym___SUB__] = ACTIONS(3025),
    [anon_sym_break] = ACTIONS(3025),
    [anon_sym_fork] = ACTIONS(3025),
    [anon_sym_getppid] = ACTIONS(3025),
    [anon_sym_time] = ACTIONS(3025),
    [anon_sym_times] = ACTIONS(3025),
    [anon_sym_wait] = ACTIONS(3025),
    [anon_sym_wantarray] = ACTIONS(3025),
    [anon_sym_abs] = ACTIONS(3025),
    [anon_sym_alarm] = ACTIONS(3025),
    [anon_sym_chop] = ACTIONS(3025),
    [anon_sym_chdir] = ACTIONS(3025),
    [anon_sym_close] = ACTIONS(3025),
    [anon_sym_closedir] = ACTIONS(3025),
    [anon_sym_caller] = ACTIONS(3025),
    [anon_sym_chomp] = ACTIONS(3025),
    [anon_sym_chr] = ACTIONS(3025),
    [anon_sym_cos] = ACTIONS(3025),
    [anon_sym_chroot] = ACTIONS(3025),
    [anon_sym_defined] = ACTIONS(3025),
    [anon_sym_delete] = ACTIONS(3025),
    [anon_sym_dbmclose] = ACTIONS(3025),
    [anon_sym_exists] = ACTIONS(3025),
    [anon_sym_exit] = ACTIONS(3025),
    [anon_sym_eof] = ACTIONS(3025),
    [anon_sym_exp] = ACTIONS(3025),
    [anon_sym_each] = ACTIONS(3025),
    [anon_sym_fc] = ACTIONS(3025),
    [anon_sym_fileno] = ACTIONS(3025),
    [anon_sym_gmtime] = ACTIONS(3025),
    [anon_sym_getc] = ACTIONS(3025),
    [anon_sym_getpgrp] = ACTIONS(3025),
    [anon_sym_getprotobyname] = ACTIONS(3025),
    [anon_sym_getpwname] = ACTIONS(3025),
    [anon_sym_getpwuid] = ACTIONS(3025),
    [anon_sym_getpeername] = ACTIONS(3025),
    [anon_sym_getnetbyname] = ACTIONS(3025),
    [anon_sym_getsockname] = ACTIONS(3025),
    [anon_sym_getgrnam] = ACTIONS(3025),
    [anon_sym_getgrgid] = ACTIONS(3025),
    [anon_sym_hex] = ACTIONS(3025),
    [anon_sym_int] = ACTIONS(3025),
    [anon_sym_keys] = ACTIONS(3025),
    [anon_sym_lc] = ACTIONS(3025),
    [anon_sym_lcfirst] = ACTIONS(3025),
    [anon_sym_length] = ACTIONS(3025),
    [anon_sym_localtime] = ACTIONS(3025),
    [anon_sym_log] = ACTIONS(3025),
    [anon_sym_lock] = ACTIONS(3025),
    [anon_sym_lstat] = ACTIONS(3025),
    [anon_sym_oct] = ACTIONS(3025),
    [anon_sym_ord] = ACTIONS(3025),
    [anon_sym_prototype] = ACTIONS(3025),
    [anon_sym_pop] = ACTIONS(3025),
    [anon_sym_pos] = ACTIONS(3025),
    [anon_sym_quotemeta] = ACTIONS(3025),
    [anon_sym_reset] = ACTIONS(3025),
    [anon_sym_rand] = ACTIONS(3025),
    [anon_sym_rmdir] = ACTIONS(3025),
    [anon_sym_readdir] = ACTIONS(3025),
    [anon_sym_readline] = ACTIONS(3025),
    [anon_sym_readpipe] = ACTIONS(3025),
    [anon_sym_rewinddir] = ACTIONS(3025),
    [anon_sym_readlink] = ACTIONS(3025),
    [anon_sym_ref] = ACTIONS(3025),
    [anon_sym_scalar] = ACTIONS(3025),
    [anon_sym_shift] = ACTIONS(3025),
    [anon_sym_sin] = ACTIONS(3025),
    [anon_sym_sleep] = ACTIONS(3025),
    [anon_sym_sqrt] = ACTIONS(3025),
    [anon_sym_srand] = ACTIONS(3025),
    [anon_sym_stat] = ACTIONS(3025),
    [anon_sym_study] = ACTIONS(3025),
    [anon_sym_tell] = ACTIONS(3025),
    [anon_sym_telldir] = ACTIONS(3025),
    [anon_sym_tied] = ACTIONS(3025),
    [anon_sym_uc] = ACTIONS(3025),
    [anon_sym_ucfirst] = ACTIONS(3025),
    [anon_sym_untie] = ACTIONS(3025),
    [anon_sym_umask] = ACTIONS(3025),
    [anon_sym_values] = ACTIONS(3025),
    [anon_sym_write] = ACTIONS(3025),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3025),
    [anon_sym_qq] = ACTIONS(3025),
    [anon_sym_qw] = ACTIONS(3025),
    [anon_sym_qx] = ACTIONS(3025),
    [anon_sym_qr] = ACTIONS(3025),
    [anon_sym_m] = ACTIONS(3025),
    [anon_sym_s] = ACTIONS(3025),
    [anon_sym_tr] = ACTIONS(3025),
    [anon_sym_y] = ACTIONS(3025),
    [aux_sym__bareword_token1] = ACTIONS(3025),
    [sym_number] = ACTIONS(3025),
    [anon_sym_true] = ACTIONS(3025),
    [anon_sym_false] = ACTIONS(3025),
    [sym__single_quote] = ACTIONS(3023),
    [sym__double_quote] = ACTIONS(3023),
    [sym__backtick_quote] = ACTIONS(3023),
    [sym__search_slash_quote] = ACTIONS(3023),
    [sym__open_readline_bracket] = ACTIONS(3023),
    [sym__open_fileglob_bracket] = ACTIONS(3023),
    [sym__PERLY_HEREDOC] = ACTIONS(3023),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3023),
    [sym__filetest] = ACTIONS(3023),
  },
  [STATE(1702)] = {
    [sym_heredoc_content] = STATE(1702),
    [sym__identifier] = ACTIONS(3053),
    [anon_sym_LBRACE] = ACTIONS(3053),
    [anon_sym_DOLLAR] = ACTIONS(3053),
    [anon_sym_AT] = ACTIONS(3053),
    [anon_sym_RPAREN] = ACTIONS(3053),
    [anon_sym_my] = ACTIONS(3053),
    [anon_sym_extended] = ACTIONS(3053),
    [anon_sym_async] = ACTIONS(3053),
    [anon_sym_sub] = ACTIONS(3053),
    [anon_sym_method] = ACTIONS(3053),
    [anon_sym_LPAREN] = ACTIONS(3053),
    [anon_sym_continue] = ACTIONS(3053),
    [anon_sym_state] = ACTIONS(3053),
    [anon_sym_our] = ACTIONS(3053),
    [anon_sym_LBRACK] = ACTIONS(3053),
    [anon_sym_LT_LT] = ACTIONS(3053),
    [anon_sym_SLASH_SLASH] = ACTIONS(3053),
    [anon_sym_PLUS] = ACTIONS(3053),
    [anon_sym_DASH] = ACTIONS(3053),
    [anon_sym_TILDE] = ACTIONS(3053),
    [anon_sym_BANG] = ACTIONS(3053),
    [anon_sym_PLUS_PLUS] = ACTIONS(3053),
    [anon_sym_DASH_DASH] = ACTIONS(3053),
    [anon_sym_BSLASH] = ACTIONS(3051),
    [anon_sym_do] = ACTIONS(3053),
    [anon_sym_eval] = ACTIONS(3053),
    [anon_sym_await] = ACTIONS(3053),
    [anon_sym_field] = ACTIONS(3053),
    [anon_sym_local] = ACTIONS(3053),
    [anon_sym_dynamically] = ACTIONS(3053),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3053),
    [anon_sym_require] = ACTIONS(3053),
    [anon_sym_map] = ACTIONS(3053),
    [anon_sym_grep] = ACTIONS(3053),
    [anon_sym_sort] = ACTIONS(3053),
    [anon_sym_goto] = ACTIONS(3053),
    [anon_sym_return] = ACTIONS(3053),
    [anon_sym_undef] = ACTIONS(3053),
    [anon_sym_PERCENT2] = ACTIONS(3053),
    [anon_sym_AMP2] = ACTIONS(3053),
    [anon_sym_STAR2] = ACTIONS(3053),
    [anon_sym_last] = ACTIONS(3053),
    [anon_sym_next] = ACTIONS(3053),
    [anon_sym_redo] = ACTIONS(3053),
    [anon_sym___FILE__] = ACTIONS(3053),
    [anon_sym___LINE__] = ACTIONS(3053),
    [anon_sym___PACKAGE__] = ACTIONS(3053),
    [anon_sym___SUB__] = ACTIONS(3053),
    [anon_sym_break] = ACTIONS(3053),
    [anon_sym_fork] = ACTIONS(3053),
    [anon_sym_getppid] = ACTIONS(3053),
    [anon_sym_time] = ACTIONS(3053),
    [anon_sym_times] = ACTIONS(3053),
    [anon_sym_wait] = ACTIONS(3053),
    [anon_sym_wantarray] = ACTIONS(3053),
    [anon_sym_abs] = ACTIONS(3053),
    [anon_sym_alarm] = ACTIONS(3053),
    [anon_sym_chop] = ACTIONS(3053),
    [anon_sym_chdir] = ACTIONS(3053),
    [anon_sym_close] = ACTIONS(3053),
    [anon_sym_closedir] = ACTIONS(3053),
    [anon_sym_caller] = ACTIONS(3053),
    [anon_sym_chomp] = ACTIONS(3053),
    [anon_sym_chr] = ACTIONS(3053),
    [anon_sym_cos] = ACTIONS(3053),
    [anon_sym_chroot] = ACTIONS(3053),
    [anon_sym_defined] = ACTIONS(3053),
    [anon_sym_delete] = ACTIONS(3053),
    [anon_sym_dbmclose] = ACTIONS(3053),
    [anon_sym_exists] = ACTIONS(3053),
    [anon_sym_exit] = ACTIONS(3053),
    [anon_sym_eof] = ACTIONS(3053),
    [anon_sym_exp] = ACTIONS(3053),
    [anon_sym_each] = ACTIONS(3053),
    [anon_sym_fc] = ACTIONS(3053),
    [anon_sym_fileno] = ACTIONS(3053),
    [anon_sym_gmtime] = ACTIONS(3053),
    [anon_sym_getc] = ACTIONS(3053),
    [anon_sym_getpgrp] = ACTIONS(3053),
    [anon_sym_getprotobyname] = ACTIONS(3053),
    [anon_sym_getpwname] = ACTIONS(3053),
    [anon_sym_getpwuid] = ACTIONS(3053),
    [anon_sym_getpeername] = ACTIONS(3053),
    [anon_sym_getnetbyname] = ACTIONS(3053),
    [anon_sym_getsockname] = ACTIONS(3053),
    [anon_sym_getgrnam] = ACTIONS(3053),
    [anon_sym_getgrgid] = ACTIONS(3053),
    [anon_sym_hex] = ACTIONS(3053),
    [anon_sym_int] = ACTIONS(3053),
    [anon_sym_keys] = ACTIONS(3053),
    [anon_sym_lc] = ACTIONS(3053),
    [anon_sym_lcfirst] = ACTIONS(3053),
    [anon_sym_length] = ACTIONS(3053),
    [anon_sym_localtime] = ACTIONS(3053),
    [anon_sym_log] = ACTIONS(3053),
    [anon_sym_lock] = ACTIONS(3053),
    [anon_sym_lstat] = ACTIONS(3053),
    [anon_sym_oct] = ACTIONS(3053),
    [anon_sym_ord] = ACTIONS(3053),
    [anon_sym_prototype] = ACTIONS(3053),
    [anon_sym_pop] = ACTIONS(3053),
    [anon_sym_pos] = ACTIONS(3053),
    [anon_sym_quotemeta] = ACTIONS(3053),
    [anon_sym_reset] = ACTIONS(3053),
    [anon_sym_rand] = ACTIONS(3053),
    [anon_sym_rmdir] = ACTIONS(3053),
    [anon_sym_readdir] = ACTIONS(3053),
    [anon_sym_readline] = ACTIONS(3053),
    [anon_sym_readpipe] = ACTIONS(3053),
    [anon_sym_rewinddir] = ACTIONS(3053),
    [anon_sym_readlink] = ACTIONS(3053),
    [anon_sym_ref] = ACTIONS(3053),
    [anon_sym_scalar] = ACTIONS(3053),
    [anon_sym_shift] = ACTIONS(3053),
    [anon_sym_sin] = ACTIONS(3053),
    [anon_sym_sleep] = ACTIONS(3053),
    [anon_sym_sqrt] = ACTIONS(3053),
    [anon_sym_srand] = ACTIONS(3053),
    [anon_sym_stat] = ACTIONS(3053),
    [anon_sym_study] = ACTIONS(3053),
    [anon_sym_tell] = ACTIONS(3053),
    [anon_sym_telldir] = ACTIONS(3053),
    [anon_sym_tied] = ACTIONS(3053),
    [anon_sym_uc] = ACTIONS(3053),
    [anon_sym_ucfirst] = ACTIONS(3053),
    [anon_sym_untie] = ACTIONS(3053),
    [anon_sym_umask] = ACTIONS(3053),
    [anon_sym_values] = ACTIONS(3053),
    [anon_sym_write] = ACTIONS(3053),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3053),
    [anon_sym_qq] = ACTIONS(3053),
    [anon_sym_qw] = ACTIONS(3053),
    [anon_sym_qx] = ACTIONS(3053),
    [anon_sym_qr] = ACTIONS(3053),
    [anon_sym_m] = ACTIONS(3053),
    [anon_sym_s] = ACTIONS(3053),
    [anon_sym_tr] = ACTIONS(3053),
    [anon_sym_y] = ACTIONS(3053),
    [aux_sym__bareword_token1] = ACTIONS(3053),
    [sym_number] = ACTIONS(3053),
    [anon_sym_true] = ACTIONS(3053),
    [anon_sym_false] = ACTIONS(3053),
    [sym__single_quote] = ACTIONS(3051),
    [sym__double_quote] = ACTIONS(3051),
    [sym__backtick_quote] = ACTIONS(3051),
    [sym__search_slash_quote] = ACTIONS(3051),
    [sym__open_readline_bracket] = ACTIONS(3051),
    [sym__open_fileglob_bracket] = ACTIONS(3051),
    [sym__PERLY_HEREDOC] = ACTIONS(3051),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3051),
    [sym__filetest] = ACTIONS(3051),
  },
  [STATE(1703)] = {
    [sym_heredoc_content] = STATE(1703),
    [sym__identifier] = ACTIONS(3169),
    [anon_sym_LBRACE] = ACTIONS(3169),
    [anon_sym_DOLLAR] = ACTIONS(3169),
    [anon_sym_AT] = ACTIONS(3169),
    [anon_sym_RPAREN] = ACTIONS(3169),
    [anon_sym_my] = ACTIONS(3169),
    [anon_sym_extended] = ACTIONS(3169),
    [anon_sym_async] = ACTIONS(3169),
    [anon_sym_sub] = ACTIONS(3169),
    [anon_sym_method] = ACTIONS(3169),
    [anon_sym_LPAREN] = ACTIONS(3169),
    [anon_sym_continue] = ACTIONS(3169),
    [anon_sym_state] = ACTIONS(3169),
    [anon_sym_our] = ACTIONS(3169),
    [anon_sym_LBRACK] = ACTIONS(3169),
    [anon_sym_LT_LT] = ACTIONS(3169),
    [anon_sym_SLASH_SLASH] = ACTIONS(3169),
    [anon_sym_PLUS] = ACTIONS(3169),
    [anon_sym_DASH] = ACTIONS(3169),
    [anon_sym_TILDE] = ACTIONS(3169),
    [anon_sym_BANG] = ACTIONS(3169),
    [anon_sym_PLUS_PLUS] = ACTIONS(3169),
    [anon_sym_DASH_DASH] = ACTIONS(3169),
    [anon_sym_BSLASH] = ACTIONS(3167),
    [anon_sym_do] = ACTIONS(3169),
    [anon_sym_eval] = ACTIONS(3169),
    [anon_sym_await] = ACTIONS(3169),
    [anon_sym_field] = ACTIONS(3169),
    [anon_sym_local] = ACTIONS(3169),
    [anon_sym_dynamically] = ACTIONS(3169),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3169),
    [anon_sym_require] = ACTIONS(3169),
    [anon_sym_map] = ACTIONS(3169),
    [anon_sym_grep] = ACTIONS(3169),
    [anon_sym_sort] = ACTIONS(3169),
    [anon_sym_goto] = ACTIONS(3169),
    [anon_sym_return] = ACTIONS(3169),
    [anon_sym_undef] = ACTIONS(3169),
    [anon_sym_PERCENT2] = ACTIONS(3169),
    [anon_sym_AMP2] = ACTIONS(3169),
    [anon_sym_STAR2] = ACTIONS(3169),
    [anon_sym_last] = ACTIONS(3169),
    [anon_sym_next] = ACTIONS(3169),
    [anon_sym_redo] = ACTIONS(3169),
    [anon_sym___FILE__] = ACTIONS(3169),
    [anon_sym___LINE__] = ACTIONS(3169),
    [anon_sym___PACKAGE__] = ACTIONS(3169),
    [anon_sym___SUB__] = ACTIONS(3169),
    [anon_sym_break] = ACTIONS(3169),
    [anon_sym_fork] = ACTIONS(3169),
    [anon_sym_getppid] = ACTIONS(3169),
    [anon_sym_time] = ACTIONS(3169),
    [anon_sym_times] = ACTIONS(3169),
    [anon_sym_wait] = ACTIONS(3169),
    [anon_sym_wantarray] = ACTIONS(3169),
    [anon_sym_abs] = ACTIONS(3169),
    [anon_sym_alarm] = ACTIONS(3169),
    [anon_sym_chop] = ACTIONS(3169),
    [anon_sym_chdir] = ACTIONS(3169),
    [anon_sym_close] = ACTIONS(3169),
    [anon_sym_closedir] = ACTIONS(3169),
    [anon_sym_caller] = ACTIONS(3169),
    [anon_sym_chomp] = ACTIONS(3169),
    [anon_sym_chr] = ACTIONS(3169),
    [anon_sym_cos] = ACTIONS(3169),
    [anon_sym_chroot] = ACTIONS(3169),
    [anon_sym_defined] = ACTIONS(3169),
    [anon_sym_delete] = ACTIONS(3169),
    [anon_sym_dbmclose] = ACTIONS(3169),
    [anon_sym_exists] = ACTIONS(3169),
    [anon_sym_exit] = ACTIONS(3169),
    [anon_sym_eof] = ACTIONS(3169),
    [anon_sym_exp] = ACTIONS(3169),
    [anon_sym_each] = ACTIONS(3169),
    [anon_sym_fc] = ACTIONS(3169),
    [anon_sym_fileno] = ACTIONS(3169),
    [anon_sym_gmtime] = ACTIONS(3169),
    [anon_sym_getc] = ACTIONS(3169),
    [anon_sym_getpgrp] = ACTIONS(3169),
    [anon_sym_getprotobyname] = ACTIONS(3169),
    [anon_sym_getpwname] = ACTIONS(3169),
    [anon_sym_getpwuid] = ACTIONS(3169),
    [anon_sym_getpeername] = ACTIONS(3169),
    [anon_sym_getnetbyname] = ACTIONS(3169),
    [anon_sym_getsockname] = ACTIONS(3169),
    [anon_sym_getgrnam] = ACTIONS(3169),
    [anon_sym_getgrgid] = ACTIONS(3169),
    [anon_sym_hex] = ACTIONS(3169),
    [anon_sym_int] = ACTIONS(3169),
    [anon_sym_keys] = ACTIONS(3169),
    [anon_sym_lc] = ACTIONS(3169),
    [anon_sym_lcfirst] = ACTIONS(3169),
    [anon_sym_length] = ACTIONS(3169),
    [anon_sym_localtime] = ACTIONS(3169),
    [anon_sym_log] = ACTIONS(3169),
    [anon_sym_lock] = ACTIONS(3169),
    [anon_sym_lstat] = ACTIONS(3169),
    [anon_sym_oct] = ACTIONS(3169),
    [anon_sym_ord] = ACTIONS(3169),
    [anon_sym_prototype] = ACTIONS(3169),
    [anon_sym_pop] = ACTIONS(3169),
    [anon_sym_pos] = ACTIONS(3169),
    [anon_sym_quotemeta] = ACTIONS(3169),
    [anon_sym_reset] = ACTIONS(3169),
    [anon_sym_rand] = ACTIONS(3169),
    [anon_sym_rmdir] = ACTIONS(3169),
    [anon_sym_readdir] = ACTIONS(3169),
    [anon_sym_readline] = ACTIONS(3169),
    [anon_sym_readpipe] = ACTIONS(3169),
    [anon_sym_rewinddir] = ACTIONS(3169),
    [anon_sym_readlink] = ACTIONS(3169),
    [anon_sym_ref] = ACTIONS(3169),
    [anon_sym_scalar] = ACTIONS(3169),
    [anon_sym_shift] = ACTIONS(3169),
    [anon_sym_sin] = ACTIONS(3169),
    [anon_sym_sleep] = ACTIONS(3169),
    [anon_sym_sqrt] = ACTIONS(3169),
    [anon_sym_srand] = ACTIONS(3169),
    [anon_sym_stat] = ACTIONS(3169),
    [anon_sym_study] = ACTIONS(3169),
    [anon_sym_tell] = ACTIONS(3169),
    [anon_sym_telldir] = ACTIONS(3169),
    [anon_sym_tied] = ACTIONS(3169),
    [anon_sym_uc] = ACTIONS(3169),
    [anon_sym_ucfirst] = ACTIONS(3169),
    [anon_sym_untie] = ACTIONS(3169),
    [anon_sym_umask] = ACTIONS(3169),
    [anon_sym_values] = ACTIONS(3169),
    [anon_sym_write] = ACTIONS(3169),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3169),
    [anon_sym_qq] = ACTIONS(3169),
    [anon_sym_qw] = ACTIONS(3169),
    [anon_sym_qx] = ACTIONS(3169),
    [anon_sym_qr] = ACTIONS(3169),
    [anon_sym_m] = ACTIONS(3169),
    [anon_sym_s] = ACTIONS(3169),
    [anon_sym_tr] = ACTIONS(3169),
    [anon_sym_y] = ACTIONS(3169),
    [aux_sym__bareword_token1] = ACTIONS(3169),
    [sym_number] = ACTIONS(3169),
    [anon_sym_true] = ACTIONS(3169),
    [anon_sym_false] = ACTIONS(3169),
    [sym__single_quote] = ACTIONS(3167),
    [sym__double_quote] = ACTIONS(3167),
    [sym__backtick_quote] = ACTIONS(3167),
    [sym__search_slash_quote] = ACTIONS(3167),
    [sym__open_readline_bracket] = ACTIONS(3167),
    [sym__open_fileglob_bracket] = ACTIONS(3167),
    [sym__PERLY_HEREDOC] = ACTIONS(3167),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3167),
    [sym__filetest] = ACTIONS(3167),
  },
  [STATE(1704)] = {
    [sym_heredoc_content] = STATE(1704),
    [sym__identifier] = ACTIONS(3097),
    [anon_sym_LBRACE] = ACTIONS(3097),
    [anon_sym_DOLLAR] = ACTIONS(3097),
    [anon_sym_AT] = ACTIONS(3097),
    [anon_sym_RPAREN] = ACTIONS(3097),
    [anon_sym_my] = ACTIONS(3097),
    [anon_sym_extended] = ACTIONS(3097),
    [anon_sym_async] = ACTIONS(3097),
    [anon_sym_sub] = ACTIONS(3097),
    [anon_sym_method] = ACTIONS(3097),
    [anon_sym_LPAREN] = ACTIONS(3097),
    [anon_sym_continue] = ACTIONS(3097),
    [anon_sym_state] = ACTIONS(3097),
    [anon_sym_our] = ACTIONS(3097),
    [anon_sym_LBRACK] = ACTIONS(3097),
    [anon_sym_LT_LT] = ACTIONS(3097),
    [anon_sym_SLASH_SLASH] = ACTIONS(3097),
    [anon_sym_PLUS] = ACTIONS(3097),
    [anon_sym_DASH] = ACTIONS(3097),
    [anon_sym_TILDE] = ACTIONS(3097),
    [anon_sym_BANG] = ACTIONS(3097),
    [anon_sym_PLUS_PLUS] = ACTIONS(3097),
    [anon_sym_DASH_DASH] = ACTIONS(3097),
    [anon_sym_BSLASH] = ACTIONS(3095),
    [anon_sym_do] = ACTIONS(3097),
    [anon_sym_eval] = ACTIONS(3097),
    [anon_sym_await] = ACTIONS(3097),
    [anon_sym_field] = ACTIONS(3097),
    [anon_sym_local] = ACTIONS(3097),
    [anon_sym_dynamically] = ACTIONS(3097),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3097),
    [anon_sym_require] = ACTIONS(3097),
    [anon_sym_map] = ACTIONS(3097),
    [anon_sym_grep] = ACTIONS(3097),
    [anon_sym_sort] = ACTIONS(3097),
    [anon_sym_goto] = ACTIONS(3097),
    [anon_sym_return] = ACTIONS(3097),
    [anon_sym_undef] = ACTIONS(3097),
    [anon_sym_PERCENT2] = ACTIONS(3097),
    [anon_sym_AMP2] = ACTIONS(3097),
    [anon_sym_STAR2] = ACTIONS(3097),
    [anon_sym_last] = ACTIONS(3097),
    [anon_sym_next] = ACTIONS(3097),
    [anon_sym_redo] = ACTIONS(3097),
    [anon_sym___FILE__] = ACTIONS(3097),
    [anon_sym___LINE__] = ACTIONS(3097),
    [anon_sym___PACKAGE__] = ACTIONS(3097),
    [anon_sym___SUB__] = ACTIONS(3097),
    [anon_sym_break] = ACTIONS(3097),
    [anon_sym_fork] = ACTIONS(3097),
    [anon_sym_getppid] = ACTIONS(3097),
    [anon_sym_time] = ACTIONS(3097),
    [anon_sym_times] = ACTIONS(3097),
    [anon_sym_wait] = ACTIONS(3097),
    [anon_sym_wantarray] = ACTIONS(3097),
    [anon_sym_abs] = ACTIONS(3097),
    [anon_sym_alarm] = ACTIONS(3097),
    [anon_sym_chop] = ACTIONS(3097),
    [anon_sym_chdir] = ACTIONS(3097),
    [anon_sym_close] = ACTIONS(3097),
    [anon_sym_closedir] = ACTIONS(3097),
    [anon_sym_caller] = ACTIONS(3097),
    [anon_sym_chomp] = ACTIONS(3097),
    [anon_sym_chr] = ACTIONS(3097),
    [anon_sym_cos] = ACTIONS(3097),
    [anon_sym_chroot] = ACTIONS(3097),
    [anon_sym_defined] = ACTIONS(3097),
    [anon_sym_delete] = ACTIONS(3097),
    [anon_sym_dbmclose] = ACTIONS(3097),
    [anon_sym_exists] = ACTIONS(3097),
    [anon_sym_exit] = ACTIONS(3097),
    [anon_sym_eof] = ACTIONS(3097),
    [anon_sym_exp] = ACTIONS(3097),
    [anon_sym_each] = ACTIONS(3097),
    [anon_sym_fc] = ACTIONS(3097),
    [anon_sym_fileno] = ACTIONS(3097),
    [anon_sym_gmtime] = ACTIONS(3097),
    [anon_sym_getc] = ACTIONS(3097),
    [anon_sym_getpgrp] = ACTIONS(3097),
    [anon_sym_getprotobyname] = ACTIONS(3097),
    [anon_sym_getpwname] = ACTIONS(3097),
    [anon_sym_getpwuid] = ACTIONS(3097),
    [anon_sym_getpeername] = ACTIONS(3097),
    [anon_sym_getnetbyname] = ACTIONS(3097),
    [anon_sym_getsockname] = ACTIONS(3097),
    [anon_sym_getgrnam] = ACTIONS(3097),
    [anon_sym_getgrgid] = ACTIONS(3097),
    [anon_sym_hex] = ACTIONS(3097),
    [anon_sym_int] = ACTIONS(3097),
    [anon_sym_keys] = ACTIONS(3097),
    [anon_sym_lc] = ACTIONS(3097),
    [anon_sym_lcfirst] = ACTIONS(3097),
    [anon_sym_length] = ACTIONS(3097),
    [anon_sym_localtime] = ACTIONS(3097),
    [anon_sym_log] = ACTIONS(3097),
    [anon_sym_lock] = ACTIONS(3097),
    [anon_sym_lstat] = ACTIONS(3097),
    [anon_sym_oct] = ACTIONS(3097),
    [anon_sym_ord] = ACTIONS(3097),
    [anon_sym_prototype] = ACTIONS(3097),
    [anon_sym_pop] = ACTIONS(3097),
    [anon_sym_pos] = ACTIONS(3097),
    [anon_sym_quotemeta] = ACTIONS(3097),
    [anon_sym_reset] = ACTIONS(3097),
    [anon_sym_rand] = ACTIONS(3097),
    [anon_sym_rmdir] = ACTIONS(3097),
    [anon_sym_readdir] = ACTIONS(3097),
    [anon_sym_readline] = ACTIONS(3097),
    [anon_sym_readpipe] = ACTIONS(3097),
    [anon_sym_rewinddir] = ACTIONS(3097),
    [anon_sym_readlink] = ACTIONS(3097),
    [anon_sym_ref] = ACTIONS(3097),
    [anon_sym_scalar] = ACTIONS(3097),
    [anon_sym_shift] = ACTIONS(3097),
    [anon_sym_sin] = ACTIONS(3097),
    [anon_sym_sleep] = ACTIONS(3097),
    [anon_sym_sqrt] = ACTIONS(3097),
    [anon_sym_srand] = ACTIONS(3097),
    [anon_sym_stat] = ACTIONS(3097),
    [anon_sym_study] = ACTIONS(3097),
    [anon_sym_tell] = ACTIONS(3097),
    [anon_sym_telldir] = ACTIONS(3097),
    [anon_sym_tied] = ACTIONS(3097),
    [anon_sym_uc] = ACTIONS(3097),
    [anon_sym_ucfirst] = ACTIONS(3097),
    [anon_sym_untie] = ACTIONS(3097),
    [anon_sym_umask] = ACTIONS(3097),
    [anon_sym_values] = ACTIONS(3097),
    [anon_sym_write] = ACTIONS(3097),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3097),
    [anon_sym_qq] = ACTIONS(3097),
    [anon_sym_qw] = ACTIONS(3097),
    [anon_sym_qx] = ACTIONS(3097),
    [anon_sym_qr] = ACTIONS(3097),
    [anon_sym_m] = ACTIONS(3097),
    [anon_sym_s] = ACTIONS(3097),
    [anon_sym_tr] = ACTIONS(3097),
    [anon_sym_y] = ACTIONS(3097),
    [aux_sym__bareword_token1] = ACTIONS(3097),
    [sym_number] = ACTIONS(3097),
    [anon_sym_true] = ACTIONS(3097),
    [anon_sym_false] = ACTIONS(3097),
    [sym__single_quote] = ACTIONS(3095),
    [sym__double_quote] = ACTIONS(3095),
    [sym__backtick_quote] = ACTIONS(3095),
    [sym__search_slash_quote] = ACTIONS(3095),
    [sym__open_readline_bracket] = ACTIONS(3095),
    [sym__open_fileglob_bracket] = ACTIONS(3095),
    [sym__PERLY_HEREDOC] = ACTIONS(3095),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3095),
    [sym__filetest] = ACTIONS(3095),
  },
  [STATE(1705)] = {
    [sym_heredoc_content] = STATE(1705),
    [sym__identifier] = ACTIONS(3109),
    [anon_sym_LBRACE] = ACTIONS(3109),
    [anon_sym_DOLLAR] = ACTIONS(3109),
    [anon_sym_AT] = ACTIONS(3109),
    [anon_sym_RPAREN] = ACTIONS(3109),
    [anon_sym_my] = ACTIONS(3109),
    [anon_sym_extended] = ACTIONS(3109),
    [anon_sym_async] = ACTIONS(3109),
    [anon_sym_sub] = ACTIONS(3109),
    [anon_sym_method] = ACTIONS(3109),
    [anon_sym_LPAREN] = ACTIONS(3109),
    [anon_sym_continue] = ACTIONS(3109),
    [anon_sym_state] = ACTIONS(3109),
    [anon_sym_our] = ACTIONS(3109),
    [anon_sym_LBRACK] = ACTIONS(3109),
    [anon_sym_LT_LT] = ACTIONS(3109),
    [anon_sym_SLASH_SLASH] = ACTIONS(3109),
    [anon_sym_PLUS] = ACTIONS(3109),
    [anon_sym_DASH] = ACTIONS(3109),
    [anon_sym_TILDE] = ACTIONS(3109),
    [anon_sym_BANG] = ACTIONS(3109),
    [anon_sym_PLUS_PLUS] = ACTIONS(3109),
    [anon_sym_DASH_DASH] = ACTIONS(3109),
    [anon_sym_BSLASH] = ACTIONS(3107),
    [anon_sym_do] = ACTIONS(3109),
    [anon_sym_eval] = ACTIONS(3109),
    [anon_sym_await] = ACTIONS(3109),
    [anon_sym_field] = ACTIONS(3109),
    [anon_sym_local] = ACTIONS(3109),
    [anon_sym_dynamically] = ACTIONS(3109),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3109),
    [anon_sym_require] = ACTIONS(3109),
    [anon_sym_map] = ACTIONS(3109),
    [anon_sym_grep] = ACTIONS(3109),
    [anon_sym_sort] = ACTIONS(3109),
    [anon_sym_goto] = ACTIONS(3109),
    [anon_sym_return] = ACTIONS(3109),
    [anon_sym_undef] = ACTIONS(3109),
    [anon_sym_PERCENT2] = ACTIONS(3109),
    [anon_sym_AMP2] = ACTIONS(3109),
    [anon_sym_STAR2] = ACTIONS(3109),
    [anon_sym_last] = ACTIONS(3109),
    [anon_sym_next] = ACTIONS(3109),
    [anon_sym_redo] = ACTIONS(3109),
    [anon_sym___FILE__] = ACTIONS(3109),
    [anon_sym___LINE__] = ACTIONS(3109),
    [anon_sym___PACKAGE__] = ACTIONS(3109),
    [anon_sym___SUB__] = ACTIONS(3109),
    [anon_sym_break] = ACTIONS(3109),
    [anon_sym_fork] = ACTIONS(3109),
    [anon_sym_getppid] = ACTIONS(3109),
    [anon_sym_time] = ACTIONS(3109),
    [anon_sym_times] = ACTIONS(3109),
    [anon_sym_wait] = ACTIONS(3109),
    [anon_sym_wantarray] = ACTIONS(3109),
    [anon_sym_abs] = ACTIONS(3109),
    [anon_sym_alarm] = ACTIONS(3109),
    [anon_sym_chop] = ACTIONS(3109),
    [anon_sym_chdir] = ACTIONS(3109),
    [anon_sym_close] = ACTIONS(3109),
    [anon_sym_closedir] = ACTIONS(3109),
    [anon_sym_caller] = ACTIONS(3109),
    [anon_sym_chomp] = ACTIONS(3109),
    [anon_sym_chr] = ACTIONS(3109),
    [anon_sym_cos] = ACTIONS(3109),
    [anon_sym_chroot] = ACTIONS(3109),
    [anon_sym_defined] = ACTIONS(3109),
    [anon_sym_delete] = ACTIONS(3109),
    [anon_sym_dbmclose] = ACTIONS(3109),
    [anon_sym_exists] = ACTIONS(3109),
    [anon_sym_exit] = ACTIONS(3109),
    [anon_sym_eof] = ACTIONS(3109),
    [anon_sym_exp] = ACTIONS(3109),
    [anon_sym_each] = ACTIONS(3109),
    [anon_sym_fc] = ACTIONS(3109),
    [anon_sym_fileno] = ACTIONS(3109),
    [anon_sym_gmtime] = ACTIONS(3109),
    [anon_sym_getc] = ACTIONS(3109),
    [anon_sym_getpgrp] = ACTIONS(3109),
    [anon_sym_getprotobyname] = ACTIONS(3109),
    [anon_sym_getpwname] = ACTIONS(3109),
    [anon_sym_getpwuid] = ACTIONS(3109),
    [anon_sym_getpeername] = ACTIONS(3109),
    [anon_sym_getnetbyname] = ACTIONS(3109),
    [anon_sym_getsockname] = ACTIONS(3109),
    [anon_sym_getgrnam] = ACTIONS(3109),
    [anon_sym_getgrgid] = ACTIONS(3109),
    [anon_sym_hex] = ACTIONS(3109),
    [anon_sym_int] = ACTIONS(3109),
    [anon_sym_keys] = ACTIONS(3109),
    [anon_sym_lc] = ACTIONS(3109),
    [anon_sym_lcfirst] = ACTIONS(3109),
    [anon_sym_length] = ACTIONS(3109),
    [anon_sym_localtime] = ACTIONS(3109),
    [anon_sym_log] = ACTIONS(3109),
    [anon_sym_lock] = ACTIONS(3109),
    [anon_sym_lstat] = ACTIONS(3109),
    [anon_sym_oct] = ACTIONS(3109),
    [anon_sym_ord] = ACTIONS(3109),
    [anon_sym_prototype] = ACTIONS(3109),
    [anon_sym_pop] = ACTIONS(3109),
    [anon_sym_pos] = ACTIONS(3109),
    [anon_sym_quotemeta] = ACTIONS(3109),
    [anon_sym_reset] = ACTIONS(3109),
    [anon_sym_rand] = ACTIONS(3109),
    [anon_sym_rmdir] = ACTIONS(3109),
    [anon_sym_readdir] = ACTIONS(3109),
    [anon_sym_readline] = ACTIONS(3109),
    [anon_sym_readpipe] = ACTIONS(3109),
    [anon_sym_rewinddir] = ACTIONS(3109),
    [anon_sym_readlink] = ACTIONS(3109),
    [anon_sym_ref] = ACTIONS(3109),
    [anon_sym_scalar] = ACTIONS(3109),
    [anon_sym_shift] = ACTIONS(3109),
    [anon_sym_sin] = ACTIONS(3109),
    [anon_sym_sleep] = ACTIONS(3109),
    [anon_sym_sqrt] = ACTIONS(3109),
    [anon_sym_srand] = ACTIONS(3109),
    [anon_sym_stat] = ACTIONS(3109),
    [anon_sym_study] = ACTIONS(3109),
    [anon_sym_tell] = ACTIONS(3109),
    [anon_sym_telldir] = ACTIONS(3109),
    [anon_sym_tied] = ACTIONS(3109),
    [anon_sym_uc] = ACTIONS(3109),
    [anon_sym_ucfirst] = ACTIONS(3109),
    [anon_sym_untie] = ACTIONS(3109),
    [anon_sym_umask] = ACTIONS(3109),
    [anon_sym_values] = ACTIONS(3109),
    [anon_sym_write] = ACTIONS(3109),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3109),
    [anon_sym_qq] = ACTIONS(3109),
    [anon_sym_qw] = ACTIONS(3109),
    [anon_sym_qx] = ACTIONS(3109),
    [anon_sym_qr] = ACTIONS(3109),
    [anon_sym_m] = ACTIONS(3109),
    [anon_sym_s] = ACTIONS(3109),
    [anon_sym_tr] = ACTIONS(3109),
    [anon_sym_y] = ACTIONS(3109),
    [aux_sym__bareword_token1] = ACTIONS(3109),
    [sym_number] = ACTIONS(3109),
    [anon_sym_true] = ACTIONS(3109),
    [anon_sym_false] = ACTIONS(3109),
    [sym__single_quote] = ACTIONS(3107),
    [sym__double_quote] = ACTIONS(3107),
    [sym__backtick_quote] = ACTIONS(3107),
    [sym__search_slash_quote] = ACTIONS(3107),
    [sym__open_readline_bracket] = ACTIONS(3107),
    [sym__open_fileglob_bracket] = ACTIONS(3107),
    [sym__PERLY_HEREDOC] = ACTIONS(3107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3107),
    [sym__filetest] = ACTIONS(3107),
  },
  [STATE(1706)] = {
    [sym_heredoc_content] = STATE(1706),
    [sym__identifier] = ACTIONS(3117),
    [anon_sym_LBRACE] = ACTIONS(3117),
    [anon_sym_DOLLAR] = ACTIONS(3117),
    [anon_sym_AT] = ACTIONS(3117),
    [anon_sym_RPAREN] = ACTIONS(3117),
    [anon_sym_my] = ACTIONS(3117),
    [anon_sym_extended] = ACTIONS(3117),
    [anon_sym_async] = ACTIONS(3117),
    [anon_sym_sub] = ACTIONS(3117),
    [anon_sym_method] = ACTIONS(3117),
    [anon_sym_LPAREN] = ACTIONS(3117),
    [anon_sym_continue] = ACTIONS(3117),
    [anon_sym_state] = ACTIONS(3117),
    [anon_sym_our] = ACTIONS(3117),
    [anon_sym_LBRACK] = ACTIONS(3117),
    [anon_sym_LT_LT] = ACTIONS(3117),
    [anon_sym_SLASH_SLASH] = ACTIONS(3117),
    [anon_sym_PLUS] = ACTIONS(3117),
    [anon_sym_DASH] = ACTIONS(3117),
    [anon_sym_TILDE] = ACTIONS(3117),
    [anon_sym_BANG] = ACTIONS(3117),
    [anon_sym_PLUS_PLUS] = ACTIONS(3117),
    [anon_sym_DASH_DASH] = ACTIONS(3117),
    [anon_sym_BSLASH] = ACTIONS(3115),
    [anon_sym_do] = ACTIONS(3117),
    [anon_sym_eval] = ACTIONS(3117),
    [anon_sym_await] = ACTIONS(3117),
    [anon_sym_field] = ACTIONS(3117),
    [anon_sym_local] = ACTIONS(3117),
    [anon_sym_dynamically] = ACTIONS(3117),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3117),
    [anon_sym_require] = ACTIONS(3117),
    [anon_sym_map] = ACTIONS(3117),
    [anon_sym_grep] = ACTIONS(3117),
    [anon_sym_sort] = ACTIONS(3117),
    [anon_sym_goto] = ACTIONS(3117),
    [anon_sym_return] = ACTIONS(3117),
    [anon_sym_undef] = ACTIONS(3117),
    [anon_sym_PERCENT2] = ACTIONS(3117),
    [anon_sym_AMP2] = ACTIONS(3117),
    [anon_sym_STAR2] = ACTIONS(3117),
    [anon_sym_last] = ACTIONS(3117),
    [anon_sym_next] = ACTIONS(3117),
    [anon_sym_redo] = ACTIONS(3117),
    [anon_sym___FILE__] = ACTIONS(3117),
    [anon_sym___LINE__] = ACTIONS(3117),
    [anon_sym___PACKAGE__] = ACTIONS(3117),
    [anon_sym___SUB__] = ACTIONS(3117),
    [anon_sym_break] = ACTIONS(3117),
    [anon_sym_fork] = ACTIONS(3117),
    [anon_sym_getppid] = ACTIONS(3117),
    [anon_sym_time] = ACTIONS(3117),
    [anon_sym_times] = ACTIONS(3117),
    [anon_sym_wait] = ACTIONS(3117),
    [anon_sym_wantarray] = ACTIONS(3117),
    [anon_sym_abs] = ACTIONS(3117),
    [anon_sym_alarm] = ACTIONS(3117),
    [anon_sym_chop] = ACTIONS(3117),
    [anon_sym_chdir] = ACTIONS(3117),
    [anon_sym_close] = ACTIONS(3117),
    [anon_sym_closedir] = ACTIONS(3117),
    [anon_sym_caller] = ACTIONS(3117),
    [anon_sym_chomp] = ACTIONS(3117),
    [anon_sym_chr] = ACTIONS(3117),
    [anon_sym_cos] = ACTIONS(3117),
    [anon_sym_chroot] = ACTIONS(3117),
    [anon_sym_defined] = ACTIONS(3117),
    [anon_sym_delete] = ACTIONS(3117),
    [anon_sym_dbmclose] = ACTIONS(3117),
    [anon_sym_exists] = ACTIONS(3117),
    [anon_sym_exit] = ACTIONS(3117),
    [anon_sym_eof] = ACTIONS(3117),
    [anon_sym_exp] = ACTIONS(3117),
    [anon_sym_each] = ACTIONS(3117),
    [anon_sym_fc] = ACTIONS(3117),
    [anon_sym_fileno] = ACTIONS(3117),
    [anon_sym_gmtime] = ACTIONS(3117),
    [anon_sym_getc] = ACTIONS(3117),
    [anon_sym_getpgrp] = ACTIONS(3117),
    [anon_sym_getprotobyname] = ACTIONS(3117),
    [anon_sym_getpwname] = ACTIONS(3117),
    [anon_sym_getpwuid] = ACTIONS(3117),
    [anon_sym_getpeername] = ACTIONS(3117),
    [anon_sym_getnetbyname] = ACTIONS(3117),
    [anon_sym_getsockname] = ACTIONS(3117),
    [anon_sym_getgrnam] = ACTIONS(3117),
    [anon_sym_getgrgid] = ACTIONS(3117),
    [anon_sym_hex] = ACTIONS(3117),
    [anon_sym_int] = ACTIONS(3117),
    [anon_sym_keys] = ACTIONS(3117),
    [anon_sym_lc] = ACTIONS(3117),
    [anon_sym_lcfirst] = ACTIONS(3117),
    [anon_sym_length] = ACTIONS(3117),
    [anon_sym_localtime] = ACTIONS(3117),
    [anon_sym_log] = ACTIONS(3117),
    [anon_sym_lock] = ACTIONS(3117),
    [anon_sym_lstat] = ACTIONS(3117),
    [anon_sym_oct] = ACTIONS(3117),
    [anon_sym_ord] = ACTIONS(3117),
    [anon_sym_prototype] = ACTIONS(3117),
    [anon_sym_pop] = ACTIONS(3117),
    [anon_sym_pos] = ACTIONS(3117),
    [anon_sym_quotemeta] = ACTIONS(3117),
    [anon_sym_reset] = ACTIONS(3117),
    [anon_sym_rand] = ACTIONS(3117),
    [anon_sym_rmdir] = ACTIONS(3117),
    [anon_sym_readdir] = ACTIONS(3117),
    [anon_sym_readline] = ACTIONS(3117),
    [anon_sym_readpipe] = ACTIONS(3117),
    [anon_sym_rewinddir] = ACTIONS(3117),
    [anon_sym_readlink] = ACTIONS(3117),
    [anon_sym_ref] = ACTIONS(3117),
    [anon_sym_scalar] = ACTIONS(3117),
    [anon_sym_shift] = ACTIONS(3117),
    [anon_sym_sin] = ACTIONS(3117),
    [anon_sym_sleep] = ACTIONS(3117),
    [anon_sym_sqrt] = ACTIONS(3117),
    [anon_sym_srand] = ACTIONS(3117),
    [anon_sym_stat] = ACTIONS(3117),
    [anon_sym_study] = ACTIONS(3117),
    [anon_sym_tell] = ACTIONS(3117),
    [anon_sym_telldir] = ACTIONS(3117),
    [anon_sym_tied] = ACTIONS(3117),
    [anon_sym_uc] = ACTIONS(3117),
    [anon_sym_ucfirst] = ACTIONS(3117),
    [anon_sym_untie] = ACTIONS(3117),
    [anon_sym_umask] = ACTIONS(3117),
    [anon_sym_values] = ACTIONS(3117),
    [anon_sym_write] = ACTIONS(3117),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3117),
    [anon_sym_qq] = ACTIONS(3117),
    [anon_sym_qw] = ACTIONS(3117),
    [anon_sym_qx] = ACTIONS(3117),
    [anon_sym_qr] = ACTIONS(3117),
    [anon_sym_m] = ACTIONS(3117),
    [anon_sym_s] = ACTIONS(3117),
    [anon_sym_tr] = ACTIONS(3117),
    [anon_sym_y] = ACTIONS(3117),
    [aux_sym__bareword_token1] = ACTIONS(3117),
    [sym_number] = ACTIONS(3117),
    [anon_sym_true] = ACTIONS(3117),
    [anon_sym_false] = ACTIONS(3117),
    [sym__single_quote] = ACTIONS(3115),
    [sym__double_quote] = ACTIONS(3115),
    [sym__backtick_quote] = ACTIONS(3115),
    [sym__search_slash_quote] = ACTIONS(3115),
    [sym__open_readline_bracket] = ACTIONS(3115),
    [sym__open_fileglob_bracket] = ACTIONS(3115),
    [sym__PERLY_HEREDOC] = ACTIONS(3115),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3115),
    [sym__filetest] = ACTIONS(3115),
  },
  [STATE(1707)] = {
    [sym_heredoc_content] = STATE(1707),
    [sym__identifier] = ACTIONS(3133),
    [anon_sym_LBRACE] = ACTIONS(3133),
    [anon_sym_DOLLAR] = ACTIONS(3133),
    [anon_sym_AT] = ACTIONS(3133),
    [anon_sym_RPAREN] = ACTIONS(3133),
    [anon_sym_my] = ACTIONS(3133),
    [anon_sym_extended] = ACTIONS(3133),
    [anon_sym_async] = ACTIONS(3133),
    [anon_sym_sub] = ACTIONS(3133),
    [anon_sym_method] = ACTIONS(3133),
    [anon_sym_LPAREN] = ACTIONS(3133),
    [anon_sym_continue] = ACTIONS(3133),
    [anon_sym_state] = ACTIONS(3133),
    [anon_sym_our] = ACTIONS(3133),
    [anon_sym_LBRACK] = ACTIONS(3133),
    [anon_sym_LT_LT] = ACTIONS(3133),
    [anon_sym_SLASH_SLASH] = ACTIONS(3133),
    [anon_sym_PLUS] = ACTIONS(3133),
    [anon_sym_DASH] = ACTIONS(3133),
    [anon_sym_TILDE] = ACTIONS(3133),
    [anon_sym_BANG] = ACTIONS(3133),
    [anon_sym_PLUS_PLUS] = ACTIONS(3133),
    [anon_sym_DASH_DASH] = ACTIONS(3133),
    [anon_sym_BSLASH] = ACTIONS(3131),
    [anon_sym_do] = ACTIONS(3133),
    [anon_sym_eval] = ACTIONS(3133),
    [anon_sym_await] = ACTIONS(3133),
    [anon_sym_field] = ACTIONS(3133),
    [anon_sym_local] = ACTIONS(3133),
    [anon_sym_dynamically] = ACTIONS(3133),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3133),
    [anon_sym_require] = ACTIONS(3133),
    [anon_sym_map] = ACTIONS(3133),
    [anon_sym_grep] = ACTIONS(3133),
    [anon_sym_sort] = ACTIONS(3133),
    [anon_sym_goto] = ACTIONS(3133),
    [anon_sym_return] = ACTIONS(3133),
    [anon_sym_undef] = ACTIONS(3133),
    [anon_sym_PERCENT2] = ACTIONS(3133),
    [anon_sym_AMP2] = ACTIONS(3133),
    [anon_sym_STAR2] = ACTIONS(3133),
    [anon_sym_last] = ACTIONS(3133),
    [anon_sym_next] = ACTIONS(3133),
    [anon_sym_redo] = ACTIONS(3133),
    [anon_sym___FILE__] = ACTIONS(3133),
    [anon_sym___LINE__] = ACTIONS(3133),
    [anon_sym___PACKAGE__] = ACTIONS(3133),
    [anon_sym___SUB__] = ACTIONS(3133),
    [anon_sym_break] = ACTIONS(3133),
    [anon_sym_fork] = ACTIONS(3133),
    [anon_sym_getppid] = ACTIONS(3133),
    [anon_sym_time] = ACTIONS(3133),
    [anon_sym_times] = ACTIONS(3133),
    [anon_sym_wait] = ACTIONS(3133),
    [anon_sym_wantarray] = ACTIONS(3133),
    [anon_sym_abs] = ACTIONS(3133),
    [anon_sym_alarm] = ACTIONS(3133),
    [anon_sym_chop] = ACTIONS(3133),
    [anon_sym_chdir] = ACTIONS(3133),
    [anon_sym_close] = ACTIONS(3133),
    [anon_sym_closedir] = ACTIONS(3133),
    [anon_sym_caller] = ACTIONS(3133),
    [anon_sym_chomp] = ACTIONS(3133),
    [anon_sym_chr] = ACTIONS(3133),
    [anon_sym_cos] = ACTIONS(3133),
    [anon_sym_chroot] = ACTIONS(3133),
    [anon_sym_defined] = ACTIONS(3133),
    [anon_sym_delete] = ACTIONS(3133),
    [anon_sym_dbmclose] = ACTIONS(3133),
    [anon_sym_exists] = ACTIONS(3133),
    [anon_sym_exit] = ACTIONS(3133),
    [anon_sym_eof] = ACTIONS(3133),
    [anon_sym_exp] = ACTIONS(3133),
    [anon_sym_each] = ACTIONS(3133),
    [anon_sym_fc] = ACTIONS(3133),
    [anon_sym_fileno] = ACTIONS(3133),
    [anon_sym_gmtime] = ACTIONS(3133),
    [anon_sym_getc] = ACTIONS(3133),
    [anon_sym_getpgrp] = ACTIONS(3133),
    [anon_sym_getprotobyname] = ACTIONS(3133),
    [anon_sym_getpwname] = ACTIONS(3133),
    [anon_sym_getpwuid] = ACTIONS(3133),
    [anon_sym_getpeername] = ACTIONS(3133),
    [anon_sym_getnetbyname] = ACTIONS(3133),
    [anon_sym_getsockname] = ACTIONS(3133),
    [anon_sym_getgrnam] = ACTIONS(3133),
    [anon_sym_getgrgid] = ACTIONS(3133),
    [anon_sym_hex] = ACTIONS(3133),
    [anon_sym_int] = ACTIONS(3133),
    [anon_sym_keys] = ACTIONS(3133),
    [anon_sym_lc] = ACTIONS(3133),
    [anon_sym_lcfirst] = ACTIONS(3133),
    [anon_sym_length] = ACTIONS(3133),
    [anon_sym_localtime] = ACTIONS(3133),
    [anon_sym_log] = ACTIONS(3133),
    [anon_sym_lock] = ACTIONS(3133),
    [anon_sym_lstat] = ACTIONS(3133),
    [anon_sym_oct] = ACTIONS(3133),
    [anon_sym_ord] = ACTIONS(3133),
    [anon_sym_prototype] = ACTIONS(3133),
    [anon_sym_pop] = ACTIONS(3133),
    [anon_sym_pos] = ACTIONS(3133),
    [anon_sym_quotemeta] = ACTIONS(3133),
    [anon_sym_reset] = ACTIONS(3133),
    [anon_sym_rand] = ACTIONS(3133),
    [anon_sym_rmdir] = ACTIONS(3133),
    [anon_sym_readdir] = ACTIONS(3133),
    [anon_sym_readline] = ACTIONS(3133),
    [anon_sym_readpipe] = ACTIONS(3133),
    [anon_sym_rewinddir] = ACTIONS(3133),
    [anon_sym_readlink] = ACTIONS(3133),
    [anon_sym_ref] = ACTIONS(3133),
    [anon_sym_scalar] = ACTIONS(3133),
    [anon_sym_shift] = ACTIONS(3133),
    [anon_sym_sin] = ACTIONS(3133),
    [anon_sym_sleep] = ACTIONS(3133),
    [anon_sym_sqrt] = ACTIONS(3133),
    [anon_sym_srand] = ACTIONS(3133),
    [anon_sym_stat] = ACTIONS(3133),
    [anon_sym_study] = ACTIONS(3133),
    [anon_sym_tell] = ACTIONS(3133),
    [anon_sym_telldir] = ACTIONS(3133),
    [anon_sym_tied] = ACTIONS(3133),
    [anon_sym_uc] = ACTIONS(3133),
    [anon_sym_ucfirst] = ACTIONS(3133),
    [anon_sym_untie] = ACTIONS(3133),
    [anon_sym_umask] = ACTIONS(3133),
    [anon_sym_values] = ACTIONS(3133),
    [anon_sym_write] = ACTIONS(3133),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3133),
    [anon_sym_qq] = ACTIONS(3133),
    [anon_sym_qw] = ACTIONS(3133),
    [anon_sym_qx] = ACTIONS(3133),
    [anon_sym_qr] = ACTIONS(3133),
    [anon_sym_m] = ACTIONS(3133),
    [anon_sym_s] = ACTIONS(3133),
    [anon_sym_tr] = ACTIONS(3133),
    [anon_sym_y] = ACTIONS(3133),
    [aux_sym__bareword_token1] = ACTIONS(3133),
    [sym_number] = ACTIONS(3133),
    [anon_sym_true] = ACTIONS(3133),
    [anon_sym_false] = ACTIONS(3133),
    [sym__single_quote] = ACTIONS(3131),
    [sym__double_quote] = ACTIONS(3131),
    [sym__backtick_quote] = ACTIONS(3131),
    [sym__search_slash_quote] = ACTIONS(3131),
    [sym__open_readline_bracket] = ACTIONS(3131),
    [sym__open_fileglob_bracket] = ACTIONS(3131),
    [sym__PERLY_HEREDOC] = ACTIONS(3131),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3131),
    [sym__filetest] = ACTIONS(3131),
  },
  [STATE(1708)] = {
    [sym_heredoc_content] = STATE(1708),
    [sym__identifier] = ACTIONS(3137),
    [anon_sym_LBRACE] = ACTIONS(3137),
    [anon_sym_DOLLAR] = ACTIONS(3137),
    [anon_sym_AT] = ACTIONS(3137),
    [anon_sym_RPAREN] = ACTIONS(3137),
    [anon_sym_my] = ACTIONS(3137),
    [anon_sym_extended] = ACTIONS(3137),
    [anon_sym_async] = ACTIONS(3137),
    [anon_sym_sub] = ACTIONS(3137),
    [anon_sym_method] = ACTIONS(3137),
    [anon_sym_LPAREN] = ACTIONS(3137),
    [anon_sym_continue] = ACTIONS(3137),
    [anon_sym_state] = ACTIONS(3137),
    [anon_sym_our] = ACTIONS(3137),
    [anon_sym_LBRACK] = ACTIONS(3137),
    [anon_sym_LT_LT] = ACTIONS(3137),
    [anon_sym_SLASH_SLASH] = ACTIONS(3137),
    [anon_sym_PLUS] = ACTIONS(3137),
    [anon_sym_DASH] = ACTIONS(3137),
    [anon_sym_TILDE] = ACTIONS(3137),
    [anon_sym_BANG] = ACTIONS(3137),
    [anon_sym_PLUS_PLUS] = ACTIONS(3137),
    [anon_sym_DASH_DASH] = ACTIONS(3137),
    [anon_sym_BSLASH] = ACTIONS(3135),
    [anon_sym_do] = ACTIONS(3137),
    [anon_sym_eval] = ACTIONS(3137),
    [anon_sym_await] = ACTIONS(3137),
    [anon_sym_field] = ACTIONS(3137),
    [anon_sym_local] = ACTIONS(3137),
    [anon_sym_dynamically] = ACTIONS(3137),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3137),
    [anon_sym_require] = ACTIONS(3137),
    [anon_sym_map] = ACTIONS(3137),
    [anon_sym_grep] = ACTIONS(3137),
    [anon_sym_sort] = ACTIONS(3137),
    [anon_sym_goto] = ACTIONS(3137),
    [anon_sym_return] = ACTIONS(3137),
    [anon_sym_undef] = ACTIONS(3137),
    [anon_sym_PERCENT2] = ACTIONS(3137),
    [anon_sym_AMP2] = ACTIONS(3137),
    [anon_sym_STAR2] = ACTIONS(3137),
    [anon_sym_last] = ACTIONS(3137),
    [anon_sym_next] = ACTIONS(3137),
    [anon_sym_redo] = ACTIONS(3137),
    [anon_sym___FILE__] = ACTIONS(3137),
    [anon_sym___LINE__] = ACTIONS(3137),
    [anon_sym___PACKAGE__] = ACTIONS(3137),
    [anon_sym___SUB__] = ACTIONS(3137),
    [anon_sym_break] = ACTIONS(3137),
    [anon_sym_fork] = ACTIONS(3137),
    [anon_sym_getppid] = ACTIONS(3137),
    [anon_sym_time] = ACTIONS(3137),
    [anon_sym_times] = ACTIONS(3137),
    [anon_sym_wait] = ACTIONS(3137),
    [anon_sym_wantarray] = ACTIONS(3137),
    [anon_sym_abs] = ACTIONS(3137),
    [anon_sym_alarm] = ACTIONS(3137),
    [anon_sym_chop] = ACTIONS(3137),
    [anon_sym_chdir] = ACTIONS(3137),
    [anon_sym_close] = ACTIONS(3137),
    [anon_sym_closedir] = ACTIONS(3137),
    [anon_sym_caller] = ACTIONS(3137),
    [anon_sym_chomp] = ACTIONS(3137),
    [anon_sym_chr] = ACTIONS(3137),
    [anon_sym_cos] = ACTIONS(3137),
    [anon_sym_chroot] = ACTIONS(3137),
    [anon_sym_defined] = ACTIONS(3137),
    [anon_sym_delete] = ACTIONS(3137),
    [anon_sym_dbmclose] = ACTIONS(3137),
    [anon_sym_exists] = ACTIONS(3137),
    [anon_sym_exit] = ACTIONS(3137),
    [anon_sym_eof] = ACTIONS(3137),
    [anon_sym_exp] = ACTIONS(3137),
    [anon_sym_each] = ACTIONS(3137),
    [anon_sym_fc] = ACTIONS(3137),
    [anon_sym_fileno] = ACTIONS(3137),
    [anon_sym_gmtime] = ACTIONS(3137),
    [anon_sym_getc] = ACTIONS(3137),
    [anon_sym_getpgrp] = ACTIONS(3137),
    [anon_sym_getprotobyname] = ACTIONS(3137),
    [anon_sym_getpwname] = ACTIONS(3137),
    [anon_sym_getpwuid] = ACTIONS(3137),
    [anon_sym_getpeername] = ACTIONS(3137),
    [anon_sym_getnetbyname] = ACTIONS(3137),
    [anon_sym_getsockname] = ACTIONS(3137),
    [anon_sym_getgrnam] = ACTIONS(3137),
    [anon_sym_getgrgid] = ACTIONS(3137),
    [anon_sym_hex] = ACTIONS(3137),
    [anon_sym_int] = ACTIONS(3137),
    [anon_sym_keys] = ACTIONS(3137),
    [anon_sym_lc] = ACTIONS(3137),
    [anon_sym_lcfirst] = ACTIONS(3137),
    [anon_sym_length] = ACTIONS(3137),
    [anon_sym_localtime] = ACTIONS(3137),
    [anon_sym_log] = ACTIONS(3137),
    [anon_sym_lock] = ACTIONS(3137),
    [anon_sym_lstat] = ACTIONS(3137),
    [anon_sym_oct] = ACTIONS(3137),
    [anon_sym_ord] = ACTIONS(3137),
    [anon_sym_prototype] = ACTIONS(3137),
    [anon_sym_pop] = ACTIONS(3137),
    [anon_sym_pos] = ACTIONS(3137),
    [anon_sym_quotemeta] = ACTIONS(3137),
    [anon_sym_reset] = ACTIONS(3137),
    [anon_sym_rand] = ACTIONS(3137),
    [anon_sym_rmdir] = ACTIONS(3137),
    [anon_sym_readdir] = ACTIONS(3137),
    [anon_sym_readline] = ACTIONS(3137),
    [anon_sym_readpipe] = ACTIONS(3137),
    [anon_sym_rewinddir] = ACTIONS(3137),
    [anon_sym_readlink] = ACTIONS(3137),
    [anon_sym_ref] = ACTIONS(3137),
    [anon_sym_scalar] = ACTIONS(3137),
    [anon_sym_shift] = ACTIONS(3137),
    [anon_sym_sin] = ACTIONS(3137),
    [anon_sym_sleep] = ACTIONS(3137),
    [anon_sym_sqrt] = ACTIONS(3137),
    [anon_sym_srand] = ACTIONS(3137),
    [anon_sym_stat] = ACTIONS(3137),
    [anon_sym_study] = ACTIONS(3137),
    [anon_sym_tell] = ACTIONS(3137),
    [anon_sym_telldir] = ACTIONS(3137),
    [anon_sym_tied] = ACTIONS(3137),
    [anon_sym_uc] = ACTIONS(3137),
    [anon_sym_ucfirst] = ACTIONS(3137),
    [anon_sym_untie] = ACTIONS(3137),
    [anon_sym_umask] = ACTIONS(3137),
    [anon_sym_values] = ACTIONS(3137),
    [anon_sym_write] = ACTIONS(3137),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3137),
    [anon_sym_qq] = ACTIONS(3137),
    [anon_sym_qw] = ACTIONS(3137),
    [anon_sym_qx] = ACTIONS(3137),
    [anon_sym_qr] = ACTIONS(3137),
    [anon_sym_m] = ACTIONS(3137),
    [anon_sym_s] = ACTIONS(3137),
    [anon_sym_tr] = ACTIONS(3137),
    [anon_sym_y] = ACTIONS(3137),
    [aux_sym__bareword_token1] = ACTIONS(3137),
    [sym_number] = ACTIONS(3137),
    [anon_sym_true] = ACTIONS(3137),
    [anon_sym_false] = ACTIONS(3137),
    [sym__single_quote] = ACTIONS(3135),
    [sym__double_quote] = ACTIONS(3135),
    [sym__backtick_quote] = ACTIONS(3135),
    [sym__search_slash_quote] = ACTIONS(3135),
    [sym__open_readline_bracket] = ACTIONS(3135),
    [sym__open_fileglob_bracket] = ACTIONS(3135),
    [sym__PERLY_HEREDOC] = ACTIONS(3135),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3135),
    [sym__filetest] = ACTIONS(3135),
  },
  [STATE(1709)] = {
    [sym_heredoc_content] = STATE(1709),
    [sym__identifier] = ACTIONS(3149),
    [anon_sym_LBRACE] = ACTIONS(3149),
    [anon_sym_DOLLAR] = ACTIONS(3149),
    [anon_sym_AT] = ACTIONS(3149),
    [anon_sym_RPAREN] = ACTIONS(3149),
    [anon_sym_my] = ACTIONS(3149),
    [anon_sym_extended] = ACTIONS(3149),
    [anon_sym_async] = ACTIONS(3149),
    [anon_sym_sub] = ACTIONS(3149),
    [anon_sym_method] = ACTIONS(3149),
    [anon_sym_LPAREN] = ACTIONS(3149),
    [anon_sym_continue] = ACTIONS(3149),
    [anon_sym_state] = ACTIONS(3149),
    [anon_sym_our] = ACTIONS(3149),
    [anon_sym_LBRACK] = ACTIONS(3149),
    [anon_sym_LT_LT] = ACTIONS(3149),
    [anon_sym_SLASH_SLASH] = ACTIONS(3149),
    [anon_sym_PLUS] = ACTIONS(3149),
    [anon_sym_DASH] = ACTIONS(3149),
    [anon_sym_TILDE] = ACTIONS(3149),
    [anon_sym_BANG] = ACTIONS(3149),
    [anon_sym_PLUS_PLUS] = ACTIONS(3149),
    [anon_sym_DASH_DASH] = ACTIONS(3149),
    [anon_sym_BSLASH] = ACTIONS(3147),
    [anon_sym_do] = ACTIONS(3149),
    [anon_sym_eval] = ACTIONS(3149),
    [anon_sym_await] = ACTIONS(3149),
    [anon_sym_field] = ACTIONS(3149),
    [anon_sym_local] = ACTIONS(3149),
    [anon_sym_dynamically] = ACTIONS(3149),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3149),
    [anon_sym_require] = ACTIONS(3149),
    [anon_sym_map] = ACTIONS(3149),
    [anon_sym_grep] = ACTIONS(3149),
    [anon_sym_sort] = ACTIONS(3149),
    [anon_sym_goto] = ACTIONS(3149),
    [anon_sym_return] = ACTIONS(3149),
    [anon_sym_undef] = ACTIONS(3149),
    [anon_sym_PERCENT2] = ACTIONS(3149),
    [anon_sym_AMP2] = ACTIONS(3149),
    [anon_sym_STAR2] = ACTIONS(3149),
    [anon_sym_last] = ACTIONS(3149),
    [anon_sym_next] = ACTIONS(3149),
    [anon_sym_redo] = ACTIONS(3149),
    [anon_sym___FILE__] = ACTIONS(3149),
    [anon_sym___LINE__] = ACTIONS(3149),
    [anon_sym___PACKAGE__] = ACTIONS(3149),
    [anon_sym___SUB__] = ACTIONS(3149),
    [anon_sym_break] = ACTIONS(3149),
    [anon_sym_fork] = ACTIONS(3149),
    [anon_sym_getppid] = ACTIONS(3149),
    [anon_sym_time] = ACTIONS(3149),
    [anon_sym_times] = ACTIONS(3149),
    [anon_sym_wait] = ACTIONS(3149),
    [anon_sym_wantarray] = ACTIONS(3149),
    [anon_sym_abs] = ACTIONS(3149),
    [anon_sym_alarm] = ACTIONS(3149),
    [anon_sym_chop] = ACTIONS(3149),
    [anon_sym_chdir] = ACTIONS(3149),
    [anon_sym_close] = ACTIONS(3149),
    [anon_sym_closedir] = ACTIONS(3149),
    [anon_sym_caller] = ACTIONS(3149),
    [anon_sym_chomp] = ACTIONS(3149),
    [anon_sym_chr] = ACTIONS(3149),
    [anon_sym_cos] = ACTIONS(3149),
    [anon_sym_chroot] = ACTIONS(3149),
    [anon_sym_defined] = ACTIONS(3149),
    [anon_sym_delete] = ACTIONS(3149),
    [anon_sym_dbmclose] = ACTIONS(3149),
    [anon_sym_exists] = ACTIONS(3149),
    [anon_sym_exit] = ACTIONS(3149),
    [anon_sym_eof] = ACTIONS(3149),
    [anon_sym_exp] = ACTIONS(3149),
    [anon_sym_each] = ACTIONS(3149),
    [anon_sym_fc] = ACTIONS(3149),
    [anon_sym_fileno] = ACTIONS(3149),
    [anon_sym_gmtime] = ACTIONS(3149),
    [anon_sym_getc] = ACTIONS(3149),
    [anon_sym_getpgrp] = ACTIONS(3149),
    [anon_sym_getprotobyname] = ACTIONS(3149),
    [anon_sym_getpwname] = ACTIONS(3149),
    [anon_sym_getpwuid] = ACTIONS(3149),
    [anon_sym_getpeername] = ACTIONS(3149),
    [anon_sym_getnetbyname] = ACTIONS(3149),
    [anon_sym_getsockname] = ACTIONS(3149),
    [anon_sym_getgrnam] = ACTIONS(3149),
    [anon_sym_getgrgid] = ACTIONS(3149),
    [anon_sym_hex] = ACTIONS(3149),
    [anon_sym_int] = ACTIONS(3149),
    [anon_sym_keys] = ACTIONS(3149),
    [anon_sym_lc] = ACTIONS(3149),
    [anon_sym_lcfirst] = ACTIONS(3149),
    [anon_sym_length] = ACTIONS(3149),
    [anon_sym_localtime] = ACTIONS(3149),
    [anon_sym_log] = ACTIONS(3149),
    [anon_sym_lock] = ACTIONS(3149),
    [anon_sym_lstat] = ACTIONS(3149),
    [anon_sym_oct] = ACTIONS(3149),
    [anon_sym_ord] = ACTIONS(3149),
    [anon_sym_prototype] = ACTIONS(3149),
    [anon_sym_pop] = ACTIONS(3149),
    [anon_sym_pos] = ACTIONS(3149),
    [anon_sym_quotemeta] = ACTIONS(3149),
    [anon_sym_reset] = ACTIONS(3149),
    [anon_sym_rand] = ACTIONS(3149),
    [anon_sym_rmdir] = ACTIONS(3149),
    [anon_sym_readdir] = ACTIONS(3149),
    [anon_sym_readline] = ACTIONS(3149),
    [anon_sym_readpipe] = ACTIONS(3149),
    [anon_sym_rewinddir] = ACTIONS(3149),
    [anon_sym_readlink] = ACTIONS(3149),
    [anon_sym_ref] = ACTIONS(3149),
    [anon_sym_scalar] = ACTIONS(3149),
    [anon_sym_shift] = ACTIONS(3149),
    [anon_sym_sin] = ACTIONS(3149),
    [anon_sym_sleep] = ACTIONS(3149),
    [anon_sym_sqrt] = ACTIONS(3149),
    [anon_sym_srand] = ACTIONS(3149),
    [anon_sym_stat] = ACTIONS(3149),
    [anon_sym_study] = ACTIONS(3149),
    [anon_sym_tell] = ACTIONS(3149),
    [anon_sym_telldir] = ACTIONS(3149),
    [anon_sym_tied] = ACTIONS(3149),
    [anon_sym_uc] = ACTIONS(3149),
    [anon_sym_ucfirst] = ACTIONS(3149),
    [anon_sym_untie] = ACTIONS(3149),
    [anon_sym_umask] = ACTIONS(3149),
    [anon_sym_values] = ACTIONS(3149),
    [anon_sym_write] = ACTIONS(3149),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3149),
    [anon_sym_qq] = ACTIONS(3149),
    [anon_sym_qw] = ACTIONS(3149),
    [anon_sym_qx] = ACTIONS(3149),
    [anon_sym_qr] = ACTIONS(3149),
    [anon_sym_m] = ACTIONS(3149),
    [anon_sym_s] = ACTIONS(3149),
    [anon_sym_tr] = ACTIONS(3149),
    [anon_sym_y] = ACTIONS(3149),
    [aux_sym__bareword_token1] = ACTIONS(3149),
    [sym_number] = ACTIONS(3149),
    [anon_sym_true] = ACTIONS(3149),
    [anon_sym_false] = ACTIONS(3149),
    [sym__single_quote] = ACTIONS(3147),
    [sym__double_quote] = ACTIONS(3147),
    [sym__backtick_quote] = ACTIONS(3147),
    [sym__search_slash_quote] = ACTIONS(3147),
    [sym__open_readline_bracket] = ACTIONS(3147),
    [sym__open_fileglob_bracket] = ACTIONS(3147),
    [sym__PERLY_HEREDOC] = ACTIONS(3147),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3147),
    [sym__filetest] = ACTIONS(3147),
  },
  [STATE(1710)] = {
    [sym_heredoc_content] = STATE(1710),
    [sym__identifier] = ACTIONS(3041),
    [anon_sym_LBRACE] = ACTIONS(3041),
    [anon_sym_DOLLAR] = ACTIONS(3041),
    [anon_sym_AT] = ACTIONS(3041),
    [anon_sym_RPAREN] = ACTIONS(3041),
    [anon_sym_my] = ACTIONS(3041),
    [anon_sym_extended] = ACTIONS(3041),
    [anon_sym_async] = ACTIONS(3041),
    [anon_sym_sub] = ACTIONS(3041),
    [anon_sym_method] = ACTIONS(3041),
    [anon_sym_LPAREN] = ACTIONS(3041),
    [anon_sym_continue] = ACTIONS(3041),
    [anon_sym_state] = ACTIONS(3041),
    [anon_sym_our] = ACTIONS(3041),
    [anon_sym_LBRACK] = ACTIONS(3041),
    [anon_sym_LT_LT] = ACTIONS(3041),
    [anon_sym_SLASH_SLASH] = ACTIONS(3041),
    [anon_sym_PLUS] = ACTIONS(3041),
    [anon_sym_DASH] = ACTIONS(3041),
    [anon_sym_TILDE] = ACTIONS(3041),
    [anon_sym_BANG] = ACTIONS(3041),
    [anon_sym_PLUS_PLUS] = ACTIONS(3041),
    [anon_sym_DASH_DASH] = ACTIONS(3041),
    [anon_sym_BSLASH] = ACTIONS(3039),
    [anon_sym_do] = ACTIONS(3041),
    [anon_sym_eval] = ACTIONS(3041),
    [anon_sym_await] = ACTIONS(3041),
    [anon_sym_field] = ACTIONS(3041),
    [anon_sym_local] = ACTIONS(3041),
    [anon_sym_dynamically] = ACTIONS(3041),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3041),
    [anon_sym_require] = ACTIONS(3041),
    [anon_sym_map] = ACTIONS(3041),
    [anon_sym_grep] = ACTIONS(3041),
    [anon_sym_sort] = ACTIONS(3041),
    [anon_sym_goto] = ACTIONS(3041),
    [anon_sym_return] = ACTIONS(3041),
    [anon_sym_undef] = ACTIONS(3041),
    [anon_sym_PERCENT2] = ACTIONS(3041),
    [anon_sym_AMP2] = ACTIONS(3041),
    [anon_sym_STAR2] = ACTIONS(3041),
    [anon_sym_last] = ACTIONS(3041),
    [anon_sym_next] = ACTIONS(3041),
    [anon_sym_redo] = ACTIONS(3041),
    [anon_sym___FILE__] = ACTIONS(3041),
    [anon_sym___LINE__] = ACTIONS(3041),
    [anon_sym___PACKAGE__] = ACTIONS(3041),
    [anon_sym___SUB__] = ACTIONS(3041),
    [anon_sym_break] = ACTIONS(3041),
    [anon_sym_fork] = ACTIONS(3041),
    [anon_sym_getppid] = ACTIONS(3041),
    [anon_sym_time] = ACTIONS(3041),
    [anon_sym_times] = ACTIONS(3041),
    [anon_sym_wait] = ACTIONS(3041),
    [anon_sym_wantarray] = ACTIONS(3041),
    [anon_sym_abs] = ACTIONS(3041),
    [anon_sym_alarm] = ACTIONS(3041),
    [anon_sym_chop] = ACTIONS(3041),
    [anon_sym_chdir] = ACTIONS(3041),
    [anon_sym_close] = ACTIONS(3041),
    [anon_sym_closedir] = ACTIONS(3041),
    [anon_sym_caller] = ACTIONS(3041),
    [anon_sym_chomp] = ACTIONS(3041),
    [anon_sym_chr] = ACTIONS(3041),
    [anon_sym_cos] = ACTIONS(3041),
    [anon_sym_chroot] = ACTIONS(3041),
    [anon_sym_defined] = ACTIONS(3041),
    [anon_sym_delete] = ACTIONS(3041),
    [anon_sym_dbmclose] = ACTIONS(3041),
    [anon_sym_exists] = ACTIONS(3041),
    [anon_sym_exit] = ACTIONS(3041),
    [anon_sym_eof] = ACTIONS(3041),
    [anon_sym_exp] = ACTIONS(3041),
    [anon_sym_each] = ACTIONS(3041),
    [anon_sym_fc] = ACTIONS(3041),
    [anon_sym_fileno] = ACTIONS(3041),
    [anon_sym_gmtime] = ACTIONS(3041),
    [anon_sym_getc] = ACTIONS(3041),
    [anon_sym_getpgrp] = ACTIONS(3041),
    [anon_sym_getprotobyname] = ACTIONS(3041),
    [anon_sym_getpwname] = ACTIONS(3041),
    [anon_sym_getpwuid] = ACTIONS(3041),
    [anon_sym_getpeername] = ACTIONS(3041),
    [anon_sym_getnetbyname] = ACTIONS(3041),
    [anon_sym_getsockname] = ACTIONS(3041),
    [anon_sym_getgrnam] = ACTIONS(3041),
    [anon_sym_getgrgid] = ACTIONS(3041),
    [anon_sym_hex] = ACTIONS(3041),
    [anon_sym_int] = ACTIONS(3041),
    [anon_sym_keys] = ACTIONS(3041),
    [anon_sym_lc] = ACTIONS(3041),
    [anon_sym_lcfirst] = ACTIONS(3041),
    [anon_sym_length] = ACTIONS(3041),
    [anon_sym_localtime] = ACTIONS(3041),
    [anon_sym_log] = ACTIONS(3041),
    [anon_sym_lock] = ACTIONS(3041),
    [anon_sym_lstat] = ACTIONS(3041),
    [anon_sym_oct] = ACTIONS(3041),
    [anon_sym_ord] = ACTIONS(3041),
    [anon_sym_prototype] = ACTIONS(3041),
    [anon_sym_pop] = ACTIONS(3041),
    [anon_sym_pos] = ACTIONS(3041),
    [anon_sym_quotemeta] = ACTIONS(3041),
    [anon_sym_reset] = ACTIONS(3041),
    [anon_sym_rand] = ACTIONS(3041),
    [anon_sym_rmdir] = ACTIONS(3041),
    [anon_sym_readdir] = ACTIONS(3041),
    [anon_sym_readline] = ACTIONS(3041),
    [anon_sym_readpipe] = ACTIONS(3041),
    [anon_sym_rewinddir] = ACTIONS(3041),
    [anon_sym_readlink] = ACTIONS(3041),
    [anon_sym_ref] = ACTIONS(3041),
    [anon_sym_scalar] = ACTIONS(3041),
    [anon_sym_shift] = ACTIONS(3041),
    [anon_sym_sin] = ACTIONS(3041),
    [anon_sym_sleep] = ACTIONS(3041),
    [anon_sym_sqrt] = ACTIONS(3041),
    [anon_sym_srand] = ACTIONS(3041),
    [anon_sym_stat] = ACTIONS(3041),
    [anon_sym_study] = ACTIONS(3041),
    [anon_sym_tell] = ACTIONS(3041),
    [anon_sym_telldir] = ACTIONS(3041),
    [anon_sym_tied] = ACTIONS(3041),
    [anon_sym_uc] = ACTIONS(3041),
    [anon_sym_ucfirst] = ACTIONS(3041),
    [anon_sym_untie] = ACTIONS(3041),
    [anon_sym_umask] = ACTIONS(3041),
    [anon_sym_values] = ACTIONS(3041),
    [anon_sym_write] = ACTIONS(3041),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3041),
    [anon_sym_qq] = ACTIONS(3041),
    [anon_sym_qw] = ACTIONS(3041),
    [anon_sym_qx] = ACTIONS(3041),
    [anon_sym_qr] = ACTIONS(3041),
    [anon_sym_m] = ACTIONS(3041),
    [anon_sym_s] = ACTIONS(3041),
    [anon_sym_tr] = ACTIONS(3041),
    [anon_sym_y] = ACTIONS(3041),
    [aux_sym__bareword_token1] = ACTIONS(3041),
    [sym_number] = ACTIONS(3041),
    [anon_sym_true] = ACTIONS(3041),
    [anon_sym_false] = ACTIONS(3041),
    [sym__single_quote] = ACTIONS(3039),
    [sym__double_quote] = ACTIONS(3039),
    [sym__backtick_quote] = ACTIONS(3039),
    [sym__search_slash_quote] = ACTIONS(3039),
    [sym__open_readline_bracket] = ACTIONS(3039),
    [sym__open_fileglob_bracket] = ACTIONS(3039),
    [sym__PERLY_HEREDOC] = ACTIONS(3039),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3039),
    [sym__filetest] = ACTIONS(3039),
  },
  [STATE(1711)] = {
    [sym_heredoc_content] = STATE(1711),
    [sym__identifier] = ACTIONS(3185),
    [anon_sym_LBRACE] = ACTIONS(3185),
    [anon_sym_DOLLAR] = ACTIONS(3185),
    [anon_sym_AT] = ACTIONS(3185),
    [anon_sym_RPAREN] = ACTIONS(3185),
    [anon_sym_my] = ACTIONS(3185),
    [anon_sym_extended] = ACTIONS(3185),
    [anon_sym_async] = ACTIONS(3185),
    [anon_sym_sub] = ACTIONS(3185),
    [anon_sym_method] = ACTIONS(3185),
    [anon_sym_LPAREN] = ACTIONS(3185),
    [anon_sym_continue] = ACTIONS(3185),
    [anon_sym_state] = ACTIONS(3185),
    [anon_sym_our] = ACTIONS(3185),
    [anon_sym_LBRACK] = ACTIONS(3185),
    [anon_sym_LT_LT] = ACTIONS(3185),
    [anon_sym_SLASH_SLASH] = ACTIONS(3185),
    [anon_sym_PLUS] = ACTIONS(3185),
    [anon_sym_DASH] = ACTIONS(3185),
    [anon_sym_TILDE] = ACTIONS(3185),
    [anon_sym_BANG] = ACTIONS(3185),
    [anon_sym_PLUS_PLUS] = ACTIONS(3185),
    [anon_sym_DASH_DASH] = ACTIONS(3185),
    [anon_sym_BSLASH] = ACTIONS(3183),
    [anon_sym_do] = ACTIONS(3185),
    [anon_sym_eval] = ACTIONS(3185),
    [anon_sym_await] = ACTIONS(3185),
    [anon_sym_field] = ACTIONS(3185),
    [anon_sym_local] = ACTIONS(3185),
    [anon_sym_dynamically] = ACTIONS(3185),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3185),
    [anon_sym_require] = ACTIONS(3185),
    [anon_sym_map] = ACTIONS(3185),
    [anon_sym_grep] = ACTIONS(3185),
    [anon_sym_sort] = ACTIONS(3185),
    [anon_sym_goto] = ACTIONS(3185),
    [anon_sym_return] = ACTIONS(3185),
    [anon_sym_undef] = ACTIONS(3185),
    [anon_sym_PERCENT2] = ACTIONS(3185),
    [anon_sym_AMP2] = ACTIONS(3185),
    [anon_sym_STAR2] = ACTIONS(3185),
    [anon_sym_last] = ACTIONS(3185),
    [anon_sym_next] = ACTIONS(3185),
    [anon_sym_redo] = ACTIONS(3185),
    [anon_sym___FILE__] = ACTIONS(3185),
    [anon_sym___LINE__] = ACTIONS(3185),
    [anon_sym___PACKAGE__] = ACTIONS(3185),
    [anon_sym___SUB__] = ACTIONS(3185),
    [anon_sym_break] = ACTIONS(3185),
    [anon_sym_fork] = ACTIONS(3185),
    [anon_sym_getppid] = ACTIONS(3185),
    [anon_sym_time] = ACTIONS(3185),
    [anon_sym_times] = ACTIONS(3185),
    [anon_sym_wait] = ACTIONS(3185),
    [anon_sym_wantarray] = ACTIONS(3185),
    [anon_sym_abs] = ACTIONS(3185),
    [anon_sym_alarm] = ACTIONS(3185),
    [anon_sym_chop] = ACTIONS(3185),
    [anon_sym_chdir] = ACTIONS(3185),
    [anon_sym_close] = ACTIONS(3185),
    [anon_sym_closedir] = ACTIONS(3185),
    [anon_sym_caller] = ACTIONS(3185),
    [anon_sym_chomp] = ACTIONS(3185),
    [anon_sym_chr] = ACTIONS(3185),
    [anon_sym_cos] = ACTIONS(3185),
    [anon_sym_chroot] = ACTIONS(3185),
    [anon_sym_defined] = ACTIONS(3185),
    [anon_sym_delete] = ACTIONS(3185),
    [anon_sym_dbmclose] = ACTIONS(3185),
    [anon_sym_exists] = ACTIONS(3185),
    [anon_sym_exit] = ACTIONS(3185),
    [anon_sym_eof] = ACTIONS(3185),
    [anon_sym_exp] = ACTIONS(3185),
    [anon_sym_each] = ACTIONS(3185),
    [anon_sym_fc] = ACTIONS(3185),
    [anon_sym_fileno] = ACTIONS(3185),
    [anon_sym_gmtime] = ACTIONS(3185),
    [anon_sym_getc] = ACTIONS(3185),
    [anon_sym_getpgrp] = ACTIONS(3185),
    [anon_sym_getprotobyname] = ACTIONS(3185),
    [anon_sym_getpwname] = ACTIONS(3185),
    [anon_sym_getpwuid] = ACTIONS(3185),
    [anon_sym_getpeername] = ACTIONS(3185),
    [anon_sym_getnetbyname] = ACTIONS(3185),
    [anon_sym_getsockname] = ACTIONS(3185),
    [anon_sym_getgrnam] = ACTIONS(3185),
    [anon_sym_getgrgid] = ACTIONS(3185),
    [anon_sym_hex] = ACTIONS(3185),
    [anon_sym_int] = ACTIONS(3185),
    [anon_sym_keys] = ACTIONS(3185),
    [anon_sym_lc] = ACTIONS(3185),
    [anon_sym_lcfirst] = ACTIONS(3185),
    [anon_sym_length] = ACTIONS(3185),
    [anon_sym_localtime] = ACTIONS(3185),
    [anon_sym_log] = ACTIONS(3185),
    [anon_sym_lock] = ACTIONS(3185),
    [anon_sym_lstat] = ACTIONS(3185),
    [anon_sym_oct] = ACTIONS(3185),
    [anon_sym_ord] = ACTIONS(3185),
    [anon_sym_prototype] = ACTIONS(3185),
    [anon_sym_pop] = ACTIONS(3185),
    [anon_sym_pos] = ACTIONS(3185),
    [anon_sym_quotemeta] = ACTIONS(3185),
    [anon_sym_reset] = ACTIONS(3185),
    [anon_sym_rand] = ACTIONS(3185),
    [anon_sym_rmdir] = ACTIONS(3185),
    [anon_sym_readdir] = ACTIONS(3185),
    [anon_sym_readline] = ACTIONS(3185),
    [anon_sym_readpipe] = ACTIONS(3185),
    [anon_sym_rewinddir] = ACTIONS(3185),
    [anon_sym_readlink] = ACTIONS(3185),
    [anon_sym_ref] = ACTIONS(3185),
    [anon_sym_scalar] = ACTIONS(3185),
    [anon_sym_shift] = ACTIONS(3185),
    [anon_sym_sin] = ACTIONS(3185),
    [anon_sym_sleep] = ACTIONS(3185),
    [anon_sym_sqrt] = ACTIONS(3185),
    [anon_sym_srand] = ACTIONS(3185),
    [anon_sym_stat] = ACTIONS(3185),
    [anon_sym_study] = ACTIONS(3185),
    [anon_sym_tell] = ACTIONS(3185),
    [anon_sym_telldir] = ACTIONS(3185),
    [anon_sym_tied] = ACTIONS(3185),
    [anon_sym_uc] = ACTIONS(3185),
    [anon_sym_ucfirst] = ACTIONS(3185),
    [anon_sym_untie] = ACTIONS(3185),
    [anon_sym_umask] = ACTIONS(3185),
    [anon_sym_values] = ACTIONS(3185),
    [anon_sym_write] = ACTIONS(3185),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3185),
    [anon_sym_qq] = ACTIONS(3185),
    [anon_sym_qw] = ACTIONS(3185),
    [anon_sym_qx] = ACTIONS(3185),
    [anon_sym_qr] = ACTIONS(3185),
    [anon_sym_m] = ACTIONS(3185),
    [anon_sym_s] = ACTIONS(3185),
    [anon_sym_tr] = ACTIONS(3185),
    [anon_sym_y] = ACTIONS(3185),
    [aux_sym__bareword_token1] = ACTIONS(3185),
    [sym_number] = ACTIONS(3185),
    [anon_sym_true] = ACTIONS(3185),
    [anon_sym_false] = ACTIONS(3185),
    [sym__single_quote] = ACTIONS(3183),
    [sym__double_quote] = ACTIONS(3183),
    [sym__backtick_quote] = ACTIONS(3183),
    [sym__search_slash_quote] = ACTIONS(3183),
    [sym__open_readline_bracket] = ACTIONS(3183),
    [sym__open_fileglob_bracket] = ACTIONS(3183),
    [sym__PERLY_HEREDOC] = ACTIONS(3183),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3183),
    [sym__filetest] = ACTIONS(3183),
  },
  [STATE(1712)] = {
    [sym_heredoc_content] = STATE(1712),
    [sym__identifier] = ACTIONS(3153),
    [anon_sym_LBRACE] = ACTIONS(3153),
    [anon_sym_DOLLAR] = ACTIONS(3153),
    [anon_sym_AT] = ACTIONS(3153),
    [anon_sym_RPAREN] = ACTIONS(3153),
    [anon_sym_my] = ACTIONS(3153),
    [anon_sym_extended] = ACTIONS(3153),
    [anon_sym_async] = ACTIONS(3153),
    [anon_sym_sub] = ACTIONS(3153),
    [anon_sym_method] = ACTIONS(3153),
    [anon_sym_LPAREN] = ACTIONS(3153),
    [anon_sym_continue] = ACTIONS(3153),
    [anon_sym_state] = ACTIONS(3153),
    [anon_sym_our] = ACTIONS(3153),
    [anon_sym_LBRACK] = ACTIONS(3153),
    [anon_sym_LT_LT] = ACTIONS(3153),
    [anon_sym_SLASH_SLASH] = ACTIONS(3153),
    [anon_sym_PLUS] = ACTIONS(3153),
    [anon_sym_DASH] = ACTIONS(3153),
    [anon_sym_TILDE] = ACTIONS(3153),
    [anon_sym_BANG] = ACTIONS(3153),
    [anon_sym_PLUS_PLUS] = ACTIONS(3153),
    [anon_sym_DASH_DASH] = ACTIONS(3153),
    [anon_sym_BSLASH] = ACTIONS(3151),
    [anon_sym_do] = ACTIONS(3153),
    [anon_sym_eval] = ACTIONS(3153),
    [anon_sym_await] = ACTIONS(3153),
    [anon_sym_field] = ACTIONS(3153),
    [anon_sym_local] = ACTIONS(3153),
    [anon_sym_dynamically] = ACTIONS(3153),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3153),
    [anon_sym_require] = ACTIONS(3153),
    [anon_sym_map] = ACTIONS(3153),
    [anon_sym_grep] = ACTIONS(3153),
    [anon_sym_sort] = ACTIONS(3153),
    [anon_sym_goto] = ACTIONS(3153),
    [anon_sym_return] = ACTIONS(3153),
    [anon_sym_undef] = ACTIONS(3153),
    [anon_sym_PERCENT2] = ACTIONS(3153),
    [anon_sym_AMP2] = ACTIONS(3153),
    [anon_sym_STAR2] = ACTIONS(3153),
    [anon_sym_last] = ACTIONS(3153),
    [anon_sym_next] = ACTIONS(3153),
    [anon_sym_redo] = ACTIONS(3153),
    [anon_sym___FILE__] = ACTIONS(3153),
    [anon_sym___LINE__] = ACTIONS(3153),
    [anon_sym___PACKAGE__] = ACTIONS(3153),
    [anon_sym___SUB__] = ACTIONS(3153),
    [anon_sym_break] = ACTIONS(3153),
    [anon_sym_fork] = ACTIONS(3153),
    [anon_sym_getppid] = ACTIONS(3153),
    [anon_sym_time] = ACTIONS(3153),
    [anon_sym_times] = ACTIONS(3153),
    [anon_sym_wait] = ACTIONS(3153),
    [anon_sym_wantarray] = ACTIONS(3153),
    [anon_sym_abs] = ACTIONS(3153),
    [anon_sym_alarm] = ACTIONS(3153),
    [anon_sym_chop] = ACTIONS(3153),
    [anon_sym_chdir] = ACTIONS(3153),
    [anon_sym_close] = ACTIONS(3153),
    [anon_sym_closedir] = ACTIONS(3153),
    [anon_sym_caller] = ACTIONS(3153),
    [anon_sym_chomp] = ACTIONS(3153),
    [anon_sym_chr] = ACTIONS(3153),
    [anon_sym_cos] = ACTIONS(3153),
    [anon_sym_chroot] = ACTIONS(3153),
    [anon_sym_defined] = ACTIONS(3153),
    [anon_sym_delete] = ACTIONS(3153),
    [anon_sym_dbmclose] = ACTIONS(3153),
    [anon_sym_exists] = ACTIONS(3153),
    [anon_sym_exit] = ACTIONS(3153),
    [anon_sym_eof] = ACTIONS(3153),
    [anon_sym_exp] = ACTIONS(3153),
    [anon_sym_each] = ACTIONS(3153),
    [anon_sym_fc] = ACTIONS(3153),
    [anon_sym_fileno] = ACTIONS(3153),
    [anon_sym_gmtime] = ACTIONS(3153),
    [anon_sym_getc] = ACTIONS(3153),
    [anon_sym_getpgrp] = ACTIONS(3153),
    [anon_sym_getprotobyname] = ACTIONS(3153),
    [anon_sym_getpwname] = ACTIONS(3153),
    [anon_sym_getpwuid] = ACTIONS(3153),
    [anon_sym_getpeername] = ACTIONS(3153),
    [anon_sym_getnetbyname] = ACTIONS(3153),
    [anon_sym_getsockname] = ACTIONS(3153),
    [anon_sym_getgrnam] = ACTIONS(3153),
    [anon_sym_getgrgid] = ACTIONS(3153),
    [anon_sym_hex] = ACTIONS(3153),
    [anon_sym_int] = ACTIONS(3153),
    [anon_sym_keys] = ACTIONS(3153),
    [anon_sym_lc] = ACTIONS(3153),
    [anon_sym_lcfirst] = ACTIONS(3153),
    [anon_sym_length] = ACTIONS(3153),
    [anon_sym_localtime] = ACTIONS(3153),
    [anon_sym_log] = ACTIONS(3153),
    [anon_sym_lock] = ACTIONS(3153),
    [anon_sym_lstat] = ACTIONS(3153),
    [anon_sym_oct] = ACTIONS(3153),
    [anon_sym_ord] = ACTIONS(3153),
    [anon_sym_prototype] = ACTIONS(3153),
    [anon_sym_pop] = ACTIONS(3153),
    [anon_sym_pos] = ACTIONS(3153),
    [anon_sym_quotemeta] = ACTIONS(3153),
    [anon_sym_reset] = ACTIONS(3153),
    [anon_sym_rand] = ACTIONS(3153),
    [anon_sym_rmdir] = ACTIONS(3153),
    [anon_sym_readdir] = ACTIONS(3153),
    [anon_sym_readline] = ACTIONS(3153),
    [anon_sym_readpipe] = ACTIONS(3153),
    [anon_sym_rewinddir] = ACTIONS(3153),
    [anon_sym_readlink] = ACTIONS(3153),
    [anon_sym_ref] = ACTIONS(3153),
    [anon_sym_scalar] = ACTIONS(3153),
    [anon_sym_shift] = ACTIONS(3153),
    [anon_sym_sin] = ACTIONS(3153),
    [anon_sym_sleep] = ACTIONS(3153),
    [anon_sym_sqrt] = ACTIONS(3153),
    [anon_sym_srand] = ACTIONS(3153),
    [anon_sym_stat] = ACTIONS(3153),
    [anon_sym_study] = ACTIONS(3153),
    [anon_sym_tell] = ACTIONS(3153),
    [anon_sym_telldir] = ACTIONS(3153),
    [anon_sym_tied] = ACTIONS(3153),
    [anon_sym_uc] = ACTIONS(3153),
    [anon_sym_ucfirst] = ACTIONS(3153),
    [anon_sym_untie] = ACTIONS(3153),
    [anon_sym_umask] = ACTIONS(3153),
    [anon_sym_values] = ACTIONS(3153),
    [anon_sym_write] = ACTIONS(3153),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3153),
    [anon_sym_qq] = ACTIONS(3153),
    [anon_sym_qw] = ACTIONS(3153),
    [anon_sym_qx] = ACTIONS(3153),
    [anon_sym_qr] = ACTIONS(3153),
    [anon_sym_m] = ACTIONS(3153),
    [anon_sym_s] = ACTIONS(3153),
    [anon_sym_tr] = ACTIONS(3153),
    [anon_sym_y] = ACTIONS(3153),
    [aux_sym__bareword_token1] = ACTIONS(3153),
    [sym_number] = ACTIONS(3153),
    [anon_sym_true] = ACTIONS(3153),
    [anon_sym_false] = ACTIONS(3153),
    [sym__single_quote] = ACTIONS(3151),
    [sym__double_quote] = ACTIONS(3151),
    [sym__backtick_quote] = ACTIONS(3151),
    [sym__search_slash_quote] = ACTIONS(3151),
    [sym__open_readline_bracket] = ACTIONS(3151),
    [sym__open_fileglob_bracket] = ACTIONS(3151),
    [sym__PERLY_HEREDOC] = ACTIONS(3151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3151),
    [sym__filetest] = ACTIONS(3151),
  },
  [STATE(1713)] = {
    [sym_heredoc_content] = STATE(1713),
    [sym__identifier] = ACTIONS(3189),
    [anon_sym_LBRACE] = ACTIONS(3189),
    [anon_sym_DOLLAR] = ACTIONS(3189),
    [anon_sym_AT] = ACTIONS(3189),
    [anon_sym_RPAREN] = ACTIONS(3189),
    [anon_sym_my] = ACTIONS(3189),
    [anon_sym_extended] = ACTIONS(3189),
    [anon_sym_async] = ACTIONS(3189),
    [anon_sym_sub] = ACTIONS(3189),
    [anon_sym_method] = ACTIONS(3189),
    [anon_sym_LPAREN] = ACTIONS(3189),
    [anon_sym_continue] = ACTIONS(3189),
    [anon_sym_state] = ACTIONS(3189),
    [anon_sym_our] = ACTIONS(3189),
    [anon_sym_LBRACK] = ACTIONS(3189),
    [anon_sym_LT_LT] = ACTIONS(3189),
    [anon_sym_SLASH_SLASH] = ACTIONS(3189),
    [anon_sym_PLUS] = ACTIONS(3189),
    [anon_sym_DASH] = ACTIONS(3189),
    [anon_sym_TILDE] = ACTIONS(3189),
    [anon_sym_BANG] = ACTIONS(3189),
    [anon_sym_PLUS_PLUS] = ACTIONS(3189),
    [anon_sym_DASH_DASH] = ACTIONS(3189),
    [anon_sym_BSLASH] = ACTIONS(3187),
    [anon_sym_do] = ACTIONS(3189),
    [anon_sym_eval] = ACTIONS(3189),
    [anon_sym_await] = ACTIONS(3189),
    [anon_sym_field] = ACTIONS(3189),
    [anon_sym_local] = ACTIONS(3189),
    [anon_sym_dynamically] = ACTIONS(3189),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3189),
    [anon_sym_require] = ACTIONS(3189),
    [anon_sym_map] = ACTIONS(3189),
    [anon_sym_grep] = ACTIONS(3189),
    [anon_sym_sort] = ACTIONS(3189),
    [anon_sym_goto] = ACTIONS(3189),
    [anon_sym_return] = ACTIONS(3189),
    [anon_sym_undef] = ACTIONS(3189),
    [anon_sym_PERCENT2] = ACTIONS(3189),
    [anon_sym_AMP2] = ACTIONS(3189),
    [anon_sym_STAR2] = ACTIONS(3189),
    [anon_sym_last] = ACTIONS(3189),
    [anon_sym_next] = ACTIONS(3189),
    [anon_sym_redo] = ACTIONS(3189),
    [anon_sym___FILE__] = ACTIONS(3189),
    [anon_sym___LINE__] = ACTIONS(3189),
    [anon_sym___PACKAGE__] = ACTIONS(3189),
    [anon_sym___SUB__] = ACTIONS(3189),
    [anon_sym_break] = ACTIONS(3189),
    [anon_sym_fork] = ACTIONS(3189),
    [anon_sym_getppid] = ACTIONS(3189),
    [anon_sym_time] = ACTIONS(3189),
    [anon_sym_times] = ACTIONS(3189),
    [anon_sym_wait] = ACTIONS(3189),
    [anon_sym_wantarray] = ACTIONS(3189),
    [anon_sym_abs] = ACTIONS(3189),
    [anon_sym_alarm] = ACTIONS(3189),
    [anon_sym_chop] = ACTIONS(3189),
    [anon_sym_chdir] = ACTIONS(3189),
    [anon_sym_close] = ACTIONS(3189),
    [anon_sym_closedir] = ACTIONS(3189),
    [anon_sym_caller] = ACTIONS(3189),
    [anon_sym_chomp] = ACTIONS(3189),
    [anon_sym_chr] = ACTIONS(3189),
    [anon_sym_cos] = ACTIONS(3189),
    [anon_sym_chroot] = ACTIONS(3189),
    [anon_sym_defined] = ACTIONS(3189),
    [anon_sym_delete] = ACTIONS(3189),
    [anon_sym_dbmclose] = ACTIONS(3189),
    [anon_sym_exists] = ACTIONS(3189),
    [anon_sym_exit] = ACTIONS(3189),
    [anon_sym_eof] = ACTIONS(3189),
    [anon_sym_exp] = ACTIONS(3189),
    [anon_sym_each] = ACTIONS(3189),
    [anon_sym_fc] = ACTIONS(3189),
    [anon_sym_fileno] = ACTIONS(3189),
    [anon_sym_gmtime] = ACTIONS(3189),
    [anon_sym_getc] = ACTIONS(3189),
    [anon_sym_getpgrp] = ACTIONS(3189),
    [anon_sym_getprotobyname] = ACTIONS(3189),
    [anon_sym_getpwname] = ACTIONS(3189),
    [anon_sym_getpwuid] = ACTIONS(3189),
    [anon_sym_getpeername] = ACTIONS(3189),
    [anon_sym_getnetbyname] = ACTIONS(3189),
    [anon_sym_getsockname] = ACTIONS(3189),
    [anon_sym_getgrnam] = ACTIONS(3189),
    [anon_sym_getgrgid] = ACTIONS(3189),
    [anon_sym_hex] = ACTIONS(3189),
    [anon_sym_int] = ACTIONS(3189),
    [anon_sym_keys] = ACTIONS(3189),
    [anon_sym_lc] = ACTIONS(3189),
    [anon_sym_lcfirst] = ACTIONS(3189),
    [anon_sym_length] = ACTIONS(3189),
    [anon_sym_localtime] = ACTIONS(3189),
    [anon_sym_log] = ACTIONS(3189),
    [anon_sym_lock] = ACTIONS(3189),
    [anon_sym_lstat] = ACTIONS(3189),
    [anon_sym_oct] = ACTIONS(3189),
    [anon_sym_ord] = ACTIONS(3189),
    [anon_sym_prototype] = ACTIONS(3189),
    [anon_sym_pop] = ACTIONS(3189),
    [anon_sym_pos] = ACTIONS(3189),
    [anon_sym_quotemeta] = ACTIONS(3189),
    [anon_sym_reset] = ACTIONS(3189),
    [anon_sym_rand] = ACTIONS(3189),
    [anon_sym_rmdir] = ACTIONS(3189),
    [anon_sym_readdir] = ACTIONS(3189),
    [anon_sym_readline] = ACTIONS(3189),
    [anon_sym_readpipe] = ACTIONS(3189),
    [anon_sym_rewinddir] = ACTIONS(3189),
    [anon_sym_readlink] = ACTIONS(3189),
    [anon_sym_ref] = ACTIONS(3189),
    [anon_sym_scalar] = ACTIONS(3189),
    [anon_sym_shift] = ACTIONS(3189),
    [anon_sym_sin] = ACTIONS(3189),
    [anon_sym_sleep] = ACTIONS(3189),
    [anon_sym_sqrt] = ACTIONS(3189),
    [anon_sym_srand] = ACTIONS(3189),
    [anon_sym_stat] = ACTIONS(3189),
    [anon_sym_study] = ACTIONS(3189),
    [anon_sym_tell] = ACTIONS(3189),
    [anon_sym_telldir] = ACTIONS(3189),
    [anon_sym_tied] = ACTIONS(3189),
    [anon_sym_uc] = ACTIONS(3189),
    [anon_sym_ucfirst] = ACTIONS(3189),
    [anon_sym_untie] = ACTIONS(3189),
    [anon_sym_umask] = ACTIONS(3189),
    [anon_sym_values] = ACTIONS(3189),
    [anon_sym_write] = ACTIONS(3189),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3189),
    [anon_sym_qq] = ACTIONS(3189),
    [anon_sym_qw] = ACTIONS(3189),
    [anon_sym_qx] = ACTIONS(3189),
    [anon_sym_qr] = ACTIONS(3189),
    [anon_sym_m] = ACTIONS(3189),
    [anon_sym_s] = ACTIONS(3189),
    [anon_sym_tr] = ACTIONS(3189),
    [anon_sym_y] = ACTIONS(3189),
    [aux_sym__bareword_token1] = ACTIONS(3189),
    [sym_number] = ACTIONS(3189),
    [anon_sym_true] = ACTIONS(3189),
    [anon_sym_false] = ACTIONS(3189),
    [sym__single_quote] = ACTIONS(3187),
    [sym__double_quote] = ACTIONS(3187),
    [sym__backtick_quote] = ACTIONS(3187),
    [sym__search_slash_quote] = ACTIONS(3187),
    [sym__open_readline_bracket] = ACTIONS(3187),
    [sym__open_fileglob_bracket] = ACTIONS(3187),
    [sym__PERLY_HEREDOC] = ACTIONS(3187),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3187),
    [sym__filetest] = ACTIONS(3187),
  },
  [STATE(1714)] = {
    [sym_heredoc_content] = STATE(1714),
    [sym__identifier] = ACTIONS(3205),
    [anon_sym_LBRACE] = ACTIONS(3205),
    [anon_sym_DOLLAR] = ACTIONS(3205),
    [anon_sym_AT] = ACTIONS(3205),
    [anon_sym_RPAREN] = ACTIONS(3205),
    [anon_sym_my] = ACTIONS(3205),
    [anon_sym_extended] = ACTIONS(3205),
    [anon_sym_async] = ACTIONS(3205),
    [anon_sym_sub] = ACTIONS(3205),
    [anon_sym_method] = ACTIONS(3205),
    [anon_sym_LPAREN] = ACTIONS(3205),
    [anon_sym_continue] = ACTIONS(3205),
    [anon_sym_state] = ACTIONS(3205),
    [anon_sym_our] = ACTIONS(3205),
    [anon_sym_LBRACK] = ACTIONS(3205),
    [anon_sym_LT_LT] = ACTIONS(3205),
    [anon_sym_SLASH_SLASH] = ACTIONS(3205),
    [anon_sym_PLUS] = ACTIONS(3205),
    [anon_sym_DASH] = ACTIONS(3205),
    [anon_sym_TILDE] = ACTIONS(3205),
    [anon_sym_BANG] = ACTIONS(3205),
    [anon_sym_PLUS_PLUS] = ACTIONS(3205),
    [anon_sym_DASH_DASH] = ACTIONS(3205),
    [anon_sym_BSLASH] = ACTIONS(3203),
    [anon_sym_do] = ACTIONS(3205),
    [anon_sym_eval] = ACTIONS(3205),
    [anon_sym_await] = ACTIONS(3205),
    [anon_sym_field] = ACTIONS(3205),
    [anon_sym_local] = ACTIONS(3205),
    [anon_sym_dynamically] = ACTIONS(3205),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3205),
    [anon_sym_require] = ACTIONS(3205),
    [anon_sym_map] = ACTIONS(3205),
    [anon_sym_grep] = ACTIONS(3205),
    [anon_sym_sort] = ACTIONS(3205),
    [anon_sym_goto] = ACTIONS(3205),
    [anon_sym_return] = ACTIONS(3205),
    [anon_sym_undef] = ACTIONS(3205),
    [anon_sym_PERCENT2] = ACTIONS(3205),
    [anon_sym_AMP2] = ACTIONS(3205),
    [anon_sym_STAR2] = ACTIONS(3205),
    [anon_sym_last] = ACTIONS(3205),
    [anon_sym_next] = ACTIONS(3205),
    [anon_sym_redo] = ACTIONS(3205),
    [anon_sym___FILE__] = ACTIONS(3205),
    [anon_sym___LINE__] = ACTIONS(3205),
    [anon_sym___PACKAGE__] = ACTIONS(3205),
    [anon_sym___SUB__] = ACTIONS(3205),
    [anon_sym_break] = ACTIONS(3205),
    [anon_sym_fork] = ACTIONS(3205),
    [anon_sym_getppid] = ACTIONS(3205),
    [anon_sym_time] = ACTIONS(3205),
    [anon_sym_times] = ACTIONS(3205),
    [anon_sym_wait] = ACTIONS(3205),
    [anon_sym_wantarray] = ACTIONS(3205),
    [anon_sym_abs] = ACTIONS(3205),
    [anon_sym_alarm] = ACTIONS(3205),
    [anon_sym_chop] = ACTIONS(3205),
    [anon_sym_chdir] = ACTIONS(3205),
    [anon_sym_close] = ACTIONS(3205),
    [anon_sym_closedir] = ACTIONS(3205),
    [anon_sym_caller] = ACTIONS(3205),
    [anon_sym_chomp] = ACTIONS(3205),
    [anon_sym_chr] = ACTIONS(3205),
    [anon_sym_cos] = ACTIONS(3205),
    [anon_sym_chroot] = ACTIONS(3205),
    [anon_sym_defined] = ACTIONS(3205),
    [anon_sym_delete] = ACTIONS(3205),
    [anon_sym_dbmclose] = ACTIONS(3205),
    [anon_sym_exists] = ACTIONS(3205),
    [anon_sym_exit] = ACTIONS(3205),
    [anon_sym_eof] = ACTIONS(3205),
    [anon_sym_exp] = ACTIONS(3205),
    [anon_sym_each] = ACTIONS(3205),
    [anon_sym_fc] = ACTIONS(3205),
    [anon_sym_fileno] = ACTIONS(3205),
    [anon_sym_gmtime] = ACTIONS(3205),
    [anon_sym_getc] = ACTIONS(3205),
    [anon_sym_getpgrp] = ACTIONS(3205),
    [anon_sym_getprotobyname] = ACTIONS(3205),
    [anon_sym_getpwname] = ACTIONS(3205),
    [anon_sym_getpwuid] = ACTIONS(3205),
    [anon_sym_getpeername] = ACTIONS(3205),
    [anon_sym_getnetbyname] = ACTIONS(3205),
    [anon_sym_getsockname] = ACTIONS(3205),
    [anon_sym_getgrnam] = ACTIONS(3205),
    [anon_sym_getgrgid] = ACTIONS(3205),
    [anon_sym_hex] = ACTIONS(3205),
    [anon_sym_int] = ACTIONS(3205),
    [anon_sym_keys] = ACTIONS(3205),
    [anon_sym_lc] = ACTIONS(3205),
    [anon_sym_lcfirst] = ACTIONS(3205),
    [anon_sym_length] = ACTIONS(3205),
    [anon_sym_localtime] = ACTIONS(3205),
    [anon_sym_log] = ACTIONS(3205),
    [anon_sym_lock] = ACTIONS(3205),
    [anon_sym_lstat] = ACTIONS(3205),
    [anon_sym_oct] = ACTIONS(3205),
    [anon_sym_ord] = ACTIONS(3205),
    [anon_sym_prototype] = ACTIONS(3205),
    [anon_sym_pop] = ACTIONS(3205),
    [anon_sym_pos] = ACTIONS(3205),
    [anon_sym_quotemeta] = ACTIONS(3205),
    [anon_sym_reset] = ACTIONS(3205),
    [anon_sym_rand] = ACTIONS(3205),
    [anon_sym_rmdir] = ACTIONS(3205),
    [anon_sym_readdir] = ACTIONS(3205),
    [anon_sym_readline] = ACTIONS(3205),
    [anon_sym_readpipe] = ACTIONS(3205),
    [anon_sym_rewinddir] = ACTIONS(3205),
    [anon_sym_readlink] = ACTIONS(3205),
    [anon_sym_ref] = ACTIONS(3205),
    [anon_sym_scalar] = ACTIONS(3205),
    [anon_sym_shift] = ACTIONS(3205),
    [anon_sym_sin] = ACTIONS(3205),
    [anon_sym_sleep] = ACTIONS(3205),
    [anon_sym_sqrt] = ACTIONS(3205),
    [anon_sym_srand] = ACTIONS(3205),
    [anon_sym_stat] = ACTIONS(3205),
    [anon_sym_study] = ACTIONS(3205),
    [anon_sym_tell] = ACTIONS(3205),
    [anon_sym_telldir] = ACTIONS(3205),
    [anon_sym_tied] = ACTIONS(3205),
    [anon_sym_uc] = ACTIONS(3205),
    [anon_sym_ucfirst] = ACTIONS(3205),
    [anon_sym_untie] = ACTIONS(3205),
    [anon_sym_umask] = ACTIONS(3205),
    [anon_sym_values] = ACTIONS(3205),
    [anon_sym_write] = ACTIONS(3205),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3205),
    [anon_sym_qq] = ACTIONS(3205),
    [anon_sym_qw] = ACTIONS(3205),
    [anon_sym_qx] = ACTIONS(3205),
    [anon_sym_qr] = ACTIONS(3205),
    [anon_sym_m] = ACTIONS(3205),
    [anon_sym_s] = ACTIONS(3205),
    [anon_sym_tr] = ACTIONS(3205),
    [anon_sym_y] = ACTIONS(3205),
    [aux_sym__bareword_token1] = ACTIONS(3205),
    [sym_number] = ACTIONS(3205),
    [anon_sym_true] = ACTIONS(3205),
    [anon_sym_false] = ACTIONS(3205),
    [sym__single_quote] = ACTIONS(3203),
    [sym__double_quote] = ACTIONS(3203),
    [sym__backtick_quote] = ACTIONS(3203),
    [sym__search_slash_quote] = ACTIONS(3203),
    [sym__open_readline_bracket] = ACTIONS(3203),
    [sym__open_fileglob_bracket] = ACTIONS(3203),
    [sym__PERLY_HEREDOC] = ACTIONS(3203),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3203),
    [sym__filetest] = ACTIONS(3203),
  },
  [STATE(1715)] = {
    [sym_heredoc_content] = STATE(1715),
    [sym__identifier] = ACTIONS(2688),
    [anon_sym_LBRACE] = ACTIONS(2688),
    [anon_sym_DOLLAR] = ACTIONS(2688),
    [anon_sym_AT] = ACTIONS(2688),
    [anon_sym_RPAREN] = ACTIONS(2688),
    [anon_sym_my] = ACTIONS(2688),
    [anon_sym_extended] = ACTIONS(2688),
    [anon_sym_async] = ACTIONS(2688),
    [anon_sym_sub] = ACTIONS(2688),
    [anon_sym_method] = ACTIONS(2688),
    [anon_sym_LPAREN] = ACTIONS(2688),
    [anon_sym_continue] = ACTIONS(2688),
    [anon_sym_state] = ACTIONS(2688),
    [anon_sym_our] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_TILDE] = ACTIONS(2688),
    [anon_sym_BANG] = ACTIONS(2688),
    [anon_sym_PLUS_PLUS] = ACTIONS(2688),
    [anon_sym_DASH_DASH] = ACTIONS(2688),
    [anon_sym_BSLASH] = ACTIONS(2690),
    [anon_sym_do] = ACTIONS(2688),
    [anon_sym_eval] = ACTIONS(2688),
    [anon_sym_await] = ACTIONS(2688),
    [anon_sym_field] = ACTIONS(2688),
    [anon_sym_local] = ACTIONS(2688),
    [anon_sym_dynamically] = ACTIONS(2688),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2688),
    [anon_sym_require] = ACTIONS(2688),
    [anon_sym_map] = ACTIONS(2688),
    [anon_sym_grep] = ACTIONS(2688),
    [anon_sym_sort] = ACTIONS(2688),
    [anon_sym_goto] = ACTIONS(2688),
    [anon_sym_return] = ACTIONS(2688),
    [anon_sym_undef] = ACTIONS(2688),
    [anon_sym_PERCENT2] = ACTIONS(2688),
    [anon_sym_AMP2] = ACTIONS(2688),
    [anon_sym_STAR2] = ACTIONS(2688),
    [anon_sym_last] = ACTIONS(2688),
    [anon_sym_next] = ACTIONS(2688),
    [anon_sym_redo] = ACTIONS(2688),
    [anon_sym___FILE__] = ACTIONS(2688),
    [anon_sym___LINE__] = ACTIONS(2688),
    [anon_sym___PACKAGE__] = ACTIONS(2688),
    [anon_sym___SUB__] = ACTIONS(2688),
    [anon_sym_break] = ACTIONS(2688),
    [anon_sym_fork] = ACTIONS(2688),
    [anon_sym_getppid] = ACTIONS(2688),
    [anon_sym_time] = ACTIONS(2688),
    [anon_sym_times] = ACTIONS(2688),
    [anon_sym_wait] = ACTIONS(2688),
    [anon_sym_wantarray] = ACTIONS(2688),
    [anon_sym_abs] = ACTIONS(2688),
    [anon_sym_alarm] = ACTIONS(2688),
    [anon_sym_chop] = ACTIONS(2688),
    [anon_sym_chdir] = ACTIONS(2688),
    [anon_sym_close] = ACTIONS(2688),
    [anon_sym_closedir] = ACTIONS(2688),
    [anon_sym_caller] = ACTIONS(2688),
    [anon_sym_chomp] = ACTIONS(2688),
    [anon_sym_chr] = ACTIONS(2688),
    [anon_sym_cos] = ACTIONS(2688),
    [anon_sym_chroot] = ACTIONS(2688),
    [anon_sym_defined] = ACTIONS(2688),
    [anon_sym_delete] = ACTIONS(2688),
    [anon_sym_dbmclose] = ACTIONS(2688),
    [anon_sym_exists] = ACTIONS(2688),
    [anon_sym_exit] = ACTIONS(2688),
    [anon_sym_eof] = ACTIONS(2688),
    [anon_sym_exp] = ACTIONS(2688),
    [anon_sym_each] = ACTIONS(2688),
    [anon_sym_fc] = ACTIONS(2688),
    [anon_sym_fileno] = ACTIONS(2688),
    [anon_sym_gmtime] = ACTIONS(2688),
    [anon_sym_getc] = ACTIONS(2688),
    [anon_sym_getpgrp] = ACTIONS(2688),
    [anon_sym_getprotobyname] = ACTIONS(2688),
    [anon_sym_getpwname] = ACTIONS(2688),
    [anon_sym_getpwuid] = ACTIONS(2688),
    [anon_sym_getpeername] = ACTIONS(2688),
    [anon_sym_getnetbyname] = ACTIONS(2688),
    [anon_sym_getsockname] = ACTIONS(2688),
    [anon_sym_getgrnam] = ACTIONS(2688),
    [anon_sym_getgrgid] = ACTIONS(2688),
    [anon_sym_hex] = ACTIONS(2688),
    [anon_sym_int] = ACTIONS(2688),
    [anon_sym_keys] = ACTIONS(2688),
    [anon_sym_lc] = ACTIONS(2688),
    [anon_sym_lcfirst] = ACTIONS(2688),
    [anon_sym_length] = ACTIONS(2688),
    [anon_sym_localtime] = ACTIONS(2688),
    [anon_sym_log] = ACTIONS(2688),
    [anon_sym_lock] = ACTIONS(2688),
    [anon_sym_lstat] = ACTIONS(2688),
    [anon_sym_oct] = ACTIONS(2688),
    [anon_sym_ord] = ACTIONS(2688),
    [anon_sym_prototype] = ACTIONS(2688),
    [anon_sym_pop] = ACTIONS(2688),
    [anon_sym_pos] = ACTIONS(2688),
    [anon_sym_quotemeta] = ACTIONS(2688),
    [anon_sym_reset] = ACTIONS(2688),
    [anon_sym_rand] = ACTIONS(2688),
    [anon_sym_rmdir] = ACTIONS(2688),
    [anon_sym_readdir] = ACTIONS(2688),
    [anon_sym_readline] = ACTIONS(2688),
    [anon_sym_readpipe] = ACTIONS(2688),
    [anon_sym_rewinddir] = ACTIONS(2688),
    [anon_sym_readlink] = ACTIONS(2688),
    [anon_sym_ref] = ACTIONS(2688),
    [anon_sym_scalar] = ACTIONS(2688),
    [anon_sym_shift] = ACTIONS(2688),
    [anon_sym_sin] = ACTIONS(2688),
    [anon_sym_sleep] = ACTIONS(2688),
    [anon_sym_sqrt] = ACTIONS(2688),
    [anon_sym_srand] = ACTIONS(2688),
    [anon_sym_stat] = ACTIONS(2688),
    [anon_sym_study] = ACTIONS(2688),
    [anon_sym_tell] = ACTIONS(2688),
    [anon_sym_telldir] = ACTIONS(2688),
    [anon_sym_tied] = ACTIONS(2688),
    [anon_sym_uc] = ACTIONS(2688),
    [anon_sym_ucfirst] = ACTIONS(2688),
    [anon_sym_untie] = ACTIONS(2688),
    [anon_sym_umask] = ACTIONS(2688),
    [anon_sym_values] = ACTIONS(2688),
    [anon_sym_write] = ACTIONS(2688),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2688),
    [anon_sym_qq] = ACTIONS(2688),
    [anon_sym_qw] = ACTIONS(2688),
    [anon_sym_qx] = ACTIONS(2688),
    [anon_sym_qr] = ACTIONS(2688),
    [anon_sym_m] = ACTIONS(2688),
    [anon_sym_s] = ACTIONS(2688),
    [anon_sym_tr] = ACTIONS(2688),
    [anon_sym_y] = ACTIONS(2688),
    [aux_sym__bareword_token1] = ACTIONS(2688),
    [sym_number] = ACTIONS(2688),
    [anon_sym_true] = ACTIONS(2688),
    [anon_sym_false] = ACTIONS(2688),
    [sym__single_quote] = ACTIONS(2690),
    [sym__double_quote] = ACTIONS(2690),
    [sym__backtick_quote] = ACTIONS(2690),
    [sym__search_slash_quote] = ACTIONS(2690),
    [sym__open_readline_bracket] = ACTIONS(2690),
    [sym__open_fileglob_bracket] = ACTIONS(2690),
    [sym__PERLY_HEREDOC] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2690),
    [sym__filetest] = ACTIONS(2690),
  },
  [STATE(1716)] = {
    [sym_heredoc_content] = STATE(1716),
    [sym__identifier] = ACTIONS(3255),
    [anon_sym_LBRACE] = ACTIONS(3255),
    [anon_sym_RBRACE] = ACTIONS(3255),
    [anon_sym_DOLLAR] = ACTIONS(3255),
    [anon_sym_AT] = ACTIONS(3255),
    [anon_sym_my] = ACTIONS(3255),
    [anon_sym_extended] = ACTIONS(3255),
    [anon_sym_async] = ACTIONS(3255),
    [anon_sym_sub] = ACTIONS(3255),
    [anon_sym_method] = ACTIONS(3255),
    [anon_sym_LPAREN] = ACTIONS(3255),
    [anon_sym_continue] = ACTIONS(3255),
    [anon_sym_state] = ACTIONS(3255),
    [anon_sym_our] = ACTIONS(3255),
    [anon_sym_LBRACK] = ACTIONS(3255),
    [anon_sym_LT_LT] = ACTIONS(3255),
    [anon_sym_SLASH_SLASH] = ACTIONS(3255),
    [anon_sym_PLUS] = ACTIONS(3255),
    [anon_sym_DASH] = ACTIONS(3255),
    [anon_sym_TILDE] = ACTIONS(3255),
    [anon_sym_BANG] = ACTIONS(3255),
    [anon_sym_PLUS_PLUS] = ACTIONS(3255),
    [anon_sym_DASH_DASH] = ACTIONS(3255),
    [anon_sym_BSLASH] = ACTIONS(3257),
    [anon_sym_do] = ACTIONS(3255),
    [anon_sym_eval] = ACTIONS(3255),
    [anon_sym_await] = ACTIONS(3255),
    [anon_sym_field] = ACTIONS(3255),
    [anon_sym_local] = ACTIONS(3255),
    [anon_sym_dynamically] = ACTIONS(3255),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3255),
    [anon_sym_require] = ACTIONS(3255),
    [anon_sym_map] = ACTIONS(3255),
    [anon_sym_grep] = ACTIONS(3255),
    [anon_sym_sort] = ACTIONS(3255),
    [anon_sym_goto] = ACTIONS(3255),
    [anon_sym_return] = ACTIONS(3255),
    [anon_sym_undef] = ACTIONS(3255),
    [anon_sym_PERCENT2] = ACTIONS(3255),
    [anon_sym_AMP2] = ACTIONS(3255),
    [anon_sym_STAR2] = ACTIONS(3255),
    [anon_sym_last] = ACTIONS(3255),
    [anon_sym_next] = ACTIONS(3255),
    [anon_sym_redo] = ACTIONS(3255),
    [anon_sym___FILE__] = ACTIONS(3255),
    [anon_sym___LINE__] = ACTIONS(3255),
    [anon_sym___PACKAGE__] = ACTIONS(3255),
    [anon_sym___SUB__] = ACTIONS(3255),
    [anon_sym_break] = ACTIONS(3255),
    [anon_sym_fork] = ACTIONS(3255),
    [anon_sym_getppid] = ACTIONS(3255),
    [anon_sym_time] = ACTIONS(3255),
    [anon_sym_times] = ACTIONS(3255),
    [anon_sym_wait] = ACTIONS(3255),
    [anon_sym_wantarray] = ACTIONS(3255),
    [anon_sym_abs] = ACTIONS(3255),
    [anon_sym_alarm] = ACTIONS(3255),
    [anon_sym_chop] = ACTIONS(3255),
    [anon_sym_chdir] = ACTIONS(3255),
    [anon_sym_close] = ACTIONS(3255),
    [anon_sym_closedir] = ACTIONS(3255),
    [anon_sym_caller] = ACTIONS(3255),
    [anon_sym_chomp] = ACTIONS(3255),
    [anon_sym_chr] = ACTIONS(3255),
    [anon_sym_cos] = ACTIONS(3255),
    [anon_sym_chroot] = ACTIONS(3255),
    [anon_sym_defined] = ACTIONS(3255),
    [anon_sym_delete] = ACTIONS(3255),
    [anon_sym_dbmclose] = ACTIONS(3255),
    [anon_sym_exists] = ACTIONS(3255),
    [anon_sym_exit] = ACTIONS(3255),
    [anon_sym_eof] = ACTIONS(3255),
    [anon_sym_exp] = ACTIONS(3255),
    [anon_sym_each] = ACTIONS(3255),
    [anon_sym_fc] = ACTIONS(3255),
    [anon_sym_fileno] = ACTIONS(3255),
    [anon_sym_gmtime] = ACTIONS(3255),
    [anon_sym_getc] = ACTIONS(3255),
    [anon_sym_getpgrp] = ACTIONS(3255),
    [anon_sym_getprotobyname] = ACTIONS(3255),
    [anon_sym_getpwname] = ACTIONS(3255),
    [anon_sym_getpwuid] = ACTIONS(3255),
    [anon_sym_getpeername] = ACTIONS(3255),
    [anon_sym_getnetbyname] = ACTIONS(3255),
    [anon_sym_getsockname] = ACTIONS(3255),
    [anon_sym_getgrnam] = ACTIONS(3255),
    [anon_sym_getgrgid] = ACTIONS(3255),
    [anon_sym_hex] = ACTIONS(3255),
    [anon_sym_int] = ACTIONS(3255),
    [anon_sym_keys] = ACTIONS(3255),
    [anon_sym_lc] = ACTIONS(3255),
    [anon_sym_lcfirst] = ACTIONS(3255),
    [anon_sym_length] = ACTIONS(3255),
    [anon_sym_localtime] = ACTIONS(3255),
    [anon_sym_log] = ACTIONS(3255),
    [anon_sym_lock] = ACTIONS(3255),
    [anon_sym_lstat] = ACTIONS(3255),
    [anon_sym_oct] = ACTIONS(3255),
    [anon_sym_ord] = ACTIONS(3255),
    [anon_sym_prototype] = ACTIONS(3255),
    [anon_sym_pop] = ACTIONS(3255),
    [anon_sym_pos] = ACTIONS(3255),
    [anon_sym_quotemeta] = ACTIONS(3255),
    [anon_sym_reset] = ACTIONS(3255),
    [anon_sym_rand] = ACTIONS(3255),
    [anon_sym_rmdir] = ACTIONS(3255),
    [anon_sym_readdir] = ACTIONS(3255),
    [anon_sym_readline] = ACTIONS(3255),
    [anon_sym_readpipe] = ACTIONS(3255),
    [anon_sym_rewinddir] = ACTIONS(3255),
    [anon_sym_readlink] = ACTIONS(3255),
    [anon_sym_ref] = ACTIONS(3255),
    [anon_sym_scalar] = ACTIONS(3255),
    [anon_sym_shift] = ACTIONS(3255),
    [anon_sym_sin] = ACTIONS(3255),
    [anon_sym_sleep] = ACTIONS(3255),
    [anon_sym_sqrt] = ACTIONS(3255),
    [anon_sym_srand] = ACTIONS(3255),
    [anon_sym_stat] = ACTIONS(3255),
    [anon_sym_study] = ACTIONS(3255),
    [anon_sym_tell] = ACTIONS(3255),
    [anon_sym_telldir] = ACTIONS(3255),
    [anon_sym_tied] = ACTIONS(3255),
    [anon_sym_uc] = ACTIONS(3255),
    [anon_sym_ucfirst] = ACTIONS(3255),
    [anon_sym_untie] = ACTIONS(3255),
    [anon_sym_umask] = ACTIONS(3255),
    [anon_sym_values] = ACTIONS(3255),
    [anon_sym_write] = ACTIONS(3255),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3255),
    [anon_sym_qq] = ACTIONS(3255),
    [anon_sym_qw] = ACTIONS(3255),
    [anon_sym_qx] = ACTIONS(3255),
    [anon_sym_qr] = ACTIONS(3255),
    [anon_sym_m] = ACTIONS(3255),
    [anon_sym_s] = ACTIONS(3255),
    [anon_sym_tr] = ACTIONS(3255),
    [anon_sym_y] = ACTIONS(3255),
    [aux_sym__bareword_token1] = ACTIONS(3255),
    [sym_number] = ACTIONS(3255),
    [anon_sym_true] = ACTIONS(3255),
    [anon_sym_false] = ACTIONS(3255),
    [sym__single_quote] = ACTIONS(3257),
    [sym__double_quote] = ACTIONS(3257),
    [sym__backtick_quote] = ACTIONS(3257),
    [sym__search_slash_quote] = ACTIONS(3257),
    [sym__open_readline_bracket] = ACTIONS(3257),
    [sym__open_fileglob_bracket] = ACTIONS(3257),
    [sym__PERLY_HEREDOC] = ACTIONS(3257),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3257),
    [sym__filetest] = ACTIONS(3257),
  },
  [STATE(1717)] = {
    [sym_heredoc_content] = STATE(1717),
    [sym__identifier] = ACTIONS(2902),
    [anon_sym_LBRACE] = ACTIONS(2902),
    [anon_sym_DOLLAR] = ACTIONS(2902),
    [anon_sym_AT] = ACTIONS(2902),
    [anon_sym_RPAREN] = ACTIONS(2902),
    [anon_sym_my] = ACTIONS(2902),
    [anon_sym_extended] = ACTIONS(2902),
    [anon_sym_async] = ACTIONS(2902),
    [anon_sym_sub] = ACTIONS(2902),
    [anon_sym_method] = ACTIONS(2902),
    [anon_sym_LPAREN] = ACTIONS(2902),
    [anon_sym_continue] = ACTIONS(2902),
    [anon_sym_state] = ACTIONS(2902),
    [anon_sym_our] = ACTIONS(2902),
    [anon_sym_LBRACK] = ACTIONS(2902),
    [anon_sym_LT_LT] = ACTIONS(2902),
    [anon_sym_SLASH_SLASH] = ACTIONS(2902),
    [anon_sym_PLUS] = ACTIONS(2902),
    [anon_sym_DASH] = ACTIONS(2902),
    [anon_sym_TILDE] = ACTIONS(2902),
    [anon_sym_BANG] = ACTIONS(2902),
    [anon_sym_PLUS_PLUS] = ACTIONS(2902),
    [anon_sym_DASH_DASH] = ACTIONS(2902),
    [anon_sym_BSLASH] = ACTIONS(2900),
    [anon_sym_do] = ACTIONS(2902),
    [anon_sym_eval] = ACTIONS(2902),
    [anon_sym_await] = ACTIONS(2902),
    [anon_sym_field] = ACTIONS(2902),
    [anon_sym_local] = ACTIONS(2902),
    [anon_sym_dynamically] = ACTIONS(2902),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2902),
    [anon_sym_require] = ACTIONS(2902),
    [anon_sym_map] = ACTIONS(2902),
    [anon_sym_grep] = ACTIONS(2902),
    [anon_sym_sort] = ACTIONS(2902),
    [anon_sym_goto] = ACTIONS(2902),
    [anon_sym_return] = ACTIONS(2902),
    [anon_sym_undef] = ACTIONS(2902),
    [anon_sym_PERCENT2] = ACTIONS(2902),
    [anon_sym_AMP2] = ACTIONS(2902),
    [anon_sym_STAR2] = ACTIONS(2902),
    [anon_sym_last] = ACTIONS(2902),
    [anon_sym_next] = ACTIONS(2902),
    [anon_sym_redo] = ACTIONS(2902),
    [anon_sym___FILE__] = ACTIONS(2902),
    [anon_sym___LINE__] = ACTIONS(2902),
    [anon_sym___PACKAGE__] = ACTIONS(2902),
    [anon_sym___SUB__] = ACTIONS(2902),
    [anon_sym_break] = ACTIONS(2902),
    [anon_sym_fork] = ACTIONS(2902),
    [anon_sym_getppid] = ACTIONS(2902),
    [anon_sym_time] = ACTIONS(2902),
    [anon_sym_times] = ACTIONS(2902),
    [anon_sym_wait] = ACTIONS(2902),
    [anon_sym_wantarray] = ACTIONS(2902),
    [anon_sym_abs] = ACTIONS(2902),
    [anon_sym_alarm] = ACTIONS(2902),
    [anon_sym_chop] = ACTIONS(2902),
    [anon_sym_chdir] = ACTIONS(2902),
    [anon_sym_close] = ACTIONS(2902),
    [anon_sym_closedir] = ACTIONS(2902),
    [anon_sym_caller] = ACTIONS(2902),
    [anon_sym_chomp] = ACTIONS(2902),
    [anon_sym_chr] = ACTIONS(2902),
    [anon_sym_cos] = ACTIONS(2902),
    [anon_sym_chroot] = ACTIONS(2902),
    [anon_sym_defined] = ACTIONS(2902),
    [anon_sym_delete] = ACTIONS(2902),
    [anon_sym_dbmclose] = ACTIONS(2902),
    [anon_sym_exists] = ACTIONS(2902),
    [anon_sym_exit] = ACTIONS(2902),
    [anon_sym_eof] = ACTIONS(2902),
    [anon_sym_exp] = ACTIONS(2902),
    [anon_sym_each] = ACTIONS(2902),
    [anon_sym_fc] = ACTIONS(2902),
    [anon_sym_fileno] = ACTIONS(2902),
    [anon_sym_gmtime] = ACTIONS(2902),
    [anon_sym_getc] = ACTIONS(2902),
    [anon_sym_getpgrp] = ACTIONS(2902),
    [anon_sym_getprotobyname] = ACTIONS(2902),
    [anon_sym_getpwname] = ACTIONS(2902),
    [anon_sym_getpwuid] = ACTIONS(2902),
    [anon_sym_getpeername] = ACTIONS(2902),
    [anon_sym_getnetbyname] = ACTIONS(2902),
    [anon_sym_getsockname] = ACTIONS(2902),
    [anon_sym_getgrnam] = ACTIONS(2902),
    [anon_sym_getgrgid] = ACTIONS(2902),
    [anon_sym_hex] = ACTIONS(2902),
    [anon_sym_int] = ACTIONS(2902),
    [anon_sym_keys] = ACTIONS(2902),
    [anon_sym_lc] = ACTIONS(2902),
    [anon_sym_lcfirst] = ACTIONS(2902),
    [anon_sym_length] = ACTIONS(2902),
    [anon_sym_localtime] = ACTIONS(2902),
    [anon_sym_log] = ACTIONS(2902),
    [anon_sym_lock] = ACTIONS(2902),
    [anon_sym_lstat] = ACTIONS(2902),
    [anon_sym_oct] = ACTIONS(2902),
    [anon_sym_ord] = ACTIONS(2902),
    [anon_sym_prototype] = ACTIONS(2902),
    [anon_sym_pop] = ACTIONS(2902),
    [anon_sym_pos] = ACTIONS(2902),
    [anon_sym_quotemeta] = ACTIONS(2902),
    [anon_sym_reset] = ACTIONS(2902),
    [anon_sym_rand] = ACTIONS(2902),
    [anon_sym_rmdir] = ACTIONS(2902),
    [anon_sym_readdir] = ACTIONS(2902),
    [anon_sym_readline] = ACTIONS(2902),
    [anon_sym_readpipe] = ACTIONS(2902),
    [anon_sym_rewinddir] = ACTIONS(2902),
    [anon_sym_readlink] = ACTIONS(2902),
    [anon_sym_ref] = ACTIONS(2902),
    [anon_sym_scalar] = ACTIONS(2902),
    [anon_sym_shift] = ACTIONS(2902),
    [anon_sym_sin] = ACTIONS(2902),
    [anon_sym_sleep] = ACTIONS(2902),
    [anon_sym_sqrt] = ACTIONS(2902),
    [anon_sym_srand] = ACTIONS(2902),
    [anon_sym_stat] = ACTIONS(2902),
    [anon_sym_study] = ACTIONS(2902),
    [anon_sym_tell] = ACTIONS(2902),
    [anon_sym_telldir] = ACTIONS(2902),
    [anon_sym_tied] = ACTIONS(2902),
    [anon_sym_uc] = ACTIONS(2902),
    [anon_sym_ucfirst] = ACTIONS(2902),
    [anon_sym_untie] = ACTIONS(2902),
    [anon_sym_umask] = ACTIONS(2902),
    [anon_sym_values] = ACTIONS(2902),
    [anon_sym_write] = ACTIONS(2902),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2902),
    [anon_sym_qq] = ACTIONS(2902),
    [anon_sym_qw] = ACTIONS(2902),
    [anon_sym_qx] = ACTIONS(2902),
    [anon_sym_qr] = ACTIONS(2902),
    [anon_sym_m] = ACTIONS(2902),
    [anon_sym_s] = ACTIONS(2902),
    [anon_sym_tr] = ACTIONS(2902),
    [anon_sym_y] = ACTIONS(2902),
    [aux_sym__bareword_token1] = ACTIONS(2902),
    [sym_number] = ACTIONS(2902),
    [anon_sym_true] = ACTIONS(2902),
    [anon_sym_false] = ACTIONS(2902),
    [sym__single_quote] = ACTIONS(2900),
    [sym__double_quote] = ACTIONS(2900),
    [sym__backtick_quote] = ACTIONS(2900),
    [sym__search_slash_quote] = ACTIONS(2900),
    [sym__open_readline_bracket] = ACTIONS(2900),
    [sym__open_fileglob_bracket] = ACTIONS(2900),
    [sym__PERLY_HEREDOC] = ACTIONS(2900),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2900),
    [sym__filetest] = ACTIONS(2900),
  },
  [STATE(1718)] = {
    [sym_heredoc_content] = STATE(1718),
    [sym__identifier] = ACTIONS(3221),
    [anon_sym_LBRACE] = ACTIONS(3221),
    [anon_sym_DOLLAR] = ACTIONS(3221),
    [anon_sym_AT] = ACTIONS(3221),
    [anon_sym_RPAREN] = ACTIONS(3221),
    [anon_sym_my] = ACTIONS(3221),
    [anon_sym_extended] = ACTIONS(3221),
    [anon_sym_async] = ACTIONS(3221),
    [anon_sym_sub] = ACTIONS(3221),
    [anon_sym_method] = ACTIONS(3221),
    [anon_sym_LPAREN] = ACTIONS(3221),
    [anon_sym_continue] = ACTIONS(3221),
    [anon_sym_state] = ACTIONS(3221),
    [anon_sym_our] = ACTIONS(3221),
    [anon_sym_LBRACK] = ACTIONS(3221),
    [anon_sym_LT_LT] = ACTIONS(3221),
    [anon_sym_SLASH_SLASH] = ACTIONS(3221),
    [anon_sym_PLUS] = ACTIONS(3221),
    [anon_sym_DASH] = ACTIONS(3221),
    [anon_sym_TILDE] = ACTIONS(3221),
    [anon_sym_BANG] = ACTIONS(3221),
    [anon_sym_PLUS_PLUS] = ACTIONS(3221),
    [anon_sym_DASH_DASH] = ACTIONS(3221),
    [anon_sym_BSLASH] = ACTIONS(3219),
    [anon_sym_do] = ACTIONS(3221),
    [anon_sym_eval] = ACTIONS(3221),
    [anon_sym_await] = ACTIONS(3221),
    [anon_sym_field] = ACTIONS(3221),
    [anon_sym_local] = ACTIONS(3221),
    [anon_sym_dynamically] = ACTIONS(3221),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3221),
    [anon_sym_require] = ACTIONS(3221),
    [anon_sym_map] = ACTIONS(3221),
    [anon_sym_grep] = ACTIONS(3221),
    [anon_sym_sort] = ACTIONS(3221),
    [anon_sym_goto] = ACTIONS(3221),
    [anon_sym_return] = ACTIONS(3221),
    [anon_sym_undef] = ACTIONS(3221),
    [anon_sym_PERCENT2] = ACTIONS(3221),
    [anon_sym_AMP2] = ACTIONS(3221),
    [anon_sym_STAR2] = ACTIONS(3221),
    [anon_sym_last] = ACTIONS(3221),
    [anon_sym_next] = ACTIONS(3221),
    [anon_sym_redo] = ACTIONS(3221),
    [anon_sym___FILE__] = ACTIONS(3221),
    [anon_sym___LINE__] = ACTIONS(3221),
    [anon_sym___PACKAGE__] = ACTIONS(3221),
    [anon_sym___SUB__] = ACTIONS(3221),
    [anon_sym_break] = ACTIONS(3221),
    [anon_sym_fork] = ACTIONS(3221),
    [anon_sym_getppid] = ACTIONS(3221),
    [anon_sym_time] = ACTIONS(3221),
    [anon_sym_times] = ACTIONS(3221),
    [anon_sym_wait] = ACTIONS(3221),
    [anon_sym_wantarray] = ACTIONS(3221),
    [anon_sym_abs] = ACTIONS(3221),
    [anon_sym_alarm] = ACTIONS(3221),
    [anon_sym_chop] = ACTIONS(3221),
    [anon_sym_chdir] = ACTIONS(3221),
    [anon_sym_close] = ACTIONS(3221),
    [anon_sym_closedir] = ACTIONS(3221),
    [anon_sym_caller] = ACTIONS(3221),
    [anon_sym_chomp] = ACTIONS(3221),
    [anon_sym_chr] = ACTIONS(3221),
    [anon_sym_cos] = ACTIONS(3221),
    [anon_sym_chroot] = ACTIONS(3221),
    [anon_sym_defined] = ACTIONS(3221),
    [anon_sym_delete] = ACTIONS(3221),
    [anon_sym_dbmclose] = ACTIONS(3221),
    [anon_sym_exists] = ACTIONS(3221),
    [anon_sym_exit] = ACTIONS(3221),
    [anon_sym_eof] = ACTIONS(3221),
    [anon_sym_exp] = ACTIONS(3221),
    [anon_sym_each] = ACTIONS(3221),
    [anon_sym_fc] = ACTIONS(3221),
    [anon_sym_fileno] = ACTIONS(3221),
    [anon_sym_gmtime] = ACTIONS(3221),
    [anon_sym_getc] = ACTIONS(3221),
    [anon_sym_getpgrp] = ACTIONS(3221),
    [anon_sym_getprotobyname] = ACTIONS(3221),
    [anon_sym_getpwname] = ACTIONS(3221),
    [anon_sym_getpwuid] = ACTIONS(3221),
    [anon_sym_getpeername] = ACTIONS(3221),
    [anon_sym_getnetbyname] = ACTIONS(3221),
    [anon_sym_getsockname] = ACTIONS(3221),
    [anon_sym_getgrnam] = ACTIONS(3221),
    [anon_sym_getgrgid] = ACTIONS(3221),
    [anon_sym_hex] = ACTIONS(3221),
    [anon_sym_int] = ACTIONS(3221),
    [anon_sym_keys] = ACTIONS(3221),
    [anon_sym_lc] = ACTIONS(3221),
    [anon_sym_lcfirst] = ACTIONS(3221),
    [anon_sym_length] = ACTIONS(3221),
    [anon_sym_localtime] = ACTIONS(3221),
    [anon_sym_log] = ACTIONS(3221),
    [anon_sym_lock] = ACTIONS(3221),
    [anon_sym_lstat] = ACTIONS(3221),
    [anon_sym_oct] = ACTIONS(3221),
    [anon_sym_ord] = ACTIONS(3221),
    [anon_sym_prototype] = ACTIONS(3221),
    [anon_sym_pop] = ACTIONS(3221),
    [anon_sym_pos] = ACTIONS(3221),
    [anon_sym_quotemeta] = ACTIONS(3221),
    [anon_sym_reset] = ACTIONS(3221),
    [anon_sym_rand] = ACTIONS(3221),
    [anon_sym_rmdir] = ACTIONS(3221),
    [anon_sym_readdir] = ACTIONS(3221),
    [anon_sym_readline] = ACTIONS(3221),
    [anon_sym_readpipe] = ACTIONS(3221),
    [anon_sym_rewinddir] = ACTIONS(3221),
    [anon_sym_readlink] = ACTIONS(3221),
    [anon_sym_ref] = ACTIONS(3221),
    [anon_sym_scalar] = ACTIONS(3221),
    [anon_sym_shift] = ACTIONS(3221),
    [anon_sym_sin] = ACTIONS(3221),
    [anon_sym_sleep] = ACTIONS(3221),
    [anon_sym_sqrt] = ACTIONS(3221),
    [anon_sym_srand] = ACTIONS(3221),
    [anon_sym_stat] = ACTIONS(3221),
    [anon_sym_study] = ACTIONS(3221),
    [anon_sym_tell] = ACTIONS(3221),
    [anon_sym_telldir] = ACTIONS(3221),
    [anon_sym_tied] = ACTIONS(3221),
    [anon_sym_uc] = ACTIONS(3221),
    [anon_sym_ucfirst] = ACTIONS(3221),
    [anon_sym_untie] = ACTIONS(3221),
    [anon_sym_umask] = ACTIONS(3221),
    [anon_sym_values] = ACTIONS(3221),
    [anon_sym_write] = ACTIONS(3221),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3221),
    [anon_sym_qq] = ACTIONS(3221),
    [anon_sym_qw] = ACTIONS(3221),
    [anon_sym_qx] = ACTIONS(3221),
    [anon_sym_qr] = ACTIONS(3221),
    [anon_sym_m] = ACTIONS(3221),
    [anon_sym_s] = ACTIONS(3221),
    [anon_sym_tr] = ACTIONS(3221),
    [anon_sym_y] = ACTIONS(3221),
    [aux_sym__bareword_token1] = ACTIONS(3221),
    [sym_number] = ACTIONS(3221),
    [anon_sym_true] = ACTIONS(3221),
    [anon_sym_false] = ACTIONS(3221),
    [sym__single_quote] = ACTIONS(3219),
    [sym__double_quote] = ACTIONS(3219),
    [sym__backtick_quote] = ACTIONS(3219),
    [sym__search_slash_quote] = ACTIONS(3219),
    [sym__open_readline_bracket] = ACTIONS(3219),
    [sym__open_fileglob_bracket] = ACTIONS(3219),
    [sym__PERLY_HEREDOC] = ACTIONS(3219),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3219),
    [sym__filetest] = ACTIONS(3219),
  },
  [STATE(1719)] = {
    [sym_heredoc_content] = STATE(1719),
    [sym__identifier] = ACTIONS(2532),
    [anon_sym_LBRACE] = ACTIONS(2532),
    [anon_sym_DOLLAR] = ACTIONS(2532),
    [anon_sym_AT] = ACTIONS(2532),
    [anon_sym_RPAREN] = ACTIONS(2532),
    [anon_sym_my] = ACTIONS(2532),
    [anon_sym_extended] = ACTIONS(2532),
    [anon_sym_async] = ACTIONS(2532),
    [anon_sym_sub] = ACTIONS(2532),
    [anon_sym_method] = ACTIONS(2532),
    [anon_sym_LPAREN] = ACTIONS(2532),
    [anon_sym_continue] = ACTIONS(2532),
    [anon_sym_state] = ACTIONS(2532),
    [anon_sym_our] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_TILDE] = ACTIONS(2532),
    [anon_sym_BANG] = ACTIONS(2532),
    [anon_sym_PLUS_PLUS] = ACTIONS(2532),
    [anon_sym_DASH_DASH] = ACTIONS(2532),
    [anon_sym_BSLASH] = ACTIONS(2530),
    [anon_sym_do] = ACTIONS(2532),
    [anon_sym_eval] = ACTIONS(2532),
    [anon_sym_await] = ACTIONS(2532),
    [anon_sym_field] = ACTIONS(2532),
    [anon_sym_local] = ACTIONS(2532),
    [anon_sym_dynamically] = ACTIONS(2532),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2532),
    [anon_sym_require] = ACTIONS(2532),
    [anon_sym_map] = ACTIONS(2532),
    [anon_sym_grep] = ACTIONS(2532),
    [anon_sym_sort] = ACTIONS(2532),
    [anon_sym_goto] = ACTIONS(2532),
    [anon_sym_return] = ACTIONS(2532),
    [anon_sym_undef] = ACTIONS(2532),
    [anon_sym_PERCENT2] = ACTIONS(2532),
    [anon_sym_AMP2] = ACTIONS(2532),
    [anon_sym_STAR2] = ACTIONS(2532),
    [anon_sym_last] = ACTIONS(2532),
    [anon_sym_next] = ACTIONS(2532),
    [anon_sym_redo] = ACTIONS(2532),
    [anon_sym___FILE__] = ACTIONS(2532),
    [anon_sym___LINE__] = ACTIONS(2532),
    [anon_sym___PACKAGE__] = ACTIONS(2532),
    [anon_sym___SUB__] = ACTIONS(2532),
    [anon_sym_break] = ACTIONS(2532),
    [anon_sym_fork] = ACTIONS(2532),
    [anon_sym_getppid] = ACTIONS(2532),
    [anon_sym_time] = ACTIONS(2532),
    [anon_sym_times] = ACTIONS(2532),
    [anon_sym_wait] = ACTIONS(2532),
    [anon_sym_wantarray] = ACTIONS(2532),
    [anon_sym_abs] = ACTIONS(2532),
    [anon_sym_alarm] = ACTIONS(2532),
    [anon_sym_chop] = ACTIONS(2532),
    [anon_sym_chdir] = ACTIONS(2532),
    [anon_sym_close] = ACTIONS(2532),
    [anon_sym_closedir] = ACTIONS(2532),
    [anon_sym_caller] = ACTIONS(2532),
    [anon_sym_chomp] = ACTIONS(2532),
    [anon_sym_chr] = ACTIONS(2532),
    [anon_sym_cos] = ACTIONS(2532),
    [anon_sym_chroot] = ACTIONS(2532),
    [anon_sym_defined] = ACTIONS(2532),
    [anon_sym_delete] = ACTIONS(2532),
    [anon_sym_dbmclose] = ACTIONS(2532),
    [anon_sym_exists] = ACTIONS(2532),
    [anon_sym_exit] = ACTIONS(2532),
    [anon_sym_eof] = ACTIONS(2532),
    [anon_sym_exp] = ACTIONS(2532),
    [anon_sym_each] = ACTIONS(2532),
    [anon_sym_fc] = ACTIONS(2532),
    [anon_sym_fileno] = ACTIONS(2532),
    [anon_sym_gmtime] = ACTIONS(2532),
    [anon_sym_getc] = ACTIONS(2532),
    [anon_sym_getpgrp] = ACTIONS(2532),
    [anon_sym_getprotobyname] = ACTIONS(2532),
    [anon_sym_getpwname] = ACTIONS(2532),
    [anon_sym_getpwuid] = ACTIONS(2532),
    [anon_sym_getpeername] = ACTIONS(2532),
    [anon_sym_getnetbyname] = ACTIONS(2532),
    [anon_sym_getsockname] = ACTIONS(2532),
    [anon_sym_getgrnam] = ACTIONS(2532),
    [anon_sym_getgrgid] = ACTIONS(2532),
    [anon_sym_hex] = ACTIONS(2532),
    [anon_sym_int] = ACTIONS(2532),
    [anon_sym_keys] = ACTIONS(2532),
    [anon_sym_lc] = ACTIONS(2532),
    [anon_sym_lcfirst] = ACTIONS(2532),
    [anon_sym_length] = ACTIONS(2532),
    [anon_sym_localtime] = ACTIONS(2532),
    [anon_sym_log] = ACTIONS(2532),
    [anon_sym_lock] = ACTIONS(2532),
    [anon_sym_lstat] = ACTIONS(2532),
    [anon_sym_oct] = ACTIONS(2532),
    [anon_sym_ord] = ACTIONS(2532),
    [anon_sym_prototype] = ACTIONS(2532),
    [anon_sym_pop] = ACTIONS(2532),
    [anon_sym_pos] = ACTIONS(2532),
    [anon_sym_quotemeta] = ACTIONS(2532),
    [anon_sym_reset] = ACTIONS(2532),
    [anon_sym_rand] = ACTIONS(2532),
    [anon_sym_rmdir] = ACTIONS(2532),
    [anon_sym_readdir] = ACTIONS(2532),
    [anon_sym_readline] = ACTIONS(2532),
    [anon_sym_readpipe] = ACTIONS(2532),
    [anon_sym_rewinddir] = ACTIONS(2532),
    [anon_sym_readlink] = ACTIONS(2532),
    [anon_sym_ref] = ACTIONS(2532),
    [anon_sym_scalar] = ACTIONS(2532),
    [anon_sym_shift] = ACTIONS(2532),
    [anon_sym_sin] = ACTIONS(2532),
    [anon_sym_sleep] = ACTIONS(2532),
    [anon_sym_sqrt] = ACTIONS(2532),
    [anon_sym_srand] = ACTIONS(2532),
    [anon_sym_stat] = ACTIONS(2532),
    [anon_sym_study] = ACTIONS(2532),
    [anon_sym_tell] = ACTIONS(2532),
    [anon_sym_telldir] = ACTIONS(2532),
    [anon_sym_tied] = ACTIONS(2532),
    [anon_sym_uc] = ACTIONS(2532),
    [anon_sym_ucfirst] = ACTIONS(2532),
    [anon_sym_untie] = ACTIONS(2532),
    [anon_sym_umask] = ACTIONS(2532),
    [anon_sym_values] = ACTIONS(2532),
    [anon_sym_write] = ACTIONS(2532),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2532),
    [anon_sym_qq] = ACTIONS(2532),
    [anon_sym_qw] = ACTIONS(2532),
    [anon_sym_qx] = ACTIONS(2532),
    [anon_sym_qr] = ACTIONS(2532),
    [anon_sym_m] = ACTIONS(2532),
    [anon_sym_s] = ACTIONS(2532),
    [anon_sym_tr] = ACTIONS(2532),
    [anon_sym_y] = ACTIONS(2532),
    [aux_sym__bareword_token1] = ACTIONS(2532),
    [sym_number] = ACTIONS(2532),
    [anon_sym_true] = ACTIONS(2532),
    [anon_sym_false] = ACTIONS(2532),
    [sym__single_quote] = ACTIONS(2530),
    [sym__double_quote] = ACTIONS(2530),
    [sym__backtick_quote] = ACTIONS(2530),
    [sym__search_slash_quote] = ACTIONS(2530),
    [sym__open_readline_bracket] = ACTIONS(2530),
    [sym__open_fileglob_bracket] = ACTIONS(2530),
    [sym__PERLY_HEREDOC] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2530),
    [sym__filetest] = ACTIONS(2530),
  },
  [STATE(1720)] = {
    [sym_heredoc_content] = STATE(1720),
    [sym__identifier] = ACTIONS(3225),
    [anon_sym_LBRACE] = ACTIONS(3225),
    [anon_sym_DOLLAR] = ACTIONS(3225),
    [anon_sym_AT] = ACTIONS(3225),
    [anon_sym_RPAREN] = ACTIONS(3225),
    [anon_sym_my] = ACTIONS(3225),
    [anon_sym_extended] = ACTIONS(3225),
    [anon_sym_async] = ACTIONS(3225),
    [anon_sym_sub] = ACTIONS(3225),
    [anon_sym_method] = ACTIONS(3225),
    [anon_sym_LPAREN] = ACTIONS(3225),
    [anon_sym_continue] = ACTIONS(3225),
    [anon_sym_state] = ACTIONS(3225),
    [anon_sym_our] = ACTIONS(3225),
    [anon_sym_LBRACK] = ACTIONS(3225),
    [anon_sym_LT_LT] = ACTIONS(3225),
    [anon_sym_SLASH_SLASH] = ACTIONS(3225),
    [anon_sym_PLUS] = ACTIONS(3225),
    [anon_sym_DASH] = ACTIONS(3225),
    [anon_sym_TILDE] = ACTIONS(3225),
    [anon_sym_BANG] = ACTIONS(3225),
    [anon_sym_PLUS_PLUS] = ACTIONS(3225),
    [anon_sym_DASH_DASH] = ACTIONS(3225),
    [anon_sym_BSLASH] = ACTIONS(3223),
    [anon_sym_do] = ACTIONS(3225),
    [anon_sym_eval] = ACTIONS(3225),
    [anon_sym_await] = ACTIONS(3225),
    [anon_sym_field] = ACTIONS(3225),
    [anon_sym_local] = ACTIONS(3225),
    [anon_sym_dynamically] = ACTIONS(3225),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3225),
    [anon_sym_require] = ACTIONS(3225),
    [anon_sym_map] = ACTIONS(3225),
    [anon_sym_grep] = ACTIONS(3225),
    [anon_sym_sort] = ACTIONS(3225),
    [anon_sym_goto] = ACTIONS(3225),
    [anon_sym_return] = ACTIONS(3225),
    [anon_sym_undef] = ACTIONS(3225),
    [anon_sym_PERCENT2] = ACTIONS(3225),
    [anon_sym_AMP2] = ACTIONS(3225),
    [anon_sym_STAR2] = ACTIONS(3225),
    [anon_sym_last] = ACTIONS(3225),
    [anon_sym_next] = ACTIONS(3225),
    [anon_sym_redo] = ACTIONS(3225),
    [anon_sym___FILE__] = ACTIONS(3225),
    [anon_sym___LINE__] = ACTIONS(3225),
    [anon_sym___PACKAGE__] = ACTIONS(3225),
    [anon_sym___SUB__] = ACTIONS(3225),
    [anon_sym_break] = ACTIONS(3225),
    [anon_sym_fork] = ACTIONS(3225),
    [anon_sym_getppid] = ACTIONS(3225),
    [anon_sym_time] = ACTIONS(3225),
    [anon_sym_times] = ACTIONS(3225),
    [anon_sym_wait] = ACTIONS(3225),
    [anon_sym_wantarray] = ACTIONS(3225),
    [anon_sym_abs] = ACTIONS(3225),
    [anon_sym_alarm] = ACTIONS(3225),
    [anon_sym_chop] = ACTIONS(3225),
    [anon_sym_chdir] = ACTIONS(3225),
    [anon_sym_close] = ACTIONS(3225),
    [anon_sym_closedir] = ACTIONS(3225),
    [anon_sym_caller] = ACTIONS(3225),
    [anon_sym_chomp] = ACTIONS(3225),
    [anon_sym_chr] = ACTIONS(3225),
    [anon_sym_cos] = ACTIONS(3225),
    [anon_sym_chroot] = ACTIONS(3225),
    [anon_sym_defined] = ACTIONS(3225),
    [anon_sym_delete] = ACTIONS(3225),
    [anon_sym_dbmclose] = ACTIONS(3225),
    [anon_sym_exists] = ACTIONS(3225),
    [anon_sym_exit] = ACTIONS(3225),
    [anon_sym_eof] = ACTIONS(3225),
    [anon_sym_exp] = ACTIONS(3225),
    [anon_sym_each] = ACTIONS(3225),
    [anon_sym_fc] = ACTIONS(3225),
    [anon_sym_fileno] = ACTIONS(3225),
    [anon_sym_gmtime] = ACTIONS(3225),
    [anon_sym_getc] = ACTIONS(3225),
    [anon_sym_getpgrp] = ACTIONS(3225),
    [anon_sym_getprotobyname] = ACTIONS(3225),
    [anon_sym_getpwname] = ACTIONS(3225),
    [anon_sym_getpwuid] = ACTIONS(3225),
    [anon_sym_getpeername] = ACTIONS(3225),
    [anon_sym_getnetbyname] = ACTIONS(3225),
    [anon_sym_getsockname] = ACTIONS(3225),
    [anon_sym_getgrnam] = ACTIONS(3225),
    [anon_sym_getgrgid] = ACTIONS(3225),
    [anon_sym_hex] = ACTIONS(3225),
    [anon_sym_int] = ACTIONS(3225),
    [anon_sym_keys] = ACTIONS(3225),
    [anon_sym_lc] = ACTIONS(3225),
    [anon_sym_lcfirst] = ACTIONS(3225),
    [anon_sym_length] = ACTIONS(3225),
    [anon_sym_localtime] = ACTIONS(3225),
    [anon_sym_log] = ACTIONS(3225),
    [anon_sym_lock] = ACTIONS(3225),
    [anon_sym_lstat] = ACTIONS(3225),
    [anon_sym_oct] = ACTIONS(3225),
    [anon_sym_ord] = ACTIONS(3225),
    [anon_sym_prototype] = ACTIONS(3225),
    [anon_sym_pop] = ACTIONS(3225),
    [anon_sym_pos] = ACTIONS(3225),
    [anon_sym_quotemeta] = ACTIONS(3225),
    [anon_sym_reset] = ACTIONS(3225),
    [anon_sym_rand] = ACTIONS(3225),
    [anon_sym_rmdir] = ACTIONS(3225),
    [anon_sym_readdir] = ACTIONS(3225),
    [anon_sym_readline] = ACTIONS(3225),
    [anon_sym_readpipe] = ACTIONS(3225),
    [anon_sym_rewinddir] = ACTIONS(3225),
    [anon_sym_readlink] = ACTIONS(3225),
    [anon_sym_ref] = ACTIONS(3225),
    [anon_sym_scalar] = ACTIONS(3225),
    [anon_sym_shift] = ACTIONS(3225),
    [anon_sym_sin] = ACTIONS(3225),
    [anon_sym_sleep] = ACTIONS(3225),
    [anon_sym_sqrt] = ACTIONS(3225),
    [anon_sym_srand] = ACTIONS(3225),
    [anon_sym_stat] = ACTIONS(3225),
    [anon_sym_study] = ACTIONS(3225),
    [anon_sym_tell] = ACTIONS(3225),
    [anon_sym_telldir] = ACTIONS(3225),
    [anon_sym_tied] = ACTIONS(3225),
    [anon_sym_uc] = ACTIONS(3225),
    [anon_sym_ucfirst] = ACTIONS(3225),
    [anon_sym_untie] = ACTIONS(3225),
    [anon_sym_umask] = ACTIONS(3225),
    [anon_sym_values] = ACTIONS(3225),
    [anon_sym_write] = ACTIONS(3225),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3225),
    [anon_sym_qq] = ACTIONS(3225),
    [anon_sym_qw] = ACTIONS(3225),
    [anon_sym_qx] = ACTIONS(3225),
    [anon_sym_qr] = ACTIONS(3225),
    [anon_sym_m] = ACTIONS(3225),
    [anon_sym_s] = ACTIONS(3225),
    [anon_sym_tr] = ACTIONS(3225),
    [anon_sym_y] = ACTIONS(3225),
    [aux_sym__bareword_token1] = ACTIONS(3225),
    [sym_number] = ACTIONS(3225),
    [anon_sym_true] = ACTIONS(3225),
    [anon_sym_false] = ACTIONS(3225),
    [sym__single_quote] = ACTIONS(3223),
    [sym__double_quote] = ACTIONS(3223),
    [sym__backtick_quote] = ACTIONS(3223),
    [sym__search_slash_quote] = ACTIONS(3223),
    [sym__open_readline_bracket] = ACTIONS(3223),
    [sym__open_fileglob_bracket] = ACTIONS(3223),
    [sym__PERLY_HEREDOC] = ACTIONS(3223),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3223),
    [sym__filetest] = ACTIONS(3223),
  },
  [STATE(1721)] = {
    [sym_heredoc_content] = STATE(1721),
    [sym__identifier] = ACTIONS(3081),
    [anon_sym_LBRACE] = ACTIONS(3081),
    [anon_sym_DOLLAR] = ACTIONS(3081),
    [anon_sym_AT] = ACTIONS(3081),
    [anon_sym_RPAREN] = ACTIONS(3081),
    [anon_sym_my] = ACTIONS(3081),
    [anon_sym_extended] = ACTIONS(3081),
    [anon_sym_async] = ACTIONS(3081),
    [anon_sym_sub] = ACTIONS(3081),
    [anon_sym_method] = ACTIONS(3081),
    [anon_sym_LPAREN] = ACTIONS(3081),
    [anon_sym_continue] = ACTIONS(3081),
    [anon_sym_state] = ACTIONS(3081),
    [anon_sym_our] = ACTIONS(3081),
    [anon_sym_LBRACK] = ACTIONS(3081),
    [anon_sym_LT_LT] = ACTIONS(3081),
    [anon_sym_SLASH_SLASH] = ACTIONS(3081),
    [anon_sym_PLUS] = ACTIONS(3081),
    [anon_sym_DASH] = ACTIONS(3081),
    [anon_sym_TILDE] = ACTIONS(3081),
    [anon_sym_BANG] = ACTIONS(3081),
    [anon_sym_PLUS_PLUS] = ACTIONS(3081),
    [anon_sym_DASH_DASH] = ACTIONS(3081),
    [anon_sym_BSLASH] = ACTIONS(3079),
    [anon_sym_do] = ACTIONS(3081),
    [anon_sym_eval] = ACTIONS(3081),
    [anon_sym_await] = ACTIONS(3081),
    [anon_sym_field] = ACTIONS(3081),
    [anon_sym_local] = ACTIONS(3081),
    [anon_sym_dynamically] = ACTIONS(3081),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3081),
    [anon_sym_require] = ACTIONS(3081),
    [anon_sym_map] = ACTIONS(3081),
    [anon_sym_grep] = ACTIONS(3081),
    [anon_sym_sort] = ACTIONS(3081),
    [anon_sym_goto] = ACTIONS(3081),
    [anon_sym_return] = ACTIONS(3081),
    [anon_sym_undef] = ACTIONS(3081),
    [anon_sym_PERCENT2] = ACTIONS(3081),
    [anon_sym_AMP2] = ACTIONS(3081),
    [anon_sym_STAR2] = ACTIONS(3081),
    [anon_sym_last] = ACTIONS(3081),
    [anon_sym_next] = ACTIONS(3081),
    [anon_sym_redo] = ACTIONS(3081),
    [anon_sym___FILE__] = ACTIONS(3081),
    [anon_sym___LINE__] = ACTIONS(3081),
    [anon_sym___PACKAGE__] = ACTIONS(3081),
    [anon_sym___SUB__] = ACTIONS(3081),
    [anon_sym_break] = ACTIONS(3081),
    [anon_sym_fork] = ACTIONS(3081),
    [anon_sym_getppid] = ACTIONS(3081),
    [anon_sym_time] = ACTIONS(3081),
    [anon_sym_times] = ACTIONS(3081),
    [anon_sym_wait] = ACTIONS(3081),
    [anon_sym_wantarray] = ACTIONS(3081),
    [anon_sym_abs] = ACTIONS(3081),
    [anon_sym_alarm] = ACTIONS(3081),
    [anon_sym_chop] = ACTIONS(3081),
    [anon_sym_chdir] = ACTIONS(3081),
    [anon_sym_close] = ACTIONS(3081),
    [anon_sym_closedir] = ACTIONS(3081),
    [anon_sym_caller] = ACTIONS(3081),
    [anon_sym_chomp] = ACTIONS(3081),
    [anon_sym_chr] = ACTIONS(3081),
    [anon_sym_cos] = ACTIONS(3081),
    [anon_sym_chroot] = ACTIONS(3081),
    [anon_sym_defined] = ACTIONS(3081),
    [anon_sym_delete] = ACTIONS(3081),
    [anon_sym_dbmclose] = ACTIONS(3081),
    [anon_sym_exists] = ACTIONS(3081),
    [anon_sym_exit] = ACTIONS(3081),
    [anon_sym_eof] = ACTIONS(3081),
    [anon_sym_exp] = ACTIONS(3081),
    [anon_sym_each] = ACTIONS(3081),
    [anon_sym_fc] = ACTIONS(3081),
    [anon_sym_fileno] = ACTIONS(3081),
    [anon_sym_gmtime] = ACTIONS(3081),
    [anon_sym_getc] = ACTIONS(3081),
    [anon_sym_getpgrp] = ACTIONS(3081),
    [anon_sym_getprotobyname] = ACTIONS(3081),
    [anon_sym_getpwname] = ACTIONS(3081),
    [anon_sym_getpwuid] = ACTIONS(3081),
    [anon_sym_getpeername] = ACTIONS(3081),
    [anon_sym_getnetbyname] = ACTIONS(3081),
    [anon_sym_getsockname] = ACTIONS(3081),
    [anon_sym_getgrnam] = ACTIONS(3081),
    [anon_sym_getgrgid] = ACTIONS(3081),
    [anon_sym_hex] = ACTIONS(3081),
    [anon_sym_int] = ACTIONS(3081),
    [anon_sym_keys] = ACTIONS(3081),
    [anon_sym_lc] = ACTIONS(3081),
    [anon_sym_lcfirst] = ACTIONS(3081),
    [anon_sym_length] = ACTIONS(3081),
    [anon_sym_localtime] = ACTIONS(3081),
    [anon_sym_log] = ACTIONS(3081),
    [anon_sym_lock] = ACTIONS(3081),
    [anon_sym_lstat] = ACTIONS(3081),
    [anon_sym_oct] = ACTIONS(3081),
    [anon_sym_ord] = ACTIONS(3081),
    [anon_sym_prototype] = ACTIONS(3081),
    [anon_sym_pop] = ACTIONS(3081),
    [anon_sym_pos] = ACTIONS(3081),
    [anon_sym_quotemeta] = ACTIONS(3081),
    [anon_sym_reset] = ACTIONS(3081),
    [anon_sym_rand] = ACTIONS(3081),
    [anon_sym_rmdir] = ACTIONS(3081),
    [anon_sym_readdir] = ACTIONS(3081),
    [anon_sym_readline] = ACTIONS(3081),
    [anon_sym_readpipe] = ACTIONS(3081),
    [anon_sym_rewinddir] = ACTIONS(3081),
    [anon_sym_readlink] = ACTIONS(3081),
    [anon_sym_ref] = ACTIONS(3081),
    [anon_sym_scalar] = ACTIONS(3081),
    [anon_sym_shift] = ACTIONS(3081),
    [anon_sym_sin] = ACTIONS(3081),
    [anon_sym_sleep] = ACTIONS(3081),
    [anon_sym_sqrt] = ACTIONS(3081),
    [anon_sym_srand] = ACTIONS(3081),
    [anon_sym_stat] = ACTIONS(3081),
    [anon_sym_study] = ACTIONS(3081),
    [anon_sym_tell] = ACTIONS(3081),
    [anon_sym_telldir] = ACTIONS(3081),
    [anon_sym_tied] = ACTIONS(3081),
    [anon_sym_uc] = ACTIONS(3081),
    [anon_sym_ucfirst] = ACTIONS(3081),
    [anon_sym_untie] = ACTIONS(3081),
    [anon_sym_umask] = ACTIONS(3081),
    [anon_sym_values] = ACTIONS(3081),
    [anon_sym_write] = ACTIONS(3081),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3081),
    [anon_sym_qq] = ACTIONS(3081),
    [anon_sym_qw] = ACTIONS(3081),
    [anon_sym_qx] = ACTIONS(3081),
    [anon_sym_qr] = ACTIONS(3081),
    [anon_sym_m] = ACTIONS(3081),
    [anon_sym_s] = ACTIONS(3081),
    [anon_sym_tr] = ACTIONS(3081),
    [anon_sym_y] = ACTIONS(3081),
    [aux_sym__bareword_token1] = ACTIONS(3081),
    [sym_number] = ACTIONS(3081),
    [anon_sym_true] = ACTIONS(3081),
    [anon_sym_false] = ACTIONS(3081),
    [sym__single_quote] = ACTIONS(3079),
    [sym__double_quote] = ACTIONS(3079),
    [sym__backtick_quote] = ACTIONS(3079),
    [sym__search_slash_quote] = ACTIONS(3079),
    [sym__open_readline_bracket] = ACTIONS(3079),
    [sym__open_fileglob_bracket] = ACTIONS(3079),
    [sym__PERLY_HEREDOC] = ACTIONS(3079),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3079),
    [sym__filetest] = ACTIONS(3079),
  },
  [STATE(1722)] = {
    [sym_heredoc_content] = STATE(1722),
    [sym__identifier] = ACTIONS(3177),
    [anon_sym_LBRACE] = ACTIONS(3177),
    [anon_sym_DOLLAR] = ACTIONS(3177),
    [anon_sym_AT] = ACTIONS(3177),
    [anon_sym_RPAREN] = ACTIONS(3177),
    [anon_sym_my] = ACTIONS(3177),
    [anon_sym_extended] = ACTIONS(3177),
    [anon_sym_async] = ACTIONS(3177),
    [anon_sym_sub] = ACTIONS(3177),
    [anon_sym_method] = ACTIONS(3177),
    [anon_sym_LPAREN] = ACTIONS(3177),
    [anon_sym_continue] = ACTIONS(3177),
    [anon_sym_state] = ACTIONS(3177),
    [anon_sym_our] = ACTIONS(3177),
    [anon_sym_LBRACK] = ACTIONS(3177),
    [anon_sym_LT_LT] = ACTIONS(3177),
    [anon_sym_SLASH_SLASH] = ACTIONS(3177),
    [anon_sym_PLUS] = ACTIONS(3177),
    [anon_sym_DASH] = ACTIONS(3177),
    [anon_sym_TILDE] = ACTIONS(3177),
    [anon_sym_BANG] = ACTIONS(3177),
    [anon_sym_PLUS_PLUS] = ACTIONS(3177),
    [anon_sym_DASH_DASH] = ACTIONS(3177),
    [anon_sym_BSLASH] = ACTIONS(3175),
    [anon_sym_do] = ACTIONS(3177),
    [anon_sym_eval] = ACTIONS(3177),
    [anon_sym_await] = ACTIONS(3177),
    [anon_sym_field] = ACTIONS(3177),
    [anon_sym_local] = ACTIONS(3177),
    [anon_sym_dynamically] = ACTIONS(3177),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3177),
    [anon_sym_require] = ACTIONS(3177),
    [anon_sym_map] = ACTIONS(3177),
    [anon_sym_grep] = ACTIONS(3177),
    [anon_sym_sort] = ACTIONS(3177),
    [anon_sym_goto] = ACTIONS(3177),
    [anon_sym_return] = ACTIONS(3177),
    [anon_sym_undef] = ACTIONS(3177),
    [anon_sym_PERCENT2] = ACTIONS(3177),
    [anon_sym_AMP2] = ACTIONS(3177),
    [anon_sym_STAR2] = ACTIONS(3177),
    [anon_sym_last] = ACTIONS(3177),
    [anon_sym_next] = ACTIONS(3177),
    [anon_sym_redo] = ACTIONS(3177),
    [anon_sym___FILE__] = ACTIONS(3177),
    [anon_sym___LINE__] = ACTIONS(3177),
    [anon_sym___PACKAGE__] = ACTIONS(3177),
    [anon_sym___SUB__] = ACTIONS(3177),
    [anon_sym_break] = ACTIONS(3177),
    [anon_sym_fork] = ACTIONS(3177),
    [anon_sym_getppid] = ACTIONS(3177),
    [anon_sym_time] = ACTIONS(3177),
    [anon_sym_times] = ACTIONS(3177),
    [anon_sym_wait] = ACTIONS(3177),
    [anon_sym_wantarray] = ACTIONS(3177),
    [anon_sym_abs] = ACTIONS(3177),
    [anon_sym_alarm] = ACTIONS(3177),
    [anon_sym_chop] = ACTIONS(3177),
    [anon_sym_chdir] = ACTIONS(3177),
    [anon_sym_close] = ACTIONS(3177),
    [anon_sym_closedir] = ACTIONS(3177),
    [anon_sym_caller] = ACTIONS(3177),
    [anon_sym_chomp] = ACTIONS(3177),
    [anon_sym_chr] = ACTIONS(3177),
    [anon_sym_cos] = ACTIONS(3177),
    [anon_sym_chroot] = ACTIONS(3177),
    [anon_sym_defined] = ACTIONS(3177),
    [anon_sym_delete] = ACTIONS(3177),
    [anon_sym_dbmclose] = ACTIONS(3177),
    [anon_sym_exists] = ACTIONS(3177),
    [anon_sym_exit] = ACTIONS(3177),
    [anon_sym_eof] = ACTIONS(3177),
    [anon_sym_exp] = ACTIONS(3177),
    [anon_sym_each] = ACTIONS(3177),
    [anon_sym_fc] = ACTIONS(3177),
    [anon_sym_fileno] = ACTIONS(3177),
    [anon_sym_gmtime] = ACTIONS(3177),
    [anon_sym_getc] = ACTIONS(3177),
    [anon_sym_getpgrp] = ACTIONS(3177),
    [anon_sym_getprotobyname] = ACTIONS(3177),
    [anon_sym_getpwname] = ACTIONS(3177),
    [anon_sym_getpwuid] = ACTIONS(3177),
    [anon_sym_getpeername] = ACTIONS(3177),
    [anon_sym_getnetbyname] = ACTIONS(3177),
    [anon_sym_getsockname] = ACTIONS(3177),
    [anon_sym_getgrnam] = ACTIONS(3177),
    [anon_sym_getgrgid] = ACTIONS(3177),
    [anon_sym_hex] = ACTIONS(3177),
    [anon_sym_int] = ACTIONS(3177),
    [anon_sym_keys] = ACTIONS(3177),
    [anon_sym_lc] = ACTIONS(3177),
    [anon_sym_lcfirst] = ACTIONS(3177),
    [anon_sym_length] = ACTIONS(3177),
    [anon_sym_localtime] = ACTIONS(3177),
    [anon_sym_log] = ACTIONS(3177),
    [anon_sym_lock] = ACTIONS(3177),
    [anon_sym_lstat] = ACTIONS(3177),
    [anon_sym_oct] = ACTIONS(3177),
    [anon_sym_ord] = ACTIONS(3177),
    [anon_sym_prototype] = ACTIONS(3177),
    [anon_sym_pop] = ACTIONS(3177),
    [anon_sym_pos] = ACTIONS(3177),
    [anon_sym_quotemeta] = ACTIONS(3177),
    [anon_sym_reset] = ACTIONS(3177),
    [anon_sym_rand] = ACTIONS(3177),
    [anon_sym_rmdir] = ACTIONS(3177),
    [anon_sym_readdir] = ACTIONS(3177),
    [anon_sym_readline] = ACTIONS(3177),
    [anon_sym_readpipe] = ACTIONS(3177),
    [anon_sym_rewinddir] = ACTIONS(3177),
    [anon_sym_readlink] = ACTIONS(3177),
    [anon_sym_ref] = ACTIONS(3177),
    [anon_sym_scalar] = ACTIONS(3177),
    [anon_sym_shift] = ACTIONS(3177),
    [anon_sym_sin] = ACTIONS(3177),
    [anon_sym_sleep] = ACTIONS(3177),
    [anon_sym_sqrt] = ACTIONS(3177),
    [anon_sym_srand] = ACTIONS(3177),
    [anon_sym_stat] = ACTIONS(3177),
    [anon_sym_study] = ACTIONS(3177),
    [anon_sym_tell] = ACTIONS(3177),
    [anon_sym_telldir] = ACTIONS(3177),
    [anon_sym_tied] = ACTIONS(3177),
    [anon_sym_uc] = ACTIONS(3177),
    [anon_sym_ucfirst] = ACTIONS(3177),
    [anon_sym_untie] = ACTIONS(3177),
    [anon_sym_umask] = ACTIONS(3177),
    [anon_sym_values] = ACTIONS(3177),
    [anon_sym_write] = ACTIONS(3177),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3177),
    [anon_sym_qq] = ACTIONS(3177),
    [anon_sym_qw] = ACTIONS(3177),
    [anon_sym_qx] = ACTIONS(3177),
    [anon_sym_qr] = ACTIONS(3177),
    [anon_sym_m] = ACTIONS(3177),
    [anon_sym_s] = ACTIONS(3177),
    [anon_sym_tr] = ACTIONS(3177),
    [anon_sym_y] = ACTIONS(3177),
    [aux_sym__bareword_token1] = ACTIONS(3177),
    [sym_number] = ACTIONS(3177),
    [anon_sym_true] = ACTIONS(3177),
    [anon_sym_false] = ACTIONS(3177),
    [sym__single_quote] = ACTIONS(3175),
    [sym__double_quote] = ACTIONS(3175),
    [sym__backtick_quote] = ACTIONS(3175),
    [sym__search_slash_quote] = ACTIONS(3175),
    [sym__open_readline_bracket] = ACTIONS(3175),
    [sym__open_fileglob_bracket] = ACTIONS(3175),
    [sym__PERLY_HEREDOC] = ACTIONS(3175),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3175),
    [sym__filetest] = ACTIONS(3175),
  },
  [STATE(1723)] = {
    [sym_heredoc_content] = STATE(1723),
    [sym__identifier] = ACTIONS(3121),
    [anon_sym_LBRACE] = ACTIONS(3121),
    [anon_sym_DOLLAR] = ACTIONS(3121),
    [anon_sym_AT] = ACTIONS(3121),
    [anon_sym_RPAREN] = ACTIONS(3121),
    [anon_sym_my] = ACTIONS(3121),
    [anon_sym_extended] = ACTIONS(3121),
    [anon_sym_async] = ACTIONS(3121),
    [anon_sym_sub] = ACTIONS(3121),
    [anon_sym_method] = ACTIONS(3121),
    [anon_sym_LPAREN] = ACTIONS(3121),
    [anon_sym_continue] = ACTIONS(3121),
    [anon_sym_state] = ACTIONS(3121),
    [anon_sym_our] = ACTIONS(3121),
    [anon_sym_LBRACK] = ACTIONS(3121),
    [anon_sym_LT_LT] = ACTIONS(3121),
    [anon_sym_SLASH_SLASH] = ACTIONS(3121),
    [anon_sym_PLUS] = ACTIONS(3121),
    [anon_sym_DASH] = ACTIONS(3121),
    [anon_sym_TILDE] = ACTIONS(3121),
    [anon_sym_BANG] = ACTIONS(3121),
    [anon_sym_PLUS_PLUS] = ACTIONS(3121),
    [anon_sym_DASH_DASH] = ACTIONS(3121),
    [anon_sym_BSLASH] = ACTIONS(3119),
    [anon_sym_do] = ACTIONS(3121),
    [anon_sym_eval] = ACTIONS(3121),
    [anon_sym_await] = ACTIONS(3121),
    [anon_sym_field] = ACTIONS(3121),
    [anon_sym_local] = ACTIONS(3121),
    [anon_sym_dynamically] = ACTIONS(3121),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3121),
    [anon_sym_require] = ACTIONS(3121),
    [anon_sym_map] = ACTIONS(3121),
    [anon_sym_grep] = ACTIONS(3121),
    [anon_sym_sort] = ACTIONS(3121),
    [anon_sym_goto] = ACTIONS(3121),
    [anon_sym_return] = ACTIONS(3121),
    [anon_sym_undef] = ACTIONS(3121),
    [anon_sym_PERCENT2] = ACTIONS(3121),
    [anon_sym_AMP2] = ACTIONS(3121),
    [anon_sym_STAR2] = ACTIONS(3121),
    [anon_sym_last] = ACTIONS(3121),
    [anon_sym_next] = ACTIONS(3121),
    [anon_sym_redo] = ACTIONS(3121),
    [anon_sym___FILE__] = ACTIONS(3121),
    [anon_sym___LINE__] = ACTIONS(3121),
    [anon_sym___PACKAGE__] = ACTIONS(3121),
    [anon_sym___SUB__] = ACTIONS(3121),
    [anon_sym_break] = ACTIONS(3121),
    [anon_sym_fork] = ACTIONS(3121),
    [anon_sym_getppid] = ACTIONS(3121),
    [anon_sym_time] = ACTIONS(3121),
    [anon_sym_times] = ACTIONS(3121),
    [anon_sym_wait] = ACTIONS(3121),
    [anon_sym_wantarray] = ACTIONS(3121),
    [anon_sym_abs] = ACTIONS(3121),
    [anon_sym_alarm] = ACTIONS(3121),
    [anon_sym_chop] = ACTIONS(3121),
    [anon_sym_chdir] = ACTIONS(3121),
    [anon_sym_close] = ACTIONS(3121),
    [anon_sym_closedir] = ACTIONS(3121),
    [anon_sym_caller] = ACTIONS(3121),
    [anon_sym_chomp] = ACTIONS(3121),
    [anon_sym_chr] = ACTIONS(3121),
    [anon_sym_cos] = ACTIONS(3121),
    [anon_sym_chroot] = ACTIONS(3121),
    [anon_sym_defined] = ACTIONS(3121),
    [anon_sym_delete] = ACTIONS(3121),
    [anon_sym_dbmclose] = ACTIONS(3121),
    [anon_sym_exists] = ACTIONS(3121),
    [anon_sym_exit] = ACTIONS(3121),
    [anon_sym_eof] = ACTIONS(3121),
    [anon_sym_exp] = ACTIONS(3121),
    [anon_sym_each] = ACTIONS(3121),
    [anon_sym_fc] = ACTIONS(3121),
    [anon_sym_fileno] = ACTIONS(3121),
    [anon_sym_gmtime] = ACTIONS(3121),
    [anon_sym_getc] = ACTIONS(3121),
    [anon_sym_getpgrp] = ACTIONS(3121),
    [anon_sym_getprotobyname] = ACTIONS(3121),
    [anon_sym_getpwname] = ACTIONS(3121),
    [anon_sym_getpwuid] = ACTIONS(3121),
    [anon_sym_getpeername] = ACTIONS(3121),
    [anon_sym_getnetbyname] = ACTIONS(3121),
    [anon_sym_getsockname] = ACTIONS(3121),
    [anon_sym_getgrnam] = ACTIONS(3121),
    [anon_sym_getgrgid] = ACTIONS(3121),
    [anon_sym_hex] = ACTIONS(3121),
    [anon_sym_int] = ACTIONS(3121),
    [anon_sym_keys] = ACTIONS(3121),
    [anon_sym_lc] = ACTIONS(3121),
    [anon_sym_lcfirst] = ACTIONS(3121),
    [anon_sym_length] = ACTIONS(3121),
    [anon_sym_localtime] = ACTIONS(3121),
    [anon_sym_log] = ACTIONS(3121),
    [anon_sym_lock] = ACTIONS(3121),
    [anon_sym_lstat] = ACTIONS(3121),
    [anon_sym_oct] = ACTIONS(3121),
    [anon_sym_ord] = ACTIONS(3121),
    [anon_sym_prototype] = ACTIONS(3121),
    [anon_sym_pop] = ACTIONS(3121),
    [anon_sym_pos] = ACTIONS(3121),
    [anon_sym_quotemeta] = ACTIONS(3121),
    [anon_sym_reset] = ACTIONS(3121),
    [anon_sym_rand] = ACTIONS(3121),
    [anon_sym_rmdir] = ACTIONS(3121),
    [anon_sym_readdir] = ACTIONS(3121),
    [anon_sym_readline] = ACTIONS(3121),
    [anon_sym_readpipe] = ACTIONS(3121),
    [anon_sym_rewinddir] = ACTIONS(3121),
    [anon_sym_readlink] = ACTIONS(3121),
    [anon_sym_ref] = ACTIONS(3121),
    [anon_sym_scalar] = ACTIONS(3121),
    [anon_sym_shift] = ACTIONS(3121),
    [anon_sym_sin] = ACTIONS(3121),
    [anon_sym_sleep] = ACTIONS(3121),
    [anon_sym_sqrt] = ACTIONS(3121),
    [anon_sym_srand] = ACTIONS(3121),
    [anon_sym_stat] = ACTIONS(3121),
    [anon_sym_study] = ACTIONS(3121),
    [anon_sym_tell] = ACTIONS(3121),
    [anon_sym_telldir] = ACTIONS(3121),
    [anon_sym_tied] = ACTIONS(3121),
    [anon_sym_uc] = ACTIONS(3121),
    [anon_sym_ucfirst] = ACTIONS(3121),
    [anon_sym_untie] = ACTIONS(3121),
    [anon_sym_umask] = ACTIONS(3121),
    [anon_sym_values] = ACTIONS(3121),
    [anon_sym_write] = ACTIONS(3121),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3121),
    [anon_sym_qq] = ACTIONS(3121),
    [anon_sym_qw] = ACTIONS(3121),
    [anon_sym_qx] = ACTIONS(3121),
    [anon_sym_qr] = ACTIONS(3121),
    [anon_sym_m] = ACTIONS(3121),
    [anon_sym_s] = ACTIONS(3121),
    [anon_sym_tr] = ACTIONS(3121),
    [anon_sym_y] = ACTIONS(3121),
    [aux_sym__bareword_token1] = ACTIONS(3121),
    [sym_number] = ACTIONS(3121),
    [anon_sym_true] = ACTIONS(3121),
    [anon_sym_false] = ACTIONS(3121),
    [sym__single_quote] = ACTIONS(3119),
    [sym__double_quote] = ACTIONS(3119),
    [sym__backtick_quote] = ACTIONS(3119),
    [sym__search_slash_quote] = ACTIONS(3119),
    [sym__open_readline_bracket] = ACTIONS(3119),
    [sym__open_fileglob_bracket] = ACTIONS(3119),
    [sym__PERLY_HEREDOC] = ACTIONS(3119),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3119),
    [sym__filetest] = ACTIONS(3119),
  },
  [STATE(1724)] = {
    [sym_heredoc_content] = STATE(1724),
    [sym__identifier] = ACTIONS(3125),
    [anon_sym_LBRACE] = ACTIONS(3125),
    [anon_sym_DOLLAR] = ACTIONS(3125),
    [anon_sym_AT] = ACTIONS(3125),
    [anon_sym_RPAREN] = ACTIONS(3125),
    [anon_sym_my] = ACTIONS(3125),
    [anon_sym_extended] = ACTIONS(3125),
    [anon_sym_async] = ACTIONS(3125),
    [anon_sym_sub] = ACTIONS(3125),
    [anon_sym_method] = ACTIONS(3125),
    [anon_sym_LPAREN] = ACTIONS(3125),
    [anon_sym_continue] = ACTIONS(3125),
    [anon_sym_state] = ACTIONS(3125),
    [anon_sym_our] = ACTIONS(3125),
    [anon_sym_LBRACK] = ACTIONS(3125),
    [anon_sym_LT_LT] = ACTIONS(3125),
    [anon_sym_SLASH_SLASH] = ACTIONS(3125),
    [anon_sym_PLUS] = ACTIONS(3125),
    [anon_sym_DASH] = ACTIONS(3125),
    [anon_sym_TILDE] = ACTIONS(3125),
    [anon_sym_BANG] = ACTIONS(3125),
    [anon_sym_PLUS_PLUS] = ACTIONS(3125),
    [anon_sym_DASH_DASH] = ACTIONS(3125),
    [anon_sym_BSLASH] = ACTIONS(3123),
    [anon_sym_do] = ACTIONS(3125),
    [anon_sym_eval] = ACTIONS(3125),
    [anon_sym_await] = ACTIONS(3125),
    [anon_sym_field] = ACTIONS(3125),
    [anon_sym_local] = ACTIONS(3125),
    [anon_sym_dynamically] = ACTIONS(3125),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3125),
    [anon_sym_require] = ACTIONS(3125),
    [anon_sym_map] = ACTIONS(3125),
    [anon_sym_grep] = ACTIONS(3125),
    [anon_sym_sort] = ACTIONS(3125),
    [anon_sym_goto] = ACTIONS(3125),
    [anon_sym_return] = ACTIONS(3125),
    [anon_sym_undef] = ACTIONS(3125),
    [anon_sym_PERCENT2] = ACTIONS(3125),
    [anon_sym_AMP2] = ACTIONS(3125),
    [anon_sym_STAR2] = ACTIONS(3125),
    [anon_sym_last] = ACTIONS(3125),
    [anon_sym_next] = ACTIONS(3125),
    [anon_sym_redo] = ACTIONS(3125),
    [anon_sym___FILE__] = ACTIONS(3125),
    [anon_sym___LINE__] = ACTIONS(3125),
    [anon_sym___PACKAGE__] = ACTIONS(3125),
    [anon_sym___SUB__] = ACTIONS(3125),
    [anon_sym_break] = ACTIONS(3125),
    [anon_sym_fork] = ACTIONS(3125),
    [anon_sym_getppid] = ACTIONS(3125),
    [anon_sym_time] = ACTIONS(3125),
    [anon_sym_times] = ACTIONS(3125),
    [anon_sym_wait] = ACTIONS(3125),
    [anon_sym_wantarray] = ACTIONS(3125),
    [anon_sym_abs] = ACTIONS(3125),
    [anon_sym_alarm] = ACTIONS(3125),
    [anon_sym_chop] = ACTIONS(3125),
    [anon_sym_chdir] = ACTIONS(3125),
    [anon_sym_close] = ACTIONS(3125),
    [anon_sym_closedir] = ACTIONS(3125),
    [anon_sym_caller] = ACTIONS(3125),
    [anon_sym_chomp] = ACTIONS(3125),
    [anon_sym_chr] = ACTIONS(3125),
    [anon_sym_cos] = ACTIONS(3125),
    [anon_sym_chroot] = ACTIONS(3125),
    [anon_sym_defined] = ACTIONS(3125),
    [anon_sym_delete] = ACTIONS(3125),
    [anon_sym_dbmclose] = ACTIONS(3125),
    [anon_sym_exists] = ACTIONS(3125),
    [anon_sym_exit] = ACTIONS(3125),
    [anon_sym_eof] = ACTIONS(3125),
    [anon_sym_exp] = ACTIONS(3125),
    [anon_sym_each] = ACTIONS(3125),
    [anon_sym_fc] = ACTIONS(3125),
    [anon_sym_fileno] = ACTIONS(3125),
    [anon_sym_gmtime] = ACTIONS(3125),
    [anon_sym_getc] = ACTIONS(3125),
    [anon_sym_getpgrp] = ACTIONS(3125),
    [anon_sym_getprotobyname] = ACTIONS(3125),
    [anon_sym_getpwname] = ACTIONS(3125),
    [anon_sym_getpwuid] = ACTIONS(3125),
    [anon_sym_getpeername] = ACTIONS(3125),
    [anon_sym_getnetbyname] = ACTIONS(3125),
    [anon_sym_getsockname] = ACTIONS(3125),
    [anon_sym_getgrnam] = ACTIONS(3125),
    [anon_sym_getgrgid] = ACTIONS(3125),
    [anon_sym_hex] = ACTIONS(3125),
    [anon_sym_int] = ACTIONS(3125),
    [anon_sym_keys] = ACTIONS(3125),
    [anon_sym_lc] = ACTIONS(3125),
    [anon_sym_lcfirst] = ACTIONS(3125),
    [anon_sym_length] = ACTIONS(3125),
    [anon_sym_localtime] = ACTIONS(3125),
    [anon_sym_log] = ACTIONS(3125),
    [anon_sym_lock] = ACTIONS(3125),
    [anon_sym_lstat] = ACTIONS(3125),
    [anon_sym_oct] = ACTIONS(3125),
    [anon_sym_ord] = ACTIONS(3125),
    [anon_sym_prototype] = ACTIONS(3125),
    [anon_sym_pop] = ACTIONS(3125),
    [anon_sym_pos] = ACTIONS(3125),
    [anon_sym_quotemeta] = ACTIONS(3125),
    [anon_sym_reset] = ACTIONS(3125),
    [anon_sym_rand] = ACTIONS(3125),
    [anon_sym_rmdir] = ACTIONS(3125),
    [anon_sym_readdir] = ACTIONS(3125),
    [anon_sym_readline] = ACTIONS(3125),
    [anon_sym_readpipe] = ACTIONS(3125),
    [anon_sym_rewinddir] = ACTIONS(3125),
    [anon_sym_readlink] = ACTIONS(3125),
    [anon_sym_ref] = ACTIONS(3125),
    [anon_sym_scalar] = ACTIONS(3125),
    [anon_sym_shift] = ACTIONS(3125),
    [anon_sym_sin] = ACTIONS(3125),
    [anon_sym_sleep] = ACTIONS(3125),
    [anon_sym_sqrt] = ACTIONS(3125),
    [anon_sym_srand] = ACTIONS(3125),
    [anon_sym_stat] = ACTIONS(3125),
    [anon_sym_study] = ACTIONS(3125),
    [anon_sym_tell] = ACTIONS(3125),
    [anon_sym_telldir] = ACTIONS(3125),
    [anon_sym_tied] = ACTIONS(3125),
    [anon_sym_uc] = ACTIONS(3125),
    [anon_sym_ucfirst] = ACTIONS(3125),
    [anon_sym_untie] = ACTIONS(3125),
    [anon_sym_umask] = ACTIONS(3125),
    [anon_sym_values] = ACTIONS(3125),
    [anon_sym_write] = ACTIONS(3125),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3125),
    [anon_sym_qq] = ACTIONS(3125),
    [anon_sym_qw] = ACTIONS(3125),
    [anon_sym_qx] = ACTIONS(3125),
    [anon_sym_qr] = ACTIONS(3125),
    [anon_sym_m] = ACTIONS(3125),
    [anon_sym_s] = ACTIONS(3125),
    [anon_sym_tr] = ACTIONS(3125),
    [anon_sym_y] = ACTIONS(3125),
    [aux_sym__bareword_token1] = ACTIONS(3125),
    [sym_number] = ACTIONS(3125),
    [anon_sym_true] = ACTIONS(3125),
    [anon_sym_false] = ACTIONS(3125),
    [sym__single_quote] = ACTIONS(3123),
    [sym__double_quote] = ACTIONS(3123),
    [sym__backtick_quote] = ACTIONS(3123),
    [sym__search_slash_quote] = ACTIONS(3123),
    [sym__open_readline_bracket] = ACTIONS(3123),
    [sym__open_fileglob_bracket] = ACTIONS(3123),
    [sym__PERLY_HEREDOC] = ACTIONS(3123),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3123),
    [sym__filetest] = ACTIONS(3123),
  },
  [STATE(1725)] = {
    [sym_heredoc_content] = STATE(1725),
    [sym__identifier] = ACTIONS(3157),
    [anon_sym_LBRACE] = ACTIONS(3157),
    [anon_sym_DOLLAR] = ACTIONS(3157),
    [anon_sym_AT] = ACTIONS(3157),
    [anon_sym_RPAREN] = ACTIONS(3157),
    [anon_sym_my] = ACTIONS(3157),
    [anon_sym_extended] = ACTIONS(3157),
    [anon_sym_async] = ACTIONS(3157),
    [anon_sym_sub] = ACTIONS(3157),
    [anon_sym_method] = ACTIONS(3157),
    [anon_sym_LPAREN] = ACTIONS(3157),
    [anon_sym_continue] = ACTIONS(3157),
    [anon_sym_state] = ACTIONS(3157),
    [anon_sym_our] = ACTIONS(3157),
    [anon_sym_LBRACK] = ACTIONS(3157),
    [anon_sym_LT_LT] = ACTIONS(3157),
    [anon_sym_SLASH_SLASH] = ACTIONS(3157),
    [anon_sym_PLUS] = ACTIONS(3157),
    [anon_sym_DASH] = ACTIONS(3157),
    [anon_sym_TILDE] = ACTIONS(3157),
    [anon_sym_BANG] = ACTIONS(3157),
    [anon_sym_PLUS_PLUS] = ACTIONS(3157),
    [anon_sym_DASH_DASH] = ACTIONS(3157),
    [anon_sym_BSLASH] = ACTIONS(3155),
    [anon_sym_do] = ACTIONS(3157),
    [anon_sym_eval] = ACTIONS(3157),
    [anon_sym_await] = ACTIONS(3157),
    [anon_sym_field] = ACTIONS(3157),
    [anon_sym_local] = ACTIONS(3157),
    [anon_sym_dynamically] = ACTIONS(3157),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3157),
    [anon_sym_require] = ACTIONS(3157),
    [anon_sym_map] = ACTIONS(3157),
    [anon_sym_grep] = ACTIONS(3157),
    [anon_sym_sort] = ACTIONS(3157),
    [anon_sym_goto] = ACTIONS(3157),
    [anon_sym_return] = ACTIONS(3157),
    [anon_sym_undef] = ACTIONS(3157),
    [anon_sym_PERCENT2] = ACTIONS(3157),
    [anon_sym_AMP2] = ACTIONS(3157),
    [anon_sym_STAR2] = ACTIONS(3157),
    [anon_sym_last] = ACTIONS(3157),
    [anon_sym_next] = ACTIONS(3157),
    [anon_sym_redo] = ACTIONS(3157),
    [anon_sym___FILE__] = ACTIONS(3157),
    [anon_sym___LINE__] = ACTIONS(3157),
    [anon_sym___PACKAGE__] = ACTIONS(3157),
    [anon_sym___SUB__] = ACTIONS(3157),
    [anon_sym_break] = ACTIONS(3157),
    [anon_sym_fork] = ACTIONS(3157),
    [anon_sym_getppid] = ACTIONS(3157),
    [anon_sym_time] = ACTIONS(3157),
    [anon_sym_times] = ACTIONS(3157),
    [anon_sym_wait] = ACTIONS(3157),
    [anon_sym_wantarray] = ACTIONS(3157),
    [anon_sym_abs] = ACTIONS(3157),
    [anon_sym_alarm] = ACTIONS(3157),
    [anon_sym_chop] = ACTIONS(3157),
    [anon_sym_chdir] = ACTIONS(3157),
    [anon_sym_close] = ACTIONS(3157),
    [anon_sym_closedir] = ACTIONS(3157),
    [anon_sym_caller] = ACTIONS(3157),
    [anon_sym_chomp] = ACTIONS(3157),
    [anon_sym_chr] = ACTIONS(3157),
    [anon_sym_cos] = ACTIONS(3157),
    [anon_sym_chroot] = ACTIONS(3157),
    [anon_sym_defined] = ACTIONS(3157),
    [anon_sym_delete] = ACTIONS(3157),
    [anon_sym_dbmclose] = ACTIONS(3157),
    [anon_sym_exists] = ACTIONS(3157),
    [anon_sym_exit] = ACTIONS(3157),
    [anon_sym_eof] = ACTIONS(3157),
    [anon_sym_exp] = ACTIONS(3157),
    [anon_sym_each] = ACTIONS(3157),
    [anon_sym_fc] = ACTIONS(3157),
    [anon_sym_fileno] = ACTIONS(3157),
    [anon_sym_gmtime] = ACTIONS(3157),
    [anon_sym_getc] = ACTIONS(3157),
    [anon_sym_getpgrp] = ACTIONS(3157),
    [anon_sym_getprotobyname] = ACTIONS(3157),
    [anon_sym_getpwname] = ACTIONS(3157),
    [anon_sym_getpwuid] = ACTIONS(3157),
    [anon_sym_getpeername] = ACTIONS(3157),
    [anon_sym_getnetbyname] = ACTIONS(3157),
    [anon_sym_getsockname] = ACTIONS(3157),
    [anon_sym_getgrnam] = ACTIONS(3157),
    [anon_sym_getgrgid] = ACTIONS(3157),
    [anon_sym_hex] = ACTIONS(3157),
    [anon_sym_int] = ACTIONS(3157),
    [anon_sym_keys] = ACTIONS(3157),
    [anon_sym_lc] = ACTIONS(3157),
    [anon_sym_lcfirst] = ACTIONS(3157),
    [anon_sym_length] = ACTIONS(3157),
    [anon_sym_localtime] = ACTIONS(3157),
    [anon_sym_log] = ACTIONS(3157),
    [anon_sym_lock] = ACTIONS(3157),
    [anon_sym_lstat] = ACTIONS(3157),
    [anon_sym_oct] = ACTIONS(3157),
    [anon_sym_ord] = ACTIONS(3157),
    [anon_sym_prototype] = ACTIONS(3157),
    [anon_sym_pop] = ACTIONS(3157),
    [anon_sym_pos] = ACTIONS(3157),
    [anon_sym_quotemeta] = ACTIONS(3157),
    [anon_sym_reset] = ACTIONS(3157),
    [anon_sym_rand] = ACTIONS(3157),
    [anon_sym_rmdir] = ACTIONS(3157),
    [anon_sym_readdir] = ACTIONS(3157),
    [anon_sym_readline] = ACTIONS(3157),
    [anon_sym_readpipe] = ACTIONS(3157),
    [anon_sym_rewinddir] = ACTIONS(3157),
    [anon_sym_readlink] = ACTIONS(3157),
    [anon_sym_ref] = ACTIONS(3157),
    [anon_sym_scalar] = ACTIONS(3157),
    [anon_sym_shift] = ACTIONS(3157),
    [anon_sym_sin] = ACTIONS(3157),
    [anon_sym_sleep] = ACTIONS(3157),
    [anon_sym_sqrt] = ACTIONS(3157),
    [anon_sym_srand] = ACTIONS(3157),
    [anon_sym_stat] = ACTIONS(3157),
    [anon_sym_study] = ACTIONS(3157),
    [anon_sym_tell] = ACTIONS(3157),
    [anon_sym_telldir] = ACTIONS(3157),
    [anon_sym_tied] = ACTIONS(3157),
    [anon_sym_uc] = ACTIONS(3157),
    [anon_sym_ucfirst] = ACTIONS(3157),
    [anon_sym_untie] = ACTIONS(3157),
    [anon_sym_umask] = ACTIONS(3157),
    [anon_sym_values] = ACTIONS(3157),
    [anon_sym_write] = ACTIONS(3157),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3157),
    [anon_sym_qq] = ACTIONS(3157),
    [anon_sym_qw] = ACTIONS(3157),
    [anon_sym_qx] = ACTIONS(3157),
    [anon_sym_qr] = ACTIONS(3157),
    [anon_sym_m] = ACTIONS(3157),
    [anon_sym_s] = ACTIONS(3157),
    [anon_sym_tr] = ACTIONS(3157),
    [anon_sym_y] = ACTIONS(3157),
    [aux_sym__bareword_token1] = ACTIONS(3157),
    [sym_number] = ACTIONS(3157),
    [anon_sym_true] = ACTIONS(3157),
    [anon_sym_false] = ACTIONS(3157),
    [sym__single_quote] = ACTIONS(3155),
    [sym__double_quote] = ACTIONS(3155),
    [sym__backtick_quote] = ACTIONS(3155),
    [sym__search_slash_quote] = ACTIONS(3155),
    [sym__open_readline_bracket] = ACTIONS(3155),
    [sym__open_fileglob_bracket] = ACTIONS(3155),
    [sym__PERLY_HEREDOC] = ACTIONS(3155),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3155),
    [sym__filetest] = ACTIONS(3155),
  },
  [STATE(1726)] = {
    [sym_heredoc_content] = STATE(1726),
    [sym__identifier] = ACTIONS(2966),
    [anon_sym_LBRACE] = ACTIONS(2966),
    [anon_sym_DOLLAR] = ACTIONS(2966),
    [anon_sym_AT] = ACTIONS(2966),
    [anon_sym_RPAREN] = ACTIONS(2966),
    [anon_sym_my] = ACTIONS(2966),
    [anon_sym_extended] = ACTIONS(2966),
    [anon_sym_async] = ACTIONS(2966),
    [anon_sym_sub] = ACTIONS(2966),
    [anon_sym_method] = ACTIONS(2966),
    [anon_sym_LPAREN] = ACTIONS(2966),
    [anon_sym_continue] = ACTIONS(2966),
    [anon_sym_state] = ACTIONS(2966),
    [anon_sym_our] = ACTIONS(2966),
    [anon_sym_LBRACK] = ACTIONS(2966),
    [anon_sym_LT_LT] = ACTIONS(2966),
    [anon_sym_SLASH_SLASH] = ACTIONS(2966),
    [anon_sym_PLUS] = ACTIONS(2966),
    [anon_sym_DASH] = ACTIONS(2966),
    [anon_sym_TILDE] = ACTIONS(2966),
    [anon_sym_BANG] = ACTIONS(2966),
    [anon_sym_PLUS_PLUS] = ACTIONS(2966),
    [anon_sym_DASH_DASH] = ACTIONS(2966),
    [anon_sym_BSLASH] = ACTIONS(2964),
    [anon_sym_do] = ACTIONS(2966),
    [anon_sym_eval] = ACTIONS(2966),
    [anon_sym_await] = ACTIONS(2966),
    [anon_sym_field] = ACTIONS(2966),
    [anon_sym_local] = ACTIONS(2966),
    [anon_sym_dynamically] = ACTIONS(2966),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2966),
    [anon_sym_require] = ACTIONS(2966),
    [anon_sym_map] = ACTIONS(2966),
    [anon_sym_grep] = ACTIONS(2966),
    [anon_sym_sort] = ACTIONS(2966),
    [anon_sym_goto] = ACTIONS(2966),
    [anon_sym_return] = ACTIONS(2966),
    [anon_sym_undef] = ACTIONS(2966),
    [anon_sym_PERCENT2] = ACTIONS(2966),
    [anon_sym_AMP2] = ACTIONS(2966),
    [anon_sym_STAR2] = ACTIONS(2966),
    [anon_sym_last] = ACTIONS(2966),
    [anon_sym_next] = ACTIONS(2966),
    [anon_sym_redo] = ACTIONS(2966),
    [anon_sym___FILE__] = ACTIONS(2966),
    [anon_sym___LINE__] = ACTIONS(2966),
    [anon_sym___PACKAGE__] = ACTIONS(2966),
    [anon_sym___SUB__] = ACTIONS(2966),
    [anon_sym_break] = ACTIONS(2966),
    [anon_sym_fork] = ACTIONS(2966),
    [anon_sym_getppid] = ACTIONS(2966),
    [anon_sym_time] = ACTIONS(2966),
    [anon_sym_times] = ACTIONS(2966),
    [anon_sym_wait] = ACTIONS(2966),
    [anon_sym_wantarray] = ACTIONS(2966),
    [anon_sym_abs] = ACTIONS(2966),
    [anon_sym_alarm] = ACTIONS(2966),
    [anon_sym_chop] = ACTIONS(2966),
    [anon_sym_chdir] = ACTIONS(2966),
    [anon_sym_close] = ACTIONS(2966),
    [anon_sym_closedir] = ACTIONS(2966),
    [anon_sym_caller] = ACTIONS(2966),
    [anon_sym_chomp] = ACTIONS(2966),
    [anon_sym_chr] = ACTIONS(2966),
    [anon_sym_cos] = ACTIONS(2966),
    [anon_sym_chroot] = ACTIONS(2966),
    [anon_sym_defined] = ACTIONS(2966),
    [anon_sym_delete] = ACTIONS(2966),
    [anon_sym_dbmclose] = ACTIONS(2966),
    [anon_sym_exists] = ACTIONS(2966),
    [anon_sym_exit] = ACTIONS(2966),
    [anon_sym_eof] = ACTIONS(2966),
    [anon_sym_exp] = ACTIONS(2966),
    [anon_sym_each] = ACTIONS(2966),
    [anon_sym_fc] = ACTIONS(2966),
    [anon_sym_fileno] = ACTIONS(2966),
    [anon_sym_gmtime] = ACTIONS(2966),
    [anon_sym_getc] = ACTIONS(2966),
    [anon_sym_getpgrp] = ACTIONS(2966),
    [anon_sym_getprotobyname] = ACTIONS(2966),
    [anon_sym_getpwname] = ACTIONS(2966),
    [anon_sym_getpwuid] = ACTIONS(2966),
    [anon_sym_getpeername] = ACTIONS(2966),
    [anon_sym_getnetbyname] = ACTIONS(2966),
    [anon_sym_getsockname] = ACTIONS(2966),
    [anon_sym_getgrnam] = ACTIONS(2966),
    [anon_sym_getgrgid] = ACTIONS(2966),
    [anon_sym_hex] = ACTIONS(2966),
    [anon_sym_int] = ACTIONS(2966),
    [anon_sym_keys] = ACTIONS(2966),
    [anon_sym_lc] = ACTIONS(2966),
    [anon_sym_lcfirst] = ACTIONS(2966),
    [anon_sym_length] = ACTIONS(2966),
    [anon_sym_localtime] = ACTIONS(2966),
    [anon_sym_log] = ACTIONS(2966),
    [anon_sym_lock] = ACTIONS(2966),
    [anon_sym_lstat] = ACTIONS(2966),
    [anon_sym_oct] = ACTIONS(2966),
    [anon_sym_ord] = ACTIONS(2966),
    [anon_sym_prototype] = ACTIONS(2966),
    [anon_sym_pop] = ACTIONS(2966),
    [anon_sym_pos] = ACTIONS(2966),
    [anon_sym_quotemeta] = ACTIONS(2966),
    [anon_sym_reset] = ACTIONS(2966),
    [anon_sym_rand] = ACTIONS(2966),
    [anon_sym_rmdir] = ACTIONS(2966),
    [anon_sym_readdir] = ACTIONS(2966),
    [anon_sym_readline] = ACTIONS(2966),
    [anon_sym_readpipe] = ACTIONS(2966),
    [anon_sym_rewinddir] = ACTIONS(2966),
    [anon_sym_readlink] = ACTIONS(2966),
    [anon_sym_ref] = ACTIONS(2966),
    [anon_sym_scalar] = ACTIONS(2966),
    [anon_sym_shift] = ACTIONS(2966),
    [anon_sym_sin] = ACTIONS(2966),
    [anon_sym_sleep] = ACTIONS(2966),
    [anon_sym_sqrt] = ACTIONS(2966),
    [anon_sym_srand] = ACTIONS(2966),
    [anon_sym_stat] = ACTIONS(2966),
    [anon_sym_study] = ACTIONS(2966),
    [anon_sym_tell] = ACTIONS(2966),
    [anon_sym_telldir] = ACTIONS(2966),
    [anon_sym_tied] = ACTIONS(2966),
    [anon_sym_uc] = ACTIONS(2966),
    [anon_sym_ucfirst] = ACTIONS(2966),
    [anon_sym_untie] = ACTIONS(2966),
    [anon_sym_umask] = ACTIONS(2966),
    [anon_sym_values] = ACTIONS(2966),
    [anon_sym_write] = ACTIONS(2966),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2966),
    [anon_sym_qq] = ACTIONS(2966),
    [anon_sym_qw] = ACTIONS(2966),
    [anon_sym_qx] = ACTIONS(2966),
    [anon_sym_qr] = ACTIONS(2966),
    [anon_sym_m] = ACTIONS(2966),
    [anon_sym_s] = ACTIONS(2966),
    [anon_sym_tr] = ACTIONS(2966),
    [anon_sym_y] = ACTIONS(2966),
    [aux_sym__bareword_token1] = ACTIONS(2966),
    [sym_number] = ACTIONS(2966),
    [anon_sym_true] = ACTIONS(2966),
    [anon_sym_false] = ACTIONS(2966),
    [sym__single_quote] = ACTIONS(2964),
    [sym__double_quote] = ACTIONS(2964),
    [sym__backtick_quote] = ACTIONS(2964),
    [sym__search_slash_quote] = ACTIONS(2964),
    [sym__open_readline_bracket] = ACTIONS(2964),
    [sym__open_fileglob_bracket] = ACTIONS(2964),
    [sym__PERLY_HEREDOC] = ACTIONS(2964),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2964),
    [sym__filetest] = ACTIONS(2964),
  },
  [STATE(1727)] = {
    [sym_heredoc_content] = STATE(1727),
    [sym__identifier] = ACTIONS(2898),
    [anon_sym_LBRACE] = ACTIONS(2898),
    [anon_sym_DOLLAR] = ACTIONS(2898),
    [anon_sym_AT] = ACTIONS(2898),
    [anon_sym_RPAREN] = ACTIONS(2898),
    [anon_sym_my] = ACTIONS(2898),
    [anon_sym_extended] = ACTIONS(2898),
    [anon_sym_async] = ACTIONS(2898),
    [anon_sym_sub] = ACTIONS(2898),
    [anon_sym_method] = ACTIONS(2898),
    [anon_sym_LPAREN] = ACTIONS(2898),
    [anon_sym_continue] = ACTIONS(2898),
    [anon_sym_state] = ACTIONS(2898),
    [anon_sym_our] = ACTIONS(2898),
    [anon_sym_LBRACK] = ACTIONS(2898),
    [anon_sym_LT_LT] = ACTIONS(2898),
    [anon_sym_SLASH_SLASH] = ACTIONS(2898),
    [anon_sym_PLUS] = ACTIONS(2898),
    [anon_sym_DASH] = ACTIONS(2898),
    [anon_sym_TILDE] = ACTIONS(2898),
    [anon_sym_BANG] = ACTIONS(2898),
    [anon_sym_PLUS_PLUS] = ACTIONS(2898),
    [anon_sym_DASH_DASH] = ACTIONS(2898),
    [anon_sym_BSLASH] = ACTIONS(2896),
    [anon_sym_do] = ACTIONS(2898),
    [anon_sym_eval] = ACTIONS(2898),
    [anon_sym_await] = ACTIONS(2898),
    [anon_sym_field] = ACTIONS(2898),
    [anon_sym_local] = ACTIONS(2898),
    [anon_sym_dynamically] = ACTIONS(2898),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2898),
    [anon_sym_require] = ACTIONS(2898),
    [anon_sym_map] = ACTIONS(2898),
    [anon_sym_grep] = ACTIONS(2898),
    [anon_sym_sort] = ACTIONS(2898),
    [anon_sym_goto] = ACTIONS(2898),
    [anon_sym_return] = ACTIONS(2898),
    [anon_sym_undef] = ACTIONS(2898),
    [anon_sym_PERCENT2] = ACTIONS(2898),
    [anon_sym_AMP2] = ACTIONS(2898),
    [anon_sym_STAR2] = ACTIONS(2898),
    [anon_sym_last] = ACTIONS(2898),
    [anon_sym_next] = ACTIONS(2898),
    [anon_sym_redo] = ACTIONS(2898),
    [anon_sym___FILE__] = ACTIONS(2898),
    [anon_sym___LINE__] = ACTIONS(2898),
    [anon_sym___PACKAGE__] = ACTIONS(2898),
    [anon_sym___SUB__] = ACTIONS(2898),
    [anon_sym_break] = ACTIONS(2898),
    [anon_sym_fork] = ACTIONS(2898),
    [anon_sym_getppid] = ACTIONS(2898),
    [anon_sym_time] = ACTIONS(2898),
    [anon_sym_times] = ACTIONS(2898),
    [anon_sym_wait] = ACTIONS(2898),
    [anon_sym_wantarray] = ACTIONS(2898),
    [anon_sym_abs] = ACTIONS(2898),
    [anon_sym_alarm] = ACTIONS(2898),
    [anon_sym_chop] = ACTIONS(2898),
    [anon_sym_chdir] = ACTIONS(2898),
    [anon_sym_close] = ACTIONS(2898),
    [anon_sym_closedir] = ACTIONS(2898),
    [anon_sym_caller] = ACTIONS(2898),
    [anon_sym_chomp] = ACTIONS(2898),
    [anon_sym_chr] = ACTIONS(2898),
    [anon_sym_cos] = ACTIONS(2898),
    [anon_sym_chroot] = ACTIONS(2898),
    [anon_sym_defined] = ACTIONS(2898),
    [anon_sym_delete] = ACTIONS(2898),
    [anon_sym_dbmclose] = ACTIONS(2898),
    [anon_sym_exists] = ACTIONS(2898),
    [anon_sym_exit] = ACTIONS(2898),
    [anon_sym_eof] = ACTIONS(2898),
    [anon_sym_exp] = ACTIONS(2898),
    [anon_sym_each] = ACTIONS(2898),
    [anon_sym_fc] = ACTIONS(2898),
    [anon_sym_fileno] = ACTIONS(2898),
    [anon_sym_gmtime] = ACTIONS(2898),
    [anon_sym_getc] = ACTIONS(2898),
    [anon_sym_getpgrp] = ACTIONS(2898),
    [anon_sym_getprotobyname] = ACTIONS(2898),
    [anon_sym_getpwname] = ACTIONS(2898),
    [anon_sym_getpwuid] = ACTIONS(2898),
    [anon_sym_getpeername] = ACTIONS(2898),
    [anon_sym_getnetbyname] = ACTIONS(2898),
    [anon_sym_getsockname] = ACTIONS(2898),
    [anon_sym_getgrnam] = ACTIONS(2898),
    [anon_sym_getgrgid] = ACTIONS(2898),
    [anon_sym_hex] = ACTIONS(2898),
    [anon_sym_int] = ACTIONS(2898),
    [anon_sym_keys] = ACTIONS(2898),
    [anon_sym_lc] = ACTIONS(2898),
    [anon_sym_lcfirst] = ACTIONS(2898),
    [anon_sym_length] = ACTIONS(2898),
    [anon_sym_localtime] = ACTIONS(2898),
    [anon_sym_log] = ACTIONS(2898),
    [anon_sym_lock] = ACTIONS(2898),
    [anon_sym_lstat] = ACTIONS(2898),
    [anon_sym_oct] = ACTIONS(2898),
    [anon_sym_ord] = ACTIONS(2898),
    [anon_sym_prototype] = ACTIONS(2898),
    [anon_sym_pop] = ACTIONS(2898),
    [anon_sym_pos] = ACTIONS(2898),
    [anon_sym_quotemeta] = ACTIONS(2898),
    [anon_sym_reset] = ACTIONS(2898),
    [anon_sym_rand] = ACTIONS(2898),
    [anon_sym_rmdir] = ACTIONS(2898),
    [anon_sym_readdir] = ACTIONS(2898),
    [anon_sym_readline] = ACTIONS(2898),
    [anon_sym_readpipe] = ACTIONS(2898),
    [anon_sym_rewinddir] = ACTIONS(2898),
    [anon_sym_readlink] = ACTIONS(2898),
    [anon_sym_ref] = ACTIONS(2898),
    [anon_sym_scalar] = ACTIONS(2898),
    [anon_sym_shift] = ACTIONS(2898),
    [anon_sym_sin] = ACTIONS(2898),
    [anon_sym_sleep] = ACTIONS(2898),
    [anon_sym_sqrt] = ACTIONS(2898),
    [anon_sym_srand] = ACTIONS(2898),
    [anon_sym_stat] = ACTIONS(2898),
    [anon_sym_study] = ACTIONS(2898),
    [anon_sym_tell] = ACTIONS(2898),
    [anon_sym_telldir] = ACTIONS(2898),
    [anon_sym_tied] = ACTIONS(2898),
    [anon_sym_uc] = ACTIONS(2898),
    [anon_sym_ucfirst] = ACTIONS(2898),
    [anon_sym_untie] = ACTIONS(2898),
    [anon_sym_umask] = ACTIONS(2898),
    [anon_sym_values] = ACTIONS(2898),
    [anon_sym_write] = ACTIONS(2898),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2898),
    [anon_sym_qq] = ACTIONS(2898),
    [anon_sym_qw] = ACTIONS(2898),
    [anon_sym_qx] = ACTIONS(2898),
    [anon_sym_qr] = ACTIONS(2898),
    [anon_sym_m] = ACTIONS(2898),
    [anon_sym_s] = ACTIONS(2898),
    [anon_sym_tr] = ACTIONS(2898),
    [anon_sym_y] = ACTIONS(2898),
    [aux_sym__bareword_token1] = ACTIONS(2898),
    [sym_number] = ACTIONS(2898),
    [anon_sym_true] = ACTIONS(2898),
    [anon_sym_false] = ACTIONS(2898),
    [sym__single_quote] = ACTIONS(2896),
    [sym__double_quote] = ACTIONS(2896),
    [sym__backtick_quote] = ACTIONS(2896),
    [sym__search_slash_quote] = ACTIONS(2896),
    [sym__open_readline_bracket] = ACTIONS(2896),
    [sym__open_fileglob_bracket] = ACTIONS(2896),
    [sym__PERLY_HEREDOC] = ACTIONS(2896),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2896),
    [sym__filetest] = ACTIONS(2896),
  },
  [STATE(1728)] = {
    [sym_heredoc_content] = STATE(1728),
    [sym__identifier] = ACTIONS(3045),
    [anon_sym_LBRACE] = ACTIONS(3045),
    [anon_sym_DOLLAR] = ACTIONS(3045),
    [anon_sym_AT] = ACTIONS(3045),
    [anon_sym_RPAREN] = ACTIONS(3045),
    [anon_sym_my] = ACTIONS(3045),
    [anon_sym_extended] = ACTIONS(3045),
    [anon_sym_async] = ACTIONS(3045),
    [anon_sym_sub] = ACTIONS(3045),
    [anon_sym_method] = ACTIONS(3045),
    [anon_sym_LPAREN] = ACTIONS(3045),
    [anon_sym_continue] = ACTIONS(3045),
    [anon_sym_state] = ACTIONS(3045),
    [anon_sym_our] = ACTIONS(3045),
    [anon_sym_LBRACK] = ACTIONS(3045),
    [anon_sym_LT_LT] = ACTIONS(3045),
    [anon_sym_SLASH_SLASH] = ACTIONS(3045),
    [anon_sym_PLUS] = ACTIONS(3045),
    [anon_sym_DASH] = ACTIONS(3045),
    [anon_sym_TILDE] = ACTIONS(3045),
    [anon_sym_BANG] = ACTIONS(3045),
    [anon_sym_PLUS_PLUS] = ACTIONS(3045),
    [anon_sym_DASH_DASH] = ACTIONS(3045),
    [anon_sym_BSLASH] = ACTIONS(3043),
    [anon_sym_do] = ACTIONS(3045),
    [anon_sym_eval] = ACTIONS(3045),
    [anon_sym_await] = ACTIONS(3045),
    [anon_sym_field] = ACTIONS(3045),
    [anon_sym_local] = ACTIONS(3045),
    [anon_sym_dynamically] = ACTIONS(3045),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3045),
    [anon_sym_require] = ACTIONS(3045),
    [anon_sym_map] = ACTIONS(3045),
    [anon_sym_grep] = ACTIONS(3045),
    [anon_sym_sort] = ACTIONS(3045),
    [anon_sym_goto] = ACTIONS(3045),
    [anon_sym_return] = ACTIONS(3045),
    [anon_sym_undef] = ACTIONS(3045),
    [anon_sym_PERCENT2] = ACTIONS(3045),
    [anon_sym_AMP2] = ACTIONS(3045),
    [anon_sym_STAR2] = ACTIONS(3045),
    [anon_sym_last] = ACTIONS(3045),
    [anon_sym_next] = ACTIONS(3045),
    [anon_sym_redo] = ACTIONS(3045),
    [anon_sym___FILE__] = ACTIONS(3045),
    [anon_sym___LINE__] = ACTIONS(3045),
    [anon_sym___PACKAGE__] = ACTIONS(3045),
    [anon_sym___SUB__] = ACTIONS(3045),
    [anon_sym_break] = ACTIONS(3045),
    [anon_sym_fork] = ACTIONS(3045),
    [anon_sym_getppid] = ACTIONS(3045),
    [anon_sym_time] = ACTIONS(3045),
    [anon_sym_times] = ACTIONS(3045),
    [anon_sym_wait] = ACTIONS(3045),
    [anon_sym_wantarray] = ACTIONS(3045),
    [anon_sym_abs] = ACTIONS(3045),
    [anon_sym_alarm] = ACTIONS(3045),
    [anon_sym_chop] = ACTIONS(3045),
    [anon_sym_chdir] = ACTIONS(3045),
    [anon_sym_close] = ACTIONS(3045),
    [anon_sym_closedir] = ACTIONS(3045),
    [anon_sym_caller] = ACTIONS(3045),
    [anon_sym_chomp] = ACTIONS(3045),
    [anon_sym_chr] = ACTIONS(3045),
    [anon_sym_cos] = ACTIONS(3045),
    [anon_sym_chroot] = ACTIONS(3045),
    [anon_sym_defined] = ACTIONS(3045),
    [anon_sym_delete] = ACTIONS(3045),
    [anon_sym_dbmclose] = ACTIONS(3045),
    [anon_sym_exists] = ACTIONS(3045),
    [anon_sym_exit] = ACTIONS(3045),
    [anon_sym_eof] = ACTIONS(3045),
    [anon_sym_exp] = ACTIONS(3045),
    [anon_sym_each] = ACTIONS(3045),
    [anon_sym_fc] = ACTIONS(3045),
    [anon_sym_fileno] = ACTIONS(3045),
    [anon_sym_gmtime] = ACTIONS(3045),
    [anon_sym_getc] = ACTIONS(3045),
    [anon_sym_getpgrp] = ACTIONS(3045),
    [anon_sym_getprotobyname] = ACTIONS(3045),
    [anon_sym_getpwname] = ACTIONS(3045),
    [anon_sym_getpwuid] = ACTIONS(3045),
    [anon_sym_getpeername] = ACTIONS(3045),
    [anon_sym_getnetbyname] = ACTIONS(3045),
    [anon_sym_getsockname] = ACTIONS(3045),
    [anon_sym_getgrnam] = ACTIONS(3045),
    [anon_sym_getgrgid] = ACTIONS(3045),
    [anon_sym_hex] = ACTIONS(3045),
    [anon_sym_int] = ACTIONS(3045),
    [anon_sym_keys] = ACTIONS(3045),
    [anon_sym_lc] = ACTIONS(3045),
    [anon_sym_lcfirst] = ACTIONS(3045),
    [anon_sym_length] = ACTIONS(3045),
    [anon_sym_localtime] = ACTIONS(3045),
    [anon_sym_log] = ACTIONS(3045),
    [anon_sym_lock] = ACTIONS(3045),
    [anon_sym_lstat] = ACTIONS(3045),
    [anon_sym_oct] = ACTIONS(3045),
    [anon_sym_ord] = ACTIONS(3045),
    [anon_sym_prototype] = ACTIONS(3045),
    [anon_sym_pop] = ACTIONS(3045),
    [anon_sym_pos] = ACTIONS(3045),
    [anon_sym_quotemeta] = ACTIONS(3045),
    [anon_sym_reset] = ACTIONS(3045),
    [anon_sym_rand] = ACTIONS(3045),
    [anon_sym_rmdir] = ACTIONS(3045),
    [anon_sym_readdir] = ACTIONS(3045),
    [anon_sym_readline] = ACTIONS(3045),
    [anon_sym_readpipe] = ACTIONS(3045),
    [anon_sym_rewinddir] = ACTIONS(3045),
    [anon_sym_readlink] = ACTIONS(3045),
    [anon_sym_ref] = ACTIONS(3045),
    [anon_sym_scalar] = ACTIONS(3045),
    [anon_sym_shift] = ACTIONS(3045),
    [anon_sym_sin] = ACTIONS(3045),
    [anon_sym_sleep] = ACTIONS(3045),
    [anon_sym_sqrt] = ACTIONS(3045),
    [anon_sym_srand] = ACTIONS(3045),
    [anon_sym_stat] = ACTIONS(3045),
    [anon_sym_study] = ACTIONS(3045),
    [anon_sym_tell] = ACTIONS(3045),
    [anon_sym_telldir] = ACTIONS(3045),
    [anon_sym_tied] = ACTIONS(3045),
    [anon_sym_uc] = ACTIONS(3045),
    [anon_sym_ucfirst] = ACTIONS(3045),
    [anon_sym_untie] = ACTIONS(3045),
    [anon_sym_umask] = ACTIONS(3045),
    [anon_sym_values] = ACTIONS(3045),
    [anon_sym_write] = ACTIONS(3045),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3045),
    [anon_sym_qq] = ACTIONS(3045),
    [anon_sym_qw] = ACTIONS(3045),
    [anon_sym_qx] = ACTIONS(3045),
    [anon_sym_qr] = ACTIONS(3045),
    [anon_sym_m] = ACTIONS(3045),
    [anon_sym_s] = ACTIONS(3045),
    [anon_sym_tr] = ACTIONS(3045),
    [anon_sym_y] = ACTIONS(3045),
    [aux_sym__bareword_token1] = ACTIONS(3045),
    [sym_number] = ACTIONS(3045),
    [anon_sym_true] = ACTIONS(3045),
    [anon_sym_false] = ACTIONS(3045),
    [sym__single_quote] = ACTIONS(3043),
    [sym__double_quote] = ACTIONS(3043),
    [sym__backtick_quote] = ACTIONS(3043),
    [sym__search_slash_quote] = ACTIONS(3043),
    [sym__open_readline_bracket] = ACTIONS(3043),
    [sym__open_fileglob_bracket] = ACTIONS(3043),
    [sym__PERLY_HEREDOC] = ACTIONS(3043),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3043),
    [sym__filetest] = ACTIONS(3043),
  },
  [STATE(1729)] = {
    [sym_heredoc_content] = STATE(1729),
    [sym__identifier] = ACTIONS(3241),
    [anon_sym_LBRACE] = ACTIONS(3241),
    [anon_sym_DOLLAR] = ACTIONS(3241),
    [anon_sym_AT] = ACTIONS(3241),
    [anon_sym_RPAREN] = ACTIONS(3241),
    [anon_sym_my] = ACTIONS(3241),
    [anon_sym_extended] = ACTIONS(3241),
    [anon_sym_async] = ACTIONS(3241),
    [anon_sym_sub] = ACTIONS(3241),
    [anon_sym_method] = ACTIONS(3241),
    [anon_sym_LPAREN] = ACTIONS(3241),
    [anon_sym_continue] = ACTIONS(3241),
    [anon_sym_state] = ACTIONS(3241),
    [anon_sym_our] = ACTIONS(3241),
    [anon_sym_LBRACK] = ACTIONS(3241),
    [anon_sym_LT_LT] = ACTIONS(3241),
    [anon_sym_SLASH_SLASH] = ACTIONS(3241),
    [anon_sym_PLUS] = ACTIONS(3241),
    [anon_sym_DASH] = ACTIONS(3241),
    [anon_sym_TILDE] = ACTIONS(3241),
    [anon_sym_BANG] = ACTIONS(3241),
    [anon_sym_PLUS_PLUS] = ACTIONS(3241),
    [anon_sym_DASH_DASH] = ACTIONS(3241),
    [anon_sym_BSLASH] = ACTIONS(3239),
    [anon_sym_do] = ACTIONS(3241),
    [anon_sym_eval] = ACTIONS(3241),
    [anon_sym_await] = ACTIONS(3241),
    [anon_sym_field] = ACTIONS(3241),
    [anon_sym_local] = ACTIONS(3241),
    [anon_sym_dynamically] = ACTIONS(3241),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3241),
    [anon_sym_require] = ACTIONS(3241),
    [anon_sym_map] = ACTIONS(3241),
    [anon_sym_grep] = ACTIONS(3241),
    [anon_sym_sort] = ACTIONS(3241),
    [anon_sym_goto] = ACTIONS(3241),
    [anon_sym_return] = ACTIONS(3241),
    [anon_sym_undef] = ACTIONS(3241),
    [anon_sym_PERCENT2] = ACTIONS(3241),
    [anon_sym_AMP2] = ACTIONS(3241),
    [anon_sym_STAR2] = ACTIONS(3241),
    [anon_sym_last] = ACTIONS(3241),
    [anon_sym_next] = ACTIONS(3241),
    [anon_sym_redo] = ACTIONS(3241),
    [anon_sym___FILE__] = ACTIONS(3241),
    [anon_sym___LINE__] = ACTIONS(3241),
    [anon_sym___PACKAGE__] = ACTIONS(3241),
    [anon_sym___SUB__] = ACTIONS(3241),
    [anon_sym_break] = ACTIONS(3241),
    [anon_sym_fork] = ACTIONS(3241),
    [anon_sym_getppid] = ACTIONS(3241),
    [anon_sym_time] = ACTIONS(3241),
    [anon_sym_times] = ACTIONS(3241),
    [anon_sym_wait] = ACTIONS(3241),
    [anon_sym_wantarray] = ACTIONS(3241),
    [anon_sym_abs] = ACTIONS(3241),
    [anon_sym_alarm] = ACTIONS(3241),
    [anon_sym_chop] = ACTIONS(3241),
    [anon_sym_chdir] = ACTIONS(3241),
    [anon_sym_close] = ACTIONS(3241),
    [anon_sym_closedir] = ACTIONS(3241),
    [anon_sym_caller] = ACTIONS(3241),
    [anon_sym_chomp] = ACTIONS(3241),
    [anon_sym_chr] = ACTIONS(3241),
    [anon_sym_cos] = ACTIONS(3241),
    [anon_sym_chroot] = ACTIONS(3241),
    [anon_sym_defined] = ACTIONS(3241),
    [anon_sym_delete] = ACTIONS(3241),
    [anon_sym_dbmclose] = ACTIONS(3241),
    [anon_sym_exists] = ACTIONS(3241),
    [anon_sym_exit] = ACTIONS(3241),
    [anon_sym_eof] = ACTIONS(3241),
    [anon_sym_exp] = ACTIONS(3241),
    [anon_sym_each] = ACTIONS(3241),
    [anon_sym_fc] = ACTIONS(3241),
    [anon_sym_fileno] = ACTIONS(3241),
    [anon_sym_gmtime] = ACTIONS(3241),
    [anon_sym_getc] = ACTIONS(3241),
    [anon_sym_getpgrp] = ACTIONS(3241),
    [anon_sym_getprotobyname] = ACTIONS(3241),
    [anon_sym_getpwname] = ACTIONS(3241),
    [anon_sym_getpwuid] = ACTIONS(3241),
    [anon_sym_getpeername] = ACTIONS(3241),
    [anon_sym_getnetbyname] = ACTIONS(3241),
    [anon_sym_getsockname] = ACTIONS(3241),
    [anon_sym_getgrnam] = ACTIONS(3241),
    [anon_sym_getgrgid] = ACTIONS(3241),
    [anon_sym_hex] = ACTIONS(3241),
    [anon_sym_int] = ACTIONS(3241),
    [anon_sym_keys] = ACTIONS(3241),
    [anon_sym_lc] = ACTIONS(3241),
    [anon_sym_lcfirst] = ACTIONS(3241),
    [anon_sym_length] = ACTIONS(3241),
    [anon_sym_localtime] = ACTIONS(3241),
    [anon_sym_log] = ACTIONS(3241),
    [anon_sym_lock] = ACTIONS(3241),
    [anon_sym_lstat] = ACTIONS(3241),
    [anon_sym_oct] = ACTIONS(3241),
    [anon_sym_ord] = ACTIONS(3241),
    [anon_sym_prototype] = ACTIONS(3241),
    [anon_sym_pop] = ACTIONS(3241),
    [anon_sym_pos] = ACTIONS(3241),
    [anon_sym_quotemeta] = ACTIONS(3241),
    [anon_sym_reset] = ACTIONS(3241),
    [anon_sym_rand] = ACTIONS(3241),
    [anon_sym_rmdir] = ACTIONS(3241),
    [anon_sym_readdir] = ACTIONS(3241),
    [anon_sym_readline] = ACTIONS(3241),
    [anon_sym_readpipe] = ACTIONS(3241),
    [anon_sym_rewinddir] = ACTIONS(3241),
    [anon_sym_readlink] = ACTIONS(3241),
    [anon_sym_ref] = ACTIONS(3241),
    [anon_sym_scalar] = ACTIONS(3241),
    [anon_sym_shift] = ACTIONS(3241),
    [anon_sym_sin] = ACTIONS(3241),
    [anon_sym_sleep] = ACTIONS(3241),
    [anon_sym_sqrt] = ACTIONS(3241),
    [anon_sym_srand] = ACTIONS(3241),
    [anon_sym_stat] = ACTIONS(3241),
    [anon_sym_study] = ACTIONS(3241),
    [anon_sym_tell] = ACTIONS(3241),
    [anon_sym_telldir] = ACTIONS(3241),
    [anon_sym_tied] = ACTIONS(3241),
    [anon_sym_uc] = ACTIONS(3241),
    [anon_sym_ucfirst] = ACTIONS(3241),
    [anon_sym_untie] = ACTIONS(3241),
    [anon_sym_umask] = ACTIONS(3241),
    [anon_sym_values] = ACTIONS(3241),
    [anon_sym_write] = ACTIONS(3241),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3241),
    [anon_sym_qq] = ACTIONS(3241),
    [anon_sym_qw] = ACTIONS(3241),
    [anon_sym_qx] = ACTIONS(3241),
    [anon_sym_qr] = ACTIONS(3241),
    [anon_sym_m] = ACTIONS(3241),
    [anon_sym_s] = ACTIONS(3241),
    [anon_sym_tr] = ACTIONS(3241),
    [anon_sym_y] = ACTIONS(3241),
    [aux_sym__bareword_token1] = ACTIONS(3241),
    [sym_number] = ACTIONS(3241),
    [anon_sym_true] = ACTIONS(3241),
    [anon_sym_false] = ACTIONS(3241),
    [sym__single_quote] = ACTIONS(3239),
    [sym__double_quote] = ACTIONS(3239),
    [sym__backtick_quote] = ACTIONS(3239),
    [sym__search_slash_quote] = ACTIONS(3239),
    [sym__open_readline_bracket] = ACTIONS(3239),
    [sym__open_fileglob_bracket] = ACTIONS(3239),
    [sym__PERLY_HEREDOC] = ACTIONS(3239),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3239),
    [sym__filetest] = ACTIONS(3239),
  },
  [STATE(1730)] = {
    [sym_heredoc_content] = STATE(1730),
    [sym__identifier] = ACTIONS(3073),
    [anon_sym_LBRACE] = ACTIONS(3073),
    [anon_sym_DOLLAR] = ACTIONS(3073),
    [anon_sym_AT] = ACTIONS(3073),
    [anon_sym_RPAREN] = ACTIONS(3073),
    [anon_sym_my] = ACTIONS(3073),
    [anon_sym_extended] = ACTIONS(3073),
    [anon_sym_async] = ACTIONS(3073),
    [anon_sym_sub] = ACTIONS(3073),
    [anon_sym_method] = ACTIONS(3073),
    [anon_sym_LPAREN] = ACTIONS(3073),
    [anon_sym_continue] = ACTIONS(3073),
    [anon_sym_state] = ACTIONS(3073),
    [anon_sym_our] = ACTIONS(3073),
    [anon_sym_LBRACK] = ACTIONS(3073),
    [anon_sym_LT_LT] = ACTIONS(3073),
    [anon_sym_SLASH_SLASH] = ACTIONS(3073),
    [anon_sym_PLUS] = ACTIONS(3073),
    [anon_sym_DASH] = ACTIONS(3073),
    [anon_sym_TILDE] = ACTIONS(3073),
    [anon_sym_BANG] = ACTIONS(3073),
    [anon_sym_PLUS_PLUS] = ACTIONS(3073),
    [anon_sym_DASH_DASH] = ACTIONS(3073),
    [anon_sym_BSLASH] = ACTIONS(3071),
    [anon_sym_do] = ACTIONS(3073),
    [anon_sym_eval] = ACTIONS(3073),
    [anon_sym_await] = ACTIONS(3073),
    [anon_sym_field] = ACTIONS(3073),
    [anon_sym_local] = ACTIONS(3073),
    [anon_sym_dynamically] = ACTIONS(3073),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3073),
    [anon_sym_require] = ACTIONS(3073),
    [anon_sym_map] = ACTIONS(3073),
    [anon_sym_grep] = ACTIONS(3073),
    [anon_sym_sort] = ACTIONS(3073),
    [anon_sym_goto] = ACTIONS(3073),
    [anon_sym_return] = ACTIONS(3073),
    [anon_sym_undef] = ACTIONS(3073),
    [anon_sym_PERCENT2] = ACTIONS(3073),
    [anon_sym_AMP2] = ACTIONS(3073),
    [anon_sym_STAR2] = ACTIONS(3073),
    [anon_sym_last] = ACTIONS(3073),
    [anon_sym_next] = ACTIONS(3073),
    [anon_sym_redo] = ACTIONS(3073),
    [anon_sym___FILE__] = ACTIONS(3073),
    [anon_sym___LINE__] = ACTIONS(3073),
    [anon_sym___PACKAGE__] = ACTIONS(3073),
    [anon_sym___SUB__] = ACTIONS(3073),
    [anon_sym_break] = ACTIONS(3073),
    [anon_sym_fork] = ACTIONS(3073),
    [anon_sym_getppid] = ACTIONS(3073),
    [anon_sym_time] = ACTIONS(3073),
    [anon_sym_times] = ACTIONS(3073),
    [anon_sym_wait] = ACTIONS(3073),
    [anon_sym_wantarray] = ACTIONS(3073),
    [anon_sym_abs] = ACTIONS(3073),
    [anon_sym_alarm] = ACTIONS(3073),
    [anon_sym_chop] = ACTIONS(3073),
    [anon_sym_chdir] = ACTIONS(3073),
    [anon_sym_close] = ACTIONS(3073),
    [anon_sym_closedir] = ACTIONS(3073),
    [anon_sym_caller] = ACTIONS(3073),
    [anon_sym_chomp] = ACTIONS(3073),
    [anon_sym_chr] = ACTIONS(3073),
    [anon_sym_cos] = ACTIONS(3073),
    [anon_sym_chroot] = ACTIONS(3073),
    [anon_sym_defined] = ACTIONS(3073),
    [anon_sym_delete] = ACTIONS(3073),
    [anon_sym_dbmclose] = ACTIONS(3073),
    [anon_sym_exists] = ACTIONS(3073),
    [anon_sym_exit] = ACTIONS(3073),
    [anon_sym_eof] = ACTIONS(3073),
    [anon_sym_exp] = ACTIONS(3073),
    [anon_sym_each] = ACTIONS(3073),
    [anon_sym_fc] = ACTIONS(3073),
    [anon_sym_fileno] = ACTIONS(3073),
    [anon_sym_gmtime] = ACTIONS(3073),
    [anon_sym_getc] = ACTIONS(3073),
    [anon_sym_getpgrp] = ACTIONS(3073),
    [anon_sym_getprotobyname] = ACTIONS(3073),
    [anon_sym_getpwname] = ACTIONS(3073),
    [anon_sym_getpwuid] = ACTIONS(3073),
    [anon_sym_getpeername] = ACTIONS(3073),
    [anon_sym_getnetbyname] = ACTIONS(3073),
    [anon_sym_getsockname] = ACTIONS(3073),
    [anon_sym_getgrnam] = ACTIONS(3073),
    [anon_sym_getgrgid] = ACTIONS(3073),
    [anon_sym_hex] = ACTIONS(3073),
    [anon_sym_int] = ACTIONS(3073),
    [anon_sym_keys] = ACTIONS(3073),
    [anon_sym_lc] = ACTIONS(3073),
    [anon_sym_lcfirst] = ACTIONS(3073),
    [anon_sym_length] = ACTIONS(3073),
    [anon_sym_localtime] = ACTIONS(3073),
    [anon_sym_log] = ACTIONS(3073),
    [anon_sym_lock] = ACTIONS(3073),
    [anon_sym_lstat] = ACTIONS(3073),
    [anon_sym_oct] = ACTIONS(3073),
    [anon_sym_ord] = ACTIONS(3073),
    [anon_sym_prototype] = ACTIONS(3073),
    [anon_sym_pop] = ACTIONS(3073),
    [anon_sym_pos] = ACTIONS(3073),
    [anon_sym_quotemeta] = ACTIONS(3073),
    [anon_sym_reset] = ACTIONS(3073),
    [anon_sym_rand] = ACTIONS(3073),
    [anon_sym_rmdir] = ACTIONS(3073),
    [anon_sym_readdir] = ACTIONS(3073),
    [anon_sym_readline] = ACTIONS(3073),
    [anon_sym_readpipe] = ACTIONS(3073),
    [anon_sym_rewinddir] = ACTIONS(3073),
    [anon_sym_readlink] = ACTIONS(3073),
    [anon_sym_ref] = ACTIONS(3073),
    [anon_sym_scalar] = ACTIONS(3073),
    [anon_sym_shift] = ACTIONS(3073),
    [anon_sym_sin] = ACTIONS(3073),
    [anon_sym_sleep] = ACTIONS(3073),
    [anon_sym_sqrt] = ACTIONS(3073),
    [anon_sym_srand] = ACTIONS(3073),
    [anon_sym_stat] = ACTIONS(3073),
    [anon_sym_study] = ACTIONS(3073),
    [anon_sym_tell] = ACTIONS(3073),
    [anon_sym_telldir] = ACTIONS(3073),
    [anon_sym_tied] = ACTIONS(3073),
    [anon_sym_uc] = ACTIONS(3073),
    [anon_sym_ucfirst] = ACTIONS(3073),
    [anon_sym_untie] = ACTIONS(3073),
    [anon_sym_umask] = ACTIONS(3073),
    [anon_sym_values] = ACTIONS(3073),
    [anon_sym_write] = ACTIONS(3073),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3073),
    [anon_sym_qq] = ACTIONS(3073),
    [anon_sym_qw] = ACTIONS(3073),
    [anon_sym_qx] = ACTIONS(3073),
    [anon_sym_qr] = ACTIONS(3073),
    [anon_sym_m] = ACTIONS(3073),
    [anon_sym_s] = ACTIONS(3073),
    [anon_sym_tr] = ACTIONS(3073),
    [anon_sym_y] = ACTIONS(3073),
    [aux_sym__bareword_token1] = ACTIONS(3073),
    [sym_number] = ACTIONS(3073),
    [anon_sym_true] = ACTIONS(3073),
    [anon_sym_false] = ACTIONS(3073),
    [sym__single_quote] = ACTIONS(3071),
    [sym__double_quote] = ACTIONS(3071),
    [sym__backtick_quote] = ACTIONS(3071),
    [sym__search_slash_quote] = ACTIONS(3071),
    [sym__open_readline_bracket] = ACTIONS(3071),
    [sym__open_fileglob_bracket] = ACTIONS(3071),
    [sym__PERLY_HEREDOC] = ACTIONS(3071),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3071),
    [sym__filetest] = ACTIONS(3071),
  },
  [STATE(1731)] = {
    [sym_heredoc_content] = STATE(1731),
    [sym__identifier] = ACTIONS(3289),
    [anon_sym_LBRACE] = ACTIONS(3289),
    [anon_sym_DOLLAR] = ACTIONS(3289),
    [anon_sym_AT] = ACTIONS(3289),
    [anon_sym_my] = ACTIONS(3289),
    [anon_sym_extended] = ACTIONS(3289),
    [anon_sym_async] = ACTIONS(3289),
    [anon_sym_sub] = ACTIONS(3289),
    [anon_sym_method] = ACTIONS(3289),
    [anon_sym_LPAREN] = ACTIONS(3289),
    [anon_sym_continue] = ACTIONS(3289),
    [anon_sym_state] = ACTIONS(3289),
    [anon_sym_our] = ACTIONS(3289),
    [anon_sym_LBRACK] = ACTIONS(3289),
    [anon_sym_LT_LT] = ACTIONS(3289),
    [anon_sym_SLASH_SLASH] = ACTIONS(3289),
    [anon_sym_PLUS] = ACTIONS(3289),
    [anon_sym_DASH] = ACTIONS(3289),
    [anon_sym_TILDE] = ACTIONS(3289),
    [anon_sym_BANG] = ACTIONS(3289),
    [anon_sym_PLUS_PLUS] = ACTIONS(3289),
    [anon_sym_DASH_DASH] = ACTIONS(3289),
    [anon_sym_BSLASH] = ACTIONS(3291),
    [anon_sym_do] = ACTIONS(3289),
    [anon_sym_eval] = ACTIONS(3289),
    [anon_sym_await] = ACTIONS(3289),
    [anon_sym_field] = ACTIONS(3289),
    [anon_sym_local] = ACTIONS(3289),
    [anon_sym_dynamically] = ACTIONS(3289),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3289),
    [anon_sym_require] = ACTIONS(3289),
    [anon_sym_map] = ACTIONS(3289),
    [anon_sym_grep] = ACTIONS(3289),
    [anon_sym_sort] = ACTIONS(3289),
    [anon_sym_goto] = ACTIONS(3289),
    [anon_sym_return] = ACTIONS(3289),
    [anon_sym_undef] = ACTIONS(3289),
    [anon_sym_PERCENT2] = ACTIONS(3289),
    [anon_sym_AMP2] = ACTIONS(3289),
    [anon_sym_STAR2] = ACTIONS(3289),
    [anon_sym_last] = ACTIONS(3289),
    [anon_sym_next] = ACTIONS(3289),
    [anon_sym_redo] = ACTIONS(3289),
    [anon_sym___FILE__] = ACTIONS(3289),
    [anon_sym___LINE__] = ACTIONS(3289),
    [anon_sym___PACKAGE__] = ACTIONS(3289),
    [anon_sym___SUB__] = ACTIONS(3289),
    [anon_sym_break] = ACTIONS(3289),
    [anon_sym_fork] = ACTIONS(3289),
    [anon_sym_getppid] = ACTIONS(3289),
    [anon_sym_time] = ACTIONS(3289),
    [anon_sym_times] = ACTIONS(3289),
    [anon_sym_wait] = ACTIONS(3289),
    [anon_sym_wantarray] = ACTIONS(3289),
    [anon_sym_abs] = ACTIONS(3289),
    [anon_sym_alarm] = ACTIONS(3289),
    [anon_sym_chop] = ACTIONS(3289),
    [anon_sym_chdir] = ACTIONS(3289),
    [anon_sym_close] = ACTIONS(3289),
    [anon_sym_closedir] = ACTIONS(3289),
    [anon_sym_caller] = ACTIONS(3289),
    [anon_sym_chomp] = ACTIONS(3289),
    [anon_sym_chr] = ACTIONS(3289),
    [anon_sym_cos] = ACTIONS(3289),
    [anon_sym_chroot] = ACTIONS(3289),
    [anon_sym_defined] = ACTIONS(3289),
    [anon_sym_delete] = ACTIONS(3289),
    [anon_sym_dbmclose] = ACTIONS(3289),
    [anon_sym_exists] = ACTIONS(3289),
    [anon_sym_exit] = ACTIONS(3289),
    [anon_sym_eof] = ACTIONS(3289),
    [anon_sym_exp] = ACTIONS(3289),
    [anon_sym_each] = ACTIONS(3289),
    [anon_sym_fc] = ACTIONS(3289),
    [anon_sym_fileno] = ACTIONS(3289),
    [anon_sym_gmtime] = ACTIONS(3289),
    [anon_sym_getc] = ACTIONS(3289),
    [anon_sym_getpgrp] = ACTIONS(3289),
    [anon_sym_getprotobyname] = ACTIONS(3289),
    [anon_sym_getpwname] = ACTIONS(3289),
    [anon_sym_getpwuid] = ACTIONS(3289),
    [anon_sym_getpeername] = ACTIONS(3289),
    [anon_sym_getnetbyname] = ACTIONS(3289),
    [anon_sym_getsockname] = ACTIONS(3289),
    [anon_sym_getgrnam] = ACTIONS(3289),
    [anon_sym_getgrgid] = ACTIONS(3289),
    [anon_sym_hex] = ACTIONS(3289),
    [anon_sym_int] = ACTIONS(3289),
    [anon_sym_keys] = ACTIONS(3289),
    [anon_sym_lc] = ACTIONS(3289),
    [anon_sym_lcfirst] = ACTIONS(3289),
    [anon_sym_length] = ACTIONS(3289),
    [anon_sym_localtime] = ACTIONS(3289),
    [anon_sym_log] = ACTIONS(3289),
    [anon_sym_lock] = ACTIONS(3289),
    [anon_sym_lstat] = ACTIONS(3289),
    [anon_sym_oct] = ACTIONS(3289),
    [anon_sym_ord] = ACTIONS(3289),
    [anon_sym_prototype] = ACTIONS(3289),
    [anon_sym_pop] = ACTIONS(3289),
    [anon_sym_pos] = ACTIONS(3289),
    [anon_sym_quotemeta] = ACTIONS(3289),
    [anon_sym_reset] = ACTIONS(3289),
    [anon_sym_rand] = ACTIONS(3289),
    [anon_sym_rmdir] = ACTIONS(3289),
    [anon_sym_readdir] = ACTIONS(3289),
    [anon_sym_readline] = ACTIONS(3289),
    [anon_sym_readpipe] = ACTIONS(3289),
    [anon_sym_rewinddir] = ACTIONS(3289),
    [anon_sym_readlink] = ACTIONS(3289),
    [anon_sym_ref] = ACTIONS(3289),
    [anon_sym_scalar] = ACTIONS(3289),
    [anon_sym_shift] = ACTIONS(3289),
    [anon_sym_sin] = ACTIONS(3289),
    [anon_sym_sleep] = ACTIONS(3289),
    [anon_sym_sqrt] = ACTIONS(3289),
    [anon_sym_srand] = ACTIONS(3289),
    [anon_sym_stat] = ACTIONS(3289),
    [anon_sym_study] = ACTIONS(3289),
    [anon_sym_tell] = ACTIONS(3289),
    [anon_sym_telldir] = ACTIONS(3289),
    [anon_sym_tied] = ACTIONS(3289),
    [anon_sym_uc] = ACTIONS(3289),
    [anon_sym_ucfirst] = ACTIONS(3289),
    [anon_sym_untie] = ACTIONS(3289),
    [anon_sym_umask] = ACTIONS(3289),
    [anon_sym_values] = ACTIONS(3289),
    [anon_sym_write] = ACTIONS(3289),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3289),
    [anon_sym_qq] = ACTIONS(3289),
    [anon_sym_qw] = ACTIONS(3289),
    [anon_sym_qx] = ACTIONS(3289),
    [anon_sym_qr] = ACTIONS(3289),
    [anon_sym_m] = ACTIONS(3289),
    [anon_sym_s] = ACTIONS(3289),
    [anon_sym_tr] = ACTIONS(3289),
    [anon_sym_y] = ACTIONS(3289),
    [aux_sym__bareword_token1] = ACTIONS(3289),
    [sym_number] = ACTIONS(3289),
    [anon_sym_true] = ACTIONS(3289),
    [anon_sym_false] = ACTIONS(3289),
    [sym__single_quote] = ACTIONS(3291),
    [sym__double_quote] = ACTIONS(3291),
    [sym__backtick_quote] = ACTIONS(3291),
    [sym__search_slash_quote] = ACTIONS(3291),
    [sym__open_readline_bracket] = ACTIONS(3291),
    [sym__open_fileglob_bracket] = ACTIONS(3291),
    [sym__PERLY_HEREDOC] = ACTIONS(3291),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3291),
    [sym__filetest] = ACTIONS(3291),
  },
  [STATE(1732)] = {
    [sym_heredoc_content] = STATE(1732),
    [sym__identifier] = ACTIONS(2700),
    [anon_sym_LBRACE] = ACTIONS(2700),
    [anon_sym_DOLLAR] = ACTIONS(2700),
    [anon_sym_AT] = ACTIONS(2700),
    [anon_sym_my] = ACTIONS(2700),
    [anon_sym_extended] = ACTIONS(2700),
    [anon_sym_async] = ACTIONS(2700),
    [anon_sym_sub] = ACTIONS(2700),
    [anon_sym_method] = ACTIONS(2700),
    [anon_sym_LPAREN] = ACTIONS(2700),
    [anon_sym_continue] = ACTIONS(2700),
    [anon_sym_state] = ACTIONS(2700),
    [anon_sym_our] = ACTIONS(2700),
    [anon_sym_LBRACK] = ACTIONS(2700),
    [anon_sym_LT_LT] = ACTIONS(2700),
    [anon_sym_SLASH_SLASH] = ACTIONS(2700),
    [anon_sym_PLUS] = ACTIONS(2700),
    [anon_sym_DASH] = ACTIONS(2700),
    [anon_sym_TILDE] = ACTIONS(2700),
    [anon_sym_BANG] = ACTIONS(2700),
    [anon_sym_PLUS_PLUS] = ACTIONS(2700),
    [anon_sym_DASH_DASH] = ACTIONS(2700),
    [anon_sym_BSLASH] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2700),
    [anon_sym_eval] = ACTIONS(2700),
    [anon_sym_await] = ACTIONS(2700),
    [anon_sym_field] = ACTIONS(2700),
    [anon_sym_local] = ACTIONS(2700),
    [anon_sym_dynamically] = ACTIONS(2700),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2700),
    [anon_sym_require] = ACTIONS(2700),
    [anon_sym_map] = ACTIONS(2700),
    [anon_sym_grep] = ACTIONS(2700),
    [anon_sym_sort] = ACTIONS(2700),
    [anon_sym_goto] = ACTIONS(2700),
    [anon_sym_return] = ACTIONS(2700),
    [anon_sym_undef] = ACTIONS(2700),
    [anon_sym_PERCENT2] = ACTIONS(2700),
    [anon_sym_AMP2] = ACTIONS(2700),
    [anon_sym_STAR2] = ACTIONS(2700),
    [anon_sym_last] = ACTIONS(2700),
    [anon_sym_next] = ACTIONS(2700),
    [anon_sym_redo] = ACTIONS(2700),
    [anon_sym___FILE__] = ACTIONS(2700),
    [anon_sym___LINE__] = ACTIONS(2700),
    [anon_sym___PACKAGE__] = ACTIONS(2700),
    [anon_sym___SUB__] = ACTIONS(2700),
    [anon_sym_break] = ACTIONS(2700),
    [anon_sym_fork] = ACTIONS(2700),
    [anon_sym_getppid] = ACTIONS(2700),
    [anon_sym_time] = ACTIONS(2700),
    [anon_sym_times] = ACTIONS(2700),
    [anon_sym_wait] = ACTIONS(2700),
    [anon_sym_wantarray] = ACTIONS(2700),
    [anon_sym_abs] = ACTIONS(2700),
    [anon_sym_alarm] = ACTIONS(2700),
    [anon_sym_chop] = ACTIONS(2700),
    [anon_sym_chdir] = ACTIONS(2700),
    [anon_sym_close] = ACTIONS(2700),
    [anon_sym_closedir] = ACTIONS(2700),
    [anon_sym_caller] = ACTIONS(2700),
    [anon_sym_chomp] = ACTIONS(2700),
    [anon_sym_chr] = ACTIONS(2700),
    [anon_sym_cos] = ACTIONS(2700),
    [anon_sym_chroot] = ACTIONS(2700),
    [anon_sym_defined] = ACTIONS(2700),
    [anon_sym_delete] = ACTIONS(2700),
    [anon_sym_dbmclose] = ACTIONS(2700),
    [anon_sym_exists] = ACTIONS(2700),
    [anon_sym_exit] = ACTIONS(2700),
    [anon_sym_eof] = ACTIONS(2700),
    [anon_sym_exp] = ACTIONS(2700),
    [anon_sym_each] = ACTIONS(2700),
    [anon_sym_fc] = ACTIONS(2700),
    [anon_sym_fileno] = ACTIONS(2700),
    [anon_sym_gmtime] = ACTIONS(2700),
    [anon_sym_getc] = ACTIONS(2700),
    [anon_sym_getpgrp] = ACTIONS(2700),
    [anon_sym_getprotobyname] = ACTIONS(2700),
    [anon_sym_getpwname] = ACTIONS(2700),
    [anon_sym_getpwuid] = ACTIONS(2700),
    [anon_sym_getpeername] = ACTIONS(2700),
    [anon_sym_getnetbyname] = ACTIONS(2700),
    [anon_sym_getsockname] = ACTIONS(2700),
    [anon_sym_getgrnam] = ACTIONS(2700),
    [anon_sym_getgrgid] = ACTIONS(2700),
    [anon_sym_hex] = ACTIONS(2700),
    [anon_sym_int] = ACTIONS(2700),
    [anon_sym_keys] = ACTIONS(2700),
    [anon_sym_lc] = ACTIONS(2700),
    [anon_sym_lcfirst] = ACTIONS(2700),
    [anon_sym_length] = ACTIONS(2700),
    [anon_sym_localtime] = ACTIONS(2700),
    [anon_sym_log] = ACTIONS(2700),
    [anon_sym_lock] = ACTIONS(2700),
    [anon_sym_lstat] = ACTIONS(2700),
    [anon_sym_oct] = ACTIONS(2700),
    [anon_sym_ord] = ACTIONS(2700),
    [anon_sym_prototype] = ACTIONS(2700),
    [anon_sym_pop] = ACTIONS(2700),
    [anon_sym_pos] = ACTIONS(2700),
    [anon_sym_quotemeta] = ACTIONS(2700),
    [anon_sym_reset] = ACTIONS(2700),
    [anon_sym_rand] = ACTIONS(2700),
    [anon_sym_rmdir] = ACTIONS(2700),
    [anon_sym_readdir] = ACTIONS(2700),
    [anon_sym_readline] = ACTIONS(2700),
    [anon_sym_readpipe] = ACTIONS(2700),
    [anon_sym_rewinddir] = ACTIONS(2700),
    [anon_sym_readlink] = ACTIONS(2700),
    [anon_sym_ref] = ACTIONS(2700),
    [anon_sym_scalar] = ACTIONS(2700),
    [anon_sym_shift] = ACTIONS(2700),
    [anon_sym_sin] = ACTIONS(2700),
    [anon_sym_sleep] = ACTIONS(2700),
    [anon_sym_sqrt] = ACTIONS(2700),
    [anon_sym_srand] = ACTIONS(2700),
    [anon_sym_stat] = ACTIONS(2700),
    [anon_sym_study] = ACTIONS(2700),
    [anon_sym_tell] = ACTIONS(2700),
    [anon_sym_telldir] = ACTIONS(2700),
    [anon_sym_tied] = ACTIONS(2700),
    [anon_sym_uc] = ACTIONS(2700),
    [anon_sym_ucfirst] = ACTIONS(2700),
    [anon_sym_untie] = ACTIONS(2700),
    [anon_sym_umask] = ACTIONS(2700),
    [anon_sym_values] = ACTIONS(2700),
    [anon_sym_write] = ACTIONS(2700),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2700),
    [anon_sym_qq] = ACTIONS(2700),
    [anon_sym_qw] = ACTIONS(2700),
    [anon_sym_qx] = ACTIONS(2700),
    [anon_sym_qr] = ACTIONS(2700),
    [anon_sym_m] = ACTIONS(2700),
    [anon_sym_s] = ACTIONS(2700),
    [anon_sym_tr] = ACTIONS(2700),
    [anon_sym_y] = ACTIONS(2700),
    [aux_sym__bareword_token1] = ACTIONS(2700),
    [sym_number] = ACTIONS(2700),
    [anon_sym_true] = ACTIONS(2700),
    [anon_sym_false] = ACTIONS(2700),
    [sym__single_quote] = ACTIONS(2707),
    [sym__double_quote] = ACTIONS(2707),
    [sym__backtick_quote] = ACTIONS(2707),
    [sym__search_slash_quote] = ACTIONS(2707),
    [sym__open_readline_bracket] = ACTIONS(2707),
    [sym__open_fileglob_bracket] = ACTIONS(2707),
    [sym__PERLY_HEREDOC] = ACTIONS(2707),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2707),
    [sym__filetest] = ACTIONS(2707),
  },
  [STATE(1733)] = {
    [sym_heredoc_content] = STATE(1733),
    [sym__identifier] = ACTIONS(2769),
    [anon_sym_LBRACE] = ACTIONS(2769),
    [anon_sym_DOLLAR] = ACTIONS(2769),
    [anon_sym_AT] = ACTIONS(2769),
    [anon_sym_my] = ACTIONS(2769),
    [anon_sym_extended] = ACTIONS(2769),
    [anon_sym_async] = ACTIONS(2769),
    [anon_sym_sub] = ACTIONS(2769),
    [anon_sym_method] = ACTIONS(2769),
    [anon_sym_LPAREN] = ACTIONS(2769),
    [anon_sym_continue] = ACTIONS(2769),
    [anon_sym_state] = ACTIONS(2769),
    [anon_sym_our] = ACTIONS(2769),
    [anon_sym_LBRACK] = ACTIONS(2769),
    [anon_sym_LT_LT] = ACTIONS(2769),
    [anon_sym_SLASH_SLASH] = ACTIONS(2769),
    [anon_sym_PLUS] = ACTIONS(2769),
    [anon_sym_DASH] = ACTIONS(2769),
    [anon_sym_TILDE] = ACTIONS(2769),
    [anon_sym_BANG] = ACTIONS(2769),
    [anon_sym_PLUS_PLUS] = ACTIONS(2769),
    [anon_sym_DASH_DASH] = ACTIONS(2769),
    [anon_sym_BSLASH] = ACTIONS(2771),
    [anon_sym_do] = ACTIONS(2769),
    [anon_sym_eval] = ACTIONS(2769),
    [anon_sym_await] = ACTIONS(2769),
    [anon_sym_field] = ACTIONS(2769),
    [anon_sym_local] = ACTIONS(2769),
    [anon_sym_dynamically] = ACTIONS(2769),
    [anon_sym_DOLLAR_POUND] = ACTIONS(2769),
    [anon_sym_require] = ACTIONS(2769),
    [anon_sym_map] = ACTIONS(2769),
    [anon_sym_grep] = ACTIONS(2769),
    [anon_sym_sort] = ACTIONS(2769),
    [anon_sym_goto] = ACTIONS(2769),
    [anon_sym_return] = ACTIONS(2769),
    [anon_sym_undef] = ACTIONS(2769),
    [anon_sym_PERCENT2] = ACTIONS(2769),
    [anon_sym_AMP2] = ACTIONS(2769),
    [anon_sym_STAR2] = ACTIONS(2769),
    [anon_sym_last] = ACTIONS(2769),
    [anon_sym_next] = ACTIONS(2769),
    [anon_sym_redo] = ACTIONS(2769),
    [anon_sym___FILE__] = ACTIONS(2769),
    [anon_sym___LINE__] = ACTIONS(2769),
    [anon_sym___PACKAGE__] = ACTIONS(2769),
    [anon_sym___SUB__] = ACTIONS(2769),
    [anon_sym_break] = ACTIONS(2769),
    [anon_sym_fork] = ACTIONS(2769),
    [anon_sym_getppid] = ACTIONS(2769),
    [anon_sym_time] = ACTIONS(2769),
    [anon_sym_times] = ACTIONS(2769),
    [anon_sym_wait] = ACTIONS(2769),
    [anon_sym_wantarray] = ACTIONS(2769),
    [anon_sym_abs] = ACTIONS(2769),
    [anon_sym_alarm] = ACTIONS(2769),
    [anon_sym_chop] = ACTIONS(2769),
    [anon_sym_chdir] = ACTIONS(2769),
    [anon_sym_close] = ACTIONS(2769),
    [anon_sym_closedir] = ACTIONS(2769),
    [anon_sym_caller] = ACTIONS(2769),
    [anon_sym_chomp] = ACTIONS(2769),
    [anon_sym_chr] = ACTIONS(2769),
    [anon_sym_cos] = ACTIONS(2769),
    [anon_sym_chroot] = ACTIONS(2769),
    [anon_sym_defined] = ACTIONS(2769),
    [anon_sym_delete] = ACTIONS(2769),
    [anon_sym_dbmclose] = ACTIONS(2769),
    [anon_sym_exists] = ACTIONS(2769),
    [anon_sym_exit] = ACTIONS(2769),
    [anon_sym_eof] = ACTIONS(2769),
    [anon_sym_exp] = ACTIONS(2769),
    [anon_sym_each] = ACTIONS(2769),
    [anon_sym_fc] = ACTIONS(2769),
    [anon_sym_fileno] = ACTIONS(2769),
    [anon_sym_gmtime] = ACTIONS(2769),
    [anon_sym_getc] = ACTIONS(2769),
    [anon_sym_getpgrp] = ACTIONS(2769),
    [anon_sym_getprotobyname] = ACTIONS(2769),
    [anon_sym_getpwname] = ACTIONS(2769),
    [anon_sym_getpwuid] = ACTIONS(2769),
    [anon_sym_getpeername] = ACTIONS(2769),
    [anon_sym_getnetbyname] = ACTIONS(2769),
    [anon_sym_getsockname] = ACTIONS(2769),
    [anon_sym_getgrnam] = ACTIONS(2769),
    [anon_sym_getgrgid] = ACTIONS(2769),
    [anon_sym_hex] = ACTIONS(2769),
    [anon_sym_int] = ACTIONS(2769),
    [anon_sym_keys] = ACTIONS(2769),
    [anon_sym_lc] = ACTIONS(2769),
    [anon_sym_lcfirst] = ACTIONS(2769),
    [anon_sym_length] = ACTIONS(2769),
    [anon_sym_localtime] = ACTIONS(2769),
    [anon_sym_log] = ACTIONS(2769),
    [anon_sym_lock] = ACTIONS(2769),
    [anon_sym_lstat] = ACTIONS(2769),
    [anon_sym_oct] = ACTIONS(2769),
    [anon_sym_ord] = ACTIONS(2769),
    [anon_sym_prototype] = ACTIONS(2769),
    [anon_sym_pop] = ACTIONS(2769),
    [anon_sym_pos] = ACTIONS(2769),
    [anon_sym_quotemeta] = ACTIONS(2769),
    [anon_sym_reset] = ACTIONS(2769),
    [anon_sym_rand] = ACTIONS(2769),
    [anon_sym_rmdir] = ACTIONS(2769),
    [anon_sym_readdir] = ACTIONS(2769),
    [anon_sym_readline] = ACTIONS(2769),
    [anon_sym_readpipe] = ACTIONS(2769),
    [anon_sym_rewinddir] = ACTIONS(2769),
    [anon_sym_readlink] = ACTIONS(2769),
    [anon_sym_ref] = ACTIONS(2769),
    [anon_sym_scalar] = ACTIONS(2769),
    [anon_sym_shift] = ACTIONS(2769),
    [anon_sym_sin] = ACTIONS(2769),
    [anon_sym_sleep] = ACTIONS(2769),
    [anon_sym_sqrt] = ACTIONS(2769),
    [anon_sym_srand] = ACTIONS(2769),
    [anon_sym_stat] = ACTIONS(2769),
    [anon_sym_study] = ACTIONS(2769),
    [anon_sym_tell] = ACTIONS(2769),
    [anon_sym_telldir] = ACTIONS(2769),
    [anon_sym_tied] = ACTIONS(2769),
    [anon_sym_uc] = ACTIONS(2769),
    [anon_sym_ucfirst] = ACTIONS(2769),
    [anon_sym_untie] = ACTIONS(2769),
    [anon_sym_umask] = ACTIONS(2769),
    [anon_sym_values] = ACTIONS(2769),
    [anon_sym_write] = ACTIONS(2769),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(2769),
    [anon_sym_qq] = ACTIONS(2769),
    [anon_sym_qw] = ACTIONS(2769),
    [anon_sym_qx] = ACTIONS(2769),
    [anon_sym_qr] = ACTIONS(2769),
    [anon_sym_m] = ACTIONS(2769),
    [anon_sym_s] = ACTIONS(2769),
    [anon_sym_tr] = ACTIONS(2769),
    [anon_sym_y] = ACTIONS(2769),
    [aux_sym__bareword_token1] = ACTIONS(2769),
    [sym_number] = ACTIONS(2769),
    [anon_sym_true] = ACTIONS(2769),
    [anon_sym_false] = ACTIONS(2769),
    [sym__single_quote] = ACTIONS(2771),
    [sym__double_quote] = ACTIONS(2771),
    [sym__backtick_quote] = ACTIONS(2771),
    [sym__search_slash_quote] = ACTIONS(2771),
    [sym__open_readline_bracket] = ACTIONS(2771),
    [sym__open_fileglob_bracket] = ACTIONS(2771),
    [sym__PERLY_HEREDOC] = ACTIONS(2771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(2771),
    [sym__filetest] = ACTIONS(2771),
  },
  [STATE(1734)] = {
    [sym_heredoc_content] = STATE(1734),
    [sym__identifier] = ACTIONS(3293),
    [anon_sym_LBRACE] = ACTIONS(3293),
    [anon_sym_DOLLAR] = ACTIONS(3293),
    [anon_sym_AT] = ACTIONS(3293),
    [anon_sym_my] = ACTIONS(3293),
    [anon_sym_extended] = ACTIONS(3293),
    [anon_sym_async] = ACTIONS(3293),
    [anon_sym_sub] = ACTIONS(3293),
    [anon_sym_method] = ACTIONS(3293),
    [anon_sym_LPAREN] = ACTIONS(3293),
    [anon_sym_continue] = ACTIONS(3293),
    [anon_sym_state] = ACTIONS(3293),
    [anon_sym_our] = ACTIONS(3293),
    [anon_sym_LBRACK] = ACTIONS(3293),
    [anon_sym_LT_LT] = ACTIONS(3293),
    [anon_sym_SLASH_SLASH] = ACTIONS(3293),
    [anon_sym_PLUS] = ACTIONS(3293),
    [anon_sym_DASH] = ACTIONS(3293),
    [anon_sym_TILDE] = ACTIONS(3293),
    [anon_sym_BANG] = ACTIONS(3293),
    [anon_sym_PLUS_PLUS] = ACTIONS(3293),
    [anon_sym_DASH_DASH] = ACTIONS(3293),
    [anon_sym_BSLASH] = ACTIONS(3295),
    [anon_sym_do] = ACTIONS(3293),
    [anon_sym_eval] = ACTIONS(3293),
    [anon_sym_await] = ACTIONS(3293),
    [anon_sym_field] = ACTIONS(3293),
    [anon_sym_local] = ACTIONS(3293),
    [anon_sym_dynamically] = ACTIONS(3293),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3293),
    [anon_sym_require] = ACTIONS(3293),
    [anon_sym_map] = ACTIONS(3293),
    [anon_sym_grep] = ACTIONS(3293),
    [anon_sym_sort] = ACTIONS(3293),
    [anon_sym_goto] = ACTIONS(3293),
    [anon_sym_return] = ACTIONS(3293),
    [anon_sym_undef] = ACTIONS(3293),
    [anon_sym_PERCENT2] = ACTIONS(3293),
    [anon_sym_AMP2] = ACTIONS(3293),
    [anon_sym_STAR2] = ACTIONS(3293),
    [anon_sym_last] = ACTIONS(3293),
    [anon_sym_next] = ACTIONS(3293),
    [anon_sym_redo] = ACTIONS(3293),
    [anon_sym___FILE__] = ACTIONS(3293),
    [anon_sym___LINE__] = ACTIONS(3293),
    [anon_sym___PACKAGE__] = ACTIONS(3293),
    [anon_sym___SUB__] = ACTIONS(3293),
    [anon_sym_break] = ACTIONS(3293),
    [anon_sym_fork] = ACTIONS(3293),
    [anon_sym_getppid] = ACTIONS(3293),
    [anon_sym_time] = ACTIONS(3293),
    [anon_sym_times] = ACTIONS(3293),
    [anon_sym_wait] = ACTIONS(3293),
    [anon_sym_wantarray] = ACTIONS(3293),
    [anon_sym_abs] = ACTIONS(3293),
    [anon_sym_alarm] = ACTIONS(3293),
    [anon_sym_chop] = ACTIONS(3293),
    [anon_sym_chdir] = ACTIONS(3293),
    [anon_sym_close] = ACTIONS(3293),
    [anon_sym_closedir] = ACTIONS(3293),
    [anon_sym_caller] = ACTIONS(3293),
    [anon_sym_chomp] = ACTIONS(3293),
    [anon_sym_chr] = ACTIONS(3293),
    [anon_sym_cos] = ACTIONS(3293),
    [anon_sym_chroot] = ACTIONS(3293),
    [anon_sym_defined] = ACTIONS(3293),
    [anon_sym_delete] = ACTIONS(3293),
    [anon_sym_dbmclose] = ACTIONS(3293),
    [anon_sym_exists] = ACTIONS(3293),
    [anon_sym_exit] = ACTIONS(3293),
    [anon_sym_eof] = ACTIONS(3293),
    [anon_sym_exp] = ACTIONS(3293),
    [anon_sym_each] = ACTIONS(3293),
    [anon_sym_fc] = ACTIONS(3293),
    [anon_sym_fileno] = ACTIONS(3293),
    [anon_sym_gmtime] = ACTIONS(3293),
    [anon_sym_getc] = ACTIONS(3293),
    [anon_sym_getpgrp] = ACTIONS(3293),
    [anon_sym_getprotobyname] = ACTIONS(3293),
    [anon_sym_getpwname] = ACTIONS(3293),
    [anon_sym_getpwuid] = ACTIONS(3293),
    [anon_sym_getpeername] = ACTIONS(3293),
    [anon_sym_getnetbyname] = ACTIONS(3293),
    [anon_sym_getsockname] = ACTIONS(3293),
    [anon_sym_getgrnam] = ACTIONS(3293),
    [anon_sym_getgrgid] = ACTIONS(3293),
    [anon_sym_hex] = ACTIONS(3293),
    [anon_sym_int] = ACTIONS(3293),
    [anon_sym_keys] = ACTIONS(3293),
    [anon_sym_lc] = ACTIONS(3293),
    [anon_sym_lcfirst] = ACTIONS(3293),
    [anon_sym_length] = ACTIONS(3293),
    [anon_sym_localtime] = ACTIONS(3293),
    [anon_sym_log] = ACTIONS(3293),
    [anon_sym_lock] = ACTIONS(3293),
    [anon_sym_lstat] = ACTIONS(3293),
    [anon_sym_oct] = ACTIONS(3293),
    [anon_sym_ord] = ACTIONS(3293),
    [anon_sym_prototype] = ACTIONS(3293),
    [anon_sym_pop] = ACTIONS(3293),
    [anon_sym_pos] = ACTIONS(3293),
    [anon_sym_quotemeta] = ACTIONS(3293),
    [anon_sym_reset] = ACTIONS(3293),
    [anon_sym_rand] = ACTIONS(3293),
    [anon_sym_rmdir] = ACTIONS(3293),
    [anon_sym_readdir] = ACTIONS(3293),
    [anon_sym_readline] = ACTIONS(3293),
    [anon_sym_readpipe] = ACTIONS(3293),
    [anon_sym_rewinddir] = ACTIONS(3293),
    [anon_sym_readlink] = ACTIONS(3293),
    [anon_sym_ref] = ACTIONS(3293),
    [anon_sym_scalar] = ACTIONS(3293),
    [anon_sym_shift] = ACTIONS(3293),
    [anon_sym_sin] = ACTIONS(3293),
    [anon_sym_sleep] = ACTIONS(3293),
    [anon_sym_sqrt] = ACTIONS(3293),
    [anon_sym_srand] = ACTIONS(3293),
    [anon_sym_stat] = ACTIONS(3293),
    [anon_sym_study] = ACTIONS(3293),
    [anon_sym_tell] = ACTIONS(3293),
    [anon_sym_telldir] = ACTIONS(3293),
    [anon_sym_tied] = ACTIONS(3293),
    [anon_sym_uc] = ACTIONS(3293),
    [anon_sym_ucfirst] = ACTIONS(3293),
    [anon_sym_untie] = ACTIONS(3293),
    [anon_sym_umask] = ACTIONS(3293),
    [anon_sym_values] = ACTIONS(3293),
    [anon_sym_write] = ACTIONS(3293),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3293),
    [anon_sym_qq] = ACTIONS(3293),
    [anon_sym_qw] = ACTIONS(3293),
    [anon_sym_qx] = ACTIONS(3293),
    [anon_sym_qr] = ACTIONS(3293),
    [anon_sym_m] = ACTIONS(3293),
    [anon_sym_s] = ACTIONS(3293),
    [anon_sym_tr] = ACTIONS(3293),
    [anon_sym_y] = ACTIONS(3293),
    [aux_sym__bareword_token1] = ACTIONS(3293),
    [sym_number] = ACTIONS(3293),
    [anon_sym_true] = ACTIONS(3293),
    [anon_sym_false] = ACTIONS(3293),
    [sym__single_quote] = ACTIONS(3295),
    [sym__double_quote] = ACTIONS(3295),
    [sym__backtick_quote] = ACTIONS(3295),
    [sym__search_slash_quote] = ACTIONS(3295),
    [sym__open_readline_bracket] = ACTIONS(3295),
    [sym__open_fileglob_bracket] = ACTIONS(3295),
    [sym__PERLY_HEREDOC] = ACTIONS(3295),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3295),
    [sym__filetest] = ACTIONS(3295),
  },
  [STATE(1735)] = {
    [sym_heredoc_content] = STATE(1735),
    [sym__identifier] = ACTIONS(3297),
    [anon_sym_LBRACE] = ACTIONS(3297),
    [anon_sym_DOLLAR] = ACTIONS(3297),
    [anon_sym_AT] = ACTIONS(3297),
    [anon_sym_my] = ACTIONS(3297),
    [anon_sym_extended] = ACTIONS(3297),
    [anon_sym_async] = ACTIONS(3297),
    [anon_sym_sub] = ACTIONS(3297),
    [anon_sym_method] = ACTIONS(3297),
    [anon_sym_LPAREN] = ACTIONS(3297),
    [anon_sym_continue] = ACTIONS(3297),
    [anon_sym_state] = ACTIONS(3297),
    [anon_sym_our] = ACTIONS(3297),
    [anon_sym_LBRACK] = ACTIONS(3297),
    [anon_sym_LT_LT] = ACTIONS(3297),
    [anon_sym_SLASH_SLASH] = ACTIONS(3297),
    [anon_sym_PLUS] = ACTIONS(3297),
    [anon_sym_DASH] = ACTIONS(3297),
    [anon_sym_TILDE] = ACTIONS(3297),
    [anon_sym_BANG] = ACTIONS(3297),
    [anon_sym_PLUS_PLUS] = ACTIONS(3297),
    [anon_sym_DASH_DASH] = ACTIONS(3297),
    [anon_sym_BSLASH] = ACTIONS(3299),
    [anon_sym_do] = ACTIONS(3297),
    [anon_sym_eval] = ACTIONS(3297),
    [anon_sym_await] = ACTIONS(3297),
    [anon_sym_field] = ACTIONS(3297),
    [anon_sym_local] = ACTIONS(3297),
    [anon_sym_dynamically] = ACTIONS(3297),
    [anon_sym_DOLLAR_POUND] = ACTIONS(3297),
    [anon_sym_require] = ACTIONS(3297),
    [anon_sym_map] = ACTIONS(3297),
    [anon_sym_grep] = ACTIONS(3297),
    [anon_sym_sort] = ACTIONS(3297),
    [anon_sym_goto] = ACTIONS(3297),
    [anon_sym_return] = ACTIONS(3297),
    [anon_sym_undef] = ACTIONS(3297),
    [anon_sym_PERCENT2] = ACTIONS(3297),
    [anon_sym_AMP2] = ACTIONS(3297),
    [anon_sym_STAR2] = ACTIONS(3297),
    [anon_sym_last] = ACTIONS(3297),
    [anon_sym_next] = ACTIONS(3297),
    [anon_sym_redo] = ACTIONS(3297),
    [anon_sym___FILE__] = ACTIONS(3297),
    [anon_sym___LINE__] = ACTIONS(3297),
    [anon_sym___PACKAGE__] = ACTIONS(3297),
    [anon_sym___SUB__] = ACTIONS(3297),
    [anon_sym_break] = ACTIONS(3297),
    [anon_sym_fork] = ACTIONS(3297),
    [anon_sym_getppid] = ACTIONS(3297),
    [anon_sym_time] = ACTIONS(3297),
    [anon_sym_times] = ACTIONS(3297),
    [anon_sym_wait] = ACTIONS(3297),
    [anon_sym_wantarray] = ACTIONS(3297),
    [anon_sym_abs] = ACTIONS(3297),
    [anon_sym_alarm] = ACTIONS(3297),
    [anon_sym_chop] = ACTIONS(3297),
    [anon_sym_chdir] = ACTIONS(3297),
    [anon_sym_close] = ACTIONS(3297),
    [anon_sym_closedir] = ACTIONS(3297),
    [anon_sym_caller] = ACTIONS(3297),
    [anon_sym_chomp] = ACTIONS(3297),
    [anon_sym_chr] = ACTIONS(3297),
    [anon_sym_cos] = ACTIONS(3297),
    [anon_sym_chroot] = ACTIONS(3297),
    [anon_sym_defined] = ACTIONS(3297),
    [anon_sym_delete] = ACTIONS(3297),
    [anon_sym_dbmclose] = ACTIONS(3297),
    [anon_sym_exists] = ACTIONS(3297),
    [anon_sym_exit] = ACTIONS(3297),
    [anon_sym_eof] = ACTIONS(3297),
    [anon_sym_exp] = ACTIONS(3297),
    [anon_sym_each] = ACTIONS(3297),
    [anon_sym_fc] = ACTIONS(3297),
    [anon_sym_fileno] = ACTIONS(3297),
    [anon_sym_gmtime] = ACTIONS(3297),
    [anon_sym_getc] = ACTIONS(3297),
    [anon_sym_getpgrp] = ACTIONS(3297),
    [anon_sym_getprotobyname] = ACTIONS(3297),
    [anon_sym_getpwname] = ACTIONS(3297),
    [anon_sym_getpwuid] = ACTIONS(3297),
    [anon_sym_getpeername] = ACTIONS(3297),
    [anon_sym_getnetbyname] = ACTIONS(3297),
    [anon_sym_getsockname] = ACTIONS(3297),
    [anon_sym_getgrnam] = ACTIONS(3297),
    [anon_sym_getgrgid] = ACTIONS(3297),
    [anon_sym_hex] = ACTIONS(3297),
    [anon_sym_int] = ACTIONS(3297),
    [anon_sym_keys] = ACTIONS(3297),
    [anon_sym_lc] = ACTIONS(3297),
    [anon_sym_lcfirst] = ACTIONS(3297),
    [anon_sym_length] = ACTIONS(3297),
    [anon_sym_localtime] = ACTIONS(3297),
    [anon_sym_log] = ACTIONS(3297),
    [anon_sym_lock] = ACTIONS(3297),
    [anon_sym_lstat] = ACTIONS(3297),
    [anon_sym_oct] = ACTIONS(3297),
    [anon_sym_ord] = ACTIONS(3297),
    [anon_sym_prototype] = ACTIONS(3297),
    [anon_sym_pop] = ACTIONS(3297),
    [anon_sym_pos] = ACTIONS(3297),
    [anon_sym_quotemeta] = ACTIONS(3297),
    [anon_sym_reset] = ACTIONS(3297),
    [anon_sym_rand] = ACTIONS(3297),
    [anon_sym_rmdir] = ACTIONS(3297),
    [anon_sym_readdir] = ACTIONS(3297),
    [anon_sym_readline] = ACTIONS(3297),
    [anon_sym_readpipe] = ACTIONS(3297),
    [anon_sym_rewinddir] = ACTIONS(3297),
    [anon_sym_readlink] = ACTIONS(3297),
    [anon_sym_ref] = ACTIONS(3297),
    [anon_sym_scalar] = ACTIONS(3297),
    [anon_sym_shift] = ACTIONS(3297),
    [anon_sym_sin] = ACTIONS(3297),
    [anon_sym_sleep] = ACTIONS(3297),
    [anon_sym_sqrt] = ACTIONS(3297),
    [anon_sym_srand] = ACTIONS(3297),
    [anon_sym_stat] = ACTIONS(3297),
    [anon_sym_study] = ACTIONS(3297),
    [anon_sym_tell] = ACTIONS(3297),
    [anon_sym_telldir] = ACTIONS(3297),
    [anon_sym_tied] = ACTIONS(3297),
    [anon_sym_uc] = ACTIONS(3297),
    [anon_sym_ucfirst] = ACTIONS(3297),
    [anon_sym_untie] = ACTIONS(3297),
    [anon_sym_umask] = ACTIONS(3297),
    [anon_sym_values] = ACTIONS(3297),
    [anon_sym_write] = ACTIONS(3297),
    [sym_comment] = ACTIONS(3),
    [anon_sym_q] = ACTIONS(3297),
    [anon_sym_qq] = ACTIONS(3297),
    [anon_sym_qw] = ACTIONS(3297),
    [anon_sym_qx] = ACTIONS(3297),
    [anon_sym_qr] = ACTIONS(3297),
    [anon_sym_m] = ACTIONS(3297),
    [anon_sym_s] = ACTIONS(3297),
    [anon_sym_tr] = ACTIONS(3297),
    [anon_sym_y] = ACTIONS(3297),
    [aux_sym__bareword_token1] = ACTIONS(3297),
    [sym_number] = ACTIONS(3297),
    [anon_sym_true] = ACTIONS(3297),
    [anon_sym_false] = ACTIONS(3297),
    [sym__single_quote] = ACTIONS(3299),
    [sym__double_quote] = ACTIONS(3299),
    [sym__backtick_quote] = ACTIONS(3299),
    [sym__search_slash_quote] = ACTIONS(3299),
    [sym__open_readline_bracket] = ACTIONS(3299),
    [sym__open_fileglob_bracket] = ACTIONS(3299),
    [sym__PERLY_HEREDOC] = ACTIONS(3299),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__fat_comma_autoquoted] = ACTIONS(3299),
    [sym__filetest] = ACTIONS(3299),
  },
  [STATE(1736)] = {
    [sym_attribute] = STATE(1769),
    [sym_attribute_name] = STATE(1744),
    [sym_heredoc_content] = STATE(1736),
    [aux_sym_attrlist_repeat1] = STATE(1736),
    [sym__identifier] = ACTIONS(3301),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(3306),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [anon_sym_for] = ACTIONS(3309),
    [anon_sym_foreach] = ACTIONS(3309),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3309),
    [anon_sym___END__] = ACTIONS(3309),
    [anon_sym_EOT] = ACTIONS(3304),
    [anon_sym_if] = ACTIONS(3309),
    [anon_sym_unless] = ACTIONS(3309),
    [anon_sym_while] = ACTIONS(3309),
    [anon_sym_until] = ACTIONS(3309),
    [aux_sym__bareword_token1] = ACTIONS(3301),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym__ctrl_z_hack] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1737)] = {
    [sym_attribute] = STATE(1769),
    [sym_attribute_name] = STATE(1744),
    [sym_heredoc_content] = STATE(1737),
    [aux_sym_attrlist_repeat1] = STATE(1738),
    [sym__identifier] = ACTIONS(3311),
    [anon_sym_RBRACE] = ACTIONS(3313),
    [anon_sym_COLON] = ACTIONS(3315),
    [anon_sym_SEMI] = ACTIONS(3313),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_RPAREN] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [anon_sym_for] = ACTIONS(3317),
    [anon_sym_foreach] = ACTIONS(3317),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3317),
    [anon_sym___END__] = ACTIONS(3317),
    [anon_sym_EOT] = ACTIONS(3313),
    [anon_sym_if] = ACTIONS(3317),
    [anon_sym_unless] = ACTIONS(3317),
    [anon_sym_while] = ACTIONS(3317),
    [anon_sym_until] = ACTIONS(3317),
    [aux_sym__bareword_token1] = ACTIONS(3311),
    [sym__PERLY_SEMICOLON] = ACTIONS(3313),
    [sym__ctrl_z_hack] = ACTIONS(3313),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1738)] = {
    [sym_attribute] = STATE(1769),
    [sym_attribute_name] = STATE(1744),
    [sym_heredoc_content] = STATE(1738),
    [aux_sym_attrlist_repeat1] = STATE(1736),
    [sym__identifier] = ACTIONS(3311),
    [anon_sym_RBRACE] = ACTIONS(3319),
    [anon_sym_COLON] = ACTIONS(3315),
    [anon_sym_SEMI] = ACTIONS(3319),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_RPAREN] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [anon_sym_for] = ACTIONS(3321),
    [anon_sym_foreach] = ACTIONS(3321),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3321),
    [anon_sym___END__] = ACTIONS(3321),
    [anon_sym_EOT] = ACTIONS(3319),
    [anon_sym_if] = ACTIONS(3321),
    [anon_sym_unless] = ACTIONS(3321),
    [anon_sym_while] = ACTIONS(3321),
    [anon_sym_until] = ACTIONS(3321),
    [aux_sym__bareword_token1] = ACTIONS(3311),
    [sym__PERLY_SEMICOLON] = ACTIONS(3319),
    [sym__ctrl_z_hack] = ACTIONS(3319),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1739)] = {
    [sym_attrlist] = STATE(1889),
    [sym_attribute] = STATE(1737),
    [sym_attribute_name] = STATE(1744),
    [sym_heredoc_content] = STATE(1739),
    [sym__identifier] = ACTIONS(3311),
    [anon_sym_RBRACE] = ACTIONS(3323),
    [anon_sym_SEMI] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_RPAREN] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [anon_sym_for] = ACTIONS(3325),
    [anon_sym_foreach] = ACTIONS(3325),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3325),
    [anon_sym___END__] = ACTIONS(3325),
    [anon_sym_EOT] = ACTIONS(3323),
    [anon_sym_if] = ACTIONS(3325),
    [anon_sym_unless] = ACTIONS(3325),
    [anon_sym_while] = ACTIONS(3325),
    [anon_sym_until] = ACTIONS(3325),
    [aux_sym__bareword_token1] = ACTIONS(3311),
    [sym__PERLY_SEMICOLON] = ACTIONS(3323),
    [sym__ctrl_z_hack] = ACTIONS(3323),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1740)] = {
    [sym_attrlist] = STATE(1888),
    [sym_attribute] = STATE(1737),
    [sym_attribute_name] = STATE(1744),
    [sym_heredoc_content] = STATE(1740),
    [sym__identifier] = ACTIONS(3311),
    [anon_sym_RBRACE] = ACTIONS(3327),
    [anon_sym_SEMI] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_RPAREN] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [anon_sym_for] = ACTIONS(3329),
    [anon_sym_foreach] = ACTIONS(3329),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3329),
    [anon_sym___END__] = ACTIONS(3329),
    [anon_sym_EOT] = ACTIONS(3327),
    [anon_sym_if] = ACTIONS(3329),
    [anon_sym_unless] = ACTIONS(3329),
    [anon_sym_while] = ACTIONS(3329),
    [anon_sym_until] = ACTIONS(3329),
    [aux_sym__bareword_token1] = ACTIONS(3311),
    [sym__PERLY_SEMICOLON] = ACTIONS(3327),
    [sym__ctrl_z_hack] = ACTIONS(3327),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1741)] = {
    [sym_heredoc_content] = STATE(1741),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_RBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_SEMI] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_RPAREN] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_and] = ACTIONS(2715),
    [anon_sym_or] = ACTIONS(2715),
    [anon_sym_xor] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [anon_sym_for] = ACTIONS(2713),
    [anon_sym_foreach] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2715),
    [anon_sym___END__] = ACTIONS(2715),
    [anon_sym_EOT] = ACTIONS(2715),
    [anon_sym_if] = ACTIONS(2715),
    [anon_sym_unless] = ACTIONS(2715),
    [anon_sym_while] = ACTIONS(2715),
    [anon_sym_until] = ACTIONS(2715),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym__ctrl_z_hack] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1742)] = {
    [sym_heredoc_content] = STATE(1742),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_RBRACE] = ACTIONS(3333),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_SEMI] = ACTIONS(3333),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_RPAREN] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [anon_sym_for] = ACTIONS(3331),
    [anon_sym_foreach] = ACTIONS(3331),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3331),
    [anon_sym___END__] = ACTIONS(3331),
    [anon_sym_EOT] = ACTIONS(3333),
    [anon_sym_if] = ACTIONS(3331),
    [anon_sym_unless] = ACTIONS(3331),
    [anon_sym_while] = ACTIONS(3331),
    [anon_sym_until] = ACTIONS(3331),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym__PERLY_SEMICOLON] = ACTIONS(3333),
    [sym__ctrl_z_hack] = ACTIONS(3333),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1743)] = {
    [sym_heredoc_content] = STATE(1743),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_RBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_SEMI] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_RPAREN] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_and] = ACTIONS(2719),
    [anon_sym_or] = ACTIONS(2719),
    [anon_sym_xor] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [anon_sym_for] = ACTIONS(2717),
    [anon_sym_foreach] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2719),
    [anon_sym___END__] = ACTIONS(2719),
    [anon_sym_EOT] = ACTIONS(2719),
    [anon_sym_if] = ACTIONS(2719),
    [anon_sym_unless] = ACTIONS(2719),
    [anon_sym_while] = ACTIONS(2719),
    [anon_sym_until] = ACTIONS(2719),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym__ctrl_z_hack] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1744)] = {
    [sym_heredoc_content] = STATE(1744),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_RBRACE] = ACTIONS(3337),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_SEMI] = ACTIONS(3337),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_RPAREN] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [anon_sym_for] = ACTIONS(3335),
    [anon_sym_foreach] = ACTIONS(3335),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3335),
    [anon_sym___END__] = ACTIONS(3335),
    [anon_sym_EOT] = ACTIONS(3337),
    [anon_sym_if] = ACTIONS(3335),
    [anon_sym_unless] = ACTIONS(3335),
    [anon_sym_while] = ACTIONS(3335),
    [anon_sym_until] = ACTIONS(3335),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym__PERLY_SEMICOLON] = ACTIONS(3337),
    [sym__ctrl_z_hack] = ACTIONS(3337),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3339),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1745)] = {
    [sym_heredoc_content] = STATE(1745),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_RBRACE] = ACTIONS(3341),
    [anon_sym_SEMI] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_RPAREN] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_and] = ACTIONS(3341),
    [anon_sym_or] = ACTIONS(3341),
    [anon_sym_xor] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [anon_sym_for] = ACTIONS(3343),
    [anon_sym_foreach] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3341),
    [anon_sym___END__] = ACTIONS(3341),
    [anon_sym_EOT] = ACTIONS(3341),
    [anon_sym_if] = ACTIONS(3341),
    [anon_sym_unless] = ACTIONS(3341),
    [anon_sym_while] = ACTIONS(3341),
    [anon_sym_until] = ACTIONS(3341),
    [sym__PERLY_SEMICOLON] = ACTIONS(3341),
    [sym__ctrl_z_hack] = ACTIONS(3341),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1746)] = {
    [sym_heredoc_content] = STATE(1746),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_RBRACE] = ACTIONS(3345),
    [anon_sym_SEMI] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_RPAREN] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_and] = ACTIONS(3345),
    [anon_sym_or] = ACTIONS(3345),
    [anon_sym_xor] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [anon_sym_for] = ACTIONS(3347),
    [anon_sym_foreach] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3345),
    [anon_sym___END__] = ACTIONS(3345),
    [anon_sym_EOT] = ACTIONS(3345),
    [anon_sym_if] = ACTIONS(3345),
    [anon_sym_unless] = ACTIONS(3345),
    [anon_sym_while] = ACTIONS(3345),
    [anon_sym_until] = ACTIONS(3345),
    [sym__PERLY_SEMICOLON] = ACTIONS(3345),
    [sym__ctrl_z_hack] = ACTIONS(3345),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1747)] = {
    [sym_heredoc_content] = STATE(1747),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_RBRACE] = ACTIONS(3349),
    [anon_sym_SEMI] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_RPAREN] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_and] = ACTIONS(3349),
    [anon_sym_or] = ACTIONS(3349),
    [anon_sym_xor] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [anon_sym_for] = ACTIONS(3351),
    [anon_sym_foreach] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3349),
    [anon_sym___END__] = ACTIONS(3349),
    [anon_sym_EOT] = ACTIONS(3349),
    [anon_sym_if] = ACTIONS(3349),
    [anon_sym_unless] = ACTIONS(3349),
    [anon_sym_while] = ACTIONS(3349),
    [anon_sym_until] = ACTIONS(3349),
    [sym__PERLY_SEMICOLON] = ACTIONS(3349),
    [sym__ctrl_z_hack] = ACTIONS(3349),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1748)] = {
    [sym_heredoc_content] = STATE(1748),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_RBRACE] = ACTIONS(3353),
    [anon_sym_SEMI] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_RPAREN] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_and] = ACTIONS(3353),
    [anon_sym_or] = ACTIONS(3353),
    [anon_sym_xor] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [anon_sym_for] = ACTIONS(3355),
    [anon_sym_foreach] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3353),
    [anon_sym___END__] = ACTIONS(3353),
    [anon_sym_EOT] = ACTIONS(3353),
    [anon_sym_if] = ACTIONS(3353),
    [anon_sym_unless] = ACTIONS(3353),
    [anon_sym_while] = ACTIONS(3353),
    [anon_sym_until] = ACTIONS(3353),
    [sym__PERLY_SEMICOLON] = ACTIONS(3353),
    [sym__ctrl_z_hack] = ACTIONS(3353),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1749)] = {
    [sym_heredoc_content] = STATE(1749),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_RBRACE] = ACTIONS(2530),
    [anon_sym_SEMI] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_RPAREN] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_and] = ACTIONS(2530),
    [anon_sym_or] = ACTIONS(2530),
    [anon_sym_xor] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [anon_sym_for] = ACTIONS(2532),
    [anon_sym_foreach] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2530),
    [anon_sym___END__] = ACTIONS(2530),
    [anon_sym_EOT] = ACTIONS(2530),
    [anon_sym_if] = ACTIONS(2530),
    [anon_sym_unless] = ACTIONS(2530),
    [anon_sym_while] = ACTIONS(2530),
    [anon_sym_until] = ACTIONS(2530),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym__ctrl_z_hack] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1750)] = {
    [sym_heredoc_content] = STATE(1750),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_RBRACE] = ACTIONS(3357),
    [anon_sym_SEMI] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_RPAREN] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_and] = ACTIONS(3357),
    [anon_sym_or] = ACTIONS(3357),
    [anon_sym_xor] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [anon_sym_for] = ACTIONS(3359),
    [anon_sym_foreach] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3357),
    [anon_sym___END__] = ACTIONS(3357),
    [anon_sym_EOT] = ACTIONS(3357),
    [anon_sym_if] = ACTIONS(3357),
    [anon_sym_unless] = ACTIONS(3357),
    [anon_sym_while] = ACTIONS(3357),
    [anon_sym_until] = ACTIONS(3357),
    [sym__PERLY_SEMICOLON] = ACTIONS(3357),
    [sym__ctrl_z_hack] = ACTIONS(3357),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1751)] = {
    [sym_heredoc_content] = STATE(1751),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_RBRACE] = ACTIONS(3361),
    [anon_sym_SEMI] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_RPAREN] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_and] = ACTIONS(3361),
    [anon_sym_or] = ACTIONS(3361),
    [anon_sym_xor] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [anon_sym_for] = ACTIONS(3363),
    [anon_sym_foreach] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3361),
    [anon_sym___END__] = ACTIONS(3361),
    [anon_sym_EOT] = ACTIONS(3361),
    [anon_sym_if] = ACTIONS(3361),
    [anon_sym_unless] = ACTIONS(3361),
    [anon_sym_while] = ACTIONS(3361),
    [anon_sym_until] = ACTIONS(3361),
    [sym__PERLY_SEMICOLON] = ACTIONS(3361),
    [sym__ctrl_z_hack] = ACTIONS(3361),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1752)] = {
    [sym_heredoc_content] = STATE(1752),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_RBRACE] = ACTIONS(3365),
    [anon_sym_SEMI] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_RPAREN] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_and] = ACTIONS(3365),
    [anon_sym_or] = ACTIONS(3365),
    [anon_sym_xor] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [anon_sym_for] = ACTIONS(3367),
    [anon_sym_foreach] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3365),
    [anon_sym___END__] = ACTIONS(3365),
    [anon_sym_EOT] = ACTIONS(3365),
    [anon_sym_if] = ACTIONS(3365),
    [anon_sym_unless] = ACTIONS(3365),
    [anon_sym_while] = ACTIONS(3365),
    [anon_sym_until] = ACTIONS(3365),
    [sym__PERLY_SEMICOLON] = ACTIONS(3365),
    [sym__ctrl_z_hack] = ACTIONS(3365),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1753)] = {
    [sym_heredoc_content] = STATE(1753),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_RBRACE] = ACTIONS(3369),
    [anon_sym_SEMI] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_RPAREN] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_and] = ACTIONS(3369),
    [anon_sym_or] = ACTIONS(3369),
    [anon_sym_xor] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [anon_sym_for] = ACTIONS(3371),
    [anon_sym_foreach] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3369),
    [anon_sym___END__] = ACTIONS(3369),
    [anon_sym_EOT] = ACTIONS(3369),
    [anon_sym_if] = ACTIONS(3369),
    [anon_sym_unless] = ACTIONS(3369),
    [anon_sym_while] = ACTIONS(3369),
    [anon_sym_until] = ACTIONS(3369),
    [sym__PERLY_SEMICOLON] = ACTIONS(3369),
    [sym__ctrl_z_hack] = ACTIONS(3369),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1754)] = {
    [sym_heredoc_content] = STATE(1754),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_RBRACE] = ACTIONS(3373),
    [anon_sym_SEMI] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_RPAREN] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_and] = ACTIONS(3373),
    [anon_sym_or] = ACTIONS(3373),
    [anon_sym_xor] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [anon_sym_for] = ACTIONS(3375),
    [anon_sym_foreach] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3373),
    [anon_sym___END__] = ACTIONS(3373),
    [anon_sym_EOT] = ACTIONS(3373),
    [anon_sym_if] = ACTIONS(3373),
    [anon_sym_unless] = ACTIONS(3373),
    [anon_sym_while] = ACTIONS(3373),
    [anon_sym_until] = ACTIONS(3373),
    [sym__PERLY_SEMICOLON] = ACTIONS(3373),
    [sym__ctrl_z_hack] = ACTIONS(3373),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1755)] = {
    [sym_heredoc_content] = STATE(1755),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2686),
    [anon_sym___END__] = ACTIONS(2686),
    [anon_sym_EOT] = ACTIONS(2686),
    [anon_sym_if] = ACTIONS(2686),
    [anon_sym_unless] = ACTIONS(2686),
    [anon_sym_while] = ACTIONS(2686),
    [anon_sym_until] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1756)] = {
    [sym_heredoc_content] = STATE(1756),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_RBRACE] = ACTIONS(3377),
    [anon_sym_SEMI] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_RPAREN] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_and] = ACTIONS(3377),
    [anon_sym_or] = ACTIONS(3377),
    [anon_sym_xor] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [anon_sym_for] = ACTIONS(3379),
    [anon_sym_foreach] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3377),
    [anon_sym___END__] = ACTIONS(3377),
    [anon_sym_EOT] = ACTIONS(3377),
    [anon_sym_if] = ACTIONS(3377),
    [anon_sym_unless] = ACTIONS(3377),
    [anon_sym_while] = ACTIONS(3377),
    [anon_sym_until] = ACTIONS(3377),
    [sym__PERLY_SEMICOLON] = ACTIONS(3377),
    [sym__ctrl_z_hack] = ACTIONS(3377),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1757)] = {
    [sym_heredoc_content] = STATE(1757),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_RBRACE] = ACTIONS(3381),
    [anon_sym_SEMI] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_RPAREN] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_and] = ACTIONS(3381),
    [anon_sym_or] = ACTIONS(3381),
    [anon_sym_xor] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [anon_sym_for] = ACTIONS(3383),
    [anon_sym_foreach] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3381),
    [anon_sym___END__] = ACTIONS(3381),
    [anon_sym_EOT] = ACTIONS(3381),
    [anon_sym_if] = ACTIONS(3381),
    [anon_sym_unless] = ACTIONS(3381),
    [anon_sym_while] = ACTIONS(3381),
    [anon_sym_until] = ACTIONS(3381),
    [sym__PERLY_SEMICOLON] = ACTIONS(3381),
    [sym__ctrl_z_hack] = ACTIONS(3381),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1758)] = {
    [sym_heredoc_content] = STATE(1758),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_RBRACE] = ACTIONS(3385),
    [anon_sym_SEMI] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_RPAREN] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_and] = ACTIONS(3385),
    [anon_sym_or] = ACTIONS(3385),
    [anon_sym_xor] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [anon_sym_for] = ACTIONS(3387),
    [anon_sym_foreach] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3385),
    [anon_sym___END__] = ACTIONS(3385),
    [anon_sym_EOT] = ACTIONS(3385),
    [anon_sym_if] = ACTIONS(3385),
    [anon_sym_unless] = ACTIONS(3385),
    [anon_sym_while] = ACTIONS(3385),
    [anon_sym_until] = ACTIONS(3385),
    [sym__PERLY_SEMICOLON] = ACTIONS(3385),
    [sym__ctrl_z_hack] = ACTIONS(3385),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1759)] = {
    [sym_heredoc_content] = STATE(1759),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_RBRACE] = ACTIONS(3389),
    [anon_sym_SEMI] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_RPAREN] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_and] = ACTIONS(3389),
    [anon_sym_or] = ACTIONS(3389),
    [anon_sym_xor] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [anon_sym_for] = ACTIONS(3391),
    [anon_sym_foreach] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3389),
    [anon_sym___END__] = ACTIONS(3389),
    [anon_sym_EOT] = ACTIONS(3389),
    [anon_sym_if] = ACTIONS(3389),
    [anon_sym_unless] = ACTIONS(3389),
    [anon_sym_while] = ACTIONS(3389),
    [anon_sym_until] = ACTIONS(3389),
    [sym__PERLY_SEMICOLON] = ACTIONS(3389),
    [sym__ctrl_z_hack] = ACTIONS(3389),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1760)] = {
    [sym_heredoc_content] = STATE(1760),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_RBRACE] = ACTIONS(3393),
    [anon_sym_SEMI] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_RPAREN] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_and] = ACTIONS(3393),
    [anon_sym_or] = ACTIONS(3393),
    [anon_sym_xor] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [anon_sym_for] = ACTIONS(3395),
    [anon_sym_foreach] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3393),
    [anon_sym___END__] = ACTIONS(3393),
    [anon_sym_EOT] = ACTIONS(3393),
    [anon_sym_if] = ACTIONS(3393),
    [anon_sym_unless] = ACTIONS(3393),
    [anon_sym_while] = ACTIONS(3393),
    [anon_sym_until] = ACTIONS(3393),
    [sym__PERLY_SEMICOLON] = ACTIONS(3393),
    [sym__ctrl_z_hack] = ACTIONS(3393),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1761)] = {
    [sym_heredoc_content] = STATE(1761),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_RBRACE] = ACTIONS(2723),
    [anon_sym_SEMI] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_RPAREN] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_and] = ACTIONS(2723),
    [anon_sym_or] = ACTIONS(2723),
    [anon_sym_xor] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [anon_sym_for] = ACTIONS(2721),
    [anon_sym_foreach] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2723),
    [anon_sym___END__] = ACTIONS(2723),
    [anon_sym_EOT] = ACTIONS(2723),
    [anon_sym_if] = ACTIONS(2723),
    [anon_sym_unless] = ACTIONS(2723),
    [anon_sym_while] = ACTIONS(2723),
    [anon_sym_until] = ACTIONS(2723),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym__ctrl_z_hack] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1762)] = {
    [sym_heredoc_content] = STATE(1762),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [anon_sym_for] = ACTIONS(3309),
    [anon_sym_foreach] = ACTIONS(3309),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3309),
    [anon_sym___END__] = ACTIONS(3309),
    [anon_sym_EOT] = ACTIONS(3304),
    [anon_sym_if] = ACTIONS(3309),
    [anon_sym_unless] = ACTIONS(3309),
    [anon_sym_while] = ACTIONS(3309),
    [anon_sym_until] = ACTIONS(3309),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym__ctrl_z_hack] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1763)] = {
    [sym_heredoc_content] = STATE(1763),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_RBRACE] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_RPAREN] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_and] = ACTIONS(2694),
    [anon_sym_or] = ACTIONS(2694),
    [anon_sym_xor] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [anon_sym_for] = ACTIONS(2692),
    [anon_sym_foreach] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2694),
    [anon_sym___END__] = ACTIONS(2694),
    [anon_sym_EOT] = ACTIONS(2694),
    [anon_sym_if] = ACTIONS(2694),
    [anon_sym_unless] = ACTIONS(2694),
    [anon_sym_while] = ACTIONS(2694),
    [anon_sym_until] = ACTIONS(2694),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym__ctrl_z_hack] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1764)] = {
    [sym_heredoc_content] = STATE(1764),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_RBRACE] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_RPAREN] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_and] = ACTIONS(2690),
    [anon_sym_or] = ACTIONS(2690),
    [anon_sym_xor] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [anon_sym_for] = ACTIONS(2688),
    [anon_sym_foreach] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2690),
    [anon_sym___END__] = ACTIONS(2690),
    [anon_sym_EOT] = ACTIONS(2690),
    [anon_sym_if] = ACTIONS(2690),
    [anon_sym_unless] = ACTIONS(2690),
    [anon_sym_while] = ACTIONS(2690),
    [anon_sym_until] = ACTIONS(2690),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym__ctrl_z_hack] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1765)] = {
    [sym_heredoc_content] = STATE(1765),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_RBRACE] = ACTIONS(3397),
    [anon_sym_SEMI] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_RPAREN] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_and] = ACTIONS(3397),
    [anon_sym_or] = ACTIONS(3397),
    [anon_sym_xor] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [anon_sym_for] = ACTIONS(3399),
    [anon_sym_foreach] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3397),
    [anon_sym___END__] = ACTIONS(3397),
    [anon_sym_EOT] = ACTIONS(3397),
    [anon_sym_if] = ACTIONS(3397),
    [anon_sym_unless] = ACTIONS(3397),
    [anon_sym_while] = ACTIONS(3397),
    [anon_sym_until] = ACTIONS(3397),
    [sym__PERLY_SEMICOLON] = ACTIONS(3397),
    [sym__ctrl_z_hack] = ACTIONS(3397),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1766)] = {
    [sym_heredoc_content] = STATE(1766),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_RBRACE] = ACTIONS(3403),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_SEMI] = ACTIONS(3403),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_RPAREN] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [anon_sym_for] = ACTIONS(3401),
    [anon_sym_foreach] = ACTIONS(3401),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3401),
    [anon_sym___END__] = ACTIONS(3401),
    [anon_sym_EOT] = ACTIONS(3403),
    [anon_sym_if] = ACTIONS(3401),
    [anon_sym_unless] = ACTIONS(3401),
    [anon_sym_while] = ACTIONS(3401),
    [anon_sym_until] = ACTIONS(3401),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym__PERLY_SEMICOLON] = ACTIONS(3403),
    [sym__ctrl_z_hack] = ACTIONS(3403),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1767)] = {
    [sym_heredoc_content] = STATE(1767),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_RBRACE] = ACTIONS(3405),
    [anon_sym_SEMI] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_RPAREN] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_and] = ACTIONS(3405),
    [anon_sym_or] = ACTIONS(3405),
    [anon_sym_xor] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [anon_sym_for] = ACTIONS(3407),
    [anon_sym_foreach] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3405),
    [anon_sym___END__] = ACTIONS(3405),
    [anon_sym_EOT] = ACTIONS(3405),
    [anon_sym_if] = ACTIONS(3405),
    [anon_sym_unless] = ACTIONS(3405),
    [anon_sym_while] = ACTIONS(3405),
    [anon_sym_until] = ACTIONS(3405),
    [sym__PERLY_SEMICOLON] = ACTIONS(3405),
    [sym__ctrl_z_hack] = ACTIONS(3405),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1768)] = {
    [sym_heredoc_content] = STATE(1768),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_RBRACE] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_RPAREN] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_and] = ACTIONS(2698),
    [anon_sym_or] = ACTIONS(2698),
    [anon_sym_xor] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [anon_sym_for] = ACTIONS(2696),
    [anon_sym_foreach] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2698),
    [anon_sym___END__] = ACTIONS(2698),
    [anon_sym_EOT] = ACTIONS(2698),
    [anon_sym_if] = ACTIONS(2698),
    [anon_sym_unless] = ACTIONS(2698),
    [anon_sym_while] = ACTIONS(2698),
    [anon_sym_until] = ACTIONS(2698),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym__ctrl_z_hack] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1769)] = {
    [sym_heredoc_content] = STATE(1769),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_RBRACE] = ACTIONS(3411),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_SEMI] = ACTIONS(3411),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_RPAREN] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [anon_sym_for] = ACTIONS(3409),
    [anon_sym_foreach] = ACTIONS(3409),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3409),
    [anon_sym___END__] = ACTIONS(3409),
    [anon_sym_EOT] = ACTIONS(3411),
    [anon_sym_if] = ACTIONS(3409),
    [anon_sym_unless] = ACTIONS(3409),
    [anon_sym_while] = ACTIONS(3409),
    [anon_sym_until] = ACTIONS(3409),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym__PERLY_SEMICOLON] = ACTIONS(3411),
    [sym__ctrl_z_hack] = ACTIONS(3411),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1770)] = {
    [sym_heredoc_content] = STATE(1770),
    [anon_sym_LBRACE] = ACTIONS(3413),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(3419),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(3421),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [anon_sym_for] = ACTIONS(3417),
    [anon_sym_foreach] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3415),
    [anon_sym___END__] = ACTIONS(3415),
    [anon_sym_EOT] = ACTIONS(3415),
    [anon_sym_if] = ACTIONS(3415),
    [anon_sym_unless] = ACTIONS(3415),
    [anon_sym_while] = ACTIONS(3415),
    [anon_sym_until] = ACTIONS(3415),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym__ctrl_z_hack] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1771)] = {
    [sym_heredoc_content] = STATE(1771),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_RBRACE] = ACTIONS(3425),
    [anon_sym_SEMI] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_RPAREN] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_and] = ACTIONS(3425),
    [anon_sym_or] = ACTIONS(3425),
    [anon_sym_xor] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [anon_sym_for] = ACTIONS(3427),
    [anon_sym_foreach] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3425),
    [anon_sym___END__] = ACTIONS(3425),
    [anon_sym_EOT] = ACTIONS(3425),
    [anon_sym_if] = ACTIONS(3425),
    [anon_sym_unless] = ACTIONS(3425),
    [anon_sym_while] = ACTIONS(3425),
    [anon_sym_until] = ACTIONS(3425),
    [sym__PERLY_SEMICOLON] = ACTIONS(3425),
    [sym__ctrl_z_hack] = ACTIONS(3425),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1772)] = {
    [sym_heredoc_content] = STATE(1772),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_RBRACE] = ACTIONS(3431),
    [anon_sym_SEMI] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_RPAREN] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_and] = ACTIONS(3431),
    [anon_sym_or] = ACTIONS(3431),
    [anon_sym_xor] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [anon_sym_for] = ACTIONS(3433),
    [anon_sym_foreach] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3431),
    [anon_sym___END__] = ACTIONS(3431),
    [anon_sym_EOT] = ACTIONS(3431),
    [anon_sym_if] = ACTIONS(3431),
    [anon_sym_unless] = ACTIONS(3431),
    [anon_sym_while] = ACTIONS(3431),
    [anon_sym_until] = ACTIONS(3431),
    [sym__PERLY_SEMICOLON] = ACTIONS(3431),
    [sym__ctrl_z_hack] = ACTIONS(3431),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1773)] = {
    [sym_heredoc_content] = STATE(1773),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3435),
    [anon_sym_or] = ACTIONS(3435),
    [anon_sym_xor] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [anon_sym_for] = ACTIONS(3437),
    [anon_sym_foreach] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3435),
    [anon_sym___END__] = ACTIONS(3435),
    [anon_sym_EOT] = ACTIONS(3435),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [anon_sym_if] = ACTIONS(3435),
    [anon_sym_unless] = ACTIONS(3435),
    [anon_sym_while] = ACTIONS(3435),
    [anon_sym_until] = ACTIONS(3435),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym__ctrl_z_hack] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1774)] = {
    [sym_heredoc_content] = STATE(1774),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [anon_sym_for] = ACTIONS(2682),
    [anon_sym_foreach] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2686),
    [anon_sym___END__] = ACTIONS(2686),
    [anon_sym_EOT] = ACTIONS(2686),
    [anon_sym_if] = ACTIONS(2686),
    [anon_sym_unless] = ACTIONS(2686),
    [anon_sym_while] = ACTIONS(2686),
    [anon_sym_until] = ACTIONS(2686),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym__ctrl_z_hack] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1775)] = {
    [sym_heredoc_content] = STATE(1775),
    [anon_sym_RBRACE] = ACTIONS(3441),
    [anon_sym_SEMI] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_RPAREN] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_and] = ACTIONS(3441),
    [anon_sym_or] = ACTIONS(3441),
    [anon_sym_xor] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [anon_sym_for] = ACTIONS(3443),
    [anon_sym_foreach] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3441),
    [anon_sym___END__] = ACTIONS(3441),
    [anon_sym_EOT] = ACTIONS(3441),
    [sym_transliteration_modifiers] = ACTIONS(3445),
    [anon_sym_if] = ACTIONS(3441),
    [anon_sym_unless] = ACTIONS(3441),
    [anon_sym_while] = ACTIONS(3441),
    [anon_sym_until] = ACTIONS(3441),
    [sym__PERLY_SEMICOLON] = ACTIONS(3441),
    [sym__ctrl_z_hack] = ACTIONS(3441),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1776)] = {
    [sym_heredoc_content] = STATE(1776),
    [anon_sym_RBRACE] = ACTIONS(3447),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_SEMI] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_RPAREN] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_and] = ACTIONS(3447),
    [anon_sym_or] = ACTIONS(3447),
    [anon_sym_xor] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [anon_sym_for] = ACTIONS(3449),
    [anon_sym_foreach] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3447),
    [anon_sym___END__] = ACTIONS(3447),
    [anon_sym_EOT] = ACTIONS(3447),
    [anon_sym_if] = ACTIONS(3447),
    [anon_sym_unless] = ACTIONS(3447),
    [anon_sym_while] = ACTIONS(3447),
    [anon_sym_until] = ACTIONS(3447),
    [sym__PERLY_SEMICOLON] = ACTIONS(3447),
    [sym__ctrl_z_hack] = ACTIONS(3447),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1777)] = {
    [sym_heredoc_content] = STATE(1777),
    [anon_sym_RBRACE] = ACTIONS(3451),
    [anon_sym_SEMI] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_RPAREN] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_and] = ACTIONS(3451),
    [anon_sym_or] = ACTIONS(3451),
    [anon_sym_xor] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(3455),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [anon_sym_for] = ACTIONS(3453),
    [anon_sym_foreach] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3451),
    [anon_sym___END__] = ACTIONS(3451),
    [anon_sym_EOT] = ACTIONS(3451),
    [anon_sym_if] = ACTIONS(3451),
    [anon_sym_unless] = ACTIONS(3451),
    [anon_sym_while] = ACTIONS(3451),
    [anon_sym_until] = ACTIONS(3451),
    [sym__PERLY_SEMICOLON] = ACTIONS(3451),
    [sym__ctrl_z_hack] = ACTIONS(3451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1778)] = {
    [sym_heredoc_content] = STATE(1778),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [anon_sym_for] = ACTIONS(3437),
    [anon_sym_foreach] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3435),
    [anon_sym___END__] = ACTIONS(3435),
    [anon_sym_EOT] = ACTIONS(3435),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [anon_sym_if] = ACTIONS(3437),
    [anon_sym_unless] = ACTIONS(3437),
    [anon_sym_while] = ACTIONS(3435),
    [anon_sym_until] = ACTIONS(3437),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym__ctrl_z_hack] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1779)] = {
    [sym_heredoc_content] = STATE(1779),
    [anon_sym_RBRACE] = ACTIONS(3457),
    [anon_sym_SEMI] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_RPAREN] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_and] = ACTIONS(3457),
    [anon_sym_or] = ACTIONS(3457),
    [anon_sym_xor] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(3461),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [anon_sym_for] = ACTIONS(3459),
    [anon_sym_foreach] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3457),
    [anon_sym___END__] = ACTIONS(3457),
    [anon_sym_EOT] = ACTIONS(3457),
    [anon_sym_if] = ACTIONS(3457),
    [anon_sym_unless] = ACTIONS(3457),
    [anon_sym_while] = ACTIONS(3457),
    [anon_sym_until] = ACTIONS(3457),
    [sym__PERLY_SEMICOLON] = ACTIONS(3457),
    [sym__ctrl_z_hack] = ACTIONS(3457),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1780)] = {
    [sym_heredoc_content] = STATE(1780),
    [anon_sym_RBRACE] = ACTIONS(3463),
    [anon_sym_SEMI] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_RPAREN] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_and] = ACTIONS(3465),
    [anon_sym_or] = ACTIONS(3465),
    [anon_sym_xor] = ACTIONS(3465),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [anon_sym_for] = ACTIONS(3465),
    [anon_sym_foreach] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3463),
    [anon_sym___END__] = ACTIONS(3463),
    [anon_sym_EOT] = ACTIONS(3463),
    [sym_match_regexp_modifiers] = ACTIONS(3467),
    [anon_sym_if] = ACTIONS(3465),
    [anon_sym_unless] = ACTIONS(3465),
    [anon_sym_while] = ACTIONS(3463),
    [anon_sym_until] = ACTIONS(3465),
    [sym__PERLY_SEMICOLON] = ACTIONS(3463),
    [sym__ctrl_z_hack] = ACTIONS(3463),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1781)] = {
    [sym_heredoc_content] = STATE(1781),
    [anon_sym_RBRACE] = ACTIONS(3469),
    [anon_sym_SEMI] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_RPAREN] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_and] = ACTIONS(3471),
    [anon_sym_or] = ACTIONS(3471),
    [anon_sym_xor] = ACTIONS(3471),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [anon_sym_for] = ACTIONS(3471),
    [anon_sym_foreach] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3469),
    [anon_sym___END__] = ACTIONS(3469),
    [anon_sym_EOT] = ACTIONS(3469),
    [sym_match_regexp_modifiers] = ACTIONS(3473),
    [anon_sym_if] = ACTIONS(3471),
    [anon_sym_unless] = ACTIONS(3471),
    [anon_sym_while] = ACTIONS(3469),
    [anon_sym_until] = ACTIONS(3471),
    [sym__PERLY_SEMICOLON] = ACTIONS(3469),
    [sym__ctrl_z_hack] = ACTIONS(3469),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1782)] = {
    [sym_heredoc_content] = STATE(1782),
    [anon_sym_RBRACE] = ACTIONS(3475),
    [anon_sym_SEMI] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_RPAREN] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_and] = ACTIONS(3475),
    [anon_sym_or] = ACTIONS(3475),
    [anon_sym_xor] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [anon_sym_for] = ACTIONS(3477),
    [anon_sym_foreach] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3475),
    [anon_sym___END__] = ACTIONS(3475),
    [anon_sym_EOT] = ACTIONS(3475),
    [anon_sym_if] = ACTIONS(3475),
    [anon_sym_unless] = ACTIONS(3475),
    [anon_sym_while] = ACTIONS(3475),
    [anon_sym_until] = ACTIONS(3475),
    [sym__PERLY_SEMICOLON] = ACTIONS(3475),
    [sym__ctrl_z_hack] = ACTIONS(3475),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1783)] = {
    [sym_heredoc_content] = STATE(1783),
    [anon_sym_RBRACE] = ACTIONS(3479),
    [anon_sym_SEMI] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_RPAREN] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(3483),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_and] = ACTIONS(3479),
    [anon_sym_or] = ACTIONS(3479),
    [anon_sym_xor] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [anon_sym_for] = ACTIONS(3481),
    [anon_sym_foreach] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3479),
    [anon_sym___END__] = ACTIONS(3479),
    [anon_sym_EOT] = ACTIONS(3479),
    [anon_sym_if] = ACTIONS(3479),
    [anon_sym_unless] = ACTIONS(3479),
    [anon_sym_while] = ACTIONS(3479),
    [anon_sym_until] = ACTIONS(3479),
    [sym__PERLY_SEMICOLON] = ACTIONS(3479),
    [sym__ctrl_z_hack] = ACTIONS(3479),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1784)] = {
    [sym_heredoc_content] = STATE(1784),
    [aux_sym__term_rightward_repeat1] = STATE(1790),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(3491),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3491),
    [anon_sym_for] = ACTIONS(3536),
    [anon_sym_foreach] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3485),
    [anon_sym___END__] = ACTIONS(3485),
    [anon_sym_EOT] = ACTIONS(3485),
    [anon_sym_if] = ACTIONS(3485),
    [anon_sym_unless] = ACTIONS(3485),
    [anon_sym_while] = ACTIONS(3485),
    [anon_sym_until] = ACTIONS(3485),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym__ctrl_z_hack] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1785)] = {
    [sym_heredoc_content] = STATE(1785),
    [anon_sym_RBRACE] = ACTIONS(3538),
    [anon_sym_SEMI] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_RPAREN] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_and] = ACTIONS(3540),
    [anon_sym_or] = ACTIONS(3540),
    [anon_sym_xor] = ACTIONS(3540),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [anon_sym_for] = ACTIONS(3540),
    [anon_sym_foreach] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3538),
    [anon_sym___END__] = ACTIONS(3538),
    [anon_sym_EOT] = ACTIONS(3538),
    [sym_substitution_regexp_modifiers] = ACTIONS(3542),
    [anon_sym_if] = ACTIONS(3540),
    [anon_sym_unless] = ACTIONS(3540),
    [anon_sym_while] = ACTIONS(3538),
    [anon_sym_until] = ACTIONS(3540),
    [sym__PERLY_SEMICOLON] = ACTIONS(3538),
    [sym__ctrl_z_hack] = ACTIONS(3538),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1786)] = {
    [sym_heredoc_content] = STATE(1786),
    [anon_sym_RBRACE] = ACTIONS(3544),
    [anon_sym_SEMI] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_RPAREN] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_and] = ACTIONS(3546),
    [anon_sym_or] = ACTIONS(3546),
    [anon_sym_xor] = ACTIONS(3546),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [anon_sym_for] = ACTIONS(3546),
    [anon_sym_foreach] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3544),
    [anon_sym___END__] = ACTIONS(3544),
    [anon_sym_EOT] = ACTIONS(3544),
    [sym_substitution_regexp_modifiers] = ACTIONS(3548),
    [anon_sym_if] = ACTIONS(3546),
    [anon_sym_unless] = ACTIONS(3546),
    [anon_sym_while] = ACTIONS(3544),
    [anon_sym_until] = ACTIONS(3546),
    [sym__PERLY_SEMICOLON] = ACTIONS(3544),
    [sym__ctrl_z_hack] = ACTIONS(3544),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1787)] = {
    [sym_heredoc_content] = STATE(1787),
    [anon_sym_RBRACE] = ACTIONS(3550),
    [anon_sym_SEMI] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_RPAREN] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_and] = ACTIONS(3550),
    [anon_sym_or] = ACTIONS(3550),
    [anon_sym_xor] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [anon_sym_for] = ACTIONS(3552),
    [anon_sym_foreach] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3550),
    [anon_sym___END__] = ACTIONS(3550),
    [anon_sym_EOT] = ACTIONS(3550),
    [sym_transliteration_modifiers] = ACTIONS(3554),
    [anon_sym_if] = ACTIONS(3550),
    [anon_sym_unless] = ACTIONS(3550),
    [anon_sym_while] = ACTIONS(3550),
    [anon_sym_until] = ACTIONS(3550),
    [sym__PERLY_SEMICOLON] = ACTIONS(3550),
    [sym__ctrl_z_hack] = ACTIONS(3550),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1788)] = {
    [sym_heredoc_content] = STATE(1788),
    [anon_sym_RBRACE] = ACTIONS(3556),
    [anon_sym_SEMI] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_RPAREN] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_and] = ACTIONS(3558),
    [anon_sym_or] = ACTIONS(3558),
    [anon_sym_xor] = ACTIONS(3558),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [anon_sym_for] = ACTIONS(3558),
    [anon_sym_foreach] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3556),
    [anon_sym___END__] = ACTIONS(3556),
    [anon_sym_EOT] = ACTIONS(3556),
    [sym_match_regexp_modifiers] = ACTIONS(3560),
    [anon_sym_if] = ACTIONS(3558),
    [anon_sym_unless] = ACTIONS(3558),
    [anon_sym_while] = ACTIONS(3556),
    [anon_sym_until] = ACTIONS(3558),
    [sym__PERLY_SEMICOLON] = ACTIONS(3556),
    [sym__ctrl_z_hack] = ACTIONS(3556),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1789)] = {
    [sym_heredoc_content] = STATE(1789),
    [anon_sym_RBRACE] = ACTIONS(3562),
    [anon_sym_SEMI] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_RPAREN] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(3566),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_and] = ACTIONS(3562),
    [anon_sym_or] = ACTIONS(3562),
    [anon_sym_xor] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [anon_sym_for] = ACTIONS(3564),
    [anon_sym_foreach] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3562),
    [anon_sym___END__] = ACTIONS(3562),
    [anon_sym_EOT] = ACTIONS(3562),
    [anon_sym_if] = ACTIONS(3562),
    [anon_sym_unless] = ACTIONS(3562),
    [anon_sym_while] = ACTIONS(3562),
    [anon_sym_until] = ACTIONS(3562),
    [sym__PERLY_SEMICOLON] = ACTIONS(3562),
    [sym__ctrl_z_hack] = ACTIONS(3562),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1790)] = {
    [sym_heredoc_content] = STATE(1790),
    [aux_sym__term_rightward_repeat1] = STATE(1794),
    [anon_sym_RBRACE] = ACTIONS(3568),
    [anon_sym_SEMI] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(3572),
    [anon_sym_RPAREN] = ACTIONS(3568),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(3572),
    [anon_sym_for] = ACTIONS(3570),
    [anon_sym_foreach] = ACTIONS(3568),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3568),
    [anon_sym___END__] = ACTIONS(3568),
    [anon_sym_EOT] = ACTIONS(3568),
    [anon_sym_if] = ACTIONS(3568),
    [anon_sym_unless] = ACTIONS(3568),
    [anon_sym_while] = ACTIONS(3568),
    [anon_sym_until] = ACTIONS(3568),
    [sym__PERLY_SEMICOLON] = ACTIONS(3568),
    [sym__ctrl_z_hack] = ACTIONS(3568),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1791)] = {
    [sym_heredoc_content] = STATE(1791),
    [anon_sym_RBRACE] = ACTIONS(3574),
    [anon_sym_SEMI] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_RPAREN] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_and] = ACTIONS(3574),
    [anon_sym_or] = ACTIONS(3574),
    [anon_sym_xor] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(3455),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [anon_sym_for] = ACTIONS(3576),
    [anon_sym_foreach] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3574),
    [anon_sym___END__] = ACTIONS(3574),
    [anon_sym_EOT] = ACTIONS(3574),
    [anon_sym_if] = ACTIONS(3574),
    [anon_sym_unless] = ACTIONS(3574),
    [anon_sym_while] = ACTIONS(3574),
    [anon_sym_until] = ACTIONS(3574),
    [sym__PERLY_SEMICOLON] = ACTIONS(3574),
    [sym__ctrl_z_hack] = ACTIONS(3574),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1792)] = {
    [sym_heredoc_content] = STATE(1792),
    [anon_sym_RBRACE] = ACTIONS(3578),
    [anon_sym_SEMI] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_RPAREN] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_and] = ACTIONS(3578),
    [anon_sym_or] = ACTIONS(3578),
    [anon_sym_xor] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [anon_sym_for] = ACTIONS(3580),
    [anon_sym_foreach] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3578),
    [anon_sym___END__] = ACTIONS(3578),
    [anon_sym_EOT] = ACTIONS(3578),
    [sym_transliteration_modifiers] = ACTIONS(3582),
    [anon_sym_if] = ACTIONS(3578),
    [anon_sym_unless] = ACTIONS(3578),
    [anon_sym_while] = ACTIONS(3578),
    [anon_sym_until] = ACTIONS(3578),
    [sym__PERLY_SEMICOLON] = ACTIONS(3578),
    [sym__ctrl_z_hack] = ACTIONS(3578),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1793)] = {
    [sym_heredoc_content] = STATE(1793),
    [anon_sym_RBRACE] = ACTIONS(3584),
    [anon_sym_SEMI] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_RPAREN] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_and] = ACTIONS(3586),
    [anon_sym_or] = ACTIONS(3586),
    [anon_sym_xor] = ACTIONS(3586),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [anon_sym_for] = ACTIONS(3586),
    [anon_sym_foreach] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3584),
    [anon_sym___END__] = ACTIONS(3584),
    [anon_sym_EOT] = ACTIONS(3584),
    [sym_substitution_regexp_modifiers] = ACTIONS(3588),
    [anon_sym_if] = ACTIONS(3586),
    [anon_sym_unless] = ACTIONS(3586),
    [anon_sym_while] = ACTIONS(3584),
    [anon_sym_until] = ACTIONS(3586),
    [sym__PERLY_SEMICOLON] = ACTIONS(3584),
    [sym__ctrl_z_hack] = ACTIONS(3584),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1794)] = {
    [sym_heredoc_content] = STATE(1794),
    [aux_sym__term_rightward_repeat1] = STATE(1794),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(3594),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(3594),
    [anon_sym_for] = ACTIONS(3592),
    [anon_sym_foreach] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3590),
    [anon_sym___END__] = ACTIONS(3590),
    [anon_sym_EOT] = ACTIONS(3590),
    [anon_sym_if] = ACTIONS(3590),
    [anon_sym_unless] = ACTIONS(3590),
    [anon_sym_while] = ACTIONS(3590),
    [anon_sym_until] = ACTIONS(3590),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym__ctrl_z_hack] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1795)] = {
    [sym_heredoc_content] = STATE(1795),
    [anon_sym_RBRACE] = ACTIONS(3597),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_SEMI] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_RPAREN] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_and] = ACTIONS(3597),
    [anon_sym_or] = ACTIONS(3597),
    [anon_sym_xor] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [anon_sym_for] = ACTIONS(3599),
    [anon_sym_foreach] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3597),
    [anon_sym___END__] = ACTIONS(3597),
    [anon_sym_EOT] = ACTIONS(3597),
    [anon_sym_if] = ACTIONS(3597),
    [anon_sym_unless] = ACTIONS(3597),
    [anon_sym_while] = ACTIONS(3597),
    [anon_sym_until] = ACTIONS(3597),
    [sym__PERLY_SEMICOLON] = ACTIONS(3597),
    [sym__ctrl_z_hack] = ACTIONS(3597),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1796)] = {
    [sym_heredoc_content] = STATE(1796),
    [anon_sym_RBRACE] = ACTIONS(3601),
    [anon_sym_SEMI] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_RPAREN] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_and] = ACTIONS(3601),
    [anon_sym_or] = ACTIONS(3601),
    [anon_sym_xor] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [anon_sym_for] = ACTIONS(3603),
    [anon_sym_foreach] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3601),
    [anon_sym___END__] = ACTIONS(3601),
    [anon_sym_EOT] = ACTIONS(3601),
    [anon_sym_if] = ACTIONS(3601),
    [anon_sym_unless] = ACTIONS(3601),
    [anon_sym_while] = ACTIONS(3601),
    [anon_sym_until] = ACTIONS(3601),
    [sym__PERLY_SEMICOLON] = ACTIONS(3601),
    [sym__ctrl_z_hack] = ACTIONS(3601),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1797)] = {
    [sym_heredoc_content] = STATE(1797),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_RBRACE] = ACTIONS(3605),
    [anon_sym_SEMI] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_RPAREN] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_and] = ACTIONS(3605),
    [anon_sym_or] = ACTIONS(3605),
    [anon_sym_xor] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [anon_sym_for] = ACTIONS(3607),
    [anon_sym_foreach] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3605),
    [anon_sym___END__] = ACTIONS(3605),
    [anon_sym_EOT] = ACTIONS(3605),
    [anon_sym_if] = ACTIONS(3605),
    [anon_sym_unless] = ACTIONS(3605),
    [anon_sym_while] = ACTIONS(3605),
    [anon_sym_until] = ACTIONS(3605),
    [sym__PERLY_SEMICOLON] = ACTIONS(3605),
    [sym__ctrl_z_hack] = ACTIONS(3605),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1798)] = {
    [sym_heredoc_content] = STATE(1798),
    [anon_sym_LBRACE] = ACTIONS(3609),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_for] = ACTIONS(2702),
    [anon_sym_foreach] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2711),
    [anon_sym___END__] = ACTIONS(2711),
    [anon_sym_EOT] = ACTIONS(2711),
    [anon_sym_if] = ACTIONS(2711),
    [anon_sym_unless] = ACTIONS(2711),
    [anon_sym_while] = ACTIONS(2711),
    [anon_sym_until] = ACTIONS(2711),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1799)] = {
    [sym_heredoc_content] = STATE(1799),
    [anon_sym_RBRACE] = ACTIONS(3611),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_SEMI] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_RPAREN] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_and] = ACTIONS(3611),
    [anon_sym_or] = ACTIONS(3611),
    [anon_sym_xor] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [anon_sym_for] = ACTIONS(3613),
    [anon_sym_foreach] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3611),
    [anon_sym___END__] = ACTIONS(3611),
    [anon_sym_EOT] = ACTIONS(3611),
    [anon_sym_if] = ACTIONS(3611),
    [anon_sym_unless] = ACTIONS(3611),
    [anon_sym_while] = ACTIONS(3611),
    [anon_sym_until] = ACTIONS(3611),
    [sym__PERLY_SEMICOLON] = ACTIONS(3611),
    [sym__ctrl_z_hack] = ACTIONS(3611),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1800)] = {
    [sym_heredoc_content] = STATE(1800),
    [anon_sym_RBRACE] = ACTIONS(3615),
    [anon_sym_SEMI] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_RPAREN] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_and] = ACTIONS(3617),
    [anon_sym_or] = ACTIONS(3617),
    [anon_sym_xor] = ACTIONS(3617),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [anon_sym_for] = ACTIONS(3617),
    [anon_sym_foreach] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3615),
    [anon_sym___END__] = ACTIONS(3615),
    [anon_sym_EOT] = ACTIONS(3615),
    [sym_substitution_regexp_modifiers] = ACTIONS(3619),
    [anon_sym_if] = ACTIONS(3617),
    [anon_sym_unless] = ACTIONS(3617),
    [anon_sym_while] = ACTIONS(3615),
    [anon_sym_until] = ACTIONS(3617),
    [sym__PERLY_SEMICOLON] = ACTIONS(3615),
    [sym__ctrl_z_hack] = ACTIONS(3615),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1801)] = {
    [sym_heredoc_content] = STATE(1801),
    [anon_sym_RBRACE] = ACTIONS(3621),
    [anon_sym_SEMI] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_RPAREN] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(3625),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_and] = ACTIONS(3621),
    [anon_sym_or] = ACTIONS(3621),
    [anon_sym_xor] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [anon_sym_for] = ACTIONS(3623),
    [anon_sym_foreach] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3621),
    [anon_sym___END__] = ACTIONS(3621),
    [anon_sym_EOT] = ACTIONS(3621),
    [anon_sym_if] = ACTIONS(3621),
    [anon_sym_unless] = ACTIONS(3621),
    [anon_sym_while] = ACTIONS(3621),
    [anon_sym_until] = ACTIONS(3621),
    [sym__PERLY_SEMICOLON] = ACTIONS(3621),
    [sym__ctrl_z_hack] = ACTIONS(3621),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1802)] = {
    [sym_heredoc_content] = STATE(1802),
    [anon_sym_RBRACE] = ACTIONS(3627),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_SEMI] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_RPAREN] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_and] = ACTIONS(3627),
    [anon_sym_or] = ACTIONS(3627),
    [anon_sym_xor] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [anon_sym_for] = ACTIONS(3629),
    [anon_sym_foreach] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3627),
    [anon_sym___END__] = ACTIONS(3627),
    [anon_sym_EOT] = ACTIONS(3627),
    [anon_sym_if] = ACTIONS(3627),
    [anon_sym_unless] = ACTIONS(3627),
    [anon_sym_while] = ACTIONS(3627),
    [anon_sym_until] = ACTIONS(3627),
    [sym__PERLY_SEMICOLON] = ACTIONS(3627),
    [sym__ctrl_z_hack] = ACTIONS(3627),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1803)] = {
    [sym_heredoc_content] = STATE(1803),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(3631),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [anon_sym_for] = ACTIONS(3417),
    [anon_sym_foreach] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3415),
    [anon_sym___END__] = ACTIONS(3415),
    [anon_sym_EOT] = ACTIONS(3415),
    [anon_sym_if] = ACTIONS(3415),
    [anon_sym_unless] = ACTIONS(3415),
    [anon_sym_while] = ACTIONS(3415),
    [anon_sym_until] = ACTIONS(3415),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym__ctrl_z_hack] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1804)] = {
    [sym_heredoc_content] = STATE(1804),
    [anon_sym_RBRACE] = ACTIONS(3633),
    [anon_sym_SEMI] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_RPAREN] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(3637),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_and] = ACTIONS(3633),
    [anon_sym_or] = ACTIONS(3633),
    [anon_sym_xor] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [anon_sym_for] = ACTIONS(3635),
    [anon_sym_foreach] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3633),
    [anon_sym___END__] = ACTIONS(3633),
    [anon_sym_EOT] = ACTIONS(3633),
    [anon_sym_if] = ACTIONS(3633),
    [anon_sym_unless] = ACTIONS(3633),
    [anon_sym_while] = ACTIONS(3633),
    [anon_sym_until] = ACTIONS(3633),
    [sym__PERLY_SEMICOLON] = ACTIONS(3633),
    [sym__ctrl_z_hack] = ACTIONS(3633),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1805)] = {
    [sym_heredoc_content] = STATE(1805),
    [anon_sym_RBRACE] = ACTIONS(3639),
    [anon_sym_SEMI] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_RPAREN] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_and] = ACTIONS(3639),
    [anon_sym_or] = ACTIONS(3639),
    [anon_sym_xor] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [anon_sym_for] = ACTIONS(3641),
    [anon_sym_foreach] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3639),
    [anon_sym___END__] = ACTIONS(3639),
    [anon_sym_EOT] = ACTIONS(3639),
    [anon_sym_if] = ACTIONS(3639),
    [anon_sym_unless] = ACTIONS(3639),
    [anon_sym_while] = ACTIONS(3639),
    [anon_sym_until] = ACTIONS(3639),
    [sym__PERLY_SEMICOLON] = ACTIONS(3639),
    [sym__ctrl_z_hack] = ACTIONS(3639),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1806)] = {
    [sym_heredoc_content] = STATE(1806),
    [anon_sym_RBRACE] = ACTIONS(3643),
    [anon_sym_SEMI] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_RPAREN] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_and] = ACTIONS(3645),
    [anon_sym_or] = ACTIONS(3645),
    [anon_sym_xor] = ACTIONS(3645),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [anon_sym_for] = ACTIONS(3645),
    [anon_sym_foreach] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3643),
    [anon_sym___END__] = ACTIONS(3643),
    [anon_sym_EOT] = ACTIONS(3643),
    [sym_quoted_regexp_modifiers] = ACTIONS(3647),
    [anon_sym_if] = ACTIONS(3645),
    [anon_sym_unless] = ACTIONS(3645),
    [anon_sym_while] = ACTIONS(3643),
    [anon_sym_until] = ACTIONS(3645),
    [sym__PERLY_SEMICOLON] = ACTIONS(3643),
    [sym__ctrl_z_hack] = ACTIONS(3643),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1807)] = {
    [sym_heredoc_content] = STATE(1807),
    [anon_sym_RBRACE] = ACTIONS(3649),
    [anon_sym_SEMI] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_RPAREN] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_and] = ACTIONS(3649),
    [anon_sym_or] = ACTIONS(3649),
    [anon_sym_xor] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(3461),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [anon_sym_for] = ACTIONS(3651),
    [anon_sym_foreach] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3649),
    [anon_sym___END__] = ACTIONS(3649),
    [anon_sym_EOT] = ACTIONS(3649),
    [anon_sym_if] = ACTIONS(3649),
    [anon_sym_unless] = ACTIONS(3649),
    [anon_sym_while] = ACTIONS(3649),
    [anon_sym_until] = ACTIONS(3649),
    [sym__PERLY_SEMICOLON] = ACTIONS(3649),
    [sym__ctrl_z_hack] = ACTIONS(3649),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1808)] = {
    [sym_heredoc_content] = STATE(1808),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [anon_sym_for] = ACTIONS(3437),
    [anon_sym_foreach] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3435),
    [anon_sym___END__] = ACTIONS(3435),
    [anon_sym_EOT] = ACTIONS(3435),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [anon_sym_if] = ACTIONS(3437),
    [anon_sym_unless] = ACTIONS(3437),
    [anon_sym_while] = ACTIONS(3435),
    [anon_sym_until] = ACTIONS(3437),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym__ctrl_z_hack] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1809)] = {
    [sym_heredoc_content] = STATE(1809),
    [aux_sym__term_rightward_repeat1] = STATE(3994),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(3653),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3653),
    [anon_sym_for] = ACTIONS(3536),
    [anon_sym_foreach] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3485),
    [anon_sym___END__] = ACTIONS(3485),
    [anon_sym_EOT] = ACTIONS(3485),
    [anon_sym_if] = ACTIONS(3485),
    [anon_sym_unless] = ACTIONS(3485),
    [anon_sym_while] = ACTIONS(3485),
    [anon_sym_until] = ACTIONS(3485),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym__ctrl_z_hack] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1810)] = {
    [sym_heredoc_content] = STATE(1810),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [anon_sym_for] = ACTIONS(3437),
    [anon_sym_foreach] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3435),
    [anon_sym___END__] = ACTIONS(3435),
    [anon_sym_EOT] = ACTIONS(3435),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [anon_sym_if] = ACTIONS(3437),
    [anon_sym_unless] = ACTIONS(3437),
    [anon_sym_while] = ACTIONS(3435),
    [anon_sym_until] = ACTIONS(3437),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym__ctrl_z_hack] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1811)] = {
    [sym_heredoc_content] = STATE(1811),
    [anon_sym_RBRACE] = ACTIONS(3655),
    [anon_sym_SEMI] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_RPAREN] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_and] = ACTIONS(3657),
    [anon_sym_or] = ACTIONS(3657),
    [anon_sym_xor] = ACTIONS(3657),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [anon_sym_for] = ACTIONS(3657),
    [anon_sym_foreach] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3655),
    [anon_sym___END__] = ACTIONS(3655),
    [anon_sym_EOT] = ACTIONS(3655),
    [sym_match_regexp_modifiers] = ACTIONS(3659),
    [anon_sym_if] = ACTIONS(3657),
    [anon_sym_unless] = ACTIONS(3657),
    [anon_sym_while] = ACTIONS(3655),
    [anon_sym_until] = ACTIONS(3657),
    [sym__PERLY_SEMICOLON] = ACTIONS(3655),
    [sym__ctrl_z_hack] = ACTIONS(3655),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1812)] = {
    [sym_heredoc_content] = STATE(1812),
    [anon_sym_RBRACE] = ACTIONS(3661),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_SEMI] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_RPAREN] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_and] = ACTIONS(3661),
    [anon_sym_or] = ACTIONS(3661),
    [anon_sym_xor] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [anon_sym_for] = ACTIONS(3663),
    [anon_sym_foreach] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3661),
    [anon_sym___END__] = ACTIONS(3661),
    [anon_sym_EOT] = ACTIONS(3661),
    [anon_sym_if] = ACTIONS(3661),
    [anon_sym_unless] = ACTIONS(3661),
    [anon_sym_while] = ACTIONS(3661),
    [anon_sym_until] = ACTIONS(3661),
    [sym__PERLY_SEMICOLON] = ACTIONS(3661),
    [sym__ctrl_z_hack] = ACTIONS(3661),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1813)] = {
    [sym_heredoc_content] = STATE(1813),
    [anon_sym_RBRACE] = ACTIONS(3665),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_SEMI] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_RPAREN] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_and] = ACTIONS(3665),
    [anon_sym_or] = ACTIONS(3665),
    [anon_sym_xor] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [anon_sym_for] = ACTIONS(3667),
    [anon_sym_foreach] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3665),
    [anon_sym___END__] = ACTIONS(3665),
    [anon_sym_EOT] = ACTIONS(3665),
    [anon_sym_if] = ACTIONS(3665),
    [anon_sym_unless] = ACTIONS(3665),
    [anon_sym_while] = ACTIONS(3665),
    [anon_sym_until] = ACTIONS(3665),
    [sym__PERLY_SEMICOLON] = ACTIONS(3665),
    [sym__ctrl_z_hack] = ACTIONS(3665),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1814)] = {
    [sym_heredoc_content] = STATE(1814),
    [anon_sym_RBRACE] = ACTIONS(3669),
    [anon_sym_COLON] = ACTIONS(3671),
    [anon_sym_SEMI] = ACTIONS(3669),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_RPAREN] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [anon_sym_for] = ACTIONS(3673),
    [anon_sym_foreach] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3669),
    [anon_sym___END__] = ACTIONS(3669),
    [anon_sym_EOT] = ACTIONS(3669),
    [anon_sym_if] = ACTIONS(3669),
    [anon_sym_unless] = ACTIONS(3669),
    [anon_sym_while] = ACTIONS(3669),
    [anon_sym_until] = ACTIONS(3669),
    [sym__PERLY_SEMICOLON] = ACTIONS(3669),
    [sym__ctrl_z_hack] = ACTIONS(3669),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1815)] = {
    [sym_heredoc_content] = STATE(1815),
    [anon_sym_RBRACE] = ACTIONS(3675),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_SEMI] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_RPAREN] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_and] = ACTIONS(3675),
    [anon_sym_or] = ACTIONS(3675),
    [anon_sym_xor] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [anon_sym_for] = ACTIONS(3677),
    [anon_sym_foreach] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3675),
    [anon_sym___END__] = ACTIONS(3675),
    [anon_sym_EOT] = ACTIONS(3675),
    [anon_sym_if] = ACTIONS(3675),
    [anon_sym_unless] = ACTIONS(3675),
    [anon_sym_while] = ACTIONS(3675),
    [anon_sym_until] = ACTIONS(3675),
    [sym__PERLY_SEMICOLON] = ACTIONS(3675),
    [sym__ctrl_z_hack] = ACTIONS(3675),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1816)] = {
    [sym_heredoc_content] = STATE(1816),
    [anon_sym_RBRACE] = ACTIONS(3679),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_SEMI] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_RPAREN] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_and] = ACTIONS(3679),
    [anon_sym_or] = ACTIONS(3679),
    [anon_sym_xor] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [anon_sym_for] = ACTIONS(3681),
    [anon_sym_foreach] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3679),
    [anon_sym___END__] = ACTIONS(3679),
    [anon_sym_EOT] = ACTIONS(3679),
    [anon_sym_if] = ACTIONS(3679),
    [anon_sym_unless] = ACTIONS(3679),
    [anon_sym_while] = ACTIONS(3679),
    [anon_sym_until] = ACTIONS(3679),
    [sym__PERLY_SEMICOLON] = ACTIONS(3679),
    [sym__ctrl_z_hack] = ACTIONS(3679),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1817)] = {
    [sym_heredoc_content] = STATE(1817),
    [anon_sym_RBRACE] = ACTIONS(3683),
    [anon_sym_SEMI] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_RPAREN] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_and] = ACTIONS(3683),
    [anon_sym_or] = ACTIONS(3683),
    [anon_sym_xor] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [anon_sym_for] = ACTIONS(3685),
    [anon_sym_foreach] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3683),
    [anon_sym___END__] = ACTIONS(3683),
    [anon_sym_EOT] = ACTIONS(3683),
    [sym_transliteration_modifiers] = ACTIONS(3687),
    [anon_sym_if] = ACTIONS(3683),
    [anon_sym_unless] = ACTIONS(3683),
    [anon_sym_while] = ACTIONS(3683),
    [anon_sym_until] = ACTIONS(3683),
    [sym__PERLY_SEMICOLON] = ACTIONS(3683),
    [sym__ctrl_z_hack] = ACTIONS(3683),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1818)] = {
    [sym_heredoc_content] = STATE(1818),
    [anon_sym_RBRACE] = ACTIONS(3689),
    [anon_sym_COLON] = ACTIONS(3691),
    [anon_sym_SEMI] = ACTIONS(3689),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_RPAREN] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [anon_sym_for] = ACTIONS(3693),
    [anon_sym_foreach] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3689),
    [anon_sym___END__] = ACTIONS(3689),
    [anon_sym_EOT] = ACTIONS(3689),
    [anon_sym_if] = ACTIONS(3689),
    [anon_sym_unless] = ACTIONS(3689),
    [anon_sym_while] = ACTIONS(3689),
    [anon_sym_until] = ACTIONS(3689),
    [sym__PERLY_SEMICOLON] = ACTIONS(3689),
    [sym__ctrl_z_hack] = ACTIONS(3689),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1819)] = {
    [sym_heredoc_content] = STATE(1819),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_SEMI] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(3625),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [anon_sym_for] = ACTIONS(3697),
    [anon_sym_foreach] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3695),
    [anon_sym___END__] = ACTIONS(3695),
    [anon_sym_EOT] = ACTIONS(3695),
    [anon_sym_if] = ACTIONS(3695),
    [anon_sym_unless] = ACTIONS(3695),
    [anon_sym_while] = ACTIONS(3695),
    [anon_sym_until] = ACTIONS(3695),
    [sym__PERLY_SEMICOLON] = ACTIONS(3695),
    [sym__ctrl_z_hack] = ACTIONS(3695),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1820)] = {
    [sym_heredoc_content] = STATE(1820),
    [anon_sym_RBRACE] = ACTIONS(3699),
    [anon_sym_SEMI] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_RPAREN] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_and] = ACTIONS(3699),
    [anon_sym_or] = ACTIONS(3699),
    [anon_sym_xor] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [anon_sym_for] = ACTIONS(3701),
    [anon_sym_foreach] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3699),
    [anon_sym___END__] = ACTIONS(3699),
    [anon_sym_EOT] = ACTIONS(3699),
    [anon_sym_if] = ACTIONS(3699),
    [anon_sym_unless] = ACTIONS(3699),
    [anon_sym_while] = ACTIONS(3699),
    [anon_sym_until] = ACTIONS(3699),
    [sym__PERLY_SEMICOLON] = ACTIONS(3699),
    [sym__ctrl_z_hack] = ACTIONS(3699),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1821)] = {
    [sym_heredoc_content] = STATE(1821),
    [anon_sym_RBRACE] = ACTIONS(3703),
    [anon_sym_SEMI] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_RPAREN] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_and] = ACTIONS(3705),
    [anon_sym_or] = ACTIONS(3705),
    [anon_sym_xor] = ACTIONS(3705),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [anon_sym_for] = ACTIONS(3705),
    [anon_sym_foreach] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3703),
    [anon_sym___END__] = ACTIONS(3703),
    [anon_sym_EOT] = ACTIONS(3703),
    [sym_quoted_regexp_modifiers] = ACTIONS(3707),
    [anon_sym_if] = ACTIONS(3705),
    [anon_sym_unless] = ACTIONS(3705),
    [anon_sym_while] = ACTIONS(3703),
    [anon_sym_until] = ACTIONS(3705),
    [sym__PERLY_SEMICOLON] = ACTIONS(3703),
    [sym__ctrl_z_hack] = ACTIONS(3703),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1822)] = {
    [sym_heredoc_content] = STATE(1822),
    [anon_sym_RBRACE] = ACTIONS(3709),
    [anon_sym_SEMI] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_RPAREN] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_and] = ACTIONS(3711),
    [anon_sym_or] = ACTIONS(3711),
    [anon_sym_xor] = ACTIONS(3711),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [anon_sym_for] = ACTIONS(3711),
    [anon_sym_foreach] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3709),
    [anon_sym___END__] = ACTIONS(3709),
    [anon_sym_EOT] = ACTIONS(3709),
    [sym_match_regexp_modifiers] = ACTIONS(3713),
    [anon_sym_if] = ACTIONS(3711),
    [anon_sym_unless] = ACTIONS(3711),
    [anon_sym_while] = ACTIONS(3709),
    [anon_sym_until] = ACTIONS(3711),
    [sym__PERLY_SEMICOLON] = ACTIONS(3709),
    [sym__ctrl_z_hack] = ACTIONS(3709),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1823)] = {
    [sym_heredoc_content] = STATE(1823),
    [anon_sym_RBRACE] = ACTIONS(3715),
    [anon_sym_SEMI] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_RPAREN] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_and] = ACTIONS(3715),
    [anon_sym_or] = ACTIONS(3715),
    [anon_sym_xor] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [anon_sym_for] = ACTIONS(3717),
    [anon_sym_foreach] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3715),
    [anon_sym___END__] = ACTIONS(3715),
    [anon_sym_EOT] = ACTIONS(3715),
    [anon_sym_if] = ACTIONS(3715),
    [anon_sym_unless] = ACTIONS(3715),
    [anon_sym_while] = ACTIONS(3715),
    [anon_sym_until] = ACTIONS(3715),
    [sym__PERLY_SEMICOLON] = ACTIONS(3715),
    [sym__ctrl_z_hack] = ACTIONS(3715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1824)] = {
    [sym_heredoc_content] = STATE(1824),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1825)] = {
    [sym_heredoc_content] = STATE(1825),
    [anon_sym_RBRACE] = ACTIONS(3723),
    [anon_sym_SEMI] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_RPAREN] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_and] = ACTIONS(3723),
    [anon_sym_or] = ACTIONS(3723),
    [anon_sym_xor] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [anon_sym_for] = ACTIONS(3725),
    [anon_sym_foreach] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3723),
    [anon_sym___END__] = ACTIONS(3723),
    [anon_sym_EOT] = ACTIONS(3723),
    [anon_sym_if] = ACTIONS(3723),
    [anon_sym_unless] = ACTIONS(3723),
    [anon_sym_while] = ACTIONS(3723),
    [anon_sym_until] = ACTIONS(3723),
    [sym__PERLY_SEMICOLON] = ACTIONS(3723),
    [sym__ctrl_z_hack] = ACTIONS(3723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1826)] = {
    [sym_heredoc_content] = STATE(1826),
    [anon_sym_RBRACE] = ACTIONS(3727),
    [anon_sym_SEMI] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_RPAREN] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_and] = ACTIONS(3727),
    [anon_sym_or] = ACTIONS(3727),
    [anon_sym_xor] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [anon_sym_for] = ACTIONS(3729),
    [anon_sym_foreach] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3727),
    [anon_sym___END__] = ACTIONS(3727),
    [anon_sym_EOT] = ACTIONS(3727),
    [anon_sym_if] = ACTIONS(3727),
    [anon_sym_unless] = ACTIONS(3727),
    [anon_sym_while] = ACTIONS(3727),
    [anon_sym_until] = ACTIONS(3727),
    [sym__PERLY_SEMICOLON] = ACTIONS(3727),
    [sym__ctrl_z_hack] = ACTIONS(3727),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1827)] = {
    [sym_heredoc_content] = STATE(1827),
    [anon_sym_RBRACE] = ACTIONS(3731),
    [anon_sym_SEMI] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_RPAREN] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_and] = ACTIONS(3731),
    [anon_sym_or] = ACTIONS(3731),
    [anon_sym_xor] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [anon_sym_for] = ACTIONS(3733),
    [anon_sym_foreach] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3731),
    [anon_sym___END__] = ACTIONS(3731),
    [anon_sym_EOT] = ACTIONS(3731),
    [anon_sym_if] = ACTIONS(3731),
    [anon_sym_unless] = ACTIONS(3731),
    [anon_sym_while] = ACTIONS(3731),
    [anon_sym_until] = ACTIONS(3731),
    [sym__PERLY_SEMICOLON] = ACTIONS(3731),
    [sym__ctrl_z_hack] = ACTIONS(3731),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1828)] = {
    [sym_heredoc_content] = STATE(1828),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [anon_sym_for] = ACTIONS(3737),
    [anon_sym_foreach] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3735),
    [anon_sym___END__] = ACTIONS(3735),
    [anon_sym_EOT] = ACTIONS(3735),
    [anon_sym_if] = ACTIONS(3735),
    [anon_sym_unless] = ACTIONS(3735),
    [anon_sym_while] = ACTIONS(3735),
    [anon_sym_until] = ACTIONS(3735),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym__ctrl_z_hack] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1829)] = {
    [sym_heredoc_content] = STATE(1829),
    [anon_sym_RBRACE] = ACTIONS(3739),
    [anon_sym_SEMI] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [anon_sym_for] = ACTIONS(3741),
    [anon_sym_foreach] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3739),
    [anon_sym___END__] = ACTIONS(3739),
    [anon_sym_EOT] = ACTIONS(3739),
    [anon_sym_if] = ACTIONS(3739),
    [anon_sym_unless] = ACTIONS(3739),
    [anon_sym_while] = ACTIONS(3739),
    [anon_sym_until] = ACTIONS(3739),
    [sym__PERLY_SEMICOLON] = ACTIONS(3739),
    [sym__ctrl_z_hack] = ACTIONS(3739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1830)] = {
    [sym_heredoc_content] = STATE(1830),
    [anon_sym_RBRACE] = ACTIONS(3743),
    [anon_sym_SEMI] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_RPAREN] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_and] = ACTIONS(3743),
    [anon_sym_or] = ACTIONS(3743),
    [anon_sym_xor] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [anon_sym_for] = ACTIONS(3745),
    [anon_sym_foreach] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3743),
    [anon_sym___END__] = ACTIONS(3743),
    [anon_sym_EOT] = ACTIONS(3743),
    [anon_sym_if] = ACTIONS(3743),
    [anon_sym_unless] = ACTIONS(3743),
    [anon_sym_while] = ACTIONS(3743),
    [anon_sym_until] = ACTIONS(3743),
    [sym__PERLY_SEMICOLON] = ACTIONS(3743),
    [sym__ctrl_z_hack] = ACTIONS(3743),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1831)] = {
    [sym_heredoc_content] = STATE(1831),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1832)] = {
    [sym_heredoc_content] = STATE(1832),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1833)] = {
    [sym_heredoc_content] = STATE(1833),
    [anon_sym_RBRACE] = ACTIONS(3747),
    [anon_sym_SEMI] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_RPAREN] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_and] = ACTIONS(3747),
    [anon_sym_or] = ACTIONS(3747),
    [anon_sym_xor] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [anon_sym_for] = ACTIONS(3749),
    [anon_sym_foreach] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3747),
    [anon_sym___END__] = ACTIONS(3747),
    [anon_sym_EOT] = ACTIONS(3747),
    [anon_sym_if] = ACTIONS(3747),
    [anon_sym_unless] = ACTIONS(3747),
    [anon_sym_while] = ACTIONS(3747),
    [anon_sym_until] = ACTIONS(3747),
    [sym__PERLY_SEMICOLON] = ACTIONS(3747),
    [sym__ctrl_z_hack] = ACTIONS(3747),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1834)] = {
    [sym_heredoc_content] = STATE(1834),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1835)] = {
    [sym_heredoc_content] = STATE(1835),
    [anon_sym_RBRACE] = ACTIONS(3751),
    [anon_sym_SEMI] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_RPAREN] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_and] = ACTIONS(3751),
    [anon_sym_or] = ACTIONS(3751),
    [anon_sym_xor] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [anon_sym_for] = ACTIONS(3753),
    [anon_sym_foreach] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3751),
    [anon_sym___END__] = ACTIONS(3751),
    [anon_sym_EOT] = ACTIONS(3751),
    [anon_sym_if] = ACTIONS(3751),
    [anon_sym_unless] = ACTIONS(3751),
    [anon_sym_while] = ACTIONS(3751),
    [anon_sym_until] = ACTIONS(3751),
    [sym__PERLY_SEMICOLON] = ACTIONS(3751),
    [sym__ctrl_z_hack] = ACTIONS(3751),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1836)] = {
    [sym_heredoc_content] = STATE(1836),
    [anon_sym_RBRACE] = ACTIONS(3755),
    [anon_sym_SEMI] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_RPAREN] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_and] = ACTIONS(3755),
    [anon_sym_or] = ACTIONS(3755),
    [anon_sym_xor] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [anon_sym_for] = ACTIONS(3757),
    [anon_sym_foreach] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3755),
    [anon_sym___END__] = ACTIONS(3755),
    [anon_sym_EOT] = ACTIONS(3755),
    [anon_sym_if] = ACTIONS(3755),
    [anon_sym_unless] = ACTIONS(3755),
    [anon_sym_while] = ACTIONS(3755),
    [anon_sym_until] = ACTIONS(3755),
    [sym__PERLY_SEMICOLON] = ACTIONS(3755),
    [sym__ctrl_z_hack] = ACTIONS(3755),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1837)] = {
    [sym_heredoc_content] = STATE(1837),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1838)] = {
    [sym_heredoc_content] = STATE(1838),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1839)] = {
    [sym_heredoc_content] = STATE(1839),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1840)] = {
    [sym_heredoc_content] = STATE(1840),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1841)] = {
    [sym_heredoc_content] = STATE(1841),
    [anon_sym_RBRACE] = ACTIONS(3759),
    [anon_sym_SEMI] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_RPAREN] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_and] = ACTIONS(3759),
    [anon_sym_or] = ACTIONS(3759),
    [anon_sym_xor] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [anon_sym_for] = ACTIONS(3761),
    [anon_sym_foreach] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3759),
    [anon_sym___END__] = ACTIONS(3759),
    [anon_sym_EOT] = ACTIONS(3759),
    [anon_sym_if] = ACTIONS(3759),
    [anon_sym_unless] = ACTIONS(3759),
    [anon_sym_while] = ACTIONS(3759),
    [anon_sym_until] = ACTIONS(3759),
    [sym__PERLY_SEMICOLON] = ACTIONS(3759),
    [sym__ctrl_z_hack] = ACTIONS(3759),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1842)] = {
    [sym_heredoc_content] = STATE(1842),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1843)] = {
    [sym_heredoc_content] = STATE(1843),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [anon_sym_for] = ACTIONS(3765),
    [anon_sym_foreach] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3763),
    [anon_sym___END__] = ACTIONS(3763),
    [anon_sym_EOT] = ACTIONS(3763),
    [anon_sym_if] = ACTIONS(3763),
    [anon_sym_unless] = ACTIONS(3763),
    [anon_sym_while] = ACTIONS(3763),
    [anon_sym_until] = ACTIONS(3763),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym__ctrl_z_hack] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1844)] = {
    [sym_heredoc_content] = STATE(1844),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3767),
    [anon_sym_cmp] = ACTIONS(3767),
    [anon_sym_TILDE_TILDE] = ACTIONS(3767),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [anon_sym_for] = ACTIONS(3765),
    [anon_sym_foreach] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3763),
    [anon_sym___END__] = ACTIONS(3763),
    [anon_sym_EOT] = ACTIONS(3763),
    [anon_sym_if] = ACTIONS(3763),
    [anon_sym_unless] = ACTIONS(3763),
    [anon_sym_while] = ACTIONS(3763),
    [anon_sym_until] = ACTIONS(3763),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym__ctrl_z_hack] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1845)] = {
    [sym_heredoc_content] = STATE(1845),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3769),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [anon_sym_for] = ACTIONS(3737),
    [anon_sym_foreach] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3735),
    [anon_sym___END__] = ACTIONS(3735),
    [anon_sym_EOT] = ACTIONS(3735),
    [anon_sym_if] = ACTIONS(3735),
    [anon_sym_unless] = ACTIONS(3735),
    [anon_sym_while] = ACTIONS(3735),
    [anon_sym_until] = ACTIONS(3735),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym__ctrl_z_hack] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1846)] = {
    [sym_heredoc_content] = STATE(1846),
    [anon_sym_RBRACE] = ACTIONS(3771),
    [anon_sym_SEMI] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_RPAREN] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_and] = ACTIONS(3771),
    [anon_sym_or] = ACTIONS(3771),
    [anon_sym_xor] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [anon_sym_for] = ACTIONS(3773),
    [anon_sym_foreach] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3771),
    [anon_sym___END__] = ACTIONS(3771),
    [anon_sym_EOT] = ACTIONS(3771),
    [anon_sym_if] = ACTIONS(3771),
    [anon_sym_unless] = ACTIONS(3771),
    [anon_sym_while] = ACTIONS(3771),
    [anon_sym_until] = ACTIONS(3771),
    [sym__PERLY_SEMICOLON] = ACTIONS(3771),
    [sym__ctrl_z_hack] = ACTIONS(3771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1847)] = {
    [sym_heredoc_content] = STATE(1847),
    [anon_sym_RBRACE] = ACTIONS(3775),
    [anon_sym_SEMI] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_RPAREN] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_and] = ACTIONS(3775),
    [anon_sym_or] = ACTIONS(3775),
    [anon_sym_xor] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [anon_sym_for] = ACTIONS(3777),
    [anon_sym_foreach] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3775),
    [anon_sym___END__] = ACTIONS(3775),
    [anon_sym_EOT] = ACTIONS(3775),
    [anon_sym_if] = ACTIONS(3775),
    [anon_sym_unless] = ACTIONS(3775),
    [anon_sym_while] = ACTIONS(3775),
    [anon_sym_until] = ACTIONS(3775),
    [sym__PERLY_SEMICOLON] = ACTIONS(3775),
    [sym__ctrl_z_hack] = ACTIONS(3775),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1848)] = {
    [sym_heredoc_content] = STATE(1848),
    [anon_sym_RBRACE] = ACTIONS(3779),
    [anon_sym_SEMI] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_RPAREN] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_and] = ACTIONS(3779),
    [anon_sym_or] = ACTIONS(3779),
    [anon_sym_xor] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [anon_sym_for] = ACTIONS(3781),
    [anon_sym_foreach] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3779),
    [anon_sym___END__] = ACTIONS(3779),
    [anon_sym_EOT] = ACTIONS(3779),
    [anon_sym_if] = ACTIONS(3779),
    [anon_sym_unless] = ACTIONS(3779),
    [anon_sym_while] = ACTIONS(3779),
    [anon_sym_until] = ACTIONS(3779),
    [sym__PERLY_SEMICOLON] = ACTIONS(3779),
    [sym__ctrl_z_hack] = ACTIONS(3779),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1849)] = {
    [sym_heredoc_content] = STATE(1849),
    [anon_sym_RBRACE] = ACTIONS(3783),
    [anon_sym_SEMI] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_RPAREN] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_and] = ACTIONS(3783),
    [anon_sym_or] = ACTIONS(3783),
    [anon_sym_xor] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [anon_sym_for] = ACTIONS(3785),
    [anon_sym_foreach] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3783),
    [anon_sym___END__] = ACTIONS(3783),
    [anon_sym_EOT] = ACTIONS(3783),
    [anon_sym_if] = ACTIONS(3783),
    [anon_sym_unless] = ACTIONS(3783),
    [anon_sym_while] = ACTIONS(3783),
    [anon_sym_until] = ACTIONS(3783),
    [sym__PERLY_SEMICOLON] = ACTIONS(3783),
    [sym__ctrl_z_hack] = ACTIONS(3783),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1850)] = {
    [sym_heredoc_content] = STATE(1850),
    [anon_sym_RBRACE] = ACTIONS(3787),
    [anon_sym_SEMI] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_RPAREN] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_and] = ACTIONS(3787),
    [anon_sym_or] = ACTIONS(3787),
    [anon_sym_xor] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [anon_sym_for] = ACTIONS(3789),
    [anon_sym_foreach] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3787),
    [anon_sym___END__] = ACTIONS(3787),
    [anon_sym_EOT] = ACTIONS(3787),
    [anon_sym_if] = ACTIONS(3787),
    [anon_sym_unless] = ACTIONS(3787),
    [anon_sym_while] = ACTIONS(3787),
    [anon_sym_until] = ACTIONS(3787),
    [sym__PERLY_SEMICOLON] = ACTIONS(3787),
    [sym__ctrl_z_hack] = ACTIONS(3787),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1851)] = {
    [sym_heredoc_content] = STATE(1851),
    [anon_sym_RBRACE] = ACTIONS(3791),
    [anon_sym_SEMI] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_RPAREN] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_and] = ACTIONS(3791),
    [anon_sym_or] = ACTIONS(3791),
    [anon_sym_xor] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [anon_sym_for] = ACTIONS(3793),
    [anon_sym_foreach] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3791),
    [anon_sym___END__] = ACTIONS(3791),
    [anon_sym_EOT] = ACTIONS(3791),
    [anon_sym_if] = ACTIONS(3791),
    [anon_sym_unless] = ACTIONS(3791),
    [anon_sym_while] = ACTIONS(3791),
    [anon_sym_until] = ACTIONS(3791),
    [sym__PERLY_SEMICOLON] = ACTIONS(3791),
    [sym__ctrl_z_hack] = ACTIONS(3791),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1852)] = {
    [sym_heredoc_content] = STATE(1852),
    [anon_sym_RBRACE] = ACTIONS(3795),
    [anon_sym_SEMI] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_RPAREN] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_and] = ACTIONS(3795),
    [anon_sym_or] = ACTIONS(3795),
    [anon_sym_xor] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [anon_sym_for] = ACTIONS(3797),
    [anon_sym_foreach] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3795),
    [anon_sym___END__] = ACTIONS(3795),
    [anon_sym_EOT] = ACTIONS(3795),
    [anon_sym_if] = ACTIONS(3795),
    [anon_sym_unless] = ACTIONS(3795),
    [anon_sym_while] = ACTIONS(3795),
    [anon_sym_until] = ACTIONS(3795),
    [sym__PERLY_SEMICOLON] = ACTIONS(3795),
    [sym__ctrl_z_hack] = ACTIONS(3795),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1853)] = {
    [sym_heredoc_content] = STATE(1853),
    [anon_sym_RBRACE] = ACTIONS(3799),
    [anon_sym_SEMI] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_RPAREN] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_and] = ACTIONS(3799),
    [anon_sym_or] = ACTIONS(3799),
    [anon_sym_xor] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [anon_sym_for] = ACTIONS(3801),
    [anon_sym_foreach] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3799),
    [anon_sym___END__] = ACTIONS(3799),
    [anon_sym_EOT] = ACTIONS(3799),
    [anon_sym_if] = ACTIONS(3799),
    [anon_sym_unless] = ACTIONS(3799),
    [anon_sym_while] = ACTIONS(3799),
    [anon_sym_until] = ACTIONS(3799),
    [sym__PERLY_SEMICOLON] = ACTIONS(3799),
    [sym__ctrl_z_hack] = ACTIONS(3799),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1854)] = {
    [sym_heredoc_content] = STATE(1854),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_for] = ACTIONS(2702),
    [anon_sym_foreach] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2711),
    [anon_sym___END__] = ACTIONS(2711),
    [anon_sym_EOT] = ACTIONS(2711),
    [anon_sym_if] = ACTIONS(2711),
    [anon_sym_unless] = ACTIONS(2711),
    [anon_sym_while] = ACTIONS(2711),
    [anon_sym_until] = ACTIONS(2711),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym__ctrl_z_hack] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1855)] = {
    [sym_heredoc_content] = STATE(1855),
    [anon_sym_RBRACE] = ACTIONS(3803),
    [anon_sym_SEMI] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [anon_sym_for] = ACTIONS(3805),
    [anon_sym_foreach] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3803),
    [anon_sym___END__] = ACTIONS(3803),
    [anon_sym_EOT] = ACTIONS(3803),
    [anon_sym_if] = ACTIONS(3803),
    [anon_sym_unless] = ACTIONS(3803),
    [anon_sym_while] = ACTIONS(3803),
    [anon_sym_until] = ACTIONS(3803),
    [sym__PERLY_SEMICOLON] = ACTIONS(3803),
    [sym__ctrl_z_hack] = ACTIONS(3803),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1856)] = {
    [sym_heredoc_content] = STATE(1856),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_SEMI] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [anon_sym_for] = ACTIONS(3697),
    [anon_sym_foreach] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3695),
    [anon_sym___END__] = ACTIONS(3695),
    [anon_sym_EOT] = ACTIONS(3695),
    [anon_sym_if] = ACTIONS(3695),
    [anon_sym_unless] = ACTIONS(3695),
    [anon_sym_while] = ACTIONS(3695),
    [anon_sym_until] = ACTIONS(3695),
    [sym__PERLY_SEMICOLON] = ACTIONS(3695),
    [sym__ctrl_z_hack] = ACTIONS(3695),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1857)] = {
    [sym_heredoc_content] = STATE(1857),
    [anon_sym_RBRACE] = ACTIONS(3807),
    [anon_sym_SEMI] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_RPAREN] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_and] = ACTIONS(3807),
    [anon_sym_or] = ACTIONS(3807),
    [anon_sym_xor] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [anon_sym_for] = ACTIONS(3809),
    [anon_sym_foreach] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3807),
    [anon_sym___END__] = ACTIONS(3807),
    [anon_sym_EOT] = ACTIONS(3807),
    [anon_sym_if] = ACTIONS(3807),
    [anon_sym_unless] = ACTIONS(3807),
    [anon_sym_while] = ACTIONS(3807),
    [anon_sym_until] = ACTIONS(3807),
    [sym__PERLY_SEMICOLON] = ACTIONS(3807),
    [sym__ctrl_z_hack] = ACTIONS(3807),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1858)] = {
    [sym_heredoc_content] = STATE(1858),
    [anon_sym_RBRACE] = ACTIONS(3811),
    [anon_sym_SEMI] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_RPAREN] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_and] = ACTIONS(3811),
    [anon_sym_or] = ACTIONS(3811),
    [anon_sym_xor] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [anon_sym_for] = ACTIONS(3813),
    [anon_sym_foreach] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3811),
    [anon_sym___END__] = ACTIONS(3811),
    [anon_sym_EOT] = ACTIONS(3811),
    [anon_sym_if] = ACTIONS(3811),
    [anon_sym_unless] = ACTIONS(3811),
    [anon_sym_while] = ACTIONS(3811),
    [anon_sym_until] = ACTIONS(3811),
    [sym__PERLY_SEMICOLON] = ACTIONS(3811),
    [sym__ctrl_z_hack] = ACTIONS(3811),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1859)] = {
    [sym_heredoc_content] = STATE(1859),
    [anon_sym_RBRACE] = ACTIONS(3815),
    [anon_sym_SEMI] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_RPAREN] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_and] = ACTIONS(3815),
    [anon_sym_or] = ACTIONS(3815),
    [anon_sym_xor] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [anon_sym_for] = ACTIONS(3817),
    [anon_sym_foreach] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3815),
    [anon_sym___END__] = ACTIONS(3815),
    [anon_sym_EOT] = ACTIONS(3815),
    [anon_sym_if] = ACTIONS(3815),
    [anon_sym_unless] = ACTIONS(3815),
    [anon_sym_while] = ACTIONS(3815),
    [anon_sym_until] = ACTIONS(3815),
    [sym__PERLY_SEMICOLON] = ACTIONS(3815),
    [sym__ctrl_z_hack] = ACTIONS(3815),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1860)] = {
    [sym_heredoc_content] = STATE(1860),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_SEMI] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3823),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3823),
    [anon_sym_for] = ACTIONS(3821),
    [anon_sym_foreach] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3819),
    [anon_sym___END__] = ACTIONS(3819),
    [anon_sym_EOT] = ACTIONS(3819),
    [anon_sym_if] = ACTIONS(3819),
    [anon_sym_unless] = ACTIONS(3819),
    [anon_sym_while] = ACTIONS(3819),
    [anon_sym_until] = ACTIONS(3819),
    [sym__PERLY_SEMICOLON] = ACTIONS(3819),
    [sym__ctrl_z_hack] = ACTIONS(3819),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1861)] = {
    [sym_heredoc_content] = STATE(1861),
    [anon_sym_RBRACE] = ACTIONS(3825),
    [anon_sym_SEMI] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_RPAREN] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_and] = ACTIONS(3825),
    [anon_sym_or] = ACTIONS(3825),
    [anon_sym_xor] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [anon_sym_for] = ACTIONS(3827),
    [anon_sym_foreach] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3825),
    [anon_sym___END__] = ACTIONS(3825),
    [anon_sym_EOT] = ACTIONS(3825),
    [anon_sym_if] = ACTIONS(3825),
    [anon_sym_unless] = ACTIONS(3825),
    [anon_sym_while] = ACTIONS(3825),
    [anon_sym_until] = ACTIONS(3825),
    [sym__PERLY_SEMICOLON] = ACTIONS(3825),
    [sym__ctrl_z_hack] = ACTIONS(3825),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1862)] = {
    [sym_heredoc_content] = STATE(1862),
    [anon_sym_RBRACE] = ACTIONS(3829),
    [anon_sym_SEMI] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_RPAREN] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_and] = ACTIONS(3829),
    [anon_sym_or] = ACTIONS(3829),
    [anon_sym_xor] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [anon_sym_for] = ACTIONS(3831),
    [anon_sym_foreach] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3829),
    [anon_sym___END__] = ACTIONS(3829),
    [anon_sym_EOT] = ACTIONS(3829),
    [anon_sym_if] = ACTIONS(3829),
    [anon_sym_unless] = ACTIONS(3829),
    [anon_sym_while] = ACTIONS(3829),
    [anon_sym_until] = ACTIONS(3829),
    [sym__PERLY_SEMICOLON] = ACTIONS(3829),
    [sym__ctrl_z_hack] = ACTIONS(3829),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1863)] = {
    [sym_heredoc_content] = STATE(1863),
    [anon_sym_RBRACE] = ACTIONS(3833),
    [anon_sym_SEMI] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_RPAREN] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_and] = ACTIONS(3833),
    [anon_sym_or] = ACTIONS(3833),
    [anon_sym_xor] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [anon_sym_for] = ACTIONS(3835),
    [anon_sym_foreach] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3833),
    [anon_sym___END__] = ACTIONS(3833),
    [anon_sym_EOT] = ACTIONS(3833),
    [anon_sym_if] = ACTIONS(3833),
    [anon_sym_unless] = ACTIONS(3833),
    [anon_sym_while] = ACTIONS(3833),
    [anon_sym_until] = ACTIONS(3833),
    [sym__PERLY_SEMICOLON] = ACTIONS(3833),
    [sym__ctrl_z_hack] = ACTIONS(3833),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1864)] = {
    [sym_heredoc_content] = STATE(1864),
    [anon_sym_RBRACE] = ACTIONS(3837),
    [anon_sym_SEMI] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_RPAREN] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_and] = ACTIONS(3837),
    [anon_sym_or] = ACTIONS(3837),
    [anon_sym_xor] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [anon_sym_for] = ACTIONS(3839),
    [anon_sym_foreach] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3837),
    [anon_sym___END__] = ACTIONS(3837),
    [anon_sym_EOT] = ACTIONS(3837),
    [anon_sym_if] = ACTIONS(3837),
    [anon_sym_unless] = ACTIONS(3837),
    [anon_sym_while] = ACTIONS(3837),
    [anon_sym_until] = ACTIONS(3837),
    [sym__PERLY_SEMICOLON] = ACTIONS(3837),
    [sym__ctrl_z_hack] = ACTIONS(3837),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1865)] = {
    [sym_heredoc_content] = STATE(1865),
    [anon_sym_RBRACE] = ACTIONS(3841),
    [anon_sym_SEMI] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_RPAREN] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_and] = ACTIONS(3841),
    [anon_sym_or] = ACTIONS(3841),
    [anon_sym_xor] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [anon_sym_for] = ACTIONS(3843),
    [anon_sym_foreach] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3841),
    [anon_sym___END__] = ACTIONS(3841),
    [anon_sym_EOT] = ACTIONS(3841),
    [anon_sym_if] = ACTIONS(3841),
    [anon_sym_unless] = ACTIONS(3841),
    [anon_sym_while] = ACTIONS(3841),
    [anon_sym_until] = ACTIONS(3841),
    [sym__PERLY_SEMICOLON] = ACTIONS(3841),
    [sym__ctrl_z_hack] = ACTIONS(3841),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1866)] = {
    [sym_heredoc_content] = STATE(1866),
    [anon_sym_RBRACE] = ACTIONS(3845),
    [anon_sym_SEMI] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [anon_sym_for] = ACTIONS(3847),
    [anon_sym_foreach] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3845),
    [anon_sym___END__] = ACTIONS(3845),
    [anon_sym_EOT] = ACTIONS(3845),
    [anon_sym_if] = ACTIONS(3845),
    [anon_sym_unless] = ACTIONS(3845),
    [anon_sym_while] = ACTIONS(3845),
    [anon_sym_until] = ACTIONS(3845),
    [sym__PERLY_SEMICOLON] = ACTIONS(3845),
    [sym__ctrl_z_hack] = ACTIONS(3845),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1867)] = {
    [sym_heredoc_content] = STATE(1867),
    [anon_sym_RBRACE] = ACTIONS(3849),
    [anon_sym_SEMI] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_RPAREN] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_and] = ACTIONS(3849),
    [anon_sym_or] = ACTIONS(3849),
    [anon_sym_xor] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [anon_sym_for] = ACTIONS(3851),
    [anon_sym_foreach] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3849),
    [anon_sym___END__] = ACTIONS(3849),
    [anon_sym_EOT] = ACTIONS(3849),
    [anon_sym_if] = ACTIONS(3849),
    [anon_sym_unless] = ACTIONS(3849),
    [anon_sym_while] = ACTIONS(3849),
    [anon_sym_until] = ACTIONS(3849),
    [sym__PERLY_SEMICOLON] = ACTIONS(3849),
    [sym__ctrl_z_hack] = ACTIONS(3849),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1868)] = {
    [sym_heredoc_content] = STATE(1868),
    [anon_sym_RBRACE] = ACTIONS(3853),
    [anon_sym_SEMI] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_RPAREN] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_and] = ACTIONS(3853),
    [anon_sym_or] = ACTIONS(3853),
    [anon_sym_xor] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [anon_sym_for] = ACTIONS(3855),
    [anon_sym_foreach] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3853),
    [anon_sym___END__] = ACTIONS(3853),
    [anon_sym_EOT] = ACTIONS(3853),
    [anon_sym_if] = ACTIONS(3853),
    [anon_sym_unless] = ACTIONS(3853),
    [anon_sym_while] = ACTIONS(3853),
    [anon_sym_until] = ACTIONS(3853),
    [sym__PERLY_SEMICOLON] = ACTIONS(3853),
    [sym__ctrl_z_hack] = ACTIONS(3853),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1869)] = {
    [sym_heredoc_content] = STATE(1869),
    [anon_sym_RBRACE] = ACTIONS(3857),
    [anon_sym_SEMI] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_RPAREN] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_and] = ACTIONS(3857),
    [anon_sym_or] = ACTIONS(3857),
    [anon_sym_xor] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [anon_sym_for] = ACTIONS(3859),
    [anon_sym_foreach] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3857),
    [anon_sym___END__] = ACTIONS(3857),
    [anon_sym_EOT] = ACTIONS(3857),
    [anon_sym_if] = ACTIONS(3857),
    [anon_sym_unless] = ACTIONS(3857),
    [anon_sym_while] = ACTIONS(3857),
    [anon_sym_until] = ACTIONS(3857),
    [sym__PERLY_SEMICOLON] = ACTIONS(3857),
    [sym__ctrl_z_hack] = ACTIONS(3857),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1870)] = {
    [sym_heredoc_content] = STATE(1870),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [anon_sym_for] = ACTIONS(2854),
    [anon_sym_foreach] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2858),
    [anon_sym___END__] = ACTIONS(2858),
    [anon_sym_EOT] = ACTIONS(2858),
    [anon_sym_if] = ACTIONS(2858),
    [anon_sym_unless] = ACTIONS(2858),
    [anon_sym_while] = ACTIONS(2858),
    [anon_sym_until] = ACTIONS(2858),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym__ctrl_z_hack] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1871)] = {
    [sym_heredoc_content] = STATE(1871),
    [anon_sym_RBRACE] = ACTIONS(3861),
    [anon_sym_SEMI] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_RPAREN] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_and] = ACTIONS(3861),
    [anon_sym_or] = ACTIONS(3861),
    [anon_sym_xor] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [anon_sym_for] = ACTIONS(3863),
    [anon_sym_foreach] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3861),
    [anon_sym___END__] = ACTIONS(3861),
    [anon_sym_EOT] = ACTIONS(3861),
    [anon_sym_if] = ACTIONS(3861),
    [anon_sym_unless] = ACTIONS(3861),
    [anon_sym_while] = ACTIONS(3861),
    [anon_sym_until] = ACTIONS(3861),
    [sym__PERLY_SEMICOLON] = ACTIONS(3861),
    [sym__ctrl_z_hack] = ACTIONS(3861),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1872)] = {
    [sym_heredoc_content] = STATE(1872),
    [anon_sym_RBRACE] = ACTIONS(3865),
    [anon_sym_SEMI] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_RPAREN] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_and] = ACTIONS(3865),
    [anon_sym_or] = ACTIONS(3865),
    [anon_sym_xor] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [anon_sym_for] = ACTIONS(3867),
    [anon_sym_foreach] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3865),
    [anon_sym___END__] = ACTIONS(3865),
    [anon_sym_EOT] = ACTIONS(3865),
    [anon_sym_if] = ACTIONS(3865),
    [anon_sym_unless] = ACTIONS(3865),
    [anon_sym_while] = ACTIONS(3865),
    [anon_sym_until] = ACTIONS(3865),
    [sym__PERLY_SEMICOLON] = ACTIONS(3865),
    [sym__ctrl_z_hack] = ACTIONS(3865),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1873)] = {
    [sym_heredoc_content] = STATE(1873),
    [anon_sym_RBRACE] = ACTIONS(3869),
    [anon_sym_SEMI] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_RPAREN] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_and] = ACTIONS(3869),
    [anon_sym_or] = ACTIONS(3869),
    [anon_sym_xor] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [anon_sym_for] = ACTIONS(3871),
    [anon_sym_foreach] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3869),
    [anon_sym___END__] = ACTIONS(3869),
    [anon_sym_EOT] = ACTIONS(3869),
    [anon_sym_if] = ACTIONS(3869),
    [anon_sym_unless] = ACTIONS(3869),
    [anon_sym_while] = ACTIONS(3869),
    [anon_sym_until] = ACTIONS(3869),
    [sym__PERLY_SEMICOLON] = ACTIONS(3869),
    [sym__ctrl_z_hack] = ACTIONS(3869),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1874)] = {
    [sym_heredoc_content] = STATE(1874),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [anon_sym_for] = ACTIONS(2854),
    [anon_sym_foreach] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2858),
    [anon_sym___END__] = ACTIONS(2858),
    [anon_sym_EOT] = ACTIONS(2858),
    [anon_sym_if] = ACTIONS(2858),
    [anon_sym_unless] = ACTIONS(2858),
    [anon_sym_while] = ACTIONS(2858),
    [anon_sym_until] = ACTIONS(2858),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym__ctrl_z_hack] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1875)] = {
    [sym_heredoc_content] = STATE(1875),
    [anon_sym_RBRACE] = ACTIONS(3873),
    [anon_sym_SEMI] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_RPAREN] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_and] = ACTIONS(3873),
    [anon_sym_or] = ACTIONS(3873),
    [anon_sym_xor] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [anon_sym_for] = ACTIONS(3875),
    [anon_sym_foreach] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3873),
    [anon_sym___END__] = ACTIONS(3873),
    [anon_sym_EOT] = ACTIONS(3873),
    [anon_sym_if] = ACTIONS(3873),
    [anon_sym_unless] = ACTIONS(3873),
    [anon_sym_while] = ACTIONS(3873),
    [anon_sym_until] = ACTIONS(3873),
    [sym__PERLY_SEMICOLON] = ACTIONS(3873),
    [sym__ctrl_z_hack] = ACTIONS(3873),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1876)] = {
    [sym_heredoc_content] = STATE(1876),
    [anon_sym_RBRACE] = ACTIONS(3877),
    [anon_sym_SEMI] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_RPAREN] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_and] = ACTIONS(3877),
    [anon_sym_or] = ACTIONS(3877),
    [anon_sym_xor] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [anon_sym_for] = ACTIONS(3879),
    [anon_sym_foreach] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3877),
    [anon_sym___END__] = ACTIONS(3877),
    [anon_sym_EOT] = ACTIONS(3877),
    [anon_sym_if] = ACTIONS(3877),
    [anon_sym_unless] = ACTIONS(3877),
    [anon_sym_while] = ACTIONS(3877),
    [anon_sym_until] = ACTIONS(3877),
    [sym__PERLY_SEMICOLON] = ACTIONS(3877),
    [sym__ctrl_z_hack] = ACTIONS(3877),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1877)] = {
    [sym_heredoc_content] = STATE(1877),
    [anon_sym_RBRACE] = ACTIONS(3881),
    [anon_sym_SEMI] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_RPAREN] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_and] = ACTIONS(3881),
    [anon_sym_or] = ACTIONS(3881),
    [anon_sym_xor] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [anon_sym_for] = ACTIONS(3883),
    [anon_sym_foreach] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3881),
    [anon_sym___END__] = ACTIONS(3881),
    [anon_sym_EOT] = ACTIONS(3881),
    [anon_sym_if] = ACTIONS(3881),
    [anon_sym_unless] = ACTIONS(3881),
    [anon_sym_while] = ACTIONS(3881),
    [anon_sym_until] = ACTIONS(3881),
    [sym__PERLY_SEMICOLON] = ACTIONS(3881),
    [sym__ctrl_z_hack] = ACTIONS(3881),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1878)] = {
    [sym_heredoc_content] = STATE(1878),
    [anon_sym_RBRACE] = ACTIONS(3885),
    [anon_sym_SEMI] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_RPAREN] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_and] = ACTIONS(3885),
    [anon_sym_or] = ACTIONS(3885),
    [anon_sym_xor] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [anon_sym_for] = ACTIONS(3887),
    [anon_sym_foreach] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3885),
    [anon_sym___END__] = ACTIONS(3885),
    [anon_sym_EOT] = ACTIONS(3885),
    [anon_sym_if] = ACTIONS(3885),
    [anon_sym_unless] = ACTIONS(3885),
    [anon_sym_while] = ACTIONS(3885),
    [anon_sym_until] = ACTIONS(3885),
    [sym__PERLY_SEMICOLON] = ACTIONS(3885),
    [sym__ctrl_z_hack] = ACTIONS(3885),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1879)] = {
    [sym_heredoc_content] = STATE(1879),
    [anon_sym_RBRACE] = ACTIONS(3889),
    [anon_sym_SEMI] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_RPAREN] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_and] = ACTIONS(3889),
    [anon_sym_or] = ACTIONS(3889),
    [anon_sym_xor] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [anon_sym_for] = ACTIONS(3891),
    [anon_sym_foreach] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3889),
    [anon_sym___END__] = ACTIONS(3889),
    [anon_sym_EOT] = ACTIONS(3889),
    [anon_sym_if] = ACTIONS(3889),
    [anon_sym_unless] = ACTIONS(3889),
    [anon_sym_while] = ACTIONS(3889),
    [anon_sym_until] = ACTIONS(3889),
    [sym__PERLY_SEMICOLON] = ACTIONS(3889),
    [sym__ctrl_z_hack] = ACTIONS(3889),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1880)] = {
    [sym_heredoc_content] = STATE(1880),
    [anon_sym_RBRACE] = ACTIONS(3893),
    [anon_sym_SEMI] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_RPAREN] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_and] = ACTIONS(3893),
    [anon_sym_or] = ACTIONS(3893),
    [anon_sym_xor] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [anon_sym_for] = ACTIONS(3895),
    [anon_sym_foreach] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3893),
    [anon_sym___END__] = ACTIONS(3893),
    [anon_sym_EOT] = ACTIONS(3893),
    [anon_sym_if] = ACTIONS(3893),
    [anon_sym_unless] = ACTIONS(3893),
    [anon_sym_while] = ACTIONS(3893),
    [anon_sym_until] = ACTIONS(3893),
    [sym__PERLY_SEMICOLON] = ACTIONS(3893),
    [sym__ctrl_z_hack] = ACTIONS(3893),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1881)] = {
    [sym_heredoc_content] = STATE(1881),
    [anon_sym_RBRACE] = ACTIONS(3897),
    [anon_sym_SEMI] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [anon_sym_for] = ACTIONS(3899),
    [anon_sym_foreach] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3897),
    [anon_sym___END__] = ACTIONS(3897),
    [anon_sym_EOT] = ACTIONS(3897),
    [anon_sym_if] = ACTIONS(3897),
    [anon_sym_unless] = ACTIONS(3897),
    [anon_sym_while] = ACTIONS(3897),
    [anon_sym_until] = ACTIONS(3897),
    [sym__PERLY_SEMICOLON] = ACTIONS(3897),
    [sym__ctrl_z_hack] = ACTIONS(3897),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1882)] = {
    [sym_heredoc_content] = STATE(1882),
    [anon_sym_RBRACE] = ACTIONS(3901),
    [anon_sym_SEMI] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_RPAREN] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_and] = ACTIONS(3901),
    [anon_sym_or] = ACTIONS(3901),
    [anon_sym_xor] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [anon_sym_for] = ACTIONS(3903),
    [anon_sym_foreach] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3901),
    [anon_sym___END__] = ACTIONS(3901),
    [anon_sym_EOT] = ACTIONS(3901),
    [anon_sym_if] = ACTIONS(3901),
    [anon_sym_unless] = ACTIONS(3901),
    [anon_sym_while] = ACTIONS(3901),
    [anon_sym_until] = ACTIONS(3901),
    [sym__PERLY_SEMICOLON] = ACTIONS(3901),
    [sym__ctrl_z_hack] = ACTIONS(3901),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1883)] = {
    [sym_heredoc_content] = STATE(1883),
    [anon_sym_RBRACE] = ACTIONS(3905),
    [anon_sym_SEMI] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_RPAREN] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_and] = ACTIONS(3905),
    [anon_sym_or] = ACTIONS(3905),
    [anon_sym_xor] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [anon_sym_for] = ACTIONS(3907),
    [anon_sym_foreach] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3905),
    [anon_sym___END__] = ACTIONS(3905),
    [anon_sym_EOT] = ACTIONS(3905),
    [anon_sym_if] = ACTIONS(3905),
    [anon_sym_unless] = ACTIONS(3905),
    [anon_sym_while] = ACTIONS(3905),
    [anon_sym_until] = ACTIONS(3905),
    [sym__PERLY_SEMICOLON] = ACTIONS(3905),
    [sym__ctrl_z_hack] = ACTIONS(3905),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1884)] = {
    [sym_heredoc_content] = STATE(1884),
    [anon_sym_RBRACE] = ACTIONS(3909),
    [anon_sym_SEMI] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [anon_sym_for] = ACTIONS(3911),
    [anon_sym_foreach] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3909),
    [anon_sym___END__] = ACTIONS(3909),
    [anon_sym_EOT] = ACTIONS(3909),
    [anon_sym_if] = ACTIONS(3909),
    [anon_sym_unless] = ACTIONS(3909),
    [anon_sym_while] = ACTIONS(3909),
    [anon_sym_until] = ACTIONS(3909),
    [sym__PERLY_SEMICOLON] = ACTIONS(3909),
    [sym__ctrl_z_hack] = ACTIONS(3909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1885)] = {
    [sym_heredoc_content] = STATE(1885),
    [anon_sym_RBRACE] = ACTIONS(3913),
    [anon_sym_SEMI] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_RPAREN] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_and] = ACTIONS(3913),
    [anon_sym_or] = ACTIONS(3913),
    [anon_sym_xor] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [anon_sym_for] = ACTIONS(3915),
    [anon_sym_foreach] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3913),
    [anon_sym___END__] = ACTIONS(3913),
    [anon_sym_EOT] = ACTIONS(3913),
    [anon_sym_if] = ACTIONS(3913),
    [anon_sym_unless] = ACTIONS(3913),
    [anon_sym_while] = ACTIONS(3913),
    [anon_sym_until] = ACTIONS(3913),
    [sym__PERLY_SEMICOLON] = ACTIONS(3913),
    [sym__ctrl_z_hack] = ACTIONS(3913),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1886)] = {
    [sym_heredoc_content] = STATE(1886),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_SEMI] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_RPAREN] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [anon_sym_for] = ACTIONS(3919),
    [anon_sym_foreach] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3917),
    [anon_sym___END__] = ACTIONS(3917),
    [anon_sym_EOT] = ACTIONS(3917),
    [anon_sym_if] = ACTIONS(3917),
    [anon_sym_unless] = ACTIONS(3917),
    [anon_sym_while] = ACTIONS(3917),
    [anon_sym_until] = ACTIONS(3917),
    [sym__PERLY_SEMICOLON] = ACTIONS(3917),
    [sym__ctrl_z_hack] = ACTIONS(3917),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1887)] = {
    [sym_heredoc_content] = STATE(1887),
    [anon_sym_RBRACE] = ACTIONS(3921),
    [anon_sym_SEMI] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_RPAREN] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_and] = ACTIONS(3921),
    [anon_sym_or] = ACTIONS(3921),
    [anon_sym_xor] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [anon_sym_for] = ACTIONS(3923),
    [anon_sym_foreach] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3921),
    [anon_sym___END__] = ACTIONS(3921),
    [anon_sym_EOT] = ACTIONS(3921),
    [anon_sym_if] = ACTIONS(3921),
    [anon_sym_unless] = ACTIONS(3921),
    [anon_sym_while] = ACTIONS(3921),
    [anon_sym_until] = ACTIONS(3921),
    [sym__PERLY_SEMICOLON] = ACTIONS(3921),
    [sym__ctrl_z_hack] = ACTIONS(3921),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1888)] = {
    [sym_heredoc_content] = STATE(1888),
    [anon_sym_RBRACE] = ACTIONS(3925),
    [anon_sym_SEMI] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_RPAREN] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_and] = ACTIONS(3925),
    [anon_sym_or] = ACTIONS(3925),
    [anon_sym_xor] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [anon_sym_for] = ACTIONS(3927),
    [anon_sym_foreach] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3925),
    [anon_sym___END__] = ACTIONS(3925),
    [anon_sym_EOT] = ACTIONS(3925),
    [anon_sym_if] = ACTIONS(3925),
    [anon_sym_unless] = ACTIONS(3925),
    [anon_sym_while] = ACTIONS(3925),
    [anon_sym_until] = ACTIONS(3925),
    [sym__PERLY_SEMICOLON] = ACTIONS(3925),
    [sym__ctrl_z_hack] = ACTIONS(3925),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1889)] = {
    [sym_heredoc_content] = STATE(1889),
    [anon_sym_RBRACE] = ACTIONS(3929),
    [anon_sym_SEMI] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_RPAREN] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_and] = ACTIONS(3929),
    [anon_sym_or] = ACTIONS(3929),
    [anon_sym_xor] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [anon_sym_for] = ACTIONS(3931),
    [anon_sym_foreach] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3929),
    [anon_sym___END__] = ACTIONS(3929),
    [anon_sym_EOT] = ACTIONS(3929),
    [anon_sym_if] = ACTIONS(3929),
    [anon_sym_unless] = ACTIONS(3929),
    [anon_sym_while] = ACTIONS(3929),
    [anon_sym_until] = ACTIONS(3929),
    [sym__PERLY_SEMICOLON] = ACTIONS(3929),
    [sym__ctrl_z_hack] = ACTIONS(3929),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1890)] = {
    [sym_heredoc_content] = STATE(1890),
    [anon_sym_RBRACE] = ACTIONS(3933),
    [anon_sym_SEMI] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_RPAREN] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_and] = ACTIONS(3933),
    [anon_sym_or] = ACTIONS(3933),
    [anon_sym_xor] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [anon_sym_for] = ACTIONS(3935),
    [anon_sym_foreach] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3933),
    [anon_sym___END__] = ACTIONS(3933),
    [anon_sym_EOT] = ACTIONS(3933),
    [anon_sym_if] = ACTIONS(3933),
    [anon_sym_unless] = ACTIONS(3933),
    [anon_sym_while] = ACTIONS(3933),
    [anon_sym_until] = ACTIONS(3933),
    [sym__PERLY_SEMICOLON] = ACTIONS(3933),
    [sym__ctrl_z_hack] = ACTIONS(3933),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1891)] = {
    [sym_heredoc_content] = STATE(1891),
    [anon_sym_RBRACE] = ACTIONS(3937),
    [anon_sym_SEMI] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_RPAREN] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_and] = ACTIONS(3937),
    [anon_sym_or] = ACTIONS(3937),
    [anon_sym_xor] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [anon_sym_for] = ACTIONS(3939),
    [anon_sym_foreach] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3937),
    [anon_sym___END__] = ACTIONS(3937),
    [anon_sym_EOT] = ACTIONS(3937),
    [anon_sym_if] = ACTIONS(3937),
    [anon_sym_unless] = ACTIONS(3937),
    [anon_sym_while] = ACTIONS(3937),
    [anon_sym_until] = ACTIONS(3937),
    [sym__PERLY_SEMICOLON] = ACTIONS(3937),
    [sym__ctrl_z_hack] = ACTIONS(3937),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1892)] = {
    [sym_heredoc_content] = STATE(1892),
    [anon_sym_RBRACE] = ACTIONS(3941),
    [anon_sym_SEMI] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_RPAREN] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_and] = ACTIONS(3941),
    [anon_sym_or] = ACTIONS(3941),
    [anon_sym_xor] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [anon_sym_for] = ACTIONS(3943),
    [anon_sym_foreach] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3941),
    [anon_sym___END__] = ACTIONS(3941),
    [anon_sym_EOT] = ACTIONS(3941),
    [anon_sym_if] = ACTIONS(3941),
    [anon_sym_unless] = ACTIONS(3941),
    [anon_sym_while] = ACTIONS(3941),
    [anon_sym_until] = ACTIONS(3941),
    [sym__PERLY_SEMICOLON] = ACTIONS(3941),
    [sym__ctrl_z_hack] = ACTIONS(3941),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1893)] = {
    [sym_heredoc_content] = STATE(1893),
    [anon_sym_RBRACE] = ACTIONS(3945),
    [anon_sym_SEMI] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_RPAREN] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_and] = ACTIONS(3945),
    [anon_sym_or] = ACTIONS(3945),
    [anon_sym_xor] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [anon_sym_for] = ACTIONS(3947),
    [anon_sym_foreach] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3945),
    [anon_sym___END__] = ACTIONS(3945),
    [anon_sym_EOT] = ACTIONS(3945),
    [anon_sym_if] = ACTIONS(3945),
    [anon_sym_unless] = ACTIONS(3945),
    [anon_sym_while] = ACTIONS(3945),
    [anon_sym_until] = ACTIONS(3945),
    [sym__PERLY_SEMICOLON] = ACTIONS(3945),
    [sym__ctrl_z_hack] = ACTIONS(3945),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1894)] = {
    [sym_heredoc_content] = STATE(1894),
    [anon_sym_RBRACE] = ACTIONS(3949),
    [anon_sym_SEMI] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_RPAREN] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_and] = ACTIONS(3949),
    [anon_sym_or] = ACTIONS(3949),
    [anon_sym_xor] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [anon_sym_for] = ACTIONS(3951),
    [anon_sym_foreach] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3949),
    [anon_sym___END__] = ACTIONS(3949),
    [anon_sym_EOT] = ACTIONS(3949),
    [anon_sym_if] = ACTIONS(3949),
    [anon_sym_unless] = ACTIONS(3949),
    [anon_sym_while] = ACTIONS(3949),
    [anon_sym_until] = ACTIONS(3949),
    [sym__PERLY_SEMICOLON] = ACTIONS(3949),
    [sym__ctrl_z_hack] = ACTIONS(3949),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1895)] = {
    [sym_heredoc_content] = STATE(1895),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_and] = ACTIONS(3953),
    [anon_sym_or] = ACTIONS(3953),
    [anon_sym_xor] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [anon_sym_for] = ACTIONS(3955),
    [anon_sym_foreach] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3953),
    [anon_sym___END__] = ACTIONS(3953),
    [anon_sym_EOT] = ACTIONS(3953),
    [anon_sym_if] = ACTIONS(3953),
    [anon_sym_unless] = ACTIONS(3953),
    [anon_sym_while] = ACTIONS(3953),
    [anon_sym_until] = ACTIONS(3953),
    [sym__PERLY_SEMICOLON] = ACTIONS(3953),
    [sym__ctrl_z_hack] = ACTIONS(3953),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1896)] = {
    [sym_heredoc_content] = STATE(1896),
    [anon_sym_RBRACE] = ACTIONS(3957),
    [anon_sym_SEMI] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_RPAREN] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_and] = ACTIONS(3957),
    [anon_sym_or] = ACTIONS(3957),
    [anon_sym_xor] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [anon_sym_for] = ACTIONS(3959),
    [anon_sym_foreach] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3957),
    [anon_sym___END__] = ACTIONS(3957),
    [anon_sym_EOT] = ACTIONS(3957),
    [anon_sym_if] = ACTIONS(3957),
    [anon_sym_unless] = ACTIONS(3957),
    [anon_sym_while] = ACTIONS(3957),
    [anon_sym_until] = ACTIONS(3957),
    [sym__PERLY_SEMICOLON] = ACTIONS(3957),
    [sym__ctrl_z_hack] = ACTIONS(3957),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1897)] = {
    [sym_heredoc_content] = STATE(1897),
    [anon_sym_RBRACE] = ACTIONS(3961),
    [anon_sym_SEMI] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_RPAREN] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_and] = ACTIONS(3961),
    [anon_sym_or] = ACTIONS(3961),
    [anon_sym_xor] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [anon_sym_for] = ACTIONS(3963),
    [anon_sym_foreach] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3961),
    [anon_sym___END__] = ACTIONS(3961),
    [anon_sym_EOT] = ACTIONS(3961),
    [anon_sym_if] = ACTIONS(3961),
    [anon_sym_unless] = ACTIONS(3961),
    [anon_sym_while] = ACTIONS(3961),
    [anon_sym_until] = ACTIONS(3961),
    [sym__PERLY_SEMICOLON] = ACTIONS(3961),
    [sym__ctrl_z_hack] = ACTIONS(3961),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1898)] = {
    [sym_heredoc_content] = STATE(1898),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_and] = ACTIONS(3965),
    [anon_sym_or] = ACTIONS(3965),
    [anon_sym_xor] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [anon_sym_for] = ACTIONS(3967),
    [anon_sym_foreach] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3965),
    [anon_sym___END__] = ACTIONS(3965),
    [anon_sym_EOT] = ACTIONS(3965),
    [anon_sym_if] = ACTIONS(3965),
    [anon_sym_unless] = ACTIONS(3965),
    [anon_sym_while] = ACTIONS(3965),
    [anon_sym_until] = ACTIONS(3965),
    [sym__PERLY_SEMICOLON] = ACTIONS(3965),
    [sym__ctrl_z_hack] = ACTIONS(3965),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1899)] = {
    [sym_heredoc_content] = STATE(1899),
    [anon_sym_RBRACE] = ACTIONS(3969),
    [anon_sym_SEMI] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_RPAREN] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_and] = ACTIONS(3969),
    [anon_sym_or] = ACTIONS(3969),
    [anon_sym_xor] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [anon_sym_for] = ACTIONS(3971),
    [anon_sym_foreach] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3969),
    [anon_sym___END__] = ACTIONS(3969),
    [anon_sym_EOT] = ACTIONS(3969),
    [anon_sym_if] = ACTIONS(3969),
    [anon_sym_unless] = ACTIONS(3969),
    [anon_sym_while] = ACTIONS(3969),
    [anon_sym_until] = ACTIONS(3969),
    [sym__PERLY_SEMICOLON] = ACTIONS(3969),
    [sym__ctrl_z_hack] = ACTIONS(3969),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1900)] = {
    [sym_heredoc_content] = STATE(1900),
    [anon_sym_RBRACE] = ACTIONS(3973),
    [anon_sym_SEMI] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_and] = ACTIONS(3973),
    [anon_sym_or] = ACTIONS(3973),
    [anon_sym_xor] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [anon_sym_for] = ACTIONS(3975),
    [anon_sym_foreach] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3973),
    [anon_sym___END__] = ACTIONS(3973),
    [anon_sym_EOT] = ACTIONS(3973),
    [anon_sym_if] = ACTIONS(3973),
    [anon_sym_unless] = ACTIONS(3973),
    [anon_sym_while] = ACTIONS(3973),
    [anon_sym_until] = ACTIONS(3973),
    [sym__PERLY_SEMICOLON] = ACTIONS(3973),
    [sym__ctrl_z_hack] = ACTIONS(3973),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1901)] = {
    [sym_heredoc_content] = STATE(1901),
    [anon_sym_RBRACE] = ACTIONS(3977),
    [anon_sym_SEMI] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [anon_sym_for] = ACTIONS(3979),
    [anon_sym_foreach] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3977),
    [anon_sym___END__] = ACTIONS(3977),
    [anon_sym_EOT] = ACTIONS(3977),
    [anon_sym_if] = ACTIONS(3977),
    [anon_sym_unless] = ACTIONS(3977),
    [anon_sym_while] = ACTIONS(3977),
    [anon_sym_until] = ACTIONS(3977),
    [sym__PERLY_SEMICOLON] = ACTIONS(3977),
    [sym__ctrl_z_hack] = ACTIONS(3977),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1902)] = {
    [sym_heredoc_content] = STATE(1902),
    [anon_sym_RBRACE] = ACTIONS(3981),
    [anon_sym_SEMI] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_RPAREN] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_and] = ACTIONS(3981),
    [anon_sym_or] = ACTIONS(3981),
    [anon_sym_xor] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [anon_sym_for] = ACTIONS(3983),
    [anon_sym_foreach] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3981),
    [anon_sym___END__] = ACTIONS(3981),
    [anon_sym_EOT] = ACTIONS(3981),
    [anon_sym_if] = ACTIONS(3981),
    [anon_sym_unless] = ACTIONS(3981),
    [anon_sym_while] = ACTIONS(3981),
    [anon_sym_until] = ACTIONS(3981),
    [sym__PERLY_SEMICOLON] = ACTIONS(3981),
    [sym__ctrl_z_hack] = ACTIONS(3981),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1903)] = {
    [sym_heredoc_content] = STATE(1903),
    [anon_sym_RBRACE] = ACTIONS(3985),
    [anon_sym_SEMI] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [anon_sym_for] = ACTIONS(3987),
    [anon_sym_foreach] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3985),
    [anon_sym___END__] = ACTIONS(3985),
    [anon_sym_EOT] = ACTIONS(3985),
    [anon_sym_if] = ACTIONS(3985),
    [anon_sym_unless] = ACTIONS(3985),
    [anon_sym_while] = ACTIONS(3985),
    [anon_sym_until] = ACTIONS(3985),
    [sym__PERLY_SEMICOLON] = ACTIONS(3985),
    [sym__ctrl_z_hack] = ACTIONS(3985),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1904)] = {
    [sym_heredoc_content] = STATE(1904),
    [anon_sym_RBRACE] = ACTIONS(3989),
    [anon_sym_SEMI] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_RPAREN] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_and] = ACTIONS(3989),
    [anon_sym_or] = ACTIONS(3989),
    [anon_sym_xor] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [anon_sym_for] = ACTIONS(3991),
    [anon_sym_foreach] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3989),
    [anon_sym___END__] = ACTIONS(3989),
    [anon_sym_EOT] = ACTIONS(3989),
    [anon_sym_if] = ACTIONS(3989),
    [anon_sym_unless] = ACTIONS(3989),
    [anon_sym_while] = ACTIONS(3989),
    [anon_sym_until] = ACTIONS(3989),
    [sym__PERLY_SEMICOLON] = ACTIONS(3989),
    [sym__ctrl_z_hack] = ACTIONS(3989),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1905)] = {
    [sym_heredoc_content] = STATE(1905),
    [anon_sym_RBRACE] = ACTIONS(3993),
    [anon_sym_SEMI] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_RPAREN] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_and] = ACTIONS(3993),
    [anon_sym_or] = ACTIONS(3993),
    [anon_sym_xor] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [anon_sym_for] = ACTIONS(3995),
    [anon_sym_foreach] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3993),
    [anon_sym___END__] = ACTIONS(3993),
    [anon_sym_EOT] = ACTIONS(3993),
    [anon_sym_if] = ACTIONS(3993),
    [anon_sym_unless] = ACTIONS(3993),
    [anon_sym_while] = ACTIONS(3993),
    [anon_sym_until] = ACTIONS(3993),
    [sym__PERLY_SEMICOLON] = ACTIONS(3993),
    [sym__ctrl_z_hack] = ACTIONS(3993),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1906)] = {
    [sym_heredoc_content] = STATE(1906),
    [anon_sym_RBRACE] = ACTIONS(3997),
    [anon_sym_SEMI] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [anon_sym_for] = ACTIONS(3999),
    [anon_sym_foreach] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3997),
    [anon_sym___END__] = ACTIONS(3997),
    [anon_sym_EOT] = ACTIONS(3997),
    [anon_sym_if] = ACTIONS(3997),
    [anon_sym_unless] = ACTIONS(3997),
    [anon_sym_while] = ACTIONS(3997),
    [anon_sym_until] = ACTIONS(3997),
    [sym__PERLY_SEMICOLON] = ACTIONS(3997),
    [sym__ctrl_z_hack] = ACTIONS(3997),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1907)] = {
    [sym_heredoc_content] = STATE(1907),
    [anon_sym_RBRACE] = ACTIONS(4001),
    [anon_sym_SEMI] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_RPAREN] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_and] = ACTIONS(4001),
    [anon_sym_or] = ACTIONS(4001),
    [anon_sym_xor] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [anon_sym_for] = ACTIONS(4003),
    [anon_sym_foreach] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4001),
    [anon_sym___END__] = ACTIONS(4001),
    [anon_sym_EOT] = ACTIONS(4001),
    [anon_sym_if] = ACTIONS(4001),
    [anon_sym_unless] = ACTIONS(4001),
    [anon_sym_while] = ACTIONS(4001),
    [anon_sym_until] = ACTIONS(4001),
    [sym__PERLY_SEMICOLON] = ACTIONS(4001),
    [sym__ctrl_z_hack] = ACTIONS(4001),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1908)] = {
    [sym_heredoc_content] = STATE(1908),
    [anon_sym_RBRACE] = ACTIONS(4005),
    [anon_sym_SEMI] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_RPAREN] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_and] = ACTIONS(4005),
    [anon_sym_or] = ACTIONS(4005),
    [anon_sym_xor] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [anon_sym_for] = ACTIONS(4007),
    [anon_sym_foreach] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4005),
    [anon_sym___END__] = ACTIONS(4005),
    [anon_sym_EOT] = ACTIONS(4005),
    [anon_sym_if] = ACTIONS(4005),
    [anon_sym_unless] = ACTIONS(4005),
    [anon_sym_while] = ACTIONS(4005),
    [anon_sym_until] = ACTIONS(4005),
    [sym__PERLY_SEMICOLON] = ACTIONS(4005),
    [sym__ctrl_z_hack] = ACTIONS(4005),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1909)] = {
    [sym_heredoc_content] = STATE(1909),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_SEMI] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [anon_sym_for] = ACTIONS(3821),
    [anon_sym_foreach] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3819),
    [anon_sym___END__] = ACTIONS(3819),
    [anon_sym_EOT] = ACTIONS(3819),
    [anon_sym_if] = ACTIONS(3819),
    [anon_sym_unless] = ACTIONS(3819),
    [anon_sym_while] = ACTIONS(3819),
    [anon_sym_until] = ACTIONS(3819),
    [sym__PERLY_SEMICOLON] = ACTIONS(3819),
    [sym__ctrl_z_hack] = ACTIONS(3819),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1910)] = {
    [sym_heredoc_content] = STATE(1910),
    [anon_sym_RBRACE] = ACTIONS(4009),
    [anon_sym_SEMI] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_RPAREN] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_and] = ACTIONS(4009),
    [anon_sym_or] = ACTIONS(4009),
    [anon_sym_xor] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [anon_sym_for] = ACTIONS(4011),
    [anon_sym_foreach] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4009),
    [anon_sym___END__] = ACTIONS(4009),
    [anon_sym_EOT] = ACTIONS(4009),
    [anon_sym_if] = ACTIONS(4009),
    [anon_sym_unless] = ACTIONS(4009),
    [anon_sym_while] = ACTIONS(4009),
    [anon_sym_until] = ACTIONS(4009),
    [sym__PERLY_SEMICOLON] = ACTIONS(4009),
    [sym__ctrl_z_hack] = ACTIONS(4009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1911)] = {
    [sym_heredoc_content] = STATE(1911),
    [anon_sym_RBRACE] = ACTIONS(4013),
    [anon_sym_SEMI] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_RPAREN] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_and] = ACTIONS(4013),
    [anon_sym_or] = ACTIONS(4013),
    [anon_sym_xor] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [anon_sym_for] = ACTIONS(4015),
    [anon_sym_foreach] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4013),
    [anon_sym___END__] = ACTIONS(4013),
    [anon_sym_EOT] = ACTIONS(4013),
    [anon_sym_if] = ACTIONS(4013),
    [anon_sym_unless] = ACTIONS(4013),
    [anon_sym_while] = ACTIONS(4013),
    [anon_sym_until] = ACTIONS(4013),
    [sym__PERLY_SEMICOLON] = ACTIONS(4013),
    [sym__ctrl_z_hack] = ACTIONS(4013),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1912)] = {
    [sym_heredoc_content] = STATE(1912),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_SEMI] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [anon_sym_for] = ACTIONS(4019),
    [anon_sym_foreach] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4017),
    [anon_sym___END__] = ACTIONS(4017),
    [anon_sym_EOT] = ACTIONS(4017),
    [anon_sym_if] = ACTIONS(4017),
    [anon_sym_unless] = ACTIONS(4017),
    [anon_sym_while] = ACTIONS(4017),
    [anon_sym_until] = ACTIONS(4017),
    [sym__PERLY_SEMICOLON] = ACTIONS(4017),
    [sym__ctrl_z_hack] = ACTIONS(4017),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1913)] = {
    [sym_heredoc_content] = STATE(1913),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_SEMI] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [anon_sym_for] = ACTIONS(4019),
    [anon_sym_foreach] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4017),
    [anon_sym___END__] = ACTIONS(4017),
    [anon_sym_EOT] = ACTIONS(4017),
    [anon_sym_if] = ACTIONS(4017),
    [anon_sym_unless] = ACTIONS(4017),
    [anon_sym_while] = ACTIONS(4017),
    [anon_sym_until] = ACTIONS(4017),
    [sym__PERLY_SEMICOLON] = ACTIONS(4017),
    [sym__ctrl_z_hack] = ACTIONS(4017),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1914)] = {
    [sym_heredoc_content] = STATE(1914),
    [anon_sym_RBRACE] = ACTIONS(4021),
    [anon_sym_SEMI] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_RPAREN] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_and] = ACTIONS(4021),
    [anon_sym_or] = ACTIONS(4021),
    [anon_sym_xor] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [anon_sym_for] = ACTIONS(4023),
    [anon_sym_foreach] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4021),
    [anon_sym___END__] = ACTIONS(4021),
    [anon_sym_EOT] = ACTIONS(4021),
    [anon_sym_if] = ACTIONS(4021),
    [anon_sym_unless] = ACTIONS(4021),
    [anon_sym_while] = ACTIONS(4021),
    [anon_sym_until] = ACTIONS(4021),
    [sym__PERLY_SEMICOLON] = ACTIONS(4021),
    [sym__ctrl_z_hack] = ACTIONS(4021),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1915)] = {
    [sym_heredoc_content] = STATE(1915),
    [anon_sym_RBRACE] = ACTIONS(4025),
    [anon_sym_SEMI] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_RPAREN] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_and] = ACTIONS(4025),
    [anon_sym_or] = ACTIONS(4025),
    [anon_sym_xor] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [anon_sym_for] = ACTIONS(4027),
    [anon_sym_foreach] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4025),
    [anon_sym___END__] = ACTIONS(4025),
    [anon_sym_EOT] = ACTIONS(4025),
    [anon_sym_if] = ACTIONS(4025),
    [anon_sym_unless] = ACTIONS(4025),
    [anon_sym_while] = ACTIONS(4025),
    [anon_sym_until] = ACTIONS(4025),
    [sym__PERLY_SEMICOLON] = ACTIONS(4025),
    [sym__ctrl_z_hack] = ACTIONS(4025),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1916)] = {
    [sym_heredoc_content] = STATE(1916),
    [anon_sym_RBRACE] = ACTIONS(4029),
    [anon_sym_SEMI] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_RPAREN] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_and] = ACTIONS(4029),
    [anon_sym_or] = ACTIONS(4029),
    [anon_sym_xor] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [anon_sym_for] = ACTIONS(4031),
    [anon_sym_foreach] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4029),
    [anon_sym___END__] = ACTIONS(4029),
    [anon_sym_EOT] = ACTIONS(4029),
    [anon_sym_if] = ACTIONS(4029),
    [anon_sym_unless] = ACTIONS(4029),
    [anon_sym_while] = ACTIONS(4029),
    [anon_sym_until] = ACTIONS(4029),
    [sym__PERLY_SEMICOLON] = ACTIONS(4029),
    [sym__ctrl_z_hack] = ACTIONS(4029),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1917)] = {
    [sym_heredoc_content] = STATE(1917),
    [anon_sym_RBRACE] = ACTIONS(4033),
    [anon_sym_SEMI] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_RPAREN] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_and] = ACTIONS(4033),
    [anon_sym_or] = ACTIONS(4033),
    [anon_sym_xor] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [anon_sym_for] = ACTIONS(4035),
    [anon_sym_foreach] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4033),
    [anon_sym___END__] = ACTIONS(4033),
    [anon_sym_EOT] = ACTIONS(4033),
    [anon_sym_if] = ACTIONS(4033),
    [anon_sym_unless] = ACTIONS(4033),
    [anon_sym_while] = ACTIONS(4033),
    [anon_sym_until] = ACTIONS(4033),
    [sym__PERLY_SEMICOLON] = ACTIONS(4033),
    [sym__ctrl_z_hack] = ACTIONS(4033),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1918)] = {
    [sym_heredoc_content] = STATE(1918),
    [anon_sym_RBRACE] = ACTIONS(4037),
    [anon_sym_SEMI] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_RPAREN] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_and] = ACTIONS(4037),
    [anon_sym_or] = ACTIONS(4037),
    [anon_sym_xor] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [anon_sym_for] = ACTIONS(4039),
    [anon_sym_foreach] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4037),
    [anon_sym___END__] = ACTIONS(4037),
    [anon_sym_EOT] = ACTIONS(4037),
    [anon_sym_if] = ACTIONS(4037),
    [anon_sym_unless] = ACTIONS(4037),
    [anon_sym_while] = ACTIONS(4037),
    [anon_sym_until] = ACTIONS(4037),
    [sym__PERLY_SEMICOLON] = ACTIONS(4037),
    [sym__ctrl_z_hack] = ACTIONS(4037),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1919)] = {
    [sym_heredoc_content] = STATE(1919),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2536),
    [anon_sym___END__] = ACTIONS(2536),
    [anon_sym_EOT] = ACTIONS(2536),
    [anon_sym_if] = ACTIONS(2536),
    [anon_sym_unless] = ACTIONS(2536),
    [anon_sym_while] = ACTIONS(2536),
    [anon_sym_until] = ACTIONS(2536),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1920)] = {
    [sym_heredoc_content] = STATE(1920),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [anon_sym_for] = ACTIONS(3417),
    [anon_sym_foreach] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3415),
    [anon_sym___END__] = ACTIONS(3415),
    [anon_sym_EOT] = ACTIONS(3415),
    [anon_sym_if] = ACTIONS(3415),
    [anon_sym_unless] = ACTIONS(3415),
    [anon_sym_while] = ACTIONS(3415),
    [anon_sym_until] = ACTIONS(3415),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym__ctrl_z_hack] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1921)] = {
    [sym_heredoc_content] = STATE(1921),
    [anon_sym_RBRACE] = ACTIONS(4041),
    [anon_sym_SEMI] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_RPAREN] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_and] = ACTIONS(4041),
    [anon_sym_or] = ACTIONS(4041),
    [anon_sym_xor] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [anon_sym_for] = ACTIONS(4043),
    [anon_sym_foreach] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4041),
    [anon_sym___END__] = ACTIONS(4041),
    [anon_sym_EOT] = ACTIONS(4041),
    [anon_sym_if] = ACTIONS(4041),
    [anon_sym_unless] = ACTIONS(4041),
    [anon_sym_while] = ACTIONS(4041),
    [anon_sym_until] = ACTIONS(4041),
    [sym__PERLY_SEMICOLON] = ACTIONS(4041),
    [sym__ctrl_z_hack] = ACTIONS(4041),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1922)] = {
    [sym_heredoc_content] = STATE(1922),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_SEMI] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_RPAREN] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_foreach] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4045),
    [anon_sym___END__] = ACTIONS(4045),
    [anon_sym_EOT] = ACTIONS(4045),
    [anon_sym_if] = ACTIONS(4045),
    [anon_sym_unless] = ACTIONS(4045),
    [anon_sym_while] = ACTIONS(4045),
    [anon_sym_until] = ACTIONS(4045),
    [sym__PERLY_SEMICOLON] = ACTIONS(4045),
    [sym__ctrl_z_hack] = ACTIONS(4045),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1923)] = {
    [sym_heredoc_content] = STATE(1923),
    [anon_sym_RBRACE] = ACTIONS(4047),
    [anon_sym_SEMI] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_RPAREN] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_and] = ACTIONS(4047),
    [anon_sym_or] = ACTIONS(4047),
    [anon_sym_xor] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [anon_sym_for] = ACTIONS(4049),
    [anon_sym_foreach] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4047),
    [anon_sym___END__] = ACTIONS(4047),
    [anon_sym_EOT] = ACTIONS(4047),
    [anon_sym_if] = ACTIONS(4047),
    [anon_sym_unless] = ACTIONS(4047),
    [anon_sym_while] = ACTIONS(4047),
    [anon_sym_until] = ACTIONS(4047),
    [sym__PERLY_SEMICOLON] = ACTIONS(4047),
    [sym__ctrl_z_hack] = ACTIONS(4047),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1924)] = {
    [sym_heredoc_content] = STATE(1924),
    [anon_sym_RBRACE] = ACTIONS(4051),
    [anon_sym_SEMI] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_RPAREN] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_and] = ACTIONS(4051),
    [anon_sym_or] = ACTIONS(4051),
    [anon_sym_xor] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [anon_sym_for] = ACTIONS(4053),
    [anon_sym_foreach] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4051),
    [anon_sym___END__] = ACTIONS(4051),
    [anon_sym_EOT] = ACTIONS(4051),
    [anon_sym_if] = ACTIONS(4051),
    [anon_sym_unless] = ACTIONS(4051),
    [anon_sym_while] = ACTIONS(4051),
    [anon_sym_until] = ACTIONS(4051),
    [sym__PERLY_SEMICOLON] = ACTIONS(4051),
    [sym__ctrl_z_hack] = ACTIONS(4051),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1925)] = {
    [sym_heredoc_content] = STATE(1925),
    [anon_sym_RBRACE] = ACTIONS(4055),
    [anon_sym_SEMI] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_RPAREN] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_and] = ACTIONS(4055),
    [anon_sym_or] = ACTIONS(4055),
    [anon_sym_xor] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [anon_sym_for] = ACTIONS(4057),
    [anon_sym_foreach] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4055),
    [anon_sym___END__] = ACTIONS(4055),
    [anon_sym_EOT] = ACTIONS(4055),
    [anon_sym_if] = ACTIONS(4055),
    [anon_sym_unless] = ACTIONS(4055),
    [anon_sym_while] = ACTIONS(4055),
    [anon_sym_until] = ACTIONS(4055),
    [sym__PERLY_SEMICOLON] = ACTIONS(4055),
    [sym__ctrl_z_hack] = ACTIONS(4055),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1926)] = {
    [sym_heredoc_content] = STATE(1926),
    [anon_sym_RBRACE] = ACTIONS(4059),
    [anon_sym_SEMI] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_RPAREN] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_and] = ACTIONS(4059),
    [anon_sym_or] = ACTIONS(4059),
    [anon_sym_xor] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [anon_sym_for] = ACTIONS(4061),
    [anon_sym_foreach] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4059),
    [anon_sym___END__] = ACTIONS(4059),
    [anon_sym_EOT] = ACTIONS(4059),
    [anon_sym_if] = ACTIONS(4059),
    [anon_sym_unless] = ACTIONS(4059),
    [anon_sym_while] = ACTIONS(4059),
    [anon_sym_until] = ACTIONS(4059),
    [sym__PERLY_SEMICOLON] = ACTIONS(4059),
    [sym__ctrl_z_hack] = ACTIONS(4059),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1927)] = {
    [sym_heredoc_content] = STATE(1927),
    [anon_sym_RBRACE] = ACTIONS(4063),
    [anon_sym_SEMI] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_RPAREN] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_and] = ACTIONS(4063),
    [anon_sym_or] = ACTIONS(4063),
    [anon_sym_xor] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [anon_sym_for] = ACTIONS(4065),
    [anon_sym_foreach] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4063),
    [anon_sym___END__] = ACTIONS(4063),
    [anon_sym_EOT] = ACTIONS(4063),
    [anon_sym_if] = ACTIONS(4063),
    [anon_sym_unless] = ACTIONS(4063),
    [anon_sym_while] = ACTIONS(4063),
    [anon_sym_until] = ACTIONS(4063),
    [sym__PERLY_SEMICOLON] = ACTIONS(4063),
    [sym__ctrl_z_hack] = ACTIONS(4063),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1928)] = {
    [sym_heredoc_content] = STATE(1928),
    [anon_sym_RBRACE] = ACTIONS(4067),
    [anon_sym_SEMI] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_RPAREN] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_and] = ACTIONS(4067),
    [anon_sym_or] = ACTIONS(4067),
    [anon_sym_xor] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [anon_sym_for] = ACTIONS(4069),
    [anon_sym_foreach] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4067),
    [anon_sym___END__] = ACTIONS(4067),
    [anon_sym_EOT] = ACTIONS(4067),
    [anon_sym_if] = ACTIONS(4067),
    [anon_sym_unless] = ACTIONS(4067),
    [anon_sym_while] = ACTIONS(4067),
    [anon_sym_until] = ACTIONS(4067),
    [sym__PERLY_SEMICOLON] = ACTIONS(4067),
    [sym__ctrl_z_hack] = ACTIONS(4067),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1929)] = {
    [sym_heredoc_content] = STATE(1929),
    [anon_sym_RBRACE] = ACTIONS(4071),
    [anon_sym_SEMI] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_RPAREN] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_and] = ACTIONS(4071),
    [anon_sym_or] = ACTIONS(4071),
    [anon_sym_xor] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [anon_sym_for] = ACTIONS(4073),
    [anon_sym_foreach] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4071),
    [anon_sym___END__] = ACTIONS(4071),
    [anon_sym_EOT] = ACTIONS(4071),
    [anon_sym_if] = ACTIONS(4071),
    [anon_sym_unless] = ACTIONS(4071),
    [anon_sym_while] = ACTIONS(4071),
    [anon_sym_until] = ACTIONS(4071),
    [sym__PERLY_SEMICOLON] = ACTIONS(4071),
    [sym__ctrl_z_hack] = ACTIONS(4071),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1930)] = {
    [sym_heredoc_content] = STATE(1930),
    [anon_sym_RBRACE] = ACTIONS(4075),
    [anon_sym_SEMI] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_RPAREN] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_and] = ACTIONS(4075),
    [anon_sym_or] = ACTIONS(4075),
    [anon_sym_xor] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [anon_sym_for] = ACTIONS(4077),
    [anon_sym_foreach] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4075),
    [anon_sym___END__] = ACTIONS(4075),
    [anon_sym_EOT] = ACTIONS(4075),
    [anon_sym_if] = ACTIONS(4075),
    [anon_sym_unless] = ACTIONS(4075),
    [anon_sym_while] = ACTIONS(4075),
    [anon_sym_until] = ACTIONS(4075),
    [sym__PERLY_SEMICOLON] = ACTIONS(4075),
    [sym__ctrl_z_hack] = ACTIONS(4075),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1931)] = {
    [sym_heredoc_content] = STATE(1931),
    [anon_sym_RBRACE] = ACTIONS(4079),
    [anon_sym_SEMI] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_RPAREN] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_and] = ACTIONS(4079),
    [anon_sym_or] = ACTIONS(4079),
    [anon_sym_xor] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [anon_sym_for] = ACTIONS(4081),
    [anon_sym_foreach] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4079),
    [anon_sym___END__] = ACTIONS(4079),
    [anon_sym_EOT] = ACTIONS(4079),
    [anon_sym_if] = ACTIONS(4079),
    [anon_sym_unless] = ACTIONS(4079),
    [anon_sym_while] = ACTIONS(4079),
    [anon_sym_until] = ACTIONS(4079),
    [sym__PERLY_SEMICOLON] = ACTIONS(4079),
    [sym__ctrl_z_hack] = ACTIONS(4079),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1932)] = {
    [sym_heredoc_content] = STATE(1932),
    [anon_sym_RBRACE] = ACTIONS(4083),
    [anon_sym_SEMI] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_RPAREN] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_and] = ACTIONS(4083),
    [anon_sym_or] = ACTIONS(4083),
    [anon_sym_xor] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [anon_sym_for] = ACTIONS(4085),
    [anon_sym_foreach] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4083),
    [anon_sym___END__] = ACTIONS(4083),
    [anon_sym_EOT] = ACTIONS(4083),
    [anon_sym_if] = ACTIONS(4083),
    [anon_sym_unless] = ACTIONS(4083),
    [anon_sym_while] = ACTIONS(4083),
    [anon_sym_until] = ACTIONS(4083),
    [sym__PERLY_SEMICOLON] = ACTIONS(4083),
    [sym__ctrl_z_hack] = ACTIONS(4083),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1933)] = {
    [sym_heredoc_content] = STATE(1933),
    [anon_sym_RBRACE] = ACTIONS(4087),
    [anon_sym_SEMI] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_RPAREN] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_and] = ACTIONS(4087),
    [anon_sym_or] = ACTIONS(4087),
    [anon_sym_xor] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [anon_sym_for] = ACTIONS(4089),
    [anon_sym_foreach] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4087),
    [anon_sym___END__] = ACTIONS(4087),
    [anon_sym_EOT] = ACTIONS(4087),
    [anon_sym_if] = ACTIONS(4087),
    [anon_sym_unless] = ACTIONS(4087),
    [anon_sym_while] = ACTIONS(4087),
    [anon_sym_until] = ACTIONS(4087),
    [sym__PERLY_SEMICOLON] = ACTIONS(4087),
    [sym__ctrl_z_hack] = ACTIONS(4087),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1934)] = {
    [sym_heredoc_content] = STATE(1934),
    [anon_sym_RBRACE] = ACTIONS(4091),
    [anon_sym_SEMI] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_RPAREN] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_and] = ACTIONS(4091),
    [anon_sym_or] = ACTIONS(4091),
    [anon_sym_xor] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [anon_sym_for] = ACTIONS(4093),
    [anon_sym_foreach] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4091),
    [anon_sym___END__] = ACTIONS(4091),
    [anon_sym_EOT] = ACTIONS(4091),
    [anon_sym_if] = ACTIONS(4091),
    [anon_sym_unless] = ACTIONS(4091),
    [anon_sym_while] = ACTIONS(4091),
    [anon_sym_until] = ACTIONS(4091),
    [sym__PERLY_SEMICOLON] = ACTIONS(4091),
    [sym__ctrl_z_hack] = ACTIONS(4091),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1935)] = {
    [sym_heredoc_content] = STATE(1935),
    [anon_sym_RBRACE] = ACTIONS(4095),
    [anon_sym_SEMI] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_RPAREN] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_and] = ACTIONS(4095),
    [anon_sym_or] = ACTIONS(4095),
    [anon_sym_xor] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [anon_sym_for] = ACTIONS(4097),
    [anon_sym_foreach] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4095),
    [anon_sym___END__] = ACTIONS(4095),
    [anon_sym_EOT] = ACTIONS(4095),
    [anon_sym_if] = ACTIONS(4095),
    [anon_sym_unless] = ACTIONS(4095),
    [anon_sym_while] = ACTIONS(4095),
    [anon_sym_until] = ACTIONS(4095),
    [sym__PERLY_SEMICOLON] = ACTIONS(4095),
    [sym__ctrl_z_hack] = ACTIONS(4095),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1936)] = {
    [sym_heredoc_content] = STATE(1936),
    [anon_sym_RBRACE] = ACTIONS(4099),
    [anon_sym_SEMI] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_RPAREN] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_and] = ACTIONS(4099),
    [anon_sym_or] = ACTIONS(4099),
    [anon_sym_xor] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [anon_sym_for] = ACTIONS(4101),
    [anon_sym_foreach] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4099),
    [anon_sym___END__] = ACTIONS(4099),
    [anon_sym_EOT] = ACTIONS(4099),
    [anon_sym_if] = ACTIONS(4099),
    [anon_sym_unless] = ACTIONS(4099),
    [anon_sym_while] = ACTIONS(4099),
    [anon_sym_until] = ACTIONS(4099),
    [sym__PERLY_SEMICOLON] = ACTIONS(4099),
    [sym__ctrl_z_hack] = ACTIONS(4099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1937)] = {
    [sym_heredoc_content] = STATE(1937),
    [anon_sym_RBRACE] = ACTIONS(4103),
    [anon_sym_SEMI] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_RPAREN] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_and] = ACTIONS(4103),
    [anon_sym_or] = ACTIONS(4103),
    [anon_sym_xor] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [anon_sym_for] = ACTIONS(4105),
    [anon_sym_foreach] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4103),
    [anon_sym___END__] = ACTIONS(4103),
    [anon_sym_EOT] = ACTIONS(4103),
    [anon_sym_if] = ACTIONS(4103),
    [anon_sym_unless] = ACTIONS(4103),
    [anon_sym_while] = ACTIONS(4103),
    [anon_sym_until] = ACTIONS(4103),
    [sym__PERLY_SEMICOLON] = ACTIONS(4103),
    [sym__ctrl_z_hack] = ACTIONS(4103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1938)] = {
    [sym_heredoc_content] = STATE(1938),
    [anon_sym_RBRACE] = ACTIONS(4107),
    [anon_sym_SEMI] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [anon_sym_for] = ACTIONS(4109),
    [anon_sym_foreach] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4107),
    [anon_sym___END__] = ACTIONS(4107),
    [anon_sym_EOT] = ACTIONS(4107),
    [anon_sym_if] = ACTIONS(4107),
    [anon_sym_unless] = ACTIONS(4107),
    [anon_sym_while] = ACTIONS(4107),
    [anon_sym_until] = ACTIONS(4107),
    [sym__PERLY_SEMICOLON] = ACTIONS(4107),
    [sym__ctrl_z_hack] = ACTIONS(4107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1939)] = {
    [sym_heredoc_content] = STATE(1939),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3487),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3494),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3489),
    [anon_sym_PLUS_EQ] = ACTIONS(3489),
    [anon_sym_DASH_EQ] = ACTIONS(3489),
    [anon_sym_DOT_EQ] = ACTIONS(3489),
    [anon_sym_STAR_EQ] = ACTIONS(3489),
    [anon_sym_SLASH_EQ] = ACTIONS(3489),
    [anon_sym_PERCENT_EQ] = ACTIONS(3489),
    [anon_sym_x_EQ] = ACTIONS(3489),
    [anon_sym_AMP_EQ] = ACTIONS(3489),
    [anon_sym_PIPE_EQ] = ACTIONS(3489),
    [anon_sym_CARET_EQ] = ACTIONS(3489),
    [anon_sym_LT_LT_EQ] = ACTIONS(3489),
    [anon_sym_GT_GT_EQ] = ACTIONS(3489),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3489),
    [anon_sym_DOT_DOT] = ACTIONS(3504),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3534),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [anon_sym_for] = ACTIONS(3592),
    [anon_sym_foreach] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3590),
    [anon_sym___END__] = ACTIONS(3590),
    [anon_sym_EOT] = ACTIONS(3590),
    [anon_sym_if] = ACTIONS(3590),
    [anon_sym_unless] = ACTIONS(3590),
    [anon_sym_while] = ACTIONS(3590),
    [anon_sym_until] = ACTIONS(3590),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym__ctrl_z_hack] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1940)] = {
    [sym_heredoc_content] = STATE(1940),
    [anon_sym_RBRACE] = ACTIONS(4111),
    [anon_sym_SEMI] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_RPAREN] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_and] = ACTIONS(4111),
    [anon_sym_or] = ACTIONS(4111),
    [anon_sym_xor] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [anon_sym_for] = ACTIONS(4113),
    [anon_sym_foreach] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4111),
    [anon_sym___END__] = ACTIONS(4111),
    [anon_sym_EOT] = ACTIONS(4111),
    [anon_sym_if] = ACTIONS(4111),
    [anon_sym_unless] = ACTIONS(4111),
    [anon_sym_while] = ACTIONS(4111),
    [anon_sym_until] = ACTIONS(4111),
    [sym__PERLY_SEMICOLON] = ACTIONS(4111),
    [sym__ctrl_z_hack] = ACTIONS(4111),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1941)] = {
    [sym_heredoc_content] = STATE(1941),
    [anon_sym_RBRACE] = ACTIONS(4115),
    [anon_sym_SEMI] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_RPAREN] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_and] = ACTIONS(4115),
    [anon_sym_or] = ACTIONS(4115),
    [anon_sym_xor] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [anon_sym_for] = ACTIONS(4117),
    [anon_sym_foreach] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4115),
    [anon_sym___END__] = ACTIONS(4115),
    [anon_sym_EOT] = ACTIONS(4115),
    [anon_sym_if] = ACTIONS(4115),
    [anon_sym_unless] = ACTIONS(4115),
    [anon_sym_while] = ACTIONS(4115),
    [anon_sym_until] = ACTIONS(4115),
    [sym__PERLY_SEMICOLON] = ACTIONS(4115),
    [sym__ctrl_z_hack] = ACTIONS(4115),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1942)] = {
    [sym_heredoc_content] = STATE(1942),
    [anon_sym_RBRACE] = ACTIONS(4119),
    [anon_sym_SEMI] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_RPAREN] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_and] = ACTIONS(4119),
    [anon_sym_or] = ACTIONS(4119),
    [anon_sym_xor] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [anon_sym_for] = ACTIONS(4121),
    [anon_sym_foreach] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4119),
    [anon_sym___END__] = ACTIONS(4119),
    [anon_sym_EOT] = ACTIONS(4119),
    [anon_sym_if] = ACTIONS(4119),
    [anon_sym_unless] = ACTIONS(4119),
    [anon_sym_while] = ACTIONS(4119),
    [anon_sym_until] = ACTIONS(4119),
    [sym__PERLY_SEMICOLON] = ACTIONS(4119),
    [sym__ctrl_z_hack] = ACTIONS(4119),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1943)] = {
    [sym_heredoc_content] = STATE(1943),
    [anon_sym_RBRACE] = ACTIONS(4123),
    [anon_sym_SEMI] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_RPAREN] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_and] = ACTIONS(4123),
    [anon_sym_or] = ACTIONS(4123),
    [anon_sym_xor] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [anon_sym_for] = ACTIONS(4125),
    [anon_sym_foreach] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4123),
    [anon_sym___END__] = ACTIONS(4123),
    [anon_sym_EOT] = ACTIONS(4123),
    [anon_sym_if] = ACTIONS(4123),
    [anon_sym_unless] = ACTIONS(4123),
    [anon_sym_while] = ACTIONS(4123),
    [anon_sym_until] = ACTIONS(4123),
    [sym__PERLY_SEMICOLON] = ACTIONS(4123),
    [sym__ctrl_z_hack] = ACTIONS(4123),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1944)] = {
    [sym_heredoc_content] = STATE(1944),
    [anon_sym_RBRACE] = ACTIONS(4127),
    [anon_sym_SEMI] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_RPAREN] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_and] = ACTIONS(4127),
    [anon_sym_or] = ACTIONS(4127),
    [anon_sym_xor] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [anon_sym_for] = ACTIONS(4129),
    [anon_sym_foreach] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4127),
    [anon_sym___END__] = ACTIONS(4127),
    [anon_sym_EOT] = ACTIONS(4127),
    [anon_sym_if] = ACTIONS(4127),
    [anon_sym_unless] = ACTIONS(4127),
    [anon_sym_while] = ACTIONS(4127),
    [anon_sym_until] = ACTIONS(4127),
    [sym__PERLY_SEMICOLON] = ACTIONS(4127),
    [sym__ctrl_z_hack] = ACTIONS(4127),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1945)] = {
    [sym_heredoc_content] = STATE(1945),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [anon_sym_for] = ACTIONS(3536),
    [anon_sym_foreach] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3485),
    [anon_sym___END__] = ACTIONS(3485),
    [anon_sym_EOT] = ACTIONS(3485),
    [anon_sym_if] = ACTIONS(3485),
    [anon_sym_unless] = ACTIONS(3485),
    [anon_sym_while] = ACTIONS(3485),
    [anon_sym_until] = ACTIONS(3485),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym__ctrl_z_hack] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1946)] = {
    [sym_heredoc_content] = STATE(1946),
    [anon_sym_RBRACE] = ACTIONS(4131),
    [anon_sym_SEMI] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_RPAREN] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_and] = ACTIONS(4131),
    [anon_sym_or] = ACTIONS(4131),
    [anon_sym_xor] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [anon_sym_for] = ACTIONS(4133),
    [anon_sym_foreach] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4131),
    [anon_sym___END__] = ACTIONS(4131),
    [anon_sym_EOT] = ACTIONS(4131),
    [anon_sym_if] = ACTIONS(4131),
    [anon_sym_unless] = ACTIONS(4131),
    [anon_sym_while] = ACTIONS(4131),
    [anon_sym_until] = ACTIONS(4131),
    [sym__PERLY_SEMICOLON] = ACTIONS(4131),
    [sym__ctrl_z_hack] = ACTIONS(4131),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1947)] = {
    [sym_heredoc_content] = STATE(1947),
    [anon_sym_RBRACE] = ACTIONS(4135),
    [anon_sym_SEMI] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_RPAREN] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_and] = ACTIONS(4135),
    [anon_sym_or] = ACTIONS(4135),
    [anon_sym_xor] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [anon_sym_for] = ACTIONS(4137),
    [anon_sym_foreach] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4135),
    [anon_sym___END__] = ACTIONS(4135),
    [anon_sym_EOT] = ACTIONS(4135),
    [anon_sym_if] = ACTIONS(4135),
    [anon_sym_unless] = ACTIONS(4135),
    [anon_sym_while] = ACTIONS(4135),
    [anon_sym_until] = ACTIONS(4135),
    [sym__PERLY_SEMICOLON] = ACTIONS(4135),
    [sym__ctrl_z_hack] = ACTIONS(4135),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1948)] = {
    [sym_heredoc_content] = STATE(1948),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4139),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(3500),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4141),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(3508),
    [anon_sym_SLASH_SLASH] = ACTIONS(3508),
    [anon_sym_CARET_CARET] = ACTIONS(3510),
    [anon_sym_AMP_AMP] = ACTIONS(3512),
    [anon_sym_PIPE] = ACTIONS(3514),
    [anon_sym_CARET] = ACTIONS(3514),
    [anon_sym_AMP] = ACTIONS(3516),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(3522),
    [anon_sym_BANG_EQ] = ACTIONS(3524),
    [anon_sym_eq] = ACTIONS(3522),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3524),
    [anon_sym_equ] = ACTIONS(3524),
    [anon_sym_eqr] = ACTIONS(3524),
    [anon_sym_ne] = ACTIONS(3524),
    [anon_sym_LT_EQ_GT] = ACTIONS(3526),
    [anon_sym_cmp] = ACTIONS(3526),
    [anon_sym_TILDE_TILDE] = ACTIONS(3526),
    [anon_sym_LT] = ACTIONS(3500),
    [anon_sym_LT_EQ] = ACTIONS(3500),
    [anon_sym_GT_EQ] = ACTIONS(3528),
    [anon_sym_lt] = ACTIONS(3528),
    [anon_sym_le] = ACTIONS(3528),
    [anon_sym_ge] = ACTIONS(3528),
    [anon_sym_gt] = ACTIONS(3528),
    [anon_sym_isa] = ACTIONS(3530),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [anon_sym_for] = ACTIONS(3721),
    [anon_sym_foreach] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [anon_sym_if] = ACTIONS(3719),
    [anon_sym_unless] = ACTIONS(3719),
    [anon_sym_while] = ACTIONS(3719),
    [anon_sym_until] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1949)] = {
    [sym_heredoc_content] = STATE(1949),
    [anon_sym_RBRACE] = ACTIONS(4143),
    [anon_sym_SEMI] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_RPAREN] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_and] = ACTIONS(4143),
    [anon_sym_or] = ACTIONS(4143),
    [anon_sym_xor] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [anon_sym_for] = ACTIONS(4145),
    [anon_sym_foreach] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4143),
    [anon_sym___END__] = ACTIONS(4143),
    [anon_sym_EOT] = ACTIONS(4143),
    [anon_sym_if] = ACTIONS(4143),
    [anon_sym_unless] = ACTIONS(4143),
    [anon_sym_while] = ACTIONS(4143),
    [anon_sym_until] = ACTIONS(4143),
    [sym__PERLY_SEMICOLON] = ACTIONS(4143),
    [sym__ctrl_z_hack] = ACTIONS(4143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1950)] = {
    [sym_heredoc_content] = STATE(1950),
    [anon_sym_RBRACE] = ACTIONS(4147),
    [anon_sym_SEMI] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_RPAREN] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_and] = ACTIONS(4147),
    [anon_sym_or] = ACTIONS(4147),
    [anon_sym_xor] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [anon_sym_for] = ACTIONS(4149),
    [anon_sym_foreach] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4147),
    [anon_sym___END__] = ACTIONS(4147),
    [anon_sym_EOT] = ACTIONS(4147),
    [anon_sym_if] = ACTIONS(4147),
    [anon_sym_unless] = ACTIONS(4147),
    [anon_sym_while] = ACTIONS(4147),
    [anon_sym_until] = ACTIONS(4147),
    [sym__PERLY_SEMICOLON] = ACTIONS(4147),
    [sym__ctrl_z_hack] = ACTIONS(4147),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1951)] = {
    [sym_heredoc_content] = STATE(1951),
    [anon_sym_RBRACE] = ACTIONS(4151),
    [anon_sym_SEMI] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3498),
    [anon_sym_PERCENT] = ACTIONS(3498),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(3502),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(3506),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(3502),
    [anon_sym_PLUS] = ACTIONS(3518),
    [anon_sym_DASH] = ACTIONS(3518),
    [anon_sym_DOT] = ACTIONS(3518),
    [anon_sym_SLASH] = ACTIONS(3498),
    [anon_sym_x] = ACTIONS(3498),
    [anon_sym_EQ_TILDE] = ACTIONS(3520),
    [anon_sym_BANG_TILDE] = ACTIONS(3520),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [anon_sym_for] = ACTIONS(4153),
    [anon_sym_foreach] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4151),
    [anon_sym___END__] = ACTIONS(4151),
    [anon_sym_EOT] = ACTIONS(4151),
    [anon_sym_if] = ACTIONS(4151),
    [anon_sym_unless] = ACTIONS(4151),
    [anon_sym_while] = ACTIONS(4151),
    [anon_sym_until] = ACTIONS(4151),
    [sym__PERLY_SEMICOLON] = ACTIONS(4151),
    [sym__ctrl_z_hack] = ACTIONS(4151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1952)] = {
    [sym_heredoc_content] = STATE(1952),
    [anon_sym_RBRACE] = ACTIONS(4155),
    [anon_sym_SEMI] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_RPAREN] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_and] = ACTIONS(4155),
    [anon_sym_or] = ACTIONS(4155),
    [anon_sym_xor] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [anon_sym_for] = ACTIONS(4157),
    [anon_sym_foreach] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4155),
    [anon_sym___END__] = ACTIONS(4155),
    [anon_sym_EOT] = ACTIONS(4155),
    [anon_sym_if] = ACTIONS(4155),
    [anon_sym_unless] = ACTIONS(4155),
    [anon_sym_while] = ACTIONS(4155),
    [anon_sym_until] = ACTIONS(4155),
    [sym__PERLY_SEMICOLON] = ACTIONS(4155),
    [sym__ctrl_z_hack] = ACTIONS(4155),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1953)] = {
    [sym_heredoc_content] = STATE(1953),
    [anon_sym_RBRACE] = ACTIONS(4159),
    [anon_sym_SEMI] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_RPAREN] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_and] = ACTIONS(4159),
    [anon_sym_or] = ACTIONS(4159),
    [anon_sym_xor] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [anon_sym_for] = ACTIONS(4162),
    [anon_sym_foreach] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4159),
    [anon_sym___END__] = ACTIONS(4159),
    [anon_sym_EOT] = ACTIONS(4159),
    [anon_sym_if] = ACTIONS(4159),
    [anon_sym_unless] = ACTIONS(4159),
    [anon_sym_while] = ACTIONS(4159),
    [anon_sym_until] = ACTIONS(4159),
    [sym__PERLY_SEMICOLON] = ACTIONS(4159),
    [sym__ctrl_z_hack] = ACTIONS(4159),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1954)] = {
    [sym_heredoc_content] = STATE(1954),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [anon_sym_for] = ACTIONS(2534),
    [anon_sym_foreach] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(2536),
    [anon_sym___END__] = ACTIONS(2536),
    [anon_sym_EOT] = ACTIONS(2536),
    [anon_sym_if] = ACTIONS(2536),
    [anon_sym_unless] = ACTIONS(2536),
    [anon_sym_while] = ACTIONS(2536),
    [anon_sym_until] = ACTIONS(2536),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym__ctrl_z_hack] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1955)] = {
    [sym_heredoc_content] = STATE(1955),
    [anon_sym_RBRACE] = ACTIONS(4165),
    [anon_sym_SEMI] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_RPAREN] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_and] = ACTIONS(4165),
    [anon_sym_or] = ACTIONS(4165),
    [anon_sym_xor] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [anon_sym_for] = ACTIONS(4167),
    [anon_sym_foreach] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4165),
    [anon_sym___END__] = ACTIONS(4165),
    [anon_sym_EOT] = ACTIONS(4165),
    [anon_sym_if] = ACTIONS(4165),
    [anon_sym_unless] = ACTIONS(4165),
    [anon_sym_while] = ACTIONS(4165),
    [anon_sym_until] = ACTIONS(4165),
    [sym__PERLY_SEMICOLON] = ACTIONS(4165),
    [sym__ctrl_z_hack] = ACTIONS(4165),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1956)] = {
    [sym_heredoc_content] = STATE(1956),
    [anon_sym_RBRACE] = ACTIONS(4169),
    [anon_sym_SEMI] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_RPAREN] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_and] = ACTIONS(4169),
    [anon_sym_or] = ACTIONS(4169),
    [anon_sym_xor] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(4173),
    [anon_sym_DASH_DASH] = ACTIONS(4173),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [anon_sym_for] = ACTIONS(4171),
    [anon_sym_foreach] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4169),
    [anon_sym___END__] = ACTIONS(4169),
    [anon_sym_EOT] = ACTIONS(4169),
    [anon_sym_if] = ACTIONS(4169),
    [anon_sym_unless] = ACTIONS(4169),
    [anon_sym_while] = ACTIONS(4169),
    [anon_sym_until] = ACTIONS(4169),
    [sym__PERLY_SEMICOLON] = ACTIONS(4169),
    [sym__ctrl_z_hack] = ACTIONS(4169),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1957)] = {
    [sym_heredoc_content] = STATE(1957),
    [anon_sym_RBRACE] = ACTIONS(4176),
    [anon_sym_SEMI] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_RPAREN] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_and] = ACTIONS(4176),
    [anon_sym_or] = ACTIONS(4176),
    [anon_sym_xor] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [anon_sym_for] = ACTIONS(4178),
    [anon_sym_foreach] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4176),
    [anon_sym___END__] = ACTIONS(4176),
    [anon_sym_EOT] = ACTIONS(4176),
    [anon_sym_if] = ACTIONS(4176),
    [anon_sym_unless] = ACTIONS(4176),
    [anon_sym_while] = ACTIONS(4176),
    [anon_sym_until] = ACTIONS(4176),
    [sym__PERLY_SEMICOLON] = ACTIONS(4176),
    [sym__ctrl_z_hack] = ACTIONS(4176),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1958)] = {
    [sym_heredoc_content] = STATE(1958),
    [anon_sym_RBRACE] = ACTIONS(4180),
    [anon_sym_SEMI] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_RPAREN] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_and] = ACTIONS(4180),
    [anon_sym_or] = ACTIONS(4180),
    [anon_sym_xor] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [anon_sym_for] = ACTIONS(4182),
    [anon_sym_foreach] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4180),
    [anon_sym___END__] = ACTIONS(4180),
    [anon_sym_EOT] = ACTIONS(4180),
    [anon_sym_if] = ACTIONS(4180),
    [anon_sym_unless] = ACTIONS(4180),
    [anon_sym_while] = ACTIONS(4180),
    [anon_sym_until] = ACTIONS(4180),
    [sym__PERLY_SEMICOLON] = ACTIONS(4180),
    [sym__ctrl_z_hack] = ACTIONS(4180),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1959)] = {
    [sym_heredoc_content] = STATE(1959),
    [anon_sym_RBRACE] = ACTIONS(4184),
    [anon_sym_SEMI] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_RPAREN] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_and] = ACTIONS(4184),
    [anon_sym_or] = ACTIONS(4184),
    [anon_sym_xor] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [anon_sym_for] = ACTIONS(4186),
    [anon_sym_foreach] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4184),
    [anon_sym___END__] = ACTIONS(4184),
    [anon_sym_EOT] = ACTIONS(4184),
    [anon_sym_if] = ACTIONS(4184),
    [anon_sym_unless] = ACTIONS(4184),
    [anon_sym_while] = ACTIONS(4184),
    [anon_sym_until] = ACTIONS(4184),
    [sym__PERLY_SEMICOLON] = ACTIONS(4184),
    [sym__ctrl_z_hack] = ACTIONS(4184),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1960)] = {
    [sym_heredoc_content] = STATE(1960),
    [anon_sym_RBRACE] = ACTIONS(4188),
    [anon_sym_SEMI] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_RPAREN] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_and] = ACTIONS(4188),
    [anon_sym_or] = ACTIONS(4188),
    [anon_sym_xor] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [anon_sym_for] = ACTIONS(4190),
    [anon_sym_foreach] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4188),
    [anon_sym___END__] = ACTIONS(4188),
    [anon_sym_EOT] = ACTIONS(4188),
    [anon_sym_if] = ACTIONS(4188),
    [anon_sym_unless] = ACTIONS(4188),
    [anon_sym_while] = ACTIONS(4188),
    [anon_sym_until] = ACTIONS(4188),
    [sym__PERLY_SEMICOLON] = ACTIONS(4188),
    [sym__ctrl_z_hack] = ACTIONS(4188),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1961)] = {
    [sym_heredoc_content] = STATE(1961),
    [anon_sym_RBRACE] = ACTIONS(4192),
    [anon_sym_SEMI] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_RPAREN] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_and] = ACTIONS(4192),
    [anon_sym_or] = ACTIONS(4192),
    [anon_sym_xor] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [anon_sym_for] = ACTIONS(4194),
    [anon_sym_foreach] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4192),
    [anon_sym___END__] = ACTIONS(4192),
    [anon_sym_EOT] = ACTIONS(4192),
    [anon_sym_if] = ACTIONS(4192),
    [anon_sym_unless] = ACTIONS(4192),
    [anon_sym_while] = ACTIONS(4192),
    [anon_sym_until] = ACTIONS(4192),
    [sym__PERLY_SEMICOLON] = ACTIONS(4192),
    [sym__ctrl_z_hack] = ACTIONS(4192),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1962)] = {
    [sym_heredoc_content] = STATE(1962),
    [anon_sym_RBRACE] = ACTIONS(4196),
    [anon_sym_SEMI] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_RPAREN] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_and] = ACTIONS(4196),
    [anon_sym_or] = ACTIONS(4196),
    [anon_sym_xor] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [anon_sym_for] = ACTIONS(4198),
    [anon_sym_foreach] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4196),
    [anon_sym___END__] = ACTIONS(4196),
    [anon_sym_EOT] = ACTIONS(4196),
    [anon_sym_if] = ACTIONS(4196),
    [anon_sym_unless] = ACTIONS(4196),
    [anon_sym_while] = ACTIONS(4196),
    [anon_sym_until] = ACTIONS(4196),
    [sym__PERLY_SEMICOLON] = ACTIONS(4196),
    [sym__ctrl_z_hack] = ACTIONS(4196),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1963)] = {
    [sym_heredoc_content] = STATE(1963),
    [anon_sym_RBRACE] = ACTIONS(4200),
    [anon_sym_SEMI] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_RPAREN] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_and] = ACTIONS(4200),
    [anon_sym_or] = ACTIONS(4200),
    [anon_sym_xor] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [anon_sym_for] = ACTIONS(4202),
    [anon_sym_foreach] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4200),
    [anon_sym___END__] = ACTIONS(4200),
    [anon_sym_EOT] = ACTIONS(4200),
    [anon_sym_if] = ACTIONS(4200),
    [anon_sym_unless] = ACTIONS(4200),
    [anon_sym_while] = ACTIONS(4200),
    [anon_sym_until] = ACTIONS(4200),
    [sym__PERLY_SEMICOLON] = ACTIONS(4200),
    [sym__ctrl_z_hack] = ACTIONS(4200),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1964)] = {
    [sym_heredoc_content] = STATE(1964),
    [anon_sym_RBRACE] = ACTIONS(4204),
    [anon_sym_SEMI] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_RPAREN] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_and] = ACTIONS(4204),
    [anon_sym_or] = ACTIONS(4204),
    [anon_sym_xor] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [anon_sym_for] = ACTIONS(4206),
    [anon_sym_foreach] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4204),
    [anon_sym___END__] = ACTIONS(4204),
    [anon_sym_EOT] = ACTIONS(4204),
    [anon_sym_if] = ACTIONS(4204),
    [anon_sym_unless] = ACTIONS(4204),
    [anon_sym_while] = ACTIONS(4204),
    [anon_sym_until] = ACTIONS(4204),
    [sym__PERLY_SEMICOLON] = ACTIONS(4204),
    [sym__ctrl_z_hack] = ACTIONS(4204),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1965)] = {
    [sym_heredoc_content] = STATE(1965),
    [anon_sym_RBRACE] = ACTIONS(4208),
    [anon_sym_SEMI] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_RPAREN] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_and] = ACTIONS(4208),
    [anon_sym_or] = ACTIONS(4208),
    [anon_sym_xor] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [anon_sym_for] = ACTIONS(4210),
    [anon_sym_foreach] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4208),
    [anon_sym___END__] = ACTIONS(4208),
    [anon_sym_EOT] = ACTIONS(4208),
    [anon_sym_if] = ACTIONS(4208),
    [anon_sym_unless] = ACTIONS(4208),
    [anon_sym_while] = ACTIONS(4208),
    [anon_sym_until] = ACTIONS(4208),
    [sym__PERLY_SEMICOLON] = ACTIONS(4208),
    [sym__ctrl_z_hack] = ACTIONS(4208),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1966)] = {
    [sym_heredoc_content] = STATE(1966),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [anon_sym_for] = ACTIONS(3417),
    [anon_sym_foreach] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3415),
    [anon_sym___END__] = ACTIONS(3415),
    [anon_sym_EOT] = ACTIONS(3415),
    [anon_sym_if] = ACTIONS(3415),
    [anon_sym_unless] = ACTIONS(3415),
    [anon_sym_while] = ACTIONS(3415),
    [anon_sym_until] = ACTIONS(3415),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym__ctrl_z_hack] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1967)] = {
    [sym_heredoc_content] = STATE(1967),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_SEMI] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [anon_sym_for] = ACTIONS(2856),
    [anon_sym_foreach] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4045),
    [anon_sym___END__] = ACTIONS(4045),
    [anon_sym_EOT] = ACTIONS(4045),
    [anon_sym_if] = ACTIONS(4045),
    [anon_sym_unless] = ACTIONS(4045),
    [anon_sym_while] = ACTIONS(4045),
    [anon_sym_until] = ACTIONS(4045),
    [sym__PERLY_SEMICOLON] = ACTIONS(4045),
    [sym__ctrl_z_hack] = ACTIONS(4045),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1968)] = {
    [sym_heredoc_content] = STATE(1968),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_SEMI] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [anon_sym_for] = ACTIONS(3919),
    [anon_sym_foreach] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3917),
    [anon_sym___END__] = ACTIONS(3917),
    [anon_sym_EOT] = ACTIONS(3917),
    [anon_sym_if] = ACTIONS(3917),
    [anon_sym_unless] = ACTIONS(3917),
    [anon_sym_while] = ACTIONS(3917),
    [anon_sym_until] = ACTIONS(3917),
    [sym__PERLY_SEMICOLON] = ACTIONS(3917),
    [sym__ctrl_z_hack] = ACTIONS(3917),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1969)] = {
    [sym_attribute] = STATE(1981),
    [sym_attribute_name] = STATE(1975),
    [sym_heredoc_content] = STATE(1969),
    [aux_sym_attrlist_repeat1] = STATE(1971),
    [sym__identifier] = ACTIONS(4214),
    [anon_sym_COLON] = ACTIONS(4216),
    [anon_sym_SEMI] = ACTIONS(3313),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3317),
    [anon_sym___END__] = ACTIONS(3317),
    [anon_sym_EOT] = ACTIONS(3313),
    [aux_sym__bareword_token1] = ACTIONS(4214),
    [sym__PERLY_SEMICOLON] = ACTIONS(3313),
    [sym__ctrl_z_hack] = ACTIONS(3313),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1970)] = {
    [sym_attribute] = STATE(1981),
    [sym_attribute_name] = STATE(1975),
    [sym_heredoc_content] = STATE(1970),
    [aux_sym_attrlist_repeat1] = STATE(1970),
    [sym__identifier] = ACTIONS(4218),
    [anon_sym_COLON] = ACTIONS(4221),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3309),
    [anon_sym___END__] = ACTIONS(3309),
    [anon_sym_EOT] = ACTIONS(3304),
    [aux_sym__bareword_token1] = ACTIONS(4218),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym__ctrl_z_hack] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1971)] = {
    [sym_attribute] = STATE(1981),
    [sym_attribute_name] = STATE(1975),
    [sym_heredoc_content] = STATE(1971),
    [aux_sym_attrlist_repeat1] = STATE(1970),
    [sym__identifier] = ACTIONS(4214),
    [anon_sym_COLON] = ACTIONS(4216),
    [anon_sym_SEMI] = ACTIONS(3319),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3321),
    [anon_sym___END__] = ACTIONS(3321),
    [anon_sym_EOT] = ACTIONS(3319),
    [aux_sym__bareword_token1] = ACTIONS(4214),
    [sym__PERLY_SEMICOLON] = ACTIONS(3319),
    [sym__ctrl_z_hack] = ACTIONS(3319),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1972)] = {
    [sym_attrlist] = STATE(1889),
    [sym_attribute] = STATE(1969),
    [sym_attribute_name] = STATE(1975),
    [sym_heredoc_content] = STATE(1972),
    [sym__identifier] = ACTIONS(4214),
    [anon_sym_SEMI] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3325),
    [anon_sym___END__] = ACTIONS(3325),
    [anon_sym_EOT] = ACTIONS(3323),
    [aux_sym__bareword_token1] = ACTIONS(4214),
    [sym__PERLY_SEMICOLON] = ACTIONS(3323),
    [sym__ctrl_z_hack] = ACTIONS(3323),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1973)] = {
    [sym_attrlist] = STATE(1888),
    [sym_attribute] = STATE(1969),
    [sym_attribute_name] = STATE(1975),
    [sym_heredoc_content] = STATE(1973),
    [sym__identifier] = ACTIONS(4214),
    [anon_sym_SEMI] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3329),
    [anon_sym___END__] = ACTIONS(3329),
    [anon_sym_EOT] = ACTIONS(3327),
    [aux_sym__bareword_token1] = ACTIONS(4214),
    [sym__PERLY_SEMICOLON] = ACTIONS(3327),
    [sym__ctrl_z_hack] = ACTIONS(3327),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1974)] = {
    [sym_heredoc_content] = STATE(1974),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_SEMI] = ACTIONS(3333),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3331),
    [anon_sym___END__] = ACTIONS(3331),
    [anon_sym_EOT] = ACTIONS(3333),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym__PERLY_SEMICOLON] = ACTIONS(3333),
    [sym__ctrl_z_hack] = ACTIONS(3333),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1975)] = {
    [sym_heredoc_content] = STATE(1975),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_SEMI] = ACTIONS(3337),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3335),
    [anon_sym___END__] = ACTIONS(3335),
    [anon_sym_EOT] = ACTIONS(3337),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym__PERLY_SEMICOLON] = ACTIONS(3337),
    [sym__ctrl_z_hack] = ACTIONS(3337),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4224),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1976)] = {
    [sym_attribute] = STATE(2057),
    [sym_attribute_name] = STATE(2006),
    [sym_heredoc_content] = STATE(1976),
    [aux_sym_attrlist_repeat1] = STATE(1980),
    [sym__identifier] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(3319),
    [anon_sym_COLON] = ACTIONS(4228),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_RPAREN] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_RBRACK] = ACTIONS(3319),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4226),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1977)] = {
    [sym_heredoc_content] = STATE(1977),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_SEMI] = ACTIONS(3403),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3401),
    [anon_sym___END__] = ACTIONS(3401),
    [anon_sym_EOT] = ACTIONS(3403),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym__PERLY_SEMICOLON] = ACTIONS(3403),
    [sym__ctrl_z_hack] = ACTIONS(3403),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1978)] = {
    [sym_heredoc_content] = STATE(1978),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3309),
    [anon_sym___END__] = ACTIONS(3309),
    [anon_sym_EOT] = ACTIONS(3304),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym__ctrl_z_hack] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1979)] = {
    [sym_attribute] = STATE(2057),
    [sym_attribute_name] = STATE(2006),
    [sym_heredoc_content] = STATE(1979),
    [aux_sym_attrlist_repeat1] = STATE(1976),
    [sym__identifier] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(3313),
    [anon_sym_COLON] = ACTIONS(4228),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_RPAREN] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_RBRACK] = ACTIONS(3313),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4226),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1980)] = {
    [sym_attribute] = STATE(2057),
    [sym_attribute_name] = STATE(2006),
    [sym_heredoc_content] = STATE(1980),
    [aux_sym_attrlist_repeat1] = STATE(1980),
    [sym__identifier] = ACTIONS(4230),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(4233),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_RBRACK] = ACTIONS(3304),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4230),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1981)] = {
    [sym_heredoc_content] = STATE(1981),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_SEMI] = ACTIONS(3411),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3409),
    [anon_sym___END__] = ACTIONS(3409),
    [anon_sym_EOT] = ACTIONS(3411),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym__PERLY_SEMICOLON] = ACTIONS(3411),
    [sym__ctrl_z_hack] = ACTIONS(3411),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1982)] = {
    [sym_attrlist] = STATE(2403),
    [sym_attribute] = STATE(1979),
    [sym_attribute_name] = STATE(2006),
    [sym_heredoc_content] = STATE(1982),
    [sym__identifier] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_RPAREN] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_RBRACK] = ACTIONS(3327),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4226),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1983)] = {
    [sym_attrlist] = STATE(2404),
    [sym_attribute] = STATE(1979),
    [sym_attribute_name] = STATE(2006),
    [sym_heredoc_content] = STATE(1983),
    [sym__identifier] = ACTIONS(4226),
    [anon_sym_RBRACE] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_RPAREN] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_RBRACK] = ACTIONS(3323),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4226),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1984)] = {
    [sym_attribute] = STATE(2133),
    [sym_attribute_name] = STATE(2064),
    [sym_heredoc_content] = STATE(1984),
    [aux_sym_attrlist_repeat1] = STATE(1991),
    [sym__identifier] = ACTIONS(4236),
    [anon_sym_COLON] = ACTIONS(4238),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_RBRACK] = ACTIONS(3313),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4236),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3313),
  },
  [STATE(1985)] = {
    [sym_attribute] = STATE(2134),
    [sym_attribute_name] = STATE(2063),
    [sym_heredoc_content] = STATE(1985),
    [aux_sym_attrlist_repeat1] = STATE(1986),
    [sym__identifier] = ACTIONS(4240),
    [anon_sym_RBRACE] = ACTIONS(3313),
    [anon_sym_COLON] = ACTIONS(4242),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4240),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3313),
  },
  [STATE(1986)] = {
    [sym_attribute] = STATE(2134),
    [sym_attribute_name] = STATE(2063),
    [sym_heredoc_content] = STATE(1986),
    [aux_sym_attrlist_repeat1] = STATE(1987),
    [sym__identifier] = ACTIONS(4240),
    [anon_sym_RBRACE] = ACTIONS(3319),
    [anon_sym_COLON] = ACTIONS(4242),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4240),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3319),
  },
  [STATE(1987)] = {
    [sym_attribute] = STATE(2134),
    [sym_attribute_name] = STATE(2063),
    [sym_heredoc_content] = STATE(1987),
    [aux_sym_attrlist_repeat1] = STATE(1987),
    [sym__identifier] = ACTIONS(4244),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(4247),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4244),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3304),
  },
  [STATE(1988)] = {
    [sym_attribute] = STATE(2163),
    [sym_attribute_name] = STATE(2019),
    [sym_heredoc_content] = STATE(1988),
    [aux_sym_attrlist_repeat1] = STATE(1989),
    [sym__identifier] = ACTIONS(4250),
    [anon_sym_COLON] = ACTIONS(4252),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_RPAREN] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_and] = ACTIONS(3317),
    [anon_sym_or] = ACTIONS(3317),
    [anon_sym_xor] = ACTIONS(3317),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4250),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3313),
  },
  [STATE(1989)] = {
    [sym_attribute] = STATE(2163),
    [sym_attribute_name] = STATE(2019),
    [sym_heredoc_content] = STATE(1989),
    [aux_sym_attrlist_repeat1] = STATE(1990),
    [sym__identifier] = ACTIONS(4250),
    [anon_sym_COLON] = ACTIONS(4252),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_RPAREN] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4250),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3319),
  },
  [STATE(1990)] = {
    [sym_attribute] = STATE(2163),
    [sym_attribute_name] = STATE(2019),
    [sym_heredoc_content] = STATE(1990),
    [aux_sym_attrlist_repeat1] = STATE(1990),
    [sym__identifier] = ACTIONS(4254),
    [anon_sym_COLON] = ACTIONS(4257),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4254),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3304),
  },
  [STATE(1991)] = {
    [sym_attribute] = STATE(2133),
    [sym_attribute_name] = STATE(2064),
    [sym_heredoc_content] = STATE(1991),
    [aux_sym_attrlist_repeat1] = STATE(1992),
    [sym__identifier] = ACTIONS(4236),
    [anon_sym_COLON] = ACTIONS(4238),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_and] = ACTIONS(3321),
    [anon_sym_or] = ACTIONS(3321),
    [anon_sym_xor] = ACTIONS(3321),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_RBRACK] = ACTIONS(3319),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4236),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3319),
  },
  [STATE(1992)] = {
    [sym_attribute] = STATE(2133),
    [sym_attribute_name] = STATE(2064),
    [sym_heredoc_content] = STATE(1992),
    [aux_sym_attrlist_repeat1] = STATE(1992),
    [sym__identifier] = ACTIONS(4260),
    [anon_sym_COLON] = ACTIONS(4263),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_RBRACK] = ACTIONS(3304),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4260),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3304),
  },
  [STATE(1993)] = {
    [sym_heredoc_content] = STATE(1993),
    [aux_sym__term_rightward_repeat1] = STATE(2003),
    [anon_sym_SEMI] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4266),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4266),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3568),
    [anon_sym___END__] = ACTIONS(3568),
    [anon_sym_EOT] = ACTIONS(3568),
    [sym__PERLY_SEMICOLON] = ACTIONS(3568),
    [sym__ctrl_z_hack] = ACTIONS(3568),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1994)] = {
    [sym_heredoc_content] = STATE(1994),
    [anon_sym_COLON] = ACTIONS(4268),
    [anon_sym_SEMI] = ACTIONS(3669),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3669),
    [anon_sym___END__] = ACTIONS(3669),
    [anon_sym_EOT] = ACTIONS(3669),
    [sym__PERLY_SEMICOLON] = ACTIONS(3669),
    [sym__ctrl_z_hack] = ACTIONS(3669),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1995)] = {
    [sym_heredoc_content] = STATE(1995),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_RBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_RPAREN] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_and] = ACTIONS(2719),
    [anon_sym_or] = ACTIONS(2719),
    [anon_sym_xor] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_RBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1996)] = {
    [sym_heredoc_content] = STATE(1996),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_RBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_RPAREN] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_and] = ACTIONS(2715),
    [anon_sym_or] = ACTIONS(2715),
    [anon_sym_xor] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_RBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1997)] = {
    [sym_attribute] = STATE(2251),
    [sym_attribute_name] = STATE(2167),
    [sym_heredoc_content] = STATE(1997),
    [aux_sym_attrlist_repeat1] = STATE(1997),
    [sym__identifier] = ACTIONS(4270),
    [anon_sym_COLON] = ACTIONS(4273),
    [anon_sym_DOLLAR] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_AT] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3309),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_PERCENT2] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4270),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1998)] = {
    [sym_heredoc_content] = STATE(1998),
    [aux_sym__term_rightward_repeat1] = STATE(1993),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(4280),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(4280),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3485),
    [anon_sym___END__] = ACTIONS(3485),
    [anon_sym_EOT] = ACTIONS(3485),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym__ctrl_z_hack] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(1999)] = {
    [sym_attribute] = STATE(2251),
    [sym_attribute_name] = STATE(2167),
    [sym_heredoc_content] = STATE(1999),
    [aux_sym_attrlist_repeat1] = STATE(2000),
    [sym__identifier] = ACTIONS(4321),
    [anon_sym_COLON] = ACTIONS(3317),
    [anon_sym_DOLLAR] = ACTIONS(3313),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_AT] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_RPAREN] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3317),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_PERCENT2] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4321),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2000)] = {
    [sym_attribute] = STATE(2251),
    [sym_attribute_name] = STATE(2167),
    [sym_heredoc_content] = STATE(2000),
    [aux_sym_attrlist_repeat1] = STATE(1997),
    [sym__identifier] = ACTIONS(4321),
    [anon_sym_COLON] = ACTIONS(3321),
    [anon_sym_DOLLAR] = ACTIONS(3319),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_AT] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_RPAREN] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3321),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_PERCENT2] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4321),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2001)] = {
    [sym_attrlist] = STATE(3205),
    [sym_attribute] = STATE(1985),
    [sym_attribute_name] = STATE(2063),
    [sym_heredoc_content] = STATE(2001),
    [sym__identifier] = ACTIONS(4240),
    [anon_sym_RBRACE] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4240),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3327),
  },
  [STATE(2002)] = {
    [sym_attrlist] = STATE(3206),
    [sym_attribute] = STATE(1985),
    [sym_attribute_name] = STATE(2063),
    [sym_heredoc_content] = STATE(2002),
    [sym__identifier] = ACTIONS(4240),
    [anon_sym_RBRACE] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4240),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3323),
  },
  [STATE(2003)] = {
    [sym_heredoc_content] = STATE(2003),
    [aux_sym__term_rightward_repeat1] = STATE(2003),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4323),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3590),
    [anon_sym___END__] = ACTIONS(3590),
    [anon_sym_EOT] = ACTIONS(3590),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym__ctrl_z_hack] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2004)] = {
    [sym_attrlist] = STATE(3020),
    [sym_attribute] = STATE(1984),
    [sym_attribute_name] = STATE(2064),
    [sym_heredoc_content] = STATE(2004),
    [sym__identifier] = ACTIONS(4236),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_RBRACK] = ACTIONS(3327),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4236),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3327),
  },
  [STATE(2005)] = {
    [sym_attrlist] = STATE(3021),
    [sym_attribute] = STATE(1984),
    [sym_attribute_name] = STATE(2064),
    [sym_heredoc_content] = STATE(2005),
    [sym__identifier] = ACTIONS(4236),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_RBRACK] = ACTIONS(3323),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4236),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3323),
  },
  [STATE(2006)] = {
    [sym_heredoc_content] = STATE(2006),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_RBRACE] = ACTIONS(3337),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_RPAREN] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_RBRACK] = ACTIONS(3337),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4326),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2007)] = {
    [sym_heredoc_content] = STATE(2007),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_RBRACE] = ACTIONS(3333),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_RPAREN] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_RBRACK] = ACTIONS(3333),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2008)] = {
    [sym_attrlist] = STATE(2744),
    [sym_attribute] = STATE(1988),
    [sym_attribute_name] = STATE(2019),
    [sym_heredoc_content] = STATE(2008),
    [sym__identifier] = ACTIONS(4250),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_RPAREN] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_and] = ACTIONS(3329),
    [anon_sym_or] = ACTIONS(3329),
    [anon_sym_xor] = ACTIONS(3329),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4250),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3327),
  },
  [STATE(2009)] = {
    [sym_attrlist] = STATE(2745),
    [sym_attribute] = STATE(1988),
    [sym_attribute_name] = STATE(2019),
    [sym_heredoc_content] = STATE(2009),
    [sym__identifier] = ACTIONS(4250),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_RPAREN] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_and] = ACTIONS(3325),
    [anon_sym_or] = ACTIONS(3325),
    [anon_sym_xor] = ACTIONS(3325),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4250),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3323),
  },
  [STATE(2010)] = {
    [sym_heredoc_content] = STATE(2010),
    [aux_sym__term_rightward_repeat1] = STATE(4251),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(4328),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(4328),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3485),
    [anon_sym___END__] = ACTIONS(3485),
    [anon_sym_EOT] = ACTIONS(3485),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym__ctrl_z_hack] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2011)] = {
    [sym_heredoc_content] = STATE(2011),
    [anon_sym_COLON] = ACTIONS(4330),
    [anon_sym_SEMI] = ACTIONS(3689),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3689),
    [anon_sym___END__] = ACTIONS(3689),
    [anon_sym_EOT] = ACTIONS(3689),
    [sym__PERLY_SEMICOLON] = ACTIONS(3689),
    [sym__ctrl_z_hack] = ACTIONS(3689),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2012)] = {
    [sym_heredoc_content] = STATE(2012),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3590),
    [anon_sym___END__] = ACTIONS(3590),
    [anon_sym_EOT] = ACTIONS(3590),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym__ctrl_z_hack] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2013)] = {
    [sym_heredoc_content] = STATE(2013),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_RBRACE] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_RPAREN] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_and] = ACTIONS(2690),
    [anon_sym_or] = ACTIONS(2690),
    [anon_sym_xor] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_RBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2014)] = {
    [sym_heredoc_content] = STATE(2014),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_RBRACE] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_RPAREN] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_and] = ACTIONS(2694),
    [anon_sym_or] = ACTIONS(2694),
    [anon_sym_xor] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_RBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2015)] = {
    [sym_heredoc_content] = STATE(2015),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_RBRACE] = ACTIONS(3403),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_RPAREN] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_RBRACK] = ACTIONS(3403),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2016)] = {
    [sym_heredoc_content] = STATE(2016),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_RBRACE] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_RPAREN] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_and] = ACTIONS(3381),
    [anon_sym_or] = ACTIONS(3381),
    [anon_sym_xor] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_RBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2017)] = {
    [sym_heredoc_content] = STATE(2017),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_RPAREN] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3333),
  },
  [STATE(2018)] = {
    [sym_heredoc_content] = STATE(2018),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_RBRACE] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_RPAREN] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_and] = ACTIONS(3397),
    [anon_sym_or] = ACTIONS(3397),
    [anon_sym_xor] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_RBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2019)] = {
    [sym_heredoc_content] = STATE(2019),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_RPAREN] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4332),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3337),
  },
  [STATE(2020)] = {
    [sym_heredoc_content] = STATE(2020),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_RBRACE] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_RPAREN] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_and] = ACTIONS(3349),
    [anon_sym_or] = ACTIONS(3349),
    [anon_sym_xor] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_RBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2021)] = {
    [sym_heredoc_content] = STATE(2021),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_RBRACE] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_RPAREN] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_and] = ACTIONS(3365),
    [anon_sym_or] = ACTIONS(3365),
    [anon_sym_xor] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_RBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2022)] = {
    [sym_heredoc_content] = STATE(2022),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_RBRACE] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_RPAREN] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_and] = ACTIONS(3373),
    [anon_sym_or] = ACTIONS(3373),
    [anon_sym_xor] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_RBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2023)] = {
    [sym_heredoc_content] = STATE(2023),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_RBRACE] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_RPAREN] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_and] = ACTIONS(3341),
    [anon_sym_or] = ACTIONS(3341),
    [anon_sym_xor] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_RBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2024)] = {
    [sym_heredoc_content] = STATE(2024),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_RBRACE] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_RPAREN] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_and] = ACTIONS(3353),
    [anon_sym_or] = ACTIONS(3353),
    [anon_sym_xor] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_RBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2025)] = {
    [sym_heredoc_content] = STATE(2025),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_RBRACE] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_RPAREN] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_and] = ACTIONS(2530),
    [anon_sym_or] = ACTIONS(2530),
    [anon_sym_xor] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_RBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2026)] = {
    [sym_heredoc_content] = STATE(2026),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_and] = ACTIONS(2715),
    [anon_sym_or] = ACTIONS(2715),
    [anon_sym_xor] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_RBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2715),
  },
  [STATE(2027)] = {
    [sym_heredoc_content] = STATE(2027),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_RBRACE] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_RPAREN] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_and] = ACTIONS(3357),
    [anon_sym_or] = ACTIONS(3357),
    [anon_sym_xor] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_RBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2028)] = {
    [sym_heredoc_content] = STATE(2028),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_RBRACE] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_RPAREN] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_and] = ACTIONS(3369),
    [anon_sym_or] = ACTIONS(3369),
    [anon_sym_xor] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_RBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2029)] = {
    [sym_heredoc_content] = STATE(2029),
    [anon_sym_LBRACE] = ACTIONS(4334),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4336),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4338),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2030)] = {
    [sym_heredoc_content] = STATE(2030),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_RBRACE] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_RPAREN] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_and] = ACTIONS(2698),
    [anon_sym_or] = ACTIONS(2698),
    [anon_sym_xor] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_RBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2031)] = {
    [sym_heredoc_content] = STATE(2031),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_RBRACE] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_RPAREN] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_and] = ACTIONS(3345),
    [anon_sym_or] = ACTIONS(3345),
    [anon_sym_xor] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_RBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2032)] = {
    [sym_heredoc_content] = STATE(2032),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_RBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2033)] = {
    [sym_heredoc_content] = STATE(2033),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_RBRACK] = ACTIONS(3333),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3333),
  },
  [STATE(2034)] = {
    [sym_heredoc_content] = STATE(2034),
    [anon_sym_SEMI] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4107),
    [anon_sym___END__] = ACTIONS(4107),
    [anon_sym_EOT] = ACTIONS(4107),
    [sym__PERLY_SEMICOLON] = ACTIONS(4107),
    [sym__ctrl_z_hack] = ACTIONS(4107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2035)] = {
    [sym_heredoc_content] = STATE(2035),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_RBRACK] = ACTIONS(3304),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2036)] = {
    [sym_heredoc_content] = STATE(2036),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_RBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_and] = ACTIONS(2719),
    [anon_sym_or] = ACTIONS(2719),
    [anon_sym_xor] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2719),
  },
  [STATE(2037)] = {
    [sym_heredoc_content] = STATE(2037),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_RBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_and] = ACTIONS(2715),
    [anon_sym_or] = ACTIONS(2715),
    [anon_sym_xor] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2715),
  },
  [STATE(2038)] = {
    [sym_heredoc_content] = STATE(2038),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_RPAREN] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_and] = ACTIONS(2719),
    [anon_sym_or] = ACTIONS(2719),
    [anon_sym_xor] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2719),
  },
  [STATE(2039)] = {
    [sym_heredoc_content] = STATE(2039),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_RPAREN] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_and] = ACTIONS(2715),
    [anon_sym_or] = ACTIONS(2715),
    [anon_sym_xor] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2715),
  },
  [STATE(2040)] = {
    [sym_heredoc_content] = STATE(2040),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_RBRACE] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_RPAREN] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_and] = ACTIONS(3405),
    [anon_sym_or] = ACTIONS(3405),
    [anon_sym_xor] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_RBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2041)] = {
    [sym_heredoc_content] = STATE(2041),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4340),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4342),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2042)] = {
    [sym_heredoc_content] = STATE(2042),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2043)] = {
    [sym_heredoc_content] = STATE(2043),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3735),
    [anon_sym___END__] = ACTIONS(3735),
    [anon_sym_EOT] = ACTIONS(3735),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym__ctrl_z_hack] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2044)] = {
    [sym_heredoc_content] = STATE(2044),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2045)] = {
    [sym_heredoc_content] = STATE(2045),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2046)] = {
    [sym_heredoc_content] = STATE(2046),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2047)] = {
    [sym_heredoc_content] = STATE(2047),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2048)] = {
    [sym_heredoc_content] = STATE(2048),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_and] = ACTIONS(2719),
    [anon_sym_or] = ACTIONS(2719),
    [anon_sym_xor] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_RBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2719),
  },
  [STATE(2049)] = {
    [sym_heredoc_content] = STATE(2049),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2050)] = {
    [sym_heredoc_content] = STATE(2050),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2051)] = {
    [sym_heredoc_content] = STATE(2051),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2052)] = {
    [sym_heredoc_content] = STATE(2052),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3763),
    [anon_sym___END__] = ACTIONS(3763),
    [anon_sym_EOT] = ACTIONS(3763),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym__ctrl_z_hack] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2053)] = {
    [sym_heredoc_content] = STATE(2053),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4344),
    [anon_sym_cmp] = ACTIONS(4344),
    [anon_sym_TILDE_TILDE] = ACTIONS(4344),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3763),
    [anon_sym___END__] = ACTIONS(3763),
    [anon_sym_EOT] = ACTIONS(3763),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym__ctrl_z_hack] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2054)] = {
    [sym_heredoc_content] = STATE(2054),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4346),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3735),
    [anon_sym___END__] = ACTIONS(3735),
    [anon_sym_EOT] = ACTIONS(3735),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym__ctrl_z_hack] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2055)] = {
    [sym_heredoc_content] = STATE(2055),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_RBRACE] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_RPAREN] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_and] = ACTIONS(3393),
    [anon_sym_or] = ACTIONS(3393),
    [anon_sym_xor] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_RBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2056)] = {
    [sym_heredoc_content] = STATE(2056),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_RBRACE] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_RPAREN] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_and] = ACTIONS(3385),
    [anon_sym_or] = ACTIONS(3385),
    [anon_sym_xor] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_RBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2057)] = {
    [sym_heredoc_content] = STATE(2057),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_RBRACE] = ACTIONS(3411),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_RPAREN] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_RBRACK] = ACTIONS(3411),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2058)] = {
    [sym_heredoc_content] = STATE(2058),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_RBRACE] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_RPAREN] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_and] = ACTIONS(3389),
    [anon_sym_or] = ACTIONS(3389),
    [anon_sym_xor] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_RBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2059)] = {
    [sym_heredoc_content] = STATE(2059),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_RBRACE] = ACTIONS(3333),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_and] = ACTIONS(3331),
    [anon_sym_or] = ACTIONS(3331),
    [anon_sym_xor] = ACTIONS(3331),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3333),
  },
  [STATE(2060)] = {
    [sym_heredoc_content] = STATE(2060),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_RBRACE] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_RPAREN] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_and] = ACTIONS(3361),
    [anon_sym_or] = ACTIONS(3361),
    [anon_sym_xor] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_RBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2061)] = {
    [sym_heredoc_content] = STATE(2061),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_RBRACE] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_RPAREN] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_and] = ACTIONS(3377),
    [anon_sym_or] = ACTIONS(3377),
    [anon_sym_xor] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_RBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2062)] = {
    [sym_heredoc_content] = STATE(2062),
    [anon_sym_SEMI] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3977),
    [anon_sym___END__] = ACTIONS(3977),
    [anon_sym_EOT] = ACTIONS(3977),
    [sym__PERLY_SEMICOLON] = ACTIONS(3977),
    [sym__ctrl_z_hack] = ACTIONS(3977),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2063)] = {
    [sym_heredoc_content] = STATE(2063),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_RBRACE] = ACTIONS(3337),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4348),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3337),
  },
  [STATE(2064)] = {
    [sym_heredoc_content] = STATE(2064),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_and] = ACTIONS(3335),
    [anon_sym_or] = ACTIONS(3335),
    [anon_sym_xor] = ACTIONS(3335),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_RBRACK] = ACTIONS(3337),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4350),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3337),
  },
  [STATE(2065)] = {
    [sym_heredoc_content] = STATE(2065),
    [anon_sym_SEMI] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3997),
    [anon_sym___END__] = ACTIONS(3997),
    [anon_sym_EOT] = ACTIONS(3997),
    [sym__PERLY_SEMICOLON] = ACTIONS(3997),
    [sym__ctrl_z_hack] = ACTIONS(3997),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2066)] = {
    [sym_heredoc_content] = STATE(2066),
    [anon_sym_SEMI] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3695),
    [anon_sym___END__] = ACTIONS(3695),
    [anon_sym_EOT] = ACTIONS(3695),
    [sym__PERLY_SEMICOLON] = ACTIONS(3695),
    [sym__ctrl_z_hack] = ACTIONS(3695),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2067)] = {
    [sym_heredoc_content] = STATE(2067),
    [anon_sym_SEMI] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3909),
    [anon_sym___END__] = ACTIONS(3909),
    [anon_sym_EOT] = ACTIONS(3909),
    [sym__PERLY_SEMICOLON] = ACTIONS(3909),
    [sym__ctrl_z_hack] = ACTIONS(3909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2068)] = {
    [sym_heredoc_content] = STATE(2068),
    [anon_sym_SEMI] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4017),
    [anon_sym___END__] = ACTIONS(4017),
    [anon_sym_EOT] = ACTIONS(4017),
    [sym__PERLY_SEMICOLON] = ACTIONS(4017),
    [sym__ctrl_z_hack] = ACTIONS(4017),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2069)] = {
    [sym_heredoc_content] = STATE(2069),
    [anon_sym_SEMI] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3803),
    [anon_sym___END__] = ACTIONS(3803),
    [anon_sym_EOT] = ACTIONS(3803),
    [sym__PERLY_SEMICOLON] = ACTIONS(3803),
    [sym__ctrl_z_hack] = ACTIONS(3803),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2070)] = {
    [sym_heredoc_content] = STATE(2070),
    [anon_sym_SEMI] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3845),
    [anon_sym___END__] = ACTIONS(3845),
    [anon_sym_EOT] = ACTIONS(3845),
    [sym__PERLY_SEMICOLON] = ACTIONS(3845),
    [sym__ctrl_z_hack] = ACTIONS(3845),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2071)] = {
    [sym_heredoc_content] = STATE(2071),
    [anon_sym_SEMI] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3739),
    [anon_sym___END__] = ACTIONS(3739),
    [anon_sym_EOT] = ACTIONS(3739),
    [sym__PERLY_SEMICOLON] = ACTIONS(3739),
    [sym__ctrl_z_hack] = ACTIONS(3739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2072)] = {
    [sym_heredoc_content] = STATE(2072),
    [anon_sym_SEMI] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(4276),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4278),
    [anon_sym_PLUS_EQ] = ACTIONS(4278),
    [anon_sym_DASH_EQ] = ACTIONS(4278),
    [anon_sym_DOT_EQ] = ACTIONS(4278),
    [anon_sym_STAR_EQ] = ACTIONS(4278),
    [anon_sym_SLASH_EQ] = ACTIONS(4278),
    [anon_sym_PERCENT_EQ] = ACTIONS(4278),
    [anon_sym_x_EQ] = ACTIONS(4278),
    [anon_sym_AMP_EQ] = ACTIONS(4278),
    [anon_sym_PIPE_EQ] = ACTIONS(4278),
    [anon_sym_CARET_EQ] = ACTIONS(4278),
    [anon_sym_LT_LT_EQ] = ACTIONS(4278),
    [anon_sym_GT_GT_EQ] = ACTIONS(4278),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4278),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3897),
    [anon_sym___END__] = ACTIONS(3897),
    [anon_sym_EOT] = ACTIONS(3897),
    [sym__PERLY_SEMICOLON] = ACTIONS(3897),
    [sym__ctrl_z_hack] = ACTIONS(3897),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2073)] = {
    [sym_heredoc_content] = STATE(2073),
    [anon_sym_SEMI] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(4151),
    [anon_sym___END__] = ACTIONS(4151),
    [anon_sym_EOT] = ACTIONS(4151),
    [sym__PERLY_SEMICOLON] = ACTIONS(4151),
    [sym__ctrl_z_hack] = ACTIONS(4151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2074)] = {
    [sym_heredoc_content] = STATE(2074),
    [anon_sym_SEMI] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(4352),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(4352),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3819),
    [anon_sym___END__] = ACTIONS(3819),
    [anon_sym_EOT] = ACTIONS(3819),
    [sym__PERLY_SEMICOLON] = ACTIONS(3819),
    [sym__ctrl_z_hack] = ACTIONS(3819),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2075)] = {
    [sym_heredoc_content] = STATE(2075),
    [anon_sym_SEMI] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4283),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4291),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(4295),
    [anon_sym_SLASH_SLASH] = ACTIONS(4295),
    [anon_sym_CARET_CARET] = ACTIONS(4297),
    [anon_sym_AMP_AMP] = ACTIONS(4299),
    [anon_sym_PIPE] = ACTIONS(4301),
    [anon_sym_CARET] = ACTIONS(4301),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(4319),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3985),
    [anon_sym___END__] = ACTIONS(3985),
    [anon_sym_EOT] = ACTIONS(3985),
    [sym__PERLY_SEMICOLON] = ACTIONS(3985),
    [sym__ctrl_z_hack] = ACTIONS(3985),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2076)] = {
    [sym_heredoc_content] = STATE(2076),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_RBRACE] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_RPAREN] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_and] = ACTIONS(2723),
    [anon_sym_or] = ACTIONS(2723),
    [anon_sym_xor] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_RBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2077)] = {
    [sym_heredoc_content] = STATE(2077),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(3496),
    [anon_sym_STAR] = ACTIONS(4285),
    [anon_sym_PERCENT] = ACTIONS(4285),
    [anon_sym_GT] = ACTIONS(4287),
    [anon_sym_LT_LT] = ACTIONS(4289),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4293),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4303),
    [anon_sym_GT_GT2] = ACTIONS(4289),
    [anon_sym_PLUS] = ACTIONS(4305),
    [anon_sym_DASH] = ACTIONS(4305),
    [anon_sym_DOT] = ACTIONS(4305),
    [anon_sym_SLASH] = ACTIONS(4285),
    [anon_sym_x] = ACTIONS(4285),
    [anon_sym_EQ_TILDE] = ACTIONS(4307),
    [anon_sym_BANG_TILDE] = ACTIONS(4307),
    [anon_sym_EQ_EQ] = ACTIONS(4309),
    [anon_sym_BANG_EQ] = ACTIONS(4311),
    [anon_sym_eq] = ACTIONS(4309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4311),
    [anon_sym_equ] = ACTIONS(4311),
    [anon_sym_eqr] = ACTIONS(4311),
    [anon_sym_ne] = ACTIONS(4311),
    [anon_sym_LT_EQ_GT] = ACTIONS(4313),
    [anon_sym_cmp] = ACTIONS(4313),
    [anon_sym_TILDE_TILDE] = ACTIONS(4313),
    [anon_sym_LT] = ACTIONS(4287),
    [anon_sym_LT_EQ] = ACTIONS(4287),
    [anon_sym_GT_EQ] = ACTIONS(4315),
    [anon_sym_lt] = ACTIONS(4315),
    [anon_sym_le] = ACTIONS(4315),
    [anon_sym_ge] = ACTIONS(4315),
    [anon_sym_gt] = ACTIONS(4315),
    [anon_sym_isa] = ACTIONS(4317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3532),
    [anon_sym_DASH_DASH] = ACTIONS(3532),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [anon_sym___DATA__] = ACTIONS(3719),
    [anon_sym___END__] = ACTIONS(3719),
    [anon_sym_EOT] = ACTIONS(3719),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym__ctrl_z_hack] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2078)] = {
    [sym_heredoc_content] = STATE(2078),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_RPAREN] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_and] = ACTIONS(3357),
    [anon_sym_or] = ACTIONS(3357),
    [anon_sym_xor] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3357),
  },
  [STATE(2079)] = {
    [sym_heredoc_content] = STATE(2079),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_RBRACK] = ACTIONS(3403),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3403),
  },
  [STATE(2080)] = {
    [sym_heredoc_content] = STATE(2080),
    [anon_sym_LBRACE] = ACTIONS(4354),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4356),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4358),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3415),
  },
  [STATE(2081)] = {
    [sym_heredoc_content] = STATE(2081),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_and] = ACTIONS(3345),
    [anon_sym_or] = ACTIONS(3345),
    [anon_sym_xor] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_RBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3345),
  },
  [STATE(2082)] = {
    [sym_attribute] = STATE(2743),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2082),
    [aux_sym_attrlist_repeat1] = STATE(2087),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_RBRACE] = ACTIONS(3313),
    [anon_sym_COLON] = ACTIONS(4362),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_RPAREN] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2083)] = {
    [sym_heredoc_content] = STATE(2083),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_and] = ACTIONS(2723),
    [anon_sym_or] = ACTIONS(2723),
    [anon_sym_xor] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_RBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2723),
  },
  [STATE(2084)] = {
    [sym_heredoc_content] = STATE(2084),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_and] = ACTIONS(2698),
    [anon_sym_or] = ACTIONS(2698),
    [anon_sym_xor] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_RBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2698),
  },
  [STATE(2085)] = {
    [sym_heredoc_content] = STATE(2085),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3435),
    [anon_sym_or] = ACTIONS(3435),
    [anon_sym_xor] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2086)] = {
    [sym_heredoc_content] = STATE(2086),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_RPAREN] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_and] = ACTIONS(3361),
    [anon_sym_or] = ACTIONS(3361),
    [anon_sym_xor] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3361),
  },
  [STATE(2087)] = {
    [sym_attribute] = STATE(2743),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2087),
    [aux_sym_attrlist_repeat1] = STATE(2098),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_RBRACE] = ACTIONS(3319),
    [anon_sym_COLON] = ACTIONS(4362),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_RPAREN] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2088)] = {
    [sym_heredoc_content] = STATE(2088),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_RPAREN] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_and] = ACTIONS(3397),
    [anon_sym_or] = ACTIONS(3397),
    [anon_sym_xor] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3397),
  },
  [STATE(2089)] = {
    [sym_heredoc_content] = STATE(2089),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_RPAREN] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_and] = ACTIONS(3349),
    [anon_sym_or] = ACTIONS(3349),
    [anon_sym_xor] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3349),
  },
  [STATE(2090)] = {
    [sym_heredoc_content] = STATE(2090),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_RPAREN] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_and] = ACTIONS(3365),
    [anon_sym_or] = ACTIONS(3365),
    [anon_sym_xor] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3365),
  },
  [STATE(2091)] = {
    [sym_heredoc_content] = STATE(2091),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_and] = ACTIONS(3361),
    [anon_sym_or] = ACTIONS(3361),
    [anon_sym_xor] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_RBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3361),
  },
  [STATE(2092)] = {
    [sym_heredoc_content] = STATE(2092),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_RPAREN] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_and] = ACTIONS(3385),
    [anon_sym_or] = ACTIONS(3385),
    [anon_sym_xor] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3385),
  },
  [STATE(2093)] = {
    [sym_heredoc_content] = STATE(2093),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_RPAREN] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_and] = ACTIONS(3389),
    [anon_sym_or] = ACTIONS(3389),
    [anon_sym_xor] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3389),
  },
  [STATE(2094)] = {
    [sym_heredoc_content] = STATE(2094),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_RPAREN] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_and] = ACTIONS(3405),
    [anon_sym_or] = ACTIONS(3405),
    [anon_sym_xor] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3405),
  },
  [STATE(2095)] = {
    [sym_heredoc_content] = STATE(2095),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_RBRACE] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_and] = ACTIONS(3361),
    [anon_sym_or] = ACTIONS(3361),
    [anon_sym_xor] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3361),
  },
  [STATE(2096)] = {
    [sym_heredoc_content] = STATE(2096),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_RPAREN] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_and] = ACTIONS(3373),
    [anon_sym_or] = ACTIONS(3373),
    [anon_sym_xor] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3373),
  },
  [STATE(2097)] = {
    [sym_heredoc_content] = STATE(2097),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_RPAREN] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_and] = ACTIONS(3381),
    [anon_sym_or] = ACTIONS(3381),
    [anon_sym_xor] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3381),
  },
  [STATE(2098)] = {
    [sym_attribute] = STATE(2743),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2098),
    [aux_sym_attrlist_repeat1] = STATE(2098),
    [sym__identifier] = ACTIONS(4364),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(4367),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4364),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2099)] = {
    [sym_heredoc_content] = STATE(2099),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_and] = ACTIONS(3397),
    [anon_sym_or] = ACTIONS(3397),
    [anon_sym_xor] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_RBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3397),
  },
  [STATE(2100)] = {
    [sym_heredoc_content] = STATE(2100),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_and] = ACTIONS(3349),
    [anon_sym_or] = ACTIONS(3349),
    [anon_sym_xor] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_RBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3349),
  },
  [STATE(2101)] = {
    [sym_heredoc_content] = STATE(2101),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_and] = ACTIONS(3365),
    [anon_sym_or] = ACTIONS(3365),
    [anon_sym_xor] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_RBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3365),
  },
  [STATE(2102)] = {
    [sym_heredoc_content] = STATE(2102),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_and] = ACTIONS(3385),
    [anon_sym_or] = ACTIONS(3385),
    [anon_sym_xor] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_RBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3385),
  },
  [STATE(2103)] = {
    [sym_heredoc_content] = STATE(2103),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_and] = ACTIONS(3389),
    [anon_sym_or] = ACTIONS(3389),
    [anon_sym_xor] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_RBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3389),
  },
  [STATE(2104)] = {
    [sym_heredoc_content] = STATE(2104),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_and] = ACTIONS(3405),
    [anon_sym_or] = ACTIONS(3405),
    [anon_sym_xor] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_RBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3405),
  },
  [STATE(2105)] = {
    [sym_heredoc_content] = STATE(2105),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_RBRACE] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_and] = ACTIONS(2723),
    [anon_sym_or] = ACTIONS(2723),
    [anon_sym_xor] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2723),
  },
  [STATE(2106)] = {
    [sym_heredoc_content] = STATE(2106),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_RBRACE] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_and] = ACTIONS(2698),
    [anon_sym_or] = ACTIONS(2698),
    [anon_sym_xor] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2698),
  },
  [STATE(2107)] = {
    [sym_heredoc_content] = STATE(2107),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_and] = ACTIONS(3373),
    [anon_sym_or] = ACTIONS(3373),
    [anon_sym_xor] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_RBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3373),
  },
  [STATE(2108)] = {
    [sym_heredoc_content] = STATE(2108),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_and] = ACTIONS(3381),
    [anon_sym_or] = ACTIONS(3381),
    [anon_sym_xor] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_RBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3381),
  },
  [STATE(2109)] = {
    [sym_heredoc_content] = STATE(2109),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_RPAREN] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_and] = ACTIONS(3341),
    [anon_sym_or] = ACTIONS(3341),
    [anon_sym_xor] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3341),
  },
  [STATE(2110)] = {
    [sym_heredoc_content] = STATE(2110),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_RPAREN] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_and] = ACTIONS(2723),
    [anon_sym_or] = ACTIONS(2723),
    [anon_sym_xor] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2723),
  },
  [STATE(2111)] = {
    [sym_heredoc_content] = STATE(2111),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_RPAREN] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_and] = ACTIONS(2698),
    [anon_sym_or] = ACTIONS(2698),
    [anon_sym_xor] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2698),
  },
  [STATE(2112)] = {
    [sym_heredoc_content] = STATE(2112),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_RPAREN] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_and] = ACTIONS(3353),
    [anon_sym_or] = ACTIONS(3353),
    [anon_sym_xor] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3353),
  },
  [STATE(2113)] = {
    [sym_heredoc_content] = STATE(2113),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_and] = ACTIONS(3341),
    [anon_sym_or] = ACTIONS(3341),
    [anon_sym_xor] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_RBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3341),
  },
  [STATE(2114)] = {
    [sym_heredoc_content] = STATE(2114),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(2115)] = {
    [sym_heredoc_content] = STATE(2115),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_and] = ACTIONS(3353),
    [anon_sym_or] = ACTIONS(3353),
    [anon_sym_xor] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_RBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3353),
  },
  [STATE(2116)] = {
    [sym_heredoc_content] = STATE(2116),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_and] = ACTIONS(3357),
    [anon_sym_or] = ACTIONS(3357),
    [anon_sym_xor] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_RBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3357),
  },
  [STATE(2117)] = {
    [sym_heredoc_content] = STATE(2117),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_and] = ACTIONS(3369),
    [anon_sym_or] = ACTIONS(3369),
    [anon_sym_xor] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_RBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3369),
  },
  [STATE(2118)] = {
    [sym_heredoc_content] = STATE(2118),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_RPAREN] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_and] = ACTIONS(3369),
    [anon_sym_or] = ACTIONS(3369),
    [anon_sym_xor] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3369),
  },
  [STATE(2119)] = {
    [sym_heredoc_content] = STATE(2119),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_RBRACE] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_and] = ACTIONS(2694),
    [anon_sym_or] = ACTIONS(2694),
    [anon_sym_xor] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2694),
  },
  [STATE(2120)] = {
    [sym_heredoc_content] = STATE(2120),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_RPAREN] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_and] = ACTIONS(3393),
    [anon_sym_or] = ACTIONS(3393),
    [anon_sym_xor] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3393),
  },
  [STATE(2121)] = {
    [sym_heredoc_content] = STATE(2121),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(2122)] = {
    [sym_heredoc_content] = STATE(2122),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_RPAREN] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_and] = ACTIONS(2694),
    [anon_sym_or] = ACTIONS(2694),
    [anon_sym_xor] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2694),
  },
  [STATE(2123)] = {
    [sym_heredoc_content] = STATE(2123),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_RPAREN] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_and] = ACTIONS(3377),
    [anon_sym_or] = ACTIONS(3377),
    [anon_sym_xor] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3377),
  },
  [STATE(2124)] = {
    [sym_heredoc_content] = STATE(2124),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_and] = ACTIONS(3393),
    [anon_sym_or] = ACTIONS(3393),
    [anon_sym_xor] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_RBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3393),
  },
  [STATE(2125)] = {
    [sym_heredoc_content] = STATE(2125),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_and] = ACTIONS(3377),
    [anon_sym_or] = ACTIONS(3377),
    [anon_sym_xor] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_RBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3377),
  },
  [STATE(2126)] = {
    [sym_heredoc_content] = STATE(2126),
    [anon_sym_LBRACE] = ACTIONS(4370),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4372),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4374),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3415),
  },
  [STATE(2127)] = {
    [sym_heredoc_content] = STATE(2127),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_RBRACE] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_and] = ACTIONS(3345),
    [anon_sym_or] = ACTIONS(3345),
    [anon_sym_xor] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3345),
  },
  [STATE(2128)] = {
    [sym_heredoc_content] = STATE(2128),
    [anon_sym_LBRACE] = ACTIONS(4376),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4378),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4380),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3415),
  },
  [STATE(2129)] = {
    [sym_heredoc_content] = STATE(2129),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_RPAREN] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_and] = ACTIONS(3345),
    [anon_sym_or] = ACTIONS(3345),
    [anon_sym_xor] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3345),
  },
  [STATE(2130)] = {
    [sym_heredoc_content] = STATE(2130),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_RBRACE] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_and] = ACTIONS(2530),
    [anon_sym_or] = ACTIONS(2530),
    [anon_sym_xor] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2530),
  },
  [STATE(2131)] = {
    [sym_heredoc_content] = STATE(2131),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_COLON] = ACTIONS(3345),
    [anon_sym_DOLLAR] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_AT] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_RPAREN] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3347),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_PERCENT2] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2132)] = {
    [sym_heredoc_content] = STATE(2132),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_RPAREN] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_and] = ACTIONS(2530),
    [anon_sym_or] = ACTIONS(2530),
    [anon_sym_xor] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2530),
  },
  [STATE(2133)] = {
    [sym_heredoc_content] = STATE(2133),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_RBRACK] = ACTIONS(3411),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3411),
  },
  [STATE(2134)] = {
    [sym_heredoc_content] = STATE(2134),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_RBRACE] = ACTIONS(3411),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3411),
  },
  [STATE(2135)] = {
    [sym_heredoc_content] = STATE(2135),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_RBRACE] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_and] = ACTIONS(2690),
    [anon_sym_or] = ACTIONS(2690),
    [anon_sym_xor] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2690),
  },
  [STATE(2136)] = {
    [sym_heredoc_content] = STATE(2136),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_RBRACE] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_and] = ACTIONS(3397),
    [anon_sym_or] = ACTIONS(3397),
    [anon_sym_xor] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3397),
  },
  [STATE(2137)] = {
    [sym_heredoc_content] = STATE(2137),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_RBRACE] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_and] = ACTIONS(3349),
    [anon_sym_or] = ACTIONS(3349),
    [anon_sym_xor] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3349),
  },
  [STATE(2138)] = {
    [sym_heredoc_content] = STATE(2138),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_RBRACE] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_and] = ACTIONS(3365),
    [anon_sym_or] = ACTIONS(3365),
    [anon_sym_xor] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3365),
  },
  [STATE(2139)] = {
    [sym_heredoc_content] = STATE(2139),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_RBRACE] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_and] = ACTIONS(3385),
    [anon_sym_or] = ACTIONS(3385),
    [anon_sym_xor] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3385),
  },
  [STATE(2140)] = {
    [sym_heredoc_content] = STATE(2140),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_RBRACE] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_and] = ACTIONS(3389),
    [anon_sym_or] = ACTIONS(3389),
    [anon_sym_xor] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3389),
  },
  [STATE(2141)] = {
    [sym_heredoc_content] = STATE(2141),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_RBRACE] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_and] = ACTIONS(3405),
    [anon_sym_or] = ACTIONS(3405),
    [anon_sym_xor] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3405),
  },
  [STATE(2142)] = {
    [sym_heredoc_content] = STATE(2142),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_RPAREN] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_and] = ACTIONS(2690),
    [anon_sym_or] = ACTIONS(2690),
    [anon_sym_xor] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2690),
  },
  [STATE(2143)] = {
    [sym_heredoc_content] = STATE(2143),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_COLON] = ACTIONS(2723),
    [anon_sym_DOLLAR] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_AT] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_RPAREN] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2721),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_PERCENT2] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2144)] = {
    [sym_heredoc_content] = STATE(2144),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_COLON] = ACTIONS(2698),
    [anon_sym_DOLLAR] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_AT] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_RPAREN] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2696),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_PERCENT2] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2145)] = {
    [sym_heredoc_content] = STATE(2145),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_RBRACE] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_and] = ACTIONS(3373),
    [anon_sym_or] = ACTIONS(3373),
    [anon_sym_xor] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3373),
  },
  [STATE(2146)] = {
    [sym_heredoc_content] = STATE(2146),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_RBRACE] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_and] = ACTIONS(3381),
    [anon_sym_or] = ACTIONS(3381),
    [anon_sym_xor] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3381),
  },
  [STATE(2147)] = {
    [sym_heredoc_content] = STATE(2147),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_RBRACK] = ACTIONS(3304),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3304),
  },
  [STATE(2148)] = {
    [sym_heredoc_content] = STATE(2148),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_DOLLAR] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_AT] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_RPAREN] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2717),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_PERCENT2] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2149)] = {
    [sym_heredoc_content] = STATE(2149),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2686),
    [anon_sym_DOLLAR] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_PERCENT2] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2150)] = {
    [sym_heredoc_content] = STATE(2150),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3304),
  },
  [STATE(2151)] = {
    [sym_heredoc_content] = STATE(2151),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_DOLLAR] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_AT] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_RPAREN] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2713),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_PERCENT2] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2152)] = {
    [sym_heredoc_content] = STATE(2152),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_and] = ACTIONS(2694),
    [anon_sym_or] = ACTIONS(2694),
    [anon_sym_xor] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_RBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2694),
  },
  [STATE(2153)] = {
    [sym_heredoc_content] = STATE(2153),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_RBRACE] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_and] = ACTIONS(3341),
    [anon_sym_or] = ACTIONS(3341),
    [anon_sym_xor] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3341),
  },
  [STATE(2154)] = {
    [sym_heredoc_content] = STATE(2154),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_COLON] = ACTIONS(2694),
    [anon_sym_DOLLAR] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_AT] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_RPAREN] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2692),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_PERCENT2] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2155)] = {
    [sym_attribute] = STATE(2923),
    [sym_attribute_name] = STATE(2483),
    [sym_heredoc_content] = STATE(2155),
    [aux_sym_attrlist_repeat1] = STATE(2169),
    [sym__identifier] = ACTIONS(4382),
    [anon_sym_COLON] = ACTIONS(4384),
    [anon_sym_SEMI] = ACTIONS(3313),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4382),
    [sym__PERLY_SEMICOLON] = ACTIONS(3313),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2156)] = {
    [sym_heredoc_content] = STATE(2156),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_and] = ACTIONS(2530),
    [anon_sym_or] = ACTIONS(2530),
    [anon_sym_xor] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_RBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2530),
  },
  [STATE(2157)] = {
    [sym_heredoc_content] = STATE(2157),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_RBRACE] = ACTIONS(3403),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3403),
  },
  [STATE(2158)] = {
    [sym_heredoc_content] = STATE(2158),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_COLON] = ACTIONS(2530),
    [anon_sym_DOLLAR] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_AT] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_RPAREN] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2532),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_PERCENT2] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2159)] = {
    [sym_heredoc_content] = STATE(2159),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_COLON] = ACTIONS(2690),
    [anon_sym_DOLLAR] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_AT] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_RPAREN] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2688),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_PERCENT2] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2160)] = {
    [sym_heredoc_content] = STATE(2160),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_COLON] = ACTIONS(3361),
    [anon_sym_DOLLAR] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_AT] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_RPAREN] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3363),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_PERCENT2] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2161)] = {
    [sym_heredoc_content] = STATE(2161),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_RBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2162)] = {
    [sym_heredoc_content] = STATE(2162),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_RBRACE] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_RPAREN] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_and] = ACTIONS(3425),
    [anon_sym_or] = ACTIONS(3425),
    [anon_sym_xor] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_RBRACK] = ACTIONS(3425),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2163)] = {
    [sym_heredoc_content] = STATE(2163),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_RPAREN] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_and] = ACTIONS(3409),
    [anon_sym_or] = ACTIONS(3409),
    [anon_sym_xor] = ACTIONS(3409),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3411),
  },
  [STATE(2164)] = {
    [sym_heredoc_content] = STATE(2164),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_and] = ACTIONS(3309),
    [anon_sym_or] = ACTIONS(3309),
    [anon_sym_xor] = ACTIONS(3309),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3304),
  },
  [STATE(2165)] = {
    [sym_heredoc_content] = STATE(2165),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_RPAREN] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_and] = ACTIONS(3401),
    [anon_sym_or] = ACTIONS(3401),
    [anon_sym_xor] = ACTIONS(3401),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3403),
  },
  [STATE(2166)] = {
    [sym_heredoc_content] = STATE(2166),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_DOLLAR] = ACTIONS(3333),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_AT] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_RPAREN] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3331),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_PERCENT2] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2167)] = {
    [sym_heredoc_content] = STATE(2167),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_DOLLAR] = ACTIONS(3337),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_AT] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_RPAREN] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3335),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_PERCENT2] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4386),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2168)] = {
    [sym_heredoc_content] = STATE(2168),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_RBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(2169)] = {
    [sym_attribute] = STATE(2923),
    [sym_attribute_name] = STATE(2483),
    [sym_heredoc_content] = STATE(2169),
    [aux_sym_attrlist_repeat1] = STATE(2182),
    [sym__identifier] = ACTIONS(4382),
    [anon_sym_COLON] = ACTIONS(4384),
    [anon_sym_SEMI] = ACTIONS(3319),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4382),
    [sym__PERLY_SEMICOLON] = ACTIONS(3319),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2170)] = {
    [sym_heredoc_content] = STATE(2170),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_RBRACE] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_RPAREN] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_and] = ACTIONS(3431),
    [anon_sym_or] = ACTIONS(3431),
    [anon_sym_xor] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_RBRACK] = ACTIONS(3431),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2171)] = {
    [sym_heredoc_content] = STATE(2171),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_COLON] = ACTIONS(3397),
    [anon_sym_DOLLAR] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_AT] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_RPAREN] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3399),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_PERCENT2] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2172)] = {
    [sym_heredoc_content] = STATE(2172),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_COLON] = ACTIONS(3349),
    [anon_sym_DOLLAR] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_AT] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_RPAREN] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3351),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_PERCENT2] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2173)] = {
    [sym_heredoc_content] = STATE(2173),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_COLON] = ACTIONS(3365),
    [anon_sym_DOLLAR] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_AT] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_RPAREN] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3367),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_PERCENT2] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2174)] = {
    [sym_heredoc_content] = STATE(2174),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_COLON] = ACTIONS(3385),
    [anon_sym_DOLLAR] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_AT] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_RPAREN] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3387),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_PERCENT2] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2175)] = {
    [sym_heredoc_content] = STATE(2175),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_COLON] = ACTIONS(3389),
    [anon_sym_DOLLAR] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_AT] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_RPAREN] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3391),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_PERCENT2] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2176)] = {
    [sym_heredoc_content] = STATE(2176),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_COLON] = ACTIONS(3405),
    [anon_sym_DOLLAR] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_AT] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_RPAREN] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3407),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_PERCENT2] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2177)] = {
    [sym_heredoc_content] = STATE(2177),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_RBRACE] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_and] = ACTIONS(3353),
    [anon_sym_or] = ACTIONS(3353),
    [anon_sym_xor] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3353),
  },
  [STATE(2178)] = {
    [sym_heredoc_content] = STATE(2178),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_RBRACE] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_and] = ACTIONS(3357),
    [anon_sym_or] = ACTIONS(3357),
    [anon_sym_xor] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3357),
  },
  [STATE(2179)] = {
    [sym_heredoc_content] = STATE(2179),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_RBRACE] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_and] = ACTIONS(3369),
    [anon_sym_or] = ACTIONS(3369),
    [anon_sym_xor] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3369),
  },
  [STATE(2180)] = {
    [sym_heredoc_content] = STATE(2180),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_COLON] = ACTIONS(3373),
    [anon_sym_DOLLAR] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_AT] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_RPAREN] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3375),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_PERCENT2] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2181)] = {
    [sym_heredoc_content] = STATE(2181),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_COLON] = ACTIONS(3381),
    [anon_sym_DOLLAR] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_AT] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_RPAREN] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3383),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_PERCENT2] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2182)] = {
    [sym_attribute] = STATE(2923),
    [sym_attribute_name] = STATE(2483),
    [sym_heredoc_content] = STATE(2182),
    [aux_sym_attrlist_repeat1] = STATE(2182),
    [sym__identifier] = ACTIONS(4388),
    [anon_sym_COLON] = ACTIONS(4391),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4388),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2183)] = {
    [sym_heredoc_content] = STATE(2183),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_and] = ACTIONS(2690),
    [anon_sym_or] = ACTIONS(2690),
    [anon_sym_xor] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_RBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2690),
  },
  [STATE(2184)] = {
    [sym_heredoc_content] = STATE(2184),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_COLON] = ACTIONS(3341),
    [anon_sym_DOLLAR] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_AT] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_RPAREN] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3343),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_PERCENT2] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2185)] = {
    [sym_heredoc_content] = STATE(2185),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_COLON] = ACTIONS(3353),
    [anon_sym_DOLLAR] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_AT] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_RPAREN] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3355),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_PERCENT2] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2186)] = {
    [sym_heredoc_content] = STATE(2186),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_COLON] = ACTIONS(3357),
    [anon_sym_DOLLAR] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_AT] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_RPAREN] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3359),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_PERCENT2] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2187)] = {
    [sym_heredoc_content] = STATE(2187),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_COLON] = ACTIONS(3369),
    [anon_sym_DOLLAR] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_AT] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_RPAREN] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3371),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_PERCENT2] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2188)] = {
    [sym_heredoc_content] = STATE(2188),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_COLON] = ACTIONS(3393),
    [anon_sym_DOLLAR] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_AT] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_RPAREN] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3395),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_PERCENT2] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2189)] = {
    [sym_heredoc_content] = STATE(2189),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_COLON] = ACTIONS(3377),
    [anon_sym_DOLLAR] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_AT] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_RPAREN] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3379),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_PERCENT2] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2190)] = {
    [sym_heredoc_content] = STATE(2190),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_RBRACE] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_and] = ACTIONS(3393),
    [anon_sym_or] = ACTIONS(3393),
    [anon_sym_xor] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3393),
  },
  [STATE(2191)] = {
    [sym_heredoc_content] = STATE(2191),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_RBRACE] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_and] = ACTIONS(3377),
    [anon_sym_or] = ACTIONS(3377),
    [anon_sym_xor] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3377),
  },
  [STATE(2192)] = {
    [sym_heredoc_content] = STATE(2192),
    [anon_sym_LBRACE] = ACTIONS(4394),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_DOLLAR] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_AT] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4396),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4398),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3417),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_PERCENT2] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2193)] = {
    [sym_heredoc_content] = STATE(2193),
    [anon_sym_RBRACE] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_RPAREN] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_and] = ACTIONS(3711),
    [anon_sym_or] = ACTIONS(3711),
    [anon_sym_xor] = ACTIONS(3711),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_RBRACK] = ACTIONS(3709),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4400),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2194)] = {
    [sym_heredoc_content] = STATE(2194),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_RPAREN] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_and] = ACTIONS(3431),
    [anon_sym_or] = ACTIONS(3431),
    [anon_sym_xor] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3431),
  },
  [STATE(2195)] = {
    [sym_heredoc_content] = STATE(2195),
    [anon_sym_RBRACE] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_RPAREN] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_and] = ACTIONS(3540),
    [anon_sym_or] = ACTIONS(3540),
    [anon_sym_xor] = ACTIONS(3540),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_RBRACK] = ACTIONS(3538),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4402),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2196)] = {
    [sym_heredoc_content] = STATE(2196),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2197)] = {
    [sym_heredoc_content] = STATE(2197),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2198)] = {
    [sym_attrlist] = STATE(3608),
    [sym_attribute] = STATE(2155),
    [sym_attribute_name] = STATE(2483),
    [sym_heredoc_content] = STATE(2198),
    [sym__identifier] = ACTIONS(4382),
    [anon_sym_SEMI] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4382),
    [sym__PERLY_SEMICOLON] = ACTIONS(3323),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2199)] = {
    [sym_heredoc_content] = STATE(2199),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_RPAREN] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_and] = ACTIONS(3425),
    [anon_sym_or] = ACTIONS(3425),
    [anon_sym_xor] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3425),
  },
  [STATE(2200)] = {
    [sym_heredoc_content] = STATE(2200),
    [anon_sym_RBRACE] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_RPAREN] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_and] = ACTIONS(3471),
    [anon_sym_or] = ACTIONS(3471),
    [anon_sym_xor] = ACTIONS(3471),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_RBRACK] = ACTIONS(3469),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4404),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2201)] = {
    [sym_heredoc_content] = STATE(2201),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_COLON] = ACTIONS(2686),
    [anon_sym_DOLLAR] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_AT] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2682),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_PERCENT2] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2202)] = {
    [sym_heredoc_content] = STATE(2202),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2203)] = {
    [sym_attrlist] = STATE(3372),
    [sym_attribute] = STATE(2082),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2203),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_RBRACE] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_RPAREN] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2204)] = {
    [sym_attrlist] = STATE(3373),
    [sym_attribute] = STATE(2082),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2204),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_RBRACE] = ACTIONS(3323),
    [anon_sym_EQ] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_COMMA] = ACTIONS(3323),
    [anon_sym_RPAREN] = ACTIONS(3323),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3323),
    [anon_sym_DASH_GT] = ACTIONS(3323),
    [anon_sym_STAR] = ACTIONS(3325),
    [anon_sym_PERCENT] = ACTIONS(3325),
    [anon_sym_GT] = ACTIONS(3325),
    [anon_sym_LT_LT] = ACTIONS(3325),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3323),
    [anon_sym_PLUS_EQ] = ACTIONS(3323),
    [anon_sym_DASH_EQ] = ACTIONS(3323),
    [anon_sym_DOT_EQ] = ACTIONS(3323),
    [anon_sym_STAR_EQ] = ACTIONS(3323),
    [anon_sym_SLASH_EQ] = ACTIONS(3323),
    [anon_sym_PERCENT_EQ] = ACTIONS(3323),
    [anon_sym_x_EQ] = ACTIONS(3323),
    [anon_sym_AMP_EQ] = ACTIONS(3323),
    [anon_sym_PIPE_EQ] = ACTIONS(3323),
    [anon_sym_CARET_EQ] = ACTIONS(3323),
    [anon_sym_LT_LT_EQ] = ACTIONS(3323),
    [anon_sym_GT_GT_EQ] = ACTIONS(3323),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3323),
    [anon_sym_DOT_DOT] = ACTIONS(3325),
    [anon_sym_STAR_STAR] = ACTIONS(3325),
    [anon_sym_PIPE_PIPE] = ACTIONS(3325),
    [anon_sym_SLASH_SLASH] = ACTIONS(3325),
    [anon_sym_CARET_CARET] = ACTIONS(3323),
    [anon_sym_AMP_AMP] = ACTIONS(3325),
    [anon_sym_PIPE] = ACTIONS(3325),
    [anon_sym_CARET] = ACTIONS(3325),
    [anon_sym_AMP] = ACTIONS(3325),
    [anon_sym_GT_GT2] = ACTIONS(3325),
    [anon_sym_PLUS] = ACTIONS(3325),
    [anon_sym_DASH] = ACTIONS(3325),
    [anon_sym_DOT] = ACTIONS(3325),
    [anon_sym_SLASH] = ACTIONS(3325),
    [anon_sym_x] = ACTIONS(3325),
    [anon_sym_EQ_TILDE] = ACTIONS(3323),
    [anon_sym_BANG_TILDE] = ACTIONS(3323),
    [anon_sym_EQ_EQ] = ACTIONS(3325),
    [anon_sym_BANG_EQ] = ACTIONS(3323),
    [anon_sym_eq] = ACTIONS(3325),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3323),
    [anon_sym_equ] = ACTIONS(3325),
    [anon_sym_eqr] = ACTIONS(3325),
    [anon_sym_ne] = ACTIONS(3325),
    [anon_sym_LT_EQ_GT] = ACTIONS(3323),
    [anon_sym_cmp] = ACTIONS(3325),
    [anon_sym_TILDE_TILDE] = ACTIONS(3323),
    [anon_sym_LT] = ACTIONS(3325),
    [anon_sym_LT_EQ] = ACTIONS(3325),
    [anon_sym_GT_EQ] = ACTIONS(3323),
    [anon_sym_lt] = ACTIONS(3325),
    [anon_sym_le] = ACTIONS(3325),
    [anon_sym_ge] = ACTIONS(3325),
    [anon_sym_gt] = ACTIONS(3325),
    [anon_sym_isa] = ACTIONS(3325),
    [anon_sym_PLUS_PLUS] = ACTIONS(3323),
    [anon_sym_DASH_DASH] = ACTIONS(3323),
    [anon_sym_QMARK] = ACTIONS(3323),
    [anon_sym_EQ_GT] = ACTIONS(3323),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2205)] = {
    [sym_heredoc_content] = STATE(2205),
    [anon_sym_RBRACE] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_RPAREN] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_and] = ACTIONS(3578),
    [anon_sym_or] = ACTIONS(3578),
    [anon_sym_xor] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_RBRACK] = ACTIONS(3578),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4406),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2206)] = {
    [sym_heredoc_content] = STATE(2206),
    [anon_sym_RBRACE] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_RPAREN] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_and] = ACTIONS(3639),
    [anon_sym_or] = ACTIONS(3639),
    [anon_sym_xor] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_RBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2207)] = {
    [sym_attrlist] = STATE(3607),
    [sym_attribute] = STATE(2155),
    [sym_attribute_name] = STATE(2483),
    [sym_heredoc_content] = STATE(2207),
    [sym__identifier] = ACTIONS(4382),
    [anon_sym_SEMI] = ACTIONS(3327),
    [anon_sym_EQ] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_COMMA] = ACTIONS(3327),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3327),
    [anon_sym_DASH_GT] = ACTIONS(3327),
    [anon_sym_STAR] = ACTIONS(3329),
    [anon_sym_PERCENT] = ACTIONS(3329),
    [anon_sym_GT] = ACTIONS(3329),
    [anon_sym_LT_LT] = ACTIONS(3329),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3327),
    [anon_sym_PLUS_EQ] = ACTIONS(3327),
    [anon_sym_DASH_EQ] = ACTIONS(3327),
    [anon_sym_DOT_EQ] = ACTIONS(3327),
    [anon_sym_STAR_EQ] = ACTIONS(3327),
    [anon_sym_SLASH_EQ] = ACTIONS(3327),
    [anon_sym_PERCENT_EQ] = ACTIONS(3327),
    [anon_sym_x_EQ] = ACTIONS(3327),
    [anon_sym_AMP_EQ] = ACTIONS(3327),
    [anon_sym_PIPE_EQ] = ACTIONS(3327),
    [anon_sym_CARET_EQ] = ACTIONS(3327),
    [anon_sym_LT_LT_EQ] = ACTIONS(3327),
    [anon_sym_GT_GT_EQ] = ACTIONS(3327),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3327),
    [anon_sym_DOT_DOT] = ACTIONS(3329),
    [anon_sym_STAR_STAR] = ACTIONS(3329),
    [anon_sym_PIPE_PIPE] = ACTIONS(3329),
    [anon_sym_SLASH_SLASH] = ACTIONS(3329),
    [anon_sym_CARET_CARET] = ACTIONS(3327),
    [anon_sym_AMP_AMP] = ACTIONS(3329),
    [anon_sym_PIPE] = ACTIONS(3329),
    [anon_sym_CARET] = ACTIONS(3329),
    [anon_sym_AMP] = ACTIONS(3329),
    [anon_sym_GT_GT2] = ACTIONS(3329),
    [anon_sym_PLUS] = ACTIONS(3329),
    [anon_sym_DASH] = ACTIONS(3329),
    [anon_sym_DOT] = ACTIONS(3329),
    [anon_sym_SLASH] = ACTIONS(3329),
    [anon_sym_x] = ACTIONS(3329),
    [anon_sym_EQ_TILDE] = ACTIONS(3327),
    [anon_sym_BANG_TILDE] = ACTIONS(3327),
    [anon_sym_EQ_EQ] = ACTIONS(3329),
    [anon_sym_BANG_EQ] = ACTIONS(3327),
    [anon_sym_eq] = ACTIONS(3329),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3327),
    [anon_sym_equ] = ACTIONS(3329),
    [anon_sym_eqr] = ACTIONS(3329),
    [anon_sym_ne] = ACTIONS(3329),
    [anon_sym_LT_EQ_GT] = ACTIONS(3327),
    [anon_sym_cmp] = ACTIONS(3329),
    [anon_sym_TILDE_TILDE] = ACTIONS(3327),
    [anon_sym_LT] = ACTIONS(3329),
    [anon_sym_LT_EQ] = ACTIONS(3329),
    [anon_sym_GT_EQ] = ACTIONS(3327),
    [anon_sym_lt] = ACTIONS(3329),
    [anon_sym_le] = ACTIONS(3329),
    [anon_sym_ge] = ACTIONS(3329),
    [anon_sym_gt] = ACTIONS(3329),
    [anon_sym_isa] = ACTIONS(3329),
    [anon_sym_PLUS_PLUS] = ACTIONS(3327),
    [anon_sym_DASH_DASH] = ACTIONS(3327),
    [anon_sym_QMARK] = ACTIONS(3327),
    [anon_sym_EQ_GT] = ACTIONS(3327),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4382),
    [sym__PERLY_SEMICOLON] = ACTIONS(3327),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2208)] = {
    [sym_heredoc_content] = STATE(2208),
    [anon_sym_RBRACE] = ACTIONS(3661),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_RPAREN] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_and] = ACTIONS(3661),
    [anon_sym_or] = ACTIONS(3661),
    [anon_sym_xor] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_RBRACK] = ACTIONS(3661),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2209)] = {
    [sym_heredoc_content] = STATE(2209),
    [aux_sym__term_rightward_repeat1] = STATE(4549),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(4412),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(4412),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2210)] = {
    [sym_heredoc_content] = STATE(2210),
    [anon_sym_RBRACE] = ACTIONS(3627),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_RPAREN] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_and] = ACTIONS(3627),
    [anon_sym_or] = ACTIONS(3627),
    [anon_sym_xor] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_RBRACK] = ACTIONS(3627),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2211)] = {
    [sym_heredoc_content] = STATE(2211),
    [anon_sym_RBRACE] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_RPAREN] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_and] = ACTIONS(3558),
    [anon_sym_or] = ACTIONS(3558),
    [anon_sym_xor] = ACTIONS(3558),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_RBRACK] = ACTIONS(3556),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4456),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2212)] = {
    [sym_heredoc_content] = STATE(2212),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_and] = ACTIONS(3431),
    [anon_sym_or] = ACTIONS(3431),
    [anon_sym_xor] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_RBRACK] = ACTIONS(3431),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3431),
  },
  [STATE(2213)] = {
    [sym_heredoc_content] = STATE(2213),
    [anon_sym_RBRACE] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_RPAREN] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_and] = ACTIONS(3645),
    [anon_sym_or] = ACTIONS(3645),
    [anon_sym_xor] = ACTIONS(3645),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_RBRACK] = ACTIONS(3643),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4458),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2214)] = {
    [sym_heredoc_content] = STATE(2214),
    [anon_sym_RBRACE] = ACTIONS(3689),
    [anon_sym_COLON] = ACTIONS(4460),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_RPAREN] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_RBRACK] = ACTIONS(3689),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2215)] = {
    [sym_heredoc_content] = STATE(2215),
    [anon_sym_RBRACE] = ACTIONS(3669),
    [anon_sym_COLON] = ACTIONS(4462),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_RPAREN] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_RBRACK] = ACTIONS(3669),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2216)] = {
    [sym_heredoc_content] = STATE(2216),
    [anon_sym_RBRACE] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_RPAREN] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_and] = ACTIONS(3465),
    [anon_sym_or] = ACTIONS(3465),
    [anon_sym_xor] = ACTIONS(3465),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_RBRACK] = ACTIONS(3463),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4464),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2217)] = {
    [sym_heredoc_content] = STATE(2217),
    [anon_sym_RBRACE] = ACTIONS(3675),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_RPAREN] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_and] = ACTIONS(3675),
    [anon_sym_or] = ACTIONS(3675),
    [anon_sym_xor] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_RBRACK] = ACTIONS(3675),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2218)] = {
    [sym_heredoc_content] = STATE(2218),
    [anon_sym_RBRACE] = ACTIONS(3679),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_RPAREN] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_and] = ACTIONS(3679),
    [anon_sym_or] = ACTIONS(3679),
    [anon_sym_xor] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_RBRACK] = ACTIONS(3679),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2219)] = {
    [sym_heredoc_content] = STATE(2219),
    [anon_sym_RBRACE] = ACTIONS(3447),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_RPAREN] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_and] = ACTIONS(3447),
    [anon_sym_or] = ACTIONS(3447),
    [anon_sym_xor] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_RBRACK] = ACTIONS(3447),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2220)] = {
    [sym_heredoc_content] = STATE(2220),
    [anon_sym_RBRACE] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_RPAREN] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_and] = ACTIONS(3699),
    [anon_sym_or] = ACTIONS(3699),
    [anon_sym_xor] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_RBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2221)] = {
    [sym_heredoc_content] = STATE(2221),
    [anon_sym_RBRACE] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_RPAREN] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(4466),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_and] = ACTIONS(3562),
    [anon_sym_or] = ACTIONS(3562),
    [anon_sym_xor] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_RBRACK] = ACTIONS(3562),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2222)] = {
    [sym_heredoc_content] = STATE(2222),
    [anon_sym_RBRACE] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_RPAREN] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_and] = ACTIONS(3586),
    [anon_sym_or] = ACTIONS(3586),
    [anon_sym_xor] = ACTIONS(3586),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_RBRACK] = ACTIONS(3584),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4468),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2223)] = {
    [sym_heredoc_content] = STATE(2223),
    [anon_sym_RBRACE] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_RPAREN] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_and] = ACTIONS(3475),
    [anon_sym_or] = ACTIONS(3475),
    [anon_sym_xor] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_RBRACK] = ACTIONS(3475),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2224)] = {
    [sym_heredoc_content] = STATE(2224),
    [anon_sym_RBRACE] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_RPAREN] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_and] = ACTIONS(3550),
    [anon_sym_or] = ACTIONS(3550),
    [anon_sym_xor] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_RBRACK] = ACTIONS(3550),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4470),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2225)] = {
    [sym_heredoc_content] = STATE(2225),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3435),
    [anon_sym_or] = ACTIONS(3435),
    [anon_sym_xor] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3435),
  },
  [STATE(2226)] = {
    [sym_heredoc_content] = STATE(2226),
    [anon_sym_LBRACE] = ACTIONS(4472),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_RBRACK] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2227)] = {
    [sym_heredoc_content] = STATE(2227),
    [anon_sym_RBRACE] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_RPAREN] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_and] = ACTIONS(3705),
    [anon_sym_or] = ACTIONS(3705),
    [anon_sym_xor] = ACTIONS(3705),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_RBRACK] = ACTIONS(3703),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4474),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2228)] = {
    [sym_heredoc_content] = STATE(2228),
    [anon_sym_RBRACE] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_RPAREN] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(4476),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_and] = ACTIONS(3479),
    [anon_sym_or] = ACTIONS(3479),
    [anon_sym_xor] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_RBRACK] = ACTIONS(3479),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2229)] = {
    [sym_heredoc_content] = STATE(2229),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3435),
    [anon_sym_or] = ACTIONS(3435),
    [anon_sym_xor] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3435),
  },
  [STATE(2230)] = {
    [sym_heredoc_content] = STATE(2230),
    [anon_sym_RBRACE] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_RPAREN] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_and] = ACTIONS(3441),
    [anon_sym_or] = ACTIONS(3441),
    [anon_sym_xor] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_RBRACK] = ACTIONS(3441),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4478),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2231)] = {
    [sym_heredoc_content] = STATE(2231),
    [anon_sym_RBRACE] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_RPAREN] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_and] = ACTIONS(3617),
    [anon_sym_or] = ACTIONS(3617),
    [anon_sym_xor] = ACTIONS(3617),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_RBRACK] = ACTIONS(3615),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4480),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2232)] = {
    [sym_heredoc_content] = STATE(2232),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(4482),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_RBRACK] = ACTIONS(3695),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2233)] = {
    [sym_heredoc_content] = STATE(2233),
    [anon_sym_RBRACE] = ACTIONS(3597),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_RPAREN] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_and] = ACTIONS(3597),
    [anon_sym_or] = ACTIONS(3597),
    [anon_sym_xor] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_RBRACK] = ACTIONS(3597),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2234)] = {
    [sym_heredoc_content] = STATE(2234),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_RBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2686),
  },
  [STATE(2235)] = {
    [sym_heredoc_content] = STATE(2235),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3435),
    [anon_sym_or] = ACTIONS(3435),
    [anon_sym_xor] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3435),
  },
  [STATE(2236)] = {
    [sym_heredoc_content] = STATE(2236),
    [anon_sym_RBRACE] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_RPAREN] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_and] = ACTIONS(3683),
    [anon_sym_or] = ACTIONS(3683),
    [anon_sym_xor] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_RBRACK] = ACTIONS(3683),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4484),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2237)] = {
    [sym_heredoc_content] = STATE(2237),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_COLON] = ACTIONS(3425),
    [anon_sym_DOLLAR] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_AT] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_RPAREN] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3427),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_PERCENT2] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2238)] = {
    [sym_heredoc_content] = STATE(2238),
    [anon_sym_RBRACE] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_RPAREN] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_and] = ACTIONS(3574),
    [anon_sym_or] = ACTIONS(3574),
    [anon_sym_xor] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(4486),
    [anon_sym_RBRACK] = ACTIONS(3574),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2239)] = {
    [sym_heredoc_content] = STATE(2239),
    [anon_sym_RBRACE] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_RPAREN] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_and] = ACTIONS(3451),
    [anon_sym_or] = ACTIONS(3451),
    [anon_sym_xor] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(4486),
    [anon_sym_RBRACK] = ACTIONS(3451),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2240)] = {
    [sym_heredoc_content] = STATE(2240),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_and] = ACTIONS(3425),
    [anon_sym_or] = ACTIONS(3425),
    [anon_sym_xor] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_RBRACK] = ACTIONS(3425),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3425),
  },
  [STATE(2241)] = {
    [sym_heredoc_content] = STATE(2241),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_COLON] = ACTIONS(3431),
    [anon_sym_DOLLAR] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_AT] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_RPAREN] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3433),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_PERCENT2] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2242)] = {
    [sym_heredoc_content] = STATE(2242),
    [anon_sym_RBRACE] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_RPAREN] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_and] = ACTIONS(3546),
    [anon_sym_or] = ACTIONS(3546),
    [anon_sym_xor] = ACTIONS(3546),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_RBRACK] = ACTIONS(3544),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4488),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2243)] = {
    [sym_heredoc_content] = STATE(2243),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_DOLLAR] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_AT] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3437),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_PERCENT2] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2244)] = {
    [sym_heredoc_content] = STATE(2244),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2686),
  },
  [STATE(2245)] = {
    [sym_heredoc_content] = STATE(2245),
    [anon_sym_RBRACE] = ACTIONS(3665),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_RPAREN] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_and] = ACTIONS(3665),
    [anon_sym_or] = ACTIONS(3665),
    [anon_sym_xor] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_RBRACK] = ACTIONS(3665),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2246)] = {
    [sym_heredoc_content] = STATE(2246),
    [anon_sym_RBRACE] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_RPAREN] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(4490),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_and] = ACTIONS(3633),
    [anon_sym_or] = ACTIONS(3633),
    [anon_sym_xor] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_RBRACK] = ACTIONS(3633),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2247)] = {
    [sym_heredoc_content] = STATE(2247),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4492),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2248)] = {
    [sym_heredoc_content] = STATE(2248),
    [anon_sym_RBRACE] = ACTIONS(3611),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_RPAREN] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_and] = ACTIONS(3611),
    [anon_sym_or] = ACTIONS(3611),
    [anon_sym_xor] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_RBRACK] = ACTIONS(3611),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2249)] = {
    [sym_heredoc_content] = STATE(2249),
    [aux_sym__term_rightward_repeat1] = STATE(2249),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4494),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_RBRACK] = ACTIONS(3590),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4494),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2250)] = {
    [sym_heredoc_content] = STATE(2250),
    [aux_sym__term_rightward_repeat1] = STATE(2254),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(4497),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(4497),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2251)] = {
    [sym_heredoc_content] = STATE(2251),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_DOLLAR] = ACTIONS(3411),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_AT] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_RPAREN] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3409),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_PERCENT2] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2252)] = {
    [sym_heredoc_content] = STATE(2252),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_DOLLAR] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_AT] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3309),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_PERCENT2] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2253)] = {
    [sym_heredoc_content] = STATE(2253),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_RBRACE] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_and] = ACTIONS(3425),
    [anon_sym_or] = ACTIONS(3425),
    [anon_sym_xor] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3425),
  },
  [STATE(2254)] = {
    [sym_heredoc_content] = STATE(2254),
    [aux_sym__term_rightward_repeat1] = STATE(2249),
    [anon_sym_RBRACE] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4500),
    [anon_sym_RPAREN] = ACTIONS(3568),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_RBRACK] = ACTIONS(3568),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4500),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2255)] = {
    [sym_heredoc_content] = STATE(2255),
    [anon_sym_RBRACE] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_RPAREN] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_and] = ACTIONS(3601),
    [anon_sym_or] = ACTIONS(3601),
    [anon_sym_xor] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_RBRACK] = ACTIONS(3601),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2256)] = {
    [sym_heredoc_content] = STATE(2256),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_RBRACE] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_RPAREN] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_and] = ACTIONS(3605),
    [anon_sym_or] = ACTIONS(3605),
    [anon_sym_xor] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_RBRACK] = ACTIONS(3605),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2257)] = {
    [sym_heredoc_content] = STATE(2257),
    [anon_sym_RBRACE] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_RPAREN] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_and] = ACTIONS(3657),
    [anon_sym_or] = ACTIONS(3657),
    [anon_sym_xor] = ACTIONS(3657),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_RBRACK] = ACTIONS(3655),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4502),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2258)] = {
    [sym_heredoc_content] = STATE(2258),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_DOLLAR] = ACTIONS(3403),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_AT] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_RPAREN] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3401),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_PERCENT2] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2259)] = {
    [sym_heredoc_content] = STATE(2259),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_and] = ACTIONS(2686),
    [anon_sym_or] = ACTIONS(2686),
    [anon_sym_xor] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2686),
  },
  [STATE(2260)] = {
    [sym_heredoc_content] = STATE(2260),
    [anon_sym_RBRACE] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_RPAREN] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_and] = ACTIONS(3649),
    [anon_sym_or] = ACTIONS(3649),
    [anon_sym_xor] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(4504),
    [anon_sym_RBRACK] = ACTIONS(3649),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2261)] = {
    [sym_heredoc_content] = STATE(2261),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_RBRACE] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_and] = ACTIONS(3431),
    [anon_sym_or] = ACTIONS(3431),
    [anon_sym_xor] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3431),
  },
  [STATE(2262)] = {
    [sym_heredoc_content] = STATE(2262),
    [anon_sym_RBRACE] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_RPAREN] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(4482),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_and] = ACTIONS(3621),
    [anon_sym_or] = ACTIONS(3621),
    [anon_sym_xor] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_RBRACK] = ACTIONS(3621),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2263)] = {
    [sym_heredoc_content] = STATE(2263),
    [anon_sym_RBRACE] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_RPAREN] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_and] = ACTIONS(3457),
    [anon_sym_or] = ACTIONS(3457),
    [anon_sym_xor] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(4504),
    [anon_sym_RBRACK] = ACTIONS(3457),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2264)] = {
    [sym_heredoc_content] = STATE(2264),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_RPAREN] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_and] = ACTIONS(3661),
    [anon_sym_or] = ACTIONS(3661),
    [anon_sym_xor] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3661),
  },
  [STATE(2265)] = {
    [sym_heredoc_content] = STATE(2265),
    [anon_sym_RBRACE] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_and] = ACTIONS(3639),
    [anon_sym_or] = ACTIONS(3639),
    [anon_sym_xor] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3639),
  },
  [STATE(2266)] = {
    [sym_heredoc_content] = STATE(2266),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_RPAREN] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_and] = ACTIONS(3639),
    [anon_sym_or] = ACTIONS(3639),
    [anon_sym_xor] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3639),
  },
  [STATE(2267)] = {
    [sym_heredoc_content] = STATE(2267),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_RBRACE] = ACTIONS(2690),
    [anon_sym_COLON] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_RPAREN] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2268)] = {
    [sym_heredoc_content] = STATE(2268),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4506),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3415),
  },
  [STATE(2269)] = {
    [sym_heredoc_content] = STATE(2269),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_RPAREN] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_and] = ACTIONS(3645),
    [anon_sym_or] = ACTIONS(3645),
    [anon_sym_xor] = ACTIONS(3645),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4508),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3643),
  },
  [STATE(2270)] = {
    [sym_heredoc_content] = STATE(2270),
    [aux_sym__term_rightward_repeat1] = STATE(2299),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(4514),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(4514),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3485),
  },
  [STATE(2271)] = {
    [sym_heredoc_content] = STATE(2271),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_RPAREN] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_and] = ACTIONS(3465),
    [anon_sym_or] = ACTIONS(3465),
    [anon_sym_xor] = ACTIONS(3465),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4559),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3463),
  },
  [STATE(2272)] = {
    [sym_heredoc_content] = STATE(2272),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2273)] = {
    [sym_heredoc_content] = STATE(2273),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_RPAREN] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_and] = ACTIONS(3475),
    [anon_sym_or] = ACTIONS(3475),
    [anon_sym_xor] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3475),
  },
  [STATE(2274)] = {
    [sym_heredoc_content] = STATE(2274),
    [anon_sym_COLON] = ACTIONS(4561),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_RPAREN] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3689),
  },
  [STATE(2275)] = {
    [sym_heredoc_content] = STATE(2275),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_RPAREN] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(4563),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_and] = ACTIONS(3479),
    [anon_sym_or] = ACTIONS(3479),
    [anon_sym_xor] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3479),
  },
  [STATE(2276)] = {
    [sym_heredoc_content] = STATE(2276),
    [anon_sym_RBRACE] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_RPAREN] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_and] = ACTIONS(3949),
    [anon_sym_or] = ACTIONS(3949),
    [anon_sym_xor] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_RBRACK] = ACTIONS(3949),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2277)] = {
    [sym_heredoc_content] = STATE(2277),
    [aux_sym__term_rightward_repeat1] = STATE(2281),
    [anon_sym_RBRACE] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4565),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4565),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3568),
  },
  [STATE(2278)] = {
    [sym_heredoc_content] = STATE(2278),
    [anon_sym_RBRACE] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_and] = ACTIONS(3645),
    [anon_sym_or] = ACTIONS(3645),
    [anon_sym_xor] = ACTIONS(3645),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4567),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3643),
  },
  [STATE(2279)] = {
    [sym_heredoc_content] = STATE(2279),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_RPAREN] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_and] = ACTIONS(3558),
    [anon_sym_or] = ACTIONS(3558),
    [anon_sym_xor] = ACTIONS(3558),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4569),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3556),
  },
  [STATE(2280)] = {
    [sym_heredoc_content] = STATE(2280),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_RPAREN] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_and] = ACTIONS(3574),
    [anon_sym_or] = ACTIONS(3574),
    [anon_sym_xor] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(4571),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3574),
  },
  [STATE(2281)] = {
    [sym_heredoc_content] = STATE(2281),
    [aux_sym__term_rightward_repeat1] = STATE(2281),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4573),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4573),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3590),
  },
  [STATE(2282)] = {
    [sym_heredoc_content] = STATE(2282),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_and] = ACTIONS(3601),
    [anon_sym_or] = ACTIONS(3601),
    [anon_sym_xor] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_RBRACK] = ACTIONS(3601),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3601),
  },
  [STATE(2283)] = {
    [sym_heredoc_content] = STATE(2283),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_and] = ACTIONS(3605),
    [anon_sym_or] = ACTIONS(3605),
    [anon_sym_xor] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_RBRACK] = ACTIONS(3605),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3605),
  },
  [STATE(2284)] = {
    [sym_heredoc_content] = STATE(2284),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_and] = ACTIONS(3657),
    [anon_sym_or] = ACTIONS(3657),
    [anon_sym_xor] = ACTIONS(3657),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_RBRACK] = ACTIONS(3655),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4576),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3655),
  },
  [STATE(2285)] = {
    [sym_heredoc_content] = STATE(2285),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_and] = ACTIONS(3649),
    [anon_sym_or] = ACTIONS(3649),
    [anon_sym_xor] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(4578),
    [anon_sym_RBRACK] = ACTIONS(3649),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3649),
  },
  [STATE(2286)] = {
    [sym_heredoc_content] = STATE(2286),
    [anon_sym_COLON] = ACTIONS(4580),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_RPAREN] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3669),
  },
  [STATE(2287)] = {
    [sym_heredoc_content] = STATE(2287),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_RPAREN] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_and] = ACTIONS(3627),
    [anon_sym_or] = ACTIONS(3627),
    [anon_sym_xor] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3627),
  },
  [STATE(2288)] = {
    [sym_heredoc_content] = STATE(2288),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_RPAREN] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_and] = ACTIONS(3699),
    [anon_sym_or] = ACTIONS(3699),
    [anon_sym_xor] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3699),
  },
  [STATE(2289)] = {
    [sym_heredoc_content] = STATE(2289),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_RPAREN] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_and] = ACTIONS(3705),
    [anon_sym_or] = ACTIONS(3705),
    [anon_sym_xor] = ACTIONS(3705),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4582),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3703),
  },
  [STATE(2290)] = {
    [sym_heredoc_content] = STATE(2290),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_RPAREN] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_and] = ACTIONS(3711),
    [anon_sym_or] = ACTIONS(3711),
    [anon_sym_xor] = ACTIONS(3711),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4584),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3709),
  },
  [STATE(2291)] = {
    [sym_heredoc_content] = STATE(2291),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_and] = ACTIONS(3665),
    [anon_sym_or] = ACTIONS(3665),
    [anon_sym_xor] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_RBRACK] = ACTIONS(3665),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3665),
  },
  [STATE(2292)] = {
    [sym_heredoc_content] = STATE(2292),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_and] = ACTIONS(3597),
    [anon_sym_or] = ACTIONS(3597),
    [anon_sym_xor] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_RBRACK] = ACTIONS(3597),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3597),
  },
  [STATE(2293)] = {
    [sym_heredoc_content] = STATE(2293),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_and] = ACTIONS(3611),
    [anon_sym_or] = ACTIONS(3611),
    [anon_sym_xor] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_RBRACK] = ACTIONS(3611),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3611),
  },
  [STATE(2294)] = {
    [sym_heredoc_content] = STATE(2294),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_and] = ACTIONS(3679),
    [anon_sym_or] = ACTIONS(3679),
    [anon_sym_xor] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_RBRACK] = ACTIONS(3679),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3679),
  },
  [STATE(2295)] = {
    [sym_heredoc_content] = STATE(2295),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_and] = ACTIONS(3447),
    [anon_sym_or] = ACTIONS(3447),
    [anon_sym_xor] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_RBRACK] = ACTIONS(3447),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3447),
  },
  [STATE(2296)] = {
    [sym_heredoc_content] = STATE(2296),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_RPAREN] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_and] = ACTIONS(3457),
    [anon_sym_or] = ACTIONS(3457),
    [anon_sym_xor] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(4586),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3457),
  },
  [STATE(2297)] = {
    [sym_heredoc_content] = STATE(2297),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2298)] = {
    [sym_heredoc_content] = STATE(2298),
    [anon_sym_RBRACE] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_RPAREN] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_and] = ACTIONS(3957),
    [anon_sym_or] = ACTIONS(3957),
    [anon_sym_xor] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_RBRACK] = ACTIONS(3957),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2299)] = {
    [sym_heredoc_content] = STATE(2299),
    [aux_sym__term_rightward_repeat1] = STATE(2304),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4588),
    [anon_sym_RPAREN] = ACTIONS(3568),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4588),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3568),
  },
  [STATE(2300)] = {
    [sym_heredoc_content] = STATE(2300),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_and] = ACTIONS(3451),
    [anon_sym_or] = ACTIONS(3451),
    [anon_sym_xor] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(4590),
    [anon_sym_RBRACK] = ACTIONS(3451),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3451),
  },
  [STATE(2301)] = {
    [sym_heredoc_content] = STATE(2301),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_RPAREN] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_and] = ACTIONS(3617),
    [anon_sym_or] = ACTIONS(3617),
    [anon_sym_xor] = ACTIONS(3617),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4592),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3615),
  },
  [STATE(2302)] = {
    [sym_heredoc_content] = STATE(2302),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_RPAREN] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_and] = ACTIONS(3683),
    [anon_sym_or] = ACTIONS(3683),
    [anon_sym_xor] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4594),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3683),
  },
  [STATE(2303)] = {
    [sym_heredoc_content] = STATE(2303),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_RPAREN] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_RBRACK] = ACTIONS(4045),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2304)] = {
    [sym_heredoc_content] = STATE(2304),
    [aux_sym__term_rightward_repeat1] = STATE(2304),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4596),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4596),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3590),
  },
  [STATE(2305)] = {
    [sym_heredoc_content] = STATE(2305),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(4599),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3695),
  },
  [STATE(2306)] = {
    [sym_heredoc_content] = STATE(2306),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_and] = ACTIONS(3639),
    [anon_sym_or] = ACTIONS(3639),
    [anon_sym_xor] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_RBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3639),
  },
  [STATE(2307)] = {
    [sym_heredoc_content] = STATE(2307),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_and] = ACTIONS(3645),
    [anon_sym_or] = ACTIONS(3645),
    [anon_sym_xor] = ACTIONS(3645),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_RBRACK] = ACTIONS(3643),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4601),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3643),
  },
  [STATE(2308)] = {
    [sym_heredoc_content] = STATE(2308),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_and] = ACTIONS(3465),
    [anon_sym_or] = ACTIONS(3465),
    [anon_sym_xor] = ACTIONS(3465),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_RBRACK] = ACTIONS(3463),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4603),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3463),
  },
  [STATE(2309)] = {
    [sym_heredoc_content] = STATE(2309),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3763),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2310)] = {
    [sym_heredoc_content] = STATE(2310),
    [anon_sym_RBRACE] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_RPAREN] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_and] = ACTIONS(4047),
    [anon_sym_or] = ACTIONS(4047),
    [anon_sym_xor] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_RBRACK] = ACTIONS(4047),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2311)] = {
    [sym_heredoc_content] = STATE(2311),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_and] = ACTIONS(3475),
    [anon_sym_or] = ACTIONS(3475),
    [anon_sym_xor] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_RBRACK] = ACTIONS(3475),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3475),
  },
  [STATE(2312)] = {
    [sym_heredoc_content] = STATE(2312),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(4605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_and] = ACTIONS(3479),
    [anon_sym_or] = ACTIONS(3479),
    [anon_sym_xor] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_RBRACK] = ACTIONS(3479),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3479),
  },
  [STATE(2313)] = {
    [sym_heredoc_content] = STATE(2313),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3763),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4607),
    [anon_sym_cmp] = ACTIONS(4607),
    [anon_sym_TILDE_TILDE] = ACTIONS(4607),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2314)] = {
    [sym_heredoc_content] = STATE(2314),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_and] = ACTIONS(3558),
    [anon_sym_or] = ACTIONS(3558),
    [anon_sym_xor] = ACTIONS(3558),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_RBRACK] = ACTIONS(3556),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4609),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3556),
  },
  [STATE(2315)] = {
    [sym_heredoc_content] = STATE(2315),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_and] = ACTIONS(3574),
    [anon_sym_or] = ACTIONS(3574),
    [anon_sym_xor] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(4590),
    [anon_sym_RBRACK] = ACTIONS(3574),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3574),
  },
  [STATE(2316)] = {
    [sym_heredoc_content] = STATE(2316),
    [anon_sym_RBRACE] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_RPAREN] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_and] = ACTIONS(4208),
    [anon_sym_or] = ACTIONS(4208),
    [anon_sym_xor] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_RBRACK] = ACTIONS(4208),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2317)] = {
    [sym_heredoc_content] = STATE(2317),
    [aux_sym__term_rightward_repeat1] = STATE(2322),
    [anon_sym_COLON] = ACTIONS(3568),
    [anon_sym_DOLLAR] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_AT] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4611),
    [anon_sym_RPAREN] = ACTIONS(3568),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3570),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_PERCENT2] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2318)] = {
    [sym_heredoc_content] = STATE(2318),
    [anon_sym_RBRACE] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_RPAREN] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_and] = ACTIONS(3865),
    [anon_sym_or] = ACTIONS(3865),
    [anon_sym_xor] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_RBRACK] = ACTIONS(3865),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2319)] = {
    [sym_heredoc_content] = STATE(2319),
    [anon_sym_RBRACE] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_and] = ACTIONS(3465),
    [anon_sym_or] = ACTIONS(3465),
    [anon_sym_xor] = ACTIONS(3465),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4613),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3463),
  },
  [STATE(2320)] = {
    [sym_heredoc_content] = STATE(2320),
    [aux_sym__term_rightward_repeat1] = STATE(2418),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(4615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_and] = ACTIONS(3568),
    [anon_sym_or] = ACTIONS(3568),
    [anon_sym_xor] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_RBRACK] = ACTIONS(3568),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(4615),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3568),
  },
  [STATE(2321)] = {
    [sym_heredoc_content] = STATE(2321),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_RPAREN] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(4617),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_and] = ACTIONS(3562),
    [anon_sym_or] = ACTIONS(3562),
    [anon_sym_xor] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3562),
  },
  [STATE(2322)] = {
    [sym_heredoc_content] = STATE(2322),
    [aux_sym__term_rightward_repeat1] = STATE(2322),
    [anon_sym_COLON] = ACTIONS(3590),
    [anon_sym_DOLLAR] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_AT] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4619),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3592),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_PERCENT2] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4619),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2323)] = {
    [sym_heredoc_content] = STATE(2323),
    [anon_sym_RBRACE] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_and] = ACTIONS(3457),
    [anon_sym_or] = ACTIONS(3457),
    [anon_sym_xor] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(4622),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3457),
  },
  [STATE(2324)] = {
    [sym_heredoc_content] = STATE(2324),
    [aux_sym__term_rightward_repeat1] = STATE(2317),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_DOLLAR] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(4628),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(4611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2325)] = {
    [sym_heredoc_content] = STATE(2325),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_and] = ACTIONS(3627),
    [anon_sym_or] = ACTIONS(3627),
    [anon_sym_xor] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_RBRACK] = ACTIONS(3627),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3627),
  },
  [STATE(2326)] = {
    [sym_heredoc_content] = STATE(2326),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_SEMI] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2327)] = {
    [sym_heredoc_content] = STATE(2327),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3735),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4673),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2328)] = {
    [sym_heredoc_content] = STATE(2328),
    [anon_sym_RBRACE] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_RPAREN] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_and] = ACTIONS(3829),
    [anon_sym_or] = ACTIONS(3829),
    [anon_sym_xor] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_RBRACK] = ACTIONS(3829),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2329)] = {
    [sym_heredoc_content] = STATE(2329),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_RBRACE] = ACTIONS(2530),
    [anon_sym_COLON] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_RPAREN] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2330)] = {
    [sym_heredoc_content] = STATE(2330),
    [anon_sym_RBRACE] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_RPAREN] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_and] = ACTIONS(3771),
    [anon_sym_or] = ACTIONS(3771),
    [anon_sym_xor] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_RBRACK] = ACTIONS(3771),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2331)] = {
    [sym_heredoc_content] = STATE(2331),
    [anon_sym_RBRACE] = ACTIONS(3597),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_and] = ACTIONS(3597),
    [anon_sym_or] = ACTIONS(3597),
    [anon_sym_xor] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3597),
  },
  [STATE(2332)] = {
    [sym_heredoc_content] = STATE(2332),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_and] = ACTIONS(3699),
    [anon_sym_or] = ACTIONS(3699),
    [anon_sym_xor] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_RBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3699),
  },
  [STATE(2333)] = {
    [sym_heredoc_content] = STATE(2333),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_and] = ACTIONS(3705),
    [anon_sym_or] = ACTIONS(3705),
    [anon_sym_xor] = ACTIONS(3705),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_RBRACK] = ACTIONS(3703),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4675),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3703),
  },
  [STATE(2334)] = {
    [sym_heredoc_content] = STATE(2334),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_SEMI] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2335)] = {
    [sym_heredoc_content] = STATE(2335),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_RBRACE] = ACTIONS(2694),
    [anon_sym_COLON] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_RPAREN] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2336)] = {
    [sym_heredoc_content] = STATE(2336),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_and] = ACTIONS(3711),
    [anon_sym_or] = ACTIONS(3711),
    [anon_sym_xor] = ACTIONS(3711),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_RBRACK] = ACTIONS(3709),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4677),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3709),
  },
  [STATE(2337)] = {
    [sym_heredoc_content] = STATE(2337),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_and] = ACTIONS(3617),
    [anon_sym_or] = ACTIONS(3617),
    [anon_sym_xor] = ACTIONS(3617),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_RBRACK] = ACTIONS(3615),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4679),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3615),
  },
  [STATE(2338)] = {
    [sym_heredoc_content] = STATE(2338),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_and] = ACTIONS(3683),
    [anon_sym_or] = ACTIONS(3683),
    [anon_sym_xor] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_RBRACK] = ACTIONS(3683),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4681),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3683),
  },
  [STATE(2339)] = {
    [sym_heredoc_content] = STATE(2339),
    [anon_sym_RBRACE] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_RPAREN] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_and] = ACTIONS(4095),
    [anon_sym_or] = ACTIONS(4095),
    [anon_sym_xor] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_RBRACK] = ACTIONS(4095),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2340)] = {
    [sym_heredoc_content] = STATE(2340),
    [anon_sym_RBRACE] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_RPAREN] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_and] = ACTIONS(3783),
    [anon_sym_or] = ACTIONS(3783),
    [anon_sym_xor] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_RBRACK] = ACTIONS(3783),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2341)] = {
    [sym_heredoc_content] = STATE(2341),
    [anon_sym_RBRACE] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_RPAREN] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_and] = ACTIONS(3787),
    [anon_sym_or] = ACTIONS(3787),
    [anon_sym_xor] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_RBRACK] = ACTIONS(3787),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2342)] = {
    [sym_heredoc_content] = STATE(2342),
    [anon_sym_RBRACE] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_and] = ACTIONS(3601),
    [anon_sym_or] = ACTIONS(3601),
    [anon_sym_xor] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3601),
  },
  [STATE(2343)] = {
    [sym_heredoc_content] = STATE(2343),
    [anon_sym_RBRACE] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_RPAREN] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_and] = ACTIONS(4184),
    [anon_sym_or] = ACTIONS(4184),
    [anon_sym_xor] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_RBRACK] = ACTIONS(4184),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2344)] = {
    [sym_heredoc_content] = STATE(2344),
    [aux_sym__term_rightward_repeat1] = STATE(2317),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_DOLLAR] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(4628),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(4628),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2345)] = {
    [sym_heredoc_content] = STATE(2345),
    [anon_sym_RBRACE] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_RPAREN] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_and] = ACTIONS(3727),
    [anon_sym_or] = ACTIONS(3727),
    [anon_sym_xor] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_RBRACK] = ACTIONS(3727),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2346)] = {
    [sym_heredoc_content] = STATE(2346),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(4683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_and] = ACTIONS(3562),
    [anon_sym_or] = ACTIONS(3562),
    [anon_sym_xor] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_RBRACK] = ACTIONS(3562),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3562),
  },
  [STATE(2347)] = {
    [sym_heredoc_content] = STATE(2347),
    [anon_sym_RBRACE] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_RPAREN] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_and] = ACTIONS(3743),
    [anon_sym_or] = ACTIONS(3743),
    [anon_sym_xor] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_RBRACK] = ACTIONS(3743),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2348)] = {
    [sym_heredoc_content] = STATE(2348),
    [anon_sym_RBRACE] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_RPAREN] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_and] = ACTIONS(3791),
    [anon_sym_or] = ACTIONS(3791),
    [anon_sym_xor] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_RBRACK] = ACTIONS(3791),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2349)] = {
    [sym_heredoc_content] = STATE(2349),
    [anon_sym_RBRACE] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_RPAREN] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_and] = ACTIONS(3795),
    [anon_sym_or] = ACTIONS(3795),
    [anon_sym_xor] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_RBRACK] = ACTIONS(3795),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2350)] = {
    [sym_heredoc_content] = STATE(2350),
    [anon_sym_RBRACE] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_RPAREN] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_and] = ACTIONS(3811),
    [anon_sym_or] = ACTIONS(3811),
    [anon_sym_xor] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_RBRACK] = ACTIONS(3811),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2351)] = {
    [sym_heredoc_content] = STATE(2351),
    [anon_sym_RBRACE] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_RPAREN] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_and] = ACTIONS(3861),
    [anon_sym_or] = ACTIONS(3861),
    [anon_sym_xor] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_RBRACK] = ACTIONS(3861),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2352)] = {
    [sym_heredoc_content] = STATE(2352),
    [anon_sym_RBRACE] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3977),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2353)] = {
    [sym_heredoc_content] = STATE(2353),
    [anon_sym_RBRACE] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_RPAREN] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_and] = ACTIONS(3747),
    [anon_sym_or] = ACTIONS(3747),
    [anon_sym_xor] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_RBRACK] = ACTIONS(3747),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2354)] = {
    [sym_heredoc_content] = STATE(2354),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_RPAREN] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_and] = ACTIONS(3540),
    [anon_sym_or] = ACTIONS(3540),
    [anon_sym_xor] = ACTIONS(3540),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4685),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3538),
  },
  [STATE(2355)] = {
    [sym_heredoc_content] = STATE(2355),
    [aux_sym__term_rightward_repeat1] = STATE(2317),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_DOLLAR] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(4611),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(4611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2356)] = {
    [sym_heredoc_content] = STATE(2356),
    [anon_sym_RBRACE] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_RPAREN] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_and] = ACTIONS(3873),
    [anon_sym_or] = ACTIONS(3873),
    [anon_sym_xor] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_RBRACK] = ACTIONS(3873),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2357)] = {
    [sym_heredoc_content] = STATE(2357),
    [anon_sym_RBRACE] = ACTIONS(3611),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_and] = ACTIONS(3611),
    [anon_sym_or] = ACTIONS(3611),
    [anon_sym_xor] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3611),
  },
  [STATE(2358)] = {
    [sym_heredoc_content] = STATE(2358),
    [aux_sym__term_rightward_repeat1] = STATE(4555),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(4691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(4691),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3485),
  },
  [STATE(2359)] = {
    [sym_heredoc_content] = STATE(2359),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_RPAREN] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_and] = ACTIONS(3546),
    [anon_sym_or] = ACTIONS(3546),
    [anon_sym_xor] = ACTIONS(3546),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3544),
  },
  [STATE(2360)] = {
    [sym_heredoc_content] = STATE(2360),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_RPAREN] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_and] = ACTIONS(3578),
    [anon_sym_or] = ACTIONS(3578),
    [anon_sym_xor] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4737),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3578),
  },
  [STATE(2361)] = {
    [sym_heredoc_content] = STATE(2361),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_and] = ACTIONS(3661),
    [anon_sym_or] = ACTIONS(3661),
    [anon_sym_xor] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_RBRACK] = ACTIONS(3661),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3661),
  },
  [STATE(2362)] = {
    [sym_heredoc_content] = STATE(2362),
    [anon_sym_RBRACE] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_RPAREN] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_and] = ACTIONS(3731),
    [anon_sym_or] = ACTIONS(3731),
    [anon_sym_xor] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_RBRACK] = ACTIONS(3731),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2363)] = {
    [sym_heredoc_content] = STATE(2363),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_RPAREN] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_and] = ACTIONS(3441),
    [anon_sym_or] = ACTIONS(3441),
    [anon_sym_xor] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3441),
  },
  [STATE(2364)] = {
    [sym_heredoc_content] = STATE(2364),
    [anon_sym_RBRACE] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_RPAREN] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_and] = ACTIONS(3981),
    [anon_sym_or] = ACTIONS(3981),
    [anon_sym_xor] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_RBRACK] = ACTIONS(3981),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2365)] = {
    [sym_heredoc_content] = STATE(2365),
    [anon_sym_RBRACE] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3985),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2366)] = {
    [sym_heredoc_content] = STATE(2366),
    [anon_sym_RBRACE] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_RPAREN] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_and] = ACTIONS(4001),
    [anon_sym_or] = ACTIONS(4001),
    [anon_sym_xor] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_RBRACK] = ACTIONS(4001),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2367)] = {
    [sym_heredoc_content] = STATE(2367),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3435),
  },
  [STATE(2368)] = {
    [sym_heredoc_content] = STATE(2368),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_RBRACE] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_and] = ACTIONS(3605),
    [anon_sym_or] = ACTIONS(3605),
    [anon_sym_xor] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3605),
  },
  [STATE(2369)] = {
    [sym_heredoc_content] = STATE(2369),
    [anon_sym_RBRACE] = ACTIONS(3447),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_and] = ACTIONS(3447),
    [anon_sym_or] = ACTIONS(3447),
    [anon_sym_xor] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3447),
  },
  [STATE(2370)] = {
    [sym_heredoc_content] = STATE(2370),
    [anon_sym_RBRACE] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_RPAREN] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_and] = ACTIONS(4009),
    [anon_sym_or] = ACTIONS(4009),
    [anon_sym_xor] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_RBRACK] = ACTIONS(4009),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2371)] = {
    [sym_heredoc_content] = STATE(2371),
    [anon_sym_RBRACE] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_RPAREN] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_and] = ACTIONS(4013),
    [anon_sym_or] = ACTIONS(4013),
    [anon_sym_xor] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_RBRACK] = ACTIONS(4013),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2372)] = {
    [sym_heredoc_content] = STATE(2372),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_and] = ACTIONS(3540),
    [anon_sym_or] = ACTIONS(3540),
    [anon_sym_xor] = ACTIONS(3540),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_RBRACK] = ACTIONS(3538),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4741),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3538),
  },
  [STATE(2373)] = {
    [sym_heredoc_content] = STATE(2373),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_and] = ACTIONS(3546),
    [anon_sym_or] = ACTIONS(3546),
    [anon_sym_xor] = ACTIONS(3546),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_RBRACK] = ACTIONS(3544),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4743),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3544),
  },
  [STATE(2374)] = {
    [sym_heredoc_content] = STATE(2374),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_and] = ACTIONS(3578),
    [anon_sym_or] = ACTIONS(3578),
    [anon_sym_xor] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_RBRACK] = ACTIONS(3578),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4745),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3578),
  },
  [STATE(2375)] = {
    [sym_heredoc_content] = STATE(2375),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_and] = ACTIONS(3441),
    [anon_sym_or] = ACTIONS(3441),
    [anon_sym_xor] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_RBRACK] = ACTIONS(3441),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4747),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3441),
  },
  [STATE(2376)] = {
    [sym_heredoc_content] = STATE(2376),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2377)] = {
    [sym_heredoc_content] = STATE(2377),
    [anon_sym_RBRACE] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_RPAREN] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_and] = ACTIONS(4025),
    [anon_sym_or] = ACTIONS(4025),
    [anon_sym_xor] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_RBRACK] = ACTIONS(4025),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2378)] = {
    [sym_heredoc_content] = STATE(2378),
    [anon_sym_RBRACE] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_RPAREN] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_and] = ACTIONS(4029),
    [anon_sym_or] = ACTIONS(4029),
    [anon_sym_xor] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_RBRACK] = ACTIONS(4029),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2379)] = {
    [sym_heredoc_content] = STATE(2379),
    [anon_sym_RBRACE] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_RPAREN] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_and] = ACTIONS(3837),
    [anon_sym_or] = ACTIONS(3837),
    [anon_sym_xor] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_RBRACK] = ACTIONS(3837),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2380)] = {
    [sym_heredoc_content] = STATE(2380),
    [anon_sym_RBRACE] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_RPAREN] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_and] = ACTIONS(3993),
    [anon_sym_or] = ACTIONS(3993),
    [anon_sym_xor] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_RBRACK] = ACTIONS(3993),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2381)] = {
    [sym_heredoc_content] = STATE(2381),
    [anon_sym_RBRACE] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_and] = ACTIONS(3657),
    [anon_sym_or] = ACTIONS(3657),
    [anon_sym_xor] = ACTIONS(3657),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4749),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3655),
  },
  [STATE(2382)] = {
    [sym_heredoc_content] = STATE(2382),
    [anon_sym_RBRACE] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(4751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_and] = ACTIONS(3633),
    [anon_sym_or] = ACTIONS(3633),
    [anon_sym_xor] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3633),
  },
  [STATE(2383)] = {
    [sym_heredoc_content] = STATE(2383),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2384)] = {
    [sym_heredoc_content] = STATE(2384),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_RPAREN] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_and] = ACTIONS(3586),
    [anon_sym_or] = ACTIONS(3586),
    [anon_sym_xor] = ACTIONS(3586),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4753),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3584),
  },
  [STATE(2385)] = {
    [sym_heredoc_content] = STATE(2385),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_RPAREN] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_and] = ACTIONS(3550),
    [anon_sym_or] = ACTIONS(3550),
    [anon_sym_xor] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4755),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3550),
  },
  [STATE(2386)] = {
    [sym_heredoc_content] = STATE(2386),
    [anon_sym_RBRACE] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_and] = ACTIONS(3649),
    [anon_sym_or] = ACTIONS(3649),
    [anon_sym_xor] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(4622),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3649),
  },
  [STATE(2387)] = {
    [sym_heredoc_content] = STATE(2387),
    [anon_sym_RBRACE] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_RPAREN] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_and] = ACTIONS(4071),
    [anon_sym_or] = ACTIONS(4071),
    [anon_sym_xor] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_RBRACK] = ACTIONS(4071),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2388)] = {
    [sym_heredoc_content] = STATE(2388),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_RPAREN] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_and] = ACTIONS(3471),
    [anon_sym_or] = ACTIONS(3471),
    [anon_sym_xor] = ACTIONS(3471),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4757),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3469),
  },
  [STATE(2389)] = {
    [sym_heredoc_content] = STATE(2389),
    [anon_sym_COLON] = ACTIONS(3633),
    [anon_sym_DOLLAR] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_AT] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_RPAREN] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(4759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3635),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_PERCENT2] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2390)] = {
    [sym_heredoc_content] = STATE(2390),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3435),
  },
  [STATE(2391)] = {
    [sym_heredoc_content] = STATE(2391),
    [anon_sym_COLON] = ACTIONS(3469),
    [anon_sym_DOLLAR] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_AT] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_RPAREN] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3471),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_PERCENT2] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4761),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2392)] = {
    [sym_heredoc_content] = STATE(2392),
    [aux_sym__term_rightward_repeat1] = STATE(4584),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(4763),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(4763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3485),
  },
  [STATE(2393)] = {
    [sym_heredoc_content] = STATE(2393),
    [anon_sym_RBRACE] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_and] = ACTIONS(3475),
    [anon_sym_or] = ACTIONS(3475),
    [anon_sym_xor] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3475),
  },
  [STATE(2394)] = {
    [sym_heredoc_content] = STATE(2394),
    [anon_sym_RBRACE] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(4765),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_and] = ACTIONS(3479),
    [anon_sym_or] = ACTIONS(3479),
    [anon_sym_xor] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3479),
  },
  [STATE(2395)] = {
    [sym_heredoc_content] = STATE(2395),
    [anon_sym_RBRACE] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_RPAREN] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_and] = ACTIONS(4005),
    [anon_sym_or] = ACTIONS(4005),
    [anon_sym_xor] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_RBRACK] = ACTIONS(4005),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2396)] = {
    [sym_heredoc_content] = STATE(2396),
    [aux_sym__term_rightward_repeat1] = STATE(2320),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(4771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(4771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3485),
  },
  [STATE(2397)] = {
    [sym_heredoc_content] = STATE(2397),
    [anon_sym_RBRACE] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_RPAREN] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_and] = ACTIONS(3779),
    [anon_sym_or] = ACTIONS(3779),
    [anon_sym_xor] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_RBRACK] = ACTIONS(3779),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2398)] = {
    [sym_heredoc_content] = STATE(2398),
    [anon_sym_RBRACE] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_RPAREN] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_and] = ACTIONS(4135),
    [anon_sym_or] = ACTIONS(4135),
    [anon_sym_xor] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_RBRACK] = ACTIONS(4135),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2399)] = {
    [sym_heredoc_content] = STATE(2399),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_and] = ACTIONS(3586),
    [anon_sym_or] = ACTIONS(3586),
    [anon_sym_xor] = ACTIONS(3586),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_RBRACK] = ACTIONS(3584),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4816),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3584),
  },
  [STATE(2400)] = {
    [sym_heredoc_content] = STATE(2400),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_and] = ACTIONS(3550),
    [anon_sym_or] = ACTIONS(3550),
    [anon_sym_xor] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_RBRACK] = ACTIONS(3550),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4818),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3550),
  },
  [STATE(2401)] = {
    [sym_heredoc_content] = STATE(2401),
    [anon_sym_RBRACE] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_RPAREN] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_and] = ACTIONS(4111),
    [anon_sym_or] = ACTIONS(4111),
    [anon_sym_xor] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_RBRACK] = ACTIONS(4111),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2402)] = {
    [sym_heredoc_content] = STATE(2402),
    [anon_sym_RBRACE] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_RPAREN] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_and] = ACTIONS(4123),
    [anon_sym_or] = ACTIONS(4123),
    [anon_sym_xor] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_RBRACK] = ACTIONS(4123),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2403)] = {
    [sym_heredoc_content] = STATE(2403),
    [anon_sym_RBRACE] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_RPAREN] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_and] = ACTIONS(3925),
    [anon_sym_or] = ACTIONS(3925),
    [anon_sym_xor] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_RBRACK] = ACTIONS(3925),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2404)] = {
    [sym_heredoc_content] = STATE(2404),
    [anon_sym_RBRACE] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_RPAREN] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_and] = ACTIONS(3929),
    [anon_sym_or] = ACTIONS(3929),
    [anon_sym_xor] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_RBRACK] = ACTIONS(3929),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2405)] = {
    [sym_heredoc_content] = STATE(2405),
    [anon_sym_RBRACE] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_RPAREN] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_and] = ACTIONS(3945),
    [anon_sym_or] = ACTIONS(3945),
    [anon_sym_xor] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_RBRACK] = ACTIONS(3945),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2406)] = {
    [sym_heredoc_content] = STATE(2406),
    [anon_sym_RBRACE] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_RPAREN] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_and] = ACTIONS(4196),
    [anon_sym_or] = ACTIONS(4196),
    [anon_sym_xor] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_RBRACK] = ACTIONS(4196),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2407)] = {
    [sym_heredoc_content] = STATE(2407),
    [anon_sym_RBRACE] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_RPAREN] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_and] = ACTIONS(3933),
    [anon_sym_or] = ACTIONS(3933),
    [anon_sym_xor] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_RBRACK] = ACTIONS(3933),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2408)] = {
    [sym_heredoc_content] = STATE(2408),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3435),
  },
  [STATE(2409)] = {
    [sym_heredoc_content] = STATE(2409),
    [anon_sym_RBRACE] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_RPAREN] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_and] = ACTIONS(3905),
    [anon_sym_or] = ACTIONS(3905),
    [anon_sym_xor] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_RBRACK] = ACTIONS(3905),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2410)] = {
    [sym_heredoc_content] = STATE(2410),
    [anon_sym_RBRACE] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_RPAREN] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_and] = ACTIONS(4067),
    [anon_sym_or] = ACTIONS(4067),
    [anon_sym_xor] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_RBRACK] = ACTIONS(4067),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2411)] = {
    [sym_heredoc_content] = STATE(2411),
    [anon_sym_RBRACE] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_RPAREN] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_and] = ACTIONS(3937),
    [anon_sym_or] = ACTIONS(3937),
    [anon_sym_xor] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_RBRACK] = ACTIONS(3937),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2412)] = {
    [sym_heredoc_content] = STATE(2412),
    [anon_sym_RBRACE] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_RPAREN] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_and] = ACTIONS(3825),
    [anon_sym_or] = ACTIONS(3825),
    [anon_sym_xor] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_RBRACK] = ACTIONS(3825),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2413)] = {
    [sym_heredoc_content] = STATE(2413),
    [anon_sym_RBRACE] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_RPAREN] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_and] = ACTIONS(3849),
    [anon_sym_or] = ACTIONS(3849),
    [anon_sym_xor] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_RBRACK] = ACTIONS(3849),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2414)] = {
    [sym_heredoc_content] = STATE(2414),
    [anon_sym_RBRACE] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_RPAREN] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_and] = ACTIONS(4051),
    [anon_sym_or] = ACTIONS(4051),
    [anon_sym_xor] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_RBRACK] = ACTIONS(4051),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2415)] = {
    [sym_heredoc_content] = STATE(2415),
    [anon_sym_RBRACE] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_RPAREN] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_and] = ACTIONS(4059),
    [anon_sym_or] = ACTIONS(4059),
    [anon_sym_xor] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_RBRACK] = ACTIONS(4059),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2416)] = {
    [sym_heredoc_content] = STATE(2416),
    [anon_sym_RBRACE] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_and] = ACTIONS(3471),
    [anon_sym_or] = ACTIONS(3471),
    [anon_sym_xor] = ACTIONS(3471),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4820),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3469),
  },
  [STATE(2417)] = {
    [sym_heredoc_content] = STATE(2417),
    [anon_sym_RBRACE] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_RPAREN] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_and] = ACTIONS(4099),
    [anon_sym_or] = ACTIONS(4099),
    [anon_sym_xor] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_RBRACK] = ACTIONS(4099),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2418)] = {
    [sym_heredoc_content] = STATE(2418),
    [aux_sym__term_rightward_repeat1] = STATE(2418),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(4822),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_RBRACK] = ACTIONS(3590),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(4822),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3590),
  },
  [STATE(2419)] = {
    [sym_heredoc_content] = STATE(2419),
    [anon_sym_LBRACE] = ACTIONS(4825),
    [anon_sym_COLON] = ACTIONS(2711),
    [anon_sym_DOLLAR] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_PERCENT2] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2420)] = {
    [sym_heredoc_content] = STATE(2420),
    [anon_sym_COLON] = ACTIONS(3621),
    [anon_sym_DOLLAR] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_AT] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_RPAREN] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(4827),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3623),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_PERCENT2] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2421)] = {
    [sym_heredoc_content] = STATE(2421),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_DOLLAR] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_AT] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4829),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3417),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_PERCENT2] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2422)] = {
    [sym_heredoc_content] = STATE(2422),
    [anon_sym_RBRACE] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_RPAREN] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_and] = ACTIONS(4127),
    [anon_sym_or] = ACTIONS(4127),
    [anon_sym_xor] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_RBRACK] = ACTIONS(4127),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2423)] = {
    [sym_heredoc_content] = STATE(2423),
    [anon_sym_RBRACE] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_RPAREN] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_and] = ACTIONS(4147),
    [anon_sym_or] = ACTIONS(4147),
    [anon_sym_xor] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_RBRACK] = ACTIONS(4147),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2424)] = {
    [sym_heredoc_content] = STATE(2424),
    [anon_sym_LBRACE] = ACTIONS(2719),
    [anon_sym_RBRACE] = ACTIONS(2719),
    [anon_sym_COLON] = ACTIONS(2719),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_COMMA] = ACTIONS(2719),
    [anon_sym_RPAREN] = ACTIONS(2719),
    [anon_sym_LPAREN] = ACTIONS(2719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2719),
    [anon_sym_DASH_GT] = ACTIONS(2719),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2719),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2719),
    [anon_sym_PLUS_EQ] = ACTIONS(2719),
    [anon_sym_DASH_EQ] = ACTIONS(2719),
    [anon_sym_DOT_EQ] = ACTIONS(2719),
    [anon_sym_STAR_EQ] = ACTIONS(2719),
    [anon_sym_SLASH_EQ] = ACTIONS(2719),
    [anon_sym_PERCENT_EQ] = ACTIONS(2719),
    [anon_sym_x_EQ] = ACTIONS(2719),
    [anon_sym_AMP_EQ] = ACTIONS(2719),
    [anon_sym_PIPE_EQ] = ACTIONS(2719),
    [anon_sym_CARET_EQ] = ACTIONS(2719),
    [anon_sym_LT_LT_EQ] = ACTIONS(2719),
    [anon_sym_GT_GT_EQ] = ACTIONS(2719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2719),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH_SLASH] = ACTIONS(2717),
    [anon_sym_CARET_CARET] = ACTIONS(2719),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_GT_GT2] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_DOT] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_x] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2719),
    [anon_sym_BANG_TILDE] = ACTIONS(2719),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2719),
    [anon_sym_eq] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2719),
    [anon_sym_equ] = ACTIONS(2719),
    [anon_sym_eqr] = ACTIONS(2719),
    [anon_sym_ne] = ACTIONS(2719),
    [anon_sym_LT_EQ_GT] = ACTIONS(2719),
    [anon_sym_cmp] = ACTIONS(2719),
    [anon_sym_TILDE_TILDE] = ACTIONS(2719),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2719),
    [anon_sym_lt] = ACTIONS(2719),
    [anon_sym_le] = ACTIONS(2719),
    [anon_sym_ge] = ACTIONS(2719),
    [anon_sym_gt] = ACTIONS(2719),
    [anon_sym_isa] = ACTIONS(2719),
    [anon_sym_PLUS_PLUS] = ACTIONS(2719),
    [anon_sym_DASH_DASH] = ACTIONS(2719),
    [anon_sym_QMARK] = ACTIONS(2719),
    [anon_sym_EQ_GT] = ACTIONS(2719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2425)] = {
    [sym_heredoc_content] = STATE(2425),
    [anon_sym_RBRACE] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_and] = ACTIONS(3558),
    [anon_sym_or] = ACTIONS(3558),
    [anon_sym_xor] = ACTIONS(3558),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4831),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3556),
  },
  [STATE(2426)] = {
    [sym_heredoc_content] = STATE(2426),
    [anon_sym_RBRACE] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_and] = ACTIONS(3574),
    [anon_sym_or] = ACTIONS(3574),
    [anon_sym_xor] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(4833),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3574),
  },
  [STATE(2427)] = {
    [sym_heredoc_content] = STATE(2427),
    [anon_sym_RBRACE] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_RPAREN] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_and] = ACTIONS(3961),
    [anon_sym_or] = ACTIONS(3961),
    [anon_sym_xor] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_RBRACK] = ACTIONS(3961),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2428)] = {
    [sym_heredoc_content] = STATE(2428),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3435),
  },
  [STATE(2429)] = {
    [sym_heredoc_content] = STATE(2429),
    [anon_sym_COLON] = ACTIONS(3457),
    [anon_sym_DOLLAR] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_AT] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_RPAREN] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(4835),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3459),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_PERCENT2] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2430)] = {
    [sym_heredoc_content] = STATE(2430),
    [anon_sym_RBRACE] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_RPAREN] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_and] = ACTIONS(3913),
    [anon_sym_or] = ACTIONS(3913),
    [anon_sym_xor] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_RBRACK] = ACTIONS(3913),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2431)] = {
    [sym_heredoc_content] = STATE(2431),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_DOLLAR] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_AT] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(4827),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3697),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_PERCENT2] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2432)] = {
    [sym_heredoc_content] = STATE(2432),
    [anon_sym_RBRACE] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_RPAREN] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_and] = ACTIONS(4143),
    [anon_sym_or] = ACTIONS(4143),
    [anon_sym_xor] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_RBRACK] = ACTIONS(4143),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2433)] = {
    [sym_heredoc_content] = STATE(2433),
    [anon_sym_RBRACE] = ACTIONS(3627),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_and] = ACTIONS(3627),
    [anon_sym_or] = ACTIONS(3627),
    [anon_sym_xor] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3627),
  },
  [STATE(2434)] = {
    [sym_heredoc_content] = STATE(2434),
    [anon_sym_RBRACE] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_and] = ACTIONS(3699),
    [anon_sym_or] = ACTIONS(3699),
    [anon_sym_xor] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3699),
  },
  [STATE(2435)] = {
    [sym_heredoc_content] = STATE(2435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3435),
  },
  [STATE(2436)] = {
    [sym_heredoc_content] = STATE(2436),
    [anon_sym_RBRACE] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_and] = ACTIONS(3705),
    [anon_sym_or] = ACTIONS(3705),
    [anon_sym_xor] = ACTIONS(3705),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4837),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3703),
  },
  [STATE(2437)] = {
    [sym_heredoc_content] = STATE(2437),
    [anon_sym_RBRACE] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_and] = ACTIONS(3711),
    [anon_sym_or] = ACTIONS(3711),
    [anon_sym_xor] = ACTIONS(3711),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4839),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3709),
  },
  [STATE(2438)] = {
    [sym_heredoc_content] = STATE(2438),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_and] = ACTIONS(3953),
    [anon_sym_or] = ACTIONS(3953),
    [anon_sym_xor] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_RBRACK] = ACTIONS(3953),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2439)] = {
    [sym_heredoc_content] = STATE(2439),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_and] = ACTIONS(3965),
    [anon_sym_or] = ACTIONS(3965),
    [anon_sym_xor] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_RBRACK] = ACTIONS(3965),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2440)] = {
    [sym_heredoc_content] = STATE(2440),
    [anon_sym_RBRACE] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_and] = ACTIONS(3617),
    [anon_sym_or] = ACTIONS(3617),
    [anon_sym_xor] = ACTIONS(3617),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4841),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3615),
  },
  [STATE(2441)] = {
    [sym_heredoc_content] = STATE(2441),
    [anon_sym_RBRACE] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_and] = ACTIONS(3683),
    [anon_sym_or] = ACTIONS(3683),
    [anon_sym_xor] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4843),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3683),
  },
  [STATE(2442)] = {
    [sym_heredoc_content] = STATE(2442),
    [anon_sym_RBRACE] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_RPAREN] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_and] = ACTIONS(3799),
    [anon_sym_or] = ACTIONS(3799),
    [anon_sym_xor] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_RBRACK] = ACTIONS(3799),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2443)] = {
    [sym_heredoc_content] = STATE(2443),
    [anon_sym_RBRACE] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_RPAREN] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_and] = ACTIONS(3941),
    [anon_sym_or] = ACTIONS(3941),
    [anon_sym_xor] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_RBRACK] = ACTIONS(3941),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2444)] = {
    [sym_heredoc_content] = STATE(2444),
    [anon_sym_LBRACE] = ACTIONS(4845),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4847),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4849),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2445)] = {
    [sym_heredoc_content] = STATE(2445),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_RBRACE] = ACTIONS(3345),
    [anon_sym_COLON] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_RPAREN] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2446)] = {
    [sym_heredoc_content] = STATE(2446),
    [anon_sym_LBRACE] = ACTIONS(4851),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2711),
  },
  [STATE(2447)] = {
    [sym_heredoc_content] = STATE(2447),
    [anon_sym_RBRACE] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_RPAREN] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_and] = ACTIONS(3921),
    [anon_sym_or] = ACTIONS(3921),
    [anon_sym_xor] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_RBRACK] = ACTIONS(3921),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2448)] = {
    [sym_heredoc_content] = STATE(2448),
    [anon_sym_RBRACE] = ACTIONS(3665),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_and] = ACTIONS(3665),
    [anon_sym_or] = ACTIONS(3665),
    [anon_sym_xor] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3665),
  },
  [STATE(2449)] = {
    [sym_heredoc_content] = STATE(2449),
    [anon_sym_RBRACE] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_RPAREN] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_and] = ACTIONS(4087),
    [anon_sym_or] = ACTIONS(4087),
    [anon_sym_xor] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_RBRACK] = ACTIONS(4087),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2450)] = {
    [sym_heredoc_content] = STATE(2450),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(4853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3695),
  },
  [STATE(2451)] = {
    [sym_heredoc_content] = STATE(2451),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_DOLLAR] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_AT] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3437),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_PERCENT2] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2452)] = {
    [sym_heredoc_content] = STATE(2452),
    [anon_sym_RBRACE] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_RPAREN] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_and] = ACTIONS(4155),
    [anon_sym_or] = ACTIONS(4155),
    [anon_sym_xor] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_RBRACK] = ACTIONS(4155),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2453)] = {
    [sym_heredoc_content] = STATE(2453),
    [anon_sym_RBRACE] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(4855),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_and] = ACTIONS(3562),
    [anon_sym_or] = ACTIONS(3562),
    [anon_sym_xor] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3562),
  },
  [STATE(2454)] = {
    [sym_heredoc_content] = STATE(2454),
    [anon_sym_RBRACE] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(4853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_and] = ACTIONS(3621),
    [anon_sym_or] = ACTIONS(3621),
    [anon_sym_xor] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3621),
  },
  [STATE(2455)] = {
    [sym_heredoc_content] = STATE(2455),
    [anon_sym_LBRACE] = ACTIONS(4857),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2711),
  },
  [STATE(2456)] = {
    [sym_heredoc_content] = STATE(2456),
    [anon_sym_RBRACE] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_RPAREN] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_and] = ACTIONS(4176),
    [anon_sym_or] = ACTIONS(4176),
    [anon_sym_xor] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_RBRACK] = ACTIONS(4176),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2457)] = {
    [sym_heredoc_content] = STATE(2457),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_RPAREN] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_and] = ACTIONS(3665),
    [anon_sym_or] = ACTIONS(3665),
    [anon_sym_xor] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3665),
  },
  [STATE(2458)] = {
    [sym_heredoc_content] = STATE(2458),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_RPAREN] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_and] = ACTIONS(3597),
    [anon_sym_or] = ACTIONS(3597),
    [anon_sym_xor] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3597),
  },
  [STATE(2459)] = {
    [sym_heredoc_content] = STATE(2459),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_RPAREN] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_and] = ACTIONS(3611),
    [anon_sym_or] = ACTIONS(3611),
    [anon_sym_xor] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3611),
  },
  [STATE(2460)] = {
    [sym_heredoc_content] = STATE(2460),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3435),
  },
  [STATE(2461)] = {
    [sym_heredoc_content] = STATE(2461),
    [anon_sym_RBRACE] = ACTIONS(3679),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_and] = ACTIONS(3679),
    [anon_sym_or] = ACTIONS(3679),
    [anon_sym_xor] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3679),
  },
  [STATE(2462)] = {
    [sym_heredoc_content] = STATE(2462),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_RPAREN] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_and] = ACTIONS(3447),
    [anon_sym_or] = ACTIONS(3447),
    [anon_sym_xor] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3447),
  },
  [STATE(2463)] = {
    [sym_heredoc_content] = STATE(2463),
    [anon_sym_RBRACE] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_RPAREN] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_and] = ACTIONS(3751),
    [anon_sym_or] = ACTIONS(3751),
    [anon_sym_xor] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_RBRACK] = ACTIONS(3751),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2464)] = {
    [sym_heredoc_content] = STATE(2464),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3435),
  },
  [STATE(2465)] = {
    [sym_heredoc_content] = STATE(2465),
    [anon_sym_RBRACE] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_RPAREN] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_and] = ACTIONS(3755),
    [anon_sym_or] = ACTIONS(3755),
    [anon_sym_xor] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_RBRACK] = ACTIONS(3755),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2466)] = {
    [sym_heredoc_content] = STATE(2466),
    [anon_sym_RBRACE] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_RPAREN] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_and] = ACTIONS(3759),
    [anon_sym_or] = ACTIONS(3759),
    [anon_sym_xor] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_RBRACK] = ACTIONS(3759),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2467)] = {
    [sym_heredoc_content] = STATE(2467),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_RBRACK] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2468)] = {
    [sym_heredoc_content] = STATE(2468),
    [anon_sym_RBRACE] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_RPAREN] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_and] = ACTIONS(4180),
    [anon_sym_or] = ACTIONS(4180),
    [anon_sym_xor] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_RBRACK] = ACTIONS(4180),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2469)] = {
    [sym_heredoc_content] = STATE(2469),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3435),
  },
  [STATE(2470)] = {
    [sym_heredoc_content] = STATE(2470),
    [anon_sym_RBRACE] = ACTIONS(3661),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_and] = ACTIONS(3661),
    [anon_sym_or] = ACTIONS(3661),
    [anon_sym_xor] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3661),
  },
  [STATE(2471)] = {
    [sym_heredoc_content] = STATE(2471),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_RPAREN] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(4859),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_and] = ACTIONS(3633),
    [anon_sym_or] = ACTIONS(3633),
    [anon_sym_xor] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3633),
  },
  [STATE(2472)] = {
    [sym_heredoc_content] = STATE(2472),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_DOLLAR] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_AT] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3437),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_PERCENT2] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2473)] = {
    [sym_heredoc_content] = STATE(2473),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_DOLLAR] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_AT] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3437),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_PERCENT2] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2474)] = {
    [sym_heredoc_content] = STATE(2474),
    [anon_sym_COLON] = ACTIONS(3601),
    [anon_sym_DOLLAR] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_AT] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_RPAREN] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3603),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_PERCENT2] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2475)] = {
    [sym_heredoc_content] = STATE(2475),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_COLON] = ACTIONS(3605),
    [anon_sym_DOLLAR] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_AT] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_RPAREN] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3607),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_PERCENT2] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2476)] = {
    [sym_heredoc_content] = STATE(2476),
    [anon_sym_RBRACE] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_RPAREN] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_and] = ACTIONS(4188),
    [anon_sym_or] = ACTIONS(4188),
    [anon_sym_xor] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_RBRACK] = ACTIONS(4188),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2477)] = {
    [sym_heredoc_content] = STATE(2477),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_and] = ACTIONS(3437),
    [anon_sym_or] = ACTIONS(3437),
    [anon_sym_xor] = ACTIONS(3437),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_RBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3435),
  },
  [STATE(2478)] = {
    [sym_heredoc_content] = STATE(2478),
    [anon_sym_COLON] = ACTIONS(3655),
    [anon_sym_DOLLAR] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_AT] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_RPAREN] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3657),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_PERCENT2] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4861),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2479)] = {
    [sym_heredoc_content] = STATE(2479),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2480)] = {
    [sym_heredoc_content] = STATE(2480),
    [anon_sym_COLON] = ACTIONS(3649),
    [anon_sym_DOLLAR] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_AT] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_RPAREN] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(4835),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3651),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_PERCENT2] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2481)] = {
    [sym_heredoc_content] = STATE(2481),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_SEMI] = ACTIONS(3333),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym__PERLY_SEMICOLON] = ACTIONS(3333),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2482)] = {
    [sym_heredoc_content] = STATE(2482),
    [anon_sym_RBRACE] = ACTIONS(3689),
    [anon_sym_COLON] = ACTIONS(4863),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3689),
  },
  [STATE(2483)] = {
    [sym_heredoc_content] = STATE(2483),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_SEMI] = ACTIONS(3337),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym__PERLY_SEMICOLON] = ACTIONS(3337),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4865),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2484)] = {
    [sym_heredoc_content] = STATE(2484),
    [anon_sym_RBRACE] = ACTIONS(3669),
    [anon_sym_COLON] = ACTIONS(4867),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3669),
  },
  [STATE(2485)] = {
    [sym_heredoc_content] = STATE(2485),
    [anon_sym_RBRACE] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_RPAREN] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_and] = ACTIONS(4192),
    [anon_sym_or] = ACTIONS(4192),
    [anon_sym_xor] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_RBRACK] = ACTIONS(4192),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2486)] = {
    [sym_heredoc_content] = STATE(2486),
    [anon_sym_RBRACE] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_RPAREN] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_and] = ACTIONS(3833),
    [anon_sym_or] = ACTIONS(3833),
    [anon_sym_xor] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_RBRACK] = ACTIONS(3833),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2487)] = {
    [sym_heredoc_content] = STATE(2487),
    [anon_sym_LBRACE] = ACTIONS(2715),
    [anon_sym_RBRACE] = ACTIONS(2715),
    [anon_sym_COLON] = ACTIONS(2715),
    [anon_sym_EQ] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_COMMA] = ACTIONS(2715),
    [anon_sym_RPAREN] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2715),
    [anon_sym_DASH_GT] = ACTIONS(2715),
    [anon_sym_STAR] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(2715),
    [anon_sym_PERCENT] = ACTIONS(2713),
    [anon_sym_GT] = ACTIONS(2713),
    [anon_sym_LT_LT] = ACTIONS(2713),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2715),
    [anon_sym_PLUS_EQ] = ACTIONS(2715),
    [anon_sym_DASH_EQ] = ACTIONS(2715),
    [anon_sym_DOT_EQ] = ACTIONS(2715),
    [anon_sym_STAR_EQ] = ACTIONS(2715),
    [anon_sym_SLASH_EQ] = ACTIONS(2715),
    [anon_sym_PERCENT_EQ] = ACTIONS(2715),
    [anon_sym_x_EQ] = ACTIONS(2715),
    [anon_sym_AMP_EQ] = ACTIONS(2715),
    [anon_sym_PIPE_EQ] = ACTIONS(2715),
    [anon_sym_CARET_EQ] = ACTIONS(2715),
    [anon_sym_LT_LT_EQ] = ACTIONS(2715),
    [anon_sym_GT_GT_EQ] = ACTIONS(2715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2715),
    [anon_sym_DOT_DOT] = ACTIONS(2713),
    [anon_sym_STAR_STAR] = ACTIONS(2713),
    [anon_sym_PIPE_PIPE] = ACTIONS(2713),
    [anon_sym_SLASH_SLASH] = ACTIONS(2713),
    [anon_sym_CARET_CARET] = ACTIONS(2715),
    [anon_sym_AMP_AMP] = ACTIONS(2713),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_CARET] = ACTIONS(2713),
    [anon_sym_AMP] = ACTIONS(2713),
    [anon_sym_GT_GT2] = ACTIONS(2713),
    [anon_sym_PLUS] = ACTIONS(2713),
    [anon_sym_DASH] = ACTIONS(2713),
    [anon_sym_DOT] = ACTIONS(2713),
    [anon_sym_SLASH] = ACTIONS(2713),
    [anon_sym_x] = ACTIONS(2713),
    [anon_sym_EQ_TILDE] = ACTIONS(2715),
    [anon_sym_BANG_TILDE] = ACTIONS(2715),
    [anon_sym_EQ_EQ] = ACTIONS(2713),
    [anon_sym_BANG_EQ] = ACTIONS(2715),
    [anon_sym_eq] = ACTIONS(2713),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2715),
    [anon_sym_equ] = ACTIONS(2715),
    [anon_sym_eqr] = ACTIONS(2715),
    [anon_sym_ne] = ACTIONS(2715),
    [anon_sym_LT_EQ_GT] = ACTIONS(2715),
    [anon_sym_cmp] = ACTIONS(2715),
    [anon_sym_TILDE_TILDE] = ACTIONS(2715),
    [anon_sym_LT] = ACTIONS(2713),
    [anon_sym_LT_EQ] = ACTIONS(2713),
    [anon_sym_GT_EQ] = ACTIONS(2715),
    [anon_sym_lt] = ACTIONS(2715),
    [anon_sym_le] = ACTIONS(2715),
    [anon_sym_ge] = ACTIONS(2715),
    [anon_sym_gt] = ACTIONS(2715),
    [anon_sym_isa] = ACTIONS(2715),
    [anon_sym_PLUS_PLUS] = ACTIONS(2715),
    [anon_sym_DASH_DASH] = ACTIONS(2715),
    [anon_sym_QMARK] = ACTIONS(2715),
    [anon_sym_EQ_GT] = ACTIONS(2715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2488)] = {
    [sym_heredoc_content] = STATE(2488),
    [anon_sym_RBRACE] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_RPAREN] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_and] = ACTIONS(4200),
    [anon_sym_or] = ACTIONS(4200),
    [anon_sym_xor] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_RBRACK] = ACTIONS(4200),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2489)] = {
    [sym_heredoc_content] = STATE(2489),
    [anon_sym_RBRACE] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(4107),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2490)] = {
    [sym_heredoc_content] = STATE(2490),
    [anon_sym_COLON] = ACTIONS(3451),
    [anon_sym_DOLLAR] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_AT] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_RPAREN] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(4869),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3453),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_PERCENT2] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2491)] = {
    [sym_heredoc_content] = STATE(2491),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_RBRACE] = ACTIONS(3361),
    [anon_sym_COLON] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_RPAREN] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2492)] = {
    [sym_heredoc_content] = STATE(2492),
    [anon_sym_RBRACE] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_and] = ACTIONS(3540),
    [anon_sym_or] = ACTIONS(3540),
    [anon_sym_xor] = ACTIONS(3540),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4871),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3538),
  },
  [STATE(2493)] = {
    [sym_heredoc_content] = STATE(2493),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3590),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2494)] = {
    [sym_heredoc_content] = STATE(2494),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(4873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_and] = ACTIONS(3633),
    [anon_sym_or] = ACTIONS(3633),
    [anon_sym_xor] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_RBRACK] = ACTIONS(3633),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3633),
  },
  [STATE(2495)] = {
    [sym_heredoc_content] = STATE(2495),
    [anon_sym_COLON] = ACTIONS(3639),
    [anon_sym_DOLLAR] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_AT] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_RPAREN] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3641),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_PERCENT2] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2496)] = {
    [sym_heredoc_content] = STATE(2496),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_RBRACE] = ACTIONS(3397),
    [anon_sym_COLON] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_RPAREN] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2497)] = {
    [sym_heredoc_content] = STATE(2497),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_RBRACE] = ACTIONS(3349),
    [anon_sym_COLON] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_RPAREN] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2498)] = {
    [sym_heredoc_content] = STATE(2498),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_RBRACE] = ACTIONS(3365),
    [anon_sym_COLON] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_RPAREN] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2499)] = {
    [sym_heredoc_content] = STATE(2499),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_RBRACE] = ACTIONS(3385),
    [anon_sym_COLON] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_RPAREN] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2500)] = {
    [sym_heredoc_content] = STATE(2500),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_RBRACE] = ACTIONS(3389),
    [anon_sym_COLON] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_RPAREN] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2501)] = {
    [sym_heredoc_content] = STATE(2501),
    [anon_sym_COLON] = ACTIONS(3643),
    [anon_sym_DOLLAR] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_AT] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_RPAREN] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3645),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_PERCENT2] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4875),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2502)] = {
    [sym_heredoc_content] = STATE(2502),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_RBRACE] = ACTIONS(3405),
    [anon_sym_COLON] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_RPAREN] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2503)] = {
    [sym_heredoc_content] = STATE(2503),
    [anon_sym_COLON] = ACTIONS(3463),
    [anon_sym_DOLLAR] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_AT] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_RPAREN] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3465),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_PERCENT2] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4877),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2504)] = {
    [sym_heredoc_content] = STATE(2504),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_and] = ACTIONS(3471),
    [anon_sym_or] = ACTIONS(3471),
    [anon_sym_xor] = ACTIONS(3471),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_RBRACK] = ACTIONS(3469),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4879),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3469),
  },
  [STATE(2505)] = {
    [sym_heredoc_content] = STATE(2505),
    [anon_sym_COLON] = ACTIONS(3475),
    [anon_sym_DOLLAR] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_AT] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_RPAREN] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3477),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_PERCENT2] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2506)] = {
    [sym_heredoc_content] = STATE(2506),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_RBRACE] = ACTIONS(3373),
    [anon_sym_COLON] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_RPAREN] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2507)] = {
    [sym_heredoc_content] = STATE(2507),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_RBRACE] = ACTIONS(3381),
    [anon_sym_COLON] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_RPAREN] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2508)] = {
    [sym_heredoc_content] = STATE(2508),
    [anon_sym_COLON] = ACTIONS(3479),
    [anon_sym_DOLLAR] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_AT] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_RPAREN] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(4881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3481),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_PERCENT2] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2509)] = {
    [sym_heredoc_content] = STATE(2509),
    [anon_sym_RBRACE] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_and] = ACTIONS(3546),
    [anon_sym_or] = ACTIONS(3546),
    [anon_sym_xor] = ACTIONS(3546),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4883),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3544),
  },
  [STATE(2510)] = {
    [sym_heredoc_content] = STATE(2510),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_RBRACE] = ACTIONS(3341),
    [anon_sym_COLON] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_RPAREN] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2511)] = {
    [sym_heredoc_content] = STATE(2511),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_RBRACE] = ACTIONS(3353),
    [anon_sym_COLON] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_RPAREN] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2512)] = {
    [sym_heredoc_content] = STATE(2512),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_RBRACE] = ACTIONS(3357),
    [anon_sym_COLON] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_RPAREN] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2513)] = {
    [sym_heredoc_content] = STATE(2513),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_RBRACE] = ACTIONS(3369),
    [anon_sym_COLON] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_RPAREN] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2514)] = {
    [sym_heredoc_content] = STATE(2514),
    [anon_sym_RBRACE] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_and] = ACTIONS(3578),
    [anon_sym_or] = ACTIONS(3578),
    [anon_sym_xor] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4885),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3578),
  },
  [STATE(2515)] = {
    [sym_heredoc_content] = STATE(2515),
    [anon_sym_RBRACE] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_and] = ACTIONS(3451),
    [anon_sym_or] = ACTIONS(3451),
    [anon_sym_xor] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(4833),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3451),
  },
  [STATE(2516)] = {
    [sym_heredoc_content] = STATE(2516),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_RBRACE] = ACTIONS(3393),
    [anon_sym_COLON] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_RPAREN] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2517)] = {
    [sym_heredoc_content] = STATE(2517),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_RBRACE] = ACTIONS(3377),
    [anon_sym_COLON] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_RPAREN] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2518)] = {
    [sym_heredoc_content] = STATE(2518),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_RPAREN] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_and] = ACTIONS(3675),
    [anon_sym_or] = ACTIONS(3675),
    [anon_sym_xor] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3675),
  },
  [STATE(2519)] = {
    [sym_heredoc_content] = STATE(2519),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_RPAREN] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_and] = ACTIONS(3601),
    [anon_sym_or] = ACTIONS(3601),
    [anon_sym_xor] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3601),
  },
  [STATE(2520)] = {
    [sym_heredoc_content] = STATE(2520),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_RPAREN] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_and] = ACTIONS(3605),
    [anon_sym_or] = ACTIONS(3605),
    [anon_sym_xor] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3605),
  },
  [STATE(2521)] = {
    [sym_heredoc_content] = STATE(2521),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4887),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4889),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2522)] = {
    [sym_heredoc_content] = STATE(2522),
    [anon_sym_RBRACE] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_RPAREN] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_and] = ACTIONS(3969),
    [anon_sym_or] = ACTIONS(3969),
    [anon_sym_xor] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_RBRACK] = ACTIONS(3969),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2523)] = {
    [sym_heredoc_content] = STATE(2523),
    [anon_sym_RBRACE] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_RPAREN] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_and] = ACTIONS(3841),
    [anon_sym_or] = ACTIONS(3841),
    [anon_sym_xor] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_RBRACK] = ACTIONS(3841),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2524)] = {
    [sym_heredoc_content] = STATE(2524),
    [anon_sym_RBRACE] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_RPAREN] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_and] = ACTIONS(4165),
    [anon_sym_or] = ACTIONS(4165),
    [anon_sym_xor] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_RBRACK] = ACTIONS(4165),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2525)] = {
    [sym_attribute] = STATE(2743),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2525),
    [aux_sym_attrlist_repeat1] = STATE(2531),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_COLON] = ACTIONS(3317),
    [anon_sym_EQ] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_COMMA] = ACTIONS(3313),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3313),
    [anon_sym_DASH_GT] = ACTIONS(3313),
    [anon_sym_STAR] = ACTIONS(3317),
    [anon_sym_PERCENT] = ACTIONS(3317),
    [anon_sym_GT] = ACTIONS(3317),
    [anon_sym_LT_LT] = ACTIONS(3317),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3313),
    [anon_sym_PLUS_EQ] = ACTIONS(3313),
    [anon_sym_DASH_EQ] = ACTIONS(3313),
    [anon_sym_DOT_EQ] = ACTIONS(3313),
    [anon_sym_STAR_EQ] = ACTIONS(3313),
    [anon_sym_SLASH_EQ] = ACTIONS(3313),
    [anon_sym_PERCENT_EQ] = ACTIONS(3313),
    [anon_sym_x_EQ] = ACTIONS(3313),
    [anon_sym_AMP_EQ] = ACTIONS(3313),
    [anon_sym_PIPE_EQ] = ACTIONS(3313),
    [anon_sym_CARET_EQ] = ACTIONS(3313),
    [anon_sym_LT_LT_EQ] = ACTIONS(3313),
    [anon_sym_GT_GT_EQ] = ACTIONS(3313),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3313),
    [anon_sym_DOT_DOT] = ACTIONS(3317),
    [anon_sym_STAR_STAR] = ACTIONS(3317),
    [anon_sym_PIPE_PIPE] = ACTIONS(3317),
    [anon_sym_SLASH_SLASH] = ACTIONS(3317),
    [anon_sym_CARET_CARET] = ACTIONS(3313),
    [anon_sym_AMP_AMP] = ACTIONS(3317),
    [anon_sym_PIPE] = ACTIONS(3317),
    [anon_sym_CARET] = ACTIONS(3317),
    [anon_sym_AMP] = ACTIONS(3317),
    [anon_sym_GT_GT2] = ACTIONS(3317),
    [anon_sym_PLUS] = ACTIONS(3317),
    [anon_sym_DASH] = ACTIONS(3317),
    [anon_sym_DOT] = ACTIONS(3317),
    [anon_sym_SLASH] = ACTIONS(3317),
    [anon_sym_x] = ACTIONS(3317),
    [anon_sym_EQ_TILDE] = ACTIONS(3313),
    [anon_sym_BANG_TILDE] = ACTIONS(3313),
    [anon_sym_EQ_EQ] = ACTIONS(3317),
    [anon_sym_BANG_EQ] = ACTIONS(3313),
    [anon_sym_eq] = ACTIONS(3317),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3313),
    [anon_sym_equ] = ACTIONS(3317),
    [anon_sym_eqr] = ACTIONS(3317),
    [anon_sym_ne] = ACTIONS(3317),
    [anon_sym_LT_EQ_GT] = ACTIONS(3313),
    [anon_sym_cmp] = ACTIONS(3317),
    [anon_sym_TILDE_TILDE] = ACTIONS(3313),
    [anon_sym_LT] = ACTIONS(3317),
    [anon_sym_LT_EQ] = ACTIONS(3317),
    [anon_sym_GT_EQ] = ACTIONS(3313),
    [anon_sym_lt] = ACTIONS(3317),
    [anon_sym_le] = ACTIONS(3317),
    [anon_sym_ge] = ACTIONS(3317),
    [anon_sym_gt] = ACTIONS(3317),
    [anon_sym_isa] = ACTIONS(3317),
    [anon_sym_PLUS_PLUS] = ACTIONS(3313),
    [anon_sym_DASH_DASH] = ACTIONS(3313),
    [anon_sym_QMARK] = ACTIONS(3313),
    [anon_sym_EQ_GT] = ACTIONS(3313),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2526)] = {
    [sym_heredoc_content] = STATE(2526),
    [anon_sym_RBRACE] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_RPAREN] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_and] = ACTIONS(3853),
    [anon_sym_or] = ACTIONS(3853),
    [anon_sym_xor] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_RBRACK] = ACTIONS(3853),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2527)] = {
    [sym_heredoc_content] = STATE(2527),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_RPAREN] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_and] = ACTIONS(3657),
    [anon_sym_or] = ACTIONS(3657),
    [anon_sym_xor] = ACTIONS(3657),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4891),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3655),
  },
  [STATE(2528)] = {
    [sym_heredoc_content] = STATE(2528),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2529)] = {
    [sym_heredoc_content] = STATE(2529),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3735),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2530)] = {
    [sym_heredoc_content] = STATE(2530),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2531)] = {
    [sym_attribute] = STATE(2743),
    [sym_attribute_name] = STATE(2581),
    [sym_heredoc_content] = STATE(2531),
    [aux_sym_attrlist_repeat1] = STATE(2098),
    [sym__identifier] = ACTIONS(4360),
    [anon_sym_COLON] = ACTIONS(3321),
    [anon_sym_EQ] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_COMMA] = ACTIONS(3319),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3319),
    [anon_sym_DASH_GT] = ACTIONS(3319),
    [anon_sym_STAR] = ACTIONS(3321),
    [anon_sym_PERCENT] = ACTIONS(3321),
    [anon_sym_GT] = ACTIONS(3321),
    [anon_sym_LT_LT] = ACTIONS(3321),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3319),
    [anon_sym_PLUS_EQ] = ACTIONS(3319),
    [anon_sym_DASH_EQ] = ACTIONS(3319),
    [anon_sym_DOT_EQ] = ACTIONS(3319),
    [anon_sym_STAR_EQ] = ACTIONS(3319),
    [anon_sym_SLASH_EQ] = ACTIONS(3319),
    [anon_sym_PERCENT_EQ] = ACTIONS(3319),
    [anon_sym_x_EQ] = ACTIONS(3319),
    [anon_sym_AMP_EQ] = ACTIONS(3319),
    [anon_sym_PIPE_EQ] = ACTIONS(3319),
    [anon_sym_CARET_EQ] = ACTIONS(3319),
    [anon_sym_LT_LT_EQ] = ACTIONS(3319),
    [anon_sym_GT_GT_EQ] = ACTIONS(3319),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3319),
    [anon_sym_DOT_DOT] = ACTIONS(3321),
    [anon_sym_STAR_STAR] = ACTIONS(3321),
    [anon_sym_PIPE_PIPE] = ACTIONS(3321),
    [anon_sym_SLASH_SLASH] = ACTIONS(3321),
    [anon_sym_CARET_CARET] = ACTIONS(3319),
    [anon_sym_AMP_AMP] = ACTIONS(3321),
    [anon_sym_PIPE] = ACTIONS(3321),
    [anon_sym_CARET] = ACTIONS(3321),
    [anon_sym_AMP] = ACTIONS(3321),
    [anon_sym_GT_GT2] = ACTIONS(3321),
    [anon_sym_PLUS] = ACTIONS(3321),
    [anon_sym_DASH] = ACTIONS(3321),
    [anon_sym_DOT] = ACTIONS(3321),
    [anon_sym_SLASH] = ACTIONS(3321),
    [anon_sym_x] = ACTIONS(3321),
    [anon_sym_EQ_TILDE] = ACTIONS(3319),
    [anon_sym_BANG_TILDE] = ACTIONS(3319),
    [anon_sym_EQ_EQ] = ACTIONS(3321),
    [anon_sym_BANG_EQ] = ACTIONS(3319),
    [anon_sym_eq] = ACTIONS(3321),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3319),
    [anon_sym_equ] = ACTIONS(3321),
    [anon_sym_eqr] = ACTIONS(3321),
    [anon_sym_ne] = ACTIONS(3321),
    [anon_sym_LT_EQ_GT] = ACTIONS(3319),
    [anon_sym_cmp] = ACTIONS(3321),
    [anon_sym_TILDE_TILDE] = ACTIONS(3319),
    [anon_sym_LT] = ACTIONS(3321),
    [anon_sym_LT_EQ] = ACTIONS(3321),
    [anon_sym_GT_EQ] = ACTIONS(3319),
    [anon_sym_lt] = ACTIONS(3321),
    [anon_sym_le] = ACTIONS(3321),
    [anon_sym_ge] = ACTIONS(3321),
    [anon_sym_gt] = ACTIONS(3321),
    [anon_sym_isa] = ACTIONS(3321),
    [anon_sym_PLUS_PLUS] = ACTIONS(3319),
    [anon_sym_DASH_DASH] = ACTIONS(3319),
    [anon_sym_QMARK] = ACTIONS(3319),
    [anon_sym_EQ_GT] = ACTIONS(3319),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(4360),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2532)] = {
    [sym_heredoc_content] = STATE(2532),
    [anon_sym_RBRACE] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_RPAREN] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_and] = ACTIONS(4079),
    [anon_sym_or] = ACTIONS(4079),
    [anon_sym_xor] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_RBRACK] = ACTIONS(4079),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2533)] = {
    [sym_heredoc_content] = STATE(2533),
    [anon_sym_RBRACE] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_RPAREN] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_and] = ACTIONS(4083),
    [anon_sym_or] = ACTIONS(4083),
    [anon_sym_xor] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_RBRACK] = ACTIONS(4083),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2534)] = {
    [sym_heredoc_content] = STATE(2534),
    [anon_sym_RBRACE] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_RPAREN] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_and] = ACTIONS(4091),
    [anon_sym_or] = ACTIONS(4091),
    [anon_sym_xor] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_RBRACK] = ACTIONS(4091),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2535)] = {
    [sym_heredoc_content] = STATE(2535),
    [anon_sym_RBRACE] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_RPAREN] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_and] = ACTIONS(3715),
    [anon_sym_or] = ACTIONS(3715),
    [anon_sym_xor] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_RBRACK] = ACTIONS(3715),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2536)] = {
    [sym_heredoc_content] = STATE(2536),
    [anon_sym_RBRACE] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_RPAREN] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_and] = ACTIONS(4204),
    [anon_sym_or] = ACTIONS(4204),
    [anon_sym_xor] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_RBRACK] = ACTIONS(4204),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2537)] = {
    [sym_heredoc_content] = STATE(2537),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_RPAREN] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(4599),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_and] = ACTIONS(3621),
    [anon_sym_or] = ACTIONS(3621),
    [anon_sym_xor] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3621),
  },
  [STATE(2538)] = {
    [sym_heredoc_content] = STATE(2538),
    [anon_sym_COLON] = ACTIONS(3556),
    [anon_sym_DOLLAR] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_AT] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_RPAREN] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3558),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_PERCENT2] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4893),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2539)] = {
    [sym_heredoc_content] = STATE(2539),
    [anon_sym_COLON] = ACTIONS(3574),
    [anon_sym_DOLLAR] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_AT] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_RPAREN] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(4869),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3576),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_PERCENT2] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2540)] = {
    [sym_heredoc_content] = STATE(2540),
    [aux_sym__term_rightward_repeat1] = STATE(4575),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(4895),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(4895),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3485),
  },
  [STATE(2541)] = {
    [sym_heredoc_content] = STATE(2541),
    [anon_sym_RBRACE] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_RPAREN] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_and] = ACTIONS(3869),
    [anon_sym_or] = ACTIONS(3869),
    [anon_sym_xor] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_RBRACK] = ACTIONS(3869),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2542)] = {
    [sym_heredoc_content] = STATE(2542),
    [anon_sym_RBRACE] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_RPAREN] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_and] = ACTIONS(4103),
    [anon_sym_or] = ACTIONS(4103),
    [anon_sym_xor] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_RBRACK] = ACTIONS(4103),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2543)] = {
    [sym_heredoc_content] = STATE(2543),
    [anon_sym_RBRACE] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_RBRACK] = ACTIONS(3997),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2544)] = {
    [sym_heredoc_content] = STATE(2544),
    [anon_sym_RBRACE] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_RPAREN] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_and] = ACTIONS(4159),
    [anon_sym_or] = ACTIONS(4159),
    [anon_sym_xor] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_RBRACK] = ACTIONS(4159),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2545)] = {
    [sym_heredoc_content] = STATE(2545),
    [anon_sym_RBRACE] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_RPAREN] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_and] = ACTIONS(3893),
    [anon_sym_or] = ACTIONS(3893),
    [anon_sym_xor] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_RBRACK] = ACTIONS(3893),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2546)] = {
    [sym_heredoc_content] = STATE(2546),
    [anon_sym_RBRACE] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_RPAREN] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_and] = ACTIONS(4169),
    [anon_sym_or] = ACTIONS(4169),
    [anon_sym_xor] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_RBRACK] = ACTIONS(4169),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(4897),
    [anon_sym_DASH_DASH] = ACTIONS(4897),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2547)] = {
    [sym_heredoc_content] = STATE(2547),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_RBRACK] = ACTIONS(3695),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2548)] = {
    [sym_heredoc_content] = STATE(2548),
    [anon_sym_LBRACE] = ACTIONS(4900),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_RBRACK] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2711),
  },
  [STATE(2549)] = {
    [sym_heredoc_content] = STATE(2549),
    [anon_sym_RBRACE] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_RPAREN] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_and] = ACTIONS(4037),
    [anon_sym_or] = ACTIONS(4037),
    [anon_sym_xor] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_RBRACK] = ACTIONS(4037),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2550)] = {
    [sym_heredoc_content] = STATE(2550),
    [anon_sym_RBRACE] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3909),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2551)] = {
    [sym_heredoc_content] = STATE(2551),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_RBRACK] = ACTIONS(4017),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2552)] = {
    [sym_heredoc_content] = STATE(2552),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(4017),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2553)] = {
    [sym_heredoc_content] = STATE(2553),
    [anon_sym_RBRACE] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3803),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2554)] = {
    [sym_heredoc_content] = STATE(2554),
    [anon_sym_RBRACE] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3845),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2555)] = {
    [sym_heredoc_content] = STATE(2555),
    [anon_sym_RBRACE] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_RPAREN] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_and] = ACTIONS(3857),
    [anon_sym_or] = ACTIONS(3857),
    [anon_sym_xor] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_RBRACK] = ACTIONS(3857),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2556)] = {
    [sym_heredoc_content] = STATE(2556),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_RBRACK] = ACTIONS(2858),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2557)] = {
    [sym_heredoc_content] = STATE(2557),
    [anon_sym_RBRACE] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3739),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2558)] = {
    [sym_heredoc_content] = STATE(2558),
    [anon_sym_RBRACE] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_RPAREN] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_and] = ACTIONS(3815),
    [anon_sym_or] = ACTIONS(3815),
    [anon_sym_xor] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_RBRACK] = ACTIONS(3815),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2559)] = {
    [sym_heredoc_content] = STATE(2559),
    [anon_sym_COLON] = ACTIONS(3699),
    [anon_sym_DOLLAR] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_AT] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_RPAREN] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3701),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_PERCENT2] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2560)] = {
    [sym_heredoc_content] = STATE(2560),
    [anon_sym_COLON] = ACTIONS(3703),
    [anon_sym_DOLLAR] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_AT] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_RPAREN] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3705),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_PERCENT2] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(4902),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2561)] = {
    [sym_heredoc_content] = STATE(2561),
    [anon_sym_COLON] = ACTIONS(3709),
    [anon_sym_DOLLAR] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_AT] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_RPAREN] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3711),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_PERCENT2] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(4904),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2562)] = {
    [sym_heredoc_content] = STATE(2562),
    [anon_sym_RBRACE] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_RPAREN] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_and] = ACTIONS(3775),
    [anon_sym_or] = ACTIONS(3775),
    [anon_sym_xor] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_RBRACK] = ACTIONS(3775),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2563)] = {
    [sym_heredoc_content] = STATE(2563),
    [anon_sym_COLON] = ACTIONS(3615),
    [anon_sym_DOLLAR] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_AT] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_RPAREN] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3617),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_PERCENT2] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4906),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2564)] = {
    [sym_heredoc_content] = STATE(2564),
    [anon_sym_COLON] = ACTIONS(3683),
    [anon_sym_DOLLAR] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_AT] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_RPAREN] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3685),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_PERCENT2] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4908),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2565)] = {
    [sym_heredoc_content] = STATE(2565),
    [anon_sym_RBRACE] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(4408),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4414),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3897),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4410),
    [anon_sym_PLUS_EQ] = ACTIONS(4410),
    [anon_sym_DASH_EQ] = ACTIONS(4410),
    [anon_sym_DOT_EQ] = ACTIONS(4410),
    [anon_sym_STAR_EQ] = ACTIONS(4410),
    [anon_sym_SLASH_EQ] = ACTIONS(4410),
    [anon_sym_PERCENT_EQ] = ACTIONS(4410),
    [anon_sym_x_EQ] = ACTIONS(4410),
    [anon_sym_AMP_EQ] = ACTIONS(4410),
    [anon_sym_PIPE_EQ] = ACTIONS(4410),
    [anon_sym_CARET_EQ] = ACTIONS(4410),
    [anon_sym_LT_LT_EQ] = ACTIONS(4410),
    [anon_sym_GT_GT_EQ] = ACTIONS(4410),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4410),
    [anon_sym_DOT_DOT] = ACTIONS(4424),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4428),
    [anon_sym_SLASH_SLASH] = ACTIONS(4428),
    [anon_sym_CARET_CARET] = ACTIONS(4430),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4454),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2566)] = {
    [sym_heredoc_content] = STATE(2566),
    [anon_sym_RBRACE] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_RPAREN] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_and] = ACTIONS(3901),
    [anon_sym_or] = ACTIONS(3901),
    [anon_sym_xor] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_RBRACK] = ACTIONS(3901),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2567)] = {
    [sym_heredoc_content] = STATE(2567),
    [anon_sym_RBRACE] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_RPAREN] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_and] = ACTIONS(4119),
    [anon_sym_or] = ACTIONS(4119),
    [anon_sym_xor] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_RBRACK] = ACTIONS(4119),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2568)] = {
    [sym_heredoc_content] = STATE(2568),
    [anon_sym_RBRACE] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_RPAREN] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_and] = ACTIONS(4131),
    [anon_sym_or] = ACTIONS(4131),
    [anon_sym_xor] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_RBRACK] = ACTIONS(4131),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2569)] = {
    [sym_heredoc_content] = STATE(2569),
    [anon_sym_RBRACE] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(4151),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2570)] = {
    [sym_heredoc_content] = STATE(2570),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(4910),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_and] = ACTIONS(3621),
    [anon_sym_or] = ACTIONS(3621),
    [anon_sym_xor] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_RBRACK] = ACTIONS(3621),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3621),
  },
  [STATE(2571)] = {
    [sym_heredoc_content] = STATE(2571),
    [anon_sym_RBRACE] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_RPAREN] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_and] = ACTIONS(3881),
    [anon_sym_or] = ACTIONS(3881),
    [anon_sym_xor] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_RBRACK] = ACTIONS(3881),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2572)] = {
    [sym_heredoc_content] = STATE(2572),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4912),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3415),
  },
  [STATE(2573)] = {
    [sym_heredoc_content] = STATE(2573),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_RPAREN] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_and] = ACTIONS(3649),
    [anon_sym_or] = ACTIONS(3649),
    [anon_sym_xor] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(4586),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3649),
  },
  [STATE(2574)] = {
    [sym_heredoc_content] = STATE(2574),
    [sym__identifier] = ACTIONS(3331),
    [anon_sym_RBRACE] = ACTIONS(3333),
    [anon_sym_COLON] = ACTIONS(3331),
    [anon_sym_EQ] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_COMMA] = ACTIONS(3333),
    [anon_sym_RPAREN] = ACTIONS(3333),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3333),
    [anon_sym_DASH_GT] = ACTIONS(3333),
    [anon_sym_STAR] = ACTIONS(3331),
    [anon_sym_PERCENT] = ACTIONS(3331),
    [anon_sym_GT] = ACTIONS(3331),
    [anon_sym_LT_LT] = ACTIONS(3331),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3333),
    [anon_sym_PLUS_EQ] = ACTIONS(3333),
    [anon_sym_DASH_EQ] = ACTIONS(3333),
    [anon_sym_DOT_EQ] = ACTIONS(3333),
    [anon_sym_STAR_EQ] = ACTIONS(3333),
    [anon_sym_SLASH_EQ] = ACTIONS(3333),
    [anon_sym_PERCENT_EQ] = ACTIONS(3333),
    [anon_sym_x_EQ] = ACTIONS(3333),
    [anon_sym_AMP_EQ] = ACTIONS(3333),
    [anon_sym_PIPE_EQ] = ACTIONS(3333),
    [anon_sym_CARET_EQ] = ACTIONS(3333),
    [anon_sym_LT_LT_EQ] = ACTIONS(3333),
    [anon_sym_GT_GT_EQ] = ACTIONS(3333),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3333),
    [anon_sym_DOT_DOT] = ACTIONS(3331),
    [anon_sym_STAR_STAR] = ACTIONS(3331),
    [anon_sym_PIPE_PIPE] = ACTIONS(3331),
    [anon_sym_SLASH_SLASH] = ACTIONS(3331),
    [anon_sym_CARET_CARET] = ACTIONS(3333),
    [anon_sym_AMP_AMP] = ACTIONS(3331),
    [anon_sym_PIPE] = ACTIONS(3331),
    [anon_sym_CARET] = ACTIONS(3331),
    [anon_sym_AMP] = ACTIONS(3331),
    [anon_sym_GT_GT2] = ACTIONS(3331),
    [anon_sym_PLUS] = ACTIONS(3331),
    [anon_sym_DASH] = ACTIONS(3331),
    [anon_sym_DOT] = ACTIONS(3331),
    [anon_sym_SLASH] = ACTIONS(3331),
    [anon_sym_x] = ACTIONS(3331),
    [anon_sym_EQ_TILDE] = ACTIONS(3333),
    [anon_sym_BANG_TILDE] = ACTIONS(3333),
    [anon_sym_EQ_EQ] = ACTIONS(3331),
    [anon_sym_BANG_EQ] = ACTIONS(3333),
    [anon_sym_eq] = ACTIONS(3331),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3333),
    [anon_sym_equ] = ACTIONS(3331),
    [anon_sym_eqr] = ACTIONS(3331),
    [anon_sym_ne] = ACTIONS(3331),
    [anon_sym_LT_EQ_GT] = ACTIONS(3333),
    [anon_sym_cmp] = ACTIONS(3331),
    [anon_sym_TILDE_TILDE] = ACTIONS(3333),
    [anon_sym_LT] = ACTIONS(3331),
    [anon_sym_LT_EQ] = ACTIONS(3331),
    [anon_sym_GT_EQ] = ACTIONS(3333),
    [anon_sym_lt] = ACTIONS(3331),
    [anon_sym_le] = ACTIONS(3331),
    [anon_sym_ge] = ACTIONS(3331),
    [anon_sym_gt] = ACTIONS(3331),
    [anon_sym_isa] = ACTIONS(3331),
    [anon_sym_PLUS_PLUS] = ACTIONS(3333),
    [anon_sym_DASH_DASH] = ACTIONS(3333),
    [anon_sym_QMARK] = ACTIONS(3333),
    [anon_sym_EQ_GT] = ACTIONS(3333),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3331),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(3333),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2575)] = {
    [sym_heredoc_content] = STATE(2575),
    [anon_sym_RBRACE] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_RPAREN] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_and] = ACTIONS(4063),
    [anon_sym_or] = ACTIONS(4063),
    [anon_sym_xor] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_RBRACK] = ACTIONS(4063),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2576)] = {
    [sym_heredoc_content] = STATE(2576),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2577)] = {
    [sym_heredoc_content] = STATE(2577),
    [anon_sym_RBRACE] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_and] = ACTIONS(3441),
    [anon_sym_or] = ACTIONS(3441),
    [anon_sym_xor] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4914),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3441),
  },
  [STATE(2578)] = {
    [sym_heredoc_content] = STATE(2578),
    [anon_sym_COLON] = ACTIONS(4916),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_and] = ACTIONS(3689),
    [anon_sym_or] = ACTIONS(3689),
    [anon_sym_xor] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_RBRACK] = ACTIONS(3689),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3689),
  },
  [STATE(2579)] = {
    [sym_heredoc_content] = STATE(2579),
    [anon_sym_COLON] = ACTIONS(3562),
    [anon_sym_DOLLAR] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_AT] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_RPAREN] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(4918),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3564),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_PERCENT2] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2580)] = {
    [sym_heredoc_content] = STATE(2580),
    [anon_sym_RBRACE] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_RPAREN] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_and] = ACTIONS(3877),
    [anon_sym_or] = ACTIONS(3877),
    [anon_sym_xor] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_RBRACK] = ACTIONS(3877),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2581)] = {
    [sym_heredoc_content] = STATE(2581),
    [sym__identifier] = ACTIONS(3335),
    [anon_sym_RBRACE] = ACTIONS(3337),
    [anon_sym_COLON] = ACTIONS(3335),
    [anon_sym_EQ] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_COMMA] = ACTIONS(3337),
    [anon_sym_RPAREN] = ACTIONS(3337),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3337),
    [anon_sym_DASH_GT] = ACTIONS(3337),
    [anon_sym_STAR] = ACTIONS(3335),
    [anon_sym_PERCENT] = ACTIONS(3335),
    [anon_sym_GT] = ACTIONS(3335),
    [anon_sym_LT_LT] = ACTIONS(3335),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3337),
    [anon_sym_PLUS_EQ] = ACTIONS(3337),
    [anon_sym_DASH_EQ] = ACTIONS(3337),
    [anon_sym_DOT_EQ] = ACTIONS(3337),
    [anon_sym_STAR_EQ] = ACTIONS(3337),
    [anon_sym_SLASH_EQ] = ACTIONS(3337),
    [anon_sym_PERCENT_EQ] = ACTIONS(3337),
    [anon_sym_x_EQ] = ACTIONS(3337),
    [anon_sym_AMP_EQ] = ACTIONS(3337),
    [anon_sym_PIPE_EQ] = ACTIONS(3337),
    [anon_sym_CARET_EQ] = ACTIONS(3337),
    [anon_sym_LT_LT_EQ] = ACTIONS(3337),
    [anon_sym_GT_GT_EQ] = ACTIONS(3337),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3337),
    [anon_sym_DOT_DOT] = ACTIONS(3335),
    [anon_sym_STAR_STAR] = ACTIONS(3335),
    [anon_sym_PIPE_PIPE] = ACTIONS(3335),
    [anon_sym_SLASH_SLASH] = ACTIONS(3335),
    [anon_sym_CARET_CARET] = ACTIONS(3337),
    [anon_sym_AMP_AMP] = ACTIONS(3335),
    [anon_sym_PIPE] = ACTIONS(3335),
    [anon_sym_CARET] = ACTIONS(3335),
    [anon_sym_AMP] = ACTIONS(3335),
    [anon_sym_GT_GT2] = ACTIONS(3335),
    [anon_sym_PLUS] = ACTIONS(3335),
    [anon_sym_DASH] = ACTIONS(3335),
    [anon_sym_DOT] = ACTIONS(3335),
    [anon_sym_SLASH] = ACTIONS(3335),
    [anon_sym_x] = ACTIONS(3335),
    [anon_sym_EQ_TILDE] = ACTIONS(3337),
    [anon_sym_BANG_TILDE] = ACTIONS(3337),
    [anon_sym_EQ_EQ] = ACTIONS(3335),
    [anon_sym_BANG_EQ] = ACTIONS(3337),
    [anon_sym_eq] = ACTIONS(3335),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3337),
    [anon_sym_equ] = ACTIONS(3335),
    [anon_sym_eqr] = ACTIONS(3335),
    [anon_sym_ne] = ACTIONS(3335),
    [anon_sym_LT_EQ_GT] = ACTIONS(3337),
    [anon_sym_cmp] = ACTIONS(3335),
    [anon_sym_TILDE_TILDE] = ACTIONS(3337),
    [anon_sym_LT] = ACTIONS(3335),
    [anon_sym_LT_EQ] = ACTIONS(3335),
    [anon_sym_GT_EQ] = ACTIONS(3337),
    [anon_sym_lt] = ACTIONS(3335),
    [anon_sym_le] = ACTIONS(3335),
    [anon_sym_ge] = ACTIONS(3335),
    [anon_sym_gt] = ACTIONS(3335),
    [anon_sym_isa] = ACTIONS(3335),
    [anon_sym_PLUS_PLUS] = ACTIONS(3337),
    [anon_sym_DASH_DASH] = ACTIONS(3337),
    [anon_sym_QMARK] = ACTIONS(3337),
    [anon_sym_EQ_GT] = ACTIONS(3337),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3335),
    [sym_pod] = ACTIONS(5),
    [sym__attribute_value_begin] = ACTIONS(4920),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2582)] = {
    [sym_heredoc_content] = STATE(2582),
    [anon_sym_COLON] = ACTIONS(4922),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_and] = ACTIONS(3669),
    [anon_sym_or] = ACTIONS(3669),
    [anon_sym_xor] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_RBRACK] = ACTIONS(3669),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3669),
  },
  [STATE(2583)] = {
    [sym_heredoc_content] = STATE(2583),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_and] = ACTIONS(3675),
    [anon_sym_or] = ACTIONS(3675),
    [anon_sym_xor] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_RBRACK] = ACTIONS(3675),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3675),
  },
  [STATE(2584)] = {
    [sym_heredoc_content] = STATE(2584),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_and] = ACTIONS(3457),
    [anon_sym_or] = ACTIONS(3457),
    [anon_sym_xor] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(4578),
    [anon_sym_RBRACK] = ACTIONS(3457),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3457),
  },
  [STATE(2585)] = {
    [sym_heredoc_content] = STATE(2585),
    [anon_sym_RBRACE] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_RPAREN] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_and] = ACTIONS(3989),
    [anon_sym_or] = ACTIONS(3989),
    [anon_sym_xor] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_RBRACK] = ACTIONS(3989),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2586)] = {
    [sym_heredoc_content] = STATE(2586),
    [anon_sym_RBRACE] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_RPAREN] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_and] = ACTIONS(4115),
    [anon_sym_or] = ACTIONS(4115),
    [anon_sym_xor] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_RBRACK] = ACTIONS(4115),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2587)] = {
    [sym_heredoc_content] = STATE(2587),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4432),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2588)] = {
    [sym_heredoc_content] = STATE(2588),
    [anon_sym_COLON] = ACTIONS(3538),
    [anon_sym_DOLLAR] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_AT] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_RPAREN] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3540),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_PERCENT2] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4924),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2589)] = {
    [sym_heredoc_content] = STATE(2589),
    [anon_sym_COLON] = ACTIONS(3544),
    [anon_sym_DOLLAR] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_AT] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_RPAREN] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3546),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_PERCENT2] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4926),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2590)] = {
    [sym_heredoc_content] = STATE(2590),
    [anon_sym_COLON] = ACTIONS(3578),
    [anon_sym_DOLLAR] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_AT] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_RPAREN] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3580),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_PERCENT2] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4928),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2591)] = {
    [sym_heredoc_content] = STATE(2591),
    [anon_sym_COLON] = ACTIONS(3441),
    [anon_sym_DOLLAR] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_AT] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_RPAREN] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3443),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_PERCENT2] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4930),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2592)] = {
    [sym_heredoc_content] = STATE(2592),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(4910),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_RBRACK] = ACTIONS(3695),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3695),
  },
  [STATE(2593)] = {
    [sym_heredoc_content] = STATE(2593),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_RBRACK] = ACTIONS(2858),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2594)] = {
    [sym_heredoc_content] = STATE(2594),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_RPAREN] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_RBRACK] = ACTIONS(3917),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2595)] = {
    [sym_heredoc_content] = STATE(2595),
    [anon_sym_RBRACE] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_RPAREN] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_and] = ACTIONS(4075),
    [anon_sym_or] = ACTIONS(4075),
    [anon_sym_xor] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_RBRACK] = ACTIONS(4075),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2596)] = {
    [sym_heredoc_content] = STATE(2596),
    [anon_sym_RBRACE] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_RPAREN] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_and] = ACTIONS(3807),
    [anon_sym_or] = ACTIONS(3807),
    [anon_sym_xor] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_RBRACK] = ACTIONS(3807),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2597)] = {
    [sym_heredoc_content] = STATE(2597),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4932),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3415),
  },
  [STATE(2598)] = {
    [sym_heredoc_content] = STATE(2598),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_RBRACE] = ACTIONS(2723),
    [anon_sym_COLON] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_RPAREN] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2599)] = {
    [sym_heredoc_content] = STATE(2599),
    [anon_sym_RBRACE] = ACTIONS(3675),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_and] = ACTIONS(3675),
    [anon_sym_or] = ACTIONS(3675),
    [anon_sym_xor] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3675),
  },
  [STATE(2600)] = {
    [sym_heredoc_content] = STATE(2600),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_RBRACK] = ACTIONS(3819),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2601)] = {
    [sym_heredoc_content] = STATE(2601),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_RPAREN] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_and] = ACTIONS(3679),
    [anon_sym_or] = ACTIONS(3679),
    [anon_sym_xor] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3679),
  },
  [STATE(2602)] = {
    [sym_heredoc_content] = STATE(2602),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(4934),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_RBRACK] = ACTIONS(3819),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(4934),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2603)] = {
    [sym_heredoc_content] = STATE(2603),
    [anon_sym_COLON] = ACTIONS(3584),
    [anon_sym_DOLLAR] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_AT] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_RPAREN] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3586),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_PERCENT2] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4936),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2604)] = {
    [sym_heredoc_content] = STATE(2604),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4434),
    [anon_sym_CARET] = ACTIONS(4434),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2605)] = {
    [sym_heredoc_content] = STATE(2605),
    [aux_sym__term_rightward_repeat1] = STATE(2277),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(4938),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(4938),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3485),
  },
  [STATE(2606)] = {
    [sym_heredoc_content] = STATE(2606),
    [anon_sym_COLON] = ACTIONS(3550),
    [anon_sym_DOLLAR] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_AT] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_RPAREN] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3552),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_PERCENT2] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4941),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2607)] = {
    [sym_heredoc_content] = STATE(2607),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_RBRACK] = ACTIONS(2536),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2608)] = {
    [sym_heredoc_content] = STATE(2608),
    [anon_sym_RBRACE] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_RPAREN] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_and] = ACTIONS(3885),
    [anon_sym_or] = ACTIONS(3885),
    [anon_sym_xor] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_RBRACK] = ACTIONS(3885),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2609)] = {
    [sym_heredoc_content] = STATE(2609),
    [anon_sym_RBRACE] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_RPAREN] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_and] = ACTIONS(3889),
    [anon_sym_or] = ACTIONS(3889),
    [anon_sym_xor] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_RBRACK] = ACTIONS(3889),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2610)] = {
    [sym_heredoc_content] = STATE(2610),
    [anon_sym_RBRACE] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_and] = ACTIONS(3973),
    [anon_sym_or] = ACTIONS(3973),
    [anon_sym_xor] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_RBRACK] = ACTIONS(3973),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2611)] = {
    [sym_heredoc_content] = STATE(2611),
    [anon_sym_RBRACE] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_RPAREN] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_and] = ACTIONS(4041),
    [anon_sym_or] = ACTIONS(4041),
    [anon_sym_xor] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_RBRACK] = ACTIONS(4041),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2612)] = {
    [sym_heredoc_content] = STATE(2612),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4436),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2613)] = {
    [sym_heredoc_content] = STATE(2613),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_RPAREN] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_and] = ACTIONS(3451),
    [anon_sym_or] = ACTIONS(3451),
    [anon_sym_xor] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(4571),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3451),
  },
  [STATE(2614)] = {
    [sym_heredoc_content] = STATE(2614),
    [anon_sym_RBRACE] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_RPAREN] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_and] = ACTIONS(4021),
    [anon_sym_or] = ACTIONS(4021),
    [anon_sym_xor] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_RBRACK] = ACTIONS(4021),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2615)] = {
    [sym_heredoc_content] = STATE(2615),
    [anon_sym_RBRACE] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_RPAREN] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_and] = ACTIONS(4033),
    [anon_sym_or] = ACTIONS(4033),
    [anon_sym_xor] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_RBRACK] = ACTIONS(4033),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2616)] = {
    [sym_heredoc_content] = STATE(2616),
    [anon_sym_RBRACE] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_RPAREN] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_and] = ACTIONS(4055),
    [anon_sym_or] = ACTIONS(4055),
    [anon_sym_xor] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_RBRACK] = ACTIONS(4055),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2617)] = {
    [sym_heredoc_content] = STATE(2617),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_RBRACK] = ACTIONS(2536),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2618)] = {
    [sym_heredoc_content] = STATE(2618),
    [anon_sym_RBRACE] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_and] = ACTIONS(3586),
    [anon_sym_or] = ACTIONS(3586),
    [anon_sym_xor] = ACTIONS(3586),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(4943),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3584),
  },
  [STATE(2619)] = {
    [sym_heredoc_content] = STATE(2619),
    [anon_sym_RBRACE] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_and] = ACTIONS(3550),
    [anon_sym_or] = ACTIONS(3550),
    [anon_sym_xor] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(4945),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3550),
  },
  [STATE(2620)] = {
    [sym_heredoc_content] = STATE(2620),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4416),
    [anon_sym_STAR] = ACTIONS(4418),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4418),
    [anon_sym_GT] = ACTIONS(4420),
    [anon_sym_LT_LT] = ACTIONS(4422),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4426),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4422),
    [anon_sym_PLUS] = ACTIONS(4438),
    [anon_sym_DASH] = ACTIONS(4438),
    [anon_sym_DOT] = ACTIONS(4438),
    [anon_sym_SLASH] = ACTIONS(4418),
    [anon_sym_x] = ACTIONS(4418),
    [anon_sym_EQ_TILDE] = ACTIONS(4440),
    [anon_sym_BANG_TILDE] = ACTIONS(4440),
    [anon_sym_EQ_EQ] = ACTIONS(4442),
    [anon_sym_BANG_EQ] = ACTIONS(4444),
    [anon_sym_eq] = ACTIONS(4442),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4444),
    [anon_sym_equ] = ACTIONS(4444),
    [anon_sym_eqr] = ACTIONS(4444),
    [anon_sym_ne] = ACTIONS(4444),
    [anon_sym_LT_EQ_GT] = ACTIONS(4446),
    [anon_sym_cmp] = ACTIONS(4446),
    [anon_sym_TILDE_TILDE] = ACTIONS(4446),
    [anon_sym_LT] = ACTIONS(4420),
    [anon_sym_LT_EQ] = ACTIONS(4420),
    [anon_sym_GT_EQ] = ACTIONS(4448),
    [anon_sym_lt] = ACTIONS(4448),
    [anon_sym_le] = ACTIONS(4448),
    [anon_sym_ge] = ACTIONS(4448),
    [anon_sym_gt] = ACTIONS(4448),
    [anon_sym_isa] = ACTIONS(4450),
    [anon_sym_PLUS_PLUS] = ACTIONS(4452),
    [anon_sym_DASH_DASH] = ACTIONS(4452),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2621)] = {
    [sym_heredoc_content] = STATE(2621),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_RBRACE] = ACTIONS(2698),
    [anon_sym_COLON] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_RPAREN] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2622)] = {
    [sym_heredoc_content] = STATE(2622),
    [anon_sym_RBRACE] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_RPAREN] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_and] = ACTIONS(3723),
    [anon_sym_or] = ACTIONS(3723),
    [anon_sym_xor] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_RBRACK] = ACTIONS(3723),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2623)] = {
    [sym_heredoc_content] = STATE(2623),
    [anon_sym_COLON] = ACTIONS(3783),
    [anon_sym_DOLLAR] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_AT] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_RPAREN] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3785),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_PERCENT2] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2624)] = {
    [sym_heredoc_content] = STATE(2624),
    [anon_sym_RBRACE] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_and] = ACTIONS(4025),
    [anon_sym_or] = ACTIONS(4025),
    [anon_sym_xor] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4025),
  },
  [STATE(2625)] = {
    [sym_heredoc_content] = STATE(2625),
    [anon_sym_RBRACE] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_and] = ACTIONS(4029),
    [anon_sym_or] = ACTIONS(4029),
    [anon_sym_xor] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4029),
  },
  [STATE(2626)] = {
    [sym_heredoc_content] = STATE(2626),
    [anon_sym_RBRACE] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_and] = ACTIONS(4111),
    [anon_sym_or] = ACTIONS(4111),
    [anon_sym_xor] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4111),
  },
  [STATE(2627)] = {
    [sym_heredoc_content] = STATE(2627),
    [anon_sym_RBRACE] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_and] = ACTIONS(4123),
    [anon_sym_or] = ACTIONS(4123),
    [anon_sym_xor] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4123),
  },
  [STATE(2628)] = {
    [sym_heredoc_content] = STATE(2628),
    [anon_sym_RBRACE] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_and] = ACTIONS(3945),
    [anon_sym_or] = ACTIONS(3945),
    [anon_sym_xor] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3945),
  },
  [STATE(2629)] = {
    [sym_heredoc_content] = STATE(2629),
    [anon_sym_RBRACE] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_and] = ACTIONS(4196),
    [anon_sym_or] = ACTIONS(4196),
    [anon_sym_xor] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4196),
  },
  [STATE(2630)] = {
    [sym_heredoc_content] = STATE(2630),
    [anon_sym_RBRACE] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_and] = ACTIONS(3905),
    [anon_sym_or] = ACTIONS(3905),
    [anon_sym_xor] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3905),
  },
  [STATE(2631)] = {
    [sym_heredoc_content] = STATE(2631),
    [anon_sym_RBRACE] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_and] = ACTIONS(4067),
    [anon_sym_or] = ACTIONS(4067),
    [anon_sym_xor] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4067),
  },
  [STATE(2632)] = {
    [sym_heredoc_content] = STATE(2632),
    [anon_sym_RBRACE] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_and] = ACTIONS(3825),
    [anon_sym_or] = ACTIONS(3825),
    [anon_sym_xor] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3825),
  },
  [STATE(2633)] = {
    [sym_heredoc_content] = STATE(2633),
    [anon_sym_RBRACE] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_and] = ACTIONS(3849),
    [anon_sym_or] = ACTIONS(3849),
    [anon_sym_xor] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3849),
  },
  [STATE(2634)] = {
    [sym_heredoc_content] = STATE(2634),
    [anon_sym_RBRACE] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_and] = ACTIONS(4051),
    [anon_sym_or] = ACTIONS(4051),
    [anon_sym_xor] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4051),
  },
  [STATE(2635)] = {
    [sym_heredoc_content] = STATE(2635),
    [anon_sym_RBRACE] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_and] = ACTIONS(4059),
    [anon_sym_or] = ACTIONS(4059),
    [anon_sym_xor] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4059),
  },
  [STATE(2636)] = {
    [sym_heredoc_content] = STATE(2636),
    [anon_sym_RBRACE] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_and] = ACTIONS(4099),
    [anon_sym_or] = ACTIONS(4099),
    [anon_sym_xor] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4099),
  },
  [STATE(2637)] = {
    [sym_heredoc_content] = STATE(2637),
    [anon_sym_RBRACE] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_and] = ACTIONS(4127),
    [anon_sym_or] = ACTIONS(4127),
    [anon_sym_xor] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4127),
  },
  [STATE(2638)] = {
    [sym_heredoc_content] = STATE(2638),
    [anon_sym_RBRACE] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_and] = ACTIONS(4147),
    [anon_sym_or] = ACTIONS(4147),
    [anon_sym_xor] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4147),
  },
  [STATE(2639)] = {
    [sym_heredoc_content] = STATE(2639),
    [anon_sym_RBRACE] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_and] = ACTIONS(3961),
    [anon_sym_or] = ACTIONS(3961),
    [anon_sym_xor] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3961),
  },
  [STATE(2640)] = {
    [sym_heredoc_content] = STATE(2640),
    [anon_sym_RBRACE] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_and] = ACTIONS(3913),
    [anon_sym_or] = ACTIONS(3913),
    [anon_sym_xor] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3913),
  },
  [STATE(2641)] = {
    [sym_heredoc_content] = STATE(2641),
    [anon_sym_RBRACE] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_and] = ACTIONS(4143),
    [anon_sym_or] = ACTIONS(4143),
    [anon_sym_xor] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4143),
  },
  [STATE(2642)] = {
    [sym_heredoc_content] = STATE(2642),
    [anon_sym_RBRACE] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_and] = ACTIONS(3799),
    [anon_sym_or] = ACTIONS(3799),
    [anon_sym_xor] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3799),
  },
  [STATE(2643)] = {
    [sym_heredoc_content] = STATE(2643),
    [anon_sym_RBRACE] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_and] = ACTIONS(3921),
    [anon_sym_or] = ACTIONS(3921),
    [anon_sym_xor] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3921),
  },
  [STATE(2644)] = {
    [sym_heredoc_content] = STATE(2644),
    [anon_sym_RBRACE] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_and] = ACTIONS(3723),
    [anon_sym_or] = ACTIONS(3723),
    [anon_sym_xor] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3723),
  },
  [STATE(2645)] = {
    [sym_heredoc_content] = STATE(2645),
    [anon_sym_RBRACE] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_and] = ACTIONS(3751),
    [anon_sym_or] = ACTIONS(3751),
    [anon_sym_xor] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3751),
  },
  [STATE(2646)] = {
    [sym_heredoc_content] = STATE(2646),
    [anon_sym_RBRACE] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_and] = ACTIONS(3755),
    [anon_sym_or] = ACTIONS(3755),
    [anon_sym_xor] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3755),
  },
  [STATE(2647)] = {
    [sym_heredoc_content] = STATE(2647),
    [anon_sym_RBRACE] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_and] = ACTIONS(3759),
    [anon_sym_or] = ACTIONS(3759),
    [anon_sym_xor] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3759),
  },
  [STATE(2648)] = {
    [sym_heredoc_content] = STATE(2648),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_RPAREN] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_and] = ACTIONS(3731),
    [anon_sym_or] = ACTIONS(3731),
    [anon_sym_xor] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3731),
  },
  [STATE(2649)] = {
    [sym_heredoc_content] = STATE(2649),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_RPAREN] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4045),
  },
  [STATE(2650)] = {
    [sym_heredoc_content] = STATE(2650),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_RPAREN] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_and] = ACTIONS(4047),
    [anon_sym_or] = ACTIONS(4047),
    [anon_sym_xor] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4047),
  },
  [STATE(2651)] = {
    [sym_heredoc_content] = STATE(2651),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3485),
  },
  [STATE(2652)] = {
    [sym_heredoc_content] = STATE(2652),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3415),
  },
  [STATE(2653)] = {
    [sym_heredoc_content] = STATE(2653),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_RPAREN] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_and] = ACTIONS(4071),
    [anon_sym_or] = ACTIONS(4071),
    [anon_sym_xor] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4071),
  },
  [STATE(2654)] = {
    [sym_heredoc_content] = STATE(2654),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_RPAREN] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_and] = ACTIONS(4135),
    [anon_sym_or] = ACTIONS(4135),
    [anon_sym_xor] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4135),
  },
  [STATE(2655)] = {
    [sym_heredoc_content] = STATE(2655),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_RPAREN] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_and] = ACTIONS(3941),
    [anon_sym_or] = ACTIONS(3941),
    [anon_sym_xor] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3941),
  },
  [STATE(2656)] = {
    [sym_heredoc_content] = STATE(2656),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2711),
  },
  [STATE(2657)] = {
    [sym_heredoc_content] = STATE(2657),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_RPAREN] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3917),
  },
  [STATE(2658)] = {
    [sym_heredoc_content] = STATE(2658),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_RPAREN] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_and] = ACTIONS(3949),
    [anon_sym_or] = ACTIONS(3949),
    [anon_sym_xor] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3949),
  },
  [STATE(2659)] = {
    [sym_heredoc_content] = STATE(2659),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_RPAREN] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_and] = ACTIONS(3957),
    [anon_sym_or] = ACTIONS(3957),
    [anon_sym_xor] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3957),
  },
  [STATE(2660)] = {
    [sym_heredoc_content] = STATE(2660),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_RPAREN] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_and] = ACTIONS(4208),
    [anon_sym_or] = ACTIONS(4208),
    [anon_sym_xor] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4208),
  },
  [STATE(2661)] = {
    [sym_heredoc_content] = STATE(2661),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3977),
  },
  [STATE(2662)] = {
    [sym_heredoc_content] = STATE(2662),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_and] = ACTIONS(3731),
    [anon_sym_or] = ACTIONS(3731),
    [anon_sym_xor] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_RBRACK] = ACTIONS(3731),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3731),
  },
  [STATE(2663)] = {
    [sym_heredoc_content] = STATE(2663),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_RPAREN] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_and] = ACTIONS(4165),
    [anon_sym_or] = ACTIONS(4165),
    [anon_sym_xor] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4165),
  },
  [STATE(2664)] = {
    [sym_heredoc_content] = STATE(2664),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_RPAREN] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_and] = ACTIONS(4204),
    [anon_sym_or] = ACTIONS(4204),
    [anon_sym_xor] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4204),
  },
  [STATE(2665)] = {
    [sym_heredoc_content] = STATE(2665),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_RPAREN] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_and] = ACTIONS(3869),
    [anon_sym_or] = ACTIONS(3869),
    [anon_sym_xor] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3869),
  },
  [STATE(2666)] = {
    [sym_heredoc_content] = STATE(2666),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_RPAREN] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_and] = ACTIONS(4103),
    [anon_sym_or] = ACTIONS(4103),
    [anon_sym_xor] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4103),
  },
  [STATE(2667)] = {
    [sym_heredoc_content] = STATE(2667),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3997),
  },
  [STATE(2668)] = {
    [sym_heredoc_content] = STATE(2668),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_RPAREN] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_and] = ACTIONS(4159),
    [anon_sym_or] = ACTIONS(4159),
    [anon_sym_xor] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4159),
  },
  [STATE(2669)] = {
    [sym_heredoc_content] = STATE(2669),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_RPAREN] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_and] = ACTIONS(4169),
    [anon_sym_or] = ACTIONS(4169),
    [anon_sym_xor] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(4947),
    [anon_sym_DASH_DASH] = ACTIONS(4947),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4169),
  },
  [STATE(2670)] = {
    [sym_heredoc_content] = STATE(2670),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3695),
  },
  [STATE(2671)] = {
    [sym_heredoc_content] = STATE(2671),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_RPAREN] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_and] = ACTIONS(4037),
    [anon_sym_or] = ACTIONS(4037),
    [anon_sym_xor] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4037),
  },
  [STATE(2672)] = {
    [sym_heredoc_content] = STATE(2672),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3909),
  },
  [STATE(2673)] = {
    [sym_heredoc_content] = STATE(2673),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4017),
  },
  [STATE(2674)] = {
    [sym_heredoc_content] = STATE(2674),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4017),
  },
  [STATE(2675)] = {
    [sym_heredoc_content] = STATE(2675),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3803),
  },
  [STATE(2676)] = {
    [sym_heredoc_content] = STATE(2676),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3845),
  },
  [STATE(2677)] = {
    [sym_heredoc_content] = STATE(2677),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_RPAREN] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_and] = ACTIONS(3857),
    [anon_sym_or] = ACTIONS(3857),
    [anon_sym_xor] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3857),
  },
  [STATE(2678)] = {
    [sym_heredoc_content] = STATE(2678),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2858),
  },
  [STATE(2679)] = {
    [sym_heredoc_content] = STATE(2679),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2858),
  },
  [STATE(2680)] = {
    [sym_heredoc_content] = STATE(2680),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3739),
  },
  [STATE(2681)] = {
    [sym_heredoc_content] = STATE(2681),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_RPAREN] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_and] = ACTIONS(3815),
    [anon_sym_or] = ACTIONS(3815),
    [anon_sym_xor] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3815),
  },
  [STATE(2682)] = {
    [sym_heredoc_content] = STATE(2682),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_RPAREN] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_and] = ACTIONS(3775),
    [anon_sym_or] = ACTIONS(3775),
    [anon_sym_xor] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3775),
  },
  [STATE(2683)] = {
    [sym_heredoc_content] = STATE(2683),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3897),
  },
  [STATE(2684)] = {
    [sym_heredoc_content] = STATE(2684),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_RPAREN] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_and] = ACTIONS(3901),
    [anon_sym_or] = ACTIONS(3901),
    [anon_sym_xor] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3901),
  },
  [STATE(2685)] = {
    [sym_heredoc_content] = STATE(2685),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_RPAREN] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_and] = ACTIONS(4119),
    [anon_sym_or] = ACTIONS(4119),
    [anon_sym_xor] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4119),
  },
  [STATE(2686)] = {
    [sym_heredoc_content] = STATE(2686),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_RPAREN] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_and] = ACTIONS(4131),
    [anon_sym_or] = ACTIONS(4131),
    [anon_sym_xor] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4131),
  },
  [STATE(2687)] = {
    [sym_heredoc_content] = STATE(2687),
    [anon_sym_RBRACE] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4151),
  },
  [STATE(2688)] = {
    [sym_heredoc_content] = STATE(2688),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_RPAREN] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_and] = ACTIONS(3881),
    [anon_sym_or] = ACTIONS(3881),
    [anon_sym_xor] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3881),
  },
  [STATE(2689)] = {
    [sym_heredoc_content] = STATE(2689),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_RBRACK] = ACTIONS(2536),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2536),
  },
  [STATE(2690)] = {
    [sym_heredoc_content] = STATE(2690),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_RBRACK] = ACTIONS(4045),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4045),
  },
  [STATE(2691)] = {
    [sym_heredoc_content] = STATE(2691),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_and] = ACTIONS(4047),
    [anon_sym_or] = ACTIONS(4047),
    [anon_sym_xor] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_RBRACK] = ACTIONS(4047),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4047),
  },
  [STATE(2692)] = {
    [sym_heredoc_content] = STATE(2692),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_RBRACK] = ACTIONS(3485),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3485),
  },
  [STATE(2693)] = {
    [sym_heredoc_content] = STATE(2693),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_RBRACK] = ACTIONS(3415),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3415),
  },
  [STATE(2694)] = {
    [sym_heredoc_content] = STATE(2694),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_and] = ACTIONS(4071),
    [anon_sym_or] = ACTIONS(4071),
    [anon_sym_xor] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_RBRACK] = ACTIONS(4071),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4071),
  },
  [STATE(2695)] = {
    [sym_heredoc_content] = STATE(2695),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_RPAREN] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_and] = ACTIONS(4075),
    [anon_sym_or] = ACTIONS(4075),
    [anon_sym_xor] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4075),
  },
  [STATE(2696)] = {
    [sym_heredoc_content] = STATE(2696),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_and] = ACTIONS(4135),
    [anon_sym_or] = ACTIONS(4135),
    [anon_sym_xor] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_RBRACK] = ACTIONS(4135),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4135),
  },
  [STATE(2697)] = {
    [sym_heredoc_content] = STATE(2697),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3819),
  },
  [STATE(2698)] = {
    [sym_heredoc_content] = STATE(2698),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_and] = ACTIONS(3941),
    [anon_sym_or] = ACTIONS(3941),
    [anon_sym_xor] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_RBRACK] = ACTIONS(3941),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3941),
  },
  [STATE(2699)] = {
    [sym_heredoc_content] = STATE(2699),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(4950),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(4950),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3819),
  },
  [STATE(2700)] = {
    [sym_heredoc_content] = STATE(2700),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_RBRACK] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2711),
  },
  [STATE(2701)] = {
    [sym_heredoc_content] = STATE(2701),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_RPAREN] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_and] = ACTIONS(3885),
    [anon_sym_or] = ACTIONS(3885),
    [anon_sym_xor] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3885),
  },
  [STATE(2702)] = {
    [sym_heredoc_content] = STATE(2702),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_RPAREN] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_and] = ACTIONS(3889),
    [anon_sym_or] = ACTIONS(3889),
    [anon_sym_xor] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3889),
  },
  [STATE(2703)] = {
    [sym_heredoc_content] = STATE(2703),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_and] = ACTIONS(3973),
    [anon_sym_or] = ACTIONS(3973),
    [anon_sym_xor] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3973),
  },
  [STATE(2704)] = {
    [sym_heredoc_content] = STATE(2704),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_RBRACK] = ACTIONS(3917),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3917),
  },
  [STATE(2705)] = {
    [sym_heredoc_content] = STATE(2705),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_RPAREN] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_and] = ACTIONS(4021),
    [anon_sym_or] = ACTIONS(4021),
    [anon_sym_xor] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4021),
  },
  [STATE(2706)] = {
    [sym_heredoc_content] = STATE(2706),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_RPAREN] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_and] = ACTIONS(4033),
    [anon_sym_or] = ACTIONS(4033),
    [anon_sym_xor] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4033),
  },
  [STATE(2707)] = {
    [sym_heredoc_content] = STATE(2707),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2536),
  },
  [STATE(2708)] = {
    [sym_heredoc_content] = STATE(2708),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_RPAREN] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_and] = ACTIONS(3747),
    [anon_sym_or] = ACTIONS(3747),
    [anon_sym_xor] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3747),
  },
  [STATE(2709)] = {
    [sym_heredoc_content] = STATE(2709),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_RPAREN] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_and] = ACTIONS(3779),
    [anon_sym_or] = ACTIONS(3779),
    [anon_sym_xor] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3779),
  },
  [STATE(2710)] = {
    [sym_heredoc_content] = STATE(2710),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_RPAREN] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_and] = ACTIONS(3833),
    [anon_sym_or] = ACTIONS(3833),
    [anon_sym_xor] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3833),
  },
  [STATE(2711)] = {
    [sym_heredoc_content] = STATE(2711),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_RPAREN] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_and] = ACTIONS(3841),
    [anon_sym_or] = ACTIONS(3841),
    [anon_sym_xor] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3841),
  },
  [STATE(2712)] = {
    [sym_heredoc_content] = STATE(2712),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_RPAREN] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_and] = ACTIONS(3853),
    [anon_sym_or] = ACTIONS(3853),
    [anon_sym_xor] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3853),
  },
  [STATE(2713)] = {
    [sym_heredoc_content] = STATE(2713),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_and] = ACTIONS(3949),
    [anon_sym_or] = ACTIONS(3949),
    [anon_sym_xor] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_RBRACK] = ACTIONS(3949),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3949),
  },
  [STATE(2714)] = {
    [sym_heredoc_content] = STATE(2714),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_RPAREN] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_and] = ACTIONS(3861),
    [anon_sym_or] = ACTIONS(3861),
    [anon_sym_xor] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3861),
  },
  [STATE(2715)] = {
    [sym_heredoc_content] = STATE(2715),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_RPAREN] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_and] = ACTIONS(3837),
    [anon_sym_or] = ACTIONS(3837),
    [anon_sym_xor] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3837),
  },
  [STATE(2716)] = {
    [sym_heredoc_content] = STATE(2716),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_RPAREN] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_and] = ACTIONS(3993),
    [anon_sym_or] = ACTIONS(3993),
    [anon_sym_xor] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3993),
  },
  [STATE(2717)] = {
    [sym_heredoc_content] = STATE(2717),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_and] = ACTIONS(3957),
    [anon_sym_or] = ACTIONS(3957),
    [anon_sym_xor] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_RBRACK] = ACTIONS(3957),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3957),
  },
  [STATE(2718)] = {
    [sym_heredoc_content] = STATE(2718),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_RPAREN] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_and] = ACTIONS(4005),
    [anon_sym_or] = ACTIONS(4005),
    [anon_sym_xor] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4005),
  },
  [STATE(2719)] = {
    [sym_heredoc_content] = STATE(2719),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_and] = ACTIONS(4208),
    [anon_sym_or] = ACTIONS(4208),
    [anon_sym_xor] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_RBRACK] = ACTIONS(4208),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4208),
  },
  [STATE(2720)] = {
    [sym_heredoc_content] = STATE(2720),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_RPAREN] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_and] = ACTIONS(4087),
    [anon_sym_or] = ACTIONS(4087),
    [anon_sym_xor] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4087),
  },
  [STATE(2721)] = {
    [sym_heredoc_content] = STATE(2721),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3977),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3977),
  },
  [STATE(2722)] = {
    [sym_heredoc_content] = STATE(2722),
    [anon_sym_RBRACE] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4107),
  },
  [STATE(2723)] = {
    [sym_heredoc_content] = STATE(2723),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3590),
  },
  [STATE(2724)] = {
    [sym_heredoc_content] = STATE(2724),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4952),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4954),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2725)] = {
    [sym_heredoc_content] = STATE(2725),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2726)] = {
    [sym_heredoc_content] = STATE(2726),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3735),
  },
  [STATE(2727)] = {
    [sym_heredoc_content] = STATE(2727),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2728)] = {
    [sym_heredoc_content] = STATE(2728),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2729)] = {
    [sym_heredoc_content] = STATE(2729),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2730)] = {
    [sym_heredoc_content] = STATE(2730),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2731)] = {
    [sym_heredoc_content] = STATE(2731),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2732)] = {
    [sym_heredoc_content] = STATE(2732),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2733)] = {
    [sym_heredoc_content] = STATE(2733),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2734)] = {
    [sym_heredoc_content] = STATE(2734),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3719),
  },
  [STATE(2735)] = {
    [sym_heredoc_content] = STATE(2735),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3763),
  },
  [STATE(2736)] = {
    [sym_heredoc_content] = STATE(2736),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4956),
    [anon_sym_cmp] = ACTIONS(4956),
    [anon_sym_TILDE_TILDE] = ACTIONS(4956),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3763),
  },
  [STATE(2737)] = {
    [sym_heredoc_content] = STATE(2737),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4958),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3735),
  },
  [STATE(2738)] = {
    [sym_heredoc_content] = STATE(2738),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_RPAREN] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_and] = ACTIONS(3771),
    [anon_sym_or] = ACTIONS(3771),
    [anon_sym_xor] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3771),
  },
  [STATE(2739)] = {
    [sym_heredoc_content] = STATE(2739),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_RPAREN] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_and] = ACTIONS(3783),
    [anon_sym_or] = ACTIONS(3783),
    [anon_sym_xor] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3783),
  },
  [STATE(2740)] = {
    [sym_heredoc_content] = STATE(2740),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_RPAREN] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_and] = ACTIONS(3787),
    [anon_sym_or] = ACTIONS(3787),
    [anon_sym_xor] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3787),
  },
  [STATE(2741)] = {
    [sym_heredoc_content] = STATE(2741),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_RPAREN] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_and] = ACTIONS(3791),
    [anon_sym_or] = ACTIONS(3791),
    [anon_sym_xor] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3791),
  },
  [STATE(2742)] = {
    [sym_heredoc_content] = STATE(2742),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_RPAREN] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_and] = ACTIONS(3795),
    [anon_sym_or] = ACTIONS(3795),
    [anon_sym_xor] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3795),
  },
  [STATE(2743)] = {
    [sym_heredoc_content] = STATE(2743),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_RBRACE] = ACTIONS(3411),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_RPAREN] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2744)] = {
    [sym_heredoc_content] = STATE(2744),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_RPAREN] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_and] = ACTIONS(3925),
    [anon_sym_or] = ACTIONS(3925),
    [anon_sym_xor] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3925),
  },
  [STATE(2745)] = {
    [sym_heredoc_content] = STATE(2745),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_RPAREN] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_and] = ACTIONS(3929),
    [anon_sym_or] = ACTIONS(3929),
    [anon_sym_xor] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3929),
  },
  [STATE(2746)] = {
    [sym_heredoc_content] = STATE(2746),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_RPAREN] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_and] = ACTIONS(3933),
    [anon_sym_or] = ACTIONS(3933),
    [anon_sym_xor] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3933),
  },
  [STATE(2747)] = {
    [sym_heredoc_content] = STATE(2747),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_RPAREN] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_and] = ACTIONS(3937),
    [anon_sym_or] = ACTIONS(3937),
    [anon_sym_xor] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3937),
  },
  [STATE(2748)] = {
    [sym_heredoc_content] = STATE(2748),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_and] = ACTIONS(3953),
    [anon_sym_or] = ACTIONS(3953),
    [anon_sym_xor] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3953),
  },
  [STATE(2749)] = {
    [sym_heredoc_content] = STATE(2749),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_and] = ACTIONS(3965),
    [anon_sym_or] = ACTIONS(3965),
    [anon_sym_xor] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3965),
  },
  [STATE(2750)] = {
    [sym_heredoc_content] = STATE(2750),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_RPAREN] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_and] = ACTIONS(4155),
    [anon_sym_or] = ACTIONS(4155),
    [anon_sym_xor] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4155),
  },
  [STATE(2751)] = {
    [sym_heredoc_content] = STATE(2751),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_RPAREN] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_and] = ACTIONS(4176),
    [anon_sym_or] = ACTIONS(4176),
    [anon_sym_xor] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4176),
  },
  [STATE(2752)] = {
    [sym_heredoc_content] = STATE(2752),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_RPAREN] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_and] = ACTIONS(4180),
    [anon_sym_or] = ACTIONS(4180),
    [anon_sym_xor] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4180),
  },
  [STATE(2753)] = {
    [sym_heredoc_content] = STATE(2753),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_RPAREN] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_and] = ACTIONS(4188),
    [anon_sym_or] = ACTIONS(4188),
    [anon_sym_xor] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4188),
  },
  [STATE(2754)] = {
    [sym_heredoc_content] = STATE(2754),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_RPAREN] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_and] = ACTIONS(4192),
    [anon_sym_or] = ACTIONS(4192),
    [anon_sym_xor] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4192),
  },
  [STATE(2755)] = {
    [sym_heredoc_content] = STATE(2755),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_RPAREN] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_and] = ACTIONS(4200),
    [anon_sym_or] = ACTIONS(4200),
    [anon_sym_xor] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4200),
  },
  [STATE(2756)] = {
    [sym_heredoc_content] = STATE(2756),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_and] = ACTIONS(4165),
    [anon_sym_or] = ACTIONS(4165),
    [anon_sym_xor] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_RBRACK] = ACTIONS(4165),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4165),
  },
  [STATE(2757)] = {
    [sym_heredoc_content] = STATE(2757),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_and] = ACTIONS(4204),
    [anon_sym_or] = ACTIONS(4204),
    [anon_sym_xor] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_RBRACK] = ACTIONS(4204),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4204),
  },
  [STATE(2758)] = {
    [sym_heredoc_content] = STATE(2758),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_RPAREN] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_and] = ACTIONS(3969),
    [anon_sym_or] = ACTIONS(3969),
    [anon_sym_xor] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3969),
  },
  [STATE(2759)] = {
    [sym_heredoc_content] = STATE(2759),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_and] = ACTIONS(3869),
    [anon_sym_or] = ACTIONS(3869),
    [anon_sym_xor] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_RBRACK] = ACTIONS(3869),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3869),
  },
  [STATE(2760)] = {
    [sym_heredoc_content] = STATE(2760),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_and] = ACTIONS(4103),
    [anon_sym_or] = ACTIONS(4103),
    [anon_sym_xor] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_RBRACK] = ACTIONS(4103),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4103),
  },
  [STATE(2761)] = {
    [sym_heredoc_content] = STATE(2761),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_RBRACK] = ACTIONS(3997),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3997),
  },
  [STATE(2762)] = {
    [sym_heredoc_content] = STATE(2762),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_RPAREN] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_and] = ACTIONS(4079),
    [anon_sym_or] = ACTIONS(4079),
    [anon_sym_xor] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4079),
  },
  [STATE(2763)] = {
    [sym_heredoc_content] = STATE(2763),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_RPAREN] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_and] = ACTIONS(4083),
    [anon_sym_or] = ACTIONS(4083),
    [anon_sym_xor] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4083),
  },
  [STATE(2764)] = {
    [sym_heredoc_content] = STATE(2764),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_RPAREN] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_and] = ACTIONS(4091),
    [anon_sym_or] = ACTIONS(4091),
    [anon_sym_xor] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4091),
  },
  [STATE(2765)] = {
    [sym_heredoc_content] = STATE(2765),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_RPAREN] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_and] = ACTIONS(3715),
    [anon_sym_or] = ACTIONS(3715),
    [anon_sym_xor] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3715),
  },
  [STATE(2766)] = {
    [sym_heredoc_content] = STATE(2766),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_and] = ACTIONS(4159),
    [anon_sym_or] = ACTIONS(4159),
    [anon_sym_xor] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_RBRACK] = ACTIONS(4159),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4159),
  },
  [STATE(2767)] = {
    [sym_heredoc_content] = STATE(2767),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_RPAREN] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_and] = ACTIONS(3893),
    [anon_sym_or] = ACTIONS(3893),
    [anon_sym_xor] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3893),
  },
  [STATE(2768)] = {
    [sym_heredoc_content] = STATE(2768),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_RPAREN] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_and] = ACTIONS(4063),
    [anon_sym_or] = ACTIONS(4063),
    [anon_sym_xor] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4063),
  },
  [STATE(2769)] = {
    [sym_heredoc_content] = STATE(2769),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_RPAREN] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_and] = ACTIONS(3877),
    [anon_sym_or] = ACTIONS(3877),
    [anon_sym_xor] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3877),
  },
  [STATE(2770)] = {
    [sym_heredoc_content] = STATE(2770),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_RPAREN] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_and] = ACTIONS(3989),
    [anon_sym_or] = ACTIONS(3989),
    [anon_sym_xor] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3989),
  },
  [STATE(2771)] = {
    [sym_heredoc_content] = STATE(2771),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_RPAREN] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_and] = ACTIONS(4115),
    [anon_sym_or] = ACTIONS(4115),
    [anon_sym_xor] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4115),
  },
  [STATE(2772)] = {
    [sym_heredoc_content] = STATE(2772),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_and] = ACTIONS(4169),
    [anon_sym_or] = ACTIONS(4169),
    [anon_sym_xor] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_RBRACK] = ACTIONS(4169),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(4960),
    [anon_sym_DASH_DASH] = ACTIONS(4960),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4169),
  },
  [STATE(2773)] = {
    [sym_heredoc_content] = STATE(2773),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_RPAREN] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_and] = ACTIONS(3807),
    [anon_sym_or] = ACTIONS(3807),
    [anon_sym_xor] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3807),
  },
  [STATE(2774)] = {
    [sym_heredoc_content] = STATE(2774),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_RBRACK] = ACTIONS(3695),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3695),
  },
  [STATE(2775)] = {
    [sym_heredoc_content] = STATE(2775),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_and] = ACTIONS(4037),
    [anon_sym_or] = ACTIONS(4037),
    [anon_sym_xor] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_RBRACK] = ACTIONS(4037),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4037),
  },
  [STATE(2776)] = {
    [sym_heredoc_content] = STATE(2776),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_RPAREN] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_and] = ACTIONS(4041),
    [anon_sym_or] = ACTIONS(4041),
    [anon_sym_xor] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4041),
  },
  [STATE(2777)] = {
    [sym_heredoc_content] = STATE(2777),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3909),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3909),
  },
  [STATE(2778)] = {
    [sym_heredoc_content] = STATE(2778),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_RPAREN] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_and] = ACTIONS(4055),
    [anon_sym_or] = ACTIONS(4055),
    [anon_sym_xor] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4055),
  },
  [STATE(2779)] = {
    [sym_heredoc_content] = STATE(2779),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_RBRACK] = ACTIONS(4017),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4017),
  },
  [STATE(2780)] = {
    [sym_heredoc_content] = STATE(2780),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_RBRACE] = ACTIONS(3304),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_RPAREN] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2781)] = {
    [sym_heredoc_content] = STATE(2781),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(4017),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4017),
  },
  [STATE(2782)] = {
    [sym_heredoc_content] = STATE(2782),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3803),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3803),
  },
  [STATE(2783)] = {
    [sym_heredoc_content] = STATE(2783),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3845),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3845),
  },
  [STATE(2784)] = {
    [sym_heredoc_content] = STATE(2784),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_RPAREN] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_and] = ACTIONS(3865),
    [anon_sym_or] = ACTIONS(3865),
    [anon_sym_xor] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3865),
  },
  [STATE(2785)] = {
    [sym_heredoc_content] = STATE(2785),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_RPAREN] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_and] = ACTIONS(3829),
    [anon_sym_or] = ACTIONS(3829),
    [anon_sym_xor] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3829),
  },
  [STATE(2786)] = {
    [sym_heredoc_content] = STATE(2786),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_and] = ACTIONS(3857),
    [anon_sym_or] = ACTIONS(3857),
    [anon_sym_xor] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_RBRACK] = ACTIONS(3857),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3857),
  },
  [STATE(2787)] = {
    [sym_heredoc_content] = STATE(2787),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_RPAREN] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_and] = ACTIONS(4095),
    [anon_sym_or] = ACTIONS(4095),
    [anon_sym_xor] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4095),
  },
  [STATE(2788)] = {
    [sym_heredoc_content] = STATE(2788),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_RPAREN] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_and] = ACTIONS(4184),
    [anon_sym_or] = ACTIONS(4184),
    [anon_sym_xor] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4184),
  },
  [STATE(2789)] = {
    [sym_heredoc_content] = STATE(2789),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_RPAREN] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_and] = ACTIONS(3727),
    [anon_sym_or] = ACTIONS(3727),
    [anon_sym_xor] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3727),
  },
  [STATE(2790)] = {
    [sym_heredoc_content] = STATE(2790),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_RPAREN] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_and] = ACTIONS(3743),
    [anon_sym_or] = ACTIONS(3743),
    [anon_sym_xor] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3743),
  },
  [STATE(2791)] = {
    [sym_heredoc_content] = STATE(2791),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_RBRACK] = ACTIONS(2858),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2858),
  },
  [STATE(2792)] = {
    [sym_heredoc_content] = STATE(2792),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_RPAREN] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_and] = ACTIONS(3811),
    [anon_sym_or] = ACTIONS(3811),
    [anon_sym_xor] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3811),
  },
  [STATE(2793)] = {
    [sym_heredoc_content] = STATE(2793),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3739),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3739),
  },
  [STATE(2794)] = {
    [sym_heredoc_content] = STATE(2794),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_and] = ACTIONS(3815),
    [anon_sym_or] = ACTIONS(3815),
    [anon_sym_xor] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_RBRACK] = ACTIONS(3815),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3815),
  },
  [STATE(2795)] = {
    [sym_heredoc_content] = STATE(2795),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_RPAREN] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_and] = ACTIONS(3873),
    [anon_sym_or] = ACTIONS(3873),
    [anon_sym_xor] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3873),
  },
  [STATE(2796)] = {
    [sym_heredoc_content] = STATE(2796),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_RPAREN] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_and] = ACTIONS(3981),
    [anon_sym_or] = ACTIONS(3981),
    [anon_sym_xor] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3981),
  },
  [STATE(2797)] = {
    [sym_heredoc_content] = STATE(2797),
    [anon_sym_RBRACE] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3985),
  },
  [STATE(2798)] = {
    [sym_heredoc_content] = STATE(2798),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_RPAREN] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_and] = ACTIONS(4001),
    [anon_sym_or] = ACTIONS(4001),
    [anon_sym_xor] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4001),
  },
  [STATE(2799)] = {
    [sym_heredoc_content] = STATE(2799),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_RPAREN] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_and] = ACTIONS(4009),
    [anon_sym_or] = ACTIONS(4009),
    [anon_sym_xor] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4009),
  },
  [STATE(2800)] = {
    [sym_heredoc_content] = STATE(2800),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_RPAREN] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_and] = ACTIONS(4013),
    [anon_sym_or] = ACTIONS(4013),
    [anon_sym_xor] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4013),
  },
  [STATE(2801)] = {
    [sym_heredoc_content] = STATE(2801),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_RPAREN] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_and] = ACTIONS(4025),
    [anon_sym_or] = ACTIONS(4025),
    [anon_sym_xor] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4025),
  },
  [STATE(2802)] = {
    [sym_heredoc_content] = STATE(2802),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_RPAREN] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_and] = ACTIONS(4029),
    [anon_sym_or] = ACTIONS(4029),
    [anon_sym_xor] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4029),
  },
  [STATE(2803)] = {
    [sym_heredoc_content] = STATE(2803),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_and] = ACTIONS(3775),
    [anon_sym_or] = ACTIONS(3775),
    [anon_sym_xor] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_RBRACK] = ACTIONS(3775),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3775),
  },
  [STATE(2804)] = {
    [sym_heredoc_content] = STATE(2804),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_RPAREN] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_and] = ACTIONS(4111),
    [anon_sym_or] = ACTIONS(4111),
    [anon_sym_xor] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4111),
  },
  [STATE(2805)] = {
    [sym_heredoc_content] = STATE(2805),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_RPAREN] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_and] = ACTIONS(4123),
    [anon_sym_or] = ACTIONS(4123),
    [anon_sym_xor] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4123),
  },
  [STATE(2806)] = {
    [sym_heredoc_content] = STATE(2806),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_RPAREN] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_and] = ACTIONS(3945),
    [anon_sym_or] = ACTIONS(3945),
    [anon_sym_xor] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3945),
  },
  [STATE(2807)] = {
    [sym_heredoc_content] = STATE(2807),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_RPAREN] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_and] = ACTIONS(4196),
    [anon_sym_or] = ACTIONS(4196),
    [anon_sym_xor] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4196),
  },
  [STATE(2808)] = {
    [sym_heredoc_content] = STATE(2808),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_RPAREN] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_and] = ACTIONS(3905),
    [anon_sym_or] = ACTIONS(3905),
    [anon_sym_xor] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3905),
  },
  [STATE(2809)] = {
    [sym_heredoc_content] = STATE(2809),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_RPAREN] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_and] = ACTIONS(4067),
    [anon_sym_or] = ACTIONS(4067),
    [anon_sym_xor] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4067),
  },
  [STATE(2810)] = {
    [sym_heredoc_content] = STATE(2810),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_RPAREN] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_and] = ACTIONS(3825),
    [anon_sym_or] = ACTIONS(3825),
    [anon_sym_xor] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3825),
  },
  [STATE(2811)] = {
    [sym_heredoc_content] = STATE(2811),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_RPAREN] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_and] = ACTIONS(3849),
    [anon_sym_or] = ACTIONS(3849),
    [anon_sym_xor] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3849),
  },
  [STATE(2812)] = {
    [sym_heredoc_content] = STATE(2812),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_RPAREN] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_and] = ACTIONS(4051),
    [anon_sym_or] = ACTIONS(4051),
    [anon_sym_xor] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4051),
  },
  [STATE(2813)] = {
    [sym_heredoc_content] = STATE(2813),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_RPAREN] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_and] = ACTIONS(4059),
    [anon_sym_or] = ACTIONS(4059),
    [anon_sym_xor] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4059),
  },
  [STATE(2814)] = {
    [sym_heredoc_content] = STATE(2814),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3897),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3897),
  },
  [STATE(2815)] = {
    [sym_heredoc_content] = STATE(2815),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_RPAREN] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_and] = ACTIONS(4099),
    [anon_sym_or] = ACTIONS(4099),
    [anon_sym_xor] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4099),
  },
  [STATE(2816)] = {
    [sym_heredoc_content] = STATE(2816),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_RPAREN] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_and] = ACTIONS(4127),
    [anon_sym_or] = ACTIONS(4127),
    [anon_sym_xor] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4127),
  },
  [STATE(2817)] = {
    [sym_heredoc_content] = STATE(2817),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_RPAREN] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_and] = ACTIONS(4147),
    [anon_sym_or] = ACTIONS(4147),
    [anon_sym_xor] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4147),
  },
  [STATE(2818)] = {
    [sym_heredoc_content] = STATE(2818),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_RPAREN] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_and] = ACTIONS(3961),
    [anon_sym_or] = ACTIONS(3961),
    [anon_sym_xor] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3961),
  },
  [STATE(2819)] = {
    [sym_heredoc_content] = STATE(2819),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_RPAREN] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_and] = ACTIONS(3913),
    [anon_sym_or] = ACTIONS(3913),
    [anon_sym_xor] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3913),
  },
  [STATE(2820)] = {
    [sym_heredoc_content] = STATE(2820),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_RPAREN] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_and] = ACTIONS(4143),
    [anon_sym_or] = ACTIONS(4143),
    [anon_sym_xor] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4143),
  },
  [STATE(2821)] = {
    [sym_heredoc_content] = STATE(2821),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_RPAREN] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_and] = ACTIONS(3799),
    [anon_sym_or] = ACTIONS(3799),
    [anon_sym_xor] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3799),
  },
  [STATE(2822)] = {
    [sym_heredoc_content] = STATE(2822),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_RPAREN] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_and] = ACTIONS(3921),
    [anon_sym_or] = ACTIONS(3921),
    [anon_sym_xor] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3921),
  },
  [STATE(2823)] = {
    [sym_heredoc_content] = STATE(2823),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_RBRACE] = ACTIONS(3403),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_RPAREN] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2824)] = {
    [sym_heredoc_content] = STATE(2824),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_RPAREN] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_and] = ACTIONS(3723),
    [anon_sym_or] = ACTIONS(3723),
    [anon_sym_xor] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3723),
  },
  [STATE(2825)] = {
    [sym_heredoc_content] = STATE(2825),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_RPAREN] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_and] = ACTIONS(3751),
    [anon_sym_or] = ACTIONS(3751),
    [anon_sym_xor] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3751),
  },
  [STATE(2826)] = {
    [sym_heredoc_content] = STATE(2826),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_RPAREN] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_and] = ACTIONS(3755),
    [anon_sym_or] = ACTIONS(3755),
    [anon_sym_xor] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3755),
  },
  [STATE(2827)] = {
    [sym_heredoc_content] = STATE(2827),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_RPAREN] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_and] = ACTIONS(3759),
    [anon_sym_or] = ACTIONS(3759),
    [anon_sym_xor] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3759),
  },
  [STATE(2828)] = {
    [sym_heredoc_content] = STATE(2828),
    [anon_sym_COLON] = ACTIONS(3731),
    [anon_sym_DOLLAR] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_AT] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_RPAREN] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3733),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_PERCENT2] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2829)] = {
    [sym_heredoc_content] = STATE(2829),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_and] = ACTIONS(3901),
    [anon_sym_or] = ACTIONS(3901),
    [anon_sym_xor] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_RBRACK] = ACTIONS(3901),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3901),
  },
  [STATE(2830)] = {
    [sym_heredoc_content] = STATE(2830),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_and] = ACTIONS(4119),
    [anon_sym_or] = ACTIONS(4119),
    [anon_sym_xor] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_RBRACK] = ACTIONS(4119),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4119),
  },
  [STATE(2831)] = {
    [sym_heredoc_content] = STATE(2831),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_and] = ACTIONS(4131),
    [anon_sym_or] = ACTIONS(4131),
    [anon_sym_xor] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_RBRACK] = ACTIONS(4131),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4131),
  },
  [STATE(2832)] = {
    [sym_heredoc_content] = STATE(2832),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(4151),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4151),
  },
  [STATE(2833)] = {
    [sym_heredoc_content] = STATE(2833),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_and] = ACTIONS(3881),
    [anon_sym_or] = ACTIONS(3881),
    [anon_sym_xor] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_RBRACK] = ACTIONS(3881),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3881),
  },
  [STATE(2834)] = {
    [sym_heredoc_content] = STATE(2834),
    [anon_sym_COLON] = ACTIONS(4045),
    [anon_sym_DOLLAR] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_AT] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_RPAREN] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(2856),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_PERCENT2] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2835)] = {
    [sym_heredoc_content] = STATE(2835),
    [anon_sym_COLON] = ACTIONS(4047),
    [anon_sym_DOLLAR] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_AT] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_RPAREN] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4049),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_PERCENT2] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2836)] = {
    [sym_heredoc_content] = STATE(2836),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_DOLLAR] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_AT] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3536),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_PERCENT2] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2837)] = {
    [sym_heredoc_content] = STATE(2837),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_DOLLAR] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_AT] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3417),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_PERCENT2] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2838)] = {
    [sym_heredoc_content] = STATE(2838),
    [anon_sym_COLON] = ACTIONS(4071),
    [anon_sym_DOLLAR] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_AT] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_RPAREN] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4073),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_PERCENT2] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2839)] = {
    [sym_heredoc_content] = STATE(2839),
    [anon_sym_COLON] = ACTIONS(4135),
    [anon_sym_DOLLAR] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_AT] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_RPAREN] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4137),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_PERCENT2] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2840)] = {
    [sym_heredoc_content] = STATE(2840),
    [anon_sym_COLON] = ACTIONS(3941),
    [anon_sym_DOLLAR] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_AT] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_RPAREN] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3943),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_PERCENT2] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2841)] = {
    [sym_heredoc_content] = STATE(2841),
    [anon_sym_COLON] = ACTIONS(2711),
    [anon_sym_DOLLAR] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2702),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_PERCENT2] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2842)] = {
    [sym_heredoc_content] = STATE(2842),
    [anon_sym_COLON] = ACTIONS(3917),
    [anon_sym_DOLLAR] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_AT] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_RPAREN] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3919),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_PERCENT2] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2843)] = {
    [sym_heredoc_content] = STATE(2843),
    [anon_sym_COLON] = ACTIONS(3949),
    [anon_sym_DOLLAR] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_AT] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_RPAREN] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3951),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_PERCENT2] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2844)] = {
    [sym_heredoc_content] = STATE(2844),
    [anon_sym_COLON] = ACTIONS(3957),
    [anon_sym_DOLLAR] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_AT] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_RPAREN] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3959),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_PERCENT2] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2845)] = {
    [sym_heredoc_content] = STATE(2845),
    [anon_sym_COLON] = ACTIONS(4208),
    [anon_sym_DOLLAR] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_AT] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_RPAREN] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4210),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_PERCENT2] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2846)] = {
    [sym_heredoc_content] = STATE(2846),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_and] = ACTIONS(4075),
    [anon_sym_or] = ACTIONS(4075),
    [anon_sym_xor] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_RBRACK] = ACTIONS(4075),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4075),
  },
  [STATE(2847)] = {
    [sym_heredoc_content] = STATE(2847),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_DOLLAR] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_AT] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_RPAREN] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3677),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_PERCENT2] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2848)] = {
    [sym_heredoc_content] = STATE(2848),
    [anon_sym_COLON] = ACTIONS(4165),
    [anon_sym_DOLLAR] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_AT] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_RPAREN] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4167),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_PERCENT2] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2849)] = {
    [sym_heredoc_content] = STATE(2849),
    [anon_sym_COLON] = ACTIONS(4204),
    [anon_sym_DOLLAR] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_AT] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_RPAREN] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4206),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_PERCENT2] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2850)] = {
    [sym_heredoc_content] = STATE(2850),
    [anon_sym_COLON] = ACTIONS(3869),
    [anon_sym_DOLLAR] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_AT] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_RPAREN] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3871),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_PERCENT2] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2851)] = {
    [sym_heredoc_content] = STATE(2851),
    [anon_sym_COLON] = ACTIONS(4103),
    [anon_sym_DOLLAR] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_AT] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_RPAREN] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4105),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_PERCENT2] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2852)] = {
    [sym_heredoc_content] = STATE(2852),
    [anon_sym_COLON] = ACTIONS(4159),
    [anon_sym_DOLLAR] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_AT] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_RPAREN] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4162),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_PERCENT2] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2853)] = {
    [sym_heredoc_content] = STATE(2853),
    [anon_sym_COLON] = ACTIONS(4169),
    [anon_sym_DOLLAR] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_AT] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_RPAREN] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4171),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(4963),
    [anon_sym_DASH_DASH] = ACTIONS(4963),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_PERCENT2] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2854)] = {
    [sym_heredoc_content] = STATE(2854),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_RBRACK] = ACTIONS(3819),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3819),
  },
  [STATE(2855)] = {
    [sym_heredoc_content] = STATE(2855),
    [anon_sym_COLON] = ACTIONS(4037),
    [anon_sym_DOLLAR] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_AT] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_RPAREN] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4039),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_PERCENT2] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2856)] = {
    [sym_heredoc_content] = STATE(2856),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_DOLLAR] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_AT] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4019),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_PERCENT2] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2857)] = {
    [sym_heredoc_content] = STATE(2857),
    [anon_sym_COLON] = ACTIONS(3857),
    [anon_sym_DOLLAR] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_AT] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_RPAREN] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3859),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_PERCENT2] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2858)] = {
    [sym_heredoc_content] = STATE(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2858),
  },
  [STATE(2859)] = {
    [sym_heredoc_content] = STATE(2859),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2858),
  },
  [STATE(2860)] = {
    [sym_heredoc_content] = STATE(2860),
    [anon_sym_COLON] = ACTIONS(3815),
    [anon_sym_DOLLAR] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_AT] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_RPAREN] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3817),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_PERCENT2] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2861)] = {
    [sym_heredoc_content] = STATE(2861),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(4966),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_RBRACK] = ACTIONS(3819),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(4966),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3819),
  },
  [STATE(2862)] = {
    [sym_heredoc_content] = STATE(2862),
    [anon_sym_COLON] = ACTIONS(3775),
    [anon_sym_DOLLAR] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_AT] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_RPAREN] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3777),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_PERCENT2] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2863)] = {
    [sym_heredoc_content] = STATE(2863),
    [anon_sym_COLON] = ACTIONS(3901),
    [anon_sym_DOLLAR] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_AT] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_RPAREN] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3903),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_PERCENT2] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2864)] = {
    [sym_heredoc_content] = STATE(2864),
    [anon_sym_COLON] = ACTIONS(4119),
    [anon_sym_DOLLAR] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_AT] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_RPAREN] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4121),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_PERCENT2] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2865)] = {
    [sym_heredoc_content] = STATE(2865),
    [anon_sym_COLON] = ACTIONS(4131),
    [anon_sym_DOLLAR] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_AT] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_RPAREN] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4133),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_PERCENT2] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2866)] = {
    [sym_heredoc_content] = STATE(2866),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_and] = ACTIONS(4151),
    [anon_sym_or] = ACTIONS(4151),
    [anon_sym_xor] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4151),
  },
  [STATE(2867)] = {
    [sym_heredoc_content] = STATE(2867),
    [anon_sym_COLON] = ACTIONS(3881),
    [anon_sym_DOLLAR] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_AT] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_RPAREN] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3883),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_PERCENT2] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2868)] = {
    [sym_heredoc_content] = STATE(2868),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_and] = ACTIONS(3885),
    [anon_sym_or] = ACTIONS(3885),
    [anon_sym_xor] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_RBRACK] = ACTIONS(3885),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3885),
  },
  [STATE(2869)] = {
    [sym_heredoc_content] = STATE(2869),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_and] = ACTIONS(3889),
    [anon_sym_or] = ACTIONS(3889),
    [anon_sym_xor] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_RBRACK] = ACTIONS(3889),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3889),
  },
  [STATE(2870)] = {
    [sym_heredoc_content] = STATE(2870),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_and] = ACTIONS(3973),
    [anon_sym_or] = ACTIONS(3973),
    [anon_sym_xor] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_RBRACK] = ACTIONS(3973),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3973),
  },
  [STATE(2871)] = {
    [sym_heredoc_content] = STATE(2871),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_and] = ACTIONS(4021),
    [anon_sym_or] = ACTIONS(4021),
    [anon_sym_xor] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_RBRACK] = ACTIONS(4021),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4021),
  },
  [STATE(2872)] = {
    [sym_heredoc_content] = STATE(2872),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_and] = ACTIONS(4033),
    [anon_sym_or] = ACTIONS(4033),
    [anon_sym_xor] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_RBRACK] = ACTIONS(4033),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4033),
  },
  [STATE(2873)] = {
    [sym_heredoc_content] = STATE(2873),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_RBRACK] = ACTIONS(2536),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2536),
  },
  [STATE(2874)] = {
    [sym_heredoc_content] = STATE(2874),
    [anon_sym_COLON] = ACTIONS(4075),
    [anon_sym_DOLLAR] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_AT] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_RPAREN] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4077),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_PERCENT2] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2875)] = {
    [sym_heredoc_content] = STATE(2875),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_DOLLAR] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_AT] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3821),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_PERCENT2] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2876)] = {
    [sym_heredoc_content] = STATE(2876),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_DOLLAR] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_AT] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(4968),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3821),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_PERCENT2] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(4968),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2877)] = {
    [sym_heredoc_content] = STATE(2877),
    [anon_sym_COLON] = ACTIONS(3885),
    [anon_sym_DOLLAR] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_AT] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_RPAREN] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3887),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_PERCENT2] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2878)] = {
    [sym_heredoc_content] = STATE(2878),
    [anon_sym_COLON] = ACTIONS(3889),
    [anon_sym_DOLLAR] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_AT] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_RPAREN] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3891),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_PERCENT2] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2879)] = {
    [sym_heredoc_content] = STATE(2879),
    [anon_sym_COLON] = ACTIONS(3973),
    [anon_sym_DOLLAR] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_AT] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3975),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_PERCENT2] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2880)] = {
    [sym_heredoc_content] = STATE(2880),
    [anon_sym_COLON] = ACTIONS(4021),
    [anon_sym_DOLLAR] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_AT] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_RPAREN] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4023),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_PERCENT2] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2881)] = {
    [sym_heredoc_content] = STATE(2881),
    [anon_sym_COLON] = ACTIONS(4033),
    [anon_sym_DOLLAR] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_AT] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_RPAREN] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4035),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_PERCENT2] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2882)] = {
    [sym_heredoc_content] = STATE(2882),
    [anon_sym_COLON] = ACTIONS(2536),
    [anon_sym_DOLLAR] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_PERCENT2] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2883)] = {
    [sym_heredoc_content] = STATE(2883),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_DOLLAR] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_AT] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_RPAREN] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3667),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_PERCENT2] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2884)] = {
    [sym_heredoc_content] = STATE(2884),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_DOLLAR] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_AT] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_RPAREN] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3599),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_PERCENT2] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2885)] = {
    [sym_heredoc_content] = STATE(2885),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_DOLLAR] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_AT] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_RPAREN] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3613),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_PERCENT2] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2886)] = {
    [sym_heredoc_content] = STATE(2886),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_DOLLAR] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_AT] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_RPAREN] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3681),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_PERCENT2] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2887)] = {
    [sym_heredoc_content] = STATE(2887),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_DOLLAR] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_AT] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_RPAREN] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3449),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_PERCENT2] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2888)] = {
    [sym_heredoc_content] = STATE(2888),
    [anon_sym_COLON] = ACTIONS(3747),
    [anon_sym_DOLLAR] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_AT] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_RPAREN] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3749),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_PERCENT2] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2889)] = {
    [sym_heredoc_content] = STATE(2889),
    [anon_sym_COLON] = ACTIONS(3779),
    [anon_sym_DOLLAR] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_AT] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_RPAREN] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3781),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_PERCENT2] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2890)] = {
    [sym_heredoc_content] = STATE(2890),
    [anon_sym_COLON] = ACTIONS(3833),
    [anon_sym_DOLLAR] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_AT] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_RPAREN] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3835),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_PERCENT2] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2891)] = {
    [sym_heredoc_content] = STATE(2891),
    [anon_sym_COLON] = ACTIONS(3841),
    [anon_sym_DOLLAR] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_AT] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_RPAREN] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3843),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_PERCENT2] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2892)] = {
    [sym_heredoc_content] = STATE(2892),
    [anon_sym_COLON] = ACTIONS(3853),
    [anon_sym_DOLLAR] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_AT] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_RPAREN] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3855),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_PERCENT2] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2893)] = {
    [sym_heredoc_content] = STATE(2893),
    [anon_sym_COLON] = ACTIONS(3861),
    [anon_sym_DOLLAR] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_AT] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_RPAREN] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3863),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_PERCENT2] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2894)] = {
    [sym_heredoc_content] = STATE(2894),
    [anon_sym_COLON] = ACTIONS(3837),
    [anon_sym_DOLLAR] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_AT] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_RPAREN] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3839),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_PERCENT2] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2895)] = {
    [sym_heredoc_content] = STATE(2895),
    [anon_sym_COLON] = ACTIONS(3993),
    [anon_sym_DOLLAR] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_AT] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_RPAREN] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3995),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_PERCENT2] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2896)] = {
    [sym_heredoc_content] = STATE(2896),
    [anon_sym_COLON] = ACTIONS(4005),
    [anon_sym_DOLLAR] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_AT] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_RPAREN] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4007),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_PERCENT2] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2897)] = {
    [sym_heredoc_content] = STATE(2897),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_and] = ACTIONS(3747),
    [anon_sym_or] = ACTIONS(3747),
    [anon_sym_xor] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_RBRACK] = ACTIONS(3747),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3747),
  },
  [STATE(2898)] = {
    [sym_heredoc_content] = STATE(2898),
    [anon_sym_COLON] = ACTIONS(4087),
    [anon_sym_DOLLAR] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_AT] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_RPAREN] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4089),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_PERCENT2] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2899)] = {
    [sym_heredoc_content] = STATE(2899),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_and] = ACTIONS(3779),
    [anon_sym_or] = ACTIONS(3779),
    [anon_sym_xor] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_RBRACK] = ACTIONS(3779),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3779),
  },
  [STATE(2900)] = {
    [sym_heredoc_content] = STATE(2900),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(4107),
  },
  [STATE(2901)] = {
    [sym_heredoc_content] = STATE(2901),
    [anon_sym_EQ] = ACTIONS(4510),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4512),
    [anon_sym_PLUS_EQ] = ACTIONS(4512),
    [anon_sym_DASH_EQ] = ACTIONS(4512),
    [anon_sym_DOT_EQ] = ACTIONS(4512),
    [anon_sym_STAR_EQ] = ACTIONS(4512),
    [anon_sym_SLASH_EQ] = ACTIONS(4512),
    [anon_sym_PERCENT_EQ] = ACTIONS(4512),
    [anon_sym_x_EQ] = ACTIONS(4512),
    [anon_sym_AMP_EQ] = ACTIONS(4512),
    [anon_sym_PIPE_EQ] = ACTIONS(4512),
    [anon_sym_CARET_EQ] = ACTIONS(4512),
    [anon_sym_LT_LT_EQ] = ACTIONS(4512),
    [anon_sym_GT_GT_EQ] = ACTIONS(4512),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4512),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3590),
  },
  [STATE(2902)] = {
    [sym_heredoc_content] = STATE(2902),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4970),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4972),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2903)] = {
    [sym_heredoc_content] = STATE(2903),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2904)] = {
    [sym_heredoc_content] = STATE(2904),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3735),
  },
  [STATE(2905)] = {
    [sym_heredoc_content] = STATE(2905),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2906)] = {
    [sym_heredoc_content] = STATE(2906),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2907)] = {
    [sym_heredoc_content] = STATE(2907),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2908)] = {
    [sym_heredoc_content] = STATE(2908),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2909)] = {
    [sym_heredoc_content] = STATE(2909),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2910)] = {
    [sym_heredoc_content] = STATE(2910),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2911)] = {
    [sym_heredoc_content] = STATE(2911),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2912)] = {
    [sym_heredoc_content] = STATE(2912),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3719),
  },
  [STATE(2913)] = {
    [sym_heredoc_content] = STATE(2913),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3763),
  },
  [STATE(2914)] = {
    [sym_heredoc_content] = STATE(2914),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4974),
    [anon_sym_cmp] = ACTIONS(4974),
    [anon_sym_TILDE_TILDE] = ACTIONS(4974),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3763),
  },
  [STATE(2915)] = {
    [sym_heredoc_content] = STATE(2915),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4976),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3735),
  },
  [STATE(2916)] = {
    [sym_heredoc_content] = STATE(2916),
    [anon_sym_COLON] = ACTIONS(3771),
    [anon_sym_DOLLAR] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_AT] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_RPAREN] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3773),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_PERCENT2] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2917)] = {
    [sym_heredoc_content] = STATE(2917),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_and] = ACTIONS(3833),
    [anon_sym_or] = ACTIONS(3833),
    [anon_sym_xor] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_RBRACK] = ACTIONS(3833),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3833),
  },
  [STATE(2918)] = {
    [sym_heredoc_content] = STATE(2918),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_and] = ACTIONS(3841),
    [anon_sym_or] = ACTIONS(3841),
    [anon_sym_xor] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_RBRACK] = ACTIONS(3841),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3841),
  },
  [STATE(2919)] = {
    [sym_heredoc_content] = STATE(2919),
    [anon_sym_COLON] = ACTIONS(3787),
    [anon_sym_DOLLAR] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_AT] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_RPAREN] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3789),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_PERCENT2] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2920)] = {
    [sym_heredoc_content] = STATE(2920),
    [anon_sym_COLON] = ACTIONS(3791),
    [anon_sym_DOLLAR] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_AT] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_RPAREN] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3793),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_PERCENT2] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2921)] = {
    [sym_heredoc_content] = STATE(2921),
    [anon_sym_COLON] = ACTIONS(3795),
    [anon_sym_DOLLAR] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_AT] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_RPAREN] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3797),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_PERCENT2] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2922)] = {
    [sym_heredoc_content] = STATE(2922),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_and] = ACTIONS(3853),
    [anon_sym_or] = ACTIONS(3853),
    [anon_sym_xor] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_RBRACK] = ACTIONS(3853),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3853),
  },
  [STATE(2923)] = {
    [sym_heredoc_content] = STATE(2923),
    [sym__identifier] = ACTIONS(3409),
    [anon_sym_COLON] = ACTIONS(3409),
    [anon_sym_SEMI] = ACTIONS(3411),
    [anon_sym_EQ] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_COMMA] = ACTIONS(3411),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3411),
    [anon_sym_DASH_GT] = ACTIONS(3411),
    [anon_sym_STAR] = ACTIONS(3409),
    [anon_sym_PERCENT] = ACTIONS(3409),
    [anon_sym_GT] = ACTIONS(3409),
    [anon_sym_LT_LT] = ACTIONS(3409),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3411),
    [anon_sym_PLUS_EQ] = ACTIONS(3411),
    [anon_sym_DASH_EQ] = ACTIONS(3411),
    [anon_sym_DOT_EQ] = ACTIONS(3411),
    [anon_sym_STAR_EQ] = ACTIONS(3411),
    [anon_sym_SLASH_EQ] = ACTIONS(3411),
    [anon_sym_PERCENT_EQ] = ACTIONS(3411),
    [anon_sym_x_EQ] = ACTIONS(3411),
    [anon_sym_AMP_EQ] = ACTIONS(3411),
    [anon_sym_PIPE_EQ] = ACTIONS(3411),
    [anon_sym_CARET_EQ] = ACTIONS(3411),
    [anon_sym_LT_LT_EQ] = ACTIONS(3411),
    [anon_sym_GT_GT_EQ] = ACTIONS(3411),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3411),
    [anon_sym_DOT_DOT] = ACTIONS(3409),
    [anon_sym_STAR_STAR] = ACTIONS(3409),
    [anon_sym_PIPE_PIPE] = ACTIONS(3409),
    [anon_sym_SLASH_SLASH] = ACTIONS(3409),
    [anon_sym_CARET_CARET] = ACTIONS(3411),
    [anon_sym_AMP_AMP] = ACTIONS(3409),
    [anon_sym_PIPE] = ACTIONS(3409),
    [anon_sym_CARET] = ACTIONS(3409),
    [anon_sym_AMP] = ACTIONS(3409),
    [anon_sym_GT_GT2] = ACTIONS(3409),
    [anon_sym_PLUS] = ACTIONS(3409),
    [anon_sym_DASH] = ACTIONS(3409),
    [anon_sym_DOT] = ACTIONS(3409),
    [anon_sym_SLASH] = ACTIONS(3409),
    [anon_sym_x] = ACTIONS(3409),
    [anon_sym_EQ_TILDE] = ACTIONS(3411),
    [anon_sym_BANG_TILDE] = ACTIONS(3411),
    [anon_sym_EQ_EQ] = ACTIONS(3409),
    [anon_sym_BANG_EQ] = ACTIONS(3411),
    [anon_sym_eq] = ACTIONS(3409),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3411),
    [anon_sym_equ] = ACTIONS(3409),
    [anon_sym_eqr] = ACTIONS(3409),
    [anon_sym_ne] = ACTIONS(3409),
    [anon_sym_LT_EQ_GT] = ACTIONS(3411),
    [anon_sym_cmp] = ACTIONS(3409),
    [anon_sym_TILDE_TILDE] = ACTIONS(3411),
    [anon_sym_LT] = ACTIONS(3409),
    [anon_sym_LT_EQ] = ACTIONS(3409),
    [anon_sym_GT_EQ] = ACTIONS(3411),
    [anon_sym_lt] = ACTIONS(3409),
    [anon_sym_le] = ACTIONS(3409),
    [anon_sym_ge] = ACTIONS(3409),
    [anon_sym_gt] = ACTIONS(3409),
    [anon_sym_isa] = ACTIONS(3409),
    [anon_sym_PLUS_PLUS] = ACTIONS(3411),
    [anon_sym_DASH_DASH] = ACTIONS(3411),
    [anon_sym_QMARK] = ACTIONS(3411),
    [anon_sym_EQ_GT] = ACTIONS(3411),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3409),
    [sym__PERLY_SEMICOLON] = ACTIONS(3411),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2924)] = {
    [sym_heredoc_content] = STATE(2924),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_and] = ACTIONS(3861),
    [anon_sym_or] = ACTIONS(3861),
    [anon_sym_xor] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_RBRACK] = ACTIONS(3861),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3861),
  },
  [STATE(2925)] = {
    [sym_heredoc_content] = STATE(2925),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_DOLLAR] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_AT] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_RPAREN] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3629),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_PERCENT2] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2926)] = {
    [sym_heredoc_content] = STATE(2926),
    [anon_sym_COLON] = ACTIONS(3925),
    [anon_sym_DOLLAR] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_AT] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_RPAREN] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3927),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_PERCENT2] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2927)] = {
    [sym_heredoc_content] = STATE(2927),
    [anon_sym_COLON] = ACTIONS(3929),
    [anon_sym_DOLLAR] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_AT] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_RPAREN] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3931),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_PERCENT2] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2928)] = {
    [sym_heredoc_content] = STATE(2928),
    [anon_sym_COLON] = ACTIONS(3933),
    [anon_sym_DOLLAR] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_AT] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_RPAREN] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3935),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_PERCENT2] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2929)] = {
    [sym_heredoc_content] = STATE(2929),
    [anon_sym_COLON] = ACTIONS(3937),
    [anon_sym_DOLLAR] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_AT] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_RPAREN] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3939),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_PERCENT2] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2930)] = {
    [sym_heredoc_content] = STATE(2930),
    [anon_sym_COLON] = ACTIONS(3953),
    [anon_sym_DOLLAR] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_AT] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3955),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_PERCENT2] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2931)] = {
    [sym_heredoc_content] = STATE(2931),
    [anon_sym_COLON] = ACTIONS(3965),
    [anon_sym_DOLLAR] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_AT] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3967),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_PERCENT2] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2932)] = {
    [sym_heredoc_content] = STATE(2932),
    [anon_sym_COLON] = ACTIONS(4155),
    [anon_sym_DOLLAR] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_AT] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_RPAREN] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4157),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_PERCENT2] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2933)] = {
    [sym_heredoc_content] = STATE(2933),
    [anon_sym_COLON] = ACTIONS(4176),
    [anon_sym_DOLLAR] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_AT] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_RPAREN] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4178),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_PERCENT2] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2934)] = {
    [sym_heredoc_content] = STATE(2934),
    [anon_sym_COLON] = ACTIONS(4180),
    [anon_sym_DOLLAR] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_AT] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_RPAREN] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4182),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_PERCENT2] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2935)] = {
    [sym_heredoc_content] = STATE(2935),
    [anon_sym_COLON] = ACTIONS(4188),
    [anon_sym_DOLLAR] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_AT] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_RPAREN] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4190),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_PERCENT2] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2936)] = {
    [sym_heredoc_content] = STATE(2936),
    [anon_sym_COLON] = ACTIONS(4192),
    [anon_sym_DOLLAR] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_AT] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_RPAREN] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4194),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_PERCENT2] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2937)] = {
    [sym_heredoc_content] = STATE(2937),
    [anon_sym_COLON] = ACTIONS(4200),
    [anon_sym_DOLLAR] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_AT] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_RPAREN] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4202),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_PERCENT2] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2938)] = {
    [sym_heredoc_content] = STATE(2938),
    [anon_sym_COLON] = ACTIONS(3969),
    [anon_sym_DOLLAR] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_AT] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_RPAREN] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3971),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_PERCENT2] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2939)] = {
    [sym_heredoc_content] = STATE(2939),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_and] = ACTIONS(3837),
    [anon_sym_or] = ACTIONS(3837),
    [anon_sym_xor] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_RBRACK] = ACTIONS(3837),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3837),
  },
  [STATE(2940)] = {
    [sym_heredoc_content] = STATE(2940),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_and] = ACTIONS(3993),
    [anon_sym_or] = ACTIONS(3993),
    [anon_sym_xor] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_RBRACK] = ACTIONS(3993),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3993),
  },
  [STATE(2941)] = {
    [sym_heredoc_content] = STATE(2941),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_and] = ACTIONS(4005),
    [anon_sym_or] = ACTIONS(4005),
    [anon_sym_xor] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_RBRACK] = ACTIONS(4005),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4005),
  },
  [STATE(2942)] = {
    [sym_heredoc_content] = STATE(2942),
    [anon_sym_COLON] = ACTIONS(4079),
    [anon_sym_DOLLAR] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_AT] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_RPAREN] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4081),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_PERCENT2] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2943)] = {
    [sym_heredoc_content] = STATE(2943),
    [anon_sym_COLON] = ACTIONS(4083),
    [anon_sym_DOLLAR] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_AT] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_RPAREN] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4085),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_PERCENT2] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2944)] = {
    [sym_heredoc_content] = STATE(2944),
    [anon_sym_COLON] = ACTIONS(4091),
    [anon_sym_DOLLAR] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_AT] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_RPAREN] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4093),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_PERCENT2] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2945)] = {
    [sym_heredoc_content] = STATE(2945),
    [anon_sym_COLON] = ACTIONS(3715),
    [anon_sym_DOLLAR] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_AT] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_RPAREN] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3717),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_PERCENT2] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2946)] = {
    [sym_heredoc_content] = STATE(2946),
    [anon_sym_COLON] = ACTIONS(3893),
    [anon_sym_DOLLAR] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_AT] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_RPAREN] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3895),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_PERCENT2] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2947)] = {
    [sym_heredoc_content] = STATE(2947),
    [anon_sym_COLON] = ACTIONS(4063),
    [anon_sym_DOLLAR] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_AT] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_RPAREN] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4065),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_PERCENT2] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2948)] = {
    [sym_heredoc_content] = STATE(2948),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_and] = ACTIONS(4087),
    [anon_sym_or] = ACTIONS(4087),
    [anon_sym_xor] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_RBRACK] = ACTIONS(4087),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4087),
  },
  [STATE(2949)] = {
    [sym_heredoc_content] = STATE(2949),
    [anon_sym_COLON] = ACTIONS(3877),
    [anon_sym_DOLLAR] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_AT] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_RPAREN] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3879),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_PERCENT2] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2950)] = {
    [sym_heredoc_content] = STATE(2950),
    [anon_sym_COLON] = ACTIONS(3989),
    [anon_sym_DOLLAR] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_AT] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_RPAREN] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3991),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_PERCENT2] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2951)] = {
    [sym_heredoc_content] = STATE(2951),
    [anon_sym_COLON] = ACTIONS(4115),
    [anon_sym_DOLLAR] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_AT] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_RPAREN] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4117),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_PERCENT2] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2952)] = {
    [sym_heredoc_content] = STATE(2952),
    [anon_sym_COLON] = ACTIONS(3807),
    [anon_sym_DOLLAR] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_AT] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_RPAREN] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3809),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_PERCENT2] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2953)] = {
    [sym_heredoc_content] = STATE(2953),
    [anon_sym_COLON] = ACTIONS(4041),
    [anon_sym_DOLLAR] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_AT] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_RPAREN] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4043),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_PERCENT2] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2954)] = {
    [sym_heredoc_content] = STATE(2954),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(4107),
    [anon_sym_or] = ACTIONS(4107),
    [anon_sym_xor] = ACTIONS(4107),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(4107),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4107),
  },
  [STATE(2955)] = {
    [sym_heredoc_content] = STATE(2955),
    [anon_sym_COLON] = ACTIONS(4055),
    [anon_sym_DOLLAR] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_AT] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_RPAREN] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4057),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_PERCENT2] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2956)] = {
    [sym_heredoc_content] = STATE(2956),
    [anon_sym_EQ] = ACTIONS(4767),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3590),
    [anon_sym_or] = ACTIONS(3590),
    [anon_sym_xor] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3590),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4769),
    [anon_sym_PLUS_EQ] = ACTIONS(4769),
    [anon_sym_DASH_EQ] = ACTIONS(4769),
    [anon_sym_DOT_EQ] = ACTIONS(4769),
    [anon_sym_STAR_EQ] = ACTIONS(4769),
    [anon_sym_SLASH_EQ] = ACTIONS(4769),
    [anon_sym_PERCENT_EQ] = ACTIONS(4769),
    [anon_sym_x_EQ] = ACTIONS(4769),
    [anon_sym_AMP_EQ] = ACTIONS(4769),
    [anon_sym_PIPE_EQ] = ACTIONS(4769),
    [anon_sym_CARET_EQ] = ACTIONS(4769),
    [anon_sym_LT_LT_EQ] = ACTIONS(4769),
    [anon_sym_GT_GT_EQ] = ACTIONS(4769),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4769),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3590),
  },
  [STATE(2957)] = {
    [sym_heredoc_content] = STATE(2957),
    [sym__identifier] = ACTIONS(3309),
    [anon_sym_COLON] = ACTIONS(3309),
    [anon_sym_SEMI] = ACTIONS(3304),
    [anon_sym_EQ] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_COMMA] = ACTIONS(3304),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3304),
    [anon_sym_DASH_GT] = ACTIONS(3304),
    [anon_sym_STAR] = ACTIONS(3309),
    [anon_sym_PERCENT] = ACTIONS(3309),
    [anon_sym_GT] = ACTIONS(3309),
    [anon_sym_LT_LT] = ACTIONS(3309),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3304),
    [anon_sym_PLUS_EQ] = ACTIONS(3304),
    [anon_sym_DASH_EQ] = ACTIONS(3304),
    [anon_sym_DOT_EQ] = ACTIONS(3304),
    [anon_sym_STAR_EQ] = ACTIONS(3304),
    [anon_sym_SLASH_EQ] = ACTIONS(3304),
    [anon_sym_PERCENT_EQ] = ACTIONS(3304),
    [anon_sym_x_EQ] = ACTIONS(3304),
    [anon_sym_AMP_EQ] = ACTIONS(3304),
    [anon_sym_PIPE_EQ] = ACTIONS(3304),
    [anon_sym_CARET_EQ] = ACTIONS(3304),
    [anon_sym_LT_LT_EQ] = ACTIONS(3304),
    [anon_sym_GT_GT_EQ] = ACTIONS(3304),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3304),
    [anon_sym_DOT_DOT] = ACTIONS(3309),
    [anon_sym_STAR_STAR] = ACTIONS(3309),
    [anon_sym_PIPE_PIPE] = ACTIONS(3309),
    [anon_sym_SLASH_SLASH] = ACTIONS(3309),
    [anon_sym_CARET_CARET] = ACTIONS(3304),
    [anon_sym_AMP_AMP] = ACTIONS(3309),
    [anon_sym_PIPE] = ACTIONS(3309),
    [anon_sym_CARET] = ACTIONS(3309),
    [anon_sym_AMP] = ACTIONS(3309),
    [anon_sym_GT_GT2] = ACTIONS(3309),
    [anon_sym_PLUS] = ACTIONS(3309),
    [anon_sym_DASH] = ACTIONS(3309),
    [anon_sym_DOT] = ACTIONS(3309),
    [anon_sym_SLASH] = ACTIONS(3309),
    [anon_sym_x] = ACTIONS(3309),
    [anon_sym_EQ_TILDE] = ACTIONS(3304),
    [anon_sym_BANG_TILDE] = ACTIONS(3304),
    [anon_sym_EQ_EQ] = ACTIONS(3309),
    [anon_sym_BANG_EQ] = ACTIONS(3304),
    [anon_sym_eq] = ACTIONS(3309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3304),
    [anon_sym_equ] = ACTIONS(3309),
    [anon_sym_eqr] = ACTIONS(3309),
    [anon_sym_ne] = ACTIONS(3309),
    [anon_sym_LT_EQ_GT] = ACTIONS(3304),
    [anon_sym_cmp] = ACTIONS(3309),
    [anon_sym_TILDE_TILDE] = ACTIONS(3304),
    [anon_sym_LT] = ACTIONS(3309),
    [anon_sym_LT_EQ] = ACTIONS(3309),
    [anon_sym_GT_EQ] = ACTIONS(3304),
    [anon_sym_lt] = ACTIONS(3309),
    [anon_sym_le] = ACTIONS(3309),
    [anon_sym_ge] = ACTIONS(3309),
    [anon_sym_gt] = ACTIONS(3309),
    [anon_sym_isa] = ACTIONS(3309),
    [anon_sym_PLUS_PLUS] = ACTIONS(3304),
    [anon_sym_DASH_DASH] = ACTIONS(3304),
    [anon_sym_QMARK] = ACTIONS(3304),
    [anon_sym_EQ_GT] = ACTIONS(3304),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3309),
    [sym__PERLY_SEMICOLON] = ACTIONS(3304),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2958)] = {
    [sym_heredoc_content] = STATE(2958),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4978),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4980),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2959)] = {
    [sym_heredoc_content] = STATE(2959),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_DOLLAR] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_AT] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_RPAREN] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3663),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_PERCENT2] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2960)] = {
    [sym_heredoc_content] = STATE(2960),
    [anon_sym_COLON] = ACTIONS(3865),
    [anon_sym_DOLLAR] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_AT] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_RPAREN] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3867),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_PERCENT2] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2961)] = {
    [sym_heredoc_content] = STATE(2961),
    [anon_sym_COLON] = ACTIONS(3829),
    [anon_sym_DOLLAR] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_AT] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_RPAREN] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3831),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_PERCENT2] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2962)] = {
    [sym_heredoc_content] = STATE(2962),
    [anon_sym_COLON] = ACTIONS(4095),
    [anon_sym_DOLLAR] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_AT] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_RPAREN] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4097),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_PERCENT2] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2963)] = {
    [sym_heredoc_content] = STATE(2963),
    [anon_sym_COLON] = ACTIONS(4184),
    [anon_sym_DOLLAR] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_AT] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_RPAREN] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4186),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_PERCENT2] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2964)] = {
    [sym_heredoc_content] = STATE(2964),
    [anon_sym_COLON] = ACTIONS(3727),
    [anon_sym_DOLLAR] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_AT] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_RPAREN] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3729),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_PERCENT2] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2965)] = {
    [sym_heredoc_content] = STATE(2965),
    [anon_sym_COLON] = ACTIONS(3743),
    [anon_sym_DOLLAR] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_AT] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_RPAREN] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3745),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_PERCENT2] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2966)] = {
    [sym_heredoc_content] = STATE(2966),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2967)] = {
    [sym_heredoc_content] = STATE(2967),
    [anon_sym_COLON] = ACTIONS(3811),
    [anon_sym_DOLLAR] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_AT] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_RPAREN] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3813),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_PERCENT2] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2968)] = {
    [sym_heredoc_content] = STATE(2968),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3735),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3735),
  },
  [STATE(2969)] = {
    [sym_heredoc_content] = STATE(2969),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2970)] = {
    [sym_heredoc_content] = STATE(2970),
    [anon_sym_COLON] = ACTIONS(3873),
    [anon_sym_DOLLAR] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_AT] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_RPAREN] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3875),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_PERCENT2] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2971)] = {
    [sym_heredoc_content] = STATE(2971),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2972)] = {
    [sym_heredoc_content] = STATE(2972),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2973)] = {
    [sym_heredoc_content] = STATE(2973),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2974)] = {
    [sym_heredoc_content] = STATE(2974),
    [anon_sym_COLON] = ACTIONS(3981),
    [anon_sym_DOLLAR] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_AT] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_RPAREN] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3983),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_PERCENT2] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2975)] = {
    [sym_heredoc_content] = STATE(2975),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4517),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(4519),
    [anon_sym_STAR] = ACTIONS(4521),
    [anon_sym_PERCENT] = ACTIONS(4521),
    [anon_sym_GT] = ACTIONS(4523),
    [anon_sym_LT_LT] = ACTIONS(4525),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4527),
    [anon_sym_STAR_STAR] = ACTIONS(4529),
    [anon_sym_PIPE_PIPE] = ACTIONS(4531),
    [anon_sym_SLASH_SLASH] = ACTIONS(4531),
    [anon_sym_CARET_CARET] = ACTIONS(4533),
    [anon_sym_AMP_AMP] = ACTIONS(4535),
    [anon_sym_PIPE] = ACTIONS(4537),
    [anon_sym_CARET] = ACTIONS(4537),
    [anon_sym_AMP] = ACTIONS(4539),
    [anon_sym_GT_GT2] = ACTIONS(4525),
    [anon_sym_PLUS] = ACTIONS(4541),
    [anon_sym_DASH] = ACTIONS(4541),
    [anon_sym_DOT] = ACTIONS(4541),
    [anon_sym_SLASH] = ACTIONS(4521),
    [anon_sym_x] = ACTIONS(4521),
    [anon_sym_EQ_TILDE] = ACTIONS(4543),
    [anon_sym_BANG_TILDE] = ACTIONS(4543),
    [anon_sym_EQ_EQ] = ACTIONS(4545),
    [anon_sym_BANG_EQ] = ACTIONS(4547),
    [anon_sym_eq] = ACTIONS(4545),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4547),
    [anon_sym_equ] = ACTIONS(4547),
    [anon_sym_eqr] = ACTIONS(4547),
    [anon_sym_ne] = ACTIONS(4547),
    [anon_sym_LT_EQ_GT] = ACTIONS(4549),
    [anon_sym_cmp] = ACTIONS(4549),
    [anon_sym_TILDE_TILDE] = ACTIONS(4549),
    [anon_sym_LT] = ACTIONS(4523),
    [anon_sym_LT_EQ] = ACTIONS(4523),
    [anon_sym_GT_EQ] = ACTIONS(4551),
    [anon_sym_lt] = ACTIONS(4551),
    [anon_sym_le] = ACTIONS(4551),
    [anon_sym_ge] = ACTIONS(4551),
    [anon_sym_gt] = ACTIONS(4551),
    [anon_sym_isa] = ACTIONS(4553),
    [anon_sym_PLUS_PLUS] = ACTIONS(4555),
    [anon_sym_DASH_DASH] = ACTIONS(4555),
    [anon_sym_QMARK] = ACTIONS(4557),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(3985),
  },
  [STATE(2976)] = {
    [sym_heredoc_content] = STATE(2976),
    [anon_sym_COLON] = ACTIONS(4001),
    [anon_sym_DOLLAR] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_AT] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_RPAREN] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4003),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_PERCENT2] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2977)] = {
    [sym_heredoc_content] = STATE(2977),
    [anon_sym_COLON] = ACTIONS(4009),
    [anon_sym_DOLLAR] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_AT] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_RPAREN] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4011),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_PERCENT2] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2978)] = {
    [sym_heredoc_content] = STATE(2978),
    [anon_sym_COLON] = ACTIONS(4013),
    [anon_sym_DOLLAR] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_AT] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_RPAREN] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4015),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_PERCENT2] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2979)] = {
    [sym_heredoc_content] = STATE(2979),
    [anon_sym_COLON] = ACTIONS(4025),
    [anon_sym_DOLLAR] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_AT] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_RPAREN] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4027),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_PERCENT2] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2980)] = {
    [sym_heredoc_content] = STATE(2980),
    [anon_sym_COLON] = ACTIONS(4029),
    [anon_sym_DOLLAR] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_AT] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_RPAREN] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4031),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_PERCENT2] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2981)] = {
    [sym_heredoc_content] = STATE(2981),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2982)] = {
    [sym_heredoc_content] = STATE(2982),
    [anon_sym_COLON] = ACTIONS(4111),
    [anon_sym_DOLLAR] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_AT] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_RPAREN] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4113),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_PERCENT2] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2983)] = {
    [sym_heredoc_content] = STATE(2983),
    [anon_sym_COLON] = ACTIONS(4123),
    [anon_sym_DOLLAR] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_AT] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_RPAREN] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4125),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_PERCENT2] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2984)] = {
    [sym_heredoc_content] = STATE(2984),
    [anon_sym_COLON] = ACTIONS(3945),
    [anon_sym_DOLLAR] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_AT] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_RPAREN] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3947),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_PERCENT2] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2985)] = {
    [sym_heredoc_content] = STATE(2985),
    [anon_sym_COLON] = ACTIONS(4196),
    [anon_sym_DOLLAR] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_AT] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_RPAREN] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4198),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_PERCENT2] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2986)] = {
    [sym_heredoc_content] = STATE(2986),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2987)] = {
    [sym_heredoc_content] = STATE(2987),
    [anon_sym_COLON] = ACTIONS(3905),
    [anon_sym_DOLLAR] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_AT] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_RPAREN] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3907),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_PERCENT2] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2988)] = {
    [sym_heredoc_content] = STATE(2988),
    [anon_sym_COLON] = ACTIONS(4067),
    [anon_sym_DOLLAR] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_AT] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_RPAREN] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4069),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_PERCENT2] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2989)] = {
    [sym_heredoc_content] = STATE(2989),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2990)] = {
    [sym_heredoc_content] = STATE(2990),
    [anon_sym_COLON] = ACTIONS(3825),
    [anon_sym_DOLLAR] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_AT] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_RPAREN] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3827),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_PERCENT2] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2991)] = {
    [sym_heredoc_content] = STATE(2991),
    [anon_sym_COLON] = ACTIONS(3849),
    [anon_sym_DOLLAR] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_AT] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_RPAREN] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3851),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_PERCENT2] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2992)] = {
    [sym_heredoc_content] = STATE(2992),
    [anon_sym_COLON] = ACTIONS(4051),
    [anon_sym_DOLLAR] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_AT] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_RPAREN] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4053),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_PERCENT2] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2993)] = {
    [sym_heredoc_content] = STATE(2993),
    [anon_sym_COLON] = ACTIONS(4059),
    [anon_sym_DOLLAR] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_AT] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_RPAREN] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4061),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_PERCENT2] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2994)] = {
    [sym_heredoc_content] = STATE(2994),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_and] = ACTIONS(3719),
    [anon_sym_or] = ACTIONS(3719),
    [anon_sym_xor] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_RBRACK] = ACTIONS(3719),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3719),
  },
  [STATE(2995)] = {
    [sym_heredoc_content] = STATE(2995),
    [anon_sym_COLON] = ACTIONS(4099),
    [anon_sym_DOLLAR] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_AT] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_RPAREN] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4101),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_PERCENT2] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2996)] = {
    [sym_heredoc_content] = STATE(2996),
    [anon_sym_COLON] = ACTIONS(4127),
    [anon_sym_DOLLAR] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_AT] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_RPAREN] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4129),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_PERCENT2] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2997)] = {
    [sym_heredoc_content] = STATE(2997),
    [anon_sym_COLON] = ACTIONS(4147),
    [anon_sym_DOLLAR] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_AT] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_RPAREN] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4149),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_PERCENT2] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2998)] = {
    [sym_heredoc_content] = STATE(2998),
    [anon_sym_COLON] = ACTIONS(3961),
    [anon_sym_DOLLAR] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_AT] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_RPAREN] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3963),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_PERCENT2] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(2999)] = {
    [sym_heredoc_content] = STATE(2999),
    [anon_sym_COLON] = ACTIONS(3913),
    [anon_sym_DOLLAR] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_AT] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_RPAREN] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3915),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_PERCENT2] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3000)] = {
    [sym_heredoc_content] = STATE(3000),
    [anon_sym_COLON] = ACTIONS(4143),
    [anon_sym_DOLLAR] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_AT] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_RPAREN] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4145),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_PERCENT2] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3001)] = {
    [sym_heredoc_content] = STATE(3001),
    [anon_sym_COLON] = ACTIONS(3799),
    [anon_sym_DOLLAR] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_AT] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_RPAREN] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3801),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_PERCENT2] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3002)] = {
    [sym_heredoc_content] = STATE(3002),
    [anon_sym_COLON] = ACTIONS(3921),
    [anon_sym_DOLLAR] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_AT] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_RPAREN] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3923),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_PERCENT2] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3003)] = {
    [sym_heredoc_content] = STATE(3003),
    [sym__identifier] = ACTIONS(3401),
    [anon_sym_COLON] = ACTIONS(3401),
    [anon_sym_SEMI] = ACTIONS(3403),
    [anon_sym_EQ] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_COMMA] = ACTIONS(3403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3403),
    [anon_sym_DASH_GT] = ACTIONS(3403),
    [anon_sym_STAR] = ACTIONS(3401),
    [anon_sym_PERCENT] = ACTIONS(3401),
    [anon_sym_GT] = ACTIONS(3401),
    [anon_sym_LT_LT] = ACTIONS(3401),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3403),
    [anon_sym_PLUS_EQ] = ACTIONS(3403),
    [anon_sym_DASH_EQ] = ACTIONS(3403),
    [anon_sym_DOT_EQ] = ACTIONS(3403),
    [anon_sym_STAR_EQ] = ACTIONS(3403),
    [anon_sym_SLASH_EQ] = ACTIONS(3403),
    [anon_sym_PERCENT_EQ] = ACTIONS(3403),
    [anon_sym_x_EQ] = ACTIONS(3403),
    [anon_sym_AMP_EQ] = ACTIONS(3403),
    [anon_sym_PIPE_EQ] = ACTIONS(3403),
    [anon_sym_CARET_EQ] = ACTIONS(3403),
    [anon_sym_LT_LT_EQ] = ACTIONS(3403),
    [anon_sym_GT_GT_EQ] = ACTIONS(3403),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3403),
    [anon_sym_DOT_DOT] = ACTIONS(3401),
    [anon_sym_STAR_STAR] = ACTIONS(3401),
    [anon_sym_PIPE_PIPE] = ACTIONS(3401),
    [anon_sym_SLASH_SLASH] = ACTIONS(3401),
    [anon_sym_CARET_CARET] = ACTIONS(3403),
    [anon_sym_AMP_AMP] = ACTIONS(3401),
    [anon_sym_PIPE] = ACTIONS(3401),
    [anon_sym_CARET] = ACTIONS(3401),
    [anon_sym_AMP] = ACTIONS(3401),
    [anon_sym_GT_GT2] = ACTIONS(3401),
    [anon_sym_PLUS] = ACTIONS(3401),
    [anon_sym_DASH] = ACTIONS(3401),
    [anon_sym_DOT] = ACTIONS(3401),
    [anon_sym_SLASH] = ACTIONS(3401),
    [anon_sym_x] = ACTIONS(3401),
    [anon_sym_EQ_TILDE] = ACTIONS(3403),
    [anon_sym_BANG_TILDE] = ACTIONS(3403),
    [anon_sym_EQ_EQ] = ACTIONS(3401),
    [anon_sym_BANG_EQ] = ACTIONS(3403),
    [anon_sym_eq] = ACTIONS(3401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3403),
    [anon_sym_equ] = ACTIONS(3401),
    [anon_sym_eqr] = ACTIONS(3401),
    [anon_sym_ne] = ACTIONS(3401),
    [anon_sym_LT_EQ_GT] = ACTIONS(3403),
    [anon_sym_cmp] = ACTIONS(3401),
    [anon_sym_TILDE_TILDE] = ACTIONS(3403),
    [anon_sym_LT] = ACTIONS(3401),
    [anon_sym_LT_EQ] = ACTIONS(3401),
    [anon_sym_GT_EQ] = ACTIONS(3403),
    [anon_sym_lt] = ACTIONS(3401),
    [anon_sym_le] = ACTIONS(3401),
    [anon_sym_ge] = ACTIONS(3401),
    [anon_sym_gt] = ACTIONS(3401),
    [anon_sym_isa] = ACTIONS(3401),
    [anon_sym_PLUS_PLUS] = ACTIONS(3403),
    [anon_sym_DASH_DASH] = ACTIONS(3403),
    [anon_sym_QMARK] = ACTIONS(3403),
    [anon_sym_EQ_GT] = ACTIONS(3403),
    [sym_comment] = ACTIONS(5),
    [aux_sym__bareword_token1] = ACTIONS(3401),
    [sym__PERLY_SEMICOLON] = ACTIONS(3403),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3004)] = {
    [sym_heredoc_content] = STATE(3004),
    [anon_sym_COLON] = ACTIONS(3723),
    [anon_sym_DOLLAR] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_AT] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_RPAREN] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3725),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_PERCENT2] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3005)] = {
    [sym_heredoc_content] = STATE(3005),
    [anon_sym_COLON] = ACTIONS(3751),
    [anon_sym_DOLLAR] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_AT] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_RPAREN] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3753),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_PERCENT2] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3006)] = {
    [sym_heredoc_content] = STATE(3006),
    [anon_sym_COLON] = ACTIONS(3755),
    [anon_sym_DOLLAR] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_AT] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_RPAREN] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3757),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_PERCENT2] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3007)] = {
    [sym_heredoc_content] = STATE(3007),
    [anon_sym_COLON] = ACTIONS(3759),
    [anon_sym_DOLLAR] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_AT] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_RPAREN] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3761),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_PERCENT2] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3008)] = {
    [sym_heredoc_content] = STATE(3008),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3763),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3763),
  },
  [STATE(3009)] = {
    [sym_heredoc_content] = STATE(3009),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_and] = ACTIONS(3763),
    [anon_sym_or] = ACTIONS(3763),
    [anon_sym_xor] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3763),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4982),
    [anon_sym_cmp] = ACTIONS(4982),
    [anon_sym_TILDE_TILDE] = ACTIONS(4982),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3763),
  },
  [STATE(3010)] = {
    [sym_heredoc_content] = STATE(3010),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_and] = ACTIONS(3735),
    [anon_sym_or] = ACTIONS(3735),
    [anon_sym_xor] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3735),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4984),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3735),
  },
  [STATE(3011)] = {
    [sym_heredoc_content] = STATE(3011),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_and] = ACTIONS(3771),
    [anon_sym_or] = ACTIONS(3771),
    [anon_sym_xor] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_RBRACK] = ACTIONS(3771),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3771),
  },
  [STATE(3012)] = {
    [sym_heredoc_content] = STATE(3012),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_and] = ACTIONS(3783),
    [anon_sym_or] = ACTIONS(3783),
    [anon_sym_xor] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_RBRACK] = ACTIONS(3783),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3783),
  },
  [STATE(3013)] = {
    [sym_heredoc_content] = STATE(3013),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_and] = ACTIONS(3787),
    [anon_sym_or] = ACTIONS(3787),
    [anon_sym_xor] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_RBRACK] = ACTIONS(3787),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3787),
  },
  [STATE(3014)] = {
    [sym_heredoc_content] = STATE(3014),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_and] = ACTIONS(3791),
    [anon_sym_or] = ACTIONS(3791),
    [anon_sym_xor] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_RBRACK] = ACTIONS(3791),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3791),
  },
  [STATE(3015)] = {
    [sym_heredoc_content] = STATE(3015),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_and] = ACTIONS(3795),
    [anon_sym_or] = ACTIONS(3795),
    [anon_sym_xor] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_RBRACK] = ACTIONS(3795),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3795),
  },
  [STATE(3016)] = {
    [sym_heredoc_content] = STATE(3016),
    [anon_sym_COLON] = ACTIONS(3689),
    [anon_sym_DOLLAR] = ACTIONS(3689),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_AT] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_RPAREN] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3693),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_PERCENT2] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3017)] = {
    [sym_heredoc_content] = STATE(3017),
    [anon_sym_COLON] = ACTIONS(3669),
    [anon_sym_DOLLAR] = ACTIONS(3669),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_AT] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_RPAREN] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3673),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_PERCENT2] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3018)] = {
    [sym_heredoc_content] = STATE(3018),
    [anon_sym_COLON] = ACTIONS(2858),
    [anon_sym_DOLLAR] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2854),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_PERCENT2] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3019)] = {
    [sym_heredoc_content] = STATE(3019),
    [anon_sym_COLON] = ACTIONS(2858),
    [anon_sym_DOLLAR] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_AT] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2854),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_PERCENT2] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3020)] = {
    [sym_heredoc_content] = STATE(3020),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_and] = ACTIONS(3925),
    [anon_sym_or] = ACTIONS(3925),
    [anon_sym_xor] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_RBRACK] = ACTIONS(3925),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3925),
  },
  [STATE(3021)] = {
    [sym_heredoc_content] = STATE(3021),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_and] = ACTIONS(3929),
    [anon_sym_or] = ACTIONS(3929),
    [anon_sym_xor] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_RBRACK] = ACTIONS(3929),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3929),
  },
  [STATE(3022)] = {
    [sym_heredoc_content] = STATE(3022),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_and] = ACTIONS(3933),
    [anon_sym_or] = ACTIONS(3933),
    [anon_sym_xor] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_RBRACK] = ACTIONS(3933),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3933),
  },
  [STATE(3023)] = {
    [sym_heredoc_content] = STATE(3023),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_and] = ACTIONS(3937),
    [anon_sym_or] = ACTIONS(3937),
    [anon_sym_xor] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_RBRACK] = ACTIONS(3937),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3937),
  },
  [STATE(3024)] = {
    [sym_heredoc_content] = STATE(3024),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_and] = ACTIONS(3953),
    [anon_sym_or] = ACTIONS(3953),
    [anon_sym_xor] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_RBRACK] = ACTIONS(3953),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3953),
  },
  [STATE(3025)] = {
    [sym_heredoc_content] = STATE(3025),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_and] = ACTIONS(3965),
    [anon_sym_or] = ACTIONS(3965),
    [anon_sym_xor] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_RBRACK] = ACTIONS(3965),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3965),
  },
  [STATE(3026)] = {
    [sym_heredoc_content] = STATE(3026),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3027)] = {
    [sym_heredoc_content] = STATE(3027),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_and] = ACTIONS(4155),
    [anon_sym_or] = ACTIONS(4155),
    [anon_sym_xor] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_RBRACK] = ACTIONS(4155),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4155),
  },
  [STATE(3028)] = {
    [sym_heredoc_content] = STATE(3028),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_and] = ACTIONS(4176),
    [anon_sym_or] = ACTIONS(4176),
    [anon_sym_xor] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_RBRACK] = ACTIONS(4176),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4176),
  },
  [STATE(3029)] = {
    [sym_heredoc_content] = STATE(3029),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_and] = ACTIONS(4180),
    [anon_sym_or] = ACTIONS(4180),
    [anon_sym_xor] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_RBRACK] = ACTIONS(4180),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4180),
  },
  [STATE(3030)] = {
    [sym_heredoc_content] = STATE(3030),
    [anon_sym_COLON] = ACTIONS(3590),
    [anon_sym_DOLLAR] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3031)] = {
    [sym_heredoc_content] = STATE(3031),
    [anon_sym_RBRACE] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_and] = ACTIONS(4013),
    [anon_sym_or] = ACTIONS(4013),
    [anon_sym_xor] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4013),
  },
  [STATE(3032)] = {
    [sym_heredoc_content] = STATE(3032),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_and] = ACTIONS(4192),
    [anon_sym_or] = ACTIONS(4192),
    [anon_sym_xor] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_RBRACK] = ACTIONS(4192),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4192),
  },
  [STATE(3033)] = {
    [sym_heredoc_content] = STATE(3033),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_and] = ACTIONS(4200),
    [anon_sym_or] = ACTIONS(4200),
    [anon_sym_xor] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_RBRACK] = ACTIONS(4200),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4200),
  },
  [STATE(3034)] = {
    [sym_heredoc_content] = STATE(3034),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_and] = ACTIONS(3969),
    [anon_sym_or] = ACTIONS(3969),
    [anon_sym_xor] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_RBRACK] = ACTIONS(3969),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3969),
  },
  [STATE(3035)] = {
    [sym_heredoc_content] = STATE(3035),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_and] = ACTIONS(4079),
    [anon_sym_or] = ACTIONS(4079),
    [anon_sym_xor] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_RBRACK] = ACTIONS(4079),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4079),
  },
  [STATE(3036)] = {
    [sym_heredoc_content] = STATE(3036),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_and] = ACTIONS(4083),
    [anon_sym_or] = ACTIONS(4083),
    [anon_sym_xor] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_RBRACK] = ACTIONS(4083),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4083),
  },
  [STATE(3037)] = {
    [sym_heredoc_content] = STATE(3037),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_and] = ACTIONS(4091),
    [anon_sym_or] = ACTIONS(4091),
    [anon_sym_xor] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_RBRACK] = ACTIONS(4091),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4091),
  },
  [STATE(3038)] = {
    [sym_heredoc_content] = STATE(3038),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_and] = ACTIONS(3715),
    [anon_sym_or] = ACTIONS(3715),
    [anon_sym_xor] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_RBRACK] = ACTIONS(3715),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3715),
  },
  [STATE(3039)] = {
    [sym_heredoc_content] = STATE(3039),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_and] = ACTIONS(3893),
    [anon_sym_or] = ACTIONS(3893),
    [anon_sym_xor] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_RBRACK] = ACTIONS(3893),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3893),
  },
  [STATE(3040)] = {
    [sym_heredoc_content] = STATE(3040),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_and] = ACTIONS(4063),
    [anon_sym_or] = ACTIONS(4063),
    [anon_sym_xor] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_RBRACK] = ACTIONS(4063),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4063),
  },
  [STATE(3041)] = {
    [sym_heredoc_content] = STATE(3041),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_and] = ACTIONS(3877),
    [anon_sym_or] = ACTIONS(3877),
    [anon_sym_xor] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_RBRACK] = ACTIONS(3877),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3877),
  },
  [STATE(3042)] = {
    [sym_heredoc_content] = STATE(3042),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_and] = ACTIONS(3989),
    [anon_sym_or] = ACTIONS(3989),
    [anon_sym_xor] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_RBRACK] = ACTIONS(3989),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3989),
  },
  [STATE(3043)] = {
    [sym_heredoc_content] = STATE(3043),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_and] = ACTIONS(4115),
    [anon_sym_or] = ACTIONS(4115),
    [anon_sym_xor] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_RBRACK] = ACTIONS(4115),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4115),
  },
  [STATE(3044)] = {
    [sym_heredoc_content] = STATE(3044),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_and] = ACTIONS(3807),
    [anon_sym_or] = ACTIONS(3807),
    [anon_sym_xor] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_RBRACK] = ACTIONS(3807),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3807),
  },
  [STATE(3045)] = {
    [sym_heredoc_content] = STATE(3045),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_and] = ACTIONS(4041),
    [anon_sym_or] = ACTIONS(4041),
    [anon_sym_xor] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_RBRACK] = ACTIONS(4041),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4041),
  },
  [STATE(3046)] = {
    [sym_heredoc_content] = STATE(3046),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_and] = ACTIONS(4055),
    [anon_sym_or] = ACTIONS(4055),
    [anon_sym_xor] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_RBRACK] = ACTIONS(4055),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4055),
  },
  [STATE(3047)] = {
    [sym_heredoc_content] = STATE(3047),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_and] = ACTIONS(3865),
    [anon_sym_or] = ACTIONS(3865),
    [anon_sym_xor] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_RBRACK] = ACTIONS(3865),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3865),
  },
  [STATE(3048)] = {
    [sym_heredoc_content] = STATE(3048),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_and] = ACTIONS(3829),
    [anon_sym_or] = ACTIONS(3829),
    [anon_sym_xor] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_RBRACK] = ACTIONS(3829),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3829),
  },
  [STATE(3049)] = {
    [sym_heredoc_content] = STATE(3049),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_and] = ACTIONS(4095),
    [anon_sym_or] = ACTIONS(4095),
    [anon_sym_xor] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_RBRACK] = ACTIONS(4095),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4095),
  },
  [STATE(3050)] = {
    [sym_heredoc_content] = STATE(3050),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_and] = ACTIONS(4184),
    [anon_sym_or] = ACTIONS(4184),
    [anon_sym_xor] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_RBRACK] = ACTIONS(4184),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4184),
  },
  [STATE(3051)] = {
    [sym_heredoc_content] = STATE(3051),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_and] = ACTIONS(3727),
    [anon_sym_or] = ACTIONS(3727),
    [anon_sym_xor] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_RBRACK] = ACTIONS(3727),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3727),
  },
  [STATE(3052)] = {
    [sym_heredoc_content] = STATE(3052),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_and] = ACTIONS(3743),
    [anon_sym_or] = ACTIONS(3743),
    [anon_sym_xor] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_RBRACK] = ACTIONS(3743),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3743),
  },
  [STATE(3053)] = {
    [sym_heredoc_content] = STATE(3053),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_and] = ACTIONS(3811),
    [anon_sym_or] = ACTIONS(3811),
    [anon_sym_xor] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_RBRACK] = ACTIONS(3811),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3811),
  },
  [STATE(3054)] = {
    [sym_heredoc_content] = STATE(3054),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_and] = ACTIONS(3873),
    [anon_sym_or] = ACTIONS(3873),
    [anon_sym_xor] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_RBRACK] = ACTIONS(3873),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3873),
  },
  [STATE(3055)] = {
    [sym_heredoc_content] = STATE(3055),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2536),
  },
  [STATE(3056)] = {
    [sym_heredoc_content] = STATE(3056),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_and] = ACTIONS(3981),
    [anon_sym_or] = ACTIONS(3981),
    [anon_sym_xor] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_RBRACK] = ACTIONS(3981),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3981),
  },
  [STATE(3057)] = {
    [sym_heredoc_content] = STATE(3057),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4774),
    [anon_sym_and] = ACTIONS(3985),
    [anon_sym_or] = ACTIONS(3985),
    [anon_sym_xor] = ACTIONS(3985),
    [anon_sym_DASH_GT] = ACTIONS(4776),
    [anon_sym_STAR] = ACTIONS(4778),
    [anon_sym_RBRACK] = ACTIONS(3985),
    [anon_sym_PERCENT] = ACTIONS(4778),
    [anon_sym_GT] = ACTIONS(4780),
    [anon_sym_LT_LT] = ACTIONS(4782),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4784),
    [anon_sym_STAR_STAR] = ACTIONS(4786),
    [anon_sym_PIPE_PIPE] = ACTIONS(4788),
    [anon_sym_SLASH_SLASH] = ACTIONS(4788),
    [anon_sym_CARET_CARET] = ACTIONS(4790),
    [anon_sym_AMP_AMP] = ACTIONS(4792),
    [anon_sym_PIPE] = ACTIONS(4794),
    [anon_sym_CARET] = ACTIONS(4794),
    [anon_sym_AMP] = ACTIONS(4796),
    [anon_sym_GT_GT2] = ACTIONS(4782),
    [anon_sym_PLUS] = ACTIONS(4798),
    [anon_sym_DASH] = ACTIONS(4798),
    [anon_sym_DOT] = ACTIONS(4798),
    [anon_sym_SLASH] = ACTIONS(4778),
    [anon_sym_x] = ACTIONS(4778),
    [anon_sym_EQ_TILDE] = ACTIONS(4800),
    [anon_sym_BANG_TILDE] = ACTIONS(4800),
    [anon_sym_EQ_EQ] = ACTIONS(4802),
    [anon_sym_BANG_EQ] = ACTIONS(4804),
    [anon_sym_eq] = ACTIONS(4802),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4804),
    [anon_sym_equ] = ACTIONS(4804),
    [anon_sym_eqr] = ACTIONS(4804),
    [anon_sym_ne] = ACTIONS(4804),
    [anon_sym_LT_EQ_GT] = ACTIONS(4806),
    [anon_sym_cmp] = ACTIONS(4806),
    [anon_sym_TILDE_TILDE] = ACTIONS(4806),
    [anon_sym_LT] = ACTIONS(4780),
    [anon_sym_LT_EQ] = ACTIONS(4780),
    [anon_sym_GT_EQ] = ACTIONS(4808),
    [anon_sym_lt] = ACTIONS(4808),
    [anon_sym_le] = ACTIONS(4808),
    [anon_sym_ge] = ACTIONS(4808),
    [anon_sym_gt] = ACTIONS(4808),
    [anon_sym_isa] = ACTIONS(4810),
    [anon_sym_PLUS_PLUS] = ACTIONS(4812),
    [anon_sym_DASH_DASH] = ACTIONS(4812),
    [anon_sym_QMARK] = ACTIONS(4814),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3985),
  },
  [STATE(3058)] = {
    [sym_heredoc_content] = STATE(3058),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_and] = ACTIONS(4001),
    [anon_sym_or] = ACTIONS(4001),
    [anon_sym_xor] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_RBRACK] = ACTIONS(4001),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4001),
  },
  [STATE(3059)] = {
    [sym_heredoc_content] = STATE(3059),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_and] = ACTIONS(4009),
    [anon_sym_or] = ACTIONS(4009),
    [anon_sym_xor] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_RBRACK] = ACTIONS(4009),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4009),
  },
  [STATE(3060)] = {
    [sym_heredoc_content] = STATE(3060),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_and] = ACTIONS(4013),
    [anon_sym_or] = ACTIONS(4013),
    [anon_sym_xor] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_RBRACK] = ACTIONS(4013),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4013),
  },
  [STATE(3061)] = {
    [sym_heredoc_content] = STATE(3061),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_and] = ACTIONS(4025),
    [anon_sym_or] = ACTIONS(4025),
    [anon_sym_xor] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_RBRACK] = ACTIONS(4025),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4025),
  },
  [STATE(3062)] = {
    [sym_heredoc_content] = STATE(3062),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_and] = ACTIONS(4029),
    [anon_sym_or] = ACTIONS(4029),
    [anon_sym_xor] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_RBRACK] = ACTIONS(4029),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4029),
  },
  [STATE(3063)] = {
    [sym_heredoc_content] = STATE(3063),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_PAREN_CLOSE] = ACTIONS(2536),
  },
  [STATE(3064)] = {
    [sym_heredoc_content] = STATE(3064),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_and] = ACTIONS(4111),
    [anon_sym_or] = ACTIONS(4111),
    [anon_sym_xor] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_RBRACK] = ACTIONS(4111),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4111),
  },
  [STATE(3065)] = {
    [sym_heredoc_content] = STATE(3065),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_and] = ACTIONS(4123),
    [anon_sym_or] = ACTIONS(4123),
    [anon_sym_xor] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_RBRACK] = ACTIONS(4123),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4123),
  },
  [STATE(3066)] = {
    [sym_heredoc_content] = STATE(3066),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_and] = ACTIONS(3945),
    [anon_sym_or] = ACTIONS(3945),
    [anon_sym_xor] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_RBRACK] = ACTIONS(3945),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3945),
  },
  [STATE(3067)] = {
    [sym_heredoc_content] = STATE(3067),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_and] = ACTIONS(4196),
    [anon_sym_or] = ACTIONS(4196),
    [anon_sym_xor] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_RBRACK] = ACTIONS(4196),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4196),
  },
  [STATE(3068)] = {
    [sym_heredoc_content] = STATE(3068),
    [anon_sym_LBRACE] = ACTIONS(2723),
    [anon_sym_SEMI] = ACTIONS(2723),
    [anon_sym_EQ] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_COMMA] = ACTIONS(2723),
    [anon_sym_LPAREN] = ACTIONS(2723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2723),
    [anon_sym_DASH_GT] = ACTIONS(2723),
    [anon_sym_STAR] = ACTIONS(2721),
    [anon_sym_LBRACK] = ACTIONS(2723),
    [anon_sym_PERCENT] = ACTIONS(2721),
    [anon_sym_GT] = ACTIONS(2721),
    [anon_sym_LT_LT] = ACTIONS(2721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2723),
    [anon_sym_PLUS_EQ] = ACTIONS(2723),
    [anon_sym_DASH_EQ] = ACTIONS(2723),
    [anon_sym_DOT_EQ] = ACTIONS(2723),
    [anon_sym_STAR_EQ] = ACTIONS(2723),
    [anon_sym_SLASH_EQ] = ACTIONS(2723),
    [anon_sym_PERCENT_EQ] = ACTIONS(2723),
    [anon_sym_x_EQ] = ACTIONS(2723),
    [anon_sym_AMP_EQ] = ACTIONS(2723),
    [anon_sym_PIPE_EQ] = ACTIONS(2723),
    [anon_sym_CARET_EQ] = ACTIONS(2723),
    [anon_sym_LT_LT_EQ] = ACTIONS(2723),
    [anon_sym_GT_GT_EQ] = ACTIONS(2723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2723),
    [anon_sym_DOT_DOT] = ACTIONS(2721),
    [anon_sym_STAR_STAR] = ACTIONS(2721),
    [anon_sym_PIPE_PIPE] = ACTIONS(2721),
    [anon_sym_SLASH_SLASH] = ACTIONS(2721),
    [anon_sym_CARET_CARET] = ACTIONS(2723),
    [anon_sym_AMP_AMP] = ACTIONS(2721),
    [anon_sym_PIPE] = ACTIONS(2721),
    [anon_sym_CARET] = ACTIONS(2721),
    [anon_sym_AMP] = ACTIONS(2721),
    [anon_sym_GT_GT2] = ACTIONS(2721),
    [anon_sym_PLUS] = ACTIONS(2721),
    [anon_sym_DASH] = ACTIONS(2721),
    [anon_sym_DOT] = ACTIONS(2721),
    [anon_sym_SLASH] = ACTIONS(2721),
    [anon_sym_x] = ACTIONS(2721),
    [anon_sym_EQ_TILDE] = ACTIONS(2723),
    [anon_sym_BANG_TILDE] = ACTIONS(2723),
    [anon_sym_EQ_EQ] = ACTIONS(2721),
    [anon_sym_BANG_EQ] = ACTIONS(2723),
    [anon_sym_eq] = ACTIONS(2721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2723),
    [anon_sym_equ] = ACTIONS(2723),
    [anon_sym_eqr] = ACTIONS(2723),
    [anon_sym_ne] = ACTIONS(2723),
    [anon_sym_LT_EQ_GT] = ACTIONS(2723),
    [anon_sym_cmp] = ACTIONS(2723),
    [anon_sym_TILDE_TILDE] = ACTIONS(2723),
    [anon_sym_LT] = ACTIONS(2721),
    [anon_sym_LT_EQ] = ACTIONS(2721),
    [anon_sym_GT_EQ] = ACTIONS(2723),
    [anon_sym_lt] = ACTIONS(2723),
    [anon_sym_le] = ACTIONS(2723),
    [anon_sym_ge] = ACTIONS(2723),
    [anon_sym_gt] = ACTIONS(2723),
    [anon_sym_isa] = ACTIONS(2723),
    [anon_sym_PLUS_PLUS] = ACTIONS(2723),
    [anon_sym_DASH_DASH] = ACTIONS(2723),
    [anon_sym_QMARK] = ACTIONS(2723),
    [anon_sym_EQ_GT] = ACTIONS(2723),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3069)] = {
    [sym_heredoc_content] = STATE(3069),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_and] = ACTIONS(3905),
    [anon_sym_or] = ACTIONS(3905),
    [anon_sym_xor] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_RBRACK] = ACTIONS(3905),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3905),
  },
  [STATE(3070)] = {
    [sym_heredoc_content] = STATE(3070),
    [anon_sym_LBRACE] = ACTIONS(2698),
    [anon_sym_SEMI] = ACTIONS(2698),
    [anon_sym_EQ] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_COMMA] = ACTIONS(2698),
    [anon_sym_LPAREN] = ACTIONS(2698),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2698),
    [anon_sym_DASH_GT] = ACTIONS(2698),
    [anon_sym_STAR] = ACTIONS(2696),
    [anon_sym_LBRACK] = ACTIONS(2698),
    [anon_sym_PERCENT] = ACTIONS(2696),
    [anon_sym_GT] = ACTIONS(2696),
    [anon_sym_LT_LT] = ACTIONS(2696),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2698),
    [anon_sym_PLUS_EQ] = ACTIONS(2698),
    [anon_sym_DASH_EQ] = ACTIONS(2698),
    [anon_sym_DOT_EQ] = ACTIONS(2698),
    [anon_sym_STAR_EQ] = ACTIONS(2698),
    [anon_sym_SLASH_EQ] = ACTIONS(2698),
    [anon_sym_PERCENT_EQ] = ACTIONS(2698),
    [anon_sym_x_EQ] = ACTIONS(2698),
    [anon_sym_AMP_EQ] = ACTIONS(2698),
    [anon_sym_PIPE_EQ] = ACTIONS(2698),
    [anon_sym_CARET_EQ] = ACTIONS(2698),
    [anon_sym_LT_LT_EQ] = ACTIONS(2698),
    [anon_sym_GT_GT_EQ] = ACTIONS(2698),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2698),
    [anon_sym_DOT_DOT] = ACTIONS(2696),
    [anon_sym_STAR_STAR] = ACTIONS(2696),
    [anon_sym_PIPE_PIPE] = ACTIONS(2696),
    [anon_sym_SLASH_SLASH] = ACTIONS(2696),
    [anon_sym_CARET_CARET] = ACTIONS(2698),
    [anon_sym_AMP_AMP] = ACTIONS(2696),
    [anon_sym_PIPE] = ACTIONS(2696),
    [anon_sym_CARET] = ACTIONS(2696),
    [anon_sym_AMP] = ACTIONS(2696),
    [anon_sym_GT_GT2] = ACTIONS(2696),
    [anon_sym_PLUS] = ACTIONS(2696),
    [anon_sym_DASH] = ACTIONS(2696),
    [anon_sym_DOT] = ACTIONS(2696),
    [anon_sym_SLASH] = ACTIONS(2696),
    [anon_sym_x] = ACTIONS(2696),
    [anon_sym_EQ_TILDE] = ACTIONS(2698),
    [anon_sym_BANG_TILDE] = ACTIONS(2698),
    [anon_sym_EQ_EQ] = ACTIONS(2696),
    [anon_sym_BANG_EQ] = ACTIONS(2698),
    [anon_sym_eq] = ACTIONS(2696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2698),
    [anon_sym_equ] = ACTIONS(2698),
    [anon_sym_eqr] = ACTIONS(2698),
    [anon_sym_ne] = ACTIONS(2698),
    [anon_sym_LT_EQ_GT] = ACTIONS(2698),
    [anon_sym_cmp] = ACTIONS(2698),
    [anon_sym_TILDE_TILDE] = ACTIONS(2698),
    [anon_sym_LT] = ACTIONS(2696),
    [anon_sym_LT_EQ] = ACTIONS(2696),
    [anon_sym_GT_EQ] = ACTIONS(2698),
    [anon_sym_lt] = ACTIONS(2698),
    [anon_sym_le] = ACTIONS(2698),
    [anon_sym_ge] = ACTIONS(2698),
    [anon_sym_gt] = ACTIONS(2698),
    [anon_sym_isa] = ACTIONS(2698),
    [anon_sym_PLUS_PLUS] = ACTIONS(2698),
    [anon_sym_DASH_DASH] = ACTIONS(2698),
    [anon_sym_QMARK] = ACTIONS(2698),
    [anon_sym_EQ_GT] = ACTIONS(2698),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2698),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3071)] = {
    [sym_heredoc_content] = STATE(3071),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_and] = ACTIONS(4067),
    [anon_sym_or] = ACTIONS(4067),
    [anon_sym_xor] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_RBRACK] = ACTIONS(4067),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4067),
  },
  [STATE(3072)] = {
    [sym_heredoc_content] = STATE(3072),
    [anon_sym_COLON] = ACTIONS(2536),
    [anon_sym_DOLLAR] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_AT] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2534),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_PERCENT2] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3073)] = {
    [sym_heredoc_content] = STATE(3073),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_and] = ACTIONS(3825),
    [anon_sym_or] = ACTIONS(3825),
    [anon_sym_xor] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_RBRACK] = ACTIONS(3825),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3825),
  },
  [STATE(3074)] = {
    [sym_heredoc_content] = STATE(3074),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_and] = ACTIONS(3849),
    [anon_sym_or] = ACTIONS(3849),
    [anon_sym_xor] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_RBRACK] = ACTIONS(3849),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3849),
  },
  [STATE(3075)] = {
    [sym_heredoc_content] = STATE(3075),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_and] = ACTIONS(4051),
    [anon_sym_or] = ACTIONS(4051),
    [anon_sym_xor] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_RBRACK] = ACTIONS(4051),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4051),
  },
  [STATE(3076)] = {
    [sym_heredoc_content] = STATE(3076),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_and] = ACTIONS(4059),
    [anon_sym_or] = ACTIONS(4059),
    [anon_sym_xor] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_RBRACK] = ACTIONS(4059),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4059),
  },
  [STATE(3077)] = {
    [sym_heredoc_content] = STATE(3077),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_and] = ACTIONS(4099),
    [anon_sym_or] = ACTIONS(4099),
    [anon_sym_xor] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_RBRACK] = ACTIONS(4099),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4099),
  },
  [STATE(3078)] = {
    [sym_heredoc_content] = STATE(3078),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_and] = ACTIONS(4127),
    [anon_sym_or] = ACTIONS(4127),
    [anon_sym_xor] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_RBRACK] = ACTIONS(4127),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4127),
  },
  [STATE(3079)] = {
    [sym_heredoc_content] = STATE(3079),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_and] = ACTIONS(4147),
    [anon_sym_or] = ACTIONS(4147),
    [anon_sym_xor] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_RBRACK] = ACTIONS(4147),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4147),
  },
  [STATE(3080)] = {
    [sym_heredoc_content] = STATE(3080),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_and] = ACTIONS(3961),
    [anon_sym_or] = ACTIONS(3961),
    [anon_sym_xor] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_RBRACK] = ACTIONS(3961),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3961),
  },
  [STATE(3081)] = {
    [sym_heredoc_content] = STATE(3081),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_and] = ACTIONS(3913),
    [anon_sym_or] = ACTIONS(3913),
    [anon_sym_xor] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_RBRACK] = ACTIONS(3913),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3913),
  },
  [STATE(3082)] = {
    [sym_heredoc_content] = STATE(3082),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_and] = ACTIONS(4143),
    [anon_sym_or] = ACTIONS(4143),
    [anon_sym_xor] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_RBRACK] = ACTIONS(4143),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4143),
  },
  [STATE(3083)] = {
    [sym_heredoc_content] = STATE(3083),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_and] = ACTIONS(3799),
    [anon_sym_or] = ACTIONS(3799),
    [anon_sym_xor] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_RBRACK] = ACTIONS(3799),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3799),
  },
  [STATE(3084)] = {
    [sym_heredoc_content] = STATE(3084),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_and] = ACTIONS(3921),
    [anon_sym_or] = ACTIONS(3921),
    [anon_sym_xor] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_RBRACK] = ACTIONS(3921),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3921),
  },
  [STATE(3085)] = {
    [sym_heredoc_content] = STATE(3085),
    [anon_sym_LBRACE] = ACTIONS(2694),
    [anon_sym_SEMI] = ACTIONS(2694),
    [anon_sym_EQ] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_COMMA] = ACTIONS(2694),
    [anon_sym_LPAREN] = ACTIONS(2694),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2694),
    [anon_sym_DASH_GT] = ACTIONS(2694),
    [anon_sym_STAR] = ACTIONS(2692),
    [anon_sym_LBRACK] = ACTIONS(2694),
    [anon_sym_PERCENT] = ACTIONS(2692),
    [anon_sym_GT] = ACTIONS(2692),
    [anon_sym_LT_LT] = ACTIONS(2692),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2694),
    [anon_sym_PLUS_EQ] = ACTIONS(2694),
    [anon_sym_DASH_EQ] = ACTIONS(2694),
    [anon_sym_DOT_EQ] = ACTIONS(2694),
    [anon_sym_STAR_EQ] = ACTIONS(2694),
    [anon_sym_SLASH_EQ] = ACTIONS(2694),
    [anon_sym_PERCENT_EQ] = ACTIONS(2694),
    [anon_sym_x_EQ] = ACTIONS(2694),
    [anon_sym_AMP_EQ] = ACTIONS(2694),
    [anon_sym_PIPE_EQ] = ACTIONS(2694),
    [anon_sym_CARET_EQ] = ACTIONS(2694),
    [anon_sym_LT_LT_EQ] = ACTIONS(2694),
    [anon_sym_GT_GT_EQ] = ACTIONS(2694),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2694),
    [anon_sym_DOT_DOT] = ACTIONS(2692),
    [anon_sym_STAR_STAR] = ACTIONS(2692),
    [anon_sym_PIPE_PIPE] = ACTIONS(2692),
    [anon_sym_SLASH_SLASH] = ACTIONS(2692),
    [anon_sym_CARET_CARET] = ACTIONS(2694),
    [anon_sym_AMP_AMP] = ACTIONS(2692),
    [anon_sym_PIPE] = ACTIONS(2692),
    [anon_sym_CARET] = ACTIONS(2692),
    [anon_sym_AMP] = ACTIONS(2692),
    [anon_sym_GT_GT2] = ACTIONS(2692),
    [anon_sym_PLUS] = ACTIONS(2692),
    [anon_sym_DASH] = ACTIONS(2692),
    [anon_sym_DOT] = ACTIONS(2692),
    [anon_sym_SLASH] = ACTIONS(2692),
    [anon_sym_x] = ACTIONS(2692),
    [anon_sym_EQ_TILDE] = ACTIONS(2694),
    [anon_sym_BANG_TILDE] = ACTIONS(2694),
    [anon_sym_EQ_EQ] = ACTIONS(2692),
    [anon_sym_BANG_EQ] = ACTIONS(2694),
    [anon_sym_eq] = ACTIONS(2692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2694),
    [anon_sym_equ] = ACTIONS(2694),
    [anon_sym_eqr] = ACTIONS(2694),
    [anon_sym_ne] = ACTIONS(2694),
    [anon_sym_LT_EQ_GT] = ACTIONS(2694),
    [anon_sym_cmp] = ACTIONS(2694),
    [anon_sym_TILDE_TILDE] = ACTIONS(2694),
    [anon_sym_LT] = ACTIONS(2692),
    [anon_sym_LT_EQ] = ACTIONS(2692),
    [anon_sym_GT_EQ] = ACTIONS(2694),
    [anon_sym_lt] = ACTIONS(2694),
    [anon_sym_le] = ACTIONS(2694),
    [anon_sym_ge] = ACTIONS(2694),
    [anon_sym_gt] = ACTIONS(2694),
    [anon_sym_isa] = ACTIONS(2694),
    [anon_sym_PLUS_PLUS] = ACTIONS(2694),
    [anon_sym_DASH_DASH] = ACTIONS(2694),
    [anon_sym_QMARK] = ACTIONS(2694),
    [anon_sym_EQ_GT] = ACTIONS(2694),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2694),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3086)] = {
    [sym_heredoc_content] = STATE(3086),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_and] = ACTIONS(3723),
    [anon_sym_or] = ACTIONS(3723),
    [anon_sym_xor] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_RBRACK] = ACTIONS(3723),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3723),
  },
  [STATE(3087)] = {
    [sym_heredoc_content] = STATE(3087),
    [anon_sym_LBRACE] = ACTIONS(2686),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_LPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(2686),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3088)] = {
    [sym_heredoc_content] = STATE(3088),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_and] = ACTIONS(3751),
    [anon_sym_or] = ACTIONS(3751),
    [anon_sym_xor] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_RBRACK] = ACTIONS(3751),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3751),
  },
  [STATE(3089)] = {
    [sym_heredoc_content] = STATE(3089),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_and] = ACTIONS(3755),
    [anon_sym_or] = ACTIONS(3755),
    [anon_sym_xor] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_RBRACK] = ACTIONS(3755),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3755),
  },
  [STATE(3090)] = {
    [sym_heredoc_content] = STATE(3090),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_and] = ACTIONS(3759),
    [anon_sym_or] = ACTIONS(3759),
    [anon_sym_xor] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_RBRACK] = ACTIONS(3759),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(3759),
  },
  [STATE(3091)] = {
    [sym_heredoc_content] = STATE(3091),
    [anon_sym_LBRACE] = ACTIONS(2530),
    [anon_sym_SEMI] = ACTIONS(2530),
    [anon_sym_EQ] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_COMMA] = ACTIONS(2530),
    [anon_sym_LPAREN] = ACTIONS(2530),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2530),
    [anon_sym_DASH_GT] = ACTIONS(2530),
    [anon_sym_STAR] = ACTIONS(2532),
    [anon_sym_LBRACK] = ACTIONS(2530),
    [anon_sym_PERCENT] = ACTIONS(2532),
    [anon_sym_GT] = ACTIONS(2532),
    [anon_sym_LT_LT] = ACTIONS(2532),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2530),
    [anon_sym_PLUS_EQ] = ACTIONS(2530),
    [anon_sym_DASH_EQ] = ACTIONS(2530),
    [anon_sym_DOT_EQ] = ACTIONS(2530),
    [anon_sym_STAR_EQ] = ACTIONS(2530),
    [anon_sym_SLASH_EQ] = ACTIONS(2530),
    [anon_sym_PERCENT_EQ] = ACTIONS(2530),
    [anon_sym_x_EQ] = ACTIONS(2530),
    [anon_sym_AMP_EQ] = ACTIONS(2530),
    [anon_sym_PIPE_EQ] = ACTIONS(2530),
    [anon_sym_CARET_EQ] = ACTIONS(2530),
    [anon_sym_LT_LT_EQ] = ACTIONS(2530),
    [anon_sym_GT_GT_EQ] = ACTIONS(2530),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2530),
    [anon_sym_DOT_DOT] = ACTIONS(2532),
    [anon_sym_STAR_STAR] = ACTIONS(2532),
    [anon_sym_PIPE_PIPE] = ACTIONS(2532),
    [anon_sym_SLASH_SLASH] = ACTIONS(2532),
    [anon_sym_CARET_CARET] = ACTIONS(2530),
    [anon_sym_AMP_AMP] = ACTIONS(2532),
    [anon_sym_PIPE] = ACTIONS(2532),
    [anon_sym_CARET] = ACTIONS(2532),
    [anon_sym_AMP] = ACTIONS(2532),
    [anon_sym_GT_GT2] = ACTIONS(2532),
    [anon_sym_PLUS] = ACTIONS(2532),
    [anon_sym_DASH] = ACTIONS(2532),
    [anon_sym_DOT] = ACTIONS(2532),
    [anon_sym_SLASH] = ACTIONS(2532),
    [anon_sym_x] = ACTIONS(2532),
    [anon_sym_EQ_TILDE] = ACTIONS(2530),
    [anon_sym_BANG_TILDE] = ACTIONS(2530),
    [anon_sym_EQ_EQ] = ACTIONS(2532),
    [anon_sym_BANG_EQ] = ACTIONS(2530),
    [anon_sym_eq] = ACTIONS(2532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2530),
    [anon_sym_equ] = ACTIONS(2530),
    [anon_sym_eqr] = ACTIONS(2530),
    [anon_sym_ne] = ACTIONS(2530),
    [anon_sym_LT_EQ_GT] = ACTIONS(2530),
    [anon_sym_cmp] = ACTIONS(2530),
    [anon_sym_TILDE_TILDE] = ACTIONS(2530),
    [anon_sym_LT] = ACTIONS(2532),
    [anon_sym_LT_EQ] = ACTIONS(2532),
    [anon_sym_GT_EQ] = ACTIONS(2530),
    [anon_sym_lt] = ACTIONS(2530),
    [anon_sym_le] = ACTIONS(2530),
    [anon_sym_ge] = ACTIONS(2530),
    [anon_sym_gt] = ACTIONS(2530),
    [anon_sym_isa] = ACTIONS(2530),
    [anon_sym_PLUS_PLUS] = ACTIONS(2530),
    [anon_sym_DASH_DASH] = ACTIONS(2530),
    [anon_sym_QMARK] = ACTIONS(2530),
    [anon_sym_EQ_GT] = ACTIONS(2530),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2530),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3092)] = {
    [sym_heredoc_content] = STATE(3092),
    [anon_sym_LBRACE] = ACTIONS(2690),
    [anon_sym_SEMI] = ACTIONS(2690),
    [anon_sym_EQ] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_COMMA] = ACTIONS(2690),
    [anon_sym_LPAREN] = ACTIONS(2690),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2690),
    [anon_sym_DASH_GT] = ACTIONS(2690),
    [anon_sym_STAR] = ACTIONS(2688),
    [anon_sym_LBRACK] = ACTIONS(2690),
    [anon_sym_PERCENT] = ACTIONS(2688),
    [anon_sym_GT] = ACTIONS(2688),
    [anon_sym_LT_LT] = ACTIONS(2688),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2690),
    [anon_sym_PLUS_EQ] = ACTIONS(2690),
    [anon_sym_DASH_EQ] = ACTIONS(2690),
    [anon_sym_DOT_EQ] = ACTIONS(2690),
    [anon_sym_STAR_EQ] = ACTIONS(2690),
    [anon_sym_SLASH_EQ] = ACTIONS(2690),
    [anon_sym_PERCENT_EQ] = ACTIONS(2690),
    [anon_sym_x_EQ] = ACTIONS(2690),
    [anon_sym_AMP_EQ] = ACTIONS(2690),
    [anon_sym_PIPE_EQ] = ACTIONS(2690),
    [anon_sym_CARET_EQ] = ACTIONS(2690),
    [anon_sym_LT_LT_EQ] = ACTIONS(2690),
    [anon_sym_GT_GT_EQ] = ACTIONS(2690),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2690),
    [anon_sym_DOT_DOT] = ACTIONS(2688),
    [anon_sym_STAR_STAR] = ACTIONS(2688),
    [anon_sym_PIPE_PIPE] = ACTIONS(2688),
    [anon_sym_SLASH_SLASH] = ACTIONS(2688),
    [anon_sym_CARET_CARET] = ACTIONS(2690),
    [anon_sym_AMP_AMP] = ACTIONS(2688),
    [anon_sym_PIPE] = ACTIONS(2688),
    [anon_sym_CARET] = ACTIONS(2688),
    [anon_sym_AMP] = ACTIONS(2688),
    [anon_sym_GT_GT2] = ACTIONS(2688),
    [anon_sym_PLUS] = ACTIONS(2688),
    [anon_sym_DASH] = ACTIONS(2688),
    [anon_sym_DOT] = ACTIONS(2688),
    [anon_sym_SLASH] = ACTIONS(2688),
    [anon_sym_x] = ACTIONS(2688),
    [anon_sym_EQ_TILDE] = ACTIONS(2690),
    [anon_sym_BANG_TILDE] = ACTIONS(2690),
    [anon_sym_EQ_EQ] = ACTIONS(2688),
    [anon_sym_BANG_EQ] = ACTIONS(2690),
    [anon_sym_eq] = ACTIONS(2688),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2690),
    [anon_sym_equ] = ACTIONS(2690),
    [anon_sym_eqr] = ACTIONS(2690),
    [anon_sym_ne] = ACTIONS(2690),
    [anon_sym_LT_EQ_GT] = ACTIONS(2690),
    [anon_sym_cmp] = ACTIONS(2690),
    [anon_sym_TILDE_TILDE] = ACTIONS(2690),
    [anon_sym_LT] = ACTIONS(2688),
    [anon_sym_LT_EQ] = ACTIONS(2688),
    [anon_sym_GT_EQ] = ACTIONS(2690),
    [anon_sym_lt] = ACTIONS(2690),
    [anon_sym_le] = ACTIONS(2690),
    [anon_sym_ge] = ACTIONS(2690),
    [anon_sym_gt] = ACTIONS(2690),
    [anon_sym_isa] = ACTIONS(2690),
    [anon_sym_PLUS_PLUS] = ACTIONS(2690),
    [anon_sym_DASH_DASH] = ACTIONS(2690),
    [anon_sym_QMARK] = ACTIONS(2690),
    [anon_sym_EQ_GT] = ACTIONS(2690),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2690),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3093)] = {
    [sym_heredoc_content] = STATE(3093),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_RBRACE] = ACTIONS(2686),
    [anon_sym_COLON] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_RPAREN] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3094)] = {
    [sym_heredoc_content] = STATE(3094),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_RBRACE] = ACTIONS(3425),
    [anon_sym_COLON] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_RPAREN] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3095)] = {
    [sym_heredoc_content] = STATE(3095),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_and] = ACTIONS(2858),
    [anon_sym_or] = ACTIONS(2858),
    [anon_sym_xor] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_RBRACK] = ACTIONS(2858),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(2858),
  },
  [STATE(3096)] = {
    [sym_heredoc_content] = STATE(3096),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_RBRACE] = ACTIONS(3431),
    [anon_sym_COLON] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_RPAREN] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3097)] = {
    [sym_heredoc_content] = STATE(3097),
    [anon_sym_COLON] = ACTIONS(4107),
    [anon_sym_DOLLAR] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(4107),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(4107),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3098)] = {
    [sym_heredoc_content] = STATE(3098),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4986),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(4988),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3099)] = {
    [sym_heredoc_content] = STATE(3099),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3100)] = {
    [sym_heredoc_content] = STATE(3100),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3101)] = {
    [sym_heredoc_content] = STATE(3101),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3102)] = {
    [sym_heredoc_content] = STATE(3102),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3103)] = {
    [sym_heredoc_content] = STATE(3103),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3104)] = {
    [sym_heredoc_content] = STATE(3104),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3105)] = {
    [sym_heredoc_content] = STATE(3105),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3106)] = {
    [sym_heredoc_content] = STATE(3106),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3107)] = {
    [sym_heredoc_content] = STATE(3107),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3108)] = {
    [sym_heredoc_content] = STATE(3108),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3109)] = {
    [sym_heredoc_content] = STATE(3109),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3110)] = {
    [sym_heredoc_content] = STATE(3110),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4990),
    [anon_sym_cmp] = ACTIONS(4990),
    [anon_sym_TILDE_TILDE] = ACTIONS(4990),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3111)] = {
    [sym_heredoc_content] = STATE(3111),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4992),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3112)] = {
    [sym_heredoc_content] = STATE(3112),
    [anon_sym_COLON] = ACTIONS(3985),
    [anon_sym_DOLLAR] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_AT] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3987),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3985),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3113)] = {
    [sym_heredoc_content] = STATE(3113),
    [anon_sym_LBRACE] = ACTIONS(4994),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_LPAREN] = ACTIONS(4996),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(4998),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3114)] = {
    [sym_heredoc_content] = STATE(3114),
    [anon_sym_LBRACE] = ACTIONS(3345),
    [anon_sym_SEMI] = ACTIONS(3345),
    [anon_sym_EQ] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_COMMA] = ACTIONS(3345),
    [anon_sym_LPAREN] = ACTIONS(3345),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3345),
    [anon_sym_DASH_GT] = ACTIONS(3345),
    [anon_sym_STAR] = ACTIONS(3347),
    [anon_sym_LBRACK] = ACTIONS(3345),
    [anon_sym_PERCENT] = ACTIONS(3347),
    [anon_sym_GT] = ACTIONS(3347),
    [anon_sym_LT_LT] = ACTIONS(3347),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3345),
    [anon_sym_PLUS_EQ] = ACTIONS(3345),
    [anon_sym_DASH_EQ] = ACTIONS(3345),
    [anon_sym_DOT_EQ] = ACTIONS(3345),
    [anon_sym_STAR_EQ] = ACTIONS(3345),
    [anon_sym_SLASH_EQ] = ACTIONS(3345),
    [anon_sym_PERCENT_EQ] = ACTIONS(3345),
    [anon_sym_x_EQ] = ACTIONS(3345),
    [anon_sym_AMP_EQ] = ACTIONS(3345),
    [anon_sym_PIPE_EQ] = ACTIONS(3345),
    [anon_sym_CARET_EQ] = ACTIONS(3345),
    [anon_sym_LT_LT_EQ] = ACTIONS(3345),
    [anon_sym_GT_GT_EQ] = ACTIONS(3345),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3345),
    [anon_sym_DOT_DOT] = ACTIONS(3347),
    [anon_sym_STAR_STAR] = ACTIONS(3347),
    [anon_sym_PIPE_PIPE] = ACTIONS(3347),
    [anon_sym_SLASH_SLASH] = ACTIONS(3347),
    [anon_sym_CARET_CARET] = ACTIONS(3345),
    [anon_sym_AMP_AMP] = ACTIONS(3347),
    [anon_sym_PIPE] = ACTIONS(3347),
    [anon_sym_CARET] = ACTIONS(3347),
    [anon_sym_AMP] = ACTIONS(3347),
    [anon_sym_GT_GT2] = ACTIONS(3347),
    [anon_sym_PLUS] = ACTIONS(3347),
    [anon_sym_DASH] = ACTIONS(3347),
    [anon_sym_DOT] = ACTIONS(3347),
    [anon_sym_SLASH] = ACTIONS(3347),
    [anon_sym_x] = ACTIONS(3347),
    [anon_sym_EQ_TILDE] = ACTIONS(3345),
    [anon_sym_BANG_TILDE] = ACTIONS(3345),
    [anon_sym_EQ_EQ] = ACTIONS(3347),
    [anon_sym_BANG_EQ] = ACTIONS(3345),
    [anon_sym_eq] = ACTIONS(3347),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3345),
    [anon_sym_equ] = ACTIONS(3345),
    [anon_sym_eqr] = ACTIONS(3345),
    [anon_sym_ne] = ACTIONS(3345),
    [anon_sym_LT_EQ_GT] = ACTIONS(3345),
    [anon_sym_cmp] = ACTIONS(3345),
    [anon_sym_TILDE_TILDE] = ACTIONS(3345),
    [anon_sym_LT] = ACTIONS(3347),
    [anon_sym_LT_EQ] = ACTIONS(3347),
    [anon_sym_GT_EQ] = ACTIONS(3345),
    [anon_sym_lt] = ACTIONS(3345),
    [anon_sym_le] = ACTIONS(3345),
    [anon_sym_ge] = ACTIONS(3345),
    [anon_sym_gt] = ACTIONS(3345),
    [anon_sym_isa] = ACTIONS(3345),
    [anon_sym_PLUS_PLUS] = ACTIONS(3345),
    [anon_sym_DASH_DASH] = ACTIONS(3345),
    [anon_sym_QMARK] = ACTIONS(3345),
    [anon_sym_EQ_GT] = ACTIONS(3345),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3345),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3115)] = {
    [sym_heredoc_content] = STATE(3115),
    [anon_sym_LBRACE] = ACTIONS(3361),
    [anon_sym_SEMI] = ACTIONS(3361),
    [anon_sym_EQ] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_COMMA] = ACTIONS(3361),
    [anon_sym_LPAREN] = ACTIONS(3361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3361),
    [anon_sym_DASH_GT] = ACTIONS(3361),
    [anon_sym_STAR] = ACTIONS(3363),
    [anon_sym_LBRACK] = ACTIONS(3361),
    [anon_sym_PERCENT] = ACTIONS(3363),
    [anon_sym_GT] = ACTIONS(3363),
    [anon_sym_LT_LT] = ACTIONS(3363),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3361),
    [anon_sym_PLUS_EQ] = ACTIONS(3361),
    [anon_sym_DASH_EQ] = ACTIONS(3361),
    [anon_sym_DOT_EQ] = ACTIONS(3361),
    [anon_sym_STAR_EQ] = ACTIONS(3361),
    [anon_sym_SLASH_EQ] = ACTIONS(3361),
    [anon_sym_PERCENT_EQ] = ACTIONS(3361),
    [anon_sym_x_EQ] = ACTIONS(3361),
    [anon_sym_AMP_EQ] = ACTIONS(3361),
    [anon_sym_PIPE_EQ] = ACTIONS(3361),
    [anon_sym_CARET_EQ] = ACTIONS(3361),
    [anon_sym_LT_LT_EQ] = ACTIONS(3361),
    [anon_sym_GT_GT_EQ] = ACTIONS(3361),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3361),
    [anon_sym_DOT_DOT] = ACTIONS(3363),
    [anon_sym_STAR_STAR] = ACTIONS(3363),
    [anon_sym_PIPE_PIPE] = ACTIONS(3363),
    [anon_sym_SLASH_SLASH] = ACTIONS(3363),
    [anon_sym_CARET_CARET] = ACTIONS(3361),
    [anon_sym_AMP_AMP] = ACTIONS(3363),
    [anon_sym_PIPE] = ACTIONS(3363),
    [anon_sym_CARET] = ACTIONS(3363),
    [anon_sym_AMP] = ACTIONS(3363),
    [anon_sym_GT_GT2] = ACTIONS(3363),
    [anon_sym_PLUS] = ACTIONS(3363),
    [anon_sym_DASH] = ACTIONS(3363),
    [anon_sym_DOT] = ACTIONS(3363),
    [anon_sym_SLASH] = ACTIONS(3363),
    [anon_sym_x] = ACTIONS(3363),
    [anon_sym_EQ_TILDE] = ACTIONS(3361),
    [anon_sym_BANG_TILDE] = ACTIONS(3361),
    [anon_sym_EQ_EQ] = ACTIONS(3363),
    [anon_sym_BANG_EQ] = ACTIONS(3361),
    [anon_sym_eq] = ACTIONS(3363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3361),
    [anon_sym_equ] = ACTIONS(3361),
    [anon_sym_eqr] = ACTIONS(3361),
    [anon_sym_ne] = ACTIONS(3361),
    [anon_sym_LT_EQ_GT] = ACTIONS(3361),
    [anon_sym_cmp] = ACTIONS(3361),
    [anon_sym_TILDE_TILDE] = ACTIONS(3361),
    [anon_sym_LT] = ACTIONS(3363),
    [anon_sym_LT_EQ] = ACTIONS(3363),
    [anon_sym_GT_EQ] = ACTIONS(3361),
    [anon_sym_lt] = ACTIONS(3361),
    [anon_sym_le] = ACTIONS(3361),
    [anon_sym_ge] = ACTIONS(3361),
    [anon_sym_gt] = ACTIONS(3361),
    [anon_sym_isa] = ACTIONS(3361),
    [anon_sym_PLUS_PLUS] = ACTIONS(3361),
    [anon_sym_DASH_DASH] = ACTIONS(3361),
    [anon_sym_QMARK] = ACTIONS(3361),
    [anon_sym_EQ_GT] = ACTIONS(3361),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3361),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3116)] = {
    [sym_heredoc_content] = STATE(3116),
    [anon_sym_LBRACE] = ACTIONS(3397),
    [anon_sym_SEMI] = ACTIONS(3397),
    [anon_sym_EQ] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_COMMA] = ACTIONS(3397),
    [anon_sym_LPAREN] = ACTIONS(3397),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3397),
    [anon_sym_DASH_GT] = ACTIONS(3397),
    [anon_sym_STAR] = ACTIONS(3399),
    [anon_sym_LBRACK] = ACTIONS(3397),
    [anon_sym_PERCENT] = ACTIONS(3399),
    [anon_sym_GT] = ACTIONS(3399),
    [anon_sym_LT_LT] = ACTIONS(3399),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3397),
    [anon_sym_PLUS_EQ] = ACTIONS(3397),
    [anon_sym_DASH_EQ] = ACTIONS(3397),
    [anon_sym_DOT_EQ] = ACTIONS(3397),
    [anon_sym_STAR_EQ] = ACTIONS(3397),
    [anon_sym_SLASH_EQ] = ACTIONS(3397),
    [anon_sym_PERCENT_EQ] = ACTIONS(3397),
    [anon_sym_x_EQ] = ACTIONS(3397),
    [anon_sym_AMP_EQ] = ACTIONS(3397),
    [anon_sym_PIPE_EQ] = ACTIONS(3397),
    [anon_sym_CARET_EQ] = ACTIONS(3397),
    [anon_sym_LT_LT_EQ] = ACTIONS(3397),
    [anon_sym_GT_GT_EQ] = ACTIONS(3397),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3397),
    [anon_sym_DOT_DOT] = ACTIONS(3399),
    [anon_sym_STAR_STAR] = ACTIONS(3399),
    [anon_sym_PIPE_PIPE] = ACTIONS(3399),
    [anon_sym_SLASH_SLASH] = ACTIONS(3399),
    [anon_sym_CARET_CARET] = ACTIONS(3397),
    [anon_sym_AMP_AMP] = ACTIONS(3399),
    [anon_sym_PIPE] = ACTIONS(3399),
    [anon_sym_CARET] = ACTIONS(3399),
    [anon_sym_AMP] = ACTIONS(3399),
    [anon_sym_GT_GT2] = ACTIONS(3399),
    [anon_sym_PLUS] = ACTIONS(3399),
    [anon_sym_DASH] = ACTIONS(3399),
    [anon_sym_DOT] = ACTIONS(3399),
    [anon_sym_SLASH] = ACTIONS(3399),
    [anon_sym_x] = ACTIONS(3399),
    [anon_sym_EQ_TILDE] = ACTIONS(3397),
    [anon_sym_BANG_TILDE] = ACTIONS(3397),
    [anon_sym_EQ_EQ] = ACTIONS(3399),
    [anon_sym_BANG_EQ] = ACTIONS(3397),
    [anon_sym_eq] = ACTIONS(3399),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3397),
    [anon_sym_equ] = ACTIONS(3397),
    [anon_sym_eqr] = ACTIONS(3397),
    [anon_sym_ne] = ACTIONS(3397),
    [anon_sym_LT_EQ_GT] = ACTIONS(3397),
    [anon_sym_cmp] = ACTIONS(3397),
    [anon_sym_TILDE_TILDE] = ACTIONS(3397),
    [anon_sym_LT] = ACTIONS(3399),
    [anon_sym_LT_EQ] = ACTIONS(3399),
    [anon_sym_GT_EQ] = ACTIONS(3397),
    [anon_sym_lt] = ACTIONS(3397),
    [anon_sym_le] = ACTIONS(3397),
    [anon_sym_ge] = ACTIONS(3397),
    [anon_sym_gt] = ACTIONS(3397),
    [anon_sym_isa] = ACTIONS(3397),
    [anon_sym_PLUS_PLUS] = ACTIONS(3397),
    [anon_sym_DASH_DASH] = ACTIONS(3397),
    [anon_sym_QMARK] = ACTIONS(3397),
    [anon_sym_EQ_GT] = ACTIONS(3397),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3397),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3117)] = {
    [sym_heredoc_content] = STATE(3117),
    [anon_sym_LBRACE] = ACTIONS(3349),
    [anon_sym_SEMI] = ACTIONS(3349),
    [anon_sym_EQ] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_COMMA] = ACTIONS(3349),
    [anon_sym_LPAREN] = ACTIONS(3349),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3349),
    [anon_sym_DASH_GT] = ACTIONS(3349),
    [anon_sym_STAR] = ACTIONS(3351),
    [anon_sym_LBRACK] = ACTIONS(3349),
    [anon_sym_PERCENT] = ACTIONS(3351),
    [anon_sym_GT] = ACTIONS(3351),
    [anon_sym_LT_LT] = ACTIONS(3351),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3349),
    [anon_sym_PLUS_EQ] = ACTIONS(3349),
    [anon_sym_DASH_EQ] = ACTIONS(3349),
    [anon_sym_DOT_EQ] = ACTIONS(3349),
    [anon_sym_STAR_EQ] = ACTIONS(3349),
    [anon_sym_SLASH_EQ] = ACTIONS(3349),
    [anon_sym_PERCENT_EQ] = ACTIONS(3349),
    [anon_sym_x_EQ] = ACTIONS(3349),
    [anon_sym_AMP_EQ] = ACTIONS(3349),
    [anon_sym_PIPE_EQ] = ACTIONS(3349),
    [anon_sym_CARET_EQ] = ACTIONS(3349),
    [anon_sym_LT_LT_EQ] = ACTIONS(3349),
    [anon_sym_GT_GT_EQ] = ACTIONS(3349),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3349),
    [anon_sym_DOT_DOT] = ACTIONS(3351),
    [anon_sym_STAR_STAR] = ACTIONS(3351),
    [anon_sym_PIPE_PIPE] = ACTIONS(3351),
    [anon_sym_SLASH_SLASH] = ACTIONS(3351),
    [anon_sym_CARET_CARET] = ACTIONS(3349),
    [anon_sym_AMP_AMP] = ACTIONS(3351),
    [anon_sym_PIPE] = ACTIONS(3351),
    [anon_sym_CARET] = ACTIONS(3351),
    [anon_sym_AMP] = ACTIONS(3351),
    [anon_sym_GT_GT2] = ACTIONS(3351),
    [anon_sym_PLUS] = ACTIONS(3351),
    [anon_sym_DASH] = ACTIONS(3351),
    [anon_sym_DOT] = ACTIONS(3351),
    [anon_sym_SLASH] = ACTIONS(3351),
    [anon_sym_x] = ACTIONS(3351),
    [anon_sym_EQ_TILDE] = ACTIONS(3349),
    [anon_sym_BANG_TILDE] = ACTIONS(3349),
    [anon_sym_EQ_EQ] = ACTIONS(3351),
    [anon_sym_BANG_EQ] = ACTIONS(3349),
    [anon_sym_eq] = ACTIONS(3351),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3349),
    [anon_sym_equ] = ACTIONS(3349),
    [anon_sym_eqr] = ACTIONS(3349),
    [anon_sym_ne] = ACTIONS(3349),
    [anon_sym_LT_EQ_GT] = ACTIONS(3349),
    [anon_sym_cmp] = ACTIONS(3349),
    [anon_sym_TILDE_TILDE] = ACTIONS(3349),
    [anon_sym_LT] = ACTIONS(3351),
    [anon_sym_LT_EQ] = ACTIONS(3351),
    [anon_sym_GT_EQ] = ACTIONS(3349),
    [anon_sym_lt] = ACTIONS(3349),
    [anon_sym_le] = ACTIONS(3349),
    [anon_sym_ge] = ACTIONS(3349),
    [anon_sym_gt] = ACTIONS(3349),
    [anon_sym_isa] = ACTIONS(3349),
    [anon_sym_PLUS_PLUS] = ACTIONS(3349),
    [anon_sym_DASH_DASH] = ACTIONS(3349),
    [anon_sym_QMARK] = ACTIONS(3349),
    [anon_sym_EQ_GT] = ACTIONS(3349),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3349),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3118)] = {
    [sym_heredoc_content] = STATE(3118),
    [anon_sym_LBRACE] = ACTIONS(3365),
    [anon_sym_SEMI] = ACTIONS(3365),
    [anon_sym_EQ] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_COMMA] = ACTIONS(3365),
    [anon_sym_LPAREN] = ACTIONS(3365),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3365),
    [anon_sym_DASH_GT] = ACTIONS(3365),
    [anon_sym_STAR] = ACTIONS(3367),
    [anon_sym_LBRACK] = ACTIONS(3365),
    [anon_sym_PERCENT] = ACTIONS(3367),
    [anon_sym_GT] = ACTIONS(3367),
    [anon_sym_LT_LT] = ACTIONS(3367),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3365),
    [anon_sym_PLUS_EQ] = ACTIONS(3365),
    [anon_sym_DASH_EQ] = ACTIONS(3365),
    [anon_sym_DOT_EQ] = ACTIONS(3365),
    [anon_sym_STAR_EQ] = ACTIONS(3365),
    [anon_sym_SLASH_EQ] = ACTIONS(3365),
    [anon_sym_PERCENT_EQ] = ACTIONS(3365),
    [anon_sym_x_EQ] = ACTIONS(3365),
    [anon_sym_AMP_EQ] = ACTIONS(3365),
    [anon_sym_PIPE_EQ] = ACTIONS(3365),
    [anon_sym_CARET_EQ] = ACTIONS(3365),
    [anon_sym_LT_LT_EQ] = ACTIONS(3365),
    [anon_sym_GT_GT_EQ] = ACTIONS(3365),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3365),
    [anon_sym_DOT_DOT] = ACTIONS(3367),
    [anon_sym_STAR_STAR] = ACTIONS(3367),
    [anon_sym_PIPE_PIPE] = ACTIONS(3367),
    [anon_sym_SLASH_SLASH] = ACTIONS(3367),
    [anon_sym_CARET_CARET] = ACTIONS(3365),
    [anon_sym_AMP_AMP] = ACTIONS(3367),
    [anon_sym_PIPE] = ACTIONS(3367),
    [anon_sym_CARET] = ACTIONS(3367),
    [anon_sym_AMP] = ACTIONS(3367),
    [anon_sym_GT_GT2] = ACTIONS(3367),
    [anon_sym_PLUS] = ACTIONS(3367),
    [anon_sym_DASH] = ACTIONS(3367),
    [anon_sym_DOT] = ACTIONS(3367),
    [anon_sym_SLASH] = ACTIONS(3367),
    [anon_sym_x] = ACTIONS(3367),
    [anon_sym_EQ_TILDE] = ACTIONS(3365),
    [anon_sym_BANG_TILDE] = ACTIONS(3365),
    [anon_sym_EQ_EQ] = ACTIONS(3367),
    [anon_sym_BANG_EQ] = ACTIONS(3365),
    [anon_sym_eq] = ACTIONS(3367),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3365),
    [anon_sym_equ] = ACTIONS(3365),
    [anon_sym_eqr] = ACTIONS(3365),
    [anon_sym_ne] = ACTIONS(3365),
    [anon_sym_LT_EQ_GT] = ACTIONS(3365),
    [anon_sym_cmp] = ACTIONS(3365),
    [anon_sym_TILDE_TILDE] = ACTIONS(3365),
    [anon_sym_LT] = ACTIONS(3367),
    [anon_sym_LT_EQ] = ACTIONS(3367),
    [anon_sym_GT_EQ] = ACTIONS(3365),
    [anon_sym_lt] = ACTIONS(3365),
    [anon_sym_le] = ACTIONS(3365),
    [anon_sym_ge] = ACTIONS(3365),
    [anon_sym_gt] = ACTIONS(3365),
    [anon_sym_isa] = ACTIONS(3365),
    [anon_sym_PLUS_PLUS] = ACTIONS(3365),
    [anon_sym_DASH_DASH] = ACTIONS(3365),
    [anon_sym_QMARK] = ACTIONS(3365),
    [anon_sym_EQ_GT] = ACTIONS(3365),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3365),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3119)] = {
    [sym_heredoc_content] = STATE(3119),
    [anon_sym_LBRACE] = ACTIONS(3385),
    [anon_sym_SEMI] = ACTIONS(3385),
    [anon_sym_EQ] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_COMMA] = ACTIONS(3385),
    [anon_sym_LPAREN] = ACTIONS(3385),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3385),
    [anon_sym_DASH_GT] = ACTIONS(3385),
    [anon_sym_STAR] = ACTIONS(3387),
    [anon_sym_LBRACK] = ACTIONS(3385),
    [anon_sym_PERCENT] = ACTIONS(3387),
    [anon_sym_GT] = ACTIONS(3387),
    [anon_sym_LT_LT] = ACTIONS(3387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3385),
    [anon_sym_PLUS_EQ] = ACTIONS(3385),
    [anon_sym_DASH_EQ] = ACTIONS(3385),
    [anon_sym_DOT_EQ] = ACTIONS(3385),
    [anon_sym_STAR_EQ] = ACTIONS(3385),
    [anon_sym_SLASH_EQ] = ACTIONS(3385),
    [anon_sym_PERCENT_EQ] = ACTIONS(3385),
    [anon_sym_x_EQ] = ACTIONS(3385),
    [anon_sym_AMP_EQ] = ACTIONS(3385),
    [anon_sym_PIPE_EQ] = ACTIONS(3385),
    [anon_sym_CARET_EQ] = ACTIONS(3385),
    [anon_sym_LT_LT_EQ] = ACTIONS(3385),
    [anon_sym_GT_GT_EQ] = ACTIONS(3385),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3385),
    [anon_sym_DOT_DOT] = ACTIONS(3387),
    [anon_sym_STAR_STAR] = ACTIONS(3387),
    [anon_sym_PIPE_PIPE] = ACTIONS(3387),
    [anon_sym_SLASH_SLASH] = ACTIONS(3387),
    [anon_sym_CARET_CARET] = ACTIONS(3385),
    [anon_sym_AMP_AMP] = ACTIONS(3387),
    [anon_sym_PIPE] = ACTIONS(3387),
    [anon_sym_CARET] = ACTIONS(3387),
    [anon_sym_AMP] = ACTIONS(3387),
    [anon_sym_GT_GT2] = ACTIONS(3387),
    [anon_sym_PLUS] = ACTIONS(3387),
    [anon_sym_DASH] = ACTIONS(3387),
    [anon_sym_DOT] = ACTIONS(3387),
    [anon_sym_SLASH] = ACTIONS(3387),
    [anon_sym_x] = ACTIONS(3387),
    [anon_sym_EQ_TILDE] = ACTIONS(3385),
    [anon_sym_BANG_TILDE] = ACTIONS(3385),
    [anon_sym_EQ_EQ] = ACTIONS(3387),
    [anon_sym_BANG_EQ] = ACTIONS(3385),
    [anon_sym_eq] = ACTIONS(3387),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3385),
    [anon_sym_equ] = ACTIONS(3385),
    [anon_sym_eqr] = ACTIONS(3385),
    [anon_sym_ne] = ACTIONS(3385),
    [anon_sym_LT_EQ_GT] = ACTIONS(3385),
    [anon_sym_cmp] = ACTIONS(3385),
    [anon_sym_TILDE_TILDE] = ACTIONS(3385),
    [anon_sym_LT] = ACTIONS(3387),
    [anon_sym_LT_EQ] = ACTIONS(3387),
    [anon_sym_GT_EQ] = ACTIONS(3385),
    [anon_sym_lt] = ACTIONS(3385),
    [anon_sym_le] = ACTIONS(3385),
    [anon_sym_ge] = ACTIONS(3385),
    [anon_sym_gt] = ACTIONS(3385),
    [anon_sym_isa] = ACTIONS(3385),
    [anon_sym_PLUS_PLUS] = ACTIONS(3385),
    [anon_sym_DASH_DASH] = ACTIONS(3385),
    [anon_sym_QMARK] = ACTIONS(3385),
    [anon_sym_EQ_GT] = ACTIONS(3385),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3385),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3120)] = {
    [sym_heredoc_content] = STATE(3120),
    [anon_sym_LBRACE] = ACTIONS(3389),
    [anon_sym_SEMI] = ACTIONS(3389),
    [anon_sym_EQ] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_COMMA] = ACTIONS(3389),
    [anon_sym_LPAREN] = ACTIONS(3389),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3389),
    [anon_sym_DASH_GT] = ACTIONS(3389),
    [anon_sym_STAR] = ACTIONS(3391),
    [anon_sym_LBRACK] = ACTIONS(3389),
    [anon_sym_PERCENT] = ACTIONS(3391),
    [anon_sym_GT] = ACTIONS(3391),
    [anon_sym_LT_LT] = ACTIONS(3391),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3389),
    [anon_sym_PLUS_EQ] = ACTIONS(3389),
    [anon_sym_DASH_EQ] = ACTIONS(3389),
    [anon_sym_DOT_EQ] = ACTIONS(3389),
    [anon_sym_STAR_EQ] = ACTIONS(3389),
    [anon_sym_SLASH_EQ] = ACTIONS(3389),
    [anon_sym_PERCENT_EQ] = ACTIONS(3389),
    [anon_sym_x_EQ] = ACTIONS(3389),
    [anon_sym_AMP_EQ] = ACTIONS(3389),
    [anon_sym_PIPE_EQ] = ACTIONS(3389),
    [anon_sym_CARET_EQ] = ACTIONS(3389),
    [anon_sym_LT_LT_EQ] = ACTIONS(3389),
    [anon_sym_GT_GT_EQ] = ACTIONS(3389),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3389),
    [anon_sym_DOT_DOT] = ACTIONS(3391),
    [anon_sym_STAR_STAR] = ACTIONS(3391),
    [anon_sym_PIPE_PIPE] = ACTIONS(3391),
    [anon_sym_SLASH_SLASH] = ACTIONS(3391),
    [anon_sym_CARET_CARET] = ACTIONS(3389),
    [anon_sym_AMP_AMP] = ACTIONS(3391),
    [anon_sym_PIPE] = ACTIONS(3391),
    [anon_sym_CARET] = ACTIONS(3391),
    [anon_sym_AMP] = ACTIONS(3391),
    [anon_sym_GT_GT2] = ACTIONS(3391),
    [anon_sym_PLUS] = ACTIONS(3391),
    [anon_sym_DASH] = ACTIONS(3391),
    [anon_sym_DOT] = ACTIONS(3391),
    [anon_sym_SLASH] = ACTIONS(3391),
    [anon_sym_x] = ACTIONS(3391),
    [anon_sym_EQ_TILDE] = ACTIONS(3389),
    [anon_sym_BANG_TILDE] = ACTIONS(3389),
    [anon_sym_EQ_EQ] = ACTIONS(3391),
    [anon_sym_BANG_EQ] = ACTIONS(3389),
    [anon_sym_eq] = ACTIONS(3391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3389),
    [anon_sym_equ] = ACTIONS(3389),
    [anon_sym_eqr] = ACTIONS(3389),
    [anon_sym_ne] = ACTIONS(3389),
    [anon_sym_LT_EQ_GT] = ACTIONS(3389),
    [anon_sym_cmp] = ACTIONS(3389),
    [anon_sym_TILDE_TILDE] = ACTIONS(3389),
    [anon_sym_LT] = ACTIONS(3391),
    [anon_sym_LT_EQ] = ACTIONS(3391),
    [anon_sym_GT_EQ] = ACTIONS(3389),
    [anon_sym_lt] = ACTIONS(3389),
    [anon_sym_le] = ACTIONS(3389),
    [anon_sym_ge] = ACTIONS(3389),
    [anon_sym_gt] = ACTIONS(3389),
    [anon_sym_isa] = ACTIONS(3389),
    [anon_sym_PLUS_PLUS] = ACTIONS(3389),
    [anon_sym_DASH_DASH] = ACTIONS(3389),
    [anon_sym_QMARK] = ACTIONS(3389),
    [anon_sym_EQ_GT] = ACTIONS(3389),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3389),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3121)] = {
    [sym_heredoc_content] = STATE(3121),
    [anon_sym_COLON] = ACTIONS(3977),
    [anon_sym_DOLLAR] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_AT] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3979),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_PERCENT2] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3122)] = {
    [sym_heredoc_content] = STATE(3122),
    [anon_sym_COLON] = ACTIONS(3997),
    [anon_sym_DOLLAR] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_AT] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3999),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_PERCENT2] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3123)] = {
    [sym_heredoc_content] = STATE(3123),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_DOLLAR] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_AT] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3697),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_PERCENT2] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3124)] = {
    [sym_heredoc_content] = STATE(3124),
    [anon_sym_COLON] = ACTIONS(3909),
    [anon_sym_DOLLAR] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_AT] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3911),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_PERCENT2] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3125)] = {
    [sym_heredoc_content] = STATE(3125),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_DOLLAR] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_AT] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4019),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_PERCENT2] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3126)] = {
    [sym_heredoc_content] = STATE(3126),
    [anon_sym_COLON] = ACTIONS(3803),
    [anon_sym_DOLLAR] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3803),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3803),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3127)] = {
    [sym_heredoc_content] = STATE(3127),
    [anon_sym_COLON] = ACTIONS(3845),
    [anon_sym_DOLLAR] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_AT] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3847),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_PERCENT2] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3128)] = {
    [sym_heredoc_content] = STATE(3128),
    [anon_sym_COLON] = ACTIONS(3739),
    [anon_sym_DOLLAR] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_AT] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3741),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_PERCENT2] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3129)] = {
    [sym_heredoc_content] = STATE(3129),
    [anon_sym_COLON] = ACTIONS(3897),
    [anon_sym_DOLLAR] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(4624),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_AT] = ACTIONS(3897),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4631),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4637),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4626),
    [anon_sym_PLUS_EQ] = ACTIONS(4626),
    [anon_sym_DASH_EQ] = ACTIONS(4626),
    [anon_sym_DOT_EQ] = ACTIONS(4626),
    [anon_sym_STAR_EQ] = ACTIONS(4626),
    [anon_sym_SLASH_EQ] = ACTIONS(4626),
    [anon_sym_PERCENT_EQ] = ACTIONS(4624),
    [anon_sym_x_EQ] = ACTIONS(4626),
    [anon_sym_AMP_EQ] = ACTIONS(4626),
    [anon_sym_PIPE_EQ] = ACTIONS(4626),
    [anon_sym_CARET_EQ] = ACTIONS(4626),
    [anon_sym_LT_LT_EQ] = ACTIONS(4626),
    [anon_sym_GT_GT_EQ] = ACTIONS(4626),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4626),
    [anon_sym_DOT_DOT] = ACTIONS(4641),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4645),
    [anon_sym_SLASH_SLASH] = ACTIONS(4645),
    [anon_sym_CARET_CARET] = ACTIONS(4647),
    [anon_sym_AMP_AMP] = ACTIONS(4649),
    [anon_sym_PIPE] = ACTIONS(4651),
    [anon_sym_CARET] = ACTIONS(4651),
    [anon_sym_AMP] = ACTIONS(4653),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4659),
    [anon_sym_BANG_EQ] = ACTIONS(4661),
    [anon_sym_eq] = ACTIONS(4659),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4661),
    [anon_sym_equ] = ACTIONS(4661),
    [anon_sym_eqr] = ACTIONS(4661),
    [anon_sym_ne] = ACTIONS(4661),
    [anon_sym_LT_EQ_GT] = ACTIONS(4663),
    [anon_sym_cmp] = ACTIONS(4663),
    [anon_sym_TILDE_TILDE] = ACTIONS(4663),
    [anon_sym_LT] = ACTIONS(4637),
    [anon_sym_LT_EQ] = ACTIONS(4637),
    [anon_sym_GT_EQ] = ACTIONS(4665),
    [anon_sym_lt] = ACTIONS(4665),
    [anon_sym_le] = ACTIONS(4665),
    [anon_sym_ge] = ACTIONS(4665),
    [anon_sym_gt] = ACTIONS(4665),
    [anon_sym_isa] = ACTIONS(4667),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4671),
    [anon_sym_PERCENT2] = ACTIONS(3897),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3130)] = {
    [sym_heredoc_content] = STATE(3130),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_DOLLAR] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_AT] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5000),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3821),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_PERCENT2] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5000),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3131)] = {
    [sym_heredoc_content] = STATE(3131),
    [anon_sym_LBRACE] = ACTIONS(3405),
    [anon_sym_SEMI] = ACTIONS(3405),
    [anon_sym_EQ] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_COMMA] = ACTIONS(3405),
    [anon_sym_LPAREN] = ACTIONS(3405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3405),
    [anon_sym_DASH_GT] = ACTIONS(3405),
    [anon_sym_STAR] = ACTIONS(3407),
    [anon_sym_LBRACK] = ACTIONS(3405),
    [anon_sym_PERCENT] = ACTIONS(3407),
    [anon_sym_GT] = ACTIONS(3407),
    [anon_sym_LT_LT] = ACTIONS(3407),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3405),
    [anon_sym_PLUS_EQ] = ACTIONS(3405),
    [anon_sym_DASH_EQ] = ACTIONS(3405),
    [anon_sym_DOT_EQ] = ACTIONS(3405),
    [anon_sym_STAR_EQ] = ACTIONS(3405),
    [anon_sym_SLASH_EQ] = ACTIONS(3405),
    [anon_sym_PERCENT_EQ] = ACTIONS(3405),
    [anon_sym_x_EQ] = ACTIONS(3405),
    [anon_sym_AMP_EQ] = ACTIONS(3405),
    [anon_sym_PIPE_EQ] = ACTIONS(3405),
    [anon_sym_CARET_EQ] = ACTIONS(3405),
    [anon_sym_LT_LT_EQ] = ACTIONS(3405),
    [anon_sym_GT_GT_EQ] = ACTIONS(3405),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3405),
    [anon_sym_DOT_DOT] = ACTIONS(3407),
    [anon_sym_STAR_STAR] = ACTIONS(3407),
    [anon_sym_PIPE_PIPE] = ACTIONS(3407),
    [anon_sym_SLASH_SLASH] = ACTIONS(3407),
    [anon_sym_CARET_CARET] = ACTIONS(3405),
    [anon_sym_AMP_AMP] = ACTIONS(3407),
    [anon_sym_PIPE] = ACTIONS(3407),
    [anon_sym_CARET] = ACTIONS(3407),
    [anon_sym_AMP] = ACTIONS(3407),
    [anon_sym_GT_GT2] = ACTIONS(3407),
    [anon_sym_PLUS] = ACTIONS(3407),
    [anon_sym_DASH] = ACTIONS(3407),
    [anon_sym_DOT] = ACTIONS(3407),
    [anon_sym_SLASH] = ACTIONS(3407),
    [anon_sym_x] = ACTIONS(3407),
    [anon_sym_EQ_TILDE] = ACTIONS(3405),
    [anon_sym_BANG_TILDE] = ACTIONS(3405),
    [anon_sym_EQ_EQ] = ACTIONS(3407),
    [anon_sym_BANG_EQ] = ACTIONS(3405),
    [anon_sym_eq] = ACTIONS(3407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3405),
    [anon_sym_equ] = ACTIONS(3405),
    [anon_sym_eqr] = ACTIONS(3405),
    [anon_sym_ne] = ACTIONS(3405),
    [anon_sym_LT_EQ_GT] = ACTIONS(3405),
    [anon_sym_cmp] = ACTIONS(3405),
    [anon_sym_TILDE_TILDE] = ACTIONS(3405),
    [anon_sym_LT] = ACTIONS(3407),
    [anon_sym_LT_EQ] = ACTIONS(3407),
    [anon_sym_GT_EQ] = ACTIONS(3405),
    [anon_sym_lt] = ACTIONS(3405),
    [anon_sym_le] = ACTIONS(3405),
    [anon_sym_ge] = ACTIONS(3405),
    [anon_sym_gt] = ACTIONS(3405),
    [anon_sym_isa] = ACTIONS(3405),
    [anon_sym_PLUS_PLUS] = ACTIONS(3405),
    [anon_sym_DASH_DASH] = ACTIONS(3405),
    [anon_sym_QMARK] = ACTIONS(3405),
    [anon_sym_EQ_GT] = ACTIONS(3405),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3405),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3132)] = {
    [sym_heredoc_content] = STATE(3132),
    [anon_sym_COLON] = ACTIONS(4151),
    [anon_sym_DOLLAR] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_AT] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(4635),
    [anon_sym_PERCENT] = ACTIONS(4635),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(4639),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4153),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(4643),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(4639),
    [anon_sym_PLUS] = ACTIONS(4655),
    [anon_sym_DASH] = ACTIONS(4655),
    [anon_sym_DOT] = ACTIONS(4655),
    [anon_sym_SLASH] = ACTIONS(4635),
    [anon_sym_x] = ACTIONS(4635),
    [anon_sym_EQ_TILDE] = ACTIONS(4657),
    [anon_sym_BANG_TILDE] = ACTIONS(4657),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_PERCENT2] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3133)] = {
    [sym_heredoc_content] = STATE(3133),
    [anon_sym_LBRACE] = ACTIONS(3373),
    [anon_sym_SEMI] = ACTIONS(3373),
    [anon_sym_EQ] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_COMMA] = ACTIONS(3373),
    [anon_sym_LPAREN] = ACTIONS(3373),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3373),
    [anon_sym_DASH_GT] = ACTIONS(3373),
    [anon_sym_STAR] = ACTIONS(3375),
    [anon_sym_LBRACK] = ACTIONS(3373),
    [anon_sym_PERCENT] = ACTIONS(3375),
    [anon_sym_GT] = ACTIONS(3375),
    [anon_sym_LT_LT] = ACTIONS(3375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3373),
    [anon_sym_PLUS_EQ] = ACTIONS(3373),
    [anon_sym_DASH_EQ] = ACTIONS(3373),
    [anon_sym_DOT_EQ] = ACTIONS(3373),
    [anon_sym_STAR_EQ] = ACTIONS(3373),
    [anon_sym_SLASH_EQ] = ACTIONS(3373),
    [anon_sym_PERCENT_EQ] = ACTIONS(3373),
    [anon_sym_x_EQ] = ACTIONS(3373),
    [anon_sym_AMP_EQ] = ACTIONS(3373),
    [anon_sym_PIPE_EQ] = ACTIONS(3373),
    [anon_sym_CARET_EQ] = ACTIONS(3373),
    [anon_sym_LT_LT_EQ] = ACTIONS(3373),
    [anon_sym_GT_GT_EQ] = ACTIONS(3373),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3373),
    [anon_sym_DOT_DOT] = ACTIONS(3375),
    [anon_sym_STAR_STAR] = ACTIONS(3375),
    [anon_sym_PIPE_PIPE] = ACTIONS(3375),
    [anon_sym_SLASH_SLASH] = ACTIONS(3375),
    [anon_sym_CARET_CARET] = ACTIONS(3373),
    [anon_sym_AMP_AMP] = ACTIONS(3375),
    [anon_sym_PIPE] = ACTIONS(3375),
    [anon_sym_CARET] = ACTIONS(3375),
    [anon_sym_AMP] = ACTIONS(3375),
    [anon_sym_GT_GT2] = ACTIONS(3375),
    [anon_sym_PLUS] = ACTIONS(3375),
    [anon_sym_DASH] = ACTIONS(3375),
    [anon_sym_DOT] = ACTIONS(3375),
    [anon_sym_SLASH] = ACTIONS(3375),
    [anon_sym_x] = ACTIONS(3375),
    [anon_sym_EQ_TILDE] = ACTIONS(3373),
    [anon_sym_BANG_TILDE] = ACTIONS(3373),
    [anon_sym_EQ_EQ] = ACTIONS(3375),
    [anon_sym_BANG_EQ] = ACTIONS(3373),
    [anon_sym_eq] = ACTIONS(3375),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3373),
    [anon_sym_equ] = ACTIONS(3373),
    [anon_sym_eqr] = ACTIONS(3373),
    [anon_sym_ne] = ACTIONS(3373),
    [anon_sym_LT_EQ_GT] = ACTIONS(3373),
    [anon_sym_cmp] = ACTIONS(3373),
    [anon_sym_TILDE_TILDE] = ACTIONS(3373),
    [anon_sym_LT] = ACTIONS(3375),
    [anon_sym_LT_EQ] = ACTIONS(3375),
    [anon_sym_GT_EQ] = ACTIONS(3373),
    [anon_sym_lt] = ACTIONS(3373),
    [anon_sym_le] = ACTIONS(3373),
    [anon_sym_ge] = ACTIONS(3373),
    [anon_sym_gt] = ACTIONS(3373),
    [anon_sym_isa] = ACTIONS(3373),
    [anon_sym_PLUS_PLUS] = ACTIONS(3373),
    [anon_sym_DASH_DASH] = ACTIONS(3373),
    [anon_sym_QMARK] = ACTIONS(3373),
    [anon_sym_EQ_GT] = ACTIONS(3373),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3373),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3134)] = {
    [sym_heredoc_content] = STATE(3134),
    [anon_sym_LBRACE] = ACTIONS(3381),
    [anon_sym_SEMI] = ACTIONS(3381),
    [anon_sym_EQ] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_COMMA] = ACTIONS(3381),
    [anon_sym_LPAREN] = ACTIONS(3381),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3381),
    [anon_sym_DASH_GT] = ACTIONS(3381),
    [anon_sym_STAR] = ACTIONS(3383),
    [anon_sym_LBRACK] = ACTIONS(3381),
    [anon_sym_PERCENT] = ACTIONS(3383),
    [anon_sym_GT] = ACTIONS(3383),
    [anon_sym_LT_LT] = ACTIONS(3383),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3381),
    [anon_sym_PLUS_EQ] = ACTIONS(3381),
    [anon_sym_DASH_EQ] = ACTIONS(3381),
    [anon_sym_DOT_EQ] = ACTIONS(3381),
    [anon_sym_STAR_EQ] = ACTIONS(3381),
    [anon_sym_SLASH_EQ] = ACTIONS(3381),
    [anon_sym_PERCENT_EQ] = ACTIONS(3381),
    [anon_sym_x_EQ] = ACTIONS(3381),
    [anon_sym_AMP_EQ] = ACTIONS(3381),
    [anon_sym_PIPE_EQ] = ACTIONS(3381),
    [anon_sym_CARET_EQ] = ACTIONS(3381),
    [anon_sym_LT_LT_EQ] = ACTIONS(3381),
    [anon_sym_GT_GT_EQ] = ACTIONS(3381),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3381),
    [anon_sym_DOT_DOT] = ACTIONS(3383),
    [anon_sym_STAR_STAR] = ACTIONS(3383),
    [anon_sym_PIPE_PIPE] = ACTIONS(3383),
    [anon_sym_SLASH_SLASH] = ACTIONS(3383),
    [anon_sym_CARET_CARET] = ACTIONS(3381),
    [anon_sym_AMP_AMP] = ACTIONS(3383),
    [anon_sym_PIPE] = ACTIONS(3383),
    [anon_sym_CARET] = ACTIONS(3383),
    [anon_sym_AMP] = ACTIONS(3383),
    [anon_sym_GT_GT2] = ACTIONS(3383),
    [anon_sym_PLUS] = ACTIONS(3383),
    [anon_sym_DASH] = ACTIONS(3383),
    [anon_sym_DOT] = ACTIONS(3383),
    [anon_sym_SLASH] = ACTIONS(3383),
    [anon_sym_x] = ACTIONS(3383),
    [anon_sym_EQ_TILDE] = ACTIONS(3381),
    [anon_sym_BANG_TILDE] = ACTIONS(3381),
    [anon_sym_EQ_EQ] = ACTIONS(3383),
    [anon_sym_BANG_EQ] = ACTIONS(3381),
    [anon_sym_eq] = ACTIONS(3383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3381),
    [anon_sym_equ] = ACTIONS(3381),
    [anon_sym_eqr] = ACTIONS(3381),
    [anon_sym_ne] = ACTIONS(3381),
    [anon_sym_LT_EQ_GT] = ACTIONS(3381),
    [anon_sym_cmp] = ACTIONS(3381),
    [anon_sym_TILDE_TILDE] = ACTIONS(3381),
    [anon_sym_LT] = ACTIONS(3383),
    [anon_sym_LT_EQ] = ACTIONS(3383),
    [anon_sym_GT_EQ] = ACTIONS(3381),
    [anon_sym_lt] = ACTIONS(3381),
    [anon_sym_le] = ACTIONS(3381),
    [anon_sym_ge] = ACTIONS(3381),
    [anon_sym_gt] = ACTIONS(3381),
    [anon_sym_isa] = ACTIONS(3381),
    [anon_sym_PLUS_PLUS] = ACTIONS(3381),
    [anon_sym_DASH_DASH] = ACTIONS(3381),
    [anon_sym_QMARK] = ACTIONS(3381),
    [anon_sym_EQ_GT] = ACTIONS(3381),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3381),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3135)] = {
    [sym_heredoc_content] = STATE(3135),
    [anon_sym_LBRACE] = ACTIONS(3341),
    [anon_sym_SEMI] = ACTIONS(3341),
    [anon_sym_EQ] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_COMMA] = ACTIONS(3341),
    [anon_sym_LPAREN] = ACTIONS(3341),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3341),
    [anon_sym_DASH_GT] = ACTIONS(3341),
    [anon_sym_STAR] = ACTIONS(3343),
    [anon_sym_LBRACK] = ACTIONS(3341),
    [anon_sym_PERCENT] = ACTIONS(3343),
    [anon_sym_GT] = ACTIONS(3343),
    [anon_sym_LT_LT] = ACTIONS(3343),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3341),
    [anon_sym_PLUS_EQ] = ACTIONS(3341),
    [anon_sym_DASH_EQ] = ACTIONS(3341),
    [anon_sym_DOT_EQ] = ACTIONS(3341),
    [anon_sym_STAR_EQ] = ACTIONS(3341),
    [anon_sym_SLASH_EQ] = ACTIONS(3341),
    [anon_sym_PERCENT_EQ] = ACTIONS(3341),
    [anon_sym_x_EQ] = ACTIONS(3341),
    [anon_sym_AMP_EQ] = ACTIONS(3341),
    [anon_sym_PIPE_EQ] = ACTIONS(3341),
    [anon_sym_CARET_EQ] = ACTIONS(3341),
    [anon_sym_LT_LT_EQ] = ACTIONS(3341),
    [anon_sym_GT_GT_EQ] = ACTIONS(3341),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3341),
    [anon_sym_DOT_DOT] = ACTIONS(3343),
    [anon_sym_STAR_STAR] = ACTIONS(3343),
    [anon_sym_PIPE_PIPE] = ACTIONS(3343),
    [anon_sym_SLASH_SLASH] = ACTIONS(3343),
    [anon_sym_CARET_CARET] = ACTIONS(3341),
    [anon_sym_AMP_AMP] = ACTIONS(3343),
    [anon_sym_PIPE] = ACTIONS(3343),
    [anon_sym_CARET] = ACTIONS(3343),
    [anon_sym_AMP] = ACTIONS(3343),
    [anon_sym_GT_GT2] = ACTIONS(3343),
    [anon_sym_PLUS] = ACTIONS(3343),
    [anon_sym_DASH] = ACTIONS(3343),
    [anon_sym_DOT] = ACTIONS(3343),
    [anon_sym_SLASH] = ACTIONS(3343),
    [anon_sym_x] = ACTIONS(3343),
    [anon_sym_EQ_TILDE] = ACTIONS(3341),
    [anon_sym_BANG_TILDE] = ACTIONS(3341),
    [anon_sym_EQ_EQ] = ACTIONS(3343),
    [anon_sym_BANG_EQ] = ACTIONS(3341),
    [anon_sym_eq] = ACTIONS(3343),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3341),
    [anon_sym_equ] = ACTIONS(3341),
    [anon_sym_eqr] = ACTIONS(3341),
    [anon_sym_ne] = ACTIONS(3341),
    [anon_sym_LT_EQ_GT] = ACTIONS(3341),
    [anon_sym_cmp] = ACTIONS(3341),
    [anon_sym_TILDE_TILDE] = ACTIONS(3341),
    [anon_sym_LT] = ACTIONS(3343),
    [anon_sym_LT_EQ] = ACTIONS(3343),
    [anon_sym_GT_EQ] = ACTIONS(3341),
    [anon_sym_lt] = ACTIONS(3341),
    [anon_sym_le] = ACTIONS(3341),
    [anon_sym_ge] = ACTIONS(3341),
    [anon_sym_gt] = ACTIONS(3341),
    [anon_sym_isa] = ACTIONS(3341),
    [anon_sym_PLUS_PLUS] = ACTIONS(3341),
    [anon_sym_DASH_DASH] = ACTIONS(3341),
    [anon_sym_QMARK] = ACTIONS(3341),
    [anon_sym_EQ_GT] = ACTIONS(3341),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3341),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3136)] = {
    [sym_heredoc_content] = STATE(3136),
    [anon_sym_LBRACE] = ACTIONS(3353),
    [anon_sym_SEMI] = ACTIONS(3353),
    [anon_sym_EQ] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_COMMA] = ACTIONS(3353),
    [anon_sym_LPAREN] = ACTIONS(3353),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3353),
    [anon_sym_DASH_GT] = ACTIONS(3353),
    [anon_sym_STAR] = ACTIONS(3355),
    [anon_sym_LBRACK] = ACTIONS(3353),
    [anon_sym_PERCENT] = ACTIONS(3355),
    [anon_sym_GT] = ACTIONS(3355),
    [anon_sym_LT_LT] = ACTIONS(3355),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3353),
    [anon_sym_PLUS_EQ] = ACTIONS(3353),
    [anon_sym_DASH_EQ] = ACTIONS(3353),
    [anon_sym_DOT_EQ] = ACTIONS(3353),
    [anon_sym_STAR_EQ] = ACTIONS(3353),
    [anon_sym_SLASH_EQ] = ACTIONS(3353),
    [anon_sym_PERCENT_EQ] = ACTIONS(3353),
    [anon_sym_x_EQ] = ACTIONS(3353),
    [anon_sym_AMP_EQ] = ACTIONS(3353),
    [anon_sym_PIPE_EQ] = ACTIONS(3353),
    [anon_sym_CARET_EQ] = ACTIONS(3353),
    [anon_sym_LT_LT_EQ] = ACTIONS(3353),
    [anon_sym_GT_GT_EQ] = ACTIONS(3353),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3353),
    [anon_sym_DOT_DOT] = ACTIONS(3355),
    [anon_sym_STAR_STAR] = ACTIONS(3355),
    [anon_sym_PIPE_PIPE] = ACTIONS(3355),
    [anon_sym_SLASH_SLASH] = ACTIONS(3355),
    [anon_sym_CARET_CARET] = ACTIONS(3353),
    [anon_sym_AMP_AMP] = ACTIONS(3355),
    [anon_sym_PIPE] = ACTIONS(3355),
    [anon_sym_CARET] = ACTIONS(3355),
    [anon_sym_AMP] = ACTIONS(3355),
    [anon_sym_GT_GT2] = ACTIONS(3355),
    [anon_sym_PLUS] = ACTIONS(3355),
    [anon_sym_DASH] = ACTIONS(3355),
    [anon_sym_DOT] = ACTIONS(3355),
    [anon_sym_SLASH] = ACTIONS(3355),
    [anon_sym_x] = ACTIONS(3355),
    [anon_sym_EQ_TILDE] = ACTIONS(3353),
    [anon_sym_BANG_TILDE] = ACTIONS(3353),
    [anon_sym_EQ_EQ] = ACTIONS(3355),
    [anon_sym_BANG_EQ] = ACTIONS(3353),
    [anon_sym_eq] = ACTIONS(3355),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3353),
    [anon_sym_equ] = ACTIONS(3353),
    [anon_sym_eqr] = ACTIONS(3353),
    [anon_sym_ne] = ACTIONS(3353),
    [anon_sym_LT_EQ_GT] = ACTIONS(3353),
    [anon_sym_cmp] = ACTIONS(3353),
    [anon_sym_TILDE_TILDE] = ACTIONS(3353),
    [anon_sym_LT] = ACTIONS(3355),
    [anon_sym_LT_EQ] = ACTIONS(3355),
    [anon_sym_GT_EQ] = ACTIONS(3353),
    [anon_sym_lt] = ACTIONS(3353),
    [anon_sym_le] = ACTIONS(3353),
    [anon_sym_ge] = ACTIONS(3353),
    [anon_sym_gt] = ACTIONS(3353),
    [anon_sym_isa] = ACTIONS(3353),
    [anon_sym_PLUS_PLUS] = ACTIONS(3353),
    [anon_sym_DASH_DASH] = ACTIONS(3353),
    [anon_sym_QMARK] = ACTIONS(3353),
    [anon_sym_EQ_GT] = ACTIONS(3353),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3353),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3137)] = {
    [sym_heredoc_content] = STATE(3137),
    [anon_sym_LBRACE] = ACTIONS(3357),
    [anon_sym_SEMI] = ACTIONS(3357),
    [anon_sym_EQ] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_COMMA] = ACTIONS(3357),
    [anon_sym_LPAREN] = ACTIONS(3357),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3357),
    [anon_sym_DASH_GT] = ACTIONS(3357),
    [anon_sym_STAR] = ACTIONS(3359),
    [anon_sym_LBRACK] = ACTIONS(3357),
    [anon_sym_PERCENT] = ACTIONS(3359),
    [anon_sym_GT] = ACTIONS(3359),
    [anon_sym_LT_LT] = ACTIONS(3359),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3357),
    [anon_sym_PLUS_EQ] = ACTIONS(3357),
    [anon_sym_DASH_EQ] = ACTIONS(3357),
    [anon_sym_DOT_EQ] = ACTIONS(3357),
    [anon_sym_STAR_EQ] = ACTIONS(3357),
    [anon_sym_SLASH_EQ] = ACTIONS(3357),
    [anon_sym_PERCENT_EQ] = ACTIONS(3357),
    [anon_sym_x_EQ] = ACTIONS(3357),
    [anon_sym_AMP_EQ] = ACTIONS(3357),
    [anon_sym_PIPE_EQ] = ACTIONS(3357),
    [anon_sym_CARET_EQ] = ACTIONS(3357),
    [anon_sym_LT_LT_EQ] = ACTIONS(3357),
    [anon_sym_GT_GT_EQ] = ACTIONS(3357),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3357),
    [anon_sym_DOT_DOT] = ACTIONS(3359),
    [anon_sym_STAR_STAR] = ACTIONS(3359),
    [anon_sym_PIPE_PIPE] = ACTIONS(3359),
    [anon_sym_SLASH_SLASH] = ACTIONS(3359),
    [anon_sym_CARET_CARET] = ACTIONS(3357),
    [anon_sym_AMP_AMP] = ACTIONS(3359),
    [anon_sym_PIPE] = ACTIONS(3359),
    [anon_sym_CARET] = ACTIONS(3359),
    [anon_sym_AMP] = ACTIONS(3359),
    [anon_sym_GT_GT2] = ACTIONS(3359),
    [anon_sym_PLUS] = ACTIONS(3359),
    [anon_sym_DASH] = ACTIONS(3359),
    [anon_sym_DOT] = ACTIONS(3359),
    [anon_sym_SLASH] = ACTIONS(3359),
    [anon_sym_x] = ACTIONS(3359),
    [anon_sym_EQ_TILDE] = ACTIONS(3357),
    [anon_sym_BANG_TILDE] = ACTIONS(3357),
    [anon_sym_EQ_EQ] = ACTIONS(3359),
    [anon_sym_BANG_EQ] = ACTIONS(3357),
    [anon_sym_eq] = ACTIONS(3359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3357),
    [anon_sym_equ] = ACTIONS(3357),
    [anon_sym_eqr] = ACTIONS(3357),
    [anon_sym_ne] = ACTIONS(3357),
    [anon_sym_LT_EQ_GT] = ACTIONS(3357),
    [anon_sym_cmp] = ACTIONS(3357),
    [anon_sym_TILDE_TILDE] = ACTIONS(3357),
    [anon_sym_LT] = ACTIONS(3359),
    [anon_sym_LT_EQ] = ACTIONS(3359),
    [anon_sym_GT_EQ] = ACTIONS(3357),
    [anon_sym_lt] = ACTIONS(3357),
    [anon_sym_le] = ACTIONS(3357),
    [anon_sym_ge] = ACTIONS(3357),
    [anon_sym_gt] = ACTIONS(3357),
    [anon_sym_isa] = ACTIONS(3357),
    [anon_sym_PLUS_PLUS] = ACTIONS(3357),
    [anon_sym_DASH_DASH] = ACTIONS(3357),
    [anon_sym_QMARK] = ACTIONS(3357),
    [anon_sym_EQ_GT] = ACTIONS(3357),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3357),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3138)] = {
    [sym_heredoc_content] = STATE(3138),
    [anon_sym_LBRACE] = ACTIONS(3369),
    [anon_sym_SEMI] = ACTIONS(3369),
    [anon_sym_EQ] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_COMMA] = ACTIONS(3369),
    [anon_sym_LPAREN] = ACTIONS(3369),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3369),
    [anon_sym_DASH_GT] = ACTIONS(3369),
    [anon_sym_STAR] = ACTIONS(3371),
    [anon_sym_LBRACK] = ACTIONS(3369),
    [anon_sym_PERCENT] = ACTIONS(3371),
    [anon_sym_GT] = ACTIONS(3371),
    [anon_sym_LT_LT] = ACTIONS(3371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3369),
    [anon_sym_PLUS_EQ] = ACTIONS(3369),
    [anon_sym_DASH_EQ] = ACTIONS(3369),
    [anon_sym_DOT_EQ] = ACTIONS(3369),
    [anon_sym_STAR_EQ] = ACTIONS(3369),
    [anon_sym_SLASH_EQ] = ACTIONS(3369),
    [anon_sym_PERCENT_EQ] = ACTIONS(3369),
    [anon_sym_x_EQ] = ACTIONS(3369),
    [anon_sym_AMP_EQ] = ACTIONS(3369),
    [anon_sym_PIPE_EQ] = ACTIONS(3369),
    [anon_sym_CARET_EQ] = ACTIONS(3369),
    [anon_sym_LT_LT_EQ] = ACTIONS(3369),
    [anon_sym_GT_GT_EQ] = ACTIONS(3369),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3369),
    [anon_sym_DOT_DOT] = ACTIONS(3371),
    [anon_sym_STAR_STAR] = ACTIONS(3371),
    [anon_sym_PIPE_PIPE] = ACTIONS(3371),
    [anon_sym_SLASH_SLASH] = ACTIONS(3371),
    [anon_sym_CARET_CARET] = ACTIONS(3369),
    [anon_sym_AMP_AMP] = ACTIONS(3371),
    [anon_sym_PIPE] = ACTIONS(3371),
    [anon_sym_CARET] = ACTIONS(3371),
    [anon_sym_AMP] = ACTIONS(3371),
    [anon_sym_GT_GT2] = ACTIONS(3371),
    [anon_sym_PLUS] = ACTIONS(3371),
    [anon_sym_DASH] = ACTIONS(3371),
    [anon_sym_DOT] = ACTIONS(3371),
    [anon_sym_SLASH] = ACTIONS(3371),
    [anon_sym_x] = ACTIONS(3371),
    [anon_sym_EQ_TILDE] = ACTIONS(3369),
    [anon_sym_BANG_TILDE] = ACTIONS(3369),
    [anon_sym_EQ_EQ] = ACTIONS(3371),
    [anon_sym_BANG_EQ] = ACTIONS(3369),
    [anon_sym_eq] = ACTIONS(3371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3369),
    [anon_sym_equ] = ACTIONS(3369),
    [anon_sym_eqr] = ACTIONS(3369),
    [anon_sym_ne] = ACTIONS(3369),
    [anon_sym_LT_EQ_GT] = ACTIONS(3369),
    [anon_sym_cmp] = ACTIONS(3369),
    [anon_sym_TILDE_TILDE] = ACTIONS(3369),
    [anon_sym_LT] = ACTIONS(3371),
    [anon_sym_LT_EQ] = ACTIONS(3371),
    [anon_sym_GT_EQ] = ACTIONS(3369),
    [anon_sym_lt] = ACTIONS(3369),
    [anon_sym_le] = ACTIONS(3369),
    [anon_sym_ge] = ACTIONS(3369),
    [anon_sym_gt] = ACTIONS(3369),
    [anon_sym_isa] = ACTIONS(3369),
    [anon_sym_PLUS_PLUS] = ACTIONS(3369),
    [anon_sym_DASH_DASH] = ACTIONS(3369),
    [anon_sym_QMARK] = ACTIONS(3369),
    [anon_sym_EQ_GT] = ACTIONS(3369),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3369),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3139)] = {
    [sym_heredoc_content] = STATE(3139),
    [anon_sym_LBRACE] = ACTIONS(3393),
    [anon_sym_SEMI] = ACTIONS(3393),
    [anon_sym_EQ] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_COMMA] = ACTIONS(3393),
    [anon_sym_LPAREN] = ACTIONS(3393),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3393),
    [anon_sym_DASH_GT] = ACTIONS(3393),
    [anon_sym_STAR] = ACTIONS(3395),
    [anon_sym_LBRACK] = ACTIONS(3393),
    [anon_sym_PERCENT] = ACTIONS(3395),
    [anon_sym_GT] = ACTIONS(3395),
    [anon_sym_LT_LT] = ACTIONS(3395),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3393),
    [anon_sym_PLUS_EQ] = ACTIONS(3393),
    [anon_sym_DASH_EQ] = ACTIONS(3393),
    [anon_sym_DOT_EQ] = ACTIONS(3393),
    [anon_sym_STAR_EQ] = ACTIONS(3393),
    [anon_sym_SLASH_EQ] = ACTIONS(3393),
    [anon_sym_PERCENT_EQ] = ACTIONS(3393),
    [anon_sym_x_EQ] = ACTIONS(3393),
    [anon_sym_AMP_EQ] = ACTIONS(3393),
    [anon_sym_PIPE_EQ] = ACTIONS(3393),
    [anon_sym_CARET_EQ] = ACTIONS(3393),
    [anon_sym_LT_LT_EQ] = ACTIONS(3393),
    [anon_sym_GT_GT_EQ] = ACTIONS(3393),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3393),
    [anon_sym_DOT_DOT] = ACTIONS(3395),
    [anon_sym_STAR_STAR] = ACTIONS(3395),
    [anon_sym_PIPE_PIPE] = ACTIONS(3395),
    [anon_sym_SLASH_SLASH] = ACTIONS(3395),
    [anon_sym_CARET_CARET] = ACTIONS(3393),
    [anon_sym_AMP_AMP] = ACTIONS(3395),
    [anon_sym_PIPE] = ACTIONS(3395),
    [anon_sym_CARET] = ACTIONS(3395),
    [anon_sym_AMP] = ACTIONS(3395),
    [anon_sym_GT_GT2] = ACTIONS(3395),
    [anon_sym_PLUS] = ACTIONS(3395),
    [anon_sym_DASH] = ACTIONS(3395),
    [anon_sym_DOT] = ACTIONS(3395),
    [anon_sym_SLASH] = ACTIONS(3395),
    [anon_sym_x] = ACTIONS(3395),
    [anon_sym_EQ_TILDE] = ACTIONS(3393),
    [anon_sym_BANG_TILDE] = ACTIONS(3393),
    [anon_sym_EQ_EQ] = ACTIONS(3395),
    [anon_sym_BANG_EQ] = ACTIONS(3393),
    [anon_sym_eq] = ACTIONS(3395),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3393),
    [anon_sym_equ] = ACTIONS(3393),
    [anon_sym_eqr] = ACTIONS(3393),
    [anon_sym_ne] = ACTIONS(3393),
    [anon_sym_LT_EQ_GT] = ACTIONS(3393),
    [anon_sym_cmp] = ACTIONS(3393),
    [anon_sym_TILDE_TILDE] = ACTIONS(3393),
    [anon_sym_LT] = ACTIONS(3395),
    [anon_sym_LT_EQ] = ACTIONS(3395),
    [anon_sym_GT_EQ] = ACTIONS(3393),
    [anon_sym_lt] = ACTIONS(3393),
    [anon_sym_le] = ACTIONS(3393),
    [anon_sym_ge] = ACTIONS(3393),
    [anon_sym_gt] = ACTIONS(3393),
    [anon_sym_isa] = ACTIONS(3393),
    [anon_sym_PLUS_PLUS] = ACTIONS(3393),
    [anon_sym_DASH_DASH] = ACTIONS(3393),
    [anon_sym_QMARK] = ACTIONS(3393),
    [anon_sym_EQ_GT] = ACTIONS(3393),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3393),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3140)] = {
    [sym_heredoc_content] = STATE(3140),
    [anon_sym_RBRACE] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_and] = ACTIONS(4009),
    [anon_sym_or] = ACTIONS(4009),
    [anon_sym_xor] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4009),
  },
  [STATE(3141)] = {
    [sym_heredoc_content] = STATE(3141),
    [anon_sym_LBRACE] = ACTIONS(3377),
    [anon_sym_SEMI] = ACTIONS(3377),
    [anon_sym_EQ] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_COMMA] = ACTIONS(3377),
    [anon_sym_LPAREN] = ACTIONS(3377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3377),
    [anon_sym_DASH_GT] = ACTIONS(3377),
    [anon_sym_STAR] = ACTIONS(3379),
    [anon_sym_LBRACK] = ACTIONS(3377),
    [anon_sym_PERCENT] = ACTIONS(3379),
    [anon_sym_GT] = ACTIONS(3379),
    [anon_sym_LT_LT] = ACTIONS(3379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3377),
    [anon_sym_PLUS_EQ] = ACTIONS(3377),
    [anon_sym_DASH_EQ] = ACTIONS(3377),
    [anon_sym_DOT_EQ] = ACTIONS(3377),
    [anon_sym_STAR_EQ] = ACTIONS(3377),
    [anon_sym_SLASH_EQ] = ACTIONS(3377),
    [anon_sym_PERCENT_EQ] = ACTIONS(3377),
    [anon_sym_x_EQ] = ACTIONS(3377),
    [anon_sym_AMP_EQ] = ACTIONS(3377),
    [anon_sym_PIPE_EQ] = ACTIONS(3377),
    [anon_sym_CARET_EQ] = ACTIONS(3377),
    [anon_sym_LT_LT_EQ] = ACTIONS(3377),
    [anon_sym_GT_GT_EQ] = ACTIONS(3377),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3377),
    [anon_sym_DOT_DOT] = ACTIONS(3379),
    [anon_sym_STAR_STAR] = ACTIONS(3379),
    [anon_sym_PIPE_PIPE] = ACTIONS(3379),
    [anon_sym_SLASH_SLASH] = ACTIONS(3379),
    [anon_sym_CARET_CARET] = ACTIONS(3377),
    [anon_sym_AMP_AMP] = ACTIONS(3379),
    [anon_sym_PIPE] = ACTIONS(3379),
    [anon_sym_CARET] = ACTIONS(3379),
    [anon_sym_AMP] = ACTIONS(3379),
    [anon_sym_GT_GT2] = ACTIONS(3379),
    [anon_sym_PLUS] = ACTIONS(3379),
    [anon_sym_DASH] = ACTIONS(3379),
    [anon_sym_DOT] = ACTIONS(3379),
    [anon_sym_SLASH] = ACTIONS(3379),
    [anon_sym_x] = ACTIONS(3379),
    [anon_sym_EQ_TILDE] = ACTIONS(3377),
    [anon_sym_BANG_TILDE] = ACTIONS(3377),
    [anon_sym_EQ_EQ] = ACTIONS(3379),
    [anon_sym_BANG_EQ] = ACTIONS(3377),
    [anon_sym_eq] = ACTIONS(3379),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3377),
    [anon_sym_equ] = ACTIONS(3377),
    [anon_sym_eqr] = ACTIONS(3377),
    [anon_sym_ne] = ACTIONS(3377),
    [anon_sym_LT_EQ_GT] = ACTIONS(3377),
    [anon_sym_cmp] = ACTIONS(3377),
    [anon_sym_TILDE_TILDE] = ACTIONS(3377),
    [anon_sym_LT] = ACTIONS(3379),
    [anon_sym_LT_EQ] = ACTIONS(3379),
    [anon_sym_GT_EQ] = ACTIONS(3377),
    [anon_sym_lt] = ACTIONS(3377),
    [anon_sym_le] = ACTIONS(3377),
    [anon_sym_ge] = ACTIONS(3377),
    [anon_sym_gt] = ACTIONS(3377),
    [anon_sym_isa] = ACTIONS(3377),
    [anon_sym_PLUS_PLUS] = ACTIONS(3377),
    [anon_sym_DASH_DASH] = ACTIONS(3377),
    [anon_sym_QMARK] = ACTIONS(3377),
    [anon_sym_EQ_GT] = ACTIONS(3377),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3377),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3142)] = {
    [sym_heredoc_content] = STATE(3142),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_DOLLAR] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_AT] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5002),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3821),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_PERCENT2] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5002),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3143)] = {
    [sym_heredoc_content] = STATE(3143),
    [anon_sym_RBRACE] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_and] = ACTIONS(3731),
    [anon_sym_or] = ACTIONS(3731),
    [anon_sym_xor] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3731),
  },
  [STATE(3144)] = {
    [sym_heredoc_content] = STATE(3144),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4045),
  },
  [STATE(3145)] = {
    [sym_heredoc_content] = STATE(3145),
    [anon_sym_RBRACE] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_and] = ACTIONS(4047),
    [anon_sym_or] = ACTIONS(4047),
    [anon_sym_xor] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4047),
  },
  [STATE(3146)] = {
    [sym_heredoc_content] = STATE(3146),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_and] = ACTIONS(3485),
    [anon_sym_or] = ACTIONS(3485),
    [anon_sym_xor] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3485),
  },
  [STATE(3147)] = {
    [sym_heredoc_content] = STATE(3147),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3415),
  },
  [STATE(3148)] = {
    [sym_heredoc_content] = STATE(3148),
    [anon_sym_RBRACE] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_and] = ACTIONS(4071),
    [anon_sym_or] = ACTIONS(4071),
    [anon_sym_xor] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4071),
  },
  [STATE(3149)] = {
    [sym_heredoc_content] = STATE(3149),
    [anon_sym_RBRACE] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_and] = ACTIONS(4135),
    [anon_sym_or] = ACTIONS(4135),
    [anon_sym_xor] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4135),
  },
  [STATE(3150)] = {
    [sym_heredoc_content] = STATE(3150),
    [anon_sym_RBRACE] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_and] = ACTIONS(3941),
    [anon_sym_or] = ACTIONS(3941),
    [anon_sym_xor] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3941),
  },
  [STATE(3151)] = {
    [sym_heredoc_content] = STATE(3151),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_xor] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2711),
  },
  [STATE(3152)] = {
    [sym_heredoc_content] = STATE(3152),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3917),
  },
  [STATE(3153)] = {
    [sym_heredoc_content] = STATE(3153),
    [anon_sym_RBRACE] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_and] = ACTIONS(3949),
    [anon_sym_or] = ACTIONS(3949),
    [anon_sym_xor] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3949),
  },
  [STATE(3154)] = {
    [sym_heredoc_content] = STATE(3154),
    [anon_sym_RBRACE] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_and] = ACTIONS(3957),
    [anon_sym_or] = ACTIONS(3957),
    [anon_sym_xor] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3957),
  },
  [STATE(3155)] = {
    [sym_heredoc_content] = STATE(3155),
    [anon_sym_RBRACE] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_and] = ACTIONS(4208),
    [anon_sym_or] = ACTIONS(4208),
    [anon_sym_xor] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4208),
  },
  [STATE(3156)] = {
    [sym_heredoc_content] = STATE(3156),
    [anon_sym_RBRACE] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_and] = ACTIONS(3977),
    [anon_sym_or] = ACTIONS(3977),
    [anon_sym_xor] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3977),
  },
  [STATE(3157)] = {
    [sym_heredoc_content] = STATE(3157),
    [anon_sym_RBRACE] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_and] = ACTIONS(4165),
    [anon_sym_or] = ACTIONS(4165),
    [anon_sym_xor] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4165),
  },
  [STATE(3158)] = {
    [sym_heredoc_content] = STATE(3158),
    [anon_sym_RBRACE] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_and] = ACTIONS(4204),
    [anon_sym_or] = ACTIONS(4204),
    [anon_sym_xor] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4204),
  },
  [STATE(3159)] = {
    [sym_heredoc_content] = STATE(3159),
    [anon_sym_RBRACE] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_and] = ACTIONS(3869),
    [anon_sym_or] = ACTIONS(3869),
    [anon_sym_xor] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3869),
  },
  [STATE(3160)] = {
    [sym_heredoc_content] = STATE(3160),
    [anon_sym_RBRACE] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_and] = ACTIONS(4103),
    [anon_sym_or] = ACTIONS(4103),
    [anon_sym_xor] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4103),
  },
  [STATE(3161)] = {
    [sym_heredoc_content] = STATE(3161),
    [anon_sym_RBRACE] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_and] = ACTIONS(3997),
    [anon_sym_or] = ACTIONS(3997),
    [anon_sym_xor] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3997),
  },
  [STATE(3162)] = {
    [sym_heredoc_content] = STATE(3162),
    [anon_sym_RBRACE] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_and] = ACTIONS(4159),
    [anon_sym_or] = ACTIONS(4159),
    [anon_sym_xor] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4159),
  },
  [STATE(3163)] = {
    [sym_heredoc_content] = STATE(3163),
    [anon_sym_RBRACE] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_and] = ACTIONS(4169),
    [anon_sym_or] = ACTIONS(4169),
    [anon_sym_xor] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(5004),
    [anon_sym_DASH_DASH] = ACTIONS(5004),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4169),
  },
  [STATE(3164)] = {
    [sym_heredoc_content] = STATE(3164),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_and] = ACTIONS(3695),
    [anon_sym_or] = ACTIONS(3695),
    [anon_sym_xor] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3695),
  },
  [STATE(3165)] = {
    [sym_heredoc_content] = STATE(3165),
    [anon_sym_RBRACE] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_and] = ACTIONS(4037),
    [anon_sym_or] = ACTIONS(4037),
    [anon_sym_xor] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4037),
  },
  [STATE(3166)] = {
    [sym_heredoc_content] = STATE(3166),
    [anon_sym_RBRACE] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_and] = ACTIONS(3909),
    [anon_sym_or] = ACTIONS(3909),
    [anon_sym_xor] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3909),
  },
  [STATE(3167)] = {
    [sym_heredoc_content] = STATE(3167),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4017),
  },
  [STATE(3168)] = {
    [sym_heredoc_content] = STATE(3168),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_and] = ACTIONS(4017),
    [anon_sym_or] = ACTIONS(4017),
    [anon_sym_xor] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4017),
  },
  [STATE(3169)] = {
    [sym_heredoc_content] = STATE(3169),
    [anon_sym_RBRACE] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3803),
    [anon_sym_or] = ACTIONS(3803),
    [anon_sym_xor] = ACTIONS(3803),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3803),
  },
  [STATE(3170)] = {
    [sym_heredoc_content] = STATE(3170),
    [anon_sym_RBRACE] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_and] = ACTIONS(3845),
    [anon_sym_or] = ACTIONS(3845),
    [anon_sym_xor] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3845),
  },
  [STATE(3171)] = {
    [sym_heredoc_content] = STATE(3171),
    [anon_sym_RBRACE] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_and] = ACTIONS(3857),
    [anon_sym_or] = ACTIONS(3857),
    [anon_sym_xor] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3857),
  },
  [STATE(3172)] = {
    [sym_heredoc_content] = STATE(3172),
    [anon_sym_RBRACE] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_and] = ACTIONS(3739),
    [anon_sym_or] = ACTIONS(3739),
    [anon_sym_xor] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3739),
  },
  [STATE(3173)] = {
    [sym_heredoc_content] = STATE(3173),
    [anon_sym_RBRACE] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_and] = ACTIONS(3815),
    [anon_sym_or] = ACTIONS(3815),
    [anon_sym_xor] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3815),
  },
  [STATE(3174)] = {
    [sym_heredoc_content] = STATE(3174),
    [anon_sym_RBRACE] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_and] = ACTIONS(3775),
    [anon_sym_or] = ACTIONS(3775),
    [anon_sym_xor] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3775),
  },
  [STATE(3175)] = {
    [sym_heredoc_content] = STATE(3175),
    [anon_sym_RBRACE] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(4687),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4693),
    [anon_sym_and] = ACTIONS(3897),
    [anon_sym_or] = ACTIONS(3897),
    [anon_sym_xor] = ACTIONS(3897),
    [anon_sym_DASH_GT] = ACTIONS(4695),
    [anon_sym_STAR] = ACTIONS(4697),
    [anon_sym_PERCENT] = ACTIONS(4697),
    [anon_sym_GT] = ACTIONS(4699),
    [anon_sym_LT_LT] = ACTIONS(4701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4689),
    [anon_sym_PLUS_EQ] = ACTIONS(4689),
    [anon_sym_DASH_EQ] = ACTIONS(4689),
    [anon_sym_DOT_EQ] = ACTIONS(4689),
    [anon_sym_STAR_EQ] = ACTIONS(4689),
    [anon_sym_SLASH_EQ] = ACTIONS(4689),
    [anon_sym_PERCENT_EQ] = ACTIONS(4689),
    [anon_sym_x_EQ] = ACTIONS(4689),
    [anon_sym_AMP_EQ] = ACTIONS(4689),
    [anon_sym_PIPE_EQ] = ACTIONS(4689),
    [anon_sym_CARET_EQ] = ACTIONS(4689),
    [anon_sym_LT_LT_EQ] = ACTIONS(4689),
    [anon_sym_GT_GT_EQ] = ACTIONS(4689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4689),
    [anon_sym_DOT_DOT] = ACTIONS(4703),
    [anon_sym_STAR_STAR] = ACTIONS(4705),
    [anon_sym_PIPE_PIPE] = ACTIONS(4707),
    [anon_sym_SLASH_SLASH] = ACTIONS(4707),
    [anon_sym_CARET_CARET] = ACTIONS(4709),
    [anon_sym_AMP_AMP] = ACTIONS(4711),
    [anon_sym_PIPE] = ACTIONS(4713),
    [anon_sym_CARET] = ACTIONS(4713),
    [anon_sym_AMP] = ACTIONS(4715),
    [anon_sym_GT_GT2] = ACTIONS(4701),
    [anon_sym_PLUS] = ACTIONS(4717),
    [anon_sym_DASH] = ACTIONS(4717),
    [anon_sym_DOT] = ACTIONS(4717),
    [anon_sym_SLASH] = ACTIONS(4697),
    [anon_sym_x] = ACTIONS(4697),
    [anon_sym_EQ_TILDE] = ACTIONS(4719),
    [anon_sym_BANG_TILDE] = ACTIONS(4719),
    [anon_sym_EQ_EQ] = ACTIONS(4721),
    [anon_sym_BANG_EQ] = ACTIONS(4723),
    [anon_sym_eq] = ACTIONS(4721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4723),
    [anon_sym_equ] = ACTIONS(4723),
    [anon_sym_eqr] = ACTIONS(4723),
    [anon_sym_ne] = ACTIONS(4723),
    [anon_sym_LT_EQ_GT] = ACTIONS(4725),
    [anon_sym_cmp] = ACTIONS(4725),
    [anon_sym_TILDE_TILDE] = ACTIONS(4725),
    [anon_sym_LT] = ACTIONS(4699),
    [anon_sym_LT_EQ] = ACTIONS(4699),
    [anon_sym_GT_EQ] = ACTIONS(4727),
    [anon_sym_lt] = ACTIONS(4727),
    [anon_sym_le] = ACTIONS(4727),
    [anon_sym_ge] = ACTIONS(4727),
    [anon_sym_gt] = ACTIONS(4727),
    [anon_sym_isa] = ACTIONS(4729),
    [anon_sym_PLUS_PLUS] = ACTIONS(4731),
    [anon_sym_DASH_DASH] = ACTIONS(4731),
    [anon_sym_QMARK] = ACTIONS(4733),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3897),
  },
  [STATE(3176)] = {
    [sym_heredoc_content] = STATE(3176),
    [anon_sym_RBRACE] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_and] = ACTIONS(3901),
    [anon_sym_or] = ACTIONS(3901),
    [anon_sym_xor] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3901),
  },
  [STATE(3177)] = {
    [sym_heredoc_content] = STATE(3177),
    [anon_sym_RBRACE] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_and] = ACTIONS(4119),
    [anon_sym_or] = ACTIONS(4119),
    [anon_sym_xor] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4119),
  },
  [STATE(3178)] = {
    [sym_heredoc_content] = STATE(3178),
    [anon_sym_RBRACE] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_and] = ACTIONS(4131),
    [anon_sym_or] = ACTIONS(4131),
    [anon_sym_xor] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4131),
  },
  [STATE(3179)] = {
    [sym_heredoc_content] = STATE(3179),
    [anon_sym_RBRACE] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_and] = ACTIONS(3881),
    [anon_sym_or] = ACTIONS(3881),
    [anon_sym_xor] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3881),
  },
  [STATE(3180)] = {
    [sym_heredoc_content] = STATE(3180),
    [anon_sym_RBRACE] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_and] = ACTIONS(4075),
    [anon_sym_or] = ACTIONS(4075),
    [anon_sym_xor] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4075),
  },
  [STATE(3181)] = {
    [sym_heredoc_content] = STATE(3181),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3819),
  },
  [STATE(3182)] = {
    [sym_heredoc_content] = STATE(3182),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5007),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_and] = ACTIONS(3819),
    [anon_sym_or] = ACTIONS(3819),
    [anon_sym_xor] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5007),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3819),
  },
  [STATE(3183)] = {
    [sym_heredoc_content] = STATE(3183),
    [anon_sym_RBRACE] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_and] = ACTIONS(3885),
    [anon_sym_or] = ACTIONS(3885),
    [anon_sym_xor] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3885),
  },
  [STATE(3184)] = {
    [sym_heredoc_content] = STATE(3184),
    [anon_sym_RBRACE] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_and] = ACTIONS(3889),
    [anon_sym_or] = ACTIONS(3889),
    [anon_sym_xor] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3889),
  },
  [STATE(3185)] = {
    [sym_heredoc_content] = STATE(3185),
    [anon_sym_RBRACE] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_and] = ACTIONS(3973),
    [anon_sym_or] = ACTIONS(3973),
    [anon_sym_xor] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3973),
  },
  [STATE(3186)] = {
    [sym_heredoc_content] = STATE(3186),
    [anon_sym_RBRACE] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_and] = ACTIONS(4021),
    [anon_sym_or] = ACTIONS(4021),
    [anon_sym_xor] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4021),
  },
  [STATE(3187)] = {
    [sym_heredoc_content] = STATE(3187),
    [anon_sym_RBRACE] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_and] = ACTIONS(4033),
    [anon_sym_or] = ACTIONS(4033),
    [anon_sym_xor] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4033),
  },
  [STATE(3188)] = {
    [sym_heredoc_content] = STATE(3188),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_and] = ACTIONS(2536),
    [anon_sym_or] = ACTIONS(2536),
    [anon_sym_xor] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(2536),
  },
  [STATE(3189)] = {
    [sym_heredoc_content] = STATE(3189),
    [anon_sym_RBRACE] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_and] = ACTIONS(3747),
    [anon_sym_or] = ACTIONS(3747),
    [anon_sym_xor] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3747),
  },
  [STATE(3190)] = {
    [sym_heredoc_content] = STATE(3190),
    [anon_sym_RBRACE] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_and] = ACTIONS(3779),
    [anon_sym_or] = ACTIONS(3779),
    [anon_sym_xor] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3779),
  },
  [STATE(3191)] = {
    [sym_heredoc_content] = STATE(3191),
    [anon_sym_RBRACE] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_and] = ACTIONS(3833),
    [anon_sym_or] = ACTIONS(3833),
    [anon_sym_xor] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3833),
  },
  [STATE(3192)] = {
    [sym_heredoc_content] = STATE(3192),
    [anon_sym_RBRACE] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_and] = ACTIONS(3841),
    [anon_sym_or] = ACTIONS(3841),
    [anon_sym_xor] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3841),
  },
  [STATE(3193)] = {
    [sym_heredoc_content] = STATE(3193),
    [anon_sym_RBRACE] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_and] = ACTIONS(3853),
    [anon_sym_or] = ACTIONS(3853),
    [anon_sym_xor] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3853),
  },
  [STATE(3194)] = {
    [sym_heredoc_content] = STATE(3194),
    [anon_sym_RBRACE] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_and] = ACTIONS(3861),
    [anon_sym_or] = ACTIONS(3861),
    [anon_sym_xor] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3861),
  },
  [STATE(3195)] = {
    [sym_heredoc_content] = STATE(3195),
    [anon_sym_RBRACE] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_and] = ACTIONS(3837),
    [anon_sym_or] = ACTIONS(3837),
    [anon_sym_xor] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3837),
  },
  [STATE(3196)] = {
    [sym_heredoc_content] = STATE(3196),
    [anon_sym_RBRACE] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_and] = ACTIONS(3993),
    [anon_sym_or] = ACTIONS(3993),
    [anon_sym_xor] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3993),
  },
  [STATE(3197)] = {
    [sym_heredoc_content] = STATE(3197),
    [anon_sym_RBRACE] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_and] = ACTIONS(4005),
    [anon_sym_or] = ACTIONS(4005),
    [anon_sym_xor] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4005),
  },
  [STATE(3198)] = {
    [sym_heredoc_content] = STATE(3198),
    [anon_sym_RBRACE] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_and] = ACTIONS(4001),
    [anon_sym_or] = ACTIONS(4001),
    [anon_sym_xor] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4001),
  },
  [STATE(3199)] = {
    [sym_heredoc_content] = STATE(3199),
    [anon_sym_RBRACE] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_and] = ACTIONS(4087),
    [anon_sym_or] = ACTIONS(4087),
    [anon_sym_xor] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4087),
  },
  [STATE(3200)] = {
    [sym_heredoc_content] = STATE(3200),
    [anon_sym_RBRACE] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_and] = ACTIONS(3771),
    [anon_sym_or] = ACTIONS(3771),
    [anon_sym_xor] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3771),
  },
  [STATE(3201)] = {
    [sym_heredoc_content] = STATE(3201),
    [anon_sym_RBRACE] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_and] = ACTIONS(3783),
    [anon_sym_or] = ACTIONS(3783),
    [anon_sym_xor] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3783),
  },
  [STATE(3202)] = {
    [sym_heredoc_content] = STATE(3202),
    [anon_sym_RBRACE] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_and] = ACTIONS(3787),
    [anon_sym_or] = ACTIONS(3787),
    [anon_sym_xor] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3787),
  },
  [STATE(3203)] = {
    [sym_heredoc_content] = STATE(3203),
    [anon_sym_RBRACE] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_and] = ACTIONS(3791),
    [anon_sym_or] = ACTIONS(3791),
    [anon_sym_xor] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3791),
  },
  [STATE(3204)] = {
    [sym_heredoc_content] = STATE(3204),
    [anon_sym_RBRACE] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_and] = ACTIONS(3795),
    [anon_sym_or] = ACTIONS(3795),
    [anon_sym_xor] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3795),
  },
  [STATE(3205)] = {
    [sym_heredoc_content] = STATE(3205),
    [anon_sym_RBRACE] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_and] = ACTIONS(3925),
    [anon_sym_or] = ACTIONS(3925),
    [anon_sym_xor] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3925),
  },
  [STATE(3206)] = {
    [sym_heredoc_content] = STATE(3206),
    [anon_sym_RBRACE] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_and] = ACTIONS(3929),
    [anon_sym_or] = ACTIONS(3929),
    [anon_sym_xor] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3929),
  },
  [STATE(3207)] = {
    [sym_heredoc_content] = STATE(3207),
    [anon_sym_RBRACE] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_and] = ACTIONS(3933),
    [anon_sym_or] = ACTIONS(3933),
    [anon_sym_xor] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3933),
  },
  [STATE(3208)] = {
    [sym_heredoc_content] = STATE(3208),
    [anon_sym_RBRACE] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_and] = ACTIONS(3937),
    [anon_sym_or] = ACTIONS(3937),
    [anon_sym_xor] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3937),
  },
  [STATE(3209)] = {
    [sym_heredoc_content] = STATE(3209),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_and] = ACTIONS(3953),
    [anon_sym_or] = ACTIONS(3953),
    [anon_sym_xor] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3953),
  },
  [STATE(3210)] = {
    [sym_heredoc_content] = STATE(3210),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_and] = ACTIONS(3965),
    [anon_sym_or] = ACTIONS(3965),
    [anon_sym_xor] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3965),
  },
  [STATE(3211)] = {
    [sym_heredoc_content] = STATE(3211),
    [anon_sym_RBRACE] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_and] = ACTIONS(4155),
    [anon_sym_or] = ACTIONS(4155),
    [anon_sym_xor] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4155),
  },
  [STATE(3212)] = {
    [sym_heredoc_content] = STATE(3212),
    [anon_sym_RBRACE] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_and] = ACTIONS(4176),
    [anon_sym_or] = ACTIONS(4176),
    [anon_sym_xor] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4176),
  },
  [STATE(3213)] = {
    [sym_heredoc_content] = STATE(3213),
    [anon_sym_RBRACE] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_and] = ACTIONS(4180),
    [anon_sym_or] = ACTIONS(4180),
    [anon_sym_xor] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4180),
  },
  [STATE(3214)] = {
    [sym_heredoc_content] = STATE(3214),
    [anon_sym_RBRACE] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_and] = ACTIONS(4188),
    [anon_sym_or] = ACTIONS(4188),
    [anon_sym_xor] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4188),
  },
  [STATE(3215)] = {
    [sym_heredoc_content] = STATE(3215),
    [anon_sym_RBRACE] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_and] = ACTIONS(4192),
    [anon_sym_or] = ACTIONS(4192),
    [anon_sym_xor] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4192),
  },
  [STATE(3216)] = {
    [sym_heredoc_content] = STATE(3216),
    [anon_sym_RBRACE] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_and] = ACTIONS(4200),
    [anon_sym_or] = ACTIONS(4200),
    [anon_sym_xor] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4200),
  },
  [STATE(3217)] = {
    [sym_heredoc_content] = STATE(3217),
    [anon_sym_RBRACE] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_and] = ACTIONS(3969),
    [anon_sym_or] = ACTIONS(3969),
    [anon_sym_xor] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3969),
  },
  [STATE(3218)] = {
    [sym_heredoc_content] = STATE(3218),
    [anon_sym_RBRACE] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_and] = ACTIONS(4079),
    [anon_sym_or] = ACTIONS(4079),
    [anon_sym_xor] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4079),
  },
  [STATE(3219)] = {
    [sym_heredoc_content] = STATE(3219),
    [anon_sym_RBRACE] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_and] = ACTIONS(4083),
    [anon_sym_or] = ACTIONS(4083),
    [anon_sym_xor] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4083),
  },
  [STATE(3220)] = {
    [sym_heredoc_content] = STATE(3220),
    [anon_sym_RBRACE] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_and] = ACTIONS(4091),
    [anon_sym_or] = ACTIONS(4091),
    [anon_sym_xor] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4091),
  },
  [STATE(3221)] = {
    [sym_heredoc_content] = STATE(3221),
    [anon_sym_RBRACE] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_and] = ACTIONS(3715),
    [anon_sym_or] = ACTIONS(3715),
    [anon_sym_xor] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3715),
  },
  [STATE(3222)] = {
    [sym_heredoc_content] = STATE(3222),
    [anon_sym_RBRACE] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_and] = ACTIONS(3893),
    [anon_sym_or] = ACTIONS(3893),
    [anon_sym_xor] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3893),
  },
  [STATE(3223)] = {
    [sym_heredoc_content] = STATE(3223),
    [anon_sym_RBRACE] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_and] = ACTIONS(4063),
    [anon_sym_or] = ACTIONS(4063),
    [anon_sym_xor] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4063),
  },
  [STATE(3224)] = {
    [sym_heredoc_content] = STATE(3224),
    [anon_sym_RBRACE] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_and] = ACTIONS(3877),
    [anon_sym_or] = ACTIONS(3877),
    [anon_sym_xor] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3877),
  },
  [STATE(3225)] = {
    [sym_heredoc_content] = STATE(3225),
    [anon_sym_RBRACE] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_and] = ACTIONS(3989),
    [anon_sym_or] = ACTIONS(3989),
    [anon_sym_xor] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3989),
  },
  [STATE(3226)] = {
    [sym_heredoc_content] = STATE(3226),
    [anon_sym_RBRACE] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_and] = ACTIONS(4115),
    [anon_sym_or] = ACTIONS(4115),
    [anon_sym_xor] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4115),
  },
  [STATE(3227)] = {
    [sym_heredoc_content] = STATE(3227),
    [anon_sym_RBRACE] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_and] = ACTIONS(3807),
    [anon_sym_or] = ACTIONS(3807),
    [anon_sym_xor] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3807),
  },
  [STATE(3228)] = {
    [sym_heredoc_content] = STATE(3228),
    [anon_sym_RBRACE] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_and] = ACTIONS(4041),
    [anon_sym_or] = ACTIONS(4041),
    [anon_sym_xor] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4041),
  },
  [STATE(3229)] = {
    [sym_heredoc_content] = STATE(3229),
    [anon_sym_RBRACE] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_and] = ACTIONS(4055),
    [anon_sym_or] = ACTIONS(4055),
    [anon_sym_xor] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4055),
  },
  [STATE(3230)] = {
    [sym_heredoc_content] = STATE(3230),
    [anon_sym_RBRACE] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_and] = ACTIONS(3865),
    [anon_sym_or] = ACTIONS(3865),
    [anon_sym_xor] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3865),
  },
  [STATE(3231)] = {
    [sym_heredoc_content] = STATE(3231),
    [anon_sym_RBRACE] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_and] = ACTIONS(3829),
    [anon_sym_or] = ACTIONS(3829),
    [anon_sym_xor] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3829),
  },
  [STATE(3232)] = {
    [sym_heredoc_content] = STATE(3232),
    [anon_sym_RBRACE] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_and] = ACTIONS(4095),
    [anon_sym_or] = ACTIONS(4095),
    [anon_sym_xor] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4095),
  },
  [STATE(3233)] = {
    [sym_heredoc_content] = STATE(3233),
    [anon_sym_RBRACE] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_and] = ACTIONS(4184),
    [anon_sym_or] = ACTIONS(4184),
    [anon_sym_xor] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(4184),
  },
  [STATE(3234)] = {
    [sym_heredoc_content] = STATE(3234),
    [anon_sym_RBRACE] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_and] = ACTIONS(3727),
    [anon_sym_or] = ACTIONS(3727),
    [anon_sym_xor] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3727),
  },
  [STATE(3235)] = {
    [sym_heredoc_content] = STATE(3235),
    [anon_sym_RBRACE] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_and] = ACTIONS(3743),
    [anon_sym_or] = ACTIONS(3743),
    [anon_sym_xor] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3743),
  },
  [STATE(3236)] = {
    [sym_heredoc_content] = STATE(3236),
    [anon_sym_RBRACE] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_and] = ACTIONS(3811),
    [anon_sym_or] = ACTIONS(3811),
    [anon_sym_xor] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3811),
  },
  [STATE(3237)] = {
    [sym_heredoc_content] = STATE(3237),
    [anon_sym_RBRACE] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_and] = ACTIONS(3873),
    [anon_sym_or] = ACTIONS(3873),
    [anon_sym_xor] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3873),
  },
  [STATE(3238)] = {
    [sym_heredoc_content] = STATE(3238),
    [anon_sym_RBRACE] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_and] = ACTIONS(3981),
    [anon_sym_or] = ACTIONS(3981),
    [anon_sym_xor] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACE_CLOSE] = ACTIONS(3981),
  },
  [STATE(3239)] = {
    [sym_heredoc_content] = STATE(3239),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_and] = ACTIONS(4188),
    [anon_sym_or] = ACTIONS(4188),
    [anon_sym_xor] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_RBRACK] = ACTIONS(4188),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
    [sym__RECOVER_BRACKET_CLOSE] = ACTIONS(4188),
  },
  [STATE(3240)] = {
    [sym_heredoc_content] = STATE(3240),
    [anon_sym_COLON] = ACTIONS(3739),
    [anon_sym_DOLLAR] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_AT] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3741),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_PERCENT2] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3241)] = {
    [sym_heredoc_content] = STATE(3241),
    [anon_sym_RBRACE] = ACTIONS(3556),
    [anon_sym_COLON] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_RPAREN] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5019),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3242)] = {
    [sym_heredoc_content] = STATE(3242),
    [anon_sym_RBRACE] = ACTIONS(3574),
    [anon_sym_COLON] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_RPAREN] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(5021),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3243)] = {
    [sym_heredoc_content] = STATE(3243),
    [anon_sym_RBRACE] = ACTIONS(3699),
    [anon_sym_COLON] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_RPAREN] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3244)] = {
    [sym_heredoc_content] = STATE(3244),
    [anon_sym_RBRACE] = ACTIONS(3703),
    [anon_sym_COLON] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_RPAREN] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(5023),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3245)] = {
    [sym_heredoc_content] = STATE(3245),
    [anon_sym_RBRACE] = ACTIONS(3709),
    [anon_sym_COLON] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_RPAREN] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5025),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3246)] = {
    [sym_heredoc_content] = STATE(3246),
    [anon_sym_RBRACE] = ACTIONS(3615),
    [anon_sym_COLON] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_RPAREN] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5027),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3247)] = {
    [sym_heredoc_content] = STATE(3247),
    [anon_sym_RBRACE] = ACTIONS(3683),
    [anon_sym_COLON] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_RPAREN] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5029),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3248)] = {
    [sym_heredoc_content] = STATE(3248),
    [anon_sym_RBRACE] = ACTIONS(3562),
    [anon_sym_COLON] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_RPAREN] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(5031),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3249)] = {
    [sym_heredoc_content] = STATE(3249),
    [anon_sym_RBRACE] = ACTIONS(3538),
    [anon_sym_COLON] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_RPAREN] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5033),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3250)] = {
    [sym_heredoc_content] = STATE(3250),
    [anon_sym_RBRACE] = ACTIONS(3544),
    [anon_sym_COLON] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_RPAREN] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5035),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3251)] = {
    [sym_heredoc_content] = STATE(3251),
    [anon_sym_RBRACE] = ACTIONS(3578),
    [anon_sym_COLON] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_RPAREN] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5037),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3252)] = {
    [sym_heredoc_content] = STATE(3252),
    [anon_sym_RBRACE] = ACTIONS(3441),
    [anon_sym_COLON] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_RPAREN] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5039),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3253)] = {
    [sym_heredoc_content] = STATE(3253),
    [anon_sym_RBRACE] = ACTIONS(3584),
    [anon_sym_COLON] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_RPAREN] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5041),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3254)] = {
    [sym_heredoc_content] = STATE(3254),
    [anon_sym_RBRACE] = ACTIONS(3550),
    [anon_sym_COLON] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_RPAREN] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5043),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3255)] = {
    [sym_heredoc_content] = STATE(3255),
    [anon_sym_COLON] = ACTIONS(5045),
    [anon_sym_DOLLAR] = ACTIONS(5045),
    [anon_sym_EQ] = ACTIONS(5047),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_AT] = ACTIONS(5045),
    [anon_sym_COMMA] = ACTIONS(5045),
    [anon_sym_RPAREN] = ACTIONS(5045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5049),
    [anon_sym_PLUS_EQ] = ACTIONS(5049),
    [anon_sym_DASH_EQ] = ACTIONS(5049),
    [anon_sym_DOT_EQ] = ACTIONS(5049),
    [anon_sym_STAR_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_PERCENT_EQ] = ACTIONS(5047),
    [anon_sym_x_EQ] = ACTIONS(5049),
    [anon_sym_AMP_EQ] = ACTIONS(5049),
    [anon_sym_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_CARET_EQ] = ACTIONS(5049),
    [anon_sym_LT_LT_EQ] = ACTIONS(5049),
    [anon_sym_GT_GT_EQ] = ACTIONS(5049),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5049),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(5045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3256)] = {
    [sym_heredoc_content] = STATE(3256),
    [anon_sym_LBRACE] = ACTIONS(3423),
    [anon_sym_SEMI] = ACTIONS(3425),
    [anon_sym_EQ] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_COMMA] = ACTIONS(3425),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3425),
    [anon_sym_DASH_GT] = ACTIONS(3425),
    [anon_sym_STAR] = ACTIONS(3427),
    [anon_sym_LBRACK] = ACTIONS(3423),
    [anon_sym_PERCENT] = ACTIONS(3427),
    [anon_sym_GT] = ACTIONS(3427),
    [anon_sym_LT_LT] = ACTIONS(3427),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3425),
    [anon_sym_PLUS_EQ] = ACTIONS(3425),
    [anon_sym_DASH_EQ] = ACTIONS(3425),
    [anon_sym_DOT_EQ] = ACTIONS(3425),
    [anon_sym_STAR_EQ] = ACTIONS(3425),
    [anon_sym_SLASH_EQ] = ACTIONS(3425),
    [anon_sym_PERCENT_EQ] = ACTIONS(3425),
    [anon_sym_x_EQ] = ACTIONS(3425),
    [anon_sym_AMP_EQ] = ACTIONS(3425),
    [anon_sym_PIPE_EQ] = ACTIONS(3425),
    [anon_sym_CARET_EQ] = ACTIONS(3425),
    [anon_sym_LT_LT_EQ] = ACTIONS(3425),
    [anon_sym_GT_GT_EQ] = ACTIONS(3425),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3425),
    [anon_sym_DOT_DOT] = ACTIONS(3427),
    [anon_sym_STAR_STAR] = ACTIONS(3427),
    [anon_sym_PIPE_PIPE] = ACTIONS(3427),
    [anon_sym_SLASH_SLASH] = ACTIONS(3427),
    [anon_sym_CARET_CARET] = ACTIONS(3425),
    [anon_sym_AMP_AMP] = ACTIONS(3427),
    [anon_sym_PIPE] = ACTIONS(3427),
    [anon_sym_CARET] = ACTIONS(3427),
    [anon_sym_AMP] = ACTIONS(3427),
    [anon_sym_GT_GT2] = ACTIONS(3427),
    [anon_sym_PLUS] = ACTIONS(3427),
    [anon_sym_DASH] = ACTIONS(3427),
    [anon_sym_DOT] = ACTIONS(3427),
    [anon_sym_SLASH] = ACTIONS(3427),
    [anon_sym_x] = ACTIONS(3427),
    [anon_sym_EQ_TILDE] = ACTIONS(3425),
    [anon_sym_BANG_TILDE] = ACTIONS(3425),
    [anon_sym_EQ_EQ] = ACTIONS(3427),
    [anon_sym_BANG_EQ] = ACTIONS(3425),
    [anon_sym_eq] = ACTIONS(3427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3425),
    [anon_sym_equ] = ACTIONS(3425),
    [anon_sym_eqr] = ACTIONS(3425),
    [anon_sym_ne] = ACTIONS(3425),
    [anon_sym_LT_EQ_GT] = ACTIONS(3425),
    [anon_sym_cmp] = ACTIONS(3425),
    [anon_sym_TILDE_TILDE] = ACTIONS(3425),
    [anon_sym_LT] = ACTIONS(3427),
    [anon_sym_LT_EQ] = ACTIONS(3427),
    [anon_sym_GT_EQ] = ACTIONS(3425),
    [anon_sym_lt] = ACTIONS(3425),
    [anon_sym_le] = ACTIONS(3425),
    [anon_sym_ge] = ACTIONS(3425),
    [anon_sym_gt] = ACTIONS(3425),
    [anon_sym_isa] = ACTIONS(3425),
    [anon_sym_PLUS_PLUS] = ACTIONS(3425),
    [anon_sym_DASH_DASH] = ACTIONS(3425),
    [anon_sym_QMARK] = ACTIONS(3425),
    [anon_sym_EQ_GT] = ACTIONS(3425),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3425),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3257)] = {
    [sym_heredoc_content] = STATE(3257),
    [anon_sym_LBRACE] = ACTIONS(3429),
    [anon_sym_SEMI] = ACTIONS(3431),
    [anon_sym_EQ] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_COMMA] = ACTIONS(3431),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3431),
    [anon_sym_DASH_GT] = ACTIONS(3431),
    [anon_sym_STAR] = ACTIONS(3433),
    [anon_sym_LBRACK] = ACTIONS(3429),
    [anon_sym_PERCENT] = ACTIONS(3433),
    [anon_sym_GT] = ACTIONS(3433),
    [anon_sym_LT_LT] = ACTIONS(3433),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3431),
    [anon_sym_PLUS_EQ] = ACTIONS(3431),
    [anon_sym_DASH_EQ] = ACTIONS(3431),
    [anon_sym_DOT_EQ] = ACTIONS(3431),
    [anon_sym_STAR_EQ] = ACTIONS(3431),
    [anon_sym_SLASH_EQ] = ACTIONS(3431),
    [anon_sym_PERCENT_EQ] = ACTIONS(3431),
    [anon_sym_x_EQ] = ACTIONS(3431),
    [anon_sym_AMP_EQ] = ACTIONS(3431),
    [anon_sym_PIPE_EQ] = ACTIONS(3431),
    [anon_sym_CARET_EQ] = ACTIONS(3431),
    [anon_sym_LT_LT_EQ] = ACTIONS(3431),
    [anon_sym_GT_GT_EQ] = ACTIONS(3431),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3431),
    [anon_sym_DOT_DOT] = ACTIONS(3433),
    [anon_sym_STAR_STAR] = ACTIONS(3433),
    [anon_sym_PIPE_PIPE] = ACTIONS(3433),
    [anon_sym_SLASH_SLASH] = ACTIONS(3433),
    [anon_sym_CARET_CARET] = ACTIONS(3431),
    [anon_sym_AMP_AMP] = ACTIONS(3433),
    [anon_sym_PIPE] = ACTIONS(3433),
    [anon_sym_CARET] = ACTIONS(3433),
    [anon_sym_AMP] = ACTIONS(3433),
    [anon_sym_GT_GT2] = ACTIONS(3433),
    [anon_sym_PLUS] = ACTIONS(3433),
    [anon_sym_DASH] = ACTIONS(3433),
    [anon_sym_DOT] = ACTIONS(3433),
    [anon_sym_SLASH] = ACTIONS(3433),
    [anon_sym_x] = ACTIONS(3433),
    [anon_sym_EQ_TILDE] = ACTIONS(3431),
    [anon_sym_BANG_TILDE] = ACTIONS(3431),
    [anon_sym_EQ_EQ] = ACTIONS(3433),
    [anon_sym_BANG_EQ] = ACTIONS(3431),
    [anon_sym_eq] = ACTIONS(3433),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3431),
    [anon_sym_equ] = ACTIONS(3431),
    [anon_sym_eqr] = ACTIONS(3431),
    [anon_sym_ne] = ACTIONS(3431),
    [anon_sym_LT_EQ_GT] = ACTIONS(3431),
    [anon_sym_cmp] = ACTIONS(3431),
    [anon_sym_TILDE_TILDE] = ACTIONS(3431),
    [anon_sym_LT] = ACTIONS(3433),
    [anon_sym_LT_EQ] = ACTIONS(3433),
    [anon_sym_GT_EQ] = ACTIONS(3431),
    [anon_sym_lt] = ACTIONS(3431),
    [anon_sym_le] = ACTIONS(3431),
    [anon_sym_ge] = ACTIONS(3431),
    [anon_sym_gt] = ACTIONS(3431),
    [anon_sym_isa] = ACTIONS(3431),
    [anon_sym_PLUS_PLUS] = ACTIONS(3431),
    [anon_sym_DASH_DASH] = ACTIONS(3431),
    [anon_sym_QMARK] = ACTIONS(3431),
    [anon_sym_EQ_GT] = ACTIONS(3431),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3431),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3258)] = {
    [sym_heredoc_content] = STATE(3258),
    [anon_sym_COLON] = ACTIONS(3897),
    [anon_sym_DOLLAR] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5047),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_AT] = ACTIONS(3897),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5049),
    [anon_sym_PLUS_EQ] = ACTIONS(5049),
    [anon_sym_DASH_EQ] = ACTIONS(5049),
    [anon_sym_DOT_EQ] = ACTIONS(5049),
    [anon_sym_STAR_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_PERCENT_EQ] = ACTIONS(5047),
    [anon_sym_x_EQ] = ACTIONS(5049),
    [anon_sym_AMP_EQ] = ACTIONS(5049),
    [anon_sym_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_CARET_EQ] = ACTIONS(5049),
    [anon_sym_LT_LT_EQ] = ACTIONS(5049),
    [anon_sym_GT_GT_EQ] = ACTIONS(5049),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5049),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3259)] = {
    [sym_heredoc_content] = STATE(3259),
    [anon_sym_COLON] = ACTIONS(3997),
    [anon_sym_DOLLAR] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_AT] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3999),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_PERCENT2] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3260)] = {
    [sym_heredoc_content] = STATE(3260),
    [anon_sym_COLON] = ACTIONS(3985),
    [anon_sym_DOLLAR] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_AT] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3987),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3261)] = {
    [sym_heredoc_content] = STATE(3261),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_DOLLAR] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_AT] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3697),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_PERCENT2] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3262)] = {
    [sym_heredoc_content] = STATE(3262),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3263)] = {
    [sym_heredoc_content] = STATE(3263),
    [anon_sym_COLON] = ACTIONS(4151),
    [anon_sym_DOLLAR] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_AT] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4153),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_PERCENT2] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3264)] = {
    [sym_heredoc_content] = STATE(3264),
    [anon_sym_COLON] = ACTIONS(3909),
    [anon_sym_DOLLAR] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_AT] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3911),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_PERCENT2] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3265)] = {
    [sym_heredoc_content] = STATE(3265),
    [anon_sym_COLON] = ACTIONS(3977),
    [anon_sym_DOLLAR] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_AT] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3979),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_PERCENT2] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3266)] = {
    [sym_heredoc_content] = STATE(3266),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_DOLLAR] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_AT] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4019),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_PERCENT2] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3267)] = {
    [sym_heredoc_content] = STATE(3267),
    [anon_sym_COLON] = ACTIONS(3803),
    [anon_sym_DOLLAR] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5047),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_AT] = ACTIONS(3803),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5049),
    [anon_sym_PLUS_EQ] = ACTIONS(5049),
    [anon_sym_DASH_EQ] = ACTIONS(5049),
    [anon_sym_DOT_EQ] = ACTIONS(5049),
    [anon_sym_STAR_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_PERCENT_EQ] = ACTIONS(5047),
    [anon_sym_x_EQ] = ACTIONS(5049),
    [anon_sym_AMP_EQ] = ACTIONS(5049),
    [anon_sym_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_CARET_EQ] = ACTIONS(5049),
    [anon_sym_LT_LT_EQ] = ACTIONS(5049),
    [anon_sym_GT_GT_EQ] = ACTIONS(5049),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5049),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3268)] = {
    [sym_heredoc_content] = STATE(3268),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_and] = ACTIONS(4045),
    [anon_sym_or] = ACTIONS(4045),
    [anon_sym_xor] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3269)] = {
    [sym_heredoc_content] = STATE(3269),
    [anon_sym_COLON] = ACTIONS(3845),
    [anon_sym_DOLLAR] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_AT] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3847),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_PERCENT2] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3270)] = {
    [sym_heredoc_content] = STATE(3270),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_and] = ACTIONS(3917),
    [anon_sym_or] = ACTIONS(3917),
    [anon_sym_xor] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3271)] = {
    [sym_heredoc_content] = STATE(3271),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(4212),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_and] = ACTIONS(3415),
    [anon_sym_or] = ACTIONS(3415),
    [anon_sym_xor] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(4212),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3272)] = {
    [sym_heredoc_content] = STATE(3272),
    [anon_sym_LBRACE] = ACTIONS(3439),
    [anon_sym_SEMI] = ACTIONS(2686),
    [anon_sym_EQ] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_COMMA] = ACTIONS(2686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2686),
    [anon_sym_DASH_GT] = ACTIONS(2686),
    [anon_sym_STAR] = ACTIONS(2682),
    [anon_sym_LBRACK] = ACTIONS(3439),
    [anon_sym_PERCENT] = ACTIONS(2682),
    [anon_sym_GT] = ACTIONS(2682),
    [anon_sym_LT_LT] = ACTIONS(2682),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2686),
    [anon_sym_PLUS_EQ] = ACTIONS(2686),
    [anon_sym_DASH_EQ] = ACTIONS(2686),
    [anon_sym_DOT_EQ] = ACTIONS(2686),
    [anon_sym_STAR_EQ] = ACTIONS(2686),
    [anon_sym_SLASH_EQ] = ACTIONS(2686),
    [anon_sym_PERCENT_EQ] = ACTIONS(2686),
    [anon_sym_x_EQ] = ACTIONS(2686),
    [anon_sym_AMP_EQ] = ACTIONS(2686),
    [anon_sym_PIPE_EQ] = ACTIONS(2686),
    [anon_sym_CARET_EQ] = ACTIONS(2686),
    [anon_sym_LT_LT_EQ] = ACTIONS(2686),
    [anon_sym_GT_GT_EQ] = ACTIONS(2686),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2686),
    [anon_sym_DOT_DOT] = ACTIONS(2682),
    [anon_sym_STAR_STAR] = ACTIONS(2682),
    [anon_sym_PIPE_PIPE] = ACTIONS(2682),
    [anon_sym_SLASH_SLASH] = ACTIONS(2682),
    [anon_sym_CARET_CARET] = ACTIONS(2686),
    [anon_sym_AMP_AMP] = ACTIONS(2682),
    [anon_sym_PIPE] = ACTIONS(2682),
    [anon_sym_CARET] = ACTIONS(2682),
    [anon_sym_AMP] = ACTIONS(2682),
    [anon_sym_GT_GT2] = ACTIONS(2682),
    [anon_sym_PLUS] = ACTIONS(2682),
    [anon_sym_DASH] = ACTIONS(2682),
    [anon_sym_DOT] = ACTIONS(2682),
    [anon_sym_SLASH] = ACTIONS(2682),
    [anon_sym_x] = ACTIONS(2682),
    [anon_sym_EQ_TILDE] = ACTIONS(2686),
    [anon_sym_BANG_TILDE] = ACTIONS(2686),
    [anon_sym_EQ_EQ] = ACTIONS(2682),
    [anon_sym_BANG_EQ] = ACTIONS(2686),
    [anon_sym_eq] = ACTIONS(2682),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2686),
    [anon_sym_equ] = ACTIONS(2686),
    [anon_sym_eqr] = ACTIONS(2686),
    [anon_sym_ne] = ACTIONS(2686),
    [anon_sym_LT_EQ_GT] = ACTIONS(2686),
    [anon_sym_cmp] = ACTIONS(2686),
    [anon_sym_TILDE_TILDE] = ACTIONS(2686),
    [anon_sym_LT] = ACTIONS(2682),
    [anon_sym_LT_EQ] = ACTIONS(2682),
    [anon_sym_GT_EQ] = ACTIONS(2686),
    [anon_sym_lt] = ACTIONS(2686),
    [anon_sym_le] = ACTIONS(2686),
    [anon_sym_ge] = ACTIONS(2686),
    [anon_sym_gt] = ACTIONS(2686),
    [anon_sym_isa] = ACTIONS(2686),
    [anon_sym_PLUS_PLUS] = ACTIONS(2686),
    [anon_sym_DASH_DASH] = ACTIONS(2686),
    [anon_sym_QMARK] = ACTIONS(2686),
    [anon_sym_EQ_GT] = ACTIONS(2686),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2686),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3273)] = {
    [sym_heredoc_content] = STATE(3273),
    [anon_sym_RBRACE] = ACTIONS(3633),
    [anon_sym_COLON] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_RPAREN] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(5079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3274)] = {
    [sym_heredoc_content] = STATE(3274),
    [anon_sym_RBRACE] = ACTIONS(3469),
    [anon_sym_COLON] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_RPAREN] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5081),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3275)] = {
    [sym_heredoc_content] = STATE(3275),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_LBRACK] = ACTIONS(3435),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3435),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3435),
    [anon_sym_le] = ACTIONS(3435),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3435),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(3435),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3276)] = {
    [sym_heredoc_content] = STATE(3276),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3277)] = {
    [sym_heredoc_content] = STATE(3277),
    [anon_sym_LBRACE] = ACTIONS(5085),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_COLON] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3278)] = {
    [sym_heredoc_content] = STATE(3278),
    [anon_sym_RBRACE] = ACTIONS(3621),
    [anon_sym_COLON] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_RPAREN] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(5087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3279)] = {
    [sym_heredoc_content] = STATE(3279),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(5089),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3280)] = {
    [sym_heredoc_content] = STATE(3280),
    [anon_sym_COLON] = ACTIONS(4107),
    [anon_sym_DOLLAR] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5047),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_AT] = ACTIONS(4107),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5049),
    [anon_sym_PLUS_EQ] = ACTIONS(5049),
    [anon_sym_DASH_EQ] = ACTIONS(5049),
    [anon_sym_DOT_EQ] = ACTIONS(5049),
    [anon_sym_STAR_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_PERCENT_EQ] = ACTIONS(5047),
    [anon_sym_x_EQ] = ACTIONS(5049),
    [anon_sym_AMP_EQ] = ACTIONS(5049),
    [anon_sym_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_CARET_EQ] = ACTIONS(5049),
    [anon_sym_LT_LT_EQ] = ACTIONS(5049),
    [anon_sym_GT_GT_EQ] = ACTIONS(5049),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5049),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3281)] = {
    [sym_heredoc_content] = STATE(3281),
    [anon_sym_RBRACE] = ACTIONS(3457),
    [anon_sym_COLON] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_RPAREN] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(5091),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3282)] = {
    [sym_heredoc_content] = STATE(3282),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(5087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3283)] = {
    [sym_heredoc_content] = STATE(3283),
    [anon_sym_COLON] = ACTIONS(5093),
    [anon_sym_DOLLAR] = ACTIONS(5093),
    [anon_sym_EQ] = ACTIONS(5047),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_AT] = ACTIONS(5093),
    [anon_sym_COMMA] = ACTIONS(5093),
    [anon_sym_RPAREN] = ACTIONS(5093),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5051),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5049),
    [anon_sym_PLUS_EQ] = ACTIONS(5049),
    [anon_sym_DASH_EQ] = ACTIONS(5049),
    [anon_sym_DOT_EQ] = ACTIONS(5049),
    [anon_sym_STAR_EQ] = ACTIONS(5049),
    [anon_sym_SLASH_EQ] = ACTIONS(5049),
    [anon_sym_PERCENT_EQ] = ACTIONS(5047),
    [anon_sym_x_EQ] = ACTIONS(5049),
    [anon_sym_AMP_EQ] = ACTIONS(5049),
    [anon_sym_PIPE_EQ] = ACTIONS(5049),
    [anon_sym_CARET_EQ] = ACTIONS(5049),
    [anon_sym_LT_LT_EQ] = ACTIONS(5049),
    [anon_sym_GT_GT_EQ] = ACTIONS(5049),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5049),
    [anon_sym_DOT_DOT] = ACTIONS(5055),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5077),
    [anon_sym_PERCENT2] = ACTIONS(5093),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3284)] = {
    [sym_heredoc_content] = STATE(3284),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5095),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5097),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(5057),
    [anon_sym_SLASH_SLASH] = ACTIONS(5057),
    [anon_sym_CARET_CARET] = ACTIONS(5059),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3285)] = {
    [sym_heredoc_content] = STATE(3285),
    [anon_sym_RBRACE] = ACTIONS(3601),
    [anon_sym_COLON] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_RPAREN] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3286)] = {
    [sym_heredoc_content] = STATE(3286),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_RBRACE] = ACTIONS(3605),
    [anon_sym_COLON] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_RPAREN] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3287)] = {
    [sym_heredoc_content] = STATE(3287),
    [anon_sym_RBRACE] = ACTIONS(3655),
    [anon_sym_COLON] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_RPAREN] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3288)] = {
    [sym_heredoc_content] = STATE(3288),
    [anon_sym_RBRACE] = ACTIONS(3649),
    [anon_sym_COLON] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_RPAREN] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(5091),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3289)] = {
    [sym_heredoc_content] = STATE(3289),
    [anon_sym_RBRACE] = ACTIONS(3451),
    [anon_sym_COLON] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_RPAREN] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(5021),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3290)] = {
    [sym_heredoc_content] = STATE(3290),
    [anon_sym_RBRACE] = ACTIONS(3639),
    [anon_sym_COLON] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_RPAREN] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3291)] = {
    [sym_heredoc_content] = STATE(3291),
    [anon_sym_RBRACE] = ACTIONS(3643),
    [anon_sym_COLON] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_RPAREN] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(5101),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3292)] = {
    [sym_heredoc_content] = STATE(3292),
    [anon_sym_RBRACE] = ACTIONS(3463),
    [anon_sym_COLON] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_RPAREN] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3293)] = {
    [sym_heredoc_content] = STATE(3293),
    [anon_sym_RBRACE] = ACTIONS(3475),
    [anon_sym_COLON] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_RPAREN] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3294)] = {
    [sym_heredoc_content] = STATE(3294),
    [anon_sym_RBRACE] = ACTIONS(3479),
    [anon_sym_COLON] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_RPAREN] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(5105),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3295)] = {
    [sym_heredoc_content] = STATE(3295),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3296)] = {
    [sym_heredoc_content] = STATE(3296),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3297)] = {
    [sym_heredoc_content] = STATE(3297),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3298)] = {
    [sym_heredoc_content] = STATE(3298),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3299)] = {
    [sym_heredoc_content] = STATE(3299),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5061),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3300)] = {
    [sym_heredoc_content] = STATE(3300),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5063),
    [anon_sym_CARET] = ACTIONS(5063),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3301)] = {
    [sym_heredoc_content] = STATE(3301),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5065),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3302)] = {
    [sym_heredoc_content] = STATE(3302),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5071),
    [anon_sym_cmp] = ACTIONS(5071),
    [anon_sym_TILDE_TILDE] = ACTIONS(5071),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3303)] = {
    [sym_heredoc_content] = STATE(3303),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3304)] = {
    [sym_heredoc_content] = STATE(3304),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3305)] = {
    [sym_heredoc_content] = STATE(3305),
    [anon_sym_LBRACE] = ACTIONS(2858),
    [anon_sym_COLON] = ACTIONS(2858),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3306)] = {
    [sym_heredoc_content] = STATE(3306),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3307)] = {
    [sym_heredoc_content] = STATE(3307),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3308)] = {
    [sym_heredoc_content] = STATE(3308),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5009),
    [anon_sym_PERCENT] = ACTIONS(5009),
    [anon_sym_GT] = ACTIONS(5053),
    [anon_sym_LT_LT] = ACTIONS(5011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5013),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5011),
    [anon_sym_PLUS] = ACTIONS(5015),
    [anon_sym_DASH] = ACTIONS(5015),
    [anon_sym_DOT] = ACTIONS(5015),
    [anon_sym_SLASH] = ACTIONS(5009),
    [anon_sym_x] = ACTIONS(5009),
    [anon_sym_EQ_TILDE] = ACTIONS(5017),
    [anon_sym_BANG_TILDE] = ACTIONS(5017),
    [anon_sym_EQ_EQ] = ACTIONS(5067),
    [anon_sym_BANG_EQ] = ACTIONS(5069),
    [anon_sym_eq] = ACTIONS(5067),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5069),
    [anon_sym_equ] = ACTIONS(5069),
    [anon_sym_eqr] = ACTIONS(5069),
    [anon_sym_ne] = ACTIONS(5069),
    [anon_sym_LT_EQ_GT] = ACTIONS(5107),
    [anon_sym_cmp] = ACTIONS(5107),
    [anon_sym_TILDE_TILDE] = ACTIONS(5107),
    [anon_sym_LT] = ACTIONS(5053),
    [anon_sym_LT_EQ] = ACTIONS(5053),
    [anon_sym_GT_EQ] = ACTIONS(5073),
    [anon_sym_lt] = ACTIONS(5073),
    [anon_sym_le] = ACTIONS(5073),
    [anon_sym_ge] = ACTIONS(5073),
    [anon_sym_gt] = ACTIONS(5073),
    [anon_sym_isa] = ACTIONS(5075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3309)] = {
    [sym_heredoc_content] = STATE(3309),
    [anon_sym_RBRACE] = ACTIONS(3435),
    [anon_sym_COLON] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_RPAREN] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3310)] = {
    [sym_heredoc_content] = STATE(3310),
    [anon_sym_SEMI] = ACTIONS(3584),
    [anon_sym_EQ] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_COMMA] = ACTIONS(3584),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3584),
    [anon_sym_DASH_GT] = ACTIONS(3584),
    [anon_sym_STAR] = ACTIONS(3586),
    [anon_sym_PERCENT] = ACTIONS(3586),
    [anon_sym_GT] = ACTIONS(3586),
    [anon_sym_LT_LT] = ACTIONS(3586),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3584),
    [anon_sym_PLUS_EQ] = ACTIONS(3584),
    [anon_sym_DASH_EQ] = ACTIONS(3584),
    [anon_sym_DOT_EQ] = ACTIONS(3584),
    [anon_sym_STAR_EQ] = ACTIONS(3584),
    [anon_sym_SLASH_EQ] = ACTIONS(3584),
    [anon_sym_PERCENT_EQ] = ACTIONS(3584),
    [anon_sym_x_EQ] = ACTIONS(3586),
    [anon_sym_AMP_EQ] = ACTIONS(3584),
    [anon_sym_PIPE_EQ] = ACTIONS(3584),
    [anon_sym_CARET_EQ] = ACTIONS(3584),
    [anon_sym_LT_LT_EQ] = ACTIONS(3584),
    [anon_sym_GT_GT_EQ] = ACTIONS(3584),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3584),
    [anon_sym_DOT_DOT] = ACTIONS(3586),
    [anon_sym_STAR_STAR] = ACTIONS(3586),
    [anon_sym_PIPE_PIPE] = ACTIONS(3586),
    [anon_sym_SLASH_SLASH] = ACTIONS(3586),
    [anon_sym_CARET_CARET] = ACTIONS(3584),
    [anon_sym_AMP_AMP] = ACTIONS(3586),
    [anon_sym_PIPE] = ACTIONS(3586),
    [anon_sym_CARET] = ACTIONS(3586),
    [anon_sym_AMP] = ACTIONS(3586),
    [anon_sym_GT_GT2] = ACTIONS(3586),
    [anon_sym_PLUS] = ACTIONS(3586),
    [anon_sym_DASH] = ACTIONS(3586),
    [anon_sym_DOT] = ACTIONS(3586),
    [anon_sym_SLASH] = ACTIONS(3586),
    [anon_sym_x] = ACTIONS(3586),
    [anon_sym_EQ_TILDE] = ACTIONS(3584),
    [anon_sym_BANG_TILDE] = ACTIONS(3584),
    [anon_sym_EQ_EQ] = ACTIONS(3586),
    [anon_sym_BANG_EQ] = ACTIONS(3584),
    [anon_sym_eq] = ACTIONS(3586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3584),
    [anon_sym_equ] = ACTIONS(3586),
    [anon_sym_eqr] = ACTIONS(3586),
    [anon_sym_ne] = ACTIONS(3584),
    [anon_sym_LT_EQ_GT] = ACTIONS(3584),
    [anon_sym_cmp] = ACTIONS(3586),
    [anon_sym_TILDE_TILDE] = ACTIONS(3584),
    [anon_sym_LT] = ACTIONS(3586),
    [anon_sym_LT_EQ] = ACTIONS(3586),
    [anon_sym_GT_EQ] = ACTIONS(3584),
    [anon_sym_lt] = ACTIONS(3586),
    [anon_sym_le] = ACTIONS(3586),
    [anon_sym_ge] = ACTIONS(3586),
    [anon_sym_gt] = ACTIONS(3586),
    [anon_sym_isa] = ACTIONS(3586),
    [anon_sym_PLUS_PLUS] = ACTIONS(3584),
    [anon_sym_DASH_DASH] = ACTIONS(3584),
    [anon_sym_QMARK] = ACTIONS(3584),
    [anon_sym_EQ_GT] = ACTIONS(3584),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5109),
    [sym__PERLY_SEMICOLON] = ACTIONS(3584),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3311)] = {
    [sym_heredoc_content] = STATE(3311),
    [anon_sym_RBRACE] = ACTIONS(4029),
    [anon_sym_COLON] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_RPAREN] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3312)] = {
    [sym_heredoc_content] = STATE(3312),
    [anon_sym_RBRACE] = ACTIONS(4111),
    [anon_sym_COLON] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_RPAREN] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3313)] = {
    [sym_heredoc_content] = STATE(3313),
    [anon_sym_RBRACE] = ACTIONS(4123),
    [anon_sym_COLON] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_RPAREN] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3314)] = {
    [sym_heredoc_content] = STATE(3314),
    [anon_sym_RBRACE] = ACTIONS(3945),
    [anon_sym_COLON] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_RPAREN] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3315)] = {
    [sym_heredoc_content] = STATE(3315),
    [anon_sym_RBRACE] = ACTIONS(4196),
    [anon_sym_COLON] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_RPAREN] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3316)] = {
    [sym_heredoc_content] = STATE(3316),
    [anon_sym_RBRACE] = ACTIONS(4180),
    [anon_sym_COLON] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_RPAREN] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3317)] = {
    [sym_heredoc_content] = STATE(3317),
    [anon_sym_RBRACE] = ACTIONS(3905),
    [anon_sym_COLON] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_RPAREN] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3318)] = {
    [sym_heredoc_content] = STATE(3318),
    [anon_sym_RBRACE] = ACTIONS(4067),
    [anon_sym_COLON] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_RPAREN] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3319)] = {
    [sym_heredoc_content] = STATE(3319),
    [anon_sym_RBRACE] = ACTIONS(3795),
    [anon_sym_COLON] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_RPAREN] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3320)] = {
    [sym_heredoc_content] = STATE(3320),
    [anon_sym_RBRACE] = ACTIONS(3825),
    [anon_sym_COLON] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_RPAREN] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3321)] = {
    [sym_heredoc_content] = STATE(3321),
    [anon_sym_RBRACE] = ACTIONS(3849),
    [anon_sym_COLON] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_RPAREN] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3322)] = {
    [sym_heredoc_content] = STATE(3322),
    [anon_sym_RBRACE] = ACTIONS(4051),
    [anon_sym_COLON] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_RPAREN] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3323)] = {
    [sym_heredoc_content] = STATE(3323),
    [anon_sym_RBRACE] = ACTIONS(4059),
    [anon_sym_COLON] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_RPAREN] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3324)] = {
    [sym_heredoc_content] = STATE(3324),
    [anon_sym_RBRACE] = ACTIONS(4099),
    [anon_sym_COLON] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_RPAREN] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3325)] = {
    [sym_heredoc_content] = STATE(3325),
    [anon_sym_RBRACE] = ACTIONS(4127),
    [anon_sym_COLON] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_RPAREN] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3326)] = {
    [sym_heredoc_content] = STATE(3326),
    [anon_sym_RBRACE] = ACTIONS(4147),
    [anon_sym_COLON] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_RPAREN] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3327)] = {
    [sym_heredoc_content] = STATE(3327),
    [anon_sym_RBRACE] = ACTIONS(3961),
    [anon_sym_COLON] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_RPAREN] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3328)] = {
    [sym_heredoc_content] = STATE(3328),
    [anon_sym_RBRACE] = ACTIONS(3913),
    [anon_sym_COLON] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_RPAREN] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3329)] = {
    [sym_heredoc_content] = STATE(3329),
    [anon_sym_RBRACE] = ACTIONS(4143),
    [anon_sym_COLON] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_RPAREN] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3330)] = {
    [sym_heredoc_content] = STATE(3330),
    [anon_sym_RBRACE] = ACTIONS(3799),
    [anon_sym_COLON] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_RPAREN] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3331)] = {
    [sym_heredoc_content] = STATE(3331),
    [anon_sym_RBRACE] = ACTIONS(3921),
    [anon_sym_COLON] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_RPAREN] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3332)] = {
    [sym_heredoc_content] = STATE(3332),
    [anon_sym_RBRACE] = ACTIONS(3723),
    [anon_sym_COLON] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_RPAREN] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3333)] = {
    [sym_heredoc_content] = STATE(3333),
    [anon_sym_RBRACE] = ACTIONS(3751),
    [anon_sym_COLON] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_RPAREN] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3334)] = {
    [sym_heredoc_content] = STATE(3334),
    [anon_sym_RBRACE] = ACTIONS(3755),
    [anon_sym_COLON] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_RPAREN] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3335)] = {
    [sym_heredoc_content] = STATE(3335),
    [anon_sym_RBRACE] = ACTIONS(3759),
    [anon_sym_COLON] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_RPAREN] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3336)] = {
    [sym_heredoc_content] = STATE(3336),
    [anon_sym_SEMI] = ACTIONS(3633),
    [anon_sym_EQ] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_COMMA] = ACTIONS(3633),
    [anon_sym_LPAREN] = ACTIONS(5111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3633),
    [anon_sym_DASH_GT] = ACTIONS(3633),
    [anon_sym_STAR] = ACTIONS(3635),
    [anon_sym_PERCENT] = ACTIONS(3635),
    [anon_sym_GT] = ACTIONS(3635),
    [anon_sym_LT_LT] = ACTIONS(3635),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3633),
    [anon_sym_PLUS_EQ] = ACTIONS(3633),
    [anon_sym_DASH_EQ] = ACTIONS(3633),
    [anon_sym_DOT_EQ] = ACTIONS(3633),
    [anon_sym_STAR_EQ] = ACTIONS(3633),
    [anon_sym_SLASH_EQ] = ACTIONS(3633),
    [anon_sym_PERCENT_EQ] = ACTIONS(3633),
    [anon_sym_x_EQ] = ACTIONS(3633),
    [anon_sym_AMP_EQ] = ACTIONS(3633),
    [anon_sym_PIPE_EQ] = ACTIONS(3633),
    [anon_sym_CARET_EQ] = ACTIONS(3633),
    [anon_sym_LT_LT_EQ] = ACTIONS(3633),
    [anon_sym_GT_GT_EQ] = ACTIONS(3633),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3633),
    [anon_sym_DOT_DOT] = ACTIONS(3635),
    [anon_sym_STAR_STAR] = ACTIONS(3635),
    [anon_sym_PIPE_PIPE] = ACTIONS(3635),
    [anon_sym_SLASH_SLASH] = ACTIONS(3635),
    [anon_sym_CARET_CARET] = ACTIONS(3633),
    [anon_sym_AMP_AMP] = ACTIONS(3635),
    [anon_sym_PIPE] = ACTIONS(3635),
    [anon_sym_CARET] = ACTIONS(3635),
    [anon_sym_AMP] = ACTIONS(3635),
    [anon_sym_GT_GT2] = ACTIONS(3635),
    [anon_sym_PLUS] = ACTIONS(3635),
    [anon_sym_DASH] = ACTIONS(3635),
    [anon_sym_DOT] = ACTIONS(3635),
    [anon_sym_SLASH] = ACTIONS(3635),
    [anon_sym_x] = ACTIONS(3635),
    [anon_sym_EQ_TILDE] = ACTIONS(3633),
    [anon_sym_BANG_TILDE] = ACTIONS(3633),
    [anon_sym_EQ_EQ] = ACTIONS(3635),
    [anon_sym_BANG_EQ] = ACTIONS(3633),
    [anon_sym_eq] = ACTIONS(3635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3633),
    [anon_sym_equ] = ACTIONS(3633),
    [anon_sym_eqr] = ACTIONS(3633),
    [anon_sym_ne] = ACTIONS(3633),
    [anon_sym_LT_EQ_GT] = ACTIONS(3633),
    [anon_sym_cmp] = ACTIONS(3633),
    [anon_sym_TILDE_TILDE] = ACTIONS(3633),
    [anon_sym_LT] = ACTIONS(3635),
    [anon_sym_LT_EQ] = ACTIONS(3635),
    [anon_sym_GT_EQ] = ACTIONS(3633),
    [anon_sym_lt] = ACTIONS(3633),
    [anon_sym_le] = ACTIONS(3633),
    [anon_sym_ge] = ACTIONS(3633),
    [anon_sym_gt] = ACTIONS(3633),
    [anon_sym_isa] = ACTIONS(3633),
    [anon_sym_PLUS_PLUS] = ACTIONS(3633),
    [anon_sym_DASH_DASH] = ACTIONS(3633),
    [anon_sym_QMARK] = ACTIONS(3633),
    [anon_sym_EQ_GT] = ACTIONS(3633),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3633),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3337)] = {
    [sym_heredoc_content] = STATE(3337),
    [anon_sym_SEMI] = ACTIONS(3469),
    [anon_sym_EQ] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_COMMA] = ACTIONS(3469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3469),
    [anon_sym_DASH_GT] = ACTIONS(3469),
    [anon_sym_STAR] = ACTIONS(3471),
    [anon_sym_PERCENT] = ACTIONS(3471),
    [anon_sym_GT] = ACTIONS(3471),
    [anon_sym_LT_LT] = ACTIONS(3471),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3469),
    [anon_sym_PLUS_EQ] = ACTIONS(3469),
    [anon_sym_DASH_EQ] = ACTIONS(3469),
    [anon_sym_DOT_EQ] = ACTIONS(3469),
    [anon_sym_STAR_EQ] = ACTIONS(3469),
    [anon_sym_SLASH_EQ] = ACTIONS(3469),
    [anon_sym_PERCENT_EQ] = ACTIONS(3469),
    [anon_sym_x_EQ] = ACTIONS(3471),
    [anon_sym_AMP_EQ] = ACTIONS(3469),
    [anon_sym_PIPE_EQ] = ACTIONS(3469),
    [anon_sym_CARET_EQ] = ACTIONS(3469),
    [anon_sym_LT_LT_EQ] = ACTIONS(3469),
    [anon_sym_GT_GT_EQ] = ACTIONS(3469),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3469),
    [anon_sym_DOT_DOT] = ACTIONS(3471),
    [anon_sym_STAR_STAR] = ACTIONS(3471),
    [anon_sym_PIPE_PIPE] = ACTIONS(3471),
    [anon_sym_SLASH_SLASH] = ACTIONS(3471),
    [anon_sym_CARET_CARET] = ACTIONS(3469),
    [anon_sym_AMP_AMP] = ACTIONS(3471),
    [anon_sym_PIPE] = ACTIONS(3471),
    [anon_sym_CARET] = ACTIONS(3471),
    [anon_sym_AMP] = ACTIONS(3471),
    [anon_sym_GT_GT2] = ACTIONS(3471),
    [anon_sym_PLUS] = ACTIONS(3471),
    [anon_sym_DASH] = ACTIONS(3471),
    [anon_sym_DOT] = ACTIONS(3471),
    [anon_sym_SLASH] = ACTIONS(3471),
    [anon_sym_x] = ACTIONS(3471),
    [anon_sym_EQ_TILDE] = ACTIONS(3469),
    [anon_sym_BANG_TILDE] = ACTIONS(3469),
    [anon_sym_EQ_EQ] = ACTIONS(3471),
    [anon_sym_BANG_EQ] = ACTIONS(3469),
    [anon_sym_eq] = ACTIONS(3471),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3469),
    [anon_sym_equ] = ACTIONS(3469),
    [anon_sym_eqr] = ACTIONS(3469),
    [anon_sym_ne] = ACTIONS(3471),
    [anon_sym_LT_EQ_GT] = ACTIONS(3469),
    [anon_sym_cmp] = ACTIONS(3471),
    [anon_sym_TILDE_TILDE] = ACTIONS(3469),
    [anon_sym_LT] = ACTIONS(3471),
    [anon_sym_LT_EQ] = ACTIONS(3471),
    [anon_sym_GT_EQ] = ACTIONS(3469),
    [anon_sym_lt] = ACTIONS(3471),
    [anon_sym_le] = ACTIONS(3471),
    [anon_sym_ge] = ACTIONS(3471),
    [anon_sym_gt] = ACTIONS(3471),
    [anon_sym_isa] = ACTIONS(3471),
    [anon_sym_PLUS_PLUS] = ACTIONS(3469),
    [anon_sym_DASH_DASH] = ACTIONS(3469),
    [anon_sym_QMARK] = ACTIONS(3469),
    [anon_sym_EQ_GT] = ACTIONS(3469),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5113),
    [sym__PERLY_SEMICOLON] = ACTIONS(3469),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3338)] = {
    [sym_heredoc_content] = STATE(3338),
    [anon_sym_RBRACE] = ACTIONS(4188),
    [anon_sym_COLON] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_RPAREN] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3339)] = {
    [sym_heredoc_content] = STATE(3339),
    [anon_sym_RBRACE] = ACTIONS(3675),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_RPAREN] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3340)] = {
    [sym_heredoc_content] = STATE(3340),
    [anon_sym_LBRACE] = ACTIONS(5115),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3341)] = {
    [sym_heredoc_content] = STATE(3341),
    [anon_sym_SEMI] = ACTIONS(3621),
    [anon_sym_EQ] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_COMMA] = ACTIONS(3621),
    [anon_sym_LPAREN] = ACTIONS(5117),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3621),
    [anon_sym_DASH_GT] = ACTIONS(3621),
    [anon_sym_STAR] = ACTIONS(3623),
    [anon_sym_PERCENT] = ACTIONS(3623),
    [anon_sym_GT] = ACTIONS(3623),
    [anon_sym_LT_LT] = ACTIONS(3623),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3621),
    [anon_sym_PLUS_EQ] = ACTIONS(3621),
    [anon_sym_DASH_EQ] = ACTIONS(3621),
    [anon_sym_DOT_EQ] = ACTIONS(3621),
    [anon_sym_STAR_EQ] = ACTIONS(3621),
    [anon_sym_SLASH_EQ] = ACTIONS(3621),
    [anon_sym_PERCENT_EQ] = ACTIONS(3621),
    [anon_sym_x_EQ] = ACTIONS(3621),
    [anon_sym_AMP_EQ] = ACTIONS(3621),
    [anon_sym_PIPE_EQ] = ACTIONS(3621),
    [anon_sym_CARET_EQ] = ACTIONS(3621),
    [anon_sym_LT_LT_EQ] = ACTIONS(3621),
    [anon_sym_GT_GT_EQ] = ACTIONS(3621),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3621),
    [anon_sym_DOT_DOT] = ACTIONS(3623),
    [anon_sym_STAR_STAR] = ACTIONS(3623),
    [anon_sym_PIPE_PIPE] = ACTIONS(3623),
    [anon_sym_SLASH_SLASH] = ACTIONS(3623),
    [anon_sym_CARET_CARET] = ACTIONS(3621),
    [anon_sym_AMP_AMP] = ACTIONS(3623),
    [anon_sym_PIPE] = ACTIONS(3623),
    [anon_sym_CARET] = ACTIONS(3623),
    [anon_sym_AMP] = ACTIONS(3623),
    [anon_sym_GT_GT2] = ACTIONS(3623),
    [anon_sym_PLUS] = ACTIONS(3623),
    [anon_sym_DASH] = ACTIONS(3623),
    [anon_sym_DOT] = ACTIONS(3623),
    [anon_sym_SLASH] = ACTIONS(3623),
    [anon_sym_x] = ACTIONS(3623),
    [anon_sym_EQ_TILDE] = ACTIONS(3621),
    [anon_sym_BANG_TILDE] = ACTIONS(3621),
    [anon_sym_EQ_EQ] = ACTIONS(3623),
    [anon_sym_BANG_EQ] = ACTIONS(3621),
    [anon_sym_eq] = ACTIONS(3623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3621),
    [anon_sym_equ] = ACTIONS(3621),
    [anon_sym_eqr] = ACTIONS(3621),
    [anon_sym_ne] = ACTIONS(3621),
    [anon_sym_LT_EQ_GT] = ACTIONS(3621),
    [anon_sym_cmp] = ACTIONS(3621),
    [anon_sym_TILDE_TILDE] = ACTIONS(3621),
    [anon_sym_LT] = ACTIONS(3623),
    [anon_sym_LT_EQ] = ACTIONS(3623),
    [anon_sym_GT_EQ] = ACTIONS(3621),
    [anon_sym_lt] = ACTIONS(3621),
    [anon_sym_le] = ACTIONS(3621),
    [anon_sym_ge] = ACTIONS(3621),
    [anon_sym_gt] = ACTIONS(3621),
    [anon_sym_isa] = ACTIONS(3621),
    [anon_sym_PLUS_PLUS] = ACTIONS(3621),
    [anon_sym_DASH_DASH] = ACTIONS(3621),
    [anon_sym_QMARK] = ACTIONS(3621),
    [anon_sym_EQ_GT] = ACTIONS(3621),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3621),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3342)] = {
    [sym_heredoc_content] = STATE(3342),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_LBRACK] = ACTIONS(5119),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3343)] = {
    [sym_heredoc_content] = STATE(3343),
    [anon_sym_RBRACE] = ACTIONS(4192),
    [anon_sym_COLON] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_RPAREN] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3344)] = {
    [sym_heredoc_content] = STATE(3344),
    [anon_sym_COLON] = ACTIONS(5121),
    [anon_sym_SEMI] = ACTIONS(3689),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3689),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3345)] = {
    [sym_heredoc_content] = STATE(3345),
    [anon_sym_COLON] = ACTIONS(5123),
    [anon_sym_SEMI] = ACTIONS(3669),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3669),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3346)] = {
    [sym_heredoc_content] = STATE(3346),
    [anon_sym_SEMI] = ACTIONS(3457),
    [anon_sym_EQ] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_COMMA] = ACTIONS(3457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3457),
    [anon_sym_DASH_GT] = ACTIONS(3457),
    [anon_sym_STAR] = ACTIONS(3459),
    [anon_sym_LBRACK] = ACTIONS(5125),
    [anon_sym_PERCENT] = ACTIONS(3459),
    [anon_sym_GT] = ACTIONS(3459),
    [anon_sym_LT_LT] = ACTIONS(3459),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3457),
    [anon_sym_PLUS_EQ] = ACTIONS(3457),
    [anon_sym_DASH_EQ] = ACTIONS(3457),
    [anon_sym_DOT_EQ] = ACTIONS(3457),
    [anon_sym_STAR_EQ] = ACTIONS(3457),
    [anon_sym_SLASH_EQ] = ACTIONS(3457),
    [anon_sym_PERCENT_EQ] = ACTIONS(3457),
    [anon_sym_x_EQ] = ACTIONS(3457),
    [anon_sym_AMP_EQ] = ACTIONS(3457),
    [anon_sym_PIPE_EQ] = ACTIONS(3457),
    [anon_sym_CARET_EQ] = ACTIONS(3457),
    [anon_sym_LT_LT_EQ] = ACTIONS(3457),
    [anon_sym_GT_GT_EQ] = ACTIONS(3457),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3457),
    [anon_sym_DOT_DOT] = ACTIONS(3459),
    [anon_sym_STAR_STAR] = ACTIONS(3459),
    [anon_sym_PIPE_PIPE] = ACTIONS(3459),
    [anon_sym_SLASH_SLASH] = ACTIONS(3459),
    [anon_sym_CARET_CARET] = ACTIONS(3457),
    [anon_sym_AMP_AMP] = ACTIONS(3459),
    [anon_sym_PIPE] = ACTIONS(3459),
    [anon_sym_CARET] = ACTIONS(3459),
    [anon_sym_AMP] = ACTIONS(3459),
    [anon_sym_GT_GT2] = ACTIONS(3459),
    [anon_sym_PLUS] = ACTIONS(3459),
    [anon_sym_DASH] = ACTIONS(3459),
    [anon_sym_DOT] = ACTIONS(3459),
    [anon_sym_SLASH] = ACTIONS(3459),
    [anon_sym_x] = ACTIONS(3459),
    [anon_sym_EQ_TILDE] = ACTIONS(3457),
    [anon_sym_BANG_TILDE] = ACTIONS(3457),
    [anon_sym_EQ_EQ] = ACTIONS(3459),
    [anon_sym_BANG_EQ] = ACTIONS(3457),
    [anon_sym_eq] = ACTIONS(3459),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3457),
    [anon_sym_equ] = ACTIONS(3457),
    [anon_sym_eqr] = ACTIONS(3457),
    [anon_sym_ne] = ACTIONS(3457),
    [anon_sym_LT_EQ_GT] = ACTIONS(3457),
    [anon_sym_cmp] = ACTIONS(3457),
    [anon_sym_TILDE_TILDE] = ACTIONS(3457),
    [anon_sym_LT] = ACTIONS(3459),
    [anon_sym_LT_EQ] = ACTIONS(3459),
    [anon_sym_GT_EQ] = ACTIONS(3457),
    [anon_sym_lt] = ACTIONS(3457),
    [anon_sym_le] = ACTIONS(3457),
    [anon_sym_ge] = ACTIONS(3457),
    [anon_sym_gt] = ACTIONS(3457),
    [anon_sym_isa] = ACTIONS(3457),
    [anon_sym_PLUS_PLUS] = ACTIONS(3457),
    [anon_sym_DASH_DASH] = ACTIONS(3457),
    [anon_sym_QMARK] = ACTIONS(3457),
    [anon_sym_EQ_GT] = ACTIONS(3457),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3457),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3347)] = {
    [sym_heredoc_content] = STATE(3347),
    [anon_sym_SEMI] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_LPAREN] = ACTIONS(5117),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(3695),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(3695),
    [anon_sym_DASH_DASH] = ACTIONS(3695),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3695),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3348)] = {
    [sym_heredoc_content] = STATE(3348),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_COLON] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3349)] = {
    [sym_heredoc_content] = STATE(3349),
    [anon_sym_RBRACE] = ACTIONS(3627),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_RPAREN] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3350)] = {
    [sym_heredoc_content] = STATE(3350),
    [anon_sym_SEMI] = ACTIONS(3601),
    [anon_sym_EQ] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_COMMA] = ACTIONS(3601),
    [anon_sym_LPAREN] = ACTIONS(3601),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3601),
    [anon_sym_DASH_GT] = ACTIONS(3601),
    [anon_sym_STAR] = ACTIONS(3603),
    [anon_sym_PERCENT] = ACTIONS(3603),
    [anon_sym_GT] = ACTIONS(3603),
    [anon_sym_LT_LT] = ACTIONS(3603),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3601),
    [anon_sym_PLUS_EQ] = ACTIONS(3601),
    [anon_sym_DASH_EQ] = ACTIONS(3601),
    [anon_sym_DOT_EQ] = ACTIONS(3601),
    [anon_sym_STAR_EQ] = ACTIONS(3601),
    [anon_sym_SLASH_EQ] = ACTIONS(3601),
    [anon_sym_PERCENT_EQ] = ACTIONS(3601),
    [anon_sym_x_EQ] = ACTIONS(3601),
    [anon_sym_AMP_EQ] = ACTIONS(3601),
    [anon_sym_PIPE_EQ] = ACTIONS(3601),
    [anon_sym_CARET_EQ] = ACTIONS(3601),
    [anon_sym_LT_LT_EQ] = ACTIONS(3601),
    [anon_sym_GT_GT_EQ] = ACTIONS(3601),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3601),
    [anon_sym_DOT_DOT] = ACTIONS(3603),
    [anon_sym_STAR_STAR] = ACTIONS(3603),
    [anon_sym_PIPE_PIPE] = ACTIONS(3603),
    [anon_sym_SLASH_SLASH] = ACTIONS(3603),
    [anon_sym_CARET_CARET] = ACTIONS(3601),
    [anon_sym_AMP_AMP] = ACTIONS(3603),
    [anon_sym_PIPE] = ACTIONS(3603),
    [anon_sym_CARET] = ACTIONS(3603),
    [anon_sym_AMP] = ACTIONS(3603),
    [anon_sym_GT_GT2] = ACTIONS(3603),
    [anon_sym_PLUS] = ACTIONS(3603),
    [anon_sym_DASH] = ACTIONS(3603),
    [anon_sym_DOT] = ACTIONS(3603),
    [anon_sym_SLASH] = ACTIONS(3603),
    [anon_sym_x] = ACTIONS(3603),
    [anon_sym_EQ_TILDE] = ACTIONS(3601),
    [anon_sym_BANG_TILDE] = ACTIONS(3601),
    [anon_sym_EQ_EQ] = ACTIONS(3603),
    [anon_sym_BANG_EQ] = ACTIONS(3601),
    [anon_sym_eq] = ACTIONS(3603),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3601),
    [anon_sym_equ] = ACTIONS(3601),
    [anon_sym_eqr] = ACTIONS(3601),
    [anon_sym_ne] = ACTIONS(3601),
    [anon_sym_LT_EQ_GT] = ACTIONS(3601),
    [anon_sym_cmp] = ACTIONS(3601),
    [anon_sym_TILDE_TILDE] = ACTIONS(3601),
    [anon_sym_LT] = ACTIONS(3603),
    [anon_sym_LT_EQ] = ACTIONS(3603),
    [anon_sym_GT_EQ] = ACTIONS(3601),
    [anon_sym_lt] = ACTIONS(3601),
    [anon_sym_le] = ACTIONS(3601),
    [anon_sym_ge] = ACTIONS(3601),
    [anon_sym_gt] = ACTIONS(3601),
    [anon_sym_isa] = ACTIONS(3601),
    [anon_sym_PLUS_PLUS] = ACTIONS(3601),
    [anon_sym_DASH_DASH] = ACTIONS(3601),
    [anon_sym_QMARK] = ACTIONS(3601),
    [anon_sym_EQ_GT] = ACTIONS(3601),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3601),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3351)] = {
    [sym_heredoc_content] = STATE(3351),
    [anon_sym_LBRACE] = ACTIONS(3605),
    [anon_sym_SEMI] = ACTIONS(3605),
    [anon_sym_EQ] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_COMMA] = ACTIONS(3605),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3605),
    [anon_sym_DASH_GT] = ACTIONS(3605),
    [anon_sym_STAR] = ACTIONS(3607),
    [anon_sym_PERCENT] = ACTIONS(3607),
    [anon_sym_GT] = ACTIONS(3607),
    [anon_sym_LT_LT] = ACTIONS(3607),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3605),
    [anon_sym_PLUS_EQ] = ACTIONS(3605),
    [anon_sym_DASH_EQ] = ACTIONS(3605),
    [anon_sym_DOT_EQ] = ACTIONS(3605),
    [anon_sym_STAR_EQ] = ACTIONS(3605),
    [anon_sym_SLASH_EQ] = ACTIONS(3605),
    [anon_sym_PERCENT_EQ] = ACTIONS(3605),
    [anon_sym_x_EQ] = ACTIONS(3605),
    [anon_sym_AMP_EQ] = ACTIONS(3605),
    [anon_sym_PIPE_EQ] = ACTIONS(3605),
    [anon_sym_CARET_EQ] = ACTIONS(3605),
    [anon_sym_LT_LT_EQ] = ACTIONS(3605),
    [anon_sym_GT_GT_EQ] = ACTIONS(3605),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3605),
    [anon_sym_DOT_DOT] = ACTIONS(3607),
    [anon_sym_STAR_STAR] = ACTIONS(3607),
    [anon_sym_PIPE_PIPE] = ACTIONS(3607),
    [anon_sym_SLASH_SLASH] = ACTIONS(3607),
    [anon_sym_CARET_CARET] = ACTIONS(3605),
    [anon_sym_AMP_AMP] = ACTIONS(3607),
    [anon_sym_PIPE] = ACTIONS(3607),
    [anon_sym_CARET] = ACTIONS(3607),
    [anon_sym_AMP] = ACTIONS(3607),
    [anon_sym_GT_GT2] = ACTIONS(3607),
    [anon_sym_PLUS] = ACTIONS(3607),
    [anon_sym_DASH] = ACTIONS(3607),
    [anon_sym_DOT] = ACTIONS(3607),
    [anon_sym_SLASH] = ACTIONS(3607),
    [anon_sym_x] = ACTIONS(3607),
    [anon_sym_EQ_TILDE] = ACTIONS(3605),
    [anon_sym_BANG_TILDE] = ACTIONS(3605),
    [anon_sym_EQ_EQ] = ACTIONS(3607),
    [anon_sym_BANG_EQ] = ACTIONS(3605),
    [anon_sym_eq] = ACTIONS(3607),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3605),
    [anon_sym_equ] = ACTIONS(3605),
    [anon_sym_eqr] = ACTIONS(3605),
    [anon_sym_ne] = ACTIONS(3605),
    [anon_sym_LT_EQ_GT] = ACTIONS(3605),
    [anon_sym_cmp] = ACTIONS(3605),
    [anon_sym_TILDE_TILDE] = ACTIONS(3605),
    [anon_sym_LT] = ACTIONS(3607),
    [anon_sym_LT_EQ] = ACTIONS(3607),
    [anon_sym_GT_EQ] = ACTIONS(3605),
    [anon_sym_lt] = ACTIONS(3605),
    [anon_sym_le] = ACTIONS(3605),
    [anon_sym_ge] = ACTIONS(3605),
    [anon_sym_gt] = ACTIONS(3605),
    [anon_sym_isa] = ACTIONS(3605),
    [anon_sym_PLUS_PLUS] = ACTIONS(3605),
    [anon_sym_DASH_DASH] = ACTIONS(3605),
    [anon_sym_QMARK] = ACTIONS(3605),
    [anon_sym_EQ_GT] = ACTIONS(3605),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3605),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3352)] = {
    [sym_heredoc_content] = STATE(3352),
    [anon_sym_SEMI] = ACTIONS(3655),
    [anon_sym_EQ] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_COMMA] = ACTIONS(3655),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3655),
    [anon_sym_DASH_GT] = ACTIONS(3655),
    [anon_sym_STAR] = ACTIONS(3657),
    [anon_sym_PERCENT] = ACTIONS(3657),
    [anon_sym_GT] = ACTIONS(3657),
    [anon_sym_LT_LT] = ACTIONS(3657),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3655),
    [anon_sym_PLUS_EQ] = ACTIONS(3655),
    [anon_sym_DASH_EQ] = ACTIONS(3655),
    [anon_sym_DOT_EQ] = ACTIONS(3655),
    [anon_sym_STAR_EQ] = ACTIONS(3655),
    [anon_sym_SLASH_EQ] = ACTIONS(3655),
    [anon_sym_PERCENT_EQ] = ACTIONS(3655),
    [anon_sym_x_EQ] = ACTIONS(3657),
    [anon_sym_AMP_EQ] = ACTIONS(3655),
    [anon_sym_PIPE_EQ] = ACTIONS(3655),
    [anon_sym_CARET_EQ] = ACTIONS(3655),
    [anon_sym_LT_LT_EQ] = ACTIONS(3655),
    [anon_sym_GT_GT_EQ] = ACTIONS(3655),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3655),
    [anon_sym_DOT_DOT] = ACTIONS(3657),
    [anon_sym_STAR_STAR] = ACTIONS(3657),
    [anon_sym_PIPE_PIPE] = ACTIONS(3657),
    [anon_sym_SLASH_SLASH] = ACTIONS(3657),
    [anon_sym_CARET_CARET] = ACTIONS(3655),
    [anon_sym_AMP_AMP] = ACTIONS(3657),
    [anon_sym_PIPE] = ACTIONS(3657),
    [anon_sym_CARET] = ACTIONS(3657),
    [anon_sym_AMP] = ACTIONS(3657),
    [anon_sym_GT_GT2] = ACTIONS(3657),
    [anon_sym_PLUS] = ACTIONS(3657),
    [anon_sym_DASH] = ACTIONS(3657),
    [anon_sym_DOT] = ACTIONS(3657),
    [anon_sym_SLASH] = ACTIONS(3657),
    [anon_sym_x] = ACTIONS(3657),
    [anon_sym_EQ_TILDE] = ACTIONS(3655),
    [anon_sym_BANG_TILDE] = ACTIONS(3655),
    [anon_sym_EQ_EQ] = ACTIONS(3657),
    [anon_sym_BANG_EQ] = ACTIONS(3655),
    [anon_sym_eq] = ACTIONS(3657),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3655),
    [anon_sym_equ] = ACTIONS(3655),
    [anon_sym_eqr] = ACTIONS(3655),
    [anon_sym_ne] = ACTIONS(3657),
    [anon_sym_LT_EQ_GT] = ACTIONS(3655),
    [anon_sym_cmp] = ACTIONS(3657),
    [anon_sym_TILDE_TILDE] = ACTIONS(3655),
    [anon_sym_LT] = ACTIONS(3657),
    [anon_sym_LT_EQ] = ACTIONS(3657),
    [anon_sym_GT_EQ] = ACTIONS(3655),
    [anon_sym_lt] = ACTIONS(3657),
    [anon_sym_le] = ACTIONS(3657),
    [anon_sym_ge] = ACTIONS(3657),
    [anon_sym_gt] = ACTIONS(3657),
    [anon_sym_isa] = ACTIONS(3657),
    [anon_sym_PLUS_PLUS] = ACTIONS(3655),
    [anon_sym_DASH_DASH] = ACTIONS(3655),
    [anon_sym_QMARK] = ACTIONS(3655),
    [anon_sym_EQ_GT] = ACTIONS(3655),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5127),
    [sym__PERLY_SEMICOLON] = ACTIONS(3655),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3353)] = {
    [sym_heredoc_content] = STATE(3353),
    [anon_sym_SEMI] = ACTIONS(3649),
    [anon_sym_EQ] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_COMMA] = ACTIONS(3649),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3649),
    [anon_sym_DASH_GT] = ACTIONS(3649),
    [anon_sym_STAR] = ACTIONS(3651),
    [anon_sym_LBRACK] = ACTIONS(5125),
    [anon_sym_PERCENT] = ACTIONS(3651),
    [anon_sym_GT] = ACTIONS(3651),
    [anon_sym_LT_LT] = ACTIONS(3651),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3649),
    [anon_sym_PLUS_EQ] = ACTIONS(3649),
    [anon_sym_DASH_EQ] = ACTIONS(3649),
    [anon_sym_DOT_EQ] = ACTIONS(3649),
    [anon_sym_STAR_EQ] = ACTIONS(3649),
    [anon_sym_SLASH_EQ] = ACTIONS(3649),
    [anon_sym_PERCENT_EQ] = ACTIONS(3649),
    [anon_sym_x_EQ] = ACTIONS(3649),
    [anon_sym_AMP_EQ] = ACTIONS(3649),
    [anon_sym_PIPE_EQ] = ACTIONS(3649),
    [anon_sym_CARET_EQ] = ACTIONS(3649),
    [anon_sym_LT_LT_EQ] = ACTIONS(3649),
    [anon_sym_GT_GT_EQ] = ACTIONS(3649),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3649),
    [anon_sym_DOT_DOT] = ACTIONS(3651),
    [anon_sym_STAR_STAR] = ACTIONS(3651),
    [anon_sym_PIPE_PIPE] = ACTIONS(3651),
    [anon_sym_SLASH_SLASH] = ACTIONS(3651),
    [anon_sym_CARET_CARET] = ACTIONS(3649),
    [anon_sym_AMP_AMP] = ACTIONS(3651),
    [anon_sym_PIPE] = ACTIONS(3651),
    [anon_sym_CARET] = ACTIONS(3651),
    [anon_sym_AMP] = ACTIONS(3651),
    [anon_sym_GT_GT2] = ACTIONS(3651),
    [anon_sym_PLUS] = ACTIONS(3651),
    [anon_sym_DASH] = ACTIONS(3651),
    [anon_sym_DOT] = ACTIONS(3651),
    [anon_sym_SLASH] = ACTIONS(3651),
    [anon_sym_x] = ACTIONS(3651),
    [anon_sym_EQ_TILDE] = ACTIONS(3649),
    [anon_sym_BANG_TILDE] = ACTIONS(3649),
    [anon_sym_EQ_EQ] = ACTIONS(3651),
    [anon_sym_BANG_EQ] = ACTIONS(3649),
    [anon_sym_eq] = ACTIONS(3651),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3649),
    [anon_sym_equ] = ACTIONS(3649),
    [anon_sym_eqr] = ACTIONS(3649),
    [anon_sym_ne] = ACTIONS(3649),
    [anon_sym_LT_EQ_GT] = ACTIONS(3649),
    [anon_sym_cmp] = ACTIONS(3649),
    [anon_sym_TILDE_TILDE] = ACTIONS(3649),
    [anon_sym_LT] = ACTIONS(3651),
    [anon_sym_LT_EQ] = ACTIONS(3651),
    [anon_sym_GT_EQ] = ACTIONS(3649),
    [anon_sym_lt] = ACTIONS(3649),
    [anon_sym_le] = ACTIONS(3649),
    [anon_sym_ge] = ACTIONS(3649),
    [anon_sym_gt] = ACTIONS(3649),
    [anon_sym_isa] = ACTIONS(3649),
    [anon_sym_PLUS_PLUS] = ACTIONS(3649),
    [anon_sym_DASH_DASH] = ACTIONS(3649),
    [anon_sym_QMARK] = ACTIONS(3649),
    [anon_sym_EQ_GT] = ACTIONS(3649),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3649),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3354)] = {
    [sym_heredoc_content] = STATE(3354),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_SEMI] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3679),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3355)] = {
    [sym_heredoc_content] = STATE(3355),
    [anon_sym_SEMI] = ACTIONS(3451),
    [anon_sym_EQ] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_COMMA] = ACTIONS(3451),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3451),
    [anon_sym_DASH_GT] = ACTIONS(3451),
    [anon_sym_STAR] = ACTIONS(3453),
    [anon_sym_LBRACK] = ACTIONS(5129),
    [anon_sym_PERCENT] = ACTIONS(3453),
    [anon_sym_GT] = ACTIONS(3453),
    [anon_sym_LT_LT] = ACTIONS(3453),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3451),
    [anon_sym_PLUS_EQ] = ACTIONS(3451),
    [anon_sym_DASH_EQ] = ACTIONS(3451),
    [anon_sym_DOT_EQ] = ACTIONS(3451),
    [anon_sym_STAR_EQ] = ACTIONS(3451),
    [anon_sym_SLASH_EQ] = ACTIONS(3451),
    [anon_sym_PERCENT_EQ] = ACTIONS(3451),
    [anon_sym_x_EQ] = ACTIONS(3451),
    [anon_sym_AMP_EQ] = ACTIONS(3451),
    [anon_sym_PIPE_EQ] = ACTIONS(3451),
    [anon_sym_CARET_EQ] = ACTIONS(3451),
    [anon_sym_LT_LT_EQ] = ACTIONS(3451),
    [anon_sym_GT_GT_EQ] = ACTIONS(3451),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3451),
    [anon_sym_DOT_DOT] = ACTIONS(3453),
    [anon_sym_STAR_STAR] = ACTIONS(3453),
    [anon_sym_PIPE_PIPE] = ACTIONS(3453),
    [anon_sym_SLASH_SLASH] = ACTIONS(3453),
    [anon_sym_CARET_CARET] = ACTIONS(3451),
    [anon_sym_AMP_AMP] = ACTIONS(3453),
    [anon_sym_PIPE] = ACTIONS(3453),
    [anon_sym_CARET] = ACTIONS(3453),
    [anon_sym_AMP] = ACTIONS(3453),
    [anon_sym_GT_GT2] = ACTIONS(3453),
    [anon_sym_PLUS] = ACTIONS(3453),
    [anon_sym_DASH] = ACTIONS(3453),
    [anon_sym_DOT] = ACTIONS(3453),
    [anon_sym_SLASH] = ACTIONS(3453),
    [anon_sym_x] = ACTIONS(3453),
    [anon_sym_EQ_TILDE] = ACTIONS(3451),
    [anon_sym_BANG_TILDE] = ACTIONS(3451),
    [anon_sym_EQ_EQ] = ACTIONS(3453),
    [anon_sym_BANG_EQ] = ACTIONS(3451),
    [anon_sym_eq] = ACTIONS(3453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3451),
    [anon_sym_equ] = ACTIONS(3451),
    [anon_sym_eqr] = ACTIONS(3451),
    [anon_sym_ne] = ACTIONS(3451),
    [anon_sym_LT_EQ_GT] = ACTIONS(3451),
    [anon_sym_cmp] = ACTIONS(3451),
    [anon_sym_TILDE_TILDE] = ACTIONS(3451),
    [anon_sym_LT] = ACTIONS(3453),
    [anon_sym_LT_EQ] = ACTIONS(3453),
    [anon_sym_GT_EQ] = ACTIONS(3451),
    [anon_sym_lt] = ACTIONS(3451),
    [anon_sym_le] = ACTIONS(3451),
    [anon_sym_ge] = ACTIONS(3451),
    [anon_sym_gt] = ACTIONS(3451),
    [anon_sym_isa] = ACTIONS(3451),
    [anon_sym_PLUS_PLUS] = ACTIONS(3451),
    [anon_sym_DASH_DASH] = ACTIONS(3451),
    [anon_sym_QMARK] = ACTIONS(3451),
    [anon_sym_EQ_GT] = ACTIONS(3451),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3451),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3356)] = {
    [sym_heredoc_content] = STATE(3356),
    [anon_sym_SEMI] = ACTIONS(3639),
    [anon_sym_EQ] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_COMMA] = ACTIONS(3639),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3639),
    [anon_sym_DASH_GT] = ACTIONS(3639),
    [anon_sym_STAR] = ACTIONS(3641),
    [anon_sym_LBRACK] = ACTIONS(3639),
    [anon_sym_PERCENT] = ACTIONS(3641),
    [anon_sym_GT] = ACTIONS(3641),
    [anon_sym_LT_LT] = ACTIONS(3641),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3639),
    [anon_sym_PLUS_EQ] = ACTIONS(3639),
    [anon_sym_DASH_EQ] = ACTIONS(3639),
    [anon_sym_DOT_EQ] = ACTIONS(3639),
    [anon_sym_STAR_EQ] = ACTIONS(3639),
    [anon_sym_SLASH_EQ] = ACTIONS(3639),
    [anon_sym_PERCENT_EQ] = ACTIONS(3639),
    [anon_sym_x_EQ] = ACTIONS(3639),
    [anon_sym_AMP_EQ] = ACTIONS(3639),
    [anon_sym_PIPE_EQ] = ACTIONS(3639),
    [anon_sym_CARET_EQ] = ACTIONS(3639),
    [anon_sym_LT_LT_EQ] = ACTIONS(3639),
    [anon_sym_GT_GT_EQ] = ACTIONS(3639),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3639),
    [anon_sym_DOT_DOT] = ACTIONS(3641),
    [anon_sym_STAR_STAR] = ACTIONS(3641),
    [anon_sym_PIPE_PIPE] = ACTIONS(3641),
    [anon_sym_SLASH_SLASH] = ACTIONS(3641),
    [anon_sym_CARET_CARET] = ACTIONS(3639),
    [anon_sym_AMP_AMP] = ACTIONS(3641),
    [anon_sym_PIPE] = ACTIONS(3641),
    [anon_sym_CARET] = ACTIONS(3641),
    [anon_sym_AMP] = ACTIONS(3641),
    [anon_sym_GT_GT2] = ACTIONS(3641),
    [anon_sym_PLUS] = ACTIONS(3641),
    [anon_sym_DASH] = ACTIONS(3641),
    [anon_sym_DOT] = ACTIONS(3641),
    [anon_sym_SLASH] = ACTIONS(3641),
    [anon_sym_x] = ACTIONS(3641),
    [anon_sym_EQ_TILDE] = ACTIONS(3639),
    [anon_sym_BANG_TILDE] = ACTIONS(3639),
    [anon_sym_EQ_EQ] = ACTIONS(3641),
    [anon_sym_BANG_EQ] = ACTIONS(3639),
    [anon_sym_eq] = ACTIONS(3641),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3639),
    [anon_sym_equ] = ACTIONS(3639),
    [anon_sym_eqr] = ACTIONS(3639),
    [anon_sym_ne] = ACTIONS(3639),
    [anon_sym_LT_EQ_GT] = ACTIONS(3639),
    [anon_sym_cmp] = ACTIONS(3639),
    [anon_sym_TILDE_TILDE] = ACTIONS(3639),
    [anon_sym_LT] = ACTIONS(3641),
    [anon_sym_LT_EQ] = ACTIONS(3641),
    [anon_sym_GT_EQ] = ACTIONS(3639),
    [anon_sym_lt] = ACTIONS(3639),
    [anon_sym_le] = ACTIONS(3639),
    [anon_sym_ge] = ACTIONS(3639),
    [anon_sym_gt] = ACTIONS(3639),
    [anon_sym_isa] = ACTIONS(3639),
    [anon_sym_PLUS_PLUS] = ACTIONS(3639),
    [anon_sym_DASH_DASH] = ACTIONS(3639),
    [anon_sym_QMARK] = ACTIONS(3639),
    [anon_sym_EQ_GT] = ACTIONS(3639),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3639),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3357)] = {
    [sym_heredoc_content] = STATE(3357),
    [anon_sym_SEMI] = ACTIONS(3643),
    [anon_sym_EQ] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_COMMA] = ACTIONS(3643),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3643),
    [anon_sym_DASH_GT] = ACTIONS(3643),
    [anon_sym_STAR] = ACTIONS(3645),
    [anon_sym_PERCENT] = ACTIONS(3645),
    [anon_sym_GT] = ACTIONS(3645),
    [anon_sym_LT_LT] = ACTIONS(3645),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3643),
    [anon_sym_PLUS_EQ] = ACTIONS(3643),
    [anon_sym_DASH_EQ] = ACTIONS(3643),
    [anon_sym_DOT_EQ] = ACTIONS(3643),
    [anon_sym_STAR_EQ] = ACTIONS(3643),
    [anon_sym_SLASH_EQ] = ACTIONS(3643),
    [anon_sym_PERCENT_EQ] = ACTIONS(3643),
    [anon_sym_x_EQ] = ACTIONS(3645),
    [anon_sym_AMP_EQ] = ACTIONS(3643),
    [anon_sym_PIPE_EQ] = ACTIONS(3643),
    [anon_sym_CARET_EQ] = ACTIONS(3643),
    [anon_sym_LT_LT_EQ] = ACTIONS(3643),
    [anon_sym_GT_GT_EQ] = ACTIONS(3643),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3643),
    [anon_sym_DOT_DOT] = ACTIONS(3645),
    [anon_sym_STAR_STAR] = ACTIONS(3645),
    [anon_sym_PIPE_PIPE] = ACTIONS(3645),
    [anon_sym_SLASH_SLASH] = ACTIONS(3645),
    [anon_sym_CARET_CARET] = ACTIONS(3643),
    [anon_sym_AMP_AMP] = ACTIONS(3645),
    [anon_sym_PIPE] = ACTIONS(3645),
    [anon_sym_CARET] = ACTIONS(3645),
    [anon_sym_AMP] = ACTIONS(3645),
    [anon_sym_GT_GT2] = ACTIONS(3645),
    [anon_sym_PLUS] = ACTIONS(3645),
    [anon_sym_DASH] = ACTIONS(3645),
    [anon_sym_DOT] = ACTIONS(3645),
    [anon_sym_SLASH] = ACTIONS(3645),
    [anon_sym_x] = ACTIONS(3645),
    [anon_sym_EQ_TILDE] = ACTIONS(3643),
    [anon_sym_BANG_TILDE] = ACTIONS(3643),
    [anon_sym_EQ_EQ] = ACTIONS(3645),
    [anon_sym_BANG_EQ] = ACTIONS(3643),
    [anon_sym_eq] = ACTIONS(3645),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3643),
    [anon_sym_equ] = ACTIONS(3643),
    [anon_sym_eqr] = ACTIONS(3643),
    [anon_sym_ne] = ACTIONS(3645),
    [anon_sym_LT_EQ_GT] = ACTIONS(3643),
    [anon_sym_cmp] = ACTIONS(3643),
    [anon_sym_TILDE_TILDE] = ACTIONS(3643),
    [anon_sym_LT] = ACTIONS(3645),
    [anon_sym_LT_EQ] = ACTIONS(3645),
    [anon_sym_GT_EQ] = ACTIONS(3643),
    [anon_sym_lt] = ACTIONS(3645),
    [anon_sym_le] = ACTIONS(3645),
    [anon_sym_ge] = ACTIONS(3643),
    [anon_sym_gt] = ACTIONS(3643),
    [anon_sym_isa] = ACTIONS(3645),
    [anon_sym_PLUS_PLUS] = ACTIONS(3643),
    [anon_sym_DASH_DASH] = ACTIONS(3643),
    [anon_sym_QMARK] = ACTIONS(3643),
    [anon_sym_EQ_GT] = ACTIONS(3643),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(5131),
    [sym__PERLY_SEMICOLON] = ACTIONS(3643),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3358)] = {
    [sym_heredoc_content] = STATE(3358),
    [anon_sym_SEMI] = ACTIONS(3463),
    [anon_sym_EQ] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_COMMA] = ACTIONS(3463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3463),
    [anon_sym_DASH_GT] = ACTIONS(3463),
    [anon_sym_STAR] = ACTIONS(3465),
    [anon_sym_PERCENT] = ACTIONS(3465),
    [anon_sym_GT] = ACTIONS(3465),
    [anon_sym_LT_LT] = ACTIONS(3465),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3463),
    [anon_sym_PLUS_EQ] = ACTIONS(3463),
    [anon_sym_DASH_EQ] = ACTIONS(3463),
    [anon_sym_DOT_EQ] = ACTIONS(3463),
    [anon_sym_STAR_EQ] = ACTIONS(3463),
    [anon_sym_SLASH_EQ] = ACTIONS(3463),
    [anon_sym_PERCENT_EQ] = ACTIONS(3463),
    [anon_sym_x_EQ] = ACTIONS(3465),
    [anon_sym_AMP_EQ] = ACTIONS(3463),
    [anon_sym_PIPE_EQ] = ACTIONS(3463),
    [anon_sym_CARET_EQ] = ACTIONS(3463),
    [anon_sym_LT_LT_EQ] = ACTIONS(3463),
    [anon_sym_GT_GT_EQ] = ACTIONS(3463),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3463),
    [anon_sym_DOT_DOT] = ACTIONS(3465),
    [anon_sym_STAR_STAR] = ACTIONS(3465),
    [anon_sym_PIPE_PIPE] = ACTIONS(3465),
    [anon_sym_SLASH_SLASH] = ACTIONS(3465),
    [anon_sym_CARET_CARET] = ACTIONS(3463),
    [anon_sym_AMP_AMP] = ACTIONS(3465),
    [anon_sym_PIPE] = ACTIONS(3465),
    [anon_sym_CARET] = ACTIONS(3465),
    [anon_sym_AMP] = ACTIONS(3465),
    [anon_sym_GT_GT2] = ACTIONS(3465),
    [anon_sym_PLUS] = ACTIONS(3465),
    [anon_sym_DASH] = ACTIONS(3465),
    [anon_sym_DOT] = ACTIONS(3465),
    [anon_sym_SLASH] = ACTIONS(3465),
    [anon_sym_x] = ACTIONS(3465),
    [anon_sym_EQ_TILDE] = ACTIONS(3463),
    [anon_sym_BANG_TILDE] = ACTIONS(3463),
    [anon_sym_EQ_EQ] = ACTIONS(3465),
    [anon_sym_BANG_EQ] = ACTIONS(3463),
    [anon_sym_eq] = ACTIONS(3465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3463),
    [anon_sym_equ] = ACTIONS(3463),
    [anon_sym_eqr] = ACTIONS(3463),
    [anon_sym_ne] = ACTIONS(3465),
    [anon_sym_LT_EQ_GT] = ACTIONS(3463),
    [anon_sym_cmp] = ACTIONS(3465),
    [anon_sym_TILDE_TILDE] = ACTIONS(3463),
    [anon_sym_LT] = ACTIONS(3465),
    [anon_sym_LT_EQ] = ACTIONS(3465),
    [anon_sym_GT_EQ] = ACTIONS(3463),
    [anon_sym_lt] = ACTIONS(3465),
    [anon_sym_le] = ACTIONS(3465),
    [anon_sym_ge] = ACTIONS(3465),
    [anon_sym_gt] = ACTIONS(3465),
    [anon_sym_isa] = ACTIONS(3465),
    [anon_sym_PLUS_PLUS] = ACTIONS(3463),
    [anon_sym_DASH_DASH] = ACTIONS(3463),
    [anon_sym_QMARK] = ACTIONS(3463),
    [anon_sym_EQ_GT] = ACTIONS(3463),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5133),
    [sym__PERLY_SEMICOLON] = ACTIONS(3463),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3359)] = {
    [sym_heredoc_content] = STATE(3359),
    [anon_sym_RBRACE] = ACTIONS(4200),
    [anon_sym_COLON] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_RPAREN] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3360)] = {
    [sym_heredoc_content] = STATE(3360),
    [anon_sym_SEMI] = ACTIONS(3475),
    [anon_sym_EQ] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_COMMA] = ACTIONS(3475),
    [anon_sym_LPAREN] = ACTIONS(3475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3475),
    [anon_sym_DASH_GT] = ACTIONS(3475),
    [anon_sym_STAR] = ACTIONS(3477),
    [anon_sym_PERCENT] = ACTIONS(3477),
    [anon_sym_GT] = ACTIONS(3477),
    [anon_sym_LT_LT] = ACTIONS(3477),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3475),
    [anon_sym_PLUS_EQ] = ACTIONS(3475),
    [anon_sym_DASH_EQ] = ACTIONS(3475),
    [anon_sym_DOT_EQ] = ACTIONS(3475),
    [anon_sym_STAR_EQ] = ACTIONS(3475),
    [anon_sym_SLASH_EQ] = ACTIONS(3475),
    [anon_sym_PERCENT_EQ] = ACTIONS(3475),
    [anon_sym_x_EQ] = ACTIONS(3475),
    [anon_sym_AMP_EQ] = ACTIONS(3475),
    [anon_sym_PIPE_EQ] = ACTIONS(3475),
    [anon_sym_CARET_EQ] = ACTIONS(3475),
    [anon_sym_LT_LT_EQ] = ACTIONS(3475),
    [anon_sym_GT_GT_EQ] = ACTIONS(3475),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3475),
    [anon_sym_DOT_DOT] = ACTIONS(3477),
    [anon_sym_STAR_STAR] = ACTIONS(3477),
    [anon_sym_PIPE_PIPE] = ACTIONS(3477),
    [anon_sym_SLASH_SLASH] = ACTIONS(3477),
    [anon_sym_CARET_CARET] = ACTIONS(3475),
    [anon_sym_AMP_AMP] = ACTIONS(3477),
    [anon_sym_PIPE] = ACTIONS(3477),
    [anon_sym_CARET] = ACTIONS(3477),
    [anon_sym_AMP] = ACTIONS(3477),
    [anon_sym_GT_GT2] = ACTIONS(3477),
    [anon_sym_PLUS] = ACTIONS(3477),
    [anon_sym_DASH] = ACTIONS(3477),
    [anon_sym_DOT] = ACTIONS(3477),
    [anon_sym_SLASH] = ACTIONS(3477),
    [anon_sym_x] = ACTIONS(3477),
    [anon_sym_EQ_TILDE] = ACTIONS(3475),
    [anon_sym_BANG_TILDE] = ACTIONS(3475),
    [anon_sym_EQ_EQ] = ACTIONS(3477),
    [anon_sym_BANG_EQ] = ACTIONS(3475),
    [anon_sym_eq] = ACTIONS(3477),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3475),
    [anon_sym_equ] = ACTIONS(3475),
    [anon_sym_eqr] = ACTIONS(3475),
    [anon_sym_ne] = ACTIONS(3475),
    [anon_sym_LT_EQ_GT] = ACTIONS(3475),
    [anon_sym_cmp] = ACTIONS(3475),
    [anon_sym_TILDE_TILDE] = ACTIONS(3475),
    [anon_sym_LT] = ACTIONS(3477),
    [anon_sym_LT_EQ] = ACTIONS(3477),
    [anon_sym_GT_EQ] = ACTIONS(3475),
    [anon_sym_lt] = ACTIONS(3475),
    [anon_sym_le] = ACTIONS(3475),
    [anon_sym_ge] = ACTIONS(3475),
    [anon_sym_gt] = ACTIONS(3475),
    [anon_sym_isa] = ACTIONS(3475),
    [anon_sym_PLUS_PLUS] = ACTIONS(3475),
    [anon_sym_DASH_DASH] = ACTIONS(3475),
    [anon_sym_QMARK] = ACTIONS(3475),
    [anon_sym_EQ_GT] = ACTIONS(3475),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3475),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3361)] = {
    [sym_heredoc_content] = STATE(3361),
    [anon_sym_SEMI] = ACTIONS(3479),
    [anon_sym_EQ] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_COMMA] = ACTIONS(3479),
    [anon_sym_LPAREN] = ACTIONS(5135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3479),
    [anon_sym_DASH_GT] = ACTIONS(3479),
    [anon_sym_STAR] = ACTIONS(3481),
    [anon_sym_PERCENT] = ACTIONS(3481),
    [anon_sym_GT] = ACTIONS(3481),
    [anon_sym_LT_LT] = ACTIONS(3481),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3479),
    [anon_sym_PLUS_EQ] = ACTIONS(3479),
    [anon_sym_DASH_EQ] = ACTIONS(3479),
    [anon_sym_DOT_EQ] = ACTIONS(3479),
    [anon_sym_STAR_EQ] = ACTIONS(3479),
    [anon_sym_SLASH_EQ] = ACTIONS(3479),
    [anon_sym_PERCENT_EQ] = ACTIONS(3479),
    [anon_sym_x_EQ] = ACTIONS(3479),
    [anon_sym_AMP_EQ] = ACTIONS(3479),
    [anon_sym_PIPE_EQ] = ACTIONS(3479),
    [anon_sym_CARET_EQ] = ACTIONS(3479),
    [anon_sym_LT_LT_EQ] = ACTIONS(3479),
    [anon_sym_GT_GT_EQ] = ACTIONS(3479),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3479),
    [anon_sym_DOT_DOT] = ACTIONS(3481),
    [anon_sym_STAR_STAR] = ACTIONS(3481),
    [anon_sym_PIPE_PIPE] = ACTIONS(3481),
    [anon_sym_SLASH_SLASH] = ACTIONS(3481),
    [anon_sym_CARET_CARET] = ACTIONS(3479),
    [anon_sym_AMP_AMP] = ACTIONS(3481),
    [anon_sym_PIPE] = ACTIONS(3481),
    [anon_sym_CARET] = ACTIONS(3481),
    [anon_sym_AMP] = ACTIONS(3481),
    [anon_sym_GT_GT2] = ACTIONS(3481),
    [anon_sym_PLUS] = ACTIONS(3481),
    [anon_sym_DASH] = ACTIONS(3481),
    [anon_sym_DOT] = ACTIONS(3481),
    [anon_sym_SLASH] = ACTIONS(3481),
    [anon_sym_x] = ACTIONS(3481),
    [anon_sym_EQ_TILDE] = ACTIONS(3479),
    [anon_sym_BANG_TILDE] = ACTIONS(3479),
    [anon_sym_EQ_EQ] = ACTIONS(3481),
    [anon_sym_BANG_EQ] = ACTIONS(3479),
    [anon_sym_eq] = ACTIONS(3481),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3479),
    [anon_sym_equ] = ACTIONS(3479),
    [anon_sym_eqr] = ACTIONS(3479),
    [anon_sym_ne] = ACTIONS(3479),
    [anon_sym_LT_EQ_GT] = ACTIONS(3479),
    [anon_sym_cmp] = ACTIONS(3479),
    [anon_sym_TILDE_TILDE] = ACTIONS(3479),
    [anon_sym_LT] = ACTIONS(3481),
    [anon_sym_LT_EQ] = ACTIONS(3481),
    [anon_sym_GT_EQ] = ACTIONS(3479),
    [anon_sym_lt] = ACTIONS(3479),
    [anon_sym_le] = ACTIONS(3479),
    [anon_sym_ge] = ACTIONS(3479),
    [anon_sym_gt] = ACTIONS(3479),
    [anon_sym_isa] = ACTIONS(3479),
    [anon_sym_PLUS_PLUS] = ACTIONS(3479),
    [anon_sym_DASH_DASH] = ACTIONS(3479),
    [anon_sym_QMARK] = ACTIONS(3479),
    [anon_sym_EQ_GT] = ACTIONS(3479),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3479),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3362)] = {
    [sym_heredoc_content] = STATE(3362),
    [anon_sym_COLON] = ACTIONS(5093),
    [anon_sym_DOLLAR] = ACTIONS(5093),
    [anon_sym_EQ] = ACTIONS(5137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_AT] = ACTIONS(5093),
    [anon_sym_RPAREN] = ACTIONS(5093),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5139),
    [anon_sym_PLUS_EQ] = ACTIONS(5139),
    [anon_sym_DASH_EQ] = ACTIONS(5139),
    [anon_sym_DOT_EQ] = ACTIONS(5139),
    [anon_sym_STAR_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_PERCENT_EQ] = ACTIONS(5137),
    [anon_sym_x_EQ] = ACTIONS(5139),
    [anon_sym_AMP_EQ] = ACTIONS(5139),
    [anon_sym_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_CARET_EQ] = ACTIONS(5139),
    [anon_sym_LT_LT_EQ] = ACTIONS(5139),
    [anon_sym_GT_GT_EQ] = ACTIONS(5139),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5139),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(5093),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3363)] = {
    [sym_heredoc_content] = STATE(3363),
    [anon_sym_COLON] = ACTIONS(5045),
    [anon_sym_DOLLAR] = ACTIONS(5045),
    [anon_sym_EQ] = ACTIONS(5137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_AT] = ACTIONS(5045),
    [anon_sym_RPAREN] = ACTIONS(5045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5139),
    [anon_sym_PLUS_EQ] = ACTIONS(5139),
    [anon_sym_DASH_EQ] = ACTIONS(5139),
    [anon_sym_DOT_EQ] = ACTIONS(5139),
    [anon_sym_STAR_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_PERCENT_EQ] = ACTIONS(5137),
    [anon_sym_x_EQ] = ACTIONS(5139),
    [anon_sym_AMP_EQ] = ACTIONS(5139),
    [anon_sym_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_CARET_EQ] = ACTIONS(5139),
    [anon_sym_LT_LT_EQ] = ACTIONS(5139),
    [anon_sym_GT_GT_EQ] = ACTIONS(5139),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5139),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(5045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3364)] = {
    [sym_heredoc_content] = STATE(3364),
    [anon_sym_SEMI] = ACTIONS(3556),
    [anon_sym_EQ] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_COMMA] = ACTIONS(3556),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3556),
    [anon_sym_DASH_GT] = ACTIONS(3556),
    [anon_sym_STAR] = ACTIONS(3558),
    [anon_sym_PERCENT] = ACTIONS(3558),
    [anon_sym_GT] = ACTIONS(3558),
    [anon_sym_LT_LT] = ACTIONS(3558),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3556),
    [anon_sym_PLUS_EQ] = ACTIONS(3556),
    [anon_sym_DASH_EQ] = ACTIONS(3556),
    [anon_sym_DOT_EQ] = ACTIONS(3556),
    [anon_sym_STAR_EQ] = ACTIONS(3556),
    [anon_sym_SLASH_EQ] = ACTIONS(3556),
    [anon_sym_PERCENT_EQ] = ACTIONS(3556),
    [anon_sym_x_EQ] = ACTIONS(3558),
    [anon_sym_AMP_EQ] = ACTIONS(3556),
    [anon_sym_PIPE_EQ] = ACTIONS(3556),
    [anon_sym_CARET_EQ] = ACTIONS(3556),
    [anon_sym_LT_LT_EQ] = ACTIONS(3556),
    [anon_sym_GT_GT_EQ] = ACTIONS(3556),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3556),
    [anon_sym_DOT_DOT] = ACTIONS(3558),
    [anon_sym_STAR_STAR] = ACTIONS(3558),
    [anon_sym_PIPE_PIPE] = ACTIONS(3558),
    [anon_sym_SLASH_SLASH] = ACTIONS(3558),
    [anon_sym_CARET_CARET] = ACTIONS(3556),
    [anon_sym_AMP_AMP] = ACTIONS(3558),
    [anon_sym_PIPE] = ACTIONS(3558),
    [anon_sym_CARET] = ACTIONS(3558),
    [anon_sym_AMP] = ACTIONS(3558),
    [anon_sym_GT_GT2] = ACTIONS(3558),
    [anon_sym_PLUS] = ACTIONS(3558),
    [anon_sym_DASH] = ACTIONS(3558),
    [anon_sym_DOT] = ACTIONS(3558),
    [anon_sym_SLASH] = ACTIONS(3558),
    [anon_sym_x] = ACTIONS(3558),
    [anon_sym_EQ_TILDE] = ACTIONS(3556),
    [anon_sym_BANG_TILDE] = ACTIONS(3556),
    [anon_sym_EQ_EQ] = ACTIONS(3558),
    [anon_sym_BANG_EQ] = ACTIONS(3556),
    [anon_sym_eq] = ACTIONS(3558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3556),
    [anon_sym_equ] = ACTIONS(3556),
    [anon_sym_eqr] = ACTIONS(3556),
    [anon_sym_ne] = ACTIONS(3558),
    [anon_sym_LT_EQ_GT] = ACTIONS(3556),
    [anon_sym_cmp] = ACTIONS(3558),
    [anon_sym_TILDE_TILDE] = ACTIONS(3556),
    [anon_sym_LT] = ACTIONS(3558),
    [anon_sym_LT_EQ] = ACTIONS(3558),
    [anon_sym_GT_EQ] = ACTIONS(3556),
    [anon_sym_lt] = ACTIONS(3558),
    [anon_sym_le] = ACTIONS(3558),
    [anon_sym_ge] = ACTIONS(3558),
    [anon_sym_gt] = ACTIONS(3558),
    [anon_sym_isa] = ACTIONS(3558),
    [anon_sym_PLUS_PLUS] = ACTIONS(3556),
    [anon_sym_DASH_DASH] = ACTIONS(3556),
    [anon_sym_QMARK] = ACTIONS(3556),
    [anon_sym_EQ_GT] = ACTIONS(3556),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5179),
    [sym__PERLY_SEMICOLON] = ACTIONS(3556),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3365)] = {
    [sym_heredoc_content] = STATE(3365),
    [anon_sym_SEMI] = ACTIONS(3574),
    [anon_sym_EQ] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_COMMA] = ACTIONS(3574),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3574),
    [anon_sym_DASH_GT] = ACTIONS(3574),
    [anon_sym_STAR] = ACTIONS(3576),
    [anon_sym_LBRACK] = ACTIONS(5129),
    [anon_sym_PERCENT] = ACTIONS(3576),
    [anon_sym_GT] = ACTIONS(3576),
    [anon_sym_LT_LT] = ACTIONS(3576),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3574),
    [anon_sym_PLUS_EQ] = ACTIONS(3574),
    [anon_sym_DASH_EQ] = ACTIONS(3574),
    [anon_sym_DOT_EQ] = ACTIONS(3574),
    [anon_sym_STAR_EQ] = ACTIONS(3574),
    [anon_sym_SLASH_EQ] = ACTIONS(3574),
    [anon_sym_PERCENT_EQ] = ACTIONS(3574),
    [anon_sym_x_EQ] = ACTIONS(3574),
    [anon_sym_AMP_EQ] = ACTIONS(3574),
    [anon_sym_PIPE_EQ] = ACTIONS(3574),
    [anon_sym_CARET_EQ] = ACTIONS(3574),
    [anon_sym_LT_LT_EQ] = ACTIONS(3574),
    [anon_sym_GT_GT_EQ] = ACTIONS(3574),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3574),
    [anon_sym_DOT_DOT] = ACTIONS(3576),
    [anon_sym_STAR_STAR] = ACTIONS(3576),
    [anon_sym_PIPE_PIPE] = ACTIONS(3576),
    [anon_sym_SLASH_SLASH] = ACTIONS(3576),
    [anon_sym_CARET_CARET] = ACTIONS(3574),
    [anon_sym_AMP_AMP] = ACTIONS(3576),
    [anon_sym_PIPE] = ACTIONS(3576),
    [anon_sym_CARET] = ACTIONS(3576),
    [anon_sym_AMP] = ACTIONS(3576),
    [anon_sym_GT_GT2] = ACTIONS(3576),
    [anon_sym_PLUS] = ACTIONS(3576),
    [anon_sym_DASH] = ACTIONS(3576),
    [anon_sym_DOT] = ACTIONS(3576),
    [anon_sym_SLASH] = ACTIONS(3576),
    [anon_sym_x] = ACTIONS(3576),
    [anon_sym_EQ_TILDE] = ACTIONS(3574),
    [anon_sym_BANG_TILDE] = ACTIONS(3574),
    [anon_sym_EQ_EQ] = ACTIONS(3576),
    [anon_sym_BANG_EQ] = ACTIONS(3574),
    [anon_sym_eq] = ACTIONS(3576),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3574),
    [anon_sym_equ] = ACTIONS(3574),
    [anon_sym_eqr] = ACTIONS(3574),
    [anon_sym_ne] = ACTIONS(3574),
    [anon_sym_LT_EQ_GT] = ACTIONS(3574),
    [anon_sym_cmp] = ACTIONS(3574),
    [anon_sym_TILDE_TILDE] = ACTIONS(3574),
    [anon_sym_LT] = ACTIONS(3576),
    [anon_sym_LT_EQ] = ACTIONS(3576),
    [anon_sym_GT_EQ] = ACTIONS(3574),
    [anon_sym_lt] = ACTIONS(3574),
    [anon_sym_le] = ACTIONS(3574),
    [anon_sym_ge] = ACTIONS(3574),
    [anon_sym_gt] = ACTIONS(3574),
    [anon_sym_isa] = ACTIONS(3574),
    [anon_sym_PLUS_PLUS] = ACTIONS(3574),
    [anon_sym_DASH_DASH] = ACTIONS(3574),
    [anon_sym_QMARK] = ACTIONS(3574),
    [anon_sym_EQ_GT] = ACTIONS(3574),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3574),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3366)] = {
    [sym_heredoc_content] = STATE(3366),
    [anon_sym_COLON] = ACTIONS(3627),
    [anon_sym_SEMI] = ACTIONS(3627),
    [anon_sym_EQ] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_COMMA] = ACTIONS(3627),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3627),
    [anon_sym_DASH_GT] = ACTIONS(3627),
    [anon_sym_STAR] = ACTIONS(3629),
    [anon_sym_PERCENT] = ACTIONS(3629),
    [anon_sym_GT] = ACTIONS(3629),
    [anon_sym_LT_LT] = ACTIONS(3629),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3627),
    [anon_sym_PLUS_EQ] = ACTIONS(3627),
    [anon_sym_DASH_EQ] = ACTIONS(3627),
    [anon_sym_DOT_EQ] = ACTIONS(3627),
    [anon_sym_STAR_EQ] = ACTIONS(3627),
    [anon_sym_SLASH_EQ] = ACTIONS(3627),
    [anon_sym_PERCENT_EQ] = ACTIONS(3627),
    [anon_sym_x_EQ] = ACTIONS(3627),
    [anon_sym_AMP_EQ] = ACTIONS(3627),
    [anon_sym_PIPE_EQ] = ACTIONS(3627),
    [anon_sym_CARET_EQ] = ACTIONS(3627),
    [anon_sym_LT_LT_EQ] = ACTIONS(3627),
    [anon_sym_GT_GT_EQ] = ACTIONS(3627),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3627),
    [anon_sym_DOT_DOT] = ACTIONS(3629),
    [anon_sym_STAR_STAR] = ACTIONS(3629),
    [anon_sym_PIPE_PIPE] = ACTIONS(3629),
    [anon_sym_SLASH_SLASH] = ACTIONS(3629),
    [anon_sym_CARET_CARET] = ACTIONS(3627),
    [anon_sym_AMP_AMP] = ACTIONS(3629),
    [anon_sym_PIPE] = ACTIONS(3629),
    [anon_sym_CARET] = ACTIONS(3629),
    [anon_sym_AMP] = ACTIONS(3629),
    [anon_sym_GT_GT2] = ACTIONS(3629),
    [anon_sym_PLUS] = ACTIONS(3629),
    [anon_sym_DASH] = ACTIONS(3629),
    [anon_sym_DOT] = ACTIONS(3629),
    [anon_sym_SLASH] = ACTIONS(3629),
    [anon_sym_x] = ACTIONS(3629),
    [anon_sym_EQ_TILDE] = ACTIONS(3627),
    [anon_sym_BANG_TILDE] = ACTIONS(3627),
    [anon_sym_EQ_EQ] = ACTIONS(3629),
    [anon_sym_BANG_EQ] = ACTIONS(3627),
    [anon_sym_eq] = ACTIONS(3629),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3627),
    [anon_sym_equ] = ACTIONS(3627),
    [anon_sym_eqr] = ACTIONS(3627),
    [anon_sym_ne] = ACTIONS(3627),
    [anon_sym_LT_EQ_GT] = ACTIONS(3627),
    [anon_sym_cmp] = ACTIONS(3627),
    [anon_sym_TILDE_TILDE] = ACTIONS(3627),
    [anon_sym_LT] = ACTIONS(3629),
    [anon_sym_LT_EQ] = ACTIONS(3629),
    [anon_sym_GT_EQ] = ACTIONS(3627),
    [anon_sym_lt] = ACTIONS(3627),
    [anon_sym_le] = ACTIONS(3627),
    [anon_sym_ge] = ACTIONS(3627),
    [anon_sym_gt] = ACTIONS(3627),
    [anon_sym_isa] = ACTIONS(3627),
    [anon_sym_PLUS_PLUS] = ACTIONS(3627),
    [anon_sym_DASH_DASH] = ACTIONS(3627),
    [anon_sym_QMARK] = ACTIONS(3627),
    [anon_sym_EQ_GT] = ACTIONS(3627),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3627),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3367)] = {
    [sym_heredoc_content] = STATE(3367),
    [anon_sym_SEMI] = ACTIONS(3699),
    [anon_sym_EQ] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_COMMA] = ACTIONS(3699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3699),
    [anon_sym_DASH_GT] = ACTIONS(3699),
    [anon_sym_STAR] = ACTIONS(3701),
    [anon_sym_LBRACK] = ACTIONS(3699),
    [anon_sym_PERCENT] = ACTIONS(3701),
    [anon_sym_GT] = ACTIONS(3701),
    [anon_sym_LT_LT] = ACTIONS(3701),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3699),
    [anon_sym_PLUS_EQ] = ACTIONS(3699),
    [anon_sym_DASH_EQ] = ACTIONS(3699),
    [anon_sym_DOT_EQ] = ACTIONS(3699),
    [anon_sym_STAR_EQ] = ACTIONS(3699),
    [anon_sym_SLASH_EQ] = ACTIONS(3699),
    [anon_sym_PERCENT_EQ] = ACTIONS(3699),
    [anon_sym_x_EQ] = ACTIONS(3699),
    [anon_sym_AMP_EQ] = ACTIONS(3699),
    [anon_sym_PIPE_EQ] = ACTIONS(3699),
    [anon_sym_CARET_EQ] = ACTIONS(3699),
    [anon_sym_LT_LT_EQ] = ACTIONS(3699),
    [anon_sym_GT_GT_EQ] = ACTIONS(3699),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3699),
    [anon_sym_DOT_DOT] = ACTIONS(3701),
    [anon_sym_STAR_STAR] = ACTIONS(3701),
    [anon_sym_PIPE_PIPE] = ACTIONS(3701),
    [anon_sym_SLASH_SLASH] = ACTIONS(3701),
    [anon_sym_CARET_CARET] = ACTIONS(3699),
    [anon_sym_AMP_AMP] = ACTIONS(3701),
    [anon_sym_PIPE] = ACTIONS(3701),
    [anon_sym_CARET] = ACTIONS(3701),
    [anon_sym_AMP] = ACTIONS(3701),
    [anon_sym_GT_GT2] = ACTIONS(3701),
    [anon_sym_PLUS] = ACTIONS(3701),
    [anon_sym_DASH] = ACTIONS(3701),
    [anon_sym_DOT] = ACTIONS(3701),
    [anon_sym_SLASH] = ACTIONS(3701),
    [anon_sym_x] = ACTIONS(3701),
    [anon_sym_EQ_TILDE] = ACTIONS(3699),
    [anon_sym_BANG_TILDE] = ACTIONS(3699),
    [anon_sym_EQ_EQ] = ACTIONS(3701),
    [anon_sym_BANG_EQ] = ACTIONS(3699),
    [anon_sym_eq] = ACTIONS(3701),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3699),
    [anon_sym_equ] = ACTIONS(3699),
    [anon_sym_eqr] = ACTIONS(3699),
    [anon_sym_ne] = ACTIONS(3699),
    [anon_sym_LT_EQ_GT] = ACTIONS(3699),
    [anon_sym_cmp] = ACTIONS(3699),
    [anon_sym_TILDE_TILDE] = ACTIONS(3699),
    [anon_sym_LT] = ACTIONS(3701),
    [anon_sym_LT_EQ] = ACTIONS(3701),
    [anon_sym_GT_EQ] = ACTIONS(3699),
    [anon_sym_lt] = ACTIONS(3699),
    [anon_sym_le] = ACTIONS(3699),
    [anon_sym_ge] = ACTIONS(3699),
    [anon_sym_gt] = ACTIONS(3699),
    [anon_sym_isa] = ACTIONS(3699),
    [anon_sym_PLUS_PLUS] = ACTIONS(3699),
    [anon_sym_DASH_DASH] = ACTIONS(3699),
    [anon_sym_QMARK] = ACTIONS(3699),
    [anon_sym_EQ_GT] = ACTIONS(3699),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3699),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3368)] = {
    [sym_heredoc_content] = STATE(3368),
    [anon_sym_SEMI] = ACTIONS(3703),
    [anon_sym_EQ] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_COMMA] = ACTIONS(3703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3703),
    [anon_sym_DASH_GT] = ACTIONS(3703),
    [anon_sym_STAR] = ACTIONS(3705),
    [anon_sym_PERCENT] = ACTIONS(3705),
    [anon_sym_GT] = ACTIONS(3705),
    [anon_sym_LT_LT] = ACTIONS(3705),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3703),
    [anon_sym_PLUS_EQ] = ACTIONS(3703),
    [anon_sym_DASH_EQ] = ACTIONS(3703),
    [anon_sym_DOT_EQ] = ACTIONS(3703),
    [anon_sym_STAR_EQ] = ACTIONS(3703),
    [anon_sym_SLASH_EQ] = ACTIONS(3703),
    [anon_sym_PERCENT_EQ] = ACTIONS(3703),
    [anon_sym_x_EQ] = ACTIONS(3705),
    [anon_sym_AMP_EQ] = ACTIONS(3703),
    [anon_sym_PIPE_EQ] = ACTIONS(3703),
    [anon_sym_CARET_EQ] = ACTIONS(3703),
    [anon_sym_LT_LT_EQ] = ACTIONS(3703),
    [anon_sym_GT_GT_EQ] = ACTIONS(3703),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3703),
    [anon_sym_DOT_DOT] = ACTIONS(3705),
    [anon_sym_STAR_STAR] = ACTIONS(3705),
    [anon_sym_PIPE_PIPE] = ACTIONS(3705),
    [anon_sym_SLASH_SLASH] = ACTIONS(3705),
    [anon_sym_CARET_CARET] = ACTIONS(3703),
    [anon_sym_AMP_AMP] = ACTIONS(3705),
    [anon_sym_PIPE] = ACTIONS(3705),
    [anon_sym_CARET] = ACTIONS(3705),
    [anon_sym_AMP] = ACTIONS(3705),
    [anon_sym_GT_GT2] = ACTIONS(3705),
    [anon_sym_PLUS] = ACTIONS(3705),
    [anon_sym_DASH] = ACTIONS(3705),
    [anon_sym_DOT] = ACTIONS(3705),
    [anon_sym_SLASH] = ACTIONS(3705),
    [anon_sym_x] = ACTIONS(3705),
    [anon_sym_EQ_TILDE] = ACTIONS(3703),
    [anon_sym_BANG_TILDE] = ACTIONS(3703),
    [anon_sym_EQ_EQ] = ACTIONS(3705),
    [anon_sym_BANG_EQ] = ACTIONS(3703),
    [anon_sym_eq] = ACTIONS(3705),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3703),
    [anon_sym_equ] = ACTIONS(3703),
    [anon_sym_eqr] = ACTIONS(3703),
    [anon_sym_ne] = ACTIONS(3705),
    [anon_sym_LT_EQ_GT] = ACTIONS(3703),
    [anon_sym_cmp] = ACTIONS(3703),
    [anon_sym_TILDE_TILDE] = ACTIONS(3703),
    [anon_sym_LT] = ACTIONS(3705),
    [anon_sym_LT_EQ] = ACTIONS(3705),
    [anon_sym_GT_EQ] = ACTIONS(3703),
    [anon_sym_lt] = ACTIONS(3705),
    [anon_sym_le] = ACTIONS(3705),
    [anon_sym_ge] = ACTIONS(3703),
    [anon_sym_gt] = ACTIONS(3703),
    [anon_sym_isa] = ACTIONS(3705),
    [anon_sym_PLUS_PLUS] = ACTIONS(3703),
    [anon_sym_DASH_DASH] = ACTIONS(3703),
    [anon_sym_QMARK] = ACTIONS(3703),
    [anon_sym_EQ_GT] = ACTIONS(3703),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(5181),
    [sym__PERLY_SEMICOLON] = ACTIONS(3703),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3369)] = {
    [sym_heredoc_content] = STATE(3369),
    [anon_sym_SEMI] = ACTIONS(3709),
    [anon_sym_EQ] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_COMMA] = ACTIONS(3709),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3709),
    [anon_sym_DASH_GT] = ACTIONS(3709),
    [anon_sym_STAR] = ACTIONS(3711),
    [anon_sym_PERCENT] = ACTIONS(3711),
    [anon_sym_GT] = ACTIONS(3711),
    [anon_sym_LT_LT] = ACTIONS(3711),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3709),
    [anon_sym_PLUS_EQ] = ACTIONS(3709),
    [anon_sym_DASH_EQ] = ACTIONS(3709),
    [anon_sym_DOT_EQ] = ACTIONS(3709),
    [anon_sym_STAR_EQ] = ACTIONS(3709),
    [anon_sym_SLASH_EQ] = ACTIONS(3709),
    [anon_sym_PERCENT_EQ] = ACTIONS(3709),
    [anon_sym_x_EQ] = ACTIONS(3711),
    [anon_sym_AMP_EQ] = ACTIONS(3709),
    [anon_sym_PIPE_EQ] = ACTIONS(3709),
    [anon_sym_CARET_EQ] = ACTIONS(3709),
    [anon_sym_LT_LT_EQ] = ACTIONS(3709),
    [anon_sym_GT_GT_EQ] = ACTIONS(3709),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3709),
    [anon_sym_DOT_DOT] = ACTIONS(3711),
    [anon_sym_STAR_STAR] = ACTIONS(3711),
    [anon_sym_PIPE_PIPE] = ACTIONS(3711),
    [anon_sym_SLASH_SLASH] = ACTIONS(3711),
    [anon_sym_CARET_CARET] = ACTIONS(3709),
    [anon_sym_AMP_AMP] = ACTIONS(3711),
    [anon_sym_PIPE] = ACTIONS(3711),
    [anon_sym_CARET] = ACTIONS(3711),
    [anon_sym_AMP] = ACTIONS(3711),
    [anon_sym_GT_GT2] = ACTIONS(3711),
    [anon_sym_PLUS] = ACTIONS(3711),
    [anon_sym_DASH] = ACTIONS(3711),
    [anon_sym_DOT] = ACTIONS(3711),
    [anon_sym_SLASH] = ACTIONS(3711),
    [anon_sym_x] = ACTIONS(3711),
    [anon_sym_EQ_TILDE] = ACTIONS(3709),
    [anon_sym_BANG_TILDE] = ACTIONS(3709),
    [anon_sym_EQ_EQ] = ACTIONS(3711),
    [anon_sym_BANG_EQ] = ACTIONS(3709),
    [anon_sym_eq] = ACTIONS(3711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3709),
    [anon_sym_equ] = ACTIONS(3709),
    [anon_sym_eqr] = ACTIONS(3709),
    [anon_sym_ne] = ACTIONS(3711),
    [anon_sym_LT_EQ_GT] = ACTIONS(3709),
    [anon_sym_cmp] = ACTIONS(3711),
    [anon_sym_TILDE_TILDE] = ACTIONS(3709),
    [anon_sym_LT] = ACTIONS(3711),
    [anon_sym_LT_EQ] = ACTIONS(3711),
    [anon_sym_GT_EQ] = ACTIONS(3709),
    [anon_sym_lt] = ACTIONS(3711),
    [anon_sym_le] = ACTIONS(3711),
    [anon_sym_ge] = ACTIONS(3711),
    [anon_sym_gt] = ACTIONS(3711),
    [anon_sym_isa] = ACTIONS(3711),
    [anon_sym_PLUS_PLUS] = ACTIONS(3709),
    [anon_sym_DASH_DASH] = ACTIONS(3709),
    [anon_sym_QMARK] = ACTIONS(3709),
    [anon_sym_EQ_GT] = ACTIONS(3709),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(5183),
    [sym__PERLY_SEMICOLON] = ACTIONS(3709),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3370)] = {
    [sym_heredoc_content] = STATE(3370),
    [anon_sym_SEMI] = ACTIONS(3615),
    [anon_sym_EQ] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_COMMA] = ACTIONS(3615),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3615),
    [anon_sym_DASH_GT] = ACTIONS(3615),
    [anon_sym_STAR] = ACTIONS(3617),
    [anon_sym_PERCENT] = ACTIONS(3617),
    [anon_sym_GT] = ACTIONS(3617),
    [anon_sym_LT_LT] = ACTIONS(3617),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3615),
    [anon_sym_PLUS_EQ] = ACTIONS(3615),
    [anon_sym_DASH_EQ] = ACTIONS(3615),
    [anon_sym_DOT_EQ] = ACTIONS(3615),
    [anon_sym_STAR_EQ] = ACTIONS(3615),
    [anon_sym_SLASH_EQ] = ACTIONS(3615),
    [anon_sym_PERCENT_EQ] = ACTIONS(3615),
    [anon_sym_x_EQ] = ACTIONS(3617),
    [anon_sym_AMP_EQ] = ACTIONS(3615),
    [anon_sym_PIPE_EQ] = ACTIONS(3615),
    [anon_sym_CARET_EQ] = ACTIONS(3615),
    [anon_sym_LT_LT_EQ] = ACTIONS(3615),
    [anon_sym_GT_GT_EQ] = ACTIONS(3615),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3615),
    [anon_sym_DOT_DOT] = ACTIONS(3617),
    [anon_sym_STAR_STAR] = ACTIONS(3617),
    [anon_sym_PIPE_PIPE] = ACTIONS(3617),
    [anon_sym_SLASH_SLASH] = ACTIONS(3617),
    [anon_sym_CARET_CARET] = ACTIONS(3615),
    [anon_sym_AMP_AMP] = ACTIONS(3617),
    [anon_sym_PIPE] = ACTIONS(3617),
    [anon_sym_CARET] = ACTIONS(3617),
    [anon_sym_AMP] = ACTIONS(3617),
    [anon_sym_GT_GT2] = ACTIONS(3617),
    [anon_sym_PLUS] = ACTIONS(3617),
    [anon_sym_DASH] = ACTIONS(3617),
    [anon_sym_DOT] = ACTIONS(3617),
    [anon_sym_SLASH] = ACTIONS(3617),
    [anon_sym_x] = ACTIONS(3617),
    [anon_sym_EQ_TILDE] = ACTIONS(3615),
    [anon_sym_BANG_TILDE] = ACTIONS(3615),
    [anon_sym_EQ_EQ] = ACTIONS(3617),
    [anon_sym_BANG_EQ] = ACTIONS(3615),
    [anon_sym_eq] = ACTIONS(3617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3615),
    [anon_sym_equ] = ACTIONS(3617),
    [anon_sym_eqr] = ACTIONS(3617),
    [anon_sym_ne] = ACTIONS(3615),
    [anon_sym_LT_EQ_GT] = ACTIONS(3615),
    [anon_sym_cmp] = ACTIONS(3617),
    [anon_sym_TILDE_TILDE] = ACTIONS(3615),
    [anon_sym_LT] = ACTIONS(3617),
    [anon_sym_LT_EQ] = ACTIONS(3617),
    [anon_sym_GT_EQ] = ACTIONS(3615),
    [anon_sym_lt] = ACTIONS(3617),
    [anon_sym_le] = ACTIONS(3617),
    [anon_sym_ge] = ACTIONS(3617),
    [anon_sym_gt] = ACTIONS(3617),
    [anon_sym_isa] = ACTIONS(3617),
    [anon_sym_PLUS_PLUS] = ACTIONS(3615),
    [anon_sym_DASH_DASH] = ACTIONS(3615),
    [anon_sym_QMARK] = ACTIONS(3615),
    [anon_sym_EQ_GT] = ACTIONS(3615),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5185),
    [sym__PERLY_SEMICOLON] = ACTIONS(3615),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3371)] = {
    [sym_heredoc_content] = STATE(3371),
    [anon_sym_SEMI] = ACTIONS(3683),
    [anon_sym_EQ] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_COMMA] = ACTIONS(3683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3683),
    [anon_sym_DASH_GT] = ACTIONS(3683),
    [anon_sym_STAR] = ACTIONS(3685),
    [anon_sym_PERCENT] = ACTIONS(3685),
    [anon_sym_GT] = ACTIONS(3685),
    [anon_sym_LT_LT] = ACTIONS(3685),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3683),
    [anon_sym_PLUS_EQ] = ACTIONS(3683),
    [anon_sym_DASH_EQ] = ACTIONS(3683),
    [anon_sym_DOT_EQ] = ACTIONS(3683),
    [anon_sym_STAR_EQ] = ACTIONS(3683),
    [anon_sym_SLASH_EQ] = ACTIONS(3683),
    [anon_sym_PERCENT_EQ] = ACTIONS(3683),
    [anon_sym_x_EQ] = ACTIONS(3683),
    [anon_sym_AMP_EQ] = ACTIONS(3683),
    [anon_sym_PIPE_EQ] = ACTIONS(3683),
    [anon_sym_CARET_EQ] = ACTIONS(3683),
    [anon_sym_LT_LT_EQ] = ACTIONS(3683),
    [anon_sym_GT_GT_EQ] = ACTIONS(3683),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3683),
    [anon_sym_DOT_DOT] = ACTIONS(3685),
    [anon_sym_STAR_STAR] = ACTIONS(3685),
    [anon_sym_PIPE_PIPE] = ACTIONS(3685),
    [anon_sym_SLASH_SLASH] = ACTIONS(3685),
    [anon_sym_CARET_CARET] = ACTIONS(3683),
    [anon_sym_AMP_AMP] = ACTIONS(3685),
    [anon_sym_PIPE] = ACTIONS(3685),
    [anon_sym_CARET] = ACTIONS(3685),
    [anon_sym_AMP] = ACTIONS(3685),
    [anon_sym_GT_GT2] = ACTIONS(3685),
    [anon_sym_PLUS] = ACTIONS(3685),
    [anon_sym_DASH] = ACTIONS(3685),
    [anon_sym_DOT] = ACTIONS(3685),
    [anon_sym_SLASH] = ACTIONS(3685),
    [anon_sym_x] = ACTIONS(3685),
    [anon_sym_EQ_TILDE] = ACTIONS(3683),
    [anon_sym_BANG_TILDE] = ACTIONS(3683),
    [anon_sym_EQ_EQ] = ACTIONS(3685),
    [anon_sym_BANG_EQ] = ACTIONS(3683),
    [anon_sym_eq] = ACTIONS(3685),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3683),
    [anon_sym_equ] = ACTIONS(3683),
    [anon_sym_eqr] = ACTIONS(3683),
    [anon_sym_ne] = ACTIONS(3683),
    [anon_sym_LT_EQ_GT] = ACTIONS(3683),
    [anon_sym_cmp] = ACTIONS(3685),
    [anon_sym_TILDE_TILDE] = ACTIONS(3683),
    [anon_sym_LT] = ACTIONS(3685),
    [anon_sym_LT_EQ] = ACTIONS(3685),
    [anon_sym_GT_EQ] = ACTIONS(3683),
    [anon_sym_lt] = ACTIONS(3683),
    [anon_sym_le] = ACTIONS(3683),
    [anon_sym_ge] = ACTIONS(3683),
    [anon_sym_gt] = ACTIONS(3683),
    [anon_sym_isa] = ACTIONS(3683),
    [anon_sym_PLUS_PLUS] = ACTIONS(3683),
    [anon_sym_DASH_DASH] = ACTIONS(3683),
    [anon_sym_QMARK] = ACTIONS(3683),
    [anon_sym_EQ_GT] = ACTIONS(3683),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5187),
    [sym__PERLY_SEMICOLON] = ACTIONS(3683),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3372)] = {
    [sym_heredoc_content] = STATE(3372),
    [anon_sym_RBRACE] = ACTIONS(3925),
    [anon_sym_COLON] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_RPAREN] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3373)] = {
    [sym_heredoc_content] = STATE(3373),
    [anon_sym_RBRACE] = ACTIONS(3929),
    [anon_sym_COLON] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_RPAREN] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3374)] = {
    [sym_heredoc_content] = STATE(3374),
    [aux_sym__term_rightward_repeat1] = STATE(3382),
    [anon_sym_RBRACE] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(5189),
    [anon_sym_RPAREN] = ACTIONS(3568),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(5189),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3375)] = {
    [sym_heredoc_content] = STATE(3375),
    [anon_sym_RBRACE] = ACTIONS(3665),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_RPAREN] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3376)] = {
    [sym_heredoc_content] = STATE(3376),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_match_regexp_modifiers] = ACTIONS(3435),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3377)] = {
    [sym_heredoc_content] = STATE(3377),
    [anon_sym_RBRACE] = ACTIONS(3933),
    [anon_sym_COLON] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_RPAREN] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3378)] = {
    [sym_heredoc_content] = STATE(3378),
    [anon_sym_RBRACE] = ACTIONS(3597),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_RPAREN] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3379)] = {
    [sym_heredoc_content] = STATE(3379),
    [anon_sym_RBRACE] = ACTIONS(3969),
    [anon_sym_COLON] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_RPAREN] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3380)] = {
    [sym_heredoc_content] = STATE(3380),
    [anon_sym_RBRACE] = ACTIONS(4079),
    [anon_sym_COLON] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_RPAREN] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3381)] = {
    [sym_heredoc_content] = STATE(3381),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_COLON] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3382)] = {
    [sym_heredoc_content] = STATE(3382),
    [aux_sym__term_rightward_repeat1] = STATE(3382),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(5191),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(5191),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3383)] = {
    [sym_heredoc_content] = STATE(3383),
    [anon_sym_RBRACE] = ACTIONS(4083),
    [anon_sym_COLON] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_RPAREN] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3384)] = {
    [sym_heredoc_content] = STATE(3384),
    [anon_sym_RBRACE] = ACTIONS(4091),
    [anon_sym_COLON] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_RPAREN] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3385)] = {
    [sym_heredoc_content] = STATE(3385),
    [anon_sym_RBRACE] = ACTIONS(3715),
    [anon_sym_COLON] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_RPAREN] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3386)] = {
    [sym_heredoc_content] = STATE(3386),
    [aux_sym__term_rightward_repeat1] = STATE(5111),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5198),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5198),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3387)] = {
    [sym_heredoc_content] = STATE(3387),
    [anon_sym_RBRACE] = ACTIONS(3893),
    [anon_sym_COLON] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_RPAREN] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3388)] = {
    [sym_heredoc_content] = STATE(3388),
    [anon_sym_SEMI] = ACTIONS(3562),
    [anon_sym_EQ] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_COMMA] = ACTIONS(3562),
    [anon_sym_LPAREN] = ACTIONS(5242),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3562),
    [anon_sym_DASH_GT] = ACTIONS(3562),
    [anon_sym_STAR] = ACTIONS(3564),
    [anon_sym_PERCENT] = ACTIONS(3564),
    [anon_sym_GT] = ACTIONS(3564),
    [anon_sym_LT_LT] = ACTIONS(3564),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3562),
    [anon_sym_PLUS_EQ] = ACTIONS(3562),
    [anon_sym_DASH_EQ] = ACTIONS(3562),
    [anon_sym_DOT_EQ] = ACTIONS(3562),
    [anon_sym_STAR_EQ] = ACTIONS(3562),
    [anon_sym_SLASH_EQ] = ACTIONS(3562),
    [anon_sym_PERCENT_EQ] = ACTIONS(3562),
    [anon_sym_x_EQ] = ACTIONS(3562),
    [anon_sym_AMP_EQ] = ACTIONS(3562),
    [anon_sym_PIPE_EQ] = ACTIONS(3562),
    [anon_sym_CARET_EQ] = ACTIONS(3562),
    [anon_sym_LT_LT_EQ] = ACTIONS(3562),
    [anon_sym_GT_GT_EQ] = ACTIONS(3562),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3562),
    [anon_sym_DOT_DOT] = ACTIONS(3564),
    [anon_sym_STAR_STAR] = ACTIONS(3564),
    [anon_sym_PIPE_PIPE] = ACTIONS(3564),
    [anon_sym_SLASH_SLASH] = ACTIONS(3564),
    [anon_sym_CARET_CARET] = ACTIONS(3562),
    [anon_sym_AMP_AMP] = ACTIONS(3564),
    [anon_sym_PIPE] = ACTIONS(3564),
    [anon_sym_CARET] = ACTIONS(3564),
    [anon_sym_AMP] = ACTIONS(3564),
    [anon_sym_GT_GT2] = ACTIONS(3564),
    [anon_sym_PLUS] = ACTIONS(3564),
    [anon_sym_DASH] = ACTIONS(3564),
    [anon_sym_DOT] = ACTIONS(3564),
    [anon_sym_SLASH] = ACTIONS(3564),
    [anon_sym_x] = ACTIONS(3564),
    [anon_sym_EQ_TILDE] = ACTIONS(3562),
    [anon_sym_BANG_TILDE] = ACTIONS(3562),
    [anon_sym_EQ_EQ] = ACTIONS(3564),
    [anon_sym_BANG_EQ] = ACTIONS(3562),
    [anon_sym_eq] = ACTIONS(3564),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3562),
    [anon_sym_equ] = ACTIONS(3562),
    [anon_sym_eqr] = ACTIONS(3562),
    [anon_sym_ne] = ACTIONS(3562),
    [anon_sym_LT_EQ_GT] = ACTIONS(3562),
    [anon_sym_cmp] = ACTIONS(3562),
    [anon_sym_TILDE_TILDE] = ACTIONS(3562),
    [anon_sym_LT] = ACTIONS(3564),
    [anon_sym_LT_EQ] = ACTIONS(3564),
    [anon_sym_GT_EQ] = ACTIONS(3562),
    [anon_sym_lt] = ACTIONS(3562),
    [anon_sym_le] = ACTIONS(3562),
    [anon_sym_ge] = ACTIONS(3562),
    [anon_sym_gt] = ACTIONS(3562),
    [anon_sym_isa] = ACTIONS(3562),
    [anon_sym_PLUS_PLUS] = ACTIONS(3562),
    [anon_sym_DASH_DASH] = ACTIONS(3562),
    [anon_sym_QMARK] = ACTIONS(3562),
    [anon_sym_EQ_GT] = ACTIONS(3562),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3562),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3389)] = {
    [sym_heredoc_content] = STATE(3389),
    [anon_sym_RBRACE] = ACTIONS(4063),
    [anon_sym_COLON] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_RPAREN] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3390)] = {
    [sym_heredoc_content] = STATE(3390),
    [anon_sym_RBRACE] = ACTIONS(3877),
    [anon_sym_COLON] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_RPAREN] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3391)] = {
    [sym_heredoc_content] = STATE(3391),
    [anon_sym_RBRACE] = ACTIONS(3989),
    [anon_sym_COLON] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_RPAREN] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3392)] = {
    [sym_heredoc_content] = STATE(3392),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_SEMI] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3661),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3393)] = {
    [sym_heredoc_content] = STATE(3393),
    [anon_sym_RBRACE] = ACTIONS(4115),
    [anon_sym_COLON] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_RPAREN] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3394)] = {
    [sym_heredoc_content] = STATE(3394),
    [anon_sym_SEMI] = ACTIONS(3538),
    [anon_sym_EQ] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_COMMA] = ACTIONS(3538),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3538),
    [anon_sym_DASH_GT] = ACTIONS(3538),
    [anon_sym_STAR] = ACTIONS(3540),
    [anon_sym_PERCENT] = ACTIONS(3540),
    [anon_sym_GT] = ACTIONS(3540),
    [anon_sym_LT_LT] = ACTIONS(3540),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3538),
    [anon_sym_PLUS_EQ] = ACTIONS(3538),
    [anon_sym_DASH_EQ] = ACTIONS(3538),
    [anon_sym_DOT_EQ] = ACTIONS(3538),
    [anon_sym_STAR_EQ] = ACTIONS(3538),
    [anon_sym_SLASH_EQ] = ACTIONS(3538),
    [anon_sym_PERCENT_EQ] = ACTIONS(3538),
    [anon_sym_x_EQ] = ACTIONS(3540),
    [anon_sym_AMP_EQ] = ACTIONS(3538),
    [anon_sym_PIPE_EQ] = ACTIONS(3538),
    [anon_sym_CARET_EQ] = ACTIONS(3538),
    [anon_sym_LT_LT_EQ] = ACTIONS(3538),
    [anon_sym_GT_GT_EQ] = ACTIONS(3538),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3538),
    [anon_sym_DOT_DOT] = ACTIONS(3540),
    [anon_sym_STAR_STAR] = ACTIONS(3540),
    [anon_sym_PIPE_PIPE] = ACTIONS(3540),
    [anon_sym_SLASH_SLASH] = ACTIONS(3540),
    [anon_sym_CARET_CARET] = ACTIONS(3538),
    [anon_sym_AMP_AMP] = ACTIONS(3540),
    [anon_sym_PIPE] = ACTIONS(3540),
    [anon_sym_CARET] = ACTIONS(3540),
    [anon_sym_AMP] = ACTIONS(3540),
    [anon_sym_GT_GT2] = ACTIONS(3540),
    [anon_sym_PLUS] = ACTIONS(3540),
    [anon_sym_DASH] = ACTIONS(3540),
    [anon_sym_DOT] = ACTIONS(3540),
    [anon_sym_SLASH] = ACTIONS(3540),
    [anon_sym_x] = ACTIONS(3540),
    [anon_sym_EQ_TILDE] = ACTIONS(3538),
    [anon_sym_BANG_TILDE] = ACTIONS(3538),
    [anon_sym_EQ_EQ] = ACTIONS(3540),
    [anon_sym_BANG_EQ] = ACTIONS(3538),
    [anon_sym_eq] = ACTIONS(3540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3538),
    [anon_sym_equ] = ACTIONS(3540),
    [anon_sym_eqr] = ACTIONS(3540),
    [anon_sym_ne] = ACTIONS(3538),
    [anon_sym_LT_EQ_GT] = ACTIONS(3538),
    [anon_sym_cmp] = ACTIONS(3540),
    [anon_sym_TILDE_TILDE] = ACTIONS(3538),
    [anon_sym_LT] = ACTIONS(3540),
    [anon_sym_LT_EQ] = ACTIONS(3540),
    [anon_sym_GT_EQ] = ACTIONS(3538),
    [anon_sym_lt] = ACTIONS(3540),
    [anon_sym_le] = ACTIONS(3540),
    [anon_sym_ge] = ACTIONS(3540),
    [anon_sym_gt] = ACTIONS(3540),
    [anon_sym_isa] = ACTIONS(3540),
    [anon_sym_PLUS_PLUS] = ACTIONS(3538),
    [anon_sym_DASH_DASH] = ACTIONS(3538),
    [anon_sym_QMARK] = ACTIONS(3538),
    [anon_sym_EQ_GT] = ACTIONS(3538),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5244),
    [sym__PERLY_SEMICOLON] = ACTIONS(3538),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3395)] = {
    [sym_heredoc_content] = STATE(3395),
    [anon_sym_SEMI] = ACTIONS(3544),
    [anon_sym_EQ] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_COMMA] = ACTIONS(3544),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3544),
    [anon_sym_DASH_GT] = ACTIONS(3544),
    [anon_sym_STAR] = ACTIONS(3546),
    [anon_sym_PERCENT] = ACTIONS(3546),
    [anon_sym_GT] = ACTIONS(3546),
    [anon_sym_LT_LT] = ACTIONS(3546),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3544),
    [anon_sym_PLUS_EQ] = ACTIONS(3544),
    [anon_sym_DASH_EQ] = ACTIONS(3544),
    [anon_sym_DOT_EQ] = ACTIONS(3544),
    [anon_sym_STAR_EQ] = ACTIONS(3544),
    [anon_sym_SLASH_EQ] = ACTIONS(3544),
    [anon_sym_PERCENT_EQ] = ACTIONS(3544),
    [anon_sym_x_EQ] = ACTIONS(3546),
    [anon_sym_AMP_EQ] = ACTIONS(3544),
    [anon_sym_PIPE_EQ] = ACTIONS(3544),
    [anon_sym_CARET_EQ] = ACTIONS(3544),
    [anon_sym_LT_LT_EQ] = ACTIONS(3544),
    [anon_sym_GT_GT_EQ] = ACTIONS(3544),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3544),
    [anon_sym_DOT_DOT] = ACTIONS(3546),
    [anon_sym_STAR_STAR] = ACTIONS(3546),
    [anon_sym_PIPE_PIPE] = ACTIONS(3546),
    [anon_sym_SLASH_SLASH] = ACTIONS(3546),
    [anon_sym_CARET_CARET] = ACTIONS(3544),
    [anon_sym_AMP_AMP] = ACTIONS(3546),
    [anon_sym_PIPE] = ACTIONS(3546),
    [anon_sym_CARET] = ACTIONS(3546),
    [anon_sym_AMP] = ACTIONS(3546),
    [anon_sym_GT_GT2] = ACTIONS(3546),
    [anon_sym_PLUS] = ACTIONS(3546),
    [anon_sym_DASH] = ACTIONS(3546),
    [anon_sym_DOT] = ACTIONS(3546),
    [anon_sym_SLASH] = ACTIONS(3546),
    [anon_sym_x] = ACTIONS(3546),
    [anon_sym_EQ_TILDE] = ACTIONS(3544),
    [anon_sym_BANG_TILDE] = ACTIONS(3544),
    [anon_sym_EQ_EQ] = ACTIONS(3546),
    [anon_sym_BANG_EQ] = ACTIONS(3544),
    [anon_sym_eq] = ACTIONS(3546),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3544),
    [anon_sym_equ] = ACTIONS(3546),
    [anon_sym_eqr] = ACTIONS(3546),
    [anon_sym_ne] = ACTIONS(3544),
    [anon_sym_LT_EQ_GT] = ACTIONS(3544),
    [anon_sym_cmp] = ACTIONS(3546),
    [anon_sym_TILDE_TILDE] = ACTIONS(3544),
    [anon_sym_LT] = ACTIONS(3546),
    [anon_sym_LT_EQ] = ACTIONS(3546),
    [anon_sym_GT_EQ] = ACTIONS(3544),
    [anon_sym_lt] = ACTIONS(3546),
    [anon_sym_le] = ACTIONS(3546),
    [anon_sym_ge] = ACTIONS(3546),
    [anon_sym_gt] = ACTIONS(3546),
    [anon_sym_isa] = ACTIONS(3546),
    [anon_sym_PLUS_PLUS] = ACTIONS(3544),
    [anon_sym_DASH_DASH] = ACTIONS(3544),
    [anon_sym_QMARK] = ACTIONS(3544),
    [anon_sym_EQ_GT] = ACTIONS(3544),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(5246),
    [sym__PERLY_SEMICOLON] = ACTIONS(3544),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3396)] = {
    [sym_heredoc_content] = STATE(3396),
    [anon_sym_SEMI] = ACTIONS(3578),
    [anon_sym_EQ] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_COMMA] = ACTIONS(3578),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3578),
    [anon_sym_DASH_GT] = ACTIONS(3578),
    [anon_sym_STAR] = ACTIONS(3580),
    [anon_sym_PERCENT] = ACTIONS(3580),
    [anon_sym_GT] = ACTIONS(3580),
    [anon_sym_LT_LT] = ACTIONS(3580),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3578),
    [anon_sym_PLUS_EQ] = ACTIONS(3578),
    [anon_sym_DASH_EQ] = ACTIONS(3578),
    [anon_sym_DOT_EQ] = ACTIONS(3578),
    [anon_sym_STAR_EQ] = ACTIONS(3578),
    [anon_sym_SLASH_EQ] = ACTIONS(3578),
    [anon_sym_PERCENT_EQ] = ACTIONS(3578),
    [anon_sym_x_EQ] = ACTIONS(3578),
    [anon_sym_AMP_EQ] = ACTIONS(3578),
    [anon_sym_PIPE_EQ] = ACTIONS(3578),
    [anon_sym_CARET_EQ] = ACTIONS(3578),
    [anon_sym_LT_LT_EQ] = ACTIONS(3578),
    [anon_sym_GT_GT_EQ] = ACTIONS(3578),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3578),
    [anon_sym_DOT_DOT] = ACTIONS(3580),
    [anon_sym_STAR_STAR] = ACTIONS(3580),
    [anon_sym_PIPE_PIPE] = ACTIONS(3580),
    [anon_sym_SLASH_SLASH] = ACTIONS(3580),
    [anon_sym_CARET_CARET] = ACTIONS(3578),
    [anon_sym_AMP_AMP] = ACTIONS(3580),
    [anon_sym_PIPE] = ACTIONS(3580),
    [anon_sym_CARET] = ACTIONS(3580),
    [anon_sym_AMP] = ACTIONS(3580),
    [anon_sym_GT_GT2] = ACTIONS(3580),
    [anon_sym_PLUS] = ACTIONS(3580),
    [anon_sym_DASH] = ACTIONS(3580),
    [anon_sym_DOT] = ACTIONS(3580),
    [anon_sym_SLASH] = ACTIONS(3580),
    [anon_sym_x] = ACTIONS(3580),
    [anon_sym_EQ_TILDE] = ACTIONS(3578),
    [anon_sym_BANG_TILDE] = ACTIONS(3578),
    [anon_sym_EQ_EQ] = ACTIONS(3580),
    [anon_sym_BANG_EQ] = ACTIONS(3578),
    [anon_sym_eq] = ACTIONS(3580),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3578),
    [anon_sym_equ] = ACTIONS(3578),
    [anon_sym_eqr] = ACTIONS(3578),
    [anon_sym_ne] = ACTIONS(3578),
    [anon_sym_LT_EQ_GT] = ACTIONS(3578),
    [anon_sym_cmp] = ACTIONS(3580),
    [anon_sym_TILDE_TILDE] = ACTIONS(3578),
    [anon_sym_LT] = ACTIONS(3580),
    [anon_sym_LT_EQ] = ACTIONS(3580),
    [anon_sym_GT_EQ] = ACTIONS(3578),
    [anon_sym_lt] = ACTIONS(3578),
    [anon_sym_le] = ACTIONS(3578),
    [anon_sym_ge] = ACTIONS(3578),
    [anon_sym_gt] = ACTIONS(3578),
    [anon_sym_isa] = ACTIONS(3578),
    [anon_sym_PLUS_PLUS] = ACTIONS(3578),
    [anon_sym_DASH_DASH] = ACTIONS(3578),
    [anon_sym_QMARK] = ACTIONS(3578),
    [anon_sym_EQ_GT] = ACTIONS(3578),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5248),
    [sym__PERLY_SEMICOLON] = ACTIONS(3578),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3397)] = {
    [sym_heredoc_content] = STATE(3397),
    [anon_sym_SEMI] = ACTIONS(3441),
    [anon_sym_EQ] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_COMMA] = ACTIONS(3441),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3441),
    [anon_sym_DASH_GT] = ACTIONS(3441),
    [anon_sym_STAR] = ACTIONS(3443),
    [anon_sym_PERCENT] = ACTIONS(3443),
    [anon_sym_GT] = ACTIONS(3443),
    [anon_sym_LT_LT] = ACTIONS(3443),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3441),
    [anon_sym_PLUS_EQ] = ACTIONS(3441),
    [anon_sym_DASH_EQ] = ACTIONS(3441),
    [anon_sym_DOT_EQ] = ACTIONS(3441),
    [anon_sym_STAR_EQ] = ACTIONS(3441),
    [anon_sym_SLASH_EQ] = ACTIONS(3441),
    [anon_sym_PERCENT_EQ] = ACTIONS(3441),
    [anon_sym_x_EQ] = ACTIONS(3441),
    [anon_sym_AMP_EQ] = ACTIONS(3441),
    [anon_sym_PIPE_EQ] = ACTIONS(3441),
    [anon_sym_CARET_EQ] = ACTIONS(3441),
    [anon_sym_LT_LT_EQ] = ACTIONS(3441),
    [anon_sym_GT_GT_EQ] = ACTIONS(3441),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3441),
    [anon_sym_DOT_DOT] = ACTIONS(3443),
    [anon_sym_STAR_STAR] = ACTIONS(3443),
    [anon_sym_PIPE_PIPE] = ACTIONS(3443),
    [anon_sym_SLASH_SLASH] = ACTIONS(3443),
    [anon_sym_CARET_CARET] = ACTIONS(3441),
    [anon_sym_AMP_AMP] = ACTIONS(3443),
    [anon_sym_PIPE] = ACTIONS(3443),
    [anon_sym_CARET] = ACTIONS(3443),
    [anon_sym_AMP] = ACTIONS(3443),
    [anon_sym_GT_GT2] = ACTIONS(3443),
    [anon_sym_PLUS] = ACTIONS(3443),
    [anon_sym_DASH] = ACTIONS(3443),
    [anon_sym_DOT] = ACTIONS(3443),
    [anon_sym_SLASH] = ACTIONS(3443),
    [anon_sym_x] = ACTIONS(3443),
    [anon_sym_EQ_TILDE] = ACTIONS(3441),
    [anon_sym_BANG_TILDE] = ACTIONS(3441),
    [anon_sym_EQ_EQ] = ACTIONS(3443),
    [anon_sym_BANG_EQ] = ACTIONS(3441),
    [anon_sym_eq] = ACTIONS(3443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3441),
    [anon_sym_equ] = ACTIONS(3441),
    [anon_sym_eqr] = ACTIONS(3441),
    [anon_sym_ne] = ACTIONS(3441),
    [anon_sym_LT_EQ_GT] = ACTIONS(3441),
    [anon_sym_cmp] = ACTIONS(3443),
    [anon_sym_TILDE_TILDE] = ACTIONS(3441),
    [anon_sym_LT] = ACTIONS(3443),
    [anon_sym_LT_EQ] = ACTIONS(3443),
    [anon_sym_GT_EQ] = ACTIONS(3441),
    [anon_sym_lt] = ACTIONS(3441),
    [anon_sym_le] = ACTIONS(3441),
    [anon_sym_ge] = ACTIONS(3441),
    [anon_sym_gt] = ACTIONS(3441),
    [anon_sym_isa] = ACTIONS(3441),
    [anon_sym_PLUS_PLUS] = ACTIONS(3441),
    [anon_sym_DASH_DASH] = ACTIONS(3441),
    [anon_sym_QMARK] = ACTIONS(3441),
    [anon_sym_EQ_GT] = ACTIONS(3441),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5250),
    [sym__PERLY_SEMICOLON] = ACTIONS(3441),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3398)] = {
    [sym_heredoc_content] = STATE(3398),
    [aux_sym__term_rightward_repeat1] = STATE(3418),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(5256),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(5256),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3399)] = {
    [sym_heredoc_content] = STATE(3399),
    [anon_sym_RBRACE] = ACTIONS(3611),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_RPAREN] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3400)] = {
    [sym_heredoc_content] = STATE(3400),
    [anon_sym_RBRACE] = ACTIONS(3447),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_RPAREN] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3401)] = {
    [sym_heredoc_content] = STATE(3401),
    [aux_sym__term_rightward_repeat1] = STATE(5041),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(5301),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(5301),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3402)] = {
    [sym_heredoc_content] = STATE(3402),
    [anon_sym_RBRACE] = ACTIONS(3937),
    [anon_sym_COLON] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_RPAREN] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3403)] = {
    [sym_heredoc_content] = STATE(3403),
    [anon_sym_RBRACE] = ACTIONS(3787),
    [anon_sym_COLON] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_RPAREN] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3404)] = {
    [sym_heredoc_content] = STATE(3404),
    [anon_sym_SEMI] = ACTIONS(3550),
    [anon_sym_EQ] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_COMMA] = ACTIONS(3550),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3550),
    [anon_sym_DASH_GT] = ACTIONS(3550),
    [anon_sym_STAR] = ACTIONS(3552),
    [anon_sym_PERCENT] = ACTIONS(3552),
    [anon_sym_GT] = ACTIONS(3552),
    [anon_sym_LT_LT] = ACTIONS(3552),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3550),
    [anon_sym_PLUS_EQ] = ACTIONS(3550),
    [anon_sym_DASH_EQ] = ACTIONS(3550),
    [anon_sym_DOT_EQ] = ACTIONS(3550),
    [anon_sym_STAR_EQ] = ACTIONS(3550),
    [anon_sym_SLASH_EQ] = ACTIONS(3550),
    [anon_sym_PERCENT_EQ] = ACTIONS(3550),
    [anon_sym_x_EQ] = ACTIONS(3550),
    [anon_sym_AMP_EQ] = ACTIONS(3550),
    [anon_sym_PIPE_EQ] = ACTIONS(3550),
    [anon_sym_CARET_EQ] = ACTIONS(3550),
    [anon_sym_LT_LT_EQ] = ACTIONS(3550),
    [anon_sym_GT_GT_EQ] = ACTIONS(3550),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3550),
    [anon_sym_DOT_DOT] = ACTIONS(3552),
    [anon_sym_STAR_STAR] = ACTIONS(3552),
    [anon_sym_PIPE_PIPE] = ACTIONS(3552),
    [anon_sym_SLASH_SLASH] = ACTIONS(3552),
    [anon_sym_CARET_CARET] = ACTIONS(3550),
    [anon_sym_AMP_AMP] = ACTIONS(3552),
    [anon_sym_PIPE] = ACTIONS(3552),
    [anon_sym_CARET] = ACTIONS(3552),
    [anon_sym_AMP] = ACTIONS(3552),
    [anon_sym_GT_GT2] = ACTIONS(3552),
    [anon_sym_PLUS] = ACTIONS(3552),
    [anon_sym_DASH] = ACTIONS(3552),
    [anon_sym_DOT] = ACTIONS(3552),
    [anon_sym_SLASH] = ACTIONS(3552),
    [anon_sym_x] = ACTIONS(3552),
    [anon_sym_EQ_TILDE] = ACTIONS(3550),
    [anon_sym_BANG_TILDE] = ACTIONS(3550),
    [anon_sym_EQ_EQ] = ACTIONS(3552),
    [anon_sym_BANG_EQ] = ACTIONS(3550),
    [anon_sym_eq] = ACTIONS(3552),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3550),
    [anon_sym_equ] = ACTIONS(3550),
    [anon_sym_eqr] = ACTIONS(3550),
    [anon_sym_ne] = ACTIONS(3550),
    [anon_sym_LT_EQ_GT] = ACTIONS(3550),
    [anon_sym_cmp] = ACTIONS(3552),
    [anon_sym_TILDE_TILDE] = ACTIONS(3550),
    [anon_sym_LT] = ACTIONS(3552),
    [anon_sym_LT_EQ] = ACTIONS(3552),
    [anon_sym_GT_EQ] = ACTIONS(3550),
    [anon_sym_lt] = ACTIONS(3550),
    [anon_sym_le] = ACTIONS(3550),
    [anon_sym_ge] = ACTIONS(3550),
    [anon_sym_gt] = ACTIONS(3550),
    [anon_sym_isa] = ACTIONS(3550),
    [anon_sym_PLUS_PLUS] = ACTIONS(3550),
    [anon_sym_DASH_DASH] = ACTIONS(3550),
    [anon_sym_QMARK] = ACTIONS(3550),
    [anon_sym_EQ_GT] = ACTIONS(3550),
    [sym_comment] = ACTIONS(5),
    [sym_transliteration_modifiers] = ACTIONS(5303),
    [sym__PERLY_SEMICOLON] = ACTIONS(3550),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3405)] = {
    [sym_heredoc_content] = STATE(3405),
    [anon_sym_RBRACE] = ACTIONS(3731),
    [anon_sym_COLON] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_RPAREN] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3406)] = {
    [sym_heredoc_content] = STATE(3406),
    [anon_sym_COLON] = ACTIONS(3977),
    [anon_sym_DOLLAR] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_AT] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3979),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_PERCENT2] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3407)] = {
    [sym_heredoc_content] = STATE(3407),
    [anon_sym_COLON] = ACTIONS(3997),
    [anon_sym_DOLLAR] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_AT] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3999),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_PERCENT2] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3408)] = {
    [sym_heredoc_content] = STATE(3408),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_DOLLAR] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_AT] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3697),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_PERCENT2] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3409)] = {
    [sym_heredoc_content] = STATE(3409),
    [anon_sym_COLON] = ACTIONS(3909),
    [anon_sym_DOLLAR] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_AT] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3911),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_PERCENT2] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3410)] = {
    [sym_heredoc_content] = STATE(3410),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_DOLLAR] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_AT] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4019),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_PERCENT2] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3411)] = {
    [sym_heredoc_content] = STATE(3411),
    [anon_sym_COLON] = ACTIONS(3803),
    [anon_sym_DOLLAR] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_AT] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5139),
    [anon_sym_PLUS_EQ] = ACTIONS(5139),
    [anon_sym_DASH_EQ] = ACTIONS(5139),
    [anon_sym_DOT_EQ] = ACTIONS(5139),
    [anon_sym_STAR_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_PERCENT_EQ] = ACTIONS(5137),
    [anon_sym_x_EQ] = ACTIONS(5139),
    [anon_sym_AMP_EQ] = ACTIONS(5139),
    [anon_sym_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_CARET_EQ] = ACTIONS(5139),
    [anon_sym_LT_LT_EQ] = ACTIONS(5139),
    [anon_sym_GT_GT_EQ] = ACTIONS(5139),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5139),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3412)] = {
    [sym_heredoc_content] = STATE(3412),
    [anon_sym_COLON] = ACTIONS(3845),
    [anon_sym_DOLLAR] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_AT] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3847),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_PERCENT2] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3413)] = {
    [sym_heredoc_content] = STATE(3413),
    [anon_sym_COLON] = ACTIONS(3739),
    [anon_sym_DOLLAR] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_AT] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3741),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_PERCENT2] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3414)] = {
    [sym_heredoc_content] = STATE(3414),
    [anon_sym_COLON] = ACTIONS(3897),
    [anon_sym_DOLLAR] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_AT] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5139),
    [anon_sym_PLUS_EQ] = ACTIONS(5139),
    [anon_sym_DASH_EQ] = ACTIONS(5139),
    [anon_sym_DOT_EQ] = ACTIONS(5139),
    [anon_sym_STAR_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_PERCENT_EQ] = ACTIONS(5137),
    [anon_sym_x_EQ] = ACTIONS(5139),
    [anon_sym_AMP_EQ] = ACTIONS(5139),
    [anon_sym_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_CARET_EQ] = ACTIONS(5139),
    [anon_sym_LT_LT_EQ] = ACTIONS(5139),
    [anon_sym_GT_GT_EQ] = ACTIONS(5139),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5139),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3415)] = {
    [sym_heredoc_content] = STATE(3415),
    [anon_sym_RBRACE] = ACTIONS(3807),
    [anon_sym_COLON] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_RPAREN] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3416)] = {
    [sym_heredoc_content] = STATE(3416),
    [anon_sym_COLON] = ACTIONS(4151),
    [anon_sym_DOLLAR] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_AT] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4153),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_PERCENT2] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3417)] = {
    [sym_heredoc_content] = STATE(3417),
    [anon_sym_RBRACE] = ACTIONS(4041),
    [anon_sym_COLON] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_RPAREN] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3418)] = {
    [sym_heredoc_content] = STATE(3418),
    [aux_sym__term_rightward_repeat1] = STATE(3455),
    [anon_sym_SEMI] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(5305),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(5305),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3568),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3419)] = {
    [sym_heredoc_content] = STATE(3419),
    [anon_sym_RBRACE] = ACTIONS(4045),
    [anon_sym_COLON] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_RPAREN] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3420)] = {
    [sym_heredoc_content] = STATE(3420),
    [anon_sym_RBRACE] = ACTIONS(4047),
    [anon_sym_COLON] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_RPAREN] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3421)] = {
    [sym_heredoc_content] = STATE(3421),
    [anon_sym_RBRACE] = ACTIONS(4055),
    [anon_sym_COLON] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_RPAREN] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3422)] = {
    [sym_heredoc_content] = STATE(3422),
    [anon_sym_RBRACE] = ACTIONS(3791),
    [anon_sym_COLON] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_RPAREN] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3423)] = {
    [sym_heredoc_content] = STATE(3423),
    [anon_sym_RBRACE] = ACTIONS(3661),
    [anon_sym_COLON] = ACTIONS(3661),
    [anon_sym_EQ] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_COMMA] = ACTIONS(3661),
    [anon_sym_RPAREN] = ACTIONS(3661),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3661),
    [anon_sym_DASH_GT] = ACTIONS(3661),
    [anon_sym_STAR] = ACTIONS(3663),
    [anon_sym_PERCENT] = ACTIONS(3663),
    [anon_sym_GT] = ACTIONS(3663),
    [anon_sym_LT_LT] = ACTIONS(3663),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3661),
    [anon_sym_PLUS_EQ] = ACTIONS(3661),
    [anon_sym_DASH_EQ] = ACTIONS(3661),
    [anon_sym_DOT_EQ] = ACTIONS(3661),
    [anon_sym_STAR_EQ] = ACTIONS(3661),
    [anon_sym_SLASH_EQ] = ACTIONS(3661),
    [anon_sym_PERCENT_EQ] = ACTIONS(3661),
    [anon_sym_x_EQ] = ACTIONS(3661),
    [anon_sym_AMP_EQ] = ACTIONS(3661),
    [anon_sym_PIPE_EQ] = ACTIONS(3661),
    [anon_sym_CARET_EQ] = ACTIONS(3661),
    [anon_sym_LT_LT_EQ] = ACTIONS(3661),
    [anon_sym_GT_GT_EQ] = ACTIONS(3661),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3661),
    [anon_sym_DOT_DOT] = ACTIONS(3663),
    [anon_sym_STAR_STAR] = ACTIONS(3663),
    [anon_sym_PIPE_PIPE] = ACTIONS(3663),
    [anon_sym_SLASH_SLASH] = ACTIONS(3663),
    [anon_sym_CARET_CARET] = ACTIONS(3661),
    [anon_sym_AMP_AMP] = ACTIONS(3663),
    [anon_sym_PIPE] = ACTIONS(3663),
    [anon_sym_CARET] = ACTIONS(3663),
    [anon_sym_AMP] = ACTIONS(3663),
    [anon_sym_GT_GT2] = ACTIONS(3663),
    [anon_sym_PLUS] = ACTIONS(3663),
    [anon_sym_DASH] = ACTIONS(3663),
    [anon_sym_DOT] = ACTIONS(3663),
    [anon_sym_SLASH] = ACTIONS(3663),
    [anon_sym_x] = ACTIONS(3663),
    [anon_sym_EQ_TILDE] = ACTIONS(3661),
    [anon_sym_BANG_TILDE] = ACTIONS(3661),
    [anon_sym_EQ_EQ] = ACTIONS(3663),
    [anon_sym_BANG_EQ] = ACTIONS(3661),
    [anon_sym_eq] = ACTIONS(3663),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3661),
    [anon_sym_equ] = ACTIONS(3661),
    [anon_sym_eqr] = ACTIONS(3661),
    [anon_sym_ne] = ACTIONS(3661),
    [anon_sym_LT_EQ_GT] = ACTIONS(3661),
    [anon_sym_cmp] = ACTIONS(3661),
    [anon_sym_TILDE_TILDE] = ACTIONS(3661),
    [anon_sym_LT] = ACTIONS(3663),
    [anon_sym_LT_EQ] = ACTIONS(3663),
    [anon_sym_GT_EQ] = ACTIONS(3661),
    [anon_sym_lt] = ACTIONS(3661),
    [anon_sym_le] = ACTIONS(3661),
    [anon_sym_ge] = ACTIONS(3661),
    [anon_sym_gt] = ACTIONS(3661),
    [anon_sym_isa] = ACTIONS(3661),
    [anon_sym_PLUS_PLUS] = ACTIONS(3661),
    [anon_sym_DASH_DASH] = ACTIONS(3661),
    [anon_sym_QMARK] = ACTIONS(3661),
    [anon_sym_EQ_GT] = ACTIONS(3661),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3424)] = {
    [sym_heredoc_content] = STATE(3424),
    [anon_sym_RBRACE] = ACTIONS(3415),
    [anon_sym_COLON] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_RPAREN] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3425)] = {
    [sym_heredoc_content] = STATE(3425),
    [anon_sym_RBRACE] = ACTIONS(4071),
    [anon_sym_COLON] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_RPAREN] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3426)] = {
    [sym_heredoc_content] = STATE(3426),
    [anon_sym_RBRACE] = ACTIONS(4135),
    [anon_sym_COLON] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_RPAREN] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3427)] = {
    [sym_heredoc_content] = STATE(3427),
    [anon_sym_RBRACE] = ACTIONS(3941),
    [anon_sym_COLON] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_RPAREN] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3428)] = {
    [sym_heredoc_content] = STATE(3428),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_COLON] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3429)] = {
    [sym_heredoc_content] = STATE(3429),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3435),
    [anon_sym_eqr] = ACTIONS(3435),
    [anon_sym_ne] = ACTIONS(3437),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3435),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3435),
    [anon_sym_gt] = ACTIONS(3435),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_quoted_regexp_modifiers] = ACTIONS(3435),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3430)] = {
    [sym_heredoc_content] = STATE(3430),
    [anon_sym_RBRACE] = ACTIONS(3865),
    [anon_sym_COLON] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_RPAREN] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3431)] = {
    [sym_heredoc_content] = STATE(3431),
    [anon_sym_RBRACE] = ACTIONS(3829),
    [anon_sym_COLON] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_RPAREN] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3432)] = {
    [sym_heredoc_content] = STATE(3432),
    [anon_sym_RBRACE] = ACTIONS(3917),
    [anon_sym_COLON] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_RPAREN] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3433)] = {
    [sym_heredoc_content] = STATE(3433),
    [anon_sym_RBRACE] = ACTIONS(4095),
    [anon_sym_COLON] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_RPAREN] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3434)] = {
    [sym_heredoc_content] = STATE(3434),
    [anon_sym_RBRACE] = ACTIONS(3949),
    [anon_sym_COLON] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_RPAREN] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3435)] = {
    [sym_heredoc_content] = STATE(3435),
    [anon_sym_RBRACE] = ACTIONS(3957),
    [anon_sym_COLON] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_RPAREN] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3436)] = {
    [sym_heredoc_content] = STATE(3436),
    [anon_sym_RBRACE] = ACTIONS(4208),
    [anon_sym_COLON] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_RPAREN] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3437)] = {
    [sym_heredoc_content] = STATE(3437),
    [anon_sym_RBRACE] = ACTIONS(4184),
    [anon_sym_COLON] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_RPAREN] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3438)] = {
    [sym_heredoc_content] = STATE(3438),
    [aux_sym__term_rightward_repeat1] = STATE(3374),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5307),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5307),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3439)] = {
    [sym_heredoc_content] = STATE(3439),
    [anon_sym_RBRACE] = ACTIONS(3689),
    [anon_sym_COLON] = ACTIONS(5310),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_RPAREN] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3440)] = {
    [sym_heredoc_content] = STATE(3440),
    [anon_sym_RBRACE] = ACTIONS(3669),
    [anon_sym_COLON] = ACTIONS(5312),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_RPAREN] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3441)] = {
    [sym_heredoc_content] = STATE(3441),
    [anon_sym_COLON] = ACTIONS(3675),
    [anon_sym_SEMI] = ACTIONS(3675),
    [anon_sym_EQ] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_COMMA] = ACTIONS(3675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3675),
    [anon_sym_DASH_GT] = ACTIONS(3675),
    [anon_sym_STAR] = ACTIONS(3677),
    [anon_sym_PERCENT] = ACTIONS(3677),
    [anon_sym_GT] = ACTIONS(3677),
    [anon_sym_LT_LT] = ACTIONS(3677),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3675),
    [anon_sym_PLUS_EQ] = ACTIONS(3675),
    [anon_sym_DASH_EQ] = ACTIONS(3675),
    [anon_sym_DOT_EQ] = ACTIONS(3675),
    [anon_sym_STAR_EQ] = ACTIONS(3675),
    [anon_sym_SLASH_EQ] = ACTIONS(3675),
    [anon_sym_PERCENT_EQ] = ACTIONS(3675),
    [anon_sym_x_EQ] = ACTIONS(3675),
    [anon_sym_AMP_EQ] = ACTIONS(3675),
    [anon_sym_PIPE_EQ] = ACTIONS(3675),
    [anon_sym_CARET_EQ] = ACTIONS(3675),
    [anon_sym_LT_LT_EQ] = ACTIONS(3675),
    [anon_sym_GT_GT_EQ] = ACTIONS(3675),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3675),
    [anon_sym_DOT_DOT] = ACTIONS(3677),
    [anon_sym_STAR_STAR] = ACTIONS(3677),
    [anon_sym_PIPE_PIPE] = ACTIONS(3677),
    [anon_sym_SLASH_SLASH] = ACTIONS(3677),
    [anon_sym_CARET_CARET] = ACTIONS(3675),
    [anon_sym_AMP_AMP] = ACTIONS(3677),
    [anon_sym_PIPE] = ACTIONS(3677),
    [anon_sym_CARET] = ACTIONS(3677),
    [anon_sym_AMP] = ACTIONS(3677),
    [anon_sym_GT_GT2] = ACTIONS(3677),
    [anon_sym_PLUS] = ACTIONS(3677),
    [anon_sym_DASH] = ACTIONS(3677),
    [anon_sym_DOT] = ACTIONS(3677),
    [anon_sym_SLASH] = ACTIONS(3677),
    [anon_sym_x] = ACTIONS(3677),
    [anon_sym_EQ_TILDE] = ACTIONS(3675),
    [anon_sym_BANG_TILDE] = ACTIONS(3675),
    [anon_sym_EQ_EQ] = ACTIONS(3677),
    [anon_sym_BANG_EQ] = ACTIONS(3675),
    [anon_sym_eq] = ACTIONS(3677),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3675),
    [anon_sym_equ] = ACTIONS(3675),
    [anon_sym_eqr] = ACTIONS(3675),
    [anon_sym_ne] = ACTIONS(3675),
    [anon_sym_LT_EQ_GT] = ACTIONS(3675),
    [anon_sym_cmp] = ACTIONS(3675),
    [anon_sym_TILDE_TILDE] = ACTIONS(3675),
    [anon_sym_LT] = ACTIONS(3677),
    [anon_sym_LT_EQ] = ACTIONS(3677),
    [anon_sym_GT_EQ] = ACTIONS(3675),
    [anon_sym_lt] = ACTIONS(3675),
    [anon_sym_le] = ACTIONS(3675),
    [anon_sym_ge] = ACTIONS(3675),
    [anon_sym_gt] = ACTIONS(3675),
    [anon_sym_isa] = ACTIONS(3675),
    [anon_sym_PLUS_PLUS] = ACTIONS(3675),
    [anon_sym_DASH_DASH] = ACTIONS(3675),
    [anon_sym_QMARK] = ACTIONS(3675),
    [anon_sym_EQ_GT] = ACTIONS(3675),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3675),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3442)] = {
    [sym_heredoc_content] = STATE(3442),
    [anon_sym_RBRACE] = ACTIONS(4165),
    [anon_sym_COLON] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_RPAREN] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3443)] = {
    [sym_heredoc_content] = STATE(3443),
    [anon_sym_RBRACE] = ACTIONS(4204),
    [anon_sym_COLON] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_RPAREN] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3444)] = {
    [sym_heredoc_content] = STATE(3444),
    [anon_sym_RBRACE] = ACTIONS(3727),
    [anon_sym_COLON] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_RPAREN] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3445)] = {
    [sym_heredoc_content] = STATE(3445),
    [anon_sym_RBRACE] = ACTIONS(3869),
    [anon_sym_COLON] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_RPAREN] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3446)] = {
    [sym_heredoc_content] = STATE(3446),
    [anon_sym_RBRACE] = ACTIONS(4103),
    [anon_sym_COLON] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_RPAREN] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3447)] = {
    [sym_heredoc_content] = STATE(3447),
    [anon_sym_RBRACE] = ACTIONS(4159),
    [anon_sym_COLON] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_RPAREN] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3448)] = {
    [sym_heredoc_content] = STATE(3448),
    [anon_sym_RBRACE] = ACTIONS(4169),
    [anon_sym_COLON] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_RPAREN] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(5314),
    [anon_sym_DASH_DASH] = ACTIONS(5314),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3449)] = {
    [sym_heredoc_content] = STATE(3449),
    [anon_sym_SEMI] = ACTIONS(3435),
    [anon_sym_EQ] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_COMMA] = ACTIONS(3435),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3435),
    [anon_sym_DASH_GT] = ACTIONS(3435),
    [anon_sym_STAR] = ACTIONS(3437),
    [anon_sym_PERCENT] = ACTIONS(3437),
    [anon_sym_GT] = ACTIONS(3437),
    [anon_sym_LT_LT] = ACTIONS(3437),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3435),
    [anon_sym_PLUS_EQ] = ACTIONS(3435),
    [anon_sym_DASH_EQ] = ACTIONS(3435),
    [anon_sym_DOT_EQ] = ACTIONS(3435),
    [anon_sym_STAR_EQ] = ACTIONS(3435),
    [anon_sym_SLASH_EQ] = ACTIONS(3435),
    [anon_sym_PERCENT_EQ] = ACTIONS(3435),
    [anon_sym_x_EQ] = ACTIONS(3437),
    [anon_sym_AMP_EQ] = ACTIONS(3435),
    [anon_sym_PIPE_EQ] = ACTIONS(3435),
    [anon_sym_CARET_EQ] = ACTIONS(3435),
    [anon_sym_LT_LT_EQ] = ACTIONS(3435),
    [anon_sym_GT_GT_EQ] = ACTIONS(3435),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3435),
    [anon_sym_DOT_DOT] = ACTIONS(3437),
    [anon_sym_STAR_STAR] = ACTIONS(3437),
    [anon_sym_PIPE_PIPE] = ACTIONS(3437),
    [anon_sym_SLASH_SLASH] = ACTIONS(3437),
    [anon_sym_CARET_CARET] = ACTIONS(3435),
    [anon_sym_AMP_AMP] = ACTIONS(3437),
    [anon_sym_PIPE] = ACTIONS(3437),
    [anon_sym_CARET] = ACTIONS(3437),
    [anon_sym_AMP] = ACTIONS(3437),
    [anon_sym_GT_GT2] = ACTIONS(3437),
    [anon_sym_PLUS] = ACTIONS(3437),
    [anon_sym_DASH] = ACTIONS(3437),
    [anon_sym_DOT] = ACTIONS(3437),
    [anon_sym_SLASH] = ACTIONS(3437),
    [anon_sym_x] = ACTIONS(3437),
    [anon_sym_EQ_TILDE] = ACTIONS(3435),
    [anon_sym_BANG_TILDE] = ACTIONS(3435),
    [anon_sym_EQ_EQ] = ACTIONS(3437),
    [anon_sym_BANG_EQ] = ACTIONS(3435),
    [anon_sym_eq] = ACTIONS(3437),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3435),
    [anon_sym_equ] = ACTIONS(3437),
    [anon_sym_eqr] = ACTIONS(3437),
    [anon_sym_ne] = ACTIONS(3435),
    [anon_sym_LT_EQ_GT] = ACTIONS(3435),
    [anon_sym_cmp] = ACTIONS(3437),
    [anon_sym_TILDE_TILDE] = ACTIONS(3435),
    [anon_sym_LT] = ACTIONS(3437),
    [anon_sym_LT_EQ] = ACTIONS(3437),
    [anon_sym_GT_EQ] = ACTIONS(3435),
    [anon_sym_lt] = ACTIONS(3437),
    [anon_sym_le] = ACTIONS(3437),
    [anon_sym_ge] = ACTIONS(3437),
    [anon_sym_gt] = ACTIONS(3437),
    [anon_sym_isa] = ACTIONS(3437),
    [anon_sym_PLUS_PLUS] = ACTIONS(3435),
    [anon_sym_DASH_DASH] = ACTIONS(3435),
    [anon_sym_QMARK] = ACTIONS(3435),
    [anon_sym_EQ_GT] = ACTIONS(3435),
    [sym_comment] = ACTIONS(5),
    [sym_substitution_regexp_modifiers] = ACTIONS(3435),
    [sym__PERLY_SEMICOLON] = ACTIONS(3435),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3450)] = {
    [sym_heredoc_content] = STATE(3450),
    [anon_sym_COLON] = ACTIONS(4107),
    [anon_sym_DOLLAR] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_AT] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5139),
    [anon_sym_PLUS_EQ] = ACTIONS(5139),
    [anon_sym_DASH_EQ] = ACTIONS(5139),
    [anon_sym_DOT_EQ] = ACTIONS(5139),
    [anon_sym_STAR_EQ] = ACTIONS(5139),
    [anon_sym_SLASH_EQ] = ACTIONS(5139),
    [anon_sym_PERCENT_EQ] = ACTIONS(5137),
    [anon_sym_x_EQ] = ACTIONS(5139),
    [anon_sym_AMP_EQ] = ACTIONS(5139),
    [anon_sym_PIPE_EQ] = ACTIONS(5139),
    [anon_sym_CARET_EQ] = ACTIONS(5139),
    [anon_sym_LT_LT_EQ] = ACTIONS(5139),
    [anon_sym_GT_GT_EQ] = ACTIONS(5139),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5139),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3451)] = {
    [sym_heredoc_content] = STATE(3451),
    [anon_sym_RBRACE] = ACTIONS(4037),
    [anon_sym_COLON] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_RPAREN] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3452)] = {
    [sym_heredoc_content] = STATE(3452),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3453)] = {
    [sym_heredoc_content] = STATE(3453),
    [anon_sym_RBRACE] = ACTIONS(3857),
    [anon_sym_COLON] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_RPAREN] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3454)] = {
    [sym_heredoc_content] = STATE(3454),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5317),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5319),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3455)] = {
    [sym_heredoc_content] = STATE(3455),
    [aux_sym__term_rightward_repeat1] = STATE(3455),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(5321),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(5321),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3456)] = {
    [sym_heredoc_content] = STATE(3456),
    [anon_sym_RBRACE] = ACTIONS(2858),
    [anon_sym_COLON] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_RPAREN] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3457)] = {
    [sym_heredoc_content] = STATE(3457),
    [anon_sym_RBRACE] = ACTIONS(3815),
    [anon_sym_COLON] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_RPAREN] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3458)] = {
    [sym_heredoc_content] = STATE(3458),
    [anon_sym_RBRACE] = ACTIONS(3775),
    [anon_sym_COLON] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_RPAREN] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3459)] = {
    [sym_heredoc_content] = STATE(3459),
    [anon_sym_RBRACE] = ACTIONS(3901),
    [anon_sym_COLON] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_RPAREN] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3460)] = {
    [sym_heredoc_content] = STATE(3460),
    [anon_sym_RBRACE] = ACTIONS(4119),
    [anon_sym_COLON] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_RPAREN] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3461)] = {
    [sym_heredoc_content] = STATE(3461),
    [anon_sym_RBRACE] = ACTIONS(4131),
    [anon_sym_COLON] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_RPAREN] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3462)] = {
    [sym_heredoc_content] = STATE(3462),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3463)] = {
    [sym_heredoc_content] = STATE(3463),
    [anon_sym_RBRACE] = ACTIONS(3881),
    [anon_sym_COLON] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_RPAREN] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3464)] = {
    [sym_heredoc_content] = STATE(3464),
    [anon_sym_RBRACE] = ACTIONS(4075),
    [anon_sym_COLON] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_RPAREN] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3465)] = {
    [sym_heredoc_content] = STATE(3465),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3466)] = {
    [sym_heredoc_content] = STATE(3466),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3467)] = {
    [sym_heredoc_content] = STATE(3467),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3468)] = {
    [sym_heredoc_content] = STATE(3468),
    [anon_sym_RBRACE] = ACTIONS(3885),
    [anon_sym_COLON] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_RPAREN] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3469)] = {
    [sym_heredoc_content] = STATE(3469),
    [anon_sym_RBRACE] = ACTIONS(3889),
    [anon_sym_COLON] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_RPAREN] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3470)] = {
    [sym_heredoc_content] = STATE(3470),
    [anon_sym_RBRACE] = ACTIONS(3973),
    [anon_sym_COLON] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_RPAREN] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3471)] = {
    [sym_heredoc_content] = STATE(3471),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3472)] = {
    [sym_heredoc_content] = STATE(3472),
    [anon_sym_RBRACE] = ACTIONS(4021),
    [anon_sym_COLON] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_RPAREN] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3473)] = {
    [sym_heredoc_content] = STATE(3473),
    [anon_sym_RBRACE] = ACTIONS(4033),
    [anon_sym_COLON] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_RPAREN] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3474)] = {
    [sym_heredoc_content] = STATE(3474),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3475)] = {
    [sym_heredoc_content] = STATE(3475),
    [anon_sym_RBRACE] = ACTIONS(2536),
    [anon_sym_COLON] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_RPAREN] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3476)] = {
    [sym_heredoc_content] = STATE(3476),
    [anon_sym_COLON] = ACTIONS(3665),
    [anon_sym_SEMI] = ACTIONS(3665),
    [anon_sym_EQ] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_COMMA] = ACTIONS(3665),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3665),
    [anon_sym_DASH_GT] = ACTIONS(3665),
    [anon_sym_STAR] = ACTIONS(3667),
    [anon_sym_PERCENT] = ACTIONS(3667),
    [anon_sym_GT] = ACTIONS(3667),
    [anon_sym_LT_LT] = ACTIONS(3667),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3665),
    [anon_sym_PLUS_EQ] = ACTIONS(3665),
    [anon_sym_DASH_EQ] = ACTIONS(3665),
    [anon_sym_DOT_EQ] = ACTIONS(3665),
    [anon_sym_STAR_EQ] = ACTIONS(3665),
    [anon_sym_SLASH_EQ] = ACTIONS(3665),
    [anon_sym_PERCENT_EQ] = ACTIONS(3665),
    [anon_sym_x_EQ] = ACTIONS(3665),
    [anon_sym_AMP_EQ] = ACTIONS(3665),
    [anon_sym_PIPE_EQ] = ACTIONS(3665),
    [anon_sym_CARET_EQ] = ACTIONS(3665),
    [anon_sym_LT_LT_EQ] = ACTIONS(3665),
    [anon_sym_GT_GT_EQ] = ACTIONS(3665),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3665),
    [anon_sym_DOT_DOT] = ACTIONS(3667),
    [anon_sym_STAR_STAR] = ACTIONS(3667),
    [anon_sym_PIPE_PIPE] = ACTIONS(3667),
    [anon_sym_SLASH_SLASH] = ACTIONS(3667),
    [anon_sym_CARET_CARET] = ACTIONS(3665),
    [anon_sym_AMP_AMP] = ACTIONS(3667),
    [anon_sym_PIPE] = ACTIONS(3667),
    [anon_sym_CARET] = ACTIONS(3667),
    [anon_sym_AMP] = ACTIONS(3667),
    [anon_sym_GT_GT2] = ACTIONS(3667),
    [anon_sym_PLUS] = ACTIONS(3667),
    [anon_sym_DASH] = ACTIONS(3667),
    [anon_sym_DOT] = ACTIONS(3667),
    [anon_sym_SLASH] = ACTIONS(3667),
    [anon_sym_x] = ACTIONS(3667),
    [anon_sym_EQ_TILDE] = ACTIONS(3665),
    [anon_sym_BANG_TILDE] = ACTIONS(3665),
    [anon_sym_EQ_EQ] = ACTIONS(3667),
    [anon_sym_BANG_EQ] = ACTIONS(3665),
    [anon_sym_eq] = ACTIONS(3667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3665),
    [anon_sym_equ] = ACTIONS(3665),
    [anon_sym_eqr] = ACTIONS(3665),
    [anon_sym_ne] = ACTIONS(3665),
    [anon_sym_LT_EQ_GT] = ACTIONS(3665),
    [anon_sym_cmp] = ACTIONS(3665),
    [anon_sym_TILDE_TILDE] = ACTIONS(3665),
    [anon_sym_LT] = ACTIONS(3667),
    [anon_sym_LT_EQ] = ACTIONS(3667),
    [anon_sym_GT_EQ] = ACTIONS(3665),
    [anon_sym_lt] = ACTIONS(3665),
    [anon_sym_le] = ACTIONS(3665),
    [anon_sym_ge] = ACTIONS(3665),
    [anon_sym_gt] = ACTIONS(3665),
    [anon_sym_isa] = ACTIONS(3665),
    [anon_sym_PLUS_PLUS] = ACTIONS(3665),
    [anon_sym_DASH_DASH] = ACTIONS(3665),
    [anon_sym_QMARK] = ACTIONS(3665),
    [anon_sym_EQ_GT] = ACTIONS(3665),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3665),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3477)] = {
    [sym_heredoc_content] = STATE(3477),
    [anon_sym_COLON] = ACTIONS(3597),
    [anon_sym_SEMI] = ACTIONS(3597),
    [anon_sym_EQ] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_COMMA] = ACTIONS(3597),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3597),
    [anon_sym_DASH_GT] = ACTIONS(3597),
    [anon_sym_STAR] = ACTIONS(3599),
    [anon_sym_PERCENT] = ACTIONS(3599),
    [anon_sym_GT] = ACTIONS(3599),
    [anon_sym_LT_LT] = ACTIONS(3599),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3597),
    [anon_sym_PLUS_EQ] = ACTIONS(3597),
    [anon_sym_DASH_EQ] = ACTIONS(3597),
    [anon_sym_DOT_EQ] = ACTIONS(3597),
    [anon_sym_STAR_EQ] = ACTIONS(3597),
    [anon_sym_SLASH_EQ] = ACTIONS(3597),
    [anon_sym_PERCENT_EQ] = ACTIONS(3597),
    [anon_sym_x_EQ] = ACTIONS(3597),
    [anon_sym_AMP_EQ] = ACTIONS(3597),
    [anon_sym_PIPE_EQ] = ACTIONS(3597),
    [anon_sym_CARET_EQ] = ACTIONS(3597),
    [anon_sym_LT_LT_EQ] = ACTIONS(3597),
    [anon_sym_GT_GT_EQ] = ACTIONS(3597),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3597),
    [anon_sym_DOT_DOT] = ACTIONS(3599),
    [anon_sym_STAR_STAR] = ACTIONS(3599),
    [anon_sym_PIPE_PIPE] = ACTIONS(3599),
    [anon_sym_SLASH_SLASH] = ACTIONS(3599),
    [anon_sym_CARET_CARET] = ACTIONS(3597),
    [anon_sym_AMP_AMP] = ACTIONS(3599),
    [anon_sym_PIPE] = ACTIONS(3599),
    [anon_sym_CARET] = ACTIONS(3599),
    [anon_sym_AMP] = ACTIONS(3599),
    [anon_sym_GT_GT2] = ACTIONS(3599),
    [anon_sym_PLUS] = ACTIONS(3599),
    [anon_sym_DASH] = ACTIONS(3599),
    [anon_sym_DOT] = ACTIONS(3599),
    [anon_sym_SLASH] = ACTIONS(3599),
    [anon_sym_x] = ACTIONS(3599),
    [anon_sym_EQ_TILDE] = ACTIONS(3597),
    [anon_sym_BANG_TILDE] = ACTIONS(3597),
    [anon_sym_EQ_EQ] = ACTIONS(3599),
    [anon_sym_BANG_EQ] = ACTIONS(3597),
    [anon_sym_eq] = ACTIONS(3599),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3597),
    [anon_sym_equ] = ACTIONS(3597),
    [anon_sym_eqr] = ACTIONS(3597),
    [anon_sym_ne] = ACTIONS(3597),
    [anon_sym_LT_EQ_GT] = ACTIONS(3597),
    [anon_sym_cmp] = ACTIONS(3597),
    [anon_sym_TILDE_TILDE] = ACTIONS(3597),
    [anon_sym_LT] = ACTIONS(3599),
    [anon_sym_LT_EQ] = ACTIONS(3599),
    [anon_sym_GT_EQ] = ACTIONS(3597),
    [anon_sym_lt] = ACTIONS(3597),
    [anon_sym_le] = ACTIONS(3597),
    [anon_sym_ge] = ACTIONS(3597),
    [anon_sym_gt] = ACTIONS(3597),
    [anon_sym_isa] = ACTIONS(3597),
    [anon_sym_PLUS_PLUS] = ACTIONS(3597),
    [anon_sym_DASH_DASH] = ACTIONS(3597),
    [anon_sym_QMARK] = ACTIONS(3597),
    [anon_sym_EQ_GT] = ACTIONS(3597),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3597),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3478)] = {
    [sym_heredoc_content] = STATE(3478),
    [anon_sym_COLON] = ACTIONS(3611),
    [anon_sym_SEMI] = ACTIONS(3611),
    [anon_sym_EQ] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_COMMA] = ACTIONS(3611),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3611),
    [anon_sym_DASH_GT] = ACTIONS(3611),
    [anon_sym_STAR] = ACTIONS(3613),
    [anon_sym_PERCENT] = ACTIONS(3613),
    [anon_sym_GT] = ACTIONS(3613),
    [anon_sym_LT_LT] = ACTIONS(3613),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3611),
    [anon_sym_PLUS_EQ] = ACTIONS(3611),
    [anon_sym_DASH_EQ] = ACTIONS(3611),
    [anon_sym_DOT_EQ] = ACTIONS(3611),
    [anon_sym_STAR_EQ] = ACTIONS(3611),
    [anon_sym_SLASH_EQ] = ACTIONS(3611),
    [anon_sym_PERCENT_EQ] = ACTIONS(3611),
    [anon_sym_x_EQ] = ACTIONS(3611),
    [anon_sym_AMP_EQ] = ACTIONS(3611),
    [anon_sym_PIPE_EQ] = ACTIONS(3611),
    [anon_sym_CARET_EQ] = ACTIONS(3611),
    [anon_sym_LT_LT_EQ] = ACTIONS(3611),
    [anon_sym_GT_GT_EQ] = ACTIONS(3611),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3611),
    [anon_sym_DOT_DOT] = ACTIONS(3613),
    [anon_sym_STAR_STAR] = ACTIONS(3613),
    [anon_sym_PIPE_PIPE] = ACTIONS(3613),
    [anon_sym_SLASH_SLASH] = ACTIONS(3613),
    [anon_sym_CARET_CARET] = ACTIONS(3611),
    [anon_sym_AMP_AMP] = ACTIONS(3613),
    [anon_sym_PIPE] = ACTIONS(3613),
    [anon_sym_CARET] = ACTIONS(3613),
    [anon_sym_AMP] = ACTIONS(3613),
    [anon_sym_GT_GT2] = ACTIONS(3613),
    [anon_sym_PLUS] = ACTIONS(3613),
    [anon_sym_DASH] = ACTIONS(3613),
    [anon_sym_DOT] = ACTIONS(3613),
    [anon_sym_SLASH] = ACTIONS(3613),
    [anon_sym_x] = ACTIONS(3613),
    [anon_sym_EQ_TILDE] = ACTIONS(3611),
    [anon_sym_BANG_TILDE] = ACTIONS(3611),
    [anon_sym_EQ_EQ] = ACTIONS(3613),
    [anon_sym_BANG_EQ] = ACTIONS(3611),
    [anon_sym_eq] = ACTIONS(3613),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3611),
    [anon_sym_equ] = ACTIONS(3611),
    [anon_sym_eqr] = ACTIONS(3611),
    [anon_sym_ne] = ACTIONS(3611),
    [anon_sym_LT_EQ_GT] = ACTIONS(3611),
    [anon_sym_cmp] = ACTIONS(3611),
    [anon_sym_TILDE_TILDE] = ACTIONS(3611),
    [anon_sym_LT] = ACTIONS(3613),
    [anon_sym_LT_EQ] = ACTIONS(3613),
    [anon_sym_GT_EQ] = ACTIONS(3611),
    [anon_sym_lt] = ACTIONS(3611),
    [anon_sym_le] = ACTIONS(3611),
    [anon_sym_ge] = ACTIONS(3611),
    [anon_sym_gt] = ACTIONS(3611),
    [anon_sym_isa] = ACTIONS(3611),
    [anon_sym_PLUS_PLUS] = ACTIONS(3611),
    [anon_sym_DASH_DASH] = ACTIONS(3611),
    [anon_sym_QMARK] = ACTIONS(3611),
    [anon_sym_EQ_GT] = ACTIONS(3611),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3611),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3479)] = {
    [sym_heredoc_content] = STATE(3479),
    [anon_sym_RBRACE] = ACTIONS(3679),
    [anon_sym_COLON] = ACTIONS(3679),
    [anon_sym_EQ] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_COMMA] = ACTIONS(3679),
    [anon_sym_RPAREN] = ACTIONS(3679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3679),
    [anon_sym_DASH_GT] = ACTIONS(3679),
    [anon_sym_STAR] = ACTIONS(3681),
    [anon_sym_PERCENT] = ACTIONS(3681),
    [anon_sym_GT] = ACTIONS(3681),
    [anon_sym_LT_LT] = ACTIONS(3681),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3679),
    [anon_sym_PLUS_EQ] = ACTIONS(3679),
    [anon_sym_DASH_EQ] = ACTIONS(3679),
    [anon_sym_DOT_EQ] = ACTIONS(3679),
    [anon_sym_STAR_EQ] = ACTIONS(3679),
    [anon_sym_SLASH_EQ] = ACTIONS(3679),
    [anon_sym_PERCENT_EQ] = ACTIONS(3679),
    [anon_sym_x_EQ] = ACTIONS(3679),
    [anon_sym_AMP_EQ] = ACTIONS(3679),
    [anon_sym_PIPE_EQ] = ACTIONS(3679),
    [anon_sym_CARET_EQ] = ACTIONS(3679),
    [anon_sym_LT_LT_EQ] = ACTIONS(3679),
    [anon_sym_GT_GT_EQ] = ACTIONS(3679),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3679),
    [anon_sym_DOT_DOT] = ACTIONS(3681),
    [anon_sym_STAR_STAR] = ACTIONS(3681),
    [anon_sym_PIPE_PIPE] = ACTIONS(3681),
    [anon_sym_SLASH_SLASH] = ACTIONS(3681),
    [anon_sym_CARET_CARET] = ACTIONS(3679),
    [anon_sym_AMP_AMP] = ACTIONS(3681),
    [anon_sym_PIPE] = ACTIONS(3681),
    [anon_sym_CARET] = ACTIONS(3681),
    [anon_sym_AMP] = ACTIONS(3681),
    [anon_sym_GT_GT2] = ACTIONS(3681),
    [anon_sym_PLUS] = ACTIONS(3681),
    [anon_sym_DASH] = ACTIONS(3681),
    [anon_sym_DOT] = ACTIONS(3681),
    [anon_sym_SLASH] = ACTIONS(3681),
    [anon_sym_x] = ACTIONS(3681),
    [anon_sym_EQ_TILDE] = ACTIONS(3679),
    [anon_sym_BANG_TILDE] = ACTIONS(3679),
    [anon_sym_EQ_EQ] = ACTIONS(3681),
    [anon_sym_BANG_EQ] = ACTIONS(3679),
    [anon_sym_eq] = ACTIONS(3681),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3679),
    [anon_sym_equ] = ACTIONS(3679),
    [anon_sym_eqr] = ACTIONS(3679),
    [anon_sym_ne] = ACTIONS(3679),
    [anon_sym_LT_EQ_GT] = ACTIONS(3679),
    [anon_sym_cmp] = ACTIONS(3679),
    [anon_sym_TILDE_TILDE] = ACTIONS(3679),
    [anon_sym_LT] = ACTIONS(3681),
    [anon_sym_LT_EQ] = ACTIONS(3681),
    [anon_sym_GT_EQ] = ACTIONS(3679),
    [anon_sym_lt] = ACTIONS(3679),
    [anon_sym_le] = ACTIONS(3679),
    [anon_sym_ge] = ACTIONS(3679),
    [anon_sym_gt] = ACTIONS(3679),
    [anon_sym_isa] = ACTIONS(3679),
    [anon_sym_PLUS_PLUS] = ACTIONS(3679),
    [anon_sym_DASH_DASH] = ACTIONS(3679),
    [anon_sym_QMARK] = ACTIONS(3679),
    [anon_sym_EQ_GT] = ACTIONS(3679),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3480)] = {
    [sym_heredoc_content] = STATE(3480),
    [anon_sym_COLON] = ACTIONS(3447),
    [anon_sym_SEMI] = ACTIONS(3447),
    [anon_sym_EQ] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_COMMA] = ACTIONS(3447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3447),
    [anon_sym_DASH_GT] = ACTIONS(3447),
    [anon_sym_STAR] = ACTIONS(3449),
    [anon_sym_PERCENT] = ACTIONS(3449),
    [anon_sym_GT] = ACTIONS(3449),
    [anon_sym_LT_LT] = ACTIONS(3449),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3447),
    [anon_sym_PLUS_EQ] = ACTIONS(3447),
    [anon_sym_DASH_EQ] = ACTIONS(3447),
    [anon_sym_DOT_EQ] = ACTIONS(3447),
    [anon_sym_STAR_EQ] = ACTIONS(3447),
    [anon_sym_SLASH_EQ] = ACTIONS(3447),
    [anon_sym_PERCENT_EQ] = ACTIONS(3447),
    [anon_sym_x_EQ] = ACTIONS(3447),
    [anon_sym_AMP_EQ] = ACTIONS(3447),
    [anon_sym_PIPE_EQ] = ACTIONS(3447),
    [anon_sym_CARET_EQ] = ACTIONS(3447),
    [anon_sym_LT_LT_EQ] = ACTIONS(3447),
    [anon_sym_GT_GT_EQ] = ACTIONS(3447),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3447),
    [anon_sym_DOT_DOT] = ACTIONS(3449),
    [anon_sym_STAR_STAR] = ACTIONS(3449),
    [anon_sym_PIPE_PIPE] = ACTIONS(3449),
    [anon_sym_SLASH_SLASH] = ACTIONS(3449),
    [anon_sym_CARET_CARET] = ACTIONS(3447),
    [anon_sym_AMP_AMP] = ACTIONS(3449),
    [anon_sym_PIPE] = ACTIONS(3449),
    [anon_sym_CARET] = ACTIONS(3449),
    [anon_sym_AMP] = ACTIONS(3449),
    [anon_sym_GT_GT2] = ACTIONS(3449),
    [anon_sym_PLUS] = ACTIONS(3449),
    [anon_sym_DASH] = ACTIONS(3449),
    [anon_sym_DOT] = ACTIONS(3449),
    [anon_sym_SLASH] = ACTIONS(3449),
    [anon_sym_x] = ACTIONS(3449),
    [anon_sym_EQ_TILDE] = ACTIONS(3447),
    [anon_sym_BANG_TILDE] = ACTIONS(3447),
    [anon_sym_EQ_EQ] = ACTIONS(3449),
    [anon_sym_BANG_EQ] = ACTIONS(3447),
    [anon_sym_eq] = ACTIONS(3449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3447),
    [anon_sym_equ] = ACTIONS(3447),
    [anon_sym_eqr] = ACTIONS(3447),
    [anon_sym_ne] = ACTIONS(3447),
    [anon_sym_LT_EQ_GT] = ACTIONS(3447),
    [anon_sym_cmp] = ACTIONS(3447),
    [anon_sym_TILDE_TILDE] = ACTIONS(3447),
    [anon_sym_LT] = ACTIONS(3449),
    [anon_sym_LT_EQ] = ACTIONS(3449),
    [anon_sym_GT_EQ] = ACTIONS(3447),
    [anon_sym_lt] = ACTIONS(3447),
    [anon_sym_le] = ACTIONS(3447),
    [anon_sym_ge] = ACTIONS(3447),
    [anon_sym_gt] = ACTIONS(3447),
    [anon_sym_isa] = ACTIONS(3447),
    [anon_sym_PLUS_PLUS] = ACTIONS(3447),
    [anon_sym_DASH_DASH] = ACTIONS(3447),
    [anon_sym_QMARK] = ACTIONS(3447),
    [anon_sym_EQ_GT] = ACTIONS(3447),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3447),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3481)] = {
    [sym_heredoc_content] = STATE(3481),
    [anon_sym_RBRACE] = ACTIONS(3747),
    [anon_sym_COLON] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_RPAREN] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3482)] = {
    [sym_heredoc_content] = STATE(3482),
    [anon_sym_RBRACE] = ACTIONS(3779),
    [anon_sym_COLON] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_RPAREN] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3483)] = {
    [sym_heredoc_content] = STATE(3483),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3484)] = {
    [sym_heredoc_content] = STATE(3484),
    [anon_sym_RBRACE] = ACTIONS(3833),
    [anon_sym_COLON] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_RPAREN] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3485)] = {
    [sym_heredoc_content] = STATE(3485),
    [anon_sym_RBRACE] = ACTIONS(3841),
    [anon_sym_COLON] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_RPAREN] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3486)] = {
    [sym_heredoc_content] = STATE(3486),
    [anon_sym_RBRACE] = ACTIONS(3853),
    [anon_sym_COLON] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_RPAREN] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3487)] = {
    [sym_heredoc_content] = STATE(3487),
    [anon_sym_RBRACE] = ACTIONS(3861),
    [anon_sym_COLON] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_RPAREN] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3488)] = {
    [sym_heredoc_content] = STATE(3488),
    [anon_sym_RBRACE] = ACTIONS(3837),
    [anon_sym_COLON] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_RPAREN] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3489)] = {
    [sym_heredoc_content] = STATE(3489),
    [anon_sym_RBRACE] = ACTIONS(3993),
    [anon_sym_COLON] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_RPAREN] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3490)] = {
    [sym_heredoc_content] = STATE(3490),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3491)] = {
    [sym_heredoc_content] = STATE(3491),
    [anon_sym_RBRACE] = ACTIONS(4005),
    [anon_sym_COLON] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_RPAREN] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3492)] = {
    [sym_heredoc_content] = STATE(3492),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3493)] = {
    [sym_heredoc_content] = STATE(3493),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3494)] = {
    [sym_heredoc_content] = STATE(3494),
    [anon_sym_RBRACE] = ACTIONS(4087),
    [anon_sym_COLON] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_RPAREN] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3495)] = {
    [sym_heredoc_content] = STATE(3495),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_DOLLAR] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_AT] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3721),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_PERCENT2] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3496)] = {
    [sym_heredoc_content] = STATE(3496),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3497)] = {
    [sym_heredoc_content] = STATE(3497),
    [anon_sym_RBRACE] = ACTIONS(3771),
    [anon_sym_COLON] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_RPAREN] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3498)] = {
    [sym_heredoc_content] = STATE(3498),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_DOLLAR] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_AT] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3765),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5324),
    [anon_sym_cmp] = ACTIONS(5324),
    [anon_sym_TILDE_TILDE] = ACTIONS(5324),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_PERCENT2] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3499)] = {
    [sym_heredoc_content] = STATE(3499),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_DOLLAR] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_AT] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3737),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5326),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_PERCENT2] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3500)] = {
    [sym_heredoc_content] = STATE(3500),
    [anon_sym_RBRACE] = ACTIONS(3743),
    [anon_sym_COLON] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_RPAREN] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3501)] = {
    [sym_heredoc_content] = STATE(3501),
    [anon_sym_RBRACE] = ACTIONS(3953),
    [anon_sym_COLON] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_RPAREN] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3502)] = {
    [sym_heredoc_content] = STATE(3502),
    [anon_sym_RBRACE] = ACTIONS(3965),
    [anon_sym_COLON] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_RPAREN] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3503)] = {
    [sym_heredoc_content] = STATE(3503),
    [anon_sym_RBRACE] = ACTIONS(3811),
    [anon_sym_COLON] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_RPAREN] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3504)] = {
    [sym_heredoc_content] = STATE(3504),
    [anon_sym_RBRACE] = ACTIONS(4155),
    [anon_sym_COLON] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_RPAREN] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3505)] = {
    [sym_heredoc_content] = STATE(3505),
    [anon_sym_COLON] = ACTIONS(3985),
    [anon_sym_DOLLAR] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_AT] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5141),
    [anon_sym_DASH_GT] = ACTIONS(4633),
    [anon_sym_STAR] = ACTIONS(5143),
    [anon_sym_PERCENT] = ACTIONS(5143),
    [anon_sym_GT] = ACTIONS(5145),
    [anon_sym_LT_LT] = ACTIONS(5147),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3987),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5149),
    [anon_sym_STAR_STAR] = ACTIONS(5151),
    [anon_sym_PIPE_PIPE] = ACTIONS(5153),
    [anon_sym_SLASH_SLASH] = ACTIONS(5153),
    [anon_sym_CARET_CARET] = ACTIONS(5155),
    [anon_sym_AMP_AMP] = ACTIONS(5157),
    [anon_sym_PIPE] = ACTIONS(5159),
    [anon_sym_CARET] = ACTIONS(5159),
    [anon_sym_AMP] = ACTIONS(5161),
    [anon_sym_GT_GT2] = ACTIONS(5147),
    [anon_sym_PLUS] = ACTIONS(5163),
    [anon_sym_DASH] = ACTIONS(5163),
    [anon_sym_DOT] = ACTIONS(5163),
    [anon_sym_SLASH] = ACTIONS(5143),
    [anon_sym_x] = ACTIONS(5143),
    [anon_sym_EQ_TILDE] = ACTIONS(5165),
    [anon_sym_BANG_TILDE] = ACTIONS(5165),
    [anon_sym_EQ_EQ] = ACTIONS(5167),
    [anon_sym_BANG_EQ] = ACTIONS(5169),
    [anon_sym_eq] = ACTIONS(5167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5169),
    [anon_sym_equ] = ACTIONS(5169),
    [anon_sym_eqr] = ACTIONS(5169),
    [anon_sym_ne] = ACTIONS(5169),
    [anon_sym_LT_EQ_GT] = ACTIONS(5171),
    [anon_sym_cmp] = ACTIONS(5171),
    [anon_sym_TILDE_TILDE] = ACTIONS(5171),
    [anon_sym_LT] = ACTIONS(5145),
    [anon_sym_LT_EQ] = ACTIONS(5145),
    [anon_sym_GT_EQ] = ACTIONS(5173),
    [anon_sym_lt] = ACTIONS(5173),
    [anon_sym_le] = ACTIONS(5173),
    [anon_sym_ge] = ACTIONS(5173),
    [anon_sym_gt] = ACTIONS(5173),
    [anon_sym_isa] = ACTIONS(5175),
    [anon_sym_PLUS_PLUS] = ACTIONS(4669),
    [anon_sym_DASH_DASH] = ACTIONS(4669),
    [anon_sym_QMARK] = ACTIONS(5177),
    [anon_sym_PERCENT2] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3506)] = {
    [sym_heredoc_content] = STATE(3506),
    [anon_sym_RBRACE] = ACTIONS(4176),
    [anon_sym_COLON] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_RPAREN] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3507)] = {
    [sym_heredoc_content] = STATE(3507),
    [anon_sym_RBRACE] = ACTIONS(3873),
    [anon_sym_COLON] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_RPAREN] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3508)] = {
    [sym_heredoc_content] = STATE(3508),
    [anon_sym_RBRACE] = ACTIONS(3981),
    [anon_sym_COLON] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_RPAREN] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3509)] = {
    [sym_heredoc_content] = STATE(3509),
    [anon_sym_RBRACE] = ACTIONS(4001),
    [anon_sym_COLON] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_RPAREN] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3510)] = {
    [sym_heredoc_content] = STATE(3510),
    [anon_sym_RBRACE] = ACTIONS(4009),
    [anon_sym_COLON] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_RPAREN] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3511)] = {
    [sym_heredoc_content] = STATE(3511),
    [anon_sym_RBRACE] = ACTIONS(4013),
    [anon_sym_COLON] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_RPAREN] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3512)] = {
    [sym_heredoc_content] = STATE(3512),
    [anon_sym_RBRACE] = ACTIONS(4025),
    [anon_sym_COLON] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_RPAREN] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3513)] = {
    [sym_heredoc_content] = STATE(3513),
    [anon_sym_RBRACE] = ACTIONS(3783),
    [anon_sym_COLON] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_RPAREN] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3514)] = {
    [sym_heredoc_content] = STATE(3514),
    [anon_sym_RBRACE] = ACTIONS(3485),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3515)] = {
    [sym_heredoc_content] = STATE(3515),
    [anon_sym_SEMI] = ACTIONS(4111),
    [anon_sym_EQ] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_COMMA] = ACTIONS(4111),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4111),
    [anon_sym_DASH_GT] = ACTIONS(4111),
    [anon_sym_STAR] = ACTIONS(4113),
    [anon_sym_PERCENT] = ACTIONS(4113),
    [anon_sym_GT] = ACTIONS(4113),
    [anon_sym_LT_LT] = ACTIONS(4113),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4111),
    [anon_sym_PLUS_EQ] = ACTIONS(4111),
    [anon_sym_DASH_EQ] = ACTIONS(4111),
    [anon_sym_DOT_EQ] = ACTIONS(4111),
    [anon_sym_STAR_EQ] = ACTIONS(4111),
    [anon_sym_SLASH_EQ] = ACTIONS(4111),
    [anon_sym_PERCENT_EQ] = ACTIONS(4111),
    [anon_sym_x_EQ] = ACTIONS(4111),
    [anon_sym_AMP_EQ] = ACTIONS(4111),
    [anon_sym_PIPE_EQ] = ACTIONS(4111),
    [anon_sym_CARET_EQ] = ACTIONS(4111),
    [anon_sym_LT_LT_EQ] = ACTIONS(4111),
    [anon_sym_GT_GT_EQ] = ACTIONS(4111),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4111),
    [anon_sym_DOT_DOT] = ACTIONS(4113),
    [anon_sym_STAR_STAR] = ACTIONS(4113),
    [anon_sym_PIPE_PIPE] = ACTIONS(4113),
    [anon_sym_SLASH_SLASH] = ACTIONS(4113),
    [anon_sym_CARET_CARET] = ACTIONS(4111),
    [anon_sym_AMP_AMP] = ACTIONS(4113),
    [anon_sym_PIPE] = ACTIONS(4113),
    [anon_sym_CARET] = ACTIONS(4113),
    [anon_sym_AMP] = ACTIONS(4113),
    [anon_sym_GT_GT2] = ACTIONS(4113),
    [anon_sym_PLUS] = ACTIONS(4113),
    [anon_sym_DASH] = ACTIONS(4113),
    [anon_sym_DOT] = ACTIONS(4113),
    [anon_sym_SLASH] = ACTIONS(4113),
    [anon_sym_x] = ACTIONS(4113),
    [anon_sym_EQ_TILDE] = ACTIONS(4111),
    [anon_sym_BANG_TILDE] = ACTIONS(4111),
    [anon_sym_EQ_EQ] = ACTIONS(4113),
    [anon_sym_BANG_EQ] = ACTIONS(4111),
    [anon_sym_eq] = ACTIONS(4113),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4111),
    [anon_sym_equ] = ACTIONS(4111),
    [anon_sym_eqr] = ACTIONS(4111),
    [anon_sym_ne] = ACTIONS(4111),
    [anon_sym_LT_EQ_GT] = ACTIONS(4111),
    [anon_sym_cmp] = ACTIONS(4111),
    [anon_sym_TILDE_TILDE] = ACTIONS(4111),
    [anon_sym_LT] = ACTIONS(4113),
    [anon_sym_LT_EQ] = ACTIONS(4113),
    [anon_sym_GT_EQ] = ACTIONS(4111),
    [anon_sym_lt] = ACTIONS(4111),
    [anon_sym_le] = ACTIONS(4111),
    [anon_sym_ge] = ACTIONS(4111),
    [anon_sym_gt] = ACTIONS(4111),
    [anon_sym_isa] = ACTIONS(4111),
    [anon_sym_PLUS_PLUS] = ACTIONS(4111),
    [anon_sym_DASH_DASH] = ACTIONS(4111),
    [anon_sym_QMARK] = ACTIONS(4111),
    [anon_sym_EQ_GT] = ACTIONS(4111),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4111),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3516)] = {
    [sym_heredoc_content] = STATE(3516),
    [anon_sym_RBRACE] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3517)] = {
    [sym_heredoc_content] = STATE(3517),
    [anon_sym_RBRACE] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3518)] = {
    [sym_heredoc_content] = STATE(3518),
    [anon_sym_RBRACE] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3519)] = {
    [sym_heredoc_content] = STATE(3519),
    [anon_sym_RBRACE] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3520)] = {
    [sym_heredoc_content] = STATE(3520),
    [anon_sym_RBRACE] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3521)] = {
    [sym_heredoc_content] = STATE(3521),
    [anon_sym_RBRACE] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3522)] = {
    [sym_heredoc_content] = STATE(3522),
    [anon_sym_RBRACE] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3523)] = {
    [sym_heredoc_content] = STATE(3523),
    [anon_sym_RBRACE] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3524)] = {
    [sym_heredoc_content] = STATE(3524),
    [anon_sym_RBRACE] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3525)] = {
    [sym_heredoc_content] = STATE(3525),
    [aux_sym__term_rightward_repeat1] = STATE(3624),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(5332),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(5332),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3526)] = {
    [sym_heredoc_content] = STATE(3526),
    [anon_sym_RBRACE] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3527)] = {
    [sym_heredoc_content] = STATE(3527),
    [anon_sym_RBRACE] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3528)] = {
    [sym_heredoc_content] = STATE(3528),
    [anon_sym_RBRACE] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5372),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5372),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3529)] = {
    [sym_heredoc_content] = STATE(3529),
    [anon_sym_SEMI] = ACTIONS(3731),
    [anon_sym_EQ] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_COMMA] = ACTIONS(3731),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3731),
    [anon_sym_DASH_GT] = ACTIONS(3731),
    [anon_sym_STAR] = ACTIONS(3733),
    [anon_sym_PERCENT] = ACTIONS(3733),
    [anon_sym_GT] = ACTIONS(3733),
    [anon_sym_LT_LT] = ACTIONS(3733),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3731),
    [anon_sym_PLUS_EQ] = ACTIONS(3731),
    [anon_sym_DASH_EQ] = ACTIONS(3731),
    [anon_sym_DOT_EQ] = ACTIONS(3731),
    [anon_sym_STAR_EQ] = ACTIONS(3731),
    [anon_sym_SLASH_EQ] = ACTIONS(3731),
    [anon_sym_PERCENT_EQ] = ACTIONS(3731),
    [anon_sym_x_EQ] = ACTIONS(3731),
    [anon_sym_AMP_EQ] = ACTIONS(3731),
    [anon_sym_PIPE_EQ] = ACTIONS(3731),
    [anon_sym_CARET_EQ] = ACTIONS(3731),
    [anon_sym_LT_LT_EQ] = ACTIONS(3731),
    [anon_sym_GT_GT_EQ] = ACTIONS(3731),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3731),
    [anon_sym_DOT_DOT] = ACTIONS(3733),
    [anon_sym_STAR_STAR] = ACTIONS(3733),
    [anon_sym_PIPE_PIPE] = ACTIONS(3733),
    [anon_sym_SLASH_SLASH] = ACTIONS(3733),
    [anon_sym_CARET_CARET] = ACTIONS(3731),
    [anon_sym_AMP_AMP] = ACTIONS(3733),
    [anon_sym_PIPE] = ACTIONS(3733),
    [anon_sym_CARET] = ACTIONS(3733),
    [anon_sym_AMP] = ACTIONS(3733),
    [anon_sym_GT_GT2] = ACTIONS(3733),
    [anon_sym_PLUS] = ACTIONS(3733),
    [anon_sym_DASH] = ACTIONS(3733),
    [anon_sym_DOT] = ACTIONS(3733),
    [anon_sym_SLASH] = ACTIONS(3733),
    [anon_sym_x] = ACTIONS(3733),
    [anon_sym_EQ_TILDE] = ACTIONS(3731),
    [anon_sym_BANG_TILDE] = ACTIONS(3731),
    [anon_sym_EQ_EQ] = ACTIONS(3733),
    [anon_sym_BANG_EQ] = ACTIONS(3731),
    [anon_sym_eq] = ACTIONS(3733),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3731),
    [anon_sym_equ] = ACTIONS(3731),
    [anon_sym_eqr] = ACTIONS(3731),
    [anon_sym_ne] = ACTIONS(3731),
    [anon_sym_LT_EQ_GT] = ACTIONS(3731),
    [anon_sym_cmp] = ACTIONS(3731),
    [anon_sym_TILDE_TILDE] = ACTIONS(3731),
    [anon_sym_LT] = ACTIONS(3733),
    [anon_sym_LT_EQ] = ACTIONS(3733),
    [anon_sym_GT_EQ] = ACTIONS(3731),
    [anon_sym_lt] = ACTIONS(3731),
    [anon_sym_le] = ACTIONS(3731),
    [anon_sym_ge] = ACTIONS(3731),
    [anon_sym_gt] = ACTIONS(3731),
    [anon_sym_isa] = ACTIONS(3731),
    [anon_sym_PLUS_PLUS] = ACTIONS(3731),
    [anon_sym_DASH_DASH] = ACTIONS(3731),
    [anon_sym_QMARK] = ACTIONS(3731),
    [anon_sym_EQ_GT] = ACTIONS(3731),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3731),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3530)] = {
    [sym_heredoc_content] = STATE(3530),
    [anon_sym_SEMI] = ACTIONS(4045),
    [anon_sym_EQ] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_COMMA] = ACTIONS(4045),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4045),
    [anon_sym_DASH_GT] = ACTIONS(4045),
    [anon_sym_STAR] = ACTIONS(2856),
    [anon_sym_PERCENT] = ACTIONS(2856),
    [anon_sym_GT] = ACTIONS(2856),
    [anon_sym_LT_LT] = ACTIONS(2856),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4045),
    [anon_sym_PLUS_EQ] = ACTIONS(4045),
    [anon_sym_DASH_EQ] = ACTIONS(4045),
    [anon_sym_DOT_EQ] = ACTIONS(4045),
    [anon_sym_STAR_EQ] = ACTIONS(4045),
    [anon_sym_SLASH_EQ] = ACTIONS(4045),
    [anon_sym_PERCENT_EQ] = ACTIONS(4045),
    [anon_sym_x_EQ] = ACTIONS(4045),
    [anon_sym_AMP_EQ] = ACTIONS(4045),
    [anon_sym_PIPE_EQ] = ACTIONS(4045),
    [anon_sym_CARET_EQ] = ACTIONS(4045),
    [anon_sym_LT_LT_EQ] = ACTIONS(4045),
    [anon_sym_GT_GT_EQ] = ACTIONS(4045),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4045),
    [anon_sym_DOT_DOT] = ACTIONS(2856),
    [anon_sym_STAR_STAR] = ACTIONS(2856),
    [anon_sym_PIPE_PIPE] = ACTIONS(2856),
    [anon_sym_SLASH_SLASH] = ACTIONS(2856),
    [anon_sym_CARET_CARET] = ACTIONS(4045),
    [anon_sym_AMP_AMP] = ACTIONS(2856),
    [anon_sym_PIPE] = ACTIONS(2856),
    [anon_sym_CARET] = ACTIONS(2856),
    [anon_sym_AMP] = ACTIONS(2856),
    [anon_sym_GT_GT2] = ACTIONS(2856),
    [anon_sym_PLUS] = ACTIONS(2856),
    [anon_sym_DASH] = ACTIONS(2856),
    [anon_sym_DOT] = ACTIONS(2856),
    [anon_sym_SLASH] = ACTIONS(2856),
    [anon_sym_x] = ACTIONS(2856),
    [anon_sym_EQ_TILDE] = ACTIONS(4045),
    [anon_sym_BANG_TILDE] = ACTIONS(4045),
    [anon_sym_EQ_EQ] = ACTIONS(2856),
    [anon_sym_BANG_EQ] = ACTIONS(4045),
    [anon_sym_eq] = ACTIONS(2856),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4045),
    [anon_sym_equ] = ACTIONS(4045),
    [anon_sym_eqr] = ACTIONS(4045),
    [anon_sym_ne] = ACTIONS(4045),
    [anon_sym_LT_EQ_GT] = ACTIONS(4045),
    [anon_sym_cmp] = ACTIONS(4045),
    [anon_sym_TILDE_TILDE] = ACTIONS(4045),
    [anon_sym_LT] = ACTIONS(2856),
    [anon_sym_LT_EQ] = ACTIONS(2856),
    [anon_sym_GT_EQ] = ACTIONS(4045),
    [anon_sym_lt] = ACTIONS(4045),
    [anon_sym_le] = ACTIONS(4045),
    [anon_sym_ge] = ACTIONS(4045),
    [anon_sym_gt] = ACTIONS(4045),
    [anon_sym_isa] = ACTIONS(4045),
    [anon_sym_PLUS_PLUS] = ACTIONS(4045),
    [anon_sym_DASH_DASH] = ACTIONS(4045),
    [anon_sym_QMARK] = ACTIONS(4045),
    [anon_sym_EQ_GT] = ACTIONS(4045),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4045),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3531)] = {
    [sym_heredoc_content] = STATE(3531),
    [anon_sym_SEMI] = ACTIONS(4047),
    [anon_sym_EQ] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_COMMA] = ACTIONS(4047),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4047),
    [anon_sym_DASH_GT] = ACTIONS(4047),
    [anon_sym_STAR] = ACTIONS(4049),
    [anon_sym_PERCENT] = ACTIONS(4049),
    [anon_sym_GT] = ACTIONS(4049),
    [anon_sym_LT_LT] = ACTIONS(4049),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4047),
    [anon_sym_PLUS_EQ] = ACTIONS(4047),
    [anon_sym_DASH_EQ] = ACTIONS(4047),
    [anon_sym_DOT_EQ] = ACTIONS(4047),
    [anon_sym_STAR_EQ] = ACTIONS(4047),
    [anon_sym_SLASH_EQ] = ACTIONS(4047),
    [anon_sym_PERCENT_EQ] = ACTIONS(4047),
    [anon_sym_x_EQ] = ACTIONS(4047),
    [anon_sym_AMP_EQ] = ACTIONS(4047),
    [anon_sym_PIPE_EQ] = ACTIONS(4047),
    [anon_sym_CARET_EQ] = ACTIONS(4047),
    [anon_sym_LT_LT_EQ] = ACTIONS(4047),
    [anon_sym_GT_GT_EQ] = ACTIONS(4047),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4047),
    [anon_sym_DOT_DOT] = ACTIONS(4049),
    [anon_sym_STAR_STAR] = ACTIONS(4049),
    [anon_sym_PIPE_PIPE] = ACTIONS(4049),
    [anon_sym_SLASH_SLASH] = ACTIONS(4049),
    [anon_sym_CARET_CARET] = ACTIONS(4047),
    [anon_sym_AMP_AMP] = ACTIONS(4049),
    [anon_sym_PIPE] = ACTIONS(4049),
    [anon_sym_CARET] = ACTIONS(4049),
    [anon_sym_AMP] = ACTIONS(4049),
    [anon_sym_GT_GT2] = ACTIONS(4049),
    [anon_sym_PLUS] = ACTIONS(4049),
    [anon_sym_DASH] = ACTIONS(4049),
    [anon_sym_DOT] = ACTIONS(4049),
    [anon_sym_SLASH] = ACTIONS(4049),
    [anon_sym_x] = ACTIONS(4049),
    [anon_sym_EQ_TILDE] = ACTIONS(4047),
    [anon_sym_BANG_TILDE] = ACTIONS(4047),
    [anon_sym_EQ_EQ] = ACTIONS(4049),
    [anon_sym_BANG_EQ] = ACTIONS(4047),
    [anon_sym_eq] = ACTIONS(4049),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4047),
    [anon_sym_equ] = ACTIONS(4047),
    [anon_sym_eqr] = ACTIONS(4047),
    [anon_sym_ne] = ACTIONS(4047),
    [anon_sym_LT_EQ_GT] = ACTIONS(4047),
    [anon_sym_cmp] = ACTIONS(4047),
    [anon_sym_TILDE_TILDE] = ACTIONS(4047),
    [anon_sym_LT] = ACTIONS(4049),
    [anon_sym_LT_EQ] = ACTIONS(4049),
    [anon_sym_GT_EQ] = ACTIONS(4047),
    [anon_sym_lt] = ACTIONS(4047),
    [anon_sym_le] = ACTIONS(4047),
    [anon_sym_ge] = ACTIONS(4047),
    [anon_sym_gt] = ACTIONS(4047),
    [anon_sym_isa] = ACTIONS(4047),
    [anon_sym_PLUS_PLUS] = ACTIONS(4047),
    [anon_sym_DASH_DASH] = ACTIONS(4047),
    [anon_sym_QMARK] = ACTIONS(4047),
    [anon_sym_EQ_GT] = ACTIONS(4047),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4047),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3532)] = {
    [sym_heredoc_content] = STATE(3532),
    [anon_sym_SEMI] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_COMMA] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3485),
    [anon_sym_DASH_GT] = ACTIONS(3485),
    [anon_sym_STAR] = ACTIONS(3536),
    [anon_sym_PERCENT] = ACTIONS(3536),
    [anon_sym_GT] = ACTIONS(3536),
    [anon_sym_LT_LT] = ACTIONS(3536),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3485),
    [anon_sym_PLUS_EQ] = ACTIONS(3485),
    [anon_sym_DASH_EQ] = ACTIONS(3485),
    [anon_sym_DOT_EQ] = ACTIONS(3485),
    [anon_sym_STAR_EQ] = ACTIONS(3485),
    [anon_sym_SLASH_EQ] = ACTIONS(3485),
    [anon_sym_PERCENT_EQ] = ACTIONS(3485),
    [anon_sym_x_EQ] = ACTIONS(3485),
    [anon_sym_AMP_EQ] = ACTIONS(3485),
    [anon_sym_PIPE_EQ] = ACTIONS(3485),
    [anon_sym_CARET_EQ] = ACTIONS(3485),
    [anon_sym_LT_LT_EQ] = ACTIONS(3485),
    [anon_sym_GT_GT_EQ] = ACTIONS(3485),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3485),
    [anon_sym_DOT_DOT] = ACTIONS(3536),
    [anon_sym_STAR_STAR] = ACTIONS(3536),
    [anon_sym_PIPE_PIPE] = ACTIONS(3536),
    [anon_sym_SLASH_SLASH] = ACTIONS(3536),
    [anon_sym_CARET_CARET] = ACTIONS(3485),
    [anon_sym_AMP_AMP] = ACTIONS(3536),
    [anon_sym_PIPE] = ACTIONS(3536),
    [anon_sym_CARET] = ACTIONS(3536),
    [anon_sym_AMP] = ACTIONS(3536),
    [anon_sym_GT_GT2] = ACTIONS(3536),
    [anon_sym_PLUS] = ACTIONS(3536),
    [anon_sym_DASH] = ACTIONS(3536),
    [anon_sym_DOT] = ACTIONS(3536),
    [anon_sym_SLASH] = ACTIONS(3536),
    [anon_sym_x] = ACTIONS(3536),
    [anon_sym_EQ_TILDE] = ACTIONS(3485),
    [anon_sym_BANG_TILDE] = ACTIONS(3485),
    [anon_sym_EQ_EQ] = ACTIONS(3536),
    [anon_sym_BANG_EQ] = ACTIONS(3485),
    [anon_sym_eq] = ACTIONS(3536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3485),
    [anon_sym_equ] = ACTIONS(3485),
    [anon_sym_eqr] = ACTIONS(3485),
    [anon_sym_ne] = ACTIONS(3485),
    [anon_sym_LT_EQ_GT] = ACTIONS(3485),
    [anon_sym_cmp] = ACTIONS(3485),
    [anon_sym_TILDE_TILDE] = ACTIONS(3485),
    [anon_sym_LT] = ACTIONS(3536),
    [anon_sym_LT_EQ] = ACTIONS(3536),
    [anon_sym_GT_EQ] = ACTIONS(3485),
    [anon_sym_lt] = ACTIONS(3485),
    [anon_sym_le] = ACTIONS(3485),
    [anon_sym_ge] = ACTIONS(3485),
    [anon_sym_gt] = ACTIONS(3485),
    [anon_sym_isa] = ACTIONS(3485),
    [anon_sym_PLUS_PLUS] = ACTIONS(3485),
    [anon_sym_DASH_DASH] = ACTIONS(3485),
    [anon_sym_QMARK] = ACTIONS(3485),
    [anon_sym_EQ_GT] = ACTIONS(3485),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3485),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3533)] = {
    [sym_heredoc_content] = STATE(3533),
    [aux_sym__term_rightward_repeat1] = STATE(3374),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5307),
    [anon_sym_RPAREN] = ACTIONS(3485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5189),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3534)] = {
    [sym_heredoc_content] = STATE(3534),
    [anon_sym_SEMI] = ACTIONS(3415),
    [anon_sym_EQ] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_COMMA] = ACTIONS(3415),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3415),
    [anon_sym_DASH_GT] = ACTIONS(3415),
    [anon_sym_STAR] = ACTIONS(3417),
    [anon_sym_PERCENT] = ACTIONS(3417),
    [anon_sym_GT] = ACTIONS(3417),
    [anon_sym_LT_LT] = ACTIONS(3417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3415),
    [anon_sym_PLUS_EQ] = ACTIONS(3415),
    [anon_sym_DASH_EQ] = ACTIONS(3415),
    [anon_sym_DOT_EQ] = ACTIONS(3415),
    [anon_sym_STAR_EQ] = ACTIONS(3415),
    [anon_sym_SLASH_EQ] = ACTIONS(3415),
    [anon_sym_PERCENT_EQ] = ACTIONS(3415),
    [anon_sym_x_EQ] = ACTIONS(3415),
    [anon_sym_AMP_EQ] = ACTIONS(3415),
    [anon_sym_PIPE_EQ] = ACTIONS(3415),
    [anon_sym_CARET_EQ] = ACTIONS(3415),
    [anon_sym_LT_LT_EQ] = ACTIONS(3415),
    [anon_sym_GT_GT_EQ] = ACTIONS(3415),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3415),
    [anon_sym_DOT_DOT] = ACTIONS(3417),
    [anon_sym_STAR_STAR] = ACTIONS(3417),
    [anon_sym_PIPE_PIPE] = ACTIONS(3417),
    [anon_sym_SLASH_SLASH] = ACTIONS(3417),
    [anon_sym_CARET_CARET] = ACTIONS(3415),
    [anon_sym_AMP_AMP] = ACTIONS(3417),
    [anon_sym_PIPE] = ACTIONS(3417),
    [anon_sym_CARET] = ACTIONS(3417),
    [anon_sym_AMP] = ACTIONS(3417),
    [anon_sym_GT_GT2] = ACTIONS(3417),
    [anon_sym_PLUS] = ACTIONS(3417),
    [anon_sym_DASH] = ACTIONS(3417),
    [anon_sym_DOT] = ACTIONS(3417),
    [anon_sym_SLASH] = ACTIONS(3417),
    [anon_sym_x] = ACTIONS(3417),
    [anon_sym_EQ_TILDE] = ACTIONS(3415),
    [anon_sym_BANG_TILDE] = ACTIONS(3415),
    [anon_sym_EQ_EQ] = ACTIONS(3417),
    [anon_sym_BANG_EQ] = ACTIONS(3415),
    [anon_sym_eq] = ACTIONS(3417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3415),
    [anon_sym_equ] = ACTIONS(3415),
    [anon_sym_eqr] = ACTIONS(3415),
    [anon_sym_ne] = ACTIONS(3415),
    [anon_sym_LT_EQ_GT] = ACTIONS(3415),
    [anon_sym_cmp] = ACTIONS(3415),
    [anon_sym_TILDE_TILDE] = ACTIONS(3415),
    [anon_sym_LT] = ACTIONS(3417),
    [anon_sym_LT_EQ] = ACTIONS(3417),
    [anon_sym_GT_EQ] = ACTIONS(3415),
    [anon_sym_lt] = ACTIONS(3415),
    [anon_sym_le] = ACTIONS(3415),
    [anon_sym_ge] = ACTIONS(3415),
    [anon_sym_gt] = ACTIONS(3415),
    [anon_sym_isa] = ACTIONS(3415),
    [anon_sym_PLUS_PLUS] = ACTIONS(3415),
    [anon_sym_DASH_DASH] = ACTIONS(3415),
    [anon_sym_QMARK] = ACTIONS(3415),
    [anon_sym_EQ_GT] = ACTIONS(3415),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3415),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3535)] = {
    [sym_heredoc_content] = STATE(3535),
    [anon_sym_SEMI] = ACTIONS(4071),
    [anon_sym_EQ] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_COMMA] = ACTIONS(4071),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4071),
    [anon_sym_DASH_GT] = ACTIONS(4071),
    [anon_sym_STAR] = ACTIONS(4073),
    [anon_sym_PERCENT] = ACTIONS(4073),
    [anon_sym_GT] = ACTIONS(4073),
    [anon_sym_LT_LT] = ACTIONS(4073),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4071),
    [anon_sym_PLUS_EQ] = ACTIONS(4071),
    [anon_sym_DASH_EQ] = ACTIONS(4071),
    [anon_sym_DOT_EQ] = ACTIONS(4071),
    [anon_sym_STAR_EQ] = ACTIONS(4071),
    [anon_sym_SLASH_EQ] = ACTIONS(4071),
    [anon_sym_PERCENT_EQ] = ACTIONS(4071),
    [anon_sym_x_EQ] = ACTIONS(4071),
    [anon_sym_AMP_EQ] = ACTIONS(4071),
    [anon_sym_PIPE_EQ] = ACTIONS(4071),
    [anon_sym_CARET_EQ] = ACTIONS(4071),
    [anon_sym_LT_LT_EQ] = ACTIONS(4071),
    [anon_sym_GT_GT_EQ] = ACTIONS(4071),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4071),
    [anon_sym_DOT_DOT] = ACTIONS(4073),
    [anon_sym_STAR_STAR] = ACTIONS(4073),
    [anon_sym_PIPE_PIPE] = ACTIONS(4073),
    [anon_sym_SLASH_SLASH] = ACTIONS(4073),
    [anon_sym_CARET_CARET] = ACTIONS(4071),
    [anon_sym_AMP_AMP] = ACTIONS(4073),
    [anon_sym_PIPE] = ACTIONS(4073),
    [anon_sym_CARET] = ACTIONS(4073),
    [anon_sym_AMP] = ACTIONS(4073),
    [anon_sym_GT_GT2] = ACTIONS(4073),
    [anon_sym_PLUS] = ACTIONS(4073),
    [anon_sym_DASH] = ACTIONS(4073),
    [anon_sym_DOT] = ACTIONS(4073),
    [anon_sym_SLASH] = ACTIONS(4073),
    [anon_sym_x] = ACTIONS(4073),
    [anon_sym_EQ_TILDE] = ACTIONS(4071),
    [anon_sym_BANG_TILDE] = ACTIONS(4071),
    [anon_sym_EQ_EQ] = ACTIONS(4073),
    [anon_sym_BANG_EQ] = ACTIONS(4071),
    [anon_sym_eq] = ACTIONS(4073),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4071),
    [anon_sym_equ] = ACTIONS(4071),
    [anon_sym_eqr] = ACTIONS(4071),
    [anon_sym_ne] = ACTIONS(4071),
    [anon_sym_LT_EQ_GT] = ACTIONS(4071),
    [anon_sym_cmp] = ACTIONS(4071),
    [anon_sym_TILDE_TILDE] = ACTIONS(4071),
    [anon_sym_LT] = ACTIONS(4073),
    [anon_sym_LT_EQ] = ACTIONS(4073),
    [anon_sym_GT_EQ] = ACTIONS(4071),
    [anon_sym_lt] = ACTIONS(4071),
    [anon_sym_le] = ACTIONS(4071),
    [anon_sym_ge] = ACTIONS(4071),
    [anon_sym_gt] = ACTIONS(4071),
    [anon_sym_isa] = ACTIONS(4071),
    [anon_sym_PLUS_PLUS] = ACTIONS(4071),
    [anon_sym_DASH_DASH] = ACTIONS(4071),
    [anon_sym_QMARK] = ACTIONS(4071),
    [anon_sym_EQ_GT] = ACTIONS(4071),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4071),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3536)] = {
    [sym_heredoc_content] = STATE(3536),
    [anon_sym_SEMI] = ACTIONS(4135),
    [anon_sym_EQ] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_COMMA] = ACTIONS(4135),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4135),
    [anon_sym_DASH_GT] = ACTIONS(4135),
    [anon_sym_STAR] = ACTIONS(4137),
    [anon_sym_PERCENT] = ACTIONS(4137),
    [anon_sym_GT] = ACTIONS(4137),
    [anon_sym_LT_LT] = ACTIONS(4137),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4135),
    [anon_sym_PLUS_EQ] = ACTIONS(4135),
    [anon_sym_DASH_EQ] = ACTIONS(4135),
    [anon_sym_DOT_EQ] = ACTIONS(4135),
    [anon_sym_STAR_EQ] = ACTIONS(4135),
    [anon_sym_SLASH_EQ] = ACTIONS(4135),
    [anon_sym_PERCENT_EQ] = ACTIONS(4135),
    [anon_sym_x_EQ] = ACTIONS(4135),
    [anon_sym_AMP_EQ] = ACTIONS(4135),
    [anon_sym_PIPE_EQ] = ACTIONS(4135),
    [anon_sym_CARET_EQ] = ACTIONS(4135),
    [anon_sym_LT_LT_EQ] = ACTIONS(4135),
    [anon_sym_GT_GT_EQ] = ACTIONS(4135),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4135),
    [anon_sym_DOT_DOT] = ACTIONS(4137),
    [anon_sym_STAR_STAR] = ACTIONS(4137),
    [anon_sym_PIPE_PIPE] = ACTIONS(4137),
    [anon_sym_SLASH_SLASH] = ACTIONS(4137),
    [anon_sym_CARET_CARET] = ACTIONS(4135),
    [anon_sym_AMP_AMP] = ACTIONS(4137),
    [anon_sym_PIPE] = ACTIONS(4137),
    [anon_sym_CARET] = ACTIONS(4137),
    [anon_sym_AMP] = ACTIONS(4137),
    [anon_sym_GT_GT2] = ACTIONS(4137),
    [anon_sym_PLUS] = ACTIONS(4137),
    [anon_sym_DASH] = ACTIONS(4137),
    [anon_sym_DOT] = ACTIONS(4137),
    [anon_sym_SLASH] = ACTIONS(4137),
    [anon_sym_x] = ACTIONS(4137),
    [anon_sym_EQ_TILDE] = ACTIONS(4135),
    [anon_sym_BANG_TILDE] = ACTIONS(4135),
    [anon_sym_EQ_EQ] = ACTIONS(4137),
    [anon_sym_BANG_EQ] = ACTIONS(4135),
    [anon_sym_eq] = ACTIONS(4137),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4135),
    [anon_sym_equ] = ACTIONS(4135),
    [anon_sym_eqr] = ACTIONS(4135),
    [anon_sym_ne] = ACTIONS(4135),
    [anon_sym_LT_EQ_GT] = ACTIONS(4135),
    [anon_sym_cmp] = ACTIONS(4135),
    [anon_sym_TILDE_TILDE] = ACTIONS(4135),
    [anon_sym_LT] = ACTIONS(4137),
    [anon_sym_LT_EQ] = ACTIONS(4137),
    [anon_sym_GT_EQ] = ACTIONS(4135),
    [anon_sym_lt] = ACTIONS(4135),
    [anon_sym_le] = ACTIONS(4135),
    [anon_sym_ge] = ACTIONS(4135),
    [anon_sym_gt] = ACTIONS(4135),
    [anon_sym_isa] = ACTIONS(4135),
    [anon_sym_PLUS_PLUS] = ACTIONS(4135),
    [anon_sym_DASH_DASH] = ACTIONS(4135),
    [anon_sym_QMARK] = ACTIONS(4135),
    [anon_sym_EQ_GT] = ACTIONS(4135),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4135),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3537)] = {
    [sym_heredoc_content] = STATE(3537),
    [anon_sym_SEMI] = ACTIONS(3941),
    [anon_sym_EQ] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_COMMA] = ACTIONS(3941),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3941),
    [anon_sym_DASH_GT] = ACTIONS(3941),
    [anon_sym_STAR] = ACTIONS(3943),
    [anon_sym_PERCENT] = ACTIONS(3943),
    [anon_sym_GT] = ACTIONS(3943),
    [anon_sym_LT_LT] = ACTIONS(3943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3941),
    [anon_sym_PLUS_EQ] = ACTIONS(3941),
    [anon_sym_DASH_EQ] = ACTIONS(3941),
    [anon_sym_DOT_EQ] = ACTIONS(3941),
    [anon_sym_STAR_EQ] = ACTIONS(3941),
    [anon_sym_SLASH_EQ] = ACTIONS(3941),
    [anon_sym_PERCENT_EQ] = ACTIONS(3941),
    [anon_sym_x_EQ] = ACTIONS(3941),
    [anon_sym_AMP_EQ] = ACTIONS(3941),
    [anon_sym_PIPE_EQ] = ACTIONS(3941),
    [anon_sym_CARET_EQ] = ACTIONS(3941),
    [anon_sym_LT_LT_EQ] = ACTIONS(3941),
    [anon_sym_GT_GT_EQ] = ACTIONS(3941),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3941),
    [anon_sym_DOT_DOT] = ACTIONS(3943),
    [anon_sym_STAR_STAR] = ACTIONS(3943),
    [anon_sym_PIPE_PIPE] = ACTIONS(3943),
    [anon_sym_SLASH_SLASH] = ACTIONS(3943),
    [anon_sym_CARET_CARET] = ACTIONS(3941),
    [anon_sym_AMP_AMP] = ACTIONS(3943),
    [anon_sym_PIPE] = ACTIONS(3943),
    [anon_sym_CARET] = ACTIONS(3943),
    [anon_sym_AMP] = ACTIONS(3943),
    [anon_sym_GT_GT2] = ACTIONS(3943),
    [anon_sym_PLUS] = ACTIONS(3943),
    [anon_sym_DASH] = ACTIONS(3943),
    [anon_sym_DOT] = ACTIONS(3943),
    [anon_sym_SLASH] = ACTIONS(3943),
    [anon_sym_x] = ACTIONS(3943),
    [anon_sym_EQ_TILDE] = ACTIONS(3941),
    [anon_sym_BANG_TILDE] = ACTIONS(3941),
    [anon_sym_EQ_EQ] = ACTIONS(3943),
    [anon_sym_BANG_EQ] = ACTIONS(3941),
    [anon_sym_eq] = ACTIONS(3943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3941),
    [anon_sym_equ] = ACTIONS(3941),
    [anon_sym_eqr] = ACTIONS(3941),
    [anon_sym_ne] = ACTIONS(3941),
    [anon_sym_LT_EQ_GT] = ACTIONS(3941),
    [anon_sym_cmp] = ACTIONS(3941),
    [anon_sym_TILDE_TILDE] = ACTIONS(3941),
    [anon_sym_LT] = ACTIONS(3943),
    [anon_sym_LT_EQ] = ACTIONS(3943),
    [anon_sym_GT_EQ] = ACTIONS(3941),
    [anon_sym_lt] = ACTIONS(3941),
    [anon_sym_le] = ACTIONS(3941),
    [anon_sym_ge] = ACTIONS(3941),
    [anon_sym_gt] = ACTIONS(3941),
    [anon_sym_isa] = ACTIONS(3941),
    [anon_sym_PLUS_PLUS] = ACTIONS(3941),
    [anon_sym_DASH_DASH] = ACTIONS(3941),
    [anon_sym_QMARK] = ACTIONS(3941),
    [anon_sym_EQ_GT] = ACTIONS(3941),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3941),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3538)] = {
    [sym_heredoc_content] = STATE(3538),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2702),
    [anon_sym_PERCENT] = ACTIONS(2702),
    [anon_sym_GT] = ACTIONS(2702),
    [anon_sym_LT_LT] = ACTIONS(2702),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2711),
    [anon_sym_PLUS_EQ] = ACTIONS(2711),
    [anon_sym_DASH_EQ] = ACTIONS(2711),
    [anon_sym_DOT_EQ] = ACTIONS(2711),
    [anon_sym_STAR_EQ] = ACTIONS(2711),
    [anon_sym_SLASH_EQ] = ACTIONS(2711),
    [anon_sym_PERCENT_EQ] = ACTIONS(2711),
    [anon_sym_x_EQ] = ACTIONS(2711),
    [anon_sym_AMP_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_EQ] = ACTIONS(2711),
    [anon_sym_CARET_EQ] = ACTIONS(2711),
    [anon_sym_LT_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_GT_EQ] = ACTIONS(2711),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2702),
    [anon_sym_STAR_STAR] = ACTIONS(2702),
    [anon_sym_PIPE_PIPE] = ACTIONS(2702),
    [anon_sym_SLASH_SLASH] = ACTIONS(2702),
    [anon_sym_CARET_CARET] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2702),
    [anon_sym_PIPE] = ACTIONS(2702),
    [anon_sym_CARET] = ACTIONS(2702),
    [anon_sym_AMP] = ACTIONS(2702),
    [anon_sym_GT_GT2] = ACTIONS(2702),
    [anon_sym_PLUS] = ACTIONS(2702),
    [anon_sym_DASH] = ACTIONS(2702),
    [anon_sym_DOT] = ACTIONS(2702),
    [anon_sym_SLASH] = ACTIONS(2702),
    [anon_sym_x] = ACTIONS(2702),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_BANG_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2702),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_eq] = ACTIONS(2702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_equ] = ACTIONS(2711),
    [anon_sym_eqr] = ACTIONS(2711),
    [anon_sym_ne] = ACTIONS(2711),
    [anon_sym_LT_EQ_GT] = ACTIONS(2711),
    [anon_sym_cmp] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2702),
    [anon_sym_LT_EQ] = ACTIONS(2702),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_lt] = ACTIONS(2711),
    [anon_sym_le] = ACTIONS(2711),
    [anon_sym_ge] = ACTIONS(2711),
    [anon_sym_gt] = ACTIONS(2711),
    [anon_sym_isa] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_QMARK] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2711),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3539)] = {
    [sym_heredoc_content] = STATE(3539),
    [anon_sym_SEMI] = ACTIONS(3917),
    [anon_sym_EQ] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_COMMA] = ACTIONS(3917),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3917),
    [anon_sym_DASH_GT] = ACTIONS(3917),
    [anon_sym_STAR] = ACTIONS(3919),
    [anon_sym_PERCENT] = ACTIONS(3919),
    [anon_sym_GT] = ACTIONS(3919),
    [anon_sym_LT_LT] = ACTIONS(3919),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3917),
    [anon_sym_PLUS_EQ] = ACTIONS(3917),
    [anon_sym_DASH_EQ] = ACTIONS(3917),
    [anon_sym_DOT_EQ] = ACTIONS(3917),
    [anon_sym_STAR_EQ] = ACTIONS(3917),
    [anon_sym_SLASH_EQ] = ACTIONS(3917),
    [anon_sym_PERCENT_EQ] = ACTIONS(3917),
    [anon_sym_x_EQ] = ACTIONS(3917),
    [anon_sym_AMP_EQ] = ACTIONS(3917),
    [anon_sym_PIPE_EQ] = ACTIONS(3917),
    [anon_sym_CARET_EQ] = ACTIONS(3917),
    [anon_sym_LT_LT_EQ] = ACTIONS(3917),
    [anon_sym_GT_GT_EQ] = ACTIONS(3917),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3917),
    [anon_sym_DOT_DOT] = ACTIONS(3919),
    [anon_sym_STAR_STAR] = ACTIONS(3919),
    [anon_sym_PIPE_PIPE] = ACTIONS(3919),
    [anon_sym_SLASH_SLASH] = ACTIONS(3919),
    [anon_sym_CARET_CARET] = ACTIONS(3917),
    [anon_sym_AMP_AMP] = ACTIONS(3919),
    [anon_sym_PIPE] = ACTIONS(3919),
    [anon_sym_CARET] = ACTIONS(3919),
    [anon_sym_AMP] = ACTIONS(3919),
    [anon_sym_GT_GT2] = ACTIONS(3919),
    [anon_sym_PLUS] = ACTIONS(3919),
    [anon_sym_DASH] = ACTIONS(3919),
    [anon_sym_DOT] = ACTIONS(3919),
    [anon_sym_SLASH] = ACTIONS(3919),
    [anon_sym_x] = ACTIONS(3919),
    [anon_sym_EQ_TILDE] = ACTIONS(3917),
    [anon_sym_BANG_TILDE] = ACTIONS(3917),
    [anon_sym_EQ_EQ] = ACTIONS(3919),
    [anon_sym_BANG_EQ] = ACTIONS(3917),
    [anon_sym_eq] = ACTIONS(3919),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3917),
    [anon_sym_equ] = ACTIONS(3917),
    [anon_sym_eqr] = ACTIONS(3917),
    [anon_sym_ne] = ACTIONS(3917),
    [anon_sym_LT_EQ_GT] = ACTIONS(3917),
    [anon_sym_cmp] = ACTIONS(3917),
    [anon_sym_TILDE_TILDE] = ACTIONS(3917),
    [anon_sym_LT] = ACTIONS(3919),
    [anon_sym_LT_EQ] = ACTIONS(3919),
    [anon_sym_GT_EQ] = ACTIONS(3917),
    [anon_sym_lt] = ACTIONS(3917),
    [anon_sym_le] = ACTIONS(3917),
    [anon_sym_ge] = ACTIONS(3917),
    [anon_sym_gt] = ACTIONS(3917),
    [anon_sym_isa] = ACTIONS(3917),
    [anon_sym_PLUS_PLUS] = ACTIONS(3917),
    [anon_sym_DASH_DASH] = ACTIONS(3917),
    [anon_sym_QMARK] = ACTIONS(3917),
    [anon_sym_EQ_GT] = ACTIONS(3917),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3917),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3540)] = {
    [sym_heredoc_content] = STATE(3540),
    [anon_sym_SEMI] = ACTIONS(3949),
    [anon_sym_EQ] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_COMMA] = ACTIONS(3949),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3949),
    [anon_sym_DASH_GT] = ACTIONS(3949),
    [anon_sym_STAR] = ACTIONS(3951),
    [anon_sym_PERCENT] = ACTIONS(3951),
    [anon_sym_GT] = ACTIONS(3951),
    [anon_sym_LT_LT] = ACTIONS(3951),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3949),
    [anon_sym_PLUS_EQ] = ACTIONS(3949),
    [anon_sym_DASH_EQ] = ACTIONS(3949),
    [anon_sym_DOT_EQ] = ACTIONS(3949),
    [anon_sym_STAR_EQ] = ACTIONS(3949),
    [anon_sym_SLASH_EQ] = ACTIONS(3949),
    [anon_sym_PERCENT_EQ] = ACTIONS(3949),
    [anon_sym_x_EQ] = ACTIONS(3949),
    [anon_sym_AMP_EQ] = ACTIONS(3949),
    [anon_sym_PIPE_EQ] = ACTIONS(3949),
    [anon_sym_CARET_EQ] = ACTIONS(3949),
    [anon_sym_LT_LT_EQ] = ACTIONS(3949),
    [anon_sym_GT_GT_EQ] = ACTIONS(3949),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3949),
    [anon_sym_DOT_DOT] = ACTIONS(3951),
    [anon_sym_STAR_STAR] = ACTIONS(3951),
    [anon_sym_PIPE_PIPE] = ACTIONS(3951),
    [anon_sym_SLASH_SLASH] = ACTIONS(3951),
    [anon_sym_CARET_CARET] = ACTIONS(3949),
    [anon_sym_AMP_AMP] = ACTIONS(3951),
    [anon_sym_PIPE] = ACTIONS(3951),
    [anon_sym_CARET] = ACTIONS(3951),
    [anon_sym_AMP] = ACTIONS(3951),
    [anon_sym_GT_GT2] = ACTIONS(3951),
    [anon_sym_PLUS] = ACTIONS(3951),
    [anon_sym_DASH] = ACTIONS(3951),
    [anon_sym_DOT] = ACTIONS(3951),
    [anon_sym_SLASH] = ACTIONS(3951),
    [anon_sym_x] = ACTIONS(3951),
    [anon_sym_EQ_TILDE] = ACTIONS(3949),
    [anon_sym_BANG_TILDE] = ACTIONS(3949),
    [anon_sym_EQ_EQ] = ACTIONS(3951),
    [anon_sym_BANG_EQ] = ACTIONS(3949),
    [anon_sym_eq] = ACTIONS(3951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3949),
    [anon_sym_equ] = ACTIONS(3949),
    [anon_sym_eqr] = ACTIONS(3949),
    [anon_sym_ne] = ACTIONS(3949),
    [anon_sym_LT_EQ_GT] = ACTIONS(3949),
    [anon_sym_cmp] = ACTIONS(3949),
    [anon_sym_TILDE_TILDE] = ACTIONS(3949),
    [anon_sym_LT] = ACTIONS(3951),
    [anon_sym_LT_EQ] = ACTIONS(3951),
    [anon_sym_GT_EQ] = ACTIONS(3949),
    [anon_sym_lt] = ACTIONS(3949),
    [anon_sym_le] = ACTIONS(3949),
    [anon_sym_ge] = ACTIONS(3949),
    [anon_sym_gt] = ACTIONS(3949),
    [anon_sym_isa] = ACTIONS(3949),
    [anon_sym_PLUS_PLUS] = ACTIONS(3949),
    [anon_sym_DASH_DASH] = ACTIONS(3949),
    [anon_sym_QMARK] = ACTIONS(3949),
    [anon_sym_EQ_GT] = ACTIONS(3949),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3949),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3541)] = {
    [sym_heredoc_content] = STATE(3541),
    [anon_sym_SEMI] = ACTIONS(3957),
    [anon_sym_EQ] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_COMMA] = ACTIONS(3957),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3957),
    [anon_sym_DASH_GT] = ACTIONS(3957),
    [anon_sym_STAR] = ACTIONS(3959),
    [anon_sym_PERCENT] = ACTIONS(3959),
    [anon_sym_GT] = ACTIONS(3959),
    [anon_sym_LT_LT] = ACTIONS(3959),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3957),
    [anon_sym_PLUS_EQ] = ACTIONS(3957),
    [anon_sym_DASH_EQ] = ACTIONS(3957),
    [anon_sym_DOT_EQ] = ACTIONS(3957),
    [anon_sym_STAR_EQ] = ACTIONS(3957),
    [anon_sym_SLASH_EQ] = ACTIONS(3957),
    [anon_sym_PERCENT_EQ] = ACTIONS(3957),
    [anon_sym_x_EQ] = ACTIONS(3957),
    [anon_sym_AMP_EQ] = ACTIONS(3957),
    [anon_sym_PIPE_EQ] = ACTIONS(3957),
    [anon_sym_CARET_EQ] = ACTIONS(3957),
    [anon_sym_LT_LT_EQ] = ACTIONS(3957),
    [anon_sym_GT_GT_EQ] = ACTIONS(3957),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3957),
    [anon_sym_DOT_DOT] = ACTIONS(3959),
    [anon_sym_STAR_STAR] = ACTIONS(3959),
    [anon_sym_PIPE_PIPE] = ACTIONS(3959),
    [anon_sym_SLASH_SLASH] = ACTIONS(3959),
    [anon_sym_CARET_CARET] = ACTIONS(3957),
    [anon_sym_AMP_AMP] = ACTIONS(3959),
    [anon_sym_PIPE] = ACTIONS(3959),
    [anon_sym_CARET] = ACTIONS(3959),
    [anon_sym_AMP] = ACTIONS(3959),
    [anon_sym_GT_GT2] = ACTIONS(3959),
    [anon_sym_PLUS] = ACTIONS(3959),
    [anon_sym_DASH] = ACTIONS(3959),
    [anon_sym_DOT] = ACTIONS(3959),
    [anon_sym_SLASH] = ACTIONS(3959),
    [anon_sym_x] = ACTIONS(3959),
    [anon_sym_EQ_TILDE] = ACTIONS(3957),
    [anon_sym_BANG_TILDE] = ACTIONS(3957),
    [anon_sym_EQ_EQ] = ACTIONS(3959),
    [anon_sym_BANG_EQ] = ACTIONS(3957),
    [anon_sym_eq] = ACTIONS(3959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3957),
    [anon_sym_equ] = ACTIONS(3957),
    [anon_sym_eqr] = ACTIONS(3957),
    [anon_sym_ne] = ACTIONS(3957),
    [anon_sym_LT_EQ_GT] = ACTIONS(3957),
    [anon_sym_cmp] = ACTIONS(3957),
    [anon_sym_TILDE_TILDE] = ACTIONS(3957),
    [anon_sym_LT] = ACTIONS(3959),
    [anon_sym_LT_EQ] = ACTIONS(3959),
    [anon_sym_GT_EQ] = ACTIONS(3957),
    [anon_sym_lt] = ACTIONS(3957),
    [anon_sym_le] = ACTIONS(3957),
    [anon_sym_ge] = ACTIONS(3957),
    [anon_sym_gt] = ACTIONS(3957),
    [anon_sym_isa] = ACTIONS(3957),
    [anon_sym_PLUS_PLUS] = ACTIONS(3957),
    [anon_sym_DASH_DASH] = ACTIONS(3957),
    [anon_sym_QMARK] = ACTIONS(3957),
    [anon_sym_EQ_GT] = ACTIONS(3957),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3957),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3542)] = {
    [sym_heredoc_content] = STATE(3542),
    [anon_sym_SEMI] = ACTIONS(4208),
    [anon_sym_EQ] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_COMMA] = ACTIONS(4208),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4208),
    [anon_sym_DASH_GT] = ACTIONS(4208),
    [anon_sym_STAR] = ACTIONS(4210),
    [anon_sym_PERCENT] = ACTIONS(4210),
    [anon_sym_GT] = ACTIONS(4210),
    [anon_sym_LT_LT] = ACTIONS(4210),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4208),
    [anon_sym_PLUS_EQ] = ACTIONS(4208),
    [anon_sym_DASH_EQ] = ACTIONS(4208),
    [anon_sym_DOT_EQ] = ACTIONS(4208),
    [anon_sym_STAR_EQ] = ACTIONS(4208),
    [anon_sym_SLASH_EQ] = ACTIONS(4208),
    [anon_sym_PERCENT_EQ] = ACTIONS(4208),
    [anon_sym_x_EQ] = ACTIONS(4208),
    [anon_sym_AMP_EQ] = ACTIONS(4208),
    [anon_sym_PIPE_EQ] = ACTIONS(4208),
    [anon_sym_CARET_EQ] = ACTIONS(4208),
    [anon_sym_LT_LT_EQ] = ACTIONS(4208),
    [anon_sym_GT_GT_EQ] = ACTIONS(4208),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4208),
    [anon_sym_DOT_DOT] = ACTIONS(4210),
    [anon_sym_STAR_STAR] = ACTIONS(4210),
    [anon_sym_PIPE_PIPE] = ACTIONS(4210),
    [anon_sym_SLASH_SLASH] = ACTIONS(4210),
    [anon_sym_CARET_CARET] = ACTIONS(4208),
    [anon_sym_AMP_AMP] = ACTIONS(4210),
    [anon_sym_PIPE] = ACTIONS(4210),
    [anon_sym_CARET] = ACTIONS(4210),
    [anon_sym_AMP] = ACTIONS(4210),
    [anon_sym_GT_GT2] = ACTIONS(4210),
    [anon_sym_PLUS] = ACTIONS(4210),
    [anon_sym_DASH] = ACTIONS(4210),
    [anon_sym_DOT] = ACTIONS(4210),
    [anon_sym_SLASH] = ACTIONS(4210),
    [anon_sym_x] = ACTIONS(4210),
    [anon_sym_EQ_TILDE] = ACTIONS(4208),
    [anon_sym_BANG_TILDE] = ACTIONS(4208),
    [anon_sym_EQ_EQ] = ACTIONS(4210),
    [anon_sym_BANG_EQ] = ACTIONS(4208),
    [anon_sym_eq] = ACTIONS(4210),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4208),
    [anon_sym_equ] = ACTIONS(4208),
    [anon_sym_eqr] = ACTIONS(4208),
    [anon_sym_ne] = ACTIONS(4208),
    [anon_sym_LT_EQ_GT] = ACTIONS(4208),
    [anon_sym_cmp] = ACTIONS(4208),
    [anon_sym_TILDE_TILDE] = ACTIONS(4208),
    [anon_sym_LT] = ACTIONS(4210),
    [anon_sym_LT_EQ] = ACTIONS(4210),
    [anon_sym_GT_EQ] = ACTIONS(4208),
    [anon_sym_lt] = ACTIONS(4208),
    [anon_sym_le] = ACTIONS(4208),
    [anon_sym_ge] = ACTIONS(4208),
    [anon_sym_gt] = ACTIONS(4208),
    [anon_sym_isa] = ACTIONS(4208),
    [anon_sym_PLUS_PLUS] = ACTIONS(4208),
    [anon_sym_DASH_DASH] = ACTIONS(4208),
    [anon_sym_QMARK] = ACTIONS(4208),
    [anon_sym_EQ_GT] = ACTIONS(4208),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4208),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3543)] = {
    [sym_heredoc_content] = STATE(3543),
    [anon_sym_SEMI] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3977),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3544)] = {
    [sym_heredoc_content] = STATE(3544),
    [anon_sym_SEMI] = ACTIONS(4165),
    [anon_sym_EQ] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_COMMA] = ACTIONS(4165),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4165),
    [anon_sym_DASH_GT] = ACTIONS(4165),
    [anon_sym_STAR] = ACTIONS(4167),
    [anon_sym_PERCENT] = ACTIONS(4167),
    [anon_sym_GT] = ACTIONS(4167),
    [anon_sym_LT_LT] = ACTIONS(4167),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4165),
    [anon_sym_PLUS_EQ] = ACTIONS(4165),
    [anon_sym_DASH_EQ] = ACTIONS(4165),
    [anon_sym_DOT_EQ] = ACTIONS(4165),
    [anon_sym_STAR_EQ] = ACTIONS(4165),
    [anon_sym_SLASH_EQ] = ACTIONS(4165),
    [anon_sym_PERCENT_EQ] = ACTIONS(4165),
    [anon_sym_x_EQ] = ACTIONS(4165),
    [anon_sym_AMP_EQ] = ACTIONS(4165),
    [anon_sym_PIPE_EQ] = ACTIONS(4165),
    [anon_sym_CARET_EQ] = ACTIONS(4165),
    [anon_sym_LT_LT_EQ] = ACTIONS(4165),
    [anon_sym_GT_GT_EQ] = ACTIONS(4165),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4165),
    [anon_sym_DOT_DOT] = ACTIONS(4167),
    [anon_sym_STAR_STAR] = ACTIONS(4167),
    [anon_sym_PIPE_PIPE] = ACTIONS(4167),
    [anon_sym_SLASH_SLASH] = ACTIONS(4167),
    [anon_sym_CARET_CARET] = ACTIONS(4165),
    [anon_sym_AMP_AMP] = ACTIONS(4167),
    [anon_sym_PIPE] = ACTIONS(4167),
    [anon_sym_CARET] = ACTIONS(4167),
    [anon_sym_AMP] = ACTIONS(4167),
    [anon_sym_GT_GT2] = ACTIONS(4167),
    [anon_sym_PLUS] = ACTIONS(4167),
    [anon_sym_DASH] = ACTIONS(4167),
    [anon_sym_DOT] = ACTIONS(4167),
    [anon_sym_SLASH] = ACTIONS(4167),
    [anon_sym_x] = ACTIONS(4167),
    [anon_sym_EQ_TILDE] = ACTIONS(4165),
    [anon_sym_BANG_TILDE] = ACTIONS(4165),
    [anon_sym_EQ_EQ] = ACTIONS(4167),
    [anon_sym_BANG_EQ] = ACTIONS(4165),
    [anon_sym_eq] = ACTIONS(4167),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4165),
    [anon_sym_equ] = ACTIONS(4165),
    [anon_sym_eqr] = ACTIONS(4165),
    [anon_sym_ne] = ACTIONS(4165),
    [anon_sym_LT_EQ_GT] = ACTIONS(4165),
    [anon_sym_cmp] = ACTIONS(4165),
    [anon_sym_TILDE_TILDE] = ACTIONS(4165),
    [anon_sym_LT] = ACTIONS(4167),
    [anon_sym_LT_EQ] = ACTIONS(4167),
    [anon_sym_GT_EQ] = ACTIONS(4165),
    [anon_sym_lt] = ACTIONS(4165),
    [anon_sym_le] = ACTIONS(4165),
    [anon_sym_ge] = ACTIONS(4165),
    [anon_sym_gt] = ACTIONS(4165),
    [anon_sym_isa] = ACTIONS(4165),
    [anon_sym_PLUS_PLUS] = ACTIONS(4165),
    [anon_sym_DASH_DASH] = ACTIONS(4165),
    [anon_sym_QMARK] = ACTIONS(4165),
    [anon_sym_EQ_GT] = ACTIONS(4165),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4165),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3545)] = {
    [sym_heredoc_content] = STATE(3545),
    [anon_sym_SEMI] = ACTIONS(4204),
    [anon_sym_EQ] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_COMMA] = ACTIONS(4204),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4204),
    [anon_sym_DASH_GT] = ACTIONS(4204),
    [anon_sym_STAR] = ACTIONS(4206),
    [anon_sym_PERCENT] = ACTIONS(4206),
    [anon_sym_GT] = ACTIONS(4206),
    [anon_sym_LT_LT] = ACTIONS(4206),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4204),
    [anon_sym_PLUS_EQ] = ACTIONS(4204),
    [anon_sym_DASH_EQ] = ACTIONS(4204),
    [anon_sym_DOT_EQ] = ACTIONS(4204),
    [anon_sym_STAR_EQ] = ACTIONS(4204),
    [anon_sym_SLASH_EQ] = ACTIONS(4204),
    [anon_sym_PERCENT_EQ] = ACTIONS(4204),
    [anon_sym_x_EQ] = ACTIONS(4204),
    [anon_sym_AMP_EQ] = ACTIONS(4204),
    [anon_sym_PIPE_EQ] = ACTIONS(4204),
    [anon_sym_CARET_EQ] = ACTIONS(4204),
    [anon_sym_LT_LT_EQ] = ACTIONS(4204),
    [anon_sym_GT_GT_EQ] = ACTIONS(4204),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4204),
    [anon_sym_DOT_DOT] = ACTIONS(4206),
    [anon_sym_STAR_STAR] = ACTIONS(4206),
    [anon_sym_PIPE_PIPE] = ACTIONS(4206),
    [anon_sym_SLASH_SLASH] = ACTIONS(4206),
    [anon_sym_CARET_CARET] = ACTIONS(4204),
    [anon_sym_AMP_AMP] = ACTIONS(4206),
    [anon_sym_PIPE] = ACTIONS(4206),
    [anon_sym_CARET] = ACTIONS(4206),
    [anon_sym_AMP] = ACTIONS(4206),
    [anon_sym_GT_GT2] = ACTIONS(4206),
    [anon_sym_PLUS] = ACTIONS(4206),
    [anon_sym_DASH] = ACTIONS(4206),
    [anon_sym_DOT] = ACTIONS(4206),
    [anon_sym_SLASH] = ACTIONS(4206),
    [anon_sym_x] = ACTIONS(4206),
    [anon_sym_EQ_TILDE] = ACTIONS(4204),
    [anon_sym_BANG_TILDE] = ACTIONS(4204),
    [anon_sym_EQ_EQ] = ACTIONS(4206),
    [anon_sym_BANG_EQ] = ACTIONS(4204),
    [anon_sym_eq] = ACTIONS(4206),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4204),
    [anon_sym_equ] = ACTIONS(4204),
    [anon_sym_eqr] = ACTIONS(4204),
    [anon_sym_ne] = ACTIONS(4204),
    [anon_sym_LT_EQ_GT] = ACTIONS(4204),
    [anon_sym_cmp] = ACTIONS(4204),
    [anon_sym_TILDE_TILDE] = ACTIONS(4204),
    [anon_sym_LT] = ACTIONS(4206),
    [anon_sym_LT_EQ] = ACTIONS(4206),
    [anon_sym_GT_EQ] = ACTIONS(4204),
    [anon_sym_lt] = ACTIONS(4204),
    [anon_sym_le] = ACTIONS(4204),
    [anon_sym_ge] = ACTIONS(4204),
    [anon_sym_gt] = ACTIONS(4204),
    [anon_sym_isa] = ACTIONS(4204),
    [anon_sym_PLUS_PLUS] = ACTIONS(4204),
    [anon_sym_DASH_DASH] = ACTIONS(4204),
    [anon_sym_QMARK] = ACTIONS(4204),
    [anon_sym_EQ_GT] = ACTIONS(4204),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4204),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3546)] = {
    [sym_heredoc_content] = STATE(3546),
    [anon_sym_SEMI] = ACTIONS(3869),
    [anon_sym_EQ] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_COMMA] = ACTIONS(3869),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3869),
    [anon_sym_DASH_GT] = ACTIONS(3869),
    [anon_sym_STAR] = ACTIONS(3871),
    [anon_sym_PERCENT] = ACTIONS(3871),
    [anon_sym_GT] = ACTIONS(3871),
    [anon_sym_LT_LT] = ACTIONS(3871),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3869),
    [anon_sym_PLUS_EQ] = ACTIONS(3869),
    [anon_sym_DASH_EQ] = ACTIONS(3869),
    [anon_sym_DOT_EQ] = ACTIONS(3869),
    [anon_sym_STAR_EQ] = ACTIONS(3869),
    [anon_sym_SLASH_EQ] = ACTIONS(3869),
    [anon_sym_PERCENT_EQ] = ACTIONS(3869),
    [anon_sym_x_EQ] = ACTIONS(3869),
    [anon_sym_AMP_EQ] = ACTIONS(3869),
    [anon_sym_PIPE_EQ] = ACTIONS(3869),
    [anon_sym_CARET_EQ] = ACTIONS(3869),
    [anon_sym_LT_LT_EQ] = ACTIONS(3869),
    [anon_sym_GT_GT_EQ] = ACTIONS(3869),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3869),
    [anon_sym_DOT_DOT] = ACTIONS(3871),
    [anon_sym_STAR_STAR] = ACTIONS(3871),
    [anon_sym_PIPE_PIPE] = ACTIONS(3871),
    [anon_sym_SLASH_SLASH] = ACTIONS(3871),
    [anon_sym_CARET_CARET] = ACTIONS(3869),
    [anon_sym_AMP_AMP] = ACTIONS(3871),
    [anon_sym_PIPE] = ACTIONS(3871),
    [anon_sym_CARET] = ACTIONS(3871),
    [anon_sym_AMP] = ACTIONS(3871),
    [anon_sym_GT_GT2] = ACTIONS(3871),
    [anon_sym_PLUS] = ACTIONS(3871),
    [anon_sym_DASH] = ACTIONS(3871),
    [anon_sym_DOT] = ACTIONS(3871),
    [anon_sym_SLASH] = ACTIONS(3871),
    [anon_sym_x] = ACTIONS(3871),
    [anon_sym_EQ_TILDE] = ACTIONS(3869),
    [anon_sym_BANG_TILDE] = ACTIONS(3869),
    [anon_sym_EQ_EQ] = ACTIONS(3871),
    [anon_sym_BANG_EQ] = ACTIONS(3869),
    [anon_sym_eq] = ACTIONS(3871),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3869),
    [anon_sym_equ] = ACTIONS(3869),
    [anon_sym_eqr] = ACTIONS(3869),
    [anon_sym_ne] = ACTIONS(3869),
    [anon_sym_LT_EQ_GT] = ACTIONS(3869),
    [anon_sym_cmp] = ACTIONS(3869),
    [anon_sym_TILDE_TILDE] = ACTIONS(3869),
    [anon_sym_LT] = ACTIONS(3871),
    [anon_sym_LT_EQ] = ACTIONS(3871),
    [anon_sym_GT_EQ] = ACTIONS(3869),
    [anon_sym_lt] = ACTIONS(3869),
    [anon_sym_le] = ACTIONS(3869),
    [anon_sym_ge] = ACTIONS(3869),
    [anon_sym_gt] = ACTIONS(3869),
    [anon_sym_isa] = ACTIONS(3869),
    [anon_sym_PLUS_PLUS] = ACTIONS(3869),
    [anon_sym_DASH_DASH] = ACTIONS(3869),
    [anon_sym_QMARK] = ACTIONS(3869),
    [anon_sym_EQ_GT] = ACTIONS(3869),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3869),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3547)] = {
    [sym_heredoc_content] = STATE(3547),
    [anon_sym_SEMI] = ACTIONS(4103),
    [anon_sym_EQ] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_COMMA] = ACTIONS(4103),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4103),
    [anon_sym_DASH_GT] = ACTIONS(4103),
    [anon_sym_STAR] = ACTIONS(4105),
    [anon_sym_PERCENT] = ACTIONS(4105),
    [anon_sym_GT] = ACTIONS(4105),
    [anon_sym_LT_LT] = ACTIONS(4105),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4103),
    [anon_sym_PLUS_EQ] = ACTIONS(4103),
    [anon_sym_DASH_EQ] = ACTIONS(4103),
    [anon_sym_DOT_EQ] = ACTIONS(4103),
    [anon_sym_STAR_EQ] = ACTIONS(4103),
    [anon_sym_SLASH_EQ] = ACTIONS(4103),
    [anon_sym_PERCENT_EQ] = ACTIONS(4103),
    [anon_sym_x_EQ] = ACTIONS(4103),
    [anon_sym_AMP_EQ] = ACTIONS(4103),
    [anon_sym_PIPE_EQ] = ACTIONS(4103),
    [anon_sym_CARET_EQ] = ACTIONS(4103),
    [anon_sym_LT_LT_EQ] = ACTIONS(4103),
    [anon_sym_GT_GT_EQ] = ACTIONS(4103),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4103),
    [anon_sym_DOT_DOT] = ACTIONS(4105),
    [anon_sym_STAR_STAR] = ACTIONS(4105),
    [anon_sym_PIPE_PIPE] = ACTIONS(4105),
    [anon_sym_SLASH_SLASH] = ACTIONS(4105),
    [anon_sym_CARET_CARET] = ACTIONS(4103),
    [anon_sym_AMP_AMP] = ACTIONS(4105),
    [anon_sym_PIPE] = ACTIONS(4105),
    [anon_sym_CARET] = ACTIONS(4105),
    [anon_sym_AMP] = ACTIONS(4105),
    [anon_sym_GT_GT2] = ACTIONS(4105),
    [anon_sym_PLUS] = ACTIONS(4105),
    [anon_sym_DASH] = ACTIONS(4105),
    [anon_sym_DOT] = ACTIONS(4105),
    [anon_sym_SLASH] = ACTIONS(4105),
    [anon_sym_x] = ACTIONS(4105),
    [anon_sym_EQ_TILDE] = ACTIONS(4103),
    [anon_sym_BANG_TILDE] = ACTIONS(4103),
    [anon_sym_EQ_EQ] = ACTIONS(4105),
    [anon_sym_BANG_EQ] = ACTIONS(4103),
    [anon_sym_eq] = ACTIONS(4105),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4103),
    [anon_sym_equ] = ACTIONS(4103),
    [anon_sym_eqr] = ACTIONS(4103),
    [anon_sym_ne] = ACTIONS(4103),
    [anon_sym_LT_EQ_GT] = ACTIONS(4103),
    [anon_sym_cmp] = ACTIONS(4103),
    [anon_sym_TILDE_TILDE] = ACTIONS(4103),
    [anon_sym_LT] = ACTIONS(4105),
    [anon_sym_LT_EQ] = ACTIONS(4105),
    [anon_sym_GT_EQ] = ACTIONS(4103),
    [anon_sym_lt] = ACTIONS(4103),
    [anon_sym_le] = ACTIONS(4103),
    [anon_sym_ge] = ACTIONS(4103),
    [anon_sym_gt] = ACTIONS(4103),
    [anon_sym_isa] = ACTIONS(4103),
    [anon_sym_PLUS_PLUS] = ACTIONS(4103),
    [anon_sym_DASH_DASH] = ACTIONS(4103),
    [anon_sym_QMARK] = ACTIONS(4103),
    [anon_sym_EQ_GT] = ACTIONS(4103),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4103),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3548)] = {
    [sym_heredoc_content] = STATE(3548),
    [anon_sym_SEMI] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3997),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3549)] = {
    [sym_heredoc_content] = STATE(3549),
    [anon_sym_SEMI] = ACTIONS(4159),
    [anon_sym_EQ] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_COMMA] = ACTIONS(4159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4159),
    [anon_sym_DASH_GT] = ACTIONS(4159),
    [anon_sym_STAR] = ACTIONS(4162),
    [anon_sym_PERCENT] = ACTIONS(4162),
    [anon_sym_GT] = ACTIONS(4162),
    [anon_sym_LT_LT] = ACTIONS(4162),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4159),
    [anon_sym_PLUS_EQ] = ACTIONS(4159),
    [anon_sym_DASH_EQ] = ACTIONS(4159),
    [anon_sym_DOT_EQ] = ACTIONS(4159),
    [anon_sym_STAR_EQ] = ACTIONS(4159),
    [anon_sym_SLASH_EQ] = ACTIONS(4159),
    [anon_sym_PERCENT_EQ] = ACTIONS(4159),
    [anon_sym_x_EQ] = ACTIONS(4159),
    [anon_sym_AMP_EQ] = ACTIONS(4159),
    [anon_sym_PIPE_EQ] = ACTIONS(4159),
    [anon_sym_CARET_EQ] = ACTIONS(4159),
    [anon_sym_LT_LT_EQ] = ACTIONS(4159),
    [anon_sym_GT_GT_EQ] = ACTIONS(4159),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4159),
    [anon_sym_DOT_DOT] = ACTIONS(4162),
    [anon_sym_STAR_STAR] = ACTIONS(4162),
    [anon_sym_PIPE_PIPE] = ACTIONS(4162),
    [anon_sym_SLASH_SLASH] = ACTIONS(4162),
    [anon_sym_CARET_CARET] = ACTIONS(4159),
    [anon_sym_AMP_AMP] = ACTIONS(4162),
    [anon_sym_PIPE] = ACTIONS(4162),
    [anon_sym_CARET] = ACTIONS(4162),
    [anon_sym_AMP] = ACTIONS(4162),
    [anon_sym_GT_GT2] = ACTIONS(4162),
    [anon_sym_PLUS] = ACTIONS(4162),
    [anon_sym_DASH] = ACTIONS(4162),
    [anon_sym_DOT] = ACTIONS(4162),
    [anon_sym_SLASH] = ACTIONS(4162),
    [anon_sym_x] = ACTIONS(4162),
    [anon_sym_EQ_TILDE] = ACTIONS(4159),
    [anon_sym_BANG_TILDE] = ACTIONS(4159),
    [anon_sym_EQ_EQ] = ACTIONS(4162),
    [anon_sym_BANG_EQ] = ACTIONS(4159),
    [anon_sym_eq] = ACTIONS(4162),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4159),
    [anon_sym_equ] = ACTIONS(4159),
    [anon_sym_eqr] = ACTIONS(4159),
    [anon_sym_ne] = ACTIONS(4159),
    [anon_sym_LT_EQ_GT] = ACTIONS(4159),
    [anon_sym_cmp] = ACTIONS(4159),
    [anon_sym_TILDE_TILDE] = ACTIONS(4159),
    [anon_sym_LT] = ACTIONS(4162),
    [anon_sym_LT_EQ] = ACTIONS(4162),
    [anon_sym_GT_EQ] = ACTIONS(4159),
    [anon_sym_lt] = ACTIONS(4159),
    [anon_sym_le] = ACTIONS(4159),
    [anon_sym_ge] = ACTIONS(4159),
    [anon_sym_gt] = ACTIONS(4159),
    [anon_sym_isa] = ACTIONS(4159),
    [anon_sym_PLUS_PLUS] = ACTIONS(4159),
    [anon_sym_DASH_DASH] = ACTIONS(4159),
    [anon_sym_QMARK] = ACTIONS(4159),
    [anon_sym_EQ_GT] = ACTIONS(4159),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4159),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3550)] = {
    [sym_heredoc_content] = STATE(3550),
    [anon_sym_SEMI] = ACTIONS(4169),
    [anon_sym_EQ] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_COMMA] = ACTIONS(4169),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4169),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(4171),
    [anon_sym_PERCENT] = ACTIONS(4171),
    [anon_sym_GT] = ACTIONS(4171),
    [anon_sym_LT_LT] = ACTIONS(4171),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4169),
    [anon_sym_PLUS_EQ] = ACTIONS(4169),
    [anon_sym_DASH_EQ] = ACTIONS(4169),
    [anon_sym_DOT_EQ] = ACTIONS(4169),
    [anon_sym_STAR_EQ] = ACTIONS(4169),
    [anon_sym_SLASH_EQ] = ACTIONS(4169),
    [anon_sym_PERCENT_EQ] = ACTIONS(4169),
    [anon_sym_x_EQ] = ACTIONS(4169),
    [anon_sym_AMP_EQ] = ACTIONS(4169),
    [anon_sym_PIPE_EQ] = ACTIONS(4169),
    [anon_sym_CARET_EQ] = ACTIONS(4169),
    [anon_sym_LT_LT_EQ] = ACTIONS(4169),
    [anon_sym_GT_GT_EQ] = ACTIONS(4169),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4169),
    [anon_sym_DOT_DOT] = ACTIONS(4171),
    [anon_sym_STAR_STAR] = ACTIONS(4171),
    [anon_sym_PIPE_PIPE] = ACTIONS(4171),
    [anon_sym_SLASH_SLASH] = ACTIONS(4171),
    [anon_sym_CARET_CARET] = ACTIONS(4169),
    [anon_sym_AMP_AMP] = ACTIONS(4171),
    [anon_sym_PIPE] = ACTIONS(4171),
    [anon_sym_CARET] = ACTIONS(4171),
    [anon_sym_AMP] = ACTIONS(4171),
    [anon_sym_GT_GT2] = ACTIONS(4171),
    [anon_sym_PLUS] = ACTIONS(4171),
    [anon_sym_DASH] = ACTIONS(4171),
    [anon_sym_DOT] = ACTIONS(4171),
    [anon_sym_SLASH] = ACTIONS(4171),
    [anon_sym_x] = ACTIONS(4171),
    [anon_sym_EQ_TILDE] = ACTIONS(4169),
    [anon_sym_BANG_TILDE] = ACTIONS(4169),
    [anon_sym_EQ_EQ] = ACTIONS(4171),
    [anon_sym_BANG_EQ] = ACTIONS(4169),
    [anon_sym_eq] = ACTIONS(4171),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4169),
    [anon_sym_equ] = ACTIONS(4169),
    [anon_sym_eqr] = ACTIONS(4169),
    [anon_sym_ne] = ACTIONS(4169),
    [anon_sym_LT_EQ_GT] = ACTIONS(4169),
    [anon_sym_cmp] = ACTIONS(4169),
    [anon_sym_TILDE_TILDE] = ACTIONS(4169),
    [anon_sym_LT] = ACTIONS(4171),
    [anon_sym_LT_EQ] = ACTIONS(4171),
    [anon_sym_GT_EQ] = ACTIONS(4169),
    [anon_sym_lt] = ACTIONS(4169),
    [anon_sym_le] = ACTIONS(4169),
    [anon_sym_ge] = ACTIONS(4169),
    [anon_sym_gt] = ACTIONS(4169),
    [anon_sym_isa] = ACTIONS(4169),
    [anon_sym_PLUS_PLUS] = ACTIONS(5374),
    [anon_sym_DASH_DASH] = ACTIONS(5374),
    [anon_sym_QMARK] = ACTIONS(4169),
    [anon_sym_EQ_GT] = ACTIONS(4169),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4169),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3551)] = {
    [sym_heredoc_content] = STATE(3551),
    [anon_sym_SEMI] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3695),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3552)] = {
    [sym_heredoc_content] = STATE(3552),
    [anon_sym_SEMI] = ACTIONS(4037),
    [anon_sym_EQ] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_COMMA] = ACTIONS(4037),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4037),
    [anon_sym_DASH_GT] = ACTIONS(4037),
    [anon_sym_STAR] = ACTIONS(4039),
    [anon_sym_PERCENT] = ACTIONS(4039),
    [anon_sym_GT] = ACTIONS(4039),
    [anon_sym_LT_LT] = ACTIONS(4039),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4037),
    [anon_sym_PLUS_EQ] = ACTIONS(4037),
    [anon_sym_DASH_EQ] = ACTIONS(4037),
    [anon_sym_DOT_EQ] = ACTIONS(4037),
    [anon_sym_STAR_EQ] = ACTIONS(4037),
    [anon_sym_SLASH_EQ] = ACTIONS(4037),
    [anon_sym_PERCENT_EQ] = ACTIONS(4037),
    [anon_sym_x_EQ] = ACTIONS(4037),
    [anon_sym_AMP_EQ] = ACTIONS(4037),
    [anon_sym_PIPE_EQ] = ACTIONS(4037),
    [anon_sym_CARET_EQ] = ACTIONS(4037),
    [anon_sym_LT_LT_EQ] = ACTIONS(4037),
    [anon_sym_GT_GT_EQ] = ACTIONS(4037),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4037),
    [anon_sym_DOT_DOT] = ACTIONS(4039),
    [anon_sym_STAR_STAR] = ACTIONS(4039),
    [anon_sym_PIPE_PIPE] = ACTIONS(4039),
    [anon_sym_SLASH_SLASH] = ACTIONS(4039),
    [anon_sym_CARET_CARET] = ACTIONS(4037),
    [anon_sym_AMP_AMP] = ACTIONS(4039),
    [anon_sym_PIPE] = ACTIONS(4039),
    [anon_sym_CARET] = ACTIONS(4039),
    [anon_sym_AMP] = ACTIONS(4039),
    [anon_sym_GT_GT2] = ACTIONS(4039),
    [anon_sym_PLUS] = ACTIONS(4039),
    [anon_sym_DASH] = ACTIONS(4039),
    [anon_sym_DOT] = ACTIONS(4039),
    [anon_sym_SLASH] = ACTIONS(4039),
    [anon_sym_x] = ACTIONS(4039),
    [anon_sym_EQ_TILDE] = ACTIONS(4037),
    [anon_sym_BANG_TILDE] = ACTIONS(4037),
    [anon_sym_EQ_EQ] = ACTIONS(4039),
    [anon_sym_BANG_EQ] = ACTIONS(4037),
    [anon_sym_eq] = ACTIONS(4039),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4037),
    [anon_sym_equ] = ACTIONS(4037),
    [anon_sym_eqr] = ACTIONS(4037),
    [anon_sym_ne] = ACTIONS(4037),
    [anon_sym_LT_EQ_GT] = ACTIONS(4037),
    [anon_sym_cmp] = ACTIONS(4037),
    [anon_sym_TILDE_TILDE] = ACTIONS(4037),
    [anon_sym_LT] = ACTIONS(4039),
    [anon_sym_LT_EQ] = ACTIONS(4039),
    [anon_sym_GT_EQ] = ACTIONS(4037),
    [anon_sym_lt] = ACTIONS(4037),
    [anon_sym_le] = ACTIONS(4037),
    [anon_sym_ge] = ACTIONS(4037),
    [anon_sym_gt] = ACTIONS(4037),
    [anon_sym_isa] = ACTIONS(4037),
    [anon_sym_PLUS_PLUS] = ACTIONS(4037),
    [anon_sym_DASH_DASH] = ACTIONS(4037),
    [anon_sym_QMARK] = ACTIONS(4037),
    [anon_sym_EQ_GT] = ACTIONS(4037),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4037),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3553)] = {
    [sym_heredoc_content] = STATE(3553),
    [anon_sym_SEMI] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3909),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3554)] = {
    [sym_heredoc_content] = STATE(3554),
    [anon_sym_SEMI] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(4017),
    [anon_sym_STAR] = ACTIONS(4019),
    [anon_sym_PERCENT] = ACTIONS(4019),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(4019),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(4019),
    [anon_sym_PLUS] = ACTIONS(4019),
    [anon_sym_DASH] = ACTIONS(4019),
    [anon_sym_DOT] = ACTIONS(4019),
    [anon_sym_SLASH] = ACTIONS(4019),
    [anon_sym_x] = ACTIONS(4019),
    [anon_sym_EQ_TILDE] = ACTIONS(4017),
    [anon_sym_BANG_TILDE] = ACTIONS(4017),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(4017),
    [anon_sym_DASH_DASH] = ACTIONS(4017),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4017),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3555)] = {
    [sym_heredoc_content] = STATE(3555),
    [anon_sym_SEMI] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4017),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3556)] = {
    [sym_heredoc_content] = STATE(3556),
    [anon_sym_SEMI] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3803),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3557)] = {
    [sym_heredoc_content] = STATE(3557),
    [anon_sym_SEMI] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3845),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3558)] = {
    [sym_heredoc_content] = STATE(3558),
    [anon_sym_SEMI] = ACTIONS(3857),
    [anon_sym_EQ] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_COMMA] = ACTIONS(3857),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3857),
    [anon_sym_DASH_GT] = ACTIONS(3857),
    [anon_sym_STAR] = ACTIONS(3859),
    [anon_sym_PERCENT] = ACTIONS(3859),
    [anon_sym_GT] = ACTIONS(3859),
    [anon_sym_LT_LT] = ACTIONS(3859),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3857),
    [anon_sym_PLUS_EQ] = ACTIONS(3857),
    [anon_sym_DASH_EQ] = ACTIONS(3857),
    [anon_sym_DOT_EQ] = ACTIONS(3857),
    [anon_sym_STAR_EQ] = ACTIONS(3857),
    [anon_sym_SLASH_EQ] = ACTIONS(3857),
    [anon_sym_PERCENT_EQ] = ACTIONS(3857),
    [anon_sym_x_EQ] = ACTIONS(3857),
    [anon_sym_AMP_EQ] = ACTIONS(3857),
    [anon_sym_PIPE_EQ] = ACTIONS(3857),
    [anon_sym_CARET_EQ] = ACTIONS(3857),
    [anon_sym_LT_LT_EQ] = ACTIONS(3857),
    [anon_sym_GT_GT_EQ] = ACTIONS(3857),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3857),
    [anon_sym_DOT_DOT] = ACTIONS(3859),
    [anon_sym_STAR_STAR] = ACTIONS(3859),
    [anon_sym_PIPE_PIPE] = ACTIONS(3859),
    [anon_sym_SLASH_SLASH] = ACTIONS(3859),
    [anon_sym_CARET_CARET] = ACTIONS(3857),
    [anon_sym_AMP_AMP] = ACTIONS(3859),
    [anon_sym_PIPE] = ACTIONS(3859),
    [anon_sym_CARET] = ACTIONS(3859),
    [anon_sym_AMP] = ACTIONS(3859),
    [anon_sym_GT_GT2] = ACTIONS(3859),
    [anon_sym_PLUS] = ACTIONS(3859),
    [anon_sym_DASH] = ACTIONS(3859),
    [anon_sym_DOT] = ACTIONS(3859),
    [anon_sym_SLASH] = ACTIONS(3859),
    [anon_sym_x] = ACTIONS(3859),
    [anon_sym_EQ_TILDE] = ACTIONS(3857),
    [anon_sym_BANG_TILDE] = ACTIONS(3857),
    [anon_sym_EQ_EQ] = ACTIONS(3859),
    [anon_sym_BANG_EQ] = ACTIONS(3857),
    [anon_sym_eq] = ACTIONS(3859),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3857),
    [anon_sym_equ] = ACTIONS(3857),
    [anon_sym_eqr] = ACTIONS(3857),
    [anon_sym_ne] = ACTIONS(3857),
    [anon_sym_LT_EQ_GT] = ACTIONS(3857),
    [anon_sym_cmp] = ACTIONS(3857),
    [anon_sym_TILDE_TILDE] = ACTIONS(3857),
    [anon_sym_LT] = ACTIONS(3859),
    [anon_sym_LT_EQ] = ACTIONS(3859),
    [anon_sym_GT_EQ] = ACTIONS(3857),
    [anon_sym_lt] = ACTIONS(3857),
    [anon_sym_le] = ACTIONS(3857),
    [anon_sym_ge] = ACTIONS(3857),
    [anon_sym_gt] = ACTIONS(3857),
    [anon_sym_isa] = ACTIONS(3857),
    [anon_sym_PLUS_PLUS] = ACTIONS(3857),
    [anon_sym_DASH_DASH] = ACTIONS(3857),
    [anon_sym_QMARK] = ACTIONS(3857),
    [anon_sym_EQ_GT] = ACTIONS(3857),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3857),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3559)] = {
    [sym_heredoc_content] = STATE(3559),
    [anon_sym_SEMI] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3739),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3560)] = {
    [sym_heredoc_content] = STATE(3560),
    [anon_sym_SEMI] = ACTIONS(3815),
    [anon_sym_EQ] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_COMMA] = ACTIONS(3815),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3815),
    [anon_sym_DASH_GT] = ACTIONS(3815),
    [anon_sym_STAR] = ACTIONS(3817),
    [anon_sym_PERCENT] = ACTIONS(3817),
    [anon_sym_GT] = ACTIONS(3817),
    [anon_sym_LT_LT] = ACTIONS(3817),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3815),
    [anon_sym_PLUS_EQ] = ACTIONS(3815),
    [anon_sym_DASH_EQ] = ACTIONS(3815),
    [anon_sym_DOT_EQ] = ACTIONS(3815),
    [anon_sym_STAR_EQ] = ACTIONS(3815),
    [anon_sym_SLASH_EQ] = ACTIONS(3815),
    [anon_sym_PERCENT_EQ] = ACTIONS(3815),
    [anon_sym_x_EQ] = ACTIONS(3815),
    [anon_sym_AMP_EQ] = ACTIONS(3815),
    [anon_sym_PIPE_EQ] = ACTIONS(3815),
    [anon_sym_CARET_EQ] = ACTIONS(3815),
    [anon_sym_LT_LT_EQ] = ACTIONS(3815),
    [anon_sym_GT_GT_EQ] = ACTIONS(3815),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3815),
    [anon_sym_DOT_DOT] = ACTIONS(3817),
    [anon_sym_STAR_STAR] = ACTIONS(3817),
    [anon_sym_PIPE_PIPE] = ACTIONS(3817),
    [anon_sym_SLASH_SLASH] = ACTIONS(3817),
    [anon_sym_CARET_CARET] = ACTIONS(3815),
    [anon_sym_AMP_AMP] = ACTIONS(3817),
    [anon_sym_PIPE] = ACTIONS(3817),
    [anon_sym_CARET] = ACTIONS(3817),
    [anon_sym_AMP] = ACTIONS(3817),
    [anon_sym_GT_GT2] = ACTIONS(3817),
    [anon_sym_PLUS] = ACTIONS(3817),
    [anon_sym_DASH] = ACTIONS(3817),
    [anon_sym_DOT] = ACTIONS(3817),
    [anon_sym_SLASH] = ACTIONS(3817),
    [anon_sym_x] = ACTIONS(3817),
    [anon_sym_EQ_TILDE] = ACTIONS(3815),
    [anon_sym_BANG_TILDE] = ACTIONS(3815),
    [anon_sym_EQ_EQ] = ACTIONS(3817),
    [anon_sym_BANG_EQ] = ACTIONS(3815),
    [anon_sym_eq] = ACTIONS(3817),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3815),
    [anon_sym_equ] = ACTIONS(3815),
    [anon_sym_eqr] = ACTIONS(3815),
    [anon_sym_ne] = ACTIONS(3815),
    [anon_sym_LT_EQ_GT] = ACTIONS(3815),
    [anon_sym_cmp] = ACTIONS(3815),
    [anon_sym_TILDE_TILDE] = ACTIONS(3815),
    [anon_sym_LT] = ACTIONS(3817),
    [anon_sym_LT_EQ] = ACTIONS(3817),
    [anon_sym_GT_EQ] = ACTIONS(3815),
    [anon_sym_lt] = ACTIONS(3815),
    [anon_sym_le] = ACTIONS(3815),
    [anon_sym_ge] = ACTIONS(3815),
    [anon_sym_gt] = ACTIONS(3815),
    [anon_sym_isa] = ACTIONS(3815),
    [anon_sym_PLUS_PLUS] = ACTIONS(3815),
    [anon_sym_DASH_DASH] = ACTIONS(3815),
    [anon_sym_QMARK] = ACTIONS(3815),
    [anon_sym_EQ_GT] = ACTIONS(3815),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3815),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3561)] = {
    [sym_heredoc_content] = STATE(3561),
    [anon_sym_SEMI] = ACTIONS(3775),
    [anon_sym_EQ] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_COMMA] = ACTIONS(3775),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3775),
    [anon_sym_DASH_GT] = ACTIONS(3775),
    [anon_sym_STAR] = ACTIONS(3777),
    [anon_sym_PERCENT] = ACTIONS(3777),
    [anon_sym_GT] = ACTIONS(3777),
    [anon_sym_LT_LT] = ACTIONS(3777),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3775),
    [anon_sym_PLUS_EQ] = ACTIONS(3775),
    [anon_sym_DASH_EQ] = ACTIONS(3775),
    [anon_sym_DOT_EQ] = ACTIONS(3775),
    [anon_sym_STAR_EQ] = ACTIONS(3775),
    [anon_sym_SLASH_EQ] = ACTIONS(3775),
    [anon_sym_PERCENT_EQ] = ACTIONS(3775),
    [anon_sym_x_EQ] = ACTIONS(3775),
    [anon_sym_AMP_EQ] = ACTIONS(3775),
    [anon_sym_PIPE_EQ] = ACTIONS(3775),
    [anon_sym_CARET_EQ] = ACTIONS(3775),
    [anon_sym_LT_LT_EQ] = ACTIONS(3775),
    [anon_sym_GT_GT_EQ] = ACTIONS(3775),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3775),
    [anon_sym_DOT_DOT] = ACTIONS(3777),
    [anon_sym_STAR_STAR] = ACTIONS(3777),
    [anon_sym_PIPE_PIPE] = ACTIONS(3777),
    [anon_sym_SLASH_SLASH] = ACTIONS(3777),
    [anon_sym_CARET_CARET] = ACTIONS(3775),
    [anon_sym_AMP_AMP] = ACTIONS(3777),
    [anon_sym_PIPE] = ACTIONS(3777),
    [anon_sym_CARET] = ACTIONS(3777),
    [anon_sym_AMP] = ACTIONS(3777),
    [anon_sym_GT_GT2] = ACTIONS(3777),
    [anon_sym_PLUS] = ACTIONS(3777),
    [anon_sym_DASH] = ACTIONS(3777),
    [anon_sym_DOT] = ACTIONS(3777),
    [anon_sym_SLASH] = ACTIONS(3777),
    [anon_sym_x] = ACTIONS(3777),
    [anon_sym_EQ_TILDE] = ACTIONS(3775),
    [anon_sym_BANG_TILDE] = ACTIONS(3775),
    [anon_sym_EQ_EQ] = ACTIONS(3777),
    [anon_sym_BANG_EQ] = ACTIONS(3775),
    [anon_sym_eq] = ACTIONS(3777),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3775),
    [anon_sym_equ] = ACTIONS(3775),
    [anon_sym_eqr] = ACTIONS(3775),
    [anon_sym_ne] = ACTIONS(3775),
    [anon_sym_LT_EQ_GT] = ACTIONS(3775),
    [anon_sym_cmp] = ACTIONS(3775),
    [anon_sym_TILDE_TILDE] = ACTIONS(3775),
    [anon_sym_LT] = ACTIONS(3777),
    [anon_sym_LT_EQ] = ACTIONS(3777),
    [anon_sym_GT_EQ] = ACTIONS(3775),
    [anon_sym_lt] = ACTIONS(3775),
    [anon_sym_le] = ACTIONS(3775),
    [anon_sym_ge] = ACTIONS(3775),
    [anon_sym_gt] = ACTIONS(3775),
    [anon_sym_isa] = ACTIONS(3775),
    [anon_sym_PLUS_PLUS] = ACTIONS(3775),
    [anon_sym_DASH_DASH] = ACTIONS(3775),
    [anon_sym_QMARK] = ACTIONS(3775),
    [anon_sym_EQ_GT] = ACTIONS(3775),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3775),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3562)] = {
    [sym_heredoc_content] = STATE(3562),
    [anon_sym_SEMI] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3897),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3563)] = {
    [sym_heredoc_content] = STATE(3563),
    [anon_sym_SEMI] = ACTIONS(3901),
    [anon_sym_EQ] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_COMMA] = ACTIONS(3901),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3901),
    [anon_sym_DASH_GT] = ACTIONS(3901),
    [anon_sym_STAR] = ACTIONS(3903),
    [anon_sym_PERCENT] = ACTIONS(3903),
    [anon_sym_GT] = ACTIONS(3903),
    [anon_sym_LT_LT] = ACTIONS(3903),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3901),
    [anon_sym_PLUS_EQ] = ACTIONS(3901),
    [anon_sym_DASH_EQ] = ACTIONS(3901),
    [anon_sym_DOT_EQ] = ACTIONS(3901),
    [anon_sym_STAR_EQ] = ACTIONS(3901),
    [anon_sym_SLASH_EQ] = ACTIONS(3901),
    [anon_sym_PERCENT_EQ] = ACTIONS(3901),
    [anon_sym_x_EQ] = ACTIONS(3901),
    [anon_sym_AMP_EQ] = ACTIONS(3901),
    [anon_sym_PIPE_EQ] = ACTIONS(3901),
    [anon_sym_CARET_EQ] = ACTIONS(3901),
    [anon_sym_LT_LT_EQ] = ACTIONS(3901),
    [anon_sym_GT_GT_EQ] = ACTIONS(3901),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3901),
    [anon_sym_DOT_DOT] = ACTIONS(3903),
    [anon_sym_STAR_STAR] = ACTIONS(3903),
    [anon_sym_PIPE_PIPE] = ACTIONS(3903),
    [anon_sym_SLASH_SLASH] = ACTIONS(3903),
    [anon_sym_CARET_CARET] = ACTIONS(3901),
    [anon_sym_AMP_AMP] = ACTIONS(3903),
    [anon_sym_PIPE] = ACTIONS(3903),
    [anon_sym_CARET] = ACTIONS(3903),
    [anon_sym_AMP] = ACTIONS(3903),
    [anon_sym_GT_GT2] = ACTIONS(3903),
    [anon_sym_PLUS] = ACTIONS(3903),
    [anon_sym_DASH] = ACTIONS(3903),
    [anon_sym_DOT] = ACTIONS(3903),
    [anon_sym_SLASH] = ACTIONS(3903),
    [anon_sym_x] = ACTIONS(3903),
    [anon_sym_EQ_TILDE] = ACTIONS(3901),
    [anon_sym_BANG_TILDE] = ACTIONS(3901),
    [anon_sym_EQ_EQ] = ACTIONS(3903),
    [anon_sym_BANG_EQ] = ACTIONS(3901),
    [anon_sym_eq] = ACTIONS(3903),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3901),
    [anon_sym_equ] = ACTIONS(3901),
    [anon_sym_eqr] = ACTIONS(3901),
    [anon_sym_ne] = ACTIONS(3901),
    [anon_sym_LT_EQ_GT] = ACTIONS(3901),
    [anon_sym_cmp] = ACTIONS(3901),
    [anon_sym_TILDE_TILDE] = ACTIONS(3901),
    [anon_sym_LT] = ACTIONS(3903),
    [anon_sym_LT_EQ] = ACTIONS(3903),
    [anon_sym_GT_EQ] = ACTIONS(3901),
    [anon_sym_lt] = ACTIONS(3901),
    [anon_sym_le] = ACTIONS(3901),
    [anon_sym_ge] = ACTIONS(3901),
    [anon_sym_gt] = ACTIONS(3901),
    [anon_sym_isa] = ACTIONS(3901),
    [anon_sym_PLUS_PLUS] = ACTIONS(3901),
    [anon_sym_DASH_DASH] = ACTIONS(3901),
    [anon_sym_QMARK] = ACTIONS(3901),
    [anon_sym_EQ_GT] = ACTIONS(3901),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3901),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3564)] = {
    [sym_heredoc_content] = STATE(3564),
    [anon_sym_SEMI] = ACTIONS(4119),
    [anon_sym_EQ] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_COMMA] = ACTIONS(4119),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4119),
    [anon_sym_DASH_GT] = ACTIONS(4119),
    [anon_sym_STAR] = ACTIONS(4121),
    [anon_sym_PERCENT] = ACTIONS(4121),
    [anon_sym_GT] = ACTIONS(4121),
    [anon_sym_LT_LT] = ACTIONS(4121),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4119),
    [anon_sym_PLUS_EQ] = ACTIONS(4119),
    [anon_sym_DASH_EQ] = ACTIONS(4119),
    [anon_sym_DOT_EQ] = ACTIONS(4119),
    [anon_sym_STAR_EQ] = ACTIONS(4119),
    [anon_sym_SLASH_EQ] = ACTIONS(4119),
    [anon_sym_PERCENT_EQ] = ACTIONS(4119),
    [anon_sym_x_EQ] = ACTIONS(4119),
    [anon_sym_AMP_EQ] = ACTIONS(4119),
    [anon_sym_PIPE_EQ] = ACTIONS(4119),
    [anon_sym_CARET_EQ] = ACTIONS(4119),
    [anon_sym_LT_LT_EQ] = ACTIONS(4119),
    [anon_sym_GT_GT_EQ] = ACTIONS(4119),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4119),
    [anon_sym_DOT_DOT] = ACTIONS(4121),
    [anon_sym_STAR_STAR] = ACTIONS(4121),
    [anon_sym_PIPE_PIPE] = ACTIONS(4121),
    [anon_sym_SLASH_SLASH] = ACTIONS(4121),
    [anon_sym_CARET_CARET] = ACTIONS(4119),
    [anon_sym_AMP_AMP] = ACTIONS(4121),
    [anon_sym_PIPE] = ACTIONS(4121),
    [anon_sym_CARET] = ACTIONS(4121),
    [anon_sym_AMP] = ACTIONS(4121),
    [anon_sym_GT_GT2] = ACTIONS(4121),
    [anon_sym_PLUS] = ACTIONS(4121),
    [anon_sym_DASH] = ACTIONS(4121),
    [anon_sym_DOT] = ACTIONS(4121),
    [anon_sym_SLASH] = ACTIONS(4121),
    [anon_sym_x] = ACTIONS(4121),
    [anon_sym_EQ_TILDE] = ACTIONS(4119),
    [anon_sym_BANG_TILDE] = ACTIONS(4119),
    [anon_sym_EQ_EQ] = ACTIONS(4121),
    [anon_sym_BANG_EQ] = ACTIONS(4119),
    [anon_sym_eq] = ACTIONS(4121),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4119),
    [anon_sym_equ] = ACTIONS(4119),
    [anon_sym_eqr] = ACTIONS(4119),
    [anon_sym_ne] = ACTIONS(4119),
    [anon_sym_LT_EQ_GT] = ACTIONS(4119),
    [anon_sym_cmp] = ACTIONS(4119),
    [anon_sym_TILDE_TILDE] = ACTIONS(4119),
    [anon_sym_LT] = ACTIONS(4121),
    [anon_sym_LT_EQ] = ACTIONS(4121),
    [anon_sym_GT_EQ] = ACTIONS(4119),
    [anon_sym_lt] = ACTIONS(4119),
    [anon_sym_le] = ACTIONS(4119),
    [anon_sym_ge] = ACTIONS(4119),
    [anon_sym_gt] = ACTIONS(4119),
    [anon_sym_isa] = ACTIONS(4119),
    [anon_sym_PLUS_PLUS] = ACTIONS(4119),
    [anon_sym_DASH_DASH] = ACTIONS(4119),
    [anon_sym_QMARK] = ACTIONS(4119),
    [anon_sym_EQ_GT] = ACTIONS(4119),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4119),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3565)] = {
    [sym_heredoc_content] = STATE(3565),
    [anon_sym_SEMI] = ACTIONS(4131),
    [anon_sym_EQ] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_COMMA] = ACTIONS(4131),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4131),
    [anon_sym_DASH_GT] = ACTIONS(4131),
    [anon_sym_STAR] = ACTIONS(4133),
    [anon_sym_PERCENT] = ACTIONS(4133),
    [anon_sym_GT] = ACTIONS(4133),
    [anon_sym_LT_LT] = ACTIONS(4133),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4131),
    [anon_sym_PLUS_EQ] = ACTIONS(4131),
    [anon_sym_DASH_EQ] = ACTIONS(4131),
    [anon_sym_DOT_EQ] = ACTIONS(4131),
    [anon_sym_STAR_EQ] = ACTIONS(4131),
    [anon_sym_SLASH_EQ] = ACTIONS(4131),
    [anon_sym_PERCENT_EQ] = ACTIONS(4131),
    [anon_sym_x_EQ] = ACTIONS(4131),
    [anon_sym_AMP_EQ] = ACTIONS(4131),
    [anon_sym_PIPE_EQ] = ACTIONS(4131),
    [anon_sym_CARET_EQ] = ACTIONS(4131),
    [anon_sym_LT_LT_EQ] = ACTIONS(4131),
    [anon_sym_GT_GT_EQ] = ACTIONS(4131),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4131),
    [anon_sym_DOT_DOT] = ACTIONS(4133),
    [anon_sym_STAR_STAR] = ACTIONS(4133),
    [anon_sym_PIPE_PIPE] = ACTIONS(4133),
    [anon_sym_SLASH_SLASH] = ACTIONS(4133),
    [anon_sym_CARET_CARET] = ACTIONS(4131),
    [anon_sym_AMP_AMP] = ACTIONS(4133),
    [anon_sym_PIPE] = ACTIONS(4133),
    [anon_sym_CARET] = ACTIONS(4133),
    [anon_sym_AMP] = ACTIONS(4133),
    [anon_sym_GT_GT2] = ACTIONS(4133),
    [anon_sym_PLUS] = ACTIONS(4133),
    [anon_sym_DASH] = ACTIONS(4133),
    [anon_sym_DOT] = ACTIONS(4133),
    [anon_sym_SLASH] = ACTIONS(4133),
    [anon_sym_x] = ACTIONS(4133),
    [anon_sym_EQ_TILDE] = ACTIONS(4131),
    [anon_sym_BANG_TILDE] = ACTIONS(4131),
    [anon_sym_EQ_EQ] = ACTIONS(4133),
    [anon_sym_BANG_EQ] = ACTIONS(4131),
    [anon_sym_eq] = ACTIONS(4133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4131),
    [anon_sym_equ] = ACTIONS(4131),
    [anon_sym_eqr] = ACTIONS(4131),
    [anon_sym_ne] = ACTIONS(4131),
    [anon_sym_LT_EQ_GT] = ACTIONS(4131),
    [anon_sym_cmp] = ACTIONS(4131),
    [anon_sym_TILDE_TILDE] = ACTIONS(4131),
    [anon_sym_LT] = ACTIONS(4133),
    [anon_sym_LT_EQ] = ACTIONS(4133),
    [anon_sym_GT_EQ] = ACTIONS(4131),
    [anon_sym_lt] = ACTIONS(4131),
    [anon_sym_le] = ACTIONS(4131),
    [anon_sym_ge] = ACTIONS(4131),
    [anon_sym_gt] = ACTIONS(4131),
    [anon_sym_isa] = ACTIONS(4131),
    [anon_sym_PLUS_PLUS] = ACTIONS(4131),
    [anon_sym_DASH_DASH] = ACTIONS(4131),
    [anon_sym_QMARK] = ACTIONS(4131),
    [anon_sym_EQ_GT] = ACTIONS(4131),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4131),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3566)] = {
    [sym_heredoc_content] = STATE(3566),
    [anon_sym_SEMI] = ACTIONS(3881),
    [anon_sym_EQ] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_COMMA] = ACTIONS(3881),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3881),
    [anon_sym_DASH_GT] = ACTIONS(3881),
    [anon_sym_STAR] = ACTIONS(3883),
    [anon_sym_PERCENT] = ACTIONS(3883),
    [anon_sym_GT] = ACTIONS(3883),
    [anon_sym_LT_LT] = ACTIONS(3883),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3881),
    [anon_sym_PLUS_EQ] = ACTIONS(3881),
    [anon_sym_DASH_EQ] = ACTIONS(3881),
    [anon_sym_DOT_EQ] = ACTIONS(3881),
    [anon_sym_STAR_EQ] = ACTIONS(3881),
    [anon_sym_SLASH_EQ] = ACTIONS(3881),
    [anon_sym_PERCENT_EQ] = ACTIONS(3881),
    [anon_sym_x_EQ] = ACTIONS(3881),
    [anon_sym_AMP_EQ] = ACTIONS(3881),
    [anon_sym_PIPE_EQ] = ACTIONS(3881),
    [anon_sym_CARET_EQ] = ACTIONS(3881),
    [anon_sym_LT_LT_EQ] = ACTIONS(3881),
    [anon_sym_GT_GT_EQ] = ACTIONS(3881),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3881),
    [anon_sym_DOT_DOT] = ACTIONS(3883),
    [anon_sym_STAR_STAR] = ACTIONS(3883),
    [anon_sym_PIPE_PIPE] = ACTIONS(3883),
    [anon_sym_SLASH_SLASH] = ACTIONS(3883),
    [anon_sym_CARET_CARET] = ACTIONS(3881),
    [anon_sym_AMP_AMP] = ACTIONS(3883),
    [anon_sym_PIPE] = ACTIONS(3883),
    [anon_sym_CARET] = ACTIONS(3883),
    [anon_sym_AMP] = ACTIONS(3883),
    [anon_sym_GT_GT2] = ACTIONS(3883),
    [anon_sym_PLUS] = ACTIONS(3883),
    [anon_sym_DASH] = ACTIONS(3883),
    [anon_sym_DOT] = ACTIONS(3883),
    [anon_sym_SLASH] = ACTIONS(3883),
    [anon_sym_x] = ACTIONS(3883),
    [anon_sym_EQ_TILDE] = ACTIONS(3881),
    [anon_sym_BANG_TILDE] = ACTIONS(3881),
    [anon_sym_EQ_EQ] = ACTIONS(3883),
    [anon_sym_BANG_EQ] = ACTIONS(3881),
    [anon_sym_eq] = ACTIONS(3883),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3881),
    [anon_sym_equ] = ACTIONS(3881),
    [anon_sym_eqr] = ACTIONS(3881),
    [anon_sym_ne] = ACTIONS(3881),
    [anon_sym_LT_EQ_GT] = ACTIONS(3881),
    [anon_sym_cmp] = ACTIONS(3881),
    [anon_sym_TILDE_TILDE] = ACTIONS(3881),
    [anon_sym_LT] = ACTIONS(3883),
    [anon_sym_LT_EQ] = ACTIONS(3883),
    [anon_sym_GT_EQ] = ACTIONS(3881),
    [anon_sym_lt] = ACTIONS(3881),
    [anon_sym_le] = ACTIONS(3881),
    [anon_sym_ge] = ACTIONS(3881),
    [anon_sym_gt] = ACTIONS(3881),
    [anon_sym_isa] = ACTIONS(3881),
    [anon_sym_PLUS_PLUS] = ACTIONS(3881),
    [anon_sym_DASH_DASH] = ACTIONS(3881),
    [anon_sym_QMARK] = ACTIONS(3881),
    [anon_sym_EQ_GT] = ACTIONS(3881),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3881),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3567)] = {
    [sym_heredoc_content] = STATE(3567),
    [anon_sym_SEMI] = ACTIONS(4075),
    [anon_sym_EQ] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_COMMA] = ACTIONS(4075),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4075),
    [anon_sym_DASH_GT] = ACTIONS(4075),
    [anon_sym_STAR] = ACTIONS(4077),
    [anon_sym_PERCENT] = ACTIONS(4077),
    [anon_sym_GT] = ACTIONS(4077),
    [anon_sym_LT_LT] = ACTIONS(4077),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4075),
    [anon_sym_PLUS_EQ] = ACTIONS(4075),
    [anon_sym_DASH_EQ] = ACTIONS(4075),
    [anon_sym_DOT_EQ] = ACTIONS(4075),
    [anon_sym_STAR_EQ] = ACTIONS(4075),
    [anon_sym_SLASH_EQ] = ACTIONS(4075),
    [anon_sym_PERCENT_EQ] = ACTIONS(4075),
    [anon_sym_x_EQ] = ACTIONS(4075),
    [anon_sym_AMP_EQ] = ACTIONS(4075),
    [anon_sym_PIPE_EQ] = ACTIONS(4075),
    [anon_sym_CARET_EQ] = ACTIONS(4075),
    [anon_sym_LT_LT_EQ] = ACTIONS(4075),
    [anon_sym_GT_GT_EQ] = ACTIONS(4075),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4075),
    [anon_sym_DOT_DOT] = ACTIONS(4077),
    [anon_sym_STAR_STAR] = ACTIONS(4077),
    [anon_sym_PIPE_PIPE] = ACTIONS(4077),
    [anon_sym_SLASH_SLASH] = ACTIONS(4077),
    [anon_sym_CARET_CARET] = ACTIONS(4075),
    [anon_sym_AMP_AMP] = ACTIONS(4077),
    [anon_sym_PIPE] = ACTIONS(4077),
    [anon_sym_CARET] = ACTIONS(4077),
    [anon_sym_AMP] = ACTIONS(4077),
    [anon_sym_GT_GT2] = ACTIONS(4077),
    [anon_sym_PLUS] = ACTIONS(4077),
    [anon_sym_DASH] = ACTIONS(4077),
    [anon_sym_DOT] = ACTIONS(4077),
    [anon_sym_SLASH] = ACTIONS(4077),
    [anon_sym_x] = ACTIONS(4077),
    [anon_sym_EQ_TILDE] = ACTIONS(4075),
    [anon_sym_BANG_TILDE] = ACTIONS(4075),
    [anon_sym_EQ_EQ] = ACTIONS(4077),
    [anon_sym_BANG_EQ] = ACTIONS(4075),
    [anon_sym_eq] = ACTIONS(4077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4075),
    [anon_sym_equ] = ACTIONS(4075),
    [anon_sym_eqr] = ACTIONS(4075),
    [anon_sym_ne] = ACTIONS(4075),
    [anon_sym_LT_EQ_GT] = ACTIONS(4075),
    [anon_sym_cmp] = ACTIONS(4075),
    [anon_sym_TILDE_TILDE] = ACTIONS(4075),
    [anon_sym_LT] = ACTIONS(4077),
    [anon_sym_LT_EQ] = ACTIONS(4077),
    [anon_sym_GT_EQ] = ACTIONS(4075),
    [anon_sym_lt] = ACTIONS(4075),
    [anon_sym_le] = ACTIONS(4075),
    [anon_sym_ge] = ACTIONS(4075),
    [anon_sym_gt] = ACTIONS(4075),
    [anon_sym_isa] = ACTIONS(4075),
    [anon_sym_PLUS_PLUS] = ACTIONS(4075),
    [anon_sym_DASH_DASH] = ACTIONS(4075),
    [anon_sym_QMARK] = ACTIONS(4075),
    [anon_sym_EQ_GT] = ACTIONS(4075),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4075),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3568)] = {
    [sym_heredoc_content] = STATE(3568),
    [anon_sym_SEMI] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(3819),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3819),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3569)] = {
    [sym_heredoc_content] = STATE(3569),
    [anon_sym_SEMI] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5377),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5377),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3819),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3570)] = {
    [sym_heredoc_content] = STATE(3570),
    [anon_sym_SEMI] = ACTIONS(3885),
    [anon_sym_EQ] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_COMMA] = ACTIONS(3885),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3885),
    [anon_sym_DASH_GT] = ACTIONS(3885),
    [anon_sym_STAR] = ACTIONS(3887),
    [anon_sym_PERCENT] = ACTIONS(3887),
    [anon_sym_GT] = ACTIONS(3887),
    [anon_sym_LT_LT] = ACTIONS(3887),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3885),
    [anon_sym_PLUS_EQ] = ACTIONS(3885),
    [anon_sym_DASH_EQ] = ACTIONS(3885),
    [anon_sym_DOT_EQ] = ACTIONS(3885),
    [anon_sym_STAR_EQ] = ACTIONS(3885),
    [anon_sym_SLASH_EQ] = ACTIONS(3885),
    [anon_sym_PERCENT_EQ] = ACTIONS(3885),
    [anon_sym_x_EQ] = ACTIONS(3885),
    [anon_sym_AMP_EQ] = ACTIONS(3885),
    [anon_sym_PIPE_EQ] = ACTIONS(3885),
    [anon_sym_CARET_EQ] = ACTIONS(3885),
    [anon_sym_LT_LT_EQ] = ACTIONS(3885),
    [anon_sym_GT_GT_EQ] = ACTIONS(3885),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3885),
    [anon_sym_DOT_DOT] = ACTIONS(3887),
    [anon_sym_STAR_STAR] = ACTIONS(3887),
    [anon_sym_PIPE_PIPE] = ACTIONS(3887),
    [anon_sym_SLASH_SLASH] = ACTIONS(3887),
    [anon_sym_CARET_CARET] = ACTIONS(3885),
    [anon_sym_AMP_AMP] = ACTIONS(3887),
    [anon_sym_PIPE] = ACTIONS(3887),
    [anon_sym_CARET] = ACTIONS(3887),
    [anon_sym_AMP] = ACTIONS(3887),
    [anon_sym_GT_GT2] = ACTIONS(3887),
    [anon_sym_PLUS] = ACTIONS(3887),
    [anon_sym_DASH] = ACTIONS(3887),
    [anon_sym_DOT] = ACTIONS(3887),
    [anon_sym_SLASH] = ACTIONS(3887),
    [anon_sym_x] = ACTIONS(3887),
    [anon_sym_EQ_TILDE] = ACTIONS(3885),
    [anon_sym_BANG_TILDE] = ACTIONS(3885),
    [anon_sym_EQ_EQ] = ACTIONS(3887),
    [anon_sym_BANG_EQ] = ACTIONS(3885),
    [anon_sym_eq] = ACTIONS(3887),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3885),
    [anon_sym_equ] = ACTIONS(3885),
    [anon_sym_eqr] = ACTIONS(3885),
    [anon_sym_ne] = ACTIONS(3885),
    [anon_sym_LT_EQ_GT] = ACTIONS(3885),
    [anon_sym_cmp] = ACTIONS(3885),
    [anon_sym_TILDE_TILDE] = ACTIONS(3885),
    [anon_sym_LT] = ACTIONS(3887),
    [anon_sym_LT_EQ] = ACTIONS(3887),
    [anon_sym_GT_EQ] = ACTIONS(3885),
    [anon_sym_lt] = ACTIONS(3885),
    [anon_sym_le] = ACTIONS(3885),
    [anon_sym_ge] = ACTIONS(3885),
    [anon_sym_gt] = ACTIONS(3885),
    [anon_sym_isa] = ACTIONS(3885),
    [anon_sym_PLUS_PLUS] = ACTIONS(3885),
    [anon_sym_DASH_DASH] = ACTIONS(3885),
    [anon_sym_QMARK] = ACTIONS(3885),
    [anon_sym_EQ_GT] = ACTIONS(3885),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3885),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3571)] = {
    [sym_heredoc_content] = STATE(3571),
    [anon_sym_SEMI] = ACTIONS(3889),
    [anon_sym_EQ] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_COMMA] = ACTIONS(3889),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3889),
    [anon_sym_DASH_GT] = ACTIONS(3889),
    [anon_sym_STAR] = ACTIONS(3891),
    [anon_sym_PERCENT] = ACTIONS(3891),
    [anon_sym_GT] = ACTIONS(3891),
    [anon_sym_LT_LT] = ACTIONS(3891),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3889),
    [anon_sym_PLUS_EQ] = ACTIONS(3889),
    [anon_sym_DASH_EQ] = ACTIONS(3889),
    [anon_sym_DOT_EQ] = ACTIONS(3889),
    [anon_sym_STAR_EQ] = ACTIONS(3889),
    [anon_sym_SLASH_EQ] = ACTIONS(3889),
    [anon_sym_PERCENT_EQ] = ACTIONS(3889),
    [anon_sym_x_EQ] = ACTIONS(3889),
    [anon_sym_AMP_EQ] = ACTIONS(3889),
    [anon_sym_PIPE_EQ] = ACTIONS(3889),
    [anon_sym_CARET_EQ] = ACTIONS(3889),
    [anon_sym_LT_LT_EQ] = ACTIONS(3889),
    [anon_sym_GT_GT_EQ] = ACTIONS(3889),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3889),
    [anon_sym_DOT_DOT] = ACTIONS(3891),
    [anon_sym_STAR_STAR] = ACTIONS(3891),
    [anon_sym_PIPE_PIPE] = ACTIONS(3891),
    [anon_sym_SLASH_SLASH] = ACTIONS(3891),
    [anon_sym_CARET_CARET] = ACTIONS(3889),
    [anon_sym_AMP_AMP] = ACTIONS(3891),
    [anon_sym_PIPE] = ACTIONS(3891),
    [anon_sym_CARET] = ACTIONS(3891),
    [anon_sym_AMP] = ACTIONS(3891),
    [anon_sym_GT_GT2] = ACTIONS(3891),
    [anon_sym_PLUS] = ACTIONS(3891),
    [anon_sym_DASH] = ACTIONS(3891),
    [anon_sym_DOT] = ACTIONS(3891),
    [anon_sym_SLASH] = ACTIONS(3891),
    [anon_sym_x] = ACTIONS(3891),
    [anon_sym_EQ_TILDE] = ACTIONS(3889),
    [anon_sym_BANG_TILDE] = ACTIONS(3889),
    [anon_sym_EQ_EQ] = ACTIONS(3891),
    [anon_sym_BANG_EQ] = ACTIONS(3889),
    [anon_sym_eq] = ACTIONS(3891),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3889),
    [anon_sym_equ] = ACTIONS(3889),
    [anon_sym_eqr] = ACTIONS(3889),
    [anon_sym_ne] = ACTIONS(3889),
    [anon_sym_LT_EQ_GT] = ACTIONS(3889),
    [anon_sym_cmp] = ACTIONS(3889),
    [anon_sym_TILDE_TILDE] = ACTIONS(3889),
    [anon_sym_LT] = ACTIONS(3891),
    [anon_sym_LT_EQ] = ACTIONS(3891),
    [anon_sym_GT_EQ] = ACTIONS(3889),
    [anon_sym_lt] = ACTIONS(3889),
    [anon_sym_le] = ACTIONS(3889),
    [anon_sym_ge] = ACTIONS(3889),
    [anon_sym_gt] = ACTIONS(3889),
    [anon_sym_isa] = ACTIONS(3889),
    [anon_sym_PLUS_PLUS] = ACTIONS(3889),
    [anon_sym_DASH_DASH] = ACTIONS(3889),
    [anon_sym_QMARK] = ACTIONS(3889),
    [anon_sym_EQ_GT] = ACTIONS(3889),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3889),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3572)] = {
    [sym_heredoc_content] = STATE(3572),
    [anon_sym_SEMI] = ACTIONS(3973),
    [anon_sym_EQ] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_COMMA] = ACTIONS(3973),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3973),
    [anon_sym_DASH_GT] = ACTIONS(3973),
    [anon_sym_STAR] = ACTIONS(3975),
    [anon_sym_PERCENT] = ACTIONS(3975),
    [anon_sym_GT] = ACTIONS(3975),
    [anon_sym_LT_LT] = ACTIONS(3975),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3973),
    [anon_sym_PLUS_EQ] = ACTIONS(3973),
    [anon_sym_DASH_EQ] = ACTIONS(3973),
    [anon_sym_DOT_EQ] = ACTIONS(3973),
    [anon_sym_STAR_EQ] = ACTIONS(3973),
    [anon_sym_SLASH_EQ] = ACTIONS(3973),
    [anon_sym_PERCENT_EQ] = ACTIONS(3973),
    [anon_sym_x_EQ] = ACTIONS(3973),
    [anon_sym_AMP_EQ] = ACTIONS(3973),
    [anon_sym_PIPE_EQ] = ACTIONS(3973),
    [anon_sym_CARET_EQ] = ACTIONS(3973),
    [anon_sym_LT_LT_EQ] = ACTIONS(3973),
    [anon_sym_GT_GT_EQ] = ACTIONS(3973),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3973),
    [anon_sym_DOT_DOT] = ACTIONS(3975),
    [anon_sym_STAR_STAR] = ACTIONS(3975),
    [anon_sym_PIPE_PIPE] = ACTIONS(3975),
    [anon_sym_SLASH_SLASH] = ACTIONS(3975),
    [anon_sym_CARET_CARET] = ACTIONS(3973),
    [anon_sym_AMP_AMP] = ACTIONS(3975),
    [anon_sym_PIPE] = ACTIONS(3975),
    [anon_sym_CARET] = ACTIONS(3975),
    [anon_sym_AMP] = ACTIONS(3975),
    [anon_sym_GT_GT2] = ACTIONS(3975),
    [anon_sym_PLUS] = ACTIONS(3975),
    [anon_sym_DASH] = ACTIONS(3975),
    [anon_sym_DOT] = ACTIONS(3975),
    [anon_sym_SLASH] = ACTIONS(3975),
    [anon_sym_x] = ACTIONS(3975),
    [anon_sym_EQ_TILDE] = ACTIONS(3973),
    [anon_sym_BANG_TILDE] = ACTIONS(3973),
    [anon_sym_EQ_EQ] = ACTIONS(3975),
    [anon_sym_BANG_EQ] = ACTIONS(3973),
    [anon_sym_eq] = ACTIONS(3975),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3973),
    [anon_sym_equ] = ACTIONS(3973),
    [anon_sym_eqr] = ACTIONS(3973),
    [anon_sym_ne] = ACTIONS(3973),
    [anon_sym_LT_EQ_GT] = ACTIONS(3973),
    [anon_sym_cmp] = ACTIONS(3973),
    [anon_sym_TILDE_TILDE] = ACTIONS(3973),
    [anon_sym_LT] = ACTIONS(3975),
    [anon_sym_LT_EQ] = ACTIONS(3975),
    [anon_sym_GT_EQ] = ACTIONS(3973),
    [anon_sym_lt] = ACTIONS(3973),
    [anon_sym_le] = ACTIONS(3973),
    [anon_sym_ge] = ACTIONS(3973),
    [anon_sym_gt] = ACTIONS(3973),
    [anon_sym_isa] = ACTIONS(3973),
    [anon_sym_PLUS_PLUS] = ACTIONS(3973),
    [anon_sym_DASH_DASH] = ACTIONS(3973),
    [anon_sym_QMARK] = ACTIONS(3973),
    [anon_sym_EQ_GT] = ACTIONS(3973),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3973),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3573)] = {
    [sym_heredoc_content] = STATE(3573),
    [anon_sym_SEMI] = ACTIONS(4021),
    [anon_sym_EQ] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_COMMA] = ACTIONS(4021),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4021),
    [anon_sym_DASH_GT] = ACTIONS(4021),
    [anon_sym_STAR] = ACTIONS(4023),
    [anon_sym_PERCENT] = ACTIONS(4023),
    [anon_sym_GT] = ACTIONS(4023),
    [anon_sym_LT_LT] = ACTIONS(4023),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4021),
    [anon_sym_PLUS_EQ] = ACTIONS(4021),
    [anon_sym_DASH_EQ] = ACTIONS(4021),
    [anon_sym_DOT_EQ] = ACTIONS(4021),
    [anon_sym_STAR_EQ] = ACTIONS(4021),
    [anon_sym_SLASH_EQ] = ACTIONS(4021),
    [anon_sym_PERCENT_EQ] = ACTIONS(4021),
    [anon_sym_x_EQ] = ACTIONS(4021),
    [anon_sym_AMP_EQ] = ACTIONS(4021),
    [anon_sym_PIPE_EQ] = ACTIONS(4021),
    [anon_sym_CARET_EQ] = ACTIONS(4021),
    [anon_sym_LT_LT_EQ] = ACTIONS(4021),
    [anon_sym_GT_GT_EQ] = ACTIONS(4021),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4021),
    [anon_sym_DOT_DOT] = ACTIONS(4023),
    [anon_sym_STAR_STAR] = ACTIONS(4023),
    [anon_sym_PIPE_PIPE] = ACTIONS(4023),
    [anon_sym_SLASH_SLASH] = ACTIONS(4023),
    [anon_sym_CARET_CARET] = ACTIONS(4021),
    [anon_sym_AMP_AMP] = ACTIONS(4023),
    [anon_sym_PIPE] = ACTIONS(4023),
    [anon_sym_CARET] = ACTIONS(4023),
    [anon_sym_AMP] = ACTIONS(4023),
    [anon_sym_GT_GT2] = ACTIONS(4023),
    [anon_sym_PLUS] = ACTIONS(4023),
    [anon_sym_DASH] = ACTIONS(4023),
    [anon_sym_DOT] = ACTIONS(4023),
    [anon_sym_SLASH] = ACTIONS(4023),
    [anon_sym_x] = ACTIONS(4023),
    [anon_sym_EQ_TILDE] = ACTIONS(4021),
    [anon_sym_BANG_TILDE] = ACTIONS(4021),
    [anon_sym_EQ_EQ] = ACTIONS(4023),
    [anon_sym_BANG_EQ] = ACTIONS(4021),
    [anon_sym_eq] = ACTIONS(4023),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4021),
    [anon_sym_equ] = ACTIONS(4021),
    [anon_sym_eqr] = ACTIONS(4021),
    [anon_sym_ne] = ACTIONS(4021),
    [anon_sym_LT_EQ_GT] = ACTIONS(4021),
    [anon_sym_cmp] = ACTIONS(4021),
    [anon_sym_TILDE_TILDE] = ACTIONS(4021),
    [anon_sym_LT] = ACTIONS(4023),
    [anon_sym_LT_EQ] = ACTIONS(4023),
    [anon_sym_GT_EQ] = ACTIONS(4021),
    [anon_sym_lt] = ACTIONS(4021),
    [anon_sym_le] = ACTIONS(4021),
    [anon_sym_ge] = ACTIONS(4021),
    [anon_sym_gt] = ACTIONS(4021),
    [anon_sym_isa] = ACTIONS(4021),
    [anon_sym_PLUS_PLUS] = ACTIONS(4021),
    [anon_sym_DASH_DASH] = ACTIONS(4021),
    [anon_sym_QMARK] = ACTIONS(4021),
    [anon_sym_EQ_GT] = ACTIONS(4021),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4021),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3574)] = {
    [sym_heredoc_content] = STATE(3574),
    [anon_sym_SEMI] = ACTIONS(4033),
    [anon_sym_EQ] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_COMMA] = ACTIONS(4033),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4033),
    [anon_sym_DASH_GT] = ACTIONS(4033),
    [anon_sym_STAR] = ACTIONS(4035),
    [anon_sym_PERCENT] = ACTIONS(4035),
    [anon_sym_GT] = ACTIONS(4035),
    [anon_sym_LT_LT] = ACTIONS(4035),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4033),
    [anon_sym_PLUS_EQ] = ACTIONS(4033),
    [anon_sym_DASH_EQ] = ACTIONS(4033),
    [anon_sym_DOT_EQ] = ACTIONS(4033),
    [anon_sym_STAR_EQ] = ACTIONS(4033),
    [anon_sym_SLASH_EQ] = ACTIONS(4033),
    [anon_sym_PERCENT_EQ] = ACTIONS(4033),
    [anon_sym_x_EQ] = ACTIONS(4033),
    [anon_sym_AMP_EQ] = ACTIONS(4033),
    [anon_sym_PIPE_EQ] = ACTIONS(4033),
    [anon_sym_CARET_EQ] = ACTIONS(4033),
    [anon_sym_LT_LT_EQ] = ACTIONS(4033),
    [anon_sym_GT_GT_EQ] = ACTIONS(4033),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4033),
    [anon_sym_DOT_DOT] = ACTIONS(4035),
    [anon_sym_STAR_STAR] = ACTIONS(4035),
    [anon_sym_PIPE_PIPE] = ACTIONS(4035),
    [anon_sym_SLASH_SLASH] = ACTIONS(4035),
    [anon_sym_CARET_CARET] = ACTIONS(4033),
    [anon_sym_AMP_AMP] = ACTIONS(4035),
    [anon_sym_PIPE] = ACTIONS(4035),
    [anon_sym_CARET] = ACTIONS(4035),
    [anon_sym_AMP] = ACTIONS(4035),
    [anon_sym_GT_GT2] = ACTIONS(4035),
    [anon_sym_PLUS] = ACTIONS(4035),
    [anon_sym_DASH] = ACTIONS(4035),
    [anon_sym_DOT] = ACTIONS(4035),
    [anon_sym_SLASH] = ACTIONS(4035),
    [anon_sym_x] = ACTIONS(4035),
    [anon_sym_EQ_TILDE] = ACTIONS(4033),
    [anon_sym_BANG_TILDE] = ACTIONS(4033),
    [anon_sym_EQ_EQ] = ACTIONS(4035),
    [anon_sym_BANG_EQ] = ACTIONS(4033),
    [anon_sym_eq] = ACTIONS(4035),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4033),
    [anon_sym_equ] = ACTIONS(4033),
    [anon_sym_eqr] = ACTIONS(4033),
    [anon_sym_ne] = ACTIONS(4033),
    [anon_sym_LT_EQ_GT] = ACTIONS(4033),
    [anon_sym_cmp] = ACTIONS(4033),
    [anon_sym_TILDE_TILDE] = ACTIONS(4033),
    [anon_sym_LT] = ACTIONS(4035),
    [anon_sym_LT_EQ] = ACTIONS(4035),
    [anon_sym_GT_EQ] = ACTIONS(4033),
    [anon_sym_lt] = ACTIONS(4033),
    [anon_sym_le] = ACTIONS(4033),
    [anon_sym_ge] = ACTIONS(4033),
    [anon_sym_gt] = ACTIONS(4033),
    [anon_sym_isa] = ACTIONS(4033),
    [anon_sym_PLUS_PLUS] = ACTIONS(4033),
    [anon_sym_DASH_DASH] = ACTIONS(4033),
    [anon_sym_QMARK] = ACTIONS(4033),
    [anon_sym_EQ_GT] = ACTIONS(4033),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4033),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3575)] = {
    [sym_heredoc_content] = STATE(3575),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3576)] = {
    [sym_heredoc_content] = STATE(3576),
    [anon_sym_SEMI] = ACTIONS(3747),
    [anon_sym_EQ] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_COMMA] = ACTIONS(3747),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3747),
    [anon_sym_DASH_GT] = ACTIONS(3747),
    [anon_sym_STAR] = ACTIONS(3749),
    [anon_sym_PERCENT] = ACTIONS(3749),
    [anon_sym_GT] = ACTIONS(3749),
    [anon_sym_LT_LT] = ACTIONS(3749),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3747),
    [anon_sym_PLUS_EQ] = ACTIONS(3747),
    [anon_sym_DASH_EQ] = ACTIONS(3747),
    [anon_sym_DOT_EQ] = ACTIONS(3747),
    [anon_sym_STAR_EQ] = ACTIONS(3747),
    [anon_sym_SLASH_EQ] = ACTIONS(3747),
    [anon_sym_PERCENT_EQ] = ACTIONS(3747),
    [anon_sym_x_EQ] = ACTIONS(3747),
    [anon_sym_AMP_EQ] = ACTIONS(3747),
    [anon_sym_PIPE_EQ] = ACTIONS(3747),
    [anon_sym_CARET_EQ] = ACTIONS(3747),
    [anon_sym_LT_LT_EQ] = ACTIONS(3747),
    [anon_sym_GT_GT_EQ] = ACTIONS(3747),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3747),
    [anon_sym_DOT_DOT] = ACTIONS(3749),
    [anon_sym_STAR_STAR] = ACTIONS(3749),
    [anon_sym_PIPE_PIPE] = ACTIONS(3749),
    [anon_sym_SLASH_SLASH] = ACTIONS(3749),
    [anon_sym_CARET_CARET] = ACTIONS(3747),
    [anon_sym_AMP_AMP] = ACTIONS(3749),
    [anon_sym_PIPE] = ACTIONS(3749),
    [anon_sym_CARET] = ACTIONS(3749),
    [anon_sym_AMP] = ACTIONS(3749),
    [anon_sym_GT_GT2] = ACTIONS(3749),
    [anon_sym_PLUS] = ACTIONS(3749),
    [anon_sym_DASH] = ACTIONS(3749),
    [anon_sym_DOT] = ACTIONS(3749),
    [anon_sym_SLASH] = ACTIONS(3749),
    [anon_sym_x] = ACTIONS(3749),
    [anon_sym_EQ_TILDE] = ACTIONS(3747),
    [anon_sym_BANG_TILDE] = ACTIONS(3747),
    [anon_sym_EQ_EQ] = ACTIONS(3749),
    [anon_sym_BANG_EQ] = ACTIONS(3747),
    [anon_sym_eq] = ACTIONS(3749),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3747),
    [anon_sym_equ] = ACTIONS(3747),
    [anon_sym_eqr] = ACTIONS(3747),
    [anon_sym_ne] = ACTIONS(3747),
    [anon_sym_LT_EQ_GT] = ACTIONS(3747),
    [anon_sym_cmp] = ACTIONS(3747),
    [anon_sym_TILDE_TILDE] = ACTIONS(3747),
    [anon_sym_LT] = ACTIONS(3749),
    [anon_sym_LT_EQ] = ACTIONS(3749),
    [anon_sym_GT_EQ] = ACTIONS(3747),
    [anon_sym_lt] = ACTIONS(3747),
    [anon_sym_le] = ACTIONS(3747),
    [anon_sym_ge] = ACTIONS(3747),
    [anon_sym_gt] = ACTIONS(3747),
    [anon_sym_isa] = ACTIONS(3747),
    [anon_sym_PLUS_PLUS] = ACTIONS(3747),
    [anon_sym_DASH_DASH] = ACTIONS(3747),
    [anon_sym_QMARK] = ACTIONS(3747),
    [anon_sym_EQ_GT] = ACTIONS(3747),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3747),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3577)] = {
    [sym_heredoc_content] = STATE(3577),
    [anon_sym_SEMI] = ACTIONS(3779),
    [anon_sym_EQ] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_COMMA] = ACTIONS(3779),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3779),
    [anon_sym_DASH_GT] = ACTIONS(3779),
    [anon_sym_STAR] = ACTIONS(3781),
    [anon_sym_PERCENT] = ACTIONS(3781),
    [anon_sym_GT] = ACTIONS(3781),
    [anon_sym_LT_LT] = ACTIONS(3781),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3779),
    [anon_sym_PLUS_EQ] = ACTIONS(3779),
    [anon_sym_DASH_EQ] = ACTIONS(3779),
    [anon_sym_DOT_EQ] = ACTIONS(3779),
    [anon_sym_STAR_EQ] = ACTIONS(3779),
    [anon_sym_SLASH_EQ] = ACTIONS(3779),
    [anon_sym_PERCENT_EQ] = ACTIONS(3779),
    [anon_sym_x_EQ] = ACTIONS(3779),
    [anon_sym_AMP_EQ] = ACTIONS(3779),
    [anon_sym_PIPE_EQ] = ACTIONS(3779),
    [anon_sym_CARET_EQ] = ACTIONS(3779),
    [anon_sym_LT_LT_EQ] = ACTIONS(3779),
    [anon_sym_GT_GT_EQ] = ACTIONS(3779),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3779),
    [anon_sym_DOT_DOT] = ACTIONS(3781),
    [anon_sym_STAR_STAR] = ACTIONS(3781),
    [anon_sym_PIPE_PIPE] = ACTIONS(3781),
    [anon_sym_SLASH_SLASH] = ACTIONS(3781),
    [anon_sym_CARET_CARET] = ACTIONS(3779),
    [anon_sym_AMP_AMP] = ACTIONS(3781),
    [anon_sym_PIPE] = ACTIONS(3781),
    [anon_sym_CARET] = ACTIONS(3781),
    [anon_sym_AMP] = ACTIONS(3781),
    [anon_sym_GT_GT2] = ACTIONS(3781),
    [anon_sym_PLUS] = ACTIONS(3781),
    [anon_sym_DASH] = ACTIONS(3781),
    [anon_sym_DOT] = ACTIONS(3781),
    [anon_sym_SLASH] = ACTIONS(3781),
    [anon_sym_x] = ACTIONS(3781),
    [anon_sym_EQ_TILDE] = ACTIONS(3779),
    [anon_sym_BANG_TILDE] = ACTIONS(3779),
    [anon_sym_EQ_EQ] = ACTIONS(3781),
    [anon_sym_BANG_EQ] = ACTIONS(3779),
    [anon_sym_eq] = ACTIONS(3781),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3779),
    [anon_sym_equ] = ACTIONS(3779),
    [anon_sym_eqr] = ACTIONS(3779),
    [anon_sym_ne] = ACTIONS(3779),
    [anon_sym_LT_EQ_GT] = ACTIONS(3779),
    [anon_sym_cmp] = ACTIONS(3779),
    [anon_sym_TILDE_TILDE] = ACTIONS(3779),
    [anon_sym_LT] = ACTIONS(3781),
    [anon_sym_LT_EQ] = ACTIONS(3781),
    [anon_sym_GT_EQ] = ACTIONS(3779),
    [anon_sym_lt] = ACTIONS(3779),
    [anon_sym_le] = ACTIONS(3779),
    [anon_sym_ge] = ACTIONS(3779),
    [anon_sym_gt] = ACTIONS(3779),
    [anon_sym_isa] = ACTIONS(3779),
    [anon_sym_PLUS_PLUS] = ACTIONS(3779),
    [anon_sym_DASH_DASH] = ACTIONS(3779),
    [anon_sym_QMARK] = ACTIONS(3779),
    [anon_sym_EQ_GT] = ACTIONS(3779),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3779),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3578)] = {
    [sym_heredoc_content] = STATE(3578),
    [anon_sym_SEMI] = ACTIONS(3833),
    [anon_sym_EQ] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_COMMA] = ACTIONS(3833),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3833),
    [anon_sym_DASH_GT] = ACTIONS(3833),
    [anon_sym_STAR] = ACTIONS(3835),
    [anon_sym_PERCENT] = ACTIONS(3835),
    [anon_sym_GT] = ACTIONS(3835),
    [anon_sym_LT_LT] = ACTIONS(3835),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3833),
    [anon_sym_PLUS_EQ] = ACTIONS(3833),
    [anon_sym_DASH_EQ] = ACTIONS(3833),
    [anon_sym_DOT_EQ] = ACTIONS(3833),
    [anon_sym_STAR_EQ] = ACTIONS(3833),
    [anon_sym_SLASH_EQ] = ACTIONS(3833),
    [anon_sym_PERCENT_EQ] = ACTIONS(3833),
    [anon_sym_x_EQ] = ACTIONS(3833),
    [anon_sym_AMP_EQ] = ACTIONS(3833),
    [anon_sym_PIPE_EQ] = ACTIONS(3833),
    [anon_sym_CARET_EQ] = ACTIONS(3833),
    [anon_sym_LT_LT_EQ] = ACTIONS(3833),
    [anon_sym_GT_GT_EQ] = ACTIONS(3833),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3833),
    [anon_sym_DOT_DOT] = ACTIONS(3835),
    [anon_sym_STAR_STAR] = ACTIONS(3835),
    [anon_sym_PIPE_PIPE] = ACTIONS(3835),
    [anon_sym_SLASH_SLASH] = ACTIONS(3835),
    [anon_sym_CARET_CARET] = ACTIONS(3833),
    [anon_sym_AMP_AMP] = ACTIONS(3835),
    [anon_sym_PIPE] = ACTIONS(3835),
    [anon_sym_CARET] = ACTIONS(3835),
    [anon_sym_AMP] = ACTIONS(3835),
    [anon_sym_GT_GT2] = ACTIONS(3835),
    [anon_sym_PLUS] = ACTIONS(3835),
    [anon_sym_DASH] = ACTIONS(3835),
    [anon_sym_DOT] = ACTIONS(3835),
    [anon_sym_SLASH] = ACTIONS(3835),
    [anon_sym_x] = ACTIONS(3835),
    [anon_sym_EQ_TILDE] = ACTIONS(3833),
    [anon_sym_BANG_TILDE] = ACTIONS(3833),
    [anon_sym_EQ_EQ] = ACTIONS(3835),
    [anon_sym_BANG_EQ] = ACTIONS(3833),
    [anon_sym_eq] = ACTIONS(3835),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3833),
    [anon_sym_equ] = ACTIONS(3833),
    [anon_sym_eqr] = ACTIONS(3833),
    [anon_sym_ne] = ACTIONS(3833),
    [anon_sym_LT_EQ_GT] = ACTIONS(3833),
    [anon_sym_cmp] = ACTIONS(3833),
    [anon_sym_TILDE_TILDE] = ACTIONS(3833),
    [anon_sym_LT] = ACTIONS(3835),
    [anon_sym_LT_EQ] = ACTIONS(3835),
    [anon_sym_GT_EQ] = ACTIONS(3833),
    [anon_sym_lt] = ACTIONS(3833),
    [anon_sym_le] = ACTIONS(3833),
    [anon_sym_ge] = ACTIONS(3833),
    [anon_sym_gt] = ACTIONS(3833),
    [anon_sym_isa] = ACTIONS(3833),
    [anon_sym_PLUS_PLUS] = ACTIONS(3833),
    [anon_sym_DASH_DASH] = ACTIONS(3833),
    [anon_sym_QMARK] = ACTIONS(3833),
    [anon_sym_EQ_GT] = ACTIONS(3833),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3833),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3579)] = {
    [sym_heredoc_content] = STATE(3579),
    [anon_sym_SEMI] = ACTIONS(3841),
    [anon_sym_EQ] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_COMMA] = ACTIONS(3841),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3841),
    [anon_sym_DASH_GT] = ACTIONS(3841),
    [anon_sym_STAR] = ACTIONS(3843),
    [anon_sym_PERCENT] = ACTIONS(3843),
    [anon_sym_GT] = ACTIONS(3843),
    [anon_sym_LT_LT] = ACTIONS(3843),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3841),
    [anon_sym_PLUS_EQ] = ACTIONS(3841),
    [anon_sym_DASH_EQ] = ACTIONS(3841),
    [anon_sym_DOT_EQ] = ACTIONS(3841),
    [anon_sym_STAR_EQ] = ACTIONS(3841),
    [anon_sym_SLASH_EQ] = ACTIONS(3841),
    [anon_sym_PERCENT_EQ] = ACTIONS(3841),
    [anon_sym_x_EQ] = ACTIONS(3841),
    [anon_sym_AMP_EQ] = ACTIONS(3841),
    [anon_sym_PIPE_EQ] = ACTIONS(3841),
    [anon_sym_CARET_EQ] = ACTIONS(3841),
    [anon_sym_LT_LT_EQ] = ACTIONS(3841),
    [anon_sym_GT_GT_EQ] = ACTIONS(3841),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3841),
    [anon_sym_DOT_DOT] = ACTIONS(3843),
    [anon_sym_STAR_STAR] = ACTIONS(3843),
    [anon_sym_PIPE_PIPE] = ACTIONS(3843),
    [anon_sym_SLASH_SLASH] = ACTIONS(3843),
    [anon_sym_CARET_CARET] = ACTIONS(3841),
    [anon_sym_AMP_AMP] = ACTIONS(3843),
    [anon_sym_PIPE] = ACTIONS(3843),
    [anon_sym_CARET] = ACTIONS(3843),
    [anon_sym_AMP] = ACTIONS(3843),
    [anon_sym_GT_GT2] = ACTIONS(3843),
    [anon_sym_PLUS] = ACTIONS(3843),
    [anon_sym_DASH] = ACTIONS(3843),
    [anon_sym_DOT] = ACTIONS(3843),
    [anon_sym_SLASH] = ACTIONS(3843),
    [anon_sym_x] = ACTIONS(3843),
    [anon_sym_EQ_TILDE] = ACTIONS(3841),
    [anon_sym_BANG_TILDE] = ACTIONS(3841),
    [anon_sym_EQ_EQ] = ACTIONS(3843),
    [anon_sym_BANG_EQ] = ACTIONS(3841),
    [anon_sym_eq] = ACTIONS(3843),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3841),
    [anon_sym_equ] = ACTIONS(3841),
    [anon_sym_eqr] = ACTIONS(3841),
    [anon_sym_ne] = ACTIONS(3841),
    [anon_sym_LT_EQ_GT] = ACTIONS(3841),
    [anon_sym_cmp] = ACTIONS(3841),
    [anon_sym_TILDE_TILDE] = ACTIONS(3841),
    [anon_sym_LT] = ACTIONS(3843),
    [anon_sym_LT_EQ] = ACTIONS(3843),
    [anon_sym_GT_EQ] = ACTIONS(3841),
    [anon_sym_lt] = ACTIONS(3841),
    [anon_sym_le] = ACTIONS(3841),
    [anon_sym_ge] = ACTIONS(3841),
    [anon_sym_gt] = ACTIONS(3841),
    [anon_sym_isa] = ACTIONS(3841),
    [anon_sym_PLUS_PLUS] = ACTIONS(3841),
    [anon_sym_DASH_DASH] = ACTIONS(3841),
    [anon_sym_QMARK] = ACTIONS(3841),
    [anon_sym_EQ_GT] = ACTIONS(3841),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3841),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3580)] = {
    [sym_heredoc_content] = STATE(3580),
    [anon_sym_SEMI] = ACTIONS(3853),
    [anon_sym_EQ] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_COMMA] = ACTIONS(3853),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3853),
    [anon_sym_DASH_GT] = ACTIONS(3853),
    [anon_sym_STAR] = ACTIONS(3855),
    [anon_sym_PERCENT] = ACTIONS(3855),
    [anon_sym_GT] = ACTIONS(3855),
    [anon_sym_LT_LT] = ACTIONS(3855),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3853),
    [anon_sym_PLUS_EQ] = ACTIONS(3853),
    [anon_sym_DASH_EQ] = ACTIONS(3853),
    [anon_sym_DOT_EQ] = ACTIONS(3853),
    [anon_sym_STAR_EQ] = ACTIONS(3853),
    [anon_sym_SLASH_EQ] = ACTIONS(3853),
    [anon_sym_PERCENT_EQ] = ACTIONS(3853),
    [anon_sym_x_EQ] = ACTIONS(3853),
    [anon_sym_AMP_EQ] = ACTIONS(3853),
    [anon_sym_PIPE_EQ] = ACTIONS(3853),
    [anon_sym_CARET_EQ] = ACTIONS(3853),
    [anon_sym_LT_LT_EQ] = ACTIONS(3853),
    [anon_sym_GT_GT_EQ] = ACTIONS(3853),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3853),
    [anon_sym_DOT_DOT] = ACTIONS(3855),
    [anon_sym_STAR_STAR] = ACTIONS(3855),
    [anon_sym_PIPE_PIPE] = ACTIONS(3855),
    [anon_sym_SLASH_SLASH] = ACTIONS(3855),
    [anon_sym_CARET_CARET] = ACTIONS(3853),
    [anon_sym_AMP_AMP] = ACTIONS(3855),
    [anon_sym_PIPE] = ACTIONS(3855),
    [anon_sym_CARET] = ACTIONS(3855),
    [anon_sym_AMP] = ACTIONS(3855),
    [anon_sym_GT_GT2] = ACTIONS(3855),
    [anon_sym_PLUS] = ACTIONS(3855),
    [anon_sym_DASH] = ACTIONS(3855),
    [anon_sym_DOT] = ACTIONS(3855),
    [anon_sym_SLASH] = ACTIONS(3855),
    [anon_sym_x] = ACTIONS(3855),
    [anon_sym_EQ_TILDE] = ACTIONS(3853),
    [anon_sym_BANG_TILDE] = ACTIONS(3853),
    [anon_sym_EQ_EQ] = ACTIONS(3855),
    [anon_sym_BANG_EQ] = ACTIONS(3853),
    [anon_sym_eq] = ACTIONS(3855),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3853),
    [anon_sym_equ] = ACTIONS(3853),
    [anon_sym_eqr] = ACTIONS(3853),
    [anon_sym_ne] = ACTIONS(3853),
    [anon_sym_LT_EQ_GT] = ACTIONS(3853),
    [anon_sym_cmp] = ACTIONS(3853),
    [anon_sym_TILDE_TILDE] = ACTIONS(3853),
    [anon_sym_LT] = ACTIONS(3855),
    [anon_sym_LT_EQ] = ACTIONS(3855),
    [anon_sym_GT_EQ] = ACTIONS(3853),
    [anon_sym_lt] = ACTIONS(3853),
    [anon_sym_le] = ACTIONS(3853),
    [anon_sym_ge] = ACTIONS(3853),
    [anon_sym_gt] = ACTIONS(3853),
    [anon_sym_isa] = ACTIONS(3853),
    [anon_sym_PLUS_PLUS] = ACTIONS(3853),
    [anon_sym_DASH_DASH] = ACTIONS(3853),
    [anon_sym_QMARK] = ACTIONS(3853),
    [anon_sym_EQ_GT] = ACTIONS(3853),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3853),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3581)] = {
    [sym_heredoc_content] = STATE(3581),
    [anon_sym_SEMI] = ACTIONS(3861),
    [anon_sym_EQ] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_COMMA] = ACTIONS(3861),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3861),
    [anon_sym_DASH_GT] = ACTIONS(3861),
    [anon_sym_STAR] = ACTIONS(3863),
    [anon_sym_PERCENT] = ACTIONS(3863),
    [anon_sym_GT] = ACTIONS(3863),
    [anon_sym_LT_LT] = ACTIONS(3863),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3861),
    [anon_sym_PLUS_EQ] = ACTIONS(3861),
    [anon_sym_DASH_EQ] = ACTIONS(3861),
    [anon_sym_DOT_EQ] = ACTIONS(3861),
    [anon_sym_STAR_EQ] = ACTIONS(3861),
    [anon_sym_SLASH_EQ] = ACTIONS(3861),
    [anon_sym_PERCENT_EQ] = ACTIONS(3861),
    [anon_sym_x_EQ] = ACTIONS(3861),
    [anon_sym_AMP_EQ] = ACTIONS(3861),
    [anon_sym_PIPE_EQ] = ACTIONS(3861),
    [anon_sym_CARET_EQ] = ACTIONS(3861),
    [anon_sym_LT_LT_EQ] = ACTIONS(3861),
    [anon_sym_GT_GT_EQ] = ACTIONS(3861),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3861),
    [anon_sym_DOT_DOT] = ACTIONS(3863),
    [anon_sym_STAR_STAR] = ACTIONS(3863),
    [anon_sym_PIPE_PIPE] = ACTIONS(3863),
    [anon_sym_SLASH_SLASH] = ACTIONS(3863),
    [anon_sym_CARET_CARET] = ACTIONS(3861),
    [anon_sym_AMP_AMP] = ACTIONS(3863),
    [anon_sym_PIPE] = ACTIONS(3863),
    [anon_sym_CARET] = ACTIONS(3863),
    [anon_sym_AMP] = ACTIONS(3863),
    [anon_sym_GT_GT2] = ACTIONS(3863),
    [anon_sym_PLUS] = ACTIONS(3863),
    [anon_sym_DASH] = ACTIONS(3863),
    [anon_sym_DOT] = ACTIONS(3863),
    [anon_sym_SLASH] = ACTIONS(3863),
    [anon_sym_x] = ACTIONS(3863),
    [anon_sym_EQ_TILDE] = ACTIONS(3861),
    [anon_sym_BANG_TILDE] = ACTIONS(3861),
    [anon_sym_EQ_EQ] = ACTIONS(3863),
    [anon_sym_BANG_EQ] = ACTIONS(3861),
    [anon_sym_eq] = ACTIONS(3863),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3861),
    [anon_sym_equ] = ACTIONS(3861),
    [anon_sym_eqr] = ACTIONS(3861),
    [anon_sym_ne] = ACTIONS(3861),
    [anon_sym_LT_EQ_GT] = ACTIONS(3861),
    [anon_sym_cmp] = ACTIONS(3861),
    [anon_sym_TILDE_TILDE] = ACTIONS(3861),
    [anon_sym_LT] = ACTIONS(3863),
    [anon_sym_LT_EQ] = ACTIONS(3863),
    [anon_sym_GT_EQ] = ACTIONS(3861),
    [anon_sym_lt] = ACTIONS(3861),
    [anon_sym_le] = ACTIONS(3861),
    [anon_sym_ge] = ACTIONS(3861),
    [anon_sym_gt] = ACTIONS(3861),
    [anon_sym_isa] = ACTIONS(3861),
    [anon_sym_PLUS_PLUS] = ACTIONS(3861),
    [anon_sym_DASH_DASH] = ACTIONS(3861),
    [anon_sym_QMARK] = ACTIONS(3861),
    [anon_sym_EQ_GT] = ACTIONS(3861),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3861),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3582)] = {
    [sym_heredoc_content] = STATE(3582),
    [anon_sym_SEMI] = ACTIONS(3837),
    [anon_sym_EQ] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_COMMA] = ACTIONS(3837),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3837),
    [anon_sym_DASH_GT] = ACTIONS(3837),
    [anon_sym_STAR] = ACTIONS(3839),
    [anon_sym_PERCENT] = ACTIONS(3839),
    [anon_sym_GT] = ACTIONS(3839),
    [anon_sym_LT_LT] = ACTIONS(3839),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3837),
    [anon_sym_PLUS_EQ] = ACTIONS(3837),
    [anon_sym_DASH_EQ] = ACTIONS(3837),
    [anon_sym_DOT_EQ] = ACTIONS(3837),
    [anon_sym_STAR_EQ] = ACTIONS(3837),
    [anon_sym_SLASH_EQ] = ACTIONS(3837),
    [anon_sym_PERCENT_EQ] = ACTIONS(3837),
    [anon_sym_x_EQ] = ACTIONS(3837),
    [anon_sym_AMP_EQ] = ACTIONS(3837),
    [anon_sym_PIPE_EQ] = ACTIONS(3837),
    [anon_sym_CARET_EQ] = ACTIONS(3837),
    [anon_sym_LT_LT_EQ] = ACTIONS(3837),
    [anon_sym_GT_GT_EQ] = ACTIONS(3837),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3837),
    [anon_sym_DOT_DOT] = ACTIONS(3839),
    [anon_sym_STAR_STAR] = ACTIONS(3839),
    [anon_sym_PIPE_PIPE] = ACTIONS(3839),
    [anon_sym_SLASH_SLASH] = ACTIONS(3839),
    [anon_sym_CARET_CARET] = ACTIONS(3837),
    [anon_sym_AMP_AMP] = ACTIONS(3839),
    [anon_sym_PIPE] = ACTIONS(3839),
    [anon_sym_CARET] = ACTIONS(3839),
    [anon_sym_AMP] = ACTIONS(3839),
    [anon_sym_GT_GT2] = ACTIONS(3839),
    [anon_sym_PLUS] = ACTIONS(3839),
    [anon_sym_DASH] = ACTIONS(3839),
    [anon_sym_DOT] = ACTIONS(3839),
    [anon_sym_SLASH] = ACTIONS(3839),
    [anon_sym_x] = ACTIONS(3839),
    [anon_sym_EQ_TILDE] = ACTIONS(3837),
    [anon_sym_BANG_TILDE] = ACTIONS(3837),
    [anon_sym_EQ_EQ] = ACTIONS(3839),
    [anon_sym_BANG_EQ] = ACTIONS(3837),
    [anon_sym_eq] = ACTIONS(3839),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3837),
    [anon_sym_equ] = ACTIONS(3837),
    [anon_sym_eqr] = ACTIONS(3837),
    [anon_sym_ne] = ACTIONS(3837),
    [anon_sym_LT_EQ_GT] = ACTIONS(3837),
    [anon_sym_cmp] = ACTIONS(3837),
    [anon_sym_TILDE_TILDE] = ACTIONS(3837),
    [anon_sym_LT] = ACTIONS(3839),
    [anon_sym_LT_EQ] = ACTIONS(3839),
    [anon_sym_GT_EQ] = ACTIONS(3837),
    [anon_sym_lt] = ACTIONS(3837),
    [anon_sym_le] = ACTIONS(3837),
    [anon_sym_ge] = ACTIONS(3837),
    [anon_sym_gt] = ACTIONS(3837),
    [anon_sym_isa] = ACTIONS(3837),
    [anon_sym_PLUS_PLUS] = ACTIONS(3837),
    [anon_sym_DASH_DASH] = ACTIONS(3837),
    [anon_sym_QMARK] = ACTIONS(3837),
    [anon_sym_EQ_GT] = ACTIONS(3837),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3837),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3583)] = {
    [sym_heredoc_content] = STATE(3583),
    [anon_sym_SEMI] = ACTIONS(3993),
    [anon_sym_EQ] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_COMMA] = ACTIONS(3993),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3993),
    [anon_sym_DASH_GT] = ACTIONS(3993),
    [anon_sym_STAR] = ACTIONS(3995),
    [anon_sym_PERCENT] = ACTIONS(3995),
    [anon_sym_GT] = ACTIONS(3995),
    [anon_sym_LT_LT] = ACTIONS(3995),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3993),
    [anon_sym_PLUS_EQ] = ACTIONS(3993),
    [anon_sym_DASH_EQ] = ACTIONS(3993),
    [anon_sym_DOT_EQ] = ACTIONS(3993),
    [anon_sym_STAR_EQ] = ACTIONS(3993),
    [anon_sym_SLASH_EQ] = ACTIONS(3993),
    [anon_sym_PERCENT_EQ] = ACTIONS(3993),
    [anon_sym_x_EQ] = ACTIONS(3993),
    [anon_sym_AMP_EQ] = ACTIONS(3993),
    [anon_sym_PIPE_EQ] = ACTIONS(3993),
    [anon_sym_CARET_EQ] = ACTIONS(3993),
    [anon_sym_LT_LT_EQ] = ACTIONS(3993),
    [anon_sym_GT_GT_EQ] = ACTIONS(3993),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3993),
    [anon_sym_DOT_DOT] = ACTIONS(3995),
    [anon_sym_STAR_STAR] = ACTIONS(3995),
    [anon_sym_PIPE_PIPE] = ACTIONS(3995),
    [anon_sym_SLASH_SLASH] = ACTIONS(3995),
    [anon_sym_CARET_CARET] = ACTIONS(3993),
    [anon_sym_AMP_AMP] = ACTIONS(3995),
    [anon_sym_PIPE] = ACTIONS(3995),
    [anon_sym_CARET] = ACTIONS(3995),
    [anon_sym_AMP] = ACTIONS(3995),
    [anon_sym_GT_GT2] = ACTIONS(3995),
    [anon_sym_PLUS] = ACTIONS(3995),
    [anon_sym_DASH] = ACTIONS(3995),
    [anon_sym_DOT] = ACTIONS(3995),
    [anon_sym_SLASH] = ACTIONS(3995),
    [anon_sym_x] = ACTIONS(3995),
    [anon_sym_EQ_TILDE] = ACTIONS(3993),
    [anon_sym_BANG_TILDE] = ACTIONS(3993),
    [anon_sym_EQ_EQ] = ACTIONS(3995),
    [anon_sym_BANG_EQ] = ACTIONS(3993),
    [anon_sym_eq] = ACTIONS(3995),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3993),
    [anon_sym_equ] = ACTIONS(3993),
    [anon_sym_eqr] = ACTIONS(3993),
    [anon_sym_ne] = ACTIONS(3993),
    [anon_sym_LT_EQ_GT] = ACTIONS(3993),
    [anon_sym_cmp] = ACTIONS(3993),
    [anon_sym_TILDE_TILDE] = ACTIONS(3993),
    [anon_sym_LT] = ACTIONS(3995),
    [anon_sym_LT_EQ] = ACTIONS(3995),
    [anon_sym_GT_EQ] = ACTIONS(3993),
    [anon_sym_lt] = ACTIONS(3993),
    [anon_sym_le] = ACTIONS(3993),
    [anon_sym_ge] = ACTIONS(3993),
    [anon_sym_gt] = ACTIONS(3993),
    [anon_sym_isa] = ACTIONS(3993),
    [anon_sym_PLUS_PLUS] = ACTIONS(3993),
    [anon_sym_DASH_DASH] = ACTIONS(3993),
    [anon_sym_QMARK] = ACTIONS(3993),
    [anon_sym_EQ_GT] = ACTIONS(3993),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3993),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3584)] = {
    [sym_heredoc_content] = STATE(3584),
    [anon_sym_SEMI] = ACTIONS(4005),
    [anon_sym_EQ] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_COMMA] = ACTIONS(4005),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4005),
    [anon_sym_DASH_GT] = ACTIONS(4005),
    [anon_sym_STAR] = ACTIONS(4007),
    [anon_sym_PERCENT] = ACTIONS(4007),
    [anon_sym_GT] = ACTIONS(4007),
    [anon_sym_LT_LT] = ACTIONS(4007),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4005),
    [anon_sym_PLUS_EQ] = ACTIONS(4005),
    [anon_sym_DASH_EQ] = ACTIONS(4005),
    [anon_sym_DOT_EQ] = ACTIONS(4005),
    [anon_sym_STAR_EQ] = ACTIONS(4005),
    [anon_sym_SLASH_EQ] = ACTIONS(4005),
    [anon_sym_PERCENT_EQ] = ACTIONS(4005),
    [anon_sym_x_EQ] = ACTIONS(4005),
    [anon_sym_AMP_EQ] = ACTIONS(4005),
    [anon_sym_PIPE_EQ] = ACTIONS(4005),
    [anon_sym_CARET_EQ] = ACTIONS(4005),
    [anon_sym_LT_LT_EQ] = ACTIONS(4005),
    [anon_sym_GT_GT_EQ] = ACTIONS(4005),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4005),
    [anon_sym_DOT_DOT] = ACTIONS(4007),
    [anon_sym_STAR_STAR] = ACTIONS(4007),
    [anon_sym_PIPE_PIPE] = ACTIONS(4007),
    [anon_sym_SLASH_SLASH] = ACTIONS(4007),
    [anon_sym_CARET_CARET] = ACTIONS(4005),
    [anon_sym_AMP_AMP] = ACTIONS(4007),
    [anon_sym_PIPE] = ACTIONS(4007),
    [anon_sym_CARET] = ACTIONS(4007),
    [anon_sym_AMP] = ACTIONS(4007),
    [anon_sym_GT_GT2] = ACTIONS(4007),
    [anon_sym_PLUS] = ACTIONS(4007),
    [anon_sym_DASH] = ACTIONS(4007),
    [anon_sym_DOT] = ACTIONS(4007),
    [anon_sym_SLASH] = ACTIONS(4007),
    [anon_sym_x] = ACTIONS(4007),
    [anon_sym_EQ_TILDE] = ACTIONS(4005),
    [anon_sym_BANG_TILDE] = ACTIONS(4005),
    [anon_sym_EQ_EQ] = ACTIONS(4007),
    [anon_sym_BANG_EQ] = ACTIONS(4005),
    [anon_sym_eq] = ACTIONS(4007),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4005),
    [anon_sym_equ] = ACTIONS(4005),
    [anon_sym_eqr] = ACTIONS(4005),
    [anon_sym_ne] = ACTIONS(4005),
    [anon_sym_LT_EQ_GT] = ACTIONS(4005),
    [anon_sym_cmp] = ACTIONS(4005),
    [anon_sym_TILDE_TILDE] = ACTIONS(4005),
    [anon_sym_LT] = ACTIONS(4007),
    [anon_sym_LT_EQ] = ACTIONS(4007),
    [anon_sym_GT_EQ] = ACTIONS(4005),
    [anon_sym_lt] = ACTIONS(4005),
    [anon_sym_le] = ACTIONS(4005),
    [anon_sym_ge] = ACTIONS(4005),
    [anon_sym_gt] = ACTIONS(4005),
    [anon_sym_isa] = ACTIONS(4005),
    [anon_sym_PLUS_PLUS] = ACTIONS(4005),
    [anon_sym_DASH_DASH] = ACTIONS(4005),
    [anon_sym_QMARK] = ACTIONS(4005),
    [anon_sym_EQ_GT] = ACTIONS(4005),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4005),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3585)] = {
    [sym_heredoc_content] = STATE(3585),
    [anon_sym_SEMI] = ACTIONS(4087),
    [anon_sym_EQ] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_COMMA] = ACTIONS(4087),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4087),
    [anon_sym_DASH_GT] = ACTIONS(4087),
    [anon_sym_STAR] = ACTIONS(4089),
    [anon_sym_PERCENT] = ACTIONS(4089),
    [anon_sym_GT] = ACTIONS(4089),
    [anon_sym_LT_LT] = ACTIONS(4089),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4087),
    [anon_sym_PLUS_EQ] = ACTIONS(4087),
    [anon_sym_DASH_EQ] = ACTIONS(4087),
    [anon_sym_DOT_EQ] = ACTIONS(4087),
    [anon_sym_STAR_EQ] = ACTIONS(4087),
    [anon_sym_SLASH_EQ] = ACTIONS(4087),
    [anon_sym_PERCENT_EQ] = ACTIONS(4087),
    [anon_sym_x_EQ] = ACTIONS(4087),
    [anon_sym_AMP_EQ] = ACTIONS(4087),
    [anon_sym_PIPE_EQ] = ACTIONS(4087),
    [anon_sym_CARET_EQ] = ACTIONS(4087),
    [anon_sym_LT_LT_EQ] = ACTIONS(4087),
    [anon_sym_GT_GT_EQ] = ACTIONS(4087),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4087),
    [anon_sym_DOT_DOT] = ACTIONS(4089),
    [anon_sym_STAR_STAR] = ACTIONS(4089),
    [anon_sym_PIPE_PIPE] = ACTIONS(4089),
    [anon_sym_SLASH_SLASH] = ACTIONS(4089),
    [anon_sym_CARET_CARET] = ACTIONS(4087),
    [anon_sym_AMP_AMP] = ACTIONS(4089),
    [anon_sym_PIPE] = ACTIONS(4089),
    [anon_sym_CARET] = ACTIONS(4089),
    [anon_sym_AMP] = ACTIONS(4089),
    [anon_sym_GT_GT2] = ACTIONS(4089),
    [anon_sym_PLUS] = ACTIONS(4089),
    [anon_sym_DASH] = ACTIONS(4089),
    [anon_sym_DOT] = ACTIONS(4089),
    [anon_sym_SLASH] = ACTIONS(4089),
    [anon_sym_x] = ACTIONS(4089),
    [anon_sym_EQ_TILDE] = ACTIONS(4087),
    [anon_sym_BANG_TILDE] = ACTIONS(4087),
    [anon_sym_EQ_EQ] = ACTIONS(4089),
    [anon_sym_BANG_EQ] = ACTIONS(4087),
    [anon_sym_eq] = ACTIONS(4089),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4087),
    [anon_sym_equ] = ACTIONS(4087),
    [anon_sym_eqr] = ACTIONS(4087),
    [anon_sym_ne] = ACTIONS(4087),
    [anon_sym_LT_EQ_GT] = ACTIONS(4087),
    [anon_sym_cmp] = ACTIONS(4087),
    [anon_sym_TILDE_TILDE] = ACTIONS(4087),
    [anon_sym_LT] = ACTIONS(4089),
    [anon_sym_LT_EQ] = ACTIONS(4089),
    [anon_sym_GT_EQ] = ACTIONS(4087),
    [anon_sym_lt] = ACTIONS(4087),
    [anon_sym_le] = ACTIONS(4087),
    [anon_sym_ge] = ACTIONS(4087),
    [anon_sym_gt] = ACTIONS(4087),
    [anon_sym_isa] = ACTIONS(4087),
    [anon_sym_PLUS_PLUS] = ACTIONS(4087),
    [anon_sym_DASH_DASH] = ACTIONS(4087),
    [anon_sym_QMARK] = ACTIONS(4087),
    [anon_sym_EQ_GT] = ACTIONS(4087),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4087),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3586)] = {
    [sym_heredoc_content] = STATE(3586),
    [anon_sym_SEMI] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4107),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3587)] = {
    [sym_heredoc_content] = STATE(3587),
    [anon_sym_SEMI] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(5252),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5254),
    [anon_sym_PLUS_EQ] = ACTIONS(5254),
    [anon_sym_DASH_EQ] = ACTIONS(5254),
    [anon_sym_DOT_EQ] = ACTIONS(5254),
    [anon_sym_STAR_EQ] = ACTIONS(5254),
    [anon_sym_SLASH_EQ] = ACTIONS(5254),
    [anon_sym_PERCENT_EQ] = ACTIONS(5254),
    [anon_sym_x_EQ] = ACTIONS(5254),
    [anon_sym_AMP_EQ] = ACTIONS(5254),
    [anon_sym_PIPE_EQ] = ACTIONS(5254),
    [anon_sym_CARET_EQ] = ACTIONS(5254),
    [anon_sym_LT_LT_EQ] = ACTIONS(5254),
    [anon_sym_GT_GT_EQ] = ACTIONS(5254),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5254),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3590),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3588)] = {
    [sym_heredoc_content] = STATE(3588),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5379),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5381),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3589)] = {
    [sym_heredoc_content] = STATE(3589),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3590)] = {
    [sym_heredoc_content] = STATE(3590),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3591)] = {
    [sym_heredoc_content] = STATE(3591),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3592)] = {
    [sym_heredoc_content] = STATE(3592),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3593)] = {
    [sym_heredoc_content] = STATE(3593),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3594)] = {
    [sym_heredoc_content] = STATE(3594),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3595)] = {
    [sym_heredoc_content] = STATE(3595),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3596)] = {
    [sym_heredoc_content] = STATE(3596),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3597)] = {
    [sym_heredoc_content] = STATE(3597),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3598)] = {
    [sym_heredoc_content] = STATE(3598),
    [anon_sym_SEMI] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3719),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3599)] = {
    [sym_heredoc_content] = STATE(3599),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3600)] = {
    [sym_heredoc_content] = STATE(3600),
    [anon_sym_SEMI] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5383),
    [anon_sym_cmp] = ACTIONS(5383),
    [anon_sym_TILDE_TILDE] = ACTIONS(5383),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3763),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3601)] = {
    [sym_heredoc_content] = STATE(3601),
    [anon_sym_SEMI] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5385),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3735),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3602)] = {
    [sym_heredoc_content] = STATE(3602),
    [anon_sym_SEMI] = ACTIONS(3771),
    [anon_sym_EQ] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_COMMA] = ACTIONS(3771),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3771),
    [anon_sym_DASH_GT] = ACTIONS(3771),
    [anon_sym_STAR] = ACTIONS(3773),
    [anon_sym_PERCENT] = ACTIONS(3773),
    [anon_sym_GT] = ACTIONS(3773),
    [anon_sym_LT_LT] = ACTIONS(3773),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3771),
    [anon_sym_PLUS_EQ] = ACTIONS(3771),
    [anon_sym_DASH_EQ] = ACTIONS(3771),
    [anon_sym_DOT_EQ] = ACTIONS(3771),
    [anon_sym_STAR_EQ] = ACTIONS(3771),
    [anon_sym_SLASH_EQ] = ACTIONS(3771),
    [anon_sym_PERCENT_EQ] = ACTIONS(3771),
    [anon_sym_x_EQ] = ACTIONS(3771),
    [anon_sym_AMP_EQ] = ACTIONS(3771),
    [anon_sym_PIPE_EQ] = ACTIONS(3771),
    [anon_sym_CARET_EQ] = ACTIONS(3771),
    [anon_sym_LT_LT_EQ] = ACTIONS(3771),
    [anon_sym_GT_GT_EQ] = ACTIONS(3771),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3771),
    [anon_sym_DOT_DOT] = ACTIONS(3773),
    [anon_sym_STAR_STAR] = ACTIONS(3773),
    [anon_sym_PIPE_PIPE] = ACTIONS(3773),
    [anon_sym_SLASH_SLASH] = ACTIONS(3773),
    [anon_sym_CARET_CARET] = ACTIONS(3771),
    [anon_sym_AMP_AMP] = ACTIONS(3773),
    [anon_sym_PIPE] = ACTIONS(3773),
    [anon_sym_CARET] = ACTIONS(3773),
    [anon_sym_AMP] = ACTIONS(3773),
    [anon_sym_GT_GT2] = ACTIONS(3773),
    [anon_sym_PLUS] = ACTIONS(3773),
    [anon_sym_DASH] = ACTIONS(3773),
    [anon_sym_DOT] = ACTIONS(3773),
    [anon_sym_SLASH] = ACTIONS(3773),
    [anon_sym_x] = ACTIONS(3773),
    [anon_sym_EQ_TILDE] = ACTIONS(3771),
    [anon_sym_BANG_TILDE] = ACTIONS(3771),
    [anon_sym_EQ_EQ] = ACTIONS(3773),
    [anon_sym_BANG_EQ] = ACTIONS(3771),
    [anon_sym_eq] = ACTIONS(3773),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3771),
    [anon_sym_equ] = ACTIONS(3771),
    [anon_sym_eqr] = ACTIONS(3771),
    [anon_sym_ne] = ACTIONS(3771),
    [anon_sym_LT_EQ_GT] = ACTIONS(3771),
    [anon_sym_cmp] = ACTIONS(3771),
    [anon_sym_TILDE_TILDE] = ACTIONS(3771),
    [anon_sym_LT] = ACTIONS(3773),
    [anon_sym_LT_EQ] = ACTIONS(3773),
    [anon_sym_GT_EQ] = ACTIONS(3771),
    [anon_sym_lt] = ACTIONS(3771),
    [anon_sym_le] = ACTIONS(3771),
    [anon_sym_ge] = ACTIONS(3771),
    [anon_sym_gt] = ACTIONS(3771),
    [anon_sym_isa] = ACTIONS(3771),
    [anon_sym_PLUS_PLUS] = ACTIONS(3771),
    [anon_sym_DASH_DASH] = ACTIONS(3771),
    [anon_sym_QMARK] = ACTIONS(3771),
    [anon_sym_EQ_GT] = ACTIONS(3771),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3771),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3603)] = {
    [sym_heredoc_content] = STATE(3603),
    [anon_sym_SEMI] = ACTIONS(3783),
    [anon_sym_EQ] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_COMMA] = ACTIONS(3783),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3783),
    [anon_sym_DASH_GT] = ACTIONS(3783),
    [anon_sym_STAR] = ACTIONS(3785),
    [anon_sym_PERCENT] = ACTIONS(3785),
    [anon_sym_GT] = ACTIONS(3785),
    [anon_sym_LT_LT] = ACTIONS(3785),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3783),
    [anon_sym_PLUS_EQ] = ACTIONS(3783),
    [anon_sym_DASH_EQ] = ACTIONS(3783),
    [anon_sym_DOT_EQ] = ACTIONS(3783),
    [anon_sym_STAR_EQ] = ACTIONS(3783),
    [anon_sym_SLASH_EQ] = ACTIONS(3783),
    [anon_sym_PERCENT_EQ] = ACTIONS(3783),
    [anon_sym_x_EQ] = ACTIONS(3783),
    [anon_sym_AMP_EQ] = ACTIONS(3783),
    [anon_sym_PIPE_EQ] = ACTIONS(3783),
    [anon_sym_CARET_EQ] = ACTIONS(3783),
    [anon_sym_LT_LT_EQ] = ACTIONS(3783),
    [anon_sym_GT_GT_EQ] = ACTIONS(3783),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3783),
    [anon_sym_DOT_DOT] = ACTIONS(3785),
    [anon_sym_STAR_STAR] = ACTIONS(3785),
    [anon_sym_PIPE_PIPE] = ACTIONS(3785),
    [anon_sym_SLASH_SLASH] = ACTIONS(3785),
    [anon_sym_CARET_CARET] = ACTIONS(3783),
    [anon_sym_AMP_AMP] = ACTIONS(3785),
    [anon_sym_PIPE] = ACTIONS(3785),
    [anon_sym_CARET] = ACTIONS(3785),
    [anon_sym_AMP] = ACTIONS(3785),
    [anon_sym_GT_GT2] = ACTIONS(3785),
    [anon_sym_PLUS] = ACTIONS(3785),
    [anon_sym_DASH] = ACTIONS(3785),
    [anon_sym_DOT] = ACTIONS(3785),
    [anon_sym_SLASH] = ACTIONS(3785),
    [anon_sym_x] = ACTIONS(3785),
    [anon_sym_EQ_TILDE] = ACTIONS(3783),
    [anon_sym_BANG_TILDE] = ACTIONS(3783),
    [anon_sym_EQ_EQ] = ACTIONS(3785),
    [anon_sym_BANG_EQ] = ACTIONS(3783),
    [anon_sym_eq] = ACTIONS(3785),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3783),
    [anon_sym_equ] = ACTIONS(3783),
    [anon_sym_eqr] = ACTIONS(3783),
    [anon_sym_ne] = ACTIONS(3783),
    [anon_sym_LT_EQ_GT] = ACTIONS(3783),
    [anon_sym_cmp] = ACTIONS(3783),
    [anon_sym_TILDE_TILDE] = ACTIONS(3783),
    [anon_sym_LT] = ACTIONS(3785),
    [anon_sym_LT_EQ] = ACTIONS(3785),
    [anon_sym_GT_EQ] = ACTIONS(3783),
    [anon_sym_lt] = ACTIONS(3783),
    [anon_sym_le] = ACTIONS(3783),
    [anon_sym_ge] = ACTIONS(3783),
    [anon_sym_gt] = ACTIONS(3783),
    [anon_sym_isa] = ACTIONS(3783),
    [anon_sym_PLUS_PLUS] = ACTIONS(3783),
    [anon_sym_DASH_DASH] = ACTIONS(3783),
    [anon_sym_QMARK] = ACTIONS(3783),
    [anon_sym_EQ_GT] = ACTIONS(3783),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3783),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3604)] = {
    [sym_heredoc_content] = STATE(3604),
    [anon_sym_SEMI] = ACTIONS(3787),
    [anon_sym_EQ] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_COMMA] = ACTIONS(3787),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3787),
    [anon_sym_DASH_GT] = ACTIONS(3787),
    [anon_sym_STAR] = ACTIONS(3789),
    [anon_sym_PERCENT] = ACTIONS(3789),
    [anon_sym_GT] = ACTIONS(3789),
    [anon_sym_LT_LT] = ACTIONS(3789),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3787),
    [anon_sym_PLUS_EQ] = ACTIONS(3787),
    [anon_sym_DASH_EQ] = ACTIONS(3787),
    [anon_sym_DOT_EQ] = ACTIONS(3787),
    [anon_sym_STAR_EQ] = ACTIONS(3787),
    [anon_sym_SLASH_EQ] = ACTIONS(3787),
    [anon_sym_PERCENT_EQ] = ACTIONS(3787),
    [anon_sym_x_EQ] = ACTIONS(3787),
    [anon_sym_AMP_EQ] = ACTIONS(3787),
    [anon_sym_PIPE_EQ] = ACTIONS(3787),
    [anon_sym_CARET_EQ] = ACTIONS(3787),
    [anon_sym_LT_LT_EQ] = ACTIONS(3787),
    [anon_sym_GT_GT_EQ] = ACTIONS(3787),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3787),
    [anon_sym_DOT_DOT] = ACTIONS(3789),
    [anon_sym_STAR_STAR] = ACTIONS(3789),
    [anon_sym_PIPE_PIPE] = ACTIONS(3789),
    [anon_sym_SLASH_SLASH] = ACTIONS(3789),
    [anon_sym_CARET_CARET] = ACTIONS(3787),
    [anon_sym_AMP_AMP] = ACTIONS(3789),
    [anon_sym_PIPE] = ACTIONS(3789),
    [anon_sym_CARET] = ACTIONS(3789),
    [anon_sym_AMP] = ACTIONS(3789),
    [anon_sym_GT_GT2] = ACTIONS(3789),
    [anon_sym_PLUS] = ACTIONS(3789),
    [anon_sym_DASH] = ACTIONS(3789),
    [anon_sym_DOT] = ACTIONS(3789),
    [anon_sym_SLASH] = ACTIONS(3789),
    [anon_sym_x] = ACTIONS(3789),
    [anon_sym_EQ_TILDE] = ACTIONS(3787),
    [anon_sym_BANG_TILDE] = ACTIONS(3787),
    [anon_sym_EQ_EQ] = ACTIONS(3789),
    [anon_sym_BANG_EQ] = ACTIONS(3787),
    [anon_sym_eq] = ACTIONS(3789),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3787),
    [anon_sym_equ] = ACTIONS(3787),
    [anon_sym_eqr] = ACTIONS(3787),
    [anon_sym_ne] = ACTIONS(3787),
    [anon_sym_LT_EQ_GT] = ACTIONS(3787),
    [anon_sym_cmp] = ACTIONS(3787),
    [anon_sym_TILDE_TILDE] = ACTIONS(3787),
    [anon_sym_LT] = ACTIONS(3789),
    [anon_sym_LT_EQ] = ACTIONS(3789),
    [anon_sym_GT_EQ] = ACTIONS(3787),
    [anon_sym_lt] = ACTIONS(3787),
    [anon_sym_le] = ACTIONS(3787),
    [anon_sym_ge] = ACTIONS(3787),
    [anon_sym_gt] = ACTIONS(3787),
    [anon_sym_isa] = ACTIONS(3787),
    [anon_sym_PLUS_PLUS] = ACTIONS(3787),
    [anon_sym_DASH_DASH] = ACTIONS(3787),
    [anon_sym_QMARK] = ACTIONS(3787),
    [anon_sym_EQ_GT] = ACTIONS(3787),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3787),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3605)] = {
    [sym_heredoc_content] = STATE(3605),
    [anon_sym_SEMI] = ACTIONS(3791),
    [anon_sym_EQ] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_COMMA] = ACTIONS(3791),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3791),
    [anon_sym_DASH_GT] = ACTIONS(3791),
    [anon_sym_STAR] = ACTIONS(3793),
    [anon_sym_PERCENT] = ACTIONS(3793),
    [anon_sym_GT] = ACTIONS(3793),
    [anon_sym_LT_LT] = ACTIONS(3793),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3791),
    [anon_sym_PLUS_EQ] = ACTIONS(3791),
    [anon_sym_DASH_EQ] = ACTIONS(3791),
    [anon_sym_DOT_EQ] = ACTIONS(3791),
    [anon_sym_STAR_EQ] = ACTIONS(3791),
    [anon_sym_SLASH_EQ] = ACTIONS(3791),
    [anon_sym_PERCENT_EQ] = ACTIONS(3791),
    [anon_sym_x_EQ] = ACTIONS(3791),
    [anon_sym_AMP_EQ] = ACTIONS(3791),
    [anon_sym_PIPE_EQ] = ACTIONS(3791),
    [anon_sym_CARET_EQ] = ACTIONS(3791),
    [anon_sym_LT_LT_EQ] = ACTIONS(3791),
    [anon_sym_GT_GT_EQ] = ACTIONS(3791),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3791),
    [anon_sym_DOT_DOT] = ACTIONS(3793),
    [anon_sym_STAR_STAR] = ACTIONS(3793),
    [anon_sym_PIPE_PIPE] = ACTIONS(3793),
    [anon_sym_SLASH_SLASH] = ACTIONS(3793),
    [anon_sym_CARET_CARET] = ACTIONS(3791),
    [anon_sym_AMP_AMP] = ACTIONS(3793),
    [anon_sym_PIPE] = ACTIONS(3793),
    [anon_sym_CARET] = ACTIONS(3793),
    [anon_sym_AMP] = ACTIONS(3793),
    [anon_sym_GT_GT2] = ACTIONS(3793),
    [anon_sym_PLUS] = ACTIONS(3793),
    [anon_sym_DASH] = ACTIONS(3793),
    [anon_sym_DOT] = ACTIONS(3793),
    [anon_sym_SLASH] = ACTIONS(3793),
    [anon_sym_x] = ACTIONS(3793),
    [anon_sym_EQ_TILDE] = ACTIONS(3791),
    [anon_sym_BANG_TILDE] = ACTIONS(3791),
    [anon_sym_EQ_EQ] = ACTIONS(3793),
    [anon_sym_BANG_EQ] = ACTIONS(3791),
    [anon_sym_eq] = ACTIONS(3793),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3791),
    [anon_sym_equ] = ACTIONS(3791),
    [anon_sym_eqr] = ACTIONS(3791),
    [anon_sym_ne] = ACTIONS(3791),
    [anon_sym_LT_EQ_GT] = ACTIONS(3791),
    [anon_sym_cmp] = ACTIONS(3791),
    [anon_sym_TILDE_TILDE] = ACTIONS(3791),
    [anon_sym_LT] = ACTIONS(3793),
    [anon_sym_LT_EQ] = ACTIONS(3793),
    [anon_sym_GT_EQ] = ACTIONS(3791),
    [anon_sym_lt] = ACTIONS(3791),
    [anon_sym_le] = ACTIONS(3791),
    [anon_sym_ge] = ACTIONS(3791),
    [anon_sym_gt] = ACTIONS(3791),
    [anon_sym_isa] = ACTIONS(3791),
    [anon_sym_PLUS_PLUS] = ACTIONS(3791),
    [anon_sym_DASH_DASH] = ACTIONS(3791),
    [anon_sym_QMARK] = ACTIONS(3791),
    [anon_sym_EQ_GT] = ACTIONS(3791),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3791),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3606)] = {
    [sym_heredoc_content] = STATE(3606),
    [anon_sym_SEMI] = ACTIONS(3795),
    [anon_sym_EQ] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_COMMA] = ACTIONS(3795),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3795),
    [anon_sym_DASH_GT] = ACTIONS(3795),
    [anon_sym_STAR] = ACTIONS(3797),
    [anon_sym_PERCENT] = ACTIONS(3797),
    [anon_sym_GT] = ACTIONS(3797),
    [anon_sym_LT_LT] = ACTIONS(3797),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3795),
    [anon_sym_PLUS_EQ] = ACTIONS(3795),
    [anon_sym_DASH_EQ] = ACTIONS(3795),
    [anon_sym_DOT_EQ] = ACTIONS(3795),
    [anon_sym_STAR_EQ] = ACTIONS(3795),
    [anon_sym_SLASH_EQ] = ACTIONS(3795),
    [anon_sym_PERCENT_EQ] = ACTIONS(3795),
    [anon_sym_x_EQ] = ACTIONS(3795),
    [anon_sym_AMP_EQ] = ACTIONS(3795),
    [anon_sym_PIPE_EQ] = ACTIONS(3795),
    [anon_sym_CARET_EQ] = ACTIONS(3795),
    [anon_sym_LT_LT_EQ] = ACTIONS(3795),
    [anon_sym_GT_GT_EQ] = ACTIONS(3795),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3795),
    [anon_sym_DOT_DOT] = ACTIONS(3797),
    [anon_sym_STAR_STAR] = ACTIONS(3797),
    [anon_sym_PIPE_PIPE] = ACTIONS(3797),
    [anon_sym_SLASH_SLASH] = ACTIONS(3797),
    [anon_sym_CARET_CARET] = ACTIONS(3795),
    [anon_sym_AMP_AMP] = ACTIONS(3797),
    [anon_sym_PIPE] = ACTIONS(3797),
    [anon_sym_CARET] = ACTIONS(3797),
    [anon_sym_AMP] = ACTIONS(3797),
    [anon_sym_GT_GT2] = ACTIONS(3797),
    [anon_sym_PLUS] = ACTIONS(3797),
    [anon_sym_DASH] = ACTIONS(3797),
    [anon_sym_DOT] = ACTIONS(3797),
    [anon_sym_SLASH] = ACTIONS(3797),
    [anon_sym_x] = ACTIONS(3797),
    [anon_sym_EQ_TILDE] = ACTIONS(3795),
    [anon_sym_BANG_TILDE] = ACTIONS(3795),
    [anon_sym_EQ_EQ] = ACTIONS(3797),
    [anon_sym_BANG_EQ] = ACTIONS(3795),
    [anon_sym_eq] = ACTIONS(3797),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3795),
    [anon_sym_equ] = ACTIONS(3795),
    [anon_sym_eqr] = ACTIONS(3795),
    [anon_sym_ne] = ACTIONS(3795),
    [anon_sym_LT_EQ_GT] = ACTIONS(3795),
    [anon_sym_cmp] = ACTIONS(3795),
    [anon_sym_TILDE_TILDE] = ACTIONS(3795),
    [anon_sym_LT] = ACTIONS(3797),
    [anon_sym_LT_EQ] = ACTIONS(3797),
    [anon_sym_GT_EQ] = ACTIONS(3795),
    [anon_sym_lt] = ACTIONS(3795),
    [anon_sym_le] = ACTIONS(3795),
    [anon_sym_ge] = ACTIONS(3795),
    [anon_sym_gt] = ACTIONS(3795),
    [anon_sym_isa] = ACTIONS(3795),
    [anon_sym_PLUS_PLUS] = ACTIONS(3795),
    [anon_sym_DASH_DASH] = ACTIONS(3795),
    [anon_sym_QMARK] = ACTIONS(3795),
    [anon_sym_EQ_GT] = ACTIONS(3795),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3795),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3607)] = {
    [sym_heredoc_content] = STATE(3607),
    [anon_sym_SEMI] = ACTIONS(3925),
    [anon_sym_EQ] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_COMMA] = ACTIONS(3925),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3925),
    [anon_sym_DASH_GT] = ACTIONS(3925),
    [anon_sym_STAR] = ACTIONS(3927),
    [anon_sym_PERCENT] = ACTIONS(3927),
    [anon_sym_GT] = ACTIONS(3927),
    [anon_sym_LT_LT] = ACTIONS(3927),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3925),
    [anon_sym_PLUS_EQ] = ACTIONS(3925),
    [anon_sym_DASH_EQ] = ACTIONS(3925),
    [anon_sym_DOT_EQ] = ACTIONS(3925),
    [anon_sym_STAR_EQ] = ACTIONS(3925),
    [anon_sym_SLASH_EQ] = ACTIONS(3925),
    [anon_sym_PERCENT_EQ] = ACTIONS(3925),
    [anon_sym_x_EQ] = ACTIONS(3925),
    [anon_sym_AMP_EQ] = ACTIONS(3925),
    [anon_sym_PIPE_EQ] = ACTIONS(3925),
    [anon_sym_CARET_EQ] = ACTIONS(3925),
    [anon_sym_LT_LT_EQ] = ACTIONS(3925),
    [anon_sym_GT_GT_EQ] = ACTIONS(3925),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3925),
    [anon_sym_DOT_DOT] = ACTIONS(3927),
    [anon_sym_STAR_STAR] = ACTIONS(3927),
    [anon_sym_PIPE_PIPE] = ACTIONS(3927),
    [anon_sym_SLASH_SLASH] = ACTIONS(3927),
    [anon_sym_CARET_CARET] = ACTIONS(3925),
    [anon_sym_AMP_AMP] = ACTIONS(3927),
    [anon_sym_PIPE] = ACTIONS(3927),
    [anon_sym_CARET] = ACTIONS(3927),
    [anon_sym_AMP] = ACTIONS(3927),
    [anon_sym_GT_GT2] = ACTIONS(3927),
    [anon_sym_PLUS] = ACTIONS(3927),
    [anon_sym_DASH] = ACTIONS(3927),
    [anon_sym_DOT] = ACTIONS(3927),
    [anon_sym_SLASH] = ACTIONS(3927),
    [anon_sym_x] = ACTIONS(3927),
    [anon_sym_EQ_TILDE] = ACTIONS(3925),
    [anon_sym_BANG_TILDE] = ACTIONS(3925),
    [anon_sym_EQ_EQ] = ACTIONS(3927),
    [anon_sym_BANG_EQ] = ACTIONS(3925),
    [anon_sym_eq] = ACTIONS(3927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3925),
    [anon_sym_equ] = ACTIONS(3925),
    [anon_sym_eqr] = ACTIONS(3925),
    [anon_sym_ne] = ACTIONS(3925),
    [anon_sym_LT_EQ_GT] = ACTIONS(3925),
    [anon_sym_cmp] = ACTIONS(3925),
    [anon_sym_TILDE_TILDE] = ACTIONS(3925),
    [anon_sym_LT] = ACTIONS(3927),
    [anon_sym_LT_EQ] = ACTIONS(3927),
    [anon_sym_GT_EQ] = ACTIONS(3925),
    [anon_sym_lt] = ACTIONS(3925),
    [anon_sym_le] = ACTIONS(3925),
    [anon_sym_ge] = ACTIONS(3925),
    [anon_sym_gt] = ACTIONS(3925),
    [anon_sym_isa] = ACTIONS(3925),
    [anon_sym_PLUS_PLUS] = ACTIONS(3925),
    [anon_sym_DASH_DASH] = ACTIONS(3925),
    [anon_sym_QMARK] = ACTIONS(3925),
    [anon_sym_EQ_GT] = ACTIONS(3925),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3925),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3608)] = {
    [sym_heredoc_content] = STATE(3608),
    [anon_sym_SEMI] = ACTIONS(3929),
    [anon_sym_EQ] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_COMMA] = ACTIONS(3929),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3929),
    [anon_sym_DASH_GT] = ACTIONS(3929),
    [anon_sym_STAR] = ACTIONS(3931),
    [anon_sym_PERCENT] = ACTIONS(3931),
    [anon_sym_GT] = ACTIONS(3931),
    [anon_sym_LT_LT] = ACTIONS(3931),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3929),
    [anon_sym_PLUS_EQ] = ACTIONS(3929),
    [anon_sym_DASH_EQ] = ACTIONS(3929),
    [anon_sym_DOT_EQ] = ACTIONS(3929),
    [anon_sym_STAR_EQ] = ACTIONS(3929),
    [anon_sym_SLASH_EQ] = ACTIONS(3929),
    [anon_sym_PERCENT_EQ] = ACTIONS(3929),
    [anon_sym_x_EQ] = ACTIONS(3929),
    [anon_sym_AMP_EQ] = ACTIONS(3929),
    [anon_sym_PIPE_EQ] = ACTIONS(3929),
    [anon_sym_CARET_EQ] = ACTIONS(3929),
    [anon_sym_LT_LT_EQ] = ACTIONS(3929),
    [anon_sym_GT_GT_EQ] = ACTIONS(3929),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3929),
    [anon_sym_DOT_DOT] = ACTIONS(3931),
    [anon_sym_STAR_STAR] = ACTIONS(3931),
    [anon_sym_PIPE_PIPE] = ACTIONS(3931),
    [anon_sym_SLASH_SLASH] = ACTIONS(3931),
    [anon_sym_CARET_CARET] = ACTIONS(3929),
    [anon_sym_AMP_AMP] = ACTIONS(3931),
    [anon_sym_PIPE] = ACTIONS(3931),
    [anon_sym_CARET] = ACTIONS(3931),
    [anon_sym_AMP] = ACTIONS(3931),
    [anon_sym_GT_GT2] = ACTIONS(3931),
    [anon_sym_PLUS] = ACTIONS(3931),
    [anon_sym_DASH] = ACTIONS(3931),
    [anon_sym_DOT] = ACTIONS(3931),
    [anon_sym_SLASH] = ACTIONS(3931),
    [anon_sym_x] = ACTIONS(3931),
    [anon_sym_EQ_TILDE] = ACTIONS(3929),
    [anon_sym_BANG_TILDE] = ACTIONS(3929),
    [anon_sym_EQ_EQ] = ACTIONS(3931),
    [anon_sym_BANG_EQ] = ACTIONS(3929),
    [anon_sym_eq] = ACTIONS(3931),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3929),
    [anon_sym_equ] = ACTIONS(3929),
    [anon_sym_eqr] = ACTIONS(3929),
    [anon_sym_ne] = ACTIONS(3929),
    [anon_sym_LT_EQ_GT] = ACTIONS(3929),
    [anon_sym_cmp] = ACTIONS(3929),
    [anon_sym_TILDE_TILDE] = ACTIONS(3929),
    [anon_sym_LT] = ACTIONS(3931),
    [anon_sym_LT_EQ] = ACTIONS(3931),
    [anon_sym_GT_EQ] = ACTIONS(3929),
    [anon_sym_lt] = ACTIONS(3929),
    [anon_sym_le] = ACTIONS(3929),
    [anon_sym_ge] = ACTIONS(3929),
    [anon_sym_gt] = ACTIONS(3929),
    [anon_sym_isa] = ACTIONS(3929),
    [anon_sym_PLUS_PLUS] = ACTIONS(3929),
    [anon_sym_DASH_DASH] = ACTIONS(3929),
    [anon_sym_QMARK] = ACTIONS(3929),
    [anon_sym_EQ_GT] = ACTIONS(3929),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3929),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3609)] = {
    [sym_heredoc_content] = STATE(3609),
    [anon_sym_SEMI] = ACTIONS(3933),
    [anon_sym_EQ] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_COMMA] = ACTIONS(3933),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3933),
    [anon_sym_DASH_GT] = ACTIONS(3933),
    [anon_sym_STAR] = ACTIONS(3935),
    [anon_sym_PERCENT] = ACTIONS(3935),
    [anon_sym_GT] = ACTIONS(3935),
    [anon_sym_LT_LT] = ACTIONS(3935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3933),
    [anon_sym_PLUS_EQ] = ACTIONS(3933),
    [anon_sym_DASH_EQ] = ACTIONS(3933),
    [anon_sym_DOT_EQ] = ACTIONS(3933),
    [anon_sym_STAR_EQ] = ACTIONS(3933),
    [anon_sym_SLASH_EQ] = ACTIONS(3933),
    [anon_sym_PERCENT_EQ] = ACTIONS(3933),
    [anon_sym_x_EQ] = ACTIONS(3933),
    [anon_sym_AMP_EQ] = ACTIONS(3933),
    [anon_sym_PIPE_EQ] = ACTIONS(3933),
    [anon_sym_CARET_EQ] = ACTIONS(3933),
    [anon_sym_LT_LT_EQ] = ACTIONS(3933),
    [anon_sym_GT_GT_EQ] = ACTIONS(3933),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3933),
    [anon_sym_DOT_DOT] = ACTIONS(3935),
    [anon_sym_STAR_STAR] = ACTIONS(3935),
    [anon_sym_PIPE_PIPE] = ACTIONS(3935),
    [anon_sym_SLASH_SLASH] = ACTIONS(3935),
    [anon_sym_CARET_CARET] = ACTIONS(3933),
    [anon_sym_AMP_AMP] = ACTIONS(3935),
    [anon_sym_PIPE] = ACTIONS(3935),
    [anon_sym_CARET] = ACTIONS(3935),
    [anon_sym_AMP] = ACTIONS(3935),
    [anon_sym_GT_GT2] = ACTIONS(3935),
    [anon_sym_PLUS] = ACTIONS(3935),
    [anon_sym_DASH] = ACTIONS(3935),
    [anon_sym_DOT] = ACTIONS(3935),
    [anon_sym_SLASH] = ACTIONS(3935),
    [anon_sym_x] = ACTIONS(3935),
    [anon_sym_EQ_TILDE] = ACTIONS(3933),
    [anon_sym_BANG_TILDE] = ACTIONS(3933),
    [anon_sym_EQ_EQ] = ACTIONS(3935),
    [anon_sym_BANG_EQ] = ACTIONS(3933),
    [anon_sym_eq] = ACTIONS(3935),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3933),
    [anon_sym_equ] = ACTIONS(3933),
    [anon_sym_eqr] = ACTIONS(3933),
    [anon_sym_ne] = ACTIONS(3933),
    [anon_sym_LT_EQ_GT] = ACTIONS(3933),
    [anon_sym_cmp] = ACTIONS(3933),
    [anon_sym_TILDE_TILDE] = ACTIONS(3933),
    [anon_sym_LT] = ACTIONS(3935),
    [anon_sym_LT_EQ] = ACTIONS(3935),
    [anon_sym_GT_EQ] = ACTIONS(3933),
    [anon_sym_lt] = ACTIONS(3933),
    [anon_sym_le] = ACTIONS(3933),
    [anon_sym_ge] = ACTIONS(3933),
    [anon_sym_gt] = ACTIONS(3933),
    [anon_sym_isa] = ACTIONS(3933),
    [anon_sym_PLUS_PLUS] = ACTIONS(3933),
    [anon_sym_DASH_DASH] = ACTIONS(3933),
    [anon_sym_QMARK] = ACTIONS(3933),
    [anon_sym_EQ_GT] = ACTIONS(3933),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3933),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3610)] = {
    [sym_heredoc_content] = STATE(3610),
    [anon_sym_SEMI] = ACTIONS(3937),
    [anon_sym_EQ] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_COMMA] = ACTIONS(3937),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3937),
    [anon_sym_DASH_GT] = ACTIONS(3937),
    [anon_sym_STAR] = ACTIONS(3939),
    [anon_sym_PERCENT] = ACTIONS(3939),
    [anon_sym_GT] = ACTIONS(3939),
    [anon_sym_LT_LT] = ACTIONS(3939),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3937),
    [anon_sym_PLUS_EQ] = ACTIONS(3937),
    [anon_sym_DASH_EQ] = ACTIONS(3937),
    [anon_sym_DOT_EQ] = ACTIONS(3937),
    [anon_sym_STAR_EQ] = ACTIONS(3937),
    [anon_sym_SLASH_EQ] = ACTIONS(3937),
    [anon_sym_PERCENT_EQ] = ACTIONS(3937),
    [anon_sym_x_EQ] = ACTIONS(3937),
    [anon_sym_AMP_EQ] = ACTIONS(3937),
    [anon_sym_PIPE_EQ] = ACTIONS(3937),
    [anon_sym_CARET_EQ] = ACTIONS(3937),
    [anon_sym_LT_LT_EQ] = ACTIONS(3937),
    [anon_sym_GT_GT_EQ] = ACTIONS(3937),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3937),
    [anon_sym_DOT_DOT] = ACTIONS(3939),
    [anon_sym_STAR_STAR] = ACTIONS(3939),
    [anon_sym_PIPE_PIPE] = ACTIONS(3939),
    [anon_sym_SLASH_SLASH] = ACTIONS(3939),
    [anon_sym_CARET_CARET] = ACTIONS(3937),
    [anon_sym_AMP_AMP] = ACTIONS(3939),
    [anon_sym_PIPE] = ACTIONS(3939),
    [anon_sym_CARET] = ACTIONS(3939),
    [anon_sym_AMP] = ACTIONS(3939),
    [anon_sym_GT_GT2] = ACTIONS(3939),
    [anon_sym_PLUS] = ACTIONS(3939),
    [anon_sym_DASH] = ACTIONS(3939),
    [anon_sym_DOT] = ACTIONS(3939),
    [anon_sym_SLASH] = ACTIONS(3939),
    [anon_sym_x] = ACTIONS(3939),
    [anon_sym_EQ_TILDE] = ACTIONS(3937),
    [anon_sym_BANG_TILDE] = ACTIONS(3937),
    [anon_sym_EQ_EQ] = ACTIONS(3939),
    [anon_sym_BANG_EQ] = ACTIONS(3937),
    [anon_sym_eq] = ACTIONS(3939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3937),
    [anon_sym_equ] = ACTIONS(3937),
    [anon_sym_eqr] = ACTIONS(3937),
    [anon_sym_ne] = ACTIONS(3937),
    [anon_sym_LT_EQ_GT] = ACTIONS(3937),
    [anon_sym_cmp] = ACTIONS(3937),
    [anon_sym_TILDE_TILDE] = ACTIONS(3937),
    [anon_sym_LT] = ACTIONS(3939),
    [anon_sym_LT_EQ] = ACTIONS(3939),
    [anon_sym_GT_EQ] = ACTIONS(3937),
    [anon_sym_lt] = ACTIONS(3937),
    [anon_sym_le] = ACTIONS(3937),
    [anon_sym_ge] = ACTIONS(3937),
    [anon_sym_gt] = ACTIONS(3937),
    [anon_sym_isa] = ACTIONS(3937),
    [anon_sym_PLUS_PLUS] = ACTIONS(3937),
    [anon_sym_DASH_DASH] = ACTIONS(3937),
    [anon_sym_QMARK] = ACTIONS(3937),
    [anon_sym_EQ_GT] = ACTIONS(3937),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3937),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3611)] = {
    [sym_heredoc_content] = STATE(3611),
    [anon_sym_SEMI] = ACTIONS(3953),
    [anon_sym_EQ] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_COMMA] = ACTIONS(3953),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3953),
    [anon_sym_DASH_GT] = ACTIONS(3953),
    [anon_sym_STAR] = ACTIONS(3955),
    [anon_sym_PERCENT] = ACTIONS(3955),
    [anon_sym_GT] = ACTIONS(3955),
    [anon_sym_LT_LT] = ACTIONS(3955),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3953),
    [anon_sym_PLUS_EQ] = ACTIONS(3953),
    [anon_sym_DASH_EQ] = ACTIONS(3953),
    [anon_sym_DOT_EQ] = ACTIONS(3953),
    [anon_sym_STAR_EQ] = ACTIONS(3953),
    [anon_sym_SLASH_EQ] = ACTIONS(3953),
    [anon_sym_PERCENT_EQ] = ACTIONS(3953),
    [anon_sym_x_EQ] = ACTIONS(3953),
    [anon_sym_AMP_EQ] = ACTIONS(3953),
    [anon_sym_PIPE_EQ] = ACTIONS(3953),
    [anon_sym_CARET_EQ] = ACTIONS(3953),
    [anon_sym_LT_LT_EQ] = ACTIONS(3953),
    [anon_sym_GT_GT_EQ] = ACTIONS(3953),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3953),
    [anon_sym_DOT_DOT] = ACTIONS(3955),
    [anon_sym_STAR_STAR] = ACTIONS(3955),
    [anon_sym_PIPE_PIPE] = ACTIONS(3955),
    [anon_sym_SLASH_SLASH] = ACTIONS(3955),
    [anon_sym_CARET_CARET] = ACTIONS(3953),
    [anon_sym_AMP_AMP] = ACTIONS(3955),
    [anon_sym_PIPE] = ACTIONS(3955),
    [anon_sym_CARET] = ACTIONS(3955),
    [anon_sym_AMP] = ACTIONS(3955),
    [anon_sym_GT_GT2] = ACTIONS(3955),
    [anon_sym_PLUS] = ACTIONS(3955),
    [anon_sym_DASH] = ACTIONS(3955),
    [anon_sym_DOT] = ACTIONS(3955),
    [anon_sym_SLASH] = ACTIONS(3955),
    [anon_sym_x] = ACTIONS(3955),
    [anon_sym_EQ_TILDE] = ACTIONS(3953),
    [anon_sym_BANG_TILDE] = ACTIONS(3953),
    [anon_sym_EQ_EQ] = ACTIONS(3955),
    [anon_sym_BANG_EQ] = ACTIONS(3953),
    [anon_sym_eq] = ACTIONS(3955),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3953),
    [anon_sym_equ] = ACTIONS(3953),
    [anon_sym_eqr] = ACTIONS(3953),
    [anon_sym_ne] = ACTIONS(3953),
    [anon_sym_LT_EQ_GT] = ACTIONS(3953),
    [anon_sym_cmp] = ACTIONS(3953),
    [anon_sym_TILDE_TILDE] = ACTIONS(3953),
    [anon_sym_LT] = ACTIONS(3955),
    [anon_sym_LT_EQ] = ACTIONS(3955),
    [anon_sym_GT_EQ] = ACTIONS(3953),
    [anon_sym_lt] = ACTIONS(3953),
    [anon_sym_le] = ACTIONS(3953),
    [anon_sym_ge] = ACTIONS(3953),
    [anon_sym_gt] = ACTIONS(3953),
    [anon_sym_isa] = ACTIONS(3953),
    [anon_sym_PLUS_PLUS] = ACTIONS(3953),
    [anon_sym_DASH_DASH] = ACTIONS(3953),
    [anon_sym_QMARK] = ACTIONS(3953),
    [anon_sym_EQ_GT] = ACTIONS(3953),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3953),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3612)] = {
    [sym_heredoc_content] = STATE(3612),
    [anon_sym_SEMI] = ACTIONS(3965),
    [anon_sym_EQ] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_COMMA] = ACTIONS(3965),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3965),
    [anon_sym_DASH_GT] = ACTIONS(3965),
    [anon_sym_STAR] = ACTIONS(3967),
    [anon_sym_PERCENT] = ACTIONS(3967),
    [anon_sym_GT] = ACTIONS(3967),
    [anon_sym_LT_LT] = ACTIONS(3967),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3965),
    [anon_sym_PLUS_EQ] = ACTIONS(3965),
    [anon_sym_DASH_EQ] = ACTIONS(3965),
    [anon_sym_DOT_EQ] = ACTIONS(3965),
    [anon_sym_STAR_EQ] = ACTIONS(3965),
    [anon_sym_SLASH_EQ] = ACTIONS(3965),
    [anon_sym_PERCENT_EQ] = ACTIONS(3965),
    [anon_sym_x_EQ] = ACTIONS(3965),
    [anon_sym_AMP_EQ] = ACTIONS(3965),
    [anon_sym_PIPE_EQ] = ACTIONS(3965),
    [anon_sym_CARET_EQ] = ACTIONS(3965),
    [anon_sym_LT_LT_EQ] = ACTIONS(3965),
    [anon_sym_GT_GT_EQ] = ACTIONS(3965),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3965),
    [anon_sym_DOT_DOT] = ACTIONS(3967),
    [anon_sym_STAR_STAR] = ACTIONS(3967),
    [anon_sym_PIPE_PIPE] = ACTIONS(3967),
    [anon_sym_SLASH_SLASH] = ACTIONS(3967),
    [anon_sym_CARET_CARET] = ACTIONS(3965),
    [anon_sym_AMP_AMP] = ACTIONS(3967),
    [anon_sym_PIPE] = ACTIONS(3967),
    [anon_sym_CARET] = ACTIONS(3967),
    [anon_sym_AMP] = ACTIONS(3967),
    [anon_sym_GT_GT2] = ACTIONS(3967),
    [anon_sym_PLUS] = ACTIONS(3967),
    [anon_sym_DASH] = ACTIONS(3967),
    [anon_sym_DOT] = ACTIONS(3967),
    [anon_sym_SLASH] = ACTIONS(3967),
    [anon_sym_x] = ACTIONS(3967),
    [anon_sym_EQ_TILDE] = ACTIONS(3965),
    [anon_sym_BANG_TILDE] = ACTIONS(3965),
    [anon_sym_EQ_EQ] = ACTIONS(3967),
    [anon_sym_BANG_EQ] = ACTIONS(3965),
    [anon_sym_eq] = ACTIONS(3967),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3965),
    [anon_sym_equ] = ACTIONS(3965),
    [anon_sym_eqr] = ACTIONS(3965),
    [anon_sym_ne] = ACTIONS(3965),
    [anon_sym_LT_EQ_GT] = ACTIONS(3965),
    [anon_sym_cmp] = ACTIONS(3965),
    [anon_sym_TILDE_TILDE] = ACTIONS(3965),
    [anon_sym_LT] = ACTIONS(3967),
    [anon_sym_LT_EQ] = ACTIONS(3967),
    [anon_sym_GT_EQ] = ACTIONS(3965),
    [anon_sym_lt] = ACTIONS(3965),
    [anon_sym_le] = ACTIONS(3965),
    [anon_sym_ge] = ACTIONS(3965),
    [anon_sym_gt] = ACTIONS(3965),
    [anon_sym_isa] = ACTIONS(3965),
    [anon_sym_PLUS_PLUS] = ACTIONS(3965),
    [anon_sym_DASH_DASH] = ACTIONS(3965),
    [anon_sym_QMARK] = ACTIONS(3965),
    [anon_sym_EQ_GT] = ACTIONS(3965),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3965),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3613)] = {
    [sym_heredoc_content] = STATE(3613),
    [anon_sym_SEMI] = ACTIONS(4155),
    [anon_sym_EQ] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_COMMA] = ACTIONS(4155),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4155),
    [anon_sym_DASH_GT] = ACTIONS(4155),
    [anon_sym_STAR] = ACTIONS(4157),
    [anon_sym_PERCENT] = ACTIONS(4157),
    [anon_sym_GT] = ACTIONS(4157),
    [anon_sym_LT_LT] = ACTIONS(4157),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4155),
    [anon_sym_PLUS_EQ] = ACTIONS(4155),
    [anon_sym_DASH_EQ] = ACTIONS(4155),
    [anon_sym_DOT_EQ] = ACTIONS(4155),
    [anon_sym_STAR_EQ] = ACTIONS(4155),
    [anon_sym_SLASH_EQ] = ACTIONS(4155),
    [anon_sym_PERCENT_EQ] = ACTIONS(4155),
    [anon_sym_x_EQ] = ACTIONS(4155),
    [anon_sym_AMP_EQ] = ACTIONS(4155),
    [anon_sym_PIPE_EQ] = ACTIONS(4155),
    [anon_sym_CARET_EQ] = ACTIONS(4155),
    [anon_sym_LT_LT_EQ] = ACTIONS(4155),
    [anon_sym_GT_GT_EQ] = ACTIONS(4155),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4155),
    [anon_sym_DOT_DOT] = ACTIONS(4157),
    [anon_sym_STAR_STAR] = ACTIONS(4157),
    [anon_sym_PIPE_PIPE] = ACTIONS(4157),
    [anon_sym_SLASH_SLASH] = ACTIONS(4157),
    [anon_sym_CARET_CARET] = ACTIONS(4155),
    [anon_sym_AMP_AMP] = ACTIONS(4157),
    [anon_sym_PIPE] = ACTIONS(4157),
    [anon_sym_CARET] = ACTIONS(4157),
    [anon_sym_AMP] = ACTIONS(4157),
    [anon_sym_GT_GT2] = ACTIONS(4157),
    [anon_sym_PLUS] = ACTIONS(4157),
    [anon_sym_DASH] = ACTIONS(4157),
    [anon_sym_DOT] = ACTIONS(4157),
    [anon_sym_SLASH] = ACTIONS(4157),
    [anon_sym_x] = ACTIONS(4157),
    [anon_sym_EQ_TILDE] = ACTIONS(4155),
    [anon_sym_BANG_TILDE] = ACTIONS(4155),
    [anon_sym_EQ_EQ] = ACTIONS(4157),
    [anon_sym_BANG_EQ] = ACTIONS(4155),
    [anon_sym_eq] = ACTIONS(4157),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4155),
    [anon_sym_equ] = ACTIONS(4155),
    [anon_sym_eqr] = ACTIONS(4155),
    [anon_sym_ne] = ACTIONS(4155),
    [anon_sym_LT_EQ_GT] = ACTIONS(4155),
    [anon_sym_cmp] = ACTIONS(4155),
    [anon_sym_TILDE_TILDE] = ACTIONS(4155),
    [anon_sym_LT] = ACTIONS(4157),
    [anon_sym_LT_EQ] = ACTIONS(4157),
    [anon_sym_GT_EQ] = ACTIONS(4155),
    [anon_sym_lt] = ACTIONS(4155),
    [anon_sym_le] = ACTIONS(4155),
    [anon_sym_ge] = ACTIONS(4155),
    [anon_sym_gt] = ACTIONS(4155),
    [anon_sym_isa] = ACTIONS(4155),
    [anon_sym_PLUS_PLUS] = ACTIONS(4155),
    [anon_sym_DASH_DASH] = ACTIONS(4155),
    [anon_sym_QMARK] = ACTIONS(4155),
    [anon_sym_EQ_GT] = ACTIONS(4155),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4155),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3614)] = {
    [sym_heredoc_content] = STATE(3614),
    [anon_sym_SEMI] = ACTIONS(4176),
    [anon_sym_EQ] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_COMMA] = ACTIONS(4176),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4176),
    [anon_sym_DASH_GT] = ACTIONS(4176),
    [anon_sym_STAR] = ACTIONS(4178),
    [anon_sym_PERCENT] = ACTIONS(4178),
    [anon_sym_GT] = ACTIONS(4178),
    [anon_sym_LT_LT] = ACTIONS(4178),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4176),
    [anon_sym_PLUS_EQ] = ACTIONS(4176),
    [anon_sym_DASH_EQ] = ACTIONS(4176),
    [anon_sym_DOT_EQ] = ACTIONS(4176),
    [anon_sym_STAR_EQ] = ACTIONS(4176),
    [anon_sym_SLASH_EQ] = ACTIONS(4176),
    [anon_sym_PERCENT_EQ] = ACTIONS(4176),
    [anon_sym_x_EQ] = ACTIONS(4176),
    [anon_sym_AMP_EQ] = ACTIONS(4176),
    [anon_sym_PIPE_EQ] = ACTIONS(4176),
    [anon_sym_CARET_EQ] = ACTIONS(4176),
    [anon_sym_LT_LT_EQ] = ACTIONS(4176),
    [anon_sym_GT_GT_EQ] = ACTIONS(4176),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4176),
    [anon_sym_DOT_DOT] = ACTIONS(4178),
    [anon_sym_STAR_STAR] = ACTIONS(4178),
    [anon_sym_PIPE_PIPE] = ACTIONS(4178),
    [anon_sym_SLASH_SLASH] = ACTIONS(4178),
    [anon_sym_CARET_CARET] = ACTIONS(4176),
    [anon_sym_AMP_AMP] = ACTIONS(4178),
    [anon_sym_PIPE] = ACTIONS(4178),
    [anon_sym_CARET] = ACTIONS(4178),
    [anon_sym_AMP] = ACTIONS(4178),
    [anon_sym_GT_GT2] = ACTIONS(4178),
    [anon_sym_PLUS] = ACTIONS(4178),
    [anon_sym_DASH] = ACTIONS(4178),
    [anon_sym_DOT] = ACTIONS(4178),
    [anon_sym_SLASH] = ACTIONS(4178),
    [anon_sym_x] = ACTIONS(4178),
    [anon_sym_EQ_TILDE] = ACTIONS(4176),
    [anon_sym_BANG_TILDE] = ACTIONS(4176),
    [anon_sym_EQ_EQ] = ACTIONS(4178),
    [anon_sym_BANG_EQ] = ACTIONS(4176),
    [anon_sym_eq] = ACTIONS(4178),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4176),
    [anon_sym_equ] = ACTIONS(4176),
    [anon_sym_eqr] = ACTIONS(4176),
    [anon_sym_ne] = ACTIONS(4176),
    [anon_sym_LT_EQ_GT] = ACTIONS(4176),
    [anon_sym_cmp] = ACTIONS(4176),
    [anon_sym_TILDE_TILDE] = ACTIONS(4176),
    [anon_sym_LT] = ACTIONS(4178),
    [anon_sym_LT_EQ] = ACTIONS(4178),
    [anon_sym_GT_EQ] = ACTIONS(4176),
    [anon_sym_lt] = ACTIONS(4176),
    [anon_sym_le] = ACTIONS(4176),
    [anon_sym_ge] = ACTIONS(4176),
    [anon_sym_gt] = ACTIONS(4176),
    [anon_sym_isa] = ACTIONS(4176),
    [anon_sym_PLUS_PLUS] = ACTIONS(4176),
    [anon_sym_DASH_DASH] = ACTIONS(4176),
    [anon_sym_QMARK] = ACTIONS(4176),
    [anon_sym_EQ_GT] = ACTIONS(4176),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4176),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3615)] = {
    [sym_heredoc_content] = STATE(3615),
    [anon_sym_SEMI] = ACTIONS(4180),
    [anon_sym_EQ] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_COMMA] = ACTIONS(4180),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4180),
    [anon_sym_DASH_GT] = ACTIONS(4180),
    [anon_sym_STAR] = ACTIONS(4182),
    [anon_sym_PERCENT] = ACTIONS(4182),
    [anon_sym_GT] = ACTIONS(4182),
    [anon_sym_LT_LT] = ACTIONS(4182),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4180),
    [anon_sym_PLUS_EQ] = ACTIONS(4180),
    [anon_sym_DASH_EQ] = ACTIONS(4180),
    [anon_sym_DOT_EQ] = ACTIONS(4180),
    [anon_sym_STAR_EQ] = ACTIONS(4180),
    [anon_sym_SLASH_EQ] = ACTIONS(4180),
    [anon_sym_PERCENT_EQ] = ACTIONS(4180),
    [anon_sym_x_EQ] = ACTIONS(4180),
    [anon_sym_AMP_EQ] = ACTIONS(4180),
    [anon_sym_PIPE_EQ] = ACTIONS(4180),
    [anon_sym_CARET_EQ] = ACTIONS(4180),
    [anon_sym_LT_LT_EQ] = ACTIONS(4180),
    [anon_sym_GT_GT_EQ] = ACTIONS(4180),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4180),
    [anon_sym_DOT_DOT] = ACTIONS(4182),
    [anon_sym_STAR_STAR] = ACTIONS(4182),
    [anon_sym_PIPE_PIPE] = ACTIONS(4182),
    [anon_sym_SLASH_SLASH] = ACTIONS(4182),
    [anon_sym_CARET_CARET] = ACTIONS(4180),
    [anon_sym_AMP_AMP] = ACTIONS(4182),
    [anon_sym_PIPE] = ACTIONS(4182),
    [anon_sym_CARET] = ACTIONS(4182),
    [anon_sym_AMP] = ACTIONS(4182),
    [anon_sym_GT_GT2] = ACTIONS(4182),
    [anon_sym_PLUS] = ACTIONS(4182),
    [anon_sym_DASH] = ACTIONS(4182),
    [anon_sym_DOT] = ACTIONS(4182),
    [anon_sym_SLASH] = ACTIONS(4182),
    [anon_sym_x] = ACTIONS(4182),
    [anon_sym_EQ_TILDE] = ACTIONS(4180),
    [anon_sym_BANG_TILDE] = ACTIONS(4180),
    [anon_sym_EQ_EQ] = ACTIONS(4182),
    [anon_sym_BANG_EQ] = ACTIONS(4180),
    [anon_sym_eq] = ACTIONS(4182),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4180),
    [anon_sym_equ] = ACTIONS(4180),
    [anon_sym_eqr] = ACTIONS(4180),
    [anon_sym_ne] = ACTIONS(4180),
    [anon_sym_LT_EQ_GT] = ACTIONS(4180),
    [anon_sym_cmp] = ACTIONS(4180),
    [anon_sym_TILDE_TILDE] = ACTIONS(4180),
    [anon_sym_LT] = ACTIONS(4182),
    [anon_sym_LT_EQ] = ACTIONS(4182),
    [anon_sym_GT_EQ] = ACTIONS(4180),
    [anon_sym_lt] = ACTIONS(4180),
    [anon_sym_le] = ACTIONS(4180),
    [anon_sym_ge] = ACTIONS(4180),
    [anon_sym_gt] = ACTIONS(4180),
    [anon_sym_isa] = ACTIONS(4180),
    [anon_sym_PLUS_PLUS] = ACTIONS(4180),
    [anon_sym_DASH_DASH] = ACTIONS(4180),
    [anon_sym_QMARK] = ACTIONS(4180),
    [anon_sym_EQ_GT] = ACTIONS(4180),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4180),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3616)] = {
    [sym_heredoc_content] = STATE(3616),
    [anon_sym_SEMI] = ACTIONS(4192),
    [anon_sym_EQ] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_COMMA] = ACTIONS(4192),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4192),
    [anon_sym_DASH_GT] = ACTIONS(4192),
    [anon_sym_STAR] = ACTIONS(4194),
    [anon_sym_PERCENT] = ACTIONS(4194),
    [anon_sym_GT] = ACTIONS(4194),
    [anon_sym_LT_LT] = ACTIONS(4194),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4192),
    [anon_sym_PLUS_EQ] = ACTIONS(4192),
    [anon_sym_DASH_EQ] = ACTIONS(4192),
    [anon_sym_DOT_EQ] = ACTIONS(4192),
    [anon_sym_STAR_EQ] = ACTIONS(4192),
    [anon_sym_SLASH_EQ] = ACTIONS(4192),
    [anon_sym_PERCENT_EQ] = ACTIONS(4192),
    [anon_sym_x_EQ] = ACTIONS(4192),
    [anon_sym_AMP_EQ] = ACTIONS(4192),
    [anon_sym_PIPE_EQ] = ACTIONS(4192),
    [anon_sym_CARET_EQ] = ACTIONS(4192),
    [anon_sym_LT_LT_EQ] = ACTIONS(4192),
    [anon_sym_GT_GT_EQ] = ACTIONS(4192),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4192),
    [anon_sym_DOT_DOT] = ACTIONS(4194),
    [anon_sym_STAR_STAR] = ACTIONS(4194),
    [anon_sym_PIPE_PIPE] = ACTIONS(4194),
    [anon_sym_SLASH_SLASH] = ACTIONS(4194),
    [anon_sym_CARET_CARET] = ACTIONS(4192),
    [anon_sym_AMP_AMP] = ACTIONS(4194),
    [anon_sym_PIPE] = ACTIONS(4194),
    [anon_sym_CARET] = ACTIONS(4194),
    [anon_sym_AMP] = ACTIONS(4194),
    [anon_sym_GT_GT2] = ACTIONS(4194),
    [anon_sym_PLUS] = ACTIONS(4194),
    [anon_sym_DASH] = ACTIONS(4194),
    [anon_sym_DOT] = ACTIONS(4194),
    [anon_sym_SLASH] = ACTIONS(4194),
    [anon_sym_x] = ACTIONS(4194),
    [anon_sym_EQ_TILDE] = ACTIONS(4192),
    [anon_sym_BANG_TILDE] = ACTIONS(4192),
    [anon_sym_EQ_EQ] = ACTIONS(4194),
    [anon_sym_BANG_EQ] = ACTIONS(4192),
    [anon_sym_eq] = ACTIONS(4194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4192),
    [anon_sym_equ] = ACTIONS(4192),
    [anon_sym_eqr] = ACTIONS(4192),
    [anon_sym_ne] = ACTIONS(4192),
    [anon_sym_LT_EQ_GT] = ACTIONS(4192),
    [anon_sym_cmp] = ACTIONS(4192),
    [anon_sym_TILDE_TILDE] = ACTIONS(4192),
    [anon_sym_LT] = ACTIONS(4194),
    [anon_sym_LT_EQ] = ACTIONS(4194),
    [anon_sym_GT_EQ] = ACTIONS(4192),
    [anon_sym_lt] = ACTIONS(4192),
    [anon_sym_le] = ACTIONS(4192),
    [anon_sym_ge] = ACTIONS(4192),
    [anon_sym_gt] = ACTIONS(4192),
    [anon_sym_isa] = ACTIONS(4192),
    [anon_sym_PLUS_PLUS] = ACTIONS(4192),
    [anon_sym_DASH_DASH] = ACTIONS(4192),
    [anon_sym_QMARK] = ACTIONS(4192),
    [anon_sym_EQ_GT] = ACTIONS(4192),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4192),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3617)] = {
    [sym_heredoc_content] = STATE(3617),
    [anon_sym_SEMI] = ACTIONS(4200),
    [anon_sym_EQ] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_COMMA] = ACTIONS(4200),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4200),
    [anon_sym_DASH_GT] = ACTIONS(4200),
    [anon_sym_STAR] = ACTIONS(4202),
    [anon_sym_PERCENT] = ACTIONS(4202),
    [anon_sym_GT] = ACTIONS(4202),
    [anon_sym_LT_LT] = ACTIONS(4202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4200),
    [anon_sym_PLUS_EQ] = ACTIONS(4200),
    [anon_sym_DASH_EQ] = ACTIONS(4200),
    [anon_sym_DOT_EQ] = ACTIONS(4200),
    [anon_sym_STAR_EQ] = ACTIONS(4200),
    [anon_sym_SLASH_EQ] = ACTIONS(4200),
    [anon_sym_PERCENT_EQ] = ACTIONS(4200),
    [anon_sym_x_EQ] = ACTIONS(4200),
    [anon_sym_AMP_EQ] = ACTIONS(4200),
    [anon_sym_PIPE_EQ] = ACTIONS(4200),
    [anon_sym_CARET_EQ] = ACTIONS(4200),
    [anon_sym_LT_LT_EQ] = ACTIONS(4200),
    [anon_sym_GT_GT_EQ] = ACTIONS(4200),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4200),
    [anon_sym_DOT_DOT] = ACTIONS(4202),
    [anon_sym_STAR_STAR] = ACTIONS(4202),
    [anon_sym_PIPE_PIPE] = ACTIONS(4202),
    [anon_sym_SLASH_SLASH] = ACTIONS(4202),
    [anon_sym_CARET_CARET] = ACTIONS(4200),
    [anon_sym_AMP_AMP] = ACTIONS(4202),
    [anon_sym_PIPE] = ACTIONS(4202),
    [anon_sym_CARET] = ACTIONS(4202),
    [anon_sym_AMP] = ACTIONS(4202),
    [anon_sym_GT_GT2] = ACTIONS(4202),
    [anon_sym_PLUS] = ACTIONS(4202),
    [anon_sym_DASH] = ACTIONS(4202),
    [anon_sym_DOT] = ACTIONS(4202),
    [anon_sym_SLASH] = ACTIONS(4202),
    [anon_sym_x] = ACTIONS(4202),
    [anon_sym_EQ_TILDE] = ACTIONS(4200),
    [anon_sym_BANG_TILDE] = ACTIONS(4200),
    [anon_sym_EQ_EQ] = ACTIONS(4202),
    [anon_sym_BANG_EQ] = ACTIONS(4200),
    [anon_sym_eq] = ACTIONS(4202),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4200),
    [anon_sym_equ] = ACTIONS(4200),
    [anon_sym_eqr] = ACTIONS(4200),
    [anon_sym_ne] = ACTIONS(4200),
    [anon_sym_LT_EQ_GT] = ACTIONS(4200),
    [anon_sym_cmp] = ACTIONS(4200),
    [anon_sym_TILDE_TILDE] = ACTIONS(4200),
    [anon_sym_LT] = ACTIONS(4202),
    [anon_sym_LT_EQ] = ACTIONS(4202),
    [anon_sym_GT_EQ] = ACTIONS(4200),
    [anon_sym_lt] = ACTIONS(4200),
    [anon_sym_le] = ACTIONS(4200),
    [anon_sym_ge] = ACTIONS(4200),
    [anon_sym_gt] = ACTIONS(4200),
    [anon_sym_isa] = ACTIONS(4200),
    [anon_sym_PLUS_PLUS] = ACTIONS(4200),
    [anon_sym_DASH_DASH] = ACTIONS(4200),
    [anon_sym_QMARK] = ACTIONS(4200),
    [anon_sym_EQ_GT] = ACTIONS(4200),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4200),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3618)] = {
    [sym_heredoc_content] = STATE(3618),
    [anon_sym_SEMI] = ACTIONS(3969),
    [anon_sym_EQ] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_COMMA] = ACTIONS(3969),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3969),
    [anon_sym_DASH_GT] = ACTIONS(3969),
    [anon_sym_STAR] = ACTIONS(3971),
    [anon_sym_PERCENT] = ACTIONS(3971),
    [anon_sym_GT] = ACTIONS(3971),
    [anon_sym_LT_LT] = ACTIONS(3971),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3969),
    [anon_sym_PLUS_EQ] = ACTIONS(3969),
    [anon_sym_DASH_EQ] = ACTIONS(3969),
    [anon_sym_DOT_EQ] = ACTIONS(3969),
    [anon_sym_STAR_EQ] = ACTIONS(3969),
    [anon_sym_SLASH_EQ] = ACTIONS(3969),
    [anon_sym_PERCENT_EQ] = ACTIONS(3969),
    [anon_sym_x_EQ] = ACTIONS(3969),
    [anon_sym_AMP_EQ] = ACTIONS(3969),
    [anon_sym_PIPE_EQ] = ACTIONS(3969),
    [anon_sym_CARET_EQ] = ACTIONS(3969),
    [anon_sym_LT_LT_EQ] = ACTIONS(3969),
    [anon_sym_GT_GT_EQ] = ACTIONS(3969),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3969),
    [anon_sym_DOT_DOT] = ACTIONS(3971),
    [anon_sym_STAR_STAR] = ACTIONS(3971),
    [anon_sym_PIPE_PIPE] = ACTIONS(3971),
    [anon_sym_SLASH_SLASH] = ACTIONS(3971),
    [anon_sym_CARET_CARET] = ACTIONS(3969),
    [anon_sym_AMP_AMP] = ACTIONS(3971),
    [anon_sym_PIPE] = ACTIONS(3971),
    [anon_sym_CARET] = ACTIONS(3971),
    [anon_sym_AMP] = ACTIONS(3971),
    [anon_sym_GT_GT2] = ACTIONS(3971),
    [anon_sym_PLUS] = ACTIONS(3971),
    [anon_sym_DASH] = ACTIONS(3971),
    [anon_sym_DOT] = ACTIONS(3971),
    [anon_sym_SLASH] = ACTIONS(3971),
    [anon_sym_x] = ACTIONS(3971),
    [anon_sym_EQ_TILDE] = ACTIONS(3969),
    [anon_sym_BANG_TILDE] = ACTIONS(3969),
    [anon_sym_EQ_EQ] = ACTIONS(3971),
    [anon_sym_BANG_EQ] = ACTIONS(3969),
    [anon_sym_eq] = ACTIONS(3971),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3969),
    [anon_sym_equ] = ACTIONS(3969),
    [anon_sym_eqr] = ACTIONS(3969),
    [anon_sym_ne] = ACTIONS(3969),
    [anon_sym_LT_EQ_GT] = ACTIONS(3969),
    [anon_sym_cmp] = ACTIONS(3969),
    [anon_sym_TILDE_TILDE] = ACTIONS(3969),
    [anon_sym_LT] = ACTIONS(3971),
    [anon_sym_LT_EQ] = ACTIONS(3971),
    [anon_sym_GT_EQ] = ACTIONS(3969),
    [anon_sym_lt] = ACTIONS(3969),
    [anon_sym_le] = ACTIONS(3969),
    [anon_sym_ge] = ACTIONS(3969),
    [anon_sym_gt] = ACTIONS(3969),
    [anon_sym_isa] = ACTIONS(3969),
    [anon_sym_PLUS_PLUS] = ACTIONS(3969),
    [anon_sym_DASH_DASH] = ACTIONS(3969),
    [anon_sym_QMARK] = ACTIONS(3969),
    [anon_sym_EQ_GT] = ACTIONS(3969),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3969),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3619)] = {
    [sym_heredoc_content] = STATE(3619),
    [anon_sym_SEMI] = ACTIONS(4079),
    [anon_sym_EQ] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_COMMA] = ACTIONS(4079),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4079),
    [anon_sym_DASH_GT] = ACTIONS(4079),
    [anon_sym_STAR] = ACTIONS(4081),
    [anon_sym_PERCENT] = ACTIONS(4081),
    [anon_sym_GT] = ACTIONS(4081),
    [anon_sym_LT_LT] = ACTIONS(4081),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4079),
    [anon_sym_PLUS_EQ] = ACTIONS(4079),
    [anon_sym_DASH_EQ] = ACTIONS(4079),
    [anon_sym_DOT_EQ] = ACTIONS(4079),
    [anon_sym_STAR_EQ] = ACTIONS(4079),
    [anon_sym_SLASH_EQ] = ACTIONS(4079),
    [anon_sym_PERCENT_EQ] = ACTIONS(4079),
    [anon_sym_x_EQ] = ACTIONS(4079),
    [anon_sym_AMP_EQ] = ACTIONS(4079),
    [anon_sym_PIPE_EQ] = ACTIONS(4079),
    [anon_sym_CARET_EQ] = ACTIONS(4079),
    [anon_sym_LT_LT_EQ] = ACTIONS(4079),
    [anon_sym_GT_GT_EQ] = ACTIONS(4079),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4079),
    [anon_sym_DOT_DOT] = ACTIONS(4081),
    [anon_sym_STAR_STAR] = ACTIONS(4081),
    [anon_sym_PIPE_PIPE] = ACTIONS(4081),
    [anon_sym_SLASH_SLASH] = ACTIONS(4081),
    [anon_sym_CARET_CARET] = ACTIONS(4079),
    [anon_sym_AMP_AMP] = ACTIONS(4081),
    [anon_sym_PIPE] = ACTIONS(4081),
    [anon_sym_CARET] = ACTIONS(4081),
    [anon_sym_AMP] = ACTIONS(4081),
    [anon_sym_GT_GT2] = ACTIONS(4081),
    [anon_sym_PLUS] = ACTIONS(4081),
    [anon_sym_DASH] = ACTIONS(4081),
    [anon_sym_DOT] = ACTIONS(4081),
    [anon_sym_SLASH] = ACTIONS(4081),
    [anon_sym_x] = ACTIONS(4081),
    [anon_sym_EQ_TILDE] = ACTIONS(4079),
    [anon_sym_BANG_TILDE] = ACTIONS(4079),
    [anon_sym_EQ_EQ] = ACTIONS(4081),
    [anon_sym_BANG_EQ] = ACTIONS(4079),
    [anon_sym_eq] = ACTIONS(4081),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4079),
    [anon_sym_equ] = ACTIONS(4079),
    [anon_sym_eqr] = ACTIONS(4079),
    [anon_sym_ne] = ACTIONS(4079),
    [anon_sym_LT_EQ_GT] = ACTIONS(4079),
    [anon_sym_cmp] = ACTIONS(4079),
    [anon_sym_TILDE_TILDE] = ACTIONS(4079),
    [anon_sym_LT] = ACTIONS(4081),
    [anon_sym_LT_EQ] = ACTIONS(4081),
    [anon_sym_GT_EQ] = ACTIONS(4079),
    [anon_sym_lt] = ACTIONS(4079),
    [anon_sym_le] = ACTIONS(4079),
    [anon_sym_ge] = ACTIONS(4079),
    [anon_sym_gt] = ACTIONS(4079),
    [anon_sym_isa] = ACTIONS(4079),
    [anon_sym_PLUS_PLUS] = ACTIONS(4079),
    [anon_sym_DASH_DASH] = ACTIONS(4079),
    [anon_sym_QMARK] = ACTIONS(4079),
    [anon_sym_EQ_GT] = ACTIONS(4079),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4079),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3620)] = {
    [sym_heredoc_content] = STATE(3620),
    [anon_sym_SEMI] = ACTIONS(4083),
    [anon_sym_EQ] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_COMMA] = ACTIONS(4083),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4083),
    [anon_sym_DASH_GT] = ACTIONS(4083),
    [anon_sym_STAR] = ACTIONS(4085),
    [anon_sym_PERCENT] = ACTIONS(4085),
    [anon_sym_GT] = ACTIONS(4085),
    [anon_sym_LT_LT] = ACTIONS(4085),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4083),
    [anon_sym_PLUS_EQ] = ACTIONS(4083),
    [anon_sym_DASH_EQ] = ACTIONS(4083),
    [anon_sym_DOT_EQ] = ACTIONS(4083),
    [anon_sym_STAR_EQ] = ACTIONS(4083),
    [anon_sym_SLASH_EQ] = ACTIONS(4083),
    [anon_sym_PERCENT_EQ] = ACTIONS(4083),
    [anon_sym_x_EQ] = ACTIONS(4083),
    [anon_sym_AMP_EQ] = ACTIONS(4083),
    [anon_sym_PIPE_EQ] = ACTIONS(4083),
    [anon_sym_CARET_EQ] = ACTIONS(4083),
    [anon_sym_LT_LT_EQ] = ACTIONS(4083),
    [anon_sym_GT_GT_EQ] = ACTIONS(4083),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4083),
    [anon_sym_DOT_DOT] = ACTIONS(4085),
    [anon_sym_STAR_STAR] = ACTIONS(4085),
    [anon_sym_PIPE_PIPE] = ACTIONS(4085),
    [anon_sym_SLASH_SLASH] = ACTIONS(4085),
    [anon_sym_CARET_CARET] = ACTIONS(4083),
    [anon_sym_AMP_AMP] = ACTIONS(4085),
    [anon_sym_PIPE] = ACTIONS(4085),
    [anon_sym_CARET] = ACTIONS(4085),
    [anon_sym_AMP] = ACTIONS(4085),
    [anon_sym_GT_GT2] = ACTIONS(4085),
    [anon_sym_PLUS] = ACTIONS(4085),
    [anon_sym_DASH] = ACTIONS(4085),
    [anon_sym_DOT] = ACTIONS(4085),
    [anon_sym_SLASH] = ACTIONS(4085),
    [anon_sym_x] = ACTIONS(4085),
    [anon_sym_EQ_TILDE] = ACTIONS(4083),
    [anon_sym_BANG_TILDE] = ACTIONS(4083),
    [anon_sym_EQ_EQ] = ACTIONS(4085),
    [anon_sym_BANG_EQ] = ACTIONS(4083),
    [anon_sym_eq] = ACTIONS(4085),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4083),
    [anon_sym_equ] = ACTIONS(4083),
    [anon_sym_eqr] = ACTIONS(4083),
    [anon_sym_ne] = ACTIONS(4083),
    [anon_sym_LT_EQ_GT] = ACTIONS(4083),
    [anon_sym_cmp] = ACTIONS(4083),
    [anon_sym_TILDE_TILDE] = ACTIONS(4083),
    [anon_sym_LT] = ACTIONS(4085),
    [anon_sym_LT_EQ] = ACTIONS(4085),
    [anon_sym_GT_EQ] = ACTIONS(4083),
    [anon_sym_lt] = ACTIONS(4083),
    [anon_sym_le] = ACTIONS(4083),
    [anon_sym_ge] = ACTIONS(4083),
    [anon_sym_gt] = ACTIONS(4083),
    [anon_sym_isa] = ACTIONS(4083),
    [anon_sym_PLUS_PLUS] = ACTIONS(4083),
    [anon_sym_DASH_DASH] = ACTIONS(4083),
    [anon_sym_QMARK] = ACTIONS(4083),
    [anon_sym_EQ_GT] = ACTIONS(4083),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4083),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3621)] = {
    [sym_heredoc_content] = STATE(3621),
    [anon_sym_SEMI] = ACTIONS(4091),
    [anon_sym_EQ] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_COMMA] = ACTIONS(4091),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4091),
    [anon_sym_DASH_GT] = ACTIONS(4091),
    [anon_sym_STAR] = ACTIONS(4093),
    [anon_sym_PERCENT] = ACTIONS(4093),
    [anon_sym_GT] = ACTIONS(4093),
    [anon_sym_LT_LT] = ACTIONS(4093),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4091),
    [anon_sym_PLUS_EQ] = ACTIONS(4091),
    [anon_sym_DASH_EQ] = ACTIONS(4091),
    [anon_sym_DOT_EQ] = ACTIONS(4091),
    [anon_sym_STAR_EQ] = ACTIONS(4091),
    [anon_sym_SLASH_EQ] = ACTIONS(4091),
    [anon_sym_PERCENT_EQ] = ACTIONS(4091),
    [anon_sym_x_EQ] = ACTIONS(4091),
    [anon_sym_AMP_EQ] = ACTIONS(4091),
    [anon_sym_PIPE_EQ] = ACTIONS(4091),
    [anon_sym_CARET_EQ] = ACTIONS(4091),
    [anon_sym_LT_LT_EQ] = ACTIONS(4091),
    [anon_sym_GT_GT_EQ] = ACTIONS(4091),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4091),
    [anon_sym_DOT_DOT] = ACTIONS(4093),
    [anon_sym_STAR_STAR] = ACTIONS(4093),
    [anon_sym_PIPE_PIPE] = ACTIONS(4093),
    [anon_sym_SLASH_SLASH] = ACTIONS(4093),
    [anon_sym_CARET_CARET] = ACTIONS(4091),
    [anon_sym_AMP_AMP] = ACTIONS(4093),
    [anon_sym_PIPE] = ACTIONS(4093),
    [anon_sym_CARET] = ACTIONS(4093),
    [anon_sym_AMP] = ACTIONS(4093),
    [anon_sym_GT_GT2] = ACTIONS(4093),
    [anon_sym_PLUS] = ACTIONS(4093),
    [anon_sym_DASH] = ACTIONS(4093),
    [anon_sym_DOT] = ACTIONS(4093),
    [anon_sym_SLASH] = ACTIONS(4093),
    [anon_sym_x] = ACTIONS(4093),
    [anon_sym_EQ_TILDE] = ACTIONS(4091),
    [anon_sym_BANG_TILDE] = ACTIONS(4091),
    [anon_sym_EQ_EQ] = ACTIONS(4093),
    [anon_sym_BANG_EQ] = ACTIONS(4091),
    [anon_sym_eq] = ACTIONS(4093),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4091),
    [anon_sym_equ] = ACTIONS(4091),
    [anon_sym_eqr] = ACTIONS(4091),
    [anon_sym_ne] = ACTIONS(4091),
    [anon_sym_LT_EQ_GT] = ACTIONS(4091),
    [anon_sym_cmp] = ACTIONS(4091),
    [anon_sym_TILDE_TILDE] = ACTIONS(4091),
    [anon_sym_LT] = ACTIONS(4093),
    [anon_sym_LT_EQ] = ACTIONS(4093),
    [anon_sym_GT_EQ] = ACTIONS(4091),
    [anon_sym_lt] = ACTIONS(4091),
    [anon_sym_le] = ACTIONS(4091),
    [anon_sym_ge] = ACTIONS(4091),
    [anon_sym_gt] = ACTIONS(4091),
    [anon_sym_isa] = ACTIONS(4091),
    [anon_sym_PLUS_PLUS] = ACTIONS(4091),
    [anon_sym_DASH_DASH] = ACTIONS(4091),
    [anon_sym_QMARK] = ACTIONS(4091),
    [anon_sym_EQ_GT] = ACTIONS(4091),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4091),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3622)] = {
    [sym_heredoc_content] = STATE(3622),
    [anon_sym_SEMI] = ACTIONS(3715),
    [anon_sym_EQ] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_COMMA] = ACTIONS(3715),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3715),
    [anon_sym_DASH_GT] = ACTIONS(3715),
    [anon_sym_STAR] = ACTIONS(3717),
    [anon_sym_PERCENT] = ACTIONS(3717),
    [anon_sym_GT] = ACTIONS(3717),
    [anon_sym_LT_LT] = ACTIONS(3717),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3715),
    [anon_sym_PLUS_EQ] = ACTIONS(3715),
    [anon_sym_DASH_EQ] = ACTIONS(3715),
    [anon_sym_DOT_EQ] = ACTIONS(3715),
    [anon_sym_STAR_EQ] = ACTIONS(3715),
    [anon_sym_SLASH_EQ] = ACTIONS(3715),
    [anon_sym_PERCENT_EQ] = ACTIONS(3715),
    [anon_sym_x_EQ] = ACTIONS(3715),
    [anon_sym_AMP_EQ] = ACTIONS(3715),
    [anon_sym_PIPE_EQ] = ACTIONS(3715),
    [anon_sym_CARET_EQ] = ACTIONS(3715),
    [anon_sym_LT_LT_EQ] = ACTIONS(3715),
    [anon_sym_GT_GT_EQ] = ACTIONS(3715),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3715),
    [anon_sym_DOT_DOT] = ACTIONS(3717),
    [anon_sym_STAR_STAR] = ACTIONS(3717),
    [anon_sym_PIPE_PIPE] = ACTIONS(3717),
    [anon_sym_SLASH_SLASH] = ACTIONS(3717),
    [anon_sym_CARET_CARET] = ACTIONS(3715),
    [anon_sym_AMP_AMP] = ACTIONS(3717),
    [anon_sym_PIPE] = ACTIONS(3717),
    [anon_sym_CARET] = ACTIONS(3717),
    [anon_sym_AMP] = ACTIONS(3717),
    [anon_sym_GT_GT2] = ACTIONS(3717),
    [anon_sym_PLUS] = ACTIONS(3717),
    [anon_sym_DASH] = ACTIONS(3717),
    [anon_sym_DOT] = ACTIONS(3717),
    [anon_sym_SLASH] = ACTIONS(3717),
    [anon_sym_x] = ACTIONS(3717),
    [anon_sym_EQ_TILDE] = ACTIONS(3715),
    [anon_sym_BANG_TILDE] = ACTIONS(3715),
    [anon_sym_EQ_EQ] = ACTIONS(3717),
    [anon_sym_BANG_EQ] = ACTIONS(3715),
    [anon_sym_eq] = ACTIONS(3717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3715),
    [anon_sym_equ] = ACTIONS(3715),
    [anon_sym_eqr] = ACTIONS(3715),
    [anon_sym_ne] = ACTIONS(3715),
    [anon_sym_LT_EQ_GT] = ACTIONS(3715),
    [anon_sym_cmp] = ACTIONS(3715),
    [anon_sym_TILDE_TILDE] = ACTIONS(3715),
    [anon_sym_LT] = ACTIONS(3717),
    [anon_sym_LT_EQ] = ACTIONS(3717),
    [anon_sym_GT_EQ] = ACTIONS(3715),
    [anon_sym_lt] = ACTIONS(3715),
    [anon_sym_le] = ACTIONS(3715),
    [anon_sym_ge] = ACTIONS(3715),
    [anon_sym_gt] = ACTIONS(3715),
    [anon_sym_isa] = ACTIONS(3715),
    [anon_sym_PLUS_PLUS] = ACTIONS(3715),
    [anon_sym_DASH_DASH] = ACTIONS(3715),
    [anon_sym_QMARK] = ACTIONS(3715),
    [anon_sym_EQ_GT] = ACTIONS(3715),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3715),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3623)] = {
    [sym_heredoc_content] = STATE(3623),
    [aux_sym__term_rightward_repeat1] = STATE(3624),
    [anon_sym_COLON] = ACTIONS(3485),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(5387),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(5387),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3624)] = {
    [sym_heredoc_content] = STATE(3624),
    [aux_sym__term_rightward_repeat1] = STATE(3625),
    [anon_sym_COLON] = ACTIONS(3568),
    [anon_sym_EQ] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_COMMA] = ACTIONS(5332),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3568),
    [anon_sym_DASH_GT] = ACTIONS(3568),
    [anon_sym_STAR] = ACTIONS(3570),
    [anon_sym_PERCENT] = ACTIONS(3570),
    [anon_sym_GT] = ACTIONS(3570),
    [anon_sym_LT_LT] = ACTIONS(3570),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3568),
    [anon_sym_PLUS_EQ] = ACTIONS(3568),
    [anon_sym_DASH_EQ] = ACTIONS(3568),
    [anon_sym_DOT_EQ] = ACTIONS(3568),
    [anon_sym_STAR_EQ] = ACTIONS(3568),
    [anon_sym_SLASH_EQ] = ACTIONS(3568),
    [anon_sym_PERCENT_EQ] = ACTIONS(3568),
    [anon_sym_x_EQ] = ACTIONS(3568),
    [anon_sym_AMP_EQ] = ACTIONS(3568),
    [anon_sym_PIPE_EQ] = ACTIONS(3568),
    [anon_sym_CARET_EQ] = ACTIONS(3568),
    [anon_sym_LT_LT_EQ] = ACTIONS(3568),
    [anon_sym_GT_GT_EQ] = ACTIONS(3568),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3568),
    [anon_sym_DOT_DOT] = ACTIONS(3570),
    [anon_sym_STAR_STAR] = ACTIONS(3570),
    [anon_sym_PIPE_PIPE] = ACTIONS(3570),
    [anon_sym_SLASH_SLASH] = ACTIONS(3570),
    [anon_sym_CARET_CARET] = ACTIONS(3568),
    [anon_sym_AMP_AMP] = ACTIONS(3570),
    [anon_sym_PIPE] = ACTIONS(3570),
    [anon_sym_CARET] = ACTIONS(3570),
    [anon_sym_AMP] = ACTIONS(3570),
    [anon_sym_GT_GT2] = ACTIONS(3570),
    [anon_sym_PLUS] = ACTIONS(3570),
    [anon_sym_DASH] = ACTIONS(3570),
    [anon_sym_DOT] = ACTIONS(3570),
    [anon_sym_SLASH] = ACTIONS(3570),
    [anon_sym_x] = ACTIONS(3570),
    [anon_sym_EQ_TILDE] = ACTIONS(3568),
    [anon_sym_BANG_TILDE] = ACTIONS(3568),
    [anon_sym_EQ_EQ] = ACTIONS(3570),
    [anon_sym_BANG_EQ] = ACTIONS(3568),
    [anon_sym_eq] = ACTIONS(3570),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3568),
    [anon_sym_equ] = ACTIONS(3568),
    [anon_sym_eqr] = ACTIONS(3568),
    [anon_sym_ne] = ACTIONS(3568),
    [anon_sym_LT_EQ_GT] = ACTIONS(3568),
    [anon_sym_cmp] = ACTIONS(3568),
    [anon_sym_TILDE_TILDE] = ACTIONS(3568),
    [anon_sym_LT] = ACTIONS(3570),
    [anon_sym_LT_EQ] = ACTIONS(3570),
    [anon_sym_GT_EQ] = ACTIONS(3568),
    [anon_sym_lt] = ACTIONS(3568),
    [anon_sym_le] = ACTIONS(3568),
    [anon_sym_ge] = ACTIONS(3568),
    [anon_sym_gt] = ACTIONS(3568),
    [anon_sym_isa] = ACTIONS(3568),
    [anon_sym_PLUS_PLUS] = ACTIONS(3568),
    [anon_sym_DASH_DASH] = ACTIONS(3568),
    [anon_sym_QMARK] = ACTIONS(3568),
    [anon_sym_EQ_GT] = ACTIONS(5332),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3625)] = {
    [sym_heredoc_content] = STATE(3625),
    [aux_sym__term_rightward_repeat1] = STATE(3625),
    [anon_sym_COLON] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_COMMA] = ACTIONS(5390),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3590),
    [anon_sym_DASH_GT] = ACTIONS(3590),
    [anon_sym_STAR] = ACTIONS(3592),
    [anon_sym_PERCENT] = ACTIONS(3592),
    [anon_sym_GT] = ACTIONS(3592),
    [anon_sym_LT_LT] = ACTIONS(3592),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3590),
    [anon_sym_PLUS_EQ] = ACTIONS(3590),
    [anon_sym_DASH_EQ] = ACTIONS(3590),
    [anon_sym_DOT_EQ] = ACTIONS(3590),
    [anon_sym_STAR_EQ] = ACTIONS(3590),
    [anon_sym_SLASH_EQ] = ACTIONS(3590),
    [anon_sym_PERCENT_EQ] = ACTIONS(3590),
    [anon_sym_x_EQ] = ACTIONS(3590),
    [anon_sym_AMP_EQ] = ACTIONS(3590),
    [anon_sym_PIPE_EQ] = ACTIONS(3590),
    [anon_sym_CARET_EQ] = ACTIONS(3590),
    [anon_sym_LT_LT_EQ] = ACTIONS(3590),
    [anon_sym_GT_GT_EQ] = ACTIONS(3590),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3590),
    [anon_sym_DOT_DOT] = ACTIONS(3592),
    [anon_sym_STAR_STAR] = ACTIONS(3592),
    [anon_sym_PIPE_PIPE] = ACTIONS(3592),
    [anon_sym_SLASH_SLASH] = ACTIONS(3592),
    [anon_sym_CARET_CARET] = ACTIONS(3590),
    [anon_sym_AMP_AMP] = ACTIONS(3592),
    [anon_sym_PIPE] = ACTIONS(3592),
    [anon_sym_CARET] = ACTIONS(3592),
    [anon_sym_AMP] = ACTIONS(3592),
    [anon_sym_GT_GT2] = ACTIONS(3592),
    [anon_sym_PLUS] = ACTIONS(3592),
    [anon_sym_DASH] = ACTIONS(3592),
    [anon_sym_DOT] = ACTIONS(3592),
    [anon_sym_SLASH] = ACTIONS(3592),
    [anon_sym_x] = ACTIONS(3592),
    [anon_sym_EQ_TILDE] = ACTIONS(3590),
    [anon_sym_BANG_TILDE] = ACTIONS(3590),
    [anon_sym_EQ_EQ] = ACTIONS(3592),
    [anon_sym_BANG_EQ] = ACTIONS(3590),
    [anon_sym_eq] = ACTIONS(3592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3590),
    [anon_sym_equ] = ACTIONS(3590),
    [anon_sym_eqr] = ACTIONS(3590),
    [anon_sym_ne] = ACTIONS(3590),
    [anon_sym_LT_EQ_GT] = ACTIONS(3590),
    [anon_sym_cmp] = ACTIONS(3590),
    [anon_sym_TILDE_TILDE] = ACTIONS(3590),
    [anon_sym_LT] = ACTIONS(3592),
    [anon_sym_LT_EQ] = ACTIONS(3592),
    [anon_sym_GT_EQ] = ACTIONS(3590),
    [anon_sym_lt] = ACTIONS(3590),
    [anon_sym_le] = ACTIONS(3590),
    [anon_sym_ge] = ACTIONS(3590),
    [anon_sym_gt] = ACTIONS(3590),
    [anon_sym_isa] = ACTIONS(3590),
    [anon_sym_PLUS_PLUS] = ACTIONS(3590),
    [anon_sym_DASH_DASH] = ACTIONS(3590),
    [anon_sym_QMARK] = ACTIONS(3590),
    [anon_sym_EQ_GT] = ACTIONS(5390),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3626)] = {
    [sym_heredoc_content] = STATE(3626),
    [anon_sym_SEMI] = ACTIONS(3893),
    [anon_sym_EQ] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_COMMA] = ACTIONS(3893),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3893),
    [anon_sym_DASH_GT] = ACTIONS(3893),
    [anon_sym_STAR] = ACTIONS(3895),
    [anon_sym_PERCENT] = ACTIONS(3895),
    [anon_sym_GT] = ACTIONS(3895),
    [anon_sym_LT_LT] = ACTIONS(3895),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3893),
    [anon_sym_PLUS_EQ] = ACTIONS(3893),
    [anon_sym_DASH_EQ] = ACTIONS(3893),
    [anon_sym_DOT_EQ] = ACTIONS(3893),
    [anon_sym_STAR_EQ] = ACTIONS(3893),
    [anon_sym_SLASH_EQ] = ACTIONS(3893),
    [anon_sym_PERCENT_EQ] = ACTIONS(3893),
    [anon_sym_x_EQ] = ACTIONS(3893),
    [anon_sym_AMP_EQ] = ACTIONS(3893),
    [anon_sym_PIPE_EQ] = ACTIONS(3893),
    [anon_sym_CARET_EQ] = ACTIONS(3893),
    [anon_sym_LT_LT_EQ] = ACTIONS(3893),
    [anon_sym_GT_GT_EQ] = ACTIONS(3893),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3893),
    [anon_sym_DOT_DOT] = ACTIONS(3895),
    [anon_sym_STAR_STAR] = ACTIONS(3895),
    [anon_sym_PIPE_PIPE] = ACTIONS(3895),
    [anon_sym_SLASH_SLASH] = ACTIONS(3895),
    [anon_sym_CARET_CARET] = ACTIONS(3893),
    [anon_sym_AMP_AMP] = ACTIONS(3895),
    [anon_sym_PIPE] = ACTIONS(3895),
    [anon_sym_CARET] = ACTIONS(3895),
    [anon_sym_AMP] = ACTIONS(3895),
    [anon_sym_GT_GT2] = ACTIONS(3895),
    [anon_sym_PLUS] = ACTIONS(3895),
    [anon_sym_DASH] = ACTIONS(3895),
    [anon_sym_DOT] = ACTIONS(3895),
    [anon_sym_SLASH] = ACTIONS(3895),
    [anon_sym_x] = ACTIONS(3895),
    [anon_sym_EQ_TILDE] = ACTIONS(3893),
    [anon_sym_BANG_TILDE] = ACTIONS(3893),
    [anon_sym_EQ_EQ] = ACTIONS(3895),
    [anon_sym_BANG_EQ] = ACTIONS(3893),
    [anon_sym_eq] = ACTIONS(3895),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3893),
    [anon_sym_equ] = ACTIONS(3893),
    [anon_sym_eqr] = ACTIONS(3893),
    [anon_sym_ne] = ACTIONS(3893),
    [anon_sym_LT_EQ_GT] = ACTIONS(3893),
    [anon_sym_cmp] = ACTIONS(3893),
    [anon_sym_TILDE_TILDE] = ACTIONS(3893),
    [anon_sym_LT] = ACTIONS(3895),
    [anon_sym_LT_EQ] = ACTIONS(3895),
    [anon_sym_GT_EQ] = ACTIONS(3893),
    [anon_sym_lt] = ACTIONS(3893),
    [anon_sym_le] = ACTIONS(3893),
    [anon_sym_ge] = ACTIONS(3893),
    [anon_sym_gt] = ACTIONS(3893),
    [anon_sym_isa] = ACTIONS(3893),
    [anon_sym_PLUS_PLUS] = ACTIONS(3893),
    [anon_sym_DASH_DASH] = ACTIONS(3893),
    [anon_sym_QMARK] = ACTIONS(3893),
    [anon_sym_EQ_GT] = ACTIONS(3893),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3893),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3627)] = {
    [sym_heredoc_content] = STATE(3627),
    [anon_sym_SEMI] = ACTIONS(4063),
    [anon_sym_EQ] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_COMMA] = ACTIONS(4063),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4063),
    [anon_sym_DASH_GT] = ACTIONS(4063),
    [anon_sym_STAR] = ACTIONS(4065),
    [anon_sym_PERCENT] = ACTIONS(4065),
    [anon_sym_GT] = ACTIONS(4065),
    [anon_sym_LT_LT] = ACTIONS(4065),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4063),
    [anon_sym_PLUS_EQ] = ACTIONS(4063),
    [anon_sym_DASH_EQ] = ACTIONS(4063),
    [anon_sym_DOT_EQ] = ACTIONS(4063),
    [anon_sym_STAR_EQ] = ACTIONS(4063),
    [anon_sym_SLASH_EQ] = ACTIONS(4063),
    [anon_sym_PERCENT_EQ] = ACTIONS(4063),
    [anon_sym_x_EQ] = ACTIONS(4063),
    [anon_sym_AMP_EQ] = ACTIONS(4063),
    [anon_sym_PIPE_EQ] = ACTIONS(4063),
    [anon_sym_CARET_EQ] = ACTIONS(4063),
    [anon_sym_LT_LT_EQ] = ACTIONS(4063),
    [anon_sym_GT_GT_EQ] = ACTIONS(4063),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4063),
    [anon_sym_DOT_DOT] = ACTIONS(4065),
    [anon_sym_STAR_STAR] = ACTIONS(4065),
    [anon_sym_PIPE_PIPE] = ACTIONS(4065),
    [anon_sym_SLASH_SLASH] = ACTIONS(4065),
    [anon_sym_CARET_CARET] = ACTIONS(4063),
    [anon_sym_AMP_AMP] = ACTIONS(4065),
    [anon_sym_PIPE] = ACTIONS(4065),
    [anon_sym_CARET] = ACTIONS(4065),
    [anon_sym_AMP] = ACTIONS(4065),
    [anon_sym_GT_GT2] = ACTIONS(4065),
    [anon_sym_PLUS] = ACTIONS(4065),
    [anon_sym_DASH] = ACTIONS(4065),
    [anon_sym_DOT] = ACTIONS(4065),
    [anon_sym_SLASH] = ACTIONS(4065),
    [anon_sym_x] = ACTIONS(4065),
    [anon_sym_EQ_TILDE] = ACTIONS(4063),
    [anon_sym_BANG_TILDE] = ACTIONS(4063),
    [anon_sym_EQ_EQ] = ACTIONS(4065),
    [anon_sym_BANG_EQ] = ACTIONS(4063),
    [anon_sym_eq] = ACTIONS(4065),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4063),
    [anon_sym_equ] = ACTIONS(4063),
    [anon_sym_eqr] = ACTIONS(4063),
    [anon_sym_ne] = ACTIONS(4063),
    [anon_sym_LT_EQ_GT] = ACTIONS(4063),
    [anon_sym_cmp] = ACTIONS(4063),
    [anon_sym_TILDE_TILDE] = ACTIONS(4063),
    [anon_sym_LT] = ACTIONS(4065),
    [anon_sym_LT_EQ] = ACTIONS(4065),
    [anon_sym_GT_EQ] = ACTIONS(4063),
    [anon_sym_lt] = ACTIONS(4063),
    [anon_sym_le] = ACTIONS(4063),
    [anon_sym_ge] = ACTIONS(4063),
    [anon_sym_gt] = ACTIONS(4063),
    [anon_sym_isa] = ACTIONS(4063),
    [anon_sym_PLUS_PLUS] = ACTIONS(4063),
    [anon_sym_DASH_DASH] = ACTIONS(4063),
    [anon_sym_QMARK] = ACTIONS(4063),
    [anon_sym_EQ_GT] = ACTIONS(4063),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4063),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3628)] = {
    [sym_heredoc_content] = STATE(3628),
    [anon_sym_SEMI] = ACTIONS(3877),
    [anon_sym_EQ] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_COMMA] = ACTIONS(3877),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3877),
    [anon_sym_DASH_GT] = ACTIONS(3877),
    [anon_sym_STAR] = ACTIONS(3879),
    [anon_sym_PERCENT] = ACTIONS(3879),
    [anon_sym_GT] = ACTIONS(3879),
    [anon_sym_LT_LT] = ACTIONS(3879),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3877),
    [anon_sym_PLUS_EQ] = ACTIONS(3877),
    [anon_sym_DASH_EQ] = ACTIONS(3877),
    [anon_sym_DOT_EQ] = ACTIONS(3877),
    [anon_sym_STAR_EQ] = ACTIONS(3877),
    [anon_sym_SLASH_EQ] = ACTIONS(3877),
    [anon_sym_PERCENT_EQ] = ACTIONS(3877),
    [anon_sym_x_EQ] = ACTIONS(3877),
    [anon_sym_AMP_EQ] = ACTIONS(3877),
    [anon_sym_PIPE_EQ] = ACTIONS(3877),
    [anon_sym_CARET_EQ] = ACTIONS(3877),
    [anon_sym_LT_LT_EQ] = ACTIONS(3877),
    [anon_sym_GT_GT_EQ] = ACTIONS(3877),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3877),
    [anon_sym_DOT_DOT] = ACTIONS(3879),
    [anon_sym_STAR_STAR] = ACTIONS(3879),
    [anon_sym_PIPE_PIPE] = ACTIONS(3879),
    [anon_sym_SLASH_SLASH] = ACTIONS(3879),
    [anon_sym_CARET_CARET] = ACTIONS(3877),
    [anon_sym_AMP_AMP] = ACTIONS(3879),
    [anon_sym_PIPE] = ACTIONS(3879),
    [anon_sym_CARET] = ACTIONS(3879),
    [anon_sym_AMP] = ACTIONS(3879),
    [anon_sym_GT_GT2] = ACTIONS(3879),
    [anon_sym_PLUS] = ACTIONS(3879),
    [anon_sym_DASH] = ACTIONS(3879),
    [anon_sym_DOT] = ACTIONS(3879),
    [anon_sym_SLASH] = ACTIONS(3879),
    [anon_sym_x] = ACTIONS(3879),
    [anon_sym_EQ_TILDE] = ACTIONS(3877),
    [anon_sym_BANG_TILDE] = ACTIONS(3877),
    [anon_sym_EQ_EQ] = ACTIONS(3879),
    [anon_sym_BANG_EQ] = ACTIONS(3877),
    [anon_sym_eq] = ACTIONS(3879),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3877),
    [anon_sym_equ] = ACTIONS(3877),
    [anon_sym_eqr] = ACTIONS(3877),
    [anon_sym_ne] = ACTIONS(3877),
    [anon_sym_LT_EQ_GT] = ACTIONS(3877),
    [anon_sym_cmp] = ACTIONS(3877),
    [anon_sym_TILDE_TILDE] = ACTIONS(3877),
    [anon_sym_LT] = ACTIONS(3879),
    [anon_sym_LT_EQ] = ACTIONS(3879),
    [anon_sym_GT_EQ] = ACTIONS(3877),
    [anon_sym_lt] = ACTIONS(3877),
    [anon_sym_le] = ACTIONS(3877),
    [anon_sym_ge] = ACTIONS(3877),
    [anon_sym_gt] = ACTIONS(3877),
    [anon_sym_isa] = ACTIONS(3877),
    [anon_sym_PLUS_PLUS] = ACTIONS(3877),
    [anon_sym_DASH_DASH] = ACTIONS(3877),
    [anon_sym_QMARK] = ACTIONS(3877),
    [anon_sym_EQ_GT] = ACTIONS(3877),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3877),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3629)] = {
    [sym_heredoc_content] = STATE(3629),
    [anon_sym_SEMI] = ACTIONS(3989),
    [anon_sym_EQ] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_COMMA] = ACTIONS(3989),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3989),
    [anon_sym_DASH_GT] = ACTIONS(3989),
    [anon_sym_STAR] = ACTIONS(3991),
    [anon_sym_PERCENT] = ACTIONS(3991),
    [anon_sym_GT] = ACTIONS(3991),
    [anon_sym_LT_LT] = ACTIONS(3991),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3989),
    [anon_sym_PLUS_EQ] = ACTIONS(3989),
    [anon_sym_DASH_EQ] = ACTIONS(3989),
    [anon_sym_DOT_EQ] = ACTIONS(3989),
    [anon_sym_STAR_EQ] = ACTIONS(3989),
    [anon_sym_SLASH_EQ] = ACTIONS(3989),
    [anon_sym_PERCENT_EQ] = ACTIONS(3989),
    [anon_sym_x_EQ] = ACTIONS(3989),
    [anon_sym_AMP_EQ] = ACTIONS(3989),
    [anon_sym_PIPE_EQ] = ACTIONS(3989),
    [anon_sym_CARET_EQ] = ACTIONS(3989),
    [anon_sym_LT_LT_EQ] = ACTIONS(3989),
    [anon_sym_GT_GT_EQ] = ACTIONS(3989),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3989),
    [anon_sym_DOT_DOT] = ACTIONS(3991),
    [anon_sym_STAR_STAR] = ACTIONS(3991),
    [anon_sym_PIPE_PIPE] = ACTIONS(3991),
    [anon_sym_SLASH_SLASH] = ACTIONS(3991),
    [anon_sym_CARET_CARET] = ACTIONS(3989),
    [anon_sym_AMP_AMP] = ACTIONS(3991),
    [anon_sym_PIPE] = ACTIONS(3991),
    [anon_sym_CARET] = ACTIONS(3991),
    [anon_sym_AMP] = ACTIONS(3991),
    [anon_sym_GT_GT2] = ACTIONS(3991),
    [anon_sym_PLUS] = ACTIONS(3991),
    [anon_sym_DASH] = ACTIONS(3991),
    [anon_sym_DOT] = ACTIONS(3991),
    [anon_sym_SLASH] = ACTIONS(3991),
    [anon_sym_x] = ACTIONS(3991),
    [anon_sym_EQ_TILDE] = ACTIONS(3989),
    [anon_sym_BANG_TILDE] = ACTIONS(3989),
    [anon_sym_EQ_EQ] = ACTIONS(3991),
    [anon_sym_BANG_EQ] = ACTIONS(3989),
    [anon_sym_eq] = ACTIONS(3991),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3989),
    [anon_sym_equ] = ACTIONS(3989),
    [anon_sym_eqr] = ACTIONS(3989),
    [anon_sym_ne] = ACTIONS(3989),
    [anon_sym_LT_EQ_GT] = ACTIONS(3989),
    [anon_sym_cmp] = ACTIONS(3989),
    [anon_sym_TILDE_TILDE] = ACTIONS(3989),
    [anon_sym_LT] = ACTIONS(3991),
    [anon_sym_LT_EQ] = ACTIONS(3991),
    [anon_sym_GT_EQ] = ACTIONS(3989),
    [anon_sym_lt] = ACTIONS(3989),
    [anon_sym_le] = ACTIONS(3989),
    [anon_sym_ge] = ACTIONS(3989),
    [anon_sym_gt] = ACTIONS(3989),
    [anon_sym_isa] = ACTIONS(3989),
    [anon_sym_PLUS_PLUS] = ACTIONS(3989),
    [anon_sym_DASH_DASH] = ACTIONS(3989),
    [anon_sym_QMARK] = ACTIONS(3989),
    [anon_sym_EQ_GT] = ACTIONS(3989),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3989),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3630)] = {
    [sym_heredoc_content] = STATE(3630),
    [anon_sym_SEMI] = ACTIONS(4115),
    [anon_sym_EQ] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_COMMA] = ACTIONS(4115),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4115),
    [anon_sym_DASH_GT] = ACTIONS(4115),
    [anon_sym_STAR] = ACTIONS(4117),
    [anon_sym_PERCENT] = ACTIONS(4117),
    [anon_sym_GT] = ACTIONS(4117),
    [anon_sym_LT_LT] = ACTIONS(4117),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4115),
    [anon_sym_PLUS_EQ] = ACTIONS(4115),
    [anon_sym_DASH_EQ] = ACTIONS(4115),
    [anon_sym_DOT_EQ] = ACTIONS(4115),
    [anon_sym_STAR_EQ] = ACTIONS(4115),
    [anon_sym_SLASH_EQ] = ACTIONS(4115),
    [anon_sym_PERCENT_EQ] = ACTIONS(4115),
    [anon_sym_x_EQ] = ACTIONS(4115),
    [anon_sym_AMP_EQ] = ACTIONS(4115),
    [anon_sym_PIPE_EQ] = ACTIONS(4115),
    [anon_sym_CARET_EQ] = ACTIONS(4115),
    [anon_sym_LT_LT_EQ] = ACTIONS(4115),
    [anon_sym_GT_GT_EQ] = ACTIONS(4115),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4115),
    [anon_sym_DOT_DOT] = ACTIONS(4117),
    [anon_sym_STAR_STAR] = ACTIONS(4117),
    [anon_sym_PIPE_PIPE] = ACTIONS(4117),
    [anon_sym_SLASH_SLASH] = ACTIONS(4117),
    [anon_sym_CARET_CARET] = ACTIONS(4115),
    [anon_sym_AMP_AMP] = ACTIONS(4117),
    [anon_sym_PIPE] = ACTIONS(4117),
    [anon_sym_CARET] = ACTIONS(4117),
    [anon_sym_AMP] = ACTIONS(4117),
    [anon_sym_GT_GT2] = ACTIONS(4117),
    [anon_sym_PLUS] = ACTIONS(4117),
    [anon_sym_DASH] = ACTIONS(4117),
    [anon_sym_DOT] = ACTIONS(4117),
    [anon_sym_SLASH] = ACTIONS(4117),
    [anon_sym_x] = ACTIONS(4117),
    [anon_sym_EQ_TILDE] = ACTIONS(4115),
    [anon_sym_BANG_TILDE] = ACTIONS(4115),
    [anon_sym_EQ_EQ] = ACTIONS(4117),
    [anon_sym_BANG_EQ] = ACTIONS(4115),
    [anon_sym_eq] = ACTIONS(4117),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4115),
    [anon_sym_equ] = ACTIONS(4115),
    [anon_sym_eqr] = ACTIONS(4115),
    [anon_sym_ne] = ACTIONS(4115),
    [anon_sym_LT_EQ_GT] = ACTIONS(4115),
    [anon_sym_cmp] = ACTIONS(4115),
    [anon_sym_TILDE_TILDE] = ACTIONS(4115),
    [anon_sym_LT] = ACTIONS(4117),
    [anon_sym_LT_EQ] = ACTIONS(4117),
    [anon_sym_GT_EQ] = ACTIONS(4115),
    [anon_sym_lt] = ACTIONS(4115),
    [anon_sym_le] = ACTIONS(4115),
    [anon_sym_ge] = ACTIONS(4115),
    [anon_sym_gt] = ACTIONS(4115),
    [anon_sym_isa] = ACTIONS(4115),
    [anon_sym_PLUS_PLUS] = ACTIONS(4115),
    [anon_sym_DASH_DASH] = ACTIONS(4115),
    [anon_sym_QMARK] = ACTIONS(4115),
    [anon_sym_EQ_GT] = ACTIONS(4115),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4115),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3631)] = {
    [sym_heredoc_content] = STATE(3631),
    [anon_sym_SEMI] = ACTIONS(3807),
    [anon_sym_EQ] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_COMMA] = ACTIONS(3807),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3807),
    [anon_sym_DASH_GT] = ACTIONS(3807),
    [anon_sym_STAR] = ACTIONS(3809),
    [anon_sym_PERCENT] = ACTIONS(3809),
    [anon_sym_GT] = ACTIONS(3809),
    [anon_sym_LT_LT] = ACTIONS(3809),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3807),
    [anon_sym_PLUS_EQ] = ACTIONS(3807),
    [anon_sym_DASH_EQ] = ACTIONS(3807),
    [anon_sym_DOT_EQ] = ACTIONS(3807),
    [anon_sym_STAR_EQ] = ACTIONS(3807),
    [anon_sym_SLASH_EQ] = ACTIONS(3807),
    [anon_sym_PERCENT_EQ] = ACTIONS(3807),
    [anon_sym_x_EQ] = ACTIONS(3807),
    [anon_sym_AMP_EQ] = ACTIONS(3807),
    [anon_sym_PIPE_EQ] = ACTIONS(3807),
    [anon_sym_CARET_EQ] = ACTIONS(3807),
    [anon_sym_LT_LT_EQ] = ACTIONS(3807),
    [anon_sym_GT_GT_EQ] = ACTIONS(3807),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3807),
    [anon_sym_DOT_DOT] = ACTIONS(3809),
    [anon_sym_STAR_STAR] = ACTIONS(3809),
    [anon_sym_PIPE_PIPE] = ACTIONS(3809),
    [anon_sym_SLASH_SLASH] = ACTIONS(3809),
    [anon_sym_CARET_CARET] = ACTIONS(3807),
    [anon_sym_AMP_AMP] = ACTIONS(3809),
    [anon_sym_PIPE] = ACTIONS(3809),
    [anon_sym_CARET] = ACTIONS(3809),
    [anon_sym_AMP] = ACTIONS(3809),
    [anon_sym_GT_GT2] = ACTIONS(3809),
    [anon_sym_PLUS] = ACTIONS(3809),
    [anon_sym_DASH] = ACTIONS(3809),
    [anon_sym_DOT] = ACTIONS(3809),
    [anon_sym_SLASH] = ACTIONS(3809),
    [anon_sym_x] = ACTIONS(3809),
    [anon_sym_EQ_TILDE] = ACTIONS(3807),
    [anon_sym_BANG_TILDE] = ACTIONS(3807),
    [anon_sym_EQ_EQ] = ACTIONS(3809),
    [anon_sym_BANG_EQ] = ACTIONS(3807),
    [anon_sym_eq] = ACTIONS(3809),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3807),
    [anon_sym_equ] = ACTIONS(3807),
    [anon_sym_eqr] = ACTIONS(3807),
    [anon_sym_ne] = ACTIONS(3807),
    [anon_sym_LT_EQ_GT] = ACTIONS(3807),
    [anon_sym_cmp] = ACTIONS(3807),
    [anon_sym_TILDE_TILDE] = ACTIONS(3807),
    [anon_sym_LT] = ACTIONS(3809),
    [anon_sym_LT_EQ] = ACTIONS(3809),
    [anon_sym_GT_EQ] = ACTIONS(3807),
    [anon_sym_lt] = ACTIONS(3807),
    [anon_sym_le] = ACTIONS(3807),
    [anon_sym_ge] = ACTIONS(3807),
    [anon_sym_gt] = ACTIONS(3807),
    [anon_sym_isa] = ACTIONS(3807),
    [anon_sym_PLUS_PLUS] = ACTIONS(3807),
    [anon_sym_DASH_DASH] = ACTIONS(3807),
    [anon_sym_QMARK] = ACTIONS(3807),
    [anon_sym_EQ_GT] = ACTIONS(3807),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3807),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3632)] = {
    [sym_heredoc_content] = STATE(3632),
    [anon_sym_SEMI] = ACTIONS(2536),
    [anon_sym_EQ] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_COMMA] = ACTIONS(2536),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2536),
    [anon_sym_DASH_GT] = ACTIONS(2536),
    [anon_sym_STAR] = ACTIONS(2534),
    [anon_sym_PERCENT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2534),
    [anon_sym_LT_LT] = ACTIONS(2534),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2536),
    [anon_sym_PLUS_EQ] = ACTIONS(2536),
    [anon_sym_DASH_EQ] = ACTIONS(2536),
    [anon_sym_DOT_EQ] = ACTIONS(2536),
    [anon_sym_STAR_EQ] = ACTIONS(2536),
    [anon_sym_SLASH_EQ] = ACTIONS(2536),
    [anon_sym_PERCENT_EQ] = ACTIONS(2536),
    [anon_sym_x_EQ] = ACTIONS(2536),
    [anon_sym_AMP_EQ] = ACTIONS(2536),
    [anon_sym_PIPE_EQ] = ACTIONS(2536),
    [anon_sym_CARET_EQ] = ACTIONS(2536),
    [anon_sym_LT_LT_EQ] = ACTIONS(2536),
    [anon_sym_GT_GT_EQ] = ACTIONS(2536),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2536),
    [anon_sym_DOT_DOT] = ACTIONS(2534),
    [anon_sym_STAR_STAR] = ACTIONS(2534),
    [anon_sym_PIPE_PIPE] = ACTIONS(2534),
    [anon_sym_SLASH_SLASH] = ACTIONS(2534),
    [anon_sym_CARET_CARET] = ACTIONS(2536),
    [anon_sym_AMP_AMP] = ACTIONS(2534),
    [anon_sym_PIPE] = ACTIONS(2534),
    [anon_sym_CARET] = ACTIONS(2534),
    [anon_sym_AMP] = ACTIONS(2534),
    [anon_sym_GT_GT2] = ACTIONS(2534),
    [anon_sym_PLUS] = ACTIONS(2534),
    [anon_sym_DASH] = ACTIONS(2534),
    [anon_sym_DOT] = ACTIONS(2534),
    [anon_sym_SLASH] = ACTIONS(2534),
    [anon_sym_x] = ACTIONS(2534),
    [anon_sym_EQ_TILDE] = ACTIONS(2536),
    [anon_sym_BANG_TILDE] = ACTIONS(2536),
    [anon_sym_EQ_EQ] = ACTIONS(2534),
    [anon_sym_BANG_EQ] = ACTIONS(2536),
    [anon_sym_eq] = ACTIONS(2534),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2536),
    [anon_sym_equ] = ACTIONS(2536),
    [anon_sym_eqr] = ACTIONS(2536),
    [anon_sym_ne] = ACTIONS(2536),
    [anon_sym_LT_EQ_GT] = ACTIONS(2536),
    [anon_sym_cmp] = ACTIONS(2536),
    [anon_sym_TILDE_TILDE] = ACTIONS(2536),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_LT_EQ] = ACTIONS(2534),
    [anon_sym_GT_EQ] = ACTIONS(2536),
    [anon_sym_lt] = ACTIONS(2536),
    [anon_sym_le] = ACTIONS(2536),
    [anon_sym_ge] = ACTIONS(2536),
    [anon_sym_gt] = ACTIONS(2536),
    [anon_sym_isa] = ACTIONS(2536),
    [anon_sym_PLUS_PLUS] = ACTIONS(2536),
    [anon_sym_DASH_DASH] = ACTIONS(2536),
    [anon_sym_QMARK] = ACTIONS(2536),
    [anon_sym_EQ_GT] = ACTIONS(2536),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2536),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3633)] = {
    [sym_heredoc_content] = STATE(3633),
    [anon_sym_SEMI] = ACTIONS(4041),
    [anon_sym_EQ] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_COMMA] = ACTIONS(4041),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4041),
    [anon_sym_DASH_GT] = ACTIONS(4041),
    [anon_sym_STAR] = ACTIONS(4043),
    [anon_sym_PERCENT] = ACTIONS(4043),
    [anon_sym_GT] = ACTIONS(4043),
    [anon_sym_LT_LT] = ACTIONS(4043),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4041),
    [anon_sym_PLUS_EQ] = ACTIONS(4041),
    [anon_sym_DASH_EQ] = ACTIONS(4041),
    [anon_sym_DOT_EQ] = ACTIONS(4041),
    [anon_sym_STAR_EQ] = ACTIONS(4041),
    [anon_sym_SLASH_EQ] = ACTIONS(4041),
    [anon_sym_PERCENT_EQ] = ACTIONS(4041),
    [anon_sym_x_EQ] = ACTIONS(4041),
    [anon_sym_AMP_EQ] = ACTIONS(4041),
    [anon_sym_PIPE_EQ] = ACTIONS(4041),
    [anon_sym_CARET_EQ] = ACTIONS(4041),
    [anon_sym_LT_LT_EQ] = ACTIONS(4041),
    [anon_sym_GT_GT_EQ] = ACTIONS(4041),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4041),
    [anon_sym_DOT_DOT] = ACTIONS(4043),
    [anon_sym_STAR_STAR] = ACTIONS(4043),
    [anon_sym_PIPE_PIPE] = ACTIONS(4043),
    [anon_sym_SLASH_SLASH] = ACTIONS(4043),
    [anon_sym_CARET_CARET] = ACTIONS(4041),
    [anon_sym_AMP_AMP] = ACTIONS(4043),
    [anon_sym_PIPE] = ACTIONS(4043),
    [anon_sym_CARET] = ACTIONS(4043),
    [anon_sym_AMP] = ACTIONS(4043),
    [anon_sym_GT_GT2] = ACTIONS(4043),
    [anon_sym_PLUS] = ACTIONS(4043),
    [anon_sym_DASH] = ACTIONS(4043),
    [anon_sym_DOT] = ACTIONS(4043),
    [anon_sym_SLASH] = ACTIONS(4043),
    [anon_sym_x] = ACTIONS(4043),
    [anon_sym_EQ_TILDE] = ACTIONS(4041),
    [anon_sym_BANG_TILDE] = ACTIONS(4041),
    [anon_sym_EQ_EQ] = ACTIONS(4043),
    [anon_sym_BANG_EQ] = ACTIONS(4041),
    [anon_sym_eq] = ACTIONS(4043),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4041),
    [anon_sym_equ] = ACTIONS(4041),
    [anon_sym_eqr] = ACTIONS(4041),
    [anon_sym_ne] = ACTIONS(4041),
    [anon_sym_LT_EQ_GT] = ACTIONS(4041),
    [anon_sym_cmp] = ACTIONS(4041),
    [anon_sym_TILDE_TILDE] = ACTIONS(4041),
    [anon_sym_LT] = ACTIONS(4043),
    [anon_sym_LT_EQ] = ACTIONS(4043),
    [anon_sym_GT_EQ] = ACTIONS(4041),
    [anon_sym_lt] = ACTIONS(4041),
    [anon_sym_le] = ACTIONS(4041),
    [anon_sym_ge] = ACTIONS(4041),
    [anon_sym_gt] = ACTIONS(4041),
    [anon_sym_isa] = ACTIONS(4041),
    [anon_sym_PLUS_PLUS] = ACTIONS(4041),
    [anon_sym_DASH_DASH] = ACTIONS(4041),
    [anon_sym_QMARK] = ACTIONS(4041),
    [anon_sym_EQ_GT] = ACTIONS(4041),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4041),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3634)] = {
    [sym_heredoc_content] = STATE(3634),
    [anon_sym_SEMI] = ACTIONS(4055),
    [anon_sym_EQ] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_COMMA] = ACTIONS(4055),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4055),
    [anon_sym_DASH_GT] = ACTIONS(4055),
    [anon_sym_STAR] = ACTIONS(4057),
    [anon_sym_PERCENT] = ACTIONS(4057),
    [anon_sym_GT] = ACTIONS(4057),
    [anon_sym_LT_LT] = ACTIONS(4057),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4055),
    [anon_sym_PLUS_EQ] = ACTIONS(4055),
    [anon_sym_DASH_EQ] = ACTIONS(4055),
    [anon_sym_DOT_EQ] = ACTIONS(4055),
    [anon_sym_STAR_EQ] = ACTIONS(4055),
    [anon_sym_SLASH_EQ] = ACTIONS(4055),
    [anon_sym_PERCENT_EQ] = ACTIONS(4055),
    [anon_sym_x_EQ] = ACTIONS(4055),
    [anon_sym_AMP_EQ] = ACTIONS(4055),
    [anon_sym_PIPE_EQ] = ACTIONS(4055),
    [anon_sym_CARET_EQ] = ACTIONS(4055),
    [anon_sym_LT_LT_EQ] = ACTIONS(4055),
    [anon_sym_GT_GT_EQ] = ACTIONS(4055),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4055),
    [anon_sym_DOT_DOT] = ACTIONS(4057),
    [anon_sym_STAR_STAR] = ACTIONS(4057),
    [anon_sym_PIPE_PIPE] = ACTIONS(4057),
    [anon_sym_SLASH_SLASH] = ACTIONS(4057),
    [anon_sym_CARET_CARET] = ACTIONS(4055),
    [anon_sym_AMP_AMP] = ACTIONS(4057),
    [anon_sym_PIPE] = ACTIONS(4057),
    [anon_sym_CARET] = ACTIONS(4057),
    [anon_sym_AMP] = ACTIONS(4057),
    [anon_sym_GT_GT2] = ACTIONS(4057),
    [anon_sym_PLUS] = ACTIONS(4057),
    [anon_sym_DASH] = ACTIONS(4057),
    [anon_sym_DOT] = ACTIONS(4057),
    [anon_sym_SLASH] = ACTIONS(4057),
    [anon_sym_x] = ACTIONS(4057),
    [anon_sym_EQ_TILDE] = ACTIONS(4055),
    [anon_sym_BANG_TILDE] = ACTIONS(4055),
    [anon_sym_EQ_EQ] = ACTIONS(4057),
    [anon_sym_BANG_EQ] = ACTIONS(4055),
    [anon_sym_eq] = ACTIONS(4057),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4055),
    [anon_sym_equ] = ACTIONS(4055),
    [anon_sym_eqr] = ACTIONS(4055),
    [anon_sym_ne] = ACTIONS(4055),
    [anon_sym_LT_EQ_GT] = ACTIONS(4055),
    [anon_sym_cmp] = ACTIONS(4055),
    [anon_sym_TILDE_TILDE] = ACTIONS(4055),
    [anon_sym_LT] = ACTIONS(4057),
    [anon_sym_LT_EQ] = ACTIONS(4057),
    [anon_sym_GT_EQ] = ACTIONS(4055),
    [anon_sym_lt] = ACTIONS(4055),
    [anon_sym_le] = ACTIONS(4055),
    [anon_sym_ge] = ACTIONS(4055),
    [anon_sym_gt] = ACTIONS(4055),
    [anon_sym_isa] = ACTIONS(4055),
    [anon_sym_PLUS_PLUS] = ACTIONS(4055),
    [anon_sym_DASH_DASH] = ACTIONS(4055),
    [anon_sym_QMARK] = ACTIONS(4055),
    [anon_sym_EQ_GT] = ACTIONS(4055),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4055),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3635)] = {
    [sym_heredoc_content] = STATE(3635),
    [anon_sym_SEMI] = ACTIONS(3865),
    [anon_sym_EQ] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_COMMA] = ACTIONS(3865),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3865),
    [anon_sym_DASH_GT] = ACTIONS(3865),
    [anon_sym_STAR] = ACTIONS(3867),
    [anon_sym_PERCENT] = ACTIONS(3867),
    [anon_sym_GT] = ACTIONS(3867),
    [anon_sym_LT_LT] = ACTIONS(3867),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3865),
    [anon_sym_PLUS_EQ] = ACTIONS(3865),
    [anon_sym_DASH_EQ] = ACTIONS(3865),
    [anon_sym_DOT_EQ] = ACTIONS(3865),
    [anon_sym_STAR_EQ] = ACTIONS(3865),
    [anon_sym_SLASH_EQ] = ACTIONS(3865),
    [anon_sym_PERCENT_EQ] = ACTIONS(3865),
    [anon_sym_x_EQ] = ACTIONS(3865),
    [anon_sym_AMP_EQ] = ACTIONS(3865),
    [anon_sym_PIPE_EQ] = ACTIONS(3865),
    [anon_sym_CARET_EQ] = ACTIONS(3865),
    [anon_sym_LT_LT_EQ] = ACTIONS(3865),
    [anon_sym_GT_GT_EQ] = ACTIONS(3865),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3865),
    [anon_sym_DOT_DOT] = ACTIONS(3867),
    [anon_sym_STAR_STAR] = ACTIONS(3867),
    [anon_sym_PIPE_PIPE] = ACTIONS(3867),
    [anon_sym_SLASH_SLASH] = ACTIONS(3867),
    [anon_sym_CARET_CARET] = ACTIONS(3865),
    [anon_sym_AMP_AMP] = ACTIONS(3867),
    [anon_sym_PIPE] = ACTIONS(3867),
    [anon_sym_CARET] = ACTIONS(3867),
    [anon_sym_AMP] = ACTIONS(3867),
    [anon_sym_GT_GT2] = ACTIONS(3867),
    [anon_sym_PLUS] = ACTIONS(3867),
    [anon_sym_DASH] = ACTIONS(3867),
    [anon_sym_DOT] = ACTIONS(3867),
    [anon_sym_SLASH] = ACTIONS(3867),
    [anon_sym_x] = ACTIONS(3867),
    [anon_sym_EQ_TILDE] = ACTIONS(3865),
    [anon_sym_BANG_TILDE] = ACTIONS(3865),
    [anon_sym_EQ_EQ] = ACTIONS(3867),
    [anon_sym_BANG_EQ] = ACTIONS(3865),
    [anon_sym_eq] = ACTIONS(3867),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3865),
    [anon_sym_equ] = ACTIONS(3865),
    [anon_sym_eqr] = ACTIONS(3865),
    [anon_sym_ne] = ACTIONS(3865),
    [anon_sym_LT_EQ_GT] = ACTIONS(3865),
    [anon_sym_cmp] = ACTIONS(3865),
    [anon_sym_TILDE_TILDE] = ACTIONS(3865),
    [anon_sym_LT] = ACTIONS(3867),
    [anon_sym_LT_EQ] = ACTIONS(3867),
    [anon_sym_GT_EQ] = ACTIONS(3865),
    [anon_sym_lt] = ACTIONS(3865),
    [anon_sym_le] = ACTIONS(3865),
    [anon_sym_ge] = ACTIONS(3865),
    [anon_sym_gt] = ACTIONS(3865),
    [anon_sym_isa] = ACTIONS(3865),
    [anon_sym_PLUS_PLUS] = ACTIONS(3865),
    [anon_sym_DASH_DASH] = ACTIONS(3865),
    [anon_sym_QMARK] = ACTIONS(3865),
    [anon_sym_EQ_GT] = ACTIONS(3865),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3865),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3636)] = {
    [sym_heredoc_content] = STATE(3636),
    [anon_sym_SEMI] = ACTIONS(3829),
    [anon_sym_EQ] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_COMMA] = ACTIONS(3829),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3829),
    [anon_sym_DASH_GT] = ACTIONS(3829),
    [anon_sym_STAR] = ACTIONS(3831),
    [anon_sym_PERCENT] = ACTIONS(3831),
    [anon_sym_GT] = ACTIONS(3831),
    [anon_sym_LT_LT] = ACTIONS(3831),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3829),
    [anon_sym_PLUS_EQ] = ACTIONS(3829),
    [anon_sym_DASH_EQ] = ACTIONS(3829),
    [anon_sym_DOT_EQ] = ACTIONS(3829),
    [anon_sym_STAR_EQ] = ACTIONS(3829),
    [anon_sym_SLASH_EQ] = ACTIONS(3829),
    [anon_sym_PERCENT_EQ] = ACTIONS(3829),
    [anon_sym_x_EQ] = ACTIONS(3829),
    [anon_sym_AMP_EQ] = ACTIONS(3829),
    [anon_sym_PIPE_EQ] = ACTIONS(3829),
    [anon_sym_CARET_EQ] = ACTIONS(3829),
    [anon_sym_LT_LT_EQ] = ACTIONS(3829),
    [anon_sym_GT_GT_EQ] = ACTIONS(3829),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3829),
    [anon_sym_DOT_DOT] = ACTIONS(3831),
    [anon_sym_STAR_STAR] = ACTIONS(3831),
    [anon_sym_PIPE_PIPE] = ACTIONS(3831),
    [anon_sym_SLASH_SLASH] = ACTIONS(3831),
    [anon_sym_CARET_CARET] = ACTIONS(3829),
    [anon_sym_AMP_AMP] = ACTIONS(3831),
    [anon_sym_PIPE] = ACTIONS(3831),
    [anon_sym_CARET] = ACTIONS(3831),
    [anon_sym_AMP] = ACTIONS(3831),
    [anon_sym_GT_GT2] = ACTIONS(3831),
    [anon_sym_PLUS] = ACTIONS(3831),
    [anon_sym_DASH] = ACTIONS(3831),
    [anon_sym_DOT] = ACTIONS(3831),
    [anon_sym_SLASH] = ACTIONS(3831),
    [anon_sym_x] = ACTIONS(3831),
    [anon_sym_EQ_TILDE] = ACTIONS(3829),
    [anon_sym_BANG_TILDE] = ACTIONS(3829),
    [anon_sym_EQ_EQ] = ACTIONS(3831),
    [anon_sym_BANG_EQ] = ACTIONS(3829),
    [anon_sym_eq] = ACTIONS(3831),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3829),
    [anon_sym_equ] = ACTIONS(3829),
    [anon_sym_eqr] = ACTIONS(3829),
    [anon_sym_ne] = ACTIONS(3829),
    [anon_sym_LT_EQ_GT] = ACTIONS(3829),
    [anon_sym_cmp] = ACTIONS(3829),
    [anon_sym_TILDE_TILDE] = ACTIONS(3829),
    [anon_sym_LT] = ACTIONS(3831),
    [anon_sym_LT_EQ] = ACTIONS(3831),
    [anon_sym_GT_EQ] = ACTIONS(3829),
    [anon_sym_lt] = ACTIONS(3829),
    [anon_sym_le] = ACTIONS(3829),
    [anon_sym_ge] = ACTIONS(3829),
    [anon_sym_gt] = ACTIONS(3829),
    [anon_sym_isa] = ACTIONS(3829),
    [anon_sym_PLUS_PLUS] = ACTIONS(3829),
    [anon_sym_DASH_DASH] = ACTIONS(3829),
    [anon_sym_QMARK] = ACTIONS(3829),
    [anon_sym_EQ_GT] = ACTIONS(3829),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3829),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3637)] = {
    [sym_heredoc_content] = STATE(3637),
    [anon_sym_SEMI] = ACTIONS(4095),
    [anon_sym_EQ] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_COMMA] = ACTIONS(4095),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4095),
    [anon_sym_DASH_GT] = ACTIONS(4095),
    [anon_sym_STAR] = ACTIONS(4097),
    [anon_sym_PERCENT] = ACTIONS(4097),
    [anon_sym_GT] = ACTIONS(4097),
    [anon_sym_LT_LT] = ACTIONS(4097),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4095),
    [anon_sym_PLUS_EQ] = ACTIONS(4095),
    [anon_sym_DASH_EQ] = ACTIONS(4095),
    [anon_sym_DOT_EQ] = ACTIONS(4095),
    [anon_sym_STAR_EQ] = ACTIONS(4095),
    [anon_sym_SLASH_EQ] = ACTIONS(4095),
    [anon_sym_PERCENT_EQ] = ACTIONS(4095),
    [anon_sym_x_EQ] = ACTIONS(4095),
    [anon_sym_AMP_EQ] = ACTIONS(4095),
    [anon_sym_PIPE_EQ] = ACTIONS(4095),
    [anon_sym_CARET_EQ] = ACTIONS(4095),
    [anon_sym_LT_LT_EQ] = ACTIONS(4095),
    [anon_sym_GT_GT_EQ] = ACTIONS(4095),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4095),
    [anon_sym_DOT_DOT] = ACTIONS(4097),
    [anon_sym_STAR_STAR] = ACTIONS(4097),
    [anon_sym_PIPE_PIPE] = ACTIONS(4097),
    [anon_sym_SLASH_SLASH] = ACTIONS(4097),
    [anon_sym_CARET_CARET] = ACTIONS(4095),
    [anon_sym_AMP_AMP] = ACTIONS(4097),
    [anon_sym_PIPE] = ACTIONS(4097),
    [anon_sym_CARET] = ACTIONS(4097),
    [anon_sym_AMP] = ACTIONS(4097),
    [anon_sym_GT_GT2] = ACTIONS(4097),
    [anon_sym_PLUS] = ACTIONS(4097),
    [anon_sym_DASH] = ACTIONS(4097),
    [anon_sym_DOT] = ACTIONS(4097),
    [anon_sym_SLASH] = ACTIONS(4097),
    [anon_sym_x] = ACTIONS(4097),
    [anon_sym_EQ_TILDE] = ACTIONS(4095),
    [anon_sym_BANG_TILDE] = ACTIONS(4095),
    [anon_sym_EQ_EQ] = ACTIONS(4097),
    [anon_sym_BANG_EQ] = ACTIONS(4095),
    [anon_sym_eq] = ACTIONS(4097),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4095),
    [anon_sym_equ] = ACTIONS(4095),
    [anon_sym_eqr] = ACTIONS(4095),
    [anon_sym_ne] = ACTIONS(4095),
    [anon_sym_LT_EQ_GT] = ACTIONS(4095),
    [anon_sym_cmp] = ACTIONS(4095),
    [anon_sym_TILDE_TILDE] = ACTIONS(4095),
    [anon_sym_LT] = ACTIONS(4097),
    [anon_sym_LT_EQ] = ACTIONS(4097),
    [anon_sym_GT_EQ] = ACTIONS(4095),
    [anon_sym_lt] = ACTIONS(4095),
    [anon_sym_le] = ACTIONS(4095),
    [anon_sym_ge] = ACTIONS(4095),
    [anon_sym_gt] = ACTIONS(4095),
    [anon_sym_isa] = ACTIONS(4095),
    [anon_sym_PLUS_PLUS] = ACTIONS(4095),
    [anon_sym_DASH_DASH] = ACTIONS(4095),
    [anon_sym_QMARK] = ACTIONS(4095),
    [anon_sym_EQ_GT] = ACTIONS(4095),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4095),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3638)] = {
    [sym_heredoc_content] = STATE(3638),
    [anon_sym_SEMI] = ACTIONS(4184),
    [anon_sym_EQ] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_COMMA] = ACTIONS(4184),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4184),
    [anon_sym_DASH_GT] = ACTIONS(4184),
    [anon_sym_STAR] = ACTIONS(4186),
    [anon_sym_PERCENT] = ACTIONS(4186),
    [anon_sym_GT] = ACTIONS(4186),
    [anon_sym_LT_LT] = ACTIONS(4186),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4184),
    [anon_sym_PLUS_EQ] = ACTIONS(4184),
    [anon_sym_DASH_EQ] = ACTIONS(4184),
    [anon_sym_DOT_EQ] = ACTIONS(4184),
    [anon_sym_STAR_EQ] = ACTIONS(4184),
    [anon_sym_SLASH_EQ] = ACTIONS(4184),
    [anon_sym_PERCENT_EQ] = ACTIONS(4184),
    [anon_sym_x_EQ] = ACTIONS(4184),
    [anon_sym_AMP_EQ] = ACTIONS(4184),
    [anon_sym_PIPE_EQ] = ACTIONS(4184),
    [anon_sym_CARET_EQ] = ACTIONS(4184),
    [anon_sym_LT_LT_EQ] = ACTIONS(4184),
    [anon_sym_GT_GT_EQ] = ACTIONS(4184),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4184),
    [anon_sym_DOT_DOT] = ACTIONS(4186),
    [anon_sym_STAR_STAR] = ACTIONS(4186),
    [anon_sym_PIPE_PIPE] = ACTIONS(4186),
    [anon_sym_SLASH_SLASH] = ACTIONS(4186),
    [anon_sym_CARET_CARET] = ACTIONS(4184),
    [anon_sym_AMP_AMP] = ACTIONS(4186),
    [anon_sym_PIPE] = ACTIONS(4186),
    [anon_sym_CARET] = ACTIONS(4186),
    [anon_sym_AMP] = ACTIONS(4186),
    [anon_sym_GT_GT2] = ACTIONS(4186),
    [anon_sym_PLUS] = ACTIONS(4186),
    [anon_sym_DASH] = ACTIONS(4186),
    [anon_sym_DOT] = ACTIONS(4186),
    [anon_sym_SLASH] = ACTIONS(4186),
    [anon_sym_x] = ACTIONS(4186),
    [anon_sym_EQ_TILDE] = ACTIONS(4184),
    [anon_sym_BANG_TILDE] = ACTIONS(4184),
    [anon_sym_EQ_EQ] = ACTIONS(4186),
    [anon_sym_BANG_EQ] = ACTIONS(4184),
    [anon_sym_eq] = ACTIONS(4186),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4184),
    [anon_sym_equ] = ACTIONS(4184),
    [anon_sym_eqr] = ACTIONS(4184),
    [anon_sym_ne] = ACTIONS(4184),
    [anon_sym_LT_EQ_GT] = ACTIONS(4184),
    [anon_sym_cmp] = ACTIONS(4184),
    [anon_sym_TILDE_TILDE] = ACTIONS(4184),
    [anon_sym_LT] = ACTIONS(4186),
    [anon_sym_LT_EQ] = ACTIONS(4186),
    [anon_sym_GT_EQ] = ACTIONS(4184),
    [anon_sym_lt] = ACTIONS(4184),
    [anon_sym_le] = ACTIONS(4184),
    [anon_sym_ge] = ACTIONS(4184),
    [anon_sym_gt] = ACTIONS(4184),
    [anon_sym_isa] = ACTIONS(4184),
    [anon_sym_PLUS_PLUS] = ACTIONS(4184),
    [anon_sym_DASH_DASH] = ACTIONS(4184),
    [anon_sym_QMARK] = ACTIONS(4184),
    [anon_sym_EQ_GT] = ACTIONS(4184),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4184),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3639)] = {
    [sym_heredoc_content] = STATE(3639),
    [anon_sym_SEMI] = ACTIONS(3727),
    [anon_sym_EQ] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_COMMA] = ACTIONS(3727),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3727),
    [anon_sym_DASH_GT] = ACTIONS(3727),
    [anon_sym_STAR] = ACTIONS(3729),
    [anon_sym_PERCENT] = ACTIONS(3729),
    [anon_sym_GT] = ACTIONS(3729),
    [anon_sym_LT_LT] = ACTIONS(3729),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3727),
    [anon_sym_PLUS_EQ] = ACTIONS(3727),
    [anon_sym_DASH_EQ] = ACTIONS(3727),
    [anon_sym_DOT_EQ] = ACTIONS(3727),
    [anon_sym_STAR_EQ] = ACTIONS(3727),
    [anon_sym_SLASH_EQ] = ACTIONS(3727),
    [anon_sym_PERCENT_EQ] = ACTIONS(3727),
    [anon_sym_x_EQ] = ACTIONS(3727),
    [anon_sym_AMP_EQ] = ACTIONS(3727),
    [anon_sym_PIPE_EQ] = ACTIONS(3727),
    [anon_sym_CARET_EQ] = ACTIONS(3727),
    [anon_sym_LT_LT_EQ] = ACTIONS(3727),
    [anon_sym_GT_GT_EQ] = ACTIONS(3727),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3727),
    [anon_sym_DOT_DOT] = ACTIONS(3729),
    [anon_sym_STAR_STAR] = ACTIONS(3729),
    [anon_sym_PIPE_PIPE] = ACTIONS(3729),
    [anon_sym_SLASH_SLASH] = ACTIONS(3729),
    [anon_sym_CARET_CARET] = ACTIONS(3727),
    [anon_sym_AMP_AMP] = ACTIONS(3729),
    [anon_sym_PIPE] = ACTIONS(3729),
    [anon_sym_CARET] = ACTIONS(3729),
    [anon_sym_AMP] = ACTIONS(3729),
    [anon_sym_GT_GT2] = ACTIONS(3729),
    [anon_sym_PLUS] = ACTIONS(3729),
    [anon_sym_DASH] = ACTIONS(3729),
    [anon_sym_DOT] = ACTIONS(3729),
    [anon_sym_SLASH] = ACTIONS(3729),
    [anon_sym_x] = ACTIONS(3729),
    [anon_sym_EQ_TILDE] = ACTIONS(3727),
    [anon_sym_BANG_TILDE] = ACTIONS(3727),
    [anon_sym_EQ_EQ] = ACTIONS(3729),
    [anon_sym_BANG_EQ] = ACTIONS(3727),
    [anon_sym_eq] = ACTIONS(3729),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3727),
    [anon_sym_equ] = ACTIONS(3727),
    [anon_sym_eqr] = ACTIONS(3727),
    [anon_sym_ne] = ACTIONS(3727),
    [anon_sym_LT_EQ_GT] = ACTIONS(3727),
    [anon_sym_cmp] = ACTIONS(3727),
    [anon_sym_TILDE_TILDE] = ACTIONS(3727),
    [anon_sym_LT] = ACTIONS(3729),
    [anon_sym_LT_EQ] = ACTIONS(3729),
    [anon_sym_GT_EQ] = ACTIONS(3727),
    [anon_sym_lt] = ACTIONS(3727),
    [anon_sym_le] = ACTIONS(3727),
    [anon_sym_ge] = ACTIONS(3727),
    [anon_sym_gt] = ACTIONS(3727),
    [anon_sym_isa] = ACTIONS(3727),
    [anon_sym_PLUS_PLUS] = ACTIONS(3727),
    [anon_sym_DASH_DASH] = ACTIONS(3727),
    [anon_sym_QMARK] = ACTIONS(3727),
    [anon_sym_EQ_GT] = ACTIONS(3727),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3727),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3640)] = {
    [sym_heredoc_content] = STATE(3640),
    [anon_sym_SEMI] = ACTIONS(3743),
    [anon_sym_EQ] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_COMMA] = ACTIONS(3743),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3743),
    [anon_sym_DASH_GT] = ACTIONS(3743),
    [anon_sym_STAR] = ACTIONS(3745),
    [anon_sym_PERCENT] = ACTIONS(3745),
    [anon_sym_GT] = ACTIONS(3745),
    [anon_sym_LT_LT] = ACTIONS(3745),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3743),
    [anon_sym_PLUS_EQ] = ACTIONS(3743),
    [anon_sym_DASH_EQ] = ACTIONS(3743),
    [anon_sym_DOT_EQ] = ACTIONS(3743),
    [anon_sym_STAR_EQ] = ACTIONS(3743),
    [anon_sym_SLASH_EQ] = ACTIONS(3743),
    [anon_sym_PERCENT_EQ] = ACTIONS(3743),
    [anon_sym_x_EQ] = ACTIONS(3743),
    [anon_sym_AMP_EQ] = ACTIONS(3743),
    [anon_sym_PIPE_EQ] = ACTIONS(3743),
    [anon_sym_CARET_EQ] = ACTIONS(3743),
    [anon_sym_LT_LT_EQ] = ACTIONS(3743),
    [anon_sym_GT_GT_EQ] = ACTIONS(3743),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3743),
    [anon_sym_DOT_DOT] = ACTIONS(3745),
    [anon_sym_STAR_STAR] = ACTIONS(3745),
    [anon_sym_PIPE_PIPE] = ACTIONS(3745),
    [anon_sym_SLASH_SLASH] = ACTIONS(3745),
    [anon_sym_CARET_CARET] = ACTIONS(3743),
    [anon_sym_AMP_AMP] = ACTIONS(3745),
    [anon_sym_PIPE] = ACTIONS(3745),
    [anon_sym_CARET] = ACTIONS(3745),
    [anon_sym_AMP] = ACTIONS(3745),
    [anon_sym_GT_GT2] = ACTIONS(3745),
    [anon_sym_PLUS] = ACTIONS(3745),
    [anon_sym_DASH] = ACTIONS(3745),
    [anon_sym_DOT] = ACTIONS(3745),
    [anon_sym_SLASH] = ACTIONS(3745),
    [anon_sym_x] = ACTIONS(3745),
    [anon_sym_EQ_TILDE] = ACTIONS(3743),
    [anon_sym_BANG_TILDE] = ACTIONS(3743),
    [anon_sym_EQ_EQ] = ACTIONS(3745),
    [anon_sym_BANG_EQ] = ACTIONS(3743),
    [anon_sym_eq] = ACTIONS(3745),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3743),
    [anon_sym_equ] = ACTIONS(3743),
    [anon_sym_eqr] = ACTIONS(3743),
    [anon_sym_ne] = ACTIONS(3743),
    [anon_sym_LT_EQ_GT] = ACTIONS(3743),
    [anon_sym_cmp] = ACTIONS(3743),
    [anon_sym_TILDE_TILDE] = ACTIONS(3743),
    [anon_sym_LT] = ACTIONS(3745),
    [anon_sym_LT_EQ] = ACTIONS(3745),
    [anon_sym_GT_EQ] = ACTIONS(3743),
    [anon_sym_lt] = ACTIONS(3743),
    [anon_sym_le] = ACTIONS(3743),
    [anon_sym_ge] = ACTIONS(3743),
    [anon_sym_gt] = ACTIONS(3743),
    [anon_sym_isa] = ACTIONS(3743),
    [anon_sym_PLUS_PLUS] = ACTIONS(3743),
    [anon_sym_DASH_DASH] = ACTIONS(3743),
    [anon_sym_QMARK] = ACTIONS(3743),
    [anon_sym_EQ_GT] = ACTIONS(3743),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3743),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3641)] = {
    [sym_heredoc_content] = STATE(3641),
    [anon_sym_SEMI] = ACTIONS(3811),
    [anon_sym_EQ] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_COMMA] = ACTIONS(3811),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3811),
    [anon_sym_DASH_GT] = ACTIONS(3811),
    [anon_sym_STAR] = ACTIONS(3813),
    [anon_sym_PERCENT] = ACTIONS(3813),
    [anon_sym_GT] = ACTIONS(3813),
    [anon_sym_LT_LT] = ACTIONS(3813),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3811),
    [anon_sym_PLUS_EQ] = ACTIONS(3811),
    [anon_sym_DASH_EQ] = ACTIONS(3811),
    [anon_sym_DOT_EQ] = ACTIONS(3811),
    [anon_sym_STAR_EQ] = ACTIONS(3811),
    [anon_sym_SLASH_EQ] = ACTIONS(3811),
    [anon_sym_PERCENT_EQ] = ACTIONS(3811),
    [anon_sym_x_EQ] = ACTIONS(3811),
    [anon_sym_AMP_EQ] = ACTIONS(3811),
    [anon_sym_PIPE_EQ] = ACTIONS(3811),
    [anon_sym_CARET_EQ] = ACTIONS(3811),
    [anon_sym_LT_LT_EQ] = ACTIONS(3811),
    [anon_sym_GT_GT_EQ] = ACTIONS(3811),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3811),
    [anon_sym_DOT_DOT] = ACTIONS(3813),
    [anon_sym_STAR_STAR] = ACTIONS(3813),
    [anon_sym_PIPE_PIPE] = ACTIONS(3813),
    [anon_sym_SLASH_SLASH] = ACTIONS(3813),
    [anon_sym_CARET_CARET] = ACTIONS(3811),
    [anon_sym_AMP_AMP] = ACTIONS(3813),
    [anon_sym_PIPE] = ACTIONS(3813),
    [anon_sym_CARET] = ACTIONS(3813),
    [anon_sym_AMP] = ACTIONS(3813),
    [anon_sym_GT_GT2] = ACTIONS(3813),
    [anon_sym_PLUS] = ACTIONS(3813),
    [anon_sym_DASH] = ACTIONS(3813),
    [anon_sym_DOT] = ACTIONS(3813),
    [anon_sym_SLASH] = ACTIONS(3813),
    [anon_sym_x] = ACTIONS(3813),
    [anon_sym_EQ_TILDE] = ACTIONS(3811),
    [anon_sym_BANG_TILDE] = ACTIONS(3811),
    [anon_sym_EQ_EQ] = ACTIONS(3813),
    [anon_sym_BANG_EQ] = ACTIONS(3811),
    [anon_sym_eq] = ACTIONS(3813),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3811),
    [anon_sym_equ] = ACTIONS(3811),
    [anon_sym_eqr] = ACTIONS(3811),
    [anon_sym_ne] = ACTIONS(3811),
    [anon_sym_LT_EQ_GT] = ACTIONS(3811),
    [anon_sym_cmp] = ACTIONS(3811),
    [anon_sym_TILDE_TILDE] = ACTIONS(3811),
    [anon_sym_LT] = ACTIONS(3813),
    [anon_sym_LT_EQ] = ACTIONS(3813),
    [anon_sym_GT_EQ] = ACTIONS(3811),
    [anon_sym_lt] = ACTIONS(3811),
    [anon_sym_le] = ACTIONS(3811),
    [anon_sym_ge] = ACTIONS(3811),
    [anon_sym_gt] = ACTIONS(3811),
    [anon_sym_isa] = ACTIONS(3811),
    [anon_sym_PLUS_PLUS] = ACTIONS(3811),
    [anon_sym_DASH_DASH] = ACTIONS(3811),
    [anon_sym_QMARK] = ACTIONS(3811),
    [anon_sym_EQ_GT] = ACTIONS(3811),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3811),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3642)] = {
    [sym_heredoc_content] = STATE(3642),
    [anon_sym_SEMI] = ACTIONS(3873),
    [anon_sym_EQ] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_COMMA] = ACTIONS(3873),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3873),
    [anon_sym_DASH_GT] = ACTIONS(3873),
    [anon_sym_STAR] = ACTIONS(3875),
    [anon_sym_PERCENT] = ACTIONS(3875),
    [anon_sym_GT] = ACTIONS(3875),
    [anon_sym_LT_LT] = ACTIONS(3875),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3873),
    [anon_sym_PLUS_EQ] = ACTIONS(3873),
    [anon_sym_DASH_EQ] = ACTIONS(3873),
    [anon_sym_DOT_EQ] = ACTIONS(3873),
    [anon_sym_STAR_EQ] = ACTIONS(3873),
    [anon_sym_SLASH_EQ] = ACTIONS(3873),
    [anon_sym_PERCENT_EQ] = ACTIONS(3873),
    [anon_sym_x_EQ] = ACTIONS(3873),
    [anon_sym_AMP_EQ] = ACTIONS(3873),
    [anon_sym_PIPE_EQ] = ACTIONS(3873),
    [anon_sym_CARET_EQ] = ACTIONS(3873),
    [anon_sym_LT_LT_EQ] = ACTIONS(3873),
    [anon_sym_GT_GT_EQ] = ACTIONS(3873),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3873),
    [anon_sym_DOT_DOT] = ACTIONS(3875),
    [anon_sym_STAR_STAR] = ACTIONS(3875),
    [anon_sym_PIPE_PIPE] = ACTIONS(3875),
    [anon_sym_SLASH_SLASH] = ACTIONS(3875),
    [anon_sym_CARET_CARET] = ACTIONS(3873),
    [anon_sym_AMP_AMP] = ACTIONS(3875),
    [anon_sym_PIPE] = ACTIONS(3875),
    [anon_sym_CARET] = ACTIONS(3875),
    [anon_sym_AMP] = ACTIONS(3875),
    [anon_sym_GT_GT2] = ACTIONS(3875),
    [anon_sym_PLUS] = ACTIONS(3875),
    [anon_sym_DASH] = ACTIONS(3875),
    [anon_sym_DOT] = ACTIONS(3875),
    [anon_sym_SLASH] = ACTIONS(3875),
    [anon_sym_x] = ACTIONS(3875),
    [anon_sym_EQ_TILDE] = ACTIONS(3873),
    [anon_sym_BANG_TILDE] = ACTIONS(3873),
    [anon_sym_EQ_EQ] = ACTIONS(3875),
    [anon_sym_BANG_EQ] = ACTIONS(3873),
    [anon_sym_eq] = ACTIONS(3875),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3873),
    [anon_sym_equ] = ACTIONS(3873),
    [anon_sym_eqr] = ACTIONS(3873),
    [anon_sym_ne] = ACTIONS(3873),
    [anon_sym_LT_EQ_GT] = ACTIONS(3873),
    [anon_sym_cmp] = ACTIONS(3873),
    [anon_sym_TILDE_TILDE] = ACTIONS(3873),
    [anon_sym_LT] = ACTIONS(3875),
    [anon_sym_LT_EQ] = ACTIONS(3875),
    [anon_sym_GT_EQ] = ACTIONS(3873),
    [anon_sym_lt] = ACTIONS(3873),
    [anon_sym_le] = ACTIONS(3873),
    [anon_sym_ge] = ACTIONS(3873),
    [anon_sym_gt] = ACTIONS(3873),
    [anon_sym_isa] = ACTIONS(3873),
    [anon_sym_PLUS_PLUS] = ACTIONS(3873),
    [anon_sym_DASH_DASH] = ACTIONS(3873),
    [anon_sym_QMARK] = ACTIONS(3873),
    [anon_sym_EQ_GT] = ACTIONS(3873),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3873),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3643)] = {
    [sym_heredoc_content] = STATE(3643),
    [anon_sym_SEMI] = ACTIONS(3981),
    [anon_sym_EQ] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_COMMA] = ACTIONS(3981),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3981),
    [anon_sym_DASH_GT] = ACTIONS(3981),
    [anon_sym_STAR] = ACTIONS(3983),
    [anon_sym_PERCENT] = ACTIONS(3983),
    [anon_sym_GT] = ACTIONS(3983),
    [anon_sym_LT_LT] = ACTIONS(3983),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3981),
    [anon_sym_PLUS_EQ] = ACTIONS(3981),
    [anon_sym_DASH_EQ] = ACTIONS(3981),
    [anon_sym_DOT_EQ] = ACTIONS(3981),
    [anon_sym_STAR_EQ] = ACTIONS(3981),
    [anon_sym_SLASH_EQ] = ACTIONS(3981),
    [anon_sym_PERCENT_EQ] = ACTIONS(3981),
    [anon_sym_x_EQ] = ACTIONS(3981),
    [anon_sym_AMP_EQ] = ACTIONS(3981),
    [anon_sym_PIPE_EQ] = ACTIONS(3981),
    [anon_sym_CARET_EQ] = ACTIONS(3981),
    [anon_sym_LT_LT_EQ] = ACTIONS(3981),
    [anon_sym_GT_GT_EQ] = ACTIONS(3981),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3981),
    [anon_sym_DOT_DOT] = ACTIONS(3983),
    [anon_sym_STAR_STAR] = ACTIONS(3983),
    [anon_sym_PIPE_PIPE] = ACTIONS(3983),
    [anon_sym_SLASH_SLASH] = ACTIONS(3983),
    [anon_sym_CARET_CARET] = ACTIONS(3981),
    [anon_sym_AMP_AMP] = ACTIONS(3983),
    [anon_sym_PIPE] = ACTIONS(3983),
    [anon_sym_CARET] = ACTIONS(3983),
    [anon_sym_AMP] = ACTIONS(3983),
    [anon_sym_GT_GT2] = ACTIONS(3983),
    [anon_sym_PLUS] = ACTIONS(3983),
    [anon_sym_DASH] = ACTIONS(3983),
    [anon_sym_DOT] = ACTIONS(3983),
    [anon_sym_SLASH] = ACTIONS(3983),
    [anon_sym_x] = ACTIONS(3983),
    [anon_sym_EQ_TILDE] = ACTIONS(3981),
    [anon_sym_BANG_TILDE] = ACTIONS(3981),
    [anon_sym_EQ_EQ] = ACTIONS(3983),
    [anon_sym_BANG_EQ] = ACTIONS(3981),
    [anon_sym_eq] = ACTIONS(3983),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3981),
    [anon_sym_equ] = ACTIONS(3981),
    [anon_sym_eqr] = ACTIONS(3981),
    [anon_sym_ne] = ACTIONS(3981),
    [anon_sym_LT_EQ_GT] = ACTIONS(3981),
    [anon_sym_cmp] = ACTIONS(3981),
    [anon_sym_TILDE_TILDE] = ACTIONS(3981),
    [anon_sym_LT] = ACTIONS(3983),
    [anon_sym_LT_EQ] = ACTIONS(3983),
    [anon_sym_GT_EQ] = ACTIONS(3981),
    [anon_sym_lt] = ACTIONS(3981),
    [anon_sym_le] = ACTIONS(3981),
    [anon_sym_ge] = ACTIONS(3981),
    [anon_sym_gt] = ACTIONS(3981),
    [anon_sym_isa] = ACTIONS(3981),
    [anon_sym_PLUS_PLUS] = ACTIONS(3981),
    [anon_sym_DASH_DASH] = ACTIONS(3981),
    [anon_sym_QMARK] = ACTIONS(3981),
    [anon_sym_EQ_GT] = ACTIONS(3981),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3981),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3644)] = {
    [sym_heredoc_content] = STATE(3644),
    [anon_sym_SEMI] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5259),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(5265),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5269),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(5273),
    [anon_sym_SLASH_SLASH] = ACTIONS(5273),
    [anon_sym_CARET_CARET] = ACTIONS(5275),
    [anon_sym_AMP_AMP] = ACTIONS(5277),
    [anon_sym_PIPE] = ACTIONS(5279),
    [anon_sym_CARET] = ACTIONS(5279),
    [anon_sym_AMP] = ACTIONS(5281),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(5287),
    [anon_sym_BANG_EQ] = ACTIONS(5289),
    [anon_sym_eq] = ACTIONS(5287),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5289),
    [anon_sym_equ] = ACTIONS(5289),
    [anon_sym_eqr] = ACTIONS(5289),
    [anon_sym_ne] = ACTIONS(5289),
    [anon_sym_LT_EQ_GT] = ACTIONS(5291),
    [anon_sym_cmp] = ACTIONS(5291),
    [anon_sym_TILDE_TILDE] = ACTIONS(5291),
    [anon_sym_LT] = ACTIONS(5265),
    [anon_sym_LT_EQ] = ACTIONS(5265),
    [anon_sym_GT_EQ] = ACTIONS(5293),
    [anon_sym_lt] = ACTIONS(5293),
    [anon_sym_le] = ACTIONS(5293),
    [anon_sym_ge] = ACTIONS(5293),
    [anon_sym_gt] = ACTIONS(5293),
    [anon_sym_isa] = ACTIONS(5295),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(5299),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3985),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3645)] = {
    [sym_heredoc_content] = STATE(3645),
    [anon_sym_SEMI] = ACTIONS(4001),
    [anon_sym_EQ] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_COMMA] = ACTIONS(4001),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4001),
    [anon_sym_DASH_GT] = ACTIONS(4001),
    [anon_sym_STAR] = ACTIONS(4003),
    [anon_sym_PERCENT] = ACTIONS(4003),
    [anon_sym_GT] = ACTIONS(4003),
    [anon_sym_LT_LT] = ACTIONS(4003),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4001),
    [anon_sym_PLUS_EQ] = ACTIONS(4001),
    [anon_sym_DASH_EQ] = ACTIONS(4001),
    [anon_sym_DOT_EQ] = ACTIONS(4001),
    [anon_sym_STAR_EQ] = ACTIONS(4001),
    [anon_sym_SLASH_EQ] = ACTIONS(4001),
    [anon_sym_PERCENT_EQ] = ACTIONS(4001),
    [anon_sym_x_EQ] = ACTIONS(4001),
    [anon_sym_AMP_EQ] = ACTIONS(4001),
    [anon_sym_PIPE_EQ] = ACTIONS(4001),
    [anon_sym_CARET_EQ] = ACTIONS(4001),
    [anon_sym_LT_LT_EQ] = ACTIONS(4001),
    [anon_sym_GT_GT_EQ] = ACTIONS(4001),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4001),
    [anon_sym_DOT_DOT] = ACTIONS(4003),
    [anon_sym_STAR_STAR] = ACTIONS(4003),
    [anon_sym_PIPE_PIPE] = ACTIONS(4003),
    [anon_sym_SLASH_SLASH] = ACTIONS(4003),
    [anon_sym_CARET_CARET] = ACTIONS(4001),
    [anon_sym_AMP_AMP] = ACTIONS(4003),
    [anon_sym_PIPE] = ACTIONS(4003),
    [anon_sym_CARET] = ACTIONS(4003),
    [anon_sym_AMP] = ACTIONS(4003),
    [anon_sym_GT_GT2] = ACTIONS(4003),
    [anon_sym_PLUS] = ACTIONS(4003),
    [anon_sym_DASH] = ACTIONS(4003),
    [anon_sym_DOT] = ACTIONS(4003),
    [anon_sym_SLASH] = ACTIONS(4003),
    [anon_sym_x] = ACTIONS(4003),
    [anon_sym_EQ_TILDE] = ACTIONS(4001),
    [anon_sym_BANG_TILDE] = ACTIONS(4001),
    [anon_sym_EQ_EQ] = ACTIONS(4003),
    [anon_sym_BANG_EQ] = ACTIONS(4001),
    [anon_sym_eq] = ACTIONS(4003),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4001),
    [anon_sym_equ] = ACTIONS(4001),
    [anon_sym_eqr] = ACTIONS(4001),
    [anon_sym_ne] = ACTIONS(4001),
    [anon_sym_LT_EQ_GT] = ACTIONS(4001),
    [anon_sym_cmp] = ACTIONS(4001),
    [anon_sym_TILDE_TILDE] = ACTIONS(4001),
    [anon_sym_LT] = ACTIONS(4003),
    [anon_sym_LT_EQ] = ACTIONS(4003),
    [anon_sym_GT_EQ] = ACTIONS(4001),
    [anon_sym_lt] = ACTIONS(4001),
    [anon_sym_le] = ACTIONS(4001),
    [anon_sym_ge] = ACTIONS(4001),
    [anon_sym_gt] = ACTIONS(4001),
    [anon_sym_isa] = ACTIONS(4001),
    [anon_sym_PLUS_PLUS] = ACTIONS(4001),
    [anon_sym_DASH_DASH] = ACTIONS(4001),
    [anon_sym_QMARK] = ACTIONS(4001),
    [anon_sym_EQ_GT] = ACTIONS(4001),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4001),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3646)] = {
    [sym_heredoc_content] = STATE(3646),
    [anon_sym_SEMI] = ACTIONS(4009),
    [anon_sym_EQ] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_COMMA] = ACTIONS(4009),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4009),
    [anon_sym_DASH_GT] = ACTIONS(4009),
    [anon_sym_STAR] = ACTIONS(4011),
    [anon_sym_PERCENT] = ACTIONS(4011),
    [anon_sym_GT] = ACTIONS(4011),
    [anon_sym_LT_LT] = ACTIONS(4011),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4009),
    [anon_sym_PLUS_EQ] = ACTIONS(4009),
    [anon_sym_DASH_EQ] = ACTIONS(4009),
    [anon_sym_DOT_EQ] = ACTIONS(4009),
    [anon_sym_STAR_EQ] = ACTIONS(4009),
    [anon_sym_SLASH_EQ] = ACTIONS(4009),
    [anon_sym_PERCENT_EQ] = ACTIONS(4009),
    [anon_sym_x_EQ] = ACTIONS(4009),
    [anon_sym_AMP_EQ] = ACTIONS(4009),
    [anon_sym_PIPE_EQ] = ACTIONS(4009),
    [anon_sym_CARET_EQ] = ACTIONS(4009),
    [anon_sym_LT_LT_EQ] = ACTIONS(4009),
    [anon_sym_GT_GT_EQ] = ACTIONS(4009),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4009),
    [anon_sym_DOT_DOT] = ACTIONS(4011),
    [anon_sym_STAR_STAR] = ACTIONS(4011),
    [anon_sym_PIPE_PIPE] = ACTIONS(4011),
    [anon_sym_SLASH_SLASH] = ACTIONS(4011),
    [anon_sym_CARET_CARET] = ACTIONS(4009),
    [anon_sym_AMP_AMP] = ACTIONS(4011),
    [anon_sym_PIPE] = ACTIONS(4011),
    [anon_sym_CARET] = ACTIONS(4011),
    [anon_sym_AMP] = ACTIONS(4011),
    [anon_sym_GT_GT2] = ACTIONS(4011),
    [anon_sym_PLUS] = ACTIONS(4011),
    [anon_sym_DASH] = ACTIONS(4011),
    [anon_sym_DOT] = ACTIONS(4011),
    [anon_sym_SLASH] = ACTIONS(4011),
    [anon_sym_x] = ACTIONS(4011),
    [anon_sym_EQ_TILDE] = ACTIONS(4009),
    [anon_sym_BANG_TILDE] = ACTIONS(4009),
    [anon_sym_EQ_EQ] = ACTIONS(4011),
    [anon_sym_BANG_EQ] = ACTIONS(4009),
    [anon_sym_eq] = ACTIONS(4011),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4009),
    [anon_sym_equ] = ACTIONS(4009),
    [anon_sym_eqr] = ACTIONS(4009),
    [anon_sym_ne] = ACTIONS(4009),
    [anon_sym_LT_EQ_GT] = ACTIONS(4009),
    [anon_sym_cmp] = ACTIONS(4009),
    [anon_sym_TILDE_TILDE] = ACTIONS(4009),
    [anon_sym_LT] = ACTIONS(4011),
    [anon_sym_LT_EQ] = ACTIONS(4011),
    [anon_sym_GT_EQ] = ACTIONS(4009),
    [anon_sym_lt] = ACTIONS(4009),
    [anon_sym_le] = ACTIONS(4009),
    [anon_sym_ge] = ACTIONS(4009),
    [anon_sym_gt] = ACTIONS(4009),
    [anon_sym_isa] = ACTIONS(4009),
    [anon_sym_PLUS_PLUS] = ACTIONS(4009),
    [anon_sym_DASH_DASH] = ACTIONS(4009),
    [anon_sym_QMARK] = ACTIONS(4009),
    [anon_sym_EQ_GT] = ACTIONS(4009),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4009),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3647)] = {
    [sym_heredoc_content] = STATE(3647),
    [anon_sym_SEMI] = ACTIONS(4013),
    [anon_sym_EQ] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_COMMA] = ACTIONS(4013),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4013),
    [anon_sym_DASH_GT] = ACTIONS(4013),
    [anon_sym_STAR] = ACTIONS(4015),
    [anon_sym_PERCENT] = ACTIONS(4015),
    [anon_sym_GT] = ACTIONS(4015),
    [anon_sym_LT_LT] = ACTIONS(4015),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4013),
    [anon_sym_PLUS_EQ] = ACTIONS(4013),
    [anon_sym_DASH_EQ] = ACTIONS(4013),
    [anon_sym_DOT_EQ] = ACTIONS(4013),
    [anon_sym_STAR_EQ] = ACTIONS(4013),
    [anon_sym_SLASH_EQ] = ACTIONS(4013),
    [anon_sym_PERCENT_EQ] = ACTIONS(4013),
    [anon_sym_x_EQ] = ACTIONS(4013),
    [anon_sym_AMP_EQ] = ACTIONS(4013),
    [anon_sym_PIPE_EQ] = ACTIONS(4013),
    [anon_sym_CARET_EQ] = ACTIONS(4013),
    [anon_sym_LT_LT_EQ] = ACTIONS(4013),
    [anon_sym_GT_GT_EQ] = ACTIONS(4013),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4013),
    [anon_sym_DOT_DOT] = ACTIONS(4015),
    [anon_sym_STAR_STAR] = ACTIONS(4015),
    [anon_sym_PIPE_PIPE] = ACTIONS(4015),
    [anon_sym_SLASH_SLASH] = ACTIONS(4015),
    [anon_sym_CARET_CARET] = ACTIONS(4013),
    [anon_sym_AMP_AMP] = ACTIONS(4015),
    [anon_sym_PIPE] = ACTIONS(4015),
    [anon_sym_CARET] = ACTIONS(4015),
    [anon_sym_AMP] = ACTIONS(4015),
    [anon_sym_GT_GT2] = ACTIONS(4015),
    [anon_sym_PLUS] = ACTIONS(4015),
    [anon_sym_DASH] = ACTIONS(4015),
    [anon_sym_DOT] = ACTIONS(4015),
    [anon_sym_SLASH] = ACTIONS(4015),
    [anon_sym_x] = ACTIONS(4015),
    [anon_sym_EQ_TILDE] = ACTIONS(4013),
    [anon_sym_BANG_TILDE] = ACTIONS(4013),
    [anon_sym_EQ_EQ] = ACTIONS(4015),
    [anon_sym_BANG_EQ] = ACTIONS(4013),
    [anon_sym_eq] = ACTIONS(4015),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4013),
    [anon_sym_equ] = ACTIONS(4013),
    [anon_sym_eqr] = ACTIONS(4013),
    [anon_sym_ne] = ACTIONS(4013),
    [anon_sym_LT_EQ_GT] = ACTIONS(4013),
    [anon_sym_cmp] = ACTIONS(4013),
    [anon_sym_TILDE_TILDE] = ACTIONS(4013),
    [anon_sym_LT] = ACTIONS(4015),
    [anon_sym_LT_EQ] = ACTIONS(4015),
    [anon_sym_GT_EQ] = ACTIONS(4013),
    [anon_sym_lt] = ACTIONS(4013),
    [anon_sym_le] = ACTIONS(4013),
    [anon_sym_ge] = ACTIONS(4013),
    [anon_sym_gt] = ACTIONS(4013),
    [anon_sym_isa] = ACTIONS(4013),
    [anon_sym_PLUS_PLUS] = ACTIONS(4013),
    [anon_sym_DASH_DASH] = ACTIONS(4013),
    [anon_sym_QMARK] = ACTIONS(4013),
    [anon_sym_EQ_GT] = ACTIONS(4013),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4013),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3648)] = {
    [sym_heredoc_content] = STATE(3648),
    [anon_sym_SEMI] = ACTIONS(4025),
    [anon_sym_EQ] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_COMMA] = ACTIONS(4025),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4025),
    [anon_sym_DASH_GT] = ACTIONS(4025),
    [anon_sym_STAR] = ACTIONS(4027),
    [anon_sym_PERCENT] = ACTIONS(4027),
    [anon_sym_GT] = ACTIONS(4027),
    [anon_sym_LT_LT] = ACTIONS(4027),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4025),
    [anon_sym_PLUS_EQ] = ACTIONS(4025),
    [anon_sym_DASH_EQ] = ACTIONS(4025),
    [anon_sym_DOT_EQ] = ACTIONS(4025),
    [anon_sym_STAR_EQ] = ACTIONS(4025),
    [anon_sym_SLASH_EQ] = ACTIONS(4025),
    [anon_sym_PERCENT_EQ] = ACTIONS(4025),
    [anon_sym_x_EQ] = ACTIONS(4025),
    [anon_sym_AMP_EQ] = ACTIONS(4025),
    [anon_sym_PIPE_EQ] = ACTIONS(4025),
    [anon_sym_CARET_EQ] = ACTIONS(4025),
    [anon_sym_LT_LT_EQ] = ACTIONS(4025),
    [anon_sym_GT_GT_EQ] = ACTIONS(4025),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4025),
    [anon_sym_DOT_DOT] = ACTIONS(4027),
    [anon_sym_STAR_STAR] = ACTIONS(4027),
    [anon_sym_PIPE_PIPE] = ACTIONS(4027),
    [anon_sym_SLASH_SLASH] = ACTIONS(4027),
    [anon_sym_CARET_CARET] = ACTIONS(4025),
    [anon_sym_AMP_AMP] = ACTIONS(4027),
    [anon_sym_PIPE] = ACTIONS(4027),
    [anon_sym_CARET] = ACTIONS(4027),
    [anon_sym_AMP] = ACTIONS(4027),
    [anon_sym_GT_GT2] = ACTIONS(4027),
    [anon_sym_PLUS] = ACTIONS(4027),
    [anon_sym_DASH] = ACTIONS(4027),
    [anon_sym_DOT] = ACTIONS(4027),
    [anon_sym_SLASH] = ACTIONS(4027),
    [anon_sym_x] = ACTIONS(4027),
    [anon_sym_EQ_TILDE] = ACTIONS(4025),
    [anon_sym_BANG_TILDE] = ACTIONS(4025),
    [anon_sym_EQ_EQ] = ACTIONS(4027),
    [anon_sym_BANG_EQ] = ACTIONS(4025),
    [anon_sym_eq] = ACTIONS(4027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4025),
    [anon_sym_equ] = ACTIONS(4025),
    [anon_sym_eqr] = ACTIONS(4025),
    [anon_sym_ne] = ACTIONS(4025),
    [anon_sym_LT_EQ_GT] = ACTIONS(4025),
    [anon_sym_cmp] = ACTIONS(4025),
    [anon_sym_TILDE_TILDE] = ACTIONS(4025),
    [anon_sym_LT] = ACTIONS(4027),
    [anon_sym_LT_EQ] = ACTIONS(4027),
    [anon_sym_GT_EQ] = ACTIONS(4025),
    [anon_sym_lt] = ACTIONS(4025),
    [anon_sym_le] = ACTIONS(4025),
    [anon_sym_ge] = ACTIONS(4025),
    [anon_sym_gt] = ACTIONS(4025),
    [anon_sym_isa] = ACTIONS(4025),
    [anon_sym_PLUS_PLUS] = ACTIONS(4025),
    [anon_sym_DASH_DASH] = ACTIONS(4025),
    [anon_sym_QMARK] = ACTIONS(4025),
    [anon_sym_EQ_GT] = ACTIONS(4025),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4025),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3649)] = {
    [sym_heredoc_content] = STATE(3649),
    [anon_sym_SEMI] = ACTIONS(4029),
    [anon_sym_EQ] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_COMMA] = ACTIONS(4029),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4029),
    [anon_sym_DASH_GT] = ACTIONS(4029),
    [anon_sym_STAR] = ACTIONS(4031),
    [anon_sym_PERCENT] = ACTIONS(4031),
    [anon_sym_GT] = ACTIONS(4031),
    [anon_sym_LT_LT] = ACTIONS(4031),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4029),
    [anon_sym_PLUS_EQ] = ACTIONS(4029),
    [anon_sym_DASH_EQ] = ACTIONS(4029),
    [anon_sym_DOT_EQ] = ACTIONS(4029),
    [anon_sym_STAR_EQ] = ACTIONS(4029),
    [anon_sym_SLASH_EQ] = ACTIONS(4029),
    [anon_sym_PERCENT_EQ] = ACTIONS(4029),
    [anon_sym_x_EQ] = ACTIONS(4029),
    [anon_sym_AMP_EQ] = ACTIONS(4029),
    [anon_sym_PIPE_EQ] = ACTIONS(4029),
    [anon_sym_CARET_EQ] = ACTIONS(4029),
    [anon_sym_LT_LT_EQ] = ACTIONS(4029),
    [anon_sym_GT_GT_EQ] = ACTIONS(4029),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4029),
    [anon_sym_DOT_DOT] = ACTIONS(4031),
    [anon_sym_STAR_STAR] = ACTIONS(4031),
    [anon_sym_PIPE_PIPE] = ACTIONS(4031),
    [anon_sym_SLASH_SLASH] = ACTIONS(4031),
    [anon_sym_CARET_CARET] = ACTIONS(4029),
    [anon_sym_AMP_AMP] = ACTIONS(4031),
    [anon_sym_PIPE] = ACTIONS(4031),
    [anon_sym_CARET] = ACTIONS(4031),
    [anon_sym_AMP] = ACTIONS(4031),
    [anon_sym_GT_GT2] = ACTIONS(4031),
    [anon_sym_PLUS] = ACTIONS(4031),
    [anon_sym_DASH] = ACTIONS(4031),
    [anon_sym_DOT] = ACTIONS(4031),
    [anon_sym_SLASH] = ACTIONS(4031),
    [anon_sym_x] = ACTIONS(4031),
    [anon_sym_EQ_TILDE] = ACTIONS(4029),
    [anon_sym_BANG_TILDE] = ACTIONS(4029),
    [anon_sym_EQ_EQ] = ACTIONS(4031),
    [anon_sym_BANG_EQ] = ACTIONS(4029),
    [anon_sym_eq] = ACTIONS(4031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4029),
    [anon_sym_equ] = ACTIONS(4029),
    [anon_sym_eqr] = ACTIONS(4029),
    [anon_sym_ne] = ACTIONS(4029),
    [anon_sym_LT_EQ_GT] = ACTIONS(4029),
    [anon_sym_cmp] = ACTIONS(4029),
    [anon_sym_TILDE_TILDE] = ACTIONS(4029),
    [anon_sym_LT] = ACTIONS(4031),
    [anon_sym_LT_EQ] = ACTIONS(4031),
    [anon_sym_GT_EQ] = ACTIONS(4029),
    [anon_sym_lt] = ACTIONS(4029),
    [anon_sym_le] = ACTIONS(4029),
    [anon_sym_ge] = ACTIONS(4029),
    [anon_sym_gt] = ACTIONS(4029),
    [anon_sym_isa] = ACTIONS(4029),
    [anon_sym_PLUS_PLUS] = ACTIONS(4029),
    [anon_sym_DASH_DASH] = ACTIONS(4029),
    [anon_sym_QMARK] = ACTIONS(4029),
    [anon_sym_EQ_GT] = ACTIONS(4029),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4029),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3650)] = {
    [sym_heredoc_content] = STATE(3650),
    [anon_sym_SEMI] = ACTIONS(4123),
    [anon_sym_EQ] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_COMMA] = ACTIONS(4123),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4123),
    [anon_sym_DASH_GT] = ACTIONS(4123),
    [anon_sym_STAR] = ACTIONS(4125),
    [anon_sym_PERCENT] = ACTIONS(4125),
    [anon_sym_GT] = ACTIONS(4125),
    [anon_sym_LT_LT] = ACTIONS(4125),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4123),
    [anon_sym_PLUS_EQ] = ACTIONS(4123),
    [anon_sym_DASH_EQ] = ACTIONS(4123),
    [anon_sym_DOT_EQ] = ACTIONS(4123),
    [anon_sym_STAR_EQ] = ACTIONS(4123),
    [anon_sym_SLASH_EQ] = ACTIONS(4123),
    [anon_sym_PERCENT_EQ] = ACTIONS(4123),
    [anon_sym_x_EQ] = ACTIONS(4123),
    [anon_sym_AMP_EQ] = ACTIONS(4123),
    [anon_sym_PIPE_EQ] = ACTIONS(4123),
    [anon_sym_CARET_EQ] = ACTIONS(4123),
    [anon_sym_LT_LT_EQ] = ACTIONS(4123),
    [anon_sym_GT_GT_EQ] = ACTIONS(4123),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4123),
    [anon_sym_DOT_DOT] = ACTIONS(4125),
    [anon_sym_STAR_STAR] = ACTIONS(4125),
    [anon_sym_PIPE_PIPE] = ACTIONS(4125),
    [anon_sym_SLASH_SLASH] = ACTIONS(4125),
    [anon_sym_CARET_CARET] = ACTIONS(4123),
    [anon_sym_AMP_AMP] = ACTIONS(4125),
    [anon_sym_PIPE] = ACTIONS(4125),
    [anon_sym_CARET] = ACTIONS(4125),
    [anon_sym_AMP] = ACTIONS(4125),
    [anon_sym_GT_GT2] = ACTIONS(4125),
    [anon_sym_PLUS] = ACTIONS(4125),
    [anon_sym_DASH] = ACTIONS(4125),
    [anon_sym_DOT] = ACTIONS(4125),
    [anon_sym_SLASH] = ACTIONS(4125),
    [anon_sym_x] = ACTIONS(4125),
    [anon_sym_EQ_TILDE] = ACTIONS(4123),
    [anon_sym_BANG_TILDE] = ACTIONS(4123),
    [anon_sym_EQ_EQ] = ACTIONS(4125),
    [anon_sym_BANG_EQ] = ACTIONS(4123),
    [anon_sym_eq] = ACTIONS(4125),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4123),
    [anon_sym_equ] = ACTIONS(4123),
    [anon_sym_eqr] = ACTIONS(4123),
    [anon_sym_ne] = ACTIONS(4123),
    [anon_sym_LT_EQ_GT] = ACTIONS(4123),
    [anon_sym_cmp] = ACTIONS(4123),
    [anon_sym_TILDE_TILDE] = ACTIONS(4123),
    [anon_sym_LT] = ACTIONS(4125),
    [anon_sym_LT_EQ] = ACTIONS(4125),
    [anon_sym_GT_EQ] = ACTIONS(4123),
    [anon_sym_lt] = ACTIONS(4123),
    [anon_sym_le] = ACTIONS(4123),
    [anon_sym_ge] = ACTIONS(4123),
    [anon_sym_gt] = ACTIONS(4123),
    [anon_sym_isa] = ACTIONS(4123),
    [anon_sym_PLUS_PLUS] = ACTIONS(4123),
    [anon_sym_DASH_DASH] = ACTIONS(4123),
    [anon_sym_QMARK] = ACTIONS(4123),
    [anon_sym_EQ_GT] = ACTIONS(4123),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4123),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3651)] = {
    [sym_heredoc_content] = STATE(3651),
    [anon_sym_SEMI] = ACTIONS(3945),
    [anon_sym_EQ] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_COMMA] = ACTIONS(3945),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3945),
    [anon_sym_DASH_GT] = ACTIONS(3945),
    [anon_sym_STAR] = ACTIONS(3947),
    [anon_sym_PERCENT] = ACTIONS(3947),
    [anon_sym_GT] = ACTIONS(3947),
    [anon_sym_LT_LT] = ACTIONS(3947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3945),
    [anon_sym_PLUS_EQ] = ACTIONS(3945),
    [anon_sym_DASH_EQ] = ACTIONS(3945),
    [anon_sym_DOT_EQ] = ACTIONS(3945),
    [anon_sym_STAR_EQ] = ACTIONS(3945),
    [anon_sym_SLASH_EQ] = ACTIONS(3945),
    [anon_sym_PERCENT_EQ] = ACTIONS(3945),
    [anon_sym_x_EQ] = ACTIONS(3945),
    [anon_sym_AMP_EQ] = ACTIONS(3945),
    [anon_sym_PIPE_EQ] = ACTIONS(3945),
    [anon_sym_CARET_EQ] = ACTIONS(3945),
    [anon_sym_LT_LT_EQ] = ACTIONS(3945),
    [anon_sym_GT_GT_EQ] = ACTIONS(3945),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3945),
    [anon_sym_DOT_DOT] = ACTIONS(3947),
    [anon_sym_STAR_STAR] = ACTIONS(3947),
    [anon_sym_PIPE_PIPE] = ACTIONS(3947),
    [anon_sym_SLASH_SLASH] = ACTIONS(3947),
    [anon_sym_CARET_CARET] = ACTIONS(3945),
    [anon_sym_AMP_AMP] = ACTIONS(3947),
    [anon_sym_PIPE] = ACTIONS(3947),
    [anon_sym_CARET] = ACTIONS(3947),
    [anon_sym_AMP] = ACTIONS(3947),
    [anon_sym_GT_GT2] = ACTIONS(3947),
    [anon_sym_PLUS] = ACTIONS(3947),
    [anon_sym_DASH] = ACTIONS(3947),
    [anon_sym_DOT] = ACTIONS(3947),
    [anon_sym_SLASH] = ACTIONS(3947),
    [anon_sym_x] = ACTIONS(3947),
    [anon_sym_EQ_TILDE] = ACTIONS(3945),
    [anon_sym_BANG_TILDE] = ACTIONS(3945),
    [anon_sym_EQ_EQ] = ACTIONS(3947),
    [anon_sym_BANG_EQ] = ACTIONS(3945),
    [anon_sym_eq] = ACTIONS(3947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3945),
    [anon_sym_equ] = ACTIONS(3945),
    [anon_sym_eqr] = ACTIONS(3945),
    [anon_sym_ne] = ACTIONS(3945),
    [anon_sym_LT_EQ_GT] = ACTIONS(3945),
    [anon_sym_cmp] = ACTIONS(3945),
    [anon_sym_TILDE_TILDE] = ACTIONS(3945),
    [anon_sym_LT] = ACTIONS(3947),
    [anon_sym_LT_EQ] = ACTIONS(3947),
    [anon_sym_GT_EQ] = ACTIONS(3945),
    [anon_sym_lt] = ACTIONS(3945),
    [anon_sym_le] = ACTIONS(3945),
    [anon_sym_ge] = ACTIONS(3945),
    [anon_sym_gt] = ACTIONS(3945),
    [anon_sym_isa] = ACTIONS(3945),
    [anon_sym_PLUS_PLUS] = ACTIONS(3945),
    [anon_sym_DASH_DASH] = ACTIONS(3945),
    [anon_sym_QMARK] = ACTIONS(3945),
    [anon_sym_EQ_GT] = ACTIONS(3945),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3945),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3652)] = {
    [sym_heredoc_content] = STATE(3652),
    [anon_sym_SEMI] = ACTIONS(4196),
    [anon_sym_EQ] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_COMMA] = ACTIONS(4196),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4196),
    [anon_sym_DASH_GT] = ACTIONS(4196),
    [anon_sym_STAR] = ACTIONS(4198),
    [anon_sym_PERCENT] = ACTIONS(4198),
    [anon_sym_GT] = ACTIONS(4198),
    [anon_sym_LT_LT] = ACTIONS(4198),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4196),
    [anon_sym_PLUS_EQ] = ACTIONS(4196),
    [anon_sym_DASH_EQ] = ACTIONS(4196),
    [anon_sym_DOT_EQ] = ACTIONS(4196),
    [anon_sym_STAR_EQ] = ACTIONS(4196),
    [anon_sym_SLASH_EQ] = ACTIONS(4196),
    [anon_sym_PERCENT_EQ] = ACTIONS(4196),
    [anon_sym_x_EQ] = ACTIONS(4196),
    [anon_sym_AMP_EQ] = ACTIONS(4196),
    [anon_sym_PIPE_EQ] = ACTIONS(4196),
    [anon_sym_CARET_EQ] = ACTIONS(4196),
    [anon_sym_LT_LT_EQ] = ACTIONS(4196),
    [anon_sym_GT_GT_EQ] = ACTIONS(4196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4196),
    [anon_sym_DOT_DOT] = ACTIONS(4198),
    [anon_sym_STAR_STAR] = ACTIONS(4198),
    [anon_sym_PIPE_PIPE] = ACTIONS(4198),
    [anon_sym_SLASH_SLASH] = ACTIONS(4198),
    [anon_sym_CARET_CARET] = ACTIONS(4196),
    [anon_sym_AMP_AMP] = ACTIONS(4198),
    [anon_sym_PIPE] = ACTIONS(4198),
    [anon_sym_CARET] = ACTIONS(4198),
    [anon_sym_AMP] = ACTIONS(4198),
    [anon_sym_GT_GT2] = ACTIONS(4198),
    [anon_sym_PLUS] = ACTIONS(4198),
    [anon_sym_DASH] = ACTIONS(4198),
    [anon_sym_DOT] = ACTIONS(4198),
    [anon_sym_SLASH] = ACTIONS(4198),
    [anon_sym_x] = ACTIONS(4198),
    [anon_sym_EQ_TILDE] = ACTIONS(4196),
    [anon_sym_BANG_TILDE] = ACTIONS(4196),
    [anon_sym_EQ_EQ] = ACTIONS(4198),
    [anon_sym_BANG_EQ] = ACTIONS(4196),
    [anon_sym_eq] = ACTIONS(4198),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4196),
    [anon_sym_equ] = ACTIONS(4196),
    [anon_sym_eqr] = ACTIONS(4196),
    [anon_sym_ne] = ACTIONS(4196),
    [anon_sym_LT_EQ_GT] = ACTIONS(4196),
    [anon_sym_cmp] = ACTIONS(4196),
    [anon_sym_TILDE_TILDE] = ACTIONS(4196),
    [anon_sym_LT] = ACTIONS(4198),
    [anon_sym_LT_EQ] = ACTIONS(4198),
    [anon_sym_GT_EQ] = ACTIONS(4196),
    [anon_sym_lt] = ACTIONS(4196),
    [anon_sym_le] = ACTIONS(4196),
    [anon_sym_ge] = ACTIONS(4196),
    [anon_sym_gt] = ACTIONS(4196),
    [anon_sym_isa] = ACTIONS(4196),
    [anon_sym_PLUS_PLUS] = ACTIONS(4196),
    [anon_sym_DASH_DASH] = ACTIONS(4196),
    [anon_sym_QMARK] = ACTIONS(4196),
    [anon_sym_EQ_GT] = ACTIONS(4196),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4196),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3653)] = {
    [sym_heredoc_content] = STATE(3653),
    [anon_sym_SEMI] = ACTIONS(3905),
    [anon_sym_EQ] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_COMMA] = ACTIONS(3905),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3905),
    [anon_sym_DASH_GT] = ACTIONS(3905),
    [anon_sym_STAR] = ACTIONS(3907),
    [anon_sym_PERCENT] = ACTIONS(3907),
    [anon_sym_GT] = ACTIONS(3907),
    [anon_sym_LT_LT] = ACTIONS(3907),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3905),
    [anon_sym_PLUS_EQ] = ACTIONS(3905),
    [anon_sym_DASH_EQ] = ACTIONS(3905),
    [anon_sym_DOT_EQ] = ACTIONS(3905),
    [anon_sym_STAR_EQ] = ACTIONS(3905),
    [anon_sym_SLASH_EQ] = ACTIONS(3905),
    [anon_sym_PERCENT_EQ] = ACTIONS(3905),
    [anon_sym_x_EQ] = ACTIONS(3905),
    [anon_sym_AMP_EQ] = ACTIONS(3905),
    [anon_sym_PIPE_EQ] = ACTIONS(3905),
    [anon_sym_CARET_EQ] = ACTIONS(3905),
    [anon_sym_LT_LT_EQ] = ACTIONS(3905),
    [anon_sym_GT_GT_EQ] = ACTIONS(3905),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3905),
    [anon_sym_DOT_DOT] = ACTIONS(3907),
    [anon_sym_STAR_STAR] = ACTIONS(3907),
    [anon_sym_PIPE_PIPE] = ACTIONS(3907),
    [anon_sym_SLASH_SLASH] = ACTIONS(3907),
    [anon_sym_CARET_CARET] = ACTIONS(3905),
    [anon_sym_AMP_AMP] = ACTIONS(3907),
    [anon_sym_PIPE] = ACTIONS(3907),
    [anon_sym_CARET] = ACTIONS(3907),
    [anon_sym_AMP] = ACTIONS(3907),
    [anon_sym_GT_GT2] = ACTIONS(3907),
    [anon_sym_PLUS] = ACTIONS(3907),
    [anon_sym_DASH] = ACTIONS(3907),
    [anon_sym_DOT] = ACTIONS(3907),
    [anon_sym_SLASH] = ACTIONS(3907),
    [anon_sym_x] = ACTIONS(3907),
    [anon_sym_EQ_TILDE] = ACTIONS(3905),
    [anon_sym_BANG_TILDE] = ACTIONS(3905),
    [anon_sym_EQ_EQ] = ACTIONS(3907),
    [anon_sym_BANG_EQ] = ACTIONS(3905),
    [anon_sym_eq] = ACTIONS(3907),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3905),
    [anon_sym_equ] = ACTIONS(3905),
    [anon_sym_eqr] = ACTIONS(3905),
    [anon_sym_ne] = ACTIONS(3905),
    [anon_sym_LT_EQ_GT] = ACTIONS(3905),
    [anon_sym_cmp] = ACTIONS(3905),
    [anon_sym_TILDE_TILDE] = ACTIONS(3905),
    [anon_sym_LT] = ACTIONS(3907),
    [anon_sym_LT_EQ] = ACTIONS(3907),
    [anon_sym_GT_EQ] = ACTIONS(3905),
    [anon_sym_lt] = ACTIONS(3905),
    [anon_sym_le] = ACTIONS(3905),
    [anon_sym_ge] = ACTIONS(3905),
    [anon_sym_gt] = ACTIONS(3905),
    [anon_sym_isa] = ACTIONS(3905),
    [anon_sym_PLUS_PLUS] = ACTIONS(3905),
    [anon_sym_DASH_DASH] = ACTIONS(3905),
    [anon_sym_QMARK] = ACTIONS(3905),
    [anon_sym_EQ_GT] = ACTIONS(3905),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3905),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3654)] = {
    [sym_heredoc_content] = STATE(3654),
    [anon_sym_SEMI] = ACTIONS(4067),
    [anon_sym_EQ] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_COMMA] = ACTIONS(4067),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4067),
    [anon_sym_DASH_GT] = ACTIONS(4067),
    [anon_sym_STAR] = ACTIONS(4069),
    [anon_sym_PERCENT] = ACTIONS(4069),
    [anon_sym_GT] = ACTIONS(4069),
    [anon_sym_LT_LT] = ACTIONS(4069),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4067),
    [anon_sym_PLUS_EQ] = ACTIONS(4067),
    [anon_sym_DASH_EQ] = ACTIONS(4067),
    [anon_sym_DOT_EQ] = ACTIONS(4067),
    [anon_sym_STAR_EQ] = ACTIONS(4067),
    [anon_sym_SLASH_EQ] = ACTIONS(4067),
    [anon_sym_PERCENT_EQ] = ACTIONS(4067),
    [anon_sym_x_EQ] = ACTIONS(4067),
    [anon_sym_AMP_EQ] = ACTIONS(4067),
    [anon_sym_PIPE_EQ] = ACTIONS(4067),
    [anon_sym_CARET_EQ] = ACTIONS(4067),
    [anon_sym_LT_LT_EQ] = ACTIONS(4067),
    [anon_sym_GT_GT_EQ] = ACTIONS(4067),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4067),
    [anon_sym_DOT_DOT] = ACTIONS(4069),
    [anon_sym_STAR_STAR] = ACTIONS(4069),
    [anon_sym_PIPE_PIPE] = ACTIONS(4069),
    [anon_sym_SLASH_SLASH] = ACTIONS(4069),
    [anon_sym_CARET_CARET] = ACTIONS(4067),
    [anon_sym_AMP_AMP] = ACTIONS(4069),
    [anon_sym_PIPE] = ACTIONS(4069),
    [anon_sym_CARET] = ACTIONS(4069),
    [anon_sym_AMP] = ACTIONS(4069),
    [anon_sym_GT_GT2] = ACTIONS(4069),
    [anon_sym_PLUS] = ACTIONS(4069),
    [anon_sym_DASH] = ACTIONS(4069),
    [anon_sym_DOT] = ACTIONS(4069),
    [anon_sym_SLASH] = ACTIONS(4069),
    [anon_sym_x] = ACTIONS(4069),
    [anon_sym_EQ_TILDE] = ACTIONS(4067),
    [anon_sym_BANG_TILDE] = ACTIONS(4067),
    [anon_sym_EQ_EQ] = ACTIONS(4069),
    [anon_sym_BANG_EQ] = ACTIONS(4067),
    [anon_sym_eq] = ACTIONS(4069),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4067),
    [anon_sym_equ] = ACTIONS(4067),
    [anon_sym_eqr] = ACTIONS(4067),
    [anon_sym_ne] = ACTIONS(4067),
    [anon_sym_LT_EQ_GT] = ACTIONS(4067),
    [anon_sym_cmp] = ACTIONS(4067),
    [anon_sym_TILDE_TILDE] = ACTIONS(4067),
    [anon_sym_LT] = ACTIONS(4069),
    [anon_sym_LT_EQ] = ACTIONS(4069),
    [anon_sym_GT_EQ] = ACTIONS(4067),
    [anon_sym_lt] = ACTIONS(4067),
    [anon_sym_le] = ACTIONS(4067),
    [anon_sym_ge] = ACTIONS(4067),
    [anon_sym_gt] = ACTIONS(4067),
    [anon_sym_isa] = ACTIONS(4067),
    [anon_sym_PLUS_PLUS] = ACTIONS(4067),
    [anon_sym_DASH_DASH] = ACTIONS(4067),
    [anon_sym_QMARK] = ACTIONS(4067),
    [anon_sym_EQ_GT] = ACTIONS(4067),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4067),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3655)] = {
    [sym_heredoc_content] = STATE(3655),
    [anon_sym_SEMI] = ACTIONS(3825),
    [anon_sym_EQ] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_COMMA] = ACTIONS(3825),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3825),
    [anon_sym_DASH_GT] = ACTIONS(3825),
    [anon_sym_STAR] = ACTIONS(3827),
    [anon_sym_PERCENT] = ACTIONS(3827),
    [anon_sym_GT] = ACTIONS(3827),
    [anon_sym_LT_LT] = ACTIONS(3827),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3825),
    [anon_sym_PLUS_EQ] = ACTIONS(3825),
    [anon_sym_DASH_EQ] = ACTIONS(3825),
    [anon_sym_DOT_EQ] = ACTIONS(3825),
    [anon_sym_STAR_EQ] = ACTIONS(3825),
    [anon_sym_SLASH_EQ] = ACTIONS(3825),
    [anon_sym_PERCENT_EQ] = ACTIONS(3825),
    [anon_sym_x_EQ] = ACTIONS(3825),
    [anon_sym_AMP_EQ] = ACTIONS(3825),
    [anon_sym_PIPE_EQ] = ACTIONS(3825),
    [anon_sym_CARET_EQ] = ACTIONS(3825),
    [anon_sym_LT_LT_EQ] = ACTIONS(3825),
    [anon_sym_GT_GT_EQ] = ACTIONS(3825),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3825),
    [anon_sym_DOT_DOT] = ACTIONS(3827),
    [anon_sym_STAR_STAR] = ACTIONS(3827),
    [anon_sym_PIPE_PIPE] = ACTIONS(3827),
    [anon_sym_SLASH_SLASH] = ACTIONS(3827),
    [anon_sym_CARET_CARET] = ACTIONS(3825),
    [anon_sym_AMP_AMP] = ACTIONS(3827),
    [anon_sym_PIPE] = ACTIONS(3827),
    [anon_sym_CARET] = ACTIONS(3827),
    [anon_sym_AMP] = ACTIONS(3827),
    [anon_sym_GT_GT2] = ACTIONS(3827),
    [anon_sym_PLUS] = ACTIONS(3827),
    [anon_sym_DASH] = ACTIONS(3827),
    [anon_sym_DOT] = ACTIONS(3827),
    [anon_sym_SLASH] = ACTIONS(3827),
    [anon_sym_x] = ACTIONS(3827),
    [anon_sym_EQ_TILDE] = ACTIONS(3825),
    [anon_sym_BANG_TILDE] = ACTIONS(3825),
    [anon_sym_EQ_EQ] = ACTIONS(3827),
    [anon_sym_BANG_EQ] = ACTIONS(3825),
    [anon_sym_eq] = ACTIONS(3827),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3825),
    [anon_sym_equ] = ACTIONS(3825),
    [anon_sym_eqr] = ACTIONS(3825),
    [anon_sym_ne] = ACTIONS(3825),
    [anon_sym_LT_EQ_GT] = ACTIONS(3825),
    [anon_sym_cmp] = ACTIONS(3825),
    [anon_sym_TILDE_TILDE] = ACTIONS(3825),
    [anon_sym_LT] = ACTIONS(3827),
    [anon_sym_LT_EQ] = ACTIONS(3827),
    [anon_sym_GT_EQ] = ACTIONS(3825),
    [anon_sym_lt] = ACTIONS(3825),
    [anon_sym_le] = ACTIONS(3825),
    [anon_sym_ge] = ACTIONS(3825),
    [anon_sym_gt] = ACTIONS(3825),
    [anon_sym_isa] = ACTIONS(3825),
    [anon_sym_PLUS_PLUS] = ACTIONS(3825),
    [anon_sym_DASH_DASH] = ACTIONS(3825),
    [anon_sym_QMARK] = ACTIONS(3825),
    [anon_sym_EQ_GT] = ACTIONS(3825),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3825),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3656)] = {
    [sym_heredoc_content] = STATE(3656),
    [anon_sym_SEMI] = ACTIONS(3849),
    [anon_sym_EQ] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_COMMA] = ACTIONS(3849),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3849),
    [anon_sym_DASH_GT] = ACTIONS(3849),
    [anon_sym_STAR] = ACTIONS(3851),
    [anon_sym_PERCENT] = ACTIONS(3851),
    [anon_sym_GT] = ACTIONS(3851),
    [anon_sym_LT_LT] = ACTIONS(3851),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3849),
    [anon_sym_PLUS_EQ] = ACTIONS(3849),
    [anon_sym_DASH_EQ] = ACTIONS(3849),
    [anon_sym_DOT_EQ] = ACTIONS(3849),
    [anon_sym_STAR_EQ] = ACTIONS(3849),
    [anon_sym_SLASH_EQ] = ACTIONS(3849),
    [anon_sym_PERCENT_EQ] = ACTIONS(3849),
    [anon_sym_x_EQ] = ACTIONS(3849),
    [anon_sym_AMP_EQ] = ACTIONS(3849),
    [anon_sym_PIPE_EQ] = ACTIONS(3849),
    [anon_sym_CARET_EQ] = ACTIONS(3849),
    [anon_sym_LT_LT_EQ] = ACTIONS(3849),
    [anon_sym_GT_GT_EQ] = ACTIONS(3849),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3849),
    [anon_sym_DOT_DOT] = ACTIONS(3851),
    [anon_sym_STAR_STAR] = ACTIONS(3851),
    [anon_sym_PIPE_PIPE] = ACTIONS(3851),
    [anon_sym_SLASH_SLASH] = ACTIONS(3851),
    [anon_sym_CARET_CARET] = ACTIONS(3849),
    [anon_sym_AMP_AMP] = ACTIONS(3851),
    [anon_sym_PIPE] = ACTIONS(3851),
    [anon_sym_CARET] = ACTIONS(3851),
    [anon_sym_AMP] = ACTIONS(3851),
    [anon_sym_GT_GT2] = ACTIONS(3851),
    [anon_sym_PLUS] = ACTIONS(3851),
    [anon_sym_DASH] = ACTIONS(3851),
    [anon_sym_DOT] = ACTIONS(3851),
    [anon_sym_SLASH] = ACTIONS(3851),
    [anon_sym_x] = ACTIONS(3851),
    [anon_sym_EQ_TILDE] = ACTIONS(3849),
    [anon_sym_BANG_TILDE] = ACTIONS(3849),
    [anon_sym_EQ_EQ] = ACTIONS(3851),
    [anon_sym_BANG_EQ] = ACTIONS(3849),
    [anon_sym_eq] = ACTIONS(3851),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3849),
    [anon_sym_equ] = ACTIONS(3849),
    [anon_sym_eqr] = ACTIONS(3849),
    [anon_sym_ne] = ACTIONS(3849),
    [anon_sym_LT_EQ_GT] = ACTIONS(3849),
    [anon_sym_cmp] = ACTIONS(3849),
    [anon_sym_TILDE_TILDE] = ACTIONS(3849),
    [anon_sym_LT] = ACTIONS(3851),
    [anon_sym_LT_EQ] = ACTIONS(3851),
    [anon_sym_GT_EQ] = ACTIONS(3849),
    [anon_sym_lt] = ACTIONS(3849),
    [anon_sym_le] = ACTIONS(3849),
    [anon_sym_ge] = ACTIONS(3849),
    [anon_sym_gt] = ACTIONS(3849),
    [anon_sym_isa] = ACTIONS(3849),
    [anon_sym_PLUS_PLUS] = ACTIONS(3849),
    [anon_sym_DASH_DASH] = ACTIONS(3849),
    [anon_sym_QMARK] = ACTIONS(3849),
    [anon_sym_EQ_GT] = ACTIONS(3849),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3849),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3657)] = {
    [sym_heredoc_content] = STATE(3657),
    [anon_sym_SEMI] = ACTIONS(4051),
    [anon_sym_EQ] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_COMMA] = ACTIONS(4051),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4051),
    [anon_sym_DASH_GT] = ACTIONS(4051),
    [anon_sym_STAR] = ACTIONS(4053),
    [anon_sym_PERCENT] = ACTIONS(4053),
    [anon_sym_GT] = ACTIONS(4053),
    [anon_sym_LT_LT] = ACTIONS(4053),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4051),
    [anon_sym_PLUS_EQ] = ACTIONS(4051),
    [anon_sym_DASH_EQ] = ACTIONS(4051),
    [anon_sym_DOT_EQ] = ACTIONS(4051),
    [anon_sym_STAR_EQ] = ACTIONS(4051),
    [anon_sym_SLASH_EQ] = ACTIONS(4051),
    [anon_sym_PERCENT_EQ] = ACTIONS(4051),
    [anon_sym_x_EQ] = ACTIONS(4051),
    [anon_sym_AMP_EQ] = ACTIONS(4051),
    [anon_sym_PIPE_EQ] = ACTIONS(4051),
    [anon_sym_CARET_EQ] = ACTIONS(4051),
    [anon_sym_LT_LT_EQ] = ACTIONS(4051),
    [anon_sym_GT_GT_EQ] = ACTIONS(4051),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4051),
    [anon_sym_DOT_DOT] = ACTIONS(4053),
    [anon_sym_STAR_STAR] = ACTIONS(4053),
    [anon_sym_PIPE_PIPE] = ACTIONS(4053),
    [anon_sym_SLASH_SLASH] = ACTIONS(4053),
    [anon_sym_CARET_CARET] = ACTIONS(4051),
    [anon_sym_AMP_AMP] = ACTIONS(4053),
    [anon_sym_PIPE] = ACTIONS(4053),
    [anon_sym_CARET] = ACTIONS(4053),
    [anon_sym_AMP] = ACTIONS(4053),
    [anon_sym_GT_GT2] = ACTIONS(4053),
    [anon_sym_PLUS] = ACTIONS(4053),
    [anon_sym_DASH] = ACTIONS(4053),
    [anon_sym_DOT] = ACTIONS(4053),
    [anon_sym_SLASH] = ACTIONS(4053),
    [anon_sym_x] = ACTIONS(4053),
    [anon_sym_EQ_TILDE] = ACTIONS(4051),
    [anon_sym_BANG_TILDE] = ACTIONS(4051),
    [anon_sym_EQ_EQ] = ACTIONS(4053),
    [anon_sym_BANG_EQ] = ACTIONS(4051),
    [anon_sym_eq] = ACTIONS(4053),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4051),
    [anon_sym_equ] = ACTIONS(4051),
    [anon_sym_eqr] = ACTIONS(4051),
    [anon_sym_ne] = ACTIONS(4051),
    [anon_sym_LT_EQ_GT] = ACTIONS(4051),
    [anon_sym_cmp] = ACTIONS(4051),
    [anon_sym_TILDE_TILDE] = ACTIONS(4051),
    [anon_sym_LT] = ACTIONS(4053),
    [anon_sym_LT_EQ] = ACTIONS(4053),
    [anon_sym_GT_EQ] = ACTIONS(4051),
    [anon_sym_lt] = ACTIONS(4051),
    [anon_sym_le] = ACTIONS(4051),
    [anon_sym_ge] = ACTIONS(4051),
    [anon_sym_gt] = ACTIONS(4051),
    [anon_sym_isa] = ACTIONS(4051),
    [anon_sym_PLUS_PLUS] = ACTIONS(4051),
    [anon_sym_DASH_DASH] = ACTIONS(4051),
    [anon_sym_QMARK] = ACTIONS(4051),
    [anon_sym_EQ_GT] = ACTIONS(4051),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4051),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3658)] = {
    [sym_heredoc_content] = STATE(3658),
    [anon_sym_SEMI] = ACTIONS(4059),
    [anon_sym_EQ] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_COMMA] = ACTIONS(4059),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4059),
    [anon_sym_DASH_GT] = ACTIONS(4059),
    [anon_sym_STAR] = ACTIONS(4061),
    [anon_sym_PERCENT] = ACTIONS(4061),
    [anon_sym_GT] = ACTIONS(4061),
    [anon_sym_LT_LT] = ACTIONS(4061),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4059),
    [anon_sym_PLUS_EQ] = ACTIONS(4059),
    [anon_sym_DASH_EQ] = ACTIONS(4059),
    [anon_sym_DOT_EQ] = ACTIONS(4059),
    [anon_sym_STAR_EQ] = ACTIONS(4059),
    [anon_sym_SLASH_EQ] = ACTIONS(4059),
    [anon_sym_PERCENT_EQ] = ACTIONS(4059),
    [anon_sym_x_EQ] = ACTIONS(4059),
    [anon_sym_AMP_EQ] = ACTIONS(4059),
    [anon_sym_PIPE_EQ] = ACTIONS(4059),
    [anon_sym_CARET_EQ] = ACTIONS(4059),
    [anon_sym_LT_LT_EQ] = ACTIONS(4059),
    [anon_sym_GT_GT_EQ] = ACTIONS(4059),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4059),
    [anon_sym_DOT_DOT] = ACTIONS(4061),
    [anon_sym_STAR_STAR] = ACTIONS(4061),
    [anon_sym_PIPE_PIPE] = ACTIONS(4061),
    [anon_sym_SLASH_SLASH] = ACTIONS(4061),
    [anon_sym_CARET_CARET] = ACTIONS(4059),
    [anon_sym_AMP_AMP] = ACTIONS(4061),
    [anon_sym_PIPE] = ACTIONS(4061),
    [anon_sym_CARET] = ACTIONS(4061),
    [anon_sym_AMP] = ACTIONS(4061),
    [anon_sym_GT_GT2] = ACTIONS(4061),
    [anon_sym_PLUS] = ACTIONS(4061),
    [anon_sym_DASH] = ACTIONS(4061),
    [anon_sym_DOT] = ACTIONS(4061),
    [anon_sym_SLASH] = ACTIONS(4061),
    [anon_sym_x] = ACTIONS(4061),
    [anon_sym_EQ_TILDE] = ACTIONS(4059),
    [anon_sym_BANG_TILDE] = ACTIONS(4059),
    [anon_sym_EQ_EQ] = ACTIONS(4061),
    [anon_sym_BANG_EQ] = ACTIONS(4059),
    [anon_sym_eq] = ACTIONS(4061),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4059),
    [anon_sym_equ] = ACTIONS(4059),
    [anon_sym_eqr] = ACTIONS(4059),
    [anon_sym_ne] = ACTIONS(4059),
    [anon_sym_LT_EQ_GT] = ACTIONS(4059),
    [anon_sym_cmp] = ACTIONS(4059),
    [anon_sym_TILDE_TILDE] = ACTIONS(4059),
    [anon_sym_LT] = ACTIONS(4061),
    [anon_sym_LT_EQ] = ACTIONS(4061),
    [anon_sym_GT_EQ] = ACTIONS(4059),
    [anon_sym_lt] = ACTIONS(4059),
    [anon_sym_le] = ACTIONS(4059),
    [anon_sym_ge] = ACTIONS(4059),
    [anon_sym_gt] = ACTIONS(4059),
    [anon_sym_isa] = ACTIONS(4059),
    [anon_sym_PLUS_PLUS] = ACTIONS(4059),
    [anon_sym_DASH_DASH] = ACTIONS(4059),
    [anon_sym_QMARK] = ACTIONS(4059),
    [anon_sym_EQ_GT] = ACTIONS(4059),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4059),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3659)] = {
    [sym_heredoc_content] = STATE(3659),
    [anon_sym_SEMI] = ACTIONS(4099),
    [anon_sym_EQ] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_COMMA] = ACTIONS(4099),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4099),
    [anon_sym_DASH_GT] = ACTIONS(4099),
    [anon_sym_STAR] = ACTIONS(4101),
    [anon_sym_PERCENT] = ACTIONS(4101),
    [anon_sym_GT] = ACTIONS(4101),
    [anon_sym_LT_LT] = ACTIONS(4101),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4099),
    [anon_sym_PLUS_EQ] = ACTIONS(4099),
    [anon_sym_DASH_EQ] = ACTIONS(4099),
    [anon_sym_DOT_EQ] = ACTIONS(4099),
    [anon_sym_STAR_EQ] = ACTIONS(4099),
    [anon_sym_SLASH_EQ] = ACTIONS(4099),
    [anon_sym_PERCENT_EQ] = ACTIONS(4099),
    [anon_sym_x_EQ] = ACTIONS(4099),
    [anon_sym_AMP_EQ] = ACTIONS(4099),
    [anon_sym_PIPE_EQ] = ACTIONS(4099),
    [anon_sym_CARET_EQ] = ACTIONS(4099),
    [anon_sym_LT_LT_EQ] = ACTIONS(4099),
    [anon_sym_GT_GT_EQ] = ACTIONS(4099),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4099),
    [anon_sym_DOT_DOT] = ACTIONS(4101),
    [anon_sym_STAR_STAR] = ACTIONS(4101),
    [anon_sym_PIPE_PIPE] = ACTIONS(4101),
    [anon_sym_SLASH_SLASH] = ACTIONS(4101),
    [anon_sym_CARET_CARET] = ACTIONS(4099),
    [anon_sym_AMP_AMP] = ACTIONS(4101),
    [anon_sym_PIPE] = ACTIONS(4101),
    [anon_sym_CARET] = ACTIONS(4101),
    [anon_sym_AMP] = ACTIONS(4101),
    [anon_sym_GT_GT2] = ACTIONS(4101),
    [anon_sym_PLUS] = ACTIONS(4101),
    [anon_sym_DASH] = ACTIONS(4101),
    [anon_sym_DOT] = ACTIONS(4101),
    [anon_sym_SLASH] = ACTIONS(4101),
    [anon_sym_x] = ACTIONS(4101),
    [anon_sym_EQ_TILDE] = ACTIONS(4099),
    [anon_sym_BANG_TILDE] = ACTIONS(4099),
    [anon_sym_EQ_EQ] = ACTIONS(4101),
    [anon_sym_BANG_EQ] = ACTIONS(4099),
    [anon_sym_eq] = ACTIONS(4101),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4099),
    [anon_sym_equ] = ACTIONS(4099),
    [anon_sym_eqr] = ACTIONS(4099),
    [anon_sym_ne] = ACTIONS(4099),
    [anon_sym_LT_EQ_GT] = ACTIONS(4099),
    [anon_sym_cmp] = ACTIONS(4099),
    [anon_sym_TILDE_TILDE] = ACTIONS(4099),
    [anon_sym_LT] = ACTIONS(4101),
    [anon_sym_LT_EQ] = ACTIONS(4101),
    [anon_sym_GT_EQ] = ACTIONS(4099),
    [anon_sym_lt] = ACTIONS(4099),
    [anon_sym_le] = ACTIONS(4099),
    [anon_sym_ge] = ACTIONS(4099),
    [anon_sym_gt] = ACTIONS(4099),
    [anon_sym_isa] = ACTIONS(4099),
    [anon_sym_PLUS_PLUS] = ACTIONS(4099),
    [anon_sym_DASH_DASH] = ACTIONS(4099),
    [anon_sym_QMARK] = ACTIONS(4099),
    [anon_sym_EQ_GT] = ACTIONS(4099),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4099),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3660)] = {
    [sym_heredoc_content] = STATE(3660),
    [anon_sym_SEMI] = ACTIONS(4127),
    [anon_sym_EQ] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_COMMA] = ACTIONS(4127),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4127),
    [anon_sym_DASH_GT] = ACTIONS(4127),
    [anon_sym_STAR] = ACTIONS(4129),
    [anon_sym_PERCENT] = ACTIONS(4129),
    [anon_sym_GT] = ACTIONS(4129),
    [anon_sym_LT_LT] = ACTIONS(4129),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4127),
    [anon_sym_PLUS_EQ] = ACTIONS(4127),
    [anon_sym_DASH_EQ] = ACTIONS(4127),
    [anon_sym_DOT_EQ] = ACTIONS(4127),
    [anon_sym_STAR_EQ] = ACTIONS(4127),
    [anon_sym_SLASH_EQ] = ACTIONS(4127),
    [anon_sym_PERCENT_EQ] = ACTIONS(4127),
    [anon_sym_x_EQ] = ACTIONS(4127),
    [anon_sym_AMP_EQ] = ACTIONS(4127),
    [anon_sym_PIPE_EQ] = ACTIONS(4127),
    [anon_sym_CARET_EQ] = ACTIONS(4127),
    [anon_sym_LT_LT_EQ] = ACTIONS(4127),
    [anon_sym_GT_GT_EQ] = ACTIONS(4127),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4127),
    [anon_sym_DOT_DOT] = ACTIONS(4129),
    [anon_sym_STAR_STAR] = ACTIONS(4129),
    [anon_sym_PIPE_PIPE] = ACTIONS(4129),
    [anon_sym_SLASH_SLASH] = ACTIONS(4129),
    [anon_sym_CARET_CARET] = ACTIONS(4127),
    [anon_sym_AMP_AMP] = ACTIONS(4129),
    [anon_sym_PIPE] = ACTIONS(4129),
    [anon_sym_CARET] = ACTIONS(4129),
    [anon_sym_AMP] = ACTIONS(4129),
    [anon_sym_GT_GT2] = ACTIONS(4129),
    [anon_sym_PLUS] = ACTIONS(4129),
    [anon_sym_DASH] = ACTIONS(4129),
    [anon_sym_DOT] = ACTIONS(4129),
    [anon_sym_SLASH] = ACTIONS(4129),
    [anon_sym_x] = ACTIONS(4129),
    [anon_sym_EQ_TILDE] = ACTIONS(4127),
    [anon_sym_BANG_TILDE] = ACTIONS(4127),
    [anon_sym_EQ_EQ] = ACTIONS(4129),
    [anon_sym_BANG_EQ] = ACTIONS(4127),
    [anon_sym_eq] = ACTIONS(4129),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4127),
    [anon_sym_equ] = ACTIONS(4127),
    [anon_sym_eqr] = ACTIONS(4127),
    [anon_sym_ne] = ACTIONS(4127),
    [anon_sym_LT_EQ_GT] = ACTIONS(4127),
    [anon_sym_cmp] = ACTIONS(4127),
    [anon_sym_TILDE_TILDE] = ACTIONS(4127),
    [anon_sym_LT] = ACTIONS(4129),
    [anon_sym_LT_EQ] = ACTIONS(4129),
    [anon_sym_GT_EQ] = ACTIONS(4127),
    [anon_sym_lt] = ACTIONS(4127),
    [anon_sym_le] = ACTIONS(4127),
    [anon_sym_ge] = ACTIONS(4127),
    [anon_sym_gt] = ACTIONS(4127),
    [anon_sym_isa] = ACTIONS(4127),
    [anon_sym_PLUS_PLUS] = ACTIONS(4127),
    [anon_sym_DASH_DASH] = ACTIONS(4127),
    [anon_sym_QMARK] = ACTIONS(4127),
    [anon_sym_EQ_GT] = ACTIONS(4127),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4127),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3661)] = {
    [sym_heredoc_content] = STATE(3661),
    [anon_sym_SEMI] = ACTIONS(4147),
    [anon_sym_EQ] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_COMMA] = ACTIONS(4147),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4147),
    [anon_sym_DASH_GT] = ACTIONS(4147),
    [anon_sym_STAR] = ACTIONS(4149),
    [anon_sym_PERCENT] = ACTIONS(4149),
    [anon_sym_GT] = ACTIONS(4149),
    [anon_sym_LT_LT] = ACTIONS(4149),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4147),
    [anon_sym_PLUS_EQ] = ACTIONS(4147),
    [anon_sym_DASH_EQ] = ACTIONS(4147),
    [anon_sym_DOT_EQ] = ACTIONS(4147),
    [anon_sym_STAR_EQ] = ACTIONS(4147),
    [anon_sym_SLASH_EQ] = ACTIONS(4147),
    [anon_sym_PERCENT_EQ] = ACTIONS(4147),
    [anon_sym_x_EQ] = ACTIONS(4147),
    [anon_sym_AMP_EQ] = ACTIONS(4147),
    [anon_sym_PIPE_EQ] = ACTIONS(4147),
    [anon_sym_CARET_EQ] = ACTIONS(4147),
    [anon_sym_LT_LT_EQ] = ACTIONS(4147),
    [anon_sym_GT_GT_EQ] = ACTIONS(4147),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4147),
    [anon_sym_DOT_DOT] = ACTIONS(4149),
    [anon_sym_STAR_STAR] = ACTIONS(4149),
    [anon_sym_PIPE_PIPE] = ACTIONS(4149),
    [anon_sym_SLASH_SLASH] = ACTIONS(4149),
    [anon_sym_CARET_CARET] = ACTIONS(4147),
    [anon_sym_AMP_AMP] = ACTIONS(4149),
    [anon_sym_PIPE] = ACTIONS(4149),
    [anon_sym_CARET] = ACTIONS(4149),
    [anon_sym_AMP] = ACTIONS(4149),
    [anon_sym_GT_GT2] = ACTIONS(4149),
    [anon_sym_PLUS] = ACTIONS(4149),
    [anon_sym_DASH] = ACTIONS(4149),
    [anon_sym_DOT] = ACTIONS(4149),
    [anon_sym_SLASH] = ACTIONS(4149),
    [anon_sym_x] = ACTIONS(4149),
    [anon_sym_EQ_TILDE] = ACTIONS(4147),
    [anon_sym_BANG_TILDE] = ACTIONS(4147),
    [anon_sym_EQ_EQ] = ACTIONS(4149),
    [anon_sym_BANG_EQ] = ACTIONS(4147),
    [anon_sym_eq] = ACTIONS(4149),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4147),
    [anon_sym_equ] = ACTIONS(4147),
    [anon_sym_eqr] = ACTIONS(4147),
    [anon_sym_ne] = ACTIONS(4147),
    [anon_sym_LT_EQ_GT] = ACTIONS(4147),
    [anon_sym_cmp] = ACTIONS(4147),
    [anon_sym_TILDE_TILDE] = ACTIONS(4147),
    [anon_sym_LT] = ACTIONS(4149),
    [anon_sym_LT_EQ] = ACTIONS(4149),
    [anon_sym_GT_EQ] = ACTIONS(4147),
    [anon_sym_lt] = ACTIONS(4147),
    [anon_sym_le] = ACTIONS(4147),
    [anon_sym_ge] = ACTIONS(4147),
    [anon_sym_gt] = ACTIONS(4147),
    [anon_sym_isa] = ACTIONS(4147),
    [anon_sym_PLUS_PLUS] = ACTIONS(4147),
    [anon_sym_DASH_DASH] = ACTIONS(4147),
    [anon_sym_QMARK] = ACTIONS(4147),
    [anon_sym_EQ_GT] = ACTIONS(4147),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4147),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3662)] = {
    [sym_heredoc_content] = STATE(3662),
    [anon_sym_SEMI] = ACTIONS(3961),
    [anon_sym_EQ] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_COMMA] = ACTIONS(3961),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3961),
    [anon_sym_DASH_GT] = ACTIONS(3961),
    [anon_sym_STAR] = ACTIONS(3963),
    [anon_sym_PERCENT] = ACTIONS(3963),
    [anon_sym_GT] = ACTIONS(3963),
    [anon_sym_LT_LT] = ACTIONS(3963),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3961),
    [anon_sym_PLUS_EQ] = ACTIONS(3961),
    [anon_sym_DASH_EQ] = ACTIONS(3961),
    [anon_sym_DOT_EQ] = ACTIONS(3961),
    [anon_sym_STAR_EQ] = ACTIONS(3961),
    [anon_sym_SLASH_EQ] = ACTIONS(3961),
    [anon_sym_PERCENT_EQ] = ACTIONS(3961),
    [anon_sym_x_EQ] = ACTIONS(3961),
    [anon_sym_AMP_EQ] = ACTIONS(3961),
    [anon_sym_PIPE_EQ] = ACTIONS(3961),
    [anon_sym_CARET_EQ] = ACTIONS(3961),
    [anon_sym_LT_LT_EQ] = ACTIONS(3961),
    [anon_sym_GT_GT_EQ] = ACTIONS(3961),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3961),
    [anon_sym_DOT_DOT] = ACTIONS(3963),
    [anon_sym_STAR_STAR] = ACTIONS(3963),
    [anon_sym_PIPE_PIPE] = ACTIONS(3963),
    [anon_sym_SLASH_SLASH] = ACTIONS(3963),
    [anon_sym_CARET_CARET] = ACTIONS(3961),
    [anon_sym_AMP_AMP] = ACTIONS(3963),
    [anon_sym_PIPE] = ACTIONS(3963),
    [anon_sym_CARET] = ACTIONS(3963),
    [anon_sym_AMP] = ACTIONS(3963),
    [anon_sym_GT_GT2] = ACTIONS(3963),
    [anon_sym_PLUS] = ACTIONS(3963),
    [anon_sym_DASH] = ACTIONS(3963),
    [anon_sym_DOT] = ACTIONS(3963),
    [anon_sym_SLASH] = ACTIONS(3963),
    [anon_sym_x] = ACTIONS(3963),
    [anon_sym_EQ_TILDE] = ACTIONS(3961),
    [anon_sym_BANG_TILDE] = ACTIONS(3961),
    [anon_sym_EQ_EQ] = ACTIONS(3963),
    [anon_sym_BANG_EQ] = ACTIONS(3961),
    [anon_sym_eq] = ACTIONS(3963),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3961),
    [anon_sym_equ] = ACTIONS(3961),
    [anon_sym_eqr] = ACTIONS(3961),
    [anon_sym_ne] = ACTIONS(3961),
    [anon_sym_LT_EQ_GT] = ACTIONS(3961),
    [anon_sym_cmp] = ACTIONS(3961),
    [anon_sym_TILDE_TILDE] = ACTIONS(3961),
    [anon_sym_LT] = ACTIONS(3963),
    [anon_sym_LT_EQ] = ACTIONS(3963),
    [anon_sym_GT_EQ] = ACTIONS(3961),
    [anon_sym_lt] = ACTIONS(3961),
    [anon_sym_le] = ACTIONS(3961),
    [anon_sym_ge] = ACTIONS(3961),
    [anon_sym_gt] = ACTIONS(3961),
    [anon_sym_isa] = ACTIONS(3961),
    [anon_sym_PLUS_PLUS] = ACTIONS(3961),
    [anon_sym_DASH_DASH] = ACTIONS(3961),
    [anon_sym_QMARK] = ACTIONS(3961),
    [anon_sym_EQ_GT] = ACTIONS(3961),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3961),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3663)] = {
    [sym_heredoc_content] = STATE(3663),
    [anon_sym_SEMI] = ACTIONS(3913),
    [anon_sym_EQ] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_COMMA] = ACTIONS(3913),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3913),
    [anon_sym_DASH_GT] = ACTIONS(3913),
    [anon_sym_STAR] = ACTIONS(3915),
    [anon_sym_PERCENT] = ACTIONS(3915),
    [anon_sym_GT] = ACTIONS(3915),
    [anon_sym_LT_LT] = ACTIONS(3915),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3913),
    [anon_sym_PLUS_EQ] = ACTIONS(3913),
    [anon_sym_DASH_EQ] = ACTIONS(3913),
    [anon_sym_DOT_EQ] = ACTIONS(3913),
    [anon_sym_STAR_EQ] = ACTIONS(3913),
    [anon_sym_SLASH_EQ] = ACTIONS(3913),
    [anon_sym_PERCENT_EQ] = ACTIONS(3913),
    [anon_sym_x_EQ] = ACTIONS(3913),
    [anon_sym_AMP_EQ] = ACTIONS(3913),
    [anon_sym_PIPE_EQ] = ACTIONS(3913),
    [anon_sym_CARET_EQ] = ACTIONS(3913),
    [anon_sym_LT_LT_EQ] = ACTIONS(3913),
    [anon_sym_GT_GT_EQ] = ACTIONS(3913),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3913),
    [anon_sym_DOT_DOT] = ACTIONS(3915),
    [anon_sym_STAR_STAR] = ACTIONS(3915),
    [anon_sym_PIPE_PIPE] = ACTIONS(3915),
    [anon_sym_SLASH_SLASH] = ACTIONS(3915),
    [anon_sym_CARET_CARET] = ACTIONS(3913),
    [anon_sym_AMP_AMP] = ACTIONS(3915),
    [anon_sym_PIPE] = ACTIONS(3915),
    [anon_sym_CARET] = ACTIONS(3915),
    [anon_sym_AMP] = ACTIONS(3915),
    [anon_sym_GT_GT2] = ACTIONS(3915),
    [anon_sym_PLUS] = ACTIONS(3915),
    [anon_sym_DASH] = ACTIONS(3915),
    [anon_sym_DOT] = ACTIONS(3915),
    [anon_sym_SLASH] = ACTIONS(3915),
    [anon_sym_x] = ACTIONS(3915),
    [anon_sym_EQ_TILDE] = ACTIONS(3913),
    [anon_sym_BANG_TILDE] = ACTIONS(3913),
    [anon_sym_EQ_EQ] = ACTIONS(3915),
    [anon_sym_BANG_EQ] = ACTIONS(3913),
    [anon_sym_eq] = ACTIONS(3915),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3913),
    [anon_sym_equ] = ACTIONS(3913),
    [anon_sym_eqr] = ACTIONS(3913),
    [anon_sym_ne] = ACTIONS(3913),
    [anon_sym_LT_EQ_GT] = ACTIONS(3913),
    [anon_sym_cmp] = ACTIONS(3913),
    [anon_sym_TILDE_TILDE] = ACTIONS(3913),
    [anon_sym_LT] = ACTIONS(3915),
    [anon_sym_LT_EQ] = ACTIONS(3915),
    [anon_sym_GT_EQ] = ACTIONS(3913),
    [anon_sym_lt] = ACTIONS(3913),
    [anon_sym_le] = ACTIONS(3913),
    [anon_sym_ge] = ACTIONS(3913),
    [anon_sym_gt] = ACTIONS(3913),
    [anon_sym_isa] = ACTIONS(3913),
    [anon_sym_PLUS_PLUS] = ACTIONS(3913),
    [anon_sym_DASH_DASH] = ACTIONS(3913),
    [anon_sym_QMARK] = ACTIONS(3913),
    [anon_sym_EQ_GT] = ACTIONS(3913),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3913),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3664)] = {
    [sym_heredoc_content] = STATE(3664),
    [anon_sym_SEMI] = ACTIONS(4143),
    [anon_sym_EQ] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_COMMA] = ACTIONS(4143),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4143),
    [anon_sym_DASH_GT] = ACTIONS(4143),
    [anon_sym_STAR] = ACTIONS(4145),
    [anon_sym_PERCENT] = ACTIONS(4145),
    [anon_sym_GT] = ACTIONS(4145),
    [anon_sym_LT_LT] = ACTIONS(4145),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4143),
    [anon_sym_PLUS_EQ] = ACTIONS(4143),
    [anon_sym_DASH_EQ] = ACTIONS(4143),
    [anon_sym_DOT_EQ] = ACTIONS(4143),
    [anon_sym_STAR_EQ] = ACTIONS(4143),
    [anon_sym_SLASH_EQ] = ACTIONS(4143),
    [anon_sym_PERCENT_EQ] = ACTIONS(4143),
    [anon_sym_x_EQ] = ACTIONS(4143),
    [anon_sym_AMP_EQ] = ACTIONS(4143),
    [anon_sym_PIPE_EQ] = ACTIONS(4143),
    [anon_sym_CARET_EQ] = ACTIONS(4143),
    [anon_sym_LT_LT_EQ] = ACTIONS(4143),
    [anon_sym_GT_GT_EQ] = ACTIONS(4143),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4143),
    [anon_sym_DOT_DOT] = ACTIONS(4145),
    [anon_sym_STAR_STAR] = ACTIONS(4145),
    [anon_sym_PIPE_PIPE] = ACTIONS(4145),
    [anon_sym_SLASH_SLASH] = ACTIONS(4145),
    [anon_sym_CARET_CARET] = ACTIONS(4143),
    [anon_sym_AMP_AMP] = ACTIONS(4145),
    [anon_sym_PIPE] = ACTIONS(4145),
    [anon_sym_CARET] = ACTIONS(4145),
    [anon_sym_AMP] = ACTIONS(4145),
    [anon_sym_GT_GT2] = ACTIONS(4145),
    [anon_sym_PLUS] = ACTIONS(4145),
    [anon_sym_DASH] = ACTIONS(4145),
    [anon_sym_DOT] = ACTIONS(4145),
    [anon_sym_SLASH] = ACTIONS(4145),
    [anon_sym_x] = ACTIONS(4145),
    [anon_sym_EQ_TILDE] = ACTIONS(4143),
    [anon_sym_BANG_TILDE] = ACTIONS(4143),
    [anon_sym_EQ_EQ] = ACTIONS(4145),
    [anon_sym_BANG_EQ] = ACTIONS(4143),
    [anon_sym_eq] = ACTIONS(4145),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4143),
    [anon_sym_equ] = ACTIONS(4143),
    [anon_sym_eqr] = ACTIONS(4143),
    [anon_sym_ne] = ACTIONS(4143),
    [anon_sym_LT_EQ_GT] = ACTIONS(4143),
    [anon_sym_cmp] = ACTIONS(4143),
    [anon_sym_TILDE_TILDE] = ACTIONS(4143),
    [anon_sym_LT] = ACTIONS(4145),
    [anon_sym_LT_EQ] = ACTIONS(4145),
    [anon_sym_GT_EQ] = ACTIONS(4143),
    [anon_sym_lt] = ACTIONS(4143),
    [anon_sym_le] = ACTIONS(4143),
    [anon_sym_ge] = ACTIONS(4143),
    [anon_sym_gt] = ACTIONS(4143),
    [anon_sym_isa] = ACTIONS(4143),
    [anon_sym_PLUS_PLUS] = ACTIONS(4143),
    [anon_sym_DASH_DASH] = ACTIONS(4143),
    [anon_sym_QMARK] = ACTIONS(4143),
    [anon_sym_EQ_GT] = ACTIONS(4143),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4143),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3665)] = {
    [sym_heredoc_content] = STATE(3665),
    [anon_sym_SEMI] = ACTIONS(3799),
    [anon_sym_EQ] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_COMMA] = ACTIONS(3799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3799),
    [anon_sym_DASH_GT] = ACTIONS(3799),
    [anon_sym_STAR] = ACTIONS(3801),
    [anon_sym_PERCENT] = ACTIONS(3801),
    [anon_sym_GT] = ACTIONS(3801),
    [anon_sym_LT_LT] = ACTIONS(3801),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3799),
    [anon_sym_PLUS_EQ] = ACTIONS(3799),
    [anon_sym_DASH_EQ] = ACTIONS(3799),
    [anon_sym_DOT_EQ] = ACTIONS(3799),
    [anon_sym_STAR_EQ] = ACTIONS(3799),
    [anon_sym_SLASH_EQ] = ACTIONS(3799),
    [anon_sym_PERCENT_EQ] = ACTIONS(3799),
    [anon_sym_x_EQ] = ACTIONS(3799),
    [anon_sym_AMP_EQ] = ACTIONS(3799),
    [anon_sym_PIPE_EQ] = ACTIONS(3799),
    [anon_sym_CARET_EQ] = ACTIONS(3799),
    [anon_sym_LT_LT_EQ] = ACTIONS(3799),
    [anon_sym_GT_GT_EQ] = ACTIONS(3799),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3799),
    [anon_sym_DOT_DOT] = ACTIONS(3801),
    [anon_sym_STAR_STAR] = ACTIONS(3801),
    [anon_sym_PIPE_PIPE] = ACTIONS(3801),
    [anon_sym_SLASH_SLASH] = ACTIONS(3801),
    [anon_sym_CARET_CARET] = ACTIONS(3799),
    [anon_sym_AMP_AMP] = ACTIONS(3801),
    [anon_sym_PIPE] = ACTIONS(3801),
    [anon_sym_CARET] = ACTIONS(3801),
    [anon_sym_AMP] = ACTIONS(3801),
    [anon_sym_GT_GT2] = ACTIONS(3801),
    [anon_sym_PLUS] = ACTIONS(3801),
    [anon_sym_DASH] = ACTIONS(3801),
    [anon_sym_DOT] = ACTIONS(3801),
    [anon_sym_SLASH] = ACTIONS(3801),
    [anon_sym_x] = ACTIONS(3801),
    [anon_sym_EQ_TILDE] = ACTIONS(3799),
    [anon_sym_BANG_TILDE] = ACTIONS(3799),
    [anon_sym_EQ_EQ] = ACTIONS(3801),
    [anon_sym_BANG_EQ] = ACTIONS(3799),
    [anon_sym_eq] = ACTIONS(3801),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3799),
    [anon_sym_equ] = ACTIONS(3799),
    [anon_sym_eqr] = ACTIONS(3799),
    [anon_sym_ne] = ACTIONS(3799),
    [anon_sym_LT_EQ_GT] = ACTIONS(3799),
    [anon_sym_cmp] = ACTIONS(3799),
    [anon_sym_TILDE_TILDE] = ACTIONS(3799),
    [anon_sym_LT] = ACTIONS(3801),
    [anon_sym_LT_EQ] = ACTIONS(3801),
    [anon_sym_GT_EQ] = ACTIONS(3799),
    [anon_sym_lt] = ACTIONS(3799),
    [anon_sym_le] = ACTIONS(3799),
    [anon_sym_ge] = ACTIONS(3799),
    [anon_sym_gt] = ACTIONS(3799),
    [anon_sym_isa] = ACTIONS(3799),
    [anon_sym_PLUS_PLUS] = ACTIONS(3799),
    [anon_sym_DASH_DASH] = ACTIONS(3799),
    [anon_sym_QMARK] = ACTIONS(3799),
    [anon_sym_EQ_GT] = ACTIONS(3799),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3799),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3666)] = {
    [sym_heredoc_content] = STATE(3666),
    [anon_sym_SEMI] = ACTIONS(3921),
    [anon_sym_EQ] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_COMMA] = ACTIONS(3921),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3921),
    [anon_sym_DASH_GT] = ACTIONS(3921),
    [anon_sym_STAR] = ACTIONS(3923),
    [anon_sym_PERCENT] = ACTIONS(3923),
    [anon_sym_GT] = ACTIONS(3923),
    [anon_sym_LT_LT] = ACTIONS(3923),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3921),
    [anon_sym_PLUS_EQ] = ACTIONS(3921),
    [anon_sym_DASH_EQ] = ACTIONS(3921),
    [anon_sym_DOT_EQ] = ACTIONS(3921),
    [anon_sym_STAR_EQ] = ACTIONS(3921),
    [anon_sym_SLASH_EQ] = ACTIONS(3921),
    [anon_sym_PERCENT_EQ] = ACTIONS(3921),
    [anon_sym_x_EQ] = ACTIONS(3921),
    [anon_sym_AMP_EQ] = ACTIONS(3921),
    [anon_sym_PIPE_EQ] = ACTIONS(3921),
    [anon_sym_CARET_EQ] = ACTIONS(3921),
    [anon_sym_LT_LT_EQ] = ACTIONS(3921),
    [anon_sym_GT_GT_EQ] = ACTIONS(3921),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3921),
    [anon_sym_DOT_DOT] = ACTIONS(3923),
    [anon_sym_STAR_STAR] = ACTIONS(3923),
    [anon_sym_PIPE_PIPE] = ACTIONS(3923),
    [anon_sym_SLASH_SLASH] = ACTIONS(3923),
    [anon_sym_CARET_CARET] = ACTIONS(3921),
    [anon_sym_AMP_AMP] = ACTIONS(3923),
    [anon_sym_PIPE] = ACTIONS(3923),
    [anon_sym_CARET] = ACTIONS(3923),
    [anon_sym_AMP] = ACTIONS(3923),
    [anon_sym_GT_GT2] = ACTIONS(3923),
    [anon_sym_PLUS] = ACTIONS(3923),
    [anon_sym_DASH] = ACTIONS(3923),
    [anon_sym_DOT] = ACTIONS(3923),
    [anon_sym_SLASH] = ACTIONS(3923),
    [anon_sym_x] = ACTIONS(3923),
    [anon_sym_EQ_TILDE] = ACTIONS(3921),
    [anon_sym_BANG_TILDE] = ACTIONS(3921),
    [anon_sym_EQ_EQ] = ACTIONS(3923),
    [anon_sym_BANG_EQ] = ACTIONS(3921),
    [anon_sym_eq] = ACTIONS(3923),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3921),
    [anon_sym_equ] = ACTIONS(3921),
    [anon_sym_eqr] = ACTIONS(3921),
    [anon_sym_ne] = ACTIONS(3921),
    [anon_sym_LT_EQ_GT] = ACTIONS(3921),
    [anon_sym_cmp] = ACTIONS(3921),
    [anon_sym_TILDE_TILDE] = ACTIONS(3921),
    [anon_sym_LT] = ACTIONS(3923),
    [anon_sym_LT_EQ] = ACTIONS(3923),
    [anon_sym_GT_EQ] = ACTIONS(3921),
    [anon_sym_lt] = ACTIONS(3921),
    [anon_sym_le] = ACTIONS(3921),
    [anon_sym_ge] = ACTIONS(3921),
    [anon_sym_gt] = ACTIONS(3921),
    [anon_sym_isa] = ACTIONS(3921),
    [anon_sym_PLUS_PLUS] = ACTIONS(3921),
    [anon_sym_DASH_DASH] = ACTIONS(3921),
    [anon_sym_QMARK] = ACTIONS(3921),
    [anon_sym_EQ_GT] = ACTIONS(3921),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3921),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3667)] = {
    [sym_heredoc_content] = STATE(3667),
    [anon_sym_SEMI] = ACTIONS(3723),
    [anon_sym_EQ] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_COMMA] = ACTIONS(3723),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3723),
    [anon_sym_DASH_GT] = ACTIONS(3723),
    [anon_sym_STAR] = ACTIONS(3725),
    [anon_sym_PERCENT] = ACTIONS(3725),
    [anon_sym_GT] = ACTIONS(3725),
    [anon_sym_LT_LT] = ACTIONS(3725),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3723),
    [anon_sym_PLUS_EQ] = ACTIONS(3723),
    [anon_sym_DASH_EQ] = ACTIONS(3723),
    [anon_sym_DOT_EQ] = ACTIONS(3723),
    [anon_sym_STAR_EQ] = ACTIONS(3723),
    [anon_sym_SLASH_EQ] = ACTIONS(3723),
    [anon_sym_PERCENT_EQ] = ACTIONS(3723),
    [anon_sym_x_EQ] = ACTIONS(3723),
    [anon_sym_AMP_EQ] = ACTIONS(3723),
    [anon_sym_PIPE_EQ] = ACTIONS(3723),
    [anon_sym_CARET_EQ] = ACTIONS(3723),
    [anon_sym_LT_LT_EQ] = ACTIONS(3723),
    [anon_sym_GT_GT_EQ] = ACTIONS(3723),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3723),
    [anon_sym_DOT_DOT] = ACTIONS(3725),
    [anon_sym_STAR_STAR] = ACTIONS(3725),
    [anon_sym_PIPE_PIPE] = ACTIONS(3725),
    [anon_sym_SLASH_SLASH] = ACTIONS(3725),
    [anon_sym_CARET_CARET] = ACTIONS(3723),
    [anon_sym_AMP_AMP] = ACTIONS(3725),
    [anon_sym_PIPE] = ACTIONS(3725),
    [anon_sym_CARET] = ACTIONS(3725),
    [anon_sym_AMP] = ACTIONS(3725),
    [anon_sym_GT_GT2] = ACTIONS(3725),
    [anon_sym_PLUS] = ACTIONS(3725),
    [anon_sym_DASH] = ACTIONS(3725),
    [anon_sym_DOT] = ACTIONS(3725),
    [anon_sym_SLASH] = ACTIONS(3725),
    [anon_sym_x] = ACTIONS(3725),
    [anon_sym_EQ_TILDE] = ACTIONS(3723),
    [anon_sym_BANG_TILDE] = ACTIONS(3723),
    [anon_sym_EQ_EQ] = ACTIONS(3725),
    [anon_sym_BANG_EQ] = ACTIONS(3723),
    [anon_sym_eq] = ACTIONS(3725),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3723),
    [anon_sym_equ] = ACTIONS(3723),
    [anon_sym_eqr] = ACTIONS(3723),
    [anon_sym_ne] = ACTIONS(3723),
    [anon_sym_LT_EQ_GT] = ACTIONS(3723),
    [anon_sym_cmp] = ACTIONS(3723),
    [anon_sym_TILDE_TILDE] = ACTIONS(3723),
    [anon_sym_LT] = ACTIONS(3725),
    [anon_sym_LT_EQ] = ACTIONS(3725),
    [anon_sym_GT_EQ] = ACTIONS(3723),
    [anon_sym_lt] = ACTIONS(3723),
    [anon_sym_le] = ACTIONS(3723),
    [anon_sym_ge] = ACTIONS(3723),
    [anon_sym_gt] = ACTIONS(3723),
    [anon_sym_isa] = ACTIONS(3723),
    [anon_sym_PLUS_PLUS] = ACTIONS(3723),
    [anon_sym_DASH_DASH] = ACTIONS(3723),
    [anon_sym_QMARK] = ACTIONS(3723),
    [anon_sym_EQ_GT] = ACTIONS(3723),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3723),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3668)] = {
    [sym_heredoc_content] = STATE(3668),
    [anon_sym_SEMI] = ACTIONS(3751),
    [anon_sym_EQ] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_COMMA] = ACTIONS(3751),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3751),
    [anon_sym_DASH_GT] = ACTIONS(3751),
    [anon_sym_STAR] = ACTIONS(3753),
    [anon_sym_PERCENT] = ACTIONS(3753),
    [anon_sym_GT] = ACTIONS(3753),
    [anon_sym_LT_LT] = ACTIONS(3753),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3751),
    [anon_sym_PLUS_EQ] = ACTIONS(3751),
    [anon_sym_DASH_EQ] = ACTIONS(3751),
    [anon_sym_DOT_EQ] = ACTIONS(3751),
    [anon_sym_STAR_EQ] = ACTIONS(3751),
    [anon_sym_SLASH_EQ] = ACTIONS(3751),
    [anon_sym_PERCENT_EQ] = ACTIONS(3751),
    [anon_sym_x_EQ] = ACTIONS(3751),
    [anon_sym_AMP_EQ] = ACTIONS(3751),
    [anon_sym_PIPE_EQ] = ACTIONS(3751),
    [anon_sym_CARET_EQ] = ACTIONS(3751),
    [anon_sym_LT_LT_EQ] = ACTIONS(3751),
    [anon_sym_GT_GT_EQ] = ACTIONS(3751),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3751),
    [anon_sym_DOT_DOT] = ACTIONS(3753),
    [anon_sym_STAR_STAR] = ACTIONS(3753),
    [anon_sym_PIPE_PIPE] = ACTIONS(3753),
    [anon_sym_SLASH_SLASH] = ACTIONS(3753),
    [anon_sym_CARET_CARET] = ACTIONS(3751),
    [anon_sym_AMP_AMP] = ACTIONS(3753),
    [anon_sym_PIPE] = ACTIONS(3753),
    [anon_sym_CARET] = ACTIONS(3753),
    [anon_sym_AMP] = ACTIONS(3753),
    [anon_sym_GT_GT2] = ACTIONS(3753),
    [anon_sym_PLUS] = ACTIONS(3753),
    [anon_sym_DASH] = ACTIONS(3753),
    [anon_sym_DOT] = ACTIONS(3753),
    [anon_sym_SLASH] = ACTIONS(3753),
    [anon_sym_x] = ACTIONS(3753),
    [anon_sym_EQ_TILDE] = ACTIONS(3751),
    [anon_sym_BANG_TILDE] = ACTIONS(3751),
    [anon_sym_EQ_EQ] = ACTIONS(3753),
    [anon_sym_BANG_EQ] = ACTIONS(3751),
    [anon_sym_eq] = ACTIONS(3753),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3751),
    [anon_sym_equ] = ACTIONS(3751),
    [anon_sym_eqr] = ACTIONS(3751),
    [anon_sym_ne] = ACTIONS(3751),
    [anon_sym_LT_EQ_GT] = ACTIONS(3751),
    [anon_sym_cmp] = ACTIONS(3751),
    [anon_sym_TILDE_TILDE] = ACTIONS(3751),
    [anon_sym_LT] = ACTIONS(3753),
    [anon_sym_LT_EQ] = ACTIONS(3753),
    [anon_sym_GT_EQ] = ACTIONS(3751),
    [anon_sym_lt] = ACTIONS(3751),
    [anon_sym_le] = ACTIONS(3751),
    [anon_sym_ge] = ACTIONS(3751),
    [anon_sym_gt] = ACTIONS(3751),
    [anon_sym_isa] = ACTIONS(3751),
    [anon_sym_PLUS_PLUS] = ACTIONS(3751),
    [anon_sym_DASH_DASH] = ACTIONS(3751),
    [anon_sym_QMARK] = ACTIONS(3751),
    [anon_sym_EQ_GT] = ACTIONS(3751),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3751),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3669)] = {
    [sym_heredoc_content] = STATE(3669),
    [anon_sym_SEMI] = ACTIONS(3755),
    [anon_sym_EQ] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_COMMA] = ACTIONS(3755),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3755),
    [anon_sym_DASH_GT] = ACTIONS(3755),
    [anon_sym_STAR] = ACTIONS(3757),
    [anon_sym_PERCENT] = ACTIONS(3757),
    [anon_sym_GT] = ACTIONS(3757),
    [anon_sym_LT_LT] = ACTIONS(3757),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3755),
    [anon_sym_PLUS_EQ] = ACTIONS(3755),
    [anon_sym_DASH_EQ] = ACTIONS(3755),
    [anon_sym_DOT_EQ] = ACTIONS(3755),
    [anon_sym_STAR_EQ] = ACTIONS(3755),
    [anon_sym_SLASH_EQ] = ACTIONS(3755),
    [anon_sym_PERCENT_EQ] = ACTIONS(3755),
    [anon_sym_x_EQ] = ACTIONS(3755),
    [anon_sym_AMP_EQ] = ACTIONS(3755),
    [anon_sym_PIPE_EQ] = ACTIONS(3755),
    [anon_sym_CARET_EQ] = ACTIONS(3755),
    [anon_sym_LT_LT_EQ] = ACTIONS(3755),
    [anon_sym_GT_GT_EQ] = ACTIONS(3755),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3755),
    [anon_sym_DOT_DOT] = ACTIONS(3757),
    [anon_sym_STAR_STAR] = ACTIONS(3757),
    [anon_sym_PIPE_PIPE] = ACTIONS(3757),
    [anon_sym_SLASH_SLASH] = ACTIONS(3757),
    [anon_sym_CARET_CARET] = ACTIONS(3755),
    [anon_sym_AMP_AMP] = ACTIONS(3757),
    [anon_sym_PIPE] = ACTIONS(3757),
    [anon_sym_CARET] = ACTIONS(3757),
    [anon_sym_AMP] = ACTIONS(3757),
    [anon_sym_GT_GT2] = ACTIONS(3757),
    [anon_sym_PLUS] = ACTIONS(3757),
    [anon_sym_DASH] = ACTIONS(3757),
    [anon_sym_DOT] = ACTIONS(3757),
    [anon_sym_SLASH] = ACTIONS(3757),
    [anon_sym_x] = ACTIONS(3757),
    [anon_sym_EQ_TILDE] = ACTIONS(3755),
    [anon_sym_BANG_TILDE] = ACTIONS(3755),
    [anon_sym_EQ_EQ] = ACTIONS(3757),
    [anon_sym_BANG_EQ] = ACTIONS(3755),
    [anon_sym_eq] = ACTIONS(3757),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3755),
    [anon_sym_equ] = ACTIONS(3755),
    [anon_sym_eqr] = ACTIONS(3755),
    [anon_sym_ne] = ACTIONS(3755),
    [anon_sym_LT_EQ_GT] = ACTIONS(3755),
    [anon_sym_cmp] = ACTIONS(3755),
    [anon_sym_TILDE_TILDE] = ACTIONS(3755),
    [anon_sym_LT] = ACTIONS(3757),
    [anon_sym_LT_EQ] = ACTIONS(3757),
    [anon_sym_GT_EQ] = ACTIONS(3755),
    [anon_sym_lt] = ACTIONS(3755),
    [anon_sym_le] = ACTIONS(3755),
    [anon_sym_ge] = ACTIONS(3755),
    [anon_sym_gt] = ACTIONS(3755),
    [anon_sym_isa] = ACTIONS(3755),
    [anon_sym_PLUS_PLUS] = ACTIONS(3755),
    [anon_sym_DASH_DASH] = ACTIONS(3755),
    [anon_sym_QMARK] = ACTIONS(3755),
    [anon_sym_EQ_GT] = ACTIONS(3755),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3755),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3670)] = {
    [sym_heredoc_content] = STATE(3670),
    [anon_sym_SEMI] = ACTIONS(3759),
    [anon_sym_EQ] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_COMMA] = ACTIONS(3759),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3759),
    [anon_sym_DASH_GT] = ACTIONS(3759),
    [anon_sym_STAR] = ACTIONS(3761),
    [anon_sym_PERCENT] = ACTIONS(3761),
    [anon_sym_GT] = ACTIONS(3761),
    [anon_sym_LT_LT] = ACTIONS(3761),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3759),
    [anon_sym_PLUS_EQ] = ACTIONS(3759),
    [anon_sym_DASH_EQ] = ACTIONS(3759),
    [anon_sym_DOT_EQ] = ACTIONS(3759),
    [anon_sym_STAR_EQ] = ACTIONS(3759),
    [anon_sym_SLASH_EQ] = ACTIONS(3759),
    [anon_sym_PERCENT_EQ] = ACTIONS(3759),
    [anon_sym_x_EQ] = ACTIONS(3759),
    [anon_sym_AMP_EQ] = ACTIONS(3759),
    [anon_sym_PIPE_EQ] = ACTIONS(3759),
    [anon_sym_CARET_EQ] = ACTIONS(3759),
    [anon_sym_LT_LT_EQ] = ACTIONS(3759),
    [anon_sym_GT_GT_EQ] = ACTIONS(3759),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3759),
    [anon_sym_DOT_DOT] = ACTIONS(3761),
    [anon_sym_STAR_STAR] = ACTIONS(3761),
    [anon_sym_PIPE_PIPE] = ACTIONS(3761),
    [anon_sym_SLASH_SLASH] = ACTIONS(3761),
    [anon_sym_CARET_CARET] = ACTIONS(3759),
    [anon_sym_AMP_AMP] = ACTIONS(3761),
    [anon_sym_PIPE] = ACTIONS(3761),
    [anon_sym_CARET] = ACTIONS(3761),
    [anon_sym_AMP] = ACTIONS(3761),
    [anon_sym_GT_GT2] = ACTIONS(3761),
    [anon_sym_PLUS] = ACTIONS(3761),
    [anon_sym_DASH] = ACTIONS(3761),
    [anon_sym_DOT] = ACTIONS(3761),
    [anon_sym_SLASH] = ACTIONS(3761),
    [anon_sym_x] = ACTIONS(3761),
    [anon_sym_EQ_TILDE] = ACTIONS(3759),
    [anon_sym_BANG_TILDE] = ACTIONS(3759),
    [anon_sym_EQ_EQ] = ACTIONS(3761),
    [anon_sym_BANG_EQ] = ACTIONS(3759),
    [anon_sym_eq] = ACTIONS(3761),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3759),
    [anon_sym_equ] = ACTIONS(3759),
    [anon_sym_eqr] = ACTIONS(3759),
    [anon_sym_ne] = ACTIONS(3759),
    [anon_sym_LT_EQ_GT] = ACTIONS(3759),
    [anon_sym_cmp] = ACTIONS(3759),
    [anon_sym_TILDE_TILDE] = ACTIONS(3759),
    [anon_sym_LT] = ACTIONS(3761),
    [anon_sym_LT_EQ] = ACTIONS(3761),
    [anon_sym_GT_EQ] = ACTIONS(3759),
    [anon_sym_lt] = ACTIONS(3759),
    [anon_sym_le] = ACTIONS(3759),
    [anon_sym_ge] = ACTIONS(3759),
    [anon_sym_gt] = ACTIONS(3759),
    [anon_sym_isa] = ACTIONS(3759),
    [anon_sym_PLUS_PLUS] = ACTIONS(3759),
    [anon_sym_DASH_DASH] = ACTIONS(3759),
    [anon_sym_QMARK] = ACTIONS(3759),
    [anon_sym_EQ_GT] = ACTIONS(3759),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(3759),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3671)] = {
    [sym_heredoc_content] = STATE(3671),
    [anon_sym_RBRACE] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3672)] = {
    [sym_heredoc_content] = STATE(3672),
    [anon_sym_RBRACE] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_RPAREN] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3673)] = {
    [sym_heredoc_content] = STATE(3673),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5393),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5395),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3674)] = {
    [sym_heredoc_content] = STATE(3674),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3675)] = {
    [sym_heredoc_content] = STATE(3675),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3676)] = {
    [sym_heredoc_content] = STATE(3676),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3677)] = {
    [sym_heredoc_content] = STATE(3677),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3678)] = {
    [sym_heredoc_content] = STATE(3678),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3679)] = {
    [sym_heredoc_content] = STATE(3679),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3680)] = {
    [sym_heredoc_content] = STATE(3680),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3681)] = {
    [sym_heredoc_content] = STATE(3681),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3682)] = {
    [sym_heredoc_content] = STATE(3682),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3683)] = {
    [sym_heredoc_content] = STATE(3683),
    [anon_sym_RBRACE] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3684)] = {
    [sym_heredoc_content] = STATE(3684),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3685)] = {
    [sym_heredoc_content] = STATE(3685),
    [anon_sym_RBRACE] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5397),
    [anon_sym_cmp] = ACTIONS(5397),
    [anon_sym_TILDE_TILDE] = ACTIONS(5397),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3686)] = {
    [sym_heredoc_content] = STATE(3686),
    [anon_sym_RBRACE] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5399),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3687)] = {
    [sym_heredoc_content] = STATE(3687),
    [anon_sym_SEMI] = ACTIONS(2858),
    [anon_sym_EQ] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_COMMA] = ACTIONS(2858),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2858),
    [anon_sym_DASH_GT] = ACTIONS(2858),
    [anon_sym_STAR] = ACTIONS(2854),
    [anon_sym_PERCENT] = ACTIONS(2854),
    [anon_sym_GT] = ACTIONS(2854),
    [anon_sym_LT_LT] = ACTIONS(2854),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(2858),
    [anon_sym_PLUS_EQ] = ACTIONS(2858),
    [anon_sym_DASH_EQ] = ACTIONS(2858),
    [anon_sym_DOT_EQ] = ACTIONS(2858),
    [anon_sym_STAR_EQ] = ACTIONS(2858),
    [anon_sym_SLASH_EQ] = ACTIONS(2858),
    [anon_sym_PERCENT_EQ] = ACTIONS(2858),
    [anon_sym_x_EQ] = ACTIONS(2858),
    [anon_sym_AMP_EQ] = ACTIONS(2858),
    [anon_sym_PIPE_EQ] = ACTIONS(2858),
    [anon_sym_CARET_EQ] = ACTIONS(2858),
    [anon_sym_LT_LT_EQ] = ACTIONS(2858),
    [anon_sym_GT_GT_EQ] = ACTIONS(2858),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(2858),
    [anon_sym_DOT_DOT] = ACTIONS(2854),
    [anon_sym_STAR_STAR] = ACTIONS(2854),
    [anon_sym_PIPE_PIPE] = ACTIONS(2854),
    [anon_sym_SLASH_SLASH] = ACTIONS(2854),
    [anon_sym_CARET_CARET] = ACTIONS(2858),
    [anon_sym_AMP_AMP] = ACTIONS(2854),
    [anon_sym_PIPE] = ACTIONS(2854),
    [anon_sym_CARET] = ACTIONS(2854),
    [anon_sym_AMP] = ACTIONS(2854),
    [anon_sym_GT_GT2] = ACTIONS(2854),
    [anon_sym_PLUS] = ACTIONS(2854),
    [anon_sym_DASH] = ACTIONS(2854),
    [anon_sym_DOT] = ACTIONS(2854),
    [anon_sym_SLASH] = ACTIONS(2854),
    [anon_sym_x] = ACTIONS(2854),
    [anon_sym_EQ_TILDE] = ACTIONS(2858),
    [anon_sym_BANG_TILDE] = ACTIONS(2858),
    [anon_sym_EQ_EQ] = ACTIONS(2854),
    [anon_sym_BANG_EQ] = ACTIONS(2858),
    [anon_sym_eq] = ACTIONS(2854),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2858),
    [anon_sym_equ] = ACTIONS(2858),
    [anon_sym_eqr] = ACTIONS(2858),
    [anon_sym_ne] = ACTIONS(2858),
    [anon_sym_LT_EQ_GT] = ACTIONS(2858),
    [anon_sym_cmp] = ACTIONS(2858),
    [anon_sym_TILDE_TILDE] = ACTIONS(2858),
    [anon_sym_LT] = ACTIONS(2854),
    [anon_sym_LT_EQ] = ACTIONS(2854),
    [anon_sym_GT_EQ] = ACTIONS(2858),
    [anon_sym_lt] = ACTIONS(2858),
    [anon_sym_le] = ACTIONS(2858),
    [anon_sym_ge] = ACTIONS(2858),
    [anon_sym_gt] = ACTIONS(2858),
    [anon_sym_isa] = ACTIONS(2858),
    [anon_sym_PLUS_PLUS] = ACTIONS(2858),
    [anon_sym_DASH_DASH] = ACTIONS(2858),
    [anon_sym_QMARK] = ACTIONS(2858),
    [anon_sym_EQ_GT] = ACTIONS(2858),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(2858),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3688)] = {
    [sym_heredoc_content] = STATE(3688),
    [anon_sym_SEMI] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(5261),
    [anon_sym_STAR] = ACTIONS(5263),
    [anon_sym_PERCENT] = ACTIONS(5263),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5267),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5271),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5267),
    [anon_sym_PLUS] = ACTIONS(5283),
    [anon_sym_DASH] = ACTIONS(5283),
    [anon_sym_DOT] = ACTIONS(5283),
    [anon_sym_SLASH] = ACTIONS(5263),
    [anon_sym_x] = ACTIONS(5263),
    [anon_sym_EQ_TILDE] = ACTIONS(5285),
    [anon_sym_BANG_TILDE] = ACTIONS(5285),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(5297),
    [anon_sym_DASH_DASH] = ACTIONS(5297),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4151),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3689)] = {
    [sym_heredoc_content] = STATE(3689),
    [anon_sym_SEMI] = ACTIONS(4188),
    [anon_sym_EQ] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_COMMA] = ACTIONS(4188),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4188),
    [anon_sym_DASH_GT] = ACTIONS(4188),
    [anon_sym_STAR] = ACTIONS(4190),
    [anon_sym_PERCENT] = ACTIONS(4190),
    [anon_sym_GT] = ACTIONS(4190),
    [anon_sym_LT_LT] = ACTIONS(4190),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4188),
    [anon_sym_PLUS_EQ] = ACTIONS(4188),
    [anon_sym_DASH_EQ] = ACTIONS(4188),
    [anon_sym_DOT_EQ] = ACTIONS(4188),
    [anon_sym_STAR_EQ] = ACTIONS(4188),
    [anon_sym_SLASH_EQ] = ACTIONS(4188),
    [anon_sym_PERCENT_EQ] = ACTIONS(4188),
    [anon_sym_x_EQ] = ACTIONS(4188),
    [anon_sym_AMP_EQ] = ACTIONS(4188),
    [anon_sym_PIPE_EQ] = ACTIONS(4188),
    [anon_sym_CARET_EQ] = ACTIONS(4188),
    [anon_sym_LT_LT_EQ] = ACTIONS(4188),
    [anon_sym_GT_GT_EQ] = ACTIONS(4188),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4188),
    [anon_sym_DOT_DOT] = ACTIONS(4190),
    [anon_sym_STAR_STAR] = ACTIONS(4190),
    [anon_sym_PIPE_PIPE] = ACTIONS(4190),
    [anon_sym_SLASH_SLASH] = ACTIONS(4190),
    [anon_sym_CARET_CARET] = ACTIONS(4188),
    [anon_sym_AMP_AMP] = ACTIONS(4190),
    [anon_sym_PIPE] = ACTIONS(4190),
    [anon_sym_CARET] = ACTIONS(4190),
    [anon_sym_AMP] = ACTIONS(4190),
    [anon_sym_GT_GT2] = ACTIONS(4190),
    [anon_sym_PLUS] = ACTIONS(4190),
    [anon_sym_DASH] = ACTIONS(4190),
    [anon_sym_DOT] = ACTIONS(4190),
    [anon_sym_SLASH] = ACTIONS(4190),
    [anon_sym_x] = ACTIONS(4190),
    [anon_sym_EQ_TILDE] = ACTIONS(4188),
    [anon_sym_BANG_TILDE] = ACTIONS(4188),
    [anon_sym_EQ_EQ] = ACTIONS(4190),
    [anon_sym_BANG_EQ] = ACTIONS(4188),
    [anon_sym_eq] = ACTIONS(4190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4188),
    [anon_sym_equ] = ACTIONS(4188),
    [anon_sym_eqr] = ACTIONS(4188),
    [anon_sym_ne] = ACTIONS(4188),
    [anon_sym_LT_EQ_GT] = ACTIONS(4188),
    [anon_sym_cmp] = ACTIONS(4188),
    [anon_sym_TILDE_TILDE] = ACTIONS(4188),
    [anon_sym_LT] = ACTIONS(4190),
    [anon_sym_LT_EQ] = ACTIONS(4190),
    [anon_sym_GT_EQ] = ACTIONS(4188),
    [anon_sym_lt] = ACTIONS(4188),
    [anon_sym_le] = ACTIONS(4188),
    [anon_sym_ge] = ACTIONS(4188),
    [anon_sym_gt] = ACTIONS(4188),
    [anon_sym_isa] = ACTIONS(4188),
    [anon_sym_PLUS_PLUS] = ACTIONS(4188),
    [anon_sym_DASH_DASH] = ACTIONS(4188),
    [anon_sym_QMARK] = ACTIONS(4188),
    [anon_sym_EQ_GT] = ACTIONS(4188),
    [sym_comment] = ACTIONS(5),
    [sym__PERLY_SEMICOLON] = ACTIONS(4188),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3690)] = {
    [sym_heredoc_content] = STATE(3690),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5401),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5401),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3691)] = {
    [sym_heredoc_content] = STATE(3691),
    [anon_sym_COLON] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(3985),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3692)] = {
    [sym_heredoc_content] = STATE(3692),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5403),
    [anon_sym_RPAREN] = ACTIONS(3819),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5403),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3693)] = {
    [sym_heredoc_content] = STATE(3693),
    [anon_sym_COLON] = ACTIONS(3689),
    [anon_sym_EQ] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_COMMA] = ACTIONS(3689),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3689),
    [anon_sym_DASH_GT] = ACTIONS(3689),
    [anon_sym_STAR] = ACTIONS(3693),
    [anon_sym_PERCENT] = ACTIONS(3693),
    [anon_sym_GT] = ACTIONS(3693),
    [anon_sym_LT_LT] = ACTIONS(3693),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3689),
    [anon_sym_PLUS_EQ] = ACTIONS(3689),
    [anon_sym_DASH_EQ] = ACTIONS(3689),
    [anon_sym_DOT_EQ] = ACTIONS(3689),
    [anon_sym_STAR_EQ] = ACTIONS(3689),
    [anon_sym_SLASH_EQ] = ACTIONS(3689),
    [anon_sym_PERCENT_EQ] = ACTIONS(3689),
    [anon_sym_x_EQ] = ACTIONS(3689),
    [anon_sym_AMP_EQ] = ACTIONS(3689),
    [anon_sym_PIPE_EQ] = ACTIONS(3689),
    [anon_sym_CARET_EQ] = ACTIONS(3689),
    [anon_sym_LT_LT_EQ] = ACTIONS(3689),
    [anon_sym_GT_GT_EQ] = ACTIONS(3689),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3689),
    [anon_sym_DOT_DOT] = ACTIONS(3693),
    [anon_sym_STAR_STAR] = ACTIONS(3693),
    [anon_sym_PIPE_PIPE] = ACTIONS(3693),
    [anon_sym_SLASH_SLASH] = ACTIONS(3693),
    [anon_sym_CARET_CARET] = ACTIONS(3689),
    [anon_sym_AMP_AMP] = ACTIONS(3693),
    [anon_sym_PIPE] = ACTIONS(3693),
    [anon_sym_CARET] = ACTIONS(3693),
    [anon_sym_AMP] = ACTIONS(3693),
    [anon_sym_GT_GT2] = ACTIONS(3693),
    [anon_sym_PLUS] = ACTIONS(3693),
    [anon_sym_DASH] = ACTIONS(3693),
    [anon_sym_DOT] = ACTIONS(3693),
    [anon_sym_SLASH] = ACTIONS(3693),
    [anon_sym_x] = ACTIONS(3693),
    [anon_sym_EQ_TILDE] = ACTIONS(3689),
    [anon_sym_BANG_TILDE] = ACTIONS(3689),
    [anon_sym_EQ_EQ] = ACTIONS(3693),
    [anon_sym_BANG_EQ] = ACTIONS(3689),
    [anon_sym_eq] = ACTIONS(3693),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3689),
    [anon_sym_equ] = ACTIONS(3689),
    [anon_sym_eqr] = ACTIONS(3689),
    [anon_sym_ne] = ACTIONS(3689),
    [anon_sym_LT_EQ_GT] = ACTIONS(3689),
    [anon_sym_cmp] = ACTIONS(3689),
    [anon_sym_TILDE_TILDE] = ACTIONS(3689),
    [anon_sym_LT] = ACTIONS(3693),
    [anon_sym_LT_EQ] = ACTIONS(3693),
    [anon_sym_GT_EQ] = ACTIONS(3689),
    [anon_sym_lt] = ACTIONS(3689),
    [anon_sym_le] = ACTIONS(3689),
    [anon_sym_ge] = ACTIONS(3689),
    [anon_sym_gt] = ACTIONS(3689),
    [anon_sym_isa] = ACTIONS(3689),
    [anon_sym_PLUS_PLUS] = ACTIONS(3689),
    [anon_sym_DASH_DASH] = ACTIONS(3689),
    [anon_sym_QMARK] = ACTIONS(3689),
    [anon_sym_EQ_GT] = ACTIONS(3689),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3694)] = {
    [sym_heredoc_content] = STATE(3694),
    [anon_sym_COLON] = ACTIONS(3669),
    [anon_sym_EQ] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_COMMA] = ACTIONS(3669),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3669),
    [anon_sym_DASH_GT] = ACTIONS(3669),
    [anon_sym_STAR] = ACTIONS(3673),
    [anon_sym_PERCENT] = ACTIONS(3673),
    [anon_sym_GT] = ACTIONS(3673),
    [anon_sym_LT_LT] = ACTIONS(3673),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3669),
    [anon_sym_PLUS_EQ] = ACTIONS(3669),
    [anon_sym_DASH_EQ] = ACTIONS(3669),
    [anon_sym_DOT_EQ] = ACTIONS(3669),
    [anon_sym_STAR_EQ] = ACTIONS(3669),
    [anon_sym_SLASH_EQ] = ACTIONS(3669),
    [anon_sym_PERCENT_EQ] = ACTIONS(3669),
    [anon_sym_x_EQ] = ACTIONS(3669),
    [anon_sym_AMP_EQ] = ACTIONS(3669),
    [anon_sym_PIPE_EQ] = ACTIONS(3669),
    [anon_sym_CARET_EQ] = ACTIONS(3669),
    [anon_sym_LT_LT_EQ] = ACTIONS(3669),
    [anon_sym_GT_GT_EQ] = ACTIONS(3669),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3669),
    [anon_sym_DOT_DOT] = ACTIONS(3673),
    [anon_sym_STAR_STAR] = ACTIONS(3673),
    [anon_sym_PIPE_PIPE] = ACTIONS(3673),
    [anon_sym_SLASH_SLASH] = ACTIONS(3673),
    [anon_sym_CARET_CARET] = ACTIONS(3669),
    [anon_sym_AMP_AMP] = ACTIONS(3673),
    [anon_sym_PIPE] = ACTIONS(3673),
    [anon_sym_CARET] = ACTIONS(3673),
    [anon_sym_AMP] = ACTIONS(3673),
    [anon_sym_GT_GT2] = ACTIONS(3673),
    [anon_sym_PLUS] = ACTIONS(3673),
    [anon_sym_DASH] = ACTIONS(3673),
    [anon_sym_DOT] = ACTIONS(3673),
    [anon_sym_SLASH] = ACTIONS(3673),
    [anon_sym_x] = ACTIONS(3673),
    [anon_sym_EQ_TILDE] = ACTIONS(3669),
    [anon_sym_BANG_TILDE] = ACTIONS(3669),
    [anon_sym_EQ_EQ] = ACTIONS(3673),
    [anon_sym_BANG_EQ] = ACTIONS(3669),
    [anon_sym_eq] = ACTIONS(3673),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3669),
    [anon_sym_equ] = ACTIONS(3669),
    [anon_sym_eqr] = ACTIONS(3669),
    [anon_sym_ne] = ACTIONS(3669),
    [anon_sym_LT_EQ_GT] = ACTIONS(3669),
    [anon_sym_cmp] = ACTIONS(3669),
    [anon_sym_TILDE_TILDE] = ACTIONS(3669),
    [anon_sym_LT] = ACTIONS(3673),
    [anon_sym_LT_EQ] = ACTIONS(3673),
    [anon_sym_GT_EQ] = ACTIONS(3669),
    [anon_sym_lt] = ACTIONS(3669),
    [anon_sym_le] = ACTIONS(3669),
    [anon_sym_ge] = ACTIONS(3669),
    [anon_sym_gt] = ACTIONS(3669),
    [anon_sym_isa] = ACTIONS(3669),
    [anon_sym_PLUS_PLUS] = ACTIONS(3669),
    [anon_sym_DASH_DASH] = ACTIONS(3669),
    [anon_sym_QMARK] = ACTIONS(3669),
    [anon_sym_EQ_GT] = ACTIONS(3669),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3695)] = {
    [sym_heredoc_content] = STATE(3695),
    [anon_sym_COLON] = ACTIONS(3819),
    [anon_sym_EQ] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_COMMA] = ACTIONS(5405),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3819),
    [anon_sym_DASH_GT] = ACTIONS(3819),
    [anon_sym_STAR] = ACTIONS(3821),
    [anon_sym_PERCENT] = ACTIONS(3821),
    [anon_sym_GT] = ACTIONS(3821),
    [anon_sym_LT_LT] = ACTIONS(3821),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3819),
    [anon_sym_PLUS_EQ] = ACTIONS(3819),
    [anon_sym_DASH_EQ] = ACTIONS(3819),
    [anon_sym_DOT_EQ] = ACTIONS(3819),
    [anon_sym_STAR_EQ] = ACTIONS(3819),
    [anon_sym_SLASH_EQ] = ACTIONS(3819),
    [anon_sym_PERCENT_EQ] = ACTIONS(3819),
    [anon_sym_x_EQ] = ACTIONS(3819),
    [anon_sym_AMP_EQ] = ACTIONS(3819),
    [anon_sym_PIPE_EQ] = ACTIONS(3819),
    [anon_sym_CARET_EQ] = ACTIONS(3819),
    [anon_sym_LT_LT_EQ] = ACTIONS(3819),
    [anon_sym_GT_GT_EQ] = ACTIONS(3819),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3819),
    [anon_sym_DOT_DOT] = ACTIONS(3821),
    [anon_sym_STAR_STAR] = ACTIONS(3821),
    [anon_sym_PIPE_PIPE] = ACTIONS(3821),
    [anon_sym_SLASH_SLASH] = ACTIONS(3821),
    [anon_sym_CARET_CARET] = ACTIONS(3819),
    [anon_sym_AMP_AMP] = ACTIONS(3821),
    [anon_sym_PIPE] = ACTIONS(3821),
    [anon_sym_CARET] = ACTIONS(3821),
    [anon_sym_AMP] = ACTIONS(3821),
    [anon_sym_GT_GT2] = ACTIONS(3821),
    [anon_sym_PLUS] = ACTIONS(3821),
    [anon_sym_DASH] = ACTIONS(3821),
    [anon_sym_DOT] = ACTIONS(3821),
    [anon_sym_SLASH] = ACTIONS(3821),
    [anon_sym_x] = ACTIONS(3821),
    [anon_sym_EQ_TILDE] = ACTIONS(3819),
    [anon_sym_BANG_TILDE] = ACTIONS(3819),
    [anon_sym_EQ_EQ] = ACTIONS(3821),
    [anon_sym_BANG_EQ] = ACTIONS(3819),
    [anon_sym_eq] = ACTIONS(3821),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3819),
    [anon_sym_equ] = ACTIONS(3819),
    [anon_sym_eqr] = ACTIONS(3819),
    [anon_sym_ne] = ACTIONS(3819),
    [anon_sym_LT_EQ_GT] = ACTIONS(3819),
    [anon_sym_cmp] = ACTIONS(3819),
    [anon_sym_TILDE_TILDE] = ACTIONS(3819),
    [anon_sym_LT] = ACTIONS(3821),
    [anon_sym_LT_EQ] = ACTIONS(3821),
    [anon_sym_GT_EQ] = ACTIONS(3819),
    [anon_sym_lt] = ACTIONS(3819),
    [anon_sym_le] = ACTIONS(3819),
    [anon_sym_ge] = ACTIONS(3819),
    [anon_sym_gt] = ACTIONS(3819),
    [anon_sym_isa] = ACTIONS(3819),
    [anon_sym_PLUS_PLUS] = ACTIONS(3819),
    [anon_sym_DASH_DASH] = ACTIONS(3819),
    [anon_sym_QMARK] = ACTIONS(3819),
    [anon_sym_EQ_GT] = ACTIONS(5405),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3696)] = {
    [sym_heredoc_content] = STATE(3696),
    [anon_sym_COLON] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(4107),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3697)] = {
    [sym_heredoc_content] = STATE(3697),
    [anon_sym_COLON] = ACTIONS(3590),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(3590),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(3590),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3698)] = {
    [sym_heredoc_content] = STATE(3698),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5407),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5409),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3699)] = {
    [sym_heredoc_content] = STATE(3699),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3700)] = {
    [sym_heredoc_content] = STATE(3700),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3701)] = {
    [sym_heredoc_content] = STATE(3701),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3702)] = {
    [sym_heredoc_content] = STATE(3702),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3703)] = {
    [sym_heredoc_content] = STATE(3703),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3704)] = {
    [sym_heredoc_content] = STATE(3704),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3705)] = {
    [sym_heredoc_content] = STATE(3705),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3706)] = {
    [sym_heredoc_content] = STATE(3706),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3707)] = {
    [sym_heredoc_content] = STATE(3707),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3708)] = {
    [sym_heredoc_content] = STATE(3708),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [anon_sym_EQ_GT] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3709)] = {
    [sym_heredoc_content] = STATE(3709),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3710)] = {
    [sym_heredoc_content] = STATE(3710),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5411),
    [anon_sym_cmp] = ACTIONS(5411),
    [anon_sym_TILDE_TILDE] = ACTIONS(5411),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [anon_sym_EQ_GT] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3711)] = {
    [sym_heredoc_content] = STATE(3711),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5413),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [anon_sym_EQ_GT] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3712)] = {
    [sym_heredoc_content] = STATE(3712),
    [anon_sym_COLON] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3977),
    [anon_sym_EQ_GT] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3713)] = {
    [sym_heredoc_content] = STATE(3713),
    [anon_sym_COLON] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3997),
    [anon_sym_EQ_GT] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3714)] = {
    [sym_heredoc_content] = STATE(3714),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3695),
    [anon_sym_EQ_GT] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3715)] = {
    [sym_heredoc_content] = STATE(3715),
    [anon_sym_COLON] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3909),
    [anon_sym_EQ_GT] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3716)] = {
    [sym_heredoc_content] = STATE(3716),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4017),
    [anon_sym_EQ_GT] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3717)] = {
    [sym_heredoc_content] = STATE(3717),
    [anon_sym_COLON] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(3803),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3718)] = {
    [sym_heredoc_content] = STATE(3718),
    [anon_sym_COLON] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3845),
    [anon_sym_EQ_GT] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3719)] = {
    [sym_heredoc_content] = STATE(3719),
    [anon_sym_COLON] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3739),
    [anon_sym_EQ_GT] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3720)] = {
    [sym_heredoc_content] = STATE(3720),
    [anon_sym_COLON] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5328),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5334),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(5338),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5330),
    [anon_sym_PLUS_EQ] = ACTIONS(5330),
    [anon_sym_DASH_EQ] = ACTIONS(5330),
    [anon_sym_DOT_EQ] = ACTIONS(5330),
    [anon_sym_STAR_EQ] = ACTIONS(5330),
    [anon_sym_SLASH_EQ] = ACTIONS(5330),
    [anon_sym_PERCENT_EQ] = ACTIONS(5330),
    [anon_sym_x_EQ] = ACTIONS(5330),
    [anon_sym_AMP_EQ] = ACTIONS(5330),
    [anon_sym_PIPE_EQ] = ACTIONS(5330),
    [anon_sym_CARET_EQ] = ACTIONS(5330),
    [anon_sym_LT_LT_EQ] = ACTIONS(5330),
    [anon_sym_GT_GT_EQ] = ACTIONS(5330),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5330),
    [anon_sym_DOT_DOT] = ACTIONS(5342),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(5346),
    [anon_sym_SLASH_SLASH] = ACTIONS(5346),
    [anon_sym_CARET_CARET] = ACTIONS(5348),
    [anon_sym_AMP_AMP] = ACTIONS(5350),
    [anon_sym_PIPE] = ACTIONS(5352),
    [anon_sym_CARET] = ACTIONS(5352),
    [anon_sym_AMP] = ACTIONS(5354),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(5360),
    [anon_sym_BANG_EQ] = ACTIONS(5362),
    [anon_sym_eq] = ACTIONS(5360),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5362),
    [anon_sym_equ] = ACTIONS(5362),
    [anon_sym_eqr] = ACTIONS(5362),
    [anon_sym_ne] = ACTIONS(5362),
    [anon_sym_LT_EQ_GT] = ACTIONS(5364),
    [anon_sym_cmp] = ACTIONS(5364),
    [anon_sym_TILDE_TILDE] = ACTIONS(5364),
    [anon_sym_LT] = ACTIONS(5338),
    [anon_sym_LT_EQ] = ACTIONS(5338),
    [anon_sym_GT_EQ] = ACTIONS(5366),
    [anon_sym_lt] = ACTIONS(5366),
    [anon_sym_le] = ACTIONS(5366),
    [anon_sym_ge] = ACTIONS(5366),
    [anon_sym_gt] = ACTIONS(5366),
    [anon_sym_isa] = ACTIONS(5368),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5370),
    [anon_sym_EQ_GT] = ACTIONS(3897),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3721)] = {
    [sym_heredoc_content] = STATE(3721),
    [anon_sym_COLON] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5336),
    [anon_sym_PERCENT] = ACTIONS(5336),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5344),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5340),
    [anon_sym_PLUS] = ACTIONS(5356),
    [anon_sym_DASH] = ACTIONS(5356),
    [anon_sym_DOT] = ACTIONS(5356),
    [anon_sym_SLASH] = ACTIONS(5336),
    [anon_sym_x] = ACTIONS(5336),
    [anon_sym_EQ_TILDE] = ACTIONS(5358),
    [anon_sym_BANG_TILDE] = ACTIONS(5358),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4151),
    [anon_sym_EQ_GT] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3722)] = {
    [sym_heredoc_content] = STATE(3722),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3723)] = {
    [sym_heredoc_content] = STATE(3723),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3724)] = {
    [sym_heredoc_content] = STATE(3724),
    [anon_sym_EQ] = ACTIONS(5425),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_COMMA] = ACTIONS(4151),
    [anon_sym_RPAREN] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5429),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5427),
    [anon_sym_PLUS_EQ] = ACTIONS(5427),
    [anon_sym_DASH_EQ] = ACTIONS(5427),
    [anon_sym_DOT_EQ] = ACTIONS(5427),
    [anon_sym_STAR_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_PERCENT_EQ] = ACTIONS(5427),
    [anon_sym_x_EQ] = ACTIONS(5427),
    [anon_sym_AMP_EQ] = ACTIONS(5427),
    [anon_sym_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_CARET_EQ] = ACTIONS(5427),
    [anon_sym_LT_LT_EQ] = ACTIONS(5427),
    [anon_sym_GT_GT_EQ] = ACTIONS(5427),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5427),
    [anon_sym_DOT_DOT] = ACTIONS(5433),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5455),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3725)] = {
    [sym_heredoc_content] = STATE(3725),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3726)] = {
    [sym_heredoc_content] = STATE(3726),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5457),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5457),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3727)] = {
    [sym_heredoc_content] = STATE(3727),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3728)] = {
    [sym_heredoc_content] = STATE(3728),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3729)] = {
    [sym_heredoc_content] = STATE(3729),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3730)] = {
    [sym_heredoc_content] = STATE(3730),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3731)] = {
    [sym_heredoc_content] = STATE(3731),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5459),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5459),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3732)] = {
    [sym_heredoc_content] = STATE(3732),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_COMMA] = ACTIONS(3977),
    [anon_sym_RPAREN] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3733)] = {
    [sym_heredoc_content] = STATE(3733),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_COMMA] = ACTIONS(3997),
    [anon_sym_RPAREN] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3734)] = {
    [sym_heredoc_content] = STATE(3734),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_COMMA] = ACTIONS(3695),
    [anon_sym_RPAREN] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3735)] = {
    [sym_heredoc_content] = STATE(3735),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_COMMA] = ACTIONS(3909),
    [anon_sym_RPAREN] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3736)] = {
    [sym_heredoc_content] = STATE(3736),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_COMMA] = ACTIONS(4017),
    [anon_sym_RPAREN] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3737)] = {
    [sym_heredoc_content] = STATE(3737),
    [anon_sym_EQ] = ACTIONS(5425),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_COMMA] = ACTIONS(3803),
    [anon_sym_RPAREN] = ACTIONS(3803),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5429),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5427),
    [anon_sym_PLUS_EQ] = ACTIONS(5427),
    [anon_sym_DASH_EQ] = ACTIONS(5427),
    [anon_sym_DOT_EQ] = ACTIONS(5427),
    [anon_sym_STAR_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_PERCENT_EQ] = ACTIONS(5427),
    [anon_sym_x_EQ] = ACTIONS(5427),
    [anon_sym_AMP_EQ] = ACTIONS(5427),
    [anon_sym_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_CARET_EQ] = ACTIONS(5427),
    [anon_sym_LT_LT_EQ] = ACTIONS(5427),
    [anon_sym_GT_GT_EQ] = ACTIONS(5427),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5427),
    [anon_sym_DOT_DOT] = ACTIONS(5433),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5455),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3738)] = {
    [sym_heredoc_content] = STATE(3738),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_COMMA] = ACTIONS(3845),
    [anon_sym_RPAREN] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3739)] = {
    [sym_heredoc_content] = STATE(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_COMMA] = ACTIONS(3739),
    [anon_sym_RPAREN] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3740)] = {
    [sym_heredoc_content] = STATE(3740),
    [anon_sym_EQ] = ACTIONS(5425),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_COMMA] = ACTIONS(3897),
    [anon_sym_RPAREN] = ACTIONS(3897),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5429),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5427),
    [anon_sym_PLUS_EQ] = ACTIONS(5427),
    [anon_sym_DASH_EQ] = ACTIONS(5427),
    [anon_sym_DOT_EQ] = ACTIONS(5427),
    [anon_sym_STAR_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_PERCENT_EQ] = ACTIONS(5427),
    [anon_sym_x_EQ] = ACTIONS(5427),
    [anon_sym_AMP_EQ] = ACTIONS(5427),
    [anon_sym_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_CARET_EQ] = ACTIONS(5427),
    [anon_sym_LT_LT_EQ] = ACTIONS(5427),
    [anon_sym_GT_GT_EQ] = ACTIONS(5427),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5427),
    [anon_sym_DOT_DOT] = ACTIONS(5433),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5455),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3741)] = {
    [sym_heredoc_content] = STATE(3741),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5461),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5461),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3742)] = {
    [sym_heredoc_content] = STATE(3742),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3743)] = {
    [sym_heredoc_content] = STATE(3743),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5463),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5463),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3744)] = {
    [sym_heredoc_content] = STATE(3744),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5465),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5465),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3745)] = {
    [sym_heredoc_content] = STATE(3745),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5467),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5467),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3746)] = {
    [sym_heredoc_content] = STATE(3746),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5469),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5469),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3747)] = {
    [sym_heredoc_content] = STATE(3747),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5471),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5471),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3748)] = {
    [sym_heredoc_content] = STATE(3748),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3749)] = {
    [sym_heredoc_content] = STATE(3749),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5473),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5473),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3750)] = {
    [sym_heredoc_content] = STATE(3750),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5475),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3751)] = {
    [sym_heredoc_content] = STATE(3751),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_COMMA] = ACTIONS(3763),
    [anon_sym_RPAREN] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5477),
    [anon_sym_cmp] = ACTIONS(5477),
    [anon_sym_TILDE_TILDE] = ACTIONS(5477),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3752)] = {
    [sym_heredoc_content] = STATE(3752),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5479),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5479),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3753)] = {
    [sym_heredoc_content] = STATE(3753),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5481),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3754)] = {
    [sym_heredoc_content] = STATE(3754),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_COMMA] = ACTIONS(3985),
    [anon_sym_RPAREN] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5429),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5433),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5455),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3755)] = {
    [sym_heredoc_content] = STATE(3755),
    [anon_sym_EQ] = ACTIONS(5425),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_COMMA] = ACTIONS(4107),
    [anon_sym_RPAREN] = ACTIONS(4107),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5429),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5427),
    [anon_sym_PLUS_EQ] = ACTIONS(5427),
    [anon_sym_DASH_EQ] = ACTIONS(5427),
    [anon_sym_DOT_EQ] = ACTIONS(5427),
    [anon_sym_STAR_EQ] = ACTIONS(5427),
    [anon_sym_SLASH_EQ] = ACTIONS(5427),
    [anon_sym_PERCENT_EQ] = ACTIONS(5427),
    [anon_sym_x_EQ] = ACTIONS(5427),
    [anon_sym_AMP_EQ] = ACTIONS(5427),
    [anon_sym_PIPE_EQ] = ACTIONS(5427),
    [anon_sym_CARET_EQ] = ACTIONS(5427),
    [anon_sym_LT_LT_EQ] = ACTIONS(5427),
    [anon_sym_GT_GT_EQ] = ACTIONS(5427),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5427),
    [anon_sym_DOT_DOT] = ACTIONS(5433),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5455),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3756)] = {
    [sym_heredoc_content] = STATE(3756),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5483),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5483),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3757)] = {
    [sym_heredoc_content] = STATE(3757),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5485),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5485),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3758)] = {
    [sym_heredoc_content] = STATE(3758),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5487),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5487),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3759)] = {
    [sym_heredoc_content] = STATE(3759),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5489),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5489),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3760)] = {
    [sym_heredoc_content] = STATE(3760),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5491),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5491),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3761)] = {
    [sym_heredoc_content] = STATE(3761),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5493),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5493),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3762)] = {
    [sym_heredoc_content] = STATE(3762),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5495),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(5431),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5497),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(5435),
    [anon_sym_SLASH_SLASH] = ACTIONS(5435),
    [anon_sym_CARET_CARET] = ACTIONS(5437),
    [anon_sym_AMP_AMP] = ACTIONS(5439),
    [anon_sym_PIPE] = ACTIONS(5441),
    [anon_sym_CARET] = ACTIONS(5441),
    [anon_sym_AMP] = ACTIONS(5443),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(5445),
    [anon_sym_BANG_EQ] = ACTIONS(5447),
    [anon_sym_eq] = ACTIONS(5445),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5447),
    [anon_sym_equ] = ACTIONS(5447),
    [anon_sym_eqr] = ACTIONS(5447),
    [anon_sym_ne] = ACTIONS(5447),
    [anon_sym_LT_EQ_GT] = ACTIONS(5449),
    [anon_sym_cmp] = ACTIONS(5449),
    [anon_sym_TILDE_TILDE] = ACTIONS(5449),
    [anon_sym_LT] = ACTIONS(5431),
    [anon_sym_LT_EQ] = ACTIONS(5431),
    [anon_sym_GT_EQ] = ACTIONS(5451),
    [anon_sym_lt] = ACTIONS(5451),
    [anon_sym_le] = ACTIONS(5451),
    [anon_sym_ge] = ACTIONS(5451),
    [anon_sym_gt] = ACTIONS(5451),
    [anon_sym_isa] = ACTIONS(5453),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3763)] = {
    [sym_heredoc_content] = STATE(3763),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3764)] = {
    [sym_heredoc_content] = STATE(3764),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_COMMA] = ACTIONS(3735),
    [anon_sym_RPAREN] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5417),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5417),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3765)] = {
    [sym_heredoc_content] = STATE(3765),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_COMMA] = ACTIONS(3719),
    [anon_sym_RPAREN] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5415),
    [anon_sym_PERCENT] = ACTIONS(5415),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5419),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5421),
    [anon_sym_DASH] = ACTIONS(5421),
    [anon_sym_DOT] = ACTIONS(5421),
    [anon_sym_SLASH] = ACTIONS(5415),
    [anon_sym_x] = ACTIONS(5415),
    [anon_sym_EQ_TILDE] = ACTIONS(5423),
    [anon_sym_BANG_TILDE] = ACTIONS(5423),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3766)] = {
    [sym_heredoc_content] = STATE(3766),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5499),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5499),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3767)] = {
    [sym_heredoc_content] = STATE(3767),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5501),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5501),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3768)] = {
    [sym_heredoc_content] = STATE(3768),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5503),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5503),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3769)] = {
    [sym_heredoc_content] = STATE(3769),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5505),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5505),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3770)] = {
    [sym_heredoc_content] = STATE(3770),
    [anon_sym_EQ] = ACTIONS(5194),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_COMMA] = ACTIONS(5507),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5200),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5204),
    [anon_sym_PERCENT] = ACTIONS(5204),
    [anon_sym_GT] = ACTIONS(5206),
    [anon_sym_LT_LT] = ACTIONS(5208),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5196),
    [anon_sym_PLUS_EQ] = ACTIONS(5196),
    [anon_sym_DASH_EQ] = ACTIONS(5196),
    [anon_sym_DOT_EQ] = ACTIONS(5196),
    [anon_sym_STAR_EQ] = ACTIONS(5196),
    [anon_sym_SLASH_EQ] = ACTIONS(5196),
    [anon_sym_PERCENT_EQ] = ACTIONS(5196),
    [anon_sym_x_EQ] = ACTIONS(5196),
    [anon_sym_AMP_EQ] = ACTIONS(5196),
    [anon_sym_PIPE_EQ] = ACTIONS(5196),
    [anon_sym_CARET_EQ] = ACTIONS(5196),
    [anon_sym_LT_LT_EQ] = ACTIONS(5196),
    [anon_sym_GT_GT_EQ] = ACTIONS(5196),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5196),
    [anon_sym_DOT_DOT] = ACTIONS(5210),
    [anon_sym_STAR_STAR] = ACTIONS(5212),
    [anon_sym_PIPE_PIPE] = ACTIONS(5214),
    [anon_sym_SLASH_SLASH] = ACTIONS(5214),
    [anon_sym_CARET_CARET] = ACTIONS(5216),
    [anon_sym_AMP_AMP] = ACTIONS(5218),
    [anon_sym_PIPE] = ACTIONS(5220),
    [anon_sym_CARET] = ACTIONS(5220),
    [anon_sym_AMP] = ACTIONS(5222),
    [anon_sym_GT_GT2] = ACTIONS(5208),
    [anon_sym_PLUS] = ACTIONS(5224),
    [anon_sym_DASH] = ACTIONS(5224),
    [anon_sym_DOT] = ACTIONS(5224),
    [anon_sym_SLASH] = ACTIONS(5204),
    [anon_sym_x] = ACTIONS(5204),
    [anon_sym_EQ_TILDE] = ACTIONS(5226),
    [anon_sym_BANG_TILDE] = ACTIONS(5226),
    [anon_sym_EQ_EQ] = ACTIONS(5228),
    [anon_sym_BANG_EQ] = ACTIONS(5230),
    [anon_sym_eq] = ACTIONS(5228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5230),
    [anon_sym_equ] = ACTIONS(5230),
    [anon_sym_eqr] = ACTIONS(5230),
    [anon_sym_ne] = ACTIONS(5230),
    [anon_sym_LT_EQ_GT] = ACTIONS(5232),
    [anon_sym_cmp] = ACTIONS(5232),
    [anon_sym_TILDE_TILDE] = ACTIONS(5232),
    [anon_sym_LT] = ACTIONS(5206),
    [anon_sym_LT_EQ] = ACTIONS(5206),
    [anon_sym_GT_EQ] = ACTIONS(5234),
    [anon_sym_lt] = ACTIONS(5234),
    [anon_sym_le] = ACTIONS(5234),
    [anon_sym_ge] = ACTIONS(5234),
    [anon_sym_gt] = ACTIONS(5234),
    [anon_sym_isa] = ACTIONS(5236),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5240),
    [anon_sym_EQ_GT] = ACTIONS(5507),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3771)] = {
    [sym_heredoc_content] = STATE(3771),
    [anon_sym_COLON] = ACTIONS(5509),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3772)] = {
    [sym_heredoc_content] = STATE(3772),
    [anon_sym_COLON] = ACTIONS(3897),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3773)] = {
    [sym_heredoc_content] = STATE(3773),
    [anon_sym_COLON] = ACTIONS(5553),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3774)] = {
    [sym_heredoc_content] = STATE(3774),
    [anon_sym_COLON] = ACTIONS(5555),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3775)] = {
    [sym_heredoc_content] = STATE(3775),
    [anon_sym_COLON] = ACTIONS(3977),
    [anon_sym_EQ] = ACTIONS(3979),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3977),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3979),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3977),
    [anon_sym_PLUS_EQ] = ACTIONS(3977),
    [anon_sym_DASH_EQ] = ACTIONS(3977),
    [anon_sym_DOT_EQ] = ACTIONS(3977),
    [anon_sym_STAR_EQ] = ACTIONS(3977),
    [anon_sym_SLASH_EQ] = ACTIONS(3977),
    [anon_sym_PERCENT_EQ] = ACTIONS(3977),
    [anon_sym_x_EQ] = ACTIONS(3977),
    [anon_sym_AMP_EQ] = ACTIONS(3977),
    [anon_sym_PIPE_EQ] = ACTIONS(3977),
    [anon_sym_CARET_EQ] = ACTIONS(3977),
    [anon_sym_LT_LT_EQ] = ACTIONS(3977),
    [anon_sym_GT_GT_EQ] = ACTIONS(3977),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3977),
    [anon_sym_DOT_DOT] = ACTIONS(3979),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3979),
    [anon_sym_SLASH_SLASH] = ACTIONS(3979),
    [anon_sym_CARET_CARET] = ACTIONS(3977),
    [anon_sym_AMP_AMP] = ACTIONS(3979),
    [anon_sym_PIPE] = ACTIONS(3979),
    [anon_sym_CARET] = ACTIONS(3979),
    [anon_sym_AMP] = ACTIONS(3979),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3979),
    [anon_sym_BANG_EQ] = ACTIONS(3977),
    [anon_sym_eq] = ACTIONS(3979),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3977),
    [anon_sym_equ] = ACTIONS(3977),
    [anon_sym_eqr] = ACTIONS(3977),
    [anon_sym_ne] = ACTIONS(3977),
    [anon_sym_LT_EQ_GT] = ACTIONS(3977),
    [anon_sym_cmp] = ACTIONS(3977),
    [anon_sym_TILDE_TILDE] = ACTIONS(3977),
    [anon_sym_LT] = ACTIONS(3979),
    [anon_sym_LT_EQ] = ACTIONS(3979),
    [anon_sym_GT_EQ] = ACTIONS(3977),
    [anon_sym_lt] = ACTIONS(3977),
    [anon_sym_le] = ACTIONS(3977),
    [anon_sym_ge] = ACTIONS(3977),
    [anon_sym_gt] = ACTIONS(3977),
    [anon_sym_isa] = ACTIONS(3977),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3977),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3776)] = {
    [sym_heredoc_content] = STATE(3776),
    [anon_sym_COLON] = ACTIONS(5557),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3777)] = {
    [sym_heredoc_content] = STATE(3777),
    [anon_sym_COLON] = ACTIONS(3997),
    [anon_sym_EQ] = ACTIONS(3999),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3997),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3999),
    [anon_sym_PERCENT] = ACTIONS(3999),
    [anon_sym_GT] = ACTIONS(3999),
    [anon_sym_LT_LT] = ACTIONS(3999),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3997),
    [anon_sym_PLUS_EQ] = ACTIONS(3997),
    [anon_sym_DASH_EQ] = ACTIONS(3997),
    [anon_sym_DOT_EQ] = ACTIONS(3997),
    [anon_sym_STAR_EQ] = ACTIONS(3997),
    [anon_sym_SLASH_EQ] = ACTIONS(3997),
    [anon_sym_PERCENT_EQ] = ACTIONS(3997),
    [anon_sym_x_EQ] = ACTIONS(3997),
    [anon_sym_AMP_EQ] = ACTIONS(3997),
    [anon_sym_PIPE_EQ] = ACTIONS(3997),
    [anon_sym_CARET_EQ] = ACTIONS(3997),
    [anon_sym_LT_LT_EQ] = ACTIONS(3997),
    [anon_sym_GT_GT_EQ] = ACTIONS(3997),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3997),
    [anon_sym_DOT_DOT] = ACTIONS(3999),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3999),
    [anon_sym_SLASH_SLASH] = ACTIONS(3999),
    [anon_sym_CARET_CARET] = ACTIONS(3997),
    [anon_sym_AMP_AMP] = ACTIONS(3999),
    [anon_sym_PIPE] = ACTIONS(3999),
    [anon_sym_CARET] = ACTIONS(3999),
    [anon_sym_AMP] = ACTIONS(3999),
    [anon_sym_GT_GT2] = ACTIONS(3999),
    [anon_sym_PLUS] = ACTIONS(3999),
    [anon_sym_DASH] = ACTIONS(3999),
    [anon_sym_DOT] = ACTIONS(3999),
    [anon_sym_SLASH] = ACTIONS(3999),
    [anon_sym_x] = ACTIONS(3999),
    [anon_sym_EQ_TILDE] = ACTIONS(3997),
    [anon_sym_BANG_TILDE] = ACTIONS(3997),
    [anon_sym_EQ_EQ] = ACTIONS(3999),
    [anon_sym_BANG_EQ] = ACTIONS(3997),
    [anon_sym_eq] = ACTIONS(3999),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3997),
    [anon_sym_equ] = ACTIONS(3997),
    [anon_sym_eqr] = ACTIONS(3997),
    [anon_sym_ne] = ACTIONS(3997),
    [anon_sym_LT_EQ_GT] = ACTIONS(3997),
    [anon_sym_cmp] = ACTIONS(3997),
    [anon_sym_TILDE_TILDE] = ACTIONS(3997),
    [anon_sym_LT] = ACTIONS(3999),
    [anon_sym_LT_EQ] = ACTIONS(3999),
    [anon_sym_GT_EQ] = ACTIONS(3997),
    [anon_sym_lt] = ACTIONS(3997),
    [anon_sym_le] = ACTIONS(3997),
    [anon_sym_ge] = ACTIONS(3997),
    [anon_sym_gt] = ACTIONS(3997),
    [anon_sym_isa] = ACTIONS(3997),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3997),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3778)] = {
    [sym_heredoc_content] = STATE(3778),
    [anon_sym_COLON] = ACTIONS(5559),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3779)] = {
    [sym_heredoc_content] = STATE(3779),
    [anon_sym_COLON] = ACTIONS(5561),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3780)] = {
    [sym_heredoc_content] = STATE(3780),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3781)] = {
    [sym_heredoc_content] = STATE(3781),
    [anon_sym_COLON] = ACTIONS(5563),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3782)] = {
    [sym_heredoc_content] = STATE(3782),
    [anon_sym_COLON] = ACTIONS(3695),
    [anon_sym_EQ] = ACTIONS(3697),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3695),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3697),
    [anon_sym_PERCENT] = ACTIONS(3697),
    [anon_sym_GT] = ACTIONS(3697),
    [anon_sym_LT_LT] = ACTIONS(3697),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3695),
    [anon_sym_PLUS_EQ] = ACTIONS(3695),
    [anon_sym_DASH_EQ] = ACTIONS(3695),
    [anon_sym_DOT_EQ] = ACTIONS(3695),
    [anon_sym_STAR_EQ] = ACTIONS(3695),
    [anon_sym_SLASH_EQ] = ACTIONS(3695),
    [anon_sym_PERCENT_EQ] = ACTIONS(3695),
    [anon_sym_x_EQ] = ACTIONS(3695),
    [anon_sym_AMP_EQ] = ACTIONS(3695),
    [anon_sym_PIPE_EQ] = ACTIONS(3695),
    [anon_sym_CARET_EQ] = ACTIONS(3695),
    [anon_sym_LT_LT_EQ] = ACTIONS(3695),
    [anon_sym_GT_GT_EQ] = ACTIONS(3695),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3695),
    [anon_sym_DOT_DOT] = ACTIONS(3697),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3697),
    [anon_sym_SLASH_SLASH] = ACTIONS(3697),
    [anon_sym_CARET_CARET] = ACTIONS(3695),
    [anon_sym_AMP_AMP] = ACTIONS(3697),
    [anon_sym_PIPE] = ACTIONS(3697),
    [anon_sym_CARET] = ACTIONS(3697),
    [anon_sym_AMP] = ACTIONS(3697),
    [anon_sym_GT_GT2] = ACTIONS(3697),
    [anon_sym_PLUS] = ACTIONS(3697),
    [anon_sym_DASH] = ACTIONS(3697),
    [anon_sym_DOT] = ACTIONS(3697),
    [anon_sym_SLASH] = ACTIONS(3697),
    [anon_sym_x] = ACTIONS(3697),
    [anon_sym_EQ_TILDE] = ACTIONS(3695),
    [anon_sym_BANG_TILDE] = ACTIONS(3695),
    [anon_sym_EQ_EQ] = ACTIONS(3697),
    [anon_sym_BANG_EQ] = ACTIONS(3695),
    [anon_sym_eq] = ACTIONS(3697),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3695),
    [anon_sym_equ] = ACTIONS(3695),
    [anon_sym_eqr] = ACTIONS(3695),
    [anon_sym_ne] = ACTIONS(3695),
    [anon_sym_LT_EQ_GT] = ACTIONS(3695),
    [anon_sym_cmp] = ACTIONS(3695),
    [anon_sym_TILDE_TILDE] = ACTIONS(3695),
    [anon_sym_LT] = ACTIONS(3697),
    [anon_sym_LT_EQ] = ACTIONS(3697),
    [anon_sym_GT_EQ] = ACTIONS(3695),
    [anon_sym_lt] = ACTIONS(3695),
    [anon_sym_le] = ACTIONS(3695),
    [anon_sym_ge] = ACTIONS(3695),
    [anon_sym_gt] = ACTIONS(3695),
    [anon_sym_isa] = ACTIONS(3695),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3695),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3783)] = {
    [sym_heredoc_content] = STATE(3783),
    [anon_sym_COLON] = ACTIONS(3909),
    [anon_sym_EQ] = ACTIONS(3911),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3909),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3911),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3909),
    [anon_sym_PLUS_EQ] = ACTIONS(3909),
    [anon_sym_DASH_EQ] = ACTIONS(3909),
    [anon_sym_DOT_EQ] = ACTIONS(3909),
    [anon_sym_STAR_EQ] = ACTIONS(3909),
    [anon_sym_SLASH_EQ] = ACTIONS(3909),
    [anon_sym_PERCENT_EQ] = ACTIONS(3909),
    [anon_sym_x_EQ] = ACTIONS(3909),
    [anon_sym_AMP_EQ] = ACTIONS(3909),
    [anon_sym_PIPE_EQ] = ACTIONS(3909),
    [anon_sym_CARET_EQ] = ACTIONS(3909),
    [anon_sym_LT_LT_EQ] = ACTIONS(3909),
    [anon_sym_GT_GT_EQ] = ACTIONS(3909),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3909),
    [anon_sym_DOT_DOT] = ACTIONS(3911),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3911),
    [anon_sym_SLASH_SLASH] = ACTIONS(3911),
    [anon_sym_CARET_CARET] = ACTIONS(3909),
    [anon_sym_AMP_AMP] = ACTIONS(3911),
    [anon_sym_PIPE] = ACTIONS(3911),
    [anon_sym_CARET] = ACTIONS(3911),
    [anon_sym_AMP] = ACTIONS(3911),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3911),
    [anon_sym_BANG_EQ] = ACTIONS(3909),
    [anon_sym_eq] = ACTIONS(3911),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3909),
    [anon_sym_equ] = ACTIONS(3909),
    [anon_sym_eqr] = ACTIONS(3909),
    [anon_sym_ne] = ACTIONS(3909),
    [anon_sym_LT_EQ_GT] = ACTIONS(3909),
    [anon_sym_cmp] = ACTIONS(3909),
    [anon_sym_TILDE_TILDE] = ACTIONS(3909),
    [anon_sym_LT] = ACTIONS(3911),
    [anon_sym_LT_EQ] = ACTIONS(3911),
    [anon_sym_GT_EQ] = ACTIONS(3909),
    [anon_sym_lt] = ACTIONS(3909),
    [anon_sym_le] = ACTIONS(3909),
    [anon_sym_ge] = ACTIONS(3909),
    [anon_sym_gt] = ACTIONS(3909),
    [anon_sym_isa] = ACTIONS(3909),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3909),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3784)] = {
    [sym_heredoc_content] = STATE(3784),
    [anon_sym_COLON] = ACTIONS(5565),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3785)] = {
    [sym_heredoc_content] = STATE(3785),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3786)] = {
    [sym_heredoc_content] = STATE(3786),
    [anon_sym_COLON] = ACTIONS(4017),
    [anon_sym_EQ] = ACTIONS(4019),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4017),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(4019),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4017),
    [anon_sym_PLUS_EQ] = ACTIONS(4017),
    [anon_sym_DASH_EQ] = ACTIONS(4017),
    [anon_sym_DOT_EQ] = ACTIONS(4017),
    [anon_sym_STAR_EQ] = ACTIONS(4017),
    [anon_sym_SLASH_EQ] = ACTIONS(4017),
    [anon_sym_PERCENT_EQ] = ACTIONS(4017),
    [anon_sym_x_EQ] = ACTIONS(4017),
    [anon_sym_AMP_EQ] = ACTIONS(4017),
    [anon_sym_PIPE_EQ] = ACTIONS(4017),
    [anon_sym_CARET_EQ] = ACTIONS(4017),
    [anon_sym_LT_LT_EQ] = ACTIONS(4017),
    [anon_sym_GT_GT_EQ] = ACTIONS(4017),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4017),
    [anon_sym_DOT_DOT] = ACTIONS(4019),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(4019),
    [anon_sym_SLASH_SLASH] = ACTIONS(4019),
    [anon_sym_CARET_CARET] = ACTIONS(4017),
    [anon_sym_AMP_AMP] = ACTIONS(4019),
    [anon_sym_PIPE] = ACTIONS(4019),
    [anon_sym_CARET] = ACTIONS(4019),
    [anon_sym_AMP] = ACTIONS(4019),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(4019),
    [anon_sym_BANG_EQ] = ACTIONS(4017),
    [anon_sym_eq] = ACTIONS(4019),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4017),
    [anon_sym_equ] = ACTIONS(4017),
    [anon_sym_eqr] = ACTIONS(4017),
    [anon_sym_ne] = ACTIONS(4017),
    [anon_sym_LT_EQ_GT] = ACTIONS(4017),
    [anon_sym_cmp] = ACTIONS(4017),
    [anon_sym_TILDE_TILDE] = ACTIONS(4017),
    [anon_sym_LT] = ACTIONS(4019),
    [anon_sym_LT_EQ] = ACTIONS(4019),
    [anon_sym_GT_EQ] = ACTIONS(4017),
    [anon_sym_lt] = ACTIONS(4017),
    [anon_sym_le] = ACTIONS(4017),
    [anon_sym_ge] = ACTIONS(4017),
    [anon_sym_gt] = ACTIONS(4017),
    [anon_sym_isa] = ACTIONS(4017),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4017),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3787)] = {
    [sym_heredoc_content] = STATE(3787),
    [anon_sym_COLON] = ACTIONS(3803),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3788)] = {
    [sym_heredoc_content] = STATE(3788),
    [anon_sym_COLON] = ACTIONS(5567),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3789)] = {
    [sym_heredoc_content] = STATE(3789),
    [anon_sym_COLON] = ACTIONS(3845),
    [anon_sym_EQ] = ACTIONS(3847),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3845),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3847),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3845),
    [anon_sym_PLUS_EQ] = ACTIONS(3845),
    [anon_sym_DASH_EQ] = ACTIONS(3845),
    [anon_sym_DOT_EQ] = ACTIONS(3845),
    [anon_sym_STAR_EQ] = ACTIONS(3845),
    [anon_sym_SLASH_EQ] = ACTIONS(3845),
    [anon_sym_PERCENT_EQ] = ACTIONS(3845),
    [anon_sym_x_EQ] = ACTIONS(3845),
    [anon_sym_AMP_EQ] = ACTIONS(3845),
    [anon_sym_PIPE_EQ] = ACTIONS(3845),
    [anon_sym_CARET_EQ] = ACTIONS(3845),
    [anon_sym_LT_LT_EQ] = ACTIONS(3845),
    [anon_sym_GT_GT_EQ] = ACTIONS(3845),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3845),
    [anon_sym_DOT_DOT] = ACTIONS(3847),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3847),
    [anon_sym_SLASH_SLASH] = ACTIONS(3847),
    [anon_sym_CARET_CARET] = ACTIONS(3845),
    [anon_sym_AMP_AMP] = ACTIONS(3847),
    [anon_sym_PIPE] = ACTIONS(3847),
    [anon_sym_CARET] = ACTIONS(3847),
    [anon_sym_AMP] = ACTIONS(3847),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3847),
    [anon_sym_BANG_EQ] = ACTIONS(3845),
    [anon_sym_eq] = ACTIONS(3847),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3845),
    [anon_sym_equ] = ACTIONS(3845),
    [anon_sym_eqr] = ACTIONS(3845),
    [anon_sym_ne] = ACTIONS(3845),
    [anon_sym_LT_EQ_GT] = ACTIONS(3845),
    [anon_sym_cmp] = ACTIONS(3845),
    [anon_sym_TILDE_TILDE] = ACTIONS(3845),
    [anon_sym_LT] = ACTIONS(3847),
    [anon_sym_LT_EQ] = ACTIONS(3847),
    [anon_sym_GT_EQ] = ACTIONS(3845),
    [anon_sym_lt] = ACTIONS(3845),
    [anon_sym_le] = ACTIONS(3845),
    [anon_sym_ge] = ACTIONS(3845),
    [anon_sym_gt] = ACTIONS(3845),
    [anon_sym_isa] = ACTIONS(3845),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3845),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3790)] = {
    [sym_heredoc_content] = STATE(3790),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3791)] = {
    [sym_heredoc_content] = STATE(3791),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3765),
    [anon_sym_BANG_EQ] = ACTIONS(3763),
    [anon_sym_eq] = ACTIONS(3765),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3763),
    [anon_sym_equ] = ACTIONS(3763),
    [anon_sym_eqr] = ACTIONS(3763),
    [anon_sym_ne] = ACTIONS(3763),
    [anon_sym_LT_EQ_GT] = ACTIONS(3763),
    [anon_sym_cmp] = ACTIONS(3763),
    [anon_sym_TILDE_TILDE] = ACTIONS(3763),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3792)] = {
    [sym_heredoc_content] = STATE(3792),
    [anon_sym_COLON] = ACTIONS(3763),
    [anon_sym_EQ] = ACTIONS(3765),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3763),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3763),
    [anon_sym_PLUS_EQ] = ACTIONS(3763),
    [anon_sym_DASH_EQ] = ACTIONS(3763),
    [anon_sym_DOT_EQ] = ACTIONS(3763),
    [anon_sym_STAR_EQ] = ACTIONS(3763),
    [anon_sym_SLASH_EQ] = ACTIONS(3763),
    [anon_sym_PERCENT_EQ] = ACTIONS(3763),
    [anon_sym_x_EQ] = ACTIONS(3763),
    [anon_sym_AMP_EQ] = ACTIONS(3763),
    [anon_sym_PIPE_EQ] = ACTIONS(3763),
    [anon_sym_CARET_EQ] = ACTIONS(3763),
    [anon_sym_LT_LT_EQ] = ACTIONS(3763),
    [anon_sym_GT_GT_EQ] = ACTIONS(3763),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3763),
    [anon_sym_DOT_DOT] = ACTIONS(3765),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3765),
    [anon_sym_SLASH_SLASH] = ACTIONS(3765),
    [anon_sym_CARET_CARET] = ACTIONS(3763),
    [anon_sym_AMP_AMP] = ACTIONS(3765),
    [anon_sym_PIPE] = ACTIONS(3765),
    [anon_sym_CARET] = ACTIONS(3765),
    [anon_sym_AMP] = ACTIONS(3765),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5569),
    [anon_sym_cmp] = ACTIONS(5569),
    [anon_sym_TILDE_TILDE] = ACTIONS(5569),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3763),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3793)] = {
    [sym_heredoc_content] = STATE(3793),
    [anon_sym_COLON] = ACTIONS(5571),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3794)] = {
    [sym_heredoc_content] = STATE(3794),
    [anon_sym_COLON] = ACTIONS(5573),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3795)] = {
    [sym_heredoc_content] = STATE(3795),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5575),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3796)] = {
    [sym_heredoc_content] = STATE(3796),
    [anon_sym_COLON] = ACTIONS(3739),
    [anon_sym_EQ] = ACTIONS(3741),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3739),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3741),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3739),
    [anon_sym_PLUS_EQ] = ACTIONS(3739),
    [anon_sym_DASH_EQ] = ACTIONS(3739),
    [anon_sym_DOT_EQ] = ACTIONS(3739),
    [anon_sym_STAR_EQ] = ACTIONS(3739),
    [anon_sym_SLASH_EQ] = ACTIONS(3739),
    [anon_sym_PERCENT_EQ] = ACTIONS(3739),
    [anon_sym_x_EQ] = ACTIONS(3739),
    [anon_sym_AMP_EQ] = ACTIONS(3739),
    [anon_sym_PIPE_EQ] = ACTIONS(3739),
    [anon_sym_CARET_EQ] = ACTIONS(3739),
    [anon_sym_LT_LT_EQ] = ACTIONS(3739),
    [anon_sym_GT_GT_EQ] = ACTIONS(3739),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3739),
    [anon_sym_DOT_DOT] = ACTIONS(3741),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3741),
    [anon_sym_SLASH_SLASH] = ACTIONS(3741),
    [anon_sym_CARET_CARET] = ACTIONS(3739),
    [anon_sym_AMP_AMP] = ACTIONS(3741),
    [anon_sym_PIPE] = ACTIONS(3741),
    [anon_sym_CARET] = ACTIONS(3741),
    [anon_sym_AMP] = ACTIONS(3741),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3741),
    [anon_sym_BANG_EQ] = ACTIONS(3739),
    [anon_sym_eq] = ACTIONS(3741),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3739),
    [anon_sym_equ] = ACTIONS(3739),
    [anon_sym_eqr] = ACTIONS(3739),
    [anon_sym_ne] = ACTIONS(3739),
    [anon_sym_LT_EQ_GT] = ACTIONS(3739),
    [anon_sym_cmp] = ACTIONS(3739),
    [anon_sym_TILDE_TILDE] = ACTIONS(3739),
    [anon_sym_LT] = ACTIONS(3741),
    [anon_sym_LT_EQ] = ACTIONS(3741),
    [anon_sym_GT_EQ] = ACTIONS(3739),
    [anon_sym_lt] = ACTIONS(3739),
    [anon_sym_le] = ACTIONS(3739),
    [anon_sym_ge] = ACTIONS(3739),
    [anon_sym_gt] = ACTIONS(3739),
    [anon_sym_isa] = ACTIONS(3739),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3739),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3797)] = {
    [sym_heredoc_content] = STATE(3797),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3798)] = {
    [sym_heredoc_content] = STATE(3798),
    [anon_sym_COLON] = ACTIONS(3985),
    [anon_sym_EQ] = ACTIONS(3987),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3985),
    [anon_sym_PLUS_EQ] = ACTIONS(3985),
    [anon_sym_DASH_EQ] = ACTIONS(3985),
    [anon_sym_DOT_EQ] = ACTIONS(3985),
    [anon_sym_STAR_EQ] = ACTIONS(3985),
    [anon_sym_SLASH_EQ] = ACTIONS(3985),
    [anon_sym_PERCENT_EQ] = ACTIONS(3985),
    [anon_sym_x_EQ] = ACTIONS(3985),
    [anon_sym_AMP_EQ] = ACTIONS(3985),
    [anon_sym_PIPE_EQ] = ACTIONS(3985),
    [anon_sym_CARET_EQ] = ACTIONS(3985),
    [anon_sym_LT_LT_EQ] = ACTIONS(3985),
    [anon_sym_GT_GT_EQ] = ACTIONS(3985),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3985),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3799)] = {
    [sym_heredoc_content] = STATE(3799),
    [anon_sym_COLON] = ACTIONS(4107),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3800)] = {
    [sym_heredoc_content] = STATE(3800),
    [anon_sym_COLON] = ACTIONS(4151),
    [anon_sym_EQ] = ACTIONS(4153),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(4151),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(4153),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(4151),
    [anon_sym_PLUS_EQ] = ACTIONS(4151),
    [anon_sym_DASH_EQ] = ACTIONS(4151),
    [anon_sym_DOT_EQ] = ACTIONS(4151),
    [anon_sym_STAR_EQ] = ACTIONS(4151),
    [anon_sym_SLASH_EQ] = ACTIONS(4151),
    [anon_sym_PERCENT_EQ] = ACTIONS(4151),
    [anon_sym_x_EQ] = ACTIONS(4151),
    [anon_sym_AMP_EQ] = ACTIONS(4151),
    [anon_sym_PIPE_EQ] = ACTIONS(4151),
    [anon_sym_CARET_EQ] = ACTIONS(4151),
    [anon_sym_LT_LT_EQ] = ACTIONS(4151),
    [anon_sym_GT_GT_EQ] = ACTIONS(4151),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(4151),
    [anon_sym_DOT_DOT] = ACTIONS(4153),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(4153),
    [anon_sym_SLASH_SLASH] = ACTIONS(4153),
    [anon_sym_CARET_CARET] = ACTIONS(4151),
    [anon_sym_AMP_AMP] = ACTIONS(4153),
    [anon_sym_PIPE] = ACTIONS(4153),
    [anon_sym_CARET] = ACTIONS(4153),
    [anon_sym_AMP] = ACTIONS(4153),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(4153),
    [anon_sym_BANG_EQ] = ACTIONS(4151),
    [anon_sym_eq] = ACTIONS(4153),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(4151),
    [anon_sym_equ] = ACTIONS(4151),
    [anon_sym_eqr] = ACTIONS(4151),
    [anon_sym_ne] = ACTIONS(4151),
    [anon_sym_LT_EQ_GT] = ACTIONS(4151),
    [anon_sym_cmp] = ACTIONS(4151),
    [anon_sym_TILDE_TILDE] = ACTIONS(4151),
    [anon_sym_LT] = ACTIONS(4153),
    [anon_sym_LT_EQ] = ACTIONS(4153),
    [anon_sym_GT_EQ] = ACTIONS(4151),
    [anon_sym_lt] = ACTIONS(4151),
    [anon_sym_le] = ACTIONS(4151),
    [anon_sym_ge] = ACTIONS(4151),
    [anon_sym_gt] = ACTIONS(4151),
    [anon_sym_isa] = ACTIONS(4151),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(4151),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3801)] = {
    [sym_heredoc_content] = STATE(3801),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5577),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(5579),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3802)] = {
    [sym_heredoc_content] = STATE(3802),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3803)] = {
    [sym_heredoc_content] = STATE(3803),
    [anon_sym_COLON] = ACTIONS(5581),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3804)] = {
    [sym_heredoc_content] = STATE(3804),
    [anon_sym_COLON] = ACTIONS(3735),
    [anon_sym_EQ] = ACTIONS(3737),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3735),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3737),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3735),
    [anon_sym_PLUS_EQ] = ACTIONS(3735),
    [anon_sym_DASH_EQ] = ACTIONS(3735),
    [anon_sym_DOT_EQ] = ACTIONS(3735),
    [anon_sym_STAR_EQ] = ACTIONS(3735),
    [anon_sym_SLASH_EQ] = ACTIONS(3735),
    [anon_sym_PERCENT_EQ] = ACTIONS(3735),
    [anon_sym_x_EQ] = ACTIONS(3735),
    [anon_sym_AMP_EQ] = ACTIONS(3735),
    [anon_sym_PIPE_EQ] = ACTIONS(3735),
    [anon_sym_CARET_EQ] = ACTIONS(3735),
    [anon_sym_LT_LT_EQ] = ACTIONS(3735),
    [anon_sym_GT_GT_EQ] = ACTIONS(3735),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3735),
    [anon_sym_DOT_DOT] = ACTIONS(3737),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3737),
    [anon_sym_SLASH_SLASH] = ACTIONS(3737),
    [anon_sym_CARET_CARET] = ACTIONS(3735),
    [anon_sym_AMP_AMP] = ACTIONS(3737),
    [anon_sym_PIPE] = ACTIONS(3737),
    [anon_sym_CARET] = ACTIONS(3737),
    [anon_sym_AMP] = ACTIONS(3737),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3737),
    [anon_sym_BANG_EQ] = ACTIONS(3735),
    [anon_sym_eq] = ACTIONS(3737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3735),
    [anon_sym_equ] = ACTIONS(3735),
    [anon_sym_eqr] = ACTIONS(3735),
    [anon_sym_ne] = ACTIONS(3735),
    [anon_sym_LT_EQ_GT] = ACTIONS(3735),
    [anon_sym_cmp] = ACTIONS(3735),
    [anon_sym_TILDE_TILDE] = ACTIONS(3735),
    [anon_sym_LT] = ACTIONS(3737),
    [anon_sym_LT_EQ] = ACTIONS(3737),
    [anon_sym_GT_EQ] = ACTIONS(3735),
    [anon_sym_lt] = ACTIONS(3735),
    [anon_sym_le] = ACTIONS(3735),
    [anon_sym_ge] = ACTIONS(3735),
    [anon_sym_gt] = ACTIONS(3735),
    [anon_sym_isa] = ACTIONS(3735),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3735),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3805)] = {
    [sym_heredoc_content] = STATE(3805),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3806)] = {
    [sym_heredoc_content] = STATE(3806),
    [anon_sym_COLON] = ACTIONS(5583),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3807)] = {
    [sym_heredoc_content] = STATE(3807),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3808)] = {
    [sym_heredoc_content] = STATE(3808),
    [anon_sym_COLON] = ACTIONS(5585),
    [anon_sym_EQ] = ACTIONS(5511),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(5515),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(5513),
    [anon_sym_PLUS_EQ] = ACTIONS(5513),
    [anon_sym_DASH_EQ] = ACTIONS(5513),
    [anon_sym_DOT_EQ] = ACTIONS(5513),
    [anon_sym_STAR_EQ] = ACTIONS(5513),
    [anon_sym_SLASH_EQ] = ACTIONS(5513),
    [anon_sym_PERCENT_EQ] = ACTIONS(5513),
    [anon_sym_x_EQ] = ACTIONS(5513),
    [anon_sym_AMP_EQ] = ACTIONS(5513),
    [anon_sym_PIPE_EQ] = ACTIONS(5513),
    [anon_sym_CARET_EQ] = ACTIONS(5513),
    [anon_sym_LT_LT_EQ] = ACTIONS(5513),
    [anon_sym_GT_GT_EQ] = ACTIONS(5513),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(5513),
    [anon_sym_DOT_DOT] = ACTIONS(5523),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(5527),
    [anon_sym_SLASH_SLASH] = ACTIONS(5527),
    [anon_sym_CARET_CARET] = ACTIONS(5529),
    [anon_sym_AMP_AMP] = ACTIONS(5531),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(5551),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3809)] = {
    [sym_heredoc_content] = STATE(3809),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(5517),
    [anon_sym_PERCENT] = ACTIONS(5517),
    [anon_sym_GT] = ACTIONS(5519),
    [anon_sym_LT_LT] = ACTIONS(5521),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(5533),
    [anon_sym_CARET] = ACTIONS(5533),
    [anon_sym_AMP] = ACTIONS(5535),
    [anon_sym_GT_GT2] = ACTIONS(5521),
    [anon_sym_PLUS] = ACTIONS(5537),
    [anon_sym_DASH] = ACTIONS(5537),
    [anon_sym_DOT] = ACTIONS(5537),
    [anon_sym_SLASH] = ACTIONS(5517),
    [anon_sym_x] = ACTIONS(5517),
    [anon_sym_EQ_TILDE] = ACTIONS(5539),
    [anon_sym_BANG_TILDE] = ACTIONS(5539),
    [anon_sym_EQ_EQ] = ACTIONS(5541),
    [anon_sym_BANG_EQ] = ACTIONS(5543),
    [anon_sym_eq] = ACTIONS(5541),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(5543),
    [anon_sym_equ] = ACTIONS(5543),
    [anon_sym_eqr] = ACTIONS(5543),
    [anon_sym_ne] = ACTIONS(5543),
    [anon_sym_LT_EQ_GT] = ACTIONS(5545),
    [anon_sym_cmp] = ACTIONS(5545),
    [anon_sym_TILDE_TILDE] = ACTIONS(5545),
    [anon_sym_LT] = ACTIONS(5519),
    [anon_sym_LT_EQ] = ACTIONS(5519),
    [anon_sym_GT_EQ] = ACTIONS(5547),
    [anon_sym_lt] = ACTIONS(5547),
    [anon_sym_le] = ACTIONS(5547),
    [anon_sym_ge] = ACTIONS(5547),
    [anon_sym_gt] = ACTIONS(5547),
    [anon_sym_isa] = ACTIONS(5549),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
  [STATE(3810)] = {
    [sym_heredoc_content] = STATE(3810),
    [anon_sym_COLON] = ACTIONS(3719),
    [anon_sym_EQ] = ACTIONS(3721),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(3719),
    [anon_sym_DASH_GT] = ACTIONS(5202),
    [anon_sym_STAR] = ACTIONS(3721),
    [anon_sym_PERCENT] = ACTIONS(3721),
    [anon_sym_GT] = ACTIONS(3721),
    [anon_sym_LT_LT] = ACTIONS(3721),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(3719),
    [anon_sym_PLUS_EQ] = ACTIONS(3719),
    [anon_sym_DASH_EQ] = ACTIONS(3719),
    [anon_sym_DOT_EQ] = ACTIONS(3719),
    [anon_sym_STAR_EQ] = ACTIONS(3719),
    [anon_sym_SLASH_EQ] = ACTIONS(3719),
    [anon_sym_PERCENT_EQ] = ACTIONS(3719),
    [anon_sym_x_EQ] = ACTIONS(3719),
    [anon_sym_AMP_EQ] = ACTIONS(3719),
    [anon_sym_PIPE_EQ] = ACTIONS(3719),
    [anon_sym_CARET_EQ] = ACTIONS(3719),
    [anon_sym_LT_LT_EQ] = ACTIONS(3719),
    [anon_sym_GT_GT_EQ] = ACTIONS(3719),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(3719),
    [anon_sym_DOT_DOT] = ACTIONS(3721),
    [anon_sym_STAR_STAR] = ACTIONS(5525),
    [anon_sym_PIPE_PIPE] = ACTIONS(3721),
    [anon_sym_SLASH_SLASH] = ACTIONS(3721),
    [anon_sym_CARET_CARET] = ACTIONS(3719),
    [anon_sym_AMP_AMP] = ACTIONS(3721),
    [anon_sym_PIPE] = ACTIONS(3721),
    [anon_sym_CARET] = ACTIONS(3721),
    [anon_sym_AMP] = ACTIONS(3721),
    [anon_sym_GT_GT2] = ACTIONS(3721),
    [anon_sym_PLUS] = ACTIONS(3721),
    [anon_sym_DASH] = ACTIONS(3721),
    [anon_sym_DOT] = ACTIONS(3721),
    [anon_sym_SLASH] = ACTIONS(3721),
    [anon_sym_x] = ACTIONS(3721),
    [anon_sym_EQ_TILDE] = ACTIONS(3719),
    [anon_sym_BANG_TILDE] = ACTIONS(3719),
    [anon_sym_EQ_EQ] = ACTIONS(3721),
    [anon_sym_BANG_EQ] = ACTIONS(3719),
    [anon_sym_eq] = ACTIONS(3721),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3719),
    [anon_sym_equ] = ACTIONS(3719),
    [anon_sym_eqr] = ACTIONS(3719),
    [anon_sym_ne] = ACTIONS(3719),
    [anon_sym_LT_EQ_GT] = ACTIONS(3719),
    [anon_sym_cmp] = ACTIONS(3719),
    [anon_sym_TILDE_TILDE] = ACTIONS(3719),
    [anon_sym_LT] = ACTIONS(3721),
    [anon_sym_LT_EQ] = ACTIONS(3721),
    [anon_sym_GT_EQ] = ACTIONS(3719),
    [anon_sym_lt] = ACTIONS(3719),
    [anon_sym_le] = ACTIONS(3719),
    [anon_sym_ge] = ACTIONS(3719),
    [anon_sym_gt] = ACTIONS(3719),
    [anon_sym_isa] = ACTIONS(3719),
    [anon_sym_PLUS_PLUS] = ACTIONS(5238),
    [anon_sym_DASH_DASH] = ACTIONS(5238),
    [anon_sym_QMARK] = ACTIONS(3719),
    [sym_comment] = ACTIONS(5),
    [sym_pod] = ACTIONS(5),
    [sym__heredoc_start] = ACTIONS(7),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3811), 1,
      sym_heredoc_content,
    STATE(3819), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3955), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [73] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3812), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3954), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [146] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3813), 1,
      sym_heredoc_content,
    STATE(3816), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3893), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [219] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3812), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3814), 1,
      sym_heredoc_content,
    STATE(3884), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [292] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3815), 1,
      sym_heredoc_content,
    STATE(3826), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3932), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [365] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3816), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3899), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [438] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3817), 1,
      sym_heredoc_content,
    STATE(3818), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3916), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [511] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3818), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3919), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [584] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3819), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3960), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [657] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3820), 1,
      sym_heredoc_content,
    STATE(3822), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3971), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [730] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3821), 1,
      sym_heredoc_content,
    STATE(3823), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3979), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [803] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3822), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3975), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [876] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3823), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3929), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [949] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3824), 1,
      sym_heredoc_content,
    STATE(3825), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3898), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1022] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3825), 1,
      sym_heredoc_content,
    STATE(3875), 1,
      sym__quotelike_middle,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1095] = 21,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5589), 1,
      anon_sym_DOLLAR,
    ACTIONS(5591), 1,
      anon_sym_AT,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(3826), 1,
      sym_heredoc_content,
    STATE(3877), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3949), 1,
      sym__quotelike_middle,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5587), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5595), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1168] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1780), 1,
      sym__quotelike_end,
    STATE(3827), 1,
      sym_heredoc_content,
    STATE(3835), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1238] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5607), 1,
      sym__quotelike_end_quote,
    STATE(2213), 1,
      sym__quotelike_end,
    STATE(3828), 1,
      sym_heredoc_content,
    STATE(3845), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1308] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1811), 1,
      sym__quotelike_end,
    STATE(3829), 1,
      sym_heredoc_content,
    STATE(3843), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1378] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2216), 1,
      sym__quotelike_end,
    STATE(3830), 1,
      sym_heredoc_content,
    STATE(3849), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1448] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3352), 1,
      sym__quotelike_end,
    STATE(3831), 1,
      sym_heredoc_content,
    STATE(3834), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1518] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5613), 1,
      sym__quotelike_end_quote,
    STATE(3357), 1,
      sym__quotelike_end,
    STATE(3832), 1,
      sym_heredoc_content,
    STATE(3836), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1588] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3358), 1,
      sym__quotelike_end,
    STATE(3833), 1,
      sym_heredoc_content,
    STATE(3837), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1658] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3364), 1,
      sym__quotelike_end,
    STATE(3834), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1728] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1822), 1,
      sym__quotelike_end,
    STATE(3835), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1798] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5613), 1,
      sym__quotelike_end_quote,
    STATE(3368), 1,
      sym__quotelike_end,
    STATE(3836), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1868] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3369), 1,
      sym__quotelike_end,
    STATE(3837), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [1938] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2257), 1,
      sym__quotelike_end,
    STATE(3838), 1,
      sym_heredoc_content,
    STATE(3841), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2008] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2381), 1,
      sym__quotelike_end,
    STATE(3839), 1,
      sym_heredoc_content,
    STATE(3846), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2078] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5617), 1,
      sym__quotelike_end_quote,
    STATE(2278), 1,
      sym__quotelike_end,
    STATE(3840), 1,
      sym_heredoc_content,
    STATE(3874), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2148] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2211), 1,
      sym__quotelike_end,
    STATE(3841), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2218] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2319), 1,
      sym__quotelike_end,
    STATE(3842), 1,
      sym_heredoc_content,
    STATE(3848), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2288] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1788), 1,
      sym__quotelike_end,
    STATE(3843), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2358] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5619), 1,
      sym__quotelike_end_quote,
    STATE(1821), 1,
      sym__quotelike_end,
    STATE(3844), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2428] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5607), 1,
      sym__quotelike_end_quote,
    STATE(2227), 1,
      sym__quotelike_end,
    STATE(3845), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2498] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2425), 1,
      sym__quotelike_end,
    STATE(3846), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2568] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3245), 1,
      sym__quotelike_end,
    STATE(3847), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2638] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2437), 1,
      sym__quotelike_end,
    STATE(3848), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2708] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2193), 1,
      sym__quotelike_end,
    STATE(3849), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2778] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2527), 1,
      sym__quotelike_end,
    STATE(3850), 1,
      sym_heredoc_content,
    STATE(3853), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2848] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5625), 1,
      sym__quotelike_end_quote,
    STATE(2269), 1,
      sym__quotelike_end,
    STATE(3851), 1,
      sym_heredoc_content,
    STATE(3854), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2918] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2271), 1,
      sym__quotelike_end,
    STATE(3852), 1,
      sym_heredoc_content,
    STATE(3855), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [2988] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2279), 1,
      sym__quotelike_end,
    STATE(3853), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3058] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5625), 1,
      sym__quotelike_end_quote,
    STATE(2289), 1,
      sym__quotelike_end,
    STATE(3854), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3128] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2290), 1,
      sym__quotelike_end,
    STATE(3855), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3198] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3287), 1,
      sym__quotelike_end,
    STATE(3856), 1,
      sym_heredoc_content,
    STATE(3872), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3268] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2478), 1,
      sym__quotelike_end,
    STATE(3857), 1,
      sym_heredoc_content,
    STATE(3868), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3338] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5619), 1,
      sym__quotelike_end_quote,
    STATE(1806), 1,
      sym__quotelike_end,
    STATE(3844), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3858), 1,
      sym_heredoc_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3408] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5629), 1,
      sym__quotelike_end_quote,
    STATE(2501), 1,
      sym__quotelike_end,
    STATE(3859), 1,
      sym_heredoc_content,
    STATE(3869), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3478] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2503), 1,
      sym__quotelike_end,
    STATE(3860), 1,
      sym_heredoc_content,
    STATE(3870), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3548] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2284), 1,
      sym__quotelike_end,
    STATE(3861), 1,
      sym_heredoc_content,
    STATE(3865), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3618] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5633), 1,
      sym__quotelike_end_quote,
    STATE(3291), 1,
      sym__quotelike_end,
    STATE(3862), 1,
      sym_heredoc_content,
    STATE(3873), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3688] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5635), 1,
      sym__quotelike_end_quote,
    STATE(2307), 1,
      sym__quotelike_end,
    STATE(3863), 1,
      sym_heredoc_content,
    STATE(3866), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3758] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2308), 1,
      sym__quotelike_end,
    STATE(3864), 1,
      sym_heredoc_content,
    STATE(3867), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3828] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2314), 1,
      sym__quotelike_end,
    STATE(3865), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3898] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5635), 1,
      sym__quotelike_end_quote,
    STATE(2333), 1,
      sym__quotelike_end,
    STATE(3866), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [3968] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2336), 1,
      sym__quotelike_end,
    STATE(3867), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4038] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2538), 1,
      sym__quotelike_end,
    STATE(3868), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4108] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5629), 1,
      sym__quotelike_end_quote,
    STATE(2560), 1,
      sym__quotelike_end,
    STATE(3869), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4178] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2561), 1,
      sym__quotelike_end,
    STATE(3870), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4248] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3292), 1,
      sym__quotelike_end,
    STATE(3847), 1,
      aux_sym__interpolated_regexp_content,
    STATE(3871), 1,
      sym_heredoc_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4318] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3241), 1,
      sym__quotelike_end,
    STATE(3872), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4388] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5633), 1,
      sym__quotelike_end_quote,
    STATE(3244), 1,
      sym__quotelike_end,
    STATE(3873), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4458] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5599), 1,
      anon_sym_DOLLAR,
    ACTIONS(5601), 1,
      anon_sym_AT,
    ACTIONS(5617), 1,
      sym__quotelike_end_quote,
    STATE(2436), 1,
      sym__quotelike_end,
    STATE(3874), 1,
      sym_heredoc_content,
    STATE(3894), 1,
      aux_sym__interpolated_regexp_content,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5597), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5605), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4528] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2242), 1,
      sym__quotelike_end,
    STATE(3875), 1,
      sym_heredoc_content,
    STATE(3922), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [4597] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2348), 1,
      sym__quotelike_end,
    STATE(3876), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [4666] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5652), 1,
      anon_sym_DOLLAR,
    ACTIONS(5655), 1,
      anon_sym_AT,
    ACTIONS(5658), 1,
      sym__quotelike_middle_close_quote,
    STATE(4029), 1,
      sym_scalar,
    STATE(4156), 1,
      sym__hash_element_interpolation,
    STATE(4282), 1,
      sym__array_element_interpolation,
    STATE(4300), 1,
      sym_array,
    STATE(4301), 1,
      sym__subscripted_interpolations,
    STATE(4424), 1,
      sym__slice_expression_interpolation,
    STATE(4439), 1,
      sym__array_deref_interpolation,
    STATE(4484), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3877), 2,
      aux_sym__interpolated_regexp_content,
      sym_heredoc_content,
    STATE(4426), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5649), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4401), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5660), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [4731] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3185), 1,
      sym__quotelike_end,
    STATE(3878), 1,
      sym_heredoc_content,
    STATE(3970), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [4800] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1850), 1,
      sym__quotelike_end,
    STATE(3879), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [4869] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3187), 1,
      sym__quotelike_end,
    STATE(3880), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [4938] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1851), 1,
      sym__quotelike_end,
    STATE(3881), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5007] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3196), 1,
      sym__quotelike_end,
    STATE(3882), 1,
      sym_heredoc_content,
    STATE(3972), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5076] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3197), 1,
      sym__quotelike_end,
    STATE(3883), 1,
      sym_heredoc_content,
    STATE(3973), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5145] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1800), 1,
      sym__quotelike_end,
    STATE(3884), 1,
      sym_heredoc_content,
    STATE(3953), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5214] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3571), 1,
      sym__quotelike_end,
    STATE(3885), 1,
      sym_heredoc_content,
    STATE(3886), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5283] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3604), 1,
      sym__quotelike_end,
    STATE(3886), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5352] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3572), 1,
      sym__quotelike_end,
    STATE(3887), 1,
      sym_heredoc_content,
    STATE(3888), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5421] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3605), 1,
      sym__quotelike_end,
    STATE(3888), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5490] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3574), 1,
      sym__quotelike_end,
    STATE(3889), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5559] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1960), 1,
      sym__quotelike_end,
    STATE(3890), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5628] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3583), 1,
      sym__quotelike_end,
    STATE(3891), 1,
      sym_heredoc_content,
    STATE(3913), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5697] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3584), 1,
      sym__quotelike_end,
    STATE(3892), 1,
      sym_heredoc_content,
    STATE(3917), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5766] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2440), 1,
      sym__quotelike_end,
    STATE(3893), 1,
      sym_heredoc_content,
    STATE(3897), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5835] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5658), 1,
      sym__quotelike_end_quote,
    ACTIONS(5676), 1,
      anon_sym_DOLLAR,
    ACTIONS(5679), 1,
      anon_sym_AT,
    STATE(4024), 1,
      sym_scalar,
    STATE(4142), 1,
      sym__array_element_interpolation,
    STATE(4153), 1,
      sym__subscripted_interpolations,
    STATE(4162), 1,
      sym_array,
    STATE(4170), 1,
      sym__hash_element_interpolation,
    STATE(4395), 1,
      sym__array_deref_interpolation,
    STATE(4400), 1,
      sym__slice_expression_interpolation,
    STATE(4408), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3894), 2,
      aux_sym__interpolated_regexp_content,
      sym_heredoc_content,
    STATE(4482), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5673), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4471), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
    ACTIONS(5682), 4,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
  [5900] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3469), 1,
      sym__quotelike_end,
    STATE(3895), 1,
      sym_heredoc_content,
    STATE(3962), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [5969] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2380), 1,
      sym__quotelike_end,
    STATE(3896), 1,
      sym_heredoc_content,
    STATE(3933), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6038] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2492), 1,
      sym__quotelike_end,
    STATE(3897), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6107] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2231), 1,
      sym__quotelike_end,
    STATE(3898), 1,
      sym_heredoc_content,
    STATE(3915), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6176] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2509), 1,
      sym__quotelike_end,
    STATE(3899), 1,
      sym_heredoc_content,
    STATE(3900), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6245] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2618), 1,
      sym__quotelike_end,
    STATE(3900), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6314] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2740), 1,
      sym__quotelike_end,
    STATE(3901), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6383] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2741), 1,
      sym__quotelike_end,
    STATE(3902), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6452] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2752), 1,
      sym__quotelike_end,
    STATE(3903), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6521] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2753), 1,
      sym__quotelike_end,
    STATE(3904), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6590] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2702), 1,
      sym__quotelike_end,
    STATE(3901), 1,
      aux_sym__interpolated_string_content,
    STATE(3905), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6659] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2703), 1,
      sym__quotelike_end,
    STATE(3902), 1,
      aux_sym__interpolated_string_content,
    STATE(3906), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6728] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2395), 1,
      sym__quotelike_end,
    STATE(3907), 1,
      sym_heredoc_content,
    STATE(3914), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6797] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2706), 1,
      sym__quotelike_end,
    STATE(3908), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6866] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3470), 1,
      sym__quotelike_end,
    STATE(3909), 1,
      sym_heredoc_content,
    STATE(3963), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [6935] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2716), 1,
      sym__quotelike_end,
    STATE(3903), 1,
      aux_sym__interpolated_string_content,
    STATE(3910), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7004] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2718), 1,
      sym__quotelike_end,
    STATE(3904), 1,
      aux_sym__interpolated_string_content,
    STATE(3911), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7073] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2609), 1,
      sym__quotelike_end,
    STATE(3912), 1,
      sym_heredoc_content,
    STATE(3956), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7142] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3615), 1,
      sym__quotelike_end,
    STATE(3913), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7211] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2476), 1,
      sym__quotelike_end,
    STATE(3914), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7280] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2195), 1,
      sym__quotelike_end,
    STATE(3915), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7349] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2301), 1,
      sym__quotelike_end,
    STATE(3916), 1,
      sym_heredoc_content,
    STATE(3918), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7418] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3689), 1,
      sym__quotelike_end,
    STATE(3917), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7487] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2354), 1,
      sym__quotelike_end,
    STATE(3918), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7556] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2359), 1,
      sym__quotelike_end,
    STATE(3919), 1,
      sym_heredoc_content,
    STATE(3920), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7625] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2384), 1,
      sym__quotelike_end,
    STATE(3920), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7694] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3249), 1,
      sym__quotelike_end,
    STATE(3921), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7763] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2222), 1,
      sym__quotelike_end,
    STATE(3922), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7832] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2919), 1,
      sym__quotelike_end,
    STATE(3923), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7901] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2920), 1,
      sym__quotelike_end,
    STATE(3924), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [7970] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2934), 1,
      sym__quotelike_end,
    STATE(3925), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8039] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2935), 1,
      sym__quotelike_end,
    STATE(3926), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8108] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3473), 1,
      sym__quotelike_end,
    STATE(3927), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8177] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2878), 1,
      sym__quotelike_end,
    STATE(3923), 1,
      aux_sym__interpolated_string_content,
    STATE(3928), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8246] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3250), 1,
      sym__quotelike_end,
    STATE(3929), 1,
      sym_heredoc_content,
    STATE(3950), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8315] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1917), 1,
      sym__quotelike_end,
    STATE(3930), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8384] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2881), 1,
      sym__quotelike_end,
    STATE(3931), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8453] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3370), 1,
      sym__quotelike_end,
    STATE(3932), 1,
      sym_heredoc_content,
    STATE(3944), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8522] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2468), 1,
      sym__quotelike_end,
    STATE(3933), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8591] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1900), 1,
      sym__quotelike_end,
    STATE(3881), 1,
      aux_sym__interpolated_string_content,
    STATE(3934), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8660] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1958), 1,
      sym__quotelike_end,
    STATE(3935), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8729] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3489), 1,
      sym__quotelike_end,
    STATE(3936), 1,
      sym_heredoc_content,
    STATE(3964), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8798] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2895), 1,
      sym__quotelike_end,
    STATE(3925), 1,
      aux_sym__interpolated_string_content,
    STATE(3937), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8867] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2896), 1,
      sym__quotelike_end,
    STATE(3926), 1,
      aux_sym__interpolated_string_content,
    STATE(3938), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [8936] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2610), 1,
      sym__quotelike_end,
    STATE(3876), 1,
      aux_sym__interpolated_string_content,
    STATE(3939), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9005] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1905), 1,
      sym__quotelike_end,
    STATE(3935), 1,
      aux_sym__interpolated_string_content,
    STATE(3940), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9074] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2869), 1,
      sym__quotelike_end,
    STATE(3941), 1,
      sym_heredoc_content,
    STATE(3952), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9143] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1908), 1,
      sym__quotelike_end,
    STATE(3890), 1,
      aux_sym__interpolated_string_content,
    STATE(3942), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9212] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2870), 1,
      sym__quotelike_end,
    STATE(3943), 1,
      sym_heredoc_content,
    STATE(3966), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9281] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3394), 1,
      sym__quotelike_end,
    STATE(3944), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9350] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2872), 1,
      sym__quotelike_end,
    STATE(3945), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9419] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3491), 1,
      sym__quotelike_end,
    STATE(3946), 1,
      sym_heredoc_content,
    STATE(3965), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9488] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2940), 1,
      sym__quotelike_end,
    STATE(3947), 1,
      sym_heredoc_content,
    STATE(3967), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9557] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2941), 1,
      sym__quotelike_end,
    STATE(3948), 1,
      sym_heredoc_content,
    STATE(3968), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9626] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3395), 1,
      sym__quotelike_end,
    STATE(3949), 1,
      sym_heredoc_content,
    STATE(3957), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9695] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3253), 1,
      sym__quotelike_end,
    STATE(3950), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9764] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1879), 1,
      sym__quotelike_end,
    STATE(3879), 1,
      aux_sym__interpolated_string_content,
    STATE(3951), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9833] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3013), 1,
      sym__quotelike_end,
    STATE(3952), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9902] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1785), 1,
      sym__quotelike_end,
    STATE(3953), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [9971] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1786), 1,
      sym__quotelike_end,
    STATE(3954), 1,
      sym_heredoc_content,
    STATE(3958), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10040] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2337), 1,
      sym__quotelike_end,
    STATE(3955), 1,
      sym_heredoc_content,
    STATE(3959), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10109] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2341), 1,
      sym__quotelike_end,
    STATE(3956), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10178] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3310), 1,
      sym__quotelike_end,
    STATE(3957), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10247] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1793), 1,
      sym__quotelike_end,
    STATE(3958), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10316] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2372), 1,
      sym__quotelike_end,
    STATE(3959), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10385] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2373), 1,
      sym__quotelike_end,
    STATE(3960), 1,
      sym_heredoc_content,
    STATE(3961), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10454] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2399), 1,
      sym__quotelike_end,
    STATE(3961), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10523] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3403), 1,
      sym__quotelike_end,
    STATE(3962), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10592] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3422), 1,
      sym__quotelike_end,
    STATE(3963), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10661] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3316), 1,
      sym__quotelike_end,
    STATE(3964), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10730] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3338), 1,
      sym__quotelike_end,
    STATE(3965), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10799] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3014), 1,
      sym__quotelike_end,
    STATE(3966), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10868] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3029), 1,
      sym__quotelike_end,
    STATE(3967), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [10937] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3239), 1,
      sym__quotelike_end,
    STATE(3968), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11006] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3202), 1,
      sym__quotelike_end,
    STATE(3969), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11075] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3203), 1,
      sym__quotelike_end,
    STATE(3970), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11144] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2563), 1,
      sym__quotelike_end,
    STATE(3971), 1,
      sym_heredoc_content,
    STATE(3974), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11213] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3213), 1,
      sym__quotelike_end,
    STATE(3972), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11282] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3214), 1,
      sym__quotelike_end,
    STATE(3973), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11351] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2588), 1,
      sym__quotelike_end,
    STATE(3974), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11420] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2589), 1,
      sym__quotelike_end,
    STATE(3975), 1,
      sym_heredoc_content,
    STATE(3976), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11489] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2603), 1,
      sym__quotelike_end,
    STATE(3976), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11558] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2615), 1,
      sym__quotelike_end,
    STATE(3977), 1,
      sym_heredoc_content,
    STATE(3981), 1,
      aux_sym__interpolated_string_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11627] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3184), 1,
      sym__quotelike_end,
    STATE(3969), 1,
      aux_sym__interpolated_string_content,
    STATE(3978), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11696] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3246), 1,
      sym__quotelike_end,
    STATE(3921), 1,
      aux_sym__interpolated_string_content,
    STATE(3979), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11765] = 20,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2879), 1,
      sym__quotelike_end,
    STATE(3924), 1,
      aux_sym__interpolated_string_content,
    STATE(3980), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11834] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5706), 1,
      anon_sym_DOLLAR,
    ACTIONS(5709), 1,
      anon_sym_AT,
    ACTIONS(5712), 1,
      sym__quotelike_end_quote,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3981), 2,
      aux_sym__interpolated_string_content,
      sym_heredoc_content,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5703), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5714), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11898] = 19,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5719), 1,
      anon_sym_DOLLAR,
    ACTIONS(5721), 1,
      anon_sym_AT,
    ACTIONS(5725), 1,
      sym_heredoc_end,
    STATE(3982), 1,
      sym_heredoc_content,
    STATE(3991), 1,
      aux_sym_heredoc_content_repeat1,
    STATE(4151), 1,
      sym_scalar,
    STATE(4409), 1,
      sym_array,
    STATE(4416), 1,
      sym__subscripted_interpolations,
    STATE(4431), 1,
      sym__array_element_interpolation,
    STATE(4433), 1,
      sym__hash_element_interpolation,
    STATE(4558), 1,
      sym__array_deref_interpolation,
    STATE(4563), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(4585), 1,
      sym__slice_expression_interpolation,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5723), 2,
      sym_escape_sequence,
      sym__heredoc_middle,
    STATE(4583), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4576), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [11963] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3927), 1,
      aux_sym__interpolated_string_content,
    STATE(3983), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12026] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3889), 1,
      aux_sym__interpolated_string_content,
    STATE(3984), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12089] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3930), 1,
      aux_sym__interpolated_string_content,
    STATE(3985), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12152] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3880), 1,
      aux_sym__interpolated_string_content,
    STATE(3986), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12215] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3908), 1,
      aux_sym__interpolated_string_content,
    STATE(3987), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12278] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3977), 1,
      aux_sym__interpolated_string_content,
    STATE(3988), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12341] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3931), 1,
      aux_sym__interpolated_string_content,
    STATE(3989), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12404] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5639), 1,
      anon_sym_DOLLAR,
    ACTIONS(5641), 1,
      anon_sym_AT,
    STATE(3945), 1,
      aux_sym__interpolated_string_content,
    STATE(3990), 1,
      sym_heredoc_content,
    STATE(4078), 1,
      sym_scalar,
    STATE(4308), 1,
      sym_array,
    STATE(4309), 1,
      sym__subscripted_interpolations,
    STATE(4310), 1,
      sym__array_element_interpolation,
    STATE(4311), 1,
      sym__hash_element_interpolation,
    STATE(4513), 1,
      sym__slice_expression_interpolation,
    STATE(4519), 1,
      sym__array_deref_interpolation,
    STATE(4524), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4497), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5645), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
    STATE(4493), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12467] = 19,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5719), 1,
      anon_sym_DOLLAR,
    ACTIONS(5721), 1,
      anon_sym_AT,
    ACTIONS(5727), 1,
      sym_heredoc_end,
    STATE(3991), 1,
      sym_heredoc_content,
    STATE(3992), 1,
      aux_sym_heredoc_content_repeat1,
    STATE(4151), 1,
      sym_scalar,
    STATE(4409), 1,
      sym_array,
    STATE(4416), 1,
      sym__subscripted_interpolations,
    STATE(4431), 1,
      sym__array_element_interpolation,
    STATE(4433), 1,
      sym__hash_element_interpolation,
    STATE(4558), 1,
      sym__array_deref_interpolation,
    STATE(4563), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(4585), 1,
      sym__slice_expression_interpolation,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5723), 2,
      sym_escape_sequence,
      sym__heredoc_middle,
    STATE(4583), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4576), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12532] = 18,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5732), 1,
      anon_sym_DOLLAR,
    ACTIONS(5735), 1,
      anon_sym_AT,
    ACTIONS(5741), 1,
      sym_heredoc_end,
    STATE(4151), 1,
      sym_scalar,
    STATE(4409), 1,
      sym_array,
    STATE(4416), 1,
      sym__subscripted_interpolations,
    STATE(4431), 1,
      sym__array_element_interpolation,
    STATE(4433), 1,
      sym__hash_element_interpolation,
    STATE(4558), 1,
      sym__array_deref_interpolation,
    STATE(4563), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(4585), 1,
      sym__slice_expression_interpolation,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5738), 2,
      sym_escape_sequence,
      sym__heredoc_middle,
    STATE(3992), 2,
      sym_heredoc_content,
      aux_sym_heredoc_content_repeat1,
    STATE(4583), 2,
      sym__interpolations,
      sym__interpolation_fallbacks,
    ACTIONS(5729), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(4576), 3,
      sym__braced_scalar,
      sym__braced_array,
      sym__scalar_deref_interpolation,
  [12595] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3592), 1,
      anon_sym_for,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5743), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(3993), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 16,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_foreach,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
      anon_sym_if,
      anon_sym_unless,
      anon_sym_while,
      anon_sym_until,
  [12632] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3570), 1,
      anon_sym_for,
    STATE(3993), 1,
      aux_sym__term_rightward_repeat1,
    STATE(3994), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3653), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 16,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_foreach,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
      anon_sym_if,
      anon_sym_unless,
      anon_sym_while,
      anon_sym_until,
  [12671] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5746), 1,
      anon_sym_RBRACE,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(3995), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12720] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5762), 1,
      anon_sym_RPAREN,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(3996), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12769] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5764), 1,
      anon_sym_RBRACE,
    ACTIONS(5766), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(3997), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12820] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5769), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(3998), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12869] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5771), 1,
      anon_sym_RBRACE,
    ACTIONS(5773), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(3999), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12920] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5771), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4000), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [12969] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5776), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4001), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13018] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5778), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4002), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13067] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5780), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4003), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13116] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5778), 1,
      anon_sym_RBRACE,
    ACTIONS(5782), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4004), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13167] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5780), 1,
      anon_sym_RBRACE,
    ACTIONS(5785), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4005), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13218] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5788), 1,
      anon_sym_RPAREN,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4006), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13267] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5764), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4007), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13316] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5790), 1,
      anon_sym_RBRACE,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4008), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13365] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5776), 1,
      anon_sym_RBRACE,
    ACTIONS(5792), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4009), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13416] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5795), 1,
      anon_sym_RPAREN,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4010), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13465] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5790), 1,
      anon_sym_RBRACE,
    ACTIONS(5797), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4011), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13516] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5746), 1,
      anon_sym_RBRACE,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5800), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4012), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13567] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    ACTIONS(5769), 1,
      anon_sym_RBRACE,
    ACTIONS(5803), 1,
      sym__PERLY_SEMICOLON,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4013), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 5,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13618] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5810), 1,
      anon_sym_DOLLAR,
    ACTIONS(5812), 1,
      aux_sym__interpolation_fallbacks_token1,
    ACTIONS(5814), 1,
      aux_sym__ident_special_token1,
    ACTIONS(5818), 1,
      sym__quotelike_end_quote,
    STATE(44), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4014), 1,
      sym_heredoc_content,
    STATE(4331), 1,
      sym__indirob,
    STATE(4371), 1,
      sym_block,
    STATE(4372), 1,
      sym__var_indirob_autoquote,
    STATE(4451), 1,
      sym__var_indirob,
    STATE(4594), 1,
      sym__quotelike_end,
    ACTIONS(5806), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(5816), 2,
      sym__quotelike_end_zw,
      sym__no_interp_whitespace_zw,
    STATE(4325), 2,
      sym_scalar,
      sym__ident_special,
  [13676] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5822), 1,
      anon_sym_DOLLAR,
    ACTIONS(5824), 1,
      aux_sym__interpolation_fallbacks_token1,
    ACTIONS(5826), 1,
      aux_sym__ident_special_token1,
    ACTIONS(5830), 1,
      sym__quotelike_end_quote,
    STATE(80), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4015), 1,
      sym_heredoc_content,
    STATE(4054), 1,
      sym__var_indirob_autoquote,
    STATE(4086), 1,
      sym__indirob,
    STATE(4109), 1,
      sym_block,
    STATE(4246), 1,
      sym__var_indirob,
    STATE(4473), 1,
      sym__quotelike_end,
    ACTIONS(5820), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(5828), 2,
      sym__quotelike_end_zw,
      sym__no_interp_whitespace_zw,
    STATE(4085), 2,
      sym_scalar,
      sym__ident_special,
  [13734] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5834), 1,
      anon_sym_DOLLAR,
    ACTIONS(5836), 1,
      aux_sym__interpolation_fallbacks_token1,
    ACTIONS(5838), 1,
      aux_sym__ident_special_token1,
    ACTIONS(5842), 1,
      sym__quotelike_end_quote,
    STATE(61), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4016), 1,
      sym_heredoc_content,
    STATE(4087), 1,
      sym_block,
    STATE(4102), 1,
      sym__indirob,
    STATE(4107), 1,
      sym__var_indirob_autoquote,
    STATE(4229), 1,
      sym__var_indirob,
    STATE(4440), 1,
      sym__quotelike_end,
    ACTIONS(5832), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(5840), 2,
      sym__quotelike_end_zw,
      sym__no_interp_whitespace_zw,
    STATE(4095), 2,
      sym_scalar,
      sym__ident_special,
  [13792] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5846), 1,
      anon_sym_DOLLAR,
    ACTIONS(5848), 1,
      aux_sym__interpolation_fallbacks_token1,
    ACTIONS(5850), 1,
      aux_sym__ident_special_token1,
    ACTIONS(5854), 1,
      sym__quotelike_end_quote,
    STATE(46), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4017), 1,
      sym_heredoc_content,
    STATE(4168), 1,
      sym__var_indirob_autoquote,
    STATE(4210), 1,
      sym_block,
    STATE(4244), 1,
      sym__indirob,
    STATE(4314), 1,
      sym__var_indirob,
    STATE(4512), 1,
      sym__quotelike_end,
    ACTIONS(5844), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(5852), 2,
      sym__quotelike_end_zw,
      sym__no_interp_whitespace_zw,
    STATE(4165), 2,
      sym_scalar,
      sym__ident_special,
  [13850] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5754), 1,
      anon_sym_for,
    ACTIONS(5756), 1,
      anon_sym_foreach,
    STATE(427), 1,
      sym__loops,
    STATE(550), 1,
      sym__conditionals,
    STATE(4018), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5752), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(5758), 2,
      anon_sym_if,
      anon_sym_unless,
    ACTIONS(5760), 2,
      anon_sym_while,
      anon_sym_until,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [13896] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5750), 1,
      anon_sym_and,
    ACTIONS(5858), 1,
      anon_sym_for,
    STATE(4019), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 15,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_foreach,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
      anon_sym_if,
      anon_sym_unless,
      anon_sym_while,
      anon_sym_until,
  [13930] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5858), 1,
      anon_sym_for,
    STATE(4020), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 16,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_foreach,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
      anon_sym_if,
      anon_sym_unless,
      anon_sym_while,
      anon_sym_until,
  [13962] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5862), 1,
      anon_sym_for,
    STATE(4021), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5860), 16,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_foreach,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
      anon_sym_if,
      anon_sym_unless,
      anon_sym_while,
      anon_sym_until,
  [13994] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5864), 1,
      anon_sym_DOLLAR,
    ACTIONS(5866), 1,
      anon_sym_AT,
    ACTIONS(5870), 1,
      anon_sym_sub,
    ACTIONS(5872), 1,
      anon_sym_method,
    ACTIONS(5874), 1,
      anon_sym_LPAREN,
    ACTIONS(5876), 1,
      anon_sym_PERCENT2,
    STATE(1814), 1,
      sym__decl_variable_list,
    STATE(1818), 1,
      sym__declared_vars,
    STATE(4022), 1,
      sym_heredoc_content,
    STATE(5095), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
    STATE(1815), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [14041] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5880), 1,
      anon_sym_LBRACE,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR,
    ACTIONS(5884), 1,
      anon_sym_AT,
    ACTIONS(5886), 1,
      anon_sym_LPAREN,
    ACTIONS(5888), 1,
      anon_sym_STAR,
    ACTIONS(5890), 1,
      anon_sym_LBRACK,
    ACTIONS(5892), 1,
      anon_sym_PERCENT,
    ACTIONS(5894), 1,
      anon_sym_AMP,
    ACTIONS(5896), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(5898), 1,
      sym__RECOVER_ARROW,
    STATE(3360), 1,
      sym_scalar,
    STATE(3361), 1,
      sym_method,
    STATE(4023), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5878), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14092] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5904), 1,
      anon_sym_DASH_GT2,
    ACTIONS(5906), 1,
      anon_sym_LBRACK2,
    ACTIONS(5908), 1,
      anon_sym_LBRACE2,
    STATE(4024), 1,
      sym_heredoc_content,
    STATE(5077), 1,
      sym__interp_arrow,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5902), 7,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
  [14129] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5912), 1,
      anon_sym_LBRACE,
    ACTIONS(5914), 1,
      anon_sym_DOLLAR,
    ACTIONS(5916), 1,
      anon_sym_AT,
    ACTIONS(5918), 1,
      anon_sym_LPAREN,
    ACTIONS(5920), 1,
      anon_sym_STAR,
    ACTIONS(5922), 1,
      anon_sym_LBRACK,
    ACTIONS(5924), 1,
      anon_sym_PERCENT,
    ACTIONS(5926), 1,
      anon_sym_AMP,
    ACTIONS(5928), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(5930), 1,
      sym__RECOVER_ARROW,
    STATE(3293), 1,
      sym_scalar,
    STATE(3294), 1,
      sym_method,
    STATE(4025), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5910), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14180] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5934), 1,
      anon_sym_LBRACE,
    ACTIONS(5936), 1,
      anon_sym_DOLLAR,
    ACTIONS(5938), 1,
      anon_sym_AT,
    ACTIONS(5940), 1,
      anon_sym_LPAREN,
    ACTIONS(5942), 1,
      anon_sym_STAR,
    ACTIONS(5944), 1,
      anon_sym_LBRACK,
    ACTIONS(5946), 1,
      anon_sym_PERCENT,
    ACTIONS(5948), 1,
      anon_sym_AMP,
    ACTIONS(5950), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(5952), 1,
      sym__RECOVER_ARROW,
    STATE(2393), 1,
      sym_scalar,
    STATE(2394), 1,
      sym_method,
    STATE(4026), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5932), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14231] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5954), 1,
      sym__NONASSOC,
    STATE(4027), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14262] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5958), 1,
      anon_sym_LBRACE,
    ACTIONS(5960), 1,
      anon_sym_DOLLAR,
    ACTIONS(5962), 1,
      anon_sym_AT,
    ACTIONS(5964), 1,
      anon_sym_LPAREN,
    ACTIONS(5966), 1,
      anon_sym_STAR,
    ACTIONS(5968), 1,
      anon_sym_LBRACK,
    ACTIONS(5970), 1,
      anon_sym_PERCENT,
    ACTIONS(5972), 1,
      anon_sym_AMP,
    ACTIONS(5974), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(5976), 1,
      sym__RECOVER_ARROW,
    STATE(2505), 1,
      sym_scalar,
    STATE(2508), 1,
      sym_method,
    STATE(4028), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5956), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14313] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5978), 1,
      anon_sym_DASH_GT2,
    ACTIONS(5980), 1,
      anon_sym_LBRACK2,
    ACTIONS(5982), 1,
      anon_sym_LBRACE2,
    STATE(4029), 1,
      sym_heredoc_content,
    STATE(5085), 1,
      sym__interp_arrow,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5902), 7,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
  [14350] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5984), 1,
      anon_sym_COLON,
    ACTIONS(5986), 1,
      anon_sym_DOLLAR,
    ACTIONS(5988), 1,
      anon_sym_AT,
    ACTIONS(5990), 1,
      anon_sym_RPAREN,
    ACTIONS(5992), 1,
      anon_sym_PERCENT2,
    STATE(4030), 1,
      sym_heredoc_content,
    STATE(4044), 1,
      aux_sym_signature_repeat1,
    STATE(4540), 1,
      sym__signature_scalar,
    STATE(4921), 1,
      sym__signature_vars,
    STATE(4933), 1,
      sym__signature_array,
    STATE(4935), 1,
      sym__signature_hash,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4916), 4,
      sym_mandatory_parameter,
      sym_optional_parameter,
      sym_named_parameter,
      sym_slurpy_parameter,
  [14397] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4031), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 11,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14426] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5996), 1,
      anon_sym_LBRACE,
    ACTIONS(5998), 1,
      anon_sym_DOLLAR,
    ACTIONS(6000), 1,
      anon_sym_AT,
    ACTIONS(6002), 1,
      anon_sym_LPAREN,
    ACTIONS(6004), 1,
      anon_sym_STAR,
    ACTIONS(6006), 1,
      anon_sym_LBRACK,
    ACTIONS(6008), 1,
      anon_sym_PERCENT,
    ACTIONS(6010), 1,
      anon_sym_AMP,
    ACTIONS(6012), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(6014), 1,
      sym__RECOVER_ARROW,
    STATE(2223), 1,
      sym_scalar,
    STATE(2228), 1,
      sym_method,
    STATE(4032), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5994), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14477] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4033), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 11,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14506] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4034), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 11,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14535] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5954), 1,
      sym__NONASSOC,
    STATE(4035), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14566] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6016), 1,
      anon_sym_COLON,
    ACTIONS(6019), 1,
      anon_sym_DOLLAR,
    ACTIONS(6022), 1,
      anon_sym_AT,
    ACTIONS(6025), 1,
      anon_sym_RPAREN,
    ACTIONS(6027), 1,
      anon_sym_PERCENT2,
    STATE(4540), 1,
      sym__signature_scalar,
    STATE(4921), 1,
      sym__signature_vars,
    STATE(4933), 1,
      sym__signature_array,
    STATE(4935), 1,
      sym__signature_hash,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4036), 2,
      sym_heredoc_content,
      aux_sym_signature_repeat1,
    STATE(4916), 4,
      sym_mandatory_parameter,
      sym_optional_parameter,
      sym_named_parameter,
      sym_slurpy_parameter,
  [14611] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4037), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 11,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14640] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4038), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 11,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14669] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4039), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 11,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14698] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6032), 1,
      anon_sym_LBRACE,
    ACTIONS(6034), 1,
      anon_sym_DOLLAR,
    ACTIONS(6036), 1,
      anon_sym_AT,
    ACTIONS(6038), 1,
      anon_sym_LPAREN,
    ACTIONS(6040), 1,
      anon_sym_STAR,
    ACTIONS(6042), 1,
      anon_sym_LBRACK,
    ACTIONS(6044), 1,
      anon_sym_PERCENT,
    ACTIONS(6046), 1,
      anon_sym_AMP,
    ACTIONS(6048), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(6050), 1,
      sym__RECOVER_ARROW,
    STATE(2273), 1,
      sym_scalar,
    STATE(2275), 1,
      sym_method,
    STATE(4040), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6030), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14749] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6054), 1,
      anon_sym_LBRACE,
    ACTIONS(6056), 1,
      anon_sym_DOLLAR,
    ACTIONS(6058), 1,
      anon_sym_AT,
    ACTIONS(6060), 1,
      anon_sym_LPAREN,
    ACTIONS(6062), 1,
      anon_sym_STAR,
    ACTIONS(6064), 1,
      anon_sym_LBRACK,
    ACTIONS(6066), 1,
      anon_sym_PERCENT,
    ACTIONS(6068), 1,
      anon_sym_AMP,
    ACTIONS(6070), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(6072), 1,
      sym__RECOVER_ARROW,
    STATE(1782), 1,
      sym_scalar,
    STATE(1783), 1,
      sym_method,
    STATE(4041), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6052), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14800] = 16,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6076), 1,
      anon_sym_LBRACE,
    ACTIONS(6078), 1,
      anon_sym_DOLLAR,
    ACTIONS(6080), 1,
      anon_sym_AT,
    ACTIONS(6082), 1,
      anon_sym_LPAREN,
    ACTIONS(6084), 1,
      anon_sym_STAR,
    ACTIONS(6086), 1,
      anon_sym_LBRACK,
    ACTIONS(6088), 1,
      anon_sym_PERCENT,
    ACTIONS(6090), 1,
      anon_sym_AMP,
    ACTIONS(6092), 1,
      anon_sym_DOLLAR_POUND,
    ACTIONS(6094), 1,
      sym__RECOVER_ARROW,
    STATE(2311), 1,
      sym_scalar,
    STATE(2312), 1,
      sym_method,
    STATE(4042), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6074), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [14851] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6096), 1,
      sym__NONASSOC,
    STATE(4043), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14882] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5984), 1,
      anon_sym_COLON,
    ACTIONS(5986), 1,
      anon_sym_DOLLAR,
    ACTIONS(5988), 1,
      anon_sym_AT,
    ACTIONS(5992), 1,
      anon_sym_PERCENT2,
    ACTIONS(6098), 1,
      anon_sym_RPAREN,
    STATE(4036), 1,
      aux_sym_signature_repeat1,
    STATE(4044), 1,
      sym_heredoc_content,
    STATE(4540), 1,
      sym__signature_scalar,
    STATE(4921), 1,
      sym__signature_vars,
    STATE(4933), 1,
      sym__signature_array,
    STATE(4935), 1,
      sym__signature_hash,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4916), 4,
      sym_mandatory_parameter,
      sym_optional_parameter,
      sym_named_parameter,
      sym_slurpy_parameter,
  [14929] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6096), 1,
      sym__NONASSOC,
    STATE(4045), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [14960] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5864), 1,
      anon_sym_DOLLAR,
    ACTIONS(5866), 1,
      anon_sym_AT,
    ACTIONS(5874), 1,
      anon_sym_LPAREN,
    ACTIONS(5876), 1,
      anon_sym_PERCENT2,
    ACTIONS(6100), 1,
      anon_sym_sub,
    ACTIONS(6102), 1,
      anon_sym_method,
    STATE(1814), 1,
      sym__decl_variable_list,
    STATE(1818), 1,
      sym__declared_vars,
    STATE(4046), 1,
      sym_heredoc_content,
    STATE(5139), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
    STATE(1815), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [15007] = 14,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5864), 1,
      anon_sym_DOLLAR,
    ACTIONS(5866), 1,
      anon_sym_AT,
    ACTIONS(5874), 1,
      anon_sym_LPAREN,
    ACTIONS(5876), 1,
      anon_sym_PERCENT2,
    ACTIONS(6104), 1,
      anon_sym_sub,
    ACTIONS(6106), 1,
      anon_sym_method,
    STATE(1814), 1,
      sym__decl_variable_list,
    STATE(1818), 1,
      sym__declared_vars,
    STATE(4047), 1,
      sym_heredoc_content,
    STATE(5150), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
    STATE(1815), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [15054] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6110), 1,
      anon_sym_DOLLAR,
    ACTIONS(6112), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6114), 1,
      sym__dollar_ident_zw,
    STATE(74), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1334), 1,
      sym__var_indirob,
    STATE(4048), 1,
      sym_heredoc_content,
    ACTIONS(6108), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1330), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1328), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15098] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6118), 1,
      anon_sym_DOLLAR,
    ACTIONS(6120), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6122), 1,
      sym__dollar_ident_zw,
    STATE(57), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1295), 1,
      sym__var_indirob,
    STATE(4049), 1,
      sym_heredoc_content,
    ACTIONS(6116), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1281), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1279), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15142] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4050), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15170] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6124), 1,
      sym__NONASSOC,
    STATE(4051), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15200] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6128), 1,
      anon_sym_DOLLAR,
    ACTIONS(6130), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6132), 1,
      sym__dollar_ident_zw,
    STATE(58), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1396), 1,
      sym__var_indirob,
    STATE(4052), 1,
      sym_heredoc_content,
    ACTIONS(6126), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1382), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1381), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15244] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6136), 1,
      anon_sym_DOLLAR,
    ACTIONS(6138), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6140), 1,
      sym__dollar_ident_zw,
    STATE(51), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1136), 1,
      sym__var_indirob,
    STATE(4053), 1,
      sym_heredoc_content,
    ACTIONS(6134), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1137), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1141), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15288] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6142), 1,
      sym__NONASSOC,
    STATE(4054), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2698), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15318] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4055), 1,
      sym_heredoc_content,
    STATE(4064), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4245), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [15358] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6152), 1,
      anon_sym_DOLLAR,
    ACTIONS(6154), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6156), 1,
      sym__dollar_ident_zw,
    STATE(59), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1371), 1,
      sym__var_indirob,
    STATE(4056), 1,
      sym_heredoc_content,
    ACTIONS(6150), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1365), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1373), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15402] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4057), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15430] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6164), 1,
      sym__dollar_ident_zw,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3087), 1,
      sym__var_indirob,
    STATE(4058), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15474] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4059), 1,
      sym_heredoc_content,
    STATE(4101), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4261), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [15514] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5834), 1,
      anon_sym_DOLLAR,
    ACTIONS(5838), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6166), 1,
      sym__dollar_ident_zw,
    STATE(69), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4060), 1,
      sym_heredoc_content,
    STATE(4116), 1,
      sym__var_indirob,
    ACTIONS(5832), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4102), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4095), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15558] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4061), 1,
      sym_heredoc_content,
    STATE(4286), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [15598] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6174), 1,
      sym__dollar_ident_zw,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2114), 1,
      sym__var_indirob,
    STATE(4062), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15642] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5846), 1,
      anon_sym_DOLLAR,
    ACTIONS(5850), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6176), 1,
      sym__dollar_ident_zw,
    STATE(53), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4063), 1,
      sym_heredoc_content,
    STATE(4292), 1,
      sym__var_indirob,
    ACTIONS(5844), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4244), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4165), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15686] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4064), 1,
      sym_heredoc_content,
    STATE(4270), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [15726] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6184), 1,
      sym__dollar_ident_zw,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2121), 1,
      sym__var_indirob,
    STATE(4065), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15770] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6188), 1,
      anon_sym_DOLLAR,
    ACTIONS(6190), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6192), 1,
      sym__dollar_ident_zw,
    STATE(65), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4066), 1,
      sym_heredoc_content,
    STATE(4577), 1,
      sym__var_indirob,
    ACTIONS(6186), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4590), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4587), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15814] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6196), 1,
      anon_sym_DOLLAR,
    ACTIONS(6198), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6200), 1,
      sym__dollar_ident_zw,
    STATE(66), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4067), 1,
      sym_heredoc_content,
    STATE(4509), 1,
      sym__var_indirob,
    ACTIONS(6194), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4541), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4537), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15858] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4068), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2719), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15886] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6204), 1,
      anon_sym_DOLLAR,
    ACTIONS(6206), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6208), 1,
      sym__dollar_ident_zw,
    STATE(67), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1412), 1,
      sym__var_indirob,
    STATE(4069), 1,
      sym_heredoc_content,
    ACTIONS(6202), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1422), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1416), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [15930] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4070), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2713), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2715), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [15958] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6212), 1,
      anon_sym_DOLLAR,
    ACTIONS(6214), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6216), 1,
      sym__dollar_ident_zw,
    STATE(68), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1376), 1,
      sym__var_indirob,
    STATE(4071), 1,
      sym_heredoc_content,
    ACTIONS(6210), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1388), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1394), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16002] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6220), 1,
      anon_sym_DOLLAR,
    ACTIONS(6222), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6224), 1,
      sym__dollar_ident_zw,
    STATE(70), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1289), 1,
      sym__var_indirob,
    STATE(4072), 1,
      sym_heredoc_content,
    ACTIONS(6218), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1302), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1275), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16046] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6228), 1,
      anon_sym_DOLLAR,
    ACTIONS(6230), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6232), 1,
      sym__dollar_ident_zw,
    STATE(39), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1247), 1,
      sym__var_indirob,
    STATE(4073), 1,
      sym_heredoc_content,
    ACTIONS(6226), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1245), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1254), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16090] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6236), 1,
      anon_sym_DOLLAR,
    ACTIONS(6238), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6240), 1,
      sym__dollar_ident_zw,
    STATE(71), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1240), 1,
      sym__var_indirob,
    STATE(4074), 1,
      sym_heredoc_content,
    ACTIONS(6234), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1237), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1236), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16134] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4075), 1,
      sym_heredoc_content,
    STATE(4295), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [16174] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6244), 1,
      anon_sym_DOLLAR,
    ACTIONS(6246), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6248), 1,
      sym__dollar_ident_zw,
    STATE(72), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1331), 1,
      sym__var_indirob,
    STATE(4076), 1,
      sym_heredoc_content,
    ACTIONS(6242), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1325), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1324), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16218] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6252), 1,
      anon_sym_DOLLAR,
    ACTIONS(6254), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6256), 1,
      sym__dollar_ident_zw,
    STATE(73), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1286), 1,
      sym__var_indirob,
    STATE(4077), 1,
      sym_heredoc_content,
    ACTIONS(6250), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1274), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1272), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16262] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6258), 1,
      anon_sym_DASH_GT2,
    ACTIONS(6260), 1,
      anon_sym_LBRACK2,
    ACTIONS(6262), 1,
      anon_sym_LBRACE2,
    STATE(4078), 1,
      sym_heredoc_content,
    STATE(5092), 1,
      sym__interp_arrow,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5902), 6,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
  [16298] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6266), 1,
      anon_sym_DOLLAR,
    ACTIONS(6268), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6270), 1,
      sym__no_interp_whitespace_zw,
    STATE(77), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4079), 1,
      sym_heredoc_content,
    STATE(4483), 1,
      sym__var_indirob,
    ACTIONS(6264), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4465), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4462), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16342] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6274), 1,
      anon_sym_DOLLAR,
    ACTIONS(6276), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6278), 1,
      sym__dollar_ident_zw,
    STATE(75), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4080), 1,
      sym_heredoc_content,
    STATE(4477), 1,
      sym__var_indirob,
    ACTIONS(6272), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4457), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4456), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16386] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4081), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16414] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6286), 1,
      sym__dollar_ident_zw,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2149), 1,
      sym__var_indirob,
    STATE(4082), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16458] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4083), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16486] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6266), 1,
      anon_sym_DOLLAR,
    ACTIONS(6268), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6288), 1,
      sym__dollar_ident_zw,
    STATE(77), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4084), 1,
      sym_heredoc_content,
    STATE(4483), 1,
      sym__var_indirob,
    ACTIONS(6264), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4465), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4462), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16530] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4085), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16558] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4086), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16586] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6290), 1,
      sym__NONASSOC,
    STATE(4087), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2723), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16616] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4088), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16644] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5834), 1,
      anon_sym_DOLLAR,
    ACTIONS(5838), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6292), 1,
      sym__no_interp_whitespace_zw,
    STATE(61), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4027), 1,
      sym__var_indirob_autoquote,
    STATE(4035), 1,
      sym_block,
    STATE(4089), 1,
      sym_heredoc_content,
    STATE(4102), 1,
      sym__indirob,
    STATE(4116), 1,
      sym__var_indirob,
    ACTIONS(5832), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4095), 2,
      sym_scalar,
      sym__ident_special,
  [16692] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4090), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2682), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2686), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16720] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6294), 1,
      anon_sym_STAR,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2121), 1,
      sym__var_indirob,
    STATE(4091), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16764] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5822), 1,
      anon_sym_DOLLAR,
    ACTIONS(5826), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6296), 1,
      sym__no_interp_whitespace_zw,
    STATE(80), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4043), 1,
      sym_block,
    STATE(4045), 1,
      sym__var_indirob_autoquote,
    STATE(4086), 1,
      sym__indirob,
    STATE(4090), 1,
      sym__var_indirob,
    STATE(4092), 1,
      sym_heredoc_content,
    ACTIONS(5820), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4085), 2,
      sym_scalar,
      sym__ident_special,
  [16812] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6298), 1,
      anon_sym_STAR,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2114), 1,
      sym__var_indirob,
    STATE(4093), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16856] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5822), 1,
      anon_sym_DOLLAR,
    ACTIONS(5826), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6300), 1,
      sym__dollar_ident_zw,
    STATE(64), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4090), 1,
      sym__var_indirob,
    STATE(4094), 1,
      sym_heredoc_content,
    ACTIONS(5820), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4086), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4085), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [16900] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4095), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [16928] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4096), 1,
      sym_heredoc_content,
    STATE(4118), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4204), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [16968] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6306), 1,
      anon_sym_STAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1755), 1,
      sym__var_indirob,
    STATE(4097), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17012] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6022), 1,
      anon_sym_AT,
    ACTIONS(6025), 1,
      anon_sym_RPAREN,
    ACTIONS(6027), 1,
      anon_sym_PERCENT2,
    ACTIONS(6310), 1,
      anon_sym_COLON,
    ACTIONS(6313), 1,
      anon_sym_DOLLAR,
    STATE(4098), 1,
      sym_heredoc_content,
    STATE(4588), 1,
      sym__signature_scalar,
    STATE(4933), 1,
      sym__signature_array,
    STATE(4935), 1,
      sym__signature_hash,
    STATE(5115), 1,
      sym__signature_vars,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4916), 4,
      sym_mandatory_parameter,
      sym_optional_parameter,
      sym_named_parameter,
      sym_slurpy_parameter,
  [17056] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6318), 1,
      anon_sym_DOLLAR,
    ACTIONS(6320), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6322), 1,
      sym__dollar_ident_zw,
    STATE(54), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1283), 1,
      sym__var_indirob,
    STATE(4099), 1,
      sym_heredoc_content,
    ACTIONS(6316), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1288), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1284), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17100] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4100), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17128] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4101), 1,
      sym_heredoc_content,
    STATE(4266), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17168] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4102), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17196] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6324), 1,
      anon_sym_STAR,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3087), 1,
      sym__var_indirob,
    STATE(4103), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17240] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4104), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2713), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2715), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17268] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4105), 1,
      sym_heredoc_content,
    STATE(4164), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17308] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6332), 1,
      sym__dollar_ident_zw,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2479), 1,
      sym__var_indirob,
    STATE(4106), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17352] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6290), 1,
      sym__NONASSOC,
    STATE(4107), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2698), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17382] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4108), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17410] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6142), 1,
      sym__NONASSOC,
    STATE(4109), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2723), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17440] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4075), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4110), 1,
      sym_heredoc_content,
    STATE(4155), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17480] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6336), 1,
      anon_sym_DOLLAR,
    ACTIONS(6338), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6340), 1,
      sym__dollar_ident_zw,
    STATE(45), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4111), 1,
      sym_heredoc_content,
    STATE(5286), 1,
      sym__var_indirob,
    ACTIONS(6334), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5164), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(5241), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17524] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4112), 1,
      sym_heredoc_content,
    STATE(4117), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4173), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17564] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6344), 1,
      anon_sym_DOLLAR,
    ACTIONS(6346), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6348), 1,
      sym__dollar_ident_zw,
    STATE(55), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1266), 1,
      sym__var_indirob,
    STATE(4113), 1,
      sym_heredoc_content,
    ACTIONS(6342), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1261), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1262), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17608] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4114), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17636] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4115), 1,
      sym_heredoc_content,
    STATE(4120), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4297), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17676] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4116), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2682), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2686), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17704] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4117), 1,
      sym_heredoc_content,
    STATE(4178), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17744] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4118), 1,
      sym_heredoc_content,
    STATE(4224), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17784] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6354), 1,
      anon_sym_STAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2168), 1,
      sym__var_indirob,
    STATE(4119), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17828] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4120), 1,
      sym_heredoc_content,
    STATE(4187), 1,
      sym__quotelike_middle,
    STATE(4350), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [17868] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6196), 1,
      anon_sym_DOLLAR,
    ACTIONS(6198), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6358), 1,
      sym__no_interp_whitespace_zw,
    STATE(66), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4121), 1,
      sym_heredoc_content,
    STATE(4509), 1,
      sym__var_indirob,
    ACTIONS(6194), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4541), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4537), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17912] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6360), 1,
      sym__dollar_ident_zw,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1755), 1,
      sym__var_indirob,
    STATE(4122), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [17956] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6124), 1,
      sym__NONASSOC,
    STATE(4123), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [17986] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6364), 1,
      anon_sym_DOLLAR,
    ACTIONS(6366), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6368), 1,
      sym__dollar_ident_zw,
    STATE(56), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1360), 1,
      sym__var_indirob,
    STATE(4124), 1,
      sym_heredoc_content,
    ACTIONS(6362), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1350), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1349), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18030] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6274), 1,
      anon_sym_DOLLAR,
    ACTIONS(6276), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6370), 1,
      sym__no_interp_whitespace_zw,
    STATE(75), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4125), 1,
      sym_heredoc_content,
    STATE(4477), 1,
      sym__var_indirob,
    ACTIONS(6272), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4457), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4456), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18074] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6188), 1,
      anon_sym_DOLLAR,
    ACTIONS(6190), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6372), 1,
      sym__no_interp_whitespace_zw,
    STATE(65), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4126), 1,
      sym_heredoc_content,
    STATE(4577), 1,
      sym__var_indirob,
    ACTIONS(6186), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4590), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4587), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18118] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6374), 1,
      anon_sym_STAR,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2149), 1,
      sym__var_indirob,
    STATE(4127), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18162] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4061), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4128), 1,
      sym_heredoc_content,
    STATE(4269), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [18202] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5810), 1,
      anon_sym_DOLLAR,
    ACTIONS(5814), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6376), 1,
      sym__dollar_ident_zw,
    STATE(52), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4129), 1,
      sym_heredoc_content,
    STATE(4370), 1,
      sym__var_indirob,
    ACTIONS(5806), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4331), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(4325), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18246] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6378), 1,
      anon_sym_STAR,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2479), 1,
      sym__var_indirob,
    STATE(4130), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18290] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6384), 1,
      anon_sym_STAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2032), 1,
      sym__var_indirob,
    STATE(4131), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18334] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6388), 1,
      sym__dollar_ident_zw,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2032), 1,
      sym__var_indirob,
    STATE(4132), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18378] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6390), 1,
      sym__dollar_ident_zw,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2168), 1,
      sym__var_indirob,
    STATE(4133), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18422] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4105), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4134), 1,
      sym_heredoc_content,
    STATE(4194), 1,
      sym__quotelike_middle,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    STATE(5396), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6146), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(6144), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6148), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [18462] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4135), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [18490] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6392), 1,
      sym__dollar_ident_zw,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1774), 1,
      sym__var_indirob,
    STATE(4136), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18534] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6396), 1,
      anon_sym_DOLLAR,
    ACTIONS(6398), 1,
      aux_sym__ident_special_token1,
    ACTIONS(6400), 1,
      sym__dollar_ident_zw,
    STATE(50), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1142), 1,
      sym__var_indirob,
    STATE(4137), 1,
      sym_heredoc_content,
    ACTIONS(6394), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1146), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1145), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18578] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4138), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2719), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [18606] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2170), 1,
      sym__var_indirob,
    STATE(4139), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18647] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2255), 1,
      sym__var_indirob,
    STATE(4140), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18688] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6136), 1,
      anon_sym_DOLLAR,
    ACTIONS(6138), 1,
      aux_sym__ident_special_token1,
    STATE(51), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1132), 1,
      sym__var_indirob,
    STATE(4141), 1,
      sym_heredoc_content,
    ACTIONS(6134), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1137), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1141), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18729] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4142), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6402), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6404), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [18756] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2201), 1,
      sym__var_indirob,
    STATE(4143), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18797] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2253), 1,
      sym__var_indirob,
    STATE(4144), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18838] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2168), 1,
      sym__var_indirob,
    STATE(4145), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18879] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1755), 1,
      sym__var_indirob,
    STATE(4146), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18920] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6152), 1,
      anon_sym_DOLLAR,
    ACTIONS(6154), 1,
      aux_sym__ident_special_token1,
    STATE(59), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1372), 1,
      sym__var_indirob,
    STATE(4147), 1,
      sym_heredoc_content,
    ACTIONS(6150), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1365), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1373), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [18961] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3171), 1,
      sym__var_indirob,
    STATE(4148), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19002] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6406), 1,
      sym__NONASSOC,
    STATE(4149), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19031] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2374), 1,
      sym__quotelike_end,
    STATE(4150), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19068] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6412), 1,
      anon_sym_DASH_GT2,
    ACTIONS(6414), 1,
      anon_sym_LBRACK2,
    ACTIONS(6416), 1,
      anon_sym_LBRACE2,
    STATE(4151), 1,
      sym_heredoc_content,
    STATE(5096), 1,
      sym__interp_arrow,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(5902), 5,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
  [19103] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6424), 1,
      anon_sym_RPAREN,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    STATE(4152), 1,
      sym_heredoc_content,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5685), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [19144] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6430), 1,
      anon_sym_LBRACK2,
    ACTIONS(6432), 1,
      anon_sym_LBRACE2,
    STATE(4153), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 8,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [19175] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2261), 1,
      sym__var_indirob,
    STATE(4154), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19216] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1817), 1,
      sym__quotelike_end,
    STATE(4155), 1,
      sym_heredoc_content,
    STATE(4283), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19253] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4156), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6434), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6436), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19280] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6364), 1,
      anon_sym_DOLLAR,
    ACTIONS(6366), 1,
      aux_sym__ident_special_token1,
    STATE(56), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1341), 1,
      sym__var_indirob,
    STATE(4157), 1,
      sym_heredoc_content,
    ACTIONS(6362), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1350), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1349), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19321] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6406), 1,
      sym__NONASSOC,
    STATE(4158), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19350] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2342), 1,
      sym__var_indirob,
    STATE(4159), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19391] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2368), 1,
      sym__var_indirob,
    STATE(4160), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19432] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3272), 1,
      sym__var_indirob,
    STATE(4161), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19473] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6438), 1,
      anon_sym_LBRACK2,
    ACTIONS(6440), 1,
      anon_sym_LBRACE2,
    STATE(4162), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 8,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [19504] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2205), 1,
      sym__quotelike_end,
    STATE(4163), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19541] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2230), 1,
      sym__quotelike_end,
    STATE(4164), 1,
      sym_heredoc_content,
    STATE(4237), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19578] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4165), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19605] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6442), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(4166), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 9,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [19632] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2400), 1,
      sym__quotelike_end,
    STATE(4167), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19669] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6445), 1,
      sym__NONASSOC,
    STATE(4168), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2698), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19698] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4169), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6447), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6449), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19725] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4170), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6434), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6436), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19752] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1869), 1,
      sym__var_indirob,
    STATE(4171), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19793] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4172), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6451), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6453), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [19820] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2302), 1,
      sym__quotelike_end,
    STATE(4173), 1,
      sym_heredoc_content,
    STATE(4175), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19857] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3087), 1,
      sym__var_indirob,
    STATE(4174), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19898] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2360), 1,
      sym__quotelike_end,
    STATE(4175), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [19935] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2149), 1,
      sym__var_indirob,
    STATE(4176), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [19976] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6455), 1,
      anon_sym_RPAREN,
    STATE(4177), 1,
      sym_heredoc_content,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5609), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [20017] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2363), 1,
      sym__quotelike_end,
    STATE(4178), 1,
      sym_heredoc_content,
    STATE(4179), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [20054] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2385), 1,
      sym__quotelike_end,
    STATE(4179), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [20091] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6318), 1,
      anon_sym_DOLLAR,
    ACTIONS(6320), 1,
      aux_sym__ident_special_token1,
    STATE(54), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1290), 1,
      sym__var_indirob,
    STATE(4180), 1,
      sym_heredoc_content,
    ACTIONS(6316), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1288), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1284), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20132] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1774), 1,
      sym__var_indirob,
    STATE(4181), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20173] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6457), 1,
      anon_sym_RPAREN,
    STATE(4152), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4182), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5640), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [20214] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4183), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [20241] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5846), 1,
      anon_sym_DOLLAR,
    ACTIONS(5850), 1,
      aux_sym__ident_special_token1,
    STATE(46), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4051), 1,
      sym__var_indirob_autoquote,
    STATE(4123), 1,
      sym_block,
    STATE(4184), 1,
      sym_heredoc_content,
    STATE(4244), 1,
      sym__indirob,
    STATE(4292), 1,
      sym__var_indirob,
    ACTIONS(5844), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4165), 2,
      sym_scalar,
      sym__ident_special,
  [20286] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2244), 1,
      sym__var_indirob,
    STATE(4185), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20327] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3285), 1,
      sym__var_indirob,
    STATE(4186), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20368] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2375), 1,
      sym__quotelike_end,
    STATE(4167), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4187), 1,
      sym_heredoc_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [20405] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2199), 1,
      sym__var_indirob,
    STATE(4188), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20446] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3453), 1,
      sym__var_indirob,
    STATE(4189), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20487] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1797), 1,
      sym__var_indirob,
    STATE(4190), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20528] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2677), 1,
      sym__var_indirob,
    STATE(4191), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20569] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4192), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2719), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [20596] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2194), 1,
      sym__var_indirob,
    STATE(4193), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20637] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2236), 1,
      sym__quotelike_end,
    STATE(4163), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4194), 1,
      sym_heredoc_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [20674] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6459), 1,
      anon_sym_RPAREN,
    STATE(4195), 1,
      sym_heredoc_content,
    STATE(4253), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5643), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [20715] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2519), 1,
      sym__var_indirob,
    STATE(4196), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20756] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6336), 1,
      anon_sym_DOLLAR,
    ACTIONS(6338), 1,
      aux_sym__ident_special_token1,
    STATE(45), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4197), 1,
      sym_heredoc_content,
    STATE(5286), 1,
      sym__var_indirob,
    ACTIONS(6334), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5164), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(5241), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20797] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2520), 1,
      sym__var_indirob,
    STATE(4198), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20838] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(5810), 1,
      anon_sym_DOLLAR,
    ACTIONS(5814), 1,
      aux_sym__ident_special_token1,
    STATE(44), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4149), 1,
      sym_block,
    STATE(4158), 1,
      sym__var_indirob_autoquote,
    STATE(4199), 1,
      sym_heredoc_content,
    STATE(4331), 1,
      sym__indirob,
    STATE(4370), 1,
      sym__var_indirob,
    ACTIONS(5806), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4325), 2,
      sym_scalar,
      sym__ident_special,
  [20883] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2259), 1,
      sym__var_indirob,
    STATE(4200), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20924] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1796), 1,
      sym__var_indirob,
    STATE(4201), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [20965] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4202), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6461), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6463), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [20992] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4203), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6465), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6467), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21019] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3371), 1,
      sym__quotelike_end,
    STATE(4204), 1,
      sym_heredoc_content,
    STATE(4223), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [21056] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6469), 1,
      anon_sym_RPAREN,
    STATE(4205), 1,
      sym_heredoc_content,
    STATE(4230), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5581), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [21097] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6170), 1,
      anon_sym_DOLLAR,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    STATE(62), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2114), 1,
      sym__var_indirob,
    STATE(4206), 1,
      sym_heredoc_content,
    ACTIONS(6168), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2106), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2105), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21138] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3286), 1,
      sym__var_indirob,
    STATE(4207), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21179] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6471), 1,
      anon_sym_RPAREN,
    STATE(4208), 1,
      sym_heredoc_content,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5688), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [21220] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4209), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6473), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6475), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21247] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6445), 1,
      sym__NONASSOC,
    STATE(4210), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2723), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21276] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6204), 1,
      anon_sym_DOLLAR,
    ACTIONS(6206), 1,
      aux_sym__ident_special_token1,
    STATE(67), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1417), 1,
      sym__var_indirob,
    STATE(4211), 1,
      sym_heredoc_content,
    ACTIONS(6202), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1422), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1416), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21317] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4212), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6477), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6479), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21344] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2234), 1,
      sym__var_indirob,
    STATE(4213), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21385] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4214), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 9,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21412] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2240), 1,
      sym__var_indirob,
    STATE(4215), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21453] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2786), 1,
      sym__var_indirob,
    STATE(4216), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21494] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2212), 1,
      sym__var_indirob,
    STATE(4217), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21535] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2282), 1,
      sym__var_indirob,
    STATE(4218), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21576] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4219), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 9,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21603] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6352), 1,
      anon_sym_DOLLAR,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    STATE(48), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2283), 1,
      sym__var_indirob,
    STATE(4220), 1,
      sym_heredoc_content,
    ACTIONS(6350), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2084), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2083), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21644] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6481), 1,
      anon_sym_RPAREN,
    STATE(4221), 1,
      sym_heredoc_content,
    STATE(4239), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5644), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [21685] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4222), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2713), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2715), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21712] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3396), 1,
      sym__quotelike_end,
    STATE(4223), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [21749] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3397), 1,
      sym__quotelike_end,
    STATE(4224), 1,
      sym_heredoc_content,
    STATE(4225), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [21786] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3404), 1,
      sym__quotelike_end,
    STATE(4225), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [21823] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6396), 1,
      anon_sym_DOLLAR,
    ACTIONS(6398), 1,
      aux_sym__ident_special_token1,
    STATE(50), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1195), 1,
      sym__var_indirob,
    STATE(4226), 1,
      sym_heredoc_content,
    ACTIONS(6394), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1146), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1145), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21864] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4227), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6465), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6467), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21891] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2032), 1,
      sym__var_indirob,
    STATE(4228), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [21932] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4229), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3427), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(3425), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [21959] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6483), 1,
      anon_sym_RPAREN,
    STATE(4230), 1,
      sym_heredoc_content,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5603), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22000] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6212), 1,
      anon_sym_DOLLAR,
    ACTIONS(6214), 1,
      aux_sym__ident_special_token1,
    STATE(68), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1393), 1,
      sym__var_indirob,
    STATE(4231), 1,
      sym_heredoc_content,
    ACTIONS(6210), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1388), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1394), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22041] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6485), 1,
      anon_sym_RPAREN,
    STATE(4232), 1,
      sym_heredoc_content,
    STATE(4248), 1,
      aux_sym__decl_variable_list_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5666), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22082] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3256), 1,
      sym__var_indirob,
    STATE(4233), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22123] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6487), 1,
      anon_sym_RPAREN,
    STATE(4177), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4234), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5598), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22164] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6489), 1,
      anon_sym_DOLLAR,
    ACTIONS(6492), 1,
      anon_sym_AT,
    ACTIONS(6495), 1,
      anon_sym_COMMA,
    ACTIONS(6498), 1,
      anon_sym_RPAREN,
    ACTIONS(6500), 1,
      anon_sym_undef,
    ACTIONS(6503), 1,
      anon_sym_PERCENT2,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(4235), 2,
      sym_heredoc_content,
      aux_sym__decl_variable_list_repeat1,
    STATE(5799), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22203] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6118), 1,
      anon_sym_DOLLAR,
    ACTIONS(6120), 1,
      aux_sym__ident_special_token1,
    STATE(57), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1271), 1,
      sym__var_indirob,
    STATE(4236), 1,
      sym_heredoc_content,
    ACTIONS(6116), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1281), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1279), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22244] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2224), 1,
      sym__quotelike_end,
    STATE(4237), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [22281] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6506), 1,
      anon_sym_RPAREN,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4238), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5647), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22322] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6508), 1,
      anon_sym_RPAREN,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4239), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5672), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22363] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3558), 1,
      sym__var_indirob,
    STATE(4240), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22404] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4241), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6473), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6475), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [22431] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6220), 1,
      anon_sym_DOLLAR,
    ACTIONS(6222), 1,
      aux_sym__ident_special_token1,
    STATE(70), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1299), 1,
      sym__var_indirob,
    STATE(4242), 1,
      sym_heredoc_content,
    ACTIONS(6218), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1302), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1275), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22472] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3257), 1,
      sym__var_indirob,
    STATE(4243), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22513] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4244), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [22540] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2441), 1,
      sym__quotelike_end,
    STATE(4245), 1,
      sym_heredoc_content,
    STATE(4263), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [22577] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4246), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3427), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(3425), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [22604] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2161), 1,
      sym__var_indirob,
    STATE(4247), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22645] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6510), 1,
      anon_sym_RPAREN,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4248), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5583), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22686] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3350), 1,
      sym__var_indirob,
    STATE(4249), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22727] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6160), 1,
      anon_sym_DOLLAR,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    STATE(60), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3351), 1,
      sym__var_indirob,
    STATE(4250), 1,
      sym_heredoc_content,
    ACTIONS(6158), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(3070), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(3068), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22768] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4166), 1,
      aux_sym__term_rightward_repeat1,
    STATE(4251), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4328), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 9,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [22797] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6228), 1,
      anon_sym_DOLLAR,
    ACTIONS(6230), 1,
      aux_sym__ident_special_token1,
    STATE(39), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1258), 1,
      sym__var_indirob,
    STATE(4252), 1,
      sym_heredoc_content,
    ACTIONS(6226), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1245), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1254), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22838] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6512), 1,
      anon_sym_RPAREN,
    STATE(4235), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4253), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5607), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [22879] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4254), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6477), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6479), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [22906] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2256), 1,
      sym__var_indirob,
    STATE(4255), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [22947] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4256), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6447), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6449), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [22974] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4257), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6451), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6453), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [23001] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4258), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6461), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6463), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [23028] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2479), 1,
      sym__var_indirob,
    STATE(4259), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23069] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6236), 1,
      anon_sym_DOLLAR,
    ACTIONS(6238), 1,
      aux_sym__ident_special_token1,
    STATE(71), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1235), 1,
      sym__var_indirob,
    STATE(4260), 1,
      sym_heredoc_content,
    ACTIONS(6234), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1237), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1236), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23110] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2564), 1,
      sym__quotelike_end,
    STATE(4261), 1,
      sym_heredoc_content,
    STATE(4264), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23147] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3096), 1,
      sym__var_indirob,
    STATE(4262), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23188] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2514), 1,
      sym__quotelike_end,
    STATE(4263), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23225] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2590), 1,
      sym__quotelike_end,
    STATE(4264), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23262] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4265), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 9,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      sym__NONASSOC,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [23289] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2591), 1,
      sym__quotelike_end,
    STATE(4266), 1,
      sym_heredoc_content,
    STATE(4267), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23326] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2606), 1,
      sym__quotelike_end,
    STATE(4267), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23363] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1787), 1,
      sym__quotelike_end,
    STATE(4268), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23400] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3247), 1,
      sym__quotelike_end,
    STATE(4269), 1,
      sym_heredoc_content,
    STATE(4272), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23437] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2577), 1,
      sym__quotelike_end,
    STATE(4270), 1,
      sym_heredoc_content,
    STATE(4291), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23474] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6244), 1,
      anon_sym_DOLLAR,
    ACTIONS(6246), 1,
      aux_sym__ident_special_token1,
    STATE(72), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1317), 1,
      sym__var_indirob,
    STATE(4271), 1,
      sym_heredoc_content,
    ACTIONS(6242), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1325), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1324), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23515] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3251), 1,
      sym__quotelike_end,
    STATE(4272), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23552] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6344), 1,
      anon_sym_DOLLAR,
    ACTIONS(6346), 1,
      aux_sym__ident_special_token1,
    STATE(55), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1260), 1,
      sym__var_indirob,
    STATE(4273), 1,
      sym_heredoc_content,
    ACTIONS(6342), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1261), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1262), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23593] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2237), 1,
      sym__var_indirob,
    STATE(4274), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23634] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6128), 1,
      anon_sym_DOLLAR,
    ACTIONS(6130), 1,
      aux_sym__ident_special_token1,
    STATE(58), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1389), 1,
      sym__var_indirob,
    STATE(4275), 1,
      sym_heredoc_content,
    ACTIONS(6126), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1382), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1381), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23675] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2857), 1,
      sym__var_indirob,
    STATE(4276), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23716] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2162), 1,
      sym__var_indirob,
    STATE(4277), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23757] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2241), 1,
      sym__var_indirob,
    STATE(4278), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23798] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6382), 1,
      anon_sym_DOLLAR,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    STATE(47), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2555), 1,
      sym__var_indirob,
    STATE(4279), 1,
      sym_heredoc_content,
    ACTIONS(6380), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2030), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2076), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23839] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2474), 1,
      sym__var_indirob,
    STATE(4280), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23880] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6282), 1,
      anon_sym_DOLLAR,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    STATE(76), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2475), 1,
      sym__var_indirob,
    STATE(4281), 1,
      sym_heredoc_content,
    ACTIONS(6280), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2144), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2143), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [23921] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4282), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6402), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6404), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [23948] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1792), 1,
      sym__quotelike_end,
    STATE(4283), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [23985] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1772), 1,
      sym__var_indirob,
    STATE(4284), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24026] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6252), 1,
      anon_sym_DOLLAR,
    ACTIONS(6254), 1,
      aux_sym__ident_special_token1,
    STATE(73), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1313), 1,
      sym__var_indirob,
    STATE(4285), 1,
      sym_heredoc_content,
    ACTIONS(6250), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1274), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1272), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24067] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3252), 1,
      sym__quotelike_end,
    STATE(4286), 1,
      sym_heredoc_content,
    STATE(4287), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [24104] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3254), 1,
      sym__quotelike_end,
    STATE(4287), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [24141] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4288), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [24168] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6180), 1,
      anon_sym_DOLLAR,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    STATE(63), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2121), 1,
      sym__var_indirob,
    STATE(4289), 1,
      sym_heredoc_content,
    ACTIONS(6178), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2111), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2110), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24209] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6514), 1,
      anon_sym_RPAREN,
    STATE(4238), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4290), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5654), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [24250] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2619), 1,
      sym__quotelike_end,
    STATE(4291), 1,
      sym_heredoc_content,
    STATE(4374), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [24287] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4292), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2682), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2686), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [24314] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4293), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [24341] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3093), 1,
      sym__var_indirob,
    STATE(4294), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24382] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1775), 1,
      sym__quotelike_end,
    STATE(4268), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4295), 1,
      sym_heredoc_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [24419] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6110), 1,
      anon_sym_DOLLAR,
    ACTIONS(6112), 1,
      aux_sym__ident_special_token1,
    STATE(74), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1343), 1,
      sym__var_indirob,
    STATE(4296), 1,
      sym_heredoc_content,
    ACTIONS(6108), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1330), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1328), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24460] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2338), 1,
      sym__quotelike_end,
    STATE(4150), 1,
      aux_sym__interpolated_transliteration_content,
    STATE(4297), 1,
      sym_heredoc_content,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6408), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    ACTIONS(5637), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6410), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [24497] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6328), 1,
      anon_sym_DOLLAR,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    STATE(49), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3094), 1,
      sym__var_indirob,
    STATE(4298), 1,
      sym_heredoc_content,
    ACTIONS(6326), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(2621), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(2598), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24538] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6418), 1,
      anon_sym_DOLLAR,
    ACTIONS(6420), 1,
      anon_sym_AT,
    ACTIONS(6422), 1,
      anon_sym_COMMA,
    ACTIONS(6426), 1,
      anon_sym_undef,
    ACTIONS(6428), 1,
      anon_sym_PERCENT2,
    ACTIONS(6516), 1,
      anon_sym_RPAREN,
    STATE(4208), 1,
      aux_sym__decl_variable_list_repeat1,
    STATE(4299), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5588), 2,
      sym__declared_vars,
      sym_undef_expression,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [24579] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6518), 1,
      anon_sym_LBRACK2,
    ACTIONS(6520), 1,
      anon_sym_LBRACE2,
    STATE(4300), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 8,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [24610] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6522), 1,
      anon_sym_LBRACK2,
    ACTIONS(6524), 1,
      anon_sym_LBRACE2,
    STATE(4301), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 8,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [24641] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5808), 1,
      anon_sym_LBRACE,
    ACTIONS(6304), 1,
      anon_sym_DOLLAR,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    STATE(42), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1771), 1,
      sym__var_indirob,
    STATE(4302), 1,
      sym_heredoc_content,
    ACTIONS(6302), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1768), 2,
      sym__indirob,
      sym__var_indirob_autoquote,
    STATE(1761), 3,
      sym_block,
      sym_scalar,
      sym__ident_special,
  [24682] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4303), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4805), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1456), 2,
      sym_block,
      sym__semicolon,
  [24720] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6532), 1,
      sym_prototype,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1518), 1,
      sym_block,
    STATE(4304), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4899), 1,
      sym_attrlist,
    STATE(5511), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [24762] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6536), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1470), 1,
      sym_block,
    STATE(4305), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4735), 1,
      sym_attrlist,
    STATE(5344), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [24804] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6538), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3481), 1,
      sym_block,
    STATE(4306), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4708), 1,
      sym_attrlist,
    STATE(5440), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [24846] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6540), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3482), 1,
      sym_block,
    STATE(4307), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4710), 1,
      sym_attrlist,
    STATE(5448), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [24888] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6542), 1,
      anon_sym_LBRACK2,
    ACTIONS(6544), 1,
      anon_sym_LBRACE2,
    STATE(4308), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 7,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [24918] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6546), 1,
      anon_sym_LBRACK2,
    ACTIONS(6548), 1,
      anon_sym_LBRACE2,
    STATE(4309), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 7,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [24948] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4310), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6402), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6404), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [24974] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4311), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6434), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6436), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25000] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6550), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3377), 1,
      sym_block,
    STATE(4312), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4912), 1,
      sym_attrlist,
    STATE(5501), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25042] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6552), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3501), 1,
      sym_block,
    STATE(4313), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4914), 1,
      sym_attrlist,
    STATE(5514), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25084] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4314), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3427), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(3425), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25110] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4315), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4951), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1547), 2,
      sym_block,
      sym__semicolon,
  [25148] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6554), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3576), 1,
      sym_block,
    STATE(4316), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4810), 1,
      sym_attrlist,
    STATE(5518), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25190] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6556), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3577), 1,
      sym_block,
    STATE(4317), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4812), 1,
      sym_attrlist,
    STATE(5544), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25232] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4318), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4963), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1549), 2,
      sym_block,
      sym__semicolon,
  [25270] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6558), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3609), 1,
      sym_block,
    STATE(4319), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4851), 1,
      sym_attrlist,
    STATE(5473), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25312] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6560), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3611), 1,
      sym_block,
    STATE(4320), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4972), 1,
      sym_attrlist,
    STATE(5493), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25354] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6562), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2353), 1,
      sym_block,
    STATE(4321), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4944), 1,
      sym_attrlist,
    STATE(5452), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25396] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6564), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2397), 1,
      sym_block,
    STATE(4322), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4952), 1,
      sym_attrlist,
    STATE(5576), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25438] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6566), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3189), 1,
      sym_block,
    STATE(4323), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4949), 1,
      sym_attrlist,
    STATE(5356), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25480] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6568), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3190), 1,
      sym_block,
    STATE(4324), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4953), 1,
      sym_attrlist,
    STATE(5358), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25522] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4325), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2723), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25548] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4326), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2692), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2694), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25574] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6570), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3207), 1,
      sym_block,
    STATE(4327), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4786), 1,
      sym_attrlist,
    STATE(5397), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25616] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6572), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3209), 1,
      sym_block,
    STATE(4328), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4790), 1,
      sym_attrlist,
    STATE(5405), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25658] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4329), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2713), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2715), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25684] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4330), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2717), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2719), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25710] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4331), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2698), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [25736] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6574), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1555), 1,
      sym_block,
    STATE(4332), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4806), 1,
      sym_attrlist,
    STATE(5377), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25778] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6576), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1556), 1,
      sym_block,
    STATE(4333), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4816), 1,
      sym_attrlist,
    STATE(5406), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25820] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6578), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2708), 1,
      sym_block,
    STATE(4334), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4730), 1,
      sym_attrlist,
    STATE(5505), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25862] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6580), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2709), 1,
      sym_block,
    STATE(4335), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4732), 1,
      sym_attrlist,
    STATE(5507), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25904] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6582), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2746), 1,
      sym_block,
    STATE(4336), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4757), 1,
      sym_attrlist,
    STATE(5552), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25946] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6584), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2748), 1,
      sym_block,
    STATE(4337), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4759), 1,
      sym_attrlist,
    STATE(5553), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [25988] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6586), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2407), 1,
      sym_block,
    STATE(4338), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4848), 1,
      sym_attrlist,
    STATE(5359), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26030] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6588), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2438), 1,
      sym_block,
    STATE(4339), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4857), 1,
      sym_attrlist,
    STATE(5368), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26072] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6590), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1567), 1,
      sym_block,
    STATE(4340), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4863), 1,
      sym_attrlist,
    STATE(5381), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26114] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6592), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1568), 1,
      sym_block,
    STATE(4341), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4864), 1,
      sym_attrlist,
    STATE(5384), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26156] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6594), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2888), 1,
      sym_block,
    STATE(4342), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4823), 1,
      sym_attrlist,
    STATE(5332), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26198] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6596), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2889), 1,
      sym_block,
    STATE(4343), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4825), 1,
      sym_attrlist,
    STATE(5453), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26240] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6598), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2928), 1,
      sym_block,
    STATE(4344), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4842), 1,
      sym_attrlist,
    STATE(5337), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26282] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6600), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2930), 1,
      sym_block,
    STATE(4345), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4843), 1,
      sym_attrlist,
    STATE(5340), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26324] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6602), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1578), 1,
      sym_block,
    STATE(4346), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4872), 1,
      sym_attrlist,
    STATE(5421), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26366] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6604), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1579), 1,
      sym_block,
    STATE(4347), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4875), 1,
      sym_attrlist,
    STATE(5423), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26408] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6606), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1586), 1,
      sym_block,
    STATE(4348), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4876), 1,
      sym_attrlist,
    STATE(5429), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26450] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6608), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1587), 1,
      sym_block,
    STATE(4349), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4877), 1,
      sym_attrlist,
    STATE(5436), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26492] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6616), 1,
      sym__quotelike_middle_close_quote,
    STATE(4517), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6613), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    STATE(4350), 2,
      aux_sym__interpolated_transliteration_content,
      sym_heredoc_content,
    ACTIONS(6610), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6618), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [26524] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4351), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4820), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1662), 2,
      sym_block,
      sym__semicolon,
  [26562] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4352), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4886), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1563), 2,
      sym_block,
      sym__semicolon,
  [26600] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4353), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4889), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1565), 2,
      sym_block,
      sym__semicolon,
  [26638] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4354), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4931), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1643), 2,
      sym_block,
      sym__semicolon,
  [26676] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4355), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4958), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1645), 2,
      sym_block,
      sym__semicolon,
  [26714] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6621), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1890), 1,
      sym_block,
    STATE(4356), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4849), 1,
      sym_attrlist,
    STATE(5458), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26756] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6623), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1895), 1,
      sym_block,
    STATE(4357), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4852), 1,
      sym_attrlist,
    STATE(5466), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26798] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4358), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2532), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2530), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [26824] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6625), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2897), 1,
      sym_block,
    STATE(4359), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4807), 1,
      sym_attrlist,
    STATE(5411), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26866] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6627), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2899), 1,
      sym_block,
    STATE(4360), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4809), 1,
      sym_attrlist,
    STATE(5414), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [26908] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4361), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2688), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2690), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [26934] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4362), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4874), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1521), 2,
      sym_block,
      sym__semicolon,
  [26972] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6629), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1651), 1,
      sym_block,
    STATE(4363), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4828), 1,
      sym_attrlist,
    STATE(5434), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27014] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6631), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1652), 1,
      sym_block,
    STATE(4364), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4830), 1,
      sym_attrlist,
    STATE(5438), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27056] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6633), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3022), 1,
      sym_block,
    STATE(4365), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4850), 1,
      sym_attrlist,
    STATE(5461), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27098] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6635), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3024), 1,
      sym_block,
    STATE(4366), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4856), 1,
      sym_attrlist,
    STATE(5474), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27140] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4367), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4884), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1482), 2,
      sym_block,
      sym__semicolon,
  [27178] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6637), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1664), 1,
      sym_block,
    STATE(4368), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4860), 1,
      sym_attrlist,
    STATE(5478), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27220] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6639), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1665), 1,
      sym_block,
    STATE(4369), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4862), 1,
      sym_attrlist,
    STATE(5481), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27262] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4370), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2682), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2686), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH_GT2,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27288] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6641), 1,
      sym__NONASSOC,
    STATE(4371), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2721), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2723), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27316] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6641), 1,
      sym__NONASSOC,
    STATE(4372), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2696), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(2698), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27344] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6643), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(4373), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4879), 1,
      sym_attrlist,
    STATE(5526), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27386] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6616), 1,
      sym__quotelike_end_quote,
    STATE(4492), 1,
      sym__nonvar_interpolation_fallbacks,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6648), 2,
      anon_sym_DOLLAR,
      anon_sym_AT,
    STATE(4374), 2,
      aux_sym__interpolated_transliteration_content,
      sym_heredoc_content,
    ACTIONS(6645), 3,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(6651), 3,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [27418] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6654), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1484), 1,
      sym_block,
    STATE(4375), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4795), 1,
      sym_attrlist,
    STATE(5557), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27460] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6656), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1486), 1,
      sym_block,
    STATE(4376), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4737), 1,
      sym_attrlist,
    STATE(5353), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27502] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6658), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1676), 1,
      sym_block,
    STATE(4377), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4892), 1,
      sym_attrlist,
    STATE(5502), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27544] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6660), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1677), 1,
      sym_block,
    STATE(4378), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4893), 1,
      sym_attrlist,
    STATE(5504), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27586] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6662), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1684), 1,
      sym_block,
    STATE(4379), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4902), 1,
      sym_attrlist,
    STATE(5513), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27628] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6664), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1685), 1,
      sym_block,
    STATE(4380), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4905), 1,
      sym_attrlist,
    STATE(5517), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27670] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6666), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(4381), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4962), 1,
      sym_attrlist,
    STATE(5373), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27712] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4382), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6461), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6463), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27738] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4383), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6465), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6467), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27764] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4384), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6473), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6475), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27790] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4385), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6477), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6479), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27816] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6668), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1510), 1,
      sym_block,
    STATE(4386), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4734), 1,
      sym_attrlist,
    STATE(5390), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27858] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4387), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4791), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1454), 2,
      sym_block,
      sym__semicolon,
  [27896] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6670), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1513), 1,
      sym_block,
    STATE(4388), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4718), 1,
      sym_attrlist,
    STATE(5479), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [27938] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6672), 1,
      sym__identifier,
    STATE(4389), 1,
      sym_heredoc_content,
    STATE(4543), 1,
      sym__signature_varname,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6676), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6674), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [27968] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4390), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6447), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6449), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [27994] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4391), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6451), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6453), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [28020] = 11,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4392), 1,
      sym_heredoc_content,
    STATE(4518), 1,
      sym_attribute,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4782), 1,
      sym_attrlist,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(1660), 2,
      sym_block,
      sym__semicolon,
  [28058] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6678), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1502), 1,
      sym_block,
    STATE(4393), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4898), 1,
      sym_attrlist,
    STATE(5508), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28100] = 13,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6680), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1464), 1,
      sym_block,
    STATE(4394), 1,
      sym_heredoc_content,
    STATE(4539), 1,
      sym_attribute,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4946), 1,
      sym_attrlist,
    STATE(5510), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28142] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4395), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6682), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28165] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4396), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6684), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28188] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6688), 1,
      anon_sym_COLON,
    ACTIONS(6690), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1955), 1,
      sym_block,
    STATE(4397), 1,
      sym_heredoc_content,
    STATE(4697), 1,
      sym_bareword,
    STATE(5445), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28227] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6692), 1,
      anon_sym_COLON,
    ACTIONS(6694), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1964), 1,
      sym_block,
    STATE(4398), 1,
      sym_heredoc_content,
    STATE(4701), 1,
      sym_bareword,
    STATE(5465), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28266] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4399), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2719), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28289] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4400), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6696), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28312] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4401), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5902), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28335] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4402), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6698), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28358] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6643), 1,
      sym_prototype,
    ACTIONS(6700), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(4403), 1,
      sym_heredoc_content,
    STATE(4654), 1,
      sym_bareword,
    STATE(5526), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28397] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6666), 1,
      sym_prototype,
    ACTIONS(6702), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(4404), 1,
      sym_heredoc_content,
    STATE(4663), 1,
      sym_bareword,
    STATE(5373), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28436] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6704), 1,
      anon_sym_COLON,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4405), 1,
      sym_heredoc_content,
    STATE(4671), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1508), 2,
      sym_block,
      sym__semicolon,
  [28471] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4406), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6708), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28494] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4407), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6710), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28517] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4408), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6712), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28540] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6714), 1,
      anon_sym_LBRACK2,
    ACTIONS(6716), 1,
      anon_sym_LBRACE2,
    STATE(4409), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 6,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [28569] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4410), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6718), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28592] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4411), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6447), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6449), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [28617] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4412), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6720), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28640] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4413), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6722), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28663] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4414), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6724), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28686] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4415), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6726), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28709] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6728), 1,
      anon_sym_LBRACK2,
    ACTIONS(6730), 1,
      anon_sym_LBRACE2,
    STATE(4416), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5900), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(5902), 6,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
  [28738] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4417), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6732), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28761] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4418), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6734), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28784] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4419), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6736), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28807] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4420), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3435), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28830] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6738), 1,
      anon_sym_COLON,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4421), 1,
      sym_heredoc_content,
    STATE(4645), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1703), 2,
      sym_block,
      sym__semicolon,
  [28865] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6740), 1,
      anon_sym_COLON,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4422), 1,
      sym_heredoc_content,
    STATE(4704), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1713), 2,
      sym_block,
      sym__semicolon,
  [28900] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1720), 1,
      sym_block,
    STATE(4423), 1,
      sym_heredoc_content,
    STATE(4554), 1,
      sym_attribute,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5078), 1,
      sym_attrlist,
    STATE(5408), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [28939] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4424), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6696), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28962] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4425), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2715), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [28985] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4426), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6744), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29008] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6643), 1,
      sym_prototype,
    ACTIONS(6700), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(4427), 1,
      sym_heredoc_content,
    STATE(4657), 1,
      sym_bareword,
    STATE(5526), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29047] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4428), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6720), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29070] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6688), 1,
      anon_sym_COLON,
    ACTIONS(6690), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1955), 1,
      sym_block,
    STATE(4429), 1,
      sym_heredoc_content,
    STATE(4601), 1,
      sym_bareword,
    STATE(5445), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29109] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6692), 1,
      anon_sym_COLON,
    ACTIONS(6694), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1964), 1,
      sym_block,
    STATE(4430), 1,
      sym_heredoc_content,
    STATE(4602), 1,
      sym_bareword,
    STATE(5465), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29148] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4431), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6402), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6404), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29173] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4432), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2694), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29196] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4433), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6434), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6436), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29221] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6666), 1,
      sym_prototype,
    ACTIONS(6702), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(4434), 1,
      sym_heredoc_content,
    STATE(4658), 1,
      sym_bareword,
    STATE(5373), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29260] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4435), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6722), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29283] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4436), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6724), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29306] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6643), 1,
      sym_prototype,
    ACTIONS(6700), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(4437), 1,
      sym_heredoc_content,
    STATE(4626), 1,
      sym_bareword,
    STATE(5526), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29345] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6666), 1,
      sym_prototype,
    ACTIONS(6702), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(4438), 1,
      sym_heredoc_content,
    STATE(4627), 1,
      sym_bareword,
    STATE(5373), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29384] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4439), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6682), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29407] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4440), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6746), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29430] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4441), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6726), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29453] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4442), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6736), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29476] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1517), 1,
      sym_block,
    STATE(4443), 1,
      sym_heredoc_content,
    STATE(4554), 1,
      sym_attribute,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5137), 1,
      sym_attrlist,
    STATE(5551), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29515] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4444), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6684), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29538] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4445), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6698), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29561] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4446), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2694), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29584] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4447), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6718), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29607] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4448), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2530), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29630] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4449), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3435), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29653] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6692), 1,
      anon_sym_COLON,
    ACTIONS(6694), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1964), 1,
      sym_block,
    STATE(4450), 1,
      sym_heredoc_content,
    STATE(4655), 1,
      sym_bareword,
    STATE(5465), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [29692] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4451), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3427), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(3425), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29717] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4452), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2690), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29740] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4453), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6451), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6453), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29765] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4454), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2530), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29788] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4455), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2690), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29811] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4456), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2723), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29834] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4457), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2698), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29857] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4458), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6461), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6463), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29882] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4459), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6465), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6467), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [29907] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4460), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2719), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29930] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4461), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2715), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29953] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4462), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2723), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [29976] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4463), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6473), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6475), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [30001] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4464), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6477), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6479), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_DASH,
      anon_sym_LBRACK2,
      anon_sym_LBRACE2,
  [30026] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4465), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2698), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30049] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6748), 1,
      anon_sym_COLON,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4466), 1,
      sym_heredoc_content,
    STATE(4606), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1538), 2,
      sym_block,
      sym__semicolon,
  [30084] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6750), 1,
      anon_sym_COLON,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4467), 1,
      sym_heredoc_content,
    STATE(4647), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1539), 2,
      sym_block,
      sym__semicolon,
  [30119] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1540), 1,
      sym_block,
    STATE(4468), 1,
      sym_heredoc_content,
    STATE(4554), 1,
      sym_attribute,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5097), 1,
      sym_attrlist,
    STATE(5444), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [30158] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4469), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6752), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30181] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4470), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6710), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30204] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4471), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5902), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30227] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4472), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6708), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30250] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4473), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6746), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30273] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4474), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5658), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30296] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4475), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6732), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30319] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4476), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5658), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30342] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4477), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2686), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30365] = 12,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6688), 1,
      anon_sym_COLON,
    ACTIONS(6690), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1955), 1,
      sym_block,
    STATE(4478), 1,
      sym_heredoc_content,
    STATE(4638), 1,
      sym_bareword,
    STATE(5445), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [30404] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6672), 1,
      sym__identifier,
    STATE(4479), 1,
      sym_heredoc_content,
    STATE(4543), 1,
      sym__signature_varname,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6754), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6674), 5,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [30433] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6756), 1,
      anon_sym_COLON,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4480), 1,
      sym_heredoc_content,
    STATE(4670), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1503), 2,
      sym_block,
      sym__semicolon,
  [30468] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4481), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6734), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30491] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4482), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6744), 10,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30514] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4483), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2686), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30537] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4484), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6712), 10,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      sym__dollar_in_regexp,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30560] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6758), 1,
      anon_sym_COLON,
    STATE(4485), 1,
      sym_heredoc_content,
    STATE(4494), 1,
      aux_sym_attrlist_repeat1,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3319), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [30592] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4486), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6708), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30614] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4487), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2530), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30636] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6762), 1,
      anon_sym_and,
    STATE(4488), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6764), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(6760), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [30662] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6768), 1,
      anon_sym_DOLLAR,
    ACTIONS(6770), 1,
      anon_sym_AT,
    ACTIONS(6775), 1,
      aux_sym__interpolation_fallbacks_token2,
    ACTIONS(6777), 1,
      sym__no_interp_whitespace_zw,
    STATE(4414), 1,
      sym_scalar,
    STATE(4489), 1,
      sym_heredoc_content,
    ACTIONS(6766), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6773), 2,
      anon_sym_DOLLAR_STAR,
      anon_sym_AT_STAR,
  [30698] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6779), 1,
      anon_sym_DOLLAR,
    ACTIONS(6781), 1,
      anon_sym_AT,
    ACTIONS(6784), 1,
      aux_sym__interpolation_fallbacks_token2,
    ACTIONS(6786), 1,
      sym__no_interp_whitespace_zw,
    STATE(4490), 1,
      sym_heredoc_content,
    STATE(4506), 1,
      sym_scalar,
    ACTIONS(6766), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6773), 2,
      anon_sym_DOLLAR_STAR,
      anon_sym_AT_STAR,
  [30734] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6788), 1,
      anon_sym_DOLLAR,
    ACTIONS(6790), 1,
      anon_sym_AT,
    ACTIONS(6792), 1,
      anon_sym_LPAREN,
    ACTIONS(6794), 1,
      anon_sym_PERCENT2,
    STATE(2274), 1,
      sym__declared_vars,
    STATE(2286), 1,
      sym__decl_variable_list,
    STATE(4491), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(2518), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [30768] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4492), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6796), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30790] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4493), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5902), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30812] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6801), 1,
      anon_sym_COLON,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6798), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4494), 2,
      sym_heredoc_content,
      aux_sym_attrlist_repeat1,
    ACTIONS(3304), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [30842] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4495), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2690), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30864] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6804), 1,
      anon_sym_DOLLAR,
    ACTIONS(6806), 1,
      anon_sym_AT,
    ACTIONS(6808), 1,
      anon_sym_LPAREN,
    ACTIONS(6810), 1,
      anon_sym_PERCENT2,
    STATE(3693), 1,
      sym__declared_vars,
    STATE(3694), 1,
      sym__decl_variable_list,
    STATE(4496), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3339), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [30898] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4497), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6812), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30920] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6814), 1,
      anon_sym_DOLLAR,
    ACTIONS(6816), 1,
      anon_sym_AT,
    ACTIONS(6818), 1,
      anon_sym_LPAREN,
    ACTIONS(6820), 1,
      anon_sym_PERCENT2,
    STATE(3016), 1,
      sym__declared_vars,
    STATE(3017), 1,
      sym__decl_variable_list,
    STATE(4498), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(2847), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [30954] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4499), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6708), 9,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [30976] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6804), 1,
      anon_sym_DOLLAR,
    ACTIONS(6806), 1,
      anon_sym_AT,
    ACTIONS(6808), 1,
      anon_sym_LPAREN,
    ACTIONS(6810), 1,
      anon_sym_PERCENT2,
    STATE(3439), 1,
      sym__declared_vars,
    STATE(3440), 1,
      sym__decl_variable_list,
    STATE(4500), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3339), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [31010] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4501), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6718), 9,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31032] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4502), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6616), 9,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31054] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6822), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(4503), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 6,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [31078] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4504), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6720), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31100] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4505), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6722), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31122] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4506), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6724), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31144] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4507), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6726), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31166] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4508), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6736), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31188] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4509), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2686), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31210] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5864), 1,
      anon_sym_DOLLAR,
    ACTIONS(5866), 1,
      anon_sym_AT,
    ACTIONS(5874), 1,
      anon_sym_LPAREN,
    ACTIONS(5876), 1,
      anon_sym_PERCENT2,
    STATE(1994), 1,
      sym__decl_variable_list,
    STATE(2011), 1,
      sym__declared_vars,
    STATE(4510), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(1815), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [31244] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6762), 1,
      anon_sym_and,
    STATE(4511), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6764), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(6825), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [31270] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4512), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6746), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31292] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4513), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6696), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31314] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6827), 1,
      anon_sym_COLON,
    STATE(4514), 1,
      sym_heredoc_content,
    STATE(4538), 1,
      aux_sym_attrlist_repeat1,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3319), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [31346] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6762), 1,
      anon_sym_and,
    STATE(4515), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 8,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym_or,
      anon_sym_xor,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [31370] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6829), 1,
      anon_sym_DOLLAR,
    ACTIONS(6831), 1,
      anon_sym_AT,
    ACTIONS(6833), 1,
      anon_sym_LPAREN,
    ACTIONS(6835), 1,
      anon_sym_PERCENT2,
    STATE(3344), 1,
      sym__declared_vars,
    STATE(3345), 1,
      sym__decl_variable_list,
    STATE(4516), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(3441), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [31404] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4517), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6796), 9,
      sym__quotelike_middle_close_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31426] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6758), 1,
      anon_sym_COLON,
    STATE(4485), 1,
      aux_sym_attrlist_repeat1,
    STATE(4518), 1,
      sym_heredoc_content,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3313), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [31458] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4519), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6682), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31480] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6837), 1,
      anon_sym_DOLLAR,
    ACTIONS(6839), 1,
      anon_sym_AT,
    ACTIONS(6841), 1,
      anon_sym_LPAREN,
    ACTIONS(6843), 1,
      anon_sym_PERCENT2,
    STATE(2578), 1,
      sym__declared_vars,
    STATE(2582), 1,
      sym__decl_variable_list,
    STATE(4520), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(2583), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [31514] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4521), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3435), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31536] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6845), 1,
      anon_sym_DOLLAR,
    ACTIONS(6847), 1,
      anon_sym_AT,
    ACTIONS(6849), 1,
      anon_sym_LPAREN,
    ACTIONS(6851), 1,
      anon_sym_PERCENT2,
    STATE(2214), 1,
      sym__declared_vars,
    STATE(2215), 1,
      sym__decl_variable_list,
    STATE(4522), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(2217), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [31570] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4523), 1,
      sym_heredoc_content,
    STATE(4960), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1653), 2,
      sym_block,
      sym__semicolon,
  [31602] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4524), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6712), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31624] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4525), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6732), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31646] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6853), 1,
      anon_sym_DOLLAR,
    ACTIONS(6855), 1,
      anon_sym_AT,
    ACTIONS(6858), 1,
      aux_sym__interpolation_fallbacks_token2,
    ACTIONS(6860), 1,
      sym__no_interp_whitespace_zw,
    STATE(4436), 1,
      sym_scalar,
    STATE(4526), 1,
      sym_heredoc_content,
    ACTIONS(6766), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6773), 2,
      anon_sym_DOLLAR_STAR,
      anon_sym_AT_STAR,
  [31682] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4527), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6734), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31704] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6862), 1,
      anon_sym_DOLLAR,
    ACTIONS(6864), 1,
      anon_sym_AT,
    ACTIONS(6867), 1,
      aux_sym__interpolation_fallbacks_token2,
    ACTIONS(6869), 1,
      sym__no_interp_whitespace_zw,
    STATE(4528), 1,
      sym_heredoc_content,
    STATE(4574), 1,
      sym_scalar,
    ACTIONS(6766), 2,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
    ACTIONS(6773), 2,
      anon_sym_DOLLAR_STAR,
      anon_sym_AT_STAR,
  [31740] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4529), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6871), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31762] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4530), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6873), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31784] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4531), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6616), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31806] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4532), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6877), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6875), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [31830] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4533), 1,
      sym_heredoc_content,
    STATE(4752), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1585), 2,
      sym_block,
      sym__semicolon,
  [31862] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4534), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6684), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31884] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4535), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6698), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31906] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4536), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6710), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31928] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4537), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2723), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [31950] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6882), 1,
      anon_sym_COLON,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6879), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4538), 2,
      sym_heredoc_content,
      aux_sym_attrlist_repeat1,
    ACTIONS(3304), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [31980] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6827), 1,
      anon_sym_COLON,
    STATE(4514), 1,
      aux_sym_attrlist_repeat1,
    STATE(4539), 1,
      sym_heredoc_content,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3313), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [32012] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4540), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6885), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6674), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [32036] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4541), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2698), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32058] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4542), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6887), 9,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [32080] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4543), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6889), 9,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [32102] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5864), 1,
      anon_sym_DOLLAR,
    ACTIONS(5866), 1,
      anon_sym_AT,
    ACTIONS(5874), 1,
      anon_sym_LPAREN,
    ACTIONS(5876), 1,
      anon_sym_PERCENT2,
    STATE(1814), 1,
      sym__decl_variable_list,
    STATE(1818), 1,
      sym__declared_vars,
    STATE(4544), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(1815), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [32136] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4545), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2719), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32158] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6762), 1,
      anon_sym_and,
    STATE(4546), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6764), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(6891), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [32184] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4547), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2715), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32206] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6893), 1,
      anon_sym_DOLLAR,
    ACTIONS(6895), 1,
      anon_sym_AT,
    ACTIONS(6897), 1,
      anon_sym_LPAREN,
    ACTIONS(6899), 1,
      anon_sym_PERCENT2,
    STATE(2482), 1,
      sym__declared_vars,
    STATE(2484), 1,
      sym__decl_variable_list,
    STATE(4548), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(2599), 3,
      sym__declare_scalar,
      sym__declare_array,
      sym__declare_hash,
  [32240] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4503), 1,
      aux_sym__term_rightward_repeat1,
    STATE(4549), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4412), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 6,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [32266] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4550), 1,
      sym_heredoc_content,
    STATE(4900), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    STATE(1493), 2,
      sym_block,
      sym__semicolon,
  [32298] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4551), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6901), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32320] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4552), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2694), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32342] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4553), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6718), 9,
      sym__quotelike_end_quote,
      sym__qq_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32364] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6903), 1,
      anon_sym_COLON,
    STATE(4554), 1,
      sym_heredoc_content,
    STATE(4592), 1,
      aux_sym_attrlist_repeat1,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3313), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [32395] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4555), 1,
      sym_heredoc_content,
    STATE(4579), 1,
      aux_sym__term_rightward_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4691), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 5,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [32420] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4556), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6708), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32441] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4557), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6726), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32462] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4558), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6682), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32483] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4559), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6732), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32504] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4560), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6684), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32525] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4561), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6698), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32546] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4562), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6710), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32567] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4563), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6712), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32588] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4564), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6736), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32609] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4565), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6734), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32630] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4566), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2690), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32651] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6905), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(4567), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 5,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [32674] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4568), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2530), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32695] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6908), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(4569), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 5,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [32718] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4570), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6911), 2,
      anon_sym___DATA__,
      anon_sym___END__,
    ACTIONS(6913), 2,
      sym__ctrl_z_hack,
      anon_sym_EOT,
    STATE(1729), 2,
      sym__semicolon,
      sym___DATA__,
  [32745] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6918), 1,
      anon_sym_COLON,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3304), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(6915), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(4571), 2,
      sym_heredoc_content,
      aux_sym_attrlist_repeat1,
  [32774] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6672), 1,
      sym__identifier,
    STATE(4572), 1,
      sym_heredoc_content,
    STATE(4930), 1,
      sym__signature_varname,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6921), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [32799] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6672), 1,
      sym__identifier,
    STATE(4573), 1,
      sym_heredoc_content,
    STATE(4938), 1,
      sym__signature_varname,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6923), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [32824] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4574), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6724), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32845] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4567), 1,
      aux_sym__term_rightward_repeat1,
    STATE(4575), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4895), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 5,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [32870] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4576), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5902), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32891] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4577), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2686), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32912] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4578), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3435), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32933] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6925), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(4579), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
    ACTIONS(3590), 5,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [32956] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4580), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2715), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [32977] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4581), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6928), 2,
      anon_sym___DATA__,
      anon_sym___END__,
    ACTIONS(6930), 2,
      sym__ctrl_z_hack,
      anon_sym_EOT,
    STATE(1622), 2,
      sym__semicolon,
      sym___DATA__,
  [33004] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4582), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(6932), 2,
      anon_sym___DATA__,
      anon_sym___END__,
    ACTIONS(6934), 2,
      sym__ctrl_z_hack,
      anon_sym_EOT,
    STATE(1523), 2,
      sym__semicolon,
      sym___DATA__,
  [33031] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4583), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6936), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33052] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4569), 1,
      aux_sym__term_rightward_repeat1,
    STATE(4584), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4763), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(3568), 5,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [33077] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4585), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6696), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33098] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4586), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6720), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33119] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4587), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2723), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33140] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4588), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6938), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6674), 5,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [33163] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4589), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6940), 3,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_SLASH_SLASH_EQ,
    ACTIONS(6875), 5,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [33186] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4590), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2698), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33207] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4591), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6722), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33228] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6903), 1,
      anon_sym_COLON,
    STATE(4571), 1,
      aux_sym_attrlist_repeat1,
    STATE(4592), 1,
      sym_heredoc_content,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3319), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [33259] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4593), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2694), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33280] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4594), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6746), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33301] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4595), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2719), 8,
      sym_escape_sequence,
      sym__heredoc_middle,
      sym_heredoc_end,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_LBRACK,
      anon_sym_DASH,
  [33322] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4596), 1,
      sym_heredoc_content,
    STATE(4624), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4901), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33350] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6944), 1,
      anon_sym_COLON,
    ACTIONS(6946), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1554), 1,
      sym_block,
    STATE(4597), 1,
      sym_heredoc_content,
    STATE(5556), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33382] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4598), 1,
      sym_heredoc_content,
    STATE(4904), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33410] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6948), 1,
      anon_sym_COLON,
    ACTIONS(6950), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3442), 1,
      sym_block,
    STATE(4599), 1,
      sym_heredoc_content,
    STATE(5572), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33442] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6952), 1,
      anon_sym_COLON,
    ACTIONS(6954), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1445), 1,
      sym_block,
    STATE(4600), 1,
      sym_heredoc_content,
    STATE(5519), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33474] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6956), 1,
      anon_sym_COLON,
    ACTIONS(6958), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1721), 1,
      sym_block,
    STATE(4601), 1,
      sym_heredoc_content,
    STATE(5412), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33506] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6960), 1,
      anon_sym_COLON,
    ACTIONS(6962), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1723), 1,
      sym_block,
    STATE(4602), 1,
      sym_heredoc_content,
    STATE(5416), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33538] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(6966), 1,
      anon_sym_my,
    ACTIONS(6968), 1,
      anon_sym_LPAREN,
    STATE(4603), 1,
      sym_heredoc_content,
    STATE(5748), 1,
      sym_scalar,
    STATE(6043), 1,
      sym__for_initializer,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6970), 2,
      anon_sym_state,
      anon_sym_our,
  [33568] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6540), 1,
      sym_prototype,
    ACTIONS(6972), 1,
      anon_sym_COLON,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3482), 1,
      sym_block,
    STATE(4604), 1,
      sym_heredoc_content,
    STATE(5448), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33600] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6974), 1,
      anon_sym_COLON,
    ACTIONS(6976), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3157), 1,
      sym_block,
    STATE(4605), 1,
      sym_heredoc_content,
    STATE(5335), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33632] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6978), 1,
      anon_sym_COLON,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4606), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1548), 2,
      sym_block,
      sym__semicolon,
  [33660] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4607), 1,
      sym_heredoc_content,
    STATE(4621), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4815), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33688] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4608), 1,
      sym_heredoc_content,
    STATE(4628), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4967), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33716] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6980), 1,
      anon_sym_COLON,
    ACTIONS(6982), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3443), 1,
      sym_block,
    STATE(4609), 1,
      sym_heredoc_content,
    STATE(5490), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33748] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6984), 1,
      anon_sym_COLON,
    ACTIONS(6986), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2663), 1,
      sym_block,
    STATE(4610), 1,
      sym_heredoc_content,
    STATE(5483), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33780] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4611), 1,
      sym_heredoc_content,
    STATE(4880), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33808] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6988), 1,
      anon_sym_COLON,
    ACTIONS(6990), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2664), 1,
      sym_block,
    STATE(4612), 1,
      sym_heredoc_content,
    STATE(5485), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33840] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4613), 1,
      sym_heredoc_content,
    STATE(4625), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4939), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33868] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6094), 1,
      sym__RECOVER_ARROW,
    ACTIONS(6992), 1,
      anon_sym_DOLLAR,
    ACTIONS(6994), 1,
      anon_sym_STAR,
    STATE(2311), 1,
      sym_scalar,
    STATE(2346), 1,
      sym_method,
    STATE(4614), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6074), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [33898] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4615), 1,
      sym_heredoc_content,
    STATE(4661), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4929), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [33926] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6554), 1,
      sym_prototype,
    ACTIONS(6996), 1,
      anon_sym_COLON,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3576), 1,
      sym_block,
    STATE(4616), 1,
      sym_heredoc_content,
    STATE(5518), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33958] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6625), 1,
      sym_prototype,
    ACTIONS(6998), 1,
      anon_sym_COLON,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2897), 1,
      sym_block,
    STATE(4617), 1,
      sym_heredoc_content,
    STATE(5411), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [33990] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6627), 1,
      sym_prototype,
    ACTIONS(7000), 1,
      anon_sym_COLON,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2899), 1,
      sym_block,
    STATE(4618), 1,
      sym_heredoc_content,
    STATE(5414), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34022] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7002), 1,
      anon_sym_COLON,
    ACTIONS(7004), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1649), 1,
      sym_block,
    STATE(4619), 1,
      sym_heredoc_content,
    STATE(5431), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34054] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7006), 1,
      anon_sym_COLON,
    ACTIONS(7008), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1650), 1,
      sym_block,
    STATE(4620), 1,
      sym_heredoc_content,
    STATE(5432), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34086] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4621), 1,
      sym_heredoc_content,
    STATE(4845), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34114] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6556), 1,
      sym_prototype,
    ACTIONS(7010), 1,
      anon_sym_COLON,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3577), 1,
      sym_block,
    STATE(4622), 1,
      sym_heredoc_content,
    STATE(5544), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34146] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4623), 1,
      sym_heredoc_content,
    STATE(4639), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4859), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34174] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4624), 1,
      sym_heredoc_content,
    STATE(4910), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34202] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4625), 1,
      sym_heredoc_content,
    STATE(4841), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34230] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7012), 1,
      anon_sym_COLON,
    ACTIONS(7014), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1657), 1,
      sym_block,
    STATE(4626), 1,
      sym_heredoc_content,
    STATE(5469), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34262] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7016), 1,
      anon_sym_COLON,
    ACTIONS(7018), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1658), 1,
      sym_block,
    STATE(4627), 1,
      sym_heredoc_content,
    STATE(5476), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34294] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4628), 1,
      sym_heredoc_content,
    STATE(4853), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34322] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6690), 1,
      sym_prototype,
    ACTIONS(7020), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1955), 1,
      sym_block,
    STATE(4629), 1,
      sym_heredoc_content,
    STATE(5445), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34354] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7022), 1,
      anon_sym_COLON,
    ACTIONS(7024), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1666), 1,
      sym_block,
    STATE(4630), 1,
      sym_heredoc_content,
    STATE(5482), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34386] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7026), 1,
      anon_sym_COLON,
    ACTIONS(7028), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1667), 1,
      sym_block,
    STATE(4631), 1,
      sym_heredoc_content,
    STATE(5484), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34418] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4632), 1,
      sym_heredoc_content,
    STATE(4640), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4775), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34446] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7030), 1,
      anon_sym_COLON,
    ACTIONS(7032), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1463), 1,
      sym_block,
    STATE(4633), 1,
      sym_heredoc_content,
    STATE(5345), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34478] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6578), 1,
      sym_prototype,
    ACTIONS(7034), 1,
      anon_sym_COLON,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2708), 1,
      sym_block,
    STATE(4634), 1,
      sym_heredoc_content,
    STATE(5505), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34510] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6580), 1,
      sym_prototype,
    ACTIONS(7036), 1,
      anon_sym_COLON,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2709), 1,
      sym_block,
    STATE(4635), 1,
      sym_heredoc_content,
    STATE(5507), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34542] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6050), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7038), 1,
      anon_sym_DOLLAR,
    ACTIONS(7040), 1,
      anon_sym_STAR,
    STATE(2273), 1,
      sym_scalar,
    STATE(2321), 1,
      sym_method,
    STATE(4636), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6030), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [34572] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6694), 1,
      sym_prototype,
    ACTIONS(7042), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1964), 1,
      sym_block,
    STATE(4637), 1,
      sym_heredoc_content,
    STATE(5465), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34604] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7044), 1,
      anon_sym_COLON,
    ACTIONS(7046), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1481), 1,
      sym_block,
    STATE(4638), 1,
      sym_heredoc_content,
    STATE(5443), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34636] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4639), 1,
      sym_heredoc_content,
    STATE(4895), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34664] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4640), 1,
      sym_heredoc_content,
    STATE(4858), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34692] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4641), 1,
      sym_heredoc_content,
    STATE(4660), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4725), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34720] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7048), 1,
      anon_sym_COLON,
    ACTIONS(7050), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2524), 1,
      sym_block,
    STATE(4642), 1,
      sym_heredoc_content,
    STATE(5537), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34752] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4643), 1,
      sym_heredoc_content,
    STATE(4651), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4956), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34780] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4644), 1,
      sym_heredoc_content,
    STATE(4648), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4821), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34808] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7052), 1,
      anon_sym_COLON,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4645), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1644), 2,
      sym_block,
      sym__semicolon,
  [34836] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7054), 1,
      anon_sym_COLON,
    ACTIONS(7056), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2536), 1,
      sym_block,
    STATE(4646), 1,
      sym_heredoc_content,
    STATE(5543), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [34868] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7058), 1,
      anon_sym_COLON,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4647), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1550), 2,
      sym_block,
      sym__semicolon,
  [34896] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4648), 1,
      sym_heredoc_content,
    STATE(4835), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34924] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4649), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3331), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3333), 4,
      sym__attribute_value_begin,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [34946] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4650), 1,
      sym_heredoc_content,
    STATE(4656), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4724), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [34974] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4651), 1,
      sym_heredoc_content,
    STATE(4885), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35002] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7060), 1,
      anon_sym_COLON,
    ACTIONS(7062), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3158), 1,
      sym_block,
    STATE(4652), 1,
      sym_heredoc_content,
    STATE(5336), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35034] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6072), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7064), 1,
      anon_sym_DOLLAR,
    ACTIONS(7066), 1,
      anon_sym_STAR,
    STATE(1782), 1,
      sym_scalar,
    STATE(1789), 1,
      sym_method,
    STATE(4653), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6052), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [35064] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7068), 1,
      anon_sym_COLON,
    ACTIONS(7070), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1560), 1,
      sym_block,
    STATE(4654), 1,
      sym_heredoc_content,
    STATE(5362), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35096] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7072), 1,
      anon_sym_COLON,
    ACTIONS(7074), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1491), 1,
      sym_block,
    STATE(4655), 1,
      sym_heredoc_content,
    STATE(5374), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35128] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4656), 1,
      sym_heredoc_content,
    STATE(4776), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35156] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7076), 1,
      anon_sym_COLON,
    ACTIONS(7078), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1497), 1,
      sym_block,
    STATE(4657), 1,
      sym_heredoc_content,
    STATE(5463), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35188] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7080), 1,
      anon_sym_COLON,
    ACTIONS(7082), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1443), 1,
      sym_block,
    STATE(4658), 1,
      sym_heredoc_content,
    STATE(5468), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35220] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4598), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4659), 1,
      sym_heredoc_content,
    STATE(4882), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35248] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4660), 1,
      sym_heredoc_content,
    STATE(4827), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35276] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4661), 1,
      sym_heredoc_content,
    STATE(4934), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35304] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4662), 1,
      sym_heredoc_content,
    STATE(4665), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4741), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35332] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7084), 1,
      anon_sym_COLON,
    ACTIONS(7086), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1561), 1,
      sym_block,
    STATE(4663), 1,
      sym_heredoc_content,
    STATE(5372), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35364] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7088), 1,
      anon_sym_COLON,
    ACTIONS(7090), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2848), 1,
      sym_block,
    STATE(4664), 1,
      sym_heredoc_content,
    STATE(5333), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35396] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4665), 1,
      sym_heredoc_content,
    STATE(4754), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35424] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6643), 1,
      sym_prototype,
    ACTIONS(7092), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(4666), 1,
      sym_heredoc_content,
    STATE(5526), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35456] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6666), 1,
      sym_prototype,
    ACTIONS(7094), 1,
      anon_sym_COLON,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(4667), 1,
      sym_heredoc_content,
    STATE(5373), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35488] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4668), 1,
      sym_heredoc_content,
    STATE(4669), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4831), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35516] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4669), 1,
      sym_heredoc_content,
    STATE(4839), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35544] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7096), 1,
      anon_sym_COLON,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4670), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1455), 2,
      sym_block,
      sym__semicolon,
  [35572] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7098), 1,
      anon_sym_COLON,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4671), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1457), 2,
      sym_block,
      sym__semicolon,
  [35600] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5952), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7100), 1,
      anon_sym_DOLLAR,
    ACTIONS(7102), 1,
      anon_sym_STAR,
    STATE(2393), 1,
      sym_scalar,
    STATE(2453), 1,
      sym_method,
    STATE(4672), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5932), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [35630] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7104), 1,
      anon_sym_COLON,
    ACTIONS(7106), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2849), 1,
      sym_block,
    STATE(4673), 1,
      sym_heredoc_content,
    STATE(5343), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35662] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5898), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7108), 1,
      anon_sym_DOLLAR,
    ACTIONS(7110), 1,
      anon_sym_STAR,
    STATE(3360), 1,
      sym_scalar,
    STATE(3388), 1,
      sym_method,
    STATE(4674), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5878), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [35692] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6538), 1,
      sym_prototype,
    ACTIONS(7112), 1,
      anon_sym_COLON,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3481), 1,
      sym_block,
    STATE(4675), 1,
      sym_heredoc_content,
    STATE(5440), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35724] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5930), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7114), 1,
      anon_sym_DOLLAR,
    ACTIONS(7116), 1,
      anon_sym_STAR,
    STATE(3248), 1,
      sym_method,
    STATE(3293), 1,
      sym_scalar,
    STATE(4676), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5910), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [35754] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4677), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3331), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3333), 4,
      sym__PERLY_SEMICOLON,
      sym__attribute_value_begin,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [35776] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7118), 1,
      anon_sym_COLON,
    ACTIONS(7120), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1569), 1,
      sym_block,
    STATE(4678), 1,
      sym_heredoc_content,
    STATE(5389), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35808] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7122), 1,
      anon_sym_COLON,
    ACTIONS(7124), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1570), 1,
      sym_block,
    STATE(4679), 1,
      sym_heredoc_content,
    STATE(5392), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35840] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7126), 1,
      sym__attribute_value_begin,
    STATE(4680), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3335), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3337), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [35864] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4681), 1,
      sym_heredoc_content,
    STATE(4686), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4765), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35892] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6014), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7128), 1,
      anon_sym_DOLLAR,
    ACTIONS(7130), 1,
      anon_sym_STAR,
    STATE(2221), 1,
      sym_method,
    STATE(2223), 1,
      sym_scalar,
    STATE(4682), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5994), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [35922] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4611), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4683), 1,
      sym_heredoc_content,
    STATE(4738), 1,
      sym__quotelike_middle,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [35950] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6594), 1,
      sym_prototype,
    ACTIONS(7132), 1,
      anon_sym_COLON,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2888), 1,
      sym_block,
    STATE(4684), 1,
      sym_heredoc_content,
    STATE(5332), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [35982] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6596), 1,
      sym_prototype,
    ACTIONS(7134), 1,
      anon_sym_COLON,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2889), 1,
      sym_block,
    STATE(4685), 1,
      sym_heredoc_content,
    STATE(5453), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36014] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5593), 1,
      sym__quotelike_middle_close_quote,
    STATE(4686), 1,
      sym_heredoc_content,
    STATE(4768), 1,
      sym__quotelike_middle,
    STATE(5030), 1,
      aux_sym__noninterpolated_string_content,
    STATE(5426), 1,
      sym__quotelike_middle_close,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6942), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36042] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6566), 1,
      sym_prototype,
    ACTIONS(7136), 1,
      anon_sym_COLON,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3189), 1,
      sym_block,
    STATE(4687), 1,
      sym_heredoc_content,
    STATE(5356), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36074] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7138), 1,
      anon_sym_COLON,
    ACTIONS(7140), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1444), 1,
      sym_block,
    STATE(4688), 1,
      sym_heredoc_content,
    STATE(5515), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36106] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6562), 1,
      sym_prototype,
    ACTIONS(7142), 1,
      anon_sym_COLON,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2353), 1,
      sym_block,
    STATE(4689), 1,
      sym_heredoc_content,
    STATE(5452), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36138] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7144), 1,
      sym__attribute_value_begin,
    STATE(4690), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3335), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3337), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [36162] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6564), 1,
      sym_prototype,
    ACTIONS(7146), 1,
      anon_sym_COLON,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2397), 1,
      sym_block,
    STATE(4691), 1,
      sym_heredoc_content,
    STATE(5576), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36194] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7148), 1,
      anon_sym_COLON,
    ACTIONS(7150), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1553), 1,
      sym_block,
    STATE(4692), 1,
      sym_heredoc_content,
    STATE(5554), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36226] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(6568), 1,
      sym_prototype,
    ACTIONS(7152), 1,
      anon_sym_COLON,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3190), 1,
      sym_block,
    STATE(4693), 1,
      sym_heredoc_content,
    STATE(5358), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36258] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7154), 1,
      anon_sym_COLON,
    ACTIONS(7156), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3544), 1,
      sym_block,
    STATE(4694), 1,
      sym_heredoc_content,
    STATE(5454), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36290] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7158), 1,
      anon_sym_COLON,
    ACTIONS(7160), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3545), 1,
      sym_block,
    STATE(4695), 1,
      sym_heredoc_content,
    STATE(5462), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36322] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7162), 1,
      anon_sym_COLON,
    ACTIONS(7164), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2756), 1,
      sym_block,
    STATE(4696), 1,
      sym_heredoc_content,
    STATE(5385), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36354] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7166), 1,
      anon_sym_COLON,
    ACTIONS(7168), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1541), 1,
      sym_block,
    STATE(4697), 1,
      sym_heredoc_content,
    STATE(5459), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36386] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5976), 1,
      sym__RECOVER_ARROW,
    ACTIONS(7170), 1,
      anon_sym_DOLLAR,
    ACTIONS(7172), 1,
      anon_sym_STAR,
    STATE(2505), 1,
      sym_scalar,
    STATE(2579), 1,
      sym_method,
    STATE(4698), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5956), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [36416] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7174), 1,
      anon_sym_COLON,
    ACTIONS(7176), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2757), 1,
      sym_block,
    STATE(4699), 1,
      sym_heredoc_content,
    STATE(5386), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36448] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7178), 1,
      anon_sym_COLON,
    ACTIONS(7180), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1462), 1,
      sym_block,
    STATE(4700), 1,
      sym_heredoc_content,
    STATE(5568), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36480] = 10,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7182), 1,
      anon_sym_COLON,
    ACTIONS(7184), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1542), 1,
      sym_block,
    STATE(4701), 1,
      sym_heredoc_content,
    STATE(5472), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36512] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(6966), 1,
      anon_sym_my,
    ACTIONS(7186), 1,
      anon_sym_LPAREN,
    STATE(4702), 1,
      sym_heredoc_content,
    STATE(5748), 1,
      sym_scalar,
    STATE(6032), 1,
      sym__for_initializer,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6970), 2,
      anon_sym_state,
      anon_sym_our,
  [36542] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(6966), 1,
      anon_sym_my,
    ACTIONS(7188), 1,
      anon_sym_LPAREN,
    STATE(4703), 1,
      sym_heredoc_content,
    STATE(5748), 1,
      sym_scalar,
    STATE(6039), 1,
      sym__for_initializer,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6970), 2,
      anon_sym_state,
      anon_sym_our,
  [36572] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7190), 1,
      anon_sym_COLON,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4704), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1646), 2,
      sym_block,
      sym__semicolon,
  [36600] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3356), 1,
      sym__quotelike_end,
    STATE(4705), 1,
      sym_heredoc_content,
    STATE(4957), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36625] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2437), 1,
      sym__quotelike_end,
    STATE(4706), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36650] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(4707), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 5,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [36671] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7196), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3402), 1,
      sym_block,
    STATE(4708), 1,
      sym_heredoc_content,
    STATE(5470), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36700] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2476), 1,
      sym__quotelike_end,
    STATE(4709), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36725] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7198), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3502), 1,
      sym_block,
    STATE(4710), 1,
      sym_heredoc_content,
    STATE(5475), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36754] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4711), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3409), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3411), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [36775] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3404), 1,
      sym__quotelike_end,
    STATE(4712), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36800] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2715), 1,
      sym__quotelike_end,
    STATE(4713), 1,
      sym_heredoc_content,
    STATE(4818), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36825] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5607), 1,
      sym__quotelike_end_quote,
    STATE(2227), 1,
      sym__quotelike_end,
    STATE(4714), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36850] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7202), 1,
      anon_sym_GT,
    STATE(4715), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5993), 2,
      sym_scalar,
      sym_bareword,
  [36875] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2266), 1,
      sym__quotelike_end,
    STATE(4716), 1,
      sym_heredoc_content,
    STATE(4832), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36900] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2718), 1,
      sym__quotelike_end,
    STATE(4717), 1,
      sym_heredoc_content,
    STATE(4837), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [36925] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7204), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1474), 1,
      sym_block,
    STATE(4718), 1,
      sym_heredoc_content,
    STATE(5376), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [36954] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4719), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3304), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
    ACTIONS(3309), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [36975] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5625), 1,
      sym__quotelike_end_quote,
    STATE(2269), 1,
      sym__quotelike_end,
    STATE(4720), 1,
      sym_heredoc_content,
    STATE(4733), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37000] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2193), 1,
      sym__quotelike_end,
    STATE(4721), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37025] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2271), 1,
      sym__quotelike_end,
    STATE(4722), 1,
      sym_heredoc_content,
    STATE(4740), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37050] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5609), 1,
      sym__quotelike_end_quote,
    STATE(2216), 1,
      sym__quotelike_end,
    STATE(4721), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4723), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37075] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2440), 1,
      sym__quotelike_end,
    STATE(4724), 1,
      sym_heredoc_content,
    STATE(4749), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37100] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2231), 1,
      sym__quotelike_end,
    STATE(4725), 1,
      sym_heredoc_content,
    STATE(4822), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37125] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2265), 1,
      sym__quotelike_end,
    STATE(4726), 1,
      sym_heredoc_content,
    STATE(4873), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37150] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3197), 1,
      sym__quotelike_end,
    STATE(4727), 1,
      sym_heredoc_content,
    STATE(4878), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37175] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1908), 1,
      sym__quotelike_end,
    STATE(4728), 1,
      sym_heredoc_content,
    STATE(4758), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37200] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5633), 1,
      sym__quotelike_end_quote,
    STATE(3244), 1,
      sym__quotelike_end,
    STATE(4729), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37225] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7206), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2747), 1,
      sym_block,
    STATE(4730), 1,
      sym_heredoc_content,
    STATE(5545), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37254] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3468), 1,
      sym__quotelike_end,
    STATE(4731), 1,
      sym_heredoc_content,
    STATE(4918), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37279] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7208), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2749), 1,
      sym_block,
    STATE(4732), 1,
      sym_heredoc_content,
    STATE(5546), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37308] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5625), 1,
      sym__quotelike_end_quote,
    STATE(2289), 1,
      sym__quotelike_end,
    STATE(4733), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37333] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7210), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1449), 1,
      sym_block,
    STATE(4734), 1,
      sym_heredoc_content,
    STATE(5361), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37362] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7212), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1459), 1,
      sym_block,
    STATE(4735), 1,
      sym_heredoc_content,
    STATE(5533), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37391] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5619), 1,
      sym__quotelike_end_quote,
    STATE(1806), 1,
      sym__quotelike_end,
    STATE(4736), 1,
      sym_heredoc_content,
    STATE(4742), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37416] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7214), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1452), 1,
      sym_block,
    STATE(4737), 1,
      sym_heredoc_content,
    STATE(5354), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37445] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1800), 1,
      sym__quotelike_end,
    STATE(4738), 1,
      sym_heredoc_content,
    STATE(4871), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37470] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1957), 1,
      sym__quotelike_end,
    STATE(4739), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37495] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5623), 1,
      sym__quotelike_end_quote,
    STATE(2290), 1,
      sym__quotelike_end,
    STATE(4740), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37520] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2301), 1,
      sym__quotelike_end,
    STATE(4741), 1,
      sym_heredoc_content,
    STATE(4750), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37545] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5619), 1,
      sym__quotelike_end_quote,
    STATE(1821), 1,
      sym__quotelike_end,
    STATE(4742), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37570] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5617), 1,
      sym__quotelike_end_quote,
    STATE(2278), 1,
      sym__quotelike_end,
    STATE(4743), 1,
      sym_heredoc_content,
    STATE(4955), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37595] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1878), 1,
      sym__quotelike_end,
    STATE(4744), 1,
      sym_heredoc_content,
    STATE(4922), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37620] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1820), 1,
      sym__quotelike_end,
    STATE(4745), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37645] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3245), 1,
      sym__quotelike_end,
    STATE(4746), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37670] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7216), 1,
      anon_sym_GT,
    STATE(4747), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5772), 2,
      sym_scalar,
      sym_bareword,
  [37695] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7218), 1,
      anon_sym_COLON,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1718), 1,
      sym_block,
    STATE(4748), 1,
      sym_heredoc_content,
    STATE(5382), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37724] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2492), 1,
      sym__quotelike_end,
    STATE(4749), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37749] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2354), 1,
      sym__quotelike_end,
    STATE(4750), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37774] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3570), 1,
      sym__quotelike_end,
    STATE(4751), 1,
      sym_heredoc_content,
    STATE(4920), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37799] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4752), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1546), 2,
      sym_block,
      sym__semicolon,
  [37824] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7220), 1,
      anon_sym_GT,
    STATE(4753), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5858), 2,
      sym_scalar,
      sym_bareword,
  [37849] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2359), 1,
      sym__quotelike_end,
    STATE(4754), 1,
      sym_heredoc_content,
    STATE(4760), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37874] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7222), 1,
      anon_sym_COLON,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1624), 1,
      sym_block,
    STATE(4755), 1,
      sym_heredoc_content,
    STATE(5497), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37903] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2868), 1,
      sym__quotelike_end,
    STATE(4756), 1,
      sym_heredoc_content,
    STATE(4887), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37928] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7224), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2800), 1,
      sym_block,
    STATE(4757), 1,
      sym_heredoc_content,
    STATE(5570), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [37957] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1960), 1,
      sym__quotelike_end,
    STATE(4758), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [37982] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7226), 1,
      sym_prototype,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2802), 1,
      sym_block,
    STATE(4759), 1,
      sym_heredoc_content,
    STATE(5571), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [38011] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5689), 1,
      sym__quotelike_end_quote,
    STATE(2384), 1,
      sym__quotelike_end,
    STATE(4760), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38036] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2623), 1,
      sym__quotelike_end,
    STATE(4761), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38061] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2933), 1,
      sym__quotelike_end,
    STATE(4762), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38086] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2559), 1,
      sym__quotelike_end,
    STATE(4763), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38111] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2935), 1,
      sym__quotelike_end,
    STATE(4764), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38136] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2564), 1,
      sym__quotelike_end,
    STATE(4765), 1,
      sym_heredoc_content,
    STATE(4766), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38161] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2590), 1,
      sym__quotelike_end,
    STATE(4766), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38186] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2456), 1,
      sym__quotelike_end,
    STATE(4767), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38211] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2591), 1,
      sym__quotelike_end,
    STATE(4768), 1,
      sym_heredoc_content,
    STATE(4769), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38236] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2606), 1,
      sym__quotelike_end,
    STATE(4769), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38261] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4770), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3401), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3403), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [38282] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(234), 1,
      sym_package,
    STATE(4771), 1,
      sym_heredoc_content,
    STATE(5360), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    ACTIONS(7228), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [38307] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3582), 1,
      sym__quotelike_end,
    STATE(4772), 1,
      sym_heredoc_content,
    STATE(4950), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38332] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7230), 1,
      anon_sym_GT,
    STATE(4773), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5764), 2,
      sym_scalar,
      sym_bareword,
  [38357] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3183), 1,
      sym__quotelike_end,
    STATE(4774), 1,
      sym_heredoc_content,
    STATE(4861), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38382] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3246), 1,
      sym__quotelike_end,
    STATE(4775), 1,
      sym_heredoc_content,
    STATE(4854), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38407] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2509), 1,
      sym__quotelike_end,
    STATE(4776), 1,
      sym_heredoc_content,
    STATE(4792), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38432] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2939), 1,
      sym__quotelike_end,
    STATE(4777), 1,
      sym_heredoc_content,
    STATE(4891), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38457] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3584), 1,
      sym__quotelike_end,
    STATE(4778), 1,
      sym_heredoc_content,
    STATE(4959), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38482] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2306), 1,
      sym__quotelike_end,
    STATE(4779), 1,
      sym_heredoc_content,
    STATE(4894), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38507] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2941), 1,
      sym__quotelike_end,
    STATE(4780), 1,
      sym_heredoc_content,
    STATE(4897), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38532] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5635), 1,
      sym__quotelike_end_quote,
    STATE(2307), 1,
      sym__quotelike_end,
    STATE(4781), 1,
      sym_heredoc_content,
    STATE(4811), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38557] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4782), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1678), 2,
      sym_block,
      sym__semicolon,
  [38582] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2308), 1,
      sym__quotelike_end,
    STATE(4783), 1,
      sym_heredoc_content,
    STATE(4813), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38607] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5613), 1,
      sym__quotelike_end_quote,
    STATE(3357), 1,
      sym__quotelike_end,
    STATE(4784), 1,
      sym_heredoc_content,
    STATE(4814), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38632] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3290), 1,
      sym__quotelike_end,
    STATE(4785), 1,
      sym_heredoc_content,
    STATE(4926), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38657] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7232), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3031), 1,
      sym_block,
    STATE(4786), 1,
      sym_heredoc_content,
    STATE(5413), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [38686] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3358), 1,
      sym__quotelike_end,
    STATE(4787), 1,
      sym_heredoc_content,
    STATE(4819), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38711] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(243), 1,
      sym_package,
    STATE(4788), 1,
      sym_heredoc_content,
    STATE(5391), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    ACTIONS(7228), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [38736] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2877), 1,
      sym__quotelike_end,
    STATE(4761), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4789), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38761] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7234), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2625), 1,
      sym_block,
    STATE(4790), 1,
      sym_heredoc_content,
    STATE(5415), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [38790] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4791), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1476), 2,
      sym_block,
      sym__semicolon,
  [38815] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5671), 1,
      sym__quotelike_end_quote,
    STATE(2618), 1,
      sym__quotelike_end,
    STATE(4792), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38840] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4793), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6498), 6,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_undef,
      anon_sym_PERCENT2,
  [38859] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4794), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5748), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [38878] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7236), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1489), 1,
      sym_block,
    STATE(4795), 1,
      sym_heredoc_content,
    STATE(5548), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [38907] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1864), 1,
      sym__quotelike_end,
    STATE(4739), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4796), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38932] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4797), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7238), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [38951] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2894), 1,
      sym__quotelike_end,
    STATE(4762), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4798), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [38976] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2608), 1,
      sym__quotelike_end,
    STATE(4799), 1,
      sym_heredoc_content,
    STATE(4966), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39001] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2739), 1,
      sym__quotelike_end,
    STATE(4800), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39026] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2495), 1,
      sym__quotelike_end,
    STATE(4763), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4801), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39051] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5693), 1,
      sym__quotelike_end_quote,
    STATE(2896), 1,
      sym__quotelike_end,
    STATE(4764), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4802), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39076] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5629), 1,
      sym__quotelike_end_quote,
    STATE(2501), 1,
      sym__quotelike_end,
    STATE(4803), 1,
      sym_heredoc_content,
    STATE(4826), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39101] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2503), 1,
      sym__quotelike_end,
    STATE(4804), 1,
      sym_heredoc_content,
    STATE(4829), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39126] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4805), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1442), 2,
      sym_block,
      sym__semicolon,
  [39151] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7240), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1572), 1,
      sym_block,
    STATE(4806), 1,
      sym_heredoc_content,
    STATE(5398), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39180] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7242), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3023), 1,
      sym_block,
    STATE(4807), 1,
      sym_heredoc_content,
    STATE(5433), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39209] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5615), 1,
      sym__quotelike_end_quote,
    STATE(2319), 1,
      sym__quotelike_end,
    STATE(4706), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4808), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39234] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7244), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3025), 1,
      sym_block,
    STATE(4809), 1,
      sym_heredoc_content,
    STATE(5435), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39263] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7246), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3610), 1,
      sym_block,
    STATE(4810), 1,
      sym_heredoc_content,
    STATE(5355), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39292] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5635), 1,
      sym__quotelike_end_quote,
    STATE(2333), 1,
      sym__quotelike_end,
    STATE(4811), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39317] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7248), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3612), 1,
      sym_block,
    STATE(4812), 1,
      sym_heredoc_content,
    STATE(5363), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39346] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5631), 1,
      sym__quotelike_end_quote,
    STATE(2336), 1,
      sym__quotelike_end,
    STATE(4813), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39371] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5613), 1,
      sym__quotelike_end_quote,
    STATE(3368), 1,
      sym__quotelike_end,
    STATE(4814), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39396] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2337), 1,
      sym__quotelike_end,
    STATE(4815), 1,
      sym_heredoc_content,
    STATE(4844), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39421] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7250), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1574), 1,
      sym_block,
    STATE(4816), 1,
      sym_heredoc_content,
    STATE(5399), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39450] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3491), 1,
      sym__quotelike_end,
    STATE(4817), 1,
      sym_heredoc_content,
    STATE(4927), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39475] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2751), 1,
      sym__quotelike_end,
    STATE(4818), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39500] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5611), 1,
      sym__quotelike_end_quote,
    STATE(3369), 1,
      sym__quotelike_end,
    STATE(4819), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39525] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4820), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1679), 2,
      sym_block,
      sym__semicolon,
  [39550] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3370), 1,
      sym__quotelike_end,
    STATE(4821), 1,
      sym_heredoc_content,
    STATE(4834), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39575] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2195), 1,
      sym__quotelike_end,
    STATE(4822), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39600] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7252), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2929), 1,
      sym_block,
    STATE(4823), 1,
      sym_heredoc_content,
    STATE(5563), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39629] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1780), 1,
      sym__quotelike_end,
    STATE(4824), 1,
      sym_heredoc_content,
    STATE(4833), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39654] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7254), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2931), 1,
      sym_block,
    STATE(4825), 1,
      sym_heredoc_content,
    STATE(5569), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39683] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5629), 1,
      sym__quotelike_end_quote,
    STATE(2560), 1,
      sym__quotelike_end,
    STATE(4826), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39708] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2242), 1,
      sym__quotelike_end,
    STATE(4827), 1,
      sym_heredoc_content,
    STATE(4870), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39733] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7256), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1669), 1,
      sym_block,
    STATE(4828), 1,
      sym_heredoc_content,
    STATE(5486), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39762] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5627), 1,
      sym__quotelike_end_quote,
    STATE(2561), 1,
      sym__quotelike_end,
    STATE(4829), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39787] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7258), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1671), 1,
      sym_block,
    STATE(4830), 1,
      sym_heredoc_content,
    STATE(5487), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [39816] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2563), 1,
      sym__quotelike_end,
    STATE(4831), 1,
      sym_heredoc_content,
    STATE(4838), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39841] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2288), 1,
      sym__quotelike_end,
    STATE(4832), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39866] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5603), 1,
      sym__quotelike_end_quote,
    STATE(1822), 1,
      sym__quotelike_end,
    STATE(4833), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39891] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3394), 1,
      sym__quotelike_end,
    STATE(4834), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39916] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3395), 1,
      sym__quotelike_end,
    STATE(4835), 1,
      sym_heredoc_content,
    STATE(4855), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39941] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7260), 1,
      anon_sym_GT,
    STATE(4836), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5809), 2,
      sym_scalar,
      sym_bareword,
  [39966] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2753), 1,
      sym__quotelike_end,
    STATE(4837), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [39991] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2588), 1,
      sym__quotelike_end,
    STATE(4838), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40016] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2589), 1,
      sym__quotelike_end,
    STATE(4839), 1,
      sym_heredoc_content,
    STATE(4846), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40041] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2205), 1,
      sym__quotelike_end,
    STATE(4840), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40066] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2230), 1,
      sym__quotelike_end,
    STATE(4841), 1,
      sym_heredoc_content,
    STATE(4911), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40091] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7262), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2978), 1,
      sym_block,
    STATE(4842), 1,
      sym_heredoc_content,
    STATE(5351), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40120] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7264), 1,
      sym_prototype,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2980), 1,
      sym_block,
    STATE(4843), 1,
      sym_heredoc_content,
    STATE(5352), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40149] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2372), 1,
      sym__quotelike_end,
    STATE(4844), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40174] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2373), 1,
      sym__quotelike_end,
    STATE(4845), 1,
      sym_heredoc_content,
    STATE(4868), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40199] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5701), 1,
      sym__quotelike_end_quote,
    STATE(2603), 1,
      sym__quotelike_end,
    STATE(4846), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40224] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5633), 1,
      sym__quotelike_end_quote,
    STATE(3291), 1,
      sym__quotelike_end,
    STATE(4729), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4847), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40249] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7266), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2371), 1,
      sym_block,
    STATE(4848), 1,
      sym_heredoc_content,
    STATE(5420), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40278] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7268), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1911), 1,
      sym_block,
    STATE(4849), 1,
      sym_heredoc_content,
    STATE(5441), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40307] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7270), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3060), 1,
      sym_block,
    STATE(4850), 1,
      sym_heredoc_content,
    STATE(5500), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40336] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7272), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3647), 1,
      sym_block,
    STATE(4851), 1,
      sym_heredoc_content,
    STATE(5407), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40365] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7274), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1916), 1,
      sym_block,
    STATE(4852), 1,
      sym_heredoc_content,
    STATE(5489), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40394] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3397), 1,
      sym__quotelike_end,
    STATE(4712), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4853), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40419] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3249), 1,
      sym__quotelike_end,
    STATE(4854), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40444] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5695), 1,
      sym__quotelike_end_quote,
    STATE(3310), 1,
      sym__quotelike_end,
    STATE(4855), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40469] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7276), 1,
      sym_prototype,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3062), 1,
      sym_block,
    STATE(4856), 1,
      sym_heredoc_content,
    STATE(5503), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40498] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7278), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2378), 1,
      sym_block,
    STATE(4857), 1,
      sym_heredoc_content,
    STATE(5422), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40527] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3250), 1,
      sym__quotelike_end,
    STATE(4858), 1,
      sym_heredoc_content,
    STATE(4919), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40552] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2302), 1,
      sym__quotelike_end,
    STATE(4859), 1,
      sym_heredoc_content,
    STATE(4867), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40577] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7280), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1681), 1,
      sym_block,
    STATE(4860), 1,
      sym_heredoc_content,
    STATE(5506), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40606] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3201), 1,
      sym__quotelike_end,
    STATE(4861), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40631] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7282), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1683), 1,
      sym_block,
    STATE(4862), 1,
      sym_heredoc_content,
    STATE(5509), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40660] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7284), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1583), 1,
      sym_block,
    STATE(4863), 1,
      sym_heredoc_content,
    STATE(5427), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40689] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7286), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1628), 1,
      sym_block,
    STATE(4864), 1,
      sym_heredoc_content,
    STATE(5428), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40718] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7288), 1,
      anon_sym_COLON,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1516), 1,
      sym_block,
    STATE(4865), 1,
      sym_heredoc_content,
    STATE(5550), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40747] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3212), 1,
      sym__quotelike_end,
    STATE(4866), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40772] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2360), 1,
      sym__quotelike_end,
    STATE(4867), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40797] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5699), 1,
      sym__quotelike_end_quote,
    STATE(2399), 1,
      sym__quotelike_end,
    STATE(4868), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40822] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4869), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7290), 6,
      sym__PERLY_SEMICOLON,
      sym__ctrl_z_hack,
      anon_sym_SEMI,
      anon_sym___DATA__,
      anon_sym___END__,
      anon_sym_EOT,
  [40841] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5643), 1,
      sym__quotelike_end_quote,
    STATE(2222), 1,
      sym__quotelike_end,
    STATE(4870), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40866] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1785), 1,
      sym__quotelike_end,
    STATE(4871), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40891] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7292), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1596), 1,
      sym_block,
    STATE(4872), 1,
      sym_heredoc_content,
    STATE(5447), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40920] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2434), 1,
      sym__quotelike_end,
    STATE(4873), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [40945] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4874), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1487), 2,
      sym_block,
      sym__semicolon,
  [40970] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7294), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1598), 1,
      sym_block,
    STATE(4875), 1,
      sym_heredoc_content,
    STATE(5451), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [40999] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7296), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1602), 1,
      sym_block,
    STATE(4876), 1,
      sym_heredoc_content,
    STATE(5455), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41028] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7298), 1,
      sym_prototype,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1604), 1,
      sym_block,
    STATE(4877), 1,
      sym_heredoc_content,
    STATE(5457), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41057] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3214), 1,
      sym__quotelike_end,
    STATE(4878), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41082] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7300), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1891), 1,
      sym_block,
    STATE(4879), 1,
      sym_heredoc_content,
    STATE(5496), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41111] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1786), 1,
      sym__quotelike_end,
    STATE(4880), 1,
      sym_heredoc_content,
    STATE(4940), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41136] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1792), 1,
      sym__quotelike_end,
    STATE(4881), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41161] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2441), 1,
      sym__quotelike_end,
    STATE(4882), 1,
      sym_heredoc_content,
    STATE(4903), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41186] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1805), 1,
      sym__quotelike_end,
    STATE(4745), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4883), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41211] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4884), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1496), 2,
      sym_block,
      sym__semicolon,
  [41236] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1775), 1,
      sym__quotelike_end,
    STATE(4885), 1,
      sym_heredoc_content,
    STATE(4945), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41261] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4886), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1580), 2,
      sym_block,
      sym__semicolon,
  [41286] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3012), 1,
      sym__quotelike_end,
    STATE(4887), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41311] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4888), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7302), 6,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_undef,
      anon_sym_PERCENT2,
  [41330] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4889), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1581), 2,
      sym_block,
      sym__semicolon,
  [41355] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4890), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3304), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(3309), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [41376] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3028), 1,
      sym__quotelike_end,
    STATE(4891), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41401] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7304), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1694), 1,
      sym_block,
    STATE(4892), 1,
      sym_heredoc_content,
    STATE(5525), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41430] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7306), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1696), 1,
      sym_block,
    STATE(4893), 1,
      sym_heredoc_content,
    STATE(5527), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41459] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2332), 1,
      sym__quotelike_end,
    STATE(4894), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41484] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2363), 1,
      sym__quotelike_end,
    STATE(4895), 1,
      sym_heredoc_content,
    STATE(4907), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41509] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2379), 1,
      sym__quotelike_end,
    STATE(4767), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4896), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41534] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(3239), 1,
      sym__quotelike_end,
    STATE(4897), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41559] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7308), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1504), 1,
      sym_block,
    STATE(4898), 1,
      sym_heredoc_content,
    STATE(5573), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41588] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7310), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1509), 1,
      sym_block,
    STATE(4899), 1,
      sym_heredoc_content,
    STATE(5534), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41617] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4900), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1453), 2,
      sym_block,
      sym__semicolon,
  [41642] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2338), 1,
      sym__quotelike_end,
    STATE(4901), 1,
      sym_heredoc_content,
    STATE(4908), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41667] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7312), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1700), 1,
      sym_block,
    STATE(4902), 1,
      sym_heredoc_content,
    STATE(5528), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41696] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2514), 1,
      sym__quotelike_end,
    STATE(4903), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41721] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2577), 1,
      sym__quotelike_end,
    STATE(4904), 1,
      sym_heredoc_content,
    STATE(4906), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41746] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7314), 1,
      sym_prototype,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1702), 1,
      sym_block,
    STATE(4905), 1,
      sym_heredoc_content,
    STATE(5529), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41775] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(2619), 1,
      sym__quotelike_end,
    STATE(4906), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41800] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2385), 1,
      sym__quotelike_end,
    STATE(4907), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41825] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2374), 1,
      sym__quotelike_end,
    STATE(4908), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41850] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7316), 1,
      anon_sym_GT,
    STATE(4909), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5728), 2,
      sym_scalar,
      sym_bareword,
  [41875] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2375), 1,
      sym__quotelike_end,
    STATE(4910), 1,
      sym_heredoc_content,
    STATE(4915), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41900] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2224), 1,
      sym__quotelike_end,
    STATE(4911), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [41925] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7318), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3511), 1,
      sym_block,
    STATE(4912), 1,
      sym_heredoc_content,
    STATE(5338), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [41954] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4913), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3401), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3403), 3,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
  [41975] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7320), 1,
      sym_prototype,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3311), 1,
      sym_block,
    STATE(4914), 1,
      sym_heredoc_content,
    STATE(5339), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42004] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5697), 1,
      sym__quotelike_end_quote,
    STATE(2400), 1,
      sym__quotelike_end,
    STATE(4915), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42029] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4916), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7322), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42048] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7324), 1,
      anon_sym_GT,
    STATE(4917), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5949), 2,
      sym_scalar,
      sym_bareword,
  [42073] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3513), 1,
      sym__quotelike_end,
    STATE(4918), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42098] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5691), 1,
      sym__quotelike_end_quote,
    STATE(3253), 1,
      sym__quotelike_end,
    STATE(4919), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42123] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3603), 1,
      sym__quotelike_end,
    STATE(4920), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42148] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7328), 1,
      anon_sym_COMMA,
    STATE(4921), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7326), 5,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42169] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1849), 1,
      sym__quotelike_end,
    STATE(4922), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42194] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2206), 1,
      sym__quotelike_end,
    STATE(4923), 1,
      sym_heredoc_content,
    STATE(4965), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42219] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3506), 1,
      sym__quotelike_end,
    STATE(4924), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42244] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7330), 1,
      anon_sym_GT,
    STATE(4925), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7200), 2,
      sym__identifier,
      aux_sym__bareword_token1,
    STATE(5883), 2,
      sym_scalar,
      sym_bareword,
  [42269] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3243), 1,
      sym__quotelike_end,
    STATE(4926), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42294] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3338), 1,
      sym__quotelike_end,
    STATE(4927), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42319] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2395), 1,
      sym__quotelike_end,
    STATE(4709), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4928), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42344] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3247), 1,
      sym__quotelike_end,
    STATE(4929), 1,
      sym_heredoc_content,
    STATE(4932), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42369] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4930), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7332), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42388] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4931), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1659), 2,
      sym_block,
      sym__semicolon,
  [42413] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3251), 1,
      sym__quotelike_end,
    STATE(4932), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42438] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4933), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6921), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42457] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3252), 1,
      sym__quotelike_end,
    STATE(4934), 1,
      sym_heredoc_content,
    STATE(4941), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42482] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4935), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6923), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42501] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5663), 1,
      sym__quotelike_end_quote,
    STATE(3195), 1,
      sym__quotelike_end,
    STATE(4866), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4936), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42526] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4937), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3331), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3333), 3,
      sym__attribute_value_begin,
      sym__signature_start,
      anon_sym_LBRACE,
  [42547] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4938), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7334), 6,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [42566] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2236), 1,
      sym__quotelike_end,
    STATE(4840), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4939), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42591] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5667), 1,
      sym__quotelike_end_quote,
    STATE(1793), 1,
      sym__quotelike_end,
    STATE(4940), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42616] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3254), 1,
      sym__quotelike_end,
    STATE(4941), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42641] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5621), 1,
      sym__quotelike_end_quote,
    STATE(3292), 1,
      sym__quotelike_end,
    STATE(4746), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4942), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42666] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5607), 1,
      sym__quotelike_end_quote,
    STATE(2213), 1,
      sym__quotelike_end,
    STATE(4714), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4943), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42691] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7336), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2411), 1,
      sym_block,
    STATE(4944), 1,
      sym_heredoc_content,
    STATE(5357), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42720] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1787), 1,
      sym__quotelike_end,
    STATE(4945), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42745] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7338), 1,
      sym_prototype,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1450), 1,
      sym_block,
    STATE(4946), 1,
      sym_heredoc_content,
    STATE(5530), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42774] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5687), 1,
      sym__quotelike_end_quote,
    STATE(2701), 1,
      sym__quotelike_end,
    STATE(4800), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4947), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42799] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7340), 1,
      sym__attribute_value_begin,
    STATE(4948), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3337), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(3335), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [42822] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7342), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3208), 1,
      sym_block,
    STATE(4949), 1,
      sym_heredoc_content,
    STATE(5387), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42851] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3614), 1,
      sym__quotelike_end,
    STATE(4950), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [42876] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4951), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1562), 2,
      sym_block,
      sym__semicolon,
  [42901] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7344), 1,
      sym_prototype,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2439), 1,
      sym_block,
    STATE(4952), 1,
      sym_heredoc_content,
    STATE(5400), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42930] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7346), 1,
      sym_prototype,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3210), 1,
      sym_block,
    STATE(4953), 1,
      sym_heredoc_content,
    STATE(5388), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [42959] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4954), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3276), 6,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_SEMI,
      sym_version,
      sym_number,
  [42978] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5617), 1,
      sym__quotelike_end_quote,
    STATE(2436), 1,
      sym__quotelike_end,
    STATE(4955), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43003] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5665), 1,
      sym__quotelike_end_quote,
    STATE(1817), 1,
      sym__quotelike_end,
    STATE(4881), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4956), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43028] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3367), 1,
      sym__quotelike_end,
    STATE(4957), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43053] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4958), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1661), 2,
      sym_block,
      sym__semicolon,
  [43078] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3689), 1,
      sym__quotelike_end,
    STATE(4959), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43103] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4960), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1726), 2,
      sym_block,
      sym__semicolon,
  [43128] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5685), 1,
      sym__quotelike_end_quote,
    STATE(3488), 1,
      sym__quotelike_end,
    STATE(4924), 1,
      aux_sym__noninterpolated_string_content,
    STATE(4961), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43153] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7348), 1,
      sym_prototype,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1898), 1,
      sym_block,
    STATE(4962), 1,
      sym_heredoc_content,
    STATE(5549), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [43182] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(4963), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    STATE(1564), 2,
      sym_block,
      sym__semicolon,
  [43207] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4964), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5860), 6,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [43226] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2220), 1,
      sym__quotelike_end,
    STATE(4965), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43251] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5647), 1,
      sym__quotelike_end_quote,
    STATE(2340), 1,
      sym__quotelike_end,
    STATE(4966), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43276] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3371), 1,
      sym__quotelike_end,
    STATE(4967), 1,
      sym_heredoc_content,
    STATE(4968), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43301] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5669), 1,
      sym__quotelike_end_quote,
    STATE(3396), 1,
      sym__quotelike_end,
    STATE(4968), 1,
      sym_heredoc_content,
    STATE(5036), 1,
      aux_sym__noninterpolated_string_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7192), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [43326] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4969), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 6,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [43345] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(245), 1,
      sym_package,
    STATE(4970), 1,
      sym_heredoc_content,
    STATE(5347), 1,
      sym__version,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6706), 2,
      sym_version,
      sym_number,
    ACTIONS(7228), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [43370] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4971), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3409), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
    ACTIONS(3411), 3,
      sym__PERLY_SEMICOLON,
      anon_sym_LBRACE,
      anon_sym_SEMI,
  [43391] = 9,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    ACTIONS(7350), 1,
      sym_prototype,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3649), 1,
      sym_block,
    STATE(4972), 1,
      sym_heredoc_content,
    STATE(5419), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [43420] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7352), 1,
      sym__identifier,
    ACTIONS(7354), 1,
      anon_sym_DOLLAR,
    STATE(3375), 1,
      sym__ident_special,
    STATE(3378), 1,
      sym_varname,
    STATE(4973), 1,
      sym_heredoc_content,
  [43448] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4974), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7356), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
  [43470] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_RPAREN,
    STATE(4975), 1,
      sym_heredoc_content,
    STATE(5129), 1,
      aux_sym__for_initializer_repeat1,
    STATE(5668), 1,
      sym_scalar,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [43496] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4976), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7366), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43518] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4977), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7372), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43540] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4978), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7374), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43562] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4979), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7376), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43584] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4980), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7378), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43606] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6338), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7382), 1,
      anon_sym_DOLLAR,
    STATE(2233), 1,
      sym_varname,
    STATE(2245), 1,
      sym__ident_special,
    STATE(4981), 1,
      sym_heredoc_content,
  [43634] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4982), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5860), 5,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [43652] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4983), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7384), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43674] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4984), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7386), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43696] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4985), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7388), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43718] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4986), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7390), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43740] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4987), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7392), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43762] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6338), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7382), 1,
      anon_sym_DOLLAR,
    STATE(2245), 1,
      sym__ident_special,
    STATE(2248), 1,
      sym_varname,
    STATE(4988), 1,
      sym_heredoc_content,
  [43790] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6338), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7382), 1,
      anon_sym_DOLLAR,
    STATE(2219), 1,
      sym_varname,
    STATE(2245), 1,
      sym__ident_special,
    STATE(4989), 1,
      sym_heredoc_content,
  [43818] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4990), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7394), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43840] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4991), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7396), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43862] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4992), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7398), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43884] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4993), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7400), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43906] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4994), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7402), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43928] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4995), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7404), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43950] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4996), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7406), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [43972] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(4997), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7408), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [43994] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4998), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7410), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44016] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(4999), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7412), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44038] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5000), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7414), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44060] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5001), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7416), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44082] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5002), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7418), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44104] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5003), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7420), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44126] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5004), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7422), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44148] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5005), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7424), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44170] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5006), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7426), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44192] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7428), 1,
      sym__identifier,
    ACTIONS(7430), 1,
      anon_sym_DOLLAR,
    STATE(2883), 1,
      sym__ident_special,
    STATE(2884), 1,
      sym_varname,
    STATE(5007), 1,
      sym_heredoc_content,
  [44220] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7428), 1,
      sym__identifier,
    ACTIONS(7430), 1,
      anon_sym_DOLLAR,
    STATE(2883), 1,
      sym__ident_special,
    STATE(2885), 1,
      sym_varname,
    STATE(5008), 1,
      sym_heredoc_content,
  [44248] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6284), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7428), 1,
      sym__identifier,
    ACTIONS(7430), 1,
      anon_sym_DOLLAR,
    STATE(2883), 1,
      sym__ident_special,
    STATE(2887), 1,
      sym_varname,
    STATE(5009), 1,
      sym_heredoc_content,
  [44276] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5010), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7432), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44298] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5011), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7434), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44320] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5012), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7436), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44342] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3590), 2,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
    ACTIONS(7438), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(5013), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
  [44362] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5014), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7441), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44384] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5015), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7443), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44406] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5016), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7445), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44428] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5017), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7447), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44450] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5018), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7449), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44472] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7451), 1,
      sym__identifier,
    ACTIONS(7453), 1,
      anon_sym_DOLLAR,
    STATE(3476), 1,
      sym__ident_special,
    STATE(3477), 1,
      sym_varname,
    STATE(5019), 1,
      sym_heredoc_content,
  [44500] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7451), 1,
      sym__identifier,
    ACTIONS(7453), 1,
      anon_sym_DOLLAR,
    STATE(3476), 1,
      sym__ident_special,
    STATE(3478), 1,
      sym_varname,
    STATE(5020), 1,
      sym_heredoc_content,
  [44528] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5021), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7455), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44550] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6162), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7451), 1,
      sym__identifier,
    ACTIONS(7453), 1,
      anon_sym_DOLLAR,
    STATE(3476), 1,
      sym__ident_special,
    STATE(3480), 1,
      sym_varname,
    STATE(5022), 1,
      sym_heredoc_content,
  [44578] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7457), 1,
      anon_sym_DOLLAR,
    STATE(2233), 1,
      sym_varname,
    STATE(2245), 1,
      sym__ident_special,
    STATE(5023), 1,
      sym_heredoc_content,
  [44606] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7457), 1,
      anon_sym_DOLLAR,
    STATE(2245), 1,
      sym__ident_special,
    STATE(2248), 1,
      sym_varname,
    STATE(5024), 1,
      sym_heredoc_content,
  [44634] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6386), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7380), 1,
      sym__identifier,
    ACTIONS(7457), 1,
      anon_sym_DOLLAR,
    STATE(2219), 1,
      sym_varname,
    STATE(2245), 1,
      sym__ident_special,
    STATE(5025), 1,
      sym_heredoc_content,
  [44662] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5026), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7459), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44684] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5027), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7461), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44706] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5028), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7463), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44728] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5029), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7465), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44750] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7467), 1,
      sym__quotelike_middle_close_quote,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5030), 2,
      aux_sym__noninterpolated_string_content,
      sym_heredoc_content,
    ACTIONS(7469), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [44770] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7472), 1,
      sym__identifier,
    ACTIONS(7474), 1,
      anon_sym_DOLLAR,
    STATE(1795), 1,
      sym_varname,
    STATE(1813), 1,
      sym__ident_special,
    STATE(5031), 1,
      sym_heredoc_content,
  [44798] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5032), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7476), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44820] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7472), 1,
      sym__identifier,
    ACTIONS(7474), 1,
      anon_sym_DOLLAR,
    STATE(1799), 1,
      sym_varname,
    STATE(1813), 1,
      sym__ident_special,
    STATE(5033), 1,
      sym_heredoc_content,
  [44848] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5034), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7478), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44870] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7480), 1,
      anon_sym_sub,
    ACTIONS(7482), 1,
      anon_sym_method,
    STATE(5035), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [44894] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7467), 1,
      sym__quotelike_end_quote,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5036), 2,
      aux_sym__noninterpolated_string_content,
      sym_heredoc_content,
    ACTIONS(7484), 3,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [44914] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5037), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7487), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44936] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5038), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7489), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [44958] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5039), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7491), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [44980] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5040), 1,
      sym_heredoc_content,
    ACTIONS(7493), 5,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      sym__identifier,
      aux_sym__ident_special_token1,
      aux_sym__bareword_token1,
  [45000] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5041), 1,
      sym_heredoc_content,
    STATE(5101), 1,
      aux_sym__term_rightward_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3568), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(5301), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
  [45022] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5042), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7495), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45044] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5043), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7497), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45066] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7499), 1,
      anon_sym_sub,
    ACTIONS(7501), 1,
      anon_sym_method,
    STATE(5044), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [45090] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5045), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7503), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45112] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5046), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7505), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45134] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5047), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7507), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45156] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5048), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 5,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [45174] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7509), 1,
      anon_sym_and,
    STATE(5049), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 4,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
      anon_sym_or,
      anon_sym_xor,
  [45194] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5050), 1,
      sym_heredoc_content,
    ACTIONS(7511), 5,
      anon_sym_LBRACE,
      anon_sym_DOLLAR,
      sym__identifier,
      aux_sym__ident_special_token1,
      aux_sym__bareword_token1,
  [45214] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5051), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 5,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [45232] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5052), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7513), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45254] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7352), 1,
      sym__identifier,
    ACTIONS(7354), 1,
      anon_sym_DOLLAR,
    STATE(3375), 1,
      sym__ident_special,
    STATE(3399), 1,
      sym_varname,
    STATE(5053), 1,
      sym_heredoc_content,
  [45282] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6330), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7352), 1,
      sym__identifier,
    ACTIONS(7354), 1,
      anon_sym_DOLLAR,
    STATE(3375), 1,
      sym__ident_special,
    STATE(3400), 1,
      sym_varname,
    STATE(5054), 1,
      sym_heredoc_content,
  [45310] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7509), 1,
      anon_sym_and,
    STATE(5055), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7515), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
    ACTIONS(7517), 2,
      anon_sym_or,
      anon_sym_xor,
  [45332] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5056), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 5,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [45350] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6308), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7472), 1,
      sym__identifier,
    ACTIONS(7474), 1,
      anon_sym_DOLLAR,
    STATE(1776), 1,
      sym_varname,
    STATE(1813), 1,
      sym__ident_special,
    STATE(5057), 1,
      sym_heredoc_content,
  [45378] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5058), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3403), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(3401), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [45398] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7519), 1,
      anon_sym_RPAREN,
    STATE(4975), 1,
      aux_sym__for_initializer_repeat1,
    STATE(5059), 1,
      sym_heredoc_content,
    STATE(5614), 1,
      sym_scalar,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [45424] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5060), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 4,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [45444] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5061), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7521), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45466] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5062), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7523), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45488] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5063), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7525), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45510] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5064), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7527), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45532] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5065), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7529), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45554] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5066), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7531), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45576] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5067), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7533), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45598] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5068), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7535), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45620] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5069), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7537), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45642] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5070), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7539), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45664] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5071), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7541), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45686] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7543), 1,
      anon_sym_sub,
    ACTIONS(7545), 1,
      anon_sym_method,
    STATE(5072), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [45710] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7547), 1,
      sym__identifier,
    ACTIONS(7549), 1,
      anon_sym_DOLLAR,
    STATE(2457), 1,
      sym__ident_special,
    STATE(2462), 1,
      sym_varname,
    STATE(5073), 1,
      sym_heredoc_content,
  [45738] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5074), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7551), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45760] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5075), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7553), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45782] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5076), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7555), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45804] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7557), 1,
      anon_sym_LBRACE,
    ACTIONS(7559), 1,
      anon_sym_AT,
    ACTIONS(7561), 1,
      anon_sym_LBRACK,
    ACTIONS(7563), 1,
      anon_sym_DOLLAR_STAR,
    ACTIONS(7565), 1,
      anon_sym_AT_STAR,
    STATE(5077), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [45830] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1648), 1,
      sym_block,
    STATE(5078), 1,
      sym_heredoc_content,
    STATE(5430), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [45856] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5079), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7567), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [45878] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5080), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7569), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45900] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5081), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7571), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [45922] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7573), 1,
      sym__identifier,
    ACTIONS(7575), 1,
      anon_sym_DOLLAR,
    STATE(2331), 1,
      sym_varname,
    STATE(2448), 1,
      sym__ident_special,
    STATE(5082), 1,
      sym_heredoc_content,
  [45950] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7573), 1,
      sym__identifier,
    ACTIONS(7575), 1,
      anon_sym_DOLLAR,
    STATE(2357), 1,
      sym_varname,
    STATE(2448), 1,
      sym__ident_special,
    STATE(5083), 1,
      sym_heredoc_content,
  [45978] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6172), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7573), 1,
      sym__identifier,
    ACTIONS(7575), 1,
      anon_sym_DOLLAR,
    STATE(2369), 1,
      sym_varname,
    STATE(2448), 1,
      sym__ident_special,
    STATE(5084), 1,
      sym_heredoc_content,
  [46006] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7577), 1,
      anon_sym_LBRACE,
    ACTIONS(7579), 1,
      anon_sym_AT,
    ACTIONS(7581), 1,
      anon_sym_LBRACK,
    ACTIONS(7583), 1,
      anon_sym_DOLLAR_STAR,
    ACTIONS(7585), 1,
      anon_sym_AT_STAR,
    STATE(5085), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [46032] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5086), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7587), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46054] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5087), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7589), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46076] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5088), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7591), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46098] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5089), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7593), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46120] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5090), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7595), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46142] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5091), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7597), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46164] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7599), 1,
      anon_sym_LBRACE,
    ACTIONS(7601), 1,
      anon_sym_AT,
    ACTIONS(7603), 1,
      anon_sym_LBRACK,
    ACTIONS(7605), 1,
      anon_sym_DOLLAR_STAR,
    ACTIONS(7607), 1,
      anon_sym_AT_STAR,
    STATE(5092), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [46190] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7609), 1,
      anon_sym_sub,
    ACTIONS(7611), 1,
      anon_sym_method,
    STATE(5093), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46214] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5094), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5860), 5,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
      anon_sym_RBRACK,
  [46232] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7613), 1,
      anon_sym_sub,
    ACTIONS(7615), 1,
      anon_sym_method,
    STATE(5095), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46256] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7617), 1,
      anon_sym_LBRACE,
    ACTIONS(7619), 1,
      anon_sym_AT,
    ACTIONS(7621), 1,
      anon_sym_LBRACK,
    ACTIONS(7623), 1,
      anon_sym_DOLLAR_STAR,
    ACTIONS(7625), 1,
      anon_sym_AT_STAR,
    STATE(5096), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [46282] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1552), 1,
      sym_block,
    STATE(5097), 1,
      sym_heredoc_content,
    STATE(5498), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [46308] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5098), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7627), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46330] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5099), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7629), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46352] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5100), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7631), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46374] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3590), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
    ACTIONS(7633), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(5101), 2,
      sym_heredoc_content,
      aux_sym__term_rightward_repeat1,
  [46394] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7636), 1,
      anon_sym_sub,
    ACTIONS(7638), 1,
      anon_sym_method,
    STATE(5102), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46418] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5103), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7640), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46440] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7642), 1,
      anon_sym_sub,
    ACTIONS(7644), 1,
      anon_sym_method,
    STATE(5104), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46464] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5105), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7646), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46486] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5106), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7648), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46508] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5107), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7650), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46530] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5108), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7652), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46552] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5109), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7654), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46574] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5110), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7656), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46596] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5013), 1,
      aux_sym__term_rightward_repeat1,
    STATE(5111), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3568), 2,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
    ACTIONS(5198), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
  [46618] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5112), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7658), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46640] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5113), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7660), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46662] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7662), 1,
      anon_sym_sub,
    ACTIONS(7664), 1,
      anon_sym_method,
    STATE(5114), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46686] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5115), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7666), 5,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_AT,
      anon_sym_RPAREN,
      anon_sym_PERCENT2,
  [46704] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5116), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7668), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46726] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5117), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3304), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(3309), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [46746] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5118), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7670), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46768] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7672), 1,
      anon_sym_sub,
    ACTIONS(7674), 1,
      anon_sym_method,
    STATE(5119), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [46792] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5120), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7676), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46814] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5121), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7678), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46836] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5122), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7680), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46858] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5123), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5856), 4,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
      anon_sym_or,
      anon_sym_xor,
  [46878] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5124), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7682), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [46900] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5125), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7684), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46922] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5126), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7686), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46944] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5127), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7688), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46966] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5128), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7690), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [46988] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7692), 1,
      anon_sym_DOLLAR,
    ACTIONS(7695), 1,
      anon_sym_COMMA,
    ACTIONS(7698), 1,
      anon_sym_RPAREN,
    STATE(5935), 1,
      sym_scalar,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    STATE(5129), 2,
      sym_heredoc_content,
      aux_sym__for_initializer_repeat1,
  [47012] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7700), 1,
      sym__identifier,
    ACTIONS(7702), 1,
      anon_sym_DOLLAR,
    STATE(2291), 1,
      sym__ident_special,
    STATE(2292), 1,
      sym_varname,
    STATE(5130), 1,
      sym_heredoc_content,
  [47040] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7700), 1,
      sym__identifier,
    ACTIONS(7702), 1,
      anon_sym_DOLLAR,
    STATE(2291), 1,
      sym__ident_special,
    STATE(2293), 1,
      sym_varname,
    STATE(5131), 1,
      sym_heredoc_content,
  [47068] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6356), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7700), 1,
      sym__identifier,
    ACTIONS(7702), 1,
      anon_sym_DOLLAR,
    STATE(2291), 1,
      sym__ident_special,
    STATE(2295), 1,
      sym_varname,
    STATE(5132), 1,
      sym_heredoc_content,
  [47096] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5133), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7704), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47118] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5134), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7706), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47140] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5135), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7708), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47162] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5136), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7710), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47184] = 8,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(6534), 1,
      sym__signature_start,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1461), 1,
      sym_block,
    STATE(5137), 1,
      sym_heredoc_content,
    STATE(5560), 1,
      sym_signature,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [47210] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7712), 1,
      anon_sym_sub,
    ACTIONS(7714), 1,
      anon_sym_method,
    STATE(5138), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [47234] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7716), 1,
      anon_sym_sub,
    ACTIONS(7718), 1,
      anon_sym_method,
    STATE(5139), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [47258] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5140), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7720), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47280] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5141), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7722), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47302] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5142), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7724), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47324] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7726), 1,
      anon_sym_sub,
    ACTIONS(7728), 1,
      anon_sym_method,
    STATE(5143), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [47348] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5144), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7730), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47370] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5145), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7732), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47392] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5146), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7734), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47414] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5147), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3411), 2,
      sym__signature_start,
      anon_sym_LBRACE,
    ACTIONS(3409), 3,
      anon_sym_COLON,
      sym__identifier,
      aux_sym__bareword_token1,
  [47434] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5148), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7736), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47456] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7738), 1,
      anon_sym_sub,
    ACTIONS(7740), 1,
      anon_sym_method,
    STATE(5149), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [47480] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7742), 1,
      anon_sym_sub,
    ACTIONS(7744), 1,
      anon_sym_method,
    STATE(5150), 1,
      sym_heredoc_content,
    STATE(5158), 1,
      aux_sym_subroutine_declaration_statement_repeat1,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5868), 2,
      anon_sym_extended,
      anon_sym_async,
  [47504] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5151), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7746), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47526] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5152), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(5860), 5,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
      anon_sym_and,
      anon_sym_or,
      anon_sym_xor,
  [47544] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7368), 1,
      anon_sym_and,
    STATE(5153), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7370), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7748), 2,
      sym__RECOVER_BRACKET_CLOSE,
      anon_sym_RBRACK,
  [47566] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5154), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7750), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47588] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7547), 1,
      sym__identifier,
    ACTIONS(7549), 1,
      anon_sym_DOLLAR,
    STATE(2457), 1,
      sym__ident_special,
    STATE(2458), 1,
      sym_varname,
    STATE(5155), 1,
      sym_heredoc_content,
  [47616] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6182), 1,
      aux_sym__ident_special_token1,
    ACTIONS(7547), 1,
      sym__identifier,
    ACTIONS(7549), 1,
      anon_sym_DOLLAR,
    STATE(2457), 1,
      sym__ident_special,
    STATE(2459), 1,
      sym_varname,
    STATE(5156), 1,
      sym_heredoc_content,
  [47644] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7358), 1,
      anon_sym_and,
    STATE(5157), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7360), 2,
      anon_sym_or,
      anon_sym_xor,
    ACTIONS(7752), 2,
      sym__RECOVER_PAREN_CLOSE,
      anon_sym_RPAREN,
  [47666] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7754), 2,
      anon_sym_extended,
      anon_sym_async,
    ACTIONS(7757), 2,
      anon_sym_sub,
      anon_sym_method,
    STATE(5158), 2,
      sym_heredoc_content,
      aux_sym_subroutine_declaration_statement_repeat1,
  [47686] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7759), 1,
      anon_sym_RPAREN,
    STATE(5159), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47707] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1466), 1,
      sym_block,
    STATE(1479), 1,
      sym__loop_body,
    STATE(5160), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [47730] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7763), 1,
      anon_sym_RBRACK,
    STATE(5161), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47751] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3942), 1,
      sym__quotelike_begin,
    STATE(4728), 1,
      sym__apostrophe,
    STATE(5162), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [47774] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7767), 1,
      anon_sym_RBRACK,
    STATE(5163), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47795] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5164), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2698), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [47812] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1576), 1,
      sym_block,
    STATE(1592), 1,
      sym__loop_body,
    STATE(5165), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [47835] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7769), 1,
      anon_sym_RPAREN,
    STATE(5166), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47856] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7771), 1,
      anon_sym_RPAREN,
    STATE(5167), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47877] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5780), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5168), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47898] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7773), 1,
      anon_sym_RPAREN,
    STATE(5169), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47919] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7775), 1,
      anon_sym_RBRACK,
    STATE(5170), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47940] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7777), 1,
      anon_sym_RBRACK,
    STATE(5171), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47961] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7779), 1,
      anon_sym_RPAREN,
    STATE(5172), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [47982] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7781), 1,
      anon_sym_RPAREN,
    STATE(5173), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48003] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7783), 1,
      anon_sym_LPAREN,
    STATE(170), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1637), 1,
      sym_block,
    STATE(5174), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48026] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7785), 1,
      anon_sym_RPAREN,
    STATE(5175), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48047] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5176), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2530), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [48064] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7787), 1,
      anon_sym_RBRACK,
    STATE(5177), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48085] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7789), 1,
      anon_sym_RBRACK,
    STATE(5178), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48106] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7791), 1,
      anon_sym_RBRACK,
    STATE(5179), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48127] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3907), 1,
      sym__quotelike_begin,
    STATE(4928), 1,
      sym__apostrophe,
    STATE(5180), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48150] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7793), 1,
      anon_sym_RPAREN,
    STATE(5181), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48171] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3828), 1,
      sym__quotelike_begin,
    STATE(4943), 1,
      sym__apostrophe,
    STATE(5182), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48194] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7797), 1,
      anon_sym_RPAREN,
    STATE(5183), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48215] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4680), 1,
      sym_attribute_name,
    STATE(4719), 1,
      sym_attribute,
    STATE(5184), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6526), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [48236] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3830), 1,
      sym__quotelike_begin,
    STATE(4723), 1,
      sym__apostrophe,
    STATE(5185), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48259] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4134), 1,
      sym__quotelike_begin,
    STATE(4613), 1,
      sym__apostrophe,
    STATE(5186), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48282] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2167), 1,
      sym_attribute_name,
    STATE(2252), 1,
      sym_attribute,
    STATE(5187), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4321), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [48303] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7803), 1,
      anon_sym_RBRACK,
    STATE(5188), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48324] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7805), 1,
      anon_sym_RPAREN,
    STATE(5189), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48345] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7807), 1,
      anon_sym_RBRACK,
    STATE(5190), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48366] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7809), 1,
      anon_sym_RPAREN,
    STATE(5191), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48387] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5771), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5192), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48408] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7811), 1,
      anon_sym_RBRACK,
    STATE(5193), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48429] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7813), 1,
      anon_sym_RBRACK,
    STATE(5194), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48450] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7815), 1,
      anon_sym_RBRACK,
    STATE(5195), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48471] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4690), 1,
      sym_attribute_name,
    STATE(4890), 1,
      sym_attribute,
    STATE(5196), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6530), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [48492] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7817), 1,
      anon_sym_RPAREN,
    STATE(5197), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48513] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3946), 1,
      sym__quotelike_begin,
    STATE(4817), 1,
      sym__apostrophe,
    STATE(5198), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48536] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3862), 1,
      sym__quotelike_begin,
    STATE(4847), 1,
      sym__apostrophe,
    STATE(5199), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48559] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3871), 1,
      sym__quotelike_begin,
    STATE(4942), 1,
      sym__apostrophe,
    STATE(5200), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48582] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4128), 1,
      sym__quotelike_begin,
    STATE(4615), 1,
      sym__apostrophe,
    STATE(5201), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48605] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7819), 1,
      anon_sym_RPAREN,
    STATE(5202), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48626] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5764), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5203), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48647] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7821), 1,
      anon_sym_RBRACK,
    STATE(5204), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48668] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7823), 1,
      anon_sym_RBRACK,
    STATE(5205), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48689] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7825), 1,
      anon_sym_RPAREN,
    STATE(5206), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48710] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7827), 1,
      anon_sym_RBRACK,
    STATE(5207), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48731] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7829), 1,
      anon_sym_RBRACK,
    STATE(5208), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48752] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1673), 1,
      sym_block,
    STATE(1674), 1,
      sym__loop_body,
    STATE(5209), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48775] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7831), 1,
      anon_sym_RBRACK,
    STATE(5210), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48796] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7833), 1,
      anon_sym_RBRACK,
    STATE(5211), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48817] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1576), 1,
      sym_block,
    STATE(1605), 1,
      sym__loop_body,
    STATE(5212), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48840] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1466), 1,
      sym_block,
    STATE(1483), 1,
      sym__loop_body,
    STATE(5213), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48863] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1673), 1,
      sym_block,
    STATE(1675), 1,
      sym__loop_body,
    STATE(5214), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48886] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7835), 1,
      anon_sym_RBRACK,
    STATE(5215), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [48907] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3892), 1,
      sym__quotelike_begin,
    STATE(4778), 1,
      sym__apostrophe,
    STATE(5216), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48930] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3832), 1,
      sym__quotelike_begin,
    STATE(4784), 1,
      sym__apostrophe,
    STATE(5217), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48953] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3833), 1,
      sym__quotelike_begin,
    STATE(4787), 1,
      sym__apostrophe,
    STATE(5218), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48976] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4096), 1,
      sym__quotelike_begin,
    STATE(4608), 1,
      sym__apostrophe,
    STATE(5219), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [48999] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3858), 1,
      sym__quotelike_begin,
    STATE(4736), 1,
      sym__apostrophe,
    STATE(5220), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49022] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7837), 1,
      anon_sym_RPAREN,
    STATE(5221), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49043] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1466), 1,
      sym_block,
    STATE(1469), 1,
      sym__loop_body,
    STATE(5222), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49066] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7839), 1,
      anon_sym_RPAREN,
    STATE(5223), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49087] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7841), 1,
      anon_sym_RBRACK,
    STATE(5224), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49108] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7843), 1,
      anon_sym_RPAREN,
    STATE(5225), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49129] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7845), 1,
      anon_sym_LPAREN,
    STATE(160), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1536), 1,
      sym_block,
    STATE(5226), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49152] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7847), 1,
      anon_sym_RBRACK,
    STATE(5227), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49173] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7849), 1,
      anon_sym_RBRACK,
    STATE(5228), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49194] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1673), 1,
      sym_block,
    STATE(1689), 1,
      sym__loop_body,
    STATE(5229), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49217] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1673), 1,
      sym_block,
    STATE(1690), 1,
      sym__loop_body,
    STATE(5230), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49240] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3883), 1,
      sym__quotelike_begin,
    STATE(4727), 1,
      sym__apostrophe,
    STATE(5231), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49263] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3840), 1,
      sym__quotelike_begin,
    STATE(4743), 1,
      sym__apostrophe,
    STATE(5232), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49286] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3842), 1,
      sym__quotelike_begin,
    STATE(4808), 1,
      sym__apostrophe,
    STATE(5233), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49309] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4055), 1,
      sym__quotelike_begin,
    STATE(4659), 1,
      sym__apostrophe,
    STATE(5234), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49332] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7851), 1,
      anon_sym_RPAREN,
    STATE(5235), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49353] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7853), 1,
      anon_sym_RPAREN,
    STATE(5236), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49374] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7855), 1,
      anon_sym_RPAREN,
    STATE(5237), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49395] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7857), 1,
      anon_sym_RBRACK,
    STATE(5238), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49416] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1576), 1,
      sym_block,
    STATE(1577), 1,
      sym__loop_body,
    STATE(5239), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49439] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1975), 1,
      sym_attribute_name,
    STATE(1978), 1,
      sym_attribute,
    STATE(5240), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4214), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [49460] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5241), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2723), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [49477] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7859), 1,
      anon_sym_RPAREN,
    STATE(5242), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49498] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2063), 1,
      sym_attribute_name,
    STATE(2150), 1,
      sym_attribute,
    STATE(5243), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4240), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [49519] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2019), 1,
      sym_attribute_name,
    STATE(2164), 1,
      sym_attribute,
    STATE(5244), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4250), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [49540] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7861), 1,
      anon_sym_RPAREN,
    STATE(5245), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49561] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7863), 1,
      anon_sym_RBRACK,
    STATE(5246), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49582] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7865), 1,
      anon_sym_RBRACK,
    STATE(5247), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49603] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7515), 1,
      anon_sym_RBRACE,
    STATE(5248), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49624] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7867), 1,
      anon_sym_RBRACK,
    STATE(5249), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49645] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5250), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7869), 4,
      sym__quotelike_middle_close_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [49662] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5251), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2694), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [49679] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3851), 1,
      sym__quotelike_begin,
    STATE(4720), 1,
      sym__apostrophe,
    STATE(5252), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49702] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3852), 1,
      sym__quotelike_begin,
    STATE(4722), 1,
      sym__apostrophe,
    STATE(5253), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49725] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4112), 1,
      sym__quotelike_begin,
    STATE(4623), 1,
      sym__apostrophe,
    STATE(5254), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [49748] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7871), 1,
      anon_sym_RPAREN,
    STATE(5255), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49769] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7873), 1,
      anon_sym_RPAREN,
    STATE(5256), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49790] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7875), 1,
      anon_sym_RPAREN,
    STATE(5257), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49811] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5790), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5258), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49832] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5746), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5259), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49853] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7877), 1,
      anon_sym_RPAREN,
    STATE(5260), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49874] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7879), 1,
      anon_sym_RPAREN,
    STATE(5261), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49895] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7881), 1,
      anon_sym_RBRACK,
    STATE(5262), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49916] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7883), 1,
      anon_sym_RPAREN,
    STATE(5263), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49937] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5776), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5264), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49958] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7885), 1,
      anon_sym_RPAREN,
    STATE(5265), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [49979] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5266), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2715), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [49996] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3938), 1,
      sym__quotelike_begin,
    STATE(4802), 1,
      sym__apostrophe,
    STATE(5267), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50019] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3859), 1,
      sym__quotelike_begin,
    STATE(4803), 1,
      sym__apostrophe,
    STATE(5268), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50042] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3860), 1,
      sym__quotelike_begin,
    STATE(4804), 1,
      sym__apostrophe,
    STATE(5269), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50065] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4059), 1,
      sym__quotelike_begin,
    STATE(4681), 1,
      sym__apostrophe,
    STATE(5270), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50088] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7887), 1,
      anon_sym_RPAREN,
    STATE(5271), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50109] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5778), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5272), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50130] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1673), 1,
      sym_block,
    STATE(1730), 1,
      sym__loop_body,
    STATE(5273), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50153] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2006), 1,
      sym_attribute_name,
    STATE(2035), 1,
      sym_attribute,
    STATE(5274), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4226), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50174] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3827), 1,
      sym__quotelike_begin,
    STATE(4824), 1,
      sym__apostrophe,
    STATE(5275), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50197] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3814), 1,
      sym__quotelike_begin,
    STATE(4683), 1,
      sym__apostrophe,
    STATE(5276), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50220] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5277), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6718), 4,
      sym__quotelike_end_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [50237] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7891), 1,
      anon_sym_RBRACK,
    STATE(5278), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50258] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7893), 1,
      anon_sym_RBRACK,
    STATE(5279), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50279] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7895), 1,
      anon_sym_RBRACK,
    STATE(5280), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50300] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7897), 1,
      anon_sym_RBRACK,
    STATE(5281), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50321] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7899), 1,
      anon_sym_RPAREN,
    STATE(5282), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50342] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5283), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7869), 4,
      sym__quotelike_end_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [50359] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7901), 1,
      anon_sym_RBRACK,
    STATE(5284), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50380] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5285), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7903), 4,
      anon_sym_extended,
      anon_sym_async,
      anon_sym_sub,
      anon_sym_method,
  [50397] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5286), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2686), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [50414] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7905), 1,
      anon_sym_RBRACK,
    STATE(5287), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50435] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1439), 1,
      sym__loop_body,
    STATE(1466), 1,
      sym_block,
    STATE(5288), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50458] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2064), 1,
      sym_attribute_name,
    STATE(2147), 1,
      sym_attribute,
    STATE(5289), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4236), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50479] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7907), 1,
      anon_sym_RPAREN,
    STATE(5290), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50500] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7909), 1,
      anon_sym_RBRACK,
    STATE(5291), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50521] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1451), 1,
      sym__loop_body,
    STATE(1466), 1,
      sym_block,
    STATE(5292), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50544] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7911), 1,
      anon_sym_RBRACK,
    STATE(5293), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50565] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    ACTIONS(7913), 1,
      anon_sym_LPAREN,
    STATE(144), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1434), 1,
      sym_block,
    STATE(5294), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50588] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1537), 1,
      sym__loop_body,
    STATE(1576), 1,
      sym_block,
    STATE(5295), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50611] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7915), 1,
      anon_sym_RPAREN,
    STATE(5296), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50632] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5297), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7917), 4,
      sym__quotelike_end_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [50649] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5769), 1,
      anon_sym_RBRACE,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(5298), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50670] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5299), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2739), 4,
      sym_prototype,
      sym__signature_start,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [50687] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7919), 1,
      anon_sym_RBRACK,
    STATE(5300), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50708] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3824), 1,
      sym__quotelike_begin,
    STATE(4641), 1,
      sym__apostrophe,
    STATE(5301), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [50731] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7921), 1,
      anon_sym_RBRACK,
    STATE(5302), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50752] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7923), 1,
      anon_sym_RBRACK,
    STATE(5303), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50773] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7925), 1,
      anon_sym_RBRACK,
    STATE(5304), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50794] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1744), 1,
      sym_attribute_name,
    STATE(1762), 1,
      sym_attribute,
    STATE(5305), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(3311), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50815] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7927), 1,
      anon_sym_RBRACK,
    STATE(5306), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50836] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7929), 1,
      anon_sym_RPAREN,
    STATE(5307), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50857] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7931), 1,
      anon_sym_RBRACK,
    STATE(5308), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50878] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7933), 1,
      anon_sym_RBRACK,
    STATE(5309), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [50899] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2581), 1,
      sym_attribute_name,
    STATE(2780), 1,
      sym_attribute,
    STATE(5310), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4360), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50920] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4948), 1,
      sym_attribute_name,
    STATE(5117), 1,
      sym_attribute,
    STATE(5311), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6742), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50941] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5312), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2690), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [50958] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(2483), 1,
      sym_attribute_name,
    STATE(2957), 1,
      sym_attribute,
    STATE(5313), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(4382), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [50979] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3811), 1,
      sym__quotelike_begin,
    STATE(4607), 1,
      sym__apostrophe,
    STATE(5314), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51002] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4110), 1,
      sym__quotelike_begin,
    STATE(4643), 1,
      sym__apostrophe,
    STATE(5315), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51025] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5316), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6873), 4,
      sym__quotelike_end_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [51042] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7935), 1,
      anon_sym_RBRACK,
    STATE(5317), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [51063] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5318), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7917), 4,
      sym__quotelike_middle_close_quote,
      sym__q_string_content,
      sym_escape_sequence,
      sym_escaped_delimiter,
  [51080] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3821), 1,
      sym__quotelike_begin,
    STATE(4632), 1,
      sym__apostrophe,
    STATE(5319), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51103] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5320), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2719), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_GT,
  [51120] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3815), 1,
      sym__quotelike_begin,
    STATE(4644), 1,
      sym__apostrophe,
    STATE(5321), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51143] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1576), 1,
      sym_block,
    STATE(1591), 1,
      sym__loop_body,
    STATE(5322), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51166] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3813), 1,
      sym__quotelike_begin,
    STATE(4650), 1,
      sym__apostrophe,
    STATE(5323), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51189] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3817), 1,
      sym__quotelike_begin,
    STATE(4662), 1,
      sym__apostrophe,
    STATE(5324), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51212] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3948), 1,
      sym__quotelike_begin,
    STATE(4780), 1,
      sym__apostrophe,
    STATE(5325), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51235] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7889), 1,
      sym__quotelike_begin_quote,
    STATE(3820), 1,
      sym__quotelike_begin,
    STATE(4668), 1,
      sym__apostrophe,
    STATE(5326), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51258] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3863), 1,
      sym__quotelike_begin,
    STATE(4781), 1,
      sym__apostrophe,
    STATE(5327), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51281] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7795), 1,
      sym__quotelike_begin_quote,
    STATE(3864), 1,
      sym__quotelike_begin,
    STATE(4783), 1,
      sym__apostrophe,
    STATE(5328), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51304] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7799), 1,
      sym__single_quote,
    ACTIONS(7801), 1,
      sym__quotelike_begin_quote,
    STATE(4115), 1,
      sym__quotelike_begin,
    STATE(4596), 1,
      sym__apostrophe,
    STATE(5329), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51327] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7194), 1,
      anon_sym_and,
    ACTIONS(7937), 1,
      anon_sym_RBRACK,
    STATE(5330), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7761), 2,
      anon_sym_or,
      anon_sym_xor,
  [51348] = 7,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(131), 1,
      sym__single_quote,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3911), 1,
      sym__quotelike_begin,
    STATE(4717), 1,
      sym__apostrophe,
    STATE(5331), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51371] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2928), 1,
      sym_block,
    STATE(5332), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51391] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2888), 1,
      sym_block,
    STATE(5333), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51411] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(163), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1424), 1,
      sym_block,
    STATE(5334), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51431] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3189), 1,
      sym_block,
    STATE(5335), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51451] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3190), 1,
      sym_block,
    STATE(5336), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51471] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2977), 1,
      sym_block,
    STATE(5337), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51491] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3330), 1,
      sym_block,
    STATE(5338), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51511] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3331), 1,
      sym_block,
    STATE(5339), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51531] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2979), 1,
      sym_block,
    STATE(5340), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51551] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4678), 1,
      sym_bareword,
    STATE(5341), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [51569] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4679), 1,
      sym_bareword,
    STATE(5342), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [51587] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2889), 1,
      sym_block,
    STATE(5343), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51607] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1458), 1,
      sym_block,
    STATE(5344), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51627] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1518), 1,
      sym_block,
    STATE(5345), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51647] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(168), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1435), 1,
      sym_block,
    STATE(5346), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51667] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1544), 1,
      sym__semicolon,
    STATE(5347), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [51685] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5348), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7939), 3,
      anon_sym_DOLLAR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [51701] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7941), 1,
      anon_sym_LBRACE,
    ACTIONS(7943), 1,
      anon_sym_STAR,
    ACTIONS(7945), 1,
      anon_sym_LBRACK,
    STATE(5349), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51721] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1522), 1,
      sym_block,
    STATE(5350), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51741] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3001), 1,
      sym_block,
    STATE(5351), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51761] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3002), 1,
      sym_block,
    STATE(5352), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51781] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1511), 1,
      sym_block,
    STATE(5353), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51801] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1490), 1,
      sym_block,
    STATE(5354), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51821] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3635), 1,
      sym_block,
    STATE(5355), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51841] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3207), 1,
      sym_block,
    STATE(5356), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51861] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2318), 1,
      sym_block,
    STATE(5357), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51881] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3209), 1,
      sym_block,
    STATE(5358), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51901] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2370), 1,
      sym_block,
    STATE(5359), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51921] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1500), 1,
      sym__semicolon,
    STATE(5360), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [51939] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1494), 1,
      sym_block,
    STATE(5361), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51959] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1578), 1,
      sym_block,
    STATE(5362), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51979] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3636), 1,
      sym_block,
    STATE(5363), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [51999] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    ACTIONS(7947), 1,
      anon_sym_LPAREN,
    STATE(5364), 1,
      sym_heredoc_content,
    STATE(5770), 1,
      sym_scalar,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52019] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4600), 1,
      sym_bareword,
    STATE(5365), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52037] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1485), 1,
      sym_block,
    STATE(5366), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52057] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4550), 1,
      sym_package,
    STATE(5367), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52075] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2377), 1,
      sym_block,
    STATE(5368), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52095] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1480), 1,
      sym_block,
    STATE(5369), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52115] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(157), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1635), 1,
      sym_block,
    STATE(5370), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52135] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1722), 1,
      sym_block,
    STATE(5371), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52155] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1579), 1,
      sym_block,
    STATE(5372), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52175] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1895), 1,
      sym_block,
    STATE(5373), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52195] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1470), 1,
      sym_block,
    STATE(5374), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52215] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1727), 1,
      sym_block,
    STATE(5375), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52235] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1495), 1,
      sym_block,
    STATE(5376), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52255] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1571), 1,
      sym_block,
    STATE(5377), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52275] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7951), 1,
      anon_sym_LBRACE,
    ACTIONS(7953), 1,
      anon_sym_STAR,
    ACTIONS(7955), 1,
      anon_sym_LBRACK,
    STATE(5378), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52295] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7957), 1,
      anon_sym_LBRACE,
    ACTIONS(7959), 1,
      anon_sym_STAR,
    ACTIONS(7961), 1,
      anon_sym_LBRACK,
    STATE(5379), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52315] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7963), 1,
      anon_sym_LBRACE,
    ACTIONS(7965), 1,
      anon_sym_STAR,
    ACTIONS(7967), 1,
      anon_sym_LBRACK,
    STATE(5380), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52335] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1582), 1,
      sym_block,
    STATE(5381), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52355] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1720), 1,
      sym_block,
    STATE(5382), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52375] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1672), 1,
      sym__semicolon,
    STATE(5383), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [52393] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1584), 1,
      sym_block,
    STATE(5384), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52413] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2897), 1,
      sym_block,
    STATE(5385), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52433] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2899), 1,
      sym_block,
    STATE(5386), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52453] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3230), 1,
      sym_block,
    STATE(5387), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52473] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3231), 1,
      sym_block,
    STATE(5388), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52493] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1586), 1,
      sym_block,
    STATE(5389), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52513] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1440), 1,
      sym_block,
    STATE(5390), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52533] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1725), 1,
      sym__semicolon,
    STATE(5391), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [52551] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1587), 1,
      sym_block,
    STATE(5392), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52571] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4523), 1,
      sym_package,
    STATE(5393), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52589] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4421), 1,
      sym_package,
    STATE(5394), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52607] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4422), 1,
      sym_package,
    STATE(5395), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52625] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    ACTIONS(7969), 1,
      sym__quotelike_middle_skip,
    STATE(4530), 1,
      sym__quotelike_begin,
    STATE(5396), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52645] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3140), 1,
      sym_block,
    STATE(5397), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52665] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1588), 1,
      sym_block,
    STATE(5398), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52685] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1589), 1,
      sym_block,
    STATE(5399), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52705] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2328), 1,
      sym_block,
    STATE(5400), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52725] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1590), 1,
      sym_block,
    STATE(5401), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52745] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4619), 1,
      sym_bareword,
    STATE(5402), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52763] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4620), 1,
      sym_bareword,
    STATE(5403), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52781] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7971), 1,
      anon_sym_LBRACE,
    ACTIONS(7973), 1,
      anon_sym_STAR,
    ACTIONS(7975), 1,
      anon_sym_LBRACK,
    STATE(5404), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52801] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2624), 1,
      sym_block,
    STATE(5405), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52821] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1573), 1,
      sym_block,
    STATE(5406), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52841] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3665), 1,
      sym_block,
    STATE(5407), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52861] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1647), 1,
      sym_block,
    STATE(5408), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52881] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4630), 1,
      sym_bareword,
    STATE(5409), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52899] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4631), 1,
      sym_bareword,
    STATE(5410), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [52917] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3022), 1,
      sym_block,
    STATE(5411), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52937] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1651), 1,
      sym_block,
    STATE(5412), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52957] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2642), 1,
      sym_block,
    STATE(5413), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52977] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3024), 1,
      sym_block,
    STATE(5414), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [52997] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(150), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2643), 1,
      sym_block,
    STATE(5415), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53017] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1652), 1,
      sym_block,
    STATE(5416), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53037] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1654), 1,
      sym_block,
    STATE(5417), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53057] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7977), 1,
      anon_sym_LBRACE,
    ACTIONS(7979), 1,
      anon_sym_STAR,
    ACTIONS(7981), 1,
      anon_sym_LBRACK,
    STATE(5418), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53077] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3666), 1,
      sym_block,
    STATE(5419), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53097] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2442), 1,
      sym_block,
    STATE(5420), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53117] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1595), 1,
      sym_block,
    STATE(5421), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53137] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2447), 1,
      sym_block,
    STATE(5422), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53157] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1597), 1,
      sym_block,
    STATE(5423), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53177] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7983), 1,
      anon_sym_LBRACE,
    ACTIONS(7985), 1,
      anon_sym_STAR,
    ACTIONS(7987), 1,
      anon_sym_LBRACK,
    STATE(5424), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53197] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7989), 1,
      anon_sym_LBRACE,
    ACTIONS(7991), 1,
      anon_sym_STAR,
    ACTIONS(7993), 1,
      anon_sym_LBRACK,
    STATE(5425), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53217] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    ACTIONS(7997), 1,
      sym__quotelike_middle_skip,
    STATE(5316), 1,
      sym__quotelike_begin,
    STATE(5426), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53237] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1599), 1,
      sym_block,
    STATE(5427), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53257] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1600), 1,
      sym_block,
    STATE(5428), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53277] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1601), 1,
      sym_block,
    STATE(5429), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53297] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1663), 1,
      sym_block,
    STATE(5430), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53317] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1664), 1,
      sym_block,
    STATE(5431), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53337] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1665), 1,
      sym_block,
    STATE(5432), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53357] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3047), 1,
      sym_block,
    STATE(5433), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53377] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1668), 1,
      sym_block,
    STATE(5434), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53397] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3048), 1,
      sym_block,
    STATE(5435), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53417] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1603), 1,
      sym_block,
    STATE(5436), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53437] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(160), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1533), 1,
      sym_block,
    STATE(5437), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53457] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1670), 1,
      sym_block,
    STATE(5438), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53477] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1606), 1,
      sym_block,
    STATE(5439), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53497] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3377), 1,
      sym_block,
    STATE(5440), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53517] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1853), 1,
      sym_block,
    STATE(5441), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53537] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1575), 1,
      sym__semicolon,
    STATE(5442), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [53555] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1464), 1,
      sym_block,
    STATE(5443), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53575] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1551), 1,
      sym_block,
    STATE(5444), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53595] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1833), 1,
      sym_block,
    STATE(5445), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53615] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1607), 1,
      sym_block,
    STATE(5446), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53635] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1608), 1,
      sym_block,
    STATE(5447), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53655] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3501), 1,
      sym_block,
    STATE(5448), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53675] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(140), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1528), 1,
      sym_block,
    STATE(5449), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53695] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1625), 1,
      sym_block,
    STATE(5450), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53715] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1609), 1,
      sym_block,
    STATE(5451), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53735] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2407), 1,
      sym_block,
    STATE(5452), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53755] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2930), 1,
      sym_block,
    STATE(5453), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53775] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3576), 1,
      sym_block,
    STATE(5454), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53795] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1610), 1,
      sym_block,
    STATE(5455), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53815] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(163), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1423), 1,
      sym_block,
    STATE(5456), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53835] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1611), 1,
      sym_block,
    STATE(5457), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53855] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1910), 1,
      sym_block,
    STATE(5458), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53875] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1555), 1,
      sym_block,
    STATE(5459), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53895] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(206), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1629), 1,
      sym_block,
    STATE(5460), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53915] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3059), 1,
      sym_block,
    STATE(5461), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53935] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3577), 1,
      sym_block,
    STATE(5462), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53955] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1484), 1,
      sym_block,
    STATE(5463), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [53975] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4597), 1,
      sym_bareword,
    STATE(5464), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [53993] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1848), 1,
      sym_block,
    STATE(5465), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54013] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1915), 1,
      sym_block,
    STATE(5466), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54033] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1612), 1,
      sym_block,
    STATE(5467), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54053] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1486), 1,
      sym_block,
    STATE(5468), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54073] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1676), 1,
      sym_block,
    STATE(5469), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54093] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3430), 1,
      sym_block,
    STATE(5470), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54113] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(168), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1431), 1,
      sym_block,
    STATE(5471), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54133] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1556), 1,
      sym_block,
    STATE(5472), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54153] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3646), 1,
      sym_block,
    STATE(5473), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54173] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3061), 1,
      sym_block,
    STATE(5474), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54193] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3431), 1,
      sym_block,
    STATE(5475), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54213] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1677), 1,
      sym_block,
    STATE(5476), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54233] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1498), 1,
      sym_block,
    STATE(5477), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54253] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1680), 1,
      sym_block,
    STATE(5478), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54273] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1467), 1,
      sym_block,
    STATE(5479), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54293] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1465), 1,
      sym__semicolon,
    STATE(5480), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [54311] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1682), 1,
      sym_block,
    STATE(5481), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54331] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1684), 1,
      sym_block,
    STATE(5482), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54351] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2708), 1,
      sym_block,
    STATE(5483), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54371] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1685), 1,
      sym_block,
    STATE(5484), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54391] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2709), 1,
      sym_block,
    STATE(5485), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54411] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1686), 1,
      sym_block,
    STATE(5486), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54431] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1687), 1,
      sym_block,
    STATE(5487), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54451] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1688), 1,
      sym_block,
    STATE(5488), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54471] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1887), 1,
      sym_block,
    STATE(5489), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54491] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3482), 1,
      sym_block,
    STATE(5490), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54511] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4405), 1,
      sym_package,
    STATE(5491), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [54529] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1557), 1,
      sym_block,
    STATE(5492), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54549] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3648), 1,
      sym_block,
    STATE(5493), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54569] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1655), 1,
      sym__semicolon,
    STATE(5494), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2175), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [54587] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1626), 1,
      sym_block,
    STATE(5495), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54607] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1872), 1,
      sym_block,
    STATE(5496), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54627] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1540), 1,
      sym_block,
    STATE(5497), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54647] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1566), 1,
      sym_block,
    STATE(5498), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54667] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4688), 1,
      sym_bareword,
    STATE(5499), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [54685] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3083), 1,
      sym_block,
    STATE(5500), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54705] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3510), 1,
      sym_block,
    STATE(5501), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54725] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1693), 1,
      sym_block,
    STATE(5502), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54745] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(162), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3084), 1,
      sym_block,
    STATE(5503), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54765] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1695), 1,
      sym_block,
    STATE(5504), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54785] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2746), 1,
      sym_block,
    STATE(5505), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54805] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1697), 1,
      sym_block,
    STATE(5506), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54825] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2748), 1,
      sym_block,
    STATE(5507), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54845] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1501), 1,
      sym_block,
    STATE(5508), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54865] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1698), 1,
      sym_block,
    STATE(5509), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54885] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1446), 1,
      sym_block,
    STATE(5510), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54905] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1506), 1,
      sym_block,
    STATE(5511), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54925] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7999), 1,
      anon_sym_LBRACE,
    ACTIONS(8001), 1,
      anon_sym_STAR,
    ACTIONS(8003), 1,
      anon_sym_LBRACK,
    STATE(5512), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54945] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1699), 1,
      sym_block,
    STATE(5513), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54965] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3512), 1,
      sym_block,
    STATE(5514), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [54985] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1510), 1,
      sym_block,
    STATE(5515), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55005] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4480), 1,
      sym_package,
    STATE(5516), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [55023] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1701), 1,
      sym_block,
    STATE(5517), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55043] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3609), 1,
      sym_block,
    STATE(5518), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55063] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1513), 1,
      sym_block,
    STATE(5519), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55083] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(170), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1638), 1,
      sym_block,
    STATE(5520), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55103] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8005), 1,
      anon_sym_LBRACE,
    ACTIONS(8007), 1,
      anon_sym_STAR,
    ACTIONS(8009), 1,
      anon_sym_LBRACK,
    STATE(5521), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55123] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1642), 1,
      sym_block,
    STATE(5522), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55143] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8011), 1,
      anon_sym_LBRACE,
    ACTIONS(8013), 1,
      anon_sym_STAR,
    ACTIONS(8015), 1,
      anon_sym_LBRACK,
    STATE(5523), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55163] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1704), 1,
      sym_block,
    STATE(5524), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55183] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1705), 1,
      sym_block,
    STATE(5525), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55203] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1890), 1,
      sym_block,
    STATE(5526), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55223] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1706), 1,
      sym_block,
    STATE(5527), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55243] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1707), 1,
      sym_block,
    STATE(5528), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55263] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1708), 1,
      sym_block,
    STATE(5529), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55283] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1514), 1,
      sym_block,
    STATE(5530), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55303] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8017), 1,
      anon_sym_LBRACE,
    ACTIONS(8019), 1,
      anon_sym_STAR,
    ACTIONS(8021), 1,
      anon_sym_LBRACK,
    STATE(5531), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55323] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(155), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1709), 1,
      sym_block,
    STATE(5532), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55343] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1515), 1,
      sym_block,
    STATE(5533), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55363] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1477), 1,
      sym_block,
    STATE(5534), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55383] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4700), 1,
      sym_bareword,
    STATE(5535), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [55401] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1468), 1,
      sym_block,
    STATE(5536), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55421] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2353), 1,
      sym_block,
    STATE(5537), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55441] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5538), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7698), 3,
      anon_sym_DOLLAR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [55457] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(206), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1630), 1,
      sym_block,
    STATE(5539), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55477] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8023), 1,
      anon_sym_LBRACE,
    ACTIONS(8025), 1,
      anon_sym_STAR,
    ACTIONS(8027), 1,
      anon_sym_LBRACK,
    STATE(5540), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55497] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1436), 1,
      sym_block,
    STATE(5541), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55517] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8029), 1,
      anon_sym_LBRACE,
    ACTIONS(8031), 1,
      anon_sym_STAR,
    ACTIONS(8033), 1,
      anon_sym_LBRACK,
    STATE(5542), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55537] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2397), 1,
      sym_block,
    STATE(5543), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55557] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(147), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3611), 1,
      sym_block,
    STATE(5544), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55577] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2784), 1,
      sym_block,
    STATE(5545), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55597] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2785), 1,
      sym_block,
    STATE(5546), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55617] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8035), 1,
      anon_sym_LBRACE,
    ACTIONS(8037), 1,
      anon_sym_STAR,
    ACTIONS(8039), 1,
      anon_sym_LBRACK,
    STATE(5547), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55637] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1488), 1,
      sym_block,
    STATE(5548), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55657] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(139), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1862), 1,
      sym_block,
    STATE(5549), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55677] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1517), 1,
      sym_block,
    STATE(5550), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55697] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1460), 1,
      sym_block,
    STATE(5551), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55717] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2799), 1,
      sym_block,
    STATE(5552), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55737] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2801), 1,
      sym_block,
    STATE(5553), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55757] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1567), 1,
      sym_block,
    STATE(5554), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55777] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1524), 1,
      sym__semicolon,
    STATE(5555), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2171), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [55795] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(137), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1568), 1,
      sym_block,
    STATE(5556), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55815] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1526), 1,
      sym_block,
    STATE(5557), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55835] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(205), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1428), 1,
      sym_block,
    STATE(5558), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55855] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8041), 1,
      anon_sym_LBRACE,
    ACTIONS(8043), 1,
      anon_sym_STAR,
    ACTIONS(8045), 1,
      anon_sym_LBRACK,
    STATE(5559), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55875] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1438), 1,
      sym_block,
    STATE(5560), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55895] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4633), 1,
      sym_bareword,
    STATE(5561), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [55913] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(144), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1432), 1,
      sym_block,
    STATE(5562), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55933] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2960), 1,
      sym_block,
    STATE(5563), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [55953] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(1558), 1,
      sym__semicolon,
    STATE(5564), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(2179), 2,
      sym__PERLY_SEMICOLON,
      anon_sym_SEMI,
  [55971] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4533), 1,
      sym_package,
    STATE(5565), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [55989] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4466), 1,
      sym_package,
    STATE(5566), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [56007] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4467), 1,
      sym_package,
    STATE(5567), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7949), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [56025] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1502), 1,
      sym_block,
    STATE(5568), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56045] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(166), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2961), 1,
      sym_block,
    STATE(5569), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56065] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2821), 1,
      sym_block,
    STATE(5570), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56085] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2822), 1,
      sym_block,
    STATE(5571), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56105] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(3481), 1,
      sym_block,
    STATE(5572), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56125] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1475), 1,
      sym_block,
    STATE(5573), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56145] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(4692), 1,
      sym_bareword,
    STATE(5574), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(6686), 2,
      sym__identifier,
      aux_sym__bareword_token1,
  [56163] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(158), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(1505), 1,
      sym_block,
    STATE(5575), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56183] = 6,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6528), 1,
      anon_sym_LBRACE,
    STATE(149), 1,
      sym__PERLY_BRACE_OPEN,
    STATE(2438), 1,
      sym_block,
    STATE(5576), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56203] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5577), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8047), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56218] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8049), 1,
      sym__heredoc_delimiter,
    ACTIONS(8051), 1,
      sym__command_heredoc_delimiter,
    STATE(5578), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56235] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5579), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8053), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56250] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8055), 1,
      sym__heredoc_delimiter,
    ACTIONS(8057), 1,
      sym__command_heredoc_delimiter,
    STATE(5580), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56267] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6483), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5581), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56284] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8061), 1,
      sym__heredoc_delimiter,
    ACTIONS(8063), 1,
      sym__command_heredoc_delimiter,
    STATE(5582), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56301] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8065), 1,
      anon_sym_RPAREN,
    STATE(5583), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56318] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4961), 1,
      sym__quotelike_begin,
    STATE(5584), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56335] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3936), 1,
      sym__quotelike_begin,
    STATE(5585), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56352] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4785), 1,
      sym__quotelike_begin,
    STATE(5586), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56369] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5587), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8067), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56384] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6471), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5588), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56401] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5589), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8069), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56416] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5590), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8071), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56431] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5591), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8073), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56446] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5592), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8075), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56461] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5593), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8077), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56476] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8079), 1,
      sym__heredoc_delimiter,
    ACTIONS(8081), 1,
      sym__command_heredoc_delimiter,
    STATE(5594), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56493] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5595), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8083), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56508] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6964), 1,
      anon_sym_DOLLAR,
    STATE(5596), 1,
      sym_heredoc_content,
    STATE(5770), 1,
      sym_scalar,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56525] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5597), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8085), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56540] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6455), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5598), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56557] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6672), 1,
      sym__identifier,
    STATE(4543), 1,
      sym__signature_varname,
    STATE(5599), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56574] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8087), 1,
      anon_sym_LBRACE,
    ACTIONS(8089), 1,
      anon_sym_STAR,
    STATE(5600), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56591] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5601), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8091), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56606] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5602), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8093), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56621] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8095), 1,
      anon_sym_RPAREN,
    STATE(5603), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56638] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5604), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8097), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56653] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5605), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8099), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56668] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8101), 1,
      anon_sym_LBRACE,
    ACTIONS(8103), 1,
      anon_sym_LBRACK,
    STATE(5606), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56685] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8105), 1,
      anon_sym_RPAREN,
    STATE(5607), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56702] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5608), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8107), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56717] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8109), 1,
      anon_sym_RPAREN,
    STATE(5609), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56734] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4772), 1,
      sym__quotelike_begin,
    STATE(5610), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56751] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3891), 1,
      sym__quotelike_begin,
    STATE(5611), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56768] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4705), 1,
      sym__quotelike_begin,
    STATE(5612), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56785] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3940), 1,
      sym__quotelike_begin,
    STATE(5613), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56802] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7364), 1,
      anon_sym_RPAREN,
    ACTIONS(8111), 1,
      anon_sym_COMMA,
    STATE(5614), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56819] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8113), 1,
      sym__heredoc_delimiter,
    ACTIONS(8115), 1,
      sym__command_heredoc_delimiter,
    STATE(5615), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56836] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5616), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8117), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56851] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5617), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8119), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56866] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8121), 1,
      anon_sym_LBRACE,
    ACTIONS(8123), 1,
      anon_sym_LBRACK,
    STATE(5618), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56883] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4883), 1,
      sym__quotelike_begin,
    STATE(5619), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56900] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4936), 1,
      sym__quotelike_begin,
    STATE(5620), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56917] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3882), 1,
      sym__quotelike_begin,
    STATE(5621), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56934] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4726), 1,
      sym__quotelike_begin,
    STATE(5622), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56951] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8125), 1,
      anon_sym_LBRACE,
    ACTIONS(8127), 1,
      anon_sym_STAR,
    STATE(5623), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [56968] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5624), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8129), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56983] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5625), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8131), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [56998] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5626), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8133), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57013] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5627), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8135), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57028] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5628), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8137), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57043] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5629), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8139), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57058] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5630), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8141), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57073] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5631), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8143), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57088] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5632), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8145), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57103] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4777), 1,
      sym__quotelike_begin,
    STATE(5633), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57120] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4713), 1,
      sym__quotelike_begin,
    STATE(5634), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57137] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3910), 1,
      sym__quotelike_begin,
    STATE(5635), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57154] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4716), 1,
      sym__quotelike_begin,
    STATE(5636), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57171] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5637), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8147), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57186] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3947), 1,
      sym__quotelike_begin,
    STATE(5638), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57203] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4779), 1,
      sym__quotelike_begin,
    STATE(5639), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57220] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6424), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5640), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57237] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8149), 1,
      sym__heredoc_delimiter,
    ACTIONS(8151), 1,
      sym__command_heredoc_delimiter,
    STATE(5641), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57254] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5642), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8153), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57269] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6512), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5643), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57286] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6508), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5644), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57303] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5645), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8155), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57318] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5646), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8157), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57333] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8159), 1,
      anon_sym_RPAREN,
    STATE(5647), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57350] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4798), 1,
      sym__quotelike_begin,
    STATE(5648), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57367] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3937), 1,
      sym__quotelike_begin,
    STATE(5649), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57384] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4801), 1,
      sym__quotelike_begin,
    STATE(5650), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57401] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8161), 1,
      sym__heredoc_delimiter,
    ACTIONS(8163), 1,
      sym__command_heredoc_delimiter,
    STATE(5651), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57418] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5652), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8165), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57433] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8167), 1,
      anon_sym_LBRACE,
    ACTIONS(8169), 1,
      anon_sym_STAR,
    STATE(5653), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57450] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6506), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5654), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57467] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5655), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8171), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57482] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5656), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8173), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57497] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5657), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8175), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57512] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5658), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8177), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57527] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5659), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8179), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57542] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8181), 1,
      anon_sym_LBRACE,
    ACTIONS(8183), 1,
      anon_sym_STAR,
    STATE(5660), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57559] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8185), 1,
      anon_sym_DOLLAR,
    STATE(4532), 1,
      sym__signature_scalar,
    STATE(5661), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57576] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5662), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8187), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57591] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5663), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8189), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57606] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5664), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8191), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57621] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5665), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8193), 2,
      sym__quotelike_begin_quote,
      sym__quotelike_middle_skip,
  [57636] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6510), 1,
      anon_sym_RPAREN,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5666), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57653] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5667), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8195), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57668] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8111), 1,
      anon_sym_COMMA,
    ACTIONS(8197), 1,
      anon_sym_RPAREN,
    STATE(5668), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57685] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8199), 1,
      anon_sym_LBRACE,
    ACTIONS(8201), 1,
      anon_sym_STAR,
    STATE(5669), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57702] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5670), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8203), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57717] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8205), 1,
      anon_sym_LBRACE,
    ACTIONS(8207), 1,
      anon_sym_STAR,
    STATE(5671), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57734] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8209), 1,
      anon_sym_RPAREN,
    STATE(5672), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57751] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5673), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8211), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57766] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8213), 1,
      sym__heredoc_delimiter,
    ACTIONS(8215), 1,
      sym__command_heredoc_delimiter,
    STATE(5674), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57783] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5675), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8217), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57798] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5676), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8219), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57813] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8221), 1,
      anon_sym_LBRACE,
    ACTIONS(8223), 1,
      anon_sym_STAR,
    STATE(5677), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57830] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5678), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8225), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57845] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5679), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8227), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57860] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8229), 1,
      anon_sym_LBRACE,
    ACTIONS(8231), 1,
      anon_sym_LBRACK,
    STATE(5680), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57877] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8233), 1,
      anon_sym_LBRACE,
    ACTIONS(8235), 1,
      anon_sym_LBRACK,
    STATE(5681), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57894] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8237), 1,
      anon_sym_LBRACE,
    ACTIONS(8239), 1,
      anon_sym_LBRACK,
    STATE(5682), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57911] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8241), 1,
      anon_sym_LBRACE,
    ACTIONS(8243), 1,
      anon_sym_LBRACK,
    STATE(5683), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57928] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8245), 1,
      anon_sym_LBRACE,
    ACTIONS(8247), 1,
      anon_sym_LBRACK,
    STATE(5684), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57945] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8249), 1,
      anon_sym_RPAREN,
    STATE(5685), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57962] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5686), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8251), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [57977] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8253), 1,
      anon_sym_LBRACE,
    ACTIONS(8255), 1,
      anon_sym_LBRACK,
    STATE(5687), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [57994] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    ACTIONS(8257), 1,
      anon_sym_RPAREN,
    STATE(5688), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58011] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5689), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8259), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58026] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4796), 1,
      sym__quotelike_begin,
    STATE(5690), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58043] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5691), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(7515), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58058] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5692), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8261), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58073] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5693), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8263), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58088] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8265), 1,
      anon_sym_LBRACE,
    ACTIONS(8267), 1,
      anon_sym_LBRACK,
    STATE(5694), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58105] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8269), 1,
      anon_sym_LBRACE,
    ACTIONS(8271), 1,
      anon_sym_LBRACK,
    STATE(5695), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58122] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8273), 1,
      anon_sym_LBRACE,
    ACTIONS(8275), 1,
      anon_sym_LBRACK,
    STATE(5696), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58139] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5697), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8277), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58154] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5698), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8279), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58169] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5699), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8281), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58184] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5700), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8283), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58199] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8285), 1,
      anon_sym_LBRACE,
    ACTIONS(8287), 1,
      anon_sym_LBRACK,
    STATE(5701), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58216] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5702), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8289), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58231] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8291), 1,
      anon_sym_LBRACE,
    ACTIONS(8293), 1,
      anon_sym_LBRACK,
    STATE(5703), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58248] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8295), 1,
      anon_sym_LBRACE,
    ACTIONS(8297), 1,
      anon_sym_LBRACK,
    STATE(5704), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58265] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8299), 1,
      anon_sym_LBRACE,
    ACTIONS(8301), 1,
      anon_sym_LBRACK,
    STATE(5705), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58282] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5706), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8303), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58297] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8305), 1,
      anon_sym_LBRACE,
    ACTIONS(8307), 1,
      anon_sym_LBRACK,
    STATE(5707), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58314] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8309), 1,
      anon_sym_LBRACE,
    ACTIONS(8311), 1,
      anon_sym_LBRACK,
    STATE(5708), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58331] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8313), 1,
      anon_sym_LBRACE,
    ACTIONS(8315), 1,
      anon_sym_LBRACK,
    STATE(5709), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58348] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8317), 1,
      anon_sym_LBRACE,
    ACTIONS(8319), 1,
      anon_sym_LBRACK,
    STATE(5710), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58365] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4896), 1,
      sym__quotelike_begin,
    STATE(5711), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58382] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7765), 1,
      sym__quotelike_begin_quote,
    STATE(3896), 1,
      sym__quotelike_begin,
    STATE(5712), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58399] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8321), 1,
      anon_sym_LBRACE,
    ACTIONS(8323), 1,
      anon_sym_LBRACK,
    STATE(5713), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58416] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8325), 1,
      anon_sym_LBRACE,
    ACTIONS(8327), 1,
      anon_sym_LBRACK,
    STATE(5714), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58433] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8329), 1,
      anon_sym_LBRACE,
    ACTIONS(8331), 1,
      anon_sym_LBRACK,
    STATE(5715), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58450] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7995), 1,
      sym__quotelike_begin_quote,
    STATE(4923), 1,
      sym__quotelike_begin,
    STATE(5716), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58467] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8333), 1,
      anon_sym_LBRACE,
    ACTIONS(8335), 1,
      anon_sym_LBRACK,
    STATE(5717), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58484] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8337), 1,
      anon_sym_LBRACE,
    ACTIONS(8339), 1,
      anon_sym_LBRACK,
    STATE(5718), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58501] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8341), 1,
      anon_sym_LBRACE,
    ACTIONS(8343), 1,
      anon_sym_LBRACK,
    STATE(5719), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58518] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    STATE(5720), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
    ACTIONS(8345), 2,
      sym__RECOVER_BRACE_CLOSE,
      anon_sym_RBRACE,
  [58533] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8347), 1,
      anon_sym_LBRACE,
    ACTIONS(8349), 1,
      anon_sym_LBRACK,
    STATE(5721), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58550] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8351), 1,
      anon_sym_LBRACE,
    ACTIONS(8353), 1,
      anon_sym_LBRACK,
    STATE(5722), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58567] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8355), 1,
      anon_sym_LBRACE,
    ACTIONS(8357), 1,
      anon_sym_LBRACK,
    STATE(5723), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58584] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8185), 1,
      anon_sym_DOLLAR,
    STATE(4589), 1,
      sym__signature_scalar,
    STATE(5724), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58601] = 5,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8359), 1,
      anon_sym_LBRACE,
    ACTIONS(8361), 1,
      anon_sym_STAR,
    STATE(5725), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58618] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8363), 1,
      anon_sym_RPAREN,
    STATE(5726), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58632] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8365), 1,
      anon_sym_RPAREN,
    STATE(5727), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58646] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8367), 1,
      anon_sym_GT,
    STATE(5728), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58660] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8369), 1,
      aux_sym__interpolated_transliteration_content_token1,
    STATE(5729), 1,
      sym_heredoc_content,
  [58676] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8371), 1,
      sym__ERROR,
    STATE(5730), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58690] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6372), 1,
      sym__no_interp_whitespace_zw,
    STATE(5731), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58704] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8373), 1,
      anon_sym_RPAREN,
    STATE(5732), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58718] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8375), 1,
      anon_sym_STAR,
    STATE(5733), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58732] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8377), 1,
      anon_sym_RBRACE,
    STATE(5734), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58746] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8379), 1,
      anon_sym_STAR,
    STATE(5735), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58760] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3299), 1,
      anon_sym_LPAREN,
    STATE(5736), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58774] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2767), 1,
      sym__brace_end_zw,
    STATE(5737), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58788] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8381), 1,
      anon_sym_RBRACE,
    STATE(5738), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58802] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8383), 1,
      anon_sym_RBRACE,
    STATE(5739), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58816] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6184), 1,
      sym__dollar_ident_zw,
    STATE(5740), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58830] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8385), 1,
      anon_sym_RBRACE,
    STATE(5741), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58844] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8387), 1,
      anon_sym_RBRACE,
    STATE(5742), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58858] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5790), 1,
      anon_sym_RBRACE,
    STATE(5743), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58872] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8389), 1,
      anon_sym_RBRACE,
    STATE(5744), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58886] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8391), 1,
      anon_sym_RBRACE,
    STATE(5745), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58900] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8393), 1,
      anon_sym_RBRACE,
    STATE(5746), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58914] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5776), 1,
      anon_sym_RBRACE,
    STATE(5747), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58928] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8395), 1,
      anon_sym_LPAREN,
    STATE(5748), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58942] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8397), 1,
      anon_sym_RBRACE,
    STATE(5749), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58956] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5746), 1,
      anon_sym_RBRACE,
    STATE(5750), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58970] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8107), 1,
      anon_sym_RBRACE,
    STATE(5751), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58984] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6332), 1,
      sym__dollar_ident_zw,
    STATE(5752), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [58998] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8399), 1,
      aux_sym___DATA___token1,
    STATE(5753), 1,
      sym_heredoc_content,
  [59014] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7515), 1,
      anon_sym_RBRACE,
    STATE(5754), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59028] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8401), 1,
      sym_attribute_value,
    STATE(5755), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59042] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8403), 1,
      anon_sym_RBRACE,
    STATE(5756), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59056] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8405), 1,
      sym__ERROR,
    STATE(5757), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59070] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8407), 1,
      anon_sym_RBRACE,
    STATE(5758), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59084] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3291), 1,
      anon_sym_LPAREN,
    STATE(5759), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59098] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8409), 1,
      sym__ERROR,
    STATE(5760), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59112] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2759), 1,
      sym__brace_end_zw,
    STATE(5761), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59126] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8411), 1,
      anon_sym_RBRACE,
    STATE(5762), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59140] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8413), 1,
      anon_sym_RBRACE,
    STATE(5763), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59154] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8415), 1,
      anon_sym_GT,
    STATE(5764), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59168] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8417), 1,
      anon_sym_RBRACE,
    STATE(5765), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59182] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8419), 1,
      aux_sym__interpolated_transliteration_content_token1,
    STATE(5766), 1,
      sym_heredoc_content,
  [59198] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8421), 1,
      sym__ERROR,
    STATE(5767), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59212] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8423), 1,
      sym__NONASSOC,
    STATE(5768), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59226] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8425), 1,
      anon_sym_RBRACE,
    STATE(5769), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59240] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8427), 1,
      anon_sym_LPAREN,
    STATE(5770), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59254] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7330), 1,
      anon_sym_GT_GT,
    STATE(5771), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59268] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8429), 1,
      anon_sym_GT,
    STATE(5772), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59282] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8431), 1,
      anon_sym_RPAREN,
    STATE(5773), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59296] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8433), 1,
      anon_sym_RBRACE,
    STATE(5774), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59310] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8435), 1,
      anon_sym_RPAREN,
    STATE(5775), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59324] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8437), 1,
      anon_sym_RBRACE,
    STATE(5776), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59338] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6340), 1,
      sym__dollar_ident_zw,
    STATE(5777), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59352] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8439), 1,
      anon_sym_RBRACE,
    STATE(5778), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59366] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8441), 1,
      sym__gobbled_content,
    STATE(5779), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59380] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2773), 1,
      sym__brace_end_zw,
    STATE(5780), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59394] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8443), 1,
      anon_sym_RBRACE,
    STATE(5781), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59408] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8445), 1,
      sym__NONASSOC,
    STATE(5782), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59422] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8447), 1,
      sym_attribute_value,
    STATE(5783), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59436] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8449), 1,
      anon_sym_RBRACE,
    STATE(5784), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59450] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8451), 1,
      anon_sym_RPAREN,
    STATE(5785), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59464] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3151), 1,
      ts_builtin_sym_end,
    STATE(5786), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59478] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8453), 1,
      anon_sym_RPAREN,
    STATE(5787), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59492] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7260), 1,
      anon_sym_GT_GT,
    STATE(5788), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59506] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8455), 1,
      sym__gobbled_content,
    STATE(5789), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59520] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8457), 1,
      anon_sym_RBRACE,
    STATE(5790), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59534] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8459), 1,
      anon_sym_RBRACE,
    STATE(5791), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59548] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8461), 1,
      anon_sym_RBRACE,
    STATE(5792), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59562] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2755), 1,
      sym__brace_end_zw,
    STATE(5793), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59576] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8463), 1,
      sym__NONASSOC,
    STATE(5794), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59590] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8465), 1,
      anon_sym_RPAREN,
    STATE(5795), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59604] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8467), 1,
      anon_sym_RBRACE,
    STATE(5796), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59618] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5780), 1,
      anon_sym_RBRACE,
    STATE(5797), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59632] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8469), 1,
      sym_attribute_value,
    STATE(5798), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59646] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8059), 1,
      anon_sym_COMMA,
    STATE(5799), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59660] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8471), 1,
      anon_sym_RBRACE,
    STATE(5800), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59674] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8473), 1,
      anon_sym_RBRACE,
    STATE(5801), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59688] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6174), 1,
      sym__dollar_ident_zw,
    STATE(5802), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59702] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8475), 1,
      anon_sym_RBRACE,
    STATE(5803), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59716] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8477), 1,
      anon_sym_RBRACE,
    STATE(5804), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59730] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8479), 1,
      anon_sym_RBRACE,
    STATE(5805), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59744] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8481), 1,
      anon_sym_STAR,
    STATE(5806), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59758] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8483), 1,
      anon_sym_RPAREN,
    STATE(5807), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59772] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5764), 1,
      anon_sym_RBRACE,
    STATE(5808), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59786] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8485), 1,
      anon_sym_GT,
    STATE(5809), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59800] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8487), 1,
      anon_sym_RBRACE,
    STATE(5810), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59814] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8489), 1,
      sym__NONASSOC,
    STATE(5811), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59828] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8491), 1,
      anon_sym_RPAREN,
    STATE(5812), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59842] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8493), 1,
      anon_sym_RPAREN,
    STATE(5813), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59856] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8495), 1,
      anon_sym_RBRACE,
    STATE(5814), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59870] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8497), 1,
      anon_sym_RPAREN,
    STATE(5815), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59884] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7316), 1,
      anon_sym_GT_GT,
    STATE(5816), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59898] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7202), 1,
      anon_sym_GT_GT,
    STATE(5817), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59912] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5769), 1,
      anon_sym_RBRACE,
    STATE(5818), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59926] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8499), 1,
      anon_sym_RPAREN,
    STATE(5819), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59940] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8501), 1,
      sym__ERROR,
    STATE(5820), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59954] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8503), 1,
      anon_sym_RBRACE,
    STATE(5821), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59968] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8505), 1,
      anon_sym_RPAREN,
    STATE(5822), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59982] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2775), 1,
      sym__brace_end_zw,
    STATE(5823), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [59996] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8507), 1,
      aux_sym___DATA___token1,
    STATE(5824), 1,
      sym_heredoc_content,
  [60012] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8509), 1,
      anon_sym_RBRACE,
    STATE(5825), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60026] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8511), 1,
      anon_sym_RPAREN,
    STATE(5826), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60040] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8513), 1,
      anon_sym_RBRACE,
    STATE(5827), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60054] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8515), 1,
      anon_sym_RPAREN,
    STATE(5828), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60068] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8517), 1,
      anon_sym_RBRACE,
    STATE(5829), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60082] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8519), 1,
      anon_sym_RBRACE,
    STATE(5830), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60096] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8521), 1,
      sym_attribute_value,
    STATE(5831), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60110] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8523), 1,
      sym__NONASSOC,
    STATE(5832), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60124] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2757), 1,
      sym__brace_end_zw,
    STATE(5833), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60138] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8525), 1,
      anon_sym_RBRACE,
    STATE(5834), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60152] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8527), 1,
      sym__gobbled_content,
    STATE(5835), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60166] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8529), 1,
      sym__ERROR,
    STATE(5836), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60180] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8531), 1,
      anon_sym_RBRACE,
    STATE(5837), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60194] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(3203), 1,
      ts_builtin_sym_end,
    STATE(5838), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60208] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5771), 1,
      anon_sym_RBRACE,
    STATE(5839), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60222] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7324), 1,
      anon_sym_GT_GT,
    STATE(5840), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60236] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8533), 1,
      sym_attribute_value,
    STATE(5841), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60250] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6286), 1,
      sym__dollar_ident_zw,
    STATE(5842), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60264] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8535), 1,
      ts_builtin_sym_end,
    STATE(5843), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60278] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2777), 1,
      sym__brace_end_zw,
    STATE(5844), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60292] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8537), 1,
      anon_sym_RBRACE,
    STATE(5845), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60306] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(818), 1,
      ts_builtin_sym_end,
    STATE(5846), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60320] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8539), 1,
      sym_attribute_value,
    STATE(5847), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60334] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7216), 1,
      anon_sym_GT_GT,
    STATE(5848), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60348] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2779), 1,
      sym__brace_end_zw,
    STATE(5849), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60362] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8541), 1,
      sym__gobbled_content,
    STATE(5850), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60376] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8543), 1,
      anon_sym_RPAREN,
    STATE(5851), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60390] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8545), 1,
      sym_attribute_value,
    STATE(5852), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60404] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8547), 1,
      anon_sym_LBRACE,
    STATE(5853), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60418] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2781), 1,
      sym__brace_end_zw,
    STATE(5854), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60432] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8549), 1,
      anon_sym_RBRACE,
    STATE(5855), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60446] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8551), 1,
      anon_sym_RPAREN,
    STATE(5856), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60460] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8553), 1,
      sym_attribute_value,
    STATE(5857), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60474] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8555), 1,
      anon_sym_GT,
    STATE(5858), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60488] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2783), 1,
      sym__brace_end_zw,
    STATE(5859), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60502] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8557), 1,
      sym__ERROR,
    STATE(5860), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60516] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8559), 1,
      anon_sym_LPAREN,
    STATE(5861), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60530] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8561), 1,
      sym_attribute_value,
    STATE(5862), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60544] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(5778), 1,
      anon_sym_RBRACE,
    STATE(5863), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60558] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2785), 1,
      sym__brace_end_zw,
    STATE(5864), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60572] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8563), 1,
      anon_sym_RBRACE,
    STATE(5865), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60586] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8565), 1,
      sym__ERROR,
    STATE(5866), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60600] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8567), 1,
      anon_sym_RPAREN,
    STATE(5867), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60614] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2787), 1,
      sym__brace_end_zw,
    STATE(5868), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60628] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8569), 1,
      sym__ERROR,
    STATE(5869), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60642] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8571), 1,
      anon_sym_RPAREN,
    STATE(5870), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60656] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2789), 1,
      sym__brace_end_zw,
    STATE(5871), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60670] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8573), 1,
      anon_sym_RBRACE,
    STATE(5872), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60684] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2791), 1,
      sym__brace_end_zw,
    STATE(5873), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60698] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8575), 1,
      aux_sym___DATA___token1,
    STATE(5874), 1,
      sym_heredoc_content,
  [60714] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8577), 1,
      anon_sym_RBRACE,
    STATE(5875), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60728] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2793), 1,
      sym__brace_end_zw,
    STATE(5876), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60742] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8579), 1,
      anon_sym_RBRACE,
    STATE(5877), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60756] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8581), 1,
      anon_sym_RPAREN,
    STATE(5878), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60770] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2795), 1,
      sym__brace_end_zw,
    STATE(5879), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60784] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8583), 1,
      sym__gobbled_content,
    STATE(5880), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60798] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8585), 1,
      anon_sym_STAR,
    STATE(5881), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60812] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2797), 1,
      sym__brace_end_zw,
    STATE(5882), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60826] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8587), 1,
      anon_sym_GT,
    STATE(5883), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60840] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8589), 1,
      anon_sym_RBRACE,
    STATE(5884), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60854] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2799), 1,
      sym__brace_end_zw,
    STATE(5885), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60868] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6390), 1,
      sym__dollar_ident_zw,
    STATE(5886), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60882] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8591), 1,
      anon_sym_RBRACE,
    STATE(5887), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60896] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2801), 1,
      sym__brace_end_zw,
    STATE(5888), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60910] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8593), 1,
      sym__ERROR,
    STATE(5889), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60924] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8595), 1,
      anon_sym_RBRACE,
    STATE(5890), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60938] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2803), 1,
      sym__brace_end_zw,
    STATE(5891), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60952] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8597), 1,
      anon_sym_RBRACE,
    STATE(5892), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60966] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8599), 1,
      anon_sym_RPAREN,
    STATE(5893), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60980] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2805), 1,
      sym__brace_end_zw,
    STATE(5894), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [60994] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8601), 1,
      anon_sym_LPAREN,
    STATE(5895), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61008] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8603), 1,
      anon_sym_RPAREN,
    STATE(5896), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61022] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2807), 1,
      sym__brace_end_zw,
    STATE(5897), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61036] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8605), 1,
      sym_attribute_value,
    STATE(5898), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61050] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8607), 1,
      anon_sym_STAR,
    STATE(5899), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61064] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2809), 1,
      sym__brace_end_zw,
    STATE(5900), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61078] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8609), 1,
      anon_sym_RPAREN,
    STATE(5901), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61092] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8611), 1,
      anon_sym_RPAREN,
    STATE(5902), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61106] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2811), 1,
      sym__brace_end_zw,
    STATE(5903), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61120] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8613), 1,
      sym__gobbled_content,
    STATE(5904), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61134] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8615), 1,
      anon_sym_RBRACE,
    STATE(5905), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61148] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2813), 1,
      sym__brace_end_zw,
    STATE(5906), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61162] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8617), 1,
      anon_sym_RPAREN,
    STATE(5907), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61176] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8619), 1,
      anon_sym_RBRACE,
    STATE(5908), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61190] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2815), 1,
      sym__brace_end_zw,
    STATE(5909), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61204] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8621), 1,
      anon_sym_COLON,
    STATE(5910), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61218] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8623), 1,
      anon_sym_RBRACE,
    STATE(5911), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61232] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2817), 1,
      sym__brace_end_zw,
    STATE(5912), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61246] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8625), 1,
      anon_sym_RBRACE,
    STATE(5913), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61260] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8627), 1,
      anon_sym_RBRACE,
    STATE(5914), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61274] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2819), 1,
      sym__brace_end_zw,
    STATE(5915), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61288] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8629), 1,
      anon_sym_RPAREN,
    STATE(5916), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61302] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8631), 1,
      anon_sym_RBRACE,
    STATE(5917), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61316] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2821), 1,
      sym__brace_end_zw,
    STATE(5918), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61330] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8633), 1,
      anon_sym_RBRACE,
    STATE(5919), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61344] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8635), 1,
      anon_sym_LPAREN,
    STATE(5920), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61358] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2823), 1,
      sym__brace_end_zw,
    STATE(5921), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61372] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8637), 1,
      anon_sym_RPAREN,
    STATE(5922), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61386] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8639), 1,
      anon_sym_LPAREN,
    STATE(5923), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61400] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2825), 1,
      sym__brace_end_zw,
    STATE(5924), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61414] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8641), 1,
      anon_sym_STAR,
    STATE(5925), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61428] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8643), 1,
      anon_sym_RPAREN,
    STATE(5926), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61442] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2827), 1,
      sym__brace_end_zw,
    STATE(5927), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61456] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8645), 1,
      anon_sym_LBRACE,
    STATE(5928), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61470] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8647), 1,
      anon_sym_STAR,
    STATE(5929), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61484] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2829), 1,
      sym__brace_end_zw,
    STATE(5930), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61498] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7230), 1,
      anon_sym_GT_GT,
    STATE(5931), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61512] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8649), 1,
      anon_sym_LPAREN,
    STATE(5932), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61526] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2831), 1,
      sym__brace_end_zw,
    STATE(5933), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61540] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8651), 1,
      anon_sym_RBRACE,
    STATE(5934), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61554] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8111), 1,
      anon_sym_COMMA,
    STATE(5935), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61568] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2741), 1,
      sym__brace_end_zw,
    STATE(5936), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61582] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8653), 1,
      anon_sym_LPAREN,
    STATE(5937), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61596] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8655), 1,
      anon_sym_RPAREN,
    STATE(5938), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61610] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8657), 1,
      anon_sym_RPAREN,
    STATE(5939), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61624] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8659), 1,
      anon_sym_RPAREN,
    STATE(5940), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61638] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8661), 1,
      sym__ERROR,
    STATE(5941), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61652] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8663), 1,
      sym__NONASSOC,
    STATE(5942), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61666] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8665), 1,
      anon_sym_RPAREN,
    STATE(5943), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61680] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8667), 1,
      anon_sym_STAR,
    STATE(5944), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61694] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8669), 1,
      anon_sym_RBRACE,
    STATE(5945), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61708] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8671), 1,
      sym__ERROR,
    STATE(5946), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61722] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8673), 1,
      sym__ERROR,
    STATE(5947), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61736] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8675), 1,
      anon_sym_RPAREN,
    STATE(5948), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61750] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8677), 1,
      anon_sym_GT,
    STATE(5949), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61764] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8679), 1,
      sym__ERROR,
    STATE(5950), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61778] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8681), 1,
      anon_sym_RBRACE,
    STATE(5951), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61792] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6370), 1,
      sym__no_interp_whitespace_zw,
    STATE(5952), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61806] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8683), 1,
      anon_sym_RBRACE,
    STATE(5953), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61820] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8685), 1,
      sym__ERROR,
    STATE(5954), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61834] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8687), 1,
      anon_sym_LPAREN,
    STATE(5955), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61848] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2715), 1,
      sym__brace_end_zw,
    STATE(5956), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61862] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8689), 1,
      sym_attribute_value,
    STATE(5957), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61876] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8691), 1,
      anon_sym_RBRACE,
    STATE(5958), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61890] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8693), 1,
      sym__ERROR,
    STATE(5959), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61904] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8695), 1,
      sym__ERROR,
    STATE(5960), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61918] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8697), 1,
      anon_sym_RBRACE,
    STATE(5961), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61932] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8699), 1,
      sym__ERROR,
    STATE(5962), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61946] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8701), 1,
      anon_sym_RBRACE,
    STATE(5963), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61960] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8703), 1,
      anon_sym_RBRACE,
    STATE(5964), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61974] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8705), 1,
      sym__gobbled_content,
    STATE(5965), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [61988] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8707), 1,
      anon_sym_RPAREN,
    STATE(5966), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62002] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8709), 1,
      anon_sym_LPAREN,
    STATE(5967), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62016] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8711), 1,
      anon_sym_RBRACE,
    STATE(5968), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62030] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8713), 1,
      anon_sym_RBRACE,
    STATE(5969), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62044] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8715), 1,
      anon_sym_RPAREN,
    STATE(5970), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62058] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8717), 1,
      anon_sym_RBRACE,
    STATE(5971), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62072] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2761), 1,
      sym__brace_end_zw,
    STATE(5972), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62086] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8719), 1,
      anon_sym_RPAREN,
    STATE(5973), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62100] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8721), 1,
      sym__ERROR,
    STATE(5974), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62114] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8723), 1,
      anon_sym_RPAREN,
    STATE(5975), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62128] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8725), 1,
      anon_sym_RPAREN,
    STATE(5976), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62142] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8727), 1,
      anon_sym_RBRACE,
    STATE(5977), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62156] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2753), 1,
      sym__brace_end_zw,
    STATE(5978), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62170] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6388), 1,
      sym__dollar_ident_zw,
    STATE(5979), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62184] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2719), 1,
      sym__brace_end_zw,
    STATE(5980), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62198] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8729), 1,
      anon_sym_RBRACE,
    STATE(5981), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62212] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(7220), 1,
      anon_sym_GT_GT,
    STATE(5982), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62226] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8731), 1,
      anon_sym_RBRACE,
    STATE(5983), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62240] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8733), 1,
      anon_sym_LPAREN,
    STATE(5984), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62254] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8735), 1,
      sym__ERROR,
    STATE(5985), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62268] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8737), 1,
      sym__ERROR,
    STATE(5986), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62282] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8739), 1,
      anon_sym_RPAREN,
    STATE(5987), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62296] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8741), 1,
      anon_sym_RBRACE,
    STATE(5988), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62310] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8743), 1,
      anon_sym_RBRACE,
    STATE(5989), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62324] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8745), 1,
      sym__ERROR,
    STATE(5990), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62338] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8747), 1,
      anon_sym_RBRACE,
    STATE(5991), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62352] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6164), 1,
      sym__dollar_ident_zw,
    STATE(5992), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62366] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8749), 1,
      anon_sym_GT,
    STATE(5993), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62380] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8751), 1,
      anon_sym_LPAREN,
    STATE(5994), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62394] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8753), 1,
      anon_sym_RPAREN,
    STATE(5995), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62408] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_pod,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8755), 1,
      aux_sym___DATA___token1,
    STATE(5996), 1,
      sym_heredoc_content,
  [62424] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8757), 1,
      anon_sym_RPAREN,
    STATE(5997), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62438] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8759), 1,
      anon_sym_RPAREN,
    STATE(5998), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62452] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8761), 1,
      anon_sym_RBRACE,
    STATE(5999), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62466] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8763), 1,
      sym__NONASSOC,
    STATE(6000), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62480] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8765), 1,
      anon_sym_LPAREN,
    STATE(6001), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62494] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8767), 1,
      anon_sym_RPAREN,
    STATE(6002), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62508] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8769), 1,
      sym_attribute_value,
    STATE(6003), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62522] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8771), 1,
      anon_sym_RPAREN,
    STATE(6004), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62536] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6270), 1,
      sym__no_interp_whitespace_zw,
    STATE(6005), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62550] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6358), 1,
      sym__no_interp_whitespace_zw,
    STATE(6006), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62564] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8773), 1,
      sym__gobbled_content,
    STATE(6007), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62578] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8775), 1,
      anon_sym_LPAREN,
    STATE(6008), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62592] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(2739), 1,
      anon_sym_GT,
    STATE(6009), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62606] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8777), 1,
      anon_sym_RBRACE,
    STATE(6010), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62620] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8779), 1,
      anon_sym_RBRACE,
    STATE(6011), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62634] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8781), 1,
      anon_sym_RBRACE,
    STATE(6012), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62648] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8783), 1,
      anon_sym_RBRACE,
    STATE(6013), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62662] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8785), 1,
      anon_sym_RPAREN,
    STATE(6014), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62676] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8787), 1,
      anon_sym_LPAREN,
    STATE(6015), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62690] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8789), 1,
      anon_sym_RBRACE,
    STATE(6016), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62704] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8791), 1,
      anon_sym_RPAREN,
    STATE(6017), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62718] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8793), 1,
      anon_sym_RPAREN,
    STATE(6018), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62732] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8795), 1,
      anon_sym_RPAREN,
    STATE(6019), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62746] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8797), 1,
      anon_sym_RBRACE,
    STATE(6020), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62760] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8799), 1,
      ts_builtin_sym_end,
    STATE(6021), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62774] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8801), 1,
      anon_sym_LPAREN,
    STATE(6022), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62788] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8803), 1,
      anon_sym_RPAREN,
    STATE(6023), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62802] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8805), 1,
      anon_sym_LPAREN,
    STATE(6024), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62816] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8807), 1,
      anon_sym_LPAREN,
    STATE(6025), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62830] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8809), 1,
      anon_sym_LPAREN,
    STATE(6026), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62844] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8811), 1,
      anon_sym_LPAREN,
    STATE(6027), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62858] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8813), 1,
      anon_sym_RPAREN,
    STATE(6028), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62872] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8815), 1,
      anon_sym_LPAREN,
    STATE(6029), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62886] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8817), 1,
      anon_sym_LPAREN,
    STATE(6030), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62900] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8819), 1,
      anon_sym_RBRACE,
    STATE(6031), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62914] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8821), 1,
      anon_sym_LPAREN,
    STATE(6032), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62928] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8823), 1,
      anon_sym_RPAREN,
    STATE(6033), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62942] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8825), 1,
      anon_sym_LPAREN,
    STATE(6034), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62956] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8827), 1,
      anon_sym_RBRACE,
    STATE(6035), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62970] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8829), 1,
      anon_sym_LPAREN,
    STATE(6036), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62984] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8831), 1,
      anon_sym_LPAREN,
    STATE(6037), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [62998] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8833), 1,
      sym__NONASSOC,
    STATE(6038), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63012] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8835), 1,
      anon_sym_LPAREN,
    STATE(6039), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63026] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8837), 1,
      anon_sym_LPAREN,
    STATE(6040), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63040] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8839), 1,
      sym__ERROR,
    STATE(6041), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63054] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8841), 1,
      sym__ERROR,
    STATE(6042), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63068] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8843), 1,
      anon_sym_LPAREN,
    STATE(6043), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63082] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8845), 1,
      anon_sym_COLON,
    STATE(6044), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63096] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(6360), 1,
      sym__dollar_ident_zw,
    STATE(6045), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63110] = 4,
    ACTIONS(7), 1,
      sym__heredoc_start,
    ACTIONS(8847), 1,
      anon_sym_RBRACE,
    STATE(6046), 1,
      sym_heredoc_content,
    ACTIONS(5), 2,
      sym_pod,
      sym_comment,
  [63124] = 1,
    ACTIONS(8849), 1,
      ts_builtin_sym_end,
  [63128] = 1,
    ACTIONS(8851), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(3811)] = 0,
  [SMALL_STATE(3812)] = 73,
  [SMALL_STATE(3813)] = 146,
  [SMALL_STATE(3814)] = 219,
  [SMALL_STATE(3815)] = 292,
  [SMALL_STATE(3816)] = 365,
  [SMALL_STATE(3817)] = 438,
  [SMALL_STATE(3818)] = 511,
  [SMALL_STATE(3819)] = 584,
  [SMALL_STATE(3820)] = 657,
  [SMALL_STATE(3821)] = 730,
  [SMALL_STATE(3822)] = 803,
  [SMALL_STATE(3823)] = 876,
  [SMALL_STATE(3824)] = 949,
  [SMALL_STATE(3825)] = 1022,
  [SMALL_STATE(3826)] = 1095,
  [SMALL_STATE(3827)] = 1168,
  [SMALL_STATE(3828)] = 1238,
  [SMALL_STATE(3829)] = 1308,
  [SMALL_STATE(3830)] = 1378,
  [SMALL_STATE(3831)] = 1448,
  [SMALL_STATE(3832)] = 1518,
  [SMALL_STATE(3833)] = 1588,
  [SMALL_STATE(3834)] = 1658,
  [SMALL_STATE(3835)] = 1728,
  [SMALL_STATE(3836)] = 1798,
  [SMALL_STATE(3837)] = 1868,
  [SMALL_STATE(3838)] = 1938,
  [SMALL_STATE(3839)] = 2008,
  [SMALL_STATE(3840)] = 2078,
  [SMALL_STATE(3841)] = 2148,
  [SMALL_STATE(3842)] = 2218,
  [SMALL_STATE(3843)] = 2288,
  [SMALL_STATE(3844)] = 2358,
  [SMALL_STATE(3845)] = 2428,
  [SMALL_STATE(3846)] = 2498,
  [SMALL_STATE(3847)] = 2568,
  [SMALL_STATE(3848)] = 2638,
  [SMALL_STATE(3849)] = 2708,
  [SMALL_STATE(3850)] = 2778,
  [SMALL_STATE(3851)] = 2848,
  [SMALL_STATE(3852)] = 2918,
  [SMALL_STATE(3853)] = 2988,
  [SMALL_STATE(3854)] = 3058,
  [SMALL_STATE(3855)] = 3128,
  [SMALL_STATE(3856)] = 3198,
  [SMALL_STATE(3857)] = 3268,
  [SMALL_STATE(3858)] = 3338,
  [SMALL_STATE(3859)] = 3408,
  [SMALL_STATE(3860)] = 3478,
  [SMALL_STATE(3861)] = 3548,
  [SMALL_STATE(3862)] = 3618,
  [SMALL_STATE(3863)] = 3688,
  [SMALL_STATE(3864)] = 3758,
  [SMALL_STATE(3865)] = 3828,
  [SMALL_STATE(3866)] = 3898,
  [SMALL_STATE(3867)] = 3968,
  [SMALL_STATE(3868)] = 4038,
  [SMALL_STATE(3869)] = 4108,
  [SMALL_STATE(3870)] = 4178,
  [SMALL_STATE(3871)] = 4248,
  [SMALL_STATE(3872)] = 4318,
  [SMALL_STATE(3873)] = 4388,
  [SMALL_STATE(3874)] = 4458,
  [SMALL_STATE(3875)] = 4528,
  [SMALL_STATE(3876)] = 4597,
  [SMALL_STATE(3877)] = 4666,
  [SMALL_STATE(3878)] = 4731,
  [SMALL_STATE(3879)] = 4800,
  [SMALL_STATE(3880)] = 4869,
  [SMALL_STATE(3881)] = 4938,
  [SMALL_STATE(3882)] = 5007,
  [SMALL_STATE(3883)] = 5076,
  [SMALL_STATE(3884)] = 5145,
  [SMALL_STATE(3885)] = 5214,
  [SMALL_STATE(3886)] = 5283,
  [SMALL_STATE(3887)] = 5352,
  [SMALL_STATE(3888)] = 5421,
  [SMALL_STATE(3889)] = 5490,
  [SMALL_STATE(3890)] = 5559,
  [SMALL_STATE(3891)] = 5628,
  [SMALL_STATE(3892)] = 5697,
  [SMALL_STATE(3893)] = 5766,
  [SMALL_STATE(3894)] = 5835,
  [SMALL_STATE(3895)] = 5900,
  [SMALL_STATE(3896)] = 5969,
  [SMALL_STATE(3897)] = 6038,
  [SMALL_STATE(3898)] = 6107,
  [SMALL_STATE(3899)] = 6176,
  [SMALL_STATE(3900)] = 6245,
  [SMALL_STATE(3901)] = 6314,
  [SMALL_STATE(3902)] = 6383,
  [SMALL_STATE(3903)] = 6452,
  [SMALL_STATE(3904)] = 6521,
  [SMALL_STATE(3905)] = 6590,
  [SMALL_STATE(3906)] = 6659,
  [SMALL_STATE(3907)] = 6728,
  [SMALL_STATE(3908)] = 6797,
  [SMALL_STATE(3909)] = 6866,
  [SMALL_STATE(3910)] = 6935,
  [SMALL_STATE(3911)] = 7004,
  [SMALL_STATE(3912)] = 7073,
  [SMALL_STATE(3913)] = 7142,
  [SMALL_STATE(3914)] = 7211,
  [SMALL_STATE(3915)] = 7280,
  [SMALL_STATE(3916)] = 7349,
  [SMALL_STATE(3917)] = 7418,
  [SMALL_STATE(3918)] = 7487,
  [SMALL_STATE(3919)] = 7556,
  [SMALL_STATE(3920)] = 7625,
  [SMALL_STATE(3921)] = 7694,
  [SMALL_STATE(3922)] = 7763,
  [SMALL_STATE(3923)] = 7832,
  [SMALL_STATE(3924)] = 7901,
  [SMALL_STATE(3925)] = 7970,
  [SMALL_STATE(3926)] = 8039,
  [SMALL_STATE(3927)] = 8108,
  [SMALL_STATE(3928)] = 8177,
  [SMALL_STATE(3929)] = 8246,
  [SMALL_STATE(3930)] = 8315,
  [SMALL_STATE(3931)] = 8384,
  [SMALL_STATE(3932)] = 8453,
  [SMALL_STATE(3933)] = 8522,
  [SMALL_STATE(3934)] = 8591,
  [SMALL_STATE(3935)] = 8660,
  [SMALL_STATE(3936)] = 8729,
  [SMALL_STATE(3937)] = 8798,
  [SMALL_STATE(3938)] = 8867,
  [SMALL_STATE(3939)] = 8936,
  [SMALL_STATE(3940)] = 9005,
  [SMALL_STATE(3941)] = 9074,
  [SMALL_STATE(3942)] = 9143,
  [SMALL_STATE(3943)] = 9212,
  [SMALL_STATE(3944)] = 9281,
  [SMALL_STATE(3945)] = 9350,
  [SMALL_STATE(3946)] = 9419,
  [SMALL_STATE(3947)] = 9488,
  [SMALL_STATE(3948)] = 9557,
  [SMALL_STATE(3949)] = 9626,
  [SMALL_STATE(3950)] = 9695,
  [SMALL_STATE(3951)] = 9764,
  [SMALL_STATE(3952)] = 9833,
  [SMALL_STATE(3953)] = 9902,
  [SMALL_STATE(3954)] = 9971,
  [SMALL_STATE(3955)] = 10040,
  [SMALL_STATE(3956)] = 10109,
  [SMALL_STATE(3957)] = 10178,
  [SMALL_STATE(3958)] = 10247,
  [SMALL_STATE(3959)] = 10316,
  [SMALL_STATE(3960)] = 10385,
  [SMALL_STATE(3961)] = 10454,
  [SMALL_STATE(3962)] = 10523,
  [SMALL_STATE(3963)] = 10592,
  [SMALL_STATE(3964)] = 10661,
  [SMALL_STATE(3965)] = 10730,
  [SMALL_STATE(3966)] = 10799,
  [SMALL_STATE(3967)] = 10868,
  [SMALL_STATE(3968)] = 10937,
  [SMALL_STATE(3969)] = 11006,
  [SMALL_STATE(3970)] = 11075,
  [SMALL_STATE(3971)] = 11144,
  [SMALL_STATE(3972)] = 11213,
  [SMALL_STATE(3973)] = 11282,
  [SMALL_STATE(3974)] = 11351,
  [SMALL_STATE(3975)] = 11420,
  [SMALL_STATE(3976)] = 11489,
  [SMALL_STATE(3977)] = 11558,
  [SMALL_STATE(3978)] = 11627,
  [SMALL_STATE(3979)] = 11696,
  [SMALL_STATE(3980)] = 11765,
  [SMALL_STATE(3981)] = 11834,
  [SMALL_STATE(3982)] = 11898,
  [SMALL_STATE(3983)] = 11963,
  [SMALL_STATE(3984)] = 12026,
  [SMALL_STATE(3985)] = 12089,
  [SMALL_STATE(3986)] = 12152,
  [SMALL_STATE(3987)] = 12215,
  [SMALL_STATE(3988)] = 12278,
  [SMALL_STATE(3989)] = 12341,
  [SMALL_STATE(3990)] = 12404,
  [SMALL_STATE(3991)] = 12467,
  [SMALL_STATE(3992)] = 12532,
  [SMALL_STATE(3993)] = 12595,
  [SMALL_STATE(3994)] = 12632,
  [SMALL_STATE(3995)] = 12671,
  [SMALL_STATE(3996)] = 12720,
  [SMALL_STATE(3997)] = 12769,
  [SMALL_STATE(3998)] = 12820,
  [SMALL_STATE(3999)] = 12869,
  [SMALL_STATE(4000)] = 12920,
  [SMALL_STATE(4001)] = 12969,
  [SMALL_STATE(4002)] = 13018,
  [SMALL_STATE(4003)] = 13067,
  [SMALL_STATE(4004)] = 13116,
  [SMALL_STATE(4005)] = 13167,
  [SMALL_STATE(4006)] = 13218,
  [SMALL_STATE(4007)] = 13267,
  [SMALL_STATE(4008)] = 13316,
  [SMALL_STATE(4009)] = 13365,
  [SMALL_STATE(4010)] = 13416,
  [SMALL_STATE(4011)] = 13465,
  [SMALL_STATE(4012)] = 13516,
  [SMALL_STATE(4013)] = 13567,
  [SMALL_STATE(4014)] = 13618,
  [SMALL_STATE(4015)] = 13676,
  [SMALL_STATE(4016)] = 13734,
  [SMALL_STATE(4017)] = 13792,
  [SMALL_STATE(4018)] = 13850,
  [SMALL_STATE(4019)] = 13896,
  [SMALL_STATE(4020)] = 13930,
  [SMALL_STATE(4021)] = 13962,
  [SMALL_STATE(4022)] = 13994,
  [SMALL_STATE(4023)] = 14041,
  [SMALL_STATE(4024)] = 14092,
  [SMALL_STATE(4025)] = 14129,
  [SMALL_STATE(4026)] = 14180,
  [SMALL_STATE(4027)] = 14231,
  [SMALL_STATE(4028)] = 14262,
  [SMALL_STATE(4029)] = 14313,
  [SMALL_STATE(4030)] = 14350,
  [SMALL_STATE(4031)] = 14397,
  [SMALL_STATE(4032)] = 14426,
  [SMALL_STATE(4033)] = 14477,
  [SMALL_STATE(4034)] = 14506,
  [SMALL_STATE(4035)] = 14535,
  [SMALL_STATE(4036)] = 14566,
  [SMALL_STATE(4037)] = 14611,
  [SMALL_STATE(4038)] = 14640,
  [SMALL_STATE(4039)] = 14669,
  [SMALL_STATE(4040)] = 14698,
  [SMALL_STATE(4041)] = 14749,
  [SMALL_STATE(4042)] = 14800,
  [SMALL_STATE(4043)] = 14851,
  [SMALL_STATE(4044)] = 14882,
  [SMALL_STATE(4045)] = 14929,
  [SMALL_STATE(4046)] = 14960,
  [SMALL_STATE(4047)] = 15007,
  [SMALL_STATE(4048)] = 15054,
  [SMALL_STATE(4049)] = 15098,
  [SMALL_STATE(4050)] = 15142,
  [SMALL_STATE(4051)] = 15170,
  [SMALL_STATE(4052)] = 15200,
  [SMALL_STATE(4053)] = 15244,
  [SMALL_STATE(4054)] = 15288,
  [SMALL_STATE(4055)] = 15318,
  [SMALL_STATE(4056)] = 15358,
  [SMALL_STATE(4057)] = 15402,
  [SMALL_STATE(4058)] = 15430,
  [SMALL_STATE(4059)] = 15474,
  [SMALL_STATE(4060)] = 15514,
  [SMALL_STATE(4061)] = 15558,
  [SMALL_STATE(4062)] = 15598,
  [SMALL_STATE(4063)] = 15642,
  [SMALL_STATE(4064)] = 15686,
  [SMALL_STATE(4065)] = 15726,
  [SMALL_STATE(4066)] = 15770,
  [SMALL_STATE(4067)] = 15814,
  [SMALL_STATE(4068)] = 15858,
  [SMALL_STATE(4069)] = 15886,
  [SMALL_STATE(4070)] = 15930,
  [SMALL_STATE(4071)] = 15958,
  [SMALL_STATE(4072)] = 16002,
  [SMALL_STATE(4073)] = 16046,
  [SMALL_STATE(4074)] = 16090,
  [SMALL_STATE(4075)] = 16134,
  [SMALL_STATE(4076)] = 16174,
  [SMALL_STATE(4077)] = 16218,
  [SMALL_STATE(4078)] = 16262,
  [SMALL_STATE(4079)] = 16298,
  [SMALL_STATE(4080)] = 16342,
  [SMALL_STATE(4081)] = 16386,
  [SMALL_STATE(4082)] = 16414,
  [SMALL_STATE(4083)] = 16458,
  [SMALL_STATE(4084)] = 16486,
  [SMALL_STATE(4085)] = 16530,
  [SMALL_STATE(4086)] = 16558,
  [SMALL_STATE(4087)] = 16586,
  [SMALL_STATE(4088)] = 16616,
  [SMALL_STATE(4089)] = 16644,
  [SMALL_STATE(4090)] = 16692,
  [SMALL_STATE(4091)] = 16720,
  [SMALL_STATE(4092)] = 16764,
  [SMALL_STATE(4093)] = 16812,
  [SMALL_STATE(4094)] = 16856,
  [SMALL_STATE(4095)] = 16900,
  [SMALL_STATE(4096)] = 16928,
  [SMALL_STATE(4097)] = 16968,
  [SMALL_STATE(4098)] = 17012,
  [SMALL_STATE(4099)] = 17056,
  [SMALL_STATE(4100)] = 17100,
  [SMALL_STATE(4101)] = 17128,
  [SMALL_STATE(4102)] = 17168,
  [SMALL_STATE(4103)] = 17196,
  [SMALL_STATE(4104)] = 17240,
  [SMALL_STATE(4105)] = 17268,
  [SMALL_STATE(4106)] = 17308,
  [SMALL_STATE(4107)] = 17352,
  [SMALL_STATE(4108)] = 17382,
  [SMALL_STATE(4109)] = 17410,
  [SMALL_STATE(4110)] = 17440,
  [SMALL_STATE(4111)] = 17480,
  [SMALL_STATE(4112)] = 17524,
  [SMALL_STATE(4113)] = 17564,
  [SMALL_STATE(4114)] = 17608,
  [SMALL_STATE(4115)] = 17636,
  [SMALL_STATE(4116)] = 17676,
  [SMALL_STATE(4117)] = 17704,
  [SMALL_STATE(4118)] = 17744,
  [SMALL_STATE(4119)] = 17784,
  [SMALL_STATE(4120)] = 17828,
  [SMALL_STATE(4121)] = 17868,
  [SMALL_STATE(4122)] = 17912,
  [SMALL_STATE(4123)] = 17956,
  [SMALL_STATE(4124)] = 17986,
  [SMALL_STATE(4125)] = 18030,
  [SMALL_STATE(4126)] = 18074,
  [SMALL_STATE(4127)] = 18118,
  [SMALL_STATE(4128)] = 18162,
  [SMALL_STATE(4129)] = 18202,
  [SMALL_STATE(4130)] = 18246,
  [SMALL_STATE(4131)] = 18290,
  [SMALL_STATE(4132)] = 18334,
  [SMALL_STATE(4133)] = 18378,
  [SMALL_STATE(4134)] = 18422,
  [SMALL_STATE(4135)] = 18462,
  [SMALL_STATE(4136)] = 18490,
  [SMALL_STATE(4137)] = 18534,
  [SMALL_STATE(4138)] = 18578,
  [SMALL_STATE(4139)] = 18606,
  [SMALL_STATE(4140)] = 18647,
  [SMALL_STATE(4141)] = 18688,
  [SMALL_STATE(4142)] = 18729,
  [SMALL_STATE(4143)] = 18756,
  [SMALL_STATE(4144)] = 18797,
  [SMALL_STATE(4145)] = 18838,
  [SMALL_STATE(4146)] = 18879,
  [SMALL_STATE(4147)] = 18920,
  [SMALL_STATE(4148)] = 18961,
  [SMALL_STATE(4149)] = 19002,
  [SMALL_STATE(4150)] = 19031,
  [SMALL_STATE(4151)] = 19068,
  [SMALL_STATE(4152)] = 19103,
  [SMALL_STATE(4153)] = 19144,
  [SMALL_STATE(4154)] = 19175,
  [SMALL_STATE(4155)] = 19216,
  [SMALL_STATE(4156)] = 19253,
  [SMALL_STATE(4157)] = 19280,
  [SMALL_STATE(4158)] = 19321,
  [SMALL_STATE(4159)] = 19350,
  [SMALL_STATE(4160)] = 19391,
  [SMALL_STATE(4161)] = 19432,
  [SMALL_STATE(4162)] = 19473,
  [SMALL_STATE(4163)] = 19504,
  [SMALL_STATE(4164)] = 19541,
  [SMALL_STATE(4165)] = 19578,
  [SMALL_STATE(4166)] = 19605,
  [SMALL_STATE(4167)] = 19632,
  [SMALL_STATE(4168)] = 19669,
  [SMALL_STATE(4169)] = 19698,
  [SMALL_STATE(4170)] = 19725,
  [SMALL_STATE(4171)] = 19752,
  [SMALL_STATE(4172)] = 19793,
  [SMALL_STATE(4173)] = 19820,
  [SMALL_STATE(4174)] = 19857,
  [SMALL_STATE(4175)] = 19898,
  [SMALL_STATE(4176)] = 19935,
  [SMALL_STATE(4177)] = 19976,
  [SMALL_STATE(4178)] = 20017,
  [SMALL_STATE(4179)] = 20054,
  [SMALL_STATE(4180)] = 20091,
  [SMALL_STATE(4181)] = 20132,
  [SMALL_STATE(4182)] = 20173,
  [SMALL_STATE(4183)] = 20214,
  [SMALL_STATE(4184)] = 20241,
  [SMALL_STATE(4185)] = 20286,
  [SMALL_STATE(4186)] = 20327,
  [SMALL_STATE(4187)] = 20368,
  [SMALL_STATE(4188)] = 20405,
  [SMALL_STATE(4189)] = 20446,
  [SMALL_STATE(4190)] = 20487,
  [SMALL_STATE(4191)] = 20528,
  [SMALL_STATE(4192)] = 20569,
  [SMALL_STATE(4193)] = 20596,
  [SMALL_STATE(4194)] = 20637,
  [SMALL_STATE(4195)] = 20674,
  [SMALL_STATE(4196)] = 20715,
  [SMALL_STATE(4197)] = 20756,
  [SMALL_STATE(4198)] = 20797,
  [SMALL_STATE(4199)] = 20838,
  [SMALL_STATE(4200)] = 20883,
  [SMALL_STATE(4201)] = 20924,
  [SMALL_STATE(4202)] = 20965,
  [SMALL_STATE(4203)] = 20992,
  [SMALL_STATE(4204)] = 21019,
  [SMALL_STATE(4205)] = 21056,
  [SMALL_STATE(4206)] = 21097,
  [SMALL_STATE(4207)] = 21138,
  [SMALL_STATE(4208)] = 21179,
  [SMALL_STATE(4209)] = 21220,
  [SMALL_STATE(4210)] = 21247,
  [SMALL_STATE(4211)] = 21276,
  [SMALL_STATE(4212)] = 21317,
  [SMALL_STATE(4213)] = 21344,
  [SMALL_STATE(4214)] = 21385,
  [SMALL_STATE(4215)] = 21412,
  [SMALL_STATE(4216)] = 21453,
  [SMALL_STATE(4217)] = 21494,
  [SMALL_STATE(4218)] = 21535,
  [SMALL_STATE(4219)] = 21576,
  [SMALL_STATE(4220)] = 21603,
  [SMALL_STATE(4221)] = 21644,
  [SMALL_STATE(4222)] = 21685,
  [SMALL_STATE(4223)] = 21712,
  [SMALL_STATE(4224)] = 21749,
  [SMALL_STATE(4225)] = 21786,
  [SMALL_STATE(4226)] = 21823,
  [SMALL_STATE(4227)] = 21864,
  [SMALL_STATE(4228)] = 21891,
  [SMALL_STATE(4229)] = 21932,
  [SMALL_STATE(4230)] = 21959,
  [SMALL_STATE(4231)] = 22000,
  [SMALL_STATE(4232)] = 22041,
  [SMALL_STATE(4233)] = 22082,
  [SMALL_STATE(4234)] = 22123,
  [SMALL_STATE(4235)] = 22164,
  [SMALL_STATE(4236)] = 22203,
  [SMALL_STATE(4237)] = 22244,
  [SMALL_STATE(4238)] = 22281,
  [SMALL_STATE(4239)] = 22322,
  [SMALL_STATE(4240)] = 22363,
  [SMALL_STATE(4241)] = 22404,
  [SMALL_STATE(4242)] = 22431,
  [SMALL_STATE(4243)] = 22472,
  [SMALL_STATE(4244)] = 22513,
  [SMALL_STATE(4245)] = 22540,
  [SMALL_STATE(4246)] = 22577,
  [SMALL_STATE(4247)] = 22604,
  [SMALL_STATE(4248)] = 22645,
  [SMALL_STATE(4249)] = 22686,
  [SMALL_STATE(4250)] = 22727,
  [SMALL_STATE(4251)] = 22768,
  [SMALL_STATE(4252)] = 22797,
  [SMALL_STATE(4253)] = 22838,
  [SMALL_STATE(4254)] = 22879,
  [SMALL_STATE(4255)] = 22906,
  [SMALL_STATE(4256)] = 22947,
  [SMALL_STATE(4257)] = 22974,
  [SMALL_STATE(4258)] = 23001,
  [SMALL_STATE(4259)] = 23028,
  [SMALL_STATE(4260)] = 23069,
  [SMALL_STATE(4261)] = 23110,
  [SMALL_STATE(4262)] = 23147,
  [SMALL_STATE(4263)] = 23188,
  [SMALL_STATE(4264)] = 23225,
  [SMALL_STATE(4265)] = 23262,
  [SMALL_STATE(4266)] = 23289,
  [SMALL_STATE(4267)] = 23326,
  [SMALL_STATE(4268)] = 23363,
  [SMALL_STATE(4269)] = 23400,
  [SMALL_STATE(4270)] = 23437,
  [SMALL_STATE(4271)] = 23474,
  [SMALL_STATE(4272)] = 23515,
  [SMALL_STATE(4273)] = 23552,
  [SMALL_STATE(4274)] = 23593,
  [SMALL_STATE(4275)] = 23634,
  [SMALL_STATE(4276)] = 23675,
  [SMALL_STATE(4277)] = 23716,
  [SMALL_STATE(4278)] = 23757,
  [SMALL_STATE(4279)] = 23798,
  [SMALL_STATE(4280)] = 23839,
  [SMALL_STATE(4281)] = 23880,
  [SMALL_STATE(4282)] = 23921,
  [SMALL_STATE(4283)] = 23948,
  [SMALL_STATE(4284)] = 23985,
  [SMALL_STATE(4285)] = 24026,
  [SMALL_STATE(4286)] = 24067,
  [SMALL_STATE(4287)] = 24104,
  [SMALL_STATE(4288)] = 24141,
  [SMALL_STATE(4289)] = 24168,
  [SMALL_STATE(4290)] = 24209,
  [SMALL_STATE(4291)] = 24250,
  [SMALL_STATE(4292)] = 24287,
  [SMALL_STATE(4293)] = 24314,
  [SMALL_STATE(4294)] = 24341,
  [SMALL_STATE(4295)] = 24382,
  [SMALL_STATE(4296)] = 24419,
  [SMALL_STATE(4297)] = 24460,
  [SMALL_STATE(4298)] = 24497,
  [SMALL_STATE(4299)] = 24538,
  [SMALL_STATE(4300)] = 24579,
  [SMALL_STATE(4301)] = 24610,
  [SMALL_STATE(4302)] = 24641,
  [SMALL_STATE(4303)] = 24682,
  [SMALL_STATE(4304)] = 24720,
  [SMALL_STATE(4305)] = 24762,
  [SMALL_STATE(4306)] = 24804,
  [SMALL_STATE(4307)] = 24846,
  [SMALL_STATE(4308)] = 24888,
  [SMALL_STATE(4309)] = 24918,
  [SMALL_STATE(4310)] = 24948,
  [SMALL_STATE(4311)] = 24974,
  [SMALL_STATE(4312)] = 25000,
  [SMALL_STATE(4313)] = 25042,
  [SMALL_STATE(4314)] = 25084,
  [SMALL_STATE(4315)] = 25110,
  [SMALL_STATE(4316)] = 25148,
  [SMALL_STATE(4317)] = 25190,
  [SMALL_STATE(4318)] = 25232,
  [SMALL_STATE(4319)] = 25270,
  [SMALL_STATE(4320)] = 25312,
  [SMALL_STATE(4321)] = 25354,
  [SMALL_STATE(4322)] = 25396,
  [SMALL_STATE(4323)] = 25438,
  [SMALL_STATE(4324)] = 25480,
  [SMALL_STATE(4325)] = 25522,
  [SMALL_STATE(4326)] = 25548,
  [SMALL_STATE(4327)] = 25574,
  [SMALL_STATE(4328)] = 25616,
  [SMALL_STATE(4329)] = 25658,
  [SMALL_STATE(4330)] = 25684,
  [SMALL_STATE(4331)] = 25710,
  [SMALL_STATE(4332)] = 25736,
  [SMALL_STATE(4333)] = 25778,
  [SMALL_STATE(4334)] = 25820,
  [SMALL_STATE(4335)] = 25862,
  [SMALL_STATE(4336)] = 25904,
  [SMALL_STATE(4337)] = 25946,
  [SMALL_STATE(4338)] = 25988,
  [SMALL_STATE(4339)] = 26030,
  [SMALL_STATE(4340)] = 26072,
  [SMALL_STATE(4341)] = 26114,
  [SMALL_STATE(4342)] = 26156,
  [SMALL_STATE(4343)] = 26198,
  [SMALL_STATE(4344)] = 26240,
  [SMALL_STATE(4345)] = 26282,
  [SMALL_STATE(4346)] = 26324,
  [SMALL_STATE(4347)] = 26366,
  [SMALL_STATE(4348)] = 26408,
  [SMALL_STATE(4349)] = 26450,
  [SMALL_STATE(4350)] = 26492,
  [SMALL_STATE(4351)] = 26524,
  [SMALL_STATE(4352)] = 26562,
  [SMALL_STATE(4353)] = 26600,
  [SMALL_STATE(4354)] = 26638,
  [SMALL_STATE(4355)] = 26676,
  [SMALL_STATE(4356)] = 26714,
  [SMALL_STATE(4357)] = 26756,
  [SMALL_STATE(4358)] = 26798,
  [SMALL_STATE(4359)] = 26824,
  [SMALL_STATE(4360)] = 26866,
  [SMALL_STATE(4361)] = 26908,
  [SMALL_STATE(4362)] = 26934,
  [SMALL_STATE(4363)] = 26972,
  [SMALL_STATE(4364)] = 27014,
  [SMALL_STATE(4365)] = 27056,
  [SMALL_STATE(4366)] = 27098,
  [SMALL_STATE(4367)] = 27140,
  [SMALL_STATE(4368)] = 27178,
  [SMALL_STATE(4369)] = 27220,
  [SMALL_STATE(4370)] = 27262,
  [SMALL_STATE(4371)] = 27288,
  [SMALL_STATE(4372)] = 27316,
  [SMALL_STATE(4373)] = 27344,
  [SMALL_STATE(4374)] = 27386,
  [SMALL_STATE(4375)] = 27418,
  [SMALL_STATE(4376)] = 27460,
  [SMALL_STATE(4377)] = 27502,
  [SMALL_STATE(4378)] = 27544,
  [SMALL_STATE(4379)] = 27586,
  [SMALL_STATE(4380)] = 27628,
  [SMALL_STATE(4381)] = 27670,
  [SMALL_STATE(4382)] = 27712,
  [SMALL_STATE(4383)] = 27738,
  [SMALL_STATE(4384)] = 27764,
  [SMALL_STATE(4385)] = 27790,
  [SMALL_STATE(4386)] = 27816,
  [SMALL_STATE(4387)] = 27858,
  [SMALL_STATE(4388)] = 27896,
  [SMALL_STATE(4389)] = 27938,
  [SMALL_STATE(4390)] = 27968,
  [SMALL_STATE(4391)] = 27994,
  [SMALL_STATE(4392)] = 28020,
  [SMALL_STATE(4393)] = 28058,
  [SMALL_STATE(4394)] = 28100,
  [SMALL_STATE(4395)] = 28142,
  [SMALL_STATE(4396)] = 28165,
  [SMALL_STATE(4397)] = 28188,
  [SMALL_STATE(4398)] = 28227,
  [SMALL_STATE(4399)] = 28266,
  [SMALL_STATE(4400)] = 28289,
  [SMALL_STATE(4401)] = 28312,
  [SMALL_STATE(4402)] = 28335,
  [SMALL_STATE(4403)] = 28358,
  [SMALL_STATE(4404)] = 28397,
  [SMALL_STATE(4405)] = 28436,
  [SMALL_STATE(4406)] = 28471,
  [SMALL_STATE(4407)] = 28494,
  [SMALL_STATE(4408)] = 28517,
  [SMALL_STATE(4409)] = 28540,
  [SMALL_STATE(4410)] = 28569,
  [SMALL_STATE(4411)] = 28592,
  [SMALL_STATE(4412)] = 28617,
  [SMALL_STATE(4413)] = 28640,
  [SMALL_STATE(4414)] = 28663,
  [SMALL_STATE(4415)] = 28686,
  [SMALL_STATE(4416)] = 28709,
  [SMALL_STATE(4417)] = 28738,
  [SMALL_STATE(4418)] = 28761,
  [SMALL_STATE(4419)] = 28784,
  [SMALL_STATE(4420)] = 28807,
  [SMALL_STATE(4421)] = 28830,
  [SMALL_STATE(4422)] = 28865,
  [SMALL_STATE(4423)] = 28900,
  [SMALL_STATE(4424)] = 28939,
  [SMALL_STATE(4425)] = 28962,
  [SMALL_STATE(4426)] = 28985,
  [SMALL_STATE(4427)] = 29008,
  [SMALL_STATE(4428)] = 29047,
  [SMALL_STATE(4429)] = 29070,
  [SMALL_STATE(4430)] = 29109,
  [SMALL_STATE(4431)] = 29148,
  [SMALL_STATE(4432)] = 29173,
  [SMALL_STATE(4433)] = 29196,
  [SMALL_STATE(4434)] = 29221,
  [SMALL_STATE(4435)] = 29260,
  [SMALL_STATE(4436)] = 29283,
  [SMALL_STATE(4437)] = 29306,
  [SMALL_STATE(4438)] = 29345,
  [SMALL_STATE(4439)] = 29384,
  [SMALL_STATE(4440)] = 29407,
  [SMALL_STATE(4441)] = 29430,
  [SMALL_STATE(4442)] = 29453,
  [SMALL_STATE(4443)] = 29476,
  [SMALL_STATE(4444)] = 29515,
  [SMALL_STATE(4445)] = 29538,
  [SMALL_STATE(4446)] = 29561,
  [SMALL_STATE(4447)] = 29584,
  [SMALL_STATE(4448)] = 29607,
  [SMALL_STATE(4449)] = 29630,
  [SMALL_STATE(4450)] = 29653,
  [SMALL_STATE(4451)] = 29692,
  [SMALL_STATE(4452)] = 29717,
  [SMALL_STATE(4453)] = 29740,
  [SMALL_STATE(4454)] = 29765,
  [SMALL_STATE(4455)] = 29788,
  [SMALL_STATE(4456)] = 29811,
  [SMALL_STATE(4457)] = 29834,
  [SMALL_STATE(4458)] = 29857,
  [SMALL_STATE(4459)] = 29882,
  [SMALL_STATE(4460)] = 29907,
  [SMALL_STATE(4461)] = 29930,
  [SMALL_STATE(4462)] = 29953,
  [SMALL_STATE(4463)] = 29976,
  [SMALL_STATE(4464)] = 30001,
  [SMALL_STATE(4465)] = 30026,
  [SMALL_STATE(4466)] = 30049,
  [SMALL_STATE(4467)] = 30084,
  [SMALL_STATE(4468)] = 30119,
  [SMALL_STATE(4469)] = 30158,
  [SMALL_STATE(4470)] = 30181,
  [SMALL_STATE(4471)] = 30204,
  [SMALL_STATE(4472)] = 30227,
  [SMALL_STATE(4473)] = 30250,
  [SMALL_STATE(4474)] = 30273,
  [SMALL_STATE(4475)] = 30296,
  [SMALL_STATE(4476)] = 30319,
  [SMALL_STATE(4477)] = 30342,
  [SMALL_STATE(4478)] = 30365,
  [SMALL_STATE(4479)] = 30404,
  [SMALL_STATE(4480)] = 30433,
  [SMALL_STATE(4481)] = 30468,
  [SMALL_STATE(4482)] = 30491,
  [SMALL_STATE(4483)] = 30514,
  [SMALL_STATE(4484)] = 30537,
  [SMALL_STATE(4485)] = 30560,
  [SMALL_STATE(4486)] = 30592,
  [SMALL_STATE(4487)] = 30614,
  [SMALL_STATE(4488)] = 30636,
  [SMALL_STATE(4489)] = 30662,
  [SMALL_STATE(4490)] = 30698,
  [SMALL_STATE(4491)] = 30734,
  [SMALL_STATE(4492)] = 30768,
  [SMALL_STATE(4493)] = 30790,
  [SMALL_STATE(4494)] = 30812,
  [SMALL_STATE(4495)] = 30842,
  [SMALL_STATE(4496)] = 30864,
  [SMALL_STATE(4497)] = 30898,
  [SMALL_STATE(4498)] = 30920,
  [SMALL_STATE(4499)] = 30954,
  [SMALL_STATE(4500)] = 30976,
  [SMALL_STATE(4501)] = 31010,
  [SMALL_STATE(4502)] = 31032,
  [SMALL_STATE(4503)] = 31054,
  [SMALL_STATE(4504)] = 31078,
  [SMALL_STATE(4505)] = 31100,
  [SMALL_STATE(4506)] = 31122,
  [SMALL_STATE(4507)] = 31144,
  [SMALL_STATE(4508)] = 31166,
  [SMALL_STATE(4509)] = 31188,
  [SMALL_STATE(4510)] = 31210,
  [SMALL_STATE(4511)] = 31244,
  [SMALL_STATE(4512)] = 31270,
  [SMALL_STATE(4513)] = 31292,
  [SMALL_STATE(4514)] = 31314,
  [SMALL_STATE(4515)] = 31346,
  [SMALL_STATE(4516)] = 31370,
  [SMALL_STATE(4517)] = 31404,
  [SMALL_STATE(4518)] = 31426,
  [SMALL_STATE(4519)] = 31458,
  [SMALL_STATE(4520)] = 31480,
  [SMALL_STATE(4521)] = 31514,
  [SMALL_STATE(4522)] = 31536,
  [SMALL_STATE(4523)] = 31570,
  [SMALL_STATE(4524)] = 31602,
  [SMALL_STATE(4525)] = 31624,
  [SMALL_STATE(4526)] = 31646,
  [SMALL_STATE(4527)] = 31682,
  [SMALL_STATE(4528)] = 31704,
  [SMALL_STATE(4529)] = 31740,
  [SMALL_STATE(4530)] = 31762,
  [SMALL_STATE(4531)] = 31784,
  [SMALL_STATE(4532)] = 31806,
  [SMALL_STATE(4533)] = 31830,
  [SMALL_STATE(4534)] = 31862,
  [SMALL_STATE(4535)] = 31884,
  [SMALL_STATE(4536)] = 31906,
  [SMALL_STATE(4537)] = 31928,
  [SMALL_STATE(4538)] = 31950,
  [SMALL_STATE(4539)] = 31980,
  [SMALL_STATE(4540)] = 32012,
  [SMALL_STATE(4541)] = 32036,
  [SMALL_STATE(4542)] = 32058,
  [SMALL_STATE(4543)] = 32080,
  [SMALL_STATE(4544)] = 32102,
  [SMALL_STATE(4545)] = 32136,
  [SMALL_STATE(4546)] = 32158,
  [SMALL_STATE(4547)] = 32184,
  [SMALL_STATE(4548)] = 32206,
  [SMALL_STATE(4549)] = 32240,
  [SMALL_STATE(4550)] = 32266,
  [SMALL_STATE(4551)] = 32298,
  [SMALL_STATE(4552)] = 32320,
  [SMALL_STATE(4553)] = 32342,
  [SMALL_STATE(4554)] = 32364,
  [SMALL_STATE(4555)] = 32395,
  [SMALL_STATE(4556)] = 32420,
  [SMALL_STATE(4557)] = 32441,
  [SMALL_STATE(4558)] = 32462,
  [SMALL_STATE(4559)] = 32483,
  [SMALL_STATE(4560)] = 32504,
  [SMALL_STATE(4561)] = 32525,
  [SMALL_STATE(4562)] = 32546,
  [SMALL_STATE(4563)] = 32567,
  [SMALL_STATE(4564)] = 32588,
  [SMALL_STATE(4565)] = 32609,
  [SMALL_STATE(4566)] = 32630,
  [SMALL_STATE(4567)] = 32651,
  [SMALL_STATE(4568)] = 32674,
  [SMALL_STATE(4569)] = 32695,
  [SMALL_STATE(4570)] = 32718,
  [SMALL_STATE(4571)] = 32745,
  [SMALL_STATE(4572)] = 32774,
  [SMALL_STATE(4573)] = 32799,
  [SMALL_STATE(4574)] = 32824,
  [SMALL_STATE(4575)] = 32845,
  [SMALL_STATE(4576)] = 32870,
  [SMALL_STATE(4577)] = 32891,
  [SMALL_STATE(4578)] = 32912,
  [SMALL_STATE(4579)] = 32933,
  [SMALL_STATE(4580)] = 32956,
  [SMALL_STATE(4581)] = 32977,
  [SMALL_STATE(4582)] = 33004,
  [SMALL_STATE(4583)] = 33031,
  [SMALL_STATE(4584)] = 33052,
  [SMALL_STATE(4585)] = 33077,
  [SMALL_STATE(4586)] = 33098,
  [SMALL_STATE(4587)] = 33119,
  [SMALL_STATE(4588)] = 33140,
  [SMALL_STATE(4589)] = 33163,
  [SMALL_STATE(4590)] = 33186,
  [SMALL_STATE(4591)] = 33207,
  [SMALL_STATE(4592)] = 33228,
  [SMALL_STATE(4593)] = 33259,
  [SMALL_STATE(4594)] = 33280,
  [SMALL_STATE(4595)] = 33301,
  [SMALL_STATE(4596)] = 33322,
  [SMALL_STATE(4597)] = 33350,
  [SMALL_STATE(4598)] = 33382,
  [SMALL_STATE(4599)] = 33410,
  [SMALL_STATE(4600)] = 33442,
  [SMALL_STATE(4601)] = 33474,
  [SMALL_STATE(4602)] = 33506,
  [SMALL_STATE(4603)] = 33538,
  [SMALL_STATE(4604)] = 33568,
  [SMALL_STATE(4605)] = 33600,
  [SMALL_STATE(4606)] = 33632,
  [SMALL_STATE(4607)] = 33660,
  [SMALL_STATE(4608)] = 33688,
  [SMALL_STATE(4609)] = 33716,
  [SMALL_STATE(4610)] = 33748,
  [SMALL_STATE(4611)] = 33780,
  [SMALL_STATE(4612)] = 33808,
  [SMALL_STATE(4613)] = 33840,
  [SMALL_STATE(4614)] = 33868,
  [SMALL_STATE(4615)] = 33898,
  [SMALL_STATE(4616)] = 33926,
  [SMALL_STATE(4617)] = 33958,
  [SMALL_STATE(4618)] = 33990,
  [SMALL_STATE(4619)] = 34022,
  [SMALL_STATE(4620)] = 34054,
  [SMALL_STATE(4621)] = 34086,
  [SMALL_STATE(4622)] = 34114,
  [SMALL_STATE(4623)] = 34146,
  [SMALL_STATE(4624)] = 34174,
  [SMALL_STATE(4625)] = 34202,
  [SMALL_STATE(4626)] = 34230,
  [SMALL_STATE(4627)] = 34262,
  [SMALL_STATE(4628)] = 34294,
  [SMALL_STATE(4629)] = 34322,
  [SMALL_STATE(4630)] = 34354,
  [SMALL_STATE(4631)] = 34386,
  [SMALL_STATE(4632)] = 34418,
  [SMALL_STATE(4633)] = 34446,
  [SMALL_STATE(4634)] = 34478,
  [SMALL_STATE(4635)] = 34510,
  [SMALL_STATE(4636)] = 34542,
  [SMALL_STATE(4637)] = 34572,
  [SMALL_STATE(4638)] = 34604,
  [SMALL_STATE(4639)] = 34636,
  [SMALL_STATE(4640)] = 34664,
  [SMALL_STATE(4641)] = 34692,
  [SMALL_STATE(4642)] = 34720,
  [SMALL_STATE(4643)] = 34752,
  [SMALL_STATE(4644)] = 34780,
  [SMALL_STATE(4645)] = 34808,
  [SMALL_STATE(4646)] = 34836,
  [SMALL_STATE(4647)] = 34868,
  [SMALL_STATE(4648)] = 34896,
  [SMALL_STATE(4649)] = 34924,
  [SMALL_STATE(4650)] = 34946,
  [SMALL_STATE(4651)] = 34974,
  [SMALL_STATE(4652)] = 35002,
  [SMALL_STATE(4653)] = 35034,
  [SMALL_STATE(4654)] = 35064,
  [SMALL_STATE(4655)] = 35096,
  [SMALL_STATE(4656)] = 35128,
  [SMALL_STATE(4657)] = 35156,
  [SMALL_STATE(4658)] = 35188,
  [SMALL_STATE(4659)] = 35220,
  [SMALL_STATE(4660)] = 35248,
  [SMALL_STATE(4661)] = 35276,
  [SMALL_STATE(4662)] = 35304,
  [SMALL_STATE(4663)] = 35332,
  [SMALL_STATE(4664)] = 35364,
  [SMALL_STATE(4665)] = 35396,
  [SMALL_STATE(4666)] = 35424,
  [SMALL_STATE(4667)] = 35456,
  [SMALL_STATE(4668)] = 35488,
  [SMALL_STATE(4669)] = 35516,
  [SMALL_STATE(4670)] = 35544,
  [SMALL_STATE(4671)] = 35572,
  [SMALL_STATE(4672)] = 35600,
  [SMALL_STATE(4673)] = 35630,
  [SMALL_STATE(4674)] = 35662,
  [SMALL_STATE(4675)] = 35692,
  [SMALL_STATE(4676)] = 35724,
  [SMALL_STATE(4677)] = 35754,
  [SMALL_STATE(4678)] = 35776,
  [SMALL_STATE(4679)] = 35808,
  [SMALL_STATE(4680)] = 35840,
  [SMALL_STATE(4681)] = 35864,
  [SMALL_STATE(4682)] = 35892,
  [SMALL_STATE(4683)] = 35922,
  [SMALL_STATE(4684)] = 35950,
  [SMALL_STATE(4685)] = 35982,
  [SMALL_STATE(4686)] = 36014,
  [SMALL_STATE(4687)] = 36042,
  [SMALL_STATE(4688)] = 36074,
  [SMALL_STATE(4689)] = 36106,
  [SMALL_STATE(4690)] = 36138,
  [SMALL_STATE(4691)] = 36162,
  [SMALL_STATE(4692)] = 36194,
  [SMALL_STATE(4693)] = 36226,
  [SMALL_STATE(4694)] = 36258,
  [SMALL_STATE(4695)] = 36290,
  [SMALL_STATE(4696)] = 36322,
  [SMALL_STATE(4697)] = 36354,
  [SMALL_STATE(4698)] = 36386,
  [SMALL_STATE(4699)] = 36416,
  [SMALL_STATE(4700)] = 36448,
  [SMALL_STATE(4701)] = 36480,
  [SMALL_STATE(4702)] = 36512,
  [SMALL_STATE(4703)] = 36542,
  [SMALL_STATE(4704)] = 36572,
  [SMALL_STATE(4705)] = 36600,
  [SMALL_STATE(4706)] = 36625,
  [SMALL_STATE(4707)] = 36650,
  [SMALL_STATE(4708)] = 36671,
  [SMALL_STATE(4709)] = 36700,
  [SMALL_STATE(4710)] = 36725,
  [SMALL_STATE(4711)] = 36754,
  [SMALL_STATE(4712)] = 36775,
  [SMALL_STATE(4713)] = 36800,
  [SMALL_STATE(4714)] = 36825,
  [SMALL_STATE(4715)] = 36850,
  [SMALL_STATE(4716)] = 36875,
  [SMALL_STATE(4717)] = 36900,
  [SMALL_STATE(4718)] = 36925,
  [SMALL_STATE(4719)] = 36954,
  [SMALL_STATE(4720)] = 36975,
  [SMALL_STATE(4721)] = 37000,
  [SMALL_STATE(4722)] = 37025,
  [SMALL_STATE(4723)] = 37050,
  [SMALL_STATE(4724)] = 37075,
  [SMALL_STATE(4725)] = 37100,
  [SMALL_STATE(4726)] = 37125,
  [SMALL_STATE(4727)] = 37150,
  [SMALL_STATE(4728)] = 37175,
  [SMALL_STATE(4729)] = 37200,
  [SMALL_STATE(4730)] = 37225,
  [SMALL_STATE(4731)] = 37254,
  [SMALL_STATE(4732)] = 37279,
  [SMALL_STATE(4733)] = 37308,
  [SMALL_STATE(4734)] = 37333,
  [SMALL_STATE(4735)] = 37362,
  [SMALL_STATE(4736)] = 37391,
  [SMALL_STATE(4737)] = 37416,
  [SMALL_STATE(4738)] = 37445,
  [SMALL_STATE(4739)] = 37470,
  [SMALL_STATE(4740)] = 37495,
  [SMALL_STATE(4741)] = 37520,
  [SMALL_STATE(4742)] = 37545,
  [SMALL_STATE(4743)] = 37570,
  [SMALL_STATE(4744)] = 37595,
  [SMALL_STATE(4745)] = 37620,
  [SMALL_STATE(4746)] = 37645,
  [SMALL_STATE(4747)] = 37670,
  [SMALL_STATE(4748)] = 37695,
  [SMALL_STATE(4749)] = 37724,
  [SMALL_STATE(4750)] = 37749,
  [SMALL_STATE(4751)] = 37774,
  [SMALL_STATE(4752)] = 37799,
  [SMALL_STATE(4753)] = 37824,
  [SMALL_STATE(4754)] = 37849,
  [SMALL_STATE(4755)] = 37874,
  [SMALL_STATE(4756)] = 37903,
  [SMALL_STATE(4757)] = 37928,
  [SMALL_STATE(4758)] = 37957,
  [SMALL_STATE(4759)] = 37982,
  [SMALL_STATE(4760)] = 38011,
  [SMALL_STATE(4761)] = 38036,
  [SMALL_STATE(4762)] = 38061,
  [SMALL_STATE(4763)] = 38086,
  [SMALL_STATE(4764)] = 38111,
  [SMALL_STATE(4765)] = 38136,
  [SMALL_STATE(4766)] = 38161,
  [SMALL_STATE(4767)] = 38186,
  [SMALL_STATE(4768)] = 38211,
  [SMALL_STATE(4769)] = 38236,
  [SMALL_STATE(4770)] = 38261,
  [SMALL_STATE(4771)] = 38282,
  [SMALL_STATE(4772)] = 38307,
  [SMALL_STATE(4773)] = 38332,
  [SMALL_STATE(4774)] = 38357,
  [SMALL_STATE(4775)] = 38382,
  [SMALL_STATE(4776)] = 38407,
  [SMALL_STATE(4777)] = 38432,
  [SMALL_STATE(4778)] = 38457,
  [SMALL_STATE(4779)] = 38482,
  [SMALL_STATE(4780)] = 38507,
  [SMALL_STATE(4781)] = 38532,
  [SMALL_STATE(4782)] = 38557,
  [SMALL_STATE(4783)] = 38582,
  [SMALL_STATE(4784)] = 38607,
  [SMALL_STATE(4785)] = 38632,
  [SMALL_STATE(4786)] = 38657,
  [SMALL_STATE(4787)] = 38686,
  [SMALL_STATE(4788)] = 38711,
  [SMALL_STATE(4789)] = 38736,
  [SMALL_STATE(4790)] = 38761,
  [SMALL_STATE(4791)] = 38790,
  [SMALL_STATE(4792)] = 38815,
  [SMALL_STATE(4793)] = 38840,
  [SMALL_STATE(4794)] = 38859,
  [SMALL_STATE(4795)] = 38878,
  [SMALL_STATE(4796)] = 38907,
  [SMALL_STATE(4797)] = 38932,
  [SMALL_STATE(4798)] = 38951,
  [SMALL_STATE(4799)] = 38976,
  [SMALL_STATE(4800)] = 39001,
  [SMALL_STATE(4801)] = 39026,
  [SMALL_STATE(4802)] = 39051,
  [SMALL_STATE(4803)] = 39076,
  [SMALL_STATE(4804)] = 39101,
  [SMALL_STATE(4805)] = 39126,
  [SMALL_STATE(4806)] = 39151,
  [SMALL_STATE(4807)] = 39180,
  [SMALL_STATE(4808)] = 39209,
  [SMALL_STATE(4809)] = 39234,
  [SMALL_STATE(4810)] = 39263,
  [SMALL_STATE(4811)] = 39292,
  [SMALL_STATE(4812)] = 39317,
  [SMALL_STATE(4813)] = 39346,
  [SMALL_STATE(4814)] = 39371,
  [SMALL_STATE(4815)] = 39396,
  [SMALL_STATE(4816)] = 39421,
  [SMALL_STATE(4817)] = 39450,
  [SMALL_STATE(4818)] = 39475,
  [SMALL_STATE(4819)] = 39500,
  [SMALL_STATE(4820)] = 39525,
  [SMALL_STATE(4821)] = 39550,
  [SMALL_STATE(4822)] = 39575,
  [SMALL_STATE(4823)] = 39600,
  [SMALL_STATE(4824)] = 39629,
  [SMALL_STATE(4825)] = 39654,
  [SMALL_STATE(4826)] = 39683,
  [SMALL_STATE(4827)] = 39708,
  [SMALL_STATE(4828)] = 39733,
  [SMALL_STATE(4829)] = 39762,
  [SMALL_STATE(4830)] = 39787,
  [SMALL_STATE(4831)] = 39816,
  [SMALL_STATE(4832)] = 39841,
  [SMALL_STATE(4833)] = 39866,
  [SMALL_STATE(4834)] = 39891,
  [SMALL_STATE(4835)] = 39916,
  [SMALL_STATE(4836)] = 39941,
  [SMALL_STATE(4837)] = 39966,
  [SMALL_STATE(4838)] = 39991,
  [SMALL_STATE(4839)] = 40016,
  [SMALL_STATE(4840)] = 40041,
  [SMALL_STATE(4841)] = 40066,
  [SMALL_STATE(4842)] = 40091,
  [SMALL_STATE(4843)] = 40120,
  [SMALL_STATE(4844)] = 40149,
  [SMALL_STATE(4845)] = 40174,
  [SMALL_STATE(4846)] = 40199,
  [SMALL_STATE(4847)] = 40224,
  [SMALL_STATE(4848)] = 40249,
  [SMALL_STATE(4849)] = 40278,
  [SMALL_STATE(4850)] = 40307,
  [SMALL_STATE(4851)] = 40336,
  [SMALL_STATE(4852)] = 40365,
  [SMALL_STATE(4853)] = 40394,
  [SMALL_STATE(4854)] = 40419,
  [SMALL_STATE(4855)] = 40444,
  [SMALL_STATE(4856)] = 40469,
  [SMALL_STATE(4857)] = 40498,
  [SMALL_STATE(4858)] = 40527,
  [SMALL_STATE(4859)] = 40552,
  [SMALL_STATE(4860)] = 40577,
  [SMALL_STATE(4861)] = 40606,
  [SMALL_STATE(4862)] = 40631,
  [SMALL_STATE(4863)] = 40660,
  [SMALL_STATE(4864)] = 40689,
  [SMALL_STATE(4865)] = 40718,
  [SMALL_STATE(4866)] = 40747,
  [SMALL_STATE(4867)] = 40772,
  [SMALL_STATE(4868)] = 40797,
  [SMALL_STATE(4869)] = 40822,
  [SMALL_STATE(4870)] = 40841,
  [SMALL_STATE(4871)] = 40866,
  [SMALL_STATE(4872)] = 40891,
  [SMALL_STATE(4873)] = 40920,
  [SMALL_STATE(4874)] = 40945,
  [SMALL_STATE(4875)] = 40970,
  [SMALL_STATE(4876)] = 40999,
  [SMALL_STATE(4877)] = 41028,
  [SMALL_STATE(4878)] = 41057,
  [SMALL_STATE(4879)] = 41082,
  [SMALL_STATE(4880)] = 41111,
  [SMALL_STATE(4881)] = 41136,
  [SMALL_STATE(4882)] = 41161,
  [SMALL_STATE(4883)] = 41186,
  [SMALL_STATE(4884)] = 41211,
  [SMALL_STATE(4885)] = 41236,
  [SMALL_STATE(4886)] = 41261,
  [SMALL_STATE(4887)] = 41286,
  [SMALL_STATE(4888)] = 41311,
  [SMALL_STATE(4889)] = 41330,
  [SMALL_STATE(4890)] = 41355,
  [SMALL_STATE(4891)] = 41376,
  [SMALL_STATE(4892)] = 41401,
  [SMALL_STATE(4893)] = 41430,
  [SMALL_STATE(4894)] = 41459,
  [SMALL_STATE(4895)] = 41484,
  [SMALL_STATE(4896)] = 41509,
  [SMALL_STATE(4897)] = 41534,
  [SMALL_STATE(4898)] = 41559,
  [SMALL_STATE(4899)] = 41588,
  [SMALL_STATE(4900)] = 41617,
  [SMALL_STATE(4901)] = 41642,
  [SMALL_STATE(4902)] = 41667,
  [SMALL_STATE(4903)] = 41696,
  [SMALL_STATE(4904)] = 41721,
  [SMALL_STATE(4905)] = 41746,
  [SMALL_STATE(4906)] = 41775,
  [SMALL_STATE(4907)] = 41800,
  [SMALL_STATE(4908)] = 41825,
  [SMALL_STATE(4909)] = 41850,
  [SMALL_STATE(4910)] = 41875,
  [SMALL_STATE(4911)] = 41900,
  [SMALL_STATE(4912)] = 41925,
  [SMALL_STATE(4913)] = 41954,
  [SMALL_STATE(4914)] = 41975,
  [SMALL_STATE(4915)] = 42004,
  [SMALL_STATE(4916)] = 42029,
  [SMALL_STATE(4917)] = 42048,
  [SMALL_STATE(4918)] = 42073,
  [SMALL_STATE(4919)] = 42098,
  [SMALL_STATE(4920)] = 42123,
  [SMALL_STATE(4921)] = 42148,
  [SMALL_STATE(4922)] = 42169,
  [SMALL_STATE(4923)] = 42194,
  [SMALL_STATE(4924)] = 42219,
  [SMALL_STATE(4925)] = 42244,
  [SMALL_STATE(4926)] = 42269,
  [SMALL_STATE(4927)] = 42294,
  [SMALL_STATE(4928)] = 42319,
  [SMALL_STATE(4929)] = 42344,
  [SMALL_STATE(4930)] = 42369,
  [SMALL_STATE(4931)] = 42388,
  [SMALL_STATE(4932)] = 42413,
  [SMALL_STATE(4933)] = 42438,
  [SMALL_STATE(4934)] = 42457,
  [SMALL_STATE(4935)] = 42482,
  [SMALL_STATE(4936)] = 42501,
  [SMALL_STATE(4937)] = 42526,
  [SMALL_STATE(4938)] = 42547,
  [SMALL_STATE(4939)] = 42566,
  [SMALL_STATE(4940)] = 42591,
  [SMALL_STATE(4941)] = 42616,
  [SMALL_STATE(4942)] = 42641,
  [SMALL_STATE(4943)] = 42666,
  [SMALL_STATE(4944)] = 42691,
  [SMALL_STATE(4945)] = 42720,
  [SMALL_STATE(4946)] = 42745,
  [SMALL_STATE(4947)] = 42774,
  [SMALL_STATE(4948)] = 42799,
  [SMALL_STATE(4949)] = 42822,
  [SMALL_STATE(4950)] = 42851,
  [SMALL_STATE(4951)] = 42876,
  [SMALL_STATE(4952)] = 42901,
  [SMALL_STATE(4953)] = 42930,
  [SMALL_STATE(4954)] = 42959,
  [SMALL_STATE(4955)] = 42978,
  [SMALL_STATE(4956)] = 43003,
  [SMALL_STATE(4957)] = 43028,
  [SMALL_STATE(4958)] = 43053,
  [SMALL_STATE(4959)] = 43078,
  [SMALL_STATE(4960)] = 43103,
  [SMALL_STATE(4961)] = 43128,
  [SMALL_STATE(4962)] = 43153,
  [SMALL_STATE(4963)] = 43182,
  [SMALL_STATE(4964)] = 43207,
  [SMALL_STATE(4965)] = 43226,
  [SMALL_STATE(4966)] = 43251,
  [SMALL_STATE(4967)] = 43276,
  [SMALL_STATE(4968)] = 43301,
  [SMALL_STATE(4969)] = 43326,
  [SMALL_STATE(4970)] = 43345,
  [SMALL_STATE(4971)] = 43370,
  [SMALL_STATE(4972)] = 43391,
  [SMALL_STATE(4973)] = 43420,
  [SMALL_STATE(4974)] = 43448,
  [SMALL_STATE(4975)] = 43470,
  [SMALL_STATE(4976)] = 43496,
  [SMALL_STATE(4977)] = 43518,
  [SMALL_STATE(4978)] = 43540,
  [SMALL_STATE(4979)] = 43562,
  [SMALL_STATE(4980)] = 43584,
  [SMALL_STATE(4981)] = 43606,
  [SMALL_STATE(4982)] = 43634,
  [SMALL_STATE(4983)] = 43652,
  [SMALL_STATE(4984)] = 43674,
  [SMALL_STATE(4985)] = 43696,
  [SMALL_STATE(4986)] = 43718,
  [SMALL_STATE(4987)] = 43740,
  [SMALL_STATE(4988)] = 43762,
  [SMALL_STATE(4989)] = 43790,
  [SMALL_STATE(4990)] = 43818,
  [SMALL_STATE(4991)] = 43840,
  [SMALL_STATE(4992)] = 43862,
  [SMALL_STATE(4993)] = 43884,
  [SMALL_STATE(4994)] = 43906,
  [SMALL_STATE(4995)] = 43928,
  [SMALL_STATE(4996)] = 43950,
  [SMALL_STATE(4997)] = 43972,
  [SMALL_STATE(4998)] = 43994,
  [SMALL_STATE(4999)] = 44016,
  [SMALL_STATE(5000)] = 44038,
  [SMALL_STATE(5001)] = 44060,
  [SMALL_STATE(5002)] = 44082,
  [SMALL_STATE(5003)] = 44104,
  [SMALL_STATE(5004)] = 44126,
  [SMALL_STATE(5005)] = 44148,
  [SMALL_STATE(5006)] = 44170,
  [SMALL_STATE(5007)] = 44192,
  [SMALL_STATE(5008)] = 44220,
  [SMALL_STATE(5009)] = 44248,
  [SMALL_STATE(5010)] = 44276,
  [SMALL_STATE(5011)] = 44298,
  [SMALL_STATE(5012)] = 44320,
  [SMALL_STATE(5013)] = 44342,
  [SMALL_STATE(5014)] = 44362,
  [SMALL_STATE(5015)] = 44384,
  [SMALL_STATE(5016)] = 44406,
  [SMALL_STATE(5017)] = 44428,
  [SMALL_STATE(5018)] = 44450,
  [SMALL_STATE(5019)] = 44472,
  [SMALL_STATE(5020)] = 44500,
  [SMALL_STATE(5021)] = 44528,
  [SMALL_STATE(5022)] = 44550,
  [SMALL_STATE(5023)] = 44578,
  [SMALL_STATE(5024)] = 44606,
  [SMALL_STATE(5025)] = 44634,
  [SMALL_STATE(5026)] = 44662,
  [SMALL_STATE(5027)] = 44684,
  [SMALL_STATE(5028)] = 44706,
  [SMALL_STATE(5029)] = 44728,
  [SMALL_STATE(5030)] = 44750,
  [SMALL_STATE(5031)] = 44770,
  [SMALL_STATE(5032)] = 44798,
  [SMALL_STATE(5033)] = 44820,
  [SMALL_STATE(5034)] = 44848,
  [SMALL_STATE(5035)] = 44870,
  [SMALL_STATE(5036)] = 44894,
  [SMALL_STATE(5037)] = 44914,
  [SMALL_STATE(5038)] = 44936,
  [SMALL_STATE(5039)] = 44958,
  [SMALL_STATE(5040)] = 44980,
  [SMALL_STATE(5041)] = 45000,
  [SMALL_STATE(5042)] = 45022,
  [SMALL_STATE(5043)] = 45044,
  [SMALL_STATE(5044)] = 45066,
  [SMALL_STATE(5045)] = 45090,
  [SMALL_STATE(5046)] = 45112,
  [SMALL_STATE(5047)] = 45134,
  [SMALL_STATE(5048)] = 45156,
  [SMALL_STATE(5049)] = 45174,
  [SMALL_STATE(5050)] = 45194,
  [SMALL_STATE(5051)] = 45214,
  [SMALL_STATE(5052)] = 45232,
  [SMALL_STATE(5053)] = 45254,
  [SMALL_STATE(5054)] = 45282,
  [SMALL_STATE(5055)] = 45310,
  [SMALL_STATE(5056)] = 45332,
  [SMALL_STATE(5057)] = 45350,
  [SMALL_STATE(5058)] = 45378,
  [SMALL_STATE(5059)] = 45398,
  [SMALL_STATE(5060)] = 45424,
  [SMALL_STATE(5061)] = 45444,
  [SMALL_STATE(5062)] = 45466,
  [SMALL_STATE(5063)] = 45488,
  [SMALL_STATE(5064)] = 45510,
  [SMALL_STATE(5065)] = 45532,
  [SMALL_STATE(5066)] = 45554,
  [SMALL_STATE(5067)] = 45576,
  [SMALL_STATE(5068)] = 45598,
  [SMALL_STATE(5069)] = 45620,
  [SMALL_STATE(5070)] = 45642,
  [SMALL_STATE(5071)] = 45664,
  [SMALL_STATE(5072)] = 45686,
  [SMALL_STATE(5073)] = 45710,
  [SMALL_STATE(5074)] = 45738,
  [SMALL_STATE(5075)] = 45760,
  [SMALL_STATE(5076)] = 45782,
  [SMALL_STATE(5077)] = 45804,
  [SMALL_STATE(5078)] = 45830,
  [SMALL_STATE(5079)] = 45856,
  [SMALL_STATE(5080)] = 45878,
  [SMALL_STATE(5081)] = 45900,
  [SMALL_STATE(5082)] = 45922,
  [SMALL_STATE(5083)] = 45950,
  [SMALL_STATE(5084)] = 45978,
  [SMALL_STATE(5085)] = 46006,
  [SMALL_STATE(5086)] = 46032,
  [SMALL_STATE(5087)] = 46054,
  [SMALL_STATE(5088)] = 46076,
  [SMALL_STATE(5089)] = 46098,
  [SMALL_STATE(5090)] = 46120,
  [SMALL_STATE(5091)] = 46142,
  [SMALL_STATE(5092)] = 46164,
  [SMALL_STATE(5093)] = 46190,
  [SMALL_STATE(5094)] = 46214,
  [SMALL_STATE(5095)] = 46232,
  [SMALL_STATE(5096)] = 46256,
  [SMALL_STATE(5097)] = 46282,
  [SMALL_STATE(5098)] = 46308,
  [SMALL_STATE(5099)] = 46330,
  [SMALL_STATE(5100)] = 46352,
  [SMALL_STATE(5101)] = 46374,
  [SMALL_STATE(5102)] = 46394,
  [SMALL_STATE(5103)] = 46418,
  [SMALL_STATE(5104)] = 46440,
  [SMALL_STATE(5105)] = 46464,
  [SMALL_STATE(5106)] = 46486,
  [SMALL_STATE(5107)] = 46508,
  [SMALL_STATE(5108)] = 46530,
  [SMALL_STATE(5109)] = 46552,
  [SMALL_STATE(5110)] = 46574,
  [SMALL_STATE(5111)] = 46596,
  [SMALL_STATE(5112)] = 46618,
  [SMALL_STATE(5113)] = 46640,
  [SMALL_STATE(5114)] = 46662,
  [SMALL_STATE(5115)] = 46686,
  [SMALL_STATE(5116)] = 46704,
  [SMALL_STATE(5117)] = 46726,
  [SMALL_STATE(5118)] = 46746,
  [SMALL_STATE(5119)] = 46768,
  [SMALL_STATE(5120)] = 46792,
  [SMALL_STATE(5121)] = 46814,
  [SMALL_STATE(5122)] = 46836,
  [SMALL_STATE(5123)] = 46858,
  [SMALL_STATE(5124)] = 46878,
  [SMALL_STATE(5125)] = 46900,
  [SMALL_STATE(5126)] = 46922,
  [SMALL_STATE(5127)] = 46944,
  [SMALL_STATE(5128)] = 46966,
  [SMALL_STATE(5129)] = 46988,
  [SMALL_STATE(5130)] = 47012,
  [SMALL_STATE(5131)] = 47040,
  [SMALL_STATE(5132)] = 47068,
  [SMALL_STATE(5133)] = 47096,
  [SMALL_STATE(5134)] = 47118,
  [SMALL_STATE(5135)] = 47140,
  [SMALL_STATE(5136)] = 47162,
  [SMALL_STATE(5137)] = 47184,
  [SMALL_STATE(5138)] = 47210,
  [SMALL_STATE(5139)] = 47234,
  [SMALL_STATE(5140)] = 47258,
  [SMALL_STATE(5141)] = 47280,
  [SMALL_STATE(5142)] = 47302,
  [SMALL_STATE(5143)] = 47324,
  [SMALL_STATE(5144)] = 47348,
  [SMALL_STATE(5145)] = 47370,
  [SMALL_STATE(5146)] = 47392,
  [SMALL_STATE(5147)] = 47414,
  [SMALL_STATE(5148)] = 47434,
  [SMALL_STATE(5149)] = 47456,
  [SMALL_STATE(5150)] = 47480,
  [SMALL_STATE(5151)] = 47504,
  [SMALL_STATE(5152)] = 47526,
  [SMALL_STATE(5153)] = 47544,
  [SMALL_STATE(5154)] = 47566,
  [SMALL_STATE(5155)] = 47588,
  [SMALL_STATE(5156)] = 47616,
  [SMALL_STATE(5157)] = 47644,
  [SMALL_STATE(5158)] = 47666,
  [SMALL_STATE(5159)] = 47686,
  [SMALL_STATE(5160)] = 47707,
  [SMALL_STATE(5161)] = 47730,
  [SMALL_STATE(5162)] = 47751,
  [SMALL_STATE(5163)] = 47774,
  [SMALL_STATE(5164)] = 47795,
  [SMALL_STATE(5165)] = 47812,
  [SMALL_STATE(5166)] = 47835,
  [SMALL_STATE(5167)] = 47856,
  [SMALL_STATE(5168)] = 47877,
  [SMALL_STATE(5169)] = 47898,
  [SMALL_STATE(5170)] = 47919,
  [SMALL_STATE(5171)] = 47940,
  [SMALL_STATE(5172)] = 47961,
  [SMALL_STATE(5173)] = 47982,
  [SMALL_STATE(5174)] = 48003,
  [SMALL_STATE(5175)] = 48026,
  [SMALL_STATE(5176)] = 48047,
  [SMALL_STATE(5177)] = 48064,
  [SMALL_STATE(5178)] = 48085,
  [SMALL_STATE(5179)] = 48106,
  [SMALL_STATE(5180)] = 48127,
  [SMALL_STATE(5181)] = 48150,
  [SMALL_STATE(5182)] = 48171,
  [SMALL_STATE(5183)] = 48194,
  [SMALL_STATE(5184)] = 48215,
  [SMALL_STATE(5185)] = 48236,
  [SMALL_STATE(5186)] = 48259,
  [SMALL_STATE(5187)] = 48282,
  [SMALL_STATE(5188)] = 48303,
  [SMALL_STATE(5189)] = 48324,
  [SMALL_STATE(5190)] = 48345,
  [SMALL_STATE(5191)] = 48366,
  [SMALL_STATE(5192)] = 48387,
  [SMALL_STATE(5193)] = 48408,
  [SMALL_STATE(5194)] = 48429,
  [SMALL_STATE(5195)] = 48450,
  [SMALL_STATE(5196)] = 48471,
  [SMALL_STATE(5197)] = 48492,
  [SMALL_STATE(5198)] = 48513,
  [SMALL_STATE(5199)] = 48536,
  [SMALL_STATE(5200)] = 48559,
  [SMALL_STATE(5201)] = 48582,
  [SMALL_STATE(5202)] = 48605,
  [SMALL_STATE(5203)] = 48626,
  [SMALL_STATE(5204)] = 48647,
  [SMALL_STATE(5205)] = 48668,
  [SMALL_STATE(5206)] = 48689,
  [SMALL_STATE(5207)] = 48710,
  [SMALL_STATE(5208)] = 48731,
  [SMALL_STATE(5209)] = 48752,
  [SMALL_STATE(5210)] = 48775,
  [SMALL_STATE(5211)] = 48796,
  [SMALL_STATE(5212)] = 48817,
  [SMALL_STATE(5213)] = 48840,
  [SMALL_STATE(5214)] = 48863,
  [SMALL_STATE(5215)] = 48886,
  [SMALL_STATE(5216)] = 48907,
  [SMALL_STATE(5217)] = 48930,
  [SMALL_STATE(5218)] = 48953,
  [SMALL_STATE(5219)] = 48976,
  [SMALL_STATE(5220)] = 48999,
  [SMALL_STATE(5221)] = 49022,
  [SMALL_STATE(5222)] = 49043,
  [SMALL_STATE(5223)] = 49066,
  [SMALL_STATE(5224)] = 49087,
  [SMALL_STATE(5225)] = 49108,
  [SMALL_STATE(5226)] = 49129,
  [SMALL_STATE(5227)] = 49152,
  [SMALL_STATE(5228)] = 49173,
  [SMALL_STATE(5229)] = 49194,
  [SMALL_STATE(5230)] = 49217,
  [SMALL_STATE(5231)] = 49240,
  [SMALL_STATE(5232)] = 49263,
  [SMALL_STATE(5233)] = 49286,
  [SMALL_STATE(5234)] = 49309,
  [SMALL_STATE(5235)] = 49332,
  [SMALL_STATE(5236)] = 49353,
  [SMALL_STATE(5237)] = 49374,
  [SMALL_STATE(5238)] = 49395,
  [SMALL_STATE(5239)] = 49416,
  [SMALL_STATE(5240)] = 49439,
  [SMALL_STATE(5241)] = 49460,
  [SMALL_STATE(5242)] = 49477,
  [SMALL_STATE(5243)] = 49498,
  [SMALL_STATE(5244)] = 49519,
  [SMALL_STATE(5245)] = 49540,
  [SMALL_STATE(5246)] = 49561,
  [SMALL_STATE(5247)] = 49582,
  [SMALL_STATE(5248)] = 49603,
  [SMALL_STATE(5249)] = 49624,
  [SMALL_STATE(5250)] = 49645,
  [SMALL_STATE(5251)] = 49662,
  [SMALL_STATE(5252)] = 49679,
  [SMALL_STATE(5253)] = 49702,
  [SMALL_STATE(5254)] = 49725,
  [SMALL_STATE(5255)] = 49748,
  [SMALL_STATE(5256)] = 49769,
  [SMALL_STATE(5257)] = 49790,
  [SMALL_STATE(5258)] = 49811,
  [SMALL_STATE(5259)] = 49832,
  [SMALL_STATE(5260)] = 49853,
  [SMALL_STATE(5261)] = 49874,
  [SMALL_STATE(5262)] = 49895,
  [SMALL_STATE(5263)] = 49916,
  [SMALL_STATE(5264)] = 49937,
  [SMALL_STATE(5265)] = 49958,
  [SMALL_STATE(5266)] = 49979,
  [SMALL_STATE(5267)] = 49996,
  [SMALL_STATE(5268)] = 50019,
  [SMALL_STATE(5269)] = 50042,
  [SMALL_STATE(5270)] = 50065,
  [SMALL_STATE(5271)] = 50088,
  [SMALL_STATE(5272)] = 50109,
  [SMALL_STATE(5273)] = 50130,
  [SMALL_STATE(5274)] = 50153,
  [SMALL_STATE(5275)] = 50174,
  [SMALL_STATE(5276)] = 50197,
  [SMALL_STATE(5277)] = 50220,
  [SMALL_STATE(5278)] = 50237,
  [SMALL_STATE(5279)] = 50258,
  [SMALL_STATE(5280)] = 50279,
  [SMALL_STATE(5281)] = 50300,
  [SMALL_STATE(5282)] = 50321,
  [SMALL_STATE(5283)] = 50342,
  [SMALL_STATE(5284)] = 50359,
  [SMALL_STATE(5285)] = 50380,
  [SMALL_STATE(5286)] = 50397,
  [SMALL_STATE(5287)] = 50414,
  [SMALL_STATE(5288)] = 50435,
  [SMALL_STATE(5289)] = 50458,
  [SMALL_STATE(5290)] = 50479,
  [SMALL_STATE(5291)] = 50500,
  [SMALL_STATE(5292)] = 50521,
  [SMALL_STATE(5293)] = 50544,
  [SMALL_STATE(5294)] = 50565,
  [SMALL_STATE(5295)] = 50588,
  [SMALL_STATE(5296)] = 50611,
  [SMALL_STATE(5297)] = 50632,
  [SMALL_STATE(5298)] = 50649,
  [SMALL_STATE(5299)] = 50670,
  [SMALL_STATE(5300)] = 50687,
  [SMALL_STATE(5301)] = 50708,
  [SMALL_STATE(5302)] = 50731,
  [SMALL_STATE(5303)] = 50752,
  [SMALL_STATE(5304)] = 50773,
  [SMALL_STATE(5305)] = 50794,
  [SMALL_STATE(5306)] = 50815,
  [SMALL_STATE(5307)] = 50836,
  [SMALL_STATE(5308)] = 50857,
  [SMALL_STATE(5309)] = 50878,
  [SMALL_STATE(5310)] = 50899,
  [SMALL_STATE(5311)] = 50920,
  [SMALL_STATE(5312)] = 50941,
  [SMALL_STATE(5313)] = 50958,
  [SMALL_STATE(5314)] = 50979,
  [SMALL_STATE(5315)] = 51002,
  [SMALL_STATE(5316)] = 51025,
  [SMALL_STATE(5317)] = 51042,
  [SMALL_STATE(5318)] = 51063,
  [SMALL_STATE(5319)] = 51080,
  [SMALL_STATE(5320)] = 51103,
  [SMALL_STATE(5321)] = 51120,
  [SMALL_STATE(5322)] = 51143,
  [SMALL_STATE(5323)] = 51166,
  [SMALL_STATE(5324)] = 51189,
  [SMALL_STATE(5325)] = 51212,
  [SMALL_STATE(5326)] = 51235,
  [SMALL_STATE(5327)] = 51258,
  [SMALL_STATE(5328)] = 51281,
  [SMALL_STATE(5329)] = 51304,
  [SMALL_STATE(5330)] = 51327,
  [SMALL_STATE(5331)] = 51348,
  [SMALL_STATE(5332)] = 51371,
  [SMALL_STATE(5333)] = 51391,
  [SMALL_STATE(5334)] = 51411,
  [SMALL_STATE(5335)] = 51431,
  [SMALL_STATE(5336)] = 51451,
  [SMALL_STATE(5337)] = 51471,
  [SMALL_STATE(5338)] = 51491,
  [SMALL_STATE(5339)] = 51511,
  [SMALL_STATE(5340)] = 51531,
  [SMALL_STATE(5341)] = 51551,
  [SMALL_STATE(5342)] = 51569,
  [SMALL_STATE(5343)] = 51587,
  [SMALL_STATE(5344)] = 51607,
  [SMALL_STATE(5345)] = 51627,
  [SMALL_STATE(5346)] = 51647,
  [SMALL_STATE(5347)] = 51667,
  [SMALL_STATE(5348)] = 51685,
  [SMALL_STATE(5349)] = 51701,
  [SMALL_STATE(5350)] = 51721,
  [SMALL_STATE(5351)] = 51741,
  [SMALL_STATE(5352)] = 51761,
  [SMALL_STATE(5353)] = 51781,
  [SMALL_STATE(5354)] = 51801,
  [SMALL_STATE(5355)] = 51821,
  [SMALL_STATE(5356)] = 51841,
  [SMALL_STATE(5357)] = 51861,
  [SMALL_STATE(5358)] = 51881,
  [SMALL_STATE(5359)] = 51901,
  [SMALL_STATE(5360)] = 51921,
  [SMALL_STATE(5361)] = 51939,
  [SMALL_STATE(5362)] = 51959,
  [SMALL_STATE(5363)] = 51979,
  [SMALL_STATE(5364)] = 51999,
  [SMALL_STATE(5365)] = 52019,
  [SMALL_STATE(5366)] = 52037,
  [SMALL_STATE(5367)] = 52057,
  [SMALL_STATE(5368)] = 52075,
  [SMALL_STATE(5369)] = 52095,
  [SMALL_STATE(5370)] = 52115,
  [SMALL_STATE(5371)] = 52135,
  [SMALL_STATE(5372)] = 52155,
  [SMALL_STATE(5373)] = 52175,
  [SMALL_STATE(5374)] = 52195,
  [SMALL_STATE(5375)] = 52215,
  [SMALL_STATE(5376)] = 52235,
  [SMALL_STATE(5377)] = 52255,
  [SMALL_STATE(5378)] = 52275,
  [SMALL_STATE(5379)] = 52295,
  [SMALL_STATE(5380)] = 52315,
  [SMALL_STATE(5381)] = 52335,
  [SMALL_STATE(5382)] = 52355,
  [SMALL_STATE(5383)] = 52375,
  [SMALL_STATE(5384)] = 52393,
  [SMALL_STATE(5385)] = 52413,
  [SMALL_STATE(5386)] = 52433,
  [SMALL_STATE(5387)] = 52453,
  [SMALL_STATE(5388)] = 52473,
  [SMALL_STATE(5389)] = 52493,
  [SMALL_STATE(5390)] = 52513,
  [SMALL_STATE(5391)] = 52533,
  [SMALL_STATE(5392)] = 52551,
  [SMALL_STATE(5393)] = 52571,
  [SMALL_STATE(5394)] = 52589,
  [SMALL_STATE(5395)] = 52607,
  [SMALL_STATE(5396)] = 52625,
  [SMALL_STATE(5397)] = 52645,
  [SMALL_STATE(5398)] = 52665,
  [SMALL_STATE(5399)] = 52685,
  [SMALL_STATE(5400)] = 52705,
  [SMALL_STATE(5401)] = 52725,
  [SMALL_STATE(5402)] = 52745,
  [SMALL_STATE(5403)] = 52763,
  [SMALL_STATE(5404)] = 52781,
  [SMALL_STATE(5405)] = 52801,
  [SMALL_STATE(5406)] = 52821,
  [SMALL_STATE(5407)] = 52841,
  [SMALL_STATE(5408)] = 52861,
  [SMALL_STATE(5409)] = 52881,
  [SMALL_STATE(5410)] = 52899,
  [SMALL_STATE(5411)] = 52917,
  [SMALL_STATE(5412)] = 52937,
  [SMALL_STATE(5413)] = 52957,
  [SMALL_STATE(5414)] = 52977,
  [SMALL_STATE(5415)] = 52997,
  [SMALL_STATE(5416)] = 53017,
  [SMALL_STATE(5417)] = 53037,
  [SMALL_STATE(5418)] = 53057,
  [SMALL_STATE(5419)] = 53077,
  [SMALL_STATE(5420)] = 53097,
  [SMALL_STATE(5421)] = 53117,
  [SMALL_STATE(5422)] = 53137,
  [SMALL_STATE(5423)] = 53157,
  [SMALL_STATE(5424)] = 53177,
  [SMALL_STATE(5425)] = 53197,
  [SMALL_STATE(5426)] = 53217,
  [SMALL_STATE(5427)] = 53237,
  [SMALL_STATE(5428)] = 53257,
  [SMALL_STATE(5429)] = 53277,
  [SMALL_STATE(5430)] = 53297,
  [SMALL_STATE(5431)] = 53317,
  [SMALL_STATE(5432)] = 53337,
  [SMALL_STATE(5433)] = 53357,
  [SMALL_STATE(5434)] = 53377,
  [SMALL_STATE(5435)] = 53397,
  [SMALL_STATE(5436)] = 53417,
  [SMALL_STATE(5437)] = 53437,
  [SMALL_STATE(5438)] = 53457,
  [SMALL_STATE(5439)] = 53477,
  [SMALL_STATE(5440)] = 53497,
  [SMALL_STATE(5441)] = 53517,
  [SMALL_STATE(5442)] = 53537,
  [SMALL_STATE(5443)] = 53555,
  [SMALL_STATE(5444)] = 53575,
  [SMALL_STATE(5445)] = 53595,
  [SMALL_STATE(5446)] = 53615,
  [SMALL_STATE(5447)] = 53635,
  [SMALL_STATE(5448)] = 53655,
  [SMALL_STATE(5449)] = 53675,
  [SMALL_STATE(5450)] = 53695,
  [SMALL_STATE(5451)] = 53715,
  [SMALL_STATE(5452)] = 53735,
  [SMALL_STATE(5453)] = 53755,
  [SMALL_STATE(5454)] = 53775,
  [SMALL_STATE(5455)] = 53795,
  [SMALL_STATE(5456)] = 53815,
  [SMALL_STATE(5457)] = 53835,
  [SMALL_STATE(5458)] = 53855,
  [SMALL_STATE(5459)] = 53875,
  [SMALL_STATE(5460)] = 53895,
  [SMALL_STATE(5461)] = 53915,
  [SMALL_STATE(5462)] = 53935,
  [SMALL_STATE(5463)] = 53955,
  [SMALL_STATE(5464)] = 53975,
  [SMALL_STATE(5465)] = 53993,
  [SMALL_STATE(5466)] = 54013,
  [SMALL_STATE(5467)] = 54033,
  [SMALL_STATE(5468)] = 54053,
  [SMALL_STATE(5469)] = 54073,
  [SMALL_STATE(5470)] = 54093,
  [SMALL_STATE(5471)] = 54113,
  [SMALL_STATE(5472)] = 54133,
  [SMALL_STATE(5473)] = 54153,
  [SMALL_STATE(5474)] = 54173,
  [SMALL_STATE(5475)] = 54193,
  [SMALL_STATE(5476)] = 54213,
  [SMALL_STATE(5477)] = 54233,
  [SMALL_STATE(5478)] = 54253,
  [SMALL_STATE(5479)] = 54273,
  [SMALL_STATE(5480)] = 54293,
  [SMALL_STATE(5481)] = 54311,
  [SMALL_STATE(5482)] = 54331,
  [SMALL_STATE(5483)] = 54351,
  [SMALL_STATE(5484)] = 54371,
  [SMALL_STATE(5485)] = 54391,
  [SMALL_STATE(5486)] = 54411,
  [SMALL_STATE(5487)] = 54431,
  [SMALL_STATE(5488)] = 54451,
  [SMALL_STATE(5489)] = 54471,
  [SMALL_STATE(5490)] = 54491,
  [SMALL_STATE(5491)] = 54511,
  [SMALL_STATE(5492)] = 54529,
  [SMALL_STATE(5493)] = 54549,
  [SMALL_STATE(5494)] = 54569,
  [SMALL_STATE(5495)] = 54587,
  [SMALL_STATE(5496)] = 54607,
  [SMALL_STATE(5497)] = 54627,
  [SMALL_STATE(5498)] = 54647,
  [SMALL_STATE(5499)] = 54667,
  [SMALL_STATE(5500)] = 54685,
  [SMALL_STATE(5501)] = 54705,
  [SMALL_STATE(5502)] = 54725,
  [SMALL_STATE(5503)] = 54745,
  [SMALL_STATE(5504)] = 54765,
  [SMALL_STATE(5505)] = 54785,
  [SMALL_STATE(5506)] = 54805,
  [SMALL_STATE(5507)] = 54825,
  [SMALL_STATE(5508)] = 54845,
  [SMALL_STATE(5509)] = 54865,
  [SMALL_STATE(5510)] = 54885,
  [SMALL_STATE(5511)] = 54905,
  [SMALL_STATE(5512)] = 54925,
  [SMALL_STATE(5513)] = 54945,
  [SMALL_STATE(5514)] = 54965,
  [SMALL_STATE(5515)] = 54985,
  [SMALL_STATE(5516)] = 55005,
  [SMALL_STATE(5517)] = 55023,
  [SMALL_STATE(5518)] = 55043,
  [SMALL_STATE(5519)] = 55063,
  [SMALL_STATE(5520)] = 55083,
  [SMALL_STATE(5521)] = 55103,
  [SMALL_STATE(5522)] = 55123,
  [SMALL_STATE(5523)] = 55143,
  [SMALL_STATE(5524)] = 55163,
  [SMALL_STATE(5525)] = 55183,
  [SMALL_STATE(5526)] = 55203,
  [SMALL_STATE(5527)] = 55223,
  [SMALL_STATE(5528)] = 55243,
  [SMALL_STATE(5529)] = 55263,
  [SMALL_STATE(5530)] = 55283,
  [SMALL_STATE(5531)] = 55303,
  [SMALL_STATE(5532)] = 55323,
  [SMALL_STATE(5533)] = 55343,
  [SMALL_STATE(5534)] = 55363,
  [SMALL_STATE(5535)] = 55383,
  [SMALL_STATE(5536)] = 55401,
  [SMALL_STATE(5537)] = 55421,
  [SMALL_STATE(5538)] = 55441,
  [SMALL_STATE(5539)] = 55457,
  [SMALL_STATE(5540)] = 55477,
  [SMALL_STATE(5541)] = 55497,
  [SMALL_STATE(5542)] = 55517,
  [SMALL_STATE(5543)] = 55537,
  [SMALL_STATE(5544)] = 55557,
  [SMALL_STATE(5545)] = 55577,
  [SMALL_STATE(5546)] = 55597,
  [SMALL_STATE(5547)] = 55617,
  [SMALL_STATE(5548)] = 55637,
  [SMALL_STATE(5549)] = 55657,
  [SMALL_STATE(5550)] = 55677,
  [SMALL_STATE(5551)] = 55697,
  [SMALL_STATE(5552)] = 55717,
  [SMALL_STATE(5553)] = 55737,
  [SMALL_STATE(5554)] = 55757,
  [SMALL_STATE(5555)] = 55777,
  [SMALL_STATE(5556)] = 55795,
  [SMALL_STATE(5557)] = 55815,
  [SMALL_STATE(5558)] = 55835,
  [SMALL_STATE(5559)] = 55855,
  [SMALL_STATE(5560)] = 55875,
  [SMALL_STATE(5561)] = 55895,
  [SMALL_STATE(5562)] = 55913,
  [SMALL_STATE(5563)] = 55933,
  [SMALL_STATE(5564)] = 55953,
  [SMALL_STATE(5565)] = 55971,
  [SMALL_STATE(5566)] = 55989,
  [SMALL_STATE(5567)] = 56007,
  [SMALL_STATE(5568)] = 56025,
  [SMALL_STATE(5569)] = 56045,
  [SMALL_STATE(5570)] = 56065,
  [SMALL_STATE(5571)] = 56085,
  [SMALL_STATE(5572)] = 56105,
  [SMALL_STATE(5573)] = 56125,
  [SMALL_STATE(5574)] = 56145,
  [SMALL_STATE(5575)] = 56163,
  [SMALL_STATE(5576)] = 56183,
  [SMALL_STATE(5577)] = 56203,
  [SMALL_STATE(5578)] = 56218,
  [SMALL_STATE(5579)] = 56235,
  [SMALL_STATE(5580)] = 56250,
  [SMALL_STATE(5581)] = 56267,
  [SMALL_STATE(5582)] = 56284,
  [SMALL_STATE(5583)] = 56301,
  [SMALL_STATE(5584)] = 56318,
  [SMALL_STATE(5585)] = 56335,
  [SMALL_STATE(5586)] = 56352,
  [SMALL_STATE(5587)] = 56369,
  [SMALL_STATE(5588)] = 56384,
  [SMALL_STATE(5589)] = 56401,
  [SMALL_STATE(5590)] = 56416,
  [SMALL_STATE(5591)] = 56431,
  [SMALL_STATE(5592)] = 56446,
  [SMALL_STATE(5593)] = 56461,
  [SMALL_STATE(5594)] = 56476,
  [SMALL_STATE(5595)] = 56493,
  [SMALL_STATE(5596)] = 56508,
  [SMALL_STATE(5597)] = 56525,
  [SMALL_STATE(5598)] = 56540,
  [SMALL_STATE(5599)] = 56557,
  [SMALL_STATE(5600)] = 56574,
  [SMALL_STATE(5601)] = 56591,
  [SMALL_STATE(5602)] = 56606,
  [SMALL_STATE(5603)] = 56621,
  [SMALL_STATE(5604)] = 56638,
  [SMALL_STATE(5605)] = 56653,
  [SMALL_STATE(5606)] = 56668,
  [SMALL_STATE(5607)] = 56685,
  [SMALL_STATE(5608)] = 56702,
  [SMALL_STATE(5609)] = 56717,
  [SMALL_STATE(5610)] = 56734,
  [SMALL_STATE(5611)] = 56751,
  [SMALL_STATE(5612)] = 56768,
  [SMALL_STATE(5613)] = 56785,
  [SMALL_STATE(5614)] = 56802,
  [SMALL_STATE(5615)] = 56819,
  [SMALL_STATE(5616)] = 56836,
  [SMALL_STATE(5617)] = 56851,
  [SMALL_STATE(5618)] = 56866,
  [SMALL_STATE(5619)] = 56883,
  [SMALL_STATE(5620)] = 56900,
  [SMALL_STATE(5621)] = 56917,
  [SMALL_STATE(5622)] = 56934,
  [SMALL_STATE(5623)] = 56951,
  [SMALL_STATE(5624)] = 56968,
  [SMALL_STATE(5625)] = 56983,
  [SMALL_STATE(5626)] = 56998,
  [SMALL_STATE(5627)] = 57013,
  [SMALL_STATE(5628)] = 57028,
  [SMALL_STATE(5629)] = 57043,
  [SMALL_STATE(5630)] = 57058,
  [SMALL_STATE(5631)] = 57073,
  [SMALL_STATE(5632)] = 57088,
  [SMALL_STATE(5633)] = 57103,
  [SMALL_STATE(5634)] = 57120,
  [SMALL_STATE(5635)] = 57137,
  [SMALL_STATE(5636)] = 57154,
  [SMALL_STATE(5637)] = 57171,
  [SMALL_STATE(5638)] = 57186,
  [SMALL_STATE(5639)] = 57203,
  [SMALL_STATE(5640)] = 57220,
  [SMALL_STATE(5641)] = 57237,
  [SMALL_STATE(5642)] = 57254,
  [SMALL_STATE(5643)] = 57269,
  [SMALL_STATE(5644)] = 57286,
  [SMALL_STATE(5645)] = 57303,
  [SMALL_STATE(5646)] = 57318,
  [SMALL_STATE(5647)] = 57333,
  [SMALL_STATE(5648)] = 57350,
  [SMALL_STATE(5649)] = 57367,
  [SMALL_STATE(5650)] = 57384,
  [SMALL_STATE(5651)] = 57401,
  [SMALL_STATE(5652)] = 57418,
  [SMALL_STATE(5653)] = 57433,
  [SMALL_STATE(5654)] = 57450,
  [SMALL_STATE(5655)] = 57467,
  [SMALL_STATE(5656)] = 57482,
  [SMALL_STATE(5657)] = 57497,
  [SMALL_STATE(5658)] = 57512,
  [SMALL_STATE(5659)] = 57527,
  [SMALL_STATE(5660)] = 57542,
  [SMALL_STATE(5661)] = 57559,
  [SMALL_STATE(5662)] = 57576,
  [SMALL_STATE(5663)] = 57591,
  [SMALL_STATE(5664)] = 57606,
  [SMALL_STATE(5665)] = 57621,
  [SMALL_STATE(5666)] = 57636,
  [SMALL_STATE(5667)] = 57653,
  [SMALL_STATE(5668)] = 57668,
  [SMALL_STATE(5669)] = 57685,
  [SMALL_STATE(5670)] = 57702,
  [SMALL_STATE(5671)] = 57717,
  [SMALL_STATE(5672)] = 57734,
  [SMALL_STATE(5673)] = 57751,
  [SMALL_STATE(5674)] = 57766,
  [SMALL_STATE(5675)] = 57783,
  [SMALL_STATE(5676)] = 57798,
  [SMALL_STATE(5677)] = 57813,
  [SMALL_STATE(5678)] = 57830,
  [SMALL_STATE(5679)] = 57845,
  [SMALL_STATE(5680)] = 57860,
  [SMALL_STATE(5681)] = 57877,
  [SMALL_STATE(5682)] = 57894,
  [SMALL_STATE(5683)] = 57911,
  [SMALL_STATE(5684)] = 57928,
  [SMALL_STATE(5685)] = 57945,
  [SMALL_STATE(5686)] = 57962,
  [SMALL_STATE(5687)] = 57977,
  [SMALL_STATE(5688)] = 57994,
  [SMALL_STATE(5689)] = 58011,
  [SMALL_STATE(5690)] = 58026,
  [SMALL_STATE(5691)] = 58043,
  [SMALL_STATE(5692)] = 58058,
  [SMALL_STATE(5693)] = 58073,
  [SMALL_STATE(5694)] = 58088,
  [SMALL_STATE(5695)] = 58105,
  [SMALL_STATE(5696)] = 58122,
  [SMALL_STATE(5697)] = 58139,
  [SMALL_STATE(5698)] = 58154,
  [SMALL_STATE(5699)] = 58169,
  [SMALL_STATE(5700)] = 58184,
  [SMALL_STATE(5701)] = 58199,
  [SMALL_STATE(5702)] = 58216,
  [SMALL_STATE(5703)] = 58231,
  [SMALL_STATE(5704)] = 58248,
  [SMALL_STATE(5705)] = 58265,
  [SMALL_STATE(5706)] = 58282,
  [SMALL_STATE(5707)] = 58297,
  [SMALL_STATE(5708)] = 58314,
  [SMALL_STATE(5709)] = 58331,
  [SMALL_STATE(5710)] = 58348,
  [SMALL_STATE(5711)] = 58365,
  [SMALL_STATE(5712)] = 58382,
  [SMALL_STATE(5713)] = 58399,
  [SMALL_STATE(5714)] = 58416,
  [SMALL_STATE(5715)] = 58433,
  [SMALL_STATE(5716)] = 58450,
  [SMALL_STATE(5717)] = 58467,
  [SMALL_STATE(5718)] = 58484,
  [SMALL_STATE(5719)] = 58501,
  [SMALL_STATE(5720)] = 58518,
  [SMALL_STATE(5721)] = 58533,
  [SMALL_STATE(5722)] = 58550,
  [SMALL_STATE(5723)] = 58567,
  [SMALL_STATE(5724)] = 58584,
  [SMALL_STATE(5725)] = 58601,
  [SMALL_STATE(5726)] = 58618,
  [SMALL_STATE(5727)] = 58632,
  [SMALL_STATE(5728)] = 58646,
  [SMALL_STATE(5729)] = 58660,
  [SMALL_STATE(5730)] = 58676,
  [SMALL_STATE(5731)] = 58690,
  [SMALL_STATE(5732)] = 58704,
  [SMALL_STATE(5733)] = 58718,
  [SMALL_STATE(5734)] = 58732,
  [SMALL_STATE(5735)] = 58746,
  [SMALL_STATE(5736)] = 58760,
  [SMALL_STATE(5737)] = 58774,
  [SMALL_STATE(5738)] = 58788,
  [SMALL_STATE(5739)] = 58802,
  [SMALL_STATE(5740)] = 58816,
  [SMALL_STATE(5741)] = 58830,
  [SMALL_STATE(5742)] = 58844,
  [SMALL_STATE(5743)] = 58858,
  [SMALL_STATE(5744)] = 58872,
  [SMALL_STATE(5745)] = 58886,
  [SMALL_STATE(5746)] = 58900,
  [SMALL_STATE(5747)] = 58914,
  [SMALL_STATE(5748)] = 58928,
  [SMALL_STATE(5749)] = 58942,
  [SMALL_STATE(5750)] = 58956,
  [SMALL_STATE(5751)] = 58970,
  [SMALL_STATE(5752)] = 58984,
  [SMALL_STATE(5753)] = 58998,
  [SMALL_STATE(5754)] = 59014,
  [SMALL_STATE(5755)] = 59028,
  [SMALL_STATE(5756)] = 59042,
  [SMALL_STATE(5757)] = 59056,
  [SMALL_STATE(5758)] = 59070,
  [SMALL_STATE(5759)] = 59084,
  [SMALL_STATE(5760)] = 59098,
  [SMALL_STATE(5761)] = 59112,
  [SMALL_STATE(5762)] = 59126,
  [SMALL_STATE(5763)] = 59140,
  [SMALL_STATE(5764)] = 59154,
  [SMALL_STATE(5765)] = 59168,
  [SMALL_STATE(5766)] = 59182,
  [SMALL_STATE(5767)] = 59198,
  [SMALL_STATE(5768)] = 59212,
  [SMALL_STATE(5769)] = 59226,
  [SMALL_STATE(5770)] = 59240,
  [SMALL_STATE(5771)] = 59254,
  [SMALL_STATE(5772)] = 59268,
  [SMALL_STATE(5773)] = 59282,
  [SMALL_STATE(5774)] = 59296,
  [SMALL_STATE(5775)] = 59310,
  [SMALL_STATE(5776)] = 59324,
  [SMALL_STATE(5777)] = 59338,
  [SMALL_STATE(5778)] = 59352,
  [SMALL_STATE(5779)] = 59366,
  [SMALL_STATE(5780)] = 59380,
  [SMALL_STATE(5781)] = 59394,
  [SMALL_STATE(5782)] = 59408,
  [SMALL_STATE(5783)] = 59422,
  [SMALL_STATE(5784)] = 59436,
  [SMALL_STATE(5785)] = 59450,
  [SMALL_STATE(5786)] = 59464,
  [SMALL_STATE(5787)] = 59478,
  [SMALL_STATE(5788)] = 59492,
  [SMALL_STATE(5789)] = 59506,
  [SMALL_STATE(5790)] = 59520,
  [SMALL_STATE(5791)] = 59534,
  [SMALL_STATE(5792)] = 59548,
  [SMALL_STATE(5793)] = 59562,
  [SMALL_STATE(5794)] = 59576,
  [SMALL_STATE(5795)] = 59590,
  [SMALL_STATE(5796)] = 59604,
  [SMALL_STATE(5797)] = 59618,
  [SMALL_STATE(5798)] = 59632,
  [SMALL_STATE(5799)] = 59646,
  [SMALL_STATE(5800)] = 59660,
  [SMALL_STATE(5801)] = 59674,
  [SMALL_STATE(5802)] = 59688,
  [SMALL_STATE(5803)] = 59702,
  [SMALL_STATE(5804)] = 59716,
  [SMALL_STATE(5805)] = 59730,
  [SMALL_STATE(5806)] = 59744,
  [SMALL_STATE(5807)] = 59758,
  [SMALL_STATE(5808)] = 59772,
  [SMALL_STATE(5809)] = 59786,
  [SMALL_STATE(5810)] = 59800,
  [SMALL_STATE(5811)] = 59814,
  [SMALL_STATE(5812)] = 59828,
  [SMALL_STATE(5813)] = 59842,
  [SMALL_STATE(5814)] = 59856,
  [SMALL_STATE(5815)] = 59870,
  [SMALL_STATE(5816)] = 59884,
  [SMALL_STATE(5817)] = 59898,
  [SMALL_STATE(5818)] = 59912,
  [SMALL_STATE(5819)] = 59926,
  [SMALL_STATE(5820)] = 59940,
  [SMALL_STATE(5821)] = 59954,
  [SMALL_STATE(5822)] = 59968,
  [SMALL_STATE(5823)] = 59982,
  [SMALL_STATE(5824)] = 59996,
  [SMALL_STATE(5825)] = 60012,
  [SMALL_STATE(5826)] = 60026,
  [SMALL_STATE(5827)] = 60040,
  [SMALL_STATE(5828)] = 60054,
  [SMALL_STATE(5829)] = 60068,
  [SMALL_STATE(5830)] = 60082,
  [SMALL_STATE(5831)] = 60096,
  [SMALL_STATE(5832)] = 60110,
  [SMALL_STATE(5833)] = 60124,
  [SMALL_STATE(5834)] = 60138,
  [SMALL_STATE(5835)] = 60152,
  [SMALL_STATE(5836)] = 60166,
  [SMALL_STATE(5837)] = 60180,
  [SMALL_STATE(5838)] = 60194,
  [SMALL_STATE(5839)] = 60208,
  [SMALL_STATE(5840)] = 60222,
  [SMALL_STATE(5841)] = 60236,
  [SMALL_STATE(5842)] = 60250,
  [SMALL_STATE(5843)] = 60264,
  [SMALL_STATE(5844)] = 60278,
  [SMALL_STATE(5845)] = 60292,
  [SMALL_STATE(5846)] = 60306,
  [SMALL_STATE(5847)] = 60320,
  [SMALL_STATE(5848)] = 60334,
  [SMALL_STATE(5849)] = 60348,
  [SMALL_STATE(5850)] = 60362,
  [SMALL_STATE(5851)] = 60376,
  [SMALL_STATE(5852)] = 60390,
  [SMALL_STATE(5853)] = 60404,
  [SMALL_STATE(5854)] = 60418,
  [SMALL_STATE(5855)] = 60432,
  [SMALL_STATE(5856)] = 60446,
  [SMALL_STATE(5857)] = 60460,
  [SMALL_STATE(5858)] = 60474,
  [SMALL_STATE(5859)] = 60488,
  [SMALL_STATE(5860)] = 60502,
  [SMALL_STATE(5861)] = 60516,
  [SMALL_STATE(5862)] = 60530,
  [SMALL_STATE(5863)] = 60544,
  [SMALL_STATE(5864)] = 60558,
  [SMALL_STATE(5865)] = 60572,
  [SMALL_STATE(5866)] = 60586,
  [SMALL_STATE(5867)] = 60600,
  [SMALL_STATE(5868)] = 60614,
  [SMALL_STATE(5869)] = 60628,
  [SMALL_STATE(5870)] = 60642,
  [SMALL_STATE(5871)] = 60656,
  [SMALL_STATE(5872)] = 60670,
  [SMALL_STATE(5873)] = 60684,
  [SMALL_STATE(5874)] = 60698,
  [SMALL_STATE(5875)] = 60714,
  [SMALL_STATE(5876)] = 60728,
  [SMALL_STATE(5877)] = 60742,
  [SMALL_STATE(5878)] = 60756,
  [SMALL_STATE(5879)] = 60770,
  [SMALL_STATE(5880)] = 60784,
  [SMALL_STATE(5881)] = 60798,
  [SMALL_STATE(5882)] = 60812,
  [SMALL_STATE(5883)] = 60826,
  [SMALL_STATE(5884)] = 60840,
  [SMALL_STATE(5885)] = 60854,
  [SMALL_STATE(5886)] = 60868,
  [SMALL_STATE(5887)] = 60882,
  [SMALL_STATE(5888)] = 60896,
  [SMALL_STATE(5889)] = 60910,
  [SMALL_STATE(5890)] = 60924,
  [SMALL_STATE(5891)] = 60938,
  [SMALL_STATE(5892)] = 60952,
  [SMALL_STATE(5893)] = 60966,
  [SMALL_STATE(5894)] = 60980,
  [SMALL_STATE(5895)] = 60994,
  [SMALL_STATE(5896)] = 61008,
  [SMALL_STATE(5897)] = 61022,
  [SMALL_STATE(5898)] = 61036,
  [SMALL_STATE(5899)] = 61050,
  [SMALL_STATE(5900)] = 61064,
  [SMALL_STATE(5901)] = 61078,
  [SMALL_STATE(5902)] = 61092,
  [SMALL_STATE(5903)] = 61106,
  [SMALL_STATE(5904)] = 61120,
  [SMALL_STATE(5905)] = 61134,
  [SMALL_STATE(5906)] = 61148,
  [SMALL_STATE(5907)] = 61162,
  [SMALL_STATE(5908)] = 61176,
  [SMALL_STATE(5909)] = 61190,
  [SMALL_STATE(5910)] = 61204,
  [SMALL_STATE(5911)] = 61218,
  [SMALL_STATE(5912)] = 61232,
  [SMALL_STATE(5913)] = 61246,
  [SMALL_STATE(5914)] = 61260,
  [SMALL_STATE(5915)] = 61274,
  [SMALL_STATE(5916)] = 61288,
  [SMALL_STATE(5917)] = 61302,
  [SMALL_STATE(5918)] = 61316,
  [SMALL_STATE(5919)] = 61330,
  [SMALL_STATE(5920)] = 61344,
  [SMALL_STATE(5921)] = 61358,
  [SMALL_STATE(5922)] = 61372,
  [SMALL_STATE(5923)] = 61386,
  [SMALL_STATE(5924)] = 61400,
  [SMALL_STATE(5925)] = 61414,
  [SMALL_STATE(5926)] = 61428,
  [SMALL_STATE(5927)] = 61442,
  [SMALL_STATE(5928)] = 61456,
  [SMALL_STATE(5929)] = 61470,
  [SMALL_STATE(5930)] = 61484,
  [SMALL_STATE(5931)] = 61498,
  [SMALL_STATE(5932)] = 61512,
  [SMALL_STATE(5933)] = 61526,
  [SMALL_STATE(5934)] = 61540,
  [SMALL_STATE(5935)] = 61554,
  [SMALL_STATE(5936)] = 61568,
  [SMALL_STATE(5937)] = 61582,
  [SMALL_STATE(5938)] = 61596,
  [SMALL_STATE(5939)] = 61610,
  [SMALL_STATE(5940)] = 61624,
  [SMALL_STATE(5941)] = 61638,
  [SMALL_STATE(5942)] = 61652,
  [SMALL_STATE(5943)] = 61666,
  [SMALL_STATE(5944)] = 61680,
  [SMALL_STATE(5945)] = 61694,
  [SMALL_STATE(5946)] = 61708,
  [SMALL_STATE(5947)] = 61722,
  [SMALL_STATE(5948)] = 61736,
  [SMALL_STATE(5949)] = 61750,
  [SMALL_STATE(5950)] = 61764,
  [SMALL_STATE(5951)] = 61778,
  [SMALL_STATE(5952)] = 61792,
  [SMALL_STATE(5953)] = 61806,
  [SMALL_STATE(5954)] = 61820,
  [SMALL_STATE(5955)] = 61834,
  [SMALL_STATE(5956)] = 61848,
  [SMALL_STATE(5957)] = 61862,
  [SMALL_STATE(5958)] = 61876,
  [SMALL_STATE(5959)] = 61890,
  [SMALL_STATE(5960)] = 61904,
  [SMALL_STATE(5961)] = 61918,
  [SMALL_STATE(5962)] = 61932,
  [SMALL_STATE(5963)] = 61946,
  [SMALL_STATE(5964)] = 61960,
  [SMALL_STATE(5965)] = 61974,
  [SMALL_STATE(5966)] = 61988,
  [SMALL_STATE(5967)] = 62002,
  [SMALL_STATE(5968)] = 62016,
  [SMALL_STATE(5969)] = 62030,
  [SMALL_STATE(5970)] = 62044,
  [SMALL_STATE(5971)] = 62058,
  [SMALL_STATE(5972)] = 62072,
  [SMALL_STATE(5973)] = 62086,
  [SMALL_STATE(5974)] = 62100,
  [SMALL_STATE(5975)] = 62114,
  [SMALL_STATE(5976)] = 62128,
  [SMALL_STATE(5977)] = 62142,
  [SMALL_STATE(5978)] = 62156,
  [SMALL_STATE(5979)] = 62170,
  [SMALL_STATE(5980)] = 62184,
  [SMALL_STATE(5981)] = 62198,
  [SMALL_STATE(5982)] = 62212,
  [SMALL_STATE(5983)] = 62226,
  [SMALL_STATE(5984)] = 62240,
  [SMALL_STATE(5985)] = 62254,
  [SMALL_STATE(5986)] = 62268,
  [SMALL_STATE(5987)] = 62282,
  [SMALL_STATE(5988)] = 62296,
  [SMALL_STATE(5989)] = 62310,
  [SMALL_STATE(5990)] = 62324,
  [SMALL_STATE(5991)] = 62338,
  [SMALL_STATE(5992)] = 62352,
  [SMALL_STATE(5993)] = 62366,
  [SMALL_STATE(5994)] = 62380,
  [SMALL_STATE(5995)] = 62394,
  [SMALL_STATE(5996)] = 62408,
  [SMALL_STATE(5997)] = 62424,
  [SMALL_STATE(5998)] = 62438,
  [SMALL_STATE(5999)] = 62452,
  [SMALL_STATE(6000)] = 62466,
  [SMALL_STATE(6001)] = 62480,
  [SMALL_STATE(6002)] = 62494,
  [SMALL_STATE(6003)] = 62508,
  [SMALL_STATE(6004)] = 62522,
  [SMALL_STATE(6005)] = 62536,
  [SMALL_STATE(6006)] = 62550,
  [SMALL_STATE(6007)] = 62564,
  [SMALL_STATE(6008)] = 62578,
  [SMALL_STATE(6009)] = 62592,
  [SMALL_STATE(6010)] = 62606,
  [SMALL_STATE(6011)] = 62620,
  [SMALL_STATE(6012)] = 62634,
  [SMALL_STATE(6013)] = 62648,
  [SMALL_STATE(6014)] = 62662,
  [SMALL_STATE(6015)] = 62676,
  [SMALL_STATE(6016)] = 62690,
  [SMALL_STATE(6017)] = 62704,
  [SMALL_STATE(6018)] = 62718,
  [SMALL_STATE(6019)] = 62732,
  [SMALL_STATE(6020)] = 62746,
  [SMALL_STATE(6021)] = 62760,
  [SMALL_STATE(6022)] = 62774,
  [SMALL_STATE(6023)] = 62788,
  [SMALL_STATE(6024)] = 62802,
  [SMALL_STATE(6025)] = 62816,
  [SMALL_STATE(6026)] = 62830,
  [SMALL_STATE(6027)] = 62844,
  [SMALL_STATE(6028)] = 62858,
  [SMALL_STATE(6029)] = 62872,
  [SMALL_STATE(6030)] = 62886,
  [SMALL_STATE(6031)] = 62900,
  [SMALL_STATE(6032)] = 62914,
  [SMALL_STATE(6033)] = 62928,
  [SMALL_STATE(6034)] = 62942,
  [SMALL_STATE(6035)] = 62956,
  [SMALL_STATE(6036)] = 62970,
  [SMALL_STATE(6037)] = 62984,
  [SMALL_STATE(6038)] = 62998,
  [SMALL_STATE(6039)] = 63012,
  [SMALL_STATE(6040)] = 63026,
  [SMALL_STATE(6041)] = 63040,
  [SMALL_STATE(6042)] = 63054,
  [SMALL_STATE(6043)] = 63068,
  [SMALL_STATE(6044)] = 63082,
  [SMALL_STATE(6045)] = 63096,
  [SMALL_STATE(6046)] = 63110,
  [SMALL_STATE(6047)] = 63124,
  [SMALL_STATE(6048)] = 63128,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [9] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1149),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1617),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1471),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5367),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5516),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5491),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4865),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4181),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4302),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4022),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5285),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4478),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4450),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1804),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4544),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5558),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5575),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4797),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5816),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1781),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1126),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1130),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1046),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(807),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4171),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(649),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(803),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(451),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(620),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4284),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5040),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5050),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4771),
  [93] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4603),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5541),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5824),
  [103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5779),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5690),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5613),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5619),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5162),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5220),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5275),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5276),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5315),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5759),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5736),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1150),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1923),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5283),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4529),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4551),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4469),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4909),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3985),
  [143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5674),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5779),
  [147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1827),
  [149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1151),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1716),
  [155] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loopex_expression, 1, 0, 2),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4629),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4637),
  [161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loopex_expression, 1, 0, 2),
  [163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0),
  [165] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5816),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1781),
  [171] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1126),
  [174] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1130),
  [177] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(924),
  [180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_expression, 1, 0, 0),
  [182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_undef_expression, 1, 0, 0),
  [184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_undef_expression, 1, 0, 0),
  [186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func1op_call_expression, 1, 0, 1),
  [188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(423),
  [190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func1op_call_expression, 1, 0, 1),
  [192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0),
  [194] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5816),
  [197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1781),
  [200] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1126),
  [203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1130),
  [206] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(924),
  [209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0),
  [211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1251),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4510),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1090),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1117),
  [225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1122),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(786),
  [231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(640),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(10),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1257),
  [245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1090),
  [250] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1091),
  [253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1108),
  [256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1090),
  [259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1091),
  [262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1108),
  [265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4247),
  [269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4277),
  [271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4522),
  [273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4642),
  [275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4646),
  [277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2246),
  [281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5982),
  [286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(2200),
  [289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1054),
  [292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1055),
  [295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1054),
  [297] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1056),
  [300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1082),
  [302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(788),
  [304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(789),
  [306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1116),
  [308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1128),
  [310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4279),
  [312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(773),
  [316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(438),
  [318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(12),
  [322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4139),
  [326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(13),
  [328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(22),
  [330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5711),
  [332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5712),
  [334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5716),
  [336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5180),
  [338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5182),
  [340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5185),
  [342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5301),
  [344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5186),
  [346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2303),
  [348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2310),
  [350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4753),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5582),
  [356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2362),
  [358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4200),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4188),
  [368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4491),
  [370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4610),
  [372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4612),
  [374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(419),
  [376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2471),
  [378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(359),
  [380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5771),
  [383] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(2388),
  [386] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1021),
  [389] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1022),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1021),
  [394] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1025),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(781),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(782),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1034),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1038),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4191),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(657),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(546),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(660),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4193),
  [423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(18),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(27),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5634),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5635),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5636),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5331),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5252),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5253),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5324),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5254),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2649),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2650),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3987),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5594),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2648),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1351),
  [459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4185),
  [461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4144),
  [463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4548),
  [465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4605),
  [467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4652),
  [469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(415),
  [471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2382),
  [473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5840),
  [478] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(2416),
  [481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(869),
  [484] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(893),
  [487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(869),
  [489] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(894),
  [492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(755),
  [498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4148),
  [504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(701),
  [506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(491),
  [510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(705),
  [512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(15),
  [514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(28),
  [516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4154),
  [518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(17),
  [520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(23),
  [522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5620),
  [524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5621),
  [526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5622),
  [528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5231),
  [530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5232),
  [532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5233),
  [534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5323),
  [536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5234),
  [538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3144),
  [540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3145),
  [542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4917),
  [544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5651),
  [548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3143),
  [550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1342),
  [554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4213),
  [556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4215),
  [558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4520),
  [560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4696),
  [562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4699),
  [564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2494),
  [568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5848),
  [573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(2504),
  [576] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1035),
  [579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1036),
  [582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1035),
  [584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1037),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(784),
  [593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1042),
  [595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1044),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4216),
  [599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(447),
  [605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(16),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(26),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4217),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(19),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(25),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5633),
  [619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5638),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5639),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5325),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5327),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5328),
  [629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5314),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5329),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2690),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2691),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4747),
  [639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3990),
  [641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5578),
  [643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2662),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1318),
  [649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1347),
  [653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5982),
  [656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(2200),
  [659] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1054),
  [662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1055),
  [665] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1056),
  [668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [672] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5848),
  [675] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(2504),
  [678] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1035),
  [681] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1036),
  [684] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1037),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(389),
  [691] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5840),
  [694] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(2416),
  [697] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(869),
  [700] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(893),
  [703] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(894),
  [706] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5771),
  [709] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(2388),
  [712] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1021),
  [715] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1022),
  [718] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1025),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4161),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4233),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4516),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4694),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4695),
  [733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3336),
  [737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(870),
  [741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(722),
  [745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(907),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4240),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(612),
  [755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(456),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(614),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(33),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(90),
  [765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4243),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(84),
  [771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5610),
  [773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5611),
  [775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5612),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5216),
  [779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5217),
  [781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5218),
  [783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5321),
  [785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5219),
  [787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3530),
  [791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3531),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3984),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5615),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3529),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [803] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5817),
  [806] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(3337),
  [809] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(870),
  [812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(889),
  [815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(890),
  [818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1374),
  [822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4294),
  [824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4298),
  [826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4500),
  [828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4599),
  [830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4609),
  [832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(399),
  [834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3273),
  [836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [838] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5788),
  [841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(3274),
  [844] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1095),
  [847] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1098),
  [850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [852] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(1099),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(795),
  [859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(796),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4189),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(35),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(91),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4262),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(36),
  [883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(83),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5584),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5585),
  [889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5586),
  [891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5198),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5199),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5200),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5319),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5201),
  [901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3419),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3420),
  [905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3983),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5641),
  [911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1398),
  [917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1419),
  [919] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(4143),
  [922] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(4274),
  [925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4498),
  [927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4664),
  [929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4673),
  [931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(405),
  [933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2389),
  [935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [937] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(5931),
  [940] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(2391),
  [943] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(895),
  [946] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(896),
  [949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(895),
  [951] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(897),
  [954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(751),
  [958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(901),
  [962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4276),
  [966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(691),
  [968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(441),
  [972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(630),
  [974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(82),
  [978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(4278),
  [981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(38),
  [983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(93),
  [985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5648),
  [987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5649),
  [989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5650),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5267),
  [993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5268),
  [995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5269),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5326),
  [999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5270),
  [1001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2834),
  [1003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2835),
  [1005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4773),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [1009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5580),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2828),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [1015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [1017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1188),
  [1019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [1021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1616),
  [1023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5565),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5566),
  [1027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5567),
  [1029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4755),
  [1031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4136),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4046),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4397),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4398),
  [1039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5449),
  [1041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5450),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(868),
  [1045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5915),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4970),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4702),
  [1051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5495),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5980),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [1057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5915),
  [1059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1018),
  [1061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [1063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [1065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [1067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(922),
  [1069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1007),
  [1071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(712),
  [1073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [1075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [1077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [1079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(40),
  [1081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(41),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [1089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1749),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5978),
  [1095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [1097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5978),
  [1099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [1101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1149),
  [1104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1617),
  [1107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1471),
  [1110] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5367),
  [1113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5516),
  [1116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5491),
  [1119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4865),
  [1122] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4181),
  [1125] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4302),
  [1128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4022),
  [1131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5285),
  [1134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4478),
  [1137] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4450),
  [1140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(401),
  [1143] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1804),
  [1146] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4544),
  [1149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5558),
  [1152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5575),
  [1155] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4797),
  [1158] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(335),
  [1161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5816),
  [1164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1781),
  [1167] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1126),
  [1170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1130),
  [1173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(924),
  [1176] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(868),
  [1179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(756),
  [1182] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(764),
  [1185] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1046),
  [1188] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(807),
  [1191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4171),
  [1194] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(649),
  [1197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(803),
  [1200] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(451),
  [1203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(620),
  [1206] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(3),
  [1209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4),
  [1212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4284),
  [1215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5040),
  [1218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5050),
  [1221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4771),
  [1224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4603),
  [1227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2),
  [1230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5541),
  [1233] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5),
  [1236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [1238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5690),
  [1241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5613),
  [1244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5619),
  [1247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5162),
  [1250] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5220),
  [1253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5275),
  [1256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5276),
  [1259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5315),
  [1262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5759),
  [1265] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5736),
  [1268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1150),
  [1271] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1922),
  [1274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1923),
  [1277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5283),
  [1280] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4529),
  [1283] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4551),
  [1286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4469),
  [1289] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4909),
  [1292] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(3985),
  [1295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5674),
  [1298] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1827),
  [1301] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5),
  [1304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [1306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4219),
  [1308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5793),
  [1310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [1312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5793),
  [1314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1158),
  [1316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5176),
  [1318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5833),
  [1320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1198),
  [1322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5833),
  [1324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [1326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4108),
  [1328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5761),
  [1330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [1332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5761),
  [1334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [1336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2025),
  [1338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5972),
  [1340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1200),
  [1342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5972),
  [1344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1163),
  [1346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2156),
  [1348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5737),
  [1350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1201),
  [1352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5737),
  [1354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1165),
  [1356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2329),
  [1358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5780),
  [1360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1202),
  [1362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5780),
  [1364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [1366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1153),
  [1368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5823),
  [1370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1203),
  [1372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5823),
  [1374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1167),
  [1376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1133),
  [1378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5844),
  [1380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1204),
  [1382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5844),
  [1384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [1386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4358),
  [1388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5854),
  [1390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [1392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5854),
  [1394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [1396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4293),
  [1398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5859),
  [1400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1207),
  [1402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5859),
  [1404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [1406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1297),
  [1408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5864),
  [1410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1208),
  [1412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5864),
  [1414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [1416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1265),
  [1418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5868),
  [1420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1209),
  [1422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5868),
  [1424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1173),
  [1426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [1428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5871),
  [1430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1210),
  [1432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5871),
  [1434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [1436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1309),
  [1438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5873),
  [1440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1196),
  [1442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5873),
  [1444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1175),
  [1446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1379),
  [1448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5876),
  [1450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [1452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5876),
  [1454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [1456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [1458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5879),
  [1460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [1462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5879),
  [1464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [1466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3091),
  [1468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5882),
  [1470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1214),
  [1472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [1474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1178),
  [1476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4038),
  [1478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5885),
  [1480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [1482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5885),
  [1484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1179),
  [1486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2130),
  [1488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5888),
  [1490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [1492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [1494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1180),
  [1496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2132),
  [1498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5891),
  [1500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1217),
  [1502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5891),
  [1504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1181),
  [1506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4057),
  [1508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5894),
  [1510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1218),
  [1512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5894),
  [1514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1182),
  [1516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4568),
  [1518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5897),
  [1520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1219),
  [1522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5897),
  [1524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [1526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4487),
  [1528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5900),
  [1530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1220),
  [1532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5900),
  [1534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1184),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [1538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5903),
  [1540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1221),
  [1542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5903),
  [1544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1185),
  [1546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1391),
  [1548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5906),
  [1550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1222),
  [1552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5906),
  [1554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [1556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4114),
  [1558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5909),
  [1560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [1562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5909),
  [1564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1187),
  [1566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1292),
  [1568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5912),
  [1570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1224),
  [1572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5912),
  [1574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1189),
  [1576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1242),
  [1578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5918),
  [1580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [1582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5918),
  [1584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [1586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1338),
  [1588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5921),
  [1590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [1592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5921),
  [1594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1191),
  [1596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1300),
  [1598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5924),
  [1600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1228),
  [1602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5924),
  [1604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [1606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1339),
  [1608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5927),
  [1610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1230),
  [1612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5927),
  [1614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1193),
  [1616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4448),
  [1618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5930),
  [1620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [1622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5930),
  [1624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [1626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2158),
  [1628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5933),
  [1630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1232),
  [1632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5933),
  [1634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1143),
  [1636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4454),
  [1638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5936),
  [1640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1211),
  [1642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5936),
  [1644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4496),
  [1646] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(991),
  [1649] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(992),
  [1652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(991),
  [1654] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_return_expression, 1, 0, 0), SHIFT(993),
  [1657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [1659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(768),
  [1661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(769),
  [1663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [1665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [1667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(626),
  [1669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(753),
  [1671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(482),
  [1673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [1675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(78),
  [1677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(99),
  [1679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [1681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(97),
  [1683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [1685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1168),
  [1687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4039),
  [1689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5849),
  [1691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1205),
  [1693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5849),
  [1695] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(4143),
  [1698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(4274),
  [1701] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5931),
  [1704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(2391),
  [1707] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(895),
  [1710] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(896),
  [1713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(897),
  [1716] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(4278),
  [1719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [1721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [1723] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5817),
  [1726] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(3337),
  [1729] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(870),
  [1732] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(889),
  [1735] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(890),
  [1738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(940),
  [1740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [1742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(757),
  [1744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(758),
  [1746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(946),
  [1748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(947),
  [1750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(602),
  [1752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [1754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(445),
  [1756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [1758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(86),
  [1760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(106),
  [1762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(87),
  [1764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [1766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(948),
  [1770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [1772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [1774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [1776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(954),
  [1778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(955),
  [1780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(604),
  [1782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(718),
  [1784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(446),
  [1786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(676),
  [1788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(88),
  [1790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(107),
  [1792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(89),
  [1794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [1796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [1798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(5788),
  [1801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(3274),
  [1804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1095),
  [1807] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1098),
  [1810] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(1099),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [1815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(856),
  [1817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [1823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(862),
  [1825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(863),
  [1827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(681),
  [1829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(437),
  [1833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(707),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [1839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [1841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(203),
  [1843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [1845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(991),
  [1848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(992),
  [1851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 1, 0, 0), SHIFT_REPEAT(993),
  [1854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [1856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1967),
  [1858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1954),
  [1860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1273),
  [1862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [1864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1323),
  [1866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [1868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [1870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2607),
  [1872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1250),
  [1874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [1876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2689),
  [1878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(587),
  [1880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1418),
  [1882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3381),
  [1884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3632),
  [1886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1229),
  [1888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [1890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3055),
  [1892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3063),
  [1894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3072),
  [1896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2267),
  [1898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [1900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1623),
  [1902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4265),
  [1904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1529),
  [1906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5312),
  [1908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1616),
  [1911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5565),
  [1914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5566),
  [1917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5567),
  [1920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4755),
  [1923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4046),
  [1926] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4397),
  [1929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4398),
  [1932] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5449),
  [1935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5450),
  [1938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4970),
  [1941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4702),
  [1944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5495),
  [1947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1520),
  [1949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1430),
  [1951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4081),
  [1953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1764),
  [1955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2013),
  [1957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [1959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1715),
  [1961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2183),
  [1963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1719),
  [1965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1352),
  [1967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1631),
  [1969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [1971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1534),
  [1973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1384),
  [1975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [1977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1531),
  [1979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1618),
  [1981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1134),
  [1983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1532),
  [1985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4037),
  [1987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1639),
  [1989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4361),
  [1991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4183),
  [1993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [1995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [1997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1278),
  [1999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1366),
  [2001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3092),
  [2003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1404),
  [2005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4031),
  [2007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [2009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2135),
  [2011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [2013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2142),
  [2015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1335),
  [2017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4088),
  [2019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4566),
  [2021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4495),
  [2023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1400),
  [2025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1634),
  [2027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4135),
  [2029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [2031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1535),
  [2033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1243),
  [2035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [2037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4452),
  [2039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1429),
  [2041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2159),
  [2043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4455),
  [2045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1530),
  [2047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1640),
  [2049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1636),
  [2051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [2053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1427),
  [2055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1633),
  [2057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1710),
  [2059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5393),
  [2061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5394),
  [2063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5395),
  [2065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4748),
  [2067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4047),
  [2069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4429),
  [2071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4430),
  [2073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5370),
  [2075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5371),
  [2077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4788),
  [2079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4703),
  [2081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5375),
  [2083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [2085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1108),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4242),
  [2089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3618),
  [2091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5771),
  [2093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2388),
  [2095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1022),
  [2097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3618),
  [2101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2758),
  [2103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [2105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3379),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3379),
  [2109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3217),
  [2111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3217),
  [2113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2938),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2938),
  [2117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3034),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3034),
  [2121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5982),
  [2123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2200),
  [2125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1055),
  [2127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1056),
  [2129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2522),
  [2131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2522),
  [2133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1899),
  [2135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [2137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5840),
  [2139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2416),
  [2141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(893),
  [2143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(894),
  [2145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5608),
  [2147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5848),
  [2149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2504),
  [2151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1036),
  [2153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1037),
  [2155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5751),
  [2157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1472),
  [2159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5817),
  [2161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3337),
  [2163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [2165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [2167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1641),
  [2169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1421),
  [2171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [2173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1728),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [2177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1627),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [2181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1919),
  [2183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3268),
  [2185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3346),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [2189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2502),
  [2191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [2193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3508),
  [2195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3508),
  [2197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3326),
  [2199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3326),
  [2201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3546),
  [2203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3546),
  [2205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3079),
  [2207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3079),
  [2209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3281),
  [2211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [2213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3575),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1950),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [2221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1873),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [2225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1751),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1751),
  [2229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1779),
  [2231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [2233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3115),
  [2235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3115),
  [2237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2060),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2060),
  [2241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3131),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3131),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3643),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3643),
  [2249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3661),
  [2251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [2253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3159),
  [2255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3159),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2160),
  [2259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2160),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2104),
  [2263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2104),
  [2265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2323),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [2269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2541),
  [2271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2541),
  [2273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3188),
  [2275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2759),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2759),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2095),
  [2281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2095),
  [2283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3445),
  [2285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3445),
  [2287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2617),
  [2289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2584),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2141),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2141),
  [2297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1767),
  [2299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3238),
  [2303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3238),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2040),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2040),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2638),
  [2311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2638),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2665),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [2317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2873),
  [2319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2429),
  [2321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [2323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2707),
  [2325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3475),
  [2327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2263),
  [2329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [2331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1902),
  [2333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1902),
  [2335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [2337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2086),
  [2339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [2341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2364),
  [2343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2094),
  [2347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [2349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2796),
  [2351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2796),
  [2353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2817),
  [2355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2817),
  [2357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2850),
  [2359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2850),
  [2361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2091),
  [2363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2091),
  [2365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2882),
  [2367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2423),
  [2369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [2371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2491),
  [2373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2491),
  [2375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3056),
  [2377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [2379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2176),
  [2381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [2383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2974),
  [2385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2974),
  [2387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2997),
  [2389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [2391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [2393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [2395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [2397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [2399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [2403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2296),
  [2405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [2407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4147),
  [2409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4274),
  [2411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5931),
  [2413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2391),
  [2415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(896),
  [2417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(897),
  [2419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4278),
  [2421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2573),
  [2423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5322),
  [2425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2260),
  [2427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2386),
  [2429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5788),
  [2431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3274),
  [2433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1098),
  [2435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [2437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4273),
  [2439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4236),
  [2441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3353),
  [2443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4285),
  [2445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2480),
  [2447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [2449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4231),
  [2451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [2453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [2455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5229),
  [2457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4260),
  [2459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5288),
  [2461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3288),
  [2463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4141),
  [2465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1807),
  [2467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [2469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4275),
  [2471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [2473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [2475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1307),
  [2477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4180),
  [2479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [2481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1344),
  [2483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4157),
  [2485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(353),
  [2487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [2489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4226),
  [2491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [2493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [2495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4211),
  [2497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [2499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1253),
  [2501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4252),
  [2503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1354),
  [2505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4271),
  [2507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [2509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1321),
  [2511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4296),
  [2513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [2515] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_optional_parameter, 2, 0, 30),
  [2517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_optional_parameter, 2, 0, 30), SHIFT(4143),
  [2520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_optional_parameter, 2, 0, 30), SHIFT(4274),
  [2523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1047),
  [2525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1127),
  [2527] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_optional_parameter, 2, 0, 30), SHIFT(4278),
  [2530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2, 0, 0),
  [2532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2, 0, 0),
  [2534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_hash_expression, 2, 0, 0),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_hash_expression, 2, 0, 0),
  [2538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3348),
  [2540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3456),
  [2542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [2544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [2546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4143),
  [2548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(949),
  [2550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(950),
  [2552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3018),
  [2554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3019),
  [2556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3305),
  [2558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3687),
  [2560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1870),
  [2562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1874),
  [2564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2858),
  [2566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2859),
  [2568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(857),
  [2570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(858),
  [2572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3095),
  [2574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2791),
  [2576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2678),
  [2578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2679),
  [2580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2593),
  [2582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2556),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5889),
  [2586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(383),
  [2588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5959),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5866),
  [2594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5962),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5986),
  [2598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5974),
  [2600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5990),
  [2602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5946),
  [2604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5760),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5941),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5947),
  [2610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6041),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6042),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5985),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5820),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5860),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5869),
  [2624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5836),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5950),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5960),
  [2632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(384),
  [2634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [2636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5730),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5757),
  [2642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5767),
  [2644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(386),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5954),
  [2648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1380),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3447),
  [2652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1420),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3549),
  [2656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1353),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3162),
  [2660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1403),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2852),
  [2664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [2666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2668),
  [2668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2766),
  [2672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1306),
  [2674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2544),
  [2676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1249),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [2680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [2682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scalar, 2, 0, 0),
  [2684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_container_variable, 2, 0, 0),
  [2686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scalar, 2, 0, 0),
  [2688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3, 0, 0),
  [2690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3, 0, 0),
  [2692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__var_indirob_autoquote, 4, 0, 90),
  [2694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_indirob_autoquote, 4, 0, 90),
  [2696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__var_indirob, 1, 0, 0),
  [2698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_indirob, 1, 0, 0),
  [2700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indirect_object, 1, 0, 0),
  [2702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variables, 1, 0, 0),
  [2704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_indirect_object, 1, 0, 0), REDUCE(sym__variables, 1, 0, 0),
  [2707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indirect_object, 1, 0, 0),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [2711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variables, 1, 0, 0),
  [2713] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__ident_special, 2, 0, 0),
  [2715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__ident_special, 2, 0, 0),
  [2717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__ident_special, 1, 0, 0),
  [2719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__ident_special, 1, 0, 0),
  [2721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indirob, 1, 0, 0),
  [2723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indirob, 1, 0, 0),
  [2725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 1, 0, 0),
  [2727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_identifier, 1, 0, 0),
  [2729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bareword, 1, 1, 0),
  [2731] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5782),
  [2734] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), REDUCE(sym_bareword, 1, 1, 0),
  [2737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 1, 0, 0),
  [2739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bareword, 1, 1, 0),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5934),
  [2743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier, 1, 0, 0),
  [2745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ambiguous_function_call_expression, 2, 0, 25),
  [2747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ambiguous_function_call_expression, 2, 0, 25),
  [2749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_autoquoted_bareword, 2, 20, 0),
  [2751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_autoquoted_bareword, 2, 20, 0),
  [2753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5968),
  [2755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5796),
  [2757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5758),
  [2759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5872),
  [2761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5790),
  [2763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sort_routine, 1, 0, 20),
  [2765] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sort_routine, 1, 0, 20),
  [2767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5855),
  [2769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sort_routine, 1, 0, 0),
  [2771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sort_routine, 1, 0, 0),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5892),
  [2775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5951),
  [2777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5756),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5800),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5875),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5963),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5801),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5830),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6010),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5804),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5908),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5969),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6013),
  [2801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5905),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5741),
  [2805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5749),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5762),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5769),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5776),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5792),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5805),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5814),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5834),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5845),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5865),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5877),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5887),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6046),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5913),
  [2833] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5832),
  [2836] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(6038),
  [2839] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5768),
  [2842] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5794),
  [2845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(6000),
  [2848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5811),
  [2851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_function, 1, 0, 0), SHIFT(5942),
  [2854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__version, 1, 0, 0),
  [2856] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primitive, 1, 0, 0),
  [2858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__version, 1, 0, 0),
  [2860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_statement, 5, 0, 110),
  [2862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_statement, 5, 0, 110),
  [2864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5366),
  [2866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5937),
  [2868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif, 5, 0, 110),
  [2870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elsif, 5, 0, 110),
  [2872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 2, 0, 14),
  [2874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 2, 0, 14),
  [2876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5294),
  [2878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5350),
  [2880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5446),
  [2882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6034),
  [2884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 7, 0, 142),
  [2886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 7, 0, 142),
  [2888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5477),
  [2890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 61),
  [2892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 61),
  [2894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5536),
  [2896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_phaser_statement, 2, 0, 10),
  [2898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_phaser_statement, 2, 0, 10),
  [2900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__barestmt, 1, 0, 3),
  [2902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__barestmt, 1, 0, 3),
  [2904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_phaser_statement, 5, 0, 54),
  [2906] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_phaser_statement, 5, 0, 54),
  [2908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cstyle_for_statement, 6, 0, 124),
  [2910] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cstyle_for_statement, 6, 0, 124),
  [2912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 140),
  [2914] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 140),
  [2916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__fullstmt, 1, 0, 0),
  [2918] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__fullstmt, 1, 0, 0),
  [2920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 5, 0, 88),
  [2922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 5, 0, 88),
  [2924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 4, 0, 83),
  [2926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 4, 0, 83),
  [2928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 92),
  [2930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 92),
  [2932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 5, 0, 92),
  [2934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 5, 0, 92),
  [2936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 94),
  [2938] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 94),
  [2940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_label, 3, 0, 49),
  [2942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_label, 3, 0, 49),
  [2944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [2946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [2948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 141),
  [2950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 141),
  [2952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 95),
  [2954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 95),
  [2956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 125),
  [2958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 125),
  [2960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 137),
  [2962] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 137),
  [2964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_statement, 4, 0, 52),
  [2966] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_statement, 4, 0, 52),
  [2968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 4, 0, 29),
  [2970] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 4, 0, 29),
  [2972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 4, 0, 52),
  [2974] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 4, 0, 52),
  [2976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 4, 0, 29),
  [2978] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 4, 0, 29),
  [2980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 4, 0, 52),
  [2982] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 4, 0, 52),
  [2984] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 5, 0, 94),
  [2986] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 5, 0, 94),
  [2988] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 5, 0, 95),
  [2990] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 5, 0, 95),
  [2992] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_phaser_statement, 4, 0, 10),
  [2994] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_phaser_statement, 4, 0, 10),
  [2996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_phaser_statement, 4, 0, 54),
  [2998] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_phaser_statement, 4, 0, 54),
  [3000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 55),
  [3002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 55),
  [3004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 4, 0, 55),
  [3006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 4, 0, 55),
  [3008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 60),
  [3010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 60),
  [3012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_statement, 5, 0, 64),
  [3014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_statement, 5, 0, 64),
  [3016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__loop_body, 1, 0, 98),
  [3018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__loop_body, 1, 0, 98),
  [3020] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__loop_body, 1, 0, 98), SHIFT(5369),
  [3023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 7, 0, 140),
  [3025] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 7, 0, 140),
  [3027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 6, 0, 121),
  [3029] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 6, 0, 121),
  [3031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 5, 0, 99),
  [3033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 5, 0, 99),
  [3035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 4, 0, 60),
  [3037] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 4, 0, 60),
  [3039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__barestmt, 1, 0, 0),
  [3041] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__barestmt, 1, 0, 0),
  [3043] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__semicolon, 1, 0, 0),
  [3045] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__semicolon, 1, 0, 0),
  [3047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_statement, 6, 0, 110),
  [3049] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_statement, 6, 0, 110),
  [3051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 7, 0, 141),
  [3053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 7, 0, 141),
  [3055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 139),
  [3057] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 139),
  [3059] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 5, 0, 88),
  [3061] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 5, 0, 88),
  [3063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 7, 0, 139),
  [3065] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 7, 0, 139),
  [3067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__else, 1, 0, 0),
  [3069] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__else, 1, 0, 0),
  [3071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cstyle_for_statement, 7, 0, 144),
  [3073] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cstyle_for_statement, 7, 0, 144),
  [3075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__loop_body, 3, 0, 145),
  [3077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__loop_body, 3, 0, 145),
  [3079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 3, 0, 35),
  [3081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 3, 0, 35),
  [3083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 5, 0, 52),
  [3085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 5, 0, 52),
  [3087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_statement, 5, 0, 111),
  [3089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loop_statement, 5, 0, 111),
  [3091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 114),
  [3093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 114),
  [3095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else, 2, 0, 15),
  [3097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else, 2, 0, 15),
  [3099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 5, 0, 114),
  [3101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 5, 0, 114),
  [3103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 6, 0, 115),
  [3105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 6, 0, 115),
  [3107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 149),
  [3109] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 7, 0, 149),
  [3111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 137),
  [3113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 137),
  [3115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 7, 0, 149),
  [3117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 7, 0, 149),
  [3119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 3, 0, 35),
  [3121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 3, 0, 35),
  [3123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_statement, 3, 0, 37),
  [3125] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_statement, 3, 0, 37),
  [3127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_statement, 3, 0, 29),
  [3129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_statement, 3, 0, 29),
  [3131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 8, 0, 150),
  [3133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 8, 0, 150),
  [3135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 8, 0, 150),
  [3137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 8, 0, 150),
  [3139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 6, 0, 115),
  [3141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 6, 0, 115),
  [3143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 83),
  [3145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 4, 0, 83),
  [3147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 9, 0, 151),
  [3149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 9, 0, 151),
  [3151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym___DATA__, 2, 0, 0),
  [3153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym___DATA__, 2, 0, 0),
  [3155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_version_statement, 3, 0, 19),
  [3157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_version_statement, 3, 0, 19),
  [3159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 117),
  [3161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 117),
  [3163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 91),
  [3165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 5, 0, 91),
  [3167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 3, 0, 29),
  [3169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 3, 0, 29),
  [3171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 118),
  [3173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 118),
  [3175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_defer_statement, 2, 0, 15),
  [3177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_defer_statement, 2, 0, 15),
  [3179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 117),
  [3181] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 117),
  [3183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif, 6, 0, 110),
  [3185] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elsif, 6, 0, 110),
  [3187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_role_statement, 3, 0, 29),
  [3189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_role_statement, 3, 0, 29),
  [3191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 118),
  [3193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 118),
  [3195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 119),
  [3197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 119),
  [3199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 136),
  [3201] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 136),
  [3203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym___DATA__, 3, 0, 38),
  [3205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym___DATA__, 3, 0, 38),
  [3207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 119),
  [3209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 119),
  [3211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 120),
  [3213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 120),
  [3215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 6, 0, 120),
  [3217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 6, 0, 120),
  [3219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_phaser_statement, 2, 0, 10),
  [3221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_phaser_statement, 2, 0, 10),
  [3223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_phaser_statement, 3, 0, 10),
  [3225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_phaser_statement, 3, 0, 10),
  [3227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration_statement, 5, 0, 91),
  [3229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration_statement, 5, 0, 91),
  [3231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_statement, 5, 0, 52),
  [3233] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_statement, 5, 0, 52),
  [3235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 62),
  [3237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 62),
  [3239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__barestmt, 2, 0, 0),
  [3241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__barestmt, 2, 0, 0),
  [3243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_statement, 4, 0, 37),
  [3245] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_statement, 4, 0, 37),
  [3247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_statement, 4, 0, 64),
  [3249] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_statement, 4, 0, 64),
  [3251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 136),
  [3253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_declaration_statement, 6, 0, 136),
  [3255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__PERLY_BRACE_OPEN, 1, 0, 0),
  [3257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__PERLY_BRACE_OPEN, 1, 0, 0),
  [3259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5226),
  [3261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5492),
  [3263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5467),
  [3265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5401),
  [3267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__loop_body, 1, 0, 98), SHIFT(5439),
  [3270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5524),
  [3272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6040),
  [3274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package, 1, 0, 0),
  [3276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package, 1, 0, 0),
  [3278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5174),
  [3280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5417),
  [3282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5488),
  [3284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5532),
  [3286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__loop_body, 1, 0, 98), SHIFT(5522),
  [3289] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__conditionals, 1, 0, 0),
  [3291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__conditionals, 1, 0, 0),
  [3293] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indirect_object, 2, 0, 0),
  [3295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indirect_object, 2, 0, 0),
  [3297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__loops, 1, 0, 0),
  [3299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__loops, 1, 0, 0),
  [3301] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(1742),
  [3304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0),
  [3306] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5305),
  [3309] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0),
  [3311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1742),
  [3313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attrlist, 1, 0, 0),
  [3315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5305),
  [3317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attrlist, 1, 0, 0),
  [3319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attrlist, 2, 0, 0),
  [3321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attrlist, 2, 0, 0),
  [3323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 3, 0, 12),
  [3325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 3, 0, 12),
  [3327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 3, 0, 11),
  [3329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 3, 0, 11),
  [3331] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_name, 1, 0, 0),
  [3333] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_name, 1, 0, 0),
  [3335] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 1, 0, 33),
  [3337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 33),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5994),
  [3341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_slice_expression, 5, 0, 0),
  [3343] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_slice_expression, 5, 0, 0),
  [3345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscripted, 1, 0, 0),
  [3347] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscripted, 1, 0, 0),
  [3349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_coderef_call_expression, 4, 0, 73),
  [3351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_coderef_call_expression, 4, 0, 73),
  [3353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_element_expression, 5, 0, 106),
  [3355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_element_expression, 5, 0, 106),
  [3357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_coderef_call_expression, 5, 0, 107),
  [3359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_coderef_call_expression, 5, 0, 107),
  [3361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_coderef_call_expression, 3, 0, 0),
  [3363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_coderef_call_expression, 3, 0, 0),
  [3365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_expression, 4, 0, 74),
  [3367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_expression, 4, 0, 74),
  [3369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_expression, 5, 0, 108),
  [3371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_expression, 5, 0, 108),
  [3373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_glob_slot_expression, 4, 0, 0),
  [3375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_glob_slot_expression, 4, 0, 0),
  [3377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_glob_slot_expression, 6, 0, 0),
  [3379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_glob_slot_expression, 6, 0, 0),
  [3381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_slice_expression, 4, 0, 82),
  [3383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_slice_expression, 4, 0, 82),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_element_expression, 4, 0, 75),
  [3387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_element_expression, 4, 0, 75),
  [3389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_expression, 4, 0, 76),
  [3391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_expression, 4, 0, 76),
  [3393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_slice_expression, 6, 0, 21),
  [3395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_slice_expression, 6, 0, 21),
  [3397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_element_expression, 4, 0, 72),
  [3399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_element_expression, 4, 0, 72),
  [3401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 5, 0, 138),
  [3403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 5, 0, 138),
  [3405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_coderef_call_expression, 4, 0, 0),
  [3407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_coderef_call_expression, 4, 0, 0),
  [3409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 1, 0, 0),
  [3411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attrlist_repeat1, 1, 0, 0),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [3415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__term, 1, 0, 0),
  [3417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__term, 1, 0, 0),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [3421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [3423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_container_variable, 2, 0, 0),
  [3425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 2, 0, 0),
  [3427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 2, 0, 0),
  [3429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyval_container_variable, 2, 0, 0),
  [3431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash, 2, 0, 0),
  [3433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash, 2, 0, 0),
  [3435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quotelike_end, 1, 0, 0),
  [3437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quotelike_end, 1, 0, 0),
  [3439] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_container_variable, 2, 0, 0),
  [3441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 5, 0, 104),
  [3443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 5, 0, 104),
  [3445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [3447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declare_hash, 2, 0, 0),
  [3449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declare_hash, 2, 0, 0),
  [3451] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__term, 3, 0, 0),
  [3453] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__term, 3, 0, 0),
  [3455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [3457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stub_expression, 2, 0, 0),
  [3459] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_stub_expression, 2, 0, 0),
  [3461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 3, 0, 39),
  [3465] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 3, 0, 39),
  [3467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [3469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 1, 0, 0),
  [3471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 1, 0, 0),
  [3473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1937),
  [3475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method, 1, 0, 0),
  [3477] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method, 1, 0, 0),
  [3479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 3, 0, 44),
  [3481] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 3, 0, 44),
  [3483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [3485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0),
  [3487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [3489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [3491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(6),
  [3494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [3496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4041),
  [3498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [3500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1016),
  [3502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [3504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(864),
  [3506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1020),
  [3508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1032),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [3512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [3514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1049),
  [3516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1050),
  [3518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1051),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [3522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1060),
  [3524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [3526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [3528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [3530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [3532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1930),
  [3534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [3536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__listexpr, 1, 0, 0),
  [3538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 5, 0, 103),
  [3540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 5, 0, 103),
  [3542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1893),
  [3544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 5, 0, 70),
  [3546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 5, 0, 70),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [3550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 6, 0, 130),
  [3552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 6, 0, 130),
  [3554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1836),
  [3556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 3, 0, 48),
  [3558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 3, 0, 48),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1925),
  [3562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 4, 0, 79),
  [3564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 4, 0, 79),
  [3566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [3568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__term_rightward, 2, 0, 0),
  [3570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__term_rightward, 2, 0, 0),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [3574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func1op_call_expression, 4, 0, 1),
  [3576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func1op_call_expression, 4, 0, 1),
  [3578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 5, 0, 103),
  [3580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 5, 0, 103),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1883),
  [3584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 6, 0, 128),
  [3586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 6, 0, 128),
  [3588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [3590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0),
  [3592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0),
  [3594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(6),
  [3597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declare_scalar, 2, 0, 0),
  [3599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declare_scalar, 2, 0, 0),
  [3601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_amper_sub, 2, 0, 0),
  [3603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_amper_sub, 2, 0, 0),
  [3605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_glob, 2, 0, 0),
  [3607] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_glob, 2, 0, 0),
  [3609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [3611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declare_array, 2, 0, 0),
  [3613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declare_array, 2, 0, 0),
  [3615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 4, 0, 39),
  [3617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 4, 0, 39),
  [3619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [3621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 1, 0, 1),
  [3623] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 1, 0, 1),
  [3625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5782),
  [3627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decl_variable_list, 3, 0, 0),
  [3629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__decl_variable_list, 3, 0, 0),
  [3631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func0op_call_expression, 1, 0, 1),
  [3635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func0op_call_expression, 1, 0, 1),
  [3637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5826),
  [3639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_word_list, 3, 0, 0),
  [3641] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_word_list, 3, 0, 0),
  [3643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_regexp, 3, 0, 0),
  [3645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_regexp, 3, 0, 0),
  [3647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [3649] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func1op_call_expression, 3, 0, 1),
  [3651] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func1op_call_expression, 3, 0, 1),
  [3653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [3655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 2, 0, 0),
  [3657] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 2, 0, 0),
  [3659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [3661] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decl_variable_list, 4, 0, 0),
  [3663] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__decl_variable_list, 4, 0, 0),
  [3665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_varname, 1, 0, 0),
  [3667] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_varname, 1, 0, 0),
  [3669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 2, 0, 12),
  [3671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1739),
  [3673] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 2, 0, 12),
  [3675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declared_vars, 1, 0, 0),
  [3677] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declared_vars, 1, 0, 0),
  [3679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decl_variable_list, 2, 0, 0),
  [3681] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__decl_variable_list, 2, 0, 0),
  [3683] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 4, 0, 39),
  [3685] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 4, 0, 39),
  [3687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1858),
  [3689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 2, 0, 11),
  [3691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [3693] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 2, 0, 11),
  [3695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_refgen_expression, 2, 0, 0),
  [3697] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_refgen_expression, 2, 0, 0),
  [3699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_word_list, 4, 0, 66),
  [3701] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_word_list, 4, 0, 66),
  [3703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_regexp, 4, 0, 68),
  [3705] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_regexp, 4, 0, 68),
  [3707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [3709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 4, 0, 70),
  [3711] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 4, 0, 70),
  [3713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [3715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyval_expression, 4, 0, 78),
  [3717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyval_expression, 4, 0, 78),
  [3719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 40),
  [3721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 40),
  [3723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_grep_expression, 7, 0, 143),
  [3725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_grep_expression, 7, 0, 143),
  [3727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 5, 0, 101),
  [3729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 5, 0, 101),
  [3731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_autoquoted_bareword, 1, 0, 0),
  [3733] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_autoquoted_bareword, 1, 0, 0),
  [3735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relational_expression, 3, 0, 40),
  [3737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_relational_expression, 3, 0, 40),
  [3739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_expression, 2, 0, 0),
  [3741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_require_expression, 2, 0, 0),
  [3743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 5, 0, 102),
  [3745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 5, 0, 102),
  [3747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 3, 0, 34),
  [3749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 3, 0, 34),
  [3751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 7, 0, 146),
  [3753] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 7, 0, 146),
  [3755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 7, 0, 147),
  [3757] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 7, 0, 147),
  [3759] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 7, 0, 148),
  [3761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 7, 0, 148),
  [3763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_equality_expression, 3, 0, 40),
  [3765] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_equality_expression, 3, 0, 40),
  [3767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [3769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [3771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ambiguous_function_call_expression, 3, 0, 45),
  [3773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ambiguous_function_call_expression, 3, 0, 45),
  [3775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sort_expression, 2, 0, 21),
  [3777] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sort_expression, 2, 0, 21),
  [3779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 3, 0, 34),
  [3781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 3, 0, 34),
  [3783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 3, 0, 46),
  [3785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 3, 0, 46),
  [3787] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolated_string_literal, 3, 0, 46),
  [3789] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolated_string_literal, 3, 0, 46),
  [3791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_string, 3, 0, 46),
  [3793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_string, 3, 0, 46),
  [3795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 3, 0, 47),
  [3797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 3, 0, 47),
  [3799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 6, 0, 135),
  [3801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 6, 0, 135),
  [3803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_await_expression, 2, 0, 0),
  [3805] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_await_expression, 2, 0, 0),
  [3807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arraylen_deref_expression, 4, 0, 0),
  [3809] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arraylen_deref_expression, 4, 0, 0),
  [3811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 5, 0, 102),
  [3813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 5, 0, 102),
  [3815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_version_expression, 2, 0, 19),
  [3817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_require_version_expression, 2, 0, 19),
  [3819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ambiguous_function_call_expression, 2, 0, 26),
  [3821] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ambiguous_function_call_expression, 2, 0, 26),
  [3823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [3825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 6, 0, 131),
  [3827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 6, 0, 131),
  [3829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 5, 0, 93),
  [3831] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 5, 0, 93),
  [3833] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func0op_call_expression, 3, 0, 1),
  [3835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func0op_call_expression, 3, 0, 1),
  [3837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 3, 0, 0),
  [3839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 3, 0, 0),
  [3841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_array_expression, 3, 0, 0),
  [3843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_array_expression, 3, 0, 0),
  [3845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_localization_expression, 2, 0, 0),
  [3847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_localization_expression, 2, 0, 0),
  [3849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 6, 0, 132),
  [3851] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 6, 0, 132),
  [3853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_grep_expression, 3, 0, 36),
  [3855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_grep_expression, 3, 0, 36),
  [3857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arraylen, 2, 0, 0),
  [3859] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arraylen, 2, 0, 0),
  [3861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sort_expression, 3, 0, 36),
  [3863] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sort_expression, 3, 0, 36),
  [3865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 93),
  [3867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 93),
  [3869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_array_expression, 2, 0, 0),
  [3871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_array_expression, 2, 0, 0),
  [3873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 5, 0, 105),
  [3875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 5, 0, 105),
  [3877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_glob_deref_expression, 4, 0, 0),
  [3879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_glob_deref_expression, 4, 0, 0),
  [3881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loopex_expression, 2, 0, 2),
  [3883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loopex_expression, 2, 0, 2),
  [3885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 2, 0, 0),
  [3887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 2, 0, 0),
  [3889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolated_string_literal, 2, 0, 0),
  [3891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolated_string_literal, 2, 0, 0),
  [3893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scalar_deref_expression, 4, 0, 0),
  [3895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scalar_deref_expression, 4, 0, 0),
  [3897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__label_arg, 1, 0, 0),
  [3899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__label_arg, 1, 0, 0),
  [3901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_expression, 2, 0, 0),
  [3903] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_expression, 2, 0, 0),
  [3905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 6, 0, 126),
  [3907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 6, 0, 126),
  [3909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_eval_expression, 2, 0, 18),
  [3911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_eval_expression, 2, 0, 18),
  [3913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_equality_expression, 6, 0, 132),
  [3915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_equality_expression, 6, 0, 132),
  [3917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal, 1, 0, 0),
  [3919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal, 1, 0, 0),
  [3921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 6, 0, 135),
  [3923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 6, 0, 135),
  [3925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 4, 0, 56),
  [3927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 4, 0, 56),
  [3929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 4, 0, 57),
  [3931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 4, 0, 57),
  [3933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 4, 0, 58),
  [3935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 4, 0, 58),
  [3937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 4, 0, 59),
  [3939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 4, 0, 59),
  [3941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__listop, 1, 0, 0),
  [3943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__listop, 1, 0, 0),
  [3945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 6, 0, 126),
  [3947] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 6, 0, 126),
  [3949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_readline_expression, 2, 0, 9),
  [3951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_readline_expression, 2, 0, 9),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 4, 0, 58),
  [3955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 4, 0, 58),
  [3957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_token, 2, 0, 0),
  [3959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_token, 2, 0, 0),
  [3961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 6, 0, 134),
  [3963] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 6, 0, 134),
  [3965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 4, 0, 59),
  [3967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 4, 0, 59),
  [3969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 4, 0, 71),
  [3971] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 4, 0, 71),
  [3973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_string, 2, 0, 0),
  [3975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_string, 2, 0, 0),
  [3977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func1op_call_expression, 2, 0, 1),
  [3979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func1op_call_expression, 2, 0, 1),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 5, 0, 44),
  [3983] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 5, 0, 44),
  [3985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 109),
  [3987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 109),
  [3989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_deref_expression, 4, 0, 0),
  [3991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_deref_expression, 4, 0, 0),
  [3993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolated_string_literal, 3, 0, 0),
  [3995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolated_string_literal, 3, 0, 0),
  [3997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, 0, 17),
  [3999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, 0, 17),
  [4001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tricky_indirob_hashref, 4, 0, 0),
  [4003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tricky_indirob_hashref, 4, 0, 0),
  [4005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_string, 3, 0, 0),
  [4007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_string, 3, 0, 0),
  [4009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 112),
  [4011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 112),
  [4013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 113),
  [4015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 5, 0, 113),
  [4017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_eval_expression, 2, 0, 0),
  [4019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_eval_expression, 2, 0, 0),
  [4021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_readline_expression, 3, 0, 27),
  [4023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_readline_expression, 3, 0, 27),
  [4025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 5, 0, 112),
  [4027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 5, 0, 112),
  [4029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 5, 0, 113),
  [4031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 5, 0, 113),
  [4033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fileglob_expression, 3, 0, 28),
  [4035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fileglob_expression, 3, 0, 28),
  [4037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_expression, 2, 0, 0),
  [4039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_expression, 2, 0, 0),
  [4041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ambiguous_function_call_expression, 4, 0, 80),
  [4043] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ambiguous_function_call_expression, 4, 0, 80),
  [4045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primitive, 1, 0, 0),
  [4047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [4049] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [4051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 6, 0, 133),
  [4053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 6, 0, 133),
  [4055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 4, 0, 81),
  [4057] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 4, 0, 81),
  [4059] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 6, 0, 51),
  [4061] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 6, 0, 51),
  [4063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_deref_expression, 4, 0, 0),
  [4065] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_deref_expression, 4, 0, 0),
  [4067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_transliteration_expression, 6, 0, 129),
  [4069] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_transliteration_expression, 6, 0, 129),
  [4071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slices, 1, 0, 0),
  [4073] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slices, 1, 0, 0),
  [4075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postinc_expression, 2, 0, 24),
  [4077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_postinc_expression, 2, 0, 24),
  [4079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 4, 0, 77),
  [4081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 4, 0, 77),
  [4083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 4, 0, 78),
  [4085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 4, 0, 78),
  [4087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_hash_expression, 3, 0, 0),
  [4089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_hash_expression, 3, 0, 0),
  [4091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyval_expression, 4, 0, 77),
  [4093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyval_expression, 4, 0, 77),
  [4095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sort_expression, 5, 0, 96),
  [4097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sort_expression, 5, 0, 96),
  [4099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyval_expression, 6, 0, 133),
  [4101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyval_expression, 6, 0, 133),
  [4103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 2, 0, 16),
  [4105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 2, 0, 16),
  [4107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 40),
  [4109] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 3, 0, 40),
  [4111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_grep_expression, 6, 0, 122),
  [4113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_grep_expression, 6, 0, 122),
  [4115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_amper_deref_expression, 4, 0, 0),
  [4117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_amper_deref_expression, 4, 0, 0),
  [4119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__label_arg, 1, 0, 22),
  [4121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__label_arg, 1, 0, 22),
  [4123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sort_expression, 6, 0, 122),
  [4125] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sort_expression, 6, 0, 122),
  [4127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyval_expression, 6, 0, 51),
  [4129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keyval_expression, 6, 0, 51),
  [4131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_expression, 2, 0, 0),
  [4133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_expression, 2, 0, 0),
  [4135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postfix_deref, 1, 0, 0),
  [4137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_postfix_deref, 1, 0, 0),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [4141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [4143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relational_expression, 6, 0, 132),
  [4145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_relational_expression, 6, 0, 132),
  [4147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_call_expression, 6, 0, 79),
  [4149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_call_expression, 6, 0, 79),
  [4151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_undef_expression, 2, 0, 0),
  [4153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_undef_expression, 2, 0, 0),
  [4155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_grep_expression, 4, 0, 63),
  [4157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_grep_expression, 4, 0, 63),
  [4159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_autoquoted_bareword, 1, 0, 0), REDUCE(sym_autoquoted_bareword, 2, 0, 0),
  [4162] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_autoquoted_bareword, 1, 0, 0), REDUCE(sym_autoquoted_bareword, 2, 0, 0),
  [4165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_subroutine_expression, 2, 0, 13),
  [4167] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_subroutine_expression, 2, 0, 13),
  [4169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17),
  [4171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preinc_expression, 2, 0, 17),
  [4173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(1930),
  [4176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 4, 0, 66),
  [4178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 4, 0, 66),
  [4180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolated_string_literal, 4, 0, 66),
  [4182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolated_string_literal, 4, 0, 66),
  [4184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_regexp, 5, 0, 100),
  [4186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_regexp, 5, 0, 100),
  [4188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_string, 4, 0, 66),
  [4190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_string, 4, 0, 66),
  [4192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_regexp, 4, 0, 67),
  [4194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_regexp, 4, 0, 67),
  [4196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_substitution_regexp, 6, 0, 127),
  [4198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_substitution_regexp, 6, 0, 127),
  [4200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_regexp, 4, 0, 69),
  [4202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_regexp, 4, 0, 69),
  [4204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_method_expression, 2, 0, 13),
  [4206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_method_expression, 2, 0, 13),
  [4208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_heredoc_token, 2, 0, 0),
  [4210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_heredoc_token, 2, 0, 0),
  [4212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [4214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1974),
  [4216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5240),
  [4218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(1974),
  [4221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5240),
  [4224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6024),
  [4226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2007),
  [4228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5274),
  [4230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2007),
  [4233] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5274),
  [4236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2033),
  [4238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5289),
  [4240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2059),
  [4242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5243),
  [4244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2059),
  [4247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5243),
  [4250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2017),
  [4252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5244),
  [4254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2017),
  [4257] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5244),
  [4260] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2033),
  [4263] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5289),
  [4266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [4268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [4270] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2166),
  [4273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5187),
  [4276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(811),
  [4278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [4280] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(11),
  [4283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [4285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(814),
  [4287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(815),
  [4289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(816),
  [4291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(813),
  [4293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(818),
  [4295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(820),
  [4297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(822),
  [4301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(825),
  [4303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(828),
  [4305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(830),
  [4307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [4309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(835),
  [4311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [4313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [4315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [4317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [4319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [4321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2166),
  [4323] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(11),
  [4326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6001),
  [4328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [4332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6026),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [4336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [4338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [4342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(741),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [4346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [4348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6025),
  [4350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [4352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [4354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [4356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [4360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2574),
  [4362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5310),
  [4364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2574),
  [4367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5310),
  [4370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [4372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [4374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [4376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [4378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [4380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [4382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2481),
  [4384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5313),
  [4386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6027),
  [4388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(2481),
  [4391] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5313),
  [4394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [4396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [4398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [4400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2345),
  [4402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [4404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2542),
  [4406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [4408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(886),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [4412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [4414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4032),
  [4418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [4420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(906),
  [4422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(908),
  [4424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [4426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(910),
  [4428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [4430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [4432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(915),
  [4434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [4436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(918),
  [4438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [4440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [4442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(921),
  [4444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [4446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [4448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [4450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [4452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [4454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [4456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [4458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2485),
  [4460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [4462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [4464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2488),
  [4466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [4468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2463),
  [4470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2465),
  [4472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [4474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2343),
  [4476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [4478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [4480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2347),
  [4482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5832),
  [4484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [4486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2406),
  [4490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5943),
  [4492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [4494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(20),
  [4497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(20),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [4504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [4506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [4508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2754),
  [4510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1069),
  [4512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [4514] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(31),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4040),
  [4521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1075),
  [4523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1077),
  [4525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1079),
  [4527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1072),
  [4529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1081),
  [4531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [4533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [4535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1088),
  [4537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1094),
  [4539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1096),
  [4541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1097),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [4545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1110),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [4549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [4551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [4555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2695),
  [4557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2755),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [4563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [4565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [4567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3215),
  [4569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2778),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [4573] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(30),
  [4576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3015),
  [4578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [4580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [4582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2788),
  [4584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2789),
  [4586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [4588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [4590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [4592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2790),
  [4594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2792),
  [4596] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(31),
  [4599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [4601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [4603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3033),
  [4605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [4607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [4609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3046),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3216),
  [4615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [4617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [4619] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(81),
  [4622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [4624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(819),
  [4626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [4628] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(81),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [4633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4028),
  [4635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(823),
  [4637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(826),
  [4639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(827),
  [4641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(821),
  [4643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [4645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(831),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [4649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(833),
  [4651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(834),
  [4653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(836),
  [4655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2874),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [4675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3050),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3051),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3052),
  [4681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3053),
  [4683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [4685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2806),
  [4687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(939),
  [4689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [4695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4026),
  [4697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(958),
  [4699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(806),
  [4701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(962),
  [4703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(956),
  [4705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(965),
  [4707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(968),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [4711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(970),
  [4713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(972),
  [4715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(974),
  [4717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(976),
  [4719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [4721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(984),
  [4723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [4725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [4727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [4729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [4731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3180),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [4735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2807),
  [4737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2808),
  [4739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2809),
  [4741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3066),
  [4743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3067),
  [4745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3069),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3071),
  [4749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3204),
  [4751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5896),
  [4753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2825),
  [4755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2826),
  [4757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2666),
  [4759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5732),
  [4761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2851),
  [4763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [4767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1057),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [4771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(24),
  [4774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [4776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4042),
  [4778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1062),
  [4780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1063),
  [4782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1064),
  [4784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1061),
  [4786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1065),
  [4788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1066),
  [4790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [4792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [4794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1068),
  [4796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1070),
  [4798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1071),
  [4800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [4802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1074),
  [4804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [4806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [4808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [4810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [4812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2846),
  [4814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [4816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3088),
  [4818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3089),
  [4820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3160),
  [4822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(24),
  [4825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [4827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5811),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [4831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3229),
  [4833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [4835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [4837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3233),
  [4839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3234),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3235),
  [4843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3236),
  [4845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [4847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [4849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [4853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6038),
  [4855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [4859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6033),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2921),
  [4863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [4865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6022),
  [4867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [4869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [4871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [4873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5926),
  [4875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2936),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2937),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2760),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [4883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [4885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2630),
  [4887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [4889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(778),
  [4891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2742),
  [4893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2955),
  [4895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [4897] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(2595),
  [4900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [4902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2963),
  [4904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2964),
  [4906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [4908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2967),
  [4910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5794),
  [4912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [4914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [4916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [4918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [4920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6015),
  [4922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [4924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [4926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2985),
  [4928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2987),
  [4930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [4932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [4934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [4936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3005),
  [4938] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(30),
  [4941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3006),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [4947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(2695),
  [4950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [4952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [4954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(770),
  [4956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [4958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [4960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(2846),
  [4963] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(2874),
  [4966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [4968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [4970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [4972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(804),
  [4974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [4976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [4978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [4980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(790),
  [4982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [4984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [4986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [4988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [4990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [4992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [4994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [4996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [4998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [5000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [5002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [5004] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(3180),
  [5007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [5009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1119),
  [5011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(982),
  [5013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(860),
  [5015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(973),
  [5017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [5019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3421),
  [5021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [5023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3437),
  [5025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3444),
  [5027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3500),
  [5029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3503),
  [5031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3314),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3315),
  [5037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3317),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3318),
  [5041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [5043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3334),
  [5045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_parameter, 4, 0, 116),
  [5047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1045),
  [5049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [5053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1120),
  [5055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1052),
  [5057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1041),
  [5059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [5061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1111),
  [5063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(845),
  [5065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(967),
  [5067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1123),
  [5069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [5071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [5075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [5077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [5079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6019),
  [5081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3446),
  [5083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [5085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [5087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6000),
  [5089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [5091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [5093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_optional_parameter, 3, 0, 89),
  [5095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [5097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(735),
  [5099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3319),
  [5101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3343),
  [5103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3359),
  [5105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [5107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [5109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3668),
  [5111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5948),
  [5113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3547),
  [5115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [5117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5942),
  [5119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [5121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2198),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3606),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3616),
  [5133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3617),
  [5135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [5137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(977),
  [5139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [5143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [5145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(980),
  [5147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(981),
  [5149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(978),
  [5151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(983),
  [5153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(986),
  [5155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [5157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(988),
  [5159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(989),
  [5161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(990),
  [5163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [5165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [5167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [5169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [5171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [5173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [5175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [5177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [5179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3634),
  [5181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3638),
  [5183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3639),
  [5185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3640),
  [5187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3641),
  [5189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [5191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(92),
  [5194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(855),
  [5196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [5200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [5202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4025),
  [5204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(900),
  [5206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1040),
  [5208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1043),
  [5210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [5212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1059),
  [5214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [5216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1078),
  [5218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1085),
  [5220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1112),
  [5222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(808),
  [5224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(809),
  [5226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [5228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(846),
  [5230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [5232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [5234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [5236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [5238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3464),
  [5240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [5242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [5244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3651),
  [5246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [5248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3653),
  [5250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3654),
  [5252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(975),
  [5254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [5256] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(85),
  [5259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [5261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4023),
  [5263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [5265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1006),
  [5267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [5269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [5271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1010),
  [5273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1011),
  [5275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [5277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1012),
  [5279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1014),
  [5281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1019),
  [5283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [5285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [5287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1028),
  [5289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [5291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [5293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [5295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [5297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3567),
  [5299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [5301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [5303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [5305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [5307] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(92),
  [5310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [5312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [5314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(3464),
  [5317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [5319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [5321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(85),
  [5324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [5326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [5328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(871),
  [5330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [5332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [5334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [5336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(873),
  [5338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(874),
  [5340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(875),
  [5342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(872),
  [5344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [5346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [5348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [5350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [5352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(879),
  [5354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(880),
  [5356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [5358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [5360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(883),
  [5362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [5364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [5366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [5368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [5370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [5372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [5374] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_preinc_expression, 2, 0, 17), SHIFT(3567),
  [5377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [5379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [5381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [5383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [5385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [5387] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__listexpr, 1, 0, 0), SHIFT(100),
  [5390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(100),
  [5393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [5395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [5397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [5399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [5401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [5403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [5405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [5407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [5409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [5411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [5415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(926),
  [5417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [5419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(929),
  [5421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [5423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [5425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(923),
  [5427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [5431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(927),
  [5433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(925),
  [5435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [5437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [5439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(931),
  [5441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(932),
  [5443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(933),
  [5445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [5447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [5449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [5451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [5453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [5455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [5457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [5459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [5461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [5463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [5465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [5467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [5469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [5471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [5473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [5475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [5477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [5479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [5481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [5483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [5485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [5487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [5489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [5491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [5493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [5495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [5497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [5499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [5501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [5503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [5505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [5507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [5509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [5511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(861),
  [5513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [5515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [5517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(850),
  [5519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [5521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(866),
  [5523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(849),
  [5525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(909),
  [5527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [5529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [5531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(944),
  [5533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(945),
  [5535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(952),
  [5537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [5539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [5541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(959),
  [5543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [5545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [5547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [5549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [5551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [5553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [5555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [5557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [5559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [5561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [5563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [5565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [5567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [5569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [5571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [5573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [5575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [5577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [5579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [5581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [5583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [5585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [5587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4472),
  [5589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [5591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [5593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5665),
  [5595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4426),
  [5597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4406),
  [5599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4092),
  [5601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4015),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [5605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4482),
  [5607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [5609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [5611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [5613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3429),
  [5615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [5617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [5619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1808),
  [5621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3262),
  [5623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [5625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [5627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [5629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2472),
  [5631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [5633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [5635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [5637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4486),
  [5639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4184),
  [5641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4017),
  [5643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2202),
  [5645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4497),
  [5647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2085),
  [5649] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4472),
  [5652] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4089),
  [5655] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4016),
  [5658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0),
  [5660] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4426),
  [5663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2235),
  [5665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [5667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [5669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3275),
  [5671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2464),
  [5673] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4406),
  [5676] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4092),
  [5679] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4015),
  [5682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 2, 0, 0), SHIFT_REPEAT(4482),
  [5685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3026),
  [5687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [5689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2469),
  [5691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3309),
  [5693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2243),
  [5695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3449),
  [5697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [5699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [5701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [5703] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 2, 0, 0), SHIFT_REPEAT(4486),
  [5706] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 2, 0, 0), SHIFT_REPEAT(4184),
  [5709] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 2, 0, 0), SHIFT_REPEAT(4017),
  [5712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 2, 0, 0),
  [5714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 2, 0, 0), SHIFT_REPEAT(4497),
  [5717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4556),
  [5719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4199),
  [5721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4014),
  [5723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4583),
  [5725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6048),
  [5727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6047),
  [5729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4556),
  [5732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4199),
  [5735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4014),
  [5738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4583),
  [5741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 2, 0, 0),
  [5743] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(215),
  [5746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2720),
  [5748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [5750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [5752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [5754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [5756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [5758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [5760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1735),
  [5762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5209),
  [5764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [5766] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(6035),
  [5769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2898),
  [5771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3494),
  [5773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5765),
  [5776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3585),
  [5778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [5780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2948),
  [5782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5745),
  [5785] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5827),
  [5788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5295),
  [5790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1933),
  [5792] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5914),
  [5795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5222),
  [5797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5744),
  [5800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5919),
  [5803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0), SHIFT(5791),
  [5806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4325),
  [5808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1527),
  [5810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4129),
  [5812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4594),
  [5814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4330),
  [5816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4594),
  [5818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4578),
  [5820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4085),
  [5822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4094),
  [5824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4473),
  [5826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4138),
  [5828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4473),
  [5830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [5832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4095),
  [5834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4060),
  [5836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4440),
  [5838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4068),
  [5840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4440),
  [5842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4420),
  [5844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4165),
  [5846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4063),
  [5848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4512),
  [5850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4192),
  [5852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4512),
  [5854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4521),
  [5856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lowprec_logical_expression, 3, 0, 40),
  [5858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lowprec_logical_expression, 3, 0, 40),
  [5860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expr, 1, 0, 0),
  [5862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expr, 1, 0, 0),
  [5864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5031),
  [5866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5033),
  [5868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5285),
  [5870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5535),
  [5872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5561),
  [5874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4182),
  [5876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5057),
  [5878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3360),
  [5880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [5882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4103),
  [5884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5418),
  [5886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [5888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5653),
  [5890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [5892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5523),
  [5894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4674),
  [5896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5944),
  [5898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [5900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__interpolations, 1, 0, 0),
  [5902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolations, 1, 0, 0),
  [5904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4526),
  [5906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [5908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [5910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3293),
  [5912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [5914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4130),
  [5916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5512),
  [5918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [5920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5671),
  [5922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [5924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5547),
  [5926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4676),
  [5928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5925),
  [5930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3293),
  [5932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2393),
  [5934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [5936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4093),
  [5938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5379),
  [5940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [5942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5600),
  [5944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [5946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [5948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4672),
  [5950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [5952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [5954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4412),
  [5956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2505),
  [5958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [5960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4127),
  [5962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5521),
  [5964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [5966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5660),
  [5968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [5970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5531),
  [5972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4698),
  [5974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5899),
  [5976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [5978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4489),
  [5980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [5982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [5984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5661),
  [5986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4389),
  [5988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4572),
  [5990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5853),
  [5992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4573),
  [5994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2223),
  [5996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [5998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4131),
  [6000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5559),
  [6002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [6004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5677),
  [6006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [6008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5378),
  [6010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4682),
  [6012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5929),
  [6014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2223),
  [6016] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5661),
  [6019] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4389),
  [6022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4572),
  [6025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0),
  [6027] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4573),
  [6030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2273),
  [6032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [6034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4091),
  [6036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5540),
  [6038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [6040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5669),
  [6042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [6044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5542),
  [6046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4636),
  [6048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5881),
  [6050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2273),
  [6052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1782),
  [6054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [6056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4097),
  [6058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5349),
  [6060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [6062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5623),
  [6064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [6066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5404),
  [6068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4653),
  [6070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5735),
  [6072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1782),
  [6074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2311),
  [6076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [6078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4119),
  [6080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5424),
  [6082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [6084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5725),
  [6086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [6088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5425),
  [6090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4614),
  [6092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5733),
  [6094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2311),
  [6096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4428),
  [6098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5928),
  [6100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5574),
  [6102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5464),
  [6104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5402),
  [6106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5403),
  [6108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [6110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4048),
  [6112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1332),
  [6114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [6116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1279),
  [6118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4049),
  [6120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [6122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [6124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4504),
  [6126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1381),
  [6128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4052),
  [6130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1387),
  [6132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [6134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [6136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4053),
  [6138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [6140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [6142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4435),
  [6144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4499),
  [6146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5729),
  [6148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4517),
  [6150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [6152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4056),
  [6154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1368),
  [6156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [6158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3068),
  [6160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4058),
  [6162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2326),
  [6164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2334),
  [6166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4070),
  [6168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2105),
  [6170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4062),
  [6172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2036),
  [6174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [6176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4222),
  [6178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2110),
  [6180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4065),
  [6182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2038),
  [6184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2039),
  [6186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4587),
  [6188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4066),
  [6190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4595),
  [6192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4580),
  [6194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4537),
  [6196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4067),
  [6198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4545),
  [6200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4547),
  [6202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [6204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4069),
  [6206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1409),
  [6208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [6210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1394),
  [6212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4071),
  [6214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1399),
  [6216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [6218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1275),
  [6220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4072),
  [6222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [6224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [6226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [6228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4073),
  [6230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1255),
  [6232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [6234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [6236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4074),
  [6238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1238),
  [6240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [6242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1324),
  [6244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4076),
  [6246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [6248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [6250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [6252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4077),
  [6254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1280),
  [6256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1282),
  [6258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4490),
  [6260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [6262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [6264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4462),
  [6266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4084),
  [6268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4399),
  [6270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4444),
  [6272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4456),
  [6274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4080),
  [6276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4460),
  [6278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4461),
  [6280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2143),
  [6282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4082),
  [6284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2148),
  [6286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2151),
  [6288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4425),
  [6290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4413),
  [6292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4474),
  [6294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2767),
  [6296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4476),
  [6298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3222),
  [6300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4104),
  [6302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1761),
  [6304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4122),
  [6306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1880),
  [6308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1743),
  [6310] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5724),
  [6313] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4479),
  [6316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [6318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4099),
  [6320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1303),
  [6322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [6324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3626),
  [6326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2598),
  [6328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4106),
  [6330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2424),
  [6332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [6334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5241),
  [6336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4111),
  [6338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5320),
  [6340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5266),
  [6342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [6344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4113),
  [6346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [6348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [6350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2083),
  [6352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4133),
  [6354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3039),
  [6356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2048),
  [6358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4534),
  [6360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [6362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1349),
  [6364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4124),
  [6366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1355),
  [6368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [6370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4396),
  [6372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4560),
  [6374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2946),
  [6376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4329),
  [6378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3387),
  [6380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2076),
  [6382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4132),
  [6384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2545),
  [6386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1995),
  [6388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [6390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2026),
  [6392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5956),
  [6394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1145),
  [6396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4137),
  [6398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1148),
  [6400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [6402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__subscripted_interpolations, 1, 0, 5),
  [6404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__subscripted_interpolations, 1, 0, 5),
  [6406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4586),
  [6408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5766),
  [6410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4492),
  [6412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4528),
  [6414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [6416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [6418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4981),
  [6420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [6422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4888),
  [6424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [6426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [6428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4989),
  [6430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [6432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [6434] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__subscripted_interpolations, 1, 0, 6),
  [6436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__subscripted_interpolations, 1, 0, 6),
  [6438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [6440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [6442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(216),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4505),
  [6447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hash_element_interpolation, 5, 0, 106),
  [6449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hash_element_interpolation, 5, 0, 106),
  [6451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__array_element_interpolation, 5, 0, 108),
  [6453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_element_interpolation, 5, 0, 108),
  [6455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [6457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1816),
  [6459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [6461] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__array_element_interpolation, 4, 0, 84),
  [6463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_element_interpolation, 4, 0, 84),
  [6465] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hash_element_interpolation, 4, 0, 85),
  [6467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hash_element_interpolation, 4, 0, 85),
  [6469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3479),
  [6471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2925),
  [6473] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__array_element_interpolation, 4, 0, 74),
  [6475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_element_interpolation, 4, 0, 74),
  [6477] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hash_element_interpolation, 4, 0, 72),
  [6479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hash_element_interpolation, 4, 0, 72),
  [6481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [6483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3349),
  [6485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2294),
  [6487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [6489] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4981),
  [6492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4988),
  [6495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4888),
  [6498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0),
  [6500] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0), SHIFT_REPEAT(616),
  [6503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4989),
  [6506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3366),
  [6508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [6510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2325),
  [6512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [6514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3354),
  [6516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2886),
  [6518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [6520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [6522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [6524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [6526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4677),
  [6528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [6530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4649),
  [6532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5511),
  [6534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4030),
  [6536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5344),
  [6538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5440),
  [6540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5448),
  [6542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [6544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [6546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [6548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [6550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5501),
  [6552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5514),
  [6554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5518),
  [6556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5544),
  [6558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5473),
  [6560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5493),
  [6562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5452),
  [6564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5576),
  [6566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5356),
  [6568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5358),
  [6570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5397),
  [6572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5405),
  [6574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5377),
  [6576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5406),
  [6578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5505),
  [6580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5507),
  [6582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5552),
  [6584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5553),
  [6586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5359),
  [6588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5368),
  [6590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5381),
  [6592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5384),
  [6594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5332),
  [6596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5453),
  [6598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [6600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5340),
  [6602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5421),
  [6604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5423),
  [6606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5429),
  [6608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5436),
  [6610] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(4499),
  [6613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(5729),
  [6616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0),
  [6618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(4517),
  [6621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5458),
  [6623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5466),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5411),
  [6627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5414),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5434),
  [6631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5438),
  [6633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5461),
  [6635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5474),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5478),
  [6639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5481),
  [6641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4591),
  [6643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5526),
  [6645] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(4486),
  [6648] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(5766),
  [6651] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 2, 0, 0), SHIFT_REPEAT(4492),
  [6654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5557),
  [6656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5353),
  [6658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5502),
  [6660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5504),
  [6662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5513),
  [6664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5517),
  [6666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [6668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5390),
  [6670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5479),
  [6672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4542),
  [6674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mandatory_parameter, 1, 0, 30),
  [6676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [6678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5508),
  [6680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5510),
  [6682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolations, 1, 0, 8),
  [6684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolation_fallbacks, 4, 0, 50),
  [6686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5299),
  [6688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4373),
  [6690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5445),
  [6692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4381),
  [6694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5465),
  [6696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolations, 1, 0, 7),
  [6698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__slice_expression_interpolation, 4, 0, 86),
  [6700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4356),
  [6702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4357),
  [6704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4303),
  [6706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3305),
  [6708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__nonvar_interpolation_fallbacks, 1, 0, 4),
  [6710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__slice_expression_interpolation, 4, 0, 87),
  [6712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolation_fallbacks, 1, 0, 0),
  [6714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [6716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [6718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quotelike_begin, 1, 0, 0),
  [6720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__braced_scalar, 3, 0, 0),
  [6722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__braced_array, 3, 0, 0),
  [6724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolation_fallbacks, 3, 0, 50),
  [6726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__scalar_deref_interpolation, 3, 0, 0),
  [6728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [6730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [6732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__slice_expression_interpolation, 6, 0, 133),
  [6734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__slice_expression_interpolation, 6, 0, 51),
  [6736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_deref_interpolation, 3, 0, 51),
  [6738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4354),
  [6740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4355),
  [6742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4937),
  [6744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_regexp_content, 1, 0, 0),
  [6746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolation_fallbacks, 2, 0, 0),
  [6748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4315),
  [6750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4318),
  [6752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__search_slash, 1, 0, 0),
  [6754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [6756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4387),
  [6758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5184),
  [6760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postfix_loop_expression, 3, 0, 42),
  [6762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [6764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [6766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__interp_arrow, 1, 0, 0),
  [6768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4079),
  [6770] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__interp_arrow, 1, 0, 0), SHIFT(6005),
  [6773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interp_arrow, 1, 0, 0),
  [6775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4414),
  [6777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4414),
  [6779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4121),
  [6781] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__interp_arrow, 1, 0, 0), SHIFT(6006),
  [6784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4506),
  [6786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4506),
  [6788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [6790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5156),
  [6792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4221),
  [6794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5073),
  [6796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_transliteration_content, 1, 0, 0),
  [6798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(4677),
  [6801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5184),
  [6804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4973),
  [6806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5053),
  [6808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4205),
  [6810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5054),
  [6812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_content, 1, 0, 0),
  [6814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5007),
  [6816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5008),
  [6818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4299),
  [6820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5009),
  [6822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [6825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postfix_conditional_expression, 3, 0, 42),
  [6827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5196),
  [6829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5019),
  [6831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5020),
  [6833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4290),
  [6835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5022),
  [6837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5130),
  [6839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5131),
  [6841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4232),
  [6843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5132),
  [6845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5023),
  [6847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5024),
  [6849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4234),
  [6851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5025),
  [6853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4125),
  [6855] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__interp_arrow, 1, 0, 0), SHIFT(5952),
  [6858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4436),
  [6860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4436),
  [6862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4126),
  [6864] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__interp_arrow, 1, 0, 0), SHIFT(5731),
  [6867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4574),
  [6869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4574),
  [6871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quotation_mark, 1, 0, 0),
  [6873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quotelike_middle, 2, 0, 0),
  [6875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_parameter, 2, 0, 0),
  [6877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [6879] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(4649),
  [6882] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5196),
  [6885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [6887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_varname, 1, 0, 53),
  [6889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_scalar, 2, 0, 0),
  [6891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postfix_for_expression, 3, 0, 41),
  [6893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5082),
  [6895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5083),
  [6897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4195),
  [6899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5084),
  [6901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__backtick, 1, 0, 0),
  [6903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5311),
  [6905] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [6908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(244),
  [6911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5874),
  [6913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5789),
  [6915] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(4937),
  [6918] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attrlist_repeat1, 2, 0, 0), SHIFT_REPEAT(5311),
  [6921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slurpy_parameter, 1, 0, 31),
  [6923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slurpy_parameter, 1, 0, 32),
  [6925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(260),
  [6928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [6930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5904),
  [6932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5996),
  [6934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [6936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_content_repeat1, 1, 0, 0),
  [6938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [6940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [6942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [6944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4341),
  [6946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5556),
  [6948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4306),
  [6950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5572),
  [6952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4388),
  [6954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5519),
  [6956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4363),
  [6958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5412),
  [6960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4364),
  [6962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5416),
  [6964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [6966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5364),
  [6968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [6970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5596),
  [6972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4313),
  [6974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4323),
  [6976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5335),
  [6978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4352),
  [6980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4307),
  [6982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5490),
  [6984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4334),
  [6986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5483),
  [6988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4335),
  [6990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5485),
  [6992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4145),
  [6994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3043),
  [6996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4319),
  [6998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4365),
  [7000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4366),
  [7002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4368),
  [7004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5431),
  [7006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4369),
  [7008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5432),
  [7010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4320),
  [7012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4377),
  [7014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5469),
  [7016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4378),
  [7018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5476),
  [7020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4373),
  [7022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4379),
  [7024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5482),
  [7026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4380),
  [7028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5484),
  [7030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4304),
  [7032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5345),
  [7034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4336),
  [7036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4337),
  [7038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4289),
  [7040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2771),
  [7042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4381),
  [7044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4394),
  [7046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5443),
  [7048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4321),
  [7050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5537),
  [7052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4392),
  [7054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4322),
  [7056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5543),
  [7058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4353),
  [7060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4324),
  [7062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5336),
  [7064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4146),
  [7066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1941),
  [7068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4346),
  [7070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5362),
  [7072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4305),
  [7074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5374),
  [7076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4375),
  [7078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5463),
  [7080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4376),
  [7082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5468),
  [7084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4347),
  [7086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5372),
  [7088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4342),
  [7090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5333),
  [7092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4356),
  [7094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4357),
  [7096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4362),
  [7098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4367),
  [7100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4206),
  [7102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [7104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4343),
  [7106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5343),
  [7108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [7110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3630),
  [7112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4312),
  [7114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4259),
  [7116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [7118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4348),
  [7120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5389),
  [7122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4349),
  [7124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5392),
  [7126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5984),
  [7128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4228),
  [7130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [7132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4344),
  [7134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4345),
  [7136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4327),
  [7138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4386),
  [7140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5515),
  [7142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4338),
  [7144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5955),
  [7146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4339),
  [7148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4340),
  [7150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5554),
  [7152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4328),
  [7154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4316),
  [7156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5454),
  [7158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4317),
  [7160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5462),
  [7162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4359),
  [7164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5385),
  [7166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4332),
  [7168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5459),
  [7170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4176),
  [7172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2951),
  [7174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4360),
  [7176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5386),
  [7178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4393),
  [7180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5568),
  [7182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4333),
  [7184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5472),
  [7186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [7188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [7190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4351),
  [7192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5297),
  [7194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [7196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5470),
  [7198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5475),
  [7200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6009),
  [7202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3540),
  [7204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5376),
  [7206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5545),
  [7208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5546),
  [7210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5361),
  [7212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5533),
  [7214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5354),
  [7216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [7218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4423),
  [7220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2276),
  [7222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4468),
  [7224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5570),
  [7226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5571),
  [7228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1632),
  [7230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2843),
  [7232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5413),
  [7234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5415),
  [7236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5548),
  [7238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yadayada, 1, 0, 0),
  [7240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5398),
  [7242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5433),
  [7244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5435),
  [7246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5355),
  [7248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5363),
  [7250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5399),
  [7252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5563),
  [7254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5569),
  [7256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5486),
  [7258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5487),
  [7260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3434),
  [7262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5351),
  [7264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5352),
  [7266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [7268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5441),
  [7270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5500),
  [7272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5407),
  [7274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5489),
  [7276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5503),
  [7278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5422),
  [7280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5506),
  [7282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5509),
  [7284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5427),
  [7286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [7288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4443),
  [7290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__postfix_expressions, 1, 0, 0),
  [7292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5447),
  [7294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5451),
  [7296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5455),
  [7298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5457),
  [7300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5496),
  [7302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__decl_variable_list_repeat1, 1, 0, 0),
  [7304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5525),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5527),
  [7308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5573),
  [7310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5534),
  [7312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5528),
  [7314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5529),
  [7316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1894),
  [7318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5338),
  [7320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5339),
  [7322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_vars, 1, 0, 0),
  [7324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3153),
  [7326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 1, 0, 0),
  [7328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4098),
  [7330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2658),
  [7332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_array, 2, 0, 0),
  [7334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_hash, 2, 0, 0),
  [7336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5357),
  [7338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5530),
  [7340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5967),
  [7342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5387),
  [7344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5400),
  [7346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5388),
  [7348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5549),
  [7350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5419),
  [7352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3375),
  [7354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5752),
  [7356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [7358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [7360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [7362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5348),
  [7364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5861),
  [7366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3117),
  [7368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [7370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [7372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3118),
  [7374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3120),
  [7376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3620),
  [7378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3622),
  [7380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2245),
  [7382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5777),
  [7384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2891),
  [7386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3642),
  [7388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2535),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3137),
  [7392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3138),
  [7394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2172),
  [7398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2173),
  [7400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2175),
  [7402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [7404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [7406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [7408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [7410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2970),
  [7412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2186),
  [7414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [7416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3655),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2990),
  [7420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2993),
  [7422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2996),
  [7424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2998),
  [7426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3007),
  [7428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2883),
  [7430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5842),
  [7432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3658),
  [7434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3660),
  [7436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3662),
  [7438] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(430),
  [7441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2412),
  [7443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [7445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [7447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2427),
  [7449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3670),
  [7451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3476),
  [7453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5992),
  [7455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2466),
  [7457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5979),
  [7459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3507),
  [7461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [7463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2512),
  [7465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2513),
  [7467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__noninterpolated_string_content, 2, 0, 0),
  [7469] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__noninterpolated_string_content, 2, 0, 0), SHIFT_REPEAT(5318),
  [7472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1813),
  [7474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6045),
  [7476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1875),
  [7478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3485),
  [7480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4427),
  [7482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4434),
  [7484] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__noninterpolated_string_content, 2, 0, 0), SHIFT_REPEAT(5297),
  [7487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [7489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3320),
  [7491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [7493] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__SUB_AMPER, 1, 0, 0),
  [7495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3323),
  [7497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3325),
  [7499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4689),
  [7501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4691),
  [7503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3327),
  [7505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3192),
  [7507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [7509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [7511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__GLOB_STAR, 1, 0, 0),
  [7513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3335),
  [7515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hash_key, 1, 0, 0),
  [7517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [7519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5895),
  [7521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [7523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [7525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1932),
  [7527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2140),
  [7529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3219),
  [7531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3221),
  [7533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2918),
  [7535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3237),
  [7537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [7539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2178),
  [7541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2179),
  [7543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4617),
  [7545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4618),
  [7547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2457),
  [7549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5740),
  [7551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [7553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [7555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2635),
  [7557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [7559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5701),
  [7561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [7563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4441),
  [7565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4442),
  [7567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2637),
  [7569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [7571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [7573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2448),
  [7575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5802),
  [7577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [7579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5707),
  [7581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [7583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4415),
  [7585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4419),
  [7587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [7589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [7591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2101),
  [7593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [7595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3036),
  [7597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3038),
  [7599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [7601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5687),
  [7603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [7605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4507),
  [7607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4508),
  [7609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [7611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4604),
  [7613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5499),
  [7615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5365),
  [7617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [7619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5606),
  [7621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [7623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4557),
  [7625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4564),
  [7627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3054),
  [7629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2116),
  [7631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2117),
  [7633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__term_rightward_repeat1, 2, 0, 0), SHIFT_REPEAT(561),
  [7636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4616),
  [7638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4622),
  [7640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3579),
  [7642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4687),
  [7644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4693),
  [7646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2711),
  [7648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3073),
  [7650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3076),
  [7652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3078),
  [7654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2497),
  [7656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [7658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2500),
  [7660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3080),
  [7662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4634),
  [7664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4635),
  [7666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 3, 0, 0),
  [7668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3383),
  [7670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3385),
  [7672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4684),
  [7674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4685),
  [7676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [7678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [7680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3090),
  [7682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [7684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [7686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [7688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2763),
  [7690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2765),
  [7692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__for_initializer_repeat1, 2, 0, 0), SHIFT_REPEAT(4197),
  [7695] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__for_initializer_repeat1, 2, 0, 0), SHIFT_REPEAT(5348),
  [7698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__for_initializer_repeat1, 2, 0, 0),
  [7700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2291),
  [7702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5886),
  [7704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2058),
  [7706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2795),
  [7708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [7710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2118),
  [7712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4403),
  [7714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4404),
  [7716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5341),
  [7718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5342),
  [7720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1861),
  [7722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2533),
  [7724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2810),
  [7726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4666),
  [7728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4667),
  [7730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1926),
  [7732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [7734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2813),
  [7736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2816),
  [7738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4437),
  [7740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4438),
  [7742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5409),
  [7744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5410),
  [7746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2818),
  [7748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1944),
  [7750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2827),
  [7752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1897),
  [7754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subroutine_declaration_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(5285),
  [7757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_subroutine_declaration_statement_repeat1, 2, 0, 0),
  [7759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5539),
  [7761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [7763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2190),
  [7765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4553),
  [7767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [7769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5334),
  [7771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2315),
  [7773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2239),
  [7775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [7777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4241),
  [7779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5460),
  [7781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5214),
  [7783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [7785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5230),
  [7787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3139),
  [7789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2184),
  [7791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2507),
  [7793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5520),
  [7795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4410),
  [7797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5273),
  [7799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5250),
  [7801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4501),
  [7803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4384),
  [7805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5456),
  [7807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2108),
  [7809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3242),
  [7811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4202),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4402),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4209),
  [7817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2238),
  [7819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5213),
  [7821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2113),
  [7823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4527),
  [7825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [7827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [7829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [7831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4481),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4445),
  [7835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3134),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [7839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5292),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4172),
  [7843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [7845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [7847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2124),
  [7849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4453),
  [7851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5346),
  [7853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [7855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5239),
  [7857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4382),
  [7859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [7861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5165),
  [7863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4458),
  [7865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [7867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [7869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__apostrophe, 1, 0, 0),
  [7871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2426),
  [7873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5437),
  [7875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [7877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3289),
  [7879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5212),
  [7881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4418),
  [7883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3365),
  [7885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [7887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5471),
  [7889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4447),
  [7891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4258),
  [7893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [7895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4561),
  [7897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [7899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5562),
  [7901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [7903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_subroutine_declaration_statement_repeat1, 1, 0, 0),
  [7905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4463),
  [7907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5160),
  [7909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2097),
  [7911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2109),
  [7913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [7915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2539),
  [7917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__noninterpolated_string_content, 1, 0, 0),
  [7919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2120),
  [7921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4535),
  [7923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4391),
  [7925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2146),
  [7927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4565),
  [7929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2300),
  [7931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2181),
  [7933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [7935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2055),
  [7937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4257),
  [7939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__for_initializer_repeat1, 1, 0, 0),
  [7941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [7943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [7945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [7947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5059),
  [7949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4954),
  [7951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [7953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2585),
  [7955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [7957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [7959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3223),
  [7961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [7963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [7965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3225),
  [7967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [7969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4530),
  [7971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [7973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [7975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [7977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [7979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3627),
  [7981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [7983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [7985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3040),
  [7987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [7989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [7991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3042),
  [7993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [7995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5277),
  [7997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [7999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [8001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3389),
  [8003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [8005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [8007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2947),
  [8009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [8011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [8013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3629),
  [8015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [8017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [8019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2950),
  [8021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [8023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [8025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2768),
  [8027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [8029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [8031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [8033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [8035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [8037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3391),
  [8039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [8041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [8043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2575),
  [8045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [8047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [8049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2717),
  [8051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [8053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [8055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2844),
  [8057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2845),
  [8059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4793),
  [8061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [8063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2316),
  [8065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2361),
  [8067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2185),
  [8069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3136),
  [8071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [8073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2136),
  [8075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3659),
  [8077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2139),
  [8079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2659),
  [8081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2660),
  [8083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3218),
  [8085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3220),
  [8087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [8089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3224),
  [8091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2115),
  [8093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2995),
  [8095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [8097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [8099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3619),
  [8101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [8103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [8105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [8107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__brace_autoquoted, 1, 0, 43),
  [8109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [8111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5538),
  [8113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3541),
  [8115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3542),
  [8117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [8119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2177),
  [8121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [8123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [8125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [8127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1876),
  [8129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3075),
  [8131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3119),
  [8133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2511),
  [8135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3077),
  [8137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2024),
  [8139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [8141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2171),
  [8143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [8145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [8147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2174),
  [8149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3435),
  [8151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3436),
  [8153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [8155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [8157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2942),
  [8159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [8161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3154),
  [8163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3155),
  [8165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2944),
  [8167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [8169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3628),
  [8171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [8173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [8175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3621),
  [8177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2088),
  [8179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [8181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [8183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [8185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5599),
  [8187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [8189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [8191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2762),
  [8193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quotelike_middle_close, 1, 0, 0),
  [8195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2764),
  [8197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5923),
  [8199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [8201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2769),
  [8203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [8205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [8207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [8209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2264),
  [8211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3657),
  [8213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [8215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1965),
  [8217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1931),
  [8219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [8221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [8223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [8225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2112),
  [8227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [8229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [8231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [8233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [8235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [8237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [8239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [8241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [8243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [8245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [8247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [8249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [8251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [8253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [8255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [8257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2959),
  [8259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2815),
  [8261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2534),
  [8263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2099),
  [8265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [8267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [8269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [8271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [8273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [8275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [8277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2102),
  [8279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3322),
  [8281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1936),
  [8283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3116),
  [8285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [8287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [8289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [8291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [8293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [8295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [8297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [8299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [8301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [8303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3324),
  [8305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [8307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [8309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [8311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [8313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [8315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [8317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [8319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [8321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [8323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(536),
  [8325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [8327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [8329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [8331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [8333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [8335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [8337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [8339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [8341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [8343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [8345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3037),
  [8347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [8349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [8351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [8353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [8355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [8357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [8359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [8361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3041),
  [8363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2079),
  [8365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3004),
  [8367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [8369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4502),
  [8371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3074),
  [8373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2890),
  [8375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3044),
  [8377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [8379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1857),
  [8381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4212),
  [8383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4390),
  [8385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2122),
  [8387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [8389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1907),
  [8391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3198),
  [8393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4385),
  [8395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_initializer, 1, 0, 23),
  [8397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4100),
  [8399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5965),
  [8401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5726),
  [8403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [8405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3081),
  [8407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5251),
  [8409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2991),
  [8411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4593),
  [8413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [8415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2880),
  [8417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3509),
  [8419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4531),
  [8421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3082),
  [8423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [8425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4552),
  [8427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_initializer, 2, 0, 11),
  [8429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2871),
  [8431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3049),
  [8433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tricky_list, 3, 0, 0),
  [8435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1977),
  [8437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [8439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4459),
  [8441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5786),
  [8443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2107),
  [8445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [8447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5976),
  [8449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4417),
  [8451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [8453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4770),
  [8455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [8457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [8459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2976),
  [8461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [8463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [8465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [8467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4214),
  [8469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6002),
  [8471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4326),
  [8473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [8475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [8477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [8479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [8481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3227),
  [8483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [8485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3472),
  [8487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [8489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [8491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [8493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2165),
  [8495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [8497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [8499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3086),
  [8501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2633),
  [8503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4256),
  [8505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [8507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6007),
  [8509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3133),
  [8511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [8513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3058),
  [8515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2339),
  [8517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2191),
  [8519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [8521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5815),
  [8523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [8525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [8527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [8529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2413),
  [8531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4383),
  [8533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6018),
  [8535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 2, 0, 0),
  [8537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [8539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5775),
  [8541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [8543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [8545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5987),
  [8547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 2, 0, 0),
  [8549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2152),
  [8551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [8553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5813),
  [8555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [8557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2640),
  [8559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_initializer, 4, 0, 97),
  [8561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5878),
  [8563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [8565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3328),
  [8567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [8569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2641),
  [8571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3313),
  [8573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4050),
  [8575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5880),
  [8577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4288),
  [8579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [8581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [8583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [8585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2773),
  [8587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2705),
  [8589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2096),
  [8591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [8593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3656),
  [8595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4227),
  [8597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2335),
  [8599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4913),
  [8601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_initializer, 3, 0, 65),
  [8603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3191),
  [8605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5893),
  [8607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [8609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2787),
  [8611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [8613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [8615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2119),
  [8617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3637),
  [8619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [8621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [8623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [8625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2154),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3645),
  [8629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3332),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2180),
  [8633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2798),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [8639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_initializer, 5, 0, 123),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3415),
  [8643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2917),
  [8645] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 3, 0, 0),
  [8647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [8649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [8651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4446),
  [8653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [8655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2804),
  [8657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2805),
  [8659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2983),
  [8661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2999),
  [8663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [8665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2486),
  [8667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3631),
  [8669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4559),
  [8671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [8673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3000),
  [8675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3578),
  [8677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3186),
  [8679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [8681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [8683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4475),
  [8685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2811),
  [8687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5898),
  [8689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5787),
  [8691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [8693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3321),
  [8695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [8697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [8699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3329),
  [8701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [8703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4169),
  [8705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [8707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [8709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5783),
  [8711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [8713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3085),
  [8715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3232),
  [8717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4464),
  [8719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [8721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3663),
  [8723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [8725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5058),
  [8727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [8729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4254),
  [8731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4034),
  [8733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5957),
  [8735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2819),
  [8737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2820),
  [8739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [8741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2517),
  [8743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4562),
  [8745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3664),
  [8747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4411),
  [8749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3573),
  [8751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5831),
  [8753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3667),
  [8755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5835),
  [8757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3064),
  [8759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3065),
  [8761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4203),
  [8763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [8765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6003),
  [8767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [8769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5975),
  [8771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3433),
  [8773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5838),
  [8775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5755),
  [8777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [8779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3141),
  [8781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4407),
  [8783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4033),
  [8785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2824),
  [8787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5798),
  [8789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [8791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2962),
  [8793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [8795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3484),
  [8797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4536),
  [8799] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [8801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5841),
  [8803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3515),
  [8805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5847),
  [8807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [8809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5857),
  [8811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5862),
  [8813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [8815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [8817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [8819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4525),
  [8821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [8823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [8825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [8827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2366),
  [8829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [8831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [8833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [8835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [8837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [8839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [8841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [8843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [8845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [8847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4432),
  [8849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_content, 3, 0, 0),
  [8851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_content, 2, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__single_quote = 0,
  ts_external_token__double_quote = 1,
  ts_external_token__backtick_quote = 2,
  ts_external_token__search_slash_quote = 3,
  ts_external_token__no_search_slash_plz = 4,
  ts_external_token__open_readline_bracket = 5,
  ts_external_token__open_fileglob_bracket = 6,
  ts_external_token__PERLY_SEMICOLON = 7,
  ts_external_token__PERLY_HEREDOC = 8,
  ts_external_token__ctrl_z_hack = 9,
  ts_external_token__quotelike_begin_quote = 10,
  ts_external_token__quotelike_middle_close_quote = 11,
  ts_external_token__quotelike_middle_skip = 12,
  ts_external_token__quotelike_end_zw = 13,
  ts_external_token__quotelike_end_quote = 14,
  ts_external_token__q_string_content = 15,
  ts_external_token__qq_string_content = 16,
  ts_external_token_escape_sequence = 17,
  ts_external_token_escaped_delimiter = 18,
  ts_external_token__dollar_in_regexp = 19,
  ts_external_token_pod = 20,
  ts_external_token__gobbled_content = 21,
  ts_external_token__attribute_value_begin = 22,
  ts_external_token_attribute_value = 23,
  ts_external_token_prototype = 24,
  ts_external_token__signature_start = 25,
  ts_external_token__heredoc_delimiter = 26,
  ts_external_token__command_heredoc_delimiter = 27,
  ts_external_token__heredoc_start = 28,
  ts_external_token__heredoc_middle = 29,
  ts_external_token_heredoc_end = 30,
  ts_external_token__fat_comma_autoquoted = 31,
  ts_external_token__filetest = 32,
  ts_external_token__brace_autoquoted_token = 33,
  ts_external_token__brace_end_zw = 34,
  ts_external_token__dollar_ident_zw = 35,
  ts_external_token__no_interp_whitespace_zw = 36,
  ts_external_token__NONASSOC = 37,
  ts_external_token__RECOVER_PAREN_CLOSE = 38,
  ts_external_token__RECOVER_BRACKET_CLOSE = 39,
  ts_external_token__RECOVER_BRACE_CLOSE = 40,
  ts_external_token__RECOVER_ARROW = 41,
  ts_external_token__ERROR = 42,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__single_quote] = sym__single_quote,
  [ts_external_token__double_quote] = sym__double_quote,
  [ts_external_token__backtick_quote] = sym__backtick_quote,
  [ts_external_token__search_slash_quote] = sym__search_slash_quote,
  [ts_external_token__no_search_slash_plz] = sym__no_search_slash_plz,
  [ts_external_token__open_readline_bracket] = sym__open_readline_bracket,
  [ts_external_token__open_fileglob_bracket] = sym__open_fileglob_bracket,
  [ts_external_token__PERLY_SEMICOLON] = sym__PERLY_SEMICOLON,
  [ts_external_token__PERLY_HEREDOC] = sym__PERLY_HEREDOC,
  [ts_external_token__ctrl_z_hack] = sym__ctrl_z_hack,
  [ts_external_token__quotelike_begin_quote] = sym__quotelike_begin_quote,
  [ts_external_token__quotelike_middle_close_quote] = sym__quotelike_middle_close_quote,
  [ts_external_token__quotelike_middle_skip] = sym__quotelike_middle_skip,
  [ts_external_token__quotelike_end_zw] = sym__quotelike_end_zw,
  [ts_external_token__quotelike_end_quote] = sym__quotelike_end_quote,
  [ts_external_token__q_string_content] = sym__q_string_content,
  [ts_external_token__qq_string_content] = sym__qq_string_content,
  [ts_external_token_escape_sequence] = sym_escape_sequence,
  [ts_external_token_escaped_delimiter] = sym_escaped_delimiter,
  [ts_external_token__dollar_in_regexp] = sym__dollar_in_regexp,
  [ts_external_token_pod] = sym_pod,
  [ts_external_token__gobbled_content] = sym__gobbled_content,
  [ts_external_token__attribute_value_begin] = sym__attribute_value_begin,
  [ts_external_token_attribute_value] = sym_attribute_value,
  [ts_external_token_prototype] = sym_prototype,
  [ts_external_token__signature_start] = sym__signature_start,
  [ts_external_token__heredoc_delimiter] = sym__heredoc_delimiter,
  [ts_external_token__command_heredoc_delimiter] = sym__command_heredoc_delimiter,
  [ts_external_token__heredoc_start] = sym__heredoc_start,
  [ts_external_token__heredoc_middle] = sym__heredoc_middle,
  [ts_external_token_heredoc_end] = sym_heredoc_end,
  [ts_external_token__fat_comma_autoquoted] = sym__fat_comma_autoquoted,
  [ts_external_token__filetest] = sym__filetest,
  [ts_external_token__brace_autoquoted_token] = sym__brace_autoquoted_token,
  [ts_external_token__brace_end_zw] = sym__brace_end_zw,
  [ts_external_token__dollar_ident_zw] = sym__dollar_ident_zw,
  [ts_external_token__no_interp_whitespace_zw] = sym__no_interp_whitespace_zw,
  [ts_external_token__NONASSOC] = sym__NONASSOC,
  [ts_external_token__RECOVER_PAREN_CLOSE] = sym__RECOVER_PAREN_CLOSE,
  [ts_external_token__RECOVER_BRACKET_CLOSE] = sym__RECOVER_BRACKET_CLOSE,
  [ts_external_token__RECOVER_BRACE_CLOSE] = sym__RECOVER_BRACE_CLOSE,
  [ts_external_token__RECOVER_ARROW] = sym__RECOVER_ARROW,
  [ts_external_token__ERROR] = sym__ERROR,
};

static const bool ts_external_scanner_states[59][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token__quotelike_begin_quote] = true,
    [ts_external_token__quotelike_middle_close_quote] = true,
    [ts_external_token__quotelike_middle_skip] = true,
    [ts_external_token__quotelike_end_zw] = true,
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__q_string_content] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token__dollar_in_regexp] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__gobbled_content] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token_attribute_value] = true,
    [ts_external_token_prototype] = true,
    [ts_external_token__signature_start] = true,
    [ts_external_token__heredoc_delimiter] = true,
    [ts_external_token__command_heredoc_delimiter] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__heredoc_middle] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__brace_autoquoted_token] = true,
    [ts_external_token__brace_end_zw] = true,
    [ts_external_token__dollar_ident_zw] = true,
    [ts_external_token__no_interp_whitespace_zw] = true,
    [ts_external_token__NONASSOC] = true,
    [ts_external_token__RECOVER_PAREN_CLOSE] = true,
    [ts_external_token__RECOVER_BRACKET_CLOSE] = true,
    [ts_external_token__RECOVER_BRACE_CLOSE] = true,
    [ts_external_token__RECOVER_ARROW] = true,
    [ts_external_token__ERROR] = true,
  },
  [2] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [3] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [4] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [5] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_PAREN_CLOSE] = true,
  },
  [6] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_BRACE_CLOSE] = true,
  },
  [7] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_BRACKET_CLOSE] = true,
  },
  [8] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [9] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__brace_autoquoted_token] = true,
  },
  [10] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [11] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [12] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__brace_end_zw] = true,
  },
  [13] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [14] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_PAREN_CLOSE] = true,
  },
  [15] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_BRACKET_CLOSE] = true,
  },
  [16] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
    [ts_external_token__RECOVER_BRACE_CLOSE] = true,
  },
  [17] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__double_quote] = true,
    [ts_external_token__backtick_quote] = true,
    [ts_external_token__search_slash_quote] = true,
    [ts_external_token__no_search_slash_plz] = true,
    [ts_external_token__open_readline_bracket] = true,
    [ts_external_token__open_fileglob_bracket] = true,
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__PERLY_HEREDOC] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__fat_comma_autoquoted] = true,
    [ts_external_token__filetest] = true,
  },
  [18] = {
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [19] = {
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token__ctrl_z_hack] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [20] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [21] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_BRACKET_CLOSE] = true,
  },
  [22] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_BRACE_CLOSE] = true,
  },
  [23] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_PAREN_CLOSE] = true,
  },
  [24] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [25] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_PAREN_CLOSE] = true,
  },
  [26] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_BRACKET_CLOSE] = true,
  },
  [27] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_BRACE_CLOSE] = true,
  },
  [28] = {
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [29] = {
    [ts_external_token__PERLY_SEMICOLON] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [30] = {
    [ts_external_token__quotelike_middle_close_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token__dollar_in_regexp] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [31] = {
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token__dollar_in_regexp] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [32] = {
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [33] = {
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__heredoc_middle] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [34] = {
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [35] = {
    [ts_external_token__quotelike_end_zw] = true,
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__no_interp_whitespace_zw] = true,
  },
  [36] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__RECOVER_ARROW] = true,
  },
  [37] = {
    [ts_external_token__quotelike_middle_close_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token__dollar_in_regexp] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [38] = {
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token__dollar_in_regexp] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [39] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__dollar_ident_zw] = true,
  },
  [40] = {
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [41] = {
    [ts_external_token__quotelike_middle_close_quote] = true,
    [ts_external_token__qq_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [42] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__no_interp_whitespace_zw] = true,
  },
  [43] = {
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__heredoc_middle] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [44] = {
    [ts_external_token_pod] = true,
    [ts_external_token_prototype] = true,
    [ts_external_token__signature_start] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [45] = {
    [ts_external_token_pod] = true,
    [ts_external_token__signature_start] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [46] = {
    [ts_external_token__quotelike_middle_close_quote] = true,
    [ts_external_token__q_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [47] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token_prototype] = true,
    [ts_external_token__signature_start] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [48] = {
    [ts_external_token__quotelike_end_quote] = true,
    [ts_external_token__q_string_content] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_escaped_delimiter] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [49] = {
    [ts_external_token_pod] = true,
    [ts_external_token__attribute_value_begin] = true,
    [ts_external_token__signature_start] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [50] = {
    [ts_external_token__single_quote] = true,
    [ts_external_token__quotelike_begin_quote] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [51] = {
    [ts_external_token__quotelike_begin_quote] = true,
    [ts_external_token__quotelike_middle_skip] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [52] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_delimiter] = true,
    [ts_external_token__command_heredoc_delimiter] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [53] = {
    [ts_external_token__quotelike_begin_quote] = true,
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [54] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__ERROR] = true,
  },
  [55] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__brace_end_zw] = true,
  },
  [56] = {
    [ts_external_token_pod] = true,
    [ts_external_token_attribute_value] = true,
    [ts_external_token__heredoc_start] = true,
  },
  [57] = {
    [ts_external_token_pod] = true,
    [ts_external_token__heredoc_start] = true,
    [ts_external_token__NONASSOC] = true,
  },
  [58] = {
    [ts_external_token_pod] = true,
    [ts_external_token__gobbled_content] = true,
    [ts_external_token__heredoc_start] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_perl_external_scanner_create(void);
void tree_sitter_perl_external_scanner_destroy(void *);
bool tree_sitter_perl_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_perl_external_scanner_serialize(void *, char *);
void tree_sitter_perl_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_perl(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym__identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_perl_external_scanner_create,
      tree_sitter_perl_external_scanner_destroy,
      tree_sitter_perl_external_scanner_scan,
      tree_sitter_perl_external_scanner_serialize,
      tree_sitter_perl_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "perl",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 1,
      .minor_version = 0,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
